/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 7337
#define LARGE_STATE_COUNT 1406
#define SYMBOL_COUNT 456
#define ALIAS_COUNT 13
#define TOKEN_COUNT 152
#define EXTERNAL_TOKEN_COUNT 7
#define FIELD_COUNT 39
#define MAX_ALIAS_SEQUENCE_LENGTH 11
#define MAX_RESERVED_WORD_SET_SIZE 55
#define PRODUCTION_ID_COUNT 231
#define SUPERTYPE_COUNT 32

enum ts_symbol_identifiers {
  sym__lowercase_identifier = 1,
  anon_sym_SEMI_SEMI = 2,
  anon_sym_let = 3,
  anon_sym_rec = 4,
  anon_sym_and = 5,
  anon_sym_EQ = 6,
  anon_sym_TILDE = 7,
  anon_sym_QMARK = 8,
  anon_sym_COLON = 9,
  anon_sym_LPAREN = 10,
  anon_sym_RPAREN = 11,
  anon_sym_external = 12,
  anon_sym_type = 13,
  anon_sym_nonrec = 14,
  anon_sym_COLON_EQ = 15,
  anon_sym_private = 16,
  anon_sym_DOT_DOT = 17,
  anon_sym_PLUS_EQ = 18,
  anon_sym_COMMA = 19,
  anon_sym_PLUS = 20,
  anon_sym_DASH = 21,
  anon_sym_BANG = 22,
  anon_sym__ = 23,
  anon_sym_PIPE = 24,
  anon_sym_of = 25,
  anon_sym_COLON2 = 26,
  anon_sym_DOT = 27,
  anon_sym_DASH_GT = 28,
  anon_sym_LBRACK = 29,
  anon_sym_RBRACK = 30,
  anon_sym_true = 31,
  anon_sym_false = 32,
  anon_sym_STAR = 33,
  anon_sym_LBRACE = 34,
  anon_sym_SEMI = 35,
  anon_sym_RBRACE = 36,
  anon_sym_mutable = 37,
  anon_sym_constraint = 38,
  anon_sym_exception = 39,
  anon_sym_module = 40,
  anon_sym_open = 41,
  anon_sym_include = 42,
  anon_sym_class = 43,
  anon_sym_virtual = 44,
  anon_sym_val = 45,
  anon_sym_sig = 46,
  anon_sym_end = 47,
  anon_sym_with = 48,
  anon_sym_functor = 49,
  anon_sym_struct = 50,
  anon_sym_COLON_GT = 51,
  anon_sym_object = 52,
  anon_sym_inherit = 53,
  anon_sym_method = 54,
  anon_sym_in = 55,
  anon_sym_fun = 56,
  anon_sym_as = 57,
  anon_sym_initializer = 58,
  anon_sym_LBRACK_GT = 59,
  anon_sym_LBRACK_LT = 60,
  anon_sym_GT = 61,
  anon_sym_AMP = 62,
  anon_sym_LT = 63,
  anon_sym_POUND = 64,
  anon_sym_COLON_COLON = 65,
  anon_sym_LBRACK_PIPE = 66,
  anon_sym_PIPE_RBRACK = 67,
  anon_sym_LT_DASH = 68,
  anon_sym_if = 69,
  anon_sym_then = 70,
  anon_sym_else = 71,
  anon_sym_while = 72,
  anon_sym_do = 73,
  anon_sym_done = 74,
  anon_sym_for = 75,
  anon_sym_to = 76,
  anon_sym_downto = 77,
  anon_sym_match = 78,
  anon_sym_when = 79,
  anon_sym_function = 80,
  anon_sym_try = 81,
  anon_sym_assert = 82,
  anon_sym_lazy = 83,
  anon_sym_new = 84,
  anon_sym_LBRACE_LT = 85,
  anon_sym_GT_RBRACE = 86,
  anon_sym_begin = 87,
  sym_ocamlyacc_value = 88,
  anon_sym_effect = 89,
  aux_sym_attribute_token1 = 90,
  anon_sym_LBRACK_AT_AT = 91,
  anon_sym_LBRACK_AT_AT_AT = 92,
  anon_sym_LBRACK_PERCENT = 93,
  anon_sym_LBRACE_PERCENT = 94,
  aux_sym_quoted_extension_token1 = 95,
  anon_sym_LBRACK_PERCENT_PERCENT = 96,
  anon_sym_LBRACE_PERCENT_PERCENT = 97,
  anon_sym_PERCENT = 98,
  aux_sym_number_token1 = 99,
  aux_sym_signed_number_token1 = 100,
  anon_sym_SQUOTE = 101,
  anon_sym_SQUOTE2 = 102,
  aux_sym_character_content_token1 = 103,
  aux_sym_character_content_token2 = 104,
  anon_sym_DQUOTE = 105,
  aux_sym_string_content_token1 = 106,
  aux_sym_string_content_token2 = 107,
  aux_sym_string_content_token3 = 108,
  aux_sym_string_content_token4 = 109,
  aux_sym_string_content_token5 = 110,
  aux_sym_quoted_string_content_token1 = 111,
  aux_sym_escape_sequence_token1 = 112,
  aux_sym_escape_sequence_token2 = 113,
  aux_sym_escape_sequence_token3 = 114,
  aux_sym_escape_sequence_token4 = 115,
  sym_conversion_specification = 116,
  sym_pretty_printing_indication = 117,
  sym_prefix_operator = 118,
  aux_sym_sign_operator_token1 = 119,
  sym_hash_operator = 120,
  anon_sym_lsl = 121,
  anon_sym_lsr = 122,
  anon_sym_asr = 123,
  aux_sym_pow_operator_token1 = 124,
  anon_sym_mod = 125,
  anon_sym_land = 126,
  anon_sym_lor = 127,
  anon_sym_lxor = 128,
  aux_sym_mult_operator_token1 = 129,
  aux_sym_add_operator_token1 = 130,
  sym_concat_operator = 131,
  sym_rel_operator = 132,
  anon_sym_AMP_AMP = 133,
  anon_sym_or = 134,
  anon_sym_PIPE_PIPE = 135,
  sym_assign_operator = 136,
  sym_indexing_operator = 137,
  sym_let_operator = 138,
  sym_let_and_operator = 139,
  sym_match_operator = 140,
  sym__uppercase_identifier = 141,
  aux_sym_directive_token1 = 142,
  aux_sym_type_variable_token1 = 143,
  aux_sym_tag_token1 = 144,
  aux_sym_attribute_id_token1 = 145,
  sym_comment = 146,
  sym__left_quoted_string_delimiter = 147,
  sym__right_quoted_string_delimiter = 148,
  sym_line_number_directive = 149,
  sym__null = 150,
  sym__error_sentinel = 151,
  sym_compilation_unit = 152,
  sym__structure = 153,
  sym_expression_item = 154,
  sym__signature = 155,
  sym_toplevel_directive = 156,
  sym__local_structure_item = 157,
  sym__structure_item = 158,
  sym_value_definition = 159,
  sym_let_binding = 160,
  sym_parameter = 161,
  sym_external = 162,
  sym_type_definition = 163,
  sym_type_binding = 164,
  sym__type_params = 165,
  sym__type_param = 166,
  sym_variant_declaration = 167,
  sym_constructor_declaration = 168,
  sym__constructor_declaration_name = 169,
  sym__constructor_argument = 170,
  sym_record_declaration = 171,
  sym_field_declaration = 172,
  sym_external_declaration = 173,
  sym_type_constraint = 174,
  sym_exception_definition = 175,
  sym_module_definition = 176,
  sym_module_binding = 177,
  sym_module_parameter = 178,
  sym_module_type_definition = 179,
  sym_open_module = 180,
  sym_include_module = 181,
  sym_class_definition = 182,
  sym_class_binding = 183,
  sym_class_type_definition = 184,
  sym_class_type_binding = 185,
  sym__signature_item = 186,
  sym_value_specification = 187,
  sym_include_module_type = 188,
  sym__module_typed = 189,
  sym__module_type = 190,
  sym_signature = 191,
  sym_module_type_constraint = 192,
  sym_constrain_type = 193,
  sym_constrain_module = 194,
  sym_constrain_module_type = 195,
  sym_module_type_of = 196,
  sym_functor_type = 197,
  sym_parenthesized_module_type = 198,
  sym__simple_module_expression = 199,
  sym__module_expression = 200,
  sym_structure = 201,
  sym_functor = 202,
  sym_module_application = 203,
  sym_typed_module_expression = 204,
  sym_packed_module = 205,
  sym_parenthesized_module_expression = 206,
  sym__class_typed = 207,
  sym__simple_class_type = 208,
  sym__class_type = 209,
  sym_instantiated_class_type = 210,
  sym_class_body_type = 211,
  sym__class_field_specification = 212,
  sym_inheritance_specification = 213,
  sym_instance_variable_specification = 214,
  sym_method_specification = 215,
  sym_type_parameter_constraint = 216,
  sym_let_open_class_type = 217,
  sym_class_function_type = 218,
  sym__simple_class_expression = 219,
  sym__class_expression = 220,
  sym_instantiated_class = 221,
  sym_typed_class_expression = 222,
  sym_class_function = 223,
  sym_class_application = 224,
  sym_let_class_expression = 225,
  sym__class_field = 226,
  sym_inheritance_definition = 227,
  sym_instance_variable_definition = 228,
  sym_method_definition = 229,
  sym_class_initializer = 230,
  sym_let_open_class_expression = 231,
  sym_parenthesized_class_expression = 232,
  sym__typed = 233,
  sym__simple_typed = 234,
  sym__coerced = 235,
  sym__type_constrained = 236,
  sym__maybe_polymorphic_typed = 237,
  sym__polymorphic_typed = 238,
  sym__polymorphic_type = 239,
  sym_polymorphic_type = 240,
  sym__abstract_type = 241,
  sym__parenthesized_abstract_type = 242,
  sym__delimited_type = 243,
  sym__simple_type = 244,
  sym__type = 245,
  sym_function_type = 246,
  sym_labeled_argument_type = 247,
  sym__proper_tuple_type = 248,
  sym__labeled_tuple_type = 249,
  sym_labeled_tuple_element_type = 250,
  sym__tuple_type_rhs = 251,
  sym_constructed_type = 252,
  sym_aliased_type = 253,
  sym_local_open_type = 254,
  sym_polymorphic_variant_type = 255,
  sym__tag_spec = 256,
  sym_tag_specification = 257,
  sym_package_type = 258,
  sym_object_type = 259,
  sym_method_type = 260,
  sym_hash_type = 261,
  sym_parenthesized_type = 262,
  sym__delimited_expression = 263,
  sym__simple_expression = 264,
  sym__expression = 265,
  sym__sequence_expression = 266,
  sym_typed_expression = 267,
  sym_labeled_tuple_element = 268,
  sym__tuple_expression = 269,
  sym_cons_expression = 270,
  sym_list_expression = 271,
  sym_array_expression = 272,
  sym__sequence_expression_content = 273,
  sym_record_expression = 274,
  sym_field_expression = 275,
  sym_application_expression = 276,
  sym_labeled_argument = 277,
  sym_prefix_expression = 278,
  sym_sign_expression = 279,
  sym_hash_expression = 280,
  sym_infix_expression = 281,
  sym_field_get_expression = 282,
  sym_array_get_expression = 283,
  sym_string_get_expression = 284,
  sym_bigarray_get_expression = 285,
  sym_set_expression = 286,
  sym_if_expression = 287,
  sym_then_clause = 288,
  sym_else_clause = 289,
  sym_while_expression = 290,
  sym_do_clause = 291,
  sym_for_expression = 292,
  sym__sequence_expression_anonymous = 293,
  sym_match_expression = 294,
  sym__match_cases = 295,
  sym_match_case = 296,
  sym_guard = 297,
  sym_refutation_case = 298,
  sym_function_expression = 299,
  sym_fun_expression = 300,
  sym_try_expression = 301,
  sym_let_expression = 302,
  sym_assert_expression = 303,
  sym_lazy_expression = 304,
  sym_local_open_expression = 305,
  sym_package_expression = 306,
  sym_new_expression = 307,
  sym_object_copy_expression = 308,
  sym_instance_variable_expression = 309,
  sym_method_invocation = 310,
  sym_object_expression = 311,
  sym_parenthesized_expression = 312,
  sym__delimited_pattern = 313,
  sym__simple_pattern = 314,
  sym__effect_pattern = 315,
  sym__pattern = 316,
  sym__delimited_binding_pattern = 317,
  sym__simple_binding_pattern = 318,
  sym__effect_binding_pattern = 319,
  sym__binding_pattern_no_exn = 320,
  sym__binding_pattern = 321,
  sym_alias_pattern = 322,
  sym_alias_binding_pattern_no_exn = 323,
  sym_alias_binding_pattern = 324,
  sym_typed_pattern = 325,
  sym_typed_binding_pattern = 326,
  sym__or_pattern_anonymous = 327,
  sym__or_binding_pattern_no_exn_anonymous = 328,
  sym__or_binding_pattern_anonymous = 329,
  sym_constructor_pattern = 330,
  sym_constructor_binding_pattern = 331,
  sym_tag_pattern = 332,
  sym_tag_binding_pattern = 333,
  sym_polymorphic_variant_pattern = 334,
  sym_labeled_tuple_element_pattern = 335,
  sym__tuple_pattern = 336,
  sym_labeled_tuple_element_binding_pattern = 337,
  sym__tuple_binding_pattern_no_exn = 338,
  sym__tuple_binding_pattern = 339,
  sym_record_pattern = 340,
  sym_field_pattern = 341,
  sym_record_binding_pattern = 342,
  sym_field_binding_pattern = 343,
  sym_list_pattern = 344,
  sym_list_binding_pattern = 345,
  sym_cons_pattern = 346,
  sym_cons_binding_pattern_no_exn = 347,
  sym_cons_binding_pattern = 348,
  sym_array_pattern = 349,
  sym_array_binding_pattern = 350,
  sym__sequence_pattern_content = 351,
  sym__sequence_binding_pattern_content = 352,
  sym_range_pattern = 353,
  sym_lazy_pattern = 354,
  sym_lazy_binding_pattern = 355,
  sym_local_open_pattern = 356,
  sym_local_open_binding_pattern = 357,
  sym_package_pattern = 358,
  sym_parenthesized_pattern = 359,
  sym_parenthesized_binding_pattern = 360,
  sym_exception_pattern = 361,
  sym_exception_binding_pattern = 362,
  sym_effect_pattern = 363,
  sym_effect_binding_pattern = 364,
  sym_attribute = 365,
  sym_item_attribute = 366,
  sym_floating_attribute = 367,
  sym_attribute_payload = 368,
  sym_extension = 369,
  sym_quoted_extension = 370,
  sym_item_extension = 371,
  sym_quoted_item_extension = 372,
  sym__attribute = 373,
  sym__constant = 374,
  sym__signed_constant = 375,
  sym_number = 376,
  sym_signed_number = 377,
  sym_character = 378,
  sym_character_content = 379,
  sym_string = 380,
  sym_string_content = 381,
  sym_quoted_string = 382,
  sym__quoted_string = 383,
  sym_quoted_string_content = 384,
  sym_escape_sequence = 385,
  sym_boolean = 386,
  sym_unit = 387,
  sym_sign_operator = 388,
  sym__infix_operator = 389,
  sym_pow_operator = 390,
  sym_mult_operator = 391,
  sym_add_operator = 392,
  sym_and_operator = 393,
  sym_or_operator = 394,
  sym_indexing_operator_path = 395,
  sym__value_name = 396,
  sym__simple_value_pattern = 397,
  sym_parenthesized_operator = 398,
  sym_value_path = 399,
  sym_module_path = 400,
  sym_extended_module_path = 401,
  sym_module_type_path = 402,
  sym_field_path = 403,
  sym_constructor_path = 404,
  sym_type_constructor_path = 405,
  sym_class_path = 406,
  sym_class_type_path = 407,
  sym__instance_variable_name = 408,
  sym__constructor_name = 409,
  sym_directive = 410,
  sym_type_variable = 411,
  sym_tag = 412,
  sym_attribute_id = 413,
  aux_sym__structure_repeat1 = 414,
  aux_sym__structure_repeat2 = 415,
  aux_sym_expression_item_repeat1 = 416,
  aux_sym__signature_repeat1 = 417,
  aux_sym_value_definition_repeat1 = 418,
  aux_sym_let_binding_repeat1 = 419,
  aux_sym_external_repeat1 = 420,
  aux_sym_type_definition_repeat1 = 421,
  aux_sym_type_binding_repeat1 = 422,
  aux_sym__type_params_repeat1 = 423,
  aux_sym__type_param_repeat1 = 424,
  aux_sym_variant_declaration_repeat1 = 425,
  aux_sym_constructor_declaration_repeat1 = 426,
  aux_sym__constructor_argument_repeat1 = 427,
  aux_sym_record_declaration_repeat1 = 428,
  aux_sym_module_definition_repeat1 = 429,
  aux_sym_module_binding_repeat1 = 430,
  aux_sym_class_definition_repeat1 = 431,
  aux_sym_class_type_definition_repeat1 = 432,
  aux_sym_module_type_constraint_repeat1 = 433,
  aux_sym_instantiated_class_type_repeat1 = 434,
  aux_sym_class_body_type_repeat1 = 435,
  aux_sym_instance_variable_specification_repeat1 = 436,
  aux_sym_method_specification_repeat1 = 437,
  aux_sym_class_application_repeat1 = 438,
  aux_sym__abstract_type_repeat1 = 439,
  aux_sym_polymorphic_variant_type_repeat1 = 440,
  aux_sym_polymorphic_variant_type_repeat2 = 441,
  aux_sym_tag_specification_repeat1 = 442,
  aux_sym_object_type_repeat1 = 443,
  aux_sym__sequence_expression_content_repeat1 = 444,
  aux_sym_record_expression_repeat1 = 445,
  aux_sym__match_cases_repeat1 = 446,
  aux_sym_object_copy_expression_repeat1 = 447,
  aux_sym_object_expression_repeat1 = 448,
  aux_sym_record_pattern_repeat1 = 449,
  aux_sym_record_binding_pattern_repeat1 = 450,
  aux_sym__sequence_pattern_content_repeat1 = 451,
  aux_sym__sequence_binding_pattern_content_repeat1 = 452,
  aux_sym_string_content_repeat1 = 453,
  aux_sym_quoted_string_content_repeat1 = 454,
  aux_sym_attribute_id_repeat1 = 455,
  alias_sym_class_name = 456,
  alias_sym_class_type_name = 457,
  alias_sym_field_name = 458,
  alias_sym_instance_variable_name = 459,
  alias_sym_label_name = 460,
  alias_sym_method_name = 461,
  alias_sym_module_name = 462,
  alias_sym_module_type_name = 463,
  alias_sym_sequence_expression = 464,
  alias_sym_tuple_expression = 465,
  alias_sym_type_constructor = 466,
  alias_sym_value_name = 467,
  alias_sym_value_pattern = 468,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym__lowercase_identifier] = "_lowercase_identifier",
  [anon_sym_SEMI_SEMI] = ";;",
  [anon_sym_let] = "let",
  [anon_sym_rec] = "rec",
  [anon_sym_and] = "and",
  [anon_sym_EQ] = "=",
  [anon_sym_TILDE] = "~",
  [anon_sym_QMARK] = "\?",
  [anon_sym_COLON] = ":",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_external] = "external",
  [anon_sym_type] = "type",
  [anon_sym_nonrec] = "nonrec",
  [anon_sym_COLON_EQ] = ":=",
  [anon_sym_private] = "private",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_COMMA] = ",",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_BANG] = "!",
  [anon_sym__] = "_",
  [anon_sym_PIPE] = "|",
  [anon_sym_of] = "of",
  [anon_sym_COLON2] = ":",
  [anon_sym_DOT] = ".",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_true] = "true",
  [anon_sym_false] = "false",
  [anon_sym_STAR] = "*",
  [anon_sym_LBRACE] = "{",
  [anon_sym_SEMI] = ";",
  [anon_sym_RBRACE] = "}",
  [anon_sym_mutable] = "mutable",
  [anon_sym_constraint] = "constraint",
  [anon_sym_exception] = "exception",
  [anon_sym_module] = "module",
  [anon_sym_open] = "open",
  [anon_sym_include] = "include",
  [anon_sym_class] = "class",
  [anon_sym_virtual] = "virtual",
  [anon_sym_val] = "val",
  [anon_sym_sig] = "sig",
  [anon_sym_end] = "end",
  [anon_sym_with] = "with",
  [anon_sym_functor] = "functor",
  [anon_sym_struct] = "struct",
  [anon_sym_COLON_GT] = ":>",
  [anon_sym_object] = "object",
  [anon_sym_inherit] = "inherit",
  [anon_sym_method] = "method",
  [anon_sym_in] = "in",
  [anon_sym_fun] = "fun",
  [anon_sym_as] = "as",
  [anon_sym_initializer] = "initializer",
  [anon_sym_LBRACK_GT] = "[>",
  [anon_sym_LBRACK_LT] = "[<",
  [anon_sym_GT] = ">",
  [anon_sym_AMP] = "&",
  [anon_sym_LT] = "<",
  [anon_sym_POUND] = "#",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym_LBRACK_PIPE] = "[|",
  [anon_sym_PIPE_RBRACK] = "|]",
  [anon_sym_LT_DASH] = "<-",
  [anon_sym_if] = "if",
  [anon_sym_then] = "then",
  [anon_sym_else] = "else",
  [anon_sym_while] = "while",
  [anon_sym_do] = "do",
  [anon_sym_done] = "done",
  [anon_sym_for] = "for",
  [anon_sym_to] = "to",
  [anon_sym_downto] = "downto",
  [anon_sym_match] = "match",
  [anon_sym_when] = "when",
  [anon_sym_function] = "function",
  [anon_sym_try] = "try",
  [anon_sym_assert] = "assert",
  [anon_sym_lazy] = "lazy",
  [anon_sym_new] = "new",
  [anon_sym_LBRACE_LT] = "{<",
  [anon_sym_GT_RBRACE] = ">}",
  [anon_sym_begin] = "begin",
  [sym_ocamlyacc_value] = "ocamlyacc_value",
  [anon_sym_effect] = "effect",
  [aux_sym_attribute_token1] = "[@",
  [anon_sym_LBRACK_AT_AT] = "[@@",
  [anon_sym_LBRACK_AT_AT_AT] = "[@@@",
  [anon_sym_LBRACK_PERCENT] = "[%",
  [anon_sym_LBRACE_PERCENT] = "{%",
  [aux_sym_quoted_extension_token1] = "quoted_extension_token1",
  [anon_sym_LBRACK_PERCENT_PERCENT] = "[%%",
  [anon_sym_LBRACE_PERCENT_PERCENT] = "{%%",
  [anon_sym_PERCENT] = "%",
  [aux_sym_number_token1] = "number_token1",
  [aux_sym_signed_number_token1] = "signed_number_token1",
  [anon_sym_SQUOTE] = "'",
  [anon_sym_SQUOTE2] = "'",
  [aux_sym_character_content_token1] = "character_content_token1",
  [aux_sym_character_content_token2] = "character_content_token2",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym_string_content_token1] = "string_content_token1",
  [aux_sym_string_content_token2] = "string_content_token2",
  [aux_sym_string_content_token3] = "string_content_token3",
  [aux_sym_string_content_token4] = "escape_sequence",
  [aux_sym_string_content_token5] = "escape_sequence",
  [aux_sym_quoted_string_content_token1] = "quoted_string_content_token1",
  [aux_sym_escape_sequence_token1] = "escape_sequence_token1",
  [aux_sym_escape_sequence_token2] = "escape_sequence_token2",
  [aux_sym_escape_sequence_token3] = "escape_sequence_token3",
  [aux_sym_escape_sequence_token4] = "escape_sequence_token4",
  [sym_conversion_specification] = "conversion_specification",
  [sym_pretty_printing_indication] = "pretty_printing_indication",
  [sym_prefix_operator] = "prefix_operator",
  [aux_sym_sign_operator_token1] = "sign_operator_token1",
  [sym_hash_operator] = "hash_operator",
  [anon_sym_lsl] = "lsl",
  [anon_sym_lsr] = "lsr",
  [anon_sym_asr] = "asr",
  [aux_sym_pow_operator_token1] = "pow_operator_token1",
  [anon_sym_mod] = "mod",
  [anon_sym_land] = "land",
  [anon_sym_lor] = "lor",
  [anon_sym_lxor] = "lxor",
  [aux_sym_mult_operator_token1] = "mult_operator_token1",
  [aux_sym_add_operator_token1] = "add_operator_token1",
  [sym_concat_operator] = "concat_operator",
  [sym_rel_operator] = "rel_operator",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_or] = "or",
  [anon_sym_PIPE_PIPE] = "||",
  [sym_assign_operator] = "assign_operator",
  [sym_indexing_operator] = "indexing_operator",
  [sym_let_operator] = "let_operator",
  [sym_let_and_operator] = "let_and_operator",
  [sym_match_operator] = "match_operator",
  [sym__uppercase_identifier] = "_uppercase_identifier",
  [aux_sym_directive_token1] = "directive_token1",
  [aux_sym_type_variable_token1] = "type_variable_token1",
  [aux_sym_tag_token1] = "tag_token1",
  [aux_sym_attribute_id_token1] = "attribute_id_token1",
  [sym_comment] = "comment",
  [sym__left_quoted_string_delimiter] = "_left_quoted_string_delimiter",
  [sym__right_quoted_string_delimiter] = "_right_quoted_string_delimiter",
  [sym_line_number_directive] = "line_number_directive",
  [sym__null] = "_null",
  [sym__error_sentinel] = "_error_sentinel",
  [sym_compilation_unit] = "compilation_unit",
  [sym__structure] = "_structure",
  [sym_expression_item] = "expression_item",
  [sym__signature] = "_signature",
  [sym_toplevel_directive] = "toplevel_directive",
  [sym__local_structure_item] = "_local_structure_item",
  [sym__structure_item] = "_structure_item",
  [sym_value_definition] = "value_definition",
  [sym_let_binding] = "let_binding",
  [sym_parameter] = "parameter",
  [sym_external] = "external",
  [sym_type_definition] = "type_definition",
  [sym_type_binding] = "type_binding",
  [sym__type_params] = "_type_params",
  [sym__type_param] = "_type_param",
  [sym_variant_declaration] = "variant_declaration",
  [sym_constructor_declaration] = "constructor_declaration",
  [sym__constructor_declaration_name] = "constructor_name",
  [sym__constructor_argument] = "_constructor_argument",
  [sym_record_declaration] = "record_declaration",
  [sym_field_declaration] = "field_declaration",
  [sym_external_declaration] = "external_declaration",
  [sym_type_constraint] = "type_constraint",
  [sym_exception_definition] = "exception_definition",
  [sym_module_definition] = "module_definition",
  [sym_module_binding] = "module_binding",
  [sym_module_parameter] = "module_parameter",
  [sym_module_type_definition] = "module_type_definition",
  [sym_open_module] = "open_module",
  [sym_include_module] = "include_module",
  [sym_class_definition] = "class_definition",
  [sym_class_binding] = "class_binding",
  [sym_class_type_definition] = "class_type_definition",
  [sym_class_type_binding] = "class_type_binding",
  [sym__signature_item] = "_signature_item",
  [sym_value_specification] = "value_specification",
  [sym_include_module_type] = "include_module_type",
  [sym__module_typed] = "_module_typed",
  [sym__module_type] = "_module_type",
  [sym_signature] = "signature",
  [sym_module_type_constraint] = "module_type_constraint",
  [sym_constrain_type] = "constrain_type",
  [sym_constrain_module] = "constrain_module",
  [sym_constrain_module_type] = "constrain_module_type",
  [sym_module_type_of] = "module_type_of",
  [sym_functor_type] = "functor_type",
  [sym_parenthesized_module_type] = "parenthesized_module_type",
  [sym__simple_module_expression] = "_simple_module_expression",
  [sym__module_expression] = "_module_expression",
  [sym_structure] = "structure",
  [sym_functor] = "functor",
  [sym_module_application] = "module_application",
  [sym_typed_module_expression] = "typed_module_expression",
  [sym_packed_module] = "packed_module",
  [sym_parenthesized_module_expression] = "parenthesized_module_expression",
  [sym__class_typed] = "_class_typed",
  [sym__simple_class_type] = "_simple_class_type",
  [sym__class_type] = "_class_type",
  [sym_instantiated_class_type] = "instantiated_class_type",
  [sym_class_body_type] = "class_body_type",
  [sym__class_field_specification] = "_class_field_specification",
  [sym_inheritance_specification] = "inheritance_specification",
  [sym_instance_variable_specification] = "instance_variable_specification",
  [sym_method_specification] = "method_specification",
  [sym_type_parameter_constraint] = "type_parameter_constraint",
  [sym_let_open_class_type] = "let_open_class_type",
  [sym_class_function_type] = "class_function_type",
  [sym__simple_class_expression] = "_simple_class_expression",
  [sym__class_expression] = "_class_expression",
  [sym_instantiated_class] = "instantiated_class",
  [sym_typed_class_expression] = "typed_class_expression",
  [sym_class_function] = "class_function",
  [sym_class_application] = "class_application",
  [sym_let_class_expression] = "let_class_expression",
  [sym__class_field] = "_class_field",
  [sym_inheritance_definition] = "inheritance_definition",
  [sym_instance_variable_definition] = "instance_variable_definition",
  [sym_method_definition] = "method_definition",
  [sym_class_initializer] = "class_initializer",
  [sym_let_open_class_expression] = "let_open_class_expression",
  [sym_parenthesized_class_expression] = "parenthesized_class_expression",
  [sym__typed] = "_typed",
  [sym__simple_typed] = "_simple_typed",
  [sym__coerced] = "_coerced",
  [sym__type_constrained] = "_type_constrained",
  [sym__maybe_polymorphic_typed] = "_maybe_polymorphic_typed",
  [sym__polymorphic_typed] = "_polymorphic_typed",
  [sym__polymorphic_type] = "_polymorphic_type",
  [sym_polymorphic_type] = "polymorphic_type",
  [sym__abstract_type] = "_abstract_type",
  [sym__parenthesized_abstract_type] = "abstract_type",
  [sym__delimited_type] = "_delimited_type",
  [sym__simple_type] = "_simple_type",
  [sym__type] = "_type",
  [sym_function_type] = "function_type",
  [sym_labeled_argument_type] = "labeled_argument_type",
  [sym__proper_tuple_type] = "tuple_type",
  [sym__labeled_tuple_type] = "tuple_type",
  [sym_labeled_tuple_element_type] = "labeled_tuple_element_type",
  [sym__tuple_type_rhs] = "_tuple_type_rhs",
  [sym_constructed_type] = "constructed_type",
  [sym_aliased_type] = "aliased_type",
  [sym_local_open_type] = "local_open_type",
  [sym_polymorphic_variant_type] = "polymorphic_variant_type",
  [sym__tag_spec] = "_tag_spec",
  [sym_tag_specification] = "tag_specification",
  [sym_package_type] = "package_type",
  [sym_object_type] = "object_type",
  [sym_method_type] = "method_type",
  [sym_hash_type] = "hash_type",
  [sym_parenthesized_type] = "parenthesized_type",
  [sym__delimited_expression] = "_delimited_expression",
  [sym__simple_expression] = "_simple_expression",
  [sym__expression] = "_expression",
  [sym__sequence_expression] = "_sequence_expression",
  [sym_typed_expression] = "typed_expression",
  [sym_labeled_tuple_element] = "labeled_tuple_element",
  [sym__tuple_expression] = "_tuple_expression",
  [sym_cons_expression] = "cons_expression",
  [sym_list_expression] = "list_expression",
  [sym_array_expression] = "array_expression",
  [sym__sequence_expression_content] = "_sequence_expression_content",
  [sym_record_expression] = "record_expression",
  [sym_field_expression] = "field_expression",
  [sym_application_expression] = "application_expression",
  [sym_labeled_argument] = "labeled_argument",
  [sym_prefix_expression] = "prefix_expression",
  [sym_sign_expression] = "sign_expression",
  [sym_hash_expression] = "hash_expression",
  [sym_infix_expression] = "infix_expression",
  [sym_field_get_expression] = "field_get_expression",
  [sym_array_get_expression] = "array_get_expression",
  [sym_string_get_expression] = "string_get_expression",
  [sym_bigarray_get_expression] = "bigarray_get_expression",
  [sym_set_expression] = "set_expression",
  [sym_if_expression] = "if_expression",
  [sym_then_clause] = "then_clause",
  [sym_else_clause] = "else_clause",
  [sym_while_expression] = "while_expression",
  [sym_do_clause] = "do_clause",
  [sym_for_expression] = "for_expression",
  [sym__sequence_expression_anonymous] = "_sequence_expression_anonymous",
  [sym_match_expression] = "match_expression",
  [sym__match_cases] = "_match_cases",
  [sym_match_case] = "match_case",
  [sym_guard] = "guard",
  [sym_refutation_case] = "refutation_case",
  [sym_function_expression] = "function_expression",
  [sym_fun_expression] = "fun_expression",
  [sym_try_expression] = "try_expression",
  [sym_let_expression] = "let_expression",
  [sym_assert_expression] = "assert_expression",
  [sym_lazy_expression] = "lazy_expression",
  [sym_local_open_expression] = "local_open_expression",
  [sym_package_expression] = "package_expression",
  [sym_new_expression] = "new_expression",
  [sym_object_copy_expression] = "object_copy_expression",
  [sym_instance_variable_expression] = "instance_variable_expression",
  [sym_method_invocation] = "method_invocation",
  [sym_object_expression] = "object_expression",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym__delimited_pattern] = "_delimited_pattern",
  [sym__simple_pattern] = "_simple_pattern",
  [sym__effect_pattern] = "_effect_pattern",
  [sym__pattern] = "_pattern",
  [sym__delimited_binding_pattern] = "_delimited_binding_pattern",
  [sym__simple_binding_pattern] = "_simple_binding_pattern",
  [sym__effect_binding_pattern] = "_effect_binding_pattern",
  [sym__binding_pattern_no_exn] = "_binding_pattern_no_exn",
  [sym__binding_pattern] = "_binding_pattern",
  [sym_alias_pattern] = "alias_pattern",
  [sym_alias_binding_pattern_no_exn] = "alias_pattern",
  [sym_alias_binding_pattern] = "alias_pattern",
  [sym_typed_pattern] = "typed_pattern",
  [sym_typed_binding_pattern] = "typed_pattern",
  [sym__or_pattern_anonymous] = "_or_pattern_anonymous",
  [sym__or_binding_pattern_no_exn_anonymous] = "or_pattern",
  [sym__or_binding_pattern_anonymous] = "_or_binding_pattern_anonymous",
  [sym_constructor_pattern] = "constructor_pattern",
  [sym_constructor_binding_pattern] = "constructor_pattern",
  [sym_tag_pattern] = "tag_pattern",
  [sym_tag_binding_pattern] = "tag_pattern",
  [sym_polymorphic_variant_pattern] = "polymorphic_variant_pattern",
  [sym_labeled_tuple_element_pattern] = "labeled_tuple_element_pattern",
  [sym__tuple_pattern] = "_tuple_pattern",
  [sym_labeled_tuple_element_binding_pattern] = "labeled_tuple_element_binding_pattern",
  [sym__tuple_binding_pattern_no_exn] = "tuple_pattern",
  [sym__tuple_binding_pattern] = "_tuple_binding_pattern",
  [sym_record_pattern] = "record_pattern",
  [sym_field_pattern] = "field_pattern",
  [sym_record_binding_pattern] = "record_pattern",
  [sym_field_binding_pattern] = "field_pattern",
  [sym_list_pattern] = "list_pattern",
  [sym_list_binding_pattern] = "list_pattern",
  [sym_cons_pattern] = "cons_pattern",
  [sym_cons_binding_pattern_no_exn] = "cons_pattern",
  [sym_cons_binding_pattern] = "cons_pattern",
  [sym_array_pattern] = "array_pattern",
  [sym_array_binding_pattern] = "array_pattern",
  [sym__sequence_pattern_content] = "_sequence_pattern_content",
  [sym__sequence_binding_pattern_content] = "_sequence_binding_pattern_content",
  [sym_range_pattern] = "range_pattern",
  [sym_lazy_pattern] = "lazy_pattern",
  [sym_lazy_binding_pattern] = "lazy_pattern",
  [sym_local_open_pattern] = "local_open_pattern",
  [sym_local_open_binding_pattern] = "local_open_pattern",
  [sym_package_pattern] = "package_pattern",
  [sym_parenthesized_pattern] = "parenthesized_pattern",
  [sym_parenthesized_binding_pattern] = "parenthesized_pattern",
  [sym_exception_pattern] = "exception_pattern",
  [sym_exception_binding_pattern] = "exception_pattern",
  [sym_effect_pattern] = "effect_pattern",
  [sym_effect_binding_pattern] = "effect_pattern",
  [sym_attribute] = "attribute",
  [sym_item_attribute] = "item_attribute",
  [sym_floating_attribute] = "floating_attribute",
  [sym_attribute_payload] = "attribute_payload",
  [sym_extension] = "extension",
  [sym_quoted_extension] = "quoted_extension",
  [sym_item_extension] = "item_extension",
  [sym_quoted_item_extension] = "quoted_item_extension",
  [sym__attribute] = "_attribute",
  [sym__constant] = "_constant",
  [sym__signed_constant] = "_signed_constant",
  [sym_number] = "number",
  [sym_signed_number] = "signed_number",
  [sym_character] = "character",
  [sym_character_content] = "character_content",
  [sym_string] = "string",
  [sym_string_content] = "string_content",
  [sym_quoted_string] = "quoted_string",
  [sym__quoted_string] = "_quoted_string",
  [sym_quoted_string_content] = "quoted_string_content",
  [sym_escape_sequence] = "escape_sequence",
  [sym_boolean] = "boolean",
  [sym_unit] = "unit",
  [sym_sign_operator] = "sign_operator",
  [sym__infix_operator] = "_infix_operator",
  [sym_pow_operator] = "pow_operator",
  [sym_mult_operator] = "mult_operator",
  [sym_add_operator] = "add_operator",
  [sym_and_operator] = "and_operator",
  [sym_or_operator] = "or_operator",
  [sym_indexing_operator_path] = "indexing_operator_path",
  [sym__value_name] = "_value_name",
  [sym__simple_value_pattern] = "_simple_value_pattern",
  [sym_parenthesized_operator] = "parenthesized_operator",
  [sym_value_path] = "value_path",
  [sym_module_path] = "module_path",
  [sym_extended_module_path] = "extended_module_path",
  [sym_module_type_path] = "module_type_path",
  [sym_field_path] = "field_path",
  [sym_constructor_path] = "constructor_path",
  [sym_type_constructor_path] = "type_constructor_path",
  [sym_class_path] = "class_path",
  [sym_class_type_path] = "class_type_path",
  [sym__instance_variable_name] = "_instance_variable_name",
  [sym__constructor_name] = "_constructor_name",
  [sym_directive] = "directive",
  [sym_type_variable] = "type_variable",
  [sym_tag] = "tag",
  [sym_attribute_id] = "attribute_id",
  [aux_sym__structure_repeat1] = "_structure_repeat1",
  [aux_sym__structure_repeat2] = "_structure_repeat2",
  [aux_sym_expression_item_repeat1] = "expression_item_repeat1",
  [aux_sym__signature_repeat1] = "_signature_repeat1",
  [aux_sym_value_definition_repeat1] = "value_definition_repeat1",
  [aux_sym_let_binding_repeat1] = "let_binding_repeat1",
  [aux_sym_external_repeat1] = "external_repeat1",
  [aux_sym_type_definition_repeat1] = "type_definition_repeat1",
  [aux_sym_type_binding_repeat1] = "type_binding_repeat1",
  [aux_sym__type_params_repeat1] = "_type_params_repeat1",
  [aux_sym__type_param_repeat1] = "_type_param_repeat1",
  [aux_sym_variant_declaration_repeat1] = "variant_declaration_repeat1",
  [aux_sym_constructor_declaration_repeat1] = "constructor_declaration_repeat1",
  [aux_sym__constructor_argument_repeat1] = "_constructor_argument_repeat1",
  [aux_sym_record_declaration_repeat1] = "record_declaration_repeat1",
  [aux_sym_module_definition_repeat1] = "module_definition_repeat1",
  [aux_sym_module_binding_repeat1] = "module_binding_repeat1",
  [aux_sym_class_definition_repeat1] = "class_definition_repeat1",
  [aux_sym_class_type_definition_repeat1] = "class_type_definition_repeat1",
  [aux_sym_module_type_constraint_repeat1] = "module_type_constraint_repeat1",
  [aux_sym_instantiated_class_type_repeat1] = "instantiated_class_type_repeat1",
  [aux_sym_class_body_type_repeat1] = "class_body_type_repeat1",
  [aux_sym_instance_variable_specification_repeat1] = "instance_variable_specification_repeat1",
  [aux_sym_method_specification_repeat1] = "method_specification_repeat1",
  [aux_sym_class_application_repeat1] = "class_application_repeat1",
  [aux_sym__abstract_type_repeat1] = "_abstract_type_repeat1",
  [aux_sym_polymorphic_variant_type_repeat1] = "polymorphic_variant_type_repeat1",
  [aux_sym_polymorphic_variant_type_repeat2] = "polymorphic_variant_type_repeat2",
  [aux_sym_tag_specification_repeat1] = "tag_specification_repeat1",
  [aux_sym_object_type_repeat1] = "object_type_repeat1",
  [aux_sym__sequence_expression_content_repeat1] = "_sequence_expression_content_repeat1",
  [aux_sym_record_expression_repeat1] = "record_expression_repeat1",
  [aux_sym__match_cases_repeat1] = "_match_cases_repeat1",
  [aux_sym_object_copy_expression_repeat1] = "object_copy_expression_repeat1",
  [aux_sym_object_expression_repeat1] = "object_expression_repeat1",
  [aux_sym_record_pattern_repeat1] = "record_pattern_repeat1",
  [aux_sym_record_binding_pattern_repeat1] = "record_binding_pattern_repeat1",
  [aux_sym__sequence_pattern_content_repeat1] = "_sequence_pattern_content_repeat1",
  [aux_sym__sequence_binding_pattern_content_repeat1] = "_sequence_binding_pattern_content_repeat1",
  [aux_sym_string_content_repeat1] = "string_content_repeat1",
  [aux_sym_quoted_string_content_repeat1] = "quoted_string_content_repeat1",
  [aux_sym_attribute_id_repeat1] = "attribute_id_repeat1",
  [alias_sym_class_name] = "class_name",
  [alias_sym_class_type_name] = "class_type_name",
  [alias_sym_field_name] = "field_name",
  [alias_sym_instance_variable_name] = "instance_variable_name",
  [alias_sym_label_name] = "label_name",
  [alias_sym_method_name] = "method_name",
  [alias_sym_module_name] = "module_name",
  [alias_sym_module_type_name] = "module_type_name",
  [alias_sym_sequence_expression] = "sequence_expression",
  [alias_sym_tuple_expression] = "tuple_expression",
  [alias_sym_type_constructor] = "type_constructor",
  [alias_sym_value_name] = "value_name",
  [alias_sym_value_pattern] = "value_pattern",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym__lowercase_identifier] = sym__lowercase_identifier,
  [anon_sym_SEMI_SEMI] = anon_sym_SEMI_SEMI,
  [anon_sym_let] = anon_sym_let,
  [anon_sym_rec] = anon_sym_rec,
  [anon_sym_and] = anon_sym_and,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_external] = anon_sym_external,
  [anon_sym_type] = anon_sym_type,
  [anon_sym_nonrec] = anon_sym_nonrec,
  [anon_sym_COLON_EQ] = anon_sym_COLON_EQ,
  [anon_sym_private] = anon_sym_private,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym__] = anon_sym__,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_of] = anon_sym_of,
  [anon_sym_COLON2] = anon_sym_COLON,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_true] = anon_sym_true,
  [anon_sym_false] = anon_sym_false,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_mutable] = anon_sym_mutable,
  [anon_sym_constraint] = anon_sym_constraint,
  [anon_sym_exception] = anon_sym_exception,
  [anon_sym_module] = anon_sym_module,
  [anon_sym_open] = anon_sym_open,
  [anon_sym_include] = anon_sym_include,
  [anon_sym_class] = anon_sym_class,
  [anon_sym_virtual] = anon_sym_virtual,
  [anon_sym_val] = anon_sym_val,
  [anon_sym_sig] = anon_sym_sig,
  [anon_sym_end] = anon_sym_end,
  [anon_sym_with] = anon_sym_with,
  [anon_sym_functor] = anon_sym_functor,
  [anon_sym_struct] = anon_sym_struct,
  [anon_sym_COLON_GT] = anon_sym_COLON_GT,
  [anon_sym_object] = anon_sym_object,
  [anon_sym_inherit] = anon_sym_inherit,
  [anon_sym_method] = anon_sym_method,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_fun] = anon_sym_fun,
  [anon_sym_as] = anon_sym_as,
  [anon_sym_initializer] = anon_sym_initializer,
  [anon_sym_LBRACK_GT] = anon_sym_LBRACK_GT,
  [anon_sym_LBRACK_LT] = anon_sym_LBRACK_LT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_POUND] = anon_sym_POUND,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym_LBRACK_PIPE] = anon_sym_LBRACK_PIPE,
  [anon_sym_PIPE_RBRACK] = anon_sym_PIPE_RBRACK,
  [anon_sym_LT_DASH] = anon_sym_LT_DASH,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_then] = anon_sym_then,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_done] = anon_sym_done,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_to] = anon_sym_to,
  [anon_sym_downto] = anon_sym_downto,
  [anon_sym_match] = anon_sym_match,
  [anon_sym_when] = anon_sym_when,
  [anon_sym_function] = anon_sym_function,
  [anon_sym_try] = anon_sym_try,
  [anon_sym_assert] = anon_sym_assert,
  [anon_sym_lazy] = anon_sym_lazy,
  [anon_sym_new] = anon_sym_new,
  [anon_sym_LBRACE_LT] = anon_sym_LBRACE_LT,
  [anon_sym_GT_RBRACE] = anon_sym_GT_RBRACE,
  [anon_sym_begin] = anon_sym_begin,
  [sym_ocamlyacc_value] = sym_ocamlyacc_value,
  [anon_sym_effect] = anon_sym_effect,
  [aux_sym_attribute_token1] = aux_sym_attribute_token1,
  [anon_sym_LBRACK_AT_AT] = anon_sym_LBRACK_AT_AT,
  [anon_sym_LBRACK_AT_AT_AT] = anon_sym_LBRACK_AT_AT_AT,
  [anon_sym_LBRACK_PERCENT] = anon_sym_LBRACK_PERCENT,
  [anon_sym_LBRACE_PERCENT] = anon_sym_LBRACE_PERCENT,
  [aux_sym_quoted_extension_token1] = aux_sym_quoted_extension_token1,
  [anon_sym_LBRACK_PERCENT_PERCENT] = anon_sym_LBRACK_PERCENT_PERCENT,
  [anon_sym_LBRACE_PERCENT_PERCENT] = anon_sym_LBRACE_PERCENT_PERCENT,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [aux_sym_number_token1] = aux_sym_number_token1,
  [aux_sym_signed_number_token1] = aux_sym_signed_number_token1,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [anon_sym_SQUOTE2] = anon_sym_SQUOTE,
  [aux_sym_character_content_token1] = aux_sym_character_content_token1,
  [aux_sym_character_content_token2] = aux_sym_character_content_token2,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym_string_content_token1] = aux_sym_string_content_token1,
  [aux_sym_string_content_token2] = aux_sym_string_content_token2,
  [aux_sym_string_content_token3] = aux_sym_string_content_token3,
  [aux_sym_string_content_token4] = sym_escape_sequence,
  [aux_sym_string_content_token5] = sym_escape_sequence,
  [aux_sym_quoted_string_content_token1] = aux_sym_quoted_string_content_token1,
  [aux_sym_escape_sequence_token1] = aux_sym_escape_sequence_token1,
  [aux_sym_escape_sequence_token2] = aux_sym_escape_sequence_token2,
  [aux_sym_escape_sequence_token3] = aux_sym_escape_sequence_token3,
  [aux_sym_escape_sequence_token4] = aux_sym_escape_sequence_token4,
  [sym_conversion_specification] = sym_conversion_specification,
  [sym_pretty_printing_indication] = sym_pretty_printing_indication,
  [sym_prefix_operator] = sym_prefix_operator,
  [aux_sym_sign_operator_token1] = aux_sym_sign_operator_token1,
  [sym_hash_operator] = sym_hash_operator,
  [anon_sym_lsl] = anon_sym_lsl,
  [anon_sym_lsr] = anon_sym_lsr,
  [anon_sym_asr] = anon_sym_asr,
  [aux_sym_pow_operator_token1] = aux_sym_pow_operator_token1,
  [anon_sym_mod] = anon_sym_mod,
  [anon_sym_land] = anon_sym_land,
  [anon_sym_lor] = anon_sym_lor,
  [anon_sym_lxor] = anon_sym_lxor,
  [aux_sym_mult_operator_token1] = aux_sym_mult_operator_token1,
  [aux_sym_add_operator_token1] = aux_sym_add_operator_token1,
  [sym_concat_operator] = sym_concat_operator,
  [sym_rel_operator] = sym_rel_operator,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_or] = anon_sym_or,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [sym_assign_operator] = sym_assign_operator,
  [sym_indexing_operator] = sym_indexing_operator,
  [sym_let_operator] = sym_let_operator,
  [sym_let_and_operator] = sym_let_and_operator,
  [sym_match_operator] = sym_match_operator,
  [sym__uppercase_identifier] = sym__uppercase_identifier,
  [aux_sym_directive_token1] = aux_sym_directive_token1,
  [aux_sym_type_variable_token1] = aux_sym_type_variable_token1,
  [aux_sym_tag_token1] = aux_sym_tag_token1,
  [aux_sym_attribute_id_token1] = aux_sym_attribute_id_token1,
  [sym_comment] = sym_comment,
  [sym__left_quoted_string_delimiter] = sym__left_quoted_string_delimiter,
  [sym__right_quoted_string_delimiter] = sym__right_quoted_string_delimiter,
  [sym_line_number_directive] = sym_line_number_directive,
  [sym__null] = sym__null,
  [sym__error_sentinel] = sym__error_sentinel,
  [sym_compilation_unit] = sym_compilation_unit,
  [sym__structure] = sym__structure,
  [sym_expression_item] = sym_expression_item,
  [sym__signature] = sym__signature,
  [sym_toplevel_directive] = sym_toplevel_directive,
  [sym__local_structure_item] = sym__local_structure_item,
  [sym__structure_item] = sym__structure_item,
  [sym_value_definition] = sym_value_definition,
  [sym_let_binding] = sym_let_binding,
  [sym_parameter] = sym_parameter,
  [sym_external] = sym_external,
  [sym_type_definition] = sym_type_definition,
  [sym_type_binding] = sym_type_binding,
  [sym__type_params] = sym__type_params,
  [sym__type_param] = sym__type_param,
  [sym_variant_declaration] = sym_variant_declaration,
  [sym_constructor_declaration] = sym_constructor_declaration,
  [sym__constructor_declaration_name] = sym__constructor_declaration_name,
  [sym__constructor_argument] = sym__constructor_argument,
  [sym_record_declaration] = sym_record_declaration,
  [sym_field_declaration] = sym_field_declaration,
  [sym_external_declaration] = sym_external_declaration,
  [sym_type_constraint] = sym_type_constraint,
  [sym_exception_definition] = sym_exception_definition,
  [sym_module_definition] = sym_module_definition,
  [sym_module_binding] = sym_module_binding,
  [sym_module_parameter] = sym_module_parameter,
  [sym_module_type_definition] = sym_module_type_definition,
  [sym_open_module] = sym_open_module,
  [sym_include_module] = sym_include_module,
  [sym_class_definition] = sym_class_definition,
  [sym_class_binding] = sym_class_binding,
  [sym_class_type_definition] = sym_class_type_definition,
  [sym_class_type_binding] = sym_class_type_binding,
  [sym__signature_item] = sym__signature_item,
  [sym_value_specification] = sym_value_specification,
  [sym_include_module_type] = sym_include_module_type,
  [sym__module_typed] = sym__module_typed,
  [sym__module_type] = sym__module_type,
  [sym_signature] = sym_signature,
  [sym_module_type_constraint] = sym_module_type_constraint,
  [sym_constrain_type] = sym_constrain_type,
  [sym_constrain_module] = sym_constrain_module,
  [sym_constrain_module_type] = sym_constrain_module_type,
  [sym_module_type_of] = sym_module_type_of,
  [sym_functor_type] = sym_functor_type,
  [sym_parenthesized_module_type] = sym_parenthesized_module_type,
  [sym__simple_module_expression] = sym__simple_module_expression,
  [sym__module_expression] = sym__module_expression,
  [sym_structure] = sym_structure,
  [sym_functor] = sym_functor,
  [sym_module_application] = sym_module_application,
  [sym_typed_module_expression] = sym_typed_module_expression,
  [sym_packed_module] = sym_packed_module,
  [sym_parenthesized_module_expression] = sym_parenthesized_module_expression,
  [sym__class_typed] = sym__class_typed,
  [sym__simple_class_type] = sym__simple_class_type,
  [sym__class_type] = sym__class_type,
  [sym_instantiated_class_type] = sym_instantiated_class_type,
  [sym_class_body_type] = sym_class_body_type,
  [sym__class_field_specification] = sym__class_field_specification,
  [sym_inheritance_specification] = sym_inheritance_specification,
  [sym_instance_variable_specification] = sym_instance_variable_specification,
  [sym_method_specification] = sym_method_specification,
  [sym_type_parameter_constraint] = sym_type_parameter_constraint,
  [sym_let_open_class_type] = sym_let_open_class_type,
  [sym_class_function_type] = sym_class_function_type,
  [sym__simple_class_expression] = sym__simple_class_expression,
  [sym__class_expression] = sym__class_expression,
  [sym_instantiated_class] = sym_instantiated_class,
  [sym_typed_class_expression] = sym_typed_class_expression,
  [sym_class_function] = sym_class_function,
  [sym_class_application] = sym_class_application,
  [sym_let_class_expression] = sym_let_class_expression,
  [sym__class_field] = sym__class_field,
  [sym_inheritance_definition] = sym_inheritance_definition,
  [sym_instance_variable_definition] = sym_instance_variable_definition,
  [sym_method_definition] = sym_method_definition,
  [sym_class_initializer] = sym_class_initializer,
  [sym_let_open_class_expression] = sym_let_open_class_expression,
  [sym_parenthesized_class_expression] = sym_parenthesized_class_expression,
  [sym__typed] = sym__typed,
  [sym__simple_typed] = sym__simple_typed,
  [sym__coerced] = sym__coerced,
  [sym__type_constrained] = sym__type_constrained,
  [sym__maybe_polymorphic_typed] = sym__maybe_polymorphic_typed,
  [sym__polymorphic_typed] = sym__polymorphic_typed,
  [sym__polymorphic_type] = sym__polymorphic_type,
  [sym_polymorphic_type] = sym_polymorphic_type,
  [sym__abstract_type] = sym__abstract_type,
  [sym__parenthesized_abstract_type] = sym__parenthesized_abstract_type,
  [sym__delimited_type] = sym__delimited_type,
  [sym__simple_type] = sym__simple_type,
  [sym__type] = sym__type,
  [sym_function_type] = sym_function_type,
  [sym_labeled_argument_type] = sym_labeled_argument_type,
  [sym__proper_tuple_type] = sym__proper_tuple_type,
  [sym__labeled_tuple_type] = sym__proper_tuple_type,
  [sym_labeled_tuple_element_type] = sym_labeled_tuple_element_type,
  [sym__tuple_type_rhs] = sym__tuple_type_rhs,
  [sym_constructed_type] = sym_constructed_type,
  [sym_aliased_type] = sym_aliased_type,
  [sym_local_open_type] = sym_local_open_type,
  [sym_polymorphic_variant_type] = sym_polymorphic_variant_type,
  [sym__tag_spec] = sym__tag_spec,
  [sym_tag_specification] = sym_tag_specification,
  [sym_package_type] = sym_package_type,
  [sym_object_type] = sym_object_type,
  [sym_method_type] = sym_method_type,
  [sym_hash_type] = sym_hash_type,
  [sym_parenthesized_type] = sym_parenthesized_type,
  [sym__delimited_expression] = sym__delimited_expression,
  [sym__simple_expression] = sym__simple_expression,
  [sym__expression] = sym__expression,
  [sym__sequence_expression] = sym__sequence_expression,
  [sym_typed_expression] = sym_typed_expression,
  [sym_labeled_tuple_element] = sym_labeled_tuple_element,
  [sym__tuple_expression] = sym__tuple_expression,
  [sym_cons_expression] = sym_cons_expression,
  [sym_list_expression] = sym_list_expression,
  [sym_array_expression] = sym_array_expression,
  [sym__sequence_expression_content] = sym__sequence_expression_content,
  [sym_record_expression] = sym_record_expression,
  [sym_field_expression] = sym_field_expression,
  [sym_application_expression] = sym_application_expression,
  [sym_labeled_argument] = sym_labeled_argument,
  [sym_prefix_expression] = sym_prefix_expression,
  [sym_sign_expression] = sym_sign_expression,
  [sym_hash_expression] = sym_hash_expression,
  [sym_infix_expression] = sym_infix_expression,
  [sym_field_get_expression] = sym_field_get_expression,
  [sym_array_get_expression] = sym_array_get_expression,
  [sym_string_get_expression] = sym_string_get_expression,
  [sym_bigarray_get_expression] = sym_bigarray_get_expression,
  [sym_set_expression] = sym_set_expression,
  [sym_if_expression] = sym_if_expression,
  [sym_then_clause] = sym_then_clause,
  [sym_else_clause] = sym_else_clause,
  [sym_while_expression] = sym_while_expression,
  [sym_do_clause] = sym_do_clause,
  [sym_for_expression] = sym_for_expression,
  [sym__sequence_expression_anonymous] = sym__sequence_expression_anonymous,
  [sym_match_expression] = sym_match_expression,
  [sym__match_cases] = sym__match_cases,
  [sym_match_case] = sym_match_case,
  [sym_guard] = sym_guard,
  [sym_refutation_case] = sym_refutation_case,
  [sym_function_expression] = sym_function_expression,
  [sym_fun_expression] = sym_fun_expression,
  [sym_try_expression] = sym_try_expression,
  [sym_let_expression] = sym_let_expression,
  [sym_assert_expression] = sym_assert_expression,
  [sym_lazy_expression] = sym_lazy_expression,
  [sym_local_open_expression] = sym_local_open_expression,
  [sym_package_expression] = sym_package_expression,
  [sym_new_expression] = sym_new_expression,
  [sym_object_copy_expression] = sym_object_copy_expression,
  [sym_instance_variable_expression] = sym_instance_variable_expression,
  [sym_method_invocation] = sym_method_invocation,
  [sym_object_expression] = sym_object_expression,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym__delimited_pattern] = sym__delimited_pattern,
  [sym__simple_pattern] = sym__simple_pattern,
  [sym__effect_pattern] = sym__effect_pattern,
  [sym__pattern] = sym__pattern,
  [sym__delimited_binding_pattern] = sym__delimited_binding_pattern,
  [sym__simple_binding_pattern] = sym__simple_binding_pattern,
  [sym__effect_binding_pattern] = sym__effect_binding_pattern,
  [sym__binding_pattern_no_exn] = sym__binding_pattern_no_exn,
  [sym__binding_pattern] = sym__binding_pattern,
  [sym_alias_pattern] = sym_alias_pattern,
  [sym_alias_binding_pattern_no_exn] = sym_alias_pattern,
  [sym_alias_binding_pattern] = sym_alias_pattern,
  [sym_typed_pattern] = sym_typed_pattern,
  [sym_typed_binding_pattern] = sym_typed_pattern,
  [sym__or_pattern_anonymous] = sym__or_pattern_anonymous,
  [sym__or_binding_pattern_no_exn_anonymous] = sym__or_binding_pattern_no_exn_anonymous,
  [sym__or_binding_pattern_anonymous] = sym__or_binding_pattern_anonymous,
  [sym_constructor_pattern] = sym_constructor_pattern,
  [sym_constructor_binding_pattern] = sym_constructor_pattern,
  [sym_tag_pattern] = sym_tag_pattern,
  [sym_tag_binding_pattern] = sym_tag_pattern,
  [sym_polymorphic_variant_pattern] = sym_polymorphic_variant_pattern,
  [sym_labeled_tuple_element_pattern] = sym_labeled_tuple_element_pattern,
  [sym__tuple_pattern] = sym__tuple_pattern,
  [sym_labeled_tuple_element_binding_pattern] = sym_labeled_tuple_element_binding_pattern,
  [sym__tuple_binding_pattern_no_exn] = sym__tuple_binding_pattern_no_exn,
  [sym__tuple_binding_pattern] = sym__tuple_binding_pattern,
  [sym_record_pattern] = sym_record_pattern,
  [sym_field_pattern] = sym_field_pattern,
  [sym_record_binding_pattern] = sym_record_pattern,
  [sym_field_binding_pattern] = sym_field_pattern,
  [sym_list_pattern] = sym_list_pattern,
  [sym_list_binding_pattern] = sym_list_pattern,
  [sym_cons_pattern] = sym_cons_pattern,
  [sym_cons_binding_pattern_no_exn] = sym_cons_pattern,
  [sym_cons_binding_pattern] = sym_cons_pattern,
  [sym_array_pattern] = sym_array_pattern,
  [sym_array_binding_pattern] = sym_array_pattern,
  [sym__sequence_pattern_content] = sym__sequence_pattern_content,
  [sym__sequence_binding_pattern_content] = sym__sequence_binding_pattern_content,
  [sym_range_pattern] = sym_range_pattern,
  [sym_lazy_pattern] = sym_lazy_pattern,
  [sym_lazy_binding_pattern] = sym_lazy_pattern,
  [sym_local_open_pattern] = sym_local_open_pattern,
  [sym_local_open_binding_pattern] = sym_local_open_pattern,
  [sym_package_pattern] = sym_package_pattern,
  [sym_parenthesized_pattern] = sym_parenthesized_pattern,
  [sym_parenthesized_binding_pattern] = sym_parenthesized_pattern,
  [sym_exception_pattern] = sym_exception_pattern,
  [sym_exception_binding_pattern] = sym_exception_pattern,
  [sym_effect_pattern] = sym_effect_pattern,
  [sym_effect_binding_pattern] = sym_effect_pattern,
  [sym_attribute] = sym_attribute,
  [sym_item_attribute] = sym_item_attribute,
  [sym_floating_attribute] = sym_floating_attribute,
  [sym_attribute_payload] = sym_attribute_payload,
  [sym_extension] = sym_extension,
  [sym_quoted_extension] = sym_quoted_extension,
  [sym_item_extension] = sym_item_extension,
  [sym_quoted_item_extension] = sym_quoted_item_extension,
  [sym__attribute] = sym__attribute,
  [sym__constant] = sym__constant,
  [sym__signed_constant] = sym__signed_constant,
  [sym_number] = sym_number,
  [sym_signed_number] = sym_signed_number,
  [sym_character] = sym_character,
  [sym_character_content] = sym_character_content,
  [sym_string] = sym_string,
  [sym_string_content] = sym_string_content,
  [sym_quoted_string] = sym_quoted_string,
  [sym__quoted_string] = sym__quoted_string,
  [sym_quoted_string_content] = sym_quoted_string_content,
  [sym_escape_sequence] = sym_escape_sequence,
  [sym_boolean] = sym_boolean,
  [sym_unit] = sym_unit,
  [sym_sign_operator] = sym_sign_operator,
  [sym__infix_operator] = sym__infix_operator,
  [sym_pow_operator] = sym_pow_operator,
  [sym_mult_operator] = sym_mult_operator,
  [sym_add_operator] = sym_add_operator,
  [sym_and_operator] = sym_and_operator,
  [sym_or_operator] = sym_or_operator,
  [sym_indexing_operator_path] = sym_indexing_operator_path,
  [sym__value_name] = sym__value_name,
  [sym__simple_value_pattern] = sym__simple_value_pattern,
  [sym_parenthesized_operator] = sym_parenthesized_operator,
  [sym_value_path] = sym_value_path,
  [sym_module_path] = sym_module_path,
  [sym_extended_module_path] = sym_extended_module_path,
  [sym_module_type_path] = sym_module_type_path,
  [sym_field_path] = sym_field_path,
  [sym_constructor_path] = sym_constructor_path,
  [sym_type_constructor_path] = sym_type_constructor_path,
  [sym_class_path] = sym_class_path,
  [sym_class_type_path] = sym_class_type_path,
  [sym__instance_variable_name] = sym__instance_variable_name,
  [sym__constructor_name] = sym__constructor_name,
  [sym_directive] = sym_directive,
  [sym_type_variable] = sym_type_variable,
  [sym_tag] = sym_tag,
  [sym_attribute_id] = sym_attribute_id,
  [aux_sym__structure_repeat1] = aux_sym__structure_repeat1,
  [aux_sym__structure_repeat2] = aux_sym__structure_repeat2,
  [aux_sym_expression_item_repeat1] = aux_sym_expression_item_repeat1,
  [aux_sym__signature_repeat1] = aux_sym__signature_repeat1,
  [aux_sym_value_definition_repeat1] = aux_sym_value_definition_repeat1,
  [aux_sym_let_binding_repeat1] = aux_sym_let_binding_repeat1,
  [aux_sym_external_repeat1] = aux_sym_external_repeat1,
  [aux_sym_type_definition_repeat1] = aux_sym_type_definition_repeat1,
  [aux_sym_type_binding_repeat1] = aux_sym_type_binding_repeat1,
  [aux_sym__type_params_repeat1] = aux_sym__type_params_repeat1,
  [aux_sym__type_param_repeat1] = aux_sym__type_param_repeat1,
  [aux_sym_variant_declaration_repeat1] = aux_sym_variant_declaration_repeat1,
  [aux_sym_constructor_declaration_repeat1] = aux_sym_constructor_declaration_repeat1,
  [aux_sym__constructor_argument_repeat1] = aux_sym__constructor_argument_repeat1,
  [aux_sym_record_declaration_repeat1] = aux_sym_record_declaration_repeat1,
  [aux_sym_module_definition_repeat1] = aux_sym_module_definition_repeat1,
  [aux_sym_module_binding_repeat1] = aux_sym_module_binding_repeat1,
  [aux_sym_class_definition_repeat1] = aux_sym_class_definition_repeat1,
  [aux_sym_class_type_definition_repeat1] = aux_sym_class_type_definition_repeat1,
  [aux_sym_module_type_constraint_repeat1] = aux_sym_module_type_constraint_repeat1,
  [aux_sym_instantiated_class_type_repeat1] = aux_sym_instantiated_class_type_repeat1,
  [aux_sym_class_body_type_repeat1] = aux_sym_class_body_type_repeat1,
  [aux_sym_instance_variable_specification_repeat1] = aux_sym_instance_variable_specification_repeat1,
  [aux_sym_method_specification_repeat1] = aux_sym_method_specification_repeat1,
  [aux_sym_class_application_repeat1] = aux_sym_class_application_repeat1,
  [aux_sym__abstract_type_repeat1] = aux_sym__abstract_type_repeat1,
  [aux_sym_polymorphic_variant_type_repeat1] = aux_sym_polymorphic_variant_type_repeat1,
  [aux_sym_polymorphic_variant_type_repeat2] = aux_sym_polymorphic_variant_type_repeat2,
  [aux_sym_tag_specification_repeat1] = aux_sym_tag_specification_repeat1,
  [aux_sym_object_type_repeat1] = aux_sym_object_type_repeat1,
  [aux_sym__sequence_expression_content_repeat1] = aux_sym__sequence_expression_content_repeat1,
  [aux_sym_record_expression_repeat1] = aux_sym_record_expression_repeat1,
  [aux_sym__match_cases_repeat1] = aux_sym__match_cases_repeat1,
  [aux_sym_object_copy_expression_repeat1] = aux_sym_object_copy_expression_repeat1,
  [aux_sym_object_expression_repeat1] = aux_sym_object_expression_repeat1,
  [aux_sym_record_pattern_repeat1] = aux_sym_record_pattern_repeat1,
  [aux_sym_record_binding_pattern_repeat1] = aux_sym_record_binding_pattern_repeat1,
  [aux_sym__sequence_pattern_content_repeat1] = aux_sym__sequence_pattern_content_repeat1,
  [aux_sym__sequence_binding_pattern_content_repeat1] = aux_sym__sequence_binding_pattern_content_repeat1,
  [aux_sym_string_content_repeat1] = aux_sym_string_content_repeat1,
  [aux_sym_quoted_string_content_repeat1] = aux_sym_quoted_string_content_repeat1,
  [aux_sym_attribute_id_repeat1] = aux_sym_attribute_id_repeat1,
  [alias_sym_class_name] = alias_sym_class_name,
  [alias_sym_class_type_name] = alias_sym_class_type_name,
  [alias_sym_field_name] = alias_sym_field_name,
  [alias_sym_instance_variable_name] = alias_sym_instance_variable_name,
  [alias_sym_label_name] = alias_sym_label_name,
  [alias_sym_method_name] = alias_sym_method_name,
  [alias_sym_module_name] = alias_sym_module_name,
  [alias_sym_module_type_name] = alias_sym_module_type_name,
  [alias_sym_sequence_expression] = alias_sym_sequence_expression,
  [alias_sym_tuple_expression] = alias_sym_tuple_expression,
  [alias_sym_type_constructor] = alias_sym_type_constructor,
  [alias_sym_value_name] = alias_sym_value_name,
  [alias_sym_value_pattern] = alias_sym_value_pattern,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym__lowercase_identifier] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_SEMI_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_let] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_rec] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_and] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_external] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_type] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nonrec] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_private] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_of] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_mutable] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_constraint] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_exception] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_module] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_open] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_include] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_class] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_virtual] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_val] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_sig] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_end] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_with] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_functor] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_struct] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_object] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_inherit] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_method] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fun] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_as] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_initializer] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_then] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_done] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_to] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_downto] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_match] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_when] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_function] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_try] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_assert] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lazy] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_new] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_begin] = {
    .visible = true,
    .named = false,
  },
  [sym_ocamlyacc_value] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_effect] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_attribute_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_AT_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_AT_AT_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_quoted_extension_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_LBRACK_PERCENT_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE_PERCENT_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_number_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_signed_number_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_character_content_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_character_content_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_string_content_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_token4] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_string_content_token5] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_quoted_string_content_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token4] = {
    .visible = false,
    .named = false,
  },
  [sym_conversion_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_pretty_printing_indication] = {
    .visible = true,
    .named = true,
  },
  [sym_prefix_operator] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_sign_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_hash_operator] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_lsl] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lsr] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_asr] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_pow_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_mod] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_land] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lor] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lxor] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_mult_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_add_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_concat_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_rel_operator] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_or] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [sym_assign_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_indexing_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_let_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_let_and_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_match_operator] = {
    .visible = true,
    .named = true,
  },
  [sym__uppercase_identifier] = {
    .visible = false,
    .named = true,
  },
  [aux_sym_directive_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_variable_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_tag_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_id_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym__left_quoted_string_delimiter] = {
    .visible = false,
    .named = true,
  },
  [sym__right_quoted_string_delimiter] = {
    .visible = false,
    .named = true,
  },
  [sym_line_number_directive] = {
    .visible = true,
    .named = true,
  },
  [sym__null] = {
    .visible = false,
    .named = true,
  },
  [sym__error_sentinel] = {
    .visible = false,
    .named = true,
  },
  [sym_compilation_unit] = {
    .visible = true,
    .named = true,
  },
  [sym__structure] = {
    .visible = false,
    .named = true,
  },
  [sym_expression_item] = {
    .visible = true,
    .named = true,
  },
  [sym__signature] = {
    .visible = false,
    .named = true,
  },
  [sym_toplevel_directive] = {
    .visible = true,
    .named = true,
  },
  [sym__local_structure_item] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__structure_item] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_value_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_let_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_external] = {
    .visible = true,
    .named = true,
  },
  [sym_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_type_binding] = {
    .visible = true,
    .named = true,
  },
  [sym__type_params] = {
    .visible = false,
    .named = true,
  },
  [sym__type_param] = {
    .visible = false,
    .named = true,
  },
  [sym_variant_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_constructor_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__constructor_declaration_name] = {
    .visible = true,
    .named = true,
  },
  [sym__constructor_argument] = {
    .visible = false,
    .named = true,
  },
  [sym_record_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_field_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_external_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_type_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_exception_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_module_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_module_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_module_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_open_module] = {
    .visible = true,
    .named = true,
  },
  [sym_include_module] = {
    .visible = true,
    .named = true,
  },
  [sym_class_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_class_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_class_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_class_type_binding] = {
    .visible = true,
    .named = true,
  },
  [sym__signature_item] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_value_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_include_module_type] = {
    .visible = true,
    .named = true,
  },
  [sym__module_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__module_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_signature] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_constrain_type] = {
    .visible = true,
    .named = true,
  },
  [sym_constrain_module] = {
    .visible = true,
    .named = true,
  },
  [sym_constrain_module_type] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_of] = {
    .visible = true,
    .named = true,
  },
  [sym_functor_type] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_module_type] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_module_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__module_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_structure] = {
    .visible = true,
    .named = true,
  },
  [sym_functor] = {
    .visible = true,
    .named = true,
  },
  [sym_module_application] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_module_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_packed_module] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_module_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__class_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_class_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__class_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_instantiated_class_type] = {
    .visible = true,
    .named = true,
  },
  [sym_class_body_type] = {
    .visible = true,
    .named = true,
  },
  [sym__class_field_specification] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_inheritance_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_instance_variable_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_method_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_type_parameter_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_let_open_class_type] = {
    .visible = true,
    .named = true,
  },
  [sym_class_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_class_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__class_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_instantiated_class] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_class_function] = {
    .visible = true,
    .named = true,
  },
  [sym_class_application] = {
    .visible = true,
    .named = true,
  },
  [sym_let_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__class_field] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_inheritance_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_instance_variable_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_method_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_class_initializer] = {
    .visible = true,
    .named = true,
  },
  [sym_let_open_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__typed] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__coerced] = {
    .visible = false,
    .named = true,
  },
  [sym__type_constrained] = {
    .visible = false,
    .named = true,
  },
  [sym__maybe_polymorphic_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__polymorphic_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__polymorphic_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_polymorphic_type] = {
    .visible = true,
    .named = true,
  },
  [sym__abstract_type] = {
    .visible = false,
    .named = true,
  },
  [sym__parenthesized_abstract_type] = {
    .visible = true,
    .named = true,
  },
  [sym__delimited_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_argument_type] = {
    .visible = true,
    .named = true,
  },
  [sym__proper_tuple_type] = {
    .visible = true,
    .named = true,
  },
  [sym__labeled_tuple_type] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_tuple_element_type] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_type_rhs] = {
    .visible = false,
    .named = true,
  },
  [sym_constructed_type] = {
    .visible = true,
    .named = true,
  },
  [sym_aliased_type] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_type] = {
    .visible = true,
    .named = true,
  },
  [sym_polymorphic_variant_type] = {
    .visible = true,
    .named = true,
  },
  [sym__tag_spec] = {
    .visible = false,
    .named = true,
  },
  [sym_tag_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_package_type] = {
    .visible = true,
    .named = true,
  },
  [sym_object_type] = {
    .visible = true,
    .named = true,
  },
  [sym_method_type] = {
    .visible = true,
    .named = true,
  },
  [sym_hash_type] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_type] = {
    .visible = true,
    .named = true,
  },
  [sym__delimited_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__sequence_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_typed_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_tuple_element] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_cons_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_list_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__sequence_expression_content] = {
    .visible = false,
    .named = true,
  },
  [sym_record_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_field_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_application_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_prefix_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_sign_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hash_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_infix_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_field_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_string_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_bigarray_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_set_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_if_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_then_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_while_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_do_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_for_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__sequence_expression_anonymous] = {
    .visible = false,
    .named = true,
  },
  [sym_match_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__match_cases] = {
    .visible = false,
    .named = true,
  },
  [sym_match_case] = {
    .visible = true,
    .named = true,
  },
  [sym_guard] = {
    .visible = true,
    .named = true,
  },
  [sym_refutation_case] = {
    .visible = true,
    .named = true,
  },
  [sym_function_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_fun_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_try_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_let_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_assert_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_lazy_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_package_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_new_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_object_copy_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_instance_variable_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_method_invocation] = {
    .visible = true,
    .named = true,
  },
  [sym_object_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__delimited_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__effect_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__delimited_binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__effect_binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__binding_pattern_no_exn] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_alias_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_alias_binding_pattern_no_exn] = {
    .visible = true,
    .named = true,
  },
  [sym_alias_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__or_pattern_anonymous] = {
    .visible = false,
    .named = true,
  },
  [sym__or_binding_pattern_no_exn_anonymous] = {
    .visible = true,
    .named = true,
  },
  [sym__or_binding_pattern_anonymous] = {
    .visible = false,
    .named = true,
  },
  [sym_constructor_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_constructor_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_tag_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_tag_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_polymorphic_variant_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_tuple_element_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_labeled_tuple_element_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_binding_pattern_no_exn] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_binding_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_record_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_field_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_record_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_field_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_list_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_list_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_cons_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_cons_binding_pattern_no_exn] = {
    .visible = true,
    .named = true,
  },
  [sym_cons_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_array_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_array_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__sequence_pattern_content] = {
    .visible = false,
    .named = true,
  },
  [sym__sequence_binding_pattern_content] = {
    .visible = false,
    .named = true,
  },
  [sym_range_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_lazy_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_lazy_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_package_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_exception_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_exception_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_effect_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_effect_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_item_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_floating_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_payload] = {
    .visible = true,
    .named = true,
  },
  [sym_extension] = {
    .visible = true,
    .named = true,
  },
  [sym_quoted_extension] = {
    .visible = true,
    .named = true,
  },
  [sym_item_extension] = {
    .visible = true,
    .named = true,
  },
  [sym_quoted_item_extension] = {
    .visible = true,
    .named = true,
  },
  [sym__attribute] = {
    .visible = false,
    .named = true,
  },
  [sym__constant] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__signed_constant] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_number] = {
    .visible = true,
    .named = true,
  },
  [sym_signed_number] = {
    .visible = true,
    .named = true,
  },
  [sym_character] = {
    .visible = true,
    .named = true,
  },
  [sym_character_content] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_string_content] = {
    .visible = true,
    .named = true,
  },
  [sym_quoted_string] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_string] = {
    .visible = false,
    .named = true,
  },
  [sym_quoted_string_content] = {
    .visible = true,
    .named = true,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean] = {
    .visible = true,
    .named = true,
  },
  [sym_unit] = {
    .visible = true,
    .named = true,
  },
  [sym_sign_operator] = {
    .visible = true,
    .named = true,
  },
  [sym__infix_operator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_pow_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_mult_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_add_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_and_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_or_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_indexing_operator_path] = {
    .visible = true,
    .named = true,
  },
  [sym__value_name] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_value_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_parenthesized_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_value_path] = {
    .visible = true,
    .named = true,
  },
  [sym_module_path] = {
    .visible = true,
    .named = true,
  },
  [sym_extended_module_path] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_path] = {
    .visible = true,
    .named = true,
  },
  [sym_field_path] = {
    .visible = true,
    .named = true,
  },
  [sym_constructor_path] = {
    .visible = true,
    .named = true,
  },
  [sym_type_constructor_path] = {
    .visible = true,
    .named = true,
  },
  [sym_class_path] = {
    .visible = true,
    .named = true,
  },
  [sym_class_type_path] = {
    .visible = true,
    .named = true,
  },
  [sym__instance_variable_name] = {
    .visible = false,
    .named = true,
  },
  [sym__constructor_name] = {
    .visible = false,
    .named = true,
  },
  [sym_directive] = {
    .visible = true,
    .named = true,
  },
  [sym_type_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_tag] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_id] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__structure_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__structure_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expression_item_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__signature_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_value_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_let_binding_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_external_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_binding_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__type_params_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__type_param_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_variant_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_constructor_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__constructor_argument_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_binding_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_type_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_type_constraint_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_instantiated_class_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_body_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_instance_variable_specification_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_method_specification_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_application_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__abstract_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_polymorphic_variant_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_polymorphic_variant_type_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_tag_specification_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_object_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__sequence_expression_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__match_cases_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_object_copy_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_object_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_binding_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__sequence_pattern_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__sequence_binding_pattern_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_quoted_string_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_id_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_class_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_class_type_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_field_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_instance_variable_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_label_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_method_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_module_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_module_type_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_sequence_expression] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_tuple_expression] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_type_constructor] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_value_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_value_pattern] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alias = 1,
  field_argument = 2,
  field_array = 3,
  field_body = 4,
  field_class = 5,
  field_class_type = 6,
  field_codomain = 7,
  field_coercion = 8,
  field_condition = 9,
  field_constraint = 10,
  field_continuation = 11,
  field_default = 12,
  field_domain = 13,
  field_effect = 14,
  field_equation = 15,
  field_expression = 16,
  field_field = 17,
  field_from = 18,
  field_function = 19,
  field_functor = 20,
  field_index = 21,
  field_initializer = 22,
  field_left = 23,
  field_method = 24,
  field_module = 25,
  field_module_type = 26,
  field_name = 27,
  field_object = 28,
  field_operator = 29,
  field_pattern = 30,
  field_record = 31,
  field_right = 32,
  field_self = 33,
  field_self_type = 34,
  field_string = 35,
  field_synonym = 36,
  field_to = 37,
  field_type = 38,
  field_value = 39,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alias] = "alias",
  [field_argument] = "argument",
  [field_array] = "array",
  [field_body] = "body",
  [field_class] = "class",
  [field_class_type] = "class_type",
  [field_codomain] = "codomain",
  [field_coercion] = "coercion",
  [field_condition] = "condition",
  [field_constraint] = "constraint",
  [field_continuation] = "continuation",
  [field_default] = "default",
  [field_domain] = "domain",
  [field_effect] = "effect",
  [field_equation] = "equation",
  [field_expression] = "expression",
  [field_field] = "field",
  [field_from] = "from",
  [field_function] = "function",
  [field_functor] = "functor",
  [field_index] = "index",
  [field_initializer] = "initializer",
  [field_left] = "left",
  [field_method] = "method",
  [field_module] = "module",
  [field_module_type] = "module_type",
  [field_name] = "name",
  [field_object] = "object",
  [field_operator] = "operator",
  [field_pattern] = "pattern",
  [field_record] = "record",
  [field_right] = "right",
  [field_self] = "self",
  [field_self_type] = "self_type",
  [field_string] = "string",
  [field_synonym] = "synonym",
  [field_to] = "to",
  [field_type] = "type",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [3] = {.index = 0, .length = 1},
  [7] = {.index = 1, .length = 1},
  [9] = {.index = 2, .length = 1},
  [14] = {.index = 3, .length = 1},
  [17] = {.index = 4, .length = 1},
  [18] = {.index = 5, .length = 1},
  [19] = {.index = 6, .length = 2},
  [20] = {.index = 8, .length = 1},
  [21] = {.index = 9, .length = 1},
  [24] = {.index = 10, .length = 1},
  [25] = {.index = 11, .length = 1},
  [26] = {.index = 12, .length = 1},
  [31] = {.index = 13, .length = 1},
  [32] = {.index = 14, .length = 2},
  [33] = {.index = 16, .length = 1},
  [34] = {.index = 17, .length = 2},
  [35] = {.index = 19, .length = 1},
  [36] = {.index = 20, .length = 2},
  [39] = {.index = 20, .length = 2},
  [42] = {.index = 22, .length = 1},
  [43] = {.index = 23, .length = 1},
  [44] = {.index = 24, .length = 1},
  [45] = {.index = 25, .length = 1},
  [46] = {.index = 26, .length = 1},
  [47] = {.index = 27, .length = 2},
  [48] = {.index = 29, .length = 1},
  [49] = {.index = 30, .length = 1},
  [50] = {.index = 22, .length = 1},
  [51] = {.index = 31, .length = 1},
  [53] = {.index = 32, .length = 1},
  [54] = {.index = 30, .length = 1},
  [55] = {.index = 33, .length = 1},
  [56] = {.index = 34, .length = 1},
  [57] = {.index = 35, .length = 1},
  [58] = {.index = 36, .length = 2},
  [59] = {.index = 38, .length = 1},
  [61] = {.index = 39, .length = 1},
  [62] = {.index = 40, .length = 1},
  [63] = {.index = 41, .length = 1},
  [64] = {.index = 22, .length = 1},
  [65] = {.index = 42, .length = 2},
  [66] = {.index = 44, .length = 2},
  [67] = {.index = 46, .length = 3},
  [68] = {.index = 49, .length = 2},
  [69] = {.index = 51, .length = 2},
  [70] = {.index = 53, .length = 2},
  [71] = {.index = 55, .length = 2},
  [72] = {.index = 55, .length = 2},
  [73] = {.index = 53, .length = 2},
  [74] = {.index = 57, .length = 1},
  [75] = {.index = 58, .length = 2},
  [76] = {.index = 60, .length = 1},
  [77] = {.index = 61, .length = 1},
  [78] = {.index = 62, .length = 2},
  [79] = {.index = 60, .length = 1},
  [80] = {.index = 23, .length = 1},
  [81] = {.index = 64, .length = 1},
  [82] = {.index = 22, .length = 1},
  [84] = {.index = 60, .length = 1},
  [85] = {.index = 65, .length = 1},
  [86] = {.index = 66, .length = 2},
  [87] = {.index = 68, .length = 1},
  [88] = {.index = 69, .length = 2},
  [89] = {.index = 71, .length = 1},
  [90] = {.index = 60, .length = 1},
  [91] = {.index = 72, .length = 2},
  [92] = {.index = 74, .length = 1},
  [93] = {.index = 75, .length = 2},
  [94] = {.index = 77, .length = 1},
  [95] = {.index = 78, .length = 2},
  [96] = {.index = 80, .length = 3},
  [97] = {.index = 68, .length = 1},
  [98] = {.index = 83, .length = 1},
  [99] = {.index = 84, .length = 2},
  [100] = {.index = 83, .length = 1},
  [102] = {.index = 86, .length = 2},
  [103] = {.index = 88, .length = 1},
  [105] = {.index = 12, .length = 1},
  [106] = {.index = 89, .length = 1},
  [108] = {.index = 83, .length = 1},
  [109] = {.index = 68, .length = 1},
  [110] = {.index = 90, .length = 3},
  [111] = {.index = 93, .length = 2},
  [112] = {.index = 95, .length = 2},
  [113] = {.index = 95, .length = 2},
  [114] = {.index = 97, .length = 1},
  [115] = {.index = 98, .length = 2},
  [116] = {.index = 100, .length = 2},
  [117] = {.index = 60, .length = 1},
  [118] = {.index = 102, .length = 1},
  [120] = {.index = 103, .length = 2},
  [121] = {.index = 57, .length = 1},
  [122] = {.index = 105, .length = 1},
  [123] = {.index = 106, .length = 2},
  [125] = {.index = 103, .length = 2},
  [126] = {.index = 108, .length = 3},
  [127] = {.index = 111, .length = 3},
  [128] = {.index = 114, .length = 2},
  [129] = {.index = 57, .length = 1},
  [130] = {.index = 116, .length = 2},
  [131] = {.index = 118, .length = 2},
  [132] = {.index = 120, .length = 2},
  [133] = {.index = 122, .length = 1},
  [134] = {.index = 123, .length = 1},
  [135] = {.index = 124, .length = 3},
  [136] = {.index = 127, .length = 1},
  [137] = {.index = 116, .length = 2},
  [138] = {.index = 128, .length = 2},
  [139] = {.index = 33, .length = 1},
  [141] = {.index = 60, .length = 1},
  [142] = {.index = 33, .length = 1},
  [143] = {.index = 130, .length = 2},
  [144] = {.index = 132, .length = 2},
  [145] = {.index = 134, .length = 2},
  [146] = {.index = 12, .length = 1},
  [147] = {.index = 136, .length = 3},
  [148] = {.index = 139, .length = 3},
  [149] = {.index = 142, .length = 2},
  [150] = {.index = 144, .length = 1},
  [151] = {.index = 145, .length = 1},
  [152] = {.index = 97, .length = 1},
  [153] = {.index = 146, .length = 1},
  [154] = {.index = 146, .length = 1},
  [156] = {.index = 147, .length = 2},
  [157] = {.index = 149, .length = 2},
  [158] = {.index = 151, .length = 2},
  [159] = {.index = 153, .length = 1},
  [161] = {.index = 154, .length = 2},
  [162] = {.index = 33, .length = 1},
  [163] = {.index = 156, .length = 4},
  [164] = {.index = 160, .length = 3},
  [165] = {.index = 163, .length = 3},
  [166] = {.index = 166, .length = 2},
  [167] = {.index = 168, .length = 1},
  [168] = {.index = 169, .length = 4},
  [169] = {.index = 173, .length = 3},
  [170] = {.index = 176, .length = 2},
  [171] = {.index = 178, .length = 2},
  [172] = {.index = 180, .length = 3},
  [173] = {.index = 183, .length = 2},
  [174] = {.index = 57, .length = 1},
  [175] = {.index = 185, .length = 1},
  [176] = {.index = 185, .length = 1},
  [177] = {.index = 130, .length = 2},
  [178] = {.index = 57, .length = 1},
  [179] = {.index = 186, .length = 1},
  [180] = {.index = 187, .length = 2},
  [181] = {.index = 189, .length = 3},
  [182] = {.index = 192, .length = 3},
  [183] = {.index = 195, .length = 3},
  [184] = {.index = 198, .length = 3},
  [185] = {.index = 201, .length = 1},
  [186] = {.index = 202, .length = 1},
  [187] = {.index = 97, .length = 1},
  [188] = {.index = 203, .length = 2},
  [189] = {.index = 205, .length = 1},
  [190] = {.index = 205, .length = 1},
  [191] = {.index = 206, .length = 1},
  [192] = {.index = 206, .length = 1},
  [193] = {.index = 207, .length = 2},
  [194] = {.index = 209, .length = 1},
  [195] = {.index = 210, .length = 1},
  [196] = {.index = 211, .length = 4},
  [197] = {.index = 215, .length = 3},
  [198] = {.index = 218, .length = 2},
  [199] = {.index = 220, .length = 3},
  [200] = {.index = 97, .length = 1},
  [201] = {.index = 187, .length = 2},
  [202] = {.index = 97, .length = 1},
  [203] = {.index = 97, .length = 1},
  [204] = {.index = 223, .length = 1},
  [205] = {.index = 187, .length = 2},
  [206] = {.index = 224, .length = 2},
  [207] = {.index = 226, .length = 3},
  [208] = {.index = 229, .length = 1},
  [209] = {.index = 205, .length = 1},
  [210] = {.index = 230, .length = 2},
  [211] = {.index = 230, .length = 2},
  [212] = {.index = 232, .length = 1},
  [213] = {.index = 232, .length = 1},
  [214] = {.index = 233, .length = 1},
  [215] = {.index = 234, .length = 3},
  [216] = {.index = 237, .length = 2},
  [217] = {.index = 239, .length = 2},
  [218] = {.index = 241, .length = 3},
  [219] = {.index = 244, .length = 2},
  [220] = {.index = 244, .length = 2},
  [221] = {.index = 205, .length = 1},
  [222] = {.index = 246, .length = 3},
  [223] = {.index = 232, .length = 1},
  [224] = {.index = 249, .length = 2},
  [225] = {.index = 249, .length = 2},
  [226] = {.index = 251, .length = 1},
  [227] = {.index = 252, .length = 2},
  [228] = {.index = 254, .length = 2},
  [229] = {.index = 256, .length = 2},
  [230] = {.index = 258, .length = 3},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_name, 0},
  [1] =
    {field_module, 1},
  [2] =
    {field_module_type, 1},
  [3] =
    {field_name, 1},
  [4] =
    {field_module_type, 1, .inherited = true},
  [5] =
    {field_module, 2},
  [6] =
    {field_argument, 1},
    {field_functor, 0},
  [8] =
    {field_codomain, 1},
  [9] =
    {field_module_type, 2},
  [10] =
    {field_class_type, 1, .inherited = true},
  [11] =
    {field_pattern, 0},
  [12] =
    {field_type, 2, .inherited = true},
  [13] =
    {field_expression, 1},
  [14] =
    {field_expression, 1},
    {field_operator, 0},
  [16] =
    {field_argument, 0},
  [17] =
    {field_argument, 1, .inherited = true},
    {field_function, 0},
  [19] =
    {field_type, 1},
  [20] =
    {field_body, 2},
    {field_name, 0},
  [22] =
    {field_body, 2},
  [23] =
    {field_module_type, 2, .inherited = true},
  [24] =
    {field_module, 3},
  [25] =
    {field_functor, 0},
  [26] =
    {field_codomain, 2},
  [27] =
    {field_codomain, 2},
    {field_domain, 0},
  [29] =
    {field_module_type, 0},
  [30] =
    {field_class_type, 2, .inherited = true},
  [31] =
    {field_pattern, 1},
  [32] =
    {field_class_type, 1},
  [33] =
    {field_type, 3, .inherited = true},
  [34] =
    {field_type, 0, .inherited = true},
  [35] =
    {field_coercion, 0, .inherited = true},
  [36] =
    {field_coercion, 1, .inherited = true},
    {field_type, 1, .inherited = true},
  [38] =
    {field_class, 1},
  [39] =
    {field_initializer, 1},
  [40] =
    {field_condition, 1},
  [41] =
    {field_expression, 2},
  [42] =
    {field_field, 2},
    {field_record, 0},
  [44] =
    {field_method, 2},
    {field_object, 0},
  [46] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [49] =
    {field_argument, 0, .inherited = true},
    {field_argument, 1, .inherited = true},
  [51] =
    {field_left, 0},
    {field_right, 2},
  [53] =
    {field_body, 3},
    {field_name, 0},
  [55] =
    {field_body, 3},
    {field_name, 1},
  [57] =
    {field_body, 4},
  [58] =
    {field_body, 3},
    {field_module_type, 1, .inherited = true},
  [60] =
    {field_body, 3},
  [61] =
    {field_value, 2},
  [62] =
    {field_module, 1},
    {field_module_type, 2, .inherited = true},
  [64] =
    {field_codomain, 3},
  [65] =
    {field_class_type, 3, .inherited = true},
  [66] =
    {field_argument, 1, .inherited = true},
    {field_class, 0},
  [68] =
    {field_type, 1, .inherited = true},
  [69] =
    {field_body, 3},
    {field_class_type, 1, .inherited = true},
  [71] =
    {field_pattern, 2},
  [72] =
    {field_body, 2},
    {field_pattern, 0},
  [74] =
    {field_coercion, 1},
  [75] =
    {field_alias, 2},
    {field_pattern, 0},
  [77] =
    {field_expression, 3},
  [78] =
    {field_coercion, 1, .inherited = true},
    {field_type, 0, .inherited = true},
  [80] =
    {field_coercion, 2, .inherited = true},
    {field_expression, 1},
    {field_type, 2, .inherited = true},
  [83] =
    {field_type, 2},
  [84] =
    {field_alias, 2},
    {field_type, 0},
  [86] =
    {field_coercion, 2, .inherited = true},
    {field_type, 2, .inherited = true},
  [88] =
    {field_class, 2},
  [89] =
    {field_condition, 2},
  [90] =
    {field_body, 4},
    {field_name, 0},
    {field_synonym, 2},
  [93] =
    {field_constraint, 3},
    {field_type, 1},
  [95] =
    {field_body, 4},
    {field_name, 1},
  [97] =
    {field_body, 5},
  [98] =
    {field_body, 4},
    {field_module_type, 2, .inherited = true},
  [100] =
    {field_module_type, 3, .inherited = true},
    {field_value, 2},
  [102] =
    {field_class_type, 4, .inherited = true},
  [103] =
    {field_body, 4},
    {field_class_type, 2, .inherited = true},
  [105] =
    {field_pattern, 3},
  [106] =
    {field_pattern, 1},
    {field_type, 2, .inherited = true},
  [108] =
    {field_body, 3},
    {field_coercion, 1, .inherited = true},
    {field_pattern, 0},
  [111] =
    {field_body, 3},
    {field_pattern, 0},
    {field_type, 1, .inherited = true},
  [114] =
    {field_body, 3},
    {field_pattern, 0},
  [116] =
    {field_coercion, 3, .inherited = true},
    {field_type, 3, .inherited = true},
  [118] =
    {field_continuation, 3},
    {field_effect, 1},
  [120] =
    {field_module, 2},
    {field_module_type, 3, .inherited = true},
  [122] =
    {field_type, 3},
  [123] =
    {field_record, 1},
  [124] =
    {field_body, 3},
    {field_coercion, 1, .inherited = true},
    {field_type, 1, .inherited = true},
  [127] =
    {field_self, 2},
  [128] =
    {field_alias, 3},
    {field_class, 1},
  [130] =
    {field_body, 4},
    {field_type, 2, .inherited = true},
  [132] =
    {field_array, 0},
    {field_index, 3},
  [134] =
    {field_index, 3},
    {field_string, 0},
  [136] =
    {field_body, 5},
    {field_name, 0},
    {field_synonym, 2},
  [139] =
    {field_body, 5},
    {field_name, 1},
    {field_synonym, 3},
  [142] =
    {field_coercion, 4},
    {field_value, 2},
  [144] =
    {field_equation, 3},
  [145] =
    {field_constraint, 3},
  [146] =
    {field_class_type, 5, .inherited = true},
  [147] =
    {field_body, 5},
    {field_class_type, 3, .inherited = true},
  [149] =
    {field_class, 1},
    {field_class_type, 2, .inherited = true},
  [151] =
    {field_pattern, 2},
    {field_type, 3, .inherited = true},
  [153] =
    {field_module_type, 3, .inherited = true},
  [154] =
    {field_pattern, 3},
    {field_type, 1, .inherited = true},
  [156] =
    {field_body, 4},
    {field_coercion, 2, .inherited = true},
    {field_pattern, 0},
    {field_type, 1, .inherited = true},
  [160] =
    {field_body, 4},
    {field_coercion, 2, .inherited = true},
    {field_pattern, 0},
  [163] =
    {field_body, 4},
    {field_pattern, 0},
    {field_type, 2, .inherited = true},
  [166] =
    {field_module, 3},
    {field_module_type, 4, .inherited = true},
  [168] =
    {field_module_type, 3},
  [169] =
    {field_codomain, 4},
    {field_domain, 0},
    {field_domain, 1},
    {field_domain, 2},
  [173] =
    {field_type, 2},
    {field_type, 3},
    {field_type, 4},
  [176] =
    {field_self, 2},
    {field_type, 3, .inherited = true},
  [178] =
    {field_coercion, 4, .inherited = true},
    {field_type, 4, .inherited = true},
  [180] =
    {field_body, 4},
    {field_coercion, 2, .inherited = true},
    {field_type, 2, .inherited = true},
  [183] =
    {field_alias, 4},
    {field_class, 2},
  [185] =
    {field_type, 4, .inherited = true},
  [186] =
    {field_self, 3},
  [187] =
    {field_body, 5},
    {field_type, 3, .inherited = true},
  [189] =
    {field_array, 0},
    {field_index, 4},
    {field_operator, 2},
  [192] =
    {field_index, 4},
    {field_operator, 2},
    {field_string, 0},
  [195] =
    {field_body, 6},
    {field_name, 1},
    {field_synonym, 3},
  [198] =
    {field_coercion, 5},
    {field_module_type, 3, .inherited = true},
    {field_value, 2},
  [201] =
    {field_equation, 4},
  [202] =
    {field_constraint, 4},
  [203] =
    {field_body, 6},
    {field_class_type, 4, .inherited = true},
  [205] =
    {field_body, 6},
  [206] =
    {field_class_type, 6, .inherited = true},
  [207] =
    {field_default, 4},
    {field_pattern, 2},
  [209] =
    {field_module_type, 4, .inherited = true},
  [210] =
    {field_self_type, 2},
  [211] =
    {field_body, 5},
    {field_coercion, 3, .inherited = true},
    {field_pattern, 0},
    {field_type, 2, .inherited = true},
  [215] =
    {field_type, 3},
    {field_type, 4},
    {field_type, 5},
  [218] =
    {field_module, 2},
    {field_module_type, 4},
  [220] =
    {field_body, 5},
    {field_coercion, 3, .inherited = true},
    {field_type, 3, .inherited = true},
  [223] =
    {field_type, 5, .inherited = true},
  [224] =
    {field_self, 3},
    {field_type, 4, .inherited = true},
  [226] =
    {field_from, 3},
    {field_name, 1},
    {field_to, 5},
  [229] =
    {field_equation, 5},
  [230] =
    {field_body, 7},
    {field_class_type, 5, .inherited = true},
  [232] =
    {field_body, 7},
  [233] =
    {field_class_type, 7, .inherited = true},
  [234] =
    {field_default, 5},
    {field_pattern, 2},
    {field_type, 3, .inherited = true},
  [237] =
    {field_pattern, 4},
    {field_type, 5, .inherited = true},
  [239] =
    {field_module, 3},
    {field_module_type, 5},
  [241] =
    {field_body, 6},
    {field_coercion, 4, .inherited = true},
    {field_type, 4, .inherited = true},
  [244] =
    {field_body, 6},
    {field_type, 4, .inherited = true},
  [246] =
    {field_from, 4},
    {field_name, 2},
    {field_to, 6},
  [249] =
    {field_body, 8},
    {field_class_type, 6, .inherited = true},
  [251] =
    {field_body, 8},
  [252] =
    {field_default, 6},
    {field_pattern, 4},
  [254] =
    {field_body, 7},
    {field_type, 5, .inherited = true},
  [256] =
    {field_body, 9},
    {field_class_type, 7, .inherited = true},
  [258] =
    {field_default, 7},
    {field_pattern, 4},
    {field_type, 5, .inherited = true},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = alias_sym_value_name,
  },
  [2] = {
    [0] = sym_type_variable,
  },
  [3] = {
    [0] = alias_sym_type_constructor,
  },
  [4] = {
    [0] = alias_sym_type_constructor,
  },
  [5] = {
    [0] = alias_sym_module_name,
  },
  [6] = {
    [0] = sym__constructor_declaration_name,
  },
  [8] = {
    [0] = alias_sym_module_type_name,
  },
  [10] = {
    [0] = alias_sym_class_name,
  },
  [11] = {
    [0] = alias_sym_instance_variable_name,
  },
  [12] = {
    [0] = alias_sym_tuple_expression,
  },
  [13] = {
    [0] = alias_sym_sequence_expression,
  },
  [14] = {
    [1] = alias_sym_type_constructor,
  },
  [15] = {
    [1] = sym_type_variable,
  },
  [16] = {
    [2] = alias_sym_module_type_name,
  },
  [17] = {
    [0] = alias_sym_module_name,
  },
  [22] = {
    [1] = alias_sym_class_name,
  },
  [23] = {
    [0] = alias_sym_value_pattern,
  },
  [24] = {
    [0] = alias_sym_class_name,
  },
  [27] = {
    [1] = alias_sym_label_name,
  },
  [28] = {
    [0] = sym__or_binding_pattern_no_exn_anonymous,
  },
  [29] = {
    [0] = sym__tuple_binding_pattern_no_exn,
  },
  [30] = {
    [0] = alias_sym_field_name,
  },
  [36] = {
    [0] = alias_sym_type_constructor,
  },
  [37] = {
    [2] = alias_sym_type_constructor,
  },
  [38] = {
    [2] = alias_sym_module_name,
  },
  [40] = {
    [1] = sym__constructor_declaration_name,
  },
  [41] = {
    [3] = alias_sym_module_type_name,
  },
  [42] = {
    [0] = alias_sym_module_name,
  },
  [43] = {
    [0] = alias_sym_module_name,
  },
  [49] = {
    [1] = alias_sym_class_name,
  },
  [50] = {
    [0] = alias_sym_class_name,
  },
  [52] = {
    [0] = alias_sym_class_type_name,
  },
  [54] = {
    [0] = alias_sym_class_name,
  },
  [60] = {
    [1] = alias_sym_method_name,
  },
  [66] = {
    [2] = alias_sym_method_name,
  },
  [70] = {
    [0] = alias_sym_type_constructor,
  },
  [71] = {
    [1] = alias_sym_type_constructor,
  },
  [74] = {
    [2] = alias_sym_module_type_name,
  },
  [75] = {
    [0] = alias_sym_module_name,
  },
  [76] = {
    [0] = alias_sym_module_name,
  },
  [80] = {
    [1] = alias_sym_module_name,
  },
  [82] = {
    [0] = alias_sym_class_type_name,
  },
  [83] = {
    [3] = alias_sym_class_name,
  },
  [84] = {
    [1] = alias_sym_class_name,
  },
  [85] = {
    [1] = alias_sym_class_name,
  },
  [88] = {
    [0] = alias_sym_class_name,
  },
  [90] = {
    [0] = alias_sym_class_name,
  },
  [94] = {
    [1] = alias_sym_label_name,
  },
  [97] = {
    [0] = alias_sym_method_name,
  },
  [98] = {
    [0] = alias_sym_label_name,
  },
  [101] = {
    [2] = alias_sym_field_name,
  },
  [104] = {
    [2] = alias_sym_method_name,
  },
  [105] = {
    [1] = alias_sym_method_name,
  },
  [107] = {
    [2] = alias_sym_class_name,
  },
  [108] = {
    [0] = sym__parenthesized_abstract_type,
  },
  [109] = {
    [0] = alias_sym_field_name,
  },
  [110] = {
    [0] = alias_sym_type_constructor,
  },
  [112] = {
    [1] = alias_sym_type_constructor,
  },
  [114] = {
    [3] = alias_sym_module_type_name,
  },
  [115] = {
    [0] = alias_sym_module_name,
  },
  [117] = {
    [1] = alias_sym_class_type_name,
  },
  [118] = {
    [3] = alias_sym_class_name,
  },
  [119] = {
    [4] = alias_sym_class_name,
  },
  [120] = {
    [1] = alias_sym_class_name,
  },
  [121] = {
    [1] = alias_sym_class_name,
  },
  [122] = {
    [1] = alias_sym_label_name,
  },
  [124] = {
    [2] = alias_sym_class_type_name,
  },
  [125] = {
    [0] = alias_sym_class_name,
  },
  [130] = {
    [2] = alias_sym_label_name,
  },
  [133] = {
    [1] = alias_sym_label_name,
  },
  [139] = {
    [2] = alias_sym_method_name,
  },
  [140] = {
    [3] = alias_sym_method_name,
  },
  [141] = {
    [1] = alias_sym_method_name,
  },
  [142] = {
    [1] = alias_sym_method_name,
  },
  [146] = {
    [1] = alias_sym_field_name,
  },
  [147] = {
    [0] = alias_sym_type_constructor,
  },
  [148] = {
    [1] = alias_sym_type_constructor,
  },
  [152] = {
    [3] = alias_sym_class_name,
  },
  [153] = {
    [3] = alias_sym_class_name,
  },
  [154] = {
    [4] = alias_sym_class_name,
  },
  [155] = {
    [5] = alias_sym_class_name,
  },
  [156] = {
    [1] = alias_sym_class_name,
  },
  [159] = {
    [2] = alias_sym_module_name,
  },
  [160] = {
    [3] = alias_sym_module_name,
  },
  [162] = {
    [2] = alias_sym_label_name,
  },
  [169] = {
    [0] = alias_sym_label_name,
  },
  [174] = {
    [2] = alias_sym_method_name,
  },
  [175] = {
    [2] = alias_sym_method_name,
  },
  [176] = {
    [3] = alias_sym_method_name,
  },
  [177] = {
    [1] = alias_sym_method_name,
  },
  [178] = {
    [1] = alias_sym_method_name,
  },
  [183] = {
    [1] = alias_sym_type_constructor,
  },
  [187] = {
    [3] = alias_sym_class_type_name,
  },
  [188] = {
    [3] = alias_sym_class_name,
  },
  [189] = {
    [3] = alias_sym_class_name,
  },
  [190] = {
    [4] = alias_sym_class_name,
  },
  [191] = {
    [4] = alias_sym_class_name,
  },
  [192] = {
    [5] = alias_sym_class_name,
  },
  [194] = {
    [3] = alias_sym_module_name,
  },
  [197] = {
    [1] = alias_sym_label_name,
  },
  [198] = {
    [2] = alias_sym_module_name,
  },
  [201] = {
    [2] = alias_sym_method_name,
  },
  [202] = {
    [2] = alias_sym_method_name,
  },
  [203] = {
    [3] = alias_sym_method_name,
  },
  [204] = {
    [3] = alias_sym_method_name,
  },
  [205] = {
    [1] = alias_sym_method_name,
  },
  [209] = {
    [4] = alias_sym_class_type_name,
  },
  [210] = {
    [3] = alias_sym_class_name,
  },
  [211] = {
    [4] = alias_sym_class_name,
  },
  [212] = {
    [4] = alias_sym_class_name,
  },
  [213] = {
    [5] = alias_sym_class_name,
  },
  [214] = {
    [5] = alias_sym_class_name,
  },
  [216] = {
    [1] = alias_sym_label_name,
  },
  [217] = {
    [3] = alias_sym_module_name,
  },
  [219] = {
    [2] = alias_sym_method_name,
  },
  [220] = {
    [3] = alias_sym_method_name,
  },
  [221] = {
    [3] = alias_sym_method_name,
  },
  [223] = {
    [5] = alias_sym_class_type_name,
  },
  [224] = {
    [4] = alias_sym_class_name,
  },
  [225] = {
    [5] = alias_sym_class_name,
  },
  [226] = {
    [5] = alias_sym_class_name,
  },
  [227] = {
    [1] = alias_sym_label_name,
  },
  [228] = {
    [3] = alias_sym_method_name,
  },
  [229] = {
    [5] = alias_sym_class_name,
  },
  [230] = {
    [1] = alias_sym_label_name,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym__abstract_type, 2,
    sym__abstract_type,
    sym__parenthesized_abstract_type,
  sym__tuple_expression, 2,
    sym__tuple_expression,
    alias_sym_tuple_expression,
  sym__sequence_expression_anonymous, 2,
    sym__sequence_expression_anonymous,
    alias_sym_sequence_expression,
  sym__or_pattern_anonymous, 2,
    sym__or_pattern_anonymous,
    sym__or_binding_pattern_no_exn_anonymous,
  sym__or_binding_pattern_anonymous, 2,
    sym__or_binding_pattern_anonymous,
    sym__or_binding_pattern_no_exn_anonymous,
  sym__tuple_pattern, 2,
    sym__tuple_pattern,
    sym__tuple_binding_pattern_no_exn,
  sym__tuple_binding_pattern, 2,
    sym__tuple_binding_pattern,
    sym__tuple_binding_pattern_no_exn,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 2,
  [4] = 2,
  [5] = 2,
  [6] = 6,
  [7] = 6,
  [8] = 8,
  [9] = 6,
  [10] = 10,
  [11] = 11,
  [12] = 6,
  [13] = 8,
  [14] = 10,
  [15] = 6,
  [16] = 6,
  [17] = 8,
  [18] = 6,
  [19] = 6,
  [20] = 6,
  [21] = 6,
  [22] = 22,
  [23] = 6,
  [24] = 6,
  [25] = 6,
  [26] = 6,
  [27] = 11,
  [28] = 6,
  [29] = 6,
  [30] = 6,
  [31] = 6,
  [32] = 6,
  [33] = 6,
  [34] = 8,
  [35] = 35,
  [36] = 36,
  [37] = 37,
  [38] = 38,
  [39] = 38,
  [40] = 40,
  [41] = 2,
  [42] = 2,
  [43] = 43,
  [44] = 44,
  [45] = 44,
  [46] = 46,
  [47] = 44,
  [48] = 46,
  [49] = 46,
  [50] = 44,
  [51] = 46,
  [52] = 44,
  [53] = 44,
  [54] = 46,
  [55] = 44,
  [56] = 46,
  [57] = 44,
  [58] = 46,
  [59] = 44,
  [60] = 46,
  [61] = 44,
  [62] = 46,
  [63] = 44,
  [64] = 46,
  [65] = 46,
  [66] = 2,
  [67] = 2,
  [68] = 2,
  [69] = 2,
  [70] = 2,
  [71] = 2,
  [72] = 2,
  [73] = 2,
  [74] = 2,
  [75] = 75,
  [76] = 2,
  [77] = 75,
  [78] = 75,
  [79] = 79,
  [80] = 80,
  [81] = 79,
  [82] = 80,
  [83] = 80,
  [84] = 79,
  [85] = 2,
  [86] = 75,
  [87] = 2,
  [88] = 80,
  [89] = 79,
  [90] = 2,
  [91] = 80,
  [92] = 2,
  [93] = 93,
  [94] = 94,
  [95] = 94,
  [96] = 94,
  [97] = 97,
  [98] = 94,
  [99] = 97,
  [100] = 94,
  [101] = 94,
  [102] = 94,
  [103] = 103,
  [104] = 93,
  [105] = 94,
  [106] = 97,
  [107] = 94,
  [108] = 103,
  [109] = 93,
  [110] = 93,
  [111] = 93,
  [112] = 94,
  [113] = 94,
  [114] = 114,
  [115] = 115,
  [116] = 116,
  [117] = 115,
  [118] = 116,
  [119] = 115,
  [120] = 115,
  [121] = 116,
  [122] = 115,
  [123] = 116,
  [124] = 115,
  [125] = 125,
  [126] = 2,
  [127] = 116,
  [128] = 128,
  [129] = 114,
  [130] = 116,
  [131] = 115,
  [132] = 115,
  [133] = 133,
  [134] = 125,
  [135] = 116,
  [136] = 133,
  [137] = 125,
  [138] = 114,
  [139] = 125,
  [140] = 133,
  [141] = 133,
  [142] = 133,
  [143] = 133,
  [144] = 133,
  [145] = 133,
  [146] = 133,
  [147] = 133,
  [148] = 133,
  [149] = 128,
  [150] = 114,
  [151] = 2,
  [152] = 116,
  [153] = 133,
  [154] = 128,
  [155] = 114,
  [156] = 128,
  [157] = 114,
  [158] = 128,
  [159] = 114,
  [160] = 128,
  [161] = 133,
  [162] = 128,
  [163] = 114,
  [164] = 128,
  [165] = 114,
  [166] = 128,
  [167] = 114,
  [168] = 128,
  [169] = 133,
  [170] = 170,
  [171] = 170,
  [172] = 172,
  [173] = 173,
  [174] = 170,
  [175] = 175,
  [176] = 176,
  [177] = 172,
  [178] = 170,
  [179] = 173,
  [180] = 176,
  [181] = 173,
  [182] = 170,
  [183] = 172,
  [184] = 176,
  [185] = 173,
  [186] = 176,
  [187] = 173,
  [188] = 188,
  [189] = 172,
  [190] = 176,
  [191] = 176,
  [192] = 176,
  [193] = 173,
  [194] = 173,
  [195] = 175,
  [196] = 170,
  [197] = 176,
  [198] = 188,
  [199] = 176,
  [200] = 170,
  [201] = 176,
  [202] = 173,
  [203] = 173,
  [204] = 173,
  [205] = 172,
  [206] = 170,
  [207] = 172,
  [208] = 188,
  [209] = 173,
  [210] = 176,
  [211] = 188,
  [212] = 212,
  [213] = 213,
  [214] = 214,
  [215] = 215,
  [216] = 216,
  [217] = 217,
  [218] = 218,
  [219] = 219,
  [220] = 220,
  [221] = 221,
  [222] = 222,
  [223] = 214,
  [224] = 215,
  [225] = 216,
  [226] = 218,
  [227] = 227,
  [228] = 228,
  [229] = 229,
  [230] = 230,
  [231] = 212,
  [232] = 232,
  [233] = 233,
  [234] = 234,
  [235] = 235,
  [236] = 236,
  [237] = 237,
  [238] = 238,
  [239] = 239,
  [240] = 219,
  [241] = 222,
  [242] = 219,
  [243] = 221,
  [244] = 221,
  [245] = 222,
  [246] = 246,
  [247] = 222,
  [248] = 214,
  [249] = 215,
  [250] = 216,
  [251] = 218,
  [252] = 219,
  [253] = 253,
  [254] = 221,
  [255] = 222,
  [256] = 214,
  [257] = 215,
  [258] = 216,
  [259] = 218,
  [260] = 260,
  [261] = 212,
  [262] = 232,
  [263] = 212,
  [264] = 264,
  [265] = 265,
  [266] = 238,
  [267] = 267,
  [268] = 214,
  [269] = 215,
  [270] = 216,
  [271] = 271,
  [272] = 272,
  [273] = 273,
  [274] = 212,
  [275] = 232,
  [276] = 276,
  [277] = 218,
  [278] = 278,
  [279] = 279,
  [280] = 238,
  [281] = 219,
  [282] = 282,
  [283] = 283,
  [284] = 212,
  [285] = 232,
  [286] = 221,
  [287] = 222,
  [288] = 219,
  [289] = 238,
  [290] = 214,
  [291] = 215,
  [292] = 212,
  [293] = 232,
  [294] = 216,
  [295] = 218,
  [296] = 238,
  [297] = 219,
  [298] = 232,
  [299] = 221,
  [300] = 212,
  [301] = 232,
  [302] = 222,
  [303] = 303,
  [304] = 238,
  [305] = 214,
  [306] = 215,
  [307] = 216,
  [308] = 232,
  [309] = 264,
  [310] = 218,
  [311] = 265,
  [312] = 219,
  [313] = 267,
  [314] = 214,
  [315] = 212,
  [316] = 232,
  [317] = 221,
  [318] = 222,
  [319] = 214,
  [320] = 215,
  [321] = 216,
  [322] = 218,
  [323] = 323,
  [324] = 324,
  [325] = 215,
  [326] = 216,
  [327] = 327,
  [328] = 221,
  [329] = 246,
  [330] = 330,
  [331] = 331,
  [332] = 332,
  [333] = 212,
  [334] = 232,
  [335] = 335,
  [336] = 336,
  [337] = 335,
  [338] = 338,
  [339] = 279,
  [340] = 282,
  [341] = 283,
  [342] = 330,
  [343] = 331,
  [344] = 332,
  [345] = 336,
  [346] = 346,
  [347] = 347,
  [348] = 228,
  [349] = 229,
  [350] = 236,
  [351] = 219,
  [352] = 272,
  [353] = 221,
  [354] = 222,
  [355] = 273,
  [356] = 214,
  [357] = 215,
  [358] = 216,
  [359] = 218,
  [360] = 276,
  [361] = 238,
  [362] = 362,
  [363] = 222,
  [364] = 238,
  [365] = 246,
  [366] = 366,
  [367] = 238,
  [368] = 238,
  [369] = 238,
  [370] = 338,
  [371] = 371,
  [372] = 372,
  [373] = 335,
  [374] = 279,
  [375] = 282,
  [376] = 283,
  [377] = 330,
  [378] = 331,
  [379] = 332,
  [380] = 346,
  [381] = 228,
  [382] = 238,
  [383] = 264,
  [384] = 265,
  [385] = 335,
  [386] = 279,
  [387] = 282,
  [388] = 283,
  [389] = 330,
  [390] = 331,
  [391] = 332,
  [392] = 346,
  [393] = 228,
  [394] = 238,
  [395] = 267,
  [396] = 279,
  [397] = 282,
  [398] = 283,
  [399] = 330,
  [400] = 331,
  [401] = 332,
  [402] = 218,
  [403] = 279,
  [404] = 282,
  [405] = 283,
  [406] = 330,
  [407] = 331,
  [408] = 332,
  [409] = 272,
  [410] = 279,
  [411] = 282,
  [412] = 283,
  [413] = 330,
  [414] = 331,
  [415] = 332,
  [416] = 273,
  [417] = 279,
  [418] = 282,
  [419] = 283,
  [420] = 330,
  [421] = 331,
  [422] = 332,
  [423] = 276,
  [424] = 279,
  [425] = 282,
  [426] = 283,
  [427] = 330,
  [428] = 331,
  [429] = 332,
  [430] = 278,
  [431] = 279,
  [432] = 282,
  [433] = 283,
  [434] = 330,
  [435] = 331,
  [436] = 332,
  [437] = 219,
  [438] = 346,
  [439] = 327,
  [440] = 221,
  [441] = 222,
  [442] = 271,
  [443] = 239,
  [444] = 253,
  [445] = 219,
  [446] = 347,
  [447] = 221,
  [448] = 222,
  [449] = 214,
  [450] = 215,
  [451] = 216,
  [452] = 218,
  [453] = 238,
  [454] = 278,
  [455] = 214,
  [456] = 215,
  [457] = 271,
  [458] = 239,
  [459] = 253,
  [460] = 216,
  [461] = 461,
  [462] = 271,
  [463] = 239,
  [464] = 253,
  [465] = 218,
  [466] = 219,
  [467] = 271,
  [468] = 239,
  [469] = 253,
  [470] = 221,
  [471] = 222,
  [472] = 271,
  [473] = 239,
  [474] = 253,
  [475] = 214,
  [476] = 271,
  [477] = 239,
  [478] = 253,
  [479] = 215,
  [480] = 216,
  [481] = 271,
  [482] = 239,
  [483] = 253,
  [484] = 218,
  [485] = 219,
  [486] = 271,
  [487] = 239,
  [488] = 253,
  [489] = 489,
  [490] = 221,
  [491] = 271,
  [492] = 239,
  [493] = 253,
  [494] = 260,
  [495] = 327,
  [496] = 260,
  [497] = 327,
  [498] = 260,
  [499] = 499,
  [500] = 500,
  [501] = 500,
  [502] = 500,
  [503] = 500,
  [504] = 500,
  [505] = 500,
  [506] = 506,
  [507] = 500,
  [508] = 500,
  [509] = 500,
  [510] = 500,
  [511] = 500,
  [512] = 500,
  [513] = 500,
  [514] = 500,
  [515] = 500,
  [516] = 500,
  [517] = 517,
  [518] = 500,
  [519] = 506,
  [520] = 500,
  [521] = 500,
  [522] = 500,
  [523] = 517,
  [524] = 500,
  [525] = 500,
  [526] = 526,
  [527] = 527,
  [528] = 528,
  [529] = 529,
  [530] = 530,
  [531] = 531,
  [532] = 532,
  [533] = 533,
  [534] = 526,
  [535] = 535,
  [536] = 536,
  [537] = 537,
  [538] = 538,
  [539] = 539,
  [540] = 527,
  [541] = 528,
  [542] = 529,
  [543] = 530,
  [544] = 532,
  [545] = 533,
  [546] = 535,
  [547] = 536,
  [548] = 537,
  [549] = 549,
  [550] = 550,
  [551] = 551,
  [552] = 538,
  [553] = 539,
  [554] = 527,
  [555] = 528,
  [556] = 529,
  [557] = 530,
  [558] = 532,
  [559] = 533,
  [560] = 526,
  [561] = 535,
  [562] = 536,
  [563] = 537,
  [564] = 538,
  [565] = 549,
  [566] = 550,
  [567] = 549,
  [568] = 539,
  [569] = 527,
  [570] = 528,
  [571] = 529,
  [572] = 530,
  [573] = 532,
  [574] = 533,
  [575] = 526,
  [576] = 535,
  [577] = 536,
  [578] = 537,
  [579] = 549,
  [580] = 580,
  [581] = 550,
  [582] = 538,
  [583] = 539,
  [584] = 584,
  [585] = 527,
  [586] = 528,
  [587] = 529,
  [588] = 530,
  [589] = 532,
  [590] = 533,
  [591] = 526,
  [592] = 535,
  [593] = 536,
  [594] = 537,
  [595] = 531,
  [596] = 549,
  [597] = 538,
  [598] = 539,
  [599] = 527,
  [600] = 528,
  [601] = 529,
  [602] = 530,
  [603] = 532,
  [604] = 533,
  [605] = 526,
  [606] = 535,
  [607] = 536,
  [608] = 537,
  [609] = 550,
  [610] = 549,
  [611] = 539,
  [612] = 550,
  [613] = 538,
  [614] = 539,
  [615] = 527,
  [616] = 528,
  [617] = 529,
  [618] = 530,
  [619] = 532,
  [620] = 533,
  [621] = 526,
  [622] = 535,
  [623] = 536,
  [624] = 537,
  [625] = 549,
  [626] = 550,
  [627] = 538,
  [628] = 539,
  [629] = 527,
  [630] = 528,
  [631] = 584,
  [632] = 530,
  [633] = 532,
  [634] = 533,
  [635] = 526,
  [636] = 535,
  [637] = 536,
  [638] = 537,
  [639] = 549,
  [640] = 550,
  [641] = 538,
  [642] = 539,
  [643] = 527,
  [644] = 528,
  [645] = 529,
  [646] = 530,
  [647] = 532,
  [648] = 533,
  [649] = 526,
  [650] = 535,
  [651] = 536,
  [652] = 537,
  [653] = 549,
  [654] = 538,
  [655] = 550,
  [656] = 539,
  [657] = 527,
  [658] = 528,
  [659] = 529,
  [660] = 530,
  [661] = 532,
  [662] = 533,
  [663] = 526,
  [664] = 535,
  [665] = 536,
  [666] = 537,
  [667] = 538,
  [668] = 539,
  [669] = 527,
  [670] = 528,
  [671] = 529,
  [672] = 530,
  [673] = 532,
  [674] = 533,
  [675] = 526,
  [676] = 535,
  [677] = 536,
  [678] = 537,
  [679] = 549,
  [680] = 550,
  [681] = 549,
  [682] = 682,
  [683] = 538,
  [684] = 550,
  [685] = 538,
  [686] = 539,
  [687] = 527,
  [688] = 528,
  [689] = 529,
  [690] = 530,
  [691] = 532,
  [692] = 533,
  [693] = 526,
  [694] = 535,
  [695] = 536,
  [696] = 537,
  [697] = 549,
  [698] = 550,
  [699] = 538,
  [700] = 550,
  [701] = 539,
  [702] = 527,
  [703] = 528,
  [704] = 529,
  [705] = 530,
  [706] = 532,
  [707] = 533,
  [708] = 526,
  [709] = 535,
  [710] = 536,
  [711] = 537,
  [712] = 549,
  [713] = 538,
  [714] = 539,
  [715] = 527,
  [716] = 528,
  [717] = 529,
  [718] = 530,
  [719] = 532,
  [720] = 533,
  [721] = 526,
  [722] = 535,
  [723] = 536,
  [724] = 537,
  [725] = 549,
  [726] = 550,
  [727] = 550,
  [728] = 529,
  [729] = 75,
  [730] = 730,
  [731] = 731,
  [732] = 75,
  [733] = 80,
  [734] = 79,
  [735] = 79,
  [736] = 736,
  [737] = 736,
  [738] = 736,
  [739] = 736,
  [740] = 736,
  [741] = 736,
  [742] = 736,
  [743] = 736,
  [744] = 736,
  [745] = 736,
  [746] = 736,
  [747] = 736,
  [748] = 736,
  [749] = 736,
  [750] = 750,
  [751] = 751,
  [752] = 751,
  [753] = 751,
  [754] = 751,
  [755] = 751,
  [756] = 751,
  [757] = 751,
  [758] = 750,
  [759] = 751,
  [760] = 751,
  [761] = 751,
  [762] = 751,
  [763] = 751,
  [764] = 751,
  [765] = 751,
  [766] = 766,
  [767] = 767,
  [768] = 768,
  [769] = 769,
  [770] = 770,
  [771] = 771,
  [772] = 772,
  [773] = 773,
  [774] = 774,
  [775] = 775,
  [776] = 776,
  [777] = 777,
  [778] = 778,
  [779] = 779,
  [780] = 780,
  [781] = 781,
  [782] = 782,
  [783] = 783,
  [784] = 784,
  [785] = 785,
  [786] = 786,
  [787] = 787,
  [788] = 788,
  [789] = 789,
  [790] = 790,
  [791] = 791,
  [792] = 792,
  [793] = 793,
  [794] = 794,
  [795] = 766,
  [796] = 796,
  [797] = 797,
  [798] = 798,
  [799] = 799,
  [800] = 800,
  [801] = 801,
  [802] = 802,
  [803] = 803,
  [804] = 804,
  [805] = 805,
  [806] = 806,
  [807] = 807,
  [808] = 808,
  [809] = 809,
  [810] = 810,
  [811] = 811,
  [812] = 812,
  [813] = 80,
  [814] = 814,
  [815] = 815,
  [816] = 816,
  [817] = 817,
  [818] = 818,
  [819] = 819,
  [820] = 820,
  [821] = 821,
  [822] = 822,
  [823] = 823,
  [824] = 767,
  [825] = 825,
  [826] = 826,
  [827] = 827,
  [828] = 828,
  [829] = 829,
  [830] = 830,
  [831] = 831,
  [832] = 832,
  [833] = 833,
  [834] = 834,
  [835] = 835,
  [836] = 836,
  [837] = 837,
  [838] = 838,
  [839] = 839,
  [840] = 840,
  [841] = 841,
  [842] = 842,
  [843] = 843,
  [844] = 844,
  [845] = 845,
  [846] = 846,
  [847] = 847,
  [848] = 848,
  [849] = 767,
  [850] = 850,
  [851] = 730,
  [852] = 766,
  [853] = 853,
  [854] = 854,
  [855] = 855,
  [856] = 856,
  [857] = 776,
  [858] = 768,
  [859] = 854,
  [860] = 860,
  [861] = 861,
  [862] = 862,
  [863] = 850,
  [864] = 731,
  [865] = 865,
  [866] = 775,
  [867] = 853,
  [868] = 868,
  [869] = 869,
  [870] = 870,
  [871] = 770,
  [872] = 872,
  [873] = 873,
  [874] = 769,
  [875] = 774,
  [876] = 771,
  [877] = 877,
  [878] = 772,
  [879] = 879,
  [880] = 773,
  [881] = 881,
  [882] = 853,
  [883] = 821,
  [884] = 811,
  [885] = 847,
  [886] = 808,
  [887] = 835,
  [888] = 836,
  [889] = 889,
  [890] = 840,
  [891] = 804,
  [892] = 816,
  [893] = 807,
  [894] = 828,
  [895] = 777,
  [896] = 812,
  [897] = 818,
  [898] = 823,
  [899] = 825,
  [900] = 827,
  [901] = 792,
  [902] = 829,
  [903] = 830,
  [904] = 817,
  [905] = 832,
  [906] = 833,
  [907] = 837,
  [908] = 793,
  [909] = 846,
  [910] = 834,
  [911] = 822,
  [912] = 797,
  [913] = 806,
  [914] = 778,
  [915] = 779,
  [916] = 784,
  [917] = 789,
  [918] = 791,
  [919] = 799,
  [920] = 796,
  [921] = 802,
  [922] = 798,
  [923] = 787,
  [924] = 790,
  [925] = 810,
  [926] = 819,
  [927] = 814,
  [928] = 820,
  [929] = 841,
  [930] = 842,
  [931] = 843,
  [932] = 844,
  [933] = 794,
  [934] = 815,
  [935] = 780,
  [936] = 801,
  [937] = 805,
  [938] = 854,
  [939] = 803,
  [940] = 826,
  [941] = 785,
  [942] = 839,
  [943] = 809,
  [944] = 870,
  [945] = 800,
  [946] = 831,
  [947] = 838,
  [948] = 845,
  [949] = 781,
  [950] = 860,
  [951] = 782,
  [952] = 783,
  [953] = 788,
  [954] = 786,
  [955] = 868,
  [956] = 773,
  [957] = 879,
  [958] = 855,
  [959] = 856,
  [960] = 960,
  [961] = 869,
  [962] = 962,
  [963] = 881,
  [964] = 776,
  [965] = 768,
  [966] = 850,
  [967] = 967,
  [968] = 872,
  [969] = 775,
  [970] = 865,
  [971] = 873,
  [972] = 862,
  [973] = 770,
  [974] = 769,
  [975] = 774,
  [976] = 771,
  [977] = 772,
  [978] = 807,
  [979] = 786,
  [980] = 821,
  [981] = 860,
  [982] = 787,
  [983] = 847,
  [984] = 835,
  [985] = 836,
  [986] = 840,
  [987] = 785,
  [988] = 870,
  [989] = 810,
  [990] = 803,
  [991] = 798,
  [992] = 793,
  [993] = 804,
  [994] = 805,
  [995] = 853,
  [996] = 828,
  [997] = 777,
  [998] = 811,
  [999] = 812,
  [1000] = 839,
  [1001] = 841,
  [1002] = 842,
  [1003] = 790,
  [1004] = 818,
  [1005] = 823,
  [1006] = 825,
  [1007] = 826,
  [1008] = 827,
  [1009] = 829,
  [1010] = 830,
  [1011] = 843,
  [1012] = 848,
  [1013] = 797,
  [1014] = 832,
  [1015] = 833,
  [1016] = 837,
  [1017] = 792,
  [1018] = 822,
  [1019] = 794,
  [1020] = 844,
  [1021] = 846,
  [1022] = 834,
  [1023] = 806,
  [1024] = 809,
  [1025] = 778,
  [1026] = 779,
  [1027] = 784,
  [1028] = 789,
  [1029] = 791,
  [1030] = 796,
  [1031] = 800,
  [1032] = 802,
  [1033] = 815,
  [1034] = 817,
  [1035] = 799,
  [1036] = 814,
  [1037] = 820,
  [1038] = 831,
  [1039] = 838,
  [1040] = 845,
  [1041] = 819,
  [1042] = 780,
  [1043] = 781,
  [1044] = 782,
  [1045] = 783,
  [1046] = 788,
  [1047] = 801,
  [1048] = 854,
  [1049] = 808,
  [1050] = 816,
  [1051] = 960,
  [1052] = 967,
  [1053] = 877,
  [1054] = 861,
  [1055] = 773,
  [1056] = 776,
  [1057] = 730,
  [1058] = 770,
  [1059] = 769,
  [1060] = 774,
  [1061] = 768,
  [1062] = 775,
  [1063] = 771,
  [1064] = 772,
  [1065] = 889,
  [1066] = 967,
  [1067] = 814,
  [1068] = 842,
  [1069] = 808,
  [1070] = 790,
  [1071] = 811,
  [1072] = 731,
  [1073] = 730,
  [1074] = 807,
  [1075] = 797,
  [1076] = 786,
  [1077] = 837,
  [1078] = 801,
  [1079] = 792,
  [1080] = 846,
  [1081] = 781,
  [1082] = 889,
  [1083] = 850,
  [1084] = 823,
  [1085] = 825,
  [1086] = 815,
  [1087] = 802,
  [1088] = 870,
  [1089] = 730,
  [1090] = 834,
  [1091] = 847,
  [1092] = 828,
  [1093] = 839,
  [1094] = 827,
  [1095] = 799,
  [1096] = 840,
  [1097] = 829,
  [1098] = 782,
  [1099] = 853,
  [1100] = 822,
  [1101] = 804,
  [1102] = 780,
  [1103] = 832,
  [1104] = 783,
  [1105] = 787,
  [1106] = 778,
  [1107] = 826,
  [1108] = 821,
  [1109] = 806,
  [1110] = 960,
  [1111] = 779,
  [1112] = 816,
  [1113] = 798,
  [1114] = 860,
  [1115] = 793,
  [1116] = 777,
  [1117] = 833,
  [1118] = 830,
  [1119] = 818,
  [1120] = 784,
  [1121] = 809,
  [1122] = 794,
  [1123] = 789,
  [1124] = 810,
  [1125] = 841,
  [1126] = 791,
  [1127] = 805,
  [1128] = 796,
  [1129] = 800,
  [1130] = 835,
  [1131] = 803,
  [1132] = 836,
  [1133] = 854,
  [1134] = 967,
  [1135] = 843,
  [1136] = 812,
  [1137] = 820,
  [1138] = 845,
  [1139] = 817,
  [1140] = 844,
  [1141] = 785,
  [1142] = 831,
  [1143] = 962,
  [1144] = 838,
  [1145] = 819,
  [1146] = 788,
  [1147] = 861,
  [1148] = 731,
  [1149] = 731,
  [1150] = 877,
  [1151] = 962,
  [1152] = 1152,
  [1153] = 1152,
  [1154] = 1152,
  [1155] = 1152,
  [1156] = 1152,
  [1157] = 1152,
  [1158] = 1152,
  [1159] = 960,
  [1160] = 1152,
  [1161] = 1152,
  [1162] = 1152,
  [1163] = 877,
  [1164] = 861,
  [1165] = 1165,
  [1166] = 1166,
  [1167] = 1166,
  [1168] = 1168,
  [1169] = 1169,
  [1170] = 1170,
  [1171] = 1171,
  [1172] = 1172,
  [1173] = 1173,
  [1174] = 1165,
  [1175] = 1172,
  [1176] = 1170,
  [1177] = 1171,
  [1178] = 1172,
  [1179] = 1165,
  [1180] = 1166,
  [1181] = 1166,
  [1182] = 1165,
  [1183] = 1170,
  [1184] = 1170,
  [1185] = 1171,
  [1186] = 750,
  [1187] = 1172,
  [1188] = 1171,
  [1189] = 1172,
  [1190] = 1171,
  [1191] = 1166,
  [1192] = 1192,
  [1193] = 1193,
  [1194] = 1171,
  [1195] = 1170,
  [1196] = 1166,
  [1197] = 1166,
  [1198] = 1166,
  [1199] = 1165,
  [1200] = 1172,
  [1201] = 1170,
  [1202] = 1171,
  [1203] = 1172,
  [1204] = 1204,
  [1205] = 1170,
  [1206] = 1165,
  [1207] = 1166,
  [1208] = 1165,
  [1209] = 1171,
  [1210] = 1172,
  [1211] = 1170,
  [1212] = 1165,
  [1213] = 1170,
  [1214] = 1170,
  [1215] = 1171,
  [1216] = 1171,
  [1217] = 1165,
  [1218] = 1172,
  [1219] = 1165,
  [1220] = 1172,
  [1221] = 1166,
  [1222] = 1222,
  [1223] = 1223,
  [1224] = 1224,
  [1225] = 1225,
  [1226] = 1226,
  [1227] = 1227,
  [1228] = 1228,
  [1229] = 889,
  [1230] = 1230,
  [1231] = 1222,
  [1232] = 1232,
  [1233] = 1233,
  [1234] = 1234,
  [1235] = 1223,
  [1236] = 1224,
  [1237] = 1225,
  [1238] = 1226,
  [1239] = 1228,
  [1240] = 1230,
  [1241] = 1230,
  [1242] = 1232,
  [1243] = 1233,
  [1244] = 1226,
  [1245] = 1230,
  [1246] = 1232,
  [1247] = 1233,
  [1248] = 1226,
  [1249] = 1232,
  [1250] = 1233,
  [1251] = 860,
  [1252] = 870,
  [1253] = 1224,
  [1254] = 1234,
  [1255] = 1226,
  [1256] = 1232,
  [1257] = 1233,
  [1258] = 1232,
  [1259] = 1233,
  [1260] = 1232,
  [1261] = 1233,
  [1262] = 1234,
  [1263] = 1223,
  [1264] = 1224,
  [1265] = 1225,
  [1266] = 1226,
  [1267] = 1227,
  [1268] = 1228,
  [1269] = 1227,
  [1270] = 1228,
  [1271] = 1230,
  [1272] = 1232,
  [1273] = 1233,
  [1274] = 1274,
  [1275] = 730,
  [1276] = 1276,
  [1277] = 1234,
  [1278] = 1223,
  [1279] = 1274,
  [1280] = 1274,
  [1281] = 1274,
  [1282] = 1222,
  [1283] = 1222,
  [1284] = 1225,
  [1285] = 1285,
  [1286] = 1286,
  [1287] = 1287,
  [1288] = 1288,
  [1289] = 1289,
  [1290] = 1288,
  [1291] = 1291,
  [1292] = 1292,
  [1293] = 1289,
  [1294] = 1294,
  [1295] = 1288,
  [1296] = 1285,
  [1297] = 1289,
  [1298] = 1289,
  [1299] = 1292,
  [1300] = 1287,
  [1301] = 1301,
  [1302] = 1301,
  [1303] = 1285,
  [1304] = 1288,
  [1305] = 1291,
  [1306] = 1289,
  [1307] = 1286,
  [1308] = 1294,
  [1309] = 1287,
  [1310] = 1289,
  [1311] = 731,
  [1312] = 1287,
  [1313] = 1288,
  [1314] = 962,
  [1315] = 1287,
  [1316] = 1287,
  [1317] = 1289,
  [1318] = 1289,
  [1319] = 1287,
  [1320] = 1287,
  [1321] = 1321,
  [1322] = 1322,
  [1323] = 1323,
  [1324] = 1324,
  [1325] = 1325,
  [1326] = 1326,
  [1327] = 1327,
  [1328] = 1328,
  [1329] = 1329,
  [1330] = 1330,
  [1331] = 1331,
  [1332] = 1332,
  [1333] = 1333,
  [1334] = 1334,
  [1335] = 1335,
  [1336] = 1336,
  [1337] = 1328,
  [1338] = 1321,
  [1339] = 1332,
  [1340] = 1334,
  [1341] = 1325,
  [1342] = 1333,
  [1343] = 1334,
  [1344] = 1336,
  [1345] = 1332,
  [1346] = 1335,
  [1347] = 1347,
  [1348] = 1348,
  [1349] = 1349,
  [1350] = 1326,
  [1351] = 1327,
  [1352] = 1325,
  [1353] = 1336,
  [1354] = 1326,
  [1355] = 1327,
  [1356] = 1333,
  [1357] = 1357,
  [1358] = 1358,
  [1359] = 1333,
  [1360] = 1330,
  [1361] = 1331,
  [1362] = 1322,
  [1363] = 1335,
  [1364] = 1336,
  [1365] = 1328,
  [1366] = 1321,
  [1367] = 1332,
  [1368] = 1333,
  [1369] = 1334,
  [1370] = 1325,
  [1371] = 1326,
  [1372] = 1327,
  [1373] = 1325,
  [1374] = 1326,
  [1375] = 1375,
  [1376] = 1323,
  [1377] = 1377,
  [1378] = 1327,
  [1379] = 1347,
  [1380] = 960,
  [1381] = 1322,
  [1382] = 1377,
  [1383] = 1377,
  [1384] = 1347,
  [1385] = 1377,
  [1386] = 1333,
  [1387] = 1347,
  [1388] = 1334,
  [1389] = 1377,
  [1390] = 1347,
  [1391] = 1377,
  [1392] = 1347,
  [1393] = 1377,
  [1394] = 1347,
  [1395] = 1377,
  [1396] = 1347,
  [1397] = 1377,
  [1398] = 1322,
  [1399] = 1347,
  [1400] = 1375,
  [1401] = 1377,
  [1402] = 1347,
  [1403] = 1377,
  [1404] = 1347,
  [1405] = 1334,
  [1406] = 1406,
  [1407] = 1406,
  [1408] = 1406,
  [1409] = 1406,
  [1410] = 1410,
  [1411] = 1411,
  [1412] = 967,
  [1413] = 967,
  [1414] = 1406,
  [1415] = 1411,
  [1416] = 1411,
  [1417] = 1406,
  [1418] = 1406,
  [1419] = 1411,
  [1420] = 1406,
  [1421] = 1411,
  [1422] = 1406,
  [1423] = 1411,
  [1424] = 1411,
  [1425] = 1406,
  [1426] = 1411,
  [1427] = 1411,
  [1428] = 1406,
  [1429] = 1410,
  [1430] = 1411,
  [1431] = 1411,
  [1432] = 1432,
  [1433] = 1433,
  [1434] = 1434,
  [1435] = 1433,
  [1436] = 1436,
  [1437] = 1432,
  [1438] = 1436,
  [1439] = 1434,
  [1440] = 1432,
  [1441] = 1436,
  [1442] = 1433,
  [1443] = 1434,
  [1444] = 1444,
  [1445] = 1444,
  [1446] = 1444,
  [1447] = 1447,
  [1448] = 1444,
  [1449] = 1449,
  [1450] = 1447,
  [1451] = 1444,
  [1452] = 865,
  [1453] = 1444,
  [1454] = 1444,
  [1455] = 869,
  [1456] = 872,
  [1457] = 873,
  [1458] = 881,
  [1459] = 1449,
  [1460] = 1449,
  [1461] = 1447,
  [1462] = 1444,
  [1463] = 1449,
  [1464] = 1447,
  [1465] = 862,
  [1466] = 879,
  [1467] = 1444,
  [1468] = 855,
  [1469] = 1444,
  [1470] = 856,
  [1471] = 868,
  [1472] = 1472,
  [1473] = 1472,
  [1474] = 1472,
  [1475] = 1472,
  [1476] = 1472,
  [1477] = 1472,
  [1478] = 1472,
  [1479] = 1472,
  [1480] = 1472,
  [1481] = 1472,
  [1482] = 1472,
  [1483] = 1483,
  [1484] = 1484,
  [1485] = 1485,
  [1486] = 1484,
  [1487] = 1487,
  [1488] = 1488,
  [1489] = 1489,
  [1490] = 1488,
  [1491] = 1484,
  [1492] = 1487,
  [1493] = 1485,
  [1494] = 1489,
  [1495] = 1483,
  [1496] = 1496,
  [1497] = 1483,
  [1498] = 1488,
  [1499] = 1485,
  [1500] = 1484,
  [1501] = 1484,
  [1502] = 1484,
  [1503] = 1503,
  [1504] = 1483,
  [1505] = 1488,
  [1506] = 1487,
  [1507] = 1507,
  [1508] = 1483,
  [1509] = 1484,
  [1510] = 1483,
  [1511] = 1488,
  [1512] = 1483,
  [1513] = 1484,
  [1514] = 1484,
  [1515] = 1489,
  [1516] = 1489,
  [1517] = 1483,
  [1518] = 1488,
  [1519] = 1487,
  [1520] = 1484,
  [1521] = 1489,
  [1522] = 1484,
  [1523] = 1483,
  [1524] = 1483,
  [1525] = 1483,
  [1526] = 1483,
  [1527] = 1488,
  [1528] = 1484,
  [1529] = 1485,
  [1530] = 768,
  [1531] = 776,
  [1532] = 850,
  [1533] = 1533,
  [1534] = 1534,
  [1535] = 1534,
  [1536] = 771,
  [1537] = 769,
  [1538] = 772,
  [1539] = 808,
  [1540] = 773,
  [1541] = 775,
  [1542] = 1542,
  [1543] = 774,
  [1544] = 1544,
  [1545] = 770,
  [1546] = 1534,
  [1547] = 967,
  [1548] = 797,
  [1549] = 789,
  [1550] = 804,
  [1551] = 830,
  [1552] = 807,
  [1553] = 810,
  [1554] = 841,
  [1555] = 842,
  [1556] = 843,
  [1557] = 844,
  [1558] = 778,
  [1559] = 779,
  [1560] = 784,
  [1561] = 822,
  [1562] = 791,
  [1563] = 786,
  [1564] = 787,
  [1565] = 808,
  [1566] = 790,
  [1567] = 816,
  [1568] = 792,
  [1569] = 794,
  [1570] = 796,
  [1571] = 967,
  [1572] = 802,
  [1573] = 815,
  [1574] = 817,
  [1575] = 819,
  [1576] = 821,
  [1577] = 797,
  [1578] = 799,
  [1579] = 839,
  [1580] = 814,
  [1581] = 809,
  [1582] = 816,
  [1583] = 853,
  [1584] = 811,
  [1585] = 847,
  [1586] = 835,
  [1587] = 820,
  [1588] = 836,
  [1589] = 798,
  [1590] = 818,
  [1591] = 860,
  [1592] = 800,
  [1593] = 828,
  [1594] = 777,
  [1595] = 831,
  [1596] = 822,
  [1597] = 838,
  [1598] = 845,
  [1599] = 812,
  [1600] = 781,
  [1601] = 823,
  [1602] = 782,
  [1603] = 780,
  [1604] = 783,
  [1605] = 788,
  [1606] = 785,
  [1607] = 793,
  [1608] = 826,
  [1609] = 967,
  [1610] = 825,
  [1611] = 827,
  [1612] = 801,
  [1613] = 829,
  [1614] = 832,
  [1615] = 833,
  [1616] = 837,
  [1617] = 805,
  [1618] = 854,
  [1619] = 799,
  [1620] = 846,
  [1621] = 834,
  [1622] = 811,
  [1623] = 806,
  [1624] = 870,
  [1625] = 1625,
  [1626] = 803,
  [1627] = 840,
  [1628] = 1628,
  [1629] = 1628,
  [1630] = 1630,
  [1631] = 1630,
  [1632] = 1628,
  [1633] = 1628,
  [1634] = 1630,
  [1635] = 1630,
  [1636] = 1630,
  [1637] = 1628,
  [1638] = 1630,
  [1639] = 1628,
  [1640] = 1630,
  [1641] = 1628,
  [1642] = 1630,
  [1643] = 1630,
  [1644] = 967,
  [1645] = 1628,
  [1646] = 1630,
  [1647] = 1628,
  [1648] = 1630,
  [1649] = 1628,
  [1650] = 1628,
  [1651] = 1628,
  [1652] = 1630,
  [1653] = 1630,
  [1654] = 1628,
  [1655] = 1655,
  [1656] = 1656,
  [1657] = 1656,
  [1658] = 960,
  [1659] = 1656,
  [1660] = 1656,
  [1661] = 1656,
  [1662] = 1656,
  [1663] = 967,
  [1664] = 1656,
  [1665] = 1656,
  [1666] = 967,
  [1667] = 1656,
  [1668] = 967,
  [1669] = 1656,
  [1670] = 1656,
  [1671] = 1656,
  [1672] = 1656,
  [1673] = 1673,
  [1674] = 1674,
  [1675] = 1675,
  [1676] = 1676,
  [1677] = 1677,
  [1678] = 1678,
  [1679] = 1679,
  [1680] = 1680,
  [1681] = 1681,
  [1682] = 793,
  [1683] = 861,
  [1684] = 785,
  [1685] = 1677,
  [1686] = 1686,
  [1687] = 1687,
  [1688] = 1688,
  [1689] = 1677,
  [1690] = 797,
  [1691] = 822,
  [1692] = 877,
  [1693] = 1677,
  [1694] = 1694,
  [1695] = 1695,
  [1696] = 1696,
  [1697] = 1697,
  [1698] = 1698,
  [1699] = 1699,
  [1700] = 967,
  [1701] = 808,
  [1702] = 1702,
  [1703] = 1703,
  [1704] = 781,
  [1705] = 1703,
  [1706] = 791,
  [1707] = 782,
  [1708] = 783,
  [1709] = 796,
  [1710] = 811,
  [1711] = 788,
  [1712] = 800,
  [1713] = 803,
  [1714] = 1674,
  [1715] = 853,
  [1716] = 1703,
  [1717] = 834,
  [1718] = 806,
  [1719] = 1703,
  [1720] = 1675,
  [1721] = 809,
  [1722] = 1703,
  [1723] = 798,
  [1724] = 1724,
  [1725] = 802,
  [1726] = 770,
  [1727] = 1703,
  [1728] = 1676,
  [1729] = 1702,
  [1730] = 1678,
  [1731] = 1679,
  [1732] = 1703,
  [1733] = 769,
  [1734] = 801,
  [1735] = 816,
  [1736] = 789,
  [1737] = 804,
  [1738] = 1703,
  [1739] = 1739,
  [1740] = 805,
  [1741] = 807,
  [1742] = 774,
  [1743] = 814,
  [1744] = 854,
  [1745] = 1703,
  [1746] = 808,
  [1747] = 1703,
  [1748] = 1695,
  [1749] = 828,
  [1750] = 1694,
  [1751] = 1724,
  [1752] = 1752,
  [1753] = 1739,
  [1754] = 1703,
  [1755] = 780,
  [1756] = 837,
  [1757] = 812,
  [1758] = 822,
  [1759] = 784,
  [1760] = 797,
  [1761] = 799,
  [1762] = 1696,
  [1763] = 1697,
  [1764] = 1698,
  [1765] = 1699,
  [1766] = 1687,
  [1767] = 846,
  [1768] = 1703,
  [1769] = 1680,
  [1770] = 820,
  [1771] = 816,
  [1772] = 1703,
  [1773] = 1681,
  [1774] = 1724,
  [1775] = 1752,
  [1776] = 1739,
  [1777] = 1673,
  [1778] = 831,
  [1779] = 1752,
  [1780] = 1739,
  [1781] = 838,
  [1782] = 1739,
  [1783] = 845,
  [1784] = 778,
  [1785] = 818,
  [1786] = 823,
  [1787] = 779,
  [1788] = 825,
  [1789] = 839,
  [1790] = 1686,
  [1791] = 826,
  [1792] = 1752,
  [1793] = 827,
  [1794] = 829,
  [1795] = 830,
  [1796] = 775,
  [1797] = 832,
  [1798] = 799,
  [1799] = 771,
  [1800] = 772,
  [1801] = 773,
  [1802] = 777,
  [1803] = 1681,
  [1804] = 821,
  [1805] = 806,
  [1806] = 804,
  [1807] = 816,
  [1808] = 1695,
  [1809] = 843,
  [1810] = 844,
  [1811] = 807,
  [1812] = 778,
  [1813] = 768,
  [1814] = 1814,
  [1815] = 1815,
  [1816] = 1679,
  [1817] = 779,
  [1818] = 833,
  [1819] = 810,
  [1820] = 822,
  [1821] = 1675,
  [1822] = 797,
  [1823] = 784,
  [1824] = 789,
  [1825] = 828,
  [1826] = 1678,
  [1827] = 799,
  [1828] = 777,
  [1829] = 810,
  [1830] = 791,
  [1831] = 796,
  [1832] = 812,
  [1833] = 802,
  [1834] = 794,
  [1835] = 776,
  [1836] = 1694,
  [1837] = 843,
  [1838] = 1696,
  [1839] = 1697,
  [1840] = 1698,
  [1841] = 1699,
  [1842] = 1687,
  [1843] = 1680,
  [1844] = 844,
  [1845] = 1845,
  [1846] = 770,
  [1847] = 769,
  [1848] = 782,
  [1849] = 835,
  [1850] = 1695,
  [1851] = 774,
  [1852] = 785,
  [1853] = 814,
  [1854] = 835,
  [1855] = 783,
  [1856] = 820,
  [1857] = 1857,
  [1858] = 815,
  [1859] = 1496,
  [1860] = 818,
  [1861] = 790,
  [1862] = 823,
  [1863] = 836,
  [1864] = 840,
  [1865] = 1673,
  [1866] = 841,
  [1867] = 771,
  [1868] = 786,
  [1869] = 772,
  [1870] = 1870,
  [1871] = 773,
  [1872] = 780,
  [1873] = 847,
  [1874] = 825,
  [1875] = 1875,
  [1876] = 1876,
  [1877] = 793,
  [1878] = 1878,
  [1879] = 1845,
  [1880] = 811,
  [1881] = 768,
  [1882] = 1676,
  [1883] = 781,
  [1884] = 827,
  [1885] = 829,
  [1886] = 842,
  [1887] = 838,
  [1888] = 1888,
  [1889] = 786,
  [1890] = 787,
  [1891] = 790,
  [1892] = 792,
  [1893] = 801,
  [1894] = 787,
  [1895] = 840,
  [1896] = 830,
  [1897] = 847,
  [1898] = 794,
  [1899] = 775,
  [1900] = 1686,
  [1901] = 832,
  [1902] = 1675,
  [1903] = 841,
  [1904] = 815,
  [1905] = 842,
  [1906] = 792,
  [1907] = 833,
  [1908] = 808,
  [1909] = 805,
  [1910] = 854,
  [1911] = 1815,
  [1912] = 788,
  [1913] = 794,
  [1914] = 800,
  [1915] = 819,
  [1916] = 1503,
  [1917] = 817,
  [1918] = 821,
  [1919] = 776,
  [1920] = 1674,
  [1921] = 803,
  [1922] = 1814,
  [1923] = 837,
  [1924] = 1674,
  [1925] = 1507,
  [1926] = 798,
  [1927] = 815,
  [1928] = 1888,
  [1929] = 811,
  [1930] = 826,
  [1931] = 817,
  [1932] = 819,
  [1933] = 819,
  [1934] = 845,
  [1935] = 831,
  [1936] = 786,
  [1937] = 1875,
  [1938] = 787,
  [1939] = 846,
  [1940] = 790,
  [1941] = 1702,
  [1942] = 821,
  [1943] = 839,
  [1944] = 809,
  [1945] = 853,
  [1946] = 834,
  [1947] = 836,
  [1948] = 1948,
  [1949] = 777,
  [1950] = 812,
  [1951] = 818,
  [1952] = 823,
  [1953] = 825,
  [1954] = 827,
  [1955] = 829,
  [1956] = 830,
  [1957] = 775,
  [1958] = 832,
  [1959] = 833,
  [1960] = 837,
  [1961] = 846,
  [1962] = 834,
  [1963] = 806,
  [1964] = 778,
  [1965] = 779,
  [1966] = 784,
  [1967] = 789,
  [1968] = 791,
  [1969] = 796,
  [1970] = 802,
  [1971] = 770,
  [1972] = 769,
  [1973] = 774,
  [1974] = 814,
  [1975] = 820,
  [1976] = 771,
  [1977] = 772,
  [1978] = 773,
  [1979] = 780,
  [1980] = 1980,
  [1981] = 801,
  [1982] = 805,
  [1983] = 854,
  [1984] = 776,
  [1985] = 826,
  [1986] = 839,
  [1987] = 804,
  [1988] = 809,
  [1989] = 853,
  [1990] = 768,
  [1991] = 800,
  [1992] = 831,
  [1993] = 838,
  [1994] = 845,
  [1995] = 781,
  [1996] = 782,
  [1997] = 783,
  [1998] = 788,
  [1999] = 1999,
  [2000] = 2000,
  [2001] = 2001,
  [2002] = 2002,
  [2003] = 1676,
  [2004] = 785,
  [2005] = 793,
  [2006] = 2006,
  [2007] = 2007,
  [2008] = 2008,
  [2009] = 792,
  [2010] = 2010,
  [2011] = 2011,
  [2012] = 817,
  [2013] = 2013,
  [2014] = 2014,
  [2015] = 786,
  [2016] = 787,
  [2017] = 790,
  [2018] = 810,
  [2019] = 819,
  [2020] = 2020,
  [2021] = 841,
  [2022] = 842,
  [2023] = 843,
  [2024] = 844,
  [2025] = 794,
  [2026] = 822,
  [2027] = 815,
  [2028] = 821,
  [2029] = 808,
  [2030] = 816,
  [2031] = 2031,
  [2032] = 797,
  [2033] = 799,
  [2034] = 2034,
  [2035] = 2035,
  [2036] = 2036,
  [2037] = 811,
  [2038] = 847,
  [2039] = 835,
  [2040] = 828,
  [2041] = 840,
  [2042] = 2042,
  [2043] = 2043,
  [2044] = 2044,
  [2045] = 1695,
  [2046] = 2046,
  [2047] = 2047,
  [2048] = 2048,
  [2049] = 793,
  [2050] = 2050,
  [2051] = 792,
  [2052] = 1888,
  [2053] = 1815,
  [2054] = 2054,
  [2055] = 1673,
  [2056] = 2056,
  [2057] = 2057,
  [2058] = 2058,
  [2059] = 2059,
  [2060] = 2060,
  [2061] = 1875,
  [2062] = 1845,
  [2063] = 817,
  [2064] = 1674,
  [2065] = 1674,
  [2066] = 2066,
  [2067] = 1675,
  [2068] = 1814,
  [2069] = 2069,
  [2070] = 1675,
  [2071] = 2071,
  [2072] = 2072,
  [2073] = 2073,
  [2074] = 2074,
  [2075] = 2075,
  [2076] = 2076,
  [2077] = 2077,
  [2078] = 2078,
  [2079] = 2079,
  [2080] = 2080,
  [2081] = 2081,
  [2082] = 2082,
  [2083] = 2083,
  [2084] = 2084,
  [2085] = 1695,
  [2086] = 810,
  [2087] = 1686,
  [2088] = 847,
  [2089] = 2089,
  [2090] = 2090,
  [2091] = 835,
  [2092] = 836,
  [2093] = 840,
  [2094] = 841,
  [2095] = 842,
  [2096] = 843,
  [2097] = 844,
  [2098] = 2098,
  [2099] = 2099,
  [2100] = 2100,
  [2101] = 2101,
  [2102] = 2102,
  [2103] = 2103,
  [2104] = 2104,
  [2105] = 2105,
  [2106] = 2106,
  [2107] = 785,
  [2108] = 1694,
  [2109] = 1696,
  [2110] = 1697,
  [2111] = 1698,
  [2112] = 2112,
  [2113] = 1699,
  [2114] = 1702,
  [2115] = 1678,
  [2116] = 1679,
  [2117] = 1687,
  [2118] = 1680,
  [2119] = 1681,
  [2120] = 807,
  [2121] = 803,
  [2122] = 798,
  [2123] = 836,
  [2124] = 2124,
  [2125] = 2125,
  [2126] = 2126,
  [2127] = 2127,
  [2128] = 2128,
  [2129] = 2129,
  [2130] = 2130,
  [2131] = 2131,
  [2132] = 2125,
  [2133] = 2133,
  [2134] = 2134,
  [2135] = 2125,
  [2136] = 2136,
  [2137] = 2136,
  [2138] = 2136,
  [2139] = 2127,
  [2140] = 2140,
  [2141] = 835,
  [2142] = 836,
  [2143] = 2143,
  [2144] = 2144,
  [2145] = 840,
  [2146] = 2146,
  [2147] = 2147,
  [2148] = 840,
  [2149] = 2149,
  [2150] = 2150,
  [2151] = 2151,
  [2152] = 2152,
  [2153] = 2153,
  [2154] = 2154,
  [2155] = 2155,
  [2156] = 2156,
  [2157] = 1999,
  [2158] = 2158,
  [2159] = 2159,
  [2160] = 2160,
  [2161] = 841,
  [2162] = 2127,
  [2163] = 1695,
  [2164] = 2164,
  [2165] = 842,
  [2166] = 2166,
  [2167] = 843,
  [2168] = 2151,
  [2169] = 844,
  [2170] = 2153,
  [2171] = 2171,
  [2172] = 2172,
  [2173] = 2173,
  [2174] = 2112,
  [2175] = 2175,
  [2176] = 1870,
  [2177] = 2177,
  [2178] = 2155,
  [2179] = 2166,
  [2180] = 2180,
  [2181] = 2181,
  [2182] = 2182,
  [2183] = 2172,
  [2184] = 2180,
  [2185] = 2182,
  [2186] = 2186,
  [2187] = 2187,
  [2188] = 2188,
  [2189] = 2187,
  [2190] = 2190,
  [2191] = 2191,
  [2192] = 810,
  [2193] = 2193,
  [2194] = 2008,
  [2195] = 2010,
  [2196] = 2196,
  [2197] = 2197,
  [2198] = 2147,
  [2199] = 847,
  [2200] = 2181,
  [2201] = 836,
  [2202] = 2202,
  [2203] = 2044,
  [2204] = 847,
  [2205] = 2130,
  [2206] = 2150,
  [2207] = 2207,
  [2208] = 2208,
  [2209] = 2209,
  [2210] = 2188,
  [2211] = 2152,
  [2212] = 2212,
  [2213] = 2213,
  [2214] = 2190,
  [2215] = 2134,
  [2216] = 2133,
  [2217] = 786,
  [2218] = 2191,
  [2219] = 2124,
  [2220] = 2126,
  [2221] = 787,
  [2222] = 2164,
  [2223] = 2177,
  [2224] = 2128,
  [2225] = 2129,
  [2226] = 790,
  [2227] = 2196,
  [2228] = 2228,
  [2229] = 792,
  [2230] = 2230,
  [2231] = 2230,
  [2232] = 2232,
  [2233] = 794,
  [2234] = 2140,
  [2235] = 2235,
  [2236] = 2236,
  [2237] = 2237,
  [2238] = 2238,
  [2239] = 1674,
  [2240] = 2056,
  [2241] = 2241,
  [2242] = 2057,
  [2243] = 2243,
  [2244] = 2244,
  [2245] = 2245,
  [2246] = 2246,
  [2247] = 2247,
  [2248] = 2247,
  [2249] = 2249,
  [2250] = 2250,
  [2251] = 1857,
  [2252] = 815,
  [2253] = 2228,
  [2254] = 1675,
  [2255] = 2232,
  [2256] = 817,
  [2257] = 819,
  [2258] = 821,
  [2259] = 2238,
  [2260] = 835,
  [2261] = 2146,
  [2262] = 2262,
  [2263] = 2263,
  [2264] = 2129,
  [2265] = 2153,
  [2266] = 2164,
  [2267] = 2196,
  [2268] = 2247,
  [2269] = 2269,
  [2270] = 2146,
  [2271] = 2155,
  [2272] = 2166,
  [2273] = 2172,
  [2274] = 2180,
  [2275] = 2275,
  [2276] = 2182,
  [2277] = 2277,
  [2278] = 2187,
  [2279] = 2269,
  [2280] = 2280,
  [2281] = 2237,
  [2282] = 2124,
  [2283] = 2177,
  [2284] = 2277,
  [2285] = 1507,
  [2286] = 2269,
  [2287] = 2277,
  [2288] = 2186,
  [2289] = 2289,
  [2290] = 2128,
  [2291] = 2275,
  [2292] = 1503,
  [2293] = 2293,
  [2294] = 2280,
  [2295] = 2280,
  [2296] = 2188,
  [2297] = 2181,
  [2298] = 2289,
  [2299] = 2126,
  [2300] = 2289,
  [2301] = 2147,
  [2302] = 2277,
  [2303] = 2275,
  [2304] = 2134,
  [2305] = 2228,
  [2306] = 2293,
  [2307] = 2232,
  [2308] = 2245,
  [2309] = 2230,
  [2310] = 2277,
  [2311] = 2140,
  [2312] = 2312,
  [2313] = 2150,
  [2314] = 2190,
  [2315] = 2191,
  [2316] = 2133,
  [2317] = 2152,
  [2318] = 2275,
  [2319] = 2249,
  [2320] = 2293,
  [2321] = 2238,
  [2322] = 2244,
  [2323] = 2151,
  [2324] = 2312,
  [2325] = 2280,
  [2326] = 2130,
  [2327] = 2280,
  [2328] = 1496,
  [2329] = 2280,
  [2330] = 2124,
  [2331] = 2134,
  [2332] = 2332,
  [2333] = 2333,
  [2334] = 2129,
  [2335] = 2333,
  [2336] = 2128,
  [2337] = 2333,
  [2338] = 2332,
  [2339] = 2130,
  [2340] = 2332,
  [2341] = 2249,
  [2342] = 2126,
  [2343] = 2133,
  [2344] = 799,
  [2345] = 870,
  [2346] = 822,
  [2347] = 786,
  [2348] = 790,
  [2349] = 792,
  [2350] = 803,
  [2351] = 798,
  [2352] = 804,
  [2353] = 805,
  [2354] = 807,
  [2355] = 828,
  [2356] = 777,
  [2357] = 812,
  [2358] = 794,
  [2359] = 815,
  [2360] = 817,
  [2361] = 819,
  [2362] = 821,
  [2363] = 818,
  [2364] = 823,
  [2365] = 825,
  [2366] = 826,
  [2367] = 827,
  [2368] = 829,
  [2369] = 830,
  [2370] = 775,
  [2371] = 832,
  [2372] = 833,
  [2373] = 837,
  [2374] = 846,
  [2375] = 834,
  [2376] = 806,
  [2377] = 809,
  [2378] = 778,
  [2379] = 779,
  [2380] = 784,
  [2381] = 789,
  [2382] = 791,
  [2383] = 796,
  [2384] = 800,
  [2385] = 802,
  [2386] = 770,
  [2387] = 769,
  [2388] = 774,
  [2389] = 814,
  [2390] = 820,
  [2391] = 831,
  [2392] = 838,
  [2393] = 845,
  [2394] = 771,
  [2395] = 772,
  [2396] = 773,
  [2397] = 780,
  [2398] = 781,
  [2399] = 782,
  [2400] = 783,
  [2401] = 788,
  [2402] = 801,
  [2403] = 839,
  [2404] = 860,
  [2405] = 785,
  [2406] = 793,
  [2407] = 776,
  [2408] = 811,
  [2409] = 768,
  [2410] = 797,
  [2411] = 2249,
  [2412] = 847,
  [2413] = 808,
  [2414] = 835,
  [2415] = 836,
  [2416] = 840,
  [2417] = 810,
  [2418] = 841,
  [2419] = 816,
  [2420] = 842,
  [2421] = 843,
  [2422] = 844,
  [2423] = 787,
  [2424] = 845,
  [2425] = 781,
  [2426] = 782,
  [2427] = 783,
  [2428] = 788,
  [2429] = 2069,
  [2430] = 2056,
  [2431] = 2104,
  [2432] = 847,
  [2433] = 854,
  [2434] = 805,
  [2435] = 797,
  [2436] = 835,
  [2437] = 836,
  [2438] = 840,
  [2439] = 2084,
  [2440] = 2131,
  [2441] = 826,
  [2442] = 2099,
  [2443] = 2100,
  [2444] = 809,
  [2445] = 853,
  [2446] = 808,
  [2447] = 800,
  [2448] = 831,
  [2449] = 838,
  [2450] = 2241,
  [2451] = 2056,
  [2452] = 2235,
  [2453] = 2173,
  [2454] = 2207,
  [2455] = 2208,
  [2456] = 2149,
  [2457] = 2057,
  [2458] = 2156,
  [2459] = 2159,
  [2460] = 811,
  [2461] = 2461,
  [2462] = 847,
  [2463] = 815,
  [2464] = 1857,
  [2465] = 2158,
  [2466] = 786,
  [2467] = 835,
  [2468] = 1496,
  [2469] = 836,
  [2470] = 2160,
  [2471] = 2112,
  [2472] = 1870,
  [2473] = 2175,
  [2474] = 840,
  [2475] = 2236,
  [2476] = 2243,
  [2477] = 817,
  [2478] = 2250,
  [2479] = 2202,
  [2480] = 819,
  [2481] = 1999,
  [2482] = 821,
  [2483] = 2008,
  [2484] = 2010,
  [2485] = 794,
  [2486] = 2461,
  [2487] = 2212,
  [2488] = 2197,
  [2489] = 2213,
  [2490] = 787,
  [2491] = 810,
  [2492] = 799,
  [2493] = 2461,
  [2494] = 1503,
  [2495] = 960,
  [2496] = 2263,
  [2497] = 2246,
  [2498] = 2262,
  [2499] = 2143,
  [2500] = 2209,
  [2501] = 2171,
  [2502] = 2144,
  [2503] = 2193,
  [2504] = 790,
  [2505] = 841,
  [2506] = 842,
  [2507] = 843,
  [2508] = 844,
  [2509] = 822,
  [2510] = 2057,
  [2511] = 1507,
  [2512] = 1980,
  [2513] = 2154,
  [2514] = 792,
  [2515] = 816,
  [2516] = 2044,
  [2517] = 2517,
  [2518] = 2518,
  [2519] = 877,
  [2520] = 2518,
  [2521] = 2521,
  [2522] = 861,
  [2523] = 2517,
  [2524] = 2518,
  [2525] = 2525,
  [2526] = 2526,
  [2527] = 2517,
  [2528] = 1699,
  [2529] = 2529,
  [2530] = 2530,
  [2531] = 2531,
  [2532] = 2532,
  [2533] = 2533,
  [2534] = 2534,
  [2535] = 2535,
  [2536] = 2536,
  [2537] = 1676,
  [2538] = 2538,
  [2539] = 1675,
  [2540] = 2535,
  [2541] = 1695,
  [2542] = 2542,
  [2543] = 1673,
  [2544] = 1686,
  [2545] = 1695,
  [2546] = 1674,
  [2547] = 2547,
  [2548] = 1676,
  [2549] = 1694,
  [2550] = 2530,
  [2551] = 2534,
  [2552] = 1696,
  [2553] = 1697,
  [2554] = 2554,
  [2555] = 2529,
  [2556] = 2533,
  [2557] = 2532,
  [2558] = 1702,
  [2559] = 1678,
  [2560] = 1679,
  [2561] = 1687,
  [2562] = 1680,
  [2563] = 2533,
  [2564] = 1681,
  [2565] = 1674,
  [2566] = 1694,
  [2567] = 1675,
  [2568] = 1673,
  [2569] = 1696,
  [2570] = 1697,
  [2571] = 1698,
  [2572] = 1699,
  [2573] = 1702,
  [2574] = 1678,
  [2575] = 1679,
  [2576] = 1687,
  [2577] = 1680,
  [2578] = 1681,
  [2579] = 2536,
  [2580] = 2536,
  [2581] = 2534,
  [2582] = 1686,
  [2583] = 2529,
  [2584] = 2530,
  [2585] = 2535,
  [2586] = 2586,
  [2587] = 2532,
  [2588] = 2530,
  [2589] = 2589,
  [2590] = 2529,
  [2591] = 1698,
  [2592] = 2592,
  [2593] = 2593,
  [2594] = 2594,
  [2595] = 2595,
  [2596] = 2596,
  [2597] = 2597,
  [2598] = 2598,
  [2599] = 2599,
  [2600] = 2600,
  [2601] = 2601,
  [2602] = 2600,
  [2603] = 2603,
  [2604] = 2604,
  [2605] = 2605,
  [2606] = 2601,
  [2607] = 2597,
  [2608] = 2598,
  [2609] = 2599,
  [2610] = 2610,
  [2611] = 2592,
  [2612] = 2592,
  [2613] = 2600,
  [2614] = 2614,
  [2615] = 2615,
  [2616] = 2616,
  [2617] = 2597,
  [2618] = 2618,
  [2619] = 2619,
  [2620] = 2620,
  [2621] = 2621,
  [2622] = 2622,
  [2623] = 2623,
  [2624] = 2624,
  [2625] = 2625,
  [2626] = 2597,
  [2627] = 2627,
  [2628] = 1674,
  [2629] = 2629,
  [2630] = 1695,
  [2631] = 1675,
  [2632] = 2601,
  [2633] = 2633,
  [2634] = 2634,
  [2635] = 2635,
  [2636] = 2636,
  [2637] = 2637,
  [2638] = 2638,
  [2639] = 2639,
  [2640] = 2640,
  [2641] = 2598,
  [2642] = 2642,
  [2643] = 2643,
  [2644] = 2599,
  [2645] = 2645,
  [2646] = 2646,
  [2647] = 2592,
  [2648] = 2648,
  [2649] = 2649,
  [2650] = 2601,
  [2651] = 2651,
  [2652] = 2652,
  [2653] = 2653,
  [2654] = 2654,
  [2655] = 2655,
  [2656] = 2656,
  [2657] = 2657,
  [2658] = 1674,
  [2659] = 2659,
  [2660] = 2660,
  [2661] = 2661,
  [2662] = 2662,
  [2663] = 2663,
  [2664] = 2664,
  [2665] = 1675,
  [2666] = 2666,
  [2667] = 2667,
  [2668] = 2668,
  [2669] = 1695,
  [2670] = 2670,
  [2671] = 818,
  [2672] = 801,
  [2673] = 2673,
  [2674] = 841,
  [2675] = 842,
  [2676] = 823,
  [2677] = 826,
  [2678] = 825,
  [2679] = 843,
  [2680] = 837,
  [2681] = 831,
  [2682] = 803,
  [2683] = 829,
  [2684] = 800,
  [2685] = 846,
  [2686] = 811,
  [2687] = 847,
  [2688] = 828,
  [2689] = 834,
  [2690] = 2690,
  [2691] = 806,
  [2692] = 830,
  [2693] = 819,
  [2694] = 835,
  [2695] = 2690,
  [2696] = 809,
  [2697] = 794,
  [2698] = 836,
  [2699] = 853,
  [2700] = 773,
  [2701] = 781,
  [2702] = 768,
  [2703] = 1542,
  [2704] = 782,
  [2705] = 822,
  [2706] = 777,
  [2707] = 840,
  [2708] = 2708,
  [2709] = 2709,
  [2710] = 787,
  [2711] = 785,
  [2712] = 793,
  [2713] = 798,
  [2714] = 827,
  [2715] = 783,
  [2716] = 1544,
  [2717] = 778,
  [2718] = 779,
  [2719] = 2719,
  [2720] = 808,
  [2721] = 1533,
  [2722] = 784,
  [2723] = 810,
  [2724] = 789,
  [2725] = 1814,
  [2726] = 1875,
  [2727] = 772,
  [2728] = 2134,
  [2729] = 791,
  [2730] = 2708,
  [2731] = 808,
  [2732] = 796,
  [2733] = 1845,
  [2734] = 786,
  [2735] = 802,
  [2736] = 770,
  [2737] = 792,
  [2738] = 2130,
  [2739] = 2673,
  [2740] = 769,
  [2741] = 815,
  [2742] = 2134,
  [2743] = 838,
  [2744] = 780,
  [2745] = 845,
  [2746] = 805,
  [2747] = 797,
  [2748] = 2719,
  [2749] = 844,
  [2750] = 775,
  [2751] = 797,
  [2752] = 788,
  [2753] = 799,
  [2754] = 821,
  [2755] = 816,
  [2756] = 832,
  [2757] = 790,
  [2758] = 774,
  [2759] = 2673,
  [2760] = 817,
  [2761] = 2690,
  [2762] = 771,
  [2763] = 854,
  [2764] = 833,
  [2765] = 814,
  [2766] = 2130,
  [2767] = 812,
  [2768] = 820,
  [2769] = 1888,
  [2770] = 776,
  [2771] = 807,
  [2772] = 839,
  [2773] = 1815,
  [2774] = 804,
  [2775] = 2719,
  [2776] = 2708,
  [2777] = 2777,
  [2778] = 2778,
  [2779] = 2779,
  [2780] = 2780,
  [2781] = 2781,
  [2782] = 2782,
  [2783] = 2783,
  [2784] = 2784,
  [2785] = 2779,
  [2786] = 2786,
  [2787] = 2787,
  [2788] = 2788,
  [2789] = 1674,
  [2790] = 1675,
  [2791] = 2791,
  [2792] = 2792,
  [2793] = 2787,
  [2794] = 2794,
  [2795] = 2795,
  [2796] = 2796,
  [2797] = 1694,
  [2798] = 1696,
  [2799] = 1697,
  [2800] = 1698,
  [2801] = 1699,
  [2802] = 1687,
  [2803] = 1680,
  [2804] = 1681,
  [2805] = 1673,
  [2806] = 1686,
  [2807] = 2807,
  [2808] = 2796,
  [2809] = 2809,
  [2810] = 2777,
  [2811] = 2782,
  [2812] = 2812,
  [2813] = 2777,
  [2814] = 2778,
  [2815] = 1676,
  [2816] = 2787,
  [2817] = 1702,
  [2818] = 1678,
  [2819] = 1679,
  [2820] = 2820,
  [2821] = 1625,
  [2822] = 1695,
  [2823] = 2823,
  [2824] = 2792,
  [2825] = 2787,
  [2826] = 2782,
  [2827] = 2807,
  [2828] = 2778,
  [2829] = 2779,
  [2830] = 2820,
  [2831] = 811,
  [2832] = 2788,
  [2833] = 2795,
  [2834] = 2796,
  [2835] = 2792,
  [2836] = 816,
  [2837] = 2780,
  [2838] = 2781,
  [2839] = 822,
  [2840] = 2249,
  [2841] = 2796,
  [2842] = 799,
  [2843] = 2780,
  [2844] = 2781,
  [2845] = 2249,
  [2846] = 1694,
  [2847] = 1696,
  [2848] = 1697,
  [2849] = 1698,
  [2850] = 1699,
  [2851] = 1687,
  [2852] = 1680,
  [2853] = 1681,
  [2854] = 2653,
  [2855] = 1673,
  [2856] = 2525,
  [2857] = 1695,
  [2858] = 2526,
  [2859] = 811,
  [2860] = 1676,
  [2861] = 1675,
  [2862] = 1686,
  [2863] = 1674,
  [2864] = 1542,
  [2865] = 1686,
  [2866] = 1673,
  [2867] = 1676,
  [2868] = 1702,
  [2869] = 1678,
  [2870] = 1679,
  [2871] = 2871,
  [2872] = 1674,
  [2873] = 1675,
  [2874] = 2127,
  [2875] = 1681,
  [2876] = 1544,
  [2877] = 1694,
  [2878] = 1695,
  [2879] = 1696,
  [2880] = 1697,
  [2881] = 1698,
  [2882] = 1699,
  [2883] = 1702,
  [2884] = 1678,
  [2885] = 1679,
  [2886] = 1687,
  [2887] = 1680,
  [2888] = 1533,
  [2889] = 1681,
  [2890] = 2237,
  [2891] = 1625,
  [2892] = 2244,
  [2893] = 2245,
  [2894] = 1674,
  [2895] = 1675,
  [2896] = 2247,
  [2897] = 2897,
  [2898] = 2589,
  [2899] = 2151,
  [2900] = 2153,
  [2901] = 1695,
  [2902] = 1496,
  [2903] = 2903,
  [2904] = 2155,
  [2905] = 2166,
  [2906] = 2172,
  [2907] = 2903,
  [2908] = 1694,
  [2909] = 1674,
  [2910] = 1507,
  [2911] = 1696,
  [2912] = 1697,
  [2913] = 1698,
  [2914] = 1699,
  [2915] = 1687,
  [2916] = 1680,
  [2917] = 2180,
  [2918] = 1673,
  [2919] = 2542,
  [2920] = 2182,
  [2921] = 2186,
  [2922] = 2903,
  [2923] = 2187,
  [2924] = 1695,
  [2925] = 2164,
  [2926] = 2177,
  [2927] = 2927,
  [2928] = 1675,
  [2929] = 2903,
  [2930] = 1503,
  [2931] = 808,
  [2932] = 2181,
  [2933] = 2196,
  [2934] = 1676,
  [2935] = 2521,
  [2936] = 1876,
  [2937] = 2230,
  [2938] = 2150,
  [2939] = 1702,
  [2940] = 2152,
  [2941] = 1678,
  [2942] = 1679,
  [2943] = 2903,
  [2944] = 2944,
  [2945] = 2146,
  [2946] = 2946,
  [2947] = 2947,
  [2948] = 1878,
  [2949] = 1857,
  [2950] = 797,
  [2951] = 786,
  [2952] = 787,
  [2953] = 790,
  [2954] = 794,
  [2955] = 2188,
  [2956] = 2228,
  [2957] = 815,
  [2958] = 819,
  [2959] = 821,
  [2960] = 2232,
  [2961] = 2147,
  [2962] = 2140,
  [2963] = 2238,
  [2964] = 2964,
  [2965] = 2965,
  [2966] = 2966,
  [2967] = 2190,
  [2968] = 2191,
  [2969] = 2969,
  [2970] = 1870,
  [2971] = 1686,
  [2972] = 2080,
  [2973] = 1697,
  [2974] = 1698,
  [2975] = 1699,
  [2976] = 1702,
  [2977] = 1678,
  [2978] = 1679,
  [2979] = 1687,
  [2980] = 1680,
  [2981] = 1681,
  [2982] = 1673,
  [2983] = 1686,
  [2984] = 2594,
  [2985] = 1680,
  [2986] = 2054,
  [2987] = 1681,
  [2988] = 835,
  [2989] = 1673,
  [2990] = 836,
  [2991] = 2991,
  [2992] = 2992,
  [2993] = 847,
  [2994] = 2071,
  [2995] = 2072,
  [2996] = 2622,
  [2997] = 2538,
  [2998] = 2629,
  [2999] = 2073,
  [3000] = 2074,
  [3001] = 2075,
  [3002] = 2076,
  [3003] = 785,
  [3004] = 2077,
  [3005] = 2078,
  [3006] = 1999,
  [3007] = 3007,
  [3008] = 840,
  [3009] = 1674,
  [3010] = 2008,
  [3011] = 2640,
  [3012] = 1694,
  [3013] = 1674,
  [3014] = 2010,
  [3015] = 1696,
  [3016] = 1697,
  [3017] = 2006,
  [3018] = 2081,
  [3019] = 2101,
  [3020] = 2046,
  [3021] = 1698,
  [3022] = 1699,
  [3023] = 1687,
  [3024] = 3024,
  [3025] = 841,
  [3026] = 816,
  [3027] = 2105,
  [3028] = 3028,
  [3029] = 1680,
  [3030] = 2001,
  [3031] = 2002,
  [3032] = 842,
  [3033] = 1681,
  [3034] = 1673,
  [3035] = 2007,
  [3036] = 843,
  [3037] = 844,
  [3038] = 2013,
  [3039] = 1695,
  [3040] = 2603,
  [3041] = 2057,
  [3042] = 2586,
  [3043] = 3043,
  [3044] = 1675,
  [3045] = 2014,
  [3046] = 2106,
  [3047] = 1686,
  [3048] = 1675,
  [3049] = 1674,
  [3050] = 1675,
  [3051] = 793,
  [3052] = 3052,
  [3053] = 2079,
  [3054] = 3054,
  [3055] = 3055,
  [3056] = 2011,
  [3057] = 835,
  [3058] = 2044,
  [3059] = 2047,
  [3060] = 2048,
  [3061] = 2082,
  [3062] = 840,
  [3063] = 3063,
  [3064] = 1698,
  [3065] = 836,
  [3066] = 3066,
  [3067] = 2050,
  [3068] = 2633,
  [3069] = 3043,
  [3070] = 2636,
  [3071] = 2637,
  [3072] = 2605,
  [3073] = 2643,
  [3074] = 2645,
  [3075] = 2616,
  [3076] = 799,
  [3077] = 2593,
  [3078] = 3078,
  [3079] = 2058,
  [3080] = 2059,
  [3081] = 1676,
  [3082] = 1702,
  [3083] = 1678,
  [3084] = 1679,
  [3085] = 2060,
  [3086] = 2089,
  [3087] = 2090,
  [3088] = 2098,
  [3089] = 2043,
  [3090] = 2102,
  [3091] = 1695,
  [3092] = 2103,
  [3093] = 2083,
  [3094] = 2610,
  [3095] = 792,
  [3096] = 2638,
  [3097] = 1695,
  [3098] = 3078,
  [3099] = 1686,
  [3100] = 1699,
  [3101] = 2648,
  [3102] = 1702,
  [3103] = 2623,
  [3104] = 817,
  [3105] = 1678,
  [3106] = 1679,
  [3107] = 2649,
  [3108] = 2639,
  [3109] = 2646,
  [3110] = 3110,
  [3111] = 2066,
  [3112] = 2635,
  [3113] = 2653,
  [3114] = 2614,
  [3115] = 2619,
  [3116] = 2625,
  [3117] = 847,
  [3118] = 1687,
  [3119] = 810,
  [3120] = 2056,
  [3121] = 2057,
  [3122] = 1948,
  [3123] = 1676,
  [3124] = 3124,
  [3125] = 2620,
  [3126] = 3126,
  [3127] = 3127,
  [3128] = 2783,
  [3129] = 2112,
  [3130] = 1694,
  [3131] = 2000,
  [3132] = 1676,
  [3133] = 1696,
  [3134] = 2627,
  [3135] = 2642,
  [3136] = 1694,
  [3137] = 2020,
  [3138] = 1696,
  [3139] = 2031,
  [3140] = 1697,
  [3141] = 2034,
  [3142] = 2621,
  [3143] = 2035,
  [3144] = 2036,
  [3145] = 2042,
  [3146] = 2056,
  [3147] = 3147,
  [3148] = 3148,
  [3149] = 3149,
  [3150] = 808,
  [3151] = 3151,
  [3152] = 2134,
  [3153] = 797,
  [3154] = 3154,
  [3155] = 3155,
  [3156] = 3054,
  [3157] = 2784,
  [3158] = 2666,
  [3159] = 3159,
  [3160] = 3160,
  [3161] = 3161,
  [3162] = 3148,
  [3163] = 2128,
  [3164] = 1676,
  [3165] = 1694,
  [3166] = 1696,
  [3167] = 3147,
  [3168] = 1698,
  [3169] = 1699,
  [3170] = 1702,
  [3171] = 1678,
  [3172] = 1679,
  [3173] = 1687,
  [3174] = 1680,
  [3175] = 1681,
  [3176] = 2129,
  [3177] = 2133,
  [3178] = 1673,
  [3179] = 3055,
  [3180] = 1686,
  [3181] = 3181,
  [3182] = 1674,
  [3183] = 1675,
  [3184] = 1695,
  [3185] = 2130,
  [3186] = 2124,
  [3187] = 2126,
  [3188] = 3188,
  [3189] = 1697,
  [3190] = 822,
  [3191] = 3159,
  [3192] = 3192,
  [3193] = 808,
  [3194] = 3160,
  [3195] = 2128,
  [3196] = 3196,
  [3197] = 3197,
  [3198] = 3197,
  [3199] = 811,
  [3200] = 3200,
  [3201] = 3201,
  [3202] = 2525,
  [3203] = 2126,
  [3204] = 797,
  [3205] = 3205,
  [3206] = 3188,
  [3207] = 816,
  [3208] = 3197,
  [3209] = 3209,
  [3210] = 1695,
  [3211] = 1675,
  [3212] = 2130,
  [3213] = 3213,
  [3214] = 3209,
  [3215] = 3215,
  [3216] = 3216,
  [3217] = 1674,
  [3218] = 2134,
  [3219] = 2130,
  [3220] = 3201,
  [3221] = 3221,
  [3222] = 799,
  [3223] = 1674,
  [3224] = 3224,
  [3225] = 3225,
  [3226] = 3209,
  [3227] = 1675,
  [3228] = 2124,
  [3229] = 3201,
  [3230] = 2249,
  [3231] = 3215,
  [3232] = 2129,
  [3233] = 3233,
  [3234] = 2133,
  [3235] = 3235,
  [3236] = 3236,
  [3237] = 3161,
  [3238] = 3238,
  [3239] = 3239,
  [3240] = 3240,
  [3241] = 3240,
  [3242] = 2134,
  [3243] = 3243,
  [3244] = 3244,
  [3245] = 1695,
  [3246] = 3213,
  [3247] = 3247,
  [3248] = 3215,
  [3249] = 3249,
  [3250] = 3250,
  [3251] = 3251,
  [3252] = 3252,
  [3253] = 3253,
  [3254] = 3254,
  [3255] = 3255,
  [3256] = 3252,
  [3257] = 3257,
  [3258] = 3258,
  [3259] = 836,
  [3260] = 3260,
  [3261] = 3250,
  [3262] = 3262,
  [3263] = 3155,
  [3264] = 3254,
  [3265] = 811,
  [3266] = 2134,
  [3267] = 3267,
  [3268] = 3268,
  [3269] = 799,
  [3270] = 3270,
  [3271] = 3258,
  [3272] = 2249,
  [3273] = 3273,
  [3274] = 3274,
  [3275] = 822,
  [3276] = 2249,
  [3277] = 3258,
  [3278] = 3255,
  [3279] = 847,
  [3280] = 3252,
  [3281] = 3281,
  [3282] = 3154,
  [3283] = 840,
  [3284] = 2130,
  [3285] = 3285,
  [3286] = 816,
  [3287] = 3252,
  [3288] = 3274,
  [3289] = 3251,
  [3290] = 3181,
  [3291] = 3285,
  [3292] = 3274,
  [3293] = 3258,
  [3294] = 3274,
  [3295] = 835,
  [3296] = 3296,
  [3297] = 842,
  [3298] = 2130,
  [3299] = 2124,
  [3300] = 3300,
  [3301] = 2126,
  [3302] = 3302,
  [3303] = 3303,
  [3304] = 3304,
  [3305] = 3305,
  [3306] = 3306,
  [3307] = 3307,
  [3308] = 3247,
  [3309] = 3303,
  [3310] = 3310,
  [3311] = 3311,
  [3312] = 2130,
  [3313] = 841,
  [3314] = 3296,
  [3315] = 3315,
  [3316] = 3316,
  [3317] = 3317,
  [3318] = 2128,
  [3319] = 2134,
  [3320] = 3320,
  [3321] = 2128,
  [3322] = 2249,
  [3323] = 3216,
  [3324] = 3324,
  [3325] = 2126,
  [3326] = 3326,
  [3327] = 2133,
  [3328] = 3328,
  [3329] = 2134,
  [3330] = 3326,
  [3331] = 3331,
  [3332] = 3332,
  [3333] = 843,
  [3334] = 2134,
  [3335] = 3310,
  [3336] = 810,
  [3337] = 2129,
  [3338] = 3205,
  [3339] = 2133,
  [3340] = 3340,
  [3341] = 3341,
  [3342] = 2130,
  [3343] = 3243,
  [3344] = 844,
  [3345] = 3311,
  [3346] = 2794,
  [3347] = 2124,
  [3348] = 3348,
  [3349] = 2129,
  [3350] = 3350,
  [3351] = 3351,
  [3352] = 3352,
  [3353] = 3353,
  [3354] = 2249,
  [3355] = 3355,
  [3356] = 3356,
  [3357] = 3357,
  [3358] = 3358,
  [3359] = 3316,
  [3360] = 3360,
  [3361] = 3307,
  [3362] = 3362,
  [3363] = 3363,
  [3364] = 3364,
  [3365] = 3365,
  [3366] = 3363,
  [3367] = 3367,
  [3368] = 3368,
  [3369] = 3369,
  [3370] = 3370,
  [3371] = 3371,
  [3372] = 3372,
  [3373] = 3373,
  [3374] = 3374,
  [3375] = 3375,
  [3376] = 3376,
  [3377] = 3377,
  [3378] = 3378,
  [3379] = 3372,
  [3380] = 3380,
  [3381] = 3381,
  [3382] = 3351,
  [3383] = 2871,
  [3384] = 3384,
  [3385] = 3281,
  [3386] = 3386,
  [3387] = 3387,
  [3388] = 3388,
  [3389] = 3389,
  [3390] = 3381,
  [3391] = 3351,
  [3392] = 3367,
  [3393] = 3384,
  [3394] = 3372,
  [3395] = 3384,
  [3396] = 3371,
  [3397] = 3397,
  [3398] = 2237,
  [3399] = 2244,
  [3400] = 3257,
  [3401] = 3401,
  [3402] = 2130,
  [3403] = 3403,
  [3404] = 3371,
  [3405] = 3405,
  [3406] = 3384,
  [3407] = 3407,
  [3408] = 3355,
  [3409] = 3376,
  [3410] = 3355,
  [3411] = 3378,
  [3412] = 3363,
  [3413] = 3368,
  [3414] = 3414,
  [3415] = 3415,
  [3416] = 3407,
  [3417] = 3417,
  [3418] = 3418,
  [3419] = 3362,
  [3420] = 3420,
  [3421] = 3421,
  [3422] = 3420,
  [3423] = 3423,
  [3424] = 3369,
  [3425] = 3368,
  [3426] = 3401,
  [3427] = 3427,
  [3428] = 3428,
  [3429] = 3415,
  [3430] = 3421,
  [3431] = 3352,
  [3432] = 3370,
  [3433] = 3433,
  [3434] = 3427,
  [3435] = 3352,
  [3436] = 3423,
  [3437] = 3376,
  [3438] = 3438,
  [3439] = 3375,
  [3440] = 3370,
  [3441] = 3369,
  [3442] = 3442,
  [3443] = 2186,
  [3444] = 3444,
  [3445] = 3414,
  [3446] = 3446,
  [3447] = 3401,
  [3448] = 3362,
  [3449] = 2249,
  [3450] = 3415,
  [3451] = 3421,
  [3452] = 3442,
  [3453] = 3453,
  [3454] = 2249,
  [3455] = 3455,
  [3456] = 3456,
  [3457] = 3457,
  [3458] = 3458,
  [3459] = 2245,
  [3460] = 3446,
  [3461] = 3456,
  [3462] = 3455,
  [3463] = 3427,
  [3464] = 3456,
  [3465] = 3378,
  [3466] = 3367,
  [3467] = 3407,
  [3468] = 3468,
  [3469] = 3423,
  [3470] = 3375,
  [3471] = 2134,
  [3472] = 3472,
  [3473] = 3442,
  [3474] = 3389,
  [3475] = 3389,
  [3476] = 3476,
  [3477] = 3377,
  [3478] = 3351,
  [3479] = 3372,
  [3480] = 3480,
  [3481] = 3341,
  [3482] = 3482,
  [3483] = 3483,
  [3484] = 3484,
  [3485] = 3485,
  [3486] = 3486,
  [3487] = 3487,
  [3488] = 3488,
  [3489] = 3489,
  [3490] = 3490,
  [3491] = 2249,
  [3492] = 3492,
  [3493] = 3328,
  [3494] = 3494,
  [3495] = 3495,
  [3496] = 3496,
  [3497] = 3497,
  [3498] = 3498,
  [3499] = 3499,
  [3500] = 3500,
  [3501] = 3501,
  [3502] = 3502,
  [3503] = 3503,
  [3504] = 3504,
  [3505] = 3505,
  [3506] = 3506,
  [3507] = 811,
  [3508] = 3508,
  [3509] = 3509,
  [3510] = 3510,
  [3511] = 3511,
  [3512] = 3512,
  [3513] = 3513,
  [3514] = 3514,
  [3515] = 3515,
  [3516] = 3516,
  [3517] = 3517,
  [3518] = 3518,
  [3519] = 3519,
  [3520] = 3520,
  [3521] = 3521,
  [3522] = 3522,
  [3523] = 3523,
  [3524] = 3524,
  [3525] = 3525,
  [3526] = 3526,
  [3527] = 3527,
  [3528] = 3528,
  [3529] = 3529,
  [3530] = 3530,
  [3531] = 3523,
  [3532] = 3517,
  [3533] = 3458,
  [3534] = 3534,
  [3535] = 3315,
  [3536] = 3536,
  [3537] = 3517,
  [3538] = 3538,
  [3539] = 3539,
  [3540] = 785,
  [3541] = 793,
  [3542] = 3364,
  [3543] = 3543,
  [3544] = 3544,
  [3545] = 3545,
  [3546] = 3546,
  [3547] = 3547,
  [3548] = 3548,
  [3549] = 3549,
  [3550] = 3550,
  [3551] = 3551,
  [3552] = 3552,
  [3553] = 3553,
  [3554] = 3554,
  [3555] = 3498,
  [3556] = 3556,
  [3557] = 3499,
  [3558] = 3500,
  [3559] = 3559,
  [3560] = 3560,
  [3561] = 3561,
  [3562] = 3519,
  [3563] = 3563,
  [3564] = 3564,
  [3565] = 3504,
  [3566] = 3566,
  [3567] = 3567,
  [3568] = 3568,
  [3569] = 3569,
  [3570] = 3570,
  [3571] = 3571,
  [3572] = 3484,
  [3573] = 3573,
  [3574] = 3574,
  [3575] = 3575,
  [3576] = 3576,
  [3577] = 3577,
  [3578] = 3457,
  [3579] = 3365,
  [3580] = 3483,
  [3581] = 3581,
  [3582] = 3582,
  [3583] = 3487,
  [3584] = 3576,
  [3585] = 3585,
  [3586] = 3586,
  [3587] = 3587,
  [3588] = 3588,
  [3589] = 3576,
  [3590] = 3590,
  [3591] = 3494,
  [3592] = 3592,
  [3593] = 3593,
  [3594] = 3501,
  [3595] = 3480,
  [3596] = 3596,
  [3597] = 3512,
  [3598] = 3598,
  [3599] = 3599,
  [3600] = 3600,
  [3601] = 3545,
  [3602] = 3602,
  [3603] = 3603,
  [3604] = 3604,
  [3605] = 3576,
  [3606] = 3508,
  [3607] = 3492,
  [3608] = 3608,
  [3609] = 3609,
  [3610] = 3610,
  [3611] = 3576,
  [3612] = 3612,
  [3613] = 3613,
  [3614] = 3614,
  [3615] = 3615,
  [3616] = 3576,
  [3617] = 3617,
  [3618] = 3576,
  [3619] = 3619,
  [3620] = 3620,
  [3621] = 3576,
  [3622] = 3622,
  [3623] = 3623,
  [3624] = 3624,
  [3625] = 3576,
  [3626] = 3576,
  [3627] = 3627,
  [3628] = 3628,
  [3629] = 3629,
  [3630] = 3630,
  [3631] = 3631,
  [3632] = 3632,
  [3633] = 3633,
  [3634] = 3634,
  [3635] = 3635,
  [3636] = 3636,
  [3637] = 3637,
  [3638] = 3638,
  [3639] = 3639,
  [3640] = 3640,
  [3641] = 3641,
  [3642] = 3642,
  [3643] = 3643,
  [3644] = 3644,
  [3645] = 3645,
  [3646] = 2235,
  [3647] = 3647,
  [3648] = 3489,
  [3649] = 3649,
  [3650] = 3496,
  [3651] = 3651,
  [3652] = 3506,
  [3653] = 3653,
  [3654] = 3486,
  [3655] = 3655,
  [3656] = 3656,
  [3657] = 3657,
  [3658] = 3658,
  [3659] = 3659,
  [3660] = 3660,
  [3661] = 3661,
  [3662] = 3662,
  [3663] = 3663,
  [3664] = 3664,
  [3665] = 3665,
  [3666] = 3666,
  [3667] = 3667,
  [3668] = 3576,
  [3669] = 3669,
  [3670] = 1542,
  [3671] = 1533,
  [3672] = 1544,
  [3673] = 3503,
  [3674] = 3674,
  [3675] = 3488,
  [3676] = 3676,
  [3677] = 3677,
  [3678] = 3485,
  [3679] = 3679,
  [3680] = 3497,
  [3681] = 3681,
  [3682] = 3682,
  [3683] = 3683,
  [3684] = 3684,
  [3685] = 3685,
  [3686] = 3686,
  [3687] = 3687,
  [3688] = 3688,
  [3689] = 3689,
  [3690] = 3690,
  [3691] = 3691,
  [3692] = 3692,
  [3693] = 3693,
  [3694] = 3490,
  [3695] = 3695,
  [3696] = 3696,
  [3697] = 3697,
  [3698] = 3698,
  [3699] = 3699,
  [3700] = 3700,
  [3701] = 3701,
  [3702] = 3702,
  [3703] = 3703,
  [3704] = 3704,
  [3705] = 3705,
  [3706] = 3706,
  [3707] = 3707,
  [3708] = 3708,
  [3709] = 3709,
  [3710] = 3710,
  [3711] = 3711,
  [3712] = 3712,
  [3713] = 3713,
  [3714] = 3714,
  [3715] = 3715,
  [3716] = 3716,
  [3717] = 3717,
  [3718] = 3718,
  [3719] = 3719,
  [3720] = 3720,
  [3721] = 3721,
  [3722] = 3722,
  [3723] = 3723,
  [3724] = 3724,
  [3725] = 3725,
  [3726] = 3726,
  [3727] = 3727,
  [3728] = 3728,
  [3729] = 3729,
  [3730] = 3730,
  [3731] = 3495,
  [3732] = 3732,
  [3733] = 3733,
  [3734] = 3734,
  [3735] = 3735,
  [3736] = 3736,
  [3737] = 3737,
  [3738] = 3738,
  [3739] = 3518,
  [3740] = 3740,
  [3741] = 815,
  [3742] = 2646,
  [3743] = 3743,
  [3744] = 3730,
  [3745] = 3457,
  [3746] = 821,
  [3747] = 3747,
  [3748] = 3748,
  [3749] = 3365,
  [3750] = 3750,
  [3751] = 3751,
  [3752] = 3752,
  [3753] = 3753,
  [3754] = 3754,
  [3755] = 3755,
  [3756] = 3756,
  [3757] = 3757,
  [3758] = 794,
  [3759] = 815,
  [3760] = 821,
  [3761] = 3754,
  [3762] = 3762,
  [3763] = 3763,
  [3764] = 3764,
  [3765] = 3765,
  [3766] = 3766,
  [3767] = 3767,
  [3768] = 3740,
  [3769] = 3769,
  [3770] = 3736,
  [3771] = 1625,
  [3772] = 3505,
  [3773] = 3773,
  [3774] = 3774,
  [3775] = 3775,
  [3776] = 3732,
  [3777] = 3738,
  [3778] = 3740,
  [3779] = 3779,
  [3780] = 3780,
  [3781] = 3781,
  [3782] = 2594,
  [3783] = 840,
  [3784] = 3784,
  [3785] = 3785,
  [3786] = 3757,
  [3787] = 3787,
  [3788] = 835,
  [3789] = 3789,
  [3790] = 3732,
  [3791] = 3779,
  [3792] = 3792,
  [3793] = 3793,
  [3794] = 3730,
  [3795] = 3751,
  [3796] = 3735,
  [3797] = 3797,
  [3798] = 3738,
  [3799] = 3735,
  [3800] = 3547,
  [3801] = 3762,
  [3802] = 3521,
  [3803] = 3551,
  [3804] = 3762,
  [3805] = 794,
  [3806] = 3754,
  [3807] = 3785,
  [3808] = 3757,
  [3809] = 3730,
  [3810] = 3751,
  [3811] = 3762,
  [3812] = 3779,
  [3813] = 3779,
  [3814] = 3730,
  [3815] = 3751,
  [3816] = 3502,
  [3817] = 3762,
  [3818] = 3779,
  [3819] = 3793,
  [3820] = 3730,
  [3821] = 3751,
  [3822] = 3516,
  [3823] = 3793,
  [3824] = 847,
  [3825] = 3779,
  [3826] = 3779,
  [3827] = 836,
  [3828] = 3751,
  [3829] = 3762,
  [3830] = 3757,
  [3831] = 3730,
  [3832] = 3751,
  [3833] = 3762,
  [3834] = 3793,
  [3835] = 3719,
  [3836] = 794,
  [3837] = 3837,
  [3838] = 815,
  [3839] = 821,
  [3840] = 2653,
  [3841] = 3634,
  [3842] = 3725,
  [3843] = 3781,
  [3844] = 3844,
  [3845] = 3631,
  [3846] = 3846,
  [3847] = 3639,
  [3848] = 3633,
  [3849] = 3849,
  [3850] = 3844,
  [3851] = 3640,
  [3852] = 3773,
  [3853] = 3728,
  [3854] = 3854,
  [3855] = 3638,
  [3856] = 3765,
  [3857] = 3857,
  [3858] = 3858,
  [3859] = 3849,
  [3860] = 3860,
  [3861] = 3752,
  [3862] = 3849,
  [3863] = 3644,
  [3864] = 3864,
  [3865] = 3784,
  [3866] = 3849,
  [3867] = 3867,
  [3868] = 3868,
  [3869] = 3869,
  [3870] = 3641,
  [3871] = 3630,
  [3872] = 3642,
  [3873] = 3849,
  [3874] = 3792,
  [3875] = 3849,
  [3876] = 3632,
  [3877] = 3627,
  [3878] = 3878,
  [3879] = 3849,
  [3880] = 3637,
  [3881] = 3844,
  [3882] = 3635,
  [3883] = 3553,
  [3884] = 3849,
  [3885] = 3885,
  [3886] = 3636,
  [3887] = 3849,
  [3888] = 3769,
  [3889] = 3849,
  [3890] = 3724,
  [3891] = 3849,
  [3892] = 3892,
  [3893] = 3893,
  [3894] = 3894,
  [3895] = 3895,
  [3896] = 3896,
  [3897] = 3897,
  [3898] = 3898,
  [3899] = 3899,
  [3900] = 3900,
  [3901] = 3901,
  [3902] = 3893,
  [3903] = 3892,
  [3904] = 3894,
  [3905] = 3895,
  [3906] = 3896,
  [3907] = 3897,
  [3908] = 3898,
  [3909] = 3899,
  [3910] = 3900,
  [3911] = 3901,
  [3912] = 3893,
  [3913] = 3892,
  [3914] = 3894,
  [3915] = 3915,
  [3916] = 3916,
  [3917] = 3917,
  [3918] = 3918,
  [3919] = 3509,
  [3920] = 3920,
  [3921] = 3921,
  [3922] = 3922,
  [3923] = 3897,
  [3924] = 3895,
  [3925] = 3925,
  [3926] = 3896,
  [3927] = 3927,
  [3928] = 3897,
  [3929] = 3898,
  [3930] = 3899,
  [3931] = 3900,
  [3932] = 3898,
  [3933] = 3901,
  [3934] = 3893,
  [3935] = 3899,
  [3936] = 3892,
  [3937] = 3894,
  [3938] = 3938,
  [3939] = 3900,
  [3940] = 3940,
  [3941] = 3941,
  [3942] = 3925,
  [3943] = 3553,
  [3944] = 3895,
  [3945] = 3945,
  [3946] = 3946,
  [3947] = 3925,
  [3948] = 3948,
  [3949] = 3918,
  [3950] = 3950,
  [3951] = 3896,
  [3952] = 3952,
  [3953] = 3895,
  [3954] = 3925,
  [3955] = 3897,
  [3956] = 3925,
  [3957] = 3898,
  [3958] = 3901,
  [3959] = 3899,
  [3960] = 3901,
  [3961] = 3901,
  [3962] = 3917,
  [3963] = 3893,
  [3964] = 3892,
  [3965] = 3893,
  [3966] = 3892,
  [3967] = 3894,
  [3968] = 3895,
  [3969] = 3925,
  [3970] = 3896,
  [3971] = 3897,
  [3972] = 3896,
  [3973] = 3898,
  [3974] = 3899,
  [3975] = 3900,
  [3976] = 3901,
  [3977] = 3893,
  [3978] = 3892,
  [3979] = 3894,
  [3980] = 3917,
  [3981] = 3917,
  [3982] = 3917,
  [3983] = 3894,
  [3984] = 3925,
  [3985] = 3917,
  [3986] = 3897,
  [3987] = 3895,
  [3988] = 3513,
  [3989] = 3925,
  [3990] = 3917,
  [3991] = 3991,
  [3992] = 3511,
  [3993] = 3896,
  [3994] = 3897,
  [3995] = 3898,
  [3996] = 3917,
  [3997] = 3917,
  [3998] = 3899,
  [3999] = 3945,
  [4000] = 3900,
  [4001] = 3901,
  [4002] = 3893,
  [4003] = 3892,
  [4004] = 3894,
  [4005] = 4005,
  [4006] = 3945,
  [4007] = 3898,
  [4008] = 3899,
  [4009] = 4009,
  [4010] = 3895,
  [4011] = 3925,
  [4012] = 3896,
  [4013] = 3897,
  [4014] = 3898,
  [4015] = 3899,
  [4016] = 3900,
  [4017] = 3901,
  [4018] = 3893,
  [4019] = 3892,
  [4020] = 3894,
  [4021] = 3900,
  [4022] = 3896,
  [4023] = 4023,
  [4024] = 3901,
  [4025] = 3917,
  [4026] = 3895,
  [4027] = 3925,
  [4028] = 3896,
  [4029] = 3897,
  [4030] = 3898,
  [4031] = 3899,
  [4032] = 3900,
  [4033] = 3901,
  [4034] = 3893,
  [4035] = 3892,
  [4036] = 3894,
  [4037] = 3893,
  [4038] = 4038,
  [4039] = 4039,
  [4040] = 3892,
  [4041] = 3894,
  [4042] = 3895,
  [4043] = 3925,
  [4044] = 4044,
  [4045] = 4045,
  [4046] = 4046,
  [4047] = 4047,
  [4048] = 4048,
  [4049] = 3896,
  [4050] = 4050,
  [4051] = 4051,
  [4052] = 3897,
  [4053] = 3895,
  [4054] = 3898,
  [4055] = 3899,
  [4056] = 3900,
  [4057] = 3900,
  [4058] = 4058,
  [4059] = 3503,
  [4060] = 4060,
  [4061] = 3551,
  [4062] = 4062,
  [4063] = 4063,
  [4064] = 3516,
  [4065] = 4065,
  [4066] = 4065,
  [4067] = 835,
  [4068] = 4068,
  [4069] = 4069,
  [4070] = 3521,
  [4071] = 836,
  [4072] = 4058,
  [4073] = 4073,
  [4074] = 840,
  [4075] = 4073,
  [4076] = 847,
  [4077] = 4077,
  [4078] = 4078,
  [4079] = 4073,
  [4080] = 3547,
  [4081] = 4081,
  [4082] = 3946,
  [4083] = 786,
  [4084] = 3927,
  [4085] = 4045,
  [4086] = 844,
  [4087] = 4087,
  [4088] = 4088,
  [4089] = 786,
  [4090] = 4081,
  [4091] = 4050,
  [4092] = 4088,
  [4093] = 4088,
  [4094] = 4088,
  [4095] = 787,
  [4096] = 787,
  [4097] = 4087,
  [4098] = 4088,
  [4099] = 3920,
  [4100] = 790,
  [4101] = 4101,
  [4102] = 4102,
  [4103] = 810,
  [4104] = 4104,
  [4105] = 4102,
  [4106] = 4088,
  [4107] = 4005,
  [4108] = 4108,
  [4109] = 790,
  [4110] = 4088,
  [4111] = 4111,
  [4112] = 4088,
  [4113] = 4088,
  [4114] = 4114,
  [4115] = 843,
  [4116] = 4046,
  [4117] = 4047,
  [4118] = 801,
  [4119] = 4119,
  [4120] = 4048,
  [4121] = 819,
  [4122] = 4104,
  [4123] = 3948,
  [4124] = 3952,
  [4125] = 4125,
  [4126] = 4126,
  [4127] = 839,
  [4128] = 4108,
  [4129] = 4129,
  [4130] = 3921,
  [4131] = 4131,
  [4132] = 4132,
  [4133] = 4125,
  [4134] = 4101,
  [4135] = 4104,
  [4136] = 4119,
  [4137] = 4137,
  [4138] = 4111,
  [4139] = 4139,
  [4140] = 4140,
  [4141] = 4141,
  [4142] = 3916,
  [4143] = 4104,
  [4144] = 841,
  [4145] = 4104,
  [4146] = 4119,
  [4147] = 4119,
  [4148] = 3922,
  [4149] = 4126,
  [4150] = 4137,
  [4151] = 4088,
  [4152] = 4051,
  [4153] = 4101,
  [4154] = 4119,
  [4155] = 819,
  [4156] = 4156,
  [4157] = 4157,
  [4158] = 4129,
  [4159] = 4140,
  [4160] = 4044,
  [4161] = 4137,
  [4162] = 4038,
  [4163] = 4039,
  [4164] = 842,
  [4165] = 4009,
  [4166] = 4166,
  [4167] = 4166,
  [4168] = 4168,
  [4169] = 4166,
  [4170] = 4166,
  [4171] = 1655,
  [4172] = 2783,
  [4173] = 4166,
  [4174] = 4166,
  [4175] = 4166,
  [4176] = 4166,
  [4177] = 817,
  [4178] = 4166,
  [4179] = 4166,
  [4180] = 4166,
  [4181] = 4168,
  [4182] = 4166,
  [4183] = 4168,
  [4184] = 4166,
  [4185] = 4166,
  [4186] = 4166,
  [4187] = 4166,
  [4188] = 4166,
  [4189] = 4166,
  [4190] = 792,
  [4191] = 2784,
  [4192] = 4192,
  [4193] = 4193,
  [4194] = 4194,
  [4195] = 4195,
  [4196] = 2542,
  [4197] = 4197,
  [4198] = 4198,
  [4199] = 4192,
  [4200] = 2056,
  [4201] = 1688,
  [4202] = 4202,
  [4203] = 4203,
  [4204] = 4204,
  [4205] = 4205,
  [4206] = 4193,
  [4207] = 4203,
  [4208] = 4203,
  [4209] = 4209,
  [4210] = 4192,
  [4211] = 4211,
  [4212] = 4212,
  [4213] = 4202,
  [4214] = 4214,
  [4215] = 4215,
  [4216] = 2589,
  [4217] = 4195,
  [4218] = 4214,
  [4219] = 4209,
  [4220] = 2057,
  [4221] = 4193,
  [4222] = 4211,
  [4223] = 4212,
  [4224] = 4198,
  [4225] = 4209,
  [4226] = 4214,
  [4227] = 4214,
  [4228] = 4228,
  [4229] = 4229,
  [4230] = 2783,
  [4231] = 2784,
  [4232] = 822,
  [4233] = 4233,
  [4234] = 4234,
  [4235] = 4233,
  [4236] = 2897,
  [4237] = 2946,
  [4238] = 2947,
  [4239] = 811,
  [4240] = 4234,
  [4241] = 4241,
  [4242] = 4233,
  [4243] = 4241,
  [4244] = 4244,
  [4245] = 4245,
  [4246] = 2964,
  [4247] = 4247,
  [4248] = 4248,
  [4249] = 2965,
  [4250] = 2966,
  [4251] = 4234,
  [4252] = 4241,
  [4253] = 1857,
  [4254] = 2662,
  [4255] = 2668,
  [4256] = 4256,
  [4257] = 4257,
  [4258] = 4256,
  [4259] = 4257,
  [4260] = 3181,
  [4261] = 2521,
  [4262] = 4256,
  [4263] = 4257,
  [4264] = 4264,
  [4265] = 2525,
  [4266] = 2656,
  [4267] = 4267,
  [4268] = 2655,
  [4269] = 4269,
  [4270] = 4269,
  [4271] = 4271,
  [4272] = 4272,
  [4273] = 4267,
  [4274] = 4256,
  [4275] = 4275,
  [4276] = 4256,
  [4277] = 4277,
  [4278] = 4267,
  [4279] = 4279,
  [4280] = 2661,
  [4281] = 1870,
  [4282] = 2589,
  [4283] = 2660,
  [4284] = 4257,
  [4285] = 2525,
  [4286] = 4286,
  [4287] = 4287,
  [4288] = 4288,
  [4289] = 4269,
  [4290] = 4256,
  [4291] = 4257,
  [4292] = 4256,
  [4293] = 4257,
  [4294] = 4294,
  [4295] = 4295,
  [4296] = 4296,
  [4297] = 2783,
  [4298] = 4298,
  [4299] = 4256,
  [4300] = 4300,
  [4301] = 4256,
  [4302] = 3154,
  [4303] = 4257,
  [4304] = 4304,
  [4305] = 2784,
  [4306] = 4306,
  [4307] = 4257,
  [4308] = 2542,
  [4309] = 2526,
  [4310] = 3155,
  [4311] = 2667,
  [4312] = 4257,
  [4313] = 2538,
  [4314] = 4314,
  [4315] = 2010,
  [4316] = 4316,
  [4317] = 4317,
  [4318] = 4318,
  [4319] = 2158,
  [4320] = 4320,
  [4321] = 4321,
  [4322] = 4322,
  [4323] = 4323,
  [4324] = 2197,
  [4325] = 4325,
  [4326] = 1999,
  [4327] = 4327,
  [4328] = 4328,
  [4329] = 4325,
  [4330] = 2047,
  [4331] = 4328,
  [4332] = 4325,
  [4333] = 4333,
  [4334] = 2160,
  [4335] = 4335,
  [4336] = 4336,
  [4337] = 4337,
  [4338] = 4338,
  [4339] = 4339,
  [4340] = 2044,
  [4341] = 4328,
  [4342] = 4342,
  [4343] = 4343,
  [4344] = 4344,
  [4345] = 4345,
  [4346] = 4346,
  [4347] = 4347,
  [4348] = 4348,
  [4349] = 4325,
  [4350] = 4350,
  [4351] = 3205,
  [4352] = 4352,
  [4353] = 4353,
  [4354] = 4354,
  [4355] = 4355,
  [4356] = 4356,
  [4357] = 2586,
  [4358] = 4358,
  [4359] = 4359,
  [4360] = 4328,
  [4361] = 2175,
  [4362] = 4328,
  [4363] = 4328,
  [4364] = 2263,
  [4365] = 4328,
  [4366] = 3216,
  [4367] = 4367,
  [4368] = 4368,
  [4369] = 4369,
  [4370] = 4370,
  [4371] = 4371,
  [4372] = 4372,
  [4373] = 4373,
  [4374] = 4374,
  [4375] = 4375,
  [4376] = 4376,
  [4377] = 4377,
  [4378] = 4378,
  [4379] = 4379,
  [4380] = 4380,
  [4381] = 4381,
  [4382] = 4382,
  [4383] = 4383,
  [4384] = 4384,
  [4385] = 2112,
  [4386] = 2209,
  [4387] = 4387,
  [4388] = 4388,
  [4389] = 4389,
  [4390] = 4390,
  [4391] = 4391,
  [4392] = 4392,
  [4393] = 4393,
  [4394] = 4394,
  [4395] = 4395,
  [4396] = 4396,
  [4397] = 4397,
  [4398] = 4398,
  [4399] = 4399,
  [4400] = 4400,
  [4401] = 4325,
  [4402] = 4325,
  [4403] = 4403,
  [4404] = 4404,
  [4405] = 4405,
  [4406] = 4406,
  [4407] = 4407,
  [4408] = 4408,
  [4409] = 808,
  [4410] = 4410,
  [4411] = 4411,
  [4412] = 4412,
  [4413] = 4325,
  [4414] = 797,
  [4415] = 3247,
  [4416] = 4325,
  [4417] = 4328,
  [4418] = 4328,
  [4419] = 4419,
  [4420] = 4420,
  [4421] = 4328,
  [4422] = 4422,
  [4423] = 4423,
  [4424] = 4325,
  [4425] = 4425,
  [4426] = 4426,
  [4427] = 3243,
  [4428] = 4428,
  [4429] = 4429,
  [4430] = 2241,
  [4431] = 2250,
  [4432] = 4325,
  [4433] = 4335,
  [4434] = 2173,
  [4435] = 2246,
  [4436] = 2262,
  [4437] = 2143,
  [4438] = 2144,
  [4439] = 2154,
  [4440] = 2008,
  [4441] = 4441,
  [4442] = 3251,
  [4443] = 4443,
  [4444] = 3285,
  [4445] = 2593,
  [4446] = 4446,
  [4447] = 4447,
  [4448] = 4448,
  [4449] = 2610,
  [4450] = 4450,
  [4451] = 2625,
  [4452] = 2666,
  [4453] = 2627,
  [4454] = 2642,
  [4455] = 816,
  [4456] = 2646,
  [4457] = 2614,
  [4458] = 2619,
  [4459] = 2620,
  [4460] = 4460,
  [4461] = 4461,
  [4462] = 4462,
  [4463] = 4463,
  [4464] = 2594,
  [4465] = 799,
  [4466] = 4466,
  [4467] = 2622,
  [4468] = 4468,
  [4469] = 3254,
  [4470] = 2623,
  [4471] = 2636,
  [4472] = 2637,
  [4473] = 4473,
  [4474] = 4474,
  [4475] = 2638,
  [4476] = 2639,
  [4477] = 2643,
  [4478] = 2648,
  [4479] = 2649,
  [4480] = 4466,
  [4481] = 2927,
  [4482] = 2603,
  [4483] = 2605,
  [4484] = 3257,
  [4485] = 2969,
  [4486] = 2944,
  [4487] = 4466,
  [4488] = 2645,
  [4489] = 2616,
  [4490] = 2629,
  [4491] = 4468,
  [4492] = 2640,
  [4493] = 3281,
  [4494] = 4468,
  [4495] = 2633,
  [4496] = 2635,
  [4497] = 2621,
  [4498] = 4498,
  [4499] = 3328,
  [4500] = 3066,
  [4501] = 3303,
  [4502] = 4502,
  [4503] = 4503,
  [4504] = 4504,
  [4505] = 4505,
  [4506] = 4506,
  [4507] = 4507,
  [4508] = 3341,
  [4509] = 2991,
  [4510] = 4510,
  [4511] = 4502,
  [4512] = 3310,
  [4513] = 4513,
  [4514] = 2992,
  [4515] = 4498,
  [4516] = 3126,
  [4517] = 3311,
  [4518] = 768,
  [4519] = 3127,
  [4520] = 3110,
  [4521] = 776,
  [4522] = 3326,
  [4523] = 3007,
  [4524] = 4507,
  [4525] = 4498,
  [4526] = 4526,
  [4527] = 4527,
  [4528] = 3124,
  [4529] = 3052,
  [4530] = 4513,
  [4531] = 3273,
  [4532] = 4502,
  [4533] = 3063,
  [4534] = 4507,
  [4535] = 3024,
  [4536] = 3028,
  [4537] = 2794,
  [4538] = 4538,
  [4539] = 1876,
  [4540] = 4538,
  [4541] = 4513,
  [4542] = 4507,
  [4543] = 4505,
  [4544] = 4544,
  [4545] = 4545,
  [4546] = 4545,
  [4547] = 4547,
  [4548] = 4547,
  [4549] = 2102,
  [4550] = 4545,
  [4551] = 3458,
  [4552] = 2103,
  [4553] = 2078,
  [4554] = 3364,
  [4555] = 4547,
  [4556] = 4545,
  [4557] = 2101,
  [4558] = 2082,
  [4559] = 2011,
  [4560] = 2081,
  [4561] = 2058,
  [4562] = 4545,
  [4563] = 4547,
  [4564] = 2083,
  [4565] = 4547,
  [4566] = 4566,
  [4567] = 4547,
  [4568] = 4547,
  [4569] = 4547,
  [4570] = 4547,
  [4571] = 4547,
  [4572] = 4547,
  [4573] = 4573,
  [4574] = 3233,
  [4575] = 4545,
  [4576] = 3224,
  [4577] = 4545,
  [4578] = 3365,
  [4579] = 2059,
  [4580] = 3235,
  [4581] = 2098,
  [4582] = 2105,
  [4583] = 2000,
  [4584] = 4547,
  [4585] = 3249,
  [4586] = 2072,
  [4587] = 4547,
  [4588] = 3192,
  [4589] = 2050,
  [4590] = 3238,
  [4591] = 2075,
  [4592] = 2871,
  [4593] = 2077,
  [4594] = 4545,
  [4595] = 4545,
  [4596] = 4547,
  [4597] = 4547,
  [4598] = 2060,
  [4599] = 3270,
  [4600] = 4547,
  [4601] = 4545,
  [4602] = 2048,
  [4603] = 2001,
  [4604] = 2002,
  [4605] = 2080,
  [4606] = 2106,
  [4607] = 4547,
  [4608] = 3236,
  [4609] = 3457,
  [4610] = 2090,
  [4611] = 2089,
  [4612] = 1948,
  [4613] = 4613,
  [4614] = 4614,
  [4615] = 2006,
  [4616] = 2007,
  [4617] = 2013,
  [4618] = 2014,
  [4619] = 4545,
  [4620] = 4620,
  [4621] = 4621,
  [4622] = 3494,
  [4623] = 4623,
  [4624] = 3499,
  [4625] = 4625,
  [4626] = 4626,
  [4627] = 3486,
  [4628] = 3501,
  [4629] = 4629,
  [4630] = 4629,
  [4631] = 2794,
  [4632] = 3480,
  [4633] = 3488,
  [4634] = 3483,
  [4635] = 4635,
  [4636] = 3492,
  [4637] = 4637,
  [4638] = 3485,
  [4639] = 3512,
  [4640] = 4635,
  [4641] = 4641,
  [4642] = 4642,
  [4643] = 4643,
  [4644] = 3504,
  [4645] = 4642,
  [4646] = 4443,
  [4647] = 3497,
  [4648] = 3490,
  [4649] = 3438,
  [4650] = 3418,
  [4651] = 3503,
  [4652] = 4637,
  [4653] = 4637,
  [4654] = 3487,
  [4655] = 4643,
  [4656] = 4656,
  [4657] = 3267,
  [4658] = 4656,
  [4659] = 3484,
  [4660] = 3508,
  [4661] = 4637,
  [4662] = 3498,
  [4663] = 3388,
  [4664] = 4620,
  [4665] = 4665,
  [4666] = 3506,
  [4667] = 3500,
  [4668] = 4635,
  [4669] = 4643,
  [4670] = 4670,
  [4671] = 4671,
  [4672] = 3496,
  [4673] = 3489,
  [4674] = 4625,
  [4675] = 4629,
  [4676] = 4656,
  [4677] = 1878,
  [4678] = 787,
  [4679] = 4679,
  [4680] = 3331,
  [4681] = 4681,
  [4682] = 4682,
  [4683] = 4683,
  [4684] = 4684,
  [4685] = 4685,
  [4686] = 4679,
  [4687] = 3545,
  [4688] = 4688,
  [4689] = 793,
  [4690] = 4690,
  [4691] = 4691,
  [4692] = 1878,
  [4693] = 4684,
  [4694] = 3433,
  [4695] = 4683,
  [4696] = 4684,
  [4697] = 4697,
  [4698] = 4698,
  [4699] = 4699,
  [4700] = 3516,
  [4701] = 4506,
  [4702] = 4702,
  [4703] = 4699,
  [4704] = 1878,
  [4705] = 4705,
  [4706] = 4683,
  [4707] = 4679,
  [4708] = 4685,
  [4709] = 3340,
  [4710] = 3304,
  [4711] = 2071,
  [4712] = 4702,
  [4713] = 2073,
  [4714] = 4684,
  [4715] = 4699,
  [4716] = 4510,
  [4717] = 3547,
  [4718] = 4682,
  [4719] = 4683,
  [4720] = 4679,
  [4721] = 4685,
  [4722] = 4683,
  [4723] = 4679,
  [4724] = 4705,
  [4725] = 2074,
  [4726] = 2076,
  [4727] = 4727,
  [4728] = 4683,
  [4729] = 2079,
  [4730] = 3324,
  [4731] = 3519,
  [4732] = 4732,
  [4733] = 3476,
  [4734] = 3320,
  [4735] = 4683,
  [4736] = 786,
  [4737] = 4443,
  [4738] = 790,
  [4739] = 3510,
  [4740] = 4684,
  [4741] = 4741,
  [4742] = 794,
  [4743] = 4727,
  [4744] = 815,
  [4745] = 1876,
  [4746] = 819,
  [4747] = 821,
  [4748] = 4748,
  [4749] = 4684,
  [4750] = 4504,
  [4751] = 3551,
  [4752] = 3306,
  [4753] = 4741,
  [4754] = 4679,
  [4755] = 4705,
  [4756] = 4732,
  [4757] = 4688,
  [4758] = 4683,
  [4759] = 4759,
  [4760] = 4741,
  [4761] = 4761,
  [4762] = 4684,
  [4763] = 3482,
  [4764] = 3302,
  [4765] = 4690,
  [4766] = 4691,
  [4767] = 4691,
  [4768] = 4768,
  [4769] = 3397,
  [4770] = 3521,
  [4771] = 3300,
  [4772] = 4684,
  [4773] = 4690,
  [4774] = 4688,
  [4775] = 4691,
  [4776] = 4776,
  [4777] = 4688,
  [4778] = 4683,
  [4779] = 4684,
  [4780] = 4702,
  [4781] = 4684,
  [4782] = 4782,
  [4783] = 4295,
  [4784] = 785,
  [4785] = 4690,
  [4786] = 2871,
  [4787] = 4527,
  [4788] = 1857,
  [4789] = 4679,
  [4790] = 4698,
  [4791] = 3553,
  [4792] = 4741,
  [4793] = 4699,
  [4794] = 4705,
  [4795] = 4683,
  [4796] = 4796,
  [4797] = 4768,
  [4798] = 4768,
  [4799] = 4799,
  [4800] = 3518,
  [4801] = 4685,
  [4802] = 4702,
  [4803] = 808,
  [4804] = 4684,
  [4805] = 2066,
  [4806] = 1870,
  [4807] = 4727,
  [4808] = 4526,
  [4809] = 4679,
  [4810] = 797,
  [4811] = 4503,
  [4812] = 4683,
  [4813] = 817,
  [4814] = 3633,
  [4815] = 4815,
  [4816] = 2073,
  [4817] = 4817,
  [4818] = 2008,
  [4819] = 2010,
  [4820] = 4820,
  [4821] = 2011,
  [4822] = 2074,
  [4823] = 4823,
  [4824] = 2044,
  [4825] = 2076,
  [4826] = 4826,
  [4827] = 811,
  [4828] = 3635,
  [4829] = 4829,
  [4830] = 2073,
  [4831] = 4817,
  [4832] = 2074,
  [4833] = 2263,
  [4834] = 2105,
  [4835] = 3636,
  [4836] = 4836,
  [4837] = 2075,
  [4838] = 4829,
  [4839] = 2089,
  [4840] = 4840,
  [4841] = 2090,
  [4842] = 4842,
  [4843] = 4843,
  [4844] = 2079,
  [4845] = 4843,
  [4846] = 2076,
  [4847] = 2077,
  [4848] = 3725,
  [4849] = 4849,
  [4850] = 2078,
  [4851] = 2098,
  [4852] = 2079,
  [4853] = 4853,
  [4854] = 2102,
  [4855] = 4817,
  [4856] = 2103,
  [4857] = 2209,
  [4858] = 2106,
  [4859] = 4849,
  [4860] = 4817,
  [4861] = 4836,
  [4862] = 4853,
  [4863] = 2173,
  [4864] = 3550,
  [4865] = 2057,
  [4866] = 3538,
  [4867] = 4842,
  [4868] = 822,
  [4869] = 3552,
  [4870] = 3641,
  [4871] = 3642,
  [4872] = 2158,
  [4873] = 4843,
  [4874] = 3522,
  [4875] = 3530,
  [4876] = 3524,
  [4877] = 3526,
  [4878] = 3525,
  [4879] = 4879,
  [4880] = 4526,
  [4881] = 4881,
  [4882] = 4380,
  [4883] = 4398,
  [4884] = 4849,
  [4885] = 2066,
  [4886] = 4410,
  [4887] = 2160,
  [4888] = 3637,
  [4889] = 4817,
  [4890] = 4817,
  [4891] = 2246,
  [4892] = 785,
  [4893] = 4836,
  [4894] = 793,
  [4895] = 4829,
  [4896] = 2262,
  [4897] = 2082,
  [4898] = 2143,
  [4899] = 2083,
  [4900] = 2144,
  [4901] = 4817,
  [4902] = 2154,
  [4903] = 4836,
  [4904] = 4317,
  [4905] = 4836,
  [4906] = 3644,
  [4907] = 4836,
  [4908] = 3528,
  [4909] = 4909,
  [4910] = 1948,
  [4911] = 2006,
  [4912] = 4842,
  [4913] = 2007,
  [4914] = 2013,
  [4915] = 4843,
  [4916] = 2014,
  [4917] = 4917,
  [4918] = 4840,
  [4919] = 4849,
  [4920] = 4829,
  [4921] = 4921,
  [4922] = 4503,
  [4923] = 4817,
  [4924] = 2175,
  [4925] = 2047,
  [4926] = 4817,
  [4927] = 2048,
  [4928] = 4504,
  [4929] = 4527,
  [4930] = 4917,
  [4931] = 4817,
  [4932] = 4842,
  [4933] = 4933,
  [4934] = 4829,
  [4935] = 4321,
  [4936] = 4843,
  [4937] = 810,
  [4938] = 3724,
  [4939] = 3315,
  [4940] = 2050,
  [4941] = 4829,
  [4942] = 4506,
  [4943] = 4836,
  [4944] = 2112,
  [4945] = 3634,
  [4946] = 4836,
  [4947] = 4823,
  [4948] = 4909,
  [4949] = 4842,
  [4950] = 4842,
  [4951] = 2235,
  [4952] = 3638,
  [4953] = 4953,
  [4954] = 4843,
  [4955] = 3728,
  [4956] = 3632,
  [4957] = 4849,
  [4958] = 2241,
  [4959] = 2250,
  [4960] = 3639,
  [4961] = 4322,
  [4962] = 3640,
  [4963] = 792,
  [4964] = 4829,
  [4965] = 2001,
  [4966] = 4323,
  [4967] = 2002,
  [4968] = 3520,
  [4969] = 2080,
  [4970] = 4836,
  [4971] = 4510,
  [4972] = 4817,
  [4973] = 841,
  [4974] = 3630,
  [4975] = 2197,
  [4976] = 2101,
  [4977] = 4327,
  [4978] = 4829,
  [4979] = 4842,
  [4980] = 816,
  [4981] = 4336,
  [4982] = 4843,
  [4983] = 3719,
  [4984] = 842,
  [4985] = 4823,
  [4986] = 4909,
  [4987] = 4358,
  [4988] = 4843,
  [4989] = 4849,
  [4990] = 4990,
  [4991] = 2058,
  [4992] = 4836,
  [4993] = 843,
  [4994] = 4359,
  [4995] = 3627,
  [4996] = 4842,
  [4997] = 2081,
  [4998] = 4998,
  [4999] = 799,
  [5000] = 4842,
  [5001] = 4829,
  [5002] = 3546,
  [5003] = 844,
  [5004] = 3529,
  [5005] = 2059,
  [5006] = 4849,
  [5007] = 5007,
  [5008] = 4379,
  [5009] = 4441,
  [5010] = 4387,
  [5011] = 4388,
  [5012] = 4842,
  [5013] = 5007,
  [5014] = 4933,
  [5015] = 3548,
  [5016] = 2056,
  [5017] = 3544,
  [5018] = 2060,
  [5019] = 3527,
  [5020] = 3539,
  [5021] = 4843,
  [5022] = 4333,
  [5023] = 4843,
  [5024] = 3631,
  [5025] = 5007,
  [5026] = 5026,
  [5027] = 5007,
  [5028] = 1999,
  [5029] = 4849,
  [5030] = 2000,
  [5031] = 2066,
  [5032] = 4849,
  [5033] = 3504,
  [5034] = 3492,
  [5035] = 4829,
  [5036] = 2071,
  [5037] = 3508,
  [5038] = 2071,
  [5039] = 2072,
  [5040] = 3536,
  [5041] = 4849,
  [5042] = 3588,
  [5043] = 5043,
  [5044] = 5044,
  [5045] = 5045,
  [5046] = 3655,
  [5047] = 5047,
  [5048] = 5045,
  [5049] = 5049,
  [5050] = 3717,
  [5051] = 5051,
  [5052] = 5052,
  [5053] = 5045,
  [5054] = 3718,
  [5055] = 5055,
  [5056] = 3656,
  [5057] = 5057,
  [5058] = 3657,
  [5059] = 3692,
  [5060] = 5060,
  [5061] = 3693,
  [5062] = 5062,
  [5063] = 3720,
  [5064] = 5064,
  [5065] = 3721,
  [5066] = 5066,
  [5067] = 3722,
  [5068] = 3723,
  [5069] = 3726,
  [5070] = 5070,
  [5071] = 3727,
  [5072] = 5045,
  [5073] = 3585,
  [5074] = 5074,
  [5075] = 3599,
  [5076] = 3600,
  [5077] = 3645,
  [5078] = 5055,
  [5079] = 5060,
  [5080] = 3514,
  [5081] = 5081,
  [5082] = 5082,
  [5083] = 5083,
  [5084] = 5084,
  [5085] = 5055,
  [5086] = 5086,
  [5087] = 3695,
  [5088] = 5088,
  [5089] = 3566,
  [5090] = 3567,
  [5091] = 3568,
  [5092] = 5092,
  [5093] = 3608,
  [5094] = 5064,
  [5095] = 3582,
  [5096] = 5096,
  [5097] = 4244,
  [5098] = 3315,
  [5099] = 3643,
  [5100] = 5100,
  [5101] = 5047,
  [5102] = 3609,
  [5103] = 5103,
  [5104] = 3556,
  [5105] = 5105,
  [5106] = 5057,
  [5107] = 5045,
  [5108] = 5081,
  [5109] = 5044,
  [5110] = 3593,
  [5111] = 3659,
  [5112] = 5045,
  [5113] = 4247,
  [5114] = 5055,
  [5115] = 3660,
  [5116] = 3677,
  [5117] = 3610,
  [5118] = 3661,
  [5119] = 3587,
  [5120] = 3569,
  [5121] = 3696,
  [5122] = 5057,
  [5123] = 5049,
  [5124] = 5057,
  [5125] = 5125,
  [5126] = 3674,
  [5127] = 3612,
  [5128] = 3613,
  [5129] = 3681,
  [5130] = 3682,
  [5131] = 3614,
  [5132] = 3662,
  [5133] = 5055,
  [5134] = 3683,
  [5135] = 833,
  [5136] = 5100,
  [5137] = 3663,
  [5138] = 3684,
  [5139] = 5096,
  [5140] = 5140,
  [5141] = 5105,
  [5142] = 3596,
  [5143] = 3685,
  [5144] = 5081,
  [5145] = 3664,
  [5146] = 5062,
  [5147] = 3586,
  [5148] = 3697,
  [5149] = 3686,
  [5150] = 3570,
  [5151] = 3698,
  [5152] = 3665,
  [5153] = 5049,
  [5154] = 3666,
  [5155] = 5045,
  [5156] = 5045,
  [5157] = 5043,
  [5158] = 3571,
  [5159] = 3615,
  [5160] = 5160,
  [5161] = 5049,
  [5162] = 5074,
  [5163] = 5049,
  [5164] = 3687,
  [5165] = 5055,
  [5166] = 3563,
  [5167] = 3700,
  [5168] = 3701,
  [5169] = 3702,
  [5170] = 3703,
  [5171] = 5044,
  [5172] = 5043,
  [5173] = 5045,
  [5174] = 3704,
  [5175] = 3577,
  [5176] = 3547,
  [5177] = 3551,
  [5178] = 3516,
  [5179] = 5179,
  [5180] = 3521,
  [5181] = 5062,
  [5182] = 3705,
  [5183] = 3617,
  [5184] = 3706,
  [5185] = 5049,
  [5186] = 5100,
  [5187] = 3559,
  [5188] = 3688,
  [5189] = 3707,
  [5190] = 3624,
  [5191] = 3669,
  [5192] = 5062,
  [5193] = 5060,
  [5194] = 5100,
  [5195] = 3581,
  [5196] = 3658,
  [5197] = 3689,
  [5198] = 3647,
  [5199] = 3690,
  [5200] = 3543,
  [5201] = 3564,
  [5202] = 5047,
  [5203] = 5203,
  [5204] = 3708,
  [5205] = 3549,
  [5206] = 3709,
  [5207] = 5049,
  [5208] = 5064,
  [5209] = 3710,
  [5210] = 3711,
  [5211] = 3590,
  [5212] = 3560,
  [5213] = 3592,
  [5214] = 3554,
  [5215] = 5074,
  [5216] = 3712,
  [5217] = 3649,
  [5218] = 3561,
  [5219] = 3602,
  [5220] = 5045,
  [5221] = 3713,
  [5222] = 3679,
  [5223] = 3714,
  [5224] = 5049,
  [5225] = 3651,
  [5226] = 5226,
  [5227] = 5227,
  [5228] = 5228,
  [5229] = 5044,
  [5230] = 5105,
  [5231] = 5231,
  [5232] = 3676,
  [5233] = 5233,
  [5234] = 3573,
  [5235] = 5235,
  [5236] = 5051,
  [5237] = 5096,
  [5238] = 3603,
  [5239] = 3574,
  [5240] = 3653,
  [5241] = 3620,
  [5242] = 5140,
  [5243] = 3575,
  [5244] = 5096,
  [5245] = 5051,
  [5246] = 5140,
  [5247] = 5051,
  [5248] = 5140,
  [5249] = 5249,
  [5250] = 5074,
  [5251] = 3622,
  [5252] = 3623,
  [5253] = 5060,
  [5254] = 5043,
  [5255] = 3715,
  [5256] = 5049,
  [5257] = 5047,
  [5258] = 3716,
  [5259] = 5081,
  [5260] = 5049,
  [5261] = 3604,
  [5262] = 3598,
  [5263] = 3515,
  [5264] = 3699,
  [5265] = 5265,
  [5266] = 5266,
  [5267] = 5267,
  [5268] = 5268,
  [5269] = 5269,
  [5270] = 5270,
  [5271] = 3755,
  [5272] = 5272,
  [5273] = 5273,
  [5274] = 5274,
  [5275] = 5275,
  [5276] = 5276,
  [5277] = 5277,
  [5278] = 5266,
  [5279] = 5275,
  [5280] = 5280,
  [5281] = 5281,
  [5282] = 5282,
  [5283] = 5283,
  [5284] = 5268,
  [5285] = 5273,
  [5286] = 3387,
  [5287] = 3619,
  [5288] = 3765,
  [5289] = 5289,
  [5290] = 5274,
  [5291] = 5272,
  [5292] = 5292,
  [5293] = 5275,
  [5294] = 3734,
  [5295] = 5276,
  [5296] = 3468,
  [5297] = 5297,
  [5298] = 5298,
  [5299] = 3691,
  [5300] = 3358,
  [5301] = 5275,
  [5302] = 3360,
  [5303] = 5303,
  [5304] = 5304,
  [5305] = 5305,
  [5306] = 3737,
  [5307] = 5274,
  [5308] = 5275,
  [5309] = 5276,
  [5310] = 3444,
  [5311] = 5311,
  [5312] = 5270,
  [5313] = 3766,
  [5314] = 5314,
  [5315] = 5315,
  [5316] = 5282,
  [5317] = 5274,
  [5318] = 5276,
  [5319] = 5266,
  [5320] = 5303,
  [5321] = 5321,
  [5322] = 5322,
  [5323] = 5277,
  [5324] = 3792,
  [5325] = 5311,
  [5326] = 5326,
  [5327] = 3502,
  [5328] = 5328,
  [5329] = 3787,
  [5330] = 5330,
  [5331] = 3628,
  [5332] = 5266,
  [5333] = 5333,
  [5334] = 3629,
  [5335] = 3472,
  [5336] = 5289,
  [5337] = 3518,
  [5338] = 5289,
  [5339] = 5339,
  [5340] = 5340,
  [5341] = 5341,
  [5342] = 5321,
  [5343] = 5343,
  [5344] = 5344,
  [5345] = 5345,
  [5346] = 5346,
  [5347] = 5347,
  [5348] = 5348,
  [5349] = 5339,
  [5350] = 5274,
  [5351] = 5351,
  [5352] = 5352,
  [5353] = 5340,
  [5354] = 5333,
  [5355] = 5355,
  [5356] = 5356,
  [5357] = 5275,
  [5358] = 5275,
  [5359] = 5359,
  [5360] = 5297,
  [5361] = 5298,
  [5362] = 5362,
  [5363] = 2646,
  [5364] = 5364,
  [5365] = 5276,
  [5366] = 5314,
  [5367] = 5322,
  [5368] = 5347,
  [5369] = 5348,
  [5370] = 5311,
  [5371] = 5274,
  [5372] = 3495,
  [5373] = 5373,
  [5374] = 5374,
  [5375] = 3513,
  [5376] = 3511,
  [5377] = 3753,
  [5378] = 5281,
  [5379] = 5267,
  [5380] = 1814,
  [5381] = 3505,
  [5382] = 5382,
  [5383] = 5383,
  [5384] = 5275,
  [5385] = 5385,
  [5386] = 5276,
  [5387] = 2594,
  [5388] = 3750,
  [5389] = 5389,
  [5390] = 5326,
  [5391] = 5391,
  [5392] = 5272,
  [5393] = 5269,
  [5394] = 5374,
  [5395] = 5395,
  [5396] = 4336,
  [5397] = 4358,
  [5398] = 4359,
  [5399] = 3374,
  [5400] = 5391,
  [5401] = 3743,
  [5402] = 5344,
  [5403] = 5266,
  [5404] = 5404,
  [5405] = 5321,
  [5406] = 5356,
  [5407] = 3353,
  [5408] = 4264,
  [5409] = 3756,
  [5410] = 3356,
  [5411] = 2098,
  [5412] = 2102,
  [5413] = 2103,
  [5414] = 3357,
  [5415] = 5339,
  [5416] = 5416,
  [5417] = 3509,
  [5418] = 5418,
  [5419] = 5419,
  [5420] = 5420,
  [5421] = 3763,
  [5422] = 5321,
  [5423] = 5351,
  [5424] = 2014,
  [5425] = 5352,
  [5426] = 2050,
  [5427] = 5340,
  [5428] = 5274,
  [5429] = 5267,
  [5430] = 5282,
  [5431] = 5431,
  [5432] = 5297,
  [5433] = 5298,
  [5434] = 5321,
  [5435] = 5356,
  [5436] = 5322,
  [5437] = 5321,
  [5438] = 5276,
  [5439] = 5420,
  [5440] = 5276,
  [5441] = 3781,
  [5442] = 5416,
  [5443] = 5275,
  [5444] = 5265,
  [5445] = 5359,
  [5446] = 5446,
  [5447] = 5447,
  [5448] = 5448,
  [5449] = 5266,
  [5450] = 5364,
  [5451] = 3373,
  [5452] = 5268,
  [5453] = 5446,
  [5454] = 5265,
  [5455] = 5446,
  [5456] = 5456,
  [5457] = 5304,
  [5458] = 5269,
  [5459] = 4410,
  [5460] = 5460,
  [5461] = 4327,
  [5462] = 1542,
  [5463] = 5274,
  [5464] = 5351,
  [5465] = 5356,
  [5466] = 3774,
  [5467] = 5431,
  [5468] = 5347,
  [5469] = 5276,
  [5470] = 3547,
  [5471] = 3551,
  [5472] = 5364,
  [5473] = 5348,
  [5474] = 1533,
  [5475] = 3428,
  [5476] = 3516,
  [5477] = 5303,
  [5478] = 3453,
  [5479] = 5265,
  [5480] = 5446,
  [5481] = 5304,
  [5482] = 5269,
  [5483] = 5275,
  [5484] = 5359,
  [5485] = 5274,
  [5486] = 5373,
  [5487] = 3380,
  [5488] = 5265,
  [5489] = 5446,
  [5490] = 5304,
  [5491] = 3417,
  [5492] = 5269,
  [5493] = 5460,
  [5494] = 3775,
  [5495] = 3780,
  [5496] = 5385,
  [5497] = 5266,
  [5498] = 3752,
  [5499] = 5315,
  [5500] = 5460,
  [5501] = 5356,
  [5502] = 3733,
  [5503] = 5382,
  [5504] = 5460,
  [5505] = 5447,
  [5506] = 5460,
  [5507] = 5281,
  [5508] = 5460,
  [5509] = 3405,
  [5510] = 1544,
  [5511] = 3773,
  [5512] = 5385,
  [5513] = 5460,
  [5514] = 5460,
  [5515] = 5460,
  [5516] = 5460,
  [5517] = 5383,
  [5518] = 5420,
  [5519] = 3521,
  [5520] = 5274,
  [5521] = 5265,
  [5522] = 5446,
  [5523] = 5269,
  [5524] = 5383,
  [5525] = 5275,
  [5526] = 5383,
  [5527] = 5431,
  [5528] = 5374,
  [5529] = 3784,
  [5530] = 5276,
  [5531] = 5352,
  [5532] = 3547,
  [5533] = 3551,
  [5534] = 5304,
  [5535] = 3516,
  [5536] = 5352,
  [5537] = 3521,
  [5538] = 3797,
  [5539] = 5265,
  [5540] = 5446,
  [5541] = 5269,
  [5542] = 3769,
  [5543] = 5543,
  [5544] = 5544,
  [5545] = 5545,
  [5546] = 5546,
  [5547] = 5547,
  [5548] = 5548,
  [5549] = 5549,
  [5550] = 5550,
  [5551] = 5544,
  [5552] = 5552,
  [5553] = 5553,
  [5554] = 5554,
  [5555] = 5555,
  [5556] = 5556,
  [5557] = 5557,
  [5558] = 5558,
  [5559] = 5546,
  [5560] = 5560,
  [5561] = 5561,
  [5562] = 5562,
  [5563] = 5563,
  [5564] = 5564,
  [5565] = 5565,
  [5566] = 5562,
  [5567] = 5567,
  [5568] = 5568,
  [5569] = 5562,
  [5570] = 5570,
  [5571] = 5548,
  [5572] = 5572,
  [5573] = 5573,
  [5574] = 5574,
  [5575] = 5575,
  [5576] = 5576,
  [5577] = 5577,
  [5578] = 3846,
  [5579] = 5579,
  [5580] = 5580,
  [5581] = 5554,
  [5582] = 3854,
  [5583] = 5583,
  [5584] = 4045,
  [5585] = 5585,
  [5586] = 5585,
  [5587] = 3921,
  [5588] = 5553,
  [5589] = 5546,
  [5590] = 3268,
  [5591] = 5591,
  [5592] = 5592,
  [5593] = 5565,
  [5594] = 5594,
  [5595] = 5595,
  [5596] = 5596,
  [5597] = 5597,
  [5598] = 5598,
  [5599] = 5599,
  [5600] = 5591,
  [5601] = 5583,
  [5602] = 5563,
  [5603] = 5603,
  [5604] = 5604,
  [5605] = 5565,
  [5606] = 5606,
  [5607] = 5592,
  [5608] = 5577,
  [5609] = 5573,
  [5610] = 5610,
  [5611] = 5579,
  [5612] = 5612,
  [5613] = 5613,
  [5614] = 5614,
  [5615] = 5557,
  [5616] = 5560,
  [5617] = 5573,
  [5618] = 3868,
  [5619] = 5619,
  [5620] = 5585,
  [5621] = 5621,
  [5622] = 5562,
  [5623] = 5623,
  [5624] = 5624,
  [5625] = 5625,
  [5626] = 5626,
  [5627] = 5627,
  [5628] = 5604,
  [5629] = 5629,
  [5630] = 5610,
  [5631] = 5631,
  [5632] = 5610,
  [5633] = 5633,
  [5634] = 3927,
  [5635] = 5543,
  [5636] = 3869,
  [5637] = 5637,
  [5638] = 5568,
  [5639] = 5639,
  [5640] = 5585,
  [5641] = 5546,
  [5642] = 5642,
  [5643] = 5643,
  [5644] = 5562,
  [5645] = 5645,
  [5646] = 5594,
  [5647] = 4046,
  [5648] = 3948,
  [5649] = 3952,
  [5650] = 5556,
  [5651] = 5565,
  [5652] = 5625,
  [5653] = 5576,
  [5654] = 5596,
  [5655] = 5627,
  [5656] = 5573,
  [5657] = 4047,
  [5658] = 5606,
  [5659] = 5585,
  [5660] = 5573,
  [5661] = 5661,
  [5662] = 5662,
  [5663] = 5663,
  [5664] = 5664,
  [5665] = 5580,
  [5666] = 5666,
  [5667] = 5667,
  [5668] = 5662,
  [5669] = 5669,
  [5670] = 5613,
  [5671] = 5564,
  [5672] = 5543,
  [5673] = 3920,
  [5674] = 5674,
  [5675] = 5610,
  [5676] = 5553,
  [5677] = 5677,
  [5678] = 5583,
  [5679] = 3253,
  [5680] = 5563,
  [5681] = 5543,
  [5682] = 5606,
  [5683] = 5683,
  [5684] = 5612,
  [5685] = 5610,
  [5686] = 5564,
  [5687] = 5574,
  [5688] = 5579,
  [5689] = 5689,
  [5690] = 5543,
  [5691] = 5546,
  [5692] = 5637,
  [5693] = 5667,
  [5694] = 5662,
  [5695] = 5669,
  [5696] = 5696,
  [5697] = 5548,
  [5698] = 5698,
  [5699] = 5562,
  [5700] = 5700,
  [5701] = 5701,
  [5702] = 5639,
  [5703] = 5565,
  [5704] = 5546,
  [5705] = 5612,
  [5706] = 5592,
  [5707] = 5554,
  [5708] = 5633,
  [5709] = 5624,
  [5710] = 5562,
  [5711] = 5573,
  [5712] = 5594,
  [5713] = 5568,
  [5714] = 5596,
  [5715] = 5576,
  [5716] = 5667,
  [5717] = 5591,
  [5718] = 5580,
  [5719] = 5585,
  [5720] = 5575,
  [5721] = 5669,
  [5722] = 5576,
  [5723] = 5604,
  [5724] = 5565,
  [5725] = 5553,
  [5726] = 5574,
  [5727] = 5667,
  [5728] = 5565,
  [5729] = 5729,
  [5730] = 5669,
  [5731] = 5689,
  [5732] = 5564,
  [5733] = 5573,
  [5734] = 5599,
  [5735] = 5667,
  [5736] = 5613,
  [5737] = 5546,
  [5738] = 5669,
  [5739] = 5573,
  [5740] = 5585,
  [5741] = 3864,
  [5742] = 5667,
  [5743] = 5669,
  [5744] = 5677,
  [5745] = 5696,
  [5746] = 5746,
  [5747] = 5747,
  [5748] = 5667,
  [5749] = 5669,
  [5750] = 5669,
  [5751] = 5751,
  [5752] = 5752,
  [5753] = 3495,
  [5754] = 5751,
  [5755] = 5549,
  [5756] = 5756,
  [5757] = 5565,
  [5758] = 5667,
  [5759] = 5669,
  [5760] = 5760,
  [5761] = 5761,
  [5762] = 5762,
  [5763] = 5763,
  [5764] = 5667,
  [5765] = 5564,
  [5766] = 5669,
  [5767] = 5667,
  [5768] = 5613,
  [5769] = 5573,
  [5770] = 5669,
  [5771] = 5667,
  [5772] = 5669,
  [5773] = 5667,
  [5774] = 5610,
  [5775] = 5669,
  [5776] = 5667,
  [5777] = 5669,
  [5778] = 5667,
  [5779] = 5597,
  [5780] = 5669,
  [5781] = 5667,
  [5782] = 5782,
  [5783] = 5669,
  [5784] = 5667,
  [5785] = 5669,
  [5786] = 5667,
  [5787] = 5669,
  [5788] = 5667,
  [5789] = 5669,
  [5790] = 5700,
  [5791] = 5543,
  [5792] = 5792,
  [5793] = 5580,
  [5794] = 5794,
  [5795] = 4048,
  [5796] = 3260,
  [5797] = 5568,
  [5798] = 5552,
  [5799] = 3922,
  [5800] = 5546,
  [5801] = 5547,
  [5802] = 5802,
  [5803] = 5550,
  [5804] = 5544,
  [5805] = 3350,
  [5806] = 5612,
  [5807] = 5729,
  [5808] = 5562,
  [5809] = 5809,
  [5810] = 3262,
  [5811] = 5572,
  [5812] = 5576,
  [5813] = 4005,
  [5814] = 5752,
  [5815] = 5815,
  [5816] = 5614,
  [5817] = 5557,
  [5818] = 4050,
  [5819] = 5565,
  [5820] = 5560,
  [5821] = 5821,
  [5822] = 5573,
  [5823] = 5613,
  [5824] = 3946,
  [5825] = 5570,
  [5826] = 5585,
  [5827] = 5610,
  [5828] = 5667,
  [5829] = 5792,
  [5830] = 5585,
  [5831] = 5637,
  [5832] = 4009,
  [5833] = 5802,
  [5834] = 5834,
  [5835] = 5835,
  [5836] = 5543,
  [5837] = 5626,
  [5838] = 5562,
  [5839] = 5792,
  [5840] = 5661,
  [5841] = 5663,
  [5842] = 3916,
  [5843] = 5547,
  [5844] = 4051,
  [5845] = 5809,
  [5846] = 5802,
  [5847] = 5564,
  [5848] = 5667,
  [5849] = 5792,
  [5850] = 5760,
  [5851] = 4044,
  [5852] = 5642,
  [5853] = 5547,
  [5854] = 5610,
  [5855] = 5809,
  [5856] = 5792,
  [5857] = 5834,
  [5858] = 3878,
  [5859] = 5543,
  [5860] = 5547,
  [5861] = 5595,
  [5862] = 5809,
  [5863] = 5792,
  [5864] = 5543,
  [5865] = 5547,
  [5866] = 5809,
  [5867] = 5792,
  [5868] = 5662,
  [5869] = 1625,
  [5870] = 5547,
  [5871] = 5871,
  [5872] = 5809,
  [5873] = 5792,
  [5874] = 5597,
  [5875] = 5547,
  [5876] = 5809,
  [5877] = 5792,
  [5878] = 5547,
  [5879] = 5809,
  [5880] = 5792,
  [5881] = 5610,
  [5882] = 5547,
  [5883] = 5809,
  [5884] = 5792,
  [5885] = 5547,
  [5886] = 5886,
  [5887] = 5792,
  [5888] = 5547,
  [5889] = 5792,
  [5890] = 5547,
  [5891] = 5792,
  [5892] = 5547,
  [5893] = 5792,
  [5894] = 5547,
  [5895] = 5792,
  [5896] = 5547,
  [5897] = 5792,
  [5898] = 5547,
  [5899] = 5792,
  [5900] = 5792,
  [5901] = 5792,
  [5902] = 5546,
  [5903] = 5669,
  [5904] = 5802,
  [5905] = 5550,
  [5906] = 5809,
  [5907] = 5835,
  [5908] = 5761,
  [5909] = 5762,
  [5910] = 3505,
  [5911] = 5821,
  [5912] = 5610,
  [5913] = 5552,
  [5914] = 5562,
  [5915] = 5574,
  [5916] = 5576,
  [5917] = 5556,
  [5918] = 5802,
  [5919] = 3502,
  [5920] = 5625,
  [5921] = 5565,
  [5922] = 5627,
  [5923] = 5633,
  [5924] = 5760,
  [5925] = 5642,
  [5926] = 5637,
  [5927] = 5573,
  [5928] = 5613,
  [5929] = 5572,
  [5930] = 5585,
  [5931] = 5689,
  [5932] = 5558,
  [5933] = 5549,
  [5934] = 3332,
  [5935] = 5645,
  [5936] = 5598,
  [5937] = 5543,
  [5938] = 5666,
  [5939] = 5614,
  [5940] = 5549,
  [5941] = 5546,
  [5942] = 4038,
  [5943] = 4039,
  [5944] = 5631,
  [5945] = 5945,
  [5946] = 5946,
  [5947] = 5947,
  [5948] = 5948,
  [5949] = 5949,
  [5950] = 5950,
  [5951] = 5951,
  [5952] = 5952,
  [5953] = 5953,
  [5954] = 5954,
  [5955] = 5955,
  [5956] = 5956,
  [5957] = 5957,
  [5958] = 5958,
  [5959] = 5959,
  [5960] = 5960,
  [5961] = 5961,
  [5962] = 5962,
  [5963] = 5963,
  [5964] = 5964,
  [5965] = 5946,
  [5966] = 5966,
  [5967] = 5951,
  [5968] = 5964,
  [5969] = 5945,
  [5970] = 5953,
  [5971] = 5971,
  [5972] = 5972,
  [5973] = 5973,
  [5974] = 5972,
  [5975] = 5975,
  [5976] = 5976,
  [5977] = 5958,
  [5978] = 5959,
  [5979] = 5960,
  [5980] = 4443,
  [5981] = 5946,
  [5982] = 5982,
  [5983] = 5983,
  [5984] = 5984,
  [5985] = 5959,
  [5986] = 5986,
  [5987] = 5987,
  [5988] = 5988,
  [5989] = 5983,
  [5990] = 5990,
  [5991] = 5953,
  [5992] = 5992,
  [5993] = 5984,
  [5994] = 5947,
  [5995] = 5995,
  [5996] = 5958,
  [5997] = 5959,
  [5998] = 5960,
  [5999] = 5999,
  [6000] = 5971,
  [6001] = 6001,
  [6002] = 5972,
  [6003] = 6003,
  [6004] = 6004,
  [6005] = 5986,
  [6006] = 6006,
  [6007] = 6007,
  [6008] = 5953,
  [6009] = 5958,
  [6010] = 5959,
  [6011] = 5960,
  [6012] = 6012,
  [6013] = 5953,
  [6014] = 6014,
  [6015] = 5958,
  [6016] = 5953,
  [6017] = 5958,
  [6018] = 5959,
  [6019] = 5960,
  [6020] = 5961,
  [6021] = 5951,
  [6022] = 5960,
  [6023] = 5953,
  [6024] = 5958,
  [6025] = 5959,
  [6026] = 5960,
  [6027] = 5972,
  [6028] = 5947,
  [6029] = 5972,
  [6030] = 5953,
  [6031] = 5958,
  [6032] = 5959,
  [6033] = 5960,
  [6034] = 5948,
  [6035] = 5959,
  [6036] = 5953,
  [6037] = 5958,
  [6038] = 5959,
  [6039] = 5960,
  [6040] = 5999,
  [6041] = 6041,
  [6042] = 5951,
  [6043] = 5953,
  [6044] = 5958,
  [6045] = 5959,
  [6046] = 5960,
  [6047] = 6047,
  [6048] = 6048,
  [6049] = 6049,
  [6050] = 5953,
  [6051] = 5958,
  [6052] = 5959,
  [6053] = 5960,
  [6054] = 5955,
  [6055] = 6055,
  [6056] = 5953,
  [6057] = 5955,
  [6058] = 6058,
  [6059] = 5953,
  [6060] = 6060,
  [6061] = 5946,
  [6062] = 5953,
  [6063] = 5990,
  [6064] = 5972,
  [6065] = 5976,
  [6066] = 6041,
  [6067] = 5984,
  [6068] = 6068,
  [6069] = 5953,
  [6070] = 6007,
  [6071] = 6071,
  [6072] = 6048,
  [6073] = 6001,
  [6074] = 6074,
  [6075] = 5945,
  [6076] = 6076,
  [6077] = 6077,
  [6078] = 6001,
  [6079] = 5961,
  [6080] = 6080,
  [6081] = 6081,
  [6082] = 5957,
  [6083] = 5987,
  [6084] = 6068,
  [6085] = 6085,
  [6086] = 5961,
  [6087] = 6087,
  [6088] = 6088,
  [6089] = 6085,
  [6090] = 5947,
  [6091] = 5986,
  [6092] = 5946,
  [6093] = 5958,
  [6094] = 5951,
  [6095] = 6095,
  [6096] = 5959,
  [6097] = 5960,
  [6098] = 5954,
  [6099] = 5949,
  [6100] = 6100,
  [6101] = 6101,
  [6102] = 5961,
  [6103] = 5972,
  [6104] = 6007,
  [6105] = 6105,
  [6106] = 5960,
  [6107] = 6007,
  [6108] = 6108,
  [6109] = 5947,
  [6110] = 5947,
  [6111] = 5947,
  [6112] = 5961,
  [6113] = 6113,
  [6114] = 6114,
  [6115] = 6115,
  [6116] = 6047,
  [6117] = 6117,
  [6118] = 5983,
  [6119] = 5951,
  [6120] = 6120,
  [6121] = 6121,
  [6122] = 6115,
  [6123] = 5961,
  [6124] = 5983,
  [6125] = 5958,
  [6126] = 5947,
  [6127] = 5945,
  [6128] = 6117,
  [6129] = 5951,
  [6130] = 5984,
  [6131] = 5961,
  [6132] = 5951,
  [6133] = 6133,
  [6134] = 5946,
  [6135] = 5946,
  [6136] = 5947,
  [6137] = 5947,
  [6138] = 5946,
  [6139] = 5951,
  [6140] = 5950,
  [6141] = 5961,
  [6142] = 5946,
  [6143] = 6055,
  [6144] = 5952,
  [6145] = 5964,
  [6146] = 6007,
  [6147] = 1878,
  [6148] = 5947,
  [6149] = 5947,
  [6150] = 6001,
  [6151] = 5951,
  [6152] = 6152,
  [6153] = 5961,
  [6154] = 5947,
  [6155] = 6007,
  [6156] = 6007,
  [6157] = 5951,
  [6158] = 5961,
  [6159] = 5972,
  [6160] = 5959,
  [6161] = 5960,
  [6162] = 5961,
  [6163] = 6081,
  [6164] = 5972,
  [6165] = 5947,
  [6166] = 5982,
  [6167] = 5972,
  [6168] = 5948,
  [6169] = 6169,
  [6170] = 6170,
  [6171] = 6068,
  [6172] = 6172,
  [6173] = 6121,
  [6174] = 5949,
  [6175] = 5957,
  [6176] = 5947,
  [6177] = 6077,
  [6178] = 5972,
  [6179] = 5947,
  [6180] = 5948,
  [6181] = 6087,
  [6182] = 5956,
  [6183] = 6101,
  [6184] = 6108,
  [6185] = 3941,
  [6186] = 6007,
  [6187] = 5999,
  [6188] = 5946,
  [6189] = 6189,
  [6190] = 6115,
  [6191] = 5961,
  [6192] = 6047,
  [6193] = 6041,
  [6194] = 5953,
  [6195] = 6195,
  [6196] = 6077,
  [6197] = 6108,
  [6198] = 5951,
  [6199] = 6199,
  [6200] = 6058,
  [6201] = 6074,
  [6202] = 5987,
  [6203] = 6203,
  [6204] = 5947,
  [6205] = 5999,
  [6206] = 5958,
  [6207] = 6115,
  [6208] = 6047,
  [6209] = 6209,
  [6210] = 5959,
  [6211] = 5972,
  [6212] = 5960,
  [6213] = 5953,
  [6214] = 5954,
  [6215] = 5949,
  [6216] = 5971,
  [6217] = 5982,
  [6218] = 3940,
  [6219] = 6071,
  [6220] = 6001,
  [6221] = 5945,
  [6222] = 5946,
  [6223] = 5964,
  [6224] = 6224,
  [6225] = 5945,
  [6226] = 5992,
  [6227] = 5966,
  [6228] = 6041,
  [6229] = 6074,
  [6230] = 5946,
  [6231] = 2127,
  [6232] = 6232,
  [6233] = 5945,
  [6234] = 6234,
  [6235] = 6087,
  [6236] = 6117,
  [6237] = 6007,
  [6238] = 6203,
  [6239] = 5988,
  [6240] = 5999,
  [6241] = 5955,
  [6242] = 5950,
  [6243] = 5972,
  [6244] = 6203,
  [6245] = 5951,
  [6246] = 6234,
  [6247] = 6209,
  [6248] = 5999,
  [6249] = 6087,
  [6250] = 6071,
  [6251] = 6105,
  [6252] = 5972,
  [6253] = 5946,
  [6254] = 5948,
  [6255] = 6255,
  [6256] = 6101,
  [6257] = 6108,
  [6258] = 3915,
  [6259] = 5973,
  [6260] = 6007,
  [6261] = 6261,
  [6262] = 5951,
  [6263] = 6041,
  [6264] = 5953,
  [6265] = 5961,
  [6266] = 6121,
  [6267] = 6074,
  [6268] = 5947,
  [6269] = 5983,
  [6270] = 6006,
  [6271] = 6068,
  [6272] = 6272,
  [6273] = 6273,
  [6274] = 5952,
  [6275] = 6077,
  [6276] = 6189,
  [6277] = 5958,
  [6278] = 5959,
  [6279] = 5960,
  [6280] = 5954,
  [6281] = 5950,
  [6282] = 6121,
  [6283] = 5976,
  [6284] = 5948,
  [6285] = 5966,
  [6286] = 6007,
  [6287] = 6272,
  [6288] = 6232,
  [6289] = 5995,
  [6290] = 5956,
  [6291] = 6291,
  [6292] = 5955,
  [6293] = 6121,
  [6294] = 6041,
  [6295] = 5953,
  [6296] = 5984,
  [6297] = 5952,
  [6298] = 3938,
  [6299] = 5999,
  [6300] = 5966,
  [6301] = 5958,
  [6302] = 5959,
  [6303] = 5999,
  [6304] = 5960,
  [6305] = 5954,
  [6306] = 6041,
  [6307] = 5954,
  [6308] = 5951,
  [6309] = 6095,
  [6310] = 5951,
  [6311] = 6311,
  [6312] = 5955,
  [6313] = 5948,
  [6314] = 6189,
  [6315] = 5961,
  [6316] = 5957,
  [6317] = 6291,
  [6318] = 5976,
  [6319] = 5961,
  [6320] = 5990,
  [6321] = 5946,
  [6322] = 6055,
  [6323] = 6041,
  [6324] = 5953,
  [6325] = 5986,
  [6326] = 6048,
  [6327] = 5961,
  [6328] = 5990,
  [6329] = 5948,
  [6330] = 6074,
  [6331] = 5951,
  [6332] = 5958,
  [6333] = 6333,
  [6334] = 6334,
  [6335] = 6335,
  [6336] = 6336,
  [6337] = 6337,
  [6338] = 6338,
  [6339] = 2657,
  [6340] = 6340,
  [6341] = 6341,
  [6342] = 2664,
  [6343] = 6343,
  [6344] = 6344,
  [6345] = 6345,
  [6346] = 6346,
  [6347] = 6347,
  [6348] = 6348,
  [6349] = 6349,
  [6350] = 6350,
  [6351] = 6351,
  [6352] = 6349,
  [6353] = 6353,
  [6354] = 6354,
  [6355] = 6355,
  [6356] = 6341,
  [6357] = 6357,
  [6358] = 6358,
  [6359] = 6357,
  [6360] = 6344,
  [6361] = 6361,
  [6362] = 6345,
  [6363] = 6363,
  [6364] = 6364,
  [6365] = 6365,
  [6366] = 6366,
  [6367] = 6361,
  [6368] = 6368,
  [6369] = 6369,
  [6370] = 6370,
  [6371] = 6371,
  [6372] = 6372,
  [6373] = 6373,
  [6374] = 6374,
  [6375] = 6375,
  [6376] = 6341,
  [6377] = 6370,
  [6378] = 6378,
  [6379] = 6379,
  [6380] = 6380,
  [6381] = 6381,
  [6382] = 6382,
  [6383] = 6383,
  [6384] = 6374,
  [6385] = 6344,
  [6386] = 6386,
  [6387] = 6387,
  [6388] = 6388,
  [6389] = 6389,
  [6390] = 1507,
  [6391] = 6391,
  [6392] = 6392,
  [6393] = 6378,
  [6394] = 6394,
  [6395] = 6395,
  [6396] = 6396,
  [6397] = 6397,
  [6398] = 6398,
  [6399] = 6345,
  [6400] = 6363,
  [6401] = 6381,
  [6402] = 6338,
  [6403] = 6382,
  [6404] = 6404,
  [6405] = 6405,
  [6406] = 6406,
  [6407] = 6347,
  [6408] = 6346,
  [6409] = 6404,
  [6410] = 6410,
  [6411] = 6396,
  [6412] = 6412,
  [6413] = 6404,
  [6414] = 6414,
  [6415] = 6415,
  [6416] = 6416,
  [6417] = 6336,
  [6418] = 6383,
  [6419] = 6419,
  [6420] = 6420,
  [6421] = 6421,
  [6422] = 6422,
  [6423] = 6335,
  [6424] = 6424,
  [6425] = 6364,
  [6426] = 6426,
  [6427] = 6388,
  [6428] = 6428,
  [6429] = 6415,
  [6430] = 6371,
  [6431] = 6431,
  [6432] = 6432,
  [6433] = 6433,
  [6434] = 6434,
  [6435] = 6435,
  [6436] = 6431,
  [6437] = 6437,
  [6438] = 6383,
  [6439] = 6394,
  [6440] = 6440,
  [6441] = 6441,
  [6442] = 6419,
  [6443] = 6420,
  [6444] = 6444,
  [6445] = 6338,
  [6446] = 6343,
  [6447] = 6447,
  [6448] = 6398,
  [6449] = 6398,
  [6450] = 6349,
  [6451] = 6369,
  [6452] = 6351,
  [6453] = 6453,
  [6454] = 6353,
  [6455] = 6354,
  [6456] = 6355,
  [6457] = 6373,
  [6458] = 6458,
  [6459] = 6340,
  [6460] = 6460,
  [6461] = 6361,
  [6462] = 6341,
  [6463] = 6463,
  [6464] = 6464,
  [6465] = 6444,
  [6466] = 6466,
  [6467] = 6467,
  [6468] = 1503,
  [6469] = 6469,
  [6470] = 6470,
  [6471] = 6471,
  [6472] = 6472,
  [6473] = 6473,
  [6474] = 6373,
  [6475] = 6441,
  [6476] = 6476,
  [6477] = 6344,
  [6478] = 6478,
  [6479] = 6422,
  [6480] = 6422,
  [6481] = 6364,
  [6482] = 6345,
  [6483] = 6483,
  [6484] = 6483,
  [6485] = 6485,
  [6486] = 6366,
  [6487] = 6365,
  [6488] = 6369,
  [6489] = 6361,
  [6490] = 6391,
  [6491] = 6392,
  [6492] = 6371,
  [6493] = 6493,
  [6494] = 2659,
  [6495] = 6495,
  [6496] = 6496,
  [6497] = 6370,
  [6498] = 6374,
  [6499] = 6499,
  [6500] = 6500,
  [6501] = 6405,
  [6502] = 6347,
  [6503] = 6398,
  [6504] = 6396,
  [6505] = 6404,
  [6506] = 6335,
  [6507] = 6460,
  [6508] = 6383,
  [6509] = 6419,
  [6510] = 6420,
  [6511] = 6424,
  [6512] = 6512,
  [6513] = 6513,
  [6514] = 6426,
  [6515] = 6496,
  [6516] = 6453,
  [6517] = 6391,
  [6518] = 6485,
  [6519] = 6432,
  [6520] = 6433,
  [6521] = 6337,
  [6522] = 6522,
  [6523] = 6371,
  [6524] = 6437,
  [6525] = 6426,
  [6526] = 6419,
  [6527] = 6357,
  [6528] = 6410,
  [6529] = 6338,
  [6530] = 6343,
  [6531] = 6387,
  [6532] = 6532,
  [6533] = 6406,
  [6534] = 6534,
  [6535] = 6373,
  [6536] = 6351,
  [6537] = 6353,
  [6538] = 6354,
  [6539] = 6355,
  [6540] = 6387,
  [6541] = 6363,
  [6542] = 6344,
  [6543] = 6543,
  [6544] = 6447,
  [6545] = 6545,
  [6546] = 6340,
  [6547] = 6547,
  [6548] = 6347,
  [6549] = 6335,
  [6550] = 6396,
  [6551] = 6404,
  [6552] = 6387,
  [6553] = 6545,
  [6554] = 6383,
  [6555] = 6419,
  [6556] = 6420,
  [6557] = 6350,
  [6558] = 6349,
  [6559] = 6375,
  [6560] = 6560,
  [6561] = 6392,
  [6562] = 6340,
  [6563] = 6432,
  [6564] = 6433,
  [6565] = 6412,
  [6566] = 6416,
  [6567] = 6341,
  [6568] = 6437,
  [6569] = 6569,
  [6570] = 6344,
  [6571] = 6345,
  [6572] = 6432,
  [6573] = 6343,
  [6574] = 6421,
  [6575] = 6370,
  [6576] = 6433,
  [6577] = 6483,
  [6578] = 6361,
  [6579] = 6351,
  [6580] = 6353,
  [6581] = 6354,
  [6582] = 6355,
  [6583] = 6370,
  [6584] = 6584,
  [6585] = 6374,
  [6586] = 6437,
  [6587] = 6398,
  [6588] = 6374,
  [6589] = 6371,
  [6590] = 6412,
  [6591] = 6347,
  [6592] = 6592,
  [6593] = 6396,
  [6594] = 6404,
  [6595] = 6395,
  [6596] = 6416,
  [6597] = 6383,
  [6598] = 6419,
  [6599] = 6420,
  [6600] = 6397,
  [6601] = 6601,
  [6602] = 6460,
  [6603] = 6421,
  [6604] = 6431,
  [6605] = 6432,
  [6606] = 6433,
  [6607] = 6371,
  [6608] = 6466,
  [6609] = 6609,
  [6610] = 6437,
  [6611] = 6466,
  [6612] = 6340,
  [6613] = 6500,
  [6614] = 6343,
  [6615] = 6412,
  [6616] = 6364,
  [6617] = 6617,
  [6618] = 6464,
  [6619] = 6351,
  [6620] = 6353,
  [6621] = 6354,
  [6622] = 6355,
  [6623] = 6467,
  [6624] = 6471,
  [6625] = 6416,
  [6626] = 6421,
  [6627] = 6472,
  [6628] = 6628,
  [6629] = 6629,
  [6630] = 6366,
  [6631] = 6347,
  [6632] = 6349,
  [6633] = 6396,
  [6634] = 6404,
  [6635] = 6635,
  [6636] = 6378,
  [6637] = 6383,
  [6638] = 6419,
  [6639] = 6420,
  [6640] = 6364,
  [6641] = 3729,
  [6642] = 6364,
  [6643] = 6369,
  [6644] = 6432,
  [6645] = 6433,
  [6646] = 6373,
  [6647] = 6414,
  [6648] = 6369,
  [6649] = 6437,
  [6650] = 6650,
  [6651] = 6584,
  [6652] = 6652,
  [6653] = 6343,
  [6654] = 6369,
  [6655] = 6463,
  [6656] = 6391,
  [6657] = 6657,
  [6658] = 6351,
  [6659] = 6353,
  [6660] = 6354,
  [6661] = 6355,
  [6662] = 6441,
  [6663] = 6392,
  [6664] = 6387,
  [6665] = 6444,
  [6666] = 6534,
  [6667] = 6460,
  [6668] = 6483,
  [6669] = 6347,
  [6670] = 6670,
  [6671] = 6396,
  [6672] = 6404,
  [6673] = 6673,
  [6674] = 6453,
  [6675] = 6383,
  [6676] = 6419,
  [6677] = 6420,
  [6678] = 6373,
  [6679] = 6431,
  [6680] = 6412,
  [6681] = 6416,
  [6682] = 6432,
  [6683] = 6433,
  [6684] = 6495,
  [6685] = 6338,
  [6686] = 6421,
  [6687] = 6437,
  [6688] = 6476,
  [6689] = 6348,
  [6690] = 6368,
  [6691] = 6343,
  [6692] = 6341,
  [6693] = 6431,
  [6694] = 6694,
  [6695] = 6469,
  [6696] = 6351,
  [6697] = 6353,
  [6698] = 6354,
  [6699] = 6355,
  [6700] = 6700,
  [6701] = 6387,
  [6702] = 6476,
  [6703] = 6512,
  [6704] = 6704,
  [6705] = 6705,
  [6706] = 6706,
  [6707] = 6512,
  [6708] = 6464,
  [6709] = 6467,
  [6710] = 6383,
  [6711] = 6419,
  [6712] = 6420,
  [6713] = 6343,
  [6714] = 6714,
  [6715] = 6715,
  [6716] = 6471,
  [6717] = 6432,
  [6718] = 6433,
  [6719] = 6472,
  [6720] = 6720,
  [6721] = 6391,
  [6722] = 6437,
  [6723] = 6365,
  [6724] = 6337,
  [6725] = 2663,
  [6726] = 6343,
  [6727] = 6392,
  [6728] = 6460,
  [6729] = 6381,
  [6730] = 6512,
  [6731] = 6351,
  [6732] = 6353,
  [6733] = 6354,
  [6734] = 6355,
  [6735] = 6735,
  [6736] = 6485,
  [6737] = 6387,
  [6738] = 6378,
  [6739] = 6406,
  [6740] = 6424,
  [6741] = 6720,
  [6742] = 6344,
  [6743] = 6373,
  [6744] = 6420,
  [6745] = 6483,
  [6746] = 6584,
  [6747] = 6437,
  [6748] = 6412,
  [6749] = 6416,
  [6750] = 6476,
  [6751] = 6343,
  [6752] = 6496,
  [6753] = 6351,
  [6754] = 6353,
  [6755] = 6354,
  [6756] = 6355,
  [6757] = 6757,
  [6758] = 6421,
  [6759] = 6720,
  [6760] = 6760,
  [6761] = 6706,
  [6762] = 6336,
  [6763] = 6345,
  [6764] = 6464,
  [6765] = 6760,
  [6766] = 6431,
  [6767] = 6353,
  [6768] = 6354,
  [6769] = 6355,
  [6770] = 6460,
  [6771] = 6512,
  [6772] = 6357,
  [6773] = 6381,
  [6774] = 6774,
  [6775] = 6496,
  [6776] = 6467,
  [6777] = 6471,
  [6778] = 6353,
  [6779] = 6354,
  [6780] = 6355,
  [6781] = 6382,
  [6782] = 6351,
  [6783] = 6472,
  [6784] = 6353,
  [6785] = 6354,
  [6786] = 6355,
  [6787] = 6363,
  [6788] = 6353,
  [6789] = 6354,
  [6790] = 6355,
  [6791] = 6496,
  [6792] = 6363,
  [6793] = 6435,
  [6794] = 6336,
  [6795] = 6340,
  [6796] = 6348,
  [6797] = 6797,
  [6798] = 6353,
  [6799] = 6354,
  [6800] = 6355,
  [6801] = 6387,
  [6802] = 6545,
  [6803] = 6464,
  [6804] = 6344,
  [6805] = 6522,
  [6806] = 6345,
  [6807] = 6382,
  [6808] = 6353,
  [6809] = 6354,
  [6810] = 6355,
  [6811] = 6370,
  [6812] = 6412,
  [6813] = 6416,
  [6814] = 6421,
  [6815] = 6374,
  [6816] = 6380,
  [6817] = 6398,
  [6818] = 6353,
  [6819] = 6354,
  [6820] = 6355,
  [6821] = 6467,
  [6822] = 6471,
  [6823] = 6368,
  [6824] = 6340,
  [6825] = 6472,
  [6826] = 6584,
  [6827] = 6369,
  [6828] = 6361,
  [6829] = 6387,
  [6830] = 6512,
  [6831] = 6335,
  [6832] = 6395,
  [6833] = 6545,
  [6834] = 6351,
  [6835] = 6361,
  [6836] = 6395,
  [6837] = 6466,
  [6838] = 6349,
  [6839] = 6431,
  [6840] = 6532,
  [6841] = 6341,
  [6842] = 6657,
  [6843] = 6370,
  [6844] = 6844,
  [6845] = 6344,
  [6846] = 6345,
  [6847] = 6374,
  [6848] = 6373,
  [6849] = 6472,
  [6850] = 6601,
  [6851] = 6851,
  [6852] = 6852,
  [6853] = 6350,
  [6854] = 6361,
  [6855] = 6483,
  [6856] = 6478,
  [6857] = 6370,
  [6858] = 6336,
  [6859] = 6374,
  [6860] = 6860,
  [6861] = 6483,
  [6862] = 6382,
  [6863] = 6863,
  [6864] = 6398,
  [6865] = 6617,
  [6866] = 6483,
  [6867] = 6720,
  [6868] = 6398,
  [6869] = 6869,
  [6870] = 6470,
  [6871] = 6851,
  [6872] = 6852,
  [6873] = 6336,
  [6874] = 6460,
  [6875] = 6464,
  [6876] = 6371,
  [6877] = 6467,
  [6878] = 6471,
  [6879] = 6512,
  [6880] = 6447,
  [6881] = 6495,
  [6882] = 6396,
  [6883] = 6340,
  [6884] = 6472,
  [6885] = 6353,
  [6886] = 6343,
  [6887] = 6354,
  [6888] = 6363,
  [6889] = 6466,
  [6890] = 6355,
  [6891] = 6364,
  [6892] = 6617,
  [6893] = 6344,
  [6894] = 6757,
  [6895] = 6485,
  [6896] = 6464,
  [6897] = 6441,
  [6898] = 6345,
  [6899] = 6899,
  [6900] = 6375,
  [6901] = 6389,
  [6902] = 6370,
  [6903] = 6374,
  [6904] = 6904,
  [6905] = 6467,
  [6906] = 6364,
  [6907] = 6471,
  [6908] = 6704,
  [6909] = 6472,
  [6910] = 6369,
  [6911] = 6673,
  [6912] = 6369,
  [6913] = 6469,
  [6914] = 6628,
  [6915] = 6629,
  [6916] = 6916,
  [6917] = 6917,
  [6918] = 6334,
  [6919] = 6919,
  [6920] = 6760,
  [6921] = 6921,
  [6922] = 6472,
  [6923] = 6774,
  [6924] = 6924,
  [6925] = 6434,
  [6926] = 6358,
  [6927] = 6431,
  [6928] = 6928,
  [6929] = 6635,
  [6930] = 6670,
  [6931] = 6335,
  [6932] = 6373,
  [6933] = 6483,
  [6934] = 6715,
  [6935] = 6428,
  [6936] = 6460,
  [6937] = 6512,
  [6938] = 6938,
  [6939] = 6371,
  [6940] = 6387,
  [6941] = 6919,
  [6942] = 6420,
  [6943] = 6363,
  [6944] = 6412,
  [6945] = 6416,
  [6946] = 6421,
  [6947] = 6797,
  [6948] = 6924,
  [6949] = 6344,
  [6950] = 6345,
  [6951] = 6584,
  [6952] = 6370,
  [6953] = 6374,
  [6954] = 6406,
  [6955] = 6426,
  [6956] = 6369,
  [6957] = 6496,
  [6958] = 6472,
  [6959] = 6431,
  [6960] = 6960,
  [6961] = 6340,
  [6962] = 6962,
  [6963] = 6601,
  [6964] = 6463,
  [6965] = 6336,
  [6966] = 6444,
  [6967] = 6386,
  [6968] = 6673,
  [6969] = 6969,
  [6970] = 6483,
  [6971] = 6617,
  [6972] = 6428,
  [6973] = 4141,
  [6974] = 6469,
  [6975] = 6464,
  [6976] = 6467,
  [6977] = 6471,
  [6978] = 6472,
  [6979] = 6432,
  [6980] = 6469,
  [6981] = 6628,
  [6982] = 6460,
  [6983] = 6512,
  [6984] = 6483,
  [6985] = 6466,
  [6986] = 6391,
  [6987] = 6987,
  [6988] = 6433,
  [6989] = 6388,
  [6990] = 6990,
  [6991] = 6460,
  [6992] = 6366,
  [6993] = 6391,
  [6994] = 6363,
  [6995] = 6344,
  [6996] = 6345,
  [6997] = 6391,
  [6998] = 6392,
  [6999] = 6369,
  [7000] = 6860,
  [7001] = 7001,
  [7002] = 6496,
  [7003] = 6394,
  [7004] = 6347,
  [7005] = 6370,
  [7006] = 6483,
  [7007] = 6374,
  [7008] = 6398,
  [7009] = 6369,
  [7010] = 6340,
  [7011] = 6387,
  [7012] = 6512,
  [7013] = 6860,
  [7014] = 6628,
  [7015] = 7015,
  [7016] = 6472,
  [7017] = 6617,
  [7018] = 6464,
  [7019] = 6467,
  [7020] = 6469,
  [7021] = 6471,
  [7022] = 6463,
  [7023] = 6472,
  [7024] = 7024,
  [7025] = 6706,
  [7026] = 6483,
  [7027] = 6357,
  [7028] = 6496,
  [7029] = 6392,
  [7030] = 6345,
  [7031] = 6460,
  [7032] = 6512,
  [7033] = 7024,
  [7034] = 6460,
  [7035] = 6363,
  [7036] = 6344,
  [7037] = 6569,
  [7038] = 6629,
  [7039] = 7039,
  [7040] = 6860,
  [7041] = 6345,
  [7042] = 6470,
  [7043] = 6652,
  [7044] = 6391,
  [7045] = 6370,
  [7046] = 6392,
  [7047] = 6543,
  [7048] = 6374,
  [7049] = 6412,
  [7050] = 6369,
  [7051] = 6460,
  [7052] = 6928,
  [7053] = 6635,
  [7054] = 6389,
  [7055] = 6512,
  [7056] = 6904,
  [7057] = 7039,
  [7058] = 6472,
  [7059] = 6760,
  [7060] = 6916,
  [7061] = 6917,
  [7062] = 6334,
  [7063] = 6483,
  [7064] = 6336,
  [7065] = 6357,
  [7066] = 6720,
  [7067] = 6499,
  [7068] = 6496,
  [7069] = 6416,
  [7070] = 6460,
  [7071] = 6512,
  [7072] = 6410,
  [7073] = 6363,
  [7074] = 7074,
  [7075] = 6470,
  [7076] = 6363,
  [7077] = 6532,
  [7078] = 6336,
  [7079] = 6398,
  [7080] = 6421,
  [7081] = 6357,
  [7082] = 6344,
  [7083] = 6928,
  [7084] = 7084,
  [7085] = 6363,
  [7086] = 6904,
  [7087] = 6345,
  [7088] = 6370,
  [7089] = 6916,
  [7090] = 6917,
  [7091] = 6334,
  [7092] = 6495,
  [7093] = 6921,
  [7094] = 7001,
  [7095] = 6670,
  [7096] = 6714,
  [7097] = 6335,
  [7098] = 6469,
  [7099] = 6374,
  [7100] = 6369,
  [7101] = 6428,
  [7102] = 6357,
  [7103] = 6472,
  [7104] = 6628,
  [7105] = 6460,
  [7106] = 6904,
  [7107] = 6483,
  [7108] = 6916,
  [7109] = 6917,
  [7110] = 6334,
  [7111] = 6349,
  [7112] = 6349,
  [7113] = 6364,
  [7114] = 6391,
  [7115] = 6347,
  [7116] = 6415,
  [7117] = 6341,
  [7118] = 4304,
  [7119] = 6392,
  [7120] = 6388,
  [7121] = 6904,
  [7122] = 6496,
  [7123] = 6916,
  [7124] = 6917,
  [7125] = 6334,
  [7126] = 6363,
  [7127] = 6344,
  [7128] = 6345,
  [7129] = 3667,
  [7130] = 6460,
  [7131] = 6560,
  [7132] = 6512,
  [7133] = 6340,
  [7134] = 6938,
  [7135] = 6899,
  [7136] = 6513,
  [7137] = 6916,
  [7138] = 6917,
  [7139] = 6334,
  [7140] = 6398,
  [7141] = 6363,
  [7142] = 7142,
  [7143] = 6361,
  [7144] = 6340,
  [7145] = 6344,
  [7146] = 6478,
  [7147] = 6370,
  [7148] = 6345,
  [7149] = 6374,
  [7150] = 6370,
  [7151] = 6916,
  [7152] = 6917,
  [7153] = 6334,
  [7154] = 6374,
  [7155] = 7155,
  [7156] = 6382,
  [7157] = 6369,
  [7158] = 7158,
  [7159] = 6629,
  [7160] = 6398,
  [7161] = 6472,
  [7162] = 6387,
  [7163] = 6500,
  [7164] = 6363,
  [7165] = 6916,
  [7166] = 6917,
  [7167] = 6334,
  [7168] = 6483,
  [7169] = 6357,
  [7170] = 6904,
  [7171] = 6512,
  [7172] = 1496,
  [7173] = 6370,
  [7174] = 6391,
  [7175] = 6363,
  [7176] = 6392,
  [7177] = 6371,
  [7178] = 6547,
  [7179] = 6916,
  [7180] = 6917,
  [7181] = 6334,
  [7182] = 6335,
  [7183] = 6635,
  [7184] = 6512,
  [7185] = 6512,
  [7186] = 6916,
  [7187] = 6917,
  [7188] = 6334,
  [7189] = 6335,
  [7190] = 6363,
  [7191] = 6916,
  [7192] = 6917,
  [7193] = 6334,
  [7194] = 6345,
  [7195] = 6370,
  [7196] = 6916,
  [7197] = 6917,
  [7198] = 6334,
  [7199] = 6374,
  [7200] = 6469,
  [7201] = 6916,
  [7202] = 6917,
  [7203] = 6334,
  [7204] = 6369,
  [7205] = 6391,
  [7206] = 6916,
  [7207] = 6917,
  [7208] = 6334,
  [7209] = 6392,
  [7210] = 6512,
  [7211] = 6916,
  [7212] = 6917,
  [7213] = 2670,
  [7214] = 6349,
  [7215] = 6391,
  [7216] = 6916,
  [7217] = 6917,
  [7218] = 6334,
  [7219] = 6392,
  [7220] = 6370,
  [7221] = 6374,
  [7222] = 6391,
  [7223] = 6374,
  [7224] = 6392,
  [7225] = 6460,
  [7226] = 6369,
  [7227] = 7015,
  [7228] = 6496,
  [7229] = 7229,
  [7230] = 6512,
  [7231] = 7074,
  [7232] = 6512,
  [7233] = 6370,
  [7234] = 6340,
  [7235] = 6336,
  [7236] = 6469,
  [7237] = 6916,
  [7238] = 6437,
  [7239] = 7239,
  [7240] = 6347,
  [7241] = 6350,
  [7242] = 6673,
  [7243] = 6374,
  [7244] = 6369,
  [7245] = 6341,
  [7246] = 6466,
  [7247] = 6652,
  [7248] = 6396,
  [7249] = 6469,
  [7250] = 6917,
  [7251] = 6347,
  [7252] = 6404,
  [7253] = 6364,
  [7254] = 6705,
  [7255] = 6369,
  [7256] = 6414,
  [7257] = 6720,
  [7258] = 6336,
  [7259] = 6344,
  [7260] = 6334,
  [7261] = 6345,
  [7262] = 7015,
  [7263] = 7229,
  [7264] = 6373,
  [7265] = 6392,
  [7266] = 7229,
  [7267] = 6415,
  [7268] = 7015,
  [7269] = 7229,
  [7270] = 7229,
  [7271] = 7229,
  [7272] = 7229,
  [7273] = 7229,
  [7274] = 7229,
  [7275] = 7229,
  [7276] = 7229,
  [7277] = 7229,
  [7278] = 7229,
  [7279] = 7229,
  [7280] = 7229,
  [7281] = 7229,
  [7282] = 6394,
  [7283] = 6496,
  [7284] = 6383,
  [7285] = 6496,
  [7286] = 7286,
  [7287] = 6419,
  [7288] = 6420,
  [7289] = 6361,
  [7290] = 6387,
  [7291] = 6938,
  [7292] = 6447,
  [7293] = 6422,
  [7294] = 6412,
  [7295] = 6416,
  [7296] = 6421,
  [7297] = 6369,
  [7298] = 6609,
  [7299] = 6424,
  [7300] = 6584,
  [7301] = 6370,
  [7302] = 6431,
  [7303] = 6426,
  [7304] = 6374,
  [7305] = 6335,
  [7306] = 6547,
  [7307] = 6341,
  [7308] = 6336,
  [7309] = 6617,
  [7310] = 6375,
  [7311] = 6464,
  [7312] = 6467,
  [7313] = 6337,
  [7314] = 6471,
  [7315] = 6472,
  [7316] = 6628,
  [7317] = 6357,
  [7318] = 7229,
  [7319] = 6398,
  [7320] = 6336,
  [7321] = 6363,
  [7322] = 6432,
  [7323] = 6513,
  [7324] = 6433,
  [7325] = 6774,
  [7326] = 6493,
  [7327] = 6349,
  [7328] = 7015,
  [7329] = 6437,
  [7330] = 6483,
  [7331] = 6860,
  [7332] = 6371,
  [7333] = 6466,
  [7334] = 6928,
  [7335] = 7335,
  [7336] = 7336,
};

static const TSSymbol ts_supertype_symbols[SUPERTYPE_COUNT] = {
  sym__binding_pattern,
  sym__binding_pattern_no_exn,
  sym__class_expression,
  sym__class_field,
  sym__class_field_specification,
  sym__class_type,
  sym__constant,
  sym__delimited_binding_pattern,
  sym__delimited_expression,
  sym__delimited_pattern,
  sym__delimited_type,
  sym__effect_binding_pattern,
  sym__effect_pattern,
  sym__expression,
  sym__infix_operator,
  sym__local_structure_item,
  sym__module_expression,
  sym__module_type,
  sym__pattern,
  sym__polymorphic_type,
  sym__sequence_expression,
  sym__signature_item,
  sym__signed_constant,
  sym__simple_binding_pattern,
  sym__simple_class_expression,
  sym__simple_class_type,
  sym__simple_expression,
  sym__simple_module_expression,
  sym__simple_pattern,
  sym__simple_type,
  sym__structure_item,
  sym__type,
};

static const TSMapSlice ts_supertype_map_slices[] = {
  [sym__binding_pattern] = {.index = 0, .length = 14},
  [sym__binding_pattern_no_exn] = {.index = 14, .length = 10},
  [sym__class_expression] = {.index = 24, .length = 5},
  [sym__class_field] = {.index = 29, .length = 7},
  [sym__class_field_specification] = {.index = 36, .length = 6},
  [sym__class_type] = {.index = 42, .length = 2},
  [sym__constant] = {.index = 44, .length = 5},
  [sym__delimited_binding_pattern] = {.index = 49, .length = 9},
  [sym__delimited_expression] = {.index = 58, .length = 7},
  [sym__delimited_pattern] = {.index = 65, .length = 5},
  [sym__delimited_type] = {.index = 70, .length = 3},
  [sym__effect_binding_pattern] = {.index = 73, .length = 7},
  [sym__effect_pattern] = {.index = 80, .length = 4},
  [sym__expression] = {.index = 84, .length = 17},
  [sym__infix_operator] = {.index = 101, .length = 8},
  [sym__local_structure_item] = {.index = 109, .length = 11},
  [sym__module_expression] = {.index = 120, .length = 5},
  [sym__module_type] = {.index = 125, .length = 8},
  [sym__pattern] = {.index = 133, .length = 8},
  [sym__polymorphic_type] = {.index = 141, .length = 2},
  [sym__sequence_expression] = {.index = 143, .length = 2},
  [sym__signature_item] = {.index = 145, .length = 13},
  [sym__signed_constant] = {.index = 158, .length = 2},
  [sym__simple_binding_pattern] = {.index = 160, .length = 14},
  [sym__simple_class_expression] = {.index = 174, .length = 7},
  [sym__simple_class_type] = {.index = 181, .length = 6},
  [sym__simple_expression] = {.index = 187, .length = 19},
  [sym__simple_module_expression] = {.index = 206, .length = 5},
  [sym__simple_pattern] = {.index = 211, .length = 12},
  [sym__simple_type] = {.index = 223, .length = 9},
  [sym__structure_item] = {.index = 232, .length = 4},
  [sym__type] = {.index = 236, .length = 5},
};

static const TSSymbol ts_supertype_map_entries[] = {
  [0] =
    sym__effect_binding_pattern,
    sym__or_binding_pattern_no_exn_anonymous,
    sym__tuple_binding_pattern_no_exn,
    sym_alias_binding_pattern,
    sym_alias_binding_pattern_no_exn,
    sym_alias_pattern,
    sym_cons_binding_pattern,
    sym_cons_binding_pattern_no_exn,
    sym_cons_pattern,
    sym_effect_binding_pattern,
    sym_effect_pattern,
    sym_exception_binding_pattern,
    sym_exception_pattern,
    sym_range_pattern,
  [14] =
    sym__effect_binding_pattern,
    sym__or_binding_pattern_no_exn_anonymous,
    sym__tuple_binding_pattern_no_exn,
    sym_alias_binding_pattern,
    sym_alias_binding_pattern_no_exn,
    sym_alias_pattern,
    sym_cons_binding_pattern,
    sym_cons_binding_pattern_no_exn,
    sym_cons_pattern,
    sym_range_pattern,
  [24] =
    sym__simple_class_expression,
    sym_class_application,
    sym_class_function,
    sym_let_class_expression,
    sym_let_open_class_expression,
  [29] =
    sym_class_initializer,
    sym_inheritance_definition,
    sym_instance_variable_definition,
    sym_item_extension,
    sym_method_definition,
    sym_quoted_item_extension,
    sym_type_parameter_constraint,
  [36] =
    sym_inheritance_specification,
    sym_instance_variable_specification,
    sym_item_extension,
    sym_method_specification,
    sym_quoted_item_extension,
    sym_type_parameter_constraint,
  [42] =
    sym__simple_class_type,
    sym_class_function_type,
  [44] =
    sym_boolean,
    sym_character,
    sym_number,
    sym_quoted_string,
    sym_string,
  [49] =
    sym_array_binding_pattern,
    sym_array_pattern,
    sym_list_binding_pattern,
    sym_list_pattern,
    sym_parenthesized_binding_pattern,
    sym_parenthesized_pattern,
    sym_record_binding_pattern,
    sym_record_pattern,
    sym_unit,
  [58] =
    sym_array_expression,
    sym_list_expression,
    sym_object_copy_expression,
    sym_package_expression,
    sym_parenthesized_expression,
    sym_record_expression,
    sym_unit,
  [65] =
    sym_array_pattern,
    sym_list_pattern,
    sym_parenthesized_pattern,
    sym_record_pattern,
    sym_unit,
  [70] =
    sym_package_type,
    sym_parenthesized_type,
    sym_polymorphic_variant_type,
  [73] =
    sym__simple_binding_pattern,
    sym_constructor_binding_pattern,
    sym_constructor_pattern,
    sym_lazy_binding_pattern,
    sym_lazy_pattern,
    sym_tag_binding_pattern,
    sym_tag_pattern,
  [80] =
    sym__simple_pattern,
    sym_constructor_pattern,
    sym_lazy_pattern,
    sym_tag_pattern,
  [84] =
    alias_sym_tuple_expression,
    sym__simple_expression,
    sym_application_expression,
    sym_assert_expression,
    sym_cons_expression,
    sym_for_expression,
    sym_fun_expression,
    sym_function_expression,
    sym_if_expression,
    sym_infix_expression,
    sym_lazy_expression,
    sym_let_expression,
    sym_match_expression,
    sym_set_expression,
    sym_sign_expression,
    sym_try_expression,
    sym_while_expression,
  [101] =
    sym_add_operator,
    sym_and_operator,
    sym_assign_operator,
    sym_concat_operator,
    sym_mult_operator,
    sym_or_operator,
    sym_pow_operator,
    sym_rel_operator,
  [109] =
    sym_class_definition,
    sym_class_type_definition,
    sym_exception_definition,
    sym_external,
    sym_floating_attribute,
    sym_item_extension,
    sym_module_definition,
    sym_module_type_definition,
    sym_open_module,
    sym_quoted_item_extension,
    sym_type_definition,
  [120] =
    sym__simple_module_expression,
    sym_functor,
    sym_module_application,
    sym_module_path,
    sym_structure,
  [125] =
    sym_extension,
    sym_functor_type,
    sym_module_type_constraint,
    sym_module_type_of,
    sym_module_type_path,
    sym_parenthesized_module_type,
    sym_quoted_extension,
    sym_signature,
  [133] =
    sym__effect_pattern,
    sym__or_binding_pattern_no_exn_anonymous,
    sym__tuple_binding_pattern_no_exn,
    sym_alias_pattern,
    sym_cons_pattern,
    sym_effect_pattern,
    sym_exception_pattern,
    sym_range_pattern,
  [141] =
    sym__type,
    sym_polymorphic_type,
  [143] =
    alias_sym_sequence_expression,
    sym__expression,
  [145] =
    sym_class_definition,
    sym_class_type_definition,
    sym_exception_definition,
    sym_external,
    sym_floating_attribute,
    sym_include_module_type,
    sym_item_extension,
    sym_module_definition,
    sym_module_type_definition,
    sym_open_module,
    sym_quoted_item_extension,
    sym_type_definition,
    sym_value_specification,
  [158] =
    sym__constant,
    sym_signed_number,
  [160] =
    alias_sym_value_name,
    sym__delimited_binding_pattern,
    sym__signed_constant,
    sym_constructor_path,
    sym_extension,
    sym_local_open_binding_pattern,
    sym_local_open_pattern,
    sym_package_pattern,
    sym_parenthesized_operator,
    sym_polymorphic_variant_pattern,
    sym_quoted_extension,
    sym_tag,
    sym_typed_binding_pattern,
    sym_typed_pattern,
  [174] =
    sym_class_path,
    sym_extension,
    sym_instantiated_class,
    sym_object_expression,
    sym_parenthesized_class_expression,
    sym_quoted_extension,
    sym_typed_class_expression,
  [181] =
    sym_class_body_type,
    sym_class_type_path,
    sym_extension,
    sym_instantiated_class_type,
    sym_let_open_class_type,
    sym_quoted_extension,
  [187] =
    sym__constant,
    sym__delimited_expression,
    sym_array_get_expression,
    sym_bigarray_get_expression,
    sym_constructor_path,
    sym_extension,
    sym_field_get_expression,
    sym_hash_expression,
    sym_local_open_expression,
    sym_method_invocation,
    sym_new_expression,
    sym_object_expression,
    sym_ocamlyacc_value,
    sym_prefix_expression,
    sym_quoted_extension,
    sym_string_get_expression,
    sym_tag,
    sym_typed_expression,
    sym_value_path,
  [206] =
    sym_extension,
    sym_packed_module,
    sym_parenthesized_module_expression,
    sym_quoted_extension,
    sym_typed_module_expression,
  [211] =
    alias_sym_value_pattern,
    sym__delimited_pattern,
    sym__signed_constant,
    sym_constructor_path,
    sym_extension,
    sym_local_open_pattern,
    sym_package_pattern,
    sym_parenthesized_operator,
    sym_polymorphic_variant_pattern,
    sym_quoted_extension,
    sym_tag,
    sym_typed_pattern,
  [223] =
    sym__delimited_type,
    sym_constructed_type,
    sym_extension,
    sym_hash_type,
    sym_local_open_type,
    sym_object_type,
    sym_quoted_extension,
    sym_type_constructor_path,
    sym_type_variable,
  [232] =
    sym__local_structure_item,
    sym_include_module,
    sym_value_definition,
    sym_value_specification,
  [236] =
    sym__labeled_tuple_type,
    sym__proper_tuple_type,
    sym__simple_type,
    sym_aliased_type,
    sym_function_type,
};

static const TSCharacterRange sym_conversion_specification_character_set_6[] = {
  {'!', '!'}, {'%', '%'}, {',', ','}, {'@', '@'}, {'B', 'C'}, {'E', 'H'}, {'L', 'L'}, {'N', 'N'},
  {'S', 'S'}, {'X', 'X'}, {'a', 'i'}, {'l', 'l'}, {'n', 'o'}, {'s', 'u'}, {'x', 'x'},
};

static const TSCharacterRange sym_conversion_specification_character_set_7[] = {
  {'!', '!'}, {'%', '%'}, {',', ','}, {'0', '9'}, {'@', '@'}, {'B', 'C'}, {'E', 'H'}, {'L', 'L'},
  {'N', 'N'}, {'S', 'S'}, {'X', 'X'}, {'a', 'i'}, {'l', 'l'}, {'n', 'o'}, {'s', 'u'}, {'x', 'x'},
};

static const TSCharacterRange sym_prefix_operator_character_set_1[] = {
  {'!', '!'}, {'#', '&'}, {'*', '+'}, {'-', '/'}, {':', ':'}, {'<', '@'}, {'^', '^'}, {'|', '|'},
  {'~', '~'},
};

static const TSCharacterRange aux_sym_pow_operator_token1_character_set_1[] = {
  {'!', '!'}, {'$', '&'}, {'*', '+'}, {'-', '/'}, {':', ':'}, {'<', '@'}, {'^', '^'}, {'|', '|'},
  {'~', '~'},
};

static const TSCharacterRange sym_indexing_operator_character_set_1[] = {
  {'!', '!'}, {'$', '&'}, {'*', '+'}, {'-', '-'}, {'/', '/'}, {':', ':'}, {'=', '@'}, {'^', '^'},
  {'|', '|'},
};

static const TSCharacterRange sym_let_operator_character_set_1[] = {
  {'$', '$'}, {'&', '&'}, {'*', '+'}, {'-', '-'}, {'/', '/'}, {'<', '>'}, {'@', '@'}, {'^', '^'},
  {'|', '|'},
};

static const TSCharacterRange sym__lowercase_identifier_character_set_2[] = {
  {'\'', '\''}, {'0', '9'}, {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xb7, 0xb7},
  {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee},
  {0x300, 0x374}, {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f}, {0x386, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5},
  {0x3f7, 0x481}, {0x483, 0x487}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd}, {0x5bf, 0x5bf},
  {0x5c1, 0x5c2}, {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669}, {0x66e, 0x6d3},
  {0x6d5, 0x6dc}, {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5}, {0x7fa, 0x7fa},
  {0x7fd, 0x7fd}, {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1}, {0x8e3, 0x963},
  {0x966, 0x96f}, {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9},
  {0x9bc, 0x9c4}, {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1}, {0x9fc, 0x9fc},
  {0x9fe, 0x9fe}, {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36},
  {0xa38, 0xa39}, {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c}, {0xa5e, 0xa5e},
  {0xa66, 0xa75}, {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9},
  {0xabc, 0xac5}, {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff}, {0xb01, 0xb03},
  {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44}, {0xb47, 0xb48},
  {0xb4b, 0xb4d}, {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb71}, {0xb82, 0xb83}, {0xb85, 0xb8a},
  {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9},
  {0xbbe, 0xbc2}, {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbef}, {0xc00, 0xc0c}, {0xc0e, 0xc10},
  {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a}, {0xc5d, 0xc5d},
  {0xc60, 0xc63}, {0xc66, 0xc6f}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9},
  {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef}, {0xcf1, 0xcf3},
  {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd57}, {0xd5f, 0xd63}, {0xd66, 0xd6f},
  {0xd7a, 0xd7f}, {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xdca, 0xdca},
  {0xdcf, 0xdd4}, {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e}, {0xe50, 0xe59},
  {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6},
  {0xec8, 0xece}, {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf29}, {0xf35, 0xf35}, {0xf37, 0xf37},
  {0xf39, 0xf39}, {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6}, {0x1000, 0x1049},
  {0x1050, 0x109d}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256},
  {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0},
  {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x1371}, {0x1380, 0x138f},
  {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1715},
  {0x171f, 0x1734}, {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dd},
  {0x17e0, 0x17e9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1920, 0x192b},
  {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b}, {0x1a20, 0x1a5e},
  {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1abd}, {0x1abf, 0x1ace}, {0x1b00, 0x1b4c}, {0x1b50, 0x1b59},
  {0x1b6b, 0x1b73}, {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf},
  {0x1cd0, 0x1cd2}, {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59},
  {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc},
  {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x200c, 0x200d}, {0x203f, 0x2040}, {0x2054, 0x2054},
  {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x20d0, 0x20dc}, {0x20e1, 0x20e1}, {0x20e5, 0x20f0}, {0x2102, 0x2102}, {0x2107, 0x2107},
  {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f},
  {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d},
  {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6},
  {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035}, {0x3038, 0x303c},
  {0x3041, 0x3096}, {0x3099, 0x309a}, {0x309d, 0x309f}, {0x30a1, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff},
  {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa66f}, {0xa674, 0xa67d}, {0xa67f, 0xa6f1},
  {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827}, {0xa82c, 0xa82c},
  {0xa840, 0xa873}, {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953}, {0xa960, 0xa97c},
  {0xa980, 0xa9c0}, {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76}, {0xaa7a, 0xaac2},
  {0xaadb, 0xaadd}, {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e},
  {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb},
  {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e},
  {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdf9},
  {0xfe00, 0xfe0f}, {0xfe20, 0xfe2f}, {0xfe33, 0xfe34}, {0xfe4d, 0xfe4f}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77}, {0xfe79, 0xfe79},
  {0xfe7b, 0xfe7b}, {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff3f, 0xff3f}, {0xff41, 0xff5a}, {0xff65, 0xffbe},
  {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d},
  {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x101fd, 0x101fd}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x102e0, 0x102e0},
  {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d},
  {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592},
  {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755},
  {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838},
  {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939},
  {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a38, 0x10a3a},
  {0x10a3f, 0x10a3f}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b60, 0x10b72},
  {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65}, {0x10d69, 0x10d6d},
  {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f50},
  {0x10f70, 0x10f85}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11066, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2}, {0x110d0, 0x110e8},
  {0x110f0, 0x110f9}, {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4}, {0x111c9, 0x111cc},
  {0x111ce, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d},
  {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328},
  {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350}, {0x11357, 0x11357},
  {0x1135d, 0x11363}, {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113c0},
  {0x113c2, 0x113c2}, {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459}, {0x1145e, 0x11461},
  {0x11480, 0x114c5}, {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640}, {0x11644, 0x11644},
  {0x11650, 0x11659}, {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x11739}, {0x11740, 0x11746},
  {0x11800, 0x1183a}, {0x118a0, 0x118e9}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935}, {0x11937, 0x11938},
  {0x1193b, 0x11943}, {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e}, {0x11a47, 0x11a47},
  {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36}, {0x11c38, 0x11c40},
  {0x11c50, 0x11c59}, {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36}, {0x11d3a, 0x11d3a},
  {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91}, {0x11d93, 0x11d98},
  {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399},
  {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x16139},
  {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4}, {0x16b00, 0x16b36},
  {0x16b40, 0x16b43}, {0x16b50, 0x16b59}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e7f}, {0x16f00, 0x16f4a},
  {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08},
  {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167},
  {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9}, {0x1cf00, 0x1cf2d},
  {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244}, {0x1d400, 0x1d454},
  {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3},
  {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546},
  {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e},
  {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36}, {0x1da3b, 0x1da6c},
  {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e000, 0x1e006}, {0x1e008, 0x1e018},
  {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d}, {0x1e140, 0x1e149},
  {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee},
  {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8d0, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22},
  {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47},
  {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b},
  {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c},
  {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1fbf0, 0x1fbf9}, {0x20000, 0x2a6df},
  {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af},
  {0xe0100, 0xe01ef},
};

static const TSCharacterRange sym__lowercase_identifier_character_set_3[] = {
  {'_', '_'}, {'a', 'z'}, {0xb5, 0xb5}, {0xdf, 0xf6}, {0xf8, 0xff}, {0x101, 0x101}, {0x103, 0x103}, {0x105, 0x105},
  {0x107, 0x107}, {0x109, 0x109}, {0x10b, 0x10b}, {0x10d, 0x10d}, {0x10f, 0x10f}, {0x111, 0x111}, {0x113, 0x113}, {0x115, 0x115},
  {0x117, 0x117}, {0x119, 0x119}, {0x11b, 0x11b}, {0x11d, 0x11d}, {0x11f, 0x11f}, {0x121, 0x121}, {0x123, 0x123}, {0x125, 0x125},
  {0x127, 0x127}, {0x129, 0x129}, {0x12b, 0x12b}, {0x12d, 0x12d}, {0x12f, 0x12f}, {0x131, 0x131}, {0x133, 0x133}, {0x135, 0x135},
  {0x137, 0x138}, {0x13a, 0x13a}, {0x13c, 0x13c}, {0x13e, 0x13e}, {0x140, 0x140}, {0x142, 0x142}, {0x144, 0x144}, {0x146, 0x146},
  {0x148, 0x149}, {0x14b, 0x14b}, {0x14d, 0x14d}, {0x14f, 0x14f}, {0x151, 0x151}, {0x153, 0x153}, {0x155, 0x155}, {0x157, 0x157},
  {0x159, 0x159}, {0x15b, 0x15b}, {0x15d, 0x15d}, {0x15f, 0x15f}, {0x161, 0x161}, {0x163, 0x163}, {0x165, 0x165}, {0x167, 0x167},
  {0x169, 0x169}, {0x16b, 0x16b}, {0x16d, 0x16d}, {0x16f, 0x16f}, {0x171, 0x171}, {0x173, 0x173}, {0x175, 0x175}, {0x177, 0x177},
  {0x17a, 0x17a}, {0x17c, 0x17c}, {0x17e, 0x180}, {0x183, 0x183}, {0x185, 0x185}, {0x188, 0x188}, {0x18c, 0x18d}, {0x192, 0x192},
  {0x195, 0x195}, {0x199, 0x19b}, {0x19e, 0x19e}, {0x1a1, 0x1a1}, {0x1a3, 0x1a3}, {0x1a5, 0x1a5}, {0x1a8, 0x1a8}, {0x1aa, 0x1ab},
  {0x1ad, 0x1ad}, {0x1b0, 0x1b0}, {0x1b4, 0x1b4}, {0x1b6, 0x1b6}, {0x1b9, 0x1ba}, {0x1bd, 0x1bf}, {0x1c6, 0x1c6}, {0x1c9, 0x1c9},
  {0x1cc, 0x1cc}, {0x1ce, 0x1ce}, {0x1d0, 0x1d0}, {0x1d2, 0x1d2}, {0x1d4, 0x1d4}, {0x1d6, 0x1d6}, {0x1d8, 0x1d8}, {0x1da, 0x1da},
  {0x1dc, 0x1dd}, {0x1df, 0x1df}, {0x1e1, 0x1e1}, {0x1e3, 0x1e3}, {0x1e5, 0x1e5}, {0x1e7, 0x1e7}, {0x1e9, 0x1e9}, {0x1eb, 0x1eb},
  {0x1ed, 0x1ed}, {0x1ef, 0x1f0}, {0x1f3, 0x1f3}, {0x1f5, 0x1f5}, {0x1f9, 0x1f9}, {0x1fb, 0x1fb}, {0x1fd, 0x1fd}, {0x1ff, 0x1ff},
  {0x201, 0x201}, {0x203, 0x203}, {0x205, 0x205}, {0x207, 0x207}, {0x209, 0x209}, {0x20b, 0x20b}, {0x20d, 0x20d}, {0x20f, 0x20f},
  {0x211, 0x211}, {0x213, 0x213}, {0x215, 0x215}, {0x217, 0x217}, {0x219, 0x219}, {0x21b, 0x21b}, {0x21d, 0x21d}, {0x21f, 0x21f},
  {0x221, 0x221}, {0x223, 0x223}, {0x225, 0x225}, {0x227, 0x227}, {0x229, 0x229}, {0x22b, 0x22b}, {0x22d, 0x22d}, {0x22f, 0x22f},
  {0x231, 0x231}, {0x233, 0x239}, {0x23c, 0x23c}, {0x23f, 0x240}, {0x242, 0x242}, {0x247, 0x247}, {0x249, 0x249}, {0x24b, 0x24b},
  {0x24d, 0x24d}, {0x24f, 0x293}, {0x295, 0x2af}, {0x371, 0x371}, {0x373, 0x373}, {0x377, 0x377}, {0x37b, 0x37d}, {0x390, 0x390},
  {0x3ac, 0x3ce}, {0x3d0, 0x3d1}, {0x3d5, 0x3d7}, {0x3d9, 0x3d9}, {0x3db, 0x3db}, {0x3dd, 0x3dd}, {0x3df, 0x3df}, {0x3e1, 0x3e1},
  {0x3e3, 0x3e3}, {0x3e5, 0x3e5}, {0x3e7, 0x3e7}, {0x3e9, 0x3e9}, {0x3eb, 0x3eb}, {0x3ed, 0x3ed}, {0x3ef, 0x3f3}, {0x3f5, 0x3f5},
  {0x3f8, 0x3f8}, {0x3fb, 0x3fc}, {0x430, 0x45f}, {0x461, 0x461}, {0x463, 0x463}, {0x465, 0x465}, {0x467, 0x467}, {0x469, 0x469},
  {0x46b, 0x46b}, {0x46d, 0x46d}, {0x46f, 0x46f}, {0x471, 0x471}, {0x473, 0x473}, {0x475, 0x475}, {0x477, 0x477}, {0x479, 0x479},
  {0x47b, 0x47b}, {0x47d, 0x47d}, {0x47f, 0x47f}, {0x481, 0x481}, {0x48b, 0x48b}, {0x48d, 0x48d}, {0x48f, 0x48f}, {0x491, 0x491},
  {0x493, 0x493}, {0x495, 0x495}, {0x497, 0x497}, {0x499, 0x499}, {0x49b, 0x49b}, {0x49d, 0x49d}, {0x49f, 0x49f}, {0x4a1, 0x4a1},
  {0x4a3, 0x4a3}, {0x4a5, 0x4a5}, {0x4a7, 0x4a7}, {0x4a9, 0x4a9}, {0x4ab, 0x4ab}, {0x4ad, 0x4ad}, {0x4af, 0x4af}, {0x4b1, 0x4b1},
  {0x4b3, 0x4b3}, {0x4b5, 0x4b5}, {0x4b7, 0x4b7}, {0x4b9, 0x4b9}, {0x4bb, 0x4bb}, {0x4bd, 0x4bd}, {0x4bf, 0x4bf}, {0x4c2, 0x4c2},
  {0x4c4, 0x4c4}, {0x4c6, 0x4c6}, {0x4c8, 0x4c8}, {0x4ca, 0x4ca}, {0x4cc, 0x4cc}, {0x4ce, 0x4cf}, {0x4d1, 0x4d1}, {0x4d3, 0x4d3},
  {0x4d5, 0x4d5}, {0x4d7, 0x4d7}, {0x4d9, 0x4d9}, {0x4db, 0x4db}, {0x4dd, 0x4dd}, {0x4df, 0x4df}, {0x4e1, 0x4e1}, {0x4e3, 0x4e3},
  {0x4e5, 0x4e5}, {0x4e7, 0x4e7}, {0x4e9, 0x4e9}, {0x4eb, 0x4eb}, {0x4ed, 0x4ed}, {0x4ef, 0x4ef}, {0x4f1, 0x4f1}, {0x4f3, 0x4f3},
  {0x4f5, 0x4f5}, {0x4f7, 0x4f7}, {0x4f9, 0x4f9}, {0x4fb, 0x4fb}, {0x4fd, 0x4fd}, {0x4ff, 0x4ff}, {0x501, 0x501}, {0x503, 0x503},
  {0x505, 0x505}, {0x507, 0x507}, {0x509, 0x509}, {0x50b, 0x50b}, {0x50d, 0x50d}, {0x50f, 0x50f}, {0x511, 0x511}, {0x513, 0x513},
  {0x515, 0x515}, {0x517, 0x517}, {0x519, 0x519}, {0x51b, 0x51b}, {0x51d, 0x51d}, {0x51f, 0x51f}, {0x521, 0x521}, {0x523, 0x523},
  {0x525, 0x525}, {0x527, 0x527}, {0x529, 0x529}, {0x52b, 0x52b}, {0x52d, 0x52d}, {0x52f, 0x52f}, {0x560, 0x588}, {0x10d0, 0x10fa},
  {0x10fd, 0x10ff}, {0x13f8, 0x13fd}, {0x1c80, 0x1c88}, {0x1c8a, 0x1c8a}, {0x1d00, 0x1d2b}, {0x1d6b, 0x1d77}, {0x1d79, 0x1d9a}, {0x1e01, 0x1e01},
  {0x1e03, 0x1e03}, {0x1e05, 0x1e05}, {0x1e07, 0x1e07}, {0x1e09, 0x1e09}, {0x1e0b, 0x1e0b}, {0x1e0d, 0x1e0d}, {0x1e0f, 0x1e0f}, {0x1e11, 0x1e11},
  {0x1e13, 0x1e13}, {0x1e15, 0x1e15}, {0x1e17, 0x1e17}, {0x1e19, 0x1e19}, {0x1e1b, 0x1e1b}, {0x1e1d, 0x1e1d}, {0x1e1f, 0x1e1f}, {0x1e21, 0x1e21},
  {0x1e23, 0x1e23}, {0x1e25, 0x1e25}, {0x1e27, 0x1e27}, {0x1e29, 0x1e29}, {0x1e2b, 0x1e2b}, {0x1e2d, 0x1e2d}, {0x1e2f, 0x1e2f}, {0x1e31, 0x1e31},
  {0x1e33, 0x1e33}, {0x1e35, 0x1e35}, {0x1e37, 0x1e37}, {0x1e39, 0x1e39}, {0x1e3b, 0x1e3b}, {0x1e3d, 0x1e3d}, {0x1e3f, 0x1e3f}, {0x1e41, 0x1e41},
  {0x1e43, 0x1e43}, {0x1e45, 0x1e45}, {0x1e47, 0x1e47}, {0x1e49, 0x1e49}, {0x1e4b, 0x1e4b}, {0x1e4d, 0x1e4d}, {0x1e4f, 0x1e4f}, {0x1e51, 0x1e51},
  {0x1e53, 0x1e53}, {0x1e55, 0x1e55}, {0x1e57, 0x1e57}, {0x1e59, 0x1e59}, {0x1e5b, 0x1e5b}, {0x1e5d, 0x1e5d}, {0x1e5f, 0x1e5f}, {0x1e61, 0x1e61},
  {0x1e63, 0x1e63}, {0x1e65, 0x1e65}, {0x1e67, 0x1e67}, {0x1e69, 0x1e69}, {0x1e6b, 0x1e6b}, {0x1e6d, 0x1e6d}, {0x1e6f, 0x1e6f}, {0x1e71, 0x1e71},
  {0x1e73, 0x1e73}, {0x1e75, 0x1e75}, {0x1e77, 0x1e77}, {0x1e79, 0x1e79}, {0x1e7b, 0x1e7b}, {0x1e7d, 0x1e7d}, {0x1e7f, 0x1e7f}, {0x1e81, 0x1e81},
  {0x1e83, 0x1e83}, {0x1e85, 0x1e85}, {0x1e87, 0x1e87}, {0x1e89, 0x1e89}, {0x1e8b, 0x1e8b}, {0x1e8d, 0x1e8d}, {0x1e8f, 0x1e8f}, {0x1e91, 0x1e91},
  {0x1e93, 0x1e93}, {0x1e95, 0x1e9d}, {0x1e9f, 0x1e9f}, {0x1ea1, 0x1ea1}, {0x1ea3, 0x1ea3}, {0x1ea5, 0x1ea5}, {0x1ea7, 0x1ea7}, {0x1ea9, 0x1ea9},
  {0x1eab, 0x1eab}, {0x1ead, 0x1ead}, {0x1eaf, 0x1eaf}, {0x1eb1, 0x1eb1}, {0x1eb3, 0x1eb3}, {0x1eb5, 0x1eb5}, {0x1eb7, 0x1eb7}, {0x1eb9, 0x1eb9},
  {0x1ebb, 0x1ebb}, {0x1ebd, 0x1ebd}, {0x1ebf, 0x1ebf}, {0x1ec1, 0x1ec1}, {0x1ec3, 0x1ec3}, {0x1ec5, 0x1ec5}, {0x1ec7, 0x1ec7}, {0x1ec9, 0x1ec9},
  {0x1ecb, 0x1ecb}, {0x1ecd, 0x1ecd}, {0x1ecf, 0x1ecf}, {0x1ed1, 0x1ed1}, {0x1ed3, 0x1ed3}, {0x1ed5, 0x1ed5}, {0x1ed7, 0x1ed7}, {0x1ed9, 0x1ed9},
  {0x1edb, 0x1edb}, {0x1edd, 0x1edd}, {0x1edf, 0x1edf}, {0x1ee1, 0x1ee1}, {0x1ee3, 0x1ee3}, {0x1ee5, 0x1ee5}, {0x1ee7, 0x1ee7}, {0x1ee9, 0x1ee9},
  {0x1eeb, 0x1eeb}, {0x1eed, 0x1eed}, {0x1eef, 0x1eef}, {0x1ef1, 0x1ef1}, {0x1ef3, 0x1ef3}, {0x1ef5, 0x1ef5}, {0x1ef7, 0x1ef7}, {0x1ef9, 0x1ef9},
  {0x1efb, 0x1efb}, {0x1efd, 0x1efd}, {0x1eff, 0x1f07}, {0x1f10, 0x1f15}, {0x1f20, 0x1f27}, {0x1f30, 0x1f37}, {0x1f40, 0x1f45}, {0x1f50, 0x1f57},
  {0x1f60, 0x1f67}, {0x1f70, 0x1f7d}, {0x1f80, 0x1f87}, {0x1f90, 0x1f97}, {0x1fa0, 0x1fa7}, {0x1fb0, 0x1fb4}, {0x1fb6, 0x1fb7}, {0x1fbe, 0x1fbe},
  {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fc7}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fd7}, {0x1fe0, 0x1fe7}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ff7}, {0x210a, 0x210a},
  {0x210e, 0x210f}, {0x2113, 0x2113}, {0x212f, 0x212f}, {0x2134, 0x2134}, {0x2139, 0x2139}, {0x213c, 0x213d}, {0x2146, 0x2149}, {0x214e, 0x214e},
  {0x2184, 0x2184}, {0x2c30, 0x2c5f}, {0x2c61, 0x2c61}, {0x2c65, 0x2c66}, {0x2c68, 0x2c68}, {0x2c6a, 0x2c6a}, {0x2c6c, 0x2c6c}, {0x2c71, 0x2c71},
  {0x2c73, 0x2c74}, {0x2c76, 0x2c7b}, {0x2c81, 0x2c81}, {0x2c83, 0x2c83}, {0x2c85, 0x2c85}, {0x2c87, 0x2c87}, {0x2c89, 0x2c89}, {0x2c8b, 0x2c8b},
  {0x2c8d, 0x2c8d}, {0x2c8f, 0x2c8f}, {0x2c91, 0x2c91}, {0x2c93, 0x2c93}, {0x2c95, 0x2c95}, {0x2c97, 0x2c97}, {0x2c99, 0x2c99}, {0x2c9b, 0x2c9b},
  {0x2c9d, 0x2c9d}, {0x2c9f, 0x2c9f}, {0x2ca1, 0x2ca1}, {0x2ca3, 0x2ca3}, {0x2ca5, 0x2ca5}, {0x2ca7, 0x2ca7}, {0x2ca9, 0x2ca9}, {0x2cab, 0x2cab},
  {0x2cad, 0x2cad}, {0x2caf, 0x2caf}, {0x2cb1, 0x2cb1}, {0x2cb3, 0x2cb3}, {0x2cb5, 0x2cb5}, {0x2cb7, 0x2cb7}, {0x2cb9, 0x2cb9}, {0x2cbb, 0x2cbb},
  {0x2cbd, 0x2cbd}, {0x2cbf, 0x2cbf}, {0x2cc1, 0x2cc1}, {0x2cc3, 0x2cc3}, {0x2cc5, 0x2cc5}, {0x2cc7, 0x2cc7}, {0x2cc9, 0x2cc9}, {0x2ccb, 0x2ccb},
  {0x2ccd, 0x2ccd}, {0x2ccf, 0x2ccf}, {0x2cd1, 0x2cd1}, {0x2cd3, 0x2cd3}, {0x2cd5, 0x2cd5}, {0x2cd7, 0x2cd7}, {0x2cd9, 0x2cd9}, {0x2cdb, 0x2cdb},
  {0x2cdd, 0x2cdd}, {0x2cdf, 0x2cdf}, {0x2ce1, 0x2ce1}, {0x2ce3, 0x2ce4}, {0x2cec, 0x2cec}, {0x2cee, 0x2cee}, {0x2cf3, 0x2cf3}, {0x2d00, 0x2d25},
  {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d}, {0xa641, 0xa641}, {0xa643, 0xa643}, {0xa645, 0xa645}, {0xa647, 0xa647}, {0xa649, 0xa649}, {0xa64b, 0xa64b},
  {0xa64d, 0xa64d}, {0xa64f, 0xa64f}, {0xa651, 0xa651}, {0xa653, 0xa653}, {0xa655, 0xa655}, {0xa657, 0xa657}, {0xa659, 0xa659}, {0xa65b, 0xa65b},
  {0xa65d, 0xa65d}, {0xa65f, 0xa65f}, {0xa661, 0xa661}, {0xa663, 0xa663}, {0xa665, 0xa665}, {0xa667, 0xa667}, {0xa669, 0xa669}, {0xa66b, 0xa66b},
  {0xa66d, 0xa66d}, {0xa681, 0xa681}, {0xa683, 0xa683}, {0xa685, 0xa685}, {0xa687, 0xa687}, {0xa689, 0xa689}, {0xa68b, 0xa68b}, {0xa68d, 0xa68d},
  {0xa68f, 0xa68f}, {0xa691, 0xa691}, {0xa693, 0xa693}, {0xa695, 0xa695}, {0xa697, 0xa697}, {0xa699, 0xa699}, {0xa69b, 0xa69b}, {0xa723, 0xa723},
  {0xa725, 0xa725}, {0xa727, 0xa727}, {0xa729, 0xa729}, {0xa72b, 0xa72b}, {0xa72d, 0xa72d}, {0xa72f, 0xa731}, {0xa733, 0xa733}, {0xa735, 0xa735},
  {0xa737, 0xa737}, {0xa739, 0xa739}, {0xa73b, 0xa73b}, {0xa73d, 0xa73d}, {0xa73f, 0xa73f}, {0xa741, 0xa741}, {0xa743, 0xa743}, {0xa745, 0xa745},
  {0xa747, 0xa747}, {0xa749, 0xa749}, {0xa74b, 0xa74b}, {0xa74d, 0xa74d}, {0xa74f, 0xa74f}, {0xa751, 0xa751}, {0xa753, 0xa753}, {0xa755, 0xa755},
  {0xa757, 0xa757}, {0xa759, 0xa759}, {0xa75b, 0xa75b}, {0xa75d, 0xa75d}, {0xa75f, 0xa75f}, {0xa761, 0xa761}, {0xa763, 0xa763}, {0xa765, 0xa765},
  {0xa767, 0xa767}, {0xa769, 0xa769}, {0xa76b, 0xa76b}, {0xa76d, 0xa76d}, {0xa76f, 0xa76f}, {0xa771, 0xa778}, {0xa77a, 0xa77a}, {0xa77c, 0xa77c},
  {0xa77f, 0xa77f}, {0xa781, 0xa781}, {0xa783, 0xa783}, {0xa785, 0xa785}, {0xa787, 0xa787}, {0xa78c, 0xa78c}, {0xa78e, 0xa78e}, {0xa791, 0xa791},
  {0xa793, 0xa795}, {0xa797, 0xa797}, {0xa799, 0xa799}, {0xa79b, 0xa79b}, {0xa79d, 0xa79d}, {0xa79f, 0xa79f}, {0xa7a1, 0xa7a1}, {0xa7a3, 0xa7a3},
  {0xa7a5, 0xa7a5}, {0xa7a7, 0xa7a7}, {0xa7a9, 0xa7a9}, {0xa7af, 0xa7af}, {0xa7b5, 0xa7b5}, {0xa7b7, 0xa7b7}, {0xa7b9, 0xa7b9}, {0xa7bb, 0xa7bb},
  {0xa7bd, 0xa7bd}, {0xa7bf, 0xa7bf}, {0xa7c1, 0xa7c1}, {0xa7c3, 0xa7c3}, {0xa7c8, 0xa7c8}, {0xa7ca, 0xa7ca}, {0xa7cd, 0xa7cd}, {0xa7d1, 0xa7d1},
  {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7d5}, {0xa7d7, 0xa7d7}, {0xa7d9, 0xa7d9}, {0xa7db, 0xa7db}, {0xa7f6, 0xa7f6}, {0xa7fa, 0xa7fa}, {0xab30, 0xab5a},
  {0xab60, 0xab68}, {0xab70, 0xabbf}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xff41, 0xff5a}, {0x10428, 0x1044f}, {0x104d8, 0x104fb}, {0x10597, 0x105a1},
  {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x10cc0, 0x10cf2}, {0x10d70, 0x10d85}, {0x118c0, 0x118df}, {0x16e60, 0x16e7f}, {0x1d41a, 0x1d433},
  {0x1d44e, 0x1d454}, {0x1d456, 0x1d467}, {0x1d482, 0x1d49b}, {0x1d4b6, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d4cf}, {0x1d4ea, 0x1d503},
  {0x1d51e, 0x1d537}, {0x1d552, 0x1d56b}, {0x1d586, 0x1d59f}, {0x1d5ba, 0x1d5d3}, {0x1d5ee, 0x1d607}, {0x1d622, 0x1d63b}, {0x1d656, 0x1d66f}, {0x1d68a, 0x1d6a5},
  {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6e1}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d71b}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d755}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d78f},
  {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7c9}, {0x1d7cb, 0x1d7cb}, {0x1df00, 0x1df09}, {0x1df0b, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e922, 0x1e943},
};

static const TSCharacterRange sym__uppercase_identifier_character_set_1[] = {
  {'A', 'Z'}, {0xc0, 0xd6}, {0xd8, 0xde}, {0x100, 0x100}, {0x102, 0x102}, {0x104, 0x104}, {0x106, 0x106}, {0x108, 0x108},
  {0x10a, 0x10a}, {0x10c, 0x10c}, {0x10e, 0x10e}, {0x110, 0x110}, {0x112, 0x112}, {0x114, 0x114}, {0x116, 0x116}, {0x118, 0x118},
  {0x11a, 0x11a}, {0x11c, 0x11c}, {0x11e, 0x11e}, {0x120, 0x120}, {0x122, 0x122}, {0x124, 0x124}, {0x126, 0x126}, {0x128, 0x128},
  {0x12a, 0x12a}, {0x12c, 0x12c}, {0x12e, 0x12e}, {0x130, 0x130}, {0x132, 0x132}, {0x134, 0x134}, {0x136, 0x136}, {0x139, 0x139},
  {0x13b, 0x13b}, {0x13d, 0x13d}, {0x13f, 0x13f}, {0x141, 0x141}, {0x143, 0x143}, {0x145, 0x145}, {0x147, 0x147}, {0x14a, 0x14a},
  {0x14c, 0x14c}, {0x14e, 0x14e}, {0x150, 0x150}, {0x152, 0x152}, {0x154, 0x154}, {0x156, 0x156}, {0x158, 0x158}, {0x15a, 0x15a},
  {0x15c, 0x15c}, {0x15e, 0x15e}, {0x160, 0x160}, {0x162, 0x162}, {0x164, 0x164}, {0x166, 0x166}, {0x168, 0x168}, {0x16a, 0x16a},
  {0x16c, 0x16c}, {0x16e, 0x16e}, {0x170, 0x170}, {0x172, 0x172}, {0x174, 0x174}, {0x176, 0x176}, {0x178, 0x179}, {0x17b, 0x17b},
  {0x17d, 0x17d}, {0x181, 0x182}, {0x184, 0x184}, {0x186, 0x187}, {0x189, 0x18b}, {0x18e, 0x191}, {0x193, 0x194}, {0x196, 0x198},
  {0x19c, 0x19d}, {0x19f, 0x1a0}, {0x1a2, 0x1a2}, {0x1a4, 0x1a4}, {0x1a6, 0x1a7}, {0x1a9, 0x1a9}, {0x1ac, 0x1ac}, {0x1ae, 0x1af},
  {0x1b1, 0x1b3}, {0x1b5, 0x1b5}, {0x1b7, 0x1b8}, {0x1bc, 0x1bc}, {0x1c4, 0x1c4}, {0x1c7, 0x1c7}, {0x1ca, 0x1ca}, {0x1cd, 0x1cd},
  {0x1cf, 0x1cf}, {0x1d1, 0x1d1}, {0x1d3, 0x1d3}, {0x1d5, 0x1d5}, {0x1d7, 0x1d7}, {0x1d9, 0x1d9}, {0x1db, 0x1db}, {0x1de, 0x1de},
  {0x1e0, 0x1e0}, {0x1e2, 0x1e2}, {0x1e4, 0x1e4}, {0x1e6, 0x1e6}, {0x1e8, 0x1e8}, {0x1ea, 0x1ea}, {0x1ec, 0x1ec}, {0x1ee, 0x1ee},
  {0x1f1, 0x1f1}, {0x1f4, 0x1f4}, {0x1f6, 0x1f8}, {0x1fa, 0x1fa}, {0x1fc, 0x1fc}, {0x1fe, 0x1fe}, {0x200, 0x200}, {0x202, 0x202},
  {0x204, 0x204}, {0x206, 0x206}, {0x208, 0x208}, {0x20a, 0x20a}, {0x20c, 0x20c}, {0x20e, 0x20e}, {0x210, 0x210}, {0x212, 0x212},
  {0x214, 0x214}, {0x216, 0x216}, {0x218, 0x218}, {0x21a, 0x21a}, {0x21c, 0x21c}, {0x21e, 0x21e}, {0x220, 0x220}, {0x222, 0x222},
  {0x224, 0x224}, {0x226, 0x226}, {0x228, 0x228}, {0x22a, 0x22a}, {0x22c, 0x22c}, {0x22e, 0x22e}, {0x230, 0x230}, {0x232, 0x232},
  {0x23a, 0x23b}, {0x23d, 0x23e}, {0x241, 0x241}, {0x243, 0x246}, {0x248, 0x248}, {0x24a, 0x24a}, {0x24c, 0x24c}, {0x24e, 0x24e},
  {0x370, 0x370}, {0x372, 0x372}, {0x376, 0x376}, {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x38f},
  {0x391, 0x3a1}, {0x3a3, 0x3ab}, {0x3cf, 0x3cf}, {0x3d2, 0x3d4}, {0x3d8, 0x3d8}, {0x3da, 0x3da}, {0x3dc, 0x3dc}, {0x3de, 0x3de},
  {0x3e0, 0x3e0}, {0x3e2, 0x3e2}, {0x3e4, 0x3e4}, {0x3e6, 0x3e6}, {0x3e8, 0x3e8}, {0x3ea, 0x3ea}, {0x3ec, 0x3ec}, {0x3ee, 0x3ee},
  {0x3f4, 0x3f4}, {0x3f7, 0x3f7}, {0x3f9, 0x3fa}, {0x3fd, 0x42f}, {0x460, 0x460}, {0x462, 0x462}, {0x464, 0x464}, {0x466, 0x466},
  {0x468, 0x468}, {0x46a, 0x46a}, {0x46c, 0x46c}, {0x46e, 0x46e}, {0x470, 0x470}, {0x472, 0x472}, {0x474, 0x474}, {0x476, 0x476},
  {0x478, 0x478}, {0x47a, 0x47a}, {0x47c, 0x47c}, {0x47e, 0x47e}, {0x480, 0x480}, {0x48a, 0x48a}, {0x48c, 0x48c}, {0x48e, 0x48e},
  {0x490, 0x490}, {0x492, 0x492}, {0x494, 0x494}, {0x496, 0x496}, {0x498, 0x498}, {0x49a, 0x49a}, {0x49c, 0x49c}, {0x49e, 0x49e},
  {0x4a0, 0x4a0}, {0x4a2, 0x4a2}, {0x4a4, 0x4a4}, {0x4a6, 0x4a6}, {0x4a8, 0x4a8}, {0x4aa, 0x4aa}, {0x4ac, 0x4ac}, {0x4ae, 0x4ae},
  {0x4b0, 0x4b0}, {0x4b2, 0x4b2}, {0x4b4, 0x4b4}, {0x4b6, 0x4b6}, {0x4b8, 0x4b8}, {0x4ba, 0x4ba}, {0x4bc, 0x4bc}, {0x4be, 0x4be},
  {0x4c0, 0x4c1}, {0x4c3, 0x4c3}, {0x4c5, 0x4c5}, {0x4c7, 0x4c7}, {0x4c9, 0x4c9}, {0x4cb, 0x4cb}, {0x4cd, 0x4cd}, {0x4d0, 0x4d0},
  {0x4d2, 0x4d2}, {0x4d4, 0x4d4}, {0x4d6, 0x4d6}, {0x4d8, 0x4d8}, {0x4da, 0x4da}, {0x4dc, 0x4dc}, {0x4de, 0x4de}, {0x4e0, 0x4e0},
  {0x4e2, 0x4e2}, {0x4e4, 0x4e4}, {0x4e6, 0x4e6}, {0x4e8, 0x4e8}, {0x4ea, 0x4ea}, {0x4ec, 0x4ec}, {0x4ee, 0x4ee}, {0x4f0, 0x4f0},
  {0x4f2, 0x4f2}, {0x4f4, 0x4f4}, {0x4f6, 0x4f6}, {0x4f8, 0x4f8}, {0x4fa, 0x4fa}, {0x4fc, 0x4fc}, {0x4fe, 0x4fe}, {0x500, 0x500},
  {0x502, 0x502}, {0x504, 0x504}, {0x506, 0x506}, {0x508, 0x508}, {0x50a, 0x50a}, {0x50c, 0x50c}, {0x50e, 0x50e}, {0x510, 0x510},
  {0x512, 0x512}, {0x514, 0x514}, {0x516, 0x516}, {0x518, 0x518}, {0x51a, 0x51a}, {0x51c, 0x51c}, {0x51e, 0x51e}, {0x520, 0x520},
  {0x522, 0x522}, {0x524, 0x524}, {0x526, 0x526}, {0x528, 0x528}, {0x52a, 0x52a}, {0x52c, 0x52c}, {0x52e, 0x52e}, {0x531, 0x556},
  {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x13a0, 0x13f5}, {0x1c89, 0x1c89}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1e00, 0x1e00},
  {0x1e02, 0x1e02}, {0x1e04, 0x1e04}, {0x1e06, 0x1e06}, {0x1e08, 0x1e08}, {0x1e0a, 0x1e0a}, {0x1e0c, 0x1e0c}, {0x1e0e, 0x1e0e}, {0x1e10, 0x1e10},
  {0x1e12, 0x1e12}, {0x1e14, 0x1e14}, {0x1e16, 0x1e16}, {0x1e18, 0x1e18}, {0x1e1a, 0x1e1a}, {0x1e1c, 0x1e1c}, {0x1e1e, 0x1e1e}, {0x1e20, 0x1e20},
  {0x1e22, 0x1e22}, {0x1e24, 0x1e24}, {0x1e26, 0x1e26}, {0x1e28, 0x1e28}, {0x1e2a, 0x1e2a}, {0x1e2c, 0x1e2c}, {0x1e2e, 0x1e2e}, {0x1e30, 0x1e30},
  {0x1e32, 0x1e32}, {0x1e34, 0x1e34}, {0x1e36, 0x1e36}, {0x1e38, 0x1e38}, {0x1e3a, 0x1e3a}, {0x1e3c, 0x1e3c}, {0x1e3e, 0x1e3e}, {0x1e40, 0x1e40},
  {0x1e42, 0x1e42}, {0x1e44, 0x1e44}, {0x1e46, 0x1e46}, {0x1e48, 0x1e48}, {0x1e4a, 0x1e4a}, {0x1e4c, 0x1e4c}, {0x1e4e, 0x1e4e}, {0x1e50, 0x1e50},
  {0x1e52, 0x1e52}, {0x1e54, 0x1e54}, {0x1e56, 0x1e56}, {0x1e58, 0x1e58}, {0x1e5a, 0x1e5a}, {0x1e5c, 0x1e5c}, {0x1e5e, 0x1e5e}, {0x1e60, 0x1e60},
  {0x1e62, 0x1e62}, {0x1e64, 0x1e64}, {0x1e66, 0x1e66}, {0x1e68, 0x1e68}, {0x1e6a, 0x1e6a}, {0x1e6c, 0x1e6c}, {0x1e6e, 0x1e6e}, {0x1e70, 0x1e70},
  {0x1e72, 0x1e72}, {0x1e74, 0x1e74}, {0x1e76, 0x1e76}, {0x1e78, 0x1e78}, {0x1e7a, 0x1e7a}, {0x1e7c, 0x1e7c}, {0x1e7e, 0x1e7e}, {0x1e80, 0x1e80},
  {0x1e82, 0x1e82}, {0x1e84, 0x1e84}, {0x1e86, 0x1e86}, {0x1e88, 0x1e88}, {0x1e8a, 0x1e8a}, {0x1e8c, 0x1e8c}, {0x1e8e, 0x1e8e}, {0x1e90, 0x1e90},
  {0x1e92, 0x1e92}, {0x1e94, 0x1e94}, {0x1e9e, 0x1e9e}, {0x1ea0, 0x1ea0}, {0x1ea2, 0x1ea2}, {0x1ea4, 0x1ea4}, {0x1ea6, 0x1ea6}, {0x1ea8, 0x1ea8},
  {0x1eaa, 0x1eaa}, {0x1eac, 0x1eac}, {0x1eae, 0x1eae}, {0x1eb0, 0x1eb0}, {0x1eb2, 0x1eb2}, {0x1eb4, 0x1eb4}, {0x1eb6, 0x1eb6}, {0x1eb8, 0x1eb8},
  {0x1eba, 0x1eba}, {0x1ebc, 0x1ebc}, {0x1ebe, 0x1ebe}, {0x1ec0, 0x1ec0}, {0x1ec2, 0x1ec2}, {0x1ec4, 0x1ec4}, {0x1ec6, 0x1ec6}, {0x1ec8, 0x1ec8},
  {0x1eca, 0x1eca}, {0x1ecc, 0x1ecc}, {0x1ece, 0x1ece}, {0x1ed0, 0x1ed0}, {0x1ed2, 0x1ed2}, {0x1ed4, 0x1ed4}, {0x1ed6, 0x1ed6}, {0x1ed8, 0x1ed8},
  {0x1eda, 0x1eda}, {0x1edc, 0x1edc}, {0x1ede, 0x1ede}, {0x1ee0, 0x1ee0}, {0x1ee2, 0x1ee2}, {0x1ee4, 0x1ee4}, {0x1ee6, 0x1ee6}, {0x1ee8, 0x1ee8},
  {0x1eea, 0x1eea}, {0x1eec, 0x1eec}, {0x1eee, 0x1eee}, {0x1ef0, 0x1ef0}, {0x1ef2, 0x1ef2}, {0x1ef4, 0x1ef4}, {0x1ef6, 0x1ef6}, {0x1ef8, 0x1ef8},
  {0x1efa, 0x1efa}, {0x1efc, 0x1efc}, {0x1efe, 0x1efe}, {0x1f08, 0x1f0f}, {0x1f18, 0x1f1d}, {0x1f28, 0x1f2f}, {0x1f38, 0x1f3f}, {0x1f48, 0x1f4d},
  {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f5f}, {0x1f68, 0x1f6f}, {0x1fb8, 0x1fbb}, {0x1fc8, 0x1fcb}, {0x1fd8, 0x1fdb},
  {0x1fe8, 0x1fec}, {0x1ff8, 0x1ffb}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210b, 0x210d}, {0x2110, 0x2112}, {0x2115, 0x2115}, {0x2119, 0x211d},
  {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x212d}, {0x2130, 0x2133}, {0x213e, 0x213f}, {0x2145, 0x2145}, {0x2183, 0x2183},
  {0x2c00, 0x2c2f}, {0x2c60, 0x2c60}, {0x2c62, 0x2c64}, {0x2c67, 0x2c67}, {0x2c69, 0x2c69}, {0x2c6b, 0x2c6b}, {0x2c6d, 0x2c70}, {0x2c72, 0x2c72},
  {0x2c75, 0x2c75}, {0x2c7e, 0x2c80}, {0x2c82, 0x2c82}, {0x2c84, 0x2c84}, {0x2c86, 0x2c86}, {0x2c88, 0x2c88}, {0x2c8a, 0x2c8a}, {0x2c8c, 0x2c8c},
  {0x2c8e, 0x2c8e}, {0x2c90, 0x2c90}, {0x2c92, 0x2c92}, {0x2c94, 0x2c94}, {0x2c96, 0x2c96}, {0x2c98, 0x2c98}, {0x2c9a, 0x2c9a}, {0x2c9c, 0x2c9c},
  {0x2c9e, 0x2c9e}, {0x2ca0, 0x2ca0}, {0x2ca2, 0x2ca2}, {0x2ca4, 0x2ca4}, {0x2ca6, 0x2ca6}, {0x2ca8, 0x2ca8}, {0x2caa, 0x2caa}, {0x2cac, 0x2cac},
  {0x2cae, 0x2cae}, {0x2cb0, 0x2cb0}, {0x2cb2, 0x2cb2}, {0x2cb4, 0x2cb4}, {0x2cb6, 0x2cb6}, {0x2cb8, 0x2cb8}, {0x2cba, 0x2cba}, {0x2cbc, 0x2cbc},
  {0x2cbe, 0x2cbe}, {0x2cc0, 0x2cc0}, {0x2cc2, 0x2cc2}, {0x2cc4, 0x2cc4}, {0x2cc6, 0x2cc6}, {0x2cc8, 0x2cc8}, {0x2cca, 0x2cca}, {0x2ccc, 0x2ccc},
  {0x2cce, 0x2cce}, {0x2cd0, 0x2cd0}, {0x2cd2, 0x2cd2}, {0x2cd4, 0x2cd4}, {0x2cd6, 0x2cd6}, {0x2cd8, 0x2cd8}, {0x2cda, 0x2cda}, {0x2cdc, 0x2cdc},
  {0x2cde, 0x2cde}, {0x2ce0, 0x2ce0}, {0x2ce2, 0x2ce2}, {0x2ceb, 0x2ceb}, {0x2ced, 0x2ced}, {0x2cf2, 0x2cf2}, {0xa640, 0xa640}, {0xa642, 0xa642},
  {0xa644, 0xa644}, {0xa646, 0xa646}, {0xa648, 0xa648}, {0xa64a, 0xa64a}, {0xa64c, 0xa64c}, {0xa64e, 0xa64e}, {0xa650, 0xa650}, {0xa652, 0xa652},
  {0xa654, 0xa654}, {0xa656, 0xa656}, {0xa658, 0xa658}, {0xa65a, 0xa65a}, {0xa65c, 0xa65c}, {0xa65e, 0xa65e}, {0xa660, 0xa660}, {0xa662, 0xa662},
  {0xa664, 0xa664}, {0xa666, 0xa666}, {0xa668, 0xa668}, {0xa66a, 0xa66a}, {0xa66c, 0xa66c}, {0xa680, 0xa680}, {0xa682, 0xa682}, {0xa684, 0xa684},
  {0xa686, 0xa686}, {0xa688, 0xa688}, {0xa68a, 0xa68a}, {0xa68c, 0xa68c}, {0xa68e, 0xa68e}, {0xa690, 0xa690}, {0xa692, 0xa692}, {0xa694, 0xa694},
  {0xa696, 0xa696}, {0xa698, 0xa698}, {0xa69a, 0xa69a}, {0xa722, 0xa722}, {0xa724, 0xa724}, {0xa726, 0xa726}, {0xa728, 0xa728}, {0xa72a, 0xa72a},
  {0xa72c, 0xa72c}, {0xa72e, 0xa72e}, {0xa732, 0xa732}, {0xa734, 0xa734}, {0xa736, 0xa736}, {0xa738, 0xa738}, {0xa73a, 0xa73a}, {0xa73c, 0xa73c},
  {0xa73e, 0xa73e}, {0xa740, 0xa740}, {0xa742, 0xa742}, {0xa744, 0xa744}, {0xa746, 0xa746}, {0xa748, 0xa748}, {0xa74a, 0xa74a}, {0xa74c, 0xa74c},
  {0xa74e, 0xa74e}, {0xa750, 0xa750}, {0xa752, 0xa752}, {0xa754, 0xa754}, {0xa756, 0xa756}, {0xa758, 0xa758}, {0xa75a, 0xa75a}, {0xa75c, 0xa75c},
  {0xa75e, 0xa75e}, {0xa760, 0xa760}, {0xa762, 0xa762}, {0xa764, 0xa764}, {0xa766, 0xa766}, {0xa768, 0xa768}, {0xa76a, 0xa76a}, {0xa76c, 0xa76c},
  {0xa76e, 0xa76e}, {0xa779, 0xa779}, {0xa77b, 0xa77b}, {0xa77d, 0xa77e}, {0xa780, 0xa780}, {0xa782, 0xa782}, {0xa784, 0xa784}, {0xa786, 0xa786},
  {0xa78b, 0xa78b}, {0xa78d, 0xa78d}, {0xa790, 0xa790}, {0xa792, 0xa792}, {0xa796, 0xa796}, {0xa798, 0xa798}, {0xa79a, 0xa79a}, {0xa79c, 0xa79c},
  {0xa79e, 0xa79e}, {0xa7a0, 0xa7a0}, {0xa7a2, 0xa7a2}, {0xa7a4, 0xa7a4}, {0xa7a6, 0xa7a6}, {0xa7a8, 0xa7a8}, {0xa7aa, 0xa7ae}, {0xa7b0, 0xa7b4},
  {0xa7b6, 0xa7b6}, {0xa7b8, 0xa7b8}, {0xa7ba, 0xa7ba}, {0xa7bc, 0xa7bc}, {0xa7be, 0xa7be}, {0xa7c0, 0xa7c0}, {0xa7c2, 0xa7c2}, {0xa7c4, 0xa7c7},
  {0xa7c9, 0xa7c9}, {0xa7cb, 0xa7cc}, {0xa7d0, 0xa7d0}, {0xa7d6, 0xa7d6}, {0xa7d8, 0xa7d8}, {0xa7da, 0xa7da}, {0xa7dc, 0xa7dc}, {0xa7f5, 0xa7f5},
  {0xff21, 0xff3a}, {0x10400, 0x10427}, {0x104b0, 0x104d3}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10c80, 0x10cb2},
  {0x10d50, 0x10d65}, {0x118a0, 0x118bf}, {0x16e40, 0x16e5f}, {0x1d400, 0x1d419}, {0x1d434, 0x1d44d}, {0x1d468, 0x1d481}, {0x1d49c, 0x1d49c}, {0x1d49e, 0x1d49f},
  {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b5}, {0x1d4d0, 0x1d4e9}, {0x1d504, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514},
  {0x1d516, 0x1d51c}, {0x1d538, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d56c, 0x1d585}, {0x1d5a0, 0x1d5b9},
  {0x1d5d4, 0x1d5ed}, {0x1d608, 0x1d621}, {0x1d63c, 0x1d655}, {0x1d670, 0x1d689}, {0x1d6a8, 0x1d6c0}, {0x1d6e2, 0x1d6fa}, {0x1d71c, 0x1d734}, {0x1d756, 0x1d76e},
  {0x1d790, 0x1d7a8}, {0x1d7ca, 0x1d7ca}, {0x1e900, 0x1e921},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 184,
        '"', 285,
        '#', 236,
        '$', 123,
        '%', 259,
        '&', 231,
        '\'', 281,
        '(', 173,
        ')', 174,
        '*', 216,
        '+', 180,
        ',', 178,
        '-', 182,
        '.', 200,
        '0', 262,
        ':', 168,
        ';', 224,
        '<', 234,
        '=', 161,
        '>', 229,
        '?', 164,
        '@', 319,
        '[', 203,
        '\\', 2,
        ']', 214,
        '^', 321,
        '`', 361,
        'a', 347,
        'l', 343,
        'm', 338,
        '{', 219,
        '|', 187,
        '}', 225,
        '~', 162,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(136);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 1:
      if (lookahead == '\n') ADVANCE(293);
      if (lookahead == '\r') ADVANCE(1);
      END_STATE();
    case 2:
      ADVANCE_MAP(
        '\n', 293,
        '\r', 1,
        '#', 135,
        'o', 120,
        'u', 112,
        'x', 134,
        ' ', 299,
        '"', 299,
        '\'', 299,
        '\\', 299,
        'b', 299,
        'n', 299,
        'r', 299,
        't', 299,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(127);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '\n', 293,
        '\r', 1,
        'o', 120,
        'u', 112,
        'x', 134,
        ' ', 299,
        '"', 299,
        '\'', 299,
        '\\', 299,
        'b', 299,
        'n', 299,
        'r', 299,
        't', 299,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(127);
      END_STATE();
    case 4:
      if (lookahead == '\n') ADVANCE(282);
      if (lookahead == '\r') ADVANCE(5);
      if (lookahead == '[') ADVANCE(284);
      if (lookahead == '\\') ADVANCE(111);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(283);
      if (lookahead != 0 &&
          lookahead != '\'') ADVANCE(283);
      END_STATE();
    case 5:
      if (lookahead == '\n') ADVANCE(282);
      if (lookahead == '\r') ADVANCE(5);
      if (lookahead == '[') ADVANCE(104);
      if (lookahead == '\\') ADVANCE(111);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(107);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '!', 184,
        '%', 259,
        '\'', 360,
        '(', 173,
        '+', 179,
        '-', 181,
        '[', 104,
        '\\', 66,
        'a', 348,
        'l', 344,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(6);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '!', 184,
        '\'', 360,
        '(', 173,
        '+', 179,
        '-', 183,
        '.', 362,
        '[', 208,
        '\\', 66,
        'a', 348,
        'l', 344,
        '{', 70,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(7);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '!', 184,
        '\'', 360,
        '+', 179,
        ',', 178,
        '-', 181,
        '.', 88,
        ':', 92,
        ';', 223,
        '[', 104,
        '\\', 66,
        '|', 186,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(8);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '!', 184,
        '\'', 360,
        '+', 179,
        ',', 178,
        '-', 181,
        '.', 88,
        ':', 167,
        ';', 223,
        '[', 104,
        '\\', 66,
        '|', 186,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(8);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        '!', 99,
        '#', 236,
        '&', 232,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        '.', 199,
        ':', 194,
        ';', 223,
        '<', 324,
        '>', 325,
        '[', 212,
        '\\', 66,
        ']', 214,
        '{', 77,
        '|', 189,
        '}', 225,
        '$', 327,
        '=', 327,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(10);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        '!', 99,
        '#', 236,
        '&', 232,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        '.', 199,
        ':', 194,
        ';', 223,
        '<', 324,
        '>', 325,
        '[', 75,
        '\\', 66,
        ']', 214,
        '{', 77,
        '|', 188,
        '}', 225,
        '$', 327,
        '=', 327,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(11);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        '!', 99,
        '#', 236,
        '&', 232,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '.', 199,
        ':', 169,
        ';', 223,
        '<', 324,
        '[', 106,
        '\\', 66,
        'a', 347,
        '|', 189,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '$', 327,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(13);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 13:
      ADVANCE_MAP(
        '!', 99,
        '#', 236,
        '&', 232,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '.', 199,
        ':', 94,
        ';', 223,
        '<', 324,
        '[', 106,
        '\\', 66,
        'a', 347,
        '|', 189,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '$', 327,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(13);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 14:
      ADVANCE_MAP(
        '!', 99,
        '#', 236,
        '&', 232,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '.', 199,
        ':', 193,
        ';', 224,
        '<', 324,
        '[', 75,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 347,
        'l', 343,
        '{', 77,
        '|', 189,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '$', 327,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(14);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 15:
      ADVANCE_MAP(
        '!', 99,
        '#', 359,
        '&', 232,
        '\'', 281,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        ':', 194,
        ';', 224,
        '<', 324,
        '[', 212,
        '\\', 66,
        ']', 214,
        'a', 347,
        'l', 343,
        '{', 222,
        '|', 189,
        '}', 225,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '$', 327,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(16);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 16:
      ADVANCE_MAP(
        '!', 99,
        '#', 359,
        '&', 232,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        ':', 194,
        ';', 224,
        '<', 324,
        '[', 212,
        '\\', 66,
        ']', 214,
        'a', 347,
        'l', 343,
        '{', 222,
        '|', 189,
        '}', 225,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '$', 327,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(16);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 17:
      ADVANCE_MAP(
        '!', 99,
        '#', 359,
        '&', 232,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        ':', 169,
        ';', 224,
        '<', 324,
        '[', 75,
        '\\', 66,
        ']', 214,
        'a', 347,
        'l', 343,
        '{', 77,
        '|', 189,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '$', 327,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(18);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 18:
      ADVANCE_MAP(
        '!', 99,
        '#', 359,
        '&', 232,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        ':', 94,
        ';', 224,
        '<', 324,
        '[', 75,
        '\\', 66,
        ']', 214,
        'a', 347,
        'l', 343,
        '{', 77,
        '|', 189,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '$', 327,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(18);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 19:
      ADVANCE_MAP(
        '!', 99,
        '&', 232,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        ':', 170,
        ';', 223,
        '<', 324,
        '>', 325,
        '[', 75,
        '\\', 66,
        ']', 214,
        '{', 77,
        '|', 189,
        '}', 225,
        '$', 327,
        '=', 327,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(21);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 20:
      ADVANCE_MAP(
        '!', 99,
        '&', 232,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        ':', 170,
        ';', 223,
        '<', 324,
        '>', 325,
        '[', 75,
        '\\', 66,
        ']', 214,
        '{', 77,
        '|', 188,
        '}', 225,
        '$', 327,
        '=', 327,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(22);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 21:
      ADVANCE_MAP(
        '!', 99,
        '&', 232,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        ':', 194,
        ';', 223,
        '<', 324,
        '>', 325,
        '[', 75,
        '\\', 66,
        ']', 214,
        '{', 77,
        '|', 189,
        '}', 225,
        '$', 327,
        '=', 327,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(21);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 22:
      ADVANCE_MAP(
        '!', 99,
        '&', 232,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        ':', 194,
        ';', 223,
        '<', 324,
        '>', 325,
        '[', 75,
        '\\', 66,
        ']', 214,
        '{', 77,
        '|', 188,
        '}', 225,
        '$', 327,
        '=', 327,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(22);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 23:
      ADVANCE_MAP(
        '!', 99,
        '&', 232,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        ':', 194,
        ';', 223,
        '<', 324,
        '>', 325,
        '[', 106,
        '\\', 66,
        ']', 214,
        'a', 347,
        '|', 188,
        '}', 225,
        '$', 327,
        '=', 327,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(23);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 24:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 236,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        '.', 199,
        '0', 262,
        ':', 194,
        ';', 223,
        '<', 323,
        '=', 327,
        '>', 325,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 219,
        '|', 189,
        '}', 225,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(24);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 25:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 236,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        '.', 199,
        '0', 262,
        ':', 194,
        ';', 223,
        '<', 323,
        '=', 327,
        '>', 325,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 219,
        '|', 188,
        '}', 225,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(25);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 26:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 236,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '.', 199,
        '0', 262,
        ':', 169,
        ';', 223,
        '<', 323,
        '?', 165,
        '[', 211,
        '\\', 66,
        '`', 361,
        'a', 347,
        'l', 344,
        '{', 221,
        '|', 189,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(29);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 27:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 236,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '.', 199,
        '0', 262,
        ':', 94,
        ';', 224,
        '<', 323,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 347,
        'l', 343,
        '{', 219,
        '|', 189,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(27);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 28:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 236,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '.', 199,
        '0', 262,
        ':', 94,
        ';', 224,
        '<', 323,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 343,
        '{', 219,
        '|', 189,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(28);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 29:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 236,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '.', 199,
        '0', 262,
        ':', 94,
        ';', 223,
        '<', 323,
        '?', 165,
        '[', 211,
        '\\', 66,
        '`', 361,
        'a', 347,
        'l', 344,
        '{', 221,
        '|', 189,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(29);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 30:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 236,
        '&', 232,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        '-', 277,
        '.', 199,
        '0', 262,
        ':', 94,
        '<', 324,
        '?', 131,
        '[', 210,
        '\\', 66,
        '`', 361,
        'a', 347,
        'l', 343,
        'm', 338,
        '{', 220,
        '|', 113,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '$', 327,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(30);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 31:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 359,
        '$', 326,
        '%', 260,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '/', 316,
        '0', 262,
        ':', 169,
        ';', 224,
        '<', 324,
        '?', 131,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 347,
        'l', 343,
        'm', 338,
        '{', 219,
        '|', 189,
        '~', 163,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(32);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 32:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 359,
        '$', 326,
        '%', 260,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '/', 316,
        '0', 262,
        ':', 94,
        ';', 224,
        '<', 324,
        '?', 131,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 347,
        'l', 343,
        'm', 338,
        '{', 219,
        '|', 189,
        '~', 163,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(32);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 33:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 359,
        '$', 326,
        '%', 260,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '/', 316,
        '0', 262,
        ':', 94,
        ';', 224,
        '<', 324,
        '?', 131,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 343,
        'm', 338,
        '{', 219,
        '|', 189,
        '~', 163,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(33);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 34:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 359,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '0', 262,
        ':', 169,
        ';', 224,
        '<', 324,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 347,
        'l', 343,
        '{', 219,
        '|', 189,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(35);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 35:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 359,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '0', 262,
        ':', 94,
        ';', 224,
        '<', 324,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 347,
        'l', 343,
        '{', 219,
        '|', 189,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(35);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 36:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 130,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        '-', 277,
        '.', 199,
        '0', 262,
        ':', 94,
        ';', 223,
        '<', 324,
        '?', 131,
        '[', 210,
        '\\', 66,
        '`', 361,
        'a', 347,
        'l', 343,
        'm', 338,
        '{', 221,
        '|', 109,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(36);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 37:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '$', 326,
        '%', 260,
        '&', 232,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        '.', 88,
        '/', 316,
        '0', 262,
        ':', 194,
        ';', 223,
        '<', 324,
        '=', 327,
        '>', 325,
        '?', 131,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 343,
        'm', 338,
        '{', 219,
        '|', 189,
        '}', 225,
        '~', 163,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(37);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 38:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '$', 326,
        '%', 260,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '/', 316,
        '0', 262,
        ':', 94,
        ';', 223,
        '<', 324,
        '?', 131,
        '[', 210,
        '\\', 66,
        '`', 361,
        'a', 348,
        'l', 343,
        'm', 338,
        '{', 221,
        '|', 188,
        '~', 163,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(38);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 39:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '$', 326,
        '%', 260,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '/', 316,
        '0', 262,
        ':', 94,
        ';', 223,
        '<', 324,
        '?', 131,
        '[', 211,
        '\\', 66,
        '`', 361,
        'a', 347,
        'l', 343,
        'm', 338,
        '{', 221,
        '|', 189,
        '~', 163,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(39);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 40:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        '.', 362,
        '0', 262,
        ':', 170,
        ';', 223,
        '<', 324,
        '=', 327,
        '>', 325,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 219,
        '|', 188,
        '}', 225,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(41);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 41:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        '.', 362,
        '0', 262,
        ':', 194,
        ';', 223,
        '<', 324,
        '=', 327,
        '>', 325,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 219,
        '|', 188,
        '}', 225,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(41);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 42:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        '0', 262,
        ':', 170,
        ';', 223,
        '<', 324,
        '=', 327,
        '>', 325,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 219,
        '|', 189,
        '}', 225,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(43);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 43:
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 278,
        '0', 262,
        ':', 194,
        ';', 223,
        '<', 324,
        '=', 327,
        '>', 325,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 219,
        '|', 189,
        '}', 225,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(43);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 44:
      ADVANCE_MAP(
        '!', 308,
        '#', 130,
        '&', 232,
        '*', 315,
        '+', 276,
        '-', 277,
        '.', 199,
        ':', 100,
        '<', 324,
        '[', 104,
        '\\', 66,
        'a', 349,
        'l', 345,
        'm', 338,
        '|', 113,
        '%', 316,
        '/', 316,
        '?', 131,
        '~', 131,
        '@', 321,
        '^', 321,
        '$', 327,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(44);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 45:
      ADVANCE_MAP(
        '!', 307,
        '"', 285,
        '#', 359,
        '$', 123,
        '%', 259,
        '&', 230,
        '\'', 280,
        '(', 173,
        '0', 262,
        ';', 95,
        '>', 229,
        '?', 131,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 347,
        'l', 343,
        'm', 338,
        '{', 219,
        '|', 185,
        '~', 163,
        '+', 274,
        '-', 274,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(45);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 46:
      ADVANCE_MAP(
        '!', 307,
        '"', 285,
        '#', 359,
        '$', 123,
        '%', 259,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 215,
        '+', 274,
        '-', 275,
        '.', 199,
        '0', 262,
        ':', 197,
        ';', 95,
        '<', 87,
        '=', 161,
        '>', 229,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 343,
        'm', 338,
        '{', 219,
        '|', 108,
        '}', 225,
        '~', 163,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(46);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 47:
      ADVANCE_MAP(
        '!', 307,
        '"', 285,
        '#', 359,
        '$', 123,
        '\'', 280,
        '(', 173,
        '.', 362,
        '0', 262,
        ':', 190,
        ';', 95,
        '?', 165,
        '[', 209,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 343,
        'm', 338,
        '{', 219,
        '~', 163,
        '+', 274,
        '-', 274,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(47);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 48:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        '*', 215,
        '+', 273,
        ',', 178,
        '-', 279,
        '.', 362,
        '0', 262,
        ':', 195,
        ';', 223,
        '=', 161,
        '>', 229,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 218,
        '|', 185,
        '}', 225,
        '~', 162,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(48);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 49:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        '*', 215,
        '+', 273,
        '-', 279,
        '.', 199,
        '0', 262,
        ':', 196,
        ';', 95,
        '=', 161,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 343,
        '{', 218,
        '|', 185,
        '~', 162,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(49);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 50:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        ',', 178,
        '.', 199,
        '0', 262,
        ':', 195,
        ';', 95,
        '=', 161,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 347,
        'l', 343,
        '{', 218,
        '|', 185,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(50);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 51:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        ',', 178,
        '.', 199,
        '0', 262,
        ':', 195,
        ';', 223,
        '=', 161,
        '>', 229,
        '?', 164,
        '[', 211,
        '\\', 66,
        '`', 361,
        'a', 347,
        'l', 344,
        '{', 220,
        '|', 185,
        '}', 225,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(51);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 52:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        ',', 178,
        '.', 199,
        '0', 262,
        ':', 92,
        ';', 223,
        '=', 161,
        '>', 115,
        '[', 210,
        '\\', 66,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 220,
        '|', 186,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(52);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 53:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        ',', 178,
        '.', 199,
        '0', 262,
        ':', 92,
        ';', 223,
        '[', 210,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 220,
        '|', 185,
        '}', 225,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(53);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 54:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        ',', 178,
        '.', 88,
        '0', 262,
        ':', 195,
        ';', 95,
        '=', 161,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 347,
        'l', 343,
        '{', 218,
        '|', 185,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(54);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 55:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        ',', 178,
        '.', 88,
        '0', 262,
        ':', 195,
        '=', 161,
        '?', 164,
        '[', 211,
        '\\', 66,
        '`', 361,
        'a', 347,
        'l', 344,
        '{', 220,
        '|', 185,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(55);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 56:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        ',', 178,
        '0', 262,
        ':', 195,
        ';', 95,
        '=', 161,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 347,
        'l', 343,
        '{', 218,
        '|', 185,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(56);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 57:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        ',', 178,
        '0', 262,
        ':', 195,
        '=', 161,
        '?', 164,
        '[', 211,
        '\\', 66,
        '`', 361,
        'a', 347,
        'l', 344,
        '{', 220,
        '|', 185,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(57);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 58:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        ',', 178,
        '0', 262,
        ':', 171,
        ';', 95,
        '=', 161,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 347,
        'l', 343,
        '{', 218,
        '|', 185,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(56);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 59:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        ',', 178,
        '0', 262,
        ':', 171,
        '=', 161,
        '?', 164,
        '[', 211,
        '\\', 66,
        '`', 361,
        'a', 347,
        'l', 344,
        '{', 220,
        '|', 185,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(57);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 60:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        '0', 262,
        ':', 190,
        ';', 95,
        '=', 161,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 343,
        '{', 218,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(60);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 61:
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        '0', 262,
        ':', 166,
        ';', 95,
        '=', 161,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 343,
        '{', 218,
        '~', 162,
        '+', 273,
        '-', 273,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(60);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 62:
      if (lookahead == '"') ADVANCE(285);
      if (lookahead == '%') ADVANCE(288);
      if (lookahead == '@') ADVANCE(289);
      if (lookahead == '[') ADVANCE(290);
      if (lookahead == '\\') ADVANCE(3);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(286);
      if (lookahead != 0) ADVANCE(291);
      END_STATE();
    case 63:
      ADVANCE_MAP(
        '#', 236,
        '(', 173,
        ')', 174,
        ',', 178,
        '-', 101,
        '.', 199,
        ':', 195,
        ';', 223,
        '=', 161,
        '>', 115,
        '[', 75,
        '\\', 66,
        ']', 214,
        '{', 77,
        '|', 185,
        '}', 225,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(63);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 64:
      ADVANCE_MAP(
        '#', 236,
        ')', 174,
        '*', 215,
        ',', 178,
        '-', 101,
        '.', 200,
        ':', 195,
        ';', 223,
        '=', 161,
        '>', 229,
        '[', 104,
        '\\', 66,
        ']', 214,
        '|', 185,
        '}', 225,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(64);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 65:
      ADVANCE_MAP(
        '#', 236,
        ')', 174,
        ',', 178,
        '-', 101,
        '.', 200,
        ':', 192,
        ';', 223,
        '=', 161,
        '>', 229,
        '[', 106,
        '\\', 66,
        ']', 214,
        'a', 348,
        '|', 185,
        '}', 225,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(65);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 66:
      if (lookahead == '#') ADVANCE(135);
      END_STATE();
    case 67:
      ADVANCE_MAP(
        '#', 235,
        '\'', 360,
        '*', 215,
        '-', 101,
        '.', 199,
        ':', 102,
        ';', 223,
        '=', 161,
        '>', 229,
        '[', 104,
        '\\', 66,
        'a', 348,
        'l', 344,
        '}', 225,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(67);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 68:
      ADVANCE_MAP(
        '#', 235,
        '\'', 360,
        '*', 215,
        '-', 101,
        '.', 199,
        ':', 97,
        ';', 95,
        '=', 161,
        '[', 75,
        '\\', 66,
        ']', 214,
        'a', 348,
        'l', 343,
        '{', 77,
        '|', 185,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(68);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 69:
      if (lookahead == '%') ADVANCE(253);
      if (lookahead == '@') ADVANCE(246);
      END_STATE();
    case 70:
      if (lookahead == '%') ADVANCE(255);
      END_STATE();
    case 71:
      if (lookahead == '%') ADVANCE(257);
      END_STATE();
    case 72:
      if (lookahead == '%') ADVANCE(258);
      END_STATE();
    case 73:
      if (lookahead == '%') ADVANCE(252);
      if (lookahead == '@') ADVANCE(247);
      END_STATE();
    case 74:
      if (lookahead == '%') ADVANCE(254);
      END_STATE();
    case 75:
      if (lookahead == '%') ADVANCE(71);
      if (lookahead == '@') ADVANCE(246);
      END_STATE();
    case 76:
      if (lookahead == '%') ADVANCE(295);
      if (lookahead == '@') ADVANCE(296);
      if (lookahead == '[') ADVANCE(297);
      if (lookahead == '|') ADVANCE(294);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(286);
      if (lookahead != 0) ADVANCE(298);
      END_STATE();
    case 77:
      if (lookahead == '%') ADVANCE(72);
      END_STATE();
    case 78:
      ADVANCE_MAP(
        '(', 173,
        ')', 174,
        ',', 178,
        '-', 101,
        ':', 93,
        ';', 223,
        '=', 161,
        '[', 73,
        '\\', 66,
        'a', 348,
        '{', 74,
        '|', 186,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(78);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 79:
      if (lookahead == '(') ADVANCE(173);
      if (lookahead == '[') ADVANCE(213);
      if (lookahead == '\\') ADVANCE(66);
      if (lookahead == 'a') ADVANCE(348);
      if (lookahead == 'l') ADVANCE(344);
      if (lookahead == '{') ADVANCE(217);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(79);
      if (set_contains(sym_indexing_operator_character_set_1, 9, lookahead)) ADVANCE(334);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 80:
      ADVANCE_MAP(
        ')', 174,
        ',', 178,
        '-', 101,
        '.', 88,
        ':', 191,
        ';', 223,
        '=', 161,
        '[', 104,
        '\\', 66,
        ']', 214,
        '|', 185,
        '}', 225,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(80);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 81:
      ADVANCE_MAP(
        ')', 174,
        ',', 178,
        '-', 101,
        ':', 191,
        ';', 223,
        '=', 161,
        '[', 104,
        '\\', 66,
        ']', 214,
        '|', 185,
        '}', 225,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(81);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 82:
      ADVANCE_MAP(
        ')', 174,
        ',', 178,
        '-', 101,
        ':', 167,
        ';', 223,
        '=', 161,
        '[', 104,
        '\\', 66,
        ']', 214,
        '|', 185,
        '}', 225,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(81);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 83:
      if (lookahead == '*') ADVANCE(91);
      if (lookahead == '.') ADVANCE(84);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(304);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(90);
      if (set_contains(sym_conversion_specification_character_set_6, 15, lookahead)) ADVANCE(303);
      END_STATE();
    case 84:
      if (lookahead == '*') ADVANCE(119);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(304);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(118);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(303);
      END_STATE();
    case 85:
      ADVANCE_MAP(
        ',', 178,
        '.', 199,
        ':', 92,
        ';', 223,
        '[', 106,
        '\\', 66,
        ']', 214,
        'a', 348,
        '|', 185,
        '}', 225,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(85);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 86:
      if (lookahead == ',') ADVANCE(178);
      if (lookahead == '.') ADVANCE(199);
      if (lookahead == ':') ADVANCE(92);
      if (lookahead == ';') ADVANCE(223);
      if (lookahead == '[') ADVANCE(104);
      if (lookahead == '\\') ADVANCE(66);
      if (lookahead == '|') ADVANCE(186);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(86);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 87:
      if (lookahead == '-') ADVANCE(240);
      END_STATE();
    case 88:
      if (lookahead == '.') ADVANCE(176);
      END_STATE();
    case 89:
      if (lookahead == '.') ADVANCE(362);
      if (lookahead == '[') ADVANCE(104);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(256);
      END_STATE();
    case 90:
      if (lookahead == '.') ADVANCE(84);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(304);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(90);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(303);
      END_STATE();
    case 91:
      if (lookahead == '.') ADVANCE(84);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(304);
      if (set_contains(sym_conversion_specification_character_set_6, 15, lookahead)) ADVANCE(303);
      END_STATE();
    case 92:
      if (lookahead == ':') ADVANCE(237);
      END_STATE();
    case 93:
      if (lookahead == ':') ADVANCE(237);
      if (lookahead == '=') ADVANCE(175);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 94:
      if (lookahead == ':') ADVANCE(237);
      if (lookahead == '=') ADVANCE(333);
      END_STATE();
    case 95:
      if (lookahead == ';') ADVANCE(156);
      END_STATE();
    case 96:
      if (lookahead == '=') ADVANCE(177);
      END_STATE();
    case 97:
      if (lookahead == '=') ADVANCE(175);
      END_STATE();
    case 98:
      if (lookahead == '=') ADVANCE(175);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 99:
      if (lookahead == '=') ADVANCE(322);
      END_STATE();
    case 100:
      if (lookahead == '=') ADVANCE(333);
      END_STATE();
    case 101:
      if (lookahead == '>') ADVANCE(201);
      END_STATE();
    case 102:
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 103:
      if (lookahead == '>') ADVANCE(305);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(103);
      END_STATE();
    case 104:
      if (lookahead == '@') ADVANCE(245);
      END_STATE();
    case 105:
      if (lookahead == '@') ADVANCE(251);
      END_STATE();
    case 106:
      if (lookahead == '@') ADVANCE(247);
      END_STATE();
    case 107:
      if (lookahead == '[') ADVANCE(104);
      if (lookahead == '\\') ADVANCE(111);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(107);
      END_STATE();
    case 108:
      if (lookahead == ']') ADVANCE(239);
      END_STATE();
    case 109:
      if (lookahead == ']') ADVANCE(239);
      if (lookahead == '|') ADVANCE(332);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 110:
      if (lookahead == 'n') ADVANCE(305);
      END_STATE();
    case 111:
      ADVANCE_MAP(
        'o', 120,
        'x', 134,
        ' ', 299,
        '"', 299,
        '\'', 299,
        '\\', 299,
        'b', 299,
        'n', 299,
        'r', 299,
        't', 299,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(127);
      END_STATE();
    case 112:
      if (lookahead == '{') ADVANCE(132);
      END_STATE();
    case 113:
      if (lookahead == '|') ADVANCE(332);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 114:
      if (lookahead == '}') ADVANCE(292);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(114);
      END_STATE();
    case 115:
      if (lookahead == '}') ADVANCE(243);
      END_STATE();
    case 116:
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(124);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(271);
      END_STATE();
    case 117:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(268);
      END_STATE();
    case 118:
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(304);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(118);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(303);
      END_STATE();
    case 119:
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(304);
      if (set_contains(sym_conversion_specification_character_set_6, 15, lookahead)) ADVANCE(303);
      END_STATE();
    case 120:
      if (('0' <= lookahead && lookahead <= '3')) ADVANCE(122);
      END_STATE();
    case 121:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(302);
      END_STATE();
    case 122:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(121);
      END_STATE();
    case 123:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(244);
      END_STATE();
    case 124:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(271);
      END_STATE();
    case 125:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(103);
      END_STATE();
    case 126:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(300);
      END_STATE();
    case 127:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(126);
      END_STATE();
    case 128:
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 129:
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(318);
      END_STATE();
    case 130:
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(313);
      END_STATE();
    case 131:
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(309);
      END_STATE();
    case 132:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(114);
      END_STATE();
    case 133:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(301);
      END_STATE();
    case 134:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(133);
      END_STATE();
    case 135:
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 136:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 184,
        '"', 285,
        '#', 236,
        '$', 123,
        '%', 259,
        '&', 231,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 216,
        '+', 180,
        ',', 178,
        '-', 182,
        '.', 200,
        '0', 262,
        ':', 192,
        ';', 224,
        '<', 234,
        '=', 161,
        '>', 229,
        '?', 164,
        '@', 319,
        '[', 204,
        '\\', 2,
        ']', 214,
        '^', 321,
        '`', 361,
        'a', 347,
        'l', 343,
        'm', 338,
        '{', 219,
        '|', 187,
        '}', 225,
        '~', 162,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(136);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 137:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 184,
        '"', 285,
        '#', 235,
        '%', 259,
        '\'', 280,
        '(', 173,
        '+', 273,
        '-', 279,
        '0', 262,
        ':', 198,
        ';', 95,
        '=', 161,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 218,
        '|', 108,
        '~', 162,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(137);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 138:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 184,
        '"', 285,
        '%', 259,
        '&', 230,
        '\'', 280,
        '(', 173,
        ')', 174,
        '+', 273,
        ',', 178,
        '-', 279,
        '.', 88,
        '0', 262,
        ':', 192,
        ';', 95,
        '=', 161,
        '>', 229,
        '[', 206,
        '\\', 66,
        ']', 214,
        '{', 218,
        '|', 185,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(138);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 139:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 184,
        '#', 359,
        '&', 230,
        '(', 173,
        ')', 174,
        '*', 215,
        '+', 96,
        ',', 178,
        '-', 101,
        '.', 199,
        ':', 192,
        ';', 95,
        '=', 161,
        '>', 229,
        '[', 206,
        '\\', 66,
        ']', 214,
        'a', 348,
        'l', 343,
        '{', 70,
        '|', 185,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(139);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 140:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 99,
        '"', 285,
        '#', 236,
        '&', 232,
        '\'', 360,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '.', 199,
        '0', 262,
        ':', 194,
        ';', 95,
        '<', 324,
        '>', 325,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        '{', 218,
        '|', 109,
        '}', 225,
        '$', 327,
        '=', 327,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(140);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 141:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 99,
        '"', 285,
        '#', 359,
        '&', 232,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '.', 199,
        ':', 94,
        ';', 224,
        '<', 324,
        '[', 69,
        '\\', 66,
        ']', 214,
        'l', 343,
        '{', 218,
        '|', 189,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '$', 327,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(141);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 142:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 236,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '.', 199,
        '0', 262,
        ':', 194,
        ';', 95,
        '<', 324,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 219,
        '|', 113,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(142);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 143:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 359,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '0', 262,
        ':', 169,
        ';', 224,
        '<', 324,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 343,
        '{', 219,
        '|', 189,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(144);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 144:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 308,
        '"', 285,
        '#', 359,
        '$', 326,
        '&', 232,
        '\'', 280,
        '(', 173,
        '*', 315,
        '+', 276,
        ',', 178,
        '-', 277,
        '0', 262,
        ':', 94,
        ';', 224,
        '<', 324,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 343,
        '{', 219,
        '|', 189,
        '~', 163,
        '%', 316,
        '/', 316,
        '@', 321,
        '^', 321,
        '=', 327,
        '>', 327,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(144);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 145:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 307,
        '"', 285,
        '#', 235,
        '$', 123,
        '%', 259,
        '&', 230,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 215,
        '+', 273,
        ',', 178,
        '-', 279,
        '.', 88,
        '0', 262,
        ':', 192,
        ';', 95,
        '=', 161,
        '>', 229,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 219,
        '|', 185,
        '~', 163,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(145);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 146:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 307,
        '"', 285,
        '$', 123,
        '\'', 280,
        '(', 173,
        '0', 262,
        ':', 166,
        ';', 95,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 219,
        '~', 163,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(147);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 147:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '!', 307,
        '"', 285,
        '$', 123,
        '\'', 280,
        '(', 173,
        '0', 262,
        ';', 95,
        '?', 165,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 219,
        '~', 163,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(147);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 148:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '"', 285,
        '#', 236,
        '&', 230,
        '\'', 280,
        '(', 173,
        ')', 174,
        '*', 215,
        '+', 273,
        ',', 178,
        '-', 279,
        '.', 199,
        '0', 262,
        ':', 192,
        ';', 95,
        '=', 161,
        '>', 229,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 218,
        '|', 185,
        '~', 162,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(148);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 149:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        '+', 273,
        '-', 279,
        '0', 262,
        ':', 198,
        ';', 95,
        '=', 161,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 218,
        '~', 162,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(149);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 150:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '"', 285,
        '#', 235,
        '\'', 280,
        '(', 173,
        '+', 273,
        '-', 279,
        '0', 262,
        ':', 172,
        ';', 95,
        '=', 161,
        '?', 164,
        '[', 207,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 218,
        '~', 162,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(149);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 151:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '#', 236,
        '&', 230,
        '(', 173,
        ')', 174,
        '*', 215,
        '+', 96,
        ',', 178,
        '-', 101,
        '.', 199,
        ':', 192,
        ';', 95,
        '=', 161,
        '>', 229,
        '[', 206,
        '\\', 66,
        ']', 214,
        'a', 348,
        '{', 70,
        '|', 185,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(151);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 152:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '#', 359,
        ')', 174,
        ',', 178,
        '-', 101,
        ':', 192,
        ';', 95,
        '=', 161,
        '[', 75,
        '\\', 66,
        ']', 214,
        'l', 343,
        '{', 77,
        '|', 185,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(152);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 153:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '#', 235,
        '%', 259,
        '&', 230,
        '\'', 360,
        ')', 174,
        '*', 215,
        ',', 178,
        '-', 101,
        '.', 199,
        ':', 98,
        ';', 95,
        '=', 161,
        '>', 229,
        '[', 206,
        '\\', 66,
        ']', 214,
        'a', 348,
        'l', 344,
        '{', 70,
        '|', 185,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(153);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 154:
      if (eof) ADVANCE(155);
      ADVANCE_MAP(
        '#', 235,
        '&', 230,
        '\'', 360,
        '(', 173,
        ')', 174,
        '*', 215,
        '+', 96,
        ',', 178,
        '-', 101,
        '.', 88,
        ':', 192,
        ';', 95,
        '<', 233,
        '=', 161,
        '>', 229,
        '?', 164,
        '[', 205,
        '\\', 66,
        ']', 214,
        '`', 361,
        'a', 348,
        'l', 344,
        '{', 218,
        '|', 185,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(154);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(358);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(357);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_SEMI_SEMI);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(anon_sym_let);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(335);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(anon_sym_let);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(anon_sym_and);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(336);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_and);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_TILDE);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(309);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(309);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(237);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(237);
      if (lookahead == '=') ADVANCE(175);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(237);
      if (lookahead == '=') ADVANCE(333);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(237);
      if (lookahead == '=') ADVANCE(333);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(237);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_COLON_EQ);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '.') ADVANCE(310);
      if (lookahead == '=') ADVANCE(177);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(310);
      if (lookahead == '>') ADVANCE(201);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '>') ADVANCE(201);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ']') ADVANCE(239);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ']') ADVANCE(239);
      if (lookahead == '|') ADVANCE(331);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ']') ADVANCE(239);
      if (lookahead == '|') ADVANCE(332);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(332);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_COLON2);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(237);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(237);
      if (lookahead == '=') ADVANCE(175);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(237);
      if (lookahead == '=') ADVANCE(333);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(237);
      if (lookahead == '=') ADVANCE(333);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(237);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == '=') ADVANCE(175);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == '=') ADVANCE(175);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == '>') ADVANCE(226);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(176);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(318);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(253);
      if (lookahead == '<') ADVANCE(228);
      if (lookahead == '>') ADVANCE(227);
      if (lookahead == '@') ADVANCE(287);
      if (lookahead == '|') ADVANCE(238);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(253);
      if (lookahead == '<') ADVANCE(228);
      if (lookahead == '>') ADVANCE(227);
      if (lookahead == '@') ADVANCE(245);
      if (lookahead == '|') ADVANCE(238);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(253);
      if (lookahead == '<') ADVANCE(228);
      if (lookahead == '>') ADVANCE(227);
      if (lookahead == '@') ADVANCE(246);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(253);
      if (lookahead == '@') ADVANCE(246);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(253);
      if (lookahead == '@') ADVANCE(246);
      if (lookahead == '|') ADVANCE(238);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(253);
      if (lookahead == '@') ADVANCE(248);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(253);
      if (lookahead == '@') ADVANCE(248);
      if (lookahead == '|') ADVANCE(238);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(252);
      if (lookahead == '@') ADVANCE(245);
      if (lookahead == '|') ADVANCE(238);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(252);
      if (lookahead == '@') ADVANCE(247);
      if (lookahead == '|') ADVANCE(238);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(71);
      if (lookahead == '@') ADVANCE(246);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '@') ADVANCE(245);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(314);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(255);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(255);
      if (lookahead == '<') ADVANCE(242);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(254);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(254);
      if (lookahead == '<') ADVANCE(242);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(72);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym_SEMI);
      if (lookahead == ';') ADVANCE(156);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_COLON_GT);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_LBRACK_GT);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(anon_sym_LBRACK_LT);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(329);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(330);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '-') ADVANCE(240);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(anon_sym_POUND);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(anon_sym_POUND);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(313);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(anon_sym_LBRACK_PIPE);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(anon_sym_PIPE_RBRACK);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(anon_sym_LBRACE_LT);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(anon_sym_GT_RBRACE);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(sym_ocamlyacc_value);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(244);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      if (lookahead == '@') ADVANCE(250);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      if (lookahead == '@') ADVANCE(249);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      if (lookahead == '@') ADVANCE(105);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(anon_sym_LBRACK_AT_AT);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(anon_sym_LBRACK_AT_AT);
      if (lookahead == '@') ADVANCE(251);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(anon_sym_LBRACK_AT_AT_AT);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(anon_sym_LBRACK_PERCENT);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(anon_sym_LBRACK_PERCENT);
      if (lookahead == '%') ADVANCE(257);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(anon_sym_LBRACE_PERCENT);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(anon_sym_LBRACE_PERCENT);
      if (lookahead == '%') ADVANCE(258);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(aux_sym_quoted_extension_token1);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(256);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(anon_sym_LBRACK_PERCENT_PERCENT);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(anon_sym_LBRACE_PERCENT_PERCENT);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(316);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(aux_sym_number_token1);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(aux_sym_number_token1);
      ADVANCE_MAP(
        '.', 266,
        'B', 117,
        'b', 117,
        'E', 116,
        'e', 116,
        'O', 269,
        'o', 269,
        'X', 272,
        'x', 272,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(263);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(261);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '.') ADVANCE(266);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(116);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(263);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(261);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '.') ADVANCE(267);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(265);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(264);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(261);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(124);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(271);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(116);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(266);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(261);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(265);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(267);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(261);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '0' ||
          lookahead == '1' ||
          lookahead == '_') ADVANCE(268);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(261);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(270);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '7') ||
          lookahead == '_') ADVANCE(270);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(261);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(271);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(261);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(264);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(310);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(310);
      if (lookahead == '>') ADVANCE(201);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(312);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(318);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(311);
      if (lookahead == '>') ADVANCE(129);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(317);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(311);
      if (lookahead == '>') ADVANCE(202);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(317);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '>') ADVANCE(201);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(anon_sym_SQUOTE2);
      END_STATE();
    case 282:
      ACCEPT_TOKEN(aux_sym_character_content_token1);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(aux_sym_character_content_token2);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(aux_sym_character_content_token2);
      if (lookahead == '@') ADVANCE(245);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(aux_sym_string_content_token2);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      ADVANCE_MAP(
        '*', 91,
        '.', 84,
        'L', 304,
        'l', 304,
        'n', 304,
        ' ', 83,
        '#', 83,
        '+', 83,
        '-', 83,
        '0', 83,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(90);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(303);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      ADVANCE_MAP(
        '<', 125,
        '\\', 110,
        ' ', 305,
        ',', 305,
        '.', 305,
        ';', 305,
        '?', 305,
        '[', 305,
        ']', 305,
        '{', 305,
        '}', 305,
      );
      END_STATE();
    case 290:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      if (lookahead == '@') ADVANCE(287);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '%' &&
          lookahead != '\\') ADVANCE(291);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '%' &&
          lookahead != '@' &&
          lookahead != '\\') ADVANCE(291);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(aux_sym_string_content_token4);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(aux_sym_string_content_token5);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(293);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      ADVANCE_MAP(
        '*', 91,
        '.', 84,
        'L', 304,
        'l', 304,
        'n', 304,
        ' ', 83,
        '#', 83,
        '+', 83,
        '-', 83,
        '0', 83,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(90);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(303);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      ADVANCE_MAP(
        '<', 125,
        '\\', 110,
        ' ', 305,
        ',', 305,
        '.', 305,
        ';', 305,
        '?', 305,
        '[', 305,
        ']', 305,
        '{', 305,
        '}', 305,
      );
      END_STATE();
    case 297:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      if (lookahead == '@') ADVANCE(287);
      if (lookahead != 0 &&
          lookahead != '%' &&
          lookahead != '|') ADVANCE(298);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      if (lookahead != 0 &&
          lookahead != '%' &&
          lookahead != '@' &&
          lookahead != '|') ADVANCE(298);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token1);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token2);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token3);
      END_STATE();
    case 302:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token4);
      END_STATE();
    case 303:
      ACCEPT_TOKEN(sym_conversion_specification);
      END_STATE();
    case 304:
      ACCEPT_TOKEN(sym_conversion_specification);
      if (lookahead == 'X' ||
          lookahead == 'd' ||
          lookahead == 'i' ||
          lookahead == 'o' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(303);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(sym_pretty_printing_indication);
      END_STATE();
    case 306:
      ACCEPT_TOKEN(sym_pretty_printing_indication);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(321);
      END_STATE();
    case 307:
      ACCEPT_TOKEN(sym_prefix_operator);
      if (lookahead == '=') ADVANCE(131);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(309);
      END_STATE();
    case 308:
      ACCEPT_TOKEN(sym_prefix_operator);
      if (lookahead == '=') ADVANCE(328);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(309);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(sym_prefix_operator);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(309);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(aux_sym_sign_operator_token1);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(aux_sym_sign_operator_token1);
      if (lookahead == '>') ADVANCE(318);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(317);
      END_STATE();
    case 312:
      ACCEPT_TOKEN(aux_sym_sign_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(318);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(sym_hash_operator);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(313);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(aux_sym_pow_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(314);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(aux_sym_mult_operator_token1);
      if (lookahead == '*') ADVANCE(314);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(316);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(aux_sym_mult_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(316);
      END_STATE();
    case 317:
      ACCEPT_TOKEN(aux_sym_add_operator_token1);
      if (lookahead == '>') ADVANCE(318);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(317);
      END_STATE();
    case 318:
      ACCEPT_TOKEN(aux_sym_add_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(318);
      END_STATE();
    case 319:
      ACCEPT_TOKEN(sym_concat_operator);
      ADVANCE_MAP(
        '<', 320,
        '\\', 110,
        '.', 306,
        '?', 306,
        ' ', 305,
        ',', 305,
        ';', 305,
        '[', 305,
        ']', 305,
        '{', 305,
        '}', 305,
      );
      if (lookahead == '!' ||
          ('$' <= lookahead && lookahead <= '&') ||
          ('*' <= lookahead && lookahead <= '/') ||
          (':' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          ('|' <= lookahead && lookahead <= '~')) ADVANCE(321);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(sym_concat_operator);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(103);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(321);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(sym_concat_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(321);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(sym_rel_operator);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(sym_rel_operator);
      if (lookahead == '-') ADVANCE(241);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 324:
      ACCEPT_TOKEN(sym_rel_operator);
      if (lookahead == '-') ADVANCE(128);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 325:
      ACCEPT_TOKEN(sym_rel_operator);
      if (lookahead == '}') ADVANCE(243);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(sym_rel_operator);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(244);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 327:
      ACCEPT_TOKEN(sym_rel_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(sym_rel_operator);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(309);
      END_STATE();
    case 329:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 330:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 331:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 332:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(327);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(sym_assign_operator);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(sym_indexing_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(334);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(sym_let_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(335);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(sym_let_and_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(336);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(sym_match_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(337);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'a') ADVANCE(351);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 339:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'c') ADVANCE(346);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 340:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'd') ADVANCE(159);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'd') ADVANCE(160);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 342:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'd') ADVANCE(354);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 343:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'e') ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'e') ADVANCE(352);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 345:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'e') ADVANCE(353);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 346:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'h') ADVANCE(356);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 347:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'n') ADVANCE(340);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 348:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'n') ADVANCE(341);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 349:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'n') ADVANCE(342);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(157);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(339);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(158);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 353:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(336);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(335);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 356:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(337);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 357:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(357);
      END_STATE();
    case 358:
      ACCEPT_TOKEN(sym__uppercase_identifier);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(358);
      END_STATE();
    case 359:
      ACCEPT_TOKEN(aux_sym_directive_token1);
      END_STATE();
    case 360:
      ACCEPT_TOKEN(aux_sym_type_variable_token1);
      END_STATE();
    case 361:
      ACCEPT_TOKEN(aux_sym_tag_token1);
      END_STATE();
    case 362:
      ACCEPT_TOKEN(aux_sym_attribute_id_token1);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      ADVANCE_MAP(
        '_', 1,
        'a', 2,
        'b', 3,
        'c', 4,
        'd', 5,
        'e', 6,
        'f', 7,
        'i', 8,
        'l', 9,
        'm', 10,
        'n', 11,
        'o', 12,
        'p', 13,
        'r', 14,
        's', 15,
        't', 16,
        'v', 17,
        'w', 18,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(0);
      END_STATE();
    case 1:
      ACCEPT_TOKEN(anon_sym__);
      END_STATE();
    case 2:
      if (lookahead == 's') ADVANCE(19);
      END_STATE();
    case 3:
      if (lookahead == 'e') ADVANCE(20);
      END_STATE();
    case 4:
      if (lookahead == 'l') ADVANCE(21);
      if (lookahead == 'o') ADVANCE(22);
      END_STATE();
    case 5:
      if (lookahead == 'o') ADVANCE(23);
      END_STATE();
    case 6:
      if (lookahead == 'f') ADVANCE(24);
      if (lookahead == 'l') ADVANCE(25);
      if (lookahead == 'n') ADVANCE(26);
      if (lookahead == 'x') ADVANCE(27);
      END_STATE();
    case 7:
      if (lookahead == 'a') ADVANCE(28);
      if (lookahead == 'o') ADVANCE(29);
      if (lookahead == 'u') ADVANCE(30);
      END_STATE();
    case 8:
      if (lookahead == 'f') ADVANCE(31);
      if (lookahead == 'n') ADVANCE(32);
      END_STATE();
    case 9:
      if (lookahead == 'a') ADVANCE(33);
      if (lookahead == 'o') ADVANCE(34);
      if (lookahead == 's') ADVANCE(35);
      if (lookahead == 'x') ADVANCE(36);
      END_STATE();
    case 10:
      if (lookahead == 'a') ADVANCE(37);
      if (lookahead == 'e') ADVANCE(38);
      if (lookahead == 'o') ADVANCE(39);
      if (lookahead == 'u') ADVANCE(40);
      END_STATE();
    case 11:
      if (lookahead == 'e') ADVANCE(41);
      if (lookahead == 'o') ADVANCE(42);
      END_STATE();
    case 12:
      if (lookahead == 'b') ADVANCE(43);
      if (lookahead == 'f') ADVANCE(44);
      if (lookahead == 'p') ADVANCE(45);
      if (lookahead == 'r') ADVANCE(46);
      END_STATE();
    case 13:
      if (lookahead == 'r') ADVANCE(47);
      END_STATE();
    case 14:
      if (lookahead == 'e') ADVANCE(48);
      END_STATE();
    case 15:
      if (lookahead == 'i') ADVANCE(49);
      if (lookahead == 't') ADVANCE(50);
      END_STATE();
    case 16:
      if (lookahead == 'h') ADVANCE(51);
      if (lookahead == 'o') ADVANCE(52);
      if (lookahead == 'r') ADVANCE(53);
      if (lookahead == 'y') ADVANCE(54);
      END_STATE();
    case 17:
      if (lookahead == 'a') ADVANCE(55);
      if (lookahead == 'i') ADVANCE(56);
      END_STATE();
    case 18:
      if (lookahead == 'h') ADVANCE(57);
      if (lookahead == 'i') ADVANCE(58);
      END_STATE();
    case 19:
      ACCEPT_TOKEN(anon_sym_as);
      if (lookahead == 'r') ADVANCE(59);
      if (lookahead == 's') ADVANCE(60);
      END_STATE();
    case 20:
      if (lookahead == 'g') ADVANCE(61);
      END_STATE();
    case 21:
      if (lookahead == 'a') ADVANCE(62);
      END_STATE();
    case 22:
      if (lookahead == 'n') ADVANCE(63);
      END_STATE();
    case 23:
      ACCEPT_TOKEN(anon_sym_do);
      if (lookahead == 'n') ADVANCE(64);
      if (lookahead == 'w') ADVANCE(65);
      END_STATE();
    case 24:
      if (lookahead == 'f') ADVANCE(66);
      END_STATE();
    case 25:
      if (lookahead == 's') ADVANCE(67);
      END_STATE();
    case 26:
      if (lookahead == 'd') ADVANCE(68);
      END_STATE();
    case 27:
      if (lookahead == 'c') ADVANCE(69);
      if (lookahead == 't') ADVANCE(70);
      END_STATE();
    case 28:
      if (lookahead == 'l') ADVANCE(71);
      END_STATE();
    case 29:
      if (lookahead == 'r') ADVANCE(72);
      END_STATE();
    case 30:
      if (lookahead == 'n') ADVANCE(73);
      END_STATE();
    case 31:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 32:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == 'c') ADVANCE(74);
      if (lookahead == 'h') ADVANCE(75);
      if (lookahead == 'i') ADVANCE(76);
      END_STATE();
    case 33:
      if (lookahead == 'n') ADVANCE(77);
      if (lookahead == 'z') ADVANCE(78);
      END_STATE();
    case 34:
      if (lookahead == 'r') ADVANCE(79);
      END_STATE();
    case 35:
      if (lookahead == 'l') ADVANCE(80);
      if (lookahead == 'r') ADVANCE(81);
      END_STATE();
    case 36:
      if (lookahead == 'o') ADVANCE(82);
      END_STATE();
    case 37:
      if (lookahead == 't') ADVANCE(83);
      END_STATE();
    case 38:
      if (lookahead == 't') ADVANCE(84);
      END_STATE();
    case 39:
      if (lookahead == 'd') ADVANCE(85);
      END_STATE();
    case 40:
      if (lookahead == 't') ADVANCE(86);
      END_STATE();
    case 41:
      if (lookahead == 'w') ADVANCE(87);
      END_STATE();
    case 42:
      if (lookahead == 'n') ADVANCE(88);
      END_STATE();
    case 43:
      if (lookahead == 'j') ADVANCE(89);
      END_STATE();
    case 44:
      ACCEPT_TOKEN(anon_sym_of);
      END_STATE();
    case 45:
      if (lookahead == 'e') ADVANCE(90);
      END_STATE();
    case 46:
      ACCEPT_TOKEN(anon_sym_or);
      END_STATE();
    case 47:
      if (lookahead == 'i') ADVANCE(91);
      END_STATE();
    case 48:
      if (lookahead == 'c') ADVANCE(92);
      END_STATE();
    case 49:
      if (lookahead == 'g') ADVANCE(93);
      END_STATE();
    case 50:
      if (lookahead == 'r') ADVANCE(94);
      END_STATE();
    case 51:
      if (lookahead == 'e') ADVANCE(95);
      END_STATE();
    case 52:
      ACCEPT_TOKEN(anon_sym_to);
      END_STATE();
    case 53:
      if (lookahead == 'u') ADVANCE(96);
      if (lookahead == 'y') ADVANCE(97);
      END_STATE();
    case 54:
      if (lookahead == 'p') ADVANCE(98);
      END_STATE();
    case 55:
      if (lookahead == 'l') ADVANCE(99);
      END_STATE();
    case 56:
      if (lookahead == 'r') ADVANCE(100);
      END_STATE();
    case 57:
      if (lookahead == 'e') ADVANCE(101);
      if (lookahead == 'i') ADVANCE(102);
      END_STATE();
    case 58:
      if (lookahead == 't') ADVANCE(103);
      END_STATE();
    case 59:
      ACCEPT_TOKEN(anon_sym_asr);
      END_STATE();
    case 60:
      if (lookahead == 'e') ADVANCE(104);
      END_STATE();
    case 61:
      if (lookahead == 'i') ADVANCE(105);
      END_STATE();
    case 62:
      if (lookahead == 's') ADVANCE(106);
      END_STATE();
    case 63:
      if (lookahead == 's') ADVANCE(107);
      END_STATE();
    case 64:
      if (lookahead == 'e') ADVANCE(108);
      END_STATE();
    case 65:
      if (lookahead == 'n') ADVANCE(109);
      END_STATE();
    case 66:
      if (lookahead == 'e') ADVANCE(110);
      END_STATE();
    case 67:
      if (lookahead == 'e') ADVANCE(111);
      END_STATE();
    case 68:
      ACCEPT_TOKEN(anon_sym_end);
      END_STATE();
    case 69:
      if (lookahead == 'e') ADVANCE(112);
      END_STATE();
    case 70:
      if (lookahead == 'e') ADVANCE(113);
      END_STATE();
    case 71:
      if (lookahead == 's') ADVANCE(114);
      END_STATE();
    case 72:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 73:
      ACCEPT_TOKEN(anon_sym_fun);
      if (lookahead == 'c') ADVANCE(115);
      END_STATE();
    case 74:
      if (lookahead == 'l') ADVANCE(116);
      END_STATE();
    case 75:
      if (lookahead == 'e') ADVANCE(117);
      END_STATE();
    case 76:
      if (lookahead == 't') ADVANCE(118);
      END_STATE();
    case 77:
      if (lookahead == 'd') ADVANCE(119);
      END_STATE();
    case 78:
      if (lookahead == 'y') ADVANCE(120);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(anon_sym_lor);
      END_STATE();
    case 80:
      ACCEPT_TOKEN(anon_sym_lsl);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_lsr);
      END_STATE();
    case 82:
      if (lookahead == 'r') ADVANCE(121);
      END_STATE();
    case 83:
      if (lookahead == 'c') ADVANCE(122);
      END_STATE();
    case 84:
      if (lookahead == 'h') ADVANCE(123);
      END_STATE();
    case 85:
      ACCEPT_TOKEN(anon_sym_mod);
      if (lookahead == 'u') ADVANCE(124);
      END_STATE();
    case 86:
      if (lookahead == 'a') ADVANCE(125);
      END_STATE();
    case 87:
      ACCEPT_TOKEN(anon_sym_new);
      END_STATE();
    case 88:
      if (lookahead == 'r') ADVANCE(126);
      END_STATE();
    case 89:
      if (lookahead == 'e') ADVANCE(127);
      END_STATE();
    case 90:
      if (lookahead == 'n') ADVANCE(128);
      END_STATE();
    case 91:
      if (lookahead == 'v') ADVANCE(129);
      END_STATE();
    case 92:
      ACCEPT_TOKEN(anon_sym_rec);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(anon_sym_sig);
      END_STATE();
    case 94:
      if (lookahead == 'u') ADVANCE(130);
      END_STATE();
    case 95:
      if (lookahead == 'n') ADVANCE(131);
      END_STATE();
    case 96:
      if (lookahead == 'e') ADVANCE(132);
      END_STATE();
    case 97:
      ACCEPT_TOKEN(anon_sym_try);
      END_STATE();
    case 98:
      if (lookahead == 'e') ADVANCE(133);
      END_STATE();
    case 99:
      ACCEPT_TOKEN(anon_sym_val);
      END_STATE();
    case 100:
      if (lookahead == 't') ADVANCE(134);
      END_STATE();
    case 101:
      if (lookahead == 'n') ADVANCE(135);
      END_STATE();
    case 102:
      if (lookahead == 'l') ADVANCE(136);
      END_STATE();
    case 103:
      if (lookahead == 'h') ADVANCE(137);
      END_STATE();
    case 104:
      if (lookahead == 'r') ADVANCE(138);
      END_STATE();
    case 105:
      if (lookahead == 'n') ADVANCE(139);
      END_STATE();
    case 106:
      if (lookahead == 's') ADVANCE(140);
      END_STATE();
    case 107:
      if (lookahead == 't') ADVANCE(141);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(anon_sym_done);
      END_STATE();
    case 109:
      if (lookahead == 't') ADVANCE(142);
      END_STATE();
    case 110:
      if (lookahead == 'c') ADVANCE(143);
      END_STATE();
    case 111:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 112:
      if (lookahead == 'p') ADVANCE(144);
      END_STATE();
    case 113:
      if (lookahead == 'r') ADVANCE(145);
      END_STATE();
    case 114:
      if (lookahead == 'e') ADVANCE(146);
      END_STATE();
    case 115:
      if (lookahead == 't') ADVANCE(147);
      END_STATE();
    case 116:
      if (lookahead == 'u') ADVANCE(148);
      END_STATE();
    case 117:
      if (lookahead == 'r') ADVANCE(149);
      END_STATE();
    case 118:
      if (lookahead == 'i') ADVANCE(150);
      END_STATE();
    case 119:
      ACCEPT_TOKEN(anon_sym_land);
      END_STATE();
    case 120:
      ACCEPT_TOKEN(anon_sym_lazy);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(anon_sym_lxor);
      END_STATE();
    case 122:
      if (lookahead == 'h') ADVANCE(151);
      END_STATE();
    case 123:
      if (lookahead == 'o') ADVANCE(152);
      END_STATE();
    case 124:
      if (lookahead == 'l') ADVANCE(153);
      END_STATE();
    case 125:
      if (lookahead == 'b') ADVANCE(154);
      END_STATE();
    case 126:
      if (lookahead == 'e') ADVANCE(155);
      END_STATE();
    case 127:
      if (lookahead == 'c') ADVANCE(156);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_open);
      END_STATE();
    case 129:
      if (lookahead == 'a') ADVANCE(157);
      END_STATE();
    case 130:
      if (lookahead == 'c') ADVANCE(158);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_then);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(anon_sym_true);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_type);
      END_STATE();
    case 134:
      if (lookahead == 'u') ADVANCE(159);
      END_STATE();
    case 135:
      ACCEPT_TOKEN(anon_sym_when);
      END_STATE();
    case 136:
      if (lookahead == 'e') ADVANCE(160);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(anon_sym_with);
      END_STATE();
    case 138:
      if (lookahead == 't') ADVANCE(161);
      END_STATE();
    case 139:
      ACCEPT_TOKEN(anon_sym_begin);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(anon_sym_class);
      END_STATE();
    case 141:
      if (lookahead == 'r') ADVANCE(162);
      END_STATE();
    case 142:
      if (lookahead == 'o') ADVANCE(163);
      END_STATE();
    case 143:
      if (lookahead == 't') ADVANCE(164);
      END_STATE();
    case 144:
      if (lookahead == 't') ADVANCE(165);
      END_STATE();
    case 145:
      if (lookahead == 'n') ADVANCE(166);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(anon_sym_false);
      END_STATE();
    case 147:
      if (lookahead == 'i') ADVANCE(167);
      if (lookahead == 'o') ADVANCE(168);
      END_STATE();
    case 148:
      if (lookahead == 'd') ADVANCE(169);
      END_STATE();
    case 149:
      if (lookahead == 'i') ADVANCE(170);
      END_STATE();
    case 150:
      if (lookahead == 'a') ADVANCE(171);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(anon_sym_match);
      END_STATE();
    case 152:
      if (lookahead == 'd') ADVANCE(172);
      END_STATE();
    case 153:
      if (lookahead == 'e') ADVANCE(173);
      END_STATE();
    case 154:
      if (lookahead == 'l') ADVANCE(174);
      END_STATE();
    case 155:
      if (lookahead == 'c') ADVANCE(175);
      END_STATE();
    case 156:
      if (lookahead == 't') ADVANCE(176);
      END_STATE();
    case 157:
      if (lookahead == 't') ADVANCE(177);
      END_STATE();
    case 158:
      if (lookahead == 't') ADVANCE(178);
      END_STATE();
    case 159:
      if (lookahead == 'a') ADVANCE(179);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_assert);
      END_STATE();
    case 162:
      if (lookahead == 'a') ADVANCE(180);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_downto);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(anon_sym_effect);
      END_STATE();
    case 165:
      if (lookahead == 'i') ADVANCE(181);
      END_STATE();
    case 166:
      if (lookahead == 'a') ADVANCE(182);
      END_STATE();
    case 167:
      if (lookahead == 'o') ADVANCE(183);
      END_STATE();
    case 168:
      if (lookahead == 'r') ADVANCE(184);
      END_STATE();
    case 169:
      if (lookahead == 'e') ADVANCE(185);
      END_STATE();
    case 170:
      if (lookahead == 't') ADVANCE(186);
      END_STATE();
    case 171:
      if (lookahead == 'l') ADVANCE(187);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_method);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_module);
      END_STATE();
    case 174:
      if (lookahead == 'e') ADVANCE(188);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_nonrec);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_object);
      END_STATE();
    case 177:
      if (lookahead == 'e') ADVANCE(189);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_struct);
      END_STATE();
    case 179:
      if (lookahead == 'l') ADVANCE(190);
      END_STATE();
    case 180:
      if (lookahead == 'i') ADVANCE(191);
      END_STATE();
    case 181:
      if (lookahead == 'o') ADVANCE(192);
      END_STATE();
    case 182:
      if (lookahead == 'l') ADVANCE(193);
      END_STATE();
    case 183:
      if (lookahead == 'n') ADVANCE(194);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_functor);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_include);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_inherit);
      END_STATE();
    case 187:
      if (lookahead == 'i') ADVANCE(195);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_mutable);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_private);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_virtual);
      END_STATE();
    case 191:
      if (lookahead == 'n') ADVANCE(196);
      END_STATE();
    case 192:
      if (lookahead == 'n') ADVANCE(197);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_external);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_function);
      END_STATE();
    case 195:
      if (lookahead == 'z') ADVANCE(198);
      END_STATE();
    case 196:
      if (lookahead == 't') ADVANCE(199);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_exception);
      END_STATE();
    case 198:
      if (lookahead == 'e') ADVANCE(200);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_constraint);
      END_STATE();
    case 200:
      if (lookahead == 'r') ADVANCE(201);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_initializer);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 140, .external_lex_state = 2},
  [2] = {.lex_state = 37, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3] = {.lex_state = 31, .external_lex_state = 3, .reserved_word_set_id = 2},
  [4] = {.lex_state = 31, .external_lex_state = 3, .reserved_word_set_id = 3},
  [5] = {.lex_state = 33, .external_lex_state = 3, .reserved_word_set_id = 4},
  [6] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [7] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [8] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [9] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [10] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [11] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [12] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [13] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [14] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [15] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [16] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [17] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [18] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [19] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [20] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [21] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [22] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [23] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [24] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [25] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [26] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [27] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [28] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [29] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [30] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [31] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [32] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [33] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [34] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [35] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 7},
  [36] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 7},
  [37] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 7},
  [38] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 7},
  [39] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 7},
  [40] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 7},
  [41] = {.lex_state = 37, .external_lex_state = 3, .reserved_word_set_id = 8},
  [42] = {.lex_state = 37, .external_lex_state = 3, .reserved_word_set_id = 9},
  [43] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 5},
  [44] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [45] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [46] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [47] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [48] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [49] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [50] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [51] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [52] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [53] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [54] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [55] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [56] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [57] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [58] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [59] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [60] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [61] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [62] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [63] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [64] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [65] = {.lex_state = 36, .external_lex_state = 3, .reserved_word_set_id = 10},
  [66] = {.lex_state = 39, .external_lex_state = 3, .reserved_word_set_id = 11},
  [67] = {.lex_state = 39, .external_lex_state = 3, .reserved_word_set_id = 11},
  [68] = {.lex_state = 37, .external_lex_state = 3, .reserved_word_set_id = 12},
  [69] = {.lex_state = 37, .external_lex_state = 3, .reserved_word_set_id = 12},
  [70] = {.lex_state = 38, .external_lex_state = 3, .reserved_word_set_id = 13},
  [71] = {.lex_state = 37, .external_lex_state = 3, .reserved_word_set_id = 14},
  [72] = {.lex_state = 37, .external_lex_state = 3, .reserved_word_set_id = 14},
  [73] = {.lex_state = 45, .external_lex_state = 3, .reserved_word_set_id = 15},
  [74] = {.lex_state = 45, .external_lex_state = 3, .reserved_word_set_id = 16},
  [75] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 17},
  [76] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 7},
  [77] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [78] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [79] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 17},
  [80] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 17},
  [81] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [82] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [83] = {.lex_state = 34, .external_lex_state = 3, .reserved_word_set_id = 18},
  [84] = {.lex_state = 34, .external_lex_state = 3, .reserved_word_set_id = 18},
  [85] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [86] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 19},
  [87] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 20},
  [88] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 21},
  [89] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 19},
  [90] = {.lex_state = 45, .external_lex_state = 3, .reserved_word_set_id = 6},
  [91] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 22},
  [92] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 23},
  [93] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 24},
  [94] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [95] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [96] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [97] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 6},
  [98] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [99] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 24},
  [100] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [101] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [102] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [103] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 24},
  [104] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 6},
  [105] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [106] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 24},
  [107] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [108] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 24},
  [109] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 6},
  [110] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 24},
  [111] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 24},
  [112] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [113] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [114] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [115] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [116] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [117] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [118] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [119] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [120] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [121] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [122] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [123] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [124] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [125] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [126] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [127] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [128] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [129] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [130] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [131] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [132] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [133] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [134] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [135] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [136] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [137] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [138] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [139] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [140] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [141] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [142] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [143] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [144] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [145] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [146] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [147] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [148] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [149] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [150] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [151] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [152] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [153] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [154] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [155] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [156] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [157] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [158] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [159] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [160] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [161] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [162] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [163] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [164] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [165] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [166] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [167] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [168] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [169] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [170] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [171] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [172] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 6},
  [173] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [174] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 6},
  [175] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [176] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [177] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [178] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [179] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [180] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [181] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [182] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 6},
  [183] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [184] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [185] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [186] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [187] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [188] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 28},
  [189] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [190] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [191] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [192] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [193] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [194] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [195] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [196] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [197] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [198] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 28},
  [199] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [200] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [201] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [202] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [203] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [204] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [205] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [206] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 27},
  [207] = {.lex_state = 30, .external_lex_state = 3, .reserved_word_set_id = 6},
  [208] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 28},
  [209] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [210] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [211] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 28},
  [212] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [213] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [214] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [215] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [216] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [217] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [218] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [219] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [220] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [221] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [222] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [223] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [224] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [225] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [226] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [227] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [228] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [229] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [230] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [231] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [232] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [233] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [234] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [235] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [236] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [237] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [238] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [239] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [240] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [241] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [242] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [243] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [244] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [245] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [246] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [247] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [248] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [249] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [250] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [251] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [252] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [253] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [254] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [255] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [256] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [257] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [258] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [259] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [260] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [261] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [262] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [263] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [264] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [265] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [266] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [267] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [268] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [269] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [270] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [271] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [272] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [273] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [274] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [275] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [276] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [277] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [278] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [279] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [280] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [281] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [282] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [283] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [284] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [285] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [286] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [287] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [288] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [289] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [290] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [291] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [292] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [293] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [294] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [295] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [296] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [297] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [298] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [299] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [300] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [301] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [302] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [303] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [304] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [305] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [306] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [307] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [308] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [309] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [310] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [311] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [312] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [313] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [314] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [315] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [316] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [317] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [318] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [319] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [320] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [321] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [322] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [323] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [324] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [325] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [326] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [327] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [328] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [329] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [330] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [331] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [332] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [333] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [334] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [335] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [336] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [337] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [338] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [339] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [340] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [341] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [342] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [343] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [344] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [345] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [346] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [347] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [348] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [349] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [350] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [351] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [352] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [353] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [354] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [355] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [356] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [357] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [358] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [359] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [360] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [361] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [362] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [363] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [364] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [365] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [366] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [367] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [368] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [369] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [370] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [371] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [372] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [373] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [374] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [375] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [376] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [377] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [378] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [379] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [380] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [381] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [382] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [383] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [384] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [385] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [386] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [387] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [388] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [389] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [390] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [391] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [392] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [393] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [394] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [395] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [396] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [397] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [398] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [399] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [400] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [401] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [402] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [403] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [404] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [405] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [406] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [407] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [408] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [409] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [410] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [411] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [412] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [413] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [414] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [415] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [416] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [417] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [418] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [419] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [420] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [421] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [422] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [423] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [424] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [425] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [426] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [427] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [428] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [429] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [430] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [431] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [432] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [433] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [434] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [435] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [436] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [437] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [438] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [439] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [440] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [441] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [442] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [443] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [444] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [445] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [446] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [447] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [448] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [449] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [450] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [451] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [452] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [453] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [454] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [455] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [456] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [457] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [458] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [459] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [460] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [461] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [462] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [463] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [464] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [465] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [466] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [467] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [468] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [469] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [470] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [471] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [472] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [473] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [474] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [475] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [476] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [477] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [478] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [479] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [480] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [481] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [482] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [483] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [484] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [485] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [486] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [487] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [488] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [489] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [490] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [491] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [492] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [493] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [494] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [495] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [496] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [497] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [498] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [499] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [500] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [501] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [502] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [503] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [504] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [505] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [506] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [507] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [508] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [509] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [510] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [511] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [512] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [513] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [514] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [515] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [516] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [517] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [518] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [519] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [520] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [521] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [522] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [523] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [524] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [525] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [526] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [527] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [528] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [529] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [530] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [531] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [532] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [533] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [534] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [535] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [536] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [537] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [538] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [539] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [540] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [541] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [542] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [543] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [544] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [545] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [546] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [547] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [548] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [549] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [550] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [551] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [552] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [553] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [554] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [555] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [556] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [557] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [558] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [559] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [560] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [561] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [562] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [563] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [564] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [565] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [566] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [567] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [568] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [569] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [570] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [571] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [572] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [573] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [574] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [575] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [576] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [577] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [578] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [579] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [580] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [581] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [582] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [583] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [584] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [585] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [586] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [587] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [588] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [589] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [590] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [591] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [592] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [593] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [594] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [595] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [596] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [597] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [598] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [599] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [600] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [601] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [602] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [603] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [604] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [605] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [606] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [607] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [608] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [609] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [610] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [611] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [612] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [613] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [614] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [615] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [616] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [617] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [618] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [619] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [620] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [621] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [622] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [623] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [624] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [625] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [626] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [627] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [628] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [629] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [630] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [631] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [632] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [633] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [634] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [635] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [636] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [637] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [638] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [639] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [640] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [641] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [642] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [643] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [644] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [645] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [646] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [647] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [648] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [649] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [650] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [651] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [652] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [653] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [654] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [655] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [656] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [657] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [658] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [659] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [660] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [661] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [662] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [663] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [664] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [665] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [666] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [667] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [668] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [669] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [670] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [671] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [672] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [673] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [674] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [675] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [676] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [677] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [678] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [679] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [680] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [681] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [682] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [683] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [684] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [685] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [686] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [687] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [688] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [689] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [690] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [691] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [692] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [693] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [694] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [695] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [696] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [697] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [698] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [699] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [700] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [701] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [702] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [703] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [704] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [705] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [706] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [707] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [708] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [709] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [710] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [711] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [712] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [713] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [714] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [715] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [716] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [717] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [718] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [719] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [720] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [721] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [722] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [723] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [724] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [725] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [726] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [727] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [728] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 26},
  [729] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [730] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 30},
  [731] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 30},
  [732] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 31},
  [733] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [734] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [735] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 31},
  [736] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [737] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [738] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [739] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [740] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [741] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [742] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [743] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [744] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [745] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [746] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [747] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [748] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [749] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [750] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [751] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [752] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [753] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [754] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [755] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [756] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [757] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [758] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 34},
  [759] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [760] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [761] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [762] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [763] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [764] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [765] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 32},
  [766] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [767] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [768] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [769] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [770] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [771] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [772] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [773] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [774] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [775] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [776] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [777] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [778] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [779] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [780] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [781] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [782] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [783] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [784] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [785] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [786] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [787] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [788] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [789] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [790] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [791] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [792] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [793] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [794] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [795] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 37},
  [796] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [797] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [798] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [799] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [800] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [801] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [802] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [803] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [804] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [805] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [806] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [807] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [808] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [809] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [810] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [811] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [812] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [813] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 37},
  [814] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [815] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [816] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [817] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [818] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [819] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [820] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [821] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [822] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [823] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [824] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 37},
  [825] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [826] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [827] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [828] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [829] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [830] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [831] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [832] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [833] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [834] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [835] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [836] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [837] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [838] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [839] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [840] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [841] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [842] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [843] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [844] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [845] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [846] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [847] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [848] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [849] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 39},
  [850] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 17},
  [851] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [852] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 39},
  [853] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 17},
  [854] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 17},
  [855] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [856] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [857] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [858] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [859] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 41},
  [860] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 17},
  [861] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [862] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [863] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [864] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [865] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [866] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [867] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 41},
  [868] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [869] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [870] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 17},
  [871] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [872] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [873] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [874] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [875] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [876] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [877] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 36},
  [878] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [879] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [880] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [881] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [882] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [883] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [884] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [885] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [886] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [887] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [888] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [889] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 42},
  [890] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [891] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [892] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [893] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [894] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [895] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [896] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [897] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [898] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [899] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [900] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [901] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [902] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [903] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [904] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [905] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [906] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [907] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [908] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [909] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [910] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [911] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [912] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [913] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [914] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [915] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [916] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [917] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [918] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [919] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [920] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [921] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [922] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [923] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [924] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [925] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [926] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [927] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [928] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [929] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [930] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [931] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [932] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [933] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [934] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [935] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [936] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [937] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [938] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [939] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [940] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [941] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [942] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [943] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [944] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [945] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [946] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [947] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [948] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [949] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [950] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [951] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [952] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [953] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [954] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [955] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 43},
  [956] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [957] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 43},
  [958] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 43},
  [959] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 43},
  [960] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 17},
  [961] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 43},
  [962] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 42},
  [963] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 43},
  [964] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [965] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [966] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [967] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [968] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 43},
  [969] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [970] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 43},
  [971] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 43},
  [972] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 43},
  [973] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [974] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [975] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [976] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [977] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [978] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [979] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [980] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [981] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [982] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [983] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [984] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [985] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [986] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [987] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [988] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [989] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [990] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [991] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [992] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [993] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [994] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [995] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [996] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [997] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [998] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [999] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1000] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1001] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1002] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1003] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1004] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1005] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1006] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1007] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1008] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1009] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1010] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1011] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1012] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [1013] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1014] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1015] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1016] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1017] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1018] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1019] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1020] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1021] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1022] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1023] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1024] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1025] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1026] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1027] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1028] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1029] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1030] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1031] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1032] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1033] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1034] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1035] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1036] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1037] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1038] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1039] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1040] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1041] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1042] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1043] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1044] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1045] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1046] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1047] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1048] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1049] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1050] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1051] = {.lex_state = 42, .external_lex_state = 3, .reserved_word_set_id = 17},
  [1052] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1053] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [1054] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 17},
  [1055] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1056] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1057] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 44},
  [1058] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1059] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1060] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1061] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1062] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1063] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1064] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1065] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 44},
  [1066] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 45},
  [1067] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1068] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1069] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1070] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1071] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1072] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 44},
  [1073] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 44},
  [1074] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1075] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1076] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1077] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1078] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1079] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1080] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1081] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1082] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1083] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1084] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1085] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1086] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1087] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1088] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1089] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1090] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1091] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1092] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1093] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1094] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1095] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1096] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1097] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1098] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1099] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1100] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1101] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1102] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1103] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1104] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1105] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1106] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1107] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1108] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1109] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1110] = {.lex_state = 34, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1111] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1112] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1113] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1114] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1115] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1116] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1117] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1118] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1119] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1120] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1121] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1122] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1123] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1124] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1125] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1126] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1127] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1128] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1129] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1130] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1131] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1132] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1133] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1134] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1135] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1136] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1137] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1138] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1139] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1140] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1141] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1142] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1143] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 44},
  [1144] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1145] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1146] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1147] = {.lex_state = 34, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1148] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 44},
  [1149] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1150] = {.lex_state = 34, .external_lex_state = 3, .reserved_word_set_id = 18},
  [1151] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1152] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1153] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1154] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1155] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1156] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1157] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1158] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1159] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1160] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1161] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1162] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1163] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1164] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1165] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1166] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1167] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1168] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1169] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1170] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1171] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1172] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1173] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1174] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1175] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1176] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1177] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1178] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1179] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1180] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1181] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1182] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1183] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1184] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1185] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1186] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1187] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1188] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1189] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1190] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1191] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1192] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1193] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1194] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1195] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1196] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1197] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1198] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1199] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1200] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1201] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1202] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1203] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1204] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1205] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1206] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1207] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1208] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1209] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1210] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1211] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1212] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1213] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1214] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1215] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1216] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1217] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1218] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1219] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1220] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1221] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1222] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1223] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1224] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1225] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1226] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1227] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1228] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1229] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1230] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1231] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1232] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1233] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1234] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1235] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1236] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1237] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1238] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1239] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1240] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1241] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1242] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1243] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1244] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1245] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1246] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1247] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1248] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1249] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1250] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1251] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 22},
  [1252] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 22},
  [1253] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1254] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1255] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1256] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1257] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1258] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1259] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1260] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1261] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1262] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1263] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1264] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1265] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1266] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1267] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1268] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1269] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1270] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1271] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1272] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1273] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1274] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 47},
  [1275] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1276] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [1277] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1278] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1279] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 47},
  [1280] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1281] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 47},
  [1282] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1283] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1284] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1285] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1286] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1287] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1288] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1289] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1290] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1291] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1292] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1293] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1294] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1295] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1296] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1297] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1298] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1299] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1300] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1301] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1302] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1303] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1304] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1305] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1306] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1307] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1308] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1309] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1310] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1311] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1312] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1313] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1314] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1315] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1316] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1317] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1318] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1319] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1320] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1321] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1322] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1323] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1324] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1325] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1326] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1327] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1328] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1329] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1330] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1331] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1332] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1333] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1334] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1335] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1336] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1337] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1338] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1339] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1340] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1341] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1342] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1343] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1344] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1345] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1346] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1347] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1348] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1349] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1350] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1351] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1352] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1353] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1354] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1355] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1356] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1357] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1358] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1359] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1360] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1361] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1362] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1363] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1364] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1365] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1366] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1367] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1368] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1369] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1370] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1371] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1372] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1373] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1374] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1375] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1376] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1377] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1378] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1379] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1380] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 22},
  [1381] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1382] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1383] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1384] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1385] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1386] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1387] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1388] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1389] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1390] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1391] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1392] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1393] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1394] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1395] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1396] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1397] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1398] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1399] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1400] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1401] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1402] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1403] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1404] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1405] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1406] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1407] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1408] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1409] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1410] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1411] = {.lex_state = 25, .external_lex_state = 4, .reserved_word_set_id = 50},
  [1412] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 51},
  [1413] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 52},
  [1414] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1415] = {.lex_state = 25, .external_lex_state = 4, .reserved_word_set_id = 50},
  [1416] = {.lex_state = 25, .external_lex_state = 4, .reserved_word_set_id = 50},
  [1417] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1418] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1419] = {.lex_state = 25, .external_lex_state = 4, .reserved_word_set_id = 50},
  [1420] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1421] = {.lex_state = 25, .external_lex_state = 4, .reserved_word_set_id = 50},
  [1422] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1423] = {.lex_state = 25, .external_lex_state = 4, .reserved_word_set_id = 50},
  [1424] = {.lex_state = 25, .external_lex_state = 4, .reserved_word_set_id = 50},
  [1425] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1426] = {.lex_state = 25, .external_lex_state = 4, .reserved_word_set_id = 50},
  [1427] = {.lex_state = 25, .external_lex_state = 4, .reserved_word_set_id = 50},
  [1428] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1429] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1430] = {.lex_state = 25, .external_lex_state = 4, .reserved_word_set_id = 50},
  [1431] = {.lex_state = 25, .external_lex_state = 4, .reserved_word_set_id = 50},
  [1432] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1433] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1434] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1435] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1436] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1437] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1438] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1439] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1440] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1441] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1442] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1443] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1444] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1445] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1446] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1447] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1448] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1449] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1450] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1451] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1452] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1453] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1454] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1455] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1456] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1457] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1458] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1459] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1460] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1461] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1462] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1463] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1464] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1465] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1466] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1467] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1468] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1469] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1470] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1471] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1472] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1473] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1474] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1475] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1476] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1477] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1478] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1479] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1480] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1481] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1482] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1483] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1484] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1485] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1486] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1487] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1488] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1489] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1490] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1491] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1492] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1493] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1494] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1495] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1496] = {.lex_state = 15, .external_lex_state = 2},
  [1497] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1498] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1499] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1500] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1501] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1502] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1503] = {.lex_state = 15, .external_lex_state = 2},
  [1504] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1505] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1506] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1507] = {.lex_state = 15, .external_lex_state = 2},
  [1508] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1509] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1510] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1511] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1512] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1513] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1514] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1515] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1516] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1517] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1518] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1519] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1520] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1521] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1522] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1523] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1524] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1525] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1526] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1527] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1528] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1529] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1530] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1531] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1532] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1533] = {.lex_state = 47, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1534] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1535] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1536] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1537] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1538] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1539] = {.lex_state = 148, .external_lex_state = 3, .reserved_word_set_id = 54},
  [1540] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1541] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1542] = {.lex_state = 47, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1543] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1544] = {.lex_state = 47, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1545] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1546] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1547] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1548] = {.lex_state = 148, .external_lex_state = 3, .reserved_word_set_id = 54},
  [1549] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1550] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1551] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1552] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1553] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1554] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1555] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1556] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1557] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1558] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1559] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1560] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1561] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1562] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1563] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1564] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1565] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1566] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1567] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1568] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1569] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1570] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1571] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 55},
  [1572] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1573] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1574] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1575] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1576] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1577] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1578] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1579] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1580] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1581] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1582] = {.lex_state = 148, .external_lex_state = 3, .reserved_word_set_id = 54},
  [1583] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1584] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1585] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1586] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1587] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1588] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1589] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1590] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1591] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1592] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1593] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1594] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1595] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1596] = {.lex_state = 148, .external_lex_state = 3, .reserved_word_set_id = 54},
  [1597] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1598] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1599] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1600] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1601] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1602] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1603] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1604] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1605] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1606] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1607] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1608] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1609] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1610] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1611] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1612] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1613] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1614] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1615] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1616] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1617] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1618] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1619] = {.lex_state = 148, .external_lex_state = 3, .reserved_word_set_id = 54},
  [1620] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1621] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1622] = {.lex_state = 148, .external_lex_state = 3, .reserved_word_set_id = 54},
  [1623] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1624] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1625] = {.lex_state = 47, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1626] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1627] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1628] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1629] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1630] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1631] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1632] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1633] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1634] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1635] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1636] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1637] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1638] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1639] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1640] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1641] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1642] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1643] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1644] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 55},
  [1645] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1646] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1647] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1648] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1649] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1650] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1651] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1652] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1653] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1654] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1655] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1656] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1657] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1658] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1659] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1660] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1661] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1662] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1663] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 31},
  [1664] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1665] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1666] = {.lex_state = 24, .external_lex_state = 3, .reserved_word_set_id = 56},
  [1667] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1668] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 31},
  [1669] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1670] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1671] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1672] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1673] = {.lex_state = 15, .external_lex_state = 2},
  [1674] = {.lex_state = 15, .external_lex_state = 2},
  [1675] = {.lex_state = 15, .external_lex_state = 2},
  [1676] = {.lex_state = 15, .external_lex_state = 2},
  [1677] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1678] = {.lex_state = 15, .external_lex_state = 2},
  [1679] = {.lex_state = 15, .external_lex_state = 2},
  [1680] = {.lex_state = 15, .external_lex_state = 2},
  [1681] = {.lex_state = 15, .external_lex_state = 2},
  [1682] = {.lex_state = 14, .external_lex_state = 2},
  [1683] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1684] = {.lex_state = 14, .external_lex_state = 2},
  [1685] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1686] = {.lex_state = 15, .external_lex_state = 2},
  [1687] = {.lex_state = 15, .external_lex_state = 2},
  [1688] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1689] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1690] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1691] = {.lex_state = 14, .external_lex_state = 2},
  [1692] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1693] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1694] = {.lex_state = 15, .external_lex_state = 2},
  [1695] = {.lex_state = 15, .external_lex_state = 2},
  [1696] = {.lex_state = 15, .external_lex_state = 2},
  [1697] = {.lex_state = 15, .external_lex_state = 2},
  [1698] = {.lex_state = 15, .external_lex_state = 2},
  [1699] = {.lex_state = 15, .external_lex_state = 2},
  [1700] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 56},
  [1701] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1702] = {.lex_state = 15, .external_lex_state = 2},
  [1703] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1704] = {.lex_state = 14, .external_lex_state = 2},
  [1705] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1706] = {.lex_state = 14, .external_lex_state = 2},
  [1707] = {.lex_state = 14, .external_lex_state = 2},
  [1708] = {.lex_state = 14, .external_lex_state = 2},
  [1709] = {.lex_state = 14, .external_lex_state = 2},
  [1710] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1711] = {.lex_state = 14, .external_lex_state = 2},
  [1712] = {.lex_state = 14, .external_lex_state = 2},
  [1713] = {.lex_state = 14, .external_lex_state = 2},
  [1714] = {.lex_state = 15, .external_lex_state = 2},
  [1715] = {.lex_state = 14, .external_lex_state = 2},
  [1716] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1717] = {.lex_state = 14, .external_lex_state = 2},
  [1718] = {.lex_state = 14, .external_lex_state = 2},
  [1719] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1720] = {.lex_state = 15, .external_lex_state = 2},
  [1721] = {.lex_state = 14, .external_lex_state = 2},
  [1722] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1723] = {.lex_state = 14, .external_lex_state = 2},
  [1724] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1725] = {.lex_state = 14, .external_lex_state = 2},
  [1726] = {.lex_state = 14, .external_lex_state = 2},
  [1727] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1728] = {.lex_state = 15, .external_lex_state = 2},
  [1729] = {.lex_state = 15, .external_lex_state = 2},
  [1730] = {.lex_state = 15, .external_lex_state = 2},
  [1731] = {.lex_state = 15, .external_lex_state = 2},
  [1732] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1733] = {.lex_state = 14, .external_lex_state = 2},
  [1734] = {.lex_state = 14, .external_lex_state = 2},
  [1735] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1736] = {.lex_state = 14, .external_lex_state = 2},
  [1737] = {.lex_state = 14, .external_lex_state = 2},
  [1738] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1739] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1740] = {.lex_state = 14, .external_lex_state = 2},
  [1741] = {.lex_state = 14, .external_lex_state = 2},
  [1742] = {.lex_state = 14, .external_lex_state = 2},
  [1743] = {.lex_state = 14, .external_lex_state = 2},
  [1744] = {.lex_state = 14, .external_lex_state = 2},
  [1745] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1746] = {.lex_state = 14, .external_lex_state = 2},
  [1747] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1748] = {.lex_state = 15, .external_lex_state = 2},
  [1749] = {.lex_state = 14, .external_lex_state = 2},
  [1750] = {.lex_state = 15, .external_lex_state = 2},
  [1751] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1752] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1753] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1754] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1755] = {.lex_state = 14, .external_lex_state = 2},
  [1756] = {.lex_state = 14, .external_lex_state = 2},
  [1757] = {.lex_state = 14, .external_lex_state = 2},
  [1758] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1759] = {.lex_state = 14, .external_lex_state = 2},
  [1760] = {.lex_state = 14, .external_lex_state = 2},
  [1761] = {.lex_state = 14, .external_lex_state = 2},
  [1762] = {.lex_state = 15, .external_lex_state = 2},
  [1763] = {.lex_state = 15, .external_lex_state = 2},
  [1764] = {.lex_state = 15, .external_lex_state = 2},
  [1765] = {.lex_state = 15, .external_lex_state = 2},
  [1766] = {.lex_state = 15, .external_lex_state = 2},
  [1767] = {.lex_state = 14, .external_lex_state = 2},
  [1768] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1769] = {.lex_state = 15, .external_lex_state = 2},
  [1770] = {.lex_state = 14, .external_lex_state = 2},
  [1771] = {.lex_state = 14, .external_lex_state = 2},
  [1772] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1773] = {.lex_state = 15, .external_lex_state = 2},
  [1774] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1775] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1776] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1777] = {.lex_state = 15, .external_lex_state = 2},
  [1778] = {.lex_state = 14, .external_lex_state = 2},
  [1779] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1780] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1781] = {.lex_state = 14, .external_lex_state = 2},
  [1782] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [1783] = {.lex_state = 14, .external_lex_state = 2},
  [1784] = {.lex_state = 14, .external_lex_state = 2},
  [1785] = {.lex_state = 14, .external_lex_state = 2},
  [1786] = {.lex_state = 14, .external_lex_state = 2},
  [1787] = {.lex_state = 14, .external_lex_state = 2},
  [1788] = {.lex_state = 14, .external_lex_state = 2},
  [1789] = {.lex_state = 14, .external_lex_state = 2},
  [1790] = {.lex_state = 15, .external_lex_state = 2},
  [1791] = {.lex_state = 14, .external_lex_state = 2},
  [1792] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1793] = {.lex_state = 14, .external_lex_state = 2},
  [1794] = {.lex_state = 14, .external_lex_state = 2},
  [1795] = {.lex_state = 14, .external_lex_state = 2},
  [1796] = {.lex_state = 14, .external_lex_state = 2},
  [1797] = {.lex_state = 14, .external_lex_state = 2},
  [1798] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1799] = {.lex_state = 14, .external_lex_state = 2},
  [1800] = {.lex_state = 14, .external_lex_state = 2},
  [1801] = {.lex_state = 14, .external_lex_state = 2},
  [1802] = {.lex_state = 14, .external_lex_state = 2},
  [1803] = {.lex_state = 10, .external_lex_state = 2},
  [1804] = {.lex_state = 14, .external_lex_state = 2},
  [1805] = {.lex_state = 11, .external_lex_state = 2},
  [1806] = {.lex_state = 11, .external_lex_state = 2},
  [1807] = {.lex_state = 11, .external_lex_state = 2},
  [1808] = {.lex_state = 10, .external_lex_state = 2},
  [1809] = {.lex_state = 11, .external_lex_state = 2},
  [1810] = {.lex_state = 11, .external_lex_state = 2},
  [1811] = {.lex_state = 11, .external_lex_state = 2},
  [1812] = {.lex_state = 11, .external_lex_state = 2},
  [1813] = {.lex_state = 14, .external_lex_state = 2},
  [1814] = {.lex_state = 11, .external_lex_state = 2},
  [1815] = {.lex_state = 14, .external_lex_state = 2},
  [1816] = {.lex_state = 10, .external_lex_state = 2},
  [1817] = {.lex_state = 11, .external_lex_state = 2},
  [1818] = {.lex_state = 14, .external_lex_state = 2},
  [1819] = {.lex_state = 14, .external_lex_state = 2},
  [1820] = {.lex_state = 11, .external_lex_state = 2},
  [1821] = {.lex_state = 10, .external_lex_state = 2},
  [1822] = {.lex_state = 11, .external_lex_state = 2},
  [1823] = {.lex_state = 11, .external_lex_state = 2},
  [1824] = {.lex_state = 11, .external_lex_state = 2},
  [1825] = {.lex_state = 11, .external_lex_state = 2},
  [1826] = {.lex_state = 10, .external_lex_state = 2},
  [1827] = {.lex_state = 11, .external_lex_state = 2},
  [1828] = {.lex_state = 11, .external_lex_state = 2},
  [1829] = {.lex_state = 11, .external_lex_state = 2},
  [1830] = {.lex_state = 11, .external_lex_state = 2},
  [1831] = {.lex_state = 11, .external_lex_state = 2},
  [1832] = {.lex_state = 11, .external_lex_state = 2},
  [1833] = {.lex_state = 11, .external_lex_state = 2},
  [1834] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1835] = {.lex_state = 11, .external_lex_state = 2},
  [1836] = {.lex_state = 10, .external_lex_state = 2},
  [1837] = {.lex_state = 14, .external_lex_state = 2},
  [1838] = {.lex_state = 10, .external_lex_state = 2},
  [1839] = {.lex_state = 10, .external_lex_state = 2},
  [1840] = {.lex_state = 10, .external_lex_state = 2},
  [1841] = {.lex_state = 10, .external_lex_state = 2},
  [1842] = {.lex_state = 10, .external_lex_state = 2},
  [1843] = {.lex_state = 10, .external_lex_state = 2},
  [1844] = {.lex_state = 14, .external_lex_state = 2},
  [1845] = {.lex_state = 14, .external_lex_state = 2},
  [1846] = {.lex_state = 11, .external_lex_state = 2},
  [1847] = {.lex_state = 11, .external_lex_state = 2},
  [1848] = {.lex_state = 11, .external_lex_state = 2},
  [1849] = {.lex_state = 14, .external_lex_state = 2},
  [1850] = {.lex_state = 15, .external_lex_state = 2},
  [1851] = {.lex_state = 11, .external_lex_state = 2},
  [1852] = {.lex_state = 11, .external_lex_state = 2},
  [1853] = {.lex_state = 11, .external_lex_state = 2},
  [1854] = {.lex_state = 11, .external_lex_state = 2},
  [1855] = {.lex_state = 11, .external_lex_state = 2},
  [1856] = {.lex_state = 11, .external_lex_state = 2},
  [1857] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1858] = {.lex_state = 14, .external_lex_state = 2},
  [1859] = {.lex_state = 15, .external_lex_state = 2},
  [1860] = {.lex_state = 11, .external_lex_state = 2},
  [1861] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1862] = {.lex_state = 11, .external_lex_state = 2},
  [1863] = {.lex_state = 11, .external_lex_state = 2},
  [1864] = {.lex_state = 11, .external_lex_state = 2},
  [1865] = {.lex_state = 10, .external_lex_state = 2},
  [1866] = {.lex_state = 14, .external_lex_state = 2},
  [1867] = {.lex_state = 11, .external_lex_state = 2},
  [1868] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1869] = {.lex_state = 11, .external_lex_state = 2},
  [1870] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1871] = {.lex_state = 11, .external_lex_state = 2},
  [1872] = {.lex_state = 11, .external_lex_state = 2},
  [1873] = {.lex_state = 14, .external_lex_state = 2},
  [1874] = {.lex_state = 11, .external_lex_state = 2},
  [1875] = {.lex_state = 11, .external_lex_state = 2},
  [1876] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1877] = {.lex_state = 11, .external_lex_state = 2},
  [1878] = {.lex_state = 58, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1879] = {.lex_state = 11, .external_lex_state = 2},
  [1880] = {.lex_state = 11, .external_lex_state = 2},
  [1881] = {.lex_state = 11, .external_lex_state = 2},
  [1882] = {.lex_state = 10, .external_lex_state = 2},
  [1883] = {.lex_state = 11, .external_lex_state = 2},
  [1884] = {.lex_state = 11, .external_lex_state = 2},
  [1885] = {.lex_state = 11, .external_lex_state = 2},
  [1886] = {.lex_state = 14, .external_lex_state = 2},
  [1887] = {.lex_state = 11, .external_lex_state = 2},
  [1888] = {.lex_state = 11, .external_lex_state = 2},
  [1889] = {.lex_state = 11, .external_lex_state = 2},
  [1890] = {.lex_state = 11, .external_lex_state = 2},
  [1891] = {.lex_state = 11, .external_lex_state = 2},
  [1892] = {.lex_state = 14, .external_lex_state = 2},
  [1893] = {.lex_state = 11, .external_lex_state = 2},
  [1894] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1895] = {.lex_state = 14, .external_lex_state = 2},
  [1896] = {.lex_state = 11, .external_lex_state = 2},
  [1897] = {.lex_state = 11, .external_lex_state = 2},
  [1898] = {.lex_state = 14, .external_lex_state = 2},
  [1899] = {.lex_state = 11, .external_lex_state = 2},
  [1900] = {.lex_state = 10, .external_lex_state = 2},
  [1901] = {.lex_state = 11, .external_lex_state = 2},
  [1902] = {.lex_state = 15, .external_lex_state = 2},
  [1903] = {.lex_state = 11, .external_lex_state = 2},
  [1904] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1905] = {.lex_state = 11, .external_lex_state = 2},
  [1906] = {.lex_state = 11, .external_lex_state = 2},
  [1907] = {.lex_state = 11, .external_lex_state = 2},
  [1908] = {.lex_state = 11, .external_lex_state = 2},
  [1909] = {.lex_state = 11, .external_lex_state = 2},
  [1910] = {.lex_state = 11, .external_lex_state = 2},
  [1911] = {.lex_state = 11, .external_lex_state = 2},
  [1912] = {.lex_state = 11, .external_lex_state = 2},
  [1913] = {.lex_state = 11, .external_lex_state = 2},
  [1914] = {.lex_state = 11, .external_lex_state = 2},
  [1915] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1916] = {.lex_state = 15, .external_lex_state = 2},
  [1917] = {.lex_state = 14, .external_lex_state = 2},
  [1918] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1919] = {.lex_state = 14, .external_lex_state = 2},
  [1920] = {.lex_state = 15, .external_lex_state = 2},
  [1921] = {.lex_state = 11, .external_lex_state = 2},
  [1922] = {.lex_state = 14, .external_lex_state = 2},
  [1923] = {.lex_state = 11, .external_lex_state = 2},
  [1924] = {.lex_state = 10, .external_lex_state = 2},
  [1925] = {.lex_state = 15, .external_lex_state = 2},
  [1926] = {.lex_state = 11, .external_lex_state = 2},
  [1927] = {.lex_state = 11, .external_lex_state = 2},
  [1928] = {.lex_state = 14, .external_lex_state = 2},
  [1929] = {.lex_state = 14, .external_lex_state = 2},
  [1930] = {.lex_state = 11, .external_lex_state = 2},
  [1931] = {.lex_state = 11, .external_lex_state = 2},
  [1932] = {.lex_state = 11, .external_lex_state = 2},
  [1933] = {.lex_state = 14, .external_lex_state = 2},
  [1934] = {.lex_state = 11, .external_lex_state = 2},
  [1935] = {.lex_state = 11, .external_lex_state = 2},
  [1936] = {.lex_state = 14, .external_lex_state = 2},
  [1937] = {.lex_state = 14, .external_lex_state = 2},
  [1938] = {.lex_state = 14, .external_lex_state = 2},
  [1939] = {.lex_state = 11, .external_lex_state = 2},
  [1940] = {.lex_state = 14, .external_lex_state = 2},
  [1941] = {.lex_state = 10, .external_lex_state = 2},
  [1942] = {.lex_state = 11, .external_lex_state = 2},
  [1943] = {.lex_state = 11, .external_lex_state = 2},
  [1944] = {.lex_state = 11, .external_lex_state = 2},
  [1945] = {.lex_state = 11, .external_lex_state = 2},
  [1946] = {.lex_state = 11, .external_lex_state = 2},
  [1947] = {.lex_state = 14, .external_lex_state = 2},
  [1948] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [1949] = {.lex_state = 10, .external_lex_state = 2},
  [1950] = {.lex_state = 10, .external_lex_state = 2},
  [1951] = {.lex_state = 10, .external_lex_state = 2},
  [1952] = {.lex_state = 10, .external_lex_state = 2},
  [1953] = {.lex_state = 10, .external_lex_state = 2},
  [1954] = {.lex_state = 10, .external_lex_state = 2},
  [1955] = {.lex_state = 10, .external_lex_state = 2},
  [1956] = {.lex_state = 10, .external_lex_state = 2},
  [1957] = {.lex_state = 10, .external_lex_state = 2},
  [1958] = {.lex_state = 10, .external_lex_state = 2},
  [1959] = {.lex_state = 10, .external_lex_state = 2},
  [1960] = {.lex_state = 10, .external_lex_state = 2},
  [1961] = {.lex_state = 10, .external_lex_state = 2},
  [1962] = {.lex_state = 10, .external_lex_state = 2},
  [1963] = {.lex_state = 10, .external_lex_state = 2},
  [1964] = {.lex_state = 10, .external_lex_state = 2},
  [1965] = {.lex_state = 10, .external_lex_state = 2},
  [1966] = {.lex_state = 10, .external_lex_state = 2},
  [1967] = {.lex_state = 10, .external_lex_state = 2},
  [1968] = {.lex_state = 10, .external_lex_state = 2},
  [1969] = {.lex_state = 10, .external_lex_state = 2},
  [1970] = {.lex_state = 10, .external_lex_state = 2},
  [1971] = {.lex_state = 10, .external_lex_state = 2},
  [1972] = {.lex_state = 10, .external_lex_state = 2},
  [1973] = {.lex_state = 10, .external_lex_state = 2},
  [1974] = {.lex_state = 10, .external_lex_state = 2},
  [1975] = {.lex_state = 10, .external_lex_state = 2},
  [1976] = {.lex_state = 10, .external_lex_state = 2},
  [1977] = {.lex_state = 10, .external_lex_state = 2},
  [1978] = {.lex_state = 10, .external_lex_state = 2},
  [1979] = {.lex_state = 10, .external_lex_state = 2},
  [1980] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 57},
  [1981] = {.lex_state = 10, .external_lex_state = 2},
  [1982] = {.lex_state = 10, .external_lex_state = 2},
  [1983] = {.lex_state = 10, .external_lex_state = 2},
  [1984] = {.lex_state = 10, .external_lex_state = 2},
  [1985] = {.lex_state = 10, .external_lex_state = 2},
  [1986] = {.lex_state = 10, .external_lex_state = 2},
  [1987] = {.lex_state = 10, .external_lex_state = 2},
  [1988] = {.lex_state = 10, .external_lex_state = 2},
  [1989] = {.lex_state = 10, .external_lex_state = 2},
  [1990] = {.lex_state = 10, .external_lex_state = 2},
  [1991] = {.lex_state = 10, .external_lex_state = 2},
  [1992] = {.lex_state = 10, .external_lex_state = 2},
  [1993] = {.lex_state = 10, .external_lex_state = 2},
  [1994] = {.lex_state = 10, .external_lex_state = 2},
  [1995] = {.lex_state = 10, .external_lex_state = 2},
  [1996] = {.lex_state = 10, .external_lex_state = 2},
  [1997] = {.lex_state = 10, .external_lex_state = 2},
  [1998] = {.lex_state = 10, .external_lex_state = 2},
  [1999] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2000] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2001] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2002] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2003] = {.lex_state = 141, .external_lex_state = 2},
  [2004] = {.lex_state = 10, .external_lex_state = 2},
  [2005] = {.lex_state = 10, .external_lex_state = 2},
  [2006] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2007] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2008] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2009] = {.lex_state = 10, .external_lex_state = 2},
  [2010] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2011] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2012] = {.lex_state = 10, .external_lex_state = 2},
  [2013] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2014] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2015] = {.lex_state = 10, .external_lex_state = 2},
  [2016] = {.lex_state = 10, .external_lex_state = 2},
  [2017] = {.lex_state = 10, .external_lex_state = 2},
  [2018] = {.lex_state = 10, .external_lex_state = 2},
  [2019] = {.lex_state = 10, .external_lex_state = 2},
  [2020] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2021] = {.lex_state = 10, .external_lex_state = 2},
  [2022] = {.lex_state = 10, .external_lex_state = 2},
  [2023] = {.lex_state = 10, .external_lex_state = 2},
  [2024] = {.lex_state = 10, .external_lex_state = 2},
  [2025] = {.lex_state = 10, .external_lex_state = 2},
  [2026] = {.lex_state = 10, .external_lex_state = 2},
  [2027] = {.lex_state = 10, .external_lex_state = 2},
  [2028] = {.lex_state = 10, .external_lex_state = 2},
  [2029] = {.lex_state = 10, .external_lex_state = 2},
  [2030] = {.lex_state = 10, .external_lex_state = 2},
  [2031] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2032] = {.lex_state = 10, .external_lex_state = 2},
  [2033] = {.lex_state = 10, .external_lex_state = 2},
  [2034] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2035] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2036] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2037] = {.lex_state = 10, .external_lex_state = 2},
  [2038] = {.lex_state = 10, .external_lex_state = 2},
  [2039] = {.lex_state = 10, .external_lex_state = 2},
  [2040] = {.lex_state = 10, .external_lex_state = 2},
  [2041] = {.lex_state = 10, .external_lex_state = 2},
  [2042] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2043] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2044] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2045] = {.lex_state = 141, .external_lex_state = 2},
  [2046] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2047] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2048] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2049] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2050] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2051] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2052] = {.lex_state = 10, .external_lex_state = 2},
  [2053] = {.lex_state = 10, .external_lex_state = 2},
  [2054] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2055] = {.lex_state = 141, .external_lex_state = 2},
  [2056] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2057] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2058] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2059] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2060] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2061] = {.lex_state = 10, .external_lex_state = 2},
  [2062] = {.lex_state = 10, .external_lex_state = 2},
  [2063] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2064] = {.lex_state = 141, .external_lex_state = 2},
  [2065] = {.lex_state = 15, .external_lex_state = 2},
  [2066] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2067] = {.lex_state = 15, .external_lex_state = 2},
  [2068] = {.lex_state = 10, .external_lex_state = 2},
  [2069] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2070] = {.lex_state = 141, .external_lex_state = 2},
  [2071] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2072] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2073] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2074] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2075] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2076] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2077] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2078] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2079] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2080] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2081] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2082] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2083] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2084] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2085] = {.lex_state = 15, .external_lex_state = 2},
  [2086] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2087] = {.lex_state = 141, .external_lex_state = 2},
  [2088] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2089] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2090] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2091] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2092] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2093] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2094] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2095] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2096] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2097] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2098] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2099] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2100] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2101] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2102] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2103] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2104] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2105] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2106] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2107] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2108] = {.lex_state = 141, .external_lex_state = 2},
  [2109] = {.lex_state = 141, .external_lex_state = 2},
  [2110] = {.lex_state = 141, .external_lex_state = 2},
  [2111] = {.lex_state = 141, .external_lex_state = 2},
  [2112] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 33},
  [2113] = {.lex_state = 141, .external_lex_state = 2},
  [2114] = {.lex_state = 141, .external_lex_state = 2},
  [2115] = {.lex_state = 141, .external_lex_state = 2},
  [2116] = {.lex_state = 141, .external_lex_state = 2},
  [2117] = {.lex_state = 141, .external_lex_state = 2},
  [2118] = {.lex_state = 141, .external_lex_state = 2},
  [2119] = {.lex_state = 141, .external_lex_state = 2},
  [2120] = {.lex_state = 10, .external_lex_state = 2},
  [2121] = {.lex_state = 10, .external_lex_state = 2},
  [2122] = {.lex_state = 10, .external_lex_state = 2},
  [2123] = {.lex_state = 10, .external_lex_state = 2},
  [2124] = {.lex_state = 15, .external_lex_state = 2},
  [2125] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 59},
  [2126] = {.lex_state = 15, .external_lex_state = 2},
  [2127] = {.lex_state = 17, .external_lex_state = 2},
  [2128] = {.lex_state = 15, .external_lex_state = 2},
  [2129] = {.lex_state = 15, .external_lex_state = 2},
  [2130] = {.lex_state = 15, .external_lex_state = 2},
  [2131] = {.lex_state = 150, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2132] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2133] = {.lex_state = 15, .external_lex_state = 2},
  [2134] = {.lex_state = 15, .external_lex_state = 2},
  [2135] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 59},
  [2136] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 59},
  [2137] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 59},
  [2138] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2139] = {.lex_state = 20, .external_lex_state = 2},
  [2140] = {.lex_state = 11, .external_lex_state = 2},
  [2141] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2142] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2143] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2144] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2145] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2146] = {.lex_state = 15, .external_lex_state = 2},
  [2147] = {.lex_state = 11, .external_lex_state = 2},
  [2148] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2149] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2150] = {.lex_state = 11, .external_lex_state = 2},
  [2151] = {.lex_state = 11, .external_lex_state = 2},
  [2152] = {.lex_state = 11, .external_lex_state = 2},
  [2153] = {.lex_state = 11, .external_lex_state = 2},
  [2154] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2155] = {.lex_state = 11, .external_lex_state = 2},
  [2156] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2157] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2158] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2159] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2160] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2161] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2162] = {.lex_state = 19, .external_lex_state = 2},
  [2163] = {.lex_state = 141, .external_lex_state = 2},
  [2164] = {.lex_state = 15, .external_lex_state = 2},
  [2165] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2166] = {.lex_state = 11, .external_lex_state = 2},
  [2167] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2168] = {.lex_state = 15, .external_lex_state = 2},
  [2169] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2170] = {.lex_state = 15, .external_lex_state = 2},
  [2171] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2172] = {.lex_state = 11, .external_lex_state = 2},
  [2173] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2174] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2175] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2176] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2177] = {.lex_state = 15, .external_lex_state = 2},
  [2178] = {.lex_state = 15, .external_lex_state = 2},
  [2179] = {.lex_state = 15, .external_lex_state = 2},
  [2180] = {.lex_state = 11, .external_lex_state = 2},
  [2181] = {.lex_state = 11, .external_lex_state = 2},
  [2182] = {.lex_state = 11, .external_lex_state = 2},
  [2183] = {.lex_state = 15, .external_lex_state = 2},
  [2184] = {.lex_state = 15, .external_lex_state = 2},
  [2185] = {.lex_state = 15, .external_lex_state = 2},
  [2186] = {.lex_state = 15, .external_lex_state = 2},
  [2187] = {.lex_state = 15, .external_lex_state = 2},
  [2188] = {.lex_state = 15, .external_lex_state = 2},
  [2189] = {.lex_state = 11, .external_lex_state = 2},
  [2190] = {.lex_state = 11, .external_lex_state = 2},
  [2191] = {.lex_state = 11, .external_lex_state = 2},
  [2192] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2193] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2194] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2195] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2196] = {.lex_state = 11, .external_lex_state = 2},
  [2197] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2198] = {.lex_state = 15, .external_lex_state = 2},
  [2199] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2200] = {.lex_state = 15, .external_lex_state = 2},
  [2201] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2202] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2203] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2204] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2205] = {.lex_state = 15, .external_lex_state = 2},
  [2206] = {.lex_state = 15, .external_lex_state = 2},
  [2207] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2208] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2209] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2210] = {.lex_state = 11, .external_lex_state = 2},
  [2211] = {.lex_state = 15, .external_lex_state = 2},
  [2212] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2213] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2214] = {.lex_state = 15, .external_lex_state = 2},
  [2215] = {.lex_state = 15, .external_lex_state = 2},
  [2216] = {.lex_state = 15, .external_lex_state = 2},
  [2217] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2218] = {.lex_state = 15, .external_lex_state = 2},
  [2219] = {.lex_state = 15, .external_lex_state = 2},
  [2220] = {.lex_state = 15, .external_lex_state = 2},
  [2221] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2222] = {.lex_state = 11, .external_lex_state = 2},
  [2223] = {.lex_state = 11, .external_lex_state = 2},
  [2224] = {.lex_state = 15, .external_lex_state = 2},
  [2225] = {.lex_state = 15, .external_lex_state = 2},
  [2226] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2227] = {.lex_state = 15, .external_lex_state = 2},
  [2228] = {.lex_state = 15, .external_lex_state = 2},
  [2229] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2230] = {.lex_state = 11, .external_lex_state = 2},
  [2231] = {.lex_state = 15, .external_lex_state = 2},
  [2232] = {.lex_state = 15, .external_lex_state = 2},
  [2233] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2234] = {.lex_state = 15, .external_lex_state = 2},
  [2235] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2236] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2237] = {.lex_state = 15, .external_lex_state = 2},
  [2238] = {.lex_state = 15, .external_lex_state = 2},
  [2239] = {.lex_state = 141, .external_lex_state = 2},
  [2240] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2241] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2242] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2243] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2244] = {.lex_state = 15, .external_lex_state = 2},
  [2245] = {.lex_state = 15, .external_lex_state = 2},
  [2246] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2247] = {.lex_state = 15, .external_lex_state = 2},
  [2248] = {.lex_state = 11, .external_lex_state = 2},
  [2249] = {.lex_state = 15, .external_lex_state = 2},
  [2250] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2251] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2252] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2253] = {.lex_state = 11, .external_lex_state = 2},
  [2254] = {.lex_state = 141, .external_lex_state = 2},
  [2255] = {.lex_state = 11, .external_lex_state = 2},
  [2256] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2257] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2258] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2259] = {.lex_state = 11, .external_lex_state = 2},
  [2260] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2261] = {.lex_state = 11, .external_lex_state = 2},
  [2262] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2263] = {.lex_state = 137, .external_lex_state = 3, .reserved_word_set_id = 38},
  [2264] = {.lex_state = 10, .external_lex_state = 2},
  [2265] = {.lex_state = 10, .external_lex_state = 2},
  [2266] = {.lex_state = 10, .external_lex_state = 2},
  [2267] = {.lex_state = 10, .external_lex_state = 2},
  [2268] = {.lex_state = 10, .external_lex_state = 2},
  [2269] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 61},
  [2270] = {.lex_state = 10, .external_lex_state = 2},
  [2271] = {.lex_state = 10, .external_lex_state = 2},
  [2272] = {.lex_state = 10, .external_lex_state = 2},
  [2273] = {.lex_state = 10, .external_lex_state = 2},
  [2274] = {.lex_state = 10, .external_lex_state = 2},
  [2275] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2276] = {.lex_state = 10, .external_lex_state = 2},
  [2277] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2278] = {.lex_state = 10, .external_lex_state = 2},
  [2279] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 61},
  [2280] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2281] = {.lex_state = 10, .external_lex_state = 2},
  [2282] = {.lex_state = 10, .external_lex_state = 2},
  [2283] = {.lex_state = 10, .external_lex_state = 2},
  [2284] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2285] = {.lex_state = 10, .external_lex_state = 2},
  [2286] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2287] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2288] = {.lex_state = 10, .external_lex_state = 2},
  [2289] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 61},
  [2290] = {.lex_state = 10, .external_lex_state = 2},
  [2291] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2292] = {.lex_state = 10, .external_lex_state = 2},
  [2293] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2294] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2295] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2296] = {.lex_state = 10, .external_lex_state = 2},
  [2297] = {.lex_state = 10, .external_lex_state = 2},
  [2298] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 61},
  [2299] = {.lex_state = 10, .external_lex_state = 2},
  [2300] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2301] = {.lex_state = 10, .external_lex_state = 2},
  [2302] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2303] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2304] = {.lex_state = 10, .external_lex_state = 2},
  [2305] = {.lex_state = 10, .external_lex_state = 2},
  [2306] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2307] = {.lex_state = 10, .external_lex_state = 2},
  [2308] = {.lex_state = 10, .external_lex_state = 2},
  [2309] = {.lex_state = 10, .external_lex_state = 2},
  [2310] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2311] = {.lex_state = 10, .external_lex_state = 2},
  [2312] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2313] = {.lex_state = 10, .external_lex_state = 2},
  [2314] = {.lex_state = 10, .external_lex_state = 2},
  [2315] = {.lex_state = 10, .external_lex_state = 2},
  [2316] = {.lex_state = 10, .external_lex_state = 2},
  [2317] = {.lex_state = 10, .external_lex_state = 2},
  [2318] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2319] = {.lex_state = 15, .external_lex_state = 2},
  [2320] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2321] = {.lex_state = 10, .external_lex_state = 2},
  [2322] = {.lex_state = 10, .external_lex_state = 2},
  [2323] = {.lex_state = 10, .external_lex_state = 2},
  [2324] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2325] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2326] = {.lex_state = 10, .external_lex_state = 2},
  [2327] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2328] = {.lex_state = 10, .external_lex_state = 2},
  [2329] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 53},
  [2330] = {.lex_state = 141, .external_lex_state = 2},
  [2331] = {.lex_state = 141, .external_lex_state = 2},
  [2332] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2333] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2334] = {.lex_state = 141, .external_lex_state = 2},
  [2335] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2336] = {.lex_state = 141, .external_lex_state = 2},
  [2337] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2338] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2339] = {.lex_state = 141, .external_lex_state = 2},
  [2340] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2341] = {.lex_state = 10, .external_lex_state = 2},
  [2342] = {.lex_state = 141, .external_lex_state = 2},
  [2343] = {.lex_state = 141, .external_lex_state = 2},
  [2344] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2345] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2346] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2347] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2348] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2349] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2350] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2351] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2352] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2353] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2354] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2355] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2356] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2357] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2358] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2359] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2360] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2361] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2362] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2363] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2364] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2365] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2366] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2367] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2368] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2369] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2370] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2371] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2372] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2373] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2374] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2375] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2376] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2377] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2378] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2379] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2380] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2381] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2382] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2383] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2384] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2385] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2386] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2387] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2388] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2389] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2390] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2391] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2392] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2393] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2394] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2395] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2396] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2397] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2398] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2399] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2400] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2401] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2402] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2403] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2404] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2405] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2406] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2407] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2408] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2409] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2410] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2411] = {.lex_state = 141, .external_lex_state = 2},
  [2412] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2413] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2414] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2415] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2416] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2417] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2418] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2419] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2420] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2421] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2422] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2423] = {.lex_state = 142, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2424] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2425] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2426] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2427] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2428] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2429] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2430] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2431] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2432] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2433] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2434] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2435] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2436] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2437] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2438] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2439] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2440] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2441] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2442] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2443] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2444] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2445] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2446] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2447] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2448] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2449] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2450] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2451] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2452] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2453] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2454] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2455] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2456] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2457] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2458] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2459] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2460] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2461] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2462] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2463] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2464] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2465] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2466] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2467] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2468] = {.lex_state = 23, .external_lex_state = 2},
  [2469] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2470] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2471] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2472] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2473] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2474] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2475] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2476] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2477] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2478] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2479] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2480] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2481] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2482] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2483] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2484] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2485] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2486] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2487] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2488] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2489] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2490] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2491] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2492] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2493] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2494] = {.lex_state = 23, .external_lex_state = 2},
  [2495] = {.lex_state = 146, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2496] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2497] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2498] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2499] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2500] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2501] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2502] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2503] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2504] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2505] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2506] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2507] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2508] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2509] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2510] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2511] = {.lex_state = 23, .external_lex_state = 2},
  [2512] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2513] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2514] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2515] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2516] = {.lex_state = 49, .external_lex_state = 3, .reserved_word_set_id = 40},
  [2517] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2518] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2519] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2520] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2521] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2522] = {.lex_state = 145, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2523] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2524] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2525] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2526] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2527] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2528] = {.lex_state = 10, .external_lex_state = 2},
  [2529] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2530] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2531] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2532] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2533] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2534] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2535] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 65},
  [2536] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2537] = {.lex_state = 10, .external_lex_state = 2},
  [2538] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2539] = {.lex_state = 10, .external_lex_state = 2},
  [2540] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 65},
  [2541] = {.lex_state = 10, .external_lex_state = 2},
  [2542] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 66},
  [2543] = {.lex_state = 10, .external_lex_state = 2},
  [2544] = {.lex_state = 10, .external_lex_state = 2},
  [2545] = {.lex_state = 10, .external_lex_state = 2},
  [2546] = {.lex_state = 10, .external_lex_state = 2},
  [2547] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2548] = {.lex_state = 10, .external_lex_state = 2},
  [2549] = {.lex_state = 10, .external_lex_state = 2},
  [2550] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2551] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2552] = {.lex_state = 10, .external_lex_state = 2},
  [2553] = {.lex_state = 10, .external_lex_state = 2},
  [2554] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2555] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2556] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2557] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2558] = {.lex_state = 10, .external_lex_state = 2},
  [2559] = {.lex_state = 10, .external_lex_state = 2},
  [2560] = {.lex_state = 10, .external_lex_state = 2},
  [2561] = {.lex_state = 10, .external_lex_state = 2},
  [2562] = {.lex_state = 10, .external_lex_state = 2},
  [2563] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2564] = {.lex_state = 10, .external_lex_state = 2},
  [2565] = {.lex_state = 10, .external_lex_state = 2},
  [2566] = {.lex_state = 10, .external_lex_state = 2},
  [2567] = {.lex_state = 10, .external_lex_state = 2},
  [2568] = {.lex_state = 10, .external_lex_state = 2},
  [2569] = {.lex_state = 10, .external_lex_state = 2},
  [2570] = {.lex_state = 10, .external_lex_state = 2},
  [2571] = {.lex_state = 10, .external_lex_state = 2},
  [2572] = {.lex_state = 10, .external_lex_state = 2},
  [2573] = {.lex_state = 10, .external_lex_state = 2},
  [2574] = {.lex_state = 10, .external_lex_state = 2},
  [2575] = {.lex_state = 10, .external_lex_state = 2},
  [2576] = {.lex_state = 10, .external_lex_state = 2},
  [2577] = {.lex_state = 10, .external_lex_state = 2},
  [2578] = {.lex_state = 10, .external_lex_state = 2},
  [2579] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2580] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2581] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2582] = {.lex_state = 10, .external_lex_state = 2},
  [2583] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2584] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2585] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 65},
  [2586] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2587] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2588] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2589] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 66},
  [2590] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2591] = {.lex_state = 10, .external_lex_state = 2},
  [2592] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2593] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2594] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2595] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2596] = {.lex_state = 141, .external_lex_state = 2},
  [2597] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 67},
  [2598] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2599] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2600] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2601] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2602] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2603] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2604] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2605] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2606] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 67},
  [2607] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2608] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2609] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2610] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2611] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 67},
  [2612] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2613] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2614] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2615] = {.lex_state = 141, .external_lex_state = 2},
  [2616] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2617] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 67},
  [2618] = {.lex_state = 141, .external_lex_state = 2},
  [2619] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2620] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2621] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2622] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2623] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2624] = {.lex_state = 141, .external_lex_state = 2},
  [2625] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2626] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2627] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2628] = {.lex_state = 10, .external_lex_state = 2},
  [2629] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2630] = {.lex_state = 10, .external_lex_state = 2},
  [2631] = {.lex_state = 10, .external_lex_state = 2},
  [2632] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 67},
  [2633] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2634] = {.lex_state = 141, .external_lex_state = 2},
  [2635] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2636] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2637] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2638] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2639] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2640] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2641] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2642] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2643] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2644] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2645] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2646] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2647] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 67},
  [2648] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2649] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2650] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2651] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2652] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 68},
  [2653] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 25},
  [2654] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2655] = {.lex_state = 139, .external_lex_state = 2},
  [2656] = {.lex_state = 139, .external_lex_state = 2},
  [2657] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2658] = {.lex_state = 10, .external_lex_state = 2},
  [2659] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2660] = {.lex_state = 139, .external_lex_state = 2},
  [2661] = {.lex_state = 139, .external_lex_state = 2},
  [2662] = {.lex_state = 139, .external_lex_state = 2},
  [2663] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2664] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2665] = {.lex_state = 10, .external_lex_state = 2},
  [2666] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 69},
  [2667] = {.lex_state = 139, .external_lex_state = 2},
  [2668] = {.lex_state = 139, .external_lex_state = 2},
  [2669] = {.lex_state = 10, .external_lex_state = 2},
  [2670] = {.lex_state = 46, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2671] = {.lex_state = 12, .external_lex_state = 2},
  [2672] = {.lex_state = 12, .external_lex_state = 2},
  [2673] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2674] = {.lex_state = 12, .external_lex_state = 2},
  [2675] = {.lex_state = 12, .external_lex_state = 2},
  [2676] = {.lex_state = 12, .external_lex_state = 2},
  [2677] = {.lex_state = 12, .external_lex_state = 2},
  [2678] = {.lex_state = 12, .external_lex_state = 2},
  [2679] = {.lex_state = 12, .external_lex_state = 2},
  [2680] = {.lex_state = 12, .external_lex_state = 2},
  [2681] = {.lex_state = 12, .external_lex_state = 2},
  [2682] = {.lex_state = 12, .external_lex_state = 2},
  [2683] = {.lex_state = 12, .external_lex_state = 2},
  [2684] = {.lex_state = 12, .external_lex_state = 2},
  [2685] = {.lex_state = 12, .external_lex_state = 2},
  [2686] = {.lex_state = 12, .external_lex_state = 2},
  [2687] = {.lex_state = 12, .external_lex_state = 2},
  [2688] = {.lex_state = 12, .external_lex_state = 2},
  [2689] = {.lex_state = 12, .external_lex_state = 2},
  [2690] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 71},
  [2691] = {.lex_state = 12, .external_lex_state = 2},
  [2692] = {.lex_state = 12, .external_lex_state = 2},
  [2693] = {.lex_state = 12, .external_lex_state = 2},
  [2694] = {.lex_state = 12, .external_lex_state = 2},
  [2695] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 71},
  [2696] = {.lex_state = 12, .external_lex_state = 2},
  [2697] = {.lex_state = 12, .external_lex_state = 2},
  [2698] = {.lex_state = 12, .external_lex_state = 2},
  [2699] = {.lex_state = 12, .external_lex_state = 2},
  [2700] = {.lex_state = 12, .external_lex_state = 2},
  [2701] = {.lex_state = 12, .external_lex_state = 2},
  [2702] = {.lex_state = 12, .external_lex_state = 2},
  [2703] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2704] = {.lex_state = 12, .external_lex_state = 2},
  [2705] = {.lex_state = 12, .external_lex_state = 2},
  [2706] = {.lex_state = 12, .external_lex_state = 2},
  [2707] = {.lex_state = 12, .external_lex_state = 2},
  [2708] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 71},
  [2709] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2710] = {.lex_state = 12, .external_lex_state = 2},
  [2711] = {.lex_state = 12, .external_lex_state = 2},
  [2712] = {.lex_state = 12, .external_lex_state = 2},
  [2713] = {.lex_state = 12, .external_lex_state = 2},
  [2714] = {.lex_state = 12, .external_lex_state = 2},
  [2715] = {.lex_state = 12, .external_lex_state = 2},
  [2716] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2717] = {.lex_state = 12, .external_lex_state = 2},
  [2718] = {.lex_state = 12, .external_lex_state = 2},
  [2719] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2720] = {.lex_state = 12, .external_lex_state = 2},
  [2721] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2722] = {.lex_state = 12, .external_lex_state = 2},
  [2723] = {.lex_state = 12, .external_lex_state = 2},
  [2724] = {.lex_state = 12, .external_lex_state = 2},
  [2725] = {.lex_state = 12, .external_lex_state = 2},
  [2726] = {.lex_state = 12, .external_lex_state = 2},
  [2727] = {.lex_state = 12, .external_lex_state = 2},
  [2728] = {.lex_state = 10, .external_lex_state = 2},
  [2729] = {.lex_state = 12, .external_lex_state = 2},
  [2730] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 71},
  [2731] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2732] = {.lex_state = 12, .external_lex_state = 2},
  [2733] = {.lex_state = 12, .external_lex_state = 2},
  [2734] = {.lex_state = 12, .external_lex_state = 2},
  [2735] = {.lex_state = 12, .external_lex_state = 2},
  [2736] = {.lex_state = 12, .external_lex_state = 2},
  [2737] = {.lex_state = 12, .external_lex_state = 2},
  [2738] = {.lex_state = 10, .external_lex_state = 2},
  [2739] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2740] = {.lex_state = 12, .external_lex_state = 2},
  [2741] = {.lex_state = 12, .external_lex_state = 2},
  [2742] = {.lex_state = 10, .external_lex_state = 2},
  [2743] = {.lex_state = 12, .external_lex_state = 2},
  [2744] = {.lex_state = 12, .external_lex_state = 2},
  [2745] = {.lex_state = 12, .external_lex_state = 2},
  [2746] = {.lex_state = 12, .external_lex_state = 2},
  [2747] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2748] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2749] = {.lex_state = 12, .external_lex_state = 2},
  [2750] = {.lex_state = 12, .external_lex_state = 2},
  [2751] = {.lex_state = 12, .external_lex_state = 2},
  [2752] = {.lex_state = 12, .external_lex_state = 2},
  [2753] = {.lex_state = 12, .external_lex_state = 2},
  [2754] = {.lex_state = 12, .external_lex_state = 2},
  [2755] = {.lex_state = 12, .external_lex_state = 2},
  [2756] = {.lex_state = 12, .external_lex_state = 2},
  [2757] = {.lex_state = 12, .external_lex_state = 2},
  [2758] = {.lex_state = 12, .external_lex_state = 2},
  [2759] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2760] = {.lex_state = 12, .external_lex_state = 2},
  [2761] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 71},
  [2762] = {.lex_state = 12, .external_lex_state = 2},
  [2763] = {.lex_state = 12, .external_lex_state = 2},
  [2764] = {.lex_state = 12, .external_lex_state = 2},
  [2765] = {.lex_state = 12, .external_lex_state = 2},
  [2766] = {.lex_state = 10, .external_lex_state = 2},
  [2767] = {.lex_state = 12, .external_lex_state = 2},
  [2768] = {.lex_state = 12, .external_lex_state = 2},
  [2769] = {.lex_state = 12, .external_lex_state = 2},
  [2770] = {.lex_state = 12, .external_lex_state = 2},
  [2771] = {.lex_state = 12, .external_lex_state = 2},
  [2772] = {.lex_state = 12, .external_lex_state = 2},
  [2773] = {.lex_state = 12, .external_lex_state = 2},
  [2774] = {.lex_state = 12, .external_lex_state = 2},
  [2775] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2776] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 71},
  [2777] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2778] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2779] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2780] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2781] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2782] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2783] = {.lex_state = 139, .external_lex_state = 2},
  [2784] = {.lex_state = 139, .external_lex_state = 2},
  [2785] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2786] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2787] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2788] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2789] = {.lex_state = 12, .external_lex_state = 2},
  [2790] = {.lex_state = 12, .external_lex_state = 2},
  [2791] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2792] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2793] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2794] = {.lex_state = 151, .external_lex_state = 2},
  [2795] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2796] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2797] = {.lex_state = 12, .external_lex_state = 2},
  [2798] = {.lex_state = 12, .external_lex_state = 2},
  [2799] = {.lex_state = 12, .external_lex_state = 2},
  [2800] = {.lex_state = 12, .external_lex_state = 2},
  [2801] = {.lex_state = 12, .external_lex_state = 2},
  [2802] = {.lex_state = 12, .external_lex_state = 2},
  [2803] = {.lex_state = 12, .external_lex_state = 2},
  [2804] = {.lex_state = 12, .external_lex_state = 2},
  [2805] = {.lex_state = 12, .external_lex_state = 2},
  [2806] = {.lex_state = 12, .external_lex_state = 2},
  [2807] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2808] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2809] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2810] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2811] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2812] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2813] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2814] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2815] = {.lex_state = 12, .external_lex_state = 2},
  [2816] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2817] = {.lex_state = 12, .external_lex_state = 2},
  [2818] = {.lex_state = 12, .external_lex_state = 2},
  [2819] = {.lex_state = 12, .external_lex_state = 2},
  [2820] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2821] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2822] = {.lex_state = 12, .external_lex_state = 2},
  [2823] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2824] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2825] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2826] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2827] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2828] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2829] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2830] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2831] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2832] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2833] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2834] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2835] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2836] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2837] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2838] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2839] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2840] = {.lex_state = 10, .external_lex_state = 2},
  [2841] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2842] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2843] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2844] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [2845] = {.lex_state = 10, .external_lex_state = 2},
  [2846] = {.lex_state = 23, .external_lex_state = 2},
  [2847] = {.lex_state = 10, .external_lex_state = 2},
  [2848] = {.lex_state = 10, .external_lex_state = 2},
  [2849] = {.lex_state = 10, .external_lex_state = 2},
  [2850] = {.lex_state = 10, .external_lex_state = 2},
  [2851] = {.lex_state = 10, .external_lex_state = 2},
  [2852] = {.lex_state = 10, .external_lex_state = 2},
  [2853] = {.lex_state = 10, .external_lex_state = 2},
  [2854] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2855] = {.lex_state = 10, .external_lex_state = 2},
  [2856] = {.lex_state = 68, .external_lex_state = 2, .reserved_word_set_id = 73},
  [2857] = {.lex_state = 10, .external_lex_state = 2},
  [2858] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [2859] = {.lex_state = 151, .external_lex_state = 2},
  [2860] = {.lex_state = 10, .external_lex_state = 2},
  [2861] = {.lex_state = 10, .external_lex_state = 2},
  [2862] = {.lex_state = 10, .external_lex_state = 2},
  [2863] = {.lex_state = 23, .external_lex_state = 2},
  [2864] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 74},
  [2865] = {.lex_state = 23, .external_lex_state = 2},
  [2866] = {.lex_state = 23, .external_lex_state = 2},
  [2867] = {.lex_state = 23, .external_lex_state = 2},
  [2868] = {.lex_state = 23, .external_lex_state = 2},
  [2869] = {.lex_state = 23, .external_lex_state = 2},
  [2870] = {.lex_state = 23, .external_lex_state = 2},
  [2871] = {.lex_state = 151, .external_lex_state = 2},
  [2872] = {.lex_state = 10, .external_lex_state = 2},
  [2873] = {.lex_state = 23, .external_lex_state = 2},
  [2874] = {.lex_state = 12, .external_lex_state = 2},
  [2875] = {.lex_state = 23, .external_lex_state = 2},
  [2876] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 74},
  [2877] = {.lex_state = 10, .external_lex_state = 2},
  [2878] = {.lex_state = 23, .external_lex_state = 2},
  [2879] = {.lex_state = 23, .external_lex_state = 2},
  [2880] = {.lex_state = 23, .external_lex_state = 2},
  [2881] = {.lex_state = 23, .external_lex_state = 2},
  [2882] = {.lex_state = 23, .external_lex_state = 2},
  [2883] = {.lex_state = 10, .external_lex_state = 2},
  [2884] = {.lex_state = 10, .external_lex_state = 2},
  [2885] = {.lex_state = 10, .external_lex_state = 2},
  [2886] = {.lex_state = 23, .external_lex_state = 2},
  [2887] = {.lex_state = 23, .external_lex_state = 2},
  [2888] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 74},
  [2889] = {.lex_state = 10, .external_lex_state = 2},
  [2890] = {.lex_state = 12, .external_lex_state = 2},
  [2891] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 74},
  [2892] = {.lex_state = 12, .external_lex_state = 2},
  [2893] = {.lex_state = 12, .external_lex_state = 2},
  [2894] = {.lex_state = 10, .external_lex_state = 2},
  [2895] = {.lex_state = 10, .external_lex_state = 2},
  [2896] = {.lex_state = 12, .external_lex_state = 2},
  [2897] = {.lex_state = 143, .external_lex_state = 2},
  [2898] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [2899] = {.lex_state = 12, .external_lex_state = 2},
  [2900] = {.lex_state = 12, .external_lex_state = 2},
  [2901] = {.lex_state = 23, .external_lex_state = 2},
  [2902] = {.lex_state = 12, .external_lex_state = 2},
  [2903] = {.lex_state = 44, .external_lex_state = 2},
  [2904] = {.lex_state = 12, .external_lex_state = 2},
  [2905] = {.lex_state = 12, .external_lex_state = 2},
  [2906] = {.lex_state = 12, .external_lex_state = 2},
  [2907] = {.lex_state = 44, .external_lex_state = 2},
  [2908] = {.lex_state = 10, .external_lex_state = 2},
  [2909] = {.lex_state = 23, .external_lex_state = 2},
  [2910] = {.lex_state = 12, .external_lex_state = 2},
  [2911] = {.lex_state = 10, .external_lex_state = 2},
  [2912] = {.lex_state = 10, .external_lex_state = 2},
  [2913] = {.lex_state = 10, .external_lex_state = 2},
  [2914] = {.lex_state = 10, .external_lex_state = 2},
  [2915] = {.lex_state = 10, .external_lex_state = 2},
  [2916] = {.lex_state = 10, .external_lex_state = 2},
  [2917] = {.lex_state = 12, .external_lex_state = 2},
  [2918] = {.lex_state = 10, .external_lex_state = 2},
  [2919] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [2920] = {.lex_state = 12, .external_lex_state = 2},
  [2921] = {.lex_state = 12, .external_lex_state = 2},
  [2922] = {.lex_state = 44, .external_lex_state = 2},
  [2923] = {.lex_state = 12, .external_lex_state = 2},
  [2924] = {.lex_state = 10, .external_lex_state = 2},
  [2925] = {.lex_state = 12, .external_lex_state = 2},
  [2926] = {.lex_state = 12, .external_lex_state = 2},
  [2927] = {.lex_state = 139, .external_lex_state = 2},
  [2928] = {.lex_state = 23, .external_lex_state = 2},
  [2929] = {.lex_state = 44, .external_lex_state = 2},
  [2930] = {.lex_state = 12, .external_lex_state = 2},
  [2931] = {.lex_state = 139, .external_lex_state = 2},
  [2932] = {.lex_state = 12, .external_lex_state = 2},
  [2933] = {.lex_state = 12, .external_lex_state = 2},
  [2934] = {.lex_state = 10, .external_lex_state = 2},
  [2935] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [2936] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2937] = {.lex_state = 12, .external_lex_state = 2},
  [2938] = {.lex_state = 12, .external_lex_state = 2},
  [2939] = {.lex_state = 10, .external_lex_state = 2},
  [2940] = {.lex_state = 12, .external_lex_state = 2},
  [2941] = {.lex_state = 10, .external_lex_state = 2},
  [2942] = {.lex_state = 10, .external_lex_state = 2},
  [2943] = {.lex_state = 44, .external_lex_state = 2},
  [2944] = {.lex_state = 139, .external_lex_state = 2},
  [2945] = {.lex_state = 12, .external_lex_state = 2},
  [2946] = {.lex_state = 143, .external_lex_state = 2},
  [2947] = {.lex_state = 143, .external_lex_state = 2},
  [2948] = {.lex_state = 59, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2949] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 75},
  [2950] = {.lex_state = 139, .external_lex_state = 2},
  [2951] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 75},
  [2952] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 75},
  [2953] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 75},
  [2954] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 75},
  [2955] = {.lex_state = 12, .external_lex_state = 2},
  [2956] = {.lex_state = 12, .external_lex_state = 2},
  [2957] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 75},
  [2958] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 75},
  [2959] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 75},
  [2960] = {.lex_state = 12, .external_lex_state = 2},
  [2961] = {.lex_state = 12, .external_lex_state = 2},
  [2962] = {.lex_state = 12, .external_lex_state = 2},
  [2963] = {.lex_state = 12, .external_lex_state = 2},
  [2964] = {.lex_state = 143, .external_lex_state = 2},
  [2965] = {.lex_state = 143, .external_lex_state = 2},
  [2966] = {.lex_state = 143, .external_lex_state = 2},
  [2967] = {.lex_state = 12, .external_lex_state = 2},
  [2968] = {.lex_state = 12, .external_lex_state = 2},
  [2969] = {.lex_state = 139, .external_lex_state = 2},
  [2970] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 75},
  [2971] = {.lex_state = 10, .external_lex_state = 2},
  [2972] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2973] = {.lex_state = 140, .external_lex_state = 2},
  [2974] = {.lex_state = 140, .external_lex_state = 2},
  [2975] = {.lex_state = 140, .external_lex_state = 2},
  [2976] = {.lex_state = 140, .external_lex_state = 2},
  [2977] = {.lex_state = 140, .external_lex_state = 2},
  [2978] = {.lex_state = 140, .external_lex_state = 2},
  [2979] = {.lex_state = 140, .external_lex_state = 2},
  [2980] = {.lex_state = 140, .external_lex_state = 2},
  [2981] = {.lex_state = 140, .external_lex_state = 2},
  [2982] = {.lex_state = 140, .external_lex_state = 2},
  [2983] = {.lex_state = 140, .external_lex_state = 2},
  [2984] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [2985] = {.lex_state = 11, .external_lex_state = 2},
  [2986] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2987] = {.lex_state = 11, .external_lex_state = 2},
  [2988] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [2989] = {.lex_state = 11, .external_lex_state = 2},
  [2990] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [2991] = {.lex_state = 139, .external_lex_state = 2},
  [2992] = {.lex_state = 139, .external_lex_state = 2},
  [2993] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2994] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2995] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [2996] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [2997] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [2998] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [2999] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3000] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3001] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3002] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3003] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3004] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3005] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3006] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3007] = {.lex_state = 139, .external_lex_state = 2},
  [3008] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3009] = {.lex_state = 10, .external_lex_state = 2},
  [3010] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3011] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3012] = {.lex_state = 10, .external_lex_state = 2},
  [3013] = {.lex_state = 11, .external_lex_state = 2},
  [3014] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3015] = {.lex_state = 10, .external_lex_state = 2},
  [3016] = {.lex_state = 10, .external_lex_state = 2},
  [3017] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3018] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3019] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3020] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3021] = {.lex_state = 10, .external_lex_state = 2},
  [3022] = {.lex_state = 10, .external_lex_state = 2},
  [3023] = {.lex_state = 10, .external_lex_state = 2},
  [3024] = {.lex_state = 139, .external_lex_state = 2},
  [3025] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3026] = {.lex_state = 139, .external_lex_state = 2},
  [3027] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3028] = {.lex_state = 139, .external_lex_state = 2},
  [3029] = {.lex_state = 10, .external_lex_state = 2},
  [3030] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3031] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3032] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3033] = {.lex_state = 10, .external_lex_state = 2},
  [3034] = {.lex_state = 10, .external_lex_state = 2},
  [3035] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3036] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3037] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3038] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3039] = {.lex_state = 10, .external_lex_state = 2},
  [3040] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3041] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3042] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3043] = {.lex_state = 140, .external_lex_state = 2},
  [3044] = {.lex_state = 11, .external_lex_state = 2},
  [3045] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3046] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3047] = {.lex_state = 10, .external_lex_state = 2},
  [3048] = {.lex_state = 10, .external_lex_state = 2},
  [3049] = {.lex_state = 10, .external_lex_state = 2},
  [3050] = {.lex_state = 10, .external_lex_state = 2},
  [3051] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3052] = {.lex_state = 139, .external_lex_state = 2},
  [3053] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3054] = {.lex_state = 140, .external_lex_state = 2},
  [3055] = {.lex_state = 140, .external_lex_state = 2},
  [3056] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3057] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3058] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3059] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3060] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3061] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3062] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3063] = {.lex_state = 139, .external_lex_state = 2},
  [3064] = {.lex_state = 11, .external_lex_state = 2},
  [3065] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3066] = {.lex_state = 139, .external_lex_state = 2},
  [3067] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3068] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3069] = {.lex_state = 140, .external_lex_state = 2},
  [3070] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3071] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3072] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3073] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3074] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3075] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3076] = {.lex_state = 139, .external_lex_state = 2},
  [3077] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3078] = {.lex_state = 140, .external_lex_state = 2},
  [3079] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3080] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3081] = {.lex_state = 10, .external_lex_state = 2},
  [3082] = {.lex_state = 10, .external_lex_state = 2},
  [3083] = {.lex_state = 10, .external_lex_state = 2},
  [3084] = {.lex_state = 10, .external_lex_state = 2},
  [3085] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3086] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3087] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3088] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3089] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3090] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3091] = {.lex_state = 11, .external_lex_state = 2},
  [3092] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3093] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3094] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3095] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3096] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3097] = {.lex_state = 10, .external_lex_state = 2},
  [3098] = {.lex_state = 140, .external_lex_state = 2},
  [3099] = {.lex_state = 11, .external_lex_state = 2},
  [3100] = {.lex_state = 11, .external_lex_state = 2},
  [3101] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3102] = {.lex_state = 11, .external_lex_state = 2},
  [3103] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3104] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3105] = {.lex_state = 11, .external_lex_state = 2},
  [3106] = {.lex_state = 11, .external_lex_state = 2},
  [3107] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3108] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3109] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3110] = {.lex_state = 139, .external_lex_state = 2},
  [3111] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3112] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3113] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 74},
  [3114] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3115] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3116] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3117] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3118] = {.lex_state = 11, .external_lex_state = 2},
  [3119] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3120] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3121] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3122] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3123] = {.lex_state = 140, .external_lex_state = 2},
  [3124] = {.lex_state = 139, .external_lex_state = 2},
  [3125] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3126] = {.lex_state = 139, .external_lex_state = 2},
  [3127] = {.lex_state = 139, .external_lex_state = 2},
  [3128] = {.lex_state = 138, .external_lex_state = 2},
  [3129] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3130] = {.lex_state = 140, .external_lex_state = 2},
  [3131] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3132] = {.lex_state = 11, .external_lex_state = 2},
  [3133] = {.lex_state = 140, .external_lex_state = 2},
  [3134] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3135] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3136] = {.lex_state = 11, .external_lex_state = 2},
  [3137] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3138] = {.lex_state = 11, .external_lex_state = 2},
  [3139] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3140] = {.lex_state = 11, .external_lex_state = 2},
  [3141] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3142] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3143] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3144] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3145] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3146] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3147] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3148] = {.lex_state = 10, .external_lex_state = 2},
  [3149] = {.lex_state = 141, .external_lex_state = 2},
  [3150] = {.lex_state = 53, .external_lex_state = 3, .reserved_word_set_id = 44},
  [3151] = {.lex_state = 141, .external_lex_state = 2},
  [3152] = {.lex_state = 12, .external_lex_state = 2},
  [3153] = {.lex_state = 53, .external_lex_state = 3, .reserved_word_set_id = 44},
  [3154] = {.lex_state = 141, .external_lex_state = 2},
  [3155] = {.lex_state = 141, .external_lex_state = 2},
  [3156] = {.lex_state = 140, .external_lex_state = 2},
  [3157] = {.lex_state = 138, .external_lex_state = 2},
  [3158] = {.lex_state = 68, .external_lex_state = 2, .reserved_word_set_id = 76},
  [3159] = {.lex_state = 140, .external_lex_state = 2},
  [3160] = {.lex_state = 140, .external_lex_state = 2},
  [3161] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3162] = {.lex_state = 11, .external_lex_state = 2},
  [3163] = {.lex_state = 12, .external_lex_state = 2},
  [3164] = {.lex_state = 10, .external_lex_state = 2},
  [3165] = {.lex_state = 10, .external_lex_state = 2},
  [3166] = {.lex_state = 10, .external_lex_state = 2},
  [3167] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3168] = {.lex_state = 10, .external_lex_state = 2},
  [3169] = {.lex_state = 10, .external_lex_state = 2},
  [3170] = {.lex_state = 10, .external_lex_state = 2},
  [3171] = {.lex_state = 10, .external_lex_state = 2},
  [3172] = {.lex_state = 10, .external_lex_state = 2},
  [3173] = {.lex_state = 10, .external_lex_state = 2},
  [3174] = {.lex_state = 10, .external_lex_state = 2},
  [3175] = {.lex_state = 10, .external_lex_state = 2},
  [3176] = {.lex_state = 12, .external_lex_state = 2},
  [3177] = {.lex_state = 12, .external_lex_state = 2},
  [3178] = {.lex_state = 10, .external_lex_state = 2},
  [3179] = {.lex_state = 140, .external_lex_state = 2},
  [3180] = {.lex_state = 10, .external_lex_state = 2},
  [3181] = {.lex_state = 141, .external_lex_state = 2},
  [3182] = {.lex_state = 10, .external_lex_state = 2},
  [3183] = {.lex_state = 10, .external_lex_state = 2},
  [3184] = {.lex_state = 10, .external_lex_state = 2},
  [3185] = {.lex_state = 12, .external_lex_state = 2},
  [3186] = {.lex_state = 12, .external_lex_state = 2},
  [3187] = {.lex_state = 12, .external_lex_state = 2},
  [3188] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3189] = {.lex_state = 10, .external_lex_state = 2},
  [3190] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 44},
  [3191] = {.lex_state = 140, .external_lex_state = 2},
  [3192] = {.lex_state = 139, .external_lex_state = 2},
  [3193] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3194] = {.lex_state = 140, .external_lex_state = 2},
  [3195] = {.lex_state = 10, .external_lex_state = 2},
  [3196] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3197] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3198] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3199] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 44},
  [3200] = {.lex_state = 10, .external_lex_state = 2},
  [3201] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3202] = {.lex_state = 139, .external_lex_state = 2},
  [3203] = {.lex_state = 10, .external_lex_state = 2},
  [3204] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3205] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3206] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3207] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 44},
  [3208] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3209] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3210] = {.lex_state = 10, .external_lex_state = 2},
  [3211] = {.lex_state = 10, .external_lex_state = 2},
  [3212] = {.lex_state = 10, .external_lex_state = 2},
  [3213] = {.lex_state = 11, .external_lex_state = 2},
  [3214] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3215] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3216] = {.lex_state = 139, .external_lex_state = 2},
  [3217] = {.lex_state = 10, .external_lex_state = 2},
  [3218] = {.lex_state = 10, .external_lex_state = 2},
  [3219] = {.lex_state = 23, .external_lex_state = 2},
  [3220] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3221] = {.lex_state = 10, .external_lex_state = 2},
  [3222] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 44},
  [3223] = {.lex_state = 10, .external_lex_state = 2},
  [3224] = {.lex_state = 139, .external_lex_state = 2},
  [3225] = {.lex_state = 10, .external_lex_state = 2},
  [3226] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3227] = {.lex_state = 10, .external_lex_state = 2},
  [3228] = {.lex_state = 10, .external_lex_state = 2},
  [3229] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3230] = {.lex_state = 12, .external_lex_state = 2},
  [3231] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3232] = {.lex_state = 10, .external_lex_state = 2},
  [3233] = {.lex_state = 139, .external_lex_state = 2},
  [3234] = {.lex_state = 10, .external_lex_state = 2},
  [3235] = {.lex_state = 139, .external_lex_state = 2},
  [3236] = {.lex_state = 139, .external_lex_state = 2},
  [3237] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3238] = {.lex_state = 139, .external_lex_state = 2},
  [3239] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3240] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3241] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3242] = {.lex_state = 23, .external_lex_state = 2},
  [3243] = {.lex_state = 139, .external_lex_state = 2},
  [3244] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3245] = {.lex_state = 10, .external_lex_state = 2},
  [3246] = {.lex_state = 10, .external_lex_state = 2},
  [3247] = {.lex_state = 139, .external_lex_state = 2},
  [3248] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3249] = {.lex_state = 139, .external_lex_state = 2},
  [3250] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3251] = {.lex_state = 139, .external_lex_state = 2},
  [3252] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3253] = {.lex_state = 139, .external_lex_state = 2},
  [3254] = {.lex_state = 139, .external_lex_state = 2},
  [3255] = {.lex_state = 140, .external_lex_state = 2},
  [3256] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3257] = {.lex_state = 139, .external_lex_state = 2},
  [3258] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3259] = {.lex_state = 139, .external_lex_state = 2},
  [3260] = {.lex_state = 139, .external_lex_state = 2},
  [3261] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3262] = {.lex_state = 139, .external_lex_state = 2},
  [3263] = {.lex_state = 140, .external_lex_state = 2},
  [3264] = {.lex_state = 151, .external_lex_state = 2},
  [3265] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3266] = {.lex_state = 10, .external_lex_state = 2},
  [3267] = {.lex_state = 139, .external_lex_state = 2},
  [3268] = {.lex_state = 139, .external_lex_state = 2},
  [3269] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3270] = {.lex_state = 139, .external_lex_state = 2},
  [3271] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3272] = {.lex_state = 10, .external_lex_state = 2},
  [3273] = {.lex_state = 139, .external_lex_state = 2},
  [3274] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3275] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3276] = {.lex_state = 23, .external_lex_state = 2},
  [3277] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3278] = {.lex_state = 140, .external_lex_state = 2},
  [3279] = {.lex_state = 139, .external_lex_state = 2},
  [3280] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3281] = {.lex_state = 139, .external_lex_state = 2},
  [3282] = {.lex_state = 140, .external_lex_state = 2},
  [3283] = {.lex_state = 139, .external_lex_state = 2},
  [3284] = {.lex_state = 10, .external_lex_state = 2},
  [3285] = {.lex_state = 151, .external_lex_state = 2},
  [3286] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3287] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3288] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3289] = {.lex_state = 151, .external_lex_state = 2},
  [3290] = {.lex_state = 140, .external_lex_state = 2},
  [3291] = {.lex_state = 139, .external_lex_state = 2},
  [3292] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3293] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3294] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3295] = {.lex_state = 139, .external_lex_state = 2},
  [3296] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 79},
  [3297] = {.lex_state = 63, .external_lex_state = 2},
  [3298] = {.lex_state = 11, .external_lex_state = 2},
  [3299] = {.lex_state = 10, .external_lex_state = 2},
  [3300] = {.lex_state = 139, .external_lex_state = 2},
  [3301] = {.lex_state = 10, .external_lex_state = 2},
  [3302] = {.lex_state = 139, .external_lex_state = 2},
  [3303] = {.lex_state = 151, .external_lex_state = 2},
  [3304] = {.lex_state = 139, .external_lex_state = 2},
  [3305] = {.lex_state = 143, .external_lex_state = 3, .reserved_word_set_id = 68},
  [3306] = {.lex_state = 139, .external_lex_state = 2},
  [3307] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3308] = {.lex_state = 151, .external_lex_state = 2},
  [3309] = {.lex_state = 139, .external_lex_state = 2},
  [3310] = {.lex_state = 139, .external_lex_state = 2},
  [3311] = {.lex_state = 139, .external_lex_state = 2},
  [3312] = {.lex_state = 140, .external_lex_state = 2},
  [3313] = {.lex_state = 63, .external_lex_state = 2},
  [3314] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 80},
  [3315] = {.lex_state = 143, .external_lex_state = 2},
  [3316] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3317] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3318] = {.lex_state = 11, .external_lex_state = 2},
  [3319] = {.lex_state = 11, .external_lex_state = 2},
  [3320] = {.lex_state = 139, .external_lex_state = 2},
  [3321] = {.lex_state = 10, .external_lex_state = 2},
  [3322] = {.lex_state = 10, .external_lex_state = 2},
  [3323] = {.lex_state = 151, .external_lex_state = 2},
  [3324] = {.lex_state = 139, .external_lex_state = 2},
  [3325] = {.lex_state = 11, .external_lex_state = 2},
  [3326] = {.lex_state = 151, .external_lex_state = 2},
  [3327] = {.lex_state = 11, .external_lex_state = 2},
  [3328] = {.lex_state = 139, .external_lex_state = 2},
  [3329] = {.lex_state = 10, .external_lex_state = 2},
  [3330] = {.lex_state = 139, .external_lex_state = 2},
  [3331] = {.lex_state = 139, .external_lex_state = 2},
  [3332] = {.lex_state = 152, .external_lex_state = 2},
  [3333] = {.lex_state = 63, .external_lex_state = 2},
  [3334] = {.lex_state = 140, .external_lex_state = 2},
  [3335] = {.lex_state = 151, .external_lex_state = 2},
  [3336] = {.lex_state = 63, .external_lex_state = 2},
  [3337] = {.lex_state = 10, .external_lex_state = 2},
  [3338] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3339] = {.lex_state = 10, .external_lex_state = 2},
  [3340] = {.lex_state = 139, .external_lex_state = 2},
  [3341] = {.lex_state = 139, .external_lex_state = 2},
  [3342] = {.lex_state = 10, .external_lex_state = 2},
  [3343] = {.lex_state = 151, .external_lex_state = 2},
  [3344] = {.lex_state = 63, .external_lex_state = 2},
  [3345] = {.lex_state = 151, .external_lex_state = 2},
  [3346] = {.lex_state = 139, .external_lex_state = 2},
  [3347] = {.lex_state = 11, .external_lex_state = 2},
  [3348] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3349] = {.lex_state = 11, .external_lex_state = 2},
  [3350] = {.lex_state = 152, .external_lex_state = 2},
  [3351] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3352] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3353] = {.lex_state = 141, .external_lex_state = 2},
  [3354] = {.lex_state = 140, .external_lex_state = 2},
  [3355] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3356] = {.lex_state = 141, .external_lex_state = 2},
  [3357] = {.lex_state = 141, .external_lex_state = 2},
  [3358] = {.lex_state = 139, .external_lex_state = 2},
  [3359] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3360] = {.lex_state = 139, .external_lex_state = 2},
  [3361] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3362] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3363] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3364] = {.lex_state = 143, .external_lex_state = 2},
  [3365] = {.lex_state = 141, .external_lex_state = 3},
  [3366] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3367] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3368] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3369] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3370] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3371] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3372] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3373] = {.lex_state = 141, .external_lex_state = 2},
  [3374] = {.lex_state = 139, .external_lex_state = 2},
  [3375] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3376] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3377] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 80},
  [3378] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3379] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3380] = {.lex_state = 139, .external_lex_state = 2},
  [3381] = {.lex_state = 49, .external_lex_state = 2, .reserved_word_set_id = 80},
  [3382] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3383] = {.lex_state = 139, .external_lex_state = 2},
  [3384] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3385] = {.lex_state = 151, .external_lex_state = 2},
  [3386] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3387] = {.lex_state = 139, .external_lex_state = 2},
  [3388] = {.lex_state = 139, .external_lex_state = 2},
  [3389] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3390] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 79},
  [3391] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3392] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3393] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3394] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3395] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3396] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3397] = {.lex_state = 139, .external_lex_state = 2},
  [3398] = {.lex_state = 11, .external_lex_state = 2},
  [3399] = {.lex_state = 11, .external_lex_state = 2},
  [3400] = {.lex_state = 151, .external_lex_state = 2},
  [3401] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3402] = {.lex_state = 10, .external_lex_state = 2},
  [3403] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3404] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3405] = {.lex_state = 141, .external_lex_state = 2},
  [3406] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3407] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3408] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3409] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3410] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3411] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3412] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3413] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3414] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3415] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3416] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3417] = {.lex_state = 139, .external_lex_state = 2},
  [3418] = {.lex_state = 139, .external_lex_state = 2},
  [3419] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3420] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3421] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3422] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3423] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3424] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3425] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3426] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3427] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3428] = {.lex_state = 141, .external_lex_state = 2},
  [3429] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3430] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3431] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3432] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3433] = {.lex_state = 139, .external_lex_state = 2},
  [3434] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3435] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3436] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3437] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3438] = {.lex_state = 139, .external_lex_state = 2},
  [3439] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3440] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3441] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3442] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3443] = {.lex_state = 11, .external_lex_state = 2},
  [3444] = {.lex_state = 139, .external_lex_state = 2},
  [3445] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3446] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3447] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3448] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3449] = {.lex_state = 10, .external_lex_state = 2},
  [3450] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3451] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3452] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3453] = {.lex_state = 141, .external_lex_state = 2},
  [3454] = {.lex_state = 11, .external_lex_state = 2},
  [3455] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3456] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3457] = {.lex_state = 141, .external_lex_state = 3},
  [3458] = {.lex_state = 143, .external_lex_state = 2},
  [3459] = {.lex_state = 11, .external_lex_state = 2},
  [3460] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3461] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3462] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3463] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3464] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3465] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3466] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3467] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3468] = {.lex_state = 141, .external_lex_state = 2},
  [3469] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3470] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3471] = {.lex_state = 10, .external_lex_state = 2},
  [3472] = {.lex_state = 139, .external_lex_state = 2},
  [3473] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3474] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3475] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3476] = {.lex_state = 139, .external_lex_state = 2},
  [3477] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 79},
  [3478] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3479] = {.lex_state = 139, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3480] = {.lex_state = 143, .external_lex_state = 2},
  [3481] = {.lex_state = 151, .external_lex_state = 2},
  [3482] = {.lex_state = 139, .external_lex_state = 2},
  [3483] = {.lex_state = 143, .external_lex_state = 2},
  [3484] = {.lex_state = 143, .external_lex_state = 2},
  [3485] = {.lex_state = 143, .external_lex_state = 2},
  [3486] = {.lex_state = 143, .external_lex_state = 2},
  [3487] = {.lex_state = 143, .external_lex_state = 2},
  [3488] = {.lex_state = 143, .external_lex_state = 2},
  [3489] = {.lex_state = 143, .external_lex_state = 2},
  [3490] = {.lex_state = 143, .external_lex_state = 2},
  [3491] = {.lex_state = 10, .external_lex_state = 2},
  [3492] = {.lex_state = 139, .external_lex_state = 2},
  [3493] = {.lex_state = 151, .external_lex_state = 2},
  [3494] = {.lex_state = 143, .external_lex_state = 2},
  [3495] = {.lex_state = 143, .external_lex_state = 2},
  [3496] = {.lex_state = 143, .external_lex_state = 2},
  [3497] = {.lex_state = 143, .external_lex_state = 2},
  [3498] = {.lex_state = 143, .external_lex_state = 2},
  [3499] = {.lex_state = 143, .external_lex_state = 2},
  [3500] = {.lex_state = 143, .external_lex_state = 2},
  [3501] = {.lex_state = 143, .external_lex_state = 2},
  [3502] = {.lex_state = 143, .external_lex_state = 2},
  [3503] = {.lex_state = 139, .external_lex_state = 2},
  [3504] = {.lex_state = 139, .external_lex_state = 2},
  [3505] = {.lex_state = 143, .external_lex_state = 2},
  [3506] = {.lex_state = 143, .external_lex_state = 2},
  [3507] = {.lex_state = 139, .external_lex_state = 2},
  [3508] = {.lex_state = 139, .external_lex_state = 2},
  [3509] = {.lex_state = 138, .external_lex_state = 2},
  [3510] = {.lex_state = 139, .external_lex_state = 2},
  [3511] = {.lex_state = 138, .external_lex_state = 2},
  [3512] = {.lex_state = 143, .external_lex_state = 2},
  [3513] = {.lex_state = 138, .external_lex_state = 2},
  [3514] = {.lex_state = 139, .external_lex_state = 2},
  [3515] = {.lex_state = 139, .external_lex_state = 2},
  [3516] = {.lex_state = 139, .external_lex_state = 2},
  [3517] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3518] = {.lex_state = 139, .external_lex_state = 2},
  [3519] = {.lex_state = 139, .external_lex_state = 2},
  [3520] = {.lex_state = 34, .external_lex_state = 2},
  [3521] = {.lex_state = 139, .external_lex_state = 2},
  [3522] = {.lex_state = 34, .external_lex_state = 2},
  [3523] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3524] = {.lex_state = 34, .external_lex_state = 2},
  [3525] = {.lex_state = 34, .external_lex_state = 2},
  [3526] = {.lex_state = 34, .external_lex_state = 2},
  [3527] = {.lex_state = 34, .external_lex_state = 2},
  [3528] = {.lex_state = 34, .external_lex_state = 2},
  [3529] = {.lex_state = 34, .external_lex_state = 2},
  [3530] = {.lex_state = 34, .external_lex_state = 2},
  [3531] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3532] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3533] = {.lex_state = 151, .external_lex_state = 2},
  [3534] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3535] = {.lex_state = 34, .external_lex_state = 2},
  [3536] = {.lex_state = 34, .external_lex_state = 2},
  [3537] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3538] = {.lex_state = 34, .external_lex_state = 2},
  [3539] = {.lex_state = 34, .external_lex_state = 2},
  [3540] = {.lex_state = 152, .external_lex_state = 2},
  [3541] = {.lex_state = 152, .external_lex_state = 2},
  [3542] = {.lex_state = 151, .external_lex_state = 2},
  [3543] = {.lex_state = 139, .external_lex_state = 2},
  [3544] = {.lex_state = 34, .external_lex_state = 2},
  [3545] = {.lex_state = 139, .external_lex_state = 2},
  [3546] = {.lex_state = 34, .external_lex_state = 2},
  [3547] = {.lex_state = 139, .external_lex_state = 2},
  [3548] = {.lex_state = 34, .external_lex_state = 2},
  [3549] = {.lex_state = 139, .external_lex_state = 2},
  [3550] = {.lex_state = 34, .external_lex_state = 2},
  [3551] = {.lex_state = 139, .external_lex_state = 2},
  [3552] = {.lex_state = 34, .external_lex_state = 2},
  [3553] = {.lex_state = 141, .external_lex_state = 3},
  [3554] = {.lex_state = 143, .external_lex_state = 2},
  [3555] = {.lex_state = 151, .external_lex_state = 2},
  [3556] = {.lex_state = 143, .external_lex_state = 2},
  [3557] = {.lex_state = 151, .external_lex_state = 2},
  [3558] = {.lex_state = 151, .external_lex_state = 2},
  [3559] = {.lex_state = 143, .external_lex_state = 2},
  [3560] = {.lex_state = 143, .external_lex_state = 2},
  [3561] = {.lex_state = 143, .external_lex_state = 2},
  [3562] = {.lex_state = 151, .external_lex_state = 2},
  [3563] = {.lex_state = 143, .external_lex_state = 2},
  [3564] = {.lex_state = 143, .external_lex_state = 2},
  [3565] = {.lex_state = 151, .external_lex_state = 2},
  [3566] = {.lex_state = 143, .external_lex_state = 2},
  [3567] = {.lex_state = 143, .external_lex_state = 2},
  [3568] = {.lex_state = 143, .external_lex_state = 2},
  [3569] = {.lex_state = 143, .external_lex_state = 2},
  [3570] = {.lex_state = 143, .external_lex_state = 2},
  [3571] = {.lex_state = 143, .external_lex_state = 2},
  [3572] = {.lex_state = 151, .external_lex_state = 2},
  [3573] = {.lex_state = 143, .external_lex_state = 2},
  [3574] = {.lex_state = 143, .external_lex_state = 2},
  [3575] = {.lex_state = 143, .external_lex_state = 2},
  [3576] = {.lex_state = 138, .external_lex_state = 2},
  [3577] = {.lex_state = 143, .external_lex_state = 2},
  [3578] = {.lex_state = 140, .external_lex_state = 3},
  [3579] = {.lex_state = 140, .external_lex_state = 3},
  [3580] = {.lex_state = 151, .external_lex_state = 2},
  [3581] = {.lex_state = 143, .external_lex_state = 2},
  [3582] = {.lex_state = 143, .external_lex_state = 2},
  [3583] = {.lex_state = 151, .external_lex_state = 2},
  [3584] = {.lex_state = 138, .external_lex_state = 2},
  [3585] = {.lex_state = 143, .external_lex_state = 2},
  [3586] = {.lex_state = 139, .external_lex_state = 2},
  [3587] = {.lex_state = 143, .external_lex_state = 2},
  [3588] = {.lex_state = 143, .external_lex_state = 2},
  [3589] = {.lex_state = 138, .external_lex_state = 2},
  [3590] = {.lex_state = 143, .external_lex_state = 2},
  [3591] = {.lex_state = 151, .external_lex_state = 2},
  [3592] = {.lex_state = 143, .external_lex_state = 2},
  [3593] = {.lex_state = 139, .external_lex_state = 2},
  [3594] = {.lex_state = 151, .external_lex_state = 2},
  [3595] = {.lex_state = 151, .external_lex_state = 2},
  [3596] = {.lex_state = 143, .external_lex_state = 2},
  [3597] = {.lex_state = 151, .external_lex_state = 2},
  [3598] = {.lex_state = 143, .external_lex_state = 2},
  [3599] = {.lex_state = 143, .external_lex_state = 2},
  [3600] = {.lex_state = 143, .external_lex_state = 2},
  [3601] = {.lex_state = 151, .external_lex_state = 2},
  [3602] = {.lex_state = 143, .external_lex_state = 2},
  [3603] = {.lex_state = 143, .external_lex_state = 2},
  [3604] = {.lex_state = 143, .external_lex_state = 2},
  [3605] = {.lex_state = 138, .external_lex_state = 2},
  [3606] = {.lex_state = 151, .external_lex_state = 2},
  [3607] = {.lex_state = 151, .external_lex_state = 2},
  [3608] = {.lex_state = 143, .external_lex_state = 2},
  [3609] = {.lex_state = 143, .external_lex_state = 2},
  [3610] = {.lex_state = 143, .external_lex_state = 2},
  [3611] = {.lex_state = 138, .external_lex_state = 2},
  [3612] = {.lex_state = 143, .external_lex_state = 2},
  [3613] = {.lex_state = 143, .external_lex_state = 2},
  [3614] = {.lex_state = 143, .external_lex_state = 2},
  [3615] = {.lex_state = 143, .external_lex_state = 2},
  [3616] = {.lex_state = 138, .external_lex_state = 2},
  [3617] = {.lex_state = 143, .external_lex_state = 2},
  [3618] = {.lex_state = 138, .external_lex_state = 2},
  [3619] = {.lex_state = 139, .external_lex_state = 2},
  [3620] = {.lex_state = 143, .external_lex_state = 2},
  [3621] = {.lex_state = 138, .external_lex_state = 2},
  [3622] = {.lex_state = 143, .external_lex_state = 2},
  [3623] = {.lex_state = 143, .external_lex_state = 2},
  [3624] = {.lex_state = 143, .external_lex_state = 2},
  [3625] = {.lex_state = 138, .external_lex_state = 2},
  [3626] = {.lex_state = 138, .external_lex_state = 2},
  [3627] = {.lex_state = 139, .external_lex_state = 2},
  [3628] = {.lex_state = 139, .external_lex_state = 2},
  [3629] = {.lex_state = 139, .external_lex_state = 2},
  [3630] = {.lex_state = 139, .external_lex_state = 2},
  [3631] = {.lex_state = 139, .external_lex_state = 2},
  [3632] = {.lex_state = 139, .external_lex_state = 2},
  [3633] = {.lex_state = 152, .external_lex_state = 2},
  [3634] = {.lex_state = 139, .external_lex_state = 2},
  [3635] = {.lex_state = 139, .external_lex_state = 2},
  [3636] = {.lex_state = 139, .external_lex_state = 2},
  [3637] = {.lex_state = 152, .external_lex_state = 2},
  [3638] = {.lex_state = 139, .external_lex_state = 2},
  [3639] = {.lex_state = 139, .external_lex_state = 2},
  [3640] = {.lex_state = 139, .external_lex_state = 2},
  [3641] = {.lex_state = 139, .external_lex_state = 2},
  [3642] = {.lex_state = 139, .external_lex_state = 2},
  [3643] = {.lex_state = 143, .external_lex_state = 2},
  [3644] = {.lex_state = 139, .external_lex_state = 2},
  [3645] = {.lex_state = 143, .external_lex_state = 2},
  [3646] = {.lex_state = 63, .external_lex_state = 2},
  [3647] = {.lex_state = 143, .external_lex_state = 2},
  [3648] = {.lex_state = 151, .external_lex_state = 2},
  [3649] = {.lex_state = 143, .external_lex_state = 2},
  [3650] = {.lex_state = 151, .external_lex_state = 2},
  [3651] = {.lex_state = 143, .external_lex_state = 2},
  [3652] = {.lex_state = 151, .external_lex_state = 2},
  [3653] = {.lex_state = 143, .external_lex_state = 2},
  [3654] = {.lex_state = 151, .external_lex_state = 2},
  [3655] = {.lex_state = 143, .external_lex_state = 2},
  [3656] = {.lex_state = 143, .external_lex_state = 2},
  [3657] = {.lex_state = 143, .external_lex_state = 2},
  [3658] = {.lex_state = 143, .external_lex_state = 2},
  [3659] = {.lex_state = 143, .external_lex_state = 2},
  [3660] = {.lex_state = 143, .external_lex_state = 2},
  [3661] = {.lex_state = 143, .external_lex_state = 2},
  [3662] = {.lex_state = 143, .external_lex_state = 2},
  [3663] = {.lex_state = 143, .external_lex_state = 2},
  [3664] = {.lex_state = 143, .external_lex_state = 2},
  [3665] = {.lex_state = 143, .external_lex_state = 2},
  [3666] = {.lex_state = 143, .external_lex_state = 2},
  [3667] = {.lex_state = 141, .external_lex_state = 2},
  [3668] = {.lex_state = 138, .external_lex_state = 2},
  [3669] = {.lex_state = 143, .external_lex_state = 2},
  [3670] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3671] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3672] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3673] = {.lex_state = 151, .external_lex_state = 2},
  [3674] = {.lex_state = 143, .external_lex_state = 2},
  [3675] = {.lex_state = 151, .external_lex_state = 2},
  [3676] = {.lex_state = 143, .external_lex_state = 2},
  [3677] = {.lex_state = 143, .external_lex_state = 2},
  [3678] = {.lex_state = 151, .external_lex_state = 2},
  [3679] = {.lex_state = 143, .external_lex_state = 2},
  [3680] = {.lex_state = 151, .external_lex_state = 2},
  [3681] = {.lex_state = 143, .external_lex_state = 2},
  [3682] = {.lex_state = 143, .external_lex_state = 2},
  [3683] = {.lex_state = 143, .external_lex_state = 2},
  [3684] = {.lex_state = 143, .external_lex_state = 2},
  [3685] = {.lex_state = 143, .external_lex_state = 2},
  [3686] = {.lex_state = 143, .external_lex_state = 2},
  [3687] = {.lex_state = 143, .external_lex_state = 2},
  [3688] = {.lex_state = 143, .external_lex_state = 2},
  [3689] = {.lex_state = 143, .external_lex_state = 2},
  [3690] = {.lex_state = 143, .external_lex_state = 2},
  [3691] = {.lex_state = 139, .external_lex_state = 2},
  [3692] = {.lex_state = 143, .external_lex_state = 2},
  [3693] = {.lex_state = 143, .external_lex_state = 2},
  [3694] = {.lex_state = 151, .external_lex_state = 2},
  [3695] = {.lex_state = 143, .external_lex_state = 2},
  [3696] = {.lex_state = 143, .external_lex_state = 2},
  [3697] = {.lex_state = 143, .external_lex_state = 2},
  [3698] = {.lex_state = 143, .external_lex_state = 2},
  [3699] = {.lex_state = 143, .external_lex_state = 2},
  [3700] = {.lex_state = 143, .external_lex_state = 2},
  [3701] = {.lex_state = 143, .external_lex_state = 2},
  [3702] = {.lex_state = 143, .external_lex_state = 2},
  [3703] = {.lex_state = 143, .external_lex_state = 2},
  [3704] = {.lex_state = 143, .external_lex_state = 2},
  [3705] = {.lex_state = 143, .external_lex_state = 2},
  [3706] = {.lex_state = 143, .external_lex_state = 2},
  [3707] = {.lex_state = 143, .external_lex_state = 2},
  [3708] = {.lex_state = 143, .external_lex_state = 2},
  [3709] = {.lex_state = 143, .external_lex_state = 2},
  [3710] = {.lex_state = 143, .external_lex_state = 2},
  [3711] = {.lex_state = 143, .external_lex_state = 2},
  [3712] = {.lex_state = 143, .external_lex_state = 2},
  [3713] = {.lex_state = 143, .external_lex_state = 2},
  [3714] = {.lex_state = 143, .external_lex_state = 2},
  [3715] = {.lex_state = 143, .external_lex_state = 2},
  [3716] = {.lex_state = 143, .external_lex_state = 2},
  [3717] = {.lex_state = 143, .external_lex_state = 2},
  [3718] = {.lex_state = 143, .external_lex_state = 2},
  [3719] = {.lex_state = 152, .external_lex_state = 2},
  [3720] = {.lex_state = 143, .external_lex_state = 2},
  [3721] = {.lex_state = 143, .external_lex_state = 2},
  [3722] = {.lex_state = 143, .external_lex_state = 2},
  [3723] = {.lex_state = 143, .external_lex_state = 2},
  [3724] = {.lex_state = 139, .external_lex_state = 2},
  [3725] = {.lex_state = 152, .external_lex_state = 2},
  [3726] = {.lex_state = 143, .external_lex_state = 2},
  [3727] = {.lex_state = 143, .external_lex_state = 2},
  [3728] = {.lex_state = 139, .external_lex_state = 2},
  [3729] = {.lex_state = 141, .external_lex_state = 2},
  [3730] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3731] = {.lex_state = 34, .external_lex_state = 2},
  [3732] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3733] = {.lex_state = 141, .external_lex_state = 2},
  [3734] = {.lex_state = 141, .external_lex_state = 2},
  [3735] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3736] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3737] = {.lex_state = 141, .external_lex_state = 2},
  [3738] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3739] = {.lex_state = 151, .external_lex_state = 2},
  [3740] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3741] = {.lex_state = 65, .external_lex_state = 2},
  [3742] = {.lex_state = 139, .external_lex_state = 2},
  [3743] = {.lex_state = 141, .external_lex_state = 2},
  [3744] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3745] = {.lex_state = 140, .external_lex_state = 3},
  [3746] = {.lex_state = 65, .external_lex_state = 2},
  [3747] = {.lex_state = 141, .external_lex_state = 2},
  [3748] = {.lex_state = 141, .external_lex_state = 2},
  [3749] = {.lex_state = 140, .external_lex_state = 3},
  [3750] = {.lex_state = 141, .external_lex_state = 2},
  [3751] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3752] = {.lex_state = 34, .external_lex_state = 2},
  [3753] = {.lex_state = 141, .external_lex_state = 2},
  [3754] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3755] = {.lex_state = 143, .external_lex_state = 2},
  [3756] = {.lex_state = 141, .external_lex_state = 2},
  [3757] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3758] = {.lex_state = 143, .external_lex_state = 2},
  [3759] = {.lex_state = 143, .external_lex_state = 2},
  [3760] = {.lex_state = 143, .external_lex_state = 2},
  [3761] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3762] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3763] = {.lex_state = 141, .external_lex_state = 2},
  [3764] = {.lex_state = 141, .external_lex_state = 2},
  [3765] = {.lex_state = 34, .external_lex_state = 2},
  [3766] = {.lex_state = 141, .external_lex_state = 2},
  [3767] = {.lex_state = 141, .external_lex_state = 2},
  [3768] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3769] = {.lex_state = 34, .external_lex_state = 2},
  [3770] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3771] = {.lex_state = 40, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3772] = {.lex_state = 34, .external_lex_state = 2},
  [3773] = {.lex_state = 34, .external_lex_state = 2},
  [3774] = {.lex_state = 141, .external_lex_state = 2},
  [3775] = {.lex_state = 141, .external_lex_state = 2},
  [3776] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3777] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3778] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3779] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3780] = {.lex_state = 141, .external_lex_state = 2},
  [3781] = {.lex_state = 34, .external_lex_state = 2},
  [3782] = {.lex_state = 139, .external_lex_state = 2},
  [3783] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 84},
  [3784] = {.lex_state = 34, .external_lex_state = 2},
  [3785] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3786] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3787] = {.lex_state = 141, .external_lex_state = 2},
  [3788] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 84},
  [3789] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3790] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3791] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3792] = {.lex_state = 34, .external_lex_state = 2},
  [3793] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3794] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3795] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3796] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3797] = {.lex_state = 141, .external_lex_state = 3},
  [3798] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3799] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3800] = {.lex_state = 151, .external_lex_state = 2},
  [3801] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3802] = {.lex_state = 151, .external_lex_state = 2},
  [3803] = {.lex_state = 151, .external_lex_state = 2},
  [3804] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3805] = {.lex_state = 65, .external_lex_state = 2},
  [3806] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3807] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3808] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3809] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3810] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3811] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3812] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3813] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3814] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3815] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3816] = {.lex_state = 34, .external_lex_state = 2},
  [3817] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3818] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3819] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3820] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3821] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3822] = {.lex_state = 151, .external_lex_state = 2},
  [3823] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3824] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 84},
  [3825] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3826] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3827] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 84},
  [3828] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3829] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3830] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3831] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3832] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3833] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3834] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3835] = {.lex_state = 138, .external_lex_state = 2},
  [3836] = {.lex_state = 141, .external_lex_state = 3},
  [3837] = {.lex_state = 141, .external_lex_state = 2},
  [3838] = {.lex_state = 141, .external_lex_state = 3},
  [3839] = {.lex_state = 141, .external_lex_state = 3},
  [3840] = {.lex_state = 25, .external_lex_state = 3, .reserved_word_set_id = 6},
  [3841] = {.lex_state = 151, .external_lex_state = 2},
  [3842] = {.lex_state = 138, .external_lex_state = 2},
  [3843] = {.lex_state = 34, .external_lex_state = 2},
  [3844] = {.lex_state = 138, .external_lex_state = 2},
  [3845] = {.lex_state = 151, .external_lex_state = 2},
  [3846] = {.lex_state = 143, .external_lex_state = 2},
  [3847] = {.lex_state = 151, .external_lex_state = 2},
  [3848] = {.lex_state = 138, .external_lex_state = 2},
  [3849] = {.lex_state = 140, .external_lex_state = 2},
  [3850] = {.lex_state = 138, .external_lex_state = 2},
  [3851] = {.lex_state = 151, .external_lex_state = 2},
  [3852] = {.lex_state = 34, .external_lex_state = 2},
  [3853] = {.lex_state = 151, .external_lex_state = 2},
  [3854] = {.lex_state = 143, .external_lex_state = 2},
  [3855] = {.lex_state = 151, .external_lex_state = 2},
  [3856] = {.lex_state = 34, .external_lex_state = 2},
  [3857] = {.lex_state = 138, .external_lex_state = 2},
  [3858] = {.lex_state = 141, .external_lex_state = 2},
  [3859] = {.lex_state = 140, .external_lex_state = 2},
  [3860] = {.lex_state = 141, .external_lex_state = 2},
  [3861] = {.lex_state = 34, .external_lex_state = 2},
  [3862] = {.lex_state = 140, .external_lex_state = 2},
  [3863] = {.lex_state = 151, .external_lex_state = 2},
  [3864] = {.lex_state = 34, .external_lex_state = 2},
  [3865] = {.lex_state = 34, .external_lex_state = 2},
  [3866] = {.lex_state = 140, .external_lex_state = 2},
  [3867] = {.lex_state = 141, .external_lex_state = 2},
  [3868] = {.lex_state = 143, .external_lex_state = 2},
  [3869] = {.lex_state = 143, .external_lex_state = 2},
  [3870] = {.lex_state = 151, .external_lex_state = 2},
  [3871] = {.lex_state = 151, .external_lex_state = 2},
  [3872] = {.lex_state = 151, .external_lex_state = 2},
  [3873] = {.lex_state = 140, .external_lex_state = 2},
  [3874] = {.lex_state = 34, .external_lex_state = 2},
  [3875] = {.lex_state = 140, .external_lex_state = 2},
  [3876] = {.lex_state = 151, .external_lex_state = 2},
  [3877] = {.lex_state = 151, .external_lex_state = 2},
  [3878] = {.lex_state = 143, .external_lex_state = 2},
  [3879] = {.lex_state = 140, .external_lex_state = 2},
  [3880] = {.lex_state = 138, .external_lex_state = 2},
  [3881] = {.lex_state = 138, .external_lex_state = 2},
  [3882] = {.lex_state = 151, .external_lex_state = 2},
  [3883] = {.lex_state = 140, .external_lex_state = 3},
  [3884] = {.lex_state = 140, .external_lex_state = 2},
  [3885] = {.lex_state = 138, .external_lex_state = 2},
  [3886] = {.lex_state = 151, .external_lex_state = 2},
  [3887] = {.lex_state = 140, .external_lex_state = 2},
  [3888] = {.lex_state = 34, .external_lex_state = 2},
  [3889] = {.lex_state = 140, .external_lex_state = 2},
  [3890] = {.lex_state = 151, .external_lex_state = 2},
  [3891] = {.lex_state = 140, .external_lex_state = 2},
  [3892] = {.lex_state = 140, .external_lex_state = 2},
  [3893] = {.lex_state = 140, .external_lex_state = 2},
  [3894] = {.lex_state = 140, .external_lex_state = 2},
  [3895] = {.lex_state = 140, .external_lex_state = 2},
  [3896] = {.lex_state = 140, .external_lex_state = 2},
  [3897] = {.lex_state = 140, .external_lex_state = 2},
  [3898] = {.lex_state = 140, .external_lex_state = 2},
  [3899] = {.lex_state = 140, .external_lex_state = 2},
  [3900] = {.lex_state = 140, .external_lex_state = 2},
  [3901] = {.lex_state = 140, .external_lex_state = 2},
  [3902] = {.lex_state = 140, .external_lex_state = 2},
  [3903] = {.lex_state = 140, .external_lex_state = 2},
  [3904] = {.lex_state = 140, .external_lex_state = 2},
  [3905] = {.lex_state = 140, .external_lex_state = 2},
  [3906] = {.lex_state = 140, .external_lex_state = 2},
  [3907] = {.lex_state = 140, .external_lex_state = 2},
  [3908] = {.lex_state = 140, .external_lex_state = 2},
  [3909] = {.lex_state = 140, .external_lex_state = 2},
  [3910] = {.lex_state = 140, .external_lex_state = 2},
  [3911] = {.lex_state = 140, .external_lex_state = 2},
  [3912] = {.lex_state = 140, .external_lex_state = 2},
  [3913] = {.lex_state = 140, .external_lex_state = 2},
  [3914] = {.lex_state = 140, .external_lex_state = 2},
  [3915] = {.lex_state = 143, .external_lex_state = 2},
  [3916] = {.lex_state = 143, .external_lex_state = 2},
  [3917] = {.lex_state = 138, .external_lex_state = 2},
  [3918] = {.lex_state = 140, .external_lex_state = 2},
  [3919] = {.lex_state = 141, .external_lex_state = 2},
  [3920] = {.lex_state = 143, .external_lex_state = 2},
  [3921] = {.lex_state = 143, .external_lex_state = 2},
  [3922] = {.lex_state = 143, .external_lex_state = 2},
  [3923] = {.lex_state = 140, .external_lex_state = 2},
  [3924] = {.lex_state = 140, .external_lex_state = 2},
  [3925] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3926] = {.lex_state = 140, .external_lex_state = 2},
  [3927] = {.lex_state = 143, .external_lex_state = 2},
  [3928] = {.lex_state = 140, .external_lex_state = 2},
  [3929] = {.lex_state = 140, .external_lex_state = 2},
  [3930] = {.lex_state = 140, .external_lex_state = 2},
  [3931] = {.lex_state = 140, .external_lex_state = 2},
  [3932] = {.lex_state = 140, .external_lex_state = 2},
  [3933] = {.lex_state = 140, .external_lex_state = 2},
  [3934] = {.lex_state = 140, .external_lex_state = 2},
  [3935] = {.lex_state = 140, .external_lex_state = 2},
  [3936] = {.lex_state = 140, .external_lex_state = 2},
  [3937] = {.lex_state = 140, .external_lex_state = 2},
  [3938] = {.lex_state = 143, .external_lex_state = 2},
  [3939] = {.lex_state = 140, .external_lex_state = 2},
  [3940] = {.lex_state = 143, .external_lex_state = 2},
  [3941] = {.lex_state = 143, .external_lex_state = 2},
  [3942] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3943] = {.lex_state = 140, .external_lex_state = 3},
  [3944] = {.lex_state = 140, .external_lex_state = 2},
  [3945] = {.lex_state = 138, .external_lex_state = 2},
  [3946] = {.lex_state = 143, .external_lex_state = 2},
  [3947] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3948] = {.lex_state = 143, .external_lex_state = 2},
  [3949] = {.lex_state = 140, .external_lex_state = 2},
  [3950] = {.lex_state = 138, .external_lex_state = 2},
  [3951] = {.lex_state = 140, .external_lex_state = 2},
  [3952] = {.lex_state = 143, .external_lex_state = 2},
  [3953] = {.lex_state = 140, .external_lex_state = 2},
  [3954] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3955] = {.lex_state = 140, .external_lex_state = 2},
  [3956] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3957] = {.lex_state = 140, .external_lex_state = 2},
  [3958] = {.lex_state = 140, .external_lex_state = 2},
  [3959] = {.lex_state = 140, .external_lex_state = 2},
  [3960] = {.lex_state = 140, .external_lex_state = 2},
  [3961] = {.lex_state = 140, .external_lex_state = 2},
  [3962] = {.lex_state = 138, .external_lex_state = 2},
  [3963] = {.lex_state = 140, .external_lex_state = 2},
  [3964] = {.lex_state = 140, .external_lex_state = 2},
  [3965] = {.lex_state = 140, .external_lex_state = 2},
  [3966] = {.lex_state = 140, .external_lex_state = 2},
  [3967] = {.lex_state = 140, .external_lex_state = 2},
  [3968] = {.lex_state = 140, .external_lex_state = 2},
  [3969] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3970] = {.lex_state = 140, .external_lex_state = 2},
  [3971] = {.lex_state = 140, .external_lex_state = 2},
  [3972] = {.lex_state = 140, .external_lex_state = 2},
  [3973] = {.lex_state = 140, .external_lex_state = 2},
  [3974] = {.lex_state = 140, .external_lex_state = 2},
  [3975] = {.lex_state = 140, .external_lex_state = 2},
  [3976] = {.lex_state = 140, .external_lex_state = 2},
  [3977] = {.lex_state = 140, .external_lex_state = 2},
  [3978] = {.lex_state = 140, .external_lex_state = 2},
  [3979] = {.lex_state = 140, .external_lex_state = 2},
  [3980] = {.lex_state = 138, .external_lex_state = 2},
  [3981] = {.lex_state = 138, .external_lex_state = 2},
  [3982] = {.lex_state = 138, .external_lex_state = 2},
  [3983] = {.lex_state = 140, .external_lex_state = 2},
  [3984] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3985] = {.lex_state = 138, .external_lex_state = 2},
  [3986] = {.lex_state = 140, .external_lex_state = 2},
  [3987] = {.lex_state = 140, .external_lex_state = 2},
  [3988] = {.lex_state = 141, .external_lex_state = 2},
  [3989] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3990] = {.lex_state = 138, .external_lex_state = 2},
  [3991] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [3992] = {.lex_state = 141, .external_lex_state = 2},
  [3993] = {.lex_state = 140, .external_lex_state = 2},
  [3994] = {.lex_state = 140, .external_lex_state = 2},
  [3995] = {.lex_state = 140, .external_lex_state = 2},
  [3996] = {.lex_state = 138, .external_lex_state = 2},
  [3997] = {.lex_state = 138, .external_lex_state = 2},
  [3998] = {.lex_state = 140, .external_lex_state = 2},
  [3999] = {.lex_state = 138, .external_lex_state = 2},
  [4000] = {.lex_state = 140, .external_lex_state = 2},
  [4001] = {.lex_state = 140, .external_lex_state = 2},
  [4002] = {.lex_state = 140, .external_lex_state = 2},
  [4003] = {.lex_state = 140, .external_lex_state = 2},
  [4004] = {.lex_state = 140, .external_lex_state = 2},
  [4005] = {.lex_state = 143, .external_lex_state = 2},
  [4006] = {.lex_state = 138, .external_lex_state = 2},
  [4007] = {.lex_state = 140, .external_lex_state = 2},
  [4008] = {.lex_state = 140, .external_lex_state = 2},
  [4009] = {.lex_state = 143, .external_lex_state = 2},
  [4010] = {.lex_state = 140, .external_lex_state = 2},
  [4011] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 85},
  [4012] = {.lex_state = 140, .external_lex_state = 2},
  [4013] = {.lex_state = 140, .external_lex_state = 2},
  [4014] = {.lex_state = 140, .external_lex_state = 2},
  [4015] = {.lex_state = 140, .external_lex_state = 2},
  [4016] = {.lex_state = 140, .external_lex_state = 2},
  [4017] = {.lex_state = 140, .external_lex_state = 2},
  [4018] = {.lex_state = 140, .external_lex_state = 2},
  [4019] = {.lex_state = 140, .external_lex_state = 2},
  [4020] = {.lex_state = 140, .external_lex_state = 2},
  [4021] = {.lex_state = 140, .external_lex_state = 2},
  [4022] = {.lex_state = 140, .external_lex_state = 2},
  [4023] = {.lex_state = 140, .external_lex_state = 2},
  [4024] = {.lex_state = 140, .external_lex_state = 2},
  [4025] = {.lex_state = 138, .external_lex_state = 2},
  [4026] = {.lex_state = 140, .external_lex_state = 2},
  [4027] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 85},
  [4028] = {.lex_state = 140, .external_lex_state = 2},
  [4029] = {.lex_state = 140, .external_lex_state = 2},
  [4030] = {.lex_state = 140, .external_lex_state = 2},
  [4031] = {.lex_state = 140, .external_lex_state = 2},
  [4032] = {.lex_state = 140, .external_lex_state = 2},
  [4033] = {.lex_state = 140, .external_lex_state = 2},
  [4034] = {.lex_state = 140, .external_lex_state = 2},
  [4035] = {.lex_state = 140, .external_lex_state = 2},
  [4036] = {.lex_state = 140, .external_lex_state = 2},
  [4037] = {.lex_state = 140, .external_lex_state = 2},
  [4038] = {.lex_state = 143, .external_lex_state = 2},
  [4039] = {.lex_state = 143, .external_lex_state = 2},
  [4040] = {.lex_state = 140, .external_lex_state = 2},
  [4041] = {.lex_state = 140, .external_lex_state = 2},
  [4042] = {.lex_state = 140, .external_lex_state = 2},
  [4043] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 85},
  [4044] = {.lex_state = 143, .external_lex_state = 2},
  [4045] = {.lex_state = 143, .external_lex_state = 2},
  [4046] = {.lex_state = 143, .external_lex_state = 2},
  [4047] = {.lex_state = 143, .external_lex_state = 2},
  [4048] = {.lex_state = 143, .external_lex_state = 2},
  [4049] = {.lex_state = 140, .external_lex_state = 2},
  [4050] = {.lex_state = 143, .external_lex_state = 2},
  [4051] = {.lex_state = 143, .external_lex_state = 2},
  [4052] = {.lex_state = 140, .external_lex_state = 2},
  [4053] = {.lex_state = 140, .external_lex_state = 2},
  [4054] = {.lex_state = 140, .external_lex_state = 2},
  [4055] = {.lex_state = 140, .external_lex_state = 2},
  [4056] = {.lex_state = 140, .external_lex_state = 2},
  [4057] = {.lex_state = 140, .external_lex_state = 2},
  [4058] = {.lex_state = 140, .external_lex_state = 2},
  [4059] = {.lex_state = 138, .external_lex_state = 2},
  [4060] = {.lex_state = 138, .external_lex_state = 2},
  [4061] = {.lex_state = 138, .external_lex_state = 2},
  [4062] = {.lex_state = 141, .external_lex_state = 2},
  [4063] = {.lex_state = 140, .external_lex_state = 2},
  [4064] = {.lex_state = 138, .external_lex_state = 2},
  [4065] = {.lex_state = 140, .external_lex_state = 2},
  [4066] = {.lex_state = 140, .external_lex_state = 2},
  [4067] = {.lex_state = 78, .external_lex_state = 2},
  [4068] = {.lex_state = 138, .external_lex_state = 2},
  [4069] = {.lex_state = 138, .external_lex_state = 2},
  [4070] = {.lex_state = 138, .external_lex_state = 2},
  [4071] = {.lex_state = 78, .external_lex_state = 2},
  [4072] = {.lex_state = 140, .external_lex_state = 2},
  [4073] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4074] = {.lex_state = 78, .external_lex_state = 2},
  [4075] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4076] = {.lex_state = 78, .external_lex_state = 2},
  [4077] = {.lex_state = 141, .external_lex_state = 2},
  [4078] = {.lex_state = 140, .external_lex_state = 2},
  [4079] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4080] = {.lex_state = 138, .external_lex_state = 2},
  [4081] = {.lex_state = 140, .external_lex_state = 2},
  [4082] = {.lex_state = 151, .external_lex_state = 2},
  [4083] = {.lex_state = 64, .external_lex_state = 2},
  [4084] = {.lex_state = 151, .external_lex_state = 2},
  [4085] = {.lex_state = 151, .external_lex_state = 2},
  [4086] = {.lex_state = 141, .external_lex_state = 2},
  [4087] = {.lex_state = 140, .external_lex_state = 2},
  [4088] = {.lex_state = 140, .external_lex_state = 2},
  [4089] = {.lex_state = 141, .external_lex_state = 2},
  [4090] = {.lex_state = 140, .external_lex_state = 2},
  [4091] = {.lex_state = 151, .external_lex_state = 2},
  [4092] = {.lex_state = 140, .external_lex_state = 2},
  [4093] = {.lex_state = 140, .external_lex_state = 2},
  [4094] = {.lex_state = 140, .external_lex_state = 2},
  [4095] = {.lex_state = 64, .external_lex_state = 2},
  [4096] = {.lex_state = 141, .external_lex_state = 2},
  [4097] = {.lex_state = 140, .external_lex_state = 2},
  [4098] = {.lex_state = 140, .external_lex_state = 2},
  [4099] = {.lex_state = 151, .external_lex_state = 2},
  [4100] = {.lex_state = 141, .external_lex_state = 2},
  [4101] = {.lex_state = 140, .external_lex_state = 2},
  [4102] = {.lex_state = 140, .external_lex_state = 2},
  [4103] = {.lex_state = 141, .external_lex_state = 2},
  [4104] = {.lex_state = 140, .external_lex_state = 2},
  [4105] = {.lex_state = 140, .external_lex_state = 2},
  [4106] = {.lex_state = 140, .external_lex_state = 2},
  [4107] = {.lex_state = 151, .external_lex_state = 2},
  [4108] = {.lex_state = 140, .external_lex_state = 2},
  [4109] = {.lex_state = 64, .external_lex_state = 2},
  [4110] = {.lex_state = 140, .external_lex_state = 2},
  [4111] = {.lex_state = 140, .external_lex_state = 2},
  [4112] = {.lex_state = 140, .external_lex_state = 2},
  [4113] = {.lex_state = 140, .external_lex_state = 2},
  [4114] = {.lex_state = 141, .external_lex_state = 2},
  [4115] = {.lex_state = 141, .external_lex_state = 2},
  [4116] = {.lex_state = 151, .external_lex_state = 2},
  [4117] = {.lex_state = 151, .external_lex_state = 2},
  [4118] = {.lex_state = 141, .external_lex_state = 2},
  [4119] = {.lex_state = 140, .external_lex_state = 2},
  [4120] = {.lex_state = 151, .external_lex_state = 2},
  [4121] = {.lex_state = 64, .external_lex_state = 2},
  [4122] = {.lex_state = 140, .external_lex_state = 2},
  [4123] = {.lex_state = 151, .external_lex_state = 2},
  [4124] = {.lex_state = 151, .external_lex_state = 2},
  [4125] = {.lex_state = 140, .external_lex_state = 2},
  [4126] = {.lex_state = 140, .external_lex_state = 2},
  [4127] = {.lex_state = 141, .external_lex_state = 2},
  [4128] = {.lex_state = 140, .external_lex_state = 2},
  [4129] = {.lex_state = 140, .external_lex_state = 2},
  [4130] = {.lex_state = 151, .external_lex_state = 2},
  [4131] = {.lex_state = 141, .external_lex_state = 2},
  [4132] = {.lex_state = 140, .external_lex_state = 2},
  [4133] = {.lex_state = 140, .external_lex_state = 2},
  [4134] = {.lex_state = 140, .external_lex_state = 2},
  [4135] = {.lex_state = 140, .external_lex_state = 2},
  [4136] = {.lex_state = 140, .external_lex_state = 2},
  [4137] = {.lex_state = 140, .external_lex_state = 2},
  [4138] = {.lex_state = 140, .external_lex_state = 2},
  [4139] = {.lex_state = 141, .external_lex_state = 2},
  [4140] = {.lex_state = 140, .external_lex_state = 2},
  [4141] = {.lex_state = 141, .external_lex_state = 2},
  [4142] = {.lex_state = 151, .external_lex_state = 2},
  [4143] = {.lex_state = 140, .external_lex_state = 2},
  [4144] = {.lex_state = 141, .external_lex_state = 2},
  [4145] = {.lex_state = 140, .external_lex_state = 2},
  [4146] = {.lex_state = 140, .external_lex_state = 2},
  [4147] = {.lex_state = 140, .external_lex_state = 2},
  [4148] = {.lex_state = 151, .external_lex_state = 2},
  [4149] = {.lex_state = 140, .external_lex_state = 2},
  [4150] = {.lex_state = 140, .external_lex_state = 2},
  [4151] = {.lex_state = 140, .external_lex_state = 2},
  [4152] = {.lex_state = 151, .external_lex_state = 2},
  [4153] = {.lex_state = 140, .external_lex_state = 2},
  [4154] = {.lex_state = 140, .external_lex_state = 2},
  [4155] = {.lex_state = 141, .external_lex_state = 2},
  [4156] = {.lex_state = 141, .external_lex_state = 2},
  [4157] = {.lex_state = 140, .external_lex_state = 2},
  [4158] = {.lex_state = 140, .external_lex_state = 2},
  [4159] = {.lex_state = 140, .external_lex_state = 2},
  [4160] = {.lex_state = 151, .external_lex_state = 2},
  [4161] = {.lex_state = 140, .external_lex_state = 2},
  [4162] = {.lex_state = 151, .external_lex_state = 2},
  [4163] = {.lex_state = 151, .external_lex_state = 2},
  [4164] = {.lex_state = 141, .external_lex_state = 2},
  [4165] = {.lex_state = 151, .external_lex_state = 2},
  [4166] = {.lex_state = 62, .external_lex_state = 5},
  [4167] = {.lex_state = 62, .external_lex_state = 5},
  [4168] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4169] = {.lex_state = 62, .external_lex_state = 5},
  [4170] = {.lex_state = 62, .external_lex_state = 5},
  [4171] = {.lex_state = 140, .external_lex_state = 2},
  [4172] = {.lex_state = 151, .external_lex_state = 2},
  [4173] = {.lex_state = 62, .external_lex_state = 5},
  [4174] = {.lex_state = 62, .external_lex_state = 5},
  [4175] = {.lex_state = 62, .external_lex_state = 5},
  [4176] = {.lex_state = 62, .external_lex_state = 5},
  [4177] = {.lex_state = 63, .external_lex_state = 2},
  [4178] = {.lex_state = 62, .external_lex_state = 5},
  [4179] = {.lex_state = 62, .external_lex_state = 5},
  [4180] = {.lex_state = 62, .external_lex_state = 5},
  [4181] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4182] = {.lex_state = 62, .external_lex_state = 5},
  [4183] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4184] = {.lex_state = 62, .external_lex_state = 5},
  [4185] = {.lex_state = 62, .external_lex_state = 5},
  [4186] = {.lex_state = 62, .external_lex_state = 5},
  [4187] = {.lex_state = 62, .external_lex_state = 5},
  [4188] = {.lex_state = 62, .external_lex_state = 5},
  [4189] = {.lex_state = 62, .external_lex_state = 5},
  [4190] = {.lex_state = 63, .external_lex_state = 2},
  [4191] = {.lex_state = 151, .external_lex_state = 2},
  [4192] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4193] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4194] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4195] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 67},
  [4196] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 86},
  [4197] = {.lex_state = 63, .external_lex_state = 2},
  [4198] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 67},
  [4199] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4200] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4201] = {.lex_state = 140, .external_lex_state = 2},
  [4202] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4203] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 67},
  [4204] = {.lex_state = 140, .external_lex_state = 2},
  [4205] = {.lex_state = 62, .external_lex_state = 5},
  [4206] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4207] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 67},
  [4208] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 67},
  [4209] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4210] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4211] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4212] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4213] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 67},
  [4214] = {.lex_state = 138, .external_lex_state = 3},
  [4215] = {.lex_state = 140, .external_lex_state = 2},
  [4216] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 86},
  [4217] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4218] = {.lex_state = 138, .external_lex_state = 3},
  [4219] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4220] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4221] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4222] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 67},
  [4223] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 67},
  [4224] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4225] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4226] = {.lex_state = 138, .external_lex_state = 3},
  [4227] = {.lex_state = 138, .external_lex_state = 3},
  [4228] = {.lex_state = 62, .external_lex_state = 5},
  [4229] = {.lex_state = 140, .external_lex_state = 2},
  [4230] = {.lex_state = 151, .external_lex_state = 2},
  [4231] = {.lex_state = 151, .external_lex_state = 2},
  [4232] = {.lex_state = 64, .external_lex_state = 2},
  [4233] = {.lex_state = 138, .external_lex_state = 2},
  [4234] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4235] = {.lex_state = 138, .external_lex_state = 2},
  [4236] = {.lex_state = 151, .external_lex_state = 2},
  [4237] = {.lex_state = 151, .external_lex_state = 2},
  [4238] = {.lex_state = 151, .external_lex_state = 2},
  [4239] = {.lex_state = 65, .external_lex_state = 2},
  [4240] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4241] = {.lex_state = 138, .external_lex_state = 2},
  [4242] = {.lex_state = 138, .external_lex_state = 2},
  [4243] = {.lex_state = 138, .external_lex_state = 2},
  [4244] = {.lex_state = 138, .external_lex_state = 2},
  [4245] = {.lex_state = 63, .external_lex_state = 2},
  [4246] = {.lex_state = 151, .external_lex_state = 2},
  [4247] = {.lex_state = 138, .external_lex_state = 2},
  [4248] = {.lex_state = 63, .external_lex_state = 2},
  [4249] = {.lex_state = 151, .external_lex_state = 2},
  [4250] = {.lex_state = 151, .external_lex_state = 2},
  [4251] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4252] = {.lex_state = 138, .external_lex_state = 2},
  [4253] = {.lex_state = 80, .external_lex_state = 2},
  [4254] = {.lex_state = 65, .external_lex_state = 2},
  [4255] = {.lex_state = 65, .external_lex_state = 2},
  [4256] = {.lex_state = 140, .external_lex_state = 2},
  [4257] = {.lex_state = 140, .external_lex_state = 2},
  [4258] = {.lex_state = 140, .external_lex_state = 2},
  [4259] = {.lex_state = 140, .external_lex_state = 2},
  [4260] = {.lex_state = 140, .external_lex_state = 2},
  [4261] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4262] = {.lex_state = 140, .external_lex_state = 2},
  [4263] = {.lex_state = 140, .external_lex_state = 2},
  [4264] = {.lex_state = 138, .external_lex_state = 2},
  [4265] = {.lex_state = 65, .external_lex_state = 2},
  [4266] = {.lex_state = 65, .external_lex_state = 2},
  [4267] = {.lex_state = 138, .external_lex_state = 2},
  [4268] = {.lex_state = 65, .external_lex_state = 2},
  [4269] = {.lex_state = 138, .external_lex_state = 2},
  [4270] = {.lex_state = 138, .external_lex_state = 2},
  [4271] = {.lex_state = 140, .external_lex_state = 2},
  [4272] = {.lex_state = 138, .external_lex_state = 2},
  [4273] = {.lex_state = 138, .external_lex_state = 2},
  [4274] = {.lex_state = 140, .external_lex_state = 2},
  [4275] = {.lex_state = 138, .external_lex_state = 2},
  [4276] = {.lex_state = 140, .external_lex_state = 2},
  [4277] = {.lex_state = 138, .external_lex_state = 2},
  [4278] = {.lex_state = 138, .external_lex_state = 2},
  [4279] = {.lex_state = 62, .external_lex_state = 5},
  [4280] = {.lex_state = 65, .external_lex_state = 2},
  [4281] = {.lex_state = 80, .external_lex_state = 2},
  [4282] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4283] = {.lex_state = 65, .external_lex_state = 2},
  [4284] = {.lex_state = 140, .external_lex_state = 2},
  [4285] = {.lex_state = 67, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4286] = {.lex_state = 138, .external_lex_state = 2},
  [4287] = {.lex_state = 138, .external_lex_state = 2},
  [4288] = {.lex_state = 138, .external_lex_state = 2},
  [4289] = {.lex_state = 138, .external_lex_state = 2},
  [4290] = {.lex_state = 140, .external_lex_state = 2},
  [4291] = {.lex_state = 140, .external_lex_state = 2},
  [4292] = {.lex_state = 140, .external_lex_state = 2},
  [4293] = {.lex_state = 140, .external_lex_state = 2},
  [4294] = {.lex_state = 138, .external_lex_state = 2},
  [4295] = {.lex_state = 80, .external_lex_state = 2},
  [4296] = {.lex_state = 140, .external_lex_state = 2},
  [4297] = {.lex_state = 138, .external_lex_state = 2},
  [4298] = {.lex_state = 138, .external_lex_state = 2},
  [4299] = {.lex_state = 140, .external_lex_state = 2},
  [4300] = {.lex_state = 140, .external_lex_state = 2},
  [4301] = {.lex_state = 140, .external_lex_state = 2},
  [4302] = {.lex_state = 140, .external_lex_state = 2},
  [4303] = {.lex_state = 140, .external_lex_state = 2},
  [4304] = {.lex_state = 62, .external_lex_state = 5},
  [4305] = {.lex_state = 138, .external_lex_state = 2},
  [4306] = {.lex_state = 138, .external_lex_state = 2},
  [4307] = {.lex_state = 140, .external_lex_state = 2},
  [4308] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4309] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4310] = {.lex_state = 140, .external_lex_state = 2},
  [4311] = {.lex_state = 65, .external_lex_state = 2},
  [4312] = {.lex_state = 140, .external_lex_state = 2},
  [4313] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4314] = {.lex_state = 140, .external_lex_state = 2},
  [4315] = {.lex_state = 63, .external_lex_state = 2},
  [4316] = {.lex_state = 140, .external_lex_state = 2},
  [4317] = {.lex_state = 63, .external_lex_state = 2},
  [4318] = {.lex_state = 140, .external_lex_state = 2},
  [4319] = {.lex_state = 63, .external_lex_state = 2},
  [4320] = {.lex_state = 140, .external_lex_state = 2},
  [4321] = {.lex_state = 63, .external_lex_state = 2},
  [4322] = {.lex_state = 63, .external_lex_state = 2},
  [4323] = {.lex_state = 63, .external_lex_state = 2},
  [4324] = {.lex_state = 63, .external_lex_state = 2},
  [4325] = {.lex_state = 140, .external_lex_state = 2},
  [4326] = {.lex_state = 63, .external_lex_state = 2},
  [4327] = {.lex_state = 63, .external_lex_state = 2},
  [4328] = {.lex_state = 140, .external_lex_state = 2},
  [4329] = {.lex_state = 140, .external_lex_state = 2},
  [4330] = {.lex_state = 63, .external_lex_state = 2},
  [4331] = {.lex_state = 140, .external_lex_state = 2},
  [4332] = {.lex_state = 140, .external_lex_state = 2},
  [4333] = {.lex_state = 63, .external_lex_state = 2},
  [4334] = {.lex_state = 63, .external_lex_state = 2},
  [4335] = {.lex_state = 140, .external_lex_state = 2},
  [4336] = {.lex_state = 63, .external_lex_state = 2},
  [4337] = {.lex_state = 140, .external_lex_state = 2},
  [4338] = {.lex_state = 140, .external_lex_state = 2},
  [4339] = {.lex_state = 140, .external_lex_state = 2},
  [4340] = {.lex_state = 63, .external_lex_state = 2},
  [4341] = {.lex_state = 140, .external_lex_state = 2},
  [4342] = {.lex_state = 140, .external_lex_state = 2},
  [4343] = {.lex_state = 140, .external_lex_state = 2},
  [4344] = {.lex_state = 140, .external_lex_state = 2},
  [4345] = {.lex_state = 140, .external_lex_state = 2},
  [4346] = {.lex_state = 140, .external_lex_state = 2},
  [4347] = {.lex_state = 140, .external_lex_state = 2},
  [4348] = {.lex_state = 140, .external_lex_state = 2},
  [4349] = {.lex_state = 140, .external_lex_state = 2},
  [4350] = {.lex_state = 140, .external_lex_state = 2},
  [4351] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 88},
  [4352] = {.lex_state = 140, .external_lex_state = 2},
  [4353] = {.lex_state = 140, .external_lex_state = 2},
  [4354] = {.lex_state = 140, .external_lex_state = 2},
  [4355] = {.lex_state = 140, .external_lex_state = 2},
  [4356] = {.lex_state = 140, .external_lex_state = 2},
  [4357] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4358] = {.lex_state = 63, .external_lex_state = 2},
  [4359] = {.lex_state = 63, .external_lex_state = 2},
  [4360] = {.lex_state = 140, .external_lex_state = 2},
  [4361] = {.lex_state = 63, .external_lex_state = 2},
  [4362] = {.lex_state = 140, .external_lex_state = 2},
  [4363] = {.lex_state = 140, .external_lex_state = 2},
  [4364] = {.lex_state = 63, .external_lex_state = 2},
  [4365] = {.lex_state = 140, .external_lex_state = 2},
  [4366] = {.lex_state = 151, .external_lex_state = 2},
  [4367] = {.lex_state = 140, .external_lex_state = 2},
  [4368] = {.lex_state = 140, .external_lex_state = 2},
  [4369] = {.lex_state = 140, .external_lex_state = 2},
  [4370] = {.lex_state = 140, .external_lex_state = 2},
  [4371] = {.lex_state = 140, .external_lex_state = 2},
  [4372] = {.lex_state = 140, .external_lex_state = 2},
  [4373] = {.lex_state = 140, .external_lex_state = 2},
  [4374] = {.lex_state = 140, .external_lex_state = 2},
  [4375] = {.lex_state = 140, .external_lex_state = 2},
  [4376] = {.lex_state = 140, .external_lex_state = 2},
  [4377] = {.lex_state = 140, .external_lex_state = 2},
  [4378] = {.lex_state = 140, .external_lex_state = 2},
  [4379] = {.lex_state = 63, .external_lex_state = 2},
  [4380] = {.lex_state = 63, .external_lex_state = 2},
  [4381] = {.lex_state = 140, .external_lex_state = 2},
  [4382] = {.lex_state = 140, .external_lex_state = 2},
  [4383] = {.lex_state = 140, .external_lex_state = 2},
  [4384] = {.lex_state = 140, .external_lex_state = 2},
  [4385] = {.lex_state = 63, .external_lex_state = 2},
  [4386] = {.lex_state = 63, .external_lex_state = 2},
  [4387] = {.lex_state = 63, .external_lex_state = 2},
  [4388] = {.lex_state = 63, .external_lex_state = 2},
  [4389] = {.lex_state = 140, .external_lex_state = 2},
  [4390] = {.lex_state = 140, .external_lex_state = 2},
  [4391] = {.lex_state = 140, .external_lex_state = 2},
  [4392] = {.lex_state = 140, .external_lex_state = 2},
  [4393] = {.lex_state = 140, .external_lex_state = 2},
  [4394] = {.lex_state = 140, .external_lex_state = 2},
  [4395] = {.lex_state = 140, .external_lex_state = 2},
  [4396] = {.lex_state = 140, .external_lex_state = 2},
  [4397] = {.lex_state = 140, .external_lex_state = 2},
  [4398] = {.lex_state = 63, .external_lex_state = 2},
  [4399] = {.lex_state = 140, .external_lex_state = 2},
  [4400] = {.lex_state = 140, .external_lex_state = 2},
  [4401] = {.lex_state = 140, .external_lex_state = 2},
  [4402] = {.lex_state = 140, .external_lex_state = 2},
  [4403] = {.lex_state = 140, .external_lex_state = 2},
  [4404] = {.lex_state = 140, .external_lex_state = 2},
  [4405] = {.lex_state = 140, .external_lex_state = 2},
  [4406] = {.lex_state = 140, .external_lex_state = 2},
  [4407] = {.lex_state = 140, .external_lex_state = 2},
  [4408] = {.lex_state = 140, .external_lex_state = 2},
  [4409] = {.lex_state = 85, .external_lex_state = 2},
  [4410] = {.lex_state = 63, .external_lex_state = 2},
  [4411] = {.lex_state = 140, .external_lex_state = 2},
  [4412] = {.lex_state = 140, .external_lex_state = 2},
  [4413] = {.lex_state = 140, .external_lex_state = 2},
  [4414] = {.lex_state = 85, .external_lex_state = 2},
  [4415] = {.lex_state = 151, .external_lex_state = 2},
  [4416] = {.lex_state = 140, .external_lex_state = 2},
  [4417] = {.lex_state = 140, .external_lex_state = 2},
  [4418] = {.lex_state = 140, .external_lex_state = 2},
  [4419] = {.lex_state = 140, .external_lex_state = 2},
  [4420] = {.lex_state = 140, .external_lex_state = 2},
  [4421] = {.lex_state = 140, .external_lex_state = 2},
  [4422] = {.lex_state = 140, .external_lex_state = 2},
  [4423] = {.lex_state = 140, .external_lex_state = 2},
  [4424] = {.lex_state = 140, .external_lex_state = 2},
  [4425] = {.lex_state = 140, .external_lex_state = 2},
  [4426] = {.lex_state = 140, .external_lex_state = 2},
  [4427] = {.lex_state = 151, .external_lex_state = 2},
  [4428] = {.lex_state = 140, .external_lex_state = 2},
  [4429] = {.lex_state = 140, .external_lex_state = 2},
  [4430] = {.lex_state = 63, .external_lex_state = 2},
  [4431] = {.lex_state = 63, .external_lex_state = 2},
  [4432] = {.lex_state = 140, .external_lex_state = 2},
  [4433] = {.lex_state = 140, .external_lex_state = 2},
  [4434] = {.lex_state = 63, .external_lex_state = 2},
  [4435] = {.lex_state = 63, .external_lex_state = 2},
  [4436] = {.lex_state = 63, .external_lex_state = 2},
  [4437] = {.lex_state = 63, .external_lex_state = 2},
  [4438] = {.lex_state = 63, .external_lex_state = 2},
  [4439] = {.lex_state = 63, .external_lex_state = 2},
  [4440] = {.lex_state = 63, .external_lex_state = 2},
  [4441] = {.lex_state = 63, .external_lex_state = 2},
  [4442] = {.lex_state = 151, .external_lex_state = 2},
  [4443] = {.lex_state = 82, .external_lex_state = 2},
  [4444] = {.lex_state = 151, .external_lex_state = 2},
  [4445] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4446] = {.lex_state = 140, .external_lex_state = 2},
  [4447] = {.lex_state = 140, .external_lex_state = 2},
  [4448] = {.lex_state = 140, .external_lex_state = 2},
  [4449] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4450] = {.lex_state = 140, .external_lex_state = 2},
  [4451] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4452] = {.lex_state = 67, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4453] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4454] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4455] = {.lex_state = 65, .external_lex_state = 2},
  [4456] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4457] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4458] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4459] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4460] = {.lex_state = 140, .external_lex_state = 2},
  [4461] = {.lex_state = 140, .external_lex_state = 2},
  [4462] = {.lex_state = 140, .external_lex_state = 2},
  [4463] = {.lex_state = 140, .external_lex_state = 2},
  [4464] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4465] = {.lex_state = 65, .external_lex_state = 2},
  [4466] = {.lex_state = 138, .external_lex_state = 2},
  [4467] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4468] = {.lex_state = 138, .external_lex_state = 2},
  [4469] = {.lex_state = 151, .external_lex_state = 2},
  [4470] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4471] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4472] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4473] = {.lex_state = 140, .external_lex_state = 2},
  [4474] = {.lex_state = 140, .external_lex_state = 2},
  [4475] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4476] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4477] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4478] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4479] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4480] = {.lex_state = 138, .external_lex_state = 2},
  [4481] = {.lex_state = 151, .external_lex_state = 2},
  [4482] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4483] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4484] = {.lex_state = 151, .external_lex_state = 2},
  [4485] = {.lex_state = 151, .external_lex_state = 2},
  [4486] = {.lex_state = 151, .external_lex_state = 2},
  [4487] = {.lex_state = 138, .external_lex_state = 2},
  [4488] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4489] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4490] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4491] = {.lex_state = 138, .external_lex_state = 2},
  [4492] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4493] = {.lex_state = 151, .external_lex_state = 2},
  [4494] = {.lex_state = 138, .external_lex_state = 2},
  [4495] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4496] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4497] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 87},
  [4498] = {.lex_state = 138, .external_lex_state = 2},
  [4499] = {.lex_state = 151, .external_lex_state = 2},
  [4500] = {.lex_state = 151, .external_lex_state = 2},
  [4501] = {.lex_state = 151, .external_lex_state = 2},
  [4502] = {.lex_state = 138, .external_lex_state = 2},
  [4503] = {.lex_state = 138, .external_lex_state = 2},
  [4504] = {.lex_state = 138, .external_lex_state = 2},
  [4505] = {.lex_state = 138, .external_lex_state = 2},
  [4506] = {.lex_state = 138, .external_lex_state = 2},
  [4507] = {.lex_state = 138, .external_lex_state = 2},
  [4508] = {.lex_state = 151, .external_lex_state = 2},
  [4509] = {.lex_state = 151, .external_lex_state = 2},
  [4510] = {.lex_state = 138, .external_lex_state = 2},
  [4511] = {.lex_state = 138, .external_lex_state = 2},
  [4512] = {.lex_state = 151, .external_lex_state = 2},
  [4513] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4514] = {.lex_state = 151, .external_lex_state = 2},
  [4515] = {.lex_state = 138, .external_lex_state = 2},
  [4516] = {.lex_state = 151, .external_lex_state = 2},
  [4517] = {.lex_state = 151, .external_lex_state = 2},
  [4518] = {.lex_state = 63, .external_lex_state = 2},
  [4519] = {.lex_state = 151, .external_lex_state = 2},
  [4520] = {.lex_state = 151, .external_lex_state = 2},
  [4521] = {.lex_state = 63, .external_lex_state = 2},
  [4522] = {.lex_state = 151, .external_lex_state = 2},
  [4523] = {.lex_state = 151, .external_lex_state = 2},
  [4524] = {.lex_state = 138, .external_lex_state = 2},
  [4525] = {.lex_state = 138, .external_lex_state = 2},
  [4526] = {.lex_state = 138, .external_lex_state = 2},
  [4527] = {.lex_state = 138, .external_lex_state = 2},
  [4528] = {.lex_state = 151, .external_lex_state = 2},
  [4529] = {.lex_state = 151, .external_lex_state = 2},
  [4530] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4531] = {.lex_state = 151, .external_lex_state = 2},
  [4532] = {.lex_state = 138, .external_lex_state = 2},
  [4533] = {.lex_state = 151, .external_lex_state = 2},
  [4534] = {.lex_state = 138, .external_lex_state = 2},
  [4535] = {.lex_state = 151, .external_lex_state = 2},
  [4536] = {.lex_state = 151, .external_lex_state = 2},
  [4537] = {.lex_state = 151, .external_lex_state = 2},
  [4538] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4539] = {.lex_state = 80, .external_lex_state = 2},
  [4540] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4541] = {.lex_state = 154, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4542] = {.lex_state = 138, .external_lex_state = 2},
  [4543] = {.lex_state = 138, .external_lex_state = 2},
  [4544] = {.lex_state = 63, .external_lex_state = 2},
  [4545] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4546] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4547] = {.lex_state = 4, .external_lex_state = 6},
  [4548] = {.lex_state = 4, .external_lex_state = 6},
  [4549] = {.lex_state = 63, .external_lex_state = 2},
  [4550] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4551] = {.lex_state = 151, .external_lex_state = 2},
  [4552] = {.lex_state = 63, .external_lex_state = 2},
  [4553] = {.lex_state = 63, .external_lex_state = 2},
  [4554] = {.lex_state = 151, .external_lex_state = 2},
  [4555] = {.lex_state = 4, .external_lex_state = 6},
  [4556] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4557] = {.lex_state = 63, .external_lex_state = 2},
  [4558] = {.lex_state = 63, .external_lex_state = 2},
  [4559] = {.lex_state = 63, .external_lex_state = 2},
  [4560] = {.lex_state = 63, .external_lex_state = 2},
  [4561] = {.lex_state = 63, .external_lex_state = 2},
  [4562] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4563] = {.lex_state = 4, .external_lex_state = 6},
  [4564] = {.lex_state = 63, .external_lex_state = 2},
  [4565] = {.lex_state = 4, .external_lex_state = 6},
  [4566] = {.lex_state = 76, .external_lex_state = 7},
  [4567] = {.lex_state = 4, .external_lex_state = 6},
  [4568] = {.lex_state = 4, .external_lex_state = 6},
  [4569] = {.lex_state = 4, .external_lex_state = 6},
  [4570] = {.lex_state = 4, .external_lex_state = 6},
  [4571] = {.lex_state = 4, .external_lex_state = 6},
  [4572] = {.lex_state = 4, .external_lex_state = 6},
  [4573] = {.lex_state = 63, .external_lex_state = 2},
  [4574] = {.lex_state = 151, .external_lex_state = 2},
  [4575] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4576] = {.lex_state = 151, .external_lex_state = 2},
  [4577] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4578] = {.lex_state = 140, .external_lex_state = 3},
  [4579] = {.lex_state = 63, .external_lex_state = 2},
  [4580] = {.lex_state = 151, .external_lex_state = 2},
  [4581] = {.lex_state = 63, .external_lex_state = 2},
  [4582] = {.lex_state = 63, .external_lex_state = 2},
  [4583] = {.lex_state = 63, .external_lex_state = 2},
  [4584] = {.lex_state = 4, .external_lex_state = 6},
  [4585] = {.lex_state = 151, .external_lex_state = 2},
  [4586] = {.lex_state = 63, .external_lex_state = 2},
  [4587] = {.lex_state = 4, .external_lex_state = 6},
  [4588] = {.lex_state = 151, .external_lex_state = 2},
  [4589] = {.lex_state = 63, .external_lex_state = 2},
  [4590] = {.lex_state = 151, .external_lex_state = 2},
  [4591] = {.lex_state = 63, .external_lex_state = 2},
  [4592] = {.lex_state = 151, .external_lex_state = 2},
  [4593] = {.lex_state = 63, .external_lex_state = 2},
  [4594] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4595] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4596] = {.lex_state = 4, .external_lex_state = 6},
  [4597] = {.lex_state = 4, .external_lex_state = 6},
  [4598] = {.lex_state = 63, .external_lex_state = 2},
  [4599] = {.lex_state = 151, .external_lex_state = 2},
  [4600] = {.lex_state = 4, .external_lex_state = 6},
  [4601] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4602] = {.lex_state = 63, .external_lex_state = 2},
  [4603] = {.lex_state = 63, .external_lex_state = 2},
  [4604] = {.lex_state = 63, .external_lex_state = 2},
  [4605] = {.lex_state = 63, .external_lex_state = 2},
  [4606] = {.lex_state = 63, .external_lex_state = 2},
  [4607] = {.lex_state = 4, .external_lex_state = 6},
  [4608] = {.lex_state = 151, .external_lex_state = 2},
  [4609] = {.lex_state = 140, .external_lex_state = 3},
  [4610] = {.lex_state = 63, .external_lex_state = 2},
  [4611] = {.lex_state = 63, .external_lex_state = 2},
  [4612] = {.lex_state = 63, .external_lex_state = 2},
  [4613] = {.lex_state = 140, .external_lex_state = 2},
  [4614] = {.lex_state = 140, .external_lex_state = 2},
  [4615] = {.lex_state = 63, .external_lex_state = 2},
  [4616] = {.lex_state = 63, .external_lex_state = 2},
  [4617] = {.lex_state = 63, .external_lex_state = 2},
  [4618] = {.lex_state = 63, .external_lex_state = 2},
  [4619] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4620] = {.lex_state = 9, .external_lex_state = 2},
  [4621] = {.lex_state = 140, .external_lex_state = 2},
  [4622] = {.lex_state = 151, .external_lex_state = 2},
  [4623] = {.lex_state = 76, .external_lex_state = 7},
  [4624] = {.lex_state = 151, .external_lex_state = 2},
  [4625] = {.lex_state = 9, .external_lex_state = 2},
  [4626] = {.lex_state = 9, .external_lex_state = 2},
  [4627] = {.lex_state = 151, .external_lex_state = 2},
  [4628] = {.lex_state = 151, .external_lex_state = 2},
  [4629] = {.lex_state = 10, .external_lex_state = 2},
  [4630] = {.lex_state = 10, .external_lex_state = 2},
  [4631] = {.lex_state = 64, .external_lex_state = 2},
  [4632] = {.lex_state = 151, .external_lex_state = 2},
  [4633] = {.lex_state = 151, .external_lex_state = 2},
  [4634] = {.lex_state = 151, .external_lex_state = 2},
  [4635] = {.lex_state = 9, .external_lex_state = 2},
  [4636] = {.lex_state = 151, .external_lex_state = 2},
  [4637] = {.lex_state = 10, .external_lex_state = 2},
  [4638] = {.lex_state = 151, .external_lex_state = 2},
  [4639] = {.lex_state = 151, .external_lex_state = 2},
  [4640] = {.lex_state = 9, .external_lex_state = 2},
  [4641] = {.lex_state = 140, .external_lex_state = 2},
  [4642] = {.lex_state = 138, .external_lex_state = 2},
  [4643] = {.lex_state = 9, .external_lex_state = 2},
  [4644] = {.lex_state = 151, .external_lex_state = 2},
  [4645] = {.lex_state = 138, .external_lex_state = 2},
  [4646] = {.lex_state = 82, .external_lex_state = 2},
  [4647] = {.lex_state = 151, .external_lex_state = 2},
  [4648] = {.lex_state = 151, .external_lex_state = 2},
  [4649] = {.lex_state = 151, .external_lex_state = 2},
  [4650] = {.lex_state = 151, .external_lex_state = 2},
  [4651] = {.lex_state = 151, .external_lex_state = 2},
  [4652] = {.lex_state = 10, .external_lex_state = 2},
  [4653] = {.lex_state = 10, .external_lex_state = 2},
  [4654] = {.lex_state = 151, .external_lex_state = 2},
  [4655] = {.lex_state = 9, .external_lex_state = 2},
  [4656] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4657] = {.lex_state = 151, .external_lex_state = 2},
  [4658] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4659] = {.lex_state = 151, .external_lex_state = 2},
  [4660] = {.lex_state = 151, .external_lex_state = 2},
  [4661] = {.lex_state = 10, .external_lex_state = 2},
  [4662] = {.lex_state = 151, .external_lex_state = 2},
  [4663] = {.lex_state = 151, .external_lex_state = 2},
  [4664] = {.lex_state = 9, .external_lex_state = 2},
  [4665] = {.lex_state = 76, .external_lex_state = 7},
  [4666] = {.lex_state = 151, .external_lex_state = 2},
  [4667] = {.lex_state = 151, .external_lex_state = 2},
  [4668] = {.lex_state = 9, .external_lex_state = 2},
  [4669] = {.lex_state = 9, .external_lex_state = 2},
  [4670] = {.lex_state = 51, .external_lex_state = 2},
  [4671] = {.lex_state = 9, .external_lex_state = 2},
  [4672] = {.lex_state = 151, .external_lex_state = 2},
  [4673] = {.lex_state = 151, .external_lex_state = 2},
  [4674] = {.lex_state = 9, .external_lex_state = 2},
  [4675] = {.lex_state = 10, .external_lex_state = 2},
  [4676] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4677] = {.lex_state = 82, .external_lex_state = 2},
  [4678] = {.lex_state = 9, .external_lex_state = 2},
  [4679] = {.lex_state = 138, .external_lex_state = 2},
  [4680] = {.lex_state = 151, .external_lex_state = 2},
  [4681] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4682] = {.lex_state = 9, .external_lex_state = 2},
  [4683] = {.lex_state = 138, .external_lex_state = 2},
  [4684] = {.lex_state = 138, .external_lex_state = 2},
  [4685] = {.lex_state = 138, .external_lex_state = 2},
  [4686] = {.lex_state = 138, .external_lex_state = 2},
  [4687] = {.lex_state = 151, .external_lex_state = 2},
  [4688] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4689] = {.lex_state = 63, .external_lex_state = 2},
  [4690] = {.lex_state = 25, .external_lex_state = 8, .reserved_word_set_id = 6},
  [4691] = {.lex_state = 25, .external_lex_state = 8, .reserved_word_set_id = 6},
  [4692] = {.lex_state = 9, .external_lex_state = 2},
  [4693] = {.lex_state = 138, .external_lex_state = 2},
  [4694] = {.lex_state = 151, .external_lex_state = 2},
  [4695] = {.lex_state = 138, .external_lex_state = 2},
  [4696] = {.lex_state = 138, .external_lex_state = 2},
  [4697] = {.lex_state = 138, .external_lex_state = 2},
  [4698] = {.lex_state = 63, .external_lex_state = 2},
  [4699] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4700] = {.lex_state = 151, .external_lex_state = 2},
  [4701] = {.lex_state = 63, .external_lex_state = 2},
  [4702] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4703] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4704] = {.lex_state = 82, .external_lex_state = 2},
  [4705] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4706] = {.lex_state = 138, .external_lex_state = 2},
  [4707] = {.lex_state = 138, .external_lex_state = 2},
  [4708] = {.lex_state = 138, .external_lex_state = 2},
  [4709] = {.lex_state = 151, .external_lex_state = 2},
  [4710] = {.lex_state = 151, .external_lex_state = 2},
  [4711] = {.lex_state = 63, .external_lex_state = 2},
  [4712] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4713] = {.lex_state = 63, .external_lex_state = 2},
  [4714] = {.lex_state = 138, .external_lex_state = 2},
  [4715] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4716] = {.lex_state = 63, .external_lex_state = 2},
  [4717] = {.lex_state = 151, .external_lex_state = 2},
  [4718] = {.lex_state = 63, .external_lex_state = 2},
  [4719] = {.lex_state = 138, .external_lex_state = 2},
  [4720] = {.lex_state = 138, .external_lex_state = 2},
  [4721] = {.lex_state = 138, .external_lex_state = 2},
  [4722] = {.lex_state = 138, .external_lex_state = 2},
  [4723] = {.lex_state = 138, .external_lex_state = 2},
  [4724] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4725] = {.lex_state = 63, .external_lex_state = 2},
  [4726] = {.lex_state = 63, .external_lex_state = 2},
  [4727] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4728] = {.lex_state = 138, .external_lex_state = 2},
  [4729] = {.lex_state = 63, .external_lex_state = 2},
  [4730] = {.lex_state = 151, .external_lex_state = 2},
  [4731] = {.lex_state = 151, .external_lex_state = 2},
  [4732] = {.lex_state = 138, .external_lex_state = 2},
  [4733] = {.lex_state = 151, .external_lex_state = 2},
  [4734] = {.lex_state = 151, .external_lex_state = 2},
  [4735] = {.lex_state = 138, .external_lex_state = 2},
  [4736] = {.lex_state = 9, .external_lex_state = 2},
  [4737] = {.lex_state = 9, .external_lex_state = 2},
  [4738] = {.lex_state = 9, .external_lex_state = 2},
  [4739] = {.lex_state = 151, .external_lex_state = 2},
  [4740] = {.lex_state = 138, .external_lex_state = 2},
  [4741] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4742] = {.lex_state = 9, .external_lex_state = 2},
  [4743] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4744] = {.lex_state = 9, .external_lex_state = 2},
  [4745] = {.lex_state = 9, .external_lex_state = 2},
  [4746] = {.lex_state = 9, .external_lex_state = 2},
  [4747] = {.lex_state = 9, .external_lex_state = 2},
  [4748] = {.lex_state = 9, .external_lex_state = 2},
  [4749] = {.lex_state = 138, .external_lex_state = 2},
  [4750] = {.lex_state = 63, .external_lex_state = 2},
  [4751] = {.lex_state = 151, .external_lex_state = 2},
  [4752] = {.lex_state = 151, .external_lex_state = 2},
  [4753] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4754] = {.lex_state = 138, .external_lex_state = 2},
  [4755] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4756] = {.lex_state = 138, .external_lex_state = 2},
  [4757] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4758] = {.lex_state = 138, .external_lex_state = 2},
  [4759] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4760] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4761] = {.lex_state = 76, .external_lex_state = 7},
  [4762] = {.lex_state = 138, .external_lex_state = 2},
  [4763] = {.lex_state = 151, .external_lex_state = 2},
  [4764] = {.lex_state = 151, .external_lex_state = 2},
  [4765] = {.lex_state = 25, .external_lex_state = 8, .reserved_word_set_id = 6},
  [4766] = {.lex_state = 25, .external_lex_state = 8, .reserved_word_set_id = 6},
  [4767] = {.lex_state = 25, .external_lex_state = 8, .reserved_word_set_id = 6},
  [4768] = {.lex_state = 138, .external_lex_state = 2},
  [4769] = {.lex_state = 151, .external_lex_state = 2},
  [4770] = {.lex_state = 151, .external_lex_state = 2},
  [4771] = {.lex_state = 151, .external_lex_state = 2},
  [4772] = {.lex_state = 138, .external_lex_state = 2},
  [4773] = {.lex_state = 25, .external_lex_state = 8, .reserved_word_set_id = 6},
  [4774] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4775] = {.lex_state = 25, .external_lex_state = 8, .reserved_word_set_id = 6},
  [4776] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4777] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4778] = {.lex_state = 138, .external_lex_state = 2},
  [4779] = {.lex_state = 138, .external_lex_state = 2},
  [4780] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4781] = {.lex_state = 138, .external_lex_state = 2},
  [4782] = {.lex_state = 63, .external_lex_state = 2},
  [4783] = {.lex_state = 9, .external_lex_state = 2},
  [4784] = {.lex_state = 63, .external_lex_state = 2},
  [4785] = {.lex_state = 25, .external_lex_state = 8, .reserved_word_set_id = 6},
  [4786] = {.lex_state = 64, .external_lex_state = 2},
  [4787] = {.lex_state = 63, .external_lex_state = 2},
  [4788] = {.lex_state = 9, .external_lex_state = 2},
  [4789] = {.lex_state = 138, .external_lex_state = 2},
  [4790] = {.lex_state = 9, .external_lex_state = 2},
  [4791] = {.lex_state = 140, .external_lex_state = 3},
  [4792] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4793] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4794] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4795] = {.lex_state = 138, .external_lex_state = 2},
  [4796] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4797] = {.lex_state = 138, .external_lex_state = 2},
  [4798] = {.lex_state = 138, .external_lex_state = 2},
  [4799] = {.lex_state = 48, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4800] = {.lex_state = 151, .external_lex_state = 2},
  [4801] = {.lex_state = 138, .external_lex_state = 2},
  [4802] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4803] = {.lex_state = 86, .external_lex_state = 2},
  [4804] = {.lex_state = 138, .external_lex_state = 2},
  [4805] = {.lex_state = 63, .external_lex_state = 2},
  [4806] = {.lex_state = 9, .external_lex_state = 2},
  [4807] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4808] = {.lex_state = 63, .external_lex_state = 2},
  [4809] = {.lex_state = 138, .external_lex_state = 2},
  [4810] = {.lex_state = 86, .external_lex_state = 2},
  [4811] = {.lex_state = 63, .external_lex_state = 2},
  [4812] = {.lex_state = 138, .external_lex_state = 2},
  [4813] = {.lex_state = 9, .external_lex_state = 2},
  [4814] = {.lex_state = 138, .external_lex_state = 2},
  [4815] = {.lex_state = 63, .external_lex_state = 2},
  [4816] = {.lex_state = 138, .external_lex_state = 2},
  [4817] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4818] = {.lex_state = 9, .external_lex_state = 2},
  [4819] = {.lex_state = 9, .external_lex_state = 2},
  [4820] = {.lex_state = 63, .external_lex_state = 2},
  [4821] = {.lex_state = 9, .external_lex_state = 2},
  [4822] = {.lex_state = 138, .external_lex_state = 2},
  [4823] = {.lex_state = 63, .external_lex_state = 2},
  [4824] = {.lex_state = 9, .external_lex_state = 2},
  [4825] = {.lex_state = 138, .external_lex_state = 2},
  [4826] = {.lex_state = 138, .external_lex_state = 2},
  [4827] = {.lex_state = 9, .external_lex_state = 2},
  [4828] = {.lex_state = 151, .external_lex_state = 2},
  [4829] = {.lex_state = 140, .external_lex_state = 2},
  [4830] = {.lex_state = 9, .external_lex_state = 2},
  [4831] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4832] = {.lex_state = 9, .external_lex_state = 2},
  [4833] = {.lex_state = 9, .external_lex_state = 2},
  [4834] = {.lex_state = 9, .external_lex_state = 2},
  [4835] = {.lex_state = 151, .external_lex_state = 2},
  [4836] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4837] = {.lex_state = 9, .external_lex_state = 2},
  [4838] = {.lex_state = 140, .external_lex_state = 2},
  [4839] = {.lex_state = 9, .external_lex_state = 2},
  [4840] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 90},
  [4841] = {.lex_state = 9, .external_lex_state = 2},
  [4842] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4843] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4844] = {.lex_state = 138, .external_lex_state = 2},
  [4845] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4846] = {.lex_state = 9, .external_lex_state = 2},
  [4847] = {.lex_state = 9, .external_lex_state = 2},
  [4848] = {.lex_state = 138, .external_lex_state = 2},
  [4849] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4850] = {.lex_state = 9, .external_lex_state = 2},
  [4851] = {.lex_state = 9, .external_lex_state = 2},
  [4852] = {.lex_state = 9, .external_lex_state = 2},
  [4853] = {.lex_state = 63, .external_lex_state = 2},
  [4854] = {.lex_state = 9, .external_lex_state = 2},
  [4855] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4856] = {.lex_state = 9, .external_lex_state = 2},
  [4857] = {.lex_state = 9, .external_lex_state = 2},
  [4858] = {.lex_state = 9, .external_lex_state = 2},
  [4859] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4860] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4861] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4862] = {.lex_state = 9, .external_lex_state = 2},
  [4863] = {.lex_state = 9, .external_lex_state = 2},
  [4864] = {.lex_state = 23, .external_lex_state = 2},
  [4865] = {.lex_state = 9, .external_lex_state = 2},
  [4866] = {.lex_state = 23, .external_lex_state = 2},
  [4867] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4868] = {.lex_state = 9, .external_lex_state = 2},
  [4869] = {.lex_state = 23, .external_lex_state = 2},
  [4870] = {.lex_state = 151, .external_lex_state = 2},
  [4871] = {.lex_state = 151, .external_lex_state = 2},
  [4872] = {.lex_state = 9, .external_lex_state = 2},
  [4873] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4874] = {.lex_state = 23, .external_lex_state = 2},
  [4875] = {.lex_state = 23, .external_lex_state = 2},
  [4876] = {.lex_state = 23, .external_lex_state = 2},
  [4877] = {.lex_state = 23, .external_lex_state = 2},
  [4878] = {.lex_state = 23, .external_lex_state = 2},
  [4879] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4880] = {.lex_state = 9, .external_lex_state = 2},
  [4881] = {.lex_state = 63, .external_lex_state = 2},
  [4882] = {.lex_state = 9, .external_lex_state = 2},
  [4883] = {.lex_state = 9, .external_lex_state = 2},
  [4884] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4885] = {.lex_state = 138, .external_lex_state = 2},
  [4886] = {.lex_state = 9, .external_lex_state = 2},
  [4887] = {.lex_state = 9, .external_lex_state = 2},
  [4888] = {.lex_state = 138, .external_lex_state = 2},
  [4889] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4890] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4891] = {.lex_state = 9, .external_lex_state = 2},
  [4892] = {.lex_state = 9, .external_lex_state = 2},
  [4893] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4894] = {.lex_state = 9, .external_lex_state = 2},
  [4895] = {.lex_state = 140, .external_lex_state = 2},
  [4896] = {.lex_state = 9, .external_lex_state = 2},
  [4897] = {.lex_state = 9, .external_lex_state = 2},
  [4898] = {.lex_state = 9, .external_lex_state = 2},
  [4899] = {.lex_state = 9, .external_lex_state = 2},
  [4900] = {.lex_state = 9, .external_lex_state = 2},
  [4901] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4902] = {.lex_state = 9, .external_lex_state = 2},
  [4903] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4904] = {.lex_state = 9, .external_lex_state = 2},
  [4905] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4906] = {.lex_state = 151, .external_lex_state = 2},
  [4907] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4908] = {.lex_state = 23, .external_lex_state = 2},
  [4909] = {.lex_state = 63, .external_lex_state = 2},
  [4910] = {.lex_state = 9, .external_lex_state = 2},
  [4911] = {.lex_state = 9, .external_lex_state = 2},
  [4912] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4913] = {.lex_state = 9, .external_lex_state = 2},
  [4914] = {.lex_state = 9, .external_lex_state = 2},
  [4915] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4916] = {.lex_state = 9, .external_lex_state = 2},
  [4917] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4918] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4919] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4920] = {.lex_state = 140, .external_lex_state = 2},
  [4921] = {.lex_state = 138, .external_lex_state = 2},
  [4922] = {.lex_state = 9, .external_lex_state = 2},
  [4923] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4924] = {.lex_state = 9, .external_lex_state = 2},
  [4925] = {.lex_state = 9, .external_lex_state = 2},
  [4926] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4927] = {.lex_state = 9, .external_lex_state = 2},
  [4928] = {.lex_state = 9, .external_lex_state = 2},
  [4929] = {.lex_state = 9, .external_lex_state = 2},
  [4930] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4931] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4932] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4933] = {.lex_state = 63, .external_lex_state = 2},
  [4934] = {.lex_state = 140, .external_lex_state = 2},
  [4935] = {.lex_state = 9, .external_lex_state = 2},
  [4936] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4937] = {.lex_state = 9, .external_lex_state = 2},
  [4938] = {.lex_state = 151, .external_lex_state = 2},
  [4939] = {.lex_state = 23, .external_lex_state = 2},
  [4940] = {.lex_state = 9, .external_lex_state = 2},
  [4941] = {.lex_state = 140, .external_lex_state = 2},
  [4942] = {.lex_state = 9, .external_lex_state = 2},
  [4943] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4944] = {.lex_state = 9, .external_lex_state = 2},
  [4945] = {.lex_state = 151, .external_lex_state = 2},
  [4946] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4947] = {.lex_state = 63, .external_lex_state = 2},
  [4948] = {.lex_state = 63, .external_lex_state = 2},
  [4949] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4950] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4951] = {.lex_state = 9, .external_lex_state = 2},
  [4952] = {.lex_state = 151, .external_lex_state = 2},
  [4953] = {.lex_state = 63, .external_lex_state = 2},
  [4954] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4955] = {.lex_state = 151, .external_lex_state = 2},
  [4956] = {.lex_state = 151, .external_lex_state = 2},
  [4957] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4958] = {.lex_state = 9, .external_lex_state = 2},
  [4959] = {.lex_state = 9, .external_lex_state = 2},
  [4960] = {.lex_state = 151, .external_lex_state = 2},
  [4961] = {.lex_state = 9, .external_lex_state = 2},
  [4962] = {.lex_state = 151, .external_lex_state = 2},
  [4963] = {.lex_state = 9, .external_lex_state = 2},
  [4964] = {.lex_state = 140, .external_lex_state = 2},
  [4965] = {.lex_state = 9, .external_lex_state = 2},
  [4966] = {.lex_state = 9, .external_lex_state = 2},
  [4967] = {.lex_state = 9, .external_lex_state = 2},
  [4968] = {.lex_state = 23, .external_lex_state = 2},
  [4969] = {.lex_state = 9, .external_lex_state = 2},
  [4970] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4971] = {.lex_state = 9, .external_lex_state = 2},
  [4972] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4973] = {.lex_state = 9, .external_lex_state = 2},
  [4974] = {.lex_state = 151, .external_lex_state = 2},
  [4975] = {.lex_state = 9, .external_lex_state = 2},
  [4976] = {.lex_state = 9, .external_lex_state = 2},
  [4977] = {.lex_state = 9, .external_lex_state = 2},
  [4978] = {.lex_state = 140, .external_lex_state = 2},
  [4979] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4980] = {.lex_state = 9, .external_lex_state = 2},
  [4981] = {.lex_state = 9, .external_lex_state = 2},
  [4982] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4983] = {.lex_state = 138, .external_lex_state = 2},
  [4984] = {.lex_state = 9, .external_lex_state = 2},
  [4985] = {.lex_state = 63, .external_lex_state = 2},
  [4986] = {.lex_state = 63, .external_lex_state = 2},
  [4987] = {.lex_state = 9, .external_lex_state = 2},
  [4988] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4989] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4990] = {.lex_state = 9, .external_lex_state = 2},
  [4991] = {.lex_state = 9, .external_lex_state = 2},
  [4992] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4993] = {.lex_state = 9, .external_lex_state = 2},
  [4994] = {.lex_state = 9, .external_lex_state = 2},
  [4995] = {.lex_state = 151, .external_lex_state = 2},
  [4996] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4997] = {.lex_state = 9, .external_lex_state = 2},
  [4998] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [4999] = {.lex_state = 9, .external_lex_state = 2},
  [5000] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5001] = {.lex_state = 140, .external_lex_state = 2},
  [5002] = {.lex_state = 23, .external_lex_state = 2},
  [5003] = {.lex_state = 9, .external_lex_state = 2},
  [5004] = {.lex_state = 23, .external_lex_state = 2},
  [5005] = {.lex_state = 9, .external_lex_state = 2},
  [5006] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5007] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5008] = {.lex_state = 9, .external_lex_state = 2},
  [5009] = {.lex_state = 9, .external_lex_state = 2},
  [5010] = {.lex_state = 9, .external_lex_state = 2},
  [5011] = {.lex_state = 9, .external_lex_state = 2},
  [5012] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5013] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5014] = {.lex_state = 9, .external_lex_state = 2},
  [5015] = {.lex_state = 23, .external_lex_state = 2},
  [5016] = {.lex_state = 9, .external_lex_state = 2},
  [5017] = {.lex_state = 23, .external_lex_state = 2},
  [5018] = {.lex_state = 9, .external_lex_state = 2},
  [5019] = {.lex_state = 23, .external_lex_state = 2},
  [5020] = {.lex_state = 23, .external_lex_state = 2},
  [5021] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5022] = {.lex_state = 9, .external_lex_state = 2},
  [5023] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5024] = {.lex_state = 151, .external_lex_state = 2},
  [5025] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5026] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5027] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5028] = {.lex_state = 9, .external_lex_state = 2},
  [5029] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5030] = {.lex_state = 9, .external_lex_state = 2},
  [5031] = {.lex_state = 9, .external_lex_state = 2},
  [5032] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5033] = {.lex_state = 151, .external_lex_state = 2},
  [5034] = {.lex_state = 151, .external_lex_state = 2},
  [5035] = {.lex_state = 140, .external_lex_state = 2},
  [5036] = {.lex_state = 9, .external_lex_state = 2},
  [5037] = {.lex_state = 151, .external_lex_state = 2},
  [5038] = {.lex_state = 138, .external_lex_state = 2},
  [5039] = {.lex_state = 9, .external_lex_state = 2},
  [5040] = {.lex_state = 23, .external_lex_state = 2},
  [5041] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5042] = {.lex_state = 151, .external_lex_state = 2},
  [5043] = {.lex_state = 140, .external_lex_state = 3},
  [5044] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5045] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5046] = {.lex_state = 151, .external_lex_state = 2},
  [5047] = {.lex_state = 138, .external_lex_state = 2},
  [5048] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5049] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5050] = {.lex_state = 151, .external_lex_state = 2},
  [5051] = {.lex_state = 138, .external_lex_state = 2},
  [5052] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5053] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5054] = {.lex_state = 151, .external_lex_state = 2},
  [5055] = {.lex_state = 140, .external_lex_state = 2},
  [5056] = {.lex_state = 151, .external_lex_state = 2},
  [5057] = {.lex_state = 140, .external_lex_state = 2},
  [5058] = {.lex_state = 151, .external_lex_state = 2},
  [5059] = {.lex_state = 151, .external_lex_state = 2},
  [5060] = {.lex_state = 140, .external_lex_state = 2},
  [5061] = {.lex_state = 151, .external_lex_state = 2},
  [5062] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5063] = {.lex_state = 151, .external_lex_state = 2},
  [5064] = {.lex_state = 138, .external_lex_state = 2},
  [5065] = {.lex_state = 151, .external_lex_state = 2},
  [5066] = {.lex_state = 9, .external_lex_state = 2},
  [5067] = {.lex_state = 151, .external_lex_state = 2},
  [5068] = {.lex_state = 151, .external_lex_state = 2},
  [5069] = {.lex_state = 151, .external_lex_state = 2},
  [5070] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5071] = {.lex_state = 151, .external_lex_state = 2},
  [5072] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5073] = {.lex_state = 151, .external_lex_state = 2},
  [5074] = {.lex_state = 140, .external_lex_state = 2},
  [5075] = {.lex_state = 151, .external_lex_state = 2},
  [5076] = {.lex_state = 151, .external_lex_state = 2},
  [5077] = {.lex_state = 151, .external_lex_state = 2},
  [5078] = {.lex_state = 140, .external_lex_state = 2},
  [5079] = {.lex_state = 140, .external_lex_state = 2},
  [5080] = {.lex_state = 151, .external_lex_state = 2},
  [5081] = {.lex_state = 140, .external_lex_state = 2},
  [5082] = {.lex_state = 51, .external_lex_state = 2},
  [5083] = {.lex_state = 45, .external_lex_state = 2},
  [5084] = {.lex_state = 138, .external_lex_state = 2},
  [5085] = {.lex_state = 140, .external_lex_state = 2},
  [5086] = {.lex_state = 45, .external_lex_state = 2},
  [5087] = {.lex_state = 151, .external_lex_state = 2},
  [5088] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5089] = {.lex_state = 151, .external_lex_state = 2},
  [5090] = {.lex_state = 151, .external_lex_state = 2},
  [5091] = {.lex_state = 151, .external_lex_state = 2},
  [5092] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5093] = {.lex_state = 151, .external_lex_state = 2},
  [5094] = {.lex_state = 138, .external_lex_state = 2},
  [5095] = {.lex_state = 151, .external_lex_state = 2},
  [5096] = {.lex_state = 138, .external_lex_state = 2},
  [5097] = {.lex_state = 63, .external_lex_state = 2},
  [5098] = {.lex_state = 151, .external_lex_state = 2},
  [5099] = {.lex_state = 151, .external_lex_state = 2},
  [5100] = {.lex_state = 140, .external_lex_state = 3},
  [5101] = {.lex_state = 138, .external_lex_state = 2},
  [5102] = {.lex_state = 151, .external_lex_state = 2},
  [5103] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5104] = {.lex_state = 151, .external_lex_state = 2},
  [5105] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5106] = {.lex_state = 140, .external_lex_state = 2},
  [5107] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5108] = {.lex_state = 140, .external_lex_state = 2},
  [5109] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5110] = {.lex_state = 151, .external_lex_state = 2},
  [5111] = {.lex_state = 151, .external_lex_state = 2},
  [5112] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5113] = {.lex_state = 51, .external_lex_state = 2},
  [5114] = {.lex_state = 140, .external_lex_state = 2},
  [5115] = {.lex_state = 151, .external_lex_state = 2},
  [5116] = {.lex_state = 151, .external_lex_state = 2},
  [5117] = {.lex_state = 151, .external_lex_state = 2},
  [5118] = {.lex_state = 151, .external_lex_state = 2},
  [5119] = {.lex_state = 151, .external_lex_state = 2},
  [5120] = {.lex_state = 151, .external_lex_state = 2},
  [5121] = {.lex_state = 151, .external_lex_state = 2},
  [5122] = {.lex_state = 140, .external_lex_state = 2},
  [5123] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5124] = {.lex_state = 140, .external_lex_state = 2},
  [5125] = {.lex_state = 45, .external_lex_state = 2},
  [5126] = {.lex_state = 151, .external_lex_state = 2},
  [5127] = {.lex_state = 151, .external_lex_state = 2},
  [5128] = {.lex_state = 151, .external_lex_state = 2},
  [5129] = {.lex_state = 151, .external_lex_state = 2},
  [5130] = {.lex_state = 151, .external_lex_state = 2},
  [5131] = {.lex_state = 151, .external_lex_state = 2},
  [5132] = {.lex_state = 151, .external_lex_state = 2},
  [5133] = {.lex_state = 140, .external_lex_state = 2},
  [5134] = {.lex_state = 151, .external_lex_state = 2},
  [5135] = {.lex_state = 140, .external_lex_state = 2},
  [5136] = {.lex_state = 140, .external_lex_state = 3},
  [5137] = {.lex_state = 151, .external_lex_state = 2},
  [5138] = {.lex_state = 151, .external_lex_state = 2},
  [5139] = {.lex_state = 138, .external_lex_state = 2},
  [5140] = {.lex_state = 138, .external_lex_state = 2},
  [5141] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5142] = {.lex_state = 151, .external_lex_state = 2},
  [5143] = {.lex_state = 151, .external_lex_state = 2},
  [5144] = {.lex_state = 140, .external_lex_state = 2},
  [5145] = {.lex_state = 151, .external_lex_state = 2},
  [5146] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5147] = {.lex_state = 151, .external_lex_state = 2},
  [5148] = {.lex_state = 151, .external_lex_state = 2},
  [5149] = {.lex_state = 151, .external_lex_state = 2},
  [5150] = {.lex_state = 151, .external_lex_state = 2},
  [5151] = {.lex_state = 151, .external_lex_state = 2},
  [5152] = {.lex_state = 151, .external_lex_state = 2},
  [5153] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5154] = {.lex_state = 151, .external_lex_state = 2},
  [5155] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5156] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5157] = {.lex_state = 140, .external_lex_state = 3},
  [5158] = {.lex_state = 151, .external_lex_state = 2},
  [5159] = {.lex_state = 151, .external_lex_state = 2},
  [5160] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5161] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5162] = {.lex_state = 140, .external_lex_state = 2},
  [5163] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5164] = {.lex_state = 151, .external_lex_state = 2},
  [5165] = {.lex_state = 140, .external_lex_state = 2},
  [5166] = {.lex_state = 151, .external_lex_state = 2},
  [5167] = {.lex_state = 151, .external_lex_state = 2},
  [5168] = {.lex_state = 151, .external_lex_state = 2},
  [5169] = {.lex_state = 151, .external_lex_state = 2},
  [5170] = {.lex_state = 151, .external_lex_state = 2},
  [5171] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5172] = {.lex_state = 140, .external_lex_state = 3},
  [5173] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5174] = {.lex_state = 151, .external_lex_state = 2},
  [5175] = {.lex_state = 151, .external_lex_state = 2},
  [5176] = {.lex_state = 151, .external_lex_state = 2},
  [5177] = {.lex_state = 151, .external_lex_state = 2},
  [5178] = {.lex_state = 151, .external_lex_state = 2},
  [5179] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5180] = {.lex_state = 151, .external_lex_state = 2},
  [5181] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5182] = {.lex_state = 151, .external_lex_state = 2},
  [5183] = {.lex_state = 151, .external_lex_state = 2},
  [5184] = {.lex_state = 151, .external_lex_state = 2},
  [5185] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5186] = {.lex_state = 140, .external_lex_state = 3},
  [5187] = {.lex_state = 151, .external_lex_state = 2},
  [5188] = {.lex_state = 151, .external_lex_state = 2},
  [5189] = {.lex_state = 151, .external_lex_state = 2},
  [5190] = {.lex_state = 151, .external_lex_state = 2},
  [5191] = {.lex_state = 151, .external_lex_state = 2},
  [5192] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5193] = {.lex_state = 140, .external_lex_state = 2},
  [5194] = {.lex_state = 140, .external_lex_state = 3},
  [5195] = {.lex_state = 151, .external_lex_state = 2},
  [5196] = {.lex_state = 151, .external_lex_state = 2},
  [5197] = {.lex_state = 151, .external_lex_state = 2},
  [5198] = {.lex_state = 151, .external_lex_state = 2},
  [5199] = {.lex_state = 151, .external_lex_state = 2},
  [5200] = {.lex_state = 151, .external_lex_state = 2},
  [5201] = {.lex_state = 151, .external_lex_state = 2},
  [5202] = {.lex_state = 138, .external_lex_state = 2},
  [5203] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5204] = {.lex_state = 151, .external_lex_state = 2},
  [5205] = {.lex_state = 151, .external_lex_state = 2},
  [5206] = {.lex_state = 151, .external_lex_state = 2},
  [5207] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5208] = {.lex_state = 138, .external_lex_state = 2},
  [5209] = {.lex_state = 151, .external_lex_state = 2},
  [5210] = {.lex_state = 151, .external_lex_state = 2},
  [5211] = {.lex_state = 151, .external_lex_state = 2},
  [5212] = {.lex_state = 151, .external_lex_state = 2},
  [5213] = {.lex_state = 151, .external_lex_state = 2},
  [5214] = {.lex_state = 151, .external_lex_state = 2},
  [5215] = {.lex_state = 140, .external_lex_state = 2},
  [5216] = {.lex_state = 151, .external_lex_state = 2},
  [5217] = {.lex_state = 151, .external_lex_state = 2},
  [5218] = {.lex_state = 151, .external_lex_state = 2},
  [5219] = {.lex_state = 151, .external_lex_state = 2},
  [5220] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5221] = {.lex_state = 151, .external_lex_state = 2},
  [5222] = {.lex_state = 151, .external_lex_state = 2},
  [5223] = {.lex_state = 151, .external_lex_state = 2},
  [5224] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5225] = {.lex_state = 151, .external_lex_state = 2},
  [5226] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5227] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5228] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5229] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5230] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5231] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5232] = {.lex_state = 151, .external_lex_state = 2},
  [5233] = {.lex_state = 137, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5234] = {.lex_state = 151, .external_lex_state = 2},
  [5235] = {.lex_state = 51, .external_lex_state = 2},
  [5236] = {.lex_state = 138, .external_lex_state = 2},
  [5237] = {.lex_state = 138, .external_lex_state = 2},
  [5238] = {.lex_state = 151, .external_lex_state = 2},
  [5239] = {.lex_state = 151, .external_lex_state = 2},
  [5240] = {.lex_state = 151, .external_lex_state = 2},
  [5241] = {.lex_state = 151, .external_lex_state = 2},
  [5242] = {.lex_state = 138, .external_lex_state = 2},
  [5243] = {.lex_state = 151, .external_lex_state = 2},
  [5244] = {.lex_state = 138, .external_lex_state = 2},
  [5245] = {.lex_state = 138, .external_lex_state = 2},
  [5246] = {.lex_state = 138, .external_lex_state = 2},
  [5247] = {.lex_state = 138, .external_lex_state = 2},
  [5248] = {.lex_state = 138, .external_lex_state = 2},
  [5249] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5250] = {.lex_state = 140, .external_lex_state = 2},
  [5251] = {.lex_state = 151, .external_lex_state = 2},
  [5252] = {.lex_state = 151, .external_lex_state = 2},
  [5253] = {.lex_state = 140, .external_lex_state = 2},
  [5254] = {.lex_state = 140, .external_lex_state = 3},
  [5255] = {.lex_state = 151, .external_lex_state = 2},
  [5256] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5257] = {.lex_state = 138, .external_lex_state = 2},
  [5258] = {.lex_state = 151, .external_lex_state = 2},
  [5259] = {.lex_state = 140, .external_lex_state = 2},
  [5260] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5261] = {.lex_state = 151, .external_lex_state = 2},
  [5262] = {.lex_state = 151, .external_lex_state = 2},
  [5263] = {.lex_state = 151, .external_lex_state = 2},
  [5264] = {.lex_state = 151, .external_lex_state = 2},
  [5265] = {.lex_state = 46, .external_lex_state = 2},
  [5266] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5267] = {.lex_state = 140, .external_lex_state = 2},
  [5268] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5269] = {.lex_state = 46, .external_lex_state = 2},
  [5270] = {.lex_state = 140, .external_lex_state = 3},
  [5271] = {.lex_state = 151, .external_lex_state = 2},
  [5272] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5273] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5274] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5275] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5276] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5277] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5278] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5279] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5280] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5281] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5282] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5283] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5284] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5285] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 90},
  [5286] = {.lex_state = 151, .external_lex_state = 2},
  [5287] = {.lex_state = 151, .external_lex_state = 2},
  [5288] = {.lex_state = 23, .external_lex_state = 2},
  [5289] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5290] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5291] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5292] = {.lex_state = 140, .external_lex_state = 2},
  [5293] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5294] = {.lex_state = 140, .external_lex_state = 2},
  [5295] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5296] = {.lex_state = 140, .external_lex_state = 2},
  [5297] = {.lex_state = 45, .external_lex_state = 2},
  [5298] = {.lex_state = 45, .external_lex_state = 2},
  [5299] = {.lex_state = 151, .external_lex_state = 2},
  [5300] = {.lex_state = 151, .external_lex_state = 2},
  [5301] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5302] = {.lex_state = 151, .external_lex_state = 2},
  [5303] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5304] = {.lex_state = 46, .external_lex_state = 2},
  [5305] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5306] = {.lex_state = 140, .external_lex_state = 2},
  [5307] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5308] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5309] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5310] = {.lex_state = 151, .external_lex_state = 2},
  [5311] = {.lex_state = 140, .external_lex_state = 2},
  [5312] = {.lex_state = 140, .external_lex_state = 3},
  [5313] = {.lex_state = 140, .external_lex_state = 2},
  [5314] = {.lex_state = 140, .external_lex_state = 2},
  [5315] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5316] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5317] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5318] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5319] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5320] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5321] = {.lex_state = 138, .external_lex_state = 2},
  [5322] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5323] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5324] = {.lex_state = 23, .external_lex_state = 2},
  [5325] = {.lex_state = 140, .external_lex_state = 2},
  [5326] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5327] = {.lex_state = 23, .external_lex_state = 2},
  [5328] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5329] = {.lex_state = 140, .external_lex_state = 2},
  [5330] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5331] = {.lex_state = 151, .external_lex_state = 2},
  [5332] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5333] = {.lex_state = 140, .external_lex_state = 2},
  [5334] = {.lex_state = 151, .external_lex_state = 2},
  [5335] = {.lex_state = 151, .external_lex_state = 2},
  [5336] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5337] = {.lex_state = 138, .external_lex_state = 2},
  [5338] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5339] = {.lex_state = 140, .external_lex_state = 2},
  [5340] = {.lex_state = 45, .external_lex_state = 2},
  [5341] = {.lex_state = 140, .external_lex_state = 2},
  [5342] = {.lex_state = 138, .external_lex_state = 2},
  [5343] = {.lex_state = 140, .external_lex_state = 2},
  [5344] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5345] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5346] = {.lex_state = 45, .external_lex_state = 2},
  [5347] = {.lex_state = 45, .external_lex_state = 2},
  [5348] = {.lex_state = 140, .external_lex_state = 2},
  [5349] = {.lex_state = 140, .external_lex_state = 2},
  [5350] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5351] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5352] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5353] = {.lex_state = 45, .external_lex_state = 2},
  [5354] = {.lex_state = 140, .external_lex_state = 2},
  [5355] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5356] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5357] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5358] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5359] = {.lex_state = 140, .external_lex_state = 2},
  [5360] = {.lex_state = 45, .external_lex_state = 2},
  [5361] = {.lex_state = 45, .external_lex_state = 2},
  [5362] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5363] = {.lex_state = 151, .external_lex_state = 2},
  [5364] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 90},
  [5365] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5366] = {.lex_state = 140, .external_lex_state = 2},
  [5367] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5368] = {.lex_state = 45, .external_lex_state = 2},
  [5369] = {.lex_state = 140, .external_lex_state = 2},
  [5370] = {.lex_state = 140, .external_lex_state = 2},
  [5371] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5372] = {.lex_state = 23, .external_lex_state = 2},
  [5373] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 91},
  [5374] = {.lex_state = 140, .external_lex_state = 2},
  [5375] = {.lex_state = 64, .external_lex_state = 2},
  [5376] = {.lex_state = 64, .external_lex_state = 2},
  [5377] = {.lex_state = 140, .external_lex_state = 2},
  [5378] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 90},
  [5379] = {.lex_state = 140, .external_lex_state = 2},
  [5380] = {.lex_state = 140, .external_lex_state = 2},
  [5381] = {.lex_state = 23, .external_lex_state = 2},
  [5382] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5383] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5384] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5385] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5386] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5387] = {.lex_state = 151, .external_lex_state = 2},
  [5388] = {.lex_state = 140, .external_lex_state = 2},
  [5389] = {.lex_state = 140, .external_lex_state = 2},
  [5390] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5391] = {.lex_state = 46, .external_lex_state = 2},
  [5392] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5393] = {.lex_state = 46, .external_lex_state = 2},
  [5394] = {.lex_state = 140, .external_lex_state = 2},
  [5395] = {.lex_state = 140, .external_lex_state = 2},
  [5396] = {.lex_state = 138, .external_lex_state = 2},
  [5397] = {.lex_state = 138, .external_lex_state = 2},
  [5398] = {.lex_state = 138, .external_lex_state = 2},
  [5399] = {.lex_state = 151, .external_lex_state = 2},
  [5400] = {.lex_state = 46, .external_lex_state = 2},
  [5401] = {.lex_state = 140, .external_lex_state = 2},
  [5402] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5403] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5404] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5405] = {.lex_state = 138, .external_lex_state = 2},
  [5406] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5407] = {.lex_state = 140, .external_lex_state = 2},
  [5408] = {.lex_state = 63, .external_lex_state = 2},
  [5409] = {.lex_state = 140, .external_lex_state = 2},
  [5410] = {.lex_state = 140, .external_lex_state = 2},
  [5411] = {.lex_state = 138, .external_lex_state = 2},
  [5412] = {.lex_state = 138, .external_lex_state = 2},
  [5413] = {.lex_state = 138, .external_lex_state = 2},
  [5414] = {.lex_state = 140, .external_lex_state = 2},
  [5415] = {.lex_state = 140, .external_lex_state = 2},
  [5416] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 91},
  [5417] = {.lex_state = 64, .external_lex_state = 2},
  [5418] = {.lex_state = 138, .external_lex_state = 2},
  [5419] = {.lex_state = 138, .external_lex_state = 2},
  [5420] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5421] = {.lex_state = 140, .external_lex_state = 2},
  [5422] = {.lex_state = 138, .external_lex_state = 2},
  [5423] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5424] = {.lex_state = 138, .external_lex_state = 2},
  [5425] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5426] = {.lex_state = 138, .external_lex_state = 2},
  [5427] = {.lex_state = 45, .external_lex_state = 2},
  [5428] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5429] = {.lex_state = 140, .external_lex_state = 2},
  [5430] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5431] = {.lex_state = 140, .external_lex_state = 2},
  [5432] = {.lex_state = 45, .external_lex_state = 2},
  [5433] = {.lex_state = 45, .external_lex_state = 2},
  [5434] = {.lex_state = 138, .external_lex_state = 2},
  [5435] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5436] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5437] = {.lex_state = 138, .external_lex_state = 2},
  [5438] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5439] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5440] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5441] = {.lex_state = 23, .external_lex_state = 2},
  [5442] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 91},
  [5443] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5444] = {.lex_state = 46, .external_lex_state = 2},
  [5445] = {.lex_state = 140, .external_lex_state = 2},
  [5446] = {.lex_state = 46, .external_lex_state = 2},
  [5447] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5448] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5449] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5450] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5451] = {.lex_state = 140, .external_lex_state = 2},
  [5452] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5453] = {.lex_state = 46, .external_lex_state = 2},
  [5454] = {.lex_state = 46, .external_lex_state = 2},
  [5455] = {.lex_state = 46, .external_lex_state = 2},
  [5456] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5457] = {.lex_state = 46, .external_lex_state = 2},
  [5458] = {.lex_state = 46, .external_lex_state = 2},
  [5459] = {.lex_state = 138, .external_lex_state = 2},
  [5460] = {.lex_state = 140, .external_lex_state = 2},
  [5461] = {.lex_state = 138, .external_lex_state = 2},
  [5462] = {.lex_state = 89, .external_lex_state = 8},
  [5463] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5464] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5465] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5466] = {.lex_state = 140, .external_lex_state = 2},
  [5467] = {.lex_state = 140, .external_lex_state = 2},
  [5468] = {.lex_state = 45, .external_lex_state = 2},
  [5469] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5470] = {.lex_state = 138, .external_lex_state = 2},
  [5471] = {.lex_state = 138, .external_lex_state = 2},
  [5472] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 90},
  [5473] = {.lex_state = 140, .external_lex_state = 2},
  [5474] = {.lex_state = 89, .external_lex_state = 8},
  [5475] = {.lex_state = 140, .external_lex_state = 2},
  [5476] = {.lex_state = 138, .external_lex_state = 2},
  [5477] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5478] = {.lex_state = 140, .external_lex_state = 2},
  [5479] = {.lex_state = 46, .external_lex_state = 2},
  [5480] = {.lex_state = 46, .external_lex_state = 2},
  [5481] = {.lex_state = 46, .external_lex_state = 2},
  [5482] = {.lex_state = 46, .external_lex_state = 2},
  [5483] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5484] = {.lex_state = 140, .external_lex_state = 2},
  [5485] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5486] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 91},
  [5487] = {.lex_state = 151, .external_lex_state = 2},
  [5488] = {.lex_state = 46, .external_lex_state = 2},
  [5489] = {.lex_state = 46, .external_lex_state = 2},
  [5490] = {.lex_state = 46, .external_lex_state = 2},
  [5491] = {.lex_state = 151, .external_lex_state = 2},
  [5492] = {.lex_state = 46, .external_lex_state = 2},
  [5493] = {.lex_state = 140, .external_lex_state = 2},
  [5494] = {.lex_state = 140, .external_lex_state = 2},
  [5495] = {.lex_state = 140, .external_lex_state = 2},
  [5496] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5497] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5498] = {.lex_state = 23, .external_lex_state = 2},
  [5499] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 90},
  [5500] = {.lex_state = 140, .external_lex_state = 2},
  [5501] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5502] = {.lex_state = 140, .external_lex_state = 2},
  [5503] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5504] = {.lex_state = 140, .external_lex_state = 2},
  [5505] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5506] = {.lex_state = 140, .external_lex_state = 2},
  [5507] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 90},
  [5508] = {.lex_state = 140, .external_lex_state = 2},
  [5509] = {.lex_state = 140, .external_lex_state = 2},
  [5510] = {.lex_state = 89, .external_lex_state = 8},
  [5511] = {.lex_state = 23, .external_lex_state = 2},
  [5512] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5513] = {.lex_state = 140, .external_lex_state = 2},
  [5514] = {.lex_state = 140, .external_lex_state = 2},
  [5515] = {.lex_state = 140, .external_lex_state = 2},
  [5516] = {.lex_state = 140, .external_lex_state = 2},
  [5517] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5518] = {.lex_state = 145, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5519] = {.lex_state = 138, .external_lex_state = 2},
  [5520] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5521] = {.lex_state = 46, .external_lex_state = 2},
  [5522] = {.lex_state = 46, .external_lex_state = 2},
  [5523] = {.lex_state = 46, .external_lex_state = 2},
  [5524] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5525] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5526] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5527] = {.lex_state = 140, .external_lex_state = 2},
  [5528] = {.lex_state = 140, .external_lex_state = 2},
  [5529] = {.lex_state = 23, .external_lex_state = 2},
  [5530] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5531] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5532] = {.lex_state = 138, .external_lex_state = 2},
  [5533] = {.lex_state = 138, .external_lex_state = 2},
  [5534] = {.lex_state = 46, .external_lex_state = 2},
  [5535] = {.lex_state = 138, .external_lex_state = 2},
  [5536] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5537] = {.lex_state = 138, .external_lex_state = 2},
  [5538] = {.lex_state = 140, .external_lex_state = 3},
  [5539] = {.lex_state = 46, .external_lex_state = 2},
  [5540] = {.lex_state = 46, .external_lex_state = 2},
  [5541] = {.lex_state = 46, .external_lex_state = 2},
  [5542] = {.lex_state = 23, .external_lex_state = 2},
  [5543] = {.lex_state = 37, .external_lex_state = 2},
  [5544] = {.lex_state = 140, .external_lex_state = 2},
  [5545] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5546] = {.lex_state = 37, .external_lex_state = 2},
  [5547] = {.lex_state = 15, .external_lex_state = 2},
  [5548] = {.lex_state = 45, .external_lex_state = 2},
  [5549] = {.lex_state = 140, .external_lex_state = 2},
  [5550] = {.lex_state = 140, .external_lex_state = 2},
  [5551] = {.lex_state = 140, .external_lex_state = 2},
  [5552] = {.lex_state = 140, .external_lex_state = 2},
  [5553] = {.lex_state = 37, .external_lex_state = 2},
  [5554] = {.lex_state = 51, .external_lex_state = 2},
  [5555] = {.lex_state = 37, .external_lex_state = 2},
  [5556] = {.lex_state = 46, .external_lex_state = 2},
  [5557] = {.lex_state = 45, .external_lex_state = 2},
  [5558] = {.lex_state = 140, .external_lex_state = 2},
  [5559] = {.lex_state = 37, .external_lex_state = 2},
  [5560] = {.lex_state = 45, .external_lex_state = 2},
  [5561] = {.lex_state = 45, .external_lex_state = 2},
  [5562] = {.lex_state = 37, .external_lex_state = 2},
  [5563] = {.lex_state = 140, .external_lex_state = 2},
  [5564] = {.lex_state = 138, .external_lex_state = 2},
  [5565] = {.lex_state = 37, .external_lex_state = 2},
  [5566] = {.lex_state = 37, .external_lex_state = 2},
  [5567] = {.lex_state = 51, .external_lex_state = 2},
  [5568] = {.lex_state = 140, .external_lex_state = 2},
  [5569] = {.lex_state = 37, .external_lex_state = 2},
  [5570] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5571] = {.lex_state = 45, .external_lex_state = 2},
  [5572] = {.lex_state = 140, .external_lex_state = 2},
  [5573] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5574] = {.lex_state = 37, .external_lex_state = 2},
  [5575] = {.lex_state = 36, .external_lex_state = 2},
  [5576] = {.lex_state = 140, .external_lex_state = 2},
  [5577] = {.lex_state = 36, .external_lex_state = 2},
  [5578] = {.lex_state = 151, .external_lex_state = 2},
  [5579] = {.lex_state = 46, .external_lex_state = 2},
  [5580] = {.lex_state = 37, .external_lex_state = 2},
  [5581] = {.lex_state = 51, .external_lex_state = 2},
  [5582] = {.lex_state = 151, .external_lex_state = 2},
  [5583] = {.lex_state = 140, .external_lex_state = 2},
  [5584] = {.lex_state = 151, .external_lex_state = 2},
  [5585] = {.lex_state = 37, .external_lex_state = 2},
  [5586] = {.lex_state = 37, .external_lex_state = 2},
  [5587] = {.lex_state = 151, .external_lex_state = 2},
  [5588] = {.lex_state = 37, .external_lex_state = 2},
  [5589] = {.lex_state = 37, .external_lex_state = 2},
  [5590] = {.lex_state = 151, .external_lex_state = 2},
  [5591] = {.lex_state = 45, .external_lex_state = 2},
  [5592] = {.lex_state = 138, .external_lex_state = 2},
  [5593] = {.lex_state = 37, .external_lex_state = 2},
  [5594] = {.lex_state = 140, .external_lex_state = 2},
  [5595] = {.lex_state = 140, .external_lex_state = 2},
  [5596] = {.lex_state = 45, .external_lex_state = 2},
  [5597] = {.lex_state = 140, .external_lex_state = 2},
  [5598] = {.lex_state = 36, .external_lex_state = 2},
  [5599] = {.lex_state = 36, .external_lex_state = 2},
  [5600] = {.lex_state = 45, .external_lex_state = 2},
  [5601] = {.lex_state = 140, .external_lex_state = 2},
  [5602] = {.lex_state = 140, .external_lex_state = 2},
  [5603] = {.lex_state = 140, .external_lex_state = 2},
  [5604] = {.lex_state = 51, .external_lex_state = 2},
  [5605] = {.lex_state = 37, .external_lex_state = 2},
  [5606] = {.lex_state = 140, .external_lex_state = 2},
  [5607] = {.lex_state = 138, .external_lex_state = 2},
  [5608] = {.lex_state = 37, .external_lex_state = 2},
  [5609] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5610] = {.lex_state = 37, .external_lex_state = 2},
  [5611] = {.lex_state = 46, .external_lex_state = 2},
  [5612] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5613] = {.lex_state = 140, .external_lex_state = 2},
  [5614] = {.lex_state = 138, .external_lex_state = 2},
  [5615] = {.lex_state = 45, .external_lex_state = 2},
  [5616] = {.lex_state = 45, .external_lex_state = 2},
  [5617] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5618] = {.lex_state = 151, .external_lex_state = 2},
  [5619] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5620] = {.lex_state = 37, .external_lex_state = 2},
  [5621] = {.lex_state = 15, .external_lex_state = 2},
  [5622] = {.lex_state = 37, .external_lex_state = 2},
  [5623] = {.lex_state = 37, .external_lex_state = 2},
  [5624] = {.lex_state = 138, .external_lex_state = 2},
  [5625] = {.lex_state = 138, .external_lex_state = 2},
  [5626] = {.lex_state = 138, .external_lex_state = 2},
  [5627] = {.lex_state = 45, .external_lex_state = 2},
  [5628] = {.lex_state = 51, .external_lex_state = 2},
  [5629] = {.lex_state = 140, .external_lex_state = 2},
  [5630] = {.lex_state = 37, .external_lex_state = 2},
  [5631] = {.lex_state = 37, .external_lex_state = 2},
  [5632] = {.lex_state = 37, .external_lex_state = 2},
  [5633] = {.lex_state = 138, .external_lex_state = 2},
  [5634] = {.lex_state = 151, .external_lex_state = 2},
  [5635] = {.lex_state = 37, .external_lex_state = 2},
  [5636] = {.lex_state = 151, .external_lex_state = 2},
  [5637] = {.lex_state = 37, .external_lex_state = 2},
  [5638] = {.lex_state = 140, .external_lex_state = 2},
  [5639] = {.lex_state = 46, .external_lex_state = 2},
  [5640] = {.lex_state = 37, .external_lex_state = 2},
  [5641] = {.lex_state = 37, .external_lex_state = 2},
  [5642] = {.lex_state = 140, .external_lex_state = 2},
  [5643] = {.lex_state = 37, .external_lex_state = 2},
  [5644] = {.lex_state = 37, .external_lex_state = 2},
  [5645] = {.lex_state = 140, .external_lex_state = 2},
  [5646] = {.lex_state = 140, .external_lex_state = 2},
  [5647] = {.lex_state = 151, .external_lex_state = 2},
  [5648] = {.lex_state = 151, .external_lex_state = 2},
  [5649] = {.lex_state = 151, .external_lex_state = 2},
  [5650] = {.lex_state = 46, .external_lex_state = 2},
  [5651] = {.lex_state = 37, .external_lex_state = 2},
  [5652] = {.lex_state = 138, .external_lex_state = 2},
  [5653] = {.lex_state = 140, .external_lex_state = 2},
  [5654] = {.lex_state = 45, .external_lex_state = 2},
  [5655] = {.lex_state = 45, .external_lex_state = 2},
  [5656] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5657] = {.lex_state = 151, .external_lex_state = 2},
  [5658] = {.lex_state = 140, .external_lex_state = 2},
  [5659] = {.lex_state = 37, .external_lex_state = 2},
  [5660] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5661] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5662] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5663] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5664] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5665] = {.lex_state = 37, .external_lex_state = 2},
  [5666] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 91},
  [5667] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5668] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5669] = {.lex_state = 89, .external_lex_state = 8},
  [5670] = {.lex_state = 140, .external_lex_state = 2},
  [5671] = {.lex_state = 138, .external_lex_state = 2},
  [5672] = {.lex_state = 37, .external_lex_state = 2},
  [5673] = {.lex_state = 151, .external_lex_state = 2},
  [5674] = {.lex_state = 52, .external_lex_state = 2},
  [5675] = {.lex_state = 37, .external_lex_state = 2},
  [5676] = {.lex_state = 37, .external_lex_state = 2},
  [5677] = {.lex_state = 140, .external_lex_state = 2},
  [5678] = {.lex_state = 140, .external_lex_state = 2},
  [5679] = {.lex_state = 151, .external_lex_state = 2},
  [5680] = {.lex_state = 140, .external_lex_state = 2},
  [5681] = {.lex_state = 37, .external_lex_state = 2},
  [5682] = {.lex_state = 140, .external_lex_state = 2},
  [5683] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5684] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5685] = {.lex_state = 37, .external_lex_state = 2},
  [5686] = {.lex_state = 138, .external_lex_state = 2},
  [5687] = {.lex_state = 37, .external_lex_state = 2},
  [5688] = {.lex_state = 46, .external_lex_state = 2},
  [5689] = {.lex_state = 140, .external_lex_state = 2},
  [5690] = {.lex_state = 37, .external_lex_state = 2},
  [5691] = {.lex_state = 37, .external_lex_state = 2},
  [5692] = {.lex_state = 37, .external_lex_state = 2},
  [5693] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5694] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5695] = {.lex_state = 89, .external_lex_state = 8},
  [5696] = {.lex_state = 37, .external_lex_state = 2},
  [5697] = {.lex_state = 45, .external_lex_state = 2},
  [5698] = {.lex_state = 45, .external_lex_state = 2},
  [5699] = {.lex_state = 37, .external_lex_state = 2},
  [5700] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5701] = {.lex_state = 140, .external_lex_state = 2},
  [5702] = {.lex_state = 46, .external_lex_state = 2},
  [5703] = {.lex_state = 37, .external_lex_state = 2},
  [5704] = {.lex_state = 37, .external_lex_state = 2},
  [5705] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5706] = {.lex_state = 138, .external_lex_state = 2},
  [5707] = {.lex_state = 51, .external_lex_state = 2},
  [5708] = {.lex_state = 138, .external_lex_state = 2},
  [5709] = {.lex_state = 138, .external_lex_state = 2},
  [5710] = {.lex_state = 37, .external_lex_state = 2},
  [5711] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5712] = {.lex_state = 140, .external_lex_state = 2},
  [5713] = {.lex_state = 140, .external_lex_state = 2},
  [5714] = {.lex_state = 45, .external_lex_state = 2},
  [5715] = {.lex_state = 140, .external_lex_state = 2},
  [5716] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5717] = {.lex_state = 45, .external_lex_state = 2},
  [5718] = {.lex_state = 37, .external_lex_state = 2},
  [5719] = {.lex_state = 37, .external_lex_state = 2},
  [5720] = {.lex_state = 37, .external_lex_state = 2},
  [5721] = {.lex_state = 89, .external_lex_state = 8},
  [5722] = {.lex_state = 140, .external_lex_state = 2},
  [5723] = {.lex_state = 51, .external_lex_state = 2},
  [5724] = {.lex_state = 37, .external_lex_state = 2},
  [5725] = {.lex_state = 37, .external_lex_state = 2},
  [5726] = {.lex_state = 37, .external_lex_state = 2},
  [5727] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5728] = {.lex_state = 37, .external_lex_state = 2},
  [5729] = {.lex_state = 37, .external_lex_state = 2},
  [5730] = {.lex_state = 89, .external_lex_state = 8},
  [5731] = {.lex_state = 140, .external_lex_state = 2},
  [5732] = {.lex_state = 138, .external_lex_state = 2},
  [5733] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5734] = {.lex_state = 37, .external_lex_state = 2},
  [5735] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5736] = {.lex_state = 140, .external_lex_state = 2},
  [5737] = {.lex_state = 37, .external_lex_state = 2},
  [5738] = {.lex_state = 89, .external_lex_state = 8},
  [5739] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5740] = {.lex_state = 37, .external_lex_state = 2},
  [5741] = {.lex_state = 23, .external_lex_state = 2},
  [5742] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5743] = {.lex_state = 89, .external_lex_state = 8},
  [5744] = {.lex_state = 140, .external_lex_state = 2},
  [5745] = {.lex_state = 37, .external_lex_state = 2},
  [5746] = {.lex_state = 51, .external_lex_state = 2},
  [5747] = {.lex_state = 138, .external_lex_state = 2},
  [5748] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5749] = {.lex_state = 89, .external_lex_state = 8},
  [5750] = {.lex_state = 89, .external_lex_state = 8},
  [5751] = {.lex_state = 138, .external_lex_state = 2},
  [5752] = {.lex_state = 37, .external_lex_state = 2},
  [5753] = {.lex_state = 151, .external_lex_state = 2},
  [5754] = {.lex_state = 138, .external_lex_state = 2},
  [5755] = {.lex_state = 140, .external_lex_state = 2},
  [5756] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5757] = {.lex_state = 37, .external_lex_state = 2},
  [5758] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5759] = {.lex_state = 89, .external_lex_state = 8},
  [5760] = {.lex_state = 140, .external_lex_state = 2},
  [5761] = {.lex_state = 140, .external_lex_state = 2},
  [5762] = {.lex_state = 140, .external_lex_state = 2},
  [5763] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5764] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5765] = {.lex_state = 138, .external_lex_state = 2},
  [5766] = {.lex_state = 89, .external_lex_state = 8},
  [5767] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5768] = {.lex_state = 140, .external_lex_state = 2},
  [5769] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5770] = {.lex_state = 89, .external_lex_state = 8},
  [5771] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5772] = {.lex_state = 89, .external_lex_state = 8},
  [5773] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5774] = {.lex_state = 37, .external_lex_state = 2},
  [5775] = {.lex_state = 89, .external_lex_state = 8},
  [5776] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5777] = {.lex_state = 89, .external_lex_state = 8},
  [5778] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5779] = {.lex_state = 140, .external_lex_state = 2},
  [5780] = {.lex_state = 89, .external_lex_state = 8},
  [5781] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5782] = {.lex_state = 37, .external_lex_state = 2},
  [5783] = {.lex_state = 89, .external_lex_state = 8},
  [5784] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5785] = {.lex_state = 89, .external_lex_state = 8},
  [5786] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5787] = {.lex_state = 89, .external_lex_state = 8},
  [5788] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5789] = {.lex_state = 89, .external_lex_state = 8},
  [5790] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5791] = {.lex_state = 37, .external_lex_state = 2},
  [5792] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5793] = {.lex_state = 37, .external_lex_state = 2},
  [5794] = {.lex_state = 51, .external_lex_state = 2},
  [5795] = {.lex_state = 151, .external_lex_state = 2},
  [5796] = {.lex_state = 151, .external_lex_state = 2},
  [5797] = {.lex_state = 140, .external_lex_state = 2},
  [5798] = {.lex_state = 140, .external_lex_state = 2},
  [5799] = {.lex_state = 151, .external_lex_state = 2},
  [5800] = {.lex_state = 37, .external_lex_state = 2},
  [5801] = {.lex_state = 15, .external_lex_state = 2},
  [5802] = {.lex_state = 140, .external_lex_state = 2},
  [5803] = {.lex_state = 140, .external_lex_state = 2},
  [5804] = {.lex_state = 140, .external_lex_state = 2},
  [5805] = {.lex_state = 51, .external_lex_state = 2},
  [5806] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5807] = {.lex_state = 36, .external_lex_state = 2},
  [5808] = {.lex_state = 37, .external_lex_state = 2},
  [5809] = {.lex_state = 15, .external_lex_state = 2},
  [5810] = {.lex_state = 151, .external_lex_state = 2},
  [5811] = {.lex_state = 140, .external_lex_state = 2},
  [5812] = {.lex_state = 140, .external_lex_state = 2},
  [5813] = {.lex_state = 151, .external_lex_state = 2},
  [5814] = {.lex_state = 36, .external_lex_state = 2},
  [5815] = {.lex_state = 37, .external_lex_state = 2},
  [5816] = {.lex_state = 138, .external_lex_state = 2},
  [5817] = {.lex_state = 45, .external_lex_state = 2},
  [5818] = {.lex_state = 151, .external_lex_state = 2},
  [5819] = {.lex_state = 37, .external_lex_state = 2},
  [5820] = {.lex_state = 45, .external_lex_state = 2},
  [5821] = {.lex_state = 140, .external_lex_state = 2},
  [5822] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5823] = {.lex_state = 140, .external_lex_state = 2},
  [5824] = {.lex_state = 151, .external_lex_state = 2},
  [5825] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5826] = {.lex_state = 37, .external_lex_state = 2},
  [5827] = {.lex_state = 37, .external_lex_state = 2},
  [5828] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5829] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5830] = {.lex_state = 37, .external_lex_state = 2},
  [5831] = {.lex_state = 37, .external_lex_state = 2},
  [5832] = {.lex_state = 151, .external_lex_state = 2},
  [5833] = {.lex_state = 140, .external_lex_state = 2},
  [5834] = {.lex_state = 89, .external_lex_state = 8},
  [5835] = {.lex_state = 140, .external_lex_state = 2},
  [5836] = {.lex_state = 37, .external_lex_state = 2},
  [5837] = {.lex_state = 138, .external_lex_state = 2},
  [5838] = {.lex_state = 37, .external_lex_state = 2},
  [5839] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5840] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5841] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5842] = {.lex_state = 151, .external_lex_state = 2},
  [5843] = {.lex_state = 15, .external_lex_state = 2},
  [5844] = {.lex_state = 151, .external_lex_state = 2},
  [5845] = {.lex_state = 15, .external_lex_state = 2},
  [5846] = {.lex_state = 140, .external_lex_state = 2},
  [5847] = {.lex_state = 138, .external_lex_state = 2},
  [5848] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5849] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5850] = {.lex_state = 140, .external_lex_state = 2},
  [5851] = {.lex_state = 151, .external_lex_state = 2},
  [5852] = {.lex_state = 140, .external_lex_state = 2},
  [5853] = {.lex_state = 15, .external_lex_state = 2},
  [5854] = {.lex_state = 37, .external_lex_state = 2},
  [5855] = {.lex_state = 15, .external_lex_state = 2},
  [5856] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5857] = {.lex_state = 89, .external_lex_state = 8},
  [5858] = {.lex_state = 151, .external_lex_state = 2},
  [5859] = {.lex_state = 37, .external_lex_state = 2},
  [5860] = {.lex_state = 15, .external_lex_state = 2},
  [5861] = {.lex_state = 140, .external_lex_state = 2},
  [5862] = {.lex_state = 15, .external_lex_state = 2},
  [5863] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5864] = {.lex_state = 37, .external_lex_state = 2},
  [5865] = {.lex_state = 15, .external_lex_state = 2},
  [5866] = {.lex_state = 15, .external_lex_state = 2},
  [5867] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5868] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5869] = {.lex_state = 89, .external_lex_state = 8},
  [5870] = {.lex_state = 15, .external_lex_state = 2},
  [5871] = {.lex_state = 51, .external_lex_state = 2},
  [5872] = {.lex_state = 15, .external_lex_state = 2},
  [5873] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5874] = {.lex_state = 140, .external_lex_state = 2},
  [5875] = {.lex_state = 15, .external_lex_state = 2},
  [5876] = {.lex_state = 15, .external_lex_state = 2},
  [5877] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5878] = {.lex_state = 15, .external_lex_state = 2},
  [5879] = {.lex_state = 15, .external_lex_state = 2},
  [5880] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5881] = {.lex_state = 37, .external_lex_state = 2},
  [5882] = {.lex_state = 15, .external_lex_state = 2},
  [5883] = {.lex_state = 15, .external_lex_state = 2},
  [5884] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5885] = {.lex_state = 15, .external_lex_state = 2},
  [5886] = {.lex_state = 15, .external_lex_state = 2},
  [5887] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5888] = {.lex_state = 15, .external_lex_state = 2},
  [5889] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5890] = {.lex_state = 15, .external_lex_state = 2},
  [5891] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5892] = {.lex_state = 15, .external_lex_state = 2},
  [5893] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5894] = {.lex_state = 15, .external_lex_state = 2},
  [5895] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5896] = {.lex_state = 15, .external_lex_state = 2},
  [5897] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5898] = {.lex_state = 15, .external_lex_state = 2},
  [5899] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5900] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5901] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5902] = {.lex_state = 37, .external_lex_state = 2},
  [5903] = {.lex_state = 89, .external_lex_state = 8},
  [5904] = {.lex_state = 140, .external_lex_state = 2},
  [5905] = {.lex_state = 140, .external_lex_state = 2},
  [5906] = {.lex_state = 15, .external_lex_state = 2},
  [5907] = {.lex_state = 140, .external_lex_state = 2},
  [5908] = {.lex_state = 140, .external_lex_state = 2},
  [5909] = {.lex_state = 140, .external_lex_state = 2},
  [5910] = {.lex_state = 151, .external_lex_state = 2},
  [5911] = {.lex_state = 140, .external_lex_state = 2},
  [5912] = {.lex_state = 37, .external_lex_state = 2},
  [5913] = {.lex_state = 140, .external_lex_state = 2},
  [5914] = {.lex_state = 37, .external_lex_state = 2},
  [5915] = {.lex_state = 37, .external_lex_state = 2},
  [5916] = {.lex_state = 140, .external_lex_state = 2},
  [5917] = {.lex_state = 46, .external_lex_state = 2},
  [5918] = {.lex_state = 140, .external_lex_state = 2},
  [5919] = {.lex_state = 151, .external_lex_state = 2},
  [5920] = {.lex_state = 138, .external_lex_state = 2},
  [5921] = {.lex_state = 37, .external_lex_state = 2},
  [5922] = {.lex_state = 45, .external_lex_state = 2},
  [5923] = {.lex_state = 138, .external_lex_state = 2},
  [5924] = {.lex_state = 140, .external_lex_state = 2},
  [5925] = {.lex_state = 140, .external_lex_state = 2},
  [5926] = {.lex_state = 37, .external_lex_state = 2},
  [5927] = {.lex_state = 79, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5928] = {.lex_state = 140, .external_lex_state = 2},
  [5929] = {.lex_state = 140, .external_lex_state = 2},
  [5930] = {.lex_state = 37, .external_lex_state = 2},
  [5931] = {.lex_state = 140, .external_lex_state = 2},
  [5932] = {.lex_state = 140, .external_lex_state = 2},
  [5933] = {.lex_state = 140, .external_lex_state = 2},
  [5934] = {.lex_state = 51, .external_lex_state = 2},
  [5935] = {.lex_state = 140, .external_lex_state = 2},
  [5936] = {.lex_state = 37, .external_lex_state = 2},
  [5937] = {.lex_state = 37, .external_lex_state = 2},
  [5938] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 91},
  [5939] = {.lex_state = 138, .external_lex_state = 2},
  [5940] = {.lex_state = 140, .external_lex_state = 2},
  [5941] = {.lex_state = 37, .external_lex_state = 2},
  [5942] = {.lex_state = 151, .external_lex_state = 2},
  [5943] = {.lex_state = 151, .external_lex_state = 2},
  [5944] = {.lex_state = 37, .external_lex_state = 2},
  [5945] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5946] = {.lex_state = 140, .external_lex_state = 2},
  [5947] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5948] = {.lex_state = 140, .external_lex_state = 2},
  [5949] = {.lex_state = 140, .external_lex_state = 2},
  [5950] = {.lex_state = 37, .external_lex_state = 2},
  [5951] = {.lex_state = 46, .external_lex_state = 2},
  [5952] = {.lex_state = 37, .external_lex_state = 2},
  [5953] = {.lex_state = 140, .external_lex_state = 8},
  [5954] = {.lex_state = 140, .external_lex_state = 2},
  [5955] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5956] = {.lex_state = 140, .external_lex_state = 2},
  [5957] = {.lex_state = 140, .external_lex_state = 2},
  [5958] = {.lex_state = 37, .external_lex_state = 2},
  [5959] = {.lex_state = 37, .external_lex_state = 2},
  [5960] = {.lex_state = 37, .external_lex_state = 2},
  [5961] = {.lex_state = 46, .external_lex_state = 2},
  [5962] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5963] = {.lex_state = 46, .external_lex_state = 2},
  [5964] = {.lex_state = 140, .external_lex_state = 2},
  [5965] = {.lex_state = 140, .external_lex_state = 2},
  [5966] = {.lex_state = 140, .external_lex_state = 2},
  [5967] = {.lex_state = 46, .external_lex_state = 2},
  [5968] = {.lex_state = 140, .external_lex_state = 2},
  [5969] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5970] = {.lex_state = 140, .external_lex_state = 8},
  [5971] = {.lex_state = 46, .external_lex_state = 2},
  [5972] = {.lex_state = 140, .external_lex_state = 2},
  [5973] = {.lex_state = 140, .external_lex_state = 2},
  [5974] = {.lex_state = 140, .external_lex_state = 2},
  [5975] = {.lex_state = 138, .external_lex_state = 2},
  [5976] = {.lex_state = 46, .external_lex_state = 2},
  [5977] = {.lex_state = 37, .external_lex_state = 2},
  [5978] = {.lex_state = 37, .external_lex_state = 2},
  [5979] = {.lex_state = 37, .external_lex_state = 2},
  [5980] = {.lex_state = 31, .external_lex_state = 2},
  [5981] = {.lex_state = 140, .external_lex_state = 2},
  [5982] = {.lex_state = 45, .external_lex_state = 2},
  [5983] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5984] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5985] = {.lex_state = 37, .external_lex_state = 2},
  [5986] = {.lex_state = 37, .external_lex_state = 2},
  [5987] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5988] = {.lex_state = 140, .external_lex_state = 2},
  [5989] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5990] = {.lex_state = 37, .external_lex_state = 2},
  [5991] = {.lex_state = 140, .external_lex_state = 8},
  [5992] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5993] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5994] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [5995] = {.lex_state = 46, .external_lex_state = 2},
  [5996] = {.lex_state = 37, .external_lex_state = 2},
  [5997] = {.lex_state = 37, .external_lex_state = 2},
  [5998] = {.lex_state = 37, .external_lex_state = 2},
  [5999] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6000] = {.lex_state = 46, .external_lex_state = 2},
  [6001] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6002] = {.lex_state = 140, .external_lex_state = 2},
  [6003] = {.lex_state = 37, .external_lex_state = 2},
  [6004] = {.lex_state = 37, .external_lex_state = 2},
  [6005] = {.lex_state = 37, .external_lex_state = 2},
  [6006] = {.lex_state = 140, .external_lex_state = 2},
  [6007] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6008] = {.lex_state = 140, .external_lex_state = 8},
  [6009] = {.lex_state = 37, .external_lex_state = 2},
  [6010] = {.lex_state = 37, .external_lex_state = 2},
  [6011] = {.lex_state = 37, .external_lex_state = 2},
  [6012] = {.lex_state = 140, .external_lex_state = 2},
  [6013] = {.lex_state = 140, .external_lex_state = 8},
  [6014] = {.lex_state = 140, .external_lex_state = 2},
  [6015] = {.lex_state = 37, .external_lex_state = 2},
  [6016] = {.lex_state = 140, .external_lex_state = 8},
  [6017] = {.lex_state = 37, .external_lex_state = 2},
  [6018] = {.lex_state = 37, .external_lex_state = 2},
  [6019] = {.lex_state = 37, .external_lex_state = 2},
  [6020] = {.lex_state = 46, .external_lex_state = 2},
  [6021] = {.lex_state = 46, .external_lex_state = 2},
  [6022] = {.lex_state = 37, .external_lex_state = 2},
  [6023] = {.lex_state = 140, .external_lex_state = 8},
  [6024] = {.lex_state = 37, .external_lex_state = 2},
  [6025] = {.lex_state = 37, .external_lex_state = 2},
  [6026] = {.lex_state = 37, .external_lex_state = 2},
  [6027] = {.lex_state = 140, .external_lex_state = 2},
  [6028] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6029] = {.lex_state = 140, .external_lex_state = 2},
  [6030] = {.lex_state = 140, .external_lex_state = 8},
  [6031] = {.lex_state = 37, .external_lex_state = 2},
  [6032] = {.lex_state = 37, .external_lex_state = 2},
  [6033] = {.lex_state = 37, .external_lex_state = 2},
  [6034] = {.lex_state = 140, .external_lex_state = 2},
  [6035] = {.lex_state = 37, .external_lex_state = 2},
  [6036] = {.lex_state = 140, .external_lex_state = 8},
  [6037] = {.lex_state = 37, .external_lex_state = 2},
  [6038] = {.lex_state = 37, .external_lex_state = 2},
  [6039] = {.lex_state = 37, .external_lex_state = 2},
  [6040] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6041] = {.lex_state = 140, .external_lex_state = 8},
  [6042] = {.lex_state = 46, .external_lex_state = 2},
  [6043] = {.lex_state = 140, .external_lex_state = 8},
  [6044] = {.lex_state = 37, .external_lex_state = 2},
  [6045] = {.lex_state = 37, .external_lex_state = 2},
  [6046] = {.lex_state = 37, .external_lex_state = 2},
  [6047] = {.lex_state = 140, .external_lex_state = 2},
  [6048] = {.lex_state = 138, .external_lex_state = 2},
  [6049] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6050] = {.lex_state = 140, .external_lex_state = 8},
  [6051] = {.lex_state = 37, .external_lex_state = 2},
  [6052] = {.lex_state = 37, .external_lex_state = 2},
  [6053] = {.lex_state = 37, .external_lex_state = 2},
  [6054] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6055] = {.lex_state = 140, .external_lex_state = 2},
  [6056] = {.lex_state = 140, .external_lex_state = 8},
  [6057] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6058] = {.lex_state = 140, .external_lex_state = 2},
  [6059] = {.lex_state = 140, .external_lex_state = 8},
  [6060] = {.lex_state = 140, .external_lex_state = 2},
  [6061] = {.lex_state = 140, .external_lex_state = 2},
  [6062] = {.lex_state = 140, .external_lex_state = 8},
  [6063] = {.lex_state = 37, .external_lex_state = 2},
  [6064] = {.lex_state = 140, .external_lex_state = 2},
  [6065] = {.lex_state = 46, .external_lex_state = 2},
  [6066] = {.lex_state = 140, .external_lex_state = 8},
  [6067] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [6068] = {.lex_state = 140, .external_lex_state = 2},
  [6069] = {.lex_state = 140, .external_lex_state = 8},
  [6070] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6071] = {.lex_state = 140, .external_lex_state = 2},
  [6072] = {.lex_state = 138, .external_lex_state = 2},
  [6073] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6074] = {.lex_state = 140, .external_lex_state = 2},
  [6075] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6076] = {.lex_state = 51, .external_lex_state = 2},
  [6077] = {.lex_state = 140, .external_lex_state = 2},
  [6078] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6079] = {.lex_state = 46, .external_lex_state = 2},
  [6080] = {.lex_state = 46, .external_lex_state = 2},
  [6081] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6082] = {.lex_state = 140, .external_lex_state = 2},
  [6083] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6084] = {.lex_state = 140, .external_lex_state = 2},
  [6085] = {.lex_state = 140, .external_lex_state = 2},
  [6086] = {.lex_state = 46, .external_lex_state = 2},
  [6087] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6088] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6089] = {.lex_state = 140, .external_lex_state = 2},
  [6090] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6091] = {.lex_state = 37, .external_lex_state = 2},
  [6092] = {.lex_state = 140, .external_lex_state = 2},
  [6093] = {.lex_state = 37, .external_lex_state = 2},
  [6094] = {.lex_state = 46, .external_lex_state = 2},
  [6095] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6096] = {.lex_state = 37, .external_lex_state = 2},
  [6097] = {.lex_state = 37, .external_lex_state = 2},
  [6098] = {.lex_state = 140, .external_lex_state = 2},
  [6099] = {.lex_state = 140, .external_lex_state = 2},
  [6100] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6101] = {.lex_state = 140, .external_lex_state = 2},
  [6102] = {.lex_state = 46, .external_lex_state = 2},
  [6103] = {.lex_state = 140, .external_lex_state = 2},
  [6104] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6105] = {.lex_state = 140, .external_lex_state = 2},
  [6106] = {.lex_state = 37, .external_lex_state = 2},
  [6107] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6108] = {.lex_state = 140, .external_lex_state = 2},
  [6109] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6110] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6111] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6112] = {.lex_state = 46, .external_lex_state = 2},
  [6113] = {.lex_state = 37, .external_lex_state = 2},
  [6114] = {.lex_state = 37, .external_lex_state = 2},
  [6115] = {.lex_state = 140, .external_lex_state = 2},
  [6116] = {.lex_state = 140, .external_lex_state = 2},
  [6117] = {.lex_state = 140, .external_lex_state = 2},
  [6118] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6119] = {.lex_state = 46, .external_lex_state = 2},
  [6120] = {.lex_state = 140, .external_lex_state = 2},
  [6121] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6122] = {.lex_state = 140, .external_lex_state = 2},
  [6123] = {.lex_state = 46, .external_lex_state = 2},
  [6124] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6125] = {.lex_state = 37, .external_lex_state = 2},
  [6126] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6127] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6128] = {.lex_state = 140, .external_lex_state = 2},
  [6129] = {.lex_state = 46, .external_lex_state = 2},
  [6130] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [6131] = {.lex_state = 46, .external_lex_state = 2},
  [6132] = {.lex_state = 46, .external_lex_state = 2},
  [6133] = {.lex_state = 37, .external_lex_state = 2},
  [6134] = {.lex_state = 140, .external_lex_state = 2},
  [6135] = {.lex_state = 140, .external_lex_state = 2},
  [6136] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6137] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6138] = {.lex_state = 140, .external_lex_state = 2},
  [6139] = {.lex_state = 46, .external_lex_state = 2},
  [6140] = {.lex_state = 37, .external_lex_state = 2},
  [6141] = {.lex_state = 46, .external_lex_state = 2},
  [6142] = {.lex_state = 140, .external_lex_state = 2},
  [6143] = {.lex_state = 140, .external_lex_state = 2},
  [6144] = {.lex_state = 37, .external_lex_state = 2},
  [6145] = {.lex_state = 140, .external_lex_state = 2},
  [6146] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6147] = {.lex_state = 31, .external_lex_state = 2},
  [6148] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6149] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6150] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6151] = {.lex_state = 46, .external_lex_state = 2},
  [6152] = {.lex_state = 140, .external_lex_state = 2},
  [6153] = {.lex_state = 46, .external_lex_state = 2},
  [6154] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6155] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6156] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6157] = {.lex_state = 46, .external_lex_state = 2},
  [6158] = {.lex_state = 46, .external_lex_state = 2},
  [6159] = {.lex_state = 140, .external_lex_state = 2},
  [6160] = {.lex_state = 37, .external_lex_state = 2},
  [6161] = {.lex_state = 37, .external_lex_state = 2},
  [6162] = {.lex_state = 46, .external_lex_state = 2},
  [6163] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6164] = {.lex_state = 140, .external_lex_state = 2},
  [6165] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6166] = {.lex_state = 45, .external_lex_state = 2},
  [6167] = {.lex_state = 140, .external_lex_state = 2},
  [6168] = {.lex_state = 140, .external_lex_state = 2},
  [6169] = {.lex_state = 140, .external_lex_state = 2},
  [6170] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6171] = {.lex_state = 140, .external_lex_state = 2},
  [6172] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6173] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6174] = {.lex_state = 140, .external_lex_state = 2},
  [6175] = {.lex_state = 140, .external_lex_state = 2},
  [6176] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6177] = {.lex_state = 140, .external_lex_state = 2},
  [6178] = {.lex_state = 140, .external_lex_state = 2},
  [6179] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6180] = {.lex_state = 140, .external_lex_state = 2},
  [6181] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6182] = {.lex_state = 140, .external_lex_state = 2},
  [6183] = {.lex_state = 140, .external_lex_state = 2},
  [6184] = {.lex_state = 140, .external_lex_state = 2},
  [6185] = {.lex_state = 151, .external_lex_state = 2},
  [6186] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6187] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6188] = {.lex_state = 140, .external_lex_state = 2},
  [6189] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6190] = {.lex_state = 140, .external_lex_state = 2},
  [6191] = {.lex_state = 46, .external_lex_state = 2},
  [6192] = {.lex_state = 140, .external_lex_state = 2},
  [6193] = {.lex_state = 140, .external_lex_state = 8},
  [6194] = {.lex_state = 140, .external_lex_state = 8},
  [6195] = {.lex_state = 140, .external_lex_state = 2},
  [6196] = {.lex_state = 140, .external_lex_state = 2},
  [6197] = {.lex_state = 140, .external_lex_state = 2},
  [6198] = {.lex_state = 46, .external_lex_state = 2},
  [6199] = {.lex_state = 140, .external_lex_state = 2},
  [6200] = {.lex_state = 140, .external_lex_state = 2},
  [6201] = {.lex_state = 140, .external_lex_state = 2},
  [6202] = {.lex_state = 153, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6203] = {.lex_state = 46, .external_lex_state = 2},
  [6204] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6205] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6206] = {.lex_state = 37, .external_lex_state = 2},
  [6207] = {.lex_state = 140, .external_lex_state = 2},
  [6208] = {.lex_state = 140, .external_lex_state = 2},
  [6209] = {.lex_state = 140, .external_lex_state = 2},
  [6210] = {.lex_state = 37, .external_lex_state = 2},
  [6211] = {.lex_state = 140, .external_lex_state = 2},
  [6212] = {.lex_state = 37, .external_lex_state = 2},
  [6213] = {.lex_state = 140, .external_lex_state = 8},
  [6214] = {.lex_state = 140, .external_lex_state = 2},
  [6215] = {.lex_state = 140, .external_lex_state = 2},
  [6216] = {.lex_state = 46, .external_lex_state = 2},
  [6217] = {.lex_state = 45, .external_lex_state = 2},
  [6218] = {.lex_state = 151, .external_lex_state = 2},
  [6219] = {.lex_state = 140, .external_lex_state = 2},
  [6220] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6221] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6222] = {.lex_state = 140, .external_lex_state = 2},
  [6223] = {.lex_state = 140, .external_lex_state = 2},
  [6224] = {.lex_state = 37, .external_lex_state = 2},
  [6225] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6226] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6227] = {.lex_state = 140, .external_lex_state = 2},
  [6228] = {.lex_state = 140, .external_lex_state = 8},
  [6229] = {.lex_state = 140, .external_lex_state = 2},
  [6230] = {.lex_state = 140, .external_lex_state = 2},
  [6231] = {.lex_state = 31, .external_lex_state = 2},
  [6232] = {.lex_state = 140, .external_lex_state = 2},
  [6233] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6234] = {.lex_state = 140, .external_lex_state = 2},
  [6235] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6236] = {.lex_state = 140, .external_lex_state = 2},
  [6237] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6238] = {.lex_state = 46, .external_lex_state = 2},
  [6239] = {.lex_state = 140, .external_lex_state = 2},
  [6240] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6241] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6242] = {.lex_state = 37, .external_lex_state = 2},
  [6243] = {.lex_state = 140, .external_lex_state = 2},
  [6244] = {.lex_state = 46, .external_lex_state = 2},
  [6245] = {.lex_state = 46, .external_lex_state = 2},
  [6246] = {.lex_state = 140, .external_lex_state = 2},
  [6247] = {.lex_state = 140, .external_lex_state = 2},
  [6248] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6249] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6250] = {.lex_state = 140, .external_lex_state = 2},
  [6251] = {.lex_state = 140, .external_lex_state = 2},
  [6252] = {.lex_state = 140, .external_lex_state = 2},
  [6253] = {.lex_state = 140, .external_lex_state = 2},
  [6254] = {.lex_state = 140, .external_lex_state = 2},
  [6255] = {.lex_state = 37, .external_lex_state = 2},
  [6256] = {.lex_state = 140, .external_lex_state = 2},
  [6257] = {.lex_state = 140, .external_lex_state = 2},
  [6258] = {.lex_state = 151, .external_lex_state = 2},
  [6259] = {.lex_state = 140, .external_lex_state = 2},
  [6260] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6261] = {.lex_state = 140, .external_lex_state = 2},
  [6262] = {.lex_state = 46, .external_lex_state = 2},
  [6263] = {.lex_state = 140, .external_lex_state = 8},
  [6264] = {.lex_state = 140, .external_lex_state = 8},
  [6265] = {.lex_state = 46, .external_lex_state = 2},
  [6266] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6267] = {.lex_state = 140, .external_lex_state = 2},
  [6268] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6269] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6270] = {.lex_state = 140, .external_lex_state = 2},
  [6271] = {.lex_state = 140, .external_lex_state = 2},
  [6272] = {.lex_state = 140, .external_lex_state = 2},
  [6273] = {.lex_state = 140, .external_lex_state = 2},
  [6274] = {.lex_state = 37, .external_lex_state = 2},
  [6275] = {.lex_state = 140, .external_lex_state = 2},
  [6276] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6277] = {.lex_state = 37, .external_lex_state = 2},
  [6278] = {.lex_state = 37, .external_lex_state = 2},
  [6279] = {.lex_state = 37, .external_lex_state = 2},
  [6280] = {.lex_state = 140, .external_lex_state = 2},
  [6281] = {.lex_state = 37, .external_lex_state = 2},
  [6282] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6283] = {.lex_state = 46, .external_lex_state = 2},
  [6284] = {.lex_state = 140, .external_lex_state = 2},
  [6285] = {.lex_state = 140, .external_lex_state = 2},
  [6286] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6287] = {.lex_state = 140, .external_lex_state = 2},
  [6288] = {.lex_state = 140, .external_lex_state = 2},
  [6289] = {.lex_state = 46, .external_lex_state = 2},
  [6290] = {.lex_state = 140, .external_lex_state = 2},
  [6291] = {.lex_state = 140, .external_lex_state = 8},
  [6292] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6293] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6294] = {.lex_state = 140, .external_lex_state = 8},
  [6295] = {.lex_state = 140, .external_lex_state = 8},
  [6296] = {.lex_state = 140, .external_lex_state = 2, .reserved_word_set_id = 92},
  [6297] = {.lex_state = 37, .external_lex_state = 2},
  [6298] = {.lex_state = 151, .external_lex_state = 2},
  [6299] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6300] = {.lex_state = 140, .external_lex_state = 2},
  [6301] = {.lex_state = 37, .external_lex_state = 2},
  [6302] = {.lex_state = 37, .external_lex_state = 2},
  [6303] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6304] = {.lex_state = 37, .external_lex_state = 2},
  [6305] = {.lex_state = 140, .external_lex_state = 2},
  [6306] = {.lex_state = 140, .external_lex_state = 8},
  [6307] = {.lex_state = 140, .external_lex_state = 2},
  [6308] = {.lex_state = 46, .external_lex_state = 2},
  [6309] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6310] = {.lex_state = 46, .external_lex_state = 2},
  [6311] = {.lex_state = 46, .external_lex_state = 2},
  [6312] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6313] = {.lex_state = 140, .external_lex_state = 2},
  [6314] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6315] = {.lex_state = 46, .external_lex_state = 2},
  [6316] = {.lex_state = 140, .external_lex_state = 2},
  [6317] = {.lex_state = 140, .external_lex_state = 8},
  [6318] = {.lex_state = 46, .external_lex_state = 2},
  [6319] = {.lex_state = 46, .external_lex_state = 2},
  [6320] = {.lex_state = 37, .external_lex_state = 2},
  [6321] = {.lex_state = 140, .external_lex_state = 2},
  [6322] = {.lex_state = 140, .external_lex_state = 2},
  [6323] = {.lex_state = 140, .external_lex_state = 8},
  [6324] = {.lex_state = 140, .external_lex_state = 8},
  [6325] = {.lex_state = 37, .external_lex_state = 2},
  [6326] = {.lex_state = 138, .external_lex_state = 2},
  [6327] = {.lex_state = 46, .external_lex_state = 2},
  [6328] = {.lex_state = 37, .external_lex_state = 2},
  [6329] = {.lex_state = 140, .external_lex_state = 2},
  [6330] = {.lex_state = 140, .external_lex_state = 2},
  [6331] = {.lex_state = 46, .external_lex_state = 2},
  [6332] = {.lex_state = 37, .external_lex_state = 2},
  [6333] = {.lex_state = 51, .external_lex_state = 2},
  [6334] = {.lex_state = 37, .external_lex_state = 2},
  [6335] = {.lex_state = 140, .external_lex_state = 2},
  [6336] = {.lex_state = 140, .external_lex_state = 2},
  [6337] = {.lex_state = 140, .external_lex_state = 2},
  [6338] = {.lex_state = 140, .external_lex_state = 2},
  [6339] = {.lex_state = 140, .external_lex_state = 2},
  [6340] = {.lex_state = 46, .external_lex_state = 2},
  [6341] = {.lex_state = 140, .external_lex_state = 2},
  [6342] = {.lex_state = 140, .external_lex_state = 2},
  [6343] = {.lex_state = 140, .external_lex_state = 2},
  [6344] = {.lex_state = 15, .external_lex_state = 2},
  [6345] = {.lex_state = 140, .external_lex_state = 3},
  [6346] = {.lex_state = 140, .external_lex_state = 2},
  [6347] = {.lex_state = 140, .external_lex_state = 2},
  [6348] = {.lex_state = 46, .external_lex_state = 2},
  [6349] = {.lex_state = 140, .external_lex_state = 2},
  [6350] = {.lex_state = 46, .external_lex_state = 2},
  [6351] = {.lex_state = 140, .external_lex_state = 2},
  [6352] = {.lex_state = 140, .external_lex_state = 2},
  [6353] = {.lex_state = 140, .external_lex_state = 2},
  [6354] = {.lex_state = 140, .external_lex_state = 2},
  [6355] = {.lex_state = 140, .external_lex_state = 2},
  [6356] = {.lex_state = 140, .external_lex_state = 2},
  [6357] = {.lex_state = 140, .external_lex_state = 2},
  [6358] = {.lex_state = 140, .external_lex_state = 2},
  [6359] = {.lex_state = 140, .external_lex_state = 2},
  [6360] = {.lex_state = 15, .external_lex_state = 2},
  [6361] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6362] = {.lex_state = 140, .external_lex_state = 3},
  [6363] = {.lex_state = 140, .external_lex_state = 2},
  [6364] = {.lex_state = 140, .external_lex_state = 2},
  [6365] = {.lex_state = 140, .external_lex_state = 2},
  [6366] = {.lex_state = 140, .external_lex_state = 2},
  [6367] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6368] = {.lex_state = 46, .external_lex_state = 2},
  [6369] = {.lex_state = 140, .external_lex_state = 2},
  [6370] = {.lex_state = 140, .external_lex_state = 2},
  [6371] = {.lex_state = 140, .external_lex_state = 2},
  [6372] = {.lex_state = 140, .external_lex_state = 2},
  [6373] = {.lex_state = 140, .external_lex_state = 2},
  [6374] = {.lex_state = 140, .external_lex_state = 2},
  [6375] = {.lex_state = 140, .external_lex_state = 2},
  [6376] = {.lex_state = 140, .external_lex_state = 2},
  [6377] = {.lex_state = 140, .external_lex_state = 2},
  [6378] = {.lex_state = 140, .external_lex_state = 2},
  [6379] = {.lex_state = 140, .external_lex_state = 2},
  [6380] = {.lex_state = 46, .external_lex_state = 2},
  [6381] = {.lex_state = 140, .external_lex_state = 2},
  [6382] = {.lex_state = 140, .external_lex_state = 2},
  [6383] = {.lex_state = 140, .external_lex_state = 2},
  [6384] = {.lex_state = 140, .external_lex_state = 2},
  [6385] = {.lex_state = 15, .external_lex_state = 2},
  [6386] = {.lex_state = 46, .external_lex_state = 2},
  [6387] = {.lex_state = 46, .external_lex_state = 2},
  [6388] = {.lex_state = 140, .external_lex_state = 2},
  [6389] = {.lex_state = 45, .external_lex_state = 2},
  [6390] = {.lex_state = 46, .external_lex_state = 2},
  [6391] = {.lex_state = 140, .external_lex_state = 2},
  [6392] = {.lex_state = 46, .external_lex_state = 2},
  [6393] = {.lex_state = 140, .external_lex_state = 2},
  [6394] = {.lex_state = 140, .external_lex_state = 2},
  [6395] = {.lex_state = 140, .external_lex_state = 2},
  [6396] = {.lex_state = 140, .external_lex_state = 2},
  [6397] = {.lex_state = 140, .external_lex_state = 2},
  [6398] = {.lex_state = 140, .external_lex_state = 2},
  [6399] = {.lex_state = 140, .external_lex_state = 3},
  [6400] = {.lex_state = 140, .external_lex_state = 2},
  [6401] = {.lex_state = 140, .external_lex_state = 2},
  [6402] = {.lex_state = 140, .external_lex_state = 2},
  [6403] = {.lex_state = 140, .external_lex_state = 2},
  [6404] = {.lex_state = 140, .external_lex_state = 2},
  [6405] = {.lex_state = 140, .external_lex_state = 2},
  [6406] = {.lex_state = 140, .external_lex_state = 2},
  [6407] = {.lex_state = 140, .external_lex_state = 2},
  [6408] = {.lex_state = 140, .external_lex_state = 2},
  [6409] = {.lex_state = 140, .external_lex_state = 2},
  [6410] = {.lex_state = 46, .external_lex_state = 2},
  [6411] = {.lex_state = 140, .external_lex_state = 2},
  [6412] = {.lex_state = 140, .external_lex_state = 2},
  [6413] = {.lex_state = 140, .external_lex_state = 2},
  [6414] = {.lex_state = 140, .external_lex_state = 2},
  [6415] = {.lex_state = 46, .external_lex_state = 2},
  [6416] = {.lex_state = 140, .external_lex_state = 2},
  [6417] = {.lex_state = 140, .external_lex_state = 2},
  [6418] = {.lex_state = 140, .external_lex_state = 2},
  [6419] = {.lex_state = 140, .external_lex_state = 2},
  [6420] = {.lex_state = 140, .external_lex_state = 2},
  [6421] = {.lex_state = 140, .external_lex_state = 2},
  [6422] = {.lex_state = 140, .external_lex_state = 2},
  [6423] = {.lex_state = 140, .external_lex_state = 2},
  [6424] = {.lex_state = 140, .external_lex_state = 2},
  [6425] = {.lex_state = 140, .external_lex_state = 2},
  [6426] = {.lex_state = 140, .external_lex_state = 2},
  [6427] = {.lex_state = 140, .external_lex_state = 2},
  [6428] = {.lex_state = 46, .external_lex_state = 2},
  [6429] = {.lex_state = 46, .external_lex_state = 2},
  [6430] = {.lex_state = 140, .external_lex_state = 2},
  [6431] = {.lex_state = 140, .external_lex_state = 2},
  [6432] = {.lex_state = 140, .external_lex_state = 2},
  [6433] = {.lex_state = 140, .external_lex_state = 2},
  [6434] = {.lex_state = 46, .external_lex_state = 2},
  [6435] = {.lex_state = 140, .external_lex_state = 2},
  [6436] = {.lex_state = 140, .external_lex_state = 2},
  [6437] = {.lex_state = 140, .external_lex_state = 2},
  [6438] = {.lex_state = 140, .external_lex_state = 2},
  [6439] = {.lex_state = 140, .external_lex_state = 2},
  [6440] = {.lex_state = 46, .external_lex_state = 2},
  [6441] = {.lex_state = 140, .external_lex_state = 2},
  [6442] = {.lex_state = 140, .external_lex_state = 2},
  [6443] = {.lex_state = 140, .external_lex_state = 2},
  [6444] = {.lex_state = 140, .external_lex_state = 2},
  [6445] = {.lex_state = 140, .external_lex_state = 2},
  [6446] = {.lex_state = 140, .external_lex_state = 2},
  [6447] = {.lex_state = 46, .external_lex_state = 2},
  [6448] = {.lex_state = 140, .external_lex_state = 2},
  [6449] = {.lex_state = 140, .external_lex_state = 2},
  [6450] = {.lex_state = 140, .external_lex_state = 2},
  [6451] = {.lex_state = 140, .external_lex_state = 2},
  [6452] = {.lex_state = 140, .external_lex_state = 2},
  [6453] = {.lex_state = 46, .external_lex_state = 2},
  [6454] = {.lex_state = 140, .external_lex_state = 2},
  [6455] = {.lex_state = 140, .external_lex_state = 2},
  [6456] = {.lex_state = 140, .external_lex_state = 2},
  [6457] = {.lex_state = 140, .external_lex_state = 2},
  [6458] = {.lex_state = 140, .external_lex_state = 2},
  [6459] = {.lex_state = 46, .external_lex_state = 2},
  [6460] = {.lex_state = 140, .external_lex_state = 2},
  [6461] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6462] = {.lex_state = 140, .external_lex_state = 2},
  [6463] = {.lex_state = 140, .external_lex_state = 2},
  [6464] = {.lex_state = 140, .external_lex_state = 2},
  [6465] = {.lex_state = 140, .external_lex_state = 2},
  [6466] = {.lex_state = 46, .external_lex_state = 2},
  [6467] = {.lex_state = 140, .external_lex_state = 2},
  [6468] = {.lex_state = 46, .external_lex_state = 2},
  [6469] = {.lex_state = 140, .external_lex_state = 2},
  [6470] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6471] = {.lex_state = 140, .external_lex_state = 2},
  [6472] = {.lex_state = 140, .external_lex_state = 2},
  [6473] = {.lex_state = 46, .external_lex_state = 2},
  [6474] = {.lex_state = 140, .external_lex_state = 2},
  [6475] = {.lex_state = 140, .external_lex_state = 2},
  [6476] = {.lex_state = 46, .external_lex_state = 2},
  [6477] = {.lex_state = 15, .external_lex_state = 2},
  [6478] = {.lex_state = 154, .external_lex_state = 2},
  [6479] = {.lex_state = 140, .external_lex_state = 2},
  [6480] = {.lex_state = 140, .external_lex_state = 2},
  [6481] = {.lex_state = 140, .external_lex_state = 2},
  [6482] = {.lex_state = 140, .external_lex_state = 3},
  [6483] = {.lex_state = 140, .external_lex_state = 2},
  [6484] = {.lex_state = 140, .external_lex_state = 2},
  [6485] = {.lex_state = 46, .external_lex_state = 2},
  [6486] = {.lex_state = 140, .external_lex_state = 2},
  [6487] = {.lex_state = 140, .external_lex_state = 2},
  [6488] = {.lex_state = 140, .external_lex_state = 2},
  [6489] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6490] = {.lex_state = 140, .external_lex_state = 2},
  [6491] = {.lex_state = 46, .external_lex_state = 2},
  [6492] = {.lex_state = 140, .external_lex_state = 2},
  [6493] = {.lex_state = 140, .external_lex_state = 2},
  [6494] = {.lex_state = 140, .external_lex_state = 2},
  [6495] = {.lex_state = 140, .external_lex_state = 2},
  [6496] = {.lex_state = 140, .external_lex_state = 2},
  [6497] = {.lex_state = 140, .external_lex_state = 2},
  [6498] = {.lex_state = 140, .external_lex_state = 2},
  [6499] = {.lex_state = 46, .external_lex_state = 2},
  [6500] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6501] = {.lex_state = 140, .external_lex_state = 2},
  [6502] = {.lex_state = 140, .external_lex_state = 2},
  [6503] = {.lex_state = 140, .external_lex_state = 2},
  [6504] = {.lex_state = 140, .external_lex_state = 2},
  [6505] = {.lex_state = 140, .external_lex_state = 2},
  [6506] = {.lex_state = 140, .external_lex_state = 2},
  [6507] = {.lex_state = 140, .external_lex_state = 2},
  [6508] = {.lex_state = 140, .external_lex_state = 2},
  [6509] = {.lex_state = 140, .external_lex_state = 2},
  [6510] = {.lex_state = 140, .external_lex_state = 2},
  [6511] = {.lex_state = 140, .external_lex_state = 2},
  [6512] = {.lex_state = 140, .external_lex_state = 2},
  [6513] = {.lex_state = 140, .external_lex_state = 2},
  [6514] = {.lex_state = 140, .external_lex_state = 2},
  [6515] = {.lex_state = 140, .external_lex_state = 2},
  [6516] = {.lex_state = 46, .external_lex_state = 2},
  [6517] = {.lex_state = 140, .external_lex_state = 2},
  [6518] = {.lex_state = 46, .external_lex_state = 2},
  [6519] = {.lex_state = 140, .external_lex_state = 2},
  [6520] = {.lex_state = 140, .external_lex_state = 2},
  [6521] = {.lex_state = 140, .external_lex_state = 2},
  [6522] = {.lex_state = 46, .external_lex_state = 2},
  [6523] = {.lex_state = 140, .external_lex_state = 2},
  [6524] = {.lex_state = 140, .external_lex_state = 2},
  [6525] = {.lex_state = 140, .external_lex_state = 2},
  [6526] = {.lex_state = 140, .external_lex_state = 2},
  [6527] = {.lex_state = 140, .external_lex_state = 2},
  [6528] = {.lex_state = 46, .external_lex_state = 2},
  [6529] = {.lex_state = 140, .external_lex_state = 2},
  [6530] = {.lex_state = 140, .external_lex_state = 2},
  [6531] = {.lex_state = 46, .external_lex_state = 2},
  [6532] = {.lex_state = 154, .external_lex_state = 2},
  [6533] = {.lex_state = 140, .external_lex_state = 2},
  [6534] = {.lex_state = 46, .external_lex_state = 2},
  [6535] = {.lex_state = 140, .external_lex_state = 2},
  [6536] = {.lex_state = 140, .external_lex_state = 2},
  [6537] = {.lex_state = 140, .external_lex_state = 2},
  [6538] = {.lex_state = 140, .external_lex_state = 2},
  [6539] = {.lex_state = 140, .external_lex_state = 2},
  [6540] = {.lex_state = 46, .external_lex_state = 2},
  [6541] = {.lex_state = 140, .external_lex_state = 2},
  [6542] = {.lex_state = 15, .external_lex_state = 2},
  [6543] = {.lex_state = 140, .external_lex_state = 2},
  [6544] = {.lex_state = 46, .external_lex_state = 2},
  [6545] = {.lex_state = 140, .external_lex_state = 2},
  [6546] = {.lex_state = 46, .external_lex_state = 2},
  [6547] = {.lex_state = 140, .external_lex_state = 2},
  [6548] = {.lex_state = 140, .external_lex_state = 2},
  [6549] = {.lex_state = 140, .external_lex_state = 2},
  [6550] = {.lex_state = 140, .external_lex_state = 2},
  [6551] = {.lex_state = 140, .external_lex_state = 2},
  [6552] = {.lex_state = 46, .external_lex_state = 2},
  [6553] = {.lex_state = 140, .external_lex_state = 2},
  [6554] = {.lex_state = 140, .external_lex_state = 2},
  [6555] = {.lex_state = 140, .external_lex_state = 2},
  [6556] = {.lex_state = 140, .external_lex_state = 2},
  [6557] = {.lex_state = 46, .external_lex_state = 2},
  [6558] = {.lex_state = 140, .external_lex_state = 2},
  [6559] = {.lex_state = 140, .external_lex_state = 2},
  [6560] = {.lex_state = 140, .external_lex_state = 2},
  [6561] = {.lex_state = 46, .external_lex_state = 2},
  [6562] = {.lex_state = 46, .external_lex_state = 2},
  [6563] = {.lex_state = 140, .external_lex_state = 2},
  [6564] = {.lex_state = 140, .external_lex_state = 2},
  [6565] = {.lex_state = 140, .external_lex_state = 2},
  [6566] = {.lex_state = 140, .external_lex_state = 2},
  [6567] = {.lex_state = 140, .external_lex_state = 2},
  [6568] = {.lex_state = 140, .external_lex_state = 2},
  [6569] = {.lex_state = 140, .external_lex_state = 2},
  [6570] = {.lex_state = 15, .external_lex_state = 2},
  [6571] = {.lex_state = 140, .external_lex_state = 3},
  [6572] = {.lex_state = 140, .external_lex_state = 2},
  [6573] = {.lex_state = 140, .external_lex_state = 2},
  [6574] = {.lex_state = 140, .external_lex_state = 2},
  [6575] = {.lex_state = 140, .external_lex_state = 2},
  [6576] = {.lex_state = 140, .external_lex_state = 2},
  [6577] = {.lex_state = 140, .external_lex_state = 2},
  [6578] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6579] = {.lex_state = 140, .external_lex_state = 2},
  [6580] = {.lex_state = 140, .external_lex_state = 2},
  [6581] = {.lex_state = 140, .external_lex_state = 2},
  [6582] = {.lex_state = 140, .external_lex_state = 2},
  [6583] = {.lex_state = 140, .external_lex_state = 2},
  [6584] = {.lex_state = 140, .external_lex_state = 2},
  [6585] = {.lex_state = 140, .external_lex_state = 2},
  [6586] = {.lex_state = 140, .external_lex_state = 2},
  [6587] = {.lex_state = 140, .external_lex_state = 2},
  [6588] = {.lex_state = 140, .external_lex_state = 2},
  [6589] = {.lex_state = 140, .external_lex_state = 2},
  [6590] = {.lex_state = 140, .external_lex_state = 2},
  [6591] = {.lex_state = 140, .external_lex_state = 2},
  [6592] = {.lex_state = 140, .external_lex_state = 2},
  [6593] = {.lex_state = 140, .external_lex_state = 2},
  [6594] = {.lex_state = 140, .external_lex_state = 2},
  [6595] = {.lex_state = 140, .external_lex_state = 2},
  [6596] = {.lex_state = 140, .external_lex_state = 2},
  [6597] = {.lex_state = 140, .external_lex_state = 2},
  [6598] = {.lex_state = 140, .external_lex_state = 2},
  [6599] = {.lex_state = 140, .external_lex_state = 2},
  [6600] = {.lex_state = 140, .external_lex_state = 2},
  [6601] = {.lex_state = 46, .external_lex_state = 2},
  [6602] = {.lex_state = 140, .external_lex_state = 2},
  [6603] = {.lex_state = 140, .external_lex_state = 2},
  [6604] = {.lex_state = 140, .external_lex_state = 2},
  [6605] = {.lex_state = 140, .external_lex_state = 2},
  [6606] = {.lex_state = 140, .external_lex_state = 2},
  [6607] = {.lex_state = 140, .external_lex_state = 2},
  [6608] = {.lex_state = 46, .external_lex_state = 2},
  [6609] = {.lex_state = 140, .external_lex_state = 2},
  [6610] = {.lex_state = 140, .external_lex_state = 2},
  [6611] = {.lex_state = 46, .external_lex_state = 2},
  [6612] = {.lex_state = 46, .external_lex_state = 2},
  [6613] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6614] = {.lex_state = 140, .external_lex_state = 2},
  [6615] = {.lex_state = 140, .external_lex_state = 2},
  [6616] = {.lex_state = 140, .external_lex_state = 2},
  [6617] = {.lex_state = 140, .external_lex_state = 2},
  [6618] = {.lex_state = 140, .external_lex_state = 2},
  [6619] = {.lex_state = 140, .external_lex_state = 2},
  [6620] = {.lex_state = 140, .external_lex_state = 2},
  [6621] = {.lex_state = 140, .external_lex_state = 2},
  [6622] = {.lex_state = 140, .external_lex_state = 2},
  [6623] = {.lex_state = 140, .external_lex_state = 2},
  [6624] = {.lex_state = 140, .external_lex_state = 2},
  [6625] = {.lex_state = 140, .external_lex_state = 2},
  [6626] = {.lex_state = 140, .external_lex_state = 2},
  [6627] = {.lex_state = 140, .external_lex_state = 2},
  [6628] = {.lex_state = 140, .external_lex_state = 2},
  [6629] = {.lex_state = 140, .external_lex_state = 2},
  [6630] = {.lex_state = 140, .external_lex_state = 2},
  [6631] = {.lex_state = 140, .external_lex_state = 2},
  [6632] = {.lex_state = 140, .external_lex_state = 2},
  [6633] = {.lex_state = 140, .external_lex_state = 2},
  [6634] = {.lex_state = 140, .external_lex_state = 2},
  [6635] = {.lex_state = 140, .external_lex_state = 2},
  [6636] = {.lex_state = 140, .external_lex_state = 2},
  [6637] = {.lex_state = 140, .external_lex_state = 2},
  [6638] = {.lex_state = 140, .external_lex_state = 2},
  [6639] = {.lex_state = 140, .external_lex_state = 2},
  [6640] = {.lex_state = 140, .external_lex_state = 2},
  [6641] = {.lex_state = 140, .external_lex_state = 2},
  [6642] = {.lex_state = 140, .external_lex_state = 2},
  [6643] = {.lex_state = 140, .external_lex_state = 2},
  [6644] = {.lex_state = 140, .external_lex_state = 2},
  [6645] = {.lex_state = 140, .external_lex_state = 2},
  [6646] = {.lex_state = 140, .external_lex_state = 2},
  [6647] = {.lex_state = 140, .external_lex_state = 2},
  [6648] = {.lex_state = 140, .external_lex_state = 2},
  [6649] = {.lex_state = 140, .external_lex_state = 2},
  [6650] = {.lex_state = 140, .external_lex_state = 2},
  [6651] = {.lex_state = 140, .external_lex_state = 2},
  [6652] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6653] = {.lex_state = 140, .external_lex_state = 2},
  [6654] = {.lex_state = 140, .external_lex_state = 2},
  [6655] = {.lex_state = 140, .external_lex_state = 2},
  [6656] = {.lex_state = 140, .external_lex_state = 2},
  [6657] = {.lex_state = 140, .external_lex_state = 2},
  [6658] = {.lex_state = 140, .external_lex_state = 2},
  [6659] = {.lex_state = 140, .external_lex_state = 2},
  [6660] = {.lex_state = 140, .external_lex_state = 2},
  [6661] = {.lex_state = 140, .external_lex_state = 2},
  [6662] = {.lex_state = 140, .external_lex_state = 2},
  [6663] = {.lex_state = 46, .external_lex_state = 2},
  [6664] = {.lex_state = 46, .external_lex_state = 2},
  [6665] = {.lex_state = 140, .external_lex_state = 2},
  [6666] = {.lex_state = 46, .external_lex_state = 2},
  [6667] = {.lex_state = 140, .external_lex_state = 2},
  [6668] = {.lex_state = 140, .external_lex_state = 2},
  [6669] = {.lex_state = 140, .external_lex_state = 2},
  [6670] = {.lex_state = 140, .external_lex_state = 2},
  [6671] = {.lex_state = 140, .external_lex_state = 2},
  [6672] = {.lex_state = 140, .external_lex_state = 2},
  [6673] = {.lex_state = 140, .external_lex_state = 2},
  [6674] = {.lex_state = 46, .external_lex_state = 2},
  [6675] = {.lex_state = 140, .external_lex_state = 2},
  [6676] = {.lex_state = 140, .external_lex_state = 2},
  [6677] = {.lex_state = 140, .external_lex_state = 2},
  [6678] = {.lex_state = 140, .external_lex_state = 2},
  [6679] = {.lex_state = 140, .external_lex_state = 2},
  [6680] = {.lex_state = 140, .external_lex_state = 2},
  [6681] = {.lex_state = 140, .external_lex_state = 2},
  [6682] = {.lex_state = 140, .external_lex_state = 2},
  [6683] = {.lex_state = 140, .external_lex_state = 2},
  [6684] = {.lex_state = 140, .external_lex_state = 2},
  [6685] = {.lex_state = 140, .external_lex_state = 2},
  [6686] = {.lex_state = 140, .external_lex_state = 2},
  [6687] = {.lex_state = 140, .external_lex_state = 2},
  [6688] = {.lex_state = 46, .external_lex_state = 2},
  [6689] = {.lex_state = 46, .external_lex_state = 2},
  [6690] = {.lex_state = 46, .external_lex_state = 2},
  [6691] = {.lex_state = 140, .external_lex_state = 2},
  [6692] = {.lex_state = 140, .external_lex_state = 2},
  [6693] = {.lex_state = 140, .external_lex_state = 2},
  [6694] = {.lex_state = 140, .external_lex_state = 2},
  [6695] = {.lex_state = 140, .external_lex_state = 2},
  [6696] = {.lex_state = 140, .external_lex_state = 2},
  [6697] = {.lex_state = 140, .external_lex_state = 2},
  [6698] = {.lex_state = 140, .external_lex_state = 2},
  [6699] = {.lex_state = 140, .external_lex_state = 2},
  [6700] = {.lex_state = 140, .external_lex_state = 2},
  [6701] = {.lex_state = 46, .external_lex_state = 2},
  [6702] = {.lex_state = 46, .external_lex_state = 2},
  [6703] = {.lex_state = 140, .external_lex_state = 2},
  [6704] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6705] = {.lex_state = 140, .external_lex_state = 2},
  [6706] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6707] = {.lex_state = 140, .external_lex_state = 2},
  [6708] = {.lex_state = 140, .external_lex_state = 2},
  [6709] = {.lex_state = 140, .external_lex_state = 2},
  [6710] = {.lex_state = 140, .external_lex_state = 2},
  [6711] = {.lex_state = 140, .external_lex_state = 2},
  [6712] = {.lex_state = 140, .external_lex_state = 2},
  [6713] = {.lex_state = 140, .external_lex_state = 2},
  [6714] = {.lex_state = 140, .external_lex_state = 2},
  [6715] = {.lex_state = 140, .external_lex_state = 2},
  [6716] = {.lex_state = 140, .external_lex_state = 2},
  [6717] = {.lex_state = 140, .external_lex_state = 2},
  [6718] = {.lex_state = 140, .external_lex_state = 2},
  [6719] = {.lex_state = 140, .external_lex_state = 2},
  [6720] = {.lex_state = 140, .external_lex_state = 2},
  [6721] = {.lex_state = 140, .external_lex_state = 2},
  [6722] = {.lex_state = 140, .external_lex_state = 2},
  [6723] = {.lex_state = 140, .external_lex_state = 2},
  [6724] = {.lex_state = 140, .external_lex_state = 2},
  [6725] = {.lex_state = 140, .external_lex_state = 2},
  [6726] = {.lex_state = 140, .external_lex_state = 2},
  [6727] = {.lex_state = 46, .external_lex_state = 2},
  [6728] = {.lex_state = 140, .external_lex_state = 2},
  [6729] = {.lex_state = 140, .external_lex_state = 2},
  [6730] = {.lex_state = 140, .external_lex_state = 2},
  [6731] = {.lex_state = 140, .external_lex_state = 2},
  [6732] = {.lex_state = 140, .external_lex_state = 2},
  [6733] = {.lex_state = 140, .external_lex_state = 2},
  [6734] = {.lex_state = 140, .external_lex_state = 2},
  [6735] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6736] = {.lex_state = 46, .external_lex_state = 2},
  [6737] = {.lex_state = 46, .external_lex_state = 2},
  [6738] = {.lex_state = 140, .external_lex_state = 2},
  [6739] = {.lex_state = 140, .external_lex_state = 2},
  [6740] = {.lex_state = 140, .external_lex_state = 2},
  [6741] = {.lex_state = 140, .external_lex_state = 2},
  [6742] = {.lex_state = 15, .external_lex_state = 2},
  [6743] = {.lex_state = 140, .external_lex_state = 2},
  [6744] = {.lex_state = 140, .external_lex_state = 2},
  [6745] = {.lex_state = 140, .external_lex_state = 2},
  [6746] = {.lex_state = 140, .external_lex_state = 2},
  [6747] = {.lex_state = 140, .external_lex_state = 2},
  [6748] = {.lex_state = 140, .external_lex_state = 2},
  [6749] = {.lex_state = 140, .external_lex_state = 2},
  [6750] = {.lex_state = 46, .external_lex_state = 2},
  [6751] = {.lex_state = 140, .external_lex_state = 2},
  [6752] = {.lex_state = 140, .external_lex_state = 2},
  [6753] = {.lex_state = 140, .external_lex_state = 2},
  [6754] = {.lex_state = 140, .external_lex_state = 2},
  [6755] = {.lex_state = 140, .external_lex_state = 2},
  [6756] = {.lex_state = 140, .external_lex_state = 2},
  [6757] = {.lex_state = 46, .external_lex_state = 2},
  [6758] = {.lex_state = 140, .external_lex_state = 2},
  [6759] = {.lex_state = 140, .external_lex_state = 2},
  [6760] = {.lex_state = 46, .external_lex_state = 2},
  [6761] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6762] = {.lex_state = 140, .external_lex_state = 2},
  [6763] = {.lex_state = 140, .external_lex_state = 3},
  [6764] = {.lex_state = 140, .external_lex_state = 2},
  [6765] = {.lex_state = 46, .external_lex_state = 2},
  [6766] = {.lex_state = 140, .external_lex_state = 2},
  [6767] = {.lex_state = 140, .external_lex_state = 2},
  [6768] = {.lex_state = 140, .external_lex_state = 2},
  [6769] = {.lex_state = 140, .external_lex_state = 2},
  [6770] = {.lex_state = 140, .external_lex_state = 2},
  [6771] = {.lex_state = 140, .external_lex_state = 2},
  [6772] = {.lex_state = 140, .external_lex_state = 2},
  [6773] = {.lex_state = 140, .external_lex_state = 2},
  [6774] = {.lex_state = 46, .external_lex_state = 2},
  [6775] = {.lex_state = 140, .external_lex_state = 2},
  [6776] = {.lex_state = 140, .external_lex_state = 2},
  [6777] = {.lex_state = 140, .external_lex_state = 2},
  [6778] = {.lex_state = 140, .external_lex_state = 2},
  [6779] = {.lex_state = 140, .external_lex_state = 2},
  [6780] = {.lex_state = 140, .external_lex_state = 2},
  [6781] = {.lex_state = 140, .external_lex_state = 2},
  [6782] = {.lex_state = 140, .external_lex_state = 2},
  [6783] = {.lex_state = 140, .external_lex_state = 2},
  [6784] = {.lex_state = 140, .external_lex_state = 2},
  [6785] = {.lex_state = 140, .external_lex_state = 2},
  [6786] = {.lex_state = 140, .external_lex_state = 2},
  [6787] = {.lex_state = 140, .external_lex_state = 2},
  [6788] = {.lex_state = 140, .external_lex_state = 2},
  [6789] = {.lex_state = 140, .external_lex_state = 2},
  [6790] = {.lex_state = 140, .external_lex_state = 2},
  [6791] = {.lex_state = 140, .external_lex_state = 2},
  [6792] = {.lex_state = 140, .external_lex_state = 2},
  [6793] = {.lex_state = 140, .external_lex_state = 2},
  [6794] = {.lex_state = 140, .external_lex_state = 2},
  [6795] = {.lex_state = 46, .external_lex_state = 2},
  [6796] = {.lex_state = 46, .external_lex_state = 2},
  [6797] = {.lex_state = 46, .external_lex_state = 2},
  [6798] = {.lex_state = 140, .external_lex_state = 2},
  [6799] = {.lex_state = 140, .external_lex_state = 2},
  [6800] = {.lex_state = 140, .external_lex_state = 2},
  [6801] = {.lex_state = 46, .external_lex_state = 2},
  [6802] = {.lex_state = 140, .external_lex_state = 2},
  [6803] = {.lex_state = 140, .external_lex_state = 2},
  [6804] = {.lex_state = 15, .external_lex_state = 2},
  [6805] = {.lex_state = 46, .external_lex_state = 2},
  [6806] = {.lex_state = 140, .external_lex_state = 3},
  [6807] = {.lex_state = 140, .external_lex_state = 2},
  [6808] = {.lex_state = 140, .external_lex_state = 2},
  [6809] = {.lex_state = 140, .external_lex_state = 2},
  [6810] = {.lex_state = 140, .external_lex_state = 2},
  [6811] = {.lex_state = 140, .external_lex_state = 2},
  [6812] = {.lex_state = 140, .external_lex_state = 2},
  [6813] = {.lex_state = 140, .external_lex_state = 2},
  [6814] = {.lex_state = 140, .external_lex_state = 2},
  [6815] = {.lex_state = 140, .external_lex_state = 2},
  [6816] = {.lex_state = 46, .external_lex_state = 2},
  [6817] = {.lex_state = 140, .external_lex_state = 2},
  [6818] = {.lex_state = 140, .external_lex_state = 2},
  [6819] = {.lex_state = 140, .external_lex_state = 2},
  [6820] = {.lex_state = 140, .external_lex_state = 2},
  [6821] = {.lex_state = 140, .external_lex_state = 2},
  [6822] = {.lex_state = 140, .external_lex_state = 2},
  [6823] = {.lex_state = 46, .external_lex_state = 2},
  [6824] = {.lex_state = 46, .external_lex_state = 2},
  [6825] = {.lex_state = 140, .external_lex_state = 2},
  [6826] = {.lex_state = 140, .external_lex_state = 2},
  [6827] = {.lex_state = 140, .external_lex_state = 2},
  [6828] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6829] = {.lex_state = 46, .external_lex_state = 2},
  [6830] = {.lex_state = 140, .external_lex_state = 2},
  [6831] = {.lex_state = 140, .external_lex_state = 2},
  [6832] = {.lex_state = 140, .external_lex_state = 2},
  [6833] = {.lex_state = 140, .external_lex_state = 2},
  [6834] = {.lex_state = 140, .external_lex_state = 2},
  [6835] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6836] = {.lex_state = 140, .external_lex_state = 2},
  [6837] = {.lex_state = 46, .external_lex_state = 2},
  [6838] = {.lex_state = 140, .external_lex_state = 2},
  [6839] = {.lex_state = 140, .external_lex_state = 2},
  [6840] = {.lex_state = 154, .external_lex_state = 2},
  [6841] = {.lex_state = 140, .external_lex_state = 2},
  [6842] = {.lex_state = 140, .external_lex_state = 2},
  [6843] = {.lex_state = 140, .external_lex_state = 2},
  [6844] = {.lex_state = 15, .external_lex_state = 2},
  [6845] = {.lex_state = 15, .external_lex_state = 2},
  [6846] = {.lex_state = 140, .external_lex_state = 3},
  [6847] = {.lex_state = 140, .external_lex_state = 2},
  [6848] = {.lex_state = 140, .external_lex_state = 2},
  [6849] = {.lex_state = 140, .external_lex_state = 2},
  [6850] = {.lex_state = 46, .external_lex_state = 2},
  [6851] = {.lex_state = 140, .external_lex_state = 2},
  [6852] = {.lex_state = 140, .external_lex_state = 2},
  [6853] = {.lex_state = 46, .external_lex_state = 2},
  [6854] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6855] = {.lex_state = 140, .external_lex_state = 2},
  [6856] = {.lex_state = 154, .external_lex_state = 2},
  [6857] = {.lex_state = 140, .external_lex_state = 2},
  [6858] = {.lex_state = 140, .external_lex_state = 2},
  [6859] = {.lex_state = 140, .external_lex_state = 2},
  [6860] = {.lex_state = 140, .external_lex_state = 2},
  [6861] = {.lex_state = 140, .external_lex_state = 2},
  [6862] = {.lex_state = 140, .external_lex_state = 2},
  [6863] = {.lex_state = 140, .external_lex_state = 2},
  [6864] = {.lex_state = 140, .external_lex_state = 2},
  [6865] = {.lex_state = 140, .external_lex_state = 2},
  [6866] = {.lex_state = 140, .external_lex_state = 2},
  [6867] = {.lex_state = 140, .external_lex_state = 2},
  [6868] = {.lex_state = 140, .external_lex_state = 2},
  [6869] = {.lex_state = 140, .external_lex_state = 9},
  [6870] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6871] = {.lex_state = 140, .external_lex_state = 2},
  [6872] = {.lex_state = 140, .external_lex_state = 2},
  [6873] = {.lex_state = 140, .external_lex_state = 2},
  [6874] = {.lex_state = 140, .external_lex_state = 2},
  [6875] = {.lex_state = 140, .external_lex_state = 2},
  [6876] = {.lex_state = 140, .external_lex_state = 2},
  [6877] = {.lex_state = 140, .external_lex_state = 2},
  [6878] = {.lex_state = 140, .external_lex_state = 2},
  [6879] = {.lex_state = 140, .external_lex_state = 2},
  [6880] = {.lex_state = 46, .external_lex_state = 2},
  [6881] = {.lex_state = 140, .external_lex_state = 2},
  [6882] = {.lex_state = 140, .external_lex_state = 2},
  [6883] = {.lex_state = 46, .external_lex_state = 2},
  [6884] = {.lex_state = 140, .external_lex_state = 2},
  [6885] = {.lex_state = 140, .external_lex_state = 2},
  [6886] = {.lex_state = 140, .external_lex_state = 2},
  [6887] = {.lex_state = 140, .external_lex_state = 2},
  [6888] = {.lex_state = 140, .external_lex_state = 2},
  [6889] = {.lex_state = 46, .external_lex_state = 2},
  [6890] = {.lex_state = 140, .external_lex_state = 2},
  [6891] = {.lex_state = 140, .external_lex_state = 2},
  [6892] = {.lex_state = 140, .external_lex_state = 2},
  [6893] = {.lex_state = 15, .external_lex_state = 2},
  [6894] = {.lex_state = 46, .external_lex_state = 2},
  [6895] = {.lex_state = 46, .external_lex_state = 2},
  [6896] = {.lex_state = 140, .external_lex_state = 2},
  [6897] = {.lex_state = 140, .external_lex_state = 2},
  [6898] = {.lex_state = 140, .external_lex_state = 3},
  [6899] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6900] = {.lex_state = 140, .external_lex_state = 2},
  [6901] = {.lex_state = 45, .external_lex_state = 2},
  [6902] = {.lex_state = 140, .external_lex_state = 2},
  [6903] = {.lex_state = 140, .external_lex_state = 2},
  [6904] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6905] = {.lex_state = 140, .external_lex_state = 2},
  [6906] = {.lex_state = 140, .external_lex_state = 2},
  [6907] = {.lex_state = 140, .external_lex_state = 2},
  [6908] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6909] = {.lex_state = 140, .external_lex_state = 2},
  [6910] = {.lex_state = 140, .external_lex_state = 2},
  [6911] = {.lex_state = 140, .external_lex_state = 2},
  [6912] = {.lex_state = 140, .external_lex_state = 2},
  [6913] = {.lex_state = 140, .external_lex_state = 2},
  [6914] = {.lex_state = 140, .external_lex_state = 2},
  [6915] = {.lex_state = 140, .external_lex_state = 2},
  [6916] = {.lex_state = 37, .external_lex_state = 2},
  [6917] = {.lex_state = 37, .external_lex_state = 2},
  [6918] = {.lex_state = 37, .external_lex_state = 2},
  [6919] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6920] = {.lex_state = 46, .external_lex_state = 2},
  [6921] = {.lex_state = 140, .external_lex_state = 2},
  [6922] = {.lex_state = 140, .external_lex_state = 2},
  [6923] = {.lex_state = 46, .external_lex_state = 2},
  [6924] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6925] = {.lex_state = 46, .external_lex_state = 2},
  [6926] = {.lex_state = 140, .external_lex_state = 2},
  [6927] = {.lex_state = 140, .external_lex_state = 2},
  [6928] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6929] = {.lex_state = 140, .external_lex_state = 2},
  [6930] = {.lex_state = 140, .external_lex_state = 2},
  [6931] = {.lex_state = 140, .external_lex_state = 2},
  [6932] = {.lex_state = 140, .external_lex_state = 2},
  [6933] = {.lex_state = 140, .external_lex_state = 2},
  [6934] = {.lex_state = 140, .external_lex_state = 2},
  [6935] = {.lex_state = 46, .external_lex_state = 2},
  [6936] = {.lex_state = 140, .external_lex_state = 2},
  [6937] = {.lex_state = 140, .external_lex_state = 2},
  [6938] = {.lex_state = 46, .external_lex_state = 2},
  [6939] = {.lex_state = 140, .external_lex_state = 2},
  [6940] = {.lex_state = 46, .external_lex_state = 2},
  [6941] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6942] = {.lex_state = 140, .external_lex_state = 2},
  [6943] = {.lex_state = 140, .external_lex_state = 2},
  [6944] = {.lex_state = 140, .external_lex_state = 2},
  [6945] = {.lex_state = 140, .external_lex_state = 2},
  [6946] = {.lex_state = 140, .external_lex_state = 2},
  [6947] = {.lex_state = 46, .external_lex_state = 2},
  [6948] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6949] = {.lex_state = 15, .external_lex_state = 2},
  [6950] = {.lex_state = 140, .external_lex_state = 3},
  [6951] = {.lex_state = 140, .external_lex_state = 2},
  [6952] = {.lex_state = 140, .external_lex_state = 2},
  [6953] = {.lex_state = 140, .external_lex_state = 2},
  [6954] = {.lex_state = 140, .external_lex_state = 2},
  [6955] = {.lex_state = 140, .external_lex_state = 2},
  [6956] = {.lex_state = 140, .external_lex_state = 2},
  [6957] = {.lex_state = 140, .external_lex_state = 2},
  [6958] = {.lex_state = 140, .external_lex_state = 2},
  [6959] = {.lex_state = 140, .external_lex_state = 2},
  [6960] = {.lex_state = 140, .external_lex_state = 2},
  [6961] = {.lex_state = 46, .external_lex_state = 2},
  [6962] = {.lex_state = 140, .external_lex_state = 2},
  [6963] = {.lex_state = 46, .external_lex_state = 2},
  [6964] = {.lex_state = 140, .external_lex_state = 2},
  [6965] = {.lex_state = 140, .external_lex_state = 2},
  [6966] = {.lex_state = 140, .external_lex_state = 2},
  [6967] = {.lex_state = 46, .external_lex_state = 2},
  [6968] = {.lex_state = 140, .external_lex_state = 2},
  [6969] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [6970] = {.lex_state = 140, .external_lex_state = 2},
  [6971] = {.lex_state = 140, .external_lex_state = 2},
  [6972] = {.lex_state = 46, .external_lex_state = 2},
  [6973] = {.lex_state = 140, .external_lex_state = 2},
  [6974] = {.lex_state = 140, .external_lex_state = 2},
  [6975] = {.lex_state = 140, .external_lex_state = 2},
  [6976] = {.lex_state = 140, .external_lex_state = 2},
  [6977] = {.lex_state = 140, .external_lex_state = 2},
  [6978] = {.lex_state = 140, .external_lex_state = 2},
  [6979] = {.lex_state = 140, .external_lex_state = 2},
  [6980] = {.lex_state = 140, .external_lex_state = 2},
  [6981] = {.lex_state = 140, .external_lex_state = 2},
  [6982] = {.lex_state = 140, .external_lex_state = 2},
  [6983] = {.lex_state = 140, .external_lex_state = 2},
  [6984] = {.lex_state = 140, .external_lex_state = 2},
  [6985] = {.lex_state = 46, .external_lex_state = 2},
  [6986] = {.lex_state = 140, .external_lex_state = 2},
  [6987] = {.lex_state = 140, .external_lex_state = 2},
  [6988] = {.lex_state = 140, .external_lex_state = 2},
  [6989] = {.lex_state = 140, .external_lex_state = 2},
  [6990] = {.lex_state = 140, .external_lex_state = 2},
  [6991] = {.lex_state = 140, .external_lex_state = 2},
  [6992] = {.lex_state = 140, .external_lex_state = 2},
  [6993] = {.lex_state = 140, .external_lex_state = 2},
  [6994] = {.lex_state = 140, .external_lex_state = 2},
  [6995] = {.lex_state = 15, .external_lex_state = 2},
  [6996] = {.lex_state = 140, .external_lex_state = 3},
  [6997] = {.lex_state = 140, .external_lex_state = 2},
  [6998] = {.lex_state = 46, .external_lex_state = 2},
  [6999] = {.lex_state = 140, .external_lex_state = 2},
  [7000] = {.lex_state = 140, .external_lex_state = 2},
  [7001] = {.lex_state = 140, .external_lex_state = 2},
  [7002] = {.lex_state = 140, .external_lex_state = 2},
  [7003] = {.lex_state = 140, .external_lex_state = 2},
  [7004] = {.lex_state = 140, .external_lex_state = 2},
  [7005] = {.lex_state = 140, .external_lex_state = 2},
  [7006] = {.lex_state = 140, .external_lex_state = 2},
  [7007] = {.lex_state = 140, .external_lex_state = 2},
  [7008] = {.lex_state = 140, .external_lex_state = 2},
  [7009] = {.lex_state = 140, .external_lex_state = 2},
  [7010] = {.lex_state = 46, .external_lex_state = 2},
  [7011] = {.lex_state = 46, .external_lex_state = 2},
  [7012] = {.lex_state = 140, .external_lex_state = 2},
  [7013] = {.lex_state = 140, .external_lex_state = 2},
  [7014] = {.lex_state = 140, .external_lex_state = 2},
  [7015] = {.lex_state = 140, .external_lex_state = 2},
  [7016] = {.lex_state = 140, .external_lex_state = 2},
  [7017] = {.lex_state = 140, .external_lex_state = 2},
  [7018] = {.lex_state = 140, .external_lex_state = 2},
  [7019] = {.lex_state = 140, .external_lex_state = 2},
  [7020] = {.lex_state = 140, .external_lex_state = 2},
  [7021] = {.lex_state = 140, .external_lex_state = 2},
  [7022] = {.lex_state = 140, .external_lex_state = 2},
  [7023] = {.lex_state = 140, .external_lex_state = 2},
  [7024] = {.lex_state = 140, .external_lex_state = 2},
  [7025] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7026] = {.lex_state = 140, .external_lex_state = 2},
  [7027] = {.lex_state = 140, .external_lex_state = 2},
  [7028] = {.lex_state = 140, .external_lex_state = 2},
  [7029] = {.lex_state = 46, .external_lex_state = 2},
  [7030] = {.lex_state = 140, .external_lex_state = 3},
  [7031] = {.lex_state = 140, .external_lex_state = 2},
  [7032] = {.lex_state = 140, .external_lex_state = 2},
  [7033] = {.lex_state = 140, .external_lex_state = 2},
  [7034] = {.lex_state = 140, .external_lex_state = 2},
  [7035] = {.lex_state = 140, .external_lex_state = 2},
  [7036] = {.lex_state = 15, .external_lex_state = 2},
  [7037] = {.lex_state = 140, .external_lex_state = 2},
  [7038] = {.lex_state = 140, .external_lex_state = 2},
  [7039] = {.lex_state = 140, .external_lex_state = 2},
  [7040] = {.lex_state = 140, .external_lex_state = 2},
  [7041] = {.lex_state = 140, .external_lex_state = 3},
  [7042] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7043] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7044] = {.lex_state = 140, .external_lex_state = 2},
  [7045] = {.lex_state = 140, .external_lex_state = 2},
  [7046] = {.lex_state = 46, .external_lex_state = 2},
  [7047] = {.lex_state = 140, .external_lex_state = 2},
  [7048] = {.lex_state = 140, .external_lex_state = 2},
  [7049] = {.lex_state = 140, .external_lex_state = 2},
  [7050] = {.lex_state = 140, .external_lex_state = 2},
  [7051] = {.lex_state = 140, .external_lex_state = 2},
  [7052] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7053] = {.lex_state = 140, .external_lex_state = 2},
  [7054] = {.lex_state = 45, .external_lex_state = 2},
  [7055] = {.lex_state = 140, .external_lex_state = 2},
  [7056] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7057] = {.lex_state = 140, .external_lex_state = 2},
  [7058] = {.lex_state = 140, .external_lex_state = 2},
  [7059] = {.lex_state = 46, .external_lex_state = 2},
  [7060] = {.lex_state = 37, .external_lex_state = 2},
  [7061] = {.lex_state = 37, .external_lex_state = 2},
  [7062] = {.lex_state = 37, .external_lex_state = 2},
  [7063] = {.lex_state = 140, .external_lex_state = 2},
  [7064] = {.lex_state = 140, .external_lex_state = 2},
  [7065] = {.lex_state = 140, .external_lex_state = 2},
  [7066] = {.lex_state = 140, .external_lex_state = 2},
  [7067] = {.lex_state = 46, .external_lex_state = 2},
  [7068] = {.lex_state = 140, .external_lex_state = 2},
  [7069] = {.lex_state = 140, .external_lex_state = 2},
  [7070] = {.lex_state = 140, .external_lex_state = 2},
  [7071] = {.lex_state = 140, .external_lex_state = 2},
  [7072] = {.lex_state = 46, .external_lex_state = 2},
  [7073] = {.lex_state = 140, .external_lex_state = 2},
  [7074] = {.lex_state = 140, .external_lex_state = 2},
  [7075] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7076] = {.lex_state = 140, .external_lex_state = 2},
  [7077] = {.lex_state = 154, .external_lex_state = 2},
  [7078] = {.lex_state = 140, .external_lex_state = 2},
  [7079] = {.lex_state = 140, .external_lex_state = 2},
  [7080] = {.lex_state = 140, .external_lex_state = 2},
  [7081] = {.lex_state = 140, .external_lex_state = 2},
  [7082] = {.lex_state = 15, .external_lex_state = 2},
  [7083] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7084] = {.lex_state = 140, .external_lex_state = 2},
  [7085] = {.lex_state = 140, .external_lex_state = 2},
  [7086] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7087] = {.lex_state = 140, .external_lex_state = 3},
  [7088] = {.lex_state = 140, .external_lex_state = 2},
  [7089] = {.lex_state = 37, .external_lex_state = 2},
  [7090] = {.lex_state = 37, .external_lex_state = 2},
  [7091] = {.lex_state = 37, .external_lex_state = 2},
  [7092] = {.lex_state = 140, .external_lex_state = 2},
  [7093] = {.lex_state = 140, .external_lex_state = 2},
  [7094] = {.lex_state = 140, .external_lex_state = 2},
  [7095] = {.lex_state = 140, .external_lex_state = 2},
  [7096] = {.lex_state = 140, .external_lex_state = 2},
  [7097] = {.lex_state = 140, .external_lex_state = 2},
  [7098] = {.lex_state = 140, .external_lex_state = 2},
  [7099] = {.lex_state = 140, .external_lex_state = 2},
  [7100] = {.lex_state = 140, .external_lex_state = 2},
  [7101] = {.lex_state = 46, .external_lex_state = 2},
  [7102] = {.lex_state = 140, .external_lex_state = 2},
  [7103] = {.lex_state = 140, .external_lex_state = 2},
  [7104] = {.lex_state = 140, .external_lex_state = 2},
  [7105] = {.lex_state = 140, .external_lex_state = 2},
  [7106] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7107] = {.lex_state = 140, .external_lex_state = 2},
  [7108] = {.lex_state = 37, .external_lex_state = 2},
  [7109] = {.lex_state = 37, .external_lex_state = 2},
  [7110] = {.lex_state = 37, .external_lex_state = 2},
  [7111] = {.lex_state = 140, .external_lex_state = 2},
  [7112] = {.lex_state = 140, .external_lex_state = 2},
  [7113] = {.lex_state = 140, .external_lex_state = 2},
  [7114] = {.lex_state = 140, .external_lex_state = 2},
  [7115] = {.lex_state = 140, .external_lex_state = 2},
  [7116] = {.lex_state = 46, .external_lex_state = 2},
  [7117] = {.lex_state = 140, .external_lex_state = 2},
  [7118] = {.lex_state = 15, .external_lex_state = 2},
  [7119] = {.lex_state = 46, .external_lex_state = 2},
  [7120] = {.lex_state = 140, .external_lex_state = 2},
  [7121] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7122] = {.lex_state = 140, .external_lex_state = 2},
  [7123] = {.lex_state = 37, .external_lex_state = 2},
  [7124] = {.lex_state = 37, .external_lex_state = 2},
  [7125] = {.lex_state = 37, .external_lex_state = 2},
  [7126] = {.lex_state = 140, .external_lex_state = 2},
  [7127] = {.lex_state = 15, .external_lex_state = 2},
  [7128] = {.lex_state = 140, .external_lex_state = 3},
  [7129] = {.lex_state = 140, .external_lex_state = 2},
  [7130] = {.lex_state = 140, .external_lex_state = 2},
  [7131] = {.lex_state = 140, .external_lex_state = 2},
  [7132] = {.lex_state = 140, .external_lex_state = 2},
  [7133] = {.lex_state = 46, .external_lex_state = 2},
  [7134] = {.lex_state = 46, .external_lex_state = 2},
  [7135] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7136] = {.lex_state = 140, .external_lex_state = 2},
  [7137] = {.lex_state = 37, .external_lex_state = 2},
  [7138] = {.lex_state = 37, .external_lex_state = 2},
  [7139] = {.lex_state = 37, .external_lex_state = 2},
  [7140] = {.lex_state = 140, .external_lex_state = 2},
  [7141] = {.lex_state = 140, .external_lex_state = 2},
  [7142] = {.lex_state = 46, .external_lex_state = 2},
  [7143] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7144] = {.lex_state = 46, .external_lex_state = 2},
  [7145] = {.lex_state = 15, .external_lex_state = 2},
  [7146] = {.lex_state = 154, .external_lex_state = 2},
  [7147] = {.lex_state = 140, .external_lex_state = 2},
  [7148] = {.lex_state = 140, .external_lex_state = 3},
  [7149] = {.lex_state = 140, .external_lex_state = 2},
  [7150] = {.lex_state = 140, .external_lex_state = 2},
  [7151] = {.lex_state = 37, .external_lex_state = 2},
  [7152] = {.lex_state = 37, .external_lex_state = 2},
  [7153] = {.lex_state = 37, .external_lex_state = 2},
  [7154] = {.lex_state = 140, .external_lex_state = 2},
  [7155] = {.lex_state = 140, .external_lex_state = 2},
  [7156] = {.lex_state = 140, .external_lex_state = 2},
  [7157] = {.lex_state = 140, .external_lex_state = 2},
  [7158] = {.lex_state = 46, .external_lex_state = 2},
  [7159] = {.lex_state = 140, .external_lex_state = 2},
  [7160] = {.lex_state = 140, .external_lex_state = 2},
  [7161] = {.lex_state = 140, .external_lex_state = 2},
  [7162] = {.lex_state = 46, .external_lex_state = 2},
  [7163] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7164] = {.lex_state = 140, .external_lex_state = 2},
  [7165] = {.lex_state = 37, .external_lex_state = 2},
  [7166] = {.lex_state = 37, .external_lex_state = 2},
  [7167] = {.lex_state = 37, .external_lex_state = 2},
  [7168] = {.lex_state = 140, .external_lex_state = 2},
  [7169] = {.lex_state = 140, .external_lex_state = 2},
  [7170] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7171] = {.lex_state = 140, .external_lex_state = 2},
  [7172] = {.lex_state = 46, .external_lex_state = 2},
  [7173] = {.lex_state = 140, .external_lex_state = 2},
  [7174] = {.lex_state = 140, .external_lex_state = 2},
  [7175] = {.lex_state = 140, .external_lex_state = 2},
  [7176] = {.lex_state = 46, .external_lex_state = 2},
  [7177] = {.lex_state = 140, .external_lex_state = 2},
  [7178] = {.lex_state = 140, .external_lex_state = 2},
  [7179] = {.lex_state = 37, .external_lex_state = 2},
  [7180] = {.lex_state = 37, .external_lex_state = 2},
  [7181] = {.lex_state = 37, .external_lex_state = 2},
  [7182] = {.lex_state = 140, .external_lex_state = 2},
  [7183] = {.lex_state = 140, .external_lex_state = 2},
  [7184] = {.lex_state = 140, .external_lex_state = 2},
  [7185] = {.lex_state = 140, .external_lex_state = 2},
  [7186] = {.lex_state = 37, .external_lex_state = 2},
  [7187] = {.lex_state = 37, .external_lex_state = 2},
  [7188] = {.lex_state = 37, .external_lex_state = 2},
  [7189] = {.lex_state = 140, .external_lex_state = 2},
  [7190] = {.lex_state = 140, .external_lex_state = 2},
  [7191] = {.lex_state = 37, .external_lex_state = 2},
  [7192] = {.lex_state = 37, .external_lex_state = 2},
  [7193] = {.lex_state = 37, .external_lex_state = 2},
  [7194] = {.lex_state = 140, .external_lex_state = 3},
  [7195] = {.lex_state = 140, .external_lex_state = 2},
  [7196] = {.lex_state = 37, .external_lex_state = 2},
  [7197] = {.lex_state = 37, .external_lex_state = 2},
  [7198] = {.lex_state = 37, .external_lex_state = 2},
  [7199] = {.lex_state = 140, .external_lex_state = 2},
  [7200] = {.lex_state = 140, .external_lex_state = 2},
  [7201] = {.lex_state = 37, .external_lex_state = 2},
  [7202] = {.lex_state = 37, .external_lex_state = 2},
  [7203] = {.lex_state = 37, .external_lex_state = 2},
  [7204] = {.lex_state = 140, .external_lex_state = 2},
  [7205] = {.lex_state = 140, .external_lex_state = 2},
  [7206] = {.lex_state = 37, .external_lex_state = 2},
  [7207] = {.lex_state = 37, .external_lex_state = 2},
  [7208] = {.lex_state = 37, .external_lex_state = 2},
  [7209] = {.lex_state = 46, .external_lex_state = 2},
  [7210] = {.lex_state = 140, .external_lex_state = 2},
  [7211] = {.lex_state = 37, .external_lex_state = 2},
  [7212] = {.lex_state = 37, .external_lex_state = 2},
  [7213] = {.lex_state = 140, .external_lex_state = 2},
  [7214] = {.lex_state = 140, .external_lex_state = 2},
  [7215] = {.lex_state = 140, .external_lex_state = 2},
  [7216] = {.lex_state = 37, .external_lex_state = 2},
  [7217] = {.lex_state = 37, .external_lex_state = 2},
  [7218] = {.lex_state = 37, .external_lex_state = 2},
  [7219] = {.lex_state = 46, .external_lex_state = 2},
  [7220] = {.lex_state = 140, .external_lex_state = 2},
  [7221] = {.lex_state = 140, .external_lex_state = 2},
  [7222] = {.lex_state = 140, .external_lex_state = 2},
  [7223] = {.lex_state = 140, .external_lex_state = 2},
  [7224] = {.lex_state = 46, .external_lex_state = 2},
  [7225] = {.lex_state = 140, .external_lex_state = 2},
  [7226] = {.lex_state = 140, .external_lex_state = 2},
  [7227] = {.lex_state = 140, .external_lex_state = 2},
  [7228] = {.lex_state = 140, .external_lex_state = 2},
  [7229] = {.lex_state = 79, .external_lex_state = 2},
  [7230] = {.lex_state = 140, .external_lex_state = 2},
  [7231] = {.lex_state = 140, .external_lex_state = 2},
  [7232] = {.lex_state = 140, .external_lex_state = 2},
  [7233] = {.lex_state = 140, .external_lex_state = 2},
  [7234] = {.lex_state = 46, .external_lex_state = 2},
  [7235] = {.lex_state = 140, .external_lex_state = 2},
  [7236] = {.lex_state = 140, .external_lex_state = 2},
  [7237] = {.lex_state = 37, .external_lex_state = 2},
  [7238] = {.lex_state = 140, .external_lex_state = 2},
  [7239] = {.lex_state = 140, .external_lex_state = 2},
  [7240] = {.lex_state = 140, .external_lex_state = 2},
  [7241] = {.lex_state = 46, .external_lex_state = 2},
  [7242] = {.lex_state = 140, .external_lex_state = 2},
  [7243] = {.lex_state = 140, .external_lex_state = 2},
  [7244] = {.lex_state = 140, .external_lex_state = 2},
  [7245] = {.lex_state = 140, .external_lex_state = 2},
  [7246] = {.lex_state = 46, .external_lex_state = 2},
  [7247] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7248] = {.lex_state = 140, .external_lex_state = 2},
  [7249] = {.lex_state = 140, .external_lex_state = 2},
  [7250] = {.lex_state = 37, .external_lex_state = 2},
  [7251] = {.lex_state = 140, .external_lex_state = 2},
  [7252] = {.lex_state = 140, .external_lex_state = 2},
  [7253] = {.lex_state = 140, .external_lex_state = 2},
  [7254] = {.lex_state = 140, .external_lex_state = 2},
  [7255] = {.lex_state = 140, .external_lex_state = 2},
  [7256] = {.lex_state = 140, .external_lex_state = 2},
  [7257] = {.lex_state = 140, .external_lex_state = 2},
  [7258] = {.lex_state = 140, .external_lex_state = 2},
  [7259] = {.lex_state = 15, .external_lex_state = 2},
  [7260] = {.lex_state = 37, .external_lex_state = 2},
  [7261] = {.lex_state = 140, .external_lex_state = 3},
  [7262] = {.lex_state = 140, .external_lex_state = 2},
  [7263] = {.lex_state = 79, .external_lex_state = 2},
  [7264] = {.lex_state = 140, .external_lex_state = 2},
  [7265] = {.lex_state = 46, .external_lex_state = 2},
  [7266] = {.lex_state = 79, .external_lex_state = 2},
  [7267] = {.lex_state = 46, .external_lex_state = 2},
  [7268] = {.lex_state = 140, .external_lex_state = 2},
  [7269] = {.lex_state = 79, .external_lex_state = 2},
  [7270] = {.lex_state = 79, .external_lex_state = 2},
  [7271] = {.lex_state = 79, .external_lex_state = 2},
  [7272] = {.lex_state = 79, .external_lex_state = 2},
  [7273] = {.lex_state = 79, .external_lex_state = 2},
  [7274] = {.lex_state = 79, .external_lex_state = 2},
  [7275] = {.lex_state = 79, .external_lex_state = 2},
  [7276] = {.lex_state = 79, .external_lex_state = 2},
  [7277] = {.lex_state = 79, .external_lex_state = 2},
  [7278] = {.lex_state = 79, .external_lex_state = 2},
  [7279] = {.lex_state = 79, .external_lex_state = 2},
  [7280] = {.lex_state = 79, .external_lex_state = 2},
  [7281] = {.lex_state = 79, .external_lex_state = 2},
  [7282] = {.lex_state = 140, .external_lex_state = 2},
  [7283] = {.lex_state = 140, .external_lex_state = 2},
  [7284] = {.lex_state = 140, .external_lex_state = 2},
  [7285] = {.lex_state = 140, .external_lex_state = 2},
  [7286] = {.lex_state = 140, .external_lex_state = 2},
  [7287] = {.lex_state = 140, .external_lex_state = 2},
  [7288] = {.lex_state = 140, .external_lex_state = 2},
  [7289] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7290] = {.lex_state = 46, .external_lex_state = 2},
  [7291] = {.lex_state = 46, .external_lex_state = 2},
  [7292] = {.lex_state = 46, .external_lex_state = 2},
  [7293] = {.lex_state = 140, .external_lex_state = 2},
  [7294] = {.lex_state = 140, .external_lex_state = 2},
  [7295] = {.lex_state = 140, .external_lex_state = 2},
  [7296] = {.lex_state = 140, .external_lex_state = 2},
  [7297] = {.lex_state = 140, .external_lex_state = 2},
  [7298] = {.lex_state = 140, .external_lex_state = 2},
  [7299] = {.lex_state = 140, .external_lex_state = 2},
  [7300] = {.lex_state = 140, .external_lex_state = 2},
  [7301] = {.lex_state = 140, .external_lex_state = 2},
  [7302] = {.lex_state = 140, .external_lex_state = 2},
  [7303] = {.lex_state = 140, .external_lex_state = 2},
  [7304] = {.lex_state = 140, .external_lex_state = 2},
  [7305] = {.lex_state = 140, .external_lex_state = 2},
  [7306] = {.lex_state = 140, .external_lex_state = 2},
  [7307] = {.lex_state = 140, .external_lex_state = 2},
  [7308] = {.lex_state = 140, .external_lex_state = 2},
  [7309] = {.lex_state = 140, .external_lex_state = 2},
  [7310] = {.lex_state = 140, .external_lex_state = 2},
  [7311] = {.lex_state = 140, .external_lex_state = 2},
  [7312] = {.lex_state = 140, .external_lex_state = 2},
  [7313] = {.lex_state = 140, .external_lex_state = 2},
  [7314] = {.lex_state = 140, .external_lex_state = 2},
  [7315] = {.lex_state = 140, .external_lex_state = 2},
  [7316] = {.lex_state = 140, .external_lex_state = 2},
  [7317] = {.lex_state = 140, .external_lex_state = 2},
  [7318] = {.lex_state = 79, .external_lex_state = 2},
  [7319] = {.lex_state = 140, .external_lex_state = 2},
  [7320] = {.lex_state = 140, .external_lex_state = 2},
  [7321] = {.lex_state = 140, .external_lex_state = 2},
  [7322] = {.lex_state = 140, .external_lex_state = 2},
  [7323] = {.lex_state = 140, .external_lex_state = 2},
  [7324] = {.lex_state = 140, .external_lex_state = 2},
  [7325] = {.lex_state = 46, .external_lex_state = 2},
  [7326] = {.lex_state = 140, .external_lex_state = 2},
  [7327] = {.lex_state = 140, .external_lex_state = 2},
  [7328] = {.lex_state = 140, .external_lex_state = 2},
  [7329] = {.lex_state = 140, .external_lex_state = 2},
  [7330] = {.lex_state = 140, .external_lex_state = 2},
  [7331] = {.lex_state = 140, .external_lex_state = 2},
  [7332] = {.lex_state = 140, .external_lex_state = 2},
  [7333] = {.lex_state = 46, .external_lex_state = 2},
  [7334] = {.lex_state = 25, .external_lex_state = 2, .reserved_word_set_id = 6},
  [7335] = {(TSStateId)(-1),},
  [7336] = {(TSStateId)(-1),},
};

static const TSSymbol ts_reserved_words[93][MAX_RESERVED_WORD_SET_SIZE] = {
  [1] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_as,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [2] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [3] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [4] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [5] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [6] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [7] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [8] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_as,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [9] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [10] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [11] = {
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [12] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_when,
  },
  [13] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [14] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [15] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [16] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [17] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_while,
    anon_sym_for,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [18] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [19] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [20] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [21] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [22] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [23] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [24] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
  },
  [25] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [26] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [27] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
  },
  [28] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [29] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [30] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [31] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [32] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_include,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [33] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [34] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [35] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [36] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_while,
    anon_sym_for,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [37] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [38] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [39] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [40] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [41] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [42] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [43] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [44] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [45] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [46] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [47] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [48] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [49] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [50] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [51] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [52] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [53] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [54] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [55] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_for,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [56] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [57] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [58] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [59] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_type,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [60] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [61] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [62] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [63] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [64] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [65] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [66] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [67] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [68] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [69] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [70] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [71] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [72] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [73] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [74] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_with,
    anon_sym_object,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [75] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [76] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [77] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [78] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [79] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [80] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [81] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [82] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [83] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_end,
    anon_sym_with,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [84] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [85] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [86] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [87] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [88] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [89] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [90] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [91] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [92] = {
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
  },
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [sym_attribute] = STATE(0),
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym__lowercase_identifier] = ACTIONS(1),
    [anon_sym_SEMI_SEMI] = ACTIONS(1),
    [anon_sym_let] = ACTIONS(1),
    [anon_sym_rec] = ACTIONS(1),
    [anon_sym_and] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_external] = ACTIONS(1),
    [anon_sym_type] = ACTIONS(1),
    [anon_sym_nonrec] = ACTIONS(1),
    [anon_sym_COLON_EQ] = ACTIONS(1),
    [anon_sym_private] = ACTIONS(1),
    [anon_sym_DOT_DOT] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym__] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_of] = ACTIONS(1),
    [anon_sym_COLON2] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_mutable] = ACTIONS(1),
    [anon_sym_constraint] = ACTIONS(1),
    [anon_sym_exception] = ACTIONS(1),
    [anon_sym_module] = ACTIONS(1),
    [anon_sym_open] = ACTIONS(1),
    [anon_sym_include] = ACTIONS(1),
    [anon_sym_class] = ACTIONS(1),
    [anon_sym_virtual] = ACTIONS(1),
    [anon_sym_val] = ACTIONS(1),
    [anon_sym_sig] = ACTIONS(1),
    [anon_sym_end] = ACTIONS(1),
    [anon_sym_with] = ACTIONS(1),
    [anon_sym_functor] = ACTIONS(1),
    [anon_sym_struct] = ACTIONS(1),
    [anon_sym_COLON_GT] = ACTIONS(1),
    [anon_sym_object] = ACTIONS(1),
    [anon_sym_inherit] = ACTIONS(1),
    [anon_sym_method] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_fun] = ACTIONS(1),
    [anon_sym_as] = ACTIONS(1),
    [anon_sym_initializer] = ACTIONS(1),
    [anon_sym_LBRACK_GT] = ACTIONS(1),
    [anon_sym_LBRACK_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_POUND] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1),
    [anon_sym_LT_DASH] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_then] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_done] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_to] = ACTIONS(1),
    [anon_sym_downto] = ACTIONS(1),
    [anon_sym_match] = ACTIONS(1),
    [anon_sym_when] = ACTIONS(1),
    [anon_sym_function] = ACTIONS(1),
    [anon_sym_try] = ACTIONS(1),
    [anon_sym_assert] = ACTIONS(1),
    [anon_sym_lazy] = ACTIONS(1),
    [anon_sym_new] = ACTIONS(1),
    [anon_sym_LBRACE_LT] = ACTIONS(1),
    [anon_sym_begin] = ACTIONS(1),
    [sym_ocamlyacc_value] = ACTIONS(1),
    [anon_sym_effect] = ACTIONS(1),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [aux_sym_number_token1] = ACTIONS(1),
    [aux_sym_signed_number_token1] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_SQUOTE2] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [aux_sym_string_content_token2] = ACTIONS(1),
    [aux_sym_string_content_token4] = ACTIONS(1),
    [aux_sym_string_content_token5] = ACTIONS(1),
    [aux_sym_escape_sequence_token1] = ACTIONS(1),
    [aux_sym_escape_sequence_token2] = ACTIONS(1),
    [aux_sym_escape_sequence_token3] = ACTIONS(1),
    [aux_sym_escape_sequence_token4] = ACTIONS(1),
    [sym_pretty_printing_indication] = ACTIONS(1),
    [aux_sym_sign_operator_token1] = ACTIONS(1),
    [sym_hash_operator] = ACTIONS(1),
    [anon_sym_lsl] = ACTIONS(1),
    [anon_sym_lsr] = ACTIONS(1),
    [anon_sym_asr] = ACTIONS(1),
    [aux_sym_pow_operator_token1] = ACTIONS(1),
    [anon_sym_mod] = ACTIONS(1),
    [anon_sym_land] = ACTIONS(1),
    [anon_sym_lor] = ACTIONS(1),
    [anon_sym_lxor] = ACTIONS(1),
    [sym_concat_operator] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_or] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [sym_assign_operator] = ACTIONS(1),
    [sym_let_operator] = ACTIONS(1),
    [sym_let_and_operator] = ACTIONS(1),
    [sym_match_operator] = ACTIONS(1),
    [sym__uppercase_identifier] = ACTIONS(1),
    [aux_sym_directive_token1] = ACTIONS(1),
    [aux_sym_type_variable_token1] = ACTIONS(1),
    [aux_sym_tag_token1] = ACTIONS(1),
    [aux_sym_attribute_id_token1] = ACTIONS(1),
    [sym_comment] = ACTIONS(5),
    [sym__left_quoted_string_delimiter] = ACTIONS(1),
    [sym__right_quoted_string_delimiter] = ACTIONS(1),
    [sym_line_number_directive] = ACTIONS(5),
    [sym__null] = ACTIONS(1),
    [sym__error_sentinel] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_compilation_unit] = STATE(6962),
    [sym__signature] = STATE(6990),
    [sym_external] = STATE(4204),
    [sym_type_definition] = STATE(4204),
    [sym_exception_definition] = STATE(4204),
    [sym_module_definition] = STATE(4204),
    [sym_module_type_definition] = STATE(4204),
    [sym_open_module] = STATE(4204),
    [sym_class_definition] = STATE(4204),
    [sym_class_type_definition] = STATE(4204),
    [sym__signature_item] = STATE(4229),
    [sym_value_specification] = STATE(4204),
    [sym_include_module_type] = STATE(4204),
    [sym_attribute] = STATE(1),
    [sym_floating_attribute] = STATE(4204),
    [sym_item_extension] = STATE(4204),
    [sym_quoted_item_extension] = STATE(4204),
    [aux_sym__structure_repeat1] = STATE(3191),
    [aux_sym__signature_repeat1] = STATE(3156),
    [ts_builtin_sym_end] = ACTIONS(7),
    [anon_sym_SEMI_SEMI] = ACTIONS(9),
    [anon_sym_external] = ACTIONS(11),
    [anon_sym_type] = ACTIONS(13),
    [anon_sym_exception] = ACTIONS(15),
    [anon_sym_module] = ACTIONS(17),
    [anon_sym_open] = ACTIONS(19),
    [anon_sym_include] = ACTIONS(21),
    [anon_sym_class] = ACTIONS(23),
    [anon_sym_val] = ACTIONS(25),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(2)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1924),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2285),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(2),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__attribute] = STATE(513),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_RPAREN] = ACTIONS(41),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_COLON2] = ACTIONS(43),
    [anon_sym_DASH_GT] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(41),
    [anon_sym_constraint] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_with] = ACTIONS(43),
    [anon_sym_COLON_GT] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_inherit] = ACTIONS(43),
    [anon_sym_method] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_initializer] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_then] = ACTIONS(43),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_GT_RBRACE] = ACTIONS(41),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(81),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(83),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(3)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1674),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1925),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(3),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__attribute] = STATE(516),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_SEMI] = ACTIONS(43),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_in] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(4)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1714),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1507),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(4),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__attribute] = STATE(510),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_SEMI] = ACTIONS(43),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(5)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2064),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1507),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(5),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__attribute] = STATE(500),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_SEMI] = ACTIONS(43),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(6)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(6),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6952),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(243),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(7)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(7),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7233),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(263),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(8)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(8),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6366),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(265),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(9)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(9),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7173),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(267),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(10)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(10),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7033),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(269),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(11)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(11),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7057),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(271),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(12)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(12),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6575),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(273),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(13)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(13),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6992),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(275),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(14)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(14),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7024),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(277),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(15)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(15),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6377),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(279),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(16)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(16),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6370),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(281),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(17)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(17),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6486),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(283),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(18)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(18),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6857),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(285),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(19)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(19),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7147),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(287),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(20)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(20),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6843),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(289),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(21)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(21),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6497),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(291),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(22)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(22),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7155),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(293),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(23)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(23),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7301),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(295),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(24)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(24),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6583),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(297),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(25)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(25),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6811),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(299),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(26)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(26),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6902),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(301),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(27)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(27),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7039),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(303),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(28)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(28),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7005),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(305),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(29)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(29),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7045),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(307),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(30)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(30),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7088),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(309),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(31)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(31),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7150),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(311),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(32)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(32),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7195),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(313),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(33)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(33),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(7220),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(315),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(34)] = {
    [sym__structure] = STATE(6960),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(34),
    [sym_floating_attribute] = STATE(4141),
    [sym_attribute_payload] = STATE(6630),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(317),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(35)] = {
    [sym_expression_item] = STATE(4156),
    [sym_toplevel_directive] = STATE(4156),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(4156),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(35),
    [sym_floating_attribute] = STATE(4141),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(1655),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(319),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(321),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(36)] = {
    [sym_expression_item] = STATE(4156),
    [sym_toplevel_directive] = STATE(4156),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(4156),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(36),
    [sym_floating_attribute] = STATE(4141),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(1655),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(323),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(325),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(37)] = {
    [sym_expression_item] = STATE(2624),
    [sym_toplevel_directive] = STATE(2624),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2624),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(37),
    [sym_floating_attribute] = STATE(4141),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(1655),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(327),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(329),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(38)] = {
    [sym__structure] = STATE(6493),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(38),
    [sym_floating_attribute] = STATE(4141),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(331),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(39)] = {
    [sym__structure] = STATE(7326),
    [sym_expression_item] = STATE(2596),
    [sym_toplevel_directive] = STATE(2596),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(2596),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(39),
    [sym_floating_attribute] = STATE(4141),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(37),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(333),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(40)] = {
    [sym_expression_item] = STATE(4156),
    [sym_toplevel_directive] = STATE(4156),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(4156),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(40),
    [sym_floating_attribute] = STATE(4141),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(1655),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(335),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(337),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(41)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2565),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2285),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(41),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(509),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_constraint] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_inherit] = ACTIONS(43),
    [anon_sym_method] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_initializer] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(381),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(383),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(42)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2546),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2511),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(42),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(504),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(41),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_COLON2] = ACTIONS(43),
    [anon_sym_DASH_GT] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(41),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_with] = ACTIONS(43),
    [anon_sym_COLON_GT] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_then] = ACTIONS(43),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_GT_RBRACE] = ACTIONS(41),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(43)] = {
    [sym_expression_item] = STATE(4156),
    [sym_toplevel_directive] = STATE(4156),
    [sym__local_structure_item] = STATE(4139),
    [sym__structure_item] = STATE(4156),
    [sym_value_definition] = STATE(4077),
    [sym_external] = STATE(4141),
    [sym_type_definition] = STATE(4141),
    [sym_exception_definition] = STATE(4141),
    [sym_module_definition] = STATE(4141),
    [sym_module_type_definition] = STATE(4141),
    [sym_open_module] = STATE(4141),
    [sym_include_module] = STATE(4139),
    [sym_class_definition] = STATE(4141),
    [sym_class_type_definition] = STATE(4141),
    [sym_value_specification] = STATE(4139),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2163),
    [sym__sequence_expression] = STATE(3867),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(43),
    [sym_floating_attribute] = STATE(4141),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym_item_extension] = STATE(4141),
    [sym_quoted_item_extension] = STATE(4141),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_directive] = STATE(2652),
    [sym_tag] = STATE(1131),
    [aux_sym__structure_repeat1] = STATE(1655),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(259),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(261),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(44)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(4941),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(44),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(415),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(419),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(423),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(441),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(443),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(45)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(5035),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(45),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6728),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(445),
    [anon_sym_DOT] = ACTIONS(447),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(449),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(451),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(453),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(455),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(457),
    [sym_let_and_operator] = ACTIONS(455),
    [sym_match_operator] = ACTIONS(459),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(46)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7236),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(46),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(7051),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(461),
    [anon_sym_DOT] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(415),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(465),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(467),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(469),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(471),
    [sym_let_and_operator] = ACTIONS(469),
    [sym_match_operator] = ACTIONS(473),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(47)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(4941),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(47),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(7051),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(461),
    [anon_sym_DOT] = ACTIONS(463),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(415),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(465),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(467),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(469),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(471),
    [sym_let_and_operator] = ACTIONS(469),
    [sym_match_operator] = ACTIONS(473),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(48)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6913),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(48),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6728),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(445),
    [anon_sym_DOT] = ACTIONS(447),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(449),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(451),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(453),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(455),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(457),
    [sym_let_and_operator] = ACTIONS(455),
    [sym_match_operator] = ACTIONS(459),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(49)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6980),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(49),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6507),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(475),
    [anon_sym_DOT] = ACTIONS(477),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(479),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(481),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(483),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(485),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(487),
    [sym_let_and_operator] = ACTIONS(485),
    [sym_match_operator] = ACTIONS(489),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(50)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(5001),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(50),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6507),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(475),
    [anon_sym_DOT] = ACTIONS(477),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(479),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(481),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(483),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(485),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(487),
    [sym_let_and_operator] = ACTIONS(485),
    [sym_match_operator] = ACTIONS(489),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(51)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6469),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(51),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6874),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(491),
    [anon_sym_DOT] = ACTIONS(493),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(497),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(499),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(501),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(503),
    [sym_let_and_operator] = ACTIONS(501),
    [sym_match_operator] = ACTIONS(505),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(52)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(4920),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(52),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6874),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(491),
    [anon_sym_DOT] = ACTIONS(493),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(495),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(497),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(499),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(501),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(503),
    [sym_let_and_operator] = ACTIONS(501),
    [sym_match_operator] = ACTIONS(505),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(53)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(4964),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(53),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(7034),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(507),
    [anon_sym_DOT] = ACTIONS(509),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(511),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(515),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(517),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(519),
    [sym_let_and_operator] = ACTIONS(517),
    [sym_match_operator] = ACTIONS(521),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(54)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7249),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(54),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6936),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(523),
    [anon_sym_DOT] = ACTIONS(525),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(527),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(529),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(531),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(533),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(535),
    [sym_let_and_operator] = ACTIONS(533),
    [sym_match_operator] = ACTIONS(537),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(55)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(4838),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(55),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6936),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(523),
    [anon_sym_DOT] = ACTIONS(525),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(527),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(529),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(531),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(533),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(535),
    [sym_let_and_operator] = ACTIONS(533),
    [sym_match_operator] = ACTIONS(537),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(56)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7098),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(56),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6982),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(539),
    [anon_sym_DOT] = ACTIONS(541),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(543),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(545),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(547),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(549),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(551),
    [sym_let_and_operator] = ACTIONS(549),
    [sym_match_operator] = ACTIONS(553),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(57)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(4895),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(57),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6982),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(539),
    [anon_sym_DOT] = ACTIONS(541),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(543),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(545),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(547),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(549),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(551),
    [sym_let_and_operator] = ACTIONS(549),
    [sym_match_operator] = ACTIONS(553),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(58)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6695),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(58),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(7031),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(555),
    [anon_sym_DOT] = ACTIONS(557),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(559),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(561),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(563),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(565),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(567),
    [sym_let_and_operator] = ACTIONS(565),
    [sym_match_operator] = ACTIONS(569),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(59)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(4934),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(59),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(7031),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(555),
    [anon_sym_DOT] = ACTIONS(557),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(559),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(561),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(563),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(565),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(567),
    [sym_let_and_operator] = ACTIONS(565),
    [sym_match_operator] = ACTIONS(569),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(60)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7020),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(60),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(7070),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(571),
    [anon_sym_DOT] = ACTIONS(573),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(575),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(577),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(579),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(581),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(583),
    [sym_let_and_operator] = ACTIONS(581),
    [sym_match_operator] = ACTIONS(585),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(61)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(4978),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(61),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(7070),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(571),
    [anon_sym_DOT] = ACTIONS(573),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(575),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(577),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(579),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(581),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(583),
    [sym_let_and_operator] = ACTIONS(581),
    [sym_match_operator] = ACTIONS(585),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(62)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6974),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(62),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(7130),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(587),
    [anon_sym_DOT] = ACTIONS(589),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(591),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(593),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(595),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(597),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(599),
    [sym_let_and_operator] = ACTIONS(597),
    [sym_match_operator] = ACTIONS(601),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(63)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(4829),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(63),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(7130),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(587),
    [anon_sym_DOT] = ACTIONS(589),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(591),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(593),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(595),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(597),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(599),
    [sym_let_and_operator] = ACTIONS(597),
    [sym_match_operator] = ACTIONS(601),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(64)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7200),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(64),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(7034),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(507),
    [anon_sym_DOT] = ACTIONS(509),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(511),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(513),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(515),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(517),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(519),
    [sym_let_and_operator] = ACTIONS(517),
    [sym_match_operator] = ACTIONS(521),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(65)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7236),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(65),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_module] = ACTIONS(415),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_COLON_COLON] = ACTIONS(419),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(421),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(423),
    [aux_sym_sign_operator_token1] = ACTIONS(421),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(441),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(443),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(66)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2789),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2910),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(66),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__attribute] = STATE(522),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_in] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(67)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2863),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2511),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(67),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__attribute] = STATE(515),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_in] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(68)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2872),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2910),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(68),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__attribute] = STATE(502),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_done] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_to] = ACTIONS(43),
    [anon_sym_downto] = ACTIONS(43),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(69)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2894),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2910),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(69),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(503),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_done] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_to] = ACTIONS(43),
    [anon_sym_downto] = ACTIONS(43),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(70)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3013),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2511),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(70),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(512),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(41),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(71)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3049),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2285),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(71),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__attribute] = STATE(525),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(72)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3182),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2511),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(72),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(508),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(73)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1920),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1925),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(73),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__attribute] = STATE(507),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_in] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(74)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(2065),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1507),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(74),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__attribute] = STATE(501),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(75)] = {
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(860),
    [sym_typed_expression] = STATE(803),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_labeled_argument] = STATE(861),
    [sym_prefix_expression] = STATE(803),
    [sym_hash_expression] = STATE(803),
    [sym_field_get_expression] = STATE(803),
    [sym_array_get_expression] = STATE(803),
    [sym_string_get_expression] = STATE(803),
    [sym_bigarray_get_expression] = STATE(803),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(75),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [aux_sym_class_application_repeat1] = STATE(79),
    [sym__lowercase_identifier] = ACTIONS(737),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_QMARK] = ACTIONS(739),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(741),
    [anon_sym_COMMA] = ACTIONS(741),
    [anon_sym_PIPE] = ACTIONS(743),
    [anon_sym_COLON2] = ACTIONS(743),
    [anon_sym_DOT] = ACTIONS(745),
    [anon_sym_DASH_GT] = ACTIONS(743),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(741),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_SEMI] = ACTIONS(741),
    [anon_sym_RBRACE] = ACTIONS(741),
    [anon_sym_constraint] = ACTIONS(743),
    [anon_sym_val] = ACTIONS(743),
    [anon_sym_end] = ACTIONS(743),
    [anon_sym_with] = ACTIONS(743),
    [anon_sym_COLON_GT] = ACTIONS(741),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_inherit] = ACTIONS(743),
    [anon_sym_method] = ACTIONS(743),
    [anon_sym_initializer] = ACTIONS(743),
    [anon_sym_AMP] = ACTIONS(743),
    [anon_sym_POUND] = ACTIONS(747),
    [anon_sym_COLON_COLON] = ACTIONS(741),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(741),
    [anon_sym_then] = ACTIONS(743),
    [anon_sym_else] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_done] = ACTIONS(743),
    [anon_sym_to] = ACTIONS(743),
    [anon_sym_downto] = ACTIONS(743),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_GT_RBRACE] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(743),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(741),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(381),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(383),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(741),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(743),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(743),
    [sym_hash_operator] = ACTIONS(749),
    [anon_sym_lsl] = ACTIONS(743),
    [anon_sym_lsr] = ACTIONS(743),
    [anon_sym_asr] = ACTIONS(743),
    [aux_sym_pow_operator_token1] = ACTIONS(741),
    [anon_sym_mod] = ACTIONS(743),
    [anon_sym_land] = ACTIONS(743),
    [anon_sym_lor] = ACTIONS(743),
    [anon_sym_lxor] = ACTIONS(743),
    [aux_sym_mult_operator_token1] = ACTIONS(743),
    [aux_sym_add_operator_token1] = ACTIONS(743),
    [sym_concat_operator] = ACTIONS(741),
    [sym_rel_operator] = ACTIONS(743),
    [anon_sym_AMP_AMP] = ACTIONS(743),
    [anon_sym_or] = ACTIONS(743),
    [anon_sym_PIPE_PIPE] = ACTIONS(743),
    [sym_assign_operator] = ACTIONS(741),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(76)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2239),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1507),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(76),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__attribute] = STATE(520),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(77)] = {
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(950),
    [sym_typed_expression] = STATE(939),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_labeled_argument] = STATE(1054),
    [sym_prefix_expression] = STATE(939),
    [sym_hash_expression] = STATE(939),
    [sym_field_get_expression] = STATE(939),
    [sym_array_get_expression] = STATE(939),
    [sym_string_get_expression] = STATE(939),
    [sym_bigarray_get_expression] = STATE(939),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(77),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [aux_sym_class_application_repeat1] = STATE(81),
    [sym__lowercase_identifier] = ACTIONS(751),
    [anon_sym_TILDE] = ACTIONS(753),
    [anon_sym_QMARK] = ACTIONS(753),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_RPAREN] = ACTIONS(741),
    [anon_sym_COMMA] = ACTIONS(741),
    [anon_sym_PIPE] = ACTIONS(743),
    [anon_sym_COLON2] = ACTIONS(743),
    [anon_sym_DOT] = ACTIONS(755),
    [anon_sym_DASH_GT] = ACTIONS(743),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_RBRACK] = ACTIONS(741),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_SEMI] = ACTIONS(741),
    [anon_sym_RBRACE] = ACTIONS(741),
    [anon_sym_constraint] = ACTIONS(743),
    [anon_sym_val] = ACTIONS(743),
    [anon_sym_end] = ACTIONS(743),
    [anon_sym_with] = ACTIONS(743),
    [anon_sym_COLON_GT] = ACTIONS(741),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_inherit] = ACTIONS(743),
    [anon_sym_method] = ACTIONS(743),
    [anon_sym_initializer] = ACTIONS(743),
    [anon_sym_AMP] = ACTIONS(743),
    [anon_sym_POUND] = ACTIONS(757),
    [anon_sym_COLON_COLON] = ACTIONS(741),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_then] = ACTIONS(743),
    [anon_sym_else] = ACTIONS(743),
    [anon_sym_do] = ACTIONS(743),
    [anon_sym_done] = ACTIONS(743),
    [anon_sym_to] = ACTIONS(743),
    [anon_sym_downto] = ACTIONS(743),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_GT_RBRACE] = ACTIONS(741),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(743),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(741),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(81),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(83),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(741),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(743),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(743),
    [sym_hash_operator] = ACTIONS(759),
    [anon_sym_lsl] = ACTIONS(743),
    [anon_sym_lsr] = ACTIONS(743),
    [anon_sym_asr] = ACTIONS(743),
    [aux_sym_pow_operator_token1] = ACTIONS(741),
    [anon_sym_mod] = ACTIONS(743),
    [anon_sym_land] = ACTIONS(743),
    [anon_sym_lor] = ACTIONS(743),
    [anon_sym_lxor] = ACTIONS(743),
    [aux_sym_mult_operator_token1] = ACTIONS(743),
    [aux_sym_add_operator_token1] = ACTIONS(743),
    [sym_concat_operator] = ACTIONS(741),
    [sym_rel_operator] = ACTIONS(743),
    [anon_sym_AMP_AMP] = ACTIONS(743),
    [anon_sym_or] = ACTIONS(743),
    [anon_sym_PIPE_PIPE] = ACTIONS(743),
    [sym_assign_operator] = ACTIONS(741),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(78)] = {
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(981),
    [sym_typed_expression] = STATE(990),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_labeled_argument] = STATE(1147),
    [sym_prefix_expression] = STATE(990),
    [sym_hash_expression] = STATE(990),
    [sym_field_get_expression] = STATE(990),
    [sym_array_get_expression] = STATE(990),
    [sym_string_get_expression] = STATE(990),
    [sym_bigarray_get_expression] = STATE(990),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(78),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [aux_sym_class_application_repeat1] = STATE(84),
    [sym__lowercase_identifier] = ACTIONS(761),
    [anon_sym_SEMI_SEMI] = ACTIONS(741),
    [anon_sym_let] = ACTIONS(743),
    [anon_sym_and] = ACTIONS(743),
    [anon_sym_TILDE] = ACTIONS(763),
    [anon_sym_QMARK] = ACTIONS(763),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(743),
    [anon_sym_type] = ACTIONS(743),
    [anon_sym_COMMA] = ACTIONS(741),
    [anon_sym_PIPE] = ACTIONS(743),
    [anon_sym_DOT] = ACTIONS(765),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(741),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_SEMI] = ACTIONS(743),
    [anon_sym_exception] = ACTIONS(743),
    [anon_sym_module] = ACTIONS(743),
    [anon_sym_open] = ACTIONS(743),
    [anon_sym_include] = ACTIONS(743),
    [anon_sym_class] = ACTIONS(743),
    [anon_sym_val] = ACTIONS(743),
    [anon_sym_end] = ACTIONS(743),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_in] = ACTIONS(743),
    [anon_sym_AMP] = ACTIONS(743),
    [anon_sym_POUND] = ACTIONS(767),
    [anon_sym_COLON_COLON] = ACTIONS(741),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_else] = ACTIONS(743),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(743),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(741),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(741),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(743),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(743),
    [sym_hash_operator] = ACTIONS(769),
    [anon_sym_lsl] = ACTIONS(743),
    [anon_sym_lsr] = ACTIONS(743),
    [anon_sym_asr] = ACTIONS(743),
    [aux_sym_pow_operator_token1] = ACTIONS(741),
    [anon_sym_mod] = ACTIONS(743),
    [anon_sym_land] = ACTIONS(743),
    [anon_sym_lor] = ACTIONS(743),
    [anon_sym_lxor] = ACTIONS(743),
    [aux_sym_mult_operator_token1] = ACTIONS(743),
    [aux_sym_add_operator_token1] = ACTIONS(743),
    [sym_concat_operator] = ACTIONS(741),
    [sym_rel_operator] = ACTIONS(743),
    [anon_sym_AMP_AMP] = ACTIONS(743),
    [anon_sym_or] = ACTIONS(743),
    [anon_sym_PIPE_PIPE] = ACTIONS(743),
    [sym_assign_operator] = ACTIONS(741),
    [sym_let_operator] = ACTIONS(741),
    [sym_let_and_operator] = ACTIONS(741),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(743),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(79)] = {
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(860),
    [sym_typed_expression] = STATE(803),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_labeled_argument] = STATE(861),
    [sym_prefix_expression] = STATE(803),
    [sym_hash_expression] = STATE(803),
    [sym_field_get_expression] = STATE(803),
    [sym_array_get_expression] = STATE(803),
    [sym_string_get_expression] = STATE(803),
    [sym_bigarray_get_expression] = STATE(803),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(79),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [aux_sym_class_application_repeat1] = STATE(80),
    [sym__lowercase_identifier] = ACTIONS(737),
    [anon_sym_TILDE] = ACTIONS(739),
    [anon_sym_QMARK] = ACTIONS(739),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(771),
    [anon_sym_COMMA] = ACTIONS(771),
    [anon_sym_PIPE] = ACTIONS(773),
    [anon_sym_COLON2] = ACTIONS(773),
    [anon_sym_DASH_GT] = ACTIONS(773),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(771),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_SEMI] = ACTIONS(771),
    [anon_sym_RBRACE] = ACTIONS(771),
    [anon_sym_constraint] = ACTIONS(773),
    [anon_sym_val] = ACTIONS(773),
    [anon_sym_end] = ACTIONS(773),
    [anon_sym_with] = ACTIONS(773),
    [anon_sym_COLON_GT] = ACTIONS(771),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_inherit] = ACTIONS(773),
    [anon_sym_method] = ACTIONS(773),
    [anon_sym_initializer] = ACTIONS(773),
    [anon_sym_AMP] = ACTIONS(773),
    [anon_sym_COLON_COLON] = ACTIONS(771),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(771),
    [anon_sym_then] = ACTIONS(773),
    [anon_sym_else] = ACTIONS(773),
    [anon_sym_do] = ACTIONS(773),
    [anon_sym_done] = ACTIONS(773),
    [anon_sym_to] = ACTIONS(773),
    [anon_sym_downto] = ACTIONS(773),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_GT_RBRACE] = ACTIONS(771),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(773),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(771),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(381),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(383),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(771),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(771),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(773),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(773),
    [anon_sym_lsl] = ACTIONS(773),
    [anon_sym_lsr] = ACTIONS(773),
    [anon_sym_asr] = ACTIONS(773),
    [aux_sym_pow_operator_token1] = ACTIONS(771),
    [anon_sym_mod] = ACTIONS(773),
    [anon_sym_land] = ACTIONS(773),
    [anon_sym_lor] = ACTIONS(773),
    [anon_sym_lxor] = ACTIONS(773),
    [aux_sym_mult_operator_token1] = ACTIONS(773),
    [aux_sym_add_operator_token1] = ACTIONS(773),
    [sym_concat_operator] = ACTIONS(771),
    [sym_rel_operator] = ACTIONS(773),
    [anon_sym_AMP_AMP] = ACTIONS(773),
    [anon_sym_or] = ACTIONS(773),
    [anon_sym_PIPE_PIPE] = ACTIONS(773),
    [sym_assign_operator] = ACTIONS(771),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(80)] = {
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(860),
    [sym_typed_expression] = STATE(803),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_labeled_argument] = STATE(861),
    [sym_prefix_expression] = STATE(803),
    [sym_hash_expression] = STATE(803),
    [sym_field_get_expression] = STATE(803),
    [sym_array_get_expression] = STATE(803),
    [sym_string_get_expression] = STATE(803),
    [sym_bigarray_get_expression] = STATE(803),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(80),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [aux_sym_class_application_repeat1] = STATE(80),
    [sym__lowercase_identifier] = ACTIONS(775),
    [anon_sym_TILDE] = ACTIONS(778),
    [anon_sym_QMARK] = ACTIONS(778),
    [anon_sym_LPAREN] = ACTIONS(781),
    [anon_sym_RPAREN] = ACTIONS(784),
    [anon_sym_COMMA] = ACTIONS(784),
    [anon_sym_PIPE] = ACTIONS(786),
    [anon_sym_COLON2] = ACTIONS(786),
    [anon_sym_DASH_GT] = ACTIONS(786),
    [anon_sym_LBRACK] = ACTIONS(788),
    [anon_sym_RBRACK] = ACTIONS(784),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [anon_sym_LBRACE] = ACTIONS(794),
    [anon_sym_SEMI] = ACTIONS(784),
    [anon_sym_RBRACE] = ACTIONS(784),
    [anon_sym_constraint] = ACTIONS(786),
    [anon_sym_val] = ACTIONS(786),
    [anon_sym_end] = ACTIONS(786),
    [anon_sym_with] = ACTIONS(786),
    [anon_sym_COLON_GT] = ACTIONS(784),
    [anon_sym_object] = ACTIONS(797),
    [anon_sym_inherit] = ACTIONS(786),
    [anon_sym_method] = ACTIONS(786),
    [anon_sym_initializer] = ACTIONS(786),
    [anon_sym_AMP] = ACTIONS(786),
    [anon_sym_COLON_COLON] = ACTIONS(784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(800),
    [anon_sym_PIPE_RBRACK] = ACTIONS(784),
    [anon_sym_then] = ACTIONS(786),
    [anon_sym_else] = ACTIONS(786),
    [anon_sym_do] = ACTIONS(786),
    [anon_sym_done] = ACTIONS(786),
    [anon_sym_to] = ACTIONS(786),
    [anon_sym_downto] = ACTIONS(786),
    [anon_sym_new] = ACTIONS(803),
    [anon_sym_LBRACE_LT] = ACTIONS(806),
    [anon_sym_GT_RBRACE] = ACTIONS(784),
    [anon_sym_begin] = ACTIONS(809),
    [sym_ocamlyacc_value] = ACTIONS(812),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(786),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(784),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(815),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(818),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(784),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(784),
    [aux_sym_number_token1] = ACTIONS(821),
    [aux_sym_signed_number_token1] = ACTIONS(786),
    [anon_sym_SQUOTE] = ACTIONS(824),
    [anon_sym_DQUOTE] = ACTIONS(827),
    [sym_prefix_operator] = ACTIONS(830),
    [aux_sym_sign_operator_token1] = ACTIONS(786),
    [anon_sym_lsl] = ACTIONS(786),
    [anon_sym_lsr] = ACTIONS(786),
    [anon_sym_asr] = ACTIONS(786),
    [aux_sym_pow_operator_token1] = ACTIONS(784),
    [anon_sym_mod] = ACTIONS(786),
    [anon_sym_land] = ACTIONS(786),
    [anon_sym_lor] = ACTIONS(786),
    [anon_sym_lxor] = ACTIONS(786),
    [aux_sym_mult_operator_token1] = ACTIONS(786),
    [aux_sym_add_operator_token1] = ACTIONS(786),
    [sym_concat_operator] = ACTIONS(784),
    [sym_rel_operator] = ACTIONS(786),
    [anon_sym_AMP_AMP] = ACTIONS(786),
    [anon_sym_or] = ACTIONS(786),
    [anon_sym_PIPE_PIPE] = ACTIONS(786),
    [sym_assign_operator] = ACTIONS(784),
    [sym__uppercase_identifier] = ACTIONS(833),
    [aux_sym_tag_token1] = ACTIONS(836),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(81)] = {
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(950),
    [sym_typed_expression] = STATE(939),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_labeled_argument] = STATE(1054),
    [sym_prefix_expression] = STATE(939),
    [sym_hash_expression] = STATE(939),
    [sym_field_get_expression] = STATE(939),
    [sym_array_get_expression] = STATE(939),
    [sym_string_get_expression] = STATE(939),
    [sym_bigarray_get_expression] = STATE(939),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(81),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [aux_sym_class_application_repeat1] = STATE(82),
    [sym__lowercase_identifier] = ACTIONS(751),
    [anon_sym_TILDE] = ACTIONS(753),
    [anon_sym_QMARK] = ACTIONS(753),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_RPAREN] = ACTIONS(771),
    [anon_sym_COMMA] = ACTIONS(771),
    [anon_sym_PIPE] = ACTIONS(773),
    [anon_sym_COLON2] = ACTIONS(773),
    [anon_sym_DASH_GT] = ACTIONS(773),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_RBRACK] = ACTIONS(771),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_SEMI] = ACTIONS(771),
    [anon_sym_RBRACE] = ACTIONS(771),
    [anon_sym_constraint] = ACTIONS(773),
    [anon_sym_val] = ACTIONS(773),
    [anon_sym_end] = ACTIONS(773),
    [anon_sym_with] = ACTIONS(773),
    [anon_sym_COLON_GT] = ACTIONS(771),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_inherit] = ACTIONS(773),
    [anon_sym_method] = ACTIONS(773),
    [anon_sym_initializer] = ACTIONS(773),
    [anon_sym_AMP] = ACTIONS(773),
    [anon_sym_COLON_COLON] = ACTIONS(771),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_then] = ACTIONS(773),
    [anon_sym_else] = ACTIONS(773),
    [anon_sym_do] = ACTIONS(773),
    [anon_sym_done] = ACTIONS(773),
    [anon_sym_to] = ACTIONS(773),
    [anon_sym_downto] = ACTIONS(773),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_GT_RBRACE] = ACTIONS(771),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(773),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(771),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(81),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(83),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(771),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(771),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(773),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(773),
    [anon_sym_lsl] = ACTIONS(773),
    [anon_sym_lsr] = ACTIONS(773),
    [anon_sym_asr] = ACTIONS(773),
    [aux_sym_pow_operator_token1] = ACTIONS(771),
    [anon_sym_mod] = ACTIONS(773),
    [anon_sym_land] = ACTIONS(773),
    [anon_sym_lor] = ACTIONS(773),
    [anon_sym_lxor] = ACTIONS(773),
    [aux_sym_mult_operator_token1] = ACTIONS(773),
    [aux_sym_add_operator_token1] = ACTIONS(773),
    [sym_concat_operator] = ACTIONS(771),
    [sym_rel_operator] = ACTIONS(773),
    [anon_sym_AMP_AMP] = ACTIONS(773),
    [anon_sym_or] = ACTIONS(773),
    [anon_sym_PIPE_PIPE] = ACTIONS(773),
    [sym_assign_operator] = ACTIONS(771),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(82)] = {
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(950),
    [sym_typed_expression] = STATE(939),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_labeled_argument] = STATE(1054),
    [sym_prefix_expression] = STATE(939),
    [sym_hash_expression] = STATE(939),
    [sym_field_get_expression] = STATE(939),
    [sym_array_get_expression] = STATE(939),
    [sym_string_get_expression] = STATE(939),
    [sym_bigarray_get_expression] = STATE(939),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(82),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [aux_sym_class_application_repeat1] = STATE(82),
    [sym__lowercase_identifier] = ACTIONS(839),
    [anon_sym_TILDE] = ACTIONS(842),
    [anon_sym_QMARK] = ACTIONS(842),
    [anon_sym_LPAREN] = ACTIONS(845),
    [anon_sym_RPAREN] = ACTIONS(784),
    [anon_sym_COMMA] = ACTIONS(784),
    [anon_sym_PIPE] = ACTIONS(786),
    [anon_sym_COLON2] = ACTIONS(786),
    [anon_sym_DASH_GT] = ACTIONS(786),
    [anon_sym_LBRACK] = ACTIONS(848),
    [anon_sym_RBRACK] = ACTIONS(784),
    [anon_sym_true] = ACTIONS(851),
    [anon_sym_false] = ACTIONS(851),
    [anon_sym_LBRACE] = ACTIONS(854),
    [anon_sym_SEMI] = ACTIONS(784),
    [anon_sym_RBRACE] = ACTIONS(784),
    [anon_sym_constraint] = ACTIONS(786),
    [anon_sym_val] = ACTIONS(786),
    [anon_sym_end] = ACTIONS(786),
    [anon_sym_with] = ACTIONS(786),
    [anon_sym_COLON_GT] = ACTIONS(784),
    [anon_sym_object] = ACTIONS(857),
    [anon_sym_inherit] = ACTIONS(786),
    [anon_sym_method] = ACTIONS(786),
    [anon_sym_initializer] = ACTIONS(786),
    [anon_sym_AMP] = ACTIONS(786),
    [anon_sym_COLON_COLON] = ACTIONS(784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(860),
    [anon_sym_then] = ACTIONS(786),
    [anon_sym_else] = ACTIONS(786),
    [anon_sym_do] = ACTIONS(786),
    [anon_sym_done] = ACTIONS(786),
    [anon_sym_to] = ACTIONS(786),
    [anon_sym_downto] = ACTIONS(786),
    [anon_sym_new] = ACTIONS(863),
    [anon_sym_LBRACE_LT] = ACTIONS(866),
    [anon_sym_GT_RBRACE] = ACTIONS(784),
    [anon_sym_begin] = ACTIONS(869),
    [sym_ocamlyacc_value] = ACTIONS(872),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(786),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(784),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(875),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(878),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(784),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(784),
    [aux_sym_number_token1] = ACTIONS(881),
    [aux_sym_signed_number_token1] = ACTIONS(786),
    [anon_sym_SQUOTE] = ACTIONS(884),
    [anon_sym_DQUOTE] = ACTIONS(887),
    [sym_prefix_operator] = ACTIONS(890),
    [aux_sym_sign_operator_token1] = ACTIONS(786),
    [anon_sym_lsl] = ACTIONS(786),
    [anon_sym_lsr] = ACTIONS(786),
    [anon_sym_asr] = ACTIONS(786),
    [aux_sym_pow_operator_token1] = ACTIONS(784),
    [anon_sym_mod] = ACTIONS(786),
    [anon_sym_land] = ACTIONS(786),
    [anon_sym_lor] = ACTIONS(786),
    [anon_sym_lxor] = ACTIONS(786),
    [aux_sym_mult_operator_token1] = ACTIONS(786),
    [aux_sym_add_operator_token1] = ACTIONS(786),
    [sym_concat_operator] = ACTIONS(784),
    [sym_rel_operator] = ACTIONS(786),
    [anon_sym_AMP_AMP] = ACTIONS(786),
    [anon_sym_or] = ACTIONS(786),
    [anon_sym_PIPE_PIPE] = ACTIONS(786),
    [sym_assign_operator] = ACTIONS(784),
    [sym__uppercase_identifier] = ACTIONS(893),
    [aux_sym_tag_token1] = ACTIONS(896),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(83)] = {
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(981),
    [sym_typed_expression] = STATE(990),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_labeled_argument] = STATE(1147),
    [sym_prefix_expression] = STATE(990),
    [sym_hash_expression] = STATE(990),
    [sym_field_get_expression] = STATE(990),
    [sym_array_get_expression] = STATE(990),
    [sym_string_get_expression] = STATE(990),
    [sym_bigarray_get_expression] = STATE(990),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(83),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [aux_sym_class_application_repeat1] = STATE(83),
    [sym__lowercase_identifier] = ACTIONS(899),
    [anon_sym_SEMI_SEMI] = ACTIONS(784),
    [anon_sym_let] = ACTIONS(786),
    [anon_sym_and] = ACTIONS(786),
    [anon_sym_TILDE] = ACTIONS(902),
    [anon_sym_QMARK] = ACTIONS(902),
    [anon_sym_LPAREN] = ACTIONS(905),
    [anon_sym_external] = ACTIONS(786),
    [anon_sym_type] = ACTIONS(786),
    [anon_sym_COMMA] = ACTIONS(784),
    [anon_sym_PIPE] = ACTIONS(786),
    [anon_sym_LBRACK] = ACTIONS(908),
    [anon_sym_RBRACK] = ACTIONS(784),
    [anon_sym_true] = ACTIONS(911),
    [anon_sym_false] = ACTIONS(911),
    [anon_sym_LBRACE] = ACTIONS(914),
    [anon_sym_SEMI] = ACTIONS(786),
    [anon_sym_exception] = ACTIONS(786),
    [anon_sym_module] = ACTIONS(786),
    [anon_sym_open] = ACTIONS(786),
    [anon_sym_include] = ACTIONS(786),
    [anon_sym_class] = ACTIONS(786),
    [anon_sym_val] = ACTIONS(786),
    [anon_sym_end] = ACTIONS(786),
    [anon_sym_object] = ACTIONS(917),
    [anon_sym_in] = ACTIONS(786),
    [anon_sym_AMP] = ACTIONS(786),
    [anon_sym_COLON_COLON] = ACTIONS(784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(920),
    [anon_sym_else] = ACTIONS(786),
    [anon_sym_new] = ACTIONS(923),
    [anon_sym_LBRACE_LT] = ACTIONS(926),
    [anon_sym_begin] = ACTIONS(929),
    [sym_ocamlyacc_value] = ACTIONS(932),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(786),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(784),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(935),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(938),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(784),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(784),
    [aux_sym_number_token1] = ACTIONS(941),
    [aux_sym_signed_number_token1] = ACTIONS(786),
    [anon_sym_SQUOTE] = ACTIONS(944),
    [anon_sym_DQUOTE] = ACTIONS(947),
    [sym_prefix_operator] = ACTIONS(950),
    [aux_sym_sign_operator_token1] = ACTIONS(786),
    [anon_sym_lsl] = ACTIONS(786),
    [anon_sym_lsr] = ACTIONS(786),
    [anon_sym_asr] = ACTIONS(786),
    [aux_sym_pow_operator_token1] = ACTIONS(784),
    [anon_sym_mod] = ACTIONS(786),
    [anon_sym_land] = ACTIONS(786),
    [anon_sym_lor] = ACTIONS(786),
    [anon_sym_lxor] = ACTIONS(786),
    [aux_sym_mult_operator_token1] = ACTIONS(786),
    [aux_sym_add_operator_token1] = ACTIONS(786),
    [sym_concat_operator] = ACTIONS(784),
    [sym_rel_operator] = ACTIONS(786),
    [anon_sym_AMP_AMP] = ACTIONS(786),
    [anon_sym_or] = ACTIONS(786),
    [anon_sym_PIPE_PIPE] = ACTIONS(786),
    [sym_assign_operator] = ACTIONS(784),
    [sym_let_operator] = ACTIONS(784),
    [sym_let_and_operator] = ACTIONS(784),
    [sym__uppercase_identifier] = ACTIONS(953),
    [aux_sym_directive_token1] = ACTIONS(784),
    [aux_sym_tag_token1] = ACTIONS(956),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(84)] = {
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(981),
    [sym_typed_expression] = STATE(990),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_labeled_argument] = STATE(1147),
    [sym_prefix_expression] = STATE(990),
    [sym_hash_expression] = STATE(990),
    [sym_field_get_expression] = STATE(990),
    [sym_array_get_expression] = STATE(990),
    [sym_string_get_expression] = STATE(990),
    [sym_bigarray_get_expression] = STATE(990),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(84),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [aux_sym_class_application_repeat1] = STATE(83),
    [sym__lowercase_identifier] = ACTIONS(761),
    [anon_sym_SEMI_SEMI] = ACTIONS(771),
    [anon_sym_let] = ACTIONS(773),
    [anon_sym_and] = ACTIONS(773),
    [anon_sym_TILDE] = ACTIONS(763),
    [anon_sym_QMARK] = ACTIONS(763),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(773),
    [anon_sym_type] = ACTIONS(773),
    [anon_sym_COMMA] = ACTIONS(771),
    [anon_sym_PIPE] = ACTIONS(773),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(771),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_SEMI] = ACTIONS(773),
    [anon_sym_exception] = ACTIONS(773),
    [anon_sym_module] = ACTIONS(773),
    [anon_sym_open] = ACTIONS(773),
    [anon_sym_include] = ACTIONS(773),
    [anon_sym_class] = ACTIONS(773),
    [anon_sym_val] = ACTIONS(773),
    [anon_sym_end] = ACTIONS(773),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_in] = ACTIONS(773),
    [anon_sym_AMP] = ACTIONS(773),
    [anon_sym_COLON_COLON] = ACTIONS(771),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_else] = ACTIONS(773),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(773),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(771),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(771),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(771),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(773),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(773),
    [anon_sym_lsl] = ACTIONS(773),
    [anon_sym_lsr] = ACTIONS(773),
    [anon_sym_asr] = ACTIONS(773),
    [aux_sym_pow_operator_token1] = ACTIONS(771),
    [anon_sym_mod] = ACTIONS(773),
    [anon_sym_land] = ACTIONS(773),
    [anon_sym_lor] = ACTIONS(773),
    [anon_sym_lxor] = ACTIONS(773),
    [aux_sym_mult_operator_token1] = ACTIONS(773),
    [aux_sym_add_operator_token1] = ACTIONS(773),
    [sym_concat_operator] = ACTIONS(771),
    [sym_rel_operator] = ACTIONS(773),
    [anon_sym_AMP_AMP] = ACTIONS(773),
    [anon_sym_or] = ACTIONS(773),
    [anon_sym_PIPE_PIPE] = ACTIONS(773),
    [sym_assign_operator] = ACTIONS(771),
    [sym_let_operator] = ACTIONS(771),
    [sym_let_and_operator] = ACTIONS(771),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(771),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(85)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2628),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1507),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(85),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(518),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_constraint] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_inherit] = ACTIONS(43),
    [anon_sym_method] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_initializer] = ACTIONS(43),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(381),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(383),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(86)] = {
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(1114),
    [sym_typed_expression] = STATE(1131),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_labeled_argument] = STATE(1164),
    [sym_prefix_expression] = STATE(1131),
    [sym_hash_expression] = STATE(1131),
    [sym_field_get_expression] = STATE(1131),
    [sym_array_get_expression] = STATE(1131),
    [sym_string_get_expression] = STATE(1131),
    [sym_bigarray_get_expression] = STATE(1131),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(86),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [aux_sym_class_application_repeat1] = STATE(89),
    [sym__lowercase_identifier] = ACTIONS(959),
    [anon_sym_SEMI_SEMI] = ACTIONS(741),
    [anon_sym_let] = ACTIONS(743),
    [anon_sym_TILDE] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(961),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(743),
    [anon_sym_type] = ACTIONS(743),
    [anon_sym_COMMA] = ACTIONS(741),
    [anon_sym_PIPE] = ACTIONS(743),
    [anon_sym_DOT] = ACTIONS(963),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(741),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_SEMI] = ACTIONS(743),
    [anon_sym_exception] = ACTIONS(743),
    [anon_sym_module] = ACTIONS(743),
    [anon_sym_open] = ACTIONS(743),
    [anon_sym_include] = ACTIONS(743),
    [anon_sym_class] = ACTIONS(743),
    [anon_sym_val] = ACTIONS(743),
    [anon_sym_end] = ACTIONS(743),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_AMP] = ACTIONS(743),
    [anon_sym_POUND] = ACTIONS(965),
    [anon_sym_COLON_COLON] = ACTIONS(741),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_else] = ACTIONS(743),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(743),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(741),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(741),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(743),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(743),
    [sym_hash_operator] = ACTIONS(967),
    [anon_sym_lsl] = ACTIONS(743),
    [anon_sym_lsr] = ACTIONS(743),
    [anon_sym_asr] = ACTIONS(743),
    [aux_sym_pow_operator_token1] = ACTIONS(741),
    [anon_sym_mod] = ACTIONS(743),
    [anon_sym_land] = ACTIONS(743),
    [anon_sym_lor] = ACTIONS(743),
    [anon_sym_lxor] = ACTIONS(743),
    [aux_sym_mult_operator_token1] = ACTIONS(743),
    [aux_sym_add_operator_token1] = ACTIONS(743),
    [sym_concat_operator] = ACTIONS(741),
    [sym_rel_operator] = ACTIONS(743),
    [anon_sym_AMP_AMP] = ACTIONS(743),
    [anon_sym_or] = ACTIONS(743),
    [anon_sym_PIPE_PIPE] = ACTIONS(743),
    [sym_assign_operator] = ACTIONS(741),
    [sym_let_operator] = ACTIONS(741),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(743),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(87)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2658),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1507),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(87),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(521),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(41),
    [anon_sym_COLON2] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_RBRACE] = ACTIONS(41),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_with] = ACTIONS(43),
    [anon_sym_COLON_GT] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_then] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(88)] = {
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(1114),
    [sym_typed_expression] = STATE(1131),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_labeled_argument] = STATE(1164),
    [sym_prefix_expression] = STATE(1131),
    [sym_hash_expression] = STATE(1131),
    [sym_field_get_expression] = STATE(1131),
    [sym_array_get_expression] = STATE(1131),
    [sym_string_get_expression] = STATE(1131),
    [sym_bigarray_get_expression] = STATE(1131),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(88),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [aux_sym_class_application_repeat1] = STATE(88),
    [sym__lowercase_identifier] = ACTIONS(969),
    [anon_sym_SEMI_SEMI] = ACTIONS(784),
    [anon_sym_let] = ACTIONS(786),
    [anon_sym_and] = ACTIONS(786),
    [anon_sym_TILDE] = ACTIONS(972),
    [anon_sym_QMARK] = ACTIONS(972),
    [anon_sym_LPAREN] = ACTIONS(975),
    [anon_sym_external] = ACTIONS(786),
    [anon_sym_type] = ACTIONS(786),
    [anon_sym_COMMA] = ACTIONS(784),
    [anon_sym_PIPE] = ACTIONS(786),
    [anon_sym_LBRACK] = ACTIONS(978),
    [anon_sym_RBRACK] = ACTIONS(784),
    [anon_sym_true] = ACTIONS(981),
    [anon_sym_false] = ACTIONS(981),
    [anon_sym_LBRACE] = ACTIONS(984),
    [anon_sym_SEMI] = ACTIONS(786),
    [anon_sym_exception] = ACTIONS(786),
    [anon_sym_module] = ACTIONS(786),
    [anon_sym_open] = ACTIONS(786),
    [anon_sym_include] = ACTIONS(786),
    [anon_sym_class] = ACTIONS(786),
    [anon_sym_val] = ACTIONS(786),
    [anon_sym_end] = ACTIONS(786),
    [anon_sym_object] = ACTIONS(987),
    [anon_sym_AMP] = ACTIONS(786),
    [anon_sym_COLON_COLON] = ACTIONS(784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(990),
    [anon_sym_else] = ACTIONS(786),
    [anon_sym_new] = ACTIONS(993),
    [anon_sym_LBRACE_LT] = ACTIONS(996),
    [anon_sym_begin] = ACTIONS(999),
    [sym_ocamlyacc_value] = ACTIONS(1002),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(786),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(784),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1005),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1008),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(784),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(784),
    [aux_sym_number_token1] = ACTIONS(1011),
    [aux_sym_signed_number_token1] = ACTIONS(786),
    [anon_sym_SQUOTE] = ACTIONS(1014),
    [anon_sym_DQUOTE] = ACTIONS(1017),
    [sym_prefix_operator] = ACTIONS(1020),
    [aux_sym_sign_operator_token1] = ACTIONS(786),
    [anon_sym_lsl] = ACTIONS(786),
    [anon_sym_lsr] = ACTIONS(786),
    [anon_sym_asr] = ACTIONS(786),
    [aux_sym_pow_operator_token1] = ACTIONS(784),
    [anon_sym_mod] = ACTIONS(786),
    [anon_sym_land] = ACTIONS(786),
    [anon_sym_lor] = ACTIONS(786),
    [anon_sym_lxor] = ACTIONS(786),
    [aux_sym_mult_operator_token1] = ACTIONS(786),
    [aux_sym_add_operator_token1] = ACTIONS(786),
    [sym_concat_operator] = ACTIONS(784),
    [sym_rel_operator] = ACTIONS(786),
    [anon_sym_AMP_AMP] = ACTIONS(786),
    [anon_sym_or] = ACTIONS(786),
    [anon_sym_PIPE_PIPE] = ACTIONS(786),
    [sym_assign_operator] = ACTIONS(784),
    [sym_let_operator] = ACTIONS(784),
    [sym__uppercase_identifier] = ACTIONS(1023),
    [aux_sym_directive_token1] = ACTIONS(784),
    [aux_sym_tag_token1] = ACTIONS(1026),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(89)] = {
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(1114),
    [sym_typed_expression] = STATE(1131),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_labeled_argument] = STATE(1164),
    [sym_prefix_expression] = STATE(1131),
    [sym_hash_expression] = STATE(1131),
    [sym_field_get_expression] = STATE(1131),
    [sym_array_get_expression] = STATE(1131),
    [sym_string_get_expression] = STATE(1131),
    [sym_bigarray_get_expression] = STATE(1131),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(89),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [aux_sym_class_application_repeat1] = STATE(88),
    [sym__lowercase_identifier] = ACTIONS(959),
    [anon_sym_SEMI_SEMI] = ACTIONS(771),
    [anon_sym_let] = ACTIONS(773),
    [anon_sym_TILDE] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(961),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(773),
    [anon_sym_type] = ACTIONS(773),
    [anon_sym_COMMA] = ACTIONS(771),
    [anon_sym_PIPE] = ACTIONS(773),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(771),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_SEMI] = ACTIONS(773),
    [anon_sym_exception] = ACTIONS(773),
    [anon_sym_module] = ACTIONS(773),
    [anon_sym_open] = ACTIONS(773),
    [anon_sym_include] = ACTIONS(773),
    [anon_sym_class] = ACTIONS(773),
    [anon_sym_val] = ACTIONS(773),
    [anon_sym_end] = ACTIONS(773),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_AMP] = ACTIONS(773),
    [anon_sym_COLON_COLON] = ACTIONS(771),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_else] = ACTIONS(773),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(773),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(771),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(771),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(771),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(773),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(773),
    [anon_sym_lsl] = ACTIONS(773),
    [anon_sym_lsr] = ACTIONS(773),
    [anon_sym_asr] = ACTIONS(773),
    [aux_sym_pow_operator_token1] = ACTIONS(771),
    [anon_sym_mod] = ACTIONS(773),
    [anon_sym_land] = ACTIONS(773),
    [anon_sym_lor] = ACTIONS(773),
    [anon_sym_lxor] = ACTIONS(773),
    [aux_sym_mult_operator_token1] = ACTIONS(773),
    [aux_sym_add_operator_token1] = ACTIONS(773),
    [sym_concat_operator] = ACTIONS(771),
    [sym_rel_operator] = ACTIONS(773),
    [anon_sym_AMP_AMP] = ACTIONS(773),
    [anon_sym_or] = ACTIONS(773),
    [anon_sym_PIPE_PIPE] = ACTIONS(773),
    [sym_assign_operator] = ACTIONS(771),
    [sym_let_operator] = ACTIONS(771),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(771),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(90)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2909),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2511),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(90),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__attribute] = STATE(514),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_in] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(91)] = {
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(1251),
    [sym_typed_expression] = STATE(803),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_labeled_argument] = STATE(861),
    [sym_prefix_expression] = STATE(803),
    [sym_hash_expression] = STATE(803),
    [sym_field_get_expression] = STATE(803),
    [sym_array_get_expression] = STATE(803),
    [sym_string_get_expression] = STATE(803),
    [sym_bigarray_get_expression] = STATE(803),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(91),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [aux_sym_class_application_repeat1] = STATE(91),
    [sym__lowercase_identifier] = ACTIONS(775),
    [anon_sym_and] = ACTIONS(786),
    [anon_sym_TILDE] = ACTIONS(1029),
    [anon_sym_QMARK] = ACTIONS(1029),
    [anon_sym_LPAREN] = ACTIONS(781),
    [anon_sym_RPAREN] = ACTIONS(784),
    [anon_sym_COMMA] = ACTIONS(784),
    [anon_sym_COLON2] = ACTIONS(786),
    [anon_sym_LBRACK] = ACTIONS(788),
    [anon_sym_true] = ACTIONS(791),
    [anon_sym_false] = ACTIONS(791),
    [anon_sym_LBRACE] = ACTIONS(794),
    [anon_sym_constraint] = ACTIONS(786),
    [anon_sym_val] = ACTIONS(786),
    [anon_sym_end] = ACTIONS(786),
    [anon_sym_COLON_GT] = ACTIONS(784),
    [anon_sym_object] = ACTIONS(797),
    [anon_sym_inherit] = ACTIONS(786),
    [anon_sym_method] = ACTIONS(786),
    [anon_sym_in] = ACTIONS(786),
    [anon_sym_as] = ACTIONS(786),
    [anon_sym_initializer] = ACTIONS(786),
    [anon_sym_AMP] = ACTIONS(786),
    [anon_sym_COLON_COLON] = ACTIONS(784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(800),
    [anon_sym_else] = ACTIONS(786),
    [anon_sym_new] = ACTIONS(1032),
    [anon_sym_LBRACE_LT] = ACTIONS(806),
    [anon_sym_begin] = ACTIONS(809),
    [sym_ocamlyacc_value] = ACTIONS(812),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(786),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(784),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(815),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(818),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(784),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(784),
    [aux_sym_number_token1] = ACTIONS(821),
    [aux_sym_signed_number_token1] = ACTIONS(786),
    [anon_sym_SQUOTE] = ACTIONS(824),
    [anon_sym_DQUOTE] = ACTIONS(827),
    [sym_prefix_operator] = ACTIONS(1035),
    [aux_sym_sign_operator_token1] = ACTIONS(786),
    [anon_sym_lsl] = ACTIONS(786),
    [anon_sym_lsr] = ACTIONS(786),
    [anon_sym_asr] = ACTIONS(786),
    [aux_sym_pow_operator_token1] = ACTIONS(784),
    [anon_sym_mod] = ACTIONS(786),
    [anon_sym_land] = ACTIONS(786),
    [anon_sym_lor] = ACTIONS(786),
    [anon_sym_lxor] = ACTIONS(786),
    [aux_sym_mult_operator_token1] = ACTIONS(786),
    [aux_sym_add_operator_token1] = ACTIONS(786),
    [sym_concat_operator] = ACTIONS(784),
    [sym_rel_operator] = ACTIONS(786),
    [anon_sym_AMP_AMP] = ACTIONS(786),
    [anon_sym_or] = ACTIONS(786),
    [anon_sym_PIPE_PIPE] = ACTIONS(786),
    [sym_assign_operator] = ACTIONS(784),
    [sym__uppercase_identifier] = ACTIONS(833),
    [aux_sym_tag_token1] = ACTIONS(836),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(92)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3009),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1925),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(92),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(505),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_done] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_to] = ACTIONS(43),
    [anon_sym_downto] = ACTIONS(43),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(93)] = {
    [sym__abstract_type] = STATE(7254),
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4801),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(93),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(7225),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1044),
    [anon_sym_type] = ACTIONS(1046),
    [anon_sym_DOT] = ACTIONS(1048),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1058),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1062),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1084),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1084),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1084),
    [sym_let_and_operator] = ACTIONS(1084),
    [sym_match_operator] = ACTIONS(1084),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(94)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6341),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(94),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(185),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1090),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(95)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6356),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(95),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(209),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1092),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(96)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6567),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(96),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(193),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1094),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(97)] = {
    [sym__abstract_type] = STATE(7254),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4732),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(97),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(7105),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(1102),
    [anon_sym_type] = ACTIONS(1046),
    [anon_sym_DOT] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1114),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1122),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1122),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1122),
    [sym_let_and_operator] = ACTIONS(1122),
    [sym_match_operator] = ACTIONS(1122),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(98)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6692),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(98),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(203),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1124),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(99)] = {
    [sym__abstract_type] = STATE(7254),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4756),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(99),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_type] = ACTIONS(1046),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1128),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(100)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7307),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(100),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(179),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1130),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(101)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6341),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(101),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(202),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1132),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(102)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6841),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(102),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(204),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1134),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(103)] = {
    [sym__abstract_type] = STATE(7254),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4645),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(103),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(6667),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(1136),
    [anon_sym_type] = ACTIONS(1046),
    [anon_sym_DOT] = ACTIONS(1138),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1140),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1128),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1142),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1142),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1142),
    [sym_let_and_operator] = ACTIONS(1142),
    [sym_match_operator] = ACTIONS(1142),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(104)] = {
    [sym__abstract_type] = STATE(7254),
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4721),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(104),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(7105),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1102),
    [anon_sym_type] = ACTIONS(1046),
    [anon_sym_DOT] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1114),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1122),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1122),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1122),
    [sym_let_and_operator] = ACTIONS(1122),
    [sym_match_operator] = ACTIONS(1122),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(105)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6376),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(105),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(173),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1144),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(106)] = {
    [sym__abstract_type] = STATE(7254),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4756),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(106),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_type] = ACTIONS(1046),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1146),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(107)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6462),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(107),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(181),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1148),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(108)] = {
    [sym__abstract_type] = STATE(6705),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4642),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(108),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(6770),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(1150),
    [anon_sym_type] = ACTIONS(1046),
    [anon_sym_DOT] = ACTIONS(1152),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1154),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1156),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1158),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1158),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1158),
    [sym_let_and_operator] = ACTIONS(1158),
    [sym_match_operator] = ACTIONS(1158),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(109)] = {
    [sym__abstract_type] = STATE(7254),
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4685),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(109),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(6991),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1160),
    [anon_sym_type] = ACTIONS(1046),
    [anon_sym_DOT] = ACTIONS(1162),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1164),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1166),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1168),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1168),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1168),
    [sym_let_and_operator] = ACTIONS(1168),
    [sym_match_operator] = ACTIONS(1168),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(110)] = {
    [sym__abstract_type] = STATE(7254),
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4708),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(110),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_type] = ACTIONS(1046),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1128),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(111)] = {
    [sym__abstract_type] = STATE(7254),
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4708),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(111),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_type] = ACTIONS(1046),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1146),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(112)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7117),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(112),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(194),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1170),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(113)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7245),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(113),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(187),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1172),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(114)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6555),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(114),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(474),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(115)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1695),
    [sym__sequence_expression] = STATE(2186),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_refutation_case] = STATE(2186),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(115),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_DOT] = ACTIONS(1174),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(116)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3039),
    [sym__sequence_expression] = STATE(2308),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_refutation_case] = STATE(2308),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(116),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1180),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(117)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3039),
    [sym__sequence_expression] = STATE(2288),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_refutation_case] = STATE(2288),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(117),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1180),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(118)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1748),
    [sym__sequence_expression] = STATE(2245),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_refutation_case] = STATE(2245),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(118),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_DOT] = ACTIONS(1174),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(119)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1748),
    [sym__sequence_expression] = STATE(2186),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_refutation_case] = STATE(2186),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(119),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_DOT] = ACTIONS(1174),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(120)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2857),
    [sym__sequence_expression] = STATE(2921),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_refutation_case] = STATE(2921),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(120),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1182),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(121)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2045),
    [sym__sequence_expression] = STATE(2245),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_refutation_case] = STATE(2245),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(121),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_DOT] = ACTIONS(1174),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(122)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2822),
    [sym__sequence_expression] = STATE(2921),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_refutation_case] = STATE(2921),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(122),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_DOT] = ACTIONS(1182),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(123)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1808),
    [sym__sequence_expression] = STATE(2308),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_refutation_case] = STATE(2308),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(123),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1180),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(124)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1808),
    [sym__sequence_expression] = STATE(2288),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_refutation_case] = STATE(2288),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(124),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1180),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(125)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6197),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(125),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(335),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(126)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3223),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(6390),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(126),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(511),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_DASH_GT] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(127)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2857),
    [sym__sequence_expression] = STATE(2893),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_refutation_case] = STATE(2893),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(127),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1182),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(128)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6383),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(128),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(239),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(129)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6526),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(129),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(253),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(130)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3091),
    [sym__sequence_expression] = STATE(3459),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_refutation_case] = STATE(3459),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(130),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_DOT] = ACTIONS(1188),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(131)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3091),
    [sym__sequence_expression] = STATE(3443),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_refutation_case] = STATE(3443),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(131),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_DOT] = ACTIONS(1188),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(132)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2045),
    [sym__sequence_expression] = STATE(2186),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_refutation_case] = STATE(2186),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(132),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_DOT] = ACTIONS(1174),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(133)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6243),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(133),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(361),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(134)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6257),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(134),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(337),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(135)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2822),
    [sym__sequence_expression] = STATE(2893),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_refutation_case] = STATE(2893),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(135),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_DOT] = ACTIONS(1182),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(136)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(5972),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(136),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(364),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(137)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6184),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(137),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(373),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(138)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6711),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(138),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(493),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(139)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6108),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(139),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(385),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(140)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6252),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(140),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(394),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(141)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6178),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(141),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(238),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(142)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6103),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(142),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(266),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(143)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6064),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(143),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(280),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(144)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6211),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(144),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(289),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(145)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6029),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(145),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(369),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(146)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6002),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(146),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(367),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(147)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6167),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(147),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(296),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(148)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6027),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(148),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(368),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(149)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(149),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(443),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(150)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6442),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(150),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(444),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(151)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3217),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1507),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(151),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(524),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(152)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1695),
    [sym__sequence_expression] = STATE(2245),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_refutation_case] = STATE(2245),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(152),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_DOT] = ACTIONS(1174),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(153)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6159),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(153),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(453),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(154)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7284),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(154),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(458),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(155)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7287),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(155),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(459),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(156)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6418),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(156),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(463),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(157)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(157),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(464),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(158)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6508),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(158),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(468),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(159)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6509),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(159),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(469),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(160)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6554),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(160),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(473),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(161)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(5974),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(161),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(304),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(162)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6597),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(162),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(477),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(163)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6598),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(163),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(478),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(164)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6637),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(164),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(482),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(165)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6638),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(165),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(483),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(166)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6675),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(166),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(487),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(167)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6676),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(167),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(488),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(168)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6710),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(168),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(492),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(169)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6164),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(169),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__attribute] = STATE(382),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(735),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(170)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4708),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(170),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1128),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(171)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4708),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(171),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1146),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(172)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4732),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(172),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(7105),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(1102),
    [anon_sym_DOT] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1190),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1122),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1122),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1122),
    [sym_let_and_operator] = ACTIONS(1122),
    [sym_match_operator] = ACTIONS(1122),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(173)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6891),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(173),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1192),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(174)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4721),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(174),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(7105),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1102),
    [anon_sym_DOT] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1190),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1122),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1122),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1122),
    [sym_let_and_operator] = ACTIONS(1122),
    [sym_match_operator] = ACTIONS(1122),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(175)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4543),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(175),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(6667),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(1136),
    [anon_sym_DOT] = ACTIONS(1138),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1140),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1128),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1142),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1142),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1142),
    [sym_let_and_operator] = ACTIONS(1142),
    [sym_match_operator] = ACTIONS(1142),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(176)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(176),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1194),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(177)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4756),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(177),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1196),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(178)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4708),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(178),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1196),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(179)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7113),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(179),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1198),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(180)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(180),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1200),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(181)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6642),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(181),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1202),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(182)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4721),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(182),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(7105),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1102),
    [anon_sym_DOT] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1114),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1122),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1122),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1122),
    [sym_let_and_operator] = ACTIONS(1122),
    [sym_match_operator] = ACTIONS(1122),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(183)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4756),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(183),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(419),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(184)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(184),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1204),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(185)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6616),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(185),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1206),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(186)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(186),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1208),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(187)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6364),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(187),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1210),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(188)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6545),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(188),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_done] = ACTIONS(1212),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(189)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4756),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(189),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1128),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(190)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(190),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1214),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(191)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(191),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1216),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(192)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(192),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1218),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(193)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6640),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(193),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1220),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(194)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7253),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(194),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1222),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(195)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4505),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(195),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(6770),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(1150),
    [anon_sym_DOT] = ACTIONS(1152),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1154),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1156),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1158),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1158),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1158),
    [sym_let_and_operator] = ACTIONS(1158),
    [sym_match_operator] = ACTIONS(1158),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(196)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4801),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(196),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(7225),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1044),
    [anon_sym_DOT] = ACTIONS(1048),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1058),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1062),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1084),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1084),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1084),
    [sym_let_and_operator] = ACTIONS(1084),
    [sym_match_operator] = ACTIONS(1084),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(197)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(197),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1224),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(198)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6833),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(198),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_done] = ACTIONS(1226),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(199)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(199),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(200)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4685),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(200),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(6991),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1160),
    [anon_sym_DOT] = ACTIONS(1162),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1164),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1166),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1168),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1168),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1168),
    [sym_let_and_operator] = ACTIONS(1168),
    [sym_match_operator] = ACTIONS(1168),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(201)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(201),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1230),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(202)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6616),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(202),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1232),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(203)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6425),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(203),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1234),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(204)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6906),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(204),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1236),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(205)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4756),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(205),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1146),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(206)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4708),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(206),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__infix_operator] = STATE(6460),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1040),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_DOT] = ACTIONS(413),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_module] = ACTIONS(1126),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(419),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(425),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(425),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(425),
    [sym_let_and_operator] = ACTIONS(425),
    [sym_match_operator] = ACTIONS(425),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(207)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4732),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(207),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__infix_operator] = STATE(7105),
    [sym_pow_operator] = STATE(6700),
    [sym_mult_operator] = STATE(6700),
    [sym_add_operator] = STATE(6700),
    [sym_and_operator] = STATE(6700),
    [sym_or_operator] = STATE(6700),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(1098),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(1102),
    [anon_sym_DOT] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_module] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(417),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1114),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym_prefix_operator] = ACTIONS(1122),
    [aux_sym_sign_operator_token1] = ACTIONS(433),
    [sym_hash_operator] = ACTIONS(1122),
    [anon_sym_lsl] = ACTIONS(427),
    [anon_sym_lsr] = ACTIONS(427),
    [anon_sym_asr] = ACTIONS(427),
    [aux_sym_pow_operator_token1] = ACTIONS(429),
    [anon_sym_mod] = ACTIONS(431),
    [anon_sym_land] = ACTIONS(431),
    [anon_sym_lor] = ACTIONS(431),
    [anon_sym_lxor] = ACTIONS(431),
    [aux_sym_mult_operator_token1] = ACTIONS(431),
    [aux_sym_add_operator_token1] = ACTIONS(433),
    [sym_concat_operator] = ACTIONS(435),
    [sym_rel_operator] = ACTIONS(437),
    [anon_sym_AMP_AMP] = ACTIONS(417),
    [anon_sym_or] = ACTIONS(439),
    [anon_sym_PIPE_PIPE] = ACTIONS(439),
    [sym_assign_operator] = ACTIONS(435),
    [sym_let_operator] = ACTIONS(1122),
    [sym_let_and_operator] = ACTIONS(1122),
    [sym_match_operator] = ACTIONS(1122),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(208)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6553),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(208),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_done] = ACTIONS(1238),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(209)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6481),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(209),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_end] = ACTIONS(1240),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(210)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(210),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1242),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(211)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6802),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(211),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_done] = ACTIONS(1244),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(212)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(7102),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(212),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1246),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(213)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4374),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(213),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(214)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2822),
    [sym__sequence_expression] = STATE(2899),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(214),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(215)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2822),
    [sym__sequence_expression] = STATE(2904),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(215),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(216)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2822),
    [sym__sequence_expression] = STATE(2905),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(216),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(217)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4375),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(217),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(218)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2822),
    [sym__sequence_expression] = STATE(2932),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(218),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(219)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3039),
    [sym__sequence_expression] = STATE(2311),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(219),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(220)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4377),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(220),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(221)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3039),
    [sym__sequence_expression] = STATE(2270),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(221),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(222)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3039),
    [sym__sequence_expression] = STATE(2296),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(222),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(223)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3039),
    [sym__sequence_expression] = STATE(2323),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(223),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(224)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3039),
    [sym__sequence_expression] = STATE(2271),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(224),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(225)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3039),
    [sym__sequence_expression] = STATE(2272),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(225),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(226)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3039),
    [sym__sequence_expression] = STATE(2297),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(226),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(227)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4384),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(227),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(228)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6300),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(228),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(229)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6657),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(229),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(230)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4391),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(230),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(231)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(7027),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(231),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1248),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(232)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3162),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(7189),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(232),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1250),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(233)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4394),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(233),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(234)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4396),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(234),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(235)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4397),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(235),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(236)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6851),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(236),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(237)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4408),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(237),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(238)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6061),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(238),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(239)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6979),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(239),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(240)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2541),
    [sym__sequence_expression] = STATE(2140),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(240),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(241)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2045),
    [sym__sequence_expression] = STATE(2188),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(241),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(242)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2924),
    [sym__sequence_expression] = STATE(2962),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(242),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(243)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2541),
    [sym__sequence_expression] = STATE(2261),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(243),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(244)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2924),
    [sym__sequence_expression] = STATE(2945),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(244),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(245)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2924),
    [sym__sequence_expression] = STATE(2955),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(245),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(246)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2901),
    [sym__sequence_expression] = STATE(5040),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(246),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(247)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2541),
    [sym__sequence_expression] = STATE(2210),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(247),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(248)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2924),
    [sym__sequence_expression] = STATE(2899),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(248),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(249)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2924),
    [sym__sequence_expression] = STATE(2904),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(249),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(250)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2924),
    [sym__sequence_expression] = STATE(2905),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(250),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(251)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2924),
    [sym__sequence_expression] = STATE(2932),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(251),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(252)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2857),
    [sym__sequence_expression] = STATE(2962),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(252),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(253)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6988),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(253),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(254)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2857),
    [sym__sequence_expression] = STATE(2945),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(254),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(255)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2857),
    [sym__sequence_expression] = STATE(2955),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(255),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(256)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2857),
    [sym__sequence_expression] = STATE(2899),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(256),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(257)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2857),
    [sym__sequence_expression] = STATE(2904),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(257),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(258)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2857),
    [sym__sequence_expression] = STATE(2905),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(258),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(259)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2857),
    [sym__sequence_expression] = STATE(2932),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(259),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(260)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6122),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(260),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(261)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(7317),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(261),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1252),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(262)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3162),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(6335),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(262),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(263)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(7169),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(263),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1256),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(264)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2901),
    [sym__sequence_expression] = STATE(4866),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(264),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(265)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2901),
    [sym__sequence_expression] = STATE(4869),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(265),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(266)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6222),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(266),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(267)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2901),
    [sym__sequence_expression] = STATE(4878),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(267),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(268)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2541),
    [sym__sequence_expression] = STATE(2151),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(268),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(269)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2541),
    [sym__sequence_expression] = STATE(2155),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(269),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(270)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2541),
    [sym__sequence_expression] = STATE(2166),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(270),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(271)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6383),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(271),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(272)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2901),
    [sym__sequence_expression] = STATE(5015),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(272),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(273)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2901),
    [sym__sequence_expression] = STATE(5019),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(273),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(274)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(6772),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(274),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1258),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(275)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3162),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(6831),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(275),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1260),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(276)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2901),
    [sym__sequence_expression] = STATE(5020),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(276),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(277)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2541),
    [sym__sequence_expression] = STATE(2181),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(277),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(278)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2901),
    [sym__sequence_expression] = STATE(4877),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(278),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(279)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7049),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(279),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(280)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(5965),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(280),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(281)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1695),
    [sym__sequence_expression] = STATE(2234),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(281),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(282)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7069),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(282),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(283)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7080),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(283),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(284)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(7081),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(284),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1262),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(285)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3162),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(7097),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(285),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1264),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(286)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1695),
    [sym__sequence_expression] = STATE(2146),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(286),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(287)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1695),
    [sym__sequence_expression] = STATE(2188),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(287),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(288)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2045),
    [sym__sequence_expression] = STATE(2234),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(288),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(289)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6138),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(289),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(290)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1695),
    [sym__sequence_expression] = STATE(2168),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(290),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(291)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1695),
    [sym__sequence_expression] = STATE(2178),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(291),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(292)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(6359),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(292),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1266),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(293)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3162),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(6506),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(293),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1268),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(294)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1695),
    [sym__sequence_expression] = STATE(2179),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(294),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(295)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1695),
    [sym__sequence_expression] = STATE(2200),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(295),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(296)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(5981),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(296),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(297)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2878),
    [sym__sequence_expression] = STATE(2962),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(297),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(298)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3162),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(6931),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(298),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(299)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2878),
    [sym__sequence_expression] = STATE(2945),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(299),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(300)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(6357),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(300),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1272),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(301)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3162),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(6423),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(301),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1274),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(302)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2878),
    [sym__sequence_expression] = STATE(2955),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(302),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(303)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4403),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(303),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(304)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6142),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(304),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(305)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2878),
    [sym__sequence_expression] = STATE(2899),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(305),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(306)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2878),
    [sym__sequence_expression] = STATE(2904),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(306),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(307)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2878),
    [sym__sequence_expression] = STATE(2905),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(307),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(308)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3162),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(7182),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(308),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1276),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(309)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1850),
    [sym__sequence_expression] = STATE(3538),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(309),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(310)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2878),
    [sym__sequence_expression] = STATE(2932),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(310),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(311)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1850),
    [sym__sequence_expression] = STATE(3552),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(311),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(312)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1808),
    [sym__sequence_expression] = STATE(2311),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(312),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(313)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1850),
    [sym__sequence_expression] = STATE(3525),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(313),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(314)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2045),
    [sym__sequence_expression] = STATE(2168),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(314),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(315)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(6527),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(315),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1278),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(316)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3162),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(6549),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(316),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1280),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(317)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1808),
    [sym__sequence_expression] = STATE(2270),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(317),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(318)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1808),
    [sym__sequence_expression] = STATE(2296),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(318),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(319)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1808),
    [sym__sequence_expression] = STATE(2323),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(319),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(320)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1808),
    [sym__sequence_expression] = STATE(2271),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(320),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(321)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1808),
    [sym__sequence_expression] = STATE(2272),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(321),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(322)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1808),
    [sym__sequence_expression] = STATE(2297),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(322),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(323)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4422),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(323),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(324)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4316),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(324),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(325)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2045),
    [sym__sequence_expression] = STATE(2178),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(325),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(326)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2045),
    [sym__sequence_expression] = STATE(2179),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(326),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(327)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6047),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(327),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(328)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2045),
    [sym__sequence_expression] = STATE(2146),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(328),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(329)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1850),
    [sym__sequence_expression] = STATE(3536),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(329),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(330)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6875),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(330),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(331)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6877),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(331),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(332)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6878),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(332),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(333)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(7065),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(333),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1282),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(334)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3162),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym__sequence_expression_content] = STATE(7305),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(334),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1284),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(335)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6145),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(335),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(336)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7001),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(336),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(337)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(5964),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(337),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(338)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6311),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(338),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(339)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6590),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(339),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(340)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6596),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(340),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(341)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6603),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(341),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(342)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7018),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(342),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(343)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7019),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(343),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(344)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7021),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(344),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(345)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7094),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(345),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(346)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6316),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(346),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(347)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6501),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(347),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(348)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6285),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(348),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(349)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6842),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(349),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(350)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6871),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(350),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(351)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3091),
    [sym__sequence_expression] = STATE(2140),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(351),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(352)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1850),
    [sym__sequence_expression] = STATE(3548),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(352),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(353)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3091),
    [sym__sequence_expression] = STATE(2261),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(353),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(354)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3091),
    [sym__sequence_expression] = STATE(2210),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(354),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(355)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1850),
    [sym__sequence_expression] = STATE(3527),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(355),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(356)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3091),
    [sym__sequence_expression] = STATE(2151),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(356),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(357)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3091),
    [sym__sequence_expression] = STATE(2155),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(357),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(358)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3091),
    [sym__sequence_expression] = STATE(2166),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(358),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(359)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3091),
    [sym__sequence_expression] = STATE(2181),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(359),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(360)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1850),
    [sym__sequence_expression] = STATE(3539),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(360),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(361)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(5946),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(361),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(362)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4338),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(362),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(363)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2822),
    [sym__sequence_expression] = STATE(2955),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(363),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(364)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6092),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(364),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(365)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(2085),
    [sym__sequence_expression] = STATE(3536),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(365),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(366)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4347),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(366),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(367)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6188),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(367),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(368)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6253),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(368),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(369)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6321),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(369),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(370)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3210),
    [sym__sequence_expression] = STATE(6311),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(7172),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(370),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(371)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4354),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(371),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(372)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4355),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(372),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(373)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(5968),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(373),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(374)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6812),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(374),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(375)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6813),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(375),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(376)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6814),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(376),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(377)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6896),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(377),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(378)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6905),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(378),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(379)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6907),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(379),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(380)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6175),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(380),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(381)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6227),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(381),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(382)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6134),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(382),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(383)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(2085),
    [sym__sequence_expression] = STATE(3538),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(383),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(384)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(2085),
    [sym__sequence_expression] = STATE(3552),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(384),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(385)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6223),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(385),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(386)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6565),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(386),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(387)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6566),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(387),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(388)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6574),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(388),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(389)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6618),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(389),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(390)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6623),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(390),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(391)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6624),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(391),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(392)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(5957),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(392),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(393)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(5966),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(393),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(394)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6230),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(394),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(395)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(2085),
    [sym__sequence_expression] = STATE(3525),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(395),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(396)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6944),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(396),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(397)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6945),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(397),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(398)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6946),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(398),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(399)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6975),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(399),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(400)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6976),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(400),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(401)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6977),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(401),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(402)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2045),
    [sym__sequence_expression] = STATE(2200),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(402),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(403)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7294),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(403),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(404)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7295),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(404),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(405)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7296),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(405),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(406)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7311),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(406),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(407)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7312),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(407),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(408)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7314),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(408),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(409)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(2085),
    [sym__sequence_expression] = STATE(3548),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(409),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(410)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6615),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(410),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(411)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6625),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(411),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(412)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6626),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(412),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(413)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6764),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(413),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(414)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6776),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(414),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(415)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6777),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(415),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(416)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(2085),
    [sym__sequence_expression] = STATE(3527),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(416),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(417)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6748),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(417),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(418)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6749),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(418),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(419)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6758),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(419),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(420)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6803),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(420),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(421)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6821),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(421),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(422)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6822),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(422),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(423)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(2085),
    [sym__sequence_expression] = STATE(3539),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(423),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(424)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6412),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(424),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(425)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6416),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(425),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(426)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6421),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(426),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(427)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6464),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(427),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(428)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6467),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(428),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(429)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6471),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(429),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(430)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(2085),
    [sym__sequence_expression] = STATE(3526),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(430),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(431)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6680),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(431),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(432)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6681),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(432),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(433)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6686),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(433),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(434)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6708),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(434),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(435)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6709),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(435),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(436)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6716),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(436),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(437)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2545),
    [sym__sequence_expression] = STATE(2140),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(437),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(438)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3245),
    [sym__sequence_expression] = STATE(6082),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(438),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(439)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6208),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(439),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(440)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2545),
    [sym__sequence_expression] = STATE(2261),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(440),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(441)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2545),
    [sym__sequence_expression] = STATE(2210),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(441),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(442)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6438),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(442),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(443)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6572),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(443),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(444)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6576),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(444),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(445)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3184),
    [sym__sequence_expression] = STATE(2140),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(445),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(446)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6405),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(446),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(447)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3184),
    [sym__sequence_expression] = STATE(2261),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(447),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(448)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3184),
    [sym__sequence_expression] = STATE(2210),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(448),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(449)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3184),
    [sym__sequence_expression] = STATE(2151),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(449),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(450)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3184),
    [sym__sequence_expression] = STATE(2155),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(450),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(451)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3184),
    [sym__sequence_expression] = STATE(2166),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(451),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(452)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3184),
    [sym__sequence_expression] = STATE(2181),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(452),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(453)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6135),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(453),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(454)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1850),
    [sym__sequence_expression] = STATE(3526),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(454),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(455)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2545),
    [sym__sequence_expression] = STATE(2151),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(455),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(456)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2545),
    [sym__sequence_expression] = STATE(2155),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(456),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(457)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7284),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(457),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(458)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7322),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(458),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(459)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(7324),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(459),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(460)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2545),
    [sym__sequence_expression] = STATE(2166),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(460),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(461)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4368),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(461),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(462)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6418),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(462),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(463)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6432),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(463),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(464)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6433),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(464),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(465)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2545),
    [sym__sequence_expression] = STATE(2181),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2328),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(465),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(466)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1748),
    [sym__sequence_expression] = STATE(2234),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(466),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(467)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6508),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(467),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(468)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6519),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(468),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(469)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6520),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(469),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(470)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1748),
    [sym__sequence_expression] = STATE(2146),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(470),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(471)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1748),
    [sym__sequence_expression] = STATE(2188),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(471),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(472)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6554),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(472),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(473)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6563),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(473),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(474)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6564),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(474),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(475)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1748),
    [sym__sequence_expression] = STATE(2168),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(475),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(476)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6597),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(476),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(477)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6605),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(477),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(478)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6606),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(478),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(479)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1748),
    [sym__sequence_expression] = STATE(2178),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(479),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(480)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1748),
    [sym__sequence_expression] = STATE(2179),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(480),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(481)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6637),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(481),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(482)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6644),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(482),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(483)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6645),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(483),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(484)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1748),
    [sym__sequence_expression] = STATE(2200),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(484),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(485)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2822),
    [sym__sequence_expression] = STATE(2962),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(485),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(486)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6675),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(486),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(487)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6682),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(487),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(488)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6683),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(488),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(489)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4369),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(489),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(490)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2822),
    [sym__sequence_expression] = STATE(2945),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2902),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(490),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(491)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6710),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(491),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(492)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6717),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(492),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(493)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2669),
    [sym__sequence_expression] = STATE(6718),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(493),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(494)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6115),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(494),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(495)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6116),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(495),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(496)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6190),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(496),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(497)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6192),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(497),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(498)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3097),
    [sym__sequence_expression] = STATE(6207),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1859),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(498),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(499)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4344),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1496),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(499),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(500)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2070),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1503),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(500),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(501)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(2067),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1503),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(501),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(502)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2861),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2930),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(502),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(503)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2895),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2930),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(503),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(504)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2539),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2494),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(504),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(505)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3048),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(1916),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(505),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(506)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3246),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(506),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1286),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(507)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1902),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1916),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(507),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(508)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3183),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2494),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(508),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(509)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2567),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2292),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(509),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(510)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1720),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1503),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(510),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(511)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3211),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(6468),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(511),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(512)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3044),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(2494),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(512),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(513)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1821),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2292),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(513),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(514)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2928),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2494),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(514),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(515)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2873),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2494),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(515),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(516)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1675),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1916),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(516),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(517)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3246),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(517),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_RBRACK] = ACTIONS(1288),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(518)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2631),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1503),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(518),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(519)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3213),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(519),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1286),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(520)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2254),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym__sequence_expression_anonymous] = STATE(1503),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(520),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(521)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2665),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1503),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(521),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(522)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2790),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym__sequence_expression_anonymous] = STATE(2930),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(522),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(523)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3213),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(523),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1288),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(524)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3227),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym__sequence_expression_anonymous] = STATE(1503),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(524),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(525)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3050),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym__sequence_expression_anonymous] = STATE(2292),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(525),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(526)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2942),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(526),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(527)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2109),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(527),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(528)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2110),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(528),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(529)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2111),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(529),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(530)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2113),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(530),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(531)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3246),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(531),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(532)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2114),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(532),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(533)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2115),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(533),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(534)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2116),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(534),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(535)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2117),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(535),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(536)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2118),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(536),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(537)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2119),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(537),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(538)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2934),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(538),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(539)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2908),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(3322),
    [sym__tuple_expression] = STATE(2963),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(539),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(1290),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(540)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2911),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(540),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(541)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2912),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(541),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(542)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2913),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(542),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(543)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2914),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(543),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(544)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2939),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(544),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(545)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2941),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(545),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(546)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2915),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(546),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(547)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2916),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(547),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(548)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2889),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(548),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(549)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2918),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(549),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(550)] = {
    [sym_value_definition] = STATE(7258),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2971),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7222),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1644),
    [sym_array_get_expression] = STATE(1644),
    [sym_string_get_expression] = STATE(1644),
    [sym_bigarray_get_expression] = STATE(1644),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(550),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(538),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7224),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(551)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3221),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(551),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(552)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2860),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(552),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(553)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2877),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(3272),
    [sym__tuple_expression] = STATE(2963),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(553),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(1290),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(554)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2847),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(554),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(555)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2848),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(555),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(556)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2849),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(556),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(557)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2850),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(557),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(558)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2883),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(558),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(559)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2884),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(559),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(560)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2885),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(560),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(561)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2851),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(561),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(562)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2852),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(562),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(563)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2853),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(563),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(564)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2537),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(564),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(565)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2855),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(565),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(566)] = {
    [sym_value_definition] = STATE(6336),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2862),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(6517),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1571),
    [sym_array_get_expression] = STATE(1571),
    [sym_string_get_expression] = STATE(1571),
    [sym_bigarray_get_expression] = STATE(1571),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(566),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(552),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(6561),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(567)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2055),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(567),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(568)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2566),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(2840),
    [sym__tuple_expression] = STATE(2259),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(568),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(1292),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(569)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2569),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(569),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(570)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2570),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(570),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(571)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2571),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(571),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(572)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2572),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(572),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(573)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2573),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(573),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(574)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2574),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(574),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(575)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2575),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(575),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(576)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2576),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(576),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(577)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2577),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(577),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(578)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2578),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(578),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(579)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2543),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(579),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(580)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3225),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(580),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(581)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2544),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(581),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(582)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1676),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(582),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(583)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1694),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(2249),
    [sym__tuple_expression] = STATE(2238),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(583),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(1294),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(584)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(3098),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(584),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(585)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1696),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(585),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(586)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1697),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(586),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(587)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1698),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(587),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(588)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1699),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(588),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(589)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1702),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(589),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(590)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1678),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(590),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(591)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1679),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(591),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(592)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1687),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(592),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(593)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1680),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(593),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(594)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1681),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(594),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(595)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3213),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(595),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(596)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1673),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(596),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(597)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3132),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(597),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(598)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3136),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(3454),
    [sym__tuple_expression] = STATE(2259),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(598),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(1292),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(599)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3138),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(599),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(600)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3140),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(600),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(601)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3064),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(601),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(602)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3100),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(602),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(603)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3102),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(603),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(604)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3105),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(604),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(605)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3106),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(605),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(606)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3118),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(606),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(607)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2985),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(607),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(608)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2987),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(608),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(609)] = {
    [sym_value_definition] = STATE(6858),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1686),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(6656),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(967),
    [sym_array_get_expression] = STATE(967),
    [sym_string_get_expression] = STATE(967),
    [sym_bigarray_get_expression] = STATE(967),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(609),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(582),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(6663),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(610)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2989),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(610),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(611)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2108),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(2411),
    [sym__tuple_expression] = STATE(2238),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(611),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(1294),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(612)] = {
    [sym_value_definition] = STATE(6873),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3099),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(7114),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1663),
    [sym_array_get_expression] = STATE(1663),
    [sym_string_get_expression] = STATE(1663),
    [sym_bigarray_get_expression] = STATE(1663),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(612),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(597),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7119),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(699),
    [anon_sym_function] = ACTIONS(701),
    [anon_sym_try] = ACTIONS(703),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(705),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(613)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1728),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(613),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(614)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1750),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(2319),
    [sym__tuple_expression] = STATE(2238),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(614),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(1294),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(615)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1762),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(615),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(616)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1763),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(616),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(617)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1764),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(617),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(618)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1765),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(618),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(619)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1729),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(619),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(620)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1730),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(620),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(621)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1731),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(621),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(622)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1766),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(622),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(623)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1769),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(623),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(624)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1773),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(624),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(625)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1777),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(625),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(626)] = {
    [sym_value_definition] = STATE(7320),
    [sym__delimited_expression] = STATE(990),
    [sym__simple_expression] = STATE(78),
    [sym__expression] = STATE(1790),
    [sym_typed_expression] = STATE(990),
    [sym_labeled_tuple_element] = STATE(7174),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(991),
    [sym_array_expression] = STATE(991),
    [sym_record_expression] = STATE(991),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(990),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(990),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(990),
    [sym_package_expression] = STATE(991),
    [sym_new_expression] = STATE(990),
    [sym_object_copy_expression] = STATE(991),
    [sym_method_invocation] = STATE(990),
    [sym_object_expression] = STATE(990),
    [sym_parenthesized_expression] = STATE(991),
    [sym_attribute] = STATE(626),
    [sym_extension] = STATE(990),
    [sym_quoted_extension] = STATE(990),
    [sym__constant] = STATE(990),
    [sym_number] = STATE(1003),
    [sym_character] = STATE(1003),
    [sym_string] = STATE(1003),
    [sym_quoted_string] = STATE(1003),
    [sym_boolean] = STATE(1003),
    [sym_unit] = STATE(991),
    [sym_sign_operator] = STATE(613),
    [sym__value_name] = STATE(1047),
    [sym_parenthesized_operator] = STATE(992),
    [sym_value_path] = STATE(990),
    [sym_module_path] = STATE(6502),
    [sym_constructor_path] = STATE(990),
    [sym__instance_variable_name] = STATE(7176),
    [sym__constructor_name] = STATE(1050),
    [sym_tag] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1176),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1178),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(627)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2815),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(627),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(628)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2797),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(3230),
    [sym__tuple_expression] = STATE(2963),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(628),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(1290),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(629)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2798),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(629),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(630)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2799),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(630),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(631)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(3078),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(631),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(632)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2801),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(632),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(633)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2817),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(633),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(634)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2818),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(634),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(635)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2819),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(635),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(636)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2802),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(636),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(637)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2803),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(637),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(638)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2804),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(638),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(639)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2805),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(639),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(640)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2806),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(640),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(641)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3081),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(641),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(642)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3012),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(3449),
    [sym__tuple_expression] = STATE(2321),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(642),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(1302),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(643)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3015),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(643),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(644)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3016),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(644),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(645)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3021),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(645),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(646)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3022),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(646),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(647)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3082),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(647),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(648)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3083),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(648),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(649)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3084),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(649),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(650)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3023),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(650),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(651)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3029),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(651),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(652)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3033),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(652),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(653)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3034),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(653),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(654)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2867),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(654),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(655)] = {
    [sym_value_definition] = STATE(7078),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3047),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7215),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1666),
    [sym_array_get_expression] = STATE(1666),
    [sym_string_get_expression] = STATE(1666),
    [sym_bigarray_get_expression] = STATE(1666),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(655),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7219),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(707),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(709),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(711),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(713),
    [anon_sym_function] = ACTIONS(715),
    [anon_sym_try] = ACTIONS(717),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(719),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(656)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2846),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(3276),
    [sym__tuple_expression] = STATE(2963),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(656),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(1290),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(657)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2879),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(657),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(658)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2880),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(658),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(659)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2881),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(659),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(660)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2882),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(660),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(661)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2868),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(661),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(662)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2869),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(662),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(663)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2870),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(663),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(664)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2886),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(664),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(665)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2887),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(665),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(666)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2875),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(666),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(667)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(3123),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(667),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(668)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(3130),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(3354),
    [sym__tuple_expression] = STATE(2259),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(668),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(1292),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(669)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(3133),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(669),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(670)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(2973),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(670),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(671)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(2974),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(671),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(672)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(2975),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(672),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(673)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(2976),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(673),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(674)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(2977),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(674),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(675)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(2978),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(675),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(676)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(2979),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(676),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(677)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(2980),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(677),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(678)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(2981),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(678),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(679)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(2982),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(679),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(680)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(732),
    [sym__expression] = STATE(2983),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6391),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1668),
    [sym_array_get_expression] = STATE(1668),
    [sym_string_get_expression] = STATE(1668),
    [sym_bigarray_get_expression] = STATE(1668),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(680),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(667),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6392),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(681)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2866),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(681),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(682)] = {
    [sym_value_definition] = STATE(6965),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3200),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6993),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1412),
    [sym_array_get_expression] = STATE(1412),
    [sym_string_get_expression] = STATE(1412),
    [sym_bigarray_get_expression] = STATE(1412),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(682),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(564),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(399),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(401),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(403),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(683)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2003),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(683),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(684)] = {
    [sym_value_definition] = STATE(7308),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2865),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(6721),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1609),
    [sym_array_get_expression] = STATE(1609),
    [sym_string_get_expression] = STATE(1609),
    [sym_bigarray_get_expression] = STATE(1609),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(684),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(654),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(6727),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(663),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(667),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(685)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2548),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(685),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(686)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2549),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(2845),
    [sym__tuple_expression] = STATE(2259),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(686),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(1292),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(687)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2552),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(687),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(688)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2553),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(688),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(689)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2591),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(689),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(690)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2528),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(690),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(691)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2558),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(691),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(692)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2559),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(692),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(693)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2560),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(693),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(694)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2561),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(694),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(695)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2562),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(695),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(696)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2564),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(696),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(697)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2568),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(697),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(698)] = {
    [sym_value_definition] = STATE(6417),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(2582),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6490),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1413),
    [sym_array_get_expression] = STATE(1413),
    [sym_string_get_expression] = STATE(1413),
    [sym_bigarray_get_expression] = STATE(1413),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(698),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6491),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(341),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(353),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(357),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(363),
    [anon_sym_function] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(367),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(393),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(699)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1882),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(699),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(700)] = {
    [sym_value_definition] = STATE(6794),
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(86),
    [sym__expression] = STATE(2087),
    [sym_typed_expression] = STATE(1131),
    [sym_labeled_tuple_element] = STATE(6986),
    [sym__tuple_expression] = STATE(2164),
    [sym_cons_expression] = STATE(2177),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_application_expression] = STATE(2177),
    [sym_prefix_expression] = STATE(1131),
    [sym_sign_expression] = STATE(2177),
    [sym_hash_expression] = STATE(1131),
    [sym_infix_expression] = STATE(2177),
    [sym_field_get_expression] = STATE(1134),
    [sym_array_get_expression] = STATE(1134),
    [sym_string_get_expression] = STATE(1134),
    [sym_bigarray_get_expression] = STATE(1134),
    [sym_set_expression] = STATE(2177),
    [sym_if_expression] = STATE(2177),
    [sym_while_expression] = STATE(2177),
    [sym_for_expression] = STATE(2177),
    [sym_match_expression] = STATE(2177),
    [sym_function_expression] = STATE(2177),
    [sym_fun_expression] = STATE(2177),
    [sym_try_expression] = STATE(2177),
    [sym_let_expression] = STATE(2177),
    [sym_assert_expression] = STATE(2177),
    [sym_lazy_expression] = STATE(2177),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(700),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym_sign_operator] = STATE(683),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__instance_variable_name] = STATE(7265),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1184),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1186),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(701)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1836),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(2341),
    [sym__tuple_expression] = STATE(2321),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(701),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1302),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(702)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1838),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(702),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(703)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1839),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(703),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(704)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1840),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(704),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(705)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1841),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(705),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(706)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1941),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(706),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(707)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1826),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(707),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(708)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1816),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(708),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(709)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1842),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(709),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(710)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1843),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(710),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(711)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1803),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(711),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(712)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1865),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(712),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(713)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3164),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(713),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(714)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3165),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(3491),
    [sym__tuple_expression] = STATE(2259),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(714),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(1292),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(715)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3166),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(715),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(716)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3189),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(716),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(717)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3168),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(717),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(718)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3169),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(718),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(719)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3170),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(719),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(720)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3171),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(720),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(721)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3172),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(721),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(722)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3173),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(722),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(723)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3174),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(723),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(724)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3175),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(724),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(725)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3178),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(725),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(726)] = {
    [sym_value_definition] = STATE(7235),
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(75),
    [sym__expression] = STATE(3180),
    [sym_typed_expression] = STATE(803),
    [sym_labeled_tuple_element] = STATE(6997),
    [sym__tuple_expression] = STATE(2222),
    [sym_cons_expression] = STATE(2223),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_application_expression] = STATE(2223),
    [sym_prefix_expression] = STATE(803),
    [sym_sign_expression] = STATE(2223),
    [sym_hash_expression] = STATE(803),
    [sym_infix_expression] = STATE(2223),
    [sym_field_get_expression] = STATE(1700),
    [sym_array_get_expression] = STATE(1700),
    [sym_string_get_expression] = STATE(1700),
    [sym_bigarray_get_expression] = STATE(1700),
    [sym_set_expression] = STATE(2223),
    [sym_if_expression] = STATE(2223),
    [sym_while_expression] = STATE(2223),
    [sym_for_expression] = STATE(2223),
    [sym_match_expression] = STATE(2223),
    [sym_function_expression] = STATE(2223),
    [sym_fun_expression] = STATE(2223),
    [sym_try_expression] = STATE(2223),
    [sym_let_expression] = STATE(2223),
    [sym_assert_expression] = STATE(2223),
    [sym_lazy_expression] = STATE(2223),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(726),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym_sign_operator] = STATE(713),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__instance_variable_name] = STATE(6998),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(339),
    [anon_sym_let] = ACTIONS(721),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_fun] = ACTIONS(723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_if] = ACTIONS(725),
    [anon_sym_while] = ACTIONS(359),
    [anon_sym_for] = ACTIONS(361),
    [anon_sym_match] = ACTIONS(727),
    [anon_sym_function] = ACTIONS(729),
    [anon_sym_try] = ACTIONS(731),
    [anon_sym_assert] = ACTIONS(369),
    [anon_sym_lazy] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(373),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(391),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(733),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(727)] = {
    [sym_value_definition] = STATE(6762),
    [sym__delimited_expression] = STATE(939),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(1900),
    [sym_typed_expression] = STATE(939),
    [sym_labeled_tuple_element] = STATE(7044),
    [sym__tuple_expression] = STATE(2266),
    [sym_cons_expression] = STATE(2283),
    [sym_list_expression] = STATE(922),
    [sym_array_expression] = STATE(922),
    [sym_record_expression] = STATE(922),
    [sym_application_expression] = STATE(2283),
    [sym_prefix_expression] = STATE(939),
    [sym_sign_expression] = STATE(2283),
    [sym_hash_expression] = STATE(939),
    [sym_infix_expression] = STATE(2283),
    [sym_field_get_expression] = STATE(1066),
    [sym_array_get_expression] = STATE(1066),
    [sym_string_get_expression] = STATE(1066),
    [sym_bigarray_get_expression] = STATE(1066),
    [sym_set_expression] = STATE(2283),
    [sym_if_expression] = STATE(2283),
    [sym_while_expression] = STATE(2283),
    [sym_for_expression] = STATE(2283),
    [sym_match_expression] = STATE(2283),
    [sym_function_expression] = STATE(2283),
    [sym_fun_expression] = STATE(2283),
    [sym_try_expression] = STATE(2283),
    [sym_let_expression] = STATE(2283),
    [sym_assert_expression] = STATE(2283),
    [sym_lazy_expression] = STATE(2283),
    [sym_local_open_expression] = STATE(939),
    [sym_package_expression] = STATE(922),
    [sym_new_expression] = STATE(939),
    [sym_object_copy_expression] = STATE(922),
    [sym_method_invocation] = STATE(939),
    [sym_object_expression] = STATE(939),
    [sym_parenthesized_expression] = STATE(922),
    [sym_attribute] = STATE(727),
    [sym_extension] = STATE(939),
    [sym_quoted_extension] = STATE(939),
    [sym__constant] = STATE(939),
    [sym_number] = STATE(924),
    [sym_character] = STATE(924),
    [sym_string] = STATE(924),
    [sym_quoted_string] = STATE(924),
    [sym_boolean] = STATE(924),
    [sym_unit] = STATE(922),
    [sym_sign_operator] = STATE(699),
    [sym__value_name] = STATE(936),
    [sym_parenthesized_operator] = STATE(908),
    [sym_value_path] = STATE(939),
    [sym_module_path] = STATE(6591),
    [sym_constructor_path] = STATE(939),
    [sym__instance_variable_name] = STATE(7046),
    [sym__constructor_name] = STATE(892),
    [sym_tag] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(728)] = {
    [sym_value_definition] = STATE(7064),
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(729),
    [sym__expression] = STATE(2800),
    [sym_typed_expression] = STATE(1626),
    [sym_labeled_tuple_element] = STATE(7205),
    [sym__tuple_expression] = STATE(2925),
    [sym_cons_expression] = STATE(2926),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_application_expression] = STATE(2926),
    [sym_prefix_expression] = STATE(1626),
    [sym_sign_expression] = STATE(2926),
    [sym_hash_expression] = STATE(1626),
    [sym_infix_expression] = STATE(2926),
    [sym_field_get_expression] = STATE(1547),
    [sym_array_get_expression] = STATE(1547),
    [sym_string_get_expression] = STATE(1547),
    [sym_bigarray_get_expression] = STATE(1547),
    [sym_set_expression] = STATE(2926),
    [sym_if_expression] = STATE(2926),
    [sym_while_expression] = STATE(2926),
    [sym_for_expression] = STATE(2926),
    [sym_match_expression] = STATE(2926),
    [sym_function_expression] = STATE(2926),
    [sym_fun_expression] = STATE(2926),
    [sym_try_expression] = STATE(2926),
    [sym_let_expression] = STATE(2926),
    [sym_assert_expression] = STATE(2926),
    [sym_lazy_expression] = STATE(2926),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(728),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__instance_variable_name] = STATE(7209),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(603),
    [anon_sym_let] = ACTIONS(605),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_fun] = ACTIONS(617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_if] = ACTIONS(621),
    [anon_sym_while] = ACTIONS(623),
    [anon_sym_for] = ACTIONS(625),
    [anon_sym_match] = ACTIONS(627),
    [anon_sym_function] = ACTIONS(629),
    [anon_sym_try] = ACTIONS(631),
    [anon_sym_assert] = ACTIONS(633),
    [anon_sym_lazy] = ACTIONS(635),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(257),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(657),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(729)] = {
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(1591),
    [sym_typed_expression] = STATE(1626),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_labeled_argument] = STATE(1683),
    [sym_prefix_expression] = STATE(1626),
    [sym_hash_expression] = STATE(1626),
    [sym_field_get_expression] = STATE(1626),
    [sym_array_get_expression] = STATE(1626),
    [sym_string_get_expression] = STATE(1626),
    [sym_bigarray_get_expression] = STATE(1626),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(729),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [aux_sym_class_application_repeat1] = STATE(734),
    [sym__lowercase_identifier] = ACTIONS(1304),
    [anon_sym_and] = ACTIONS(743),
    [anon_sym_TILDE] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_COMMA] = ACTIONS(741),
    [anon_sym_PIPE] = ACTIONS(743),
    [anon_sym_DOT] = ACTIONS(1308),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_SEMI] = ACTIONS(741),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_in] = ACTIONS(743),
    [anon_sym_AMP] = ACTIONS(743),
    [anon_sym_POUND] = ACTIONS(1310),
    [anon_sym_COLON_COLON] = ACTIONS(741),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_else] = ACTIONS(743),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(741),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(743),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(743),
    [sym_hash_operator] = ACTIONS(1312),
    [anon_sym_lsl] = ACTIONS(743),
    [anon_sym_lsr] = ACTIONS(743),
    [anon_sym_asr] = ACTIONS(743),
    [aux_sym_pow_operator_token1] = ACTIONS(741),
    [anon_sym_mod] = ACTIONS(743),
    [anon_sym_land] = ACTIONS(743),
    [anon_sym_lor] = ACTIONS(743),
    [anon_sym_lxor] = ACTIONS(743),
    [aux_sym_mult_operator_token1] = ACTIONS(743),
    [aux_sym_add_operator_token1] = ACTIONS(743),
    [sym_concat_operator] = ACTIONS(741),
    [sym_rel_operator] = ACTIONS(743),
    [anon_sym_AMP_AMP] = ACTIONS(743),
    [anon_sym_or] = ACTIONS(743),
    [anon_sym_PIPE_PIPE] = ACTIONS(743),
    [sym_assign_operator] = ACTIONS(741),
    [sym_let_and_operator] = ACTIONS(741),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(730)] = {
    [sym__parenthesized_abstract_type] = STATE(1334),
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2102),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2059),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6426),
    [sym__tuple_binding_pattern] = STATE(2060),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(730),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_SEMI_SEMI] = ACTIONS(1316),
    [anon_sym_let] = ACTIONS(1318),
    [anon_sym_and] = ACTIONS(1318),
    [anon_sym_EQ] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_QMARK] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(1322),
    [anon_sym_external] = ACTIONS(1318),
    [anon_sym_type] = ACTIONS(1318),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_COLON2] = ACTIONS(1318),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_RBRACK] = ACTIONS(1316),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_module] = ACTIONS(1318),
    [anon_sym_open] = ACTIONS(1318),
    [anon_sym_include] = ACTIONS(1318),
    [anon_sym_class] = ACTIONS(1318),
    [anon_sym_val] = ACTIONS(1318),
    [anon_sym_end] = ACTIONS(1318),
    [anon_sym_COLON_GT] = ACTIONS(1316),
    [anon_sym_in] = ACTIONS(1318),
    [anon_sym_as] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_COLON_COLON] = ACTIONS(1316),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1318),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1316),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1342),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1344),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1316),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1316),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym_let_operator] = ACTIONS(1316),
    [sym_let_and_operator] = ACTIONS(1316),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_directive_token1] = ACTIONS(1318),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(731)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2103),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2059),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6426),
    [sym__tuple_binding_pattern] = STATE(2060),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(731),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_SEMI_SEMI] = ACTIONS(1316),
    [anon_sym_let] = ACTIONS(1318),
    [anon_sym_and] = ACTIONS(1318),
    [anon_sym_EQ] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_QMARK] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_external] = ACTIONS(1318),
    [anon_sym_type] = ACTIONS(1318),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_COLON2] = ACTIONS(1318),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_RBRACK] = ACTIONS(1316),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_module] = ACTIONS(1318),
    [anon_sym_open] = ACTIONS(1318),
    [anon_sym_include] = ACTIONS(1318),
    [anon_sym_class] = ACTIONS(1318),
    [anon_sym_val] = ACTIONS(1318),
    [anon_sym_end] = ACTIONS(1318),
    [anon_sym_COLON_GT] = ACTIONS(1316),
    [anon_sym_in] = ACTIONS(1318),
    [anon_sym_as] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_COLON_COLON] = ACTIONS(1316),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1318),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1316),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1342),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1344),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1316),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1316),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym_let_operator] = ACTIONS(1316),
    [sym_let_and_operator] = ACTIONS(1316),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_directive_token1] = ACTIONS(1318),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(732)] = {
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(1251),
    [sym_typed_expression] = STATE(803),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_labeled_argument] = STATE(861),
    [sym_prefix_expression] = STATE(803),
    [sym_hash_expression] = STATE(803),
    [sym_field_get_expression] = STATE(803),
    [sym_array_get_expression] = STATE(803),
    [sym_string_get_expression] = STATE(803),
    [sym_bigarray_get_expression] = STATE(803),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(732),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [aux_sym_class_application_repeat1] = STATE(735),
    [sym__lowercase_identifier] = ACTIONS(737),
    [anon_sym_TILDE] = ACTIONS(1360),
    [anon_sym_QMARK] = ACTIONS(1360),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(741),
    [anon_sym_COMMA] = ACTIONS(741),
    [anon_sym_COLON2] = ACTIONS(743),
    [anon_sym_DOT] = ACTIONS(745),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_COLON_GT] = ACTIONS(741),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_AMP] = ACTIONS(743),
    [anon_sym_POUND] = ACTIONS(747),
    [anon_sym_COLON_COLON] = ACTIONS(741),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_else] = ACTIONS(743),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(743),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(743),
    [sym_hash_operator] = ACTIONS(1362),
    [anon_sym_lsl] = ACTIONS(743),
    [anon_sym_lsr] = ACTIONS(743),
    [anon_sym_asr] = ACTIONS(743),
    [aux_sym_pow_operator_token1] = ACTIONS(741),
    [anon_sym_mod] = ACTIONS(743),
    [anon_sym_land] = ACTIONS(743),
    [anon_sym_lor] = ACTIONS(743),
    [anon_sym_lxor] = ACTIONS(743),
    [aux_sym_mult_operator_token1] = ACTIONS(743),
    [aux_sym_add_operator_token1] = ACTIONS(743),
    [sym_concat_operator] = ACTIONS(741),
    [sym_rel_operator] = ACTIONS(743),
    [anon_sym_AMP_AMP] = ACTIONS(743),
    [anon_sym_or] = ACTIONS(743),
    [anon_sym_PIPE_PIPE] = ACTIONS(743),
    [sym_assign_operator] = ACTIONS(741),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(733)] = {
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(1591),
    [sym_typed_expression] = STATE(1626),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_labeled_argument] = STATE(1683),
    [sym_prefix_expression] = STATE(1626),
    [sym_hash_expression] = STATE(1626),
    [sym_field_get_expression] = STATE(1626),
    [sym_array_get_expression] = STATE(1626),
    [sym_string_get_expression] = STATE(1626),
    [sym_bigarray_get_expression] = STATE(1626),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(733),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [aux_sym_class_application_repeat1] = STATE(733),
    [sym__lowercase_identifier] = ACTIONS(1364),
    [anon_sym_and] = ACTIONS(786),
    [anon_sym_TILDE] = ACTIONS(1367),
    [anon_sym_QMARK] = ACTIONS(1367),
    [anon_sym_LPAREN] = ACTIONS(1370),
    [anon_sym_COMMA] = ACTIONS(784),
    [anon_sym_PIPE] = ACTIONS(786),
    [anon_sym_LBRACK] = ACTIONS(1373),
    [anon_sym_true] = ACTIONS(1376),
    [anon_sym_false] = ACTIONS(1376),
    [anon_sym_LBRACE] = ACTIONS(1379),
    [anon_sym_SEMI] = ACTIONS(784),
    [anon_sym_object] = ACTIONS(1382),
    [anon_sym_in] = ACTIONS(786),
    [anon_sym_AMP] = ACTIONS(786),
    [anon_sym_COLON_COLON] = ACTIONS(784),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1385),
    [anon_sym_else] = ACTIONS(786),
    [anon_sym_new] = ACTIONS(1388),
    [anon_sym_LBRACE_LT] = ACTIONS(1391),
    [anon_sym_begin] = ACTIONS(1394),
    [sym_ocamlyacc_value] = ACTIONS(1397),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(784),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1400),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1403),
    [aux_sym_number_token1] = ACTIONS(1406),
    [aux_sym_signed_number_token1] = ACTIONS(786),
    [anon_sym_SQUOTE] = ACTIONS(1409),
    [anon_sym_DQUOTE] = ACTIONS(1412),
    [sym_prefix_operator] = ACTIONS(1415),
    [aux_sym_sign_operator_token1] = ACTIONS(786),
    [anon_sym_lsl] = ACTIONS(786),
    [anon_sym_lsr] = ACTIONS(786),
    [anon_sym_asr] = ACTIONS(786),
    [aux_sym_pow_operator_token1] = ACTIONS(784),
    [anon_sym_mod] = ACTIONS(786),
    [anon_sym_land] = ACTIONS(786),
    [anon_sym_lor] = ACTIONS(786),
    [anon_sym_lxor] = ACTIONS(786),
    [aux_sym_mult_operator_token1] = ACTIONS(786),
    [aux_sym_add_operator_token1] = ACTIONS(786),
    [sym_concat_operator] = ACTIONS(784),
    [sym_rel_operator] = ACTIONS(786),
    [anon_sym_AMP_AMP] = ACTIONS(786),
    [anon_sym_or] = ACTIONS(786),
    [anon_sym_PIPE_PIPE] = ACTIONS(786),
    [sym_assign_operator] = ACTIONS(784),
    [sym_let_and_operator] = ACTIONS(784),
    [sym__uppercase_identifier] = ACTIONS(1418),
    [aux_sym_tag_token1] = ACTIONS(1421),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(734)] = {
    [sym__delimited_expression] = STATE(1626),
    [sym__simple_expression] = STATE(1591),
    [sym_typed_expression] = STATE(1626),
    [sym_list_expression] = STATE(1589),
    [sym_array_expression] = STATE(1589),
    [sym_record_expression] = STATE(1589),
    [sym_labeled_argument] = STATE(1683),
    [sym_prefix_expression] = STATE(1626),
    [sym_hash_expression] = STATE(1626),
    [sym_field_get_expression] = STATE(1626),
    [sym_array_get_expression] = STATE(1626),
    [sym_string_get_expression] = STATE(1626),
    [sym_bigarray_get_expression] = STATE(1626),
    [sym_local_open_expression] = STATE(1626),
    [sym_package_expression] = STATE(1589),
    [sym_new_expression] = STATE(1626),
    [sym_object_copy_expression] = STATE(1589),
    [sym_method_invocation] = STATE(1626),
    [sym_object_expression] = STATE(1626),
    [sym_parenthesized_expression] = STATE(1589),
    [sym_attribute] = STATE(734),
    [sym_extension] = STATE(1626),
    [sym_quoted_extension] = STATE(1626),
    [sym__constant] = STATE(1626),
    [sym_number] = STATE(1566),
    [sym_character] = STATE(1566),
    [sym_string] = STATE(1566),
    [sym_quoted_string] = STATE(1566),
    [sym_boolean] = STATE(1566),
    [sym_unit] = STATE(1589),
    [sym__value_name] = STATE(1612),
    [sym_parenthesized_operator] = STATE(1607),
    [sym_value_path] = STATE(1626),
    [sym_module_path] = STATE(6548),
    [sym_constructor_path] = STATE(1626),
    [sym__constructor_name] = STATE(1567),
    [sym_tag] = STATE(1626),
    [aux_sym_class_application_repeat1] = STATE(733),
    [sym__lowercase_identifier] = ACTIONS(1304),
    [anon_sym_and] = ACTIONS(773),
    [anon_sym_TILDE] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(607),
    [anon_sym_COMMA] = ACTIONS(771),
    [anon_sym_PIPE] = ACTIONS(773),
    [anon_sym_LBRACK] = ACTIONS(609),
    [anon_sym_true] = ACTIONS(611),
    [anon_sym_false] = ACTIONS(611),
    [anon_sym_LBRACE] = ACTIONS(613),
    [anon_sym_SEMI] = ACTIONS(771),
    [anon_sym_object] = ACTIONS(615),
    [anon_sym_in] = ACTIONS(773),
    [anon_sym_AMP] = ACTIONS(773),
    [anon_sym_COLON_COLON] = ACTIONS(771),
    [anon_sym_LBRACK_PIPE] = ACTIONS(619),
    [anon_sym_else] = ACTIONS(773),
    [anon_sym_new] = ACTIONS(637),
    [anon_sym_LBRACE_LT] = ACTIONS(639),
    [anon_sym_begin] = ACTIONS(641),
    [sym_ocamlyacc_value] = ACTIONS(643),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(771),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(645),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(647),
    [aux_sym_number_token1] = ACTIONS(649),
    [aux_sym_signed_number_token1] = ACTIONS(773),
    [anon_sym_SQUOTE] = ACTIONS(651),
    [anon_sym_DQUOTE] = ACTIONS(653),
    [sym_prefix_operator] = ACTIONS(655),
    [aux_sym_sign_operator_token1] = ACTIONS(773),
    [anon_sym_lsl] = ACTIONS(773),
    [anon_sym_lsr] = ACTIONS(773),
    [anon_sym_asr] = ACTIONS(773),
    [aux_sym_pow_operator_token1] = ACTIONS(771),
    [anon_sym_mod] = ACTIONS(773),
    [anon_sym_land] = ACTIONS(773),
    [anon_sym_lor] = ACTIONS(773),
    [anon_sym_lxor] = ACTIONS(773),
    [aux_sym_mult_operator_token1] = ACTIONS(773),
    [aux_sym_add_operator_token1] = ACTIONS(773),
    [sym_concat_operator] = ACTIONS(771),
    [sym_rel_operator] = ACTIONS(773),
    [anon_sym_AMP_AMP] = ACTIONS(773),
    [anon_sym_or] = ACTIONS(773),
    [anon_sym_PIPE_PIPE] = ACTIONS(773),
    [sym_assign_operator] = ACTIONS(771),
    [sym_let_and_operator] = ACTIONS(771),
    [sym__uppercase_identifier] = ACTIONS(659),
    [aux_sym_tag_token1] = ACTIONS(661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(735)] = {
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(1251),
    [sym_typed_expression] = STATE(803),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_labeled_argument] = STATE(861),
    [sym_prefix_expression] = STATE(803),
    [sym_hash_expression] = STATE(803),
    [sym_field_get_expression] = STATE(803),
    [sym_array_get_expression] = STATE(803),
    [sym_string_get_expression] = STATE(803),
    [sym_bigarray_get_expression] = STATE(803),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(735),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [aux_sym_class_application_repeat1] = STATE(91),
    [sym__lowercase_identifier] = ACTIONS(737),
    [anon_sym_TILDE] = ACTIONS(1360),
    [anon_sym_QMARK] = ACTIONS(1360),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(771),
    [anon_sym_COMMA] = ACTIONS(771),
    [anon_sym_COLON2] = ACTIONS(773),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_COLON_GT] = ACTIONS(771),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_AMP] = ACTIONS(773),
    [anon_sym_COLON_COLON] = ACTIONS(771),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_else] = ACTIONS(773),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(407),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(409),
    [aux_sym_number_token1] = ACTIONS(385),
    [aux_sym_signed_number_token1] = ACTIONS(773),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [aux_sym_sign_operator_token1] = ACTIONS(773),
    [anon_sym_lsl] = ACTIONS(773),
    [anon_sym_lsr] = ACTIONS(773),
    [anon_sym_asr] = ACTIONS(773),
    [aux_sym_pow_operator_token1] = ACTIONS(771),
    [anon_sym_mod] = ACTIONS(773),
    [anon_sym_land] = ACTIONS(773),
    [anon_sym_lor] = ACTIONS(773),
    [anon_sym_lxor] = ACTIONS(773),
    [aux_sym_mult_operator_token1] = ACTIONS(773),
    [aux_sym_add_operator_token1] = ACTIONS(773),
    [sym_concat_operator] = ACTIONS(771),
    [sym_rel_operator] = ACTIONS(773),
    [anon_sym_AMP_AMP] = ACTIONS(773),
    [anon_sym_or] = ACTIONS(773),
    [anon_sym_PIPE_PIPE] = ACTIONS(773),
    [sym_assign_operator] = ACTIONS(771),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(736)] = {
    [sym__local_structure_item] = STATE(6515),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(736),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(764),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(737)] = {
    [sym__local_structure_item] = STATE(6775),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(737),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(763),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(738)] = {
    [sym__local_structure_item] = STATE(7283),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(738),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(757),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(739)] = {
    [sym__local_structure_item] = STATE(6957),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(739),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(756),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(740)] = {
    [sym__local_structure_item] = STATE(7285),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(740),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(753),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(741)] = {
    [sym__local_structure_item] = STATE(7028),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(741),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(761),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(742)] = {
    [sym__local_structure_item] = STATE(6496),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(742),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(759),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(743)] = {
    [sym__local_structure_item] = STATE(6791),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(743),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(765),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(744)] = {
    [sym__local_structure_item] = STATE(7228),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(744),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(755),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(745)] = {
    [sym__local_structure_item] = STATE(6752),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(745),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(752),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(746)] = {
    [sym__local_structure_item] = STATE(7122),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(746),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(762),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(747)] = {
    [sym__local_structure_item] = STATE(7068),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(747),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(754),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(748)] = {
    [sym__local_structure_item] = STATE(7228),
    [sym_let_binding] = STATE(3773),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2054),
    [sym__binding_pattern_no_exn] = STATE(750),
    [sym_alias_binding_pattern_no_exn] = STATE(2054),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2054),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6346),
    [sym__tuple_binding_pattern_no_exn] = STATE(2054),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern_no_exn] = STATE(2054),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2054),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_attribute] = STATE(748),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(760),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_rec] = ACTIONS(1480),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1342),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1344),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(749)] = {
    [sym__local_structure_item] = STATE(7002),
    [sym_let_binding] = STATE(5511),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(749),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__attribute] = STATE(751),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(750)] = {
    [sym_parameter] = STATE(2236),
    [sym__coerced] = STATE(7072),
    [sym__maybe_polymorphic_typed] = STATE(5650),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(750),
    [sym_item_attribute] = STATE(3816),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3520),
    [aux_sym_let_binding_repeat1] = STATE(1535),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(1484),
    [anon_sym_let] = ACTIONS(1486),
    [anon_sym_and] = ACTIONS(1486),
    [anon_sym_EQ] = ACTIONS(1488),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(1486),
    [anon_sym_type] = ACTIONS(1486),
    [anon_sym_COMMA] = ACTIONS(1494),
    [anon_sym_PIPE] = ACTIONS(1496),
    [anon_sym_COLON2] = ACTIONS(1498),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(1484),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(1486),
    [anon_sym_module] = ACTIONS(1486),
    [anon_sym_open] = ACTIONS(1486),
    [anon_sym_include] = ACTIONS(1486),
    [anon_sym_class] = ACTIONS(1486),
    [anon_sym_val] = ACTIONS(1486),
    [anon_sym_end] = ACTIONS(1486),
    [anon_sym_COLON_GT] = ACTIONS(1506),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_as] = ACTIONS(1508),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_COLON_COLON] = ACTIONS(1512),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1518),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1484),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1484),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1484),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym_let_operator] = ACTIONS(1484),
    [sym_let_and_operator] = ACTIONS(1484),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_directive_token1] = ACTIONS(1486),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(751)] = {
    [sym__local_structure_item] = STATE(7008),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(751),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(752)] = {
    [sym__local_structure_item] = STATE(6868),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(752),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(753)] = {
    [sym__local_structure_item] = STATE(6398),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(753),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(754)] = {
    [sym__local_structure_item] = STATE(7160),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(754),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(755)] = {
    [sym__local_structure_item] = STATE(7079),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(755),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(756)] = {
    [sym__local_structure_item] = STATE(6864),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(756),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(757)] = {
    [sym__local_structure_item] = STATE(6503),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(757),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(758)] = {
    [sym_parameter] = STATE(2236),
    [sym__coerced] = STATE(6528),
    [sym__maybe_polymorphic_typed] = STATE(5556),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(758),
    [sym_item_attribute] = STATE(3816),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3520),
    [aux_sym_let_binding_repeat1] = STATE(1534),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(1484),
    [anon_sym_let] = ACTIONS(1486),
    [anon_sym_and] = ACTIONS(1486),
    [anon_sym_EQ] = ACTIONS(1534),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(1486),
    [anon_sym_type] = ACTIONS(1486),
    [anon_sym_COMMA] = ACTIONS(1494),
    [anon_sym_PIPE] = ACTIONS(1496),
    [anon_sym_COLON2] = ACTIONS(1498),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(1484),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(1486),
    [anon_sym_module] = ACTIONS(1486),
    [anon_sym_open] = ACTIONS(1486),
    [anon_sym_include] = ACTIONS(1486),
    [anon_sym_class] = ACTIONS(1486),
    [anon_sym_val] = ACTIONS(1486),
    [anon_sym_end] = ACTIONS(1486),
    [anon_sym_COLON_GT] = ACTIONS(1506),
    [anon_sym_as] = ACTIONS(1508),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_COLON_COLON] = ACTIONS(1512),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1518),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1484),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1484),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1484),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym_let_operator] = ACTIONS(1484),
    [sym_let_and_operator] = ACTIONS(1484),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_directive_token1] = ACTIONS(1486),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(759)] = {
    [sym__local_structure_item] = STATE(6448),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(759),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(760)] = {
    [sym__local_structure_item] = STATE(7079),
    [sym_let_binding] = STATE(3752),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2054),
    [sym__binding_pattern_no_exn] = STATE(750),
    [sym_alias_binding_pattern_no_exn] = STATE(2054),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2054),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6346),
    [sym__tuple_binding_pattern_no_exn] = STATE(2054),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern_no_exn] = STATE(2054),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2054),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_attribute] = STATE(760),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_rec] = ACTIONS(1536),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1342),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1344),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(761)] = {
    [sym__local_structure_item] = STATE(7319),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(761),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(762)] = {
    [sym__local_structure_item] = STATE(7140),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(762),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(763)] = {
    [sym__local_structure_item] = STATE(6817),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(763),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(764)] = {
    [sym__local_structure_item] = STATE(6587),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(764),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(765)] = {
    [sym__local_structure_item] = STATE(6449),
    [sym_let_binding] = STATE(5498),
    [sym_external] = STATE(6973),
    [sym_type_definition] = STATE(6973),
    [sym_exception_definition] = STATE(6973),
    [sym_module_definition] = STATE(6973),
    [sym_module_type_definition] = STATE(6973),
    [sym_open_module] = STATE(6973),
    [sym_class_definition] = STATE(6973),
    [sym_class_type_definition] = STATE(6973),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(765),
    [sym_floating_attribute] = STATE(6973),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym_item_extension] = STATE(6973),
    [sym_quoted_item_extension] = STATE(6973),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_external] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(1440),
    [anon_sym_module] = ACTIONS(1442),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_class] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1456),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1458),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1460),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1462),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1464),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(766)] = {
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(1114),
    [sym_typed_expression] = STATE(1131),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_labeled_argument] = STATE(1164),
    [sym_prefix_expression] = STATE(1131),
    [sym_hash_expression] = STATE(1131),
    [sym_field_get_expression] = STATE(1131),
    [sym_array_get_expression] = STATE(1131),
    [sym_string_get_expression] = STATE(1131),
    [sym_bigarray_get_expression] = STATE(1131),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(766),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [aux_sym_class_application_repeat1] = STATE(767),
    [sym__lowercase_identifier] = ACTIONS(959),
    [anon_sym_SEMI_SEMI] = ACTIONS(1538),
    [anon_sym_let] = ACTIONS(1540),
    [anon_sym_and] = ACTIONS(1540),
    [anon_sym_TILDE] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(961),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(1540),
    [anon_sym_type] = ACTIONS(1540),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(1538),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(1540),
    [anon_sym_module] = ACTIONS(1540),
    [anon_sym_open] = ACTIONS(1540),
    [anon_sym_include] = ACTIONS(1540),
    [anon_sym_class] = ACTIONS(1540),
    [anon_sym_val] = ACTIONS(1540),
    [anon_sym_end] = ACTIONS(1540),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1540),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1538),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1538),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1538),
    [aux_sym_number_token1] = ACTIONS(217),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [sym_let_operator] = ACTIONS(1538),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(1538),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(767)] = {
    [sym__delimited_expression] = STATE(1131),
    [sym__simple_expression] = STATE(1114),
    [sym_typed_expression] = STATE(1131),
    [sym_list_expression] = STATE(1113),
    [sym_array_expression] = STATE(1113),
    [sym_record_expression] = STATE(1113),
    [sym_labeled_argument] = STATE(1164),
    [sym_prefix_expression] = STATE(1131),
    [sym_hash_expression] = STATE(1131),
    [sym_field_get_expression] = STATE(1131),
    [sym_array_get_expression] = STATE(1131),
    [sym_string_get_expression] = STATE(1131),
    [sym_bigarray_get_expression] = STATE(1131),
    [sym_local_open_expression] = STATE(1131),
    [sym_package_expression] = STATE(1113),
    [sym_new_expression] = STATE(1131),
    [sym_object_copy_expression] = STATE(1113),
    [sym_method_invocation] = STATE(1131),
    [sym_object_expression] = STATE(1131),
    [sym_parenthesized_expression] = STATE(1113),
    [sym_attribute] = STATE(767),
    [sym_extension] = STATE(1131),
    [sym_quoted_extension] = STATE(1131),
    [sym__constant] = STATE(1131),
    [sym_number] = STATE(1070),
    [sym_character] = STATE(1070),
    [sym_string] = STATE(1070),
    [sym_quoted_string] = STATE(1070),
    [sym_boolean] = STATE(1070),
    [sym_unit] = STATE(1113),
    [sym__value_name] = STATE(1078),
    [sym_parenthesized_operator] = STATE(1115),
    [sym_value_path] = STATE(1131),
    [sym_module_path] = STATE(7251),
    [sym_constructor_path] = STATE(1131),
    [sym__constructor_name] = STATE(1112),
    [sym_tag] = STATE(1131),
    [aux_sym_class_application_repeat1] = STATE(88),
    [sym__lowercase_identifier] = ACTIONS(959),
    [anon_sym_SEMI_SEMI] = ACTIONS(1542),
    [anon_sym_let] = ACTIONS(1544),
    [anon_sym_and] = ACTIONS(1544),
    [anon_sym_TILDE] = ACTIONS(961),
    [anon_sym_QMARK] = ACTIONS(961),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(1544),
    [anon_sym_type] = ACTIONS(1544),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(1542),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(1544),
    [anon_sym_module] = ACTIONS(1544),
    [anon_sym_open] = ACTIONS(1544),
    [anon_sym_include] = ACTIONS(1544),
    [anon_sym_class] = ACTIONS(1544),
    [anon_sym_val] = ACTIONS(1544),
    [anon_sym_end] = ACTIONS(1544),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1544),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1542),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1542),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1542),
    [aux_sym_number_token1] = ACTIONS(217),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [sym_let_operator] = ACTIONS(1542),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(1542),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(768)] = {
    [sym_attribute] = STATE(768),
    [sym__lowercase_identifier] = ACTIONS(1546),
    [anon_sym_and] = ACTIONS(1546),
    [anon_sym_TILDE] = ACTIONS(1546),
    [anon_sym_QMARK] = ACTIONS(1546),
    [anon_sym_LPAREN] = ACTIONS(1548),
    [anon_sym_RPAREN] = ACTIONS(1548),
    [anon_sym_COMMA] = ACTIONS(1548),
    [anon_sym_PIPE] = ACTIONS(1546),
    [anon_sym_COLON2] = ACTIONS(1546),
    [anon_sym_DOT] = ACTIONS(1548),
    [anon_sym_DASH_GT] = ACTIONS(1546),
    [anon_sym_LBRACK] = ACTIONS(1546),
    [anon_sym_RBRACK] = ACTIONS(1548),
    [anon_sym_true] = ACTIONS(1546),
    [anon_sym_false] = ACTIONS(1546),
    [anon_sym_LBRACE] = ACTIONS(1546),
    [anon_sym_SEMI] = ACTIONS(1548),
    [anon_sym_RBRACE] = ACTIONS(1548),
    [anon_sym_constraint] = ACTIONS(1546),
    [anon_sym_val] = ACTIONS(1546),
    [anon_sym_end] = ACTIONS(1546),
    [anon_sym_with] = ACTIONS(1546),
    [anon_sym_COLON_GT] = ACTIONS(1548),
    [anon_sym_object] = ACTIONS(1546),
    [anon_sym_inherit] = ACTIONS(1546),
    [anon_sym_method] = ACTIONS(1546),
    [anon_sym_in] = ACTIONS(1546),
    [anon_sym_as] = ACTIONS(1546),
    [anon_sym_initializer] = ACTIONS(1546),
    [anon_sym_AMP] = ACTIONS(1546),
    [anon_sym_POUND] = ACTIONS(1546),
    [anon_sym_COLON_COLON] = ACTIONS(1548),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1548),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1548),
    [anon_sym_LT_DASH] = ACTIONS(1546),
    [anon_sym_then] = ACTIONS(1546),
    [anon_sym_else] = ACTIONS(1546),
    [anon_sym_do] = ACTIONS(1546),
    [anon_sym_done] = ACTIONS(1546),
    [anon_sym_to] = ACTIONS(1546),
    [anon_sym_downto] = ACTIONS(1546),
    [anon_sym_new] = ACTIONS(1546),
    [anon_sym_LBRACE_LT] = ACTIONS(1548),
    [anon_sym_GT_RBRACE] = ACTIONS(1548),
    [anon_sym_begin] = ACTIONS(1546),
    [sym_ocamlyacc_value] = ACTIONS(1548),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1546),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1548),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1546),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1546),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1548),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1548),
    [aux_sym_number_token1] = ACTIONS(1548),
    [aux_sym_signed_number_token1] = ACTIONS(1546),
    [anon_sym_SQUOTE] = ACTIONS(1548),
    [anon_sym_DQUOTE] = ACTIONS(1548),
    [sym_prefix_operator] = ACTIONS(1548),
    [aux_sym_sign_operator_token1] = ACTIONS(1546),
    [sym_hash_operator] = ACTIONS(1548),
    [anon_sym_lsl] = ACTIONS(1546),
    [anon_sym_lsr] = ACTIONS(1546),
    [anon_sym_asr] = ACTIONS(1546),
    [aux_sym_pow_operator_token1] = ACTIONS(1548),
    [anon_sym_mod] = ACTIONS(1546),
    [anon_sym_land] = ACTIONS(1546),
    [anon_sym_lor] = ACTIONS(1546),
    [anon_sym_lxor] = ACTIONS(1546),
    [aux_sym_mult_operator_token1] = ACTIONS(1546),
    [aux_sym_add_operator_token1] = ACTIONS(1546),
    [sym_concat_operator] = ACTIONS(1548),
    [sym_rel_operator] = ACTIONS(1546),
    [anon_sym_AMP_AMP] = ACTIONS(1546),
    [anon_sym_or] = ACTIONS(1546),
    [anon_sym_PIPE_PIPE] = ACTIONS(1546),
    [sym_assign_operator] = ACTIONS(1548),
    [sym__uppercase_identifier] = ACTIONS(1548),
    [aux_sym_tag_token1] = ACTIONS(1548),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(769)] = {
    [sym_attribute] = STATE(769),
    [sym__lowercase_identifier] = ACTIONS(1550),
    [anon_sym_and] = ACTIONS(1550),
    [anon_sym_TILDE] = ACTIONS(1550),
    [anon_sym_QMARK] = ACTIONS(1550),
    [anon_sym_LPAREN] = ACTIONS(1552),
    [anon_sym_RPAREN] = ACTIONS(1552),
    [anon_sym_COMMA] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1550),
    [anon_sym_COLON2] = ACTIONS(1550),
    [anon_sym_DOT] = ACTIONS(1552),
    [anon_sym_DASH_GT] = ACTIONS(1550),
    [anon_sym_LBRACK] = ACTIONS(1550),
    [anon_sym_RBRACK] = ACTIONS(1552),
    [anon_sym_true] = ACTIONS(1550),
    [anon_sym_false] = ACTIONS(1550),
    [anon_sym_LBRACE] = ACTIONS(1550),
    [anon_sym_SEMI] = ACTIONS(1552),
    [anon_sym_RBRACE] = ACTIONS(1552),
    [anon_sym_constraint] = ACTIONS(1550),
    [anon_sym_val] = ACTIONS(1550),
    [anon_sym_end] = ACTIONS(1550),
    [anon_sym_with] = ACTIONS(1550),
    [anon_sym_COLON_GT] = ACTIONS(1552),
    [anon_sym_object] = ACTIONS(1550),
    [anon_sym_inherit] = ACTIONS(1550),
    [anon_sym_method] = ACTIONS(1550),
    [anon_sym_in] = ACTIONS(1550),
    [anon_sym_as] = ACTIONS(1550),
    [anon_sym_initializer] = ACTIONS(1550),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_POUND] = ACTIONS(1550),
    [anon_sym_COLON_COLON] = ACTIONS(1552),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1552),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1552),
    [anon_sym_LT_DASH] = ACTIONS(1550),
    [anon_sym_then] = ACTIONS(1550),
    [anon_sym_else] = ACTIONS(1550),
    [anon_sym_do] = ACTIONS(1550),
    [anon_sym_done] = ACTIONS(1550),
    [anon_sym_to] = ACTIONS(1550),
    [anon_sym_downto] = ACTIONS(1550),
    [anon_sym_new] = ACTIONS(1550),
    [anon_sym_LBRACE_LT] = ACTIONS(1552),
    [anon_sym_GT_RBRACE] = ACTIONS(1552),
    [anon_sym_begin] = ACTIONS(1550),
    [sym_ocamlyacc_value] = ACTIONS(1552),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1550),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1552),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1550),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1550),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1552),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1552),
    [aux_sym_number_token1] = ACTIONS(1552),
    [aux_sym_signed_number_token1] = ACTIONS(1550),
    [anon_sym_SQUOTE] = ACTIONS(1552),
    [anon_sym_DQUOTE] = ACTIONS(1552),
    [sym_prefix_operator] = ACTIONS(1552),
    [aux_sym_sign_operator_token1] = ACTIONS(1550),
    [sym_hash_operator] = ACTIONS(1552),
    [anon_sym_lsl] = ACTIONS(1550),
    [anon_sym_lsr] = ACTIONS(1550),
    [anon_sym_asr] = ACTIONS(1550),
    [aux_sym_pow_operator_token1] = ACTIONS(1552),
    [anon_sym_mod] = ACTIONS(1550),
    [anon_sym_land] = ACTIONS(1550),
    [anon_sym_lor] = ACTIONS(1550),
    [anon_sym_lxor] = ACTIONS(1550),
    [aux_sym_mult_operator_token1] = ACTIONS(1550),
    [aux_sym_add_operator_token1] = ACTIONS(1550),
    [sym_concat_operator] = ACTIONS(1552),
    [sym_rel_operator] = ACTIONS(1550),
    [anon_sym_AMP_AMP] = ACTIONS(1550),
    [anon_sym_or] = ACTIONS(1550),
    [anon_sym_PIPE_PIPE] = ACTIONS(1550),
    [sym_assign_operator] = ACTIONS(1552),
    [sym__uppercase_identifier] = ACTIONS(1552),
    [aux_sym_tag_token1] = ACTIONS(1552),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(770)] = {
    [sym_attribute] = STATE(770),
    [sym__lowercase_identifier] = ACTIONS(1554),
    [anon_sym_and] = ACTIONS(1554),
    [anon_sym_TILDE] = ACTIONS(1554),
    [anon_sym_QMARK] = ACTIONS(1554),
    [anon_sym_LPAREN] = ACTIONS(1556),
    [anon_sym_RPAREN] = ACTIONS(1556),
    [anon_sym_COMMA] = ACTIONS(1556),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_COLON2] = ACTIONS(1554),
    [anon_sym_DOT] = ACTIONS(1556),
    [anon_sym_DASH_GT] = ACTIONS(1554),
    [anon_sym_LBRACK] = ACTIONS(1554),
    [anon_sym_RBRACK] = ACTIONS(1556),
    [anon_sym_true] = ACTIONS(1554),
    [anon_sym_false] = ACTIONS(1554),
    [anon_sym_LBRACE] = ACTIONS(1554),
    [anon_sym_SEMI] = ACTIONS(1556),
    [anon_sym_RBRACE] = ACTIONS(1556),
    [anon_sym_constraint] = ACTIONS(1554),
    [anon_sym_val] = ACTIONS(1554),
    [anon_sym_end] = ACTIONS(1554),
    [anon_sym_with] = ACTIONS(1554),
    [anon_sym_COLON_GT] = ACTIONS(1556),
    [anon_sym_object] = ACTIONS(1554),
    [anon_sym_inherit] = ACTIONS(1554),
    [anon_sym_method] = ACTIONS(1554),
    [anon_sym_in] = ACTIONS(1554),
    [anon_sym_as] = ACTIONS(1554),
    [anon_sym_initializer] = ACTIONS(1554),
    [anon_sym_AMP] = ACTIONS(1554),
    [anon_sym_POUND] = ACTIONS(1554),
    [anon_sym_COLON_COLON] = ACTIONS(1556),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1556),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1556),
    [anon_sym_LT_DASH] = ACTIONS(1554),
    [anon_sym_then] = ACTIONS(1554),
    [anon_sym_else] = ACTIONS(1554),
    [anon_sym_do] = ACTIONS(1554),
    [anon_sym_done] = ACTIONS(1554),
    [anon_sym_to] = ACTIONS(1554),
    [anon_sym_downto] = ACTIONS(1554),
    [anon_sym_new] = ACTIONS(1554),
    [anon_sym_LBRACE_LT] = ACTIONS(1556),
    [anon_sym_GT_RBRACE] = ACTIONS(1556),
    [anon_sym_begin] = ACTIONS(1554),
    [sym_ocamlyacc_value] = ACTIONS(1556),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1554),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1556),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1554),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1554),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1556),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1556),
    [aux_sym_number_token1] = ACTIONS(1556),
    [aux_sym_signed_number_token1] = ACTIONS(1554),
    [anon_sym_SQUOTE] = ACTIONS(1556),
    [anon_sym_DQUOTE] = ACTIONS(1556),
    [sym_prefix_operator] = ACTIONS(1556),
    [aux_sym_sign_operator_token1] = ACTIONS(1554),
    [sym_hash_operator] = ACTIONS(1556),
    [anon_sym_lsl] = ACTIONS(1554),
    [anon_sym_lsr] = ACTIONS(1554),
    [anon_sym_asr] = ACTIONS(1554),
    [aux_sym_pow_operator_token1] = ACTIONS(1556),
    [anon_sym_mod] = ACTIONS(1554),
    [anon_sym_land] = ACTIONS(1554),
    [anon_sym_lor] = ACTIONS(1554),
    [anon_sym_lxor] = ACTIONS(1554),
    [aux_sym_mult_operator_token1] = ACTIONS(1554),
    [aux_sym_add_operator_token1] = ACTIONS(1554),
    [sym_concat_operator] = ACTIONS(1556),
    [sym_rel_operator] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1554),
    [anon_sym_or] = ACTIONS(1554),
    [anon_sym_PIPE_PIPE] = ACTIONS(1554),
    [sym_assign_operator] = ACTIONS(1556),
    [sym__uppercase_identifier] = ACTIONS(1556),
    [aux_sym_tag_token1] = ACTIONS(1556),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(771)] = {
    [sym_attribute] = STATE(771),
    [sym__lowercase_identifier] = ACTIONS(1558),
    [anon_sym_and] = ACTIONS(1558),
    [anon_sym_TILDE] = ACTIONS(1558),
    [anon_sym_QMARK] = ACTIONS(1558),
    [anon_sym_LPAREN] = ACTIONS(1560),
    [anon_sym_RPAREN] = ACTIONS(1560),
    [anon_sym_COMMA] = ACTIONS(1560),
    [anon_sym_PIPE] = ACTIONS(1558),
    [anon_sym_COLON2] = ACTIONS(1558),
    [anon_sym_DOT] = ACTIONS(1560),
    [anon_sym_DASH_GT] = ACTIONS(1558),
    [anon_sym_LBRACK] = ACTIONS(1558),
    [anon_sym_RBRACK] = ACTIONS(1560),
    [anon_sym_true] = ACTIONS(1558),
    [anon_sym_false] = ACTIONS(1558),
    [anon_sym_LBRACE] = ACTIONS(1558),
    [anon_sym_SEMI] = ACTIONS(1560),
    [anon_sym_RBRACE] = ACTIONS(1560),
    [anon_sym_constraint] = ACTIONS(1558),
    [anon_sym_val] = ACTIONS(1558),
    [anon_sym_end] = ACTIONS(1558),
    [anon_sym_with] = ACTIONS(1558),
    [anon_sym_COLON_GT] = ACTIONS(1560),
    [anon_sym_object] = ACTIONS(1558),
    [anon_sym_inherit] = ACTIONS(1558),
    [anon_sym_method] = ACTIONS(1558),
    [anon_sym_in] = ACTIONS(1558),
    [anon_sym_as] = ACTIONS(1558),
    [anon_sym_initializer] = ACTIONS(1558),
    [anon_sym_AMP] = ACTIONS(1558),
    [anon_sym_POUND] = ACTIONS(1558),
    [anon_sym_COLON_COLON] = ACTIONS(1560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1560),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1560),
    [anon_sym_LT_DASH] = ACTIONS(1558),
    [anon_sym_then] = ACTIONS(1558),
    [anon_sym_else] = ACTIONS(1558),
    [anon_sym_do] = ACTIONS(1558),
    [anon_sym_done] = ACTIONS(1558),
    [anon_sym_to] = ACTIONS(1558),
    [anon_sym_downto] = ACTIONS(1558),
    [anon_sym_new] = ACTIONS(1558),
    [anon_sym_LBRACE_LT] = ACTIONS(1560),
    [anon_sym_GT_RBRACE] = ACTIONS(1560),
    [anon_sym_begin] = ACTIONS(1558),
    [sym_ocamlyacc_value] = ACTIONS(1560),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1558),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1560),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1558),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1558),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1560),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1560),
    [aux_sym_number_token1] = ACTIONS(1560),
    [aux_sym_signed_number_token1] = ACTIONS(1558),
    [anon_sym_SQUOTE] = ACTIONS(1560),
    [anon_sym_DQUOTE] = ACTIONS(1560),
    [sym_prefix_operator] = ACTIONS(1560),
    [aux_sym_sign_operator_token1] = ACTIONS(1558),
    [sym_hash_operator] = ACTIONS(1560),
    [anon_sym_lsl] = ACTIONS(1558),
    [anon_sym_lsr] = ACTIONS(1558),
    [anon_sym_asr] = ACTIONS(1558),
    [aux_sym_pow_operator_token1] = ACTIONS(1560),
    [anon_sym_mod] = ACTIONS(1558),
    [anon_sym_land] = ACTIONS(1558),
    [anon_sym_lor] = ACTIONS(1558),
    [anon_sym_lxor] = ACTIONS(1558),
    [aux_sym_mult_operator_token1] = ACTIONS(1558),
    [aux_sym_add_operator_token1] = ACTIONS(1558),
    [sym_concat_operator] = ACTIONS(1560),
    [sym_rel_operator] = ACTIONS(1558),
    [anon_sym_AMP_AMP] = ACTIONS(1558),
    [anon_sym_or] = ACTIONS(1558),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [sym_assign_operator] = ACTIONS(1560),
    [sym__uppercase_identifier] = ACTIONS(1560),
    [aux_sym_tag_token1] = ACTIONS(1560),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(772)] = {
    [sym_attribute] = STATE(772),
    [sym__lowercase_identifier] = ACTIONS(1562),
    [anon_sym_and] = ACTIONS(1562),
    [anon_sym_TILDE] = ACTIONS(1562),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_LPAREN] = ACTIONS(1564),
    [anon_sym_RPAREN] = ACTIONS(1564),
    [anon_sym_COMMA] = ACTIONS(1564),
    [anon_sym_PIPE] = ACTIONS(1562),
    [anon_sym_COLON2] = ACTIONS(1562),
    [anon_sym_DOT] = ACTIONS(1564),
    [anon_sym_DASH_GT] = ACTIONS(1562),
    [anon_sym_LBRACK] = ACTIONS(1562),
    [anon_sym_RBRACK] = ACTIONS(1564),
    [anon_sym_true] = ACTIONS(1562),
    [anon_sym_false] = ACTIONS(1562),
    [anon_sym_LBRACE] = ACTIONS(1562),
    [anon_sym_SEMI] = ACTIONS(1564),
    [anon_sym_RBRACE] = ACTIONS(1564),
    [anon_sym_constraint] = ACTIONS(1562),
    [anon_sym_val] = ACTIONS(1562),
    [anon_sym_end] = ACTIONS(1562),
    [anon_sym_with] = ACTIONS(1562),
    [anon_sym_COLON_GT] = ACTIONS(1564),
    [anon_sym_object] = ACTIONS(1562),
    [anon_sym_inherit] = ACTIONS(1562),
    [anon_sym_method] = ACTIONS(1562),
    [anon_sym_in] = ACTIONS(1562),
    [anon_sym_as] = ACTIONS(1562),
    [anon_sym_initializer] = ACTIONS(1562),
    [anon_sym_AMP] = ACTIONS(1562),
    [anon_sym_POUND] = ACTIONS(1562),
    [anon_sym_COLON_COLON] = ACTIONS(1564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1564),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1564),
    [anon_sym_LT_DASH] = ACTIONS(1562),
    [anon_sym_then] = ACTIONS(1562),
    [anon_sym_else] = ACTIONS(1562),
    [anon_sym_do] = ACTIONS(1562),
    [anon_sym_done] = ACTIONS(1562),
    [anon_sym_to] = ACTIONS(1562),
    [anon_sym_downto] = ACTIONS(1562),
    [anon_sym_new] = ACTIONS(1562),
    [anon_sym_LBRACE_LT] = ACTIONS(1564),
    [anon_sym_GT_RBRACE] = ACTIONS(1564),
    [anon_sym_begin] = ACTIONS(1562),
    [sym_ocamlyacc_value] = ACTIONS(1564),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1562),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1564),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1562),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1562),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1564),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1564),
    [aux_sym_number_token1] = ACTIONS(1564),
    [aux_sym_signed_number_token1] = ACTIONS(1562),
    [anon_sym_SQUOTE] = ACTIONS(1564),
    [anon_sym_DQUOTE] = ACTIONS(1564),
    [sym_prefix_operator] = ACTIONS(1564),
    [aux_sym_sign_operator_token1] = ACTIONS(1562),
    [sym_hash_operator] = ACTIONS(1564),
    [anon_sym_lsl] = ACTIONS(1562),
    [anon_sym_lsr] = ACTIONS(1562),
    [anon_sym_asr] = ACTIONS(1562),
    [aux_sym_pow_operator_token1] = ACTIONS(1564),
    [anon_sym_mod] = ACTIONS(1562),
    [anon_sym_land] = ACTIONS(1562),
    [anon_sym_lor] = ACTIONS(1562),
    [anon_sym_lxor] = ACTIONS(1562),
    [aux_sym_mult_operator_token1] = ACTIONS(1562),
    [aux_sym_add_operator_token1] = ACTIONS(1562),
    [sym_concat_operator] = ACTIONS(1564),
    [sym_rel_operator] = ACTIONS(1562),
    [anon_sym_AMP_AMP] = ACTIONS(1562),
    [anon_sym_or] = ACTIONS(1562),
    [anon_sym_PIPE_PIPE] = ACTIONS(1562),
    [sym_assign_operator] = ACTIONS(1564),
    [sym__uppercase_identifier] = ACTIONS(1564),
    [aux_sym_tag_token1] = ACTIONS(1564),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(773)] = {
    [sym_attribute] = STATE(773),
    [sym__lowercase_identifier] = ACTIONS(1566),
    [anon_sym_and] = ACTIONS(1566),
    [anon_sym_TILDE] = ACTIONS(1566),
    [anon_sym_QMARK] = ACTIONS(1566),
    [anon_sym_LPAREN] = ACTIONS(1568),
    [anon_sym_RPAREN] = ACTIONS(1568),
    [anon_sym_COMMA] = ACTIONS(1568),
    [anon_sym_PIPE] = ACTIONS(1566),
    [anon_sym_COLON2] = ACTIONS(1566),
    [anon_sym_DOT] = ACTIONS(1568),
    [anon_sym_DASH_GT] = ACTIONS(1566),
    [anon_sym_LBRACK] = ACTIONS(1566),
    [anon_sym_RBRACK] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1566),
    [anon_sym_false] = ACTIONS(1566),
    [anon_sym_LBRACE] = ACTIONS(1566),
    [anon_sym_SEMI] = ACTIONS(1568),
    [anon_sym_RBRACE] = ACTIONS(1568),
    [anon_sym_constraint] = ACTIONS(1566),
    [anon_sym_val] = ACTIONS(1566),
    [anon_sym_end] = ACTIONS(1566),
    [anon_sym_with] = ACTIONS(1566),
    [anon_sym_COLON_GT] = ACTIONS(1568),
    [anon_sym_object] = ACTIONS(1566),
    [anon_sym_inherit] = ACTIONS(1566),
    [anon_sym_method] = ACTIONS(1566),
    [anon_sym_in] = ACTIONS(1566),
    [anon_sym_as] = ACTIONS(1566),
    [anon_sym_initializer] = ACTIONS(1566),
    [anon_sym_AMP] = ACTIONS(1566),
    [anon_sym_POUND] = ACTIONS(1566),
    [anon_sym_COLON_COLON] = ACTIONS(1568),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1568),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1568),
    [anon_sym_LT_DASH] = ACTIONS(1566),
    [anon_sym_then] = ACTIONS(1566),
    [anon_sym_else] = ACTIONS(1566),
    [anon_sym_do] = ACTIONS(1566),
    [anon_sym_done] = ACTIONS(1566),
    [anon_sym_to] = ACTIONS(1566),
    [anon_sym_downto] = ACTIONS(1566),
    [anon_sym_new] = ACTIONS(1566),
    [anon_sym_LBRACE_LT] = ACTIONS(1568),
    [anon_sym_GT_RBRACE] = ACTIONS(1568),
    [anon_sym_begin] = ACTIONS(1566),
    [sym_ocamlyacc_value] = ACTIONS(1568),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1566),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1568),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1566),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1566),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1568),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1568),
    [aux_sym_number_token1] = ACTIONS(1568),
    [aux_sym_signed_number_token1] = ACTIONS(1566),
    [anon_sym_SQUOTE] = ACTIONS(1568),
    [anon_sym_DQUOTE] = ACTIONS(1568),
    [sym_prefix_operator] = ACTIONS(1568),
    [aux_sym_sign_operator_token1] = ACTIONS(1566),
    [sym_hash_operator] = ACTIONS(1568),
    [anon_sym_lsl] = ACTIONS(1566),
    [anon_sym_lsr] = ACTIONS(1566),
    [anon_sym_asr] = ACTIONS(1566),
    [aux_sym_pow_operator_token1] = ACTIONS(1568),
    [anon_sym_mod] = ACTIONS(1566),
    [anon_sym_land] = ACTIONS(1566),
    [anon_sym_lor] = ACTIONS(1566),
    [anon_sym_lxor] = ACTIONS(1566),
    [aux_sym_mult_operator_token1] = ACTIONS(1566),
    [aux_sym_add_operator_token1] = ACTIONS(1566),
    [sym_concat_operator] = ACTIONS(1568),
    [sym_rel_operator] = ACTIONS(1566),
    [anon_sym_AMP_AMP] = ACTIONS(1566),
    [anon_sym_or] = ACTIONS(1566),
    [anon_sym_PIPE_PIPE] = ACTIONS(1566),
    [sym_assign_operator] = ACTIONS(1568),
    [sym__uppercase_identifier] = ACTIONS(1568),
    [aux_sym_tag_token1] = ACTIONS(1568),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(774)] = {
    [sym_attribute] = STATE(774),
    [sym__lowercase_identifier] = ACTIONS(1570),
    [anon_sym_and] = ACTIONS(1570),
    [anon_sym_TILDE] = ACTIONS(1570),
    [anon_sym_QMARK] = ACTIONS(1570),
    [anon_sym_LPAREN] = ACTIONS(1572),
    [anon_sym_RPAREN] = ACTIONS(1572),
    [anon_sym_COMMA] = ACTIONS(1572),
    [anon_sym_PIPE] = ACTIONS(1570),
    [anon_sym_COLON2] = ACTIONS(1570),
    [anon_sym_DOT] = ACTIONS(1572),
    [anon_sym_DASH_GT] = ACTIONS(1570),
    [anon_sym_LBRACK] = ACTIONS(1570),
    [anon_sym_RBRACK] = ACTIONS(1572),
    [anon_sym_true] = ACTIONS(1570),
    [anon_sym_false] = ACTIONS(1570),
    [anon_sym_LBRACE] = ACTIONS(1570),
    [anon_sym_SEMI] = ACTIONS(1572),
    [anon_sym_RBRACE] = ACTIONS(1572),
    [anon_sym_constraint] = ACTIONS(1570),
    [anon_sym_val] = ACTIONS(1570),
    [anon_sym_end] = ACTIONS(1570),
    [anon_sym_with] = ACTIONS(1570),
    [anon_sym_COLON_GT] = ACTIONS(1572),
    [anon_sym_object] = ACTIONS(1570),
    [anon_sym_inherit] = ACTIONS(1570),
    [anon_sym_method] = ACTIONS(1570),
    [anon_sym_in] = ACTIONS(1570),
    [anon_sym_as] = ACTIONS(1570),
    [anon_sym_initializer] = ACTIONS(1570),
    [anon_sym_AMP] = ACTIONS(1570),
    [anon_sym_POUND] = ACTIONS(1570),
    [anon_sym_COLON_COLON] = ACTIONS(1572),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1572),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1572),
    [anon_sym_LT_DASH] = ACTIONS(1570),
    [anon_sym_then] = ACTIONS(1570),
    [anon_sym_else] = ACTIONS(1570),
    [anon_sym_do] = ACTIONS(1570),
    [anon_sym_done] = ACTIONS(1570),
    [anon_sym_to] = ACTIONS(1570),
    [anon_sym_downto] = ACTIONS(1570),
    [anon_sym_new] = ACTIONS(1570),
    [anon_sym_LBRACE_LT] = ACTIONS(1572),
    [anon_sym_GT_RBRACE] = ACTIONS(1572),
    [anon_sym_begin] = ACTIONS(1570),
    [sym_ocamlyacc_value] = ACTIONS(1572),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1570),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1572),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1570),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1570),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1572),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1572),
    [aux_sym_number_token1] = ACTIONS(1572),
    [aux_sym_signed_number_token1] = ACTIONS(1570),
    [anon_sym_SQUOTE] = ACTIONS(1572),
    [anon_sym_DQUOTE] = ACTIONS(1572),
    [sym_prefix_operator] = ACTIONS(1572),
    [aux_sym_sign_operator_token1] = ACTIONS(1570),
    [sym_hash_operator] = ACTIONS(1572),
    [anon_sym_lsl] = ACTIONS(1570),
    [anon_sym_lsr] = ACTIONS(1570),
    [anon_sym_asr] = ACTIONS(1570),
    [aux_sym_pow_operator_token1] = ACTIONS(1572),
    [anon_sym_mod] = ACTIONS(1570),
    [anon_sym_land] = ACTIONS(1570),
    [anon_sym_lor] = ACTIONS(1570),
    [anon_sym_lxor] = ACTIONS(1570),
    [aux_sym_mult_operator_token1] = ACTIONS(1570),
    [aux_sym_add_operator_token1] = ACTIONS(1570),
    [sym_concat_operator] = ACTIONS(1572),
    [sym_rel_operator] = ACTIONS(1570),
    [anon_sym_AMP_AMP] = ACTIONS(1570),
    [anon_sym_or] = ACTIONS(1570),
    [anon_sym_PIPE_PIPE] = ACTIONS(1570),
    [sym_assign_operator] = ACTIONS(1572),
    [sym__uppercase_identifier] = ACTIONS(1572),
    [aux_sym_tag_token1] = ACTIONS(1572),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(775)] = {
    [sym_attribute] = STATE(775),
    [sym__lowercase_identifier] = ACTIONS(1574),
    [anon_sym_and] = ACTIONS(1574),
    [anon_sym_TILDE] = ACTIONS(1574),
    [anon_sym_QMARK] = ACTIONS(1574),
    [anon_sym_LPAREN] = ACTIONS(1576),
    [anon_sym_RPAREN] = ACTIONS(1576),
    [anon_sym_COMMA] = ACTIONS(1576),
    [anon_sym_PIPE] = ACTIONS(1574),
    [anon_sym_COLON2] = ACTIONS(1574),
    [anon_sym_DOT] = ACTIONS(1576),
    [anon_sym_DASH_GT] = ACTIONS(1574),
    [anon_sym_LBRACK] = ACTIONS(1574),
    [anon_sym_RBRACK] = ACTIONS(1576),
    [anon_sym_true] = ACTIONS(1574),
    [anon_sym_false] = ACTIONS(1574),
    [anon_sym_LBRACE] = ACTIONS(1574),
    [anon_sym_SEMI] = ACTIONS(1576),
    [anon_sym_RBRACE] = ACTIONS(1576),
    [anon_sym_constraint] = ACTIONS(1574),
    [anon_sym_val] = ACTIONS(1574),
    [anon_sym_end] = ACTIONS(1574),
    [anon_sym_with] = ACTIONS(1574),
    [anon_sym_COLON_GT] = ACTIONS(1576),
    [anon_sym_object] = ACTIONS(1574),
    [anon_sym_inherit] = ACTIONS(1574),
    [anon_sym_method] = ACTIONS(1574),
    [anon_sym_in] = ACTIONS(1574),
    [anon_sym_as] = ACTIONS(1574),
    [anon_sym_initializer] = ACTIONS(1574),
    [anon_sym_AMP] = ACTIONS(1574),
    [anon_sym_POUND] = ACTIONS(1574),
    [anon_sym_COLON_COLON] = ACTIONS(1576),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1576),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1576),
    [anon_sym_LT_DASH] = ACTIONS(1574),
    [anon_sym_then] = ACTIONS(1574),
    [anon_sym_else] = ACTIONS(1574),
    [anon_sym_do] = ACTIONS(1574),
    [anon_sym_done] = ACTIONS(1574),
    [anon_sym_to] = ACTIONS(1574),
    [anon_sym_downto] = ACTIONS(1574),
    [anon_sym_new] = ACTIONS(1574),
    [anon_sym_LBRACE_LT] = ACTIONS(1576),
    [anon_sym_GT_RBRACE] = ACTIONS(1576),
    [anon_sym_begin] = ACTIONS(1574),
    [sym_ocamlyacc_value] = ACTIONS(1576),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1574),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1576),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1574),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1574),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1576),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1576),
    [aux_sym_number_token1] = ACTIONS(1576),
    [aux_sym_signed_number_token1] = ACTIONS(1574),
    [anon_sym_SQUOTE] = ACTIONS(1576),
    [anon_sym_DQUOTE] = ACTIONS(1576),
    [sym_prefix_operator] = ACTIONS(1576),
    [aux_sym_sign_operator_token1] = ACTIONS(1574),
    [sym_hash_operator] = ACTIONS(1576),
    [anon_sym_lsl] = ACTIONS(1574),
    [anon_sym_lsr] = ACTIONS(1574),
    [anon_sym_asr] = ACTIONS(1574),
    [aux_sym_pow_operator_token1] = ACTIONS(1576),
    [anon_sym_mod] = ACTIONS(1574),
    [anon_sym_land] = ACTIONS(1574),
    [anon_sym_lor] = ACTIONS(1574),
    [anon_sym_lxor] = ACTIONS(1574),
    [aux_sym_mult_operator_token1] = ACTIONS(1574),
    [aux_sym_add_operator_token1] = ACTIONS(1574),
    [sym_concat_operator] = ACTIONS(1576),
    [sym_rel_operator] = ACTIONS(1574),
    [anon_sym_AMP_AMP] = ACTIONS(1574),
    [anon_sym_or] = ACTIONS(1574),
    [anon_sym_PIPE_PIPE] = ACTIONS(1574),
    [sym_assign_operator] = ACTIONS(1576),
    [sym__uppercase_identifier] = ACTIONS(1576),
    [aux_sym_tag_token1] = ACTIONS(1576),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(776)] = {
    [sym_attribute] = STATE(776),
    [sym__lowercase_identifier] = ACTIONS(1578),
    [anon_sym_and] = ACTIONS(1578),
    [anon_sym_TILDE] = ACTIONS(1578),
    [anon_sym_QMARK] = ACTIONS(1578),
    [anon_sym_LPAREN] = ACTIONS(1580),
    [anon_sym_RPAREN] = ACTIONS(1580),
    [anon_sym_COMMA] = ACTIONS(1580),
    [anon_sym_PIPE] = ACTIONS(1578),
    [anon_sym_COLON2] = ACTIONS(1578),
    [anon_sym_DOT] = ACTIONS(1580),
    [anon_sym_DASH_GT] = ACTIONS(1578),
    [anon_sym_LBRACK] = ACTIONS(1578),
    [anon_sym_RBRACK] = ACTIONS(1580),
    [anon_sym_true] = ACTIONS(1578),
    [anon_sym_false] = ACTIONS(1578),
    [anon_sym_LBRACE] = ACTIONS(1578),
    [anon_sym_SEMI] = ACTIONS(1580),
    [anon_sym_RBRACE] = ACTIONS(1580),
    [anon_sym_constraint] = ACTIONS(1578),
    [anon_sym_val] = ACTIONS(1578),
    [anon_sym_end] = ACTIONS(1578),
    [anon_sym_with] = ACTIONS(1578),
    [anon_sym_COLON_GT] = ACTIONS(1580),
    [anon_sym_object] = ACTIONS(1578),
    [anon_sym_inherit] = ACTIONS(1578),
    [anon_sym_method] = ACTIONS(1578),
    [anon_sym_in] = ACTIONS(1578),
    [anon_sym_as] = ACTIONS(1578),
    [anon_sym_initializer] = ACTIONS(1578),
    [anon_sym_AMP] = ACTIONS(1578),
    [anon_sym_POUND] = ACTIONS(1578),
    [anon_sym_COLON_COLON] = ACTIONS(1580),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1580),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1580),
    [anon_sym_LT_DASH] = ACTIONS(1578),
    [anon_sym_then] = ACTIONS(1578),
    [anon_sym_else] = ACTIONS(1578),
    [anon_sym_do] = ACTIONS(1578),
    [anon_sym_done] = ACTIONS(1578),
    [anon_sym_to] = ACTIONS(1578),
    [anon_sym_downto] = ACTIONS(1578),
    [anon_sym_new] = ACTIONS(1578),
    [anon_sym_LBRACE_LT] = ACTIONS(1580),
    [anon_sym_GT_RBRACE] = ACTIONS(1580),
    [anon_sym_begin] = ACTIONS(1578),
    [sym_ocamlyacc_value] = ACTIONS(1580),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1578),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1580),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1578),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1578),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1580),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1580),
    [aux_sym_number_token1] = ACTIONS(1580),
    [aux_sym_signed_number_token1] = ACTIONS(1578),
    [anon_sym_SQUOTE] = ACTIONS(1580),
    [anon_sym_DQUOTE] = ACTIONS(1580),
    [sym_prefix_operator] = ACTIONS(1580),
    [aux_sym_sign_operator_token1] = ACTIONS(1578),
    [sym_hash_operator] = ACTIONS(1580),
    [anon_sym_lsl] = ACTIONS(1578),
    [anon_sym_lsr] = ACTIONS(1578),
    [anon_sym_asr] = ACTIONS(1578),
    [aux_sym_pow_operator_token1] = ACTIONS(1580),
    [anon_sym_mod] = ACTIONS(1578),
    [anon_sym_land] = ACTIONS(1578),
    [anon_sym_lor] = ACTIONS(1578),
    [anon_sym_lxor] = ACTIONS(1578),
    [aux_sym_mult_operator_token1] = ACTIONS(1578),
    [aux_sym_add_operator_token1] = ACTIONS(1578),
    [sym_concat_operator] = ACTIONS(1580),
    [sym_rel_operator] = ACTIONS(1578),
    [anon_sym_AMP_AMP] = ACTIONS(1578),
    [anon_sym_or] = ACTIONS(1578),
    [anon_sym_PIPE_PIPE] = ACTIONS(1578),
    [sym_assign_operator] = ACTIONS(1580),
    [sym__uppercase_identifier] = ACTIONS(1580),
    [aux_sym_tag_token1] = ACTIONS(1580),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(777)] = {
    [sym_attribute] = STATE(777),
    [sym__lowercase_identifier] = ACTIONS(1582),
    [anon_sym_and] = ACTIONS(1582),
    [anon_sym_TILDE] = ACTIONS(1582),
    [anon_sym_QMARK] = ACTIONS(1582),
    [anon_sym_LPAREN] = ACTIONS(1584),
    [anon_sym_RPAREN] = ACTIONS(1584),
    [anon_sym_COMMA] = ACTIONS(1584),
    [anon_sym_PIPE] = ACTIONS(1582),
    [anon_sym_COLON2] = ACTIONS(1582),
    [anon_sym_DOT] = ACTIONS(1584),
    [anon_sym_DASH_GT] = ACTIONS(1582),
    [anon_sym_LBRACK] = ACTIONS(1582),
    [anon_sym_RBRACK] = ACTIONS(1584),
    [anon_sym_true] = ACTIONS(1582),
    [anon_sym_false] = ACTIONS(1582),
    [anon_sym_LBRACE] = ACTIONS(1582),
    [anon_sym_SEMI] = ACTIONS(1584),
    [anon_sym_RBRACE] = ACTIONS(1584),
    [anon_sym_constraint] = ACTIONS(1582),
    [anon_sym_val] = ACTIONS(1582),
    [anon_sym_end] = ACTIONS(1582),
    [anon_sym_with] = ACTIONS(1582),
    [anon_sym_COLON_GT] = ACTIONS(1584),
    [anon_sym_object] = ACTIONS(1582),
    [anon_sym_inherit] = ACTIONS(1582),
    [anon_sym_method] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1582),
    [anon_sym_as] = ACTIONS(1582),
    [anon_sym_initializer] = ACTIONS(1582),
    [anon_sym_AMP] = ACTIONS(1582),
    [anon_sym_POUND] = ACTIONS(1582),
    [anon_sym_COLON_COLON] = ACTIONS(1584),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1584),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1584),
    [anon_sym_then] = ACTIONS(1582),
    [anon_sym_else] = ACTIONS(1582),
    [anon_sym_do] = ACTIONS(1582),
    [anon_sym_done] = ACTIONS(1582),
    [anon_sym_to] = ACTIONS(1582),
    [anon_sym_downto] = ACTIONS(1582),
    [anon_sym_new] = ACTIONS(1582),
    [anon_sym_LBRACE_LT] = ACTIONS(1584),
    [anon_sym_GT_RBRACE] = ACTIONS(1584),
    [anon_sym_begin] = ACTIONS(1582),
    [sym_ocamlyacc_value] = ACTIONS(1584),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1582),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1584),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1582),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1582),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1584),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1584),
    [aux_sym_number_token1] = ACTIONS(1584),
    [aux_sym_signed_number_token1] = ACTIONS(1582),
    [anon_sym_SQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1584),
    [sym_prefix_operator] = ACTIONS(1584),
    [aux_sym_sign_operator_token1] = ACTIONS(1582),
    [sym_hash_operator] = ACTIONS(1584),
    [anon_sym_lsl] = ACTIONS(1582),
    [anon_sym_lsr] = ACTIONS(1582),
    [anon_sym_asr] = ACTIONS(1582),
    [aux_sym_pow_operator_token1] = ACTIONS(1584),
    [anon_sym_mod] = ACTIONS(1582),
    [anon_sym_land] = ACTIONS(1582),
    [anon_sym_lor] = ACTIONS(1582),
    [anon_sym_lxor] = ACTIONS(1582),
    [aux_sym_mult_operator_token1] = ACTIONS(1582),
    [aux_sym_add_operator_token1] = ACTIONS(1582),
    [sym_concat_operator] = ACTIONS(1584),
    [sym_rel_operator] = ACTIONS(1582),
    [anon_sym_AMP_AMP] = ACTIONS(1582),
    [anon_sym_or] = ACTIONS(1582),
    [anon_sym_PIPE_PIPE] = ACTIONS(1582),
    [sym_assign_operator] = ACTIONS(1584),
    [sym__uppercase_identifier] = ACTIONS(1584),
    [aux_sym_tag_token1] = ACTIONS(1584),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(778)] = {
    [sym_attribute] = STATE(778),
    [sym__lowercase_identifier] = ACTIONS(1586),
    [anon_sym_and] = ACTIONS(1586),
    [anon_sym_TILDE] = ACTIONS(1586),
    [anon_sym_QMARK] = ACTIONS(1586),
    [anon_sym_LPAREN] = ACTIONS(1588),
    [anon_sym_RPAREN] = ACTIONS(1588),
    [anon_sym_COMMA] = ACTIONS(1588),
    [anon_sym_PIPE] = ACTIONS(1586),
    [anon_sym_COLON2] = ACTIONS(1586),
    [anon_sym_DOT] = ACTIONS(1588),
    [anon_sym_DASH_GT] = ACTIONS(1586),
    [anon_sym_LBRACK] = ACTIONS(1586),
    [anon_sym_RBRACK] = ACTIONS(1588),
    [anon_sym_true] = ACTIONS(1586),
    [anon_sym_false] = ACTIONS(1586),
    [anon_sym_LBRACE] = ACTIONS(1586),
    [anon_sym_SEMI] = ACTIONS(1588),
    [anon_sym_RBRACE] = ACTIONS(1588),
    [anon_sym_constraint] = ACTIONS(1586),
    [anon_sym_val] = ACTIONS(1586),
    [anon_sym_end] = ACTIONS(1586),
    [anon_sym_with] = ACTIONS(1586),
    [anon_sym_COLON_GT] = ACTIONS(1588),
    [anon_sym_object] = ACTIONS(1586),
    [anon_sym_inherit] = ACTIONS(1586),
    [anon_sym_method] = ACTIONS(1586),
    [anon_sym_in] = ACTIONS(1586),
    [anon_sym_as] = ACTIONS(1586),
    [anon_sym_initializer] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1586),
    [anon_sym_POUND] = ACTIONS(1586),
    [anon_sym_COLON_COLON] = ACTIONS(1588),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1588),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1588),
    [anon_sym_then] = ACTIONS(1586),
    [anon_sym_else] = ACTIONS(1586),
    [anon_sym_do] = ACTIONS(1586),
    [anon_sym_done] = ACTIONS(1586),
    [anon_sym_to] = ACTIONS(1586),
    [anon_sym_downto] = ACTIONS(1586),
    [anon_sym_new] = ACTIONS(1586),
    [anon_sym_LBRACE_LT] = ACTIONS(1588),
    [anon_sym_GT_RBRACE] = ACTIONS(1588),
    [anon_sym_begin] = ACTIONS(1586),
    [sym_ocamlyacc_value] = ACTIONS(1588),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1586),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1588),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1586),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1586),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1588),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1588),
    [aux_sym_number_token1] = ACTIONS(1588),
    [aux_sym_signed_number_token1] = ACTIONS(1586),
    [anon_sym_SQUOTE] = ACTIONS(1588),
    [anon_sym_DQUOTE] = ACTIONS(1588),
    [sym_prefix_operator] = ACTIONS(1588),
    [aux_sym_sign_operator_token1] = ACTIONS(1586),
    [sym_hash_operator] = ACTIONS(1588),
    [anon_sym_lsl] = ACTIONS(1586),
    [anon_sym_lsr] = ACTIONS(1586),
    [anon_sym_asr] = ACTIONS(1586),
    [aux_sym_pow_operator_token1] = ACTIONS(1588),
    [anon_sym_mod] = ACTIONS(1586),
    [anon_sym_land] = ACTIONS(1586),
    [anon_sym_lor] = ACTIONS(1586),
    [anon_sym_lxor] = ACTIONS(1586),
    [aux_sym_mult_operator_token1] = ACTIONS(1586),
    [aux_sym_add_operator_token1] = ACTIONS(1586),
    [sym_concat_operator] = ACTIONS(1588),
    [sym_rel_operator] = ACTIONS(1586),
    [anon_sym_AMP_AMP] = ACTIONS(1586),
    [anon_sym_or] = ACTIONS(1586),
    [anon_sym_PIPE_PIPE] = ACTIONS(1586),
    [sym_assign_operator] = ACTIONS(1588),
    [sym__uppercase_identifier] = ACTIONS(1588),
    [aux_sym_tag_token1] = ACTIONS(1588),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(779)] = {
    [sym_attribute] = STATE(779),
    [sym__lowercase_identifier] = ACTIONS(1590),
    [anon_sym_and] = ACTIONS(1590),
    [anon_sym_TILDE] = ACTIONS(1590),
    [anon_sym_QMARK] = ACTIONS(1590),
    [anon_sym_LPAREN] = ACTIONS(1592),
    [anon_sym_RPAREN] = ACTIONS(1592),
    [anon_sym_COMMA] = ACTIONS(1592),
    [anon_sym_PIPE] = ACTIONS(1590),
    [anon_sym_COLON2] = ACTIONS(1590),
    [anon_sym_DOT] = ACTIONS(1592),
    [anon_sym_DASH_GT] = ACTIONS(1590),
    [anon_sym_LBRACK] = ACTIONS(1590),
    [anon_sym_RBRACK] = ACTIONS(1592),
    [anon_sym_true] = ACTIONS(1590),
    [anon_sym_false] = ACTIONS(1590),
    [anon_sym_LBRACE] = ACTIONS(1590),
    [anon_sym_SEMI] = ACTIONS(1592),
    [anon_sym_RBRACE] = ACTIONS(1592),
    [anon_sym_constraint] = ACTIONS(1590),
    [anon_sym_val] = ACTIONS(1590),
    [anon_sym_end] = ACTIONS(1590),
    [anon_sym_with] = ACTIONS(1590),
    [anon_sym_COLON_GT] = ACTIONS(1592),
    [anon_sym_object] = ACTIONS(1590),
    [anon_sym_inherit] = ACTIONS(1590),
    [anon_sym_method] = ACTIONS(1590),
    [anon_sym_in] = ACTIONS(1590),
    [anon_sym_as] = ACTIONS(1590),
    [anon_sym_initializer] = ACTIONS(1590),
    [anon_sym_AMP] = ACTIONS(1590),
    [anon_sym_POUND] = ACTIONS(1590),
    [anon_sym_COLON_COLON] = ACTIONS(1592),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1592),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1592),
    [anon_sym_then] = ACTIONS(1590),
    [anon_sym_else] = ACTIONS(1590),
    [anon_sym_do] = ACTIONS(1590),
    [anon_sym_done] = ACTIONS(1590),
    [anon_sym_to] = ACTIONS(1590),
    [anon_sym_downto] = ACTIONS(1590),
    [anon_sym_new] = ACTIONS(1590),
    [anon_sym_LBRACE_LT] = ACTIONS(1592),
    [anon_sym_GT_RBRACE] = ACTIONS(1592),
    [anon_sym_begin] = ACTIONS(1590),
    [sym_ocamlyacc_value] = ACTIONS(1592),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1590),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1592),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1590),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1590),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1592),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1592),
    [aux_sym_number_token1] = ACTIONS(1592),
    [aux_sym_signed_number_token1] = ACTIONS(1590),
    [anon_sym_SQUOTE] = ACTIONS(1592),
    [anon_sym_DQUOTE] = ACTIONS(1592),
    [sym_prefix_operator] = ACTIONS(1592),
    [aux_sym_sign_operator_token1] = ACTIONS(1590),
    [sym_hash_operator] = ACTIONS(1592),
    [anon_sym_lsl] = ACTIONS(1590),
    [anon_sym_lsr] = ACTIONS(1590),
    [anon_sym_asr] = ACTIONS(1590),
    [aux_sym_pow_operator_token1] = ACTIONS(1592),
    [anon_sym_mod] = ACTIONS(1590),
    [anon_sym_land] = ACTIONS(1590),
    [anon_sym_lor] = ACTIONS(1590),
    [anon_sym_lxor] = ACTIONS(1590),
    [aux_sym_mult_operator_token1] = ACTIONS(1590),
    [aux_sym_add_operator_token1] = ACTIONS(1590),
    [sym_concat_operator] = ACTIONS(1592),
    [sym_rel_operator] = ACTIONS(1590),
    [anon_sym_AMP_AMP] = ACTIONS(1590),
    [anon_sym_or] = ACTIONS(1590),
    [anon_sym_PIPE_PIPE] = ACTIONS(1590),
    [sym_assign_operator] = ACTIONS(1592),
    [sym__uppercase_identifier] = ACTIONS(1592),
    [aux_sym_tag_token1] = ACTIONS(1592),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(780)] = {
    [sym_attribute] = STATE(780),
    [sym__lowercase_identifier] = ACTIONS(1594),
    [anon_sym_and] = ACTIONS(1594),
    [anon_sym_TILDE] = ACTIONS(1594),
    [anon_sym_QMARK] = ACTIONS(1594),
    [anon_sym_LPAREN] = ACTIONS(1596),
    [anon_sym_RPAREN] = ACTIONS(1596),
    [anon_sym_COMMA] = ACTIONS(1596),
    [anon_sym_PIPE] = ACTIONS(1594),
    [anon_sym_COLON2] = ACTIONS(1594),
    [anon_sym_DOT] = ACTIONS(1596),
    [anon_sym_DASH_GT] = ACTIONS(1594),
    [anon_sym_LBRACK] = ACTIONS(1594),
    [anon_sym_RBRACK] = ACTIONS(1596),
    [anon_sym_true] = ACTIONS(1594),
    [anon_sym_false] = ACTIONS(1594),
    [anon_sym_LBRACE] = ACTIONS(1594),
    [anon_sym_SEMI] = ACTIONS(1596),
    [anon_sym_RBRACE] = ACTIONS(1596),
    [anon_sym_constraint] = ACTIONS(1594),
    [anon_sym_val] = ACTIONS(1594),
    [anon_sym_end] = ACTIONS(1594),
    [anon_sym_with] = ACTIONS(1594),
    [anon_sym_COLON_GT] = ACTIONS(1596),
    [anon_sym_object] = ACTIONS(1594),
    [anon_sym_inherit] = ACTIONS(1594),
    [anon_sym_method] = ACTIONS(1594),
    [anon_sym_in] = ACTIONS(1594),
    [anon_sym_as] = ACTIONS(1594),
    [anon_sym_initializer] = ACTIONS(1594),
    [anon_sym_AMP] = ACTIONS(1594),
    [anon_sym_POUND] = ACTIONS(1594),
    [anon_sym_COLON_COLON] = ACTIONS(1596),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1596),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1596),
    [anon_sym_then] = ACTIONS(1594),
    [anon_sym_else] = ACTIONS(1594),
    [anon_sym_do] = ACTIONS(1594),
    [anon_sym_done] = ACTIONS(1594),
    [anon_sym_to] = ACTIONS(1594),
    [anon_sym_downto] = ACTIONS(1594),
    [anon_sym_new] = ACTIONS(1594),
    [anon_sym_LBRACE_LT] = ACTIONS(1596),
    [anon_sym_GT_RBRACE] = ACTIONS(1596),
    [anon_sym_begin] = ACTIONS(1594),
    [sym_ocamlyacc_value] = ACTIONS(1596),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1594),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1596),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1594),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1594),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1596),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1596),
    [aux_sym_number_token1] = ACTIONS(1596),
    [aux_sym_signed_number_token1] = ACTIONS(1594),
    [anon_sym_SQUOTE] = ACTIONS(1596),
    [anon_sym_DQUOTE] = ACTIONS(1596),
    [sym_prefix_operator] = ACTIONS(1596),
    [aux_sym_sign_operator_token1] = ACTIONS(1594),
    [sym_hash_operator] = ACTIONS(1596),
    [anon_sym_lsl] = ACTIONS(1594),
    [anon_sym_lsr] = ACTIONS(1594),
    [anon_sym_asr] = ACTIONS(1594),
    [aux_sym_pow_operator_token1] = ACTIONS(1596),
    [anon_sym_mod] = ACTIONS(1594),
    [anon_sym_land] = ACTIONS(1594),
    [anon_sym_lor] = ACTIONS(1594),
    [anon_sym_lxor] = ACTIONS(1594),
    [aux_sym_mult_operator_token1] = ACTIONS(1594),
    [aux_sym_add_operator_token1] = ACTIONS(1594),
    [sym_concat_operator] = ACTIONS(1596),
    [sym_rel_operator] = ACTIONS(1594),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_or] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1594),
    [sym_assign_operator] = ACTIONS(1596),
    [sym__uppercase_identifier] = ACTIONS(1596),
    [aux_sym_tag_token1] = ACTIONS(1596),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(781)] = {
    [sym_attribute] = STATE(781),
    [sym__lowercase_identifier] = ACTIONS(1598),
    [anon_sym_and] = ACTIONS(1598),
    [anon_sym_TILDE] = ACTIONS(1598),
    [anon_sym_QMARK] = ACTIONS(1598),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_RPAREN] = ACTIONS(1600),
    [anon_sym_COMMA] = ACTIONS(1600),
    [anon_sym_PIPE] = ACTIONS(1598),
    [anon_sym_COLON2] = ACTIONS(1598),
    [anon_sym_DOT] = ACTIONS(1600),
    [anon_sym_DASH_GT] = ACTIONS(1598),
    [anon_sym_LBRACK] = ACTIONS(1598),
    [anon_sym_RBRACK] = ACTIONS(1600),
    [anon_sym_true] = ACTIONS(1598),
    [anon_sym_false] = ACTIONS(1598),
    [anon_sym_LBRACE] = ACTIONS(1598),
    [anon_sym_SEMI] = ACTIONS(1600),
    [anon_sym_RBRACE] = ACTIONS(1600),
    [anon_sym_constraint] = ACTIONS(1598),
    [anon_sym_val] = ACTIONS(1598),
    [anon_sym_end] = ACTIONS(1598),
    [anon_sym_with] = ACTIONS(1598),
    [anon_sym_COLON_GT] = ACTIONS(1600),
    [anon_sym_object] = ACTIONS(1598),
    [anon_sym_inherit] = ACTIONS(1598),
    [anon_sym_method] = ACTIONS(1598),
    [anon_sym_in] = ACTIONS(1598),
    [anon_sym_as] = ACTIONS(1598),
    [anon_sym_initializer] = ACTIONS(1598),
    [anon_sym_AMP] = ACTIONS(1598),
    [anon_sym_POUND] = ACTIONS(1598),
    [anon_sym_COLON_COLON] = ACTIONS(1600),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1600),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1600),
    [anon_sym_then] = ACTIONS(1598),
    [anon_sym_else] = ACTIONS(1598),
    [anon_sym_do] = ACTIONS(1598),
    [anon_sym_done] = ACTIONS(1598),
    [anon_sym_to] = ACTIONS(1598),
    [anon_sym_downto] = ACTIONS(1598),
    [anon_sym_new] = ACTIONS(1598),
    [anon_sym_LBRACE_LT] = ACTIONS(1600),
    [anon_sym_GT_RBRACE] = ACTIONS(1600),
    [anon_sym_begin] = ACTIONS(1598),
    [sym_ocamlyacc_value] = ACTIONS(1600),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1598),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1600),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1598),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1598),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1600),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1600),
    [aux_sym_number_token1] = ACTIONS(1600),
    [aux_sym_signed_number_token1] = ACTIONS(1598),
    [anon_sym_SQUOTE] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1600),
    [sym_prefix_operator] = ACTIONS(1600),
    [aux_sym_sign_operator_token1] = ACTIONS(1598),
    [sym_hash_operator] = ACTIONS(1600),
    [anon_sym_lsl] = ACTIONS(1598),
    [anon_sym_lsr] = ACTIONS(1598),
    [anon_sym_asr] = ACTIONS(1598),
    [aux_sym_pow_operator_token1] = ACTIONS(1600),
    [anon_sym_mod] = ACTIONS(1598),
    [anon_sym_land] = ACTIONS(1598),
    [anon_sym_lor] = ACTIONS(1598),
    [anon_sym_lxor] = ACTIONS(1598),
    [aux_sym_mult_operator_token1] = ACTIONS(1598),
    [aux_sym_add_operator_token1] = ACTIONS(1598),
    [sym_concat_operator] = ACTIONS(1600),
    [sym_rel_operator] = ACTIONS(1598),
    [anon_sym_AMP_AMP] = ACTIONS(1598),
    [anon_sym_or] = ACTIONS(1598),
    [anon_sym_PIPE_PIPE] = ACTIONS(1598),
    [sym_assign_operator] = ACTIONS(1600),
    [sym__uppercase_identifier] = ACTIONS(1600),
    [aux_sym_tag_token1] = ACTIONS(1600),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(782)] = {
    [sym_attribute] = STATE(782),
    [sym__lowercase_identifier] = ACTIONS(1602),
    [anon_sym_and] = ACTIONS(1602),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_QMARK] = ACTIONS(1602),
    [anon_sym_LPAREN] = ACTIONS(1604),
    [anon_sym_RPAREN] = ACTIONS(1604),
    [anon_sym_COMMA] = ACTIONS(1604),
    [anon_sym_PIPE] = ACTIONS(1602),
    [anon_sym_COLON2] = ACTIONS(1602),
    [anon_sym_DOT] = ACTIONS(1604),
    [anon_sym_DASH_GT] = ACTIONS(1602),
    [anon_sym_LBRACK] = ACTIONS(1602),
    [anon_sym_RBRACK] = ACTIONS(1604),
    [anon_sym_true] = ACTIONS(1602),
    [anon_sym_false] = ACTIONS(1602),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [anon_sym_SEMI] = ACTIONS(1604),
    [anon_sym_RBRACE] = ACTIONS(1604),
    [anon_sym_constraint] = ACTIONS(1602),
    [anon_sym_val] = ACTIONS(1602),
    [anon_sym_end] = ACTIONS(1602),
    [anon_sym_with] = ACTIONS(1602),
    [anon_sym_COLON_GT] = ACTIONS(1604),
    [anon_sym_object] = ACTIONS(1602),
    [anon_sym_inherit] = ACTIONS(1602),
    [anon_sym_method] = ACTIONS(1602),
    [anon_sym_in] = ACTIONS(1602),
    [anon_sym_as] = ACTIONS(1602),
    [anon_sym_initializer] = ACTIONS(1602),
    [anon_sym_AMP] = ACTIONS(1602),
    [anon_sym_POUND] = ACTIONS(1602),
    [anon_sym_COLON_COLON] = ACTIONS(1604),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1604),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1604),
    [anon_sym_then] = ACTIONS(1602),
    [anon_sym_else] = ACTIONS(1602),
    [anon_sym_do] = ACTIONS(1602),
    [anon_sym_done] = ACTIONS(1602),
    [anon_sym_to] = ACTIONS(1602),
    [anon_sym_downto] = ACTIONS(1602),
    [anon_sym_new] = ACTIONS(1602),
    [anon_sym_LBRACE_LT] = ACTIONS(1604),
    [anon_sym_GT_RBRACE] = ACTIONS(1604),
    [anon_sym_begin] = ACTIONS(1602),
    [sym_ocamlyacc_value] = ACTIONS(1604),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1602),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1604),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1602),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1602),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1604),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1604),
    [aux_sym_number_token1] = ACTIONS(1604),
    [aux_sym_signed_number_token1] = ACTIONS(1602),
    [anon_sym_SQUOTE] = ACTIONS(1604),
    [anon_sym_DQUOTE] = ACTIONS(1604),
    [sym_prefix_operator] = ACTIONS(1604),
    [aux_sym_sign_operator_token1] = ACTIONS(1602),
    [sym_hash_operator] = ACTIONS(1604),
    [anon_sym_lsl] = ACTIONS(1602),
    [anon_sym_lsr] = ACTIONS(1602),
    [anon_sym_asr] = ACTIONS(1602),
    [aux_sym_pow_operator_token1] = ACTIONS(1604),
    [anon_sym_mod] = ACTIONS(1602),
    [anon_sym_land] = ACTIONS(1602),
    [anon_sym_lor] = ACTIONS(1602),
    [anon_sym_lxor] = ACTIONS(1602),
    [aux_sym_mult_operator_token1] = ACTIONS(1602),
    [aux_sym_add_operator_token1] = ACTIONS(1602),
    [sym_concat_operator] = ACTIONS(1604),
    [sym_rel_operator] = ACTIONS(1602),
    [anon_sym_AMP_AMP] = ACTIONS(1602),
    [anon_sym_or] = ACTIONS(1602),
    [anon_sym_PIPE_PIPE] = ACTIONS(1602),
    [sym_assign_operator] = ACTIONS(1604),
    [sym__uppercase_identifier] = ACTIONS(1604),
    [aux_sym_tag_token1] = ACTIONS(1604),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(783)] = {
    [sym_attribute] = STATE(783),
    [sym__lowercase_identifier] = ACTIONS(1606),
    [anon_sym_and] = ACTIONS(1606),
    [anon_sym_TILDE] = ACTIONS(1606),
    [anon_sym_QMARK] = ACTIONS(1606),
    [anon_sym_LPAREN] = ACTIONS(1608),
    [anon_sym_RPAREN] = ACTIONS(1608),
    [anon_sym_COMMA] = ACTIONS(1608),
    [anon_sym_PIPE] = ACTIONS(1606),
    [anon_sym_COLON2] = ACTIONS(1606),
    [anon_sym_DOT] = ACTIONS(1608),
    [anon_sym_DASH_GT] = ACTIONS(1606),
    [anon_sym_LBRACK] = ACTIONS(1606),
    [anon_sym_RBRACK] = ACTIONS(1608),
    [anon_sym_true] = ACTIONS(1606),
    [anon_sym_false] = ACTIONS(1606),
    [anon_sym_LBRACE] = ACTIONS(1606),
    [anon_sym_SEMI] = ACTIONS(1608),
    [anon_sym_RBRACE] = ACTIONS(1608),
    [anon_sym_constraint] = ACTIONS(1606),
    [anon_sym_val] = ACTIONS(1606),
    [anon_sym_end] = ACTIONS(1606),
    [anon_sym_with] = ACTIONS(1606),
    [anon_sym_COLON_GT] = ACTIONS(1608),
    [anon_sym_object] = ACTIONS(1606),
    [anon_sym_inherit] = ACTIONS(1606),
    [anon_sym_method] = ACTIONS(1606),
    [anon_sym_in] = ACTIONS(1606),
    [anon_sym_as] = ACTIONS(1606),
    [anon_sym_initializer] = ACTIONS(1606),
    [anon_sym_AMP] = ACTIONS(1606),
    [anon_sym_POUND] = ACTIONS(1606),
    [anon_sym_COLON_COLON] = ACTIONS(1608),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1608),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1608),
    [anon_sym_then] = ACTIONS(1606),
    [anon_sym_else] = ACTIONS(1606),
    [anon_sym_do] = ACTIONS(1606),
    [anon_sym_done] = ACTIONS(1606),
    [anon_sym_to] = ACTIONS(1606),
    [anon_sym_downto] = ACTIONS(1606),
    [anon_sym_new] = ACTIONS(1606),
    [anon_sym_LBRACE_LT] = ACTIONS(1608),
    [anon_sym_GT_RBRACE] = ACTIONS(1608),
    [anon_sym_begin] = ACTIONS(1606),
    [sym_ocamlyacc_value] = ACTIONS(1608),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1606),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1608),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1606),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1606),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1608),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1608),
    [aux_sym_number_token1] = ACTIONS(1608),
    [aux_sym_signed_number_token1] = ACTIONS(1606),
    [anon_sym_SQUOTE] = ACTIONS(1608),
    [anon_sym_DQUOTE] = ACTIONS(1608),
    [sym_prefix_operator] = ACTIONS(1608),
    [aux_sym_sign_operator_token1] = ACTIONS(1606),
    [sym_hash_operator] = ACTIONS(1608),
    [anon_sym_lsl] = ACTIONS(1606),
    [anon_sym_lsr] = ACTIONS(1606),
    [anon_sym_asr] = ACTIONS(1606),
    [aux_sym_pow_operator_token1] = ACTIONS(1608),
    [anon_sym_mod] = ACTIONS(1606),
    [anon_sym_land] = ACTIONS(1606),
    [anon_sym_lor] = ACTIONS(1606),
    [anon_sym_lxor] = ACTIONS(1606),
    [aux_sym_mult_operator_token1] = ACTIONS(1606),
    [aux_sym_add_operator_token1] = ACTIONS(1606),
    [sym_concat_operator] = ACTIONS(1608),
    [sym_rel_operator] = ACTIONS(1606),
    [anon_sym_AMP_AMP] = ACTIONS(1606),
    [anon_sym_or] = ACTIONS(1606),
    [anon_sym_PIPE_PIPE] = ACTIONS(1606),
    [sym_assign_operator] = ACTIONS(1608),
    [sym__uppercase_identifier] = ACTIONS(1608),
    [aux_sym_tag_token1] = ACTIONS(1608),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(784)] = {
    [sym_attribute] = STATE(784),
    [sym__lowercase_identifier] = ACTIONS(1610),
    [anon_sym_and] = ACTIONS(1610),
    [anon_sym_TILDE] = ACTIONS(1610),
    [anon_sym_QMARK] = ACTIONS(1610),
    [anon_sym_LPAREN] = ACTIONS(1612),
    [anon_sym_RPAREN] = ACTIONS(1612),
    [anon_sym_COMMA] = ACTIONS(1612),
    [anon_sym_PIPE] = ACTIONS(1610),
    [anon_sym_COLON2] = ACTIONS(1610),
    [anon_sym_DOT] = ACTIONS(1612),
    [anon_sym_DASH_GT] = ACTIONS(1610),
    [anon_sym_LBRACK] = ACTIONS(1610),
    [anon_sym_RBRACK] = ACTIONS(1612),
    [anon_sym_true] = ACTIONS(1610),
    [anon_sym_false] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1610),
    [anon_sym_SEMI] = ACTIONS(1612),
    [anon_sym_RBRACE] = ACTIONS(1612),
    [anon_sym_constraint] = ACTIONS(1610),
    [anon_sym_val] = ACTIONS(1610),
    [anon_sym_end] = ACTIONS(1610),
    [anon_sym_with] = ACTIONS(1610),
    [anon_sym_COLON_GT] = ACTIONS(1612),
    [anon_sym_object] = ACTIONS(1610),
    [anon_sym_inherit] = ACTIONS(1610),
    [anon_sym_method] = ACTIONS(1610),
    [anon_sym_in] = ACTIONS(1610),
    [anon_sym_as] = ACTIONS(1610),
    [anon_sym_initializer] = ACTIONS(1610),
    [anon_sym_AMP] = ACTIONS(1610),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_COLON_COLON] = ACTIONS(1612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1612),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1612),
    [anon_sym_then] = ACTIONS(1610),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_do] = ACTIONS(1610),
    [anon_sym_done] = ACTIONS(1610),
    [anon_sym_to] = ACTIONS(1610),
    [anon_sym_downto] = ACTIONS(1610),
    [anon_sym_new] = ACTIONS(1610),
    [anon_sym_LBRACE_LT] = ACTIONS(1612),
    [anon_sym_GT_RBRACE] = ACTIONS(1612),
    [anon_sym_begin] = ACTIONS(1610),
    [sym_ocamlyacc_value] = ACTIONS(1612),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1610),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1612),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1610),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1610),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1612),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1612),
    [aux_sym_number_token1] = ACTIONS(1612),
    [aux_sym_signed_number_token1] = ACTIONS(1610),
    [anon_sym_SQUOTE] = ACTIONS(1612),
    [anon_sym_DQUOTE] = ACTIONS(1612),
    [sym_prefix_operator] = ACTIONS(1612),
    [aux_sym_sign_operator_token1] = ACTIONS(1610),
    [sym_hash_operator] = ACTIONS(1612),
    [anon_sym_lsl] = ACTIONS(1610),
    [anon_sym_lsr] = ACTIONS(1610),
    [anon_sym_asr] = ACTIONS(1610),
    [aux_sym_pow_operator_token1] = ACTIONS(1612),
    [anon_sym_mod] = ACTIONS(1610),
    [anon_sym_land] = ACTIONS(1610),
    [anon_sym_lor] = ACTIONS(1610),
    [anon_sym_lxor] = ACTIONS(1610),
    [aux_sym_mult_operator_token1] = ACTIONS(1610),
    [aux_sym_add_operator_token1] = ACTIONS(1610),
    [sym_concat_operator] = ACTIONS(1612),
    [sym_rel_operator] = ACTIONS(1610),
    [anon_sym_AMP_AMP] = ACTIONS(1610),
    [anon_sym_or] = ACTIONS(1610),
    [anon_sym_PIPE_PIPE] = ACTIONS(1610),
    [sym_assign_operator] = ACTIONS(1612),
    [sym__uppercase_identifier] = ACTIONS(1612),
    [aux_sym_tag_token1] = ACTIONS(1612),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(785)] = {
    [sym_attribute] = STATE(785),
    [sym__lowercase_identifier] = ACTIONS(1614),
    [anon_sym_and] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_RPAREN] = ACTIONS(1616),
    [anon_sym_COMMA] = ACTIONS(1616),
    [anon_sym_PIPE] = ACTIONS(1614),
    [anon_sym_COLON2] = ACTIONS(1614),
    [anon_sym_DOT] = ACTIONS(1616),
    [anon_sym_DASH_GT] = ACTIONS(1614),
    [anon_sym_LBRACK] = ACTIONS(1614),
    [anon_sym_RBRACK] = ACTIONS(1616),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1616),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [anon_sym_constraint] = ACTIONS(1614),
    [anon_sym_val] = ACTIONS(1614),
    [anon_sym_end] = ACTIONS(1614),
    [anon_sym_with] = ACTIONS(1614),
    [anon_sym_COLON_GT] = ACTIONS(1616),
    [anon_sym_object] = ACTIONS(1614),
    [anon_sym_inherit] = ACTIONS(1614),
    [anon_sym_method] = ACTIONS(1614),
    [anon_sym_in] = ACTIONS(1614),
    [anon_sym_as] = ACTIONS(1614),
    [anon_sym_initializer] = ACTIONS(1614),
    [anon_sym_AMP] = ACTIONS(1614),
    [anon_sym_POUND] = ACTIONS(1614),
    [anon_sym_COLON_COLON] = ACTIONS(1616),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1616),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1616),
    [anon_sym_then] = ACTIONS(1614),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_do] = ACTIONS(1614),
    [anon_sym_done] = ACTIONS(1614),
    [anon_sym_to] = ACTIONS(1614),
    [anon_sym_downto] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
    [anon_sym_LBRACE_LT] = ACTIONS(1616),
    [anon_sym_GT_RBRACE] = ACTIONS(1616),
    [anon_sym_begin] = ACTIONS(1614),
    [sym_ocamlyacc_value] = ACTIONS(1616),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1614),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1616),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1616),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1616),
    [aux_sym_number_token1] = ACTIONS(1616),
    [aux_sym_signed_number_token1] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1616),
    [sym_prefix_operator] = ACTIONS(1616),
    [aux_sym_sign_operator_token1] = ACTIONS(1614),
    [sym_hash_operator] = ACTIONS(1616),
    [anon_sym_lsl] = ACTIONS(1614),
    [anon_sym_lsr] = ACTIONS(1614),
    [anon_sym_asr] = ACTIONS(1614),
    [aux_sym_pow_operator_token1] = ACTIONS(1616),
    [anon_sym_mod] = ACTIONS(1614),
    [anon_sym_land] = ACTIONS(1614),
    [anon_sym_lor] = ACTIONS(1614),
    [anon_sym_lxor] = ACTIONS(1614),
    [aux_sym_mult_operator_token1] = ACTIONS(1614),
    [aux_sym_add_operator_token1] = ACTIONS(1614),
    [sym_concat_operator] = ACTIONS(1616),
    [sym_rel_operator] = ACTIONS(1614),
    [anon_sym_AMP_AMP] = ACTIONS(1614),
    [anon_sym_or] = ACTIONS(1614),
    [anon_sym_PIPE_PIPE] = ACTIONS(1614),
    [sym_assign_operator] = ACTIONS(1616),
    [sym__uppercase_identifier] = ACTIONS(1616),
    [aux_sym_tag_token1] = ACTIONS(1616),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(786)] = {
    [sym_attribute] = STATE(786),
    [sym__lowercase_identifier] = ACTIONS(1618),
    [anon_sym_and] = ACTIONS(1618),
    [anon_sym_TILDE] = ACTIONS(1618),
    [anon_sym_QMARK] = ACTIONS(1618),
    [anon_sym_LPAREN] = ACTIONS(1620),
    [anon_sym_RPAREN] = ACTIONS(1620),
    [anon_sym_COMMA] = ACTIONS(1620),
    [anon_sym_PIPE] = ACTIONS(1618),
    [anon_sym_COLON2] = ACTIONS(1618),
    [anon_sym_DOT] = ACTIONS(1620),
    [anon_sym_DASH_GT] = ACTIONS(1618),
    [anon_sym_LBRACK] = ACTIONS(1618),
    [anon_sym_RBRACK] = ACTIONS(1620),
    [anon_sym_true] = ACTIONS(1618),
    [anon_sym_false] = ACTIONS(1618),
    [anon_sym_LBRACE] = ACTIONS(1618),
    [anon_sym_SEMI] = ACTIONS(1620),
    [anon_sym_RBRACE] = ACTIONS(1620),
    [anon_sym_constraint] = ACTIONS(1618),
    [anon_sym_val] = ACTIONS(1618),
    [anon_sym_end] = ACTIONS(1618),
    [anon_sym_with] = ACTIONS(1618),
    [anon_sym_COLON_GT] = ACTIONS(1620),
    [anon_sym_object] = ACTIONS(1618),
    [anon_sym_inherit] = ACTIONS(1618),
    [anon_sym_method] = ACTIONS(1618),
    [anon_sym_in] = ACTIONS(1618),
    [anon_sym_as] = ACTIONS(1618),
    [anon_sym_initializer] = ACTIONS(1618),
    [anon_sym_AMP] = ACTIONS(1618),
    [anon_sym_POUND] = ACTIONS(1618),
    [anon_sym_COLON_COLON] = ACTIONS(1620),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1620),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1620),
    [anon_sym_then] = ACTIONS(1618),
    [anon_sym_else] = ACTIONS(1618),
    [anon_sym_do] = ACTIONS(1618),
    [anon_sym_done] = ACTIONS(1618),
    [anon_sym_to] = ACTIONS(1618),
    [anon_sym_downto] = ACTIONS(1618),
    [anon_sym_new] = ACTIONS(1618),
    [anon_sym_LBRACE_LT] = ACTIONS(1620),
    [anon_sym_GT_RBRACE] = ACTIONS(1620),
    [anon_sym_begin] = ACTIONS(1618),
    [sym_ocamlyacc_value] = ACTIONS(1620),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1618),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1620),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1618),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1618),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1620),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1620),
    [aux_sym_number_token1] = ACTIONS(1620),
    [aux_sym_signed_number_token1] = ACTIONS(1618),
    [anon_sym_SQUOTE] = ACTIONS(1620),
    [anon_sym_DQUOTE] = ACTIONS(1620),
    [sym_prefix_operator] = ACTIONS(1620),
    [aux_sym_sign_operator_token1] = ACTIONS(1618),
    [sym_hash_operator] = ACTIONS(1620),
    [anon_sym_lsl] = ACTIONS(1618),
    [anon_sym_lsr] = ACTIONS(1618),
    [anon_sym_asr] = ACTIONS(1618),
    [aux_sym_pow_operator_token1] = ACTIONS(1620),
    [anon_sym_mod] = ACTIONS(1618),
    [anon_sym_land] = ACTIONS(1618),
    [anon_sym_lor] = ACTIONS(1618),
    [anon_sym_lxor] = ACTIONS(1618),
    [aux_sym_mult_operator_token1] = ACTIONS(1618),
    [aux_sym_add_operator_token1] = ACTIONS(1618),
    [sym_concat_operator] = ACTIONS(1620),
    [sym_rel_operator] = ACTIONS(1618),
    [anon_sym_AMP_AMP] = ACTIONS(1618),
    [anon_sym_or] = ACTIONS(1618),
    [anon_sym_PIPE_PIPE] = ACTIONS(1618),
    [sym_assign_operator] = ACTIONS(1620),
    [sym__uppercase_identifier] = ACTIONS(1620),
    [aux_sym_tag_token1] = ACTIONS(1620),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(787)] = {
    [sym_attribute] = STATE(787),
    [sym__lowercase_identifier] = ACTIONS(1622),
    [anon_sym_and] = ACTIONS(1622),
    [anon_sym_TILDE] = ACTIONS(1622),
    [anon_sym_QMARK] = ACTIONS(1622),
    [anon_sym_LPAREN] = ACTIONS(1624),
    [anon_sym_RPAREN] = ACTIONS(1624),
    [anon_sym_COMMA] = ACTIONS(1624),
    [anon_sym_PIPE] = ACTIONS(1622),
    [anon_sym_COLON2] = ACTIONS(1622),
    [anon_sym_DOT] = ACTIONS(1624),
    [anon_sym_DASH_GT] = ACTIONS(1622),
    [anon_sym_LBRACK] = ACTIONS(1622),
    [anon_sym_RBRACK] = ACTIONS(1624),
    [anon_sym_true] = ACTIONS(1622),
    [anon_sym_false] = ACTIONS(1622),
    [anon_sym_LBRACE] = ACTIONS(1622),
    [anon_sym_SEMI] = ACTIONS(1624),
    [anon_sym_RBRACE] = ACTIONS(1624),
    [anon_sym_constraint] = ACTIONS(1622),
    [anon_sym_val] = ACTIONS(1622),
    [anon_sym_end] = ACTIONS(1622),
    [anon_sym_with] = ACTIONS(1622),
    [anon_sym_COLON_GT] = ACTIONS(1624),
    [anon_sym_object] = ACTIONS(1622),
    [anon_sym_inherit] = ACTIONS(1622),
    [anon_sym_method] = ACTIONS(1622),
    [anon_sym_in] = ACTIONS(1622),
    [anon_sym_as] = ACTIONS(1622),
    [anon_sym_initializer] = ACTIONS(1622),
    [anon_sym_AMP] = ACTIONS(1622),
    [anon_sym_POUND] = ACTIONS(1622),
    [anon_sym_COLON_COLON] = ACTIONS(1624),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1624),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1624),
    [anon_sym_then] = ACTIONS(1622),
    [anon_sym_else] = ACTIONS(1622),
    [anon_sym_do] = ACTIONS(1622),
    [anon_sym_done] = ACTIONS(1622),
    [anon_sym_to] = ACTIONS(1622),
    [anon_sym_downto] = ACTIONS(1622),
    [anon_sym_new] = ACTIONS(1622),
    [anon_sym_LBRACE_LT] = ACTIONS(1624),
    [anon_sym_GT_RBRACE] = ACTIONS(1624),
    [anon_sym_begin] = ACTIONS(1622),
    [sym_ocamlyacc_value] = ACTIONS(1624),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1622),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1624),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1622),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1622),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1624),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1624),
    [aux_sym_number_token1] = ACTIONS(1624),
    [aux_sym_signed_number_token1] = ACTIONS(1622),
    [anon_sym_SQUOTE] = ACTIONS(1624),
    [anon_sym_DQUOTE] = ACTIONS(1624),
    [sym_prefix_operator] = ACTIONS(1624),
    [aux_sym_sign_operator_token1] = ACTIONS(1622),
    [sym_hash_operator] = ACTIONS(1624),
    [anon_sym_lsl] = ACTIONS(1622),
    [anon_sym_lsr] = ACTIONS(1622),
    [anon_sym_asr] = ACTIONS(1622),
    [aux_sym_pow_operator_token1] = ACTIONS(1624),
    [anon_sym_mod] = ACTIONS(1622),
    [anon_sym_land] = ACTIONS(1622),
    [anon_sym_lor] = ACTIONS(1622),
    [anon_sym_lxor] = ACTIONS(1622),
    [aux_sym_mult_operator_token1] = ACTIONS(1622),
    [aux_sym_add_operator_token1] = ACTIONS(1622),
    [sym_concat_operator] = ACTIONS(1624),
    [sym_rel_operator] = ACTIONS(1622),
    [anon_sym_AMP_AMP] = ACTIONS(1622),
    [anon_sym_or] = ACTIONS(1622),
    [anon_sym_PIPE_PIPE] = ACTIONS(1622),
    [sym_assign_operator] = ACTIONS(1624),
    [sym__uppercase_identifier] = ACTIONS(1624),
    [aux_sym_tag_token1] = ACTIONS(1624),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(788)] = {
    [sym_attribute] = STATE(788),
    [sym__lowercase_identifier] = ACTIONS(1626),
    [anon_sym_and] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1628),
    [anon_sym_RPAREN] = ACTIONS(1628),
    [anon_sym_COMMA] = ACTIONS(1628),
    [anon_sym_PIPE] = ACTIONS(1626),
    [anon_sym_COLON2] = ACTIONS(1626),
    [anon_sym_DOT] = ACTIONS(1628),
    [anon_sym_DASH_GT] = ACTIONS(1626),
    [anon_sym_LBRACK] = ACTIONS(1626),
    [anon_sym_RBRACK] = ACTIONS(1628),
    [anon_sym_true] = ACTIONS(1626),
    [anon_sym_false] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1628),
    [anon_sym_RBRACE] = ACTIONS(1628),
    [anon_sym_constraint] = ACTIONS(1626),
    [anon_sym_val] = ACTIONS(1626),
    [anon_sym_end] = ACTIONS(1626),
    [anon_sym_with] = ACTIONS(1626),
    [anon_sym_COLON_GT] = ACTIONS(1628),
    [anon_sym_object] = ACTIONS(1626),
    [anon_sym_inherit] = ACTIONS(1626),
    [anon_sym_method] = ACTIONS(1626),
    [anon_sym_in] = ACTIONS(1626),
    [anon_sym_as] = ACTIONS(1626),
    [anon_sym_initializer] = ACTIONS(1626),
    [anon_sym_AMP] = ACTIONS(1626),
    [anon_sym_POUND] = ACTIONS(1626),
    [anon_sym_COLON_COLON] = ACTIONS(1628),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1628),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1628),
    [anon_sym_then] = ACTIONS(1626),
    [anon_sym_else] = ACTIONS(1626),
    [anon_sym_do] = ACTIONS(1626),
    [anon_sym_done] = ACTIONS(1626),
    [anon_sym_to] = ACTIONS(1626),
    [anon_sym_downto] = ACTIONS(1626),
    [anon_sym_new] = ACTIONS(1626),
    [anon_sym_LBRACE_LT] = ACTIONS(1628),
    [anon_sym_GT_RBRACE] = ACTIONS(1628),
    [anon_sym_begin] = ACTIONS(1626),
    [sym_ocamlyacc_value] = ACTIONS(1628),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1626),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1628),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1626),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1626),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1628),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1628),
    [aux_sym_number_token1] = ACTIONS(1628),
    [aux_sym_signed_number_token1] = ACTIONS(1626),
    [anon_sym_SQUOTE] = ACTIONS(1628),
    [anon_sym_DQUOTE] = ACTIONS(1628),
    [sym_prefix_operator] = ACTIONS(1628),
    [aux_sym_sign_operator_token1] = ACTIONS(1626),
    [sym_hash_operator] = ACTIONS(1628),
    [anon_sym_lsl] = ACTIONS(1626),
    [anon_sym_lsr] = ACTIONS(1626),
    [anon_sym_asr] = ACTIONS(1626),
    [aux_sym_pow_operator_token1] = ACTIONS(1628),
    [anon_sym_mod] = ACTIONS(1626),
    [anon_sym_land] = ACTIONS(1626),
    [anon_sym_lor] = ACTIONS(1626),
    [anon_sym_lxor] = ACTIONS(1626),
    [aux_sym_mult_operator_token1] = ACTIONS(1626),
    [aux_sym_add_operator_token1] = ACTIONS(1626),
    [sym_concat_operator] = ACTIONS(1628),
    [sym_rel_operator] = ACTIONS(1626),
    [anon_sym_AMP_AMP] = ACTIONS(1626),
    [anon_sym_or] = ACTIONS(1626),
    [anon_sym_PIPE_PIPE] = ACTIONS(1626),
    [sym_assign_operator] = ACTIONS(1628),
    [sym__uppercase_identifier] = ACTIONS(1628),
    [aux_sym_tag_token1] = ACTIONS(1628),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(789)] = {
    [sym_attribute] = STATE(789),
    [sym__lowercase_identifier] = ACTIONS(1630),
    [anon_sym_and] = ACTIONS(1630),
    [anon_sym_TILDE] = ACTIONS(1630),
    [anon_sym_QMARK] = ACTIONS(1630),
    [anon_sym_LPAREN] = ACTIONS(1632),
    [anon_sym_RPAREN] = ACTIONS(1632),
    [anon_sym_COMMA] = ACTIONS(1632),
    [anon_sym_PIPE] = ACTIONS(1630),
    [anon_sym_COLON2] = ACTIONS(1630),
    [anon_sym_DOT] = ACTIONS(1632),
    [anon_sym_DASH_GT] = ACTIONS(1630),
    [anon_sym_LBRACK] = ACTIONS(1630),
    [anon_sym_RBRACK] = ACTIONS(1632),
    [anon_sym_true] = ACTIONS(1630),
    [anon_sym_false] = ACTIONS(1630),
    [anon_sym_LBRACE] = ACTIONS(1630),
    [anon_sym_SEMI] = ACTIONS(1632),
    [anon_sym_RBRACE] = ACTIONS(1632),
    [anon_sym_constraint] = ACTIONS(1630),
    [anon_sym_val] = ACTIONS(1630),
    [anon_sym_end] = ACTIONS(1630),
    [anon_sym_with] = ACTIONS(1630),
    [anon_sym_COLON_GT] = ACTIONS(1632),
    [anon_sym_object] = ACTIONS(1630),
    [anon_sym_inherit] = ACTIONS(1630),
    [anon_sym_method] = ACTIONS(1630),
    [anon_sym_in] = ACTIONS(1630),
    [anon_sym_as] = ACTIONS(1630),
    [anon_sym_initializer] = ACTIONS(1630),
    [anon_sym_AMP] = ACTIONS(1630),
    [anon_sym_POUND] = ACTIONS(1630),
    [anon_sym_COLON_COLON] = ACTIONS(1632),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1632),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1632),
    [anon_sym_then] = ACTIONS(1630),
    [anon_sym_else] = ACTIONS(1630),
    [anon_sym_do] = ACTIONS(1630),
    [anon_sym_done] = ACTIONS(1630),
    [anon_sym_to] = ACTIONS(1630),
    [anon_sym_downto] = ACTIONS(1630),
    [anon_sym_new] = ACTIONS(1630),
    [anon_sym_LBRACE_LT] = ACTIONS(1632),
    [anon_sym_GT_RBRACE] = ACTIONS(1632),
    [anon_sym_begin] = ACTIONS(1630),
    [sym_ocamlyacc_value] = ACTIONS(1632),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1630),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1632),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1630),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1630),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1632),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1632),
    [aux_sym_number_token1] = ACTIONS(1632),
    [aux_sym_signed_number_token1] = ACTIONS(1630),
    [anon_sym_SQUOTE] = ACTIONS(1632),
    [anon_sym_DQUOTE] = ACTIONS(1632),
    [sym_prefix_operator] = ACTIONS(1632),
    [aux_sym_sign_operator_token1] = ACTIONS(1630),
    [sym_hash_operator] = ACTIONS(1632),
    [anon_sym_lsl] = ACTIONS(1630),
    [anon_sym_lsr] = ACTIONS(1630),
    [anon_sym_asr] = ACTIONS(1630),
    [aux_sym_pow_operator_token1] = ACTIONS(1632),
    [anon_sym_mod] = ACTIONS(1630),
    [anon_sym_land] = ACTIONS(1630),
    [anon_sym_lor] = ACTIONS(1630),
    [anon_sym_lxor] = ACTIONS(1630),
    [aux_sym_mult_operator_token1] = ACTIONS(1630),
    [aux_sym_add_operator_token1] = ACTIONS(1630),
    [sym_concat_operator] = ACTIONS(1632),
    [sym_rel_operator] = ACTIONS(1630),
    [anon_sym_AMP_AMP] = ACTIONS(1630),
    [anon_sym_or] = ACTIONS(1630),
    [anon_sym_PIPE_PIPE] = ACTIONS(1630),
    [sym_assign_operator] = ACTIONS(1632),
    [sym__uppercase_identifier] = ACTIONS(1632),
    [aux_sym_tag_token1] = ACTIONS(1632),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(790)] = {
    [sym_attribute] = STATE(790),
    [sym__lowercase_identifier] = ACTIONS(1634),
    [anon_sym_and] = ACTIONS(1634),
    [anon_sym_TILDE] = ACTIONS(1634),
    [anon_sym_QMARK] = ACTIONS(1634),
    [anon_sym_LPAREN] = ACTIONS(1636),
    [anon_sym_RPAREN] = ACTIONS(1636),
    [anon_sym_COMMA] = ACTIONS(1636),
    [anon_sym_PIPE] = ACTIONS(1634),
    [anon_sym_COLON2] = ACTIONS(1634),
    [anon_sym_DOT] = ACTIONS(1636),
    [anon_sym_DASH_GT] = ACTIONS(1634),
    [anon_sym_LBRACK] = ACTIONS(1634),
    [anon_sym_RBRACK] = ACTIONS(1636),
    [anon_sym_true] = ACTIONS(1634),
    [anon_sym_false] = ACTIONS(1634),
    [anon_sym_LBRACE] = ACTIONS(1634),
    [anon_sym_SEMI] = ACTIONS(1636),
    [anon_sym_RBRACE] = ACTIONS(1636),
    [anon_sym_constraint] = ACTIONS(1634),
    [anon_sym_val] = ACTIONS(1634),
    [anon_sym_end] = ACTIONS(1634),
    [anon_sym_with] = ACTIONS(1634),
    [anon_sym_COLON_GT] = ACTIONS(1636),
    [anon_sym_object] = ACTIONS(1634),
    [anon_sym_inherit] = ACTIONS(1634),
    [anon_sym_method] = ACTIONS(1634),
    [anon_sym_in] = ACTIONS(1634),
    [anon_sym_as] = ACTIONS(1634),
    [anon_sym_initializer] = ACTIONS(1634),
    [anon_sym_AMP] = ACTIONS(1634),
    [anon_sym_POUND] = ACTIONS(1634),
    [anon_sym_COLON_COLON] = ACTIONS(1636),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1636),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1636),
    [anon_sym_then] = ACTIONS(1634),
    [anon_sym_else] = ACTIONS(1634),
    [anon_sym_do] = ACTIONS(1634),
    [anon_sym_done] = ACTIONS(1634),
    [anon_sym_to] = ACTIONS(1634),
    [anon_sym_downto] = ACTIONS(1634),
    [anon_sym_new] = ACTIONS(1634),
    [anon_sym_LBRACE_LT] = ACTIONS(1636),
    [anon_sym_GT_RBRACE] = ACTIONS(1636),
    [anon_sym_begin] = ACTIONS(1634),
    [sym_ocamlyacc_value] = ACTIONS(1636),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1634),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1636),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1634),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1634),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1636),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1636),
    [aux_sym_number_token1] = ACTIONS(1636),
    [aux_sym_signed_number_token1] = ACTIONS(1634),
    [anon_sym_SQUOTE] = ACTIONS(1636),
    [anon_sym_DQUOTE] = ACTIONS(1636),
    [sym_prefix_operator] = ACTIONS(1636),
    [aux_sym_sign_operator_token1] = ACTIONS(1634),
    [sym_hash_operator] = ACTIONS(1636),
    [anon_sym_lsl] = ACTIONS(1634),
    [anon_sym_lsr] = ACTIONS(1634),
    [anon_sym_asr] = ACTIONS(1634),
    [aux_sym_pow_operator_token1] = ACTIONS(1636),
    [anon_sym_mod] = ACTIONS(1634),
    [anon_sym_land] = ACTIONS(1634),
    [anon_sym_lor] = ACTIONS(1634),
    [anon_sym_lxor] = ACTIONS(1634),
    [aux_sym_mult_operator_token1] = ACTIONS(1634),
    [aux_sym_add_operator_token1] = ACTIONS(1634),
    [sym_concat_operator] = ACTIONS(1636),
    [sym_rel_operator] = ACTIONS(1634),
    [anon_sym_AMP_AMP] = ACTIONS(1634),
    [anon_sym_or] = ACTIONS(1634),
    [anon_sym_PIPE_PIPE] = ACTIONS(1634),
    [sym_assign_operator] = ACTIONS(1636),
    [sym__uppercase_identifier] = ACTIONS(1636),
    [aux_sym_tag_token1] = ACTIONS(1636),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(791)] = {
    [sym_attribute] = STATE(791),
    [sym__lowercase_identifier] = ACTIONS(1638),
    [anon_sym_and] = ACTIONS(1638),
    [anon_sym_TILDE] = ACTIONS(1638),
    [anon_sym_QMARK] = ACTIONS(1638),
    [anon_sym_LPAREN] = ACTIONS(1640),
    [anon_sym_RPAREN] = ACTIONS(1640),
    [anon_sym_COMMA] = ACTIONS(1640),
    [anon_sym_PIPE] = ACTIONS(1638),
    [anon_sym_COLON2] = ACTIONS(1638),
    [anon_sym_DOT] = ACTIONS(1640),
    [anon_sym_DASH_GT] = ACTIONS(1638),
    [anon_sym_LBRACK] = ACTIONS(1638),
    [anon_sym_RBRACK] = ACTIONS(1640),
    [anon_sym_true] = ACTIONS(1638),
    [anon_sym_false] = ACTIONS(1638),
    [anon_sym_LBRACE] = ACTIONS(1638),
    [anon_sym_SEMI] = ACTIONS(1640),
    [anon_sym_RBRACE] = ACTIONS(1640),
    [anon_sym_constraint] = ACTIONS(1638),
    [anon_sym_val] = ACTIONS(1638),
    [anon_sym_end] = ACTIONS(1638),
    [anon_sym_with] = ACTIONS(1638),
    [anon_sym_COLON_GT] = ACTIONS(1640),
    [anon_sym_object] = ACTIONS(1638),
    [anon_sym_inherit] = ACTIONS(1638),
    [anon_sym_method] = ACTIONS(1638),
    [anon_sym_in] = ACTIONS(1638),
    [anon_sym_as] = ACTIONS(1638),
    [anon_sym_initializer] = ACTIONS(1638),
    [anon_sym_AMP] = ACTIONS(1638),
    [anon_sym_POUND] = ACTIONS(1638),
    [anon_sym_COLON_COLON] = ACTIONS(1640),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1640),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1640),
    [anon_sym_then] = ACTIONS(1638),
    [anon_sym_else] = ACTIONS(1638),
    [anon_sym_do] = ACTIONS(1638),
    [anon_sym_done] = ACTIONS(1638),
    [anon_sym_to] = ACTIONS(1638),
    [anon_sym_downto] = ACTIONS(1638),
    [anon_sym_new] = ACTIONS(1638),
    [anon_sym_LBRACE_LT] = ACTIONS(1640),
    [anon_sym_GT_RBRACE] = ACTIONS(1640),
    [anon_sym_begin] = ACTIONS(1638),
    [sym_ocamlyacc_value] = ACTIONS(1640),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1638),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1640),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1638),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1638),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1640),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1640),
    [aux_sym_number_token1] = ACTIONS(1640),
    [aux_sym_signed_number_token1] = ACTIONS(1638),
    [anon_sym_SQUOTE] = ACTIONS(1640),
    [anon_sym_DQUOTE] = ACTIONS(1640),
    [sym_prefix_operator] = ACTIONS(1640),
    [aux_sym_sign_operator_token1] = ACTIONS(1638),
    [sym_hash_operator] = ACTIONS(1640),
    [anon_sym_lsl] = ACTIONS(1638),
    [anon_sym_lsr] = ACTIONS(1638),
    [anon_sym_asr] = ACTIONS(1638),
    [aux_sym_pow_operator_token1] = ACTIONS(1640),
    [anon_sym_mod] = ACTIONS(1638),
    [anon_sym_land] = ACTIONS(1638),
    [anon_sym_lor] = ACTIONS(1638),
    [anon_sym_lxor] = ACTIONS(1638),
    [aux_sym_mult_operator_token1] = ACTIONS(1638),
    [aux_sym_add_operator_token1] = ACTIONS(1638),
    [sym_concat_operator] = ACTIONS(1640),
    [sym_rel_operator] = ACTIONS(1638),
    [anon_sym_AMP_AMP] = ACTIONS(1638),
    [anon_sym_or] = ACTIONS(1638),
    [anon_sym_PIPE_PIPE] = ACTIONS(1638),
    [sym_assign_operator] = ACTIONS(1640),
    [sym__uppercase_identifier] = ACTIONS(1640),
    [aux_sym_tag_token1] = ACTIONS(1640),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(792)] = {
    [sym_attribute] = STATE(792),
    [sym__lowercase_identifier] = ACTIONS(1642),
    [anon_sym_and] = ACTIONS(1642),
    [anon_sym_TILDE] = ACTIONS(1642),
    [anon_sym_QMARK] = ACTIONS(1642),
    [anon_sym_LPAREN] = ACTIONS(1644),
    [anon_sym_RPAREN] = ACTIONS(1644),
    [anon_sym_COMMA] = ACTIONS(1644),
    [anon_sym_PIPE] = ACTIONS(1642),
    [anon_sym_COLON2] = ACTIONS(1642),
    [anon_sym_DOT] = ACTIONS(1644),
    [anon_sym_DASH_GT] = ACTIONS(1642),
    [anon_sym_LBRACK] = ACTIONS(1642),
    [anon_sym_RBRACK] = ACTIONS(1644),
    [anon_sym_true] = ACTIONS(1642),
    [anon_sym_false] = ACTIONS(1642),
    [anon_sym_LBRACE] = ACTIONS(1642),
    [anon_sym_SEMI] = ACTIONS(1644),
    [anon_sym_RBRACE] = ACTIONS(1644),
    [anon_sym_constraint] = ACTIONS(1642),
    [anon_sym_val] = ACTIONS(1642),
    [anon_sym_end] = ACTIONS(1642),
    [anon_sym_with] = ACTIONS(1642),
    [anon_sym_COLON_GT] = ACTIONS(1644),
    [anon_sym_object] = ACTIONS(1642),
    [anon_sym_inherit] = ACTIONS(1642),
    [anon_sym_method] = ACTIONS(1642),
    [anon_sym_in] = ACTIONS(1642),
    [anon_sym_as] = ACTIONS(1642),
    [anon_sym_initializer] = ACTIONS(1642),
    [anon_sym_AMP] = ACTIONS(1642),
    [anon_sym_POUND] = ACTIONS(1642),
    [anon_sym_COLON_COLON] = ACTIONS(1644),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1644),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1644),
    [anon_sym_then] = ACTIONS(1642),
    [anon_sym_else] = ACTIONS(1642),
    [anon_sym_do] = ACTIONS(1642),
    [anon_sym_done] = ACTIONS(1642),
    [anon_sym_to] = ACTIONS(1642),
    [anon_sym_downto] = ACTIONS(1642),
    [anon_sym_new] = ACTIONS(1642),
    [anon_sym_LBRACE_LT] = ACTIONS(1644),
    [anon_sym_GT_RBRACE] = ACTIONS(1644),
    [anon_sym_begin] = ACTIONS(1642),
    [sym_ocamlyacc_value] = ACTIONS(1644),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1642),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1644),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1642),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1642),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1644),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1644),
    [aux_sym_number_token1] = ACTIONS(1644),
    [aux_sym_signed_number_token1] = ACTIONS(1642),
    [anon_sym_SQUOTE] = ACTIONS(1644),
    [anon_sym_DQUOTE] = ACTIONS(1644),
    [sym_prefix_operator] = ACTIONS(1644),
    [aux_sym_sign_operator_token1] = ACTIONS(1642),
    [sym_hash_operator] = ACTIONS(1644),
    [anon_sym_lsl] = ACTIONS(1642),
    [anon_sym_lsr] = ACTIONS(1642),
    [anon_sym_asr] = ACTIONS(1642),
    [aux_sym_pow_operator_token1] = ACTIONS(1644),
    [anon_sym_mod] = ACTIONS(1642),
    [anon_sym_land] = ACTIONS(1642),
    [anon_sym_lor] = ACTIONS(1642),
    [anon_sym_lxor] = ACTIONS(1642),
    [aux_sym_mult_operator_token1] = ACTIONS(1642),
    [aux_sym_add_operator_token1] = ACTIONS(1642),
    [sym_concat_operator] = ACTIONS(1644),
    [sym_rel_operator] = ACTIONS(1642),
    [anon_sym_AMP_AMP] = ACTIONS(1642),
    [anon_sym_or] = ACTIONS(1642),
    [anon_sym_PIPE_PIPE] = ACTIONS(1642),
    [sym_assign_operator] = ACTIONS(1644),
    [sym__uppercase_identifier] = ACTIONS(1644),
    [aux_sym_tag_token1] = ACTIONS(1644),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(793)] = {
    [sym_attribute] = STATE(793),
    [sym__lowercase_identifier] = ACTIONS(1646),
    [anon_sym_and] = ACTIONS(1646),
    [anon_sym_TILDE] = ACTIONS(1646),
    [anon_sym_QMARK] = ACTIONS(1646),
    [anon_sym_LPAREN] = ACTIONS(1648),
    [anon_sym_RPAREN] = ACTIONS(1648),
    [anon_sym_COMMA] = ACTIONS(1648),
    [anon_sym_PIPE] = ACTIONS(1646),
    [anon_sym_COLON2] = ACTIONS(1646),
    [anon_sym_DOT] = ACTIONS(1648),
    [anon_sym_DASH_GT] = ACTIONS(1646),
    [anon_sym_LBRACK] = ACTIONS(1646),
    [anon_sym_RBRACK] = ACTIONS(1648),
    [anon_sym_true] = ACTIONS(1646),
    [anon_sym_false] = ACTIONS(1646),
    [anon_sym_LBRACE] = ACTIONS(1646),
    [anon_sym_SEMI] = ACTIONS(1648),
    [anon_sym_RBRACE] = ACTIONS(1648),
    [anon_sym_constraint] = ACTIONS(1646),
    [anon_sym_val] = ACTIONS(1646),
    [anon_sym_end] = ACTIONS(1646),
    [anon_sym_with] = ACTIONS(1646),
    [anon_sym_COLON_GT] = ACTIONS(1648),
    [anon_sym_object] = ACTIONS(1646),
    [anon_sym_inherit] = ACTIONS(1646),
    [anon_sym_method] = ACTIONS(1646),
    [anon_sym_in] = ACTIONS(1646),
    [anon_sym_as] = ACTIONS(1646),
    [anon_sym_initializer] = ACTIONS(1646),
    [anon_sym_AMP] = ACTIONS(1646),
    [anon_sym_POUND] = ACTIONS(1646),
    [anon_sym_COLON_COLON] = ACTIONS(1648),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1648),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1648),
    [anon_sym_then] = ACTIONS(1646),
    [anon_sym_else] = ACTIONS(1646),
    [anon_sym_do] = ACTIONS(1646),
    [anon_sym_done] = ACTIONS(1646),
    [anon_sym_to] = ACTIONS(1646),
    [anon_sym_downto] = ACTIONS(1646),
    [anon_sym_new] = ACTIONS(1646),
    [anon_sym_LBRACE_LT] = ACTIONS(1648),
    [anon_sym_GT_RBRACE] = ACTIONS(1648),
    [anon_sym_begin] = ACTIONS(1646),
    [sym_ocamlyacc_value] = ACTIONS(1648),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1646),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1648),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1646),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1646),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1648),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1648),
    [aux_sym_number_token1] = ACTIONS(1648),
    [aux_sym_signed_number_token1] = ACTIONS(1646),
    [anon_sym_SQUOTE] = ACTIONS(1648),
    [anon_sym_DQUOTE] = ACTIONS(1648),
    [sym_prefix_operator] = ACTIONS(1648),
    [aux_sym_sign_operator_token1] = ACTIONS(1646),
    [sym_hash_operator] = ACTIONS(1648),
    [anon_sym_lsl] = ACTIONS(1646),
    [anon_sym_lsr] = ACTIONS(1646),
    [anon_sym_asr] = ACTIONS(1646),
    [aux_sym_pow_operator_token1] = ACTIONS(1648),
    [anon_sym_mod] = ACTIONS(1646),
    [anon_sym_land] = ACTIONS(1646),
    [anon_sym_lor] = ACTIONS(1646),
    [anon_sym_lxor] = ACTIONS(1646),
    [aux_sym_mult_operator_token1] = ACTIONS(1646),
    [aux_sym_add_operator_token1] = ACTIONS(1646),
    [sym_concat_operator] = ACTIONS(1648),
    [sym_rel_operator] = ACTIONS(1646),
    [anon_sym_AMP_AMP] = ACTIONS(1646),
    [anon_sym_or] = ACTIONS(1646),
    [anon_sym_PIPE_PIPE] = ACTIONS(1646),
    [sym_assign_operator] = ACTIONS(1648),
    [sym__uppercase_identifier] = ACTIONS(1648),
    [aux_sym_tag_token1] = ACTIONS(1648),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(794)] = {
    [sym_attribute] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(1650),
    [anon_sym_and] = ACTIONS(1650),
    [anon_sym_TILDE] = ACTIONS(1650),
    [anon_sym_QMARK] = ACTIONS(1650),
    [anon_sym_LPAREN] = ACTIONS(1652),
    [anon_sym_RPAREN] = ACTIONS(1652),
    [anon_sym_COMMA] = ACTIONS(1652),
    [anon_sym_PIPE] = ACTIONS(1650),
    [anon_sym_COLON2] = ACTIONS(1650),
    [anon_sym_DOT] = ACTIONS(1652),
    [anon_sym_DASH_GT] = ACTIONS(1650),
    [anon_sym_LBRACK] = ACTIONS(1650),
    [anon_sym_RBRACK] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1650),
    [anon_sym_false] = ACTIONS(1650),
    [anon_sym_LBRACE] = ACTIONS(1650),
    [anon_sym_SEMI] = ACTIONS(1652),
    [anon_sym_RBRACE] = ACTIONS(1652),
    [anon_sym_constraint] = ACTIONS(1650),
    [anon_sym_val] = ACTIONS(1650),
    [anon_sym_end] = ACTIONS(1650),
    [anon_sym_with] = ACTIONS(1650),
    [anon_sym_COLON_GT] = ACTIONS(1652),
    [anon_sym_object] = ACTIONS(1650),
    [anon_sym_inherit] = ACTIONS(1650),
    [anon_sym_method] = ACTIONS(1650),
    [anon_sym_in] = ACTIONS(1650),
    [anon_sym_as] = ACTIONS(1650),
    [anon_sym_initializer] = ACTIONS(1650),
    [anon_sym_AMP] = ACTIONS(1650),
    [anon_sym_POUND] = ACTIONS(1650),
    [anon_sym_COLON_COLON] = ACTIONS(1652),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1652),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1652),
    [anon_sym_then] = ACTIONS(1650),
    [anon_sym_else] = ACTIONS(1650),
    [anon_sym_do] = ACTIONS(1650),
    [anon_sym_done] = ACTIONS(1650),
    [anon_sym_to] = ACTIONS(1650),
    [anon_sym_downto] = ACTIONS(1650),
    [anon_sym_new] = ACTIONS(1650),
    [anon_sym_LBRACE_LT] = ACTIONS(1652),
    [anon_sym_GT_RBRACE] = ACTIONS(1652),
    [anon_sym_begin] = ACTIONS(1650),
    [sym_ocamlyacc_value] = ACTIONS(1652),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1650),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1652),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1650),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1650),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1652),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1652),
    [aux_sym_number_token1] = ACTIONS(1652),
    [aux_sym_signed_number_token1] = ACTIONS(1650),
    [anon_sym_SQUOTE] = ACTIONS(1652),
    [anon_sym_DQUOTE] = ACTIONS(1652),
    [sym_prefix_operator] = ACTIONS(1652),
    [aux_sym_sign_operator_token1] = ACTIONS(1650),
    [sym_hash_operator] = ACTIONS(1652),
    [anon_sym_lsl] = ACTIONS(1650),
    [anon_sym_lsr] = ACTIONS(1650),
    [anon_sym_asr] = ACTIONS(1650),
    [aux_sym_pow_operator_token1] = ACTIONS(1652),
    [anon_sym_mod] = ACTIONS(1650),
    [anon_sym_land] = ACTIONS(1650),
    [anon_sym_lor] = ACTIONS(1650),
    [anon_sym_lxor] = ACTIONS(1650),
    [aux_sym_mult_operator_token1] = ACTIONS(1650),
    [aux_sym_add_operator_token1] = ACTIONS(1650),
    [sym_concat_operator] = ACTIONS(1652),
    [sym_rel_operator] = ACTIONS(1650),
    [anon_sym_AMP_AMP] = ACTIONS(1650),
    [anon_sym_or] = ACTIONS(1650),
    [anon_sym_PIPE_PIPE] = ACTIONS(1650),
    [sym_assign_operator] = ACTIONS(1652),
    [sym__uppercase_identifier] = ACTIONS(1652),
    [aux_sym_tag_token1] = ACTIONS(1652),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(795)] = {
    [sym__delimited_expression] = STATE(2350),
    [sym__simple_expression] = STATE(2404),
    [sym_typed_expression] = STATE(2350),
    [sym_list_expression] = STATE(2351),
    [sym_array_expression] = STATE(2351),
    [sym_record_expression] = STATE(2351),
    [sym_labeled_argument] = STATE(2522),
    [sym_prefix_expression] = STATE(2350),
    [sym_hash_expression] = STATE(2350),
    [sym_field_get_expression] = STATE(2350),
    [sym_array_get_expression] = STATE(2350),
    [sym_string_get_expression] = STATE(2350),
    [sym_bigarray_get_expression] = STATE(2350),
    [sym_local_open_expression] = STATE(2350),
    [sym_package_expression] = STATE(2351),
    [sym_new_expression] = STATE(2350),
    [sym_object_copy_expression] = STATE(2351),
    [sym_method_invocation] = STATE(2350),
    [sym_object_expression] = STATE(2350),
    [sym_parenthesized_expression] = STATE(2351),
    [sym_attribute] = STATE(795),
    [sym_extension] = STATE(2350),
    [sym_quoted_extension] = STATE(2350),
    [sym__constant] = STATE(2350),
    [sym_number] = STATE(2348),
    [sym_character] = STATE(2348),
    [sym_string] = STATE(2348),
    [sym_quoted_string] = STATE(2348),
    [sym_boolean] = STATE(2348),
    [sym_unit] = STATE(2351),
    [sym__value_name] = STATE(2402),
    [sym_parenthesized_operator] = STATE(2406),
    [sym_value_path] = STATE(2350),
    [sym_module_path] = STATE(6407),
    [sym_constructor_path] = STATE(2350),
    [sym__constructor_name] = STATE(2419),
    [sym_tag] = STATE(2350),
    [aux_sym_class_application_repeat1] = STATE(824),
    [ts_builtin_sym_end] = ACTIONS(1538),
    [sym__lowercase_identifier] = ACTIONS(1654),
    [anon_sym_SEMI_SEMI] = ACTIONS(1538),
    [anon_sym_and] = ACTIONS(1540),
    [anon_sym_TILDE] = ACTIONS(1656),
    [anon_sym_QMARK] = ACTIONS(1656),
    [anon_sym_LPAREN] = ACTIONS(1658),
    [anon_sym_external] = ACTIONS(1540),
    [anon_sym_type] = ACTIONS(1540),
    [anon_sym_LBRACK] = ACTIONS(1660),
    [anon_sym_RBRACK] = ACTIONS(1538),
    [anon_sym_true] = ACTIONS(1662),
    [anon_sym_false] = ACTIONS(1662),
    [anon_sym_LBRACE] = ACTIONS(1664),
    [anon_sym_exception] = ACTIONS(1540),
    [anon_sym_module] = ACTIONS(1540),
    [anon_sym_open] = ACTIONS(1540),
    [anon_sym_include] = ACTIONS(1540),
    [anon_sym_class] = ACTIONS(1540),
    [anon_sym_val] = ACTIONS(1540),
    [anon_sym_end] = ACTIONS(1540),
    [anon_sym_object] = ACTIONS(1666),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1668),
    [anon_sym_new] = ACTIONS(1670),
    [anon_sym_LBRACE_LT] = ACTIONS(1672),
    [anon_sym_begin] = ACTIONS(1674),
    [sym_ocamlyacc_value] = ACTIONS(1676),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1540),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1538),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1678),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1680),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1538),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1538),
    [aux_sym_number_token1] = ACTIONS(1682),
    [anon_sym_SQUOTE] = ACTIONS(1684),
    [anon_sym_DQUOTE] = ACTIONS(1686),
    [sym_prefix_operator] = ACTIONS(1688),
    [sym__uppercase_identifier] = ACTIONS(1690),
    [aux_sym_tag_token1] = ACTIONS(1692),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(796)] = {
    [sym_attribute] = STATE(796),
    [sym__lowercase_identifier] = ACTIONS(1694),
    [anon_sym_and] = ACTIONS(1694),
    [anon_sym_TILDE] = ACTIONS(1694),
    [anon_sym_QMARK] = ACTIONS(1694),
    [anon_sym_LPAREN] = ACTIONS(1696),
    [anon_sym_RPAREN] = ACTIONS(1696),
    [anon_sym_COMMA] = ACTIONS(1696),
    [anon_sym_PIPE] = ACTIONS(1694),
    [anon_sym_COLON2] = ACTIONS(1694),
    [anon_sym_DOT] = ACTIONS(1696),
    [anon_sym_DASH_GT] = ACTIONS(1694),
    [anon_sym_LBRACK] = ACTIONS(1694),
    [anon_sym_RBRACK] = ACTIONS(1696),
    [anon_sym_true] = ACTIONS(1694),
    [anon_sym_false] = ACTIONS(1694),
    [anon_sym_LBRACE] = ACTIONS(1694),
    [anon_sym_SEMI] = ACTIONS(1696),
    [anon_sym_RBRACE] = ACTIONS(1696),
    [anon_sym_constraint] = ACTIONS(1694),
    [anon_sym_val] = ACTIONS(1694),
    [anon_sym_end] = ACTIONS(1694),
    [anon_sym_with] = ACTIONS(1694),
    [anon_sym_COLON_GT] = ACTIONS(1696),
    [anon_sym_object] = ACTIONS(1694),
    [anon_sym_inherit] = ACTIONS(1694),
    [anon_sym_method] = ACTIONS(1694),
    [anon_sym_in] = ACTIONS(1694),
    [anon_sym_as] = ACTIONS(1694),
    [anon_sym_initializer] = ACTIONS(1694),
    [anon_sym_AMP] = ACTIONS(1694),
    [anon_sym_POUND] = ACTIONS(1694),
    [anon_sym_COLON_COLON] = ACTIONS(1696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1696),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1696),
    [anon_sym_then] = ACTIONS(1694),
    [anon_sym_else] = ACTIONS(1694),
    [anon_sym_do] = ACTIONS(1694),
    [anon_sym_done] = ACTIONS(1694),
    [anon_sym_to] = ACTIONS(1694),
    [anon_sym_downto] = ACTIONS(1694),
    [anon_sym_new] = ACTIONS(1694),
    [anon_sym_LBRACE_LT] = ACTIONS(1696),
    [anon_sym_GT_RBRACE] = ACTIONS(1696),
    [anon_sym_begin] = ACTIONS(1694),
    [sym_ocamlyacc_value] = ACTIONS(1696),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1694),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1696),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1694),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1694),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1696),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1696),
    [aux_sym_number_token1] = ACTIONS(1696),
    [aux_sym_signed_number_token1] = ACTIONS(1694),
    [anon_sym_SQUOTE] = ACTIONS(1696),
    [anon_sym_DQUOTE] = ACTIONS(1696),
    [sym_prefix_operator] = ACTIONS(1696),
    [aux_sym_sign_operator_token1] = ACTIONS(1694),
    [sym_hash_operator] = ACTIONS(1696),
    [anon_sym_lsl] = ACTIONS(1694),
    [anon_sym_lsr] = ACTIONS(1694),
    [anon_sym_asr] = ACTIONS(1694),
    [aux_sym_pow_operator_token1] = ACTIONS(1696),
    [anon_sym_mod] = ACTIONS(1694),
    [anon_sym_land] = ACTIONS(1694),
    [anon_sym_lor] = ACTIONS(1694),
    [anon_sym_lxor] = ACTIONS(1694),
    [aux_sym_mult_operator_token1] = ACTIONS(1694),
    [aux_sym_add_operator_token1] = ACTIONS(1694),
    [sym_concat_operator] = ACTIONS(1696),
    [sym_rel_operator] = ACTIONS(1694),
    [anon_sym_AMP_AMP] = ACTIONS(1694),
    [anon_sym_or] = ACTIONS(1694),
    [anon_sym_PIPE_PIPE] = ACTIONS(1694),
    [sym_assign_operator] = ACTIONS(1696),
    [sym__uppercase_identifier] = ACTIONS(1696),
    [aux_sym_tag_token1] = ACTIONS(1696),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(797)] = {
    [sym_attribute] = STATE(797),
    [sym__lowercase_identifier] = ACTIONS(1698),
    [anon_sym_and] = ACTIONS(1698),
    [anon_sym_TILDE] = ACTIONS(1698),
    [anon_sym_QMARK] = ACTIONS(1698),
    [anon_sym_LPAREN] = ACTIONS(1700),
    [anon_sym_RPAREN] = ACTIONS(1700),
    [anon_sym_COMMA] = ACTIONS(1700),
    [anon_sym_PIPE] = ACTIONS(1698),
    [anon_sym_COLON2] = ACTIONS(1698),
    [anon_sym_DOT] = ACTIONS(1702),
    [anon_sym_DASH_GT] = ACTIONS(1698),
    [anon_sym_LBRACK] = ACTIONS(1698),
    [anon_sym_RBRACK] = ACTIONS(1700),
    [anon_sym_true] = ACTIONS(1698),
    [anon_sym_false] = ACTIONS(1698),
    [anon_sym_LBRACE] = ACTIONS(1698),
    [anon_sym_SEMI] = ACTIONS(1700),
    [anon_sym_RBRACE] = ACTIONS(1700),
    [anon_sym_constraint] = ACTIONS(1698),
    [anon_sym_val] = ACTIONS(1698),
    [anon_sym_end] = ACTIONS(1698),
    [anon_sym_with] = ACTIONS(1698),
    [anon_sym_COLON_GT] = ACTIONS(1700),
    [anon_sym_object] = ACTIONS(1698),
    [anon_sym_inherit] = ACTIONS(1698),
    [anon_sym_method] = ACTIONS(1698),
    [anon_sym_in] = ACTIONS(1698),
    [anon_sym_as] = ACTIONS(1698),
    [anon_sym_initializer] = ACTIONS(1698),
    [anon_sym_AMP] = ACTIONS(1698),
    [anon_sym_POUND] = ACTIONS(1698),
    [anon_sym_COLON_COLON] = ACTIONS(1700),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1700),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1700),
    [anon_sym_then] = ACTIONS(1698),
    [anon_sym_else] = ACTIONS(1698),
    [anon_sym_do] = ACTIONS(1698),
    [anon_sym_done] = ACTIONS(1698),
    [anon_sym_to] = ACTIONS(1698),
    [anon_sym_downto] = ACTIONS(1698),
    [anon_sym_new] = ACTIONS(1698),
    [anon_sym_LBRACE_LT] = ACTIONS(1700),
    [anon_sym_GT_RBRACE] = ACTIONS(1700),
    [anon_sym_begin] = ACTIONS(1698),
    [sym_ocamlyacc_value] = ACTIONS(1700),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1698),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1700),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1700),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1700),
    [aux_sym_number_token1] = ACTIONS(1700),
    [aux_sym_signed_number_token1] = ACTIONS(1698),
    [anon_sym_SQUOTE] = ACTIONS(1700),
    [anon_sym_DQUOTE] = ACTIONS(1700),
    [sym_prefix_operator] = ACTIONS(1700),
    [aux_sym_sign_operator_token1] = ACTIONS(1698),
    [sym_hash_operator] = ACTIONS(1700),
    [anon_sym_lsl] = ACTIONS(1698),
    [anon_sym_lsr] = ACTIONS(1698),
    [anon_sym_asr] = ACTIONS(1698),
    [aux_sym_pow_operator_token1] = ACTIONS(1700),
    [anon_sym_mod] = ACTIONS(1698),
    [anon_sym_land] = ACTIONS(1698),
    [anon_sym_lor] = ACTIONS(1698),
    [anon_sym_lxor] = ACTIONS(1698),
    [aux_sym_mult_operator_token1] = ACTIONS(1698),
    [aux_sym_add_operator_token1] = ACTIONS(1698),
    [sym_concat_operator] = ACTIONS(1700),
    [sym_rel_operator] = ACTIONS(1698),
    [anon_sym_AMP_AMP] = ACTIONS(1698),
    [anon_sym_or] = ACTIONS(1698),
    [anon_sym_PIPE_PIPE] = ACTIONS(1698),
    [sym_assign_operator] = ACTIONS(1700),
    [sym__uppercase_identifier] = ACTIONS(1700),
    [aux_sym_tag_token1] = ACTIONS(1700),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(798)] = {
    [sym_attribute] = STATE(798),
    [sym__lowercase_identifier] = ACTIONS(1704),
    [anon_sym_and] = ACTIONS(1704),
    [anon_sym_TILDE] = ACTIONS(1704),
    [anon_sym_QMARK] = ACTIONS(1704),
    [anon_sym_LPAREN] = ACTIONS(1706),
    [anon_sym_RPAREN] = ACTIONS(1706),
    [anon_sym_COMMA] = ACTIONS(1706),
    [anon_sym_PIPE] = ACTIONS(1704),
    [anon_sym_COLON2] = ACTIONS(1704),
    [anon_sym_DOT] = ACTIONS(1706),
    [anon_sym_DASH_GT] = ACTIONS(1704),
    [anon_sym_LBRACK] = ACTIONS(1704),
    [anon_sym_RBRACK] = ACTIONS(1706),
    [anon_sym_true] = ACTIONS(1704),
    [anon_sym_false] = ACTIONS(1704),
    [anon_sym_LBRACE] = ACTIONS(1704),
    [anon_sym_SEMI] = ACTIONS(1706),
    [anon_sym_RBRACE] = ACTIONS(1706),
    [anon_sym_constraint] = ACTIONS(1704),
    [anon_sym_val] = ACTIONS(1704),
    [anon_sym_end] = ACTIONS(1704),
    [anon_sym_with] = ACTIONS(1704),
    [anon_sym_COLON_GT] = ACTIONS(1706),
    [anon_sym_object] = ACTIONS(1704),
    [anon_sym_inherit] = ACTIONS(1704),
    [anon_sym_method] = ACTIONS(1704),
    [anon_sym_in] = ACTIONS(1704),
    [anon_sym_as] = ACTIONS(1704),
    [anon_sym_initializer] = ACTIONS(1704),
    [anon_sym_AMP] = ACTIONS(1704),
    [anon_sym_POUND] = ACTIONS(1704),
    [anon_sym_COLON_COLON] = ACTIONS(1706),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1706),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1706),
    [anon_sym_then] = ACTIONS(1704),
    [anon_sym_else] = ACTIONS(1704),
    [anon_sym_do] = ACTIONS(1704),
    [anon_sym_done] = ACTIONS(1704),
    [anon_sym_to] = ACTIONS(1704),
    [anon_sym_downto] = ACTIONS(1704),
    [anon_sym_new] = ACTIONS(1704),
    [anon_sym_LBRACE_LT] = ACTIONS(1706),
    [anon_sym_GT_RBRACE] = ACTIONS(1706),
    [anon_sym_begin] = ACTIONS(1704),
    [sym_ocamlyacc_value] = ACTIONS(1706),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1704),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1706),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1704),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1704),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1706),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1706),
    [aux_sym_number_token1] = ACTIONS(1706),
    [aux_sym_signed_number_token1] = ACTIONS(1704),
    [anon_sym_SQUOTE] = ACTIONS(1706),
    [anon_sym_DQUOTE] = ACTIONS(1706),
    [sym_prefix_operator] = ACTIONS(1706),
    [aux_sym_sign_operator_token1] = ACTIONS(1704),
    [sym_hash_operator] = ACTIONS(1706),
    [anon_sym_lsl] = ACTIONS(1704),
    [anon_sym_lsr] = ACTIONS(1704),
    [anon_sym_asr] = ACTIONS(1704),
    [aux_sym_pow_operator_token1] = ACTIONS(1706),
    [anon_sym_mod] = ACTIONS(1704),
    [anon_sym_land] = ACTIONS(1704),
    [anon_sym_lor] = ACTIONS(1704),
    [anon_sym_lxor] = ACTIONS(1704),
    [aux_sym_mult_operator_token1] = ACTIONS(1704),
    [aux_sym_add_operator_token1] = ACTIONS(1704),
    [sym_concat_operator] = ACTIONS(1706),
    [sym_rel_operator] = ACTIONS(1704),
    [anon_sym_AMP_AMP] = ACTIONS(1704),
    [anon_sym_or] = ACTIONS(1704),
    [anon_sym_PIPE_PIPE] = ACTIONS(1704),
    [sym_assign_operator] = ACTIONS(1706),
    [sym__uppercase_identifier] = ACTIONS(1706),
    [aux_sym_tag_token1] = ACTIONS(1706),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(799)] = {
    [sym_attribute] = STATE(799),
    [sym__lowercase_identifier] = ACTIONS(1708),
    [anon_sym_and] = ACTIONS(1708),
    [anon_sym_TILDE] = ACTIONS(1708),
    [anon_sym_QMARK] = ACTIONS(1708),
    [anon_sym_LPAREN] = ACTIONS(1710),
    [anon_sym_RPAREN] = ACTIONS(1710),
    [anon_sym_COMMA] = ACTIONS(1710),
    [anon_sym_PIPE] = ACTIONS(1708),
    [anon_sym_COLON2] = ACTIONS(1708),
    [anon_sym_DOT] = ACTIONS(1710),
    [anon_sym_DASH_GT] = ACTIONS(1708),
    [anon_sym_LBRACK] = ACTIONS(1708),
    [anon_sym_RBRACK] = ACTIONS(1710),
    [anon_sym_true] = ACTIONS(1708),
    [anon_sym_false] = ACTIONS(1708),
    [anon_sym_LBRACE] = ACTIONS(1708),
    [anon_sym_SEMI] = ACTIONS(1710),
    [anon_sym_RBRACE] = ACTIONS(1710),
    [anon_sym_constraint] = ACTIONS(1708),
    [anon_sym_val] = ACTIONS(1708),
    [anon_sym_end] = ACTIONS(1708),
    [anon_sym_with] = ACTIONS(1708),
    [anon_sym_COLON_GT] = ACTIONS(1710),
    [anon_sym_object] = ACTIONS(1708),
    [anon_sym_inherit] = ACTIONS(1708),
    [anon_sym_method] = ACTIONS(1708),
    [anon_sym_in] = ACTIONS(1708),
    [anon_sym_as] = ACTIONS(1708),
    [anon_sym_initializer] = ACTIONS(1708),
    [anon_sym_AMP] = ACTIONS(1708),
    [anon_sym_POUND] = ACTIONS(1708),
    [anon_sym_COLON_COLON] = ACTIONS(1710),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1710),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1710),
    [anon_sym_then] = ACTIONS(1708),
    [anon_sym_else] = ACTIONS(1708),
    [anon_sym_do] = ACTIONS(1708),
    [anon_sym_done] = ACTIONS(1708),
    [anon_sym_to] = ACTIONS(1708),
    [anon_sym_downto] = ACTIONS(1708),
    [anon_sym_new] = ACTIONS(1708),
    [anon_sym_LBRACE_LT] = ACTIONS(1710),
    [anon_sym_GT_RBRACE] = ACTIONS(1710),
    [anon_sym_begin] = ACTIONS(1708),
    [sym_ocamlyacc_value] = ACTIONS(1710),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1708),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1710),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1708),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1708),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1710),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1710),
    [aux_sym_number_token1] = ACTIONS(1710),
    [aux_sym_signed_number_token1] = ACTIONS(1708),
    [anon_sym_SQUOTE] = ACTIONS(1710),
    [anon_sym_DQUOTE] = ACTIONS(1710),
    [sym_prefix_operator] = ACTIONS(1710),
    [aux_sym_sign_operator_token1] = ACTIONS(1708),
    [sym_hash_operator] = ACTIONS(1710),
    [anon_sym_lsl] = ACTIONS(1708),
    [anon_sym_lsr] = ACTIONS(1708),
    [anon_sym_asr] = ACTIONS(1708),
    [aux_sym_pow_operator_token1] = ACTIONS(1710),
    [anon_sym_mod] = ACTIONS(1708),
    [anon_sym_land] = ACTIONS(1708),
    [anon_sym_lor] = ACTIONS(1708),
    [anon_sym_lxor] = ACTIONS(1708),
    [aux_sym_mult_operator_token1] = ACTIONS(1708),
    [aux_sym_add_operator_token1] = ACTIONS(1708),
    [sym_concat_operator] = ACTIONS(1710),
    [sym_rel_operator] = ACTIONS(1708),
    [anon_sym_AMP_AMP] = ACTIONS(1708),
    [anon_sym_or] = ACTIONS(1708),
    [anon_sym_PIPE_PIPE] = ACTIONS(1708),
    [sym_assign_operator] = ACTIONS(1710),
    [sym__uppercase_identifier] = ACTIONS(1710),
    [aux_sym_tag_token1] = ACTIONS(1710),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(800)] = {
    [sym_attribute] = STATE(800),
    [sym__lowercase_identifier] = ACTIONS(1712),
    [anon_sym_and] = ACTIONS(1712),
    [anon_sym_TILDE] = ACTIONS(1712),
    [anon_sym_QMARK] = ACTIONS(1712),
    [anon_sym_LPAREN] = ACTIONS(1714),
    [anon_sym_RPAREN] = ACTIONS(1714),
    [anon_sym_COMMA] = ACTIONS(1714),
    [anon_sym_PIPE] = ACTIONS(1712),
    [anon_sym_COLON2] = ACTIONS(1712),
    [anon_sym_DOT] = ACTIONS(1714),
    [anon_sym_DASH_GT] = ACTIONS(1712),
    [anon_sym_LBRACK] = ACTIONS(1712),
    [anon_sym_RBRACK] = ACTIONS(1714),
    [anon_sym_true] = ACTIONS(1712),
    [anon_sym_false] = ACTIONS(1712),
    [anon_sym_LBRACE] = ACTIONS(1712),
    [anon_sym_SEMI] = ACTIONS(1714),
    [anon_sym_RBRACE] = ACTIONS(1714),
    [anon_sym_constraint] = ACTIONS(1712),
    [anon_sym_val] = ACTIONS(1712),
    [anon_sym_end] = ACTIONS(1712),
    [anon_sym_with] = ACTIONS(1712),
    [anon_sym_COLON_GT] = ACTIONS(1714),
    [anon_sym_object] = ACTIONS(1712),
    [anon_sym_inherit] = ACTIONS(1712),
    [anon_sym_method] = ACTIONS(1712),
    [anon_sym_in] = ACTIONS(1712),
    [anon_sym_as] = ACTIONS(1712),
    [anon_sym_initializer] = ACTIONS(1712),
    [anon_sym_AMP] = ACTIONS(1712),
    [anon_sym_POUND] = ACTIONS(1712),
    [anon_sym_COLON_COLON] = ACTIONS(1714),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1714),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1714),
    [anon_sym_then] = ACTIONS(1712),
    [anon_sym_else] = ACTIONS(1712),
    [anon_sym_do] = ACTIONS(1712),
    [anon_sym_done] = ACTIONS(1712),
    [anon_sym_to] = ACTIONS(1712),
    [anon_sym_downto] = ACTIONS(1712),
    [anon_sym_new] = ACTIONS(1712),
    [anon_sym_LBRACE_LT] = ACTIONS(1714),
    [anon_sym_GT_RBRACE] = ACTIONS(1714),
    [anon_sym_begin] = ACTIONS(1712),
    [sym_ocamlyacc_value] = ACTIONS(1714),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1712),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1714),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1712),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1712),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1714),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1714),
    [aux_sym_number_token1] = ACTIONS(1714),
    [aux_sym_signed_number_token1] = ACTIONS(1712),
    [anon_sym_SQUOTE] = ACTIONS(1714),
    [anon_sym_DQUOTE] = ACTIONS(1714),
    [sym_prefix_operator] = ACTIONS(1714),
    [aux_sym_sign_operator_token1] = ACTIONS(1712),
    [sym_hash_operator] = ACTIONS(1714),
    [anon_sym_lsl] = ACTIONS(1712),
    [anon_sym_lsr] = ACTIONS(1712),
    [anon_sym_asr] = ACTIONS(1712),
    [aux_sym_pow_operator_token1] = ACTIONS(1714),
    [anon_sym_mod] = ACTIONS(1712),
    [anon_sym_land] = ACTIONS(1712),
    [anon_sym_lor] = ACTIONS(1712),
    [anon_sym_lxor] = ACTIONS(1712),
    [aux_sym_mult_operator_token1] = ACTIONS(1712),
    [aux_sym_add_operator_token1] = ACTIONS(1712),
    [sym_concat_operator] = ACTIONS(1714),
    [sym_rel_operator] = ACTIONS(1712),
    [anon_sym_AMP_AMP] = ACTIONS(1712),
    [anon_sym_or] = ACTIONS(1712),
    [anon_sym_PIPE_PIPE] = ACTIONS(1712),
    [sym_assign_operator] = ACTIONS(1714),
    [sym__uppercase_identifier] = ACTIONS(1714),
    [aux_sym_tag_token1] = ACTIONS(1714),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(801)] = {
    [sym_attribute] = STATE(801),
    [sym__lowercase_identifier] = ACTIONS(1716),
    [anon_sym_and] = ACTIONS(1716),
    [anon_sym_TILDE] = ACTIONS(1716),
    [anon_sym_QMARK] = ACTIONS(1716),
    [anon_sym_LPAREN] = ACTIONS(1718),
    [anon_sym_RPAREN] = ACTIONS(1718),
    [anon_sym_COMMA] = ACTIONS(1718),
    [anon_sym_PIPE] = ACTIONS(1716),
    [anon_sym_COLON2] = ACTIONS(1716),
    [anon_sym_DOT] = ACTIONS(1718),
    [anon_sym_DASH_GT] = ACTIONS(1716),
    [anon_sym_LBRACK] = ACTIONS(1716),
    [anon_sym_RBRACK] = ACTIONS(1718),
    [anon_sym_true] = ACTIONS(1716),
    [anon_sym_false] = ACTIONS(1716),
    [anon_sym_LBRACE] = ACTIONS(1716),
    [anon_sym_SEMI] = ACTIONS(1718),
    [anon_sym_RBRACE] = ACTIONS(1718),
    [anon_sym_constraint] = ACTIONS(1716),
    [anon_sym_val] = ACTIONS(1716),
    [anon_sym_end] = ACTIONS(1716),
    [anon_sym_with] = ACTIONS(1716),
    [anon_sym_COLON_GT] = ACTIONS(1718),
    [anon_sym_object] = ACTIONS(1716),
    [anon_sym_inherit] = ACTIONS(1716),
    [anon_sym_method] = ACTIONS(1716),
    [anon_sym_in] = ACTIONS(1716),
    [anon_sym_as] = ACTIONS(1716),
    [anon_sym_initializer] = ACTIONS(1716),
    [anon_sym_AMP] = ACTIONS(1716),
    [anon_sym_POUND] = ACTIONS(1716),
    [anon_sym_COLON_COLON] = ACTIONS(1718),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1718),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1718),
    [anon_sym_then] = ACTIONS(1716),
    [anon_sym_else] = ACTIONS(1716),
    [anon_sym_do] = ACTIONS(1716),
    [anon_sym_done] = ACTIONS(1716),
    [anon_sym_to] = ACTIONS(1716),
    [anon_sym_downto] = ACTIONS(1716),
    [anon_sym_new] = ACTIONS(1716),
    [anon_sym_LBRACE_LT] = ACTIONS(1718),
    [anon_sym_GT_RBRACE] = ACTIONS(1718),
    [anon_sym_begin] = ACTIONS(1716),
    [sym_ocamlyacc_value] = ACTIONS(1718),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1716),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1718),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1716),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1716),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1718),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1718),
    [aux_sym_number_token1] = ACTIONS(1718),
    [aux_sym_signed_number_token1] = ACTIONS(1716),
    [anon_sym_SQUOTE] = ACTIONS(1718),
    [anon_sym_DQUOTE] = ACTIONS(1718),
    [sym_prefix_operator] = ACTIONS(1718),
    [aux_sym_sign_operator_token1] = ACTIONS(1716),
    [sym_hash_operator] = ACTIONS(1718),
    [anon_sym_lsl] = ACTIONS(1716),
    [anon_sym_lsr] = ACTIONS(1716),
    [anon_sym_asr] = ACTIONS(1716),
    [aux_sym_pow_operator_token1] = ACTIONS(1718),
    [anon_sym_mod] = ACTIONS(1716),
    [anon_sym_land] = ACTIONS(1716),
    [anon_sym_lor] = ACTIONS(1716),
    [anon_sym_lxor] = ACTIONS(1716),
    [aux_sym_mult_operator_token1] = ACTIONS(1716),
    [aux_sym_add_operator_token1] = ACTIONS(1716),
    [sym_concat_operator] = ACTIONS(1718),
    [sym_rel_operator] = ACTIONS(1716),
    [anon_sym_AMP_AMP] = ACTIONS(1716),
    [anon_sym_or] = ACTIONS(1716),
    [anon_sym_PIPE_PIPE] = ACTIONS(1716),
    [sym_assign_operator] = ACTIONS(1718),
    [sym__uppercase_identifier] = ACTIONS(1718),
    [aux_sym_tag_token1] = ACTIONS(1718),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(802)] = {
    [sym_attribute] = STATE(802),
    [sym__lowercase_identifier] = ACTIONS(1720),
    [anon_sym_and] = ACTIONS(1720),
    [anon_sym_TILDE] = ACTIONS(1720),
    [anon_sym_QMARK] = ACTIONS(1720),
    [anon_sym_LPAREN] = ACTIONS(1722),
    [anon_sym_RPAREN] = ACTIONS(1722),
    [anon_sym_COMMA] = ACTIONS(1722),
    [anon_sym_PIPE] = ACTIONS(1720),
    [anon_sym_COLON2] = ACTIONS(1720),
    [anon_sym_DOT] = ACTIONS(1722),
    [anon_sym_DASH_GT] = ACTIONS(1720),
    [anon_sym_LBRACK] = ACTIONS(1720),
    [anon_sym_RBRACK] = ACTIONS(1722),
    [anon_sym_true] = ACTIONS(1720),
    [anon_sym_false] = ACTIONS(1720),
    [anon_sym_LBRACE] = ACTIONS(1720),
    [anon_sym_SEMI] = ACTIONS(1722),
    [anon_sym_RBRACE] = ACTIONS(1722),
    [anon_sym_constraint] = ACTIONS(1720),
    [anon_sym_val] = ACTIONS(1720),
    [anon_sym_end] = ACTIONS(1720),
    [anon_sym_with] = ACTIONS(1720),
    [anon_sym_COLON_GT] = ACTIONS(1722),
    [anon_sym_object] = ACTIONS(1720),
    [anon_sym_inherit] = ACTIONS(1720),
    [anon_sym_method] = ACTIONS(1720),
    [anon_sym_in] = ACTIONS(1720),
    [anon_sym_as] = ACTIONS(1720),
    [anon_sym_initializer] = ACTIONS(1720),
    [anon_sym_AMP] = ACTIONS(1720),
    [anon_sym_POUND] = ACTIONS(1720),
    [anon_sym_COLON_COLON] = ACTIONS(1722),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1722),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1722),
    [anon_sym_then] = ACTIONS(1720),
    [anon_sym_else] = ACTIONS(1720),
    [anon_sym_do] = ACTIONS(1720),
    [anon_sym_done] = ACTIONS(1720),
    [anon_sym_to] = ACTIONS(1720),
    [anon_sym_downto] = ACTIONS(1720),
    [anon_sym_new] = ACTIONS(1720),
    [anon_sym_LBRACE_LT] = ACTIONS(1722),
    [anon_sym_GT_RBRACE] = ACTIONS(1722),
    [anon_sym_begin] = ACTIONS(1720),
    [sym_ocamlyacc_value] = ACTIONS(1722),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1720),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1722),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1720),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1720),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1722),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1722),
    [aux_sym_number_token1] = ACTIONS(1722),
    [aux_sym_signed_number_token1] = ACTIONS(1720),
    [anon_sym_SQUOTE] = ACTIONS(1722),
    [anon_sym_DQUOTE] = ACTIONS(1722),
    [sym_prefix_operator] = ACTIONS(1722),
    [aux_sym_sign_operator_token1] = ACTIONS(1720),
    [sym_hash_operator] = ACTIONS(1722),
    [anon_sym_lsl] = ACTIONS(1720),
    [anon_sym_lsr] = ACTIONS(1720),
    [anon_sym_asr] = ACTIONS(1720),
    [aux_sym_pow_operator_token1] = ACTIONS(1722),
    [anon_sym_mod] = ACTIONS(1720),
    [anon_sym_land] = ACTIONS(1720),
    [anon_sym_lor] = ACTIONS(1720),
    [anon_sym_lxor] = ACTIONS(1720),
    [aux_sym_mult_operator_token1] = ACTIONS(1720),
    [aux_sym_add_operator_token1] = ACTIONS(1720),
    [sym_concat_operator] = ACTIONS(1722),
    [sym_rel_operator] = ACTIONS(1720),
    [anon_sym_AMP_AMP] = ACTIONS(1720),
    [anon_sym_or] = ACTIONS(1720),
    [anon_sym_PIPE_PIPE] = ACTIONS(1720),
    [sym_assign_operator] = ACTIONS(1722),
    [sym__uppercase_identifier] = ACTIONS(1722),
    [aux_sym_tag_token1] = ACTIONS(1722),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(803)] = {
    [sym_attribute] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(1724),
    [anon_sym_and] = ACTIONS(1724),
    [anon_sym_TILDE] = ACTIONS(1724),
    [anon_sym_QMARK] = ACTIONS(1724),
    [anon_sym_LPAREN] = ACTIONS(1726),
    [anon_sym_RPAREN] = ACTIONS(1726),
    [anon_sym_COMMA] = ACTIONS(1726),
    [anon_sym_PIPE] = ACTIONS(1724),
    [anon_sym_COLON2] = ACTIONS(1724),
    [anon_sym_DOT] = ACTIONS(1726),
    [anon_sym_DASH_GT] = ACTIONS(1724),
    [anon_sym_LBRACK] = ACTIONS(1724),
    [anon_sym_RBRACK] = ACTIONS(1726),
    [anon_sym_true] = ACTIONS(1724),
    [anon_sym_false] = ACTIONS(1724),
    [anon_sym_LBRACE] = ACTIONS(1724),
    [anon_sym_SEMI] = ACTIONS(1726),
    [anon_sym_RBRACE] = ACTIONS(1726),
    [anon_sym_constraint] = ACTIONS(1724),
    [anon_sym_val] = ACTIONS(1724),
    [anon_sym_end] = ACTIONS(1724),
    [anon_sym_with] = ACTIONS(1724),
    [anon_sym_COLON_GT] = ACTIONS(1726),
    [anon_sym_object] = ACTIONS(1724),
    [anon_sym_inherit] = ACTIONS(1724),
    [anon_sym_method] = ACTIONS(1724),
    [anon_sym_in] = ACTIONS(1724),
    [anon_sym_as] = ACTIONS(1724),
    [anon_sym_initializer] = ACTIONS(1724),
    [anon_sym_AMP] = ACTIONS(1724),
    [anon_sym_POUND] = ACTIONS(1724),
    [anon_sym_COLON_COLON] = ACTIONS(1726),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1726),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1726),
    [anon_sym_then] = ACTIONS(1724),
    [anon_sym_else] = ACTIONS(1724),
    [anon_sym_do] = ACTIONS(1724),
    [anon_sym_done] = ACTIONS(1724),
    [anon_sym_to] = ACTIONS(1724),
    [anon_sym_downto] = ACTIONS(1724),
    [anon_sym_new] = ACTIONS(1724),
    [anon_sym_LBRACE_LT] = ACTIONS(1726),
    [anon_sym_GT_RBRACE] = ACTIONS(1726),
    [anon_sym_begin] = ACTIONS(1724),
    [sym_ocamlyacc_value] = ACTIONS(1726),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1724),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1726),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1726),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1726),
    [aux_sym_number_token1] = ACTIONS(1726),
    [aux_sym_signed_number_token1] = ACTIONS(1724),
    [anon_sym_SQUOTE] = ACTIONS(1726),
    [anon_sym_DQUOTE] = ACTIONS(1726),
    [sym_prefix_operator] = ACTIONS(1726),
    [aux_sym_sign_operator_token1] = ACTIONS(1724),
    [sym_hash_operator] = ACTIONS(1726),
    [anon_sym_lsl] = ACTIONS(1724),
    [anon_sym_lsr] = ACTIONS(1724),
    [anon_sym_asr] = ACTIONS(1724),
    [aux_sym_pow_operator_token1] = ACTIONS(1726),
    [anon_sym_mod] = ACTIONS(1724),
    [anon_sym_land] = ACTIONS(1724),
    [anon_sym_lor] = ACTIONS(1724),
    [anon_sym_lxor] = ACTIONS(1724),
    [aux_sym_mult_operator_token1] = ACTIONS(1724),
    [aux_sym_add_operator_token1] = ACTIONS(1724),
    [sym_concat_operator] = ACTIONS(1726),
    [sym_rel_operator] = ACTIONS(1724),
    [anon_sym_AMP_AMP] = ACTIONS(1724),
    [anon_sym_or] = ACTIONS(1724),
    [anon_sym_PIPE_PIPE] = ACTIONS(1724),
    [sym_assign_operator] = ACTIONS(1726),
    [sym__uppercase_identifier] = ACTIONS(1726),
    [aux_sym_tag_token1] = ACTIONS(1726),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(804)] = {
    [sym_attribute] = STATE(804),
    [sym__lowercase_identifier] = ACTIONS(1728),
    [anon_sym_and] = ACTIONS(1728),
    [anon_sym_TILDE] = ACTIONS(1728),
    [anon_sym_QMARK] = ACTIONS(1728),
    [anon_sym_LPAREN] = ACTIONS(1730),
    [anon_sym_RPAREN] = ACTIONS(1730),
    [anon_sym_COMMA] = ACTIONS(1730),
    [anon_sym_PIPE] = ACTIONS(1728),
    [anon_sym_COLON2] = ACTIONS(1728),
    [anon_sym_DOT] = ACTIONS(1730),
    [anon_sym_DASH_GT] = ACTIONS(1728),
    [anon_sym_LBRACK] = ACTIONS(1728),
    [anon_sym_RBRACK] = ACTIONS(1730),
    [anon_sym_true] = ACTIONS(1728),
    [anon_sym_false] = ACTIONS(1728),
    [anon_sym_LBRACE] = ACTIONS(1728),
    [anon_sym_SEMI] = ACTIONS(1730),
    [anon_sym_RBRACE] = ACTIONS(1730),
    [anon_sym_constraint] = ACTIONS(1728),
    [anon_sym_val] = ACTIONS(1728),
    [anon_sym_end] = ACTIONS(1728),
    [anon_sym_with] = ACTIONS(1728),
    [anon_sym_COLON_GT] = ACTIONS(1730),
    [anon_sym_object] = ACTIONS(1728),
    [anon_sym_inherit] = ACTIONS(1728),
    [anon_sym_method] = ACTIONS(1728),
    [anon_sym_in] = ACTIONS(1728),
    [anon_sym_as] = ACTIONS(1728),
    [anon_sym_initializer] = ACTIONS(1728),
    [anon_sym_AMP] = ACTIONS(1728),
    [anon_sym_POUND] = ACTIONS(1728),
    [anon_sym_COLON_COLON] = ACTIONS(1730),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1730),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1730),
    [anon_sym_then] = ACTIONS(1728),
    [anon_sym_else] = ACTIONS(1728),
    [anon_sym_do] = ACTIONS(1728),
    [anon_sym_done] = ACTIONS(1728),
    [anon_sym_to] = ACTIONS(1728),
    [anon_sym_downto] = ACTIONS(1728),
    [anon_sym_new] = ACTIONS(1728),
    [anon_sym_LBRACE_LT] = ACTIONS(1730),
    [anon_sym_GT_RBRACE] = ACTIONS(1730),
    [anon_sym_begin] = ACTIONS(1728),
    [sym_ocamlyacc_value] = ACTIONS(1730),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1728),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1730),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1728),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1728),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1730),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1730),
    [aux_sym_number_token1] = ACTIONS(1730),
    [aux_sym_signed_number_token1] = ACTIONS(1728),
    [anon_sym_SQUOTE] = ACTIONS(1730),
    [anon_sym_DQUOTE] = ACTIONS(1730),
    [sym_prefix_operator] = ACTIONS(1730),
    [aux_sym_sign_operator_token1] = ACTIONS(1728),
    [sym_hash_operator] = ACTIONS(1730),
    [anon_sym_lsl] = ACTIONS(1728),
    [anon_sym_lsr] = ACTIONS(1728),
    [anon_sym_asr] = ACTIONS(1728),
    [aux_sym_pow_operator_token1] = ACTIONS(1730),
    [anon_sym_mod] = ACTIONS(1728),
    [anon_sym_land] = ACTIONS(1728),
    [anon_sym_lor] = ACTIONS(1728),
    [anon_sym_lxor] = ACTIONS(1728),
    [aux_sym_mult_operator_token1] = ACTIONS(1728),
    [aux_sym_add_operator_token1] = ACTIONS(1728),
    [sym_concat_operator] = ACTIONS(1730),
    [sym_rel_operator] = ACTIONS(1728),
    [anon_sym_AMP_AMP] = ACTIONS(1728),
    [anon_sym_or] = ACTIONS(1728),
    [anon_sym_PIPE_PIPE] = ACTIONS(1728),
    [sym_assign_operator] = ACTIONS(1730),
    [sym__uppercase_identifier] = ACTIONS(1730),
    [aux_sym_tag_token1] = ACTIONS(1730),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(805)] = {
    [sym_attribute] = STATE(805),
    [sym__lowercase_identifier] = ACTIONS(1732),
    [anon_sym_and] = ACTIONS(1732),
    [anon_sym_TILDE] = ACTIONS(1732),
    [anon_sym_QMARK] = ACTIONS(1732),
    [anon_sym_LPAREN] = ACTIONS(1734),
    [anon_sym_RPAREN] = ACTIONS(1734),
    [anon_sym_COMMA] = ACTIONS(1734),
    [anon_sym_PIPE] = ACTIONS(1732),
    [anon_sym_COLON2] = ACTIONS(1732),
    [anon_sym_DOT] = ACTIONS(1734),
    [anon_sym_DASH_GT] = ACTIONS(1732),
    [anon_sym_LBRACK] = ACTIONS(1732),
    [anon_sym_RBRACK] = ACTIONS(1734),
    [anon_sym_true] = ACTIONS(1732),
    [anon_sym_false] = ACTIONS(1732),
    [anon_sym_LBRACE] = ACTIONS(1732),
    [anon_sym_SEMI] = ACTIONS(1734),
    [anon_sym_RBRACE] = ACTIONS(1734),
    [anon_sym_constraint] = ACTIONS(1732),
    [anon_sym_val] = ACTIONS(1732),
    [anon_sym_end] = ACTIONS(1732),
    [anon_sym_with] = ACTIONS(1732),
    [anon_sym_COLON_GT] = ACTIONS(1734),
    [anon_sym_object] = ACTIONS(1732),
    [anon_sym_inherit] = ACTIONS(1732),
    [anon_sym_method] = ACTIONS(1732),
    [anon_sym_in] = ACTIONS(1732),
    [anon_sym_as] = ACTIONS(1732),
    [anon_sym_initializer] = ACTIONS(1732),
    [anon_sym_AMP] = ACTIONS(1732),
    [anon_sym_POUND] = ACTIONS(1732),
    [anon_sym_COLON_COLON] = ACTIONS(1734),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1734),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1734),
    [anon_sym_then] = ACTIONS(1732),
    [anon_sym_else] = ACTIONS(1732),
    [anon_sym_do] = ACTIONS(1732),
    [anon_sym_done] = ACTIONS(1732),
    [anon_sym_to] = ACTIONS(1732),
    [anon_sym_downto] = ACTIONS(1732),
    [anon_sym_new] = ACTIONS(1732),
    [anon_sym_LBRACE_LT] = ACTIONS(1734),
    [anon_sym_GT_RBRACE] = ACTIONS(1734),
    [anon_sym_begin] = ACTIONS(1732),
    [sym_ocamlyacc_value] = ACTIONS(1734),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1732),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1734),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1732),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1732),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1734),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1734),
    [aux_sym_number_token1] = ACTIONS(1734),
    [aux_sym_signed_number_token1] = ACTIONS(1732),
    [anon_sym_SQUOTE] = ACTIONS(1734),
    [anon_sym_DQUOTE] = ACTIONS(1734),
    [sym_prefix_operator] = ACTIONS(1734),
    [aux_sym_sign_operator_token1] = ACTIONS(1732),
    [sym_hash_operator] = ACTIONS(1734),
    [anon_sym_lsl] = ACTIONS(1732),
    [anon_sym_lsr] = ACTIONS(1732),
    [anon_sym_asr] = ACTIONS(1732),
    [aux_sym_pow_operator_token1] = ACTIONS(1734),
    [anon_sym_mod] = ACTIONS(1732),
    [anon_sym_land] = ACTIONS(1732),
    [anon_sym_lor] = ACTIONS(1732),
    [anon_sym_lxor] = ACTIONS(1732),
    [aux_sym_mult_operator_token1] = ACTIONS(1732),
    [aux_sym_add_operator_token1] = ACTIONS(1732),
    [sym_concat_operator] = ACTIONS(1734),
    [sym_rel_operator] = ACTIONS(1732),
    [anon_sym_AMP_AMP] = ACTIONS(1732),
    [anon_sym_or] = ACTIONS(1732),
    [anon_sym_PIPE_PIPE] = ACTIONS(1732),
    [sym_assign_operator] = ACTIONS(1734),
    [sym__uppercase_identifier] = ACTIONS(1734),
    [aux_sym_tag_token1] = ACTIONS(1734),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(806)] = {
    [sym_attribute] = STATE(806),
    [sym__lowercase_identifier] = ACTIONS(1736),
    [anon_sym_and] = ACTIONS(1736),
    [anon_sym_TILDE] = ACTIONS(1736),
    [anon_sym_QMARK] = ACTIONS(1736),
    [anon_sym_LPAREN] = ACTIONS(1738),
    [anon_sym_RPAREN] = ACTIONS(1738),
    [anon_sym_COMMA] = ACTIONS(1738),
    [anon_sym_PIPE] = ACTIONS(1736),
    [anon_sym_COLON2] = ACTIONS(1736),
    [anon_sym_DOT] = ACTIONS(1738),
    [anon_sym_DASH_GT] = ACTIONS(1736),
    [anon_sym_LBRACK] = ACTIONS(1736),
    [anon_sym_RBRACK] = ACTIONS(1738),
    [anon_sym_true] = ACTIONS(1736),
    [anon_sym_false] = ACTIONS(1736),
    [anon_sym_LBRACE] = ACTIONS(1736),
    [anon_sym_SEMI] = ACTIONS(1738),
    [anon_sym_RBRACE] = ACTIONS(1738),
    [anon_sym_constraint] = ACTIONS(1736),
    [anon_sym_val] = ACTIONS(1736),
    [anon_sym_end] = ACTIONS(1736),
    [anon_sym_with] = ACTIONS(1736),
    [anon_sym_COLON_GT] = ACTIONS(1738),
    [anon_sym_object] = ACTIONS(1736),
    [anon_sym_inherit] = ACTIONS(1736),
    [anon_sym_method] = ACTIONS(1736),
    [anon_sym_in] = ACTIONS(1736),
    [anon_sym_as] = ACTIONS(1736),
    [anon_sym_initializer] = ACTIONS(1736),
    [anon_sym_AMP] = ACTIONS(1736),
    [anon_sym_POUND] = ACTIONS(1736),
    [anon_sym_COLON_COLON] = ACTIONS(1738),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1738),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1738),
    [anon_sym_then] = ACTIONS(1736),
    [anon_sym_else] = ACTIONS(1736),
    [anon_sym_do] = ACTIONS(1736),
    [anon_sym_done] = ACTIONS(1736),
    [anon_sym_to] = ACTIONS(1736),
    [anon_sym_downto] = ACTIONS(1736),
    [anon_sym_new] = ACTIONS(1736),
    [anon_sym_LBRACE_LT] = ACTIONS(1738),
    [anon_sym_GT_RBRACE] = ACTIONS(1738),
    [anon_sym_begin] = ACTIONS(1736),
    [sym_ocamlyacc_value] = ACTIONS(1738),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1736),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1738),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1736),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1736),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1738),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1738),
    [aux_sym_number_token1] = ACTIONS(1738),
    [aux_sym_signed_number_token1] = ACTIONS(1736),
    [anon_sym_SQUOTE] = ACTIONS(1738),
    [anon_sym_DQUOTE] = ACTIONS(1738),
    [sym_prefix_operator] = ACTIONS(1738),
    [aux_sym_sign_operator_token1] = ACTIONS(1736),
    [sym_hash_operator] = ACTIONS(1738),
    [anon_sym_lsl] = ACTIONS(1736),
    [anon_sym_lsr] = ACTIONS(1736),
    [anon_sym_asr] = ACTIONS(1736),
    [aux_sym_pow_operator_token1] = ACTIONS(1738),
    [anon_sym_mod] = ACTIONS(1736),
    [anon_sym_land] = ACTIONS(1736),
    [anon_sym_lor] = ACTIONS(1736),
    [anon_sym_lxor] = ACTIONS(1736),
    [aux_sym_mult_operator_token1] = ACTIONS(1736),
    [aux_sym_add_operator_token1] = ACTIONS(1736),
    [sym_concat_operator] = ACTIONS(1738),
    [sym_rel_operator] = ACTIONS(1736),
    [anon_sym_AMP_AMP] = ACTIONS(1736),
    [anon_sym_or] = ACTIONS(1736),
    [anon_sym_PIPE_PIPE] = ACTIONS(1736),
    [sym_assign_operator] = ACTIONS(1738),
    [sym__uppercase_identifier] = ACTIONS(1738),
    [aux_sym_tag_token1] = ACTIONS(1738),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(807)] = {
    [sym_attribute] = STATE(807),
    [sym__lowercase_identifier] = ACTIONS(1740),
    [anon_sym_and] = ACTIONS(1740),
    [anon_sym_TILDE] = ACTIONS(1740),
    [anon_sym_QMARK] = ACTIONS(1740),
    [anon_sym_LPAREN] = ACTIONS(1742),
    [anon_sym_RPAREN] = ACTIONS(1742),
    [anon_sym_COMMA] = ACTIONS(1742),
    [anon_sym_PIPE] = ACTIONS(1740),
    [anon_sym_COLON2] = ACTIONS(1740),
    [anon_sym_DOT] = ACTIONS(1742),
    [anon_sym_DASH_GT] = ACTIONS(1740),
    [anon_sym_LBRACK] = ACTIONS(1740),
    [anon_sym_RBRACK] = ACTIONS(1742),
    [anon_sym_true] = ACTIONS(1740),
    [anon_sym_false] = ACTIONS(1740),
    [anon_sym_LBRACE] = ACTIONS(1740),
    [anon_sym_SEMI] = ACTIONS(1742),
    [anon_sym_RBRACE] = ACTIONS(1742),
    [anon_sym_constraint] = ACTIONS(1740),
    [anon_sym_val] = ACTIONS(1740),
    [anon_sym_end] = ACTIONS(1740),
    [anon_sym_with] = ACTIONS(1740),
    [anon_sym_COLON_GT] = ACTIONS(1742),
    [anon_sym_object] = ACTIONS(1740),
    [anon_sym_inherit] = ACTIONS(1740),
    [anon_sym_method] = ACTIONS(1740),
    [anon_sym_in] = ACTIONS(1740),
    [anon_sym_as] = ACTIONS(1740),
    [anon_sym_initializer] = ACTIONS(1740),
    [anon_sym_AMP] = ACTIONS(1740),
    [anon_sym_POUND] = ACTIONS(1740),
    [anon_sym_COLON_COLON] = ACTIONS(1742),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1742),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1742),
    [anon_sym_then] = ACTIONS(1740),
    [anon_sym_else] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1740),
    [anon_sym_done] = ACTIONS(1740),
    [anon_sym_to] = ACTIONS(1740),
    [anon_sym_downto] = ACTIONS(1740),
    [anon_sym_new] = ACTIONS(1740),
    [anon_sym_LBRACE_LT] = ACTIONS(1742),
    [anon_sym_GT_RBRACE] = ACTIONS(1742),
    [anon_sym_begin] = ACTIONS(1740),
    [sym_ocamlyacc_value] = ACTIONS(1742),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1740),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1742),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1740),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1740),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1742),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1742),
    [aux_sym_number_token1] = ACTIONS(1742),
    [aux_sym_signed_number_token1] = ACTIONS(1740),
    [anon_sym_SQUOTE] = ACTIONS(1742),
    [anon_sym_DQUOTE] = ACTIONS(1742),
    [sym_prefix_operator] = ACTIONS(1742),
    [aux_sym_sign_operator_token1] = ACTIONS(1740),
    [sym_hash_operator] = ACTIONS(1742),
    [anon_sym_lsl] = ACTIONS(1740),
    [anon_sym_lsr] = ACTIONS(1740),
    [anon_sym_asr] = ACTIONS(1740),
    [aux_sym_pow_operator_token1] = ACTIONS(1742),
    [anon_sym_mod] = ACTIONS(1740),
    [anon_sym_land] = ACTIONS(1740),
    [anon_sym_lor] = ACTIONS(1740),
    [anon_sym_lxor] = ACTIONS(1740),
    [aux_sym_mult_operator_token1] = ACTIONS(1740),
    [aux_sym_add_operator_token1] = ACTIONS(1740),
    [sym_concat_operator] = ACTIONS(1742),
    [sym_rel_operator] = ACTIONS(1740),
    [anon_sym_AMP_AMP] = ACTIONS(1740),
    [anon_sym_or] = ACTIONS(1740),
    [anon_sym_PIPE_PIPE] = ACTIONS(1740),
    [sym_assign_operator] = ACTIONS(1742),
    [sym__uppercase_identifier] = ACTIONS(1742),
    [aux_sym_tag_token1] = ACTIONS(1742),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(808)] = {
    [sym_attribute] = STATE(808),
    [sym__lowercase_identifier] = ACTIONS(1698),
    [anon_sym_and] = ACTIONS(1698),
    [anon_sym_TILDE] = ACTIONS(1698),
    [anon_sym_QMARK] = ACTIONS(1698),
    [anon_sym_LPAREN] = ACTIONS(1700),
    [anon_sym_RPAREN] = ACTIONS(1700),
    [anon_sym_COMMA] = ACTIONS(1700),
    [anon_sym_PIPE] = ACTIONS(1698),
    [anon_sym_COLON2] = ACTIONS(1698),
    [anon_sym_DOT] = ACTIONS(1744),
    [anon_sym_DASH_GT] = ACTIONS(1698),
    [anon_sym_LBRACK] = ACTIONS(1698),
    [anon_sym_RBRACK] = ACTIONS(1700),
    [anon_sym_true] = ACTIONS(1698),
    [anon_sym_false] = ACTIONS(1698),
    [anon_sym_LBRACE] = ACTIONS(1698),
    [anon_sym_SEMI] = ACTIONS(1700),
    [anon_sym_RBRACE] = ACTIONS(1700),
    [anon_sym_constraint] = ACTIONS(1698),
    [anon_sym_val] = ACTIONS(1698),
    [anon_sym_end] = ACTIONS(1698),
    [anon_sym_with] = ACTIONS(1698),
    [anon_sym_COLON_GT] = ACTIONS(1700),
    [anon_sym_object] = ACTIONS(1698),
    [anon_sym_inherit] = ACTIONS(1698),
    [anon_sym_method] = ACTIONS(1698),
    [anon_sym_in] = ACTIONS(1698),
    [anon_sym_as] = ACTIONS(1698),
    [anon_sym_initializer] = ACTIONS(1698),
    [anon_sym_AMP] = ACTIONS(1698),
    [anon_sym_POUND] = ACTIONS(1698),
    [anon_sym_COLON_COLON] = ACTIONS(1700),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1700),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1700),
    [anon_sym_then] = ACTIONS(1698),
    [anon_sym_else] = ACTIONS(1698),
    [anon_sym_do] = ACTIONS(1698),
    [anon_sym_done] = ACTIONS(1698),
    [anon_sym_to] = ACTIONS(1698),
    [anon_sym_downto] = ACTIONS(1698),
    [anon_sym_new] = ACTIONS(1698),
    [anon_sym_LBRACE_LT] = ACTIONS(1700),
    [anon_sym_GT_RBRACE] = ACTIONS(1700),
    [anon_sym_begin] = ACTIONS(1698),
    [sym_ocamlyacc_value] = ACTIONS(1700),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1698),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1700),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1700),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1700),
    [aux_sym_number_token1] = ACTIONS(1700),
    [aux_sym_signed_number_token1] = ACTIONS(1698),
    [anon_sym_SQUOTE] = ACTIONS(1700),
    [anon_sym_DQUOTE] = ACTIONS(1700),
    [sym_prefix_operator] = ACTIONS(1700),
    [aux_sym_sign_operator_token1] = ACTIONS(1698),
    [sym_hash_operator] = ACTIONS(1700),
    [anon_sym_lsl] = ACTIONS(1698),
    [anon_sym_lsr] = ACTIONS(1698),
    [anon_sym_asr] = ACTIONS(1698),
    [aux_sym_pow_operator_token1] = ACTIONS(1700),
    [anon_sym_mod] = ACTIONS(1698),
    [anon_sym_land] = ACTIONS(1698),
    [anon_sym_lor] = ACTIONS(1698),
    [anon_sym_lxor] = ACTIONS(1698),
    [aux_sym_mult_operator_token1] = ACTIONS(1698),
    [aux_sym_add_operator_token1] = ACTIONS(1698),
    [sym_concat_operator] = ACTIONS(1700),
    [sym_rel_operator] = ACTIONS(1698),
    [anon_sym_AMP_AMP] = ACTIONS(1698),
    [anon_sym_or] = ACTIONS(1698),
    [anon_sym_PIPE_PIPE] = ACTIONS(1698),
    [sym_assign_operator] = ACTIONS(1700),
    [sym__uppercase_identifier] = ACTIONS(1700),
    [aux_sym_tag_token1] = ACTIONS(1700),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(809)] = {
    [sym_attribute] = STATE(809),
    [sym__lowercase_identifier] = ACTIONS(1746),
    [anon_sym_and] = ACTIONS(1746),
    [anon_sym_TILDE] = ACTIONS(1746),
    [anon_sym_QMARK] = ACTIONS(1746),
    [anon_sym_LPAREN] = ACTIONS(1748),
    [anon_sym_RPAREN] = ACTIONS(1748),
    [anon_sym_COMMA] = ACTIONS(1748),
    [anon_sym_PIPE] = ACTIONS(1746),
    [anon_sym_COLON2] = ACTIONS(1746),
    [anon_sym_DOT] = ACTIONS(1748),
    [anon_sym_DASH_GT] = ACTIONS(1746),
    [anon_sym_LBRACK] = ACTIONS(1746),
    [anon_sym_RBRACK] = ACTIONS(1748),
    [anon_sym_true] = ACTIONS(1746),
    [anon_sym_false] = ACTIONS(1746),
    [anon_sym_LBRACE] = ACTIONS(1746),
    [anon_sym_SEMI] = ACTIONS(1748),
    [anon_sym_RBRACE] = ACTIONS(1748),
    [anon_sym_constraint] = ACTIONS(1746),
    [anon_sym_val] = ACTIONS(1746),
    [anon_sym_end] = ACTIONS(1746),
    [anon_sym_with] = ACTIONS(1746),
    [anon_sym_COLON_GT] = ACTIONS(1748),
    [anon_sym_object] = ACTIONS(1746),
    [anon_sym_inherit] = ACTIONS(1746),
    [anon_sym_method] = ACTIONS(1746),
    [anon_sym_in] = ACTIONS(1746),
    [anon_sym_as] = ACTIONS(1746),
    [anon_sym_initializer] = ACTIONS(1746),
    [anon_sym_AMP] = ACTIONS(1746),
    [anon_sym_POUND] = ACTIONS(1746),
    [anon_sym_COLON_COLON] = ACTIONS(1748),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1748),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1748),
    [anon_sym_then] = ACTIONS(1746),
    [anon_sym_else] = ACTIONS(1746),
    [anon_sym_do] = ACTIONS(1746),
    [anon_sym_done] = ACTIONS(1746),
    [anon_sym_to] = ACTIONS(1746),
    [anon_sym_downto] = ACTIONS(1746),
    [anon_sym_new] = ACTIONS(1746),
    [anon_sym_LBRACE_LT] = ACTIONS(1748),
    [anon_sym_GT_RBRACE] = ACTIONS(1748),
    [anon_sym_begin] = ACTIONS(1746),
    [sym_ocamlyacc_value] = ACTIONS(1748),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1746),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1748),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1746),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1746),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1748),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1748),
    [aux_sym_number_token1] = ACTIONS(1748),
    [aux_sym_signed_number_token1] = ACTIONS(1746),
    [anon_sym_SQUOTE] = ACTIONS(1748),
    [anon_sym_DQUOTE] = ACTIONS(1748),
    [sym_prefix_operator] = ACTIONS(1748),
    [aux_sym_sign_operator_token1] = ACTIONS(1746),
    [sym_hash_operator] = ACTIONS(1748),
    [anon_sym_lsl] = ACTIONS(1746),
    [anon_sym_lsr] = ACTIONS(1746),
    [anon_sym_asr] = ACTIONS(1746),
    [aux_sym_pow_operator_token1] = ACTIONS(1748),
    [anon_sym_mod] = ACTIONS(1746),
    [anon_sym_land] = ACTIONS(1746),
    [anon_sym_lor] = ACTIONS(1746),
    [anon_sym_lxor] = ACTIONS(1746),
    [aux_sym_mult_operator_token1] = ACTIONS(1746),
    [aux_sym_add_operator_token1] = ACTIONS(1746),
    [sym_concat_operator] = ACTIONS(1748),
    [sym_rel_operator] = ACTIONS(1746),
    [anon_sym_AMP_AMP] = ACTIONS(1746),
    [anon_sym_or] = ACTIONS(1746),
    [anon_sym_PIPE_PIPE] = ACTIONS(1746),
    [sym_assign_operator] = ACTIONS(1748),
    [sym__uppercase_identifier] = ACTIONS(1748),
    [aux_sym_tag_token1] = ACTIONS(1748),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(810)] = {
    [sym_attribute] = STATE(810),
    [sym__lowercase_identifier] = ACTIONS(1750),
    [anon_sym_and] = ACTIONS(1750),
    [anon_sym_TILDE] = ACTIONS(1750),
    [anon_sym_QMARK] = ACTIONS(1750),
    [anon_sym_LPAREN] = ACTIONS(1752),
    [anon_sym_RPAREN] = ACTIONS(1752),
    [anon_sym_COMMA] = ACTIONS(1752),
    [anon_sym_PIPE] = ACTIONS(1750),
    [anon_sym_COLON2] = ACTIONS(1750),
    [anon_sym_DOT] = ACTIONS(1752),
    [anon_sym_DASH_GT] = ACTIONS(1750),
    [anon_sym_LBRACK] = ACTIONS(1750),
    [anon_sym_RBRACK] = ACTIONS(1752),
    [anon_sym_true] = ACTIONS(1750),
    [anon_sym_false] = ACTIONS(1750),
    [anon_sym_LBRACE] = ACTIONS(1750),
    [anon_sym_SEMI] = ACTIONS(1752),
    [anon_sym_RBRACE] = ACTIONS(1752),
    [anon_sym_constraint] = ACTIONS(1750),
    [anon_sym_val] = ACTIONS(1750),
    [anon_sym_end] = ACTIONS(1750),
    [anon_sym_with] = ACTIONS(1750),
    [anon_sym_COLON_GT] = ACTIONS(1752),
    [anon_sym_object] = ACTIONS(1750),
    [anon_sym_inherit] = ACTIONS(1750),
    [anon_sym_method] = ACTIONS(1750),
    [anon_sym_in] = ACTIONS(1750),
    [anon_sym_as] = ACTIONS(1750),
    [anon_sym_initializer] = ACTIONS(1750),
    [anon_sym_AMP] = ACTIONS(1750),
    [anon_sym_POUND] = ACTIONS(1750),
    [anon_sym_COLON_COLON] = ACTIONS(1752),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1752),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1752),
    [anon_sym_then] = ACTIONS(1750),
    [anon_sym_else] = ACTIONS(1750),
    [anon_sym_do] = ACTIONS(1750),
    [anon_sym_done] = ACTIONS(1750),
    [anon_sym_to] = ACTIONS(1750),
    [anon_sym_downto] = ACTIONS(1750),
    [anon_sym_new] = ACTIONS(1750),
    [anon_sym_LBRACE_LT] = ACTIONS(1752),
    [anon_sym_GT_RBRACE] = ACTIONS(1752),
    [anon_sym_begin] = ACTIONS(1750),
    [sym_ocamlyacc_value] = ACTIONS(1752),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1750),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1752),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1750),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1750),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1752),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1752),
    [aux_sym_number_token1] = ACTIONS(1752),
    [aux_sym_signed_number_token1] = ACTIONS(1750),
    [anon_sym_SQUOTE] = ACTIONS(1752),
    [anon_sym_DQUOTE] = ACTIONS(1752),
    [sym_prefix_operator] = ACTIONS(1752),
    [aux_sym_sign_operator_token1] = ACTIONS(1750),
    [sym_hash_operator] = ACTIONS(1752),
    [anon_sym_lsl] = ACTIONS(1750),
    [anon_sym_lsr] = ACTIONS(1750),
    [anon_sym_asr] = ACTIONS(1750),
    [aux_sym_pow_operator_token1] = ACTIONS(1752),
    [anon_sym_mod] = ACTIONS(1750),
    [anon_sym_land] = ACTIONS(1750),
    [anon_sym_lor] = ACTIONS(1750),
    [anon_sym_lxor] = ACTIONS(1750),
    [aux_sym_mult_operator_token1] = ACTIONS(1750),
    [aux_sym_add_operator_token1] = ACTIONS(1750),
    [sym_concat_operator] = ACTIONS(1752),
    [sym_rel_operator] = ACTIONS(1750),
    [anon_sym_AMP_AMP] = ACTIONS(1750),
    [anon_sym_or] = ACTIONS(1750),
    [anon_sym_PIPE_PIPE] = ACTIONS(1750),
    [sym_assign_operator] = ACTIONS(1752),
    [sym__uppercase_identifier] = ACTIONS(1752),
    [aux_sym_tag_token1] = ACTIONS(1752),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(811)] = {
    [sym_attribute] = STATE(811),
    [sym__lowercase_identifier] = ACTIONS(1754),
    [anon_sym_and] = ACTIONS(1754),
    [anon_sym_TILDE] = ACTIONS(1754),
    [anon_sym_QMARK] = ACTIONS(1754),
    [anon_sym_LPAREN] = ACTIONS(1756),
    [anon_sym_RPAREN] = ACTIONS(1756),
    [anon_sym_COMMA] = ACTIONS(1756),
    [anon_sym_PIPE] = ACTIONS(1754),
    [anon_sym_COLON2] = ACTIONS(1754),
    [anon_sym_DOT] = ACTIONS(1756),
    [anon_sym_DASH_GT] = ACTIONS(1754),
    [anon_sym_LBRACK] = ACTIONS(1754),
    [anon_sym_RBRACK] = ACTIONS(1756),
    [anon_sym_true] = ACTIONS(1754),
    [anon_sym_false] = ACTIONS(1754),
    [anon_sym_LBRACE] = ACTIONS(1754),
    [anon_sym_SEMI] = ACTIONS(1756),
    [anon_sym_RBRACE] = ACTIONS(1756),
    [anon_sym_constraint] = ACTIONS(1754),
    [anon_sym_val] = ACTIONS(1754),
    [anon_sym_end] = ACTIONS(1754),
    [anon_sym_with] = ACTIONS(1754),
    [anon_sym_COLON_GT] = ACTIONS(1756),
    [anon_sym_object] = ACTIONS(1754),
    [anon_sym_inherit] = ACTIONS(1754),
    [anon_sym_method] = ACTIONS(1754),
    [anon_sym_in] = ACTIONS(1754),
    [anon_sym_as] = ACTIONS(1754),
    [anon_sym_initializer] = ACTIONS(1754),
    [anon_sym_AMP] = ACTIONS(1754),
    [anon_sym_POUND] = ACTIONS(1754),
    [anon_sym_COLON_COLON] = ACTIONS(1756),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1756),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1756),
    [anon_sym_then] = ACTIONS(1754),
    [anon_sym_else] = ACTIONS(1754),
    [anon_sym_do] = ACTIONS(1754),
    [anon_sym_done] = ACTIONS(1754),
    [anon_sym_to] = ACTIONS(1754),
    [anon_sym_downto] = ACTIONS(1754),
    [anon_sym_new] = ACTIONS(1754),
    [anon_sym_LBRACE_LT] = ACTIONS(1756),
    [anon_sym_GT_RBRACE] = ACTIONS(1756),
    [anon_sym_begin] = ACTIONS(1754),
    [sym_ocamlyacc_value] = ACTIONS(1756),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1754),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1756),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1754),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1754),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1756),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1756),
    [aux_sym_number_token1] = ACTIONS(1756),
    [aux_sym_signed_number_token1] = ACTIONS(1754),
    [anon_sym_SQUOTE] = ACTIONS(1756),
    [anon_sym_DQUOTE] = ACTIONS(1756),
    [sym_prefix_operator] = ACTIONS(1756),
    [aux_sym_sign_operator_token1] = ACTIONS(1754),
    [sym_hash_operator] = ACTIONS(1756),
    [anon_sym_lsl] = ACTIONS(1754),
    [anon_sym_lsr] = ACTIONS(1754),
    [anon_sym_asr] = ACTIONS(1754),
    [aux_sym_pow_operator_token1] = ACTIONS(1756),
    [anon_sym_mod] = ACTIONS(1754),
    [anon_sym_land] = ACTIONS(1754),
    [anon_sym_lor] = ACTIONS(1754),
    [anon_sym_lxor] = ACTIONS(1754),
    [aux_sym_mult_operator_token1] = ACTIONS(1754),
    [aux_sym_add_operator_token1] = ACTIONS(1754),
    [sym_concat_operator] = ACTIONS(1756),
    [sym_rel_operator] = ACTIONS(1754),
    [anon_sym_AMP_AMP] = ACTIONS(1754),
    [anon_sym_or] = ACTIONS(1754),
    [anon_sym_PIPE_PIPE] = ACTIONS(1754),
    [sym_assign_operator] = ACTIONS(1756),
    [sym__uppercase_identifier] = ACTIONS(1756),
    [aux_sym_tag_token1] = ACTIONS(1756),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(812)] = {
    [sym_attribute] = STATE(812),
    [sym__lowercase_identifier] = ACTIONS(1758),
    [anon_sym_and] = ACTIONS(1758),
    [anon_sym_TILDE] = ACTIONS(1758),
    [anon_sym_QMARK] = ACTIONS(1758),
    [anon_sym_LPAREN] = ACTIONS(1760),
    [anon_sym_RPAREN] = ACTIONS(1760),
    [anon_sym_COMMA] = ACTIONS(1760),
    [anon_sym_PIPE] = ACTIONS(1758),
    [anon_sym_COLON2] = ACTIONS(1758),
    [anon_sym_DOT] = ACTIONS(1760),
    [anon_sym_DASH_GT] = ACTIONS(1758),
    [anon_sym_LBRACK] = ACTIONS(1758),
    [anon_sym_RBRACK] = ACTIONS(1760),
    [anon_sym_true] = ACTIONS(1758),
    [anon_sym_false] = ACTIONS(1758),
    [anon_sym_LBRACE] = ACTIONS(1758),
    [anon_sym_SEMI] = ACTIONS(1760),
    [anon_sym_RBRACE] = ACTIONS(1760),
    [anon_sym_constraint] = ACTIONS(1758),
    [anon_sym_val] = ACTIONS(1758),
    [anon_sym_end] = ACTIONS(1758),
    [anon_sym_with] = ACTIONS(1758),
    [anon_sym_COLON_GT] = ACTIONS(1760),
    [anon_sym_object] = ACTIONS(1758),
    [anon_sym_inherit] = ACTIONS(1758),
    [anon_sym_method] = ACTIONS(1758),
    [anon_sym_in] = ACTIONS(1758),
    [anon_sym_as] = ACTIONS(1758),
    [anon_sym_initializer] = ACTIONS(1758),
    [anon_sym_AMP] = ACTIONS(1758),
    [anon_sym_POUND] = ACTIONS(1758),
    [anon_sym_COLON_COLON] = ACTIONS(1760),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1760),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1760),
    [anon_sym_then] = ACTIONS(1758),
    [anon_sym_else] = ACTIONS(1758),
    [anon_sym_do] = ACTIONS(1758),
    [anon_sym_done] = ACTIONS(1758),
    [anon_sym_to] = ACTIONS(1758),
    [anon_sym_downto] = ACTIONS(1758),
    [anon_sym_new] = ACTIONS(1758),
    [anon_sym_LBRACE_LT] = ACTIONS(1760),
    [anon_sym_GT_RBRACE] = ACTIONS(1760),
    [anon_sym_begin] = ACTIONS(1758),
    [sym_ocamlyacc_value] = ACTIONS(1760),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1758),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1760),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1758),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1758),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1760),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1760),
    [aux_sym_number_token1] = ACTIONS(1760),
    [aux_sym_signed_number_token1] = ACTIONS(1758),
    [anon_sym_SQUOTE] = ACTIONS(1760),
    [anon_sym_DQUOTE] = ACTIONS(1760),
    [sym_prefix_operator] = ACTIONS(1760),
    [aux_sym_sign_operator_token1] = ACTIONS(1758),
    [sym_hash_operator] = ACTIONS(1760),
    [anon_sym_lsl] = ACTIONS(1758),
    [anon_sym_lsr] = ACTIONS(1758),
    [anon_sym_asr] = ACTIONS(1758),
    [aux_sym_pow_operator_token1] = ACTIONS(1760),
    [anon_sym_mod] = ACTIONS(1758),
    [anon_sym_land] = ACTIONS(1758),
    [anon_sym_lor] = ACTIONS(1758),
    [anon_sym_lxor] = ACTIONS(1758),
    [aux_sym_mult_operator_token1] = ACTIONS(1758),
    [aux_sym_add_operator_token1] = ACTIONS(1758),
    [sym_concat_operator] = ACTIONS(1760),
    [sym_rel_operator] = ACTIONS(1758),
    [anon_sym_AMP_AMP] = ACTIONS(1758),
    [anon_sym_or] = ACTIONS(1758),
    [anon_sym_PIPE_PIPE] = ACTIONS(1758),
    [sym_assign_operator] = ACTIONS(1760),
    [sym__uppercase_identifier] = ACTIONS(1760),
    [aux_sym_tag_token1] = ACTIONS(1760),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(813)] = {
    [sym__delimited_expression] = STATE(2350),
    [sym__simple_expression] = STATE(2404),
    [sym_typed_expression] = STATE(2350),
    [sym_list_expression] = STATE(2351),
    [sym_array_expression] = STATE(2351),
    [sym_record_expression] = STATE(2351),
    [sym_labeled_argument] = STATE(2522),
    [sym_prefix_expression] = STATE(2350),
    [sym_hash_expression] = STATE(2350),
    [sym_field_get_expression] = STATE(2350),
    [sym_array_get_expression] = STATE(2350),
    [sym_string_get_expression] = STATE(2350),
    [sym_bigarray_get_expression] = STATE(2350),
    [sym_local_open_expression] = STATE(2350),
    [sym_package_expression] = STATE(2351),
    [sym_new_expression] = STATE(2350),
    [sym_object_copy_expression] = STATE(2351),
    [sym_method_invocation] = STATE(2350),
    [sym_object_expression] = STATE(2350),
    [sym_parenthesized_expression] = STATE(2351),
    [sym_attribute] = STATE(813),
    [sym_extension] = STATE(2350),
    [sym_quoted_extension] = STATE(2350),
    [sym__constant] = STATE(2350),
    [sym_number] = STATE(2348),
    [sym_character] = STATE(2348),
    [sym_string] = STATE(2348),
    [sym_quoted_string] = STATE(2348),
    [sym_boolean] = STATE(2348),
    [sym_unit] = STATE(2351),
    [sym__value_name] = STATE(2402),
    [sym_parenthesized_operator] = STATE(2406),
    [sym_value_path] = STATE(2350),
    [sym_module_path] = STATE(6407),
    [sym_constructor_path] = STATE(2350),
    [sym__constructor_name] = STATE(2419),
    [sym_tag] = STATE(2350),
    [aux_sym_class_application_repeat1] = STATE(813),
    [ts_builtin_sym_end] = ACTIONS(784),
    [sym__lowercase_identifier] = ACTIONS(1762),
    [anon_sym_SEMI_SEMI] = ACTIONS(784),
    [anon_sym_and] = ACTIONS(786),
    [anon_sym_TILDE] = ACTIONS(1765),
    [anon_sym_QMARK] = ACTIONS(1765),
    [anon_sym_LPAREN] = ACTIONS(1768),
    [anon_sym_external] = ACTIONS(786),
    [anon_sym_type] = ACTIONS(786),
    [anon_sym_LBRACK] = ACTIONS(1771),
    [anon_sym_RBRACK] = ACTIONS(784),
    [anon_sym_true] = ACTIONS(1774),
    [anon_sym_false] = ACTIONS(1774),
    [anon_sym_LBRACE] = ACTIONS(1777),
    [anon_sym_exception] = ACTIONS(786),
    [anon_sym_module] = ACTIONS(786),
    [anon_sym_open] = ACTIONS(786),
    [anon_sym_include] = ACTIONS(786),
    [anon_sym_class] = ACTIONS(786),
    [anon_sym_val] = ACTIONS(786),
    [anon_sym_end] = ACTIONS(786),
    [anon_sym_object] = ACTIONS(1780),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_new] = ACTIONS(1786),
    [anon_sym_LBRACE_LT] = ACTIONS(1789),
    [anon_sym_begin] = ACTIONS(1792),
    [sym_ocamlyacc_value] = ACTIONS(1795),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(786),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(784),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1798),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1801),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(784),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(784),
    [aux_sym_number_token1] = ACTIONS(1804),
    [anon_sym_SQUOTE] = ACTIONS(1807),
    [anon_sym_DQUOTE] = ACTIONS(1810),
    [sym_prefix_operator] = ACTIONS(1813),
    [sym__uppercase_identifier] = ACTIONS(1816),
    [aux_sym_tag_token1] = ACTIONS(1819),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(814)] = {
    [sym_attribute] = STATE(814),
    [sym__lowercase_identifier] = ACTIONS(1822),
    [anon_sym_and] = ACTIONS(1822),
    [anon_sym_TILDE] = ACTIONS(1822),
    [anon_sym_QMARK] = ACTIONS(1822),
    [anon_sym_LPAREN] = ACTIONS(1824),
    [anon_sym_RPAREN] = ACTIONS(1824),
    [anon_sym_COMMA] = ACTIONS(1824),
    [anon_sym_PIPE] = ACTIONS(1822),
    [anon_sym_COLON2] = ACTIONS(1822),
    [anon_sym_DOT] = ACTIONS(1824),
    [anon_sym_DASH_GT] = ACTIONS(1822),
    [anon_sym_LBRACK] = ACTIONS(1822),
    [anon_sym_RBRACK] = ACTIONS(1824),
    [anon_sym_true] = ACTIONS(1822),
    [anon_sym_false] = ACTIONS(1822),
    [anon_sym_LBRACE] = ACTIONS(1822),
    [anon_sym_SEMI] = ACTIONS(1824),
    [anon_sym_RBRACE] = ACTIONS(1824),
    [anon_sym_constraint] = ACTIONS(1822),
    [anon_sym_val] = ACTIONS(1822),
    [anon_sym_end] = ACTIONS(1822),
    [anon_sym_with] = ACTIONS(1822),
    [anon_sym_COLON_GT] = ACTIONS(1824),
    [anon_sym_object] = ACTIONS(1822),
    [anon_sym_inherit] = ACTIONS(1822),
    [anon_sym_method] = ACTIONS(1822),
    [anon_sym_in] = ACTIONS(1822),
    [anon_sym_as] = ACTIONS(1822),
    [anon_sym_initializer] = ACTIONS(1822),
    [anon_sym_AMP] = ACTIONS(1822),
    [anon_sym_POUND] = ACTIONS(1822),
    [anon_sym_COLON_COLON] = ACTIONS(1824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1824),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1824),
    [anon_sym_then] = ACTIONS(1822),
    [anon_sym_else] = ACTIONS(1822),
    [anon_sym_do] = ACTIONS(1822),
    [anon_sym_done] = ACTIONS(1822),
    [anon_sym_to] = ACTIONS(1822),
    [anon_sym_downto] = ACTIONS(1822),
    [anon_sym_new] = ACTIONS(1822),
    [anon_sym_LBRACE_LT] = ACTIONS(1824),
    [anon_sym_GT_RBRACE] = ACTIONS(1824),
    [anon_sym_begin] = ACTIONS(1822),
    [sym_ocamlyacc_value] = ACTIONS(1824),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1822),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1824),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1822),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1822),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1824),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1824),
    [aux_sym_number_token1] = ACTIONS(1824),
    [aux_sym_signed_number_token1] = ACTIONS(1822),
    [anon_sym_SQUOTE] = ACTIONS(1824),
    [anon_sym_DQUOTE] = ACTIONS(1824),
    [sym_prefix_operator] = ACTIONS(1824),
    [aux_sym_sign_operator_token1] = ACTIONS(1822),
    [sym_hash_operator] = ACTIONS(1824),
    [anon_sym_lsl] = ACTIONS(1822),
    [anon_sym_lsr] = ACTIONS(1822),
    [anon_sym_asr] = ACTIONS(1822),
    [aux_sym_pow_operator_token1] = ACTIONS(1824),
    [anon_sym_mod] = ACTIONS(1822),
    [anon_sym_land] = ACTIONS(1822),
    [anon_sym_lor] = ACTIONS(1822),
    [anon_sym_lxor] = ACTIONS(1822),
    [aux_sym_mult_operator_token1] = ACTIONS(1822),
    [aux_sym_add_operator_token1] = ACTIONS(1822),
    [sym_concat_operator] = ACTIONS(1824),
    [sym_rel_operator] = ACTIONS(1822),
    [anon_sym_AMP_AMP] = ACTIONS(1822),
    [anon_sym_or] = ACTIONS(1822),
    [anon_sym_PIPE_PIPE] = ACTIONS(1822),
    [sym_assign_operator] = ACTIONS(1824),
    [sym__uppercase_identifier] = ACTIONS(1824),
    [aux_sym_tag_token1] = ACTIONS(1824),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(815)] = {
    [sym_attribute] = STATE(815),
    [sym__lowercase_identifier] = ACTIONS(1826),
    [anon_sym_and] = ACTIONS(1826),
    [anon_sym_TILDE] = ACTIONS(1826),
    [anon_sym_QMARK] = ACTIONS(1826),
    [anon_sym_LPAREN] = ACTIONS(1828),
    [anon_sym_RPAREN] = ACTIONS(1828),
    [anon_sym_COMMA] = ACTIONS(1828),
    [anon_sym_PIPE] = ACTIONS(1826),
    [anon_sym_COLON2] = ACTIONS(1826),
    [anon_sym_DOT] = ACTIONS(1828),
    [anon_sym_DASH_GT] = ACTIONS(1826),
    [anon_sym_LBRACK] = ACTIONS(1826),
    [anon_sym_RBRACK] = ACTIONS(1828),
    [anon_sym_true] = ACTIONS(1826),
    [anon_sym_false] = ACTIONS(1826),
    [anon_sym_LBRACE] = ACTIONS(1826),
    [anon_sym_SEMI] = ACTIONS(1828),
    [anon_sym_RBRACE] = ACTIONS(1828),
    [anon_sym_constraint] = ACTIONS(1826),
    [anon_sym_val] = ACTIONS(1826),
    [anon_sym_end] = ACTIONS(1826),
    [anon_sym_with] = ACTIONS(1826),
    [anon_sym_COLON_GT] = ACTIONS(1828),
    [anon_sym_object] = ACTIONS(1826),
    [anon_sym_inherit] = ACTIONS(1826),
    [anon_sym_method] = ACTIONS(1826),
    [anon_sym_in] = ACTIONS(1826),
    [anon_sym_as] = ACTIONS(1826),
    [anon_sym_initializer] = ACTIONS(1826),
    [anon_sym_AMP] = ACTIONS(1826),
    [anon_sym_POUND] = ACTIONS(1826),
    [anon_sym_COLON_COLON] = ACTIONS(1828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1828),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1828),
    [anon_sym_then] = ACTIONS(1826),
    [anon_sym_else] = ACTIONS(1826),
    [anon_sym_do] = ACTIONS(1826),
    [anon_sym_done] = ACTIONS(1826),
    [anon_sym_to] = ACTIONS(1826),
    [anon_sym_downto] = ACTIONS(1826),
    [anon_sym_new] = ACTIONS(1826),
    [anon_sym_LBRACE_LT] = ACTIONS(1828),
    [anon_sym_GT_RBRACE] = ACTIONS(1828),
    [anon_sym_begin] = ACTIONS(1826),
    [sym_ocamlyacc_value] = ACTIONS(1828),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1826),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1828),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1826),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1826),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1828),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1828),
    [aux_sym_number_token1] = ACTIONS(1828),
    [aux_sym_signed_number_token1] = ACTIONS(1826),
    [anon_sym_SQUOTE] = ACTIONS(1828),
    [anon_sym_DQUOTE] = ACTIONS(1828),
    [sym_prefix_operator] = ACTIONS(1828),
    [aux_sym_sign_operator_token1] = ACTIONS(1826),
    [sym_hash_operator] = ACTIONS(1828),
    [anon_sym_lsl] = ACTIONS(1826),
    [anon_sym_lsr] = ACTIONS(1826),
    [anon_sym_asr] = ACTIONS(1826),
    [aux_sym_pow_operator_token1] = ACTIONS(1828),
    [anon_sym_mod] = ACTIONS(1826),
    [anon_sym_land] = ACTIONS(1826),
    [anon_sym_lor] = ACTIONS(1826),
    [anon_sym_lxor] = ACTIONS(1826),
    [aux_sym_mult_operator_token1] = ACTIONS(1826),
    [aux_sym_add_operator_token1] = ACTIONS(1826),
    [sym_concat_operator] = ACTIONS(1828),
    [sym_rel_operator] = ACTIONS(1826),
    [anon_sym_AMP_AMP] = ACTIONS(1826),
    [anon_sym_or] = ACTIONS(1826),
    [anon_sym_PIPE_PIPE] = ACTIONS(1826),
    [sym_assign_operator] = ACTIONS(1828),
    [sym__uppercase_identifier] = ACTIONS(1828),
    [aux_sym_tag_token1] = ACTIONS(1828),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(816)] = {
    [sym_attribute] = STATE(816),
    [sym__lowercase_identifier] = ACTIONS(1830),
    [anon_sym_and] = ACTIONS(1830),
    [anon_sym_TILDE] = ACTIONS(1830),
    [anon_sym_QMARK] = ACTIONS(1830),
    [anon_sym_LPAREN] = ACTIONS(1832),
    [anon_sym_RPAREN] = ACTIONS(1832),
    [anon_sym_COMMA] = ACTIONS(1832),
    [anon_sym_PIPE] = ACTIONS(1830),
    [anon_sym_COLON2] = ACTIONS(1830),
    [anon_sym_DOT] = ACTIONS(1832),
    [anon_sym_DASH_GT] = ACTIONS(1830),
    [anon_sym_LBRACK] = ACTIONS(1830),
    [anon_sym_RBRACK] = ACTIONS(1832),
    [anon_sym_true] = ACTIONS(1830),
    [anon_sym_false] = ACTIONS(1830),
    [anon_sym_LBRACE] = ACTIONS(1830),
    [anon_sym_SEMI] = ACTIONS(1832),
    [anon_sym_RBRACE] = ACTIONS(1832),
    [anon_sym_constraint] = ACTIONS(1830),
    [anon_sym_val] = ACTIONS(1830),
    [anon_sym_end] = ACTIONS(1830),
    [anon_sym_with] = ACTIONS(1830),
    [anon_sym_COLON_GT] = ACTIONS(1832),
    [anon_sym_object] = ACTIONS(1830),
    [anon_sym_inherit] = ACTIONS(1830),
    [anon_sym_method] = ACTIONS(1830),
    [anon_sym_in] = ACTIONS(1830),
    [anon_sym_as] = ACTIONS(1830),
    [anon_sym_initializer] = ACTIONS(1830),
    [anon_sym_AMP] = ACTIONS(1830),
    [anon_sym_POUND] = ACTIONS(1830),
    [anon_sym_COLON_COLON] = ACTIONS(1832),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1832),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1832),
    [anon_sym_then] = ACTIONS(1830),
    [anon_sym_else] = ACTIONS(1830),
    [anon_sym_do] = ACTIONS(1830),
    [anon_sym_done] = ACTIONS(1830),
    [anon_sym_to] = ACTIONS(1830),
    [anon_sym_downto] = ACTIONS(1830),
    [anon_sym_new] = ACTIONS(1830),
    [anon_sym_LBRACE_LT] = ACTIONS(1832),
    [anon_sym_GT_RBRACE] = ACTIONS(1832),
    [anon_sym_begin] = ACTIONS(1830),
    [sym_ocamlyacc_value] = ACTIONS(1832),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1830),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1832),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1830),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1830),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1832),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1832),
    [aux_sym_number_token1] = ACTIONS(1832),
    [aux_sym_signed_number_token1] = ACTIONS(1830),
    [anon_sym_SQUOTE] = ACTIONS(1832),
    [anon_sym_DQUOTE] = ACTIONS(1832),
    [sym_prefix_operator] = ACTIONS(1832),
    [aux_sym_sign_operator_token1] = ACTIONS(1830),
    [sym_hash_operator] = ACTIONS(1832),
    [anon_sym_lsl] = ACTIONS(1830),
    [anon_sym_lsr] = ACTIONS(1830),
    [anon_sym_asr] = ACTIONS(1830),
    [aux_sym_pow_operator_token1] = ACTIONS(1832),
    [anon_sym_mod] = ACTIONS(1830),
    [anon_sym_land] = ACTIONS(1830),
    [anon_sym_lor] = ACTIONS(1830),
    [anon_sym_lxor] = ACTIONS(1830),
    [aux_sym_mult_operator_token1] = ACTIONS(1830),
    [aux_sym_add_operator_token1] = ACTIONS(1830),
    [sym_concat_operator] = ACTIONS(1832),
    [sym_rel_operator] = ACTIONS(1830),
    [anon_sym_AMP_AMP] = ACTIONS(1830),
    [anon_sym_or] = ACTIONS(1830),
    [anon_sym_PIPE_PIPE] = ACTIONS(1830),
    [sym_assign_operator] = ACTIONS(1832),
    [sym__uppercase_identifier] = ACTIONS(1832),
    [aux_sym_tag_token1] = ACTIONS(1832),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(817)] = {
    [sym_attribute] = STATE(817),
    [sym__lowercase_identifier] = ACTIONS(1834),
    [anon_sym_and] = ACTIONS(1834),
    [anon_sym_TILDE] = ACTIONS(1834),
    [anon_sym_QMARK] = ACTIONS(1834),
    [anon_sym_LPAREN] = ACTIONS(1836),
    [anon_sym_RPAREN] = ACTIONS(1836),
    [anon_sym_COMMA] = ACTIONS(1836),
    [anon_sym_PIPE] = ACTIONS(1834),
    [anon_sym_COLON2] = ACTIONS(1834),
    [anon_sym_DOT] = ACTIONS(1836),
    [anon_sym_DASH_GT] = ACTIONS(1834),
    [anon_sym_LBRACK] = ACTIONS(1834),
    [anon_sym_RBRACK] = ACTIONS(1836),
    [anon_sym_true] = ACTIONS(1834),
    [anon_sym_false] = ACTIONS(1834),
    [anon_sym_LBRACE] = ACTIONS(1834),
    [anon_sym_SEMI] = ACTIONS(1836),
    [anon_sym_RBRACE] = ACTIONS(1836),
    [anon_sym_constraint] = ACTIONS(1834),
    [anon_sym_val] = ACTIONS(1834),
    [anon_sym_end] = ACTIONS(1834),
    [anon_sym_with] = ACTIONS(1834),
    [anon_sym_COLON_GT] = ACTIONS(1836),
    [anon_sym_object] = ACTIONS(1834),
    [anon_sym_inherit] = ACTIONS(1834),
    [anon_sym_method] = ACTIONS(1834),
    [anon_sym_in] = ACTIONS(1834),
    [anon_sym_as] = ACTIONS(1834),
    [anon_sym_initializer] = ACTIONS(1834),
    [anon_sym_AMP] = ACTIONS(1834),
    [anon_sym_POUND] = ACTIONS(1834),
    [anon_sym_COLON_COLON] = ACTIONS(1836),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1836),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1836),
    [anon_sym_then] = ACTIONS(1834),
    [anon_sym_else] = ACTIONS(1834),
    [anon_sym_do] = ACTIONS(1834),
    [anon_sym_done] = ACTIONS(1834),
    [anon_sym_to] = ACTIONS(1834),
    [anon_sym_downto] = ACTIONS(1834),
    [anon_sym_new] = ACTIONS(1834),
    [anon_sym_LBRACE_LT] = ACTIONS(1836),
    [anon_sym_GT_RBRACE] = ACTIONS(1836),
    [anon_sym_begin] = ACTIONS(1834),
    [sym_ocamlyacc_value] = ACTIONS(1836),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1834),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1836),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1834),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1834),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1836),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1836),
    [aux_sym_number_token1] = ACTIONS(1836),
    [aux_sym_signed_number_token1] = ACTIONS(1834),
    [anon_sym_SQUOTE] = ACTIONS(1836),
    [anon_sym_DQUOTE] = ACTIONS(1836),
    [sym_prefix_operator] = ACTIONS(1836),
    [aux_sym_sign_operator_token1] = ACTIONS(1834),
    [sym_hash_operator] = ACTIONS(1836),
    [anon_sym_lsl] = ACTIONS(1834),
    [anon_sym_lsr] = ACTIONS(1834),
    [anon_sym_asr] = ACTIONS(1834),
    [aux_sym_pow_operator_token1] = ACTIONS(1836),
    [anon_sym_mod] = ACTIONS(1834),
    [anon_sym_land] = ACTIONS(1834),
    [anon_sym_lor] = ACTIONS(1834),
    [anon_sym_lxor] = ACTIONS(1834),
    [aux_sym_mult_operator_token1] = ACTIONS(1834),
    [aux_sym_add_operator_token1] = ACTIONS(1834),
    [sym_concat_operator] = ACTIONS(1836),
    [sym_rel_operator] = ACTIONS(1834),
    [anon_sym_AMP_AMP] = ACTIONS(1834),
    [anon_sym_or] = ACTIONS(1834),
    [anon_sym_PIPE_PIPE] = ACTIONS(1834),
    [sym_assign_operator] = ACTIONS(1836),
    [sym__uppercase_identifier] = ACTIONS(1836),
    [aux_sym_tag_token1] = ACTIONS(1836),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(818)] = {
    [sym_attribute] = STATE(818),
    [sym__lowercase_identifier] = ACTIONS(1838),
    [anon_sym_and] = ACTIONS(1838),
    [anon_sym_TILDE] = ACTIONS(1838),
    [anon_sym_QMARK] = ACTIONS(1838),
    [anon_sym_LPAREN] = ACTIONS(1840),
    [anon_sym_RPAREN] = ACTIONS(1840),
    [anon_sym_COMMA] = ACTIONS(1840),
    [anon_sym_PIPE] = ACTIONS(1838),
    [anon_sym_COLON2] = ACTIONS(1838),
    [anon_sym_DOT] = ACTIONS(1840),
    [anon_sym_DASH_GT] = ACTIONS(1838),
    [anon_sym_LBRACK] = ACTIONS(1838),
    [anon_sym_RBRACK] = ACTIONS(1840),
    [anon_sym_true] = ACTIONS(1838),
    [anon_sym_false] = ACTIONS(1838),
    [anon_sym_LBRACE] = ACTIONS(1838),
    [anon_sym_SEMI] = ACTIONS(1840),
    [anon_sym_RBRACE] = ACTIONS(1840),
    [anon_sym_constraint] = ACTIONS(1838),
    [anon_sym_val] = ACTIONS(1838),
    [anon_sym_end] = ACTIONS(1838),
    [anon_sym_with] = ACTIONS(1838),
    [anon_sym_COLON_GT] = ACTIONS(1840),
    [anon_sym_object] = ACTIONS(1838),
    [anon_sym_inherit] = ACTIONS(1838),
    [anon_sym_method] = ACTIONS(1838),
    [anon_sym_in] = ACTIONS(1838),
    [anon_sym_as] = ACTIONS(1838),
    [anon_sym_initializer] = ACTIONS(1838),
    [anon_sym_AMP] = ACTIONS(1838),
    [anon_sym_POUND] = ACTIONS(1838),
    [anon_sym_COLON_COLON] = ACTIONS(1840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1840),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1840),
    [anon_sym_then] = ACTIONS(1838),
    [anon_sym_else] = ACTIONS(1838),
    [anon_sym_do] = ACTIONS(1838),
    [anon_sym_done] = ACTIONS(1838),
    [anon_sym_to] = ACTIONS(1838),
    [anon_sym_downto] = ACTIONS(1838),
    [anon_sym_new] = ACTIONS(1838),
    [anon_sym_LBRACE_LT] = ACTIONS(1840),
    [anon_sym_GT_RBRACE] = ACTIONS(1840),
    [anon_sym_begin] = ACTIONS(1838),
    [sym_ocamlyacc_value] = ACTIONS(1840),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1838),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1840),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1838),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1838),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1840),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1840),
    [aux_sym_number_token1] = ACTIONS(1840),
    [aux_sym_signed_number_token1] = ACTIONS(1838),
    [anon_sym_SQUOTE] = ACTIONS(1840),
    [anon_sym_DQUOTE] = ACTIONS(1840),
    [sym_prefix_operator] = ACTIONS(1840),
    [aux_sym_sign_operator_token1] = ACTIONS(1838),
    [sym_hash_operator] = ACTIONS(1840),
    [anon_sym_lsl] = ACTIONS(1838),
    [anon_sym_lsr] = ACTIONS(1838),
    [anon_sym_asr] = ACTIONS(1838),
    [aux_sym_pow_operator_token1] = ACTIONS(1840),
    [anon_sym_mod] = ACTIONS(1838),
    [anon_sym_land] = ACTIONS(1838),
    [anon_sym_lor] = ACTIONS(1838),
    [anon_sym_lxor] = ACTIONS(1838),
    [aux_sym_mult_operator_token1] = ACTIONS(1838),
    [aux_sym_add_operator_token1] = ACTIONS(1838),
    [sym_concat_operator] = ACTIONS(1840),
    [sym_rel_operator] = ACTIONS(1838),
    [anon_sym_AMP_AMP] = ACTIONS(1838),
    [anon_sym_or] = ACTIONS(1838),
    [anon_sym_PIPE_PIPE] = ACTIONS(1838),
    [sym_assign_operator] = ACTIONS(1840),
    [sym__uppercase_identifier] = ACTIONS(1840),
    [aux_sym_tag_token1] = ACTIONS(1840),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(819)] = {
    [sym_attribute] = STATE(819),
    [sym__lowercase_identifier] = ACTIONS(1842),
    [anon_sym_and] = ACTIONS(1842),
    [anon_sym_TILDE] = ACTIONS(1842),
    [anon_sym_QMARK] = ACTIONS(1842),
    [anon_sym_LPAREN] = ACTIONS(1844),
    [anon_sym_RPAREN] = ACTIONS(1844),
    [anon_sym_COMMA] = ACTIONS(1844),
    [anon_sym_PIPE] = ACTIONS(1842),
    [anon_sym_COLON2] = ACTIONS(1842),
    [anon_sym_DOT] = ACTIONS(1844),
    [anon_sym_DASH_GT] = ACTIONS(1842),
    [anon_sym_LBRACK] = ACTIONS(1842),
    [anon_sym_RBRACK] = ACTIONS(1844),
    [anon_sym_true] = ACTIONS(1842),
    [anon_sym_false] = ACTIONS(1842),
    [anon_sym_LBRACE] = ACTIONS(1842),
    [anon_sym_SEMI] = ACTIONS(1844),
    [anon_sym_RBRACE] = ACTIONS(1844),
    [anon_sym_constraint] = ACTIONS(1842),
    [anon_sym_val] = ACTIONS(1842),
    [anon_sym_end] = ACTIONS(1842),
    [anon_sym_with] = ACTIONS(1842),
    [anon_sym_COLON_GT] = ACTIONS(1844),
    [anon_sym_object] = ACTIONS(1842),
    [anon_sym_inherit] = ACTIONS(1842),
    [anon_sym_method] = ACTIONS(1842),
    [anon_sym_in] = ACTIONS(1842),
    [anon_sym_as] = ACTIONS(1842),
    [anon_sym_initializer] = ACTIONS(1842),
    [anon_sym_AMP] = ACTIONS(1842),
    [anon_sym_POUND] = ACTIONS(1842),
    [anon_sym_COLON_COLON] = ACTIONS(1844),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1844),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1844),
    [anon_sym_then] = ACTIONS(1842),
    [anon_sym_else] = ACTIONS(1842),
    [anon_sym_do] = ACTIONS(1842),
    [anon_sym_done] = ACTIONS(1842),
    [anon_sym_to] = ACTIONS(1842),
    [anon_sym_downto] = ACTIONS(1842),
    [anon_sym_new] = ACTIONS(1842),
    [anon_sym_LBRACE_LT] = ACTIONS(1844),
    [anon_sym_GT_RBRACE] = ACTIONS(1844),
    [anon_sym_begin] = ACTIONS(1842),
    [sym_ocamlyacc_value] = ACTIONS(1844),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1842),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1844),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1842),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1842),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1844),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1844),
    [aux_sym_number_token1] = ACTIONS(1844),
    [aux_sym_signed_number_token1] = ACTIONS(1842),
    [anon_sym_SQUOTE] = ACTIONS(1844),
    [anon_sym_DQUOTE] = ACTIONS(1844),
    [sym_prefix_operator] = ACTIONS(1844),
    [aux_sym_sign_operator_token1] = ACTIONS(1842),
    [sym_hash_operator] = ACTIONS(1844),
    [anon_sym_lsl] = ACTIONS(1842),
    [anon_sym_lsr] = ACTIONS(1842),
    [anon_sym_asr] = ACTIONS(1842),
    [aux_sym_pow_operator_token1] = ACTIONS(1844),
    [anon_sym_mod] = ACTIONS(1842),
    [anon_sym_land] = ACTIONS(1842),
    [anon_sym_lor] = ACTIONS(1842),
    [anon_sym_lxor] = ACTIONS(1842),
    [aux_sym_mult_operator_token1] = ACTIONS(1842),
    [aux_sym_add_operator_token1] = ACTIONS(1842),
    [sym_concat_operator] = ACTIONS(1844),
    [sym_rel_operator] = ACTIONS(1842),
    [anon_sym_AMP_AMP] = ACTIONS(1842),
    [anon_sym_or] = ACTIONS(1842),
    [anon_sym_PIPE_PIPE] = ACTIONS(1842),
    [sym_assign_operator] = ACTIONS(1844),
    [sym__uppercase_identifier] = ACTIONS(1844),
    [aux_sym_tag_token1] = ACTIONS(1844),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(820)] = {
    [sym_attribute] = STATE(820),
    [sym__lowercase_identifier] = ACTIONS(1846),
    [anon_sym_and] = ACTIONS(1846),
    [anon_sym_TILDE] = ACTIONS(1846),
    [anon_sym_QMARK] = ACTIONS(1846),
    [anon_sym_LPAREN] = ACTIONS(1848),
    [anon_sym_RPAREN] = ACTIONS(1848),
    [anon_sym_COMMA] = ACTIONS(1848),
    [anon_sym_PIPE] = ACTIONS(1846),
    [anon_sym_COLON2] = ACTIONS(1846),
    [anon_sym_DOT] = ACTIONS(1848),
    [anon_sym_DASH_GT] = ACTIONS(1846),
    [anon_sym_LBRACK] = ACTIONS(1846),
    [anon_sym_RBRACK] = ACTIONS(1848),
    [anon_sym_true] = ACTIONS(1846),
    [anon_sym_false] = ACTIONS(1846),
    [anon_sym_LBRACE] = ACTIONS(1846),
    [anon_sym_SEMI] = ACTIONS(1848),
    [anon_sym_RBRACE] = ACTIONS(1848),
    [anon_sym_constraint] = ACTIONS(1846),
    [anon_sym_val] = ACTIONS(1846),
    [anon_sym_end] = ACTIONS(1846),
    [anon_sym_with] = ACTIONS(1846),
    [anon_sym_COLON_GT] = ACTIONS(1848),
    [anon_sym_object] = ACTIONS(1846),
    [anon_sym_inherit] = ACTIONS(1846),
    [anon_sym_method] = ACTIONS(1846),
    [anon_sym_in] = ACTIONS(1846),
    [anon_sym_as] = ACTIONS(1846),
    [anon_sym_initializer] = ACTIONS(1846),
    [anon_sym_AMP] = ACTIONS(1846),
    [anon_sym_POUND] = ACTIONS(1846),
    [anon_sym_COLON_COLON] = ACTIONS(1848),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1848),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1848),
    [anon_sym_then] = ACTIONS(1846),
    [anon_sym_else] = ACTIONS(1846),
    [anon_sym_do] = ACTIONS(1846),
    [anon_sym_done] = ACTIONS(1846),
    [anon_sym_to] = ACTIONS(1846),
    [anon_sym_downto] = ACTIONS(1846),
    [anon_sym_new] = ACTIONS(1846),
    [anon_sym_LBRACE_LT] = ACTIONS(1848),
    [anon_sym_GT_RBRACE] = ACTIONS(1848),
    [anon_sym_begin] = ACTIONS(1846),
    [sym_ocamlyacc_value] = ACTIONS(1848),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1846),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1848),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1846),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1846),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1848),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1848),
    [aux_sym_number_token1] = ACTIONS(1848),
    [aux_sym_signed_number_token1] = ACTIONS(1846),
    [anon_sym_SQUOTE] = ACTIONS(1848),
    [anon_sym_DQUOTE] = ACTIONS(1848),
    [sym_prefix_operator] = ACTIONS(1848),
    [aux_sym_sign_operator_token1] = ACTIONS(1846),
    [sym_hash_operator] = ACTIONS(1848),
    [anon_sym_lsl] = ACTIONS(1846),
    [anon_sym_lsr] = ACTIONS(1846),
    [anon_sym_asr] = ACTIONS(1846),
    [aux_sym_pow_operator_token1] = ACTIONS(1848),
    [anon_sym_mod] = ACTIONS(1846),
    [anon_sym_land] = ACTIONS(1846),
    [anon_sym_lor] = ACTIONS(1846),
    [anon_sym_lxor] = ACTIONS(1846),
    [aux_sym_mult_operator_token1] = ACTIONS(1846),
    [aux_sym_add_operator_token1] = ACTIONS(1846),
    [sym_concat_operator] = ACTIONS(1848),
    [sym_rel_operator] = ACTIONS(1846),
    [anon_sym_AMP_AMP] = ACTIONS(1846),
    [anon_sym_or] = ACTIONS(1846),
    [anon_sym_PIPE_PIPE] = ACTIONS(1846),
    [sym_assign_operator] = ACTIONS(1848),
    [sym__uppercase_identifier] = ACTIONS(1848),
    [aux_sym_tag_token1] = ACTIONS(1848),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(821)] = {
    [sym_attribute] = STATE(821),
    [sym__lowercase_identifier] = ACTIONS(1850),
    [anon_sym_and] = ACTIONS(1850),
    [anon_sym_TILDE] = ACTIONS(1850),
    [anon_sym_QMARK] = ACTIONS(1850),
    [anon_sym_LPAREN] = ACTIONS(1852),
    [anon_sym_RPAREN] = ACTIONS(1852),
    [anon_sym_COMMA] = ACTIONS(1852),
    [anon_sym_PIPE] = ACTIONS(1850),
    [anon_sym_COLON2] = ACTIONS(1850),
    [anon_sym_DOT] = ACTIONS(1852),
    [anon_sym_DASH_GT] = ACTIONS(1850),
    [anon_sym_LBRACK] = ACTIONS(1850),
    [anon_sym_RBRACK] = ACTIONS(1852),
    [anon_sym_true] = ACTIONS(1850),
    [anon_sym_false] = ACTIONS(1850),
    [anon_sym_LBRACE] = ACTIONS(1850),
    [anon_sym_SEMI] = ACTIONS(1852),
    [anon_sym_RBRACE] = ACTIONS(1852),
    [anon_sym_constraint] = ACTIONS(1850),
    [anon_sym_val] = ACTIONS(1850),
    [anon_sym_end] = ACTIONS(1850),
    [anon_sym_with] = ACTIONS(1850),
    [anon_sym_COLON_GT] = ACTIONS(1852),
    [anon_sym_object] = ACTIONS(1850),
    [anon_sym_inherit] = ACTIONS(1850),
    [anon_sym_method] = ACTIONS(1850),
    [anon_sym_in] = ACTIONS(1850),
    [anon_sym_as] = ACTIONS(1850),
    [anon_sym_initializer] = ACTIONS(1850),
    [anon_sym_AMP] = ACTIONS(1850),
    [anon_sym_POUND] = ACTIONS(1850),
    [anon_sym_COLON_COLON] = ACTIONS(1852),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1852),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1852),
    [anon_sym_then] = ACTIONS(1850),
    [anon_sym_else] = ACTIONS(1850),
    [anon_sym_do] = ACTIONS(1850),
    [anon_sym_done] = ACTIONS(1850),
    [anon_sym_to] = ACTIONS(1850),
    [anon_sym_downto] = ACTIONS(1850),
    [anon_sym_new] = ACTIONS(1850),
    [anon_sym_LBRACE_LT] = ACTIONS(1852),
    [anon_sym_GT_RBRACE] = ACTIONS(1852),
    [anon_sym_begin] = ACTIONS(1850),
    [sym_ocamlyacc_value] = ACTIONS(1852),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1850),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1852),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1850),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1850),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1852),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1852),
    [aux_sym_number_token1] = ACTIONS(1852),
    [aux_sym_signed_number_token1] = ACTIONS(1850),
    [anon_sym_SQUOTE] = ACTIONS(1852),
    [anon_sym_DQUOTE] = ACTIONS(1852),
    [sym_prefix_operator] = ACTIONS(1852),
    [aux_sym_sign_operator_token1] = ACTIONS(1850),
    [sym_hash_operator] = ACTIONS(1852),
    [anon_sym_lsl] = ACTIONS(1850),
    [anon_sym_lsr] = ACTIONS(1850),
    [anon_sym_asr] = ACTIONS(1850),
    [aux_sym_pow_operator_token1] = ACTIONS(1852),
    [anon_sym_mod] = ACTIONS(1850),
    [anon_sym_land] = ACTIONS(1850),
    [anon_sym_lor] = ACTIONS(1850),
    [anon_sym_lxor] = ACTIONS(1850),
    [aux_sym_mult_operator_token1] = ACTIONS(1850),
    [aux_sym_add_operator_token1] = ACTIONS(1850),
    [sym_concat_operator] = ACTIONS(1852),
    [sym_rel_operator] = ACTIONS(1850),
    [anon_sym_AMP_AMP] = ACTIONS(1850),
    [anon_sym_or] = ACTIONS(1850),
    [anon_sym_PIPE_PIPE] = ACTIONS(1850),
    [sym_assign_operator] = ACTIONS(1852),
    [sym__uppercase_identifier] = ACTIONS(1852),
    [aux_sym_tag_token1] = ACTIONS(1852),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(822)] = {
    [sym_attribute] = STATE(822),
    [sym__lowercase_identifier] = ACTIONS(1854),
    [anon_sym_and] = ACTIONS(1854),
    [anon_sym_TILDE] = ACTIONS(1854),
    [anon_sym_QMARK] = ACTIONS(1854),
    [anon_sym_LPAREN] = ACTIONS(1856),
    [anon_sym_RPAREN] = ACTIONS(1856),
    [anon_sym_COMMA] = ACTIONS(1856),
    [anon_sym_PIPE] = ACTIONS(1854),
    [anon_sym_COLON2] = ACTIONS(1854),
    [anon_sym_DOT] = ACTIONS(1856),
    [anon_sym_DASH_GT] = ACTIONS(1854),
    [anon_sym_LBRACK] = ACTIONS(1854),
    [anon_sym_RBRACK] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(1854),
    [anon_sym_false] = ACTIONS(1854),
    [anon_sym_LBRACE] = ACTIONS(1854),
    [anon_sym_SEMI] = ACTIONS(1856),
    [anon_sym_RBRACE] = ACTIONS(1856),
    [anon_sym_constraint] = ACTIONS(1854),
    [anon_sym_val] = ACTIONS(1854),
    [anon_sym_end] = ACTIONS(1854),
    [anon_sym_with] = ACTIONS(1854),
    [anon_sym_COLON_GT] = ACTIONS(1856),
    [anon_sym_object] = ACTIONS(1854),
    [anon_sym_inherit] = ACTIONS(1854),
    [anon_sym_method] = ACTIONS(1854),
    [anon_sym_in] = ACTIONS(1854),
    [anon_sym_as] = ACTIONS(1854),
    [anon_sym_initializer] = ACTIONS(1854),
    [anon_sym_AMP] = ACTIONS(1854),
    [anon_sym_POUND] = ACTIONS(1854),
    [anon_sym_COLON_COLON] = ACTIONS(1856),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1856),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1856),
    [anon_sym_then] = ACTIONS(1854),
    [anon_sym_else] = ACTIONS(1854),
    [anon_sym_do] = ACTIONS(1854),
    [anon_sym_done] = ACTIONS(1854),
    [anon_sym_to] = ACTIONS(1854),
    [anon_sym_downto] = ACTIONS(1854),
    [anon_sym_new] = ACTIONS(1854),
    [anon_sym_LBRACE_LT] = ACTIONS(1856),
    [anon_sym_GT_RBRACE] = ACTIONS(1856),
    [anon_sym_begin] = ACTIONS(1854),
    [sym_ocamlyacc_value] = ACTIONS(1856),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1854),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1856),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1854),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1854),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1856),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1856),
    [aux_sym_number_token1] = ACTIONS(1856),
    [aux_sym_signed_number_token1] = ACTIONS(1854),
    [anon_sym_SQUOTE] = ACTIONS(1856),
    [anon_sym_DQUOTE] = ACTIONS(1856),
    [sym_prefix_operator] = ACTIONS(1856),
    [aux_sym_sign_operator_token1] = ACTIONS(1854),
    [sym_hash_operator] = ACTIONS(1856),
    [anon_sym_lsl] = ACTIONS(1854),
    [anon_sym_lsr] = ACTIONS(1854),
    [anon_sym_asr] = ACTIONS(1854),
    [aux_sym_pow_operator_token1] = ACTIONS(1856),
    [anon_sym_mod] = ACTIONS(1854),
    [anon_sym_land] = ACTIONS(1854),
    [anon_sym_lor] = ACTIONS(1854),
    [anon_sym_lxor] = ACTIONS(1854),
    [aux_sym_mult_operator_token1] = ACTIONS(1854),
    [aux_sym_add_operator_token1] = ACTIONS(1854),
    [sym_concat_operator] = ACTIONS(1856),
    [sym_rel_operator] = ACTIONS(1854),
    [anon_sym_AMP_AMP] = ACTIONS(1854),
    [anon_sym_or] = ACTIONS(1854),
    [anon_sym_PIPE_PIPE] = ACTIONS(1854),
    [sym_assign_operator] = ACTIONS(1856),
    [sym__uppercase_identifier] = ACTIONS(1856),
    [aux_sym_tag_token1] = ACTIONS(1856),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(823)] = {
    [sym_attribute] = STATE(823),
    [sym__lowercase_identifier] = ACTIONS(1858),
    [anon_sym_and] = ACTIONS(1858),
    [anon_sym_TILDE] = ACTIONS(1858),
    [anon_sym_QMARK] = ACTIONS(1858),
    [anon_sym_LPAREN] = ACTIONS(1860),
    [anon_sym_RPAREN] = ACTIONS(1860),
    [anon_sym_COMMA] = ACTIONS(1860),
    [anon_sym_PIPE] = ACTIONS(1858),
    [anon_sym_COLON2] = ACTIONS(1858),
    [anon_sym_DOT] = ACTIONS(1860),
    [anon_sym_DASH_GT] = ACTIONS(1858),
    [anon_sym_LBRACK] = ACTIONS(1858),
    [anon_sym_RBRACK] = ACTIONS(1860),
    [anon_sym_true] = ACTIONS(1858),
    [anon_sym_false] = ACTIONS(1858),
    [anon_sym_LBRACE] = ACTIONS(1858),
    [anon_sym_SEMI] = ACTIONS(1860),
    [anon_sym_RBRACE] = ACTIONS(1860),
    [anon_sym_constraint] = ACTIONS(1858),
    [anon_sym_val] = ACTIONS(1858),
    [anon_sym_end] = ACTIONS(1858),
    [anon_sym_with] = ACTIONS(1858),
    [anon_sym_COLON_GT] = ACTIONS(1860),
    [anon_sym_object] = ACTIONS(1858),
    [anon_sym_inherit] = ACTIONS(1858),
    [anon_sym_method] = ACTIONS(1858),
    [anon_sym_in] = ACTIONS(1858),
    [anon_sym_as] = ACTIONS(1858),
    [anon_sym_initializer] = ACTIONS(1858),
    [anon_sym_AMP] = ACTIONS(1858),
    [anon_sym_POUND] = ACTIONS(1858),
    [anon_sym_COLON_COLON] = ACTIONS(1860),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1860),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1860),
    [anon_sym_then] = ACTIONS(1858),
    [anon_sym_else] = ACTIONS(1858),
    [anon_sym_do] = ACTIONS(1858),
    [anon_sym_done] = ACTIONS(1858),
    [anon_sym_to] = ACTIONS(1858),
    [anon_sym_downto] = ACTIONS(1858),
    [anon_sym_new] = ACTIONS(1858),
    [anon_sym_LBRACE_LT] = ACTIONS(1860),
    [anon_sym_GT_RBRACE] = ACTIONS(1860),
    [anon_sym_begin] = ACTIONS(1858),
    [sym_ocamlyacc_value] = ACTIONS(1860),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1858),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1860),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1858),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1858),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1860),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1860),
    [aux_sym_number_token1] = ACTIONS(1860),
    [aux_sym_signed_number_token1] = ACTIONS(1858),
    [anon_sym_SQUOTE] = ACTIONS(1860),
    [anon_sym_DQUOTE] = ACTIONS(1860),
    [sym_prefix_operator] = ACTIONS(1860),
    [aux_sym_sign_operator_token1] = ACTIONS(1858),
    [sym_hash_operator] = ACTIONS(1860),
    [anon_sym_lsl] = ACTIONS(1858),
    [anon_sym_lsr] = ACTIONS(1858),
    [anon_sym_asr] = ACTIONS(1858),
    [aux_sym_pow_operator_token1] = ACTIONS(1860),
    [anon_sym_mod] = ACTIONS(1858),
    [anon_sym_land] = ACTIONS(1858),
    [anon_sym_lor] = ACTIONS(1858),
    [anon_sym_lxor] = ACTIONS(1858),
    [aux_sym_mult_operator_token1] = ACTIONS(1858),
    [aux_sym_add_operator_token1] = ACTIONS(1858),
    [sym_concat_operator] = ACTIONS(1860),
    [sym_rel_operator] = ACTIONS(1858),
    [anon_sym_AMP_AMP] = ACTIONS(1858),
    [anon_sym_or] = ACTIONS(1858),
    [anon_sym_PIPE_PIPE] = ACTIONS(1858),
    [sym_assign_operator] = ACTIONS(1860),
    [sym__uppercase_identifier] = ACTIONS(1860),
    [aux_sym_tag_token1] = ACTIONS(1860),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(824)] = {
    [sym__delimited_expression] = STATE(2350),
    [sym__simple_expression] = STATE(2404),
    [sym_typed_expression] = STATE(2350),
    [sym_list_expression] = STATE(2351),
    [sym_array_expression] = STATE(2351),
    [sym_record_expression] = STATE(2351),
    [sym_labeled_argument] = STATE(2522),
    [sym_prefix_expression] = STATE(2350),
    [sym_hash_expression] = STATE(2350),
    [sym_field_get_expression] = STATE(2350),
    [sym_array_get_expression] = STATE(2350),
    [sym_string_get_expression] = STATE(2350),
    [sym_bigarray_get_expression] = STATE(2350),
    [sym_local_open_expression] = STATE(2350),
    [sym_package_expression] = STATE(2351),
    [sym_new_expression] = STATE(2350),
    [sym_object_copy_expression] = STATE(2351),
    [sym_method_invocation] = STATE(2350),
    [sym_object_expression] = STATE(2350),
    [sym_parenthesized_expression] = STATE(2351),
    [sym_attribute] = STATE(824),
    [sym_extension] = STATE(2350),
    [sym_quoted_extension] = STATE(2350),
    [sym__constant] = STATE(2350),
    [sym_number] = STATE(2348),
    [sym_character] = STATE(2348),
    [sym_string] = STATE(2348),
    [sym_quoted_string] = STATE(2348),
    [sym_boolean] = STATE(2348),
    [sym_unit] = STATE(2351),
    [sym__value_name] = STATE(2402),
    [sym_parenthesized_operator] = STATE(2406),
    [sym_value_path] = STATE(2350),
    [sym_module_path] = STATE(6407),
    [sym_constructor_path] = STATE(2350),
    [sym__constructor_name] = STATE(2419),
    [sym_tag] = STATE(2350),
    [aux_sym_class_application_repeat1] = STATE(813),
    [ts_builtin_sym_end] = ACTIONS(1542),
    [sym__lowercase_identifier] = ACTIONS(1654),
    [anon_sym_SEMI_SEMI] = ACTIONS(1542),
    [anon_sym_and] = ACTIONS(1544),
    [anon_sym_TILDE] = ACTIONS(1656),
    [anon_sym_QMARK] = ACTIONS(1656),
    [anon_sym_LPAREN] = ACTIONS(1658),
    [anon_sym_external] = ACTIONS(1544),
    [anon_sym_type] = ACTIONS(1544),
    [anon_sym_LBRACK] = ACTIONS(1660),
    [anon_sym_RBRACK] = ACTIONS(1542),
    [anon_sym_true] = ACTIONS(1662),
    [anon_sym_false] = ACTIONS(1662),
    [anon_sym_LBRACE] = ACTIONS(1664),
    [anon_sym_exception] = ACTIONS(1544),
    [anon_sym_module] = ACTIONS(1544),
    [anon_sym_open] = ACTIONS(1544),
    [anon_sym_include] = ACTIONS(1544),
    [anon_sym_class] = ACTIONS(1544),
    [anon_sym_val] = ACTIONS(1544),
    [anon_sym_end] = ACTIONS(1544),
    [anon_sym_object] = ACTIONS(1666),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1668),
    [anon_sym_new] = ACTIONS(1670),
    [anon_sym_LBRACE_LT] = ACTIONS(1672),
    [anon_sym_begin] = ACTIONS(1674),
    [sym_ocamlyacc_value] = ACTIONS(1676),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1544),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1542),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1678),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1680),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1542),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1542),
    [aux_sym_number_token1] = ACTIONS(1682),
    [anon_sym_SQUOTE] = ACTIONS(1684),
    [anon_sym_DQUOTE] = ACTIONS(1686),
    [sym_prefix_operator] = ACTIONS(1688),
    [sym__uppercase_identifier] = ACTIONS(1690),
    [aux_sym_tag_token1] = ACTIONS(1692),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(825)] = {
    [sym_attribute] = STATE(825),
    [sym__lowercase_identifier] = ACTIONS(1862),
    [anon_sym_and] = ACTIONS(1862),
    [anon_sym_TILDE] = ACTIONS(1862),
    [anon_sym_QMARK] = ACTIONS(1862),
    [anon_sym_LPAREN] = ACTIONS(1864),
    [anon_sym_RPAREN] = ACTIONS(1864),
    [anon_sym_COMMA] = ACTIONS(1864),
    [anon_sym_PIPE] = ACTIONS(1862),
    [anon_sym_COLON2] = ACTIONS(1862),
    [anon_sym_DOT] = ACTIONS(1864),
    [anon_sym_DASH_GT] = ACTIONS(1862),
    [anon_sym_LBRACK] = ACTIONS(1862),
    [anon_sym_RBRACK] = ACTIONS(1864),
    [anon_sym_true] = ACTIONS(1862),
    [anon_sym_false] = ACTIONS(1862),
    [anon_sym_LBRACE] = ACTIONS(1862),
    [anon_sym_SEMI] = ACTIONS(1864),
    [anon_sym_RBRACE] = ACTIONS(1864),
    [anon_sym_constraint] = ACTIONS(1862),
    [anon_sym_val] = ACTIONS(1862),
    [anon_sym_end] = ACTIONS(1862),
    [anon_sym_with] = ACTIONS(1862),
    [anon_sym_COLON_GT] = ACTIONS(1864),
    [anon_sym_object] = ACTIONS(1862),
    [anon_sym_inherit] = ACTIONS(1862),
    [anon_sym_method] = ACTIONS(1862),
    [anon_sym_in] = ACTIONS(1862),
    [anon_sym_as] = ACTIONS(1862),
    [anon_sym_initializer] = ACTIONS(1862),
    [anon_sym_AMP] = ACTIONS(1862),
    [anon_sym_POUND] = ACTIONS(1862),
    [anon_sym_COLON_COLON] = ACTIONS(1864),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1864),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1864),
    [anon_sym_then] = ACTIONS(1862),
    [anon_sym_else] = ACTIONS(1862),
    [anon_sym_do] = ACTIONS(1862),
    [anon_sym_done] = ACTIONS(1862),
    [anon_sym_to] = ACTIONS(1862),
    [anon_sym_downto] = ACTIONS(1862),
    [anon_sym_new] = ACTIONS(1862),
    [anon_sym_LBRACE_LT] = ACTIONS(1864),
    [anon_sym_GT_RBRACE] = ACTIONS(1864),
    [anon_sym_begin] = ACTIONS(1862),
    [sym_ocamlyacc_value] = ACTIONS(1864),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1862),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1864),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1862),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1862),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1864),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1864),
    [aux_sym_number_token1] = ACTIONS(1864),
    [aux_sym_signed_number_token1] = ACTIONS(1862),
    [anon_sym_SQUOTE] = ACTIONS(1864),
    [anon_sym_DQUOTE] = ACTIONS(1864),
    [sym_prefix_operator] = ACTIONS(1864),
    [aux_sym_sign_operator_token1] = ACTIONS(1862),
    [sym_hash_operator] = ACTIONS(1864),
    [anon_sym_lsl] = ACTIONS(1862),
    [anon_sym_lsr] = ACTIONS(1862),
    [anon_sym_asr] = ACTIONS(1862),
    [aux_sym_pow_operator_token1] = ACTIONS(1864),
    [anon_sym_mod] = ACTIONS(1862),
    [anon_sym_land] = ACTIONS(1862),
    [anon_sym_lor] = ACTIONS(1862),
    [anon_sym_lxor] = ACTIONS(1862),
    [aux_sym_mult_operator_token1] = ACTIONS(1862),
    [aux_sym_add_operator_token1] = ACTIONS(1862),
    [sym_concat_operator] = ACTIONS(1864),
    [sym_rel_operator] = ACTIONS(1862),
    [anon_sym_AMP_AMP] = ACTIONS(1862),
    [anon_sym_or] = ACTIONS(1862),
    [anon_sym_PIPE_PIPE] = ACTIONS(1862),
    [sym_assign_operator] = ACTIONS(1864),
    [sym__uppercase_identifier] = ACTIONS(1864),
    [aux_sym_tag_token1] = ACTIONS(1864),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(826)] = {
    [sym_attribute] = STATE(826),
    [sym__lowercase_identifier] = ACTIONS(1866),
    [anon_sym_and] = ACTIONS(1866),
    [anon_sym_TILDE] = ACTIONS(1866),
    [anon_sym_QMARK] = ACTIONS(1866),
    [anon_sym_LPAREN] = ACTIONS(1868),
    [anon_sym_RPAREN] = ACTIONS(1868),
    [anon_sym_COMMA] = ACTIONS(1868),
    [anon_sym_PIPE] = ACTIONS(1866),
    [anon_sym_COLON2] = ACTIONS(1866),
    [anon_sym_DOT] = ACTIONS(1868),
    [anon_sym_DASH_GT] = ACTIONS(1866),
    [anon_sym_LBRACK] = ACTIONS(1866),
    [anon_sym_RBRACK] = ACTIONS(1868),
    [anon_sym_true] = ACTIONS(1866),
    [anon_sym_false] = ACTIONS(1866),
    [anon_sym_LBRACE] = ACTIONS(1866),
    [anon_sym_SEMI] = ACTIONS(1868),
    [anon_sym_RBRACE] = ACTIONS(1868),
    [anon_sym_constraint] = ACTIONS(1866),
    [anon_sym_val] = ACTIONS(1866),
    [anon_sym_end] = ACTIONS(1866),
    [anon_sym_with] = ACTIONS(1866),
    [anon_sym_COLON_GT] = ACTIONS(1868),
    [anon_sym_object] = ACTIONS(1866),
    [anon_sym_inherit] = ACTIONS(1866),
    [anon_sym_method] = ACTIONS(1866),
    [anon_sym_in] = ACTIONS(1866),
    [anon_sym_as] = ACTIONS(1866),
    [anon_sym_initializer] = ACTIONS(1866),
    [anon_sym_AMP] = ACTIONS(1866),
    [anon_sym_POUND] = ACTIONS(1866),
    [anon_sym_COLON_COLON] = ACTIONS(1868),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1868),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1868),
    [anon_sym_then] = ACTIONS(1866),
    [anon_sym_else] = ACTIONS(1866),
    [anon_sym_do] = ACTIONS(1866),
    [anon_sym_done] = ACTIONS(1866),
    [anon_sym_to] = ACTIONS(1866),
    [anon_sym_downto] = ACTIONS(1866),
    [anon_sym_new] = ACTIONS(1866),
    [anon_sym_LBRACE_LT] = ACTIONS(1868),
    [anon_sym_GT_RBRACE] = ACTIONS(1868),
    [anon_sym_begin] = ACTIONS(1866),
    [sym_ocamlyacc_value] = ACTIONS(1868),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1866),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1868),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1866),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1866),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1868),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1868),
    [aux_sym_number_token1] = ACTIONS(1868),
    [aux_sym_signed_number_token1] = ACTIONS(1866),
    [anon_sym_SQUOTE] = ACTIONS(1868),
    [anon_sym_DQUOTE] = ACTIONS(1868),
    [sym_prefix_operator] = ACTIONS(1868),
    [aux_sym_sign_operator_token1] = ACTIONS(1866),
    [sym_hash_operator] = ACTIONS(1868),
    [anon_sym_lsl] = ACTIONS(1866),
    [anon_sym_lsr] = ACTIONS(1866),
    [anon_sym_asr] = ACTIONS(1866),
    [aux_sym_pow_operator_token1] = ACTIONS(1868),
    [anon_sym_mod] = ACTIONS(1866),
    [anon_sym_land] = ACTIONS(1866),
    [anon_sym_lor] = ACTIONS(1866),
    [anon_sym_lxor] = ACTIONS(1866),
    [aux_sym_mult_operator_token1] = ACTIONS(1866),
    [aux_sym_add_operator_token1] = ACTIONS(1866),
    [sym_concat_operator] = ACTIONS(1868),
    [sym_rel_operator] = ACTIONS(1866),
    [anon_sym_AMP_AMP] = ACTIONS(1866),
    [anon_sym_or] = ACTIONS(1866),
    [anon_sym_PIPE_PIPE] = ACTIONS(1866),
    [sym_assign_operator] = ACTIONS(1868),
    [sym__uppercase_identifier] = ACTIONS(1868),
    [aux_sym_tag_token1] = ACTIONS(1868),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(827)] = {
    [sym_attribute] = STATE(827),
    [sym__lowercase_identifier] = ACTIONS(1870),
    [anon_sym_and] = ACTIONS(1870),
    [anon_sym_TILDE] = ACTIONS(1870),
    [anon_sym_QMARK] = ACTIONS(1870),
    [anon_sym_LPAREN] = ACTIONS(1872),
    [anon_sym_RPAREN] = ACTIONS(1872),
    [anon_sym_COMMA] = ACTIONS(1872),
    [anon_sym_PIPE] = ACTIONS(1870),
    [anon_sym_COLON2] = ACTIONS(1870),
    [anon_sym_DOT] = ACTIONS(1872),
    [anon_sym_DASH_GT] = ACTIONS(1870),
    [anon_sym_LBRACK] = ACTIONS(1870),
    [anon_sym_RBRACK] = ACTIONS(1872),
    [anon_sym_true] = ACTIONS(1870),
    [anon_sym_false] = ACTIONS(1870),
    [anon_sym_LBRACE] = ACTIONS(1870),
    [anon_sym_SEMI] = ACTIONS(1872),
    [anon_sym_RBRACE] = ACTIONS(1872),
    [anon_sym_constraint] = ACTIONS(1870),
    [anon_sym_val] = ACTIONS(1870),
    [anon_sym_end] = ACTIONS(1870),
    [anon_sym_with] = ACTIONS(1870),
    [anon_sym_COLON_GT] = ACTIONS(1872),
    [anon_sym_object] = ACTIONS(1870),
    [anon_sym_inherit] = ACTIONS(1870),
    [anon_sym_method] = ACTIONS(1870),
    [anon_sym_in] = ACTIONS(1870),
    [anon_sym_as] = ACTIONS(1870),
    [anon_sym_initializer] = ACTIONS(1870),
    [anon_sym_AMP] = ACTIONS(1870),
    [anon_sym_POUND] = ACTIONS(1870),
    [anon_sym_COLON_COLON] = ACTIONS(1872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1872),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1872),
    [anon_sym_then] = ACTIONS(1870),
    [anon_sym_else] = ACTIONS(1870),
    [anon_sym_do] = ACTIONS(1870),
    [anon_sym_done] = ACTIONS(1870),
    [anon_sym_to] = ACTIONS(1870),
    [anon_sym_downto] = ACTIONS(1870),
    [anon_sym_new] = ACTIONS(1870),
    [anon_sym_LBRACE_LT] = ACTIONS(1872),
    [anon_sym_GT_RBRACE] = ACTIONS(1872),
    [anon_sym_begin] = ACTIONS(1870),
    [sym_ocamlyacc_value] = ACTIONS(1872),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1870),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1872),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1870),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1870),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1872),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1872),
    [aux_sym_number_token1] = ACTIONS(1872),
    [aux_sym_signed_number_token1] = ACTIONS(1870),
    [anon_sym_SQUOTE] = ACTIONS(1872),
    [anon_sym_DQUOTE] = ACTIONS(1872),
    [sym_prefix_operator] = ACTIONS(1872),
    [aux_sym_sign_operator_token1] = ACTIONS(1870),
    [sym_hash_operator] = ACTIONS(1872),
    [anon_sym_lsl] = ACTIONS(1870),
    [anon_sym_lsr] = ACTIONS(1870),
    [anon_sym_asr] = ACTIONS(1870),
    [aux_sym_pow_operator_token1] = ACTIONS(1872),
    [anon_sym_mod] = ACTIONS(1870),
    [anon_sym_land] = ACTIONS(1870),
    [anon_sym_lor] = ACTIONS(1870),
    [anon_sym_lxor] = ACTIONS(1870),
    [aux_sym_mult_operator_token1] = ACTIONS(1870),
    [aux_sym_add_operator_token1] = ACTIONS(1870),
    [sym_concat_operator] = ACTIONS(1872),
    [sym_rel_operator] = ACTIONS(1870),
    [anon_sym_AMP_AMP] = ACTIONS(1870),
    [anon_sym_or] = ACTIONS(1870),
    [anon_sym_PIPE_PIPE] = ACTIONS(1870),
    [sym_assign_operator] = ACTIONS(1872),
    [sym__uppercase_identifier] = ACTIONS(1872),
    [aux_sym_tag_token1] = ACTIONS(1872),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(828)] = {
    [sym_attribute] = STATE(828),
    [sym__lowercase_identifier] = ACTIONS(1874),
    [anon_sym_and] = ACTIONS(1874),
    [anon_sym_TILDE] = ACTIONS(1874),
    [anon_sym_QMARK] = ACTIONS(1874),
    [anon_sym_LPAREN] = ACTIONS(1876),
    [anon_sym_RPAREN] = ACTIONS(1876),
    [anon_sym_COMMA] = ACTIONS(1876),
    [anon_sym_PIPE] = ACTIONS(1874),
    [anon_sym_COLON2] = ACTIONS(1874),
    [anon_sym_DOT] = ACTIONS(1876),
    [anon_sym_DASH_GT] = ACTIONS(1874),
    [anon_sym_LBRACK] = ACTIONS(1874),
    [anon_sym_RBRACK] = ACTIONS(1876),
    [anon_sym_true] = ACTIONS(1874),
    [anon_sym_false] = ACTIONS(1874),
    [anon_sym_LBRACE] = ACTIONS(1874),
    [anon_sym_SEMI] = ACTIONS(1876),
    [anon_sym_RBRACE] = ACTIONS(1876),
    [anon_sym_constraint] = ACTIONS(1874),
    [anon_sym_val] = ACTIONS(1874),
    [anon_sym_end] = ACTIONS(1874),
    [anon_sym_with] = ACTIONS(1874),
    [anon_sym_COLON_GT] = ACTIONS(1876),
    [anon_sym_object] = ACTIONS(1874),
    [anon_sym_inherit] = ACTIONS(1874),
    [anon_sym_method] = ACTIONS(1874),
    [anon_sym_in] = ACTIONS(1874),
    [anon_sym_as] = ACTIONS(1874),
    [anon_sym_initializer] = ACTIONS(1874),
    [anon_sym_AMP] = ACTIONS(1874),
    [anon_sym_POUND] = ACTIONS(1874),
    [anon_sym_COLON_COLON] = ACTIONS(1876),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1876),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1876),
    [anon_sym_then] = ACTIONS(1874),
    [anon_sym_else] = ACTIONS(1874),
    [anon_sym_do] = ACTIONS(1874),
    [anon_sym_done] = ACTIONS(1874),
    [anon_sym_to] = ACTIONS(1874),
    [anon_sym_downto] = ACTIONS(1874),
    [anon_sym_new] = ACTIONS(1874),
    [anon_sym_LBRACE_LT] = ACTIONS(1876),
    [anon_sym_GT_RBRACE] = ACTIONS(1876),
    [anon_sym_begin] = ACTIONS(1874),
    [sym_ocamlyacc_value] = ACTIONS(1876),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1874),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1876),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1874),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1874),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1876),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1876),
    [aux_sym_number_token1] = ACTIONS(1876),
    [aux_sym_signed_number_token1] = ACTIONS(1874),
    [anon_sym_SQUOTE] = ACTIONS(1876),
    [anon_sym_DQUOTE] = ACTIONS(1876),
    [sym_prefix_operator] = ACTIONS(1876),
    [aux_sym_sign_operator_token1] = ACTIONS(1874),
    [sym_hash_operator] = ACTIONS(1876),
    [anon_sym_lsl] = ACTIONS(1874),
    [anon_sym_lsr] = ACTIONS(1874),
    [anon_sym_asr] = ACTIONS(1874),
    [aux_sym_pow_operator_token1] = ACTIONS(1876),
    [anon_sym_mod] = ACTIONS(1874),
    [anon_sym_land] = ACTIONS(1874),
    [anon_sym_lor] = ACTIONS(1874),
    [anon_sym_lxor] = ACTIONS(1874),
    [aux_sym_mult_operator_token1] = ACTIONS(1874),
    [aux_sym_add_operator_token1] = ACTIONS(1874),
    [sym_concat_operator] = ACTIONS(1876),
    [sym_rel_operator] = ACTIONS(1874),
    [anon_sym_AMP_AMP] = ACTIONS(1874),
    [anon_sym_or] = ACTIONS(1874),
    [anon_sym_PIPE_PIPE] = ACTIONS(1874),
    [sym_assign_operator] = ACTIONS(1876),
    [sym__uppercase_identifier] = ACTIONS(1876),
    [aux_sym_tag_token1] = ACTIONS(1876),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(829)] = {
    [sym_attribute] = STATE(829),
    [sym__lowercase_identifier] = ACTIONS(1878),
    [anon_sym_and] = ACTIONS(1878),
    [anon_sym_TILDE] = ACTIONS(1878),
    [anon_sym_QMARK] = ACTIONS(1878),
    [anon_sym_LPAREN] = ACTIONS(1880),
    [anon_sym_RPAREN] = ACTIONS(1880),
    [anon_sym_COMMA] = ACTIONS(1880),
    [anon_sym_PIPE] = ACTIONS(1878),
    [anon_sym_COLON2] = ACTIONS(1878),
    [anon_sym_DOT] = ACTIONS(1880),
    [anon_sym_DASH_GT] = ACTIONS(1878),
    [anon_sym_LBRACK] = ACTIONS(1878),
    [anon_sym_RBRACK] = ACTIONS(1880),
    [anon_sym_true] = ACTIONS(1878),
    [anon_sym_false] = ACTIONS(1878),
    [anon_sym_LBRACE] = ACTIONS(1878),
    [anon_sym_SEMI] = ACTIONS(1880),
    [anon_sym_RBRACE] = ACTIONS(1880),
    [anon_sym_constraint] = ACTIONS(1878),
    [anon_sym_val] = ACTIONS(1878),
    [anon_sym_end] = ACTIONS(1878),
    [anon_sym_with] = ACTIONS(1878),
    [anon_sym_COLON_GT] = ACTIONS(1880),
    [anon_sym_object] = ACTIONS(1878),
    [anon_sym_inherit] = ACTIONS(1878),
    [anon_sym_method] = ACTIONS(1878),
    [anon_sym_in] = ACTIONS(1878),
    [anon_sym_as] = ACTIONS(1878),
    [anon_sym_initializer] = ACTIONS(1878),
    [anon_sym_AMP] = ACTIONS(1878),
    [anon_sym_POUND] = ACTIONS(1878),
    [anon_sym_COLON_COLON] = ACTIONS(1880),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1880),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1880),
    [anon_sym_then] = ACTIONS(1878),
    [anon_sym_else] = ACTIONS(1878),
    [anon_sym_do] = ACTIONS(1878),
    [anon_sym_done] = ACTIONS(1878),
    [anon_sym_to] = ACTIONS(1878),
    [anon_sym_downto] = ACTIONS(1878),
    [anon_sym_new] = ACTIONS(1878),
    [anon_sym_LBRACE_LT] = ACTIONS(1880),
    [anon_sym_GT_RBRACE] = ACTIONS(1880),
    [anon_sym_begin] = ACTIONS(1878),
    [sym_ocamlyacc_value] = ACTIONS(1880),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1878),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1880),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1878),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1878),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1880),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1880),
    [aux_sym_number_token1] = ACTIONS(1880),
    [aux_sym_signed_number_token1] = ACTIONS(1878),
    [anon_sym_SQUOTE] = ACTIONS(1880),
    [anon_sym_DQUOTE] = ACTIONS(1880),
    [sym_prefix_operator] = ACTIONS(1880),
    [aux_sym_sign_operator_token1] = ACTIONS(1878),
    [sym_hash_operator] = ACTIONS(1880),
    [anon_sym_lsl] = ACTIONS(1878),
    [anon_sym_lsr] = ACTIONS(1878),
    [anon_sym_asr] = ACTIONS(1878),
    [aux_sym_pow_operator_token1] = ACTIONS(1880),
    [anon_sym_mod] = ACTIONS(1878),
    [anon_sym_land] = ACTIONS(1878),
    [anon_sym_lor] = ACTIONS(1878),
    [anon_sym_lxor] = ACTIONS(1878),
    [aux_sym_mult_operator_token1] = ACTIONS(1878),
    [aux_sym_add_operator_token1] = ACTIONS(1878),
    [sym_concat_operator] = ACTIONS(1880),
    [sym_rel_operator] = ACTIONS(1878),
    [anon_sym_AMP_AMP] = ACTIONS(1878),
    [anon_sym_or] = ACTIONS(1878),
    [anon_sym_PIPE_PIPE] = ACTIONS(1878),
    [sym_assign_operator] = ACTIONS(1880),
    [sym__uppercase_identifier] = ACTIONS(1880),
    [aux_sym_tag_token1] = ACTIONS(1880),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(830)] = {
    [sym_attribute] = STATE(830),
    [sym__lowercase_identifier] = ACTIONS(1882),
    [anon_sym_and] = ACTIONS(1882),
    [anon_sym_TILDE] = ACTIONS(1882),
    [anon_sym_QMARK] = ACTIONS(1882),
    [anon_sym_LPAREN] = ACTIONS(1884),
    [anon_sym_RPAREN] = ACTIONS(1884),
    [anon_sym_COMMA] = ACTIONS(1884),
    [anon_sym_PIPE] = ACTIONS(1882),
    [anon_sym_COLON2] = ACTIONS(1882),
    [anon_sym_DOT] = ACTIONS(1884),
    [anon_sym_DASH_GT] = ACTIONS(1882),
    [anon_sym_LBRACK] = ACTIONS(1882),
    [anon_sym_RBRACK] = ACTIONS(1884),
    [anon_sym_true] = ACTIONS(1882),
    [anon_sym_false] = ACTIONS(1882),
    [anon_sym_LBRACE] = ACTIONS(1882),
    [anon_sym_SEMI] = ACTIONS(1884),
    [anon_sym_RBRACE] = ACTIONS(1884),
    [anon_sym_constraint] = ACTIONS(1882),
    [anon_sym_val] = ACTIONS(1882),
    [anon_sym_end] = ACTIONS(1882),
    [anon_sym_with] = ACTIONS(1882),
    [anon_sym_COLON_GT] = ACTIONS(1884),
    [anon_sym_object] = ACTIONS(1882),
    [anon_sym_inherit] = ACTIONS(1882),
    [anon_sym_method] = ACTIONS(1882),
    [anon_sym_in] = ACTIONS(1882),
    [anon_sym_as] = ACTIONS(1882),
    [anon_sym_initializer] = ACTIONS(1882),
    [anon_sym_AMP] = ACTIONS(1882),
    [anon_sym_POUND] = ACTIONS(1882),
    [anon_sym_COLON_COLON] = ACTIONS(1884),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1884),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1884),
    [anon_sym_then] = ACTIONS(1882),
    [anon_sym_else] = ACTIONS(1882),
    [anon_sym_do] = ACTIONS(1882),
    [anon_sym_done] = ACTIONS(1882),
    [anon_sym_to] = ACTIONS(1882),
    [anon_sym_downto] = ACTIONS(1882),
    [anon_sym_new] = ACTIONS(1882),
    [anon_sym_LBRACE_LT] = ACTIONS(1884),
    [anon_sym_GT_RBRACE] = ACTIONS(1884),
    [anon_sym_begin] = ACTIONS(1882),
    [sym_ocamlyacc_value] = ACTIONS(1884),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1882),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1884),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1882),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1882),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1884),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1884),
    [aux_sym_number_token1] = ACTIONS(1884),
    [aux_sym_signed_number_token1] = ACTIONS(1882),
    [anon_sym_SQUOTE] = ACTIONS(1884),
    [anon_sym_DQUOTE] = ACTIONS(1884),
    [sym_prefix_operator] = ACTIONS(1884),
    [aux_sym_sign_operator_token1] = ACTIONS(1882),
    [sym_hash_operator] = ACTIONS(1884),
    [anon_sym_lsl] = ACTIONS(1882),
    [anon_sym_lsr] = ACTIONS(1882),
    [anon_sym_asr] = ACTIONS(1882),
    [aux_sym_pow_operator_token1] = ACTIONS(1884),
    [anon_sym_mod] = ACTIONS(1882),
    [anon_sym_land] = ACTIONS(1882),
    [anon_sym_lor] = ACTIONS(1882),
    [anon_sym_lxor] = ACTIONS(1882),
    [aux_sym_mult_operator_token1] = ACTIONS(1882),
    [aux_sym_add_operator_token1] = ACTIONS(1882),
    [sym_concat_operator] = ACTIONS(1884),
    [sym_rel_operator] = ACTIONS(1882),
    [anon_sym_AMP_AMP] = ACTIONS(1882),
    [anon_sym_or] = ACTIONS(1882),
    [anon_sym_PIPE_PIPE] = ACTIONS(1882),
    [sym_assign_operator] = ACTIONS(1884),
    [sym__uppercase_identifier] = ACTIONS(1884),
    [aux_sym_tag_token1] = ACTIONS(1884),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(831)] = {
    [sym_attribute] = STATE(831),
    [sym__lowercase_identifier] = ACTIONS(1886),
    [anon_sym_and] = ACTIONS(1886),
    [anon_sym_TILDE] = ACTIONS(1886),
    [anon_sym_QMARK] = ACTIONS(1886),
    [anon_sym_LPAREN] = ACTIONS(1888),
    [anon_sym_RPAREN] = ACTIONS(1888),
    [anon_sym_COMMA] = ACTIONS(1888),
    [anon_sym_PIPE] = ACTIONS(1886),
    [anon_sym_COLON2] = ACTIONS(1886),
    [anon_sym_DOT] = ACTIONS(1888),
    [anon_sym_DASH_GT] = ACTIONS(1886),
    [anon_sym_LBRACK] = ACTIONS(1886),
    [anon_sym_RBRACK] = ACTIONS(1888),
    [anon_sym_true] = ACTIONS(1886),
    [anon_sym_false] = ACTIONS(1886),
    [anon_sym_LBRACE] = ACTIONS(1886),
    [anon_sym_SEMI] = ACTIONS(1888),
    [anon_sym_RBRACE] = ACTIONS(1888),
    [anon_sym_constraint] = ACTIONS(1886),
    [anon_sym_val] = ACTIONS(1886),
    [anon_sym_end] = ACTIONS(1886),
    [anon_sym_with] = ACTIONS(1886),
    [anon_sym_COLON_GT] = ACTIONS(1888),
    [anon_sym_object] = ACTIONS(1886),
    [anon_sym_inherit] = ACTIONS(1886),
    [anon_sym_method] = ACTIONS(1886),
    [anon_sym_in] = ACTIONS(1886),
    [anon_sym_as] = ACTIONS(1886),
    [anon_sym_initializer] = ACTIONS(1886),
    [anon_sym_AMP] = ACTIONS(1886),
    [anon_sym_POUND] = ACTIONS(1886),
    [anon_sym_COLON_COLON] = ACTIONS(1888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1888),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1888),
    [anon_sym_then] = ACTIONS(1886),
    [anon_sym_else] = ACTIONS(1886),
    [anon_sym_do] = ACTIONS(1886),
    [anon_sym_done] = ACTIONS(1886),
    [anon_sym_to] = ACTIONS(1886),
    [anon_sym_downto] = ACTIONS(1886),
    [anon_sym_new] = ACTIONS(1886),
    [anon_sym_LBRACE_LT] = ACTIONS(1888),
    [anon_sym_GT_RBRACE] = ACTIONS(1888),
    [anon_sym_begin] = ACTIONS(1886),
    [sym_ocamlyacc_value] = ACTIONS(1888),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1886),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1888),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1886),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1886),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1888),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1888),
    [aux_sym_number_token1] = ACTIONS(1888),
    [aux_sym_signed_number_token1] = ACTIONS(1886),
    [anon_sym_SQUOTE] = ACTIONS(1888),
    [anon_sym_DQUOTE] = ACTIONS(1888),
    [sym_prefix_operator] = ACTIONS(1888),
    [aux_sym_sign_operator_token1] = ACTIONS(1886),
    [sym_hash_operator] = ACTIONS(1888),
    [anon_sym_lsl] = ACTIONS(1886),
    [anon_sym_lsr] = ACTIONS(1886),
    [anon_sym_asr] = ACTIONS(1886),
    [aux_sym_pow_operator_token1] = ACTIONS(1888),
    [anon_sym_mod] = ACTIONS(1886),
    [anon_sym_land] = ACTIONS(1886),
    [anon_sym_lor] = ACTIONS(1886),
    [anon_sym_lxor] = ACTIONS(1886),
    [aux_sym_mult_operator_token1] = ACTIONS(1886),
    [aux_sym_add_operator_token1] = ACTIONS(1886),
    [sym_concat_operator] = ACTIONS(1888),
    [sym_rel_operator] = ACTIONS(1886),
    [anon_sym_AMP_AMP] = ACTIONS(1886),
    [anon_sym_or] = ACTIONS(1886),
    [anon_sym_PIPE_PIPE] = ACTIONS(1886),
    [sym_assign_operator] = ACTIONS(1888),
    [sym__uppercase_identifier] = ACTIONS(1888),
    [aux_sym_tag_token1] = ACTIONS(1888),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(832)] = {
    [sym_attribute] = STATE(832),
    [sym__lowercase_identifier] = ACTIONS(1890),
    [anon_sym_and] = ACTIONS(1890),
    [anon_sym_TILDE] = ACTIONS(1890),
    [anon_sym_QMARK] = ACTIONS(1890),
    [anon_sym_LPAREN] = ACTIONS(1892),
    [anon_sym_RPAREN] = ACTIONS(1892),
    [anon_sym_COMMA] = ACTIONS(1892),
    [anon_sym_PIPE] = ACTIONS(1890),
    [anon_sym_COLON2] = ACTIONS(1890),
    [anon_sym_DOT] = ACTIONS(1892),
    [anon_sym_DASH_GT] = ACTIONS(1890),
    [anon_sym_LBRACK] = ACTIONS(1890),
    [anon_sym_RBRACK] = ACTIONS(1892),
    [anon_sym_true] = ACTIONS(1890),
    [anon_sym_false] = ACTIONS(1890),
    [anon_sym_LBRACE] = ACTIONS(1890),
    [anon_sym_SEMI] = ACTIONS(1892),
    [anon_sym_RBRACE] = ACTIONS(1892),
    [anon_sym_constraint] = ACTIONS(1890),
    [anon_sym_val] = ACTIONS(1890),
    [anon_sym_end] = ACTIONS(1890),
    [anon_sym_with] = ACTIONS(1890),
    [anon_sym_COLON_GT] = ACTIONS(1892),
    [anon_sym_object] = ACTIONS(1890),
    [anon_sym_inherit] = ACTIONS(1890),
    [anon_sym_method] = ACTIONS(1890),
    [anon_sym_in] = ACTIONS(1890),
    [anon_sym_as] = ACTIONS(1890),
    [anon_sym_initializer] = ACTIONS(1890),
    [anon_sym_AMP] = ACTIONS(1890),
    [anon_sym_POUND] = ACTIONS(1890),
    [anon_sym_COLON_COLON] = ACTIONS(1892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1892),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1892),
    [anon_sym_then] = ACTIONS(1890),
    [anon_sym_else] = ACTIONS(1890),
    [anon_sym_do] = ACTIONS(1890),
    [anon_sym_done] = ACTIONS(1890),
    [anon_sym_to] = ACTIONS(1890),
    [anon_sym_downto] = ACTIONS(1890),
    [anon_sym_new] = ACTIONS(1890),
    [anon_sym_LBRACE_LT] = ACTIONS(1892),
    [anon_sym_GT_RBRACE] = ACTIONS(1892),
    [anon_sym_begin] = ACTIONS(1890),
    [sym_ocamlyacc_value] = ACTIONS(1892),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1890),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1892),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1890),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1890),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1892),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1892),
    [aux_sym_number_token1] = ACTIONS(1892),
    [aux_sym_signed_number_token1] = ACTIONS(1890),
    [anon_sym_SQUOTE] = ACTIONS(1892),
    [anon_sym_DQUOTE] = ACTIONS(1892),
    [sym_prefix_operator] = ACTIONS(1892),
    [aux_sym_sign_operator_token1] = ACTIONS(1890),
    [sym_hash_operator] = ACTIONS(1892),
    [anon_sym_lsl] = ACTIONS(1890),
    [anon_sym_lsr] = ACTIONS(1890),
    [anon_sym_asr] = ACTIONS(1890),
    [aux_sym_pow_operator_token1] = ACTIONS(1892),
    [anon_sym_mod] = ACTIONS(1890),
    [anon_sym_land] = ACTIONS(1890),
    [anon_sym_lor] = ACTIONS(1890),
    [anon_sym_lxor] = ACTIONS(1890),
    [aux_sym_mult_operator_token1] = ACTIONS(1890),
    [aux_sym_add_operator_token1] = ACTIONS(1890),
    [sym_concat_operator] = ACTIONS(1892),
    [sym_rel_operator] = ACTIONS(1890),
    [anon_sym_AMP_AMP] = ACTIONS(1890),
    [anon_sym_or] = ACTIONS(1890),
    [anon_sym_PIPE_PIPE] = ACTIONS(1890),
    [sym_assign_operator] = ACTIONS(1892),
    [sym__uppercase_identifier] = ACTIONS(1892),
    [aux_sym_tag_token1] = ACTIONS(1892),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(833)] = {
    [sym_attribute] = STATE(833),
    [sym__lowercase_identifier] = ACTIONS(1894),
    [anon_sym_and] = ACTIONS(1894),
    [anon_sym_TILDE] = ACTIONS(1894),
    [anon_sym_QMARK] = ACTIONS(1894),
    [anon_sym_LPAREN] = ACTIONS(1896),
    [anon_sym_RPAREN] = ACTIONS(1896),
    [anon_sym_COMMA] = ACTIONS(1896),
    [anon_sym_PIPE] = ACTIONS(1894),
    [anon_sym_COLON2] = ACTIONS(1894),
    [anon_sym_DOT] = ACTIONS(745),
    [anon_sym_DASH_GT] = ACTIONS(1894),
    [anon_sym_LBRACK] = ACTIONS(1894),
    [anon_sym_RBRACK] = ACTIONS(1896),
    [anon_sym_true] = ACTIONS(1894),
    [anon_sym_false] = ACTIONS(1894),
    [anon_sym_LBRACE] = ACTIONS(1894),
    [anon_sym_SEMI] = ACTIONS(1896),
    [anon_sym_RBRACE] = ACTIONS(1896),
    [anon_sym_constraint] = ACTIONS(1894),
    [anon_sym_val] = ACTIONS(1894),
    [anon_sym_end] = ACTIONS(1894),
    [anon_sym_with] = ACTIONS(1894),
    [anon_sym_COLON_GT] = ACTIONS(1896),
    [anon_sym_object] = ACTIONS(1894),
    [anon_sym_inherit] = ACTIONS(1894),
    [anon_sym_method] = ACTIONS(1894),
    [anon_sym_in] = ACTIONS(1894),
    [anon_sym_as] = ACTIONS(1894),
    [anon_sym_initializer] = ACTIONS(1894),
    [anon_sym_AMP] = ACTIONS(1894),
    [anon_sym_POUND] = ACTIONS(1894),
    [anon_sym_COLON_COLON] = ACTIONS(1896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1896),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1896),
    [anon_sym_then] = ACTIONS(1894),
    [anon_sym_else] = ACTIONS(1894),
    [anon_sym_do] = ACTIONS(1894),
    [anon_sym_done] = ACTIONS(1894),
    [anon_sym_to] = ACTIONS(1894),
    [anon_sym_downto] = ACTIONS(1894),
    [anon_sym_new] = ACTIONS(1894),
    [anon_sym_LBRACE_LT] = ACTIONS(1896),
    [anon_sym_GT_RBRACE] = ACTIONS(1896),
    [anon_sym_begin] = ACTIONS(1894),
    [sym_ocamlyacc_value] = ACTIONS(1896),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1894),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1896),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1894),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1894),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1896),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1896),
    [aux_sym_number_token1] = ACTIONS(1896),
    [aux_sym_signed_number_token1] = ACTIONS(1894),
    [anon_sym_SQUOTE] = ACTIONS(1896),
    [anon_sym_DQUOTE] = ACTIONS(1896),
    [sym_prefix_operator] = ACTIONS(1896),
    [aux_sym_sign_operator_token1] = ACTIONS(1894),
    [sym_hash_operator] = ACTIONS(1896),
    [anon_sym_lsl] = ACTIONS(1894),
    [anon_sym_lsr] = ACTIONS(1894),
    [anon_sym_asr] = ACTIONS(1894),
    [aux_sym_pow_operator_token1] = ACTIONS(1896),
    [anon_sym_mod] = ACTIONS(1894),
    [anon_sym_land] = ACTIONS(1894),
    [anon_sym_lor] = ACTIONS(1894),
    [anon_sym_lxor] = ACTIONS(1894),
    [aux_sym_mult_operator_token1] = ACTIONS(1894),
    [aux_sym_add_operator_token1] = ACTIONS(1894),
    [sym_concat_operator] = ACTIONS(1896),
    [sym_rel_operator] = ACTIONS(1894),
    [anon_sym_AMP_AMP] = ACTIONS(1894),
    [anon_sym_or] = ACTIONS(1894),
    [anon_sym_PIPE_PIPE] = ACTIONS(1894),
    [sym_assign_operator] = ACTIONS(1896),
    [sym__uppercase_identifier] = ACTIONS(1896),
    [aux_sym_tag_token1] = ACTIONS(1896),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(834)] = {
    [sym_attribute] = STATE(834),
    [sym__lowercase_identifier] = ACTIONS(1898),
    [anon_sym_and] = ACTIONS(1898),
    [anon_sym_TILDE] = ACTIONS(1898),
    [anon_sym_QMARK] = ACTIONS(1898),
    [anon_sym_LPAREN] = ACTIONS(1900),
    [anon_sym_RPAREN] = ACTIONS(1900),
    [anon_sym_COMMA] = ACTIONS(1900),
    [anon_sym_PIPE] = ACTIONS(1898),
    [anon_sym_COLON2] = ACTIONS(1898),
    [anon_sym_DOT] = ACTIONS(1900),
    [anon_sym_DASH_GT] = ACTIONS(1898),
    [anon_sym_LBRACK] = ACTIONS(1898),
    [anon_sym_RBRACK] = ACTIONS(1900),
    [anon_sym_true] = ACTIONS(1898),
    [anon_sym_false] = ACTIONS(1898),
    [anon_sym_LBRACE] = ACTIONS(1898),
    [anon_sym_SEMI] = ACTIONS(1900),
    [anon_sym_RBRACE] = ACTIONS(1900),
    [anon_sym_constraint] = ACTIONS(1898),
    [anon_sym_val] = ACTIONS(1898),
    [anon_sym_end] = ACTIONS(1898),
    [anon_sym_with] = ACTIONS(1898),
    [anon_sym_COLON_GT] = ACTIONS(1900),
    [anon_sym_object] = ACTIONS(1898),
    [anon_sym_inherit] = ACTIONS(1898),
    [anon_sym_method] = ACTIONS(1898),
    [anon_sym_in] = ACTIONS(1898),
    [anon_sym_as] = ACTIONS(1898),
    [anon_sym_initializer] = ACTIONS(1898),
    [anon_sym_AMP] = ACTIONS(1898),
    [anon_sym_POUND] = ACTIONS(1898),
    [anon_sym_COLON_COLON] = ACTIONS(1900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1900),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1900),
    [anon_sym_then] = ACTIONS(1898),
    [anon_sym_else] = ACTIONS(1898),
    [anon_sym_do] = ACTIONS(1898),
    [anon_sym_done] = ACTIONS(1898),
    [anon_sym_to] = ACTIONS(1898),
    [anon_sym_downto] = ACTIONS(1898),
    [anon_sym_new] = ACTIONS(1898),
    [anon_sym_LBRACE_LT] = ACTIONS(1900),
    [anon_sym_GT_RBRACE] = ACTIONS(1900),
    [anon_sym_begin] = ACTIONS(1898),
    [sym_ocamlyacc_value] = ACTIONS(1900),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1898),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1900),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1898),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1898),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1900),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1900),
    [aux_sym_number_token1] = ACTIONS(1900),
    [aux_sym_signed_number_token1] = ACTIONS(1898),
    [anon_sym_SQUOTE] = ACTIONS(1900),
    [anon_sym_DQUOTE] = ACTIONS(1900),
    [sym_prefix_operator] = ACTIONS(1900),
    [aux_sym_sign_operator_token1] = ACTIONS(1898),
    [sym_hash_operator] = ACTIONS(1900),
    [anon_sym_lsl] = ACTIONS(1898),
    [anon_sym_lsr] = ACTIONS(1898),
    [anon_sym_asr] = ACTIONS(1898),
    [aux_sym_pow_operator_token1] = ACTIONS(1900),
    [anon_sym_mod] = ACTIONS(1898),
    [anon_sym_land] = ACTIONS(1898),
    [anon_sym_lor] = ACTIONS(1898),
    [anon_sym_lxor] = ACTIONS(1898),
    [aux_sym_mult_operator_token1] = ACTIONS(1898),
    [aux_sym_add_operator_token1] = ACTIONS(1898),
    [sym_concat_operator] = ACTIONS(1900),
    [sym_rel_operator] = ACTIONS(1898),
    [anon_sym_AMP_AMP] = ACTIONS(1898),
    [anon_sym_or] = ACTIONS(1898),
    [anon_sym_PIPE_PIPE] = ACTIONS(1898),
    [sym_assign_operator] = ACTIONS(1900),
    [sym__uppercase_identifier] = ACTIONS(1900),
    [aux_sym_tag_token1] = ACTIONS(1900),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(835)] = {
    [sym_attribute] = STATE(835),
    [sym__lowercase_identifier] = ACTIONS(1902),
    [anon_sym_and] = ACTIONS(1902),
    [anon_sym_TILDE] = ACTIONS(1902),
    [anon_sym_QMARK] = ACTIONS(1902),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_RPAREN] = ACTIONS(1904),
    [anon_sym_COMMA] = ACTIONS(1904),
    [anon_sym_PIPE] = ACTIONS(1902),
    [anon_sym_COLON2] = ACTIONS(1902),
    [anon_sym_DOT] = ACTIONS(1904),
    [anon_sym_DASH_GT] = ACTIONS(1902),
    [anon_sym_LBRACK] = ACTIONS(1902),
    [anon_sym_RBRACK] = ACTIONS(1904),
    [anon_sym_true] = ACTIONS(1902),
    [anon_sym_false] = ACTIONS(1902),
    [anon_sym_LBRACE] = ACTIONS(1902),
    [anon_sym_SEMI] = ACTIONS(1904),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [anon_sym_constraint] = ACTIONS(1902),
    [anon_sym_val] = ACTIONS(1902),
    [anon_sym_end] = ACTIONS(1902),
    [anon_sym_with] = ACTIONS(1902),
    [anon_sym_COLON_GT] = ACTIONS(1904),
    [anon_sym_object] = ACTIONS(1902),
    [anon_sym_inherit] = ACTIONS(1902),
    [anon_sym_method] = ACTIONS(1902),
    [anon_sym_in] = ACTIONS(1902),
    [anon_sym_as] = ACTIONS(1902),
    [anon_sym_initializer] = ACTIONS(1902),
    [anon_sym_AMP] = ACTIONS(1902),
    [anon_sym_POUND] = ACTIONS(1902),
    [anon_sym_COLON_COLON] = ACTIONS(1904),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1904),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1904),
    [anon_sym_then] = ACTIONS(1902),
    [anon_sym_else] = ACTIONS(1902),
    [anon_sym_do] = ACTIONS(1902),
    [anon_sym_done] = ACTIONS(1902),
    [anon_sym_to] = ACTIONS(1902),
    [anon_sym_downto] = ACTIONS(1902),
    [anon_sym_new] = ACTIONS(1902),
    [anon_sym_LBRACE_LT] = ACTIONS(1904),
    [anon_sym_GT_RBRACE] = ACTIONS(1904),
    [anon_sym_begin] = ACTIONS(1902),
    [sym_ocamlyacc_value] = ACTIONS(1904),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1902),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1904),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1902),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1902),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1904),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1904),
    [aux_sym_number_token1] = ACTIONS(1904),
    [aux_sym_signed_number_token1] = ACTIONS(1902),
    [anon_sym_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1904),
    [sym_prefix_operator] = ACTIONS(1904),
    [aux_sym_sign_operator_token1] = ACTIONS(1902),
    [sym_hash_operator] = ACTIONS(1904),
    [anon_sym_lsl] = ACTIONS(1902),
    [anon_sym_lsr] = ACTIONS(1902),
    [anon_sym_asr] = ACTIONS(1902),
    [aux_sym_pow_operator_token1] = ACTIONS(1904),
    [anon_sym_mod] = ACTIONS(1902),
    [anon_sym_land] = ACTIONS(1902),
    [anon_sym_lor] = ACTIONS(1902),
    [anon_sym_lxor] = ACTIONS(1902),
    [aux_sym_mult_operator_token1] = ACTIONS(1902),
    [aux_sym_add_operator_token1] = ACTIONS(1902),
    [sym_concat_operator] = ACTIONS(1904),
    [sym_rel_operator] = ACTIONS(1902),
    [anon_sym_AMP_AMP] = ACTIONS(1902),
    [anon_sym_or] = ACTIONS(1902),
    [anon_sym_PIPE_PIPE] = ACTIONS(1902),
    [sym_assign_operator] = ACTIONS(1904),
    [sym__uppercase_identifier] = ACTIONS(1904),
    [aux_sym_tag_token1] = ACTIONS(1904),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(836)] = {
    [sym_attribute] = STATE(836),
    [sym__lowercase_identifier] = ACTIONS(1906),
    [anon_sym_and] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1906),
    [anon_sym_QMARK] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1908),
    [anon_sym_RPAREN] = ACTIONS(1908),
    [anon_sym_COMMA] = ACTIONS(1908),
    [anon_sym_PIPE] = ACTIONS(1906),
    [anon_sym_COLON2] = ACTIONS(1906),
    [anon_sym_DOT] = ACTIONS(1908),
    [anon_sym_DASH_GT] = ACTIONS(1906),
    [anon_sym_LBRACK] = ACTIONS(1906),
    [anon_sym_RBRACK] = ACTIONS(1908),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_LBRACE] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1908),
    [anon_sym_RBRACE] = ACTIONS(1908),
    [anon_sym_constraint] = ACTIONS(1906),
    [anon_sym_val] = ACTIONS(1906),
    [anon_sym_end] = ACTIONS(1906),
    [anon_sym_with] = ACTIONS(1906),
    [anon_sym_COLON_GT] = ACTIONS(1908),
    [anon_sym_object] = ACTIONS(1906),
    [anon_sym_inherit] = ACTIONS(1906),
    [anon_sym_method] = ACTIONS(1906),
    [anon_sym_in] = ACTIONS(1906),
    [anon_sym_as] = ACTIONS(1906),
    [anon_sym_initializer] = ACTIONS(1906),
    [anon_sym_AMP] = ACTIONS(1906),
    [anon_sym_POUND] = ACTIONS(1906),
    [anon_sym_COLON_COLON] = ACTIONS(1908),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1908),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1908),
    [anon_sym_then] = ACTIONS(1906),
    [anon_sym_else] = ACTIONS(1906),
    [anon_sym_do] = ACTIONS(1906),
    [anon_sym_done] = ACTIONS(1906),
    [anon_sym_to] = ACTIONS(1906),
    [anon_sym_downto] = ACTIONS(1906),
    [anon_sym_new] = ACTIONS(1906),
    [anon_sym_LBRACE_LT] = ACTIONS(1908),
    [anon_sym_GT_RBRACE] = ACTIONS(1908),
    [anon_sym_begin] = ACTIONS(1906),
    [sym_ocamlyacc_value] = ACTIONS(1908),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1906),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1908),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1906),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1906),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1908),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1908),
    [aux_sym_number_token1] = ACTIONS(1908),
    [aux_sym_signed_number_token1] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1908),
    [anon_sym_DQUOTE] = ACTIONS(1908),
    [sym_prefix_operator] = ACTIONS(1908),
    [aux_sym_sign_operator_token1] = ACTIONS(1906),
    [sym_hash_operator] = ACTIONS(1908),
    [anon_sym_lsl] = ACTIONS(1906),
    [anon_sym_lsr] = ACTIONS(1906),
    [anon_sym_asr] = ACTIONS(1906),
    [aux_sym_pow_operator_token1] = ACTIONS(1908),
    [anon_sym_mod] = ACTIONS(1906),
    [anon_sym_land] = ACTIONS(1906),
    [anon_sym_lor] = ACTIONS(1906),
    [anon_sym_lxor] = ACTIONS(1906),
    [aux_sym_mult_operator_token1] = ACTIONS(1906),
    [aux_sym_add_operator_token1] = ACTIONS(1906),
    [sym_concat_operator] = ACTIONS(1908),
    [sym_rel_operator] = ACTIONS(1906),
    [anon_sym_AMP_AMP] = ACTIONS(1906),
    [anon_sym_or] = ACTIONS(1906),
    [anon_sym_PIPE_PIPE] = ACTIONS(1906),
    [sym_assign_operator] = ACTIONS(1908),
    [sym__uppercase_identifier] = ACTIONS(1908),
    [aux_sym_tag_token1] = ACTIONS(1908),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(837)] = {
    [sym_attribute] = STATE(837),
    [sym__lowercase_identifier] = ACTIONS(1910),
    [anon_sym_and] = ACTIONS(1910),
    [anon_sym_TILDE] = ACTIONS(1910),
    [anon_sym_QMARK] = ACTIONS(1910),
    [anon_sym_LPAREN] = ACTIONS(1912),
    [anon_sym_RPAREN] = ACTIONS(1912),
    [anon_sym_COMMA] = ACTIONS(1912),
    [anon_sym_PIPE] = ACTIONS(1910),
    [anon_sym_COLON2] = ACTIONS(1910),
    [anon_sym_DOT] = ACTIONS(1912),
    [anon_sym_DASH_GT] = ACTIONS(1910),
    [anon_sym_LBRACK] = ACTIONS(1910),
    [anon_sym_RBRACK] = ACTIONS(1912),
    [anon_sym_true] = ACTIONS(1910),
    [anon_sym_false] = ACTIONS(1910),
    [anon_sym_LBRACE] = ACTIONS(1910),
    [anon_sym_SEMI] = ACTIONS(1912),
    [anon_sym_RBRACE] = ACTIONS(1912),
    [anon_sym_constraint] = ACTIONS(1910),
    [anon_sym_val] = ACTIONS(1910),
    [anon_sym_end] = ACTIONS(1910),
    [anon_sym_with] = ACTIONS(1910),
    [anon_sym_COLON_GT] = ACTIONS(1912),
    [anon_sym_object] = ACTIONS(1910),
    [anon_sym_inherit] = ACTIONS(1910),
    [anon_sym_method] = ACTIONS(1910),
    [anon_sym_in] = ACTIONS(1910),
    [anon_sym_as] = ACTIONS(1910),
    [anon_sym_initializer] = ACTIONS(1910),
    [anon_sym_AMP] = ACTIONS(1910),
    [anon_sym_POUND] = ACTIONS(1910),
    [anon_sym_COLON_COLON] = ACTIONS(1912),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1912),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1912),
    [anon_sym_then] = ACTIONS(1910),
    [anon_sym_else] = ACTIONS(1910),
    [anon_sym_do] = ACTIONS(1910),
    [anon_sym_done] = ACTIONS(1910),
    [anon_sym_to] = ACTIONS(1910),
    [anon_sym_downto] = ACTIONS(1910),
    [anon_sym_new] = ACTIONS(1910),
    [anon_sym_LBRACE_LT] = ACTIONS(1912),
    [anon_sym_GT_RBRACE] = ACTIONS(1912),
    [anon_sym_begin] = ACTIONS(1910),
    [sym_ocamlyacc_value] = ACTIONS(1912),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1910),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1912),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1910),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1910),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1912),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1912),
    [aux_sym_number_token1] = ACTIONS(1912),
    [aux_sym_signed_number_token1] = ACTIONS(1910),
    [anon_sym_SQUOTE] = ACTIONS(1912),
    [anon_sym_DQUOTE] = ACTIONS(1912),
    [sym_prefix_operator] = ACTIONS(1912),
    [aux_sym_sign_operator_token1] = ACTIONS(1910),
    [sym_hash_operator] = ACTIONS(1912),
    [anon_sym_lsl] = ACTIONS(1910),
    [anon_sym_lsr] = ACTIONS(1910),
    [anon_sym_asr] = ACTIONS(1910),
    [aux_sym_pow_operator_token1] = ACTIONS(1912),
    [anon_sym_mod] = ACTIONS(1910),
    [anon_sym_land] = ACTIONS(1910),
    [anon_sym_lor] = ACTIONS(1910),
    [anon_sym_lxor] = ACTIONS(1910),
    [aux_sym_mult_operator_token1] = ACTIONS(1910),
    [aux_sym_add_operator_token1] = ACTIONS(1910),
    [sym_concat_operator] = ACTIONS(1912),
    [sym_rel_operator] = ACTIONS(1910),
    [anon_sym_AMP_AMP] = ACTIONS(1910),
    [anon_sym_or] = ACTIONS(1910),
    [anon_sym_PIPE_PIPE] = ACTIONS(1910),
    [sym_assign_operator] = ACTIONS(1912),
    [sym__uppercase_identifier] = ACTIONS(1912),
    [aux_sym_tag_token1] = ACTIONS(1912),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(838)] = {
    [sym_attribute] = STATE(838),
    [sym__lowercase_identifier] = ACTIONS(1914),
    [anon_sym_and] = ACTIONS(1914),
    [anon_sym_TILDE] = ACTIONS(1914),
    [anon_sym_QMARK] = ACTIONS(1914),
    [anon_sym_LPAREN] = ACTIONS(1916),
    [anon_sym_RPAREN] = ACTIONS(1916),
    [anon_sym_COMMA] = ACTIONS(1916),
    [anon_sym_PIPE] = ACTIONS(1914),
    [anon_sym_COLON2] = ACTIONS(1914),
    [anon_sym_DOT] = ACTIONS(1916),
    [anon_sym_DASH_GT] = ACTIONS(1914),
    [anon_sym_LBRACK] = ACTIONS(1914),
    [anon_sym_RBRACK] = ACTIONS(1916),
    [anon_sym_true] = ACTIONS(1914),
    [anon_sym_false] = ACTIONS(1914),
    [anon_sym_LBRACE] = ACTIONS(1914),
    [anon_sym_SEMI] = ACTIONS(1916),
    [anon_sym_RBRACE] = ACTIONS(1916),
    [anon_sym_constraint] = ACTIONS(1914),
    [anon_sym_val] = ACTIONS(1914),
    [anon_sym_end] = ACTIONS(1914),
    [anon_sym_with] = ACTIONS(1914),
    [anon_sym_COLON_GT] = ACTIONS(1916),
    [anon_sym_object] = ACTIONS(1914),
    [anon_sym_inherit] = ACTIONS(1914),
    [anon_sym_method] = ACTIONS(1914),
    [anon_sym_in] = ACTIONS(1914),
    [anon_sym_as] = ACTIONS(1914),
    [anon_sym_initializer] = ACTIONS(1914),
    [anon_sym_AMP] = ACTIONS(1914),
    [anon_sym_POUND] = ACTIONS(1914),
    [anon_sym_COLON_COLON] = ACTIONS(1916),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1916),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1916),
    [anon_sym_then] = ACTIONS(1914),
    [anon_sym_else] = ACTIONS(1914),
    [anon_sym_do] = ACTIONS(1914),
    [anon_sym_done] = ACTIONS(1914),
    [anon_sym_to] = ACTIONS(1914),
    [anon_sym_downto] = ACTIONS(1914),
    [anon_sym_new] = ACTIONS(1914),
    [anon_sym_LBRACE_LT] = ACTIONS(1916),
    [anon_sym_GT_RBRACE] = ACTIONS(1916),
    [anon_sym_begin] = ACTIONS(1914),
    [sym_ocamlyacc_value] = ACTIONS(1916),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1914),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1916),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1914),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1914),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1916),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1916),
    [aux_sym_number_token1] = ACTIONS(1916),
    [aux_sym_signed_number_token1] = ACTIONS(1914),
    [anon_sym_SQUOTE] = ACTIONS(1916),
    [anon_sym_DQUOTE] = ACTIONS(1916),
    [sym_prefix_operator] = ACTIONS(1916),
    [aux_sym_sign_operator_token1] = ACTIONS(1914),
    [sym_hash_operator] = ACTIONS(1916),
    [anon_sym_lsl] = ACTIONS(1914),
    [anon_sym_lsr] = ACTIONS(1914),
    [anon_sym_asr] = ACTIONS(1914),
    [aux_sym_pow_operator_token1] = ACTIONS(1916),
    [anon_sym_mod] = ACTIONS(1914),
    [anon_sym_land] = ACTIONS(1914),
    [anon_sym_lor] = ACTIONS(1914),
    [anon_sym_lxor] = ACTIONS(1914),
    [aux_sym_mult_operator_token1] = ACTIONS(1914),
    [aux_sym_add_operator_token1] = ACTIONS(1914),
    [sym_concat_operator] = ACTIONS(1916),
    [sym_rel_operator] = ACTIONS(1914),
    [anon_sym_AMP_AMP] = ACTIONS(1914),
    [anon_sym_or] = ACTIONS(1914),
    [anon_sym_PIPE_PIPE] = ACTIONS(1914),
    [sym_assign_operator] = ACTIONS(1916),
    [sym__uppercase_identifier] = ACTIONS(1916),
    [aux_sym_tag_token1] = ACTIONS(1916),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(839)] = {
    [sym_attribute] = STATE(839),
    [sym__lowercase_identifier] = ACTIONS(1918),
    [anon_sym_and] = ACTIONS(1918),
    [anon_sym_TILDE] = ACTIONS(1918),
    [anon_sym_QMARK] = ACTIONS(1918),
    [anon_sym_LPAREN] = ACTIONS(1920),
    [anon_sym_RPAREN] = ACTIONS(1920),
    [anon_sym_COMMA] = ACTIONS(1920),
    [anon_sym_PIPE] = ACTIONS(1918),
    [anon_sym_COLON2] = ACTIONS(1918),
    [anon_sym_DOT] = ACTIONS(1920),
    [anon_sym_DASH_GT] = ACTIONS(1918),
    [anon_sym_LBRACK] = ACTIONS(1918),
    [anon_sym_RBRACK] = ACTIONS(1920),
    [anon_sym_true] = ACTIONS(1918),
    [anon_sym_false] = ACTIONS(1918),
    [anon_sym_LBRACE] = ACTIONS(1918),
    [anon_sym_SEMI] = ACTIONS(1920),
    [anon_sym_RBRACE] = ACTIONS(1920),
    [anon_sym_constraint] = ACTIONS(1918),
    [anon_sym_val] = ACTIONS(1918),
    [anon_sym_end] = ACTIONS(1918),
    [anon_sym_with] = ACTIONS(1918),
    [anon_sym_COLON_GT] = ACTIONS(1920),
    [anon_sym_object] = ACTIONS(1918),
    [anon_sym_inherit] = ACTIONS(1918),
    [anon_sym_method] = ACTIONS(1918),
    [anon_sym_in] = ACTIONS(1918),
    [anon_sym_as] = ACTIONS(1918),
    [anon_sym_initializer] = ACTIONS(1918),
    [anon_sym_AMP] = ACTIONS(1918),
    [anon_sym_POUND] = ACTIONS(1918),
    [anon_sym_COLON_COLON] = ACTIONS(1920),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1920),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1920),
    [anon_sym_then] = ACTIONS(1918),
    [anon_sym_else] = ACTIONS(1918),
    [anon_sym_do] = ACTIONS(1918),
    [anon_sym_done] = ACTIONS(1918),
    [anon_sym_to] = ACTIONS(1918),
    [anon_sym_downto] = ACTIONS(1918),
    [anon_sym_new] = ACTIONS(1918),
    [anon_sym_LBRACE_LT] = ACTIONS(1920),
    [anon_sym_GT_RBRACE] = ACTIONS(1920),
    [anon_sym_begin] = ACTIONS(1918),
    [sym_ocamlyacc_value] = ACTIONS(1920),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1918),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1920),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1918),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1918),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1920),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1920),
    [aux_sym_number_token1] = ACTIONS(1920),
    [aux_sym_signed_number_token1] = ACTIONS(1918),
    [anon_sym_SQUOTE] = ACTIONS(1920),
    [anon_sym_DQUOTE] = ACTIONS(1920),
    [sym_prefix_operator] = ACTIONS(1920),
    [aux_sym_sign_operator_token1] = ACTIONS(1918),
    [sym_hash_operator] = ACTIONS(1920),
    [anon_sym_lsl] = ACTIONS(1918),
    [anon_sym_lsr] = ACTIONS(1918),
    [anon_sym_asr] = ACTIONS(1918),
    [aux_sym_pow_operator_token1] = ACTIONS(1920),
    [anon_sym_mod] = ACTIONS(1918),
    [anon_sym_land] = ACTIONS(1918),
    [anon_sym_lor] = ACTIONS(1918),
    [anon_sym_lxor] = ACTIONS(1918),
    [aux_sym_mult_operator_token1] = ACTIONS(1918),
    [aux_sym_add_operator_token1] = ACTIONS(1918),
    [sym_concat_operator] = ACTIONS(1920),
    [sym_rel_operator] = ACTIONS(1918),
    [anon_sym_AMP_AMP] = ACTIONS(1918),
    [anon_sym_or] = ACTIONS(1918),
    [anon_sym_PIPE_PIPE] = ACTIONS(1918),
    [sym_assign_operator] = ACTIONS(1920),
    [sym__uppercase_identifier] = ACTIONS(1920),
    [aux_sym_tag_token1] = ACTIONS(1920),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(840)] = {
    [sym_attribute] = STATE(840),
    [sym__lowercase_identifier] = ACTIONS(1922),
    [anon_sym_and] = ACTIONS(1922),
    [anon_sym_TILDE] = ACTIONS(1922),
    [anon_sym_QMARK] = ACTIONS(1922),
    [anon_sym_LPAREN] = ACTIONS(1924),
    [anon_sym_RPAREN] = ACTIONS(1924),
    [anon_sym_COMMA] = ACTIONS(1924),
    [anon_sym_PIPE] = ACTIONS(1922),
    [anon_sym_COLON2] = ACTIONS(1922),
    [anon_sym_DOT] = ACTIONS(1924),
    [anon_sym_DASH_GT] = ACTIONS(1922),
    [anon_sym_LBRACK] = ACTIONS(1922),
    [anon_sym_RBRACK] = ACTIONS(1924),
    [anon_sym_true] = ACTIONS(1922),
    [anon_sym_false] = ACTIONS(1922),
    [anon_sym_LBRACE] = ACTIONS(1922),
    [anon_sym_SEMI] = ACTIONS(1924),
    [anon_sym_RBRACE] = ACTIONS(1924),
    [anon_sym_constraint] = ACTIONS(1922),
    [anon_sym_val] = ACTIONS(1922),
    [anon_sym_end] = ACTIONS(1922),
    [anon_sym_with] = ACTIONS(1922),
    [anon_sym_COLON_GT] = ACTIONS(1924),
    [anon_sym_object] = ACTIONS(1922),
    [anon_sym_inherit] = ACTIONS(1922),
    [anon_sym_method] = ACTIONS(1922),
    [anon_sym_in] = ACTIONS(1922),
    [anon_sym_as] = ACTIONS(1922),
    [anon_sym_initializer] = ACTIONS(1922),
    [anon_sym_AMP] = ACTIONS(1922),
    [anon_sym_POUND] = ACTIONS(1922),
    [anon_sym_COLON_COLON] = ACTIONS(1924),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1924),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1924),
    [anon_sym_then] = ACTIONS(1922),
    [anon_sym_else] = ACTIONS(1922),
    [anon_sym_do] = ACTIONS(1922),
    [anon_sym_done] = ACTIONS(1922),
    [anon_sym_to] = ACTIONS(1922),
    [anon_sym_downto] = ACTIONS(1922),
    [anon_sym_new] = ACTIONS(1922),
    [anon_sym_LBRACE_LT] = ACTIONS(1924),
    [anon_sym_GT_RBRACE] = ACTIONS(1924),
    [anon_sym_begin] = ACTIONS(1922),
    [sym_ocamlyacc_value] = ACTIONS(1924),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1922),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1924),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1922),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1922),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1924),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1924),
    [aux_sym_number_token1] = ACTIONS(1924),
    [aux_sym_signed_number_token1] = ACTIONS(1922),
    [anon_sym_SQUOTE] = ACTIONS(1924),
    [anon_sym_DQUOTE] = ACTIONS(1924),
    [sym_prefix_operator] = ACTIONS(1924),
    [aux_sym_sign_operator_token1] = ACTIONS(1922),
    [sym_hash_operator] = ACTIONS(1924),
    [anon_sym_lsl] = ACTIONS(1922),
    [anon_sym_lsr] = ACTIONS(1922),
    [anon_sym_asr] = ACTIONS(1922),
    [aux_sym_pow_operator_token1] = ACTIONS(1924),
    [anon_sym_mod] = ACTIONS(1922),
    [anon_sym_land] = ACTIONS(1922),
    [anon_sym_lor] = ACTIONS(1922),
    [anon_sym_lxor] = ACTIONS(1922),
    [aux_sym_mult_operator_token1] = ACTIONS(1922),
    [aux_sym_add_operator_token1] = ACTIONS(1922),
    [sym_concat_operator] = ACTIONS(1924),
    [sym_rel_operator] = ACTIONS(1922),
    [anon_sym_AMP_AMP] = ACTIONS(1922),
    [anon_sym_or] = ACTIONS(1922),
    [anon_sym_PIPE_PIPE] = ACTIONS(1922),
    [sym_assign_operator] = ACTIONS(1924),
    [sym__uppercase_identifier] = ACTIONS(1924),
    [aux_sym_tag_token1] = ACTIONS(1924),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(841)] = {
    [sym_attribute] = STATE(841),
    [sym__lowercase_identifier] = ACTIONS(1926),
    [anon_sym_and] = ACTIONS(1926),
    [anon_sym_TILDE] = ACTIONS(1926),
    [anon_sym_QMARK] = ACTIONS(1926),
    [anon_sym_LPAREN] = ACTIONS(1928),
    [anon_sym_RPAREN] = ACTIONS(1928),
    [anon_sym_COMMA] = ACTIONS(1928),
    [anon_sym_PIPE] = ACTIONS(1926),
    [anon_sym_COLON2] = ACTIONS(1926),
    [anon_sym_DOT] = ACTIONS(1928),
    [anon_sym_DASH_GT] = ACTIONS(1926),
    [anon_sym_LBRACK] = ACTIONS(1926),
    [anon_sym_RBRACK] = ACTIONS(1928),
    [anon_sym_true] = ACTIONS(1926),
    [anon_sym_false] = ACTIONS(1926),
    [anon_sym_LBRACE] = ACTIONS(1926),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym_RBRACE] = ACTIONS(1928),
    [anon_sym_constraint] = ACTIONS(1926),
    [anon_sym_val] = ACTIONS(1926),
    [anon_sym_end] = ACTIONS(1926),
    [anon_sym_with] = ACTIONS(1926),
    [anon_sym_COLON_GT] = ACTIONS(1928),
    [anon_sym_object] = ACTIONS(1926),
    [anon_sym_inherit] = ACTIONS(1926),
    [anon_sym_method] = ACTIONS(1926),
    [anon_sym_in] = ACTIONS(1926),
    [anon_sym_as] = ACTIONS(1926),
    [anon_sym_initializer] = ACTIONS(1926),
    [anon_sym_AMP] = ACTIONS(1926),
    [anon_sym_POUND] = ACTIONS(1926),
    [anon_sym_COLON_COLON] = ACTIONS(1928),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1928),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1928),
    [anon_sym_then] = ACTIONS(1926),
    [anon_sym_else] = ACTIONS(1926),
    [anon_sym_do] = ACTIONS(1926),
    [anon_sym_done] = ACTIONS(1926),
    [anon_sym_to] = ACTIONS(1926),
    [anon_sym_downto] = ACTIONS(1926),
    [anon_sym_new] = ACTIONS(1926),
    [anon_sym_LBRACE_LT] = ACTIONS(1928),
    [anon_sym_GT_RBRACE] = ACTIONS(1928),
    [anon_sym_begin] = ACTIONS(1926),
    [sym_ocamlyacc_value] = ACTIONS(1928),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1926),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1928),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1926),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1926),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1928),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1928),
    [aux_sym_number_token1] = ACTIONS(1928),
    [aux_sym_signed_number_token1] = ACTIONS(1926),
    [anon_sym_SQUOTE] = ACTIONS(1928),
    [anon_sym_DQUOTE] = ACTIONS(1928),
    [sym_prefix_operator] = ACTIONS(1928),
    [aux_sym_sign_operator_token1] = ACTIONS(1926),
    [sym_hash_operator] = ACTIONS(1928),
    [anon_sym_lsl] = ACTIONS(1926),
    [anon_sym_lsr] = ACTIONS(1926),
    [anon_sym_asr] = ACTIONS(1926),
    [aux_sym_pow_operator_token1] = ACTIONS(1928),
    [anon_sym_mod] = ACTIONS(1926),
    [anon_sym_land] = ACTIONS(1926),
    [anon_sym_lor] = ACTIONS(1926),
    [anon_sym_lxor] = ACTIONS(1926),
    [aux_sym_mult_operator_token1] = ACTIONS(1926),
    [aux_sym_add_operator_token1] = ACTIONS(1926),
    [sym_concat_operator] = ACTIONS(1928),
    [sym_rel_operator] = ACTIONS(1926),
    [anon_sym_AMP_AMP] = ACTIONS(1926),
    [anon_sym_or] = ACTIONS(1926),
    [anon_sym_PIPE_PIPE] = ACTIONS(1926),
    [sym_assign_operator] = ACTIONS(1928),
    [sym__uppercase_identifier] = ACTIONS(1928),
    [aux_sym_tag_token1] = ACTIONS(1928),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(842)] = {
    [sym_attribute] = STATE(842),
    [sym__lowercase_identifier] = ACTIONS(1930),
    [anon_sym_and] = ACTIONS(1930),
    [anon_sym_TILDE] = ACTIONS(1930),
    [anon_sym_QMARK] = ACTIONS(1930),
    [anon_sym_LPAREN] = ACTIONS(1932),
    [anon_sym_RPAREN] = ACTIONS(1932),
    [anon_sym_COMMA] = ACTIONS(1932),
    [anon_sym_PIPE] = ACTIONS(1930),
    [anon_sym_COLON2] = ACTIONS(1930),
    [anon_sym_DOT] = ACTIONS(1932),
    [anon_sym_DASH_GT] = ACTIONS(1930),
    [anon_sym_LBRACK] = ACTIONS(1930),
    [anon_sym_RBRACK] = ACTIONS(1932),
    [anon_sym_true] = ACTIONS(1930),
    [anon_sym_false] = ACTIONS(1930),
    [anon_sym_LBRACE] = ACTIONS(1930),
    [anon_sym_SEMI] = ACTIONS(1932),
    [anon_sym_RBRACE] = ACTIONS(1932),
    [anon_sym_constraint] = ACTIONS(1930),
    [anon_sym_val] = ACTIONS(1930),
    [anon_sym_end] = ACTIONS(1930),
    [anon_sym_with] = ACTIONS(1930),
    [anon_sym_COLON_GT] = ACTIONS(1932),
    [anon_sym_object] = ACTIONS(1930),
    [anon_sym_inherit] = ACTIONS(1930),
    [anon_sym_method] = ACTIONS(1930),
    [anon_sym_in] = ACTIONS(1930),
    [anon_sym_as] = ACTIONS(1930),
    [anon_sym_initializer] = ACTIONS(1930),
    [anon_sym_AMP] = ACTIONS(1930),
    [anon_sym_POUND] = ACTIONS(1930),
    [anon_sym_COLON_COLON] = ACTIONS(1932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1932),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1932),
    [anon_sym_then] = ACTIONS(1930),
    [anon_sym_else] = ACTIONS(1930),
    [anon_sym_do] = ACTIONS(1930),
    [anon_sym_done] = ACTIONS(1930),
    [anon_sym_to] = ACTIONS(1930),
    [anon_sym_downto] = ACTIONS(1930),
    [anon_sym_new] = ACTIONS(1930),
    [anon_sym_LBRACE_LT] = ACTIONS(1932),
    [anon_sym_GT_RBRACE] = ACTIONS(1932),
    [anon_sym_begin] = ACTIONS(1930),
    [sym_ocamlyacc_value] = ACTIONS(1932),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1930),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1932),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1930),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1930),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1932),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1932),
    [aux_sym_number_token1] = ACTIONS(1932),
    [aux_sym_signed_number_token1] = ACTIONS(1930),
    [anon_sym_SQUOTE] = ACTIONS(1932),
    [anon_sym_DQUOTE] = ACTIONS(1932),
    [sym_prefix_operator] = ACTIONS(1932),
    [aux_sym_sign_operator_token1] = ACTIONS(1930),
    [sym_hash_operator] = ACTIONS(1932),
    [anon_sym_lsl] = ACTIONS(1930),
    [anon_sym_lsr] = ACTIONS(1930),
    [anon_sym_asr] = ACTIONS(1930),
    [aux_sym_pow_operator_token1] = ACTIONS(1932),
    [anon_sym_mod] = ACTIONS(1930),
    [anon_sym_land] = ACTIONS(1930),
    [anon_sym_lor] = ACTIONS(1930),
    [anon_sym_lxor] = ACTIONS(1930),
    [aux_sym_mult_operator_token1] = ACTIONS(1930),
    [aux_sym_add_operator_token1] = ACTIONS(1930),
    [sym_concat_operator] = ACTIONS(1932),
    [sym_rel_operator] = ACTIONS(1930),
    [anon_sym_AMP_AMP] = ACTIONS(1930),
    [anon_sym_or] = ACTIONS(1930),
    [anon_sym_PIPE_PIPE] = ACTIONS(1930),
    [sym_assign_operator] = ACTIONS(1932),
    [sym__uppercase_identifier] = ACTIONS(1932),
    [aux_sym_tag_token1] = ACTIONS(1932),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(843)] = {
    [sym_attribute] = STATE(843),
    [sym__lowercase_identifier] = ACTIONS(1934),
    [anon_sym_and] = ACTIONS(1934),
    [anon_sym_TILDE] = ACTIONS(1934),
    [anon_sym_QMARK] = ACTIONS(1934),
    [anon_sym_LPAREN] = ACTIONS(1936),
    [anon_sym_RPAREN] = ACTIONS(1936),
    [anon_sym_COMMA] = ACTIONS(1936),
    [anon_sym_PIPE] = ACTIONS(1934),
    [anon_sym_COLON2] = ACTIONS(1934),
    [anon_sym_DOT] = ACTIONS(1936),
    [anon_sym_DASH_GT] = ACTIONS(1934),
    [anon_sym_LBRACK] = ACTIONS(1934),
    [anon_sym_RBRACK] = ACTIONS(1936),
    [anon_sym_true] = ACTIONS(1934),
    [anon_sym_false] = ACTIONS(1934),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_SEMI] = ACTIONS(1936),
    [anon_sym_RBRACE] = ACTIONS(1936),
    [anon_sym_constraint] = ACTIONS(1934),
    [anon_sym_val] = ACTIONS(1934),
    [anon_sym_end] = ACTIONS(1934),
    [anon_sym_with] = ACTIONS(1934),
    [anon_sym_COLON_GT] = ACTIONS(1936),
    [anon_sym_object] = ACTIONS(1934),
    [anon_sym_inherit] = ACTIONS(1934),
    [anon_sym_method] = ACTIONS(1934),
    [anon_sym_in] = ACTIONS(1934),
    [anon_sym_as] = ACTIONS(1934),
    [anon_sym_initializer] = ACTIONS(1934),
    [anon_sym_AMP] = ACTIONS(1934),
    [anon_sym_POUND] = ACTIONS(1934),
    [anon_sym_COLON_COLON] = ACTIONS(1936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1936),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1936),
    [anon_sym_then] = ACTIONS(1934),
    [anon_sym_else] = ACTIONS(1934),
    [anon_sym_do] = ACTIONS(1934),
    [anon_sym_done] = ACTIONS(1934),
    [anon_sym_to] = ACTIONS(1934),
    [anon_sym_downto] = ACTIONS(1934),
    [anon_sym_new] = ACTIONS(1934),
    [anon_sym_LBRACE_LT] = ACTIONS(1936),
    [anon_sym_GT_RBRACE] = ACTIONS(1936),
    [anon_sym_begin] = ACTIONS(1934),
    [sym_ocamlyacc_value] = ACTIONS(1936),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1934),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1934),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1934),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1936),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1936),
    [aux_sym_number_token1] = ACTIONS(1936),
    [aux_sym_signed_number_token1] = ACTIONS(1934),
    [anon_sym_SQUOTE] = ACTIONS(1936),
    [anon_sym_DQUOTE] = ACTIONS(1936),
    [sym_prefix_operator] = ACTIONS(1936),
    [aux_sym_sign_operator_token1] = ACTIONS(1934),
    [sym_hash_operator] = ACTIONS(1936),
    [anon_sym_lsl] = ACTIONS(1934),
    [anon_sym_lsr] = ACTIONS(1934),
    [anon_sym_asr] = ACTIONS(1934),
    [aux_sym_pow_operator_token1] = ACTIONS(1936),
    [anon_sym_mod] = ACTIONS(1934),
    [anon_sym_land] = ACTIONS(1934),
    [anon_sym_lor] = ACTIONS(1934),
    [anon_sym_lxor] = ACTIONS(1934),
    [aux_sym_mult_operator_token1] = ACTIONS(1934),
    [aux_sym_add_operator_token1] = ACTIONS(1934),
    [sym_concat_operator] = ACTIONS(1936),
    [sym_rel_operator] = ACTIONS(1934),
    [anon_sym_AMP_AMP] = ACTIONS(1934),
    [anon_sym_or] = ACTIONS(1934),
    [anon_sym_PIPE_PIPE] = ACTIONS(1934),
    [sym_assign_operator] = ACTIONS(1936),
    [sym__uppercase_identifier] = ACTIONS(1936),
    [aux_sym_tag_token1] = ACTIONS(1936),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(844)] = {
    [sym_attribute] = STATE(844),
    [sym__lowercase_identifier] = ACTIONS(1938),
    [anon_sym_and] = ACTIONS(1938),
    [anon_sym_TILDE] = ACTIONS(1938),
    [anon_sym_QMARK] = ACTIONS(1938),
    [anon_sym_LPAREN] = ACTIONS(1940),
    [anon_sym_RPAREN] = ACTIONS(1940),
    [anon_sym_COMMA] = ACTIONS(1940),
    [anon_sym_PIPE] = ACTIONS(1938),
    [anon_sym_COLON2] = ACTIONS(1938),
    [anon_sym_DOT] = ACTIONS(1940),
    [anon_sym_DASH_GT] = ACTIONS(1938),
    [anon_sym_LBRACK] = ACTIONS(1938),
    [anon_sym_RBRACK] = ACTIONS(1940),
    [anon_sym_true] = ACTIONS(1938),
    [anon_sym_false] = ACTIONS(1938),
    [anon_sym_LBRACE] = ACTIONS(1938),
    [anon_sym_SEMI] = ACTIONS(1940),
    [anon_sym_RBRACE] = ACTIONS(1940),
    [anon_sym_constraint] = ACTIONS(1938),
    [anon_sym_val] = ACTIONS(1938),
    [anon_sym_end] = ACTIONS(1938),
    [anon_sym_with] = ACTIONS(1938),
    [anon_sym_COLON_GT] = ACTIONS(1940),
    [anon_sym_object] = ACTIONS(1938),
    [anon_sym_inherit] = ACTIONS(1938),
    [anon_sym_method] = ACTIONS(1938),
    [anon_sym_in] = ACTIONS(1938),
    [anon_sym_as] = ACTIONS(1938),
    [anon_sym_initializer] = ACTIONS(1938),
    [anon_sym_AMP] = ACTIONS(1938),
    [anon_sym_POUND] = ACTIONS(1938),
    [anon_sym_COLON_COLON] = ACTIONS(1940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1940),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1940),
    [anon_sym_then] = ACTIONS(1938),
    [anon_sym_else] = ACTIONS(1938),
    [anon_sym_do] = ACTIONS(1938),
    [anon_sym_done] = ACTIONS(1938),
    [anon_sym_to] = ACTIONS(1938),
    [anon_sym_downto] = ACTIONS(1938),
    [anon_sym_new] = ACTIONS(1938),
    [anon_sym_LBRACE_LT] = ACTIONS(1940),
    [anon_sym_GT_RBRACE] = ACTIONS(1940),
    [anon_sym_begin] = ACTIONS(1938),
    [sym_ocamlyacc_value] = ACTIONS(1940),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1938),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1940),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1938),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1938),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1940),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1940),
    [aux_sym_number_token1] = ACTIONS(1940),
    [aux_sym_signed_number_token1] = ACTIONS(1938),
    [anon_sym_SQUOTE] = ACTIONS(1940),
    [anon_sym_DQUOTE] = ACTIONS(1940),
    [sym_prefix_operator] = ACTIONS(1940),
    [aux_sym_sign_operator_token1] = ACTIONS(1938),
    [sym_hash_operator] = ACTIONS(1940),
    [anon_sym_lsl] = ACTIONS(1938),
    [anon_sym_lsr] = ACTIONS(1938),
    [anon_sym_asr] = ACTIONS(1938),
    [aux_sym_pow_operator_token1] = ACTIONS(1940),
    [anon_sym_mod] = ACTIONS(1938),
    [anon_sym_land] = ACTIONS(1938),
    [anon_sym_lor] = ACTIONS(1938),
    [anon_sym_lxor] = ACTIONS(1938),
    [aux_sym_mult_operator_token1] = ACTIONS(1938),
    [aux_sym_add_operator_token1] = ACTIONS(1938),
    [sym_concat_operator] = ACTIONS(1940),
    [sym_rel_operator] = ACTIONS(1938),
    [anon_sym_AMP_AMP] = ACTIONS(1938),
    [anon_sym_or] = ACTIONS(1938),
    [anon_sym_PIPE_PIPE] = ACTIONS(1938),
    [sym_assign_operator] = ACTIONS(1940),
    [sym__uppercase_identifier] = ACTIONS(1940),
    [aux_sym_tag_token1] = ACTIONS(1940),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(845)] = {
    [sym_attribute] = STATE(845),
    [sym__lowercase_identifier] = ACTIONS(1942),
    [anon_sym_and] = ACTIONS(1942),
    [anon_sym_TILDE] = ACTIONS(1942),
    [anon_sym_QMARK] = ACTIONS(1942),
    [anon_sym_LPAREN] = ACTIONS(1944),
    [anon_sym_RPAREN] = ACTIONS(1944),
    [anon_sym_COMMA] = ACTIONS(1944),
    [anon_sym_PIPE] = ACTIONS(1942),
    [anon_sym_COLON2] = ACTIONS(1942),
    [anon_sym_DOT] = ACTIONS(1944),
    [anon_sym_DASH_GT] = ACTIONS(1942),
    [anon_sym_LBRACK] = ACTIONS(1942),
    [anon_sym_RBRACK] = ACTIONS(1944),
    [anon_sym_true] = ACTIONS(1942),
    [anon_sym_false] = ACTIONS(1942),
    [anon_sym_LBRACE] = ACTIONS(1942),
    [anon_sym_SEMI] = ACTIONS(1944),
    [anon_sym_RBRACE] = ACTIONS(1944),
    [anon_sym_constraint] = ACTIONS(1942),
    [anon_sym_val] = ACTIONS(1942),
    [anon_sym_end] = ACTIONS(1942),
    [anon_sym_with] = ACTIONS(1942),
    [anon_sym_COLON_GT] = ACTIONS(1944),
    [anon_sym_object] = ACTIONS(1942),
    [anon_sym_inherit] = ACTIONS(1942),
    [anon_sym_method] = ACTIONS(1942),
    [anon_sym_in] = ACTIONS(1942),
    [anon_sym_as] = ACTIONS(1942),
    [anon_sym_initializer] = ACTIONS(1942),
    [anon_sym_AMP] = ACTIONS(1942),
    [anon_sym_POUND] = ACTIONS(1942),
    [anon_sym_COLON_COLON] = ACTIONS(1944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1944),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1944),
    [anon_sym_then] = ACTIONS(1942),
    [anon_sym_else] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1942),
    [anon_sym_done] = ACTIONS(1942),
    [anon_sym_to] = ACTIONS(1942),
    [anon_sym_downto] = ACTIONS(1942),
    [anon_sym_new] = ACTIONS(1942),
    [anon_sym_LBRACE_LT] = ACTIONS(1944),
    [anon_sym_GT_RBRACE] = ACTIONS(1944),
    [anon_sym_begin] = ACTIONS(1942),
    [sym_ocamlyacc_value] = ACTIONS(1944),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1942),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1944),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1942),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1942),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1944),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1944),
    [aux_sym_number_token1] = ACTIONS(1944),
    [aux_sym_signed_number_token1] = ACTIONS(1942),
    [anon_sym_SQUOTE] = ACTIONS(1944),
    [anon_sym_DQUOTE] = ACTIONS(1944),
    [sym_prefix_operator] = ACTIONS(1944),
    [aux_sym_sign_operator_token1] = ACTIONS(1942),
    [sym_hash_operator] = ACTIONS(1944),
    [anon_sym_lsl] = ACTIONS(1942),
    [anon_sym_lsr] = ACTIONS(1942),
    [anon_sym_asr] = ACTIONS(1942),
    [aux_sym_pow_operator_token1] = ACTIONS(1944),
    [anon_sym_mod] = ACTIONS(1942),
    [anon_sym_land] = ACTIONS(1942),
    [anon_sym_lor] = ACTIONS(1942),
    [anon_sym_lxor] = ACTIONS(1942),
    [aux_sym_mult_operator_token1] = ACTIONS(1942),
    [aux_sym_add_operator_token1] = ACTIONS(1942),
    [sym_concat_operator] = ACTIONS(1944),
    [sym_rel_operator] = ACTIONS(1942),
    [anon_sym_AMP_AMP] = ACTIONS(1942),
    [anon_sym_or] = ACTIONS(1942),
    [anon_sym_PIPE_PIPE] = ACTIONS(1942),
    [sym_assign_operator] = ACTIONS(1944),
    [sym__uppercase_identifier] = ACTIONS(1944),
    [aux_sym_tag_token1] = ACTIONS(1944),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(846)] = {
    [sym_attribute] = STATE(846),
    [sym__lowercase_identifier] = ACTIONS(1946),
    [anon_sym_and] = ACTIONS(1946),
    [anon_sym_TILDE] = ACTIONS(1946),
    [anon_sym_QMARK] = ACTIONS(1946),
    [anon_sym_LPAREN] = ACTIONS(1948),
    [anon_sym_RPAREN] = ACTIONS(1948),
    [anon_sym_COMMA] = ACTIONS(1948),
    [anon_sym_PIPE] = ACTIONS(1946),
    [anon_sym_COLON2] = ACTIONS(1946),
    [anon_sym_DOT] = ACTIONS(1948),
    [anon_sym_DASH_GT] = ACTIONS(1946),
    [anon_sym_LBRACK] = ACTIONS(1946),
    [anon_sym_RBRACK] = ACTIONS(1948),
    [anon_sym_true] = ACTIONS(1946),
    [anon_sym_false] = ACTIONS(1946),
    [anon_sym_LBRACE] = ACTIONS(1946),
    [anon_sym_SEMI] = ACTIONS(1948),
    [anon_sym_RBRACE] = ACTIONS(1948),
    [anon_sym_constraint] = ACTIONS(1946),
    [anon_sym_val] = ACTIONS(1946),
    [anon_sym_end] = ACTIONS(1946),
    [anon_sym_with] = ACTIONS(1946),
    [anon_sym_COLON_GT] = ACTIONS(1948),
    [anon_sym_object] = ACTIONS(1946),
    [anon_sym_inherit] = ACTIONS(1946),
    [anon_sym_method] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(1946),
    [anon_sym_as] = ACTIONS(1946),
    [anon_sym_initializer] = ACTIONS(1946),
    [anon_sym_AMP] = ACTIONS(1946),
    [anon_sym_POUND] = ACTIONS(1946),
    [anon_sym_COLON_COLON] = ACTIONS(1948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1948),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1948),
    [anon_sym_then] = ACTIONS(1946),
    [anon_sym_else] = ACTIONS(1946),
    [anon_sym_do] = ACTIONS(1946),
    [anon_sym_done] = ACTIONS(1946),
    [anon_sym_to] = ACTIONS(1946),
    [anon_sym_downto] = ACTIONS(1946),
    [anon_sym_new] = ACTIONS(1946),
    [anon_sym_LBRACE_LT] = ACTIONS(1948),
    [anon_sym_GT_RBRACE] = ACTIONS(1948),
    [anon_sym_begin] = ACTIONS(1946),
    [sym_ocamlyacc_value] = ACTIONS(1948),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1946),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1948),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1946),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1946),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1948),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1948),
    [aux_sym_number_token1] = ACTIONS(1948),
    [aux_sym_signed_number_token1] = ACTIONS(1946),
    [anon_sym_SQUOTE] = ACTIONS(1948),
    [anon_sym_DQUOTE] = ACTIONS(1948),
    [sym_prefix_operator] = ACTIONS(1948),
    [aux_sym_sign_operator_token1] = ACTIONS(1946),
    [sym_hash_operator] = ACTIONS(1948),
    [anon_sym_lsl] = ACTIONS(1946),
    [anon_sym_lsr] = ACTIONS(1946),
    [anon_sym_asr] = ACTIONS(1946),
    [aux_sym_pow_operator_token1] = ACTIONS(1948),
    [anon_sym_mod] = ACTIONS(1946),
    [anon_sym_land] = ACTIONS(1946),
    [anon_sym_lor] = ACTIONS(1946),
    [anon_sym_lxor] = ACTIONS(1946),
    [aux_sym_mult_operator_token1] = ACTIONS(1946),
    [aux_sym_add_operator_token1] = ACTIONS(1946),
    [sym_concat_operator] = ACTIONS(1948),
    [sym_rel_operator] = ACTIONS(1946),
    [anon_sym_AMP_AMP] = ACTIONS(1946),
    [anon_sym_or] = ACTIONS(1946),
    [anon_sym_PIPE_PIPE] = ACTIONS(1946),
    [sym_assign_operator] = ACTIONS(1948),
    [sym__uppercase_identifier] = ACTIONS(1948),
    [aux_sym_tag_token1] = ACTIONS(1948),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(847)] = {
    [sym_attribute] = STATE(847),
    [sym__lowercase_identifier] = ACTIONS(1950),
    [anon_sym_and] = ACTIONS(1950),
    [anon_sym_TILDE] = ACTIONS(1950),
    [anon_sym_QMARK] = ACTIONS(1950),
    [anon_sym_LPAREN] = ACTIONS(1952),
    [anon_sym_RPAREN] = ACTIONS(1952),
    [anon_sym_COMMA] = ACTIONS(1952),
    [anon_sym_PIPE] = ACTIONS(1950),
    [anon_sym_COLON2] = ACTIONS(1950),
    [anon_sym_DOT] = ACTIONS(1952),
    [anon_sym_DASH_GT] = ACTIONS(1950),
    [anon_sym_LBRACK] = ACTIONS(1950),
    [anon_sym_RBRACK] = ACTIONS(1952),
    [anon_sym_true] = ACTIONS(1950),
    [anon_sym_false] = ACTIONS(1950),
    [anon_sym_LBRACE] = ACTIONS(1950),
    [anon_sym_SEMI] = ACTIONS(1952),
    [anon_sym_RBRACE] = ACTIONS(1952),
    [anon_sym_constraint] = ACTIONS(1950),
    [anon_sym_val] = ACTIONS(1950),
    [anon_sym_end] = ACTIONS(1950),
    [anon_sym_with] = ACTIONS(1950),
    [anon_sym_COLON_GT] = ACTIONS(1952),
    [anon_sym_object] = ACTIONS(1950),
    [anon_sym_inherit] = ACTIONS(1950),
    [anon_sym_method] = ACTIONS(1950),
    [anon_sym_in] = ACTIONS(1950),
    [anon_sym_as] = ACTIONS(1950),
    [anon_sym_initializer] = ACTIONS(1950),
    [anon_sym_AMP] = ACTIONS(1950),
    [anon_sym_POUND] = ACTIONS(1950),
    [anon_sym_COLON_COLON] = ACTIONS(1952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1952),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1952),
    [anon_sym_then] = ACTIONS(1950),
    [anon_sym_else] = ACTIONS(1950),
    [anon_sym_do] = ACTIONS(1950),
    [anon_sym_done] = ACTIONS(1950),
    [anon_sym_to] = ACTIONS(1950),
    [anon_sym_downto] = ACTIONS(1950),
    [anon_sym_new] = ACTIONS(1950),
    [anon_sym_LBRACE_LT] = ACTIONS(1952),
    [anon_sym_GT_RBRACE] = ACTIONS(1952),
    [anon_sym_begin] = ACTIONS(1950),
    [sym_ocamlyacc_value] = ACTIONS(1952),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1950),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1952),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1950),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1950),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1952),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1952),
    [aux_sym_number_token1] = ACTIONS(1952),
    [aux_sym_signed_number_token1] = ACTIONS(1950),
    [anon_sym_SQUOTE] = ACTIONS(1952),
    [anon_sym_DQUOTE] = ACTIONS(1952),
    [sym_prefix_operator] = ACTIONS(1952),
    [aux_sym_sign_operator_token1] = ACTIONS(1950),
    [sym_hash_operator] = ACTIONS(1952),
    [anon_sym_lsl] = ACTIONS(1950),
    [anon_sym_lsr] = ACTIONS(1950),
    [anon_sym_asr] = ACTIONS(1950),
    [aux_sym_pow_operator_token1] = ACTIONS(1952),
    [anon_sym_mod] = ACTIONS(1950),
    [anon_sym_land] = ACTIONS(1950),
    [anon_sym_lor] = ACTIONS(1950),
    [anon_sym_lxor] = ACTIONS(1950),
    [aux_sym_mult_operator_token1] = ACTIONS(1950),
    [aux_sym_add_operator_token1] = ACTIONS(1950),
    [sym_concat_operator] = ACTIONS(1952),
    [sym_rel_operator] = ACTIONS(1950),
    [anon_sym_AMP_AMP] = ACTIONS(1950),
    [anon_sym_or] = ACTIONS(1950),
    [anon_sym_PIPE_PIPE] = ACTIONS(1950),
    [sym_assign_operator] = ACTIONS(1952),
    [sym__uppercase_identifier] = ACTIONS(1952),
    [aux_sym_tag_token1] = ACTIONS(1952),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(848)] = {
    [sym_parameter] = STATE(2236),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(848),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_let_binding_repeat1] = STATE(848),
    [ts_builtin_sym_end] = ACTIONS(1954),
    [sym__lowercase_identifier] = ACTIONS(1956),
    [anon_sym_SEMI_SEMI] = ACTIONS(1954),
    [anon_sym_and] = ACTIONS(1959),
    [anon_sym_EQ] = ACTIONS(1954),
    [anon_sym_TILDE] = ACTIONS(1961),
    [anon_sym_QMARK] = ACTIONS(1961),
    [anon_sym_LPAREN] = ACTIONS(1964),
    [anon_sym_external] = ACTIONS(1959),
    [anon_sym_type] = ACTIONS(1959),
    [anon_sym_COLON2] = ACTIONS(1959),
    [anon_sym_DASH_GT] = ACTIONS(1954),
    [anon_sym_LBRACK] = ACTIONS(1967),
    [anon_sym_RBRACK] = ACTIONS(1954),
    [anon_sym_true] = ACTIONS(1970),
    [anon_sym_false] = ACTIONS(1970),
    [anon_sym_LBRACE] = ACTIONS(1973),
    [anon_sym_constraint] = ACTIONS(1959),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_module] = ACTIONS(1959),
    [anon_sym_open] = ACTIONS(1959),
    [anon_sym_include] = ACTIONS(1959),
    [anon_sym_class] = ACTIONS(1959),
    [anon_sym_val] = ACTIONS(1959),
    [anon_sym_end] = ACTIONS(1959),
    [anon_sym_COLON_GT] = ACTIONS(1954),
    [anon_sym_inherit] = ACTIONS(1959),
    [anon_sym_method] = ACTIONS(1959),
    [anon_sym_in] = ACTIONS(1959),
    [anon_sym_initializer] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1976),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1979),
    [anon_sym_begin] = ACTIONS(1982),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1959),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1954),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1985),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1988),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1954),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1954),
    [aux_sym_number_token1] = ACTIONS(1991),
    [aux_sym_signed_number_token1] = ACTIONS(1994),
    [anon_sym_SQUOTE] = ACTIONS(1997),
    [anon_sym_DQUOTE] = ACTIONS(2000),
    [sym__uppercase_identifier] = ACTIONS(2003),
    [aux_sym_tag_token1] = ACTIONS(2006),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(849)] = {
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(1251),
    [sym_typed_expression] = STATE(803),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_labeled_argument] = STATE(861),
    [sym_prefix_expression] = STATE(803),
    [sym_hash_expression] = STATE(803),
    [sym_field_get_expression] = STATE(803),
    [sym_array_get_expression] = STATE(803),
    [sym_string_get_expression] = STATE(803),
    [sym_bigarray_get_expression] = STATE(803),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(849),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [aux_sym_class_application_repeat1] = STATE(91),
    [sym__lowercase_identifier] = ACTIONS(737),
    [anon_sym_and] = ACTIONS(1544),
    [anon_sym_TILDE] = ACTIONS(1360),
    [anon_sym_QMARK] = ACTIONS(1360),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1542),
    [anon_sym_COLON2] = ACTIONS(1542),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_constraint] = ACTIONS(1544),
    [anon_sym_val] = ACTIONS(1544),
    [anon_sym_end] = ACTIONS(1544),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_inherit] = ACTIONS(1544),
    [anon_sym_method] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1544),
    [anon_sym_as] = ACTIONS(1544),
    [anon_sym_initializer] = ACTIONS(1544),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1544),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1542),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(381),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(383),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1542),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1542),
    [aux_sym_number_token1] = ACTIONS(385),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(850)] = {
    [sym_attribute] = STATE(850),
    [sym__lowercase_identifier] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_RPAREN] = ACTIONS(1616),
    [anon_sym_COMMA] = ACTIONS(1616),
    [anon_sym_PIPE] = ACTIONS(1614),
    [anon_sym_COLON2] = ACTIONS(1614),
    [anon_sym_DOT] = ACTIONS(1616),
    [anon_sym_DASH_GT] = ACTIONS(1614),
    [anon_sym_LBRACK] = ACTIONS(1614),
    [anon_sym_RBRACK] = ACTIONS(1616),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1616),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [anon_sym_constraint] = ACTIONS(1614),
    [anon_sym_val] = ACTIONS(1614),
    [anon_sym_end] = ACTIONS(1614),
    [anon_sym_with] = ACTIONS(1614),
    [anon_sym_COLON_GT] = ACTIONS(1616),
    [anon_sym_object] = ACTIONS(1614),
    [anon_sym_inherit] = ACTIONS(1614),
    [anon_sym_method] = ACTIONS(1614),
    [anon_sym_initializer] = ACTIONS(1614),
    [anon_sym_AMP] = ACTIONS(1614),
    [anon_sym_POUND] = ACTIONS(1614),
    [anon_sym_COLON_COLON] = ACTIONS(1616),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1616),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1616),
    [anon_sym_LT_DASH] = ACTIONS(2009),
    [anon_sym_then] = ACTIONS(1614),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_do] = ACTIONS(1614),
    [anon_sym_done] = ACTIONS(1614),
    [anon_sym_to] = ACTIONS(1614),
    [anon_sym_downto] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
    [anon_sym_LBRACE_LT] = ACTIONS(1616),
    [anon_sym_GT_RBRACE] = ACTIONS(1616),
    [anon_sym_begin] = ACTIONS(1614),
    [sym_ocamlyacc_value] = ACTIONS(1616),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1614),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1616),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1616),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1616),
    [aux_sym_number_token1] = ACTIONS(1616),
    [aux_sym_signed_number_token1] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1616),
    [sym_prefix_operator] = ACTIONS(1616),
    [aux_sym_sign_operator_token1] = ACTIONS(1614),
    [sym_hash_operator] = ACTIONS(1616),
    [anon_sym_lsl] = ACTIONS(1614),
    [anon_sym_lsr] = ACTIONS(1614),
    [anon_sym_asr] = ACTIONS(1614),
    [aux_sym_pow_operator_token1] = ACTIONS(1616),
    [anon_sym_mod] = ACTIONS(1614),
    [anon_sym_land] = ACTIONS(1614),
    [anon_sym_lor] = ACTIONS(1614),
    [anon_sym_lxor] = ACTIONS(1614),
    [aux_sym_mult_operator_token1] = ACTIONS(1614),
    [aux_sym_add_operator_token1] = ACTIONS(1614),
    [sym_concat_operator] = ACTIONS(1616),
    [sym_rel_operator] = ACTIONS(1614),
    [anon_sym_AMP_AMP] = ACTIONS(1614),
    [anon_sym_or] = ACTIONS(1614),
    [anon_sym_PIPE_PIPE] = ACTIONS(1614),
    [sym_assign_operator] = ACTIONS(1616),
    [sym__uppercase_identifier] = ACTIONS(1616),
    [aux_sym_tag_token1] = ACTIONS(1616),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(851)] = {
    [sym__parenthesized_abstract_type] = STATE(1369),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(3090),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3080),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6514),
    [sym__tuple_binding_pattern] = STATE(3085),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(851),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_and] = ACTIONS(1318),
    [anon_sym_EQ] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_QMARK] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(2011),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_COLON2] = ACTIONS(1318),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_COLON_GT] = ACTIONS(1316),
    [anon_sym_in] = ACTIONS(1318),
    [anon_sym_as] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_COLON_COLON] = ACTIONS(1316),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1316),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym_let_and_operator] = ACTIONS(1316),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(852)] = {
    [sym__delimited_expression] = STATE(803),
    [sym__simple_expression] = STATE(1251),
    [sym_typed_expression] = STATE(803),
    [sym_list_expression] = STATE(798),
    [sym_array_expression] = STATE(798),
    [sym_record_expression] = STATE(798),
    [sym_labeled_argument] = STATE(861),
    [sym_prefix_expression] = STATE(803),
    [sym_hash_expression] = STATE(803),
    [sym_field_get_expression] = STATE(803),
    [sym_array_get_expression] = STATE(803),
    [sym_string_get_expression] = STATE(803),
    [sym_bigarray_get_expression] = STATE(803),
    [sym_local_open_expression] = STATE(803),
    [sym_package_expression] = STATE(798),
    [sym_new_expression] = STATE(803),
    [sym_object_copy_expression] = STATE(798),
    [sym_method_invocation] = STATE(803),
    [sym_object_expression] = STATE(803),
    [sym_parenthesized_expression] = STATE(798),
    [sym_attribute] = STATE(852),
    [sym_extension] = STATE(803),
    [sym_quoted_extension] = STATE(803),
    [sym__constant] = STATE(803),
    [sym_number] = STATE(790),
    [sym_character] = STATE(790),
    [sym_string] = STATE(790),
    [sym_quoted_string] = STATE(790),
    [sym_boolean] = STATE(790),
    [sym_unit] = STATE(798),
    [sym__value_name] = STATE(801),
    [sym_parenthesized_operator] = STATE(793),
    [sym_value_path] = STATE(803),
    [sym_module_path] = STATE(6347),
    [sym_constructor_path] = STATE(803),
    [sym__constructor_name] = STATE(816),
    [sym_tag] = STATE(803),
    [aux_sym_class_application_repeat1] = STATE(849),
    [sym__lowercase_identifier] = ACTIONS(737),
    [anon_sym_and] = ACTIONS(1540),
    [anon_sym_TILDE] = ACTIONS(1360),
    [anon_sym_QMARK] = ACTIONS(1360),
    [anon_sym_LPAREN] = ACTIONS(343),
    [anon_sym_RPAREN] = ACTIONS(1538),
    [anon_sym_COLON2] = ACTIONS(1538),
    [anon_sym_LBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(347),
    [anon_sym_false] = ACTIONS(347),
    [anon_sym_LBRACE] = ACTIONS(349),
    [anon_sym_constraint] = ACTIONS(1540),
    [anon_sym_val] = ACTIONS(1540),
    [anon_sym_end] = ACTIONS(1540),
    [anon_sym_object] = ACTIONS(351),
    [anon_sym_inherit] = ACTIONS(1540),
    [anon_sym_method] = ACTIONS(1540),
    [anon_sym_in] = ACTIONS(1540),
    [anon_sym_as] = ACTIONS(1540),
    [anon_sym_initializer] = ACTIONS(1540),
    [anon_sym_LBRACK_PIPE] = ACTIONS(355),
    [anon_sym_new] = ACTIONS(1298),
    [anon_sym_LBRACE_LT] = ACTIONS(375),
    [anon_sym_begin] = ACTIONS(377),
    [sym_ocamlyacc_value] = ACTIONS(379),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1540),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1538),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(381),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(383),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1538),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1538),
    [aux_sym_number_token1] = ACTIONS(385),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE] = ACTIONS(389),
    [sym_prefix_operator] = ACTIONS(1300),
    [sym__uppercase_identifier] = ACTIONS(395),
    [aux_sym_tag_token1] = ACTIONS(397),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(853)] = {
    [sym_attribute] = STATE(853),
    [sym__lowercase_identifier] = ACTIONS(2021),
    [anon_sym_TILDE] = ACTIONS(2021),
    [anon_sym_QMARK] = ACTIONS(2021),
    [anon_sym_LPAREN] = ACTIONS(2023),
    [anon_sym_RPAREN] = ACTIONS(2023),
    [anon_sym_COMMA] = ACTIONS(2023),
    [anon_sym_PIPE] = ACTIONS(2021),
    [anon_sym_COLON2] = ACTIONS(2021),
    [anon_sym_DOT] = ACTIONS(2023),
    [anon_sym_DASH_GT] = ACTIONS(2021),
    [anon_sym_LBRACK] = ACTIONS(2021),
    [anon_sym_RBRACK] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(2021),
    [anon_sym_false] = ACTIONS(2021),
    [anon_sym_LBRACE] = ACTIONS(2021),
    [anon_sym_SEMI] = ACTIONS(2023),
    [anon_sym_RBRACE] = ACTIONS(2023),
    [anon_sym_constraint] = ACTIONS(2021),
    [anon_sym_val] = ACTIONS(2021),
    [anon_sym_end] = ACTIONS(2021),
    [anon_sym_with] = ACTIONS(2021),
    [anon_sym_COLON_GT] = ACTIONS(2023),
    [anon_sym_object] = ACTIONS(2021),
    [anon_sym_inherit] = ACTIONS(2021),
    [anon_sym_method] = ACTIONS(2021),
    [anon_sym_initializer] = ACTIONS(2021),
    [anon_sym_AMP] = ACTIONS(2021),
    [anon_sym_POUND] = ACTIONS(2021),
    [anon_sym_COLON_COLON] = ACTIONS(2023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2023),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2023),
    [anon_sym_then] = ACTIONS(2021),
    [anon_sym_else] = ACTIONS(2021),
    [anon_sym_do] = ACTIONS(2021),
    [anon_sym_done] = ACTIONS(2021),
    [anon_sym_to] = ACTIONS(2021),
    [anon_sym_downto] = ACTIONS(2021),
    [anon_sym_new] = ACTIONS(2021),
    [anon_sym_LBRACE_LT] = ACTIONS(2023),
    [anon_sym_GT_RBRACE] = ACTIONS(2023),
    [anon_sym_begin] = ACTIONS(2021),
    [sym_ocamlyacc_value] = ACTIONS(2023),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2023),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2023),
    [aux_sym_number_token1] = ACTIONS(2023),
    [aux_sym_signed_number_token1] = ACTIONS(2021),
    [anon_sym_SQUOTE] = ACTIONS(2023),
    [anon_sym_DQUOTE] = ACTIONS(2023),
    [sym_prefix_operator] = ACTIONS(2023),
    [aux_sym_sign_operator_token1] = ACTIONS(2021),
    [sym_hash_operator] = ACTIONS(2023),
    [anon_sym_lsl] = ACTIONS(2021),
    [anon_sym_lsr] = ACTIONS(2021),
    [anon_sym_asr] = ACTIONS(2021),
    [aux_sym_pow_operator_token1] = ACTIONS(2023),
    [anon_sym_mod] = ACTIONS(2021),
    [anon_sym_land] = ACTIONS(2021),
    [anon_sym_lor] = ACTIONS(2021),
    [anon_sym_lxor] = ACTIONS(2021),
    [aux_sym_mult_operator_token1] = ACTIONS(2021),
    [aux_sym_add_operator_token1] = ACTIONS(2021),
    [sym_concat_operator] = ACTIONS(2023),
    [sym_rel_operator] = ACTIONS(2021),
    [anon_sym_AMP_AMP] = ACTIONS(2021),
    [anon_sym_or] = ACTIONS(2021),
    [anon_sym_PIPE_PIPE] = ACTIONS(2021),
    [sym_assign_operator] = ACTIONS(2023),
    [sym__uppercase_identifier] = ACTIONS(2023),
    [aux_sym_tag_token1] = ACTIONS(2023),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(854)] = {
    [sym_attribute] = STATE(854),
    [sym__lowercase_identifier] = ACTIONS(2025),
    [anon_sym_TILDE] = ACTIONS(2025),
    [anon_sym_QMARK] = ACTIONS(2025),
    [anon_sym_LPAREN] = ACTIONS(2027),
    [anon_sym_RPAREN] = ACTIONS(2027),
    [anon_sym_COMMA] = ACTIONS(2027),
    [anon_sym_PIPE] = ACTIONS(2025),
    [anon_sym_COLON2] = ACTIONS(2025),
    [anon_sym_DOT] = ACTIONS(2027),
    [anon_sym_DASH_GT] = ACTIONS(2025),
    [anon_sym_LBRACK] = ACTIONS(2025),
    [anon_sym_RBRACK] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(2025),
    [anon_sym_false] = ACTIONS(2025),
    [anon_sym_LBRACE] = ACTIONS(2025),
    [anon_sym_SEMI] = ACTIONS(2027),
    [anon_sym_RBRACE] = ACTIONS(2027),
    [anon_sym_constraint] = ACTIONS(2025),
    [anon_sym_val] = ACTIONS(2025),
    [anon_sym_end] = ACTIONS(2025),
    [anon_sym_with] = ACTIONS(2025),
    [anon_sym_COLON_GT] = ACTIONS(2027),
    [anon_sym_object] = ACTIONS(2025),
    [anon_sym_inherit] = ACTIONS(2025),
    [anon_sym_method] = ACTIONS(2025),
    [anon_sym_initializer] = ACTIONS(2025),
    [anon_sym_AMP] = ACTIONS(2025),
    [anon_sym_POUND] = ACTIONS(2025),
    [anon_sym_COLON_COLON] = ACTIONS(2027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2027),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2027),
    [anon_sym_then] = ACTIONS(2025),
    [anon_sym_else] = ACTIONS(2025),
    [anon_sym_do] = ACTIONS(2025),
    [anon_sym_done] = ACTIONS(2025),
    [anon_sym_to] = ACTIONS(2025),
    [anon_sym_downto] = ACTIONS(2025),
    [anon_sym_new] = ACTIONS(2025),
    [anon_sym_LBRACE_LT] = ACTIONS(2027),
    [anon_sym_GT_RBRACE] = ACTIONS(2027),
    [anon_sym_begin] = ACTIONS(2025),
    [sym_ocamlyacc_value] = ACTIONS(2027),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2025),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2027),
    [aux_sym_number_token1] = ACTIONS(2027),
    [aux_sym_signed_number_token1] = ACTIONS(2025),
    [anon_sym_SQUOTE] = ACTIONS(2027),
    [anon_sym_DQUOTE] = ACTIONS(2027),
    [sym_prefix_operator] = ACTIONS(2027),
    [aux_sym_sign_operator_token1] = ACTIONS(2025),
    [sym_hash_operator] = ACTIONS(2027),
    [anon_sym_lsl] = ACTIONS(2025),
    [anon_sym_lsr] = ACTIONS(2025),
    [anon_sym_asr] = ACTIONS(2025),
    [aux_sym_pow_operator_token1] = ACTIONS(2027),
    [anon_sym_mod] = ACTIONS(2025),
    [anon_sym_land] = ACTIONS(2025),
    [anon_sym_lor] = ACTIONS(2025),
    [anon_sym_lxor] = ACTIONS(2025),
    [aux_sym_mult_operator_token1] = ACTIONS(2025),
    [aux_sym_add_operator_token1] = ACTIONS(2025),
    [sym_concat_operator] = ACTIONS(2027),
    [sym_rel_operator] = ACTIONS(2025),
    [anon_sym_AMP_AMP] = ACTIONS(2025),
    [anon_sym_or] = ACTIONS(2025),
    [anon_sym_PIPE_PIPE] = ACTIONS(2025),
    [sym_assign_operator] = ACTIONS(2027),
    [sym__uppercase_identifier] = ACTIONS(2027),
    [aux_sym_tag_token1] = ACTIONS(2027),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(855)] = {
    [sym_parameter] = STATE(2475),
    [sym__class_typed] = STATE(3630),
    [sym__parenthesized_abstract_type] = STATE(2475),
    [sym__delimited_pattern] = STATE(2450),
    [sym__simple_pattern] = STATE(2476),
    [sym_typed_pattern] = STATE(2450),
    [sym_polymorphic_variant_pattern] = STATE(2450),
    [sym_record_pattern] = STATE(2478),
    [sym_list_pattern] = STATE(2478),
    [sym_array_pattern] = STATE(2478),
    [sym_local_open_pattern] = STATE(2450),
    [sym_package_pattern] = STATE(2450),
    [sym_parenthesized_pattern] = STATE(2478),
    [sym_attribute] = STATE(855),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2450),
    [sym_quoted_extension] = STATE(2450),
    [sym__constant] = STATE(2464),
    [sym__signed_constant] = STATE(2450),
    [sym_number] = STATE(2504),
    [sym_signed_number] = STATE(2464),
    [sym_character] = STATE(2504),
    [sym_string] = STATE(2504),
    [sym_quoted_string] = STATE(2504),
    [sym_boolean] = STATE(2504),
    [sym_unit] = STATE(2478),
    [sym__simple_value_pattern] = STATE(2450),
    [sym_parenthesized_operator] = STATE(2450),
    [sym_module_path] = STATE(6504),
    [sym_constructor_path] = STATE(2450),
    [sym__constructor_name] = STATE(2515),
    [sym_tag] = STATE(2450),
    [aux_sym_expression_item_repeat1] = STATE(3617),
    [aux_sym_let_binding_repeat1] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(2029),
    [anon_sym_SEMI_SEMI] = ACTIONS(2031),
    [anon_sym_let] = ACTIONS(2033),
    [anon_sym_and] = ACTIONS(2033),
    [anon_sym_EQ] = ACTIONS(2035),
    [anon_sym_TILDE] = ACTIONS(2037),
    [anon_sym_QMARK] = ACTIONS(2037),
    [anon_sym_LPAREN] = ACTIONS(2039),
    [anon_sym_external] = ACTIONS(2033),
    [anon_sym_type] = ACTIONS(2033),
    [anon_sym_COLON2] = ACTIONS(2041),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2031),
    [anon_sym_true] = ACTIONS(2045),
    [anon_sym_false] = ACTIONS(2045),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2033),
    [anon_sym_module] = ACTIONS(2033),
    [anon_sym_open] = ACTIONS(2033),
    [anon_sym_include] = ACTIONS(2033),
    [anon_sym_class] = ACTIONS(2033),
    [anon_sym_val] = ACTIONS(2033),
    [anon_sym_end] = ACTIONS(2033),
    [anon_sym_POUND] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2051),
    [anon_sym_begin] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2031),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2031),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2067),
    [sym_let_operator] = ACTIONS(2031),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2033),
    [aux_sym_tag_token1] = ACTIONS(2071),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(856)] = {
    [sym_parameter] = STATE(2475),
    [sym__class_typed] = STATE(3631),
    [sym__parenthesized_abstract_type] = STATE(2475),
    [sym__delimited_pattern] = STATE(2450),
    [sym__simple_pattern] = STATE(2476),
    [sym_typed_pattern] = STATE(2450),
    [sym_polymorphic_variant_pattern] = STATE(2450),
    [sym_record_pattern] = STATE(2478),
    [sym_list_pattern] = STATE(2478),
    [sym_array_pattern] = STATE(2478),
    [sym_local_open_pattern] = STATE(2450),
    [sym_package_pattern] = STATE(2450),
    [sym_parenthesized_pattern] = STATE(2478),
    [sym_attribute] = STATE(856),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2450),
    [sym_quoted_extension] = STATE(2450),
    [sym__constant] = STATE(2464),
    [sym__signed_constant] = STATE(2450),
    [sym_number] = STATE(2504),
    [sym_signed_number] = STATE(2464),
    [sym_character] = STATE(2504),
    [sym_string] = STATE(2504),
    [sym_quoted_string] = STATE(2504),
    [sym_boolean] = STATE(2504),
    [sym_unit] = STATE(2478),
    [sym__simple_value_pattern] = STATE(2450),
    [sym_parenthesized_operator] = STATE(2450),
    [sym_module_path] = STATE(6504),
    [sym_constructor_path] = STATE(2450),
    [sym__constructor_name] = STATE(2515),
    [sym_tag] = STATE(2450),
    [aux_sym_expression_item_repeat1] = STATE(3645),
    [aux_sym_let_binding_repeat1] = STATE(1012),
    [sym__lowercase_identifier] = ACTIONS(2029),
    [anon_sym_SEMI_SEMI] = ACTIONS(2073),
    [anon_sym_let] = ACTIONS(2075),
    [anon_sym_and] = ACTIONS(2075),
    [anon_sym_EQ] = ACTIONS(2077),
    [anon_sym_TILDE] = ACTIONS(2037),
    [anon_sym_QMARK] = ACTIONS(2037),
    [anon_sym_LPAREN] = ACTIONS(2039),
    [anon_sym_external] = ACTIONS(2075),
    [anon_sym_type] = ACTIONS(2075),
    [anon_sym_COLON2] = ACTIONS(2041),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2073),
    [anon_sym_true] = ACTIONS(2045),
    [anon_sym_false] = ACTIONS(2045),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2075),
    [anon_sym_module] = ACTIONS(2075),
    [anon_sym_open] = ACTIONS(2075),
    [anon_sym_include] = ACTIONS(2075),
    [anon_sym_class] = ACTIONS(2075),
    [anon_sym_val] = ACTIONS(2075),
    [anon_sym_end] = ACTIONS(2075),
    [anon_sym_POUND] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2051),
    [anon_sym_begin] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2073),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2073),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2073),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2067),
    [sym_let_operator] = ACTIONS(2073),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2075),
    [aux_sym_tag_token1] = ACTIONS(2071),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(857)] = {
    [sym_attribute] = STATE(857),
    [sym__lowercase_identifier] = ACTIONS(1578),
    [anon_sym_TILDE] = ACTIONS(1578),
    [anon_sym_QMARK] = ACTIONS(1578),
    [anon_sym_LPAREN] = ACTIONS(1580),
    [anon_sym_RPAREN] = ACTIONS(1580),
    [anon_sym_COMMA] = ACTIONS(1580),
    [anon_sym_PIPE] = ACTIONS(1578),
    [anon_sym_COLON2] = ACTIONS(1578),
    [anon_sym_DOT] = ACTIONS(1580),
    [anon_sym_DASH_GT] = ACTIONS(1578),
    [anon_sym_LBRACK] = ACTIONS(1578),
    [anon_sym_RBRACK] = ACTIONS(1580),
    [anon_sym_true] = ACTIONS(1578),
    [anon_sym_false] = ACTIONS(1578),
    [anon_sym_LBRACE] = ACTIONS(1578),
    [anon_sym_SEMI] = ACTIONS(1580),
    [anon_sym_RBRACE] = ACTIONS(1580),
    [anon_sym_constraint] = ACTIONS(1578),
    [anon_sym_val] = ACTIONS(1578),
    [anon_sym_end] = ACTIONS(1578),
    [anon_sym_with] = ACTIONS(1578),
    [anon_sym_COLON_GT] = ACTIONS(1580),
    [anon_sym_object] = ACTIONS(1578),
    [anon_sym_inherit] = ACTIONS(1578),
    [anon_sym_method] = ACTIONS(1578),
    [anon_sym_initializer] = ACTIONS(1578),
    [anon_sym_AMP] = ACTIONS(1578),
    [anon_sym_POUND] = ACTIONS(1578),
    [anon_sym_COLON_COLON] = ACTIONS(1580),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1580),
    [anon_sym_LT_DASH] = ACTIONS(1578),
    [anon_sym_then] = ACTIONS(1578),
    [anon_sym_else] = ACTIONS(1578),
    [anon_sym_do] = ACTIONS(1578),
    [anon_sym_done] = ACTIONS(1578),
    [anon_sym_to] = ACTIONS(1578),
    [anon_sym_downto] = ACTIONS(1578),
    [anon_sym_new] = ACTIONS(1578),
    [anon_sym_LBRACE_LT] = ACTIONS(1580),
    [anon_sym_GT_RBRACE] = ACTIONS(1580),
    [anon_sym_begin] = ACTIONS(1578),
    [sym_ocamlyacc_value] = ACTIONS(1580),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1578),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1580),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1578),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1578),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1580),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1580),
    [aux_sym_number_token1] = ACTIONS(1580),
    [aux_sym_signed_number_token1] = ACTIONS(1578),
    [anon_sym_SQUOTE] = ACTIONS(1580),
    [anon_sym_DQUOTE] = ACTIONS(1580),
    [sym_prefix_operator] = ACTIONS(1580),
    [aux_sym_sign_operator_token1] = ACTIONS(1578),
    [sym_hash_operator] = ACTIONS(1580),
    [anon_sym_lsl] = ACTIONS(1578),
    [anon_sym_lsr] = ACTIONS(1578),
    [anon_sym_asr] = ACTIONS(1578),
    [aux_sym_pow_operator_token1] = ACTIONS(1580),
    [anon_sym_mod] = ACTIONS(1578),
    [anon_sym_land] = ACTIONS(1578),
    [anon_sym_lor] = ACTIONS(1578),
    [anon_sym_lxor] = ACTIONS(1578),
    [aux_sym_mult_operator_token1] = ACTIONS(1578),
    [aux_sym_add_operator_token1] = ACTIONS(1578),
    [sym_concat_operator] = ACTIONS(1580),
    [sym_rel_operator] = ACTIONS(1578),
    [anon_sym_AMP_AMP] = ACTIONS(1578),
    [anon_sym_or] = ACTIONS(1578),
    [anon_sym_PIPE_PIPE] = ACTIONS(1578),
    [sym_assign_operator] = ACTIONS(1580),
    [sym__uppercase_identifier] = ACTIONS(1580),
    [aux_sym_tag_token1] = ACTIONS(1580),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(858)] = {
    [sym_attribute] = STATE(858),
    [sym__lowercase_identifier] = ACTIONS(1546),
    [anon_sym_TILDE] = ACTIONS(1546),
    [anon_sym_QMARK] = ACTIONS(1546),
    [anon_sym_LPAREN] = ACTIONS(1548),
    [anon_sym_RPAREN] = ACTIONS(1548),
    [anon_sym_COMMA] = ACTIONS(1548),
    [anon_sym_PIPE] = ACTIONS(1546),
    [anon_sym_COLON2] = ACTIONS(1546),
    [anon_sym_DOT] = ACTIONS(1548),
    [anon_sym_DASH_GT] = ACTIONS(1546),
    [anon_sym_LBRACK] = ACTIONS(1546),
    [anon_sym_RBRACK] = ACTIONS(1548),
    [anon_sym_true] = ACTIONS(1546),
    [anon_sym_false] = ACTIONS(1546),
    [anon_sym_LBRACE] = ACTIONS(1546),
    [anon_sym_SEMI] = ACTIONS(1548),
    [anon_sym_RBRACE] = ACTIONS(1548),
    [anon_sym_constraint] = ACTIONS(1546),
    [anon_sym_val] = ACTIONS(1546),
    [anon_sym_end] = ACTIONS(1546),
    [anon_sym_with] = ACTIONS(1546),
    [anon_sym_COLON_GT] = ACTIONS(1548),
    [anon_sym_object] = ACTIONS(1546),
    [anon_sym_inherit] = ACTIONS(1546),
    [anon_sym_method] = ACTIONS(1546),
    [anon_sym_initializer] = ACTIONS(1546),
    [anon_sym_AMP] = ACTIONS(1546),
    [anon_sym_POUND] = ACTIONS(1546),
    [anon_sym_COLON_COLON] = ACTIONS(1548),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1548),
    [anon_sym_LT_DASH] = ACTIONS(1546),
    [anon_sym_then] = ACTIONS(1546),
    [anon_sym_else] = ACTIONS(1546),
    [anon_sym_do] = ACTIONS(1546),
    [anon_sym_done] = ACTIONS(1546),
    [anon_sym_to] = ACTIONS(1546),
    [anon_sym_downto] = ACTIONS(1546),
    [anon_sym_new] = ACTIONS(1546),
    [anon_sym_LBRACE_LT] = ACTIONS(1548),
    [anon_sym_GT_RBRACE] = ACTIONS(1548),
    [anon_sym_begin] = ACTIONS(1546),
    [sym_ocamlyacc_value] = ACTIONS(1548),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1546),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1548),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1546),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1546),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1548),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1548),
    [aux_sym_number_token1] = ACTIONS(1548),
    [aux_sym_signed_number_token1] = ACTIONS(1546),
    [anon_sym_SQUOTE] = ACTIONS(1548),
    [anon_sym_DQUOTE] = ACTIONS(1548),
    [sym_prefix_operator] = ACTIONS(1548),
    [aux_sym_sign_operator_token1] = ACTIONS(1546),
    [sym_hash_operator] = ACTIONS(1548),
    [anon_sym_lsl] = ACTIONS(1546),
    [anon_sym_lsr] = ACTIONS(1546),
    [anon_sym_asr] = ACTIONS(1546),
    [aux_sym_pow_operator_token1] = ACTIONS(1548),
    [anon_sym_mod] = ACTIONS(1546),
    [anon_sym_land] = ACTIONS(1546),
    [anon_sym_lor] = ACTIONS(1546),
    [anon_sym_lxor] = ACTIONS(1546),
    [aux_sym_mult_operator_token1] = ACTIONS(1546),
    [aux_sym_add_operator_token1] = ACTIONS(1546),
    [sym_concat_operator] = ACTIONS(1548),
    [sym_rel_operator] = ACTIONS(1546),
    [anon_sym_AMP_AMP] = ACTIONS(1546),
    [anon_sym_or] = ACTIONS(1546),
    [anon_sym_PIPE_PIPE] = ACTIONS(1546),
    [sym_assign_operator] = ACTIONS(1548),
    [sym__uppercase_identifier] = ACTIONS(1548),
    [aux_sym_tag_token1] = ACTIONS(1548),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(859)] = {
    [sym_attribute] = STATE(859),
    [ts_builtin_sym_end] = ACTIONS(2027),
    [sym__lowercase_identifier] = ACTIONS(2025),
    [anon_sym_SEMI_SEMI] = ACTIONS(2027),
    [anon_sym_and] = ACTIONS(2025),
    [anon_sym_TILDE] = ACTIONS(2025),
    [anon_sym_QMARK] = ACTIONS(2025),
    [anon_sym_LPAREN] = ACTIONS(2027),
    [anon_sym_RPAREN] = ACTIONS(2027),
    [anon_sym_external] = ACTIONS(2025),
    [anon_sym_type] = ACTIONS(2025),
    [anon_sym_COMMA] = ACTIONS(2027),
    [anon_sym_COLON2] = ACTIONS(2025),
    [anon_sym_DOT] = ACTIONS(2027),
    [anon_sym_LBRACK] = ACTIONS(2025),
    [anon_sym_RBRACK] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(2025),
    [anon_sym_false] = ACTIONS(2025),
    [anon_sym_LBRACE] = ACTIONS(2025),
    [anon_sym_constraint] = ACTIONS(2025),
    [anon_sym_exception] = ACTIONS(2025),
    [anon_sym_module] = ACTIONS(2025),
    [anon_sym_open] = ACTIONS(2025),
    [anon_sym_include] = ACTIONS(2025),
    [anon_sym_class] = ACTIONS(2025),
    [anon_sym_val] = ACTIONS(2025),
    [anon_sym_end] = ACTIONS(2025),
    [anon_sym_COLON_GT] = ACTIONS(2027),
    [anon_sym_object] = ACTIONS(2025),
    [anon_sym_inherit] = ACTIONS(2025),
    [anon_sym_method] = ACTIONS(2025),
    [anon_sym_in] = ACTIONS(2025),
    [anon_sym_as] = ACTIONS(2025),
    [anon_sym_initializer] = ACTIONS(2025),
    [anon_sym_AMP] = ACTIONS(2025),
    [anon_sym_POUND] = ACTIONS(2025),
    [anon_sym_COLON_COLON] = ACTIONS(2027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2027),
    [anon_sym_else] = ACTIONS(2025),
    [anon_sym_new] = ACTIONS(2025),
    [anon_sym_LBRACE_LT] = ACTIONS(2027),
    [anon_sym_begin] = ACTIONS(2025),
    [sym_ocamlyacc_value] = ACTIONS(2027),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2025),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2027),
    [aux_sym_number_token1] = ACTIONS(2027),
    [aux_sym_signed_number_token1] = ACTIONS(2025),
    [anon_sym_SQUOTE] = ACTIONS(2027),
    [anon_sym_DQUOTE] = ACTIONS(2027),
    [sym_prefix_operator] = ACTIONS(2027),
    [aux_sym_sign_operator_token1] = ACTIONS(2025),
    [sym_hash_operator] = ACTIONS(2027),
    [anon_sym_lsl] = ACTIONS(2025),
    [anon_sym_lsr] = ACTIONS(2025),
    [anon_sym_asr] = ACTIONS(2025),
    [aux_sym_pow_operator_token1] = ACTIONS(2027),
    [anon_sym_mod] = ACTIONS(2025),
    [anon_sym_land] = ACTIONS(2025),
    [anon_sym_lor] = ACTIONS(2025),
    [anon_sym_lxor] = ACTIONS(2025),
    [aux_sym_mult_operator_token1] = ACTIONS(2025),
    [aux_sym_add_operator_token1] = ACTIONS(2025),
    [sym_concat_operator] = ACTIONS(2027),
    [sym_rel_operator] = ACTIONS(2025),
    [anon_sym_AMP_AMP] = ACTIONS(2025),
    [anon_sym_or] = ACTIONS(2025),
    [anon_sym_PIPE_PIPE] = ACTIONS(2025),
    [sym_assign_operator] = ACTIONS(2027),
    [sym__uppercase_identifier] = ACTIONS(2027),
    [aux_sym_tag_token1] = ACTIONS(2027),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(860)] = {
    [sym_attribute] = STATE(860),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_RPAREN] = ACTIONS(2081),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_COLON2] = ACTIONS(2079),
    [anon_sym_DOT] = ACTIONS(745),
    [anon_sym_DASH_GT] = ACTIONS(2079),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2081),
    [anon_sym_RBRACE] = ACTIONS(2081),
    [anon_sym_constraint] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_with] = ACTIONS(2079),
    [anon_sym_COLON_GT] = ACTIONS(2081),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_inherit] = ACTIONS(2079),
    [anon_sym_method] = ACTIONS(2079),
    [anon_sym_initializer] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_POUND] = ACTIONS(747),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2081),
    [anon_sym_then] = ACTIONS(2079),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_do] = ACTIONS(2079),
    [anon_sym_done] = ACTIONS(2079),
    [anon_sym_to] = ACTIONS(2079),
    [anon_sym_downto] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_GT_RBRACE] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [sym_hash_operator] = ACTIONS(749),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(861)] = {
    [sym_attribute] = STATE(861),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_and] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_RPAREN] = ACTIONS(2081),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_COLON2] = ACTIONS(2079),
    [anon_sym_DASH_GT] = ACTIONS(2079),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2081),
    [anon_sym_RBRACE] = ACTIONS(2081),
    [anon_sym_constraint] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_with] = ACTIONS(2079),
    [anon_sym_COLON_GT] = ACTIONS(2081),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_inherit] = ACTIONS(2079),
    [anon_sym_method] = ACTIONS(2079),
    [anon_sym_in] = ACTIONS(2079),
    [anon_sym_as] = ACTIONS(2079),
    [anon_sym_initializer] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2081),
    [anon_sym_then] = ACTIONS(2079),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_do] = ACTIONS(2079),
    [anon_sym_done] = ACTIONS(2079),
    [anon_sym_to] = ACTIONS(2079),
    [anon_sym_downto] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_GT_RBRACE] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(862)] = {
    [sym_parameter] = STATE(2475),
    [sym__class_typed] = STATE(3634),
    [sym__parenthesized_abstract_type] = STATE(2475),
    [sym__delimited_pattern] = STATE(2450),
    [sym__simple_pattern] = STATE(2476),
    [sym_typed_pattern] = STATE(2450),
    [sym_polymorphic_variant_pattern] = STATE(2450),
    [sym_record_pattern] = STATE(2478),
    [sym_list_pattern] = STATE(2478),
    [sym_array_pattern] = STATE(2478),
    [sym_local_open_pattern] = STATE(2450),
    [sym_package_pattern] = STATE(2450),
    [sym_parenthesized_pattern] = STATE(2478),
    [sym_attribute] = STATE(862),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2450),
    [sym_quoted_extension] = STATE(2450),
    [sym__constant] = STATE(2464),
    [sym__signed_constant] = STATE(2450),
    [sym_number] = STATE(2504),
    [sym_signed_number] = STATE(2464),
    [sym_character] = STATE(2504),
    [sym_string] = STATE(2504),
    [sym_quoted_string] = STATE(2504),
    [sym_boolean] = STATE(2504),
    [sym_unit] = STATE(2478),
    [sym__simple_value_pattern] = STATE(2450),
    [sym_parenthesized_operator] = STATE(2450),
    [sym_module_path] = STATE(6504),
    [sym_constructor_path] = STATE(2450),
    [sym__constructor_name] = STATE(2515),
    [sym_tag] = STATE(2450),
    [aux_sym_expression_item_repeat1] = STATE(3569),
    [aux_sym_let_binding_repeat1] = STATE(1012),
    [sym__lowercase_identifier] = ACTIONS(2029),
    [anon_sym_SEMI_SEMI] = ACTIONS(2083),
    [anon_sym_let] = ACTIONS(2085),
    [anon_sym_and] = ACTIONS(2085),
    [anon_sym_EQ] = ACTIONS(2087),
    [anon_sym_TILDE] = ACTIONS(2037),
    [anon_sym_QMARK] = ACTIONS(2037),
    [anon_sym_LPAREN] = ACTIONS(2039),
    [anon_sym_external] = ACTIONS(2085),
    [anon_sym_type] = ACTIONS(2085),
    [anon_sym_COLON2] = ACTIONS(2041),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2083),
    [anon_sym_true] = ACTIONS(2045),
    [anon_sym_false] = ACTIONS(2045),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2085),
    [anon_sym_module] = ACTIONS(2085),
    [anon_sym_open] = ACTIONS(2085),
    [anon_sym_include] = ACTIONS(2085),
    [anon_sym_class] = ACTIONS(2085),
    [anon_sym_val] = ACTIONS(2085),
    [anon_sym_end] = ACTIONS(2085),
    [anon_sym_POUND] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2051),
    [anon_sym_begin] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2083),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2083),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2067),
    [sym_let_operator] = ACTIONS(2083),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2085),
    [aux_sym_tag_token1] = ACTIONS(2071),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(863)] = {
    [sym_attribute] = STATE(863),
    [sym__lowercase_identifier] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_RPAREN] = ACTIONS(1616),
    [anon_sym_COMMA] = ACTIONS(1616),
    [anon_sym_PIPE] = ACTIONS(1614),
    [anon_sym_COLON2] = ACTIONS(1614),
    [anon_sym_DOT] = ACTIONS(1616),
    [anon_sym_DASH_GT] = ACTIONS(1614),
    [anon_sym_LBRACK] = ACTIONS(1614),
    [anon_sym_RBRACK] = ACTIONS(1616),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1616),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [anon_sym_constraint] = ACTIONS(1614),
    [anon_sym_val] = ACTIONS(1614),
    [anon_sym_end] = ACTIONS(1614),
    [anon_sym_with] = ACTIONS(1614),
    [anon_sym_COLON_GT] = ACTIONS(1616),
    [anon_sym_object] = ACTIONS(1614),
    [anon_sym_inherit] = ACTIONS(1614),
    [anon_sym_method] = ACTIONS(1614),
    [anon_sym_initializer] = ACTIONS(1614),
    [anon_sym_AMP] = ACTIONS(1614),
    [anon_sym_POUND] = ACTIONS(1614),
    [anon_sym_COLON_COLON] = ACTIONS(1616),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1616),
    [anon_sym_LT_DASH] = ACTIONS(2009),
    [anon_sym_then] = ACTIONS(1614),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_do] = ACTIONS(1614),
    [anon_sym_done] = ACTIONS(1614),
    [anon_sym_to] = ACTIONS(1614),
    [anon_sym_downto] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
    [anon_sym_LBRACE_LT] = ACTIONS(1616),
    [anon_sym_GT_RBRACE] = ACTIONS(1616),
    [anon_sym_begin] = ACTIONS(1614),
    [sym_ocamlyacc_value] = ACTIONS(1616),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1614),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1616),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1616),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1616),
    [aux_sym_number_token1] = ACTIONS(1616),
    [aux_sym_signed_number_token1] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1616),
    [sym_prefix_operator] = ACTIONS(1616),
    [aux_sym_sign_operator_token1] = ACTIONS(1614),
    [sym_hash_operator] = ACTIONS(1616),
    [anon_sym_lsl] = ACTIONS(1614),
    [anon_sym_lsr] = ACTIONS(1614),
    [anon_sym_asr] = ACTIONS(1614),
    [aux_sym_pow_operator_token1] = ACTIONS(1616),
    [anon_sym_mod] = ACTIONS(1614),
    [anon_sym_land] = ACTIONS(1614),
    [anon_sym_lor] = ACTIONS(1614),
    [anon_sym_lxor] = ACTIONS(1614),
    [aux_sym_mult_operator_token1] = ACTIONS(1614),
    [aux_sym_add_operator_token1] = ACTIONS(1614),
    [sym_concat_operator] = ACTIONS(1616),
    [sym_rel_operator] = ACTIONS(1614),
    [anon_sym_AMP_AMP] = ACTIONS(1614),
    [anon_sym_or] = ACTIONS(1614),
    [anon_sym_PIPE_PIPE] = ACTIONS(1614),
    [sym_assign_operator] = ACTIONS(1616),
    [sym__uppercase_identifier] = ACTIONS(1616),
    [aux_sym_tag_token1] = ACTIONS(1616),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(864)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(3092),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3080),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6514),
    [sym__tuple_binding_pattern] = STATE(3085),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(864),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_and] = ACTIONS(1318),
    [anon_sym_EQ] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_QMARK] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_COLON2] = ACTIONS(1318),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_COLON_GT] = ACTIONS(1316),
    [anon_sym_in] = ACTIONS(1318),
    [anon_sym_as] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_COLON_COLON] = ACTIONS(1316),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1316),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym_let_and_operator] = ACTIONS(1316),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(865)] = {
    [sym_parameter] = STATE(2475),
    [sym__class_typed] = STATE(3638),
    [sym__parenthesized_abstract_type] = STATE(2475),
    [sym__delimited_pattern] = STATE(2450),
    [sym__simple_pattern] = STATE(2476),
    [sym_typed_pattern] = STATE(2450),
    [sym_polymorphic_variant_pattern] = STATE(2450),
    [sym_record_pattern] = STATE(2478),
    [sym_list_pattern] = STATE(2478),
    [sym_array_pattern] = STATE(2478),
    [sym_local_open_pattern] = STATE(2450),
    [sym_package_pattern] = STATE(2450),
    [sym_parenthesized_pattern] = STATE(2478),
    [sym_attribute] = STATE(865),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2450),
    [sym_quoted_extension] = STATE(2450),
    [sym__constant] = STATE(2464),
    [sym__signed_constant] = STATE(2450),
    [sym_number] = STATE(2504),
    [sym_signed_number] = STATE(2464),
    [sym_character] = STATE(2504),
    [sym_string] = STATE(2504),
    [sym_quoted_string] = STATE(2504),
    [sym_boolean] = STATE(2504),
    [sym_unit] = STATE(2478),
    [sym__simple_value_pattern] = STATE(2450),
    [sym_parenthesized_operator] = STATE(2450),
    [sym_module_path] = STATE(6504),
    [sym_constructor_path] = STATE(2450),
    [sym__constructor_name] = STATE(2515),
    [sym_tag] = STATE(2450),
    [aux_sym_expression_item_repeat1] = STATE(3610),
    [aux_sym_let_binding_repeat1] = STATE(868),
    [sym__lowercase_identifier] = ACTIONS(2029),
    [anon_sym_SEMI_SEMI] = ACTIONS(2089),
    [anon_sym_let] = ACTIONS(2091),
    [anon_sym_and] = ACTIONS(2091),
    [anon_sym_EQ] = ACTIONS(2093),
    [anon_sym_TILDE] = ACTIONS(2037),
    [anon_sym_QMARK] = ACTIONS(2037),
    [anon_sym_LPAREN] = ACTIONS(2039),
    [anon_sym_external] = ACTIONS(2091),
    [anon_sym_type] = ACTIONS(2091),
    [anon_sym_COLON2] = ACTIONS(2041),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2089),
    [anon_sym_true] = ACTIONS(2045),
    [anon_sym_false] = ACTIONS(2045),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2091),
    [anon_sym_module] = ACTIONS(2091),
    [anon_sym_open] = ACTIONS(2091),
    [anon_sym_include] = ACTIONS(2091),
    [anon_sym_class] = ACTIONS(2091),
    [anon_sym_val] = ACTIONS(2091),
    [anon_sym_end] = ACTIONS(2091),
    [anon_sym_POUND] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2051),
    [anon_sym_begin] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2089),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2089),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2089),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2067),
    [sym_let_operator] = ACTIONS(2089),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2091),
    [aux_sym_tag_token1] = ACTIONS(2071),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(866)] = {
    [sym_attribute] = STATE(866),
    [sym__lowercase_identifier] = ACTIONS(1574),
    [anon_sym_TILDE] = ACTIONS(1574),
    [anon_sym_QMARK] = ACTIONS(1574),
    [anon_sym_LPAREN] = ACTIONS(1576),
    [anon_sym_RPAREN] = ACTIONS(1576),
    [anon_sym_COMMA] = ACTIONS(1576),
    [anon_sym_PIPE] = ACTIONS(1574),
    [anon_sym_COLON2] = ACTIONS(1574),
    [anon_sym_DOT] = ACTIONS(1576),
    [anon_sym_DASH_GT] = ACTIONS(1574),
    [anon_sym_LBRACK] = ACTIONS(1574),
    [anon_sym_RBRACK] = ACTIONS(1576),
    [anon_sym_true] = ACTIONS(1574),
    [anon_sym_false] = ACTIONS(1574),
    [anon_sym_LBRACE] = ACTIONS(1574),
    [anon_sym_SEMI] = ACTIONS(1576),
    [anon_sym_RBRACE] = ACTIONS(1576),
    [anon_sym_constraint] = ACTIONS(1574),
    [anon_sym_val] = ACTIONS(1574),
    [anon_sym_end] = ACTIONS(1574),
    [anon_sym_with] = ACTIONS(1574),
    [anon_sym_COLON_GT] = ACTIONS(1576),
    [anon_sym_object] = ACTIONS(1574),
    [anon_sym_inherit] = ACTIONS(1574),
    [anon_sym_method] = ACTIONS(1574),
    [anon_sym_initializer] = ACTIONS(1574),
    [anon_sym_AMP] = ACTIONS(1574),
    [anon_sym_POUND] = ACTIONS(1574),
    [anon_sym_COLON_COLON] = ACTIONS(1576),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1576),
    [anon_sym_LT_DASH] = ACTIONS(1574),
    [anon_sym_then] = ACTIONS(1574),
    [anon_sym_else] = ACTIONS(1574),
    [anon_sym_do] = ACTIONS(1574),
    [anon_sym_done] = ACTIONS(1574),
    [anon_sym_to] = ACTIONS(1574),
    [anon_sym_downto] = ACTIONS(1574),
    [anon_sym_new] = ACTIONS(1574),
    [anon_sym_LBRACE_LT] = ACTIONS(1576),
    [anon_sym_GT_RBRACE] = ACTIONS(1576),
    [anon_sym_begin] = ACTIONS(1574),
    [sym_ocamlyacc_value] = ACTIONS(1576),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1574),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1576),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1574),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1574),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1576),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1576),
    [aux_sym_number_token1] = ACTIONS(1576),
    [aux_sym_signed_number_token1] = ACTIONS(1574),
    [anon_sym_SQUOTE] = ACTIONS(1576),
    [anon_sym_DQUOTE] = ACTIONS(1576),
    [sym_prefix_operator] = ACTIONS(1576),
    [aux_sym_sign_operator_token1] = ACTIONS(1574),
    [sym_hash_operator] = ACTIONS(1576),
    [anon_sym_lsl] = ACTIONS(1574),
    [anon_sym_lsr] = ACTIONS(1574),
    [anon_sym_asr] = ACTIONS(1574),
    [aux_sym_pow_operator_token1] = ACTIONS(1576),
    [anon_sym_mod] = ACTIONS(1574),
    [anon_sym_land] = ACTIONS(1574),
    [anon_sym_lor] = ACTIONS(1574),
    [anon_sym_lxor] = ACTIONS(1574),
    [aux_sym_mult_operator_token1] = ACTIONS(1574),
    [aux_sym_add_operator_token1] = ACTIONS(1574),
    [sym_concat_operator] = ACTIONS(1576),
    [sym_rel_operator] = ACTIONS(1574),
    [anon_sym_AMP_AMP] = ACTIONS(1574),
    [anon_sym_or] = ACTIONS(1574),
    [anon_sym_PIPE_PIPE] = ACTIONS(1574),
    [sym_assign_operator] = ACTIONS(1576),
    [sym__uppercase_identifier] = ACTIONS(1576),
    [aux_sym_tag_token1] = ACTIONS(1576),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(867)] = {
    [sym_attribute] = STATE(867),
    [ts_builtin_sym_end] = ACTIONS(2023),
    [sym__lowercase_identifier] = ACTIONS(2021),
    [anon_sym_SEMI_SEMI] = ACTIONS(2023),
    [anon_sym_and] = ACTIONS(2021),
    [anon_sym_TILDE] = ACTIONS(2021),
    [anon_sym_QMARK] = ACTIONS(2021),
    [anon_sym_LPAREN] = ACTIONS(2023),
    [anon_sym_RPAREN] = ACTIONS(2023),
    [anon_sym_external] = ACTIONS(2021),
    [anon_sym_type] = ACTIONS(2021),
    [anon_sym_COMMA] = ACTIONS(2023),
    [anon_sym_COLON2] = ACTIONS(2021),
    [anon_sym_DOT] = ACTIONS(2023),
    [anon_sym_LBRACK] = ACTIONS(2021),
    [anon_sym_RBRACK] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(2021),
    [anon_sym_false] = ACTIONS(2021),
    [anon_sym_LBRACE] = ACTIONS(2021),
    [anon_sym_constraint] = ACTIONS(2021),
    [anon_sym_exception] = ACTIONS(2021),
    [anon_sym_module] = ACTIONS(2021),
    [anon_sym_open] = ACTIONS(2021),
    [anon_sym_include] = ACTIONS(2021),
    [anon_sym_class] = ACTIONS(2021),
    [anon_sym_val] = ACTIONS(2021),
    [anon_sym_end] = ACTIONS(2021),
    [anon_sym_COLON_GT] = ACTIONS(2023),
    [anon_sym_object] = ACTIONS(2021),
    [anon_sym_inherit] = ACTIONS(2021),
    [anon_sym_method] = ACTIONS(2021),
    [anon_sym_in] = ACTIONS(2021),
    [anon_sym_as] = ACTIONS(2021),
    [anon_sym_initializer] = ACTIONS(2021),
    [anon_sym_AMP] = ACTIONS(2021),
    [anon_sym_POUND] = ACTIONS(2021),
    [anon_sym_COLON_COLON] = ACTIONS(2023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2023),
    [anon_sym_else] = ACTIONS(2021),
    [anon_sym_new] = ACTIONS(2021),
    [anon_sym_LBRACE_LT] = ACTIONS(2023),
    [anon_sym_begin] = ACTIONS(2021),
    [sym_ocamlyacc_value] = ACTIONS(2023),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2023),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2023),
    [aux_sym_number_token1] = ACTIONS(2023),
    [aux_sym_signed_number_token1] = ACTIONS(2021),
    [anon_sym_SQUOTE] = ACTIONS(2023),
    [anon_sym_DQUOTE] = ACTIONS(2023),
    [sym_prefix_operator] = ACTIONS(2023),
    [aux_sym_sign_operator_token1] = ACTIONS(2021),
    [sym_hash_operator] = ACTIONS(2023),
    [anon_sym_lsl] = ACTIONS(2021),
    [anon_sym_lsr] = ACTIONS(2021),
    [anon_sym_asr] = ACTIONS(2021),
    [aux_sym_pow_operator_token1] = ACTIONS(2023),
    [anon_sym_mod] = ACTIONS(2021),
    [anon_sym_land] = ACTIONS(2021),
    [anon_sym_lor] = ACTIONS(2021),
    [anon_sym_lxor] = ACTIONS(2021),
    [aux_sym_mult_operator_token1] = ACTIONS(2021),
    [aux_sym_add_operator_token1] = ACTIONS(2021),
    [sym_concat_operator] = ACTIONS(2023),
    [sym_rel_operator] = ACTIONS(2021),
    [anon_sym_AMP_AMP] = ACTIONS(2021),
    [anon_sym_or] = ACTIONS(2021),
    [anon_sym_PIPE_PIPE] = ACTIONS(2021),
    [sym_assign_operator] = ACTIONS(2023),
    [sym__uppercase_identifier] = ACTIONS(2023),
    [aux_sym_tag_token1] = ACTIONS(2023),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(868)] = {
    [sym_parameter] = STATE(2475),
    [sym__class_typed] = STATE(3639),
    [sym__parenthesized_abstract_type] = STATE(2475),
    [sym__delimited_pattern] = STATE(2450),
    [sym__simple_pattern] = STATE(2476),
    [sym_typed_pattern] = STATE(2450),
    [sym_polymorphic_variant_pattern] = STATE(2450),
    [sym_record_pattern] = STATE(2478),
    [sym_list_pattern] = STATE(2478),
    [sym_array_pattern] = STATE(2478),
    [sym_local_open_pattern] = STATE(2450),
    [sym_package_pattern] = STATE(2450),
    [sym_parenthesized_pattern] = STATE(2478),
    [sym_attribute] = STATE(868),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2450),
    [sym_quoted_extension] = STATE(2450),
    [sym__constant] = STATE(2464),
    [sym__signed_constant] = STATE(2450),
    [sym_number] = STATE(2504),
    [sym_signed_number] = STATE(2464),
    [sym_character] = STATE(2504),
    [sym_string] = STATE(2504),
    [sym_quoted_string] = STATE(2504),
    [sym_boolean] = STATE(2504),
    [sym_unit] = STATE(2478),
    [sym__simple_value_pattern] = STATE(2450),
    [sym_parenthesized_operator] = STATE(2450),
    [sym_module_path] = STATE(6504),
    [sym_constructor_path] = STATE(2450),
    [sym__constructor_name] = STATE(2515),
    [sym_tag] = STATE(2450),
    [aux_sym_expression_item_repeat1] = STATE(3662),
    [aux_sym_let_binding_repeat1] = STATE(1012),
    [sym__lowercase_identifier] = ACTIONS(2029),
    [anon_sym_SEMI_SEMI] = ACTIONS(2095),
    [anon_sym_let] = ACTIONS(2097),
    [anon_sym_and] = ACTIONS(2097),
    [anon_sym_EQ] = ACTIONS(2099),
    [anon_sym_TILDE] = ACTIONS(2037),
    [anon_sym_QMARK] = ACTIONS(2037),
    [anon_sym_LPAREN] = ACTIONS(2039),
    [anon_sym_external] = ACTIONS(2097),
    [anon_sym_type] = ACTIONS(2097),
    [anon_sym_COLON2] = ACTIONS(2041),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2095),
    [anon_sym_true] = ACTIONS(2045),
    [anon_sym_false] = ACTIONS(2045),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2097),
    [anon_sym_module] = ACTIONS(2097),
    [anon_sym_open] = ACTIONS(2097),
    [anon_sym_include] = ACTIONS(2097),
    [anon_sym_class] = ACTIONS(2097),
    [anon_sym_val] = ACTIONS(2097),
    [anon_sym_end] = ACTIONS(2097),
    [anon_sym_POUND] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2051),
    [anon_sym_begin] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2095),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2095),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2067),
    [sym_let_operator] = ACTIONS(2095),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2071),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(869)] = {
    [sym_parameter] = STATE(2475),
    [sym__class_typed] = STATE(3640),
    [sym__parenthesized_abstract_type] = STATE(2475),
    [sym__delimited_pattern] = STATE(2450),
    [sym__simple_pattern] = STATE(2476),
    [sym_typed_pattern] = STATE(2450),
    [sym_polymorphic_variant_pattern] = STATE(2450),
    [sym_record_pattern] = STATE(2478),
    [sym_list_pattern] = STATE(2478),
    [sym_array_pattern] = STATE(2478),
    [sym_local_open_pattern] = STATE(2450),
    [sym_package_pattern] = STATE(2450),
    [sym_parenthesized_pattern] = STATE(2478),
    [sym_attribute] = STATE(869),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2450),
    [sym_quoted_extension] = STATE(2450),
    [sym__constant] = STATE(2464),
    [sym__signed_constant] = STATE(2450),
    [sym_number] = STATE(2504),
    [sym_signed_number] = STATE(2464),
    [sym_character] = STATE(2504),
    [sym_string] = STATE(2504),
    [sym_quoted_string] = STATE(2504),
    [sym_boolean] = STATE(2504),
    [sym_unit] = STATE(2478),
    [sym__simple_value_pattern] = STATE(2450),
    [sym_parenthesized_operator] = STATE(2450),
    [sym_module_path] = STATE(6504),
    [sym_constructor_path] = STATE(2450),
    [sym__constructor_name] = STATE(2515),
    [sym_tag] = STATE(2450),
    [aux_sym_expression_item_repeat1] = STATE(3663),
    [aux_sym_let_binding_repeat1] = STATE(872),
    [sym__lowercase_identifier] = ACTIONS(2029),
    [anon_sym_SEMI_SEMI] = ACTIONS(2101),
    [anon_sym_let] = ACTIONS(2103),
    [anon_sym_and] = ACTIONS(2103),
    [anon_sym_EQ] = ACTIONS(2105),
    [anon_sym_TILDE] = ACTIONS(2037),
    [anon_sym_QMARK] = ACTIONS(2037),
    [anon_sym_LPAREN] = ACTIONS(2039),
    [anon_sym_external] = ACTIONS(2103),
    [anon_sym_type] = ACTIONS(2103),
    [anon_sym_COLON2] = ACTIONS(2041),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2101),
    [anon_sym_true] = ACTIONS(2045),
    [anon_sym_false] = ACTIONS(2045),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2103),
    [anon_sym_module] = ACTIONS(2103),
    [anon_sym_open] = ACTIONS(2103),
    [anon_sym_include] = ACTIONS(2103),
    [anon_sym_class] = ACTIONS(2103),
    [anon_sym_val] = ACTIONS(2103),
    [anon_sym_end] = ACTIONS(2103),
    [anon_sym_POUND] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2051),
    [anon_sym_begin] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2101),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2101),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2101),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2067),
    [sym_let_operator] = ACTIONS(2101),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2103),
    [aux_sym_tag_token1] = ACTIONS(2071),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(870)] = {
    [sym_attribute] = STATE(870),
    [sym__lowercase_identifier] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2107),
    [anon_sym_QMARK] = ACTIONS(2107),
    [anon_sym_LPAREN] = ACTIONS(2109),
    [anon_sym_RPAREN] = ACTIONS(2109),
    [anon_sym_COMMA] = ACTIONS(2109),
    [anon_sym_PIPE] = ACTIONS(2107),
    [anon_sym_COLON2] = ACTIONS(2107),
    [anon_sym_DOT] = ACTIONS(745),
    [anon_sym_DASH_GT] = ACTIONS(2107),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_RBRACK] = ACTIONS(2109),
    [anon_sym_true] = ACTIONS(2107),
    [anon_sym_false] = ACTIONS(2107),
    [anon_sym_LBRACE] = ACTIONS(2107),
    [anon_sym_SEMI] = ACTIONS(2109),
    [anon_sym_RBRACE] = ACTIONS(2109),
    [anon_sym_constraint] = ACTIONS(2107),
    [anon_sym_val] = ACTIONS(2107),
    [anon_sym_end] = ACTIONS(2107),
    [anon_sym_with] = ACTIONS(2107),
    [anon_sym_COLON_GT] = ACTIONS(2109),
    [anon_sym_object] = ACTIONS(2107),
    [anon_sym_inherit] = ACTIONS(2107),
    [anon_sym_method] = ACTIONS(2107),
    [anon_sym_initializer] = ACTIONS(2107),
    [anon_sym_AMP] = ACTIONS(2107),
    [anon_sym_POUND] = ACTIONS(747),
    [anon_sym_COLON_COLON] = ACTIONS(2109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2109),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2109),
    [anon_sym_then] = ACTIONS(2107),
    [anon_sym_else] = ACTIONS(2107),
    [anon_sym_do] = ACTIONS(2107),
    [anon_sym_done] = ACTIONS(2107),
    [anon_sym_to] = ACTIONS(2107),
    [anon_sym_downto] = ACTIONS(2107),
    [anon_sym_new] = ACTIONS(2107),
    [anon_sym_LBRACE_LT] = ACTIONS(2109),
    [anon_sym_GT_RBRACE] = ACTIONS(2109),
    [anon_sym_begin] = ACTIONS(2107),
    [sym_ocamlyacc_value] = ACTIONS(2109),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2107),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2109),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2109),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2109),
    [aux_sym_number_token1] = ACTIONS(2109),
    [aux_sym_signed_number_token1] = ACTIONS(2107),
    [anon_sym_SQUOTE] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(2109),
    [sym_prefix_operator] = ACTIONS(2109),
    [aux_sym_sign_operator_token1] = ACTIONS(2107),
    [sym_hash_operator] = ACTIONS(749),
    [anon_sym_lsl] = ACTIONS(2107),
    [anon_sym_lsr] = ACTIONS(2107),
    [anon_sym_asr] = ACTIONS(2107),
    [aux_sym_pow_operator_token1] = ACTIONS(2109),
    [anon_sym_mod] = ACTIONS(2107),
    [anon_sym_land] = ACTIONS(2107),
    [anon_sym_lor] = ACTIONS(2107),
    [anon_sym_lxor] = ACTIONS(2107),
    [aux_sym_mult_operator_token1] = ACTIONS(2107),
    [aux_sym_add_operator_token1] = ACTIONS(2107),
    [sym_concat_operator] = ACTIONS(2109),
    [sym_rel_operator] = ACTIONS(2107),
    [anon_sym_AMP_AMP] = ACTIONS(2107),
    [anon_sym_or] = ACTIONS(2107),
    [anon_sym_PIPE_PIPE] = ACTIONS(2107),
    [sym_assign_operator] = ACTIONS(2109),
    [sym__uppercase_identifier] = ACTIONS(2109),
    [aux_sym_tag_token1] = ACTIONS(2109),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(871)] = {
    [sym_attribute] = STATE(871),
    [sym__lowercase_identifier] = ACTIONS(1554),
    [anon_sym_TILDE] = ACTIONS(1554),
    [anon_sym_QMARK] = ACTIONS(1554),
    [anon_sym_LPAREN] = ACTIONS(1556),
    [anon_sym_RPAREN] = ACTIONS(1556),
    [anon_sym_COMMA] = ACTIONS(1556),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_COLON2] = ACTIONS(1554),
    [anon_sym_DOT] = ACTIONS(1556),
    [anon_sym_DASH_GT] = ACTIONS(1554),
    [anon_sym_LBRACK] = ACTIONS(1554),
    [anon_sym_RBRACK] = ACTIONS(1556),
    [anon_sym_true] = ACTIONS(1554),
    [anon_sym_false] = ACTIONS(1554),
    [anon_sym_LBRACE] = ACTIONS(1554),
    [anon_sym_SEMI] = ACTIONS(1556),
    [anon_sym_RBRACE] = ACTIONS(1556),
    [anon_sym_constraint] = ACTIONS(1554),
    [anon_sym_val] = ACTIONS(1554),
    [anon_sym_end] = ACTIONS(1554),
    [anon_sym_with] = ACTIONS(1554),
    [anon_sym_COLON_GT] = ACTIONS(1556),
    [anon_sym_object] = ACTIONS(1554),
    [anon_sym_inherit] = ACTIONS(1554),
    [anon_sym_method] = ACTIONS(1554),
    [anon_sym_initializer] = ACTIONS(1554),
    [anon_sym_AMP] = ACTIONS(1554),
    [anon_sym_POUND] = ACTIONS(1554),
    [anon_sym_COLON_COLON] = ACTIONS(1556),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1556),
    [anon_sym_LT_DASH] = ACTIONS(1554),
    [anon_sym_then] = ACTIONS(1554),
    [anon_sym_else] = ACTIONS(1554),
    [anon_sym_do] = ACTIONS(1554),
    [anon_sym_done] = ACTIONS(1554),
    [anon_sym_to] = ACTIONS(1554),
    [anon_sym_downto] = ACTIONS(1554),
    [anon_sym_new] = ACTIONS(1554),
    [anon_sym_LBRACE_LT] = ACTIONS(1556),
    [anon_sym_GT_RBRACE] = ACTIONS(1556),
    [anon_sym_begin] = ACTIONS(1554),
    [sym_ocamlyacc_value] = ACTIONS(1556),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1554),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1556),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1554),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1554),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1556),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1556),
    [aux_sym_number_token1] = ACTIONS(1556),
    [aux_sym_signed_number_token1] = ACTIONS(1554),
    [anon_sym_SQUOTE] = ACTIONS(1556),
    [anon_sym_DQUOTE] = ACTIONS(1556),
    [sym_prefix_operator] = ACTIONS(1556),
    [aux_sym_sign_operator_token1] = ACTIONS(1554),
    [sym_hash_operator] = ACTIONS(1556),
    [anon_sym_lsl] = ACTIONS(1554),
    [anon_sym_lsr] = ACTIONS(1554),
    [anon_sym_asr] = ACTIONS(1554),
    [aux_sym_pow_operator_token1] = ACTIONS(1556),
    [anon_sym_mod] = ACTIONS(1554),
    [anon_sym_land] = ACTIONS(1554),
    [anon_sym_lor] = ACTIONS(1554),
    [anon_sym_lxor] = ACTIONS(1554),
    [aux_sym_mult_operator_token1] = ACTIONS(1554),
    [aux_sym_add_operator_token1] = ACTIONS(1554),
    [sym_concat_operator] = ACTIONS(1556),
    [sym_rel_operator] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1554),
    [anon_sym_or] = ACTIONS(1554),
    [anon_sym_PIPE_PIPE] = ACTIONS(1554),
    [sym_assign_operator] = ACTIONS(1556),
    [sym__uppercase_identifier] = ACTIONS(1556),
    [aux_sym_tag_token1] = ACTIONS(1556),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(872)] = {
    [sym_parameter] = STATE(2475),
    [sym__class_typed] = STATE(3641),
    [sym__parenthesized_abstract_type] = STATE(2475),
    [sym__delimited_pattern] = STATE(2450),
    [sym__simple_pattern] = STATE(2476),
    [sym_typed_pattern] = STATE(2450),
    [sym_polymorphic_variant_pattern] = STATE(2450),
    [sym_record_pattern] = STATE(2478),
    [sym_list_pattern] = STATE(2478),
    [sym_array_pattern] = STATE(2478),
    [sym_local_open_pattern] = STATE(2450),
    [sym_package_pattern] = STATE(2450),
    [sym_parenthesized_pattern] = STATE(2478),
    [sym_attribute] = STATE(872),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2450),
    [sym_quoted_extension] = STATE(2450),
    [sym__constant] = STATE(2464),
    [sym__signed_constant] = STATE(2450),
    [sym_number] = STATE(2504),
    [sym_signed_number] = STATE(2464),
    [sym_character] = STATE(2504),
    [sym_string] = STATE(2504),
    [sym_quoted_string] = STATE(2504),
    [sym_boolean] = STATE(2504),
    [sym_unit] = STATE(2478),
    [sym__simple_value_pattern] = STATE(2450),
    [sym_parenthesized_operator] = STATE(2450),
    [sym_module_path] = STATE(6504),
    [sym_constructor_path] = STATE(2450),
    [sym__constructor_name] = STATE(2515),
    [sym_tag] = STATE(2450),
    [aux_sym_expression_item_repeat1] = STATE(3688),
    [aux_sym_let_binding_repeat1] = STATE(1012),
    [sym__lowercase_identifier] = ACTIONS(2029),
    [anon_sym_SEMI_SEMI] = ACTIONS(2111),
    [anon_sym_let] = ACTIONS(2113),
    [anon_sym_and] = ACTIONS(2113),
    [anon_sym_EQ] = ACTIONS(2115),
    [anon_sym_TILDE] = ACTIONS(2037),
    [anon_sym_QMARK] = ACTIONS(2037),
    [anon_sym_LPAREN] = ACTIONS(2039),
    [anon_sym_external] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2113),
    [anon_sym_COLON2] = ACTIONS(2041),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2111),
    [anon_sym_true] = ACTIONS(2045),
    [anon_sym_false] = ACTIONS(2045),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2113),
    [anon_sym_module] = ACTIONS(2113),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_include] = ACTIONS(2113),
    [anon_sym_class] = ACTIONS(2113),
    [anon_sym_val] = ACTIONS(2113),
    [anon_sym_end] = ACTIONS(2113),
    [anon_sym_POUND] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2051),
    [anon_sym_begin] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2111),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2067),
    [sym_let_operator] = ACTIONS(2111),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2113),
    [aux_sym_tag_token1] = ACTIONS(2071),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(873)] = {
    [sym_parameter] = STATE(2475),
    [sym__class_typed] = STATE(3642),
    [sym__parenthesized_abstract_type] = STATE(2475),
    [sym__delimited_pattern] = STATE(2450),
    [sym__simple_pattern] = STATE(2476),
    [sym_typed_pattern] = STATE(2450),
    [sym_polymorphic_variant_pattern] = STATE(2450),
    [sym_record_pattern] = STATE(2478),
    [sym_list_pattern] = STATE(2478),
    [sym_array_pattern] = STATE(2478),
    [sym_local_open_pattern] = STATE(2450),
    [sym_package_pattern] = STATE(2450),
    [sym_parenthesized_pattern] = STATE(2478),
    [sym_attribute] = STATE(873),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2450),
    [sym_quoted_extension] = STATE(2450),
    [sym__constant] = STATE(2464),
    [sym__signed_constant] = STATE(2450),
    [sym_number] = STATE(2504),
    [sym_signed_number] = STATE(2464),
    [sym_character] = STATE(2504),
    [sym_string] = STATE(2504),
    [sym_quoted_string] = STATE(2504),
    [sym_boolean] = STATE(2504),
    [sym_unit] = STATE(2478),
    [sym__simple_value_pattern] = STATE(2450),
    [sym_parenthesized_operator] = STATE(2450),
    [sym_module_path] = STATE(6504),
    [sym_constructor_path] = STATE(2450),
    [sym__constructor_name] = STATE(2515),
    [sym_tag] = STATE(2450),
    [aux_sym_expression_item_repeat1] = STATE(3689),
    [aux_sym_let_binding_repeat1] = STATE(881),
    [sym__lowercase_identifier] = ACTIONS(2029),
    [anon_sym_SEMI_SEMI] = ACTIONS(2117),
    [anon_sym_let] = ACTIONS(2119),
    [anon_sym_and] = ACTIONS(2119),
    [anon_sym_EQ] = ACTIONS(2121),
    [anon_sym_TILDE] = ACTIONS(2037),
    [anon_sym_QMARK] = ACTIONS(2037),
    [anon_sym_LPAREN] = ACTIONS(2039),
    [anon_sym_external] = ACTIONS(2119),
    [anon_sym_type] = ACTIONS(2119),
    [anon_sym_COLON2] = ACTIONS(2041),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2117),
    [anon_sym_true] = ACTIONS(2045),
    [anon_sym_false] = ACTIONS(2045),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2119),
    [anon_sym_module] = ACTIONS(2119),
    [anon_sym_open] = ACTIONS(2119),
    [anon_sym_include] = ACTIONS(2119),
    [anon_sym_class] = ACTIONS(2119),
    [anon_sym_val] = ACTIONS(2119),
    [anon_sym_end] = ACTIONS(2119),
    [anon_sym_POUND] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2051),
    [anon_sym_begin] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2117),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2117),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2067),
    [sym_let_operator] = ACTIONS(2117),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2119),
    [aux_sym_tag_token1] = ACTIONS(2071),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(874)] = {
    [sym_attribute] = STATE(874),
    [sym__lowercase_identifier] = ACTIONS(1550),
    [anon_sym_TILDE] = ACTIONS(1550),
    [anon_sym_QMARK] = ACTIONS(1550),
    [anon_sym_LPAREN] = ACTIONS(1552),
    [anon_sym_RPAREN] = ACTIONS(1552),
    [anon_sym_COMMA] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1550),
    [anon_sym_COLON2] = ACTIONS(1550),
    [anon_sym_DOT] = ACTIONS(1552),
    [anon_sym_DASH_GT] = ACTIONS(1550),
    [anon_sym_LBRACK] = ACTIONS(1550),
    [anon_sym_RBRACK] = ACTIONS(1552),
    [anon_sym_true] = ACTIONS(1550),
    [anon_sym_false] = ACTIONS(1550),
    [anon_sym_LBRACE] = ACTIONS(1550),
    [anon_sym_SEMI] = ACTIONS(1552),
    [anon_sym_RBRACE] = ACTIONS(1552),
    [anon_sym_constraint] = ACTIONS(1550),
    [anon_sym_val] = ACTIONS(1550),
    [anon_sym_end] = ACTIONS(1550),
    [anon_sym_with] = ACTIONS(1550),
    [anon_sym_COLON_GT] = ACTIONS(1552),
    [anon_sym_object] = ACTIONS(1550),
    [anon_sym_inherit] = ACTIONS(1550),
    [anon_sym_method] = ACTIONS(1550),
    [anon_sym_initializer] = ACTIONS(1550),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_POUND] = ACTIONS(1550),
    [anon_sym_COLON_COLON] = ACTIONS(1552),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1552),
    [anon_sym_LT_DASH] = ACTIONS(1550),
    [anon_sym_then] = ACTIONS(1550),
    [anon_sym_else] = ACTIONS(1550),
    [anon_sym_do] = ACTIONS(1550),
    [anon_sym_done] = ACTIONS(1550),
    [anon_sym_to] = ACTIONS(1550),
    [anon_sym_downto] = ACTIONS(1550),
    [anon_sym_new] = ACTIONS(1550),
    [anon_sym_LBRACE_LT] = ACTIONS(1552),
    [anon_sym_GT_RBRACE] = ACTIONS(1552),
    [anon_sym_begin] = ACTIONS(1550),
    [sym_ocamlyacc_value] = ACTIONS(1552),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1550),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1552),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1550),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1550),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1552),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1552),
    [aux_sym_number_token1] = ACTIONS(1552),
    [aux_sym_signed_number_token1] = ACTIONS(1550),
    [anon_sym_SQUOTE] = ACTIONS(1552),
    [anon_sym_DQUOTE] = ACTIONS(1552),
    [sym_prefix_operator] = ACTIONS(1552),
    [aux_sym_sign_operator_token1] = ACTIONS(1550),
    [sym_hash_operator] = ACTIONS(1552),
    [anon_sym_lsl] = ACTIONS(1550),
    [anon_sym_lsr] = ACTIONS(1550),
    [anon_sym_asr] = ACTIONS(1550),
    [aux_sym_pow_operator_token1] = ACTIONS(1552),
    [anon_sym_mod] = ACTIONS(1550),
    [anon_sym_land] = ACTIONS(1550),
    [anon_sym_lor] = ACTIONS(1550),
    [anon_sym_lxor] = ACTIONS(1550),
    [aux_sym_mult_operator_token1] = ACTIONS(1550),
    [aux_sym_add_operator_token1] = ACTIONS(1550),
    [sym_concat_operator] = ACTIONS(1552),
    [sym_rel_operator] = ACTIONS(1550),
    [anon_sym_AMP_AMP] = ACTIONS(1550),
    [anon_sym_or] = ACTIONS(1550),
    [anon_sym_PIPE_PIPE] = ACTIONS(1550),
    [sym_assign_operator] = ACTIONS(1552),
    [sym__uppercase_identifier] = ACTIONS(1552),
    [aux_sym_tag_token1] = ACTIONS(1552),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(875)] = {
    [sym_attribute] = STATE(875),
    [sym__lowercase_identifier] = ACTIONS(1570),
    [anon_sym_TILDE] = ACTIONS(1570),
    [anon_sym_QMARK] = ACTIONS(1570),
    [anon_sym_LPAREN] = ACTIONS(1572),
    [anon_sym_RPAREN] = ACTIONS(1572),
    [anon_sym_COMMA] = ACTIONS(1572),
    [anon_sym_PIPE] = ACTIONS(1570),
    [anon_sym_COLON2] = ACTIONS(1570),
    [anon_sym_DOT] = ACTIONS(1572),
    [anon_sym_DASH_GT] = ACTIONS(1570),
    [anon_sym_LBRACK] = ACTIONS(1570),
    [anon_sym_RBRACK] = ACTIONS(1572),
    [anon_sym_true] = ACTIONS(1570),
    [anon_sym_false] = ACTIONS(1570),
    [anon_sym_LBRACE] = ACTIONS(1570),
    [anon_sym_SEMI] = ACTIONS(1572),
    [anon_sym_RBRACE] = ACTIONS(1572),
    [anon_sym_constraint] = ACTIONS(1570),
    [anon_sym_val] = ACTIONS(1570),
    [anon_sym_end] = ACTIONS(1570),
    [anon_sym_with] = ACTIONS(1570),
    [anon_sym_COLON_GT] = ACTIONS(1572),
    [anon_sym_object] = ACTIONS(1570),
    [anon_sym_inherit] = ACTIONS(1570),
    [anon_sym_method] = ACTIONS(1570),
    [anon_sym_initializer] = ACTIONS(1570),
    [anon_sym_AMP] = ACTIONS(1570),
    [anon_sym_POUND] = ACTIONS(1570),
    [anon_sym_COLON_COLON] = ACTIONS(1572),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1572),
    [anon_sym_LT_DASH] = ACTIONS(1570),
    [anon_sym_then] = ACTIONS(1570),
    [anon_sym_else] = ACTIONS(1570),
    [anon_sym_do] = ACTIONS(1570),
    [anon_sym_done] = ACTIONS(1570),
    [anon_sym_to] = ACTIONS(1570),
    [anon_sym_downto] = ACTIONS(1570),
    [anon_sym_new] = ACTIONS(1570),
    [anon_sym_LBRACE_LT] = ACTIONS(1572),
    [anon_sym_GT_RBRACE] = ACTIONS(1572),
    [anon_sym_begin] = ACTIONS(1570),
    [sym_ocamlyacc_value] = ACTIONS(1572),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1570),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1572),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1570),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1570),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1572),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1572),
    [aux_sym_number_token1] = ACTIONS(1572),
    [aux_sym_signed_number_token1] = ACTIONS(1570),
    [anon_sym_SQUOTE] = ACTIONS(1572),
    [anon_sym_DQUOTE] = ACTIONS(1572),
    [sym_prefix_operator] = ACTIONS(1572),
    [aux_sym_sign_operator_token1] = ACTIONS(1570),
    [sym_hash_operator] = ACTIONS(1572),
    [anon_sym_lsl] = ACTIONS(1570),
    [anon_sym_lsr] = ACTIONS(1570),
    [anon_sym_asr] = ACTIONS(1570),
    [aux_sym_pow_operator_token1] = ACTIONS(1572),
    [anon_sym_mod] = ACTIONS(1570),
    [anon_sym_land] = ACTIONS(1570),
    [anon_sym_lor] = ACTIONS(1570),
    [anon_sym_lxor] = ACTIONS(1570),
    [aux_sym_mult_operator_token1] = ACTIONS(1570),
    [aux_sym_add_operator_token1] = ACTIONS(1570),
    [sym_concat_operator] = ACTIONS(1572),
    [sym_rel_operator] = ACTIONS(1570),
    [anon_sym_AMP_AMP] = ACTIONS(1570),
    [anon_sym_or] = ACTIONS(1570),
    [anon_sym_PIPE_PIPE] = ACTIONS(1570),
    [sym_assign_operator] = ACTIONS(1572),
    [sym__uppercase_identifier] = ACTIONS(1572),
    [aux_sym_tag_token1] = ACTIONS(1572),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(876)] = {
    [sym_attribute] = STATE(876),
    [sym__lowercase_identifier] = ACTIONS(1558),
    [anon_sym_TILDE] = ACTIONS(1558),
    [anon_sym_QMARK] = ACTIONS(1558),
    [anon_sym_LPAREN] = ACTIONS(1560),
    [anon_sym_RPAREN] = ACTIONS(1560),
    [anon_sym_COMMA] = ACTIONS(1560),
    [anon_sym_PIPE] = ACTIONS(1558),
    [anon_sym_COLON2] = ACTIONS(1558),
    [anon_sym_DOT] = ACTIONS(1560),
    [anon_sym_DASH_GT] = ACTIONS(1558),
    [anon_sym_LBRACK] = ACTIONS(1558),
    [anon_sym_RBRACK] = ACTIONS(1560),
    [anon_sym_true] = ACTIONS(1558),
    [anon_sym_false] = ACTIONS(1558),
    [anon_sym_LBRACE] = ACTIONS(1558),
    [anon_sym_SEMI] = ACTIONS(1560),
    [anon_sym_RBRACE] = ACTIONS(1560),
    [anon_sym_constraint] = ACTIONS(1558),
    [anon_sym_val] = ACTIONS(1558),
    [anon_sym_end] = ACTIONS(1558),
    [anon_sym_with] = ACTIONS(1558),
    [anon_sym_COLON_GT] = ACTIONS(1560),
    [anon_sym_object] = ACTIONS(1558),
    [anon_sym_inherit] = ACTIONS(1558),
    [anon_sym_method] = ACTIONS(1558),
    [anon_sym_initializer] = ACTIONS(1558),
    [anon_sym_AMP] = ACTIONS(1558),
    [anon_sym_POUND] = ACTIONS(1558),
    [anon_sym_COLON_COLON] = ACTIONS(1560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1560),
    [anon_sym_LT_DASH] = ACTIONS(1558),
    [anon_sym_then] = ACTIONS(1558),
    [anon_sym_else] = ACTIONS(1558),
    [anon_sym_do] = ACTIONS(1558),
    [anon_sym_done] = ACTIONS(1558),
    [anon_sym_to] = ACTIONS(1558),
    [anon_sym_downto] = ACTIONS(1558),
    [anon_sym_new] = ACTIONS(1558),
    [anon_sym_LBRACE_LT] = ACTIONS(1560),
    [anon_sym_GT_RBRACE] = ACTIONS(1560),
    [anon_sym_begin] = ACTIONS(1558),
    [sym_ocamlyacc_value] = ACTIONS(1560),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1558),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1560),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1558),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1558),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1560),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1560),
    [aux_sym_number_token1] = ACTIONS(1560),
    [aux_sym_signed_number_token1] = ACTIONS(1558),
    [anon_sym_SQUOTE] = ACTIONS(1560),
    [anon_sym_DQUOTE] = ACTIONS(1560),
    [sym_prefix_operator] = ACTIONS(1560),
    [aux_sym_sign_operator_token1] = ACTIONS(1558),
    [sym_hash_operator] = ACTIONS(1560),
    [anon_sym_lsl] = ACTIONS(1558),
    [anon_sym_lsr] = ACTIONS(1558),
    [anon_sym_asr] = ACTIONS(1558),
    [aux_sym_pow_operator_token1] = ACTIONS(1560),
    [anon_sym_mod] = ACTIONS(1558),
    [anon_sym_land] = ACTIONS(1558),
    [anon_sym_lor] = ACTIONS(1558),
    [anon_sym_lxor] = ACTIONS(1558),
    [aux_sym_mult_operator_token1] = ACTIONS(1558),
    [aux_sym_add_operator_token1] = ACTIONS(1558),
    [sym_concat_operator] = ACTIONS(1560),
    [sym_rel_operator] = ACTIONS(1558),
    [anon_sym_AMP_AMP] = ACTIONS(1558),
    [anon_sym_or] = ACTIONS(1558),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [sym_assign_operator] = ACTIONS(1560),
    [sym__uppercase_identifier] = ACTIONS(1560),
    [aux_sym_tag_token1] = ACTIONS(1560),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(877)] = {
    [sym_attribute] = STATE(877),
    [sym__lowercase_identifier] = ACTIONS(2123),
    [anon_sym_and] = ACTIONS(2123),
    [anon_sym_TILDE] = ACTIONS(2123),
    [anon_sym_QMARK] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(2125),
    [anon_sym_RPAREN] = ACTIONS(2125),
    [anon_sym_COMMA] = ACTIONS(2125),
    [anon_sym_PIPE] = ACTIONS(2123),
    [anon_sym_COLON2] = ACTIONS(2123),
    [anon_sym_DASH_GT] = ACTIONS(2123),
    [anon_sym_LBRACK] = ACTIONS(2123),
    [anon_sym_RBRACK] = ACTIONS(2125),
    [anon_sym_true] = ACTIONS(2123),
    [anon_sym_false] = ACTIONS(2123),
    [anon_sym_LBRACE] = ACTIONS(2123),
    [anon_sym_SEMI] = ACTIONS(2125),
    [anon_sym_RBRACE] = ACTIONS(2125),
    [anon_sym_constraint] = ACTIONS(2123),
    [anon_sym_val] = ACTIONS(2123),
    [anon_sym_end] = ACTIONS(2123),
    [anon_sym_with] = ACTIONS(2123),
    [anon_sym_COLON_GT] = ACTIONS(2125),
    [anon_sym_object] = ACTIONS(2123),
    [anon_sym_inherit] = ACTIONS(2123),
    [anon_sym_method] = ACTIONS(2123),
    [anon_sym_in] = ACTIONS(2123),
    [anon_sym_as] = ACTIONS(2123),
    [anon_sym_initializer] = ACTIONS(2123),
    [anon_sym_AMP] = ACTIONS(2123),
    [anon_sym_COLON_COLON] = ACTIONS(2125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2125),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2125),
    [anon_sym_then] = ACTIONS(2123),
    [anon_sym_else] = ACTIONS(2123),
    [anon_sym_do] = ACTIONS(2123),
    [anon_sym_done] = ACTIONS(2123),
    [anon_sym_to] = ACTIONS(2123),
    [anon_sym_downto] = ACTIONS(2123),
    [anon_sym_new] = ACTIONS(2123),
    [anon_sym_LBRACE_LT] = ACTIONS(2125),
    [anon_sym_GT_RBRACE] = ACTIONS(2125),
    [anon_sym_begin] = ACTIONS(2123),
    [sym_ocamlyacc_value] = ACTIONS(2125),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2123),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2125),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2125),
    [aux_sym_number_token1] = ACTIONS(2125),
    [aux_sym_signed_number_token1] = ACTIONS(2123),
    [anon_sym_SQUOTE] = ACTIONS(2125),
    [anon_sym_DQUOTE] = ACTIONS(2125),
    [sym_prefix_operator] = ACTIONS(2125),
    [aux_sym_sign_operator_token1] = ACTIONS(2123),
    [anon_sym_lsl] = ACTIONS(2123),
    [anon_sym_lsr] = ACTIONS(2123),
    [anon_sym_asr] = ACTIONS(2123),
    [aux_sym_pow_operator_token1] = ACTIONS(2125),
    [anon_sym_mod] = ACTIONS(2123),
    [anon_sym_land] = ACTIONS(2123),
    [anon_sym_lor] = ACTIONS(2123),
    [anon_sym_lxor] = ACTIONS(2123),
    [aux_sym_mult_operator_token1] = ACTIONS(2123),
    [aux_sym_add_operator_token1] = ACTIONS(2123),
    [sym_concat_operator] = ACTIONS(2125),
    [sym_rel_operator] = ACTIONS(2123),
    [anon_sym_AMP_AMP] = ACTIONS(2123),
    [anon_sym_or] = ACTIONS(2123),
    [anon_sym_PIPE_PIPE] = ACTIONS(2123),
    [sym_assign_operator] = ACTIONS(2125),
    [sym__uppercase_identifier] = ACTIONS(2125),
    [aux_sym_tag_token1] = ACTIONS(2125),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(878)] = {
    [sym_attribute] = STATE(878),
    [sym__lowercase_identifier] = ACTIONS(1562),
    [anon_sym_TILDE] = ACTIONS(1562),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_LPAREN] = ACTIONS(1564),
    [anon_sym_RPAREN] = ACTIONS(1564),
    [anon_sym_COMMA] = ACTIONS(1564),
    [anon_sym_PIPE] = ACTIONS(1562),
    [anon_sym_COLON2] = ACTIONS(1562),
    [anon_sym_DOT] = ACTIONS(1564),
    [anon_sym_DASH_GT] = ACTIONS(1562),
    [anon_sym_LBRACK] = ACTIONS(1562),
    [anon_sym_RBRACK] = ACTIONS(1564),
    [anon_sym_true] = ACTIONS(1562),
    [anon_sym_false] = ACTIONS(1562),
    [anon_sym_LBRACE] = ACTIONS(1562),
    [anon_sym_SEMI] = ACTIONS(1564),
    [anon_sym_RBRACE] = ACTIONS(1564),
    [anon_sym_constraint] = ACTIONS(1562),
    [anon_sym_val] = ACTIONS(1562),
    [anon_sym_end] = ACTIONS(1562),
    [anon_sym_with] = ACTIONS(1562),
    [anon_sym_COLON_GT] = ACTIONS(1564),
    [anon_sym_object] = ACTIONS(1562),
    [anon_sym_inherit] = ACTIONS(1562),
    [anon_sym_method] = ACTIONS(1562),
    [anon_sym_initializer] = ACTIONS(1562),
    [anon_sym_AMP] = ACTIONS(1562),
    [anon_sym_POUND] = ACTIONS(1562),
    [anon_sym_COLON_COLON] = ACTIONS(1564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1564),
    [anon_sym_LT_DASH] = ACTIONS(1562),
    [anon_sym_then] = ACTIONS(1562),
    [anon_sym_else] = ACTIONS(1562),
    [anon_sym_do] = ACTIONS(1562),
    [anon_sym_done] = ACTIONS(1562),
    [anon_sym_to] = ACTIONS(1562),
    [anon_sym_downto] = ACTIONS(1562),
    [anon_sym_new] = ACTIONS(1562),
    [anon_sym_LBRACE_LT] = ACTIONS(1564),
    [anon_sym_GT_RBRACE] = ACTIONS(1564),
    [anon_sym_begin] = ACTIONS(1562),
    [sym_ocamlyacc_value] = ACTIONS(1564),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1562),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1564),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1562),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1562),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1564),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1564),
    [aux_sym_number_token1] = ACTIONS(1564),
    [aux_sym_signed_number_token1] = ACTIONS(1562),
    [anon_sym_SQUOTE] = ACTIONS(1564),
    [anon_sym_DQUOTE] = ACTIONS(1564),
    [sym_prefix_operator] = ACTIONS(1564),
    [aux_sym_sign_operator_token1] = ACTIONS(1562),
    [sym_hash_operator] = ACTIONS(1564),
    [anon_sym_lsl] = ACTIONS(1562),
    [anon_sym_lsr] = ACTIONS(1562),
    [anon_sym_asr] = ACTIONS(1562),
    [aux_sym_pow_operator_token1] = ACTIONS(1564),
    [anon_sym_mod] = ACTIONS(1562),
    [anon_sym_land] = ACTIONS(1562),
    [anon_sym_lor] = ACTIONS(1562),
    [anon_sym_lxor] = ACTIONS(1562),
    [aux_sym_mult_operator_token1] = ACTIONS(1562),
    [aux_sym_add_operator_token1] = ACTIONS(1562),
    [sym_concat_operator] = ACTIONS(1564),
    [sym_rel_operator] = ACTIONS(1562),
    [anon_sym_AMP_AMP] = ACTIONS(1562),
    [anon_sym_or] = ACTIONS(1562),
    [anon_sym_PIPE_PIPE] = ACTIONS(1562),
    [sym_assign_operator] = ACTIONS(1564),
    [sym__uppercase_identifier] = ACTIONS(1564),
    [aux_sym_tag_token1] = ACTIONS(1564),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(879)] = {
    [sym_parameter] = STATE(2475),
    [sym__class_typed] = STATE(3627),
    [sym__parenthesized_abstract_type] = STATE(2475),
    [sym__delimited_pattern] = STATE(2450),
    [sym__simple_pattern] = STATE(2476),
    [sym_typed_pattern] = STATE(2450),
    [sym_polymorphic_variant_pattern] = STATE(2450),
    [sym_record_pattern] = STATE(2478),
    [sym_list_pattern] = STATE(2478),
    [sym_array_pattern] = STATE(2478),
    [sym_local_open_pattern] = STATE(2450),
    [sym_package_pattern] = STATE(2450),
    [sym_parenthesized_pattern] = STATE(2478),
    [sym_attribute] = STATE(879),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2450),
    [sym_quoted_extension] = STATE(2450),
    [sym__constant] = STATE(2464),
    [sym__signed_constant] = STATE(2450),
    [sym_number] = STATE(2504),
    [sym_signed_number] = STATE(2464),
    [sym_character] = STATE(2504),
    [sym_string] = STATE(2504),
    [sym_quoted_string] = STATE(2504),
    [sym_boolean] = STATE(2504),
    [sym_unit] = STATE(2478),
    [sym__simple_value_pattern] = STATE(2450),
    [sym_parenthesized_operator] = STATE(2450),
    [sym_module_path] = STATE(6504),
    [sym_constructor_path] = STATE(2450),
    [sym__constructor_name] = STATE(2515),
    [sym_tag] = STATE(2450),
    [aux_sym_expression_item_repeat1] = STATE(3658),
    [aux_sym_let_binding_repeat1] = STATE(856),
    [sym__lowercase_identifier] = ACTIONS(2029),
    [anon_sym_SEMI_SEMI] = ACTIONS(2127),
    [anon_sym_let] = ACTIONS(2129),
    [anon_sym_and] = ACTIONS(2129),
    [anon_sym_EQ] = ACTIONS(2131),
    [anon_sym_TILDE] = ACTIONS(2037),
    [anon_sym_QMARK] = ACTIONS(2037),
    [anon_sym_LPAREN] = ACTIONS(2039),
    [anon_sym_external] = ACTIONS(2129),
    [anon_sym_type] = ACTIONS(2129),
    [anon_sym_COLON2] = ACTIONS(2041),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2127),
    [anon_sym_true] = ACTIONS(2045),
    [anon_sym_false] = ACTIONS(2045),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2129),
    [anon_sym_module] = ACTIONS(2129),
    [anon_sym_open] = ACTIONS(2129),
    [anon_sym_include] = ACTIONS(2129),
    [anon_sym_class] = ACTIONS(2129),
    [anon_sym_val] = ACTIONS(2129),
    [anon_sym_end] = ACTIONS(2129),
    [anon_sym_POUND] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2051),
    [anon_sym_begin] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2127),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2127),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2127),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2067),
    [sym_let_operator] = ACTIONS(2127),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2129),
    [aux_sym_tag_token1] = ACTIONS(2071),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(880)] = {
    [sym_attribute] = STATE(880),
    [sym__lowercase_identifier] = ACTIONS(1566),
    [anon_sym_TILDE] = ACTIONS(1566),
    [anon_sym_QMARK] = ACTIONS(1566),
    [anon_sym_LPAREN] = ACTIONS(1568),
    [anon_sym_RPAREN] = ACTIONS(1568),
    [anon_sym_COMMA] = ACTIONS(1568),
    [anon_sym_PIPE] = ACTIONS(1566),
    [anon_sym_COLON2] = ACTIONS(1566),
    [anon_sym_DOT] = ACTIONS(1568),
    [anon_sym_DASH_GT] = ACTIONS(1566),
    [anon_sym_LBRACK] = ACTIONS(1566),
    [anon_sym_RBRACK] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1566),
    [anon_sym_false] = ACTIONS(1566),
    [anon_sym_LBRACE] = ACTIONS(1566),
    [anon_sym_SEMI] = ACTIONS(1568),
    [anon_sym_RBRACE] = ACTIONS(1568),
    [anon_sym_constraint] = ACTIONS(1566),
    [anon_sym_val] = ACTIONS(1566),
    [anon_sym_end] = ACTIONS(1566),
    [anon_sym_with] = ACTIONS(1566),
    [anon_sym_COLON_GT] = ACTIONS(1568),
    [anon_sym_object] = ACTIONS(1566),
    [anon_sym_inherit] = ACTIONS(1566),
    [anon_sym_method] = ACTIONS(1566),
    [anon_sym_initializer] = ACTIONS(1566),
    [anon_sym_AMP] = ACTIONS(1566),
    [anon_sym_POUND] = ACTIONS(1566),
    [anon_sym_COLON_COLON] = ACTIONS(1568),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1568),
    [anon_sym_LT_DASH] = ACTIONS(1566),
    [anon_sym_then] = ACTIONS(1566),
    [anon_sym_else] = ACTIONS(1566),
    [anon_sym_do] = ACTIONS(1566),
    [anon_sym_done] = ACTIONS(1566),
    [anon_sym_to] = ACTIONS(1566),
    [anon_sym_downto] = ACTIONS(1566),
    [anon_sym_new] = ACTIONS(1566),
    [anon_sym_LBRACE_LT] = ACTIONS(1568),
    [anon_sym_GT_RBRACE] = ACTIONS(1568),
    [anon_sym_begin] = ACTIONS(1566),
    [sym_ocamlyacc_value] = ACTIONS(1568),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1566),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1568),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1566),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1566),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1568),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1568),
    [aux_sym_number_token1] = ACTIONS(1568),
    [aux_sym_signed_number_token1] = ACTIONS(1566),
    [anon_sym_SQUOTE] = ACTIONS(1568),
    [anon_sym_DQUOTE] = ACTIONS(1568),
    [sym_prefix_operator] = ACTIONS(1568),
    [aux_sym_sign_operator_token1] = ACTIONS(1566),
    [sym_hash_operator] = ACTIONS(1568),
    [anon_sym_lsl] = ACTIONS(1566),
    [anon_sym_lsr] = ACTIONS(1566),
    [anon_sym_asr] = ACTIONS(1566),
    [aux_sym_pow_operator_token1] = ACTIONS(1568),
    [anon_sym_mod] = ACTIONS(1566),
    [anon_sym_land] = ACTIONS(1566),
    [anon_sym_lor] = ACTIONS(1566),
    [anon_sym_lxor] = ACTIONS(1566),
    [aux_sym_mult_operator_token1] = ACTIONS(1566),
    [aux_sym_add_operator_token1] = ACTIONS(1566),
    [sym_concat_operator] = ACTIONS(1568),
    [sym_rel_operator] = ACTIONS(1566),
    [anon_sym_AMP_AMP] = ACTIONS(1566),
    [anon_sym_or] = ACTIONS(1566),
    [anon_sym_PIPE_PIPE] = ACTIONS(1566),
    [sym_assign_operator] = ACTIONS(1568),
    [sym__uppercase_identifier] = ACTIONS(1568),
    [aux_sym_tag_token1] = ACTIONS(1568),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(881)] = {
    [sym_parameter] = STATE(2475),
    [sym__class_typed] = STATE(3644),
    [sym__parenthesized_abstract_type] = STATE(2475),
    [sym__delimited_pattern] = STATE(2450),
    [sym__simple_pattern] = STATE(2476),
    [sym_typed_pattern] = STATE(2450),
    [sym_polymorphic_variant_pattern] = STATE(2450),
    [sym_record_pattern] = STATE(2478),
    [sym_list_pattern] = STATE(2478),
    [sym_array_pattern] = STATE(2478),
    [sym_local_open_pattern] = STATE(2450),
    [sym_package_pattern] = STATE(2450),
    [sym_parenthesized_pattern] = STATE(2478),
    [sym_attribute] = STATE(881),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2450),
    [sym_quoted_extension] = STATE(2450),
    [sym__constant] = STATE(2464),
    [sym__signed_constant] = STATE(2450),
    [sym_number] = STATE(2504),
    [sym_signed_number] = STATE(2464),
    [sym_character] = STATE(2504),
    [sym_string] = STATE(2504),
    [sym_quoted_string] = STATE(2504),
    [sym_boolean] = STATE(2504),
    [sym_unit] = STATE(2478),
    [sym__simple_value_pattern] = STATE(2450),
    [sym_parenthesized_operator] = STATE(2450),
    [sym_module_path] = STATE(6504),
    [sym_constructor_path] = STATE(2450),
    [sym__constructor_name] = STATE(2515),
    [sym_tag] = STATE(2450),
    [aux_sym_expression_item_repeat1] = STATE(3707),
    [aux_sym_let_binding_repeat1] = STATE(1012),
    [sym__lowercase_identifier] = ACTIONS(2029),
    [anon_sym_SEMI_SEMI] = ACTIONS(2133),
    [anon_sym_let] = ACTIONS(2135),
    [anon_sym_and] = ACTIONS(2135),
    [anon_sym_EQ] = ACTIONS(2137),
    [anon_sym_TILDE] = ACTIONS(2037),
    [anon_sym_QMARK] = ACTIONS(2037),
    [anon_sym_LPAREN] = ACTIONS(2039),
    [anon_sym_external] = ACTIONS(2135),
    [anon_sym_type] = ACTIONS(2135),
    [anon_sym_COLON2] = ACTIONS(2041),
    [anon_sym_LBRACK] = ACTIONS(2043),
    [anon_sym_RBRACK] = ACTIONS(2133),
    [anon_sym_true] = ACTIONS(2045),
    [anon_sym_false] = ACTIONS(2045),
    [anon_sym_LBRACE] = ACTIONS(2047),
    [anon_sym_exception] = ACTIONS(2135),
    [anon_sym_module] = ACTIONS(2135),
    [anon_sym_open] = ACTIONS(2135),
    [anon_sym_include] = ACTIONS(2135),
    [anon_sym_class] = ACTIONS(2135),
    [anon_sym_val] = ACTIONS(2135),
    [anon_sym_end] = ACTIONS(2135),
    [anon_sym_POUND] = ACTIONS(2049),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2051),
    [anon_sym_begin] = ACTIONS(2053),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2133),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2059),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2133),
    [aux_sym_number_token1] = ACTIONS(2061),
    [aux_sym_signed_number_token1] = ACTIONS(2063),
    [anon_sym_SQUOTE] = ACTIONS(2065),
    [anon_sym_DQUOTE] = ACTIONS(2067),
    [sym_let_operator] = ACTIONS(2133),
    [sym__uppercase_identifier] = ACTIONS(2069),
    [aux_sym_directive_token1] = ACTIONS(2135),
    [aux_sym_tag_token1] = ACTIONS(2071),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(882)] = {
    [sym_attribute] = STATE(882),
    [sym__lowercase_identifier] = ACTIONS(2021),
    [anon_sym_TILDE] = ACTIONS(2021),
    [anon_sym_QMARK] = ACTIONS(2021),
    [anon_sym_LPAREN] = ACTIONS(2023),
    [anon_sym_RPAREN] = ACTIONS(2023),
    [anon_sym_COMMA] = ACTIONS(2023),
    [anon_sym_PIPE] = ACTIONS(2021),
    [anon_sym_COLON2] = ACTIONS(2021),
    [anon_sym_DOT] = ACTIONS(2023),
    [anon_sym_DASH_GT] = ACTIONS(2021),
    [anon_sym_LBRACK] = ACTIONS(2021),
    [anon_sym_RBRACK] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(2021),
    [anon_sym_false] = ACTIONS(2021),
    [anon_sym_LBRACE] = ACTIONS(2021),
    [anon_sym_SEMI] = ACTIONS(2023),
    [anon_sym_RBRACE] = ACTIONS(2023),
    [anon_sym_constraint] = ACTIONS(2021),
    [anon_sym_val] = ACTIONS(2021),
    [anon_sym_end] = ACTIONS(2021),
    [anon_sym_with] = ACTIONS(2021),
    [anon_sym_COLON_GT] = ACTIONS(2023),
    [anon_sym_object] = ACTIONS(2021),
    [anon_sym_inherit] = ACTIONS(2021),
    [anon_sym_method] = ACTIONS(2021),
    [anon_sym_initializer] = ACTIONS(2021),
    [anon_sym_AMP] = ACTIONS(2021),
    [anon_sym_POUND] = ACTIONS(2021),
    [anon_sym_COLON_COLON] = ACTIONS(2023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2023),
    [anon_sym_then] = ACTIONS(2021),
    [anon_sym_else] = ACTIONS(2021),
    [anon_sym_do] = ACTIONS(2021),
    [anon_sym_done] = ACTIONS(2021),
    [anon_sym_to] = ACTIONS(2021),
    [anon_sym_downto] = ACTIONS(2021),
    [anon_sym_new] = ACTIONS(2021),
    [anon_sym_LBRACE_LT] = ACTIONS(2023),
    [anon_sym_GT_RBRACE] = ACTIONS(2023),
    [anon_sym_begin] = ACTIONS(2021),
    [sym_ocamlyacc_value] = ACTIONS(2023),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2023),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2023),
    [aux_sym_number_token1] = ACTIONS(2023),
    [aux_sym_signed_number_token1] = ACTIONS(2021),
    [anon_sym_SQUOTE] = ACTIONS(2023),
    [anon_sym_DQUOTE] = ACTIONS(2023),
    [sym_prefix_operator] = ACTIONS(2023),
    [aux_sym_sign_operator_token1] = ACTIONS(2021),
    [sym_hash_operator] = ACTIONS(2023),
    [anon_sym_lsl] = ACTIONS(2021),
    [anon_sym_lsr] = ACTIONS(2021),
    [anon_sym_asr] = ACTIONS(2021),
    [aux_sym_pow_operator_token1] = ACTIONS(2023),
    [anon_sym_mod] = ACTIONS(2021),
    [anon_sym_land] = ACTIONS(2021),
    [anon_sym_lor] = ACTIONS(2021),
    [anon_sym_lxor] = ACTIONS(2021),
    [aux_sym_mult_operator_token1] = ACTIONS(2021),
    [aux_sym_add_operator_token1] = ACTIONS(2021),
    [sym_concat_operator] = ACTIONS(2023),
    [sym_rel_operator] = ACTIONS(2021),
    [anon_sym_AMP_AMP] = ACTIONS(2021),
    [anon_sym_or] = ACTIONS(2021),
    [anon_sym_PIPE_PIPE] = ACTIONS(2021),
    [sym_assign_operator] = ACTIONS(2023),
    [sym__uppercase_identifier] = ACTIONS(2023),
    [aux_sym_tag_token1] = ACTIONS(2023),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(883)] = {
    [sym_attribute] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1850),
    [anon_sym_TILDE] = ACTIONS(1850),
    [anon_sym_QMARK] = ACTIONS(1850),
    [anon_sym_LPAREN] = ACTIONS(1852),
    [anon_sym_RPAREN] = ACTIONS(1852),
    [anon_sym_COMMA] = ACTIONS(1852),
    [anon_sym_PIPE] = ACTIONS(1850),
    [anon_sym_COLON2] = ACTIONS(1850),
    [anon_sym_DOT] = ACTIONS(1852),
    [anon_sym_DASH_GT] = ACTIONS(1850),
    [anon_sym_LBRACK] = ACTIONS(1850),
    [anon_sym_RBRACK] = ACTIONS(1852),
    [anon_sym_true] = ACTIONS(1850),
    [anon_sym_false] = ACTIONS(1850),
    [anon_sym_LBRACE] = ACTIONS(1850),
    [anon_sym_SEMI] = ACTIONS(1852),
    [anon_sym_RBRACE] = ACTIONS(1852),
    [anon_sym_constraint] = ACTIONS(1850),
    [anon_sym_val] = ACTIONS(1850),
    [anon_sym_end] = ACTIONS(1850),
    [anon_sym_with] = ACTIONS(1850),
    [anon_sym_COLON_GT] = ACTIONS(1852),
    [anon_sym_object] = ACTIONS(1850),
    [anon_sym_inherit] = ACTIONS(1850),
    [anon_sym_method] = ACTIONS(1850),
    [anon_sym_initializer] = ACTIONS(1850),
    [anon_sym_AMP] = ACTIONS(1850),
    [anon_sym_POUND] = ACTIONS(1850),
    [anon_sym_COLON_COLON] = ACTIONS(1852),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1852),
    [anon_sym_then] = ACTIONS(1850),
    [anon_sym_else] = ACTIONS(1850),
    [anon_sym_do] = ACTIONS(1850),
    [anon_sym_done] = ACTIONS(1850),
    [anon_sym_to] = ACTIONS(1850),
    [anon_sym_downto] = ACTIONS(1850),
    [anon_sym_new] = ACTIONS(1850),
    [anon_sym_LBRACE_LT] = ACTIONS(1852),
    [anon_sym_GT_RBRACE] = ACTIONS(1852),
    [anon_sym_begin] = ACTIONS(1850),
    [sym_ocamlyacc_value] = ACTIONS(1852),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1850),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1852),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1850),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1850),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1852),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1852),
    [aux_sym_number_token1] = ACTIONS(1852),
    [aux_sym_signed_number_token1] = ACTIONS(1850),
    [anon_sym_SQUOTE] = ACTIONS(1852),
    [anon_sym_DQUOTE] = ACTIONS(1852),
    [sym_prefix_operator] = ACTIONS(1852),
    [aux_sym_sign_operator_token1] = ACTIONS(1850),
    [sym_hash_operator] = ACTIONS(1852),
    [anon_sym_lsl] = ACTIONS(1850),
    [anon_sym_lsr] = ACTIONS(1850),
    [anon_sym_asr] = ACTIONS(1850),
    [aux_sym_pow_operator_token1] = ACTIONS(1852),
    [anon_sym_mod] = ACTIONS(1850),
    [anon_sym_land] = ACTIONS(1850),
    [anon_sym_lor] = ACTIONS(1850),
    [anon_sym_lxor] = ACTIONS(1850),
    [aux_sym_mult_operator_token1] = ACTIONS(1850),
    [aux_sym_add_operator_token1] = ACTIONS(1850),
    [sym_concat_operator] = ACTIONS(1852),
    [sym_rel_operator] = ACTIONS(1850),
    [anon_sym_AMP_AMP] = ACTIONS(1850),
    [anon_sym_or] = ACTIONS(1850),
    [anon_sym_PIPE_PIPE] = ACTIONS(1850),
    [sym_assign_operator] = ACTIONS(1852),
    [sym__uppercase_identifier] = ACTIONS(1852),
    [aux_sym_tag_token1] = ACTIONS(1852),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(884)] = {
    [sym_attribute] = STATE(884),
    [sym__lowercase_identifier] = ACTIONS(1754),
    [anon_sym_TILDE] = ACTIONS(1754),
    [anon_sym_QMARK] = ACTIONS(1754),
    [anon_sym_LPAREN] = ACTIONS(1756),
    [anon_sym_RPAREN] = ACTIONS(1756),
    [anon_sym_COMMA] = ACTIONS(1756),
    [anon_sym_PIPE] = ACTIONS(1754),
    [anon_sym_COLON2] = ACTIONS(1754),
    [anon_sym_DOT] = ACTIONS(1756),
    [anon_sym_DASH_GT] = ACTIONS(1754),
    [anon_sym_LBRACK] = ACTIONS(1754),
    [anon_sym_RBRACK] = ACTIONS(1756),
    [anon_sym_true] = ACTIONS(1754),
    [anon_sym_false] = ACTIONS(1754),
    [anon_sym_LBRACE] = ACTIONS(1754),
    [anon_sym_SEMI] = ACTIONS(1756),
    [anon_sym_RBRACE] = ACTIONS(1756),
    [anon_sym_constraint] = ACTIONS(1754),
    [anon_sym_val] = ACTIONS(1754),
    [anon_sym_end] = ACTIONS(1754),
    [anon_sym_with] = ACTIONS(1754),
    [anon_sym_COLON_GT] = ACTIONS(1756),
    [anon_sym_object] = ACTIONS(1754),
    [anon_sym_inherit] = ACTIONS(1754),
    [anon_sym_method] = ACTIONS(1754),
    [anon_sym_initializer] = ACTIONS(1754),
    [anon_sym_AMP] = ACTIONS(1754),
    [anon_sym_POUND] = ACTIONS(1754),
    [anon_sym_COLON_COLON] = ACTIONS(1756),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1756),
    [anon_sym_then] = ACTIONS(1754),
    [anon_sym_else] = ACTIONS(1754),
    [anon_sym_do] = ACTIONS(1754),
    [anon_sym_done] = ACTIONS(1754),
    [anon_sym_to] = ACTIONS(1754),
    [anon_sym_downto] = ACTIONS(1754),
    [anon_sym_new] = ACTIONS(1754),
    [anon_sym_LBRACE_LT] = ACTIONS(1756),
    [anon_sym_GT_RBRACE] = ACTIONS(1756),
    [anon_sym_begin] = ACTIONS(1754),
    [sym_ocamlyacc_value] = ACTIONS(1756),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1754),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1756),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1754),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1754),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1756),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1756),
    [aux_sym_number_token1] = ACTIONS(1756),
    [aux_sym_signed_number_token1] = ACTIONS(1754),
    [anon_sym_SQUOTE] = ACTIONS(1756),
    [anon_sym_DQUOTE] = ACTIONS(1756),
    [sym_prefix_operator] = ACTIONS(1756),
    [aux_sym_sign_operator_token1] = ACTIONS(1754),
    [sym_hash_operator] = ACTIONS(1756),
    [anon_sym_lsl] = ACTIONS(1754),
    [anon_sym_lsr] = ACTIONS(1754),
    [anon_sym_asr] = ACTIONS(1754),
    [aux_sym_pow_operator_token1] = ACTIONS(1756),
    [anon_sym_mod] = ACTIONS(1754),
    [anon_sym_land] = ACTIONS(1754),
    [anon_sym_lor] = ACTIONS(1754),
    [anon_sym_lxor] = ACTIONS(1754),
    [aux_sym_mult_operator_token1] = ACTIONS(1754),
    [aux_sym_add_operator_token1] = ACTIONS(1754),
    [sym_concat_operator] = ACTIONS(1756),
    [sym_rel_operator] = ACTIONS(1754),
    [anon_sym_AMP_AMP] = ACTIONS(1754),
    [anon_sym_or] = ACTIONS(1754),
    [anon_sym_PIPE_PIPE] = ACTIONS(1754),
    [sym_assign_operator] = ACTIONS(1756),
    [sym__uppercase_identifier] = ACTIONS(1756),
    [aux_sym_tag_token1] = ACTIONS(1756),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(885)] = {
    [sym_attribute] = STATE(885),
    [sym__lowercase_identifier] = ACTIONS(1950),
    [anon_sym_TILDE] = ACTIONS(1950),
    [anon_sym_QMARK] = ACTIONS(1950),
    [anon_sym_LPAREN] = ACTIONS(1952),
    [anon_sym_RPAREN] = ACTIONS(1952),
    [anon_sym_COMMA] = ACTIONS(1952),
    [anon_sym_PIPE] = ACTIONS(1950),
    [anon_sym_COLON2] = ACTIONS(1950),
    [anon_sym_DOT] = ACTIONS(1952),
    [anon_sym_DASH_GT] = ACTIONS(1950),
    [anon_sym_LBRACK] = ACTIONS(1950),
    [anon_sym_RBRACK] = ACTIONS(1952),
    [anon_sym_true] = ACTIONS(1950),
    [anon_sym_false] = ACTIONS(1950),
    [anon_sym_LBRACE] = ACTIONS(1950),
    [anon_sym_SEMI] = ACTIONS(1952),
    [anon_sym_RBRACE] = ACTIONS(1952),
    [anon_sym_constraint] = ACTIONS(1950),
    [anon_sym_val] = ACTIONS(1950),
    [anon_sym_end] = ACTIONS(1950),
    [anon_sym_with] = ACTIONS(1950),
    [anon_sym_COLON_GT] = ACTIONS(1952),
    [anon_sym_object] = ACTIONS(1950),
    [anon_sym_inherit] = ACTIONS(1950),
    [anon_sym_method] = ACTIONS(1950),
    [anon_sym_initializer] = ACTIONS(1950),
    [anon_sym_AMP] = ACTIONS(1950),
    [anon_sym_POUND] = ACTIONS(1950),
    [anon_sym_COLON_COLON] = ACTIONS(1952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1952),
    [anon_sym_then] = ACTIONS(1950),
    [anon_sym_else] = ACTIONS(1950),
    [anon_sym_do] = ACTIONS(1950),
    [anon_sym_done] = ACTIONS(1950),
    [anon_sym_to] = ACTIONS(1950),
    [anon_sym_downto] = ACTIONS(1950),
    [anon_sym_new] = ACTIONS(1950),
    [anon_sym_LBRACE_LT] = ACTIONS(1952),
    [anon_sym_GT_RBRACE] = ACTIONS(1952),
    [anon_sym_begin] = ACTIONS(1950),
    [sym_ocamlyacc_value] = ACTIONS(1952),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1950),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1952),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1950),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1950),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1952),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1952),
    [aux_sym_number_token1] = ACTIONS(1952),
    [aux_sym_signed_number_token1] = ACTIONS(1950),
    [anon_sym_SQUOTE] = ACTIONS(1952),
    [anon_sym_DQUOTE] = ACTIONS(1952),
    [sym_prefix_operator] = ACTIONS(1952),
    [aux_sym_sign_operator_token1] = ACTIONS(1950),
    [sym_hash_operator] = ACTIONS(1952),
    [anon_sym_lsl] = ACTIONS(1950),
    [anon_sym_lsr] = ACTIONS(1950),
    [anon_sym_asr] = ACTIONS(1950),
    [aux_sym_pow_operator_token1] = ACTIONS(1952),
    [anon_sym_mod] = ACTIONS(1950),
    [anon_sym_land] = ACTIONS(1950),
    [anon_sym_lor] = ACTIONS(1950),
    [anon_sym_lxor] = ACTIONS(1950),
    [aux_sym_mult_operator_token1] = ACTIONS(1950),
    [aux_sym_add_operator_token1] = ACTIONS(1950),
    [sym_concat_operator] = ACTIONS(1952),
    [sym_rel_operator] = ACTIONS(1950),
    [anon_sym_AMP_AMP] = ACTIONS(1950),
    [anon_sym_or] = ACTIONS(1950),
    [anon_sym_PIPE_PIPE] = ACTIONS(1950),
    [sym_assign_operator] = ACTIONS(1952),
    [sym__uppercase_identifier] = ACTIONS(1952),
    [aux_sym_tag_token1] = ACTIONS(1952),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(886)] = {
    [sym_attribute] = STATE(886),
    [sym__lowercase_identifier] = ACTIONS(1698),
    [anon_sym_TILDE] = ACTIONS(1698),
    [anon_sym_QMARK] = ACTIONS(1698),
    [anon_sym_LPAREN] = ACTIONS(1700),
    [anon_sym_RPAREN] = ACTIONS(1700),
    [anon_sym_COMMA] = ACTIONS(1700),
    [anon_sym_PIPE] = ACTIONS(1698),
    [anon_sym_COLON2] = ACTIONS(1698),
    [anon_sym_DOT] = ACTIONS(1744),
    [anon_sym_DASH_GT] = ACTIONS(1698),
    [anon_sym_LBRACK] = ACTIONS(1698),
    [anon_sym_RBRACK] = ACTIONS(1700),
    [anon_sym_true] = ACTIONS(1698),
    [anon_sym_false] = ACTIONS(1698),
    [anon_sym_LBRACE] = ACTIONS(1698),
    [anon_sym_SEMI] = ACTIONS(1700),
    [anon_sym_RBRACE] = ACTIONS(1700),
    [anon_sym_constraint] = ACTIONS(1698),
    [anon_sym_val] = ACTIONS(1698),
    [anon_sym_end] = ACTIONS(1698),
    [anon_sym_with] = ACTIONS(1698),
    [anon_sym_COLON_GT] = ACTIONS(1700),
    [anon_sym_object] = ACTIONS(1698),
    [anon_sym_inherit] = ACTIONS(1698),
    [anon_sym_method] = ACTIONS(1698),
    [anon_sym_initializer] = ACTIONS(1698),
    [anon_sym_AMP] = ACTIONS(1698),
    [anon_sym_POUND] = ACTIONS(1698),
    [anon_sym_COLON_COLON] = ACTIONS(1700),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1700),
    [anon_sym_then] = ACTIONS(1698),
    [anon_sym_else] = ACTIONS(1698),
    [anon_sym_do] = ACTIONS(1698),
    [anon_sym_done] = ACTIONS(1698),
    [anon_sym_to] = ACTIONS(1698),
    [anon_sym_downto] = ACTIONS(1698),
    [anon_sym_new] = ACTIONS(1698),
    [anon_sym_LBRACE_LT] = ACTIONS(1700),
    [anon_sym_GT_RBRACE] = ACTIONS(1700),
    [anon_sym_begin] = ACTIONS(1698),
    [sym_ocamlyacc_value] = ACTIONS(1700),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1698),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1700),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1700),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1700),
    [aux_sym_number_token1] = ACTIONS(1700),
    [aux_sym_signed_number_token1] = ACTIONS(1698),
    [anon_sym_SQUOTE] = ACTIONS(1700),
    [anon_sym_DQUOTE] = ACTIONS(1700),
    [sym_prefix_operator] = ACTIONS(1700),
    [aux_sym_sign_operator_token1] = ACTIONS(1698),
    [sym_hash_operator] = ACTIONS(1700),
    [anon_sym_lsl] = ACTIONS(1698),
    [anon_sym_lsr] = ACTIONS(1698),
    [anon_sym_asr] = ACTIONS(1698),
    [aux_sym_pow_operator_token1] = ACTIONS(1700),
    [anon_sym_mod] = ACTIONS(1698),
    [anon_sym_land] = ACTIONS(1698),
    [anon_sym_lor] = ACTIONS(1698),
    [anon_sym_lxor] = ACTIONS(1698),
    [aux_sym_mult_operator_token1] = ACTIONS(1698),
    [aux_sym_add_operator_token1] = ACTIONS(1698),
    [sym_concat_operator] = ACTIONS(1700),
    [sym_rel_operator] = ACTIONS(1698),
    [anon_sym_AMP_AMP] = ACTIONS(1698),
    [anon_sym_or] = ACTIONS(1698),
    [anon_sym_PIPE_PIPE] = ACTIONS(1698),
    [sym_assign_operator] = ACTIONS(1700),
    [sym__uppercase_identifier] = ACTIONS(1700),
    [aux_sym_tag_token1] = ACTIONS(1700),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(887)] = {
    [sym_attribute] = STATE(887),
    [sym__lowercase_identifier] = ACTIONS(1902),
    [anon_sym_TILDE] = ACTIONS(1902),
    [anon_sym_QMARK] = ACTIONS(1902),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_RPAREN] = ACTIONS(1904),
    [anon_sym_COMMA] = ACTIONS(1904),
    [anon_sym_PIPE] = ACTIONS(1902),
    [anon_sym_COLON2] = ACTIONS(1902),
    [anon_sym_DOT] = ACTIONS(1904),
    [anon_sym_DASH_GT] = ACTIONS(1902),
    [anon_sym_LBRACK] = ACTIONS(1902),
    [anon_sym_RBRACK] = ACTIONS(1904),
    [anon_sym_true] = ACTIONS(1902),
    [anon_sym_false] = ACTIONS(1902),
    [anon_sym_LBRACE] = ACTIONS(1902),
    [anon_sym_SEMI] = ACTIONS(1904),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [anon_sym_constraint] = ACTIONS(1902),
    [anon_sym_val] = ACTIONS(1902),
    [anon_sym_end] = ACTIONS(1902),
    [anon_sym_with] = ACTIONS(1902),
    [anon_sym_COLON_GT] = ACTIONS(1904),
    [anon_sym_object] = ACTIONS(1902),
    [anon_sym_inherit] = ACTIONS(1902),
    [anon_sym_method] = ACTIONS(1902),
    [anon_sym_initializer] = ACTIONS(1902),
    [anon_sym_AMP] = ACTIONS(1902),
    [anon_sym_POUND] = ACTIONS(1902),
    [anon_sym_COLON_COLON] = ACTIONS(1904),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1904),
    [anon_sym_then] = ACTIONS(1902),
    [anon_sym_else] = ACTIONS(1902),
    [anon_sym_do] = ACTIONS(1902),
    [anon_sym_done] = ACTIONS(1902),
    [anon_sym_to] = ACTIONS(1902),
    [anon_sym_downto] = ACTIONS(1902),
    [anon_sym_new] = ACTIONS(1902),
    [anon_sym_LBRACE_LT] = ACTIONS(1904),
    [anon_sym_GT_RBRACE] = ACTIONS(1904),
    [anon_sym_begin] = ACTIONS(1902),
    [sym_ocamlyacc_value] = ACTIONS(1904),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1902),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1904),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1902),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1902),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1904),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1904),
    [aux_sym_number_token1] = ACTIONS(1904),
    [aux_sym_signed_number_token1] = ACTIONS(1902),
    [anon_sym_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1904),
    [sym_prefix_operator] = ACTIONS(1904),
    [aux_sym_sign_operator_token1] = ACTIONS(1902),
    [sym_hash_operator] = ACTIONS(1904),
    [anon_sym_lsl] = ACTIONS(1902),
    [anon_sym_lsr] = ACTIONS(1902),
    [anon_sym_asr] = ACTIONS(1902),
    [aux_sym_pow_operator_token1] = ACTIONS(1904),
    [anon_sym_mod] = ACTIONS(1902),
    [anon_sym_land] = ACTIONS(1902),
    [anon_sym_lor] = ACTIONS(1902),
    [anon_sym_lxor] = ACTIONS(1902),
    [aux_sym_mult_operator_token1] = ACTIONS(1902),
    [aux_sym_add_operator_token1] = ACTIONS(1902),
    [sym_concat_operator] = ACTIONS(1904),
    [sym_rel_operator] = ACTIONS(1902),
    [anon_sym_AMP_AMP] = ACTIONS(1902),
    [anon_sym_or] = ACTIONS(1902),
    [anon_sym_PIPE_PIPE] = ACTIONS(1902),
    [sym_assign_operator] = ACTIONS(1904),
    [sym__uppercase_identifier] = ACTIONS(1904),
    [aux_sym_tag_token1] = ACTIONS(1904),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(888)] = {
    [sym_attribute] = STATE(888),
    [sym__lowercase_identifier] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1906),
    [anon_sym_QMARK] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1908),
    [anon_sym_RPAREN] = ACTIONS(1908),
    [anon_sym_COMMA] = ACTIONS(1908),
    [anon_sym_PIPE] = ACTIONS(1906),
    [anon_sym_COLON2] = ACTIONS(1906),
    [anon_sym_DOT] = ACTIONS(1908),
    [anon_sym_DASH_GT] = ACTIONS(1906),
    [anon_sym_LBRACK] = ACTIONS(1906),
    [anon_sym_RBRACK] = ACTIONS(1908),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_LBRACE] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1908),
    [anon_sym_RBRACE] = ACTIONS(1908),
    [anon_sym_constraint] = ACTIONS(1906),
    [anon_sym_val] = ACTIONS(1906),
    [anon_sym_end] = ACTIONS(1906),
    [anon_sym_with] = ACTIONS(1906),
    [anon_sym_COLON_GT] = ACTIONS(1908),
    [anon_sym_object] = ACTIONS(1906),
    [anon_sym_inherit] = ACTIONS(1906),
    [anon_sym_method] = ACTIONS(1906),
    [anon_sym_initializer] = ACTIONS(1906),
    [anon_sym_AMP] = ACTIONS(1906),
    [anon_sym_POUND] = ACTIONS(1906),
    [anon_sym_COLON_COLON] = ACTIONS(1908),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1908),
    [anon_sym_then] = ACTIONS(1906),
    [anon_sym_else] = ACTIONS(1906),
    [anon_sym_do] = ACTIONS(1906),
    [anon_sym_done] = ACTIONS(1906),
    [anon_sym_to] = ACTIONS(1906),
    [anon_sym_downto] = ACTIONS(1906),
    [anon_sym_new] = ACTIONS(1906),
    [anon_sym_LBRACE_LT] = ACTIONS(1908),
    [anon_sym_GT_RBRACE] = ACTIONS(1908),
    [anon_sym_begin] = ACTIONS(1906),
    [sym_ocamlyacc_value] = ACTIONS(1908),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1906),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1908),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1906),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1906),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1908),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1908),
    [aux_sym_number_token1] = ACTIONS(1908),
    [aux_sym_signed_number_token1] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1908),
    [anon_sym_DQUOTE] = ACTIONS(1908),
    [sym_prefix_operator] = ACTIONS(1908),
    [aux_sym_sign_operator_token1] = ACTIONS(1906),
    [sym_hash_operator] = ACTIONS(1908),
    [anon_sym_lsl] = ACTIONS(1906),
    [anon_sym_lsr] = ACTIONS(1906),
    [anon_sym_asr] = ACTIONS(1906),
    [aux_sym_pow_operator_token1] = ACTIONS(1908),
    [anon_sym_mod] = ACTIONS(1906),
    [anon_sym_land] = ACTIONS(1906),
    [anon_sym_lor] = ACTIONS(1906),
    [anon_sym_lxor] = ACTIONS(1906),
    [aux_sym_mult_operator_token1] = ACTIONS(1906),
    [aux_sym_add_operator_token1] = ACTIONS(1906),
    [sym_concat_operator] = ACTIONS(1908),
    [sym_rel_operator] = ACTIONS(1906),
    [anon_sym_AMP_AMP] = ACTIONS(1906),
    [anon_sym_or] = ACTIONS(1906),
    [anon_sym_PIPE_PIPE] = ACTIONS(1906),
    [sym_assign_operator] = ACTIONS(1908),
    [sym__uppercase_identifier] = ACTIONS(1908),
    [aux_sym_tag_token1] = ACTIONS(1908),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(889)] = {
    [sym__parenthesized_abstract_type] = STATE(1339),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4358),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(889),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_EQ] = ACTIONS(2139),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2143),
    [anon_sym_RPAREN] = ACTIONS(2139),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2139),
    [anon_sym_COLON2] = ACTIONS(2145),
    [anon_sym_DASH_GT] = ACTIONS(2139),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(2139),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_as] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_when] = ACTIONS(2145),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2149),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(890)] = {
    [sym_attribute] = STATE(890),
    [sym__lowercase_identifier] = ACTIONS(1922),
    [anon_sym_TILDE] = ACTIONS(1922),
    [anon_sym_QMARK] = ACTIONS(1922),
    [anon_sym_LPAREN] = ACTIONS(1924),
    [anon_sym_RPAREN] = ACTIONS(1924),
    [anon_sym_COMMA] = ACTIONS(1924),
    [anon_sym_PIPE] = ACTIONS(1922),
    [anon_sym_COLON2] = ACTIONS(1922),
    [anon_sym_DOT] = ACTIONS(1924),
    [anon_sym_DASH_GT] = ACTIONS(1922),
    [anon_sym_LBRACK] = ACTIONS(1922),
    [anon_sym_RBRACK] = ACTIONS(1924),
    [anon_sym_true] = ACTIONS(1922),
    [anon_sym_false] = ACTIONS(1922),
    [anon_sym_LBRACE] = ACTIONS(1922),
    [anon_sym_SEMI] = ACTIONS(1924),
    [anon_sym_RBRACE] = ACTIONS(1924),
    [anon_sym_constraint] = ACTIONS(1922),
    [anon_sym_val] = ACTIONS(1922),
    [anon_sym_end] = ACTIONS(1922),
    [anon_sym_with] = ACTIONS(1922),
    [anon_sym_COLON_GT] = ACTIONS(1924),
    [anon_sym_object] = ACTIONS(1922),
    [anon_sym_inherit] = ACTIONS(1922),
    [anon_sym_method] = ACTIONS(1922),
    [anon_sym_initializer] = ACTIONS(1922),
    [anon_sym_AMP] = ACTIONS(1922),
    [anon_sym_POUND] = ACTIONS(1922),
    [anon_sym_COLON_COLON] = ACTIONS(1924),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1924),
    [anon_sym_then] = ACTIONS(1922),
    [anon_sym_else] = ACTIONS(1922),
    [anon_sym_do] = ACTIONS(1922),
    [anon_sym_done] = ACTIONS(1922),
    [anon_sym_to] = ACTIONS(1922),
    [anon_sym_downto] = ACTIONS(1922),
    [anon_sym_new] = ACTIONS(1922),
    [anon_sym_LBRACE_LT] = ACTIONS(1924),
    [anon_sym_GT_RBRACE] = ACTIONS(1924),
    [anon_sym_begin] = ACTIONS(1922),
    [sym_ocamlyacc_value] = ACTIONS(1924),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1922),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1924),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1922),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1922),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1924),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1924),
    [aux_sym_number_token1] = ACTIONS(1924),
    [aux_sym_signed_number_token1] = ACTIONS(1922),
    [anon_sym_SQUOTE] = ACTIONS(1924),
    [anon_sym_DQUOTE] = ACTIONS(1924),
    [sym_prefix_operator] = ACTIONS(1924),
    [aux_sym_sign_operator_token1] = ACTIONS(1922),
    [sym_hash_operator] = ACTIONS(1924),
    [anon_sym_lsl] = ACTIONS(1922),
    [anon_sym_lsr] = ACTIONS(1922),
    [anon_sym_asr] = ACTIONS(1922),
    [aux_sym_pow_operator_token1] = ACTIONS(1924),
    [anon_sym_mod] = ACTIONS(1922),
    [anon_sym_land] = ACTIONS(1922),
    [anon_sym_lor] = ACTIONS(1922),
    [anon_sym_lxor] = ACTIONS(1922),
    [aux_sym_mult_operator_token1] = ACTIONS(1922),
    [aux_sym_add_operator_token1] = ACTIONS(1922),
    [sym_concat_operator] = ACTIONS(1924),
    [sym_rel_operator] = ACTIONS(1922),
    [anon_sym_AMP_AMP] = ACTIONS(1922),
    [anon_sym_or] = ACTIONS(1922),
    [anon_sym_PIPE_PIPE] = ACTIONS(1922),
    [sym_assign_operator] = ACTIONS(1924),
    [sym__uppercase_identifier] = ACTIONS(1924),
    [aux_sym_tag_token1] = ACTIONS(1924),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(891)] = {
    [sym_attribute] = STATE(891),
    [sym__lowercase_identifier] = ACTIONS(1728),
    [anon_sym_TILDE] = ACTIONS(1728),
    [anon_sym_QMARK] = ACTIONS(1728),
    [anon_sym_LPAREN] = ACTIONS(1730),
    [anon_sym_RPAREN] = ACTIONS(1730),
    [anon_sym_COMMA] = ACTIONS(1730),
    [anon_sym_PIPE] = ACTIONS(1728),
    [anon_sym_COLON2] = ACTIONS(1728),
    [anon_sym_DOT] = ACTIONS(1730),
    [anon_sym_DASH_GT] = ACTIONS(1728),
    [anon_sym_LBRACK] = ACTIONS(1728),
    [anon_sym_RBRACK] = ACTIONS(1730),
    [anon_sym_true] = ACTIONS(1728),
    [anon_sym_false] = ACTIONS(1728),
    [anon_sym_LBRACE] = ACTIONS(1728),
    [anon_sym_SEMI] = ACTIONS(1730),
    [anon_sym_RBRACE] = ACTIONS(1730),
    [anon_sym_constraint] = ACTIONS(1728),
    [anon_sym_val] = ACTIONS(1728),
    [anon_sym_end] = ACTIONS(1728),
    [anon_sym_with] = ACTIONS(1728),
    [anon_sym_COLON_GT] = ACTIONS(1730),
    [anon_sym_object] = ACTIONS(1728),
    [anon_sym_inherit] = ACTIONS(1728),
    [anon_sym_method] = ACTIONS(1728),
    [anon_sym_initializer] = ACTIONS(1728),
    [anon_sym_AMP] = ACTIONS(1728),
    [anon_sym_POUND] = ACTIONS(1728),
    [anon_sym_COLON_COLON] = ACTIONS(1730),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1730),
    [anon_sym_then] = ACTIONS(1728),
    [anon_sym_else] = ACTIONS(1728),
    [anon_sym_do] = ACTIONS(1728),
    [anon_sym_done] = ACTIONS(1728),
    [anon_sym_to] = ACTIONS(1728),
    [anon_sym_downto] = ACTIONS(1728),
    [anon_sym_new] = ACTIONS(1728),
    [anon_sym_LBRACE_LT] = ACTIONS(1730),
    [anon_sym_GT_RBRACE] = ACTIONS(1730),
    [anon_sym_begin] = ACTIONS(1728),
    [sym_ocamlyacc_value] = ACTIONS(1730),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1728),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1730),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1728),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1728),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1730),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1730),
    [aux_sym_number_token1] = ACTIONS(1730),
    [aux_sym_signed_number_token1] = ACTIONS(1728),
    [anon_sym_SQUOTE] = ACTIONS(1730),
    [anon_sym_DQUOTE] = ACTIONS(1730),
    [sym_prefix_operator] = ACTIONS(1730),
    [aux_sym_sign_operator_token1] = ACTIONS(1728),
    [sym_hash_operator] = ACTIONS(1730),
    [anon_sym_lsl] = ACTIONS(1728),
    [anon_sym_lsr] = ACTIONS(1728),
    [anon_sym_asr] = ACTIONS(1728),
    [aux_sym_pow_operator_token1] = ACTIONS(1730),
    [anon_sym_mod] = ACTIONS(1728),
    [anon_sym_land] = ACTIONS(1728),
    [anon_sym_lor] = ACTIONS(1728),
    [anon_sym_lxor] = ACTIONS(1728),
    [aux_sym_mult_operator_token1] = ACTIONS(1728),
    [aux_sym_add_operator_token1] = ACTIONS(1728),
    [sym_concat_operator] = ACTIONS(1730),
    [sym_rel_operator] = ACTIONS(1728),
    [anon_sym_AMP_AMP] = ACTIONS(1728),
    [anon_sym_or] = ACTIONS(1728),
    [anon_sym_PIPE_PIPE] = ACTIONS(1728),
    [sym_assign_operator] = ACTIONS(1730),
    [sym__uppercase_identifier] = ACTIONS(1730),
    [aux_sym_tag_token1] = ACTIONS(1730),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(892)] = {
    [sym_attribute] = STATE(892),
    [sym__lowercase_identifier] = ACTIONS(1830),
    [anon_sym_TILDE] = ACTIONS(1830),
    [anon_sym_QMARK] = ACTIONS(1830),
    [anon_sym_LPAREN] = ACTIONS(1832),
    [anon_sym_RPAREN] = ACTIONS(1832),
    [anon_sym_COMMA] = ACTIONS(1832),
    [anon_sym_PIPE] = ACTIONS(1830),
    [anon_sym_COLON2] = ACTIONS(1830),
    [anon_sym_DOT] = ACTIONS(1832),
    [anon_sym_DASH_GT] = ACTIONS(1830),
    [anon_sym_LBRACK] = ACTIONS(1830),
    [anon_sym_RBRACK] = ACTIONS(1832),
    [anon_sym_true] = ACTIONS(1830),
    [anon_sym_false] = ACTIONS(1830),
    [anon_sym_LBRACE] = ACTIONS(1830),
    [anon_sym_SEMI] = ACTIONS(1832),
    [anon_sym_RBRACE] = ACTIONS(1832),
    [anon_sym_constraint] = ACTIONS(1830),
    [anon_sym_val] = ACTIONS(1830),
    [anon_sym_end] = ACTIONS(1830),
    [anon_sym_with] = ACTIONS(1830),
    [anon_sym_COLON_GT] = ACTIONS(1832),
    [anon_sym_object] = ACTIONS(1830),
    [anon_sym_inherit] = ACTIONS(1830),
    [anon_sym_method] = ACTIONS(1830),
    [anon_sym_initializer] = ACTIONS(1830),
    [anon_sym_AMP] = ACTIONS(1830),
    [anon_sym_POUND] = ACTIONS(1830),
    [anon_sym_COLON_COLON] = ACTIONS(1832),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1832),
    [anon_sym_then] = ACTIONS(1830),
    [anon_sym_else] = ACTIONS(1830),
    [anon_sym_do] = ACTIONS(1830),
    [anon_sym_done] = ACTIONS(1830),
    [anon_sym_to] = ACTIONS(1830),
    [anon_sym_downto] = ACTIONS(1830),
    [anon_sym_new] = ACTIONS(1830),
    [anon_sym_LBRACE_LT] = ACTIONS(1832),
    [anon_sym_GT_RBRACE] = ACTIONS(1832),
    [anon_sym_begin] = ACTIONS(1830),
    [sym_ocamlyacc_value] = ACTIONS(1832),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1830),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1832),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1830),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1830),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1832),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1832),
    [aux_sym_number_token1] = ACTIONS(1832),
    [aux_sym_signed_number_token1] = ACTIONS(1830),
    [anon_sym_SQUOTE] = ACTIONS(1832),
    [anon_sym_DQUOTE] = ACTIONS(1832),
    [sym_prefix_operator] = ACTIONS(1832),
    [aux_sym_sign_operator_token1] = ACTIONS(1830),
    [sym_hash_operator] = ACTIONS(1832),
    [anon_sym_lsl] = ACTIONS(1830),
    [anon_sym_lsr] = ACTIONS(1830),
    [anon_sym_asr] = ACTIONS(1830),
    [aux_sym_pow_operator_token1] = ACTIONS(1832),
    [anon_sym_mod] = ACTIONS(1830),
    [anon_sym_land] = ACTIONS(1830),
    [anon_sym_lor] = ACTIONS(1830),
    [anon_sym_lxor] = ACTIONS(1830),
    [aux_sym_mult_operator_token1] = ACTIONS(1830),
    [aux_sym_add_operator_token1] = ACTIONS(1830),
    [sym_concat_operator] = ACTIONS(1832),
    [sym_rel_operator] = ACTIONS(1830),
    [anon_sym_AMP_AMP] = ACTIONS(1830),
    [anon_sym_or] = ACTIONS(1830),
    [anon_sym_PIPE_PIPE] = ACTIONS(1830),
    [sym_assign_operator] = ACTIONS(1832),
    [sym__uppercase_identifier] = ACTIONS(1832),
    [aux_sym_tag_token1] = ACTIONS(1832),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(893)] = {
    [sym_attribute] = STATE(893),
    [sym__lowercase_identifier] = ACTIONS(1740),
    [anon_sym_TILDE] = ACTIONS(1740),
    [anon_sym_QMARK] = ACTIONS(1740),
    [anon_sym_LPAREN] = ACTIONS(1742),
    [anon_sym_RPAREN] = ACTIONS(1742),
    [anon_sym_COMMA] = ACTIONS(1742),
    [anon_sym_PIPE] = ACTIONS(1740),
    [anon_sym_COLON2] = ACTIONS(1740),
    [anon_sym_DOT] = ACTIONS(1742),
    [anon_sym_DASH_GT] = ACTIONS(1740),
    [anon_sym_LBRACK] = ACTIONS(1740),
    [anon_sym_RBRACK] = ACTIONS(1742),
    [anon_sym_true] = ACTIONS(1740),
    [anon_sym_false] = ACTIONS(1740),
    [anon_sym_LBRACE] = ACTIONS(1740),
    [anon_sym_SEMI] = ACTIONS(1742),
    [anon_sym_RBRACE] = ACTIONS(1742),
    [anon_sym_constraint] = ACTIONS(1740),
    [anon_sym_val] = ACTIONS(1740),
    [anon_sym_end] = ACTIONS(1740),
    [anon_sym_with] = ACTIONS(1740),
    [anon_sym_COLON_GT] = ACTIONS(1742),
    [anon_sym_object] = ACTIONS(1740),
    [anon_sym_inherit] = ACTIONS(1740),
    [anon_sym_method] = ACTIONS(1740),
    [anon_sym_initializer] = ACTIONS(1740),
    [anon_sym_AMP] = ACTIONS(1740),
    [anon_sym_POUND] = ACTIONS(1740),
    [anon_sym_COLON_COLON] = ACTIONS(1742),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1742),
    [anon_sym_then] = ACTIONS(1740),
    [anon_sym_else] = ACTIONS(1740),
    [anon_sym_do] = ACTIONS(1740),
    [anon_sym_done] = ACTIONS(1740),
    [anon_sym_to] = ACTIONS(1740),
    [anon_sym_downto] = ACTIONS(1740),
    [anon_sym_new] = ACTIONS(1740),
    [anon_sym_LBRACE_LT] = ACTIONS(1742),
    [anon_sym_GT_RBRACE] = ACTIONS(1742),
    [anon_sym_begin] = ACTIONS(1740),
    [sym_ocamlyacc_value] = ACTIONS(1742),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1740),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1742),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1740),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1740),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1742),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1742),
    [aux_sym_number_token1] = ACTIONS(1742),
    [aux_sym_signed_number_token1] = ACTIONS(1740),
    [anon_sym_SQUOTE] = ACTIONS(1742),
    [anon_sym_DQUOTE] = ACTIONS(1742),
    [sym_prefix_operator] = ACTIONS(1742),
    [aux_sym_sign_operator_token1] = ACTIONS(1740),
    [sym_hash_operator] = ACTIONS(1742),
    [anon_sym_lsl] = ACTIONS(1740),
    [anon_sym_lsr] = ACTIONS(1740),
    [anon_sym_asr] = ACTIONS(1740),
    [aux_sym_pow_operator_token1] = ACTIONS(1742),
    [anon_sym_mod] = ACTIONS(1740),
    [anon_sym_land] = ACTIONS(1740),
    [anon_sym_lor] = ACTIONS(1740),
    [anon_sym_lxor] = ACTIONS(1740),
    [aux_sym_mult_operator_token1] = ACTIONS(1740),
    [aux_sym_add_operator_token1] = ACTIONS(1740),
    [sym_concat_operator] = ACTIONS(1742),
    [sym_rel_operator] = ACTIONS(1740),
    [anon_sym_AMP_AMP] = ACTIONS(1740),
    [anon_sym_or] = ACTIONS(1740),
    [anon_sym_PIPE_PIPE] = ACTIONS(1740),
    [sym_assign_operator] = ACTIONS(1742),
    [sym__uppercase_identifier] = ACTIONS(1742),
    [aux_sym_tag_token1] = ACTIONS(1742),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(894)] = {
    [sym_attribute] = STATE(894),
    [sym__lowercase_identifier] = ACTIONS(1874),
    [anon_sym_TILDE] = ACTIONS(1874),
    [anon_sym_QMARK] = ACTIONS(1874),
    [anon_sym_LPAREN] = ACTIONS(1876),
    [anon_sym_RPAREN] = ACTIONS(1876),
    [anon_sym_COMMA] = ACTIONS(1876),
    [anon_sym_PIPE] = ACTIONS(1874),
    [anon_sym_COLON2] = ACTIONS(1874),
    [anon_sym_DOT] = ACTIONS(1876),
    [anon_sym_DASH_GT] = ACTIONS(1874),
    [anon_sym_LBRACK] = ACTIONS(1874),
    [anon_sym_RBRACK] = ACTIONS(1876),
    [anon_sym_true] = ACTIONS(1874),
    [anon_sym_false] = ACTIONS(1874),
    [anon_sym_LBRACE] = ACTIONS(1874),
    [anon_sym_SEMI] = ACTIONS(1876),
    [anon_sym_RBRACE] = ACTIONS(1876),
    [anon_sym_constraint] = ACTIONS(1874),
    [anon_sym_val] = ACTIONS(1874),
    [anon_sym_end] = ACTIONS(1874),
    [anon_sym_with] = ACTIONS(1874),
    [anon_sym_COLON_GT] = ACTIONS(1876),
    [anon_sym_object] = ACTIONS(1874),
    [anon_sym_inherit] = ACTIONS(1874),
    [anon_sym_method] = ACTIONS(1874),
    [anon_sym_initializer] = ACTIONS(1874),
    [anon_sym_AMP] = ACTIONS(1874),
    [anon_sym_POUND] = ACTIONS(1874),
    [anon_sym_COLON_COLON] = ACTIONS(1876),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1876),
    [anon_sym_then] = ACTIONS(1874),
    [anon_sym_else] = ACTIONS(1874),
    [anon_sym_do] = ACTIONS(1874),
    [anon_sym_done] = ACTIONS(1874),
    [anon_sym_to] = ACTIONS(1874),
    [anon_sym_downto] = ACTIONS(1874),
    [anon_sym_new] = ACTIONS(1874),
    [anon_sym_LBRACE_LT] = ACTIONS(1876),
    [anon_sym_GT_RBRACE] = ACTIONS(1876),
    [anon_sym_begin] = ACTIONS(1874),
    [sym_ocamlyacc_value] = ACTIONS(1876),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1874),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1876),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1874),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1874),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1876),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1876),
    [aux_sym_number_token1] = ACTIONS(1876),
    [aux_sym_signed_number_token1] = ACTIONS(1874),
    [anon_sym_SQUOTE] = ACTIONS(1876),
    [anon_sym_DQUOTE] = ACTIONS(1876),
    [sym_prefix_operator] = ACTIONS(1876),
    [aux_sym_sign_operator_token1] = ACTIONS(1874),
    [sym_hash_operator] = ACTIONS(1876),
    [anon_sym_lsl] = ACTIONS(1874),
    [anon_sym_lsr] = ACTIONS(1874),
    [anon_sym_asr] = ACTIONS(1874),
    [aux_sym_pow_operator_token1] = ACTIONS(1876),
    [anon_sym_mod] = ACTIONS(1874),
    [anon_sym_land] = ACTIONS(1874),
    [anon_sym_lor] = ACTIONS(1874),
    [anon_sym_lxor] = ACTIONS(1874),
    [aux_sym_mult_operator_token1] = ACTIONS(1874),
    [aux_sym_add_operator_token1] = ACTIONS(1874),
    [sym_concat_operator] = ACTIONS(1876),
    [sym_rel_operator] = ACTIONS(1874),
    [anon_sym_AMP_AMP] = ACTIONS(1874),
    [anon_sym_or] = ACTIONS(1874),
    [anon_sym_PIPE_PIPE] = ACTIONS(1874),
    [sym_assign_operator] = ACTIONS(1876),
    [sym__uppercase_identifier] = ACTIONS(1876),
    [aux_sym_tag_token1] = ACTIONS(1876),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(895)] = {
    [sym_attribute] = STATE(895),
    [sym__lowercase_identifier] = ACTIONS(1582),
    [anon_sym_TILDE] = ACTIONS(1582),
    [anon_sym_QMARK] = ACTIONS(1582),
    [anon_sym_LPAREN] = ACTIONS(1584),
    [anon_sym_RPAREN] = ACTIONS(1584),
    [anon_sym_COMMA] = ACTIONS(1584),
    [anon_sym_PIPE] = ACTIONS(1582),
    [anon_sym_COLON2] = ACTIONS(1582),
    [anon_sym_DOT] = ACTIONS(1584),
    [anon_sym_DASH_GT] = ACTIONS(1582),
    [anon_sym_LBRACK] = ACTIONS(1582),
    [anon_sym_RBRACK] = ACTIONS(1584),
    [anon_sym_true] = ACTIONS(1582),
    [anon_sym_false] = ACTIONS(1582),
    [anon_sym_LBRACE] = ACTIONS(1582),
    [anon_sym_SEMI] = ACTIONS(1584),
    [anon_sym_RBRACE] = ACTIONS(1584),
    [anon_sym_constraint] = ACTIONS(1582),
    [anon_sym_val] = ACTIONS(1582),
    [anon_sym_end] = ACTIONS(1582),
    [anon_sym_with] = ACTIONS(1582),
    [anon_sym_COLON_GT] = ACTIONS(1584),
    [anon_sym_object] = ACTIONS(1582),
    [anon_sym_inherit] = ACTIONS(1582),
    [anon_sym_method] = ACTIONS(1582),
    [anon_sym_initializer] = ACTIONS(1582),
    [anon_sym_AMP] = ACTIONS(1582),
    [anon_sym_POUND] = ACTIONS(1582),
    [anon_sym_COLON_COLON] = ACTIONS(1584),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1584),
    [anon_sym_then] = ACTIONS(1582),
    [anon_sym_else] = ACTIONS(1582),
    [anon_sym_do] = ACTIONS(1582),
    [anon_sym_done] = ACTIONS(1582),
    [anon_sym_to] = ACTIONS(1582),
    [anon_sym_downto] = ACTIONS(1582),
    [anon_sym_new] = ACTIONS(1582),
    [anon_sym_LBRACE_LT] = ACTIONS(1584),
    [anon_sym_GT_RBRACE] = ACTIONS(1584),
    [anon_sym_begin] = ACTIONS(1582),
    [sym_ocamlyacc_value] = ACTIONS(1584),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1582),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1584),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1582),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1582),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1584),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1584),
    [aux_sym_number_token1] = ACTIONS(1584),
    [aux_sym_signed_number_token1] = ACTIONS(1582),
    [anon_sym_SQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1584),
    [sym_prefix_operator] = ACTIONS(1584),
    [aux_sym_sign_operator_token1] = ACTIONS(1582),
    [sym_hash_operator] = ACTIONS(1584),
    [anon_sym_lsl] = ACTIONS(1582),
    [anon_sym_lsr] = ACTIONS(1582),
    [anon_sym_asr] = ACTIONS(1582),
    [aux_sym_pow_operator_token1] = ACTIONS(1584),
    [anon_sym_mod] = ACTIONS(1582),
    [anon_sym_land] = ACTIONS(1582),
    [anon_sym_lor] = ACTIONS(1582),
    [anon_sym_lxor] = ACTIONS(1582),
    [aux_sym_mult_operator_token1] = ACTIONS(1582),
    [aux_sym_add_operator_token1] = ACTIONS(1582),
    [sym_concat_operator] = ACTIONS(1584),
    [sym_rel_operator] = ACTIONS(1582),
    [anon_sym_AMP_AMP] = ACTIONS(1582),
    [anon_sym_or] = ACTIONS(1582),
    [anon_sym_PIPE_PIPE] = ACTIONS(1582),
    [sym_assign_operator] = ACTIONS(1584),
    [sym__uppercase_identifier] = ACTIONS(1584),
    [aux_sym_tag_token1] = ACTIONS(1584),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(896)] = {
    [sym_attribute] = STATE(896),
    [sym__lowercase_identifier] = ACTIONS(1758),
    [anon_sym_TILDE] = ACTIONS(1758),
    [anon_sym_QMARK] = ACTIONS(1758),
    [anon_sym_LPAREN] = ACTIONS(1760),
    [anon_sym_RPAREN] = ACTIONS(1760),
    [anon_sym_COMMA] = ACTIONS(1760),
    [anon_sym_PIPE] = ACTIONS(1758),
    [anon_sym_COLON2] = ACTIONS(1758),
    [anon_sym_DOT] = ACTIONS(1760),
    [anon_sym_DASH_GT] = ACTIONS(1758),
    [anon_sym_LBRACK] = ACTIONS(1758),
    [anon_sym_RBRACK] = ACTIONS(1760),
    [anon_sym_true] = ACTIONS(1758),
    [anon_sym_false] = ACTIONS(1758),
    [anon_sym_LBRACE] = ACTIONS(1758),
    [anon_sym_SEMI] = ACTIONS(1760),
    [anon_sym_RBRACE] = ACTIONS(1760),
    [anon_sym_constraint] = ACTIONS(1758),
    [anon_sym_val] = ACTIONS(1758),
    [anon_sym_end] = ACTIONS(1758),
    [anon_sym_with] = ACTIONS(1758),
    [anon_sym_COLON_GT] = ACTIONS(1760),
    [anon_sym_object] = ACTIONS(1758),
    [anon_sym_inherit] = ACTIONS(1758),
    [anon_sym_method] = ACTIONS(1758),
    [anon_sym_initializer] = ACTIONS(1758),
    [anon_sym_AMP] = ACTIONS(1758),
    [anon_sym_POUND] = ACTIONS(1758),
    [anon_sym_COLON_COLON] = ACTIONS(1760),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1760),
    [anon_sym_then] = ACTIONS(1758),
    [anon_sym_else] = ACTIONS(1758),
    [anon_sym_do] = ACTIONS(1758),
    [anon_sym_done] = ACTIONS(1758),
    [anon_sym_to] = ACTIONS(1758),
    [anon_sym_downto] = ACTIONS(1758),
    [anon_sym_new] = ACTIONS(1758),
    [anon_sym_LBRACE_LT] = ACTIONS(1760),
    [anon_sym_GT_RBRACE] = ACTIONS(1760),
    [anon_sym_begin] = ACTIONS(1758),
    [sym_ocamlyacc_value] = ACTIONS(1760),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1758),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1760),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1758),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1758),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1760),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1760),
    [aux_sym_number_token1] = ACTIONS(1760),
    [aux_sym_signed_number_token1] = ACTIONS(1758),
    [anon_sym_SQUOTE] = ACTIONS(1760),
    [anon_sym_DQUOTE] = ACTIONS(1760),
    [sym_prefix_operator] = ACTIONS(1760),
    [aux_sym_sign_operator_token1] = ACTIONS(1758),
    [sym_hash_operator] = ACTIONS(1760),
    [anon_sym_lsl] = ACTIONS(1758),
    [anon_sym_lsr] = ACTIONS(1758),
    [anon_sym_asr] = ACTIONS(1758),
    [aux_sym_pow_operator_token1] = ACTIONS(1760),
    [anon_sym_mod] = ACTIONS(1758),
    [anon_sym_land] = ACTIONS(1758),
    [anon_sym_lor] = ACTIONS(1758),
    [anon_sym_lxor] = ACTIONS(1758),
    [aux_sym_mult_operator_token1] = ACTIONS(1758),
    [aux_sym_add_operator_token1] = ACTIONS(1758),
    [sym_concat_operator] = ACTIONS(1760),
    [sym_rel_operator] = ACTIONS(1758),
    [anon_sym_AMP_AMP] = ACTIONS(1758),
    [anon_sym_or] = ACTIONS(1758),
    [anon_sym_PIPE_PIPE] = ACTIONS(1758),
    [sym_assign_operator] = ACTIONS(1760),
    [sym__uppercase_identifier] = ACTIONS(1760),
    [aux_sym_tag_token1] = ACTIONS(1760),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(897)] = {
    [sym_attribute] = STATE(897),
    [sym__lowercase_identifier] = ACTIONS(1838),
    [anon_sym_TILDE] = ACTIONS(1838),
    [anon_sym_QMARK] = ACTIONS(1838),
    [anon_sym_LPAREN] = ACTIONS(1840),
    [anon_sym_RPAREN] = ACTIONS(1840),
    [anon_sym_COMMA] = ACTIONS(1840),
    [anon_sym_PIPE] = ACTIONS(1838),
    [anon_sym_COLON2] = ACTIONS(1838),
    [anon_sym_DOT] = ACTIONS(1840),
    [anon_sym_DASH_GT] = ACTIONS(1838),
    [anon_sym_LBRACK] = ACTIONS(1838),
    [anon_sym_RBRACK] = ACTIONS(1840),
    [anon_sym_true] = ACTIONS(1838),
    [anon_sym_false] = ACTIONS(1838),
    [anon_sym_LBRACE] = ACTIONS(1838),
    [anon_sym_SEMI] = ACTIONS(1840),
    [anon_sym_RBRACE] = ACTIONS(1840),
    [anon_sym_constraint] = ACTIONS(1838),
    [anon_sym_val] = ACTIONS(1838),
    [anon_sym_end] = ACTIONS(1838),
    [anon_sym_with] = ACTIONS(1838),
    [anon_sym_COLON_GT] = ACTIONS(1840),
    [anon_sym_object] = ACTIONS(1838),
    [anon_sym_inherit] = ACTIONS(1838),
    [anon_sym_method] = ACTIONS(1838),
    [anon_sym_initializer] = ACTIONS(1838),
    [anon_sym_AMP] = ACTIONS(1838),
    [anon_sym_POUND] = ACTIONS(1838),
    [anon_sym_COLON_COLON] = ACTIONS(1840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1840),
    [anon_sym_then] = ACTIONS(1838),
    [anon_sym_else] = ACTIONS(1838),
    [anon_sym_do] = ACTIONS(1838),
    [anon_sym_done] = ACTIONS(1838),
    [anon_sym_to] = ACTIONS(1838),
    [anon_sym_downto] = ACTIONS(1838),
    [anon_sym_new] = ACTIONS(1838),
    [anon_sym_LBRACE_LT] = ACTIONS(1840),
    [anon_sym_GT_RBRACE] = ACTIONS(1840),
    [anon_sym_begin] = ACTIONS(1838),
    [sym_ocamlyacc_value] = ACTIONS(1840),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1838),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1840),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1838),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1838),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1840),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1840),
    [aux_sym_number_token1] = ACTIONS(1840),
    [aux_sym_signed_number_token1] = ACTIONS(1838),
    [anon_sym_SQUOTE] = ACTIONS(1840),
    [anon_sym_DQUOTE] = ACTIONS(1840),
    [sym_prefix_operator] = ACTIONS(1840),
    [aux_sym_sign_operator_token1] = ACTIONS(1838),
    [sym_hash_operator] = ACTIONS(1840),
    [anon_sym_lsl] = ACTIONS(1838),
    [anon_sym_lsr] = ACTIONS(1838),
    [anon_sym_asr] = ACTIONS(1838),
    [aux_sym_pow_operator_token1] = ACTIONS(1840),
    [anon_sym_mod] = ACTIONS(1838),
    [anon_sym_land] = ACTIONS(1838),
    [anon_sym_lor] = ACTIONS(1838),
    [anon_sym_lxor] = ACTIONS(1838),
    [aux_sym_mult_operator_token1] = ACTIONS(1838),
    [aux_sym_add_operator_token1] = ACTIONS(1838),
    [sym_concat_operator] = ACTIONS(1840),
    [sym_rel_operator] = ACTIONS(1838),
    [anon_sym_AMP_AMP] = ACTIONS(1838),
    [anon_sym_or] = ACTIONS(1838),
    [anon_sym_PIPE_PIPE] = ACTIONS(1838),
    [sym_assign_operator] = ACTIONS(1840),
    [sym__uppercase_identifier] = ACTIONS(1840),
    [aux_sym_tag_token1] = ACTIONS(1840),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(898)] = {
    [sym_attribute] = STATE(898),
    [sym__lowercase_identifier] = ACTIONS(1858),
    [anon_sym_TILDE] = ACTIONS(1858),
    [anon_sym_QMARK] = ACTIONS(1858),
    [anon_sym_LPAREN] = ACTIONS(1860),
    [anon_sym_RPAREN] = ACTIONS(1860),
    [anon_sym_COMMA] = ACTIONS(1860),
    [anon_sym_PIPE] = ACTIONS(1858),
    [anon_sym_COLON2] = ACTIONS(1858),
    [anon_sym_DOT] = ACTIONS(1860),
    [anon_sym_DASH_GT] = ACTIONS(1858),
    [anon_sym_LBRACK] = ACTIONS(1858),
    [anon_sym_RBRACK] = ACTIONS(1860),
    [anon_sym_true] = ACTIONS(1858),
    [anon_sym_false] = ACTIONS(1858),
    [anon_sym_LBRACE] = ACTIONS(1858),
    [anon_sym_SEMI] = ACTIONS(1860),
    [anon_sym_RBRACE] = ACTIONS(1860),
    [anon_sym_constraint] = ACTIONS(1858),
    [anon_sym_val] = ACTIONS(1858),
    [anon_sym_end] = ACTIONS(1858),
    [anon_sym_with] = ACTIONS(1858),
    [anon_sym_COLON_GT] = ACTIONS(1860),
    [anon_sym_object] = ACTIONS(1858),
    [anon_sym_inherit] = ACTIONS(1858),
    [anon_sym_method] = ACTIONS(1858),
    [anon_sym_initializer] = ACTIONS(1858),
    [anon_sym_AMP] = ACTIONS(1858),
    [anon_sym_POUND] = ACTIONS(1858),
    [anon_sym_COLON_COLON] = ACTIONS(1860),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1860),
    [anon_sym_then] = ACTIONS(1858),
    [anon_sym_else] = ACTIONS(1858),
    [anon_sym_do] = ACTIONS(1858),
    [anon_sym_done] = ACTIONS(1858),
    [anon_sym_to] = ACTIONS(1858),
    [anon_sym_downto] = ACTIONS(1858),
    [anon_sym_new] = ACTIONS(1858),
    [anon_sym_LBRACE_LT] = ACTIONS(1860),
    [anon_sym_GT_RBRACE] = ACTIONS(1860),
    [anon_sym_begin] = ACTIONS(1858),
    [sym_ocamlyacc_value] = ACTIONS(1860),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1858),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1860),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1858),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1858),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1860),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1860),
    [aux_sym_number_token1] = ACTIONS(1860),
    [aux_sym_signed_number_token1] = ACTIONS(1858),
    [anon_sym_SQUOTE] = ACTIONS(1860),
    [anon_sym_DQUOTE] = ACTIONS(1860),
    [sym_prefix_operator] = ACTIONS(1860),
    [aux_sym_sign_operator_token1] = ACTIONS(1858),
    [sym_hash_operator] = ACTIONS(1860),
    [anon_sym_lsl] = ACTIONS(1858),
    [anon_sym_lsr] = ACTIONS(1858),
    [anon_sym_asr] = ACTIONS(1858),
    [aux_sym_pow_operator_token1] = ACTIONS(1860),
    [anon_sym_mod] = ACTIONS(1858),
    [anon_sym_land] = ACTIONS(1858),
    [anon_sym_lor] = ACTIONS(1858),
    [anon_sym_lxor] = ACTIONS(1858),
    [aux_sym_mult_operator_token1] = ACTIONS(1858),
    [aux_sym_add_operator_token1] = ACTIONS(1858),
    [sym_concat_operator] = ACTIONS(1860),
    [sym_rel_operator] = ACTIONS(1858),
    [anon_sym_AMP_AMP] = ACTIONS(1858),
    [anon_sym_or] = ACTIONS(1858),
    [anon_sym_PIPE_PIPE] = ACTIONS(1858),
    [sym_assign_operator] = ACTIONS(1860),
    [sym__uppercase_identifier] = ACTIONS(1860),
    [aux_sym_tag_token1] = ACTIONS(1860),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(899)] = {
    [sym_attribute] = STATE(899),
    [sym__lowercase_identifier] = ACTIONS(1862),
    [anon_sym_TILDE] = ACTIONS(1862),
    [anon_sym_QMARK] = ACTIONS(1862),
    [anon_sym_LPAREN] = ACTIONS(1864),
    [anon_sym_RPAREN] = ACTIONS(1864),
    [anon_sym_COMMA] = ACTIONS(1864),
    [anon_sym_PIPE] = ACTIONS(1862),
    [anon_sym_COLON2] = ACTIONS(1862),
    [anon_sym_DOT] = ACTIONS(1864),
    [anon_sym_DASH_GT] = ACTIONS(1862),
    [anon_sym_LBRACK] = ACTIONS(1862),
    [anon_sym_RBRACK] = ACTIONS(1864),
    [anon_sym_true] = ACTIONS(1862),
    [anon_sym_false] = ACTIONS(1862),
    [anon_sym_LBRACE] = ACTIONS(1862),
    [anon_sym_SEMI] = ACTIONS(1864),
    [anon_sym_RBRACE] = ACTIONS(1864),
    [anon_sym_constraint] = ACTIONS(1862),
    [anon_sym_val] = ACTIONS(1862),
    [anon_sym_end] = ACTIONS(1862),
    [anon_sym_with] = ACTIONS(1862),
    [anon_sym_COLON_GT] = ACTIONS(1864),
    [anon_sym_object] = ACTIONS(1862),
    [anon_sym_inherit] = ACTIONS(1862),
    [anon_sym_method] = ACTIONS(1862),
    [anon_sym_initializer] = ACTIONS(1862),
    [anon_sym_AMP] = ACTIONS(1862),
    [anon_sym_POUND] = ACTIONS(1862),
    [anon_sym_COLON_COLON] = ACTIONS(1864),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1864),
    [anon_sym_then] = ACTIONS(1862),
    [anon_sym_else] = ACTIONS(1862),
    [anon_sym_do] = ACTIONS(1862),
    [anon_sym_done] = ACTIONS(1862),
    [anon_sym_to] = ACTIONS(1862),
    [anon_sym_downto] = ACTIONS(1862),
    [anon_sym_new] = ACTIONS(1862),
    [anon_sym_LBRACE_LT] = ACTIONS(1864),
    [anon_sym_GT_RBRACE] = ACTIONS(1864),
    [anon_sym_begin] = ACTIONS(1862),
    [sym_ocamlyacc_value] = ACTIONS(1864),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1862),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1864),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1862),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1862),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1864),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1864),
    [aux_sym_number_token1] = ACTIONS(1864),
    [aux_sym_signed_number_token1] = ACTIONS(1862),
    [anon_sym_SQUOTE] = ACTIONS(1864),
    [anon_sym_DQUOTE] = ACTIONS(1864),
    [sym_prefix_operator] = ACTIONS(1864),
    [aux_sym_sign_operator_token1] = ACTIONS(1862),
    [sym_hash_operator] = ACTIONS(1864),
    [anon_sym_lsl] = ACTIONS(1862),
    [anon_sym_lsr] = ACTIONS(1862),
    [anon_sym_asr] = ACTIONS(1862),
    [aux_sym_pow_operator_token1] = ACTIONS(1864),
    [anon_sym_mod] = ACTIONS(1862),
    [anon_sym_land] = ACTIONS(1862),
    [anon_sym_lor] = ACTIONS(1862),
    [anon_sym_lxor] = ACTIONS(1862),
    [aux_sym_mult_operator_token1] = ACTIONS(1862),
    [aux_sym_add_operator_token1] = ACTIONS(1862),
    [sym_concat_operator] = ACTIONS(1864),
    [sym_rel_operator] = ACTIONS(1862),
    [anon_sym_AMP_AMP] = ACTIONS(1862),
    [anon_sym_or] = ACTIONS(1862),
    [anon_sym_PIPE_PIPE] = ACTIONS(1862),
    [sym_assign_operator] = ACTIONS(1864),
    [sym__uppercase_identifier] = ACTIONS(1864),
    [aux_sym_tag_token1] = ACTIONS(1864),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(900)] = {
    [sym_attribute] = STATE(900),
    [sym__lowercase_identifier] = ACTIONS(1870),
    [anon_sym_TILDE] = ACTIONS(1870),
    [anon_sym_QMARK] = ACTIONS(1870),
    [anon_sym_LPAREN] = ACTIONS(1872),
    [anon_sym_RPAREN] = ACTIONS(1872),
    [anon_sym_COMMA] = ACTIONS(1872),
    [anon_sym_PIPE] = ACTIONS(1870),
    [anon_sym_COLON2] = ACTIONS(1870),
    [anon_sym_DOT] = ACTIONS(1872),
    [anon_sym_DASH_GT] = ACTIONS(1870),
    [anon_sym_LBRACK] = ACTIONS(1870),
    [anon_sym_RBRACK] = ACTIONS(1872),
    [anon_sym_true] = ACTIONS(1870),
    [anon_sym_false] = ACTIONS(1870),
    [anon_sym_LBRACE] = ACTIONS(1870),
    [anon_sym_SEMI] = ACTIONS(1872),
    [anon_sym_RBRACE] = ACTIONS(1872),
    [anon_sym_constraint] = ACTIONS(1870),
    [anon_sym_val] = ACTIONS(1870),
    [anon_sym_end] = ACTIONS(1870),
    [anon_sym_with] = ACTIONS(1870),
    [anon_sym_COLON_GT] = ACTIONS(1872),
    [anon_sym_object] = ACTIONS(1870),
    [anon_sym_inherit] = ACTIONS(1870),
    [anon_sym_method] = ACTIONS(1870),
    [anon_sym_initializer] = ACTIONS(1870),
    [anon_sym_AMP] = ACTIONS(1870),
    [anon_sym_POUND] = ACTIONS(1870),
    [anon_sym_COLON_COLON] = ACTIONS(1872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1872),
    [anon_sym_then] = ACTIONS(1870),
    [anon_sym_else] = ACTIONS(1870),
    [anon_sym_do] = ACTIONS(1870),
    [anon_sym_done] = ACTIONS(1870),
    [anon_sym_to] = ACTIONS(1870),
    [anon_sym_downto] = ACTIONS(1870),
    [anon_sym_new] = ACTIONS(1870),
    [anon_sym_LBRACE_LT] = ACTIONS(1872),
    [anon_sym_GT_RBRACE] = ACTIONS(1872),
    [anon_sym_begin] = ACTIONS(1870),
    [sym_ocamlyacc_value] = ACTIONS(1872),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1870),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1872),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1870),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1870),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1872),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1872),
    [aux_sym_number_token1] = ACTIONS(1872),
    [aux_sym_signed_number_token1] = ACTIONS(1870),
    [anon_sym_SQUOTE] = ACTIONS(1872),
    [anon_sym_DQUOTE] = ACTIONS(1872),
    [sym_prefix_operator] = ACTIONS(1872),
    [aux_sym_sign_operator_token1] = ACTIONS(1870),
    [sym_hash_operator] = ACTIONS(1872),
    [anon_sym_lsl] = ACTIONS(1870),
    [anon_sym_lsr] = ACTIONS(1870),
    [anon_sym_asr] = ACTIONS(1870),
    [aux_sym_pow_operator_token1] = ACTIONS(1872),
    [anon_sym_mod] = ACTIONS(1870),
    [anon_sym_land] = ACTIONS(1870),
    [anon_sym_lor] = ACTIONS(1870),
    [anon_sym_lxor] = ACTIONS(1870),
    [aux_sym_mult_operator_token1] = ACTIONS(1870),
    [aux_sym_add_operator_token1] = ACTIONS(1870),
    [sym_concat_operator] = ACTIONS(1872),
    [sym_rel_operator] = ACTIONS(1870),
    [anon_sym_AMP_AMP] = ACTIONS(1870),
    [anon_sym_or] = ACTIONS(1870),
    [anon_sym_PIPE_PIPE] = ACTIONS(1870),
    [sym_assign_operator] = ACTIONS(1872),
    [sym__uppercase_identifier] = ACTIONS(1872),
    [aux_sym_tag_token1] = ACTIONS(1872),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(901)] = {
    [sym_attribute] = STATE(901),
    [sym__lowercase_identifier] = ACTIONS(1642),
    [anon_sym_TILDE] = ACTIONS(1642),
    [anon_sym_QMARK] = ACTIONS(1642),
    [anon_sym_LPAREN] = ACTIONS(1644),
    [anon_sym_RPAREN] = ACTIONS(1644),
    [anon_sym_COMMA] = ACTIONS(1644),
    [anon_sym_PIPE] = ACTIONS(1642),
    [anon_sym_COLON2] = ACTIONS(1642),
    [anon_sym_DOT] = ACTIONS(1644),
    [anon_sym_DASH_GT] = ACTIONS(1642),
    [anon_sym_LBRACK] = ACTIONS(1642),
    [anon_sym_RBRACK] = ACTIONS(1644),
    [anon_sym_true] = ACTIONS(1642),
    [anon_sym_false] = ACTIONS(1642),
    [anon_sym_LBRACE] = ACTIONS(1642),
    [anon_sym_SEMI] = ACTIONS(1644),
    [anon_sym_RBRACE] = ACTIONS(1644),
    [anon_sym_constraint] = ACTIONS(1642),
    [anon_sym_val] = ACTIONS(1642),
    [anon_sym_end] = ACTIONS(1642),
    [anon_sym_with] = ACTIONS(1642),
    [anon_sym_COLON_GT] = ACTIONS(1644),
    [anon_sym_object] = ACTIONS(1642),
    [anon_sym_inherit] = ACTIONS(1642),
    [anon_sym_method] = ACTIONS(1642),
    [anon_sym_initializer] = ACTIONS(1642),
    [anon_sym_AMP] = ACTIONS(1642),
    [anon_sym_POUND] = ACTIONS(1642),
    [anon_sym_COLON_COLON] = ACTIONS(1644),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1644),
    [anon_sym_then] = ACTIONS(1642),
    [anon_sym_else] = ACTIONS(1642),
    [anon_sym_do] = ACTIONS(1642),
    [anon_sym_done] = ACTIONS(1642),
    [anon_sym_to] = ACTIONS(1642),
    [anon_sym_downto] = ACTIONS(1642),
    [anon_sym_new] = ACTIONS(1642),
    [anon_sym_LBRACE_LT] = ACTIONS(1644),
    [anon_sym_GT_RBRACE] = ACTIONS(1644),
    [anon_sym_begin] = ACTIONS(1642),
    [sym_ocamlyacc_value] = ACTIONS(1644),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1642),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1644),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1642),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1642),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1644),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1644),
    [aux_sym_number_token1] = ACTIONS(1644),
    [aux_sym_signed_number_token1] = ACTIONS(1642),
    [anon_sym_SQUOTE] = ACTIONS(1644),
    [anon_sym_DQUOTE] = ACTIONS(1644),
    [sym_prefix_operator] = ACTIONS(1644),
    [aux_sym_sign_operator_token1] = ACTIONS(1642),
    [sym_hash_operator] = ACTIONS(1644),
    [anon_sym_lsl] = ACTIONS(1642),
    [anon_sym_lsr] = ACTIONS(1642),
    [anon_sym_asr] = ACTIONS(1642),
    [aux_sym_pow_operator_token1] = ACTIONS(1644),
    [anon_sym_mod] = ACTIONS(1642),
    [anon_sym_land] = ACTIONS(1642),
    [anon_sym_lor] = ACTIONS(1642),
    [anon_sym_lxor] = ACTIONS(1642),
    [aux_sym_mult_operator_token1] = ACTIONS(1642),
    [aux_sym_add_operator_token1] = ACTIONS(1642),
    [sym_concat_operator] = ACTIONS(1644),
    [sym_rel_operator] = ACTIONS(1642),
    [anon_sym_AMP_AMP] = ACTIONS(1642),
    [anon_sym_or] = ACTIONS(1642),
    [anon_sym_PIPE_PIPE] = ACTIONS(1642),
    [sym_assign_operator] = ACTIONS(1644),
    [sym__uppercase_identifier] = ACTIONS(1644),
    [aux_sym_tag_token1] = ACTIONS(1644),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(902)] = {
    [sym_attribute] = STATE(902),
    [sym__lowercase_identifier] = ACTIONS(1878),
    [anon_sym_TILDE] = ACTIONS(1878),
    [anon_sym_QMARK] = ACTIONS(1878),
    [anon_sym_LPAREN] = ACTIONS(1880),
    [anon_sym_RPAREN] = ACTIONS(1880),
    [anon_sym_COMMA] = ACTIONS(1880),
    [anon_sym_PIPE] = ACTIONS(1878),
    [anon_sym_COLON2] = ACTIONS(1878),
    [anon_sym_DOT] = ACTIONS(1880),
    [anon_sym_DASH_GT] = ACTIONS(1878),
    [anon_sym_LBRACK] = ACTIONS(1878),
    [anon_sym_RBRACK] = ACTIONS(1880),
    [anon_sym_true] = ACTIONS(1878),
    [anon_sym_false] = ACTIONS(1878),
    [anon_sym_LBRACE] = ACTIONS(1878),
    [anon_sym_SEMI] = ACTIONS(1880),
    [anon_sym_RBRACE] = ACTIONS(1880),
    [anon_sym_constraint] = ACTIONS(1878),
    [anon_sym_val] = ACTIONS(1878),
    [anon_sym_end] = ACTIONS(1878),
    [anon_sym_with] = ACTIONS(1878),
    [anon_sym_COLON_GT] = ACTIONS(1880),
    [anon_sym_object] = ACTIONS(1878),
    [anon_sym_inherit] = ACTIONS(1878),
    [anon_sym_method] = ACTIONS(1878),
    [anon_sym_initializer] = ACTIONS(1878),
    [anon_sym_AMP] = ACTIONS(1878),
    [anon_sym_POUND] = ACTIONS(1878),
    [anon_sym_COLON_COLON] = ACTIONS(1880),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1880),
    [anon_sym_then] = ACTIONS(1878),
    [anon_sym_else] = ACTIONS(1878),
    [anon_sym_do] = ACTIONS(1878),
    [anon_sym_done] = ACTIONS(1878),
    [anon_sym_to] = ACTIONS(1878),
    [anon_sym_downto] = ACTIONS(1878),
    [anon_sym_new] = ACTIONS(1878),
    [anon_sym_LBRACE_LT] = ACTIONS(1880),
    [anon_sym_GT_RBRACE] = ACTIONS(1880),
    [anon_sym_begin] = ACTIONS(1878),
    [sym_ocamlyacc_value] = ACTIONS(1880),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1878),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1880),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1878),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1878),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1880),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1880),
    [aux_sym_number_token1] = ACTIONS(1880),
    [aux_sym_signed_number_token1] = ACTIONS(1878),
    [anon_sym_SQUOTE] = ACTIONS(1880),
    [anon_sym_DQUOTE] = ACTIONS(1880),
    [sym_prefix_operator] = ACTIONS(1880),
    [aux_sym_sign_operator_token1] = ACTIONS(1878),
    [sym_hash_operator] = ACTIONS(1880),
    [anon_sym_lsl] = ACTIONS(1878),
    [anon_sym_lsr] = ACTIONS(1878),
    [anon_sym_asr] = ACTIONS(1878),
    [aux_sym_pow_operator_token1] = ACTIONS(1880),
    [anon_sym_mod] = ACTIONS(1878),
    [anon_sym_land] = ACTIONS(1878),
    [anon_sym_lor] = ACTIONS(1878),
    [anon_sym_lxor] = ACTIONS(1878),
    [aux_sym_mult_operator_token1] = ACTIONS(1878),
    [aux_sym_add_operator_token1] = ACTIONS(1878),
    [sym_concat_operator] = ACTIONS(1880),
    [sym_rel_operator] = ACTIONS(1878),
    [anon_sym_AMP_AMP] = ACTIONS(1878),
    [anon_sym_or] = ACTIONS(1878),
    [anon_sym_PIPE_PIPE] = ACTIONS(1878),
    [sym_assign_operator] = ACTIONS(1880),
    [sym__uppercase_identifier] = ACTIONS(1880),
    [aux_sym_tag_token1] = ACTIONS(1880),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(903)] = {
    [sym_attribute] = STATE(903),
    [sym__lowercase_identifier] = ACTIONS(1882),
    [anon_sym_TILDE] = ACTIONS(1882),
    [anon_sym_QMARK] = ACTIONS(1882),
    [anon_sym_LPAREN] = ACTIONS(1884),
    [anon_sym_RPAREN] = ACTIONS(1884),
    [anon_sym_COMMA] = ACTIONS(1884),
    [anon_sym_PIPE] = ACTIONS(1882),
    [anon_sym_COLON2] = ACTIONS(1882),
    [anon_sym_DOT] = ACTIONS(1884),
    [anon_sym_DASH_GT] = ACTIONS(1882),
    [anon_sym_LBRACK] = ACTIONS(1882),
    [anon_sym_RBRACK] = ACTIONS(1884),
    [anon_sym_true] = ACTIONS(1882),
    [anon_sym_false] = ACTIONS(1882),
    [anon_sym_LBRACE] = ACTIONS(1882),
    [anon_sym_SEMI] = ACTIONS(1884),
    [anon_sym_RBRACE] = ACTIONS(1884),
    [anon_sym_constraint] = ACTIONS(1882),
    [anon_sym_val] = ACTIONS(1882),
    [anon_sym_end] = ACTIONS(1882),
    [anon_sym_with] = ACTIONS(1882),
    [anon_sym_COLON_GT] = ACTIONS(1884),
    [anon_sym_object] = ACTIONS(1882),
    [anon_sym_inherit] = ACTIONS(1882),
    [anon_sym_method] = ACTIONS(1882),
    [anon_sym_initializer] = ACTIONS(1882),
    [anon_sym_AMP] = ACTIONS(1882),
    [anon_sym_POUND] = ACTIONS(1882),
    [anon_sym_COLON_COLON] = ACTIONS(1884),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1884),
    [anon_sym_then] = ACTIONS(1882),
    [anon_sym_else] = ACTIONS(1882),
    [anon_sym_do] = ACTIONS(1882),
    [anon_sym_done] = ACTIONS(1882),
    [anon_sym_to] = ACTIONS(1882),
    [anon_sym_downto] = ACTIONS(1882),
    [anon_sym_new] = ACTIONS(1882),
    [anon_sym_LBRACE_LT] = ACTIONS(1884),
    [anon_sym_GT_RBRACE] = ACTIONS(1884),
    [anon_sym_begin] = ACTIONS(1882),
    [sym_ocamlyacc_value] = ACTIONS(1884),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1882),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1884),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1882),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1882),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1884),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1884),
    [aux_sym_number_token1] = ACTIONS(1884),
    [aux_sym_signed_number_token1] = ACTIONS(1882),
    [anon_sym_SQUOTE] = ACTIONS(1884),
    [anon_sym_DQUOTE] = ACTIONS(1884),
    [sym_prefix_operator] = ACTIONS(1884),
    [aux_sym_sign_operator_token1] = ACTIONS(1882),
    [sym_hash_operator] = ACTIONS(1884),
    [anon_sym_lsl] = ACTIONS(1882),
    [anon_sym_lsr] = ACTIONS(1882),
    [anon_sym_asr] = ACTIONS(1882),
    [aux_sym_pow_operator_token1] = ACTIONS(1884),
    [anon_sym_mod] = ACTIONS(1882),
    [anon_sym_land] = ACTIONS(1882),
    [anon_sym_lor] = ACTIONS(1882),
    [anon_sym_lxor] = ACTIONS(1882),
    [aux_sym_mult_operator_token1] = ACTIONS(1882),
    [aux_sym_add_operator_token1] = ACTIONS(1882),
    [sym_concat_operator] = ACTIONS(1884),
    [sym_rel_operator] = ACTIONS(1882),
    [anon_sym_AMP_AMP] = ACTIONS(1882),
    [anon_sym_or] = ACTIONS(1882),
    [anon_sym_PIPE_PIPE] = ACTIONS(1882),
    [sym_assign_operator] = ACTIONS(1884),
    [sym__uppercase_identifier] = ACTIONS(1884),
    [aux_sym_tag_token1] = ACTIONS(1884),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(904)] = {
    [sym_attribute] = STATE(904),
    [sym__lowercase_identifier] = ACTIONS(1834),
    [anon_sym_TILDE] = ACTIONS(1834),
    [anon_sym_QMARK] = ACTIONS(1834),
    [anon_sym_LPAREN] = ACTIONS(1836),
    [anon_sym_RPAREN] = ACTIONS(1836),
    [anon_sym_COMMA] = ACTIONS(1836),
    [anon_sym_PIPE] = ACTIONS(1834),
    [anon_sym_COLON2] = ACTIONS(1834),
    [anon_sym_DOT] = ACTIONS(1836),
    [anon_sym_DASH_GT] = ACTIONS(1834),
    [anon_sym_LBRACK] = ACTIONS(1834),
    [anon_sym_RBRACK] = ACTIONS(1836),
    [anon_sym_true] = ACTIONS(1834),
    [anon_sym_false] = ACTIONS(1834),
    [anon_sym_LBRACE] = ACTIONS(1834),
    [anon_sym_SEMI] = ACTIONS(1836),
    [anon_sym_RBRACE] = ACTIONS(1836),
    [anon_sym_constraint] = ACTIONS(1834),
    [anon_sym_val] = ACTIONS(1834),
    [anon_sym_end] = ACTIONS(1834),
    [anon_sym_with] = ACTIONS(1834),
    [anon_sym_COLON_GT] = ACTIONS(1836),
    [anon_sym_object] = ACTIONS(1834),
    [anon_sym_inherit] = ACTIONS(1834),
    [anon_sym_method] = ACTIONS(1834),
    [anon_sym_initializer] = ACTIONS(1834),
    [anon_sym_AMP] = ACTIONS(1834),
    [anon_sym_POUND] = ACTIONS(1834),
    [anon_sym_COLON_COLON] = ACTIONS(1836),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1836),
    [anon_sym_then] = ACTIONS(1834),
    [anon_sym_else] = ACTIONS(1834),
    [anon_sym_do] = ACTIONS(1834),
    [anon_sym_done] = ACTIONS(1834),
    [anon_sym_to] = ACTIONS(1834),
    [anon_sym_downto] = ACTIONS(1834),
    [anon_sym_new] = ACTIONS(1834),
    [anon_sym_LBRACE_LT] = ACTIONS(1836),
    [anon_sym_GT_RBRACE] = ACTIONS(1836),
    [anon_sym_begin] = ACTIONS(1834),
    [sym_ocamlyacc_value] = ACTIONS(1836),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1834),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1836),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1834),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1834),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1836),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1836),
    [aux_sym_number_token1] = ACTIONS(1836),
    [aux_sym_signed_number_token1] = ACTIONS(1834),
    [anon_sym_SQUOTE] = ACTIONS(1836),
    [anon_sym_DQUOTE] = ACTIONS(1836),
    [sym_prefix_operator] = ACTIONS(1836),
    [aux_sym_sign_operator_token1] = ACTIONS(1834),
    [sym_hash_operator] = ACTIONS(1836),
    [anon_sym_lsl] = ACTIONS(1834),
    [anon_sym_lsr] = ACTIONS(1834),
    [anon_sym_asr] = ACTIONS(1834),
    [aux_sym_pow_operator_token1] = ACTIONS(1836),
    [anon_sym_mod] = ACTIONS(1834),
    [anon_sym_land] = ACTIONS(1834),
    [anon_sym_lor] = ACTIONS(1834),
    [anon_sym_lxor] = ACTIONS(1834),
    [aux_sym_mult_operator_token1] = ACTIONS(1834),
    [aux_sym_add_operator_token1] = ACTIONS(1834),
    [sym_concat_operator] = ACTIONS(1836),
    [sym_rel_operator] = ACTIONS(1834),
    [anon_sym_AMP_AMP] = ACTIONS(1834),
    [anon_sym_or] = ACTIONS(1834),
    [anon_sym_PIPE_PIPE] = ACTIONS(1834),
    [sym_assign_operator] = ACTIONS(1836),
    [sym__uppercase_identifier] = ACTIONS(1836),
    [aux_sym_tag_token1] = ACTIONS(1836),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(905)] = {
    [sym_attribute] = STATE(905),
    [sym__lowercase_identifier] = ACTIONS(1890),
    [anon_sym_TILDE] = ACTIONS(1890),
    [anon_sym_QMARK] = ACTIONS(1890),
    [anon_sym_LPAREN] = ACTIONS(1892),
    [anon_sym_RPAREN] = ACTIONS(1892),
    [anon_sym_COMMA] = ACTIONS(1892),
    [anon_sym_PIPE] = ACTIONS(1890),
    [anon_sym_COLON2] = ACTIONS(1890),
    [anon_sym_DOT] = ACTIONS(1892),
    [anon_sym_DASH_GT] = ACTIONS(1890),
    [anon_sym_LBRACK] = ACTIONS(1890),
    [anon_sym_RBRACK] = ACTIONS(1892),
    [anon_sym_true] = ACTIONS(1890),
    [anon_sym_false] = ACTIONS(1890),
    [anon_sym_LBRACE] = ACTIONS(1890),
    [anon_sym_SEMI] = ACTIONS(1892),
    [anon_sym_RBRACE] = ACTIONS(1892),
    [anon_sym_constraint] = ACTIONS(1890),
    [anon_sym_val] = ACTIONS(1890),
    [anon_sym_end] = ACTIONS(1890),
    [anon_sym_with] = ACTIONS(1890),
    [anon_sym_COLON_GT] = ACTIONS(1892),
    [anon_sym_object] = ACTIONS(1890),
    [anon_sym_inherit] = ACTIONS(1890),
    [anon_sym_method] = ACTIONS(1890),
    [anon_sym_initializer] = ACTIONS(1890),
    [anon_sym_AMP] = ACTIONS(1890),
    [anon_sym_POUND] = ACTIONS(1890),
    [anon_sym_COLON_COLON] = ACTIONS(1892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1892),
    [anon_sym_then] = ACTIONS(1890),
    [anon_sym_else] = ACTIONS(1890),
    [anon_sym_do] = ACTIONS(1890),
    [anon_sym_done] = ACTIONS(1890),
    [anon_sym_to] = ACTIONS(1890),
    [anon_sym_downto] = ACTIONS(1890),
    [anon_sym_new] = ACTIONS(1890),
    [anon_sym_LBRACE_LT] = ACTIONS(1892),
    [anon_sym_GT_RBRACE] = ACTIONS(1892),
    [anon_sym_begin] = ACTIONS(1890),
    [sym_ocamlyacc_value] = ACTIONS(1892),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1890),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1892),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1890),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1890),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1892),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1892),
    [aux_sym_number_token1] = ACTIONS(1892),
    [aux_sym_signed_number_token1] = ACTIONS(1890),
    [anon_sym_SQUOTE] = ACTIONS(1892),
    [anon_sym_DQUOTE] = ACTIONS(1892),
    [sym_prefix_operator] = ACTIONS(1892),
    [aux_sym_sign_operator_token1] = ACTIONS(1890),
    [sym_hash_operator] = ACTIONS(1892),
    [anon_sym_lsl] = ACTIONS(1890),
    [anon_sym_lsr] = ACTIONS(1890),
    [anon_sym_asr] = ACTIONS(1890),
    [aux_sym_pow_operator_token1] = ACTIONS(1892),
    [anon_sym_mod] = ACTIONS(1890),
    [anon_sym_land] = ACTIONS(1890),
    [anon_sym_lor] = ACTIONS(1890),
    [anon_sym_lxor] = ACTIONS(1890),
    [aux_sym_mult_operator_token1] = ACTIONS(1890),
    [aux_sym_add_operator_token1] = ACTIONS(1890),
    [sym_concat_operator] = ACTIONS(1892),
    [sym_rel_operator] = ACTIONS(1890),
    [anon_sym_AMP_AMP] = ACTIONS(1890),
    [anon_sym_or] = ACTIONS(1890),
    [anon_sym_PIPE_PIPE] = ACTIONS(1890),
    [sym_assign_operator] = ACTIONS(1892),
    [sym__uppercase_identifier] = ACTIONS(1892),
    [aux_sym_tag_token1] = ACTIONS(1892),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(906)] = {
    [sym_attribute] = STATE(906),
    [sym__lowercase_identifier] = ACTIONS(1894),
    [anon_sym_TILDE] = ACTIONS(1894),
    [anon_sym_QMARK] = ACTIONS(1894),
    [anon_sym_LPAREN] = ACTIONS(1896),
    [anon_sym_RPAREN] = ACTIONS(1896),
    [anon_sym_COMMA] = ACTIONS(1896),
    [anon_sym_PIPE] = ACTIONS(1894),
    [anon_sym_COLON2] = ACTIONS(1894),
    [anon_sym_DOT] = ACTIONS(755),
    [anon_sym_DASH_GT] = ACTIONS(1894),
    [anon_sym_LBRACK] = ACTIONS(1894),
    [anon_sym_RBRACK] = ACTIONS(1896),
    [anon_sym_true] = ACTIONS(1894),
    [anon_sym_false] = ACTIONS(1894),
    [anon_sym_LBRACE] = ACTIONS(1894),
    [anon_sym_SEMI] = ACTIONS(1896),
    [anon_sym_RBRACE] = ACTIONS(1896),
    [anon_sym_constraint] = ACTIONS(1894),
    [anon_sym_val] = ACTIONS(1894),
    [anon_sym_end] = ACTIONS(1894),
    [anon_sym_with] = ACTIONS(1894),
    [anon_sym_COLON_GT] = ACTIONS(1896),
    [anon_sym_object] = ACTIONS(1894),
    [anon_sym_inherit] = ACTIONS(1894),
    [anon_sym_method] = ACTIONS(1894),
    [anon_sym_initializer] = ACTIONS(1894),
    [anon_sym_AMP] = ACTIONS(1894),
    [anon_sym_POUND] = ACTIONS(1894),
    [anon_sym_COLON_COLON] = ACTIONS(1896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1896),
    [anon_sym_then] = ACTIONS(1894),
    [anon_sym_else] = ACTIONS(1894),
    [anon_sym_do] = ACTIONS(1894),
    [anon_sym_done] = ACTIONS(1894),
    [anon_sym_to] = ACTIONS(1894),
    [anon_sym_downto] = ACTIONS(1894),
    [anon_sym_new] = ACTIONS(1894),
    [anon_sym_LBRACE_LT] = ACTIONS(1896),
    [anon_sym_GT_RBRACE] = ACTIONS(1896),
    [anon_sym_begin] = ACTIONS(1894),
    [sym_ocamlyacc_value] = ACTIONS(1896),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1894),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1896),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1894),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1894),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1896),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1896),
    [aux_sym_number_token1] = ACTIONS(1896),
    [aux_sym_signed_number_token1] = ACTIONS(1894),
    [anon_sym_SQUOTE] = ACTIONS(1896),
    [anon_sym_DQUOTE] = ACTIONS(1896),
    [sym_prefix_operator] = ACTIONS(1896),
    [aux_sym_sign_operator_token1] = ACTIONS(1894),
    [sym_hash_operator] = ACTIONS(1896),
    [anon_sym_lsl] = ACTIONS(1894),
    [anon_sym_lsr] = ACTIONS(1894),
    [anon_sym_asr] = ACTIONS(1894),
    [aux_sym_pow_operator_token1] = ACTIONS(1896),
    [anon_sym_mod] = ACTIONS(1894),
    [anon_sym_land] = ACTIONS(1894),
    [anon_sym_lor] = ACTIONS(1894),
    [anon_sym_lxor] = ACTIONS(1894),
    [aux_sym_mult_operator_token1] = ACTIONS(1894),
    [aux_sym_add_operator_token1] = ACTIONS(1894),
    [sym_concat_operator] = ACTIONS(1896),
    [sym_rel_operator] = ACTIONS(1894),
    [anon_sym_AMP_AMP] = ACTIONS(1894),
    [anon_sym_or] = ACTIONS(1894),
    [anon_sym_PIPE_PIPE] = ACTIONS(1894),
    [sym_assign_operator] = ACTIONS(1896),
    [sym__uppercase_identifier] = ACTIONS(1896),
    [aux_sym_tag_token1] = ACTIONS(1896),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(907)] = {
    [sym_attribute] = STATE(907),
    [sym__lowercase_identifier] = ACTIONS(1910),
    [anon_sym_TILDE] = ACTIONS(1910),
    [anon_sym_QMARK] = ACTIONS(1910),
    [anon_sym_LPAREN] = ACTIONS(1912),
    [anon_sym_RPAREN] = ACTIONS(1912),
    [anon_sym_COMMA] = ACTIONS(1912),
    [anon_sym_PIPE] = ACTIONS(1910),
    [anon_sym_COLON2] = ACTIONS(1910),
    [anon_sym_DOT] = ACTIONS(1912),
    [anon_sym_DASH_GT] = ACTIONS(1910),
    [anon_sym_LBRACK] = ACTIONS(1910),
    [anon_sym_RBRACK] = ACTIONS(1912),
    [anon_sym_true] = ACTIONS(1910),
    [anon_sym_false] = ACTIONS(1910),
    [anon_sym_LBRACE] = ACTIONS(1910),
    [anon_sym_SEMI] = ACTIONS(1912),
    [anon_sym_RBRACE] = ACTIONS(1912),
    [anon_sym_constraint] = ACTIONS(1910),
    [anon_sym_val] = ACTIONS(1910),
    [anon_sym_end] = ACTIONS(1910),
    [anon_sym_with] = ACTIONS(1910),
    [anon_sym_COLON_GT] = ACTIONS(1912),
    [anon_sym_object] = ACTIONS(1910),
    [anon_sym_inherit] = ACTIONS(1910),
    [anon_sym_method] = ACTIONS(1910),
    [anon_sym_initializer] = ACTIONS(1910),
    [anon_sym_AMP] = ACTIONS(1910),
    [anon_sym_POUND] = ACTIONS(1910),
    [anon_sym_COLON_COLON] = ACTIONS(1912),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1912),
    [anon_sym_then] = ACTIONS(1910),
    [anon_sym_else] = ACTIONS(1910),
    [anon_sym_do] = ACTIONS(1910),
    [anon_sym_done] = ACTIONS(1910),
    [anon_sym_to] = ACTIONS(1910),
    [anon_sym_downto] = ACTIONS(1910),
    [anon_sym_new] = ACTIONS(1910),
    [anon_sym_LBRACE_LT] = ACTIONS(1912),
    [anon_sym_GT_RBRACE] = ACTIONS(1912),
    [anon_sym_begin] = ACTIONS(1910),
    [sym_ocamlyacc_value] = ACTIONS(1912),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1910),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1912),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1910),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1910),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1912),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1912),
    [aux_sym_number_token1] = ACTIONS(1912),
    [aux_sym_signed_number_token1] = ACTIONS(1910),
    [anon_sym_SQUOTE] = ACTIONS(1912),
    [anon_sym_DQUOTE] = ACTIONS(1912),
    [sym_prefix_operator] = ACTIONS(1912),
    [aux_sym_sign_operator_token1] = ACTIONS(1910),
    [sym_hash_operator] = ACTIONS(1912),
    [anon_sym_lsl] = ACTIONS(1910),
    [anon_sym_lsr] = ACTIONS(1910),
    [anon_sym_asr] = ACTIONS(1910),
    [aux_sym_pow_operator_token1] = ACTIONS(1912),
    [anon_sym_mod] = ACTIONS(1910),
    [anon_sym_land] = ACTIONS(1910),
    [anon_sym_lor] = ACTIONS(1910),
    [anon_sym_lxor] = ACTIONS(1910),
    [aux_sym_mult_operator_token1] = ACTIONS(1910),
    [aux_sym_add_operator_token1] = ACTIONS(1910),
    [sym_concat_operator] = ACTIONS(1912),
    [sym_rel_operator] = ACTIONS(1910),
    [anon_sym_AMP_AMP] = ACTIONS(1910),
    [anon_sym_or] = ACTIONS(1910),
    [anon_sym_PIPE_PIPE] = ACTIONS(1910),
    [sym_assign_operator] = ACTIONS(1912),
    [sym__uppercase_identifier] = ACTIONS(1912),
    [aux_sym_tag_token1] = ACTIONS(1912),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(908)] = {
    [sym_attribute] = STATE(908),
    [sym__lowercase_identifier] = ACTIONS(1646),
    [anon_sym_TILDE] = ACTIONS(1646),
    [anon_sym_QMARK] = ACTIONS(1646),
    [anon_sym_LPAREN] = ACTIONS(1648),
    [anon_sym_RPAREN] = ACTIONS(1648),
    [anon_sym_COMMA] = ACTIONS(1648),
    [anon_sym_PIPE] = ACTIONS(1646),
    [anon_sym_COLON2] = ACTIONS(1646),
    [anon_sym_DOT] = ACTIONS(1648),
    [anon_sym_DASH_GT] = ACTIONS(1646),
    [anon_sym_LBRACK] = ACTIONS(1646),
    [anon_sym_RBRACK] = ACTIONS(1648),
    [anon_sym_true] = ACTIONS(1646),
    [anon_sym_false] = ACTIONS(1646),
    [anon_sym_LBRACE] = ACTIONS(1646),
    [anon_sym_SEMI] = ACTIONS(1648),
    [anon_sym_RBRACE] = ACTIONS(1648),
    [anon_sym_constraint] = ACTIONS(1646),
    [anon_sym_val] = ACTIONS(1646),
    [anon_sym_end] = ACTIONS(1646),
    [anon_sym_with] = ACTIONS(1646),
    [anon_sym_COLON_GT] = ACTIONS(1648),
    [anon_sym_object] = ACTIONS(1646),
    [anon_sym_inherit] = ACTIONS(1646),
    [anon_sym_method] = ACTIONS(1646),
    [anon_sym_initializer] = ACTIONS(1646),
    [anon_sym_AMP] = ACTIONS(1646),
    [anon_sym_POUND] = ACTIONS(1646),
    [anon_sym_COLON_COLON] = ACTIONS(1648),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1648),
    [anon_sym_then] = ACTIONS(1646),
    [anon_sym_else] = ACTIONS(1646),
    [anon_sym_do] = ACTIONS(1646),
    [anon_sym_done] = ACTIONS(1646),
    [anon_sym_to] = ACTIONS(1646),
    [anon_sym_downto] = ACTIONS(1646),
    [anon_sym_new] = ACTIONS(1646),
    [anon_sym_LBRACE_LT] = ACTIONS(1648),
    [anon_sym_GT_RBRACE] = ACTIONS(1648),
    [anon_sym_begin] = ACTIONS(1646),
    [sym_ocamlyacc_value] = ACTIONS(1648),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1646),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1648),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1646),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1646),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1648),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1648),
    [aux_sym_number_token1] = ACTIONS(1648),
    [aux_sym_signed_number_token1] = ACTIONS(1646),
    [anon_sym_SQUOTE] = ACTIONS(1648),
    [anon_sym_DQUOTE] = ACTIONS(1648),
    [sym_prefix_operator] = ACTIONS(1648),
    [aux_sym_sign_operator_token1] = ACTIONS(1646),
    [sym_hash_operator] = ACTIONS(1648),
    [anon_sym_lsl] = ACTIONS(1646),
    [anon_sym_lsr] = ACTIONS(1646),
    [anon_sym_asr] = ACTIONS(1646),
    [aux_sym_pow_operator_token1] = ACTIONS(1648),
    [anon_sym_mod] = ACTIONS(1646),
    [anon_sym_land] = ACTIONS(1646),
    [anon_sym_lor] = ACTIONS(1646),
    [anon_sym_lxor] = ACTIONS(1646),
    [aux_sym_mult_operator_token1] = ACTIONS(1646),
    [aux_sym_add_operator_token1] = ACTIONS(1646),
    [sym_concat_operator] = ACTIONS(1648),
    [sym_rel_operator] = ACTIONS(1646),
    [anon_sym_AMP_AMP] = ACTIONS(1646),
    [anon_sym_or] = ACTIONS(1646),
    [anon_sym_PIPE_PIPE] = ACTIONS(1646),
    [sym_assign_operator] = ACTIONS(1648),
    [sym__uppercase_identifier] = ACTIONS(1648),
    [aux_sym_tag_token1] = ACTIONS(1648),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(909)] = {
    [sym_attribute] = STATE(909),
    [sym__lowercase_identifier] = ACTIONS(1946),
    [anon_sym_TILDE] = ACTIONS(1946),
    [anon_sym_QMARK] = ACTIONS(1946),
    [anon_sym_LPAREN] = ACTIONS(1948),
    [anon_sym_RPAREN] = ACTIONS(1948),
    [anon_sym_COMMA] = ACTIONS(1948),
    [anon_sym_PIPE] = ACTIONS(1946),
    [anon_sym_COLON2] = ACTIONS(1946),
    [anon_sym_DOT] = ACTIONS(1948),
    [anon_sym_DASH_GT] = ACTIONS(1946),
    [anon_sym_LBRACK] = ACTIONS(1946),
    [anon_sym_RBRACK] = ACTIONS(1948),
    [anon_sym_true] = ACTIONS(1946),
    [anon_sym_false] = ACTIONS(1946),
    [anon_sym_LBRACE] = ACTIONS(1946),
    [anon_sym_SEMI] = ACTIONS(1948),
    [anon_sym_RBRACE] = ACTIONS(1948),
    [anon_sym_constraint] = ACTIONS(1946),
    [anon_sym_val] = ACTIONS(1946),
    [anon_sym_end] = ACTIONS(1946),
    [anon_sym_with] = ACTIONS(1946),
    [anon_sym_COLON_GT] = ACTIONS(1948),
    [anon_sym_object] = ACTIONS(1946),
    [anon_sym_inherit] = ACTIONS(1946),
    [anon_sym_method] = ACTIONS(1946),
    [anon_sym_initializer] = ACTIONS(1946),
    [anon_sym_AMP] = ACTIONS(1946),
    [anon_sym_POUND] = ACTIONS(1946),
    [anon_sym_COLON_COLON] = ACTIONS(1948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1948),
    [anon_sym_then] = ACTIONS(1946),
    [anon_sym_else] = ACTIONS(1946),
    [anon_sym_do] = ACTIONS(1946),
    [anon_sym_done] = ACTIONS(1946),
    [anon_sym_to] = ACTIONS(1946),
    [anon_sym_downto] = ACTIONS(1946),
    [anon_sym_new] = ACTIONS(1946),
    [anon_sym_LBRACE_LT] = ACTIONS(1948),
    [anon_sym_GT_RBRACE] = ACTIONS(1948),
    [anon_sym_begin] = ACTIONS(1946),
    [sym_ocamlyacc_value] = ACTIONS(1948),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1946),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1948),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1946),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1946),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1948),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1948),
    [aux_sym_number_token1] = ACTIONS(1948),
    [aux_sym_signed_number_token1] = ACTIONS(1946),
    [anon_sym_SQUOTE] = ACTIONS(1948),
    [anon_sym_DQUOTE] = ACTIONS(1948),
    [sym_prefix_operator] = ACTIONS(1948),
    [aux_sym_sign_operator_token1] = ACTIONS(1946),
    [sym_hash_operator] = ACTIONS(1948),
    [anon_sym_lsl] = ACTIONS(1946),
    [anon_sym_lsr] = ACTIONS(1946),
    [anon_sym_asr] = ACTIONS(1946),
    [aux_sym_pow_operator_token1] = ACTIONS(1948),
    [anon_sym_mod] = ACTIONS(1946),
    [anon_sym_land] = ACTIONS(1946),
    [anon_sym_lor] = ACTIONS(1946),
    [anon_sym_lxor] = ACTIONS(1946),
    [aux_sym_mult_operator_token1] = ACTIONS(1946),
    [aux_sym_add_operator_token1] = ACTIONS(1946),
    [sym_concat_operator] = ACTIONS(1948),
    [sym_rel_operator] = ACTIONS(1946),
    [anon_sym_AMP_AMP] = ACTIONS(1946),
    [anon_sym_or] = ACTIONS(1946),
    [anon_sym_PIPE_PIPE] = ACTIONS(1946),
    [sym_assign_operator] = ACTIONS(1948),
    [sym__uppercase_identifier] = ACTIONS(1948),
    [aux_sym_tag_token1] = ACTIONS(1948),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(910)] = {
    [sym_attribute] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(1898),
    [anon_sym_TILDE] = ACTIONS(1898),
    [anon_sym_QMARK] = ACTIONS(1898),
    [anon_sym_LPAREN] = ACTIONS(1900),
    [anon_sym_RPAREN] = ACTIONS(1900),
    [anon_sym_COMMA] = ACTIONS(1900),
    [anon_sym_PIPE] = ACTIONS(1898),
    [anon_sym_COLON2] = ACTIONS(1898),
    [anon_sym_DOT] = ACTIONS(1900),
    [anon_sym_DASH_GT] = ACTIONS(1898),
    [anon_sym_LBRACK] = ACTIONS(1898),
    [anon_sym_RBRACK] = ACTIONS(1900),
    [anon_sym_true] = ACTIONS(1898),
    [anon_sym_false] = ACTIONS(1898),
    [anon_sym_LBRACE] = ACTIONS(1898),
    [anon_sym_SEMI] = ACTIONS(1900),
    [anon_sym_RBRACE] = ACTIONS(1900),
    [anon_sym_constraint] = ACTIONS(1898),
    [anon_sym_val] = ACTIONS(1898),
    [anon_sym_end] = ACTIONS(1898),
    [anon_sym_with] = ACTIONS(1898),
    [anon_sym_COLON_GT] = ACTIONS(1900),
    [anon_sym_object] = ACTIONS(1898),
    [anon_sym_inherit] = ACTIONS(1898),
    [anon_sym_method] = ACTIONS(1898),
    [anon_sym_initializer] = ACTIONS(1898),
    [anon_sym_AMP] = ACTIONS(1898),
    [anon_sym_POUND] = ACTIONS(1898),
    [anon_sym_COLON_COLON] = ACTIONS(1900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1900),
    [anon_sym_then] = ACTIONS(1898),
    [anon_sym_else] = ACTIONS(1898),
    [anon_sym_do] = ACTIONS(1898),
    [anon_sym_done] = ACTIONS(1898),
    [anon_sym_to] = ACTIONS(1898),
    [anon_sym_downto] = ACTIONS(1898),
    [anon_sym_new] = ACTIONS(1898),
    [anon_sym_LBRACE_LT] = ACTIONS(1900),
    [anon_sym_GT_RBRACE] = ACTIONS(1900),
    [anon_sym_begin] = ACTIONS(1898),
    [sym_ocamlyacc_value] = ACTIONS(1900),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1898),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1900),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1898),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1898),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1900),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1900),
    [aux_sym_number_token1] = ACTIONS(1900),
    [aux_sym_signed_number_token1] = ACTIONS(1898),
    [anon_sym_SQUOTE] = ACTIONS(1900),
    [anon_sym_DQUOTE] = ACTIONS(1900),
    [sym_prefix_operator] = ACTIONS(1900),
    [aux_sym_sign_operator_token1] = ACTIONS(1898),
    [sym_hash_operator] = ACTIONS(1900),
    [anon_sym_lsl] = ACTIONS(1898),
    [anon_sym_lsr] = ACTIONS(1898),
    [anon_sym_asr] = ACTIONS(1898),
    [aux_sym_pow_operator_token1] = ACTIONS(1900),
    [anon_sym_mod] = ACTIONS(1898),
    [anon_sym_land] = ACTIONS(1898),
    [anon_sym_lor] = ACTIONS(1898),
    [anon_sym_lxor] = ACTIONS(1898),
    [aux_sym_mult_operator_token1] = ACTIONS(1898),
    [aux_sym_add_operator_token1] = ACTIONS(1898),
    [sym_concat_operator] = ACTIONS(1900),
    [sym_rel_operator] = ACTIONS(1898),
    [anon_sym_AMP_AMP] = ACTIONS(1898),
    [anon_sym_or] = ACTIONS(1898),
    [anon_sym_PIPE_PIPE] = ACTIONS(1898),
    [sym_assign_operator] = ACTIONS(1900),
    [sym__uppercase_identifier] = ACTIONS(1900),
    [aux_sym_tag_token1] = ACTIONS(1900),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(911)] = {
    [sym_attribute] = STATE(911),
    [sym__lowercase_identifier] = ACTIONS(1854),
    [anon_sym_TILDE] = ACTIONS(1854),
    [anon_sym_QMARK] = ACTIONS(1854),
    [anon_sym_LPAREN] = ACTIONS(1856),
    [anon_sym_RPAREN] = ACTIONS(1856),
    [anon_sym_COMMA] = ACTIONS(1856),
    [anon_sym_PIPE] = ACTIONS(1854),
    [anon_sym_COLON2] = ACTIONS(1854),
    [anon_sym_DOT] = ACTIONS(1856),
    [anon_sym_DASH_GT] = ACTIONS(1854),
    [anon_sym_LBRACK] = ACTIONS(1854),
    [anon_sym_RBRACK] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(1854),
    [anon_sym_false] = ACTIONS(1854),
    [anon_sym_LBRACE] = ACTIONS(1854),
    [anon_sym_SEMI] = ACTIONS(1856),
    [anon_sym_RBRACE] = ACTIONS(1856),
    [anon_sym_constraint] = ACTIONS(1854),
    [anon_sym_val] = ACTIONS(1854),
    [anon_sym_end] = ACTIONS(1854),
    [anon_sym_with] = ACTIONS(1854),
    [anon_sym_COLON_GT] = ACTIONS(1856),
    [anon_sym_object] = ACTIONS(1854),
    [anon_sym_inherit] = ACTIONS(1854),
    [anon_sym_method] = ACTIONS(1854),
    [anon_sym_initializer] = ACTIONS(1854),
    [anon_sym_AMP] = ACTIONS(1854),
    [anon_sym_POUND] = ACTIONS(1854),
    [anon_sym_COLON_COLON] = ACTIONS(1856),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1856),
    [anon_sym_then] = ACTIONS(1854),
    [anon_sym_else] = ACTIONS(1854),
    [anon_sym_do] = ACTIONS(1854),
    [anon_sym_done] = ACTIONS(1854),
    [anon_sym_to] = ACTIONS(1854),
    [anon_sym_downto] = ACTIONS(1854),
    [anon_sym_new] = ACTIONS(1854),
    [anon_sym_LBRACE_LT] = ACTIONS(1856),
    [anon_sym_GT_RBRACE] = ACTIONS(1856),
    [anon_sym_begin] = ACTIONS(1854),
    [sym_ocamlyacc_value] = ACTIONS(1856),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1854),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1856),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1854),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1854),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1856),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1856),
    [aux_sym_number_token1] = ACTIONS(1856),
    [aux_sym_signed_number_token1] = ACTIONS(1854),
    [anon_sym_SQUOTE] = ACTIONS(1856),
    [anon_sym_DQUOTE] = ACTIONS(1856),
    [sym_prefix_operator] = ACTIONS(1856),
    [aux_sym_sign_operator_token1] = ACTIONS(1854),
    [sym_hash_operator] = ACTIONS(1856),
    [anon_sym_lsl] = ACTIONS(1854),
    [anon_sym_lsr] = ACTIONS(1854),
    [anon_sym_asr] = ACTIONS(1854),
    [aux_sym_pow_operator_token1] = ACTIONS(1856),
    [anon_sym_mod] = ACTIONS(1854),
    [anon_sym_land] = ACTIONS(1854),
    [anon_sym_lor] = ACTIONS(1854),
    [anon_sym_lxor] = ACTIONS(1854),
    [aux_sym_mult_operator_token1] = ACTIONS(1854),
    [aux_sym_add_operator_token1] = ACTIONS(1854),
    [sym_concat_operator] = ACTIONS(1856),
    [sym_rel_operator] = ACTIONS(1854),
    [anon_sym_AMP_AMP] = ACTIONS(1854),
    [anon_sym_or] = ACTIONS(1854),
    [anon_sym_PIPE_PIPE] = ACTIONS(1854),
    [sym_assign_operator] = ACTIONS(1856),
    [sym__uppercase_identifier] = ACTIONS(1856),
    [aux_sym_tag_token1] = ACTIONS(1856),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(912)] = {
    [sym_attribute] = STATE(912),
    [sym__lowercase_identifier] = ACTIONS(1698),
    [anon_sym_TILDE] = ACTIONS(1698),
    [anon_sym_QMARK] = ACTIONS(1698),
    [anon_sym_LPAREN] = ACTIONS(1700),
    [anon_sym_RPAREN] = ACTIONS(1700),
    [anon_sym_COMMA] = ACTIONS(1700),
    [anon_sym_PIPE] = ACTIONS(1698),
    [anon_sym_COLON2] = ACTIONS(1698),
    [anon_sym_DOT] = ACTIONS(1702),
    [anon_sym_DASH_GT] = ACTIONS(1698),
    [anon_sym_LBRACK] = ACTIONS(1698),
    [anon_sym_RBRACK] = ACTIONS(1700),
    [anon_sym_true] = ACTIONS(1698),
    [anon_sym_false] = ACTIONS(1698),
    [anon_sym_LBRACE] = ACTIONS(1698),
    [anon_sym_SEMI] = ACTIONS(1700),
    [anon_sym_RBRACE] = ACTIONS(1700),
    [anon_sym_constraint] = ACTIONS(1698),
    [anon_sym_val] = ACTIONS(1698),
    [anon_sym_end] = ACTIONS(1698),
    [anon_sym_with] = ACTIONS(1698),
    [anon_sym_COLON_GT] = ACTIONS(1700),
    [anon_sym_object] = ACTIONS(1698),
    [anon_sym_inherit] = ACTIONS(1698),
    [anon_sym_method] = ACTIONS(1698),
    [anon_sym_initializer] = ACTIONS(1698),
    [anon_sym_AMP] = ACTIONS(1698),
    [anon_sym_POUND] = ACTIONS(1698),
    [anon_sym_COLON_COLON] = ACTIONS(1700),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1700),
    [anon_sym_then] = ACTIONS(1698),
    [anon_sym_else] = ACTIONS(1698),
    [anon_sym_do] = ACTIONS(1698),
    [anon_sym_done] = ACTIONS(1698),
    [anon_sym_to] = ACTIONS(1698),
    [anon_sym_downto] = ACTIONS(1698),
    [anon_sym_new] = ACTIONS(1698),
    [anon_sym_LBRACE_LT] = ACTIONS(1700),
    [anon_sym_GT_RBRACE] = ACTIONS(1700),
    [anon_sym_begin] = ACTIONS(1698),
    [sym_ocamlyacc_value] = ACTIONS(1700),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1698),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1700),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1700),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1700),
    [aux_sym_number_token1] = ACTIONS(1700),
    [aux_sym_signed_number_token1] = ACTIONS(1698),
    [anon_sym_SQUOTE] = ACTIONS(1700),
    [anon_sym_DQUOTE] = ACTIONS(1700),
    [sym_prefix_operator] = ACTIONS(1700),
    [aux_sym_sign_operator_token1] = ACTIONS(1698),
    [sym_hash_operator] = ACTIONS(1700),
    [anon_sym_lsl] = ACTIONS(1698),
    [anon_sym_lsr] = ACTIONS(1698),
    [anon_sym_asr] = ACTIONS(1698),
    [aux_sym_pow_operator_token1] = ACTIONS(1700),
    [anon_sym_mod] = ACTIONS(1698),
    [anon_sym_land] = ACTIONS(1698),
    [anon_sym_lor] = ACTIONS(1698),
    [anon_sym_lxor] = ACTIONS(1698),
    [aux_sym_mult_operator_token1] = ACTIONS(1698),
    [aux_sym_add_operator_token1] = ACTIONS(1698),
    [sym_concat_operator] = ACTIONS(1700),
    [sym_rel_operator] = ACTIONS(1698),
    [anon_sym_AMP_AMP] = ACTIONS(1698),
    [anon_sym_or] = ACTIONS(1698),
    [anon_sym_PIPE_PIPE] = ACTIONS(1698),
    [sym_assign_operator] = ACTIONS(1700),
    [sym__uppercase_identifier] = ACTIONS(1700),
    [aux_sym_tag_token1] = ACTIONS(1700),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(913)] = {
    [sym_attribute] = STATE(913),
    [sym__lowercase_identifier] = ACTIONS(1736),
    [anon_sym_TILDE] = ACTIONS(1736),
    [anon_sym_QMARK] = ACTIONS(1736),
    [anon_sym_LPAREN] = ACTIONS(1738),
    [anon_sym_RPAREN] = ACTIONS(1738),
    [anon_sym_COMMA] = ACTIONS(1738),
    [anon_sym_PIPE] = ACTIONS(1736),
    [anon_sym_COLON2] = ACTIONS(1736),
    [anon_sym_DOT] = ACTIONS(1738),
    [anon_sym_DASH_GT] = ACTIONS(1736),
    [anon_sym_LBRACK] = ACTIONS(1736),
    [anon_sym_RBRACK] = ACTIONS(1738),
    [anon_sym_true] = ACTIONS(1736),
    [anon_sym_false] = ACTIONS(1736),
    [anon_sym_LBRACE] = ACTIONS(1736),
    [anon_sym_SEMI] = ACTIONS(1738),
    [anon_sym_RBRACE] = ACTIONS(1738),
    [anon_sym_constraint] = ACTIONS(1736),
    [anon_sym_val] = ACTIONS(1736),
    [anon_sym_end] = ACTIONS(1736),
    [anon_sym_with] = ACTIONS(1736),
    [anon_sym_COLON_GT] = ACTIONS(1738),
    [anon_sym_object] = ACTIONS(1736),
    [anon_sym_inherit] = ACTIONS(1736),
    [anon_sym_method] = ACTIONS(1736),
    [anon_sym_initializer] = ACTIONS(1736),
    [anon_sym_AMP] = ACTIONS(1736),
    [anon_sym_POUND] = ACTIONS(1736),
    [anon_sym_COLON_COLON] = ACTIONS(1738),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1738),
    [anon_sym_then] = ACTIONS(1736),
    [anon_sym_else] = ACTIONS(1736),
    [anon_sym_do] = ACTIONS(1736),
    [anon_sym_done] = ACTIONS(1736),
    [anon_sym_to] = ACTIONS(1736),
    [anon_sym_downto] = ACTIONS(1736),
    [anon_sym_new] = ACTIONS(1736),
    [anon_sym_LBRACE_LT] = ACTIONS(1738),
    [anon_sym_GT_RBRACE] = ACTIONS(1738),
    [anon_sym_begin] = ACTIONS(1736),
    [sym_ocamlyacc_value] = ACTIONS(1738),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1736),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1738),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1736),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1736),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1738),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1738),
    [aux_sym_number_token1] = ACTIONS(1738),
    [aux_sym_signed_number_token1] = ACTIONS(1736),
    [anon_sym_SQUOTE] = ACTIONS(1738),
    [anon_sym_DQUOTE] = ACTIONS(1738),
    [sym_prefix_operator] = ACTIONS(1738),
    [aux_sym_sign_operator_token1] = ACTIONS(1736),
    [sym_hash_operator] = ACTIONS(1738),
    [anon_sym_lsl] = ACTIONS(1736),
    [anon_sym_lsr] = ACTIONS(1736),
    [anon_sym_asr] = ACTIONS(1736),
    [aux_sym_pow_operator_token1] = ACTIONS(1738),
    [anon_sym_mod] = ACTIONS(1736),
    [anon_sym_land] = ACTIONS(1736),
    [anon_sym_lor] = ACTIONS(1736),
    [anon_sym_lxor] = ACTIONS(1736),
    [aux_sym_mult_operator_token1] = ACTIONS(1736),
    [aux_sym_add_operator_token1] = ACTIONS(1736),
    [sym_concat_operator] = ACTIONS(1738),
    [sym_rel_operator] = ACTIONS(1736),
    [anon_sym_AMP_AMP] = ACTIONS(1736),
    [anon_sym_or] = ACTIONS(1736),
    [anon_sym_PIPE_PIPE] = ACTIONS(1736),
    [sym_assign_operator] = ACTIONS(1738),
    [sym__uppercase_identifier] = ACTIONS(1738),
    [aux_sym_tag_token1] = ACTIONS(1738),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(914)] = {
    [sym_attribute] = STATE(914),
    [sym__lowercase_identifier] = ACTIONS(1586),
    [anon_sym_TILDE] = ACTIONS(1586),
    [anon_sym_QMARK] = ACTIONS(1586),
    [anon_sym_LPAREN] = ACTIONS(1588),
    [anon_sym_RPAREN] = ACTIONS(1588),
    [anon_sym_COMMA] = ACTIONS(1588),
    [anon_sym_PIPE] = ACTIONS(1586),
    [anon_sym_COLON2] = ACTIONS(1586),
    [anon_sym_DOT] = ACTIONS(1588),
    [anon_sym_DASH_GT] = ACTIONS(1586),
    [anon_sym_LBRACK] = ACTIONS(1586),
    [anon_sym_RBRACK] = ACTIONS(1588),
    [anon_sym_true] = ACTIONS(1586),
    [anon_sym_false] = ACTIONS(1586),
    [anon_sym_LBRACE] = ACTIONS(1586),
    [anon_sym_SEMI] = ACTIONS(1588),
    [anon_sym_RBRACE] = ACTIONS(1588),
    [anon_sym_constraint] = ACTIONS(1586),
    [anon_sym_val] = ACTIONS(1586),
    [anon_sym_end] = ACTIONS(1586),
    [anon_sym_with] = ACTIONS(1586),
    [anon_sym_COLON_GT] = ACTIONS(1588),
    [anon_sym_object] = ACTIONS(1586),
    [anon_sym_inherit] = ACTIONS(1586),
    [anon_sym_method] = ACTIONS(1586),
    [anon_sym_initializer] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1586),
    [anon_sym_POUND] = ACTIONS(1586),
    [anon_sym_COLON_COLON] = ACTIONS(1588),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1588),
    [anon_sym_then] = ACTIONS(1586),
    [anon_sym_else] = ACTIONS(1586),
    [anon_sym_do] = ACTIONS(1586),
    [anon_sym_done] = ACTIONS(1586),
    [anon_sym_to] = ACTIONS(1586),
    [anon_sym_downto] = ACTIONS(1586),
    [anon_sym_new] = ACTIONS(1586),
    [anon_sym_LBRACE_LT] = ACTIONS(1588),
    [anon_sym_GT_RBRACE] = ACTIONS(1588),
    [anon_sym_begin] = ACTIONS(1586),
    [sym_ocamlyacc_value] = ACTIONS(1588),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1586),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1588),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1586),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1586),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1588),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1588),
    [aux_sym_number_token1] = ACTIONS(1588),
    [aux_sym_signed_number_token1] = ACTIONS(1586),
    [anon_sym_SQUOTE] = ACTIONS(1588),
    [anon_sym_DQUOTE] = ACTIONS(1588),
    [sym_prefix_operator] = ACTIONS(1588),
    [aux_sym_sign_operator_token1] = ACTIONS(1586),
    [sym_hash_operator] = ACTIONS(1588),
    [anon_sym_lsl] = ACTIONS(1586),
    [anon_sym_lsr] = ACTIONS(1586),
    [anon_sym_asr] = ACTIONS(1586),
    [aux_sym_pow_operator_token1] = ACTIONS(1588),
    [anon_sym_mod] = ACTIONS(1586),
    [anon_sym_land] = ACTIONS(1586),
    [anon_sym_lor] = ACTIONS(1586),
    [anon_sym_lxor] = ACTIONS(1586),
    [aux_sym_mult_operator_token1] = ACTIONS(1586),
    [aux_sym_add_operator_token1] = ACTIONS(1586),
    [sym_concat_operator] = ACTIONS(1588),
    [sym_rel_operator] = ACTIONS(1586),
    [anon_sym_AMP_AMP] = ACTIONS(1586),
    [anon_sym_or] = ACTIONS(1586),
    [anon_sym_PIPE_PIPE] = ACTIONS(1586),
    [sym_assign_operator] = ACTIONS(1588),
    [sym__uppercase_identifier] = ACTIONS(1588),
    [aux_sym_tag_token1] = ACTIONS(1588),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(915)] = {
    [sym_attribute] = STATE(915),
    [sym__lowercase_identifier] = ACTIONS(1590),
    [anon_sym_TILDE] = ACTIONS(1590),
    [anon_sym_QMARK] = ACTIONS(1590),
    [anon_sym_LPAREN] = ACTIONS(1592),
    [anon_sym_RPAREN] = ACTIONS(1592),
    [anon_sym_COMMA] = ACTIONS(1592),
    [anon_sym_PIPE] = ACTIONS(1590),
    [anon_sym_COLON2] = ACTIONS(1590),
    [anon_sym_DOT] = ACTIONS(1592),
    [anon_sym_DASH_GT] = ACTIONS(1590),
    [anon_sym_LBRACK] = ACTIONS(1590),
    [anon_sym_RBRACK] = ACTIONS(1592),
    [anon_sym_true] = ACTIONS(1590),
    [anon_sym_false] = ACTIONS(1590),
    [anon_sym_LBRACE] = ACTIONS(1590),
    [anon_sym_SEMI] = ACTIONS(1592),
    [anon_sym_RBRACE] = ACTIONS(1592),
    [anon_sym_constraint] = ACTIONS(1590),
    [anon_sym_val] = ACTIONS(1590),
    [anon_sym_end] = ACTIONS(1590),
    [anon_sym_with] = ACTIONS(1590),
    [anon_sym_COLON_GT] = ACTIONS(1592),
    [anon_sym_object] = ACTIONS(1590),
    [anon_sym_inherit] = ACTIONS(1590),
    [anon_sym_method] = ACTIONS(1590),
    [anon_sym_initializer] = ACTIONS(1590),
    [anon_sym_AMP] = ACTIONS(1590),
    [anon_sym_POUND] = ACTIONS(1590),
    [anon_sym_COLON_COLON] = ACTIONS(1592),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1592),
    [anon_sym_then] = ACTIONS(1590),
    [anon_sym_else] = ACTIONS(1590),
    [anon_sym_do] = ACTIONS(1590),
    [anon_sym_done] = ACTIONS(1590),
    [anon_sym_to] = ACTIONS(1590),
    [anon_sym_downto] = ACTIONS(1590),
    [anon_sym_new] = ACTIONS(1590),
    [anon_sym_LBRACE_LT] = ACTIONS(1592),
    [anon_sym_GT_RBRACE] = ACTIONS(1592),
    [anon_sym_begin] = ACTIONS(1590),
    [sym_ocamlyacc_value] = ACTIONS(1592),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1590),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1592),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1590),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1590),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1592),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1592),
    [aux_sym_number_token1] = ACTIONS(1592),
    [aux_sym_signed_number_token1] = ACTIONS(1590),
    [anon_sym_SQUOTE] = ACTIONS(1592),
    [anon_sym_DQUOTE] = ACTIONS(1592),
    [sym_prefix_operator] = ACTIONS(1592),
    [aux_sym_sign_operator_token1] = ACTIONS(1590),
    [sym_hash_operator] = ACTIONS(1592),
    [anon_sym_lsl] = ACTIONS(1590),
    [anon_sym_lsr] = ACTIONS(1590),
    [anon_sym_asr] = ACTIONS(1590),
    [aux_sym_pow_operator_token1] = ACTIONS(1592),
    [anon_sym_mod] = ACTIONS(1590),
    [anon_sym_land] = ACTIONS(1590),
    [anon_sym_lor] = ACTIONS(1590),
    [anon_sym_lxor] = ACTIONS(1590),
    [aux_sym_mult_operator_token1] = ACTIONS(1590),
    [aux_sym_add_operator_token1] = ACTIONS(1590),
    [sym_concat_operator] = ACTIONS(1592),
    [sym_rel_operator] = ACTIONS(1590),
    [anon_sym_AMP_AMP] = ACTIONS(1590),
    [anon_sym_or] = ACTIONS(1590),
    [anon_sym_PIPE_PIPE] = ACTIONS(1590),
    [sym_assign_operator] = ACTIONS(1592),
    [sym__uppercase_identifier] = ACTIONS(1592),
    [aux_sym_tag_token1] = ACTIONS(1592),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(916)] = {
    [sym_attribute] = STATE(916),
    [sym__lowercase_identifier] = ACTIONS(1610),
    [anon_sym_TILDE] = ACTIONS(1610),
    [anon_sym_QMARK] = ACTIONS(1610),
    [anon_sym_LPAREN] = ACTIONS(1612),
    [anon_sym_RPAREN] = ACTIONS(1612),
    [anon_sym_COMMA] = ACTIONS(1612),
    [anon_sym_PIPE] = ACTIONS(1610),
    [anon_sym_COLON2] = ACTIONS(1610),
    [anon_sym_DOT] = ACTIONS(1612),
    [anon_sym_DASH_GT] = ACTIONS(1610),
    [anon_sym_LBRACK] = ACTIONS(1610),
    [anon_sym_RBRACK] = ACTIONS(1612),
    [anon_sym_true] = ACTIONS(1610),
    [anon_sym_false] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1610),
    [anon_sym_SEMI] = ACTIONS(1612),
    [anon_sym_RBRACE] = ACTIONS(1612),
    [anon_sym_constraint] = ACTIONS(1610),
    [anon_sym_val] = ACTIONS(1610),
    [anon_sym_end] = ACTIONS(1610),
    [anon_sym_with] = ACTIONS(1610),
    [anon_sym_COLON_GT] = ACTIONS(1612),
    [anon_sym_object] = ACTIONS(1610),
    [anon_sym_inherit] = ACTIONS(1610),
    [anon_sym_method] = ACTIONS(1610),
    [anon_sym_initializer] = ACTIONS(1610),
    [anon_sym_AMP] = ACTIONS(1610),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_COLON_COLON] = ACTIONS(1612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1612),
    [anon_sym_then] = ACTIONS(1610),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_do] = ACTIONS(1610),
    [anon_sym_done] = ACTIONS(1610),
    [anon_sym_to] = ACTIONS(1610),
    [anon_sym_downto] = ACTIONS(1610),
    [anon_sym_new] = ACTIONS(1610),
    [anon_sym_LBRACE_LT] = ACTIONS(1612),
    [anon_sym_GT_RBRACE] = ACTIONS(1612),
    [anon_sym_begin] = ACTIONS(1610),
    [sym_ocamlyacc_value] = ACTIONS(1612),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1610),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1612),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1610),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1610),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1612),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1612),
    [aux_sym_number_token1] = ACTIONS(1612),
    [aux_sym_signed_number_token1] = ACTIONS(1610),
    [anon_sym_SQUOTE] = ACTIONS(1612),
    [anon_sym_DQUOTE] = ACTIONS(1612),
    [sym_prefix_operator] = ACTIONS(1612),
    [aux_sym_sign_operator_token1] = ACTIONS(1610),
    [sym_hash_operator] = ACTIONS(1612),
    [anon_sym_lsl] = ACTIONS(1610),
    [anon_sym_lsr] = ACTIONS(1610),
    [anon_sym_asr] = ACTIONS(1610),
    [aux_sym_pow_operator_token1] = ACTIONS(1612),
    [anon_sym_mod] = ACTIONS(1610),
    [anon_sym_land] = ACTIONS(1610),
    [anon_sym_lor] = ACTIONS(1610),
    [anon_sym_lxor] = ACTIONS(1610),
    [aux_sym_mult_operator_token1] = ACTIONS(1610),
    [aux_sym_add_operator_token1] = ACTIONS(1610),
    [sym_concat_operator] = ACTIONS(1612),
    [sym_rel_operator] = ACTIONS(1610),
    [anon_sym_AMP_AMP] = ACTIONS(1610),
    [anon_sym_or] = ACTIONS(1610),
    [anon_sym_PIPE_PIPE] = ACTIONS(1610),
    [sym_assign_operator] = ACTIONS(1612),
    [sym__uppercase_identifier] = ACTIONS(1612),
    [aux_sym_tag_token1] = ACTIONS(1612),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(917)] = {
    [sym_attribute] = STATE(917),
    [sym__lowercase_identifier] = ACTIONS(1630),
    [anon_sym_TILDE] = ACTIONS(1630),
    [anon_sym_QMARK] = ACTIONS(1630),
    [anon_sym_LPAREN] = ACTIONS(1632),
    [anon_sym_RPAREN] = ACTIONS(1632),
    [anon_sym_COMMA] = ACTIONS(1632),
    [anon_sym_PIPE] = ACTIONS(1630),
    [anon_sym_COLON2] = ACTIONS(1630),
    [anon_sym_DOT] = ACTIONS(1632),
    [anon_sym_DASH_GT] = ACTIONS(1630),
    [anon_sym_LBRACK] = ACTIONS(1630),
    [anon_sym_RBRACK] = ACTIONS(1632),
    [anon_sym_true] = ACTIONS(1630),
    [anon_sym_false] = ACTIONS(1630),
    [anon_sym_LBRACE] = ACTIONS(1630),
    [anon_sym_SEMI] = ACTIONS(1632),
    [anon_sym_RBRACE] = ACTIONS(1632),
    [anon_sym_constraint] = ACTIONS(1630),
    [anon_sym_val] = ACTIONS(1630),
    [anon_sym_end] = ACTIONS(1630),
    [anon_sym_with] = ACTIONS(1630),
    [anon_sym_COLON_GT] = ACTIONS(1632),
    [anon_sym_object] = ACTIONS(1630),
    [anon_sym_inherit] = ACTIONS(1630),
    [anon_sym_method] = ACTIONS(1630),
    [anon_sym_initializer] = ACTIONS(1630),
    [anon_sym_AMP] = ACTIONS(1630),
    [anon_sym_POUND] = ACTIONS(1630),
    [anon_sym_COLON_COLON] = ACTIONS(1632),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1632),
    [anon_sym_then] = ACTIONS(1630),
    [anon_sym_else] = ACTIONS(1630),
    [anon_sym_do] = ACTIONS(1630),
    [anon_sym_done] = ACTIONS(1630),
    [anon_sym_to] = ACTIONS(1630),
    [anon_sym_downto] = ACTIONS(1630),
    [anon_sym_new] = ACTIONS(1630),
    [anon_sym_LBRACE_LT] = ACTIONS(1632),
    [anon_sym_GT_RBRACE] = ACTIONS(1632),
    [anon_sym_begin] = ACTIONS(1630),
    [sym_ocamlyacc_value] = ACTIONS(1632),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1630),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1632),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1630),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1630),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1632),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1632),
    [aux_sym_number_token1] = ACTIONS(1632),
    [aux_sym_signed_number_token1] = ACTIONS(1630),
    [anon_sym_SQUOTE] = ACTIONS(1632),
    [anon_sym_DQUOTE] = ACTIONS(1632),
    [sym_prefix_operator] = ACTIONS(1632),
    [aux_sym_sign_operator_token1] = ACTIONS(1630),
    [sym_hash_operator] = ACTIONS(1632),
    [anon_sym_lsl] = ACTIONS(1630),
    [anon_sym_lsr] = ACTIONS(1630),
    [anon_sym_asr] = ACTIONS(1630),
    [aux_sym_pow_operator_token1] = ACTIONS(1632),
    [anon_sym_mod] = ACTIONS(1630),
    [anon_sym_land] = ACTIONS(1630),
    [anon_sym_lor] = ACTIONS(1630),
    [anon_sym_lxor] = ACTIONS(1630),
    [aux_sym_mult_operator_token1] = ACTIONS(1630),
    [aux_sym_add_operator_token1] = ACTIONS(1630),
    [sym_concat_operator] = ACTIONS(1632),
    [sym_rel_operator] = ACTIONS(1630),
    [anon_sym_AMP_AMP] = ACTIONS(1630),
    [anon_sym_or] = ACTIONS(1630),
    [anon_sym_PIPE_PIPE] = ACTIONS(1630),
    [sym_assign_operator] = ACTIONS(1632),
    [sym__uppercase_identifier] = ACTIONS(1632),
    [aux_sym_tag_token1] = ACTIONS(1632),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(918)] = {
    [sym_attribute] = STATE(918),
    [sym__lowercase_identifier] = ACTIONS(1638),
    [anon_sym_TILDE] = ACTIONS(1638),
    [anon_sym_QMARK] = ACTIONS(1638),
    [anon_sym_LPAREN] = ACTIONS(1640),
    [anon_sym_RPAREN] = ACTIONS(1640),
    [anon_sym_COMMA] = ACTIONS(1640),
    [anon_sym_PIPE] = ACTIONS(1638),
    [anon_sym_COLON2] = ACTIONS(1638),
    [anon_sym_DOT] = ACTIONS(1640),
    [anon_sym_DASH_GT] = ACTIONS(1638),
    [anon_sym_LBRACK] = ACTIONS(1638),
    [anon_sym_RBRACK] = ACTIONS(1640),
    [anon_sym_true] = ACTIONS(1638),
    [anon_sym_false] = ACTIONS(1638),
    [anon_sym_LBRACE] = ACTIONS(1638),
    [anon_sym_SEMI] = ACTIONS(1640),
    [anon_sym_RBRACE] = ACTIONS(1640),
    [anon_sym_constraint] = ACTIONS(1638),
    [anon_sym_val] = ACTIONS(1638),
    [anon_sym_end] = ACTIONS(1638),
    [anon_sym_with] = ACTIONS(1638),
    [anon_sym_COLON_GT] = ACTIONS(1640),
    [anon_sym_object] = ACTIONS(1638),
    [anon_sym_inherit] = ACTIONS(1638),
    [anon_sym_method] = ACTIONS(1638),
    [anon_sym_initializer] = ACTIONS(1638),
    [anon_sym_AMP] = ACTIONS(1638),
    [anon_sym_POUND] = ACTIONS(1638),
    [anon_sym_COLON_COLON] = ACTIONS(1640),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1640),
    [anon_sym_then] = ACTIONS(1638),
    [anon_sym_else] = ACTIONS(1638),
    [anon_sym_do] = ACTIONS(1638),
    [anon_sym_done] = ACTIONS(1638),
    [anon_sym_to] = ACTIONS(1638),
    [anon_sym_downto] = ACTIONS(1638),
    [anon_sym_new] = ACTIONS(1638),
    [anon_sym_LBRACE_LT] = ACTIONS(1640),
    [anon_sym_GT_RBRACE] = ACTIONS(1640),
    [anon_sym_begin] = ACTIONS(1638),
    [sym_ocamlyacc_value] = ACTIONS(1640),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1638),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1640),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1638),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1638),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1640),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1640),
    [aux_sym_number_token1] = ACTIONS(1640),
    [aux_sym_signed_number_token1] = ACTIONS(1638),
    [anon_sym_SQUOTE] = ACTIONS(1640),
    [anon_sym_DQUOTE] = ACTIONS(1640),
    [sym_prefix_operator] = ACTIONS(1640),
    [aux_sym_sign_operator_token1] = ACTIONS(1638),
    [sym_hash_operator] = ACTIONS(1640),
    [anon_sym_lsl] = ACTIONS(1638),
    [anon_sym_lsr] = ACTIONS(1638),
    [anon_sym_asr] = ACTIONS(1638),
    [aux_sym_pow_operator_token1] = ACTIONS(1640),
    [anon_sym_mod] = ACTIONS(1638),
    [anon_sym_land] = ACTIONS(1638),
    [anon_sym_lor] = ACTIONS(1638),
    [anon_sym_lxor] = ACTIONS(1638),
    [aux_sym_mult_operator_token1] = ACTIONS(1638),
    [aux_sym_add_operator_token1] = ACTIONS(1638),
    [sym_concat_operator] = ACTIONS(1640),
    [sym_rel_operator] = ACTIONS(1638),
    [anon_sym_AMP_AMP] = ACTIONS(1638),
    [anon_sym_or] = ACTIONS(1638),
    [anon_sym_PIPE_PIPE] = ACTIONS(1638),
    [sym_assign_operator] = ACTIONS(1640),
    [sym__uppercase_identifier] = ACTIONS(1640),
    [aux_sym_tag_token1] = ACTIONS(1640),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(919)] = {
    [sym_attribute] = STATE(919),
    [sym__lowercase_identifier] = ACTIONS(1708),
    [anon_sym_TILDE] = ACTIONS(1708),
    [anon_sym_QMARK] = ACTIONS(1708),
    [anon_sym_LPAREN] = ACTIONS(1710),
    [anon_sym_RPAREN] = ACTIONS(1710),
    [anon_sym_COMMA] = ACTIONS(1710),
    [anon_sym_PIPE] = ACTIONS(1708),
    [anon_sym_COLON2] = ACTIONS(1708),
    [anon_sym_DOT] = ACTIONS(1710),
    [anon_sym_DASH_GT] = ACTIONS(1708),
    [anon_sym_LBRACK] = ACTIONS(1708),
    [anon_sym_RBRACK] = ACTIONS(1710),
    [anon_sym_true] = ACTIONS(1708),
    [anon_sym_false] = ACTIONS(1708),
    [anon_sym_LBRACE] = ACTIONS(1708),
    [anon_sym_SEMI] = ACTIONS(1710),
    [anon_sym_RBRACE] = ACTIONS(1710),
    [anon_sym_constraint] = ACTIONS(1708),
    [anon_sym_val] = ACTIONS(1708),
    [anon_sym_end] = ACTIONS(1708),
    [anon_sym_with] = ACTIONS(1708),
    [anon_sym_COLON_GT] = ACTIONS(1710),
    [anon_sym_object] = ACTIONS(1708),
    [anon_sym_inherit] = ACTIONS(1708),
    [anon_sym_method] = ACTIONS(1708),
    [anon_sym_initializer] = ACTIONS(1708),
    [anon_sym_AMP] = ACTIONS(1708),
    [anon_sym_POUND] = ACTIONS(1708),
    [anon_sym_COLON_COLON] = ACTIONS(1710),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1710),
    [anon_sym_then] = ACTIONS(1708),
    [anon_sym_else] = ACTIONS(1708),
    [anon_sym_do] = ACTIONS(1708),
    [anon_sym_done] = ACTIONS(1708),
    [anon_sym_to] = ACTIONS(1708),
    [anon_sym_downto] = ACTIONS(1708),
    [anon_sym_new] = ACTIONS(1708),
    [anon_sym_LBRACE_LT] = ACTIONS(1710),
    [anon_sym_GT_RBRACE] = ACTIONS(1710),
    [anon_sym_begin] = ACTIONS(1708),
    [sym_ocamlyacc_value] = ACTIONS(1710),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1708),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1710),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1708),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1708),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1710),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1710),
    [aux_sym_number_token1] = ACTIONS(1710),
    [aux_sym_signed_number_token1] = ACTIONS(1708),
    [anon_sym_SQUOTE] = ACTIONS(1710),
    [anon_sym_DQUOTE] = ACTIONS(1710),
    [sym_prefix_operator] = ACTIONS(1710),
    [aux_sym_sign_operator_token1] = ACTIONS(1708),
    [sym_hash_operator] = ACTIONS(1710),
    [anon_sym_lsl] = ACTIONS(1708),
    [anon_sym_lsr] = ACTIONS(1708),
    [anon_sym_asr] = ACTIONS(1708),
    [aux_sym_pow_operator_token1] = ACTIONS(1710),
    [anon_sym_mod] = ACTIONS(1708),
    [anon_sym_land] = ACTIONS(1708),
    [anon_sym_lor] = ACTIONS(1708),
    [anon_sym_lxor] = ACTIONS(1708),
    [aux_sym_mult_operator_token1] = ACTIONS(1708),
    [aux_sym_add_operator_token1] = ACTIONS(1708),
    [sym_concat_operator] = ACTIONS(1710),
    [sym_rel_operator] = ACTIONS(1708),
    [anon_sym_AMP_AMP] = ACTIONS(1708),
    [anon_sym_or] = ACTIONS(1708),
    [anon_sym_PIPE_PIPE] = ACTIONS(1708),
    [sym_assign_operator] = ACTIONS(1710),
    [sym__uppercase_identifier] = ACTIONS(1710),
    [aux_sym_tag_token1] = ACTIONS(1710),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(920)] = {
    [sym_attribute] = STATE(920),
    [sym__lowercase_identifier] = ACTIONS(1694),
    [anon_sym_TILDE] = ACTIONS(1694),
    [anon_sym_QMARK] = ACTIONS(1694),
    [anon_sym_LPAREN] = ACTIONS(1696),
    [anon_sym_RPAREN] = ACTIONS(1696),
    [anon_sym_COMMA] = ACTIONS(1696),
    [anon_sym_PIPE] = ACTIONS(1694),
    [anon_sym_COLON2] = ACTIONS(1694),
    [anon_sym_DOT] = ACTIONS(1696),
    [anon_sym_DASH_GT] = ACTIONS(1694),
    [anon_sym_LBRACK] = ACTIONS(1694),
    [anon_sym_RBRACK] = ACTIONS(1696),
    [anon_sym_true] = ACTIONS(1694),
    [anon_sym_false] = ACTIONS(1694),
    [anon_sym_LBRACE] = ACTIONS(1694),
    [anon_sym_SEMI] = ACTIONS(1696),
    [anon_sym_RBRACE] = ACTIONS(1696),
    [anon_sym_constraint] = ACTIONS(1694),
    [anon_sym_val] = ACTIONS(1694),
    [anon_sym_end] = ACTIONS(1694),
    [anon_sym_with] = ACTIONS(1694),
    [anon_sym_COLON_GT] = ACTIONS(1696),
    [anon_sym_object] = ACTIONS(1694),
    [anon_sym_inherit] = ACTIONS(1694),
    [anon_sym_method] = ACTIONS(1694),
    [anon_sym_initializer] = ACTIONS(1694),
    [anon_sym_AMP] = ACTIONS(1694),
    [anon_sym_POUND] = ACTIONS(1694),
    [anon_sym_COLON_COLON] = ACTIONS(1696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1696),
    [anon_sym_then] = ACTIONS(1694),
    [anon_sym_else] = ACTIONS(1694),
    [anon_sym_do] = ACTIONS(1694),
    [anon_sym_done] = ACTIONS(1694),
    [anon_sym_to] = ACTIONS(1694),
    [anon_sym_downto] = ACTIONS(1694),
    [anon_sym_new] = ACTIONS(1694),
    [anon_sym_LBRACE_LT] = ACTIONS(1696),
    [anon_sym_GT_RBRACE] = ACTIONS(1696),
    [anon_sym_begin] = ACTIONS(1694),
    [sym_ocamlyacc_value] = ACTIONS(1696),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1694),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1696),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1694),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1694),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1696),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1696),
    [aux_sym_number_token1] = ACTIONS(1696),
    [aux_sym_signed_number_token1] = ACTIONS(1694),
    [anon_sym_SQUOTE] = ACTIONS(1696),
    [anon_sym_DQUOTE] = ACTIONS(1696),
    [sym_prefix_operator] = ACTIONS(1696),
    [aux_sym_sign_operator_token1] = ACTIONS(1694),
    [sym_hash_operator] = ACTIONS(1696),
    [anon_sym_lsl] = ACTIONS(1694),
    [anon_sym_lsr] = ACTIONS(1694),
    [anon_sym_asr] = ACTIONS(1694),
    [aux_sym_pow_operator_token1] = ACTIONS(1696),
    [anon_sym_mod] = ACTIONS(1694),
    [anon_sym_land] = ACTIONS(1694),
    [anon_sym_lor] = ACTIONS(1694),
    [anon_sym_lxor] = ACTIONS(1694),
    [aux_sym_mult_operator_token1] = ACTIONS(1694),
    [aux_sym_add_operator_token1] = ACTIONS(1694),
    [sym_concat_operator] = ACTIONS(1696),
    [sym_rel_operator] = ACTIONS(1694),
    [anon_sym_AMP_AMP] = ACTIONS(1694),
    [anon_sym_or] = ACTIONS(1694),
    [anon_sym_PIPE_PIPE] = ACTIONS(1694),
    [sym_assign_operator] = ACTIONS(1696),
    [sym__uppercase_identifier] = ACTIONS(1696),
    [aux_sym_tag_token1] = ACTIONS(1696),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(921)] = {
    [sym_attribute] = STATE(921),
    [sym__lowercase_identifier] = ACTIONS(1720),
    [anon_sym_TILDE] = ACTIONS(1720),
    [anon_sym_QMARK] = ACTIONS(1720),
    [anon_sym_LPAREN] = ACTIONS(1722),
    [anon_sym_RPAREN] = ACTIONS(1722),
    [anon_sym_COMMA] = ACTIONS(1722),
    [anon_sym_PIPE] = ACTIONS(1720),
    [anon_sym_COLON2] = ACTIONS(1720),
    [anon_sym_DOT] = ACTIONS(1722),
    [anon_sym_DASH_GT] = ACTIONS(1720),
    [anon_sym_LBRACK] = ACTIONS(1720),
    [anon_sym_RBRACK] = ACTIONS(1722),
    [anon_sym_true] = ACTIONS(1720),
    [anon_sym_false] = ACTIONS(1720),
    [anon_sym_LBRACE] = ACTIONS(1720),
    [anon_sym_SEMI] = ACTIONS(1722),
    [anon_sym_RBRACE] = ACTIONS(1722),
    [anon_sym_constraint] = ACTIONS(1720),
    [anon_sym_val] = ACTIONS(1720),
    [anon_sym_end] = ACTIONS(1720),
    [anon_sym_with] = ACTIONS(1720),
    [anon_sym_COLON_GT] = ACTIONS(1722),
    [anon_sym_object] = ACTIONS(1720),
    [anon_sym_inherit] = ACTIONS(1720),
    [anon_sym_method] = ACTIONS(1720),
    [anon_sym_initializer] = ACTIONS(1720),
    [anon_sym_AMP] = ACTIONS(1720),
    [anon_sym_POUND] = ACTIONS(1720),
    [anon_sym_COLON_COLON] = ACTIONS(1722),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1722),
    [anon_sym_then] = ACTIONS(1720),
    [anon_sym_else] = ACTIONS(1720),
    [anon_sym_do] = ACTIONS(1720),
    [anon_sym_done] = ACTIONS(1720),
    [anon_sym_to] = ACTIONS(1720),
    [anon_sym_downto] = ACTIONS(1720),
    [anon_sym_new] = ACTIONS(1720),
    [anon_sym_LBRACE_LT] = ACTIONS(1722),
    [anon_sym_GT_RBRACE] = ACTIONS(1722),
    [anon_sym_begin] = ACTIONS(1720),
    [sym_ocamlyacc_value] = ACTIONS(1722),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1720),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1722),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1720),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1720),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1722),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1722),
    [aux_sym_number_token1] = ACTIONS(1722),
    [aux_sym_signed_number_token1] = ACTIONS(1720),
    [anon_sym_SQUOTE] = ACTIONS(1722),
    [anon_sym_DQUOTE] = ACTIONS(1722),
    [sym_prefix_operator] = ACTIONS(1722),
    [aux_sym_sign_operator_token1] = ACTIONS(1720),
    [sym_hash_operator] = ACTIONS(1722),
    [anon_sym_lsl] = ACTIONS(1720),
    [anon_sym_lsr] = ACTIONS(1720),
    [anon_sym_asr] = ACTIONS(1720),
    [aux_sym_pow_operator_token1] = ACTIONS(1722),
    [anon_sym_mod] = ACTIONS(1720),
    [anon_sym_land] = ACTIONS(1720),
    [anon_sym_lor] = ACTIONS(1720),
    [anon_sym_lxor] = ACTIONS(1720),
    [aux_sym_mult_operator_token1] = ACTIONS(1720),
    [aux_sym_add_operator_token1] = ACTIONS(1720),
    [sym_concat_operator] = ACTIONS(1722),
    [sym_rel_operator] = ACTIONS(1720),
    [anon_sym_AMP_AMP] = ACTIONS(1720),
    [anon_sym_or] = ACTIONS(1720),
    [anon_sym_PIPE_PIPE] = ACTIONS(1720),
    [sym_assign_operator] = ACTIONS(1722),
    [sym__uppercase_identifier] = ACTIONS(1722),
    [aux_sym_tag_token1] = ACTIONS(1722),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(922)] = {
    [sym_attribute] = STATE(922),
    [sym__lowercase_identifier] = ACTIONS(1704),
    [anon_sym_TILDE] = ACTIONS(1704),
    [anon_sym_QMARK] = ACTIONS(1704),
    [anon_sym_LPAREN] = ACTIONS(1706),
    [anon_sym_RPAREN] = ACTIONS(1706),
    [anon_sym_COMMA] = ACTIONS(1706),
    [anon_sym_PIPE] = ACTIONS(1704),
    [anon_sym_COLON2] = ACTIONS(1704),
    [anon_sym_DOT] = ACTIONS(1706),
    [anon_sym_DASH_GT] = ACTIONS(1704),
    [anon_sym_LBRACK] = ACTIONS(1704),
    [anon_sym_RBRACK] = ACTIONS(1706),
    [anon_sym_true] = ACTIONS(1704),
    [anon_sym_false] = ACTIONS(1704),
    [anon_sym_LBRACE] = ACTIONS(1704),
    [anon_sym_SEMI] = ACTIONS(1706),
    [anon_sym_RBRACE] = ACTIONS(1706),
    [anon_sym_constraint] = ACTIONS(1704),
    [anon_sym_val] = ACTIONS(1704),
    [anon_sym_end] = ACTIONS(1704),
    [anon_sym_with] = ACTIONS(1704),
    [anon_sym_COLON_GT] = ACTIONS(1706),
    [anon_sym_object] = ACTIONS(1704),
    [anon_sym_inherit] = ACTIONS(1704),
    [anon_sym_method] = ACTIONS(1704),
    [anon_sym_initializer] = ACTIONS(1704),
    [anon_sym_AMP] = ACTIONS(1704),
    [anon_sym_POUND] = ACTIONS(1704),
    [anon_sym_COLON_COLON] = ACTIONS(1706),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1706),
    [anon_sym_then] = ACTIONS(1704),
    [anon_sym_else] = ACTIONS(1704),
    [anon_sym_do] = ACTIONS(1704),
    [anon_sym_done] = ACTIONS(1704),
    [anon_sym_to] = ACTIONS(1704),
    [anon_sym_downto] = ACTIONS(1704),
    [anon_sym_new] = ACTIONS(1704),
    [anon_sym_LBRACE_LT] = ACTIONS(1706),
    [anon_sym_GT_RBRACE] = ACTIONS(1706),
    [anon_sym_begin] = ACTIONS(1704),
    [sym_ocamlyacc_value] = ACTIONS(1706),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1704),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1706),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1704),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1704),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1706),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1706),
    [aux_sym_number_token1] = ACTIONS(1706),
    [aux_sym_signed_number_token1] = ACTIONS(1704),
    [anon_sym_SQUOTE] = ACTIONS(1706),
    [anon_sym_DQUOTE] = ACTIONS(1706),
    [sym_prefix_operator] = ACTIONS(1706),
    [aux_sym_sign_operator_token1] = ACTIONS(1704),
    [sym_hash_operator] = ACTIONS(1706),
    [anon_sym_lsl] = ACTIONS(1704),
    [anon_sym_lsr] = ACTIONS(1704),
    [anon_sym_asr] = ACTIONS(1704),
    [aux_sym_pow_operator_token1] = ACTIONS(1706),
    [anon_sym_mod] = ACTIONS(1704),
    [anon_sym_land] = ACTIONS(1704),
    [anon_sym_lor] = ACTIONS(1704),
    [anon_sym_lxor] = ACTIONS(1704),
    [aux_sym_mult_operator_token1] = ACTIONS(1704),
    [aux_sym_add_operator_token1] = ACTIONS(1704),
    [sym_concat_operator] = ACTIONS(1706),
    [sym_rel_operator] = ACTIONS(1704),
    [anon_sym_AMP_AMP] = ACTIONS(1704),
    [anon_sym_or] = ACTIONS(1704),
    [anon_sym_PIPE_PIPE] = ACTIONS(1704),
    [sym_assign_operator] = ACTIONS(1706),
    [sym__uppercase_identifier] = ACTIONS(1706),
    [aux_sym_tag_token1] = ACTIONS(1706),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(923)] = {
    [sym_attribute] = STATE(923),
    [sym__lowercase_identifier] = ACTIONS(1622),
    [anon_sym_TILDE] = ACTIONS(1622),
    [anon_sym_QMARK] = ACTIONS(1622),
    [anon_sym_LPAREN] = ACTIONS(1624),
    [anon_sym_RPAREN] = ACTIONS(1624),
    [anon_sym_COMMA] = ACTIONS(1624),
    [anon_sym_PIPE] = ACTIONS(1622),
    [anon_sym_COLON2] = ACTIONS(1622),
    [anon_sym_DOT] = ACTIONS(1624),
    [anon_sym_DASH_GT] = ACTIONS(1622),
    [anon_sym_LBRACK] = ACTIONS(1622),
    [anon_sym_RBRACK] = ACTIONS(1624),
    [anon_sym_true] = ACTIONS(1622),
    [anon_sym_false] = ACTIONS(1622),
    [anon_sym_LBRACE] = ACTIONS(1622),
    [anon_sym_SEMI] = ACTIONS(1624),
    [anon_sym_RBRACE] = ACTIONS(1624),
    [anon_sym_constraint] = ACTIONS(1622),
    [anon_sym_val] = ACTIONS(1622),
    [anon_sym_end] = ACTIONS(1622),
    [anon_sym_with] = ACTIONS(1622),
    [anon_sym_COLON_GT] = ACTIONS(1624),
    [anon_sym_object] = ACTIONS(1622),
    [anon_sym_inherit] = ACTIONS(1622),
    [anon_sym_method] = ACTIONS(1622),
    [anon_sym_initializer] = ACTIONS(1622),
    [anon_sym_AMP] = ACTIONS(1622),
    [anon_sym_POUND] = ACTIONS(1622),
    [anon_sym_COLON_COLON] = ACTIONS(1624),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1624),
    [anon_sym_then] = ACTIONS(1622),
    [anon_sym_else] = ACTIONS(1622),
    [anon_sym_do] = ACTIONS(1622),
    [anon_sym_done] = ACTIONS(1622),
    [anon_sym_to] = ACTIONS(1622),
    [anon_sym_downto] = ACTIONS(1622),
    [anon_sym_new] = ACTIONS(1622),
    [anon_sym_LBRACE_LT] = ACTIONS(1624),
    [anon_sym_GT_RBRACE] = ACTIONS(1624),
    [anon_sym_begin] = ACTIONS(1622),
    [sym_ocamlyacc_value] = ACTIONS(1624),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1622),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1624),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1622),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1622),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1624),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1624),
    [aux_sym_number_token1] = ACTIONS(1624),
    [aux_sym_signed_number_token1] = ACTIONS(1622),
    [anon_sym_SQUOTE] = ACTIONS(1624),
    [anon_sym_DQUOTE] = ACTIONS(1624),
    [sym_prefix_operator] = ACTIONS(1624),
    [aux_sym_sign_operator_token1] = ACTIONS(1622),
    [sym_hash_operator] = ACTIONS(1624),
    [anon_sym_lsl] = ACTIONS(1622),
    [anon_sym_lsr] = ACTIONS(1622),
    [anon_sym_asr] = ACTIONS(1622),
    [aux_sym_pow_operator_token1] = ACTIONS(1624),
    [anon_sym_mod] = ACTIONS(1622),
    [anon_sym_land] = ACTIONS(1622),
    [anon_sym_lor] = ACTIONS(1622),
    [anon_sym_lxor] = ACTIONS(1622),
    [aux_sym_mult_operator_token1] = ACTIONS(1622),
    [aux_sym_add_operator_token1] = ACTIONS(1622),
    [sym_concat_operator] = ACTIONS(1624),
    [sym_rel_operator] = ACTIONS(1622),
    [anon_sym_AMP_AMP] = ACTIONS(1622),
    [anon_sym_or] = ACTIONS(1622),
    [anon_sym_PIPE_PIPE] = ACTIONS(1622),
    [sym_assign_operator] = ACTIONS(1624),
    [sym__uppercase_identifier] = ACTIONS(1624),
    [aux_sym_tag_token1] = ACTIONS(1624),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(924)] = {
    [sym_attribute] = STATE(924),
    [sym__lowercase_identifier] = ACTIONS(1634),
    [anon_sym_TILDE] = ACTIONS(1634),
    [anon_sym_QMARK] = ACTIONS(1634),
    [anon_sym_LPAREN] = ACTIONS(1636),
    [anon_sym_RPAREN] = ACTIONS(1636),
    [anon_sym_COMMA] = ACTIONS(1636),
    [anon_sym_PIPE] = ACTIONS(1634),
    [anon_sym_COLON2] = ACTIONS(1634),
    [anon_sym_DOT] = ACTIONS(1636),
    [anon_sym_DASH_GT] = ACTIONS(1634),
    [anon_sym_LBRACK] = ACTIONS(1634),
    [anon_sym_RBRACK] = ACTIONS(1636),
    [anon_sym_true] = ACTIONS(1634),
    [anon_sym_false] = ACTIONS(1634),
    [anon_sym_LBRACE] = ACTIONS(1634),
    [anon_sym_SEMI] = ACTIONS(1636),
    [anon_sym_RBRACE] = ACTIONS(1636),
    [anon_sym_constraint] = ACTIONS(1634),
    [anon_sym_val] = ACTIONS(1634),
    [anon_sym_end] = ACTIONS(1634),
    [anon_sym_with] = ACTIONS(1634),
    [anon_sym_COLON_GT] = ACTIONS(1636),
    [anon_sym_object] = ACTIONS(1634),
    [anon_sym_inherit] = ACTIONS(1634),
    [anon_sym_method] = ACTIONS(1634),
    [anon_sym_initializer] = ACTIONS(1634),
    [anon_sym_AMP] = ACTIONS(1634),
    [anon_sym_POUND] = ACTIONS(1634),
    [anon_sym_COLON_COLON] = ACTIONS(1636),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1636),
    [anon_sym_then] = ACTIONS(1634),
    [anon_sym_else] = ACTIONS(1634),
    [anon_sym_do] = ACTIONS(1634),
    [anon_sym_done] = ACTIONS(1634),
    [anon_sym_to] = ACTIONS(1634),
    [anon_sym_downto] = ACTIONS(1634),
    [anon_sym_new] = ACTIONS(1634),
    [anon_sym_LBRACE_LT] = ACTIONS(1636),
    [anon_sym_GT_RBRACE] = ACTIONS(1636),
    [anon_sym_begin] = ACTIONS(1634),
    [sym_ocamlyacc_value] = ACTIONS(1636),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1634),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1636),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1634),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1634),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1636),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1636),
    [aux_sym_number_token1] = ACTIONS(1636),
    [aux_sym_signed_number_token1] = ACTIONS(1634),
    [anon_sym_SQUOTE] = ACTIONS(1636),
    [anon_sym_DQUOTE] = ACTIONS(1636),
    [sym_prefix_operator] = ACTIONS(1636),
    [aux_sym_sign_operator_token1] = ACTIONS(1634),
    [sym_hash_operator] = ACTIONS(1636),
    [anon_sym_lsl] = ACTIONS(1634),
    [anon_sym_lsr] = ACTIONS(1634),
    [anon_sym_asr] = ACTIONS(1634),
    [aux_sym_pow_operator_token1] = ACTIONS(1636),
    [anon_sym_mod] = ACTIONS(1634),
    [anon_sym_land] = ACTIONS(1634),
    [anon_sym_lor] = ACTIONS(1634),
    [anon_sym_lxor] = ACTIONS(1634),
    [aux_sym_mult_operator_token1] = ACTIONS(1634),
    [aux_sym_add_operator_token1] = ACTIONS(1634),
    [sym_concat_operator] = ACTIONS(1636),
    [sym_rel_operator] = ACTIONS(1634),
    [anon_sym_AMP_AMP] = ACTIONS(1634),
    [anon_sym_or] = ACTIONS(1634),
    [anon_sym_PIPE_PIPE] = ACTIONS(1634),
    [sym_assign_operator] = ACTIONS(1636),
    [sym__uppercase_identifier] = ACTIONS(1636),
    [aux_sym_tag_token1] = ACTIONS(1636),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(925)] = {
    [sym_attribute] = STATE(925),
    [sym__lowercase_identifier] = ACTIONS(1750),
    [anon_sym_TILDE] = ACTIONS(1750),
    [anon_sym_QMARK] = ACTIONS(1750),
    [anon_sym_LPAREN] = ACTIONS(1752),
    [anon_sym_RPAREN] = ACTIONS(1752),
    [anon_sym_COMMA] = ACTIONS(1752),
    [anon_sym_PIPE] = ACTIONS(1750),
    [anon_sym_COLON2] = ACTIONS(1750),
    [anon_sym_DOT] = ACTIONS(1752),
    [anon_sym_DASH_GT] = ACTIONS(1750),
    [anon_sym_LBRACK] = ACTIONS(1750),
    [anon_sym_RBRACK] = ACTIONS(1752),
    [anon_sym_true] = ACTIONS(1750),
    [anon_sym_false] = ACTIONS(1750),
    [anon_sym_LBRACE] = ACTIONS(1750),
    [anon_sym_SEMI] = ACTIONS(1752),
    [anon_sym_RBRACE] = ACTIONS(1752),
    [anon_sym_constraint] = ACTIONS(1750),
    [anon_sym_val] = ACTIONS(1750),
    [anon_sym_end] = ACTIONS(1750),
    [anon_sym_with] = ACTIONS(1750),
    [anon_sym_COLON_GT] = ACTIONS(1752),
    [anon_sym_object] = ACTIONS(1750),
    [anon_sym_inherit] = ACTIONS(1750),
    [anon_sym_method] = ACTIONS(1750),
    [anon_sym_initializer] = ACTIONS(1750),
    [anon_sym_AMP] = ACTIONS(1750),
    [anon_sym_POUND] = ACTIONS(1750),
    [anon_sym_COLON_COLON] = ACTIONS(1752),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1752),
    [anon_sym_then] = ACTIONS(1750),
    [anon_sym_else] = ACTIONS(1750),
    [anon_sym_do] = ACTIONS(1750),
    [anon_sym_done] = ACTIONS(1750),
    [anon_sym_to] = ACTIONS(1750),
    [anon_sym_downto] = ACTIONS(1750),
    [anon_sym_new] = ACTIONS(1750),
    [anon_sym_LBRACE_LT] = ACTIONS(1752),
    [anon_sym_GT_RBRACE] = ACTIONS(1752),
    [anon_sym_begin] = ACTIONS(1750),
    [sym_ocamlyacc_value] = ACTIONS(1752),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1750),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1752),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1750),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1750),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1752),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1752),
    [aux_sym_number_token1] = ACTIONS(1752),
    [aux_sym_signed_number_token1] = ACTIONS(1750),
    [anon_sym_SQUOTE] = ACTIONS(1752),
    [anon_sym_DQUOTE] = ACTIONS(1752),
    [sym_prefix_operator] = ACTIONS(1752),
    [aux_sym_sign_operator_token1] = ACTIONS(1750),
    [sym_hash_operator] = ACTIONS(1752),
    [anon_sym_lsl] = ACTIONS(1750),
    [anon_sym_lsr] = ACTIONS(1750),
    [anon_sym_asr] = ACTIONS(1750),
    [aux_sym_pow_operator_token1] = ACTIONS(1752),
    [anon_sym_mod] = ACTIONS(1750),
    [anon_sym_land] = ACTIONS(1750),
    [anon_sym_lor] = ACTIONS(1750),
    [anon_sym_lxor] = ACTIONS(1750),
    [aux_sym_mult_operator_token1] = ACTIONS(1750),
    [aux_sym_add_operator_token1] = ACTIONS(1750),
    [sym_concat_operator] = ACTIONS(1752),
    [sym_rel_operator] = ACTIONS(1750),
    [anon_sym_AMP_AMP] = ACTIONS(1750),
    [anon_sym_or] = ACTIONS(1750),
    [anon_sym_PIPE_PIPE] = ACTIONS(1750),
    [sym_assign_operator] = ACTIONS(1752),
    [sym__uppercase_identifier] = ACTIONS(1752),
    [aux_sym_tag_token1] = ACTIONS(1752),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(926)] = {
    [sym_attribute] = STATE(926),
    [sym__lowercase_identifier] = ACTIONS(1842),
    [anon_sym_TILDE] = ACTIONS(1842),
    [anon_sym_QMARK] = ACTIONS(1842),
    [anon_sym_LPAREN] = ACTIONS(1844),
    [anon_sym_RPAREN] = ACTIONS(1844),
    [anon_sym_COMMA] = ACTIONS(1844),
    [anon_sym_PIPE] = ACTIONS(1842),
    [anon_sym_COLON2] = ACTIONS(1842),
    [anon_sym_DOT] = ACTIONS(1844),
    [anon_sym_DASH_GT] = ACTIONS(1842),
    [anon_sym_LBRACK] = ACTIONS(1842),
    [anon_sym_RBRACK] = ACTIONS(1844),
    [anon_sym_true] = ACTIONS(1842),
    [anon_sym_false] = ACTIONS(1842),
    [anon_sym_LBRACE] = ACTIONS(1842),
    [anon_sym_SEMI] = ACTIONS(1844),
    [anon_sym_RBRACE] = ACTIONS(1844),
    [anon_sym_constraint] = ACTIONS(1842),
    [anon_sym_val] = ACTIONS(1842),
    [anon_sym_end] = ACTIONS(1842),
    [anon_sym_with] = ACTIONS(1842),
    [anon_sym_COLON_GT] = ACTIONS(1844),
    [anon_sym_object] = ACTIONS(1842),
    [anon_sym_inherit] = ACTIONS(1842),
    [anon_sym_method] = ACTIONS(1842),
    [anon_sym_initializer] = ACTIONS(1842),
    [anon_sym_AMP] = ACTIONS(1842),
    [anon_sym_POUND] = ACTIONS(1842),
    [anon_sym_COLON_COLON] = ACTIONS(1844),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1844),
    [anon_sym_then] = ACTIONS(1842),
    [anon_sym_else] = ACTIONS(1842),
    [anon_sym_do] = ACTIONS(1842),
    [anon_sym_done] = ACTIONS(1842),
    [anon_sym_to] = ACTIONS(1842),
    [anon_sym_downto] = ACTIONS(1842),
    [anon_sym_new] = ACTIONS(1842),
    [anon_sym_LBRACE_LT] = ACTIONS(1844),
    [anon_sym_GT_RBRACE] = ACTIONS(1844),
    [anon_sym_begin] = ACTIONS(1842),
    [sym_ocamlyacc_value] = ACTIONS(1844),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1842),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1844),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1842),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1842),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1844),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1844),
    [aux_sym_number_token1] = ACTIONS(1844),
    [aux_sym_signed_number_token1] = ACTIONS(1842),
    [anon_sym_SQUOTE] = ACTIONS(1844),
    [anon_sym_DQUOTE] = ACTIONS(1844),
    [sym_prefix_operator] = ACTIONS(1844),
    [aux_sym_sign_operator_token1] = ACTIONS(1842),
    [sym_hash_operator] = ACTIONS(1844),
    [anon_sym_lsl] = ACTIONS(1842),
    [anon_sym_lsr] = ACTIONS(1842),
    [anon_sym_asr] = ACTIONS(1842),
    [aux_sym_pow_operator_token1] = ACTIONS(1844),
    [anon_sym_mod] = ACTIONS(1842),
    [anon_sym_land] = ACTIONS(1842),
    [anon_sym_lor] = ACTIONS(1842),
    [anon_sym_lxor] = ACTIONS(1842),
    [aux_sym_mult_operator_token1] = ACTIONS(1842),
    [aux_sym_add_operator_token1] = ACTIONS(1842),
    [sym_concat_operator] = ACTIONS(1844),
    [sym_rel_operator] = ACTIONS(1842),
    [anon_sym_AMP_AMP] = ACTIONS(1842),
    [anon_sym_or] = ACTIONS(1842),
    [anon_sym_PIPE_PIPE] = ACTIONS(1842),
    [sym_assign_operator] = ACTIONS(1844),
    [sym__uppercase_identifier] = ACTIONS(1844),
    [aux_sym_tag_token1] = ACTIONS(1844),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(927)] = {
    [sym_attribute] = STATE(927),
    [sym__lowercase_identifier] = ACTIONS(1822),
    [anon_sym_TILDE] = ACTIONS(1822),
    [anon_sym_QMARK] = ACTIONS(1822),
    [anon_sym_LPAREN] = ACTIONS(1824),
    [anon_sym_RPAREN] = ACTIONS(1824),
    [anon_sym_COMMA] = ACTIONS(1824),
    [anon_sym_PIPE] = ACTIONS(1822),
    [anon_sym_COLON2] = ACTIONS(1822),
    [anon_sym_DOT] = ACTIONS(1824),
    [anon_sym_DASH_GT] = ACTIONS(1822),
    [anon_sym_LBRACK] = ACTIONS(1822),
    [anon_sym_RBRACK] = ACTIONS(1824),
    [anon_sym_true] = ACTIONS(1822),
    [anon_sym_false] = ACTIONS(1822),
    [anon_sym_LBRACE] = ACTIONS(1822),
    [anon_sym_SEMI] = ACTIONS(1824),
    [anon_sym_RBRACE] = ACTIONS(1824),
    [anon_sym_constraint] = ACTIONS(1822),
    [anon_sym_val] = ACTIONS(1822),
    [anon_sym_end] = ACTIONS(1822),
    [anon_sym_with] = ACTIONS(1822),
    [anon_sym_COLON_GT] = ACTIONS(1824),
    [anon_sym_object] = ACTIONS(1822),
    [anon_sym_inherit] = ACTIONS(1822),
    [anon_sym_method] = ACTIONS(1822),
    [anon_sym_initializer] = ACTIONS(1822),
    [anon_sym_AMP] = ACTIONS(1822),
    [anon_sym_POUND] = ACTIONS(1822),
    [anon_sym_COLON_COLON] = ACTIONS(1824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1824),
    [anon_sym_then] = ACTIONS(1822),
    [anon_sym_else] = ACTIONS(1822),
    [anon_sym_do] = ACTIONS(1822),
    [anon_sym_done] = ACTIONS(1822),
    [anon_sym_to] = ACTIONS(1822),
    [anon_sym_downto] = ACTIONS(1822),
    [anon_sym_new] = ACTIONS(1822),
    [anon_sym_LBRACE_LT] = ACTIONS(1824),
    [anon_sym_GT_RBRACE] = ACTIONS(1824),
    [anon_sym_begin] = ACTIONS(1822),
    [sym_ocamlyacc_value] = ACTIONS(1824),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1822),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1824),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1822),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1822),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1824),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1824),
    [aux_sym_number_token1] = ACTIONS(1824),
    [aux_sym_signed_number_token1] = ACTIONS(1822),
    [anon_sym_SQUOTE] = ACTIONS(1824),
    [anon_sym_DQUOTE] = ACTIONS(1824),
    [sym_prefix_operator] = ACTIONS(1824),
    [aux_sym_sign_operator_token1] = ACTIONS(1822),
    [sym_hash_operator] = ACTIONS(1824),
    [anon_sym_lsl] = ACTIONS(1822),
    [anon_sym_lsr] = ACTIONS(1822),
    [anon_sym_asr] = ACTIONS(1822),
    [aux_sym_pow_operator_token1] = ACTIONS(1824),
    [anon_sym_mod] = ACTIONS(1822),
    [anon_sym_land] = ACTIONS(1822),
    [anon_sym_lor] = ACTIONS(1822),
    [anon_sym_lxor] = ACTIONS(1822),
    [aux_sym_mult_operator_token1] = ACTIONS(1822),
    [aux_sym_add_operator_token1] = ACTIONS(1822),
    [sym_concat_operator] = ACTIONS(1824),
    [sym_rel_operator] = ACTIONS(1822),
    [anon_sym_AMP_AMP] = ACTIONS(1822),
    [anon_sym_or] = ACTIONS(1822),
    [anon_sym_PIPE_PIPE] = ACTIONS(1822),
    [sym_assign_operator] = ACTIONS(1824),
    [sym__uppercase_identifier] = ACTIONS(1824),
    [aux_sym_tag_token1] = ACTIONS(1824),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(928)] = {
    [sym_attribute] = STATE(928),
    [sym__lowercase_identifier] = ACTIONS(1846),
    [anon_sym_TILDE] = ACTIONS(1846),
    [anon_sym_QMARK] = ACTIONS(1846),
    [anon_sym_LPAREN] = ACTIONS(1848),
    [anon_sym_RPAREN] = ACTIONS(1848),
    [anon_sym_COMMA] = ACTIONS(1848),
    [anon_sym_PIPE] = ACTIONS(1846),
    [anon_sym_COLON2] = ACTIONS(1846),
    [anon_sym_DOT] = ACTIONS(1848),
    [anon_sym_DASH_GT] = ACTIONS(1846),
    [anon_sym_LBRACK] = ACTIONS(1846),
    [anon_sym_RBRACK] = ACTIONS(1848),
    [anon_sym_true] = ACTIONS(1846),
    [anon_sym_false] = ACTIONS(1846),
    [anon_sym_LBRACE] = ACTIONS(1846),
    [anon_sym_SEMI] = ACTIONS(1848),
    [anon_sym_RBRACE] = ACTIONS(1848),
    [anon_sym_constraint] = ACTIONS(1846),
    [anon_sym_val] = ACTIONS(1846),
    [anon_sym_end] = ACTIONS(1846),
    [anon_sym_with] = ACTIONS(1846),
    [anon_sym_COLON_GT] = ACTIONS(1848),
    [anon_sym_object] = ACTIONS(1846),
    [anon_sym_inherit] = ACTIONS(1846),
    [anon_sym_method] = ACTIONS(1846),
    [anon_sym_initializer] = ACTIONS(1846),
    [anon_sym_AMP] = ACTIONS(1846),
    [anon_sym_POUND] = ACTIONS(1846),
    [anon_sym_COLON_COLON] = ACTIONS(1848),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1848),
    [anon_sym_then] = ACTIONS(1846),
    [anon_sym_else] = ACTIONS(1846),
    [anon_sym_do] = ACTIONS(1846),
    [anon_sym_done] = ACTIONS(1846),
    [anon_sym_to] = ACTIONS(1846),
    [anon_sym_downto] = ACTIONS(1846),
    [anon_sym_new] = ACTIONS(1846),
    [anon_sym_LBRACE_LT] = ACTIONS(1848),
    [anon_sym_GT_RBRACE] = ACTIONS(1848),
    [anon_sym_begin] = ACTIONS(1846),
    [sym_ocamlyacc_value] = ACTIONS(1848),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1846),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1848),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1846),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1846),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1848),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1848),
    [aux_sym_number_token1] = ACTIONS(1848),
    [aux_sym_signed_number_token1] = ACTIONS(1846),
    [anon_sym_SQUOTE] = ACTIONS(1848),
    [anon_sym_DQUOTE] = ACTIONS(1848),
    [sym_prefix_operator] = ACTIONS(1848),
    [aux_sym_sign_operator_token1] = ACTIONS(1846),
    [sym_hash_operator] = ACTIONS(1848),
    [anon_sym_lsl] = ACTIONS(1846),
    [anon_sym_lsr] = ACTIONS(1846),
    [anon_sym_asr] = ACTIONS(1846),
    [aux_sym_pow_operator_token1] = ACTIONS(1848),
    [anon_sym_mod] = ACTIONS(1846),
    [anon_sym_land] = ACTIONS(1846),
    [anon_sym_lor] = ACTIONS(1846),
    [anon_sym_lxor] = ACTIONS(1846),
    [aux_sym_mult_operator_token1] = ACTIONS(1846),
    [aux_sym_add_operator_token1] = ACTIONS(1846),
    [sym_concat_operator] = ACTIONS(1848),
    [sym_rel_operator] = ACTIONS(1846),
    [anon_sym_AMP_AMP] = ACTIONS(1846),
    [anon_sym_or] = ACTIONS(1846),
    [anon_sym_PIPE_PIPE] = ACTIONS(1846),
    [sym_assign_operator] = ACTIONS(1848),
    [sym__uppercase_identifier] = ACTIONS(1848),
    [aux_sym_tag_token1] = ACTIONS(1848),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(929)] = {
    [sym_attribute] = STATE(929),
    [sym__lowercase_identifier] = ACTIONS(1926),
    [anon_sym_TILDE] = ACTIONS(1926),
    [anon_sym_QMARK] = ACTIONS(1926),
    [anon_sym_LPAREN] = ACTIONS(1928),
    [anon_sym_RPAREN] = ACTIONS(1928),
    [anon_sym_COMMA] = ACTIONS(1928),
    [anon_sym_PIPE] = ACTIONS(1926),
    [anon_sym_COLON2] = ACTIONS(1926),
    [anon_sym_DOT] = ACTIONS(1928),
    [anon_sym_DASH_GT] = ACTIONS(1926),
    [anon_sym_LBRACK] = ACTIONS(1926),
    [anon_sym_RBRACK] = ACTIONS(1928),
    [anon_sym_true] = ACTIONS(1926),
    [anon_sym_false] = ACTIONS(1926),
    [anon_sym_LBRACE] = ACTIONS(1926),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym_RBRACE] = ACTIONS(1928),
    [anon_sym_constraint] = ACTIONS(1926),
    [anon_sym_val] = ACTIONS(1926),
    [anon_sym_end] = ACTIONS(1926),
    [anon_sym_with] = ACTIONS(1926),
    [anon_sym_COLON_GT] = ACTIONS(1928),
    [anon_sym_object] = ACTIONS(1926),
    [anon_sym_inherit] = ACTIONS(1926),
    [anon_sym_method] = ACTIONS(1926),
    [anon_sym_initializer] = ACTIONS(1926),
    [anon_sym_AMP] = ACTIONS(1926),
    [anon_sym_POUND] = ACTIONS(1926),
    [anon_sym_COLON_COLON] = ACTIONS(1928),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1928),
    [anon_sym_then] = ACTIONS(1926),
    [anon_sym_else] = ACTIONS(1926),
    [anon_sym_do] = ACTIONS(1926),
    [anon_sym_done] = ACTIONS(1926),
    [anon_sym_to] = ACTIONS(1926),
    [anon_sym_downto] = ACTIONS(1926),
    [anon_sym_new] = ACTIONS(1926),
    [anon_sym_LBRACE_LT] = ACTIONS(1928),
    [anon_sym_GT_RBRACE] = ACTIONS(1928),
    [anon_sym_begin] = ACTIONS(1926),
    [sym_ocamlyacc_value] = ACTIONS(1928),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1926),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1928),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1926),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1926),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1928),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1928),
    [aux_sym_number_token1] = ACTIONS(1928),
    [aux_sym_signed_number_token1] = ACTIONS(1926),
    [anon_sym_SQUOTE] = ACTIONS(1928),
    [anon_sym_DQUOTE] = ACTIONS(1928),
    [sym_prefix_operator] = ACTIONS(1928),
    [aux_sym_sign_operator_token1] = ACTIONS(1926),
    [sym_hash_operator] = ACTIONS(1928),
    [anon_sym_lsl] = ACTIONS(1926),
    [anon_sym_lsr] = ACTIONS(1926),
    [anon_sym_asr] = ACTIONS(1926),
    [aux_sym_pow_operator_token1] = ACTIONS(1928),
    [anon_sym_mod] = ACTIONS(1926),
    [anon_sym_land] = ACTIONS(1926),
    [anon_sym_lor] = ACTIONS(1926),
    [anon_sym_lxor] = ACTIONS(1926),
    [aux_sym_mult_operator_token1] = ACTIONS(1926),
    [aux_sym_add_operator_token1] = ACTIONS(1926),
    [sym_concat_operator] = ACTIONS(1928),
    [sym_rel_operator] = ACTIONS(1926),
    [anon_sym_AMP_AMP] = ACTIONS(1926),
    [anon_sym_or] = ACTIONS(1926),
    [anon_sym_PIPE_PIPE] = ACTIONS(1926),
    [sym_assign_operator] = ACTIONS(1928),
    [sym__uppercase_identifier] = ACTIONS(1928),
    [aux_sym_tag_token1] = ACTIONS(1928),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(930)] = {
    [sym_attribute] = STATE(930),
    [sym__lowercase_identifier] = ACTIONS(1930),
    [anon_sym_TILDE] = ACTIONS(1930),
    [anon_sym_QMARK] = ACTIONS(1930),
    [anon_sym_LPAREN] = ACTIONS(1932),
    [anon_sym_RPAREN] = ACTIONS(1932),
    [anon_sym_COMMA] = ACTIONS(1932),
    [anon_sym_PIPE] = ACTIONS(1930),
    [anon_sym_COLON2] = ACTIONS(1930),
    [anon_sym_DOT] = ACTIONS(1932),
    [anon_sym_DASH_GT] = ACTIONS(1930),
    [anon_sym_LBRACK] = ACTIONS(1930),
    [anon_sym_RBRACK] = ACTIONS(1932),
    [anon_sym_true] = ACTIONS(1930),
    [anon_sym_false] = ACTIONS(1930),
    [anon_sym_LBRACE] = ACTIONS(1930),
    [anon_sym_SEMI] = ACTIONS(1932),
    [anon_sym_RBRACE] = ACTIONS(1932),
    [anon_sym_constraint] = ACTIONS(1930),
    [anon_sym_val] = ACTIONS(1930),
    [anon_sym_end] = ACTIONS(1930),
    [anon_sym_with] = ACTIONS(1930),
    [anon_sym_COLON_GT] = ACTIONS(1932),
    [anon_sym_object] = ACTIONS(1930),
    [anon_sym_inherit] = ACTIONS(1930),
    [anon_sym_method] = ACTIONS(1930),
    [anon_sym_initializer] = ACTIONS(1930),
    [anon_sym_AMP] = ACTIONS(1930),
    [anon_sym_POUND] = ACTIONS(1930),
    [anon_sym_COLON_COLON] = ACTIONS(1932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1932),
    [anon_sym_then] = ACTIONS(1930),
    [anon_sym_else] = ACTIONS(1930),
    [anon_sym_do] = ACTIONS(1930),
    [anon_sym_done] = ACTIONS(1930),
    [anon_sym_to] = ACTIONS(1930),
    [anon_sym_downto] = ACTIONS(1930),
    [anon_sym_new] = ACTIONS(1930),
    [anon_sym_LBRACE_LT] = ACTIONS(1932),
    [anon_sym_GT_RBRACE] = ACTIONS(1932),
    [anon_sym_begin] = ACTIONS(1930),
    [sym_ocamlyacc_value] = ACTIONS(1932),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1930),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1932),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1930),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1930),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1932),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1932),
    [aux_sym_number_token1] = ACTIONS(1932),
    [aux_sym_signed_number_token1] = ACTIONS(1930),
    [anon_sym_SQUOTE] = ACTIONS(1932),
    [anon_sym_DQUOTE] = ACTIONS(1932),
    [sym_prefix_operator] = ACTIONS(1932),
    [aux_sym_sign_operator_token1] = ACTIONS(1930),
    [sym_hash_operator] = ACTIONS(1932),
    [anon_sym_lsl] = ACTIONS(1930),
    [anon_sym_lsr] = ACTIONS(1930),
    [anon_sym_asr] = ACTIONS(1930),
    [aux_sym_pow_operator_token1] = ACTIONS(1932),
    [anon_sym_mod] = ACTIONS(1930),
    [anon_sym_land] = ACTIONS(1930),
    [anon_sym_lor] = ACTIONS(1930),
    [anon_sym_lxor] = ACTIONS(1930),
    [aux_sym_mult_operator_token1] = ACTIONS(1930),
    [aux_sym_add_operator_token1] = ACTIONS(1930),
    [sym_concat_operator] = ACTIONS(1932),
    [sym_rel_operator] = ACTIONS(1930),
    [anon_sym_AMP_AMP] = ACTIONS(1930),
    [anon_sym_or] = ACTIONS(1930),
    [anon_sym_PIPE_PIPE] = ACTIONS(1930),
    [sym_assign_operator] = ACTIONS(1932),
    [sym__uppercase_identifier] = ACTIONS(1932),
    [aux_sym_tag_token1] = ACTIONS(1932),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(931)] = {
    [sym_attribute] = STATE(931),
    [sym__lowercase_identifier] = ACTIONS(1934),
    [anon_sym_TILDE] = ACTIONS(1934),
    [anon_sym_QMARK] = ACTIONS(1934),
    [anon_sym_LPAREN] = ACTIONS(1936),
    [anon_sym_RPAREN] = ACTIONS(1936),
    [anon_sym_COMMA] = ACTIONS(1936),
    [anon_sym_PIPE] = ACTIONS(1934),
    [anon_sym_COLON2] = ACTIONS(1934),
    [anon_sym_DOT] = ACTIONS(1936),
    [anon_sym_DASH_GT] = ACTIONS(1934),
    [anon_sym_LBRACK] = ACTIONS(1934),
    [anon_sym_RBRACK] = ACTIONS(1936),
    [anon_sym_true] = ACTIONS(1934),
    [anon_sym_false] = ACTIONS(1934),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_SEMI] = ACTIONS(1936),
    [anon_sym_RBRACE] = ACTIONS(1936),
    [anon_sym_constraint] = ACTIONS(1934),
    [anon_sym_val] = ACTIONS(1934),
    [anon_sym_end] = ACTIONS(1934),
    [anon_sym_with] = ACTIONS(1934),
    [anon_sym_COLON_GT] = ACTIONS(1936),
    [anon_sym_object] = ACTIONS(1934),
    [anon_sym_inherit] = ACTIONS(1934),
    [anon_sym_method] = ACTIONS(1934),
    [anon_sym_initializer] = ACTIONS(1934),
    [anon_sym_AMP] = ACTIONS(1934),
    [anon_sym_POUND] = ACTIONS(1934),
    [anon_sym_COLON_COLON] = ACTIONS(1936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1936),
    [anon_sym_then] = ACTIONS(1934),
    [anon_sym_else] = ACTIONS(1934),
    [anon_sym_do] = ACTIONS(1934),
    [anon_sym_done] = ACTIONS(1934),
    [anon_sym_to] = ACTIONS(1934),
    [anon_sym_downto] = ACTIONS(1934),
    [anon_sym_new] = ACTIONS(1934),
    [anon_sym_LBRACE_LT] = ACTIONS(1936),
    [anon_sym_GT_RBRACE] = ACTIONS(1936),
    [anon_sym_begin] = ACTIONS(1934),
    [sym_ocamlyacc_value] = ACTIONS(1936),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1934),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1934),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1934),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1936),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1936),
    [aux_sym_number_token1] = ACTIONS(1936),
    [aux_sym_signed_number_token1] = ACTIONS(1934),
    [anon_sym_SQUOTE] = ACTIONS(1936),
    [anon_sym_DQUOTE] = ACTIONS(1936),
    [sym_prefix_operator] = ACTIONS(1936),
    [aux_sym_sign_operator_token1] = ACTIONS(1934),
    [sym_hash_operator] = ACTIONS(1936),
    [anon_sym_lsl] = ACTIONS(1934),
    [anon_sym_lsr] = ACTIONS(1934),
    [anon_sym_asr] = ACTIONS(1934),
    [aux_sym_pow_operator_token1] = ACTIONS(1936),
    [anon_sym_mod] = ACTIONS(1934),
    [anon_sym_land] = ACTIONS(1934),
    [anon_sym_lor] = ACTIONS(1934),
    [anon_sym_lxor] = ACTIONS(1934),
    [aux_sym_mult_operator_token1] = ACTIONS(1934),
    [aux_sym_add_operator_token1] = ACTIONS(1934),
    [sym_concat_operator] = ACTIONS(1936),
    [sym_rel_operator] = ACTIONS(1934),
    [anon_sym_AMP_AMP] = ACTIONS(1934),
    [anon_sym_or] = ACTIONS(1934),
    [anon_sym_PIPE_PIPE] = ACTIONS(1934),
    [sym_assign_operator] = ACTIONS(1936),
    [sym__uppercase_identifier] = ACTIONS(1936),
    [aux_sym_tag_token1] = ACTIONS(1936),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(932)] = {
    [sym_attribute] = STATE(932),
    [sym__lowercase_identifier] = ACTIONS(1938),
    [anon_sym_TILDE] = ACTIONS(1938),
    [anon_sym_QMARK] = ACTIONS(1938),
    [anon_sym_LPAREN] = ACTIONS(1940),
    [anon_sym_RPAREN] = ACTIONS(1940),
    [anon_sym_COMMA] = ACTIONS(1940),
    [anon_sym_PIPE] = ACTIONS(1938),
    [anon_sym_COLON2] = ACTIONS(1938),
    [anon_sym_DOT] = ACTIONS(1940),
    [anon_sym_DASH_GT] = ACTIONS(1938),
    [anon_sym_LBRACK] = ACTIONS(1938),
    [anon_sym_RBRACK] = ACTIONS(1940),
    [anon_sym_true] = ACTIONS(1938),
    [anon_sym_false] = ACTIONS(1938),
    [anon_sym_LBRACE] = ACTIONS(1938),
    [anon_sym_SEMI] = ACTIONS(1940),
    [anon_sym_RBRACE] = ACTIONS(1940),
    [anon_sym_constraint] = ACTIONS(1938),
    [anon_sym_val] = ACTIONS(1938),
    [anon_sym_end] = ACTIONS(1938),
    [anon_sym_with] = ACTIONS(1938),
    [anon_sym_COLON_GT] = ACTIONS(1940),
    [anon_sym_object] = ACTIONS(1938),
    [anon_sym_inherit] = ACTIONS(1938),
    [anon_sym_method] = ACTIONS(1938),
    [anon_sym_initializer] = ACTIONS(1938),
    [anon_sym_AMP] = ACTIONS(1938),
    [anon_sym_POUND] = ACTIONS(1938),
    [anon_sym_COLON_COLON] = ACTIONS(1940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1940),
    [anon_sym_then] = ACTIONS(1938),
    [anon_sym_else] = ACTIONS(1938),
    [anon_sym_do] = ACTIONS(1938),
    [anon_sym_done] = ACTIONS(1938),
    [anon_sym_to] = ACTIONS(1938),
    [anon_sym_downto] = ACTIONS(1938),
    [anon_sym_new] = ACTIONS(1938),
    [anon_sym_LBRACE_LT] = ACTIONS(1940),
    [anon_sym_GT_RBRACE] = ACTIONS(1940),
    [anon_sym_begin] = ACTIONS(1938),
    [sym_ocamlyacc_value] = ACTIONS(1940),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1938),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1940),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1938),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1938),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1940),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1940),
    [aux_sym_number_token1] = ACTIONS(1940),
    [aux_sym_signed_number_token1] = ACTIONS(1938),
    [anon_sym_SQUOTE] = ACTIONS(1940),
    [anon_sym_DQUOTE] = ACTIONS(1940),
    [sym_prefix_operator] = ACTIONS(1940),
    [aux_sym_sign_operator_token1] = ACTIONS(1938),
    [sym_hash_operator] = ACTIONS(1940),
    [anon_sym_lsl] = ACTIONS(1938),
    [anon_sym_lsr] = ACTIONS(1938),
    [anon_sym_asr] = ACTIONS(1938),
    [aux_sym_pow_operator_token1] = ACTIONS(1940),
    [anon_sym_mod] = ACTIONS(1938),
    [anon_sym_land] = ACTIONS(1938),
    [anon_sym_lor] = ACTIONS(1938),
    [anon_sym_lxor] = ACTIONS(1938),
    [aux_sym_mult_operator_token1] = ACTIONS(1938),
    [aux_sym_add_operator_token1] = ACTIONS(1938),
    [sym_concat_operator] = ACTIONS(1940),
    [sym_rel_operator] = ACTIONS(1938),
    [anon_sym_AMP_AMP] = ACTIONS(1938),
    [anon_sym_or] = ACTIONS(1938),
    [anon_sym_PIPE_PIPE] = ACTIONS(1938),
    [sym_assign_operator] = ACTIONS(1940),
    [sym__uppercase_identifier] = ACTIONS(1940),
    [aux_sym_tag_token1] = ACTIONS(1940),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(933)] = {
    [sym_attribute] = STATE(933),
    [sym__lowercase_identifier] = ACTIONS(1650),
    [anon_sym_TILDE] = ACTIONS(1650),
    [anon_sym_QMARK] = ACTIONS(1650),
    [anon_sym_LPAREN] = ACTIONS(1652),
    [anon_sym_RPAREN] = ACTIONS(1652),
    [anon_sym_COMMA] = ACTIONS(1652),
    [anon_sym_PIPE] = ACTIONS(1650),
    [anon_sym_COLON2] = ACTIONS(1650),
    [anon_sym_DOT] = ACTIONS(1652),
    [anon_sym_DASH_GT] = ACTIONS(1650),
    [anon_sym_LBRACK] = ACTIONS(1650),
    [anon_sym_RBRACK] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1650),
    [anon_sym_false] = ACTIONS(1650),
    [anon_sym_LBRACE] = ACTIONS(1650),
    [anon_sym_SEMI] = ACTIONS(1652),
    [anon_sym_RBRACE] = ACTIONS(1652),
    [anon_sym_constraint] = ACTIONS(1650),
    [anon_sym_val] = ACTIONS(1650),
    [anon_sym_end] = ACTIONS(1650),
    [anon_sym_with] = ACTIONS(1650),
    [anon_sym_COLON_GT] = ACTIONS(1652),
    [anon_sym_object] = ACTIONS(1650),
    [anon_sym_inherit] = ACTIONS(1650),
    [anon_sym_method] = ACTIONS(1650),
    [anon_sym_initializer] = ACTIONS(1650),
    [anon_sym_AMP] = ACTIONS(1650),
    [anon_sym_POUND] = ACTIONS(1650),
    [anon_sym_COLON_COLON] = ACTIONS(1652),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1652),
    [anon_sym_then] = ACTIONS(1650),
    [anon_sym_else] = ACTIONS(1650),
    [anon_sym_do] = ACTIONS(1650),
    [anon_sym_done] = ACTIONS(1650),
    [anon_sym_to] = ACTIONS(1650),
    [anon_sym_downto] = ACTIONS(1650),
    [anon_sym_new] = ACTIONS(1650),
    [anon_sym_LBRACE_LT] = ACTIONS(1652),
    [anon_sym_GT_RBRACE] = ACTIONS(1652),
    [anon_sym_begin] = ACTIONS(1650),
    [sym_ocamlyacc_value] = ACTIONS(1652),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1650),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1652),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1650),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1650),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1652),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1652),
    [aux_sym_number_token1] = ACTIONS(1652),
    [aux_sym_signed_number_token1] = ACTIONS(1650),
    [anon_sym_SQUOTE] = ACTIONS(1652),
    [anon_sym_DQUOTE] = ACTIONS(1652),
    [sym_prefix_operator] = ACTIONS(1652),
    [aux_sym_sign_operator_token1] = ACTIONS(1650),
    [sym_hash_operator] = ACTIONS(1652),
    [anon_sym_lsl] = ACTIONS(1650),
    [anon_sym_lsr] = ACTIONS(1650),
    [anon_sym_asr] = ACTIONS(1650),
    [aux_sym_pow_operator_token1] = ACTIONS(1652),
    [anon_sym_mod] = ACTIONS(1650),
    [anon_sym_land] = ACTIONS(1650),
    [anon_sym_lor] = ACTIONS(1650),
    [anon_sym_lxor] = ACTIONS(1650),
    [aux_sym_mult_operator_token1] = ACTIONS(1650),
    [aux_sym_add_operator_token1] = ACTIONS(1650),
    [sym_concat_operator] = ACTIONS(1652),
    [sym_rel_operator] = ACTIONS(1650),
    [anon_sym_AMP_AMP] = ACTIONS(1650),
    [anon_sym_or] = ACTIONS(1650),
    [anon_sym_PIPE_PIPE] = ACTIONS(1650),
    [sym_assign_operator] = ACTIONS(1652),
    [sym__uppercase_identifier] = ACTIONS(1652),
    [aux_sym_tag_token1] = ACTIONS(1652),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(934)] = {
    [sym_attribute] = STATE(934),
    [sym__lowercase_identifier] = ACTIONS(1826),
    [anon_sym_TILDE] = ACTIONS(1826),
    [anon_sym_QMARK] = ACTIONS(1826),
    [anon_sym_LPAREN] = ACTIONS(1828),
    [anon_sym_RPAREN] = ACTIONS(1828),
    [anon_sym_COMMA] = ACTIONS(1828),
    [anon_sym_PIPE] = ACTIONS(1826),
    [anon_sym_COLON2] = ACTIONS(1826),
    [anon_sym_DOT] = ACTIONS(1828),
    [anon_sym_DASH_GT] = ACTIONS(1826),
    [anon_sym_LBRACK] = ACTIONS(1826),
    [anon_sym_RBRACK] = ACTIONS(1828),
    [anon_sym_true] = ACTIONS(1826),
    [anon_sym_false] = ACTIONS(1826),
    [anon_sym_LBRACE] = ACTIONS(1826),
    [anon_sym_SEMI] = ACTIONS(1828),
    [anon_sym_RBRACE] = ACTIONS(1828),
    [anon_sym_constraint] = ACTIONS(1826),
    [anon_sym_val] = ACTIONS(1826),
    [anon_sym_end] = ACTIONS(1826),
    [anon_sym_with] = ACTIONS(1826),
    [anon_sym_COLON_GT] = ACTIONS(1828),
    [anon_sym_object] = ACTIONS(1826),
    [anon_sym_inherit] = ACTIONS(1826),
    [anon_sym_method] = ACTIONS(1826),
    [anon_sym_initializer] = ACTIONS(1826),
    [anon_sym_AMP] = ACTIONS(1826),
    [anon_sym_POUND] = ACTIONS(1826),
    [anon_sym_COLON_COLON] = ACTIONS(1828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1828),
    [anon_sym_then] = ACTIONS(1826),
    [anon_sym_else] = ACTIONS(1826),
    [anon_sym_do] = ACTIONS(1826),
    [anon_sym_done] = ACTIONS(1826),
    [anon_sym_to] = ACTIONS(1826),
    [anon_sym_downto] = ACTIONS(1826),
    [anon_sym_new] = ACTIONS(1826),
    [anon_sym_LBRACE_LT] = ACTIONS(1828),
    [anon_sym_GT_RBRACE] = ACTIONS(1828),
    [anon_sym_begin] = ACTIONS(1826),
    [sym_ocamlyacc_value] = ACTIONS(1828),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1826),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1828),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1826),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1826),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1828),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1828),
    [aux_sym_number_token1] = ACTIONS(1828),
    [aux_sym_signed_number_token1] = ACTIONS(1826),
    [anon_sym_SQUOTE] = ACTIONS(1828),
    [anon_sym_DQUOTE] = ACTIONS(1828),
    [sym_prefix_operator] = ACTIONS(1828),
    [aux_sym_sign_operator_token1] = ACTIONS(1826),
    [sym_hash_operator] = ACTIONS(1828),
    [anon_sym_lsl] = ACTIONS(1826),
    [anon_sym_lsr] = ACTIONS(1826),
    [anon_sym_asr] = ACTIONS(1826),
    [aux_sym_pow_operator_token1] = ACTIONS(1828),
    [anon_sym_mod] = ACTIONS(1826),
    [anon_sym_land] = ACTIONS(1826),
    [anon_sym_lor] = ACTIONS(1826),
    [anon_sym_lxor] = ACTIONS(1826),
    [aux_sym_mult_operator_token1] = ACTIONS(1826),
    [aux_sym_add_operator_token1] = ACTIONS(1826),
    [sym_concat_operator] = ACTIONS(1828),
    [sym_rel_operator] = ACTIONS(1826),
    [anon_sym_AMP_AMP] = ACTIONS(1826),
    [anon_sym_or] = ACTIONS(1826),
    [anon_sym_PIPE_PIPE] = ACTIONS(1826),
    [sym_assign_operator] = ACTIONS(1828),
    [sym__uppercase_identifier] = ACTIONS(1828),
    [aux_sym_tag_token1] = ACTIONS(1828),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(935)] = {
    [sym_attribute] = STATE(935),
    [sym__lowercase_identifier] = ACTIONS(1594),
    [anon_sym_TILDE] = ACTIONS(1594),
    [anon_sym_QMARK] = ACTIONS(1594),
    [anon_sym_LPAREN] = ACTIONS(1596),
    [anon_sym_RPAREN] = ACTIONS(1596),
    [anon_sym_COMMA] = ACTIONS(1596),
    [anon_sym_PIPE] = ACTIONS(1594),
    [anon_sym_COLON2] = ACTIONS(1594),
    [anon_sym_DOT] = ACTIONS(1596),
    [anon_sym_DASH_GT] = ACTIONS(1594),
    [anon_sym_LBRACK] = ACTIONS(1594),
    [anon_sym_RBRACK] = ACTIONS(1596),
    [anon_sym_true] = ACTIONS(1594),
    [anon_sym_false] = ACTIONS(1594),
    [anon_sym_LBRACE] = ACTIONS(1594),
    [anon_sym_SEMI] = ACTIONS(1596),
    [anon_sym_RBRACE] = ACTIONS(1596),
    [anon_sym_constraint] = ACTIONS(1594),
    [anon_sym_val] = ACTIONS(1594),
    [anon_sym_end] = ACTIONS(1594),
    [anon_sym_with] = ACTIONS(1594),
    [anon_sym_COLON_GT] = ACTIONS(1596),
    [anon_sym_object] = ACTIONS(1594),
    [anon_sym_inherit] = ACTIONS(1594),
    [anon_sym_method] = ACTIONS(1594),
    [anon_sym_initializer] = ACTIONS(1594),
    [anon_sym_AMP] = ACTIONS(1594),
    [anon_sym_POUND] = ACTIONS(1594),
    [anon_sym_COLON_COLON] = ACTIONS(1596),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1596),
    [anon_sym_then] = ACTIONS(1594),
    [anon_sym_else] = ACTIONS(1594),
    [anon_sym_do] = ACTIONS(1594),
    [anon_sym_done] = ACTIONS(1594),
    [anon_sym_to] = ACTIONS(1594),
    [anon_sym_downto] = ACTIONS(1594),
    [anon_sym_new] = ACTIONS(1594),
    [anon_sym_LBRACE_LT] = ACTIONS(1596),
    [anon_sym_GT_RBRACE] = ACTIONS(1596),
    [anon_sym_begin] = ACTIONS(1594),
    [sym_ocamlyacc_value] = ACTIONS(1596),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1594),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1596),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1594),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1594),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1596),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1596),
    [aux_sym_number_token1] = ACTIONS(1596),
    [aux_sym_signed_number_token1] = ACTIONS(1594),
    [anon_sym_SQUOTE] = ACTIONS(1596),
    [anon_sym_DQUOTE] = ACTIONS(1596),
    [sym_prefix_operator] = ACTIONS(1596),
    [aux_sym_sign_operator_token1] = ACTIONS(1594),
    [sym_hash_operator] = ACTIONS(1596),
    [anon_sym_lsl] = ACTIONS(1594),
    [anon_sym_lsr] = ACTIONS(1594),
    [anon_sym_asr] = ACTIONS(1594),
    [aux_sym_pow_operator_token1] = ACTIONS(1596),
    [anon_sym_mod] = ACTIONS(1594),
    [anon_sym_land] = ACTIONS(1594),
    [anon_sym_lor] = ACTIONS(1594),
    [anon_sym_lxor] = ACTIONS(1594),
    [aux_sym_mult_operator_token1] = ACTIONS(1594),
    [aux_sym_add_operator_token1] = ACTIONS(1594),
    [sym_concat_operator] = ACTIONS(1596),
    [sym_rel_operator] = ACTIONS(1594),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_or] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1594),
    [sym_assign_operator] = ACTIONS(1596),
    [sym__uppercase_identifier] = ACTIONS(1596),
    [aux_sym_tag_token1] = ACTIONS(1596),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(936)] = {
    [sym_attribute] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1716),
    [anon_sym_TILDE] = ACTIONS(1716),
    [anon_sym_QMARK] = ACTIONS(1716),
    [anon_sym_LPAREN] = ACTIONS(1718),
    [anon_sym_RPAREN] = ACTIONS(1718),
    [anon_sym_COMMA] = ACTIONS(1718),
    [anon_sym_PIPE] = ACTIONS(1716),
    [anon_sym_COLON2] = ACTIONS(1716),
    [anon_sym_DOT] = ACTIONS(1718),
    [anon_sym_DASH_GT] = ACTIONS(1716),
    [anon_sym_LBRACK] = ACTIONS(1716),
    [anon_sym_RBRACK] = ACTIONS(1718),
    [anon_sym_true] = ACTIONS(1716),
    [anon_sym_false] = ACTIONS(1716),
    [anon_sym_LBRACE] = ACTIONS(1716),
    [anon_sym_SEMI] = ACTIONS(1718),
    [anon_sym_RBRACE] = ACTIONS(1718),
    [anon_sym_constraint] = ACTIONS(1716),
    [anon_sym_val] = ACTIONS(1716),
    [anon_sym_end] = ACTIONS(1716),
    [anon_sym_with] = ACTIONS(1716),
    [anon_sym_COLON_GT] = ACTIONS(1718),
    [anon_sym_object] = ACTIONS(1716),
    [anon_sym_inherit] = ACTIONS(1716),
    [anon_sym_method] = ACTIONS(1716),
    [anon_sym_initializer] = ACTIONS(1716),
    [anon_sym_AMP] = ACTIONS(1716),
    [anon_sym_POUND] = ACTIONS(1716),
    [anon_sym_COLON_COLON] = ACTIONS(1718),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1718),
    [anon_sym_then] = ACTIONS(1716),
    [anon_sym_else] = ACTIONS(1716),
    [anon_sym_do] = ACTIONS(1716),
    [anon_sym_done] = ACTIONS(1716),
    [anon_sym_to] = ACTIONS(1716),
    [anon_sym_downto] = ACTIONS(1716),
    [anon_sym_new] = ACTIONS(1716),
    [anon_sym_LBRACE_LT] = ACTIONS(1718),
    [anon_sym_GT_RBRACE] = ACTIONS(1718),
    [anon_sym_begin] = ACTIONS(1716),
    [sym_ocamlyacc_value] = ACTIONS(1718),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1716),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1718),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1716),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1716),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1718),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1718),
    [aux_sym_number_token1] = ACTIONS(1718),
    [aux_sym_signed_number_token1] = ACTIONS(1716),
    [anon_sym_SQUOTE] = ACTIONS(1718),
    [anon_sym_DQUOTE] = ACTIONS(1718),
    [sym_prefix_operator] = ACTIONS(1718),
    [aux_sym_sign_operator_token1] = ACTIONS(1716),
    [sym_hash_operator] = ACTIONS(1718),
    [anon_sym_lsl] = ACTIONS(1716),
    [anon_sym_lsr] = ACTIONS(1716),
    [anon_sym_asr] = ACTIONS(1716),
    [aux_sym_pow_operator_token1] = ACTIONS(1718),
    [anon_sym_mod] = ACTIONS(1716),
    [anon_sym_land] = ACTIONS(1716),
    [anon_sym_lor] = ACTIONS(1716),
    [anon_sym_lxor] = ACTIONS(1716),
    [aux_sym_mult_operator_token1] = ACTIONS(1716),
    [aux_sym_add_operator_token1] = ACTIONS(1716),
    [sym_concat_operator] = ACTIONS(1718),
    [sym_rel_operator] = ACTIONS(1716),
    [anon_sym_AMP_AMP] = ACTIONS(1716),
    [anon_sym_or] = ACTIONS(1716),
    [anon_sym_PIPE_PIPE] = ACTIONS(1716),
    [sym_assign_operator] = ACTIONS(1718),
    [sym__uppercase_identifier] = ACTIONS(1718),
    [aux_sym_tag_token1] = ACTIONS(1718),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(937)] = {
    [sym_attribute] = STATE(937),
    [sym__lowercase_identifier] = ACTIONS(1732),
    [anon_sym_TILDE] = ACTIONS(1732),
    [anon_sym_QMARK] = ACTIONS(1732),
    [anon_sym_LPAREN] = ACTIONS(1734),
    [anon_sym_RPAREN] = ACTIONS(1734),
    [anon_sym_COMMA] = ACTIONS(1734),
    [anon_sym_PIPE] = ACTIONS(1732),
    [anon_sym_COLON2] = ACTIONS(1732),
    [anon_sym_DOT] = ACTIONS(1734),
    [anon_sym_DASH_GT] = ACTIONS(1732),
    [anon_sym_LBRACK] = ACTIONS(1732),
    [anon_sym_RBRACK] = ACTIONS(1734),
    [anon_sym_true] = ACTIONS(1732),
    [anon_sym_false] = ACTIONS(1732),
    [anon_sym_LBRACE] = ACTIONS(1732),
    [anon_sym_SEMI] = ACTIONS(1734),
    [anon_sym_RBRACE] = ACTIONS(1734),
    [anon_sym_constraint] = ACTIONS(1732),
    [anon_sym_val] = ACTIONS(1732),
    [anon_sym_end] = ACTIONS(1732),
    [anon_sym_with] = ACTIONS(1732),
    [anon_sym_COLON_GT] = ACTIONS(1734),
    [anon_sym_object] = ACTIONS(1732),
    [anon_sym_inherit] = ACTIONS(1732),
    [anon_sym_method] = ACTIONS(1732),
    [anon_sym_initializer] = ACTIONS(1732),
    [anon_sym_AMP] = ACTIONS(1732),
    [anon_sym_POUND] = ACTIONS(1732),
    [anon_sym_COLON_COLON] = ACTIONS(1734),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1734),
    [anon_sym_then] = ACTIONS(1732),
    [anon_sym_else] = ACTIONS(1732),
    [anon_sym_do] = ACTIONS(1732),
    [anon_sym_done] = ACTIONS(1732),
    [anon_sym_to] = ACTIONS(1732),
    [anon_sym_downto] = ACTIONS(1732),
    [anon_sym_new] = ACTIONS(1732),
    [anon_sym_LBRACE_LT] = ACTIONS(1734),
    [anon_sym_GT_RBRACE] = ACTIONS(1734),
    [anon_sym_begin] = ACTIONS(1732),
    [sym_ocamlyacc_value] = ACTIONS(1734),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1732),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1734),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1732),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1732),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1734),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1734),
    [aux_sym_number_token1] = ACTIONS(1734),
    [aux_sym_signed_number_token1] = ACTIONS(1732),
    [anon_sym_SQUOTE] = ACTIONS(1734),
    [anon_sym_DQUOTE] = ACTIONS(1734),
    [sym_prefix_operator] = ACTIONS(1734),
    [aux_sym_sign_operator_token1] = ACTIONS(1732),
    [sym_hash_operator] = ACTIONS(1734),
    [anon_sym_lsl] = ACTIONS(1732),
    [anon_sym_lsr] = ACTIONS(1732),
    [anon_sym_asr] = ACTIONS(1732),
    [aux_sym_pow_operator_token1] = ACTIONS(1734),
    [anon_sym_mod] = ACTIONS(1732),
    [anon_sym_land] = ACTIONS(1732),
    [anon_sym_lor] = ACTIONS(1732),
    [anon_sym_lxor] = ACTIONS(1732),
    [aux_sym_mult_operator_token1] = ACTIONS(1732),
    [aux_sym_add_operator_token1] = ACTIONS(1732),
    [sym_concat_operator] = ACTIONS(1734),
    [sym_rel_operator] = ACTIONS(1732),
    [anon_sym_AMP_AMP] = ACTIONS(1732),
    [anon_sym_or] = ACTIONS(1732),
    [anon_sym_PIPE_PIPE] = ACTIONS(1732),
    [sym_assign_operator] = ACTIONS(1734),
    [sym__uppercase_identifier] = ACTIONS(1734),
    [aux_sym_tag_token1] = ACTIONS(1734),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(938)] = {
    [sym_attribute] = STATE(938),
    [sym__lowercase_identifier] = ACTIONS(2025),
    [anon_sym_TILDE] = ACTIONS(2025),
    [anon_sym_QMARK] = ACTIONS(2025),
    [anon_sym_LPAREN] = ACTIONS(2027),
    [anon_sym_RPAREN] = ACTIONS(2027),
    [anon_sym_COMMA] = ACTIONS(2027),
    [anon_sym_PIPE] = ACTIONS(2025),
    [anon_sym_COLON2] = ACTIONS(2025),
    [anon_sym_DOT] = ACTIONS(2027),
    [anon_sym_DASH_GT] = ACTIONS(2025),
    [anon_sym_LBRACK] = ACTIONS(2025),
    [anon_sym_RBRACK] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(2025),
    [anon_sym_false] = ACTIONS(2025),
    [anon_sym_LBRACE] = ACTIONS(2025),
    [anon_sym_SEMI] = ACTIONS(2027),
    [anon_sym_RBRACE] = ACTIONS(2027),
    [anon_sym_constraint] = ACTIONS(2025),
    [anon_sym_val] = ACTIONS(2025),
    [anon_sym_end] = ACTIONS(2025),
    [anon_sym_with] = ACTIONS(2025),
    [anon_sym_COLON_GT] = ACTIONS(2027),
    [anon_sym_object] = ACTIONS(2025),
    [anon_sym_inherit] = ACTIONS(2025),
    [anon_sym_method] = ACTIONS(2025),
    [anon_sym_initializer] = ACTIONS(2025),
    [anon_sym_AMP] = ACTIONS(2025),
    [anon_sym_POUND] = ACTIONS(2025),
    [anon_sym_COLON_COLON] = ACTIONS(2027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2027),
    [anon_sym_then] = ACTIONS(2025),
    [anon_sym_else] = ACTIONS(2025),
    [anon_sym_do] = ACTIONS(2025),
    [anon_sym_done] = ACTIONS(2025),
    [anon_sym_to] = ACTIONS(2025),
    [anon_sym_downto] = ACTIONS(2025),
    [anon_sym_new] = ACTIONS(2025),
    [anon_sym_LBRACE_LT] = ACTIONS(2027),
    [anon_sym_GT_RBRACE] = ACTIONS(2027),
    [anon_sym_begin] = ACTIONS(2025),
    [sym_ocamlyacc_value] = ACTIONS(2027),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2025),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2027),
    [aux_sym_number_token1] = ACTIONS(2027),
    [aux_sym_signed_number_token1] = ACTIONS(2025),
    [anon_sym_SQUOTE] = ACTIONS(2027),
    [anon_sym_DQUOTE] = ACTIONS(2027),
    [sym_prefix_operator] = ACTIONS(2027),
    [aux_sym_sign_operator_token1] = ACTIONS(2025),
    [sym_hash_operator] = ACTIONS(2027),
    [anon_sym_lsl] = ACTIONS(2025),
    [anon_sym_lsr] = ACTIONS(2025),
    [anon_sym_asr] = ACTIONS(2025),
    [aux_sym_pow_operator_token1] = ACTIONS(2027),
    [anon_sym_mod] = ACTIONS(2025),
    [anon_sym_land] = ACTIONS(2025),
    [anon_sym_lor] = ACTIONS(2025),
    [anon_sym_lxor] = ACTIONS(2025),
    [aux_sym_mult_operator_token1] = ACTIONS(2025),
    [aux_sym_add_operator_token1] = ACTIONS(2025),
    [sym_concat_operator] = ACTIONS(2027),
    [sym_rel_operator] = ACTIONS(2025),
    [anon_sym_AMP_AMP] = ACTIONS(2025),
    [anon_sym_or] = ACTIONS(2025),
    [anon_sym_PIPE_PIPE] = ACTIONS(2025),
    [sym_assign_operator] = ACTIONS(2027),
    [sym__uppercase_identifier] = ACTIONS(2027),
    [aux_sym_tag_token1] = ACTIONS(2027),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(939)] = {
    [sym_attribute] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(1724),
    [anon_sym_TILDE] = ACTIONS(1724),
    [anon_sym_QMARK] = ACTIONS(1724),
    [anon_sym_LPAREN] = ACTIONS(1726),
    [anon_sym_RPAREN] = ACTIONS(1726),
    [anon_sym_COMMA] = ACTIONS(1726),
    [anon_sym_PIPE] = ACTIONS(1724),
    [anon_sym_COLON2] = ACTIONS(1724),
    [anon_sym_DOT] = ACTIONS(1726),
    [anon_sym_DASH_GT] = ACTIONS(1724),
    [anon_sym_LBRACK] = ACTIONS(1724),
    [anon_sym_RBRACK] = ACTIONS(1726),
    [anon_sym_true] = ACTIONS(1724),
    [anon_sym_false] = ACTIONS(1724),
    [anon_sym_LBRACE] = ACTIONS(1724),
    [anon_sym_SEMI] = ACTIONS(1726),
    [anon_sym_RBRACE] = ACTIONS(1726),
    [anon_sym_constraint] = ACTIONS(1724),
    [anon_sym_val] = ACTIONS(1724),
    [anon_sym_end] = ACTIONS(1724),
    [anon_sym_with] = ACTIONS(1724),
    [anon_sym_COLON_GT] = ACTIONS(1726),
    [anon_sym_object] = ACTIONS(1724),
    [anon_sym_inherit] = ACTIONS(1724),
    [anon_sym_method] = ACTIONS(1724),
    [anon_sym_initializer] = ACTIONS(1724),
    [anon_sym_AMP] = ACTIONS(1724),
    [anon_sym_POUND] = ACTIONS(1724),
    [anon_sym_COLON_COLON] = ACTIONS(1726),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1726),
    [anon_sym_then] = ACTIONS(1724),
    [anon_sym_else] = ACTIONS(1724),
    [anon_sym_do] = ACTIONS(1724),
    [anon_sym_done] = ACTIONS(1724),
    [anon_sym_to] = ACTIONS(1724),
    [anon_sym_downto] = ACTIONS(1724),
    [anon_sym_new] = ACTIONS(1724),
    [anon_sym_LBRACE_LT] = ACTIONS(1726),
    [anon_sym_GT_RBRACE] = ACTIONS(1726),
    [anon_sym_begin] = ACTIONS(1724),
    [sym_ocamlyacc_value] = ACTIONS(1726),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1724),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1726),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1726),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1726),
    [aux_sym_number_token1] = ACTIONS(1726),
    [aux_sym_signed_number_token1] = ACTIONS(1724),
    [anon_sym_SQUOTE] = ACTIONS(1726),
    [anon_sym_DQUOTE] = ACTIONS(1726),
    [sym_prefix_operator] = ACTIONS(1726),
    [aux_sym_sign_operator_token1] = ACTIONS(1724),
    [sym_hash_operator] = ACTIONS(1726),
    [anon_sym_lsl] = ACTIONS(1724),
    [anon_sym_lsr] = ACTIONS(1724),
    [anon_sym_asr] = ACTIONS(1724),
    [aux_sym_pow_operator_token1] = ACTIONS(1726),
    [anon_sym_mod] = ACTIONS(1724),
    [anon_sym_land] = ACTIONS(1724),
    [anon_sym_lor] = ACTIONS(1724),
    [anon_sym_lxor] = ACTIONS(1724),
    [aux_sym_mult_operator_token1] = ACTIONS(1724),
    [aux_sym_add_operator_token1] = ACTIONS(1724),
    [sym_concat_operator] = ACTIONS(1726),
    [sym_rel_operator] = ACTIONS(1724),
    [anon_sym_AMP_AMP] = ACTIONS(1724),
    [anon_sym_or] = ACTIONS(1724),
    [anon_sym_PIPE_PIPE] = ACTIONS(1724),
    [sym_assign_operator] = ACTIONS(1726),
    [sym__uppercase_identifier] = ACTIONS(1726),
    [aux_sym_tag_token1] = ACTIONS(1726),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(940)] = {
    [sym_attribute] = STATE(940),
    [sym__lowercase_identifier] = ACTIONS(1866),
    [anon_sym_TILDE] = ACTIONS(1866),
    [anon_sym_QMARK] = ACTIONS(1866),
    [anon_sym_LPAREN] = ACTIONS(1868),
    [anon_sym_RPAREN] = ACTIONS(1868),
    [anon_sym_COMMA] = ACTIONS(1868),
    [anon_sym_PIPE] = ACTIONS(1866),
    [anon_sym_COLON2] = ACTIONS(1866),
    [anon_sym_DOT] = ACTIONS(1868),
    [anon_sym_DASH_GT] = ACTIONS(1866),
    [anon_sym_LBRACK] = ACTIONS(1866),
    [anon_sym_RBRACK] = ACTIONS(1868),
    [anon_sym_true] = ACTIONS(1866),
    [anon_sym_false] = ACTIONS(1866),
    [anon_sym_LBRACE] = ACTIONS(1866),
    [anon_sym_SEMI] = ACTIONS(1868),
    [anon_sym_RBRACE] = ACTIONS(1868),
    [anon_sym_constraint] = ACTIONS(1866),
    [anon_sym_val] = ACTIONS(1866),
    [anon_sym_end] = ACTIONS(1866),
    [anon_sym_with] = ACTIONS(1866),
    [anon_sym_COLON_GT] = ACTIONS(1868),
    [anon_sym_object] = ACTIONS(1866),
    [anon_sym_inherit] = ACTIONS(1866),
    [anon_sym_method] = ACTIONS(1866),
    [anon_sym_initializer] = ACTIONS(1866),
    [anon_sym_AMP] = ACTIONS(1866),
    [anon_sym_POUND] = ACTIONS(1866),
    [anon_sym_COLON_COLON] = ACTIONS(1868),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1868),
    [anon_sym_then] = ACTIONS(1866),
    [anon_sym_else] = ACTIONS(1866),
    [anon_sym_do] = ACTIONS(1866),
    [anon_sym_done] = ACTIONS(1866),
    [anon_sym_to] = ACTIONS(1866),
    [anon_sym_downto] = ACTIONS(1866),
    [anon_sym_new] = ACTIONS(1866),
    [anon_sym_LBRACE_LT] = ACTIONS(1868),
    [anon_sym_GT_RBRACE] = ACTIONS(1868),
    [anon_sym_begin] = ACTIONS(1866),
    [sym_ocamlyacc_value] = ACTIONS(1868),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1866),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1868),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1866),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1866),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1868),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1868),
    [aux_sym_number_token1] = ACTIONS(1868),
    [aux_sym_signed_number_token1] = ACTIONS(1866),
    [anon_sym_SQUOTE] = ACTIONS(1868),
    [anon_sym_DQUOTE] = ACTIONS(1868),
    [sym_prefix_operator] = ACTIONS(1868),
    [aux_sym_sign_operator_token1] = ACTIONS(1866),
    [sym_hash_operator] = ACTIONS(1868),
    [anon_sym_lsl] = ACTIONS(1866),
    [anon_sym_lsr] = ACTIONS(1866),
    [anon_sym_asr] = ACTIONS(1866),
    [aux_sym_pow_operator_token1] = ACTIONS(1868),
    [anon_sym_mod] = ACTIONS(1866),
    [anon_sym_land] = ACTIONS(1866),
    [anon_sym_lor] = ACTIONS(1866),
    [anon_sym_lxor] = ACTIONS(1866),
    [aux_sym_mult_operator_token1] = ACTIONS(1866),
    [aux_sym_add_operator_token1] = ACTIONS(1866),
    [sym_concat_operator] = ACTIONS(1868),
    [sym_rel_operator] = ACTIONS(1866),
    [anon_sym_AMP_AMP] = ACTIONS(1866),
    [anon_sym_or] = ACTIONS(1866),
    [anon_sym_PIPE_PIPE] = ACTIONS(1866),
    [sym_assign_operator] = ACTIONS(1868),
    [sym__uppercase_identifier] = ACTIONS(1868),
    [aux_sym_tag_token1] = ACTIONS(1868),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(941)] = {
    [sym_attribute] = STATE(941),
    [sym__lowercase_identifier] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_RPAREN] = ACTIONS(1616),
    [anon_sym_COMMA] = ACTIONS(1616),
    [anon_sym_PIPE] = ACTIONS(1614),
    [anon_sym_COLON2] = ACTIONS(1614),
    [anon_sym_DOT] = ACTIONS(1616),
    [anon_sym_DASH_GT] = ACTIONS(1614),
    [anon_sym_LBRACK] = ACTIONS(1614),
    [anon_sym_RBRACK] = ACTIONS(1616),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1616),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [anon_sym_constraint] = ACTIONS(1614),
    [anon_sym_val] = ACTIONS(1614),
    [anon_sym_end] = ACTIONS(1614),
    [anon_sym_with] = ACTIONS(1614),
    [anon_sym_COLON_GT] = ACTIONS(1616),
    [anon_sym_object] = ACTIONS(1614),
    [anon_sym_inherit] = ACTIONS(1614),
    [anon_sym_method] = ACTIONS(1614),
    [anon_sym_initializer] = ACTIONS(1614),
    [anon_sym_AMP] = ACTIONS(1614),
    [anon_sym_POUND] = ACTIONS(1614),
    [anon_sym_COLON_COLON] = ACTIONS(1616),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1616),
    [anon_sym_then] = ACTIONS(1614),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_do] = ACTIONS(1614),
    [anon_sym_done] = ACTIONS(1614),
    [anon_sym_to] = ACTIONS(1614),
    [anon_sym_downto] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
    [anon_sym_LBRACE_LT] = ACTIONS(1616),
    [anon_sym_GT_RBRACE] = ACTIONS(1616),
    [anon_sym_begin] = ACTIONS(1614),
    [sym_ocamlyacc_value] = ACTIONS(1616),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1614),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1616),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1616),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1616),
    [aux_sym_number_token1] = ACTIONS(1616),
    [aux_sym_signed_number_token1] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1616),
    [sym_prefix_operator] = ACTIONS(1616),
    [aux_sym_sign_operator_token1] = ACTIONS(1614),
    [sym_hash_operator] = ACTIONS(1616),
    [anon_sym_lsl] = ACTIONS(1614),
    [anon_sym_lsr] = ACTIONS(1614),
    [anon_sym_asr] = ACTIONS(1614),
    [aux_sym_pow_operator_token1] = ACTIONS(1616),
    [anon_sym_mod] = ACTIONS(1614),
    [anon_sym_land] = ACTIONS(1614),
    [anon_sym_lor] = ACTIONS(1614),
    [anon_sym_lxor] = ACTIONS(1614),
    [aux_sym_mult_operator_token1] = ACTIONS(1614),
    [aux_sym_add_operator_token1] = ACTIONS(1614),
    [sym_concat_operator] = ACTIONS(1616),
    [sym_rel_operator] = ACTIONS(1614),
    [anon_sym_AMP_AMP] = ACTIONS(1614),
    [anon_sym_or] = ACTIONS(1614),
    [anon_sym_PIPE_PIPE] = ACTIONS(1614),
    [sym_assign_operator] = ACTIONS(1616),
    [sym__uppercase_identifier] = ACTIONS(1616),
    [aux_sym_tag_token1] = ACTIONS(1616),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(942)] = {
    [sym_attribute] = STATE(942),
    [sym__lowercase_identifier] = ACTIONS(1918),
    [anon_sym_TILDE] = ACTIONS(1918),
    [anon_sym_QMARK] = ACTIONS(1918),
    [anon_sym_LPAREN] = ACTIONS(1920),
    [anon_sym_RPAREN] = ACTIONS(1920),
    [anon_sym_COMMA] = ACTIONS(1920),
    [anon_sym_PIPE] = ACTIONS(1918),
    [anon_sym_COLON2] = ACTIONS(1918),
    [anon_sym_DOT] = ACTIONS(1920),
    [anon_sym_DASH_GT] = ACTIONS(1918),
    [anon_sym_LBRACK] = ACTIONS(1918),
    [anon_sym_RBRACK] = ACTIONS(1920),
    [anon_sym_true] = ACTIONS(1918),
    [anon_sym_false] = ACTIONS(1918),
    [anon_sym_LBRACE] = ACTIONS(1918),
    [anon_sym_SEMI] = ACTIONS(1920),
    [anon_sym_RBRACE] = ACTIONS(1920),
    [anon_sym_constraint] = ACTIONS(1918),
    [anon_sym_val] = ACTIONS(1918),
    [anon_sym_end] = ACTIONS(1918),
    [anon_sym_with] = ACTIONS(1918),
    [anon_sym_COLON_GT] = ACTIONS(1920),
    [anon_sym_object] = ACTIONS(1918),
    [anon_sym_inherit] = ACTIONS(1918),
    [anon_sym_method] = ACTIONS(1918),
    [anon_sym_initializer] = ACTIONS(1918),
    [anon_sym_AMP] = ACTIONS(1918),
    [anon_sym_POUND] = ACTIONS(1918),
    [anon_sym_COLON_COLON] = ACTIONS(1920),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1920),
    [anon_sym_then] = ACTIONS(1918),
    [anon_sym_else] = ACTIONS(1918),
    [anon_sym_do] = ACTIONS(1918),
    [anon_sym_done] = ACTIONS(1918),
    [anon_sym_to] = ACTIONS(1918),
    [anon_sym_downto] = ACTIONS(1918),
    [anon_sym_new] = ACTIONS(1918),
    [anon_sym_LBRACE_LT] = ACTIONS(1920),
    [anon_sym_GT_RBRACE] = ACTIONS(1920),
    [anon_sym_begin] = ACTIONS(1918),
    [sym_ocamlyacc_value] = ACTIONS(1920),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1918),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1920),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1918),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1918),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1920),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1920),
    [aux_sym_number_token1] = ACTIONS(1920),
    [aux_sym_signed_number_token1] = ACTIONS(1918),
    [anon_sym_SQUOTE] = ACTIONS(1920),
    [anon_sym_DQUOTE] = ACTIONS(1920),
    [sym_prefix_operator] = ACTIONS(1920),
    [aux_sym_sign_operator_token1] = ACTIONS(1918),
    [sym_hash_operator] = ACTIONS(1920),
    [anon_sym_lsl] = ACTIONS(1918),
    [anon_sym_lsr] = ACTIONS(1918),
    [anon_sym_asr] = ACTIONS(1918),
    [aux_sym_pow_operator_token1] = ACTIONS(1920),
    [anon_sym_mod] = ACTIONS(1918),
    [anon_sym_land] = ACTIONS(1918),
    [anon_sym_lor] = ACTIONS(1918),
    [anon_sym_lxor] = ACTIONS(1918),
    [aux_sym_mult_operator_token1] = ACTIONS(1918),
    [aux_sym_add_operator_token1] = ACTIONS(1918),
    [sym_concat_operator] = ACTIONS(1920),
    [sym_rel_operator] = ACTIONS(1918),
    [anon_sym_AMP_AMP] = ACTIONS(1918),
    [anon_sym_or] = ACTIONS(1918),
    [anon_sym_PIPE_PIPE] = ACTIONS(1918),
    [sym_assign_operator] = ACTIONS(1920),
    [sym__uppercase_identifier] = ACTIONS(1920),
    [aux_sym_tag_token1] = ACTIONS(1920),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(943)] = {
    [sym_attribute] = STATE(943),
    [sym__lowercase_identifier] = ACTIONS(1746),
    [anon_sym_TILDE] = ACTIONS(1746),
    [anon_sym_QMARK] = ACTIONS(1746),
    [anon_sym_LPAREN] = ACTIONS(1748),
    [anon_sym_RPAREN] = ACTIONS(1748),
    [anon_sym_COMMA] = ACTIONS(1748),
    [anon_sym_PIPE] = ACTIONS(1746),
    [anon_sym_COLON2] = ACTIONS(1746),
    [anon_sym_DOT] = ACTIONS(1748),
    [anon_sym_DASH_GT] = ACTIONS(1746),
    [anon_sym_LBRACK] = ACTIONS(1746),
    [anon_sym_RBRACK] = ACTIONS(1748),
    [anon_sym_true] = ACTIONS(1746),
    [anon_sym_false] = ACTIONS(1746),
    [anon_sym_LBRACE] = ACTIONS(1746),
    [anon_sym_SEMI] = ACTIONS(1748),
    [anon_sym_RBRACE] = ACTIONS(1748),
    [anon_sym_constraint] = ACTIONS(1746),
    [anon_sym_val] = ACTIONS(1746),
    [anon_sym_end] = ACTIONS(1746),
    [anon_sym_with] = ACTIONS(1746),
    [anon_sym_COLON_GT] = ACTIONS(1748),
    [anon_sym_object] = ACTIONS(1746),
    [anon_sym_inherit] = ACTIONS(1746),
    [anon_sym_method] = ACTIONS(1746),
    [anon_sym_initializer] = ACTIONS(1746),
    [anon_sym_AMP] = ACTIONS(1746),
    [anon_sym_POUND] = ACTIONS(1746),
    [anon_sym_COLON_COLON] = ACTIONS(1748),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1748),
    [anon_sym_then] = ACTIONS(1746),
    [anon_sym_else] = ACTIONS(1746),
    [anon_sym_do] = ACTIONS(1746),
    [anon_sym_done] = ACTIONS(1746),
    [anon_sym_to] = ACTIONS(1746),
    [anon_sym_downto] = ACTIONS(1746),
    [anon_sym_new] = ACTIONS(1746),
    [anon_sym_LBRACE_LT] = ACTIONS(1748),
    [anon_sym_GT_RBRACE] = ACTIONS(1748),
    [anon_sym_begin] = ACTIONS(1746),
    [sym_ocamlyacc_value] = ACTIONS(1748),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1746),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1748),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1746),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1746),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1748),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1748),
    [aux_sym_number_token1] = ACTIONS(1748),
    [aux_sym_signed_number_token1] = ACTIONS(1746),
    [anon_sym_SQUOTE] = ACTIONS(1748),
    [anon_sym_DQUOTE] = ACTIONS(1748),
    [sym_prefix_operator] = ACTIONS(1748),
    [aux_sym_sign_operator_token1] = ACTIONS(1746),
    [sym_hash_operator] = ACTIONS(1748),
    [anon_sym_lsl] = ACTIONS(1746),
    [anon_sym_lsr] = ACTIONS(1746),
    [anon_sym_asr] = ACTIONS(1746),
    [aux_sym_pow_operator_token1] = ACTIONS(1748),
    [anon_sym_mod] = ACTIONS(1746),
    [anon_sym_land] = ACTIONS(1746),
    [anon_sym_lor] = ACTIONS(1746),
    [anon_sym_lxor] = ACTIONS(1746),
    [aux_sym_mult_operator_token1] = ACTIONS(1746),
    [aux_sym_add_operator_token1] = ACTIONS(1746),
    [sym_concat_operator] = ACTIONS(1748),
    [sym_rel_operator] = ACTIONS(1746),
    [anon_sym_AMP_AMP] = ACTIONS(1746),
    [anon_sym_or] = ACTIONS(1746),
    [anon_sym_PIPE_PIPE] = ACTIONS(1746),
    [sym_assign_operator] = ACTIONS(1748),
    [sym__uppercase_identifier] = ACTIONS(1748),
    [aux_sym_tag_token1] = ACTIONS(1748),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(944)] = {
    [sym_attribute] = STATE(944),
    [sym__lowercase_identifier] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2107),
    [anon_sym_QMARK] = ACTIONS(2107),
    [anon_sym_LPAREN] = ACTIONS(2109),
    [anon_sym_RPAREN] = ACTIONS(2109),
    [anon_sym_COMMA] = ACTIONS(2109),
    [anon_sym_PIPE] = ACTIONS(2107),
    [anon_sym_COLON2] = ACTIONS(2107),
    [anon_sym_DOT] = ACTIONS(755),
    [anon_sym_DASH_GT] = ACTIONS(2107),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_RBRACK] = ACTIONS(2109),
    [anon_sym_true] = ACTIONS(2107),
    [anon_sym_false] = ACTIONS(2107),
    [anon_sym_LBRACE] = ACTIONS(2107),
    [anon_sym_SEMI] = ACTIONS(2109),
    [anon_sym_RBRACE] = ACTIONS(2109),
    [anon_sym_constraint] = ACTIONS(2107),
    [anon_sym_val] = ACTIONS(2107),
    [anon_sym_end] = ACTIONS(2107),
    [anon_sym_with] = ACTIONS(2107),
    [anon_sym_COLON_GT] = ACTIONS(2109),
    [anon_sym_object] = ACTIONS(2107),
    [anon_sym_inherit] = ACTIONS(2107),
    [anon_sym_method] = ACTIONS(2107),
    [anon_sym_initializer] = ACTIONS(2107),
    [anon_sym_AMP] = ACTIONS(2107),
    [anon_sym_POUND] = ACTIONS(757),
    [anon_sym_COLON_COLON] = ACTIONS(2109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2109),
    [anon_sym_then] = ACTIONS(2107),
    [anon_sym_else] = ACTIONS(2107),
    [anon_sym_do] = ACTIONS(2107),
    [anon_sym_done] = ACTIONS(2107),
    [anon_sym_to] = ACTIONS(2107),
    [anon_sym_downto] = ACTIONS(2107),
    [anon_sym_new] = ACTIONS(2107),
    [anon_sym_LBRACE_LT] = ACTIONS(2109),
    [anon_sym_GT_RBRACE] = ACTIONS(2109),
    [anon_sym_begin] = ACTIONS(2107),
    [sym_ocamlyacc_value] = ACTIONS(2109),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2107),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2109),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2109),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2109),
    [aux_sym_number_token1] = ACTIONS(2109),
    [aux_sym_signed_number_token1] = ACTIONS(2107),
    [anon_sym_SQUOTE] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(2109),
    [sym_prefix_operator] = ACTIONS(2109),
    [aux_sym_sign_operator_token1] = ACTIONS(2107),
    [sym_hash_operator] = ACTIONS(759),
    [anon_sym_lsl] = ACTIONS(2107),
    [anon_sym_lsr] = ACTIONS(2107),
    [anon_sym_asr] = ACTIONS(2107),
    [aux_sym_pow_operator_token1] = ACTIONS(2109),
    [anon_sym_mod] = ACTIONS(2107),
    [anon_sym_land] = ACTIONS(2107),
    [anon_sym_lor] = ACTIONS(2107),
    [anon_sym_lxor] = ACTIONS(2107),
    [aux_sym_mult_operator_token1] = ACTIONS(2107),
    [aux_sym_add_operator_token1] = ACTIONS(2107),
    [sym_concat_operator] = ACTIONS(2109),
    [sym_rel_operator] = ACTIONS(2107),
    [anon_sym_AMP_AMP] = ACTIONS(2107),
    [anon_sym_or] = ACTIONS(2107),
    [anon_sym_PIPE_PIPE] = ACTIONS(2107),
    [sym_assign_operator] = ACTIONS(2109),
    [sym__uppercase_identifier] = ACTIONS(2109),
    [aux_sym_tag_token1] = ACTIONS(2109),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(945)] = {
    [sym_attribute] = STATE(945),
    [sym__lowercase_identifier] = ACTIONS(1712),
    [anon_sym_TILDE] = ACTIONS(1712),
    [anon_sym_QMARK] = ACTIONS(1712),
    [anon_sym_LPAREN] = ACTIONS(1714),
    [anon_sym_RPAREN] = ACTIONS(1714),
    [anon_sym_COMMA] = ACTIONS(1714),
    [anon_sym_PIPE] = ACTIONS(1712),
    [anon_sym_COLON2] = ACTIONS(1712),
    [anon_sym_DOT] = ACTIONS(1714),
    [anon_sym_DASH_GT] = ACTIONS(1712),
    [anon_sym_LBRACK] = ACTIONS(1712),
    [anon_sym_RBRACK] = ACTIONS(1714),
    [anon_sym_true] = ACTIONS(1712),
    [anon_sym_false] = ACTIONS(1712),
    [anon_sym_LBRACE] = ACTIONS(1712),
    [anon_sym_SEMI] = ACTIONS(1714),
    [anon_sym_RBRACE] = ACTIONS(1714),
    [anon_sym_constraint] = ACTIONS(1712),
    [anon_sym_val] = ACTIONS(1712),
    [anon_sym_end] = ACTIONS(1712),
    [anon_sym_with] = ACTIONS(1712),
    [anon_sym_COLON_GT] = ACTIONS(1714),
    [anon_sym_object] = ACTIONS(1712),
    [anon_sym_inherit] = ACTIONS(1712),
    [anon_sym_method] = ACTIONS(1712),
    [anon_sym_initializer] = ACTIONS(1712),
    [anon_sym_AMP] = ACTIONS(1712),
    [anon_sym_POUND] = ACTIONS(1712),
    [anon_sym_COLON_COLON] = ACTIONS(1714),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1714),
    [anon_sym_then] = ACTIONS(1712),
    [anon_sym_else] = ACTIONS(1712),
    [anon_sym_do] = ACTIONS(1712),
    [anon_sym_done] = ACTIONS(1712),
    [anon_sym_to] = ACTIONS(1712),
    [anon_sym_downto] = ACTIONS(1712),
    [anon_sym_new] = ACTIONS(1712),
    [anon_sym_LBRACE_LT] = ACTIONS(1714),
    [anon_sym_GT_RBRACE] = ACTIONS(1714),
    [anon_sym_begin] = ACTIONS(1712),
    [sym_ocamlyacc_value] = ACTIONS(1714),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1712),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1714),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1712),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1712),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1714),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1714),
    [aux_sym_number_token1] = ACTIONS(1714),
    [aux_sym_signed_number_token1] = ACTIONS(1712),
    [anon_sym_SQUOTE] = ACTIONS(1714),
    [anon_sym_DQUOTE] = ACTIONS(1714),
    [sym_prefix_operator] = ACTIONS(1714),
    [aux_sym_sign_operator_token1] = ACTIONS(1712),
    [sym_hash_operator] = ACTIONS(1714),
    [anon_sym_lsl] = ACTIONS(1712),
    [anon_sym_lsr] = ACTIONS(1712),
    [anon_sym_asr] = ACTIONS(1712),
    [aux_sym_pow_operator_token1] = ACTIONS(1714),
    [anon_sym_mod] = ACTIONS(1712),
    [anon_sym_land] = ACTIONS(1712),
    [anon_sym_lor] = ACTIONS(1712),
    [anon_sym_lxor] = ACTIONS(1712),
    [aux_sym_mult_operator_token1] = ACTIONS(1712),
    [aux_sym_add_operator_token1] = ACTIONS(1712),
    [sym_concat_operator] = ACTIONS(1714),
    [sym_rel_operator] = ACTIONS(1712),
    [anon_sym_AMP_AMP] = ACTIONS(1712),
    [anon_sym_or] = ACTIONS(1712),
    [anon_sym_PIPE_PIPE] = ACTIONS(1712),
    [sym_assign_operator] = ACTIONS(1714),
    [sym__uppercase_identifier] = ACTIONS(1714),
    [aux_sym_tag_token1] = ACTIONS(1714),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(946)] = {
    [sym_attribute] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(1886),
    [anon_sym_TILDE] = ACTIONS(1886),
    [anon_sym_QMARK] = ACTIONS(1886),
    [anon_sym_LPAREN] = ACTIONS(1888),
    [anon_sym_RPAREN] = ACTIONS(1888),
    [anon_sym_COMMA] = ACTIONS(1888),
    [anon_sym_PIPE] = ACTIONS(1886),
    [anon_sym_COLON2] = ACTIONS(1886),
    [anon_sym_DOT] = ACTIONS(1888),
    [anon_sym_DASH_GT] = ACTIONS(1886),
    [anon_sym_LBRACK] = ACTIONS(1886),
    [anon_sym_RBRACK] = ACTIONS(1888),
    [anon_sym_true] = ACTIONS(1886),
    [anon_sym_false] = ACTIONS(1886),
    [anon_sym_LBRACE] = ACTIONS(1886),
    [anon_sym_SEMI] = ACTIONS(1888),
    [anon_sym_RBRACE] = ACTIONS(1888),
    [anon_sym_constraint] = ACTIONS(1886),
    [anon_sym_val] = ACTIONS(1886),
    [anon_sym_end] = ACTIONS(1886),
    [anon_sym_with] = ACTIONS(1886),
    [anon_sym_COLON_GT] = ACTIONS(1888),
    [anon_sym_object] = ACTIONS(1886),
    [anon_sym_inherit] = ACTIONS(1886),
    [anon_sym_method] = ACTIONS(1886),
    [anon_sym_initializer] = ACTIONS(1886),
    [anon_sym_AMP] = ACTIONS(1886),
    [anon_sym_POUND] = ACTIONS(1886),
    [anon_sym_COLON_COLON] = ACTIONS(1888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1888),
    [anon_sym_then] = ACTIONS(1886),
    [anon_sym_else] = ACTIONS(1886),
    [anon_sym_do] = ACTIONS(1886),
    [anon_sym_done] = ACTIONS(1886),
    [anon_sym_to] = ACTIONS(1886),
    [anon_sym_downto] = ACTIONS(1886),
    [anon_sym_new] = ACTIONS(1886),
    [anon_sym_LBRACE_LT] = ACTIONS(1888),
    [anon_sym_GT_RBRACE] = ACTIONS(1888),
    [anon_sym_begin] = ACTIONS(1886),
    [sym_ocamlyacc_value] = ACTIONS(1888),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1886),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1888),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1886),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1886),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1888),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1888),
    [aux_sym_number_token1] = ACTIONS(1888),
    [aux_sym_signed_number_token1] = ACTIONS(1886),
    [anon_sym_SQUOTE] = ACTIONS(1888),
    [anon_sym_DQUOTE] = ACTIONS(1888),
    [sym_prefix_operator] = ACTIONS(1888),
    [aux_sym_sign_operator_token1] = ACTIONS(1886),
    [sym_hash_operator] = ACTIONS(1888),
    [anon_sym_lsl] = ACTIONS(1886),
    [anon_sym_lsr] = ACTIONS(1886),
    [anon_sym_asr] = ACTIONS(1886),
    [aux_sym_pow_operator_token1] = ACTIONS(1888),
    [anon_sym_mod] = ACTIONS(1886),
    [anon_sym_land] = ACTIONS(1886),
    [anon_sym_lor] = ACTIONS(1886),
    [anon_sym_lxor] = ACTIONS(1886),
    [aux_sym_mult_operator_token1] = ACTIONS(1886),
    [aux_sym_add_operator_token1] = ACTIONS(1886),
    [sym_concat_operator] = ACTIONS(1888),
    [sym_rel_operator] = ACTIONS(1886),
    [anon_sym_AMP_AMP] = ACTIONS(1886),
    [anon_sym_or] = ACTIONS(1886),
    [anon_sym_PIPE_PIPE] = ACTIONS(1886),
    [sym_assign_operator] = ACTIONS(1888),
    [sym__uppercase_identifier] = ACTIONS(1888),
    [aux_sym_tag_token1] = ACTIONS(1888),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(947)] = {
    [sym_attribute] = STATE(947),
    [sym__lowercase_identifier] = ACTIONS(1914),
    [anon_sym_TILDE] = ACTIONS(1914),
    [anon_sym_QMARK] = ACTIONS(1914),
    [anon_sym_LPAREN] = ACTIONS(1916),
    [anon_sym_RPAREN] = ACTIONS(1916),
    [anon_sym_COMMA] = ACTIONS(1916),
    [anon_sym_PIPE] = ACTIONS(1914),
    [anon_sym_COLON2] = ACTIONS(1914),
    [anon_sym_DOT] = ACTIONS(1916),
    [anon_sym_DASH_GT] = ACTIONS(1914),
    [anon_sym_LBRACK] = ACTIONS(1914),
    [anon_sym_RBRACK] = ACTIONS(1916),
    [anon_sym_true] = ACTIONS(1914),
    [anon_sym_false] = ACTIONS(1914),
    [anon_sym_LBRACE] = ACTIONS(1914),
    [anon_sym_SEMI] = ACTIONS(1916),
    [anon_sym_RBRACE] = ACTIONS(1916),
    [anon_sym_constraint] = ACTIONS(1914),
    [anon_sym_val] = ACTIONS(1914),
    [anon_sym_end] = ACTIONS(1914),
    [anon_sym_with] = ACTIONS(1914),
    [anon_sym_COLON_GT] = ACTIONS(1916),
    [anon_sym_object] = ACTIONS(1914),
    [anon_sym_inherit] = ACTIONS(1914),
    [anon_sym_method] = ACTIONS(1914),
    [anon_sym_initializer] = ACTIONS(1914),
    [anon_sym_AMP] = ACTIONS(1914),
    [anon_sym_POUND] = ACTIONS(1914),
    [anon_sym_COLON_COLON] = ACTIONS(1916),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1916),
    [anon_sym_then] = ACTIONS(1914),
    [anon_sym_else] = ACTIONS(1914),
    [anon_sym_do] = ACTIONS(1914),
    [anon_sym_done] = ACTIONS(1914),
    [anon_sym_to] = ACTIONS(1914),
    [anon_sym_downto] = ACTIONS(1914),
    [anon_sym_new] = ACTIONS(1914),
    [anon_sym_LBRACE_LT] = ACTIONS(1916),
    [anon_sym_GT_RBRACE] = ACTIONS(1916),
    [anon_sym_begin] = ACTIONS(1914),
    [sym_ocamlyacc_value] = ACTIONS(1916),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1914),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1916),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1914),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1914),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1916),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1916),
    [aux_sym_number_token1] = ACTIONS(1916),
    [aux_sym_signed_number_token1] = ACTIONS(1914),
    [anon_sym_SQUOTE] = ACTIONS(1916),
    [anon_sym_DQUOTE] = ACTIONS(1916),
    [sym_prefix_operator] = ACTIONS(1916),
    [aux_sym_sign_operator_token1] = ACTIONS(1914),
    [sym_hash_operator] = ACTIONS(1916),
    [anon_sym_lsl] = ACTIONS(1914),
    [anon_sym_lsr] = ACTIONS(1914),
    [anon_sym_asr] = ACTIONS(1914),
    [aux_sym_pow_operator_token1] = ACTIONS(1916),
    [anon_sym_mod] = ACTIONS(1914),
    [anon_sym_land] = ACTIONS(1914),
    [anon_sym_lor] = ACTIONS(1914),
    [anon_sym_lxor] = ACTIONS(1914),
    [aux_sym_mult_operator_token1] = ACTIONS(1914),
    [aux_sym_add_operator_token1] = ACTIONS(1914),
    [sym_concat_operator] = ACTIONS(1916),
    [sym_rel_operator] = ACTIONS(1914),
    [anon_sym_AMP_AMP] = ACTIONS(1914),
    [anon_sym_or] = ACTIONS(1914),
    [anon_sym_PIPE_PIPE] = ACTIONS(1914),
    [sym_assign_operator] = ACTIONS(1916),
    [sym__uppercase_identifier] = ACTIONS(1916),
    [aux_sym_tag_token1] = ACTIONS(1916),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(948)] = {
    [sym_attribute] = STATE(948),
    [sym__lowercase_identifier] = ACTIONS(1942),
    [anon_sym_TILDE] = ACTIONS(1942),
    [anon_sym_QMARK] = ACTIONS(1942),
    [anon_sym_LPAREN] = ACTIONS(1944),
    [anon_sym_RPAREN] = ACTIONS(1944),
    [anon_sym_COMMA] = ACTIONS(1944),
    [anon_sym_PIPE] = ACTIONS(1942),
    [anon_sym_COLON2] = ACTIONS(1942),
    [anon_sym_DOT] = ACTIONS(1944),
    [anon_sym_DASH_GT] = ACTIONS(1942),
    [anon_sym_LBRACK] = ACTIONS(1942),
    [anon_sym_RBRACK] = ACTIONS(1944),
    [anon_sym_true] = ACTIONS(1942),
    [anon_sym_false] = ACTIONS(1942),
    [anon_sym_LBRACE] = ACTIONS(1942),
    [anon_sym_SEMI] = ACTIONS(1944),
    [anon_sym_RBRACE] = ACTIONS(1944),
    [anon_sym_constraint] = ACTIONS(1942),
    [anon_sym_val] = ACTIONS(1942),
    [anon_sym_end] = ACTIONS(1942),
    [anon_sym_with] = ACTIONS(1942),
    [anon_sym_COLON_GT] = ACTIONS(1944),
    [anon_sym_object] = ACTIONS(1942),
    [anon_sym_inherit] = ACTIONS(1942),
    [anon_sym_method] = ACTIONS(1942),
    [anon_sym_initializer] = ACTIONS(1942),
    [anon_sym_AMP] = ACTIONS(1942),
    [anon_sym_POUND] = ACTIONS(1942),
    [anon_sym_COLON_COLON] = ACTIONS(1944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1944),
    [anon_sym_then] = ACTIONS(1942),
    [anon_sym_else] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1942),
    [anon_sym_done] = ACTIONS(1942),
    [anon_sym_to] = ACTIONS(1942),
    [anon_sym_downto] = ACTIONS(1942),
    [anon_sym_new] = ACTIONS(1942),
    [anon_sym_LBRACE_LT] = ACTIONS(1944),
    [anon_sym_GT_RBRACE] = ACTIONS(1944),
    [anon_sym_begin] = ACTIONS(1942),
    [sym_ocamlyacc_value] = ACTIONS(1944),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1942),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1944),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1942),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1942),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1944),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1944),
    [aux_sym_number_token1] = ACTIONS(1944),
    [aux_sym_signed_number_token1] = ACTIONS(1942),
    [anon_sym_SQUOTE] = ACTIONS(1944),
    [anon_sym_DQUOTE] = ACTIONS(1944),
    [sym_prefix_operator] = ACTIONS(1944),
    [aux_sym_sign_operator_token1] = ACTIONS(1942),
    [sym_hash_operator] = ACTIONS(1944),
    [anon_sym_lsl] = ACTIONS(1942),
    [anon_sym_lsr] = ACTIONS(1942),
    [anon_sym_asr] = ACTIONS(1942),
    [aux_sym_pow_operator_token1] = ACTIONS(1944),
    [anon_sym_mod] = ACTIONS(1942),
    [anon_sym_land] = ACTIONS(1942),
    [anon_sym_lor] = ACTIONS(1942),
    [anon_sym_lxor] = ACTIONS(1942),
    [aux_sym_mult_operator_token1] = ACTIONS(1942),
    [aux_sym_add_operator_token1] = ACTIONS(1942),
    [sym_concat_operator] = ACTIONS(1944),
    [sym_rel_operator] = ACTIONS(1942),
    [anon_sym_AMP_AMP] = ACTIONS(1942),
    [anon_sym_or] = ACTIONS(1942),
    [anon_sym_PIPE_PIPE] = ACTIONS(1942),
    [sym_assign_operator] = ACTIONS(1944),
    [sym__uppercase_identifier] = ACTIONS(1944),
    [aux_sym_tag_token1] = ACTIONS(1944),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(949)] = {
    [sym_attribute] = STATE(949),
    [sym__lowercase_identifier] = ACTIONS(1598),
    [anon_sym_TILDE] = ACTIONS(1598),
    [anon_sym_QMARK] = ACTIONS(1598),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_RPAREN] = ACTIONS(1600),
    [anon_sym_COMMA] = ACTIONS(1600),
    [anon_sym_PIPE] = ACTIONS(1598),
    [anon_sym_COLON2] = ACTIONS(1598),
    [anon_sym_DOT] = ACTIONS(1600),
    [anon_sym_DASH_GT] = ACTIONS(1598),
    [anon_sym_LBRACK] = ACTIONS(1598),
    [anon_sym_RBRACK] = ACTIONS(1600),
    [anon_sym_true] = ACTIONS(1598),
    [anon_sym_false] = ACTIONS(1598),
    [anon_sym_LBRACE] = ACTIONS(1598),
    [anon_sym_SEMI] = ACTIONS(1600),
    [anon_sym_RBRACE] = ACTIONS(1600),
    [anon_sym_constraint] = ACTIONS(1598),
    [anon_sym_val] = ACTIONS(1598),
    [anon_sym_end] = ACTIONS(1598),
    [anon_sym_with] = ACTIONS(1598),
    [anon_sym_COLON_GT] = ACTIONS(1600),
    [anon_sym_object] = ACTIONS(1598),
    [anon_sym_inherit] = ACTIONS(1598),
    [anon_sym_method] = ACTIONS(1598),
    [anon_sym_initializer] = ACTIONS(1598),
    [anon_sym_AMP] = ACTIONS(1598),
    [anon_sym_POUND] = ACTIONS(1598),
    [anon_sym_COLON_COLON] = ACTIONS(1600),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1600),
    [anon_sym_then] = ACTIONS(1598),
    [anon_sym_else] = ACTIONS(1598),
    [anon_sym_do] = ACTIONS(1598),
    [anon_sym_done] = ACTIONS(1598),
    [anon_sym_to] = ACTIONS(1598),
    [anon_sym_downto] = ACTIONS(1598),
    [anon_sym_new] = ACTIONS(1598),
    [anon_sym_LBRACE_LT] = ACTIONS(1600),
    [anon_sym_GT_RBRACE] = ACTIONS(1600),
    [anon_sym_begin] = ACTIONS(1598),
    [sym_ocamlyacc_value] = ACTIONS(1600),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1598),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1600),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1598),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1598),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1600),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1600),
    [aux_sym_number_token1] = ACTIONS(1600),
    [aux_sym_signed_number_token1] = ACTIONS(1598),
    [anon_sym_SQUOTE] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1600),
    [sym_prefix_operator] = ACTIONS(1600),
    [aux_sym_sign_operator_token1] = ACTIONS(1598),
    [sym_hash_operator] = ACTIONS(1600),
    [anon_sym_lsl] = ACTIONS(1598),
    [anon_sym_lsr] = ACTIONS(1598),
    [anon_sym_asr] = ACTIONS(1598),
    [aux_sym_pow_operator_token1] = ACTIONS(1600),
    [anon_sym_mod] = ACTIONS(1598),
    [anon_sym_land] = ACTIONS(1598),
    [anon_sym_lor] = ACTIONS(1598),
    [anon_sym_lxor] = ACTIONS(1598),
    [aux_sym_mult_operator_token1] = ACTIONS(1598),
    [aux_sym_add_operator_token1] = ACTIONS(1598),
    [sym_concat_operator] = ACTIONS(1600),
    [sym_rel_operator] = ACTIONS(1598),
    [anon_sym_AMP_AMP] = ACTIONS(1598),
    [anon_sym_or] = ACTIONS(1598),
    [anon_sym_PIPE_PIPE] = ACTIONS(1598),
    [sym_assign_operator] = ACTIONS(1600),
    [sym__uppercase_identifier] = ACTIONS(1600),
    [aux_sym_tag_token1] = ACTIONS(1600),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(950)] = {
    [sym_attribute] = STATE(950),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_RPAREN] = ACTIONS(2081),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_COLON2] = ACTIONS(2079),
    [anon_sym_DOT] = ACTIONS(755),
    [anon_sym_DASH_GT] = ACTIONS(2079),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2081),
    [anon_sym_RBRACE] = ACTIONS(2081),
    [anon_sym_constraint] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_with] = ACTIONS(2079),
    [anon_sym_COLON_GT] = ACTIONS(2081),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_inherit] = ACTIONS(2079),
    [anon_sym_method] = ACTIONS(2079),
    [anon_sym_initializer] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_POUND] = ACTIONS(757),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_then] = ACTIONS(2079),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_do] = ACTIONS(2079),
    [anon_sym_done] = ACTIONS(2079),
    [anon_sym_to] = ACTIONS(2079),
    [anon_sym_downto] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_GT_RBRACE] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [sym_hash_operator] = ACTIONS(759),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(951)] = {
    [sym_attribute] = STATE(951),
    [sym__lowercase_identifier] = ACTIONS(1602),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_QMARK] = ACTIONS(1602),
    [anon_sym_LPAREN] = ACTIONS(1604),
    [anon_sym_RPAREN] = ACTIONS(1604),
    [anon_sym_COMMA] = ACTIONS(1604),
    [anon_sym_PIPE] = ACTIONS(1602),
    [anon_sym_COLON2] = ACTIONS(1602),
    [anon_sym_DOT] = ACTIONS(1604),
    [anon_sym_DASH_GT] = ACTIONS(1602),
    [anon_sym_LBRACK] = ACTIONS(1602),
    [anon_sym_RBRACK] = ACTIONS(1604),
    [anon_sym_true] = ACTIONS(1602),
    [anon_sym_false] = ACTIONS(1602),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [anon_sym_SEMI] = ACTIONS(1604),
    [anon_sym_RBRACE] = ACTIONS(1604),
    [anon_sym_constraint] = ACTIONS(1602),
    [anon_sym_val] = ACTIONS(1602),
    [anon_sym_end] = ACTIONS(1602),
    [anon_sym_with] = ACTIONS(1602),
    [anon_sym_COLON_GT] = ACTIONS(1604),
    [anon_sym_object] = ACTIONS(1602),
    [anon_sym_inherit] = ACTIONS(1602),
    [anon_sym_method] = ACTIONS(1602),
    [anon_sym_initializer] = ACTIONS(1602),
    [anon_sym_AMP] = ACTIONS(1602),
    [anon_sym_POUND] = ACTIONS(1602),
    [anon_sym_COLON_COLON] = ACTIONS(1604),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1604),
    [anon_sym_then] = ACTIONS(1602),
    [anon_sym_else] = ACTIONS(1602),
    [anon_sym_do] = ACTIONS(1602),
    [anon_sym_done] = ACTIONS(1602),
    [anon_sym_to] = ACTIONS(1602),
    [anon_sym_downto] = ACTIONS(1602),
    [anon_sym_new] = ACTIONS(1602),
    [anon_sym_LBRACE_LT] = ACTIONS(1604),
    [anon_sym_GT_RBRACE] = ACTIONS(1604),
    [anon_sym_begin] = ACTIONS(1602),
    [sym_ocamlyacc_value] = ACTIONS(1604),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1602),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1604),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1602),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1602),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1604),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1604),
    [aux_sym_number_token1] = ACTIONS(1604),
    [aux_sym_signed_number_token1] = ACTIONS(1602),
    [anon_sym_SQUOTE] = ACTIONS(1604),
    [anon_sym_DQUOTE] = ACTIONS(1604),
    [sym_prefix_operator] = ACTIONS(1604),
    [aux_sym_sign_operator_token1] = ACTIONS(1602),
    [sym_hash_operator] = ACTIONS(1604),
    [anon_sym_lsl] = ACTIONS(1602),
    [anon_sym_lsr] = ACTIONS(1602),
    [anon_sym_asr] = ACTIONS(1602),
    [aux_sym_pow_operator_token1] = ACTIONS(1604),
    [anon_sym_mod] = ACTIONS(1602),
    [anon_sym_land] = ACTIONS(1602),
    [anon_sym_lor] = ACTIONS(1602),
    [anon_sym_lxor] = ACTIONS(1602),
    [aux_sym_mult_operator_token1] = ACTIONS(1602),
    [aux_sym_add_operator_token1] = ACTIONS(1602),
    [sym_concat_operator] = ACTIONS(1604),
    [sym_rel_operator] = ACTIONS(1602),
    [anon_sym_AMP_AMP] = ACTIONS(1602),
    [anon_sym_or] = ACTIONS(1602),
    [anon_sym_PIPE_PIPE] = ACTIONS(1602),
    [sym_assign_operator] = ACTIONS(1604),
    [sym__uppercase_identifier] = ACTIONS(1604),
    [aux_sym_tag_token1] = ACTIONS(1604),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(952)] = {
    [sym_attribute] = STATE(952),
    [sym__lowercase_identifier] = ACTIONS(1606),
    [anon_sym_TILDE] = ACTIONS(1606),
    [anon_sym_QMARK] = ACTIONS(1606),
    [anon_sym_LPAREN] = ACTIONS(1608),
    [anon_sym_RPAREN] = ACTIONS(1608),
    [anon_sym_COMMA] = ACTIONS(1608),
    [anon_sym_PIPE] = ACTIONS(1606),
    [anon_sym_COLON2] = ACTIONS(1606),
    [anon_sym_DOT] = ACTIONS(1608),
    [anon_sym_DASH_GT] = ACTIONS(1606),
    [anon_sym_LBRACK] = ACTIONS(1606),
    [anon_sym_RBRACK] = ACTIONS(1608),
    [anon_sym_true] = ACTIONS(1606),
    [anon_sym_false] = ACTIONS(1606),
    [anon_sym_LBRACE] = ACTIONS(1606),
    [anon_sym_SEMI] = ACTIONS(1608),
    [anon_sym_RBRACE] = ACTIONS(1608),
    [anon_sym_constraint] = ACTIONS(1606),
    [anon_sym_val] = ACTIONS(1606),
    [anon_sym_end] = ACTIONS(1606),
    [anon_sym_with] = ACTIONS(1606),
    [anon_sym_COLON_GT] = ACTIONS(1608),
    [anon_sym_object] = ACTIONS(1606),
    [anon_sym_inherit] = ACTIONS(1606),
    [anon_sym_method] = ACTIONS(1606),
    [anon_sym_initializer] = ACTIONS(1606),
    [anon_sym_AMP] = ACTIONS(1606),
    [anon_sym_POUND] = ACTIONS(1606),
    [anon_sym_COLON_COLON] = ACTIONS(1608),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1608),
    [anon_sym_then] = ACTIONS(1606),
    [anon_sym_else] = ACTIONS(1606),
    [anon_sym_do] = ACTIONS(1606),
    [anon_sym_done] = ACTIONS(1606),
    [anon_sym_to] = ACTIONS(1606),
    [anon_sym_downto] = ACTIONS(1606),
    [anon_sym_new] = ACTIONS(1606),
    [anon_sym_LBRACE_LT] = ACTIONS(1608),
    [anon_sym_GT_RBRACE] = ACTIONS(1608),
    [anon_sym_begin] = ACTIONS(1606),
    [sym_ocamlyacc_value] = ACTIONS(1608),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1606),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1608),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1606),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1606),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1608),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1608),
    [aux_sym_number_token1] = ACTIONS(1608),
    [aux_sym_signed_number_token1] = ACTIONS(1606),
    [anon_sym_SQUOTE] = ACTIONS(1608),
    [anon_sym_DQUOTE] = ACTIONS(1608),
    [sym_prefix_operator] = ACTIONS(1608),
    [aux_sym_sign_operator_token1] = ACTIONS(1606),
    [sym_hash_operator] = ACTIONS(1608),
    [anon_sym_lsl] = ACTIONS(1606),
    [anon_sym_lsr] = ACTIONS(1606),
    [anon_sym_asr] = ACTIONS(1606),
    [aux_sym_pow_operator_token1] = ACTIONS(1608),
    [anon_sym_mod] = ACTIONS(1606),
    [anon_sym_land] = ACTIONS(1606),
    [anon_sym_lor] = ACTIONS(1606),
    [anon_sym_lxor] = ACTIONS(1606),
    [aux_sym_mult_operator_token1] = ACTIONS(1606),
    [aux_sym_add_operator_token1] = ACTIONS(1606),
    [sym_concat_operator] = ACTIONS(1608),
    [sym_rel_operator] = ACTIONS(1606),
    [anon_sym_AMP_AMP] = ACTIONS(1606),
    [anon_sym_or] = ACTIONS(1606),
    [anon_sym_PIPE_PIPE] = ACTIONS(1606),
    [sym_assign_operator] = ACTIONS(1608),
    [sym__uppercase_identifier] = ACTIONS(1608),
    [aux_sym_tag_token1] = ACTIONS(1608),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(953)] = {
    [sym_attribute] = STATE(953),
    [sym__lowercase_identifier] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1628),
    [anon_sym_RPAREN] = ACTIONS(1628),
    [anon_sym_COMMA] = ACTIONS(1628),
    [anon_sym_PIPE] = ACTIONS(1626),
    [anon_sym_COLON2] = ACTIONS(1626),
    [anon_sym_DOT] = ACTIONS(1628),
    [anon_sym_DASH_GT] = ACTIONS(1626),
    [anon_sym_LBRACK] = ACTIONS(1626),
    [anon_sym_RBRACK] = ACTIONS(1628),
    [anon_sym_true] = ACTIONS(1626),
    [anon_sym_false] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1628),
    [anon_sym_RBRACE] = ACTIONS(1628),
    [anon_sym_constraint] = ACTIONS(1626),
    [anon_sym_val] = ACTIONS(1626),
    [anon_sym_end] = ACTIONS(1626),
    [anon_sym_with] = ACTIONS(1626),
    [anon_sym_COLON_GT] = ACTIONS(1628),
    [anon_sym_object] = ACTIONS(1626),
    [anon_sym_inherit] = ACTIONS(1626),
    [anon_sym_method] = ACTIONS(1626),
    [anon_sym_initializer] = ACTIONS(1626),
    [anon_sym_AMP] = ACTIONS(1626),
    [anon_sym_POUND] = ACTIONS(1626),
    [anon_sym_COLON_COLON] = ACTIONS(1628),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1628),
    [anon_sym_then] = ACTIONS(1626),
    [anon_sym_else] = ACTIONS(1626),
    [anon_sym_do] = ACTIONS(1626),
    [anon_sym_done] = ACTIONS(1626),
    [anon_sym_to] = ACTIONS(1626),
    [anon_sym_downto] = ACTIONS(1626),
    [anon_sym_new] = ACTIONS(1626),
    [anon_sym_LBRACE_LT] = ACTIONS(1628),
    [anon_sym_GT_RBRACE] = ACTIONS(1628),
    [anon_sym_begin] = ACTIONS(1626),
    [sym_ocamlyacc_value] = ACTIONS(1628),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1626),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1628),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1626),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1626),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1628),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1628),
    [aux_sym_number_token1] = ACTIONS(1628),
    [aux_sym_signed_number_token1] = ACTIONS(1626),
    [anon_sym_SQUOTE] = ACTIONS(1628),
    [anon_sym_DQUOTE] = ACTIONS(1628),
    [sym_prefix_operator] = ACTIONS(1628),
    [aux_sym_sign_operator_token1] = ACTIONS(1626),
    [sym_hash_operator] = ACTIONS(1628),
    [anon_sym_lsl] = ACTIONS(1626),
    [anon_sym_lsr] = ACTIONS(1626),
    [anon_sym_asr] = ACTIONS(1626),
    [aux_sym_pow_operator_token1] = ACTIONS(1628),
    [anon_sym_mod] = ACTIONS(1626),
    [anon_sym_land] = ACTIONS(1626),
    [anon_sym_lor] = ACTIONS(1626),
    [anon_sym_lxor] = ACTIONS(1626),
    [aux_sym_mult_operator_token1] = ACTIONS(1626),
    [aux_sym_add_operator_token1] = ACTIONS(1626),
    [sym_concat_operator] = ACTIONS(1628),
    [sym_rel_operator] = ACTIONS(1626),
    [anon_sym_AMP_AMP] = ACTIONS(1626),
    [anon_sym_or] = ACTIONS(1626),
    [anon_sym_PIPE_PIPE] = ACTIONS(1626),
    [sym_assign_operator] = ACTIONS(1628),
    [sym__uppercase_identifier] = ACTIONS(1628),
    [aux_sym_tag_token1] = ACTIONS(1628),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(954)] = {
    [sym_attribute] = STATE(954),
    [sym__lowercase_identifier] = ACTIONS(1618),
    [anon_sym_TILDE] = ACTIONS(1618),
    [anon_sym_QMARK] = ACTIONS(1618),
    [anon_sym_LPAREN] = ACTIONS(1620),
    [anon_sym_RPAREN] = ACTIONS(1620),
    [anon_sym_COMMA] = ACTIONS(1620),
    [anon_sym_PIPE] = ACTIONS(1618),
    [anon_sym_COLON2] = ACTIONS(1618),
    [anon_sym_DOT] = ACTIONS(1620),
    [anon_sym_DASH_GT] = ACTIONS(1618),
    [anon_sym_LBRACK] = ACTIONS(1618),
    [anon_sym_RBRACK] = ACTIONS(1620),
    [anon_sym_true] = ACTIONS(1618),
    [anon_sym_false] = ACTIONS(1618),
    [anon_sym_LBRACE] = ACTIONS(1618),
    [anon_sym_SEMI] = ACTIONS(1620),
    [anon_sym_RBRACE] = ACTIONS(1620),
    [anon_sym_constraint] = ACTIONS(1618),
    [anon_sym_val] = ACTIONS(1618),
    [anon_sym_end] = ACTIONS(1618),
    [anon_sym_with] = ACTIONS(1618),
    [anon_sym_COLON_GT] = ACTIONS(1620),
    [anon_sym_object] = ACTIONS(1618),
    [anon_sym_inherit] = ACTIONS(1618),
    [anon_sym_method] = ACTIONS(1618),
    [anon_sym_initializer] = ACTIONS(1618),
    [anon_sym_AMP] = ACTIONS(1618),
    [anon_sym_POUND] = ACTIONS(1618),
    [anon_sym_COLON_COLON] = ACTIONS(1620),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1620),
    [anon_sym_then] = ACTIONS(1618),
    [anon_sym_else] = ACTIONS(1618),
    [anon_sym_do] = ACTIONS(1618),
    [anon_sym_done] = ACTIONS(1618),
    [anon_sym_to] = ACTIONS(1618),
    [anon_sym_downto] = ACTIONS(1618),
    [anon_sym_new] = ACTIONS(1618),
    [anon_sym_LBRACE_LT] = ACTIONS(1620),
    [anon_sym_GT_RBRACE] = ACTIONS(1620),
    [anon_sym_begin] = ACTIONS(1618),
    [sym_ocamlyacc_value] = ACTIONS(1620),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1618),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1620),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1618),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1618),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1620),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1620),
    [aux_sym_number_token1] = ACTIONS(1620),
    [aux_sym_signed_number_token1] = ACTIONS(1618),
    [anon_sym_SQUOTE] = ACTIONS(1620),
    [anon_sym_DQUOTE] = ACTIONS(1620),
    [sym_prefix_operator] = ACTIONS(1620),
    [aux_sym_sign_operator_token1] = ACTIONS(1618),
    [sym_hash_operator] = ACTIONS(1620),
    [anon_sym_lsl] = ACTIONS(1618),
    [anon_sym_lsr] = ACTIONS(1618),
    [anon_sym_asr] = ACTIONS(1618),
    [aux_sym_pow_operator_token1] = ACTIONS(1620),
    [anon_sym_mod] = ACTIONS(1618),
    [anon_sym_land] = ACTIONS(1618),
    [anon_sym_lor] = ACTIONS(1618),
    [anon_sym_lxor] = ACTIONS(1618),
    [aux_sym_mult_operator_token1] = ACTIONS(1618),
    [aux_sym_add_operator_token1] = ACTIONS(1618),
    [sym_concat_operator] = ACTIONS(1620),
    [sym_rel_operator] = ACTIONS(1618),
    [anon_sym_AMP_AMP] = ACTIONS(1618),
    [anon_sym_or] = ACTIONS(1618),
    [anon_sym_PIPE_PIPE] = ACTIONS(1618),
    [sym_assign_operator] = ACTIONS(1620),
    [sym__uppercase_identifier] = ACTIONS(1620),
    [aux_sym_tag_token1] = ACTIONS(1620),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(955)] = {
    [sym_parameter] = STATE(2236),
    [sym__class_typed] = STATE(3847),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(955),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3662),
    [aux_sym_let_binding_repeat1] = STATE(848),
    [ts_builtin_sym_end] = ACTIONS(2095),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(2095),
    [anon_sym_and] = ACTIONS(2097),
    [anon_sym_EQ] = ACTIONS(2151),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(2097),
    [anon_sym_type] = ACTIONS(2097),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(2095),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(2097),
    [anon_sym_module] = ACTIONS(2097),
    [anon_sym_open] = ACTIONS(2097),
    [anon_sym_include] = ACTIONS(2097),
    [anon_sym_class] = ACTIONS(2097),
    [anon_sym_val] = ACTIONS(2097),
    [anon_sym_end] = ACTIONS(2097),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2095),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2095),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2095),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(956)] = {
    [sym_attribute] = STATE(956),
    [sym__lowercase_identifier] = ACTIONS(1566),
    [anon_sym_SEMI_SEMI] = ACTIONS(1568),
    [anon_sym_let] = ACTIONS(1566),
    [anon_sym_and] = ACTIONS(1566),
    [anon_sym_TILDE] = ACTIONS(1566),
    [anon_sym_QMARK] = ACTIONS(1566),
    [anon_sym_LPAREN] = ACTIONS(1568),
    [anon_sym_external] = ACTIONS(1566),
    [anon_sym_type] = ACTIONS(1566),
    [anon_sym_COMMA] = ACTIONS(1568),
    [anon_sym_PIPE] = ACTIONS(1566),
    [anon_sym_DOT] = ACTIONS(1568),
    [anon_sym_LBRACK] = ACTIONS(1566),
    [anon_sym_RBRACK] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1566),
    [anon_sym_false] = ACTIONS(1566),
    [anon_sym_LBRACE] = ACTIONS(1566),
    [anon_sym_SEMI] = ACTIONS(1566),
    [anon_sym_exception] = ACTIONS(1566),
    [anon_sym_module] = ACTIONS(1566),
    [anon_sym_open] = ACTIONS(1566),
    [anon_sym_include] = ACTIONS(1566),
    [anon_sym_class] = ACTIONS(1566),
    [anon_sym_val] = ACTIONS(1566),
    [anon_sym_end] = ACTIONS(1566),
    [anon_sym_object] = ACTIONS(1566),
    [anon_sym_in] = ACTIONS(1566),
    [anon_sym_AMP] = ACTIONS(1566),
    [anon_sym_POUND] = ACTIONS(1566),
    [anon_sym_COLON_COLON] = ACTIONS(1568),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1568),
    [anon_sym_LT_DASH] = ACTIONS(1566),
    [anon_sym_else] = ACTIONS(1566),
    [anon_sym_new] = ACTIONS(1566),
    [anon_sym_LBRACE_LT] = ACTIONS(1568),
    [anon_sym_begin] = ACTIONS(1566),
    [sym_ocamlyacc_value] = ACTIONS(1568),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1566),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1568),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1566),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1566),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1568),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1568),
    [aux_sym_number_token1] = ACTIONS(1568),
    [aux_sym_signed_number_token1] = ACTIONS(1566),
    [anon_sym_SQUOTE] = ACTIONS(1568),
    [anon_sym_DQUOTE] = ACTIONS(1568),
    [sym_prefix_operator] = ACTIONS(1568),
    [aux_sym_sign_operator_token1] = ACTIONS(1566),
    [sym_hash_operator] = ACTIONS(1568),
    [anon_sym_lsl] = ACTIONS(1566),
    [anon_sym_lsr] = ACTIONS(1566),
    [anon_sym_asr] = ACTIONS(1566),
    [aux_sym_pow_operator_token1] = ACTIONS(1568),
    [anon_sym_mod] = ACTIONS(1566),
    [anon_sym_land] = ACTIONS(1566),
    [anon_sym_lor] = ACTIONS(1566),
    [anon_sym_lxor] = ACTIONS(1566),
    [aux_sym_mult_operator_token1] = ACTIONS(1566),
    [aux_sym_add_operator_token1] = ACTIONS(1566),
    [sym_concat_operator] = ACTIONS(1568),
    [sym_rel_operator] = ACTIONS(1566),
    [anon_sym_AMP_AMP] = ACTIONS(1566),
    [anon_sym_or] = ACTIONS(1566),
    [anon_sym_PIPE_PIPE] = ACTIONS(1566),
    [sym_assign_operator] = ACTIONS(1568),
    [sym_let_operator] = ACTIONS(1568),
    [sym_let_and_operator] = ACTIONS(1568),
    [sym__uppercase_identifier] = ACTIONS(1568),
    [aux_sym_directive_token1] = ACTIONS(1566),
    [aux_sym_tag_token1] = ACTIONS(1568),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(957)] = {
    [sym_parameter] = STATE(2236),
    [sym__class_typed] = STATE(3877),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(957),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3658),
    [aux_sym_let_binding_repeat1] = STATE(959),
    [ts_builtin_sym_end] = ACTIONS(2127),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(2127),
    [anon_sym_and] = ACTIONS(2129),
    [anon_sym_EQ] = ACTIONS(2155),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(2129),
    [anon_sym_type] = ACTIONS(2129),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(2127),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(2129),
    [anon_sym_module] = ACTIONS(2129),
    [anon_sym_open] = ACTIONS(2129),
    [anon_sym_include] = ACTIONS(2129),
    [anon_sym_class] = ACTIONS(2129),
    [anon_sym_val] = ACTIONS(2129),
    [anon_sym_end] = ACTIONS(2129),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2127),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2127),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2127),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(958)] = {
    [sym_parameter] = STATE(2236),
    [sym__class_typed] = STATE(3871),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(958),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3617),
    [aux_sym_let_binding_repeat1] = STATE(972),
    [ts_builtin_sym_end] = ACTIONS(2031),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(2031),
    [anon_sym_and] = ACTIONS(2033),
    [anon_sym_EQ] = ACTIONS(2157),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(2033),
    [anon_sym_type] = ACTIONS(2033),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(2031),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(2033),
    [anon_sym_module] = ACTIONS(2033),
    [anon_sym_open] = ACTIONS(2033),
    [anon_sym_include] = ACTIONS(2033),
    [anon_sym_class] = ACTIONS(2033),
    [anon_sym_val] = ACTIONS(2033),
    [anon_sym_end] = ACTIONS(2033),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2031),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2031),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2031),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(959)] = {
    [sym_parameter] = STATE(2236),
    [sym__class_typed] = STATE(3845),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(959),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3645),
    [aux_sym_let_binding_repeat1] = STATE(848),
    [ts_builtin_sym_end] = ACTIONS(2073),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(2073),
    [anon_sym_and] = ACTIONS(2075),
    [anon_sym_EQ] = ACTIONS(2159),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(2075),
    [anon_sym_type] = ACTIONS(2075),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(2073),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(2075),
    [anon_sym_module] = ACTIONS(2075),
    [anon_sym_open] = ACTIONS(2075),
    [anon_sym_include] = ACTIONS(2075),
    [anon_sym_class] = ACTIONS(2075),
    [anon_sym_val] = ACTIONS(2075),
    [anon_sym_end] = ACTIONS(2075),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2073),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2073),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2073),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(960)] = {
    [sym_attribute] = STATE(960),
    [sym__lowercase_identifier] = ACTIONS(2161),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_QMARK] = ACTIONS(2161),
    [anon_sym_COLON] = ACTIONS(2163),
    [anon_sym_LPAREN] = ACTIONS(2165),
    [anon_sym_RPAREN] = ACTIONS(2165),
    [anon_sym_COMMA] = ACTIONS(2165),
    [anon_sym_PIPE] = ACTIONS(2161),
    [anon_sym_COLON2] = ACTIONS(2161),
    [anon_sym_DASH_GT] = ACTIONS(2161),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [anon_sym_RBRACK] = ACTIONS(2165),
    [anon_sym_true] = ACTIONS(2161),
    [anon_sym_false] = ACTIONS(2161),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [anon_sym_SEMI] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2165),
    [anon_sym_constraint] = ACTIONS(2161),
    [anon_sym_val] = ACTIONS(2161),
    [anon_sym_end] = ACTIONS(2161),
    [anon_sym_with] = ACTIONS(2161),
    [anon_sym_COLON_GT] = ACTIONS(2165),
    [anon_sym_object] = ACTIONS(2161),
    [anon_sym_inherit] = ACTIONS(2161),
    [anon_sym_method] = ACTIONS(2161),
    [anon_sym_initializer] = ACTIONS(2161),
    [anon_sym_AMP] = ACTIONS(2161),
    [anon_sym_COLON_COLON] = ACTIONS(2165),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2165),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2165),
    [anon_sym_then] = ACTIONS(2161),
    [anon_sym_else] = ACTIONS(2161),
    [anon_sym_do] = ACTIONS(2161),
    [anon_sym_done] = ACTIONS(2161),
    [anon_sym_to] = ACTIONS(2161),
    [anon_sym_downto] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2161),
    [anon_sym_LBRACE_LT] = ACTIONS(2165),
    [anon_sym_GT_RBRACE] = ACTIONS(2165),
    [anon_sym_begin] = ACTIONS(2161),
    [sym_ocamlyacc_value] = ACTIONS(2165),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2161),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2165),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2165),
    [aux_sym_number_token1] = ACTIONS(2165),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2165),
    [anon_sym_DQUOTE] = ACTIONS(2165),
    [sym_prefix_operator] = ACTIONS(2165),
    [aux_sym_sign_operator_token1] = ACTIONS(2161),
    [anon_sym_lsl] = ACTIONS(2161),
    [anon_sym_lsr] = ACTIONS(2161),
    [anon_sym_asr] = ACTIONS(2161),
    [aux_sym_pow_operator_token1] = ACTIONS(2165),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_land] = ACTIONS(2161),
    [anon_sym_lor] = ACTIONS(2161),
    [anon_sym_lxor] = ACTIONS(2161),
    [aux_sym_mult_operator_token1] = ACTIONS(2161),
    [aux_sym_add_operator_token1] = ACTIONS(2161),
    [sym_concat_operator] = ACTIONS(2165),
    [sym_rel_operator] = ACTIONS(2161),
    [anon_sym_AMP_AMP] = ACTIONS(2161),
    [anon_sym_or] = ACTIONS(2161),
    [anon_sym_PIPE_PIPE] = ACTIONS(2161),
    [sym_assign_operator] = ACTIONS(2165),
    [sym__uppercase_identifier] = ACTIONS(2165),
    [aux_sym_tag_token1] = ACTIONS(2165),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(961)] = {
    [sym_parameter] = STATE(2236),
    [sym__class_typed] = STATE(3851),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(961),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3663),
    [aux_sym_let_binding_repeat1] = STATE(968),
    [ts_builtin_sym_end] = ACTIONS(2101),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(2101),
    [anon_sym_and] = ACTIONS(2103),
    [anon_sym_EQ] = ACTIONS(2167),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(2103),
    [anon_sym_type] = ACTIONS(2103),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(2101),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(2103),
    [anon_sym_module] = ACTIONS(2103),
    [anon_sym_open] = ACTIONS(2103),
    [anon_sym_include] = ACTIONS(2103),
    [anon_sym_class] = ACTIONS(2103),
    [anon_sym_val] = ACTIONS(2103),
    [anon_sym_end] = ACTIONS(2103),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2101),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2101),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2101),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(962)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4359),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(962),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_EQ] = ACTIONS(2139),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(2139),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2139),
    [anon_sym_COLON2] = ACTIONS(2145),
    [anon_sym_DASH_GT] = ACTIONS(2139),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(2139),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_as] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_when] = ACTIONS(2145),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2149),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(963)] = {
    [sym_parameter] = STATE(2236),
    [sym__class_typed] = STATE(3863),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(963),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3707),
    [aux_sym_let_binding_repeat1] = STATE(848),
    [ts_builtin_sym_end] = ACTIONS(2133),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(2133),
    [anon_sym_and] = ACTIONS(2135),
    [anon_sym_EQ] = ACTIONS(2169),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(2135),
    [anon_sym_type] = ACTIONS(2135),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(2133),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(2135),
    [anon_sym_module] = ACTIONS(2135),
    [anon_sym_open] = ACTIONS(2135),
    [anon_sym_include] = ACTIONS(2135),
    [anon_sym_class] = ACTIONS(2135),
    [anon_sym_val] = ACTIONS(2135),
    [anon_sym_end] = ACTIONS(2135),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2133),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2133),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2133),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(964)] = {
    [sym_attribute] = STATE(964),
    [sym__lowercase_identifier] = ACTIONS(1578),
    [anon_sym_SEMI_SEMI] = ACTIONS(1580),
    [anon_sym_let] = ACTIONS(1578),
    [anon_sym_and] = ACTIONS(1578),
    [anon_sym_TILDE] = ACTIONS(1578),
    [anon_sym_QMARK] = ACTIONS(1578),
    [anon_sym_LPAREN] = ACTIONS(1580),
    [anon_sym_external] = ACTIONS(1578),
    [anon_sym_type] = ACTIONS(1578),
    [anon_sym_COMMA] = ACTIONS(1580),
    [anon_sym_PIPE] = ACTIONS(1578),
    [anon_sym_DOT] = ACTIONS(1580),
    [anon_sym_LBRACK] = ACTIONS(1578),
    [anon_sym_RBRACK] = ACTIONS(1580),
    [anon_sym_true] = ACTIONS(1578),
    [anon_sym_false] = ACTIONS(1578),
    [anon_sym_LBRACE] = ACTIONS(1578),
    [anon_sym_SEMI] = ACTIONS(1578),
    [anon_sym_exception] = ACTIONS(1578),
    [anon_sym_module] = ACTIONS(1578),
    [anon_sym_open] = ACTIONS(1578),
    [anon_sym_include] = ACTIONS(1578),
    [anon_sym_class] = ACTIONS(1578),
    [anon_sym_val] = ACTIONS(1578),
    [anon_sym_end] = ACTIONS(1578),
    [anon_sym_object] = ACTIONS(1578),
    [anon_sym_in] = ACTIONS(1578),
    [anon_sym_AMP] = ACTIONS(1578),
    [anon_sym_POUND] = ACTIONS(1578),
    [anon_sym_COLON_COLON] = ACTIONS(1580),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1580),
    [anon_sym_LT_DASH] = ACTIONS(1578),
    [anon_sym_else] = ACTIONS(1578),
    [anon_sym_new] = ACTIONS(1578),
    [anon_sym_LBRACE_LT] = ACTIONS(1580),
    [anon_sym_begin] = ACTIONS(1578),
    [sym_ocamlyacc_value] = ACTIONS(1580),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1578),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1580),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1578),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1578),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1580),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1580),
    [aux_sym_number_token1] = ACTIONS(1580),
    [aux_sym_signed_number_token1] = ACTIONS(1578),
    [anon_sym_SQUOTE] = ACTIONS(1580),
    [anon_sym_DQUOTE] = ACTIONS(1580),
    [sym_prefix_operator] = ACTIONS(1580),
    [aux_sym_sign_operator_token1] = ACTIONS(1578),
    [sym_hash_operator] = ACTIONS(1580),
    [anon_sym_lsl] = ACTIONS(1578),
    [anon_sym_lsr] = ACTIONS(1578),
    [anon_sym_asr] = ACTIONS(1578),
    [aux_sym_pow_operator_token1] = ACTIONS(1580),
    [anon_sym_mod] = ACTIONS(1578),
    [anon_sym_land] = ACTIONS(1578),
    [anon_sym_lor] = ACTIONS(1578),
    [anon_sym_lxor] = ACTIONS(1578),
    [aux_sym_mult_operator_token1] = ACTIONS(1578),
    [aux_sym_add_operator_token1] = ACTIONS(1578),
    [sym_concat_operator] = ACTIONS(1580),
    [sym_rel_operator] = ACTIONS(1578),
    [anon_sym_AMP_AMP] = ACTIONS(1578),
    [anon_sym_or] = ACTIONS(1578),
    [anon_sym_PIPE_PIPE] = ACTIONS(1578),
    [sym_assign_operator] = ACTIONS(1580),
    [sym_let_operator] = ACTIONS(1580),
    [sym_let_and_operator] = ACTIONS(1580),
    [sym__uppercase_identifier] = ACTIONS(1580),
    [aux_sym_directive_token1] = ACTIONS(1578),
    [aux_sym_tag_token1] = ACTIONS(1580),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(965)] = {
    [sym_attribute] = STATE(965),
    [sym__lowercase_identifier] = ACTIONS(1546),
    [anon_sym_SEMI_SEMI] = ACTIONS(1548),
    [anon_sym_let] = ACTIONS(1546),
    [anon_sym_and] = ACTIONS(1546),
    [anon_sym_TILDE] = ACTIONS(1546),
    [anon_sym_QMARK] = ACTIONS(1546),
    [anon_sym_LPAREN] = ACTIONS(1548),
    [anon_sym_external] = ACTIONS(1546),
    [anon_sym_type] = ACTIONS(1546),
    [anon_sym_COMMA] = ACTIONS(1548),
    [anon_sym_PIPE] = ACTIONS(1546),
    [anon_sym_DOT] = ACTIONS(1548),
    [anon_sym_LBRACK] = ACTIONS(1546),
    [anon_sym_RBRACK] = ACTIONS(1548),
    [anon_sym_true] = ACTIONS(1546),
    [anon_sym_false] = ACTIONS(1546),
    [anon_sym_LBRACE] = ACTIONS(1546),
    [anon_sym_SEMI] = ACTIONS(1546),
    [anon_sym_exception] = ACTIONS(1546),
    [anon_sym_module] = ACTIONS(1546),
    [anon_sym_open] = ACTIONS(1546),
    [anon_sym_include] = ACTIONS(1546),
    [anon_sym_class] = ACTIONS(1546),
    [anon_sym_val] = ACTIONS(1546),
    [anon_sym_end] = ACTIONS(1546),
    [anon_sym_object] = ACTIONS(1546),
    [anon_sym_in] = ACTIONS(1546),
    [anon_sym_AMP] = ACTIONS(1546),
    [anon_sym_POUND] = ACTIONS(1546),
    [anon_sym_COLON_COLON] = ACTIONS(1548),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1548),
    [anon_sym_LT_DASH] = ACTIONS(1546),
    [anon_sym_else] = ACTIONS(1546),
    [anon_sym_new] = ACTIONS(1546),
    [anon_sym_LBRACE_LT] = ACTIONS(1548),
    [anon_sym_begin] = ACTIONS(1546),
    [sym_ocamlyacc_value] = ACTIONS(1548),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1546),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1548),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1546),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1546),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1548),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1548),
    [aux_sym_number_token1] = ACTIONS(1548),
    [aux_sym_signed_number_token1] = ACTIONS(1546),
    [anon_sym_SQUOTE] = ACTIONS(1548),
    [anon_sym_DQUOTE] = ACTIONS(1548),
    [sym_prefix_operator] = ACTIONS(1548),
    [aux_sym_sign_operator_token1] = ACTIONS(1546),
    [sym_hash_operator] = ACTIONS(1548),
    [anon_sym_lsl] = ACTIONS(1546),
    [anon_sym_lsr] = ACTIONS(1546),
    [anon_sym_asr] = ACTIONS(1546),
    [aux_sym_pow_operator_token1] = ACTIONS(1548),
    [anon_sym_mod] = ACTIONS(1546),
    [anon_sym_land] = ACTIONS(1546),
    [anon_sym_lor] = ACTIONS(1546),
    [anon_sym_lxor] = ACTIONS(1546),
    [aux_sym_mult_operator_token1] = ACTIONS(1546),
    [aux_sym_add_operator_token1] = ACTIONS(1546),
    [sym_concat_operator] = ACTIONS(1548),
    [sym_rel_operator] = ACTIONS(1546),
    [anon_sym_AMP_AMP] = ACTIONS(1546),
    [anon_sym_or] = ACTIONS(1546),
    [anon_sym_PIPE_PIPE] = ACTIONS(1546),
    [sym_assign_operator] = ACTIONS(1548),
    [sym_let_operator] = ACTIONS(1548),
    [sym_let_and_operator] = ACTIONS(1548),
    [sym__uppercase_identifier] = ACTIONS(1548),
    [aux_sym_directive_token1] = ACTIONS(1546),
    [aux_sym_tag_token1] = ACTIONS(1548),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(966)] = {
    [sym_attribute] = STATE(966),
    [sym__lowercase_identifier] = ACTIONS(1614),
    [anon_sym_SEMI_SEMI] = ACTIONS(1616),
    [anon_sym_let] = ACTIONS(1614),
    [anon_sym_and] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_external] = ACTIONS(1614),
    [anon_sym_type] = ACTIONS(1614),
    [anon_sym_COMMA] = ACTIONS(1616),
    [anon_sym_PIPE] = ACTIONS(1614),
    [anon_sym_DOT] = ACTIONS(1616),
    [anon_sym_LBRACK] = ACTIONS(1614),
    [anon_sym_RBRACK] = ACTIONS(1616),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1614),
    [anon_sym_exception] = ACTIONS(1614),
    [anon_sym_module] = ACTIONS(1614),
    [anon_sym_open] = ACTIONS(1614),
    [anon_sym_include] = ACTIONS(1614),
    [anon_sym_class] = ACTIONS(1614),
    [anon_sym_val] = ACTIONS(1614),
    [anon_sym_end] = ACTIONS(1614),
    [anon_sym_object] = ACTIONS(1614),
    [anon_sym_in] = ACTIONS(1614),
    [anon_sym_AMP] = ACTIONS(1614),
    [anon_sym_POUND] = ACTIONS(1614),
    [anon_sym_COLON_COLON] = ACTIONS(1616),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1616),
    [anon_sym_LT_DASH] = ACTIONS(2009),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
    [anon_sym_LBRACE_LT] = ACTIONS(1616),
    [anon_sym_begin] = ACTIONS(1614),
    [sym_ocamlyacc_value] = ACTIONS(1616),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1614),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1616),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1616),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1616),
    [aux_sym_number_token1] = ACTIONS(1616),
    [aux_sym_signed_number_token1] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1616),
    [sym_prefix_operator] = ACTIONS(1616),
    [aux_sym_sign_operator_token1] = ACTIONS(1614),
    [sym_hash_operator] = ACTIONS(1616),
    [anon_sym_lsl] = ACTIONS(1614),
    [anon_sym_lsr] = ACTIONS(1614),
    [anon_sym_asr] = ACTIONS(1614),
    [aux_sym_pow_operator_token1] = ACTIONS(1616),
    [anon_sym_mod] = ACTIONS(1614),
    [anon_sym_land] = ACTIONS(1614),
    [anon_sym_lor] = ACTIONS(1614),
    [anon_sym_lxor] = ACTIONS(1614),
    [aux_sym_mult_operator_token1] = ACTIONS(1614),
    [aux_sym_add_operator_token1] = ACTIONS(1614),
    [sym_concat_operator] = ACTIONS(1616),
    [sym_rel_operator] = ACTIONS(1614),
    [anon_sym_AMP_AMP] = ACTIONS(1614),
    [anon_sym_or] = ACTIONS(1614),
    [anon_sym_PIPE_PIPE] = ACTIONS(1614),
    [sym_assign_operator] = ACTIONS(1616),
    [sym_let_operator] = ACTIONS(1616),
    [sym_let_and_operator] = ACTIONS(1616),
    [sym__uppercase_identifier] = ACTIONS(1616),
    [aux_sym_directive_token1] = ACTIONS(1614),
    [aux_sym_tag_token1] = ACTIONS(1616),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(967)] = {
    [sym_attribute] = STATE(967),
    [sym__lowercase_identifier] = ACTIONS(1724),
    [anon_sym_SEMI_SEMI] = ACTIONS(1726),
    [anon_sym_let] = ACTIONS(1724),
    [anon_sym_and] = ACTIONS(1724),
    [anon_sym_TILDE] = ACTIONS(1724),
    [anon_sym_QMARK] = ACTIONS(1724),
    [anon_sym_LPAREN] = ACTIONS(1726),
    [anon_sym_external] = ACTIONS(1724),
    [anon_sym_type] = ACTIONS(1724),
    [anon_sym_COMMA] = ACTIONS(1726),
    [anon_sym_PIPE] = ACTIONS(1724),
    [anon_sym_DOT] = ACTIONS(1726),
    [anon_sym_LBRACK] = ACTIONS(1724),
    [anon_sym_RBRACK] = ACTIONS(1726),
    [anon_sym_true] = ACTIONS(1724),
    [anon_sym_false] = ACTIONS(1724),
    [anon_sym_LBRACE] = ACTIONS(1724),
    [anon_sym_SEMI] = ACTIONS(1724),
    [anon_sym_exception] = ACTIONS(1724),
    [anon_sym_module] = ACTIONS(1724),
    [anon_sym_open] = ACTIONS(1724),
    [anon_sym_include] = ACTIONS(1724),
    [anon_sym_class] = ACTIONS(1724),
    [anon_sym_val] = ACTIONS(1724),
    [anon_sym_end] = ACTIONS(1724),
    [anon_sym_object] = ACTIONS(1724),
    [anon_sym_in] = ACTIONS(1724),
    [anon_sym_AMP] = ACTIONS(1724),
    [anon_sym_POUND] = ACTIONS(1724),
    [anon_sym_COLON_COLON] = ACTIONS(1726),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1726),
    [anon_sym_LT_DASH] = ACTIONS(2171),
    [anon_sym_else] = ACTIONS(1724),
    [anon_sym_new] = ACTIONS(1724),
    [anon_sym_LBRACE_LT] = ACTIONS(1726),
    [anon_sym_begin] = ACTIONS(1724),
    [sym_ocamlyacc_value] = ACTIONS(1726),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1724),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1726),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1726),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1726),
    [aux_sym_number_token1] = ACTIONS(1726),
    [aux_sym_signed_number_token1] = ACTIONS(1724),
    [anon_sym_SQUOTE] = ACTIONS(1726),
    [anon_sym_DQUOTE] = ACTIONS(1726),
    [sym_prefix_operator] = ACTIONS(1726),
    [aux_sym_sign_operator_token1] = ACTIONS(1724),
    [sym_hash_operator] = ACTIONS(1726),
    [anon_sym_lsl] = ACTIONS(1724),
    [anon_sym_lsr] = ACTIONS(1724),
    [anon_sym_asr] = ACTIONS(1724),
    [aux_sym_pow_operator_token1] = ACTIONS(1726),
    [anon_sym_mod] = ACTIONS(1724),
    [anon_sym_land] = ACTIONS(1724),
    [anon_sym_lor] = ACTIONS(1724),
    [anon_sym_lxor] = ACTIONS(1724),
    [aux_sym_mult_operator_token1] = ACTIONS(1724),
    [aux_sym_add_operator_token1] = ACTIONS(1724),
    [sym_concat_operator] = ACTIONS(1726),
    [sym_rel_operator] = ACTIONS(1724),
    [anon_sym_AMP_AMP] = ACTIONS(1724),
    [anon_sym_or] = ACTIONS(1724),
    [anon_sym_PIPE_PIPE] = ACTIONS(1724),
    [sym_assign_operator] = ACTIONS(1726),
    [sym_let_operator] = ACTIONS(1726),
    [sym_let_and_operator] = ACTIONS(1726),
    [sym__uppercase_identifier] = ACTIONS(1726),
    [aux_sym_directive_token1] = ACTIONS(1724),
    [aux_sym_tag_token1] = ACTIONS(1726),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(968)] = {
    [sym_parameter] = STATE(2236),
    [sym__class_typed] = STATE(3870),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(968),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3688),
    [aux_sym_let_binding_repeat1] = STATE(848),
    [ts_builtin_sym_end] = ACTIONS(2111),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(2111),
    [anon_sym_and] = ACTIONS(2113),
    [anon_sym_EQ] = ACTIONS(2173),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(2113),
    [anon_sym_type] = ACTIONS(2113),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(2111),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(2113),
    [anon_sym_module] = ACTIONS(2113),
    [anon_sym_open] = ACTIONS(2113),
    [anon_sym_include] = ACTIONS(2113),
    [anon_sym_class] = ACTIONS(2113),
    [anon_sym_val] = ACTIONS(2113),
    [anon_sym_end] = ACTIONS(2113),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2111),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2111),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2111),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(969)] = {
    [sym_attribute] = STATE(969),
    [sym__lowercase_identifier] = ACTIONS(1574),
    [anon_sym_SEMI_SEMI] = ACTIONS(1576),
    [anon_sym_let] = ACTIONS(1574),
    [anon_sym_and] = ACTIONS(1574),
    [anon_sym_TILDE] = ACTIONS(1574),
    [anon_sym_QMARK] = ACTIONS(1574),
    [anon_sym_LPAREN] = ACTIONS(1576),
    [anon_sym_external] = ACTIONS(1574),
    [anon_sym_type] = ACTIONS(1574),
    [anon_sym_COMMA] = ACTIONS(1576),
    [anon_sym_PIPE] = ACTIONS(1574),
    [anon_sym_DOT] = ACTIONS(1576),
    [anon_sym_LBRACK] = ACTIONS(1574),
    [anon_sym_RBRACK] = ACTIONS(1576),
    [anon_sym_true] = ACTIONS(1574),
    [anon_sym_false] = ACTIONS(1574),
    [anon_sym_LBRACE] = ACTIONS(1574),
    [anon_sym_SEMI] = ACTIONS(1574),
    [anon_sym_exception] = ACTIONS(1574),
    [anon_sym_module] = ACTIONS(1574),
    [anon_sym_open] = ACTIONS(1574),
    [anon_sym_include] = ACTIONS(1574),
    [anon_sym_class] = ACTIONS(1574),
    [anon_sym_val] = ACTIONS(1574),
    [anon_sym_end] = ACTIONS(1574),
    [anon_sym_object] = ACTIONS(1574),
    [anon_sym_in] = ACTIONS(1574),
    [anon_sym_AMP] = ACTIONS(1574),
    [anon_sym_POUND] = ACTIONS(1574),
    [anon_sym_COLON_COLON] = ACTIONS(1576),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1576),
    [anon_sym_LT_DASH] = ACTIONS(1574),
    [anon_sym_else] = ACTIONS(1574),
    [anon_sym_new] = ACTIONS(1574),
    [anon_sym_LBRACE_LT] = ACTIONS(1576),
    [anon_sym_begin] = ACTIONS(1574),
    [sym_ocamlyacc_value] = ACTIONS(1576),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1574),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1576),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1574),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1574),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1576),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1576),
    [aux_sym_number_token1] = ACTIONS(1576),
    [aux_sym_signed_number_token1] = ACTIONS(1574),
    [anon_sym_SQUOTE] = ACTIONS(1576),
    [anon_sym_DQUOTE] = ACTIONS(1576),
    [sym_prefix_operator] = ACTIONS(1576),
    [aux_sym_sign_operator_token1] = ACTIONS(1574),
    [sym_hash_operator] = ACTIONS(1576),
    [anon_sym_lsl] = ACTIONS(1574),
    [anon_sym_lsr] = ACTIONS(1574),
    [anon_sym_asr] = ACTIONS(1574),
    [aux_sym_pow_operator_token1] = ACTIONS(1576),
    [anon_sym_mod] = ACTIONS(1574),
    [anon_sym_land] = ACTIONS(1574),
    [anon_sym_lor] = ACTIONS(1574),
    [anon_sym_lxor] = ACTIONS(1574),
    [aux_sym_mult_operator_token1] = ACTIONS(1574),
    [aux_sym_add_operator_token1] = ACTIONS(1574),
    [sym_concat_operator] = ACTIONS(1576),
    [sym_rel_operator] = ACTIONS(1574),
    [anon_sym_AMP_AMP] = ACTIONS(1574),
    [anon_sym_or] = ACTIONS(1574),
    [anon_sym_PIPE_PIPE] = ACTIONS(1574),
    [sym_assign_operator] = ACTIONS(1576),
    [sym_let_operator] = ACTIONS(1576),
    [sym_let_and_operator] = ACTIONS(1576),
    [sym__uppercase_identifier] = ACTIONS(1576),
    [aux_sym_directive_token1] = ACTIONS(1574),
    [aux_sym_tag_token1] = ACTIONS(1576),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(970)] = {
    [sym_parameter] = STATE(2236),
    [sym__class_typed] = STATE(3855),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(970),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3610),
    [aux_sym_let_binding_repeat1] = STATE(955),
    [ts_builtin_sym_end] = ACTIONS(2089),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(2089),
    [anon_sym_and] = ACTIONS(2091),
    [anon_sym_EQ] = ACTIONS(2175),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(2091),
    [anon_sym_type] = ACTIONS(2091),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(2089),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(2091),
    [anon_sym_module] = ACTIONS(2091),
    [anon_sym_open] = ACTIONS(2091),
    [anon_sym_include] = ACTIONS(2091),
    [anon_sym_class] = ACTIONS(2091),
    [anon_sym_val] = ACTIONS(2091),
    [anon_sym_end] = ACTIONS(2091),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2089),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2089),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2089),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(971)] = {
    [sym_parameter] = STATE(2236),
    [sym__class_typed] = STATE(3872),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(971),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3689),
    [aux_sym_let_binding_repeat1] = STATE(963),
    [ts_builtin_sym_end] = ACTIONS(2117),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(2117),
    [anon_sym_and] = ACTIONS(2119),
    [anon_sym_EQ] = ACTIONS(2177),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(2119),
    [anon_sym_type] = ACTIONS(2119),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(2117),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(2119),
    [anon_sym_module] = ACTIONS(2119),
    [anon_sym_open] = ACTIONS(2119),
    [anon_sym_include] = ACTIONS(2119),
    [anon_sym_class] = ACTIONS(2119),
    [anon_sym_val] = ACTIONS(2119),
    [anon_sym_end] = ACTIONS(2119),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2117),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2117),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2117),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(972)] = {
    [sym_parameter] = STATE(2236),
    [sym__class_typed] = STATE(3841),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(972),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(3569),
    [aux_sym_let_binding_repeat1] = STATE(848),
    [ts_builtin_sym_end] = ACTIONS(2083),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_SEMI_SEMI] = ACTIONS(2083),
    [anon_sym_and] = ACTIONS(2085),
    [anon_sym_EQ] = ACTIONS(2179),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_external] = ACTIONS(2085),
    [anon_sym_type] = ACTIONS(2085),
    [anon_sym_COLON2] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_RBRACK] = ACTIONS(2083),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_exception] = ACTIONS(2085),
    [anon_sym_module] = ACTIONS(2085),
    [anon_sym_open] = ACTIONS(2085),
    [anon_sym_include] = ACTIONS(2085),
    [anon_sym_class] = ACTIONS(2085),
    [anon_sym_val] = ACTIONS(2085),
    [anon_sym_end] = ACTIONS(2085),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2083),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2083),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2083),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(973)] = {
    [sym_attribute] = STATE(973),
    [sym__lowercase_identifier] = ACTIONS(1554),
    [anon_sym_SEMI_SEMI] = ACTIONS(1556),
    [anon_sym_let] = ACTIONS(1554),
    [anon_sym_and] = ACTIONS(1554),
    [anon_sym_TILDE] = ACTIONS(1554),
    [anon_sym_QMARK] = ACTIONS(1554),
    [anon_sym_LPAREN] = ACTIONS(1556),
    [anon_sym_external] = ACTIONS(1554),
    [anon_sym_type] = ACTIONS(1554),
    [anon_sym_COMMA] = ACTIONS(1556),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_DOT] = ACTIONS(1556),
    [anon_sym_LBRACK] = ACTIONS(1554),
    [anon_sym_RBRACK] = ACTIONS(1556),
    [anon_sym_true] = ACTIONS(1554),
    [anon_sym_false] = ACTIONS(1554),
    [anon_sym_LBRACE] = ACTIONS(1554),
    [anon_sym_SEMI] = ACTIONS(1554),
    [anon_sym_exception] = ACTIONS(1554),
    [anon_sym_module] = ACTIONS(1554),
    [anon_sym_open] = ACTIONS(1554),
    [anon_sym_include] = ACTIONS(1554),
    [anon_sym_class] = ACTIONS(1554),
    [anon_sym_val] = ACTIONS(1554),
    [anon_sym_end] = ACTIONS(1554),
    [anon_sym_object] = ACTIONS(1554),
    [anon_sym_in] = ACTIONS(1554),
    [anon_sym_AMP] = ACTIONS(1554),
    [anon_sym_POUND] = ACTIONS(1554),
    [anon_sym_COLON_COLON] = ACTIONS(1556),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1556),
    [anon_sym_LT_DASH] = ACTIONS(1554),
    [anon_sym_else] = ACTIONS(1554),
    [anon_sym_new] = ACTIONS(1554),
    [anon_sym_LBRACE_LT] = ACTIONS(1556),
    [anon_sym_begin] = ACTIONS(1554),
    [sym_ocamlyacc_value] = ACTIONS(1556),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1554),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1556),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1554),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1554),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1556),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1556),
    [aux_sym_number_token1] = ACTIONS(1556),
    [aux_sym_signed_number_token1] = ACTIONS(1554),
    [anon_sym_SQUOTE] = ACTIONS(1556),
    [anon_sym_DQUOTE] = ACTIONS(1556),
    [sym_prefix_operator] = ACTIONS(1556),
    [aux_sym_sign_operator_token1] = ACTIONS(1554),
    [sym_hash_operator] = ACTIONS(1556),
    [anon_sym_lsl] = ACTIONS(1554),
    [anon_sym_lsr] = ACTIONS(1554),
    [anon_sym_asr] = ACTIONS(1554),
    [aux_sym_pow_operator_token1] = ACTIONS(1556),
    [anon_sym_mod] = ACTIONS(1554),
    [anon_sym_land] = ACTIONS(1554),
    [anon_sym_lor] = ACTIONS(1554),
    [anon_sym_lxor] = ACTIONS(1554),
    [aux_sym_mult_operator_token1] = ACTIONS(1554),
    [aux_sym_add_operator_token1] = ACTIONS(1554),
    [sym_concat_operator] = ACTIONS(1556),
    [sym_rel_operator] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1554),
    [anon_sym_or] = ACTIONS(1554),
    [anon_sym_PIPE_PIPE] = ACTIONS(1554),
    [sym_assign_operator] = ACTIONS(1556),
    [sym_let_operator] = ACTIONS(1556),
    [sym_let_and_operator] = ACTIONS(1556),
    [sym__uppercase_identifier] = ACTIONS(1556),
    [aux_sym_directive_token1] = ACTIONS(1554),
    [aux_sym_tag_token1] = ACTIONS(1556),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(974)] = {
    [sym_attribute] = STATE(974),
    [sym__lowercase_identifier] = ACTIONS(1550),
    [anon_sym_SEMI_SEMI] = ACTIONS(1552),
    [anon_sym_let] = ACTIONS(1550),
    [anon_sym_and] = ACTIONS(1550),
    [anon_sym_TILDE] = ACTIONS(1550),
    [anon_sym_QMARK] = ACTIONS(1550),
    [anon_sym_LPAREN] = ACTIONS(1552),
    [anon_sym_external] = ACTIONS(1550),
    [anon_sym_type] = ACTIONS(1550),
    [anon_sym_COMMA] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1550),
    [anon_sym_DOT] = ACTIONS(1552),
    [anon_sym_LBRACK] = ACTIONS(1550),
    [anon_sym_RBRACK] = ACTIONS(1552),
    [anon_sym_true] = ACTIONS(1550),
    [anon_sym_false] = ACTIONS(1550),
    [anon_sym_LBRACE] = ACTIONS(1550),
    [anon_sym_SEMI] = ACTIONS(1550),
    [anon_sym_exception] = ACTIONS(1550),
    [anon_sym_module] = ACTIONS(1550),
    [anon_sym_open] = ACTIONS(1550),
    [anon_sym_include] = ACTIONS(1550),
    [anon_sym_class] = ACTIONS(1550),
    [anon_sym_val] = ACTIONS(1550),
    [anon_sym_end] = ACTIONS(1550),
    [anon_sym_object] = ACTIONS(1550),
    [anon_sym_in] = ACTIONS(1550),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_POUND] = ACTIONS(1550),
    [anon_sym_COLON_COLON] = ACTIONS(1552),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1552),
    [anon_sym_LT_DASH] = ACTIONS(1550),
    [anon_sym_else] = ACTIONS(1550),
    [anon_sym_new] = ACTIONS(1550),
    [anon_sym_LBRACE_LT] = ACTIONS(1552),
    [anon_sym_begin] = ACTIONS(1550),
    [sym_ocamlyacc_value] = ACTIONS(1552),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1550),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1552),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1550),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1550),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1552),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1552),
    [aux_sym_number_token1] = ACTIONS(1552),
    [aux_sym_signed_number_token1] = ACTIONS(1550),
    [anon_sym_SQUOTE] = ACTIONS(1552),
    [anon_sym_DQUOTE] = ACTIONS(1552),
    [sym_prefix_operator] = ACTIONS(1552),
    [aux_sym_sign_operator_token1] = ACTIONS(1550),
    [sym_hash_operator] = ACTIONS(1552),
    [anon_sym_lsl] = ACTIONS(1550),
    [anon_sym_lsr] = ACTIONS(1550),
    [anon_sym_asr] = ACTIONS(1550),
    [aux_sym_pow_operator_token1] = ACTIONS(1552),
    [anon_sym_mod] = ACTIONS(1550),
    [anon_sym_land] = ACTIONS(1550),
    [anon_sym_lor] = ACTIONS(1550),
    [anon_sym_lxor] = ACTIONS(1550),
    [aux_sym_mult_operator_token1] = ACTIONS(1550),
    [aux_sym_add_operator_token1] = ACTIONS(1550),
    [sym_concat_operator] = ACTIONS(1552),
    [sym_rel_operator] = ACTIONS(1550),
    [anon_sym_AMP_AMP] = ACTIONS(1550),
    [anon_sym_or] = ACTIONS(1550),
    [anon_sym_PIPE_PIPE] = ACTIONS(1550),
    [sym_assign_operator] = ACTIONS(1552),
    [sym_let_operator] = ACTIONS(1552),
    [sym_let_and_operator] = ACTIONS(1552),
    [sym__uppercase_identifier] = ACTIONS(1552),
    [aux_sym_directive_token1] = ACTIONS(1550),
    [aux_sym_tag_token1] = ACTIONS(1552),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(975)] = {
    [sym_attribute] = STATE(975),
    [sym__lowercase_identifier] = ACTIONS(1570),
    [anon_sym_SEMI_SEMI] = ACTIONS(1572),
    [anon_sym_let] = ACTIONS(1570),
    [anon_sym_and] = ACTIONS(1570),
    [anon_sym_TILDE] = ACTIONS(1570),
    [anon_sym_QMARK] = ACTIONS(1570),
    [anon_sym_LPAREN] = ACTIONS(1572),
    [anon_sym_external] = ACTIONS(1570),
    [anon_sym_type] = ACTIONS(1570),
    [anon_sym_COMMA] = ACTIONS(1572),
    [anon_sym_PIPE] = ACTIONS(1570),
    [anon_sym_DOT] = ACTIONS(1572),
    [anon_sym_LBRACK] = ACTIONS(1570),
    [anon_sym_RBRACK] = ACTIONS(1572),
    [anon_sym_true] = ACTIONS(1570),
    [anon_sym_false] = ACTIONS(1570),
    [anon_sym_LBRACE] = ACTIONS(1570),
    [anon_sym_SEMI] = ACTIONS(1570),
    [anon_sym_exception] = ACTIONS(1570),
    [anon_sym_module] = ACTIONS(1570),
    [anon_sym_open] = ACTIONS(1570),
    [anon_sym_include] = ACTIONS(1570),
    [anon_sym_class] = ACTIONS(1570),
    [anon_sym_val] = ACTIONS(1570),
    [anon_sym_end] = ACTIONS(1570),
    [anon_sym_object] = ACTIONS(1570),
    [anon_sym_in] = ACTIONS(1570),
    [anon_sym_AMP] = ACTIONS(1570),
    [anon_sym_POUND] = ACTIONS(1570),
    [anon_sym_COLON_COLON] = ACTIONS(1572),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1572),
    [anon_sym_LT_DASH] = ACTIONS(1570),
    [anon_sym_else] = ACTIONS(1570),
    [anon_sym_new] = ACTIONS(1570),
    [anon_sym_LBRACE_LT] = ACTIONS(1572),
    [anon_sym_begin] = ACTIONS(1570),
    [sym_ocamlyacc_value] = ACTIONS(1572),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1570),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1572),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1570),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1570),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1572),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1572),
    [aux_sym_number_token1] = ACTIONS(1572),
    [aux_sym_signed_number_token1] = ACTIONS(1570),
    [anon_sym_SQUOTE] = ACTIONS(1572),
    [anon_sym_DQUOTE] = ACTIONS(1572),
    [sym_prefix_operator] = ACTIONS(1572),
    [aux_sym_sign_operator_token1] = ACTIONS(1570),
    [sym_hash_operator] = ACTIONS(1572),
    [anon_sym_lsl] = ACTIONS(1570),
    [anon_sym_lsr] = ACTIONS(1570),
    [anon_sym_asr] = ACTIONS(1570),
    [aux_sym_pow_operator_token1] = ACTIONS(1572),
    [anon_sym_mod] = ACTIONS(1570),
    [anon_sym_land] = ACTIONS(1570),
    [anon_sym_lor] = ACTIONS(1570),
    [anon_sym_lxor] = ACTIONS(1570),
    [aux_sym_mult_operator_token1] = ACTIONS(1570),
    [aux_sym_add_operator_token1] = ACTIONS(1570),
    [sym_concat_operator] = ACTIONS(1572),
    [sym_rel_operator] = ACTIONS(1570),
    [anon_sym_AMP_AMP] = ACTIONS(1570),
    [anon_sym_or] = ACTIONS(1570),
    [anon_sym_PIPE_PIPE] = ACTIONS(1570),
    [sym_assign_operator] = ACTIONS(1572),
    [sym_let_operator] = ACTIONS(1572),
    [sym_let_and_operator] = ACTIONS(1572),
    [sym__uppercase_identifier] = ACTIONS(1572),
    [aux_sym_directive_token1] = ACTIONS(1570),
    [aux_sym_tag_token1] = ACTIONS(1572),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(976)] = {
    [sym_attribute] = STATE(976),
    [sym__lowercase_identifier] = ACTIONS(1558),
    [anon_sym_SEMI_SEMI] = ACTIONS(1560),
    [anon_sym_let] = ACTIONS(1558),
    [anon_sym_and] = ACTIONS(1558),
    [anon_sym_TILDE] = ACTIONS(1558),
    [anon_sym_QMARK] = ACTIONS(1558),
    [anon_sym_LPAREN] = ACTIONS(1560),
    [anon_sym_external] = ACTIONS(1558),
    [anon_sym_type] = ACTIONS(1558),
    [anon_sym_COMMA] = ACTIONS(1560),
    [anon_sym_PIPE] = ACTIONS(1558),
    [anon_sym_DOT] = ACTIONS(1560),
    [anon_sym_LBRACK] = ACTIONS(1558),
    [anon_sym_RBRACK] = ACTIONS(1560),
    [anon_sym_true] = ACTIONS(1558),
    [anon_sym_false] = ACTIONS(1558),
    [anon_sym_LBRACE] = ACTIONS(1558),
    [anon_sym_SEMI] = ACTIONS(1558),
    [anon_sym_exception] = ACTIONS(1558),
    [anon_sym_module] = ACTIONS(1558),
    [anon_sym_open] = ACTIONS(1558),
    [anon_sym_include] = ACTIONS(1558),
    [anon_sym_class] = ACTIONS(1558),
    [anon_sym_val] = ACTIONS(1558),
    [anon_sym_end] = ACTIONS(1558),
    [anon_sym_object] = ACTIONS(1558),
    [anon_sym_in] = ACTIONS(1558),
    [anon_sym_AMP] = ACTIONS(1558),
    [anon_sym_POUND] = ACTIONS(1558),
    [anon_sym_COLON_COLON] = ACTIONS(1560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1560),
    [anon_sym_LT_DASH] = ACTIONS(1558),
    [anon_sym_else] = ACTIONS(1558),
    [anon_sym_new] = ACTIONS(1558),
    [anon_sym_LBRACE_LT] = ACTIONS(1560),
    [anon_sym_begin] = ACTIONS(1558),
    [sym_ocamlyacc_value] = ACTIONS(1560),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1558),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1560),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1558),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1558),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1560),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1560),
    [aux_sym_number_token1] = ACTIONS(1560),
    [aux_sym_signed_number_token1] = ACTIONS(1558),
    [anon_sym_SQUOTE] = ACTIONS(1560),
    [anon_sym_DQUOTE] = ACTIONS(1560),
    [sym_prefix_operator] = ACTIONS(1560),
    [aux_sym_sign_operator_token1] = ACTIONS(1558),
    [sym_hash_operator] = ACTIONS(1560),
    [anon_sym_lsl] = ACTIONS(1558),
    [anon_sym_lsr] = ACTIONS(1558),
    [anon_sym_asr] = ACTIONS(1558),
    [aux_sym_pow_operator_token1] = ACTIONS(1560),
    [anon_sym_mod] = ACTIONS(1558),
    [anon_sym_land] = ACTIONS(1558),
    [anon_sym_lor] = ACTIONS(1558),
    [anon_sym_lxor] = ACTIONS(1558),
    [aux_sym_mult_operator_token1] = ACTIONS(1558),
    [aux_sym_add_operator_token1] = ACTIONS(1558),
    [sym_concat_operator] = ACTIONS(1560),
    [sym_rel_operator] = ACTIONS(1558),
    [anon_sym_AMP_AMP] = ACTIONS(1558),
    [anon_sym_or] = ACTIONS(1558),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [sym_assign_operator] = ACTIONS(1560),
    [sym_let_operator] = ACTIONS(1560),
    [sym_let_and_operator] = ACTIONS(1560),
    [sym__uppercase_identifier] = ACTIONS(1560),
    [aux_sym_directive_token1] = ACTIONS(1558),
    [aux_sym_tag_token1] = ACTIONS(1560),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(977)] = {
    [sym_attribute] = STATE(977),
    [sym__lowercase_identifier] = ACTIONS(1562),
    [anon_sym_SEMI_SEMI] = ACTIONS(1564),
    [anon_sym_let] = ACTIONS(1562),
    [anon_sym_and] = ACTIONS(1562),
    [anon_sym_TILDE] = ACTIONS(1562),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_LPAREN] = ACTIONS(1564),
    [anon_sym_external] = ACTIONS(1562),
    [anon_sym_type] = ACTIONS(1562),
    [anon_sym_COMMA] = ACTIONS(1564),
    [anon_sym_PIPE] = ACTIONS(1562),
    [anon_sym_DOT] = ACTIONS(1564),
    [anon_sym_LBRACK] = ACTIONS(1562),
    [anon_sym_RBRACK] = ACTIONS(1564),
    [anon_sym_true] = ACTIONS(1562),
    [anon_sym_false] = ACTIONS(1562),
    [anon_sym_LBRACE] = ACTIONS(1562),
    [anon_sym_SEMI] = ACTIONS(1562),
    [anon_sym_exception] = ACTIONS(1562),
    [anon_sym_module] = ACTIONS(1562),
    [anon_sym_open] = ACTIONS(1562),
    [anon_sym_include] = ACTIONS(1562),
    [anon_sym_class] = ACTIONS(1562),
    [anon_sym_val] = ACTIONS(1562),
    [anon_sym_end] = ACTIONS(1562),
    [anon_sym_object] = ACTIONS(1562),
    [anon_sym_in] = ACTIONS(1562),
    [anon_sym_AMP] = ACTIONS(1562),
    [anon_sym_POUND] = ACTIONS(1562),
    [anon_sym_COLON_COLON] = ACTIONS(1564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1564),
    [anon_sym_LT_DASH] = ACTIONS(1562),
    [anon_sym_else] = ACTIONS(1562),
    [anon_sym_new] = ACTIONS(1562),
    [anon_sym_LBRACE_LT] = ACTIONS(1564),
    [anon_sym_begin] = ACTIONS(1562),
    [sym_ocamlyacc_value] = ACTIONS(1564),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1562),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1564),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1562),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1562),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1564),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1564),
    [aux_sym_number_token1] = ACTIONS(1564),
    [aux_sym_signed_number_token1] = ACTIONS(1562),
    [anon_sym_SQUOTE] = ACTIONS(1564),
    [anon_sym_DQUOTE] = ACTIONS(1564),
    [sym_prefix_operator] = ACTIONS(1564),
    [aux_sym_sign_operator_token1] = ACTIONS(1562),
    [sym_hash_operator] = ACTIONS(1564),
    [anon_sym_lsl] = ACTIONS(1562),
    [anon_sym_lsr] = ACTIONS(1562),
    [anon_sym_asr] = ACTIONS(1562),
    [aux_sym_pow_operator_token1] = ACTIONS(1564),
    [anon_sym_mod] = ACTIONS(1562),
    [anon_sym_land] = ACTIONS(1562),
    [anon_sym_lor] = ACTIONS(1562),
    [anon_sym_lxor] = ACTIONS(1562),
    [aux_sym_mult_operator_token1] = ACTIONS(1562),
    [aux_sym_add_operator_token1] = ACTIONS(1562),
    [sym_concat_operator] = ACTIONS(1564),
    [sym_rel_operator] = ACTIONS(1562),
    [anon_sym_AMP_AMP] = ACTIONS(1562),
    [anon_sym_or] = ACTIONS(1562),
    [anon_sym_PIPE_PIPE] = ACTIONS(1562),
    [sym_assign_operator] = ACTIONS(1564),
    [sym_let_operator] = ACTIONS(1564),
    [sym_let_and_operator] = ACTIONS(1564),
    [sym__uppercase_identifier] = ACTIONS(1564),
    [aux_sym_directive_token1] = ACTIONS(1562),
    [aux_sym_tag_token1] = ACTIONS(1564),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(978)] = {
    [sym_attribute] = STATE(978),
    [sym__lowercase_identifier] = ACTIONS(1740),
    [anon_sym_SEMI_SEMI] = ACTIONS(1742),
    [anon_sym_let] = ACTIONS(1740),
    [anon_sym_and] = ACTIONS(1740),
    [anon_sym_TILDE] = ACTIONS(1740),
    [anon_sym_QMARK] = ACTIONS(1740),
    [anon_sym_LPAREN] = ACTIONS(1742),
    [anon_sym_external] = ACTIONS(1740),
    [anon_sym_type] = ACTIONS(1740),
    [anon_sym_COMMA] = ACTIONS(1742),
    [anon_sym_PIPE] = ACTIONS(1740),
    [anon_sym_DOT] = ACTIONS(1742),
    [anon_sym_LBRACK] = ACTIONS(1740),
    [anon_sym_RBRACK] = ACTIONS(1742),
    [anon_sym_true] = ACTIONS(1740),
    [anon_sym_false] = ACTIONS(1740),
    [anon_sym_LBRACE] = ACTIONS(1740),
    [anon_sym_SEMI] = ACTIONS(1740),
    [anon_sym_exception] = ACTIONS(1740),
    [anon_sym_module] = ACTIONS(1740),
    [anon_sym_open] = ACTIONS(1740),
    [anon_sym_include] = ACTIONS(1740),
    [anon_sym_class] = ACTIONS(1740),
    [anon_sym_val] = ACTIONS(1740),
    [anon_sym_end] = ACTIONS(1740),
    [anon_sym_object] = ACTIONS(1740),
    [anon_sym_in] = ACTIONS(1740),
    [anon_sym_AMP] = ACTIONS(1740),
    [anon_sym_POUND] = ACTIONS(1740),
    [anon_sym_COLON_COLON] = ACTIONS(1742),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1742),
    [anon_sym_else] = ACTIONS(1740),
    [anon_sym_new] = ACTIONS(1740),
    [anon_sym_LBRACE_LT] = ACTIONS(1742),
    [anon_sym_begin] = ACTIONS(1740),
    [sym_ocamlyacc_value] = ACTIONS(1742),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1740),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1742),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1740),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1740),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1742),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1742),
    [aux_sym_number_token1] = ACTIONS(1742),
    [aux_sym_signed_number_token1] = ACTIONS(1740),
    [anon_sym_SQUOTE] = ACTIONS(1742),
    [anon_sym_DQUOTE] = ACTIONS(1742),
    [sym_prefix_operator] = ACTIONS(1742),
    [aux_sym_sign_operator_token1] = ACTIONS(1740),
    [sym_hash_operator] = ACTIONS(1742),
    [anon_sym_lsl] = ACTIONS(1740),
    [anon_sym_lsr] = ACTIONS(1740),
    [anon_sym_asr] = ACTIONS(1740),
    [aux_sym_pow_operator_token1] = ACTIONS(1742),
    [anon_sym_mod] = ACTIONS(1740),
    [anon_sym_land] = ACTIONS(1740),
    [anon_sym_lor] = ACTIONS(1740),
    [anon_sym_lxor] = ACTIONS(1740),
    [aux_sym_mult_operator_token1] = ACTIONS(1740),
    [aux_sym_add_operator_token1] = ACTIONS(1740),
    [sym_concat_operator] = ACTIONS(1742),
    [sym_rel_operator] = ACTIONS(1740),
    [anon_sym_AMP_AMP] = ACTIONS(1740),
    [anon_sym_or] = ACTIONS(1740),
    [anon_sym_PIPE_PIPE] = ACTIONS(1740),
    [sym_assign_operator] = ACTIONS(1742),
    [sym_let_operator] = ACTIONS(1742),
    [sym_let_and_operator] = ACTIONS(1742),
    [sym__uppercase_identifier] = ACTIONS(1742),
    [aux_sym_directive_token1] = ACTIONS(1740),
    [aux_sym_tag_token1] = ACTIONS(1742),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(979)] = {
    [sym_attribute] = STATE(979),
    [sym__lowercase_identifier] = ACTIONS(1618),
    [anon_sym_SEMI_SEMI] = ACTIONS(1620),
    [anon_sym_let] = ACTIONS(1618),
    [anon_sym_and] = ACTIONS(1618),
    [anon_sym_TILDE] = ACTIONS(1618),
    [anon_sym_QMARK] = ACTIONS(1618),
    [anon_sym_LPAREN] = ACTIONS(1620),
    [anon_sym_external] = ACTIONS(1618),
    [anon_sym_type] = ACTIONS(1618),
    [anon_sym_COMMA] = ACTIONS(1620),
    [anon_sym_PIPE] = ACTIONS(1618),
    [anon_sym_DOT] = ACTIONS(1620),
    [anon_sym_LBRACK] = ACTIONS(1618),
    [anon_sym_RBRACK] = ACTIONS(1620),
    [anon_sym_true] = ACTIONS(1618),
    [anon_sym_false] = ACTIONS(1618),
    [anon_sym_LBRACE] = ACTIONS(1618),
    [anon_sym_SEMI] = ACTIONS(1618),
    [anon_sym_exception] = ACTIONS(1618),
    [anon_sym_module] = ACTIONS(1618),
    [anon_sym_open] = ACTIONS(1618),
    [anon_sym_include] = ACTIONS(1618),
    [anon_sym_class] = ACTIONS(1618),
    [anon_sym_val] = ACTIONS(1618),
    [anon_sym_end] = ACTIONS(1618),
    [anon_sym_object] = ACTIONS(1618),
    [anon_sym_in] = ACTIONS(1618),
    [anon_sym_AMP] = ACTIONS(1618),
    [anon_sym_POUND] = ACTIONS(1618),
    [anon_sym_COLON_COLON] = ACTIONS(1620),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1620),
    [anon_sym_else] = ACTIONS(1618),
    [anon_sym_new] = ACTIONS(1618),
    [anon_sym_LBRACE_LT] = ACTIONS(1620),
    [anon_sym_begin] = ACTIONS(1618),
    [sym_ocamlyacc_value] = ACTIONS(1620),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1618),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1620),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1618),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1618),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1620),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1620),
    [aux_sym_number_token1] = ACTIONS(1620),
    [aux_sym_signed_number_token1] = ACTIONS(1618),
    [anon_sym_SQUOTE] = ACTIONS(1620),
    [anon_sym_DQUOTE] = ACTIONS(1620),
    [sym_prefix_operator] = ACTIONS(1620),
    [aux_sym_sign_operator_token1] = ACTIONS(1618),
    [sym_hash_operator] = ACTIONS(1620),
    [anon_sym_lsl] = ACTIONS(1618),
    [anon_sym_lsr] = ACTIONS(1618),
    [anon_sym_asr] = ACTIONS(1618),
    [aux_sym_pow_operator_token1] = ACTIONS(1620),
    [anon_sym_mod] = ACTIONS(1618),
    [anon_sym_land] = ACTIONS(1618),
    [anon_sym_lor] = ACTIONS(1618),
    [anon_sym_lxor] = ACTIONS(1618),
    [aux_sym_mult_operator_token1] = ACTIONS(1618),
    [aux_sym_add_operator_token1] = ACTIONS(1618),
    [sym_concat_operator] = ACTIONS(1620),
    [sym_rel_operator] = ACTIONS(1618),
    [anon_sym_AMP_AMP] = ACTIONS(1618),
    [anon_sym_or] = ACTIONS(1618),
    [anon_sym_PIPE_PIPE] = ACTIONS(1618),
    [sym_assign_operator] = ACTIONS(1620),
    [sym_let_operator] = ACTIONS(1620),
    [sym_let_and_operator] = ACTIONS(1620),
    [sym__uppercase_identifier] = ACTIONS(1620),
    [aux_sym_directive_token1] = ACTIONS(1618),
    [aux_sym_tag_token1] = ACTIONS(1620),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(980)] = {
    [sym_attribute] = STATE(980),
    [sym__lowercase_identifier] = ACTIONS(1850),
    [anon_sym_SEMI_SEMI] = ACTIONS(1852),
    [anon_sym_let] = ACTIONS(1850),
    [anon_sym_and] = ACTIONS(1850),
    [anon_sym_TILDE] = ACTIONS(1850),
    [anon_sym_QMARK] = ACTIONS(1850),
    [anon_sym_LPAREN] = ACTIONS(1852),
    [anon_sym_external] = ACTIONS(1850),
    [anon_sym_type] = ACTIONS(1850),
    [anon_sym_COMMA] = ACTIONS(1852),
    [anon_sym_PIPE] = ACTIONS(1850),
    [anon_sym_DOT] = ACTIONS(1852),
    [anon_sym_LBRACK] = ACTIONS(1850),
    [anon_sym_RBRACK] = ACTIONS(1852),
    [anon_sym_true] = ACTIONS(1850),
    [anon_sym_false] = ACTIONS(1850),
    [anon_sym_LBRACE] = ACTIONS(1850),
    [anon_sym_SEMI] = ACTIONS(1850),
    [anon_sym_exception] = ACTIONS(1850),
    [anon_sym_module] = ACTIONS(1850),
    [anon_sym_open] = ACTIONS(1850),
    [anon_sym_include] = ACTIONS(1850),
    [anon_sym_class] = ACTIONS(1850),
    [anon_sym_val] = ACTIONS(1850),
    [anon_sym_end] = ACTIONS(1850),
    [anon_sym_object] = ACTIONS(1850),
    [anon_sym_in] = ACTIONS(1850),
    [anon_sym_AMP] = ACTIONS(1850),
    [anon_sym_POUND] = ACTIONS(1850),
    [anon_sym_COLON_COLON] = ACTIONS(1852),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1852),
    [anon_sym_else] = ACTIONS(1850),
    [anon_sym_new] = ACTIONS(1850),
    [anon_sym_LBRACE_LT] = ACTIONS(1852),
    [anon_sym_begin] = ACTIONS(1850),
    [sym_ocamlyacc_value] = ACTIONS(1852),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1850),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1852),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1850),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1850),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1852),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1852),
    [aux_sym_number_token1] = ACTIONS(1852),
    [aux_sym_signed_number_token1] = ACTIONS(1850),
    [anon_sym_SQUOTE] = ACTIONS(1852),
    [anon_sym_DQUOTE] = ACTIONS(1852),
    [sym_prefix_operator] = ACTIONS(1852),
    [aux_sym_sign_operator_token1] = ACTIONS(1850),
    [sym_hash_operator] = ACTIONS(1852),
    [anon_sym_lsl] = ACTIONS(1850),
    [anon_sym_lsr] = ACTIONS(1850),
    [anon_sym_asr] = ACTIONS(1850),
    [aux_sym_pow_operator_token1] = ACTIONS(1852),
    [anon_sym_mod] = ACTIONS(1850),
    [anon_sym_land] = ACTIONS(1850),
    [anon_sym_lor] = ACTIONS(1850),
    [anon_sym_lxor] = ACTIONS(1850),
    [aux_sym_mult_operator_token1] = ACTIONS(1850),
    [aux_sym_add_operator_token1] = ACTIONS(1850),
    [sym_concat_operator] = ACTIONS(1852),
    [sym_rel_operator] = ACTIONS(1850),
    [anon_sym_AMP_AMP] = ACTIONS(1850),
    [anon_sym_or] = ACTIONS(1850),
    [anon_sym_PIPE_PIPE] = ACTIONS(1850),
    [sym_assign_operator] = ACTIONS(1852),
    [sym_let_operator] = ACTIONS(1852),
    [sym_let_and_operator] = ACTIONS(1852),
    [sym__uppercase_identifier] = ACTIONS(1852),
    [aux_sym_directive_token1] = ACTIONS(1850),
    [aux_sym_tag_token1] = ACTIONS(1852),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(981)] = {
    [sym_attribute] = STATE(981),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_SEMI_SEMI] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2079),
    [anon_sym_and] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_external] = ACTIONS(2079),
    [anon_sym_type] = ACTIONS(2079),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_DOT] = ACTIONS(765),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2079),
    [anon_sym_exception] = ACTIONS(2079),
    [anon_sym_module] = ACTIONS(2079),
    [anon_sym_open] = ACTIONS(2079),
    [anon_sym_include] = ACTIONS(2079),
    [anon_sym_class] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_in] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_POUND] = ACTIONS(767),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [sym_hash_operator] = ACTIONS(769),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym_let_operator] = ACTIONS(2081),
    [sym_let_and_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_directive_token1] = ACTIONS(2079),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(982)] = {
    [sym_attribute] = STATE(982),
    [sym__lowercase_identifier] = ACTIONS(1622),
    [anon_sym_SEMI_SEMI] = ACTIONS(1624),
    [anon_sym_let] = ACTIONS(1622),
    [anon_sym_and] = ACTIONS(1622),
    [anon_sym_TILDE] = ACTIONS(1622),
    [anon_sym_QMARK] = ACTIONS(1622),
    [anon_sym_LPAREN] = ACTIONS(1624),
    [anon_sym_external] = ACTIONS(1622),
    [anon_sym_type] = ACTIONS(1622),
    [anon_sym_COMMA] = ACTIONS(1624),
    [anon_sym_PIPE] = ACTIONS(1622),
    [anon_sym_DOT] = ACTIONS(1624),
    [anon_sym_LBRACK] = ACTIONS(1622),
    [anon_sym_RBRACK] = ACTIONS(1624),
    [anon_sym_true] = ACTIONS(1622),
    [anon_sym_false] = ACTIONS(1622),
    [anon_sym_LBRACE] = ACTIONS(1622),
    [anon_sym_SEMI] = ACTIONS(1622),
    [anon_sym_exception] = ACTIONS(1622),
    [anon_sym_module] = ACTIONS(1622),
    [anon_sym_open] = ACTIONS(1622),
    [anon_sym_include] = ACTIONS(1622),
    [anon_sym_class] = ACTIONS(1622),
    [anon_sym_val] = ACTIONS(1622),
    [anon_sym_end] = ACTIONS(1622),
    [anon_sym_object] = ACTIONS(1622),
    [anon_sym_in] = ACTIONS(1622),
    [anon_sym_AMP] = ACTIONS(1622),
    [anon_sym_POUND] = ACTIONS(1622),
    [anon_sym_COLON_COLON] = ACTIONS(1624),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1624),
    [anon_sym_else] = ACTIONS(1622),
    [anon_sym_new] = ACTIONS(1622),
    [anon_sym_LBRACE_LT] = ACTIONS(1624),
    [anon_sym_begin] = ACTIONS(1622),
    [sym_ocamlyacc_value] = ACTIONS(1624),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1622),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1624),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1622),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1622),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1624),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1624),
    [aux_sym_number_token1] = ACTIONS(1624),
    [aux_sym_signed_number_token1] = ACTIONS(1622),
    [anon_sym_SQUOTE] = ACTIONS(1624),
    [anon_sym_DQUOTE] = ACTIONS(1624),
    [sym_prefix_operator] = ACTIONS(1624),
    [aux_sym_sign_operator_token1] = ACTIONS(1622),
    [sym_hash_operator] = ACTIONS(1624),
    [anon_sym_lsl] = ACTIONS(1622),
    [anon_sym_lsr] = ACTIONS(1622),
    [anon_sym_asr] = ACTIONS(1622),
    [aux_sym_pow_operator_token1] = ACTIONS(1624),
    [anon_sym_mod] = ACTIONS(1622),
    [anon_sym_land] = ACTIONS(1622),
    [anon_sym_lor] = ACTIONS(1622),
    [anon_sym_lxor] = ACTIONS(1622),
    [aux_sym_mult_operator_token1] = ACTIONS(1622),
    [aux_sym_add_operator_token1] = ACTIONS(1622),
    [sym_concat_operator] = ACTIONS(1624),
    [sym_rel_operator] = ACTIONS(1622),
    [anon_sym_AMP_AMP] = ACTIONS(1622),
    [anon_sym_or] = ACTIONS(1622),
    [anon_sym_PIPE_PIPE] = ACTIONS(1622),
    [sym_assign_operator] = ACTIONS(1624),
    [sym_let_operator] = ACTIONS(1624),
    [sym_let_and_operator] = ACTIONS(1624),
    [sym__uppercase_identifier] = ACTIONS(1624),
    [aux_sym_directive_token1] = ACTIONS(1622),
    [aux_sym_tag_token1] = ACTIONS(1624),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(983)] = {
    [sym_attribute] = STATE(983),
    [sym__lowercase_identifier] = ACTIONS(1950),
    [anon_sym_SEMI_SEMI] = ACTIONS(1952),
    [anon_sym_let] = ACTIONS(1950),
    [anon_sym_and] = ACTIONS(1950),
    [anon_sym_TILDE] = ACTIONS(1950),
    [anon_sym_QMARK] = ACTIONS(1950),
    [anon_sym_LPAREN] = ACTIONS(1952),
    [anon_sym_external] = ACTIONS(1950),
    [anon_sym_type] = ACTIONS(1950),
    [anon_sym_COMMA] = ACTIONS(1952),
    [anon_sym_PIPE] = ACTIONS(1950),
    [anon_sym_DOT] = ACTIONS(1952),
    [anon_sym_LBRACK] = ACTIONS(1950),
    [anon_sym_RBRACK] = ACTIONS(1952),
    [anon_sym_true] = ACTIONS(1950),
    [anon_sym_false] = ACTIONS(1950),
    [anon_sym_LBRACE] = ACTIONS(1950),
    [anon_sym_SEMI] = ACTIONS(1950),
    [anon_sym_exception] = ACTIONS(1950),
    [anon_sym_module] = ACTIONS(1950),
    [anon_sym_open] = ACTIONS(1950),
    [anon_sym_include] = ACTIONS(1950),
    [anon_sym_class] = ACTIONS(1950),
    [anon_sym_val] = ACTIONS(1950),
    [anon_sym_end] = ACTIONS(1950),
    [anon_sym_object] = ACTIONS(1950),
    [anon_sym_in] = ACTIONS(1950),
    [anon_sym_AMP] = ACTIONS(1950),
    [anon_sym_POUND] = ACTIONS(1950),
    [anon_sym_COLON_COLON] = ACTIONS(1952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1952),
    [anon_sym_else] = ACTIONS(1950),
    [anon_sym_new] = ACTIONS(1950),
    [anon_sym_LBRACE_LT] = ACTIONS(1952),
    [anon_sym_begin] = ACTIONS(1950),
    [sym_ocamlyacc_value] = ACTIONS(1952),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1950),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1952),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1950),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1950),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1952),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1952),
    [aux_sym_number_token1] = ACTIONS(1952),
    [aux_sym_signed_number_token1] = ACTIONS(1950),
    [anon_sym_SQUOTE] = ACTIONS(1952),
    [anon_sym_DQUOTE] = ACTIONS(1952),
    [sym_prefix_operator] = ACTIONS(1952),
    [aux_sym_sign_operator_token1] = ACTIONS(1950),
    [sym_hash_operator] = ACTIONS(1952),
    [anon_sym_lsl] = ACTIONS(1950),
    [anon_sym_lsr] = ACTIONS(1950),
    [anon_sym_asr] = ACTIONS(1950),
    [aux_sym_pow_operator_token1] = ACTIONS(1952),
    [anon_sym_mod] = ACTIONS(1950),
    [anon_sym_land] = ACTIONS(1950),
    [anon_sym_lor] = ACTIONS(1950),
    [anon_sym_lxor] = ACTIONS(1950),
    [aux_sym_mult_operator_token1] = ACTIONS(1950),
    [aux_sym_add_operator_token1] = ACTIONS(1950),
    [sym_concat_operator] = ACTIONS(1952),
    [sym_rel_operator] = ACTIONS(1950),
    [anon_sym_AMP_AMP] = ACTIONS(1950),
    [anon_sym_or] = ACTIONS(1950),
    [anon_sym_PIPE_PIPE] = ACTIONS(1950),
    [sym_assign_operator] = ACTIONS(1952),
    [sym_let_operator] = ACTIONS(1952),
    [sym_let_and_operator] = ACTIONS(1952),
    [sym__uppercase_identifier] = ACTIONS(1952),
    [aux_sym_directive_token1] = ACTIONS(1950),
    [aux_sym_tag_token1] = ACTIONS(1952),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(984)] = {
    [sym_attribute] = STATE(984),
    [sym__lowercase_identifier] = ACTIONS(1902),
    [anon_sym_SEMI_SEMI] = ACTIONS(1904),
    [anon_sym_let] = ACTIONS(1902),
    [anon_sym_and] = ACTIONS(1902),
    [anon_sym_TILDE] = ACTIONS(1902),
    [anon_sym_QMARK] = ACTIONS(1902),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_external] = ACTIONS(1902),
    [anon_sym_type] = ACTIONS(1902),
    [anon_sym_COMMA] = ACTIONS(1904),
    [anon_sym_PIPE] = ACTIONS(1902),
    [anon_sym_DOT] = ACTIONS(1904),
    [anon_sym_LBRACK] = ACTIONS(1902),
    [anon_sym_RBRACK] = ACTIONS(1904),
    [anon_sym_true] = ACTIONS(1902),
    [anon_sym_false] = ACTIONS(1902),
    [anon_sym_LBRACE] = ACTIONS(1902),
    [anon_sym_SEMI] = ACTIONS(1902),
    [anon_sym_exception] = ACTIONS(1902),
    [anon_sym_module] = ACTIONS(1902),
    [anon_sym_open] = ACTIONS(1902),
    [anon_sym_include] = ACTIONS(1902),
    [anon_sym_class] = ACTIONS(1902),
    [anon_sym_val] = ACTIONS(1902),
    [anon_sym_end] = ACTIONS(1902),
    [anon_sym_object] = ACTIONS(1902),
    [anon_sym_in] = ACTIONS(1902),
    [anon_sym_AMP] = ACTIONS(1902),
    [anon_sym_POUND] = ACTIONS(1902),
    [anon_sym_COLON_COLON] = ACTIONS(1904),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1904),
    [anon_sym_else] = ACTIONS(1902),
    [anon_sym_new] = ACTIONS(1902),
    [anon_sym_LBRACE_LT] = ACTIONS(1904),
    [anon_sym_begin] = ACTIONS(1902),
    [sym_ocamlyacc_value] = ACTIONS(1904),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1902),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1904),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1902),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1902),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1904),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1904),
    [aux_sym_number_token1] = ACTIONS(1904),
    [aux_sym_signed_number_token1] = ACTIONS(1902),
    [anon_sym_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1904),
    [sym_prefix_operator] = ACTIONS(1904),
    [aux_sym_sign_operator_token1] = ACTIONS(1902),
    [sym_hash_operator] = ACTIONS(1904),
    [anon_sym_lsl] = ACTIONS(1902),
    [anon_sym_lsr] = ACTIONS(1902),
    [anon_sym_asr] = ACTIONS(1902),
    [aux_sym_pow_operator_token1] = ACTIONS(1904),
    [anon_sym_mod] = ACTIONS(1902),
    [anon_sym_land] = ACTIONS(1902),
    [anon_sym_lor] = ACTIONS(1902),
    [anon_sym_lxor] = ACTIONS(1902),
    [aux_sym_mult_operator_token1] = ACTIONS(1902),
    [aux_sym_add_operator_token1] = ACTIONS(1902),
    [sym_concat_operator] = ACTIONS(1904),
    [sym_rel_operator] = ACTIONS(1902),
    [anon_sym_AMP_AMP] = ACTIONS(1902),
    [anon_sym_or] = ACTIONS(1902),
    [anon_sym_PIPE_PIPE] = ACTIONS(1902),
    [sym_assign_operator] = ACTIONS(1904),
    [sym_let_operator] = ACTIONS(1904),
    [sym_let_and_operator] = ACTIONS(1904),
    [sym__uppercase_identifier] = ACTIONS(1904),
    [aux_sym_directive_token1] = ACTIONS(1902),
    [aux_sym_tag_token1] = ACTIONS(1904),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(985)] = {
    [sym_attribute] = STATE(985),
    [sym__lowercase_identifier] = ACTIONS(1906),
    [anon_sym_SEMI_SEMI] = ACTIONS(1908),
    [anon_sym_let] = ACTIONS(1906),
    [anon_sym_and] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1906),
    [anon_sym_QMARK] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1908),
    [anon_sym_external] = ACTIONS(1906),
    [anon_sym_type] = ACTIONS(1906),
    [anon_sym_COMMA] = ACTIONS(1908),
    [anon_sym_PIPE] = ACTIONS(1906),
    [anon_sym_DOT] = ACTIONS(1908),
    [anon_sym_LBRACK] = ACTIONS(1906),
    [anon_sym_RBRACK] = ACTIONS(1908),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_LBRACE] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1906),
    [anon_sym_exception] = ACTIONS(1906),
    [anon_sym_module] = ACTIONS(1906),
    [anon_sym_open] = ACTIONS(1906),
    [anon_sym_include] = ACTIONS(1906),
    [anon_sym_class] = ACTIONS(1906),
    [anon_sym_val] = ACTIONS(1906),
    [anon_sym_end] = ACTIONS(1906),
    [anon_sym_object] = ACTIONS(1906),
    [anon_sym_in] = ACTIONS(1906),
    [anon_sym_AMP] = ACTIONS(1906),
    [anon_sym_POUND] = ACTIONS(1906),
    [anon_sym_COLON_COLON] = ACTIONS(1908),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1908),
    [anon_sym_else] = ACTIONS(1906),
    [anon_sym_new] = ACTIONS(1906),
    [anon_sym_LBRACE_LT] = ACTIONS(1908),
    [anon_sym_begin] = ACTIONS(1906),
    [sym_ocamlyacc_value] = ACTIONS(1908),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1906),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1908),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1906),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1906),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1908),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1908),
    [aux_sym_number_token1] = ACTIONS(1908),
    [aux_sym_signed_number_token1] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1908),
    [anon_sym_DQUOTE] = ACTIONS(1908),
    [sym_prefix_operator] = ACTIONS(1908),
    [aux_sym_sign_operator_token1] = ACTIONS(1906),
    [sym_hash_operator] = ACTIONS(1908),
    [anon_sym_lsl] = ACTIONS(1906),
    [anon_sym_lsr] = ACTIONS(1906),
    [anon_sym_asr] = ACTIONS(1906),
    [aux_sym_pow_operator_token1] = ACTIONS(1908),
    [anon_sym_mod] = ACTIONS(1906),
    [anon_sym_land] = ACTIONS(1906),
    [anon_sym_lor] = ACTIONS(1906),
    [anon_sym_lxor] = ACTIONS(1906),
    [aux_sym_mult_operator_token1] = ACTIONS(1906),
    [aux_sym_add_operator_token1] = ACTIONS(1906),
    [sym_concat_operator] = ACTIONS(1908),
    [sym_rel_operator] = ACTIONS(1906),
    [anon_sym_AMP_AMP] = ACTIONS(1906),
    [anon_sym_or] = ACTIONS(1906),
    [anon_sym_PIPE_PIPE] = ACTIONS(1906),
    [sym_assign_operator] = ACTIONS(1908),
    [sym_let_operator] = ACTIONS(1908),
    [sym_let_and_operator] = ACTIONS(1908),
    [sym__uppercase_identifier] = ACTIONS(1908),
    [aux_sym_directive_token1] = ACTIONS(1906),
    [aux_sym_tag_token1] = ACTIONS(1908),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(986)] = {
    [sym_attribute] = STATE(986),
    [sym__lowercase_identifier] = ACTIONS(1922),
    [anon_sym_SEMI_SEMI] = ACTIONS(1924),
    [anon_sym_let] = ACTIONS(1922),
    [anon_sym_and] = ACTIONS(1922),
    [anon_sym_TILDE] = ACTIONS(1922),
    [anon_sym_QMARK] = ACTIONS(1922),
    [anon_sym_LPAREN] = ACTIONS(1924),
    [anon_sym_external] = ACTIONS(1922),
    [anon_sym_type] = ACTIONS(1922),
    [anon_sym_COMMA] = ACTIONS(1924),
    [anon_sym_PIPE] = ACTIONS(1922),
    [anon_sym_DOT] = ACTIONS(1924),
    [anon_sym_LBRACK] = ACTIONS(1922),
    [anon_sym_RBRACK] = ACTIONS(1924),
    [anon_sym_true] = ACTIONS(1922),
    [anon_sym_false] = ACTIONS(1922),
    [anon_sym_LBRACE] = ACTIONS(1922),
    [anon_sym_SEMI] = ACTIONS(1922),
    [anon_sym_exception] = ACTIONS(1922),
    [anon_sym_module] = ACTIONS(1922),
    [anon_sym_open] = ACTIONS(1922),
    [anon_sym_include] = ACTIONS(1922),
    [anon_sym_class] = ACTIONS(1922),
    [anon_sym_val] = ACTIONS(1922),
    [anon_sym_end] = ACTIONS(1922),
    [anon_sym_object] = ACTIONS(1922),
    [anon_sym_in] = ACTIONS(1922),
    [anon_sym_AMP] = ACTIONS(1922),
    [anon_sym_POUND] = ACTIONS(1922),
    [anon_sym_COLON_COLON] = ACTIONS(1924),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1924),
    [anon_sym_else] = ACTIONS(1922),
    [anon_sym_new] = ACTIONS(1922),
    [anon_sym_LBRACE_LT] = ACTIONS(1924),
    [anon_sym_begin] = ACTIONS(1922),
    [sym_ocamlyacc_value] = ACTIONS(1924),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1922),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1924),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1922),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1922),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1924),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1924),
    [aux_sym_number_token1] = ACTIONS(1924),
    [aux_sym_signed_number_token1] = ACTIONS(1922),
    [anon_sym_SQUOTE] = ACTIONS(1924),
    [anon_sym_DQUOTE] = ACTIONS(1924),
    [sym_prefix_operator] = ACTIONS(1924),
    [aux_sym_sign_operator_token1] = ACTIONS(1922),
    [sym_hash_operator] = ACTIONS(1924),
    [anon_sym_lsl] = ACTIONS(1922),
    [anon_sym_lsr] = ACTIONS(1922),
    [anon_sym_asr] = ACTIONS(1922),
    [aux_sym_pow_operator_token1] = ACTIONS(1924),
    [anon_sym_mod] = ACTIONS(1922),
    [anon_sym_land] = ACTIONS(1922),
    [anon_sym_lor] = ACTIONS(1922),
    [anon_sym_lxor] = ACTIONS(1922),
    [aux_sym_mult_operator_token1] = ACTIONS(1922),
    [aux_sym_add_operator_token1] = ACTIONS(1922),
    [sym_concat_operator] = ACTIONS(1924),
    [sym_rel_operator] = ACTIONS(1922),
    [anon_sym_AMP_AMP] = ACTIONS(1922),
    [anon_sym_or] = ACTIONS(1922),
    [anon_sym_PIPE_PIPE] = ACTIONS(1922),
    [sym_assign_operator] = ACTIONS(1924),
    [sym_let_operator] = ACTIONS(1924),
    [sym_let_and_operator] = ACTIONS(1924),
    [sym__uppercase_identifier] = ACTIONS(1924),
    [aux_sym_directive_token1] = ACTIONS(1922),
    [aux_sym_tag_token1] = ACTIONS(1924),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(987)] = {
    [sym_attribute] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(1614),
    [anon_sym_SEMI_SEMI] = ACTIONS(1616),
    [anon_sym_let] = ACTIONS(1614),
    [anon_sym_and] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_external] = ACTIONS(1614),
    [anon_sym_type] = ACTIONS(1614),
    [anon_sym_COMMA] = ACTIONS(1616),
    [anon_sym_PIPE] = ACTIONS(1614),
    [anon_sym_DOT] = ACTIONS(1616),
    [anon_sym_LBRACK] = ACTIONS(1614),
    [anon_sym_RBRACK] = ACTIONS(1616),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1614),
    [anon_sym_exception] = ACTIONS(1614),
    [anon_sym_module] = ACTIONS(1614),
    [anon_sym_open] = ACTIONS(1614),
    [anon_sym_include] = ACTIONS(1614),
    [anon_sym_class] = ACTIONS(1614),
    [anon_sym_val] = ACTIONS(1614),
    [anon_sym_end] = ACTIONS(1614),
    [anon_sym_object] = ACTIONS(1614),
    [anon_sym_in] = ACTIONS(1614),
    [anon_sym_AMP] = ACTIONS(1614),
    [anon_sym_POUND] = ACTIONS(1614),
    [anon_sym_COLON_COLON] = ACTIONS(1616),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1616),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
    [anon_sym_LBRACE_LT] = ACTIONS(1616),
    [anon_sym_begin] = ACTIONS(1614),
    [sym_ocamlyacc_value] = ACTIONS(1616),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1614),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1616),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1616),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1616),
    [aux_sym_number_token1] = ACTIONS(1616),
    [aux_sym_signed_number_token1] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1616),
    [sym_prefix_operator] = ACTIONS(1616),
    [aux_sym_sign_operator_token1] = ACTIONS(1614),
    [sym_hash_operator] = ACTIONS(1616),
    [anon_sym_lsl] = ACTIONS(1614),
    [anon_sym_lsr] = ACTIONS(1614),
    [anon_sym_asr] = ACTIONS(1614),
    [aux_sym_pow_operator_token1] = ACTIONS(1616),
    [anon_sym_mod] = ACTIONS(1614),
    [anon_sym_land] = ACTIONS(1614),
    [anon_sym_lor] = ACTIONS(1614),
    [anon_sym_lxor] = ACTIONS(1614),
    [aux_sym_mult_operator_token1] = ACTIONS(1614),
    [aux_sym_add_operator_token1] = ACTIONS(1614),
    [sym_concat_operator] = ACTIONS(1616),
    [sym_rel_operator] = ACTIONS(1614),
    [anon_sym_AMP_AMP] = ACTIONS(1614),
    [anon_sym_or] = ACTIONS(1614),
    [anon_sym_PIPE_PIPE] = ACTIONS(1614),
    [sym_assign_operator] = ACTIONS(1616),
    [sym_let_operator] = ACTIONS(1616),
    [sym_let_and_operator] = ACTIONS(1616),
    [sym__uppercase_identifier] = ACTIONS(1616),
    [aux_sym_directive_token1] = ACTIONS(1614),
    [aux_sym_tag_token1] = ACTIONS(1616),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(988)] = {
    [sym_attribute] = STATE(988),
    [sym__lowercase_identifier] = ACTIONS(2107),
    [anon_sym_SEMI_SEMI] = ACTIONS(2109),
    [anon_sym_let] = ACTIONS(2107),
    [anon_sym_and] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2107),
    [anon_sym_QMARK] = ACTIONS(2107),
    [anon_sym_LPAREN] = ACTIONS(2109),
    [anon_sym_external] = ACTIONS(2107),
    [anon_sym_type] = ACTIONS(2107),
    [anon_sym_COMMA] = ACTIONS(2109),
    [anon_sym_PIPE] = ACTIONS(2107),
    [anon_sym_DOT] = ACTIONS(765),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_RBRACK] = ACTIONS(2109),
    [anon_sym_true] = ACTIONS(2107),
    [anon_sym_false] = ACTIONS(2107),
    [anon_sym_LBRACE] = ACTIONS(2107),
    [anon_sym_SEMI] = ACTIONS(2107),
    [anon_sym_exception] = ACTIONS(2107),
    [anon_sym_module] = ACTIONS(2107),
    [anon_sym_open] = ACTIONS(2107),
    [anon_sym_include] = ACTIONS(2107),
    [anon_sym_class] = ACTIONS(2107),
    [anon_sym_val] = ACTIONS(2107),
    [anon_sym_end] = ACTIONS(2107),
    [anon_sym_object] = ACTIONS(2107),
    [anon_sym_in] = ACTIONS(2107),
    [anon_sym_AMP] = ACTIONS(2107),
    [anon_sym_POUND] = ACTIONS(767),
    [anon_sym_COLON_COLON] = ACTIONS(2109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2109),
    [anon_sym_else] = ACTIONS(2107),
    [anon_sym_new] = ACTIONS(2107),
    [anon_sym_LBRACE_LT] = ACTIONS(2109),
    [anon_sym_begin] = ACTIONS(2107),
    [sym_ocamlyacc_value] = ACTIONS(2109),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2107),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2109),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2109),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2109),
    [aux_sym_number_token1] = ACTIONS(2109),
    [aux_sym_signed_number_token1] = ACTIONS(2107),
    [anon_sym_SQUOTE] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(2109),
    [sym_prefix_operator] = ACTIONS(2109),
    [aux_sym_sign_operator_token1] = ACTIONS(2107),
    [sym_hash_operator] = ACTIONS(769),
    [anon_sym_lsl] = ACTIONS(2107),
    [anon_sym_lsr] = ACTIONS(2107),
    [anon_sym_asr] = ACTIONS(2107),
    [aux_sym_pow_operator_token1] = ACTIONS(2109),
    [anon_sym_mod] = ACTIONS(2107),
    [anon_sym_land] = ACTIONS(2107),
    [anon_sym_lor] = ACTIONS(2107),
    [anon_sym_lxor] = ACTIONS(2107),
    [aux_sym_mult_operator_token1] = ACTIONS(2107),
    [aux_sym_add_operator_token1] = ACTIONS(2107),
    [sym_concat_operator] = ACTIONS(2109),
    [sym_rel_operator] = ACTIONS(2107),
    [anon_sym_AMP_AMP] = ACTIONS(2107),
    [anon_sym_or] = ACTIONS(2107),
    [anon_sym_PIPE_PIPE] = ACTIONS(2107),
    [sym_assign_operator] = ACTIONS(2109),
    [sym_let_operator] = ACTIONS(2109),
    [sym_let_and_operator] = ACTIONS(2109),
    [sym__uppercase_identifier] = ACTIONS(2109),
    [aux_sym_directive_token1] = ACTIONS(2107),
    [aux_sym_tag_token1] = ACTIONS(2109),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(989)] = {
    [sym_attribute] = STATE(989),
    [sym__lowercase_identifier] = ACTIONS(1750),
    [anon_sym_SEMI_SEMI] = ACTIONS(1752),
    [anon_sym_let] = ACTIONS(1750),
    [anon_sym_and] = ACTIONS(1750),
    [anon_sym_TILDE] = ACTIONS(1750),
    [anon_sym_QMARK] = ACTIONS(1750),
    [anon_sym_LPAREN] = ACTIONS(1752),
    [anon_sym_external] = ACTIONS(1750),
    [anon_sym_type] = ACTIONS(1750),
    [anon_sym_COMMA] = ACTIONS(1752),
    [anon_sym_PIPE] = ACTIONS(1750),
    [anon_sym_DOT] = ACTIONS(1752),
    [anon_sym_LBRACK] = ACTIONS(1750),
    [anon_sym_RBRACK] = ACTIONS(1752),
    [anon_sym_true] = ACTIONS(1750),
    [anon_sym_false] = ACTIONS(1750),
    [anon_sym_LBRACE] = ACTIONS(1750),
    [anon_sym_SEMI] = ACTIONS(1750),
    [anon_sym_exception] = ACTIONS(1750),
    [anon_sym_module] = ACTIONS(1750),
    [anon_sym_open] = ACTIONS(1750),
    [anon_sym_include] = ACTIONS(1750),
    [anon_sym_class] = ACTIONS(1750),
    [anon_sym_val] = ACTIONS(1750),
    [anon_sym_end] = ACTIONS(1750),
    [anon_sym_object] = ACTIONS(1750),
    [anon_sym_in] = ACTIONS(1750),
    [anon_sym_AMP] = ACTIONS(1750),
    [anon_sym_POUND] = ACTIONS(1750),
    [anon_sym_COLON_COLON] = ACTIONS(1752),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1752),
    [anon_sym_else] = ACTIONS(1750),
    [anon_sym_new] = ACTIONS(1750),
    [anon_sym_LBRACE_LT] = ACTIONS(1752),
    [anon_sym_begin] = ACTIONS(1750),
    [sym_ocamlyacc_value] = ACTIONS(1752),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1750),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1752),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1750),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1750),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1752),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1752),
    [aux_sym_number_token1] = ACTIONS(1752),
    [aux_sym_signed_number_token1] = ACTIONS(1750),
    [anon_sym_SQUOTE] = ACTIONS(1752),
    [anon_sym_DQUOTE] = ACTIONS(1752),
    [sym_prefix_operator] = ACTIONS(1752),
    [aux_sym_sign_operator_token1] = ACTIONS(1750),
    [sym_hash_operator] = ACTIONS(1752),
    [anon_sym_lsl] = ACTIONS(1750),
    [anon_sym_lsr] = ACTIONS(1750),
    [anon_sym_asr] = ACTIONS(1750),
    [aux_sym_pow_operator_token1] = ACTIONS(1752),
    [anon_sym_mod] = ACTIONS(1750),
    [anon_sym_land] = ACTIONS(1750),
    [anon_sym_lor] = ACTIONS(1750),
    [anon_sym_lxor] = ACTIONS(1750),
    [aux_sym_mult_operator_token1] = ACTIONS(1750),
    [aux_sym_add_operator_token1] = ACTIONS(1750),
    [sym_concat_operator] = ACTIONS(1752),
    [sym_rel_operator] = ACTIONS(1750),
    [anon_sym_AMP_AMP] = ACTIONS(1750),
    [anon_sym_or] = ACTIONS(1750),
    [anon_sym_PIPE_PIPE] = ACTIONS(1750),
    [sym_assign_operator] = ACTIONS(1752),
    [sym_let_operator] = ACTIONS(1752),
    [sym_let_and_operator] = ACTIONS(1752),
    [sym__uppercase_identifier] = ACTIONS(1752),
    [aux_sym_directive_token1] = ACTIONS(1750),
    [aux_sym_tag_token1] = ACTIONS(1752),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(990)] = {
    [sym_attribute] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(1724),
    [anon_sym_SEMI_SEMI] = ACTIONS(1726),
    [anon_sym_let] = ACTIONS(1724),
    [anon_sym_and] = ACTIONS(1724),
    [anon_sym_TILDE] = ACTIONS(1724),
    [anon_sym_QMARK] = ACTIONS(1724),
    [anon_sym_LPAREN] = ACTIONS(1726),
    [anon_sym_external] = ACTIONS(1724),
    [anon_sym_type] = ACTIONS(1724),
    [anon_sym_COMMA] = ACTIONS(1726),
    [anon_sym_PIPE] = ACTIONS(1724),
    [anon_sym_DOT] = ACTIONS(1726),
    [anon_sym_LBRACK] = ACTIONS(1724),
    [anon_sym_RBRACK] = ACTIONS(1726),
    [anon_sym_true] = ACTIONS(1724),
    [anon_sym_false] = ACTIONS(1724),
    [anon_sym_LBRACE] = ACTIONS(1724),
    [anon_sym_SEMI] = ACTIONS(1724),
    [anon_sym_exception] = ACTIONS(1724),
    [anon_sym_module] = ACTIONS(1724),
    [anon_sym_open] = ACTIONS(1724),
    [anon_sym_include] = ACTIONS(1724),
    [anon_sym_class] = ACTIONS(1724),
    [anon_sym_val] = ACTIONS(1724),
    [anon_sym_end] = ACTIONS(1724),
    [anon_sym_object] = ACTIONS(1724),
    [anon_sym_in] = ACTIONS(1724),
    [anon_sym_AMP] = ACTIONS(1724),
    [anon_sym_POUND] = ACTIONS(1724),
    [anon_sym_COLON_COLON] = ACTIONS(1726),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1726),
    [anon_sym_else] = ACTIONS(1724),
    [anon_sym_new] = ACTIONS(1724),
    [anon_sym_LBRACE_LT] = ACTIONS(1726),
    [anon_sym_begin] = ACTIONS(1724),
    [sym_ocamlyacc_value] = ACTIONS(1726),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1724),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1726),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1726),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1726),
    [aux_sym_number_token1] = ACTIONS(1726),
    [aux_sym_signed_number_token1] = ACTIONS(1724),
    [anon_sym_SQUOTE] = ACTIONS(1726),
    [anon_sym_DQUOTE] = ACTIONS(1726),
    [sym_prefix_operator] = ACTIONS(1726),
    [aux_sym_sign_operator_token1] = ACTIONS(1724),
    [sym_hash_operator] = ACTIONS(1726),
    [anon_sym_lsl] = ACTIONS(1724),
    [anon_sym_lsr] = ACTIONS(1724),
    [anon_sym_asr] = ACTIONS(1724),
    [aux_sym_pow_operator_token1] = ACTIONS(1726),
    [anon_sym_mod] = ACTIONS(1724),
    [anon_sym_land] = ACTIONS(1724),
    [anon_sym_lor] = ACTIONS(1724),
    [anon_sym_lxor] = ACTIONS(1724),
    [aux_sym_mult_operator_token1] = ACTIONS(1724),
    [aux_sym_add_operator_token1] = ACTIONS(1724),
    [sym_concat_operator] = ACTIONS(1726),
    [sym_rel_operator] = ACTIONS(1724),
    [anon_sym_AMP_AMP] = ACTIONS(1724),
    [anon_sym_or] = ACTIONS(1724),
    [anon_sym_PIPE_PIPE] = ACTIONS(1724),
    [sym_assign_operator] = ACTIONS(1726),
    [sym_let_operator] = ACTIONS(1726),
    [sym_let_and_operator] = ACTIONS(1726),
    [sym__uppercase_identifier] = ACTIONS(1726),
    [aux_sym_directive_token1] = ACTIONS(1724),
    [aux_sym_tag_token1] = ACTIONS(1726),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(991)] = {
    [sym_attribute] = STATE(991),
    [sym__lowercase_identifier] = ACTIONS(1704),
    [anon_sym_SEMI_SEMI] = ACTIONS(1706),
    [anon_sym_let] = ACTIONS(1704),
    [anon_sym_and] = ACTIONS(1704),
    [anon_sym_TILDE] = ACTIONS(1704),
    [anon_sym_QMARK] = ACTIONS(1704),
    [anon_sym_LPAREN] = ACTIONS(1706),
    [anon_sym_external] = ACTIONS(1704),
    [anon_sym_type] = ACTIONS(1704),
    [anon_sym_COMMA] = ACTIONS(1706),
    [anon_sym_PIPE] = ACTIONS(1704),
    [anon_sym_DOT] = ACTIONS(1706),
    [anon_sym_LBRACK] = ACTIONS(1704),
    [anon_sym_RBRACK] = ACTIONS(1706),
    [anon_sym_true] = ACTIONS(1704),
    [anon_sym_false] = ACTIONS(1704),
    [anon_sym_LBRACE] = ACTIONS(1704),
    [anon_sym_SEMI] = ACTIONS(1704),
    [anon_sym_exception] = ACTIONS(1704),
    [anon_sym_module] = ACTIONS(1704),
    [anon_sym_open] = ACTIONS(1704),
    [anon_sym_include] = ACTIONS(1704),
    [anon_sym_class] = ACTIONS(1704),
    [anon_sym_val] = ACTIONS(1704),
    [anon_sym_end] = ACTIONS(1704),
    [anon_sym_object] = ACTIONS(1704),
    [anon_sym_in] = ACTIONS(1704),
    [anon_sym_AMP] = ACTIONS(1704),
    [anon_sym_POUND] = ACTIONS(1704),
    [anon_sym_COLON_COLON] = ACTIONS(1706),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1706),
    [anon_sym_else] = ACTIONS(1704),
    [anon_sym_new] = ACTIONS(1704),
    [anon_sym_LBRACE_LT] = ACTIONS(1706),
    [anon_sym_begin] = ACTIONS(1704),
    [sym_ocamlyacc_value] = ACTIONS(1706),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1704),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1706),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1704),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1704),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1706),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1706),
    [aux_sym_number_token1] = ACTIONS(1706),
    [aux_sym_signed_number_token1] = ACTIONS(1704),
    [anon_sym_SQUOTE] = ACTIONS(1706),
    [anon_sym_DQUOTE] = ACTIONS(1706),
    [sym_prefix_operator] = ACTIONS(1706),
    [aux_sym_sign_operator_token1] = ACTIONS(1704),
    [sym_hash_operator] = ACTIONS(1706),
    [anon_sym_lsl] = ACTIONS(1704),
    [anon_sym_lsr] = ACTIONS(1704),
    [anon_sym_asr] = ACTIONS(1704),
    [aux_sym_pow_operator_token1] = ACTIONS(1706),
    [anon_sym_mod] = ACTIONS(1704),
    [anon_sym_land] = ACTIONS(1704),
    [anon_sym_lor] = ACTIONS(1704),
    [anon_sym_lxor] = ACTIONS(1704),
    [aux_sym_mult_operator_token1] = ACTIONS(1704),
    [aux_sym_add_operator_token1] = ACTIONS(1704),
    [sym_concat_operator] = ACTIONS(1706),
    [sym_rel_operator] = ACTIONS(1704),
    [anon_sym_AMP_AMP] = ACTIONS(1704),
    [anon_sym_or] = ACTIONS(1704),
    [anon_sym_PIPE_PIPE] = ACTIONS(1704),
    [sym_assign_operator] = ACTIONS(1706),
    [sym_let_operator] = ACTIONS(1706),
    [sym_let_and_operator] = ACTIONS(1706),
    [sym__uppercase_identifier] = ACTIONS(1706),
    [aux_sym_directive_token1] = ACTIONS(1704),
    [aux_sym_tag_token1] = ACTIONS(1706),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(992)] = {
    [sym_attribute] = STATE(992),
    [sym__lowercase_identifier] = ACTIONS(1646),
    [anon_sym_SEMI_SEMI] = ACTIONS(1648),
    [anon_sym_let] = ACTIONS(1646),
    [anon_sym_and] = ACTIONS(1646),
    [anon_sym_TILDE] = ACTIONS(1646),
    [anon_sym_QMARK] = ACTIONS(1646),
    [anon_sym_LPAREN] = ACTIONS(1648),
    [anon_sym_external] = ACTIONS(1646),
    [anon_sym_type] = ACTIONS(1646),
    [anon_sym_COMMA] = ACTIONS(1648),
    [anon_sym_PIPE] = ACTIONS(1646),
    [anon_sym_DOT] = ACTIONS(1648),
    [anon_sym_LBRACK] = ACTIONS(1646),
    [anon_sym_RBRACK] = ACTIONS(1648),
    [anon_sym_true] = ACTIONS(1646),
    [anon_sym_false] = ACTIONS(1646),
    [anon_sym_LBRACE] = ACTIONS(1646),
    [anon_sym_SEMI] = ACTIONS(1646),
    [anon_sym_exception] = ACTIONS(1646),
    [anon_sym_module] = ACTIONS(1646),
    [anon_sym_open] = ACTIONS(1646),
    [anon_sym_include] = ACTIONS(1646),
    [anon_sym_class] = ACTIONS(1646),
    [anon_sym_val] = ACTIONS(1646),
    [anon_sym_end] = ACTIONS(1646),
    [anon_sym_object] = ACTIONS(1646),
    [anon_sym_in] = ACTIONS(1646),
    [anon_sym_AMP] = ACTIONS(1646),
    [anon_sym_POUND] = ACTIONS(1646),
    [anon_sym_COLON_COLON] = ACTIONS(1648),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1648),
    [anon_sym_else] = ACTIONS(1646),
    [anon_sym_new] = ACTIONS(1646),
    [anon_sym_LBRACE_LT] = ACTIONS(1648),
    [anon_sym_begin] = ACTIONS(1646),
    [sym_ocamlyacc_value] = ACTIONS(1648),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1646),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1648),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1646),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1646),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1648),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1648),
    [aux_sym_number_token1] = ACTIONS(1648),
    [aux_sym_signed_number_token1] = ACTIONS(1646),
    [anon_sym_SQUOTE] = ACTIONS(1648),
    [anon_sym_DQUOTE] = ACTIONS(1648),
    [sym_prefix_operator] = ACTIONS(1648),
    [aux_sym_sign_operator_token1] = ACTIONS(1646),
    [sym_hash_operator] = ACTIONS(1648),
    [anon_sym_lsl] = ACTIONS(1646),
    [anon_sym_lsr] = ACTIONS(1646),
    [anon_sym_asr] = ACTIONS(1646),
    [aux_sym_pow_operator_token1] = ACTIONS(1648),
    [anon_sym_mod] = ACTIONS(1646),
    [anon_sym_land] = ACTIONS(1646),
    [anon_sym_lor] = ACTIONS(1646),
    [anon_sym_lxor] = ACTIONS(1646),
    [aux_sym_mult_operator_token1] = ACTIONS(1646),
    [aux_sym_add_operator_token1] = ACTIONS(1646),
    [sym_concat_operator] = ACTIONS(1648),
    [sym_rel_operator] = ACTIONS(1646),
    [anon_sym_AMP_AMP] = ACTIONS(1646),
    [anon_sym_or] = ACTIONS(1646),
    [anon_sym_PIPE_PIPE] = ACTIONS(1646),
    [sym_assign_operator] = ACTIONS(1648),
    [sym_let_operator] = ACTIONS(1648),
    [sym_let_and_operator] = ACTIONS(1648),
    [sym__uppercase_identifier] = ACTIONS(1648),
    [aux_sym_directive_token1] = ACTIONS(1646),
    [aux_sym_tag_token1] = ACTIONS(1648),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(993)] = {
    [sym_attribute] = STATE(993),
    [sym__lowercase_identifier] = ACTIONS(1728),
    [anon_sym_SEMI_SEMI] = ACTIONS(1730),
    [anon_sym_let] = ACTIONS(1728),
    [anon_sym_and] = ACTIONS(1728),
    [anon_sym_TILDE] = ACTIONS(1728),
    [anon_sym_QMARK] = ACTIONS(1728),
    [anon_sym_LPAREN] = ACTIONS(1730),
    [anon_sym_external] = ACTIONS(1728),
    [anon_sym_type] = ACTIONS(1728),
    [anon_sym_COMMA] = ACTIONS(1730),
    [anon_sym_PIPE] = ACTIONS(1728),
    [anon_sym_DOT] = ACTIONS(1730),
    [anon_sym_LBRACK] = ACTIONS(1728),
    [anon_sym_RBRACK] = ACTIONS(1730),
    [anon_sym_true] = ACTIONS(1728),
    [anon_sym_false] = ACTIONS(1728),
    [anon_sym_LBRACE] = ACTIONS(1728),
    [anon_sym_SEMI] = ACTIONS(1728),
    [anon_sym_exception] = ACTIONS(1728),
    [anon_sym_module] = ACTIONS(1728),
    [anon_sym_open] = ACTIONS(1728),
    [anon_sym_include] = ACTIONS(1728),
    [anon_sym_class] = ACTIONS(1728),
    [anon_sym_val] = ACTIONS(1728),
    [anon_sym_end] = ACTIONS(1728),
    [anon_sym_object] = ACTIONS(1728),
    [anon_sym_in] = ACTIONS(1728),
    [anon_sym_AMP] = ACTIONS(1728),
    [anon_sym_POUND] = ACTIONS(1728),
    [anon_sym_COLON_COLON] = ACTIONS(1730),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1730),
    [anon_sym_else] = ACTIONS(1728),
    [anon_sym_new] = ACTIONS(1728),
    [anon_sym_LBRACE_LT] = ACTIONS(1730),
    [anon_sym_begin] = ACTIONS(1728),
    [sym_ocamlyacc_value] = ACTIONS(1730),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1728),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1730),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1728),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1728),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1730),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1730),
    [aux_sym_number_token1] = ACTIONS(1730),
    [aux_sym_signed_number_token1] = ACTIONS(1728),
    [anon_sym_SQUOTE] = ACTIONS(1730),
    [anon_sym_DQUOTE] = ACTIONS(1730),
    [sym_prefix_operator] = ACTIONS(1730),
    [aux_sym_sign_operator_token1] = ACTIONS(1728),
    [sym_hash_operator] = ACTIONS(1730),
    [anon_sym_lsl] = ACTIONS(1728),
    [anon_sym_lsr] = ACTIONS(1728),
    [anon_sym_asr] = ACTIONS(1728),
    [aux_sym_pow_operator_token1] = ACTIONS(1730),
    [anon_sym_mod] = ACTIONS(1728),
    [anon_sym_land] = ACTIONS(1728),
    [anon_sym_lor] = ACTIONS(1728),
    [anon_sym_lxor] = ACTIONS(1728),
    [aux_sym_mult_operator_token1] = ACTIONS(1728),
    [aux_sym_add_operator_token1] = ACTIONS(1728),
    [sym_concat_operator] = ACTIONS(1730),
    [sym_rel_operator] = ACTIONS(1728),
    [anon_sym_AMP_AMP] = ACTIONS(1728),
    [anon_sym_or] = ACTIONS(1728),
    [anon_sym_PIPE_PIPE] = ACTIONS(1728),
    [sym_assign_operator] = ACTIONS(1730),
    [sym_let_operator] = ACTIONS(1730),
    [sym_let_and_operator] = ACTIONS(1730),
    [sym__uppercase_identifier] = ACTIONS(1730),
    [aux_sym_directive_token1] = ACTIONS(1728),
    [aux_sym_tag_token1] = ACTIONS(1730),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(994)] = {
    [sym_attribute] = STATE(994),
    [sym__lowercase_identifier] = ACTIONS(1732),
    [anon_sym_SEMI_SEMI] = ACTIONS(1734),
    [anon_sym_let] = ACTIONS(1732),
    [anon_sym_and] = ACTIONS(1732),
    [anon_sym_TILDE] = ACTIONS(1732),
    [anon_sym_QMARK] = ACTIONS(1732),
    [anon_sym_LPAREN] = ACTIONS(1734),
    [anon_sym_external] = ACTIONS(1732),
    [anon_sym_type] = ACTIONS(1732),
    [anon_sym_COMMA] = ACTIONS(1734),
    [anon_sym_PIPE] = ACTIONS(1732),
    [anon_sym_DOT] = ACTIONS(1734),
    [anon_sym_LBRACK] = ACTIONS(1732),
    [anon_sym_RBRACK] = ACTIONS(1734),
    [anon_sym_true] = ACTIONS(1732),
    [anon_sym_false] = ACTIONS(1732),
    [anon_sym_LBRACE] = ACTIONS(1732),
    [anon_sym_SEMI] = ACTIONS(1732),
    [anon_sym_exception] = ACTIONS(1732),
    [anon_sym_module] = ACTIONS(1732),
    [anon_sym_open] = ACTIONS(1732),
    [anon_sym_include] = ACTIONS(1732),
    [anon_sym_class] = ACTIONS(1732),
    [anon_sym_val] = ACTIONS(1732),
    [anon_sym_end] = ACTIONS(1732),
    [anon_sym_object] = ACTIONS(1732),
    [anon_sym_in] = ACTIONS(1732),
    [anon_sym_AMP] = ACTIONS(1732),
    [anon_sym_POUND] = ACTIONS(1732),
    [anon_sym_COLON_COLON] = ACTIONS(1734),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1734),
    [anon_sym_else] = ACTIONS(1732),
    [anon_sym_new] = ACTIONS(1732),
    [anon_sym_LBRACE_LT] = ACTIONS(1734),
    [anon_sym_begin] = ACTIONS(1732),
    [sym_ocamlyacc_value] = ACTIONS(1734),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1732),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1734),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1732),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1732),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1734),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1734),
    [aux_sym_number_token1] = ACTIONS(1734),
    [aux_sym_signed_number_token1] = ACTIONS(1732),
    [anon_sym_SQUOTE] = ACTIONS(1734),
    [anon_sym_DQUOTE] = ACTIONS(1734),
    [sym_prefix_operator] = ACTIONS(1734),
    [aux_sym_sign_operator_token1] = ACTIONS(1732),
    [sym_hash_operator] = ACTIONS(1734),
    [anon_sym_lsl] = ACTIONS(1732),
    [anon_sym_lsr] = ACTIONS(1732),
    [anon_sym_asr] = ACTIONS(1732),
    [aux_sym_pow_operator_token1] = ACTIONS(1734),
    [anon_sym_mod] = ACTIONS(1732),
    [anon_sym_land] = ACTIONS(1732),
    [anon_sym_lor] = ACTIONS(1732),
    [anon_sym_lxor] = ACTIONS(1732),
    [aux_sym_mult_operator_token1] = ACTIONS(1732),
    [aux_sym_add_operator_token1] = ACTIONS(1732),
    [sym_concat_operator] = ACTIONS(1734),
    [sym_rel_operator] = ACTIONS(1732),
    [anon_sym_AMP_AMP] = ACTIONS(1732),
    [anon_sym_or] = ACTIONS(1732),
    [anon_sym_PIPE_PIPE] = ACTIONS(1732),
    [sym_assign_operator] = ACTIONS(1734),
    [sym_let_operator] = ACTIONS(1734),
    [sym_let_and_operator] = ACTIONS(1734),
    [sym__uppercase_identifier] = ACTIONS(1734),
    [aux_sym_directive_token1] = ACTIONS(1732),
    [aux_sym_tag_token1] = ACTIONS(1734),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(995)] = {
    [sym_attribute] = STATE(995),
    [sym__lowercase_identifier] = ACTIONS(2021),
    [anon_sym_SEMI_SEMI] = ACTIONS(2023),
    [anon_sym_let] = ACTIONS(2021),
    [anon_sym_and] = ACTIONS(2021),
    [anon_sym_TILDE] = ACTIONS(2021),
    [anon_sym_QMARK] = ACTIONS(2021),
    [anon_sym_LPAREN] = ACTIONS(2023),
    [anon_sym_external] = ACTIONS(2021),
    [anon_sym_type] = ACTIONS(2021),
    [anon_sym_COMMA] = ACTIONS(2023),
    [anon_sym_PIPE] = ACTIONS(2021),
    [anon_sym_DOT] = ACTIONS(2023),
    [anon_sym_LBRACK] = ACTIONS(2021),
    [anon_sym_RBRACK] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(2021),
    [anon_sym_false] = ACTIONS(2021),
    [anon_sym_LBRACE] = ACTIONS(2021),
    [anon_sym_SEMI] = ACTIONS(2021),
    [anon_sym_exception] = ACTIONS(2021),
    [anon_sym_module] = ACTIONS(2021),
    [anon_sym_open] = ACTIONS(2021),
    [anon_sym_include] = ACTIONS(2021),
    [anon_sym_class] = ACTIONS(2021),
    [anon_sym_val] = ACTIONS(2021),
    [anon_sym_end] = ACTIONS(2021),
    [anon_sym_object] = ACTIONS(2021),
    [anon_sym_in] = ACTIONS(2021),
    [anon_sym_AMP] = ACTIONS(2021),
    [anon_sym_POUND] = ACTIONS(2021),
    [anon_sym_COLON_COLON] = ACTIONS(2023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2023),
    [anon_sym_else] = ACTIONS(2021),
    [anon_sym_new] = ACTIONS(2021),
    [anon_sym_LBRACE_LT] = ACTIONS(2023),
    [anon_sym_begin] = ACTIONS(2021),
    [sym_ocamlyacc_value] = ACTIONS(2023),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2023),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2023),
    [aux_sym_number_token1] = ACTIONS(2023),
    [aux_sym_signed_number_token1] = ACTIONS(2021),
    [anon_sym_SQUOTE] = ACTIONS(2023),
    [anon_sym_DQUOTE] = ACTIONS(2023),
    [sym_prefix_operator] = ACTIONS(2023),
    [aux_sym_sign_operator_token1] = ACTIONS(2021),
    [sym_hash_operator] = ACTIONS(2023),
    [anon_sym_lsl] = ACTIONS(2021),
    [anon_sym_lsr] = ACTIONS(2021),
    [anon_sym_asr] = ACTIONS(2021),
    [aux_sym_pow_operator_token1] = ACTIONS(2023),
    [anon_sym_mod] = ACTIONS(2021),
    [anon_sym_land] = ACTIONS(2021),
    [anon_sym_lor] = ACTIONS(2021),
    [anon_sym_lxor] = ACTIONS(2021),
    [aux_sym_mult_operator_token1] = ACTIONS(2021),
    [aux_sym_add_operator_token1] = ACTIONS(2021),
    [sym_concat_operator] = ACTIONS(2023),
    [sym_rel_operator] = ACTIONS(2021),
    [anon_sym_AMP_AMP] = ACTIONS(2021),
    [anon_sym_or] = ACTIONS(2021),
    [anon_sym_PIPE_PIPE] = ACTIONS(2021),
    [sym_assign_operator] = ACTIONS(2023),
    [sym_let_operator] = ACTIONS(2023),
    [sym_let_and_operator] = ACTIONS(2023),
    [sym__uppercase_identifier] = ACTIONS(2023),
    [aux_sym_directive_token1] = ACTIONS(2021),
    [aux_sym_tag_token1] = ACTIONS(2023),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(996)] = {
    [sym_attribute] = STATE(996),
    [sym__lowercase_identifier] = ACTIONS(1874),
    [anon_sym_SEMI_SEMI] = ACTIONS(1876),
    [anon_sym_let] = ACTIONS(1874),
    [anon_sym_and] = ACTIONS(1874),
    [anon_sym_TILDE] = ACTIONS(1874),
    [anon_sym_QMARK] = ACTIONS(1874),
    [anon_sym_LPAREN] = ACTIONS(1876),
    [anon_sym_external] = ACTIONS(1874),
    [anon_sym_type] = ACTIONS(1874),
    [anon_sym_COMMA] = ACTIONS(1876),
    [anon_sym_PIPE] = ACTIONS(1874),
    [anon_sym_DOT] = ACTIONS(1876),
    [anon_sym_LBRACK] = ACTIONS(1874),
    [anon_sym_RBRACK] = ACTIONS(1876),
    [anon_sym_true] = ACTIONS(1874),
    [anon_sym_false] = ACTIONS(1874),
    [anon_sym_LBRACE] = ACTIONS(1874),
    [anon_sym_SEMI] = ACTIONS(1874),
    [anon_sym_exception] = ACTIONS(1874),
    [anon_sym_module] = ACTIONS(1874),
    [anon_sym_open] = ACTIONS(1874),
    [anon_sym_include] = ACTIONS(1874),
    [anon_sym_class] = ACTIONS(1874),
    [anon_sym_val] = ACTIONS(1874),
    [anon_sym_end] = ACTIONS(1874),
    [anon_sym_object] = ACTIONS(1874),
    [anon_sym_in] = ACTIONS(1874),
    [anon_sym_AMP] = ACTIONS(1874),
    [anon_sym_POUND] = ACTIONS(1874),
    [anon_sym_COLON_COLON] = ACTIONS(1876),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1876),
    [anon_sym_else] = ACTIONS(1874),
    [anon_sym_new] = ACTIONS(1874),
    [anon_sym_LBRACE_LT] = ACTIONS(1876),
    [anon_sym_begin] = ACTIONS(1874),
    [sym_ocamlyacc_value] = ACTIONS(1876),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1874),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1876),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1874),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1874),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1876),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1876),
    [aux_sym_number_token1] = ACTIONS(1876),
    [aux_sym_signed_number_token1] = ACTIONS(1874),
    [anon_sym_SQUOTE] = ACTIONS(1876),
    [anon_sym_DQUOTE] = ACTIONS(1876),
    [sym_prefix_operator] = ACTIONS(1876),
    [aux_sym_sign_operator_token1] = ACTIONS(1874),
    [sym_hash_operator] = ACTIONS(1876),
    [anon_sym_lsl] = ACTIONS(1874),
    [anon_sym_lsr] = ACTIONS(1874),
    [anon_sym_asr] = ACTIONS(1874),
    [aux_sym_pow_operator_token1] = ACTIONS(1876),
    [anon_sym_mod] = ACTIONS(1874),
    [anon_sym_land] = ACTIONS(1874),
    [anon_sym_lor] = ACTIONS(1874),
    [anon_sym_lxor] = ACTIONS(1874),
    [aux_sym_mult_operator_token1] = ACTIONS(1874),
    [aux_sym_add_operator_token1] = ACTIONS(1874),
    [sym_concat_operator] = ACTIONS(1876),
    [sym_rel_operator] = ACTIONS(1874),
    [anon_sym_AMP_AMP] = ACTIONS(1874),
    [anon_sym_or] = ACTIONS(1874),
    [anon_sym_PIPE_PIPE] = ACTIONS(1874),
    [sym_assign_operator] = ACTIONS(1876),
    [sym_let_operator] = ACTIONS(1876),
    [sym_let_and_operator] = ACTIONS(1876),
    [sym__uppercase_identifier] = ACTIONS(1876),
    [aux_sym_directive_token1] = ACTIONS(1874),
    [aux_sym_tag_token1] = ACTIONS(1876),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(997)] = {
    [sym_attribute] = STATE(997),
    [sym__lowercase_identifier] = ACTIONS(1582),
    [anon_sym_SEMI_SEMI] = ACTIONS(1584),
    [anon_sym_let] = ACTIONS(1582),
    [anon_sym_and] = ACTIONS(1582),
    [anon_sym_TILDE] = ACTIONS(1582),
    [anon_sym_QMARK] = ACTIONS(1582),
    [anon_sym_LPAREN] = ACTIONS(1584),
    [anon_sym_external] = ACTIONS(1582),
    [anon_sym_type] = ACTIONS(1582),
    [anon_sym_COMMA] = ACTIONS(1584),
    [anon_sym_PIPE] = ACTIONS(1582),
    [anon_sym_DOT] = ACTIONS(1584),
    [anon_sym_LBRACK] = ACTIONS(1582),
    [anon_sym_RBRACK] = ACTIONS(1584),
    [anon_sym_true] = ACTIONS(1582),
    [anon_sym_false] = ACTIONS(1582),
    [anon_sym_LBRACE] = ACTIONS(1582),
    [anon_sym_SEMI] = ACTIONS(1582),
    [anon_sym_exception] = ACTIONS(1582),
    [anon_sym_module] = ACTIONS(1582),
    [anon_sym_open] = ACTIONS(1582),
    [anon_sym_include] = ACTIONS(1582),
    [anon_sym_class] = ACTIONS(1582),
    [anon_sym_val] = ACTIONS(1582),
    [anon_sym_end] = ACTIONS(1582),
    [anon_sym_object] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1582),
    [anon_sym_AMP] = ACTIONS(1582),
    [anon_sym_POUND] = ACTIONS(1582),
    [anon_sym_COLON_COLON] = ACTIONS(1584),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1584),
    [anon_sym_else] = ACTIONS(1582),
    [anon_sym_new] = ACTIONS(1582),
    [anon_sym_LBRACE_LT] = ACTIONS(1584),
    [anon_sym_begin] = ACTIONS(1582),
    [sym_ocamlyacc_value] = ACTIONS(1584),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1582),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1584),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1582),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1582),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1584),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1584),
    [aux_sym_number_token1] = ACTIONS(1584),
    [aux_sym_signed_number_token1] = ACTIONS(1582),
    [anon_sym_SQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1584),
    [sym_prefix_operator] = ACTIONS(1584),
    [aux_sym_sign_operator_token1] = ACTIONS(1582),
    [sym_hash_operator] = ACTIONS(1584),
    [anon_sym_lsl] = ACTIONS(1582),
    [anon_sym_lsr] = ACTIONS(1582),
    [anon_sym_asr] = ACTIONS(1582),
    [aux_sym_pow_operator_token1] = ACTIONS(1584),
    [anon_sym_mod] = ACTIONS(1582),
    [anon_sym_land] = ACTIONS(1582),
    [anon_sym_lor] = ACTIONS(1582),
    [anon_sym_lxor] = ACTIONS(1582),
    [aux_sym_mult_operator_token1] = ACTIONS(1582),
    [aux_sym_add_operator_token1] = ACTIONS(1582),
    [sym_concat_operator] = ACTIONS(1584),
    [sym_rel_operator] = ACTIONS(1582),
    [anon_sym_AMP_AMP] = ACTIONS(1582),
    [anon_sym_or] = ACTIONS(1582),
    [anon_sym_PIPE_PIPE] = ACTIONS(1582),
    [sym_assign_operator] = ACTIONS(1584),
    [sym_let_operator] = ACTIONS(1584),
    [sym_let_and_operator] = ACTIONS(1584),
    [sym__uppercase_identifier] = ACTIONS(1584),
    [aux_sym_directive_token1] = ACTIONS(1582),
    [aux_sym_tag_token1] = ACTIONS(1584),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(998)] = {
    [sym_attribute] = STATE(998),
    [sym__lowercase_identifier] = ACTIONS(1754),
    [anon_sym_SEMI_SEMI] = ACTIONS(1756),
    [anon_sym_let] = ACTIONS(1754),
    [anon_sym_and] = ACTIONS(1754),
    [anon_sym_TILDE] = ACTIONS(1754),
    [anon_sym_QMARK] = ACTIONS(1754),
    [anon_sym_LPAREN] = ACTIONS(1756),
    [anon_sym_external] = ACTIONS(1754),
    [anon_sym_type] = ACTIONS(1754),
    [anon_sym_COMMA] = ACTIONS(1756),
    [anon_sym_PIPE] = ACTIONS(1754),
    [anon_sym_DOT] = ACTIONS(1756),
    [anon_sym_LBRACK] = ACTIONS(1754),
    [anon_sym_RBRACK] = ACTIONS(1756),
    [anon_sym_true] = ACTIONS(1754),
    [anon_sym_false] = ACTIONS(1754),
    [anon_sym_LBRACE] = ACTIONS(1754),
    [anon_sym_SEMI] = ACTIONS(1754),
    [anon_sym_exception] = ACTIONS(1754),
    [anon_sym_module] = ACTIONS(1754),
    [anon_sym_open] = ACTIONS(1754),
    [anon_sym_include] = ACTIONS(1754),
    [anon_sym_class] = ACTIONS(1754),
    [anon_sym_val] = ACTIONS(1754),
    [anon_sym_end] = ACTIONS(1754),
    [anon_sym_object] = ACTIONS(1754),
    [anon_sym_in] = ACTIONS(1754),
    [anon_sym_AMP] = ACTIONS(1754),
    [anon_sym_POUND] = ACTIONS(1754),
    [anon_sym_COLON_COLON] = ACTIONS(1756),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1756),
    [anon_sym_else] = ACTIONS(1754),
    [anon_sym_new] = ACTIONS(1754),
    [anon_sym_LBRACE_LT] = ACTIONS(1756),
    [anon_sym_begin] = ACTIONS(1754),
    [sym_ocamlyacc_value] = ACTIONS(1756),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1754),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1756),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1754),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1754),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1756),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1756),
    [aux_sym_number_token1] = ACTIONS(1756),
    [aux_sym_signed_number_token1] = ACTIONS(1754),
    [anon_sym_SQUOTE] = ACTIONS(1756),
    [anon_sym_DQUOTE] = ACTIONS(1756),
    [sym_prefix_operator] = ACTIONS(1756),
    [aux_sym_sign_operator_token1] = ACTIONS(1754),
    [sym_hash_operator] = ACTIONS(1756),
    [anon_sym_lsl] = ACTIONS(1754),
    [anon_sym_lsr] = ACTIONS(1754),
    [anon_sym_asr] = ACTIONS(1754),
    [aux_sym_pow_operator_token1] = ACTIONS(1756),
    [anon_sym_mod] = ACTIONS(1754),
    [anon_sym_land] = ACTIONS(1754),
    [anon_sym_lor] = ACTIONS(1754),
    [anon_sym_lxor] = ACTIONS(1754),
    [aux_sym_mult_operator_token1] = ACTIONS(1754),
    [aux_sym_add_operator_token1] = ACTIONS(1754),
    [sym_concat_operator] = ACTIONS(1756),
    [sym_rel_operator] = ACTIONS(1754),
    [anon_sym_AMP_AMP] = ACTIONS(1754),
    [anon_sym_or] = ACTIONS(1754),
    [anon_sym_PIPE_PIPE] = ACTIONS(1754),
    [sym_assign_operator] = ACTIONS(1756),
    [sym_let_operator] = ACTIONS(1756),
    [sym_let_and_operator] = ACTIONS(1756),
    [sym__uppercase_identifier] = ACTIONS(1756),
    [aux_sym_directive_token1] = ACTIONS(1754),
    [aux_sym_tag_token1] = ACTIONS(1756),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(999)] = {
    [sym_attribute] = STATE(999),
    [sym__lowercase_identifier] = ACTIONS(1758),
    [anon_sym_SEMI_SEMI] = ACTIONS(1760),
    [anon_sym_let] = ACTIONS(1758),
    [anon_sym_and] = ACTIONS(1758),
    [anon_sym_TILDE] = ACTIONS(1758),
    [anon_sym_QMARK] = ACTIONS(1758),
    [anon_sym_LPAREN] = ACTIONS(1760),
    [anon_sym_external] = ACTIONS(1758),
    [anon_sym_type] = ACTIONS(1758),
    [anon_sym_COMMA] = ACTIONS(1760),
    [anon_sym_PIPE] = ACTIONS(1758),
    [anon_sym_DOT] = ACTIONS(1760),
    [anon_sym_LBRACK] = ACTIONS(1758),
    [anon_sym_RBRACK] = ACTIONS(1760),
    [anon_sym_true] = ACTIONS(1758),
    [anon_sym_false] = ACTIONS(1758),
    [anon_sym_LBRACE] = ACTIONS(1758),
    [anon_sym_SEMI] = ACTIONS(1758),
    [anon_sym_exception] = ACTIONS(1758),
    [anon_sym_module] = ACTIONS(1758),
    [anon_sym_open] = ACTIONS(1758),
    [anon_sym_include] = ACTIONS(1758),
    [anon_sym_class] = ACTIONS(1758),
    [anon_sym_val] = ACTIONS(1758),
    [anon_sym_end] = ACTIONS(1758),
    [anon_sym_object] = ACTIONS(1758),
    [anon_sym_in] = ACTIONS(1758),
    [anon_sym_AMP] = ACTIONS(1758),
    [anon_sym_POUND] = ACTIONS(1758),
    [anon_sym_COLON_COLON] = ACTIONS(1760),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1760),
    [anon_sym_else] = ACTIONS(1758),
    [anon_sym_new] = ACTIONS(1758),
    [anon_sym_LBRACE_LT] = ACTIONS(1760),
    [anon_sym_begin] = ACTIONS(1758),
    [sym_ocamlyacc_value] = ACTIONS(1760),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1758),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1760),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1758),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1758),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1760),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1760),
    [aux_sym_number_token1] = ACTIONS(1760),
    [aux_sym_signed_number_token1] = ACTIONS(1758),
    [anon_sym_SQUOTE] = ACTIONS(1760),
    [anon_sym_DQUOTE] = ACTIONS(1760),
    [sym_prefix_operator] = ACTIONS(1760),
    [aux_sym_sign_operator_token1] = ACTIONS(1758),
    [sym_hash_operator] = ACTIONS(1760),
    [anon_sym_lsl] = ACTIONS(1758),
    [anon_sym_lsr] = ACTIONS(1758),
    [anon_sym_asr] = ACTIONS(1758),
    [aux_sym_pow_operator_token1] = ACTIONS(1760),
    [anon_sym_mod] = ACTIONS(1758),
    [anon_sym_land] = ACTIONS(1758),
    [anon_sym_lor] = ACTIONS(1758),
    [anon_sym_lxor] = ACTIONS(1758),
    [aux_sym_mult_operator_token1] = ACTIONS(1758),
    [aux_sym_add_operator_token1] = ACTIONS(1758),
    [sym_concat_operator] = ACTIONS(1760),
    [sym_rel_operator] = ACTIONS(1758),
    [anon_sym_AMP_AMP] = ACTIONS(1758),
    [anon_sym_or] = ACTIONS(1758),
    [anon_sym_PIPE_PIPE] = ACTIONS(1758),
    [sym_assign_operator] = ACTIONS(1760),
    [sym_let_operator] = ACTIONS(1760),
    [sym_let_and_operator] = ACTIONS(1760),
    [sym__uppercase_identifier] = ACTIONS(1760),
    [aux_sym_directive_token1] = ACTIONS(1758),
    [aux_sym_tag_token1] = ACTIONS(1760),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1000)] = {
    [sym_attribute] = STATE(1000),
    [sym__lowercase_identifier] = ACTIONS(1918),
    [anon_sym_SEMI_SEMI] = ACTIONS(1920),
    [anon_sym_let] = ACTIONS(1918),
    [anon_sym_and] = ACTIONS(1918),
    [anon_sym_TILDE] = ACTIONS(1918),
    [anon_sym_QMARK] = ACTIONS(1918),
    [anon_sym_LPAREN] = ACTIONS(1920),
    [anon_sym_external] = ACTIONS(1918),
    [anon_sym_type] = ACTIONS(1918),
    [anon_sym_COMMA] = ACTIONS(1920),
    [anon_sym_PIPE] = ACTIONS(1918),
    [anon_sym_DOT] = ACTIONS(1920),
    [anon_sym_LBRACK] = ACTIONS(1918),
    [anon_sym_RBRACK] = ACTIONS(1920),
    [anon_sym_true] = ACTIONS(1918),
    [anon_sym_false] = ACTIONS(1918),
    [anon_sym_LBRACE] = ACTIONS(1918),
    [anon_sym_SEMI] = ACTIONS(1918),
    [anon_sym_exception] = ACTIONS(1918),
    [anon_sym_module] = ACTIONS(1918),
    [anon_sym_open] = ACTIONS(1918),
    [anon_sym_include] = ACTIONS(1918),
    [anon_sym_class] = ACTIONS(1918),
    [anon_sym_val] = ACTIONS(1918),
    [anon_sym_end] = ACTIONS(1918),
    [anon_sym_object] = ACTIONS(1918),
    [anon_sym_in] = ACTIONS(1918),
    [anon_sym_AMP] = ACTIONS(1918),
    [anon_sym_POUND] = ACTIONS(1918),
    [anon_sym_COLON_COLON] = ACTIONS(1920),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1920),
    [anon_sym_else] = ACTIONS(1918),
    [anon_sym_new] = ACTIONS(1918),
    [anon_sym_LBRACE_LT] = ACTIONS(1920),
    [anon_sym_begin] = ACTIONS(1918),
    [sym_ocamlyacc_value] = ACTIONS(1920),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1918),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1920),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1918),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1918),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1920),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1920),
    [aux_sym_number_token1] = ACTIONS(1920),
    [aux_sym_signed_number_token1] = ACTIONS(1918),
    [anon_sym_SQUOTE] = ACTIONS(1920),
    [anon_sym_DQUOTE] = ACTIONS(1920),
    [sym_prefix_operator] = ACTIONS(1920),
    [aux_sym_sign_operator_token1] = ACTIONS(1918),
    [sym_hash_operator] = ACTIONS(1920),
    [anon_sym_lsl] = ACTIONS(1918),
    [anon_sym_lsr] = ACTIONS(1918),
    [anon_sym_asr] = ACTIONS(1918),
    [aux_sym_pow_operator_token1] = ACTIONS(1920),
    [anon_sym_mod] = ACTIONS(1918),
    [anon_sym_land] = ACTIONS(1918),
    [anon_sym_lor] = ACTIONS(1918),
    [anon_sym_lxor] = ACTIONS(1918),
    [aux_sym_mult_operator_token1] = ACTIONS(1918),
    [aux_sym_add_operator_token1] = ACTIONS(1918),
    [sym_concat_operator] = ACTIONS(1920),
    [sym_rel_operator] = ACTIONS(1918),
    [anon_sym_AMP_AMP] = ACTIONS(1918),
    [anon_sym_or] = ACTIONS(1918),
    [anon_sym_PIPE_PIPE] = ACTIONS(1918),
    [sym_assign_operator] = ACTIONS(1920),
    [sym_let_operator] = ACTIONS(1920),
    [sym_let_and_operator] = ACTIONS(1920),
    [sym__uppercase_identifier] = ACTIONS(1920),
    [aux_sym_directive_token1] = ACTIONS(1918),
    [aux_sym_tag_token1] = ACTIONS(1920),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1001)] = {
    [sym_attribute] = STATE(1001),
    [sym__lowercase_identifier] = ACTIONS(1926),
    [anon_sym_SEMI_SEMI] = ACTIONS(1928),
    [anon_sym_let] = ACTIONS(1926),
    [anon_sym_and] = ACTIONS(1926),
    [anon_sym_TILDE] = ACTIONS(1926),
    [anon_sym_QMARK] = ACTIONS(1926),
    [anon_sym_LPAREN] = ACTIONS(1928),
    [anon_sym_external] = ACTIONS(1926),
    [anon_sym_type] = ACTIONS(1926),
    [anon_sym_COMMA] = ACTIONS(1928),
    [anon_sym_PIPE] = ACTIONS(1926),
    [anon_sym_DOT] = ACTIONS(1928),
    [anon_sym_LBRACK] = ACTIONS(1926),
    [anon_sym_RBRACK] = ACTIONS(1928),
    [anon_sym_true] = ACTIONS(1926),
    [anon_sym_false] = ACTIONS(1926),
    [anon_sym_LBRACE] = ACTIONS(1926),
    [anon_sym_SEMI] = ACTIONS(1926),
    [anon_sym_exception] = ACTIONS(1926),
    [anon_sym_module] = ACTIONS(1926),
    [anon_sym_open] = ACTIONS(1926),
    [anon_sym_include] = ACTIONS(1926),
    [anon_sym_class] = ACTIONS(1926),
    [anon_sym_val] = ACTIONS(1926),
    [anon_sym_end] = ACTIONS(1926),
    [anon_sym_object] = ACTIONS(1926),
    [anon_sym_in] = ACTIONS(1926),
    [anon_sym_AMP] = ACTIONS(1926),
    [anon_sym_POUND] = ACTIONS(1926),
    [anon_sym_COLON_COLON] = ACTIONS(1928),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1928),
    [anon_sym_else] = ACTIONS(1926),
    [anon_sym_new] = ACTIONS(1926),
    [anon_sym_LBRACE_LT] = ACTIONS(1928),
    [anon_sym_begin] = ACTIONS(1926),
    [sym_ocamlyacc_value] = ACTIONS(1928),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1926),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1928),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1926),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1926),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1928),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1928),
    [aux_sym_number_token1] = ACTIONS(1928),
    [aux_sym_signed_number_token1] = ACTIONS(1926),
    [anon_sym_SQUOTE] = ACTIONS(1928),
    [anon_sym_DQUOTE] = ACTIONS(1928),
    [sym_prefix_operator] = ACTIONS(1928),
    [aux_sym_sign_operator_token1] = ACTIONS(1926),
    [sym_hash_operator] = ACTIONS(1928),
    [anon_sym_lsl] = ACTIONS(1926),
    [anon_sym_lsr] = ACTIONS(1926),
    [anon_sym_asr] = ACTIONS(1926),
    [aux_sym_pow_operator_token1] = ACTIONS(1928),
    [anon_sym_mod] = ACTIONS(1926),
    [anon_sym_land] = ACTIONS(1926),
    [anon_sym_lor] = ACTIONS(1926),
    [anon_sym_lxor] = ACTIONS(1926),
    [aux_sym_mult_operator_token1] = ACTIONS(1926),
    [aux_sym_add_operator_token1] = ACTIONS(1926),
    [sym_concat_operator] = ACTIONS(1928),
    [sym_rel_operator] = ACTIONS(1926),
    [anon_sym_AMP_AMP] = ACTIONS(1926),
    [anon_sym_or] = ACTIONS(1926),
    [anon_sym_PIPE_PIPE] = ACTIONS(1926),
    [sym_assign_operator] = ACTIONS(1928),
    [sym_let_operator] = ACTIONS(1928),
    [sym_let_and_operator] = ACTIONS(1928),
    [sym__uppercase_identifier] = ACTIONS(1928),
    [aux_sym_directive_token1] = ACTIONS(1926),
    [aux_sym_tag_token1] = ACTIONS(1928),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1002)] = {
    [sym_attribute] = STATE(1002),
    [sym__lowercase_identifier] = ACTIONS(1930),
    [anon_sym_SEMI_SEMI] = ACTIONS(1932),
    [anon_sym_let] = ACTIONS(1930),
    [anon_sym_and] = ACTIONS(1930),
    [anon_sym_TILDE] = ACTIONS(1930),
    [anon_sym_QMARK] = ACTIONS(1930),
    [anon_sym_LPAREN] = ACTIONS(1932),
    [anon_sym_external] = ACTIONS(1930),
    [anon_sym_type] = ACTIONS(1930),
    [anon_sym_COMMA] = ACTIONS(1932),
    [anon_sym_PIPE] = ACTIONS(1930),
    [anon_sym_DOT] = ACTIONS(1932),
    [anon_sym_LBRACK] = ACTIONS(1930),
    [anon_sym_RBRACK] = ACTIONS(1932),
    [anon_sym_true] = ACTIONS(1930),
    [anon_sym_false] = ACTIONS(1930),
    [anon_sym_LBRACE] = ACTIONS(1930),
    [anon_sym_SEMI] = ACTIONS(1930),
    [anon_sym_exception] = ACTIONS(1930),
    [anon_sym_module] = ACTIONS(1930),
    [anon_sym_open] = ACTIONS(1930),
    [anon_sym_include] = ACTIONS(1930),
    [anon_sym_class] = ACTIONS(1930),
    [anon_sym_val] = ACTIONS(1930),
    [anon_sym_end] = ACTIONS(1930),
    [anon_sym_object] = ACTIONS(1930),
    [anon_sym_in] = ACTIONS(1930),
    [anon_sym_AMP] = ACTIONS(1930),
    [anon_sym_POUND] = ACTIONS(1930),
    [anon_sym_COLON_COLON] = ACTIONS(1932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1932),
    [anon_sym_else] = ACTIONS(1930),
    [anon_sym_new] = ACTIONS(1930),
    [anon_sym_LBRACE_LT] = ACTIONS(1932),
    [anon_sym_begin] = ACTIONS(1930),
    [sym_ocamlyacc_value] = ACTIONS(1932),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1930),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1932),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1930),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1930),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1932),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1932),
    [aux_sym_number_token1] = ACTIONS(1932),
    [aux_sym_signed_number_token1] = ACTIONS(1930),
    [anon_sym_SQUOTE] = ACTIONS(1932),
    [anon_sym_DQUOTE] = ACTIONS(1932),
    [sym_prefix_operator] = ACTIONS(1932),
    [aux_sym_sign_operator_token1] = ACTIONS(1930),
    [sym_hash_operator] = ACTIONS(1932),
    [anon_sym_lsl] = ACTIONS(1930),
    [anon_sym_lsr] = ACTIONS(1930),
    [anon_sym_asr] = ACTIONS(1930),
    [aux_sym_pow_operator_token1] = ACTIONS(1932),
    [anon_sym_mod] = ACTIONS(1930),
    [anon_sym_land] = ACTIONS(1930),
    [anon_sym_lor] = ACTIONS(1930),
    [anon_sym_lxor] = ACTIONS(1930),
    [aux_sym_mult_operator_token1] = ACTIONS(1930),
    [aux_sym_add_operator_token1] = ACTIONS(1930),
    [sym_concat_operator] = ACTIONS(1932),
    [sym_rel_operator] = ACTIONS(1930),
    [anon_sym_AMP_AMP] = ACTIONS(1930),
    [anon_sym_or] = ACTIONS(1930),
    [anon_sym_PIPE_PIPE] = ACTIONS(1930),
    [sym_assign_operator] = ACTIONS(1932),
    [sym_let_operator] = ACTIONS(1932),
    [sym_let_and_operator] = ACTIONS(1932),
    [sym__uppercase_identifier] = ACTIONS(1932),
    [aux_sym_directive_token1] = ACTIONS(1930),
    [aux_sym_tag_token1] = ACTIONS(1932),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1003)] = {
    [sym_attribute] = STATE(1003),
    [sym__lowercase_identifier] = ACTIONS(1634),
    [anon_sym_SEMI_SEMI] = ACTIONS(1636),
    [anon_sym_let] = ACTIONS(1634),
    [anon_sym_and] = ACTIONS(1634),
    [anon_sym_TILDE] = ACTIONS(1634),
    [anon_sym_QMARK] = ACTIONS(1634),
    [anon_sym_LPAREN] = ACTIONS(1636),
    [anon_sym_external] = ACTIONS(1634),
    [anon_sym_type] = ACTIONS(1634),
    [anon_sym_COMMA] = ACTIONS(1636),
    [anon_sym_PIPE] = ACTIONS(1634),
    [anon_sym_DOT] = ACTIONS(1636),
    [anon_sym_LBRACK] = ACTIONS(1634),
    [anon_sym_RBRACK] = ACTIONS(1636),
    [anon_sym_true] = ACTIONS(1634),
    [anon_sym_false] = ACTIONS(1634),
    [anon_sym_LBRACE] = ACTIONS(1634),
    [anon_sym_SEMI] = ACTIONS(1634),
    [anon_sym_exception] = ACTIONS(1634),
    [anon_sym_module] = ACTIONS(1634),
    [anon_sym_open] = ACTIONS(1634),
    [anon_sym_include] = ACTIONS(1634),
    [anon_sym_class] = ACTIONS(1634),
    [anon_sym_val] = ACTIONS(1634),
    [anon_sym_end] = ACTIONS(1634),
    [anon_sym_object] = ACTIONS(1634),
    [anon_sym_in] = ACTIONS(1634),
    [anon_sym_AMP] = ACTIONS(1634),
    [anon_sym_POUND] = ACTIONS(1634),
    [anon_sym_COLON_COLON] = ACTIONS(1636),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1636),
    [anon_sym_else] = ACTIONS(1634),
    [anon_sym_new] = ACTIONS(1634),
    [anon_sym_LBRACE_LT] = ACTIONS(1636),
    [anon_sym_begin] = ACTIONS(1634),
    [sym_ocamlyacc_value] = ACTIONS(1636),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1634),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1636),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1634),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1634),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1636),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1636),
    [aux_sym_number_token1] = ACTIONS(1636),
    [aux_sym_signed_number_token1] = ACTIONS(1634),
    [anon_sym_SQUOTE] = ACTIONS(1636),
    [anon_sym_DQUOTE] = ACTIONS(1636),
    [sym_prefix_operator] = ACTIONS(1636),
    [aux_sym_sign_operator_token1] = ACTIONS(1634),
    [sym_hash_operator] = ACTIONS(1636),
    [anon_sym_lsl] = ACTIONS(1634),
    [anon_sym_lsr] = ACTIONS(1634),
    [anon_sym_asr] = ACTIONS(1634),
    [aux_sym_pow_operator_token1] = ACTIONS(1636),
    [anon_sym_mod] = ACTIONS(1634),
    [anon_sym_land] = ACTIONS(1634),
    [anon_sym_lor] = ACTIONS(1634),
    [anon_sym_lxor] = ACTIONS(1634),
    [aux_sym_mult_operator_token1] = ACTIONS(1634),
    [aux_sym_add_operator_token1] = ACTIONS(1634),
    [sym_concat_operator] = ACTIONS(1636),
    [sym_rel_operator] = ACTIONS(1634),
    [anon_sym_AMP_AMP] = ACTIONS(1634),
    [anon_sym_or] = ACTIONS(1634),
    [anon_sym_PIPE_PIPE] = ACTIONS(1634),
    [sym_assign_operator] = ACTIONS(1636),
    [sym_let_operator] = ACTIONS(1636),
    [sym_let_and_operator] = ACTIONS(1636),
    [sym__uppercase_identifier] = ACTIONS(1636),
    [aux_sym_directive_token1] = ACTIONS(1634),
    [aux_sym_tag_token1] = ACTIONS(1636),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1004)] = {
    [sym_attribute] = STATE(1004),
    [sym__lowercase_identifier] = ACTIONS(1838),
    [anon_sym_SEMI_SEMI] = ACTIONS(1840),
    [anon_sym_let] = ACTIONS(1838),
    [anon_sym_and] = ACTIONS(1838),
    [anon_sym_TILDE] = ACTIONS(1838),
    [anon_sym_QMARK] = ACTIONS(1838),
    [anon_sym_LPAREN] = ACTIONS(1840),
    [anon_sym_external] = ACTIONS(1838),
    [anon_sym_type] = ACTIONS(1838),
    [anon_sym_COMMA] = ACTIONS(1840),
    [anon_sym_PIPE] = ACTIONS(1838),
    [anon_sym_DOT] = ACTIONS(1840),
    [anon_sym_LBRACK] = ACTIONS(1838),
    [anon_sym_RBRACK] = ACTIONS(1840),
    [anon_sym_true] = ACTIONS(1838),
    [anon_sym_false] = ACTIONS(1838),
    [anon_sym_LBRACE] = ACTIONS(1838),
    [anon_sym_SEMI] = ACTIONS(1838),
    [anon_sym_exception] = ACTIONS(1838),
    [anon_sym_module] = ACTIONS(1838),
    [anon_sym_open] = ACTIONS(1838),
    [anon_sym_include] = ACTIONS(1838),
    [anon_sym_class] = ACTIONS(1838),
    [anon_sym_val] = ACTIONS(1838),
    [anon_sym_end] = ACTIONS(1838),
    [anon_sym_object] = ACTIONS(1838),
    [anon_sym_in] = ACTIONS(1838),
    [anon_sym_AMP] = ACTIONS(1838),
    [anon_sym_POUND] = ACTIONS(1838),
    [anon_sym_COLON_COLON] = ACTIONS(1840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1840),
    [anon_sym_else] = ACTIONS(1838),
    [anon_sym_new] = ACTIONS(1838),
    [anon_sym_LBRACE_LT] = ACTIONS(1840),
    [anon_sym_begin] = ACTIONS(1838),
    [sym_ocamlyacc_value] = ACTIONS(1840),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1838),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1840),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1838),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1838),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1840),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1840),
    [aux_sym_number_token1] = ACTIONS(1840),
    [aux_sym_signed_number_token1] = ACTIONS(1838),
    [anon_sym_SQUOTE] = ACTIONS(1840),
    [anon_sym_DQUOTE] = ACTIONS(1840),
    [sym_prefix_operator] = ACTIONS(1840),
    [aux_sym_sign_operator_token1] = ACTIONS(1838),
    [sym_hash_operator] = ACTIONS(1840),
    [anon_sym_lsl] = ACTIONS(1838),
    [anon_sym_lsr] = ACTIONS(1838),
    [anon_sym_asr] = ACTIONS(1838),
    [aux_sym_pow_operator_token1] = ACTIONS(1840),
    [anon_sym_mod] = ACTIONS(1838),
    [anon_sym_land] = ACTIONS(1838),
    [anon_sym_lor] = ACTIONS(1838),
    [anon_sym_lxor] = ACTIONS(1838),
    [aux_sym_mult_operator_token1] = ACTIONS(1838),
    [aux_sym_add_operator_token1] = ACTIONS(1838),
    [sym_concat_operator] = ACTIONS(1840),
    [sym_rel_operator] = ACTIONS(1838),
    [anon_sym_AMP_AMP] = ACTIONS(1838),
    [anon_sym_or] = ACTIONS(1838),
    [anon_sym_PIPE_PIPE] = ACTIONS(1838),
    [sym_assign_operator] = ACTIONS(1840),
    [sym_let_operator] = ACTIONS(1840),
    [sym_let_and_operator] = ACTIONS(1840),
    [sym__uppercase_identifier] = ACTIONS(1840),
    [aux_sym_directive_token1] = ACTIONS(1838),
    [aux_sym_tag_token1] = ACTIONS(1840),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1005)] = {
    [sym_attribute] = STATE(1005),
    [sym__lowercase_identifier] = ACTIONS(1858),
    [anon_sym_SEMI_SEMI] = ACTIONS(1860),
    [anon_sym_let] = ACTIONS(1858),
    [anon_sym_and] = ACTIONS(1858),
    [anon_sym_TILDE] = ACTIONS(1858),
    [anon_sym_QMARK] = ACTIONS(1858),
    [anon_sym_LPAREN] = ACTIONS(1860),
    [anon_sym_external] = ACTIONS(1858),
    [anon_sym_type] = ACTIONS(1858),
    [anon_sym_COMMA] = ACTIONS(1860),
    [anon_sym_PIPE] = ACTIONS(1858),
    [anon_sym_DOT] = ACTIONS(1860),
    [anon_sym_LBRACK] = ACTIONS(1858),
    [anon_sym_RBRACK] = ACTIONS(1860),
    [anon_sym_true] = ACTIONS(1858),
    [anon_sym_false] = ACTIONS(1858),
    [anon_sym_LBRACE] = ACTIONS(1858),
    [anon_sym_SEMI] = ACTIONS(1858),
    [anon_sym_exception] = ACTIONS(1858),
    [anon_sym_module] = ACTIONS(1858),
    [anon_sym_open] = ACTIONS(1858),
    [anon_sym_include] = ACTIONS(1858),
    [anon_sym_class] = ACTIONS(1858),
    [anon_sym_val] = ACTIONS(1858),
    [anon_sym_end] = ACTIONS(1858),
    [anon_sym_object] = ACTIONS(1858),
    [anon_sym_in] = ACTIONS(1858),
    [anon_sym_AMP] = ACTIONS(1858),
    [anon_sym_POUND] = ACTIONS(1858),
    [anon_sym_COLON_COLON] = ACTIONS(1860),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1860),
    [anon_sym_else] = ACTIONS(1858),
    [anon_sym_new] = ACTIONS(1858),
    [anon_sym_LBRACE_LT] = ACTIONS(1860),
    [anon_sym_begin] = ACTIONS(1858),
    [sym_ocamlyacc_value] = ACTIONS(1860),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1858),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1860),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1858),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1858),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1860),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1860),
    [aux_sym_number_token1] = ACTIONS(1860),
    [aux_sym_signed_number_token1] = ACTIONS(1858),
    [anon_sym_SQUOTE] = ACTIONS(1860),
    [anon_sym_DQUOTE] = ACTIONS(1860),
    [sym_prefix_operator] = ACTIONS(1860),
    [aux_sym_sign_operator_token1] = ACTIONS(1858),
    [sym_hash_operator] = ACTIONS(1860),
    [anon_sym_lsl] = ACTIONS(1858),
    [anon_sym_lsr] = ACTIONS(1858),
    [anon_sym_asr] = ACTIONS(1858),
    [aux_sym_pow_operator_token1] = ACTIONS(1860),
    [anon_sym_mod] = ACTIONS(1858),
    [anon_sym_land] = ACTIONS(1858),
    [anon_sym_lor] = ACTIONS(1858),
    [anon_sym_lxor] = ACTIONS(1858),
    [aux_sym_mult_operator_token1] = ACTIONS(1858),
    [aux_sym_add_operator_token1] = ACTIONS(1858),
    [sym_concat_operator] = ACTIONS(1860),
    [sym_rel_operator] = ACTIONS(1858),
    [anon_sym_AMP_AMP] = ACTIONS(1858),
    [anon_sym_or] = ACTIONS(1858),
    [anon_sym_PIPE_PIPE] = ACTIONS(1858),
    [sym_assign_operator] = ACTIONS(1860),
    [sym_let_operator] = ACTIONS(1860),
    [sym_let_and_operator] = ACTIONS(1860),
    [sym__uppercase_identifier] = ACTIONS(1860),
    [aux_sym_directive_token1] = ACTIONS(1858),
    [aux_sym_tag_token1] = ACTIONS(1860),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1006)] = {
    [sym_attribute] = STATE(1006),
    [sym__lowercase_identifier] = ACTIONS(1862),
    [anon_sym_SEMI_SEMI] = ACTIONS(1864),
    [anon_sym_let] = ACTIONS(1862),
    [anon_sym_and] = ACTIONS(1862),
    [anon_sym_TILDE] = ACTIONS(1862),
    [anon_sym_QMARK] = ACTIONS(1862),
    [anon_sym_LPAREN] = ACTIONS(1864),
    [anon_sym_external] = ACTIONS(1862),
    [anon_sym_type] = ACTIONS(1862),
    [anon_sym_COMMA] = ACTIONS(1864),
    [anon_sym_PIPE] = ACTIONS(1862),
    [anon_sym_DOT] = ACTIONS(1864),
    [anon_sym_LBRACK] = ACTIONS(1862),
    [anon_sym_RBRACK] = ACTIONS(1864),
    [anon_sym_true] = ACTIONS(1862),
    [anon_sym_false] = ACTIONS(1862),
    [anon_sym_LBRACE] = ACTIONS(1862),
    [anon_sym_SEMI] = ACTIONS(1862),
    [anon_sym_exception] = ACTIONS(1862),
    [anon_sym_module] = ACTIONS(1862),
    [anon_sym_open] = ACTIONS(1862),
    [anon_sym_include] = ACTIONS(1862),
    [anon_sym_class] = ACTIONS(1862),
    [anon_sym_val] = ACTIONS(1862),
    [anon_sym_end] = ACTIONS(1862),
    [anon_sym_object] = ACTIONS(1862),
    [anon_sym_in] = ACTIONS(1862),
    [anon_sym_AMP] = ACTIONS(1862),
    [anon_sym_POUND] = ACTIONS(1862),
    [anon_sym_COLON_COLON] = ACTIONS(1864),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1864),
    [anon_sym_else] = ACTIONS(1862),
    [anon_sym_new] = ACTIONS(1862),
    [anon_sym_LBRACE_LT] = ACTIONS(1864),
    [anon_sym_begin] = ACTIONS(1862),
    [sym_ocamlyacc_value] = ACTIONS(1864),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1862),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1864),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1862),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1862),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1864),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1864),
    [aux_sym_number_token1] = ACTIONS(1864),
    [aux_sym_signed_number_token1] = ACTIONS(1862),
    [anon_sym_SQUOTE] = ACTIONS(1864),
    [anon_sym_DQUOTE] = ACTIONS(1864),
    [sym_prefix_operator] = ACTIONS(1864),
    [aux_sym_sign_operator_token1] = ACTIONS(1862),
    [sym_hash_operator] = ACTIONS(1864),
    [anon_sym_lsl] = ACTIONS(1862),
    [anon_sym_lsr] = ACTIONS(1862),
    [anon_sym_asr] = ACTIONS(1862),
    [aux_sym_pow_operator_token1] = ACTIONS(1864),
    [anon_sym_mod] = ACTIONS(1862),
    [anon_sym_land] = ACTIONS(1862),
    [anon_sym_lor] = ACTIONS(1862),
    [anon_sym_lxor] = ACTIONS(1862),
    [aux_sym_mult_operator_token1] = ACTIONS(1862),
    [aux_sym_add_operator_token1] = ACTIONS(1862),
    [sym_concat_operator] = ACTIONS(1864),
    [sym_rel_operator] = ACTIONS(1862),
    [anon_sym_AMP_AMP] = ACTIONS(1862),
    [anon_sym_or] = ACTIONS(1862),
    [anon_sym_PIPE_PIPE] = ACTIONS(1862),
    [sym_assign_operator] = ACTIONS(1864),
    [sym_let_operator] = ACTIONS(1864),
    [sym_let_and_operator] = ACTIONS(1864),
    [sym__uppercase_identifier] = ACTIONS(1864),
    [aux_sym_directive_token1] = ACTIONS(1862),
    [aux_sym_tag_token1] = ACTIONS(1864),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1007)] = {
    [sym_attribute] = STATE(1007),
    [sym__lowercase_identifier] = ACTIONS(1866),
    [anon_sym_SEMI_SEMI] = ACTIONS(1868),
    [anon_sym_let] = ACTIONS(1866),
    [anon_sym_and] = ACTIONS(1866),
    [anon_sym_TILDE] = ACTIONS(1866),
    [anon_sym_QMARK] = ACTIONS(1866),
    [anon_sym_LPAREN] = ACTIONS(1868),
    [anon_sym_external] = ACTIONS(1866),
    [anon_sym_type] = ACTIONS(1866),
    [anon_sym_COMMA] = ACTIONS(1868),
    [anon_sym_PIPE] = ACTIONS(1866),
    [anon_sym_DOT] = ACTIONS(1868),
    [anon_sym_LBRACK] = ACTIONS(1866),
    [anon_sym_RBRACK] = ACTIONS(1868),
    [anon_sym_true] = ACTIONS(1866),
    [anon_sym_false] = ACTIONS(1866),
    [anon_sym_LBRACE] = ACTIONS(1866),
    [anon_sym_SEMI] = ACTIONS(1866),
    [anon_sym_exception] = ACTIONS(1866),
    [anon_sym_module] = ACTIONS(1866),
    [anon_sym_open] = ACTIONS(1866),
    [anon_sym_include] = ACTIONS(1866),
    [anon_sym_class] = ACTIONS(1866),
    [anon_sym_val] = ACTIONS(1866),
    [anon_sym_end] = ACTIONS(1866),
    [anon_sym_object] = ACTIONS(1866),
    [anon_sym_in] = ACTIONS(1866),
    [anon_sym_AMP] = ACTIONS(1866),
    [anon_sym_POUND] = ACTIONS(1866),
    [anon_sym_COLON_COLON] = ACTIONS(1868),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1868),
    [anon_sym_else] = ACTIONS(1866),
    [anon_sym_new] = ACTIONS(1866),
    [anon_sym_LBRACE_LT] = ACTIONS(1868),
    [anon_sym_begin] = ACTIONS(1866),
    [sym_ocamlyacc_value] = ACTIONS(1868),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1866),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1868),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1866),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1866),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1868),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1868),
    [aux_sym_number_token1] = ACTIONS(1868),
    [aux_sym_signed_number_token1] = ACTIONS(1866),
    [anon_sym_SQUOTE] = ACTIONS(1868),
    [anon_sym_DQUOTE] = ACTIONS(1868),
    [sym_prefix_operator] = ACTIONS(1868),
    [aux_sym_sign_operator_token1] = ACTIONS(1866),
    [sym_hash_operator] = ACTIONS(1868),
    [anon_sym_lsl] = ACTIONS(1866),
    [anon_sym_lsr] = ACTIONS(1866),
    [anon_sym_asr] = ACTIONS(1866),
    [aux_sym_pow_operator_token1] = ACTIONS(1868),
    [anon_sym_mod] = ACTIONS(1866),
    [anon_sym_land] = ACTIONS(1866),
    [anon_sym_lor] = ACTIONS(1866),
    [anon_sym_lxor] = ACTIONS(1866),
    [aux_sym_mult_operator_token1] = ACTIONS(1866),
    [aux_sym_add_operator_token1] = ACTIONS(1866),
    [sym_concat_operator] = ACTIONS(1868),
    [sym_rel_operator] = ACTIONS(1866),
    [anon_sym_AMP_AMP] = ACTIONS(1866),
    [anon_sym_or] = ACTIONS(1866),
    [anon_sym_PIPE_PIPE] = ACTIONS(1866),
    [sym_assign_operator] = ACTIONS(1868),
    [sym_let_operator] = ACTIONS(1868),
    [sym_let_and_operator] = ACTIONS(1868),
    [sym__uppercase_identifier] = ACTIONS(1868),
    [aux_sym_directive_token1] = ACTIONS(1866),
    [aux_sym_tag_token1] = ACTIONS(1868),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1008)] = {
    [sym_attribute] = STATE(1008),
    [sym__lowercase_identifier] = ACTIONS(1870),
    [anon_sym_SEMI_SEMI] = ACTIONS(1872),
    [anon_sym_let] = ACTIONS(1870),
    [anon_sym_and] = ACTIONS(1870),
    [anon_sym_TILDE] = ACTIONS(1870),
    [anon_sym_QMARK] = ACTIONS(1870),
    [anon_sym_LPAREN] = ACTIONS(1872),
    [anon_sym_external] = ACTIONS(1870),
    [anon_sym_type] = ACTIONS(1870),
    [anon_sym_COMMA] = ACTIONS(1872),
    [anon_sym_PIPE] = ACTIONS(1870),
    [anon_sym_DOT] = ACTIONS(1872),
    [anon_sym_LBRACK] = ACTIONS(1870),
    [anon_sym_RBRACK] = ACTIONS(1872),
    [anon_sym_true] = ACTIONS(1870),
    [anon_sym_false] = ACTIONS(1870),
    [anon_sym_LBRACE] = ACTIONS(1870),
    [anon_sym_SEMI] = ACTIONS(1870),
    [anon_sym_exception] = ACTIONS(1870),
    [anon_sym_module] = ACTIONS(1870),
    [anon_sym_open] = ACTIONS(1870),
    [anon_sym_include] = ACTIONS(1870),
    [anon_sym_class] = ACTIONS(1870),
    [anon_sym_val] = ACTIONS(1870),
    [anon_sym_end] = ACTIONS(1870),
    [anon_sym_object] = ACTIONS(1870),
    [anon_sym_in] = ACTIONS(1870),
    [anon_sym_AMP] = ACTIONS(1870),
    [anon_sym_POUND] = ACTIONS(1870),
    [anon_sym_COLON_COLON] = ACTIONS(1872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1872),
    [anon_sym_else] = ACTIONS(1870),
    [anon_sym_new] = ACTIONS(1870),
    [anon_sym_LBRACE_LT] = ACTIONS(1872),
    [anon_sym_begin] = ACTIONS(1870),
    [sym_ocamlyacc_value] = ACTIONS(1872),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1870),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1872),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1870),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1870),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1872),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1872),
    [aux_sym_number_token1] = ACTIONS(1872),
    [aux_sym_signed_number_token1] = ACTIONS(1870),
    [anon_sym_SQUOTE] = ACTIONS(1872),
    [anon_sym_DQUOTE] = ACTIONS(1872),
    [sym_prefix_operator] = ACTIONS(1872),
    [aux_sym_sign_operator_token1] = ACTIONS(1870),
    [sym_hash_operator] = ACTIONS(1872),
    [anon_sym_lsl] = ACTIONS(1870),
    [anon_sym_lsr] = ACTIONS(1870),
    [anon_sym_asr] = ACTIONS(1870),
    [aux_sym_pow_operator_token1] = ACTIONS(1872),
    [anon_sym_mod] = ACTIONS(1870),
    [anon_sym_land] = ACTIONS(1870),
    [anon_sym_lor] = ACTIONS(1870),
    [anon_sym_lxor] = ACTIONS(1870),
    [aux_sym_mult_operator_token1] = ACTIONS(1870),
    [aux_sym_add_operator_token1] = ACTIONS(1870),
    [sym_concat_operator] = ACTIONS(1872),
    [sym_rel_operator] = ACTIONS(1870),
    [anon_sym_AMP_AMP] = ACTIONS(1870),
    [anon_sym_or] = ACTIONS(1870),
    [anon_sym_PIPE_PIPE] = ACTIONS(1870),
    [sym_assign_operator] = ACTIONS(1872),
    [sym_let_operator] = ACTIONS(1872),
    [sym_let_and_operator] = ACTIONS(1872),
    [sym__uppercase_identifier] = ACTIONS(1872),
    [aux_sym_directive_token1] = ACTIONS(1870),
    [aux_sym_tag_token1] = ACTIONS(1872),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1009)] = {
    [sym_attribute] = STATE(1009),
    [sym__lowercase_identifier] = ACTIONS(1878),
    [anon_sym_SEMI_SEMI] = ACTIONS(1880),
    [anon_sym_let] = ACTIONS(1878),
    [anon_sym_and] = ACTIONS(1878),
    [anon_sym_TILDE] = ACTIONS(1878),
    [anon_sym_QMARK] = ACTIONS(1878),
    [anon_sym_LPAREN] = ACTIONS(1880),
    [anon_sym_external] = ACTIONS(1878),
    [anon_sym_type] = ACTIONS(1878),
    [anon_sym_COMMA] = ACTIONS(1880),
    [anon_sym_PIPE] = ACTIONS(1878),
    [anon_sym_DOT] = ACTIONS(1880),
    [anon_sym_LBRACK] = ACTIONS(1878),
    [anon_sym_RBRACK] = ACTIONS(1880),
    [anon_sym_true] = ACTIONS(1878),
    [anon_sym_false] = ACTIONS(1878),
    [anon_sym_LBRACE] = ACTIONS(1878),
    [anon_sym_SEMI] = ACTIONS(1878),
    [anon_sym_exception] = ACTIONS(1878),
    [anon_sym_module] = ACTIONS(1878),
    [anon_sym_open] = ACTIONS(1878),
    [anon_sym_include] = ACTIONS(1878),
    [anon_sym_class] = ACTIONS(1878),
    [anon_sym_val] = ACTIONS(1878),
    [anon_sym_end] = ACTIONS(1878),
    [anon_sym_object] = ACTIONS(1878),
    [anon_sym_in] = ACTIONS(1878),
    [anon_sym_AMP] = ACTIONS(1878),
    [anon_sym_POUND] = ACTIONS(1878),
    [anon_sym_COLON_COLON] = ACTIONS(1880),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1880),
    [anon_sym_else] = ACTIONS(1878),
    [anon_sym_new] = ACTIONS(1878),
    [anon_sym_LBRACE_LT] = ACTIONS(1880),
    [anon_sym_begin] = ACTIONS(1878),
    [sym_ocamlyacc_value] = ACTIONS(1880),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1878),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1880),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1878),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1878),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1880),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1880),
    [aux_sym_number_token1] = ACTIONS(1880),
    [aux_sym_signed_number_token1] = ACTIONS(1878),
    [anon_sym_SQUOTE] = ACTIONS(1880),
    [anon_sym_DQUOTE] = ACTIONS(1880),
    [sym_prefix_operator] = ACTIONS(1880),
    [aux_sym_sign_operator_token1] = ACTIONS(1878),
    [sym_hash_operator] = ACTIONS(1880),
    [anon_sym_lsl] = ACTIONS(1878),
    [anon_sym_lsr] = ACTIONS(1878),
    [anon_sym_asr] = ACTIONS(1878),
    [aux_sym_pow_operator_token1] = ACTIONS(1880),
    [anon_sym_mod] = ACTIONS(1878),
    [anon_sym_land] = ACTIONS(1878),
    [anon_sym_lor] = ACTIONS(1878),
    [anon_sym_lxor] = ACTIONS(1878),
    [aux_sym_mult_operator_token1] = ACTIONS(1878),
    [aux_sym_add_operator_token1] = ACTIONS(1878),
    [sym_concat_operator] = ACTIONS(1880),
    [sym_rel_operator] = ACTIONS(1878),
    [anon_sym_AMP_AMP] = ACTIONS(1878),
    [anon_sym_or] = ACTIONS(1878),
    [anon_sym_PIPE_PIPE] = ACTIONS(1878),
    [sym_assign_operator] = ACTIONS(1880),
    [sym_let_operator] = ACTIONS(1880),
    [sym_let_and_operator] = ACTIONS(1880),
    [sym__uppercase_identifier] = ACTIONS(1880),
    [aux_sym_directive_token1] = ACTIONS(1878),
    [aux_sym_tag_token1] = ACTIONS(1880),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1010)] = {
    [sym_attribute] = STATE(1010),
    [sym__lowercase_identifier] = ACTIONS(1882),
    [anon_sym_SEMI_SEMI] = ACTIONS(1884),
    [anon_sym_let] = ACTIONS(1882),
    [anon_sym_and] = ACTIONS(1882),
    [anon_sym_TILDE] = ACTIONS(1882),
    [anon_sym_QMARK] = ACTIONS(1882),
    [anon_sym_LPAREN] = ACTIONS(1884),
    [anon_sym_external] = ACTIONS(1882),
    [anon_sym_type] = ACTIONS(1882),
    [anon_sym_COMMA] = ACTIONS(1884),
    [anon_sym_PIPE] = ACTIONS(1882),
    [anon_sym_DOT] = ACTIONS(1884),
    [anon_sym_LBRACK] = ACTIONS(1882),
    [anon_sym_RBRACK] = ACTIONS(1884),
    [anon_sym_true] = ACTIONS(1882),
    [anon_sym_false] = ACTIONS(1882),
    [anon_sym_LBRACE] = ACTIONS(1882),
    [anon_sym_SEMI] = ACTIONS(1882),
    [anon_sym_exception] = ACTIONS(1882),
    [anon_sym_module] = ACTIONS(1882),
    [anon_sym_open] = ACTIONS(1882),
    [anon_sym_include] = ACTIONS(1882),
    [anon_sym_class] = ACTIONS(1882),
    [anon_sym_val] = ACTIONS(1882),
    [anon_sym_end] = ACTIONS(1882),
    [anon_sym_object] = ACTIONS(1882),
    [anon_sym_in] = ACTIONS(1882),
    [anon_sym_AMP] = ACTIONS(1882),
    [anon_sym_POUND] = ACTIONS(1882),
    [anon_sym_COLON_COLON] = ACTIONS(1884),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1884),
    [anon_sym_else] = ACTIONS(1882),
    [anon_sym_new] = ACTIONS(1882),
    [anon_sym_LBRACE_LT] = ACTIONS(1884),
    [anon_sym_begin] = ACTIONS(1882),
    [sym_ocamlyacc_value] = ACTIONS(1884),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1882),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1884),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1882),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1882),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1884),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1884),
    [aux_sym_number_token1] = ACTIONS(1884),
    [aux_sym_signed_number_token1] = ACTIONS(1882),
    [anon_sym_SQUOTE] = ACTIONS(1884),
    [anon_sym_DQUOTE] = ACTIONS(1884),
    [sym_prefix_operator] = ACTIONS(1884),
    [aux_sym_sign_operator_token1] = ACTIONS(1882),
    [sym_hash_operator] = ACTIONS(1884),
    [anon_sym_lsl] = ACTIONS(1882),
    [anon_sym_lsr] = ACTIONS(1882),
    [anon_sym_asr] = ACTIONS(1882),
    [aux_sym_pow_operator_token1] = ACTIONS(1884),
    [anon_sym_mod] = ACTIONS(1882),
    [anon_sym_land] = ACTIONS(1882),
    [anon_sym_lor] = ACTIONS(1882),
    [anon_sym_lxor] = ACTIONS(1882),
    [aux_sym_mult_operator_token1] = ACTIONS(1882),
    [aux_sym_add_operator_token1] = ACTIONS(1882),
    [sym_concat_operator] = ACTIONS(1884),
    [sym_rel_operator] = ACTIONS(1882),
    [anon_sym_AMP_AMP] = ACTIONS(1882),
    [anon_sym_or] = ACTIONS(1882),
    [anon_sym_PIPE_PIPE] = ACTIONS(1882),
    [sym_assign_operator] = ACTIONS(1884),
    [sym_let_operator] = ACTIONS(1884),
    [sym_let_and_operator] = ACTIONS(1884),
    [sym__uppercase_identifier] = ACTIONS(1884),
    [aux_sym_directive_token1] = ACTIONS(1882),
    [aux_sym_tag_token1] = ACTIONS(1884),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1011)] = {
    [sym_attribute] = STATE(1011),
    [sym__lowercase_identifier] = ACTIONS(1934),
    [anon_sym_SEMI_SEMI] = ACTIONS(1936),
    [anon_sym_let] = ACTIONS(1934),
    [anon_sym_and] = ACTIONS(1934),
    [anon_sym_TILDE] = ACTIONS(1934),
    [anon_sym_QMARK] = ACTIONS(1934),
    [anon_sym_LPAREN] = ACTIONS(1936),
    [anon_sym_external] = ACTIONS(1934),
    [anon_sym_type] = ACTIONS(1934),
    [anon_sym_COMMA] = ACTIONS(1936),
    [anon_sym_PIPE] = ACTIONS(1934),
    [anon_sym_DOT] = ACTIONS(1936),
    [anon_sym_LBRACK] = ACTIONS(1934),
    [anon_sym_RBRACK] = ACTIONS(1936),
    [anon_sym_true] = ACTIONS(1934),
    [anon_sym_false] = ACTIONS(1934),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_SEMI] = ACTIONS(1934),
    [anon_sym_exception] = ACTIONS(1934),
    [anon_sym_module] = ACTIONS(1934),
    [anon_sym_open] = ACTIONS(1934),
    [anon_sym_include] = ACTIONS(1934),
    [anon_sym_class] = ACTIONS(1934),
    [anon_sym_val] = ACTIONS(1934),
    [anon_sym_end] = ACTIONS(1934),
    [anon_sym_object] = ACTIONS(1934),
    [anon_sym_in] = ACTIONS(1934),
    [anon_sym_AMP] = ACTIONS(1934),
    [anon_sym_POUND] = ACTIONS(1934),
    [anon_sym_COLON_COLON] = ACTIONS(1936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1936),
    [anon_sym_else] = ACTIONS(1934),
    [anon_sym_new] = ACTIONS(1934),
    [anon_sym_LBRACE_LT] = ACTIONS(1936),
    [anon_sym_begin] = ACTIONS(1934),
    [sym_ocamlyacc_value] = ACTIONS(1936),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1934),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1934),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1934),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1936),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1936),
    [aux_sym_number_token1] = ACTIONS(1936),
    [aux_sym_signed_number_token1] = ACTIONS(1934),
    [anon_sym_SQUOTE] = ACTIONS(1936),
    [anon_sym_DQUOTE] = ACTIONS(1936),
    [sym_prefix_operator] = ACTIONS(1936),
    [aux_sym_sign_operator_token1] = ACTIONS(1934),
    [sym_hash_operator] = ACTIONS(1936),
    [anon_sym_lsl] = ACTIONS(1934),
    [anon_sym_lsr] = ACTIONS(1934),
    [anon_sym_asr] = ACTIONS(1934),
    [aux_sym_pow_operator_token1] = ACTIONS(1936),
    [anon_sym_mod] = ACTIONS(1934),
    [anon_sym_land] = ACTIONS(1934),
    [anon_sym_lor] = ACTIONS(1934),
    [anon_sym_lxor] = ACTIONS(1934),
    [aux_sym_mult_operator_token1] = ACTIONS(1934),
    [aux_sym_add_operator_token1] = ACTIONS(1934),
    [sym_concat_operator] = ACTIONS(1936),
    [sym_rel_operator] = ACTIONS(1934),
    [anon_sym_AMP_AMP] = ACTIONS(1934),
    [anon_sym_or] = ACTIONS(1934),
    [anon_sym_PIPE_PIPE] = ACTIONS(1934),
    [sym_assign_operator] = ACTIONS(1936),
    [sym_let_operator] = ACTIONS(1936),
    [sym_let_and_operator] = ACTIONS(1936),
    [sym__uppercase_identifier] = ACTIONS(1936),
    [aux_sym_directive_token1] = ACTIONS(1934),
    [aux_sym_tag_token1] = ACTIONS(1936),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1012)] = {
    [sym_parameter] = STATE(2475),
    [sym__parenthesized_abstract_type] = STATE(2475),
    [sym__delimited_pattern] = STATE(2450),
    [sym__simple_pattern] = STATE(2476),
    [sym_typed_pattern] = STATE(2450),
    [sym_polymorphic_variant_pattern] = STATE(2450),
    [sym_record_pattern] = STATE(2478),
    [sym_list_pattern] = STATE(2478),
    [sym_array_pattern] = STATE(2478),
    [sym_local_open_pattern] = STATE(2450),
    [sym_package_pattern] = STATE(2450),
    [sym_parenthesized_pattern] = STATE(2478),
    [sym_attribute] = STATE(1012),
    [sym_extension] = STATE(2450),
    [sym_quoted_extension] = STATE(2450),
    [sym__constant] = STATE(2464),
    [sym__signed_constant] = STATE(2450),
    [sym_number] = STATE(2504),
    [sym_signed_number] = STATE(2464),
    [sym_character] = STATE(2504),
    [sym_string] = STATE(2504),
    [sym_quoted_string] = STATE(2504),
    [sym_boolean] = STATE(2504),
    [sym_unit] = STATE(2478),
    [sym__simple_value_pattern] = STATE(2450),
    [sym_parenthesized_operator] = STATE(2450),
    [sym_module_path] = STATE(6504),
    [sym_constructor_path] = STATE(2450),
    [sym__constructor_name] = STATE(2515),
    [sym_tag] = STATE(2450),
    [aux_sym_let_binding_repeat1] = STATE(1012),
    [sym__lowercase_identifier] = ACTIONS(2181),
    [anon_sym_SEMI_SEMI] = ACTIONS(1954),
    [anon_sym_let] = ACTIONS(1959),
    [anon_sym_and] = ACTIONS(1959),
    [anon_sym_EQ] = ACTIONS(1954),
    [anon_sym_TILDE] = ACTIONS(2184),
    [anon_sym_QMARK] = ACTIONS(2184),
    [anon_sym_LPAREN] = ACTIONS(2187),
    [anon_sym_external] = ACTIONS(1959),
    [anon_sym_type] = ACTIONS(1959),
    [anon_sym_COLON2] = ACTIONS(1954),
    [anon_sym_LBRACK] = ACTIONS(2190),
    [anon_sym_RBRACK] = ACTIONS(1954),
    [anon_sym_true] = ACTIONS(2193),
    [anon_sym_false] = ACTIONS(2193),
    [anon_sym_LBRACE] = ACTIONS(2196),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_module] = ACTIONS(1959),
    [anon_sym_open] = ACTIONS(1959),
    [anon_sym_include] = ACTIONS(1959),
    [anon_sym_class] = ACTIONS(1959),
    [anon_sym_val] = ACTIONS(1959),
    [anon_sym_end] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(2199),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2202),
    [anon_sym_begin] = ACTIONS(2205),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1959),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1954),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2208),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2211),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1954),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1954),
    [aux_sym_number_token1] = ACTIONS(2214),
    [aux_sym_signed_number_token1] = ACTIONS(2217),
    [anon_sym_SQUOTE] = ACTIONS(2220),
    [anon_sym_DQUOTE] = ACTIONS(2223),
    [sym_let_operator] = ACTIONS(1954),
    [sym__uppercase_identifier] = ACTIONS(2226),
    [aux_sym_directive_token1] = ACTIONS(1959),
    [aux_sym_tag_token1] = ACTIONS(2229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1013)] = {
    [sym_attribute] = STATE(1013),
    [sym__lowercase_identifier] = ACTIONS(1698),
    [anon_sym_SEMI_SEMI] = ACTIONS(1700),
    [anon_sym_let] = ACTIONS(1698),
    [anon_sym_and] = ACTIONS(1698),
    [anon_sym_TILDE] = ACTIONS(1698),
    [anon_sym_QMARK] = ACTIONS(1698),
    [anon_sym_LPAREN] = ACTIONS(1700),
    [anon_sym_external] = ACTIONS(1698),
    [anon_sym_type] = ACTIONS(1698),
    [anon_sym_COMMA] = ACTIONS(1700),
    [anon_sym_PIPE] = ACTIONS(1698),
    [anon_sym_DOT] = ACTIONS(1702),
    [anon_sym_LBRACK] = ACTIONS(1698),
    [anon_sym_RBRACK] = ACTIONS(1700),
    [anon_sym_true] = ACTIONS(1698),
    [anon_sym_false] = ACTIONS(1698),
    [anon_sym_LBRACE] = ACTIONS(1698),
    [anon_sym_SEMI] = ACTIONS(1698),
    [anon_sym_exception] = ACTIONS(1698),
    [anon_sym_module] = ACTIONS(1698),
    [anon_sym_open] = ACTIONS(1698),
    [anon_sym_include] = ACTIONS(1698),
    [anon_sym_class] = ACTIONS(1698),
    [anon_sym_val] = ACTIONS(1698),
    [anon_sym_end] = ACTIONS(1698),
    [anon_sym_object] = ACTIONS(1698),
    [anon_sym_in] = ACTIONS(1698),
    [anon_sym_AMP] = ACTIONS(1698),
    [anon_sym_POUND] = ACTIONS(1698),
    [anon_sym_COLON_COLON] = ACTIONS(1700),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1700),
    [anon_sym_else] = ACTIONS(1698),
    [anon_sym_new] = ACTIONS(1698),
    [anon_sym_LBRACE_LT] = ACTIONS(1700),
    [anon_sym_begin] = ACTIONS(1698),
    [sym_ocamlyacc_value] = ACTIONS(1700),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1698),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1700),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1700),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1700),
    [aux_sym_number_token1] = ACTIONS(1700),
    [aux_sym_signed_number_token1] = ACTIONS(1698),
    [anon_sym_SQUOTE] = ACTIONS(1700),
    [anon_sym_DQUOTE] = ACTIONS(1700),
    [sym_prefix_operator] = ACTIONS(1700),
    [aux_sym_sign_operator_token1] = ACTIONS(1698),
    [sym_hash_operator] = ACTIONS(1700),
    [anon_sym_lsl] = ACTIONS(1698),
    [anon_sym_lsr] = ACTIONS(1698),
    [anon_sym_asr] = ACTIONS(1698),
    [aux_sym_pow_operator_token1] = ACTIONS(1700),
    [anon_sym_mod] = ACTIONS(1698),
    [anon_sym_land] = ACTIONS(1698),
    [anon_sym_lor] = ACTIONS(1698),
    [anon_sym_lxor] = ACTIONS(1698),
    [aux_sym_mult_operator_token1] = ACTIONS(1698),
    [aux_sym_add_operator_token1] = ACTIONS(1698),
    [sym_concat_operator] = ACTIONS(1700),
    [sym_rel_operator] = ACTIONS(1698),
    [anon_sym_AMP_AMP] = ACTIONS(1698),
    [anon_sym_or] = ACTIONS(1698),
    [anon_sym_PIPE_PIPE] = ACTIONS(1698),
    [sym_assign_operator] = ACTIONS(1700),
    [sym_let_operator] = ACTIONS(1700),
    [sym_let_and_operator] = ACTIONS(1700),
    [sym__uppercase_identifier] = ACTIONS(1700),
    [aux_sym_directive_token1] = ACTIONS(1698),
    [aux_sym_tag_token1] = ACTIONS(1700),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1014)] = {
    [sym_attribute] = STATE(1014),
    [sym__lowercase_identifier] = ACTIONS(1890),
    [anon_sym_SEMI_SEMI] = ACTIONS(1892),
    [anon_sym_let] = ACTIONS(1890),
    [anon_sym_and] = ACTIONS(1890),
    [anon_sym_TILDE] = ACTIONS(1890),
    [anon_sym_QMARK] = ACTIONS(1890),
    [anon_sym_LPAREN] = ACTIONS(1892),
    [anon_sym_external] = ACTIONS(1890),
    [anon_sym_type] = ACTIONS(1890),
    [anon_sym_COMMA] = ACTIONS(1892),
    [anon_sym_PIPE] = ACTIONS(1890),
    [anon_sym_DOT] = ACTIONS(1892),
    [anon_sym_LBRACK] = ACTIONS(1890),
    [anon_sym_RBRACK] = ACTIONS(1892),
    [anon_sym_true] = ACTIONS(1890),
    [anon_sym_false] = ACTIONS(1890),
    [anon_sym_LBRACE] = ACTIONS(1890),
    [anon_sym_SEMI] = ACTIONS(1890),
    [anon_sym_exception] = ACTIONS(1890),
    [anon_sym_module] = ACTIONS(1890),
    [anon_sym_open] = ACTIONS(1890),
    [anon_sym_include] = ACTIONS(1890),
    [anon_sym_class] = ACTIONS(1890),
    [anon_sym_val] = ACTIONS(1890),
    [anon_sym_end] = ACTIONS(1890),
    [anon_sym_object] = ACTIONS(1890),
    [anon_sym_in] = ACTIONS(1890),
    [anon_sym_AMP] = ACTIONS(1890),
    [anon_sym_POUND] = ACTIONS(1890),
    [anon_sym_COLON_COLON] = ACTIONS(1892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1892),
    [anon_sym_else] = ACTIONS(1890),
    [anon_sym_new] = ACTIONS(1890),
    [anon_sym_LBRACE_LT] = ACTIONS(1892),
    [anon_sym_begin] = ACTIONS(1890),
    [sym_ocamlyacc_value] = ACTIONS(1892),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1890),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1892),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1890),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1890),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1892),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1892),
    [aux_sym_number_token1] = ACTIONS(1892),
    [aux_sym_signed_number_token1] = ACTIONS(1890),
    [anon_sym_SQUOTE] = ACTIONS(1892),
    [anon_sym_DQUOTE] = ACTIONS(1892),
    [sym_prefix_operator] = ACTIONS(1892),
    [aux_sym_sign_operator_token1] = ACTIONS(1890),
    [sym_hash_operator] = ACTIONS(1892),
    [anon_sym_lsl] = ACTIONS(1890),
    [anon_sym_lsr] = ACTIONS(1890),
    [anon_sym_asr] = ACTIONS(1890),
    [aux_sym_pow_operator_token1] = ACTIONS(1892),
    [anon_sym_mod] = ACTIONS(1890),
    [anon_sym_land] = ACTIONS(1890),
    [anon_sym_lor] = ACTIONS(1890),
    [anon_sym_lxor] = ACTIONS(1890),
    [aux_sym_mult_operator_token1] = ACTIONS(1890),
    [aux_sym_add_operator_token1] = ACTIONS(1890),
    [sym_concat_operator] = ACTIONS(1892),
    [sym_rel_operator] = ACTIONS(1890),
    [anon_sym_AMP_AMP] = ACTIONS(1890),
    [anon_sym_or] = ACTIONS(1890),
    [anon_sym_PIPE_PIPE] = ACTIONS(1890),
    [sym_assign_operator] = ACTIONS(1892),
    [sym_let_operator] = ACTIONS(1892),
    [sym_let_and_operator] = ACTIONS(1892),
    [sym__uppercase_identifier] = ACTIONS(1892),
    [aux_sym_directive_token1] = ACTIONS(1890),
    [aux_sym_tag_token1] = ACTIONS(1892),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1015)] = {
    [sym_attribute] = STATE(1015),
    [sym__lowercase_identifier] = ACTIONS(1894),
    [anon_sym_SEMI_SEMI] = ACTIONS(1896),
    [anon_sym_let] = ACTIONS(1894),
    [anon_sym_and] = ACTIONS(1894),
    [anon_sym_TILDE] = ACTIONS(1894),
    [anon_sym_QMARK] = ACTIONS(1894),
    [anon_sym_LPAREN] = ACTIONS(1896),
    [anon_sym_external] = ACTIONS(1894),
    [anon_sym_type] = ACTIONS(1894),
    [anon_sym_COMMA] = ACTIONS(1896),
    [anon_sym_PIPE] = ACTIONS(1894),
    [anon_sym_DOT] = ACTIONS(765),
    [anon_sym_LBRACK] = ACTIONS(1894),
    [anon_sym_RBRACK] = ACTIONS(1896),
    [anon_sym_true] = ACTIONS(1894),
    [anon_sym_false] = ACTIONS(1894),
    [anon_sym_LBRACE] = ACTIONS(1894),
    [anon_sym_SEMI] = ACTIONS(1894),
    [anon_sym_exception] = ACTIONS(1894),
    [anon_sym_module] = ACTIONS(1894),
    [anon_sym_open] = ACTIONS(1894),
    [anon_sym_include] = ACTIONS(1894),
    [anon_sym_class] = ACTIONS(1894),
    [anon_sym_val] = ACTIONS(1894),
    [anon_sym_end] = ACTIONS(1894),
    [anon_sym_object] = ACTIONS(1894),
    [anon_sym_in] = ACTIONS(1894),
    [anon_sym_AMP] = ACTIONS(1894),
    [anon_sym_POUND] = ACTIONS(1894),
    [anon_sym_COLON_COLON] = ACTIONS(1896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1896),
    [anon_sym_else] = ACTIONS(1894),
    [anon_sym_new] = ACTIONS(1894),
    [anon_sym_LBRACE_LT] = ACTIONS(1896),
    [anon_sym_begin] = ACTIONS(1894),
    [sym_ocamlyacc_value] = ACTIONS(1896),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1894),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1896),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1894),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1894),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1896),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1896),
    [aux_sym_number_token1] = ACTIONS(1896),
    [aux_sym_signed_number_token1] = ACTIONS(1894),
    [anon_sym_SQUOTE] = ACTIONS(1896),
    [anon_sym_DQUOTE] = ACTIONS(1896),
    [sym_prefix_operator] = ACTIONS(1896),
    [aux_sym_sign_operator_token1] = ACTIONS(1894),
    [sym_hash_operator] = ACTIONS(1896),
    [anon_sym_lsl] = ACTIONS(1894),
    [anon_sym_lsr] = ACTIONS(1894),
    [anon_sym_asr] = ACTIONS(1894),
    [aux_sym_pow_operator_token1] = ACTIONS(1896),
    [anon_sym_mod] = ACTIONS(1894),
    [anon_sym_land] = ACTIONS(1894),
    [anon_sym_lor] = ACTIONS(1894),
    [anon_sym_lxor] = ACTIONS(1894),
    [aux_sym_mult_operator_token1] = ACTIONS(1894),
    [aux_sym_add_operator_token1] = ACTIONS(1894),
    [sym_concat_operator] = ACTIONS(1896),
    [sym_rel_operator] = ACTIONS(1894),
    [anon_sym_AMP_AMP] = ACTIONS(1894),
    [anon_sym_or] = ACTIONS(1894),
    [anon_sym_PIPE_PIPE] = ACTIONS(1894),
    [sym_assign_operator] = ACTIONS(1896),
    [sym_let_operator] = ACTIONS(1896),
    [sym_let_and_operator] = ACTIONS(1896),
    [sym__uppercase_identifier] = ACTIONS(1896),
    [aux_sym_directive_token1] = ACTIONS(1894),
    [aux_sym_tag_token1] = ACTIONS(1896),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1016)] = {
    [sym_attribute] = STATE(1016),
    [sym__lowercase_identifier] = ACTIONS(1910),
    [anon_sym_SEMI_SEMI] = ACTIONS(1912),
    [anon_sym_let] = ACTIONS(1910),
    [anon_sym_and] = ACTIONS(1910),
    [anon_sym_TILDE] = ACTIONS(1910),
    [anon_sym_QMARK] = ACTIONS(1910),
    [anon_sym_LPAREN] = ACTIONS(1912),
    [anon_sym_external] = ACTIONS(1910),
    [anon_sym_type] = ACTIONS(1910),
    [anon_sym_COMMA] = ACTIONS(1912),
    [anon_sym_PIPE] = ACTIONS(1910),
    [anon_sym_DOT] = ACTIONS(1912),
    [anon_sym_LBRACK] = ACTIONS(1910),
    [anon_sym_RBRACK] = ACTIONS(1912),
    [anon_sym_true] = ACTIONS(1910),
    [anon_sym_false] = ACTIONS(1910),
    [anon_sym_LBRACE] = ACTIONS(1910),
    [anon_sym_SEMI] = ACTIONS(1910),
    [anon_sym_exception] = ACTIONS(1910),
    [anon_sym_module] = ACTIONS(1910),
    [anon_sym_open] = ACTIONS(1910),
    [anon_sym_include] = ACTIONS(1910),
    [anon_sym_class] = ACTIONS(1910),
    [anon_sym_val] = ACTIONS(1910),
    [anon_sym_end] = ACTIONS(1910),
    [anon_sym_object] = ACTIONS(1910),
    [anon_sym_in] = ACTIONS(1910),
    [anon_sym_AMP] = ACTIONS(1910),
    [anon_sym_POUND] = ACTIONS(1910),
    [anon_sym_COLON_COLON] = ACTIONS(1912),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1912),
    [anon_sym_else] = ACTIONS(1910),
    [anon_sym_new] = ACTIONS(1910),
    [anon_sym_LBRACE_LT] = ACTIONS(1912),
    [anon_sym_begin] = ACTIONS(1910),
    [sym_ocamlyacc_value] = ACTIONS(1912),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1910),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1912),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1910),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1910),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1912),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1912),
    [aux_sym_number_token1] = ACTIONS(1912),
    [aux_sym_signed_number_token1] = ACTIONS(1910),
    [anon_sym_SQUOTE] = ACTIONS(1912),
    [anon_sym_DQUOTE] = ACTIONS(1912),
    [sym_prefix_operator] = ACTIONS(1912),
    [aux_sym_sign_operator_token1] = ACTIONS(1910),
    [sym_hash_operator] = ACTIONS(1912),
    [anon_sym_lsl] = ACTIONS(1910),
    [anon_sym_lsr] = ACTIONS(1910),
    [anon_sym_asr] = ACTIONS(1910),
    [aux_sym_pow_operator_token1] = ACTIONS(1912),
    [anon_sym_mod] = ACTIONS(1910),
    [anon_sym_land] = ACTIONS(1910),
    [anon_sym_lor] = ACTIONS(1910),
    [anon_sym_lxor] = ACTIONS(1910),
    [aux_sym_mult_operator_token1] = ACTIONS(1910),
    [aux_sym_add_operator_token1] = ACTIONS(1910),
    [sym_concat_operator] = ACTIONS(1912),
    [sym_rel_operator] = ACTIONS(1910),
    [anon_sym_AMP_AMP] = ACTIONS(1910),
    [anon_sym_or] = ACTIONS(1910),
    [anon_sym_PIPE_PIPE] = ACTIONS(1910),
    [sym_assign_operator] = ACTIONS(1912),
    [sym_let_operator] = ACTIONS(1912),
    [sym_let_and_operator] = ACTIONS(1912),
    [sym__uppercase_identifier] = ACTIONS(1912),
    [aux_sym_directive_token1] = ACTIONS(1910),
    [aux_sym_tag_token1] = ACTIONS(1912),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1017)] = {
    [sym_attribute] = STATE(1017),
    [sym__lowercase_identifier] = ACTIONS(1642),
    [anon_sym_SEMI_SEMI] = ACTIONS(1644),
    [anon_sym_let] = ACTIONS(1642),
    [anon_sym_and] = ACTIONS(1642),
    [anon_sym_TILDE] = ACTIONS(1642),
    [anon_sym_QMARK] = ACTIONS(1642),
    [anon_sym_LPAREN] = ACTIONS(1644),
    [anon_sym_external] = ACTIONS(1642),
    [anon_sym_type] = ACTIONS(1642),
    [anon_sym_COMMA] = ACTIONS(1644),
    [anon_sym_PIPE] = ACTIONS(1642),
    [anon_sym_DOT] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(1642),
    [anon_sym_RBRACK] = ACTIONS(1644),
    [anon_sym_true] = ACTIONS(1642),
    [anon_sym_false] = ACTIONS(1642),
    [anon_sym_LBRACE] = ACTIONS(1642),
    [anon_sym_SEMI] = ACTIONS(1642),
    [anon_sym_exception] = ACTIONS(1642),
    [anon_sym_module] = ACTIONS(1642),
    [anon_sym_open] = ACTIONS(1642),
    [anon_sym_include] = ACTIONS(1642),
    [anon_sym_class] = ACTIONS(1642),
    [anon_sym_val] = ACTIONS(1642),
    [anon_sym_end] = ACTIONS(1642),
    [anon_sym_object] = ACTIONS(1642),
    [anon_sym_in] = ACTIONS(1642),
    [anon_sym_AMP] = ACTIONS(1642),
    [anon_sym_POUND] = ACTIONS(1642),
    [anon_sym_COLON_COLON] = ACTIONS(1644),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1644),
    [anon_sym_else] = ACTIONS(1642),
    [anon_sym_new] = ACTIONS(1642),
    [anon_sym_LBRACE_LT] = ACTIONS(1644),
    [anon_sym_begin] = ACTIONS(1642),
    [sym_ocamlyacc_value] = ACTIONS(1644),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1642),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1644),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1642),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1642),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1644),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1644),
    [aux_sym_number_token1] = ACTIONS(1644),
    [aux_sym_signed_number_token1] = ACTIONS(1642),
    [anon_sym_SQUOTE] = ACTIONS(1644),
    [anon_sym_DQUOTE] = ACTIONS(1644),
    [sym_prefix_operator] = ACTIONS(1644),
    [aux_sym_sign_operator_token1] = ACTIONS(1642),
    [sym_hash_operator] = ACTIONS(1644),
    [anon_sym_lsl] = ACTIONS(1642),
    [anon_sym_lsr] = ACTIONS(1642),
    [anon_sym_asr] = ACTIONS(1642),
    [aux_sym_pow_operator_token1] = ACTIONS(1644),
    [anon_sym_mod] = ACTIONS(1642),
    [anon_sym_land] = ACTIONS(1642),
    [anon_sym_lor] = ACTIONS(1642),
    [anon_sym_lxor] = ACTIONS(1642),
    [aux_sym_mult_operator_token1] = ACTIONS(1642),
    [aux_sym_add_operator_token1] = ACTIONS(1642),
    [sym_concat_operator] = ACTIONS(1644),
    [sym_rel_operator] = ACTIONS(1642),
    [anon_sym_AMP_AMP] = ACTIONS(1642),
    [anon_sym_or] = ACTIONS(1642),
    [anon_sym_PIPE_PIPE] = ACTIONS(1642),
    [sym_assign_operator] = ACTIONS(1644),
    [sym_let_operator] = ACTIONS(1644),
    [sym_let_and_operator] = ACTIONS(1644),
    [sym__uppercase_identifier] = ACTIONS(1644),
    [aux_sym_directive_token1] = ACTIONS(1642),
    [aux_sym_tag_token1] = ACTIONS(1644),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1018)] = {
    [sym_attribute] = STATE(1018),
    [sym__lowercase_identifier] = ACTIONS(1854),
    [anon_sym_SEMI_SEMI] = ACTIONS(1856),
    [anon_sym_let] = ACTIONS(1854),
    [anon_sym_and] = ACTIONS(1854),
    [anon_sym_TILDE] = ACTIONS(1854),
    [anon_sym_QMARK] = ACTIONS(1854),
    [anon_sym_LPAREN] = ACTIONS(1856),
    [anon_sym_external] = ACTIONS(1854),
    [anon_sym_type] = ACTIONS(1854),
    [anon_sym_COMMA] = ACTIONS(1856),
    [anon_sym_PIPE] = ACTIONS(1854),
    [anon_sym_DOT] = ACTIONS(1856),
    [anon_sym_LBRACK] = ACTIONS(1854),
    [anon_sym_RBRACK] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(1854),
    [anon_sym_false] = ACTIONS(1854),
    [anon_sym_LBRACE] = ACTIONS(1854),
    [anon_sym_SEMI] = ACTIONS(1854),
    [anon_sym_exception] = ACTIONS(1854),
    [anon_sym_module] = ACTIONS(1854),
    [anon_sym_open] = ACTIONS(1854),
    [anon_sym_include] = ACTIONS(1854),
    [anon_sym_class] = ACTIONS(1854),
    [anon_sym_val] = ACTIONS(1854),
    [anon_sym_end] = ACTIONS(1854),
    [anon_sym_object] = ACTIONS(1854),
    [anon_sym_in] = ACTIONS(1854),
    [anon_sym_AMP] = ACTIONS(1854),
    [anon_sym_POUND] = ACTIONS(1854),
    [anon_sym_COLON_COLON] = ACTIONS(1856),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1856),
    [anon_sym_else] = ACTIONS(1854),
    [anon_sym_new] = ACTIONS(1854),
    [anon_sym_LBRACE_LT] = ACTIONS(1856),
    [anon_sym_begin] = ACTIONS(1854),
    [sym_ocamlyacc_value] = ACTIONS(1856),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1854),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1856),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1854),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1854),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1856),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1856),
    [aux_sym_number_token1] = ACTIONS(1856),
    [aux_sym_signed_number_token1] = ACTIONS(1854),
    [anon_sym_SQUOTE] = ACTIONS(1856),
    [anon_sym_DQUOTE] = ACTIONS(1856),
    [sym_prefix_operator] = ACTIONS(1856),
    [aux_sym_sign_operator_token1] = ACTIONS(1854),
    [sym_hash_operator] = ACTIONS(1856),
    [anon_sym_lsl] = ACTIONS(1854),
    [anon_sym_lsr] = ACTIONS(1854),
    [anon_sym_asr] = ACTIONS(1854),
    [aux_sym_pow_operator_token1] = ACTIONS(1856),
    [anon_sym_mod] = ACTIONS(1854),
    [anon_sym_land] = ACTIONS(1854),
    [anon_sym_lor] = ACTIONS(1854),
    [anon_sym_lxor] = ACTIONS(1854),
    [aux_sym_mult_operator_token1] = ACTIONS(1854),
    [aux_sym_add_operator_token1] = ACTIONS(1854),
    [sym_concat_operator] = ACTIONS(1856),
    [sym_rel_operator] = ACTIONS(1854),
    [anon_sym_AMP_AMP] = ACTIONS(1854),
    [anon_sym_or] = ACTIONS(1854),
    [anon_sym_PIPE_PIPE] = ACTIONS(1854),
    [sym_assign_operator] = ACTIONS(1856),
    [sym_let_operator] = ACTIONS(1856),
    [sym_let_and_operator] = ACTIONS(1856),
    [sym__uppercase_identifier] = ACTIONS(1856),
    [aux_sym_directive_token1] = ACTIONS(1854),
    [aux_sym_tag_token1] = ACTIONS(1856),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1019)] = {
    [sym_attribute] = STATE(1019),
    [sym__lowercase_identifier] = ACTIONS(1650),
    [anon_sym_SEMI_SEMI] = ACTIONS(1652),
    [anon_sym_let] = ACTIONS(1650),
    [anon_sym_and] = ACTIONS(1650),
    [anon_sym_TILDE] = ACTIONS(1650),
    [anon_sym_QMARK] = ACTIONS(1650),
    [anon_sym_LPAREN] = ACTIONS(1652),
    [anon_sym_external] = ACTIONS(1650),
    [anon_sym_type] = ACTIONS(1650),
    [anon_sym_COMMA] = ACTIONS(1652),
    [anon_sym_PIPE] = ACTIONS(1650),
    [anon_sym_DOT] = ACTIONS(1652),
    [anon_sym_LBRACK] = ACTIONS(1650),
    [anon_sym_RBRACK] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1650),
    [anon_sym_false] = ACTIONS(1650),
    [anon_sym_LBRACE] = ACTIONS(1650),
    [anon_sym_SEMI] = ACTIONS(1650),
    [anon_sym_exception] = ACTIONS(1650),
    [anon_sym_module] = ACTIONS(1650),
    [anon_sym_open] = ACTIONS(1650),
    [anon_sym_include] = ACTIONS(1650),
    [anon_sym_class] = ACTIONS(1650),
    [anon_sym_val] = ACTIONS(1650),
    [anon_sym_end] = ACTIONS(1650),
    [anon_sym_object] = ACTIONS(1650),
    [anon_sym_in] = ACTIONS(1650),
    [anon_sym_AMP] = ACTIONS(1650),
    [anon_sym_POUND] = ACTIONS(1650),
    [anon_sym_COLON_COLON] = ACTIONS(1652),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1652),
    [anon_sym_else] = ACTIONS(1650),
    [anon_sym_new] = ACTIONS(1650),
    [anon_sym_LBRACE_LT] = ACTIONS(1652),
    [anon_sym_begin] = ACTIONS(1650),
    [sym_ocamlyacc_value] = ACTIONS(1652),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1650),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1652),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1650),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1650),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1652),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1652),
    [aux_sym_number_token1] = ACTIONS(1652),
    [aux_sym_signed_number_token1] = ACTIONS(1650),
    [anon_sym_SQUOTE] = ACTIONS(1652),
    [anon_sym_DQUOTE] = ACTIONS(1652),
    [sym_prefix_operator] = ACTIONS(1652),
    [aux_sym_sign_operator_token1] = ACTIONS(1650),
    [sym_hash_operator] = ACTIONS(1652),
    [anon_sym_lsl] = ACTIONS(1650),
    [anon_sym_lsr] = ACTIONS(1650),
    [anon_sym_asr] = ACTIONS(1650),
    [aux_sym_pow_operator_token1] = ACTIONS(1652),
    [anon_sym_mod] = ACTIONS(1650),
    [anon_sym_land] = ACTIONS(1650),
    [anon_sym_lor] = ACTIONS(1650),
    [anon_sym_lxor] = ACTIONS(1650),
    [aux_sym_mult_operator_token1] = ACTIONS(1650),
    [aux_sym_add_operator_token1] = ACTIONS(1650),
    [sym_concat_operator] = ACTIONS(1652),
    [sym_rel_operator] = ACTIONS(1650),
    [anon_sym_AMP_AMP] = ACTIONS(1650),
    [anon_sym_or] = ACTIONS(1650),
    [anon_sym_PIPE_PIPE] = ACTIONS(1650),
    [sym_assign_operator] = ACTIONS(1652),
    [sym_let_operator] = ACTIONS(1652),
    [sym_let_and_operator] = ACTIONS(1652),
    [sym__uppercase_identifier] = ACTIONS(1652),
    [aux_sym_directive_token1] = ACTIONS(1650),
    [aux_sym_tag_token1] = ACTIONS(1652),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1020)] = {
    [sym_attribute] = STATE(1020),
    [sym__lowercase_identifier] = ACTIONS(1938),
    [anon_sym_SEMI_SEMI] = ACTIONS(1940),
    [anon_sym_let] = ACTIONS(1938),
    [anon_sym_and] = ACTIONS(1938),
    [anon_sym_TILDE] = ACTIONS(1938),
    [anon_sym_QMARK] = ACTIONS(1938),
    [anon_sym_LPAREN] = ACTIONS(1940),
    [anon_sym_external] = ACTIONS(1938),
    [anon_sym_type] = ACTIONS(1938),
    [anon_sym_COMMA] = ACTIONS(1940),
    [anon_sym_PIPE] = ACTIONS(1938),
    [anon_sym_DOT] = ACTIONS(1940),
    [anon_sym_LBRACK] = ACTIONS(1938),
    [anon_sym_RBRACK] = ACTIONS(1940),
    [anon_sym_true] = ACTIONS(1938),
    [anon_sym_false] = ACTIONS(1938),
    [anon_sym_LBRACE] = ACTIONS(1938),
    [anon_sym_SEMI] = ACTIONS(1938),
    [anon_sym_exception] = ACTIONS(1938),
    [anon_sym_module] = ACTIONS(1938),
    [anon_sym_open] = ACTIONS(1938),
    [anon_sym_include] = ACTIONS(1938),
    [anon_sym_class] = ACTIONS(1938),
    [anon_sym_val] = ACTIONS(1938),
    [anon_sym_end] = ACTIONS(1938),
    [anon_sym_object] = ACTIONS(1938),
    [anon_sym_in] = ACTIONS(1938),
    [anon_sym_AMP] = ACTIONS(1938),
    [anon_sym_POUND] = ACTIONS(1938),
    [anon_sym_COLON_COLON] = ACTIONS(1940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1940),
    [anon_sym_else] = ACTIONS(1938),
    [anon_sym_new] = ACTIONS(1938),
    [anon_sym_LBRACE_LT] = ACTIONS(1940),
    [anon_sym_begin] = ACTIONS(1938),
    [sym_ocamlyacc_value] = ACTIONS(1940),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1938),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1940),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1938),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1938),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1940),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1940),
    [aux_sym_number_token1] = ACTIONS(1940),
    [aux_sym_signed_number_token1] = ACTIONS(1938),
    [anon_sym_SQUOTE] = ACTIONS(1940),
    [anon_sym_DQUOTE] = ACTIONS(1940),
    [sym_prefix_operator] = ACTIONS(1940),
    [aux_sym_sign_operator_token1] = ACTIONS(1938),
    [sym_hash_operator] = ACTIONS(1940),
    [anon_sym_lsl] = ACTIONS(1938),
    [anon_sym_lsr] = ACTIONS(1938),
    [anon_sym_asr] = ACTIONS(1938),
    [aux_sym_pow_operator_token1] = ACTIONS(1940),
    [anon_sym_mod] = ACTIONS(1938),
    [anon_sym_land] = ACTIONS(1938),
    [anon_sym_lor] = ACTIONS(1938),
    [anon_sym_lxor] = ACTIONS(1938),
    [aux_sym_mult_operator_token1] = ACTIONS(1938),
    [aux_sym_add_operator_token1] = ACTIONS(1938),
    [sym_concat_operator] = ACTIONS(1940),
    [sym_rel_operator] = ACTIONS(1938),
    [anon_sym_AMP_AMP] = ACTIONS(1938),
    [anon_sym_or] = ACTIONS(1938),
    [anon_sym_PIPE_PIPE] = ACTIONS(1938),
    [sym_assign_operator] = ACTIONS(1940),
    [sym_let_operator] = ACTIONS(1940),
    [sym_let_and_operator] = ACTIONS(1940),
    [sym__uppercase_identifier] = ACTIONS(1940),
    [aux_sym_directive_token1] = ACTIONS(1938),
    [aux_sym_tag_token1] = ACTIONS(1940),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1021)] = {
    [sym_attribute] = STATE(1021),
    [sym__lowercase_identifier] = ACTIONS(1946),
    [anon_sym_SEMI_SEMI] = ACTIONS(1948),
    [anon_sym_let] = ACTIONS(1946),
    [anon_sym_and] = ACTIONS(1946),
    [anon_sym_TILDE] = ACTIONS(1946),
    [anon_sym_QMARK] = ACTIONS(1946),
    [anon_sym_LPAREN] = ACTIONS(1948),
    [anon_sym_external] = ACTIONS(1946),
    [anon_sym_type] = ACTIONS(1946),
    [anon_sym_COMMA] = ACTIONS(1948),
    [anon_sym_PIPE] = ACTIONS(1946),
    [anon_sym_DOT] = ACTIONS(1948),
    [anon_sym_LBRACK] = ACTIONS(1946),
    [anon_sym_RBRACK] = ACTIONS(1948),
    [anon_sym_true] = ACTIONS(1946),
    [anon_sym_false] = ACTIONS(1946),
    [anon_sym_LBRACE] = ACTIONS(1946),
    [anon_sym_SEMI] = ACTIONS(1946),
    [anon_sym_exception] = ACTIONS(1946),
    [anon_sym_module] = ACTIONS(1946),
    [anon_sym_open] = ACTIONS(1946),
    [anon_sym_include] = ACTIONS(1946),
    [anon_sym_class] = ACTIONS(1946),
    [anon_sym_val] = ACTIONS(1946),
    [anon_sym_end] = ACTIONS(1946),
    [anon_sym_object] = ACTIONS(1946),
    [anon_sym_in] = ACTIONS(1946),
    [anon_sym_AMP] = ACTIONS(1946),
    [anon_sym_POUND] = ACTIONS(1946),
    [anon_sym_COLON_COLON] = ACTIONS(1948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1948),
    [anon_sym_else] = ACTIONS(1946),
    [anon_sym_new] = ACTIONS(1946),
    [anon_sym_LBRACE_LT] = ACTIONS(1948),
    [anon_sym_begin] = ACTIONS(1946),
    [sym_ocamlyacc_value] = ACTIONS(1948),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1946),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1948),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1946),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1946),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1948),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1948),
    [aux_sym_number_token1] = ACTIONS(1948),
    [aux_sym_signed_number_token1] = ACTIONS(1946),
    [anon_sym_SQUOTE] = ACTIONS(1948),
    [anon_sym_DQUOTE] = ACTIONS(1948),
    [sym_prefix_operator] = ACTIONS(1948),
    [aux_sym_sign_operator_token1] = ACTIONS(1946),
    [sym_hash_operator] = ACTIONS(1948),
    [anon_sym_lsl] = ACTIONS(1946),
    [anon_sym_lsr] = ACTIONS(1946),
    [anon_sym_asr] = ACTIONS(1946),
    [aux_sym_pow_operator_token1] = ACTIONS(1948),
    [anon_sym_mod] = ACTIONS(1946),
    [anon_sym_land] = ACTIONS(1946),
    [anon_sym_lor] = ACTIONS(1946),
    [anon_sym_lxor] = ACTIONS(1946),
    [aux_sym_mult_operator_token1] = ACTIONS(1946),
    [aux_sym_add_operator_token1] = ACTIONS(1946),
    [sym_concat_operator] = ACTIONS(1948),
    [sym_rel_operator] = ACTIONS(1946),
    [anon_sym_AMP_AMP] = ACTIONS(1946),
    [anon_sym_or] = ACTIONS(1946),
    [anon_sym_PIPE_PIPE] = ACTIONS(1946),
    [sym_assign_operator] = ACTIONS(1948),
    [sym_let_operator] = ACTIONS(1948),
    [sym_let_and_operator] = ACTIONS(1948),
    [sym__uppercase_identifier] = ACTIONS(1948),
    [aux_sym_directive_token1] = ACTIONS(1946),
    [aux_sym_tag_token1] = ACTIONS(1948),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1022)] = {
    [sym_attribute] = STATE(1022),
    [sym__lowercase_identifier] = ACTIONS(1898),
    [anon_sym_SEMI_SEMI] = ACTIONS(1900),
    [anon_sym_let] = ACTIONS(1898),
    [anon_sym_and] = ACTIONS(1898),
    [anon_sym_TILDE] = ACTIONS(1898),
    [anon_sym_QMARK] = ACTIONS(1898),
    [anon_sym_LPAREN] = ACTIONS(1900),
    [anon_sym_external] = ACTIONS(1898),
    [anon_sym_type] = ACTIONS(1898),
    [anon_sym_COMMA] = ACTIONS(1900),
    [anon_sym_PIPE] = ACTIONS(1898),
    [anon_sym_DOT] = ACTIONS(1900),
    [anon_sym_LBRACK] = ACTIONS(1898),
    [anon_sym_RBRACK] = ACTIONS(1900),
    [anon_sym_true] = ACTIONS(1898),
    [anon_sym_false] = ACTIONS(1898),
    [anon_sym_LBRACE] = ACTIONS(1898),
    [anon_sym_SEMI] = ACTIONS(1898),
    [anon_sym_exception] = ACTIONS(1898),
    [anon_sym_module] = ACTIONS(1898),
    [anon_sym_open] = ACTIONS(1898),
    [anon_sym_include] = ACTIONS(1898),
    [anon_sym_class] = ACTIONS(1898),
    [anon_sym_val] = ACTIONS(1898),
    [anon_sym_end] = ACTIONS(1898),
    [anon_sym_object] = ACTIONS(1898),
    [anon_sym_in] = ACTIONS(1898),
    [anon_sym_AMP] = ACTIONS(1898),
    [anon_sym_POUND] = ACTIONS(1898),
    [anon_sym_COLON_COLON] = ACTIONS(1900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1900),
    [anon_sym_else] = ACTIONS(1898),
    [anon_sym_new] = ACTIONS(1898),
    [anon_sym_LBRACE_LT] = ACTIONS(1900),
    [anon_sym_begin] = ACTIONS(1898),
    [sym_ocamlyacc_value] = ACTIONS(1900),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1898),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1900),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1898),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1898),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1900),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1900),
    [aux_sym_number_token1] = ACTIONS(1900),
    [aux_sym_signed_number_token1] = ACTIONS(1898),
    [anon_sym_SQUOTE] = ACTIONS(1900),
    [anon_sym_DQUOTE] = ACTIONS(1900),
    [sym_prefix_operator] = ACTIONS(1900),
    [aux_sym_sign_operator_token1] = ACTIONS(1898),
    [sym_hash_operator] = ACTIONS(1900),
    [anon_sym_lsl] = ACTIONS(1898),
    [anon_sym_lsr] = ACTIONS(1898),
    [anon_sym_asr] = ACTIONS(1898),
    [aux_sym_pow_operator_token1] = ACTIONS(1900),
    [anon_sym_mod] = ACTIONS(1898),
    [anon_sym_land] = ACTIONS(1898),
    [anon_sym_lor] = ACTIONS(1898),
    [anon_sym_lxor] = ACTIONS(1898),
    [aux_sym_mult_operator_token1] = ACTIONS(1898),
    [aux_sym_add_operator_token1] = ACTIONS(1898),
    [sym_concat_operator] = ACTIONS(1900),
    [sym_rel_operator] = ACTIONS(1898),
    [anon_sym_AMP_AMP] = ACTIONS(1898),
    [anon_sym_or] = ACTIONS(1898),
    [anon_sym_PIPE_PIPE] = ACTIONS(1898),
    [sym_assign_operator] = ACTIONS(1900),
    [sym_let_operator] = ACTIONS(1900),
    [sym_let_and_operator] = ACTIONS(1900),
    [sym__uppercase_identifier] = ACTIONS(1900),
    [aux_sym_directive_token1] = ACTIONS(1898),
    [aux_sym_tag_token1] = ACTIONS(1900),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1023)] = {
    [sym_attribute] = STATE(1023),
    [sym__lowercase_identifier] = ACTIONS(1736),
    [anon_sym_SEMI_SEMI] = ACTIONS(1738),
    [anon_sym_let] = ACTIONS(1736),
    [anon_sym_and] = ACTIONS(1736),
    [anon_sym_TILDE] = ACTIONS(1736),
    [anon_sym_QMARK] = ACTIONS(1736),
    [anon_sym_LPAREN] = ACTIONS(1738),
    [anon_sym_external] = ACTIONS(1736),
    [anon_sym_type] = ACTIONS(1736),
    [anon_sym_COMMA] = ACTIONS(1738),
    [anon_sym_PIPE] = ACTIONS(1736),
    [anon_sym_DOT] = ACTIONS(1738),
    [anon_sym_LBRACK] = ACTIONS(1736),
    [anon_sym_RBRACK] = ACTIONS(1738),
    [anon_sym_true] = ACTIONS(1736),
    [anon_sym_false] = ACTIONS(1736),
    [anon_sym_LBRACE] = ACTIONS(1736),
    [anon_sym_SEMI] = ACTIONS(1736),
    [anon_sym_exception] = ACTIONS(1736),
    [anon_sym_module] = ACTIONS(1736),
    [anon_sym_open] = ACTIONS(1736),
    [anon_sym_include] = ACTIONS(1736),
    [anon_sym_class] = ACTIONS(1736),
    [anon_sym_val] = ACTIONS(1736),
    [anon_sym_end] = ACTIONS(1736),
    [anon_sym_object] = ACTIONS(1736),
    [anon_sym_in] = ACTIONS(1736),
    [anon_sym_AMP] = ACTIONS(1736),
    [anon_sym_POUND] = ACTIONS(1736),
    [anon_sym_COLON_COLON] = ACTIONS(1738),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1738),
    [anon_sym_else] = ACTIONS(1736),
    [anon_sym_new] = ACTIONS(1736),
    [anon_sym_LBRACE_LT] = ACTIONS(1738),
    [anon_sym_begin] = ACTIONS(1736),
    [sym_ocamlyacc_value] = ACTIONS(1738),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1736),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1738),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1736),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1736),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1738),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1738),
    [aux_sym_number_token1] = ACTIONS(1738),
    [aux_sym_signed_number_token1] = ACTIONS(1736),
    [anon_sym_SQUOTE] = ACTIONS(1738),
    [anon_sym_DQUOTE] = ACTIONS(1738),
    [sym_prefix_operator] = ACTIONS(1738),
    [aux_sym_sign_operator_token1] = ACTIONS(1736),
    [sym_hash_operator] = ACTIONS(1738),
    [anon_sym_lsl] = ACTIONS(1736),
    [anon_sym_lsr] = ACTIONS(1736),
    [anon_sym_asr] = ACTIONS(1736),
    [aux_sym_pow_operator_token1] = ACTIONS(1738),
    [anon_sym_mod] = ACTIONS(1736),
    [anon_sym_land] = ACTIONS(1736),
    [anon_sym_lor] = ACTIONS(1736),
    [anon_sym_lxor] = ACTIONS(1736),
    [aux_sym_mult_operator_token1] = ACTIONS(1736),
    [aux_sym_add_operator_token1] = ACTIONS(1736),
    [sym_concat_operator] = ACTIONS(1738),
    [sym_rel_operator] = ACTIONS(1736),
    [anon_sym_AMP_AMP] = ACTIONS(1736),
    [anon_sym_or] = ACTIONS(1736),
    [anon_sym_PIPE_PIPE] = ACTIONS(1736),
    [sym_assign_operator] = ACTIONS(1738),
    [sym_let_operator] = ACTIONS(1738),
    [sym_let_and_operator] = ACTIONS(1738),
    [sym__uppercase_identifier] = ACTIONS(1738),
    [aux_sym_directive_token1] = ACTIONS(1736),
    [aux_sym_tag_token1] = ACTIONS(1738),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1024)] = {
    [sym_attribute] = STATE(1024),
    [sym__lowercase_identifier] = ACTIONS(1746),
    [anon_sym_SEMI_SEMI] = ACTIONS(1748),
    [anon_sym_let] = ACTIONS(1746),
    [anon_sym_and] = ACTIONS(1746),
    [anon_sym_TILDE] = ACTIONS(1746),
    [anon_sym_QMARK] = ACTIONS(1746),
    [anon_sym_LPAREN] = ACTIONS(1748),
    [anon_sym_external] = ACTIONS(1746),
    [anon_sym_type] = ACTIONS(1746),
    [anon_sym_COMMA] = ACTIONS(1748),
    [anon_sym_PIPE] = ACTIONS(1746),
    [anon_sym_DOT] = ACTIONS(1748),
    [anon_sym_LBRACK] = ACTIONS(1746),
    [anon_sym_RBRACK] = ACTIONS(1748),
    [anon_sym_true] = ACTIONS(1746),
    [anon_sym_false] = ACTIONS(1746),
    [anon_sym_LBRACE] = ACTIONS(1746),
    [anon_sym_SEMI] = ACTIONS(1746),
    [anon_sym_exception] = ACTIONS(1746),
    [anon_sym_module] = ACTIONS(1746),
    [anon_sym_open] = ACTIONS(1746),
    [anon_sym_include] = ACTIONS(1746),
    [anon_sym_class] = ACTIONS(1746),
    [anon_sym_val] = ACTIONS(1746),
    [anon_sym_end] = ACTIONS(1746),
    [anon_sym_object] = ACTIONS(1746),
    [anon_sym_in] = ACTIONS(1746),
    [anon_sym_AMP] = ACTIONS(1746),
    [anon_sym_POUND] = ACTIONS(1746),
    [anon_sym_COLON_COLON] = ACTIONS(1748),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1748),
    [anon_sym_else] = ACTIONS(1746),
    [anon_sym_new] = ACTIONS(1746),
    [anon_sym_LBRACE_LT] = ACTIONS(1748),
    [anon_sym_begin] = ACTIONS(1746),
    [sym_ocamlyacc_value] = ACTIONS(1748),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1746),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1748),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1746),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1746),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1748),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1748),
    [aux_sym_number_token1] = ACTIONS(1748),
    [aux_sym_signed_number_token1] = ACTIONS(1746),
    [anon_sym_SQUOTE] = ACTIONS(1748),
    [anon_sym_DQUOTE] = ACTIONS(1748),
    [sym_prefix_operator] = ACTIONS(1748),
    [aux_sym_sign_operator_token1] = ACTIONS(1746),
    [sym_hash_operator] = ACTIONS(1748),
    [anon_sym_lsl] = ACTIONS(1746),
    [anon_sym_lsr] = ACTIONS(1746),
    [anon_sym_asr] = ACTIONS(1746),
    [aux_sym_pow_operator_token1] = ACTIONS(1748),
    [anon_sym_mod] = ACTIONS(1746),
    [anon_sym_land] = ACTIONS(1746),
    [anon_sym_lor] = ACTIONS(1746),
    [anon_sym_lxor] = ACTIONS(1746),
    [aux_sym_mult_operator_token1] = ACTIONS(1746),
    [aux_sym_add_operator_token1] = ACTIONS(1746),
    [sym_concat_operator] = ACTIONS(1748),
    [sym_rel_operator] = ACTIONS(1746),
    [anon_sym_AMP_AMP] = ACTIONS(1746),
    [anon_sym_or] = ACTIONS(1746),
    [anon_sym_PIPE_PIPE] = ACTIONS(1746),
    [sym_assign_operator] = ACTIONS(1748),
    [sym_let_operator] = ACTIONS(1748),
    [sym_let_and_operator] = ACTIONS(1748),
    [sym__uppercase_identifier] = ACTIONS(1748),
    [aux_sym_directive_token1] = ACTIONS(1746),
    [aux_sym_tag_token1] = ACTIONS(1748),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1025)] = {
    [sym_attribute] = STATE(1025),
    [sym__lowercase_identifier] = ACTIONS(1586),
    [anon_sym_SEMI_SEMI] = ACTIONS(1588),
    [anon_sym_let] = ACTIONS(1586),
    [anon_sym_and] = ACTIONS(1586),
    [anon_sym_TILDE] = ACTIONS(1586),
    [anon_sym_QMARK] = ACTIONS(1586),
    [anon_sym_LPAREN] = ACTIONS(1588),
    [anon_sym_external] = ACTIONS(1586),
    [anon_sym_type] = ACTIONS(1586),
    [anon_sym_COMMA] = ACTIONS(1588),
    [anon_sym_PIPE] = ACTIONS(1586),
    [anon_sym_DOT] = ACTIONS(1588),
    [anon_sym_LBRACK] = ACTIONS(1586),
    [anon_sym_RBRACK] = ACTIONS(1588),
    [anon_sym_true] = ACTIONS(1586),
    [anon_sym_false] = ACTIONS(1586),
    [anon_sym_LBRACE] = ACTIONS(1586),
    [anon_sym_SEMI] = ACTIONS(1586),
    [anon_sym_exception] = ACTIONS(1586),
    [anon_sym_module] = ACTIONS(1586),
    [anon_sym_open] = ACTIONS(1586),
    [anon_sym_include] = ACTIONS(1586),
    [anon_sym_class] = ACTIONS(1586),
    [anon_sym_val] = ACTIONS(1586),
    [anon_sym_end] = ACTIONS(1586),
    [anon_sym_object] = ACTIONS(1586),
    [anon_sym_in] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1586),
    [anon_sym_POUND] = ACTIONS(1586),
    [anon_sym_COLON_COLON] = ACTIONS(1588),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1588),
    [anon_sym_else] = ACTIONS(1586),
    [anon_sym_new] = ACTIONS(1586),
    [anon_sym_LBRACE_LT] = ACTIONS(1588),
    [anon_sym_begin] = ACTIONS(1586),
    [sym_ocamlyacc_value] = ACTIONS(1588),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1586),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1588),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1586),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1586),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1588),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1588),
    [aux_sym_number_token1] = ACTIONS(1588),
    [aux_sym_signed_number_token1] = ACTIONS(1586),
    [anon_sym_SQUOTE] = ACTIONS(1588),
    [anon_sym_DQUOTE] = ACTIONS(1588),
    [sym_prefix_operator] = ACTIONS(1588),
    [aux_sym_sign_operator_token1] = ACTIONS(1586),
    [sym_hash_operator] = ACTIONS(1588),
    [anon_sym_lsl] = ACTIONS(1586),
    [anon_sym_lsr] = ACTIONS(1586),
    [anon_sym_asr] = ACTIONS(1586),
    [aux_sym_pow_operator_token1] = ACTIONS(1588),
    [anon_sym_mod] = ACTIONS(1586),
    [anon_sym_land] = ACTIONS(1586),
    [anon_sym_lor] = ACTIONS(1586),
    [anon_sym_lxor] = ACTIONS(1586),
    [aux_sym_mult_operator_token1] = ACTIONS(1586),
    [aux_sym_add_operator_token1] = ACTIONS(1586),
    [sym_concat_operator] = ACTIONS(1588),
    [sym_rel_operator] = ACTIONS(1586),
    [anon_sym_AMP_AMP] = ACTIONS(1586),
    [anon_sym_or] = ACTIONS(1586),
    [anon_sym_PIPE_PIPE] = ACTIONS(1586),
    [sym_assign_operator] = ACTIONS(1588),
    [sym_let_operator] = ACTIONS(1588),
    [sym_let_and_operator] = ACTIONS(1588),
    [sym__uppercase_identifier] = ACTIONS(1588),
    [aux_sym_directive_token1] = ACTIONS(1586),
    [aux_sym_tag_token1] = ACTIONS(1588),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1026)] = {
    [sym_attribute] = STATE(1026),
    [sym__lowercase_identifier] = ACTIONS(1590),
    [anon_sym_SEMI_SEMI] = ACTIONS(1592),
    [anon_sym_let] = ACTIONS(1590),
    [anon_sym_and] = ACTIONS(1590),
    [anon_sym_TILDE] = ACTIONS(1590),
    [anon_sym_QMARK] = ACTIONS(1590),
    [anon_sym_LPAREN] = ACTIONS(1592),
    [anon_sym_external] = ACTIONS(1590),
    [anon_sym_type] = ACTIONS(1590),
    [anon_sym_COMMA] = ACTIONS(1592),
    [anon_sym_PIPE] = ACTIONS(1590),
    [anon_sym_DOT] = ACTIONS(1592),
    [anon_sym_LBRACK] = ACTIONS(1590),
    [anon_sym_RBRACK] = ACTIONS(1592),
    [anon_sym_true] = ACTIONS(1590),
    [anon_sym_false] = ACTIONS(1590),
    [anon_sym_LBRACE] = ACTIONS(1590),
    [anon_sym_SEMI] = ACTIONS(1590),
    [anon_sym_exception] = ACTIONS(1590),
    [anon_sym_module] = ACTIONS(1590),
    [anon_sym_open] = ACTIONS(1590),
    [anon_sym_include] = ACTIONS(1590),
    [anon_sym_class] = ACTIONS(1590),
    [anon_sym_val] = ACTIONS(1590),
    [anon_sym_end] = ACTIONS(1590),
    [anon_sym_object] = ACTIONS(1590),
    [anon_sym_in] = ACTIONS(1590),
    [anon_sym_AMP] = ACTIONS(1590),
    [anon_sym_POUND] = ACTIONS(1590),
    [anon_sym_COLON_COLON] = ACTIONS(1592),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1592),
    [anon_sym_else] = ACTIONS(1590),
    [anon_sym_new] = ACTIONS(1590),
    [anon_sym_LBRACE_LT] = ACTIONS(1592),
    [anon_sym_begin] = ACTIONS(1590),
    [sym_ocamlyacc_value] = ACTIONS(1592),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1590),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1592),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1590),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1590),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1592),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1592),
    [aux_sym_number_token1] = ACTIONS(1592),
    [aux_sym_signed_number_token1] = ACTIONS(1590),
    [anon_sym_SQUOTE] = ACTIONS(1592),
    [anon_sym_DQUOTE] = ACTIONS(1592),
    [sym_prefix_operator] = ACTIONS(1592),
    [aux_sym_sign_operator_token1] = ACTIONS(1590),
    [sym_hash_operator] = ACTIONS(1592),
    [anon_sym_lsl] = ACTIONS(1590),
    [anon_sym_lsr] = ACTIONS(1590),
    [anon_sym_asr] = ACTIONS(1590),
    [aux_sym_pow_operator_token1] = ACTIONS(1592),
    [anon_sym_mod] = ACTIONS(1590),
    [anon_sym_land] = ACTIONS(1590),
    [anon_sym_lor] = ACTIONS(1590),
    [anon_sym_lxor] = ACTIONS(1590),
    [aux_sym_mult_operator_token1] = ACTIONS(1590),
    [aux_sym_add_operator_token1] = ACTIONS(1590),
    [sym_concat_operator] = ACTIONS(1592),
    [sym_rel_operator] = ACTIONS(1590),
    [anon_sym_AMP_AMP] = ACTIONS(1590),
    [anon_sym_or] = ACTIONS(1590),
    [anon_sym_PIPE_PIPE] = ACTIONS(1590),
    [sym_assign_operator] = ACTIONS(1592),
    [sym_let_operator] = ACTIONS(1592),
    [sym_let_and_operator] = ACTIONS(1592),
    [sym__uppercase_identifier] = ACTIONS(1592),
    [aux_sym_directive_token1] = ACTIONS(1590),
    [aux_sym_tag_token1] = ACTIONS(1592),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1027)] = {
    [sym_attribute] = STATE(1027),
    [sym__lowercase_identifier] = ACTIONS(1610),
    [anon_sym_SEMI_SEMI] = ACTIONS(1612),
    [anon_sym_let] = ACTIONS(1610),
    [anon_sym_and] = ACTIONS(1610),
    [anon_sym_TILDE] = ACTIONS(1610),
    [anon_sym_QMARK] = ACTIONS(1610),
    [anon_sym_LPAREN] = ACTIONS(1612),
    [anon_sym_external] = ACTIONS(1610),
    [anon_sym_type] = ACTIONS(1610),
    [anon_sym_COMMA] = ACTIONS(1612),
    [anon_sym_PIPE] = ACTIONS(1610),
    [anon_sym_DOT] = ACTIONS(1612),
    [anon_sym_LBRACK] = ACTIONS(1610),
    [anon_sym_RBRACK] = ACTIONS(1612),
    [anon_sym_true] = ACTIONS(1610),
    [anon_sym_false] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1610),
    [anon_sym_SEMI] = ACTIONS(1610),
    [anon_sym_exception] = ACTIONS(1610),
    [anon_sym_module] = ACTIONS(1610),
    [anon_sym_open] = ACTIONS(1610),
    [anon_sym_include] = ACTIONS(1610),
    [anon_sym_class] = ACTIONS(1610),
    [anon_sym_val] = ACTIONS(1610),
    [anon_sym_end] = ACTIONS(1610),
    [anon_sym_object] = ACTIONS(1610),
    [anon_sym_in] = ACTIONS(1610),
    [anon_sym_AMP] = ACTIONS(1610),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_COLON_COLON] = ACTIONS(1612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1612),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_new] = ACTIONS(1610),
    [anon_sym_LBRACE_LT] = ACTIONS(1612),
    [anon_sym_begin] = ACTIONS(1610),
    [sym_ocamlyacc_value] = ACTIONS(1612),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1610),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1612),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1610),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1610),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1612),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1612),
    [aux_sym_number_token1] = ACTIONS(1612),
    [aux_sym_signed_number_token1] = ACTIONS(1610),
    [anon_sym_SQUOTE] = ACTIONS(1612),
    [anon_sym_DQUOTE] = ACTIONS(1612),
    [sym_prefix_operator] = ACTIONS(1612),
    [aux_sym_sign_operator_token1] = ACTIONS(1610),
    [sym_hash_operator] = ACTIONS(1612),
    [anon_sym_lsl] = ACTIONS(1610),
    [anon_sym_lsr] = ACTIONS(1610),
    [anon_sym_asr] = ACTIONS(1610),
    [aux_sym_pow_operator_token1] = ACTIONS(1612),
    [anon_sym_mod] = ACTIONS(1610),
    [anon_sym_land] = ACTIONS(1610),
    [anon_sym_lor] = ACTIONS(1610),
    [anon_sym_lxor] = ACTIONS(1610),
    [aux_sym_mult_operator_token1] = ACTIONS(1610),
    [aux_sym_add_operator_token1] = ACTIONS(1610),
    [sym_concat_operator] = ACTIONS(1612),
    [sym_rel_operator] = ACTIONS(1610),
    [anon_sym_AMP_AMP] = ACTIONS(1610),
    [anon_sym_or] = ACTIONS(1610),
    [anon_sym_PIPE_PIPE] = ACTIONS(1610),
    [sym_assign_operator] = ACTIONS(1612),
    [sym_let_operator] = ACTIONS(1612),
    [sym_let_and_operator] = ACTIONS(1612),
    [sym__uppercase_identifier] = ACTIONS(1612),
    [aux_sym_directive_token1] = ACTIONS(1610),
    [aux_sym_tag_token1] = ACTIONS(1612),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1028)] = {
    [sym_attribute] = STATE(1028),
    [sym__lowercase_identifier] = ACTIONS(1630),
    [anon_sym_SEMI_SEMI] = ACTIONS(1632),
    [anon_sym_let] = ACTIONS(1630),
    [anon_sym_and] = ACTIONS(1630),
    [anon_sym_TILDE] = ACTIONS(1630),
    [anon_sym_QMARK] = ACTIONS(1630),
    [anon_sym_LPAREN] = ACTIONS(1632),
    [anon_sym_external] = ACTIONS(1630),
    [anon_sym_type] = ACTIONS(1630),
    [anon_sym_COMMA] = ACTIONS(1632),
    [anon_sym_PIPE] = ACTIONS(1630),
    [anon_sym_DOT] = ACTIONS(1632),
    [anon_sym_LBRACK] = ACTIONS(1630),
    [anon_sym_RBRACK] = ACTIONS(1632),
    [anon_sym_true] = ACTIONS(1630),
    [anon_sym_false] = ACTIONS(1630),
    [anon_sym_LBRACE] = ACTIONS(1630),
    [anon_sym_SEMI] = ACTIONS(1630),
    [anon_sym_exception] = ACTIONS(1630),
    [anon_sym_module] = ACTIONS(1630),
    [anon_sym_open] = ACTIONS(1630),
    [anon_sym_include] = ACTIONS(1630),
    [anon_sym_class] = ACTIONS(1630),
    [anon_sym_val] = ACTIONS(1630),
    [anon_sym_end] = ACTIONS(1630),
    [anon_sym_object] = ACTIONS(1630),
    [anon_sym_in] = ACTIONS(1630),
    [anon_sym_AMP] = ACTIONS(1630),
    [anon_sym_POUND] = ACTIONS(1630),
    [anon_sym_COLON_COLON] = ACTIONS(1632),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1632),
    [anon_sym_else] = ACTIONS(1630),
    [anon_sym_new] = ACTIONS(1630),
    [anon_sym_LBRACE_LT] = ACTIONS(1632),
    [anon_sym_begin] = ACTIONS(1630),
    [sym_ocamlyacc_value] = ACTIONS(1632),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1630),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1632),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1630),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1630),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1632),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1632),
    [aux_sym_number_token1] = ACTIONS(1632),
    [aux_sym_signed_number_token1] = ACTIONS(1630),
    [anon_sym_SQUOTE] = ACTIONS(1632),
    [anon_sym_DQUOTE] = ACTIONS(1632),
    [sym_prefix_operator] = ACTIONS(1632),
    [aux_sym_sign_operator_token1] = ACTIONS(1630),
    [sym_hash_operator] = ACTIONS(1632),
    [anon_sym_lsl] = ACTIONS(1630),
    [anon_sym_lsr] = ACTIONS(1630),
    [anon_sym_asr] = ACTIONS(1630),
    [aux_sym_pow_operator_token1] = ACTIONS(1632),
    [anon_sym_mod] = ACTIONS(1630),
    [anon_sym_land] = ACTIONS(1630),
    [anon_sym_lor] = ACTIONS(1630),
    [anon_sym_lxor] = ACTIONS(1630),
    [aux_sym_mult_operator_token1] = ACTIONS(1630),
    [aux_sym_add_operator_token1] = ACTIONS(1630),
    [sym_concat_operator] = ACTIONS(1632),
    [sym_rel_operator] = ACTIONS(1630),
    [anon_sym_AMP_AMP] = ACTIONS(1630),
    [anon_sym_or] = ACTIONS(1630),
    [anon_sym_PIPE_PIPE] = ACTIONS(1630),
    [sym_assign_operator] = ACTIONS(1632),
    [sym_let_operator] = ACTIONS(1632),
    [sym_let_and_operator] = ACTIONS(1632),
    [sym__uppercase_identifier] = ACTIONS(1632),
    [aux_sym_directive_token1] = ACTIONS(1630),
    [aux_sym_tag_token1] = ACTIONS(1632),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1029)] = {
    [sym_attribute] = STATE(1029),
    [sym__lowercase_identifier] = ACTIONS(1638),
    [anon_sym_SEMI_SEMI] = ACTIONS(1640),
    [anon_sym_let] = ACTIONS(1638),
    [anon_sym_and] = ACTIONS(1638),
    [anon_sym_TILDE] = ACTIONS(1638),
    [anon_sym_QMARK] = ACTIONS(1638),
    [anon_sym_LPAREN] = ACTIONS(1640),
    [anon_sym_external] = ACTIONS(1638),
    [anon_sym_type] = ACTIONS(1638),
    [anon_sym_COMMA] = ACTIONS(1640),
    [anon_sym_PIPE] = ACTIONS(1638),
    [anon_sym_DOT] = ACTIONS(1640),
    [anon_sym_LBRACK] = ACTIONS(1638),
    [anon_sym_RBRACK] = ACTIONS(1640),
    [anon_sym_true] = ACTIONS(1638),
    [anon_sym_false] = ACTIONS(1638),
    [anon_sym_LBRACE] = ACTIONS(1638),
    [anon_sym_SEMI] = ACTIONS(1638),
    [anon_sym_exception] = ACTIONS(1638),
    [anon_sym_module] = ACTIONS(1638),
    [anon_sym_open] = ACTIONS(1638),
    [anon_sym_include] = ACTIONS(1638),
    [anon_sym_class] = ACTIONS(1638),
    [anon_sym_val] = ACTIONS(1638),
    [anon_sym_end] = ACTIONS(1638),
    [anon_sym_object] = ACTIONS(1638),
    [anon_sym_in] = ACTIONS(1638),
    [anon_sym_AMP] = ACTIONS(1638),
    [anon_sym_POUND] = ACTIONS(1638),
    [anon_sym_COLON_COLON] = ACTIONS(1640),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1640),
    [anon_sym_else] = ACTIONS(1638),
    [anon_sym_new] = ACTIONS(1638),
    [anon_sym_LBRACE_LT] = ACTIONS(1640),
    [anon_sym_begin] = ACTIONS(1638),
    [sym_ocamlyacc_value] = ACTIONS(1640),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1638),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1640),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1638),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1638),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1640),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1640),
    [aux_sym_number_token1] = ACTIONS(1640),
    [aux_sym_signed_number_token1] = ACTIONS(1638),
    [anon_sym_SQUOTE] = ACTIONS(1640),
    [anon_sym_DQUOTE] = ACTIONS(1640),
    [sym_prefix_operator] = ACTIONS(1640),
    [aux_sym_sign_operator_token1] = ACTIONS(1638),
    [sym_hash_operator] = ACTIONS(1640),
    [anon_sym_lsl] = ACTIONS(1638),
    [anon_sym_lsr] = ACTIONS(1638),
    [anon_sym_asr] = ACTIONS(1638),
    [aux_sym_pow_operator_token1] = ACTIONS(1640),
    [anon_sym_mod] = ACTIONS(1638),
    [anon_sym_land] = ACTIONS(1638),
    [anon_sym_lor] = ACTIONS(1638),
    [anon_sym_lxor] = ACTIONS(1638),
    [aux_sym_mult_operator_token1] = ACTIONS(1638),
    [aux_sym_add_operator_token1] = ACTIONS(1638),
    [sym_concat_operator] = ACTIONS(1640),
    [sym_rel_operator] = ACTIONS(1638),
    [anon_sym_AMP_AMP] = ACTIONS(1638),
    [anon_sym_or] = ACTIONS(1638),
    [anon_sym_PIPE_PIPE] = ACTIONS(1638),
    [sym_assign_operator] = ACTIONS(1640),
    [sym_let_operator] = ACTIONS(1640),
    [sym_let_and_operator] = ACTIONS(1640),
    [sym__uppercase_identifier] = ACTIONS(1640),
    [aux_sym_directive_token1] = ACTIONS(1638),
    [aux_sym_tag_token1] = ACTIONS(1640),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1030)] = {
    [sym_attribute] = STATE(1030),
    [sym__lowercase_identifier] = ACTIONS(1694),
    [anon_sym_SEMI_SEMI] = ACTIONS(1696),
    [anon_sym_let] = ACTIONS(1694),
    [anon_sym_and] = ACTIONS(1694),
    [anon_sym_TILDE] = ACTIONS(1694),
    [anon_sym_QMARK] = ACTIONS(1694),
    [anon_sym_LPAREN] = ACTIONS(1696),
    [anon_sym_external] = ACTIONS(1694),
    [anon_sym_type] = ACTIONS(1694),
    [anon_sym_COMMA] = ACTIONS(1696),
    [anon_sym_PIPE] = ACTIONS(1694),
    [anon_sym_DOT] = ACTIONS(1696),
    [anon_sym_LBRACK] = ACTIONS(1694),
    [anon_sym_RBRACK] = ACTIONS(1696),
    [anon_sym_true] = ACTIONS(1694),
    [anon_sym_false] = ACTIONS(1694),
    [anon_sym_LBRACE] = ACTIONS(1694),
    [anon_sym_SEMI] = ACTIONS(1694),
    [anon_sym_exception] = ACTIONS(1694),
    [anon_sym_module] = ACTIONS(1694),
    [anon_sym_open] = ACTIONS(1694),
    [anon_sym_include] = ACTIONS(1694),
    [anon_sym_class] = ACTIONS(1694),
    [anon_sym_val] = ACTIONS(1694),
    [anon_sym_end] = ACTIONS(1694),
    [anon_sym_object] = ACTIONS(1694),
    [anon_sym_in] = ACTIONS(1694),
    [anon_sym_AMP] = ACTIONS(1694),
    [anon_sym_POUND] = ACTIONS(1694),
    [anon_sym_COLON_COLON] = ACTIONS(1696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1696),
    [anon_sym_else] = ACTIONS(1694),
    [anon_sym_new] = ACTIONS(1694),
    [anon_sym_LBRACE_LT] = ACTIONS(1696),
    [anon_sym_begin] = ACTIONS(1694),
    [sym_ocamlyacc_value] = ACTIONS(1696),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1694),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1696),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1694),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1694),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1696),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1696),
    [aux_sym_number_token1] = ACTIONS(1696),
    [aux_sym_signed_number_token1] = ACTIONS(1694),
    [anon_sym_SQUOTE] = ACTIONS(1696),
    [anon_sym_DQUOTE] = ACTIONS(1696),
    [sym_prefix_operator] = ACTIONS(1696),
    [aux_sym_sign_operator_token1] = ACTIONS(1694),
    [sym_hash_operator] = ACTIONS(1696),
    [anon_sym_lsl] = ACTIONS(1694),
    [anon_sym_lsr] = ACTIONS(1694),
    [anon_sym_asr] = ACTIONS(1694),
    [aux_sym_pow_operator_token1] = ACTIONS(1696),
    [anon_sym_mod] = ACTIONS(1694),
    [anon_sym_land] = ACTIONS(1694),
    [anon_sym_lor] = ACTIONS(1694),
    [anon_sym_lxor] = ACTIONS(1694),
    [aux_sym_mult_operator_token1] = ACTIONS(1694),
    [aux_sym_add_operator_token1] = ACTIONS(1694),
    [sym_concat_operator] = ACTIONS(1696),
    [sym_rel_operator] = ACTIONS(1694),
    [anon_sym_AMP_AMP] = ACTIONS(1694),
    [anon_sym_or] = ACTIONS(1694),
    [anon_sym_PIPE_PIPE] = ACTIONS(1694),
    [sym_assign_operator] = ACTIONS(1696),
    [sym_let_operator] = ACTIONS(1696),
    [sym_let_and_operator] = ACTIONS(1696),
    [sym__uppercase_identifier] = ACTIONS(1696),
    [aux_sym_directive_token1] = ACTIONS(1694),
    [aux_sym_tag_token1] = ACTIONS(1696),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1031)] = {
    [sym_attribute] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1712),
    [anon_sym_SEMI_SEMI] = ACTIONS(1714),
    [anon_sym_let] = ACTIONS(1712),
    [anon_sym_and] = ACTIONS(1712),
    [anon_sym_TILDE] = ACTIONS(1712),
    [anon_sym_QMARK] = ACTIONS(1712),
    [anon_sym_LPAREN] = ACTIONS(1714),
    [anon_sym_external] = ACTIONS(1712),
    [anon_sym_type] = ACTIONS(1712),
    [anon_sym_COMMA] = ACTIONS(1714),
    [anon_sym_PIPE] = ACTIONS(1712),
    [anon_sym_DOT] = ACTIONS(1714),
    [anon_sym_LBRACK] = ACTIONS(1712),
    [anon_sym_RBRACK] = ACTIONS(1714),
    [anon_sym_true] = ACTIONS(1712),
    [anon_sym_false] = ACTIONS(1712),
    [anon_sym_LBRACE] = ACTIONS(1712),
    [anon_sym_SEMI] = ACTIONS(1712),
    [anon_sym_exception] = ACTIONS(1712),
    [anon_sym_module] = ACTIONS(1712),
    [anon_sym_open] = ACTIONS(1712),
    [anon_sym_include] = ACTIONS(1712),
    [anon_sym_class] = ACTIONS(1712),
    [anon_sym_val] = ACTIONS(1712),
    [anon_sym_end] = ACTIONS(1712),
    [anon_sym_object] = ACTIONS(1712),
    [anon_sym_in] = ACTIONS(1712),
    [anon_sym_AMP] = ACTIONS(1712),
    [anon_sym_POUND] = ACTIONS(1712),
    [anon_sym_COLON_COLON] = ACTIONS(1714),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1714),
    [anon_sym_else] = ACTIONS(1712),
    [anon_sym_new] = ACTIONS(1712),
    [anon_sym_LBRACE_LT] = ACTIONS(1714),
    [anon_sym_begin] = ACTIONS(1712),
    [sym_ocamlyacc_value] = ACTIONS(1714),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1712),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1714),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1712),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1712),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1714),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1714),
    [aux_sym_number_token1] = ACTIONS(1714),
    [aux_sym_signed_number_token1] = ACTIONS(1712),
    [anon_sym_SQUOTE] = ACTIONS(1714),
    [anon_sym_DQUOTE] = ACTIONS(1714),
    [sym_prefix_operator] = ACTIONS(1714),
    [aux_sym_sign_operator_token1] = ACTIONS(1712),
    [sym_hash_operator] = ACTIONS(1714),
    [anon_sym_lsl] = ACTIONS(1712),
    [anon_sym_lsr] = ACTIONS(1712),
    [anon_sym_asr] = ACTIONS(1712),
    [aux_sym_pow_operator_token1] = ACTIONS(1714),
    [anon_sym_mod] = ACTIONS(1712),
    [anon_sym_land] = ACTIONS(1712),
    [anon_sym_lor] = ACTIONS(1712),
    [anon_sym_lxor] = ACTIONS(1712),
    [aux_sym_mult_operator_token1] = ACTIONS(1712),
    [aux_sym_add_operator_token1] = ACTIONS(1712),
    [sym_concat_operator] = ACTIONS(1714),
    [sym_rel_operator] = ACTIONS(1712),
    [anon_sym_AMP_AMP] = ACTIONS(1712),
    [anon_sym_or] = ACTIONS(1712),
    [anon_sym_PIPE_PIPE] = ACTIONS(1712),
    [sym_assign_operator] = ACTIONS(1714),
    [sym_let_operator] = ACTIONS(1714),
    [sym_let_and_operator] = ACTIONS(1714),
    [sym__uppercase_identifier] = ACTIONS(1714),
    [aux_sym_directive_token1] = ACTIONS(1712),
    [aux_sym_tag_token1] = ACTIONS(1714),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1032)] = {
    [sym_attribute] = STATE(1032),
    [sym__lowercase_identifier] = ACTIONS(1720),
    [anon_sym_SEMI_SEMI] = ACTIONS(1722),
    [anon_sym_let] = ACTIONS(1720),
    [anon_sym_and] = ACTIONS(1720),
    [anon_sym_TILDE] = ACTIONS(1720),
    [anon_sym_QMARK] = ACTIONS(1720),
    [anon_sym_LPAREN] = ACTIONS(1722),
    [anon_sym_external] = ACTIONS(1720),
    [anon_sym_type] = ACTIONS(1720),
    [anon_sym_COMMA] = ACTIONS(1722),
    [anon_sym_PIPE] = ACTIONS(1720),
    [anon_sym_DOT] = ACTIONS(1722),
    [anon_sym_LBRACK] = ACTIONS(1720),
    [anon_sym_RBRACK] = ACTIONS(1722),
    [anon_sym_true] = ACTIONS(1720),
    [anon_sym_false] = ACTIONS(1720),
    [anon_sym_LBRACE] = ACTIONS(1720),
    [anon_sym_SEMI] = ACTIONS(1720),
    [anon_sym_exception] = ACTIONS(1720),
    [anon_sym_module] = ACTIONS(1720),
    [anon_sym_open] = ACTIONS(1720),
    [anon_sym_include] = ACTIONS(1720),
    [anon_sym_class] = ACTIONS(1720),
    [anon_sym_val] = ACTIONS(1720),
    [anon_sym_end] = ACTIONS(1720),
    [anon_sym_object] = ACTIONS(1720),
    [anon_sym_in] = ACTIONS(1720),
    [anon_sym_AMP] = ACTIONS(1720),
    [anon_sym_POUND] = ACTIONS(1720),
    [anon_sym_COLON_COLON] = ACTIONS(1722),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1722),
    [anon_sym_else] = ACTIONS(1720),
    [anon_sym_new] = ACTIONS(1720),
    [anon_sym_LBRACE_LT] = ACTIONS(1722),
    [anon_sym_begin] = ACTIONS(1720),
    [sym_ocamlyacc_value] = ACTIONS(1722),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1720),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1722),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1720),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1720),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1722),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1722),
    [aux_sym_number_token1] = ACTIONS(1722),
    [aux_sym_signed_number_token1] = ACTIONS(1720),
    [anon_sym_SQUOTE] = ACTIONS(1722),
    [anon_sym_DQUOTE] = ACTIONS(1722),
    [sym_prefix_operator] = ACTIONS(1722),
    [aux_sym_sign_operator_token1] = ACTIONS(1720),
    [sym_hash_operator] = ACTIONS(1722),
    [anon_sym_lsl] = ACTIONS(1720),
    [anon_sym_lsr] = ACTIONS(1720),
    [anon_sym_asr] = ACTIONS(1720),
    [aux_sym_pow_operator_token1] = ACTIONS(1722),
    [anon_sym_mod] = ACTIONS(1720),
    [anon_sym_land] = ACTIONS(1720),
    [anon_sym_lor] = ACTIONS(1720),
    [anon_sym_lxor] = ACTIONS(1720),
    [aux_sym_mult_operator_token1] = ACTIONS(1720),
    [aux_sym_add_operator_token1] = ACTIONS(1720),
    [sym_concat_operator] = ACTIONS(1722),
    [sym_rel_operator] = ACTIONS(1720),
    [anon_sym_AMP_AMP] = ACTIONS(1720),
    [anon_sym_or] = ACTIONS(1720),
    [anon_sym_PIPE_PIPE] = ACTIONS(1720),
    [sym_assign_operator] = ACTIONS(1722),
    [sym_let_operator] = ACTIONS(1722),
    [sym_let_and_operator] = ACTIONS(1722),
    [sym__uppercase_identifier] = ACTIONS(1722),
    [aux_sym_directive_token1] = ACTIONS(1720),
    [aux_sym_tag_token1] = ACTIONS(1722),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1033)] = {
    [sym_attribute] = STATE(1033),
    [sym__lowercase_identifier] = ACTIONS(1826),
    [anon_sym_SEMI_SEMI] = ACTIONS(1828),
    [anon_sym_let] = ACTIONS(1826),
    [anon_sym_and] = ACTIONS(1826),
    [anon_sym_TILDE] = ACTIONS(1826),
    [anon_sym_QMARK] = ACTIONS(1826),
    [anon_sym_LPAREN] = ACTIONS(1828),
    [anon_sym_external] = ACTIONS(1826),
    [anon_sym_type] = ACTIONS(1826),
    [anon_sym_COMMA] = ACTIONS(1828),
    [anon_sym_PIPE] = ACTIONS(1826),
    [anon_sym_DOT] = ACTIONS(1828),
    [anon_sym_LBRACK] = ACTIONS(1826),
    [anon_sym_RBRACK] = ACTIONS(1828),
    [anon_sym_true] = ACTIONS(1826),
    [anon_sym_false] = ACTIONS(1826),
    [anon_sym_LBRACE] = ACTIONS(1826),
    [anon_sym_SEMI] = ACTIONS(1826),
    [anon_sym_exception] = ACTIONS(1826),
    [anon_sym_module] = ACTIONS(1826),
    [anon_sym_open] = ACTIONS(1826),
    [anon_sym_include] = ACTIONS(1826),
    [anon_sym_class] = ACTIONS(1826),
    [anon_sym_val] = ACTIONS(1826),
    [anon_sym_end] = ACTIONS(1826),
    [anon_sym_object] = ACTIONS(1826),
    [anon_sym_in] = ACTIONS(1826),
    [anon_sym_AMP] = ACTIONS(1826),
    [anon_sym_POUND] = ACTIONS(1826),
    [anon_sym_COLON_COLON] = ACTIONS(1828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1828),
    [anon_sym_else] = ACTIONS(1826),
    [anon_sym_new] = ACTIONS(1826),
    [anon_sym_LBRACE_LT] = ACTIONS(1828),
    [anon_sym_begin] = ACTIONS(1826),
    [sym_ocamlyacc_value] = ACTIONS(1828),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1826),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1828),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1826),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1826),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1828),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1828),
    [aux_sym_number_token1] = ACTIONS(1828),
    [aux_sym_signed_number_token1] = ACTIONS(1826),
    [anon_sym_SQUOTE] = ACTIONS(1828),
    [anon_sym_DQUOTE] = ACTIONS(1828),
    [sym_prefix_operator] = ACTIONS(1828),
    [aux_sym_sign_operator_token1] = ACTIONS(1826),
    [sym_hash_operator] = ACTIONS(1828),
    [anon_sym_lsl] = ACTIONS(1826),
    [anon_sym_lsr] = ACTIONS(1826),
    [anon_sym_asr] = ACTIONS(1826),
    [aux_sym_pow_operator_token1] = ACTIONS(1828),
    [anon_sym_mod] = ACTIONS(1826),
    [anon_sym_land] = ACTIONS(1826),
    [anon_sym_lor] = ACTIONS(1826),
    [anon_sym_lxor] = ACTIONS(1826),
    [aux_sym_mult_operator_token1] = ACTIONS(1826),
    [aux_sym_add_operator_token1] = ACTIONS(1826),
    [sym_concat_operator] = ACTIONS(1828),
    [sym_rel_operator] = ACTIONS(1826),
    [anon_sym_AMP_AMP] = ACTIONS(1826),
    [anon_sym_or] = ACTIONS(1826),
    [anon_sym_PIPE_PIPE] = ACTIONS(1826),
    [sym_assign_operator] = ACTIONS(1828),
    [sym_let_operator] = ACTIONS(1828),
    [sym_let_and_operator] = ACTIONS(1828),
    [sym__uppercase_identifier] = ACTIONS(1828),
    [aux_sym_directive_token1] = ACTIONS(1826),
    [aux_sym_tag_token1] = ACTIONS(1828),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1034)] = {
    [sym_attribute] = STATE(1034),
    [sym__lowercase_identifier] = ACTIONS(1834),
    [anon_sym_SEMI_SEMI] = ACTIONS(1836),
    [anon_sym_let] = ACTIONS(1834),
    [anon_sym_and] = ACTIONS(1834),
    [anon_sym_TILDE] = ACTIONS(1834),
    [anon_sym_QMARK] = ACTIONS(1834),
    [anon_sym_LPAREN] = ACTIONS(1836),
    [anon_sym_external] = ACTIONS(1834),
    [anon_sym_type] = ACTIONS(1834),
    [anon_sym_COMMA] = ACTIONS(1836),
    [anon_sym_PIPE] = ACTIONS(1834),
    [anon_sym_DOT] = ACTIONS(1836),
    [anon_sym_LBRACK] = ACTIONS(1834),
    [anon_sym_RBRACK] = ACTIONS(1836),
    [anon_sym_true] = ACTIONS(1834),
    [anon_sym_false] = ACTIONS(1834),
    [anon_sym_LBRACE] = ACTIONS(1834),
    [anon_sym_SEMI] = ACTIONS(1834),
    [anon_sym_exception] = ACTIONS(1834),
    [anon_sym_module] = ACTIONS(1834),
    [anon_sym_open] = ACTIONS(1834),
    [anon_sym_include] = ACTIONS(1834),
    [anon_sym_class] = ACTIONS(1834),
    [anon_sym_val] = ACTIONS(1834),
    [anon_sym_end] = ACTIONS(1834),
    [anon_sym_object] = ACTIONS(1834),
    [anon_sym_in] = ACTIONS(1834),
    [anon_sym_AMP] = ACTIONS(1834),
    [anon_sym_POUND] = ACTIONS(1834),
    [anon_sym_COLON_COLON] = ACTIONS(1836),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1836),
    [anon_sym_else] = ACTIONS(1834),
    [anon_sym_new] = ACTIONS(1834),
    [anon_sym_LBRACE_LT] = ACTIONS(1836),
    [anon_sym_begin] = ACTIONS(1834),
    [sym_ocamlyacc_value] = ACTIONS(1836),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1834),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1836),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1834),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1834),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1836),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1836),
    [aux_sym_number_token1] = ACTIONS(1836),
    [aux_sym_signed_number_token1] = ACTIONS(1834),
    [anon_sym_SQUOTE] = ACTIONS(1836),
    [anon_sym_DQUOTE] = ACTIONS(1836),
    [sym_prefix_operator] = ACTIONS(1836),
    [aux_sym_sign_operator_token1] = ACTIONS(1834),
    [sym_hash_operator] = ACTIONS(1836),
    [anon_sym_lsl] = ACTIONS(1834),
    [anon_sym_lsr] = ACTIONS(1834),
    [anon_sym_asr] = ACTIONS(1834),
    [aux_sym_pow_operator_token1] = ACTIONS(1836),
    [anon_sym_mod] = ACTIONS(1834),
    [anon_sym_land] = ACTIONS(1834),
    [anon_sym_lor] = ACTIONS(1834),
    [anon_sym_lxor] = ACTIONS(1834),
    [aux_sym_mult_operator_token1] = ACTIONS(1834),
    [aux_sym_add_operator_token1] = ACTIONS(1834),
    [sym_concat_operator] = ACTIONS(1836),
    [sym_rel_operator] = ACTIONS(1834),
    [anon_sym_AMP_AMP] = ACTIONS(1834),
    [anon_sym_or] = ACTIONS(1834),
    [anon_sym_PIPE_PIPE] = ACTIONS(1834),
    [sym_assign_operator] = ACTIONS(1836),
    [sym_let_operator] = ACTIONS(1836),
    [sym_let_and_operator] = ACTIONS(1836),
    [sym__uppercase_identifier] = ACTIONS(1836),
    [aux_sym_directive_token1] = ACTIONS(1834),
    [aux_sym_tag_token1] = ACTIONS(1836),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1035)] = {
    [sym_attribute] = STATE(1035),
    [sym__lowercase_identifier] = ACTIONS(1708),
    [anon_sym_SEMI_SEMI] = ACTIONS(1710),
    [anon_sym_let] = ACTIONS(1708),
    [anon_sym_and] = ACTIONS(1708),
    [anon_sym_TILDE] = ACTIONS(1708),
    [anon_sym_QMARK] = ACTIONS(1708),
    [anon_sym_LPAREN] = ACTIONS(1710),
    [anon_sym_external] = ACTIONS(1708),
    [anon_sym_type] = ACTIONS(1708),
    [anon_sym_COMMA] = ACTIONS(1710),
    [anon_sym_PIPE] = ACTIONS(1708),
    [anon_sym_DOT] = ACTIONS(1710),
    [anon_sym_LBRACK] = ACTIONS(1708),
    [anon_sym_RBRACK] = ACTIONS(1710),
    [anon_sym_true] = ACTIONS(1708),
    [anon_sym_false] = ACTIONS(1708),
    [anon_sym_LBRACE] = ACTIONS(1708),
    [anon_sym_SEMI] = ACTIONS(1708),
    [anon_sym_exception] = ACTIONS(1708),
    [anon_sym_module] = ACTIONS(1708),
    [anon_sym_open] = ACTIONS(1708),
    [anon_sym_include] = ACTIONS(1708),
    [anon_sym_class] = ACTIONS(1708),
    [anon_sym_val] = ACTIONS(1708),
    [anon_sym_end] = ACTIONS(1708),
    [anon_sym_object] = ACTIONS(1708),
    [anon_sym_in] = ACTIONS(1708),
    [anon_sym_AMP] = ACTIONS(1708),
    [anon_sym_POUND] = ACTIONS(1708),
    [anon_sym_COLON_COLON] = ACTIONS(1710),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1710),
    [anon_sym_else] = ACTIONS(1708),
    [anon_sym_new] = ACTIONS(1708),
    [anon_sym_LBRACE_LT] = ACTIONS(1710),
    [anon_sym_begin] = ACTIONS(1708),
    [sym_ocamlyacc_value] = ACTIONS(1710),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1708),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1710),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1708),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1708),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1710),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1710),
    [aux_sym_number_token1] = ACTIONS(1710),
    [aux_sym_signed_number_token1] = ACTIONS(1708),
    [anon_sym_SQUOTE] = ACTIONS(1710),
    [anon_sym_DQUOTE] = ACTIONS(1710),
    [sym_prefix_operator] = ACTIONS(1710),
    [aux_sym_sign_operator_token1] = ACTIONS(1708),
    [sym_hash_operator] = ACTIONS(1710),
    [anon_sym_lsl] = ACTIONS(1708),
    [anon_sym_lsr] = ACTIONS(1708),
    [anon_sym_asr] = ACTIONS(1708),
    [aux_sym_pow_operator_token1] = ACTIONS(1710),
    [anon_sym_mod] = ACTIONS(1708),
    [anon_sym_land] = ACTIONS(1708),
    [anon_sym_lor] = ACTIONS(1708),
    [anon_sym_lxor] = ACTIONS(1708),
    [aux_sym_mult_operator_token1] = ACTIONS(1708),
    [aux_sym_add_operator_token1] = ACTIONS(1708),
    [sym_concat_operator] = ACTIONS(1710),
    [sym_rel_operator] = ACTIONS(1708),
    [anon_sym_AMP_AMP] = ACTIONS(1708),
    [anon_sym_or] = ACTIONS(1708),
    [anon_sym_PIPE_PIPE] = ACTIONS(1708),
    [sym_assign_operator] = ACTIONS(1710),
    [sym_let_operator] = ACTIONS(1710),
    [sym_let_and_operator] = ACTIONS(1710),
    [sym__uppercase_identifier] = ACTIONS(1710),
    [aux_sym_directive_token1] = ACTIONS(1708),
    [aux_sym_tag_token1] = ACTIONS(1710),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1036)] = {
    [sym_attribute] = STATE(1036),
    [sym__lowercase_identifier] = ACTIONS(1822),
    [anon_sym_SEMI_SEMI] = ACTIONS(1824),
    [anon_sym_let] = ACTIONS(1822),
    [anon_sym_and] = ACTIONS(1822),
    [anon_sym_TILDE] = ACTIONS(1822),
    [anon_sym_QMARK] = ACTIONS(1822),
    [anon_sym_LPAREN] = ACTIONS(1824),
    [anon_sym_external] = ACTIONS(1822),
    [anon_sym_type] = ACTIONS(1822),
    [anon_sym_COMMA] = ACTIONS(1824),
    [anon_sym_PIPE] = ACTIONS(1822),
    [anon_sym_DOT] = ACTIONS(1824),
    [anon_sym_LBRACK] = ACTIONS(1822),
    [anon_sym_RBRACK] = ACTIONS(1824),
    [anon_sym_true] = ACTIONS(1822),
    [anon_sym_false] = ACTIONS(1822),
    [anon_sym_LBRACE] = ACTIONS(1822),
    [anon_sym_SEMI] = ACTIONS(1822),
    [anon_sym_exception] = ACTIONS(1822),
    [anon_sym_module] = ACTIONS(1822),
    [anon_sym_open] = ACTIONS(1822),
    [anon_sym_include] = ACTIONS(1822),
    [anon_sym_class] = ACTIONS(1822),
    [anon_sym_val] = ACTIONS(1822),
    [anon_sym_end] = ACTIONS(1822),
    [anon_sym_object] = ACTIONS(1822),
    [anon_sym_in] = ACTIONS(1822),
    [anon_sym_AMP] = ACTIONS(1822),
    [anon_sym_POUND] = ACTIONS(1822),
    [anon_sym_COLON_COLON] = ACTIONS(1824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1824),
    [anon_sym_else] = ACTIONS(1822),
    [anon_sym_new] = ACTIONS(1822),
    [anon_sym_LBRACE_LT] = ACTIONS(1824),
    [anon_sym_begin] = ACTIONS(1822),
    [sym_ocamlyacc_value] = ACTIONS(1824),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1822),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1824),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1822),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1822),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1824),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1824),
    [aux_sym_number_token1] = ACTIONS(1824),
    [aux_sym_signed_number_token1] = ACTIONS(1822),
    [anon_sym_SQUOTE] = ACTIONS(1824),
    [anon_sym_DQUOTE] = ACTIONS(1824),
    [sym_prefix_operator] = ACTIONS(1824),
    [aux_sym_sign_operator_token1] = ACTIONS(1822),
    [sym_hash_operator] = ACTIONS(1824),
    [anon_sym_lsl] = ACTIONS(1822),
    [anon_sym_lsr] = ACTIONS(1822),
    [anon_sym_asr] = ACTIONS(1822),
    [aux_sym_pow_operator_token1] = ACTIONS(1824),
    [anon_sym_mod] = ACTIONS(1822),
    [anon_sym_land] = ACTIONS(1822),
    [anon_sym_lor] = ACTIONS(1822),
    [anon_sym_lxor] = ACTIONS(1822),
    [aux_sym_mult_operator_token1] = ACTIONS(1822),
    [aux_sym_add_operator_token1] = ACTIONS(1822),
    [sym_concat_operator] = ACTIONS(1824),
    [sym_rel_operator] = ACTIONS(1822),
    [anon_sym_AMP_AMP] = ACTIONS(1822),
    [anon_sym_or] = ACTIONS(1822),
    [anon_sym_PIPE_PIPE] = ACTIONS(1822),
    [sym_assign_operator] = ACTIONS(1824),
    [sym_let_operator] = ACTIONS(1824),
    [sym_let_and_operator] = ACTIONS(1824),
    [sym__uppercase_identifier] = ACTIONS(1824),
    [aux_sym_directive_token1] = ACTIONS(1822),
    [aux_sym_tag_token1] = ACTIONS(1824),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1037)] = {
    [sym_attribute] = STATE(1037),
    [sym__lowercase_identifier] = ACTIONS(1846),
    [anon_sym_SEMI_SEMI] = ACTIONS(1848),
    [anon_sym_let] = ACTIONS(1846),
    [anon_sym_and] = ACTIONS(1846),
    [anon_sym_TILDE] = ACTIONS(1846),
    [anon_sym_QMARK] = ACTIONS(1846),
    [anon_sym_LPAREN] = ACTIONS(1848),
    [anon_sym_external] = ACTIONS(1846),
    [anon_sym_type] = ACTIONS(1846),
    [anon_sym_COMMA] = ACTIONS(1848),
    [anon_sym_PIPE] = ACTIONS(1846),
    [anon_sym_DOT] = ACTIONS(1848),
    [anon_sym_LBRACK] = ACTIONS(1846),
    [anon_sym_RBRACK] = ACTIONS(1848),
    [anon_sym_true] = ACTIONS(1846),
    [anon_sym_false] = ACTIONS(1846),
    [anon_sym_LBRACE] = ACTIONS(1846),
    [anon_sym_SEMI] = ACTIONS(1846),
    [anon_sym_exception] = ACTIONS(1846),
    [anon_sym_module] = ACTIONS(1846),
    [anon_sym_open] = ACTIONS(1846),
    [anon_sym_include] = ACTIONS(1846),
    [anon_sym_class] = ACTIONS(1846),
    [anon_sym_val] = ACTIONS(1846),
    [anon_sym_end] = ACTIONS(1846),
    [anon_sym_object] = ACTIONS(1846),
    [anon_sym_in] = ACTIONS(1846),
    [anon_sym_AMP] = ACTIONS(1846),
    [anon_sym_POUND] = ACTIONS(1846),
    [anon_sym_COLON_COLON] = ACTIONS(1848),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1848),
    [anon_sym_else] = ACTIONS(1846),
    [anon_sym_new] = ACTIONS(1846),
    [anon_sym_LBRACE_LT] = ACTIONS(1848),
    [anon_sym_begin] = ACTIONS(1846),
    [sym_ocamlyacc_value] = ACTIONS(1848),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1846),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1848),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1846),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1846),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1848),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1848),
    [aux_sym_number_token1] = ACTIONS(1848),
    [aux_sym_signed_number_token1] = ACTIONS(1846),
    [anon_sym_SQUOTE] = ACTIONS(1848),
    [anon_sym_DQUOTE] = ACTIONS(1848),
    [sym_prefix_operator] = ACTIONS(1848),
    [aux_sym_sign_operator_token1] = ACTIONS(1846),
    [sym_hash_operator] = ACTIONS(1848),
    [anon_sym_lsl] = ACTIONS(1846),
    [anon_sym_lsr] = ACTIONS(1846),
    [anon_sym_asr] = ACTIONS(1846),
    [aux_sym_pow_operator_token1] = ACTIONS(1848),
    [anon_sym_mod] = ACTIONS(1846),
    [anon_sym_land] = ACTIONS(1846),
    [anon_sym_lor] = ACTIONS(1846),
    [anon_sym_lxor] = ACTIONS(1846),
    [aux_sym_mult_operator_token1] = ACTIONS(1846),
    [aux_sym_add_operator_token1] = ACTIONS(1846),
    [sym_concat_operator] = ACTIONS(1848),
    [sym_rel_operator] = ACTIONS(1846),
    [anon_sym_AMP_AMP] = ACTIONS(1846),
    [anon_sym_or] = ACTIONS(1846),
    [anon_sym_PIPE_PIPE] = ACTIONS(1846),
    [sym_assign_operator] = ACTIONS(1848),
    [sym_let_operator] = ACTIONS(1848),
    [sym_let_and_operator] = ACTIONS(1848),
    [sym__uppercase_identifier] = ACTIONS(1848),
    [aux_sym_directive_token1] = ACTIONS(1846),
    [aux_sym_tag_token1] = ACTIONS(1848),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1038)] = {
    [sym_attribute] = STATE(1038),
    [sym__lowercase_identifier] = ACTIONS(1886),
    [anon_sym_SEMI_SEMI] = ACTIONS(1888),
    [anon_sym_let] = ACTIONS(1886),
    [anon_sym_and] = ACTIONS(1886),
    [anon_sym_TILDE] = ACTIONS(1886),
    [anon_sym_QMARK] = ACTIONS(1886),
    [anon_sym_LPAREN] = ACTIONS(1888),
    [anon_sym_external] = ACTIONS(1886),
    [anon_sym_type] = ACTIONS(1886),
    [anon_sym_COMMA] = ACTIONS(1888),
    [anon_sym_PIPE] = ACTIONS(1886),
    [anon_sym_DOT] = ACTIONS(1888),
    [anon_sym_LBRACK] = ACTIONS(1886),
    [anon_sym_RBRACK] = ACTIONS(1888),
    [anon_sym_true] = ACTIONS(1886),
    [anon_sym_false] = ACTIONS(1886),
    [anon_sym_LBRACE] = ACTIONS(1886),
    [anon_sym_SEMI] = ACTIONS(1886),
    [anon_sym_exception] = ACTIONS(1886),
    [anon_sym_module] = ACTIONS(1886),
    [anon_sym_open] = ACTIONS(1886),
    [anon_sym_include] = ACTIONS(1886),
    [anon_sym_class] = ACTIONS(1886),
    [anon_sym_val] = ACTIONS(1886),
    [anon_sym_end] = ACTIONS(1886),
    [anon_sym_object] = ACTIONS(1886),
    [anon_sym_in] = ACTIONS(1886),
    [anon_sym_AMP] = ACTIONS(1886),
    [anon_sym_POUND] = ACTIONS(1886),
    [anon_sym_COLON_COLON] = ACTIONS(1888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1888),
    [anon_sym_else] = ACTIONS(1886),
    [anon_sym_new] = ACTIONS(1886),
    [anon_sym_LBRACE_LT] = ACTIONS(1888),
    [anon_sym_begin] = ACTIONS(1886),
    [sym_ocamlyacc_value] = ACTIONS(1888),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1886),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1888),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1886),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1886),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1888),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1888),
    [aux_sym_number_token1] = ACTIONS(1888),
    [aux_sym_signed_number_token1] = ACTIONS(1886),
    [anon_sym_SQUOTE] = ACTIONS(1888),
    [anon_sym_DQUOTE] = ACTIONS(1888),
    [sym_prefix_operator] = ACTIONS(1888),
    [aux_sym_sign_operator_token1] = ACTIONS(1886),
    [sym_hash_operator] = ACTIONS(1888),
    [anon_sym_lsl] = ACTIONS(1886),
    [anon_sym_lsr] = ACTIONS(1886),
    [anon_sym_asr] = ACTIONS(1886),
    [aux_sym_pow_operator_token1] = ACTIONS(1888),
    [anon_sym_mod] = ACTIONS(1886),
    [anon_sym_land] = ACTIONS(1886),
    [anon_sym_lor] = ACTIONS(1886),
    [anon_sym_lxor] = ACTIONS(1886),
    [aux_sym_mult_operator_token1] = ACTIONS(1886),
    [aux_sym_add_operator_token1] = ACTIONS(1886),
    [sym_concat_operator] = ACTIONS(1888),
    [sym_rel_operator] = ACTIONS(1886),
    [anon_sym_AMP_AMP] = ACTIONS(1886),
    [anon_sym_or] = ACTIONS(1886),
    [anon_sym_PIPE_PIPE] = ACTIONS(1886),
    [sym_assign_operator] = ACTIONS(1888),
    [sym_let_operator] = ACTIONS(1888),
    [sym_let_and_operator] = ACTIONS(1888),
    [sym__uppercase_identifier] = ACTIONS(1888),
    [aux_sym_directive_token1] = ACTIONS(1886),
    [aux_sym_tag_token1] = ACTIONS(1888),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1039)] = {
    [sym_attribute] = STATE(1039),
    [sym__lowercase_identifier] = ACTIONS(1914),
    [anon_sym_SEMI_SEMI] = ACTIONS(1916),
    [anon_sym_let] = ACTIONS(1914),
    [anon_sym_and] = ACTIONS(1914),
    [anon_sym_TILDE] = ACTIONS(1914),
    [anon_sym_QMARK] = ACTIONS(1914),
    [anon_sym_LPAREN] = ACTIONS(1916),
    [anon_sym_external] = ACTIONS(1914),
    [anon_sym_type] = ACTIONS(1914),
    [anon_sym_COMMA] = ACTIONS(1916),
    [anon_sym_PIPE] = ACTIONS(1914),
    [anon_sym_DOT] = ACTIONS(1916),
    [anon_sym_LBRACK] = ACTIONS(1914),
    [anon_sym_RBRACK] = ACTIONS(1916),
    [anon_sym_true] = ACTIONS(1914),
    [anon_sym_false] = ACTIONS(1914),
    [anon_sym_LBRACE] = ACTIONS(1914),
    [anon_sym_SEMI] = ACTIONS(1914),
    [anon_sym_exception] = ACTIONS(1914),
    [anon_sym_module] = ACTIONS(1914),
    [anon_sym_open] = ACTIONS(1914),
    [anon_sym_include] = ACTIONS(1914),
    [anon_sym_class] = ACTIONS(1914),
    [anon_sym_val] = ACTIONS(1914),
    [anon_sym_end] = ACTIONS(1914),
    [anon_sym_object] = ACTIONS(1914),
    [anon_sym_in] = ACTIONS(1914),
    [anon_sym_AMP] = ACTIONS(1914),
    [anon_sym_POUND] = ACTIONS(1914),
    [anon_sym_COLON_COLON] = ACTIONS(1916),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1916),
    [anon_sym_else] = ACTIONS(1914),
    [anon_sym_new] = ACTIONS(1914),
    [anon_sym_LBRACE_LT] = ACTIONS(1916),
    [anon_sym_begin] = ACTIONS(1914),
    [sym_ocamlyacc_value] = ACTIONS(1916),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1914),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1916),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1914),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1914),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1916),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1916),
    [aux_sym_number_token1] = ACTIONS(1916),
    [aux_sym_signed_number_token1] = ACTIONS(1914),
    [anon_sym_SQUOTE] = ACTIONS(1916),
    [anon_sym_DQUOTE] = ACTIONS(1916),
    [sym_prefix_operator] = ACTIONS(1916),
    [aux_sym_sign_operator_token1] = ACTIONS(1914),
    [sym_hash_operator] = ACTIONS(1916),
    [anon_sym_lsl] = ACTIONS(1914),
    [anon_sym_lsr] = ACTIONS(1914),
    [anon_sym_asr] = ACTIONS(1914),
    [aux_sym_pow_operator_token1] = ACTIONS(1916),
    [anon_sym_mod] = ACTIONS(1914),
    [anon_sym_land] = ACTIONS(1914),
    [anon_sym_lor] = ACTIONS(1914),
    [anon_sym_lxor] = ACTIONS(1914),
    [aux_sym_mult_operator_token1] = ACTIONS(1914),
    [aux_sym_add_operator_token1] = ACTIONS(1914),
    [sym_concat_operator] = ACTIONS(1916),
    [sym_rel_operator] = ACTIONS(1914),
    [anon_sym_AMP_AMP] = ACTIONS(1914),
    [anon_sym_or] = ACTIONS(1914),
    [anon_sym_PIPE_PIPE] = ACTIONS(1914),
    [sym_assign_operator] = ACTIONS(1916),
    [sym_let_operator] = ACTIONS(1916),
    [sym_let_and_operator] = ACTIONS(1916),
    [sym__uppercase_identifier] = ACTIONS(1916),
    [aux_sym_directive_token1] = ACTIONS(1914),
    [aux_sym_tag_token1] = ACTIONS(1916),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1040)] = {
    [sym_attribute] = STATE(1040),
    [sym__lowercase_identifier] = ACTIONS(1942),
    [anon_sym_SEMI_SEMI] = ACTIONS(1944),
    [anon_sym_let] = ACTIONS(1942),
    [anon_sym_and] = ACTIONS(1942),
    [anon_sym_TILDE] = ACTIONS(1942),
    [anon_sym_QMARK] = ACTIONS(1942),
    [anon_sym_LPAREN] = ACTIONS(1944),
    [anon_sym_external] = ACTIONS(1942),
    [anon_sym_type] = ACTIONS(1942),
    [anon_sym_COMMA] = ACTIONS(1944),
    [anon_sym_PIPE] = ACTIONS(1942),
    [anon_sym_DOT] = ACTIONS(1944),
    [anon_sym_LBRACK] = ACTIONS(1942),
    [anon_sym_RBRACK] = ACTIONS(1944),
    [anon_sym_true] = ACTIONS(1942),
    [anon_sym_false] = ACTIONS(1942),
    [anon_sym_LBRACE] = ACTIONS(1942),
    [anon_sym_SEMI] = ACTIONS(1942),
    [anon_sym_exception] = ACTIONS(1942),
    [anon_sym_module] = ACTIONS(1942),
    [anon_sym_open] = ACTIONS(1942),
    [anon_sym_include] = ACTIONS(1942),
    [anon_sym_class] = ACTIONS(1942),
    [anon_sym_val] = ACTIONS(1942),
    [anon_sym_end] = ACTIONS(1942),
    [anon_sym_object] = ACTIONS(1942),
    [anon_sym_in] = ACTIONS(1942),
    [anon_sym_AMP] = ACTIONS(1942),
    [anon_sym_POUND] = ACTIONS(1942),
    [anon_sym_COLON_COLON] = ACTIONS(1944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1944),
    [anon_sym_else] = ACTIONS(1942),
    [anon_sym_new] = ACTIONS(1942),
    [anon_sym_LBRACE_LT] = ACTIONS(1944),
    [anon_sym_begin] = ACTIONS(1942),
    [sym_ocamlyacc_value] = ACTIONS(1944),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1942),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1944),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1942),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1942),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1944),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1944),
    [aux_sym_number_token1] = ACTIONS(1944),
    [aux_sym_signed_number_token1] = ACTIONS(1942),
    [anon_sym_SQUOTE] = ACTIONS(1944),
    [anon_sym_DQUOTE] = ACTIONS(1944),
    [sym_prefix_operator] = ACTIONS(1944),
    [aux_sym_sign_operator_token1] = ACTIONS(1942),
    [sym_hash_operator] = ACTIONS(1944),
    [anon_sym_lsl] = ACTIONS(1942),
    [anon_sym_lsr] = ACTIONS(1942),
    [anon_sym_asr] = ACTIONS(1942),
    [aux_sym_pow_operator_token1] = ACTIONS(1944),
    [anon_sym_mod] = ACTIONS(1942),
    [anon_sym_land] = ACTIONS(1942),
    [anon_sym_lor] = ACTIONS(1942),
    [anon_sym_lxor] = ACTIONS(1942),
    [aux_sym_mult_operator_token1] = ACTIONS(1942),
    [aux_sym_add_operator_token1] = ACTIONS(1942),
    [sym_concat_operator] = ACTIONS(1944),
    [sym_rel_operator] = ACTIONS(1942),
    [anon_sym_AMP_AMP] = ACTIONS(1942),
    [anon_sym_or] = ACTIONS(1942),
    [anon_sym_PIPE_PIPE] = ACTIONS(1942),
    [sym_assign_operator] = ACTIONS(1944),
    [sym_let_operator] = ACTIONS(1944),
    [sym_let_and_operator] = ACTIONS(1944),
    [sym__uppercase_identifier] = ACTIONS(1944),
    [aux_sym_directive_token1] = ACTIONS(1942),
    [aux_sym_tag_token1] = ACTIONS(1944),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1041)] = {
    [sym_attribute] = STATE(1041),
    [sym__lowercase_identifier] = ACTIONS(1842),
    [anon_sym_SEMI_SEMI] = ACTIONS(1844),
    [anon_sym_let] = ACTIONS(1842),
    [anon_sym_and] = ACTIONS(1842),
    [anon_sym_TILDE] = ACTIONS(1842),
    [anon_sym_QMARK] = ACTIONS(1842),
    [anon_sym_LPAREN] = ACTIONS(1844),
    [anon_sym_external] = ACTIONS(1842),
    [anon_sym_type] = ACTIONS(1842),
    [anon_sym_COMMA] = ACTIONS(1844),
    [anon_sym_PIPE] = ACTIONS(1842),
    [anon_sym_DOT] = ACTIONS(1844),
    [anon_sym_LBRACK] = ACTIONS(1842),
    [anon_sym_RBRACK] = ACTIONS(1844),
    [anon_sym_true] = ACTIONS(1842),
    [anon_sym_false] = ACTIONS(1842),
    [anon_sym_LBRACE] = ACTIONS(1842),
    [anon_sym_SEMI] = ACTIONS(1842),
    [anon_sym_exception] = ACTIONS(1842),
    [anon_sym_module] = ACTIONS(1842),
    [anon_sym_open] = ACTIONS(1842),
    [anon_sym_include] = ACTIONS(1842),
    [anon_sym_class] = ACTIONS(1842),
    [anon_sym_val] = ACTIONS(1842),
    [anon_sym_end] = ACTIONS(1842),
    [anon_sym_object] = ACTIONS(1842),
    [anon_sym_in] = ACTIONS(1842),
    [anon_sym_AMP] = ACTIONS(1842),
    [anon_sym_POUND] = ACTIONS(1842),
    [anon_sym_COLON_COLON] = ACTIONS(1844),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1844),
    [anon_sym_else] = ACTIONS(1842),
    [anon_sym_new] = ACTIONS(1842),
    [anon_sym_LBRACE_LT] = ACTIONS(1844),
    [anon_sym_begin] = ACTIONS(1842),
    [sym_ocamlyacc_value] = ACTIONS(1844),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1842),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1844),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1842),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1842),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1844),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1844),
    [aux_sym_number_token1] = ACTIONS(1844),
    [aux_sym_signed_number_token1] = ACTIONS(1842),
    [anon_sym_SQUOTE] = ACTIONS(1844),
    [anon_sym_DQUOTE] = ACTIONS(1844),
    [sym_prefix_operator] = ACTIONS(1844),
    [aux_sym_sign_operator_token1] = ACTIONS(1842),
    [sym_hash_operator] = ACTIONS(1844),
    [anon_sym_lsl] = ACTIONS(1842),
    [anon_sym_lsr] = ACTIONS(1842),
    [anon_sym_asr] = ACTIONS(1842),
    [aux_sym_pow_operator_token1] = ACTIONS(1844),
    [anon_sym_mod] = ACTIONS(1842),
    [anon_sym_land] = ACTIONS(1842),
    [anon_sym_lor] = ACTIONS(1842),
    [anon_sym_lxor] = ACTIONS(1842),
    [aux_sym_mult_operator_token1] = ACTIONS(1842),
    [aux_sym_add_operator_token1] = ACTIONS(1842),
    [sym_concat_operator] = ACTIONS(1844),
    [sym_rel_operator] = ACTIONS(1842),
    [anon_sym_AMP_AMP] = ACTIONS(1842),
    [anon_sym_or] = ACTIONS(1842),
    [anon_sym_PIPE_PIPE] = ACTIONS(1842),
    [sym_assign_operator] = ACTIONS(1844),
    [sym_let_operator] = ACTIONS(1844),
    [sym_let_and_operator] = ACTIONS(1844),
    [sym__uppercase_identifier] = ACTIONS(1844),
    [aux_sym_directive_token1] = ACTIONS(1842),
    [aux_sym_tag_token1] = ACTIONS(1844),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1042)] = {
    [sym_attribute] = STATE(1042),
    [sym__lowercase_identifier] = ACTIONS(1594),
    [anon_sym_SEMI_SEMI] = ACTIONS(1596),
    [anon_sym_let] = ACTIONS(1594),
    [anon_sym_and] = ACTIONS(1594),
    [anon_sym_TILDE] = ACTIONS(1594),
    [anon_sym_QMARK] = ACTIONS(1594),
    [anon_sym_LPAREN] = ACTIONS(1596),
    [anon_sym_external] = ACTIONS(1594),
    [anon_sym_type] = ACTIONS(1594),
    [anon_sym_COMMA] = ACTIONS(1596),
    [anon_sym_PIPE] = ACTIONS(1594),
    [anon_sym_DOT] = ACTIONS(1596),
    [anon_sym_LBRACK] = ACTIONS(1594),
    [anon_sym_RBRACK] = ACTIONS(1596),
    [anon_sym_true] = ACTIONS(1594),
    [anon_sym_false] = ACTIONS(1594),
    [anon_sym_LBRACE] = ACTIONS(1594),
    [anon_sym_SEMI] = ACTIONS(1594),
    [anon_sym_exception] = ACTIONS(1594),
    [anon_sym_module] = ACTIONS(1594),
    [anon_sym_open] = ACTIONS(1594),
    [anon_sym_include] = ACTIONS(1594),
    [anon_sym_class] = ACTIONS(1594),
    [anon_sym_val] = ACTIONS(1594),
    [anon_sym_end] = ACTIONS(1594),
    [anon_sym_object] = ACTIONS(1594),
    [anon_sym_in] = ACTIONS(1594),
    [anon_sym_AMP] = ACTIONS(1594),
    [anon_sym_POUND] = ACTIONS(1594),
    [anon_sym_COLON_COLON] = ACTIONS(1596),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1596),
    [anon_sym_else] = ACTIONS(1594),
    [anon_sym_new] = ACTIONS(1594),
    [anon_sym_LBRACE_LT] = ACTIONS(1596),
    [anon_sym_begin] = ACTIONS(1594),
    [sym_ocamlyacc_value] = ACTIONS(1596),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1594),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1596),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1594),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1594),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1596),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1596),
    [aux_sym_number_token1] = ACTIONS(1596),
    [aux_sym_signed_number_token1] = ACTIONS(1594),
    [anon_sym_SQUOTE] = ACTIONS(1596),
    [anon_sym_DQUOTE] = ACTIONS(1596),
    [sym_prefix_operator] = ACTIONS(1596),
    [aux_sym_sign_operator_token1] = ACTIONS(1594),
    [sym_hash_operator] = ACTIONS(1596),
    [anon_sym_lsl] = ACTIONS(1594),
    [anon_sym_lsr] = ACTIONS(1594),
    [anon_sym_asr] = ACTIONS(1594),
    [aux_sym_pow_operator_token1] = ACTIONS(1596),
    [anon_sym_mod] = ACTIONS(1594),
    [anon_sym_land] = ACTIONS(1594),
    [anon_sym_lor] = ACTIONS(1594),
    [anon_sym_lxor] = ACTIONS(1594),
    [aux_sym_mult_operator_token1] = ACTIONS(1594),
    [aux_sym_add_operator_token1] = ACTIONS(1594),
    [sym_concat_operator] = ACTIONS(1596),
    [sym_rel_operator] = ACTIONS(1594),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_or] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1594),
    [sym_assign_operator] = ACTIONS(1596),
    [sym_let_operator] = ACTIONS(1596),
    [sym_let_and_operator] = ACTIONS(1596),
    [sym__uppercase_identifier] = ACTIONS(1596),
    [aux_sym_directive_token1] = ACTIONS(1594),
    [aux_sym_tag_token1] = ACTIONS(1596),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1043)] = {
    [sym_attribute] = STATE(1043),
    [sym__lowercase_identifier] = ACTIONS(1598),
    [anon_sym_SEMI_SEMI] = ACTIONS(1600),
    [anon_sym_let] = ACTIONS(1598),
    [anon_sym_and] = ACTIONS(1598),
    [anon_sym_TILDE] = ACTIONS(1598),
    [anon_sym_QMARK] = ACTIONS(1598),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_external] = ACTIONS(1598),
    [anon_sym_type] = ACTIONS(1598),
    [anon_sym_COMMA] = ACTIONS(1600),
    [anon_sym_PIPE] = ACTIONS(1598),
    [anon_sym_DOT] = ACTIONS(1600),
    [anon_sym_LBRACK] = ACTIONS(1598),
    [anon_sym_RBRACK] = ACTIONS(1600),
    [anon_sym_true] = ACTIONS(1598),
    [anon_sym_false] = ACTIONS(1598),
    [anon_sym_LBRACE] = ACTIONS(1598),
    [anon_sym_SEMI] = ACTIONS(1598),
    [anon_sym_exception] = ACTIONS(1598),
    [anon_sym_module] = ACTIONS(1598),
    [anon_sym_open] = ACTIONS(1598),
    [anon_sym_include] = ACTIONS(1598),
    [anon_sym_class] = ACTIONS(1598),
    [anon_sym_val] = ACTIONS(1598),
    [anon_sym_end] = ACTIONS(1598),
    [anon_sym_object] = ACTIONS(1598),
    [anon_sym_in] = ACTIONS(1598),
    [anon_sym_AMP] = ACTIONS(1598),
    [anon_sym_POUND] = ACTIONS(1598),
    [anon_sym_COLON_COLON] = ACTIONS(1600),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1600),
    [anon_sym_else] = ACTIONS(1598),
    [anon_sym_new] = ACTIONS(1598),
    [anon_sym_LBRACE_LT] = ACTIONS(1600),
    [anon_sym_begin] = ACTIONS(1598),
    [sym_ocamlyacc_value] = ACTIONS(1600),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1598),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1600),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1598),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1598),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1600),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1600),
    [aux_sym_number_token1] = ACTIONS(1600),
    [aux_sym_signed_number_token1] = ACTIONS(1598),
    [anon_sym_SQUOTE] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1600),
    [sym_prefix_operator] = ACTIONS(1600),
    [aux_sym_sign_operator_token1] = ACTIONS(1598),
    [sym_hash_operator] = ACTIONS(1600),
    [anon_sym_lsl] = ACTIONS(1598),
    [anon_sym_lsr] = ACTIONS(1598),
    [anon_sym_asr] = ACTIONS(1598),
    [aux_sym_pow_operator_token1] = ACTIONS(1600),
    [anon_sym_mod] = ACTIONS(1598),
    [anon_sym_land] = ACTIONS(1598),
    [anon_sym_lor] = ACTIONS(1598),
    [anon_sym_lxor] = ACTIONS(1598),
    [aux_sym_mult_operator_token1] = ACTIONS(1598),
    [aux_sym_add_operator_token1] = ACTIONS(1598),
    [sym_concat_operator] = ACTIONS(1600),
    [sym_rel_operator] = ACTIONS(1598),
    [anon_sym_AMP_AMP] = ACTIONS(1598),
    [anon_sym_or] = ACTIONS(1598),
    [anon_sym_PIPE_PIPE] = ACTIONS(1598),
    [sym_assign_operator] = ACTIONS(1600),
    [sym_let_operator] = ACTIONS(1600),
    [sym_let_and_operator] = ACTIONS(1600),
    [sym__uppercase_identifier] = ACTIONS(1600),
    [aux_sym_directive_token1] = ACTIONS(1598),
    [aux_sym_tag_token1] = ACTIONS(1600),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1044)] = {
    [sym_attribute] = STATE(1044),
    [sym__lowercase_identifier] = ACTIONS(1602),
    [anon_sym_SEMI_SEMI] = ACTIONS(1604),
    [anon_sym_let] = ACTIONS(1602),
    [anon_sym_and] = ACTIONS(1602),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_QMARK] = ACTIONS(1602),
    [anon_sym_LPAREN] = ACTIONS(1604),
    [anon_sym_external] = ACTIONS(1602),
    [anon_sym_type] = ACTIONS(1602),
    [anon_sym_COMMA] = ACTIONS(1604),
    [anon_sym_PIPE] = ACTIONS(1602),
    [anon_sym_DOT] = ACTIONS(1604),
    [anon_sym_LBRACK] = ACTIONS(1602),
    [anon_sym_RBRACK] = ACTIONS(1604),
    [anon_sym_true] = ACTIONS(1602),
    [anon_sym_false] = ACTIONS(1602),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [anon_sym_SEMI] = ACTIONS(1602),
    [anon_sym_exception] = ACTIONS(1602),
    [anon_sym_module] = ACTIONS(1602),
    [anon_sym_open] = ACTIONS(1602),
    [anon_sym_include] = ACTIONS(1602),
    [anon_sym_class] = ACTIONS(1602),
    [anon_sym_val] = ACTIONS(1602),
    [anon_sym_end] = ACTIONS(1602),
    [anon_sym_object] = ACTIONS(1602),
    [anon_sym_in] = ACTIONS(1602),
    [anon_sym_AMP] = ACTIONS(1602),
    [anon_sym_POUND] = ACTIONS(1602),
    [anon_sym_COLON_COLON] = ACTIONS(1604),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1604),
    [anon_sym_else] = ACTIONS(1602),
    [anon_sym_new] = ACTIONS(1602),
    [anon_sym_LBRACE_LT] = ACTIONS(1604),
    [anon_sym_begin] = ACTIONS(1602),
    [sym_ocamlyacc_value] = ACTIONS(1604),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1602),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1604),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1602),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1602),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1604),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1604),
    [aux_sym_number_token1] = ACTIONS(1604),
    [aux_sym_signed_number_token1] = ACTIONS(1602),
    [anon_sym_SQUOTE] = ACTIONS(1604),
    [anon_sym_DQUOTE] = ACTIONS(1604),
    [sym_prefix_operator] = ACTIONS(1604),
    [aux_sym_sign_operator_token1] = ACTIONS(1602),
    [sym_hash_operator] = ACTIONS(1604),
    [anon_sym_lsl] = ACTIONS(1602),
    [anon_sym_lsr] = ACTIONS(1602),
    [anon_sym_asr] = ACTIONS(1602),
    [aux_sym_pow_operator_token1] = ACTIONS(1604),
    [anon_sym_mod] = ACTIONS(1602),
    [anon_sym_land] = ACTIONS(1602),
    [anon_sym_lor] = ACTIONS(1602),
    [anon_sym_lxor] = ACTIONS(1602),
    [aux_sym_mult_operator_token1] = ACTIONS(1602),
    [aux_sym_add_operator_token1] = ACTIONS(1602),
    [sym_concat_operator] = ACTIONS(1604),
    [sym_rel_operator] = ACTIONS(1602),
    [anon_sym_AMP_AMP] = ACTIONS(1602),
    [anon_sym_or] = ACTIONS(1602),
    [anon_sym_PIPE_PIPE] = ACTIONS(1602),
    [sym_assign_operator] = ACTIONS(1604),
    [sym_let_operator] = ACTIONS(1604),
    [sym_let_and_operator] = ACTIONS(1604),
    [sym__uppercase_identifier] = ACTIONS(1604),
    [aux_sym_directive_token1] = ACTIONS(1602),
    [aux_sym_tag_token1] = ACTIONS(1604),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1045)] = {
    [sym_attribute] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(1606),
    [anon_sym_SEMI_SEMI] = ACTIONS(1608),
    [anon_sym_let] = ACTIONS(1606),
    [anon_sym_and] = ACTIONS(1606),
    [anon_sym_TILDE] = ACTIONS(1606),
    [anon_sym_QMARK] = ACTIONS(1606),
    [anon_sym_LPAREN] = ACTIONS(1608),
    [anon_sym_external] = ACTIONS(1606),
    [anon_sym_type] = ACTIONS(1606),
    [anon_sym_COMMA] = ACTIONS(1608),
    [anon_sym_PIPE] = ACTIONS(1606),
    [anon_sym_DOT] = ACTIONS(1608),
    [anon_sym_LBRACK] = ACTIONS(1606),
    [anon_sym_RBRACK] = ACTIONS(1608),
    [anon_sym_true] = ACTIONS(1606),
    [anon_sym_false] = ACTIONS(1606),
    [anon_sym_LBRACE] = ACTIONS(1606),
    [anon_sym_SEMI] = ACTIONS(1606),
    [anon_sym_exception] = ACTIONS(1606),
    [anon_sym_module] = ACTIONS(1606),
    [anon_sym_open] = ACTIONS(1606),
    [anon_sym_include] = ACTIONS(1606),
    [anon_sym_class] = ACTIONS(1606),
    [anon_sym_val] = ACTIONS(1606),
    [anon_sym_end] = ACTIONS(1606),
    [anon_sym_object] = ACTIONS(1606),
    [anon_sym_in] = ACTIONS(1606),
    [anon_sym_AMP] = ACTIONS(1606),
    [anon_sym_POUND] = ACTIONS(1606),
    [anon_sym_COLON_COLON] = ACTIONS(1608),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1608),
    [anon_sym_else] = ACTIONS(1606),
    [anon_sym_new] = ACTIONS(1606),
    [anon_sym_LBRACE_LT] = ACTIONS(1608),
    [anon_sym_begin] = ACTIONS(1606),
    [sym_ocamlyacc_value] = ACTIONS(1608),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1606),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1608),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1606),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1606),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1608),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1608),
    [aux_sym_number_token1] = ACTIONS(1608),
    [aux_sym_signed_number_token1] = ACTIONS(1606),
    [anon_sym_SQUOTE] = ACTIONS(1608),
    [anon_sym_DQUOTE] = ACTIONS(1608),
    [sym_prefix_operator] = ACTIONS(1608),
    [aux_sym_sign_operator_token1] = ACTIONS(1606),
    [sym_hash_operator] = ACTIONS(1608),
    [anon_sym_lsl] = ACTIONS(1606),
    [anon_sym_lsr] = ACTIONS(1606),
    [anon_sym_asr] = ACTIONS(1606),
    [aux_sym_pow_operator_token1] = ACTIONS(1608),
    [anon_sym_mod] = ACTIONS(1606),
    [anon_sym_land] = ACTIONS(1606),
    [anon_sym_lor] = ACTIONS(1606),
    [anon_sym_lxor] = ACTIONS(1606),
    [aux_sym_mult_operator_token1] = ACTIONS(1606),
    [aux_sym_add_operator_token1] = ACTIONS(1606),
    [sym_concat_operator] = ACTIONS(1608),
    [sym_rel_operator] = ACTIONS(1606),
    [anon_sym_AMP_AMP] = ACTIONS(1606),
    [anon_sym_or] = ACTIONS(1606),
    [anon_sym_PIPE_PIPE] = ACTIONS(1606),
    [sym_assign_operator] = ACTIONS(1608),
    [sym_let_operator] = ACTIONS(1608),
    [sym_let_and_operator] = ACTIONS(1608),
    [sym__uppercase_identifier] = ACTIONS(1608),
    [aux_sym_directive_token1] = ACTIONS(1606),
    [aux_sym_tag_token1] = ACTIONS(1608),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1046)] = {
    [sym_attribute] = STATE(1046),
    [sym__lowercase_identifier] = ACTIONS(1626),
    [anon_sym_SEMI_SEMI] = ACTIONS(1628),
    [anon_sym_let] = ACTIONS(1626),
    [anon_sym_and] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1628),
    [anon_sym_external] = ACTIONS(1626),
    [anon_sym_type] = ACTIONS(1626),
    [anon_sym_COMMA] = ACTIONS(1628),
    [anon_sym_PIPE] = ACTIONS(1626),
    [anon_sym_DOT] = ACTIONS(1628),
    [anon_sym_LBRACK] = ACTIONS(1626),
    [anon_sym_RBRACK] = ACTIONS(1628),
    [anon_sym_true] = ACTIONS(1626),
    [anon_sym_false] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1626),
    [anon_sym_exception] = ACTIONS(1626),
    [anon_sym_module] = ACTIONS(1626),
    [anon_sym_open] = ACTIONS(1626),
    [anon_sym_include] = ACTIONS(1626),
    [anon_sym_class] = ACTIONS(1626),
    [anon_sym_val] = ACTIONS(1626),
    [anon_sym_end] = ACTIONS(1626),
    [anon_sym_object] = ACTIONS(1626),
    [anon_sym_in] = ACTIONS(1626),
    [anon_sym_AMP] = ACTIONS(1626),
    [anon_sym_POUND] = ACTIONS(1626),
    [anon_sym_COLON_COLON] = ACTIONS(1628),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1628),
    [anon_sym_else] = ACTIONS(1626),
    [anon_sym_new] = ACTIONS(1626),
    [anon_sym_LBRACE_LT] = ACTIONS(1628),
    [anon_sym_begin] = ACTIONS(1626),
    [sym_ocamlyacc_value] = ACTIONS(1628),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1626),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1628),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1626),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1626),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1628),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1628),
    [aux_sym_number_token1] = ACTIONS(1628),
    [aux_sym_signed_number_token1] = ACTIONS(1626),
    [anon_sym_SQUOTE] = ACTIONS(1628),
    [anon_sym_DQUOTE] = ACTIONS(1628),
    [sym_prefix_operator] = ACTIONS(1628),
    [aux_sym_sign_operator_token1] = ACTIONS(1626),
    [sym_hash_operator] = ACTIONS(1628),
    [anon_sym_lsl] = ACTIONS(1626),
    [anon_sym_lsr] = ACTIONS(1626),
    [anon_sym_asr] = ACTIONS(1626),
    [aux_sym_pow_operator_token1] = ACTIONS(1628),
    [anon_sym_mod] = ACTIONS(1626),
    [anon_sym_land] = ACTIONS(1626),
    [anon_sym_lor] = ACTIONS(1626),
    [anon_sym_lxor] = ACTIONS(1626),
    [aux_sym_mult_operator_token1] = ACTIONS(1626),
    [aux_sym_add_operator_token1] = ACTIONS(1626),
    [sym_concat_operator] = ACTIONS(1628),
    [sym_rel_operator] = ACTIONS(1626),
    [anon_sym_AMP_AMP] = ACTIONS(1626),
    [anon_sym_or] = ACTIONS(1626),
    [anon_sym_PIPE_PIPE] = ACTIONS(1626),
    [sym_assign_operator] = ACTIONS(1628),
    [sym_let_operator] = ACTIONS(1628),
    [sym_let_and_operator] = ACTIONS(1628),
    [sym__uppercase_identifier] = ACTIONS(1628),
    [aux_sym_directive_token1] = ACTIONS(1626),
    [aux_sym_tag_token1] = ACTIONS(1628),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1047)] = {
    [sym_attribute] = STATE(1047),
    [sym__lowercase_identifier] = ACTIONS(1716),
    [anon_sym_SEMI_SEMI] = ACTIONS(1718),
    [anon_sym_let] = ACTIONS(1716),
    [anon_sym_and] = ACTIONS(1716),
    [anon_sym_TILDE] = ACTIONS(1716),
    [anon_sym_QMARK] = ACTIONS(1716),
    [anon_sym_LPAREN] = ACTIONS(1718),
    [anon_sym_external] = ACTIONS(1716),
    [anon_sym_type] = ACTIONS(1716),
    [anon_sym_COMMA] = ACTIONS(1718),
    [anon_sym_PIPE] = ACTIONS(1716),
    [anon_sym_DOT] = ACTIONS(1718),
    [anon_sym_LBRACK] = ACTIONS(1716),
    [anon_sym_RBRACK] = ACTIONS(1718),
    [anon_sym_true] = ACTIONS(1716),
    [anon_sym_false] = ACTIONS(1716),
    [anon_sym_LBRACE] = ACTIONS(1716),
    [anon_sym_SEMI] = ACTIONS(1716),
    [anon_sym_exception] = ACTIONS(1716),
    [anon_sym_module] = ACTIONS(1716),
    [anon_sym_open] = ACTIONS(1716),
    [anon_sym_include] = ACTIONS(1716),
    [anon_sym_class] = ACTIONS(1716),
    [anon_sym_val] = ACTIONS(1716),
    [anon_sym_end] = ACTIONS(1716),
    [anon_sym_object] = ACTIONS(1716),
    [anon_sym_in] = ACTIONS(1716),
    [anon_sym_AMP] = ACTIONS(1716),
    [anon_sym_POUND] = ACTIONS(1716),
    [anon_sym_COLON_COLON] = ACTIONS(1718),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1718),
    [anon_sym_else] = ACTIONS(1716),
    [anon_sym_new] = ACTIONS(1716),
    [anon_sym_LBRACE_LT] = ACTIONS(1718),
    [anon_sym_begin] = ACTIONS(1716),
    [sym_ocamlyacc_value] = ACTIONS(1718),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1716),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1718),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1716),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1716),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1718),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1718),
    [aux_sym_number_token1] = ACTIONS(1718),
    [aux_sym_signed_number_token1] = ACTIONS(1716),
    [anon_sym_SQUOTE] = ACTIONS(1718),
    [anon_sym_DQUOTE] = ACTIONS(1718),
    [sym_prefix_operator] = ACTIONS(1718),
    [aux_sym_sign_operator_token1] = ACTIONS(1716),
    [sym_hash_operator] = ACTIONS(1718),
    [anon_sym_lsl] = ACTIONS(1716),
    [anon_sym_lsr] = ACTIONS(1716),
    [anon_sym_asr] = ACTIONS(1716),
    [aux_sym_pow_operator_token1] = ACTIONS(1718),
    [anon_sym_mod] = ACTIONS(1716),
    [anon_sym_land] = ACTIONS(1716),
    [anon_sym_lor] = ACTIONS(1716),
    [anon_sym_lxor] = ACTIONS(1716),
    [aux_sym_mult_operator_token1] = ACTIONS(1716),
    [aux_sym_add_operator_token1] = ACTIONS(1716),
    [sym_concat_operator] = ACTIONS(1718),
    [sym_rel_operator] = ACTIONS(1716),
    [anon_sym_AMP_AMP] = ACTIONS(1716),
    [anon_sym_or] = ACTIONS(1716),
    [anon_sym_PIPE_PIPE] = ACTIONS(1716),
    [sym_assign_operator] = ACTIONS(1718),
    [sym_let_operator] = ACTIONS(1718),
    [sym_let_and_operator] = ACTIONS(1718),
    [sym__uppercase_identifier] = ACTIONS(1718),
    [aux_sym_directive_token1] = ACTIONS(1716),
    [aux_sym_tag_token1] = ACTIONS(1718),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1048)] = {
    [sym_attribute] = STATE(1048),
    [sym__lowercase_identifier] = ACTIONS(2025),
    [anon_sym_SEMI_SEMI] = ACTIONS(2027),
    [anon_sym_let] = ACTIONS(2025),
    [anon_sym_and] = ACTIONS(2025),
    [anon_sym_TILDE] = ACTIONS(2025),
    [anon_sym_QMARK] = ACTIONS(2025),
    [anon_sym_LPAREN] = ACTIONS(2027),
    [anon_sym_external] = ACTIONS(2025),
    [anon_sym_type] = ACTIONS(2025),
    [anon_sym_COMMA] = ACTIONS(2027),
    [anon_sym_PIPE] = ACTIONS(2025),
    [anon_sym_DOT] = ACTIONS(2027),
    [anon_sym_LBRACK] = ACTIONS(2025),
    [anon_sym_RBRACK] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(2025),
    [anon_sym_false] = ACTIONS(2025),
    [anon_sym_LBRACE] = ACTIONS(2025),
    [anon_sym_SEMI] = ACTIONS(2025),
    [anon_sym_exception] = ACTIONS(2025),
    [anon_sym_module] = ACTIONS(2025),
    [anon_sym_open] = ACTIONS(2025),
    [anon_sym_include] = ACTIONS(2025),
    [anon_sym_class] = ACTIONS(2025),
    [anon_sym_val] = ACTIONS(2025),
    [anon_sym_end] = ACTIONS(2025),
    [anon_sym_object] = ACTIONS(2025),
    [anon_sym_in] = ACTIONS(2025),
    [anon_sym_AMP] = ACTIONS(2025),
    [anon_sym_POUND] = ACTIONS(2025),
    [anon_sym_COLON_COLON] = ACTIONS(2027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2027),
    [anon_sym_else] = ACTIONS(2025),
    [anon_sym_new] = ACTIONS(2025),
    [anon_sym_LBRACE_LT] = ACTIONS(2027),
    [anon_sym_begin] = ACTIONS(2025),
    [sym_ocamlyacc_value] = ACTIONS(2027),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2025),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2027),
    [aux_sym_number_token1] = ACTIONS(2027),
    [aux_sym_signed_number_token1] = ACTIONS(2025),
    [anon_sym_SQUOTE] = ACTIONS(2027),
    [anon_sym_DQUOTE] = ACTIONS(2027),
    [sym_prefix_operator] = ACTIONS(2027),
    [aux_sym_sign_operator_token1] = ACTIONS(2025),
    [sym_hash_operator] = ACTIONS(2027),
    [anon_sym_lsl] = ACTIONS(2025),
    [anon_sym_lsr] = ACTIONS(2025),
    [anon_sym_asr] = ACTIONS(2025),
    [aux_sym_pow_operator_token1] = ACTIONS(2027),
    [anon_sym_mod] = ACTIONS(2025),
    [anon_sym_land] = ACTIONS(2025),
    [anon_sym_lor] = ACTIONS(2025),
    [anon_sym_lxor] = ACTIONS(2025),
    [aux_sym_mult_operator_token1] = ACTIONS(2025),
    [aux_sym_add_operator_token1] = ACTIONS(2025),
    [sym_concat_operator] = ACTIONS(2027),
    [sym_rel_operator] = ACTIONS(2025),
    [anon_sym_AMP_AMP] = ACTIONS(2025),
    [anon_sym_or] = ACTIONS(2025),
    [anon_sym_PIPE_PIPE] = ACTIONS(2025),
    [sym_assign_operator] = ACTIONS(2027),
    [sym_let_operator] = ACTIONS(2027),
    [sym_let_and_operator] = ACTIONS(2027),
    [sym__uppercase_identifier] = ACTIONS(2027),
    [aux_sym_directive_token1] = ACTIONS(2025),
    [aux_sym_tag_token1] = ACTIONS(2027),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1049)] = {
    [sym_attribute] = STATE(1049),
    [sym__lowercase_identifier] = ACTIONS(1698),
    [anon_sym_SEMI_SEMI] = ACTIONS(1700),
    [anon_sym_let] = ACTIONS(1698),
    [anon_sym_and] = ACTIONS(1698),
    [anon_sym_TILDE] = ACTIONS(1698),
    [anon_sym_QMARK] = ACTIONS(1698),
    [anon_sym_LPAREN] = ACTIONS(1700),
    [anon_sym_external] = ACTIONS(1698),
    [anon_sym_type] = ACTIONS(1698),
    [anon_sym_COMMA] = ACTIONS(1700),
    [anon_sym_PIPE] = ACTIONS(1698),
    [anon_sym_DOT] = ACTIONS(1744),
    [anon_sym_LBRACK] = ACTIONS(1698),
    [anon_sym_RBRACK] = ACTIONS(1700),
    [anon_sym_true] = ACTIONS(1698),
    [anon_sym_false] = ACTIONS(1698),
    [anon_sym_LBRACE] = ACTIONS(1698),
    [anon_sym_SEMI] = ACTIONS(1698),
    [anon_sym_exception] = ACTIONS(1698),
    [anon_sym_module] = ACTIONS(1698),
    [anon_sym_open] = ACTIONS(1698),
    [anon_sym_include] = ACTIONS(1698),
    [anon_sym_class] = ACTIONS(1698),
    [anon_sym_val] = ACTIONS(1698),
    [anon_sym_end] = ACTIONS(1698),
    [anon_sym_object] = ACTIONS(1698),
    [anon_sym_in] = ACTIONS(1698),
    [anon_sym_AMP] = ACTIONS(1698),
    [anon_sym_POUND] = ACTIONS(1698),
    [anon_sym_COLON_COLON] = ACTIONS(1700),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1700),
    [anon_sym_else] = ACTIONS(1698),
    [anon_sym_new] = ACTIONS(1698),
    [anon_sym_LBRACE_LT] = ACTIONS(1700),
    [anon_sym_begin] = ACTIONS(1698),
    [sym_ocamlyacc_value] = ACTIONS(1700),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1698),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1700),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1700),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1700),
    [aux_sym_number_token1] = ACTIONS(1700),
    [aux_sym_signed_number_token1] = ACTIONS(1698),
    [anon_sym_SQUOTE] = ACTIONS(1700),
    [anon_sym_DQUOTE] = ACTIONS(1700),
    [sym_prefix_operator] = ACTIONS(1700),
    [aux_sym_sign_operator_token1] = ACTIONS(1698),
    [sym_hash_operator] = ACTIONS(1700),
    [anon_sym_lsl] = ACTIONS(1698),
    [anon_sym_lsr] = ACTIONS(1698),
    [anon_sym_asr] = ACTIONS(1698),
    [aux_sym_pow_operator_token1] = ACTIONS(1700),
    [anon_sym_mod] = ACTIONS(1698),
    [anon_sym_land] = ACTIONS(1698),
    [anon_sym_lor] = ACTIONS(1698),
    [anon_sym_lxor] = ACTIONS(1698),
    [aux_sym_mult_operator_token1] = ACTIONS(1698),
    [aux_sym_add_operator_token1] = ACTIONS(1698),
    [sym_concat_operator] = ACTIONS(1700),
    [sym_rel_operator] = ACTIONS(1698),
    [anon_sym_AMP_AMP] = ACTIONS(1698),
    [anon_sym_or] = ACTIONS(1698),
    [anon_sym_PIPE_PIPE] = ACTIONS(1698),
    [sym_assign_operator] = ACTIONS(1700),
    [sym_let_operator] = ACTIONS(1700),
    [sym_let_and_operator] = ACTIONS(1700),
    [sym__uppercase_identifier] = ACTIONS(1700),
    [aux_sym_directive_token1] = ACTIONS(1698),
    [aux_sym_tag_token1] = ACTIONS(1700),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1050)] = {
    [sym_attribute] = STATE(1050),
    [sym__lowercase_identifier] = ACTIONS(1830),
    [anon_sym_SEMI_SEMI] = ACTIONS(1832),
    [anon_sym_let] = ACTIONS(1830),
    [anon_sym_and] = ACTIONS(1830),
    [anon_sym_TILDE] = ACTIONS(1830),
    [anon_sym_QMARK] = ACTIONS(1830),
    [anon_sym_LPAREN] = ACTIONS(1832),
    [anon_sym_external] = ACTIONS(1830),
    [anon_sym_type] = ACTIONS(1830),
    [anon_sym_COMMA] = ACTIONS(1832),
    [anon_sym_PIPE] = ACTIONS(1830),
    [anon_sym_DOT] = ACTIONS(1832),
    [anon_sym_LBRACK] = ACTIONS(1830),
    [anon_sym_RBRACK] = ACTIONS(1832),
    [anon_sym_true] = ACTIONS(1830),
    [anon_sym_false] = ACTIONS(1830),
    [anon_sym_LBRACE] = ACTIONS(1830),
    [anon_sym_SEMI] = ACTIONS(1830),
    [anon_sym_exception] = ACTIONS(1830),
    [anon_sym_module] = ACTIONS(1830),
    [anon_sym_open] = ACTIONS(1830),
    [anon_sym_include] = ACTIONS(1830),
    [anon_sym_class] = ACTIONS(1830),
    [anon_sym_val] = ACTIONS(1830),
    [anon_sym_end] = ACTIONS(1830),
    [anon_sym_object] = ACTIONS(1830),
    [anon_sym_in] = ACTIONS(1830),
    [anon_sym_AMP] = ACTIONS(1830),
    [anon_sym_POUND] = ACTIONS(1830),
    [anon_sym_COLON_COLON] = ACTIONS(1832),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1832),
    [anon_sym_else] = ACTIONS(1830),
    [anon_sym_new] = ACTIONS(1830),
    [anon_sym_LBRACE_LT] = ACTIONS(1832),
    [anon_sym_begin] = ACTIONS(1830),
    [sym_ocamlyacc_value] = ACTIONS(1832),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1830),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1832),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1830),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1830),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1832),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1832),
    [aux_sym_number_token1] = ACTIONS(1832),
    [aux_sym_signed_number_token1] = ACTIONS(1830),
    [anon_sym_SQUOTE] = ACTIONS(1832),
    [anon_sym_DQUOTE] = ACTIONS(1832),
    [sym_prefix_operator] = ACTIONS(1832),
    [aux_sym_sign_operator_token1] = ACTIONS(1830),
    [sym_hash_operator] = ACTIONS(1832),
    [anon_sym_lsl] = ACTIONS(1830),
    [anon_sym_lsr] = ACTIONS(1830),
    [anon_sym_asr] = ACTIONS(1830),
    [aux_sym_pow_operator_token1] = ACTIONS(1832),
    [anon_sym_mod] = ACTIONS(1830),
    [anon_sym_land] = ACTIONS(1830),
    [anon_sym_lor] = ACTIONS(1830),
    [anon_sym_lxor] = ACTIONS(1830),
    [aux_sym_mult_operator_token1] = ACTIONS(1830),
    [aux_sym_add_operator_token1] = ACTIONS(1830),
    [sym_concat_operator] = ACTIONS(1832),
    [sym_rel_operator] = ACTIONS(1830),
    [anon_sym_AMP_AMP] = ACTIONS(1830),
    [anon_sym_or] = ACTIONS(1830),
    [anon_sym_PIPE_PIPE] = ACTIONS(1830),
    [sym_assign_operator] = ACTIONS(1832),
    [sym_let_operator] = ACTIONS(1832),
    [sym_let_and_operator] = ACTIONS(1832),
    [sym__uppercase_identifier] = ACTIONS(1832),
    [aux_sym_directive_token1] = ACTIONS(1830),
    [aux_sym_tag_token1] = ACTIONS(1832),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1051)] = {
    [sym_attribute] = STATE(1051),
    [sym__lowercase_identifier] = ACTIONS(2161),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_QMARK] = ACTIONS(2161),
    [anon_sym_COLON] = ACTIONS(2232),
    [anon_sym_LPAREN] = ACTIONS(2165),
    [anon_sym_RPAREN] = ACTIONS(2165),
    [anon_sym_COMMA] = ACTIONS(2165),
    [anon_sym_PIPE] = ACTIONS(2161),
    [anon_sym_COLON2] = ACTIONS(2161),
    [anon_sym_DASH_GT] = ACTIONS(2161),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [anon_sym_RBRACK] = ACTIONS(2165),
    [anon_sym_true] = ACTIONS(2161),
    [anon_sym_false] = ACTIONS(2161),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [anon_sym_SEMI] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2165),
    [anon_sym_constraint] = ACTIONS(2161),
    [anon_sym_val] = ACTIONS(2161),
    [anon_sym_end] = ACTIONS(2161),
    [anon_sym_with] = ACTIONS(2161),
    [anon_sym_COLON_GT] = ACTIONS(2165),
    [anon_sym_object] = ACTIONS(2161),
    [anon_sym_inherit] = ACTIONS(2161),
    [anon_sym_method] = ACTIONS(2161),
    [anon_sym_initializer] = ACTIONS(2161),
    [anon_sym_AMP] = ACTIONS(2161),
    [anon_sym_COLON_COLON] = ACTIONS(2165),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2165),
    [anon_sym_then] = ACTIONS(2161),
    [anon_sym_else] = ACTIONS(2161),
    [anon_sym_do] = ACTIONS(2161),
    [anon_sym_done] = ACTIONS(2161),
    [anon_sym_to] = ACTIONS(2161),
    [anon_sym_downto] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2161),
    [anon_sym_LBRACE_LT] = ACTIONS(2165),
    [anon_sym_GT_RBRACE] = ACTIONS(2165),
    [anon_sym_begin] = ACTIONS(2161),
    [sym_ocamlyacc_value] = ACTIONS(2165),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2161),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2165),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2165),
    [aux_sym_number_token1] = ACTIONS(2165),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2165),
    [anon_sym_DQUOTE] = ACTIONS(2165),
    [sym_prefix_operator] = ACTIONS(2165),
    [aux_sym_sign_operator_token1] = ACTIONS(2161),
    [anon_sym_lsl] = ACTIONS(2161),
    [anon_sym_lsr] = ACTIONS(2161),
    [anon_sym_asr] = ACTIONS(2161),
    [aux_sym_pow_operator_token1] = ACTIONS(2165),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_land] = ACTIONS(2161),
    [anon_sym_lor] = ACTIONS(2161),
    [anon_sym_lxor] = ACTIONS(2161),
    [aux_sym_mult_operator_token1] = ACTIONS(2161),
    [aux_sym_add_operator_token1] = ACTIONS(2161),
    [sym_concat_operator] = ACTIONS(2165),
    [sym_rel_operator] = ACTIONS(2161),
    [anon_sym_AMP_AMP] = ACTIONS(2161),
    [anon_sym_or] = ACTIONS(2161),
    [anon_sym_PIPE_PIPE] = ACTIONS(2161),
    [sym_assign_operator] = ACTIONS(2165),
    [sym__uppercase_identifier] = ACTIONS(2165),
    [aux_sym_tag_token1] = ACTIONS(2165),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1052)] = {
    [sym_attribute] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(1724),
    [anon_sym_SEMI_SEMI] = ACTIONS(1726),
    [anon_sym_let] = ACTIONS(1724),
    [anon_sym_and] = ACTIONS(1724),
    [anon_sym_TILDE] = ACTIONS(1724),
    [anon_sym_QMARK] = ACTIONS(1724),
    [anon_sym_LPAREN] = ACTIONS(1726),
    [anon_sym_external] = ACTIONS(1724),
    [anon_sym_type] = ACTIONS(1724),
    [anon_sym_COMMA] = ACTIONS(1726),
    [anon_sym_PIPE] = ACTIONS(1724),
    [anon_sym_DOT] = ACTIONS(1726),
    [anon_sym_LBRACK] = ACTIONS(1724),
    [anon_sym_RBRACK] = ACTIONS(1726),
    [anon_sym_true] = ACTIONS(1724),
    [anon_sym_false] = ACTIONS(1724),
    [anon_sym_LBRACE] = ACTIONS(1724),
    [anon_sym_SEMI] = ACTIONS(1724),
    [anon_sym_exception] = ACTIONS(1724),
    [anon_sym_module] = ACTIONS(1724),
    [anon_sym_open] = ACTIONS(1724),
    [anon_sym_include] = ACTIONS(1724),
    [anon_sym_class] = ACTIONS(1724),
    [anon_sym_val] = ACTIONS(1724),
    [anon_sym_end] = ACTIONS(1724),
    [anon_sym_object] = ACTIONS(1724),
    [anon_sym_AMP] = ACTIONS(1724),
    [anon_sym_POUND] = ACTIONS(1724),
    [anon_sym_COLON_COLON] = ACTIONS(1726),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1726),
    [anon_sym_LT_DASH] = ACTIONS(2234),
    [anon_sym_else] = ACTIONS(1724),
    [anon_sym_new] = ACTIONS(1724),
    [anon_sym_LBRACE_LT] = ACTIONS(1726),
    [anon_sym_begin] = ACTIONS(1724),
    [sym_ocamlyacc_value] = ACTIONS(1726),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1724),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1726),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1726),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1726),
    [aux_sym_number_token1] = ACTIONS(1726),
    [aux_sym_signed_number_token1] = ACTIONS(1724),
    [anon_sym_SQUOTE] = ACTIONS(1726),
    [anon_sym_DQUOTE] = ACTIONS(1726),
    [sym_prefix_operator] = ACTIONS(1726),
    [aux_sym_sign_operator_token1] = ACTIONS(1724),
    [sym_hash_operator] = ACTIONS(1726),
    [anon_sym_lsl] = ACTIONS(1724),
    [anon_sym_lsr] = ACTIONS(1724),
    [anon_sym_asr] = ACTIONS(1724),
    [aux_sym_pow_operator_token1] = ACTIONS(1726),
    [anon_sym_mod] = ACTIONS(1724),
    [anon_sym_land] = ACTIONS(1724),
    [anon_sym_lor] = ACTIONS(1724),
    [anon_sym_lxor] = ACTIONS(1724),
    [aux_sym_mult_operator_token1] = ACTIONS(1724),
    [aux_sym_add_operator_token1] = ACTIONS(1724),
    [sym_concat_operator] = ACTIONS(1726),
    [sym_rel_operator] = ACTIONS(1724),
    [anon_sym_AMP_AMP] = ACTIONS(1724),
    [anon_sym_or] = ACTIONS(1724),
    [anon_sym_PIPE_PIPE] = ACTIONS(1724),
    [sym_assign_operator] = ACTIONS(1726),
    [sym_let_operator] = ACTIONS(1726),
    [sym_let_and_operator] = ACTIONS(1726),
    [sym__uppercase_identifier] = ACTIONS(1726),
    [aux_sym_directive_token1] = ACTIONS(1724),
    [aux_sym_tag_token1] = ACTIONS(1726),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1053)] = {
    [sym_attribute] = STATE(1053),
    [sym__lowercase_identifier] = ACTIONS(2123),
    [anon_sym_TILDE] = ACTIONS(2123),
    [anon_sym_QMARK] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(2125),
    [anon_sym_RPAREN] = ACTIONS(2125),
    [anon_sym_COMMA] = ACTIONS(2125),
    [anon_sym_PIPE] = ACTIONS(2123),
    [anon_sym_COLON2] = ACTIONS(2123),
    [anon_sym_DASH_GT] = ACTIONS(2123),
    [anon_sym_LBRACK] = ACTIONS(2123),
    [anon_sym_RBRACK] = ACTIONS(2125),
    [anon_sym_true] = ACTIONS(2123),
    [anon_sym_false] = ACTIONS(2123),
    [anon_sym_LBRACE] = ACTIONS(2123),
    [anon_sym_SEMI] = ACTIONS(2125),
    [anon_sym_RBRACE] = ACTIONS(2125),
    [anon_sym_constraint] = ACTIONS(2123),
    [anon_sym_val] = ACTIONS(2123),
    [anon_sym_end] = ACTIONS(2123),
    [anon_sym_with] = ACTIONS(2123),
    [anon_sym_COLON_GT] = ACTIONS(2125),
    [anon_sym_object] = ACTIONS(2123),
    [anon_sym_inherit] = ACTIONS(2123),
    [anon_sym_method] = ACTIONS(2123),
    [anon_sym_initializer] = ACTIONS(2123),
    [anon_sym_AMP] = ACTIONS(2123),
    [anon_sym_COLON_COLON] = ACTIONS(2125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2125),
    [anon_sym_then] = ACTIONS(2123),
    [anon_sym_else] = ACTIONS(2123),
    [anon_sym_do] = ACTIONS(2123),
    [anon_sym_done] = ACTIONS(2123),
    [anon_sym_to] = ACTIONS(2123),
    [anon_sym_downto] = ACTIONS(2123),
    [anon_sym_new] = ACTIONS(2123),
    [anon_sym_LBRACE_LT] = ACTIONS(2125),
    [anon_sym_GT_RBRACE] = ACTIONS(2125),
    [anon_sym_begin] = ACTIONS(2123),
    [sym_ocamlyacc_value] = ACTIONS(2125),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2123),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2125),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2125),
    [aux_sym_number_token1] = ACTIONS(2125),
    [aux_sym_signed_number_token1] = ACTIONS(2123),
    [anon_sym_SQUOTE] = ACTIONS(2125),
    [anon_sym_DQUOTE] = ACTIONS(2125),
    [sym_prefix_operator] = ACTIONS(2125),
    [aux_sym_sign_operator_token1] = ACTIONS(2123),
    [anon_sym_lsl] = ACTIONS(2123),
    [anon_sym_lsr] = ACTIONS(2123),
    [anon_sym_asr] = ACTIONS(2123),
    [aux_sym_pow_operator_token1] = ACTIONS(2125),
    [anon_sym_mod] = ACTIONS(2123),
    [anon_sym_land] = ACTIONS(2123),
    [anon_sym_lor] = ACTIONS(2123),
    [anon_sym_lxor] = ACTIONS(2123),
    [aux_sym_mult_operator_token1] = ACTIONS(2123),
    [aux_sym_add_operator_token1] = ACTIONS(2123),
    [sym_concat_operator] = ACTIONS(2125),
    [sym_rel_operator] = ACTIONS(2123),
    [anon_sym_AMP_AMP] = ACTIONS(2123),
    [anon_sym_or] = ACTIONS(2123),
    [anon_sym_PIPE_PIPE] = ACTIONS(2123),
    [sym_assign_operator] = ACTIONS(2125),
    [sym__uppercase_identifier] = ACTIONS(2125),
    [aux_sym_tag_token1] = ACTIONS(2125),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1054)] = {
    [sym_attribute] = STATE(1054),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_RPAREN] = ACTIONS(2081),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_COLON2] = ACTIONS(2079),
    [anon_sym_DASH_GT] = ACTIONS(2079),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2081),
    [anon_sym_RBRACE] = ACTIONS(2081),
    [anon_sym_constraint] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_with] = ACTIONS(2079),
    [anon_sym_COLON_GT] = ACTIONS(2081),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_inherit] = ACTIONS(2079),
    [anon_sym_method] = ACTIONS(2079),
    [anon_sym_initializer] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_then] = ACTIONS(2079),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_do] = ACTIONS(2079),
    [anon_sym_done] = ACTIONS(2079),
    [anon_sym_to] = ACTIONS(2079),
    [anon_sym_downto] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_GT_RBRACE] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1055)] = {
    [sym_attribute] = STATE(1055),
    [sym__lowercase_identifier] = ACTIONS(1566),
    [anon_sym_SEMI_SEMI] = ACTIONS(1568),
    [anon_sym_let] = ACTIONS(1566),
    [anon_sym_and] = ACTIONS(1566),
    [anon_sym_TILDE] = ACTIONS(1566),
    [anon_sym_QMARK] = ACTIONS(1566),
    [anon_sym_LPAREN] = ACTIONS(1568),
    [anon_sym_external] = ACTIONS(1566),
    [anon_sym_type] = ACTIONS(1566),
    [anon_sym_COMMA] = ACTIONS(1568),
    [anon_sym_PIPE] = ACTIONS(1566),
    [anon_sym_DOT] = ACTIONS(1568),
    [anon_sym_LBRACK] = ACTIONS(1566),
    [anon_sym_RBRACK] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1566),
    [anon_sym_false] = ACTIONS(1566),
    [anon_sym_LBRACE] = ACTIONS(1566),
    [anon_sym_SEMI] = ACTIONS(1566),
    [anon_sym_exception] = ACTIONS(1566),
    [anon_sym_module] = ACTIONS(1566),
    [anon_sym_open] = ACTIONS(1566),
    [anon_sym_include] = ACTIONS(1566),
    [anon_sym_class] = ACTIONS(1566),
    [anon_sym_val] = ACTIONS(1566),
    [anon_sym_end] = ACTIONS(1566),
    [anon_sym_object] = ACTIONS(1566),
    [anon_sym_AMP] = ACTIONS(1566),
    [anon_sym_POUND] = ACTIONS(1566),
    [anon_sym_COLON_COLON] = ACTIONS(1568),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1568),
    [anon_sym_LT_DASH] = ACTIONS(1566),
    [anon_sym_else] = ACTIONS(1566),
    [anon_sym_new] = ACTIONS(1566),
    [anon_sym_LBRACE_LT] = ACTIONS(1568),
    [anon_sym_begin] = ACTIONS(1566),
    [sym_ocamlyacc_value] = ACTIONS(1568),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1566),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1568),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1566),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1566),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1568),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1568),
    [aux_sym_number_token1] = ACTIONS(1568),
    [aux_sym_signed_number_token1] = ACTIONS(1566),
    [anon_sym_SQUOTE] = ACTIONS(1568),
    [anon_sym_DQUOTE] = ACTIONS(1568),
    [sym_prefix_operator] = ACTIONS(1568),
    [aux_sym_sign_operator_token1] = ACTIONS(1566),
    [sym_hash_operator] = ACTIONS(1568),
    [anon_sym_lsl] = ACTIONS(1566),
    [anon_sym_lsr] = ACTIONS(1566),
    [anon_sym_asr] = ACTIONS(1566),
    [aux_sym_pow_operator_token1] = ACTIONS(1568),
    [anon_sym_mod] = ACTIONS(1566),
    [anon_sym_land] = ACTIONS(1566),
    [anon_sym_lor] = ACTIONS(1566),
    [anon_sym_lxor] = ACTIONS(1566),
    [aux_sym_mult_operator_token1] = ACTIONS(1566),
    [aux_sym_add_operator_token1] = ACTIONS(1566),
    [sym_concat_operator] = ACTIONS(1568),
    [sym_rel_operator] = ACTIONS(1566),
    [anon_sym_AMP_AMP] = ACTIONS(1566),
    [anon_sym_or] = ACTIONS(1566),
    [anon_sym_PIPE_PIPE] = ACTIONS(1566),
    [sym_assign_operator] = ACTIONS(1568),
    [sym_let_operator] = ACTIONS(1568),
    [sym__uppercase_identifier] = ACTIONS(1568),
    [aux_sym_directive_token1] = ACTIONS(1566),
    [aux_sym_tag_token1] = ACTIONS(1568),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1056)] = {
    [sym_attribute] = STATE(1056),
    [sym__lowercase_identifier] = ACTIONS(1578),
    [anon_sym_SEMI_SEMI] = ACTIONS(1580),
    [anon_sym_let] = ACTIONS(1578),
    [anon_sym_and] = ACTIONS(1578),
    [anon_sym_TILDE] = ACTIONS(1578),
    [anon_sym_QMARK] = ACTIONS(1578),
    [anon_sym_LPAREN] = ACTIONS(1580),
    [anon_sym_external] = ACTIONS(1578),
    [anon_sym_type] = ACTIONS(1578),
    [anon_sym_COMMA] = ACTIONS(1580),
    [anon_sym_PIPE] = ACTIONS(1578),
    [anon_sym_DOT] = ACTIONS(1580),
    [anon_sym_LBRACK] = ACTIONS(1578),
    [anon_sym_RBRACK] = ACTIONS(1580),
    [anon_sym_true] = ACTIONS(1578),
    [anon_sym_false] = ACTIONS(1578),
    [anon_sym_LBRACE] = ACTIONS(1578),
    [anon_sym_SEMI] = ACTIONS(1578),
    [anon_sym_exception] = ACTIONS(1578),
    [anon_sym_module] = ACTIONS(1578),
    [anon_sym_open] = ACTIONS(1578),
    [anon_sym_include] = ACTIONS(1578),
    [anon_sym_class] = ACTIONS(1578),
    [anon_sym_val] = ACTIONS(1578),
    [anon_sym_end] = ACTIONS(1578),
    [anon_sym_object] = ACTIONS(1578),
    [anon_sym_AMP] = ACTIONS(1578),
    [anon_sym_POUND] = ACTIONS(1578),
    [anon_sym_COLON_COLON] = ACTIONS(1580),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1580),
    [anon_sym_LT_DASH] = ACTIONS(1578),
    [anon_sym_else] = ACTIONS(1578),
    [anon_sym_new] = ACTIONS(1578),
    [anon_sym_LBRACE_LT] = ACTIONS(1580),
    [anon_sym_begin] = ACTIONS(1578),
    [sym_ocamlyacc_value] = ACTIONS(1580),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1578),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1580),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1578),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1578),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1580),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1580),
    [aux_sym_number_token1] = ACTIONS(1580),
    [aux_sym_signed_number_token1] = ACTIONS(1578),
    [anon_sym_SQUOTE] = ACTIONS(1580),
    [anon_sym_DQUOTE] = ACTIONS(1580),
    [sym_prefix_operator] = ACTIONS(1580),
    [aux_sym_sign_operator_token1] = ACTIONS(1578),
    [sym_hash_operator] = ACTIONS(1580),
    [anon_sym_lsl] = ACTIONS(1578),
    [anon_sym_lsr] = ACTIONS(1578),
    [anon_sym_asr] = ACTIONS(1578),
    [aux_sym_pow_operator_token1] = ACTIONS(1580),
    [anon_sym_mod] = ACTIONS(1578),
    [anon_sym_land] = ACTIONS(1578),
    [anon_sym_lor] = ACTIONS(1578),
    [anon_sym_lxor] = ACTIONS(1578),
    [aux_sym_mult_operator_token1] = ACTIONS(1578),
    [aux_sym_add_operator_token1] = ACTIONS(1578),
    [sym_concat_operator] = ACTIONS(1580),
    [sym_rel_operator] = ACTIONS(1578),
    [anon_sym_AMP_AMP] = ACTIONS(1578),
    [anon_sym_or] = ACTIONS(1578),
    [anon_sym_PIPE_PIPE] = ACTIONS(1578),
    [sym_assign_operator] = ACTIONS(1580),
    [sym_let_operator] = ACTIONS(1580),
    [sym__uppercase_identifier] = ACTIONS(1580),
    [aux_sym_directive_token1] = ACTIONS(1578),
    [aux_sym_tag_token1] = ACTIONS(1580),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1057)] = {
    [sym__parenthesized_abstract_type] = STATE(1340),
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4549),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1057),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2238),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_RBRACK] = ACTIONS(1316),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_SEMI] = ACTIONS(1316),
    [anon_sym_RBRACE] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_as] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_COLON_COLON] = ACTIONS(1316),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1058)] = {
    [sym_attribute] = STATE(1058),
    [sym__lowercase_identifier] = ACTIONS(1554),
    [anon_sym_SEMI_SEMI] = ACTIONS(1556),
    [anon_sym_let] = ACTIONS(1554),
    [anon_sym_and] = ACTIONS(1554),
    [anon_sym_TILDE] = ACTIONS(1554),
    [anon_sym_QMARK] = ACTIONS(1554),
    [anon_sym_LPAREN] = ACTIONS(1556),
    [anon_sym_external] = ACTIONS(1554),
    [anon_sym_type] = ACTIONS(1554),
    [anon_sym_COMMA] = ACTIONS(1556),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_DOT] = ACTIONS(1556),
    [anon_sym_LBRACK] = ACTIONS(1554),
    [anon_sym_RBRACK] = ACTIONS(1556),
    [anon_sym_true] = ACTIONS(1554),
    [anon_sym_false] = ACTIONS(1554),
    [anon_sym_LBRACE] = ACTIONS(1554),
    [anon_sym_SEMI] = ACTIONS(1554),
    [anon_sym_exception] = ACTIONS(1554),
    [anon_sym_module] = ACTIONS(1554),
    [anon_sym_open] = ACTIONS(1554),
    [anon_sym_include] = ACTIONS(1554),
    [anon_sym_class] = ACTIONS(1554),
    [anon_sym_val] = ACTIONS(1554),
    [anon_sym_end] = ACTIONS(1554),
    [anon_sym_object] = ACTIONS(1554),
    [anon_sym_AMP] = ACTIONS(1554),
    [anon_sym_POUND] = ACTIONS(1554),
    [anon_sym_COLON_COLON] = ACTIONS(1556),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1556),
    [anon_sym_LT_DASH] = ACTIONS(1554),
    [anon_sym_else] = ACTIONS(1554),
    [anon_sym_new] = ACTIONS(1554),
    [anon_sym_LBRACE_LT] = ACTIONS(1556),
    [anon_sym_begin] = ACTIONS(1554),
    [sym_ocamlyacc_value] = ACTIONS(1556),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1554),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1556),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1554),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1554),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1556),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1556),
    [aux_sym_number_token1] = ACTIONS(1556),
    [aux_sym_signed_number_token1] = ACTIONS(1554),
    [anon_sym_SQUOTE] = ACTIONS(1556),
    [anon_sym_DQUOTE] = ACTIONS(1556),
    [sym_prefix_operator] = ACTIONS(1556),
    [aux_sym_sign_operator_token1] = ACTIONS(1554),
    [sym_hash_operator] = ACTIONS(1556),
    [anon_sym_lsl] = ACTIONS(1554),
    [anon_sym_lsr] = ACTIONS(1554),
    [anon_sym_asr] = ACTIONS(1554),
    [aux_sym_pow_operator_token1] = ACTIONS(1556),
    [anon_sym_mod] = ACTIONS(1554),
    [anon_sym_land] = ACTIONS(1554),
    [anon_sym_lor] = ACTIONS(1554),
    [anon_sym_lxor] = ACTIONS(1554),
    [aux_sym_mult_operator_token1] = ACTIONS(1554),
    [aux_sym_add_operator_token1] = ACTIONS(1554),
    [sym_concat_operator] = ACTIONS(1556),
    [sym_rel_operator] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1554),
    [anon_sym_or] = ACTIONS(1554),
    [anon_sym_PIPE_PIPE] = ACTIONS(1554),
    [sym_assign_operator] = ACTIONS(1556),
    [sym_let_operator] = ACTIONS(1556),
    [sym__uppercase_identifier] = ACTIONS(1556),
    [aux_sym_directive_token1] = ACTIONS(1554),
    [aux_sym_tag_token1] = ACTIONS(1556),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1059)] = {
    [sym_attribute] = STATE(1059),
    [sym__lowercase_identifier] = ACTIONS(1550),
    [anon_sym_SEMI_SEMI] = ACTIONS(1552),
    [anon_sym_let] = ACTIONS(1550),
    [anon_sym_and] = ACTIONS(1550),
    [anon_sym_TILDE] = ACTIONS(1550),
    [anon_sym_QMARK] = ACTIONS(1550),
    [anon_sym_LPAREN] = ACTIONS(1552),
    [anon_sym_external] = ACTIONS(1550),
    [anon_sym_type] = ACTIONS(1550),
    [anon_sym_COMMA] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1550),
    [anon_sym_DOT] = ACTIONS(1552),
    [anon_sym_LBRACK] = ACTIONS(1550),
    [anon_sym_RBRACK] = ACTIONS(1552),
    [anon_sym_true] = ACTIONS(1550),
    [anon_sym_false] = ACTIONS(1550),
    [anon_sym_LBRACE] = ACTIONS(1550),
    [anon_sym_SEMI] = ACTIONS(1550),
    [anon_sym_exception] = ACTIONS(1550),
    [anon_sym_module] = ACTIONS(1550),
    [anon_sym_open] = ACTIONS(1550),
    [anon_sym_include] = ACTIONS(1550),
    [anon_sym_class] = ACTIONS(1550),
    [anon_sym_val] = ACTIONS(1550),
    [anon_sym_end] = ACTIONS(1550),
    [anon_sym_object] = ACTIONS(1550),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_POUND] = ACTIONS(1550),
    [anon_sym_COLON_COLON] = ACTIONS(1552),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1552),
    [anon_sym_LT_DASH] = ACTIONS(1550),
    [anon_sym_else] = ACTIONS(1550),
    [anon_sym_new] = ACTIONS(1550),
    [anon_sym_LBRACE_LT] = ACTIONS(1552),
    [anon_sym_begin] = ACTIONS(1550),
    [sym_ocamlyacc_value] = ACTIONS(1552),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1550),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1552),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1550),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1550),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1552),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1552),
    [aux_sym_number_token1] = ACTIONS(1552),
    [aux_sym_signed_number_token1] = ACTIONS(1550),
    [anon_sym_SQUOTE] = ACTIONS(1552),
    [anon_sym_DQUOTE] = ACTIONS(1552),
    [sym_prefix_operator] = ACTIONS(1552),
    [aux_sym_sign_operator_token1] = ACTIONS(1550),
    [sym_hash_operator] = ACTIONS(1552),
    [anon_sym_lsl] = ACTIONS(1550),
    [anon_sym_lsr] = ACTIONS(1550),
    [anon_sym_asr] = ACTIONS(1550),
    [aux_sym_pow_operator_token1] = ACTIONS(1552),
    [anon_sym_mod] = ACTIONS(1550),
    [anon_sym_land] = ACTIONS(1550),
    [anon_sym_lor] = ACTIONS(1550),
    [anon_sym_lxor] = ACTIONS(1550),
    [aux_sym_mult_operator_token1] = ACTIONS(1550),
    [aux_sym_add_operator_token1] = ACTIONS(1550),
    [sym_concat_operator] = ACTIONS(1552),
    [sym_rel_operator] = ACTIONS(1550),
    [anon_sym_AMP_AMP] = ACTIONS(1550),
    [anon_sym_or] = ACTIONS(1550),
    [anon_sym_PIPE_PIPE] = ACTIONS(1550),
    [sym_assign_operator] = ACTIONS(1552),
    [sym_let_operator] = ACTIONS(1552),
    [sym__uppercase_identifier] = ACTIONS(1552),
    [aux_sym_directive_token1] = ACTIONS(1550),
    [aux_sym_tag_token1] = ACTIONS(1552),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1060)] = {
    [sym_attribute] = STATE(1060),
    [sym__lowercase_identifier] = ACTIONS(1570),
    [anon_sym_SEMI_SEMI] = ACTIONS(1572),
    [anon_sym_let] = ACTIONS(1570),
    [anon_sym_and] = ACTIONS(1570),
    [anon_sym_TILDE] = ACTIONS(1570),
    [anon_sym_QMARK] = ACTIONS(1570),
    [anon_sym_LPAREN] = ACTIONS(1572),
    [anon_sym_external] = ACTIONS(1570),
    [anon_sym_type] = ACTIONS(1570),
    [anon_sym_COMMA] = ACTIONS(1572),
    [anon_sym_PIPE] = ACTIONS(1570),
    [anon_sym_DOT] = ACTIONS(1572),
    [anon_sym_LBRACK] = ACTIONS(1570),
    [anon_sym_RBRACK] = ACTIONS(1572),
    [anon_sym_true] = ACTIONS(1570),
    [anon_sym_false] = ACTIONS(1570),
    [anon_sym_LBRACE] = ACTIONS(1570),
    [anon_sym_SEMI] = ACTIONS(1570),
    [anon_sym_exception] = ACTIONS(1570),
    [anon_sym_module] = ACTIONS(1570),
    [anon_sym_open] = ACTIONS(1570),
    [anon_sym_include] = ACTIONS(1570),
    [anon_sym_class] = ACTIONS(1570),
    [anon_sym_val] = ACTIONS(1570),
    [anon_sym_end] = ACTIONS(1570),
    [anon_sym_object] = ACTIONS(1570),
    [anon_sym_AMP] = ACTIONS(1570),
    [anon_sym_POUND] = ACTIONS(1570),
    [anon_sym_COLON_COLON] = ACTIONS(1572),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1572),
    [anon_sym_LT_DASH] = ACTIONS(1570),
    [anon_sym_else] = ACTIONS(1570),
    [anon_sym_new] = ACTIONS(1570),
    [anon_sym_LBRACE_LT] = ACTIONS(1572),
    [anon_sym_begin] = ACTIONS(1570),
    [sym_ocamlyacc_value] = ACTIONS(1572),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1570),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1572),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1570),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1570),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1572),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1572),
    [aux_sym_number_token1] = ACTIONS(1572),
    [aux_sym_signed_number_token1] = ACTIONS(1570),
    [anon_sym_SQUOTE] = ACTIONS(1572),
    [anon_sym_DQUOTE] = ACTIONS(1572),
    [sym_prefix_operator] = ACTIONS(1572),
    [aux_sym_sign_operator_token1] = ACTIONS(1570),
    [sym_hash_operator] = ACTIONS(1572),
    [anon_sym_lsl] = ACTIONS(1570),
    [anon_sym_lsr] = ACTIONS(1570),
    [anon_sym_asr] = ACTIONS(1570),
    [aux_sym_pow_operator_token1] = ACTIONS(1572),
    [anon_sym_mod] = ACTIONS(1570),
    [anon_sym_land] = ACTIONS(1570),
    [anon_sym_lor] = ACTIONS(1570),
    [anon_sym_lxor] = ACTIONS(1570),
    [aux_sym_mult_operator_token1] = ACTIONS(1570),
    [aux_sym_add_operator_token1] = ACTIONS(1570),
    [sym_concat_operator] = ACTIONS(1572),
    [sym_rel_operator] = ACTIONS(1570),
    [anon_sym_AMP_AMP] = ACTIONS(1570),
    [anon_sym_or] = ACTIONS(1570),
    [anon_sym_PIPE_PIPE] = ACTIONS(1570),
    [sym_assign_operator] = ACTIONS(1572),
    [sym_let_operator] = ACTIONS(1572),
    [sym__uppercase_identifier] = ACTIONS(1572),
    [aux_sym_directive_token1] = ACTIONS(1570),
    [aux_sym_tag_token1] = ACTIONS(1572),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1061)] = {
    [sym_attribute] = STATE(1061),
    [sym__lowercase_identifier] = ACTIONS(1546),
    [anon_sym_SEMI_SEMI] = ACTIONS(1548),
    [anon_sym_let] = ACTIONS(1546),
    [anon_sym_and] = ACTIONS(1546),
    [anon_sym_TILDE] = ACTIONS(1546),
    [anon_sym_QMARK] = ACTIONS(1546),
    [anon_sym_LPAREN] = ACTIONS(1548),
    [anon_sym_external] = ACTIONS(1546),
    [anon_sym_type] = ACTIONS(1546),
    [anon_sym_COMMA] = ACTIONS(1548),
    [anon_sym_PIPE] = ACTIONS(1546),
    [anon_sym_DOT] = ACTIONS(1548),
    [anon_sym_LBRACK] = ACTIONS(1546),
    [anon_sym_RBRACK] = ACTIONS(1548),
    [anon_sym_true] = ACTIONS(1546),
    [anon_sym_false] = ACTIONS(1546),
    [anon_sym_LBRACE] = ACTIONS(1546),
    [anon_sym_SEMI] = ACTIONS(1546),
    [anon_sym_exception] = ACTIONS(1546),
    [anon_sym_module] = ACTIONS(1546),
    [anon_sym_open] = ACTIONS(1546),
    [anon_sym_include] = ACTIONS(1546),
    [anon_sym_class] = ACTIONS(1546),
    [anon_sym_val] = ACTIONS(1546),
    [anon_sym_end] = ACTIONS(1546),
    [anon_sym_object] = ACTIONS(1546),
    [anon_sym_AMP] = ACTIONS(1546),
    [anon_sym_POUND] = ACTIONS(1546),
    [anon_sym_COLON_COLON] = ACTIONS(1548),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1548),
    [anon_sym_LT_DASH] = ACTIONS(1546),
    [anon_sym_else] = ACTIONS(1546),
    [anon_sym_new] = ACTIONS(1546),
    [anon_sym_LBRACE_LT] = ACTIONS(1548),
    [anon_sym_begin] = ACTIONS(1546),
    [sym_ocamlyacc_value] = ACTIONS(1548),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1546),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1548),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1546),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1546),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1548),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1548),
    [aux_sym_number_token1] = ACTIONS(1548),
    [aux_sym_signed_number_token1] = ACTIONS(1546),
    [anon_sym_SQUOTE] = ACTIONS(1548),
    [anon_sym_DQUOTE] = ACTIONS(1548),
    [sym_prefix_operator] = ACTIONS(1548),
    [aux_sym_sign_operator_token1] = ACTIONS(1546),
    [sym_hash_operator] = ACTIONS(1548),
    [anon_sym_lsl] = ACTIONS(1546),
    [anon_sym_lsr] = ACTIONS(1546),
    [anon_sym_asr] = ACTIONS(1546),
    [aux_sym_pow_operator_token1] = ACTIONS(1548),
    [anon_sym_mod] = ACTIONS(1546),
    [anon_sym_land] = ACTIONS(1546),
    [anon_sym_lor] = ACTIONS(1546),
    [anon_sym_lxor] = ACTIONS(1546),
    [aux_sym_mult_operator_token1] = ACTIONS(1546),
    [aux_sym_add_operator_token1] = ACTIONS(1546),
    [sym_concat_operator] = ACTIONS(1548),
    [sym_rel_operator] = ACTIONS(1546),
    [anon_sym_AMP_AMP] = ACTIONS(1546),
    [anon_sym_or] = ACTIONS(1546),
    [anon_sym_PIPE_PIPE] = ACTIONS(1546),
    [sym_assign_operator] = ACTIONS(1548),
    [sym_let_operator] = ACTIONS(1548),
    [sym__uppercase_identifier] = ACTIONS(1548),
    [aux_sym_directive_token1] = ACTIONS(1546),
    [aux_sym_tag_token1] = ACTIONS(1548),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1062)] = {
    [sym_attribute] = STATE(1062),
    [sym__lowercase_identifier] = ACTIONS(1574),
    [anon_sym_SEMI_SEMI] = ACTIONS(1576),
    [anon_sym_let] = ACTIONS(1574),
    [anon_sym_and] = ACTIONS(1574),
    [anon_sym_TILDE] = ACTIONS(1574),
    [anon_sym_QMARK] = ACTIONS(1574),
    [anon_sym_LPAREN] = ACTIONS(1576),
    [anon_sym_external] = ACTIONS(1574),
    [anon_sym_type] = ACTIONS(1574),
    [anon_sym_COMMA] = ACTIONS(1576),
    [anon_sym_PIPE] = ACTIONS(1574),
    [anon_sym_DOT] = ACTIONS(1576),
    [anon_sym_LBRACK] = ACTIONS(1574),
    [anon_sym_RBRACK] = ACTIONS(1576),
    [anon_sym_true] = ACTIONS(1574),
    [anon_sym_false] = ACTIONS(1574),
    [anon_sym_LBRACE] = ACTIONS(1574),
    [anon_sym_SEMI] = ACTIONS(1574),
    [anon_sym_exception] = ACTIONS(1574),
    [anon_sym_module] = ACTIONS(1574),
    [anon_sym_open] = ACTIONS(1574),
    [anon_sym_include] = ACTIONS(1574),
    [anon_sym_class] = ACTIONS(1574),
    [anon_sym_val] = ACTIONS(1574),
    [anon_sym_end] = ACTIONS(1574),
    [anon_sym_object] = ACTIONS(1574),
    [anon_sym_AMP] = ACTIONS(1574),
    [anon_sym_POUND] = ACTIONS(1574),
    [anon_sym_COLON_COLON] = ACTIONS(1576),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1576),
    [anon_sym_LT_DASH] = ACTIONS(1574),
    [anon_sym_else] = ACTIONS(1574),
    [anon_sym_new] = ACTIONS(1574),
    [anon_sym_LBRACE_LT] = ACTIONS(1576),
    [anon_sym_begin] = ACTIONS(1574),
    [sym_ocamlyacc_value] = ACTIONS(1576),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1574),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1576),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1574),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1574),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1576),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1576),
    [aux_sym_number_token1] = ACTIONS(1576),
    [aux_sym_signed_number_token1] = ACTIONS(1574),
    [anon_sym_SQUOTE] = ACTIONS(1576),
    [anon_sym_DQUOTE] = ACTIONS(1576),
    [sym_prefix_operator] = ACTIONS(1576),
    [aux_sym_sign_operator_token1] = ACTIONS(1574),
    [sym_hash_operator] = ACTIONS(1576),
    [anon_sym_lsl] = ACTIONS(1574),
    [anon_sym_lsr] = ACTIONS(1574),
    [anon_sym_asr] = ACTIONS(1574),
    [aux_sym_pow_operator_token1] = ACTIONS(1576),
    [anon_sym_mod] = ACTIONS(1574),
    [anon_sym_land] = ACTIONS(1574),
    [anon_sym_lor] = ACTIONS(1574),
    [anon_sym_lxor] = ACTIONS(1574),
    [aux_sym_mult_operator_token1] = ACTIONS(1574),
    [aux_sym_add_operator_token1] = ACTIONS(1574),
    [sym_concat_operator] = ACTIONS(1576),
    [sym_rel_operator] = ACTIONS(1574),
    [anon_sym_AMP_AMP] = ACTIONS(1574),
    [anon_sym_or] = ACTIONS(1574),
    [anon_sym_PIPE_PIPE] = ACTIONS(1574),
    [sym_assign_operator] = ACTIONS(1576),
    [sym_let_operator] = ACTIONS(1576),
    [sym__uppercase_identifier] = ACTIONS(1576),
    [aux_sym_directive_token1] = ACTIONS(1574),
    [aux_sym_tag_token1] = ACTIONS(1576),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1063)] = {
    [sym_attribute] = STATE(1063),
    [sym__lowercase_identifier] = ACTIONS(1558),
    [anon_sym_SEMI_SEMI] = ACTIONS(1560),
    [anon_sym_let] = ACTIONS(1558),
    [anon_sym_and] = ACTIONS(1558),
    [anon_sym_TILDE] = ACTIONS(1558),
    [anon_sym_QMARK] = ACTIONS(1558),
    [anon_sym_LPAREN] = ACTIONS(1560),
    [anon_sym_external] = ACTIONS(1558),
    [anon_sym_type] = ACTIONS(1558),
    [anon_sym_COMMA] = ACTIONS(1560),
    [anon_sym_PIPE] = ACTIONS(1558),
    [anon_sym_DOT] = ACTIONS(1560),
    [anon_sym_LBRACK] = ACTIONS(1558),
    [anon_sym_RBRACK] = ACTIONS(1560),
    [anon_sym_true] = ACTIONS(1558),
    [anon_sym_false] = ACTIONS(1558),
    [anon_sym_LBRACE] = ACTIONS(1558),
    [anon_sym_SEMI] = ACTIONS(1558),
    [anon_sym_exception] = ACTIONS(1558),
    [anon_sym_module] = ACTIONS(1558),
    [anon_sym_open] = ACTIONS(1558),
    [anon_sym_include] = ACTIONS(1558),
    [anon_sym_class] = ACTIONS(1558),
    [anon_sym_val] = ACTIONS(1558),
    [anon_sym_end] = ACTIONS(1558),
    [anon_sym_object] = ACTIONS(1558),
    [anon_sym_AMP] = ACTIONS(1558),
    [anon_sym_POUND] = ACTIONS(1558),
    [anon_sym_COLON_COLON] = ACTIONS(1560),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1560),
    [anon_sym_LT_DASH] = ACTIONS(1558),
    [anon_sym_else] = ACTIONS(1558),
    [anon_sym_new] = ACTIONS(1558),
    [anon_sym_LBRACE_LT] = ACTIONS(1560),
    [anon_sym_begin] = ACTIONS(1558),
    [sym_ocamlyacc_value] = ACTIONS(1560),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1558),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1560),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1558),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1558),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1560),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1560),
    [aux_sym_number_token1] = ACTIONS(1560),
    [aux_sym_signed_number_token1] = ACTIONS(1558),
    [anon_sym_SQUOTE] = ACTIONS(1560),
    [anon_sym_DQUOTE] = ACTIONS(1560),
    [sym_prefix_operator] = ACTIONS(1560),
    [aux_sym_sign_operator_token1] = ACTIONS(1558),
    [sym_hash_operator] = ACTIONS(1560),
    [anon_sym_lsl] = ACTIONS(1558),
    [anon_sym_lsr] = ACTIONS(1558),
    [anon_sym_asr] = ACTIONS(1558),
    [aux_sym_pow_operator_token1] = ACTIONS(1560),
    [anon_sym_mod] = ACTIONS(1558),
    [anon_sym_land] = ACTIONS(1558),
    [anon_sym_lor] = ACTIONS(1558),
    [anon_sym_lxor] = ACTIONS(1558),
    [aux_sym_mult_operator_token1] = ACTIONS(1558),
    [aux_sym_add_operator_token1] = ACTIONS(1558),
    [sym_concat_operator] = ACTIONS(1560),
    [sym_rel_operator] = ACTIONS(1558),
    [anon_sym_AMP_AMP] = ACTIONS(1558),
    [anon_sym_or] = ACTIONS(1558),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [sym_assign_operator] = ACTIONS(1560),
    [sym_let_operator] = ACTIONS(1560),
    [sym__uppercase_identifier] = ACTIONS(1560),
    [aux_sym_directive_token1] = ACTIONS(1558),
    [aux_sym_tag_token1] = ACTIONS(1560),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1064)] = {
    [sym_attribute] = STATE(1064),
    [sym__lowercase_identifier] = ACTIONS(1562),
    [anon_sym_SEMI_SEMI] = ACTIONS(1564),
    [anon_sym_let] = ACTIONS(1562),
    [anon_sym_and] = ACTIONS(1562),
    [anon_sym_TILDE] = ACTIONS(1562),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_LPAREN] = ACTIONS(1564),
    [anon_sym_external] = ACTIONS(1562),
    [anon_sym_type] = ACTIONS(1562),
    [anon_sym_COMMA] = ACTIONS(1564),
    [anon_sym_PIPE] = ACTIONS(1562),
    [anon_sym_DOT] = ACTIONS(1564),
    [anon_sym_LBRACK] = ACTIONS(1562),
    [anon_sym_RBRACK] = ACTIONS(1564),
    [anon_sym_true] = ACTIONS(1562),
    [anon_sym_false] = ACTIONS(1562),
    [anon_sym_LBRACE] = ACTIONS(1562),
    [anon_sym_SEMI] = ACTIONS(1562),
    [anon_sym_exception] = ACTIONS(1562),
    [anon_sym_module] = ACTIONS(1562),
    [anon_sym_open] = ACTIONS(1562),
    [anon_sym_include] = ACTIONS(1562),
    [anon_sym_class] = ACTIONS(1562),
    [anon_sym_val] = ACTIONS(1562),
    [anon_sym_end] = ACTIONS(1562),
    [anon_sym_object] = ACTIONS(1562),
    [anon_sym_AMP] = ACTIONS(1562),
    [anon_sym_POUND] = ACTIONS(1562),
    [anon_sym_COLON_COLON] = ACTIONS(1564),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1564),
    [anon_sym_LT_DASH] = ACTIONS(1562),
    [anon_sym_else] = ACTIONS(1562),
    [anon_sym_new] = ACTIONS(1562),
    [anon_sym_LBRACE_LT] = ACTIONS(1564),
    [anon_sym_begin] = ACTIONS(1562),
    [sym_ocamlyacc_value] = ACTIONS(1564),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1562),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1564),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1562),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1562),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1564),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1564),
    [aux_sym_number_token1] = ACTIONS(1564),
    [aux_sym_signed_number_token1] = ACTIONS(1562),
    [anon_sym_SQUOTE] = ACTIONS(1564),
    [anon_sym_DQUOTE] = ACTIONS(1564),
    [sym_prefix_operator] = ACTIONS(1564),
    [aux_sym_sign_operator_token1] = ACTIONS(1562),
    [sym_hash_operator] = ACTIONS(1564),
    [anon_sym_lsl] = ACTIONS(1562),
    [anon_sym_lsr] = ACTIONS(1562),
    [anon_sym_asr] = ACTIONS(1562),
    [aux_sym_pow_operator_token1] = ACTIONS(1564),
    [anon_sym_mod] = ACTIONS(1562),
    [anon_sym_land] = ACTIONS(1562),
    [anon_sym_lor] = ACTIONS(1562),
    [anon_sym_lxor] = ACTIONS(1562),
    [aux_sym_mult_operator_token1] = ACTIONS(1562),
    [aux_sym_add_operator_token1] = ACTIONS(1562),
    [sym_concat_operator] = ACTIONS(1564),
    [sym_rel_operator] = ACTIONS(1562),
    [anon_sym_AMP_AMP] = ACTIONS(1562),
    [anon_sym_or] = ACTIONS(1562),
    [anon_sym_PIPE_PIPE] = ACTIONS(1562),
    [sym_assign_operator] = ACTIONS(1564),
    [sym_let_operator] = ACTIONS(1564),
    [sym__uppercase_identifier] = ACTIONS(1564),
    [aux_sym_directive_token1] = ACTIONS(1562),
    [aux_sym_tag_token1] = ACTIONS(1564),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1065)] = {
    [sym__parenthesized_abstract_type] = STATE(1367),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4358),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1065),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2258),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2139),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(2139),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_SEMI] = ACTIONS(2139),
    [anon_sym_RBRACE] = ACTIONS(2139),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_as] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1066)] = {
    [sym_attribute] = STATE(1066),
    [sym__lowercase_identifier] = ACTIONS(1724),
    [anon_sym_TILDE] = ACTIONS(1724),
    [anon_sym_QMARK] = ACTIONS(1724),
    [anon_sym_LPAREN] = ACTIONS(1726),
    [anon_sym_RPAREN] = ACTIONS(1726),
    [anon_sym_COMMA] = ACTIONS(1726),
    [anon_sym_PIPE] = ACTIONS(1724),
    [anon_sym_COLON2] = ACTIONS(1724),
    [anon_sym_DOT] = ACTIONS(1726),
    [anon_sym_DASH_GT] = ACTIONS(1724),
    [anon_sym_LBRACK] = ACTIONS(1724),
    [anon_sym_RBRACK] = ACTIONS(1726),
    [anon_sym_true] = ACTIONS(1724),
    [anon_sym_false] = ACTIONS(1724),
    [anon_sym_LBRACE] = ACTIONS(1724),
    [anon_sym_SEMI] = ACTIONS(1726),
    [anon_sym_RBRACE] = ACTIONS(1726),
    [anon_sym_constraint] = ACTIONS(1724),
    [anon_sym_val] = ACTIONS(1724),
    [anon_sym_end] = ACTIONS(1724),
    [anon_sym_with] = ACTIONS(1724),
    [anon_sym_COLON_GT] = ACTIONS(1726),
    [anon_sym_object] = ACTIONS(1724),
    [anon_sym_inherit] = ACTIONS(1724),
    [anon_sym_method] = ACTIONS(1724),
    [anon_sym_initializer] = ACTIONS(1724),
    [anon_sym_AMP] = ACTIONS(1724),
    [anon_sym_POUND] = ACTIONS(1724),
    [anon_sym_COLON_COLON] = ACTIONS(1726),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1726),
    [anon_sym_LT_DASH] = ACTIONS(2266),
    [anon_sym_then] = ACTIONS(1724),
    [anon_sym_else] = ACTIONS(1724),
    [anon_sym_new] = ACTIONS(1724),
    [anon_sym_LBRACE_LT] = ACTIONS(1726),
    [anon_sym_GT_RBRACE] = ACTIONS(1726),
    [anon_sym_begin] = ACTIONS(1724),
    [sym_ocamlyacc_value] = ACTIONS(1726),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1724),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1726),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1726),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1726),
    [aux_sym_number_token1] = ACTIONS(1726),
    [aux_sym_signed_number_token1] = ACTIONS(1724),
    [anon_sym_SQUOTE] = ACTIONS(1726),
    [anon_sym_DQUOTE] = ACTIONS(1726),
    [sym_prefix_operator] = ACTIONS(1726),
    [aux_sym_sign_operator_token1] = ACTIONS(1724),
    [sym_hash_operator] = ACTIONS(1726),
    [anon_sym_lsl] = ACTIONS(1724),
    [anon_sym_lsr] = ACTIONS(1724),
    [anon_sym_asr] = ACTIONS(1724),
    [aux_sym_pow_operator_token1] = ACTIONS(1726),
    [anon_sym_mod] = ACTIONS(1724),
    [anon_sym_land] = ACTIONS(1724),
    [anon_sym_lor] = ACTIONS(1724),
    [anon_sym_lxor] = ACTIONS(1724),
    [aux_sym_mult_operator_token1] = ACTIONS(1724),
    [aux_sym_add_operator_token1] = ACTIONS(1724),
    [sym_concat_operator] = ACTIONS(1726),
    [sym_rel_operator] = ACTIONS(1724),
    [anon_sym_AMP_AMP] = ACTIONS(1724),
    [anon_sym_or] = ACTIONS(1724),
    [anon_sym_PIPE_PIPE] = ACTIONS(1724),
    [sym_assign_operator] = ACTIONS(1726),
    [sym__uppercase_identifier] = ACTIONS(1726),
    [aux_sym_tag_token1] = ACTIONS(1726),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1067)] = {
    [sym_attribute] = STATE(1067),
    [sym__lowercase_identifier] = ACTIONS(1822),
    [anon_sym_SEMI_SEMI] = ACTIONS(1824),
    [anon_sym_let] = ACTIONS(1822),
    [anon_sym_and] = ACTIONS(1822),
    [anon_sym_TILDE] = ACTIONS(1822),
    [anon_sym_QMARK] = ACTIONS(1822),
    [anon_sym_LPAREN] = ACTIONS(1824),
    [anon_sym_external] = ACTIONS(1822),
    [anon_sym_type] = ACTIONS(1822),
    [anon_sym_COMMA] = ACTIONS(1824),
    [anon_sym_PIPE] = ACTIONS(1822),
    [anon_sym_DOT] = ACTIONS(1824),
    [anon_sym_LBRACK] = ACTIONS(1822),
    [anon_sym_RBRACK] = ACTIONS(1824),
    [anon_sym_true] = ACTIONS(1822),
    [anon_sym_false] = ACTIONS(1822),
    [anon_sym_LBRACE] = ACTIONS(1822),
    [anon_sym_SEMI] = ACTIONS(1822),
    [anon_sym_exception] = ACTIONS(1822),
    [anon_sym_module] = ACTIONS(1822),
    [anon_sym_open] = ACTIONS(1822),
    [anon_sym_include] = ACTIONS(1822),
    [anon_sym_class] = ACTIONS(1822),
    [anon_sym_val] = ACTIONS(1822),
    [anon_sym_end] = ACTIONS(1822),
    [anon_sym_object] = ACTIONS(1822),
    [anon_sym_AMP] = ACTIONS(1822),
    [anon_sym_POUND] = ACTIONS(1822),
    [anon_sym_COLON_COLON] = ACTIONS(1824),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1824),
    [anon_sym_else] = ACTIONS(1822),
    [anon_sym_new] = ACTIONS(1822),
    [anon_sym_LBRACE_LT] = ACTIONS(1824),
    [anon_sym_begin] = ACTIONS(1822),
    [sym_ocamlyacc_value] = ACTIONS(1824),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1822),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1824),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1822),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1822),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1824),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1824),
    [aux_sym_number_token1] = ACTIONS(1824),
    [aux_sym_signed_number_token1] = ACTIONS(1822),
    [anon_sym_SQUOTE] = ACTIONS(1824),
    [anon_sym_DQUOTE] = ACTIONS(1824),
    [sym_prefix_operator] = ACTIONS(1824),
    [aux_sym_sign_operator_token1] = ACTIONS(1822),
    [sym_hash_operator] = ACTIONS(1824),
    [anon_sym_lsl] = ACTIONS(1822),
    [anon_sym_lsr] = ACTIONS(1822),
    [anon_sym_asr] = ACTIONS(1822),
    [aux_sym_pow_operator_token1] = ACTIONS(1824),
    [anon_sym_mod] = ACTIONS(1822),
    [anon_sym_land] = ACTIONS(1822),
    [anon_sym_lor] = ACTIONS(1822),
    [anon_sym_lxor] = ACTIONS(1822),
    [aux_sym_mult_operator_token1] = ACTIONS(1822),
    [aux_sym_add_operator_token1] = ACTIONS(1822),
    [sym_concat_operator] = ACTIONS(1824),
    [sym_rel_operator] = ACTIONS(1822),
    [anon_sym_AMP_AMP] = ACTIONS(1822),
    [anon_sym_or] = ACTIONS(1822),
    [anon_sym_PIPE_PIPE] = ACTIONS(1822),
    [sym_assign_operator] = ACTIONS(1824),
    [sym_let_operator] = ACTIONS(1824),
    [sym__uppercase_identifier] = ACTIONS(1824),
    [aux_sym_directive_token1] = ACTIONS(1822),
    [aux_sym_tag_token1] = ACTIONS(1824),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1068)] = {
    [sym_attribute] = STATE(1068),
    [sym__lowercase_identifier] = ACTIONS(1930),
    [anon_sym_SEMI_SEMI] = ACTIONS(1932),
    [anon_sym_let] = ACTIONS(1930),
    [anon_sym_and] = ACTIONS(1930),
    [anon_sym_TILDE] = ACTIONS(1930),
    [anon_sym_QMARK] = ACTIONS(1930),
    [anon_sym_LPAREN] = ACTIONS(1932),
    [anon_sym_external] = ACTIONS(1930),
    [anon_sym_type] = ACTIONS(1930),
    [anon_sym_COMMA] = ACTIONS(1932),
    [anon_sym_PIPE] = ACTIONS(1930),
    [anon_sym_DOT] = ACTIONS(1932),
    [anon_sym_LBRACK] = ACTIONS(1930),
    [anon_sym_RBRACK] = ACTIONS(1932),
    [anon_sym_true] = ACTIONS(1930),
    [anon_sym_false] = ACTIONS(1930),
    [anon_sym_LBRACE] = ACTIONS(1930),
    [anon_sym_SEMI] = ACTIONS(1930),
    [anon_sym_exception] = ACTIONS(1930),
    [anon_sym_module] = ACTIONS(1930),
    [anon_sym_open] = ACTIONS(1930),
    [anon_sym_include] = ACTIONS(1930),
    [anon_sym_class] = ACTIONS(1930),
    [anon_sym_val] = ACTIONS(1930),
    [anon_sym_end] = ACTIONS(1930),
    [anon_sym_object] = ACTIONS(1930),
    [anon_sym_AMP] = ACTIONS(1930),
    [anon_sym_POUND] = ACTIONS(1930),
    [anon_sym_COLON_COLON] = ACTIONS(1932),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1932),
    [anon_sym_else] = ACTIONS(1930),
    [anon_sym_new] = ACTIONS(1930),
    [anon_sym_LBRACE_LT] = ACTIONS(1932),
    [anon_sym_begin] = ACTIONS(1930),
    [sym_ocamlyacc_value] = ACTIONS(1932),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1930),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1932),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1930),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1930),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1932),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1932),
    [aux_sym_number_token1] = ACTIONS(1932),
    [aux_sym_signed_number_token1] = ACTIONS(1930),
    [anon_sym_SQUOTE] = ACTIONS(1932),
    [anon_sym_DQUOTE] = ACTIONS(1932),
    [sym_prefix_operator] = ACTIONS(1932),
    [aux_sym_sign_operator_token1] = ACTIONS(1930),
    [sym_hash_operator] = ACTIONS(1932),
    [anon_sym_lsl] = ACTIONS(1930),
    [anon_sym_lsr] = ACTIONS(1930),
    [anon_sym_asr] = ACTIONS(1930),
    [aux_sym_pow_operator_token1] = ACTIONS(1932),
    [anon_sym_mod] = ACTIONS(1930),
    [anon_sym_land] = ACTIONS(1930),
    [anon_sym_lor] = ACTIONS(1930),
    [anon_sym_lxor] = ACTIONS(1930),
    [aux_sym_mult_operator_token1] = ACTIONS(1930),
    [aux_sym_add_operator_token1] = ACTIONS(1930),
    [sym_concat_operator] = ACTIONS(1932),
    [sym_rel_operator] = ACTIONS(1930),
    [anon_sym_AMP_AMP] = ACTIONS(1930),
    [anon_sym_or] = ACTIONS(1930),
    [anon_sym_PIPE_PIPE] = ACTIONS(1930),
    [sym_assign_operator] = ACTIONS(1932),
    [sym_let_operator] = ACTIONS(1932),
    [sym__uppercase_identifier] = ACTIONS(1932),
    [aux_sym_directive_token1] = ACTIONS(1930),
    [aux_sym_tag_token1] = ACTIONS(1932),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1069)] = {
    [sym_attribute] = STATE(1069),
    [sym__lowercase_identifier] = ACTIONS(1698),
    [anon_sym_SEMI_SEMI] = ACTIONS(1700),
    [anon_sym_let] = ACTIONS(1698),
    [anon_sym_and] = ACTIONS(1698),
    [anon_sym_TILDE] = ACTIONS(1698),
    [anon_sym_QMARK] = ACTIONS(1698),
    [anon_sym_LPAREN] = ACTIONS(1700),
    [anon_sym_external] = ACTIONS(1698),
    [anon_sym_type] = ACTIONS(1698),
    [anon_sym_COMMA] = ACTIONS(1700),
    [anon_sym_PIPE] = ACTIONS(1698),
    [anon_sym_DOT] = ACTIONS(1744),
    [anon_sym_LBRACK] = ACTIONS(1698),
    [anon_sym_RBRACK] = ACTIONS(1700),
    [anon_sym_true] = ACTIONS(1698),
    [anon_sym_false] = ACTIONS(1698),
    [anon_sym_LBRACE] = ACTIONS(1698),
    [anon_sym_SEMI] = ACTIONS(1698),
    [anon_sym_exception] = ACTIONS(1698),
    [anon_sym_module] = ACTIONS(1698),
    [anon_sym_open] = ACTIONS(1698),
    [anon_sym_include] = ACTIONS(1698),
    [anon_sym_class] = ACTIONS(1698),
    [anon_sym_val] = ACTIONS(1698),
    [anon_sym_end] = ACTIONS(1698),
    [anon_sym_object] = ACTIONS(1698),
    [anon_sym_AMP] = ACTIONS(1698),
    [anon_sym_POUND] = ACTIONS(1698),
    [anon_sym_COLON_COLON] = ACTIONS(1700),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1700),
    [anon_sym_else] = ACTIONS(1698),
    [anon_sym_new] = ACTIONS(1698),
    [anon_sym_LBRACE_LT] = ACTIONS(1700),
    [anon_sym_begin] = ACTIONS(1698),
    [sym_ocamlyacc_value] = ACTIONS(1700),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1698),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1700),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1700),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1700),
    [aux_sym_number_token1] = ACTIONS(1700),
    [aux_sym_signed_number_token1] = ACTIONS(1698),
    [anon_sym_SQUOTE] = ACTIONS(1700),
    [anon_sym_DQUOTE] = ACTIONS(1700),
    [sym_prefix_operator] = ACTIONS(1700),
    [aux_sym_sign_operator_token1] = ACTIONS(1698),
    [sym_hash_operator] = ACTIONS(1700),
    [anon_sym_lsl] = ACTIONS(1698),
    [anon_sym_lsr] = ACTIONS(1698),
    [anon_sym_asr] = ACTIONS(1698),
    [aux_sym_pow_operator_token1] = ACTIONS(1700),
    [anon_sym_mod] = ACTIONS(1698),
    [anon_sym_land] = ACTIONS(1698),
    [anon_sym_lor] = ACTIONS(1698),
    [anon_sym_lxor] = ACTIONS(1698),
    [aux_sym_mult_operator_token1] = ACTIONS(1698),
    [aux_sym_add_operator_token1] = ACTIONS(1698),
    [sym_concat_operator] = ACTIONS(1700),
    [sym_rel_operator] = ACTIONS(1698),
    [anon_sym_AMP_AMP] = ACTIONS(1698),
    [anon_sym_or] = ACTIONS(1698),
    [anon_sym_PIPE_PIPE] = ACTIONS(1698),
    [sym_assign_operator] = ACTIONS(1700),
    [sym_let_operator] = ACTIONS(1700),
    [sym__uppercase_identifier] = ACTIONS(1700),
    [aux_sym_directive_token1] = ACTIONS(1698),
    [aux_sym_tag_token1] = ACTIONS(1700),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1070)] = {
    [sym_attribute] = STATE(1070),
    [sym__lowercase_identifier] = ACTIONS(1634),
    [anon_sym_SEMI_SEMI] = ACTIONS(1636),
    [anon_sym_let] = ACTIONS(1634),
    [anon_sym_and] = ACTIONS(1634),
    [anon_sym_TILDE] = ACTIONS(1634),
    [anon_sym_QMARK] = ACTIONS(1634),
    [anon_sym_LPAREN] = ACTIONS(1636),
    [anon_sym_external] = ACTIONS(1634),
    [anon_sym_type] = ACTIONS(1634),
    [anon_sym_COMMA] = ACTIONS(1636),
    [anon_sym_PIPE] = ACTIONS(1634),
    [anon_sym_DOT] = ACTIONS(1636),
    [anon_sym_LBRACK] = ACTIONS(1634),
    [anon_sym_RBRACK] = ACTIONS(1636),
    [anon_sym_true] = ACTIONS(1634),
    [anon_sym_false] = ACTIONS(1634),
    [anon_sym_LBRACE] = ACTIONS(1634),
    [anon_sym_SEMI] = ACTIONS(1634),
    [anon_sym_exception] = ACTIONS(1634),
    [anon_sym_module] = ACTIONS(1634),
    [anon_sym_open] = ACTIONS(1634),
    [anon_sym_include] = ACTIONS(1634),
    [anon_sym_class] = ACTIONS(1634),
    [anon_sym_val] = ACTIONS(1634),
    [anon_sym_end] = ACTIONS(1634),
    [anon_sym_object] = ACTIONS(1634),
    [anon_sym_AMP] = ACTIONS(1634),
    [anon_sym_POUND] = ACTIONS(1634),
    [anon_sym_COLON_COLON] = ACTIONS(1636),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1636),
    [anon_sym_else] = ACTIONS(1634),
    [anon_sym_new] = ACTIONS(1634),
    [anon_sym_LBRACE_LT] = ACTIONS(1636),
    [anon_sym_begin] = ACTIONS(1634),
    [sym_ocamlyacc_value] = ACTIONS(1636),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1634),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1636),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1634),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1634),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1636),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1636),
    [aux_sym_number_token1] = ACTIONS(1636),
    [aux_sym_signed_number_token1] = ACTIONS(1634),
    [anon_sym_SQUOTE] = ACTIONS(1636),
    [anon_sym_DQUOTE] = ACTIONS(1636),
    [sym_prefix_operator] = ACTIONS(1636),
    [aux_sym_sign_operator_token1] = ACTIONS(1634),
    [sym_hash_operator] = ACTIONS(1636),
    [anon_sym_lsl] = ACTIONS(1634),
    [anon_sym_lsr] = ACTIONS(1634),
    [anon_sym_asr] = ACTIONS(1634),
    [aux_sym_pow_operator_token1] = ACTIONS(1636),
    [anon_sym_mod] = ACTIONS(1634),
    [anon_sym_land] = ACTIONS(1634),
    [anon_sym_lor] = ACTIONS(1634),
    [anon_sym_lxor] = ACTIONS(1634),
    [aux_sym_mult_operator_token1] = ACTIONS(1634),
    [aux_sym_add_operator_token1] = ACTIONS(1634),
    [sym_concat_operator] = ACTIONS(1636),
    [sym_rel_operator] = ACTIONS(1634),
    [anon_sym_AMP_AMP] = ACTIONS(1634),
    [anon_sym_or] = ACTIONS(1634),
    [anon_sym_PIPE_PIPE] = ACTIONS(1634),
    [sym_assign_operator] = ACTIONS(1636),
    [sym_let_operator] = ACTIONS(1636),
    [sym__uppercase_identifier] = ACTIONS(1636),
    [aux_sym_directive_token1] = ACTIONS(1634),
    [aux_sym_tag_token1] = ACTIONS(1636),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1071)] = {
    [sym_attribute] = STATE(1071),
    [sym__lowercase_identifier] = ACTIONS(1754),
    [anon_sym_SEMI_SEMI] = ACTIONS(1756),
    [anon_sym_let] = ACTIONS(1754),
    [anon_sym_and] = ACTIONS(1754),
    [anon_sym_TILDE] = ACTIONS(1754),
    [anon_sym_QMARK] = ACTIONS(1754),
    [anon_sym_LPAREN] = ACTIONS(1756),
    [anon_sym_external] = ACTIONS(1754),
    [anon_sym_type] = ACTIONS(1754),
    [anon_sym_COMMA] = ACTIONS(1756),
    [anon_sym_PIPE] = ACTIONS(1754),
    [anon_sym_DOT] = ACTIONS(1756),
    [anon_sym_LBRACK] = ACTIONS(1754),
    [anon_sym_RBRACK] = ACTIONS(1756),
    [anon_sym_true] = ACTIONS(1754),
    [anon_sym_false] = ACTIONS(1754),
    [anon_sym_LBRACE] = ACTIONS(1754),
    [anon_sym_SEMI] = ACTIONS(1754),
    [anon_sym_exception] = ACTIONS(1754),
    [anon_sym_module] = ACTIONS(1754),
    [anon_sym_open] = ACTIONS(1754),
    [anon_sym_include] = ACTIONS(1754),
    [anon_sym_class] = ACTIONS(1754),
    [anon_sym_val] = ACTIONS(1754),
    [anon_sym_end] = ACTIONS(1754),
    [anon_sym_object] = ACTIONS(1754),
    [anon_sym_AMP] = ACTIONS(1754),
    [anon_sym_POUND] = ACTIONS(1754),
    [anon_sym_COLON_COLON] = ACTIONS(1756),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1756),
    [anon_sym_else] = ACTIONS(1754),
    [anon_sym_new] = ACTIONS(1754),
    [anon_sym_LBRACE_LT] = ACTIONS(1756),
    [anon_sym_begin] = ACTIONS(1754),
    [sym_ocamlyacc_value] = ACTIONS(1756),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1754),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1756),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1754),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1754),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1756),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1756),
    [aux_sym_number_token1] = ACTIONS(1756),
    [aux_sym_signed_number_token1] = ACTIONS(1754),
    [anon_sym_SQUOTE] = ACTIONS(1756),
    [anon_sym_DQUOTE] = ACTIONS(1756),
    [sym_prefix_operator] = ACTIONS(1756),
    [aux_sym_sign_operator_token1] = ACTIONS(1754),
    [sym_hash_operator] = ACTIONS(1756),
    [anon_sym_lsl] = ACTIONS(1754),
    [anon_sym_lsr] = ACTIONS(1754),
    [anon_sym_asr] = ACTIONS(1754),
    [aux_sym_pow_operator_token1] = ACTIONS(1756),
    [anon_sym_mod] = ACTIONS(1754),
    [anon_sym_land] = ACTIONS(1754),
    [anon_sym_lor] = ACTIONS(1754),
    [anon_sym_lxor] = ACTIONS(1754),
    [aux_sym_mult_operator_token1] = ACTIONS(1754),
    [aux_sym_add_operator_token1] = ACTIONS(1754),
    [sym_concat_operator] = ACTIONS(1756),
    [sym_rel_operator] = ACTIONS(1754),
    [anon_sym_AMP_AMP] = ACTIONS(1754),
    [anon_sym_or] = ACTIONS(1754),
    [anon_sym_PIPE_PIPE] = ACTIONS(1754),
    [sym_assign_operator] = ACTIONS(1756),
    [sym_let_operator] = ACTIONS(1756),
    [sym__uppercase_identifier] = ACTIONS(1756),
    [aux_sym_directive_token1] = ACTIONS(1754),
    [aux_sym_tag_token1] = ACTIONS(1756),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1072)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4552),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1072),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_RBRACK] = ACTIONS(1316),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_SEMI] = ACTIONS(1316),
    [anon_sym_RBRACE] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_as] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_COLON_COLON] = ACTIONS(1316),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1073)] = {
    [sym__parenthesized_abstract_type] = STATE(1405),
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4549),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1073),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2270),
    [anon_sym_RPAREN] = ACTIONS(1316),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_COLON2] = ACTIONS(1318),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_as] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1316),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1074)] = {
    [sym_attribute] = STATE(1074),
    [sym__lowercase_identifier] = ACTIONS(1740),
    [anon_sym_SEMI_SEMI] = ACTIONS(1742),
    [anon_sym_let] = ACTIONS(1740),
    [anon_sym_and] = ACTIONS(1740),
    [anon_sym_TILDE] = ACTIONS(1740),
    [anon_sym_QMARK] = ACTIONS(1740),
    [anon_sym_LPAREN] = ACTIONS(1742),
    [anon_sym_external] = ACTIONS(1740),
    [anon_sym_type] = ACTIONS(1740),
    [anon_sym_COMMA] = ACTIONS(1742),
    [anon_sym_PIPE] = ACTIONS(1740),
    [anon_sym_DOT] = ACTIONS(1742),
    [anon_sym_LBRACK] = ACTIONS(1740),
    [anon_sym_RBRACK] = ACTIONS(1742),
    [anon_sym_true] = ACTIONS(1740),
    [anon_sym_false] = ACTIONS(1740),
    [anon_sym_LBRACE] = ACTIONS(1740),
    [anon_sym_SEMI] = ACTIONS(1740),
    [anon_sym_exception] = ACTIONS(1740),
    [anon_sym_module] = ACTIONS(1740),
    [anon_sym_open] = ACTIONS(1740),
    [anon_sym_include] = ACTIONS(1740),
    [anon_sym_class] = ACTIONS(1740),
    [anon_sym_val] = ACTIONS(1740),
    [anon_sym_end] = ACTIONS(1740),
    [anon_sym_object] = ACTIONS(1740),
    [anon_sym_AMP] = ACTIONS(1740),
    [anon_sym_POUND] = ACTIONS(1740),
    [anon_sym_COLON_COLON] = ACTIONS(1742),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1742),
    [anon_sym_else] = ACTIONS(1740),
    [anon_sym_new] = ACTIONS(1740),
    [anon_sym_LBRACE_LT] = ACTIONS(1742),
    [anon_sym_begin] = ACTIONS(1740),
    [sym_ocamlyacc_value] = ACTIONS(1742),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1740),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1742),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1740),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1740),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1742),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1742),
    [aux_sym_number_token1] = ACTIONS(1742),
    [aux_sym_signed_number_token1] = ACTIONS(1740),
    [anon_sym_SQUOTE] = ACTIONS(1742),
    [anon_sym_DQUOTE] = ACTIONS(1742),
    [sym_prefix_operator] = ACTIONS(1742),
    [aux_sym_sign_operator_token1] = ACTIONS(1740),
    [sym_hash_operator] = ACTIONS(1742),
    [anon_sym_lsl] = ACTIONS(1740),
    [anon_sym_lsr] = ACTIONS(1740),
    [anon_sym_asr] = ACTIONS(1740),
    [aux_sym_pow_operator_token1] = ACTIONS(1742),
    [anon_sym_mod] = ACTIONS(1740),
    [anon_sym_land] = ACTIONS(1740),
    [anon_sym_lor] = ACTIONS(1740),
    [anon_sym_lxor] = ACTIONS(1740),
    [aux_sym_mult_operator_token1] = ACTIONS(1740),
    [aux_sym_add_operator_token1] = ACTIONS(1740),
    [sym_concat_operator] = ACTIONS(1742),
    [sym_rel_operator] = ACTIONS(1740),
    [anon_sym_AMP_AMP] = ACTIONS(1740),
    [anon_sym_or] = ACTIONS(1740),
    [anon_sym_PIPE_PIPE] = ACTIONS(1740),
    [sym_assign_operator] = ACTIONS(1742),
    [sym_let_operator] = ACTIONS(1742),
    [sym__uppercase_identifier] = ACTIONS(1742),
    [aux_sym_directive_token1] = ACTIONS(1740),
    [aux_sym_tag_token1] = ACTIONS(1742),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1075)] = {
    [sym_attribute] = STATE(1075),
    [sym__lowercase_identifier] = ACTIONS(1698),
    [anon_sym_SEMI_SEMI] = ACTIONS(1700),
    [anon_sym_let] = ACTIONS(1698),
    [anon_sym_and] = ACTIONS(1698),
    [anon_sym_TILDE] = ACTIONS(1698),
    [anon_sym_QMARK] = ACTIONS(1698),
    [anon_sym_LPAREN] = ACTIONS(1700),
    [anon_sym_external] = ACTIONS(1698),
    [anon_sym_type] = ACTIONS(1698),
    [anon_sym_COMMA] = ACTIONS(1700),
    [anon_sym_PIPE] = ACTIONS(1698),
    [anon_sym_DOT] = ACTIONS(1702),
    [anon_sym_LBRACK] = ACTIONS(1698),
    [anon_sym_RBRACK] = ACTIONS(1700),
    [anon_sym_true] = ACTIONS(1698),
    [anon_sym_false] = ACTIONS(1698),
    [anon_sym_LBRACE] = ACTIONS(1698),
    [anon_sym_SEMI] = ACTIONS(1698),
    [anon_sym_exception] = ACTIONS(1698),
    [anon_sym_module] = ACTIONS(1698),
    [anon_sym_open] = ACTIONS(1698),
    [anon_sym_include] = ACTIONS(1698),
    [anon_sym_class] = ACTIONS(1698),
    [anon_sym_val] = ACTIONS(1698),
    [anon_sym_end] = ACTIONS(1698),
    [anon_sym_object] = ACTIONS(1698),
    [anon_sym_AMP] = ACTIONS(1698),
    [anon_sym_POUND] = ACTIONS(1698),
    [anon_sym_COLON_COLON] = ACTIONS(1700),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1700),
    [anon_sym_else] = ACTIONS(1698),
    [anon_sym_new] = ACTIONS(1698),
    [anon_sym_LBRACE_LT] = ACTIONS(1700),
    [anon_sym_begin] = ACTIONS(1698),
    [sym_ocamlyacc_value] = ACTIONS(1700),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1698),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1700),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1698),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1700),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1700),
    [aux_sym_number_token1] = ACTIONS(1700),
    [aux_sym_signed_number_token1] = ACTIONS(1698),
    [anon_sym_SQUOTE] = ACTIONS(1700),
    [anon_sym_DQUOTE] = ACTIONS(1700),
    [sym_prefix_operator] = ACTIONS(1700),
    [aux_sym_sign_operator_token1] = ACTIONS(1698),
    [sym_hash_operator] = ACTIONS(1700),
    [anon_sym_lsl] = ACTIONS(1698),
    [anon_sym_lsr] = ACTIONS(1698),
    [anon_sym_asr] = ACTIONS(1698),
    [aux_sym_pow_operator_token1] = ACTIONS(1700),
    [anon_sym_mod] = ACTIONS(1698),
    [anon_sym_land] = ACTIONS(1698),
    [anon_sym_lor] = ACTIONS(1698),
    [anon_sym_lxor] = ACTIONS(1698),
    [aux_sym_mult_operator_token1] = ACTIONS(1698),
    [aux_sym_add_operator_token1] = ACTIONS(1698),
    [sym_concat_operator] = ACTIONS(1700),
    [sym_rel_operator] = ACTIONS(1698),
    [anon_sym_AMP_AMP] = ACTIONS(1698),
    [anon_sym_or] = ACTIONS(1698),
    [anon_sym_PIPE_PIPE] = ACTIONS(1698),
    [sym_assign_operator] = ACTIONS(1700),
    [sym_let_operator] = ACTIONS(1700),
    [sym__uppercase_identifier] = ACTIONS(1700),
    [aux_sym_directive_token1] = ACTIONS(1698),
    [aux_sym_tag_token1] = ACTIONS(1700),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1076)] = {
    [sym_attribute] = STATE(1076),
    [sym__lowercase_identifier] = ACTIONS(1618),
    [anon_sym_SEMI_SEMI] = ACTIONS(1620),
    [anon_sym_let] = ACTIONS(1618),
    [anon_sym_and] = ACTIONS(1618),
    [anon_sym_TILDE] = ACTIONS(1618),
    [anon_sym_QMARK] = ACTIONS(1618),
    [anon_sym_LPAREN] = ACTIONS(1620),
    [anon_sym_external] = ACTIONS(1618),
    [anon_sym_type] = ACTIONS(1618),
    [anon_sym_COMMA] = ACTIONS(1620),
    [anon_sym_PIPE] = ACTIONS(1618),
    [anon_sym_DOT] = ACTIONS(1620),
    [anon_sym_LBRACK] = ACTIONS(1618),
    [anon_sym_RBRACK] = ACTIONS(1620),
    [anon_sym_true] = ACTIONS(1618),
    [anon_sym_false] = ACTIONS(1618),
    [anon_sym_LBRACE] = ACTIONS(1618),
    [anon_sym_SEMI] = ACTIONS(1618),
    [anon_sym_exception] = ACTIONS(1618),
    [anon_sym_module] = ACTIONS(1618),
    [anon_sym_open] = ACTIONS(1618),
    [anon_sym_include] = ACTIONS(1618),
    [anon_sym_class] = ACTIONS(1618),
    [anon_sym_val] = ACTIONS(1618),
    [anon_sym_end] = ACTIONS(1618),
    [anon_sym_object] = ACTIONS(1618),
    [anon_sym_AMP] = ACTIONS(1618),
    [anon_sym_POUND] = ACTIONS(1618),
    [anon_sym_COLON_COLON] = ACTIONS(1620),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1620),
    [anon_sym_else] = ACTIONS(1618),
    [anon_sym_new] = ACTIONS(1618),
    [anon_sym_LBRACE_LT] = ACTIONS(1620),
    [anon_sym_begin] = ACTIONS(1618),
    [sym_ocamlyacc_value] = ACTIONS(1620),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1618),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1620),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1618),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1618),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1620),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1620),
    [aux_sym_number_token1] = ACTIONS(1620),
    [aux_sym_signed_number_token1] = ACTIONS(1618),
    [anon_sym_SQUOTE] = ACTIONS(1620),
    [anon_sym_DQUOTE] = ACTIONS(1620),
    [sym_prefix_operator] = ACTIONS(1620),
    [aux_sym_sign_operator_token1] = ACTIONS(1618),
    [sym_hash_operator] = ACTIONS(1620),
    [anon_sym_lsl] = ACTIONS(1618),
    [anon_sym_lsr] = ACTIONS(1618),
    [anon_sym_asr] = ACTIONS(1618),
    [aux_sym_pow_operator_token1] = ACTIONS(1620),
    [anon_sym_mod] = ACTIONS(1618),
    [anon_sym_land] = ACTIONS(1618),
    [anon_sym_lor] = ACTIONS(1618),
    [anon_sym_lxor] = ACTIONS(1618),
    [aux_sym_mult_operator_token1] = ACTIONS(1618),
    [aux_sym_add_operator_token1] = ACTIONS(1618),
    [sym_concat_operator] = ACTIONS(1620),
    [sym_rel_operator] = ACTIONS(1618),
    [anon_sym_AMP_AMP] = ACTIONS(1618),
    [anon_sym_or] = ACTIONS(1618),
    [anon_sym_PIPE_PIPE] = ACTIONS(1618),
    [sym_assign_operator] = ACTIONS(1620),
    [sym_let_operator] = ACTIONS(1620),
    [sym__uppercase_identifier] = ACTIONS(1620),
    [aux_sym_directive_token1] = ACTIONS(1618),
    [aux_sym_tag_token1] = ACTIONS(1620),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1077)] = {
    [sym_attribute] = STATE(1077),
    [sym__lowercase_identifier] = ACTIONS(1910),
    [anon_sym_SEMI_SEMI] = ACTIONS(1912),
    [anon_sym_let] = ACTIONS(1910),
    [anon_sym_and] = ACTIONS(1910),
    [anon_sym_TILDE] = ACTIONS(1910),
    [anon_sym_QMARK] = ACTIONS(1910),
    [anon_sym_LPAREN] = ACTIONS(1912),
    [anon_sym_external] = ACTIONS(1910),
    [anon_sym_type] = ACTIONS(1910),
    [anon_sym_COMMA] = ACTIONS(1912),
    [anon_sym_PIPE] = ACTIONS(1910),
    [anon_sym_DOT] = ACTIONS(1912),
    [anon_sym_LBRACK] = ACTIONS(1910),
    [anon_sym_RBRACK] = ACTIONS(1912),
    [anon_sym_true] = ACTIONS(1910),
    [anon_sym_false] = ACTIONS(1910),
    [anon_sym_LBRACE] = ACTIONS(1910),
    [anon_sym_SEMI] = ACTIONS(1910),
    [anon_sym_exception] = ACTIONS(1910),
    [anon_sym_module] = ACTIONS(1910),
    [anon_sym_open] = ACTIONS(1910),
    [anon_sym_include] = ACTIONS(1910),
    [anon_sym_class] = ACTIONS(1910),
    [anon_sym_val] = ACTIONS(1910),
    [anon_sym_end] = ACTIONS(1910),
    [anon_sym_object] = ACTIONS(1910),
    [anon_sym_AMP] = ACTIONS(1910),
    [anon_sym_POUND] = ACTIONS(1910),
    [anon_sym_COLON_COLON] = ACTIONS(1912),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1912),
    [anon_sym_else] = ACTIONS(1910),
    [anon_sym_new] = ACTIONS(1910),
    [anon_sym_LBRACE_LT] = ACTIONS(1912),
    [anon_sym_begin] = ACTIONS(1910),
    [sym_ocamlyacc_value] = ACTIONS(1912),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1910),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1912),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1910),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1910),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1912),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1912),
    [aux_sym_number_token1] = ACTIONS(1912),
    [aux_sym_signed_number_token1] = ACTIONS(1910),
    [anon_sym_SQUOTE] = ACTIONS(1912),
    [anon_sym_DQUOTE] = ACTIONS(1912),
    [sym_prefix_operator] = ACTIONS(1912),
    [aux_sym_sign_operator_token1] = ACTIONS(1910),
    [sym_hash_operator] = ACTIONS(1912),
    [anon_sym_lsl] = ACTIONS(1910),
    [anon_sym_lsr] = ACTIONS(1910),
    [anon_sym_asr] = ACTIONS(1910),
    [aux_sym_pow_operator_token1] = ACTIONS(1912),
    [anon_sym_mod] = ACTIONS(1910),
    [anon_sym_land] = ACTIONS(1910),
    [anon_sym_lor] = ACTIONS(1910),
    [anon_sym_lxor] = ACTIONS(1910),
    [aux_sym_mult_operator_token1] = ACTIONS(1910),
    [aux_sym_add_operator_token1] = ACTIONS(1910),
    [sym_concat_operator] = ACTIONS(1912),
    [sym_rel_operator] = ACTIONS(1910),
    [anon_sym_AMP_AMP] = ACTIONS(1910),
    [anon_sym_or] = ACTIONS(1910),
    [anon_sym_PIPE_PIPE] = ACTIONS(1910),
    [sym_assign_operator] = ACTIONS(1912),
    [sym_let_operator] = ACTIONS(1912),
    [sym__uppercase_identifier] = ACTIONS(1912),
    [aux_sym_directive_token1] = ACTIONS(1910),
    [aux_sym_tag_token1] = ACTIONS(1912),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1078)] = {
    [sym_attribute] = STATE(1078),
    [sym__lowercase_identifier] = ACTIONS(1716),
    [anon_sym_SEMI_SEMI] = ACTIONS(1718),
    [anon_sym_let] = ACTIONS(1716),
    [anon_sym_and] = ACTIONS(1716),
    [anon_sym_TILDE] = ACTIONS(1716),
    [anon_sym_QMARK] = ACTIONS(1716),
    [anon_sym_LPAREN] = ACTIONS(1718),
    [anon_sym_external] = ACTIONS(1716),
    [anon_sym_type] = ACTIONS(1716),
    [anon_sym_COMMA] = ACTIONS(1718),
    [anon_sym_PIPE] = ACTIONS(1716),
    [anon_sym_DOT] = ACTIONS(1718),
    [anon_sym_LBRACK] = ACTIONS(1716),
    [anon_sym_RBRACK] = ACTIONS(1718),
    [anon_sym_true] = ACTIONS(1716),
    [anon_sym_false] = ACTIONS(1716),
    [anon_sym_LBRACE] = ACTIONS(1716),
    [anon_sym_SEMI] = ACTIONS(1716),
    [anon_sym_exception] = ACTIONS(1716),
    [anon_sym_module] = ACTIONS(1716),
    [anon_sym_open] = ACTIONS(1716),
    [anon_sym_include] = ACTIONS(1716),
    [anon_sym_class] = ACTIONS(1716),
    [anon_sym_val] = ACTIONS(1716),
    [anon_sym_end] = ACTIONS(1716),
    [anon_sym_object] = ACTIONS(1716),
    [anon_sym_AMP] = ACTIONS(1716),
    [anon_sym_POUND] = ACTIONS(1716),
    [anon_sym_COLON_COLON] = ACTIONS(1718),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1718),
    [anon_sym_else] = ACTIONS(1716),
    [anon_sym_new] = ACTIONS(1716),
    [anon_sym_LBRACE_LT] = ACTIONS(1718),
    [anon_sym_begin] = ACTIONS(1716),
    [sym_ocamlyacc_value] = ACTIONS(1718),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1716),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1718),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1716),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1716),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1718),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1718),
    [aux_sym_number_token1] = ACTIONS(1718),
    [aux_sym_signed_number_token1] = ACTIONS(1716),
    [anon_sym_SQUOTE] = ACTIONS(1718),
    [anon_sym_DQUOTE] = ACTIONS(1718),
    [sym_prefix_operator] = ACTIONS(1718),
    [aux_sym_sign_operator_token1] = ACTIONS(1716),
    [sym_hash_operator] = ACTIONS(1718),
    [anon_sym_lsl] = ACTIONS(1716),
    [anon_sym_lsr] = ACTIONS(1716),
    [anon_sym_asr] = ACTIONS(1716),
    [aux_sym_pow_operator_token1] = ACTIONS(1718),
    [anon_sym_mod] = ACTIONS(1716),
    [anon_sym_land] = ACTIONS(1716),
    [anon_sym_lor] = ACTIONS(1716),
    [anon_sym_lxor] = ACTIONS(1716),
    [aux_sym_mult_operator_token1] = ACTIONS(1716),
    [aux_sym_add_operator_token1] = ACTIONS(1716),
    [sym_concat_operator] = ACTIONS(1718),
    [sym_rel_operator] = ACTIONS(1716),
    [anon_sym_AMP_AMP] = ACTIONS(1716),
    [anon_sym_or] = ACTIONS(1716),
    [anon_sym_PIPE_PIPE] = ACTIONS(1716),
    [sym_assign_operator] = ACTIONS(1718),
    [sym_let_operator] = ACTIONS(1718),
    [sym__uppercase_identifier] = ACTIONS(1718),
    [aux_sym_directive_token1] = ACTIONS(1716),
    [aux_sym_tag_token1] = ACTIONS(1718),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1079)] = {
    [sym_attribute] = STATE(1079),
    [sym__lowercase_identifier] = ACTIONS(1642),
    [anon_sym_SEMI_SEMI] = ACTIONS(1644),
    [anon_sym_let] = ACTIONS(1642),
    [anon_sym_and] = ACTIONS(1642),
    [anon_sym_TILDE] = ACTIONS(1642),
    [anon_sym_QMARK] = ACTIONS(1642),
    [anon_sym_LPAREN] = ACTIONS(1644),
    [anon_sym_external] = ACTIONS(1642),
    [anon_sym_type] = ACTIONS(1642),
    [anon_sym_COMMA] = ACTIONS(1644),
    [anon_sym_PIPE] = ACTIONS(1642),
    [anon_sym_DOT] = ACTIONS(1644),
    [anon_sym_LBRACK] = ACTIONS(1642),
    [anon_sym_RBRACK] = ACTIONS(1644),
    [anon_sym_true] = ACTIONS(1642),
    [anon_sym_false] = ACTIONS(1642),
    [anon_sym_LBRACE] = ACTIONS(1642),
    [anon_sym_SEMI] = ACTIONS(1642),
    [anon_sym_exception] = ACTIONS(1642),
    [anon_sym_module] = ACTIONS(1642),
    [anon_sym_open] = ACTIONS(1642),
    [anon_sym_include] = ACTIONS(1642),
    [anon_sym_class] = ACTIONS(1642),
    [anon_sym_val] = ACTIONS(1642),
    [anon_sym_end] = ACTIONS(1642),
    [anon_sym_object] = ACTIONS(1642),
    [anon_sym_AMP] = ACTIONS(1642),
    [anon_sym_POUND] = ACTIONS(1642),
    [anon_sym_COLON_COLON] = ACTIONS(1644),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1644),
    [anon_sym_else] = ACTIONS(1642),
    [anon_sym_new] = ACTIONS(1642),
    [anon_sym_LBRACE_LT] = ACTIONS(1644),
    [anon_sym_begin] = ACTIONS(1642),
    [sym_ocamlyacc_value] = ACTIONS(1644),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1642),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1644),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1642),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1642),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1644),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1644),
    [aux_sym_number_token1] = ACTIONS(1644),
    [aux_sym_signed_number_token1] = ACTIONS(1642),
    [anon_sym_SQUOTE] = ACTIONS(1644),
    [anon_sym_DQUOTE] = ACTIONS(1644),
    [sym_prefix_operator] = ACTIONS(1644),
    [aux_sym_sign_operator_token1] = ACTIONS(1642),
    [sym_hash_operator] = ACTIONS(1644),
    [anon_sym_lsl] = ACTIONS(1642),
    [anon_sym_lsr] = ACTIONS(1642),
    [anon_sym_asr] = ACTIONS(1642),
    [aux_sym_pow_operator_token1] = ACTIONS(1644),
    [anon_sym_mod] = ACTIONS(1642),
    [anon_sym_land] = ACTIONS(1642),
    [anon_sym_lor] = ACTIONS(1642),
    [anon_sym_lxor] = ACTIONS(1642),
    [aux_sym_mult_operator_token1] = ACTIONS(1642),
    [aux_sym_add_operator_token1] = ACTIONS(1642),
    [sym_concat_operator] = ACTIONS(1644),
    [sym_rel_operator] = ACTIONS(1642),
    [anon_sym_AMP_AMP] = ACTIONS(1642),
    [anon_sym_or] = ACTIONS(1642),
    [anon_sym_PIPE_PIPE] = ACTIONS(1642),
    [sym_assign_operator] = ACTIONS(1644),
    [sym_let_operator] = ACTIONS(1644),
    [sym__uppercase_identifier] = ACTIONS(1644),
    [aux_sym_directive_token1] = ACTIONS(1642),
    [aux_sym_tag_token1] = ACTIONS(1644),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1080)] = {
    [sym_attribute] = STATE(1080),
    [sym__lowercase_identifier] = ACTIONS(1946),
    [anon_sym_SEMI_SEMI] = ACTIONS(1948),
    [anon_sym_let] = ACTIONS(1946),
    [anon_sym_and] = ACTIONS(1946),
    [anon_sym_TILDE] = ACTIONS(1946),
    [anon_sym_QMARK] = ACTIONS(1946),
    [anon_sym_LPAREN] = ACTIONS(1948),
    [anon_sym_external] = ACTIONS(1946),
    [anon_sym_type] = ACTIONS(1946),
    [anon_sym_COMMA] = ACTIONS(1948),
    [anon_sym_PIPE] = ACTIONS(1946),
    [anon_sym_DOT] = ACTIONS(1948),
    [anon_sym_LBRACK] = ACTIONS(1946),
    [anon_sym_RBRACK] = ACTIONS(1948),
    [anon_sym_true] = ACTIONS(1946),
    [anon_sym_false] = ACTIONS(1946),
    [anon_sym_LBRACE] = ACTIONS(1946),
    [anon_sym_SEMI] = ACTIONS(1946),
    [anon_sym_exception] = ACTIONS(1946),
    [anon_sym_module] = ACTIONS(1946),
    [anon_sym_open] = ACTIONS(1946),
    [anon_sym_include] = ACTIONS(1946),
    [anon_sym_class] = ACTIONS(1946),
    [anon_sym_val] = ACTIONS(1946),
    [anon_sym_end] = ACTIONS(1946),
    [anon_sym_object] = ACTIONS(1946),
    [anon_sym_AMP] = ACTIONS(1946),
    [anon_sym_POUND] = ACTIONS(1946),
    [anon_sym_COLON_COLON] = ACTIONS(1948),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1948),
    [anon_sym_else] = ACTIONS(1946),
    [anon_sym_new] = ACTIONS(1946),
    [anon_sym_LBRACE_LT] = ACTIONS(1948),
    [anon_sym_begin] = ACTIONS(1946),
    [sym_ocamlyacc_value] = ACTIONS(1948),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1946),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1948),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1946),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1946),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1948),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1948),
    [aux_sym_number_token1] = ACTIONS(1948),
    [aux_sym_signed_number_token1] = ACTIONS(1946),
    [anon_sym_SQUOTE] = ACTIONS(1948),
    [anon_sym_DQUOTE] = ACTIONS(1948),
    [sym_prefix_operator] = ACTIONS(1948),
    [aux_sym_sign_operator_token1] = ACTIONS(1946),
    [sym_hash_operator] = ACTIONS(1948),
    [anon_sym_lsl] = ACTIONS(1946),
    [anon_sym_lsr] = ACTIONS(1946),
    [anon_sym_asr] = ACTIONS(1946),
    [aux_sym_pow_operator_token1] = ACTIONS(1948),
    [anon_sym_mod] = ACTIONS(1946),
    [anon_sym_land] = ACTIONS(1946),
    [anon_sym_lor] = ACTIONS(1946),
    [anon_sym_lxor] = ACTIONS(1946),
    [aux_sym_mult_operator_token1] = ACTIONS(1946),
    [aux_sym_add_operator_token1] = ACTIONS(1946),
    [sym_concat_operator] = ACTIONS(1948),
    [sym_rel_operator] = ACTIONS(1946),
    [anon_sym_AMP_AMP] = ACTIONS(1946),
    [anon_sym_or] = ACTIONS(1946),
    [anon_sym_PIPE_PIPE] = ACTIONS(1946),
    [sym_assign_operator] = ACTIONS(1948),
    [sym_let_operator] = ACTIONS(1948),
    [sym__uppercase_identifier] = ACTIONS(1948),
    [aux_sym_directive_token1] = ACTIONS(1946),
    [aux_sym_tag_token1] = ACTIONS(1948),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1081)] = {
    [sym_attribute] = STATE(1081),
    [sym__lowercase_identifier] = ACTIONS(1598),
    [anon_sym_SEMI_SEMI] = ACTIONS(1600),
    [anon_sym_let] = ACTIONS(1598),
    [anon_sym_and] = ACTIONS(1598),
    [anon_sym_TILDE] = ACTIONS(1598),
    [anon_sym_QMARK] = ACTIONS(1598),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_external] = ACTIONS(1598),
    [anon_sym_type] = ACTIONS(1598),
    [anon_sym_COMMA] = ACTIONS(1600),
    [anon_sym_PIPE] = ACTIONS(1598),
    [anon_sym_DOT] = ACTIONS(1600),
    [anon_sym_LBRACK] = ACTIONS(1598),
    [anon_sym_RBRACK] = ACTIONS(1600),
    [anon_sym_true] = ACTIONS(1598),
    [anon_sym_false] = ACTIONS(1598),
    [anon_sym_LBRACE] = ACTIONS(1598),
    [anon_sym_SEMI] = ACTIONS(1598),
    [anon_sym_exception] = ACTIONS(1598),
    [anon_sym_module] = ACTIONS(1598),
    [anon_sym_open] = ACTIONS(1598),
    [anon_sym_include] = ACTIONS(1598),
    [anon_sym_class] = ACTIONS(1598),
    [anon_sym_val] = ACTIONS(1598),
    [anon_sym_end] = ACTIONS(1598),
    [anon_sym_object] = ACTIONS(1598),
    [anon_sym_AMP] = ACTIONS(1598),
    [anon_sym_POUND] = ACTIONS(1598),
    [anon_sym_COLON_COLON] = ACTIONS(1600),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1600),
    [anon_sym_else] = ACTIONS(1598),
    [anon_sym_new] = ACTIONS(1598),
    [anon_sym_LBRACE_LT] = ACTIONS(1600),
    [anon_sym_begin] = ACTIONS(1598),
    [sym_ocamlyacc_value] = ACTIONS(1600),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1598),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1600),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1598),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1598),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1600),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1600),
    [aux_sym_number_token1] = ACTIONS(1600),
    [aux_sym_signed_number_token1] = ACTIONS(1598),
    [anon_sym_SQUOTE] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1600),
    [sym_prefix_operator] = ACTIONS(1600),
    [aux_sym_sign_operator_token1] = ACTIONS(1598),
    [sym_hash_operator] = ACTIONS(1600),
    [anon_sym_lsl] = ACTIONS(1598),
    [anon_sym_lsr] = ACTIONS(1598),
    [anon_sym_asr] = ACTIONS(1598),
    [aux_sym_pow_operator_token1] = ACTIONS(1600),
    [anon_sym_mod] = ACTIONS(1598),
    [anon_sym_land] = ACTIONS(1598),
    [anon_sym_lor] = ACTIONS(1598),
    [anon_sym_lxor] = ACTIONS(1598),
    [aux_sym_mult_operator_token1] = ACTIONS(1598),
    [aux_sym_add_operator_token1] = ACTIONS(1598),
    [sym_concat_operator] = ACTIONS(1600),
    [sym_rel_operator] = ACTIONS(1598),
    [anon_sym_AMP_AMP] = ACTIONS(1598),
    [anon_sym_or] = ACTIONS(1598),
    [anon_sym_PIPE_PIPE] = ACTIONS(1598),
    [sym_assign_operator] = ACTIONS(1600),
    [sym_let_operator] = ACTIONS(1600),
    [sym__uppercase_identifier] = ACTIONS(1600),
    [aux_sym_directive_token1] = ACTIONS(1598),
    [aux_sym_tag_token1] = ACTIONS(1600),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1082)] = {
    [sym__parenthesized_abstract_type] = STATE(1332),
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4987),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1082),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2274),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2145),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_SEMI] = ACTIONS(2139),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_as] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2139),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1083)] = {
    [sym_attribute] = STATE(1083),
    [sym__lowercase_identifier] = ACTIONS(1614),
    [anon_sym_SEMI_SEMI] = ACTIONS(1616),
    [anon_sym_let] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_external] = ACTIONS(1614),
    [anon_sym_type] = ACTIONS(1614),
    [anon_sym_COMMA] = ACTIONS(1616),
    [anon_sym_PIPE] = ACTIONS(1614),
    [anon_sym_DOT] = ACTIONS(1616),
    [anon_sym_LBRACK] = ACTIONS(1614),
    [anon_sym_RBRACK] = ACTIONS(1616),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1614),
    [anon_sym_exception] = ACTIONS(1614),
    [anon_sym_module] = ACTIONS(1614),
    [anon_sym_open] = ACTIONS(1614),
    [anon_sym_include] = ACTIONS(1614),
    [anon_sym_class] = ACTIONS(1614),
    [anon_sym_val] = ACTIONS(1614),
    [anon_sym_end] = ACTIONS(1614),
    [anon_sym_object] = ACTIONS(1614),
    [anon_sym_AMP] = ACTIONS(1614),
    [anon_sym_POUND] = ACTIONS(1614),
    [anon_sym_COLON_COLON] = ACTIONS(1616),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1616),
    [anon_sym_LT_DASH] = ACTIONS(2009),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
    [anon_sym_LBRACE_LT] = ACTIONS(1616),
    [anon_sym_begin] = ACTIONS(1614),
    [sym_ocamlyacc_value] = ACTIONS(1616),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1614),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1616),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1616),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1616),
    [aux_sym_number_token1] = ACTIONS(1616),
    [aux_sym_signed_number_token1] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1616),
    [sym_prefix_operator] = ACTIONS(1616),
    [aux_sym_sign_operator_token1] = ACTIONS(1614),
    [sym_hash_operator] = ACTIONS(1616),
    [anon_sym_lsl] = ACTIONS(1614),
    [anon_sym_lsr] = ACTIONS(1614),
    [anon_sym_asr] = ACTIONS(1614),
    [aux_sym_pow_operator_token1] = ACTIONS(1616),
    [anon_sym_mod] = ACTIONS(1614),
    [anon_sym_land] = ACTIONS(1614),
    [anon_sym_lor] = ACTIONS(1614),
    [anon_sym_lxor] = ACTIONS(1614),
    [aux_sym_mult_operator_token1] = ACTIONS(1614),
    [aux_sym_add_operator_token1] = ACTIONS(1614),
    [sym_concat_operator] = ACTIONS(1616),
    [sym_rel_operator] = ACTIONS(1614),
    [anon_sym_AMP_AMP] = ACTIONS(1614),
    [anon_sym_or] = ACTIONS(1614),
    [anon_sym_PIPE_PIPE] = ACTIONS(1614),
    [sym_assign_operator] = ACTIONS(1616),
    [sym_let_operator] = ACTIONS(1616),
    [sym__uppercase_identifier] = ACTIONS(1616),
    [aux_sym_directive_token1] = ACTIONS(1614),
    [aux_sym_tag_token1] = ACTIONS(1616),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1084)] = {
    [sym_attribute] = STATE(1084),
    [sym__lowercase_identifier] = ACTIONS(1858),
    [anon_sym_SEMI_SEMI] = ACTIONS(1860),
    [anon_sym_let] = ACTIONS(1858),
    [anon_sym_and] = ACTIONS(1858),
    [anon_sym_TILDE] = ACTIONS(1858),
    [anon_sym_QMARK] = ACTIONS(1858),
    [anon_sym_LPAREN] = ACTIONS(1860),
    [anon_sym_external] = ACTIONS(1858),
    [anon_sym_type] = ACTIONS(1858),
    [anon_sym_COMMA] = ACTIONS(1860),
    [anon_sym_PIPE] = ACTIONS(1858),
    [anon_sym_DOT] = ACTIONS(1860),
    [anon_sym_LBRACK] = ACTIONS(1858),
    [anon_sym_RBRACK] = ACTIONS(1860),
    [anon_sym_true] = ACTIONS(1858),
    [anon_sym_false] = ACTIONS(1858),
    [anon_sym_LBRACE] = ACTIONS(1858),
    [anon_sym_SEMI] = ACTIONS(1858),
    [anon_sym_exception] = ACTIONS(1858),
    [anon_sym_module] = ACTIONS(1858),
    [anon_sym_open] = ACTIONS(1858),
    [anon_sym_include] = ACTIONS(1858),
    [anon_sym_class] = ACTIONS(1858),
    [anon_sym_val] = ACTIONS(1858),
    [anon_sym_end] = ACTIONS(1858),
    [anon_sym_object] = ACTIONS(1858),
    [anon_sym_AMP] = ACTIONS(1858),
    [anon_sym_POUND] = ACTIONS(1858),
    [anon_sym_COLON_COLON] = ACTIONS(1860),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1860),
    [anon_sym_else] = ACTIONS(1858),
    [anon_sym_new] = ACTIONS(1858),
    [anon_sym_LBRACE_LT] = ACTIONS(1860),
    [anon_sym_begin] = ACTIONS(1858),
    [sym_ocamlyacc_value] = ACTIONS(1860),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1858),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1860),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1858),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1858),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1860),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1860),
    [aux_sym_number_token1] = ACTIONS(1860),
    [aux_sym_signed_number_token1] = ACTIONS(1858),
    [anon_sym_SQUOTE] = ACTIONS(1860),
    [anon_sym_DQUOTE] = ACTIONS(1860),
    [sym_prefix_operator] = ACTIONS(1860),
    [aux_sym_sign_operator_token1] = ACTIONS(1858),
    [sym_hash_operator] = ACTIONS(1860),
    [anon_sym_lsl] = ACTIONS(1858),
    [anon_sym_lsr] = ACTIONS(1858),
    [anon_sym_asr] = ACTIONS(1858),
    [aux_sym_pow_operator_token1] = ACTIONS(1860),
    [anon_sym_mod] = ACTIONS(1858),
    [anon_sym_land] = ACTIONS(1858),
    [anon_sym_lor] = ACTIONS(1858),
    [anon_sym_lxor] = ACTIONS(1858),
    [aux_sym_mult_operator_token1] = ACTIONS(1858),
    [aux_sym_add_operator_token1] = ACTIONS(1858),
    [sym_concat_operator] = ACTIONS(1860),
    [sym_rel_operator] = ACTIONS(1858),
    [anon_sym_AMP_AMP] = ACTIONS(1858),
    [anon_sym_or] = ACTIONS(1858),
    [anon_sym_PIPE_PIPE] = ACTIONS(1858),
    [sym_assign_operator] = ACTIONS(1860),
    [sym_let_operator] = ACTIONS(1860),
    [sym__uppercase_identifier] = ACTIONS(1860),
    [aux_sym_directive_token1] = ACTIONS(1858),
    [aux_sym_tag_token1] = ACTIONS(1860),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1085)] = {
    [sym_attribute] = STATE(1085),
    [sym__lowercase_identifier] = ACTIONS(1862),
    [anon_sym_SEMI_SEMI] = ACTIONS(1864),
    [anon_sym_let] = ACTIONS(1862),
    [anon_sym_and] = ACTIONS(1862),
    [anon_sym_TILDE] = ACTIONS(1862),
    [anon_sym_QMARK] = ACTIONS(1862),
    [anon_sym_LPAREN] = ACTIONS(1864),
    [anon_sym_external] = ACTIONS(1862),
    [anon_sym_type] = ACTIONS(1862),
    [anon_sym_COMMA] = ACTIONS(1864),
    [anon_sym_PIPE] = ACTIONS(1862),
    [anon_sym_DOT] = ACTIONS(1864),
    [anon_sym_LBRACK] = ACTIONS(1862),
    [anon_sym_RBRACK] = ACTIONS(1864),
    [anon_sym_true] = ACTIONS(1862),
    [anon_sym_false] = ACTIONS(1862),
    [anon_sym_LBRACE] = ACTIONS(1862),
    [anon_sym_SEMI] = ACTIONS(1862),
    [anon_sym_exception] = ACTIONS(1862),
    [anon_sym_module] = ACTIONS(1862),
    [anon_sym_open] = ACTIONS(1862),
    [anon_sym_include] = ACTIONS(1862),
    [anon_sym_class] = ACTIONS(1862),
    [anon_sym_val] = ACTIONS(1862),
    [anon_sym_end] = ACTIONS(1862),
    [anon_sym_object] = ACTIONS(1862),
    [anon_sym_AMP] = ACTIONS(1862),
    [anon_sym_POUND] = ACTIONS(1862),
    [anon_sym_COLON_COLON] = ACTIONS(1864),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1864),
    [anon_sym_else] = ACTIONS(1862),
    [anon_sym_new] = ACTIONS(1862),
    [anon_sym_LBRACE_LT] = ACTIONS(1864),
    [anon_sym_begin] = ACTIONS(1862),
    [sym_ocamlyacc_value] = ACTIONS(1864),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1862),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1864),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1862),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1862),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1864),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1864),
    [aux_sym_number_token1] = ACTIONS(1864),
    [aux_sym_signed_number_token1] = ACTIONS(1862),
    [anon_sym_SQUOTE] = ACTIONS(1864),
    [anon_sym_DQUOTE] = ACTIONS(1864),
    [sym_prefix_operator] = ACTIONS(1864),
    [aux_sym_sign_operator_token1] = ACTIONS(1862),
    [sym_hash_operator] = ACTIONS(1864),
    [anon_sym_lsl] = ACTIONS(1862),
    [anon_sym_lsr] = ACTIONS(1862),
    [anon_sym_asr] = ACTIONS(1862),
    [aux_sym_pow_operator_token1] = ACTIONS(1864),
    [anon_sym_mod] = ACTIONS(1862),
    [anon_sym_land] = ACTIONS(1862),
    [anon_sym_lor] = ACTIONS(1862),
    [anon_sym_lxor] = ACTIONS(1862),
    [aux_sym_mult_operator_token1] = ACTIONS(1862),
    [aux_sym_add_operator_token1] = ACTIONS(1862),
    [sym_concat_operator] = ACTIONS(1864),
    [sym_rel_operator] = ACTIONS(1862),
    [anon_sym_AMP_AMP] = ACTIONS(1862),
    [anon_sym_or] = ACTIONS(1862),
    [anon_sym_PIPE_PIPE] = ACTIONS(1862),
    [sym_assign_operator] = ACTIONS(1864),
    [sym_let_operator] = ACTIONS(1864),
    [sym__uppercase_identifier] = ACTIONS(1864),
    [aux_sym_directive_token1] = ACTIONS(1862),
    [aux_sym_tag_token1] = ACTIONS(1864),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1086)] = {
    [sym_attribute] = STATE(1086),
    [sym__lowercase_identifier] = ACTIONS(1826),
    [anon_sym_SEMI_SEMI] = ACTIONS(1828),
    [anon_sym_let] = ACTIONS(1826),
    [anon_sym_and] = ACTIONS(1826),
    [anon_sym_TILDE] = ACTIONS(1826),
    [anon_sym_QMARK] = ACTIONS(1826),
    [anon_sym_LPAREN] = ACTIONS(1828),
    [anon_sym_external] = ACTIONS(1826),
    [anon_sym_type] = ACTIONS(1826),
    [anon_sym_COMMA] = ACTIONS(1828),
    [anon_sym_PIPE] = ACTIONS(1826),
    [anon_sym_DOT] = ACTIONS(1828),
    [anon_sym_LBRACK] = ACTIONS(1826),
    [anon_sym_RBRACK] = ACTIONS(1828),
    [anon_sym_true] = ACTIONS(1826),
    [anon_sym_false] = ACTIONS(1826),
    [anon_sym_LBRACE] = ACTIONS(1826),
    [anon_sym_SEMI] = ACTIONS(1826),
    [anon_sym_exception] = ACTIONS(1826),
    [anon_sym_module] = ACTIONS(1826),
    [anon_sym_open] = ACTIONS(1826),
    [anon_sym_include] = ACTIONS(1826),
    [anon_sym_class] = ACTIONS(1826),
    [anon_sym_val] = ACTIONS(1826),
    [anon_sym_end] = ACTIONS(1826),
    [anon_sym_object] = ACTIONS(1826),
    [anon_sym_AMP] = ACTIONS(1826),
    [anon_sym_POUND] = ACTIONS(1826),
    [anon_sym_COLON_COLON] = ACTIONS(1828),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1828),
    [anon_sym_else] = ACTIONS(1826),
    [anon_sym_new] = ACTIONS(1826),
    [anon_sym_LBRACE_LT] = ACTIONS(1828),
    [anon_sym_begin] = ACTIONS(1826),
    [sym_ocamlyacc_value] = ACTIONS(1828),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1826),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1828),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1826),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1826),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1828),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1828),
    [aux_sym_number_token1] = ACTIONS(1828),
    [aux_sym_signed_number_token1] = ACTIONS(1826),
    [anon_sym_SQUOTE] = ACTIONS(1828),
    [anon_sym_DQUOTE] = ACTIONS(1828),
    [sym_prefix_operator] = ACTIONS(1828),
    [aux_sym_sign_operator_token1] = ACTIONS(1826),
    [sym_hash_operator] = ACTIONS(1828),
    [anon_sym_lsl] = ACTIONS(1826),
    [anon_sym_lsr] = ACTIONS(1826),
    [anon_sym_asr] = ACTIONS(1826),
    [aux_sym_pow_operator_token1] = ACTIONS(1828),
    [anon_sym_mod] = ACTIONS(1826),
    [anon_sym_land] = ACTIONS(1826),
    [anon_sym_lor] = ACTIONS(1826),
    [anon_sym_lxor] = ACTIONS(1826),
    [aux_sym_mult_operator_token1] = ACTIONS(1826),
    [aux_sym_add_operator_token1] = ACTIONS(1826),
    [sym_concat_operator] = ACTIONS(1828),
    [sym_rel_operator] = ACTIONS(1826),
    [anon_sym_AMP_AMP] = ACTIONS(1826),
    [anon_sym_or] = ACTIONS(1826),
    [anon_sym_PIPE_PIPE] = ACTIONS(1826),
    [sym_assign_operator] = ACTIONS(1828),
    [sym_let_operator] = ACTIONS(1828),
    [sym__uppercase_identifier] = ACTIONS(1828),
    [aux_sym_directive_token1] = ACTIONS(1826),
    [aux_sym_tag_token1] = ACTIONS(1828),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1087)] = {
    [sym_attribute] = STATE(1087),
    [sym__lowercase_identifier] = ACTIONS(1720),
    [anon_sym_SEMI_SEMI] = ACTIONS(1722),
    [anon_sym_let] = ACTIONS(1720),
    [anon_sym_and] = ACTIONS(1720),
    [anon_sym_TILDE] = ACTIONS(1720),
    [anon_sym_QMARK] = ACTIONS(1720),
    [anon_sym_LPAREN] = ACTIONS(1722),
    [anon_sym_external] = ACTIONS(1720),
    [anon_sym_type] = ACTIONS(1720),
    [anon_sym_COMMA] = ACTIONS(1722),
    [anon_sym_PIPE] = ACTIONS(1720),
    [anon_sym_DOT] = ACTIONS(1722),
    [anon_sym_LBRACK] = ACTIONS(1720),
    [anon_sym_RBRACK] = ACTIONS(1722),
    [anon_sym_true] = ACTIONS(1720),
    [anon_sym_false] = ACTIONS(1720),
    [anon_sym_LBRACE] = ACTIONS(1720),
    [anon_sym_SEMI] = ACTIONS(1720),
    [anon_sym_exception] = ACTIONS(1720),
    [anon_sym_module] = ACTIONS(1720),
    [anon_sym_open] = ACTIONS(1720),
    [anon_sym_include] = ACTIONS(1720),
    [anon_sym_class] = ACTIONS(1720),
    [anon_sym_val] = ACTIONS(1720),
    [anon_sym_end] = ACTIONS(1720),
    [anon_sym_object] = ACTIONS(1720),
    [anon_sym_AMP] = ACTIONS(1720),
    [anon_sym_POUND] = ACTIONS(1720),
    [anon_sym_COLON_COLON] = ACTIONS(1722),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1722),
    [anon_sym_else] = ACTIONS(1720),
    [anon_sym_new] = ACTIONS(1720),
    [anon_sym_LBRACE_LT] = ACTIONS(1722),
    [anon_sym_begin] = ACTIONS(1720),
    [sym_ocamlyacc_value] = ACTIONS(1722),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1720),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1722),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1720),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1720),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1722),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1722),
    [aux_sym_number_token1] = ACTIONS(1722),
    [aux_sym_signed_number_token1] = ACTIONS(1720),
    [anon_sym_SQUOTE] = ACTIONS(1722),
    [anon_sym_DQUOTE] = ACTIONS(1722),
    [sym_prefix_operator] = ACTIONS(1722),
    [aux_sym_sign_operator_token1] = ACTIONS(1720),
    [sym_hash_operator] = ACTIONS(1722),
    [anon_sym_lsl] = ACTIONS(1720),
    [anon_sym_lsr] = ACTIONS(1720),
    [anon_sym_asr] = ACTIONS(1720),
    [aux_sym_pow_operator_token1] = ACTIONS(1722),
    [anon_sym_mod] = ACTIONS(1720),
    [anon_sym_land] = ACTIONS(1720),
    [anon_sym_lor] = ACTIONS(1720),
    [anon_sym_lxor] = ACTIONS(1720),
    [aux_sym_mult_operator_token1] = ACTIONS(1720),
    [aux_sym_add_operator_token1] = ACTIONS(1720),
    [sym_concat_operator] = ACTIONS(1722),
    [sym_rel_operator] = ACTIONS(1720),
    [anon_sym_AMP_AMP] = ACTIONS(1720),
    [anon_sym_or] = ACTIONS(1720),
    [anon_sym_PIPE_PIPE] = ACTIONS(1720),
    [sym_assign_operator] = ACTIONS(1722),
    [sym_let_operator] = ACTIONS(1722),
    [sym__uppercase_identifier] = ACTIONS(1722),
    [aux_sym_directive_token1] = ACTIONS(1720),
    [aux_sym_tag_token1] = ACTIONS(1722),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1088)] = {
    [sym_attribute] = STATE(1088),
    [sym__lowercase_identifier] = ACTIONS(2107),
    [anon_sym_SEMI_SEMI] = ACTIONS(2109),
    [anon_sym_let] = ACTIONS(2107),
    [anon_sym_and] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2107),
    [anon_sym_QMARK] = ACTIONS(2107),
    [anon_sym_LPAREN] = ACTIONS(2109),
    [anon_sym_external] = ACTIONS(2107),
    [anon_sym_type] = ACTIONS(2107),
    [anon_sym_COMMA] = ACTIONS(2109),
    [anon_sym_PIPE] = ACTIONS(2107),
    [anon_sym_DOT] = ACTIONS(963),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_RBRACK] = ACTIONS(2109),
    [anon_sym_true] = ACTIONS(2107),
    [anon_sym_false] = ACTIONS(2107),
    [anon_sym_LBRACE] = ACTIONS(2107),
    [anon_sym_SEMI] = ACTIONS(2107),
    [anon_sym_exception] = ACTIONS(2107),
    [anon_sym_module] = ACTIONS(2107),
    [anon_sym_open] = ACTIONS(2107),
    [anon_sym_include] = ACTIONS(2107),
    [anon_sym_class] = ACTIONS(2107),
    [anon_sym_val] = ACTIONS(2107),
    [anon_sym_end] = ACTIONS(2107),
    [anon_sym_object] = ACTIONS(2107),
    [anon_sym_AMP] = ACTIONS(2107),
    [anon_sym_POUND] = ACTIONS(965),
    [anon_sym_COLON_COLON] = ACTIONS(2109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2109),
    [anon_sym_else] = ACTIONS(2107),
    [anon_sym_new] = ACTIONS(2107),
    [anon_sym_LBRACE_LT] = ACTIONS(2109),
    [anon_sym_begin] = ACTIONS(2107),
    [sym_ocamlyacc_value] = ACTIONS(2109),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2107),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2109),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2109),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2109),
    [aux_sym_number_token1] = ACTIONS(2109),
    [aux_sym_signed_number_token1] = ACTIONS(2107),
    [anon_sym_SQUOTE] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(2109),
    [sym_prefix_operator] = ACTIONS(2109),
    [aux_sym_sign_operator_token1] = ACTIONS(2107),
    [sym_hash_operator] = ACTIONS(967),
    [anon_sym_lsl] = ACTIONS(2107),
    [anon_sym_lsr] = ACTIONS(2107),
    [anon_sym_asr] = ACTIONS(2107),
    [aux_sym_pow_operator_token1] = ACTIONS(2109),
    [anon_sym_mod] = ACTIONS(2107),
    [anon_sym_land] = ACTIONS(2107),
    [anon_sym_lor] = ACTIONS(2107),
    [anon_sym_lxor] = ACTIONS(2107),
    [aux_sym_mult_operator_token1] = ACTIONS(2107),
    [aux_sym_add_operator_token1] = ACTIONS(2107),
    [sym_concat_operator] = ACTIONS(2109),
    [sym_rel_operator] = ACTIONS(2107),
    [anon_sym_AMP_AMP] = ACTIONS(2107),
    [anon_sym_or] = ACTIONS(2107),
    [anon_sym_PIPE_PIPE] = ACTIONS(2107),
    [sym_assign_operator] = ACTIONS(2109),
    [sym_let_operator] = ACTIONS(2109),
    [sym__uppercase_identifier] = ACTIONS(2109),
    [aux_sym_directive_token1] = ACTIONS(2107),
    [aux_sym_tag_token1] = ACTIONS(2109),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1089)] = {
    [sym__parenthesized_abstract_type] = STATE(1343),
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4854),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1089),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2312),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1318),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_SEMI] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_as] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_COLON_COLON] = ACTIONS(1316),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1316),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1090)] = {
    [sym_attribute] = STATE(1090),
    [sym__lowercase_identifier] = ACTIONS(1898),
    [anon_sym_SEMI_SEMI] = ACTIONS(1900),
    [anon_sym_let] = ACTIONS(1898),
    [anon_sym_and] = ACTIONS(1898),
    [anon_sym_TILDE] = ACTIONS(1898),
    [anon_sym_QMARK] = ACTIONS(1898),
    [anon_sym_LPAREN] = ACTIONS(1900),
    [anon_sym_external] = ACTIONS(1898),
    [anon_sym_type] = ACTIONS(1898),
    [anon_sym_COMMA] = ACTIONS(1900),
    [anon_sym_PIPE] = ACTIONS(1898),
    [anon_sym_DOT] = ACTIONS(1900),
    [anon_sym_LBRACK] = ACTIONS(1898),
    [anon_sym_RBRACK] = ACTIONS(1900),
    [anon_sym_true] = ACTIONS(1898),
    [anon_sym_false] = ACTIONS(1898),
    [anon_sym_LBRACE] = ACTIONS(1898),
    [anon_sym_SEMI] = ACTIONS(1898),
    [anon_sym_exception] = ACTIONS(1898),
    [anon_sym_module] = ACTIONS(1898),
    [anon_sym_open] = ACTIONS(1898),
    [anon_sym_include] = ACTIONS(1898),
    [anon_sym_class] = ACTIONS(1898),
    [anon_sym_val] = ACTIONS(1898),
    [anon_sym_end] = ACTIONS(1898),
    [anon_sym_object] = ACTIONS(1898),
    [anon_sym_AMP] = ACTIONS(1898),
    [anon_sym_POUND] = ACTIONS(1898),
    [anon_sym_COLON_COLON] = ACTIONS(1900),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1900),
    [anon_sym_else] = ACTIONS(1898),
    [anon_sym_new] = ACTIONS(1898),
    [anon_sym_LBRACE_LT] = ACTIONS(1900),
    [anon_sym_begin] = ACTIONS(1898),
    [sym_ocamlyacc_value] = ACTIONS(1900),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1898),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1900),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1898),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1898),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1900),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1900),
    [aux_sym_number_token1] = ACTIONS(1900),
    [aux_sym_signed_number_token1] = ACTIONS(1898),
    [anon_sym_SQUOTE] = ACTIONS(1900),
    [anon_sym_DQUOTE] = ACTIONS(1900),
    [sym_prefix_operator] = ACTIONS(1900),
    [aux_sym_sign_operator_token1] = ACTIONS(1898),
    [sym_hash_operator] = ACTIONS(1900),
    [anon_sym_lsl] = ACTIONS(1898),
    [anon_sym_lsr] = ACTIONS(1898),
    [anon_sym_asr] = ACTIONS(1898),
    [aux_sym_pow_operator_token1] = ACTIONS(1900),
    [anon_sym_mod] = ACTIONS(1898),
    [anon_sym_land] = ACTIONS(1898),
    [anon_sym_lor] = ACTIONS(1898),
    [anon_sym_lxor] = ACTIONS(1898),
    [aux_sym_mult_operator_token1] = ACTIONS(1898),
    [aux_sym_add_operator_token1] = ACTIONS(1898),
    [sym_concat_operator] = ACTIONS(1900),
    [sym_rel_operator] = ACTIONS(1898),
    [anon_sym_AMP_AMP] = ACTIONS(1898),
    [anon_sym_or] = ACTIONS(1898),
    [anon_sym_PIPE_PIPE] = ACTIONS(1898),
    [sym_assign_operator] = ACTIONS(1900),
    [sym_let_operator] = ACTIONS(1900),
    [sym__uppercase_identifier] = ACTIONS(1900),
    [aux_sym_directive_token1] = ACTIONS(1898),
    [aux_sym_tag_token1] = ACTIONS(1900),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1091)] = {
    [sym_attribute] = STATE(1091),
    [sym__lowercase_identifier] = ACTIONS(1950),
    [anon_sym_SEMI_SEMI] = ACTIONS(1952),
    [anon_sym_let] = ACTIONS(1950),
    [anon_sym_and] = ACTIONS(1950),
    [anon_sym_TILDE] = ACTIONS(1950),
    [anon_sym_QMARK] = ACTIONS(1950),
    [anon_sym_LPAREN] = ACTIONS(1952),
    [anon_sym_external] = ACTIONS(1950),
    [anon_sym_type] = ACTIONS(1950),
    [anon_sym_COMMA] = ACTIONS(1952),
    [anon_sym_PIPE] = ACTIONS(1950),
    [anon_sym_DOT] = ACTIONS(1952),
    [anon_sym_LBRACK] = ACTIONS(1950),
    [anon_sym_RBRACK] = ACTIONS(1952),
    [anon_sym_true] = ACTIONS(1950),
    [anon_sym_false] = ACTIONS(1950),
    [anon_sym_LBRACE] = ACTIONS(1950),
    [anon_sym_SEMI] = ACTIONS(1950),
    [anon_sym_exception] = ACTIONS(1950),
    [anon_sym_module] = ACTIONS(1950),
    [anon_sym_open] = ACTIONS(1950),
    [anon_sym_include] = ACTIONS(1950),
    [anon_sym_class] = ACTIONS(1950),
    [anon_sym_val] = ACTIONS(1950),
    [anon_sym_end] = ACTIONS(1950),
    [anon_sym_object] = ACTIONS(1950),
    [anon_sym_AMP] = ACTIONS(1950),
    [anon_sym_POUND] = ACTIONS(1950),
    [anon_sym_COLON_COLON] = ACTIONS(1952),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1952),
    [anon_sym_else] = ACTIONS(1950),
    [anon_sym_new] = ACTIONS(1950),
    [anon_sym_LBRACE_LT] = ACTIONS(1952),
    [anon_sym_begin] = ACTIONS(1950),
    [sym_ocamlyacc_value] = ACTIONS(1952),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1950),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1952),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1950),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1950),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1952),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1952),
    [aux_sym_number_token1] = ACTIONS(1952),
    [aux_sym_signed_number_token1] = ACTIONS(1950),
    [anon_sym_SQUOTE] = ACTIONS(1952),
    [anon_sym_DQUOTE] = ACTIONS(1952),
    [sym_prefix_operator] = ACTIONS(1952),
    [aux_sym_sign_operator_token1] = ACTIONS(1950),
    [sym_hash_operator] = ACTIONS(1952),
    [anon_sym_lsl] = ACTIONS(1950),
    [anon_sym_lsr] = ACTIONS(1950),
    [anon_sym_asr] = ACTIONS(1950),
    [aux_sym_pow_operator_token1] = ACTIONS(1952),
    [anon_sym_mod] = ACTIONS(1950),
    [anon_sym_land] = ACTIONS(1950),
    [anon_sym_lor] = ACTIONS(1950),
    [anon_sym_lxor] = ACTIONS(1950),
    [aux_sym_mult_operator_token1] = ACTIONS(1950),
    [aux_sym_add_operator_token1] = ACTIONS(1950),
    [sym_concat_operator] = ACTIONS(1952),
    [sym_rel_operator] = ACTIONS(1950),
    [anon_sym_AMP_AMP] = ACTIONS(1950),
    [anon_sym_or] = ACTIONS(1950),
    [anon_sym_PIPE_PIPE] = ACTIONS(1950),
    [sym_assign_operator] = ACTIONS(1952),
    [sym_let_operator] = ACTIONS(1952),
    [sym__uppercase_identifier] = ACTIONS(1952),
    [aux_sym_directive_token1] = ACTIONS(1950),
    [aux_sym_tag_token1] = ACTIONS(1952),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1092)] = {
    [sym_attribute] = STATE(1092),
    [sym__lowercase_identifier] = ACTIONS(1874),
    [anon_sym_SEMI_SEMI] = ACTIONS(1876),
    [anon_sym_let] = ACTIONS(1874),
    [anon_sym_and] = ACTIONS(1874),
    [anon_sym_TILDE] = ACTIONS(1874),
    [anon_sym_QMARK] = ACTIONS(1874),
    [anon_sym_LPAREN] = ACTIONS(1876),
    [anon_sym_external] = ACTIONS(1874),
    [anon_sym_type] = ACTIONS(1874),
    [anon_sym_COMMA] = ACTIONS(1876),
    [anon_sym_PIPE] = ACTIONS(1874),
    [anon_sym_DOT] = ACTIONS(1876),
    [anon_sym_LBRACK] = ACTIONS(1874),
    [anon_sym_RBRACK] = ACTIONS(1876),
    [anon_sym_true] = ACTIONS(1874),
    [anon_sym_false] = ACTIONS(1874),
    [anon_sym_LBRACE] = ACTIONS(1874),
    [anon_sym_SEMI] = ACTIONS(1874),
    [anon_sym_exception] = ACTIONS(1874),
    [anon_sym_module] = ACTIONS(1874),
    [anon_sym_open] = ACTIONS(1874),
    [anon_sym_include] = ACTIONS(1874),
    [anon_sym_class] = ACTIONS(1874),
    [anon_sym_val] = ACTIONS(1874),
    [anon_sym_end] = ACTIONS(1874),
    [anon_sym_object] = ACTIONS(1874),
    [anon_sym_AMP] = ACTIONS(1874),
    [anon_sym_POUND] = ACTIONS(1874),
    [anon_sym_COLON_COLON] = ACTIONS(1876),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1876),
    [anon_sym_else] = ACTIONS(1874),
    [anon_sym_new] = ACTIONS(1874),
    [anon_sym_LBRACE_LT] = ACTIONS(1876),
    [anon_sym_begin] = ACTIONS(1874),
    [sym_ocamlyacc_value] = ACTIONS(1876),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1874),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1876),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1874),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1874),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1876),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1876),
    [aux_sym_number_token1] = ACTIONS(1876),
    [aux_sym_signed_number_token1] = ACTIONS(1874),
    [anon_sym_SQUOTE] = ACTIONS(1876),
    [anon_sym_DQUOTE] = ACTIONS(1876),
    [sym_prefix_operator] = ACTIONS(1876),
    [aux_sym_sign_operator_token1] = ACTIONS(1874),
    [sym_hash_operator] = ACTIONS(1876),
    [anon_sym_lsl] = ACTIONS(1874),
    [anon_sym_lsr] = ACTIONS(1874),
    [anon_sym_asr] = ACTIONS(1874),
    [aux_sym_pow_operator_token1] = ACTIONS(1876),
    [anon_sym_mod] = ACTIONS(1874),
    [anon_sym_land] = ACTIONS(1874),
    [anon_sym_lor] = ACTIONS(1874),
    [anon_sym_lxor] = ACTIONS(1874),
    [aux_sym_mult_operator_token1] = ACTIONS(1874),
    [aux_sym_add_operator_token1] = ACTIONS(1874),
    [sym_concat_operator] = ACTIONS(1876),
    [sym_rel_operator] = ACTIONS(1874),
    [anon_sym_AMP_AMP] = ACTIONS(1874),
    [anon_sym_or] = ACTIONS(1874),
    [anon_sym_PIPE_PIPE] = ACTIONS(1874),
    [sym_assign_operator] = ACTIONS(1876),
    [sym_let_operator] = ACTIONS(1876),
    [sym__uppercase_identifier] = ACTIONS(1876),
    [aux_sym_directive_token1] = ACTIONS(1874),
    [aux_sym_tag_token1] = ACTIONS(1876),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1093)] = {
    [sym_attribute] = STATE(1093),
    [sym__lowercase_identifier] = ACTIONS(1918),
    [anon_sym_SEMI_SEMI] = ACTIONS(1920),
    [anon_sym_let] = ACTIONS(1918),
    [anon_sym_and] = ACTIONS(1918),
    [anon_sym_TILDE] = ACTIONS(1918),
    [anon_sym_QMARK] = ACTIONS(1918),
    [anon_sym_LPAREN] = ACTIONS(1920),
    [anon_sym_external] = ACTIONS(1918),
    [anon_sym_type] = ACTIONS(1918),
    [anon_sym_COMMA] = ACTIONS(1920),
    [anon_sym_PIPE] = ACTIONS(1918),
    [anon_sym_DOT] = ACTIONS(1920),
    [anon_sym_LBRACK] = ACTIONS(1918),
    [anon_sym_RBRACK] = ACTIONS(1920),
    [anon_sym_true] = ACTIONS(1918),
    [anon_sym_false] = ACTIONS(1918),
    [anon_sym_LBRACE] = ACTIONS(1918),
    [anon_sym_SEMI] = ACTIONS(1918),
    [anon_sym_exception] = ACTIONS(1918),
    [anon_sym_module] = ACTIONS(1918),
    [anon_sym_open] = ACTIONS(1918),
    [anon_sym_include] = ACTIONS(1918),
    [anon_sym_class] = ACTIONS(1918),
    [anon_sym_val] = ACTIONS(1918),
    [anon_sym_end] = ACTIONS(1918),
    [anon_sym_object] = ACTIONS(1918),
    [anon_sym_AMP] = ACTIONS(1918),
    [anon_sym_POUND] = ACTIONS(1918),
    [anon_sym_COLON_COLON] = ACTIONS(1920),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1920),
    [anon_sym_else] = ACTIONS(1918),
    [anon_sym_new] = ACTIONS(1918),
    [anon_sym_LBRACE_LT] = ACTIONS(1920),
    [anon_sym_begin] = ACTIONS(1918),
    [sym_ocamlyacc_value] = ACTIONS(1920),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1918),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1920),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1918),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1918),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1920),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1920),
    [aux_sym_number_token1] = ACTIONS(1920),
    [aux_sym_signed_number_token1] = ACTIONS(1918),
    [anon_sym_SQUOTE] = ACTIONS(1920),
    [anon_sym_DQUOTE] = ACTIONS(1920),
    [sym_prefix_operator] = ACTIONS(1920),
    [aux_sym_sign_operator_token1] = ACTIONS(1918),
    [sym_hash_operator] = ACTIONS(1920),
    [anon_sym_lsl] = ACTIONS(1918),
    [anon_sym_lsr] = ACTIONS(1918),
    [anon_sym_asr] = ACTIONS(1918),
    [aux_sym_pow_operator_token1] = ACTIONS(1920),
    [anon_sym_mod] = ACTIONS(1918),
    [anon_sym_land] = ACTIONS(1918),
    [anon_sym_lor] = ACTIONS(1918),
    [anon_sym_lxor] = ACTIONS(1918),
    [aux_sym_mult_operator_token1] = ACTIONS(1918),
    [aux_sym_add_operator_token1] = ACTIONS(1918),
    [sym_concat_operator] = ACTIONS(1920),
    [sym_rel_operator] = ACTIONS(1918),
    [anon_sym_AMP_AMP] = ACTIONS(1918),
    [anon_sym_or] = ACTIONS(1918),
    [anon_sym_PIPE_PIPE] = ACTIONS(1918),
    [sym_assign_operator] = ACTIONS(1920),
    [sym_let_operator] = ACTIONS(1920),
    [sym__uppercase_identifier] = ACTIONS(1920),
    [aux_sym_directive_token1] = ACTIONS(1918),
    [aux_sym_tag_token1] = ACTIONS(1920),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1094)] = {
    [sym_attribute] = STATE(1094),
    [sym__lowercase_identifier] = ACTIONS(1870),
    [anon_sym_SEMI_SEMI] = ACTIONS(1872),
    [anon_sym_let] = ACTIONS(1870),
    [anon_sym_and] = ACTIONS(1870),
    [anon_sym_TILDE] = ACTIONS(1870),
    [anon_sym_QMARK] = ACTIONS(1870),
    [anon_sym_LPAREN] = ACTIONS(1872),
    [anon_sym_external] = ACTIONS(1870),
    [anon_sym_type] = ACTIONS(1870),
    [anon_sym_COMMA] = ACTIONS(1872),
    [anon_sym_PIPE] = ACTIONS(1870),
    [anon_sym_DOT] = ACTIONS(1872),
    [anon_sym_LBRACK] = ACTIONS(1870),
    [anon_sym_RBRACK] = ACTIONS(1872),
    [anon_sym_true] = ACTIONS(1870),
    [anon_sym_false] = ACTIONS(1870),
    [anon_sym_LBRACE] = ACTIONS(1870),
    [anon_sym_SEMI] = ACTIONS(1870),
    [anon_sym_exception] = ACTIONS(1870),
    [anon_sym_module] = ACTIONS(1870),
    [anon_sym_open] = ACTIONS(1870),
    [anon_sym_include] = ACTIONS(1870),
    [anon_sym_class] = ACTIONS(1870),
    [anon_sym_val] = ACTIONS(1870),
    [anon_sym_end] = ACTIONS(1870),
    [anon_sym_object] = ACTIONS(1870),
    [anon_sym_AMP] = ACTIONS(1870),
    [anon_sym_POUND] = ACTIONS(1870),
    [anon_sym_COLON_COLON] = ACTIONS(1872),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1872),
    [anon_sym_else] = ACTIONS(1870),
    [anon_sym_new] = ACTIONS(1870),
    [anon_sym_LBRACE_LT] = ACTIONS(1872),
    [anon_sym_begin] = ACTIONS(1870),
    [sym_ocamlyacc_value] = ACTIONS(1872),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1870),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1872),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1870),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1870),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1872),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1872),
    [aux_sym_number_token1] = ACTIONS(1872),
    [aux_sym_signed_number_token1] = ACTIONS(1870),
    [anon_sym_SQUOTE] = ACTIONS(1872),
    [anon_sym_DQUOTE] = ACTIONS(1872),
    [sym_prefix_operator] = ACTIONS(1872),
    [aux_sym_sign_operator_token1] = ACTIONS(1870),
    [sym_hash_operator] = ACTIONS(1872),
    [anon_sym_lsl] = ACTIONS(1870),
    [anon_sym_lsr] = ACTIONS(1870),
    [anon_sym_asr] = ACTIONS(1870),
    [aux_sym_pow_operator_token1] = ACTIONS(1872),
    [anon_sym_mod] = ACTIONS(1870),
    [anon_sym_land] = ACTIONS(1870),
    [anon_sym_lor] = ACTIONS(1870),
    [anon_sym_lxor] = ACTIONS(1870),
    [aux_sym_mult_operator_token1] = ACTIONS(1870),
    [aux_sym_add_operator_token1] = ACTIONS(1870),
    [sym_concat_operator] = ACTIONS(1872),
    [sym_rel_operator] = ACTIONS(1870),
    [anon_sym_AMP_AMP] = ACTIONS(1870),
    [anon_sym_or] = ACTIONS(1870),
    [anon_sym_PIPE_PIPE] = ACTIONS(1870),
    [sym_assign_operator] = ACTIONS(1872),
    [sym_let_operator] = ACTIONS(1872),
    [sym__uppercase_identifier] = ACTIONS(1872),
    [aux_sym_directive_token1] = ACTIONS(1870),
    [aux_sym_tag_token1] = ACTIONS(1872),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1095)] = {
    [sym_attribute] = STATE(1095),
    [sym__lowercase_identifier] = ACTIONS(1708),
    [anon_sym_SEMI_SEMI] = ACTIONS(1710),
    [anon_sym_let] = ACTIONS(1708),
    [anon_sym_and] = ACTIONS(1708),
    [anon_sym_TILDE] = ACTIONS(1708),
    [anon_sym_QMARK] = ACTIONS(1708),
    [anon_sym_LPAREN] = ACTIONS(1710),
    [anon_sym_external] = ACTIONS(1708),
    [anon_sym_type] = ACTIONS(1708),
    [anon_sym_COMMA] = ACTIONS(1710),
    [anon_sym_PIPE] = ACTIONS(1708),
    [anon_sym_DOT] = ACTIONS(1710),
    [anon_sym_LBRACK] = ACTIONS(1708),
    [anon_sym_RBRACK] = ACTIONS(1710),
    [anon_sym_true] = ACTIONS(1708),
    [anon_sym_false] = ACTIONS(1708),
    [anon_sym_LBRACE] = ACTIONS(1708),
    [anon_sym_SEMI] = ACTIONS(1708),
    [anon_sym_exception] = ACTIONS(1708),
    [anon_sym_module] = ACTIONS(1708),
    [anon_sym_open] = ACTIONS(1708),
    [anon_sym_include] = ACTIONS(1708),
    [anon_sym_class] = ACTIONS(1708),
    [anon_sym_val] = ACTIONS(1708),
    [anon_sym_end] = ACTIONS(1708),
    [anon_sym_object] = ACTIONS(1708),
    [anon_sym_AMP] = ACTIONS(1708),
    [anon_sym_POUND] = ACTIONS(1708),
    [anon_sym_COLON_COLON] = ACTIONS(1710),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1710),
    [anon_sym_else] = ACTIONS(1708),
    [anon_sym_new] = ACTIONS(1708),
    [anon_sym_LBRACE_LT] = ACTIONS(1710),
    [anon_sym_begin] = ACTIONS(1708),
    [sym_ocamlyacc_value] = ACTIONS(1710),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1708),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1710),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1708),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1708),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1710),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1710),
    [aux_sym_number_token1] = ACTIONS(1710),
    [aux_sym_signed_number_token1] = ACTIONS(1708),
    [anon_sym_SQUOTE] = ACTIONS(1710),
    [anon_sym_DQUOTE] = ACTIONS(1710),
    [sym_prefix_operator] = ACTIONS(1710),
    [aux_sym_sign_operator_token1] = ACTIONS(1708),
    [sym_hash_operator] = ACTIONS(1710),
    [anon_sym_lsl] = ACTIONS(1708),
    [anon_sym_lsr] = ACTIONS(1708),
    [anon_sym_asr] = ACTIONS(1708),
    [aux_sym_pow_operator_token1] = ACTIONS(1710),
    [anon_sym_mod] = ACTIONS(1708),
    [anon_sym_land] = ACTIONS(1708),
    [anon_sym_lor] = ACTIONS(1708),
    [anon_sym_lxor] = ACTIONS(1708),
    [aux_sym_mult_operator_token1] = ACTIONS(1708),
    [aux_sym_add_operator_token1] = ACTIONS(1708),
    [sym_concat_operator] = ACTIONS(1710),
    [sym_rel_operator] = ACTIONS(1708),
    [anon_sym_AMP_AMP] = ACTIONS(1708),
    [anon_sym_or] = ACTIONS(1708),
    [anon_sym_PIPE_PIPE] = ACTIONS(1708),
    [sym_assign_operator] = ACTIONS(1710),
    [sym_let_operator] = ACTIONS(1710),
    [sym__uppercase_identifier] = ACTIONS(1710),
    [aux_sym_directive_token1] = ACTIONS(1708),
    [aux_sym_tag_token1] = ACTIONS(1710),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1096)] = {
    [sym_attribute] = STATE(1096),
    [sym__lowercase_identifier] = ACTIONS(1922),
    [anon_sym_SEMI_SEMI] = ACTIONS(1924),
    [anon_sym_let] = ACTIONS(1922),
    [anon_sym_and] = ACTIONS(1922),
    [anon_sym_TILDE] = ACTIONS(1922),
    [anon_sym_QMARK] = ACTIONS(1922),
    [anon_sym_LPAREN] = ACTIONS(1924),
    [anon_sym_external] = ACTIONS(1922),
    [anon_sym_type] = ACTIONS(1922),
    [anon_sym_COMMA] = ACTIONS(1924),
    [anon_sym_PIPE] = ACTIONS(1922),
    [anon_sym_DOT] = ACTIONS(1924),
    [anon_sym_LBRACK] = ACTIONS(1922),
    [anon_sym_RBRACK] = ACTIONS(1924),
    [anon_sym_true] = ACTIONS(1922),
    [anon_sym_false] = ACTIONS(1922),
    [anon_sym_LBRACE] = ACTIONS(1922),
    [anon_sym_SEMI] = ACTIONS(1922),
    [anon_sym_exception] = ACTIONS(1922),
    [anon_sym_module] = ACTIONS(1922),
    [anon_sym_open] = ACTIONS(1922),
    [anon_sym_include] = ACTIONS(1922),
    [anon_sym_class] = ACTIONS(1922),
    [anon_sym_val] = ACTIONS(1922),
    [anon_sym_end] = ACTIONS(1922),
    [anon_sym_object] = ACTIONS(1922),
    [anon_sym_AMP] = ACTIONS(1922),
    [anon_sym_POUND] = ACTIONS(1922),
    [anon_sym_COLON_COLON] = ACTIONS(1924),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1924),
    [anon_sym_else] = ACTIONS(1922),
    [anon_sym_new] = ACTIONS(1922),
    [anon_sym_LBRACE_LT] = ACTIONS(1924),
    [anon_sym_begin] = ACTIONS(1922),
    [sym_ocamlyacc_value] = ACTIONS(1924),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1922),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1924),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1922),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1922),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1924),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1924),
    [aux_sym_number_token1] = ACTIONS(1924),
    [aux_sym_signed_number_token1] = ACTIONS(1922),
    [anon_sym_SQUOTE] = ACTIONS(1924),
    [anon_sym_DQUOTE] = ACTIONS(1924),
    [sym_prefix_operator] = ACTIONS(1924),
    [aux_sym_sign_operator_token1] = ACTIONS(1922),
    [sym_hash_operator] = ACTIONS(1924),
    [anon_sym_lsl] = ACTIONS(1922),
    [anon_sym_lsr] = ACTIONS(1922),
    [anon_sym_asr] = ACTIONS(1922),
    [aux_sym_pow_operator_token1] = ACTIONS(1924),
    [anon_sym_mod] = ACTIONS(1922),
    [anon_sym_land] = ACTIONS(1922),
    [anon_sym_lor] = ACTIONS(1922),
    [anon_sym_lxor] = ACTIONS(1922),
    [aux_sym_mult_operator_token1] = ACTIONS(1922),
    [aux_sym_add_operator_token1] = ACTIONS(1922),
    [sym_concat_operator] = ACTIONS(1924),
    [sym_rel_operator] = ACTIONS(1922),
    [anon_sym_AMP_AMP] = ACTIONS(1922),
    [anon_sym_or] = ACTIONS(1922),
    [anon_sym_PIPE_PIPE] = ACTIONS(1922),
    [sym_assign_operator] = ACTIONS(1924),
    [sym_let_operator] = ACTIONS(1924),
    [sym__uppercase_identifier] = ACTIONS(1924),
    [aux_sym_directive_token1] = ACTIONS(1922),
    [aux_sym_tag_token1] = ACTIONS(1924),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1097)] = {
    [sym_attribute] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1878),
    [anon_sym_SEMI_SEMI] = ACTIONS(1880),
    [anon_sym_let] = ACTIONS(1878),
    [anon_sym_and] = ACTIONS(1878),
    [anon_sym_TILDE] = ACTIONS(1878),
    [anon_sym_QMARK] = ACTIONS(1878),
    [anon_sym_LPAREN] = ACTIONS(1880),
    [anon_sym_external] = ACTIONS(1878),
    [anon_sym_type] = ACTIONS(1878),
    [anon_sym_COMMA] = ACTIONS(1880),
    [anon_sym_PIPE] = ACTIONS(1878),
    [anon_sym_DOT] = ACTIONS(1880),
    [anon_sym_LBRACK] = ACTIONS(1878),
    [anon_sym_RBRACK] = ACTIONS(1880),
    [anon_sym_true] = ACTIONS(1878),
    [anon_sym_false] = ACTIONS(1878),
    [anon_sym_LBRACE] = ACTIONS(1878),
    [anon_sym_SEMI] = ACTIONS(1878),
    [anon_sym_exception] = ACTIONS(1878),
    [anon_sym_module] = ACTIONS(1878),
    [anon_sym_open] = ACTIONS(1878),
    [anon_sym_include] = ACTIONS(1878),
    [anon_sym_class] = ACTIONS(1878),
    [anon_sym_val] = ACTIONS(1878),
    [anon_sym_end] = ACTIONS(1878),
    [anon_sym_object] = ACTIONS(1878),
    [anon_sym_AMP] = ACTIONS(1878),
    [anon_sym_POUND] = ACTIONS(1878),
    [anon_sym_COLON_COLON] = ACTIONS(1880),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1880),
    [anon_sym_else] = ACTIONS(1878),
    [anon_sym_new] = ACTIONS(1878),
    [anon_sym_LBRACE_LT] = ACTIONS(1880),
    [anon_sym_begin] = ACTIONS(1878),
    [sym_ocamlyacc_value] = ACTIONS(1880),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1878),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1880),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1878),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1878),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1880),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1880),
    [aux_sym_number_token1] = ACTIONS(1880),
    [aux_sym_signed_number_token1] = ACTIONS(1878),
    [anon_sym_SQUOTE] = ACTIONS(1880),
    [anon_sym_DQUOTE] = ACTIONS(1880),
    [sym_prefix_operator] = ACTIONS(1880),
    [aux_sym_sign_operator_token1] = ACTIONS(1878),
    [sym_hash_operator] = ACTIONS(1880),
    [anon_sym_lsl] = ACTIONS(1878),
    [anon_sym_lsr] = ACTIONS(1878),
    [anon_sym_asr] = ACTIONS(1878),
    [aux_sym_pow_operator_token1] = ACTIONS(1880),
    [anon_sym_mod] = ACTIONS(1878),
    [anon_sym_land] = ACTIONS(1878),
    [anon_sym_lor] = ACTIONS(1878),
    [anon_sym_lxor] = ACTIONS(1878),
    [aux_sym_mult_operator_token1] = ACTIONS(1878),
    [aux_sym_add_operator_token1] = ACTIONS(1878),
    [sym_concat_operator] = ACTIONS(1880),
    [sym_rel_operator] = ACTIONS(1878),
    [anon_sym_AMP_AMP] = ACTIONS(1878),
    [anon_sym_or] = ACTIONS(1878),
    [anon_sym_PIPE_PIPE] = ACTIONS(1878),
    [sym_assign_operator] = ACTIONS(1880),
    [sym_let_operator] = ACTIONS(1880),
    [sym__uppercase_identifier] = ACTIONS(1880),
    [aux_sym_directive_token1] = ACTIONS(1878),
    [aux_sym_tag_token1] = ACTIONS(1880),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1098)] = {
    [sym_attribute] = STATE(1098),
    [sym__lowercase_identifier] = ACTIONS(1602),
    [anon_sym_SEMI_SEMI] = ACTIONS(1604),
    [anon_sym_let] = ACTIONS(1602),
    [anon_sym_and] = ACTIONS(1602),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_QMARK] = ACTIONS(1602),
    [anon_sym_LPAREN] = ACTIONS(1604),
    [anon_sym_external] = ACTIONS(1602),
    [anon_sym_type] = ACTIONS(1602),
    [anon_sym_COMMA] = ACTIONS(1604),
    [anon_sym_PIPE] = ACTIONS(1602),
    [anon_sym_DOT] = ACTIONS(1604),
    [anon_sym_LBRACK] = ACTIONS(1602),
    [anon_sym_RBRACK] = ACTIONS(1604),
    [anon_sym_true] = ACTIONS(1602),
    [anon_sym_false] = ACTIONS(1602),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [anon_sym_SEMI] = ACTIONS(1602),
    [anon_sym_exception] = ACTIONS(1602),
    [anon_sym_module] = ACTIONS(1602),
    [anon_sym_open] = ACTIONS(1602),
    [anon_sym_include] = ACTIONS(1602),
    [anon_sym_class] = ACTIONS(1602),
    [anon_sym_val] = ACTIONS(1602),
    [anon_sym_end] = ACTIONS(1602),
    [anon_sym_object] = ACTIONS(1602),
    [anon_sym_AMP] = ACTIONS(1602),
    [anon_sym_POUND] = ACTIONS(1602),
    [anon_sym_COLON_COLON] = ACTIONS(1604),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1604),
    [anon_sym_else] = ACTIONS(1602),
    [anon_sym_new] = ACTIONS(1602),
    [anon_sym_LBRACE_LT] = ACTIONS(1604),
    [anon_sym_begin] = ACTIONS(1602),
    [sym_ocamlyacc_value] = ACTIONS(1604),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1602),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1604),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1602),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1602),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1604),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1604),
    [aux_sym_number_token1] = ACTIONS(1604),
    [aux_sym_signed_number_token1] = ACTIONS(1602),
    [anon_sym_SQUOTE] = ACTIONS(1604),
    [anon_sym_DQUOTE] = ACTIONS(1604),
    [sym_prefix_operator] = ACTIONS(1604),
    [aux_sym_sign_operator_token1] = ACTIONS(1602),
    [sym_hash_operator] = ACTIONS(1604),
    [anon_sym_lsl] = ACTIONS(1602),
    [anon_sym_lsr] = ACTIONS(1602),
    [anon_sym_asr] = ACTIONS(1602),
    [aux_sym_pow_operator_token1] = ACTIONS(1604),
    [anon_sym_mod] = ACTIONS(1602),
    [anon_sym_land] = ACTIONS(1602),
    [anon_sym_lor] = ACTIONS(1602),
    [anon_sym_lxor] = ACTIONS(1602),
    [aux_sym_mult_operator_token1] = ACTIONS(1602),
    [aux_sym_add_operator_token1] = ACTIONS(1602),
    [sym_concat_operator] = ACTIONS(1604),
    [sym_rel_operator] = ACTIONS(1602),
    [anon_sym_AMP_AMP] = ACTIONS(1602),
    [anon_sym_or] = ACTIONS(1602),
    [anon_sym_PIPE_PIPE] = ACTIONS(1602),
    [sym_assign_operator] = ACTIONS(1604),
    [sym_let_operator] = ACTIONS(1604),
    [sym__uppercase_identifier] = ACTIONS(1604),
    [aux_sym_directive_token1] = ACTIONS(1602),
    [aux_sym_tag_token1] = ACTIONS(1604),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1099)] = {
    [sym_attribute] = STATE(1099),
    [sym__lowercase_identifier] = ACTIONS(2021),
    [anon_sym_SEMI_SEMI] = ACTIONS(2023),
    [anon_sym_let] = ACTIONS(2021),
    [anon_sym_and] = ACTIONS(2021),
    [anon_sym_TILDE] = ACTIONS(2021),
    [anon_sym_QMARK] = ACTIONS(2021),
    [anon_sym_LPAREN] = ACTIONS(2023),
    [anon_sym_external] = ACTIONS(2021),
    [anon_sym_type] = ACTIONS(2021),
    [anon_sym_COMMA] = ACTIONS(2023),
    [anon_sym_PIPE] = ACTIONS(2021),
    [anon_sym_DOT] = ACTIONS(2023),
    [anon_sym_LBRACK] = ACTIONS(2021),
    [anon_sym_RBRACK] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(2021),
    [anon_sym_false] = ACTIONS(2021),
    [anon_sym_LBRACE] = ACTIONS(2021),
    [anon_sym_SEMI] = ACTIONS(2021),
    [anon_sym_exception] = ACTIONS(2021),
    [anon_sym_module] = ACTIONS(2021),
    [anon_sym_open] = ACTIONS(2021),
    [anon_sym_include] = ACTIONS(2021),
    [anon_sym_class] = ACTIONS(2021),
    [anon_sym_val] = ACTIONS(2021),
    [anon_sym_end] = ACTIONS(2021),
    [anon_sym_object] = ACTIONS(2021),
    [anon_sym_AMP] = ACTIONS(2021),
    [anon_sym_POUND] = ACTIONS(2021),
    [anon_sym_COLON_COLON] = ACTIONS(2023),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2023),
    [anon_sym_else] = ACTIONS(2021),
    [anon_sym_new] = ACTIONS(2021),
    [anon_sym_LBRACE_LT] = ACTIONS(2023),
    [anon_sym_begin] = ACTIONS(2021),
    [sym_ocamlyacc_value] = ACTIONS(2023),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2023),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2023),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2023),
    [aux_sym_number_token1] = ACTIONS(2023),
    [aux_sym_signed_number_token1] = ACTIONS(2021),
    [anon_sym_SQUOTE] = ACTIONS(2023),
    [anon_sym_DQUOTE] = ACTIONS(2023),
    [sym_prefix_operator] = ACTIONS(2023),
    [aux_sym_sign_operator_token1] = ACTIONS(2021),
    [sym_hash_operator] = ACTIONS(2023),
    [anon_sym_lsl] = ACTIONS(2021),
    [anon_sym_lsr] = ACTIONS(2021),
    [anon_sym_asr] = ACTIONS(2021),
    [aux_sym_pow_operator_token1] = ACTIONS(2023),
    [anon_sym_mod] = ACTIONS(2021),
    [anon_sym_land] = ACTIONS(2021),
    [anon_sym_lor] = ACTIONS(2021),
    [anon_sym_lxor] = ACTIONS(2021),
    [aux_sym_mult_operator_token1] = ACTIONS(2021),
    [aux_sym_add_operator_token1] = ACTIONS(2021),
    [sym_concat_operator] = ACTIONS(2023),
    [sym_rel_operator] = ACTIONS(2021),
    [anon_sym_AMP_AMP] = ACTIONS(2021),
    [anon_sym_or] = ACTIONS(2021),
    [anon_sym_PIPE_PIPE] = ACTIONS(2021),
    [sym_assign_operator] = ACTIONS(2023),
    [sym_let_operator] = ACTIONS(2023),
    [sym__uppercase_identifier] = ACTIONS(2023),
    [aux_sym_directive_token1] = ACTIONS(2021),
    [aux_sym_tag_token1] = ACTIONS(2023),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1100)] = {
    [sym_attribute] = STATE(1100),
    [sym__lowercase_identifier] = ACTIONS(1854),
    [anon_sym_SEMI_SEMI] = ACTIONS(1856),
    [anon_sym_let] = ACTIONS(1854),
    [anon_sym_and] = ACTIONS(1854),
    [anon_sym_TILDE] = ACTIONS(1854),
    [anon_sym_QMARK] = ACTIONS(1854),
    [anon_sym_LPAREN] = ACTIONS(1856),
    [anon_sym_external] = ACTIONS(1854),
    [anon_sym_type] = ACTIONS(1854),
    [anon_sym_COMMA] = ACTIONS(1856),
    [anon_sym_PIPE] = ACTIONS(1854),
    [anon_sym_DOT] = ACTIONS(1856),
    [anon_sym_LBRACK] = ACTIONS(1854),
    [anon_sym_RBRACK] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(1854),
    [anon_sym_false] = ACTIONS(1854),
    [anon_sym_LBRACE] = ACTIONS(1854),
    [anon_sym_SEMI] = ACTIONS(1854),
    [anon_sym_exception] = ACTIONS(1854),
    [anon_sym_module] = ACTIONS(1854),
    [anon_sym_open] = ACTIONS(1854),
    [anon_sym_include] = ACTIONS(1854),
    [anon_sym_class] = ACTIONS(1854),
    [anon_sym_val] = ACTIONS(1854),
    [anon_sym_end] = ACTIONS(1854),
    [anon_sym_object] = ACTIONS(1854),
    [anon_sym_AMP] = ACTIONS(1854),
    [anon_sym_POUND] = ACTIONS(1854),
    [anon_sym_COLON_COLON] = ACTIONS(1856),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1856),
    [anon_sym_else] = ACTIONS(1854),
    [anon_sym_new] = ACTIONS(1854),
    [anon_sym_LBRACE_LT] = ACTIONS(1856),
    [anon_sym_begin] = ACTIONS(1854),
    [sym_ocamlyacc_value] = ACTIONS(1856),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1854),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1856),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1854),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1854),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1856),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1856),
    [aux_sym_number_token1] = ACTIONS(1856),
    [aux_sym_signed_number_token1] = ACTIONS(1854),
    [anon_sym_SQUOTE] = ACTIONS(1856),
    [anon_sym_DQUOTE] = ACTIONS(1856),
    [sym_prefix_operator] = ACTIONS(1856),
    [aux_sym_sign_operator_token1] = ACTIONS(1854),
    [sym_hash_operator] = ACTIONS(1856),
    [anon_sym_lsl] = ACTIONS(1854),
    [anon_sym_lsr] = ACTIONS(1854),
    [anon_sym_asr] = ACTIONS(1854),
    [aux_sym_pow_operator_token1] = ACTIONS(1856),
    [anon_sym_mod] = ACTIONS(1854),
    [anon_sym_land] = ACTIONS(1854),
    [anon_sym_lor] = ACTIONS(1854),
    [anon_sym_lxor] = ACTIONS(1854),
    [aux_sym_mult_operator_token1] = ACTIONS(1854),
    [aux_sym_add_operator_token1] = ACTIONS(1854),
    [sym_concat_operator] = ACTIONS(1856),
    [sym_rel_operator] = ACTIONS(1854),
    [anon_sym_AMP_AMP] = ACTIONS(1854),
    [anon_sym_or] = ACTIONS(1854),
    [anon_sym_PIPE_PIPE] = ACTIONS(1854),
    [sym_assign_operator] = ACTIONS(1856),
    [sym_let_operator] = ACTIONS(1856),
    [sym__uppercase_identifier] = ACTIONS(1856),
    [aux_sym_directive_token1] = ACTIONS(1854),
    [aux_sym_tag_token1] = ACTIONS(1856),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1101)] = {
    [sym_attribute] = STATE(1101),
    [sym__lowercase_identifier] = ACTIONS(1728),
    [anon_sym_SEMI_SEMI] = ACTIONS(1730),
    [anon_sym_let] = ACTIONS(1728),
    [anon_sym_and] = ACTIONS(1728),
    [anon_sym_TILDE] = ACTIONS(1728),
    [anon_sym_QMARK] = ACTIONS(1728),
    [anon_sym_LPAREN] = ACTIONS(1730),
    [anon_sym_external] = ACTIONS(1728),
    [anon_sym_type] = ACTIONS(1728),
    [anon_sym_COMMA] = ACTIONS(1730),
    [anon_sym_PIPE] = ACTIONS(1728),
    [anon_sym_DOT] = ACTIONS(1730),
    [anon_sym_LBRACK] = ACTIONS(1728),
    [anon_sym_RBRACK] = ACTIONS(1730),
    [anon_sym_true] = ACTIONS(1728),
    [anon_sym_false] = ACTIONS(1728),
    [anon_sym_LBRACE] = ACTIONS(1728),
    [anon_sym_SEMI] = ACTIONS(1728),
    [anon_sym_exception] = ACTIONS(1728),
    [anon_sym_module] = ACTIONS(1728),
    [anon_sym_open] = ACTIONS(1728),
    [anon_sym_include] = ACTIONS(1728),
    [anon_sym_class] = ACTIONS(1728),
    [anon_sym_val] = ACTIONS(1728),
    [anon_sym_end] = ACTIONS(1728),
    [anon_sym_object] = ACTIONS(1728),
    [anon_sym_AMP] = ACTIONS(1728),
    [anon_sym_POUND] = ACTIONS(1728),
    [anon_sym_COLON_COLON] = ACTIONS(1730),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1730),
    [anon_sym_else] = ACTIONS(1728),
    [anon_sym_new] = ACTIONS(1728),
    [anon_sym_LBRACE_LT] = ACTIONS(1730),
    [anon_sym_begin] = ACTIONS(1728),
    [sym_ocamlyacc_value] = ACTIONS(1730),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1728),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1730),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1728),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1728),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1730),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1730),
    [aux_sym_number_token1] = ACTIONS(1730),
    [aux_sym_signed_number_token1] = ACTIONS(1728),
    [anon_sym_SQUOTE] = ACTIONS(1730),
    [anon_sym_DQUOTE] = ACTIONS(1730),
    [sym_prefix_operator] = ACTIONS(1730),
    [aux_sym_sign_operator_token1] = ACTIONS(1728),
    [sym_hash_operator] = ACTIONS(1730),
    [anon_sym_lsl] = ACTIONS(1728),
    [anon_sym_lsr] = ACTIONS(1728),
    [anon_sym_asr] = ACTIONS(1728),
    [aux_sym_pow_operator_token1] = ACTIONS(1730),
    [anon_sym_mod] = ACTIONS(1728),
    [anon_sym_land] = ACTIONS(1728),
    [anon_sym_lor] = ACTIONS(1728),
    [anon_sym_lxor] = ACTIONS(1728),
    [aux_sym_mult_operator_token1] = ACTIONS(1728),
    [aux_sym_add_operator_token1] = ACTIONS(1728),
    [sym_concat_operator] = ACTIONS(1730),
    [sym_rel_operator] = ACTIONS(1728),
    [anon_sym_AMP_AMP] = ACTIONS(1728),
    [anon_sym_or] = ACTIONS(1728),
    [anon_sym_PIPE_PIPE] = ACTIONS(1728),
    [sym_assign_operator] = ACTIONS(1730),
    [sym_let_operator] = ACTIONS(1730),
    [sym__uppercase_identifier] = ACTIONS(1730),
    [aux_sym_directive_token1] = ACTIONS(1728),
    [aux_sym_tag_token1] = ACTIONS(1730),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1102)] = {
    [sym_attribute] = STATE(1102),
    [sym__lowercase_identifier] = ACTIONS(1594),
    [anon_sym_SEMI_SEMI] = ACTIONS(1596),
    [anon_sym_let] = ACTIONS(1594),
    [anon_sym_and] = ACTIONS(1594),
    [anon_sym_TILDE] = ACTIONS(1594),
    [anon_sym_QMARK] = ACTIONS(1594),
    [anon_sym_LPAREN] = ACTIONS(1596),
    [anon_sym_external] = ACTIONS(1594),
    [anon_sym_type] = ACTIONS(1594),
    [anon_sym_COMMA] = ACTIONS(1596),
    [anon_sym_PIPE] = ACTIONS(1594),
    [anon_sym_DOT] = ACTIONS(1596),
    [anon_sym_LBRACK] = ACTIONS(1594),
    [anon_sym_RBRACK] = ACTIONS(1596),
    [anon_sym_true] = ACTIONS(1594),
    [anon_sym_false] = ACTIONS(1594),
    [anon_sym_LBRACE] = ACTIONS(1594),
    [anon_sym_SEMI] = ACTIONS(1594),
    [anon_sym_exception] = ACTIONS(1594),
    [anon_sym_module] = ACTIONS(1594),
    [anon_sym_open] = ACTIONS(1594),
    [anon_sym_include] = ACTIONS(1594),
    [anon_sym_class] = ACTIONS(1594),
    [anon_sym_val] = ACTIONS(1594),
    [anon_sym_end] = ACTIONS(1594),
    [anon_sym_object] = ACTIONS(1594),
    [anon_sym_AMP] = ACTIONS(1594),
    [anon_sym_POUND] = ACTIONS(1594),
    [anon_sym_COLON_COLON] = ACTIONS(1596),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1596),
    [anon_sym_else] = ACTIONS(1594),
    [anon_sym_new] = ACTIONS(1594),
    [anon_sym_LBRACE_LT] = ACTIONS(1596),
    [anon_sym_begin] = ACTIONS(1594),
    [sym_ocamlyacc_value] = ACTIONS(1596),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1594),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1596),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1594),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1594),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1596),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1596),
    [aux_sym_number_token1] = ACTIONS(1596),
    [aux_sym_signed_number_token1] = ACTIONS(1594),
    [anon_sym_SQUOTE] = ACTIONS(1596),
    [anon_sym_DQUOTE] = ACTIONS(1596),
    [sym_prefix_operator] = ACTIONS(1596),
    [aux_sym_sign_operator_token1] = ACTIONS(1594),
    [sym_hash_operator] = ACTIONS(1596),
    [anon_sym_lsl] = ACTIONS(1594),
    [anon_sym_lsr] = ACTIONS(1594),
    [anon_sym_asr] = ACTIONS(1594),
    [aux_sym_pow_operator_token1] = ACTIONS(1596),
    [anon_sym_mod] = ACTIONS(1594),
    [anon_sym_land] = ACTIONS(1594),
    [anon_sym_lor] = ACTIONS(1594),
    [anon_sym_lxor] = ACTIONS(1594),
    [aux_sym_mult_operator_token1] = ACTIONS(1594),
    [aux_sym_add_operator_token1] = ACTIONS(1594),
    [sym_concat_operator] = ACTIONS(1596),
    [sym_rel_operator] = ACTIONS(1594),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_or] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1594),
    [sym_assign_operator] = ACTIONS(1596),
    [sym_let_operator] = ACTIONS(1596),
    [sym__uppercase_identifier] = ACTIONS(1596),
    [aux_sym_directive_token1] = ACTIONS(1594),
    [aux_sym_tag_token1] = ACTIONS(1596),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1103)] = {
    [sym_attribute] = STATE(1103),
    [sym__lowercase_identifier] = ACTIONS(1890),
    [anon_sym_SEMI_SEMI] = ACTIONS(1892),
    [anon_sym_let] = ACTIONS(1890),
    [anon_sym_and] = ACTIONS(1890),
    [anon_sym_TILDE] = ACTIONS(1890),
    [anon_sym_QMARK] = ACTIONS(1890),
    [anon_sym_LPAREN] = ACTIONS(1892),
    [anon_sym_external] = ACTIONS(1890),
    [anon_sym_type] = ACTIONS(1890),
    [anon_sym_COMMA] = ACTIONS(1892),
    [anon_sym_PIPE] = ACTIONS(1890),
    [anon_sym_DOT] = ACTIONS(1892),
    [anon_sym_LBRACK] = ACTIONS(1890),
    [anon_sym_RBRACK] = ACTIONS(1892),
    [anon_sym_true] = ACTIONS(1890),
    [anon_sym_false] = ACTIONS(1890),
    [anon_sym_LBRACE] = ACTIONS(1890),
    [anon_sym_SEMI] = ACTIONS(1890),
    [anon_sym_exception] = ACTIONS(1890),
    [anon_sym_module] = ACTIONS(1890),
    [anon_sym_open] = ACTIONS(1890),
    [anon_sym_include] = ACTIONS(1890),
    [anon_sym_class] = ACTIONS(1890),
    [anon_sym_val] = ACTIONS(1890),
    [anon_sym_end] = ACTIONS(1890),
    [anon_sym_object] = ACTIONS(1890),
    [anon_sym_AMP] = ACTIONS(1890),
    [anon_sym_POUND] = ACTIONS(1890),
    [anon_sym_COLON_COLON] = ACTIONS(1892),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1892),
    [anon_sym_else] = ACTIONS(1890),
    [anon_sym_new] = ACTIONS(1890),
    [anon_sym_LBRACE_LT] = ACTIONS(1892),
    [anon_sym_begin] = ACTIONS(1890),
    [sym_ocamlyacc_value] = ACTIONS(1892),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1890),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1892),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1890),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1890),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1892),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1892),
    [aux_sym_number_token1] = ACTIONS(1892),
    [aux_sym_signed_number_token1] = ACTIONS(1890),
    [anon_sym_SQUOTE] = ACTIONS(1892),
    [anon_sym_DQUOTE] = ACTIONS(1892),
    [sym_prefix_operator] = ACTIONS(1892),
    [aux_sym_sign_operator_token1] = ACTIONS(1890),
    [sym_hash_operator] = ACTIONS(1892),
    [anon_sym_lsl] = ACTIONS(1890),
    [anon_sym_lsr] = ACTIONS(1890),
    [anon_sym_asr] = ACTIONS(1890),
    [aux_sym_pow_operator_token1] = ACTIONS(1892),
    [anon_sym_mod] = ACTIONS(1890),
    [anon_sym_land] = ACTIONS(1890),
    [anon_sym_lor] = ACTIONS(1890),
    [anon_sym_lxor] = ACTIONS(1890),
    [aux_sym_mult_operator_token1] = ACTIONS(1890),
    [aux_sym_add_operator_token1] = ACTIONS(1890),
    [sym_concat_operator] = ACTIONS(1892),
    [sym_rel_operator] = ACTIONS(1890),
    [anon_sym_AMP_AMP] = ACTIONS(1890),
    [anon_sym_or] = ACTIONS(1890),
    [anon_sym_PIPE_PIPE] = ACTIONS(1890),
    [sym_assign_operator] = ACTIONS(1892),
    [sym_let_operator] = ACTIONS(1892),
    [sym__uppercase_identifier] = ACTIONS(1892),
    [aux_sym_directive_token1] = ACTIONS(1890),
    [aux_sym_tag_token1] = ACTIONS(1892),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1104)] = {
    [sym_attribute] = STATE(1104),
    [sym__lowercase_identifier] = ACTIONS(1606),
    [anon_sym_SEMI_SEMI] = ACTIONS(1608),
    [anon_sym_let] = ACTIONS(1606),
    [anon_sym_and] = ACTIONS(1606),
    [anon_sym_TILDE] = ACTIONS(1606),
    [anon_sym_QMARK] = ACTIONS(1606),
    [anon_sym_LPAREN] = ACTIONS(1608),
    [anon_sym_external] = ACTIONS(1606),
    [anon_sym_type] = ACTIONS(1606),
    [anon_sym_COMMA] = ACTIONS(1608),
    [anon_sym_PIPE] = ACTIONS(1606),
    [anon_sym_DOT] = ACTIONS(1608),
    [anon_sym_LBRACK] = ACTIONS(1606),
    [anon_sym_RBRACK] = ACTIONS(1608),
    [anon_sym_true] = ACTIONS(1606),
    [anon_sym_false] = ACTIONS(1606),
    [anon_sym_LBRACE] = ACTIONS(1606),
    [anon_sym_SEMI] = ACTIONS(1606),
    [anon_sym_exception] = ACTIONS(1606),
    [anon_sym_module] = ACTIONS(1606),
    [anon_sym_open] = ACTIONS(1606),
    [anon_sym_include] = ACTIONS(1606),
    [anon_sym_class] = ACTIONS(1606),
    [anon_sym_val] = ACTIONS(1606),
    [anon_sym_end] = ACTIONS(1606),
    [anon_sym_object] = ACTIONS(1606),
    [anon_sym_AMP] = ACTIONS(1606),
    [anon_sym_POUND] = ACTIONS(1606),
    [anon_sym_COLON_COLON] = ACTIONS(1608),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1608),
    [anon_sym_else] = ACTIONS(1606),
    [anon_sym_new] = ACTIONS(1606),
    [anon_sym_LBRACE_LT] = ACTIONS(1608),
    [anon_sym_begin] = ACTIONS(1606),
    [sym_ocamlyacc_value] = ACTIONS(1608),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1606),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1608),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1606),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1606),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1608),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1608),
    [aux_sym_number_token1] = ACTIONS(1608),
    [aux_sym_signed_number_token1] = ACTIONS(1606),
    [anon_sym_SQUOTE] = ACTIONS(1608),
    [anon_sym_DQUOTE] = ACTIONS(1608),
    [sym_prefix_operator] = ACTIONS(1608),
    [aux_sym_sign_operator_token1] = ACTIONS(1606),
    [sym_hash_operator] = ACTIONS(1608),
    [anon_sym_lsl] = ACTIONS(1606),
    [anon_sym_lsr] = ACTIONS(1606),
    [anon_sym_asr] = ACTIONS(1606),
    [aux_sym_pow_operator_token1] = ACTIONS(1608),
    [anon_sym_mod] = ACTIONS(1606),
    [anon_sym_land] = ACTIONS(1606),
    [anon_sym_lor] = ACTIONS(1606),
    [anon_sym_lxor] = ACTIONS(1606),
    [aux_sym_mult_operator_token1] = ACTIONS(1606),
    [aux_sym_add_operator_token1] = ACTIONS(1606),
    [sym_concat_operator] = ACTIONS(1608),
    [sym_rel_operator] = ACTIONS(1606),
    [anon_sym_AMP_AMP] = ACTIONS(1606),
    [anon_sym_or] = ACTIONS(1606),
    [anon_sym_PIPE_PIPE] = ACTIONS(1606),
    [sym_assign_operator] = ACTIONS(1608),
    [sym_let_operator] = ACTIONS(1608),
    [sym__uppercase_identifier] = ACTIONS(1608),
    [aux_sym_directive_token1] = ACTIONS(1606),
    [aux_sym_tag_token1] = ACTIONS(1608),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1105)] = {
    [sym_attribute] = STATE(1105),
    [sym__lowercase_identifier] = ACTIONS(1622),
    [anon_sym_SEMI_SEMI] = ACTIONS(1624),
    [anon_sym_let] = ACTIONS(1622),
    [anon_sym_and] = ACTIONS(1622),
    [anon_sym_TILDE] = ACTIONS(1622),
    [anon_sym_QMARK] = ACTIONS(1622),
    [anon_sym_LPAREN] = ACTIONS(1624),
    [anon_sym_external] = ACTIONS(1622),
    [anon_sym_type] = ACTIONS(1622),
    [anon_sym_COMMA] = ACTIONS(1624),
    [anon_sym_PIPE] = ACTIONS(1622),
    [anon_sym_DOT] = ACTIONS(1624),
    [anon_sym_LBRACK] = ACTIONS(1622),
    [anon_sym_RBRACK] = ACTIONS(1624),
    [anon_sym_true] = ACTIONS(1622),
    [anon_sym_false] = ACTIONS(1622),
    [anon_sym_LBRACE] = ACTIONS(1622),
    [anon_sym_SEMI] = ACTIONS(1622),
    [anon_sym_exception] = ACTIONS(1622),
    [anon_sym_module] = ACTIONS(1622),
    [anon_sym_open] = ACTIONS(1622),
    [anon_sym_include] = ACTIONS(1622),
    [anon_sym_class] = ACTIONS(1622),
    [anon_sym_val] = ACTIONS(1622),
    [anon_sym_end] = ACTIONS(1622),
    [anon_sym_object] = ACTIONS(1622),
    [anon_sym_AMP] = ACTIONS(1622),
    [anon_sym_POUND] = ACTIONS(1622),
    [anon_sym_COLON_COLON] = ACTIONS(1624),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1624),
    [anon_sym_else] = ACTIONS(1622),
    [anon_sym_new] = ACTIONS(1622),
    [anon_sym_LBRACE_LT] = ACTIONS(1624),
    [anon_sym_begin] = ACTIONS(1622),
    [sym_ocamlyacc_value] = ACTIONS(1624),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1622),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1624),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1622),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1622),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1624),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1624),
    [aux_sym_number_token1] = ACTIONS(1624),
    [aux_sym_signed_number_token1] = ACTIONS(1622),
    [anon_sym_SQUOTE] = ACTIONS(1624),
    [anon_sym_DQUOTE] = ACTIONS(1624),
    [sym_prefix_operator] = ACTIONS(1624),
    [aux_sym_sign_operator_token1] = ACTIONS(1622),
    [sym_hash_operator] = ACTIONS(1624),
    [anon_sym_lsl] = ACTIONS(1622),
    [anon_sym_lsr] = ACTIONS(1622),
    [anon_sym_asr] = ACTIONS(1622),
    [aux_sym_pow_operator_token1] = ACTIONS(1624),
    [anon_sym_mod] = ACTIONS(1622),
    [anon_sym_land] = ACTIONS(1622),
    [anon_sym_lor] = ACTIONS(1622),
    [anon_sym_lxor] = ACTIONS(1622),
    [aux_sym_mult_operator_token1] = ACTIONS(1622),
    [aux_sym_add_operator_token1] = ACTIONS(1622),
    [sym_concat_operator] = ACTIONS(1624),
    [sym_rel_operator] = ACTIONS(1622),
    [anon_sym_AMP_AMP] = ACTIONS(1622),
    [anon_sym_or] = ACTIONS(1622),
    [anon_sym_PIPE_PIPE] = ACTIONS(1622),
    [sym_assign_operator] = ACTIONS(1624),
    [sym_let_operator] = ACTIONS(1624),
    [sym__uppercase_identifier] = ACTIONS(1624),
    [aux_sym_directive_token1] = ACTIONS(1622),
    [aux_sym_tag_token1] = ACTIONS(1624),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1106)] = {
    [sym_attribute] = STATE(1106),
    [sym__lowercase_identifier] = ACTIONS(1586),
    [anon_sym_SEMI_SEMI] = ACTIONS(1588),
    [anon_sym_let] = ACTIONS(1586),
    [anon_sym_and] = ACTIONS(1586),
    [anon_sym_TILDE] = ACTIONS(1586),
    [anon_sym_QMARK] = ACTIONS(1586),
    [anon_sym_LPAREN] = ACTIONS(1588),
    [anon_sym_external] = ACTIONS(1586),
    [anon_sym_type] = ACTIONS(1586),
    [anon_sym_COMMA] = ACTIONS(1588),
    [anon_sym_PIPE] = ACTIONS(1586),
    [anon_sym_DOT] = ACTIONS(1588),
    [anon_sym_LBRACK] = ACTIONS(1586),
    [anon_sym_RBRACK] = ACTIONS(1588),
    [anon_sym_true] = ACTIONS(1586),
    [anon_sym_false] = ACTIONS(1586),
    [anon_sym_LBRACE] = ACTIONS(1586),
    [anon_sym_SEMI] = ACTIONS(1586),
    [anon_sym_exception] = ACTIONS(1586),
    [anon_sym_module] = ACTIONS(1586),
    [anon_sym_open] = ACTIONS(1586),
    [anon_sym_include] = ACTIONS(1586),
    [anon_sym_class] = ACTIONS(1586),
    [anon_sym_val] = ACTIONS(1586),
    [anon_sym_end] = ACTIONS(1586),
    [anon_sym_object] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1586),
    [anon_sym_POUND] = ACTIONS(1586),
    [anon_sym_COLON_COLON] = ACTIONS(1588),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1588),
    [anon_sym_else] = ACTIONS(1586),
    [anon_sym_new] = ACTIONS(1586),
    [anon_sym_LBRACE_LT] = ACTIONS(1588),
    [anon_sym_begin] = ACTIONS(1586),
    [sym_ocamlyacc_value] = ACTIONS(1588),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1586),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1588),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1586),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1586),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1588),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1588),
    [aux_sym_number_token1] = ACTIONS(1588),
    [aux_sym_signed_number_token1] = ACTIONS(1586),
    [anon_sym_SQUOTE] = ACTIONS(1588),
    [anon_sym_DQUOTE] = ACTIONS(1588),
    [sym_prefix_operator] = ACTIONS(1588),
    [aux_sym_sign_operator_token1] = ACTIONS(1586),
    [sym_hash_operator] = ACTIONS(1588),
    [anon_sym_lsl] = ACTIONS(1586),
    [anon_sym_lsr] = ACTIONS(1586),
    [anon_sym_asr] = ACTIONS(1586),
    [aux_sym_pow_operator_token1] = ACTIONS(1588),
    [anon_sym_mod] = ACTIONS(1586),
    [anon_sym_land] = ACTIONS(1586),
    [anon_sym_lor] = ACTIONS(1586),
    [anon_sym_lxor] = ACTIONS(1586),
    [aux_sym_mult_operator_token1] = ACTIONS(1586),
    [aux_sym_add_operator_token1] = ACTIONS(1586),
    [sym_concat_operator] = ACTIONS(1588),
    [sym_rel_operator] = ACTIONS(1586),
    [anon_sym_AMP_AMP] = ACTIONS(1586),
    [anon_sym_or] = ACTIONS(1586),
    [anon_sym_PIPE_PIPE] = ACTIONS(1586),
    [sym_assign_operator] = ACTIONS(1588),
    [sym_let_operator] = ACTIONS(1588),
    [sym__uppercase_identifier] = ACTIONS(1588),
    [aux_sym_directive_token1] = ACTIONS(1586),
    [aux_sym_tag_token1] = ACTIONS(1588),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1107)] = {
    [sym_attribute] = STATE(1107),
    [sym__lowercase_identifier] = ACTIONS(1866),
    [anon_sym_SEMI_SEMI] = ACTIONS(1868),
    [anon_sym_let] = ACTIONS(1866),
    [anon_sym_and] = ACTIONS(1866),
    [anon_sym_TILDE] = ACTIONS(1866),
    [anon_sym_QMARK] = ACTIONS(1866),
    [anon_sym_LPAREN] = ACTIONS(1868),
    [anon_sym_external] = ACTIONS(1866),
    [anon_sym_type] = ACTIONS(1866),
    [anon_sym_COMMA] = ACTIONS(1868),
    [anon_sym_PIPE] = ACTIONS(1866),
    [anon_sym_DOT] = ACTIONS(1868),
    [anon_sym_LBRACK] = ACTIONS(1866),
    [anon_sym_RBRACK] = ACTIONS(1868),
    [anon_sym_true] = ACTIONS(1866),
    [anon_sym_false] = ACTIONS(1866),
    [anon_sym_LBRACE] = ACTIONS(1866),
    [anon_sym_SEMI] = ACTIONS(1866),
    [anon_sym_exception] = ACTIONS(1866),
    [anon_sym_module] = ACTIONS(1866),
    [anon_sym_open] = ACTIONS(1866),
    [anon_sym_include] = ACTIONS(1866),
    [anon_sym_class] = ACTIONS(1866),
    [anon_sym_val] = ACTIONS(1866),
    [anon_sym_end] = ACTIONS(1866),
    [anon_sym_object] = ACTIONS(1866),
    [anon_sym_AMP] = ACTIONS(1866),
    [anon_sym_POUND] = ACTIONS(1866),
    [anon_sym_COLON_COLON] = ACTIONS(1868),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1868),
    [anon_sym_else] = ACTIONS(1866),
    [anon_sym_new] = ACTIONS(1866),
    [anon_sym_LBRACE_LT] = ACTIONS(1868),
    [anon_sym_begin] = ACTIONS(1866),
    [sym_ocamlyacc_value] = ACTIONS(1868),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1866),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1868),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1866),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1866),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1868),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1868),
    [aux_sym_number_token1] = ACTIONS(1868),
    [aux_sym_signed_number_token1] = ACTIONS(1866),
    [anon_sym_SQUOTE] = ACTIONS(1868),
    [anon_sym_DQUOTE] = ACTIONS(1868),
    [sym_prefix_operator] = ACTIONS(1868),
    [aux_sym_sign_operator_token1] = ACTIONS(1866),
    [sym_hash_operator] = ACTIONS(1868),
    [anon_sym_lsl] = ACTIONS(1866),
    [anon_sym_lsr] = ACTIONS(1866),
    [anon_sym_asr] = ACTIONS(1866),
    [aux_sym_pow_operator_token1] = ACTIONS(1868),
    [anon_sym_mod] = ACTIONS(1866),
    [anon_sym_land] = ACTIONS(1866),
    [anon_sym_lor] = ACTIONS(1866),
    [anon_sym_lxor] = ACTIONS(1866),
    [aux_sym_mult_operator_token1] = ACTIONS(1866),
    [aux_sym_add_operator_token1] = ACTIONS(1866),
    [sym_concat_operator] = ACTIONS(1868),
    [sym_rel_operator] = ACTIONS(1866),
    [anon_sym_AMP_AMP] = ACTIONS(1866),
    [anon_sym_or] = ACTIONS(1866),
    [anon_sym_PIPE_PIPE] = ACTIONS(1866),
    [sym_assign_operator] = ACTIONS(1868),
    [sym_let_operator] = ACTIONS(1868),
    [sym__uppercase_identifier] = ACTIONS(1868),
    [aux_sym_directive_token1] = ACTIONS(1866),
    [aux_sym_tag_token1] = ACTIONS(1868),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1108)] = {
    [sym_attribute] = STATE(1108),
    [sym__lowercase_identifier] = ACTIONS(1850),
    [anon_sym_SEMI_SEMI] = ACTIONS(1852),
    [anon_sym_let] = ACTIONS(1850),
    [anon_sym_and] = ACTIONS(1850),
    [anon_sym_TILDE] = ACTIONS(1850),
    [anon_sym_QMARK] = ACTIONS(1850),
    [anon_sym_LPAREN] = ACTIONS(1852),
    [anon_sym_external] = ACTIONS(1850),
    [anon_sym_type] = ACTIONS(1850),
    [anon_sym_COMMA] = ACTIONS(1852),
    [anon_sym_PIPE] = ACTIONS(1850),
    [anon_sym_DOT] = ACTIONS(1852),
    [anon_sym_LBRACK] = ACTIONS(1850),
    [anon_sym_RBRACK] = ACTIONS(1852),
    [anon_sym_true] = ACTIONS(1850),
    [anon_sym_false] = ACTIONS(1850),
    [anon_sym_LBRACE] = ACTIONS(1850),
    [anon_sym_SEMI] = ACTIONS(1850),
    [anon_sym_exception] = ACTIONS(1850),
    [anon_sym_module] = ACTIONS(1850),
    [anon_sym_open] = ACTIONS(1850),
    [anon_sym_include] = ACTIONS(1850),
    [anon_sym_class] = ACTIONS(1850),
    [anon_sym_val] = ACTIONS(1850),
    [anon_sym_end] = ACTIONS(1850),
    [anon_sym_object] = ACTIONS(1850),
    [anon_sym_AMP] = ACTIONS(1850),
    [anon_sym_POUND] = ACTIONS(1850),
    [anon_sym_COLON_COLON] = ACTIONS(1852),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1852),
    [anon_sym_else] = ACTIONS(1850),
    [anon_sym_new] = ACTIONS(1850),
    [anon_sym_LBRACE_LT] = ACTIONS(1852),
    [anon_sym_begin] = ACTIONS(1850),
    [sym_ocamlyacc_value] = ACTIONS(1852),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1850),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1852),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1850),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1850),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1852),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1852),
    [aux_sym_number_token1] = ACTIONS(1852),
    [aux_sym_signed_number_token1] = ACTIONS(1850),
    [anon_sym_SQUOTE] = ACTIONS(1852),
    [anon_sym_DQUOTE] = ACTIONS(1852),
    [sym_prefix_operator] = ACTIONS(1852),
    [aux_sym_sign_operator_token1] = ACTIONS(1850),
    [sym_hash_operator] = ACTIONS(1852),
    [anon_sym_lsl] = ACTIONS(1850),
    [anon_sym_lsr] = ACTIONS(1850),
    [anon_sym_asr] = ACTIONS(1850),
    [aux_sym_pow_operator_token1] = ACTIONS(1852),
    [anon_sym_mod] = ACTIONS(1850),
    [anon_sym_land] = ACTIONS(1850),
    [anon_sym_lor] = ACTIONS(1850),
    [anon_sym_lxor] = ACTIONS(1850),
    [aux_sym_mult_operator_token1] = ACTIONS(1850),
    [aux_sym_add_operator_token1] = ACTIONS(1850),
    [sym_concat_operator] = ACTIONS(1852),
    [sym_rel_operator] = ACTIONS(1850),
    [anon_sym_AMP_AMP] = ACTIONS(1850),
    [anon_sym_or] = ACTIONS(1850),
    [anon_sym_PIPE_PIPE] = ACTIONS(1850),
    [sym_assign_operator] = ACTIONS(1852),
    [sym_let_operator] = ACTIONS(1852),
    [sym__uppercase_identifier] = ACTIONS(1852),
    [aux_sym_directive_token1] = ACTIONS(1850),
    [aux_sym_tag_token1] = ACTIONS(1852),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1109)] = {
    [sym_attribute] = STATE(1109),
    [sym__lowercase_identifier] = ACTIONS(1736),
    [anon_sym_SEMI_SEMI] = ACTIONS(1738),
    [anon_sym_let] = ACTIONS(1736),
    [anon_sym_and] = ACTIONS(1736),
    [anon_sym_TILDE] = ACTIONS(1736),
    [anon_sym_QMARK] = ACTIONS(1736),
    [anon_sym_LPAREN] = ACTIONS(1738),
    [anon_sym_external] = ACTIONS(1736),
    [anon_sym_type] = ACTIONS(1736),
    [anon_sym_COMMA] = ACTIONS(1738),
    [anon_sym_PIPE] = ACTIONS(1736),
    [anon_sym_DOT] = ACTIONS(1738),
    [anon_sym_LBRACK] = ACTIONS(1736),
    [anon_sym_RBRACK] = ACTIONS(1738),
    [anon_sym_true] = ACTIONS(1736),
    [anon_sym_false] = ACTIONS(1736),
    [anon_sym_LBRACE] = ACTIONS(1736),
    [anon_sym_SEMI] = ACTIONS(1736),
    [anon_sym_exception] = ACTIONS(1736),
    [anon_sym_module] = ACTIONS(1736),
    [anon_sym_open] = ACTIONS(1736),
    [anon_sym_include] = ACTIONS(1736),
    [anon_sym_class] = ACTIONS(1736),
    [anon_sym_val] = ACTIONS(1736),
    [anon_sym_end] = ACTIONS(1736),
    [anon_sym_object] = ACTIONS(1736),
    [anon_sym_AMP] = ACTIONS(1736),
    [anon_sym_POUND] = ACTIONS(1736),
    [anon_sym_COLON_COLON] = ACTIONS(1738),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1738),
    [anon_sym_else] = ACTIONS(1736),
    [anon_sym_new] = ACTIONS(1736),
    [anon_sym_LBRACE_LT] = ACTIONS(1738),
    [anon_sym_begin] = ACTIONS(1736),
    [sym_ocamlyacc_value] = ACTIONS(1738),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1736),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1738),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1736),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1736),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1738),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1738),
    [aux_sym_number_token1] = ACTIONS(1738),
    [aux_sym_signed_number_token1] = ACTIONS(1736),
    [anon_sym_SQUOTE] = ACTIONS(1738),
    [anon_sym_DQUOTE] = ACTIONS(1738),
    [sym_prefix_operator] = ACTIONS(1738),
    [aux_sym_sign_operator_token1] = ACTIONS(1736),
    [sym_hash_operator] = ACTIONS(1738),
    [anon_sym_lsl] = ACTIONS(1736),
    [anon_sym_lsr] = ACTIONS(1736),
    [anon_sym_asr] = ACTIONS(1736),
    [aux_sym_pow_operator_token1] = ACTIONS(1738),
    [anon_sym_mod] = ACTIONS(1736),
    [anon_sym_land] = ACTIONS(1736),
    [anon_sym_lor] = ACTIONS(1736),
    [anon_sym_lxor] = ACTIONS(1736),
    [aux_sym_mult_operator_token1] = ACTIONS(1736),
    [aux_sym_add_operator_token1] = ACTIONS(1736),
    [sym_concat_operator] = ACTIONS(1738),
    [sym_rel_operator] = ACTIONS(1736),
    [anon_sym_AMP_AMP] = ACTIONS(1736),
    [anon_sym_or] = ACTIONS(1736),
    [anon_sym_PIPE_PIPE] = ACTIONS(1736),
    [sym_assign_operator] = ACTIONS(1738),
    [sym_let_operator] = ACTIONS(1738),
    [sym__uppercase_identifier] = ACTIONS(1738),
    [aux_sym_directive_token1] = ACTIONS(1736),
    [aux_sym_tag_token1] = ACTIONS(1738),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1110)] = {
    [sym_attribute] = STATE(1110),
    [sym__lowercase_identifier] = ACTIONS(2161),
    [anon_sym_SEMI_SEMI] = ACTIONS(2165),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_and] = ACTIONS(2161),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_QMARK] = ACTIONS(2161),
    [anon_sym_COLON] = ACTIONS(2326),
    [anon_sym_LPAREN] = ACTIONS(2165),
    [anon_sym_external] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_COMMA] = ACTIONS(2165),
    [anon_sym_PIPE] = ACTIONS(2161),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [anon_sym_RBRACK] = ACTIONS(2165),
    [anon_sym_true] = ACTIONS(2161),
    [anon_sym_false] = ACTIONS(2161),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [anon_sym_SEMI] = ACTIONS(2161),
    [anon_sym_exception] = ACTIONS(2161),
    [anon_sym_module] = ACTIONS(2161),
    [anon_sym_open] = ACTIONS(2161),
    [anon_sym_include] = ACTIONS(2161),
    [anon_sym_class] = ACTIONS(2161),
    [anon_sym_val] = ACTIONS(2161),
    [anon_sym_end] = ACTIONS(2161),
    [anon_sym_object] = ACTIONS(2161),
    [anon_sym_in] = ACTIONS(2161),
    [anon_sym_AMP] = ACTIONS(2161),
    [anon_sym_COLON_COLON] = ACTIONS(2165),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2165),
    [anon_sym_else] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2161),
    [anon_sym_LBRACE_LT] = ACTIONS(2165),
    [anon_sym_begin] = ACTIONS(2161),
    [sym_ocamlyacc_value] = ACTIONS(2165),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2161),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2165),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2165),
    [aux_sym_number_token1] = ACTIONS(2165),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2165),
    [anon_sym_DQUOTE] = ACTIONS(2165),
    [sym_prefix_operator] = ACTIONS(2165),
    [aux_sym_sign_operator_token1] = ACTIONS(2161),
    [anon_sym_lsl] = ACTIONS(2161),
    [anon_sym_lsr] = ACTIONS(2161),
    [anon_sym_asr] = ACTIONS(2161),
    [aux_sym_pow_operator_token1] = ACTIONS(2165),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_land] = ACTIONS(2161),
    [anon_sym_lor] = ACTIONS(2161),
    [anon_sym_lxor] = ACTIONS(2161),
    [aux_sym_mult_operator_token1] = ACTIONS(2161),
    [aux_sym_add_operator_token1] = ACTIONS(2161),
    [sym_concat_operator] = ACTIONS(2165),
    [sym_rel_operator] = ACTIONS(2161),
    [anon_sym_AMP_AMP] = ACTIONS(2161),
    [anon_sym_or] = ACTIONS(2161),
    [anon_sym_PIPE_PIPE] = ACTIONS(2161),
    [sym_assign_operator] = ACTIONS(2165),
    [sym_let_operator] = ACTIONS(2165),
    [sym_let_and_operator] = ACTIONS(2165),
    [sym__uppercase_identifier] = ACTIONS(2165),
    [aux_sym_directive_token1] = ACTIONS(2165),
    [aux_sym_tag_token1] = ACTIONS(2165),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1111)] = {
    [sym_attribute] = STATE(1111),
    [sym__lowercase_identifier] = ACTIONS(1590),
    [anon_sym_SEMI_SEMI] = ACTIONS(1592),
    [anon_sym_let] = ACTIONS(1590),
    [anon_sym_and] = ACTIONS(1590),
    [anon_sym_TILDE] = ACTIONS(1590),
    [anon_sym_QMARK] = ACTIONS(1590),
    [anon_sym_LPAREN] = ACTIONS(1592),
    [anon_sym_external] = ACTIONS(1590),
    [anon_sym_type] = ACTIONS(1590),
    [anon_sym_COMMA] = ACTIONS(1592),
    [anon_sym_PIPE] = ACTIONS(1590),
    [anon_sym_DOT] = ACTIONS(1592),
    [anon_sym_LBRACK] = ACTIONS(1590),
    [anon_sym_RBRACK] = ACTIONS(1592),
    [anon_sym_true] = ACTIONS(1590),
    [anon_sym_false] = ACTIONS(1590),
    [anon_sym_LBRACE] = ACTIONS(1590),
    [anon_sym_SEMI] = ACTIONS(1590),
    [anon_sym_exception] = ACTIONS(1590),
    [anon_sym_module] = ACTIONS(1590),
    [anon_sym_open] = ACTIONS(1590),
    [anon_sym_include] = ACTIONS(1590),
    [anon_sym_class] = ACTIONS(1590),
    [anon_sym_val] = ACTIONS(1590),
    [anon_sym_end] = ACTIONS(1590),
    [anon_sym_object] = ACTIONS(1590),
    [anon_sym_AMP] = ACTIONS(1590),
    [anon_sym_POUND] = ACTIONS(1590),
    [anon_sym_COLON_COLON] = ACTIONS(1592),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1592),
    [anon_sym_else] = ACTIONS(1590),
    [anon_sym_new] = ACTIONS(1590),
    [anon_sym_LBRACE_LT] = ACTIONS(1592),
    [anon_sym_begin] = ACTIONS(1590),
    [sym_ocamlyacc_value] = ACTIONS(1592),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1590),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1592),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1590),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1590),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1592),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1592),
    [aux_sym_number_token1] = ACTIONS(1592),
    [aux_sym_signed_number_token1] = ACTIONS(1590),
    [anon_sym_SQUOTE] = ACTIONS(1592),
    [anon_sym_DQUOTE] = ACTIONS(1592),
    [sym_prefix_operator] = ACTIONS(1592),
    [aux_sym_sign_operator_token1] = ACTIONS(1590),
    [sym_hash_operator] = ACTIONS(1592),
    [anon_sym_lsl] = ACTIONS(1590),
    [anon_sym_lsr] = ACTIONS(1590),
    [anon_sym_asr] = ACTIONS(1590),
    [aux_sym_pow_operator_token1] = ACTIONS(1592),
    [anon_sym_mod] = ACTIONS(1590),
    [anon_sym_land] = ACTIONS(1590),
    [anon_sym_lor] = ACTIONS(1590),
    [anon_sym_lxor] = ACTIONS(1590),
    [aux_sym_mult_operator_token1] = ACTIONS(1590),
    [aux_sym_add_operator_token1] = ACTIONS(1590),
    [sym_concat_operator] = ACTIONS(1592),
    [sym_rel_operator] = ACTIONS(1590),
    [anon_sym_AMP_AMP] = ACTIONS(1590),
    [anon_sym_or] = ACTIONS(1590),
    [anon_sym_PIPE_PIPE] = ACTIONS(1590),
    [sym_assign_operator] = ACTIONS(1592),
    [sym_let_operator] = ACTIONS(1592),
    [sym__uppercase_identifier] = ACTIONS(1592),
    [aux_sym_directive_token1] = ACTIONS(1590),
    [aux_sym_tag_token1] = ACTIONS(1592),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1112)] = {
    [sym_attribute] = STATE(1112),
    [sym__lowercase_identifier] = ACTIONS(1830),
    [anon_sym_SEMI_SEMI] = ACTIONS(1832),
    [anon_sym_let] = ACTIONS(1830),
    [anon_sym_and] = ACTIONS(1830),
    [anon_sym_TILDE] = ACTIONS(1830),
    [anon_sym_QMARK] = ACTIONS(1830),
    [anon_sym_LPAREN] = ACTIONS(1832),
    [anon_sym_external] = ACTIONS(1830),
    [anon_sym_type] = ACTIONS(1830),
    [anon_sym_COMMA] = ACTIONS(1832),
    [anon_sym_PIPE] = ACTIONS(1830),
    [anon_sym_DOT] = ACTIONS(1832),
    [anon_sym_LBRACK] = ACTIONS(1830),
    [anon_sym_RBRACK] = ACTIONS(1832),
    [anon_sym_true] = ACTIONS(1830),
    [anon_sym_false] = ACTIONS(1830),
    [anon_sym_LBRACE] = ACTIONS(1830),
    [anon_sym_SEMI] = ACTIONS(1830),
    [anon_sym_exception] = ACTIONS(1830),
    [anon_sym_module] = ACTIONS(1830),
    [anon_sym_open] = ACTIONS(1830),
    [anon_sym_include] = ACTIONS(1830),
    [anon_sym_class] = ACTIONS(1830),
    [anon_sym_val] = ACTIONS(1830),
    [anon_sym_end] = ACTIONS(1830),
    [anon_sym_object] = ACTIONS(1830),
    [anon_sym_AMP] = ACTIONS(1830),
    [anon_sym_POUND] = ACTIONS(1830),
    [anon_sym_COLON_COLON] = ACTIONS(1832),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1832),
    [anon_sym_else] = ACTIONS(1830),
    [anon_sym_new] = ACTIONS(1830),
    [anon_sym_LBRACE_LT] = ACTIONS(1832),
    [anon_sym_begin] = ACTIONS(1830),
    [sym_ocamlyacc_value] = ACTIONS(1832),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1830),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1832),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1830),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1830),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1832),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1832),
    [aux_sym_number_token1] = ACTIONS(1832),
    [aux_sym_signed_number_token1] = ACTIONS(1830),
    [anon_sym_SQUOTE] = ACTIONS(1832),
    [anon_sym_DQUOTE] = ACTIONS(1832),
    [sym_prefix_operator] = ACTIONS(1832),
    [aux_sym_sign_operator_token1] = ACTIONS(1830),
    [sym_hash_operator] = ACTIONS(1832),
    [anon_sym_lsl] = ACTIONS(1830),
    [anon_sym_lsr] = ACTIONS(1830),
    [anon_sym_asr] = ACTIONS(1830),
    [aux_sym_pow_operator_token1] = ACTIONS(1832),
    [anon_sym_mod] = ACTIONS(1830),
    [anon_sym_land] = ACTIONS(1830),
    [anon_sym_lor] = ACTIONS(1830),
    [anon_sym_lxor] = ACTIONS(1830),
    [aux_sym_mult_operator_token1] = ACTIONS(1830),
    [aux_sym_add_operator_token1] = ACTIONS(1830),
    [sym_concat_operator] = ACTIONS(1832),
    [sym_rel_operator] = ACTIONS(1830),
    [anon_sym_AMP_AMP] = ACTIONS(1830),
    [anon_sym_or] = ACTIONS(1830),
    [anon_sym_PIPE_PIPE] = ACTIONS(1830),
    [sym_assign_operator] = ACTIONS(1832),
    [sym_let_operator] = ACTIONS(1832),
    [sym__uppercase_identifier] = ACTIONS(1832),
    [aux_sym_directive_token1] = ACTIONS(1830),
    [aux_sym_tag_token1] = ACTIONS(1832),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1113)] = {
    [sym_attribute] = STATE(1113),
    [sym__lowercase_identifier] = ACTIONS(1704),
    [anon_sym_SEMI_SEMI] = ACTIONS(1706),
    [anon_sym_let] = ACTIONS(1704),
    [anon_sym_and] = ACTIONS(1704),
    [anon_sym_TILDE] = ACTIONS(1704),
    [anon_sym_QMARK] = ACTIONS(1704),
    [anon_sym_LPAREN] = ACTIONS(1706),
    [anon_sym_external] = ACTIONS(1704),
    [anon_sym_type] = ACTIONS(1704),
    [anon_sym_COMMA] = ACTIONS(1706),
    [anon_sym_PIPE] = ACTIONS(1704),
    [anon_sym_DOT] = ACTIONS(1706),
    [anon_sym_LBRACK] = ACTIONS(1704),
    [anon_sym_RBRACK] = ACTIONS(1706),
    [anon_sym_true] = ACTIONS(1704),
    [anon_sym_false] = ACTIONS(1704),
    [anon_sym_LBRACE] = ACTIONS(1704),
    [anon_sym_SEMI] = ACTIONS(1704),
    [anon_sym_exception] = ACTIONS(1704),
    [anon_sym_module] = ACTIONS(1704),
    [anon_sym_open] = ACTIONS(1704),
    [anon_sym_include] = ACTIONS(1704),
    [anon_sym_class] = ACTIONS(1704),
    [anon_sym_val] = ACTIONS(1704),
    [anon_sym_end] = ACTIONS(1704),
    [anon_sym_object] = ACTIONS(1704),
    [anon_sym_AMP] = ACTIONS(1704),
    [anon_sym_POUND] = ACTIONS(1704),
    [anon_sym_COLON_COLON] = ACTIONS(1706),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1706),
    [anon_sym_else] = ACTIONS(1704),
    [anon_sym_new] = ACTIONS(1704),
    [anon_sym_LBRACE_LT] = ACTIONS(1706),
    [anon_sym_begin] = ACTIONS(1704),
    [sym_ocamlyacc_value] = ACTIONS(1706),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1704),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1706),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1704),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1704),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1706),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1706),
    [aux_sym_number_token1] = ACTIONS(1706),
    [aux_sym_signed_number_token1] = ACTIONS(1704),
    [anon_sym_SQUOTE] = ACTIONS(1706),
    [anon_sym_DQUOTE] = ACTIONS(1706),
    [sym_prefix_operator] = ACTIONS(1706),
    [aux_sym_sign_operator_token1] = ACTIONS(1704),
    [sym_hash_operator] = ACTIONS(1706),
    [anon_sym_lsl] = ACTIONS(1704),
    [anon_sym_lsr] = ACTIONS(1704),
    [anon_sym_asr] = ACTIONS(1704),
    [aux_sym_pow_operator_token1] = ACTIONS(1706),
    [anon_sym_mod] = ACTIONS(1704),
    [anon_sym_land] = ACTIONS(1704),
    [anon_sym_lor] = ACTIONS(1704),
    [anon_sym_lxor] = ACTIONS(1704),
    [aux_sym_mult_operator_token1] = ACTIONS(1704),
    [aux_sym_add_operator_token1] = ACTIONS(1704),
    [sym_concat_operator] = ACTIONS(1706),
    [sym_rel_operator] = ACTIONS(1704),
    [anon_sym_AMP_AMP] = ACTIONS(1704),
    [anon_sym_or] = ACTIONS(1704),
    [anon_sym_PIPE_PIPE] = ACTIONS(1704),
    [sym_assign_operator] = ACTIONS(1706),
    [sym_let_operator] = ACTIONS(1706),
    [sym__uppercase_identifier] = ACTIONS(1706),
    [aux_sym_directive_token1] = ACTIONS(1704),
    [aux_sym_tag_token1] = ACTIONS(1706),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1114)] = {
    [sym_attribute] = STATE(1114),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_SEMI_SEMI] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2079),
    [anon_sym_and] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_external] = ACTIONS(2079),
    [anon_sym_type] = ACTIONS(2079),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_DOT] = ACTIONS(963),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2079),
    [anon_sym_exception] = ACTIONS(2079),
    [anon_sym_module] = ACTIONS(2079),
    [anon_sym_open] = ACTIONS(2079),
    [anon_sym_include] = ACTIONS(2079),
    [anon_sym_class] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_POUND] = ACTIONS(965),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [sym_hash_operator] = ACTIONS(967),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym_let_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_directive_token1] = ACTIONS(2079),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1115)] = {
    [sym_attribute] = STATE(1115),
    [sym__lowercase_identifier] = ACTIONS(1646),
    [anon_sym_SEMI_SEMI] = ACTIONS(1648),
    [anon_sym_let] = ACTIONS(1646),
    [anon_sym_and] = ACTIONS(1646),
    [anon_sym_TILDE] = ACTIONS(1646),
    [anon_sym_QMARK] = ACTIONS(1646),
    [anon_sym_LPAREN] = ACTIONS(1648),
    [anon_sym_external] = ACTIONS(1646),
    [anon_sym_type] = ACTIONS(1646),
    [anon_sym_COMMA] = ACTIONS(1648),
    [anon_sym_PIPE] = ACTIONS(1646),
    [anon_sym_DOT] = ACTIONS(1648),
    [anon_sym_LBRACK] = ACTIONS(1646),
    [anon_sym_RBRACK] = ACTIONS(1648),
    [anon_sym_true] = ACTIONS(1646),
    [anon_sym_false] = ACTIONS(1646),
    [anon_sym_LBRACE] = ACTIONS(1646),
    [anon_sym_SEMI] = ACTIONS(1646),
    [anon_sym_exception] = ACTIONS(1646),
    [anon_sym_module] = ACTIONS(1646),
    [anon_sym_open] = ACTIONS(1646),
    [anon_sym_include] = ACTIONS(1646),
    [anon_sym_class] = ACTIONS(1646),
    [anon_sym_val] = ACTIONS(1646),
    [anon_sym_end] = ACTIONS(1646),
    [anon_sym_object] = ACTIONS(1646),
    [anon_sym_AMP] = ACTIONS(1646),
    [anon_sym_POUND] = ACTIONS(1646),
    [anon_sym_COLON_COLON] = ACTIONS(1648),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1648),
    [anon_sym_else] = ACTIONS(1646),
    [anon_sym_new] = ACTIONS(1646),
    [anon_sym_LBRACE_LT] = ACTIONS(1648),
    [anon_sym_begin] = ACTIONS(1646),
    [sym_ocamlyacc_value] = ACTIONS(1648),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1646),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1648),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1646),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1646),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1648),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1648),
    [aux_sym_number_token1] = ACTIONS(1648),
    [aux_sym_signed_number_token1] = ACTIONS(1646),
    [anon_sym_SQUOTE] = ACTIONS(1648),
    [anon_sym_DQUOTE] = ACTIONS(1648),
    [sym_prefix_operator] = ACTIONS(1648),
    [aux_sym_sign_operator_token1] = ACTIONS(1646),
    [sym_hash_operator] = ACTIONS(1648),
    [anon_sym_lsl] = ACTIONS(1646),
    [anon_sym_lsr] = ACTIONS(1646),
    [anon_sym_asr] = ACTIONS(1646),
    [aux_sym_pow_operator_token1] = ACTIONS(1648),
    [anon_sym_mod] = ACTIONS(1646),
    [anon_sym_land] = ACTIONS(1646),
    [anon_sym_lor] = ACTIONS(1646),
    [anon_sym_lxor] = ACTIONS(1646),
    [aux_sym_mult_operator_token1] = ACTIONS(1646),
    [aux_sym_add_operator_token1] = ACTIONS(1646),
    [sym_concat_operator] = ACTIONS(1648),
    [sym_rel_operator] = ACTIONS(1646),
    [anon_sym_AMP_AMP] = ACTIONS(1646),
    [anon_sym_or] = ACTIONS(1646),
    [anon_sym_PIPE_PIPE] = ACTIONS(1646),
    [sym_assign_operator] = ACTIONS(1648),
    [sym_let_operator] = ACTIONS(1648),
    [sym__uppercase_identifier] = ACTIONS(1648),
    [aux_sym_directive_token1] = ACTIONS(1646),
    [aux_sym_tag_token1] = ACTIONS(1648),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1116)] = {
    [sym_attribute] = STATE(1116),
    [sym__lowercase_identifier] = ACTIONS(1582),
    [anon_sym_SEMI_SEMI] = ACTIONS(1584),
    [anon_sym_let] = ACTIONS(1582),
    [anon_sym_and] = ACTIONS(1582),
    [anon_sym_TILDE] = ACTIONS(1582),
    [anon_sym_QMARK] = ACTIONS(1582),
    [anon_sym_LPAREN] = ACTIONS(1584),
    [anon_sym_external] = ACTIONS(1582),
    [anon_sym_type] = ACTIONS(1582),
    [anon_sym_COMMA] = ACTIONS(1584),
    [anon_sym_PIPE] = ACTIONS(1582),
    [anon_sym_DOT] = ACTIONS(1584),
    [anon_sym_LBRACK] = ACTIONS(1582),
    [anon_sym_RBRACK] = ACTIONS(1584),
    [anon_sym_true] = ACTIONS(1582),
    [anon_sym_false] = ACTIONS(1582),
    [anon_sym_LBRACE] = ACTIONS(1582),
    [anon_sym_SEMI] = ACTIONS(1582),
    [anon_sym_exception] = ACTIONS(1582),
    [anon_sym_module] = ACTIONS(1582),
    [anon_sym_open] = ACTIONS(1582),
    [anon_sym_include] = ACTIONS(1582),
    [anon_sym_class] = ACTIONS(1582),
    [anon_sym_val] = ACTIONS(1582),
    [anon_sym_end] = ACTIONS(1582),
    [anon_sym_object] = ACTIONS(1582),
    [anon_sym_AMP] = ACTIONS(1582),
    [anon_sym_POUND] = ACTIONS(1582),
    [anon_sym_COLON_COLON] = ACTIONS(1584),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1584),
    [anon_sym_else] = ACTIONS(1582),
    [anon_sym_new] = ACTIONS(1582),
    [anon_sym_LBRACE_LT] = ACTIONS(1584),
    [anon_sym_begin] = ACTIONS(1582),
    [sym_ocamlyacc_value] = ACTIONS(1584),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1582),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1584),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1582),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1582),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1584),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1584),
    [aux_sym_number_token1] = ACTIONS(1584),
    [aux_sym_signed_number_token1] = ACTIONS(1582),
    [anon_sym_SQUOTE] = ACTIONS(1584),
    [anon_sym_DQUOTE] = ACTIONS(1584),
    [sym_prefix_operator] = ACTIONS(1584),
    [aux_sym_sign_operator_token1] = ACTIONS(1582),
    [sym_hash_operator] = ACTIONS(1584),
    [anon_sym_lsl] = ACTIONS(1582),
    [anon_sym_lsr] = ACTIONS(1582),
    [anon_sym_asr] = ACTIONS(1582),
    [aux_sym_pow_operator_token1] = ACTIONS(1584),
    [anon_sym_mod] = ACTIONS(1582),
    [anon_sym_land] = ACTIONS(1582),
    [anon_sym_lor] = ACTIONS(1582),
    [anon_sym_lxor] = ACTIONS(1582),
    [aux_sym_mult_operator_token1] = ACTIONS(1582),
    [aux_sym_add_operator_token1] = ACTIONS(1582),
    [sym_concat_operator] = ACTIONS(1584),
    [sym_rel_operator] = ACTIONS(1582),
    [anon_sym_AMP_AMP] = ACTIONS(1582),
    [anon_sym_or] = ACTIONS(1582),
    [anon_sym_PIPE_PIPE] = ACTIONS(1582),
    [sym_assign_operator] = ACTIONS(1584),
    [sym_let_operator] = ACTIONS(1584),
    [sym__uppercase_identifier] = ACTIONS(1584),
    [aux_sym_directive_token1] = ACTIONS(1582),
    [aux_sym_tag_token1] = ACTIONS(1584),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1117)] = {
    [sym_attribute] = STATE(1117),
    [sym__lowercase_identifier] = ACTIONS(1894),
    [anon_sym_SEMI_SEMI] = ACTIONS(1896),
    [anon_sym_let] = ACTIONS(1894),
    [anon_sym_and] = ACTIONS(1894),
    [anon_sym_TILDE] = ACTIONS(1894),
    [anon_sym_QMARK] = ACTIONS(1894),
    [anon_sym_LPAREN] = ACTIONS(1896),
    [anon_sym_external] = ACTIONS(1894),
    [anon_sym_type] = ACTIONS(1894),
    [anon_sym_COMMA] = ACTIONS(1896),
    [anon_sym_PIPE] = ACTIONS(1894),
    [anon_sym_DOT] = ACTIONS(963),
    [anon_sym_LBRACK] = ACTIONS(1894),
    [anon_sym_RBRACK] = ACTIONS(1896),
    [anon_sym_true] = ACTIONS(1894),
    [anon_sym_false] = ACTIONS(1894),
    [anon_sym_LBRACE] = ACTIONS(1894),
    [anon_sym_SEMI] = ACTIONS(1894),
    [anon_sym_exception] = ACTIONS(1894),
    [anon_sym_module] = ACTIONS(1894),
    [anon_sym_open] = ACTIONS(1894),
    [anon_sym_include] = ACTIONS(1894),
    [anon_sym_class] = ACTIONS(1894),
    [anon_sym_val] = ACTIONS(1894),
    [anon_sym_end] = ACTIONS(1894),
    [anon_sym_object] = ACTIONS(1894),
    [anon_sym_AMP] = ACTIONS(1894),
    [anon_sym_POUND] = ACTIONS(1894),
    [anon_sym_COLON_COLON] = ACTIONS(1896),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1896),
    [anon_sym_else] = ACTIONS(1894),
    [anon_sym_new] = ACTIONS(1894),
    [anon_sym_LBRACE_LT] = ACTIONS(1896),
    [anon_sym_begin] = ACTIONS(1894),
    [sym_ocamlyacc_value] = ACTIONS(1896),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1894),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1896),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1894),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1894),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1896),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1896),
    [aux_sym_number_token1] = ACTIONS(1896),
    [aux_sym_signed_number_token1] = ACTIONS(1894),
    [anon_sym_SQUOTE] = ACTIONS(1896),
    [anon_sym_DQUOTE] = ACTIONS(1896),
    [sym_prefix_operator] = ACTIONS(1896),
    [aux_sym_sign_operator_token1] = ACTIONS(1894),
    [sym_hash_operator] = ACTIONS(1896),
    [anon_sym_lsl] = ACTIONS(1894),
    [anon_sym_lsr] = ACTIONS(1894),
    [anon_sym_asr] = ACTIONS(1894),
    [aux_sym_pow_operator_token1] = ACTIONS(1896),
    [anon_sym_mod] = ACTIONS(1894),
    [anon_sym_land] = ACTIONS(1894),
    [anon_sym_lor] = ACTIONS(1894),
    [anon_sym_lxor] = ACTIONS(1894),
    [aux_sym_mult_operator_token1] = ACTIONS(1894),
    [aux_sym_add_operator_token1] = ACTIONS(1894),
    [sym_concat_operator] = ACTIONS(1896),
    [sym_rel_operator] = ACTIONS(1894),
    [anon_sym_AMP_AMP] = ACTIONS(1894),
    [anon_sym_or] = ACTIONS(1894),
    [anon_sym_PIPE_PIPE] = ACTIONS(1894),
    [sym_assign_operator] = ACTIONS(1896),
    [sym_let_operator] = ACTIONS(1896),
    [sym__uppercase_identifier] = ACTIONS(1896),
    [aux_sym_directive_token1] = ACTIONS(1894),
    [aux_sym_tag_token1] = ACTIONS(1896),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1118)] = {
    [sym_attribute] = STATE(1118),
    [sym__lowercase_identifier] = ACTIONS(1882),
    [anon_sym_SEMI_SEMI] = ACTIONS(1884),
    [anon_sym_let] = ACTIONS(1882),
    [anon_sym_and] = ACTIONS(1882),
    [anon_sym_TILDE] = ACTIONS(1882),
    [anon_sym_QMARK] = ACTIONS(1882),
    [anon_sym_LPAREN] = ACTIONS(1884),
    [anon_sym_external] = ACTIONS(1882),
    [anon_sym_type] = ACTIONS(1882),
    [anon_sym_COMMA] = ACTIONS(1884),
    [anon_sym_PIPE] = ACTIONS(1882),
    [anon_sym_DOT] = ACTIONS(1884),
    [anon_sym_LBRACK] = ACTIONS(1882),
    [anon_sym_RBRACK] = ACTIONS(1884),
    [anon_sym_true] = ACTIONS(1882),
    [anon_sym_false] = ACTIONS(1882),
    [anon_sym_LBRACE] = ACTIONS(1882),
    [anon_sym_SEMI] = ACTIONS(1882),
    [anon_sym_exception] = ACTIONS(1882),
    [anon_sym_module] = ACTIONS(1882),
    [anon_sym_open] = ACTIONS(1882),
    [anon_sym_include] = ACTIONS(1882),
    [anon_sym_class] = ACTIONS(1882),
    [anon_sym_val] = ACTIONS(1882),
    [anon_sym_end] = ACTIONS(1882),
    [anon_sym_object] = ACTIONS(1882),
    [anon_sym_AMP] = ACTIONS(1882),
    [anon_sym_POUND] = ACTIONS(1882),
    [anon_sym_COLON_COLON] = ACTIONS(1884),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1884),
    [anon_sym_else] = ACTIONS(1882),
    [anon_sym_new] = ACTIONS(1882),
    [anon_sym_LBRACE_LT] = ACTIONS(1884),
    [anon_sym_begin] = ACTIONS(1882),
    [sym_ocamlyacc_value] = ACTIONS(1884),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1882),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1884),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1882),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1882),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1884),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1884),
    [aux_sym_number_token1] = ACTIONS(1884),
    [aux_sym_signed_number_token1] = ACTIONS(1882),
    [anon_sym_SQUOTE] = ACTIONS(1884),
    [anon_sym_DQUOTE] = ACTIONS(1884),
    [sym_prefix_operator] = ACTIONS(1884),
    [aux_sym_sign_operator_token1] = ACTIONS(1882),
    [sym_hash_operator] = ACTIONS(1884),
    [anon_sym_lsl] = ACTIONS(1882),
    [anon_sym_lsr] = ACTIONS(1882),
    [anon_sym_asr] = ACTIONS(1882),
    [aux_sym_pow_operator_token1] = ACTIONS(1884),
    [anon_sym_mod] = ACTIONS(1882),
    [anon_sym_land] = ACTIONS(1882),
    [anon_sym_lor] = ACTIONS(1882),
    [anon_sym_lxor] = ACTIONS(1882),
    [aux_sym_mult_operator_token1] = ACTIONS(1882),
    [aux_sym_add_operator_token1] = ACTIONS(1882),
    [sym_concat_operator] = ACTIONS(1884),
    [sym_rel_operator] = ACTIONS(1882),
    [anon_sym_AMP_AMP] = ACTIONS(1882),
    [anon_sym_or] = ACTIONS(1882),
    [anon_sym_PIPE_PIPE] = ACTIONS(1882),
    [sym_assign_operator] = ACTIONS(1884),
    [sym_let_operator] = ACTIONS(1884),
    [sym__uppercase_identifier] = ACTIONS(1884),
    [aux_sym_directive_token1] = ACTIONS(1882),
    [aux_sym_tag_token1] = ACTIONS(1884),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1119)] = {
    [sym_attribute] = STATE(1119),
    [sym__lowercase_identifier] = ACTIONS(1838),
    [anon_sym_SEMI_SEMI] = ACTIONS(1840),
    [anon_sym_let] = ACTIONS(1838),
    [anon_sym_and] = ACTIONS(1838),
    [anon_sym_TILDE] = ACTIONS(1838),
    [anon_sym_QMARK] = ACTIONS(1838),
    [anon_sym_LPAREN] = ACTIONS(1840),
    [anon_sym_external] = ACTIONS(1838),
    [anon_sym_type] = ACTIONS(1838),
    [anon_sym_COMMA] = ACTIONS(1840),
    [anon_sym_PIPE] = ACTIONS(1838),
    [anon_sym_DOT] = ACTIONS(1840),
    [anon_sym_LBRACK] = ACTIONS(1838),
    [anon_sym_RBRACK] = ACTIONS(1840),
    [anon_sym_true] = ACTIONS(1838),
    [anon_sym_false] = ACTIONS(1838),
    [anon_sym_LBRACE] = ACTIONS(1838),
    [anon_sym_SEMI] = ACTIONS(1838),
    [anon_sym_exception] = ACTIONS(1838),
    [anon_sym_module] = ACTIONS(1838),
    [anon_sym_open] = ACTIONS(1838),
    [anon_sym_include] = ACTIONS(1838),
    [anon_sym_class] = ACTIONS(1838),
    [anon_sym_val] = ACTIONS(1838),
    [anon_sym_end] = ACTIONS(1838),
    [anon_sym_object] = ACTIONS(1838),
    [anon_sym_AMP] = ACTIONS(1838),
    [anon_sym_POUND] = ACTIONS(1838),
    [anon_sym_COLON_COLON] = ACTIONS(1840),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1840),
    [anon_sym_else] = ACTIONS(1838),
    [anon_sym_new] = ACTIONS(1838),
    [anon_sym_LBRACE_LT] = ACTIONS(1840),
    [anon_sym_begin] = ACTIONS(1838),
    [sym_ocamlyacc_value] = ACTIONS(1840),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1838),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1840),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1838),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1838),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1840),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1840),
    [aux_sym_number_token1] = ACTIONS(1840),
    [aux_sym_signed_number_token1] = ACTIONS(1838),
    [anon_sym_SQUOTE] = ACTIONS(1840),
    [anon_sym_DQUOTE] = ACTIONS(1840),
    [sym_prefix_operator] = ACTIONS(1840),
    [aux_sym_sign_operator_token1] = ACTIONS(1838),
    [sym_hash_operator] = ACTIONS(1840),
    [anon_sym_lsl] = ACTIONS(1838),
    [anon_sym_lsr] = ACTIONS(1838),
    [anon_sym_asr] = ACTIONS(1838),
    [aux_sym_pow_operator_token1] = ACTIONS(1840),
    [anon_sym_mod] = ACTIONS(1838),
    [anon_sym_land] = ACTIONS(1838),
    [anon_sym_lor] = ACTIONS(1838),
    [anon_sym_lxor] = ACTIONS(1838),
    [aux_sym_mult_operator_token1] = ACTIONS(1838),
    [aux_sym_add_operator_token1] = ACTIONS(1838),
    [sym_concat_operator] = ACTIONS(1840),
    [sym_rel_operator] = ACTIONS(1838),
    [anon_sym_AMP_AMP] = ACTIONS(1838),
    [anon_sym_or] = ACTIONS(1838),
    [anon_sym_PIPE_PIPE] = ACTIONS(1838),
    [sym_assign_operator] = ACTIONS(1840),
    [sym_let_operator] = ACTIONS(1840),
    [sym__uppercase_identifier] = ACTIONS(1840),
    [aux_sym_directive_token1] = ACTIONS(1838),
    [aux_sym_tag_token1] = ACTIONS(1840),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1120)] = {
    [sym_attribute] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(1610),
    [anon_sym_SEMI_SEMI] = ACTIONS(1612),
    [anon_sym_let] = ACTIONS(1610),
    [anon_sym_and] = ACTIONS(1610),
    [anon_sym_TILDE] = ACTIONS(1610),
    [anon_sym_QMARK] = ACTIONS(1610),
    [anon_sym_LPAREN] = ACTIONS(1612),
    [anon_sym_external] = ACTIONS(1610),
    [anon_sym_type] = ACTIONS(1610),
    [anon_sym_COMMA] = ACTIONS(1612),
    [anon_sym_PIPE] = ACTIONS(1610),
    [anon_sym_DOT] = ACTIONS(1612),
    [anon_sym_LBRACK] = ACTIONS(1610),
    [anon_sym_RBRACK] = ACTIONS(1612),
    [anon_sym_true] = ACTIONS(1610),
    [anon_sym_false] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1610),
    [anon_sym_SEMI] = ACTIONS(1610),
    [anon_sym_exception] = ACTIONS(1610),
    [anon_sym_module] = ACTIONS(1610),
    [anon_sym_open] = ACTIONS(1610),
    [anon_sym_include] = ACTIONS(1610),
    [anon_sym_class] = ACTIONS(1610),
    [anon_sym_val] = ACTIONS(1610),
    [anon_sym_end] = ACTIONS(1610),
    [anon_sym_object] = ACTIONS(1610),
    [anon_sym_AMP] = ACTIONS(1610),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_COLON_COLON] = ACTIONS(1612),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1612),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_new] = ACTIONS(1610),
    [anon_sym_LBRACE_LT] = ACTIONS(1612),
    [anon_sym_begin] = ACTIONS(1610),
    [sym_ocamlyacc_value] = ACTIONS(1612),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1610),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1612),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1610),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1610),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1612),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1612),
    [aux_sym_number_token1] = ACTIONS(1612),
    [aux_sym_signed_number_token1] = ACTIONS(1610),
    [anon_sym_SQUOTE] = ACTIONS(1612),
    [anon_sym_DQUOTE] = ACTIONS(1612),
    [sym_prefix_operator] = ACTIONS(1612),
    [aux_sym_sign_operator_token1] = ACTIONS(1610),
    [sym_hash_operator] = ACTIONS(1612),
    [anon_sym_lsl] = ACTIONS(1610),
    [anon_sym_lsr] = ACTIONS(1610),
    [anon_sym_asr] = ACTIONS(1610),
    [aux_sym_pow_operator_token1] = ACTIONS(1612),
    [anon_sym_mod] = ACTIONS(1610),
    [anon_sym_land] = ACTIONS(1610),
    [anon_sym_lor] = ACTIONS(1610),
    [anon_sym_lxor] = ACTIONS(1610),
    [aux_sym_mult_operator_token1] = ACTIONS(1610),
    [aux_sym_add_operator_token1] = ACTIONS(1610),
    [sym_concat_operator] = ACTIONS(1612),
    [sym_rel_operator] = ACTIONS(1610),
    [anon_sym_AMP_AMP] = ACTIONS(1610),
    [anon_sym_or] = ACTIONS(1610),
    [anon_sym_PIPE_PIPE] = ACTIONS(1610),
    [sym_assign_operator] = ACTIONS(1612),
    [sym_let_operator] = ACTIONS(1612),
    [sym__uppercase_identifier] = ACTIONS(1612),
    [aux_sym_directive_token1] = ACTIONS(1610),
    [aux_sym_tag_token1] = ACTIONS(1612),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1121)] = {
    [sym_attribute] = STATE(1121),
    [sym__lowercase_identifier] = ACTIONS(1746),
    [anon_sym_SEMI_SEMI] = ACTIONS(1748),
    [anon_sym_let] = ACTIONS(1746),
    [anon_sym_and] = ACTIONS(1746),
    [anon_sym_TILDE] = ACTIONS(1746),
    [anon_sym_QMARK] = ACTIONS(1746),
    [anon_sym_LPAREN] = ACTIONS(1748),
    [anon_sym_external] = ACTIONS(1746),
    [anon_sym_type] = ACTIONS(1746),
    [anon_sym_COMMA] = ACTIONS(1748),
    [anon_sym_PIPE] = ACTIONS(1746),
    [anon_sym_DOT] = ACTIONS(1748),
    [anon_sym_LBRACK] = ACTIONS(1746),
    [anon_sym_RBRACK] = ACTIONS(1748),
    [anon_sym_true] = ACTIONS(1746),
    [anon_sym_false] = ACTIONS(1746),
    [anon_sym_LBRACE] = ACTIONS(1746),
    [anon_sym_SEMI] = ACTIONS(1746),
    [anon_sym_exception] = ACTIONS(1746),
    [anon_sym_module] = ACTIONS(1746),
    [anon_sym_open] = ACTIONS(1746),
    [anon_sym_include] = ACTIONS(1746),
    [anon_sym_class] = ACTIONS(1746),
    [anon_sym_val] = ACTIONS(1746),
    [anon_sym_end] = ACTIONS(1746),
    [anon_sym_object] = ACTIONS(1746),
    [anon_sym_AMP] = ACTIONS(1746),
    [anon_sym_POUND] = ACTIONS(1746),
    [anon_sym_COLON_COLON] = ACTIONS(1748),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1748),
    [anon_sym_else] = ACTIONS(1746),
    [anon_sym_new] = ACTIONS(1746),
    [anon_sym_LBRACE_LT] = ACTIONS(1748),
    [anon_sym_begin] = ACTIONS(1746),
    [sym_ocamlyacc_value] = ACTIONS(1748),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1746),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1748),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1746),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1746),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1748),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1748),
    [aux_sym_number_token1] = ACTIONS(1748),
    [aux_sym_signed_number_token1] = ACTIONS(1746),
    [anon_sym_SQUOTE] = ACTIONS(1748),
    [anon_sym_DQUOTE] = ACTIONS(1748),
    [sym_prefix_operator] = ACTIONS(1748),
    [aux_sym_sign_operator_token1] = ACTIONS(1746),
    [sym_hash_operator] = ACTIONS(1748),
    [anon_sym_lsl] = ACTIONS(1746),
    [anon_sym_lsr] = ACTIONS(1746),
    [anon_sym_asr] = ACTIONS(1746),
    [aux_sym_pow_operator_token1] = ACTIONS(1748),
    [anon_sym_mod] = ACTIONS(1746),
    [anon_sym_land] = ACTIONS(1746),
    [anon_sym_lor] = ACTIONS(1746),
    [anon_sym_lxor] = ACTIONS(1746),
    [aux_sym_mult_operator_token1] = ACTIONS(1746),
    [aux_sym_add_operator_token1] = ACTIONS(1746),
    [sym_concat_operator] = ACTIONS(1748),
    [sym_rel_operator] = ACTIONS(1746),
    [anon_sym_AMP_AMP] = ACTIONS(1746),
    [anon_sym_or] = ACTIONS(1746),
    [anon_sym_PIPE_PIPE] = ACTIONS(1746),
    [sym_assign_operator] = ACTIONS(1748),
    [sym_let_operator] = ACTIONS(1748),
    [sym__uppercase_identifier] = ACTIONS(1748),
    [aux_sym_directive_token1] = ACTIONS(1746),
    [aux_sym_tag_token1] = ACTIONS(1748),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1122)] = {
    [sym_attribute] = STATE(1122),
    [sym__lowercase_identifier] = ACTIONS(1650),
    [anon_sym_SEMI_SEMI] = ACTIONS(1652),
    [anon_sym_let] = ACTIONS(1650),
    [anon_sym_and] = ACTIONS(1650),
    [anon_sym_TILDE] = ACTIONS(1650),
    [anon_sym_QMARK] = ACTIONS(1650),
    [anon_sym_LPAREN] = ACTIONS(1652),
    [anon_sym_external] = ACTIONS(1650),
    [anon_sym_type] = ACTIONS(1650),
    [anon_sym_COMMA] = ACTIONS(1652),
    [anon_sym_PIPE] = ACTIONS(1650),
    [anon_sym_DOT] = ACTIONS(1652),
    [anon_sym_LBRACK] = ACTIONS(1650),
    [anon_sym_RBRACK] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1650),
    [anon_sym_false] = ACTIONS(1650),
    [anon_sym_LBRACE] = ACTIONS(1650),
    [anon_sym_SEMI] = ACTIONS(1650),
    [anon_sym_exception] = ACTIONS(1650),
    [anon_sym_module] = ACTIONS(1650),
    [anon_sym_open] = ACTIONS(1650),
    [anon_sym_include] = ACTIONS(1650),
    [anon_sym_class] = ACTIONS(1650),
    [anon_sym_val] = ACTIONS(1650),
    [anon_sym_end] = ACTIONS(1650),
    [anon_sym_object] = ACTIONS(1650),
    [anon_sym_AMP] = ACTIONS(1650),
    [anon_sym_POUND] = ACTIONS(1650),
    [anon_sym_COLON_COLON] = ACTIONS(1652),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1652),
    [anon_sym_else] = ACTIONS(1650),
    [anon_sym_new] = ACTIONS(1650),
    [anon_sym_LBRACE_LT] = ACTIONS(1652),
    [anon_sym_begin] = ACTIONS(1650),
    [sym_ocamlyacc_value] = ACTIONS(1652),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1650),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1652),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1650),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1650),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1652),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1652),
    [aux_sym_number_token1] = ACTIONS(1652),
    [aux_sym_signed_number_token1] = ACTIONS(1650),
    [anon_sym_SQUOTE] = ACTIONS(1652),
    [anon_sym_DQUOTE] = ACTIONS(1652),
    [sym_prefix_operator] = ACTIONS(1652),
    [aux_sym_sign_operator_token1] = ACTIONS(1650),
    [sym_hash_operator] = ACTIONS(1652),
    [anon_sym_lsl] = ACTIONS(1650),
    [anon_sym_lsr] = ACTIONS(1650),
    [anon_sym_asr] = ACTIONS(1650),
    [aux_sym_pow_operator_token1] = ACTIONS(1652),
    [anon_sym_mod] = ACTIONS(1650),
    [anon_sym_land] = ACTIONS(1650),
    [anon_sym_lor] = ACTIONS(1650),
    [anon_sym_lxor] = ACTIONS(1650),
    [aux_sym_mult_operator_token1] = ACTIONS(1650),
    [aux_sym_add_operator_token1] = ACTIONS(1650),
    [sym_concat_operator] = ACTIONS(1652),
    [sym_rel_operator] = ACTIONS(1650),
    [anon_sym_AMP_AMP] = ACTIONS(1650),
    [anon_sym_or] = ACTIONS(1650),
    [anon_sym_PIPE_PIPE] = ACTIONS(1650),
    [sym_assign_operator] = ACTIONS(1652),
    [sym_let_operator] = ACTIONS(1652),
    [sym__uppercase_identifier] = ACTIONS(1652),
    [aux_sym_directive_token1] = ACTIONS(1650),
    [aux_sym_tag_token1] = ACTIONS(1652),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1123)] = {
    [sym_attribute] = STATE(1123),
    [sym__lowercase_identifier] = ACTIONS(1630),
    [anon_sym_SEMI_SEMI] = ACTIONS(1632),
    [anon_sym_let] = ACTIONS(1630),
    [anon_sym_and] = ACTIONS(1630),
    [anon_sym_TILDE] = ACTIONS(1630),
    [anon_sym_QMARK] = ACTIONS(1630),
    [anon_sym_LPAREN] = ACTIONS(1632),
    [anon_sym_external] = ACTIONS(1630),
    [anon_sym_type] = ACTIONS(1630),
    [anon_sym_COMMA] = ACTIONS(1632),
    [anon_sym_PIPE] = ACTIONS(1630),
    [anon_sym_DOT] = ACTIONS(1632),
    [anon_sym_LBRACK] = ACTIONS(1630),
    [anon_sym_RBRACK] = ACTIONS(1632),
    [anon_sym_true] = ACTIONS(1630),
    [anon_sym_false] = ACTIONS(1630),
    [anon_sym_LBRACE] = ACTIONS(1630),
    [anon_sym_SEMI] = ACTIONS(1630),
    [anon_sym_exception] = ACTIONS(1630),
    [anon_sym_module] = ACTIONS(1630),
    [anon_sym_open] = ACTIONS(1630),
    [anon_sym_include] = ACTIONS(1630),
    [anon_sym_class] = ACTIONS(1630),
    [anon_sym_val] = ACTIONS(1630),
    [anon_sym_end] = ACTIONS(1630),
    [anon_sym_object] = ACTIONS(1630),
    [anon_sym_AMP] = ACTIONS(1630),
    [anon_sym_POUND] = ACTIONS(1630),
    [anon_sym_COLON_COLON] = ACTIONS(1632),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1632),
    [anon_sym_else] = ACTIONS(1630),
    [anon_sym_new] = ACTIONS(1630),
    [anon_sym_LBRACE_LT] = ACTIONS(1632),
    [anon_sym_begin] = ACTIONS(1630),
    [sym_ocamlyacc_value] = ACTIONS(1632),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1630),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1632),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1630),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1630),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1632),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1632),
    [aux_sym_number_token1] = ACTIONS(1632),
    [aux_sym_signed_number_token1] = ACTIONS(1630),
    [anon_sym_SQUOTE] = ACTIONS(1632),
    [anon_sym_DQUOTE] = ACTIONS(1632),
    [sym_prefix_operator] = ACTIONS(1632),
    [aux_sym_sign_operator_token1] = ACTIONS(1630),
    [sym_hash_operator] = ACTIONS(1632),
    [anon_sym_lsl] = ACTIONS(1630),
    [anon_sym_lsr] = ACTIONS(1630),
    [anon_sym_asr] = ACTIONS(1630),
    [aux_sym_pow_operator_token1] = ACTIONS(1632),
    [anon_sym_mod] = ACTIONS(1630),
    [anon_sym_land] = ACTIONS(1630),
    [anon_sym_lor] = ACTIONS(1630),
    [anon_sym_lxor] = ACTIONS(1630),
    [aux_sym_mult_operator_token1] = ACTIONS(1630),
    [aux_sym_add_operator_token1] = ACTIONS(1630),
    [sym_concat_operator] = ACTIONS(1632),
    [sym_rel_operator] = ACTIONS(1630),
    [anon_sym_AMP_AMP] = ACTIONS(1630),
    [anon_sym_or] = ACTIONS(1630),
    [anon_sym_PIPE_PIPE] = ACTIONS(1630),
    [sym_assign_operator] = ACTIONS(1632),
    [sym_let_operator] = ACTIONS(1632),
    [sym__uppercase_identifier] = ACTIONS(1632),
    [aux_sym_directive_token1] = ACTIONS(1630),
    [aux_sym_tag_token1] = ACTIONS(1632),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1124)] = {
    [sym_attribute] = STATE(1124),
    [sym__lowercase_identifier] = ACTIONS(1750),
    [anon_sym_SEMI_SEMI] = ACTIONS(1752),
    [anon_sym_let] = ACTIONS(1750),
    [anon_sym_and] = ACTIONS(1750),
    [anon_sym_TILDE] = ACTIONS(1750),
    [anon_sym_QMARK] = ACTIONS(1750),
    [anon_sym_LPAREN] = ACTIONS(1752),
    [anon_sym_external] = ACTIONS(1750),
    [anon_sym_type] = ACTIONS(1750),
    [anon_sym_COMMA] = ACTIONS(1752),
    [anon_sym_PIPE] = ACTIONS(1750),
    [anon_sym_DOT] = ACTIONS(1752),
    [anon_sym_LBRACK] = ACTIONS(1750),
    [anon_sym_RBRACK] = ACTIONS(1752),
    [anon_sym_true] = ACTIONS(1750),
    [anon_sym_false] = ACTIONS(1750),
    [anon_sym_LBRACE] = ACTIONS(1750),
    [anon_sym_SEMI] = ACTIONS(1750),
    [anon_sym_exception] = ACTIONS(1750),
    [anon_sym_module] = ACTIONS(1750),
    [anon_sym_open] = ACTIONS(1750),
    [anon_sym_include] = ACTIONS(1750),
    [anon_sym_class] = ACTIONS(1750),
    [anon_sym_val] = ACTIONS(1750),
    [anon_sym_end] = ACTIONS(1750),
    [anon_sym_object] = ACTIONS(1750),
    [anon_sym_AMP] = ACTIONS(1750),
    [anon_sym_POUND] = ACTIONS(1750),
    [anon_sym_COLON_COLON] = ACTIONS(1752),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1752),
    [anon_sym_else] = ACTIONS(1750),
    [anon_sym_new] = ACTIONS(1750),
    [anon_sym_LBRACE_LT] = ACTIONS(1752),
    [anon_sym_begin] = ACTIONS(1750),
    [sym_ocamlyacc_value] = ACTIONS(1752),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1750),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1752),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1750),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1750),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1752),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1752),
    [aux_sym_number_token1] = ACTIONS(1752),
    [aux_sym_signed_number_token1] = ACTIONS(1750),
    [anon_sym_SQUOTE] = ACTIONS(1752),
    [anon_sym_DQUOTE] = ACTIONS(1752),
    [sym_prefix_operator] = ACTIONS(1752),
    [aux_sym_sign_operator_token1] = ACTIONS(1750),
    [sym_hash_operator] = ACTIONS(1752),
    [anon_sym_lsl] = ACTIONS(1750),
    [anon_sym_lsr] = ACTIONS(1750),
    [anon_sym_asr] = ACTIONS(1750),
    [aux_sym_pow_operator_token1] = ACTIONS(1752),
    [anon_sym_mod] = ACTIONS(1750),
    [anon_sym_land] = ACTIONS(1750),
    [anon_sym_lor] = ACTIONS(1750),
    [anon_sym_lxor] = ACTIONS(1750),
    [aux_sym_mult_operator_token1] = ACTIONS(1750),
    [aux_sym_add_operator_token1] = ACTIONS(1750),
    [sym_concat_operator] = ACTIONS(1752),
    [sym_rel_operator] = ACTIONS(1750),
    [anon_sym_AMP_AMP] = ACTIONS(1750),
    [anon_sym_or] = ACTIONS(1750),
    [anon_sym_PIPE_PIPE] = ACTIONS(1750),
    [sym_assign_operator] = ACTIONS(1752),
    [sym_let_operator] = ACTIONS(1752),
    [sym__uppercase_identifier] = ACTIONS(1752),
    [aux_sym_directive_token1] = ACTIONS(1750),
    [aux_sym_tag_token1] = ACTIONS(1752),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1125)] = {
    [sym_attribute] = STATE(1125),
    [sym__lowercase_identifier] = ACTIONS(1926),
    [anon_sym_SEMI_SEMI] = ACTIONS(1928),
    [anon_sym_let] = ACTIONS(1926),
    [anon_sym_and] = ACTIONS(1926),
    [anon_sym_TILDE] = ACTIONS(1926),
    [anon_sym_QMARK] = ACTIONS(1926),
    [anon_sym_LPAREN] = ACTIONS(1928),
    [anon_sym_external] = ACTIONS(1926),
    [anon_sym_type] = ACTIONS(1926),
    [anon_sym_COMMA] = ACTIONS(1928),
    [anon_sym_PIPE] = ACTIONS(1926),
    [anon_sym_DOT] = ACTIONS(1928),
    [anon_sym_LBRACK] = ACTIONS(1926),
    [anon_sym_RBRACK] = ACTIONS(1928),
    [anon_sym_true] = ACTIONS(1926),
    [anon_sym_false] = ACTIONS(1926),
    [anon_sym_LBRACE] = ACTIONS(1926),
    [anon_sym_SEMI] = ACTIONS(1926),
    [anon_sym_exception] = ACTIONS(1926),
    [anon_sym_module] = ACTIONS(1926),
    [anon_sym_open] = ACTIONS(1926),
    [anon_sym_include] = ACTIONS(1926),
    [anon_sym_class] = ACTIONS(1926),
    [anon_sym_val] = ACTIONS(1926),
    [anon_sym_end] = ACTIONS(1926),
    [anon_sym_object] = ACTIONS(1926),
    [anon_sym_AMP] = ACTIONS(1926),
    [anon_sym_POUND] = ACTIONS(1926),
    [anon_sym_COLON_COLON] = ACTIONS(1928),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1928),
    [anon_sym_else] = ACTIONS(1926),
    [anon_sym_new] = ACTIONS(1926),
    [anon_sym_LBRACE_LT] = ACTIONS(1928),
    [anon_sym_begin] = ACTIONS(1926),
    [sym_ocamlyacc_value] = ACTIONS(1928),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1926),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1928),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1926),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1926),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1928),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1928),
    [aux_sym_number_token1] = ACTIONS(1928),
    [aux_sym_signed_number_token1] = ACTIONS(1926),
    [anon_sym_SQUOTE] = ACTIONS(1928),
    [anon_sym_DQUOTE] = ACTIONS(1928),
    [sym_prefix_operator] = ACTIONS(1928),
    [aux_sym_sign_operator_token1] = ACTIONS(1926),
    [sym_hash_operator] = ACTIONS(1928),
    [anon_sym_lsl] = ACTIONS(1926),
    [anon_sym_lsr] = ACTIONS(1926),
    [anon_sym_asr] = ACTIONS(1926),
    [aux_sym_pow_operator_token1] = ACTIONS(1928),
    [anon_sym_mod] = ACTIONS(1926),
    [anon_sym_land] = ACTIONS(1926),
    [anon_sym_lor] = ACTIONS(1926),
    [anon_sym_lxor] = ACTIONS(1926),
    [aux_sym_mult_operator_token1] = ACTIONS(1926),
    [aux_sym_add_operator_token1] = ACTIONS(1926),
    [sym_concat_operator] = ACTIONS(1928),
    [sym_rel_operator] = ACTIONS(1926),
    [anon_sym_AMP_AMP] = ACTIONS(1926),
    [anon_sym_or] = ACTIONS(1926),
    [anon_sym_PIPE_PIPE] = ACTIONS(1926),
    [sym_assign_operator] = ACTIONS(1928),
    [sym_let_operator] = ACTIONS(1928),
    [sym__uppercase_identifier] = ACTIONS(1928),
    [aux_sym_directive_token1] = ACTIONS(1926),
    [aux_sym_tag_token1] = ACTIONS(1928),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1126)] = {
    [sym_attribute] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1638),
    [anon_sym_SEMI_SEMI] = ACTIONS(1640),
    [anon_sym_let] = ACTIONS(1638),
    [anon_sym_and] = ACTIONS(1638),
    [anon_sym_TILDE] = ACTIONS(1638),
    [anon_sym_QMARK] = ACTIONS(1638),
    [anon_sym_LPAREN] = ACTIONS(1640),
    [anon_sym_external] = ACTIONS(1638),
    [anon_sym_type] = ACTIONS(1638),
    [anon_sym_COMMA] = ACTIONS(1640),
    [anon_sym_PIPE] = ACTIONS(1638),
    [anon_sym_DOT] = ACTIONS(1640),
    [anon_sym_LBRACK] = ACTIONS(1638),
    [anon_sym_RBRACK] = ACTIONS(1640),
    [anon_sym_true] = ACTIONS(1638),
    [anon_sym_false] = ACTIONS(1638),
    [anon_sym_LBRACE] = ACTIONS(1638),
    [anon_sym_SEMI] = ACTIONS(1638),
    [anon_sym_exception] = ACTIONS(1638),
    [anon_sym_module] = ACTIONS(1638),
    [anon_sym_open] = ACTIONS(1638),
    [anon_sym_include] = ACTIONS(1638),
    [anon_sym_class] = ACTIONS(1638),
    [anon_sym_val] = ACTIONS(1638),
    [anon_sym_end] = ACTIONS(1638),
    [anon_sym_object] = ACTIONS(1638),
    [anon_sym_AMP] = ACTIONS(1638),
    [anon_sym_POUND] = ACTIONS(1638),
    [anon_sym_COLON_COLON] = ACTIONS(1640),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1640),
    [anon_sym_else] = ACTIONS(1638),
    [anon_sym_new] = ACTIONS(1638),
    [anon_sym_LBRACE_LT] = ACTIONS(1640),
    [anon_sym_begin] = ACTIONS(1638),
    [sym_ocamlyacc_value] = ACTIONS(1640),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1638),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1640),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1638),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1638),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1640),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1640),
    [aux_sym_number_token1] = ACTIONS(1640),
    [aux_sym_signed_number_token1] = ACTIONS(1638),
    [anon_sym_SQUOTE] = ACTIONS(1640),
    [anon_sym_DQUOTE] = ACTIONS(1640),
    [sym_prefix_operator] = ACTIONS(1640),
    [aux_sym_sign_operator_token1] = ACTIONS(1638),
    [sym_hash_operator] = ACTIONS(1640),
    [anon_sym_lsl] = ACTIONS(1638),
    [anon_sym_lsr] = ACTIONS(1638),
    [anon_sym_asr] = ACTIONS(1638),
    [aux_sym_pow_operator_token1] = ACTIONS(1640),
    [anon_sym_mod] = ACTIONS(1638),
    [anon_sym_land] = ACTIONS(1638),
    [anon_sym_lor] = ACTIONS(1638),
    [anon_sym_lxor] = ACTIONS(1638),
    [aux_sym_mult_operator_token1] = ACTIONS(1638),
    [aux_sym_add_operator_token1] = ACTIONS(1638),
    [sym_concat_operator] = ACTIONS(1640),
    [sym_rel_operator] = ACTIONS(1638),
    [anon_sym_AMP_AMP] = ACTIONS(1638),
    [anon_sym_or] = ACTIONS(1638),
    [anon_sym_PIPE_PIPE] = ACTIONS(1638),
    [sym_assign_operator] = ACTIONS(1640),
    [sym_let_operator] = ACTIONS(1640),
    [sym__uppercase_identifier] = ACTIONS(1640),
    [aux_sym_directive_token1] = ACTIONS(1638),
    [aux_sym_tag_token1] = ACTIONS(1640),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1127)] = {
    [sym_attribute] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(1732),
    [anon_sym_SEMI_SEMI] = ACTIONS(1734),
    [anon_sym_let] = ACTIONS(1732),
    [anon_sym_and] = ACTIONS(1732),
    [anon_sym_TILDE] = ACTIONS(1732),
    [anon_sym_QMARK] = ACTIONS(1732),
    [anon_sym_LPAREN] = ACTIONS(1734),
    [anon_sym_external] = ACTIONS(1732),
    [anon_sym_type] = ACTIONS(1732),
    [anon_sym_COMMA] = ACTIONS(1734),
    [anon_sym_PIPE] = ACTIONS(1732),
    [anon_sym_DOT] = ACTIONS(1734),
    [anon_sym_LBRACK] = ACTIONS(1732),
    [anon_sym_RBRACK] = ACTIONS(1734),
    [anon_sym_true] = ACTIONS(1732),
    [anon_sym_false] = ACTIONS(1732),
    [anon_sym_LBRACE] = ACTIONS(1732),
    [anon_sym_SEMI] = ACTIONS(1732),
    [anon_sym_exception] = ACTIONS(1732),
    [anon_sym_module] = ACTIONS(1732),
    [anon_sym_open] = ACTIONS(1732),
    [anon_sym_include] = ACTIONS(1732),
    [anon_sym_class] = ACTIONS(1732),
    [anon_sym_val] = ACTIONS(1732),
    [anon_sym_end] = ACTIONS(1732),
    [anon_sym_object] = ACTIONS(1732),
    [anon_sym_AMP] = ACTIONS(1732),
    [anon_sym_POUND] = ACTIONS(1732),
    [anon_sym_COLON_COLON] = ACTIONS(1734),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1734),
    [anon_sym_else] = ACTIONS(1732),
    [anon_sym_new] = ACTIONS(1732),
    [anon_sym_LBRACE_LT] = ACTIONS(1734),
    [anon_sym_begin] = ACTIONS(1732),
    [sym_ocamlyacc_value] = ACTIONS(1734),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1732),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1734),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1732),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1732),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1734),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1734),
    [aux_sym_number_token1] = ACTIONS(1734),
    [aux_sym_signed_number_token1] = ACTIONS(1732),
    [anon_sym_SQUOTE] = ACTIONS(1734),
    [anon_sym_DQUOTE] = ACTIONS(1734),
    [sym_prefix_operator] = ACTIONS(1734),
    [aux_sym_sign_operator_token1] = ACTIONS(1732),
    [sym_hash_operator] = ACTIONS(1734),
    [anon_sym_lsl] = ACTIONS(1732),
    [anon_sym_lsr] = ACTIONS(1732),
    [anon_sym_asr] = ACTIONS(1732),
    [aux_sym_pow_operator_token1] = ACTIONS(1734),
    [anon_sym_mod] = ACTIONS(1732),
    [anon_sym_land] = ACTIONS(1732),
    [anon_sym_lor] = ACTIONS(1732),
    [anon_sym_lxor] = ACTIONS(1732),
    [aux_sym_mult_operator_token1] = ACTIONS(1732),
    [aux_sym_add_operator_token1] = ACTIONS(1732),
    [sym_concat_operator] = ACTIONS(1734),
    [sym_rel_operator] = ACTIONS(1732),
    [anon_sym_AMP_AMP] = ACTIONS(1732),
    [anon_sym_or] = ACTIONS(1732),
    [anon_sym_PIPE_PIPE] = ACTIONS(1732),
    [sym_assign_operator] = ACTIONS(1734),
    [sym_let_operator] = ACTIONS(1734),
    [sym__uppercase_identifier] = ACTIONS(1734),
    [aux_sym_directive_token1] = ACTIONS(1732),
    [aux_sym_tag_token1] = ACTIONS(1734),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1128)] = {
    [sym_attribute] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(1694),
    [anon_sym_SEMI_SEMI] = ACTIONS(1696),
    [anon_sym_let] = ACTIONS(1694),
    [anon_sym_and] = ACTIONS(1694),
    [anon_sym_TILDE] = ACTIONS(1694),
    [anon_sym_QMARK] = ACTIONS(1694),
    [anon_sym_LPAREN] = ACTIONS(1696),
    [anon_sym_external] = ACTIONS(1694),
    [anon_sym_type] = ACTIONS(1694),
    [anon_sym_COMMA] = ACTIONS(1696),
    [anon_sym_PIPE] = ACTIONS(1694),
    [anon_sym_DOT] = ACTIONS(1696),
    [anon_sym_LBRACK] = ACTIONS(1694),
    [anon_sym_RBRACK] = ACTIONS(1696),
    [anon_sym_true] = ACTIONS(1694),
    [anon_sym_false] = ACTIONS(1694),
    [anon_sym_LBRACE] = ACTIONS(1694),
    [anon_sym_SEMI] = ACTIONS(1694),
    [anon_sym_exception] = ACTIONS(1694),
    [anon_sym_module] = ACTIONS(1694),
    [anon_sym_open] = ACTIONS(1694),
    [anon_sym_include] = ACTIONS(1694),
    [anon_sym_class] = ACTIONS(1694),
    [anon_sym_val] = ACTIONS(1694),
    [anon_sym_end] = ACTIONS(1694),
    [anon_sym_object] = ACTIONS(1694),
    [anon_sym_AMP] = ACTIONS(1694),
    [anon_sym_POUND] = ACTIONS(1694),
    [anon_sym_COLON_COLON] = ACTIONS(1696),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1696),
    [anon_sym_else] = ACTIONS(1694),
    [anon_sym_new] = ACTIONS(1694),
    [anon_sym_LBRACE_LT] = ACTIONS(1696),
    [anon_sym_begin] = ACTIONS(1694),
    [sym_ocamlyacc_value] = ACTIONS(1696),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1694),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1696),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1694),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1694),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1696),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1696),
    [aux_sym_number_token1] = ACTIONS(1696),
    [aux_sym_signed_number_token1] = ACTIONS(1694),
    [anon_sym_SQUOTE] = ACTIONS(1696),
    [anon_sym_DQUOTE] = ACTIONS(1696),
    [sym_prefix_operator] = ACTIONS(1696),
    [aux_sym_sign_operator_token1] = ACTIONS(1694),
    [sym_hash_operator] = ACTIONS(1696),
    [anon_sym_lsl] = ACTIONS(1694),
    [anon_sym_lsr] = ACTIONS(1694),
    [anon_sym_asr] = ACTIONS(1694),
    [aux_sym_pow_operator_token1] = ACTIONS(1696),
    [anon_sym_mod] = ACTIONS(1694),
    [anon_sym_land] = ACTIONS(1694),
    [anon_sym_lor] = ACTIONS(1694),
    [anon_sym_lxor] = ACTIONS(1694),
    [aux_sym_mult_operator_token1] = ACTIONS(1694),
    [aux_sym_add_operator_token1] = ACTIONS(1694),
    [sym_concat_operator] = ACTIONS(1696),
    [sym_rel_operator] = ACTIONS(1694),
    [anon_sym_AMP_AMP] = ACTIONS(1694),
    [anon_sym_or] = ACTIONS(1694),
    [anon_sym_PIPE_PIPE] = ACTIONS(1694),
    [sym_assign_operator] = ACTIONS(1696),
    [sym_let_operator] = ACTIONS(1696),
    [sym__uppercase_identifier] = ACTIONS(1696),
    [aux_sym_directive_token1] = ACTIONS(1694),
    [aux_sym_tag_token1] = ACTIONS(1696),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1129)] = {
    [sym_attribute] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1712),
    [anon_sym_SEMI_SEMI] = ACTIONS(1714),
    [anon_sym_let] = ACTIONS(1712),
    [anon_sym_and] = ACTIONS(1712),
    [anon_sym_TILDE] = ACTIONS(1712),
    [anon_sym_QMARK] = ACTIONS(1712),
    [anon_sym_LPAREN] = ACTIONS(1714),
    [anon_sym_external] = ACTIONS(1712),
    [anon_sym_type] = ACTIONS(1712),
    [anon_sym_COMMA] = ACTIONS(1714),
    [anon_sym_PIPE] = ACTIONS(1712),
    [anon_sym_DOT] = ACTIONS(1714),
    [anon_sym_LBRACK] = ACTIONS(1712),
    [anon_sym_RBRACK] = ACTIONS(1714),
    [anon_sym_true] = ACTIONS(1712),
    [anon_sym_false] = ACTIONS(1712),
    [anon_sym_LBRACE] = ACTIONS(1712),
    [anon_sym_SEMI] = ACTIONS(1712),
    [anon_sym_exception] = ACTIONS(1712),
    [anon_sym_module] = ACTIONS(1712),
    [anon_sym_open] = ACTIONS(1712),
    [anon_sym_include] = ACTIONS(1712),
    [anon_sym_class] = ACTIONS(1712),
    [anon_sym_val] = ACTIONS(1712),
    [anon_sym_end] = ACTIONS(1712),
    [anon_sym_object] = ACTIONS(1712),
    [anon_sym_AMP] = ACTIONS(1712),
    [anon_sym_POUND] = ACTIONS(1712),
    [anon_sym_COLON_COLON] = ACTIONS(1714),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1714),
    [anon_sym_else] = ACTIONS(1712),
    [anon_sym_new] = ACTIONS(1712),
    [anon_sym_LBRACE_LT] = ACTIONS(1714),
    [anon_sym_begin] = ACTIONS(1712),
    [sym_ocamlyacc_value] = ACTIONS(1714),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1712),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1714),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1712),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1712),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1714),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1714),
    [aux_sym_number_token1] = ACTIONS(1714),
    [aux_sym_signed_number_token1] = ACTIONS(1712),
    [anon_sym_SQUOTE] = ACTIONS(1714),
    [anon_sym_DQUOTE] = ACTIONS(1714),
    [sym_prefix_operator] = ACTIONS(1714),
    [aux_sym_sign_operator_token1] = ACTIONS(1712),
    [sym_hash_operator] = ACTIONS(1714),
    [anon_sym_lsl] = ACTIONS(1712),
    [anon_sym_lsr] = ACTIONS(1712),
    [anon_sym_asr] = ACTIONS(1712),
    [aux_sym_pow_operator_token1] = ACTIONS(1714),
    [anon_sym_mod] = ACTIONS(1712),
    [anon_sym_land] = ACTIONS(1712),
    [anon_sym_lor] = ACTIONS(1712),
    [anon_sym_lxor] = ACTIONS(1712),
    [aux_sym_mult_operator_token1] = ACTIONS(1712),
    [aux_sym_add_operator_token1] = ACTIONS(1712),
    [sym_concat_operator] = ACTIONS(1714),
    [sym_rel_operator] = ACTIONS(1712),
    [anon_sym_AMP_AMP] = ACTIONS(1712),
    [anon_sym_or] = ACTIONS(1712),
    [anon_sym_PIPE_PIPE] = ACTIONS(1712),
    [sym_assign_operator] = ACTIONS(1714),
    [sym_let_operator] = ACTIONS(1714),
    [sym__uppercase_identifier] = ACTIONS(1714),
    [aux_sym_directive_token1] = ACTIONS(1712),
    [aux_sym_tag_token1] = ACTIONS(1714),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1130)] = {
    [sym_attribute] = STATE(1130),
    [sym__lowercase_identifier] = ACTIONS(1902),
    [anon_sym_SEMI_SEMI] = ACTIONS(1904),
    [anon_sym_let] = ACTIONS(1902),
    [anon_sym_and] = ACTIONS(1902),
    [anon_sym_TILDE] = ACTIONS(1902),
    [anon_sym_QMARK] = ACTIONS(1902),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_external] = ACTIONS(1902),
    [anon_sym_type] = ACTIONS(1902),
    [anon_sym_COMMA] = ACTIONS(1904),
    [anon_sym_PIPE] = ACTIONS(1902),
    [anon_sym_DOT] = ACTIONS(1904),
    [anon_sym_LBRACK] = ACTIONS(1902),
    [anon_sym_RBRACK] = ACTIONS(1904),
    [anon_sym_true] = ACTIONS(1902),
    [anon_sym_false] = ACTIONS(1902),
    [anon_sym_LBRACE] = ACTIONS(1902),
    [anon_sym_SEMI] = ACTIONS(1902),
    [anon_sym_exception] = ACTIONS(1902),
    [anon_sym_module] = ACTIONS(1902),
    [anon_sym_open] = ACTIONS(1902),
    [anon_sym_include] = ACTIONS(1902),
    [anon_sym_class] = ACTIONS(1902),
    [anon_sym_val] = ACTIONS(1902),
    [anon_sym_end] = ACTIONS(1902),
    [anon_sym_object] = ACTIONS(1902),
    [anon_sym_AMP] = ACTIONS(1902),
    [anon_sym_POUND] = ACTIONS(1902),
    [anon_sym_COLON_COLON] = ACTIONS(1904),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1904),
    [anon_sym_else] = ACTIONS(1902),
    [anon_sym_new] = ACTIONS(1902),
    [anon_sym_LBRACE_LT] = ACTIONS(1904),
    [anon_sym_begin] = ACTIONS(1902),
    [sym_ocamlyacc_value] = ACTIONS(1904),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1902),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1904),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1902),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1902),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1904),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1904),
    [aux_sym_number_token1] = ACTIONS(1904),
    [aux_sym_signed_number_token1] = ACTIONS(1902),
    [anon_sym_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1904),
    [sym_prefix_operator] = ACTIONS(1904),
    [aux_sym_sign_operator_token1] = ACTIONS(1902),
    [sym_hash_operator] = ACTIONS(1904),
    [anon_sym_lsl] = ACTIONS(1902),
    [anon_sym_lsr] = ACTIONS(1902),
    [anon_sym_asr] = ACTIONS(1902),
    [aux_sym_pow_operator_token1] = ACTIONS(1904),
    [anon_sym_mod] = ACTIONS(1902),
    [anon_sym_land] = ACTIONS(1902),
    [anon_sym_lor] = ACTIONS(1902),
    [anon_sym_lxor] = ACTIONS(1902),
    [aux_sym_mult_operator_token1] = ACTIONS(1902),
    [aux_sym_add_operator_token1] = ACTIONS(1902),
    [sym_concat_operator] = ACTIONS(1904),
    [sym_rel_operator] = ACTIONS(1902),
    [anon_sym_AMP_AMP] = ACTIONS(1902),
    [anon_sym_or] = ACTIONS(1902),
    [anon_sym_PIPE_PIPE] = ACTIONS(1902),
    [sym_assign_operator] = ACTIONS(1904),
    [sym_let_operator] = ACTIONS(1904),
    [sym__uppercase_identifier] = ACTIONS(1904),
    [aux_sym_directive_token1] = ACTIONS(1902),
    [aux_sym_tag_token1] = ACTIONS(1904),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1131)] = {
    [sym_attribute] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(1724),
    [anon_sym_SEMI_SEMI] = ACTIONS(1726),
    [anon_sym_let] = ACTIONS(1724),
    [anon_sym_and] = ACTIONS(1724),
    [anon_sym_TILDE] = ACTIONS(1724),
    [anon_sym_QMARK] = ACTIONS(1724),
    [anon_sym_LPAREN] = ACTIONS(1726),
    [anon_sym_external] = ACTIONS(1724),
    [anon_sym_type] = ACTIONS(1724),
    [anon_sym_COMMA] = ACTIONS(1726),
    [anon_sym_PIPE] = ACTIONS(1724),
    [anon_sym_DOT] = ACTIONS(1726),
    [anon_sym_LBRACK] = ACTIONS(1724),
    [anon_sym_RBRACK] = ACTIONS(1726),
    [anon_sym_true] = ACTIONS(1724),
    [anon_sym_false] = ACTIONS(1724),
    [anon_sym_LBRACE] = ACTIONS(1724),
    [anon_sym_SEMI] = ACTIONS(1724),
    [anon_sym_exception] = ACTIONS(1724),
    [anon_sym_module] = ACTIONS(1724),
    [anon_sym_open] = ACTIONS(1724),
    [anon_sym_include] = ACTIONS(1724),
    [anon_sym_class] = ACTIONS(1724),
    [anon_sym_val] = ACTIONS(1724),
    [anon_sym_end] = ACTIONS(1724),
    [anon_sym_object] = ACTIONS(1724),
    [anon_sym_AMP] = ACTIONS(1724),
    [anon_sym_POUND] = ACTIONS(1724),
    [anon_sym_COLON_COLON] = ACTIONS(1726),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1726),
    [anon_sym_else] = ACTIONS(1724),
    [anon_sym_new] = ACTIONS(1724),
    [anon_sym_LBRACE_LT] = ACTIONS(1726),
    [anon_sym_begin] = ACTIONS(1724),
    [sym_ocamlyacc_value] = ACTIONS(1726),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1724),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1726),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1726),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1726),
    [aux_sym_number_token1] = ACTIONS(1726),
    [aux_sym_signed_number_token1] = ACTIONS(1724),
    [anon_sym_SQUOTE] = ACTIONS(1726),
    [anon_sym_DQUOTE] = ACTIONS(1726),
    [sym_prefix_operator] = ACTIONS(1726),
    [aux_sym_sign_operator_token1] = ACTIONS(1724),
    [sym_hash_operator] = ACTIONS(1726),
    [anon_sym_lsl] = ACTIONS(1724),
    [anon_sym_lsr] = ACTIONS(1724),
    [anon_sym_asr] = ACTIONS(1724),
    [aux_sym_pow_operator_token1] = ACTIONS(1726),
    [anon_sym_mod] = ACTIONS(1724),
    [anon_sym_land] = ACTIONS(1724),
    [anon_sym_lor] = ACTIONS(1724),
    [anon_sym_lxor] = ACTIONS(1724),
    [aux_sym_mult_operator_token1] = ACTIONS(1724),
    [aux_sym_add_operator_token1] = ACTIONS(1724),
    [sym_concat_operator] = ACTIONS(1726),
    [sym_rel_operator] = ACTIONS(1724),
    [anon_sym_AMP_AMP] = ACTIONS(1724),
    [anon_sym_or] = ACTIONS(1724),
    [anon_sym_PIPE_PIPE] = ACTIONS(1724),
    [sym_assign_operator] = ACTIONS(1726),
    [sym_let_operator] = ACTIONS(1726),
    [sym__uppercase_identifier] = ACTIONS(1726),
    [aux_sym_directive_token1] = ACTIONS(1724),
    [aux_sym_tag_token1] = ACTIONS(1726),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1132)] = {
    [sym_attribute] = STATE(1132),
    [sym__lowercase_identifier] = ACTIONS(1906),
    [anon_sym_SEMI_SEMI] = ACTIONS(1908),
    [anon_sym_let] = ACTIONS(1906),
    [anon_sym_and] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1906),
    [anon_sym_QMARK] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1908),
    [anon_sym_external] = ACTIONS(1906),
    [anon_sym_type] = ACTIONS(1906),
    [anon_sym_COMMA] = ACTIONS(1908),
    [anon_sym_PIPE] = ACTIONS(1906),
    [anon_sym_DOT] = ACTIONS(1908),
    [anon_sym_LBRACK] = ACTIONS(1906),
    [anon_sym_RBRACK] = ACTIONS(1908),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_LBRACE] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1906),
    [anon_sym_exception] = ACTIONS(1906),
    [anon_sym_module] = ACTIONS(1906),
    [anon_sym_open] = ACTIONS(1906),
    [anon_sym_include] = ACTIONS(1906),
    [anon_sym_class] = ACTIONS(1906),
    [anon_sym_val] = ACTIONS(1906),
    [anon_sym_end] = ACTIONS(1906),
    [anon_sym_object] = ACTIONS(1906),
    [anon_sym_AMP] = ACTIONS(1906),
    [anon_sym_POUND] = ACTIONS(1906),
    [anon_sym_COLON_COLON] = ACTIONS(1908),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1908),
    [anon_sym_else] = ACTIONS(1906),
    [anon_sym_new] = ACTIONS(1906),
    [anon_sym_LBRACE_LT] = ACTIONS(1908),
    [anon_sym_begin] = ACTIONS(1906),
    [sym_ocamlyacc_value] = ACTIONS(1908),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1906),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1908),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1906),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1906),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1908),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1908),
    [aux_sym_number_token1] = ACTIONS(1908),
    [aux_sym_signed_number_token1] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1908),
    [anon_sym_DQUOTE] = ACTIONS(1908),
    [sym_prefix_operator] = ACTIONS(1908),
    [aux_sym_sign_operator_token1] = ACTIONS(1906),
    [sym_hash_operator] = ACTIONS(1908),
    [anon_sym_lsl] = ACTIONS(1906),
    [anon_sym_lsr] = ACTIONS(1906),
    [anon_sym_asr] = ACTIONS(1906),
    [aux_sym_pow_operator_token1] = ACTIONS(1908),
    [anon_sym_mod] = ACTIONS(1906),
    [anon_sym_land] = ACTIONS(1906),
    [anon_sym_lor] = ACTIONS(1906),
    [anon_sym_lxor] = ACTIONS(1906),
    [aux_sym_mult_operator_token1] = ACTIONS(1906),
    [aux_sym_add_operator_token1] = ACTIONS(1906),
    [sym_concat_operator] = ACTIONS(1908),
    [sym_rel_operator] = ACTIONS(1906),
    [anon_sym_AMP_AMP] = ACTIONS(1906),
    [anon_sym_or] = ACTIONS(1906),
    [anon_sym_PIPE_PIPE] = ACTIONS(1906),
    [sym_assign_operator] = ACTIONS(1908),
    [sym_let_operator] = ACTIONS(1908),
    [sym__uppercase_identifier] = ACTIONS(1908),
    [aux_sym_directive_token1] = ACTIONS(1906),
    [aux_sym_tag_token1] = ACTIONS(1908),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1133)] = {
    [sym_attribute] = STATE(1133),
    [sym__lowercase_identifier] = ACTIONS(2025),
    [anon_sym_SEMI_SEMI] = ACTIONS(2027),
    [anon_sym_let] = ACTIONS(2025),
    [anon_sym_and] = ACTIONS(2025),
    [anon_sym_TILDE] = ACTIONS(2025),
    [anon_sym_QMARK] = ACTIONS(2025),
    [anon_sym_LPAREN] = ACTIONS(2027),
    [anon_sym_external] = ACTIONS(2025),
    [anon_sym_type] = ACTIONS(2025),
    [anon_sym_COMMA] = ACTIONS(2027),
    [anon_sym_PIPE] = ACTIONS(2025),
    [anon_sym_DOT] = ACTIONS(2027),
    [anon_sym_LBRACK] = ACTIONS(2025),
    [anon_sym_RBRACK] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(2025),
    [anon_sym_false] = ACTIONS(2025),
    [anon_sym_LBRACE] = ACTIONS(2025),
    [anon_sym_SEMI] = ACTIONS(2025),
    [anon_sym_exception] = ACTIONS(2025),
    [anon_sym_module] = ACTIONS(2025),
    [anon_sym_open] = ACTIONS(2025),
    [anon_sym_include] = ACTIONS(2025),
    [anon_sym_class] = ACTIONS(2025),
    [anon_sym_val] = ACTIONS(2025),
    [anon_sym_end] = ACTIONS(2025),
    [anon_sym_object] = ACTIONS(2025),
    [anon_sym_AMP] = ACTIONS(2025),
    [anon_sym_POUND] = ACTIONS(2025),
    [anon_sym_COLON_COLON] = ACTIONS(2027),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2027),
    [anon_sym_else] = ACTIONS(2025),
    [anon_sym_new] = ACTIONS(2025),
    [anon_sym_LBRACE_LT] = ACTIONS(2027),
    [anon_sym_begin] = ACTIONS(2025),
    [sym_ocamlyacc_value] = ACTIONS(2027),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2025),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2025),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2027),
    [aux_sym_number_token1] = ACTIONS(2027),
    [aux_sym_signed_number_token1] = ACTIONS(2025),
    [anon_sym_SQUOTE] = ACTIONS(2027),
    [anon_sym_DQUOTE] = ACTIONS(2027),
    [sym_prefix_operator] = ACTIONS(2027),
    [aux_sym_sign_operator_token1] = ACTIONS(2025),
    [sym_hash_operator] = ACTIONS(2027),
    [anon_sym_lsl] = ACTIONS(2025),
    [anon_sym_lsr] = ACTIONS(2025),
    [anon_sym_asr] = ACTIONS(2025),
    [aux_sym_pow_operator_token1] = ACTIONS(2027),
    [anon_sym_mod] = ACTIONS(2025),
    [anon_sym_land] = ACTIONS(2025),
    [anon_sym_lor] = ACTIONS(2025),
    [anon_sym_lxor] = ACTIONS(2025),
    [aux_sym_mult_operator_token1] = ACTIONS(2025),
    [aux_sym_add_operator_token1] = ACTIONS(2025),
    [sym_concat_operator] = ACTIONS(2027),
    [sym_rel_operator] = ACTIONS(2025),
    [anon_sym_AMP_AMP] = ACTIONS(2025),
    [anon_sym_or] = ACTIONS(2025),
    [anon_sym_PIPE_PIPE] = ACTIONS(2025),
    [sym_assign_operator] = ACTIONS(2027),
    [sym_let_operator] = ACTIONS(2027),
    [sym__uppercase_identifier] = ACTIONS(2027),
    [aux_sym_directive_token1] = ACTIONS(2025),
    [aux_sym_tag_token1] = ACTIONS(2027),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1134)] = {
    [sym_attribute] = STATE(1134),
    [sym__lowercase_identifier] = ACTIONS(1724),
    [anon_sym_SEMI_SEMI] = ACTIONS(1726),
    [anon_sym_let] = ACTIONS(1724),
    [anon_sym_TILDE] = ACTIONS(1724),
    [anon_sym_QMARK] = ACTIONS(1724),
    [anon_sym_LPAREN] = ACTIONS(1726),
    [anon_sym_external] = ACTIONS(1724),
    [anon_sym_type] = ACTIONS(1724),
    [anon_sym_COMMA] = ACTIONS(1726),
    [anon_sym_PIPE] = ACTIONS(1724),
    [anon_sym_DOT] = ACTIONS(1726),
    [anon_sym_LBRACK] = ACTIONS(1724),
    [anon_sym_RBRACK] = ACTIONS(1726),
    [anon_sym_true] = ACTIONS(1724),
    [anon_sym_false] = ACTIONS(1724),
    [anon_sym_LBRACE] = ACTIONS(1724),
    [anon_sym_SEMI] = ACTIONS(1724),
    [anon_sym_exception] = ACTIONS(1724),
    [anon_sym_module] = ACTIONS(1724),
    [anon_sym_open] = ACTIONS(1724),
    [anon_sym_include] = ACTIONS(1724),
    [anon_sym_class] = ACTIONS(1724),
    [anon_sym_val] = ACTIONS(1724),
    [anon_sym_end] = ACTIONS(1724),
    [anon_sym_object] = ACTIONS(1724),
    [anon_sym_AMP] = ACTIONS(1724),
    [anon_sym_POUND] = ACTIONS(1724),
    [anon_sym_COLON_COLON] = ACTIONS(1726),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1726),
    [anon_sym_LT_DASH] = ACTIONS(2328),
    [anon_sym_else] = ACTIONS(1724),
    [anon_sym_new] = ACTIONS(1724),
    [anon_sym_LBRACE_LT] = ACTIONS(1726),
    [anon_sym_begin] = ACTIONS(1724),
    [sym_ocamlyacc_value] = ACTIONS(1726),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1724),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1726),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1724),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1726),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1726),
    [aux_sym_number_token1] = ACTIONS(1726),
    [aux_sym_signed_number_token1] = ACTIONS(1724),
    [anon_sym_SQUOTE] = ACTIONS(1726),
    [anon_sym_DQUOTE] = ACTIONS(1726),
    [sym_prefix_operator] = ACTIONS(1726),
    [aux_sym_sign_operator_token1] = ACTIONS(1724),
    [sym_hash_operator] = ACTIONS(1726),
    [anon_sym_lsl] = ACTIONS(1724),
    [anon_sym_lsr] = ACTIONS(1724),
    [anon_sym_asr] = ACTIONS(1724),
    [aux_sym_pow_operator_token1] = ACTIONS(1726),
    [anon_sym_mod] = ACTIONS(1724),
    [anon_sym_land] = ACTIONS(1724),
    [anon_sym_lor] = ACTIONS(1724),
    [anon_sym_lxor] = ACTIONS(1724),
    [aux_sym_mult_operator_token1] = ACTIONS(1724),
    [aux_sym_add_operator_token1] = ACTIONS(1724),
    [sym_concat_operator] = ACTIONS(1726),
    [sym_rel_operator] = ACTIONS(1724),
    [anon_sym_AMP_AMP] = ACTIONS(1724),
    [anon_sym_or] = ACTIONS(1724),
    [anon_sym_PIPE_PIPE] = ACTIONS(1724),
    [sym_assign_operator] = ACTIONS(1726),
    [sym_let_operator] = ACTIONS(1726),
    [sym__uppercase_identifier] = ACTIONS(1726),
    [aux_sym_directive_token1] = ACTIONS(1724),
    [aux_sym_tag_token1] = ACTIONS(1726),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1135)] = {
    [sym_attribute] = STATE(1135),
    [sym__lowercase_identifier] = ACTIONS(1934),
    [anon_sym_SEMI_SEMI] = ACTIONS(1936),
    [anon_sym_let] = ACTIONS(1934),
    [anon_sym_and] = ACTIONS(1934),
    [anon_sym_TILDE] = ACTIONS(1934),
    [anon_sym_QMARK] = ACTIONS(1934),
    [anon_sym_LPAREN] = ACTIONS(1936),
    [anon_sym_external] = ACTIONS(1934),
    [anon_sym_type] = ACTIONS(1934),
    [anon_sym_COMMA] = ACTIONS(1936),
    [anon_sym_PIPE] = ACTIONS(1934),
    [anon_sym_DOT] = ACTIONS(1936),
    [anon_sym_LBRACK] = ACTIONS(1934),
    [anon_sym_RBRACK] = ACTIONS(1936),
    [anon_sym_true] = ACTIONS(1934),
    [anon_sym_false] = ACTIONS(1934),
    [anon_sym_LBRACE] = ACTIONS(1934),
    [anon_sym_SEMI] = ACTIONS(1934),
    [anon_sym_exception] = ACTIONS(1934),
    [anon_sym_module] = ACTIONS(1934),
    [anon_sym_open] = ACTIONS(1934),
    [anon_sym_include] = ACTIONS(1934),
    [anon_sym_class] = ACTIONS(1934),
    [anon_sym_val] = ACTIONS(1934),
    [anon_sym_end] = ACTIONS(1934),
    [anon_sym_object] = ACTIONS(1934),
    [anon_sym_AMP] = ACTIONS(1934),
    [anon_sym_POUND] = ACTIONS(1934),
    [anon_sym_COLON_COLON] = ACTIONS(1936),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1936),
    [anon_sym_else] = ACTIONS(1934),
    [anon_sym_new] = ACTIONS(1934),
    [anon_sym_LBRACE_LT] = ACTIONS(1936),
    [anon_sym_begin] = ACTIONS(1934),
    [sym_ocamlyacc_value] = ACTIONS(1936),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1934),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1936),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1934),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1934),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1936),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1936),
    [aux_sym_number_token1] = ACTIONS(1936),
    [aux_sym_signed_number_token1] = ACTIONS(1934),
    [anon_sym_SQUOTE] = ACTIONS(1936),
    [anon_sym_DQUOTE] = ACTIONS(1936),
    [sym_prefix_operator] = ACTIONS(1936),
    [aux_sym_sign_operator_token1] = ACTIONS(1934),
    [sym_hash_operator] = ACTIONS(1936),
    [anon_sym_lsl] = ACTIONS(1934),
    [anon_sym_lsr] = ACTIONS(1934),
    [anon_sym_asr] = ACTIONS(1934),
    [aux_sym_pow_operator_token1] = ACTIONS(1936),
    [anon_sym_mod] = ACTIONS(1934),
    [anon_sym_land] = ACTIONS(1934),
    [anon_sym_lor] = ACTIONS(1934),
    [anon_sym_lxor] = ACTIONS(1934),
    [aux_sym_mult_operator_token1] = ACTIONS(1934),
    [aux_sym_add_operator_token1] = ACTIONS(1934),
    [sym_concat_operator] = ACTIONS(1936),
    [sym_rel_operator] = ACTIONS(1934),
    [anon_sym_AMP_AMP] = ACTIONS(1934),
    [anon_sym_or] = ACTIONS(1934),
    [anon_sym_PIPE_PIPE] = ACTIONS(1934),
    [sym_assign_operator] = ACTIONS(1936),
    [sym_let_operator] = ACTIONS(1936),
    [sym__uppercase_identifier] = ACTIONS(1936),
    [aux_sym_directive_token1] = ACTIONS(1934),
    [aux_sym_tag_token1] = ACTIONS(1936),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1136)] = {
    [sym_attribute] = STATE(1136),
    [sym__lowercase_identifier] = ACTIONS(1758),
    [anon_sym_SEMI_SEMI] = ACTIONS(1760),
    [anon_sym_let] = ACTIONS(1758),
    [anon_sym_and] = ACTIONS(1758),
    [anon_sym_TILDE] = ACTIONS(1758),
    [anon_sym_QMARK] = ACTIONS(1758),
    [anon_sym_LPAREN] = ACTIONS(1760),
    [anon_sym_external] = ACTIONS(1758),
    [anon_sym_type] = ACTIONS(1758),
    [anon_sym_COMMA] = ACTIONS(1760),
    [anon_sym_PIPE] = ACTIONS(1758),
    [anon_sym_DOT] = ACTIONS(1760),
    [anon_sym_LBRACK] = ACTIONS(1758),
    [anon_sym_RBRACK] = ACTIONS(1760),
    [anon_sym_true] = ACTIONS(1758),
    [anon_sym_false] = ACTIONS(1758),
    [anon_sym_LBRACE] = ACTIONS(1758),
    [anon_sym_SEMI] = ACTIONS(1758),
    [anon_sym_exception] = ACTIONS(1758),
    [anon_sym_module] = ACTIONS(1758),
    [anon_sym_open] = ACTIONS(1758),
    [anon_sym_include] = ACTIONS(1758),
    [anon_sym_class] = ACTIONS(1758),
    [anon_sym_val] = ACTIONS(1758),
    [anon_sym_end] = ACTIONS(1758),
    [anon_sym_object] = ACTIONS(1758),
    [anon_sym_AMP] = ACTIONS(1758),
    [anon_sym_POUND] = ACTIONS(1758),
    [anon_sym_COLON_COLON] = ACTIONS(1760),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1760),
    [anon_sym_else] = ACTIONS(1758),
    [anon_sym_new] = ACTIONS(1758),
    [anon_sym_LBRACE_LT] = ACTIONS(1760),
    [anon_sym_begin] = ACTIONS(1758),
    [sym_ocamlyacc_value] = ACTIONS(1760),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1758),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1760),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1758),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1758),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1760),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1760),
    [aux_sym_number_token1] = ACTIONS(1760),
    [aux_sym_signed_number_token1] = ACTIONS(1758),
    [anon_sym_SQUOTE] = ACTIONS(1760),
    [anon_sym_DQUOTE] = ACTIONS(1760),
    [sym_prefix_operator] = ACTIONS(1760),
    [aux_sym_sign_operator_token1] = ACTIONS(1758),
    [sym_hash_operator] = ACTIONS(1760),
    [anon_sym_lsl] = ACTIONS(1758),
    [anon_sym_lsr] = ACTIONS(1758),
    [anon_sym_asr] = ACTIONS(1758),
    [aux_sym_pow_operator_token1] = ACTIONS(1760),
    [anon_sym_mod] = ACTIONS(1758),
    [anon_sym_land] = ACTIONS(1758),
    [anon_sym_lor] = ACTIONS(1758),
    [anon_sym_lxor] = ACTIONS(1758),
    [aux_sym_mult_operator_token1] = ACTIONS(1758),
    [aux_sym_add_operator_token1] = ACTIONS(1758),
    [sym_concat_operator] = ACTIONS(1760),
    [sym_rel_operator] = ACTIONS(1758),
    [anon_sym_AMP_AMP] = ACTIONS(1758),
    [anon_sym_or] = ACTIONS(1758),
    [anon_sym_PIPE_PIPE] = ACTIONS(1758),
    [sym_assign_operator] = ACTIONS(1760),
    [sym_let_operator] = ACTIONS(1760),
    [sym__uppercase_identifier] = ACTIONS(1760),
    [aux_sym_directive_token1] = ACTIONS(1758),
    [aux_sym_tag_token1] = ACTIONS(1760),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1137)] = {
    [sym_attribute] = STATE(1137),
    [sym__lowercase_identifier] = ACTIONS(1846),
    [anon_sym_SEMI_SEMI] = ACTIONS(1848),
    [anon_sym_let] = ACTIONS(1846),
    [anon_sym_and] = ACTIONS(1846),
    [anon_sym_TILDE] = ACTIONS(1846),
    [anon_sym_QMARK] = ACTIONS(1846),
    [anon_sym_LPAREN] = ACTIONS(1848),
    [anon_sym_external] = ACTIONS(1846),
    [anon_sym_type] = ACTIONS(1846),
    [anon_sym_COMMA] = ACTIONS(1848),
    [anon_sym_PIPE] = ACTIONS(1846),
    [anon_sym_DOT] = ACTIONS(1848),
    [anon_sym_LBRACK] = ACTIONS(1846),
    [anon_sym_RBRACK] = ACTIONS(1848),
    [anon_sym_true] = ACTIONS(1846),
    [anon_sym_false] = ACTIONS(1846),
    [anon_sym_LBRACE] = ACTIONS(1846),
    [anon_sym_SEMI] = ACTIONS(1846),
    [anon_sym_exception] = ACTIONS(1846),
    [anon_sym_module] = ACTIONS(1846),
    [anon_sym_open] = ACTIONS(1846),
    [anon_sym_include] = ACTIONS(1846),
    [anon_sym_class] = ACTIONS(1846),
    [anon_sym_val] = ACTIONS(1846),
    [anon_sym_end] = ACTIONS(1846),
    [anon_sym_object] = ACTIONS(1846),
    [anon_sym_AMP] = ACTIONS(1846),
    [anon_sym_POUND] = ACTIONS(1846),
    [anon_sym_COLON_COLON] = ACTIONS(1848),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1848),
    [anon_sym_else] = ACTIONS(1846),
    [anon_sym_new] = ACTIONS(1846),
    [anon_sym_LBRACE_LT] = ACTIONS(1848),
    [anon_sym_begin] = ACTIONS(1846),
    [sym_ocamlyacc_value] = ACTIONS(1848),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1846),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1848),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1846),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1846),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1848),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1848),
    [aux_sym_number_token1] = ACTIONS(1848),
    [aux_sym_signed_number_token1] = ACTIONS(1846),
    [anon_sym_SQUOTE] = ACTIONS(1848),
    [anon_sym_DQUOTE] = ACTIONS(1848),
    [sym_prefix_operator] = ACTIONS(1848),
    [aux_sym_sign_operator_token1] = ACTIONS(1846),
    [sym_hash_operator] = ACTIONS(1848),
    [anon_sym_lsl] = ACTIONS(1846),
    [anon_sym_lsr] = ACTIONS(1846),
    [anon_sym_asr] = ACTIONS(1846),
    [aux_sym_pow_operator_token1] = ACTIONS(1848),
    [anon_sym_mod] = ACTIONS(1846),
    [anon_sym_land] = ACTIONS(1846),
    [anon_sym_lor] = ACTIONS(1846),
    [anon_sym_lxor] = ACTIONS(1846),
    [aux_sym_mult_operator_token1] = ACTIONS(1846),
    [aux_sym_add_operator_token1] = ACTIONS(1846),
    [sym_concat_operator] = ACTIONS(1848),
    [sym_rel_operator] = ACTIONS(1846),
    [anon_sym_AMP_AMP] = ACTIONS(1846),
    [anon_sym_or] = ACTIONS(1846),
    [anon_sym_PIPE_PIPE] = ACTIONS(1846),
    [sym_assign_operator] = ACTIONS(1848),
    [sym_let_operator] = ACTIONS(1848),
    [sym__uppercase_identifier] = ACTIONS(1848),
    [aux_sym_directive_token1] = ACTIONS(1846),
    [aux_sym_tag_token1] = ACTIONS(1848),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1138)] = {
    [sym_attribute] = STATE(1138),
    [sym__lowercase_identifier] = ACTIONS(1942),
    [anon_sym_SEMI_SEMI] = ACTIONS(1944),
    [anon_sym_let] = ACTIONS(1942),
    [anon_sym_and] = ACTIONS(1942),
    [anon_sym_TILDE] = ACTIONS(1942),
    [anon_sym_QMARK] = ACTIONS(1942),
    [anon_sym_LPAREN] = ACTIONS(1944),
    [anon_sym_external] = ACTIONS(1942),
    [anon_sym_type] = ACTIONS(1942),
    [anon_sym_COMMA] = ACTIONS(1944),
    [anon_sym_PIPE] = ACTIONS(1942),
    [anon_sym_DOT] = ACTIONS(1944),
    [anon_sym_LBRACK] = ACTIONS(1942),
    [anon_sym_RBRACK] = ACTIONS(1944),
    [anon_sym_true] = ACTIONS(1942),
    [anon_sym_false] = ACTIONS(1942),
    [anon_sym_LBRACE] = ACTIONS(1942),
    [anon_sym_SEMI] = ACTIONS(1942),
    [anon_sym_exception] = ACTIONS(1942),
    [anon_sym_module] = ACTIONS(1942),
    [anon_sym_open] = ACTIONS(1942),
    [anon_sym_include] = ACTIONS(1942),
    [anon_sym_class] = ACTIONS(1942),
    [anon_sym_val] = ACTIONS(1942),
    [anon_sym_end] = ACTIONS(1942),
    [anon_sym_object] = ACTIONS(1942),
    [anon_sym_AMP] = ACTIONS(1942),
    [anon_sym_POUND] = ACTIONS(1942),
    [anon_sym_COLON_COLON] = ACTIONS(1944),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1944),
    [anon_sym_else] = ACTIONS(1942),
    [anon_sym_new] = ACTIONS(1942),
    [anon_sym_LBRACE_LT] = ACTIONS(1944),
    [anon_sym_begin] = ACTIONS(1942),
    [sym_ocamlyacc_value] = ACTIONS(1944),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1942),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1944),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1942),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1942),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1944),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1944),
    [aux_sym_number_token1] = ACTIONS(1944),
    [aux_sym_signed_number_token1] = ACTIONS(1942),
    [anon_sym_SQUOTE] = ACTIONS(1944),
    [anon_sym_DQUOTE] = ACTIONS(1944),
    [sym_prefix_operator] = ACTIONS(1944),
    [aux_sym_sign_operator_token1] = ACTIONS(1942),
    [sym_hash_operator] = ACTIONS(1944),
    [anon_sym_lsl] = ACTIONS(1942),
    [anon_sym_lsr] = ACTIONS(1942),
    [anon_sym_asr] = ACTIONS(1942),
    [aux_sym_pow_operator_token1] = ACTIONS(1944),
    [anon_sym_mod] = ACTIONS(1942),
    [anon_sym_land] = ACTIONS(1942),
    [anon_sym_lor] = ACTIONS(1942),
    [anon_sym_lxor] = ACTIONS(1942),
    [aux_sym_mult_operator_token1] = ACTIONS(1942),
    [aux_sym_add_operator_token1] = ACTIONS(1942),
    [sym_concat_operator] = ACTIONS(1944),
    [sym_rel_operator] = ACTIONS(1942),
    [anon_sym_AMP_AMP] = ACTIONS(1942),
    [anon_sym_or] = ACTIONS(1942),
    [anon_sym_PIPE_PIPE] = ACTIONS(1942),
    [sym_assign_operator] = ACTIONS(1944),
    [sym_let_operator] = ACTIONS(1944),
    [sym__uppercase_identifier] = ACTIONS(1944),
    [aux_sym_directive_token1] = ACTIONS(1942),
    [aux_sym_tag_token1] = ACTIONS(1944),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1139)] = {
    [sym_attribute] = STATE(1139),
    [sym__lowercase_identifier] = ACTIONS(1834),
    [anon_sym_SEMI_SEMI] = ACTIONS(1836),
    [anon_sym_let] = ACTIONS(1834),
    [anon_sym_and] = ACTIONS(1834),
    [anon_sym_TILDE] = ACTIONS(1834),
    [anon_sym_QMARK] = ACTIONS(1834),
    [anon_sym_LPAREN] = ACTIONS(1836),
    [anon_sym_external] = ACTIONS(1834),
    [anon_sym_type] = ACTIONS(1834),
    [anon_sym_COMMA] = ACTIONS(1836),
    [anon_sym_PIPE] = ACTIONS(1834),
    [anon_sym_DOT] = ACTIONS(1836),
    [anon_sym_LBRACK] = ACTIONS(1834),
    [anon_sym_RBRACK] = ACTIONS(1836),
    [anon_sym_true] = ACTIONS(1834),
    [anon_sym_false] = ACTIONS(1834),
    [anon_sym_LBRACE] = ACTIONS(1834),
    [anon_sym_SEMI] = ACTIONS(1834),
    [anon_sym_exception] = ACTIONS(1834),
    [anon_sym_module] = ACTIONS(1834),
    [anon_sym_open] = ACTIONS(1834),
    [anon_sym_include] = ACTIONS(1834),
    [anon_sym_class] = ACTIONS(1834),
    [anon_sym_val] = ACTIONS(1834),
    [anon_sym_end] = ACTIONS(1834),
    [anon_sym_object] = ACTIONS(1834),
    [anon_sym_AMP] = ACTIONS(1834),
    [anon_sym_POUND] = ACTIONS(1834),
    [anon_sym_COLON_COLON] = ACTIONS(1836),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1836),
    [anon_sym_else] = ACTIONS(1834),
    [anon_sym_new] = ACTIONS(1834),
    [anon_sym_LBRACE_LT] = ACTIONS(1836),
    [anon_sym_begin] = ACTIONS(1834),
    [sym_ocamlyacc_value] = ACTIONS(1836),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1834),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1836),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1834),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1834),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1836),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1836),
    [aux_sym_number_token1] = ACTIONS(1836),
    [aux_sym_signed_number_token1] = ACTIONS(1834),
    [anon_sym_SQUOTE] = ACTIONS(1836),
    [anon_sym_DQUOTE] = ACTIONS(1836),
    [sym_prefix_operator] = ACTIONS(1836),
    [aux_sym_sign_operator_token1] = ACTIONS(1834),
    [sym_hash_operator] = ACTIONS(1836),
    [anon_sym_lsl] = ACTIONS(1834),
    [anon_sym_lsr] = ACTIONS(1834),
    [anon_sym_asr] = ACTIONS(1834),
    [aux_sym_pow_operator_token1] = ACTIONS(1836),
    [anon_sym_mod] = ACTIONS(1834),
    [anon_sym_land] = ACTIONS(1834),
    [anon_sym_lor] = ACTIONS(1834),
    [anon_sym_lxor] = ACTIONS(1834),
    [aux_sym_mult_operator_token1] = ACTIONS(1834),
    [aux_sym_add_operator_token1] = ACTIONS(1834),
    [sym_concat_operator] = ACTIONS(1836),
    [sym_rel_operator] = ACTIONS(1834),
    [anon_sym_AMP_AMP] = ACTIONS(1834),
    [anon_sym_or] = ACTIONS(1834),
    [anon_sym_PIPE_PIPE] = ACTIONS(1834),
    [sym_assign_operator] = ACTIONS(1836),
    [sym_let_operator] = ACTIONS(1836),
    [sym__uppercase_identifier] = ACTIONS(1836),
    [aux_sym_directive_token1] = ACTIONS(1834),
    [aux_sym_tag_token1] = ACTIONS(1836),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1140)] = {
    [sym_attribute] = STATE(1140),
    [sym__lowercase_identifier] = ACTIONS(1938),
    [anon_sym_SEMI_SEMI] = ACTIONS(1940),
    [anon_sym_let] = ACTIONS(1938),
    [anon_sym_and] = ACTIONS(1938),
    [anon_sym_TILDE] = ACTIONS(1938),
    [anon_sym_QMARK] = ACTIONS(1938),
    [anon_sym_LPAREN] = ACTIONS(1940),
    [anon_sym_external] = ACTIONS(1938),
    [anon_sym_type] = ACTIONS(1938),
    [anon_sym_COMMA] = ACTIONS(1940),
    [anon_sym_PIPE] = ACTIONS(1938),
    [anon_sym_DOT] = ACTIONS(1940),
    [anon_sym_LBRACK] = ACTIONS(1938),
    [anon_sym_RBRACK] = ACTIONS(1940),
    [anon_sym_true] = ACTIONS(1938),
    [anon_sym_false] = ACTIONS(1938),
    [anon_sym_LBRACE] = ACTIONS(1938),
    [anon_sym_SEMI] = ACTIONS(1938),
    [anon_sym_exception] = ACTIONS(1938),
    [anon_sym_module] = ACTIONS(1938),
    [anon_sym_open] = ACTIONS(1938),
    [anon_sym_include] = ACTIONS(1938),
    [anon_sym_class] = ACTIONS(1938),
    [anon_sym_val] = ACTIONS(1938),
    [anon_sym_end] = ACTIONS(1938),
    [anon_sym_object] = ACTIONS(1938),
    [anon_sym_AMP] = ACTIONS(1938),
    [anon_sym_POUND] = ACTIONS(1938),
    [anon_sym_COLON_COLON] = ACTIONS(1940),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1940),
    [anon_sym_else] = ACTIONS(1938),
    [anon_sym_new] = ACTIONS(1938),
    [anon_sym_LBRACE_LT] = ACTIONS(1940),
    [anon_sym_begin] = ACTIONS(1938),
    [sym_ocamlyacc_value] = ACTIONS(1940),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1938),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1940),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1938),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1938),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1940),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1940),
    [aux_sym_number_token1] = ACTIONS(1940),
    [aux_sym_signed_number_token1] = ACTIONS(1938),
    [anon_sym_SQUOTE] = ACTIONS(1940),
    [anon_sym_DQUOTE] = ACTIONS(1940),
    [sym_prefix_operator] = ACTIONS(1940),
    [aux_sym_sign_operator_token1] = ACTIONS(1938),
    [sym_hash_operator] = ACTIONS(1940),
    [anon_sym_lsl] = ACTIONS(1938),
    [anon_sym_lsr] = ACTIONS(1938),
    [anon_sym_asr] = ACTIONS(1938),
    [aux_sym_pow_operator_token1] = ACTIONS(1940),
    [anon_sym_mod] = ACTIONS(1938),
    [anon_sym_land] = ACTIONS(1938),
    [anon_sym_lor] = ACTIONS(1938),
    [anon_sym_lxor] = ACTIONS(1938),
    [aux_sym_mult_operator_token1] = ACTIONS(1938),
    [aux_sym_add_operator_token1] = ACTIONS(1938),
    [sym_concat_operator] = ACTIONS(1940),
    [sym_rel_operator] = ACTIONS(1938),
    [anon_sym_AMP_AMP] = ACTIONS(1938),
    [anon_sym_or] = ACTIONS(1938),
    [anon_sym_PIPE_PIPE] = ACTIONS(1938),
    [sym_assign_operator] = ACTIONS(1940),
    [sym_let_operator] = ACTIONS(1940),
    [sym__uppercase_identifier] = ACTIONS(1940),
    [aux_sym_directive_token1] = ACTIONS(1938),
    [aux_sym_tag_token1] = ACTIONS(1940),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1141)] = {
    [sym_attribute] = STATE(1141),
    [sym__lowercase_identifier] = ACTIONS(1614),
    [anon_sym_SEMI_SEMI] = ACTIONS(1616),
    [anon_sym_let] = ACTIONS(1614),
    [anon_sym_and] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_external] = ACTIONS(1614),
    [anon_sym_type] = ACTIONS(1614),
    [anon_sym_COMMA] = ACTIONS(1616),
    [anon_sym_PIPE] = ACTIONS(1614),
    [anon_sym_DOT] = ACTIONS(1616),
    [anon_sym_LBRACK] = ACTIONS(1614),
    [anon_sym_RBRACK] = ACTIONS(1616),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1614),
    [anon_sym_exception] = ACTIONS(1614),
    [anon_sym_module] = ACTIONS(1614),
    [anon_sym_open] = ACTIONS(1614),
    [anon_sym_include] = ACTIONS(1614),
    [anon_sym_class] = ACTIONS(1614),
    [anon_sym_val] = ACTIONS(1614),
    [anon_sym_end] = ACTIONS(1614),
    [anon_sym_object] = ACTIONS(1614),
    [anon_sym_AMP] = ACTIONS(1614),
    [anon_sym_POUND] = ACTIONS(1614),
    [anon_sym_COLON_COLON] = ACTIONS(1616),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1616),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
    [anon_sym_LBRACE_LT] = ACTIONS(1616),
    [anon_sym_begin] = ACTIONS(1614),
    [sym_ocamlyacc_value] = ACTIONS(1616),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1614),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1616),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1614),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1616),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1616),
    [aux_sym_number_token1] = ACTIONS(1616),
    [aux_sym_signed_number_token1] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1616),
    [sym_prefix_operator] = ACTIONS(1616),
    [aux_sym_sign_operator_token1] = ACTIONS(1614),
    [sym_hash_operator] = ACTIONS(1616),
    [anon_sym_lsl] = ACTIONS(1614),
    [anon_sym_lsr] = ACTIONS(1614),
    [anon_sym_asr] = ACTIONS(1614),
    [aux_sym_pow_operator_token1] = ACTIONS(1616),
    [anon_sym_mod] = ACTIONS(1614),
    [anon_sym_land] = ACTIONS(1614),
    [anon_sym_lor] = ACTIONS(1614),
    [anon_sym_lxor] = ACTIONS(1614),
    [aux_sym_mult_operator_token1] = ACTIONS(1614),
    [aux_sym_add_operator_token1] = ACTIONS(1614),
    [sym_concat_operator] = ACTIONS(1616),
    [sym_rel_operator] = ACTIONS(1614),
    [anon_sym_AMP_AMP] = ACTIONS(1614),
    [anon_sym_or] = ACTIONS(1614),
    [anon_sym_PIPE_PIPE] = ACTIONS(1614),
    [sym_assign_operator] = ACTIONS(1616),
    [sym_let_operator] = ACTIONS(1616),
    [sym__uppercase_identifier] = ACTIONS(1616),
    [aux_sym_directive_token1] = ACTIONS(1614),
    [aux_sym_tag_token1] = ACTIONS(1616),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1142)] = {
    [sym_attribute] = STATE(1142),
    [sym__lowercase_identifier] = ACTIONS(1886),
    [anon_sym_SEMI_SEMI] = ACTIONS(1888),
    [anon_sym_let] = ACTIONS(1886),
    [anon_sym_and] = ACTIONS(1886),
    [anon_sym_TILDE] = ACTIONS(1886),
    [anon_sym_QMARK] = ACTIONS(1886),
    [anon_sym_LPAREN] = ACTIONS(1888),
    [anon_sym_external] = ACTIONS(1886),
    [anon_sym_type] = ACTIONS(1886),
    [anon_sym_COMMA] = ACTIONS(1888),
    [anon_sym_PIPE] = ACTIONS(1886),
    [anon_sym_DOT] = ACTIONS(1888),
    [anon_sym_LBRACK] = ACTIONS(1886),
    [anon_sym_RBRACK] = ACTIONS(1888),
    [anon_sym_true] = ACTIONS(1886),
    [anon_sym_false] = ACTIONS(1886),
    [anon_sym_LBRACE] = ACTIONS(1886),
    [anon_sym_SEMI] = ACTIONS(1886),
    [anon_sym_exception] = ACTIONS(1886),
    [anon_sym_module] = ACTIONS(1886),
    [anon_sym_open] = ACTIONS(1886),
    [anon_sym_include] = ACTIONS(1886),
    [anon_sym_class] = ACTIONS(1886),
    [anon_sym_val] = ACTIONS(1886),
    [anon_sym_end] = ACTIONS(1886),
    [anon_sym_object] = ACTIONS(1886),
    [anon_sym_AMP] = ACTIONS(1886),
    [anon_sym_POUND] = ACTIONS(1886),
    [anon_sym_COLON_COLON] = ACTIONS(1888),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1888),
    [anon_sym_else] = ACTIONS(1886),
    [anon_sym_new] = ACTIONS(1886),
    [anon_sym_LBRACE_LT] = ACTIONS(1888),
    [anon_sym_begin] = ACTIONS(1886),
    [sym_ocamlyacc_value] = ACTIONS(1888),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1886),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1888),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1886),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1886),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1888),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1888),
    [aux_sym_number_token1] = ACTIONS(1888),
    [aux_sym_signed_number_token1] = ACTIONS(1886),
    [anon_sym_SQUOTE] = ACTIONS(1888),
    [anon_sym_DQUOTE] = ACTIONS(1888),
    [sym_prefix_operator] = ACTIONS(1888),
    [aux_sym_sign_operator_token1] = ACTIONS(1886),
    [sym_hash_operator] = ACTIONS(1888),
    [anon_sym_lsl] = ACTIONS(1886),
    [anon_sym_lsr] = ACTIONS(1886),
    [anon_sym_asr] = ACTIONS(1886),
    [aux_sym_pow_operator_token1] = ACTIONS(1888),
    [anon_sym_mod] = ACTIONS(1886),
    [anon_sym_land] = ACTIONS(1886),
    [anon_sym_lor] = ACTIONS(1886),
    [anon_sym_lxor] = ACTIONS(1886),
    [aux_sym_mult_operator_token1] = ACTIONS(1886),
    [aux_sym_add_operator_token1] = ACTIONS(1886),
    [sym_concat_operator] = ACTIONS(1888),
    [sym_rel_operator] = ACTIONS(1886),
    [anon_sym_AMP_AMP] = ACTIONS(1886),
    [anon_sym_or] = ACTIONS(1886),
    [anon_sym_PIPE_PIPE] = ACTIONS(1886),
    [sym_assign_operator] = ACTIONS(1888),
    [sym_let_operator] = ACTIONS(1888),
    [sym__uppercase_identifier] = ACTIONS(1888),
    [aux_sym_directive_token1] = ACTIONS(1886),
    [aux_sym_tag_token1] = ACTIONS(1888),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1143)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4359),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1143),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2139),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(2139),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_SEMI] = ACTIONS(2139),
    [anon_sym_RBRACE] = ACTIONS(2139),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_as] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1144)] = {
    [sym_attribute] = STATE(1144),
    [sym__lowercase_identifier] = ACTIONS(1914),
    [anon_sym_SEMI_SEMI] = ACTIONS(1916),
    [anon_sym_let] = ACTIONS(1914),
    [anon_sym_and] = ACTIONS(1914),
    [anon_sym_TILDE] = ACTIONS(1914),
    [anon_sym_QMARK] = ACTIONS(1914),
    [anon_sym_LPAREN] = ACTIONS(1916),
    [anon_sym_external] = ACTIONS(1914),
    [anon_sym_type] = ACTIONS(1914),
    [anon_sym_COMMA] = ACTIONS(1916),
    [anon_sym_PIPE] = ACTIONS(1914),
    [anon_sym_DOT] = ACTIONS(1916),
    [anon_sym_LBRACK] = ACTIONS(1914),
    [anon_sym_RBRACK] = ACTIONS(1916),
    [anon_sym_true] = ACTIONS(1914),
    [anon_sym_false] = ACTIONS(1914),
    [anon_sym_LBRACE] = ACTIONS(1914),
    [anon_sym_SEMI] = ACTIONS(1914),
    [anon_sym_exception] = ACTIONS(1914),
    [anon_sym_module] = ACTIONS(1914),
    [anon_sym_open] = ACTIONS(1914),
    [anon_sym_include] = ACTIONS(1914),
    [anon_sym_class] = ACTIONS(1914),
    [anon_sym_val] = ACTIONS(1914),
    [anon_sym_end] = ACTIONS(1914),
    [anon_sym_object] = ACTIONS(1914),
    [anon_sym_AMP] = ACTIONS(1914),
    [anon_sym_POUND] = ACTIONS(1914),
    [anon_sym_COLON_COLON] = ACTIONS(1916),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1916),
    [anon_sym_else] = ACTIONS(1914),
    [anon_sym_new] = ACTIONS(1914),
    [anon_sym_LBRACE_LT] = ACTIONS(1916),
    [anon_sym_begin] = ACTIONS(1914),
    [sym_ocamlyacc_value] = ACTIONS(1916),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1914),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1916),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1914),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1914),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1916),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1916),
    [aux_sym_number_token1] = ACTIONS(1916),
    [aux_sym_signed_number_token1] = ACTIONS(1914),
    [anon_sym_SQUOTE] = ACTIONS(1916),
    [anon_sym_DQUOTE] = ACTIONS(1916),
    [sym_prefix_operator] = ACTIONS(1916),
    [aux_sym_sign_operator_token1] = ACTIONS(1914),
    [sym_hash_operator] = ACTIONS(1916),
    [anon_sym_lsl] = ACTIONS(1914),
    [anon_sym_lsr] = ACTIONS(1914),
    [anon_sym_asr] = ACTIONS(1914),
    [aux_sym_pow_operator_token1] = ACTIONS(1916),
    [anon_sym_mod] = ACTIONS(1914),
    [anon_sym_land] = ACTIONS(1914),
    [anon_sym_lor] = ACTIONS(1914),
    [anon_sym_lxor] = ACTIONS(1914),
    [aux_sym_mult_operator_token1] = ACTIONS(1914),
    [aux_sym_add_operator_token1] = ACTIONS(1914),
    [sym_concat_operator] = ACTIONS(1916),
    [sym_rel_operator] = ACTIONS(1914),
    [anon_sym_AMP_AMP] = ACTIONS(1914),
    [anon_sym_or] = ACTIONS(1914),
    [anon_sym_PIPE_PIPE] = ACTIONS(1914),
    [sym_assign_operator] = ACTIONS(1916),
    [sym_let_operator] = ACTIONS(1916),
    [sym__uppercase_identifier] = ACTIONS(1916),
    [aux_sym_directive_token1] = ACTIONS(1914),
    [aux_sym_tag_token1] = ACTIONS(1916),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1145)] = {
    [sym_attribute] = STATE(1145),
    [sym__lowercase_identifier] = ACTIONS(1842),
    [anon_sym_SEMI_SEMI] = ACTIONS(1844),
    [anon_sym_let] = ACTIONS(1842),
    [anon_sym_and] = ACTIONS(1842),
    [anon_sym_TILDE] = ACTIONS(1842),
    [anon_sym_QMARK] = ACTIONS(1842),
    [anon_sym_LPAREN] = ACTIONS(1844),
    [anon_sym_external] = ACTIONS(1842),
    [anon_sym_type] = ACTIONS(1842),
    [anon_sym_COMMA] = ACTIONS(1844),
    [anon_sym_PIPE] = ACTIONS(1842),
    [anon_sym_DOT] = ACTIONS(1844),
    [anon_sym_LBRACK] = ACTIONS(1842),
    [anon_sym_RBRACK] = ACTIONS(1844),
    [anon_sym_true] = ACTIONS(1842),
    [anon_sym_false] = ACTIONS(1842),
    [anon_sym_LBRACE] = ACTIONS(1842),
    [anon_sym_SEMI] = ACTIONS(1842),
    [anon_sym_exception] = ACTIONS(1842),
    [anon_sym_module] = ACTIONS(1842),
    [anon_sym_open] = ACTIONS(1842),
    [anon_sym_include] = ACTIONS(1842),
    [anon_sym_class] = ACTIONS(1842),
    [anon_sym_val] = ACTIONS(1842),
    [anon_sym_end] = ACTIONS(1842),
    [anon_sym_object] = ACTIONS(1842),
    [anon_sym_AMP] = ACTIONS(1842),
    [anon_sym_POUND] = ACTIONS(1842),
    [anon_sym_COLON_COLON] = ACTIONS(1844),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1844),
    [anon_sym_else] = ACTIONS(1842),
    [anon_sym_new] = ACTIONS(1842),
    [anon_sym_LBRACE_LT] = ACTIONS(1844),
    [anon_sym_begin] = ACTIONS(1842),
    [sym_ocamlyacc_value] = ACTIONS(1844),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1842),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1844),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1842),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1842),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1844),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1844),
    [aux_sym_number_token1] = ACTIONS(1844),
    [aux_sym_signed_number_token1] = ACTIONS(1842),
    [anon_sym_SQUOTE] = ACTIONS(1844),
    [anon_sym_DQUOTE] = ACTIONS(1844),
    [sym_prefix_operator] = ACTIONS(1844),
    [aux_sym_sign_operator_token1] = ACTIONS(1842),
    [sym_hash_operator] = ACTIONS(1844),
    [anon_sym_lsl] = ACTIONS(1842),
    [anon_sym_lsr] = ACTIONS(1842),
    [anon_sym_asr] = ACTIONS(1842),
    [aux_sym_pow_operator_token1] = ACTIONS(1844),
    [anon_sym_mod] = ACTIONS(1842),
    [anon_sym_land] = ACTIONS(1842),
    [anon_sym_lor] = ACTIONS(1842),
    [anon_sym_lxor] = ACTIONS(1842),
    [aux_sym_mult_operator_token1] = ACTIONS(1842),
    [aux_sym_add_operator_token1] = ACTIONS(1842),
    [sym_concat_operator] = ACTIONS(1844),
    [sym_rel_operator] = ACTIONS(1842),
    [anon_sym_AMP_AMP] = ACTIONS(1842),
    [anon_sym_or] = ACTIONS(1842),
    [anon_sym_PIPE_PIPE] = ACTIONS(1842),
    [sym_assign_operator] = ACTIONS(1844),
    [sym_let_operator] = ACTIONS(1844),
    [sym__uppercase_identifier] = ACTIONS(1844),
    [aux_sym_directive_token1] = ACTIONS(1842),
    [aux_sym_tag_token1] = ACTIONS(1844),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1146)] = {
    [sym_attribute] = STATE(1146),
    [sym__lowercase_identifier] = ACTIONS(1626),
    [anon_sym_SEMI_SEMI] = ACTIONS(1628),
    [anon_sym_let] = ACTIONS(1626),
    [anon_sym_and] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1628),
    [anon_sym_external] = ACTIONS(1626),
    [anon_sym_type] = ACTIONS(1626),
    [anon_sym_COMMA] = ACTIONS(1628),
    [anon_sym_PIPE] = ACTIONS(1626),
    [anon_sym_DOT] = ACTIONS(1628),
    [anon_sym_LBRACK] = ACTIONS(1626),
    [anon_sym_RBRACK] = ACTIONS(1628),
    [anon_sym_true] = ACTIONS(1626),
    [anon_sym_false] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1626),
    [anon_sym_exception] = ACTIONS(1626),
    [anon_sym_module] = ACTIONS(1626),
    [anon_sym_open] = ACTIONS(1626),
    [anon_sym_include] = ACTIONS(1626),
    [anon_sym_class] = ACTIONS(1626),
    [anon_sym_val] = ACTIONS(1626),
    [anon_sym_end] = ACTIONS(1626),
    [anon_sym_object] = ACTIONS(1626),
    [anon_sym_AMP] = ACTIONS(1626),
    [anon_sym_POUND] = ACTIONS(1626),
    [anon_sym_COLON_COLON] = ACTIONS(1628),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1628),
    [anon_sym_else] = ACTIONS(1626),
    [anon_sym_new] = ACTIONS(1626),
    [anon_sym_LBRACE_LT] = ACTIONS(1628),
    [anon_sym_begin] = ACTIONS(1626),
    [sym_ocamlyacc_value] = ACTIONS(1628),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1626),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1628),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1626),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1626),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1628),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1628),
    [aux_sym_number_token1] = ACTIONS(1628),
    [aux_sym_signed_number_token1] = ACTIONS(1626),
    [anon_sym_SQUOTE] = ACTIONS(1628),
    [anon_sym_DQUOTE] = ACTIONS(1628),
    [sym_prefix_operator] = ACTIONS(1628),
    [aux_sym_sign_operator_token1] = ACTIONS(1626),
    [sym_hash_operator] = ACTIONS(1628),
    [anon_sym_lsl] = ACTIONS(1626),
    [anon_sym_lsr] = ACTIONS(1626),
    [anon_sym_asr] = ACTIONS(1626),
    [aux_sym_pow_operator_token1] = ACTIONS(1628),
    [anon_sym_mod] = ACTIONS(1626),
    [anon_sym_land] = ACTIONS(1626),
    [anon_sym_lor] = ACTIONS(1626),
    [anon_sym_lxor] = ACTIONS(1626),
    [aux_sym_mult_operator_token1] = ACTIONS(1626),
    [aux_sym_add_operator_token1] = ACTIONS(1626),
    [sym_concat_operator] = ACTIONS(1628),
    [sym_rel_operator] = ACTIONS(1626),
    [anon_sym_AMP_AMP] = ACTIONS(1626),
    [anon_sym_or] = ACTIONS(1626),
    [anon_sym_PIPE_PIPE] = ACTIONS(1626),
    [sym_assign_operator] = ACTIONS(1628),
    [sym_let_operator] = ACTIONS(1628),
    [sym__uppercase_identifier] = ACTIONS(1628),
    [aux_sym_directive_token1] = ACTIONS(1626),
    [aux_sym_tag_token1] = ACTIONS(1628),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1147)] = {
    [sym_attribute] = STATE(1147),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_SEMI_SEMI] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2079),
    [anon_sym_and] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_external] = ACTIONS(2079),
    [anon_sym_type] = ACTIONS(2079),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2079),
    [anon_sym_exception] = ACTIONS(2079),
    [anon_sym_module] = ACTIONS(2079),
    [anon_sym_open] = ACTIONS(2079),
    [anon_sym_include] = ACTIONS(2079),
    [anon_sym_class] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_in] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym_let_operator] = ACTIONS(2081),
    [sym_let_and_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_directive_token1] = ACTIONS(2081),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1148)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4552),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1148),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1316),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_COLON2] = ACTIONS(1318),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_as] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1316),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1149)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4856),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1149),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1318),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_SEMI] = ACTIONS(1316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_as] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_COLON_COLON] = ACTIONS(1316),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1316),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1150)] = {
    [sym_attribute] = STATE(1150),
    [sym__lowercase_identifier] = ACTIONS(2123),
    [anon_sym_SEMI_SEMI] = ACTIONS(2125),
    [anon_sym_let] = ACTIONS(2123),
    [anon_sym_and] = ACTIONS(2123),
    [anon_sym_TILDE] = ACTIONS(2123),
    [anon_sym_QMARK] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(2125),
    [anon_sym_external] = ACTIONS(2123),
    [anon_sym_type] = ACTIONS(2123),
    [anon_sym_COMMA] = ACTIONS(2125),
    [anon_sym_PIPE] = ACTIONS(2123),
    [anon_sym_LBRACK] = ACTIONS(2123),
    [anon_sym_RBRACK] = ACTIONS(2125),
    [anon_sym_true] = ACTIONS(2123),
    [anon_sym_false] = ACTIONS(2123),
    [anon_sym_LBRACE] = ACTIONS(2123),
    [anon_sym_SEMI] = ACTIONS(2123),
    [anon_sym_exception] = ACTIONS(2123),
    [anon_sym_module] = ACTIONS(2123),
    [anon_sym_open] = ACTIONS(2123),
    [anon_sym_include] = ACTIONS(2123),
    [anon_sym_class] = ACTIONS(2123),
    [anon_sym_val] = ACTIONS(2123),
    [anon_sym_end] = ACTIONS(2123),
    [anon_sym_object] = ACTIONS(2123),
    [anon_sym_in] = ACTIONS(2123),
    [anon_sym_AMP] = ACTIONS(2123),
    [anon_sym_COLON_COLON] = ACTIONS(2125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2125),
    [anon_sym_else] = ACTIONS(2123),
    [anon_sym_new] = ACTIONS(2123),
    [anon_sym_LBRACE_LT] = ACTIONS(2125),
    [anon_sym_begin] = ACTIONS(2123),
    [sym_ocamlyacc_value] = ACTIONS(2125),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2123),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2125),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2125),
    [aux_sym_number_token1] = ACTIONS(2125),
    [aux_sym_signed_number_token1] = ACTIONS(2123),
    [anon_sym_SQUOTE] = ACTIONS(2125),
    [anon_sym_DQUOTE] = ACTIONS(2125),
    [sym_prefix_operator] = ACTIONS(2125),
    [aux_sym_sign_operator_token1] = ACTIONS(2123),
    [anon_sym_lsl] = ACTIONS(2123),
    [anon_sym_lsr] = ACTIONS(2123),
    [anon_sym_asr] = ACTIONS(2123),
    [aux_sym_pow_operator_token1] = ACTIONS(2125),
    [anon_sym_mod] = ACTIONS(2123),
    [anon_sym_land] = ACTIONS(2123),
    [anon_sym_lor] = ACTIONS(2123),
    [anon_sym_lxor] = ACTIONS(2123),
    [aux_sym_mult_operator_token1] = ACTIONS(2123),
    [aux_sym_add_operator_token1] = ACTIONS(2123),
    [sym_concat_operator] = ACTIONS(2125),
    [sym_rel_operator] = ACTIONS(2123),
    [anon_sym_AMP_AMP] = ACTIONS(2123),
    [anon_sym_or] = ACTIONS(2123),
    [anon_sym_PIPE_PIPE] = ACTIONS(2123),
    [sym_assign_operator] = ACTIONS(2125),
    [sym_let_operator] = ACTIONS(2125),
    [sym_let_and_operator] = ACTIONS(2125),
    [sym__uppercase_identifier] = ACTIONS(2125),
    [aux_sym_directive_token1] = ACTIONS(2125),
    [aux_sym_tag_token1] = ACTIONS(2125),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1151)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4994),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1151),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_PIPE] = ACTIONS(2145),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_SEMI] = ACTIONS(2139),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_as] = ACTIONS(2145),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_COLON_COLON] = ACTIONS(2139),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2139),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1152)] = {
    [sym__match_cases] = STATE(2309),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1152),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1181),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1153)] = {
    [sym__match_cases] = STATE(2309),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1153),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1198),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1154)] = {
    [sym__match_cases] = STATE(2231),
    [sym_match_case] = STATE(2336),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4809),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1154),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1180),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2340),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1155)] = {
    [sym__match_cases] = STATE(2231),
    [sym_match_case] = STATE(2224),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4686),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1155),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1221),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2342),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1156)] = {
    [sym__match_cases] = STATE(2230),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1156),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1191),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1157)] = {
    [sym__match_cases] = STATE(2937),
    [sym_match_case] = STATE(3163),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4720),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1157),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1207),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2344),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1158)] = {
    [sym__match_cases] = STATE(2230),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1158),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1197),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1159)] = {
    [sym_attribute] = STATE(1159),
    [sym__lowercase_identifier] = ACTIONS(2161),
    [anon_sym_SEMI_SEMI] = ACTIONS(2165),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_and] = ACTIONS(2161),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_QMARK] = ACTIONS(2161),
    [anon_sym_COLON] = ACTIONS(2346),
    [anon_sym_LPAREN] = ACTIONS(2165),
    [anon_sym_external] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_COMMA] = ACTIONS(2165),
    [anon_sym_PIPE] = ACTIONS(2161),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [anon_sym_RBRACK] = ACTIONS(2165),
    [anon_sym_true] = ACTIONS(2161),
    [anon_sym_false] = ACTIONS(2161),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [anon_sym_SEMI] = ACTIONS(2161),
    [anon_sym_exception] = ACTIONS(2161),
    [anon_sym_module] = ACTIONS(2161),
    [anon_sym_open] = ACTIONS(2161),
    [anon_sym_include] = ACTIONS(2161),
    [anon_sym_class] = ACTIONS(2161),
    [anon_sym_val] = ACTIONS(2161),
    [anon_sym_end] = ACTIONS(2161),
    [anon_sym_object] = ACTIONS(2161),
    [anon_sym_AMP] = ACTIONS(2161),
    [anon_sym_COLON_COLON] = ACTIONS(2165),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2165),
    [anon_sym_else] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2161),
    [anon_sym_LBRACE_LT] = ACTIONS(2165),
    [anon_sym_begin] = ACTIONS(2161),
    [sym_ocamlyacc_value] = ACTIONS(2165),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2161),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2165),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2165),
    [aux_sym_number_token1] = ACTIONS(2165),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2165),
    [anon_sym_DQUOTE] = ACTIONS(2165),
    [sym_prefix_operator] = ACTIONS(2165),
    [aux_sym_sign_operator_token1] = ACTIONS(2161),
    [anon_sym_lsl] = ACTIONS(2161),
    [anon_sym_lsr] = ACTIONS(2161),
    [anon_sym_asr] = ACTIONS(2161),
    [aux_sym_pow_operator_token1] = ACTIONS(2165),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_land] = ACTIONS(2161),
    [anon_sym_lor] = ACTIONS(2161),
    [anon_sym_lxor] = ACTIONS(2161),
    [aux_sym_mult_operator_token1] = ACTIONS(2161),
    [aux_sym_add_operator_token1] = ACTIONS(2161),
    [sym_concat_operator] = ACTIONS(2165),
    [sym_rel_operator] = ACTIONS(2161),
    [anon_sym_AMP_AMP] = ACTIONS(2161),
    [anon_sym_or] = ACTIONS(2161),
    [anon_sym_PIPE_PIPE] = ACTIONS(2161),
    [sym_assign_operator] = ACTIONS(2165),
    [sym_let_operator] = ACTIONS(2165),
    [sym__uppercase_identifier] = ACTIONS(2165),
    [aux_sym_directive_token1] = ACTIONS(2165),
    [aux_sym_tag_token1] = ACTIONS(2165),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1160)] = {
    [sym__match_cases] = STATE(2231),
    [sym_match_case] = STATE(2128),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4723),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1160),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1196),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2348),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1161)] = {
    [sym__match_cases] = STATE(2937),
    [sym_match_case] = STATE(3195),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4754),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1161),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1167),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1162)] = {
    [sym__match_cases] = STATE(2230),
    [sym_match_case] = STATE(3318),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4707),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1162),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1166),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1163)] = {
    [sym_attribute] = STATE(1163),
    [sym__lowercase_identifier] = ACTIONS(2123),
    [anon_sym_SEMI_SEMI] = ACTIONS(2125),
    [anon_sym_let] = ACTIONS(2123),
    [anon_sym_and] = ACTIONS(2123),
    [anon_sym_TILDE] = ACTIONS(2123),
    [anon_sym_QMARK] = ACTIONS(2123),
    [anon_sym_LPAREN] = ACTIONS(2125),
    [anon_sym_external] = ACTIONS(2123),
    [anon_sym_type] = ACTIONS(2123),
    [anon_sym_COMMA] = ACTIONS(2125),
    [anon_sym_PIPE] = ACTIONS(2123),
    [anon_sym_LBRACK] = ACTIONS(2123),
    [anon_sym_RBRACK] = ACTIONS(2125),
    [anon_sym_true] = ACTIONS(2123),
    [anon_sym_false] = ACTIONS(2123),
    [anon_sym_LBRACE] = ACTIONS(2123),
    [anon_sym_SEMI] = ACTIONS(2123),
    [anon_sym_exception] = ACTIONS(2123),
    [anon_sym_module] = ACTIONS(2123),
    [anon_sym_open] = ACTIONS(2123),
    [anon_sym_include] = ACTIONS(2123),
    [anon_sym_class] = ACTIONS(2123),
    [anon_sym_val] = ACTIONS(2123),
    [anon_sym_end] = ACTIONS(2123),
    [anon_sym_object] = ACTIONS(2123),
    [anon_sym_AMP] = ACTIONS(2123),
    [anon_sym_COLON_COLON] = ACTIONS(2125),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2125),
    [anon_sym_else] = ACTIONS(2123),
    [anon_sym_new] = ACTIONS(2123),
    [anon_sym_LBRACE_LT] = ACTIONS(2125),
    [anon_sym_begin] = ACTIONS(2123),
    [sym_ocamlyacc_value] = ACTIONS(2125),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2123),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2125),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2123),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2125),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2125),
    [aux_sym_number_token1] = ACTIONS(2125),
    [aux_sym_signed_number_token1] = ACTIONS(2123),
    [anon_sym_SQUOTE] = ACTIONS(2125),
    [anon_sym_DQUOTE] = ACTIONS(2125),
    [sym_prefix_operator] = ACTIONS(2125),
    [aux_sym_sign_operator_token1] = ACTIONS(2123),
    [anon_sym_lsl] = ACTIONS(2123),
    [anon_sym_lsr] = ACTIONS(2123),
    [anon_sym_asr] = ACTIONS(2123),
    [aux_sym_pow_operator_token1] = ACTIONS(2125),
    [anon_sym_mod] = ACTIONS(2123),
    [anon_sym_land] = ACTIONS(2123),
    [anon_sym_lor] = ACTIONS(2123),
    [anon_sym_lxor] = ACTIONS(2123),
    [aux_sym_mult_operator_token1] = ACTIONS(2123),
    [aux_sym_add_operator_token1] = ACTIONS(2123),
    [sym_concat_operator] = ACTIONS(2125),
    [sym_rel_operator] = ACTIONS(2123),
    [anon_sym_AMP_AMP] = ACTIONS(2123),
    [anon_sym_or] = ACTIONS(2123),
    [anon_sym_PIPE_PIPE] = ACTIONS(2123),
    [sym_assign_operator] = ACTIONS(2125),
    [sym_let_operator] = ACTIONS(2125),
    [sym__uppercase_identifier] = ACTIONS(2125),
    [aux_sym_directive_token1] = ACTIONS(2125),
    [aux_sym_tag_token1] = ACTIONS(2125),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1164)] = {
    [sym_attribute] = STATE(1164),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_SEMI_SEMI] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2079),
    [anon_sym_and] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_external] = ACTIONS(2079),
    [anon_sym_type] = ACTIONS(2079),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_PIPE] = ACTIONS(2079),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_RBRACK] = ACTIONS(2081),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_SEMI] = ACTIONS(2079),
    [anon_sym_exception] = ACTIONS(2079),
    [anon_sym_module] = ACTIONS(2079),
    [anon_sym_open] = ACTIONS(2079),
    [anon_sym_include] = ACTIONS(2079),
    [anon_sym_class] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym_let_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_directive_token1] = ACTIONS(2081),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1165)] = {
    [sym__match_cases] = STATE(2278),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1165),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1166)] = {
    [sym__match_cases] = STATE(2255),
    [sym_match_case] = STATE(3318),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4707),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1166),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1167)] = {
    [sym__match_cases] = STATE(2960),
    [sym_match_case] = STATE(3195),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4754),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1167),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1168)] = {
    [sym_parameter] = STATE(2236),
    [sym__maybe_polymorphic_typed] = STATE(4275),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(1168),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(4314),
    [aux_sym_let_binding_repeat1] = STATE(1193),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_EQ] = ACTIONS(2354),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_COLON2] = ACTIONS(2356),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_constraint] = ACTIONS(2358),
    [anon_sym_val] = ACTIONS(2358),
    [anon_sym_end] = ACTIONS(2358),
    [anon_sym_inherit] = ACTIONS(2358),
    [anon_sym_method] = ACTIONS(2358),
    [anon_sym_initializer] = ACTIONS(2358),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2360),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2360),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2360),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1169)] = {
    [sym_parameter] = STATE(2236),
    [sym__maybe_polymorphic_typed] = STATE(4288),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(1169),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(4352),
    [aux_sym_let_binding_repeat1] = STATE(1192),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_EQ] = ACTIONS(2362),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_COLON2] = ACTIONS(2356),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_constraint] = ACTIONS(2364),
    [anon_sym_val] = ACTIONS(2364),
    [anon_sym_end] = ACTIONS(2364),
    [anon_sym_inherit] = ACTIONS(2364),
    [anon_sym_method] = ACTIONS(2364),
    [anon_sym_initializer] = ACTIONS(2364),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2366),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2366),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2366),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1170)] = {
    [sym__match_cases] = STATE(2227),
    [sym_match_case] = STATE(2224),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4686),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1170),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2342),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1171)] = {
    [sym__match_cases] = STATE(2247),
    [sym_match_case] = STATE(2224),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4686),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1171),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2342),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1172)] = {
    [sym__match_cases] = STATE(2185),
    [sym_match_case] = STATE(2224),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4686),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1172),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2342),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1173)] = {
    [sym_parameter] = STATE(2236),
    [sym__maybe_polymorphic_typed] = STATE(4277),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(1173),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(4320),
    [aux_sym_let_binding_repeat1] = STATE(848),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_EQ] = ACTIONS(2368),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_COLON2] = ACTIONS(2356),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_constraint] = ACTIONS(2370),
    [anon_sym_val] = ACTIONS(2370),
    [anon_sym_end] = ACTIONS(2370),
    [anon_sym_inherit] = ACTIONS(2370),
    [anon_sym_method] = ACTIONS(2370),
    [anon_sym_initializer] = ACTIONS(2370),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2372),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2372),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2372),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1174)] = {
    [sym__match_cases] = STATE(2187),
    [sym_match_case] = STATE(2224),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4686),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1174),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2342),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1175)] = {
    [sym__match_cases] = STATE(2185),
    [sym_match_case] = STATE(2336),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4809),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1175),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2340),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1176)] = {
    [sym__match_cases] = STATE(2933),
    [sym_match_case] = STATE(3195),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4754),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1176),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1177)] = {
    [sym__match_cases] = STATE(2896),
    [sym_match_case] = STATE(3195),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4754),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1177),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1178)] = {
    [sym__match_cases] = STATE(2920),
    [sym_match_case] = STATE(3195),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4754),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1178),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1179)] = {
    [sym__match_cases] = STATE(2189),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1179),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1180)] = {
    [sym__match_cases] = STATE(2232),
    [sym_match_case] = STATE(2336),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4809),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1180),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2340),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1181)] = {
    [sym__match_cases] = STATE(2307),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1181),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1182)] = {
    [sym__match_cases] = STATE(2923),
    [sym_match_case] = STATE(3195),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4754),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1182),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2350),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1183)] = {
    [sym__match_cases] = STATE(2196),
    [sym_match_case] = STATE(3318),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4707),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1183),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1184)] = {
    [sym__match_cases] = STATE(2267),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1184),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1185)] = {
    [sym__match_cases] = STATE(2268),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1185),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1186)] = {
    [sym_parameter] = STATE(2236),
    [sym__coerced] = STATE(6410),
    [sym__maybe_polymorphic_typed] = STATE(5917),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(1186),
    [sym_item_attribute] = STATE(5327),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(4968),
    [aux_sym_let_binding_repeat1] = STATE(1546),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_and] = ACTIONS(1486),
    [anon_sym_EQ] = ACTIONS(2374),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_COMMA] = ACTIONS(2376),
    [anon_sym_PIPE] = ACTIONS(2378),
    [anon_sym_COLON2] = ACTIONS(1498),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_COLON_GT] = ACTIONS(1506),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_as] = ACTIONS(2380),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_COLON_COLON] = ACTIONS(2382),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2384),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2386),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2388),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym_let_and_operator] = ACTIONS(1484),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1187)] = {
    [sym__match_cases] = STATE(2276),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1187),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1188)] = {
    [sym__match_cases] = STATE(2247),
    [sym_match_case] = STATE(2336),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4809),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1188),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2340),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1189)] = {
    [sym__match_cases] = STATE(2182),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1189),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1190)] = {
    [sym__match_cases] = STATE(2247),
    [sym_match_case] = STATE(2128),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4723),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1190),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2348),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1191)] = {
    [sym__match_cases] = STATE(2255),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1191),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1192)] = {
    [sym_parameter] = STATE(2236),
    [sym__maybe_polymorphic_typed] = STATE(4294),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(1192),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(4381),
    [aux_sym_let_binding_repeat1] = STATE(848),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_EQ] = ACTIONS(2390),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_COLON2] = ACTIONS(2356),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_constraint] = ACTIONS(2392),
    [anon_sym_val] = ACTIONS(2392),
    [anon_sym_end] = ACTIONS(2392),
    [anon_sym_inherit] = ACTIONS(2392),
    [anon_sym_method] = ACTIONS(2392),
    [anon_sym_initializer] = ACTIONS(2392),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2394),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2394),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2394),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1193)] = {
    [sym_parameter] = STATE(2236),
    [sym__maybe_polymorphic_typed] = STATE(4287),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(1193),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(4350),
    [aux_sym_let_binding_repeat1] = STATE(848),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_EQ] = ACTIONS(2396),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_COLON2] = ACTIONS(2356),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_constraint] = ACTIONS(2398),
    [anon_sym_val] = ACTIONS(2398),
    [anon_sym_end] = ACTIONS(2398),
    [anon_sym_inherit] = ACTIONS(2398),
    [anon_sym_method] = ACTIONS(2398),
    [anon_sym_initializer] = ACTIONS(2398),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2400),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2400),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2400),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1194)] = {
    [sym__match_cases] = STATE(2248),
    [sym_match_case] = STATE(3318),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4707),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1194),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1195)] = {
    [sym__match_cases] = STATE(2196),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1195),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1196)] = {
    [sym__match_cases] = STATE(2232),
    [sym_match_case] = STATE(2128),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4723),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1196),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2348),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1197)] = {
    [sym__match_cases] = STATE(2255),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1197),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1198)] = {
    [sym__match_cases] = STATE(2307),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1198),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1199)] = {
    [sym__match_cases] = STATE(2189),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1199),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1200)] = {
    [sym__match_cases] = STATE(2185),
    [sym_match_case] = STATE(2128),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4723),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1200),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2348),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1201)] = {
    [sym__match_cases] = STATE(2227),
    [sym_match_case] = STATE(2336),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4809),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1201),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2340),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1202)] = {
    [sym__match_cases] = STATE(2248),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1202),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1203)] = {
    [sym__match_cases] = STATE(2182),
    [sym_match_case] = STATE(3318),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4707),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1203),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1204)] = {
    [sym_parameter] = STATE(2236),
    [sym__maybe_polymorphic_typed] = STATE(4306),
    [sym__parenthesized_abstract_type] = STATE(2236),
    [sym__delimited_pattern] = STATE(2241),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2241),
    [sym_polymorphic_variant_pattern] = STATE(2241),
    [sym_record_pattern] = STATE(2250),
    [sym_list_pattern] = STATE(2250),
    [sym_array_pattern] = STATE(2250),
    [sym_local_open_pattern] = STATE(2241),
    [sym_package_pattern] = STATE(2241),
    [sym_parenthesized_pattern] = STATE(2250),
    [sym_attribute] = STATE(1204),
    [sym_item_attribute] = STATE(3502),
    [sym_extension] = STATE(2241),
    [sym_quoted_extension] = STATE(2241),
    [sym__constant] = STATE(2251),
    [sym__signed_constant] = STATE(2241),
    [sym_number] = STATE(2226),
    [sym_signed_number] = STATE(2251),
    [sym_character] = STATE(2226),
    [sym_string] = STATE(2226),
    [sym_quoted_string] = STATE(2226),
    [sym_boolean] = STATE(2226),
    [sym_unit] = STATE(2250),
    [sym__simple_value_pattern] = STATE(2241),
    [sym_parenthesized_operator] = STATE(2241),
    [sym_module_path] = STATE(6882),
    [sym_constructor_path] = STATE(2241),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(2241),
    [aux_sym_expression_item_repeat1] = STATE(4423),
    [aux_sym_let_binding_repeat1] = STATE(1173),
    [sym__lowercase_identifier] = ACTIONS(1482),
    [anon_sym_EQ] = ACTIONS(2402),
    [anon_sym_TILDE] = ACTIONS(1490),
    [anon_sym_QMARK] = ACTIONS(1490),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_COLON2] = ACTIONS(2356),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_constraint] = ACTIONS(2404),
    [anon_sym_val] = ACTIONS(2404),
    [anon_sym_end] = ACTIONS(2404),
    [anon_sym_inherit] = ACTIONS(2404),
    [anon_sym_method] = ACTIONS(2404),
    [anon_sym_initializer] = ACTIONS(2404),
    [anon_sym_POUND] = ACTIONS(1510),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1514),
    [anon_sym_begin] = ACTIONS(1516),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2055),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2406),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1520),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1522),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2406),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2406),
    [aux_sym_number_token1] = ACTIONS(1524),
    [aux_sym_signed_number_token1] = ACTIONS(1526),
    [anon_sym_SQUOTE] = ACTIONS(1528),
    [anon_sym_DQUOTE] = ACTIONS(1530),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1205)] = {
    [sym__match_cases] = STATE(2267),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1205),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1206)] = {
    [sym__match_cases] = STATE(2187),
    [sym_match_case] = STATE(2336),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4809),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1206),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2340),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1207)] = {
    [sym__match_cases] = STATE(2960),
    [sym_match_case] = STATE(3163),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4720),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1207),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2344),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1208)] = {
    [sym__match_cases] = STATE(2189),
    [sym_match_case] = STATE(3318),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4707),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1208),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1209)] = {
    [sym__match_cases] = STATE(2268),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1209),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1210)] = {
    [sym__match_cases] = STATE(2276),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1210),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1211)] = {
    [sym__match_cases] = STATE(2227),
    [sym_match_case] = STATE(2128),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4723),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1211),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2348),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1212)] = {
    [sym__match_cases] = STATE(2278),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1212),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1213)] = {
    [sym__match_cases] = STATE(2933),
    [sym_match_case] = STATE(3163),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4720),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1213),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2344),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1214)] = {
    [sym__match_cases] = STATE(2196),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1214),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1215)] = {
    [sym__match_cases] = STATE(2248),
    [sym_match_case] = STATE(2290),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1215),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2336),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1216)] = {
    [sym__match_cases] = STATE(2896),
    [sym_match_case] = STATE(3163),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4720),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1216),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2344),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1217)] = {
    [sym__match_cases] = STATE(2187),
    [sym_match_case] = STATE(2128),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4723),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1217),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2348),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1218)] = {
    [sym__match_cases] = STATE(2920),
    [sym_match_case] = STATE(3163),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4720),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1218),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2344),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1219)] = {
    [sym__match_cases] = STATE(2923),
    [sym_match_case] = STATE(3163),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4720),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1219),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2344),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1220)] = {
    [sym__match_cases] = STATE(2182),
    [sym_match_case] = STATE(3321),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1220),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2338),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1221)] = {
    [sym__match_cases] = STATE(2232),
    [sym_match_case] = STATE(2224),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4686),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1221),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(2342),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1222)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5051),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1222),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1398),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1223)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4790),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym__sequence_pattern_content] = STATE(6401),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1223),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2408),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1224)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4718),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym__sequence_binding_pattern_content] = STATE(6427),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1224),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_RBRACK] = ACTIONS(2410),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1225)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4682),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym__sequence_binding_pattern_content] = STATE(6439),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1225),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2412),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1226)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4581),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1226),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__attribute] = STATE(1333),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1227)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4929),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1227),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__attribute] = STATE(1346),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1228)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4981),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1228),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__attribute] = STATE(1353),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1229)] = {
    [sym__parenthesized_abstract_type] = STATE(1345),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5397),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1229),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2143),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1230)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(5031),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1230),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__attribute] = STATE(1341),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1231)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5247),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1231),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1381),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1232)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5096),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1232),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(1102),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1114),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1233)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5202),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1233),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1102),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1114),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1234)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4698),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym__sequence_pattern_content] = STATE(6378),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1234),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(2414),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1235)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4790),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym__sequence_pattern_content] = STATE(6381),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1235),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2416),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1236)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4718),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym__sequence_binding_pattern_content] = STATE(6388),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1236),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_RBRACK] = ACTIONS(2418),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1237)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4682),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym__sequence_binding_pattern_content] = STATE(6394),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1237),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2420),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1238)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4851),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1238),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__attribute] = STATE(1342),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1239)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5396),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1239),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1344),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1240)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2066),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2059),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6426),
    [sym__tuple_binding_pattern] = STATE(2060),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(1240),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__attribute] = STATE(1352),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1241)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4885),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1241),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__attribute] = STATE(1325),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1242)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5237),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1242),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(1150),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1156),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1243)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5257),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1243),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1160),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1166),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1244)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(3088),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3080),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6514),
    [sym__tuple_binding_pattern] = STATE(3085),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(1244),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__attribute] = STATE(1368),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1245)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(3111),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3080),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6514),
    [sym__tuple_binding_pattern] = STATE(3085),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(1245),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__attribute] = STATE(1373),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1246)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5244),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1246),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(419),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1247)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5047),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1247),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(419),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1248)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5411),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1248),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__attribute] = STATE(1386),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1249)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5139),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1249),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(1136),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1128),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1250)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5101),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1250),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1044),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1062),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1251)] = {
    [sym_attribute] = STATE(1251),
    [sym__lowercase_identifier] = ACTIONS(2079),
    [anon_sym_and] = ACTIONS(2079),
    [anon_sym_TILDE] = ACTIONS(2079),
    [anon_sym_QMARK] = ACTIONS(2079),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_RPAREN] = ACTIONS(2081),
    [anon_sym_COMMA] = ACTIONS(2081),
    [anon_sym_COLON2] = ACTIONS(2079),
    [anon_sym_DOT] = ACTIONS(745),
    [anon_sym_LBRACK] = ACTIONS(2079),
    [anon_sym_true] = ACTIONS(2079),
    [anon_sym_false] = ACTIONS(2079),
    [anon_sym_LBRACE] = ACTIONS(2079),
    [anon_sym_constraint] = ACTIONS(2079),
    [anon_sym_val] = ACTIONS(2079),
    [anon_sym_end] = ACTIONS(2079),
    [anon_sym_COLON_GT] = ACTIONS(2081),
    [anon_sym_object] = ACTIONS(2079),
    [anon_sym_inherit] = ACTIONS(2079),
    [anon_sym_method] = ACTIONS(2079),
    [anon_sym_in] = ACTIONS(2079),
    [anon_sym_as] = ACTIONS(2079),
    [anon_sym_initializer] = ACTIONS(2079),
    [anon_sym_AMP] = ACTIONS(2079),
    [anon_sym_POUND] = ACTIONS(747),
    [anon_sym_COLON_COLON] = ACTIONS(2081),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2081),
    [anon_sym_else] = ACTIONS(2079),
    [anon_sym_new] = ACTIONS(2079),
    [anon_sym_LBRACE_LT] = ACTIONS(2081),
    [anon_sym_begin] = ACTIONS(2079),
    [sym_ocamlyacc_value] = ACTIONS(2081),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2079),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2081),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2079),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2081),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2081),
    [aux_sym_number_token1] = ACTIONS(2081),
    [aux_sym_signed_number_token1] = ACTIONS(2079),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_DQUOTE] = ACTIONS(2081),
    [sym_prefix_operator] = ACTIONS(2081),
    [aux_sym_sign_operator_token1] = ACTIONS(2079),
    [sym_hash_operator] = ACTIONS(1362),
    [anon_sym_lsl] = ACTIONS(2079),
    [anon_sym_lsr] = ACTIONS(2079),
    [anon_sym_asr] = ACTIONS(2079),
    [aux_sym_pow_operator_token1] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2079),
    [anon_sym_land] = ACTIONS(2079),
    [anon_sym_lor] = ACTIONS(2079),
    [anon_sym_lxor] = ACTIONS(2079),
    [aux_sym_mult_operator_token1] = ACTIONS(2079),
    [aux_sym_add_operator_token1] = ACTIONS(2079),
    [sym_concat_operator] = ACTIONS(2081),
    [sym_rel_operator] = ACTIONS(2079),
    [anon_sym_AMP_AMP] = ACTIONS(2079),
    [anon_sym_or] = ACTIONS(2079),
    [anon_sym_PIPE_PIPE] = ACTIONS(2079),
    [sym_assign_operator] = ACTIONS(2081),
    [sym__uppercase_identifier] = ACTIONS(2081),
    [aux_sym_tag_token1] = ACTIONS(2081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1252)] = {
    [sym_attribute] = STATE(1252),
    [sym__lowercase_identifier] = ACTIONS(2107),
    [anon_sym_and] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2107),
    [anon_sym_QMARK] = ACTIONS(2107),
    [anon_sym_LPAREN] = ACTIONS(2109),
    [anon_sym_RPAREN] = ACTIONS(2109),
    [anon_sym_COMMA] = ACTIONS(2109),
    [anon_sym_COLON2] = ACTIONS(2107),
    [anon_sym_DOT] = ACTIONS(745),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_true] = ACTIONS(2107),
    [anon_sym_false] = ACTIONS(2107),
    [anon_sym_LBRACE] = ACTIONS(2107),
    [anon_sym_constraint] = ACTIONS(2107),
    [anon_sym_val] = ACTIONS(2107),
    [anon_sym_end] = ACTIONS(2107),
    [anon_sym_COLON_GT] = ACTIONS(2109),
    [anon_sym_object] = ACTIONS(2107),
    [anon_sym_inherit] = ACTIONS(2107),
    [anon_sym_method] = ACTIONS(2107),
    [anon_sym_in] = ACTIONS(2107),
    [anon_sym_as] = ACTIONS(2107),
    [anon_sym_initializer] = ACTIONS(2107),
    [anon_sym_AMP] = ACTIONS(2107),
    [anon_sym_POUND] = ACTIONS(747),
    [anon_sym_COLON_COLON] = ACTIONS(2109),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2109),
    [anon_sym_else] = ACTIONS(2107),
    [anon_sym_new] = ACTIONS(2107),
    [anon_sym_LBRACE_LT] = ACTIONS(2109),
    [anon_sym_begin] = ACTIONS(2107),
    [sym_ocamlyacc_value] = ACTIONS(2109),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2107),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2109),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2107),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2109),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2109),
    [aux_sym_number_token1] = ACTIONS(2109),
    [aux_sym_signed_number_token1] = ACTIONS(2107),
    [anon_sym_SQUOTE] = ACTIONS(2109),
    [anon_sym_DQUOTE] = ACTIONS(2109),
    [sym_prefix_operator] = ACTIONS(2109),
    [aux_sym_sign_operator_token1] = ACTIONS(2107),
    [sym_hash_operator] = ACTIONS(1362),
    [anon_sym_lsl] = ACTIONS(2107),
    [anon_sym_lsr] = ACTIONS(2107),
    [anon_sym_asr] = ACTIONS(2107),
    [aux_sym_pow_operator_token1] = ACTIONS(2109),
    [anon_sym_mod] = ACTIONS(2107),
    [anon_sym_land] = ACTIONS(2107),
    [anon_sym_lor] = ACTIONS(2107),
    [anon_sym_lxor] = ACTIONS(2107),
    [aux_sym_mult_operator_token1] = ACTIONS(2107),
    [aux_sym_add_operator_token1] = ACTIONS(2107),
    [sym_concat_operator] = ACTIONS(2109),
    [sym_rel_operator] = ACTIONS(2107),
    [anon_sym_AMP_AMP] = ACTIONS(2107),
    [anon_sym_or] = ACTIONS(2107),
    [anon_sym_PIPE_PIPE] = ACTIONS(2107),
    [sym_assign_operator] = ACTIONS(2109),
    [sym__uppercase_identifier] = ACTIONS(2109),
    [aux_sym_tag_token1] = ACTIONS(2109),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1253)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4718),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym__sequence_binding_pattern_content] = STATE(6989),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1253),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_RBRACK] = ACTIONS(2426),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1254)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4698),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym__sequence_pattern_content] = STATE(6393),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1254),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(2428),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1255)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2098),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2059),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6426),
    [sym__tuple_binding_pattern] = STATE(2060),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(1255),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__attribute] = STATE(1356),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1256)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5244),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1256),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1146),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1257)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5047),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1257),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1146),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1258)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5096),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1258),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(1102),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1190),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1259)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5202),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1259),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(1102),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1190),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1260)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5244),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1260),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1196),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1261)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5047),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1261),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1196),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1262)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4698),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym__sequence_pattern_content] = STATE(6738),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1262),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(2430),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1263)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4790),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym__sequence_pattern_content] = STATE(6773),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1263),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2432),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1264)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4718),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym__sequence_binding_pattern_content] = STATE(7120),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1264),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_RBRACK] = ACTIONS(2434),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1265)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4682),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym__sequence_binding_pattern_content] = STATE(7282),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1265),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2436),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1266)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4581),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1266),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__attribute] = STATE(1359),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1267)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4787),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1267),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1363),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1268)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4336),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1268),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1364),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1269)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4527),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1269),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1335),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1270)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4336),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1270),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1336),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1271)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4805),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1271),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__attribute] = STATE(1370),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1272)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5244),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1272),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1128),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1273)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5047),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1273),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_RPAREN] = ACTIONS(411),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_COLON_COLON] = ACTIONS(1128),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1274)] = {
    [sym_let_binding] = STATE(5511),
    [sym_open_module] = STATE(6662),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(1274),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__attribute] = STATE(1410),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1275)] = {
    [sym__parenthesized_abstract_type] = STATE(1388),
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5412),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1275),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2270),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1276)] = {
    [sym__signature] = STATE(6987),
    [sym_external] = STATE(4204),
    [sym_type_definition] = STATE(4204),
    [sym_exception_definition] = STATE(4204),
    [sym_module_definition] = STATE(4204),
    [sym_module_type_definition] = STATE(4204),
    [sym_open_module] = STATE(4204),
    [sym_class_definition] = STATE(4204),
    [sym_class_type_definition] = STATE(4204),
    [sym__signature_item] = STATE(4229),
    [sym_value_specification] = STATE(4204),
    [sym_include_module_type] = STATE(4204),
    [sym__delimited_type] = STATE(2629),
    [sym__simple_type] = STATE(2542),
    [sym__type] = STATE(6152),
    [sym_function_type] = STATE(2667),
    [sym_labeled_argument_type] = STATE(7101),
    [sym__proper_tuple_type] = STATE(2871),
    [sym__labeled_tuple_type] = STATE(2667),
    [sym_labeled_tuple_element_type] = STATE(7116),
    [sym_constructed_type] = STATE(2629),
    [sym_aliased_type] = STATE(2667),
    [sym_local_open_type] = STATE(2629),
    [sym_polymorphic_variant_type] = STATE(2640),
    [sym_package_type] = STATE(2640),
    [sym_object_type] = STATE(2629),
    [sym_hash_type] = STATE(2629),
    [sym_parenthesized_type] = STATE(2640),
    [sym_attribute] = STATE(1276),
    [sym_floating_attribute] = STATE(4204),
    [sym_extension] = STATE(2629),
    [sym_quoted_extension] = STATE(2629),
    [sym_item_extension] = STATE(4204),
    [sym_quoted_item_extension] = STATE(4204),
    [sym_extended_module_path] = STATE(6236),
    [sym_type_constructor_path] = STATE(2629),
    [sym_type_variable] = STATE(2629),
    [aux_sym__structure_repeat1] = STATE(3159),
    [aux_sym__signature_repeat1] = STATE(3054),
    [sym__lowercase_identifier] = ACTIONS(2438),
    [anon_sym_SEMI_SEMI] = ACTIONS(9),
    [anon_sym_QMARK] = ACTIONS(2440),
    [anon_sym_LPAREN] = ACTIONS(2442),
    [anon_sym_external] = ACTIONS(2444),
    [anon_sym_type] = ACTIONS(2446),
    [anon_sym_LBRACK] = ACTIONS(2448),
    [anon_sym_RBRACK] = ACTIONS(2450),
    [anon_sym_exception] = ACTIONS(2452),
    [anon_sym_module] = ACTIONS(2454),
    [anon_sym_open] = ACTIONS(2456),
    [anon_sym_include] = ACTIONS(2458),
    [anon_sym_class] = ACTIONS(2460),
    [anon_sym_val] = ACTIONS(2462),
    [anon_sym_LBRACK_GT] = ACTIONS(2464),
    [anon_sym_LBRACK_LT] = ACTIONS(2466),
    [anon_sym_LT] = ACTIONS(2468),
    [anon_sym_POUND] = ACTIONS(2470),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(27),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2472),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2474),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(29),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(31),
    [sym__uppercase_identifier] = ACTIONS(2476),
    [aux_sym_type_variable_token1] = ACTIONS(2478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1277)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4698),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym__sequence_pattern_content] = STATE(6636),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1277),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(2480),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1278)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4790),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym__sequence_pattern_content] = STATE(6729),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1278),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2482),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1279)] = {
    [sym_let_binding] = STATE(5511),
    [sym_open_module] = STATE(6897),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(1279),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__attribute] = STATE(1410),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1280)] = {
    [sym_let_binding] = STATE(5511),
    [sym_open_module] = STATE(6441),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(1280),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__attribute] = STATE(1410),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1281)] = {
    [sym_let_binding] = STATE(5511),
    [sym_open_module] = STATE(6475),
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(2986),
    [sym__binding_pattern_no_exn] = STATE(1186),
    [sym_alias_binding_pattern_no_exn] = STATE(2986),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2986),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6408),
    [sym__tuple_binding_pattern_no_exn] = STATE(2986),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern_no_exn] = STATE(2986),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(2986),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_attribute] = STATE(1281),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__attribute] = STATE(1410),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_rec] = ACTIONS(1426),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_open] = ACTIONS(1444),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1282)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5236),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1282),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1322),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1283)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5245),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1283),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__attribute] = STATE(1362),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1284)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4682),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym__sequence_binding_pattern_content] = STATE(7003),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1284),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2484),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1285)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4811),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(4750),
    [sym__tuple_pattern] = STATE(4321),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1285),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2486),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_DOT_DOT] = ACTIONS(2488),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1286)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4933),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1286),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(2490),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1287)] = {
    [sym_match_case] = STATE(2237),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4686),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1287),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1288)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4830),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(4832),
    [sym__tuple_binding_pattern] = STATE(4837),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1288),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(2492),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_DOT_DOT] = ACTIONS(2494),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1289)] = {
    [sym_match_case] = STATE(2334),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4809),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1289),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1290)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4816),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(4822),
    [sym__tuple_binding_pattern] = STATE(4591),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1290),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(2496),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_DOT_DOT] = ACTIONS(2498),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1291)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4853),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1291),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_RBRACK] = ACTIONS(2500),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1292)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2031),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2059),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(2034),
    [sym__tuple_binding_pattern] = STATE(2035),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(1292),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(2502),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_DOT_DOT] = ACTIONS(2504),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1293)] = {
    [sym_match_case] = STATE(2129),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4723),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1293),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1294)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4853),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1294),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_RBRACK] = ACTIONS(2506),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1295)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2073),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2059),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(2074),
    [sym__tuple_binding_pattern] = STATE(2075),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(1295),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(2502),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_DOT_DOT] = ACTIONS(2508),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1296)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4922),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(4928),
    [sym__tuple_pattern] = STATE(4935),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1296),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2510),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_DOT_DOT] = ACTIONS(2512),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1297)] = {
    [sym_match_case] = STATE(2264),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1297),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1298)] = {
    [sym_match_case] = STATE(3349),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4707),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1298),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1299)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(3139),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3080),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(3141),
    [sym__tuple_binding_pattern] = STATE(3143),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(1299),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(2514),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_DOT_DOT] = ACTIONS(2516),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1300)] = {
    [sym_match_case] = STATE(2890),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4720),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1300),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1301)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4933),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1301),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_RBRACK] = ACTIONS(2518),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1302)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(5014),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1302),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2518),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1303)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4503),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(4504),
    [sym__tuple_pattern] = STATE(4321),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1303),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2520),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(2488),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1304)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4713),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(4725),
    [sym__tuple_binding_pattern] = STATE(4591),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1304),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(2522),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_DOT_DOT] = ACTIONS(2498),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1305)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4862),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1305),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2500),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1306)] = {
    [sym_match_case] = STATE(3232),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4754),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1306),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1307)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(5014),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1307),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2490),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1308)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4862),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1308),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2506),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1309)] = {
    [sym_match_case] = STATE(2237),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4723),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1309),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1310)] = {
    [sym_match_case] = STATE(3337),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1310),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1311)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5413),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1311),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1312)] = {
    [sym_match_case] = STATE(3398),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4707),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1312),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1313)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(2999),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3080),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(3000),
    [sym__tuple_binding_pattern] = STATE(3001),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(1313),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(2514),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_DOT_DOT] = ACTIONS(2524),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1314)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5398),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1314),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_COMMA] = ACTIONS(2139),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1315)] = {
    [sym_match_case] = STATE(2281),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4679),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1315),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1316)] = {
    [sym_match_case] = STATE(2890),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4754),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1316),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1317)] = {
    [sym_match_case] = STATE(3176),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4720),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1317),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1318)] = {
    [sym_match_case] = STATE(2225),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4686),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1318),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1319)] = {
    [sym_match_case] = STATE(2237),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4809),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1319),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1320)] = {
    [sym_match_case] = STATE(2281),
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4789),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1320),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1321)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4971),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1321),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1322)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5242),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1322),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1323)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4853),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1323),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1324)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4953),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1324),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1325)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5038),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1325),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1326)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4825),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4593),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1326),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1327)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4844),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1327),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1328)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4942),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(4961),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1328),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1329)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4881),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1329),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1330)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2036),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2042),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6426),
    [sym__tuple_binding_pattern] = STATE(2060),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(1330),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1331)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2046),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2059),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6426),
    [sym__tuple_binding_pattern] = STATE(2060),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(1331),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1332)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4977),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1332),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1333)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4618),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1333),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1334)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2050),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2059),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6426),
    [sym__tuple_binding_pattern] = STATE(2060),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(1334),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1335)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4526),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1335),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1336)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4410),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1336),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1337)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4506),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1337),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1338)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4510),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1338),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1339)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4327),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1339),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1340)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4589),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1340),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1341)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(5036),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1341),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1342)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4916),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1342),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1343)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4940),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1343),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1344)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5459),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1344),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1345)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5461),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1345),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1346)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4880),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1346),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1347)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4772),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1347),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1348)] = {
    [sym_let_binding] = STATE(3852),
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2054),
    [sym__binding_pattern_no_exn] = STATE(758),
    [sym_alias_binding_pattern_no_exn] = STATE(2054),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2054),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6346),
    [sym__tuple_binding_pattern_no_exn] = STATE(2054),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern_no_exn] = STATE(2054),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2054),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_attribute] = STATE(1348),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__attribute] = STATE(1429),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_rec] = ACTIONS(2526),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [anon_sym_PERCENT] = ACTIONS(1466),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1349)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4697),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1349),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1350)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4846),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(4847),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1350),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1351)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4852),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1351),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1352)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2071),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2059),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6426),
    [sym__tuple_binding_pattern] = STATE(2060),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(1352),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1353)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(4886),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1353),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1354)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2076),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2077),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6426),
    [sym__tuple_binding_pattern] = STATE(2060),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(1354),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1355)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2079),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2059),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6426),
    [sym__tuple_binding_pattern] = STATE(2060),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(1355),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1356)] = {
    [sym__delimited_binding_pattern] = STATE(2001),
    [sym__simple_binding_pattern] = STATE(2002),
    [sym__effect_binding_pattern] = STATE(2058),
    [sym__binding_pattern] = STATE(2014),
    [sym_alias_binding_pattern] = STATE(2058),
    [sym_typed_binding_pattern] = STATE(2001),
    [sym__or_binding_pattern_anonymous] = STATE(2059),
    [sym_constructor_binding_pattern] = STATE(2002),
    [sym_tag_binding_pattern] = STATE(2002),
    [sym_polymorphic_variant_pattern] = STATE(2001),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6426),
    [sym__tuple_binding_pattern] = STATE(2060),
    [sym_record_binding_pattern] = STATE(2080),
    [sym_list_binding_pattern] = STATE(2080),
    [sym_cons_binding_pattern] = STATE(2058),
    [sym_array_binding_pattern] = STATE(2080),
    [sym_range_pattern] = STATE(2058),
    [sym_lazy_binding_pattern] = STATE(2002),
    [sym_local_open_binding_pattern] = STATE(2001),
    [sym_package_pattern] = STATE(2001),
    [sym_parenthesized_binding_pattern] = STATE(2080),
    [sym_exception_binding_pattern] = STATE(2058),
    [sym_effect_binding_pattern] = STATE(2058),
    [sym_attribute] = STATE(1356),
    [sym_extension] = STATE(2001),
    [sym_quoted_extension] = STATE(2001),
    [sym__constant] = STATE(1857),
    [sym__signed_constant] = STATE(1876),
    [sym_number] = STATE(1861),
    [sym_signed_number] = STATE(1857),
    [sym_character] = STATE(1861),
    [sym_string] = STATE(1861),
    [sym_quoted_string] = STATE(1861),
    [sym_boolean] = STATE(1861),
    [sym_unit] = STATE(2080),
    [sym__value_name] = STATE(2001),
    [sym_parenthesized_operator] = STATE(2049),
    [sym_module_path] = STATE(6404),
    [sym_constructor_path] = STATE(730),
    [sym__constructor_name] = STATE(1735),
    [sym_tag] = STATE(731),
    [sym__lowercase_identifier] = ACTIONS(1314),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_exception] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1334),
    [anon_sym_lazy] = ACTIONS(1336),
    [anon_sym_begin] = ACTIONS(1338),
    [anon_sym_effect] = ACTIONS(1340),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2422),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2424),
    [aux_sym_number_token1] = ACTIONS(1346),
    [aux_sym_signed_number_token1] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1352),
    [sym__uppercase_identifier] = ACTIONS(1354),
    [aux_sym_tag_token1] = ACTIONS(1356),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1357)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4815),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1357),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1358)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4820),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1358),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1359)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4618),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1359),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1360)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(3144),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3145),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6514),
    [sym__tuple_binding_pattern] = STATE(3085),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(1360),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1361)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(3020),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3080),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6514),
    [sym__tuple_binding_pattern] = STATE(3085),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(1361),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1362)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5246),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1362),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1363)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4808),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1363),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1364)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4410),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1364),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1365)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4701),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1365),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1366)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4716),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1366),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1367)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4327),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1367),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1368)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(3045),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3080),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6514),
    [sym__tuple_binding_pattern] = STATE(3085),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(1368),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1369)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(3067),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3080),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6514),
    [sym__tuple_binding_pattern] = STATE(3085),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(1369),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1370)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4711),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1370),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1371)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4726),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4593),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1371),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1372)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4729),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6525),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1372),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(4689),
    [sym_module_path] = STATE(7252),
    [sym_constructor_path] = STATE(1057),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(2236),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2268),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(2240),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(2244),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2246),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1373)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(2994),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3080),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6514),
    [sym__tuple_binding_pattern] = STATE(3085),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(1373),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1374)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(3002),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3004),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6514),
    [sym__tuple_binding_pattern] = STATE(3085),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(1374),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1375)] = {
    [sym__delimited_pattern] = STATE(4958),
    [sym__simple_pattern] = STATE(5008),
    [sym__effect_pattern] = STATE(5009),
    [sym__pattern] = STATE(5014),
    [sym_alias_pattern] = STATE(5009),
    [sym_typed_pattern] = STATE(4958),
    [sym__or_pattern_anonymous] = STATE(5010),
    [sym_constructor_pattern] = STATE(5008),
    [sym_tag_pattern] = STATE(5008),
    [sym_polymorphic_variant_pattern] = STATE(4958),
    [sym_labeled_tuple_element_pattern] = STATE(7256),
    [sym__tuple_pattern] = STATE(5011),
    [sym_record_pattern] = STATE(4959),
    [sym_list_pattern] = STATE(4959),
    [sym_cons_pattern] = STATE(5009),
    [sym_array_pattern] = STATE(4959),
    [sym_range_pattern] = STATE(5009),
    [sym_lazy_pattern] = STATE(5008),
    [sym_local_open_pattern] = STATE(4958),
    [sym_package_pattern] = STATE(4958),
    [sym_parenthesized_pattern] = STATE(4959),
    [sym_exception_pattern] = STATE(5009),
    [sym_effect_pattern] = STATE(5009),
    [sym_attribute] = STATE(1375),
    [sym_extension] = STATE(4958),
    [sym_quoted_extension] = STATE(4958),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4783),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4959),
    [sym__simple_value_pattern] = STATE(4958),
    [sym_parenthesized_operator] = STATE(4958),
    [sym_module_path] = STATE(6411),
    [sym_constructor_path] = STATE(1082),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1151),
    [sym__lowercase_identifier] = ACTIONS(2272),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2334),
    [anon_sym_LBRACK] = ACTIONS(2276),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2280),
    [anon_sym_exception] = ACTIONS(2282),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2286),
    [anon_sym_lazy] = ACTIONS(2288),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2292),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1376)] = {
    [sym__delimited_binding_pattern] = STATE(4965),
    [sym__simple_binding_pattern] = STATE(4967),
    [sym__effect_binding_pattern] = STATE(4991),
    [sym__binding_pattern] = STATE(4862),
    [sym_alias_binding_pattern] = STATE(4991),
    [sym_typed_binding_pattern] = STATE(4965),
    [sym__or_binding_pattern_anonymous] = STATE(5005),
    [sym_constructor_binding_pattern] = STATE(4967),
    [sym_tag_binding_pattern] = STATE(4967),
    [sym_polymorphic_variant_pattern] = STATE(4965),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7303),
    [sym__tuple_binding_pattern] = STATE(5018),
    [sym_record_binding_pattern] = STATE(4969),
    [sym_list_binding_pattern] = STATE(4969),
    [sym_cons_binding_pattern] = STATE(4991),
    [sym_array_binding_pattern] = STATE(4969),
    [sym_range_pattern] = STATE(4991),
    [sym_lazy_binding_pattern] = STATE(4967),
    [sym_local_open_binding_pattern] = STATE(4965),
    [sym_package_pattern] = STATE(4965),
    [sym_parenthesized_binding_pattern] = STATE(4969),
    [sym_exception_binding_pattern] = STATE(4991),
    [sym_effect_binding_pattern] = STATE(4991),
    [sym_attribute] = STATE(1376),
    [sym_extension] = STATE(4965),
    [sym_quoted_extension] = STATE(4965),
    [sym__constant] = STATE(4788),
    [sym__signed_constant] = STATE(4745),
    [sym_number] = STATE(4738),
    [sym_signed_number] = STATE(4788),
    [sym_character] = STATE(4738),
    [sym_string] = STATE(4738),
    [sym_quoted_string] = STATE(4738),
    [sym_boolean] = STATE(4738),
    [sym_unit] = STATE(4969),
    [sym__value_name] = STATE(4965),
    [sym_parenthesized_operator] = STATE(4894),
    [sym_module_path] = STATE(6413),
    [sym_constructor_path] = STATE(1089),
    [sym__constructor_name] = STATE(3286),
    [sym_tag] = STATE(1149),
    [sym__lowercase_identifier] = ACTIONS(2310),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(2332),
    [anon_sym_LBRACK] = ACTIONS(2314),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2316),
    [anon_sym_exception] = ACTIONS(2318),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2320),
    [anon_sym_lazy] = ACTIONS(2322),
    [anon_sym_begin] = ACTIONS(2290),
    [anon_sym_effect] = ACTIONS(2324),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2294),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [aux_sym_number_token1] = ACTIONS(2298),
    [aux_sym_signed_number_token1] = ACTIONS(2300),
    [anon_sym_SQUOTE] = ACTIONS(2302),
    [anon_sym_DQUOTE] = ACTIONS(2304),
    [sym__uppercase_identifier] = ACTIONS(2306),
    [aux_sym_tag_token1] = ACTIONS(2308),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1377)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4722),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1377),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1378)] = {
    [sym__delimited_binding_pattern] = STATE(3030),
    [sym__simple_binding_pattern] = STATE(3031),
    [sym__effect_binding_pattern] = STATE(3079),
    [sym__binding_pattern] = STATE(3053),
    [sym_alias_binding_pattern] = STATE(3079),
    [sym_typed_binding_pattern] = STATE(3030),
    [sym__or_binding_pattern_anonymous] = STATE(3080),
    [sym_constructor_binding_pattern] = STATE(3031),
    [sym_tag_binding_pattern] = STATE(3031),
    [sym_polymorphic_variant_pattern] = STATE(3030),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6514),
    [sym__tuple_binding_pattern] = STATE(3085),
    [sym_record_binding_pattern] = STATE(2972),
    [sym_list_binding_pattern] = STATE(2972),
    [sym_cons_binding_pattern] = STATE(3079),
    [sym_array_binding_pattern] = STATE(2972),
    [sym_range_pattern] = STATE(3079),
    [sym_lazy_binding_pattern] = STATE(3031),
    [sym_local_open_binding_pattern] = STATE(3030),
    [sym_package_pattern] = STATE(3030),
    [sym_parenthesized_binding_pattern] = STATE(2972),
    [sym_exception_binding_pattern] = STATE(3079),
    [sym_effect_binding_pattern] = STATE(3079),
    [sym_attribute] = STATE(1378),
    [sym_extension] = STATE(3030),
    [sym_quoted_extension] = STATE(3030),
    [sym__constant] = STATE(2949),
    [sym__signed_constant] = STATE(2936),
    [sym_number] = STATE(2953),
    [sym_signed_number] = STATE(2949),
    [sym_character] = STATE(2953),
    [sym_string] = STATE(2953),
    [sym_quoted_string] = STATE(2953),
    [sym_boolean] = STATE(2953),
    [sym_unit] = STATE(2972),
    [sym__value_name] = STATE(3030),
    [sym_parenthesized_operator] = STATE(3051),
    [sym_module_path] = STATE(6505),
    [sym_constructor_path] = STATE(851),
    [sym__constructor_name] = STATE(2836),
    [sym_tag] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1424),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1436),
    [anon_sym_false] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1438),
    [anon_sym_exception] = ACTIONS(2013),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1450),
    [anon_sym_lazy] = ACTIONS(1452),
    [anon_sym_begin] = ACTIONS(1454),
    [anon_sym_effect] = ACTIONS(2015),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(1468),
    [aux_sym_signed_number_token1] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1472),
    [anon_sym_DQUOTE] = ACTIONS(1474),
    [sym__uppercase_identifier] = ACTIONS(1476),
    [aux_sym_tag_token1] = ACTIONS(1478),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1379)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4693),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1379),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1380)] = {
    [sym_attribute] = STATE(1380),
    [sym__lowercase_identifier] = ACTIONS(2161),
    [anon_sym_and] = ACTIONS(2161),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_QMARK] = ACTIONS(2161),
    [anon_sym_COLON] = ACTIONS(2528),
    [anon_sym_LPAREN] = ACTIONS(2165),
    [anon_sym_RPAREN] = ACTIONS(2165),
    [anon_sym_COMMA] = ACTIONS(2165),
    [anon_sym_COLON2] = ACTIONS(2161),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [anon_sym_true] = ACTIONS(2161),
    [anon_sym_false] = ACTIONS(2161),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [anon_sym_constraint] = ACTIONS(2161),
    [anon_sym_val] = ACTIONS(2161),
    [anon_sym_end] = ACTIONS(2161),
    [anon_sym_COLON_GT] = ACTIONS(2165),
    [anon_sym_object] = ACTIONS(2161),
    [anon_sym_inherit] = ACTIONS(2161),
    [anon_sym_method] = ACTIONS(2161),
    [anon_sym_in] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_initializer] = ACTIONS(2161),
    [anon_sym_AMP] = ACTIONS(2161),
    [anon_sym_COLON_COLON] = ACTIONS(2165),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2165),
    [anon_sym_else] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2161),
    [anon_sym_LBRACE_LT] = ACTIONS(2165),
    [anon_sym_begin] = ACTIONS(2161),
    [sym_ocamlyacc_value] = ACTIONS(2165),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2161),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2165),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2161),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2165),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2165),
    [aux_sym_number_token1] = ACTIONS(2165),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2165),
    [anon_sym_DQUOTE] = ACTIONS(2165),
    [sym_prefix_operator] = ACTIONS(2165),
    [aux_sym_sign_operator_token1] = ACTIONS(2161),
    [anon_sym_lsl] = ACTIONS(2161),
    [anon_sym_lsr] = ACTIONS(2161),
    [anon_sym_asr] = ACTIONS(2161),
    [aux_sym_pow_operator_token1] = ACTIONS(2165),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_land] = ACTIONS(2161),
    [anon_sym_lor] = ACTIONS(2161),
    [anon_sym_lxor] = ACTIONS(2161),
    [aux_sym_mult_operator_token1] = ACTIONS(2161),
    [aux_sym_add_operator_token1] = ACTIONS(2161),
    [sym_concat_operator] = ACTIONS(2165),
    [sym_rel_operator] = ACTIONS(2161),
    [anon_sym_AMP_AMP] = ACTIONS(2161),
    [anon_sym_or] = ACTIONS(2161),
    [anon_sym_PIPE_PIPE] = ACTIONS(2161),
    [sym_assign_operator] = ACTIONS(2165),
    [sym__uppercase_identifier] = ACTIONS(2165),
    [aux_sym_tag_token1] = ACTIONS(2165),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1381)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5248),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1381),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1382)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4812),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1382),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1383)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4728),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1383),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1384)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4749),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1384),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1385)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4778),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1385),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1386)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5424),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1386),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1387)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4779),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1387),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1388)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(5426),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1388),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1389)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4719),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1389),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1390)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4781),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1390),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1391)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1391),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1392)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4696),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1392),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1393)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4706),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1393),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1394)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4714),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1394),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1395)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4735),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1395),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1396)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4740),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1396),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1397)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4758),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1397),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1398)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(5140),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1398),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1399)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4762),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1399),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1400)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4933),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6414),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1400),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(7248),
    [sym_constructor_path] = STATE(1065),
    [sym__constructor_name] = STATE(3207),
    [sym_tag] = STATE(1143),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(2330),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(2260),
    [anon_sym_POUND] = ACTIONS(2242),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(2262),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(2264),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2248),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2250),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(2254),
    [aux_sym_tag_token1] = ACTIONS(2256),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1401)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4795),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1401),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1402)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4804),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1402),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1403)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4683),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1403),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1404)] = {
    [sym__delimited_pattern] = STATE(4430),
    [sym__simple_pattern] = STATE(4379),
    [sym__effect_pattern] = STATE(4441),
    [sym__pattern] = STATE(4684),
    [sym_alias_pattern] = STATE(4441),
    [sym_typed_pattern] = STATE(4430),
    [sym__or_pattern_anonymous] = STATE(4387),
    [sym_constructor_pattern] = STATE(4379),
    [sym_tag_pattern] = STATE(4379),
    [sym_polymorphic_variant_pattern] = STATE(4430),
    [sym_labeled_tuple_element_pattern] = STATE(6647),
    [sym__tuple_pattern] = STATE(4388),
    [sym_record_pattern] = STATE(4431),
    [sym_list_pattern] = STATE(4431),
    [sym_cons_pattern] = STATE(4441),
    [sym_array_pattern] = STATE(4431),
    [sym_range_pattern] = STATE(4441),
    [sym_lazy_pattern] = STATE(4379),
    [sym_local_open_pattern] = STATE(4430),
    [sym_package_pattern] = STATE(4430),
    [sym_parenthesized_pattern] = STATE(4431),
    [sym_exception_pattern] = STATE(4441),
    [sym_effect_pattern] = STATE(4441),
    [sym_attribute] = STATE(1404),
    [sym_extension] = STATE(4430),
    [sym_quoted_extension] = STATE(4430),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4295),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4431),
    [sym__simple_value_pattern] = STATE(4430),
    [sym_parenthesized_operator] = STATE(4430),
    [sym_module_path] = STATE(6396),
    [sym_constructor_path] = STATE(889),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1096),
    [anon_sym_TILDE] = ACTIONS(2141),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_exception] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1116),
    [anon_sym_lazy] = ACTIONS(1118),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1120),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1405)] = {
    [sym__delimited_binding_pattern] = STATE(4603),
    [sym__simple_binding_pattern] = STATE(4604),
    [sym__effect_binding_pattern] = STATE(4561),
    [sym__binding_pattern] = STATE(4589),
    [sym_alias_binding_pattern] = STATE(4561),
    [sym_typed_binding_pattern] = STATE(4603),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4604),
    [sym_tag_binding_pattern] = STATE(4604),
    [sym_polymorphic_variant_pattern] = STATE(4603),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6955),
    [sym__tuple_binding_pattern] = STATE(4598),
    [sym_record_binding_pattern] = STATE(4605),
    [sym_list_binding_pattern] = STATE(4605),
    [sym_cons_binding_pattern] = STATE(4561),
    [sym_array_binding_pattern] = STATE(4605),
    [sym_range_pattern] = STATE(4561),
    [sym_lazy_binding_pattern] = STATE(4604),
    [sym_local_open_binding_pattern] = STATE(4603),
    [sym_package_pattern] = STATE(4603),
    [sym_parenthesized_binding_pattern] = STATE(4605),
    [sym_exception_binding_pattern] = STATE(4561),
    [sym_effect_binding_pattern] = STATE(4561),
    [sym_attribute] = STATE(1405),
    [sym_extension] = STATE(4603),
    [sym_quoted_extension] = STATE(4603),
    [sym__constant] = STATE(4253),
    [sym__signed_constant] = STATE(4539),
    [sym_number] = STATE(4109),
    [sym_signed_number] = STATE(4253),
    [sym_character] = STATE(4109),
    [sym_string] = STATE(4109),
    [sym_quoted_string] = STATE(4109),
    [sym_boolean] = STATE(4109),
    [sym_unit] = STATE(4605),
    [sym__value_name] = STATE(4603),
    [sym_parenthesized_operator] = STATE(3541),
    [sym_module_path] = STATE(6409),
    [sym_constructor_path] = STATE(1073),
    [sym__constructor_name] = STATE(1582),
    [sym_tag] = STATE(1148),
    [sym__lowercase_identifier] = ACTIONS(1038),
    [anon_sym_TILDE] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1042),
    [anon_sym_LBRACK] = ACTIONS(1050),
    [anon_sym_true] = ACTIONS(1052),
    [anon_sym_false] = ACTIONS(1052),
    [anon_sym_LBRACE] = ACTIONS(1054),
    [anon_sym_exception] = ACTIONS(1056),
    [anon_sym_POUND] = ACTIONS(2147),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1064),
    [anon_sym_lazy] = ACTIONS(1066),
    [anon_sym_begin] = ACTIONS(1068),
    [anon_sym_effect] = ACTIONS(1070),
    [aux_sym_attribute_token1] = ACTIONS(405),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1072),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1074),
    [aux_sym_number_token1] = ACTIONS(1076),
    [aux_sym_signed_number_token1] = ACTIONS(2252),
    [anon_sym_SQUOTE] = ACTIONS(1080),
    [anon_sym_DQUOTE] = ACTIONS(1082),
    [sym__uppercase_identifier] = ACTIONS(1086),
    [aux_sym_tag_token1] = ACTIONS(1088),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1216), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1406), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [138] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1208), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1407), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [276] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1224), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1408), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [414] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1230), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1409), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [552] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(1532), 1,
      anon_sym_rec,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1410), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5498), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [690] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(1411), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5508), 1,
      sym__simple_expression,
    STATE(5610), 1,
      sym_field_expression,
    STATE(6994), 1,
      sym__quoted_string,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [822] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2558), 1,
      anon_sym_LT_DASH,
    STATE(1412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 25,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      anon_sym_GT_RBRACE,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 33,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_then,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [898] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2560), 1,
      anon_sym_LT_DASH,
    STATE(1413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 37,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [974] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1242), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1414), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1112] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(1415), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5513), 1,
      sym__simple_expression,
    STATE(5912), 1,
      sym_field_expression,
    STATE(7035), 1,
      sym__quoted_string,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [1244] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(1416), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5500), 1,
      sym__simple_expression,
    STATE(5675), 1,
      sym_field_expression,
    STATE(6541), 1,
      sym__quoted_string,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [1376] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1218), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1417), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1514] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1214), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1418), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1652] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(1419), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5514), 1,
      sym__simple_expression,
    STATE(5630), 1,
      sym_field_expression,
    STATE(7141), 1,
      sym__quoted_string,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [1784] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1204), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1420), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1922] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(1421), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5515), 1,
      sym__simple_expression,
    STATE(5685), 1,
      sym_field_expression,
    STATE(7190), 1,
      sym__quoted_string,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2054] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1228), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1422), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [2192] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(1423), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5504), 1,
      sym__simple_expression,
    STATE(5774), 1,
      sym_field_expression,
    STATE(6888), 1,
      sym__quoted_string,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2324] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(1424), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5493), 1,
      sym__simple_expression,
    STATE(5827), 1,
      sym_field_expression,
    STATE(7085), 1,
      sym__quoted_string,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2456] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1200), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1425), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [2594] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(1426), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5460), 1,
      sym__simple_expression,
    STATE(5632), 1,
      sym_field_expression,
    STATE(6792), 1,
      sym__quoted_string,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2726] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(1427), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5516), 1,
      sym__simple_expression,
    STATE(5881), 1,
      sym_field_expression,
    STATE(7175), 1,
      sym__quoted_string,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2858] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1194), 1,
      anon_sym_RPAREN,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1428), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [2996] = 37,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1314), 1,
      sym__lowercase_identifier,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1328), 1,
      anon_sym_LBRACE,
    ACTIONS(1332), 1,
      anon_sym_POUND,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1336), 1,
      anon_sym_lazy,
    ACTIONS(1338), 1,
      anon_sym_begin,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(1354), 1,
      sym__uppercase_identifier,
    ACTIONS(1356), 1,
      aux_sym_tag_token1,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    ACTIONS(2422), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2562), 1,
      anon_sym_rec,
    STATE(730), 1,
      sym_constructor_path,
    STATE(731), 1,
      sym_tag,
    STATE(758), 1,
      sym__binding_pattern_no_exn,
    STATE(1429), 1,
      sym_attribute,
    STATE(1735), 1,
      sym__constructor_name,
    STATE(1876), 1,
      sym__signed_constant,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(3861), 1,
      sym_let_binding,
    STATE(6346), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1326), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(2002), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2054), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2001), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3134] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(1430), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5493), 1,
      sym__simple_expression,
    STATE(5827), 1,
      sym_field_expression,
    STATE(7239), 1,
      sym_module_path,
    STATE(7321), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [3266] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(1431), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5506), 1,
      sym__simple_expression,
    STATE(5854), 1,
      sym_field_expression,
    STATE(6943), 1,
      sym__quoted_string,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [3398] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1432), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5741), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3533] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1314), 1,
      sym__lowercase_identifier,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1328), 1,
      anon_sym_LBRACE,
    ACTIONS(1332), 1,
      anon_sym_POUND,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1336), 1,
      anon_sym_lazy,
    ACTIONS(1338), 1,
      anon_sym_begin,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(1354), 1,
      sym__uppercase_identifier,
    ACTIONS(1356), 1,
      aux_sym_tag_token1,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    ACTIONS(2422), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(730), 1,
      sym_constructor_path,
    STATE(731), 1,
      sym_tag,
    STATE(750), 1,
      sym__binding_pattern_no_exn,
    STATE(1433), 1,
      sym_attribute,
    STATE(1735), 1,
      sym__constructor_name,
    STATE(1876), 1,
      sym__signed_constant,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(3784), 1,
      sym_let_binding,
    STATE(6346), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1326), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(2002), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2054), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2001), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3668] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1314), 1,
      sym__lowercase_identifier,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1328), 1,
      anon_sym_LBRACE,
    ACTIONS(1332), 1,
      anon_sym_POUND,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1336), 1,
      anon_sym_lazy,
    ACTIONS(1338), 1,
      anon_sym_begin,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(1354), 1,
      sym__uppercase_identifier,
    ACTIONS(1356), 1,
      aux_sym_tag_token1,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    ACTIONS(2422), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(730), 1,
      sym_constructor_path,
    STATE(731), 1,
      sym_tag,
    STATE(750), 1,
      sym__binding_pattern_no_exn,
    STATE(1434), 1,
      sym_attribute,
    STATE(1735), 1,
      sym__constructor_name,
    STATE(1876), 1,
      sym__signed_constant,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(3752), 1,
      sym_let_binding,
    STATE(6346), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1326), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(2002), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2054), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2001), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3803] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1435), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5529), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3938] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1314), 1,
      sym__lowercase_identifier,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1328), 1,
      anon_sym_LBRACE,
    ACTIONS(1332), 1,
      anon_sym_POUND,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1336), 1,
      anon_sym_lazy,
    ACTIONS(1338), 1,
      anon_sym_begin,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(1354), 1,
      sym__uppercase_identifier,
    ACTIONS(1356), 1,
      aux_sym_tag_token1,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    ACTIONS(2422), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(730), 1,
      sym_constructor_path,
    STATE(731), 1,
      sym_tag,
    STATE(750), 1,
      sym__binding_pattern_no_exn,
    STATE(1436), 1,
      sym_attribute,
    STATE(1735), 1,
      sym__constructor_name,
    STATE(1876), 1,
      sym__signed_constant,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(3773), 1,
      sym_let_binding,
    STATE(6346), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1326), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(2002), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2054), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2001), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4073] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1314), 1,
      sym__lowercase_identifier,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1328), 1,
      anon_sym_LBRACE,
    ACTIONS(1332), 1,
      anon_sym_POUND,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1336), 1,
      anon_sym_lazy,
    ACTIONS(1338), 1,
      anon_sym_begin,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(1354), 1,
      sym__uppercase_identifier,
    ACTIONS(1356), 1,
      aux_sym_tag_token1,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    ACTIONS(2422), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(730), 1,
      sym_constructor_path,
    STATE(731), 1,
      sym_tag,
    STATE(750), 1,
      sym__binding_pattern_no_exn,
    STATE(1437), 1,
      sym_attribute,
    STATE(1735), 1,
      sym__constructor_name,
    STATE(1876), 1,
      sym__signed_constant,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(3864), 1,
      sym_let_binding,
    STATE(6346), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1326), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(2002), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2054), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2001), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4208] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1314), 1,
      sym__lowercase_identifier,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1328), 1,
      anon_sym_LBRACE,
    ACTIONS(1332), 1,
      anon_sym_POUND,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1336), 1,
      anon_sym_lazy,
    ACTIONS(1338), 1,
      anon_sym_begin,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(1354), 1,
      sym__uppercase_identifier,
    ACTIONS(1356), 1,
      aux_sym_tag_token1,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    ACTIONS(2422), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(730), 1,
      sym_constructor_path,
    STATE(731), 1,
      sym_tag,
    STATE(758), 1,
      sym__binding_pattern_no_exn,
    STATE(1438), 1,
      sym_attribute,
    STATE(1735), 1,
      sym__constructor_name,
    STATE(1876), 1,
      sym__signed_constant,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(3852), 1,
      sym_let_binding,
    STATE(6346), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1326), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(2002), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2054), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2001), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4343] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1314), 1,
      sym__lowercase_identifier,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1328), 1,
      anon_sym_LBRACE,
    ACTIONS(1332), 1,
      anon_sym_POUND,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1336), 1,
      anon_sym_lazy,
    ACTIONS(1338), 1,
      anon_sym_begin,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(1354), 1,
      sym__uppercase_identifier,
    ACTIONS(1356), 1,
      aux_sym_tag_token1,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    ACTIONS(2422), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(730), 1,
      sym_constructor_path,
    STATE(731), 1,
      sym_tag,
    STATE(758), 1,
      sym__binding_pattern_no_exn,
    STATE(1439), 1,
      sym_attribute,
    STATE(1735), 1,
      sym__constructor_name,
    STATE(1876), 1,
      sym__signed_constant,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(3861), 1,
      sym_let_binding,
    STATE(6346), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1326), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(2002), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2054), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2001), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4478] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1314), 1,
      sym__lowercase_identifier,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1328), 1,
      anon_sym_LBRACE,
    ACTIONS(1332), 1,
      anon_sym_POUND,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1336), 1,
      anon_sym_lazy,
    ACTIONS(1338), 1,
      anon_sym_begin,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(1354), 1,
      sym__uppercase_identifier,
    ACTIONS(1356), 1,
      aux_sym_tag_token1,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    ACTIONS(2422), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(730), 1,
      sym_constructor_path,
    STATE(731), 1,
      sym_tag,
    STATE(758), 1,
      sym__binding_pattern_no_exn,
    STATE(1440), 1,
      sym_attribute,
    STATE(1735), 1,
      sym__constructor_name,
    STATE(1876), 1,
      sym__signed_constant,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(3864), 1,
      sym_let_binding,
    STATE(6346), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1326), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(2002), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2054), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2001), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4613] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1441), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5511), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4748] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1314), 1,
      sym__lowercase_identifier,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1328), 1,
      anon_sym_LBRACE,
    ACTIONS(1332), 1,
      anon_sym_POUND,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1336), 1,
      anon_sym_lazy,
    ACTIONS(1338), 1,
      anon_sym_begin,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(1354), 1,
      sym__uppercase_identifier,
    ACTIONS(1356), 1,
      aux_sym_tag_token1,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    ACTIONS(2422), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(730), 1,
      sym_constructor_path,
    STATE(731), 1,
      sym_tag,
    STATE(758), 1,
      sym__binding_pattern_no_exn,
    STATE(1442), 1,
      sym_attribute,
    STATE(1735), 1,
      sym__constructor_name,
    STATE(1876), 1,
      sym__signed_constant,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(3865), 1,
      sym_let_binding,
    STATE(6346), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1326), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(2002), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2054), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2001), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4883] = 36,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1320), 1,
      anon_sym_TILDE,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1452), 1,
      anon_sym_lazy,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(851), 1,
      sym_constructor_path,
    STATE(864), 1,
      sym_tag,
    STATE(1186), 1,
      sym__binding_pattern_no_exn,
    STATE(1443), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(2936), 1,
      sym__signed_constant,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5498), 1,
      sym_let_binding,
    STATE(6408), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(3031), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2986), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3030), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [5018] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(1444), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5514), 1,
      sym__simple_expression,
    STATE(5630), 1,
      sym_field_expression,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [5144] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(1445), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5493), 1,
      sym__simple_expression,
    STATE(5827), 1,
      sym_field_expression,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [5270] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(1446), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5508), 1,
      sym__simple_expression,
    STATE(5610), 1,
      sym_field_expression,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [5396] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2564), 1,
      sym__lowercase_identifier,
    ACTIONS(2566), 1,
      anon_sym_LPAREN,
    ACTIONS(2568), 1,
      anon_sym_LBRACK,
    ACTIONS(2572), 1,
      anon_sym_LBRACE,
    ACTIONS(2574), 1,
      anon_sym_object,
    ACTIONS(2576), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2578), 1,
      anon_sym_new,
    ACTIONS(2580), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2582), 1,
      anon_sym_begin,
    ACTIONS(2584), 1,
      sym_ocamlyacc_value,
    ACTIONS(2586), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2588), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2590), 1,
      anon_sym_PERCENT,
    ACTIONS(2592), 1,
      aux_sym_number_token1,
    ACTIONS(2594), 1,
      anon_sym_SQUOTE,
    ACTIONS(2596), 1,
      anon_sym_DQUOTE,
    ACTIONS(2598), 1,
      sym_prefix_operator,
    ACTIONS(2600), 1,
      sym__uppercase_identifier,
    ACTIONS(2602), 1,
      aux_sym_tag_token1,
    STATE(1447), 1,
      sym_attribute,
    STATE(1499), 1,
      sym__attribute,
    STATE(1981), 1,
      sym__value_name,
    STATE(2005), 1,
      sym_parenthesized_operator,
    STATE(2030), 1,
      sym__constructor_name,
    STATE(2053), 1,
      sym__simple_expression,
    STATE(6669), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2570), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2017), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2122), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2121), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [5522] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(1448), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5515), 1,
      sym__simple_expression,
    STATE(5685), 1,
      sym_field_expression,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [5648] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2590), 1,
      anon_sym_PERCENT,
    ACTIONS(2604), 1,
      sym__lowercase_identifier,
    ACTIONS(2606), 1,
      anon_sym_LPAREN,
    ACTIONS(2608), 1,
      anon_sym_LBRACK,
    ACTIONS(2612), 1,
      anon_sym_LBRACE,
    ACTIONS(2614), 1,
      anon_sym_object,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2618), 1,
      anon_sym_new,
    ACTIONS(2620), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2622), 1,
      anon_sym_begin,
    ACTIONS(2624), 1,
      sym_ocamlyacc_value,
    ACTIONS(2626), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2628), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2630), 1,
      aux_sym_number_token1,
    ACTIONS(2632), 1,
      anon_sym_SQUOTE,
    ACTIONS(2634), 1,
      anon_sym_DQUOTE,
    ACTIONS(2636), 1,
      sym_prefix_operator,
    ACTIONS(2638), 1,
      sym__uppercase_identifier,
    ACTIONS(2640), 1,
      aux_sym_tag_token1,
    STATE(1449), 1,
      sym_attribute,
    STATE(1506), 1,
      sym__attribute,
    STATE(1807), 1,
      sym__constructor_name,
    STATE(1877), 1,
      sym_parenthesized_operator,
    STATE(1888), 1,
      sym__simple_expression,
    STATE(1893), 1,
      sym__value_name,
    STATE(7240), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2610), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1891), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1926), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1921), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [5774] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2590), 1,
      anon_sym_PERCENT,
    ACTIONS(2604), 1,
      sym__lowercase_identifier,
    ACTIONS(2606), 1,
      anon_sym_LPAREN,
    ACTIONS(2608), 1,
      anon_sym_LBRACK,
    ACTIONS(2612), 1,
      anon_sym_LBRACE,
    ACTIONS(2614), 1,
      anon_sym_object,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2618), 1,
      anon_sym_new,
    ACTIONS(2620), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2622), 1,
      anon_sym_begin,
    ACTIONS(2624), 1,
      sym_ocamlyacc_value,
    ACTIONS(2626), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2628), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2630), 1,
      aux_sym_number_token1,
    ACTIONS(2632), 1,
      anon_sym_SQUOTE,
    ACTIONS(2634), 1,
      anon_sym_DQUOTE,
    ACTIONS(2636), 1,
      sym_prefix_operator,
    ACTIONS(2638), 1,
      sym__uppercase_identifier,
    ACTIONS(2640), 1,
      aux_sym_tag_token1,
    STATE(1450), 1,
      sym_attribute,
    STATE(1529), 1,
      sym__attribute,
    STATE(1807), 1,
      sym__constructor_name,
    STATE(1877), 1,
      sym_parenthesized_operator,
    STATE(1893), 1,
      sym__value_name,
    STATE(1911), 1,
      sym__simple_expression,
    STATE(7240), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2610), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1891), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1926), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1921), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [5900] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(1451), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5460), 1,
      sym__simple_expression,
    STATE(5632), 1,
      sym_field_expression,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [6026] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2642), 1,
      anon_sym_EQ,
    ACTIONS(2644), 1,
      anon_sym_COLON2,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(1452), 1,
      sym_attribute,
    STATE(1471), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4952), 1,
      sym__class_typed,
    STATE(5117), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2091), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [6160] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(1453), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5504), 1,
      sym__simple_expression,
    STATE(5774), 1,
      sym_field_expression,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [6286] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(1454), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5506), 1,
      sym__simple_expression,
    STATE(5854), 1,
      sym_field_expression,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [6412] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2644), 1,
      anon_sym_COLON2,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2648), 1,
      anon_sym_EQ,
    STATE(1455), 1,
      sym_attribute,
    STATE(1456), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4962), 1,
      sym__class_typed,
    STATE(5137), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2103), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [6546] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2644), 1,
      anon_sym_COLON2,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2650), 1,
      anon_sym_EQ,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1456), 1,
      sym_attribute,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4870), 1,
      sym__class_typed,
    STATE(5188), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2113), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [6680] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2644), 1,
      anon_sym_COLON2,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2652), 1,
      anon_sym_EQ,
    STATE(1457), 1,
      sym_attribute,
    STATE(1458), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4871), 1,
      sym__class_typed,
    STATE(5197), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2119), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [6814] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2644), 1,
      anon_sym_COLON2,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2654), 1,
      anon_sym_EQ,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1458), 1,
      sym_attribute,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4906), 1,
      sym__class_typed,
    STATE(5189), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2135), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [6948] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2564), 1,
      sym__lowercase_identifier,
    ACTIONS(2566), 1,
      anon_sym_LPAREN,
    ACTIONS(2568), 1,
      anon_sym_LBRACK,
    ACTIONS(2572), 1,
      anon_sym_LBRACE,
    ACTIONS(2574), 1,
      anon_sym_object,
    ACTIONS(2576), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2578), 1,
      anon_sym_new,
    ACTIONS(2580), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2582), 1,
      anon_sym_begin,
    ACTIONS(2584), 1,
      sym_ocamlyacc_value,
    ACTIONS(2586), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2588), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2590), 1,
      anon_sym_PERCENT,
    ACTIONS(2592), 1,
      aux_sym_number_token1,
    ACTIONS(2594), 1,
      anon_sym_SQUOTE,
    ACTIONS(2596), 1,
      anon_sym_DQUOTE,
    ACTIONS(2598), 1,
      sym_prefix_operator,
    ACTIONS(2600), 1,
      sym__uppercase_identifier,
    ACTIONS(2602), 1,
      aux_sym_tag_token1,
    STATE(1459), 1,
      sym_attribute,
    STATE(1487), 1,
      sym__attribute,
    STATE(1981), 1,
      sym__value_name,
    STATE(2005), 1,
      sym_parenthesized_operator,
    STATE(2030), 1,
      sym__constructor_name,
    STATE(2052), 1,
      sym__simple_expression,
    STATE(6669), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2570), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2017), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2122), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2121), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7074] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2590), 1,
      anon_sym_PERCENT,
    ACTIONS(2656), 1,
      sym__lowercase_identifier,
    ACTIONS(2658), 1,
      anon_sym_LPAREN,
    ACTIONS(2660), 1,
      anon_sym_LBRACK,
    ACTIONS(2664), 1,
      anon_sym_LBRACE,
    ACTIONS(2666), 1,
      anon_sym_object,
    ACTIONS(2668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2670), 1,
      anon_sym_new,
    ACTIONS(2672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2674), 1,
      anon_sym_begin,
    ACTIONS(2676), 1,
      sym_ocamlyacc_value,
    ACTIONS(2678), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2680), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2682), 1,
      aux_sym_number_token1,
    ACTIONS(2684), 1,
      anon_sym_SQUOTE,
    ACTIONS(2686), 1,
      anon_sym_DQUOTE,
    ACTIONS(2688), 1,
      sym_prefix_operator,
    ACTIONS(2690), 1,
      sym__uppercase_identifier,
    ACTIONS(2692), 1,
      aux_sym_tag_token1,
    STATE(1460), 1,
      sym_attribute,
    STATE(1492), 1,
      sym__attribute,
    STATE(2672), 1,
      sym__value_name,
    STATE(2712), 1,
      sym_parenthesized_operator,
    STATE(2755), 1,
      sym__constructor_name,
    STATE(2769), 1,
      sym__simple_expression,
    STATE(6631), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2662), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2757), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2713), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2682), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7200] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2590), 1,
      anon_sym_PERCENT,
    ACTIONS(2656), 1,
      sym__lowercase_identifier,
    ACTIONS(2658), 1,
      anon_sym_LPAREN,
    ACTIONS(2660), 1,
      anon_sym_LBRACK,
    ACTIONS(2664), 1,
      anon_sym_LBRACE,
    ACTIONS(2666), 1,
      anon_sym_object,
    ACTIONS(2668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2670), 1,
      anon_sym_new,
    ACTIONS(2672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2674), 1,
      anon_sym_begin,
    ACTIONS(2676), 1,
      sym_ocamlyacc_value,
    ACTIONS(2678), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2680), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2682), 1,
      aux_sym_number_token1,
    ACTIONS(2684), 1,
      anon_sym_SQUOTE,
    ACTIONS(2686), 1,
      anon_sym_DQUOTE,
    ACTIONS(2688), 1,
      sym_prefix_operator,
    ACTIONS(2690), 1,
      sym__uppercase_identifier,
    ACTIONS(2692), 1,
      aux_sym_tag_token1,
    STATE(1461), 1,
      sym_attribute,
    STATE(1493), 1,
      sym__attribute,
    STATE(2672), 1,
      sym__value_name,
    STATE(2712), 1,
      sym_parenthesized_operator,
    STATE(2755), 1,
      sym__constructor_name,
    STATE(2773), 1,
      sym__simple_expression,
    STATE(6631), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2662), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2757), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2713), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2682), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7326] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(1462), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5513), 1,
      sym__simple_expression,
    STATE(5912), 1,
      sym_field_expression,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7452] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2590), 1,
      anon_sym_PERCENT,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2696), 1,
      anon_sym_LPAREN,
    ACTIONS(2700), 1,
      anon_sym_LBRACE,
    ACTIONS(2702), 1,
      anon_sym_begin,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2708), 1,
      aux_sym_number_token1,
    ACTIONS(2710), 1,
      anon_sym_SQUOTE,
    ACTIONS(2712), 1,
      anon_sym_DQUOTE,
    ACTIONS(2714), 1,
      sym_prefix_operator,
    STATE(1463), 1,
      sym_attribute,
    STATE(1519), 1,
      sym__attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(1928), 1,
      sym__simple_expression,
    STATE(7004), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2698), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1940), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7578] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2590), 1,
      anon_sym_PERCENT,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2696), 1,
      anon_sym_LPAREN,
    ACTIONS(2700), 1,
      anon_sym_LBRACE,
    ACTIONS(2702), 1,
      anon_sym_begin,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2708), 1,
      aux_sym_number_token1,
    ACTIONS(2710), 1,
      anon_sym_SQUOTE,
    ACTIONS(2712), 1,
      anon_sym_DQUOTE,
    ACTIONS(2714), 1,
      sym_prefix_operator,
    STATE(1464), 1,
      sym_attribute,
    STATE(1485), 1,
      sym__attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(1815), 1,
      sym__simple_expression,
    STATE(7004), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2698), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1940), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7704] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2644), 1,
      anon_sym_COLON2,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2716), 1,
      anon_sym_EQ,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1465), 1,
      sym_attribute,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4945), 1,
      sym__class_typed,
    STATE(5120), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2085), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [7838] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2644), 1,
      anon_sym_COLON2,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2718), 1,
      anon_sym_EQ,
    STATE(1466), 1,
      sym_attribute,
    STATE(1470), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4995), 1,
      sym__class_typed,
    STATE(5196), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2129), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [7972] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(1467), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5500), 1,
      sym__simple_expression,
    STATE(5675), 1,
      sym_field_expression,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8098] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2644), 1,
      anon_sym_COLON2,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2720), 1,
      anon_sym_EQ,
    STATE(1465), 1,
      aux_sym_let_binding_repeat1,
    STATE(1468), 1,
      sym_attribute,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4974), 1,
      sym__class_typed,
    STATE(5183), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2033), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [8232] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2530), 1,
      sym__lowercase_identifier,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(1469), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(4670), 1,
      sym_field_path,
    STATE(5516), 1,
      sym__simple_expression,
    STATE(5881), 1,
      sym_field_expression,
    STATE(7239), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8358] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2644), 1,
      anon_sym_COLON2,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2722), 1,
      anon_sym_EQ,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1470), 1,
      sym_attribute,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5024), 1,
      sym__class_typed,
    STATE(5077), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2075), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [8492] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2644), 1,
      anon_sym_COLON2,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2724), 1,
      anon_sym_EQ,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1471), 1,
      sym_attribute,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4960), 1,
      sym__class_typed,
    STATE(5132), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2097), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [8626] = 31,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1204), 1,
      anon_sym_RPAREN,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1472), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8749] = 31,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1218), 1,
      anon_sym_RPAREN,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1473), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8872] = 31,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1224), 1,
      anon_sym_RPAREN,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1474), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8995] = 31,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1228), 1,
      anon_sym_RPAREN,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1475), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9118] = 31,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1230), 1,
      anon_sym_RPAREN,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1476), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9241] = 31,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1216), 1,
      anon_sym_RPAREN,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1477), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9364] = 31,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1208), 1,
      anon_sym_RPAREN,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1478), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9487] = 31,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1242), 1,
      anon_sym_RPAREN,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1479), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9610] = 31,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1200), 1,
      anon_sym_RPAREN,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1480), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9733] = 31,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1194), 1,
      anon_sym_RPAREN,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1481), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9856] = 31,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1214), 1,
      anon_sym_RPAREN,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1482), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9979] = 30,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1298), 1,
      anon_sym_new,
    ACTIONS(1300), 1,
      sym_prefix_operator,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(816), 1,
      sym__constructor_name,
    STATE(833), 1,
      sym__simple_expression,
    STATE(1483), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10099] = 30,
    ACTIONS(39), 1,
      anon_sym_LPAREN,
    ACTIONS(45), 1,
      anon_sym_LBRACK,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(51), 1,
      anon_sym_object,
    ACTIONS(55), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(73), 1,
      anon_sym_new,
    ACTIONS(75), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(77), 1,
      anon_sym_begin,
    ACTIONS(79), 1,
      sym_ocamlyacc_value,
    ACTIONS(87), 1,
      aux_sym_number_token1,
    ACTIONS(91), 1,
      anon_sym_SQUOTE,
    ACTIONS(93), 1,
      anon_sym_DQUOTE,
    ACTIONS(95), 1,
      sym_prefix_operator,
    ACTIONS(101), 1,
      sym__uppercase_identifier,
    ACTIONS(103), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(681), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(683), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(751), 1,
      sym__lowercase_identifier,
    STATE(892), 1,
      sym__constructor_name,
    STATE(896), 1,
      sym__simple_expression,
    STATE(908), 1,
      sym_parenthesized_operator,
    STATE(936), 1,
      sym__value_name,
    STATE(1484), 1,
      sym_attribute,
    STATE(6591), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(47), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(924), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(922), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(939), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10219] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2696), 1,
      anon_sym_LPAREN,
    ACTIONS(2700), 1,
      anon_sym_LBRACE,
    ACTIONS(2702), 1,
      anon_sym_begin,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2708), 1,
      aux_sym_number_token1,
    ACTIONS(2710), 1,
      anon_sym_SQUOTE,
    ACTIONS(2712), 1,
      anon_sym_DQUOTE,
    ACTIONS(2714), 1,
      sym_prefix_operator,
    STATE(1485), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(1845), 1,
      sym__simple_expression,
    STATE(7004), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2698), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1940), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10339] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2604), 1,
      sym__lowercase_identifier,
    ACTIONS(2606), 1,
      anon_sym_LPAREN,
    ACTIONS(2608), 1,
      anon_sym_LBRACK,
    ACTIONS(2612), 1,
      anon_sym_LBRACE,
    ACTIONS(2614), 1,
      anon_sym_object,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2618), 1,
      anon_sym_new,
    ACTIONS(2620), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2622), 1,
      anon_sym_begin,
    ACTIONS(2624), 1,
      sym_ocamlyacc_value,
    ACTIONS(2626), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2628), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2630), 1,
      aux_sym_number_token1,
    ACTIONS(2632), 1,
      anon_sym_SQUOTE,
    ACTIONS(2634), 1,
      anon_sym_DQUOTE,
    ACTIONS(2636), 1,
      sym_prefix_operator,
    ACTIONS(2638), 1,
      sym__uppercase_identifier,
    ACTIONS(2640), 1,
      aux_sym_tag_token1,
    STATE(1486), 1,
      sym_attribute,
    STATE(1807), 1,
      sym__constructor_name,
    STATE(1832), 1,
      sym__simple_expression,
    STATE(1877), 1,
      sym_parenthesized_operator,
    STATE(1893), 1,
      sym__value_name,
    STATE(7240), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2610), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1891), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1926), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1921), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10459] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2564), 1,
      sym__lowercase_identifier,
    ACTIONS(2566), 1,
      anon_sym_LPAREN,
    ACTIONS(2568), 1,
      anon_sym_LBRACK,
    ACTIONS(2572), 1,
      anon_sym_LBRACE,
    ACTIONS(2574), 1,
      anon_sym_object,
    ACTIONS(2576), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2578), 1,
      anon_sym_new,
    ACTIONS(2580), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2582), 1,
      anon_sym_begin,
    ACTIONS(2584), 1,
      sym_ocamlyacc_value,
    ACTIONS(2586), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2588), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2592), 1,
      aux_sym_number_token1,
    ACTIONS(2594), 1,
      anon_sym_SQUOTE,
    ACTIONS(2596), 1,
      anon_sym_DQUOTE,
    ACTIONS(2598), 1,
      sym_prefix_operator,
    ACTIONS(2600), 1,
      sym__uppercase_identifier,
    ACTIONS(2602), 1,
      aux_sym_tag_token1,
    STATE(1487), 1,
      sym_attribute,
    STATE(1981), 1,
      sym__value_name,
    STATE(2005), 1,
      sym_parenthesized_operator,
    STATE(2030), 1,
      sym__constructor_name,
    STATE(2061), 1,
      sym__simple_expression,
    STATE(6669), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2570), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2017), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2122), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2121), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10579] = 30,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(816), 1,
      sym__constructor_name,
    STATE(870), 1,
      sym__simple_expression,
    STATE(1488), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10699] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2696), 1,
      anon_sym_LPAREN,
    ACTIONS(2700), 1,
      anon_sym_LBRACE,
    ACTIONS(2702), 1,
      anon_sym_begin,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2708), 1,
      aux_sym_number_token1,
    ACTIONS(2710), 1,
      anon_sym_SQUOTE,
    ACTIONS(2712), 1,
      anon_sym_DQUOTE,
    ACTIONS(2714), 1,
      sym_prefix_operator,
    STATE(1489), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(1922), 1,
      sym__simple_expression,
    STATE(7004), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2698), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1940), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10819] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(607), 1,
      anon_sym_LPAREN,
    ACTIONS(609), 1,
      anon_sym_LBRACK,
    ACTIONS(613), 1,
      anon_sym_LBRACE,
    ACTIONS(615), 1,
      anon_sym_object,
    ACTIONS(619), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(637), 1,
      anon_sym_new,
    ACTIONS(639), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(641), 1,
      anon_sym_begin,
    ACTIONS(643), 1,
      sym_ocamlyacc_value,
    ACTIONS(645), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(647), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(649), 1,
      aux_sym_number_token1,
    ACTIONS(651), 1,
      anon_sym_SQUOTE,
    ACTIONS(653), 1,
      anon_sym_DQUOTE,
    ACTIONS(655), 1,
      sym_prefix_operator,
    ACTIONS(659), 1,
      sym__uppercase_identifier,
    ACTIONS(661), 1,
      aux_sym_tag_token1,
    ACTIONS(1304), 1,
      sym__lowercase_identifier,
    STATE(1490), 1,
      sym_attribute,
    STATE(1567), 1,
      sym__constructor_name,
    STATE(1607), 1,
      sym_parenthesized_operator,
    STATE(1612), 1,
      sym__value_name,
    STATE(1624), 1,
      sym__simple_expression,
    STATE(6548), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(611), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1566), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1589), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1626), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10939] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2656), 1,
      sym__lowercase_identifier,
    ACTIONS(2658), 1,
      anon_sym_LPAREN,
    ACTIONS(2660), 1,
      anon_sym_LBRACK,
    ACTIONS(2664), 1,
      anon_sym_LBRACE,
    ACTIONS(2666), 1,
      anon_sym_object,
    ACTIONS(2668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2670), 1,
      anon_sym_new,
    ACTIONS(2672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2674), 1,
      anon_sym_begin,
    ACTIONS(2676), 1,
      sym_ocamlyacc_value,
    ACTIONS(2678), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2680), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2682), 1,
      aux_sym_number_token1,
    ACTIONS(2684), 1,
      anon_sym_SQUOTE,
    ACTIONS(2686), 1,
      anon_sym_DQUOTE,
    ACTIONS(2688), 1,
      sym_prefix_operator,
    ACTIONS(2690), 1,
      sym__uppercase_identifier,
    ACTIONS(2692), 1,
      aux_sym_tag_token1,
    STATE(1491), 1,
      sym_attribute,
    STATE(2672), 1,
      sym__value_name,
    STATE(2712), 1,
      sym_parenthesized_operator,
    STATE(2755), 1,
      sym__constructor_name,
    STATE(2767), 1,
      sym__simple_expression,
    STATE(6631), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2662), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2757), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2713), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2682), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11059] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2656), 1,
      sym__lowercase_identifier,
    ACTIONS(2658), 1,
      anon_sym_LPAREN,
    ACTIONS(2660), 1,
      anon_sym_LBRACK,
    ACTIONS(2664), 1,
      anon_sym_LBRACE,
    ACTIONS(2666), 1,
      anon_sym_object,
    ACTIONS(2668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2670), 1,
      anon_sym_new,
    ACTIONS(2672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2674), 1,
      anon_sym_begin,
    ACTIONS(2676), 1,
      sym_ocamlyacc_value,
    ACTIONS(2678), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2680), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2682), 1,
      aux_sym_number_token1,
    ACTIONS(2684), 1,
      anon_sym_SQUOTE,
    ACTIONS(2686), 1,
      anon_sym_DQUOTE,
    ACTIONS(2688), 1,
      sym_prefix_operator,
    ACTIONS(2690), 1,
      sym__uppercase_identifier,
    ACTIONS(2692), 1,
      aux_sym_tag_token1,
    STATE(1492), 1,
      sym_attribute,
    STATE(2672), 1,
      sym__value_name,
    STATE(2712), 1,
      sym_parenthesized_operator,
    STATE(2726), 1,
      sym__simple_expression,
    STATE(2755), 1,
      sym__constructor_name,
    STATE(6631), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2662), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2757), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2713), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2682), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11179] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2656), 1,
      sym__lowercase_identifier,
    ACTIONS(2658), 1,
      anon_sym_LPAREN,
    ACTIONS(2660), 1,
      anon_sym_LBRACK,
    ACTIONS(2664), 1,
      anon_sym_LBRACE,
    ACTIONS(2666), 1,
      anon_sym_object,
    ACTIONS(2668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2670), 1,
      anon_sym_new,
    ACTIONS(2672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2674), 1,
      anon_sym_begin,
    ACTIONS(2676), 1,
      sym_ocamlyacc_value,
    ACTIONS(2678), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2680), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2682), 1,
      aux_sym_number_token1,
    ACTIONS(2684), 1,
      anon_sym_SQUOTE,
    ACTIONS(2686), 1,
      anon_sym_DQUOTE,
    ACTIONS(2688), 1,
      sym_prefix_operator,
    ACTIONS(2690), 1,
      sym__uppercase_identifier,
    ACTIONS(2692), 1,
      aux_sym_tag_token1,
    STATE(1493), 1,
      sym_attribute,
    STATE(2672), 1,
      sym__value_name,
    STATE(2712), 1,
      sym_parenthesized_operator,
    STATE(2733), 1,
      sym__simple_expression,
    STATE(2755), 1,
      sym__constructor_name,
    STATE(6631), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2662), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2757), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2713), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2682), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11299] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2656), 1,
      sym__lowercase_identifier,
    ACTIONS(2658), 1,
      anon_sym_LPAREN,
    ACTIONS(2660), 1,
      anon_sym_LBRACK,
    ACTIONS(2664), 1,
      anon_sym_LBRACE,
    ACTIONS(2666), 1,
      anon_sym_object,
    ACTIONS(2668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2670), 1,
      anon_sym_new,
    ACTIONS(2672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2674), 1,
      anon_sym_begin,
    ACTIONS(2676), 1,
      sym_ocamlyacc_value,
    ACTIONS(2678), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2680), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2682), 1,
      aux_sym_number_token1,
    ACTIONS(2684), 1,
      anon_sym_SQUOTE,
    ACTIONS(2686), 1,
      anon_sym_DQUOTE,
    ACTIONS(2688), 1,
      sym_prefix_operator,
    ACTIONS(2690), 1,
      sym__uppercase_identifier,
    ACTIONS(2692), 1,
      aux_sym_tag_token1,
    STATE(1494), 1,
      sym_attribute,
    STATE(2672), 1,
      sym__value_name,
    STATE(2712), 1,
      sym_parenthesized_operator,
    STATE(2725), 1,
      sym__simple_expression,
    STATE(2755), 1,
      sym__constructor_name,
    STATE(6631), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2662), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2757), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2713), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2682), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11419] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    STATE(1495), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(5135), 1,
      sym__simple_expression,
    STATE(7115), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11539] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1496), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2728), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2726), 41,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [11609] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2656), 1,
      sym__lowercase_identifier,
    ACTIONS(2658), 1,
      anon_sym_LPAREN,
    ACTIONS(2660), 1,
      anon_sym_LBRACK,
    ACTIONS(2664), 1,
      anon_sym_LBRACE,
    ACTIONS(2666), 1,
      anon_sym_object,
    ACTIONS(2668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2670), 1,
      anon_sym_new,
    ACTIONS(2672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2674), 1,
      anon_sym_begin,
    ACTIONS(2676), 1,
      sym_ocamlyacc_value,
    ACTIONS(2678), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2680), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2682), 1,
      aux_sym_number_token1,
    ACTIONS(2684), 1,
      anon_sym_SQUOTE,
    ACTIONS(2686), 1,
      anon_sym_DQUOTE,
    ACTIONS(2688), 1,
      sym_prefix_operator,
    ACTIONS(2690), 1,
      sym__uppercase_identifier,
    ACTIONS(2692), 1,
      aux_sym_tag_token1,
    STATE(1497), 1,
      sym_attribute,
    STATE(2672), 1,
      sym__value_name,
    STATE(2712), 1,
      sym_parenthesized_operator,
    STATE(2755), 1,
      sym__constructor_name,
    STATE(2764), 1,
      sym__simple_expression,
    STATE(6631), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2662), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2757), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2713), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2682), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11729] = 30,
    ACTIONS(183), 1,
      anon_sym_LPAREN,
    ACTIONS(185), 1,
      anon_sym_LBRACK,
    ACTIONS(189), 1,
      anon_sym_LBRACE,
    ACTIONS(191), 1,
      anon_sym_object,
    ACTIONS(195), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(205), 1,
      anon_sym_new,
    ACTIONS(207), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(209), 1,
      anon_sym_begin,
    ACTIONS(211), 1,
      sym_ocamlyacc_value,
    ACTIONS(217), 1,
      aux_sym_number_token1,
    ACTIONS(219), 1,
      anon_sym_SQUOTE,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(223), 1,
      sym_prefix_operator,
    ACTIONS(227), 1,
      sym__uppercase_identifier,
    ACTIONS(229), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(959), 1,
      sym__lowercase_identifier,
    ACTIONS(1184), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1186), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1078), 1,
      sym__value_name,
    STATE(1088), 1,
      sym__simple_expression,
    STATE(1112), 1,
      sym__constructor_name,
    STATE(1115), 1,
      sym_parenthesized_operator,
    STATE(1498), 1,
      sym_attribute,
    STATE(7251), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(187), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1070), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1113), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1131), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11849] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2564), 1,
      sym__lowercase_identifier,
    ACTIONS(2566), 1,
      anon_sym_LPAREN,
    ACTIONS(2568), 1,
      anon_sym_LBRACK,
    ACTIONS(2572), 1,
      anon_sym_LBRACE,
    ACTIONS(2574), 1,
      anon_sym_object,
    ACTIONS(2576), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2578), 1,
      anon_sym_new,
    ACTIONS(2580), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2582), 1,
      anon_sym_begin,
    ACTIONS(2584), 1,
      sym_ocamlyacc_value,
    ACTIONS(2586), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2588), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2592), 1,
      aux_sym_number_token1,
    ACTIONS(2594), 1,
      anon_sym_SQUOTE,
    ACTIONS(2596), 1,
      anon_sym_DQUOTE,
    ACTIONS(2598), 1,
      sym_prefix_operator,
    ACTIONS(2600), 1,
      sym__uppercase_identifier,
    ACTIONS(2602), 1,
      aux_sym_tag_token1,
    STATE(1499), 1,
      sym_attribute,
    STATE(1981), 1,
      sym__value_name,
    STATE(2005), 1,
      sym_parenthesized_operator,
    STATE(2030), 1,
      sym__constructor_name,
    STATE(2062), 1,
      sym__simple_expression,
    STATE(6669), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2570), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2017), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2122), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2121), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11969] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(607), 1,
      anon_sym_LPAREN,
    ACTIONS(609), 1,
      anon_sym_LBRACK,
    ACTIONS(613), 1,
      anon_sym_LBRACE,
    ACTIONS(615), 1,
      anon_sym_object,
    ACTIONS(619), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(637), 1,
      anon_sym_new,
    ACTIONS(639), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(641), 1,
      anon_sym_begin,
    ACTIONS(643), 1,
      sym_ocamlyacc_value,
    ACTIONS(645), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(647), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(649), 1,
      aux_sym_number_token1,
    ACTIONS(651), 1,
      anon_sym_SQUOTE,
    ACTIONS(653), 1,
      anon_sym_DQUOTE,
    ACTIONS(655), 1,
      sym_prefix_operator,
    ACTIONS(659), 1,
      sym__uppercase_identifier,
    ACTIONS(661), 1,
      aux_sym_tag_token1,
    ACTIONS(1304), 1,
      sym__lowercase_identifier,
    STATE(1500), 1,
      sym_attribute,
    STATE(1567), 1,
      sym__constructor_name,
    STATE(1599), 1,
      sym__simple_expression,
    STATE(1607), 1,
      sym_parenthesized_operator,
    STATE(1612), 1,
      sym__value_name,
    STATE(6548), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(611), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1566), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1589), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1626), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12089] = 30,
    ACTIONS(183), 1,
      anon_sym_LPAREN,
    ACTIONS(185), 1,
      anon_sym_LBRACK,
    ACTIONS(189), 1,
      anon_sym_LBRACE,
    ACTIONS(191), 1,
      anon_sym_object,
    ACTIONS(195), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(205), 1,
      anon_sym_new,
    ACTIONS(207), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(209), 1,
      anon_sym_begin,
    ACTIONS(211), 1,
      sym_ocamlyacc_value,
    ACTIONS(217), 1,
      aux_sym_number_token1,
    ACTIONS(219), 1,
      anon_sym_SQUOTE,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(223), 1,
      sym_prefix_operator,
    ACTIONS(227), 1,
      sym__uppercase_identifier,
    ACTIONS(229), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(959), 1,
      sym__lowercase_identifier,
    ACTIONS(1184), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1186), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1078), 1,
      sym__value_name,
    STATE(1112), 1,
      sym__constructor_name,
    STATE(1115), 1,
      sym_parenthesized_operator,
    STATE(1136), 1,
      sym__simple_expression,
    STATE(1501), 1,
      sym_attribute,
    STATE(7251), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(187), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1070), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1113), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1131), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12209] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2696), 1,
      anon_sym_LPAREN,
    ACTIONS(2700), 1,
      anon_sym_LBRACE,
    ACTIONS(2702), 1,
      anon_sym_begin,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2708), 1,
      aux_sym_number_token1,
    ACTIONS(2710), 1,
      anon_sym_SQUOTE,
    ACTIONS(2712), 1,
      anon_sym_DQUOTE,
    ACTIONS(2714), 1,
      sym_prefix_operator,
    STATE(1502), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1757), 1,
      sym__simple_expression,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(7004), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2698), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1940), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12329] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2732), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2730), 41,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [12399] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2696), 1,
      anon_sym_LPAREN,
    ACTIONS(2700), 1,
      anon_sym_LBRACE,
    ACTIONS(2702), 1,
      anon_sym_begin,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2708), 1,
      aux_sym_number_token1,
    ACTIONS(2710), 1,
      anon_sym_SQUOTE,
    ACTIONS(2712), 1,
      anon_sym_DQUOTE,
    ACTIONS(2714), 1,
      sym_prefix_operator,
    STATE(1504), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(1818), 1,
      sym__simple_expression,
    STATE(7004), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2698), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1940), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12519] = 30,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1298), 1,
      anon_sym_new,
    ACTIONS(1300), 1,
      sym_prefix_operator,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1252), 1,
      sym__simple_expression,
    STATE(1505), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12639] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2604), 1,
      sym__lowercase_identifier,
    ACTIONS(2606), 1,
      anon_sym_LPAREN,
    ACTIONS(2608), 1,
      anon_sym_LBRACK,
    ACTIONS(2612), 1,
      anon_sym_LBRACE,
    ACTIONS(2614), 1,
      anon_sym_object,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2618), 1,
      anon_sym_new,
    ACTIONS(2620), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2622), 1,
      anon_sym_begin,
    ACTIONS(2624), 1,
      sym_ocamlyacc_value,
    ACTIONS(2626), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2628), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2630), 1,
      aux_sym_number_token1,
    ACTIONS(2632), 1,
      anon_sym_SQUOTE,
    ACTIONS(2634), 1,
      anon_sym_DQUOTE,
    ACTIONS(2636), 1,
      sym_prefix_operator,
    ACTIONS(2638), 1,
      sym__uppercase_identifier,
    ACTIONS(2640), 1,
      aux_sym_tag_token1,
    STATE(1506), 1,
      sym_attribute,
    STATE(1807), 1,
      sym__constructor_name,
    STATE(1875), 1,
      sym__simple_expression,
    STATE(1877), 1,
      sym_parenthesized_operator,
    STATE(1893), 1,
      sym__value_name,
    STATE(7240), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2610), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1891), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1926), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1921), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12759] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2736), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2734), 41,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [12829] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2564), 1,
      sym__lowercase_identifier,
    ACTIONS(2566), 1,
      anon_sym_LPAREN,
    ACTIONS(2568), 1,
      anon_sym_LBRACK,
    ACTIONS(2572), 1,
      anon_sym_LBRACE,
    ACTIONS(2574), 1,
      anon_sym_object,
    ACTIONS(2576), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2578), 1,
      anon_sym_new,
    ACTIONS(2580), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2582), 1,
      anon_sym_begin,
    ACTIONS(2584), 1,
      sym_ocamlyacc_value,
    ACTIONS(2586), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2588), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2592), 1,
      aux_sym_number_token1,
    ACTIONS(2594), 1,
      anon_sym_SQUOTE,
    ACTIONS(2596), 1,
      anon_sym_DQUOTE,
    ACTIONS(2598), 1,
      sym_prefix_operator,
    ACTIONS(2600), 1,
      sym__uppercase_identifier,
    ACTIONS(2602), 1,
      aux_sym_tag_token1,
    STATE(1508), 1,
      sym_attribute,
    STATE(1959), 1,
      sym__simple_expression,
    STATE(1981), 1,
      sym__value_name,
    STATE(2005), 1,
      sym_parenthesized_operator,
    STATE(2030), 1,
      sym__constructor_name,
    STATE(6669), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2570), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2017), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2122), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2121), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12949] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2564), 1,
      sym__lowercase_identifier,
    ACTIONS(2566), 1,
      anon_sym_LPAREN,
    ACTIONS(2568), 1,
      anon_sym_LBRACK,
    ACTIONS(2572), 1,
      anon_sym_LBRACE,
    ACTIONS(2574), 1,
      anon_sym_object,
    ACTIONS(2576), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2578), 1,
      anon_sym_new,
    ACTIONS(2580), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2582), 1,
      anon_sym_begin,
    ACTIONS(2584), 1,
      sym_ocamlyacc_value,
    ACTIONS(2586), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2588), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2592), 1,
      aux_sym_number_token1,
    ACTIONS(2594), 1,
      anon_sym_SQUOTE,
    ACTIONS(2596), 1,
      anon_sym_DQUOTE,
    ACTIONS(2598), 1,
      sym_prefix_operator,
    ACTIONS(2600), 1,
      sym__uppercase_identifier,
    ACTIONS(2602), 1,
      aux_sym_tag_token1,
    STATE(1509), 1,
      sym_attribute,
    STATE(1950), 1,
      sym__simple_expression,
    STATE(1981), 1,
      sym__value_name,
    STATE(2005), 1,
      sym_parenthesized_operator,
    STATE(2030), 1,
      sym__constructor_name,
    STATE(6669), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2570), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2017), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2122), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2121), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13069] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1654), 1,
      sym__lowercase_identifier,
    ACTIONS(1658), 1,
      anon_sym_LPAREN,
    ACTIONS(1660), 1,
      anon_sym_LBRACK,
    ACTIONS(1664), 1,
      anon_sym_LBRACE,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(1668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1670), 1,
      anon_sym_new,
    ACTIONS(1672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(1674), 1,
      anon_sym_begin,
    ACTIONS(1676), 1,
      sym_ocamlyacc_value,
    ACTIONS(1682), 1,
      aux_sym_number_token1,
    ACTIONS(1684), 1,
      anon_sym_SQUOTE,
    ACTIONS(1686), 1,
      anon_sym_DQUOTE,
    ACTIONS(1688), 1,
      sym_prefix_operator,
    ACTIONS(1690), 1,
      sym__uppercase_identifier,
    ACTIONS(1692), 1,
      aux_sym_tag_token1,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1510), 1,
      sym_attribute,
    STATE(2372), 1,
      sym__simple_expression,
    STATE(2402), 1,
      sym__value_name,
    STATE(2406), 1,
      sym_parenthesized_operator,
    STATE(2419), 1,
      sym__constructor_name,
    STATE(6407), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1662), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2348), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2351), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2350), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13189] = 30,
    ACTIONS(109), 1,
      anon_sym_LPAREN,
    ACTIONS(111), 1,
      anon_sym_LBRACK,
    ACTIONS(115), 1,
      anon_sym_LBRACE,
    ACTIONS(117), 1,
      anon_sym_object,
    ACTIONS(121), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(139), 1,
      anon_sym_new,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(145), 1,
      sym_ocamlyacc_value,
    ACTIONS(151), 1,
      aux_sym_number_token1,
    ACTIONS(153), 1,
      anon_sym_SQUOTE,
    ACTIONS(155), 1,
      anon_sym_DQUOTE,
    ACTIONS(157), 1,
      sym_prefix_operator,
    ACTIONS(161), 1,
      sym__uppercase_identifier,
    ACTIONS(163), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(761), 1,
      sym__lowercase_identifier,
    ACTIONS(1176), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1178), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(988), 1,
      sym__simple_expression,
    STATE(992), 1,
      sym_parenthesized_operator,
    STATE(1047), 1,
      sym__value_name,
    STATE(1050), 1,
      sym__constructor_name,
    STATE(1511), 1,
      sym_attribute,
    STATE(6502), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(113), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1003), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(991), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(990), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13309] = 30,
    ACTIONS(183), 1,
      anon_sym_LPAREN,
    ACTIONS(185), 1,
      anon_sym_LBRACK,
    ACTIONS(189), 1,
      anon_sym_LBRACE,
    ACTIONS(191), 1,
      anon_sym_object,
    ACTIONS(195), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(205), 1,
      anon_sym_new,
    ACTIONS(207), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(209), 1,
      anon_sym_begin,
    ACTIONS(211), 1,
      sym_ocamlyacc_value,
    ACTIONS(217), 1,
      aux_sym_number_token1,
    ACTIONS(219), 1,
      anon_sym_SQUOTE,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(223), 1,
      sym_prefix_operator,
    ACTIONS(227), 1,
      sym__uppercase_identifier,
    ACTIONS(229), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(959), 1,
      sym__lowercase_identifier,
    ACTIONS(1184), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1186), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1078), 1,
      sym__value_name,
    STATE(1112), 1,
      sym__constructor_name,
    STATE(1115), 1,
      sym_parenthesized_operator,
    STATE(1117), 1,
      sym__simple_expression,
    STATE(1512), 1,
      sym_attribute,
    STATE(7251), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(187), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1070), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1113), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1131), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13429] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    STATE(1513), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1757), 1,
      sym__simple_expression,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(7115), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13549] = 30,
    ACTIONS(109), 1,
      anon_sym_LPAREN,
    ACTIONS(111), 1,
      anon_sym_LBRACK,
    ACTIONS(115), 1,
      anon_sym_LBRACE,
    ACTIONS(117), 1,
      anon_sym_object,
    ACTIONS(121), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(139), 1,
      anon_sym_new,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(145), 1,
      sym_ocamlyacc_value,
    ACTIONS(151), 1,
      aux_sym_number_token1,
    ACTIONS(153), 1,
      anon_sym_SQUOTE,
    ACTIONS(155), 1,
      anon_sym_DQUOTE,
    ACTIONS(157), 1,
      sym_prefix_operator,
    ACTIONS(161), 1,
      sym__uppercase_identifier,
    ACTIONS(163), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(761), 1,
      sym__lowercase_identifier,
    ACTIONS(1176), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1178), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(992), 1,
      sym_parenthesized_operator,
    STATE(999), 1,
      sym__simple_expression,
    STATE(1047), 1,
      sym__value_name,
    STATE(1050), 1,
      sym__constructor_name,
    STATE(1514), 1,
      sym_attribute,
    STATE(6502), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(113), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1003), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(991), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(990), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13669] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2604), 1,
      sym__lowercase_identifier,
    ACTIONS(2606), 1,
      anon_sym_LPAREN,
    ACTIONS(2608), 1,
      anon_sym_LBRACK,
    ACTIONS(2612), 1,
      anon_sym_LBRACE,
    ACTIONS(2614), 1,
      anon_sym_object,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2618), 1,
      anon_sym_new,
    ACTIONS(2620), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2622), 1,
      anon_sym_begin,
    ACTIONS(2624), 1,
      sym_ocamlyacc_value,
    ACTIONS(2626), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2628), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2630), 1,
      aux_sym_number_token1,
    ACTIONS(2632), 1,
      anon_sym_SQUOTE,
    ACTIONS(2634), 1,
      anon_sym_DQUOTE,
    ACTIONS(2636), 1,
      sym_prefix_operator,
    ACTIONS(2638), 1,
      sym__uppercase_identifier,
    ACTIONS(2640), 1,
      aux_sym_tag_token1,
    STATE(1515), 1,
      sym_attribute,
    STATE(1807), 1,
      sym__constructor_name,
    STATE(1814), 1,
      sym__simple_expression,
    STATE(1877), 1,
      sym_parenthesized_operator,
    STATE(1893), 1,
      sym__value_name,
    STATE(7240), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2610), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1891), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1926), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1921), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13789] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2532), 1,
      anon_sym_LPAREN,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2536), 1,
      anon_sym_LBRACE,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2550), 1,
      sym_prefix_operator,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    STATE(1516), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(5380), 1,
      sym__simple_expression,
    STATE(7115), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13909] = 30,
    ACTIONS(109), 1,
      anon_sym_LPAREN,
    ACTIONS(111), 1,
      anon_sym_LBRACK,
    ACTIONS(115), 1,
      anon_sym_LBRACE,
    ACTIONS(117), 1,
      anon_sym_object,
    ACTIONS(121), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(139), 1,
      anon_sym_new,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(145), 1,
      sym_ocamlyacc_value,
    ACTIONS(151), 1,
      aux_sym_number_token1,
    ACTIONS(153), 1,
      anon_sym_SQUOTE,
    ACTIONS(155), 1,
      anon_sym_DQUOTE,
    ACTIONS(157), 1,
      sym_prefix_operator,
    ACTIONS(161), 1,
      sym__uppercase_identifier,
    ACTIONS(163), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(761), 1,
      sym__lowercase_identifier,
    ACTIONS(1176), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1178), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(992), 1,
      sym_parenthesized_operator,
    STATE(1015), 1,
      sym__simple_expression,
    STATE(1047), 1,
      sym__value_name,
    STATE(1050), 1,
      sym__constructor_name,
    STATE(1517), 1,
      sym_attribute,
    STATE(6502), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(113), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1003), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(991), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(990), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14029] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1654), 1,
      sym__lowercase_identifier,
    ACTIONS(1658), 1,
      anon_sym_LPAREN,
    ACTIONS(1660), 1,
      anon_sym_LBRACK,
    ACTIONS(1664), 1,
      anon_sym_LBRACE,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(1668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1670), 1,
      anon_sym_new,
    ACTIONS(1672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(1674), 1,
      anon_sym_begin,
    ACTIONS(1676), 1,
      sym_ocamlyacc_value,
    ACTIONS(1682), 1,
      aux_sym_number_token1,
    ACTIONS(1684), 1,
      anon_sym_SQUOTE,
    ACTIONS(1686), 1,
      anon_sym_DQUOTE,
    ACTIONS(1688), 1,
      sym_prefix_operator,
    ACTIONS(1690), 1,
      sym__uppercase_identifier,
    ACTIONS(1692), 1,
      aux_sym_tag_token1,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1518), 1,
      sym_attribute,
    STATE(2345), 1,
      sym__simple_expression,
    STATE(2402), 1,
      sym__value_name,
    STATE(2406), 1,
      sym_parenthesized_operator,
    STATE(2419), 1,
      sym__constructor_name,
    STATE(6407), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1662), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2348), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2351), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2350), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14149] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2538), 1,
      anon_sym_object,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2542), 1,
      anon_sym_new,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2548), 1,
      sym_ocamlyacc_value,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2696), 1,
      anon_sym_LPAREN,
    ACTIONS(2700), 1,
      anon_sym_LBRACE,
    ACTIONS(2702), 1,
      anon_sym_begin,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2708), 1,
      aux_sym_number_token1,
    ACTIONS(2710), 1,
      anon_sym_SQUOTE,
    ACTIONS(2712), 1,
      anon_sym_DQUOTE,
    ACTIONS(2714), 1,
      sym_prefix_operator,
    STATE(1519), 1,
      sym_attribute,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1734), 1,
      sym__value_name,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(1937), 1,
      sym__simple_expression,
    STATE(7004), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2698), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1940), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14269] = 30,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1520), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14389] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2564), 1,
      sym__lowercase_identifier,
    ACTIONS(2566), 1,
      anon_sym_LPAREN,
    ACTIONS(2568), 1,
      anon_sym_LBRACK,
    ACTIONS(2572), 1,
      anon_sym_LBRACE,
    ACTIONS(2574), 1,
      anon_sym_object,
    ACTIONS(2576), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2578), 1,
      anon_sym_new,
    ACTIONS(2580), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2582), 1,
      anon_sym_begin,
    ACTIONS(2584), 1,
      sym_ocamlyacc_value,
    ACTIONS(2586), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2588), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2592), 1,
      aux_sym_number_token1,
    ACTIONS(2594), 1,
      anon_sym_SQUOTE,
    ACTIONS(2596), 1,
      anon_sym_DQUOTE,
    ACTIONS(2598), 1,
      sym_prefix_operator,
    ACTIONS(2600), 1,
      sym__uppercase_identifier,
    ACTIONS(2602), 1,
      aux_sym_tag_token1,
    STATE(1521), 1,
      sym_attribute,
    STATE(1981), 1,
      sym__value_name,
    STATE(2005), 1,
      sym_parenthesized_operator,
    STATE(2030), 1,
      sym__constructor_name,
    STATE(2068), 1,
      sym__simple_expression,
    STATE(6669), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2570), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2017), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2122), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2121), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14509] = 30,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(1298), 1,
      anon_sym_new,
    ACTIONS(1300), 1,
      sym_prefix_operator,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(812), 1,
      sym__simple_expression,
    STATE(816), 1,
      sym__constructor_name,
    STATE(1522), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14629] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2604), 1,
      sym__lowercase_identifier,
    ACTIONS(2606), 1,
      anon_sym_LPAREN,
    ACTIONS(2608), 1,
      anon_sym_LBRACK,
    ACTIONS(2612), 1,
      anon_sym_LBRACE,
    ACTIONS(2614), 1,
      anon_sym_object,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2618), 1,
      anon_sym_new,
    ACTIONS(2620), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2622), 1,
      anon_sym_begin,
    ACTIONS(2624), 1,
      sym_ocamlyacc_value,
    ACTIONS(2626), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2628), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2630), 1,
      aux_sym_number_token1,
    ACTIONS(2632), 1,
      anon_sym_SQUOTE,
    ACTIONS(2634), 1,
      anon_sym_DQUOTE,
    ACTIONS(2636), 1,
      sym_prefix_operator,
    ACTIONS(2638), 1,
      sym__uppercase_identifier,
    ACTIONS(2640), 1,
      aux_sym_tag_token1,
    STATE(1523), 1,
      sym_attribute,
    STATE(1807), 1,
      sym__constructor_name,
    STATE(1877), 1,
      sym_parenthesized_operator,
    STATE(1893), 1,
      sym__value_name,
    STATE(1907), 1,
      sym__simple_expression,
    STATE(7240), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2610), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1891), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1926), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1921), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14749] = 30,
    ACTIONS(39), 1,
      anon_sym_LPAREN,
    ACTIONS(45), 1,
      anon_sym_LBRACK,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(51), 1,
      anon_sym_object,
    ACTIONS(55), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(73), 1,
      anon_sym_new,
    ACTIONS(75), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(77), 1,
      anon_sym_begin,
    ACTIONS(79), 1,
      sym_ocamlyacc_value,
    ACTIONS(87), 1,
      aux_sym_number_token1,
    ACTIONS(91), 1,
      anon_sym_SQUOTE,
    ACTIONS(93), 1,
      anon_sym_DQUOTE,
    ACTIONS(95), 1,
      sym_prefix_operator,
    ACTIONS(101), 1,
      sym__uppercase_identifier,
    ACTIONS(103), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(681), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(683), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(751), 1,
      sym__lowercase_identifier,
    STATE(892), 1,
      sym__constructor_name,
    STATE(906), 1,
      sym__simple_expression,
    STATE(908), 1,
      sym_parenthesized_operator,
    STATE(936), 1,
      sym__value_name,
    STATE(1524), 1,
      sym_attribute,
    STATE(6591), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(47), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(924), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(922), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(939), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14869] = 30,
    ACTIONS(343), 1,
      anon_sym_LPAREN,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(349), 1,
      anon_sym_LBRACE,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(373), 1,
      anon_sym_new,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(379), 1,
      sym_ocamlyacc_value,
    ACTIONS(385), 1,
      aux_sym_number_token1,
    ACTIONS(387), 1,
      anon_sym_SQUOTE,
    ACTIONS(389), 1,
      anon_sym_DQUOTE,
    ACTIONS(391), 1,
      sym_prefix_operator,
    ACTIONS(395), 1,
      sym__uppercase_identifier,
    ACTIONS(397), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(801), 1,
      sym__value_name,
    STATE(816), 1,
      sym__constructor_name,
    STATE(833), 1,
      sym__simple_expression,
    STATE(1525), 1,
      sym_attribute,
    STATE(6347), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(347), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(790), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(803), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14989] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(607), 1,
      anon_sym_LPAREN,
    ACTIONS(609), 1,
      anon_sym_LBRACK,
    ACTIONS(613), 1,
      anon_sym_LBRACE,
    ACTIONS(615), 1,
      anon_sym_object,
    ACTIONS(619), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(637), 1,
      anon_sym_new,
    ACTIONS(639), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(641), 1,
      anon_sym_begin,
    ACTIONS(643), 1,
      sym_ocamlyacc_value,
    ACTIONS(645), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(647), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(649), 1,
      aux_sym_number_token1,
    ACTIONS(651), 1,
      anon_sym_SQUOTE,
    ACTIONS(653), 1,
      anon_sym_DQUOTE,
    ACTIONS(655), 1,
      sym_prefix_operator,
    ACTIONS(659), 1,
      sym__uppercase_identifier,
    ACTIONS(661), 1,
      aux_sym_tag_token1,
    ACTIONS(1304), 1,
      sym__lowercase_identifier,
    STATE(1526), 1,
      sym_attribute,
    STATE(1567), 1,
      sym__constructor_name,
    STATE(1607), 1,
      sym_parenthesized_operator,
    STATE(1612), 1,
      sym__value_name,
    STATE(1615), 1,
      sym__simple_expression,
    STATE(6548), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(611), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1566), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1589), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1626), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15109] = 30,
    ACTIONS(39), 1,
      anon_sym_LPAREN,
    ACTIONS(45), 1,
      anon_sym_LBRACK,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(51), 1,
      anon_sym_object,
    ACTIONS(55), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(73), 1,
      anon_sym_new,
    ACTIONS(75), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(77), 1,
      anon_sym_begin,
    ACTIONS(79), 1,
      sym_ocamlyacc_value,
    ACTIONS(87), 1,
      aux_sym_number_token1,
    ACTIONS(91), 1,
      anon_sym_SQUOTE,
    ACTIONS(93), 1,
      anon_sym_DQUOTE,
    ACTIONS(95), 1,
      sym_prefix_operator,
    ACTIONS(101), 1,
      sym__uppercase_identifier,
    ACTIONS(103), 1,
      aux_sym_tag_token1,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(681), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(683), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(751), 1,
      sym__lowercase_identifier,
    STATE(892), 1,
      sym__constructor_name,
    STATE(908), 1,
      sym_parenthesized_operator,
    STATE(936), 1,
      sym__value_name,
    STATE(944), 1,
      sym__simple_expression,
    STATE(1527), 1,
      sym_attribute,
    STATE(6591), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(47), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(924), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(922), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(939), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15229] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1654), 1,
      sym__lowercase_identifier,
    ACTIONS(1658), 1,
      anon_sym_LPAREN,
    ACTIONS(1660), 1,
      anon_sym_LBRACK,
    ACTIONS(1664), 1,
      anon_sym_LBRACE,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(1668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1670), 1,
      anon_sym_new,
    ACTIONS(1672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(1674), 1,
      anon_sym_begin,
    ACTIONS(1676), 1,
      sym_ocamlyacc_value,
    ACTIONS(1682), 1,
      aux_sym_number_token1,
    ACTIONS(1684), 1,
      anon_sym_SQUOTE,
    ACTIONS(1686), 1,
      anon_sym_DQUOTE,
    ACTIONS(1688), 1,
      sym_prefix_operator,
    ACTIONS(1690), 1,
      sym__uppercase_identifier,
    ACTIONS(1692), 1,
      aux_sym_tag_token1,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1528), 1,
      sym_attribute,
    STATE(2357), 1,
      sym__simple_expression,
    STATE(2402), 1,
      sym__value_name,
    STATE(2406), 1,
      sym_parenthesized_operator,
    STATE(2419), 1,
      sym__constructor_name,
    STATE(6407), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1662), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2348), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2351), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2350), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15349] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2604), 1,
      sym__lowercase_identifier,
    ACTIONS(2606), 1,
      anon_sym_LPAREN,
    ACTIONS(2608), 1,
      anon_sym_LBRACK,
    ACTIONS(2612), 1,
      anon_sym_LBRACE,
    ACTIONS(2614), 1,
      anon_sym_object,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2618), 1,
      anon_sym_new,
    ACTIONS(2620), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2622), 1,
      anon_sym_begin,
    ACTIONS(2624), 1,
      sym_ocamlyacc_value,
    ACTIONS(2626), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2628), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2630), 1,
      aux_sym_number_token1,
    ACTIONS(2632), 1,
      anon_sym_SQUOTE,
    ACTIONS(2634), 1,
      anon_sym_DQUOTE,
    ACTIONS(2636), 1,
      sym_prefix_operator,
    ACTIONS(2638), 1,
      sym__uppercase_identifier,
    ACTIONS(2640), 1,
      aux_sym_tag_token1,
    STATE(1529), 1,
      sym_attribute,
    STATE(1807), 1,
      sym__constructor_name,
    STATE(1877), 1,
      sym_parenthesized_operator,
    STATE(1879), 1,
      sym__simple_expression,
    STATE(1893), 1,
      sym__value_name,
    STATE(7240), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2610), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1891), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1926), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1921), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15469] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1530), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1548), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1546), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [15538] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1531), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1580), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1578), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [15607] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2009), 1,
      anon_sym_LT_DASH,
    STATE(1532), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1616), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1614), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [15678] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2746), 1,
      aux_sym_attribute_id_token1,
    STATE(1533), 1,
      sym_attribute,
    STATE(1544), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2744), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2742), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [15751] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1498), 1,
      anon_sym_COLON2,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2748), 1,
      anon_sym_EQ,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1534), 1,
      sym_attribute,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5579), 1,
      sym__maybe_polymorphic_typed,
    STATE(6690), 1,
      sym__coerced,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [15878] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1498), 1,
      anon_sym_COLON2,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2750), 1,
      anon_sym_EQ,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1535), 1,
      sym_attribute,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5611), 1,
      sym__maybe_polymorphic_typed,
    STATE(6823), 1,
      sym__coerced,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [16005] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1536), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1560), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1558), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16074] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1552), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1550), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16143] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1564), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1562), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16212] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(1539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1698), 30,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [16283] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1568), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1566), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16352] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1576), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1574), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16421] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2746), 1,
      aux_sym_attribute_id_token1,
    STATE(1533), 1,
      aux_sym_attribute_id_repeat1,
    STATE(1542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2754), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2752), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [16494] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1572), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1570), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16563] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2760), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1544), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(2758), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2756), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [16634] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1556), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1554), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16703] = 34,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1498), 1,
      anon_sym_COLON2,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2763), 1,
      anon_sym_EQ,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1546), 1,
      sym_attribute,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5688), 1,
      sym__maybe_polymorphic_typed,
    STATE(6368), 1,
      sym__coerced,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [16830] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2765), 1,
      anon_sym_LT_DASH,
    STATE(1547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16901] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(1548), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1698), 30,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [16972] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1632), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1630), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17040] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1550), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1730), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1728), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17108] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1884), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1882), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17176] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1742), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1740), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17244] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1553), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1752), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1750), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17312] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1554), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1928), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1926), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17380] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1555), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1932), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1930), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17448] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1556), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1936), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1934), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17516] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1940), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1938), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17584] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1588), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1586), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17652] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1592), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1590), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17720] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1612), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1610), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17788] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1856), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1854), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17856] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1562), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1640), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1638), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17924] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1563), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1620), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1618), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17992] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1624), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1622), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18060] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(1565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1698), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18130] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1566), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1636), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1634), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18198] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1567), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1832), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1830), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18266] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1568), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1644), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1642), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18334] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1569), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1652), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1650), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18402] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1696), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1694), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18470] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2767), 1,
      anon_sym_LT_DASH,
    STATE(1571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 32,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18540] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1572), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1722), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1720), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18608] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1828), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1826), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18676] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1836), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1834), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18744] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1575), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1844), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1842), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18812] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1576), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1852), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1850), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18880] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(1577), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1698), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18950] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1710), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1708), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19018] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1920), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1918), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19086] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1824), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1822), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19154] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1748), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1746), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19222] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1832), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1830), 30,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [19290] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2023), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2021), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19358] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1756), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1754), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19426] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1952), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1950), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19494] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1904), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1902), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19562] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1587), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1848), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1846), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19630] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1588), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1908), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1906), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19698] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1589), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1706), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1704), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19766] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1840), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1838), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19834] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1308), 1,
      anon_sym_DOT,
    ACTIONS(1310), 1,
      anon_sym_POUND,
    ACTIONS(1312), 1,
      sym_hash_operator,
    STATE(1591), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2081), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2079), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19908] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1714), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1712), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19976] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1876), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1874), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20044] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1584), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1582), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20112] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1595), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1888), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1886), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20180] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1856), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1854), 30,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [20248] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1597), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1916), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1914), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20316] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1944), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1942), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20384] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1599), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1760), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1758), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20452] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1600), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1600), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1598), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20520] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1601), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1860), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1858), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20588] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1604), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1602), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20656] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1596), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1594), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20724] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1608), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1606), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20792] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1605), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1628), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1626), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20860] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1606), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1616), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1614), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20928] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1607), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1648), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1646), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20996] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1608), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1868), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1866), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21064] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2769), 1,
      anon_sym_LT_DASH,
    STATE(1609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21134] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1864), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1862), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21202] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1611), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1872), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1870), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21270] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1612), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1718), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1716), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21338] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1613), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1880), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1878), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21406] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1892), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1890), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21474] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1308), 1,
      anon_sym_DOT,
    STATE(1615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1896), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1894), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21544] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1616), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1912), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1910), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21612] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1734), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1732), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21680] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2027), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2025), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21748] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1619), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1710), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1708), 30,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [21816] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1948), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1946), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21884] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1900), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1898), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21952] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1622), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1756), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1754), 30,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [22020] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1623), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1738), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1736), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22088] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1308), 1,
      anon_sym_DOT,
    ACTIONS(1310), 1,
      anon_sym_POUND,
    ACTIONS(1312), 1,
      sym_hash_operator,
    STATE(1624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2109), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2107), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22162] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1625), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2758), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
      aux_sym_attribute_id_token1,
    ACTIONS(2756), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [22230] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1626), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22298] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1924), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1922), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [22366] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2773), 1,
      anon_sym_DASH_GT,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1628), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6546), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22487] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2777), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1629), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    STATE(7234), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22608] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2779), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1630), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6801), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22729] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2781), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1631), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    STATE(7290), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22850] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2783), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1632), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    STATE(6883), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22971] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2785), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1633), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6562), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23092] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2787), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1634), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6701), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23213] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2789), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1635), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6540), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23334] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2791), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1636), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6552), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23455] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2793), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1637), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6340), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23576] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2795), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1638), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6387), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23697] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2797), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1639), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    STATE(7144), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23818] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2799), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1640), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    STATE(7162), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23939] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2801), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1641), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6612), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24060] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2803), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1642), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6664), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24181] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2805), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1643), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    STATE(6940), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24302] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2807), 1,
      anon_sym_LT_DASH,
    STATE(1644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 31,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24371] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2809), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1645), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6824), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24492] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2811), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1646), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6829), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24613] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2813), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1647), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    STATE(7010), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24734] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2815), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1648), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    STATE(7011), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24855] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2817), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1649), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    STATE(6961), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24976] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2819), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1650), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    STATE(7133), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25097] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2821), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1651), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6459), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25218] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2823), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1652), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6531), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25339] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2825), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1653), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6737), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25460] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2771), 1,
      anon_sym_COLON2,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2827), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1654), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6795), 1,
      sym__simple_typed,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25581] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2831), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1655), 2,
      sym_attribute,
      aux_sym__structure_repeat1,
    ACTIONS(2834), 18,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2829), 31,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [25649] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1649), 1,
      aux_sym_let_binding_repeat1,
    STATE(1656), 1,
      sym_attribute,
    STATE(1768), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25767] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1637), 1,
      aux_sym_let_binding_repeat1,
    STATE(1657), 1,
      sym_attribute,
    STATE(1738), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25885] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2836), 1,
      anon_sym_COLON,
    STATE(1658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2161), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [25953] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1641), 1,
      aux_sym_let_binding_repeat1,
    STATE(1659), 1,
      sym_attribute,
    STATE(1745), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26071] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1645), 1,
      aux_sym_let_binding_repeat1,
    STATE(1660), 1,
      sym_attribute,
    STATE(1747), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26189] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1639), 1,
      aux_sym_let_binding_repeat1,
    STATE(1661), 1,
      sym_attribute,
    STATE(1754), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26307] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1629), 1,
      aux_sym_let_binding_repeat1,
    STATE(1662), 1,
      sym_attribute,
    STATE(1722), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26425] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      anon_sym_LT_DASH,
    STATE(1663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 29,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [26493] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1651), 1,
      aux_sym_let_binding_repeat1,
    STATE(1664), 1,
      sym_attribute,
    STATE(1716), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26611] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1650), 1,
      aux_sym_let_binding_repeat1,
    STATE(1665), 1,
      sym_attribute,
    STATE(1727), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26729] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2840), 1,
      anon_sym_LT_DASH,
    STATE(1666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 30,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_do,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [26797] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1647), 1,
      aux_sym_let_binding_repeat1,
    STATE(1667), 1,
      sym_attribute,
    STATE(1772), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26915] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2842), 1,
      anon_sym_LT_DASH,
    STATE(1668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 29,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [26983] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1654), 1,
      aux_sym_let_binding_repeat1,
    STATE(1669), 1,
      sym_attribute,
    STATE(1703), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27101] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1633), 1,
      aux_sym_let_binding_repeat1,
    STATE(1670), 1,
      sym_attribute,
    STATE(1705), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27219] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1628), 1,
      aux_sym_let_binding_repeat1,
    STATE(1671), 1,
      sym_attribute,
    STATE(1732), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27337] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1466), 1,
      anon_sym_PERCENT,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1632), 1,
      aux_sym_let_binding_repeat1,
    STATE(1672), 1,
      sym_attribute,
    STATE(1719), 1,
      sym__attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27455] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2870), 1,
      sym_assign_operator,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1673), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2846), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2844), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [27554] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2870), 1,
      sym_assign_operator,
    ACTIONS(2872), 1,
      anon_sym_SEMI,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1674), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2736), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2734), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [27655] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2870), 1,
      sym_assign_operator,
    ACTIONS(2872), 1,
      anon_sym_SEMI,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2732), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2730), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [27756] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [27831] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2878), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1677), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27946] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28023] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28104] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2880), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28199] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2870), 1,
      sym_assign_operator,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2886), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2884), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28298] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1646), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1648), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [28363] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2081), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2079), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [28428] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1614), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1616), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [28493] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2888), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1685), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [28608] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2870), 1,
      sym_assign_operator,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1686), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2892), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2890), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28707] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1687), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 7,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28798] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1688), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2896), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2894), 31,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [28863] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2898), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1689), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [28978] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(1690), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1698), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [29045] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1854), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1856), 34,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_tag_token1,
  [29110] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2125), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2123), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [29175] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2775), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2900), 1,
      anon_sym_DASH_GT,
    STATE(848), 1,
      aux_sym_let_binding_repeat1,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1693), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29290] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2904), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2902), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [29387] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2870), 1,
      sym_assign_operator,
    ACTIONS(2872), 1,
      anon_sym_SEMI,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2908), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2906), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [29488] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2912), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [29573] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [29660] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [29747] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2870), 1,
      sym_assign_operator,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2880), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [29846] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2914), 1,
      anon_sym_LT_DASH,
    STATE(1700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 29,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_do,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [29913] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(1701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1698), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [29980] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [30057] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1635), 1,
      aux_sym_let_binding_repeat1,
    STATE(1703), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [30169] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1598), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1600), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30233] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1630), 1,
      aux_sym_let_binding_repeat1,
    STATE(1705), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [30345] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1706), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1638), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1640), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30409] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1602), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1604), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30473] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1606), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1608), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30537] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1694), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1696), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30601] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1756), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1754), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [30665] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1626), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1628), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30729] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1712), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1714), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30793] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1724), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1726), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30857] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    ACTIONS(2918), 1,
      anon_sym_SEMI,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    ACTIONS(2926), 1,
      sym_assign_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2736), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2734), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [30957] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2021), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2023), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31021] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1636), 1,
      aux_sym_let_binding_repeat1,
    STATE(1716), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31133] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1898), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1900), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31197] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1736), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1738), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31261] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1643), 1,
      aux_sym_let_binding_repeat1,
    STATE(1719), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31373] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    ACTIONS(2918), 1,
      anon_sym_SEMI,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    ACTIONS(2926), 1,
      sym_assign_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2732), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2730), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [31473] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1746), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1748), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31537] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1631), 1,
      aux_sym_let_binding_repeat1,
    STATE(1722), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31649] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1704), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1706), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31713] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1096), 1,
      sym__lowercase_identifier,
    ACTIONS(1100), 1,
      anon_sym_LPAREN,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2928), 1,
      anon_sym_lazy,
    STATE(1229), 1,
      sym_constructor_path,
    STATE(1314), 1,
      sym_tag,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1724), 1,
      sym_attribute,
    STATE(6396), 1,
      sym_module_path,
    STATE(7178), 1,
      sym__effect_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4379), 4,
      sym__simple_pattern,
      sym_constructor_pattern,
      sym_tag_pattern,
      sym_lazy_pattern,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4430), 10,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [31827] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1720), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1722), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31891] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1554), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1556), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31955] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1652), 1,
      aux_sym_let_binding_repeat1,
    STATE(1727), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [32067] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [32141] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [32217] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1730), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [32293] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 11,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [32373] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1634), 1,
      aux_sym_let_binding_repeat1,
    STATE(1732), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [32485] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1550), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1552), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32549] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1716), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1718), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32613] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1832), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1830), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [32677] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1630), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1632), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32741] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1728), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1730), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32805] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1638), 1,
      aux_sym_let_binding_repeat1,
    STATE(1738), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [32917] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1042), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1054), 1,
      anon_sym_LBRACE,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2930), 1,
      anon_sym_lazy,
    STATE(1275), 1,
      sym_constructor_path,
    STATE(1311), 1,
      sym_tag,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1739), 1,
      sym_attribute,
    STATE(6402), 1,
      sym__effect_binding_pattern,
    STATE(6409), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4604), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4603), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [33033] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1732), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1734), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [33097] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1740), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1742), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [33161] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1570), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1572), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [33225] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1822), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1824), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [33289] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2025), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2027), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [33353] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1642), 1,
      aux_sym_let_binding_repeat1,
    STATE(1745), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [33465] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(1746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1700), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [33531] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1646), 1,
      aux_sym_let_binding_repeat1,
    STATE(1747), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [33643] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    ACTIONS(2918), 1,
      anon_sym_SEMI,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    ACTIONS(2926), 1,
      sym_assign_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1748), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2908), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2906), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33743] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1874), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1876), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [33807] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2904), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2902), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33903] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1096), 1,
      sym__lowercase_identifier,
    ACTIONS(1100), 1,
      anon_sym_LPAREN,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2928), 1,
      anon_sym_lazy,
    STATE(1229), 1,
      sym_constructor_path,
    STATE(1314), 1,
      sym_tag,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1751), 1,
      sym_attribute,
    STATE(6396), 1,
      sym_module_path,
    STATE(6547), 1,
      sym__effect_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4379), 4,
      sym__simple_pattern,
      sym_constructor_pattern,
      sym_tag_pattern,
      sym_lazy_pattern,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4430), 10,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [34017] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1693), 1,
      aux_sym_let_binding_repeat1,
    STATE(1752), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [34129] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1042), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1054), 1,
      anon_sym_LBRACE,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2930), 1,
      anon_sym_lazy,
    STATE(1275), 1,
      sym_constructor_path,
    STATE(1311), 1,
      sym_tag,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1753), 1,
      sym_attribute,
    STATE(6409), 1,
      sym_module_path,
    STATE(6685), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4604), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4603), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [34245] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1640), 1,
      aux_sym_let_binding_repeat1,
    STATE(1754), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [34357] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1594), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1596), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [34421] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1910), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1912), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [34485] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1758), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1760), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [34549] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1856), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1854), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [34613] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1759), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1610), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1612), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [34677] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(1760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1700), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [34743] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1708), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1710), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [34807] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1762), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2912), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34891] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1763), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34977] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [35063] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    ACTIONS(2926), 1,
      sym_assign_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1765), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2880), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [35161] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [35251] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1767), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1946), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1948), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35315] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1653), 1,
      aux_sym_let_binding_repeat1,
    STATE(1768), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [35427] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1769), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2880), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [35521] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1770), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1846), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1848), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35585] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1830), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1832), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35649] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1648), 1,
      aux_sym_let_binding_repeat1,
    STATE(1772), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [35761] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    ACTIONS(2926), 1,
      sym_assign_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2886), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2884), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [35859] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1096), 1,
      sym__lowercase_identifier,
    ACTIONS(1100), 1,
      anon_sym_LPAREN,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2928), 1,
      anon_sym_lazy,
    STATE(1229), 1,
      sym_constructor_path,
    STATE(1314), 1,
      sym_tag,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1774), 1,
      sym_attribute,
    STATE(6396), 1,
      sym_module_path,
    STATE(7306), 1,
      sym__effect_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4379), 4,
      sym__simple_pattern,
      sym_constructor_pattern,
      sym_tag_pattern,
      sym_lazy_pattern,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4430), 10,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [35973] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1677), 1,
      aux_sym_let_binding_repeat1,
    STATE(1775), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [36085] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1042), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1054), 1,
      anon_sym_LBRACE,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2930), 1,
      anon_sym_lazy,
    STATE(1275), 1,
      sym_constructor_path,
    STATE(1311), 1,
      sym_tag,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1776), 1,
      sym_attribute,
    STATE(6338), 1,
      sym__effect_binding_pattern,
    STATE(6409), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4604), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4603), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [36201] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    ACTIONS(2926), 1,
      sym_assign_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1777), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2846), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2844), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [36299] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1778), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1886), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1888), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36363] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1685), 1,
      aux_sym_let_binding_repeat1,
    STATE(1779), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [36475] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1042), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1054), 1,
      anon_sym_LBRACE,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2930), 1,
      anon_sym_lazy,
    STATE(1275), 1,
      sym_constructor_path,
    STATE(1311), 1,
      sym_tag,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1780), 1,
      sym_attribute,
    STATE(6409), 1,
      sym_module_path,
    STATE(6445), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4604), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4603), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [36591] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1781), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1914), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1916), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36655] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1042), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1054), 1,
      anon_sym_LBRACE,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2930), 1,
      anon_sym_lazy,
    STATE(1275), 1,
      sym_constructor_path,
    STATE(1311), 1,
      sym_tag,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1782), 1,
      sym_attribute,
    STATE(6409), 1,
      sym_module_path,
    STATE(6529), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4604), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4603), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [36771] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1942), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1944), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36835] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1586), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1588), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36899] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1785), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1838), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1840), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36963] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1858), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1860), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37027] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1590), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1592), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37091] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1862), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1864), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37155] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1789), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1918), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1920), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37219] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    ACTIONS(2926), 1,
      sym_assign_operator,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(1790), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2892), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2890), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [37317] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1791), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1866), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1868), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37381] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1492), 1,
      anon_sym_LPAREN,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(1689), 1,
      aux_sym_let_binding_repeat1,
    STATE(1792), 1,
      sym_attribute,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1490), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2236), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [37493] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1793), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1870), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1872), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37557] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1878), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1880), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37621] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1795), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1882), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1884), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37685] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1574), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1576), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37749] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1890), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1892), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37813] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1710), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1708), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [37877] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1558), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1560), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37941] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1562), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1564), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [38005] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1566), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1568), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [38069] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1582), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1584), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [38133] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2932), 1,
      anon_sym_COMMA,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    ACTIONS(2938), 1,
      sym_rel_operator,
    ACTIONS(2940), 1,
      sym_assign_operator,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2886), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2884), 18,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [38230] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1850), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1852), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [38293] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1736), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1738), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38356] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1728), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1730), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38419] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1807), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1830), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1832), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38482] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2932), 1,
      anon_sym_COMMA,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    ACTIONS(2938), 1,
      sym_rel_operator,
    ACTIONS(2940), 1,
      sym_assign_operator,
    ACTIONS(2942), 1,
      anon_sym_SEMI,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1808), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2908), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2906), 17,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [38581] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1809), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1934), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1936), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38644] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1938), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1940), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38707] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1811), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1740), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1742), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38770] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1812), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1586), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1588), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38833] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1813), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1546), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1548), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [38896] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2948), 1,
      anon_sym_DOT,
    ACTIONS(2950), 1,
      anon_sym_POUND,
    ACTIONS(2952), 1,
      sym_hash_operator,
    STATE(1814), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2946), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2944), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38965] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2958), 1,
      anon_sym_DOT,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(2962), 1,
      sym_hash_operator,
    STATE(1815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2956), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2954), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [39034] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 11,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 23,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39113] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1590), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1592), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39176] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2958), 1,
      anon_sym_DOT,
    STATE(1818), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1894), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1896), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [39241] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1750), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1752), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [39304] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1854), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1856), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39367] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2932), 1,
      anon_sym_COMMA,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    ACTIONS(2938), 1,
      sym_rel_operator,
    ACTIONS(2940), 1,
      sym_assign_operator,
    ACTIONS(2942), 1,
      anon_sym_SEMI,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2732), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2730), 17,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [39466] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(1822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1700), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39531] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1610), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1612), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39594] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1630), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1632), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39657] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1874), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1876), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39720] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 27,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39795] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1708), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1710), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39858] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1828), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1582), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1584), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39921] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1829), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1750), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1752), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39984] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1830), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1638), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1640), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40047] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1694), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1696), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40110] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1758), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1760), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40173] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1833), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1720), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1722), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40236] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1834), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1650), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1652), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [40299] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1578), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1580), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40362] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2932), 1,
      anon_sym_COMMA,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    ACTIONS(2938), 1,
      sym_rel_operator,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2904), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2902), 19,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [40457] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1934), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1936), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [40520] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2912), 8,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 22,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40603] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 8,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 21,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
  [40688] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 8,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 21,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
  [40773] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2932), 1,
      anon_sym_COMMA,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    ACTIONS(2938), 1,
      sym_rel_operator,
    ACTIONS(2940), 1,
      sym_assign_operator,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2880), 18,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [40870] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    ACTIONS(2938), 1,
      sym_rel_operator,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 5,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 21,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
  [40959] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    ACTIONS(2938), 1,
      sym_rel_operator,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1843), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2880), 20,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [41052] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1938), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1940), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [41115] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2958), 1,
      anon_sym_DOT,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(2962), 1,
      sym_hash_operator,
    STATE(1845), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2966), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2964), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [41184] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1554), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1556), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41247] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1550), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1552), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41310] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1602), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1604), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41373] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1902), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1904), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [41436] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2870), 1,
      sym_assign_operator,
    ACTIONS(2968), 1,
      anon_sym_SEMI,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2908), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2906), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [41535] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1570), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1572), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41598] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1614), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1616), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41661] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1822), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1824), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41724] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1902), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1904), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41787] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1606), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1608), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41850] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1846), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1848), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41913] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2970), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2972), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [41976] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1826), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1828), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [42039] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2728), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2726), 34,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [42102] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1838), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1840), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42165] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1634), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1636), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [42228] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1858), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1860), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42291] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1906), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1908), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42354] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1864), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1922), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1924), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42417] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2932), 1,
      anon_sym_COMMA,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    ACTIONS(2938), 1,
      sym_rel_operator,
    ACTIONS(2940), 1,
      sym_assign_operator,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2846), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2844), 18,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [42514] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1926), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1928), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [42577] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1558), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1560), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42640] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1618), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1620), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [42703] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1562), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1564), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42766] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2974), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2976), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [42829] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1566), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1568), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42892] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1594), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1596), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42955] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1873), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1950), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1952), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43018] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1862), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1864), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43081] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2948), 1,
      anon_sym_DOT,
    ACTIONS(2950), 1,
      anon_sym_POUND,
    ACTIONS(2952), 1,
      sym_hash_operator,
    STATE(1875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2980), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2978), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43150] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2982), 1,
      anon_sym_DOT_DOT,
    STATE(1876), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1316), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1318), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [43215] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1646), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1648), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43278] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2988), 1,
      anon_sym_COLON,
    STATE(1878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2984), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [43343] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2948), 1,
      anon_sym_DOT,
    ACTIONS(2950), 1,
      anon_sym_POUND,
    ACTIONS(2952), 1,
      sym_hash_operator,
    STATE(1879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2966), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2964), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43412] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1754), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1756), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43475] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1546), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1548), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43538] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43611] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1883), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1598), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1600), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43674] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1870), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1872), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43737] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1878), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1880), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43800] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1930), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1932), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43863] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1914), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1916), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43926] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2948), 1,
      anon_sym_DOT,
    ACTIONS(2950), 1,
      anon_sym_POUND,
    ACTIONS(2952), 1,
      sym_hash_operator,
    STATE(1888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2992), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2990), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43995] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1618), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1620), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44058] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1622), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1624), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44121] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1634), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1636), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44184] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1642), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1644), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44247] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1716), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1718), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44310] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1622), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1624), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [44373] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1895), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1922), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1924), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44436] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1882), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1884), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44499] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1950), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1952), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44562] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1650), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1652), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44625] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1574), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1576), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44688] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2932), 1,
      anon_sym_COMMA,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    ACTIONS(2938), 1,
      sym_rel_operator,
    ACTIONS(2940), 1,
      sym_assign_operator,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2892), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2890), 18,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [44785] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1890), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1892), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44848] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2870), 1,
      sym_assign_operator,
    ACTIONS(2968), 1,
      anon_sym_SEMI,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2732), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2730), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [44947] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1903), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1926), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1928), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45010] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1826), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1828), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [45073] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1930), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1932), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45136] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1642), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1644), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45199] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2948), 1,
      anon_sym_DOT,
    STATE(1907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1894), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1896), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45264] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(1908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1700), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45329] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1909), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1732), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1734), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45392] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2025), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2027), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45455] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2948), 1,
      anon_sym_DOT,
    ACTIONS(2950), 1,
      anon_sym_POUND,
    ACTIONS(2952), 1,
      sym_hash_operator,
    STATE(1911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2956), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2954), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45524] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1626), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1628), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45587] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1650), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1652), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45650] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1712), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1714), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45713] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1842), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1844), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [45776] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2732), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2730), 34,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [45839] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1917), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1834), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1836), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [45902] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1850), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1852), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [45965] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1578), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1580), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46028] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    ACTIONS(2852), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2862), 1,
      sym_concat_operator,
    ACTIONS(2864), 1,
      sym_rel_operator,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2870), 1,
      sym_assign_operator,
    ACTIONS(2968), 1,
      anon_sym_SEMI,
    STATE(589), 1,
      sym_pow_operator,
    STATE(590), 1,
      sym_mult_operator,
    STATE(591), 1,
      sym_add_operator,
    STATE(592), 1,
      sym_and_operator,
    STATE(593), 1,
      sym_or_operator,
    STATE(1920), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2736), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2734), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [46127] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1724), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1726), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46190] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2958), 1,
      anon_sym_DOT,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(2962), 1,
      sym_hash_operator,
    STATE(1922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2946), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2944), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46259] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1910), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1912), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46322] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2932), 1,
      anon_sym_COMMA,
    ACTIONS(2934), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2936), 1,
      sym_concat_operator,
    ACTIONS(2938), 1,
      sym_rel_operator,
    ACTIONS(2940), 1,
      sym_assign_operator,
    ACTIONS(2942), 1,
      anon_sym_SEMI,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2736), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2734), 17,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [46421] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2736), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2734), 34,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [46484] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1704), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1706), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46547] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1826), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1828), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46610] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2958), 1,
      anon_sym_DOT,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(2962), 1,
      sym_hash_operator,
    STATE(1928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2992), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2990), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46679] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1754), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1756), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46742] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1866), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1868), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46805] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1834), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1836), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46868] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1842), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1844), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46931] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1933), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1842), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1844), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46994] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1942), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1944), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47057] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1935), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1886), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1888), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47120] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1618), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1620), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [47183] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2958), 1,
      anon_sym_DOT,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(2962), 1,
      sym_hash_operator,
    STATE(1937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2980), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2978), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [47252] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1622), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1624), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [47315] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1946), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1948), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47378] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1634), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1636), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [47441] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(706), 1,
      sym_pow_operator,
    STATE(707), 1,
      sym_mult_operator,
    STATE(708), 1,
      sym_add_operator,
    STATE(709), 1,
      sym_and_operator,
    STATE(710), 1,
      sym_or_operator,
    STATE(1941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 27,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47516] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1850), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1852), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47579] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1943), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1918), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1920), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47642] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1746), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1748), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47705] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2021), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2023), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47768] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1898), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1900), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47831] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1906), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1908), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [47894] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2996), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2994), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47956] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1582), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1584), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48018] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1950), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1758), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1760), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48080] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1838), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1840), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48142] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1858), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1860), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48204] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1862), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1864), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48266] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1870), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1872), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48328] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1878), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1880), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48390] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1882), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1884), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48452] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1574), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1576), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48514] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1890), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1892), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48576] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    STATE(1959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1894), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1896), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48640] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1910), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1912), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48702] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1946), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1948), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48764] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1898), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1900), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48826] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1736), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1738), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48888] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1586), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1588), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48950] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1590), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1592), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49012] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1610), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1612), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49074] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1630), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1632), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49136] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1638), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1640), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49198] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1694), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1696), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49260] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1720), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1722), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49322] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1554), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1556), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49384] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1550), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1552), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49446] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1570), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1572), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49508] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1822), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1824), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49570] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1846), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1848), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49632] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1558), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1560), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49694] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1562), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1564), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49756] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1566), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1568), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49818] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1594), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1596), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49880] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3000), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3002), 28,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [49942] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1716), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1718), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50004] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1732), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1734), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50066] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2025), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2027), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50128] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1578), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1580), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50190] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1866), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1868), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50252] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1918), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1920), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50314] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1728), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1730), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50376] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1746), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1748), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50438] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2021), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2023), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50500] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1546), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1548), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50562] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1712), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1714), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50624] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1886), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1888), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50686] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1914), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1916), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50748] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1942), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1944), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50810] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1598), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1600), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50872] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1602), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1604), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50934] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1606), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1608), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50996] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1626), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1628), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51058] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3006), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3004), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51120] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3010), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3008), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51182] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1316), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1318), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51244] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3014), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3012), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51306] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [51378] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1614), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1616), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51440] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1646), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1648), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51502] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3018), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3016), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51564] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3022), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3020), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51626] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3026), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3024), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51688] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1642), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1644), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51750] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3030), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3028), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51812] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3034), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3032), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51874] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1834), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1836), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51936] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2013), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3038), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3036), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51998] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3042), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3040), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [52060] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1618), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1620), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52122] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1622), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1624), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52184] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1634), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1636), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52246] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1750), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1752), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52308] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1842), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1844), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52370] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3046), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3044), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [52432] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1926), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1928), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52494] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1930), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1932), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52556] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2023), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1934), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1936), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52618] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1938), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1940), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52680] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1650), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1652), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52742] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1854), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1856), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52804] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1826), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1828), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52866] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1850), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1852), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52928] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(2029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1700), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52992] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1830), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1832), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53054] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3048), 1,
      anon_sym_COMMA,
    ACTIONS(3050), 1,
      anon_sym_COLON_COLON,
    STATE(2031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3046), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3044), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53120] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(2032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1700), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53184] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1708), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1710), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53246] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3048), 1,
      anon_sym_COMMA,
    STATE(2034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3046), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3044), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53310] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3046), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3044), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53372] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3048), 1,
      anon_sym_COMMA,
    ACTIONS(3050), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3056), 1,
      anon_sym_PIPE,
    STATE(2036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3054), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3052), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53440] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1754), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1756), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53502] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1950), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1952), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53564] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1902), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1904), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53626] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2040), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1874), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1876), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53688] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1922), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1924), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53750] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3054), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3052), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53812] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3060), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3058), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53874] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3064), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3062), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53936] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    ACTIONS(3068), 1,
      anon_sym_SEMI,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    ACTIONS(3076), 1,
      sym_assign_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2908), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2906), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [54034] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3050), 1,
      anon_sym_COLON_COLON,
    STATE(2046), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3080), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3078), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54098] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3084), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3082), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54160] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3088), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3086), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54222] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1648), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1646), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54284] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3092), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3090), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54346] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1644), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1642), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54408] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3094), 1,
      anon_sym_POUND,
    ACTIONS(3096), 1,
      sym_hash_operator,
    STATE(2052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2992), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2990), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54476] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3094), 1,
      anon_sym_POUND,
    ACTIONS(3096), 1,
      sym_hash_operator,
    STATE(2053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2956), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2954), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54544] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3100), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3098), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54606] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    ACTIONS(3076), 1,
      sym_assign_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2846), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2844), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [54702] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3104), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3102), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54764] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3106), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54826] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2058), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3112), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3110), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54888] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2059), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3114), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54950] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3120), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3118), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55012] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3094), 1,
      anon_sym_POUND,
    ACTIONS(3096), 1,
      sym_hash_operator,
    STATE(2061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2980), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2978), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55080] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3094), 1,
      anon_sym_POUND,
    ACTIONS(3096), 1,
      sym_hash_operator,
    STATE(2062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2966), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2964), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55148] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1836), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1834), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55210] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    ACTIONS(3068), 1,
      anon_sym_SEMI,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    ACTIONS(3076), 1,
      sym_assign_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2736), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2734), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [55308] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    ACTIONS(2926), 1,
      sym_assign_operator,
    ACTIONS(3122), 1,
      anon_sym_SEMI,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(2065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2736), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2734), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [55406] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3048), 1,
      anon_sym_COMMA,
    ACTIONS(3050), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3056), 1,
      anon_sym_PIPE,
    ACTIONS(3128), 1,
      anon_sym_as,
    STATE(2066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3126), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3124), 23,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55476] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    ACTIONS(2926), 1,
      sym_assign_operator,
    ACTIONS(3122), 1,
      anon_sym_SEMI,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(2067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2732), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2730), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [55574] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3094), 1,
      anon_sym_POUND,
    ACTIONS(3096), 1,
      sym_hash_operator,
    STATE(2068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2946), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2944), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55642] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2069), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3130), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3132), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [55704] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    ACTIONS(3068), 1,
      anon_sym_SEMI,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    ACTIONS(3076), 1,
      sym_assign_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2732), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2730), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [55802] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3048), 1,
      anon_sym_COMMA,
    ACTIONS(3050), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3056), 1,
      anon_sym_PIPE,
    ACTIONS(3128), 1,
      anon_sym_as,
    STATE(2071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3136), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3134), 23,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55872] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3138), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55934] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3048), 1,
      anon_sym_COMMA,
    ACTIONS(3050), 1,
      anon_sym_COLON_COLON,
    STATE(2073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3138), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56000] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3048), 1,
      anon_sym_COMMA,
    STATE(2074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3138), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56064] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3138), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56126] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3048), 1,
      anon_sym_COMMA,
    ACTIONS(3050), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3056), 1,
      anon_sym_PIPE,
    STATE(2076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3144), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3142), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56194] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3144), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3142), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56256] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3148), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3146), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56318] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3050), 1,
      anon_sym_COLON_COLON,
    STATE(2079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3150), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56382] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3154), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56444] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3158), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56506] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3164), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3162), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56568] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3168), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3166), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56630] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3170), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3172), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [56692] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    ACTIONS(2920), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2922), 1,
      sym_concat_operator,
    ACTIONS(2924), 1,
      sym_rel_operator,
    ACTIONS(2926), 1,
      sym_assign_operator,
    ACTIONS(3122), 1,
      anon_sym_SEMI,
    STATE(619), 1,
      sym_pow_operator,
    STATE(620), 1,
      sym_mult_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(2085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2908), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2906), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [56790] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2086), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1752), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1750), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56852] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    ACTIONS(3076), 1,
      sym_assign_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2892), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2890), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [56948] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1952), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1950), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57010] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3176), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3174), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57072] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3180), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3178), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57134] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1904), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1902), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57196] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2092), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1908), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1906), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57258] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1924), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1922), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57320] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1928), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1926), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57382] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1932), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1930), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57444] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1936), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1934), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57506] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1940), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1938), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57568] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2098), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3184), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3182), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57630] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2099), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3186), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3188), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [57692] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3190), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3192), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [57754] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3196), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3194), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57816] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3200), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3198), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57878] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3204), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3202), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57940] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3206), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3208), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [58002] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3212), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3210), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58064] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3216), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3214), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58126] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1616), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1614), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58188] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2904), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2902), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [58282] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2912), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [58364] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [58448] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [58532] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3220), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3218), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58594] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    ACTIONS(3076), 1,
      sym_assign_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2880), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [58690] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 26,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [58764] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 26,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [58838] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 10,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [58916] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 5,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [59004] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2880), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [59096] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    ACTIONS(3076), 1,
      sym_assign_operator,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2886), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2884), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [59192] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1740), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1742), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [59254] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1724), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1726), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [59316] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1704), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1706), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [59378] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2123), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1906), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1908), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [59440] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3226), 1,
      anon_sym_PIPE,
    STATE(2124), 1,
      sym_attribute,
    STATE(2126), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3224), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3222), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [59505] = 33,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(3228), 1,
      sym__lowercase_identifier,
    ACTIONS(3230), 1,
      anon_sym_LPAREN,
    ACTIONS(3232), 1,
      anon_sym_external,
    ACTIONS(3234), 1,
      anon_sym_private,
    ACTIONS(3236), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(3240), 1,
      anon_sym_LBRACK,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    STATE(2125), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3341), 1,
      sym__type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3500), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [59622] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3266), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2126), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3264), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [59685] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3273), 1,
      anon_sym_COLON,
    STATE(2127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [59748] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3226), 1,
      anon_sym_PIPE,
    STATE(2128), 1,
      sym_attribute,
    STATE(2133), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [59813] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3226), 1,
      anon_sym_PIPE,
    STATE(2124), 1,
      aux_sym__match_cases_repeat1,
    STATE(2129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [59878] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3287), 1,
      anon_sym_else,
    STATE(2130), 1,
      sym_attribute,
    STATE(2183), 1,
      sym_else_clause,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [59943] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3293), 1,
      anon_sym_COLON,
    STATE(2131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3289), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3291), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [60006] = 33,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3295), 1,
      anon_sym_LPAREN,
    ACTIONS(3297), 1,
      anon_sym_external,
    ACTIONS(3299), 1,
      anon_sym_private,
    ACTIONS(3301), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(3305), 1,
      anon_sym_LBRACK,
    ACTIONS(3309), 1,
      anon_sym_LBRACE,
    ACTIONS(3311), 1,
      sym__uppercase_identifier,
    STATE(2132), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4508), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(4938), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3307), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4667), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [60123] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3226), 1,
      anon_sym_PIPE,
    STATE(2126), 1,
      aux_sym__match_cases_repeat1,
    STATE(2133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [60188] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3287), 1,
      anon_sym_else,
    STATE(2134), 1,
      sym_attribute,
    STATE(2198), 1,
      sym_else_clause,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [60253] = 33,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3232), 1,
      anon_sym_external,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3317), 1,
      anon_sym_LPAREN,
    ACTIONS(3319), 1,
      anon_sym_private,
    ACTIONS(3321), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(3325), 1,
      anon_sym_LBRACK,
    STATE(2135), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3481), 1,
      sym__type,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3558), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [60370] = 33,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(3228), 1,
      sym__lowercase_identifier,
    ACTIONS(3230), 1,
      anon_sym_LPAREN,
    ACTIONS(3232), 1,
      anon_sym_external,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(3240), 1,
      anon_sym_LBRACK,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3327), 1,
      anon_sym_private,
    ACTIONS(3329), 1,
      anon_sym_DOT_DOT,
    STATE(2136), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3328), 1,
      sym__type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3483), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [60487] = 33,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3232), 1,
      anon_sym_external,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3317), 1,
      anon_sym_LPAREN,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(3325), 1,
      anon_sym_LBRACK,
    ACTIONS(3331), 1,
      anon_sym_private,
    ACTIONS(3333), 1,
      anon_sym_DOT_DOT,
    STATE(2137), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3493), 1,
      sym__type,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3580), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [60604] = 33,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3295), 1,
      anon_sym_LPAREN,
    ACTIONS(3297), 1,
      anon_sym_external,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(3305), 1,
      anon_sym_LBRACK,
    ACTIONS(3309), 1,
      anon_sym_LBRACE,
    ACTIONS(3311), 1,
      sym__uppercase_identifier,
    ACTIONS(3335), 1,
      anon_sym_private,
    ACTIONS(3337), 1,
      anon_sym_DOT_DOT,
    STATE(2138), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4499), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(4938), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3307), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4634), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [60721] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3339), 1,
      anon_sym_COLON,
    STATE(2139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60784] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3343), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3341), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60844] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1904), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1902), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [60904] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1908), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1906), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [60964] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3345), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3347), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [61024] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3349), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3351), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [61084] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1924), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
    ACTIONS(1922), 24,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_POUND,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [61144] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3355), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3353), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [61204] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3359), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3357), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61264] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1924), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1922), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [61324] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2149), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3361), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3363), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [61384] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3367), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3365), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61444] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2151), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3371), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3369), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61504] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3375), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3373), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61564] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3379), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3377), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61624] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3381), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3383), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [61684] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3387), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3385), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61744] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3389), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3391), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [61804] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3006), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3004), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [61864] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2158), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3393), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3395), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [61924] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3397), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3399), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [61984] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3401), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3403), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [62044] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1928), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1926), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [62104] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3405), 1,
      anon_sym_COLON,
    STATE(2162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62166] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    ACTIONS(3076), 1,
      sym_assign_operator,
    ACTIONS(3407), 1,
      anon_sym_SEMI,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2908), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2906), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [62262] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3411), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3409), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [62322] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1932), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1930), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [62382] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3415), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3413), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62442] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2167), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1936), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1934), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [62502] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3371), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3369), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [62562] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1940), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1938), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [62622] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3379), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3377), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [62682] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2171), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3417), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3419), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [62742] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3423), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3421), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62802] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2173), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3425), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3427), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [62862] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2174), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3220), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3218), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [62922] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3429), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3431), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [62982] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2176), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2976), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2974), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [63042] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(743), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(741), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63102] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3387), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3385), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63162] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2179), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3415), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3413), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63222] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3435), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3433), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63282] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3439), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3437), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63342] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2182), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3443), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3441), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63402] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2183), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3423), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3421), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63462] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3435), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3433), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63522] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2185), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3443), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3441), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63582] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3447), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3445), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63642] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2187), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3451), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3449), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63702] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2188), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3455), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3453), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63762] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3451), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3449), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63822] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3459), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3457), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63882] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2191), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3463), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3461), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63942] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1752), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1750), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64002] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3465), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3467), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64062] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2194), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3026), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3024), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64122] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2195), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3030), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3028), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64182] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2196), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3471), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3469), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64242] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3473), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3475), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64302] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3359), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3357), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [64362] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1952), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1950), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64422] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3439), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3437), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [64482] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1908), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
    ACTIONS(1906), 24,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_POUND,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [64542] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3477), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3479), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64602] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3064), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3062), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64662] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1952), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
    ACTIONS(1950), 24,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_POUND,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [64722] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3481), 1,
      anon_sym_else,
    STATE(2183), 1,
      sym_else_clause,
    STATE(2205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [64786] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2206), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3367), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3365), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [64846] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2207), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3483), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3485), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64906] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2208), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3487), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3489), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64966] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2209), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3491), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3493), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [65026] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2210), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3455), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3453), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [65086] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3375), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3373), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [65146] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3495), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3497), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [65206] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2213), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3499), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3501), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [65266] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3459), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3457), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [65326] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3481), 1,
      anon_sym_else,
    STATE(2198), 1,
      sym_else_clause,
    STATE(2215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [65390] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3503), 1,
      anon_sym_PIPE,
    STATE(2216), 1,
      sym_attribute,
    STATE(2220), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [65454] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1620), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1618), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [65514] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3463), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3461), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [65574] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3503), 1,
      anon_sym_PIPE,
    STATE(2219), 1,
      sym_attribute,
    STATE(2220), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3224), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3222), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [65638] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3505), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2220), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3264), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [65700] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1624), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1622), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [65760] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3411), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3409), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [65820] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(743), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(741), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [65880] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3503), 1,
      anon_sym_PIPE,
    STATE(2216), 1,
      aux_sym__match_cases_repeat1,
    STATE(2224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [65944] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3503), 1,
      anon_sym_PIPE,
    STATE(2219), 1,
      aux_sym__match_cases_repeat1,
    STATE(2225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [66008] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1636), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1634), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [66068] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3471), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3469), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [66128] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2228), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3510), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3508), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [66188] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1644), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1642), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [66248] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3514), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3512), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [66308] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3514), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3512), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [66368] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3518), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3516), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [66428] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1652), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1650), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [66488] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3343), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3341), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [66548] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3289), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3291), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [66608] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3520), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3522), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [66668] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3264), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [66728] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [66788] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    ACTIONS(3076), 1,
      sym_assign_operator,
    ACTIONS(3407), 1,
      anon_sym_SEMI,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2736), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2734), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [66884] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3104), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3102), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [66944] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2145), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67004] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3106), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67064] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3524), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3526), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67124] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3530), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3528), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [67184] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3534), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3532), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [67244] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3536), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3538), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67304] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3542), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3540), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [67364] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3542), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3540), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67424] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    STATE(2249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [67486] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3544), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3546), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67546] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2972), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2970), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67606] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1828), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1826), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67666] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3510), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3508), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67726] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    ACTIONS(3070), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3072), 1,
      sym_concat_operator,
    ACTIONS(3074), 1,
      sym_rel_operator,
    ACTIONS(3076), 1,
      sym_assign_operator,
    ACTIONS(3407), 1,
      anon_sym_SEMI,
    STATE(532), 1,
      sym_pow_operator,
    STATE(533), 1,
      sym_mult_operator,
    STATE(534), 1,
      sym_add_operator,
    STATE(535), 1,
      sym_and_operator,
    STATE(536), 1,
      sym_or_operator,
    STATE(2254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2732), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2858), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2730), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [67822] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3518), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3516), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67882] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2256), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1836), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1834), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67942] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1844), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1842), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68002] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1852), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1850), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68062] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68122] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1904), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
    ACTIONS(1902), 24,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_POUND,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [68182] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3355), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3353), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68242] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3548), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3550), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68302] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3554), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68362] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3556), 1,
      anon_sym_PIPE,
    STATE(2264), 1,
      sym_attribute,
    STATE(2282), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 11,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68425] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3379), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3377), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68484] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3411), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3409), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68543] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3471), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3469), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68602] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3542), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3540), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68661] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(3228), 1,
      sym__lowercase_identifier,
    ACTIONS(3230), 1,
      anon_sym_LPAREN,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(3240), 1,
      anon_sym_LBRACK,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3327), 1,
      anon_sym_private,
    ACTIONS(3329), 1,
      anon_sym_DOT_DOT,
    STATE(2269), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3328), 1,
      sym__type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3483), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [68774] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3355), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3353), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68833] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2271), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3387), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3385), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68892] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2272), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3415), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3413), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68951] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3423), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3421), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [69010] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3435), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3433), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [69069] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2272), 1,
      sym__lowercase_identifier,
    ACTIONS(2276), 1,
      anon_sym_LBRACK,
    ACTIONS(2280), 1,
      anon_sym_LBRACE,
    ACTIONS(2284), 1,
      anon_sym_POUND,
    ACTIONS(2286), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2290), 1,
      anon_sym_begin,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2298), 1,
      aux_sym_number_token1,
    ACTIONS(2300), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2302), 1,
      anon_sym_SQUOTE,
    ACTIONS(2304), 1,
      anon_sym_DQUOTE,
    ACTIONS(3558), 1,
      anon_sym_LPAREN,
    ACTIONS(3560), 1,
      sym__uppercase_identifier,
    ACTIONS(3562), 1,
      aux_sym_tag_token1,
    STATE(2275), 1,
      sym_attribute,
    STATE(4882), 1,
      sym__simple_pattern,
    STATE(4980), 1,
      sym__constructor_name,
    STATE(6671), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2278), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4788), 2,
      sym__constant,
      sym_signed_number,
    STATE(4738), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4959), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4958), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [69170] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3443), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3441), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [69229] = 27,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1314), 1,
      sym__lowercase_identifier,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1328), 1,
      anon_sym_LBRACE,
    ACTIONS(1332), 1,
      anon_sym_POUND,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1338), 1,
      anon_sym_begin,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(1354), 1,
      sym__uppercase_identifier,
    ACTIONS(1356), 1,
      aux_sym_tag_token1,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    ACTIONS(2422), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1735), 1,
      sym__constructor_name,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(2081), 1,
      sym__simple_binding_pattern,
    STATE(2277), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1326), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2001), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [69332] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3451), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3449), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [69391] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3317), 1,
      anon_sym_LPAREN,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(3325), 1,
      anon_sym_LBRACK,
    ACTIONS(3331), 1,
      anon_sym_private,
    ACTIONS(3333), 1,
      anon_sym_DOT_DOT,
    STATE(2279), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3493), 1,
      sym__type,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3580), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [69504] = 27,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2280), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(3131), 1,
      sym__simple_binding_pattern,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3030), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [69607] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3264), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [69666] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3556), 1,
      anon_sym_PIPE,
    STATE(2282), 1,
      sym_attribute,
    STATE(2299), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3224), 11,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3222), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [69729] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(743), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(741), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [69788] = 27,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1038), 1,
      sym__lowercase_identifier,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1054), 1,
      anon_sym_LBRACE,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3564), 1,
      anon_sym_LPAREN,
    ACTIONS(3566), 1,
      sym__uppercase_identifier,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    STATE(2284), 1,
      sym_attribute,
    STATE(3026), 1,
      sym__constructor_name,
    STATE(3541), 1,
      sym_parenthesized_operator,
    STATE(4560), 1,
      sym__simple_binding_pattern,
    STATE(6594), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4603), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [69891] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2736), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2734), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [69950] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3295), 1,
      anon_sym_LPAREN,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(3305), 1,
      anon_sym_LBRACK,
    ACTIONS(3309), 1,
      anon_sym_LBRACE,
    ACTIONS(3311), 1,
      sym__uppercase_identifier,
    ACTIONS(3335), 1,
      anon_sym_private,
    ACTIONS(3337), 1,
      anon_sym_DOT_DOT,
    STATE(2286), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4499), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(4938), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3307), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4634), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [70063] = 27,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1424), 1,
      sym__lowercase_identifier,
    ACTIONS(1428), 1,
      anon_sym_LPAREN,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1438), 1,
      anon_sym_LBRACE,
    ACTIONS(1448), 1,
      anon_sym_POUND,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1454), 1,
      anon_sym_begin,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(1476), 1,
      sym__uppercase_identifier,
    ACTIONS(1478), 1,
      aux_sym_tag_token1,
    ACTIONS(2017), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2019), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2287), 1,
      sym_attribute,
    STATE(2836), 1,
      sym__constructor_name,
    STATE(3018), 1,
      sym__simple_binding_pattern,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(6505), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1436), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3030), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [70166] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3447), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3445), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70225] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3317), 1,
      anon_sym_LPAREN,
    ACTIONS(3319), 1,
      anon_sym_private,
    ACTIONS(3321), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(3325), 1,
      anon_sym_LBRACK,
    STATE(2289), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3481), 1,
      sym__type,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3558), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [70338] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3556), 1,
      anon_sym_PIPE,
    STATE(2290), 1,
      sym_attribute,
    STATE(2316), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 11,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70401] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1096), 1,
      sym__lowercase_identifier,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3566), 1,
      sym__uppercase_identifier,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3570), 1,
      anon_sym_LPAREN,
    STATE(2291), 1,
      sym_attribute,
    STATE(3026), 1,
      sym__constructor_name,
    STATE(4380), 1,
      sym__simple_pattern,
    STATE(6593), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4430), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [70502] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2292), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2732), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2730), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70561] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1096), 1,
      sym__lowercase_identifier,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2242), 1,
      anon_sym_POUND,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3572), 1,
      anon_sym_LPAREN,
    ACTIONS(3574), 1,
      sym__uppercase_identifier,
    STATE(2293), 1,
      sym_attribute,
    STATE(4398), 1,
      sym__simple_pattern,
    STATE(4455), 1,
      sym__constructor_name,
    STATE(6633), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4430), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [70662] = 27,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1054), 1,
      anon_sym_LBRACE,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2236), 1,
      sym__lowercase_identifier,
    ACTIONS(2242), 1,
      anon_sym_POUND,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3574), 1,
      sym__uppercase_identifier,
    ACTIONS(3576), 1,
      anon_sym_LPAREN,
    STATE(2294), 1,
      sym_attribute,
    STATE(4455), 1,
      sym__constructor_name,
    STATE(4583), 1,
      sym__simple_binding_pattern,
    STATE(4689), 1,
      sym_parenthesized_operator,
    STATE(6634), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4603), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [70765] = 27,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1054), 1,
      anon_sym_LBRACE,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(3564), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(2295), 1,
      sym_attribute,
    STATE(4583), 1,
      sym__simple_binding_pattern,
    STATE(6551), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4603), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [70868] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3455), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3453), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70927] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3439), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3437), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70986] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(3228), 1,
      sym__lowercase_identifier,
    ACTIONS(3230), 1,
      anon_sym_LPAREN,
    ACTIONS(3234), 1,
      anon_sym_private,
    ACTIONS(3236), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(3240), 1,
      anon_sym_LBRACK,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    STATE(2298), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3341), 1,
      sym__type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3500), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [71099] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3578), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2299), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3264), 11,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [71160] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3295), 1,
      anon_sym_LPAREN,
    ACTIONS(3299), 1,
      anon_sym_private,
    ACTIONS(3301), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(3305), 1,
      anon_sym_LBRACK,
    ACTIONS(3309), 1,
      anon_sym_LBRACE,
    ACTIONS(3311), 1,
      sym__uppercase_identifier,
    STATE(2300), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4508), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(4938), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3307), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4667), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [71273] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2301), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3359), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3357), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [71332] = 27,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1054), 1,
      anon_sym_LBRACE,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2236), 1,
      sym__lowercase_identifier,
    ACTIONS(2242), 1,
      anon_sym_POUND,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3574), 1,
      sym__uppercase_identifier,
    ACTIONS(3576), 1,
      anon_sym_LPAREN,
    STATE(2302), 1,
      sym_attribute,
    STATE(4455), 1,
      sym__constructor_name,
    STATE(4560), 1,
      sym__simple_binding_pattern,
    STATE(4689), 1,
      sym_parenthesized_operator,
    STATE(6634), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4603), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [71435] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1096), 1,
      sym__lowercase_identifier,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2242), 1,
      anon_sym_POUND,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3572), 1,
      anon_sym_LPAREN,
    ACTIONS(3574), 1,
      sym__uppercase_identifier,
    STATE(2303), 1,
      sym_attribute,
    STATE(4380), 1,
      sym__simple_pattern,
    STATE(4455), 1,
      sym__constructor_name,
    STATE(6633), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4430), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [71536] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3581), 1,
      anon_sym_else,
    STATE(2301), 1,
      sym_else_clause,
    STATE(2304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 30,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [71599] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3510), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3508), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [71658] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2272), 1,
      sym__lowercase_identifier,
    ACTIONS(2276), 1,
      anon_sym_LBRACK,
    ACTIONS(2280), 1,
      anon_sym_LBRACE,
    ACTIONS(2284), 1,
      anon_sym_POUND,
    ACTIONS(2286), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2290), 1,
      anon_sym_begin,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2298), 1,
      aux_sym_number_token1,
    ACTIONS(2300), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2302), 1,
      anon_sym_SQUOTE,
    ACTIONS(2304), 1,
      anon_sym_DQUOTE,
    ACTIONS(3558), 1,
      anon_sym_LPAREN,
    ACTIONS(3560), 1,
      sym__uppercase_identifier,
    ACTIONS(3562), 1,
      aux_sym_tag_token1,
    STATE(2306), 1,
      sym_attribute,
    STATE(4883), 1,
      sym__simple_pattern,
    STATE(4980), 1,
      sym__constructor_name,
    STATE(6671), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2278), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4788), 2,
      sym__constant,
      sym_signed_number,
    STATE(4738), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4959), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4958), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [71759] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2307), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3518), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3516), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [71818] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3534), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3532), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [71877] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2309), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3514), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3512), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [71936] = 27,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2284), 1,
      anon_sym_POUND,
    ACTIONS(2290), 1,
      anon_sym_begin,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2298), 1,
      aux_sym_number_token1,
    ACTIONS(2300), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2302), 1,
      anon_sym_SQUOTE,
    ACTIONS(2304), 1,
      anon_sym_DQUOTE,
    ACTIONS(2310), 1,
      sym__lowercase_identifier,
    ACTIONS(2314), 1,
      anon_sym_LBRACK,
    ACTIONS(2316), 1,
      anon_sym_LBRACE,
    ACTIONS(2320), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3560), 1,
      sym__uppercase_identifier,
    ACTIONS(3562), 1,
      aux_sym_tag_token1,
    ACTIONS(3583), 1,
      anon_sym_LPAREN,
    STATE(2310), 1,
      sym_attribute,
    STATE(4894), 1,
      sym_parenthesized_operator,
    STATE(4980), 1,
      sym__constructor_name,
    STATE(4997), 1,
      sym__simple_binding_pattern,
    STATE(6672), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2278), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4788), 2,
      sym__constant,
      sym_signed_number,
    STATE(4738), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4969), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4965), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [72039] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3343), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3341), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [72098] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1086), 1,
      sym__uppercase_identifier,
    ACTIONS(1088), 1,
      aux_sym_tag_token1,
    ACTIONS(1482), 1,
      sym__lowercase_identifier,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1504), 1,
      anon_sym_LBRACE,
    ACTIONS(1510), 1,
      anon_sym_POUND,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1516), 1,
      anon_sym_begin,
    ACTIONS(1524), 1,
      aux_sym_number_token1,
    ACTIONS(1526), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1528), 1,
      anon_sym_SQUOTE,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2386), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2388), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3585), 1,
      anon_sym_LPAREN,
    STATE(1582), 1,
      sym__constructor_name,
    STATE(2156), 1,
      sym__simple_pattern,
    STATE(2312), 1,
      sym_attribute,
    STATE(6882), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1502), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2251), 2,
      sym__constant,
      sym_signed_number,
    STATE(2226), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2241), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [72199] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3367), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3365), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [72258] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3459), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3457), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [72317] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3463), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3461), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [72376] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3556), 1,
      anon_sym_PIPE,
    STATE(2299), 1,
      aux_sym__match_cases_repeat1,
    STATE(2316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 11,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [72439] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3375), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3373), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [72498] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1096), 1,
      sym__lowercase_identifier,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(3570), 1,
      anon_sym_LPAREN,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(2318), 1,
      sym_attribute,
    STATE(4380), 1,
      sym__simple_pattern,
    STATE(6550), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4430), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [72599] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    STATE(2319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [72660] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(1096), 1,
      sym__lowercase_identifier,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1108), 1,
      anon_sym_LBRACE,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3566), 1,
      sym__uppercase_identifier,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3570), 1,
      anon_sym_LPAREN,
    STATE(2320), 1,
      sym_attribute,
    STATE(3026), 1,
      sym__constructor_name,
    STATE(4398), 1,
      sym__simple_pattern,
    STATE(6593), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4430), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [72761] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [72820] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3530), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3528), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [72879] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2323), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3371), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3369), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [72938] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2029), 1,
      sym__lowercase_identifier,
    ACTIONS(2043), 1,
      anon_sym_LBRACK,
    ACTIONS(2047), 1,
      anon_sym_LBRACE,
    ACTIONS(2049), 1,
      anon_sym_POUND,
    ACTIONS(2051), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2053), 1,
      anon_sym_begin,
    ACTIONS(2061), 1,
      aux_sym_number_token1,
    ACTIONS(2063), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2065), 1,
      anon_sym_SQUOTE,
    ACTIONS(2067), 1,
      anon_sym_DQUOTE,
    ACTIONS(2069), 1,
      sym__uppercase_identifier,
    ACTIONS(2071), 1,
      aux_sym_tag_token1,
    ACTIONS(3587), 1,
      anon_sym_LPAREN,
    ACTIONS(3589), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3591), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2324), 1,
      sym_attribute,
    STATE(2458), 1,
      sym__simple_pattern,
    STATE(2515), 1,
      sym__constructor_name,
    STATE(6504), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2045), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2464), 2,
      sym__constant,
      sym_signed_number,
    STATE(2478), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2504), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2450), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [73039] = 27,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1314), 1,
      sym__lowercase_identifier,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1328), 1,
      anon_sym_LBRACE,
    ACTIONS(1332), 1,
      anon_sym_POUND,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1338), 1,
      anon_sym_begin,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(1354), 1,
      sym__uppercase_identifier,
    ACTIONS(1356), 1,
      aux_sym_tag_token1,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    ACTIONS(2422), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2424), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1735), 1,
      sym__constructor_name,
    STATE(2000), 1,
      sym__simple_binding_pattern,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(2325), 1,
      sym_attribute,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1326), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2001), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [73142] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3581), 1,
      anon_sym_else,
    STATE(2273), 1,
      sym_else_clause,
    STATE(2326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 30,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [73205] = 27,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2284), 1,
      anon_sym_POUND,
    ACTIONS(2290), 1,
      anon_sym_begin,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2298), 1,
      aux_sym_number_token1,
    ACTIONS(2300), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2302), 1,
      anon_sym_SQUOTE,
    ACTIONS(2304), 1,
      anon_sym_DQUOTE,
    ACTIONS(2310), 1,
      sym__lowercase_identifier,
    ACTIONS(2314), 1,
      anon_sym_LBRACK,
    ACTIONS(2316), 1,
      anon_sym_LBRACE,
    ACTIONS(2320), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3560), 1,
      sym__uppercase_identifier,
    ACTIONS(3562), 1,
      aux_sym_tag_token1,
    ACTIONS(3583), 1,
      anon_sym_LPAREN,
    STATE(2327), 1,
      sym_attribute,
    STATE(4894), 1,
      sym_parenthesized_operator,
    STATE(4980), 1,
      sym__constructor_name,
    STATE(5030), 1,
      sym__simple_binding_pattern,
    STATE(6672), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2278), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4788), 2,
      sym__constant,
      sym_signed_number,
    STATE(4738), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4969), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4965), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [73308] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2728), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2726), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [73367] = 27,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1038), 1,
      sym__lowercase_identifier,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1054), 1,
      anon_sym_LBRACE,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1068), 1,
      anon_sym_begin,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2147), 1,
      anon_sym_POUND,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(3564), 1,
      anon_sym_LPAREN,
    ACTIONS(3566), 1,
      sym__uppercase_identifier,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    STATE(2329), 1,
      sym_attribute,
    STATE(3026), 1,
      sym__constructor_name,
    STATE(3541), 1,
      sym_parenthesized_operator,
    STATE(4583), 1,
      sym__simple_binding_pattern,
    STATE(6594), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1052), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4603), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [73470] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3593), 1,
      anon_sym_PIPE,
    STATE(2330), 1,
      sym_attribute,
    STATE(2342), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3224), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3222), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73532] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3595), 1,
      anon_sym_else,
    STATE(2198), 1,
      sym_else_clause,
    STATE(2331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 29,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73594] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3317), 1,
      anon_sym_LPAREN,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(3325), 1,
      anon_sym_LBRACK,
    ACTIONS(3597), 1,
      anon_sym_DOT_DOT,
    STATE(2332), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3533), 1,
      sym__type,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3555), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [73704] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3295), 1,
      anon_sym_LPAREN,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(3305), 1,
      anon_sym_LBRACK,
    ACTIONS(3309), 1,
      anon_sym_LBRACE,
    ACTIONS(3311), 1,
      sym__uppercase_identifier,
    ACTIONS(3599), 1,
      anon_sym_DOT_DOT,
    STATE(2333), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4554), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(4938), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3307), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4632), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [73814] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3593), 1,
      anon_sym_PIPE,
    STATE(2330), 1,
      aux_sym__match_cases_repeat1,
    STATE(2334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73876] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3317), 1,
      anon_sym_LPAREN,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(3325), 1,
      anon_sym_LBRACK,
    ACTIONS(3601), 1,
      anon_sym_DOT_DOT,
    STATE(2335), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3542), 1,
      sym__type,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3595), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [73986] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3593), 1,
      anon_sym_PIPE,
    STATE(2336), 1,
      sym_attribute,
    STATE(2343), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [74048] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(3228), 1,
      sym__lowercase_identifier,
    ACTIONS(3230), 1,
      anon_sym_LPAREN,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(3240), 1,
      anon_sym_LBRACK,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3603), 1,
      anon_sym_DOT_DOT,
    STATE(2337), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3364), 1,
      sym__type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3480), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [74158] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3295), 1,
      anon_sym_LPAREN,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(3305), 1,
      anon_sym_LBRACK,
    ACTIONS(3309), 1,
      anon_sym_LBRACE,
    ACTIONS(3311), 1,
      sym__uppercase_identifier,
    ACTIONS(3605), 1,
      anon_sym_DOT_DOT,
    STATE(2338), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4551), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(4938), 1,
      sym_constructor_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3307), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4662), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [74268] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3595), 1,
      anon_sym_else,
    STATE(2183), 1,
      sym_else_clause,
    STATE(2339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 29,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [74330] = 31,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(3228), 1,
      sym__lowercase_identifier,
    ACTIONS(3230), 1,
      anon_sym_LPAREN,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(3240), 1,
      anon_sym_LBRACK,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3258), 1,
      sym__uppercase_identifier,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3607), 1,
      anon_sym_DOT_DOT,
    STATE(2340), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(3458), 1,
      sym__type,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3242), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3498), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [74440] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2932), 1,
      anon_sym_COMMA,
    STATE(2341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 30,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [74500] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3609), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2342), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3264), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [74560] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3593), 1,
      anon_sym_PIPE,
    STATE(2342), 1,
      aux_sym__match_cases_repeat1,
    STATE(2343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [74622] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1710), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1708), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74679] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3612), 1,
      anon_sym_DOT,
    ACTIONS(3614), 1,
      anon_sym_POUND,
    ACTIONS(3616), 1,
      sym_hash_operator,
    STATE(2345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2109), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2107), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74742] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1856), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1854), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74799] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1620), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1618), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74856] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1636), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1634), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74913] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1644), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1642), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74970] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1726), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1724), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75027] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1706), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1704), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75084] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1730), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1728), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75141] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2353), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1734), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1732), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75198] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1742), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1740), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75255] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2355), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1876), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1874), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75312] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1584), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1582), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75369] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2357), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1760), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1758), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75426] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2358), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1652), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1650), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75483] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1828), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1826), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75540] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2360), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1836), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1834), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75597] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1844), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1842), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75654] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2362), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1852), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1850), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75711] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2363), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1840), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1838), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75768] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1860), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1858), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75825] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2365), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1864), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1862), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75882] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2366), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1868), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1866), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75939] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2367), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1872), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1870), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75996] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2368), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1880), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1878), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76053] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2369), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1884), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1882), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76110] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2370), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1576), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1574), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76167] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2371), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1892), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1890), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76224] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3612), 1,
      anon_sym_DOT,
    STATE(2372), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1896), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1894), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76283] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2373), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1912), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1910), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76340] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2374), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1948), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1946), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76397] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2375), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1900), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1898), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76454] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2376), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1738), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1736), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76511] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2377), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1748), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1746), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76568] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2378), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1588), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1586), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76625] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2379), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1592), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1590), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76682] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2380), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1612), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1610), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76739] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1632), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1630), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76796] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2382), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1640), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1638), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76853] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2383), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1696), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1694), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76910] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2384), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1714), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1712), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76967] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2385), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1722), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1720), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77024] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2386), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1556), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1554), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77081] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1552), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1550), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77138] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2388), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1572), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1570), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77195] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2389), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1824), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1822), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77252] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1848), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1846), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77309] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2391), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1888), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1886), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77366] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2392), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1916), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1914), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77423] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2393), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1944), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1942), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77480] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2394), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1560), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1558), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77537] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2395), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1564), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1562), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77594] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2396), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1568), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1566), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77651] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2397), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1596), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1594), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77708] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1600), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1598), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77765] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1604), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1602), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77822] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2400), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1608), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1606), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77879] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2401), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1628), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1626), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77936] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1718), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1716), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77993] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2403), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1920), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1918), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78050] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3612), 1,
      anon_sym_DOT,
    ACTIONS(3614), 1,
      anon_sym_POUND,
    ACTIONS(3616), 1,
      sym_hash_operator,
    STATE(2404), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2081), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2079), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78113] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2405), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1616), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1614), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78170] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1648), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1646), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78227] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1580), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1578), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78284] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2408), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1756), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1754), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78341] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2409), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1548), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1546), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78398] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(2410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1698), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78457] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    STATE(2411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 29,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [78516] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1952), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1950), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78573] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(2413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1698), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78632] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2414), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1904), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1902), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78689] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2415), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1908), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1906), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78746] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2416), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1924), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1922), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78803] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2417), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1752), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1750), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78860] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1928), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1926), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78917] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2419), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1832), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1830), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78974] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2420), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1932), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1930), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79031] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2421), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1936), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1934), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79088] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2422), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1940), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1938), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79145] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2423), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1624), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1622), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79202] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2424), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1944), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1942), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79258] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2425), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1600), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1598), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79314] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2426), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1604), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1602), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79370] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2427), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1608), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1606), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79426] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2428), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1628), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1626), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79482] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2429), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3130), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3132), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79538] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3104), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3102), 21,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [79594] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2431), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3206), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3208), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79650] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2432), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1952), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1950), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79706] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2027), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2025), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79762] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2434), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1734), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1732), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79818] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(2435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1698), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [79876] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2436), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1904), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1902), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79932] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2437), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1908), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1906), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79988] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2438), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1924), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1922), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [80044] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3170), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3172), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [80100] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3618), 1,
      anon_sym_COLON,
    STATE(2440), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3289), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3291), 22,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80158] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2441), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1868), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1866), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [80214] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2442), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3186), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3188), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [80270] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3190), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3192), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [80326] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1748), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1746), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [80382] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2445), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2023), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2021), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [80438] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(2446), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1698), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80496] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1714), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1712), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [80552] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2448), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1888), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1886), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [80608] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1916), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1914), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [80664] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2145), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80719] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2451), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3104), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3102), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80774] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2452), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3289), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3291), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80829] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2453), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3425), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3427), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80884] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3483), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3485), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80939] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2455), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3487), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3489), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80994] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3361), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3363), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81049] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3106), 21,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [81104] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3389), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3391), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81159] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3397), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3399), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81214] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1756), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1754), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81269] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3620), 1,
      anon_sym_RPAREN,
    ACTIONS(3622), 1,
      anon_sym_module,
    ACTIONS(3624), 1,
      anon_sym_COLON_COLON,
    STATE(2461), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5339), 1,
      sym__type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6559), 1,
      sym_polymorphic_type,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [81378] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2462), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1952), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1950), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81433] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2463), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1828), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1826), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81488] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2464), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2972), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2970), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81543] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3393), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3395), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81598] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1620), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1618), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81653] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1904), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1902), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81708] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2728), 12,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2726), 28,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [81763] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2469), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1908), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1906), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81818] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2470), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3401), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3403), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81873] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3220), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3218), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81928] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2976), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2974), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81983] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3429), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3431), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82038] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2474), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1924), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1922), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82093] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2475), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3520), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3522), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82148] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3524), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3526), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82203] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2477), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1836), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1834), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82258] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2478), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3544), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3546), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82313] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2479), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3477), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3479), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82368] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1844), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1842), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82423] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3006), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3004), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82478] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2482), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1852), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1850), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82533] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2483), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3026), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3024), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82588] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2484), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3030), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3028), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82643] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2485), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1652), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1650), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82698] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(419), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3620), 1,
      anon_sym_RPAREN,
    ACTIONS(3626), 1,
      anon_sym_module,
    STATE(2486), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5415), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6375), 1,
      sym_polymorphic_type,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [82807] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2487), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3495), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3497), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82862] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3473), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3475), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82917] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2489), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3499), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3501), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82972] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2490), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1624), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1622), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83027] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2491), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1752), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1750), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83082] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1710), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1708), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83137] = 32,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1196), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3626), 1,
      anon_sym_module,
    ACTIONS(3628), 1,
      anon_sym_RPAREN,
    STATE(2493), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5415), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(6900), 1,
      sym_polymorphic_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [83246] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2494), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2732), 12,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2730), 28,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [83301] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3630), 1,
      anon_sym_COLON,
    STATE(2495), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2165), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2161), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [83358] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2496), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3554), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83413] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2497), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3536), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3538), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83468] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2498), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3548), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3550), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83523] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2499), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3345), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3347), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83578] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3491), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3493), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83633] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2501), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3417), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3419), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83688] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3349), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3351), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83743] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3465), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3467), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83798] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1636), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1634), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83853] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2505), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1928), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1926), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83908] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2506), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1932), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1930), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83963] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1936), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1934), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [84018] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1940), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1938), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [84073] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2509), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1856), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1854), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [84128] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2510), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3106), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [84183] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2511), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2736), 12,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2734), 28,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [84238] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2512), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3000), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3002), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [84293] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3381), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3383), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [84348] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1644), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1642), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [84403] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2515), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1832), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1830), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [84458] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3064), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3062), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [84513] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3634), 1,
      anon_sym_PIPE,
    ACTIONS(3636), 1,
      anon_sym_RBRACK,
    STATE(2517), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5697), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [84617] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3620), 1,
      anon_sym_RBRACK,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3638), 1,
      anon_sym_PIPE,
    STATE(2518), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5982), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7054), 1,
      sym__tag_spec,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [84721] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2519), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2125), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2123), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [84775] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3620), 1,
      anon_sym_RBRACK,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3640), 1,
      anon_sym_PIPE,
    STATE(2520), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(6166), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6389), 1,
      sym__tag_spec,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [84879] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    STATE(2521), 1,
      sym_attribute,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3642), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3646), 18,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
  [84943] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2522), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2081), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2079), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [84997] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3650), 1,
      anon_sym_PIPE,
    ACTIONS(3652), 1,
      anon_sym_RBRACK,
    STATE(2523), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5548), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [85101] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3628), 1,
      anon_sym_RBRACK,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3640), 1,
      anon_sym_PIPE,
    STATE(2524), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(6166), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6389), 1,
      sym__tag_spec,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [85205] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3656), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(3654), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
  [85259] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3662), 1,
      anon_sym_POUND,
    STATE(2526), 1,
      sym_attribute,
    STATE(2625), 1,
      sym_type_constructor_path,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3658), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3660), 18,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
  [85323] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3664), 1,
      anon_sym_PIPE,
    ACTIONS(3666), 1,
      anon_sym_RBRACK,
    STATE(2527), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5571), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [85427] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    ACTIONS(3676), 1,
      sym_assign_operator,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2528), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [85514] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3626), 1,
      anon_sym_module,
    STATE(2529), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5415), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6375), 1,
      sym_polymorphic_type,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [85617] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3626), 1,
      anon_sym_module,
    STATE(2530), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5415), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(6968), 1,
      sym_polymorphic_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [85720] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3626), 1,
      anon_sym_module,
    STATE(2531), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5415), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7084), 1,
      sym_polymorphic_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [85823] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3678), 1,
      anon_sym_RBRACK,
    STATE(2532), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(5714), 1,
      sym__tag_spec,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [85924] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3640), 1,
      anon_sym_PIPE,
    STATE(2533), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(6166), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6389), 1,
      sym__tag_spec,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [86025] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3640), 1,
      anon_sym_PIPE,
    STATE(2534), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5094), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(6166), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6389), 1,
      sym__tag_spec,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [86126] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3680), 1,
      sym__lowercase_identifier,
    ACTIONS(3682), 1,
      anon_sym_LPAREN,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    STATE(2535), 1,
      sym_attribute,
    STATE(4308), 1,
      sym__simple_type,
    STATE(4452), 1,
      sym_type_variable,
    STATE(4786), 1,
      sym__proper_tuple_type,
    STATE(5325), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5417), 1,
      sym__type,
    STATE(5805), 1,
      sym__polymorphic_type,
    STATE(5934), 1,
      sym_polymorphic_type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6365), 1,
      sym__abstract_type,
    STATE(6972), 1,
      sym_labeled_argument_type,
    STATE(7267), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [86229] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3696), 1,
      anon_sym_PIPE,
    STATE(2536), 1,
      sym_attribute,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5427), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [86330] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 23,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [86393] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3698), 1,
      anon_sym_COLON2,
    STATE(2538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3106), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [86448] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3702), 1,
      anon_sym_SEMI,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3708), 1,
      sym_rel_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2732), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2730), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [86537] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3228), 1,
      sym__lowercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3712), 1,
      anon_sym_LPAREN,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    STATE(2540), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3158), 1,
      sym_type_variable,
    STATE(3332), 1,
      sym_polymorphic_type,
    STATE(3350), 1,
      sym__polymorphic_type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(3919), 1,
      sym__type,
    STATE(5311), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6487), 1,
      sym__abstract_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [86640] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3702), 1,
      anon_sym_SEMI,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3708), 1,
      sym_rel_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2908), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2906), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [86729] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(3720), 1,
      anon_sym_DASH_GT,
    ACTIONS(3722), 1,
      anon_sym_STAR,
    STATE(2542), 1,
      sym_attribute,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 14,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3718), 17,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
  [86796] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3708), 1,
      sym_rel_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2846), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2844), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [86883] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3708), 1,
      sym_rel_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2892), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2890), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [86970] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2908), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    ACTIONS(3676), 1,
      sym_assign_operator,
    ACTIONS(3724), 1,
      anon_sym_SEMI,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2906), 10,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [87059] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3702), 1,
      anon_sym_SEMI,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3708), 1,
      sym_rel_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2546), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2736), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2734), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [87148] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2547), 1,
      sym_attribute,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3350), 1,
      sym__polymorphic_type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5747), 1,
      sym__type,
    STATE(5963), 1,
      sym_polymorphic_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [87251] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2548), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 25,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [87314] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2904), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2902), 12,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [87399] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3626), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2550), 1,
      sym_attribute,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5415), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6673), 1,
      sym_polymorphic_type,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [87502] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3640), 1,
      anon_sym_PIPE,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2551), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5064), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(6166), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6389), 1,
      sym__tag_spec,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [87603] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2912), 4,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [87676] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2553), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [87751] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3626), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2554), 1,
      sym_attribute,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5415), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6650), 1,
      sym_polymorphic_type,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [87854] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3726), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2555), 1,
      sym_attribute,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5349), 1,
      sym__type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    STATE(7310), 1,
      sym_polymorphic_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [87957] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3638), 1,
      anon_sym_PIPE,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2556), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5982), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7054), 1,
      sym__tag_spec,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [88058] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3728), 1,
      anon_sym_RBRACK,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2557), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5596), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [88159] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 21,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [88224] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 21,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [88289] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 7,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [88358] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2882), 2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 14,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
  [88437] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2562), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [88520] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3730), 1,
      anon_sym_PIPE,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2563), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(6217), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6901), 1,
      sym__tag_spec,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [88621] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2886), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    ACTIONS(3676), 1,
      sym_assign_operator,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [88708] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2736), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    ACTIONS(3676), 1,
      sym_assign_operator,
    ACTIONS(3724), 1,
      anon_sym_SEMI,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2734), 10,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [88797] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3708), 1,
      sym_rel_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2566), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2904), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2902), 11,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_assign_operator,
  [88882] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2732), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    ACTIONS(3676), 1,
      sym_assign_operator,
    ACTIONS(3724), 1,
      anon_sym_SEMI,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2567), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2730), 10,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [88971] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    ACTIONS(3676), 1,
      sym_assign_operator,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2568), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2844), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [89058] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2569), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2912), 6,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 14,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [89131] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 6,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [89206] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 6,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [89281] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3708), 1,
      sym_rel_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2572), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2882), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [89368] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 19,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [89433] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 19,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [89498] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2575), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 9,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 15,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [89567] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3708), 1,
      sym_rel_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2576), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2882), 3,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [89646] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3708), 1,
      sym_rel_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2577), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2882), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 12,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_assign_operator,
  [89729] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3708), 1,
      sym_rel_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2886), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [89816] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3732), 1,
      anon_sym_PIPE,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2579), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5353), 1,
      sym__tag_spec,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [89917] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3734), 1,
      anon_sym_PIPE,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2580), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5340), 1,
      sym__tag_spec,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [90018] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3640), 1,
      anon_sym_PIPE,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2581), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5208), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(6166), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6389), 1,
      sym__tag_spec,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [90119] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2892), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    ACTIONS(3676), 1,
      sym_assign_operator,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2890), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [90206] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3626), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2583), 1,
      sym_attribute,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5415), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(6900), 1,
      sym_polymorphic_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [90309] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3626), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2584), 1,
      sym_attribute,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5415), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(6911), 1,
      sym_polymorphic_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [90412] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2585), 1,
      sym_attribute,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3332), 1,
      sym_polymorphic_type,
    STATE(3350), 1,
      sym__polymorphic_type,
    STATE(3509), 1,
      sym__type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [90515] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3736), 1,
      anon_sym_COLON2,
    STATE(2586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3106), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [90570] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(3738), 1,
      anon_sym_RBRACK,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2587), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5654), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [90671] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3626), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2588), 1,
      sym_attribute,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5415), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    STATE(7242), 1,
      sym_polymorphic_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [90774] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(3744), 1,
      anon_sym_STAR,
    STATE(2589), 1,
      sym_attribute,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3740), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3742), 17,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
  [90839] = 30,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3622), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2590), 1,
      sym_attribute,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5339), 1,
      sym__type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6559), 1,
      sym_polymorphic_type,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [90942] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2591), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [91017] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3746), 1,
      sym__lowercase_identifier,
    ACTIONS(3748), 1,
      anon_sym_let,
    ACTIONS(3750), 1,
      anon_sym_LPAREN,
    ACTIONS(3752), 1,
      anon_sym_LBRACK,
    ACTIONS(3754), 1,
      anon_sym_object,
    STATE(2592), 1,
      sym_attribute,
    STATE(4780), 1,
      sym__simple_type,
    STATE(5287), 1,
      sym__class_type,
    STATE(6259), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3477), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(5331), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6518), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(5335), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [91111] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3756), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3758), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91163] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3760), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3762), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91215] = 29,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1046), 1,
      anon_sym_type,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2595), 1,
      sym_attribute,
    STATE(2666), 1,
      sym_type_variable,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(4244), 1,
      sym__type,
    STATE(5370), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6379), 1,
      sym_polymorphic_type,
    STATE(6723), 1,
      sym__abstract_type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [91315] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(231), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(261), 1,
      aux_sym_directive_token1,
    ACTIONS(3764), 1,
      anon_sym_let,
    ACTIONS(3766), 1,
      anon_sym_external,
    ACTIONS(3768), 1,
      anon_sym_type,
    ACTIONS(3770), 1,
      anon_sym_exception,
    ACTIONS(3772), 1,
      anon_sym_module,
    ACTIONS(3774), 1,
      anon_sym_open,
    ACTIONS(3776), 1,
      anon_sym_include,
    ACTIONS(3778), 1,
      anon_sym_class,
    ACTIONS(3780), 1,
      anon_sym_val,
    ACTIONS(3782), 1,
      sym_let_operator,
    STATE(40), 1,
      aux_sym__structure_repeat1,
    STATE(2596), 1,
      sym_attribute,
    STATE(2618), 1,
      aux_sym__structure_repeat2,
    STATE(2652), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(327), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4131), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4139), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4141), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [91407] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3784), 1,
      sym__lowercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3788), 1,
      anon_sym_LPAREN,
    ACTIONS(3790), 1,
      anon_sym_LBRACK,
    ACTIONS(3792), 1,
      anon_sym_object,
    STATE(2597), 1,
      sym_attribute,
    STATE(3629), 1,
      sym__class_type,
    STATE(4702), 1,
      sym__simple_type,
    STATE(5973), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3377), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3628), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6736), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3472), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [91501] = 28,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2598), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5596), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [91599] = 28,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2599), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5360), 1,
      sym__tag_spec,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [91697] = 28,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2600), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5616), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [91795] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3746), 1,
      sym__lowercase_identifier,
    ACTIONS(3748), 1,
      anon_sym_let,
    ACTIONS(3750), 1,
      anon_sym_LPAREN,
    ACTIONS(3752), 1,
      anon_sym_LBRACK,
    ACTIONS(3754), 1,
      anon_sym_object,
    STATE(2601), 1,
      sym_attribute,
    STATE(4780), 1,
      sym__simple_type,
    STATE(5299), 1,
      sym__class_type,
    STATE(6259), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3477), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(5331), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6518), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(5335), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [91889] = 28,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2602), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5560), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [91987] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3794), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3796), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [92039] = 28,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2604), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5561), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [92137] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2605), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3798), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3800), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [92189] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3746), 1,
      sym__lowercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3790), 1,
      anon_sym_LBRACK,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(3802), 1,
      anon_sym_LPAREN,
    STATE(2606), 1,
      sym_attribute,
    STATE(3691), 1,
      sym__class_type,
    STATE(4802), 1,
      sym__simple_type,
    STATE(6259), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3477), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3628), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6485), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3472), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [92283] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3746), 1,
      sym__lowercase_identifier,
    ACTIONS(3748), 1,
      anon_sym_let,
    ACTIONS(3750), 1,
      anon_sym_LPAREN,
    ACTIONS(3752), 1,
      anon_sym_LBRACK,
    ACTIONS(3754), 1,
      anon_sym_object,
    STATE(2607), 1,
      sym_attribute,
    STATE(4780), 1,
      sym__simple_type,
    STATE(5334), 1,
      sym__class_type,
    STATE(6259), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3477), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(5331), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6518), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(5335), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [92377] = 28,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2608), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(5714), 1,
      sym__tag_spec,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [92475] = 28,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2609), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5432), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [92573] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3804), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3806), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [92625] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3746), 1,
      sym__lowercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3790), 1,
      anon_sym_LBRACK,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(3802), 1,
      anon_sym_LPAREN,
    STATE(2611), 1,
      sym_attribute,
    STATE(3619), 1,
      sym__class_type,
    STATE(4802), 1,
      sym__simple_type,
    STATE(6259), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3477), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3628), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6485), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3472), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [92719] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3746), 1,
      sym__lowercase_identifier,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(3808), 1,
      anon_sym_let,
    ACTIONS(3810), 1,
      anon_sym_LPAREN,
    ACTIONS(3812), 1,
      anon_sym_LBRACK,
    STATE(2612), 1,
      sym_attribute,
    STATE(3619), 1,
      sym__class_type,
    STATE(4712), 1,
      sym__simple_type,
    STATE(6259), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3477), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3628), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6895), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3472), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [92813] = 28,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2613), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(5820), 1,
      sym__tag_spec,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [92911] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3814), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3816), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [92963] = 24,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3818), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(3821), 1,
      anon_sym_let,
    ACTIONS(3824), 1,
      anon_sym_external,
    ACTIONS(3827), 1,
      anon_sym_type,
    ACTIONS(3832), 1,
      anon_sym_exception,
    ACTIONS(3835), 1,
      anon_sym_module,
    ACTIONS(3838), 1,
      anon_sym_open,
    ACTIONS(3841), 1,
      anon_sym_include,
    ACTIONS(3844), 1,
      anon_sym_class,
    ACTIONS(3847), 1,
      anon_sym_val,
    ACTIONS(3850), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(3853), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(3856), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3859), 1,
      sym_let_operator,
    ACTIONS(3862), 1,
      aux_sym_directive_token1,
    STATE(43), 1,
      aux_sym__structure_repeat1,
    STATE(2652), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3830), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(2615), 2,
      sym_attribute,
      aux_sym__structure_repeat2,
    STATE(4131), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4139), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4141), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [93053] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2616), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3865), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3867), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93105] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3746), 1,
      sym__lowercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3790), 1,
      anon_sym_LBRACK,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(3802), 1,
      anon_sym_LPAREN,
    STATE(2617), 1,
      sym_attribute,
    STATE(3629), 1,
      sym__class_type,
    STATE(4802), 1,
      sym__simple_type,
    STATE(6259), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3477), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3628), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6485), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3472), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [93199] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(231), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(261), 1,
      aux_sym_directive_token1,
    ACTIONS(3764), 1,
      anon_sym_let,
    ACTIONS(3766), 1,
      anon_sym_external,
    ACTIONS(3768), 1,
      anon_sym_type,
    ACTIONS(3770), 1,
      anon_sym_exception,
    ACTIONS(3772), 1,
      anon_sym_module,
    ACTIONS(3774), 1,
      anon_sym_open,
    ACTIONS(3776), 1,
      anon_sym_include,
    ACTIONS(3778), 1,
      anon_sym_class,
    ACTIONS(3780), 1,
      anon_sym_val,
    ACTIONS(3782), 1,
      sym_let_operator,
    STATE(36), 1,
      aux_sym__structure_repeat1,
    STATE(2615), 1,
      aux_sym__structure_repeat2,
    STATE(2618), 1,
      sym_attribute,
    STATE(2652), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(335), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4131), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4139), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4141), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [93291] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2619), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3869), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3871), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93343] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3873), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3875), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93395] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3877), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3879), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93447] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2622), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3883), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93499] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2623), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3658), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3660), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93551] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(231), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(261), 1,
      aux_sym_directive_token1,
    ACTIONS(3764), 1,
      anon_sym_let,
    ACTIONS(3766), 1,
      anon_sym_external,
    ACTIONS(3768), 1,
      anon_sym_type,
    ACTIONS(3770), 1,
      anon_sym_exception,
    ACTIONS(3772), 1,
      anon_sym_module,
    ACTIONS(3774), 1,
      anon_sym_open,
    ACTIONS(3776), 1,
      anon_sym_include,
    ACTIONS(3778), 1,
      anon_sym_class,
    ACTIONS(3780), 1,
      anon_sym_val,
    ACTIONS(3782), 1,
      sym_let_operator,
    STATE(36), 1,
      aux_sym__structure_repeat1,
    STATE(2624), 1,
      sym_attribute,
    STATE(2634), 1,
      aux_sym__structure_repeat2,
    STATE(2652), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(335), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4131), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4139), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4141), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [93643] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2625), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3885), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3887), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93695] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3746), 1,
      sym__lowercase_identifier,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(3808), 1,
      anon_sym_let,
    ACTIONS(3810), 1,
      anon_sym_LPAREN,
    ACTIONS(3812), 1,
      anon_sym_LBRACK,
    STATE(2626), 1,
      sym_attribute,
    STATE(3629), 1,
      sym__class_type,
    STATE(4712), 1,
      sym__simple_type,
    STATE(6259), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3477), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3628), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6895), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3472), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [93789] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3889), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3891), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93841] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2736), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    ACTIONS(3676), 1,
      sym_assign_operator,
    ACTIONS(3893), 1,
      anon_sym_SEMI,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2628), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2734), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [93929] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3895), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3897), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93981] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2908), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    ACTIONS(3676), 1,
      sym_assign_operator,
    ACTIONS(3893), 1,
      anon_sym_SEMI,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2906), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [94069] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2732), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    ACTIONS(3670), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3672), 1,
      sym_concat_operator,
    ACTIONS(3674), 1,
      sym_rel_operator,
    ACTIONS(3676), 1,
      sym_assign_operator,
    ACTIONS(3893), 1,
      anon_sym_SEMI,
    STATE(691), 1,
      sym_pow_operator,
    STATE(692), 1,
      sym_mult_operator,
    STATE(693), 1,
      sym_add_operator,
    STATE(694), 1,
      sym_and_operator,
    STATE(695), 1,
      sym_or_operator,
    STATE(2631), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2730), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [94157] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3784), 1,
      sym__lowercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3788), 1,
      anon_sym_LPAREN,
    ACTIONS(3790), 1,
      anon_sym_LBRACK,
    ACTIONS(3792), 1,
      anon_sym_object,
    STATE(2632), 1,
      sym_attribute,
    STATE(3691), 1,
      sym__class_type,
    STATE(4702), 1,
      sym__simple_type,
    STATE(5973), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3377), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3628), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6736), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3472), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [94251] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3899), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3901), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [94303] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(231), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(261), 1,
      aux_sym_directive_token1,
    ACTIONS(3764), 1,
      anon_sym_let,
    ACTIONS(3766), 1,
      anon_sym_external,
    ACTIONS(3768), 1,
      anon_sym_type,
    ACTIONS(3770), 1,
      anon_sym_exception,
    ACTIONS(3772), 1,
      anon_sym_module,
    ACTIONS(3774), 1,
      anon_sym_open,
    ACTIONS(3776), 1,
      anon_sym_include,
    ACTIONS(3778), 1,
      anon_sym_class,
    ACTIONS(3780), 1,
      anon_sym_val,
    ACTIONS(3782), 1,
      sym_let_operator,
    STATE(35), 1,
      aux_sym__structure_repeat1,
    STATE(2615), 1,
      aux_sym__structure_repeat2,
    STATE(2634), 1,
      sym_attribute,
    STATE(2652), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(323), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4131), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4139), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4141), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [94395] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2635), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3903), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3905), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [94447] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3907), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3909), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [94499] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2637), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3911), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3913), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [94551] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3915), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3917), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [94603] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2639), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3919), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3921), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [94655] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2640), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3923), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3925), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [94707] = 28,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2641), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5654), 1,
      sym__tag_spec,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [94805] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2642), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3927), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3929), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [94857] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3931), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3933), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [94909] = 28,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3568), 1,
      aux_sym_tag_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2644), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5297), 1,
      sym__tag_spec,
    STATE(5418), 1,
      sym__type,
    STATE(5419), 1,
      sym_tag,
    STATE(5698), 1,
      sym_tag_specification,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [95007] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3935), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3937), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [95059] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3939), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3941), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [95111] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3784), 1,
      sym__lowercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3788), 1,
      anon_sym_LPAREN,
    ACTIONS(3790), 1,
      anon_sym_LBRACK,
    ACTIONS(3792), 1,
      anon_sym_object,
    STATE(2647), 1,
      sym_attribute,
    STATE(3619), 1,
      sym__class_type,
    STATE(4702), 1,
      sym__simple_type,
    STATE(5973), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3377), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3628), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6736), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3472), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [95205] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2648), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3943), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3945), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [95257] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3947), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3949), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [95309] = 26,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3746), 1,
      sym__lowercase_identifier,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(3808), 1,
      anon_sym_let,
    ACTIONS(3810), 1,
      anon_sym_LPAREN,
    ACTIONS(3812), 1,
      anon_sym_LBRACK,
    STATE(2650), 1,
      sym_attribute,
    STATE(3691), 1,
      sym__class_type,
    STATE(4712), 1,
      sym__simple_type,
    STATE(6259), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3477), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3628), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6895), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3472), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2629), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [95403] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3955), 1,
      anon_sym_RPAREN,
    STATE(2651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3953), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3951), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95456] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1038), 1,
      sym__lowercase_identifier,
    ACTIONS(3961), 1,
      anon_sym_LPAREN,
    ACTIONS(3965), 1,
      anon_sym_LBRACE,
    ACTIONS(3967), 1,
      aux_sym_number_token1,
    ACTIONS(3969), 1,
      anon_sym_SQUOTE,
    ACTIONS(3971), 1,
      anon_sym_DQUOTE,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    STATE(2652), 1,
      sym_attribute,
    STATE(3541), 1,
      sym_parenthesized_operator,
    STATE(4062), 1,
      sym_module_path,
    STATE(4118), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3963), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4114), 2,
      sym__constant,
      sym_value_path,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    ACTIONS(3957), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
    ACTIONS(3959), 10,
      anon_sym_let,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
  [95533] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(2653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3977), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3975), 21,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95584] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(2654), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3953), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3951), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95634] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3981), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3979), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [95684] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3985), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3983), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [95734] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(2657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3989), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3987), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95784] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2736), 1,
      anon_sym_COLON2,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    ACTIONS(3991), 1,
      anon_sym_SEMI,
    ACTIONS(3993), 1,
      sym_rel_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2734), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
  [95870] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(2659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3955), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3995), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95920] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3999), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3997), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [95970] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4003), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4001), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [96020] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4007), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4005), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [96070] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(2663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4011), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4009), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [96120] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(2664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4015), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4013), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [96170] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2732), 1,
      anon_sym_COLON2,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    ACTIONS(3991), 1,
      anon_sym_SEMI,
    ACTIONS(3993), 1,
      sym_rel_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2730), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
  [96256] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4017), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
    ACTIONS(3895), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3897), 18,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [96308] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3718), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3716), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [96358] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4021), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4019), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [96408] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2908), 1,
      anon_sym_COLON2,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    ACTIONS(3991), 1,
      anon_sym_SEMI,
    ACTIONS(3993), 1,
      sym_rel_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(2669), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2906), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
  [96494] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(2670), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4025), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(4023), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [96544] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2671), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1838), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1840), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96593] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2672), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1716), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1718), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96642] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3726), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2673), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(6143), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [96731] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2674), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1926), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1928), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96780] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1930), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1932), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96829] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1858), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1860), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96878] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1866), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1868), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96927] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1862), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1864), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96976] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1934), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1936), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97025] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1910), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1912), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97074] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1886), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1888), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97123] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1724), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1726), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97172] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1878), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1880), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97221] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1712), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1714), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97270] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2685), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1946), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1948), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97319] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2686), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1754), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1756), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97368] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2687), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1950), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1952), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97417] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2688), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1874), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1876), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97466] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2689), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1898), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1900), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97515] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_private,
    STATE(2690), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4442), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [97604] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1736), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1738), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97653] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1882), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1884), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97702] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1842), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1844), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97751] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1902), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1904), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97800] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3228), 1,
      sym__lowercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3712), 1,
      anon_sym_LPAREN,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4031), 1,
      anon_sym_private,
    STATE(2695), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3251), 1,
      sym__type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [97889] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1746), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1748), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97938] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1650), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1652), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97987] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1906), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1908), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98036] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2021), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2023), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98085] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1566), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1568), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98134] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1598), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1600), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98183] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1546), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1548), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98232] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4033), 1,
      aux_sym_attribute_id_token1,
    STATE(2703), 1,
      sym_attribute,
    STATE(2721), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2754), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2752), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [98285] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1602), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1604), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98334] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2705), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1854), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1856), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98383] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2706), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1582), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1584), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98432] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1922), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1924), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98481] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4035), 1,
      anon_sym_private,
    STATE(2708), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4444), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [98570] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(4037), 1,
      anon_sym_AMP,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2709), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(4921), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [98659] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1622), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1624), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98708] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1614), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1616), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98757] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1646), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1648), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98806] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1704), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1706), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98855] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1870), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1872), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98904] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1606), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1608), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98953] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4039), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2716), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(2758), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2756), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [99004] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1586), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1588), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99053] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1590), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1592), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99102] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3622), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2719), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5339), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [99191] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(2720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1700), 23,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99242] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4033), 1,
      aux_sym_attribute_id_token1,
    STATE(2716), 1,
      aux_sym_attribute_id_repeat1,
    STATE(2721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2744), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2742), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [99295] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1610), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1612), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99344] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1750), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1752), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99393] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1630), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1632), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99442] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4042), 1,
      anon_sym_DOT,
    ACTIONS(4044), 1,
      anon_sym_POUND,
    ACTIONS(4046), 1,
      sym_hash_operator,
    STATE(2725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2946), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2944), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99497] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4042), 1,
      anon_sym_DOT,
    ACTIONS(4044), 1,
      anon_sym_POUND,
    ACTIONS(4046), 1,
      sym_hash_operator,
    STATE(2726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2980), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2978), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99552] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1562), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1564), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99601] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4048), 1,
      anon_sym_else,
    STATE(2147), 1,
      sym_else_clause,
    STATE(2728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 22,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [99654] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1638), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1640), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99703] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4050), 1,
      anon_sym_private,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2730), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3285), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [99792] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(2731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1700), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [99843] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2732), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1694), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1696), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99892] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4042), 1,
      anon_sym_DOT,
    ACTIONS(4044), 1,
      anon_sym_POUND,
    ACTIONS(4046), 1,
      sym_hash_operator,
    STATE(2733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2966), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2964), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99947] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1618), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1620), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99996] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1720), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1722), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100045] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1554), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1556), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100094] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1642), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1644), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100143] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4052), 1,
      anon_sym_else,
    STATE(2172), 1,
      sym_else_clause,
    STATE(2738), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 24,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [100196] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3626), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2739), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6322), 1,
      sym__type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [100285] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1550), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1552), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100334] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1826), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1828), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100383] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4052), 1,
      anon_sym_else,
    STATE(2147), 1,
      sym_else_clause,
    STATE(2742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 24,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [100436] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1914), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1916), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100485] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1594), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1596), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100534] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1942), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1944), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100583] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1732), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1734), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100632] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(2747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1700), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [100683] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3726), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2748), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5349), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [100772] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1938), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1940), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100821] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1574), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1576), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100870] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(2751), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1700), 23,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100921] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1626), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1628), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100970] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1708), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1710), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101019] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1850), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1852), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101068] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1830), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1832), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101117] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1890), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1892), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101166] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1634), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1636), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101215] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1570), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1572), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101264] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3622), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2759), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(6055), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [101353] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1834), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1836), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101402] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4054), 1,
      anon_sym_private,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2761), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3289), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [101491] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2762), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1558), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1560), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101540] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2763), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2025), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2027), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101589] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4042), 1,
      anon_sym_DOT,
    STATE(2764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1894), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1896), 23,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101640] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2765), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1822), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1824), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101689] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4048), 1,
      anon_sym_else,
    STATE(2172), 1,
      sym_else_clause,
    STATE(2766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 22,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [101742] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2767), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1758), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1760), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101791] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2768), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1846), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1848), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101840] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4042), 1,
      anon_sym_DOT,
    ACTIONS(4044), 1,
      anon_sym_POUND,
    ACTIONS(4046), 1,
      sym_hash_operator,
    STATE(2769), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2992), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2990), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101895] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2770), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1578), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1580), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101944] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1740), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1742), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101993] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2772), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1918), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1920), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [102042] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4042), 1,
      anon_sym_DOT,
    ACTIONS(4044), 1,
      anon_sym_POUND,
    ACTIONS(4046), 1,
      sym_hash_operator,
    STATE(2773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2956), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2954), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [102097] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2774), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1728), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1730), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [102146] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2438), 1,
      sym__lowercase_identifier,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3626), 1,
      anon_sym_module,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2775), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5415), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102235] = 25,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3228), 1,
      sym__lowercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3712), 1,
      anon_sym_LPAREN,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4056), 1,
      anon_sym_private,
    STATE(2776), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3291), 1,
      sym__type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102324] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2777), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3238), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102410] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2778), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3289), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102496] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2779), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5484), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102582] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3682), 1,
      anon_sym_LPAREN,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4058), 1,
      sym__lowercase_identifier,
    STATE(2780), 1,
      sym_attribute,
    STATE(4308), 1,
      sym__simple_type,
    STATE(4786), 1,
      sym__proper_tuple_type,
    STATE(5375), 1,
      sym__type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6972), 1,
      sym_labeled_argument_type,
    STATE(7267), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102668] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3682), 1,
      anon_sym_LPAREN,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4058), 1,
      sym__lowercase_identifier,
    STATE(2781), 1,
      sym_attribute,
    STATE(4308), 1,
      sym__simple_type,
    STATE(4786), 1,
      sym__proper_tuple_type,
    STATE(5376), 1,
      sym__type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6972), 1,
      sym_labeled_argument_type,
    STATE(7267), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102754] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2782), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(6072), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102840] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2783), 1,
      sym_attribute,
    STATE(3028), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4062), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4060), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [102898] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2784), 1,
      sym_attribute,
    STATE(3028), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4072), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4070), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [102956] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2785), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5445), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103042] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2786), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5084), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103128] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3712), 1,
      anon_sym_LPAREN,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    STATE(2661), 1,
      sym__type,
    STATE(2787), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103214] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2788), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(4264), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103300] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4076), 1,
      anon_sym_COMMA,
    ACTIONS(4078), 1,
      anon_sym_SEMI,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    ACTIONS(4084), 1,
      sym_rel_operator,
    ACTIONS(4086), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2789), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2736), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2734), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [103384] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4076), 1,
      anon_sym_COMMA,
    ACTIONS(4078), 1,
      anon_sym_SEMI,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    ACTIONS(4084), 1,
      sym_rel_operator,
    ACTIONS(4086), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2790), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2732), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2730), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [103468] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2791), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5341), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103554] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2792), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3264), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103640] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3682), 1,
      anon_sym_LPAREN,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4058), 1,
      sym__lowercase_identifier,
    STATE(2793), 1,
      sym_attribute,
    STATE(4280), 1,
      sym__type,
    STATE(4308), 1,
      sym__simple_type,
    STATE(4786), 1,
      sym__proper_tuple_type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6972), 1,
      sym_labeled_argument_type,
    STATE(7267), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103726] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3742), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3744), 1,
      anon_sym_STAR,
    STATE(2794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3740), 31,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [103776] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2795), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(4244), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103862] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    STATE(2796), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4254), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103948] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4076), 1,
      anon_sym_COMMA,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    ACTIONS(4084), 1,
      sym_rel_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2904), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2902), 6,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [104028] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2912), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [104096] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [104166] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [104236] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4076), 1,
      anon_sym_COMMA,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    ACTIONS(4084), 1,
      sym_rel_operator,
    ACTIONS(4086), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2882), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [104318] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    ACTIONS(4084), 1,
      sym_rel_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2882), 3,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_PIPE_PIPE,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [104392] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    ACTIONS(4084), 1,
      sym_rel_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2882), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [104470] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4076), 1,
      anon_sym_COMMA,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    ACTIONS(4084), 1,
      sym_rel_operator,
    ACTIONS(4086), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2886), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [104552] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4076), 1,
      anon_sym_COMMA,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    ACTIONS(4084), 1,
      sym_rel_operator,
    ACTIONS(4086), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2846), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2844), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [104634] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4076), 1,
      anon_sym_COMMA,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    ACTIONS(4084), 1,
      sym_rel_operator,
    ACTIONS(4086), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2892), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2890), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [104716] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2807), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5333), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104802] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2662), 1,
      sym__type,
    STATE(2808), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104888] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2809), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5975), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104974] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    STATE(2810), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4590), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105060] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2811), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(6048), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105146] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2812), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(4271), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105232] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3712), 1,
      anon_sym_LPAREN,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    STATE(2813), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3238), 1,
      sym__type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105318] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3712), 1,
      anon_sym_LPAREN,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    STATE(2814), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3251), 1,
      sym__type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105404] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [105462] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2661), 1,
      sym__type,
    STATE(2816), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105548] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [105608] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2818), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [105668] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 8,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [105732] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2820), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6288), 1,
      sym__type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105818] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2758), 16,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
      aux_sym_attribute_id_token1,
    ACTIONS(2756), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [105866] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4076), 1,
      anon_sym_COMMA,
    ACTIONS(4078), 1,
      anon_sym_SEMI,
    ACTIONS(4080), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4082), 1,
      sym_concat_operator,
    ACTIONS(4084), 1,
      sym_rel_operator,
    ACTIONS(4086), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2908), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2906), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [105950] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2823), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(4826), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106036] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3712), 1,
      anon_sym_LPAREN,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    STATE(2824), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3254), 1,
      sym__type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106122] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    STATE(2825), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4280), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106208] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2826), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6326), 1,
      sym__type,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106294] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2827), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5354), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106380] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    STATE(2828), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4442), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106466] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2829), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(5359), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106552] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2830), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(6232), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106638] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1754), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1756), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [106686] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3682), 1,
      anon_sym_LPAREN,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4058), 1,
      sym__lowercase_identifier,
    STATE(2832), 1,
      sym_attribute,
    STATE(4308), 1,
      sym__simple_type,
    STATE(4786), 1,
      sym__proper_tuple_type,
    STATE(5408), 1,
      sym__type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6972), 1,
      sym_labeled_argument_type,
    STATE(7267), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106772] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3682), 1,
      anon_sym_LPAREN,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4058), 1,
      sym__lowercase_identifier,
    STATE(2833), 1,
      sym_attribute,
    STATE(4308), 1,
      sym__simple_type,
    STATE(4786), 1,
      sym__proper_tuple_type,
    STATE(5097), 1,
      sym__type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6972), 1,
      sym_labeled_argument_type,
    STATE(7267), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106858] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3712), 1,
      anon_sym_LPAREN,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    STATE(2662), 1,
      sym__type,
    STATE(2834), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106944] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    STATE(2835), 1,
      sym_attribute,
    STATE(4196), 1,
      sym__simple_type,
    STATE(4469), 1,
      sym__type,
    STATE(4592), 1,
      sym__proper_tuple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6415), 1,
      sym_labeled_tuple_element_type,
    STATE(6428), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107030] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1830), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1832), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [107078] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2837), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3513), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107164] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2442), 1,
      anon_sym_LPAREN,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3632), 1,
      sym__lowercase_identifier,
    STATE(2542), 1,
      sym__simple_type,
    STATE(2838), 1,
      sym_attribute,
    STATE(2871), 1,
      sym__proper_tuple_type,
    STATE(3511), 1,
      sym__type,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(7101), 1,
      sym_labeled_argument_type,
    STATE(7116), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107250] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1854), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1856), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [107298] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    STATE(2840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 22,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [107348] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3682), 1,
      anon_sym_LPAREN,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4058), 1,
      sym__lowercase_identifier,
    STATE(2841), 1,
      sym_attribute,
    STATE(4254), 1,
      sym__type,
    STATE(4308), 1,
      sym__simple_type,
    STATE(4786), 1,
      sym__proper_tuple_type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6972), 1,
      sym_labeled_argument_type,
    STATE(7267), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4311), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107434] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1708), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1710), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [107482] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3712), 1,
      anon_sym_LPAREN,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    STATE(2843), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(3988), 1,
      sym__type,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107568] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3712), 1,
      anon_sym_LPAREN,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    STATE(2844), 1,
      sym_attribute,
    STATE(2919), 1,
      sym__simple_type,
    STATE(3383), 1,
      sym__proper_tuple_type,
    STATE(3992), 1,
      sym__type,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6429), 1,
      sym_labeled_tuple_element_type,
    STATE(6935), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2667), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107654] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    STATE(2845), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 24,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [107704] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2904), 1,
      anon_sym_and,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2902), 6,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [107783] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2912), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [107850] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [107919] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [107988] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_PIPE,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    ACTIONS(4100), 1,
      anon_sym_COMMA,
    ACTIONS(4102), 1,
      sym_rel_operator,
    ACTIONS(4104), 1,
      sym_assign_operator,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [108069] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    ACTIONS(4102), 1,
      sym_rel_operator,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2882), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_or,
      sym_assign_operator,
  [108142] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_PIPE,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    ACTIONS(4102), 1,
      sym_rel_operator,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [108219] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2886), 1,
      anon_sym_PIPE,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    ACTIONS(4100), 1,
      anon_sym_COMMA,
    ACTIONS(4102), 1,
      sym_rel_operator,
    ACTIONS(4104), 1,
      sym_assign_operator,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [108300] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3977), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3975), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [108347] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_PIPE,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    ACTIONS(4100), 1,
      anon_sym_COMMA,
    ACTIONS(4102), 1,
      sym_rel_operator,
    ACTIONS(4104), 1,
      sym_assign_operator,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2844), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [108428] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3654), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(3656), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [108475] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2908), 1,
      anon_sym_PIPE,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    ACTIONS(4100), 1,
      anon_sym_COMMA,
    ACTIONS(4102), 1,
      sym_rel_operator,
    ACTIONS(4104), 1,
      sym_assign_operator,
    ACTIONS(4106), 1,
      anon_sym_SEMI,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2906), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [108558] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4108), 1,
      sym__lowercase_identifier,
    ACTIONS(4110), 1,
      anon_sym_POUND,
    STATE(2858), 1,
      sym_attribute,
    STATE(3116), 1,
      sym_type_constructor_path,
    STATE(5948), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3658), 11,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(3660), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [108615] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1754), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1756), 29,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [108662] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [108719] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2732), 1,
      anon_sym_PIPE,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    ACTIONS(4100), 1,
      anon_sym_COMMA,
    ACTIONS(4102), 1,
      sym_rel_operator,
    ACTIONS(4104), 1,
      sym_assign_operator,
    ACTIONS(4106), 1,
      anon_sym_SEMI,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2730), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [108802] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2892), 1,
      anon_sym_PIPE,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    ACTIONS(4100), 1,
      anon_sym_COMMA,
    ACTIONS(4102), 1,
      sym_rel_operator,
    ACTIONS(4104), 1,
      sym_assign_operator,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2890), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [108883] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2736), 1,
      anon_sym_and,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    ACTIONS(4112), 1,
      anon_sym_SEMI,
    ACTIONS(4114), 1,
      sym_assign_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2734), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [108966] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4116), 1,
      aux_sym_attribute_id_token1,
    STATE(2864), 1,
      sym_attribute,
    STATE(2888), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2754), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(2752), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [109017] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2892), 1,
      anon_sym_and,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    ACTIONS(4114), 1,
      sym_assign_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2890), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [109098] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_and,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    ACTIONS(4114), 1,
      sym_assign_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2844), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [109179] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [109236] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [109295] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [109354] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 7,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [109417] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3718), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3720), 1,
      anon_sym_DASH_GT,
    STATE(2871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 30,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [109466] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2736), 1,
      anon_sym_PIPE,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    ACTIONS(4100), 1,
      anon_sym_COMMA,
    ACTIONS(4102), 1,
      sym_rel_operator,
    ACTIONS(4104), 1,
      sym_assign_operator,
    ACTIONS(4106), 1,
      anon_sym_SEMI,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2734), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [109549] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2732), 1,
      anon_sym_and,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    ACTIONS(4112), 1,
      anon_sym_SEMI,
    ACTIONS(4114), 1,
      sym_assign_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2873), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2730), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [109632] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4118), 1,
      anon_sym_COLON,
    STATE(2874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [109681] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2886), 1,
      anon_sym_and,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    ACTIONS(4114), 1,
      sym_assign_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [109762] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4120), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2876), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(2758), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(2756), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [109811] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2904), 1,
      anon_sym_PIPE,
    ACTIONS(4096), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4098), 1,
      sym_concat_operator,
    ACTIONS(4100), 1,
      anon_sym_COMMA,
    ACTIONS(4102), 1,
      sym_rel_operator,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2902), 6,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [109890] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2908), 1,
      anon_sym_and,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    ACTIONS(4112), 1,
      anon_sym_SEMI,
    ACTIONS(4114), 1,
      sym_assign_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2906), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [109973] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2912), 4,
      anon_sym_and,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [110040] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_and,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [110109] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_and,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [110178] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_and,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    ACTIONS(4114), 1,
      sym_assign_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [110259] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2883), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [110318] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [110377] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    STATE(558), 1,
      sym_pow_operator,
    STATE(559), 1,
      sym_mult_operator,
    STATE(560), 1,
      sym_add_operator,
    STATE(561), 1,
      sym_and_operator,
    STATE(562), 1,
      sym_or_operator,
    STATE(2885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 7,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [110440] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2882), 2,
      anon_sym_and,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [110513] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_and,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [110590] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4116), 1,
      aux_sym_attribute_id_token1,
    STATE(2876), 1,
      aux_sym_attribute_id_repeat1,
    STATE(2888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2744), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(2742), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [110641] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    ACTIONS(4131), 1,
      sym_assign_operator,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [110719] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3264), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [110765] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2758), 10,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
      aux_sym_attribute_id_token1,
    ACTIONS(2756), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [110811] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3530), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3528), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [110857] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3534), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3532), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [110903] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    ACTIONS(4131), 1,
      sym_assign_operator,
    ACTIONS(4133), 1,
      anon_sym_SEMI,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2734), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [110983] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    ACTIONS(4131), 1,
      sym_assign_operator,
    ACTIONS(4133), 1,
      anon_sym_SEMI,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2895), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2730), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [111063] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3542), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3540), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111109] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4137), 1,
      anon_sym_let,
    STATE(2897), 1,
      sym_attribute,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3563), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4135), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [111171] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4108), 1,
      sym__lowercase_identifier,
    ACTIONS(4139), 1,
      anon_sym_STAR,
    ACTIONS(4141), 1,
      anon_sym_POUND,
    STATE(2898), 1,
      sym_attribute,
    STATE(3077), 1,
      sym_type_constructor_path,
    STATE(5948), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3740), 9,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(3742), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [111229] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3371), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3369), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111275] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3379), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3377), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111321] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2908), 1,
      anon_sym_and,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    ACTIONS(4114), 1,
      sym_assign_operator,
    ACTIONS(4143), 1,
      anon_sym_SEMI,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2906), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [111403] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2728), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2726), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111449] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(431), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(437), 1,
      sym_rel_operator,
    ACTIONS(439), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1048), 1,
      anon_sym_DOT,
    ACTIONS(4147), 1,
      anon_sym_or,
    STATE(2903), 1,
      sym_attribute,
    STATE(7225), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(417), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(435), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(433), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(429), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4145), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(1084), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(6700), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [111517] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3387), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3385), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111563] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3415), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3413), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111609] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3423), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3421), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111655] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(431), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(437), 1,
      sym_rel_operator,
    ACTIONS(439), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1104), 1,
      anon_sym_DOT,
    ACTIONS(4147), 1,
      anon_sym_or,
    STATE(2907), 1,
      sym_attribute,
    STATE(7105), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(417), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(435), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(433), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(429), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4145), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(1122), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(6700), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [111723] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2902), 6,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [111799] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2736), 1,
      anon_sym_and,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    ACTIONS(4114), 1,
      sym_assign_operator,
    ACTIONS(4143), 1,
      anon_sym_SEMI,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2909), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2734), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [111881] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2736), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2734), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111927] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2912), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2910), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [111993] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2882), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [112061] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2882), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [112129] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    ACTIONS(4131), 1,
      sym_assign_operator,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [112207] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2882), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_or,
      sym_assign_operator,
  [112279] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [112353] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2917), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3435), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3433), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112399] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    ACTIONS(4131), 1,
      sym_assign_operator,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2844), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [112477] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4108), 1,
      sym__lowercase_identifier,
    ACTIONS(4141), 1,
      anon_sym_POUND,
    ACTIONS(4149), 1,
      anon_sym_DASH_GT,
    ACTIONS(4151), 1,
      anon_sym_STAR,
    STATE(2919), 1,
      sym_attribute,
    STATE(3077), 1,
      sym_type_constructor_path,
    STATE(5948), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 8,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(3718), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [112537] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2920), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3443), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3441), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112583] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3447), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3445), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112629] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(413), 1,
      anon_sym_DOT,
    ACTIONS(431), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(437), 1,
      sym_rel_operator,
    ACTIONS(439), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4147), 1,
      anon_sym_or,
    STATE(2922), 1,
      sym_attribute,
    STATE(6460), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(417), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(435), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(433), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(429), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4145), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(425), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(6700), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [112697] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3451), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3449), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112743] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    ACTIONS(4131), 1,
      sym_assign_operator,
    ACTIONS(4133), 1,
      anon_sym_SEMI,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2906), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [112823] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3411), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3409), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112869] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(743), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(741), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112915] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4153), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1702), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [112961] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2732), 1,
      anon_sym_and,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    ACTIONS(4090), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4092), 1,
      sym_concat_operator,
    ACTIONS(4094), 1,
      sym_rel_operator,
    ACTIONS(4114), 1,
      sym_assign_operator,
    ACTIONS(4143), 1,
      anon_sym_SEMI,
    STATE(661), 1,
      sym_pow_operator,
    STATE(662), 1,
      sym_mult_operator,
    STATE(663), 1,
      sym_add_operator,
    STATE(664), 1,
      sym_and_operator,
    STATE(665), 1,
      sym_or_operator,
    STATE(2928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2730), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [113043] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(431), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(437), 1,
      sym_rel_operator,
    ACTIONS(439), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1162), 1,
      anon_sym_DOT,
    ACTIONS(4147), 1,
      anon_sym_or,
    STATE(2929), 1,
      sym_attribute,
    STATE(6991), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(417), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(435), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(433), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(429), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4145), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(1168), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(6700), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [113111] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2732), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2730), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113157] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(2931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1700), 27,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [113205] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3439), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3437), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113251] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2933), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3471), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3469), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113297] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [113353] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4108), 1,
      sym__lowercase_identifier,
    ACTIONS(4141), 1,
      anon_sym_POUND,
    STATE(2935), 1,
      sym_attribute,
    STATE(3077), 1,
      sym_type_constructor_path,
    STATE(5948), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3642), 10,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(3646), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [113409] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4155), 1,
      anon_sym_DOT_DOT,
    STATE(2936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1318), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1316), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [113457] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3514), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3512), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113503] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3367), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3365), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113549] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [113607] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3375), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3373), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113653] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [113711] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 6,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [113773] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(431), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(437), 1,
      sym_rel_operator,
    ACTIONS(439), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4147), 1,
      anon_sym_or,
    ACTIONS(4157), 1,
      anon_sym_DOT,
    STATE(2943), 1,
      sym_attribute,
    STATE(6602), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(417), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(435), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(433), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(429), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4145), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(4159), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(6700), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [113841] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4165), 1,
      anon_sym_DOT,
    STATE(2944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4163), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4161), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [113889] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3355), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3353), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113935] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4169), 1,
      anon_sym_let,
    STATE(2946), 1,
      sym_attribute,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3656), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4167), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [113997] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4169), 1,
      anon_sym_let,
    STATE(2947), 1,
      sym_attribute,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3657), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4167), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114059] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4171), 1,
      anon_sym_COLON,
    STATE(2948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2984), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2986), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114107] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2970), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2972), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114153] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(2950), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1700), 27,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114201] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1618), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1620), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114247] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1622), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1624), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114293] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1634), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1636), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114339] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1650), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1652), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114385] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3455), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3453), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114431] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3510), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3508), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114477] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1826), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1828), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114523] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1842), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1844), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114569] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1850), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1852), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114615] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3518), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3516), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114661] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3359), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3357), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114707] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3343), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3341), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114753] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114799] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4175), 1,
      anon_sym_let,
    STATE(2964), 1,
      sym_attribute,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3602), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4173), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114861] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4175), 1,
      anon_sym_let,
    STATE(2965), 1,
      sym_attribute,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3603), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4173), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114923] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4179), 1,
      anon_sym_let,
    STATE(2966), 1,
      sym_attribute,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3604), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4177), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [114985] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3459), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3457), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115031] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3463), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3461), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115077] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4181), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1744), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [115123] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2974), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2976), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [115169] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    ACTIONS(4131), 1,
      sym_assign_operator,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(2971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2890), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [115247] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3154), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3156), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [115292] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      sym_concat_operator,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(2973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_COLON2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 7,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_COLON_GT,
      anon_sym_else,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [115359] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      sym_concat_operator,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(2974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_COLON2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 7,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_COLON_GT,
      anon_sym_else,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [115426] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_COLON2,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      sym_concat_operator,
    ACTIONS(4187), 1,
      anon_sym_COMMA,
    ACTIONS(4189), 1,
      sym_rel_operator,
    ACTIONS(4191), 1,
      sym_assign_operator,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(2975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2880), 3,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
      anon_sym_else,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [115505] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(2976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_COLON2,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [115562] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(2977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_COLON2,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [115619] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(2978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 7,
      anon_sym_COLON2,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 9,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_else,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [115680] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      sym_concat_operator,
    ACTIONS(4189), 1,
      sym_rel_operator,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(2979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2882), 2,
      anon_sym_COLON2,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 6,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_COLON_GT,
      anon_sym_else,
      anon_sym_or,
      sym_assign_operator,
  [115751] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_COLON2,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      sym_concat_operator,
    ACTIONS(4189), 1,
      sym_rel_operator,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(2980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_COLON_GT,
      anon_sym_else,
      sym_assign_operator,
  [115826] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2886), 1,
      anon_sym_COLON2,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      sym_concat_operator,
    ACTIONS(4187), 1,
      anon_sym_COMMA,
    ACTIONS(4189), 1,
      sym_rel_operator,
    ACTIONS(4191), 1,
      sym_assign_operator,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(2981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2884), 3,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
      anon_sym_else,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [115905] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_COLON2,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      sym_concat_operator,
    ACTIONS(4187), 1,
      anon_sym_COMMA,
    ACTIONS(4189), 1,
      sym_rel_operator,
    ACTIONS(4191), 1,
      sym_assign_operator,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(2982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2844), 3,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
      anon_sym_else,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [115984] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2892), 1,
      anon_sym_COLON2,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      sym_concat_operator,
    ACTIONS(4187), 1,
      anon_sym_COMMA,
    ACTIONS(4189), 1,
      sym_rel_operator,
    ACTIONS(4191), 1,
      sym_assign_operator,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(2983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2890), 3,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
      anon_sym_else,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [116063] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3760), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3762), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [116108] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_PIPE,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(2985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_assign_operator,
  [116183] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3098), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3100), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116228] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2886), 1,
      anon_sym_PIPE,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    ACTIONS(4201), 1,
      sym_assign_operator,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(2987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2884), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [116307] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1904), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(1902), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [116352] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_PIPE,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    ACTIONS(4201), 1,
      sym_assign_operator,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(2989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2844), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [116431] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1908), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(1906), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [116476] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4205), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4203), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [116521] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4209), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4207), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [116566] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1950), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1952), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116611] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4211), 1,
      anon_sym_COMMA,
    ACTIONS(4213), 1,
      anon_sym_PIPE,
    ACTIONS(4215), 1,
      anon_sym_as,
    ACTIONS(4217), 1,
      anon_sym_COLON_COLON,
    STATE(2994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3134), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3136), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116664] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3138), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3140), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116709] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3881), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3883), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [116754] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3698), 1,
      anon_sym_COLON2,
    STATE(2997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3106), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [116801] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3895), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3897), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [116846] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4211), 1,
      anon_sym_COMMA,
    ACTIONS(4217), 1,
      anon_sym_COLON_COLON,
    STATE(2999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3138), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3140), 18,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116895] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4211), 1,
      anon_sym_COMMA,
    STATE(3000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3138), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3140), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116942] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3138), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3140), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116987] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4211), 1,
      anon_sym_COMMA,
    ACTIONS(4213), 1,
      anon_sym_PIPE,
    ACTIONS(4217), 1,
      anon_sym_COLON_COLON,
    STATE(3002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3142), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3144), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117038] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1614), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1616), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117083] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3142), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3144), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117128] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3146), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3148), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117173] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3004), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3006), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117218] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4221), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4219), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [117263] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1924), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(1922), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117308] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    ACTIONS(4131), 1,
      sym_assign_operator,
    ACTIONS(4223), 1,
      anon_sym_SEMI,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(3009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2734), 3,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [117387] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3024), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3026), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117432] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3923), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3925), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117477] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2904), 1,
      anon_sym_PIPE,
    ACTIONS(4225), 1,
      anon_sym_COMMA,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    ACTIONS(4231), 1,
      sym_rel_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2902), 4,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_assign_operator,
  [117554] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2736), 1,
      anon_sym_PIPE,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    ACTIONS(4201), 1,
      sym_assign_operator,
    ACTIONS(4233), 1,
      anon_sym_SEMI,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3013), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2734), 2,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [117635] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3028), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3030), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117680] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2912), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [117745] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [117812] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3016), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3018), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117857] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3158), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3160), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117902] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3194), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3196), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117947] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4217), 1,
      anon_sym_COLON_COLON,
    STATE(3020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3078), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3080), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117994] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [118061] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_PIPE,
    ACTIONS(4225), 1,
      anon_sym_COMMA,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    ACTIONS(4231), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2880), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [118140] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    ACTIONS(4231), 1,
      sym_rel_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3023), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2882), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      anon_sym_or,
      sym_assign_operator,
  [118211] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4239), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4237), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [118256] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1926), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1928), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118301] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1830), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1832), 27,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [118346] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3210), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3212), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118391] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4243), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4241), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [118436] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_PIPE,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    ACTIONS(4231), 1,
      sym_rel_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_assign_operator,
  [118511] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1318), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1316), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118556] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3012), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3014), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118601] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1930), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1932), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118646] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2886), 1,
      anon_sym_PIPE,
    ACTIONS(4225), 1,
      anon_sym_COMMA,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    ACTIONS(4231), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2884), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [118725] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_PIPE,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4225), 1,
      anon_sym_COMMA,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    ACTIONS(4231), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2844), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [118804] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3020), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3022), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118849] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1934), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1936), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118894] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1938), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1940), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118939] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3036), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3038), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118984] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2908), 1,
      anon_sym_PIPE,
    ACTIONS(4225), 1,
      anon_sym_COMMA,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    ACTIONS(4231), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    ACTIONS(4245), 1,
      anon_sym_SEMI,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2906), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [119065] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3040), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3794), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3796), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [119110] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3106), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [119155] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4247), 1,
      anon_sym_COLON2,
    STATE(3042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3106), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [119202] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(9), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(13), 1,
      anon_sym_type,
    ACTIONS(17), 1,
      anon_sym_module,
    ACTIONS(19), 1,
      anon_sym_open,
    ACTIONS(21), 1,
      anon_sym_include,
    ACTIONS(23), 1,
      anon_sym_class,
    ACTIONS(25), 1,
      anon_sym_val,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4249), 1,
      anon_sym_external,
    ACTIONS(4251), 1,
      anon_sym_exception,
    ACTIONS(4253), 1,
      anon_sym_end,
    STATE(3043), 1,
      sym_attribute,
    STATE(3054), 1,
      aux_sym__signature_repeat1,
    STATE(3159), 1,
      aux_sym__structure_repeat1,
    STATE(4229), 1,
      sym__signature_item,
    STATE(6543), 1,
      sym__signature,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [119279] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2732), 1,
      anon_sym_PIPE,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    ACTIONS(4201), 1,
      sym_assign_operator,
    ACTIONS(4233), 1,
      anon_sym_SEMI,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2730), 2,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [119360] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3040), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3042), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119405] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3046), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3214), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3216), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119450] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2892), 1,
      anon_sym_PIPE,
    ACTIONS(4225), 1,
      anon_sym_COMMA,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    ACTIONS(4231), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2890), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [119529] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    ACTIONS(4131), 1,
      sym_assign_operator,
    ACTIONS(4223), 1,
      anon_sym_SEMI,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(3048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2730), 3,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [119608] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2736), 1,
      anon_sym_PIPE,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4225), 1,
      anon_sym_COMMA,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    ACTIONS(4231), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    ACTIONS(4245), 1,
      anon_sym_SEMI,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2734), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [119689] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2732), 1,
      anon_sym_PIPE,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4225), 1,
      anon_sym_COMMA,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4229), 1,
      sym_concat_operator,
    ACTIONS(4231), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    ACTIONS(4245), 1,
      anon_sym_SEMI,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2730), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [119770] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1646), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1648), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119815] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4257), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4255), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [119860] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4217), 1,
      anon_sym_COLON_COLON,
    STATE(3053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3150), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3152), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119907] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(9), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(13), 1,
      anon_sym_type,
    ACTIONS(17), 1,
      anon_sym_module,
    ACTIONS(19), 1,
      anon_sym_open,
    ACTIONS(21), 1,
      anon_sym_include,
    ACTIONS(23), 1,
      anon_sym_class,
    ACTIONS(25), 1,
      anon_sym_val,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4249), 1,
      anon_sym_external,
    ACTIONS(4251), 1,
      anon_sym_exception,
    STATE(3054), 1,
      sym_attribute,
    STATE(3055), 1,
      aux_sym__signature_repeat1,
    STATE(3160), 1,
      aux_sym__structure_repeat1,
    STATE(4229), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4259), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [119982] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4261), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4264), 1,
      anon_sym_external,
    ACTIONS(4267), 1,
      anon_sym_type,
    ACTIONS(4272), 1,
      anon_sym_exception,
    ACTIONS(4275), 1,
      anon_sym_module,
    ACTIONS(4278), 1,
      anon_sym_open,
    ACTIONS(4281), 1,
      anon_sym_include,
    ACTIONS(4284), 1,
      anon_sym_class,
    ACTIONS(4287), 1,
      anon_sym_val,
    ACTIONS(4290), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4293), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(4296), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    STATE(3278), 1,
      aux_sym__structure_repeat1,
    STATE(4229), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4270), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(3055), 2,
      sym_attribute,
      aux_sym__signature_repeat1,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [120055] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3032), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3034), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120100] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1902), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1904), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120145] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3058), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3062), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3064), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120190] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3059), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3082), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3084), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120235] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3086), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3088), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120280] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3162), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3164), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120325] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1922), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1924), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120370] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4301), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4299), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [120415] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [120482] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1906), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1908), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120527] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4305), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4303), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [120572] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3090), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3092), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120617] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3899), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3901), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [120662] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(9), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(13), 1,
      anon_sym_type,
    ACTIONS(17), 1,
      anon_sym_module,
    ACTIONS(19), 1,
      anon_sym_open,
    ACTIONS(21), 1,
      anon_sym_include,
    ACTIONS(23), 1,
      anon_sym_class,
    ACTIONS(25), 1,
      anon_sym_val,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4249), 1,
      anon_sym_external,
    ACTIONS(4251), 1,
      anon_sym_exception,
    ACTIONS(4307), 1,
      anon_sym_end,
    STATE(3054), 1,
      aux_sym__signature_repeat1,
    STATE(3069), 1,
      sym_attribute,
    STATE(3159), 1,
      aux_sym__structure_repeat1,
    STATE(4229), 1,
      sym__signature_item,
    STATE(7047), 1,
      sym__signature,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [120739] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3907), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3909), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [120784] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3911), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3913), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [120829] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3798), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3800), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [120874] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3931), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3933), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [120919] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3935), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3937), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [120964] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3865), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3867), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [121009] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1708), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1710), 27,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [121054] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3756), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3758), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [121099] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      sym_concat_operator,
    ACTIONS(4187), 1,
      anon_sym_COMMA,
    ACTIONS(4189), 1,
      sym_rel_operator,
    ACTIONS(4191), 1,
      sym_assign_operator,
    ACTIONS(4309), 1,
      anon_sym_RPAREN,
    ACTIONS(4311), 1,
      anon_sym_COLON2,
    ACTIONS(4313), 1,
      anon_sym_COLON_GT,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(3078), 1,
      sym_attribute,
    STATE(5988), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [121182] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3110), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3112), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121227] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3114), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3116), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121272] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [121327] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [121384] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [121441] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 7,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [121502] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3118), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3120), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121547] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3086), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3174), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3176), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121592] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3178), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3180), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121637] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3182), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3184), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121682] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3058), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3060), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121727] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3198), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3200), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121772] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2908), 1,
      anon_sym_PIPE,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    ACTIONS(4201), 1,
      sym_assign_operator,
    ACTIONS(4233), 1,
      anon_sym_SEMI,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2906), 2,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [121853] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3092), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3202), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3204), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121898] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3166), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3168), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121943] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3804), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3806), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [121988] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1642), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1644), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [122033] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3915), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3917), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122078] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4127), 1,
      sym_concat_operator,
    ACTIONS(4129), 1,
      sym_rel_operator,
    ACTIONS(4131), 1,
      sym_assign_operator,
    ACTIONS(4223), 1,
      anon_sym_SEMI,
    STATE(526), 1,
      sym_add_operator,
    STATE(544), 1,
      sym_pow_operator,
    STATE(545), 1,
      sym_mult_operator,
    STATE(546), 1,
      sym_and_operator,
    STATE(547), 1,
      sym_or_operator,
    STATE(3097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2906), 3,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [122157] = 24,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      sym_concat_operator,
    ACTIONS(4187), 1,
      anon_sym_COMMA,
    ACTIONS(4189), 1,
      sym_rel_operator,
    ACTIONS(4191), 1,
      sym_assign_operator,
    ACTIONS(4311), 1,
      anon_sym_COLON2,
    ACTIONS(4315), 1,
      anon_sym_RPAREN,
    ACTIONS(4317), 1,
      anon_sym_COLON_GT,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(3098), 1,
      sym_attribute,
    STATE(6239), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [122240] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2892), 1,
      anon_sym_PIPE,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    ACTIONS(4201), 1,
      sym_assign_operator,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3099), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2890), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [122319] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 1,
      anon_sym_PIPE,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    ACTIONS(4201), 1,
      sym_assign_operator,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2880), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [122398] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3943), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3945), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122443] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [122500] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3658), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3660), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122545] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1834), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1836), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [122590] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [122647] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 7,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [122708] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3947), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3949), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122753] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3919), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3921), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122798] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3939), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3941), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122843] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4321), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4319), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [122888] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4211), 1,
      anon_sym_COMMA,
    ACTIONS(4213), 1,
      anon_sym_PIPE,
    ACTIONS(4215), 1,
      anon_sym_as,
    ACTIONS(4217), 1,
      anon_sym_COLON_COLON,
    STATE(3111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3124), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3126), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [122941] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3903), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3905), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122986] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3977), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(3975), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [123031] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3814), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3816), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [123076] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3869), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3871), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [123121] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3885), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3887), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [123166] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1952), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(1950), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [123211] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2882), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_or,
      sym_assign_operator,
  [123282] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1750), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1752), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [123327] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3104), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3102), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [123372] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3106), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3108), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [123417] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2994), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2996), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [123462] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(3123), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 8,
      anon_sym_COLON2,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [123517] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4163), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4161), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [123562] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3873), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3875), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [123607] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4325), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4323), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [123652] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4329), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4327), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [123697] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2472), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2474), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3128), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4062), 2,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4060), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [123752] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3218), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3220), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [123797] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2904), 1,
      anon_sym_COLON2,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      sym_concat_operator,
    ACTIONS(4187), 1,
      anon_sym_COMMA,
    ACTIONS(4189), 1,
      sym_rel_operator,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(3130), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2902), 4,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
      anon_sym_else,
      sym_assign_operator,
  [123874] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3008), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3010), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [123919] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3132), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [123974] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    STATE(673), 1,
      sym_pow_operator,
    STATE(674), 1,
      sym_mult_operator,
    STATE(675), 1,
      sym_add_operator,
    STATE(676), 1,
      sym_and_operator,
    STATE(677), 1,
      sym_or_operator,
    STATE(3133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2912), 4,
      anon_sym_COLON2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_COLON_GT,
      anon_sym_else,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [124039] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3889), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3891), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [124084] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3927), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3929), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [124129] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2904), 1,
      anon_sym_PIPE,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2902), 4,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_assign_operator,
  [124206] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3044), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3046), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124251] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3138), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2912), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [124316] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4211), 1,
      anon_sym_COMMA,
    ACTIONS(4217), 1,
      anon_sym_COLON_COLON,
    STATE(3139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3044), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3046), 18,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124365] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [124432] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4211), 1,
      anon_sym_COMMA,
    STATE(3141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3044), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3046), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124479] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3877), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3879), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [124524] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3044), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3046), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124569] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4211), 1,
      anon_sym_COMMA,
    ACTIONS(4213), 1,
      anon_sym_PIPE,
    ACTIONS(4217), 1,
      anon_sym_COLON_COLON,
    STATE(3144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3054), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124620] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3054), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124665] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3102), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3104), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124710] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    STATE(2666), 1,
      sym_type_variable,
    STATE(3147), 1,
      sym_attribute,
    STATE(3237), 1,
      sym__simple_type,
    STATE(3549), 1,
      sym_record_declaration,
    STATE(5366), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6805), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [124790] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    ACTIONS(3993), 1,
      sym_rel_operator,
    ACTIONS(4335), 1,
      anon_sym_RBRACK,
    ACTIONS(4337), 1,
      anon_sym_SEMI,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(3148), 1,
      sym_attribute,
    STATE(5729), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [124870] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4341), 1,
      anon_sym_let,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3149), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3837), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4339), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [124930] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(3150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 10,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1700), 18,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124976] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4345), 1,
      anon_sym_let,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3151), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3858), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4343), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [125036] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4347), 1,
      anon_sym_else,
    STATE(2961), 1,
      sym_else_clause,
    STATE(3152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [125084] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(3153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 10,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1700), 18,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [125130] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4351), 1,
      anon_sym_let,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3154), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3733), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4349), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [125190] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4355), 1,
      anon_sym_let,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3155), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3743), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4353), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [125250] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(9), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(11), 1,
      anon_sym_external,
    ACTIONS(13), 1,
      anon_sym_type,
    ACTIONS(15), 1,
      anon_sym_exception,
    ACTIONS(17), 1,
      anon_sym_module,
    ACTIONS(19), 1,
      anon_sym_open,
    ACTIONS(21), 1,
      anon_sym_include,
    ACTIONS(23), 1,
      anon_sym_class,
    ACTIONS(25), 1,
      anon_sym_val,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4259), 1,
      ts_builtin_sym_end,
    STATE(3156), 1,
      sym_attribute,
    STATE(3179), 1,
      aux_sym__signature_repeat1,
    STATE(3194), 1,
      aux_sym__structure_repeat1,
    STATE(4229), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [125324] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2472), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2474), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4072), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4070), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [125378] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3158), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4017), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
    ACTIONS(3895), 11,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3897), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [125424] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(9), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(13), 1,
      anon_sym_type,
    ACTIONS(17), 1,
      anon_sym_module,
    ACTIONS(19), 1,
      anon_sym_open,
    ACTIONS(21), 1,
      anon_sym_include,
    ACTIONS(23), 1,
      anon_sym_class,
    ACTIONS(25), 1,
      anon_sym_val,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4249), 1,
      anon_sym_external,
    ACTIONS(4251), 1,
      anon_sym_exception,
    STATE(3159), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym__structure_repeat1,
    STATE(4215), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4259), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [125496] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(9), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(13), 1,
      anon_sym_type,
    ACTIONS(17), 1,
      anon_sym_module,
    ACTIONS(19), 1,
      anon_sym_open,
    ACTIONS(21), 1,
      anon_sym_include,
    ACTIONS(23), 1,
      anon_sym_class,
    ACTIONS(25), 1,
      anon_sym_val,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4249), 1,
      anon_sym_external,
    ACTIONS(4251), 1,
      anon_sym_exception,
    STATE(3160), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym__structure_repeat1,
    STATE(4215), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4357), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [125568] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4108), 1,
      sym__lowercase_identifier,
    ACTIONS(4141), 1,
      anon_sym_POUND,
    ACTIONS(4363), 1,
      anon_sym_DASH_GT,
    ACTIONS(4365), 1,
      anon_sym_STAR,
    STATE(3077), 1,
      sym_type_constructor_path,
    STATE(3161), 1,
      sym_attribute,
    STATE(3562), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(5948), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4359), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4361), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [125628] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    ACTIONS(4201), 1,
      sym_assign_operator,
    ACTIONS(4335), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(4367), 1,
      anon_sym_SEMI,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3162), 1,
      sym_attribute,
    STATE(5807), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [125708] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4369), 1,
      anon_sym_PIPE,
    STATE(3163), 1,
      sym_attribute,
    STATE(3177), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [125756] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2876), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2874), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [125810] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2902), 4,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_assign_operator,
  [125884] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2912), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2910), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [125948] = 23,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4379), 1,
      sym__lowercase_identifier,
    ACTIONS(4381), 1,
      anon_sym_LPAREN,
    STATE(3158), 1,
      sym_type_variable,
    STATE(3161), 1,
      sym__simple_type,
    STATE(3167), 1,
      sym_attribute,
    STATE(3549), 1,
      sym_record_declaration,
    STATE(5314), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6522), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [126028] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2882), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [126094] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    ACTIONS(4383), 1,
      sym_assign_operator,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2880), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [126170] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [126226] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3171), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2882), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [126282] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 6,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [126342] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2882), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3173), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      anon_sym_or,
      sym_assign_operator,
  [126412] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3174), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_assign_operator,
  [126484] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    ACTIONS(4383), 1,
      sym_assign_operator,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2884), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [126560] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4369), 1,
      anon_sym_PIPE,
    STATE(3176), 1,
      sym_attribute,
    STATE(3186), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [126608] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4369), 1,
      anon_sym_PIPE,
    STATE(3177), 1,
      sym_attribute,
    STATE(3187), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [126656] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    ACTIONS(4383), 1,
      sym_assign_operator,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2844), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [126732] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4261), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4267), 1,
      anon_sym_type,
    ACTIONS(4270), 1,
      ts_builtin_sym_end,
    ACTIONS(4275), 1,
      anon_sym_module,
    ACTIONS(4278), 1,
      anon_sym_open,
    ACTIONS(4281), 1,
      anon_sym_include,
    ACTIONS(4284), 1,
      anon_sym_class,
    ACTIONS(4287), 1,
      anon_sym_val,
    ACTIONS(4290), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4293), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(4296), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4385), 1,
      anon_sym_external,
    ACTIONS(4388), 1,
      anon_sym_exception,
    STATE(3255), 1,
      aux_sym__structure_repeat1,
    STATE(4229), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3179), 2,
      sym_attribute,
      aux_sym__signature_repeat1,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [126804] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    ACTIONS(4383), 1,
      sym_assign_operator,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2890), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [126880] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(4066), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4068), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4393), 1,
      anon_sym_let,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3181), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3787), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4391), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [126940] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    ACTIONS(4383), 1,
      sym_assign_operator,
    ACTIONS(4395), 1,
      anon_sym_SEMI,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3182), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2734), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [127018] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    ACTIONS(4383), 1,
      sym_assign_operator,
    ACTIONS(4395), 1,
      anon_sym_SEMI,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3183), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2730), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [127096] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    ACTIONS(4383), 1,
      sym_assign_operator,
    ACTIONS(4395), 1,
      anon_sym_SEMI,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2906), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [127174] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4347), 1,
      anon_sym_else,
    STATE(2906), 1,
      sym_else_clause,
    STATE(3185), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [127222] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4369), 1,
      anon_sym_PIPE,
    STATE(3186), 1,
      sym_attribute,
    STATE(3187), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3224), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3222), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [127270] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4397), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3187), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3264), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [127316] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4108), 1,
      sym__lowercase_identifier,
    ACTIONS(4141), 1,
      anon_sym_POUND,
    ACTIONS(4363), 1,
      anon_sym_DASH_GT,
    ACTIONS(4365), 1,
      anon_sym_STAR,
    STATE(3077), 1,
      sym_type_constructor_path,
    STATE(3188), 1,
      sym_attribute,
    STATE(3562), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(5948), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4400), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4402), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [127376] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2882), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [127442] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1854), 10,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1856), 18,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [127485] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(9), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(11), 1,
      anon_sym_external,
    ACTIONS(13), 1,
      anon_sym_type,
    ACTIONS(15), 1,
      anon_sym_exception,
    ACTIONS(17), 1,
      anon_sym_module,
    ACTIONS(19), 1,
      anon_sym_open,
    ACTIONS(21), 1,
      anon_sym_include,
    ACTIONS(23), 1,
      anon_sym_class,
    ACTIONS(25), 1,
      anon_sym_val,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4259), 1,
      ts_builtin_sym_end,
    STATE(3191), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym__structure_repeat1,
    STATE(4215), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [127556] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(4410), 1,
      anon_sym_DOT,
    STATE(3192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4406), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4404), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [127603] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(3193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1700), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [127648] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(9), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(11), 1,
      anon_sym_external,
    ACTIONS(13), 1,
      anon_sym_type,
    ACTIONS(15), 1,
      anon_sym_exception,
    ACTIONS(17), 1,
      anon_sym_module,
    ACTIONS(19), 1,
      anon_sym_open,
    ACTIONS(21), 1,
      anon_sym_include,
    ACTIONS(23), 1,
      anon_sym_class,
    ACTIONS(25), 1,
      anon_sym_val,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4357), 1,
      ts_builtin_sym_end,
    STATE(3194), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym__structure_repeat1,
    STATE(4215), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [127719] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_PIPE,
    STATE(3195), 1,
      sym_attribute,
    STATE(3234), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [127766] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4414), 1,
      sym__lowercase_identifier,
    ACTIONS(4416), 1,
      anon_sym_LPAREN,
    STATE(3196), 1,
      sym_attribute,
    STATE(4681), 1,
      sym__simple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(7158), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [127839] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4418), 1,
      sym__lowercase_identifier,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4422), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4424), 1,
      anon_sym_GT,
    STATE(3197), 1,
      sym_attribute,
    STATE(4759), 1,
      sym__simple_type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6333), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [127914] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4418), 1,
      sym__lowercase_identifier,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4426), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4428), 1,
      anon_sym_GT,
    STATE(3198), 1,
      sym_attribute,
    STATE(4759), 1,
      sym__simple_type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6333), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [127989] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1754), 10,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1756), 18,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [128032] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    ACTIONS(3993), 1,
      sym_rel_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(3200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(4430), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [128107] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4418), 1,
      sym__lowercase_identifier,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4432), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4434), 1,
      anon_sym_GT,
    STATE(3201), 1,
      sym_attribute,
    STATE(4759), 1,
      sym__simple_type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6333), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [128182] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3656), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3654), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [128225] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4436), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3203), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3264), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128270] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(3204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1700), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [128315] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4108), 1,
      sym__lowercase_identifier,
    ACTIONS(4141), 1,
      anon_sym_POUND,
    ACTIONS(4441), 1,
      anon_sym_STAR,
    STATE(3077), 1,
      sym_type_constructor_path,
    STATE(3205), 1,
      sym_attribute,
    STATE(3519), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(5948), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4363), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4439), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [128372] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(4363), 1,
      anon_sym_DASH_GT,
    ACTIONS(4365), 1,
      anon_sym_STAR,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(3206), 1,
      sym_attribute,
    STATE(3562), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4400), 7,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4402), 13,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [128431] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3207), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1830), 10,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1832), 18,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [128474] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4418), 1,
      sym__lowercase_identifier,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4443), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4445), 1,
      anon_sym_GT,
    STATE(3208), 1,
      sym_attribute,
    STATE(4759), 1,
      sym__simple_type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6333), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [128549] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4418), 1,
      sym__lowercase_identifier,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4447), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4449), 1,
      anon_sym_GT,
    STATE(3209), 1,
      sym_attribute,
    STATE(4658), 1,
      sym__simple_type,
    STATE(5581), 1,
      sym_method_type,
    STATE(6117), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [128624] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2908), 1,
      anon_sym_DASH_GT,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    ACTIONS(3993), 1,
      sym_rel_operator,
    ACTIONS(4451), 1,
      anon_sym_SEMI,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(3210), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [128701] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2732), 1,
      anon_sym_DASH_GT,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    ACTIONS(3993), 1,
      sym_rel_operator,
    ACTIONS(4451), 1,
      anon_sym_SEMI,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(3211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [128778] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4453), 1,
      anon_sym_else,
    STATE(2906), 1,
      sym_else_clause,
    STATE(3212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128825] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4193), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4195), 1,
      sym_concat_operator,
    ACTIONS(4197), 1,
      sym_rel_operator,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    ACTIONS(4201), 1,
      sym_assign_operator,
    STATE(603), 1,
      sym_pow_operator,
    STATE(604), 1,
      sym_mult_operator,
    STATE(605), 1,
      sym_add_operator,
    STATE(606), 1,
      sym_and_operator,
    STATE(607), 1,
      sym_or_operator,
    STATE(3213), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(4455), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [128900] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4418), 1,
      sym__lowercase_identifier,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4457), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4459), 1,
      anon_sym_GT,
    STATE(3214), 1,
      sym_attribute,
    STATE(4656), 1,
      sym__simple_type,
    STATE(5554), 1,
      sym_method_type,
    STATE(6117), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [128975] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    STATE(3215), 1,
      sym_attribute,
    STATE(3338), 1,
      sym__simple_type,
    STATE(3549), 1,
      sym_record_declaration,
    STATE(3586), 1,
      sym__constructor_argument,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129050] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    STATE(3216), 1,
      sym_attribute,
    STATE(3728), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4463), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    ACTIONS(4461), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [129105] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2734), 1,
      anon_sym_do,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    ACTIONS(4383), 1,
      sym_assign_operator,
    ACTIONS(4473), 1,
      anon_sym_SEMI,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [129182] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4453), 1,
      anon_sym_else,
    STATE(2961), 1,
      sym_else_clause,
    STATE(3218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [129229] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4475), 1,
      anon_sym_else,
    STATE(2906), 1,
      sym_else_clause,
    STATE(3219), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [129276] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4418), 1,
      sym__lowercase_identifier,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4477), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4479), 1,
      anon_sym_GT,
    STATE(3220), 1,
      sym_attribute,
    STATE(4759), 1,
      sym__simple_type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6333), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129351] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    ACTIONS(3993), 1,
      sym_rel_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(3221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(4481), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [129426] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1708), 10,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1710), 18,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [129469] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2736), 1,
      anon_sym_DASH_GT,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    ACTIONS(3993), 1,
      sym_rel_operator,
    ACTIONS(4451), 1,
      anon_sym_SEMI,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(3223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [129546] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4485), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4483), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [129589] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3708), 1,
      sym_rel_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(3225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(4487), 2,
      anon_sym_SEMI,
      anon_sym_GT_RBRACE,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [129664] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4418), 1,
      sym__lowercase_identifier,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4489), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4491), 1,
      anon_sym_GT,
    STATE(3226), 1,
      sym_attribute,
    STATE(4676), 1,
      sym__simple_type,
    STATE(5707), 1,
      sym_method_type,
    STATE(6117), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129739] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2730), 1,
      anon_sym_do,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    ACTIONS(4383), 1,
      sym_assign_operator,
    ACTIONS(4473), 1,
      anon_sym_SEMI,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [129816] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_PIPE,
    STATE(3203), 1,
      aux_sym__match_cases_repeat1,
    STATE(3228), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3224), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3222), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [129863] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4418), 1,
      sym__lowercase_identifier,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4493), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4495), 1,
      anon_sym_GT,
    STATE(3229), 1,
      sym_attribute,
    STATE(4759), 1,
      sym__simple_type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6333), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129938] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4076), 1,
      anon_sym_COMMA,
    STATE(3230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 18,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [129983] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3309), 1,
      anon_sym_LBRACE,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    STATE(3231), 1,
      sym_attribute,
    STATE(4351), 1,
      sym__simple_type,
    STATE(5147), 1,
      sym__constructor_argument,
    STATE(5205), 1,
      sym_record_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [130058] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_PIPE,
    STATE(3228), 1,
      aux_sym__match_cases_repeat1,
    STATE(3232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [130105] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4499), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4497), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [130148] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_PIPE,
    STATE(3203), 1,
      aux_sym__match_cases_repeat1,
    STATE(3234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [130195] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4503), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4501), 26,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [130238] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4483), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4507), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4505), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [130283] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(4363), 1,
      anon_sym_DASH_GT,
    ACTIONS(4365), 1,
      anon_sym_STAR,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(3237), 1,
      sym_attribute,
    STATE(3562), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4359), 7,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4361), 13,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [130342] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4513), 1,
      anon_sym_as,
    STATE(3238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4511), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4509), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [130387] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4414), 1,
      sym__lowercase_identifier,
    ACTIONS(4515), 1,
      anon_sym_LPAREN,
    STATE(3239), 1,
      sym_attribute,
    STATE(4796), 1,
      sym__simple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6440), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [130460] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    STATE(3206), 1,
      sym__simple_type,
    STATE(3240), 1,
      sym_attribute,
    STATE(3549), 1,
      sym_record_declaration,
    STATE(6236), 1,
      sym_extended_module_path,
    STATE(6534), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [130535] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4379), 1,
      sym__lowercase_identifier,
    ACTIONS(4381), 1,
      anon_sym_LPAREN,
    STATE(3188), 1,
      sym__simple_type,
    STATE(3241), 1,
      sym_attribute,
    STATE(3549), 1,
      sym_record_declaration,
    STATE(6128), 1,
      sym_extended_module_path,
    STATE(6666), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [130610] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4475), 1,
      anon_sym_else,
    STATE(2961), 1,
      sym_else_clause,
    STATE(3242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [130657] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4519), 1,
      anon_sym_let,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(4525), 1,
      anon_sym_COLON2,
    STATE(3243), 1,
      sym_attribute,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3510), 1,
      sym__module_typed,
    STATE(3588), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4521), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4517), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [130718] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2440), 1,
      anon_sym_QMARK,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4414), 1,
      sym__lowercase_identifier,
    ACTIONS(4416), 1,
      anon_sym_LPAREN,
    STATE(3244), 1,
      sym_attribute,
    STATE(4776), 1,
      sym__simple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(7158), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [130791] = 22,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2906), 1,
      anon_sym_do,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    ACTIONS(4373), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4375), 1,
      sym_concat_operator,
    ACTIONS(4377), 1,
      sym_rel_operator,
    ACTIONS(4383), 1,
      sym_assign_operator,
    ACTIONS(4473), 1,
      anon_sym_SEMI,
    STATE(719), 1,
      sym_pow_operator,
    STATE(720), 1,
      sym_mult_operator,
    STATE(721), 1,
      sym_add_operator,
    STATE(722), 1,
      sym_and_operator,
    STATE(723), 1,
      sym_or_operator,
    STATE(3245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [130868] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2858), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2866), 1,
      anon_sym_or,
    ACTIONS(2868), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3704), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3706), 1,
      sym_concat_operator,
    ACTIONS(3710), 1,
      sym_assign_operator,
    ACTIONS(3993), 1,
      sym_rel_operator,
    STATE(573), 1,
      sym_pow_operator,
    STATE(574), 1,
      sym_mult_operator,
    STATE(575), 1,
      sym_add_operator,
    STATE(576), 1,
      sym_and_operator,
    STATE(577), 1,
      sym_or_operator,
    STATE(3246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2850), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(4455), 2,
      anon_sym_RBRACK,
      anon_sym_SEMI,
    ACTIONS(2854), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2856), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2860), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [130943] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(4525), 1,
      anon_sym_COLON2,
    ACTIONS(4529), 1,
      anon_sym_let,
    STATE(3243), 1,
      aux_sym_module_binding_repeat1,
    STATE(3247), 1,
      sym_attribute,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3482), 1,
      sym__module_typed,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3577), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4531), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4527), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131004] = 21,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3244), 1,
      anon_sym_LBRACE,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4379), 1,
      sym__lowercase_identifier,
    ACTIONS(4381), 1,
      anon_sym_LPAREN,
    STATE(3205), 1,
      sym__simple_type,
    STATE(3248), 1,
      sym_attribute,
    STATE(3549), 1,
      sym_record_declaration,
    STATE(3586), 1,
      sym__constructor_argument,
    STATE(6128), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [131079] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4535), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4533), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131124] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(3250), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4537), 9,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4539), 13,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [131176] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    STATE(3251), 1,
      sym_attribute,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3310), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4543), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4541), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131226] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4547), 1,
      sym__lowercase_identifier,
    STATE(3252), 1,
      sym_attribute,
    STATE(4282), 1,
      sym__simple_type,
    STATE(4283), 1,
      sym__tuple_type_rhs,
    STATE(4631), 1,
      sym_labeled_tuple_element_type,
    STATE(6117), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [131298] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1540), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1538), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131340] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    STATE(3254), 1,
      sym_attribute,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3311), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4551), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4549), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131390] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(9), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(11), 1,
      anon_sym_external,
    ACTIONS(13), 1,
      anon_sym_type,
    ACTIONS(15), 1,
      anon_sym_exception,
    ACTIONS(17), 1,
      anon_sym_module,
    ACTIONS(19), 1,
      anon_sym_open,
    ACTIONS(21), 1,
      anon_sym_include,
    ACTIONS(23), 1,
      anon_sym_class,
    ACTIONS(25), 1,
      anon_sym_val,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    STATE(3255), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym__structure_repeat1,
    STATE(4215), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [131458] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    ACTIONS(4553), 1,
      sym__lowercase_identifier,
    STATE(2589), 1,
      sym__simple_type,
    STATE(2660), 1,
      sym__tuple_type_rhs,
    STATE(2794), 1,
      sym_labeled_tuple_element_type,
    STATE(3256), 1,
      sym_attribute,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [131530] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3108), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4557), 1,
      anon_sym_let,
    ACTIONS(4559), 1,
      anon_sym_EQ,
    ACTIONS(4561), 1,
      anon_sym_COLON_EQ,
    STATE(3257), 1,
      sym_attribute,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3484), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3571), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4555), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131590] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    ACTIONS(4553), 1,
      sym__lowercase_identifier,
    STATE(2589), 1,
      sym__simple_type,
    STATE(2655), 1,
      sym__tuple_type_rhs,
    STATE(2794), 1,
      sym_labeled_tuple_element_type,
    STATE(3258), 1,
      sym_attribute,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [131662] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1906), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1908), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131704] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4565), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4563), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131746] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4108), 1,
      sym__lowercase_identifier,
    ACTIONS(4141), 1,
      anon_sym_POUND,
    STATE(3077), 1,
      sym_type_constructor_path,
    STATE(3261), 1,
      sym_attribute,
    STATE(5948), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4537), 8,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4539), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [131798] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4569), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4567), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131840] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2472), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2474), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3263), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3743), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4353), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [131896] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4551), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3264), 1,
      sym_attribute,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3345), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4549), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [131946] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1754), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1756), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [131988] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4573), 1,
      anon_sym_else,
    STATE(2961), 1,
      sym_else_clause,
    STATE(3266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [132034] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4577), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4575), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132076] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4581), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4579), 25,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132118] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1708), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1710), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [132160] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4587), 1,
      anon_sym_LPAREN,
    STATE(3476), 1,
      sym_module_parameter,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3270), 2,
      sym_attribute,
      aux_sym_module_binding_repeat1,
    ACTIONS(4585), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4583), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132206] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4381), 1,
      anon_sym_LPAREN,
    ACTIONS(4590), 1,
      sym__lowercase_identifier,
    STATE(2655), 1,
      sym__tuple_type_rhs,
    STATE(2898), 1,
      sym__simple_type,
    STATE(3271), 1,
      sym_attribute,
    STATE(3346), 1,
      sym_labeled_tuple_element_type,
    STATE(6128), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [132278] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4100), 1,
      anon_sym_COMMA,
    STATE(3272), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 18,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [132322] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4501), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(1700), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132366] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    ACTIONS(4553), 1,
      sym__lowercase_identifier,
    STATE(3274), 1,
      sym_attribute,
    STATE(4216), 1,
      sym__simple_type,
    STATE(4255), 1,
      sym__tuple_type_rhs,
    STATE(4537), 1,
      sym_labeled_tuple_element_type,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [132438] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3275), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1854), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1856), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [132480] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    STATE(3276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 18,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [132524] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    ACTIONS(4553), 1,
      sym__lowercase_identifier,
    STATE(3277), 1,
      sym_attribute,
    STATE(4216), 1,
      sym__simple_type,
    STATE(4268), 1,
      sym__tuple_type_rhs,
    STATE(4537), 1,
      sym_labeled_tuple_element_type,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [132596] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(9), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(13), 1,
      anon_sym_type,
    ACTIONS(17), 1,
      anon_sym_module,
    ACTIONS(19), 1,
      anon_sym_open,
    ACTIONS(21), 1,
      anon_sym_include,
    ACTIONS(23), 1,
      anon_sym_class,
    ACTIONS(25), 1,
      anon_sym_val,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4249), 1,
      anon_sym_external,
    ACTIONS(4251), 1,
      anon_sym_exception,
    STATE(3278), 1,
      sym_attribute,
    STATE(4171), 1,
      aux_sym__structure_repeat1,
    STATE(4215), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4204), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [132664] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3279), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1950), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1952), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132706] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    ACTIONS(4553), 1,
      sym__lowercase_identifier,
    STATE(3280), 1,
      sym_attribute,
    STATE(4216), 1,
      sym__simple_type,
    STATE(4283), 1,
      sym__tuple_type_rhs,
    STATE(4537), 1,
      sym_labeled_tuple_element_type,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [132778] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3108), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4594), 1,
      anon_sym_let,
    ACTIONS(4596), 1,
      anon_sym_EQ,
    ACTIONS(4598), 1,
      anon_sym_COLON_EQ,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3281), 1,
      sym_attribute,
    STATE(3487), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3582), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4592), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132838] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2472), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2474), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3282), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3733), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4349), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [132894] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1922), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1924), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132936] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4573), 1,
      anon_sym_else,
    STATE(2906), 1,
      sym_else_clause,
    STATE(3284), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [132982] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    ACTIONS(4602), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3285), 1,
      sym_attribute,
    STATE(3303), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4600), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [133032] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1830), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1832), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [133074] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4381), 1,
      anon_sym_LPAREN,
    ACTIONS(4590), 1,
      sym__lowercase_identifier,
    STATE(2660), 1,
      sym__tuple_type_rhs,
    STATE(2898), 1,
      sym__simple_type,
    STATE(3287), 1,
      sym_attribute,
    STATE(3346), 1,
      sym_labeled_tuple_element_type,
    STATE(6128), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [133146] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4381), 1,
      anon_sym_LPAREN,
    ACTIONS(4590), 1,
      sym__lowercase_identifier,
    STATE(2668), 1,
      sym__tuple_type_rhs,
    STATE(2898), 1,
      sym__simple_type,
    STATE(3288), 1,
      sym_attribute,
    STATE(3346), 1,
      sym_labeled_tuple_element_type,
    STATE(6128), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [133218] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4543), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3289), 1,
      sym_attribute,
    STATE(3335), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4541), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [133268] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2472), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2474), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(3290), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3787), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4391), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [133324] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3291), 1,
      sym_attribute,
    STATE(3309), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4602), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4600), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [133374] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4547), 1,
      sym__lowercase_identifier,
    STATE(3292), 1,
      sym_attribute,
    STATE(4255), 1,
      sym__tuple_type_rhs,
    STATE(4282), 1,
      sym__simple_type,
    STATE(4631), 1,
      sym_labeled_tuple_element_type,
    STATE(6117), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [133446] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4547), 1,
      sym__lowercase_identifier,
    STATE(3293), 1,
      sym_attribute,
    STATE(4268), 1,
      sym__tuple_type_rhs,
    STATE(4282), 1,
      sym__simple_type,
    STATE(4631), 1,
      sym_labeled_tuple_element_type,
    STATE(6117), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [133518] = 20,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    ACTIONS(4553), 1,
      sym__lowercase_identifier,
    STATE(2589), 1,
      sym__simple_type,
    STATE(2668), 1,
      sym__tuple_type_rhs,
    STATE(2794), 1,
      sym_labeled_tuple_element_type,
    STATE(3294), 1,
      sym_attribute,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [133590] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1902), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1904), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [133632] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3106), 1,
      sym__lowercase_identifier,
    ACTIONS(4604), 1,
      anon_sym_COLON2,
    STATE(3296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(3939), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3941), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [133679] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1930), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1932), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [133720] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4606), 1,
      anon_sym_else,
    STATE(2172), 1,
      sym_else_clause,
    STATE(3298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [133765] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4608), 1,
      anon_sym_PIPE,
    STATE(3299), 1,
      sym_attribute,
    STATE(3301), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3224), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3222), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [133810] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4612), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4610), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [133851] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4614), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3301), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3264), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [133894] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4505), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [133935] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    ACTIONS(4619), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3303), 1,
      sym_attribute,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4617), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [133982] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4623), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4621), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134023] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4627), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(4625), 14,
      anon_sym_let,
      anon_sym_external,
      anon_sym_type,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      sym__lowercase_identifier,
  [134064] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3306), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4631), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4629), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134105] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4108), 1,
      sym__lowercase_identifier,
    ACTIONS(4141), 1,
      anon_sym_POUND,
    STATE(3077), 1,
      sym_type_constructor_path,
    STATE(3307), 1,
      sym_attribute,
    STATE(5948), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4633), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4635), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [134156] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(4637), 1,
      anon_sym_COLON2,
    STATE(3308), 1,
      sym_attribute,
    STATE(3343), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3482), 1,
      sym__module_typed,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3577), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4531), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4527), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [134213] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3309), 1,
      sym_attribute,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4619), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4617), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134260] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3310), 1,
      sym_attribute,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4641), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4639), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134307] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3311), 1,
      sym_attribute,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4645), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4643), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134354] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4647), 1,
      anon_sym_else,
    STATE(2172), 1,
      sym_else_clause,
    STATE(3312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 8,
      anon_sym_COLON2,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 16,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [134399] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1926), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1928), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [134440] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3106), 1,
      sym__lowercase_identifier,
    ACTIONS(4604), 1,
      anon_sym_COLON2,
    STATE(3314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(3939), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(3941), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [134487] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4651), 1,
      anon_sym_let,
    ACTIONS(4653), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3315), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
    ACTIONS(4649), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134532] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4108), 1,
      sym__lowercase_identifier,
    ACTIONS(4141), 1,
      anon_sym_POUND,
    STATE(3077), 1,
      sym_type_constructor_path,
    STATE(3316), 1,
      sym_attribute,
    STATE(5948), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4400), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4402), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [134583] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4418), 1,
      sym__lowercase_identifier,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    STATE(3317), 1,
      sym_attribute,
    STATE(4759), 1,
      sym__simple_type,
    STATE(6117), 1,
      sym_extended_module_path,
    STATE(6333), 1,
      sym_method_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [134652] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4656), 1,
      anon_sym_PIPE,
    STATE(3318), 1,
      sym_attribute,
    STATE(3327), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [134697] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4606), 1,
      anon_sym_else,
    STATE(2147), 1,
      sym_else_clause,
    STATE(3319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [134742] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4660), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4658), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134783] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4608), 1,
      anon_sym_PIPE,
    STATE(3321), 1,
      sym_attribute,
    STATE(3339), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [134828] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    STATE(3322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 18,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [134871] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4463), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    STATE(3323), 1,
      sym_attribute,
    STATE(3853), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    ACTIONS(4461), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [134924] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4535), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4533), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134965] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4664), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3325), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3264), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135008] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4669), 1,
      anon_sym_constraint,
    ACTIONS(4672), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3267), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3326), 2,
      sym_attribute,
      aux_sym_type_binding_repeat1,
    ACTIONS(4667), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [135053] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4656), 1,
      anon_sym_PIPE,
    STATE(3325), 1,
      aux_sym__match_cases_repeat1,
    STATE(3327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135098] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4676), 1,
      anon_sym_let,
    ACTIONS(4678), 1,
      anon_sym_EQ,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3328), 1,
      sym_attribute,
    STATE(3499), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3556), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4674), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135155] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4680), 1,
      anon_sym_else,
    STATE(2301), 1,
      sym_else_clause,
    STATE(3329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135200] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4682), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4672), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3330), 2,
      sym_attribute,
      aux_sym_type_binding_repeat1,
    ACTIONS(4667), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135245] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4687), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4685), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135286] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4691), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4689), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135327] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1934), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1936), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [135368] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4647), 1,
      anon_sym_else,
    STATE(2147), 1,
      sym_else_clause,
    STATE(3334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 8,
      anon_sym_COLON2,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 16,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135413] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    ACTIONS(4641), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4639), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [135460] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1750), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1752), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [135501] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4608), 1,
      anon_sym_PIPE,
    STATE(3299), 1,
      aux_sym__match_cases_repeat1,
    STATE(3337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135546] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(4365), 1,
      anon_sym_STAR,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(3338), 1,
      sym_attribute,
    STATE(3562), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4363), 7,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4439), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
  [135601] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4608), 1,
      anon_sym_PIPE,
    STATE(3301), 1,
      aux_sym__match_cases_repeat1,
    STATE(3339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135646] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4695), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4693), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135687] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4699), 1,
      anon_sym_let,
    ACTIONS(4701), 1,
      anon_sym_EQ,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3341), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3512), 1,
      aux_sym_type_binding_repeat1,
    STATE(3596), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4697), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135744] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4680), 1,
      anon_sym_else,
    STATE(2273), 1,
      sym_else_clause,
    STATE(3342), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135789] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(4637), 1,
      anon_sym_COLON2,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3343), 1,
      sym_attribute,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3510), 1,
      sym__module_typed,
    STATE(3588), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4521), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4517), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [135846] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1938), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1940), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [135887] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    ACTIONS(4645), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4643), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [135934] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4139), 1,
      anon_sym_STAR,
    STATE(3346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3742), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3740), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135977] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4656), 1,
      anon_sym_PIPE,
    STATE(3325), 1,
      aux_sym__match_cases_repeat1,
    STATE(3347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3224), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3222), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [136022] = 20,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4705), 1,
      anon_sym_let,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4709), 1,
      anon_sym_BANG,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4713), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3348), 1,
      sym_attribute,
    STATE(4296), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6511), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136093] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4656), 1,
      anon_sym_PIPE,
    STATE(3347), 1,
      aux_sym__match_cases_repeat1,
    STATE(3349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3281), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3279), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [136138] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4717), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4715), 24,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [136179] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4705), 1,
      anon_sym_let,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4713), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3260), 1,
      sym__class_expression,
    STATE(3351), 1,
      sym_attribute,
    STATE(6443), 1,
      sym_module_path,
    STATE(6511), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136247] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3352), 1,
      sym_attribute,
    STATE(5258), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136315] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4725), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3353), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4723), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [136361] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4187), 1,
      anon_sym_COMMA,
    STATE(3354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 8,
      anon_sym_COLON2,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 16,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [136403] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3355), 1,
      sym_attribute,
    STATE(3623), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136471] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4733), 1,
      anon_sym_let,
    STATE(3356), 1,
      sym_attribute,
    STATE(3468), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4731), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [136517] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4733), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3357), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4731), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [136563] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3358), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4737), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4735), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [136603] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(3359), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4400), 7,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4402), 13,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [136653] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3360), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4741), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4739), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [136693] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(3361), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4633), 7,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4635), 13,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [136743] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3362), 1,
      sym_attribute,
    STATE(3567), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136811] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3363), 1,
      sym_attribute,
    STATE(3624), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136879] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4761), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3364), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3506), 1,
      aux_sym_type_binding_repeat1,
    STATE(3651), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4759), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [136933] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4765), 1,
      anon_sym_let,
    ACTIONS(4767), 1,
      anon_sym_LBRACE,
    ACTIONS(4769), 1,
      anon_sym_DQUOTE,
    STATE(3365), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3553), 1,
      aux_sym_external_repeat1,
    STATE(3750), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4763), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [136987] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3366), 1,
      sym_attribute,
    STATE(5190), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137055] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3367), 1,
      sym_attribute,
    STATE(3614), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137123] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3368), 1,
      sym_attribute,
    STATE(3573), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137191] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3369), 1,
      sym_attribute,
    STATE(3574), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137259] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3370), 1,
      sym_attribute,
    STATE(5050), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137327] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3371), 1,
      sym_attribute,
    STATE(3683), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137395] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3372), 1,
      sym_attribute,
    STATE(5590), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137463] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4773), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3373), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4771), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [137509] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3374), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4777), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4775), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [137549] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3375), 1,
      sym_attribute,
    STATE(5069), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137617] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3376), 1,
      sym_attribute,
    STATE(3684), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137685] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3897), 1,
      sym__lowercase_identifier,
    STATE(3377), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3895), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(4779), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4781), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [137729] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3378), 1,
      sym_attribute,
    STATE(3686), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137797] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3268), 1,
      sym__class_expression,
    STATE(3379), 1,
      sym_attribute,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137865] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3380), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4785), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4783), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [137905] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3102), 1,
      sym__lowercase_identifier,
    STATE(3381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3104), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(3760), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(3762), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [137949] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3382), 1,
      sym_attribute,
    STATE(5796), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138017] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4149), 1,
      anon_sym_DASH_GT,
    STATE(3383), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3718), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3716), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [138059] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3384), 1,
      sym_attribute,
    STATE(5810), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138127] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3108), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    ACTIONS(4787), 1,
      anon_sym_EQ,
    ACTIONS(4789), 1,
      anon_sym_COLON_EQ,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3385), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3582), 1,
      aux_sym_expression_item_repeat1,
    STATE(3583), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4592), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [138183] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    STATE(3386), 1,
      sym_attribute,
    STATE(4998), 1,
      sym__simple_type,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [138249] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4793), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4791), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [138289] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3388), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4797), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4795), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [138329] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3389), 1,
      sym_attribute,
    STATE(3666), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138397] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3102), 1,
      sym__lowercase_identifier,
    STATE(3390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3104), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(3760), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3762), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [138441] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3260), 1,
      sym__class_expression,
    STATE(3391), 1,
      sym_attribute,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138509] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3392), 1,
      sym_attribute,
    STATE(5131), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138577] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3262), 1,
      sym__class_expression,
    STATE(3393), 1,
      sym_attribute,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138645] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4705), 1,
      anon_sym_let,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4713), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3268), 1,
      sym__class_expression,
    STATE(3394), 1,
      sym_attribute,
    STATE(6443), 1,
      sym_module_path,
    STATE(6511), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138713] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3262), 1,
      sym__class_expression,
    STATE(3395), 1,
      sym_attribute,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138781] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3396), 1,
      sym_attribute,
    STATE(3683), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138849] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3397), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4801), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4799), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [138889] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3264), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3262), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [138929] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3530), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3528), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [138969] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3108), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    ACTIONS(4803), 1,
      anon_sym_EQ,
    ACTIONS(4805), 1,
      anon_sym_COLON_EQ,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3400), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3571), 1,
      aux_sym_expression_item_repeat1,
    STATE(3572), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4555), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [139025] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3401), 1,
      sym_attribute,
    STATE(3700), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139093] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4807), 1,
      anon_sym_else,
    STATE(2172), 1,
      sym_else_clause,
    STATE(3402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3285), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3283), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [139137] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4705), 1,
      anon_sym_let,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4713), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3403), 1,
      sym_attribute,
    STATE(4300), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6511), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139205] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3404), 1,
      sym_attribute,
    STATE(5134), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139273] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4811), 1,
      anon_sym_let,
    STATE(3373), 1,
      aux_sym_expression_item_repeat1,
    STATE(3405), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4809), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [139319] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4705), 1,
      anon_sym_let,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4713), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3262), 1,
      sym__class_expression,
    STATE(3406), 1,
      sym_attribute,
    STATE(6443), 1,
      sym_module_path,
    STATE(6511), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139387] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3407), 1,
      sym_attribute,
    STATE(3660), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139455] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3408), 1,
      sym_attribute,
    STATE(3623), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139523] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3409), 1,
      sym_attribute,
    STATE(5138), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139591] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3410), 1,
      sym_attribute,
    STATE(5252), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139659] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3411), 1,
      sym_attribute,
    STATE(5149), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139727] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3412), 1,
      sym_attribute,
    STATE(3624), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139795] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3413), 1,
      sym_attribute,
    STATE(3573), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139863] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4705), 1,
      anon_sym_let,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4713), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3414), 1,
      sym_attribute,
    STATE(5645), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6511), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139931] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3415), 1,
      sym_attribute,
    STATE(3702), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139999] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3416), 1,
      sym_attribute,
    STATE(3660), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140067] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3417), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4815), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4813), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [140107] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1698), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1700), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [140147] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3419), 1,
      sym_attribute,
    STATE(5090), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140215] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    STATE(3361), 1,
      sym__simple_type,
    STATE(3420), 1,
      sym_attribute,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [140281] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3421), 1,
      sym_attribute,
    STATE(3703), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140349] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4379), 1,
      sym__lowercase_identifier,
    ACTIONS(4381), 1,
      anon_sym_LPAREN,
    STATE(3307), 1,
      sym__simple_type,
    STATE(3422), 1,
      sym_attribute,
    STATE(6128), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [140415] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3423), 1,
      sym_attribute,
    STATE(3705), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140483] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3424), 1,
      sym_attribute,
    STATE(3574), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140551] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3425), 1,
      sym_attribute,
    STATE(5234), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140619] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3426), 1,
      sym_attribute,
    STATE(5167), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140687] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3427), 1,
      sym_attribute,
    STATE(3714), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140755] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4773), 1,
      anon_sym_let,
    STATE(3353), 1,
      aux_sym_expression_item_repeat1,
    STATE(3428), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4771), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [140801] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3429), 1,
      sym_attribute,
    STATE(5169), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140869] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3430), 1,
      sym_attribute,
    STATE(5170), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140937] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3431), 1,
      sym_attribute,
    STATE(3716), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141005] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3432), 1,
      sym_attribute,
    STATE(3717), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141073] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4819), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4817), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [141113] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3434), 1,
      sym_attribute,
    STATE(3714), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141181] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3435), 1,
      sym_attribute,
    STATE(3716), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141249] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3436), 1,
      sym_attribute,
    STATE(5182), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141317] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3437), 1,
      sym_attribute,
    STATE(3684), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141385] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3438), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4823), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4821), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [141425] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3439), 1,
      sym_attribute,
    STATE(3726), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141493] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3440), 1,
      sym_attribute,
    STATE(3717), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141561] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3441), 1,
      sym_attribute,
    STATE(5239), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141629] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3442), 1,
      sym_attribute,
    STATE(3613), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141697] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3447), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3445), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [141737] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4827), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4825), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [141777] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4705), 1,
      anon_sym_let,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4713), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3445), 1,
      sym_attribute,
    STATE(5935), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6511), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141845] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    STATE(3250), 1,
      sym__simple_type,
    STATE(3446), 1,
      sym_attribute,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [141911] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3447), 1,
      sym_attribute,
    STATE(3700), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141979] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3448), 1,
      sym_attribute,
    STATE(3567), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142047] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4225), 1,
      anon_sym_COMMA,
    STATE(3449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 16,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [142089] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3450), 1,
      sym_attribute,
    STATE(3702), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142157] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3451), 1,
      sym_attribute,
    STATE(3703), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142225] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3452), 1,
      sym_attribute,
    STATE(5128), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142293] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4831), 1,
      anon_sym_let,
    STATE(3357), 1,
      aux_sym_expression_item_repeat1,
    STATE(3453), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4829), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [142339] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    STATE(3454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 16,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [142381] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    STATE(3359), 1,
      sym__simple_type,
    STATE(3455), 1,
      sym_attribute,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [142447] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4379), 1,
      sym__lowercase_identifier,
    ACTIONS(4381), 1,
      anon_sym_LPAREN,
    STATE(2935), 1,
      sym__simple_type,
    STATE(3456), 1,
      sym_attribute,
    STATE(6128), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [142513] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4767), 1,
      anon_sym_LBRACE,
    ACTIONS(4769), 1,
      anon_sym_DQUOTE,
    ACTIONS(4835), 1,
      anon_sym_let,
    STATE(3457), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3553), 1,
      aux_sym_external_repeat1,
    STATE(3753), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4833), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [142567] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4839), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3458), 1,
      sym_attribute,
    STATE(3494), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3590), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4837), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [142621] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3534), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3532), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [142661] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4379), 1,
      sym__lowercase_identifier,
    ACTIONS(4381), 1,
      anon_sym_LPAREN,
    STATE(3261), 1,
      sym__simple_type,
    STATE(3460), 1,
      sym_attribute,
    STATE(6128), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [142727] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(2468), 1,
      anon_sym_LT,
    ACTIONS(2470), 1,
      anon_sym_POUND,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    STATE(2521), 1,
      sym__simple_type,
    STATE(3461), 1,
      sym_attribute,
    STATE(6236), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2629), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [142793] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3250), 1,
      anon_sym_LT,
    ACTIONS(3252), 1,
      anon_sym_POUND,
    ACTIONS(3254), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3256), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3260), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(4379), 1,
      sym__lowercase_identifier,
    ACTIONS(4381), 1,
      anon_sym_LPAREN,
    STATE(3316), 1,
      sym__simple_type,
    STATE(3462), 1,
      sym_attribute,
    STATE(6128), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2998), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [142859] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3463), 1,
      sym_attribute,
    STATE(5223), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142927] = 18,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3690), 1,
      anon_sym_LT,
    ACTIONS(3692), 1,
      anon_sym_POUND,
    ACTIONS(3694), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4420), 1,
      anon_sym_LPAREN,
    ACTIONS(4841), 1,
      sym__lowercase_identifier,
    STATE(3464), 1,
      sym_attribute,
    STATE(4261), 1,
      sym__simple_type,
    STATE(6117), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4490), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [142993] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3465), 1,
      sym_attribute,
    STATE(3686), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143061] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3466), 1,
      sym_attribute,
    STATE(3614), 1,
      sym__class_expression,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143129] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3467), 1,
      sym_attribute,
    STATE(5115), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143197] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4845), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3468), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4843), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143243] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3469), 1,
      sym_attribute,
    STATE(3705), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143311] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3470), 1,
      sym_attribute,
    STATE(3726), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143379] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4807), 1,
      anon_sym_else,
    STATE(2147), 1,
      sym_else_clause,
    STATE(3471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3315), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3313), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [143423] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4781), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4779), 23,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143463] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3473), 1,
      sym_attribute,
    STATE(3613), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143531] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3474), 1,
      sym_attribute,
    STATE(3666), 1,
      sym__class_expression,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143599] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(351), 1,
      anon_sym_object,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(407), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(409), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4719), 1,
      anon_sym_let,
    ACTIONS(4721), 1,
      anon_sym_fun,
    STATE(852), 1,
      sym__simple_class_expression,
    STATE(3475), 1,
      sym_attribute,
    STATE(5154), 1,
      sym__class_expression,
    STATE(6424), 1,
      sym_value_definition,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5679), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143667] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4849), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4847), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143707] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3897), 1,
      sym__lowercase_identifier,
    STATE(3477), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3895), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(4779), 8,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4781), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [143751] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4743), 1,
      sym__lowercase_identifier,
    ACTIONS(4745), 1,
      anon_sym_let,
    ACTIONS(4747), 1,
      anon_sym_LPAREN,
    ACTIONS(4749), 1,
      anon_sym_LBRACK,
    ACTIONS(4751), 1,
      anon_sym_object,
    ACTIONS(4753), 1,
      anon_sym_fun,
    ACTIONS(4755), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4757), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(766), 1,
      sym__simple_class_expression,
    STATE(3260), 1,
      sym__class_expression,
    STATE(3478), 1,
      sym_attribute,
    STATE(6556), 1,
      sym_module_path,
    STATE(7299), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2439), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143819] = 19,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1666), 1,
      anon_sym_object,
    ACTIONS(2738), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2740), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(4703), 1,
      sym__lowercase_identifier,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    ACTIONS(4711), 1,
      anon_sym_LBRACK,
    ACTIONS(4727), 1,
      anon_sym_let,
    ACTIONS(4729), 1,
      anon_sym_fun,
    STATE(795), 1,
      sym__simple_class_expression,
    STATE(3268), 1,
      sym__class_expression,
    STATE(3479), 1,
      sym_attribute,
    STATE(6443), 1,
      sym_module_path,
    STATE(6740), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3253), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2084), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143887] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4761), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3480), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3506), 1,
      aux_sym_type_binding_repeat1,
    STATE(3651), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4759), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143938] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    ACTIONS(4851), 1,
      anon_sym_EQ,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3481), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3596), 1,
      aux_sym_expression_item_repeat1,
    STATE(3597), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4697), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [143991] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4855), 1,
      anon_sym_let,
    ACTIONS(4857), 1,
      anon_sym_EQ,
    ACTIONS(4859), 1,
      anon_sym_COLON_EQ,
    STATE(3482), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3587), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4853), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144040] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4676), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3483), 1,
      sym_attribute,
    STATE(3499), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3556), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4674), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144091] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4863), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    STATE(3484), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3581), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4861), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144142] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4867), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3485), 1,
      sym_attribute,
    STATE(3490), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3693), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4865), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144193] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4871), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3486), 1,
      sym_attribute,
    STATE(3497), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3679), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4869), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144244] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4875), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    STATE(3487), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3559), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4873), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144295] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4879), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    STATE(3488), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3692), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4877), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144346] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4883), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3488), 1,
      aux_sym_type_binding_repeat1,
    STATE(3489), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3674), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4881), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144397] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4887), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    STATE(3490), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3708), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4885), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144448] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    STATE(3491), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2904), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2902), 16,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [144489] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4889), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4687), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3492), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(4685), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144530] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    ACTIONS(4892), 1,
      anon_sym_EQ,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3493), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3556), 1,
      aux_sym_expression_item_repeat1,
    STATE(3557), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4674), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [144583] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4896), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    STATE(3494), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3647), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4894), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144634] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3495), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4900), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4898), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144673] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4904), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    STATE(3496), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3676), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4902), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144724] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4908), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    STATE(3497), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3695), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4906), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144775] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4839), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3494), 1,
      aux_sym_type_binding_repeat1,
    STATE(3498), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3590), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4837), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144826] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4912), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    STATE(3499), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3592), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4910), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144877] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4699), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3500), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3512), 1,
      aux_sym_type_binding_repeat1,
    STATE(3596), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4697), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144928] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4916), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3496), 1,
      aux_sym_type_binding_repeat1,
    STATE(3501), 1,
      sym_attribute,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3649), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4914), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144979] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4920), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4918), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145018] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4926), 1,
      anon_sym_EQ,
    ACTIONS(4928), 1,
      anon_sym_of,
    ACTIONS(4930), 1,
      anon_sym_COLON2,
    STATE(3503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4924), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4922), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145063] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4936), 1,
      anon_sym_and,
    STATE(3492), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(3504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4934), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4932), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145106] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3505), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4940), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4938), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145145] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4944), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3506), 1,
      sym_attribute,
    STATE(3677), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4942), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145196] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1754), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1756), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145235] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4936), 1,
      anon_sym_and,
    STATE(3504), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(3508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4948), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4946), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145278] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4691), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4950), 1,
      anon_sym_as,
    STATE(3509), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4689), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [145319] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4954), 1,
      anon_sym_let,
    ACTIONS(4956), 1,
      anon_sym_EQ,
    ACTIONS(4958), 1,
      anon_sym_COLON_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3510), 1,
      sym_attribute,
    STATE(3564), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4952), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145368] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(4962), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3511), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4960), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [145409] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 1,
      anon_sym_constraint,
    ACTIONS(4966), 1,
      anon_sym_let,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3330), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3512), 1,
      sym_attribute,
    STATE(3653), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4964), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145460] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(4970), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4968), 22,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [145501] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4974), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4972), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145539] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3515), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4978), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4976), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145577] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4984), 1,
      anon_sym_DASH_GT,
    ACTIONS(4986), 1,
      anon_sym_with,
    STATE(3516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4982), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4980), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145619] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5002), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3517), 1,
      sym_attribute,
    STATE(3732), 1,
      sym__attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5706), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [145685] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4984), 1,
      anon_sym_DASH_GT,
    ACTIONS(4986), 1,
      anon_sym_with,
    STATE(3518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5006), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5004), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145727] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4441), 1,
      anon_sym_STAR,
    STATE(3519), 1,
      sym_attribute,
    STATE(3545), 1,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5010), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5008), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145769] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3520), 1,
      sym_attribute,
    STATE(3535), 1,
      aux_sym_expression_item_repeat1,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5014), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5012), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [145813] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4984), 1,
      anon_sym_DASH_GT,
    ACTIONS(4986), 1,
      anon_sym_with,
    STATE(3521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5018), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5016), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145855] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3522), 1,
      sym_attribute,
    STATE(3535), 1,
      aux_sym_expression_item_repeat1,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5022), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5020), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [145899] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5024), 1,
      anon_sym_RPAREN,
    ACTIONS(5026), 1,
      anon_sym__,
    ACTIONS(5028), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3523), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5624), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [145965] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3524), 1,
      sym_attribute,
    STATE(3535), 1,
      aux_sym_expression_item_repeat1,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5032), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5030), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146009] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3525), 1,
      sym_attribute,
    STATE(3544), 1,
      aux_sym_expression_item_repeat1,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5036), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5034), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146053] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3526), 1,
      sym_attribute,
    STATE(3528), 1,
      aux_sym_expression_item_repeat1,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5040), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5038), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146097] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3527), 1,
      sym_attribute,
    STATE(3530), 1,
      aux_sym_expression_item_repeat1,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5044), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5042), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146141] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3528), 1,
      sym_attribute,
    STATE(3535), 1,
      aux_sym_expression_item_repeat1,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5048), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5046), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146185] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3529), 1,
      sym_attribute,
    STATE(3535), 1,
      aux_sym_expression_item_repeat1,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5052), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5050), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146229] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3530), 1,
      sym_attribute,
    STATE(3535), 1,
      aux_sym_expression_item_repeat1,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5056), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5054), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146273] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5024), 1,
      anon_sym_RPAREN,
    ACTIONS(5026), 1,
      anon_sym__,
    ACTIONS(5028), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3531), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5709), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [146339] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5058), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3532), 1,
      sym_attribute,
    STATE(3776), 1,
      sym__attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5592), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [146405] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3533), 1,
      sym_attribute,
    STATE(3590), 1,
      aux_sym_expression_item_repeat1,
    STATE(3591), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4837), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [146455] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5060), 1,
      anon_sym_DASH_GT,
    ACTIONS(5062), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3534), 1,
      sym_attribute,
    STATE(3789), 1,
      sym__attribute,
    STATE(3885), 1,
      sym__module_type,
    STATE(5453), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [146521] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5066), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3535), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
    ACTIONS(4651), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(4649), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146563] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3536), 1,
      sym_attribute,
    STATE(3550), 1,
      aux_sym_expression_item_repeat1,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5071), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5069), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146607] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5073), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3537), 1,
      sym_attribute,
    STATE(3790), 1,
      sym__attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5607), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [146673] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3538), 1,
      sym_attribute,
    STATE(3546), 1,
      aux_sym_expression_item_repeat1,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5077), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5075), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146717] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3524), 1,
      aux_sym_expression_item_repeat1,
    STATE(3539), 1,
      sym_attribute,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5081), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5079), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146761] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1614), 2,
      anon_sym_let,
      anon_sym_COLON2,
    ACTIONS(1616), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [146799] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1646), 2,
      anon_sym_let,
      anon_sym_COLON2,
    ACTIONS(1648), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [146837] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4513), 1,
      anon_sym_as,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3542), 1,
      sym_attribute,
    STATE(3651), 1,
      aux_sym_expression_item_repeat1,
    STATE(3652), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4759), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [146887] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5085), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5083), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [146925] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3535), 1,
      aux_sym_expression_item_repeat1,
    STATE(3544), 1,
      sym_attribute,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5089), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5087), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146969] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5091), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4539), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3545), 2,
      sym_attribute,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(4537), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147009] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3535), 1,
      aux_sym_expression_item_repeat1,
    STATE(3546), 1,
      sym_attribute,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5096), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5094), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [147053] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4984), 1,
      anon_sym_DASH_GT,
    ACTIONS(4986), 1,
      anon_sym_with,
    STATE(3547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5100), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5098), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147095] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3548), 1,
      sym_attribute,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5104), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5102), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [147139] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4439), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4363), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147177] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3535), 1,
      aux_sym_expression_item_repeat1,
    STATE(3550), 1,
      sym_attribute,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5108), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5106), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [147221] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4984), 1,
      anon_sym_DASH_GT,
    ACTIONS(4986), 1,
      anon_sym_with,
    STATE(3551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5112), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5110), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147263] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1518), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3529), 1,
      aux_sym_expression_item_repeat1,
    STATE(3552), 1,
      sym_attribute,
    STATE(3816), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5116), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5114), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [147307] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5122), 1,
      anon_sym_LBRACE,
    ACTIONS(5125), 1,
      anon_sym_DQUOTE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5120), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3553), 2,
      sym_attribute,
      aux_sym_external_repeat1,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(5118), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147351] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5130), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3554), 1,
      sym_attribute,
    STATE(3600), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5128), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147394] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3555), 1,
      sym_attribute,
    STATE(3590), 1,
      aux_sym_expression_item_repeat1,
    STATE(3591), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4837), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [147441] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4912), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3556), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4910), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147484] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3557), 1,
      sym_attribute,
    STATE(3592), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4910), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [147531] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3558), 1,
      sym_attribute,
    STATE(3596), 1,
      aux_sym_expression_item_repeat1,
    STATE(3597), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4697), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [147578] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5134), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5132), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147621] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5138), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3560), 1,
      sym_attribute,
    STATE(3599), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5136), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147664] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5142), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5140), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147707] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4365), 1,
      anon_sym_STAR,
    ACTIONS(5010), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3562), 1,
      sym_attribute,
    STATE(3601), 1,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5008), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [147748] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5146), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3563), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5144), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147791] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5150), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5148), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147834] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4934), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5152), 1,
      anon_sym_and,
    STATE(3565), 1,
      sym_attribute,
    STATE(3607), 1,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4932), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [147875] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5156), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3566), 1,
      sym_attribute,
    STATE(3609), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5154), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147918] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5160), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3567), 1,
      sym_attribute,
    STATE(3612), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5158), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147961] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5164), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3568), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5162), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148004] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5168), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3569), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5166), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148047] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5172), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5170), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148090] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4863), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4861), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148133] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3572), 1,
      sym_attribute,
    STATE(3581), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4861), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [148180] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5176), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3573), 1,
      sym_attribute,
    STATE(3620), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5174), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148223] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5180), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3574), 1,
      sym_attribute,
    STATE(3622), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5178), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148266] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5184), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3575), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5182), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148309] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5186), 1,
      anon_sym_LPAREN,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5192), 1,
      anon_sym_end,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    STATE(3576), 1,
      sym_attribute,
    STATE(3891), 1,
      sym__attribute,
    STATE(3953), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [148372] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4519), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3577), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4517), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148415] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5200), 1,
      anon_sym_LBRACE,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3578), 1,
      sym_attribute,
    STATE(3753), 1,
      aux_sym_expression_item_repeat1,
    STATE(3883), 1,
      aux_sym_external_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4833), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [148464] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5200), 1,
      anon_sym_LBRACE,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3579), 1,
      sym_attribute,
    STATE(3750), 1,
      aux_sym_expression_item_repeat1,
    STATE(3883), 1,
      aux_sym_external_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4763), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [148513] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3556), 1,
      aux_sym_expression_item_repeat1,
    STATE(3557), 1,
      aux_sym_type_binding_repeat1,
    STATE(3580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4674), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [148560] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5204), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5202), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148603] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4875), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4873), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148646] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3559), 1,
      aux_sym_expression_item_repeat1,
    STATE(3583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4873), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [148693] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5206), 1,
      anon_sym_LPAREN,
    ACTIONS(5208), 1,
      anon_sym_end,
    STATE(3584), 1,
      sym_attribute,
    STATE(3859), 1,
      sym__attribute,
    STATE(3924), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [148756] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5212), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3561), 1,
      aux_sym_expression_item_repeat1,
    STATE(3585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5210), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148799] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4361), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4359), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148836] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5216), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3587), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5214), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148879] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5220), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3588), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5218), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148922] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5222), 1,
      anon_sym_LPAREN,
    ACTIONS(5224), 1,
      anon_sym_end,
    STATE(3589), 1,
      sym_attribute,
    STATE(3873), 1,
      sym__attribute,
    STATE(3944), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [148985] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4896), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4894), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149028] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3591), 1,
      sym_attribute,
    STATE(3647), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4894), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [149075] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5228), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5226), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149118] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5232), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5230), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149155] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3594), 1,
      sym_attribute,
    STATE(3649), 1,
      aux_sym_expression_item_repeat1,
    STATE(3650), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4914), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [149202] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3595), 1,
      sym_attribute,
    STATE(3651), 1,
      aux_sym_expression_item_repeat1,
    STATE(3652), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4759), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [149249] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4966), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4964), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149292] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3597), 1,
      sym_attribute,
    STATE(3653), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4964), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [149339] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5236), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3598), 1,
      sym_attribute,
    STATE(3655), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5234), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149382] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5240), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3599), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5238), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149425] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5244), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3600), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5242), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149468] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4539), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5246), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3601), 2,
      sym_attribute,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(4537), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [149507] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5251), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5249), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149550] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5251), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5249), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149593] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5255), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5253), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149636] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5257), 1,
      anon_sym_LPAREN,
    ACTIONS(5259), 1,
      anon_sym_end,
    STATE(3605), 1,
      sym_attribute,
    STATE(3849), 1,
      sym__attribute,
    STATE(3968), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [149699] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4948), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5152), 1,
      anon_sym_and,
    STATE(3565), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(3606), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4946), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [149740] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4687), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5261), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3607), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(4685), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [149779] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5266), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3608), 1,
      sym_attribute,
    STATE(3659), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5264), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149822] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5270), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5268), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149865] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2097), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2095), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149908] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5272), 1,
      anon_sym_LPAREN,
    ACTIONS(5274), 1,
      anon_sym_end,
    STATE(3611), 1,
      sym_attribute,
    STATE(3866), 1,
      sym__attribute,
    STATE(3987), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [149971] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5278), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3612), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5276), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150014] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5282), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3613), 1,
      sym_attribute,
    STATE(3664), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5280), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150057] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5286), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3614), 1,
      sym_attribute,
    STATE(3665), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5284), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150100] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5290), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5288), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150143] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5292), 1,
      anon_sym_LPAREN,
    ACTIONS(5294), 1,
      anon_sym_end,
    STATE(3616), 1,
      sym_attribute,
    STATE(3875), 1,
      sym__attribute,
    STATE(4010), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [150206] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2085), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2083), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150249] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5296), 1,
      anon_sym_LPAREN,
    ACTIONS(5298), 1,
      anon_sym_end,
    STATE(3618), 1,
      sym_attribute,
    STATE(3879), 1,
      sym__attribute,
    STATE(4026), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [150312] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3619), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5302), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5300), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150349] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5306), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5304), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150392] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5308), 1,
      anon_sym_LPAREN,
    ACTIONS(5310), 1,
      anon_sym_end,
    STATE(3621), 1,
      sym_attribute,
    STATE(3884), 1,
      sym__attribute,
    STATE(4042), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [150455] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5314), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3622), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5312), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150498] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5318), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3623), 1,
      sym_attribute,
    STATE(3669), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5316), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150541] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5322), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3570), 1,
      aux_sym_expression_item_repeat1,
    STATE(3624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5320), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150584] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5324), 1,
      anon_sym_LPAREN,
    ACTIONS(5326), 1,
      anon_sym_end,
    STATE(3625), 1,
      sym_attribute,
    STATE(3889), 1,
      sym__attribute,
    STATE(3895), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [150647] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5328), 1,
      anon_sym_LPAREN,
    ACTIONS(5330), 1,
      anon_sym_end,
    STATE(3626), 1,
      sym_attribute,
    STATE(3862), 1,
      sym__attribute,
    STATE(3905), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [150710] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5334), 1,
      anon_sym_let,
    ACTIONS(5336), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3627), 1,
      sym_attribute,
    STATE(3643), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5332), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150755] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3628), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5340), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5338), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150792] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5344), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5342), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150829] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5348), 1,
      anon_sym_let,
    ACTIONS(5350), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3568), 1,
      aux_sym_expression_item_repeat1,
    STATE(3630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5346), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150874] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5354), 1,
      anon_sym_let,
    ACTIONS(5356), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3575), 1,
      aux_sym_expression_item_repeat1,
    STATE(3631), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5352), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150919] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5362), 1,
      anon_sym_PIPE,
    STATE(3632), 1,
      sym_attribute,
    STATE(3636), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5360), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5358), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150960] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4984), 1,
      anon_sym_DASH_GT,
    ACTIONS(4986), 1,
      anon_sym_with,
    ACTIONS(5366), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3633), 1,
      sym_attribute,
    STATE(3763), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5364), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151007] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5370), 1,
      anon_sym_let,
    ACTIONS(5372), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3615), 1,
      aux_sym_expression_item_repeat1,
    STATE(3634), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5368), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151052] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5362), 1,
      anon_sym_PIPE,
    STATE(3635), 1,
      sym_attribute,
    STATE(3636), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5376), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5374), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151093] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5378), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5232), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3636), 2,
      sym_attribute,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5230), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151132] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4984), 1,
      anon_sym_DASH_GT,
    ACTIONS(4986), 1,
      anon_sym_with,
    ACTIONS(5383), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3637), 1,
      sym_attribute,
    STATE(3756), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5381), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151179] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5387), 1,
      anon_sym_let,
    ACTIONS(5389), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3638), 1,
      sym_attribute,
    STATE(3661), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5385), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151224] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5393), 1,
      anon_sym_let,
    ACTIONS(5395), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3639), 1,
      sym_attribute,
    STATE(3685), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5391), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151269] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5399), 1,
      anon_sym_let,
    ACTIONS(5401), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3640), 1,
      sym_attribute,
    STATE(3687), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5397), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151314] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5405), 1,
      anon_sym_let,
    ACTIONS(5407), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3641), 1,
      sym_attribute,
    STATE(3704), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5403), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151359] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5411), 1,
      anon_sym_let,
    ACTIONS(5413), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3642), 1,
      sym_attribute,
    STATE(3706), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5409), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151404] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5417), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5415), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151447] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5421), 1,
      anon_sym_let,
    ACTIONS(5423), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3644), 1,
      sym_attribute,
    STATE(3718), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5419), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151492] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5427), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5425), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151535] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3291), 2,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3289), 20,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [151572] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5431), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3647), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5429), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151615] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3648), 1,
      sym_attribute,
    STATE(3674), 1,
      aux_sym_expression_item_repeat1,
    STATE(3675), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4881), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [151662] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4904), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4902), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151705] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3650), 1,
      sym_attribute,
    STATE(3676), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4902), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [151752] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4944), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4942), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151795] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3652), 1,
      sym_attribute,
    STATE(3677), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4942), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [151842] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5435), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5433), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151885] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3654), 1,
      sym_attribute,
    STATE(3679), 1,
      aux_sym_expression_item_repeat1,
    STATE(3680), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4869), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [151932] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5439), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5437), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151975] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5443), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5441), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152018] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5443), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5441), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152061] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2075), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2073), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152104] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5447), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5445), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152147] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5451), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3660), 1,
      sym_attribute,
    STATE(3682), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5449), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152190] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5455), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5453), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152233] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5459), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5457), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152276] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2113), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152319] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5463), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5461), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152362] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5467), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5465), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152405] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5471), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3666), 1,
      sym_attribute,
    STATE(3690), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5469), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152448] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5475), 1,
      anon_sym_let,
    STATE(3667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5473), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152485] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5477), 1,
      anon_sym_LPAREN,
    ACTIONS(5479), 1,
      anon_sym_end,
    STATE(3668), 1,
      sym_attribute,
    STATE(3887), 1,
      sym__attribute,
    STATE(4053), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [152548] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5483), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3669), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5481), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152591] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5485), 1,
      aux_sym_attribute_id_token1,
    STATE(3670), 1,
      sym_attribute,
    STATE(3671), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2752), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2754), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [152632] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5485), 1,
      aux_sym_attribute_id_token1,
    STATE(3671), 1,
      sym_attribute,
    STATE(3672), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2742), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2744), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [152673] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5487), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3672), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(2756), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2758), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [152712] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4924), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5490), 1,
      anon_sym_EQ,
    ACTIONS(5492), 1,
      anon_sym_of,
    ACTIONS(5494), 1,
      anon_sym_COLON2,
    STATE(3673), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4922), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [152755] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4879), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3674), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4877), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152798] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3675), 1,
      sym_attribute,
    STATE(3692), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4877), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [152845] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5498), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5496), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152888] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5502), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5500), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152931] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3678), 1,
      sym_attribute,
    STATE(3693), 1,
      aux_sym_expression_item_repeat1,
    STATE(3694), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4865), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [152978] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4908), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4906), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153021] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3680), 1,
      sym_attribute,
    STATE(3695), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4906), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [153068] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5506), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3681), 1,
      sym_attribute,
    STATE(3696), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5504), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153111] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5510), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5508), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153154] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5514), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3683), 1,
      sym_attribute,
    STATE(3698), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5512), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153197] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5518), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3684), 1,
      sym_attribute,
    STATE(3699), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5516), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153240] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5522), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3685), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5520), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153283] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5526), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3686), 1,
      sym_attribute,
    STATE(3701), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5524), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153326] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5530), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3687), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5528), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153369] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5534), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3688), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5532), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153412] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2135), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3689), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153455] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5538), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3690), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5536), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153498] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5542), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5540), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153535] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5546), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5544), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153578] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4887), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4885), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153621] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 1,
      anon_sym_constraint,
    STATE(3267), 1,
      sym_type_constraint,
    STATE(3326), 1,
      aux_sym_type_binding_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3694), 1,
      sym_attribute,
    STATE(3708), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4885), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [153668] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5550), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5548), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153711] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5554), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5552), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153754] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5558), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3697), 1,
      sym_attribute,
    STATE(3709), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5556), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153797] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5562), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5560), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153840] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5566), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5564), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153883] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5570), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3700), 1,
      sym_attribute,
    STATE(3711), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5568), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153926] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5574), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5572), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153969] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5578), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3702), 1,
      sym_attribute,
    STATE(3712), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5576), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154012] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5582), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3703), 1,
      sym_attribute,
    STATE(3713), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5580), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154055] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5586), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5584), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154098] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5590), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3705), 1,
      sym_attribute,
    STATE(3715), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5588), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154141] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5594), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3706), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5592), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154184] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5598), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5596), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154227] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5602), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5600), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154270] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5606), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5604), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154313] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5610), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3710), 1,
      sym_attribute,
    STATE(3720), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5608), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154356] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5614), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5612), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154399] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5618), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5616), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154442] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5622), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5620), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154485] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5626), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3714), 1,
      sym_attribute,
    STATE(3721), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5624), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154528] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5630), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5628), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154571] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5634), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3716), 1,
      sym_attribute,
    STATE(3722), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5632), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154614] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5638), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3717), 1,
      sym_attribute,
    STATE(3723), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5636), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154657] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5642), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5640), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154700] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5646), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3719), 1,
      sym_attribute,
    STATE(3737), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5648), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5644), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154745] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5652), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5650), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154788] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5656), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5654), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154831] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5660), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5658), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154874] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5664), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5662), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154917] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5362), 1,
      anon_sym_PIPE,
    STATE(3632), 1,
      aux_sym_variant_declaration_repeat1,
    STATE(3724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4463), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4461), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154958] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5668), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3725), 1,
      sym_attribute,
    STATE(3766), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5670), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5666), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155003] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5674), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3726), 1,
      sym_attribute,
    STATE(3727), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5672), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155046] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5678), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5676), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155089] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5362), 1,
      anon_sym_PIPE,
    STATE(3635), 1,
      aux_sym_variant_declaration_repeat1,
    STATE(3728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5360), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5358), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155130] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5682), 1,
      anon_sym_let,
    STATE(3729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5680), 21,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155167] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5684), 1,
      anon_sym_DASH_GT,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5688), 1,
      anon_sym_functor,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3547), 1,
      sym__module_type,
    STATE(3730), 1,
      sym_attribute,
    STATE(5522), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [155227] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4900), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4898), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [155263] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5694), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3732), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5816), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [155323] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5698), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5696), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155365] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5702), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5700), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155407] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5060), 1,
      anon_sym_DASH_GT,
    ACTIONS(5062), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3735), 1,
      sym_attribute,
    STATE(3848), 1,
      sym__module_type,
    STATE(5453), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [155467] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3736), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5837), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [155527] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5706), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5704), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155569] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3738), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5920), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [155629] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5006), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5708), 1,
      anon_sym_DASH_GT,
    ACTIONS(5710), 1,
      anon_sym_with,
    STATE(3739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5004), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [155669] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3740), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5923), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [155729] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1826), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1828), 18,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
  [155765] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3941), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3939), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155801] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5714), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5712), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155843] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5060), 1,
      anon_sym_DASH_GT,
    ACTIONS(5062), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3744), 1,
      sym_attribute,
    STATE(4080), 1,
      sym__module_type,
    STATE(5453), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [155903] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5716), 1,
      anon_sym_LBRACE,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3745), 1,
      sym_attribute,
    STATE(3753), 1,
      aux_sym_expression_item_repeat1,
    STATE(3943), 1,
      aux_sym_external_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4833), 13,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [155951] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1850), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1852), 18,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
  [155987] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5720), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3747), 1,
      sym_attribute,
    STATE(3767), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5718), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156029] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5724), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3748), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5722), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156071] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5716), 1,
      anon_sym_LBRACE,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3749), 1,
      sym_attribute,
    STATE(3750), 1,
      aux_sym_expression_item_repeat1,
    STATE(3943), 1,
      aux_sym_external_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4763), 13,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [156119] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5728), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5726), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156161] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5060), 1,
      anon_sym_DASH_GT,
    ACTIONS(5062), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3751), 1,
      sym_attribute,
    STATE(4061), 1,
      sym__module_type,
    STATE(5453), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [156221] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5734), 1,
      anon_sym_and,
    ACTIONS(5736), 1,
      sym_let_and_operator,
    STATE(3752), 1,
      sym_attribute,
    STATE(3769), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5732), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5730), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156263] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5740), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5738), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156305] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5746), 1,
      anon_sym_DASH_GT,
    ACTIONS(5748), 1,
      anon_sym_module,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5752), 1,
      anon_sym_functor,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3754), 1,
      sym_attribute,
    STATE(4888), 1,
      sym__module_type,
    STATE(5480), 1,
      aux_sym_module_binding_repeat1,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [156365] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5758), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5756), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156401] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5762), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5760), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156443] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5764), 1,
      anon_sym_DASH_GT,
    ACTIONS(5766), 1,
      anon_sym_functor,
    STATE(3306), 1,
      sym__module_type,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3757), 1,
      sym_attribute,
    STATE(5455), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [156503] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1650), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1652), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156539] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3759), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1826), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1828), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156575] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1850), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1852), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156611] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5060), 1,
      anon_sym_DASH_GT,
    ACTIONS(5062), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3761), 1,
      sym_attribute,
    STATE(3880), 1,
      sym__module_type,
    STATE(5453), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [156671] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5060), 1,
      anon_sym_DASH_GT,
    ACTIONS(5062), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3762), 1,
      sym_attribute,
    STATE(4064), 1,
      sym__module_type,
    STATE(5453), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [156731] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5770), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3763), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5768), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156773] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5774), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3748), 1,
      aux_sym_expression_item_repeat1,
    STATE(3764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5772), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156815] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5734), 1,
      anon_sym_and,
    ACTIONS(5736), 1,
      sym_let_and_operator,
    STATE(3765), 1,
      sym_attribute,
    STATE(3781), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5732), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5730), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156857] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5778), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5776), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156899] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5782), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3767), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5780), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156941] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3768), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5708), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157001] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5734), 1,
      anon_sym_and,
    ACTIONS(5736), 1,
      sym_let_and_operator,
    STATE(3769), 1,
      sym_attribute,
    STATE(3781), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5786), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5784), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157043] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3770), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5626), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157103] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2756), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2758), 13,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
      aux_sym_attribute_id_token1,
  [157139] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3772), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4940), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4938), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [157175] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5734), 1,
      anon_sym_and,
    ACTIONS(5736), 1,
      sym_let_and_operator,
    STATE(3765), 1,
      aux_sym_value_definition_repeat1,
    STATE(3773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5790), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5788), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157217] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5794), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3774), 1,
      sym_attribute,
    STATE(3775), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5792), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157259] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5798), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3775), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5796), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157301] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5800), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3776), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5614), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157361] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3777), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5625), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157421] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3778), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5633), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157481] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    ACTIONS(5802), 1,
      anon_sym_DASH_GT,
    ACTIONS(5804), 1,
      anon_sym_module,
    ACTIONS(5806), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3779), 1,
      sym_attribute,
    STATE(4770), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157541] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5798), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3734), 1,
      aux_sym_expression_item_repeat1,
    STATE(3780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5796), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157583] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5812), 1,
      anon_sym_and,
    ACTIONS(5815), 1,
      sym_let_and_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5810), 2,
      anon_sym_let,
      anon_sym_in,
    STATE(3781), 2,
      sym_attribute,
      aux_sym_value_definition_repeat1,
    ACTIONS(5808), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157623] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3782), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3762), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3760), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157659] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1922), 4,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(1924), 17,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__uppercase_identifier,
  [157695] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5734), 1,
      anon_sym_and,
    ACTIONS(5736), 1,
      sym_let_and_operator,
    STATE(3784), 1,
      sym_attribute,
    STATE(3792), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5786), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5784), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157737] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3785), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5754), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157797] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    ACTIONS(5802), 1,
      anon_sym_DASH_GT,
    ACTIONS(5804), 1,
      anon_sym_module,
    ACTIONS(5806), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3786), 1,
      sym_attribute,
    STATE(4752), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157857] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5820), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5818), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157899] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1902), 4,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(1904), 17,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__uppercase_identifier,
  [157935] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5060), 1,
      anon_sym_DASH_GT,
    ACTIONS(5062), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3789), 1,
      sym_attribute,
    STATE(3857), 1,
      sym__module_type,
    STATE(5453), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157995] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5822), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3790), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5939), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158055] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5764), 1,
      anon_sym_DASH_GT,
    ACTIONS(5766), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3791), 1,
      sym_attribute,
    STATE(3802), 1,
      sym__module_type,
    STATE(5455), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158115] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5734), 1,
      anon_sym_and,
    ACTIONS(5736), 1,
      sym_let_and_operator,
    STATE(3781), 1,
      aux_sym_value_definition_repeat1,
    STATE(3792), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5826), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5824), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [158157] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3793), 1,
      sym_attribute,
    STATE(5337), 1,
      sym__module_type,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158217] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5764), 1,
      anon_sym_DASH_GT,
    ACTIONS(5766), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3794), 1,
      sym_attribute,
    STATE(3800), 1,
      sym__module_type,
    STATE(5455), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158277] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5764), 1,
      anon_sym_DASH_GT,
    ACTIONS(5766), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3795), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__module_type,
    STATE(5455), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158337] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5746), 1,
      anon_sym_DASH_GT,
    ACTIONS(5748), 1,
      anon_sym_module,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5752), 1,
      anon_sym_functor,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3796), 1,
      sym_attribute,
    STATE(4814), 1,
      sym__module_type,
    STATE(5480), 1,
      aux_sym_module_binding_repeat1,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158397] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5830), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5828), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [158433] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3798), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5652), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158493] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5684), 1,
      anon_sym_DASH_GT,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5688), 1,
      anon_sym_functor,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3633), 1,
      sym__module_type,
    STATE(3799), 1,
      sym_attribute,
    STATE(5522), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158553] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5100), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5708), 1,
      anon_sym_DASH_GT,
    ACTIONS(5710), 1,
      anon_sym_with,
    STATE(3800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5098), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [158593] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5764), 1,
      anon_sym_DASH_GT,
    ACTIONS(5766), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3801), 1,
      sym_attribute,
    STATE(3822), 1,
      sym__module_type,
    STATE(5455), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158653] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5018), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5708), 1,
      anon_sym_DASH_GT,
    ACTIONS(5710), 1,
      anon_sym_with,
    STATE(3802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5016), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [158693] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5112), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5708), 1,
      anon_sym_DASH_GT,
    ACTIONS(5710), 1,
      anon_sym_with,
    STATE(3803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5110), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [158733] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    ACTIONS(5802), 1,
      anon_sym_DASH_GT,
    ACTIONS(5804), 1,
      anon_sym_module,
    ACTIONS(5806), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3804), 1,
      sym_attribute,
    STATE(4700), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158793] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1650), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1652), 18,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
  [158829] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5684), 1,
      anon_sym_DASH_GT,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5688), 1,
      anon_sym_functor,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3637), 1,
      sym__module_type,
    STATE(3806), 1,
      sym_attribute,
    STATE(5522), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158889] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3807), 1,
      sym_attribute,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(5751), 1,
      sym__module_type,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158949] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5832), 1,
      anon_sym_DASH_GT,
    ACTIONS(5834), 1,
      anon_sym_module,
    ACTIONS(5836), 1,
      anon_sym_functor,
    STATE(3306), 1,
      sym__module_type,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3808), 1,
      sym_attribute,
    STATE(5489), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159009] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5746), 1,
      anon_sym_DASH_GT,
    ACTIONS(5748), 1,
      anon_sym_module,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5752), 1,
      anon_sym_functor,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3809), 1,
      sym_attribute,
    STATE(5470), 1,
      sym__module_type,
    STATE(5480), 1,
      aux_sym_module_binding_repeat1,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159069] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5746), 1,
      anon_sym_DASH_GT,
    ACTIONS(5748), 1,
      anon_sym_module,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5752), 1,
      anon_sym_functor,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3810), 1,
      sym_attribute,
    STATE(5471), 1,
      sym__module_type,
    STATE(5480), 1,
      aux_sym_module_binding_repeat1,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159129] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5746), 1,
      anon_sym_DASH_GT,
    ACTIONS(5748), 1,
      anon_sym_module,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5752), 1,
      anon_sym_functor,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3811), 1,
      sym_attribute,
    STATE(5476), 1,
      sym__module_type,
    STATE(5480), 1,
      aux_sym_module_binding_repeat1,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159189] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5746), 1,
      anon_sym_DASH_GT,
    ACTIONS(5748), 1,
      anon_sym_module,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5752), 1,
      anon_sym_functor,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3812), 1,
      sym_attribute,
    STATE(5480), 1,
      aux_sym_module_binding_repeat1,
    STATE(5519), 1,
      sym__module_type,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159249] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5832), 1,
      anon_sym_DASH_GT,
    ACTIONS(5834), 1,
      anon_sym_module,
    ACTIONS(5836), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3813), 1,
      sym_attribute,
    STATE(5180), 1,
      sym__module_type,
    STATE(5489), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159309] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5832), 1,
      anon_sym_DASH_GT,
    ACTIONS(5834), 1,
      anon_sym_module,
    ACTIONS(5836), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3814), 1,
      sym_attribute,
    STATE(5176), 1,
      sym__module_type,
    STATE(5489), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159369] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5832), 1,
      anon_sym_DASH_GT,
    ACTIONS(5834), 1,
      anon_sym_module,
    ACTIONS(5836), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3815), 1,
      sym_attribute,
    STATE(5177), 1,
      sym__module_type,
    STATE(5489), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159429] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4920), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4918), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [159465] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5832), 1,
      anon_sym_DASH_GT,
    ACTIONS(5834), 1,
      anon_sym_module,
    ACTIONS(5836), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3817), 1,
      sym_attribute,
    STATE(5178), 1,
      sym__module_type,
    STATE(5489), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159525] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3818), 1,
      sym_attribute,
    STATE(5537), 1,
      sym__module_type,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159585] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    ACTIONS(5802), 1,
      anon_sym_DASH_GT,
    ACTIONS(5804), 1,
      anon_sym_module,
    ACTIONS(5806), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3819), 1,
      sym_attribute,
    STATE(4800), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159645] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    ACTIONS(5802), 1,
      anon_sym_DASH_GT,
    ACTIONS(5804), 1,
      anon_sym_module,
    ACTIONS(5806), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3820), 1,
      sym_attribute,
    STATE(4717), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159705] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5742), 1,
      sym__lowercase_identifier,
    ACTIONS(5744), 1,
      anon_sym_LPAREN,
    ACTIONS(5750), 1,
      anon_sym_sig,
    ACTIONS(5754), 1,
      sym__uppercase_identifier,
    ACTIONS(5802), 1,
      anon_sym_DASH_GT,
    ACTIONS(5804), 1,
      anon_sym_module,
    ACTIONS(5806), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3821), 1,
      sym_attribute,
    STATE(4751), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6200), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4734), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159765] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4982), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5708), 1,
      anon_sym_DASH_GT,
    ACTIONS(5710), 1,
      anon_sym_with,
    STATE(3822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4980), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [159805] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5684), 1,
      anon_sym_DASH_GT,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5688), 1,
      anon_sym_functor,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3518), 1,
      sym__module_type,
    STATE(3823), 1,
      sym_attribute,
    STATE(5522), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159865] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1950), 4,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(1952), 17,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__uppercase_identifier,
  [159901] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5060), 1,
      anon_sym_DASH_GT,
    ACTIONS(5062), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3825), 1,
      sym_attribute,
    STATE(4070), 1,
      sym__module_type,
    STATE(5453), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [159961] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5684), 1,
      anon_sym_DASH_GT,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5688), 1,
      anon_sym_functor,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3521), 1,
      sym__module_type,
    STATE(3826), 1,
      sym_attribute,
    STATE(5522), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [160021] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1906), 4,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(1908), 17,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__uppercase_identifier,
  [160057] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5684), 1,
      anon_sym_DASH_GT,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5688), 1,
      anon_sym_functor,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3551), 1,
      sym__module_type,
    STATE(3828), 1,
      sym_attribute,
    STATE(5522), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [160117] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5684), 1,
      anon_sym_DASH_GT,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5688), 1,
      anon_sym_functor,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3516), 1,
      sym__module_type,
    STATE(3829), 1,
      sym_attribute,
    STATE(5522), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [160177] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5684), 1,
      anon_sym_DASH_GT,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5688), 1,
      anon_sym_functor,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(3306), 1,
      sym__module_type,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3830), 1,
      sym_attribute,
    STATE(5522), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [160237] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3831), 1,
      sym_attribute,
    STATE(5532), 1,
      sym__module_type,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [160297] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3832), 1,
      sym_attribute,
    STATE(5533), 1,
      sym__module_type,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [160357] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4992), 1,
      anon_sym_DASH_GT,
    ACTIONS(4994), 1,
      anon_sym_module,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(4998), 1,
      anon_sym_functor,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3833), 1,
      sym_attribute,
    STATE(5535), 1,
      sym__module_type,
    STATE(5540), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [160417] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4988), 1,
      sym__lowercase_identifier,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4996), 1,
      anon_sym_sig,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(5686), 1,
      anon_sym_module,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5764), 1,
      anon_sym_DASH_GT,
    ACTIONS(5766), 1,
      anon_sym_functor,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(3739), 1,
      sym__module_type,
    STATE(3834), 1,
      sym_attribute,
    STATE(5455), 1,
      aux_sym_module_binding_repeat1,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3320), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [160477] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3737), 1,
      aux_sym_expression_item_repeat1,
    STATE(3835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5838), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5644), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [160518] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1650), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1652), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [160553] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5842), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5840), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160594] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1826), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1828), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [160629] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1850), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1852), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [160664] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3975), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3977), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [160699] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5844), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3615), 1,
      aux_sym_expression_item_repeat1,
    STATE(3841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5368), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [160740] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3766), 1,
      aux_sym_expression_item_repeat1,
    STATE(3842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5846), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5666), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [160781] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5810), 1,
      anon_sym_let,
    ACTIONS(5848), 1,
      anon_sym_and,
    ACTIONS(5851), 1,
      sym_let_and_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3843), 2,
      sym_attribute,
      aux_sym_value_definition_repeat1,
    ACTIONS(5808), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160820] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5856), 1,
      anon_sym_BANG,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2944), 1,
      sym_module_path,
    STATE(3154), 1,
      sym__module_expression,
    STATE(3844), 1,
      sym_attribute,
    STATE(4101), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [160877] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5862), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3575), 1,
      aux_sym_expression_item_repeat1,
    STATE(3845), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5352), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [160918] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5866), 1,
      anon_sym_let,
    ACTIONS(5868), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3846), 2,
      sym_attribute,
      aux_sym_class_type_definition_repeat1,
    ACTIONS(5864), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160955] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5871), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3685), 1,
      aux_sym_expression_item_repeat1,
    STATE(3847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5391), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [160996] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5710), 1,
      anon_sym_with,
    ACTIONS(5873), 1,
      anon_sym_DASH_GT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3763), 1,
      aux_sym_expression_item_repeat1,
    STATE(3848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5364), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [161039] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5875), 1,
      anon_sym_LPAREN,
    ACTIONS(5877), 1,
      anon_sym_end,
    STATE(3849), 1,
      sym_attribute,
    STATE(3970), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [161096] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5881), 1,
      anon_sym_BANG,
    ACTIONS(5883), 1,
      anon_sym_functor,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    STATE(3850), 1,
      sym_attribute,
    STATE(4134), 1,
      sym__attribute,
    STATE(4302), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [161153] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5889), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3687), 1,
      aux_sym_expression_item_repeat1,
    STATE(3851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5397), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [161194] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5790), 1,
      anon_sym_let,
    ACTIONS(5891), 1,
      anon_sym_and,
    ACTIONS(5893), 1,
      sym_let_and_operator,
    STATE(3852), 1,
      sym_attribute,
    STATE(3856), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5788), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161235] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5360), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5895), 1,
      anon_sym_PIPE,
    STATE(3853), 1,
      sym_attribute,
    STATE(3882), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5358), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [161274] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5899), 1,
      anon_sym_let,
    ACTIONS(5901), 1,
      anon_sym_and,
    STATE(3846), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(3854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5897), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161313] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5903), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3661), 1,
      aux_sym_expression_item_repeat1,
    STATE(3855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5385), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [161354] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5732), 1,
      anon_sym_let,
    ACTIONS(5891), 1,
      anon_sym_and,
    ACTIONS(5893), 1,
      sym_let_and_operator,
    STATE(3843), 1,
      aux_sym_value_definition_repeat1,
    STATE(3856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5730), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161395] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5710), 1,
      anon_sym_with,
    ACTIONS(5873), 1,
      anon_sym_DASH_GT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3857), 1,
      sym_attribute,
    STATE(4063), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5905), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [161438] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5909), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5907), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161479] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5911), 1,
      anon_sym_LPAREN,
    ACTIONS(5913), 1,
      anon_sym_end,
    STATE(3859), 1,
      sym_attribute,
    STATE(3926), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [161536] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5917), 1,
      anon_sym_let,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5915), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161577] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5732), 1,
      anon_sym_let,
    ACTIONS(5891), 1,
      anon_sym_and,
    ACTIONS(5893), 1,
      sym_let_and_operator,
    STATE(3861), 1,
      sym_attribute,
    STATE(3888), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5730), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161618] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5919), 1,
      anon_sym_LPAREN,
    ACTIONS(5921), 1,
      anon_sym_end,
    STATE(3862), 1,
      sym_attribute,
    STATE(3906), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [161675] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5923), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3718), 1,
      aux_sym_expression_item_repeat1,
    STATE(3863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5419), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [161716] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3864), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5810), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5808), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [161751] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5786), 1,
      anon_sym_let,
    ACTIONS(5891), 1,
      anon_sym_and,
    ACTIONS(5893), 1,
      sym_let_and_operator,
    STATE(3865), 1,
      sym_attribute,
    STATE(3874), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5784), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161792] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5925), 1,
      anon_sym_LPAREN,
    ACTIONS(5927), 1,
      anon_sym_end,
    STATE(3866), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [161849] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5931), 1,
      anon_sym_let,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3860), 1,
      aux_sym_expression_item_repeat1,
    STATE(3867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5929), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161890] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5901), 1,
      anon_sym_and,
    ACTIONS(5935), 1,
      anon_sym_let,
    STATE(3846), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(3868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5933), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161929] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5901), 1,
      anon_sym_and,
    ACTIONS(5935), 1,
      anon_sym_let,
    STATE(3854), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(3869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5933), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161968] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5937), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3704), 1,
      aux_sym_expression_item_repeat1,
    STATE(3870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5403), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162009] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5939), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3568), 1,
      aux_sym_expression_item_repeat1,
    STATE(3871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5346), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162050] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5941), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3706), 1,
      aux_sym_expression_item_repeat1,
    STATE(3872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5409), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162091] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5943), 1,
      anon_sym_LPAREN,
    ACTIONS(5945), 1,
      anon_sym_end,
    STATE(3873), 1,
      sym_attribute,
    STATE(3951), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162148] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5826), 1,
      anon_sym_let,
    ACTIONS(5891), 1,
      anon_sym_and,
    ACTIONS(5893), 1,
      sym_let_and_operator,
    STATE(3843), 1,
      aux_sym_value_definition_repeat1,
    STATE(3874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5824), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [162189] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5947), 1,
      anon_sym_LPAREN,
    ACTIONS(5949), 1,
      anon_sym_end,
    STATE(3875), 1,
      sym_attribute,
    STATE(4012), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162246] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5360), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5895), 1,
      anon_sym_PIPE,
    STATE(3876), 1,
      sym_attribute,
    STATE(3886), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5358), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162285] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5951), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3643), 1,
      aux_sym_expression_item_repeat1,
    STATE(3877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5332), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162326] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5901), 1,
      anon_sym_and,
    ACTIONS(5955), 1,
      anon_sym_let,
    STATE(3868), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(3878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5953), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [162365] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5957), 1,
      anon_sym_LPAREN,
    ACTIONS(5959), 1,
      anon_sym_end,
    STATE(3879), 1,
      sym_attribute,
    STATE(4028), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162422] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5710), 1,
      anon_sym_with,
    ACTIONS(5873), 1,
      anon_sym_DASH_GT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3756), 1,
      aux_sym_expression_item_repeat1,
    STATE(3880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5381), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162465] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5961), 1,
      anon_sym_BANG,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3282), 1,
      sym__module_expression,
    STATE(3881), 1,
      sym_attribute,
    STATE(4153), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [162522] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5376), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5895), 1,
      anon_sym_PIPE,
    STATE(3882), 1,
      sym_attribute,
    STATE(3886), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5374), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162561] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5120), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5965), 1,
      anon_sym_LBRACE,
    ACTIONS(5968), 1,
      anon_sym_DQUOTE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
    STATE(3883), 2,
      sym_attribute,
      aux_sym_external_repeat1,
    ACTIONS(5118), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162602] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5971), 1,
      anon_sym_LPAREN,
    ACTIONS(5973), 1,
      anon_sym_end,
    STATE(3884), 1,
      sym_attribute,
    STATE(4049), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162659] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5710), 1,
      anon_sym_with,
    ACTIONS(5873), 1,
      anon_sym_DASH_GT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(3885), 1,
      sym_attribute,
    STATE(4078), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5975), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162702] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5232), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5977), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3886), 2,
      sym_attribute,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5230), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162739] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5980), 1,
      anon_sym_LPAREN,
    ACTIONS(5982), 1,
      anon_sym_end,
    STATE(3887), 1,
      sym_attribute,
    STATE(4022), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162796] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5786), 1,
      anon_sym_let,
    ACTIONS(5891), 1,
      anon_sym_and,
    ACTIONS(5893), 1,
      sym_let_and_operator,
    STATE(3843), 1,
      aux_sym_value_definition_repeat1,
    STATE(3888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5784), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [162837] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5984), 1,
      anon_sym_LPAREN,
    ACTIONS(5986), 1,
      anon_sym_end,
    STATE(3889), 1,
      sym_attribute,
    STATE(3896), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162894] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4463), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5895), 1,
      anon_sym_PIPE,
    STATE(3876), 1,
      aux_sym_variant_declaration_repeat1,
    STATE(3890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4461), 17,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162933] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5988), 1,
      anon_sym_LPAREN,
    ACTIONS(5990), 1,
      anon_sym_end,
    STATE(3891), 1,
      sym_attribute,
    STATE(3972), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162990] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5992), 1,
      anon_sym_end,
    STATE(3892), 1,
      sym_attribute,
    STATE(3894), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163044] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5994), 1,
      anon_sym_end,
    STATE(3893), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163098] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5996), 1,
      anon_sym_end,
    STATE(3894), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163152] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5986), 1,
      anon_sym_end,
    STATE(3895), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163206] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5998), 1,
      anon_sym_end,
    STATE(3896), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163260] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6000), 1,
      anon_sym_end,
    STATE(3897), 1,
      sym_attribute,
    STATE(3898), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163314] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6002), 1,
      anon_sym_end,
    STATE(3898), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163368] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6004), 1,
      anon_sym_end,
    STATE(3899), 1,
      sym_attribute,
    STATE(3901), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163422] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6006), 1,
      anon_sym_end,
    STATE(3900), 1,
      sym_attribute,
    STATE(3902), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163476] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6008), 1,
      anon_sym_end,
    STATE(3901), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163530] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6010), 1,
      anon_sym_end,
    STATE(3902), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163584] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6012), 1,
      anon_sym_end,
    STATE(3903), 1,
      sym_attribute,
    STATE(3904), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163638] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6014), 1,
      anon_sym_end,
    STATE(3904), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163692] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5921), 1,
      anon_sym_end,
    STATE(3905), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163746] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6016), 1,
      anon_sym_end,
    STATE(3906), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163800] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6018), 1,
      anon_sym_end,
    STATE(3907), 1,
      sym_attribute,
    STATE(3908), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163854] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6020), 1,
      anon_sym_end,
    STATE(3908), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163908] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6022), 1,
      anon_sym_end,
    STATE(3909), 1,
      sym_attribute,
    STATE(3911), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163962] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6024), 1,
      anon_sym_end,
    STATE(3910), 1,
      sym_attribute,
    STATE(3912), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164016] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6026), 1,
      anon_sym_end,
    STATE(3911), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164070] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6028), 1,
      anon_sym_end,
    STATE(3912), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164124] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6030), 1,
      anon_sym_end,
    STATE(3913), 1,
      sym_attribute,
    STATE(3914), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164178] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6032), 1,
      anon_sym_end,
    STATE(3914), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164232] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6036), 1,
      anon_sym_let,
    STATE(3915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6034), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164266] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6040), 1,
      anon_sym_let,
    ACTIONS(6042), 1,
      anon_sym_and,
    STATE(3916), 1,
      sym_attribute,
    STATE(3921), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6038), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164304] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3917), 1,
      sym_attribute,
    STATE(4151), 1,
      sym__attribute,
    STATE(4363), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [164358] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    ACTIONS(6044), 1,
      anon_sym_RPAREN,
    ACTIONS(6046), 1,
      anon_sym_val,
    STATE(2944), 1,
      sym_module_path,
    STATE(3918), 1,
      sym_attribute,
    STATE(4335), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [164412] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4513), 1,
      anon_sym_as,
    STATE(3919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4691), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4689), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164448] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6042), 1,
      anon_sym_and,
    ACTIONS(6050), 1,
      anon_sym_let,
    STATE(3920), 1,
      sym_attribute,
    STATE(3921), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6048), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164486] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6036), 1,
      anon_sym_let,
    ACTIONS(6052), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3921), 2,
      sym_attribute,
      aux_sym_type_definition_repeat1,
    ACTIONS(6034), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164522] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6042), 1,
      anon_sym_and,
    ACTIONS(6050), 1,
      anon_sym_let,
    STATE(3922), 1,
      sym_attribute,
    STATE(3927), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6048), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164560] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6055), 1,
      anon_sym_end,
    STATE(3923), 1,
      sym_attribute,
    STATE(3932), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164614] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5913), 1,
      anon_sym_end,
    STATE(3924), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164668] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2604), 1,
      sym__lowercase_identifier,
    ACTIONS(2608), 1,
      anon_sym_LBRACK,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2620), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2622), 1,
      anon_sym_begin,
    ACTIONS(6057), 1,
      anon_sym_LPAREN,
    ACTIONS(6059), 1,
      anon_sym_LBRACE,
    ACTIONS(6061), 1,
      sym__uppercase_identifier,
    STATE(1827), 1,
      sym__constructor_name,
    STATE(1877), 1,
      sym_parenthesized_operator,
    STATE(1923), 1,
      sym__delimited_expression,
    STATE(1943), 1,
      sym__value_name,
    STATE(3925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1926), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [164724] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6063), 1,
      anon_sym_end,
    STATE(3926), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164778] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6042), 1,
      anon_sym_and,
    ACTIONS(6067), 1,
      anon_sym_let,
    STATE(3921), 1,
      aux_sym_type_definition_repeat1,
    STATE(3927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6065), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164816] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6069), 1,
      anon_sym_end,
    STATE(3928), 1,
      sym_attribute,
    STATE(3929), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164870] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6071), 1,
      anon_sym_end,
    STATE(3929), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164924] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6073), 1,
      anon_sym_end,
    STATE(3930), 1,
      sym_attribute,
    STATE(3933), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164978] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6075), 1,
      anon_sym_end,
    STATE(3931), 1,
      sym_attribute,
    STATE(3934), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165032] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6077), 1,
      anon_sym_end,
    STATE(3932), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165086] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6079), 1,
      anon_sym_end,
    STATE(3933), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165140] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6081), 1,
      anon_sym_end,
    STATE(3934), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165194] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6083), 1,
      anon_sym_end,
    STATE(3935), 1,
      sym_attribute,
    STATE(3958), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165248] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6085), 1,
      anon_sym_end,
    STATE(3936), 1,
      sym_attribute,
    STATE(3937), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165302] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6087), 1,
      anon_sym_end,
    STATE(3937), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165356] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6091), 1,
      anon_sym_let,
    STATE(3938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6089), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [165390] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6093), 1,
      anon_sym_end,
    STATE(3939), 1,
      sym_attribute,
    STATE(3965), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165444] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5866), 1,
      anon_sym_let,
    STATE(3940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5864), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [165478] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6097), 1,
      anon_sym_let,
    STATE(3941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6095), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [165512] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(6099), 1,
      anon_sym_LPAREN,
    ACTIONS(6101), 1,
      anon_sym_LBRACE,
    ACTIONS(6103), 1,
      sym__uppercase_identifier,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1756), 1,
      sym__delimited_expression,
    STATE(1761), 1,
      sym__constructor_name,
    STATE(1789), 1,
      sym__value_name,
    STATE(3942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [165568] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5120), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6105), 1,
      anon_sym_LBRACE,
    ACTIONS(6108), 1,
      anon_sym_DQUOTE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
    STATE(3943), 2,
      sym_attribute,
      aux_sym_external_repeat1,
    ACTIONS(5118), 13,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [165608] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5945), 1,
      anon_sym_end,
    STATE(3944), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165662] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5883), 1,
      anon_sym_functor,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    STATE(3945), 1,
      sym_attribute,
    STATE(4150), 1,
      sym__attribute,
    STATE(4310), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [165716] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6113), 1,
      anon_sym_let,
    ACTIONS(6115), 1,
      anon_sym_and,
    STATE(3946), 1,
      sym_attribute,
    STATE(4038), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6111), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [165754] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1654), 1,
      sym__lowercase_identifier,
    ACTIONS(1660), 1,
      anon_sym_LBRACK,
    ACTIONS(1668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(1674), 1,
      anon_sym_begin,
    ACTIONS(6117), 1,
      anon_sym_LPAREN,
    ACTIONS(6119), 1,
      anon_sym_LBRACE,
    ACTIONS(6121), 1,
      sym__uppercase_identifier,
    STATE(2344), 1,
      sym__constructor_name,
    STATE(2373), 1,
      sym__delimited_expression,
    STATE(2403), 1,
      sym__value_name,
    STATE(2406), 1,
      sym_parenthesized_operator,
    STATE(3947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2351), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [165810] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6097), 1,
      anon_sym_let,
    ACTIONS(6123), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3948), 2,
      sym_attribute,
      aux_sym_class_definition_repeat1,
    ACTIONS(6095), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [165846] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    ACTIONS(6126), 1,
      anon_sym_RPAREN,
    ACTIONS(6128), 1,
      anon_sym_val,
    STATE(2944), 1,
      sym_module_path,
    STATE(3949), 1,
      sym_attribute,
    STATE(4433), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [165900] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2944), 1,
      sym_module_path,
    STATE(3149), 1,
      sym__module_expression,
    STATE(3950), 1,
      sym_attribute,
    STATE(4157), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [165954] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6130), 1,
      anon_sym_end,
    STATE(3951), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166008] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6115), 1,
      anon_sym_and,
    ACTIONS(6134), 1,
      anon_sym_let,
    STATE(3948), 1,
      aux_sym_class_definition_repeat1,
    STATE(3952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6132), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [166046] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5990), 1,
      anon_sym_end,
    STATE(3953), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166100] = 16,
    ACTIONS(345), 1,
      anon_sym_LBRACK,
    ACTIONS(355), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(375), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(377), 1,
      anon_sym_begin,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(737), 1,
      sym__lowercase_identifier,
    ACTIONS(6136), 1,
      anon_sym_LPAREN,
    ACTIONS(6138), 1,
      anon_sym_LBRACE,
    ACTIONS(6140), 1,
      sym__uppercase_identifier,
    STATE(793), 1,
      sym_parenthesized_operator,
    STATE(799), 1,
      sym__constructor_name,
    STATE(837), 1,
      sym__delimited_expression,
    STATE(839), 1,
      sym__value_name,
    STATE(3954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(798), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [166156] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6142), 1,
      anon_sym_end,
    STATE(3955), 1,
      sym_attribute,
    STATE(3957), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166210] = 16,
    ACTIONS(185), 1,
      anon_sym_LBRACK,
    ACTIONS(195), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(207), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(209), 1,
      anon_sym_begin,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(959), 1,
      sym__lowercase_identifier,
    ACTIONS(6144), 1,
      anon_sym_LPAREN,
    ACTIONS(6146), 1,
      anon_sym_LBRACE,
    ACTIONS(6148), 1,
      sym__uppercase_identifier,
    STATE(1077), 1,
      sym__delimited_expression,
    STATE(1093), 1,
      sym__value_name,
    STATE(1095), 1,
      sym__constructor_name,
    STATE(1115), 1,
      sym_parenthesized_operator,
    STATE(3956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1113), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [166266] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6150), 1,
      anon_sym_end,
    STATE(3957), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166320] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6152), 1,
      anon_sym_end,
    STATE(3958), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166374] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6154), 1,
      anon_sym_end,
    STATE(3959), 1,
      sym_attribute,
    STATE(3961), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166428] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6156), 1,
      anon_sym_end,
    STATE(3960), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166482] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6158), 1,
      anon_sym_end,
    STATE(3961), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166536] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3962), 1,
      sym_attribute,
    STATE(4092), 1,
      sym__attribute,
    STATE(4421), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [166590] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6160), 1,
      anon_sym_end,
    STATE(3963), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166644] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6162), 1,
      anon_sym_end,
    STATE(3964), 1,
      sym_attribute,
    STATE(3967), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166698] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6164), 1,
      anon_sym_end,
    STATE(3965), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166752] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6166), 1,
      anon_sym_end,
    STATE(3966), 1,
      sym_attribute,
    STATE(3983), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166806] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6168), 1,
      anon_sym_end,
    STATE(3967), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166860] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5877), 1,
      anon_sym_end,
    STATE(3968), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166914] = 16,
    ACTIONS(111), 1,
      anon_sym_LBRACK,
    ACTIONS(121), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(761), 1,
      sym__lowercase_identifier,
    ACTIONS(6170), 1,
      anon_sym_LPAREN,
    ACTIONS(6172), 1,
      anon_sym_LBRACE,
    ACTIONS(6174), 1,
      sym__uppercase_identifier,
    STATE(992), 1,
      sym_parenthesized_operator,
    STATE(1000), 1,
      sym__value_name,
    STATE(1016), 1,
      sym__delimited_expression,
    STATE(1035), 1,
      sym__constructor_name,
    STATE(3969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(991), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [166970] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6176), 1,
      anon_sym_end,
    STATE(3970), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167024] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6178), 1,
      anon_sym_end,
    STATE(3971), 1,
      sym_attribute,
    STATE(3973), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167078] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6180), 1,
      anon_sym_end,
    STATE(3972), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167132] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6182), 1,
      anon_sym_end,
    STATE(3973), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167186] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6184), 1,
      anon_sym_end,
    STATE(3974), 1,
      sym_attribute,
    STATE(3976), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167240] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6186), 1,
      anon_sym_end,
    STATE(3975), 1,
      sym_attribute,
    STATE(3977), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167294] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6188), 1,
      anon_sym_end,
    STATE(3976), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167348] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6190), 1,
      anon_sym_end,
    STATE(3977), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167402] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6192), 1,
      anon_sym_end,
    STATE(3978), 1,
      sym_attribute,
    STATE(3979), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167456] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6194), 1,
      anon_sym_end,
    STATE(3979), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167510] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3980), 1,
      sym_attribute,
    STATE(4088), 1,
      sym__attribute,
    STATE(4362), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [167564] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3981), 1,
      sym_attribute,
    STATE(4093), 1,
      sym__attribute,
    STATE(4331), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [167618] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3982), 1,
      sym_attribute,
    STATE(4094), 1,
      sym__attribute,
    STATE(4417), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [167672] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6196), 1,
      anon_sym_end,
    STATE(3983), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167726] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2694), 1,
      sym__lowercase_identifier,
    ACTIONS(2702), 1,
      anon_sym_begin,
    ACTIONS(6101), 1,
      anon_sym_LBRACE,
    ACTIONS(6103), 1,
      sym__uppercase_identifier,
    ACTIONS(6198), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1756), 1,
      sym__delimited_expression,
    STATE(1761), 1,
      sym__constructor_name,
    STATE(1789), 1,
      sym__value_name,
    STATE(3984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [167782] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3985), 1,
      sym_attribute,
    STATE(4098), 1,
      sym__attribute,
    STATE(4365), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [167836] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6200), 1,
      anon_sym_end,
    STATE(3986), 1,
      sym_attribute,
    STATE(4007), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167890] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5927), 1,
      anon_sym_end,
    STATE(3987), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167944] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4513), 1,
      anon_sym_as,
    STATE(3988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4970), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4968), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167980] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(609), 1,
      anon_sym_LBRACK,
    ACTIONS(619), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(639), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(641), 1,
      anon_sym_begin,
    ACTIONS(1304), 1,
      sym__lowercase_identifier,
    ACTIONS(6202), 1,
      anon_sym_LPAREN,
    ACTIONS(6204), 1,
      anon_sym_LBRACE,
    ACTIONS(6206), 1,
      sym__uppercase_identifier,
    STATE(1578), 1,
      sym__constructor_name,
    STATE(1579), 1,
      sym__value_name,
    STATE(1607), 1,
      sym_parenthesized_operator,
    STATE(1616), 1,
      sym__delimited_expression,
    STATE(3989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1589), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [168036] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3990), 1,
      sym_attribute,
    STATE(4106), 1,
      sym__attribute,
    STATE(4418), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168090] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2534), 1,
      anon_sym_LBRACK,
    ACTIONS(2540), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2544), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2546), 1,
      anon_sym_begin,
    ACTIONS(6099), 1,
      anon_sym_LPAREN,
    ACTIONS(6101), 1,
      anon_sym_LBRACE,
    ACTIONS(6103), 1,
      sym__uppercase_identifier,
    ACTIONS(6208), 1,
      sym__lowercase_identifier,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(1756), 1,
      sym__delimited_expression,
    STATE(1761), 1,
      sym__constructor_name,
    STATE(1789), 1,
      sym__value_name,
    STATE(3991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1723), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [168146] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4513), 1,
      anon_sym_as,
    STATE(3992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4962), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4960), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [168182] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6210), 1,
      anon_sym_end,
    STATE(3993), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168236] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6212), 1,
      anon_sym_end,
    STATE(3994), 1,
      sym_attribute,
    STATE(3995), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168290] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6214), 1,
      anon_sym_end,
    STATE(3995), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168344] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3996), 1,
      sym_attribute,
    STATE(4110), 1,
      sym__attribute,
    STATE(4328), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168398] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3997), 1,
      sym_attribute,
    STATE(4112), 1,
      sym__attribute,
    STATE(4341), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168452] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6216), 1,
      anon_sym_end,
    STATE(3998), 1,
      sym_attribute,
    STATE(4001), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168506] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3263), 1,
      sym__module_expression,
    STATE(3999), 1,
      sym_attribute,
    STATE(4161), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168560] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6218), 1,
      anon_sym_end,
    STATE(4000), 1,
      sym_attribute,
    STATE(4002), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168614] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6220), 1,
      anon_sym_end,
    STATE(4001), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168668] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6222), 1,
      anon_sym_end,
    STATE(4002), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168722] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6224), 1,
      anon_sym_end,
    STATE(4003), 1,
      sym_attribute,
    STATE(4004), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168776] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6226), 1,
      anon_sym_end,
    STATE(4004), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168830] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6042), 1,
      anon_sym_and,
    ACTIONS(6230), 1,
      anon_sym_let,
    STATE(3916), 1,
      aux_sym_type_definition_repeat1,
    STATE(4005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6228), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [168868] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2944), 1,
      sym_module_path,
    STATE(3155), 1,
      sym__module_expression,
    STATE(4006), 1,
      sym_attribute,
    STATE(4137), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168922] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6232), 1,
      anon_sym_end,
    STATE(4007), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168976] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6234), 1,
      anon_sym_end,
    STATE(4008), 1,
      sym_attribute,
    STATE(4024), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169030] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6040), 1,
      anon_sym_let,
    ACTIONS(6042), 1,
      anon_sym_and,
    STATE(3920), 1,
      aux_sym_type_definition_repeat1,
    STATE(4009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6038), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169068] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5949), 1,
      anon_sym_end,
    STATE(4010), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169122] = 16,
    ACTIONS(45), 1,
      anon_sym_LBRACK,
    ACTIONS(55), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(75), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(77), 1,
      anon_sym_begin,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(751), 1,
      sym__lowercase_identifier,
    ACTIONS(6236), 1,
      anon_sym_LPAREN,
    ACTIONS(6238), 1,
      anon_sym_LBRACE,
    ACTIONS(6240), 1,
      sym__uppercase_identifier,
    STATE(907), 1,
      sym__delimited_expression,
    STATE(908), 1,
      sym_parenthesized_operator,
    STATE(919), 1,
      sym__constructor_name,
    STATE(942), 1,
      sym__value_name,
    STATE(4011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(922), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [169178] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6242), 1,
      anon_sym_end,
    STATE(4012), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169232] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6244), 1,
      anon_sym_end,
    STATE(4013), 1,
      sym_attribute,
    STATE(4014), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169286] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6246), 1,
      anon_sym_end,
    STATE(4014), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169340] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6248), 1,
      anon_sym_end,
    STATE(4015), 1,
      sym_attribute,
    STATE(4017), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169394] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6250), 1,
      anon_sym_end,
    STATE(4016), 1,
      sym_attribute,
    STATE(4018), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169448] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6252), 1,
      anon_sym_end,
    STATE(4017), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169502] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6254), 1,
      anon_sym_end,
    STATE(4018), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169556] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6256), 1,
      anon_sym_end,
    STATE(4019), 1,
      sym_attribute,
    STATE(4020), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169610] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6258), 1,
      anon_sym_end,
    STATE(4020), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169664] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6260), 1,
      anon_sym_end,
    STATE(4021), 1,
      sym_attribute,
    STATE(4037), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169718] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6262), 1,
      anon_sym_end,
    STATE(4022), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169772] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6264), 1,
      anon_sym_constraint,
    ACTIONS(6267), 1,
      anon_sym_val,
    ACTIONS(6270), 1,
      anon_sym_end,
    ACTIONS(6272), 1,
      anon_sym_inherit,
    ACTIONS(6275), 1,
      anon_sym_method,
    ACTIONS(6278), 1,
      anon_sym_initializer,
    ACTIONS(6281), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(6284), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(6287), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4023), 2,
      sym_attribute,
      aux_sym_object_expression_repeat1,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169824] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6290), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169878] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4025), 1,
      sym_attribute,
    STATE(4113), 1,
      sym__attribute,
    STATE(4360), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [169932] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5959), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169986] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2656), 1,
      sym__lowercase_identifier,
    ACTIONS(2660), 1,
      anon_sym_LBRACK,
    ACTIONS(2668), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2672), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2674), 1,
      anon_sym_begin,
    ACTIONS(6292), 1,
      anon_sym_LPAREN,
    ACTIONS(6294), 1,
      anon_sym_LBRACE,
    ACTIONS(6296), 1,
      sym__uppercase_identifier,
    STATE(2680), 1,
      sym__delimited_expression,
    STATE(2712), 1,
      sym_parenthesized_operator,
    STATE(2753), 1,
      sym__constructor_name,
    STATE(2772), 1,
      sym__value_name,
    STATE(4027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2713), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [170042] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6298), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170096] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6300), 1,
      anon_sym_end,
    STATE(4029), 1,
      sym_attribute,
    STATE(4030), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170150] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6302), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170204] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6304), 1,
      anon_sym_end,
    STATE(4031), 1,
      sym_attribute,
    STATE(4033), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170258] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6306), 1,
      anon_sym_end,
    STATE(4032), 1,
      sym_attribute,
    STATE(4034), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170312] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6308), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170366] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6310), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170420] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6312), 1,
      anon_sym_end,
    STATE(4035), 1,
      sym_attribute,
    STATE(4036), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170474] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6314), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170528] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6316), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170582] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6115), 1,
      anon_sym_and,
    ACTIONS(6320), 1,
      anon_sym_let,
    STATE(3948), 1,
      aux_sym_class_definition_repeat1,
    STATE(4038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6318), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170620] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6115), 1,
      anon_sym_and,
    ACTIONS(6320), 1,
      anon_sym_let,
    STATE(3952), 1,
      aux_sym_class_definition_repeat1,
    STATE(4039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6318), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170658] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6322), 1,
      anon_sym_end,
    STATE(4040), 1,
      sym_attribute,
    STATE(4041), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170712] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6324), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170766] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5973), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170820] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2564), 1,
      sym__lowercase_identifier,
    ACTIONS(2568), 1,
      anon_sym_LBRACK,
    ACTIONS(2576), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2580), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2582), 1,
      anon_sym_begin,
    ACTIONS(6326), 1,
      anon_sym_LPAREN,
    ACTIONS(6328), 1,
      anon_sym_LBRACE,
    ACTIONS(6330), 1,
      sym__uppercase_identifier,
    STATE(1960), 1,
      sym__delimited_expression,
    STATE(1986), 1,
      sym__value_name,
    STATE(2005), 1,
      sym_parenthesized_operator,
    STATE(2033), 1,
      sym__constructor_name,
    STATE(4043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2122), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [170876] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6334), 1,
      anon_sym_let,
    ACTIONS(6336), 1,
      anon_sym_and,
    STATE(4044), 1,
      sym_attribute,
    STATE(4046), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6332), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170914] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6336), 1,
      anon_sym_and,
    ACTIONS(6340), 1,
      anon_sym_let,
    STATE(4045), 1,
      sym_attribute,
    STATE(4046), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6338), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170952] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6091), 1,
      anon_sym_let,
    ACTIONS(6342), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4046), 2,
      sym_attribute,
      aux_sym_module_definition_repeat1,
    ACTIONS(6089), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170988] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6336), 1,
      anon_sym_and,
    ACTIONS(6340), 1,
      anon_sym_let,
    STATE(4047), 1,
      sym_attribute,
    STATE(4048), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6338), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171026] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6336), 1,
      anon_sym_and,
    ACTIONS(6347), 1,
      anon_sym_let,
    STATE(4046), 1,
      aux_sym_module_definition_repeat1,
    STATE(4048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6345), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171064] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6349), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171118] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6336), 1,
      anon_sym_and,
    ACTIONS(6353), 1,
      anon_sym_let,
    STATE(4044), 1,
      aux_sym_module_definition_repeat1,
    STATE(4050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6351), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171156] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6334), 1,
      anon_sym_let,
    ACTIONS(6336), 1,
      anon_sym_and,
    STATE(4045), 1,
      aux_sym_module_definition_repeat1,
    STATE(4051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6332), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171194] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6355), 1,
      anon_sym_end,
    STATE(4052), 1,
      sym_attribute,
    STATE(4054), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171248] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(5982), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171302] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6357), 1,
      anon_sym_end,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    STATE(4054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171356] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6359), 1,
      anon_sym_end,
    STATE(3960), 1,
      aux_sym_object_expression_repeat1,
    STATE(4055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171410] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6361), 1,
      anon_sym_end,
    STATE(3893), 1,
      aux_sym_object_expression_repeat1,
    STATE(4056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171464] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(5190), 1,
      anon_sym_val,
    ACTIONS(5194), 1,
      anon_sym_inherit,
    ACTIONS(5196), 1,
      anon_sym_method,
    ACTIONS(5198), 1,
      anon_sym_initializer,
    ACTIONS(6363), 1,
      anon_sym_end,
    STATE(3963), 1,
      aux_sym_object_expression_repeat1,
    STATE(4057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4614), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4613), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171518] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(6365), 1,
      anon_sym_LPAREN,
    ACTIONS(6367), 1,
      anon_sym_val,
    ACTIONS(6369), 1,
      anon_sym_end,
    ACTIONS(6371), 1,
      anon_sym_inherit,
    ACTIONS(6373), 1,
      anon_sym_method,
    STATE(4058), 1,
      sym_attribute,
    STATE(4097), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4641), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4621), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [171571] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4924), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5492), 1,
      anon_sym_of,
    ACTIONS(5494), 1,
      anon_sym_COLON2,
    ACTIONS(6375), 1,
      anon_sym_EQ,
    STATE(4059), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4922), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [171610] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6377), 1,
      anon_sym_EQ,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4060), 1,
      sym_attribute,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(4272), 1,
      sym__type_constrained,
    STATE(4420), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6381), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [171659] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5112), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5710), 1,
      anon_sym_with,
    ACTIONS(5873), 1,
      anon_sym_DASH_GT,
    STATE(4061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5110), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [171696] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6385), 1,
      anon_sym_let,
    ACTIONS(6387), 1,
      anon_sym_DOT,
    STATE(4062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6383), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171731] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6389), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [171768] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4982), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5710), 1,
      anon_sym_with,
    ACTIONS(5873), 1,
      anon_sym_DASH_GT,
    STATE(4064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4980), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [171805] = 14,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    ACTIONS(6046), 1,
      anon_sym_val,
    STATE(2944), 1,
      sym_module_path,
    STATE(4065), 1,
      sym_attribute,
    STATE(4335), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [171856] = 14,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    ACTIONS(6128), 1,
      anon_sym_val,
    STATE(2944), 1,
      sym_module_path,
    STATE(4066), 1,
      sym_attribute,
    STATE(4433), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [171907] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1902), 1,
      anon_sym_PIPE,
    STATE(4067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1904), 17,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [171940] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(6391), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4068), 1,
      sym_attribute,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(4286), 1,
      sym__type_constrained,
    STATE(4342), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6393), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [171989] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(6395), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4069), 1,
      sym_attribute,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(4298), 1,
      sym__type_constrained,
    STATE(4412), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6397), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172038] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5018), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5710), 1,
      anon_sym_with,
    ACTIONS(5873), 1,
      anon_sym_DASH_GT,
    STATE(4070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5016), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172075] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1906), 1,
      anon_sym_PIPE,
    STATE(4071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1908), 17,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [172108] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(6367), 1,
      anon_sym_val,
    ACTIONS(6371), 1,
      anon_sym_inherit,
    ACTIONS(6373), 1,
      anon_sym_method,
    ACTIONS(6399), 1,
      anon_sym_LPAREN,
    ACTIONS(6401), 1,
      anon_sym_end,
    STATE(4072), 1,
      sym_attribute,
    STATE(4087), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4641), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4621), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [172161] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(6403), 1,
      sym__lowercase_identifier,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6407), 1,
      anon_sym_nonrec,
    ACTIONS(6411), 1,
      anon_sym__,
    STATE(4005), 1,
      sym_type_binding,
    STATE(4073), 1,
      sym_attribute,
    STATE(4181), 1,
      sym__attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5322), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6840), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [172222] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1922), 1,
      anon_sym_PIPE,
    STATE(4074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1924), 17,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [172255] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    ACTIONS(6413), 1,
      sym__lowercase_identifier,
    ACTIONS(6415), 1,
      anon_sym_nonrec,
    STATE(4075), 1,
      sym_attribute,
    STATE(4168), 1,
      sym__attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5436), 1,
      sym__type_params,
    STATE(5813), 1,
      sym_type_binding,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6532), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [172316] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1950), 1,
      anon_sym_PIPE,
    STATE(4076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1952), 17,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [172349] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6419), 1,
      anon_sym_let,
    ACTIONS(6421), 1,
      anon_sym_in,
    STATE(4077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6417), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172384] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6423), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172421] = 19,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    ACTIONS(6425), 1,
      sym__lowercase_identifier,
    ACTIONS(6427), 1,
      anon_sym_nonrec,
    STATE(4079), 1,
      sym_attribute,
    STATE(4107), 1,
      sym_type_binding,
    STATE(4183), 1,
      sym__attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5367), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(7077), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [172482] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5100), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5710), 1,
      anon_sym_with,
    ACTIONS(5873), 1,
      anon_sym_DASH_GT,
    STATE(4080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5098), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172519] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(6367), 1,
      anon_sym_val,
    ACTIONS(6371), 1,
      anon_sym_inherit,
    ACTIONS(6373), 1,
      anon_sym_method,
    ACTIONS(6429), 1,
      anon_sym_end,
    STATE(4081), 1,
      sym_attribute,
    STATE(4132), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4641), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4621), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [172569] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6431), 1,
      anon_sym_and,
    STATE(4082), 1,
      sym_attribute,
    STATE(4162), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6111), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172603] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1618), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1620), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [172635] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6433), 1,
      anon_sym_and,
    STATE(4084), 1,
      sym_attribute,
    STATE(4130), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6065), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172669] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6435), 1,
      anon_sym_and,
    STATE(4085), 1,
      sym_attribute,
    STATE(4116), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6338), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172703] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1938), 1,
      anon_sym_let,
    STATE(4086), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1940), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172735] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(6367), 1,
      anon_sym_val,
    ACTIONS(6371), 1,
      anon_sym_inherit,
    ACTIONS(6373), 1,
      anon_sym_method,
    ACTIONS(6437), 1,
      anon_sym_end,
    STATE(4087), 1,
      sym_attribute,
    STATE(4132), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4641), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4621), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [172785] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4088), 1,
      sym_attribute,
    STATE(4416), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [172833] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1618), 1,
      anon_sym_let,
    STATE(4089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1620), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172865] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(6367), 1,
      anon_sym_val,
    ACTIONS(6371), 1,
      anon_sym_inherit,
    ACTIONS(6373), 1,
      anon_sym_method,
    ACTIONS(6439), 1,
      anon_sym_end,
    STATE(4090), 1,
      sym_attribute,
    STATE(4132), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4641), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4621), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [172915] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6435), 1,
      anon_sym_and,
    STATE(4091), 1,
      sym_attribute,
    STATE(4160), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6351), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172949] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4092), 1,
      sym_attribute,
    STATE(4432), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [172997] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4093), 1,
      sym_attribute,
    STATE(4402), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173045] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4094), 1,
      sym_attribute,
    STATE(4325), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173093] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1622), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1624), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [173125] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1622), 1,
      anon_sym_let,
    STATE(4096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1624), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173157] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(6367), 1,
      anon_sym_val,
    ACTIONS(6371), 1,
      anon_sym_inherit,
    ACTIONS(6373), 1,
      anon_sym_method,
    ACTIONS(6441), 1,
      anon_sym_end,
    STATE(4097), 1,
      sym_attribute,
    STATE(4132), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4641), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4621), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [173207] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4098), 1,
      sym_attribute,
    STATE(4401), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173255] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6433), 1,
      anon_sym_and,
    STATE(4099), 1,
      sym_attribute,
    STATE(4130), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6048), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [173289] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1634), 1,
      anon_sym_let,
    STATE(4100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1636), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173321] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2944), 1,
      sym_module_path,
    STATE(3155), 1,
      sym__module_expression,
    STATE(4101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173369] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    ACTIONS(6443), 1,
      anon_sym_functor,
    STATE(4102), 1,
      sym_attribute,
    STATE(4246), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173417] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1750), 1,
      anon_sym_let,
    STATE(4103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1752), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173449] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2783), 1,
      sym__module_expression,
    STATE(2944), 1,
      sym_module_path,
    STATE(4104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173497] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2944), 1,
      sym_module_path,
    STATE(2964), 1,
      sym__module_expression,
    STATE(4105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173545] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4106), 1,
      sym_attribute,
    STATE(4424), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173593] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6433), 1,
      anon_sym_and,
    STATE(4107), 1,
      sym_attribute,
    STATE(4142), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6228), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [173627] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2944), 1,
      sym_module_path,
    STATE(2965), 1,
      sym__module_expression,
    STATE(4108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173675] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1634), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1636), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [173707] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4110), 1,
      sym_attribute,
    STATE(4332), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173755] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2944), 1,
      sym_module_path,
    STATE(2966), 1,
      sym__module_expression,
    STATE(4111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173803] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4112), 1,
      sym_attribute,
    STATE(4349), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173851] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4113), 1,
      sym_attribute,
    STATE(4329), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173899] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6385), 1,
      anon_sym_let,
    STATE(4114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6383), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173931] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1934), 1,
      anon_sym_let,
    STATE(4115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1936), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173963] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6445), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4116), 2,
      sym_attribute,
      aux_sym_module_definition_repeat1,
    ACTIONS(6089), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [173995] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6435), 1,
      anon_sym_and,
    STATE(4117), 1,
      sym_attribute,
    STATE(4120), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6338), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [174029] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1716), 1,
      anon_sym_let,
    STATE(4118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1718), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174061] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2784), 1,
      sym__module_expression,
    STATE(2944), 1,
      sym_module_path,
    STATE(4119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174109] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6435), 1,
      anon_sym_and,
    STATE(4116), 1,
      aux_sym_module_definition_repeat1,
    STATE(4120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6345), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [174143] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1842), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1844), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [174175] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3128), 1,
      sym__module_expression,
    STATE(4122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174223] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6448), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4123), 2,
      sym_attribute,
      aux_sym_class_definition_repeat1,
    ACTIONS(6095), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [174255] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6431), 1,
      anon_sym_and,
    STATE(4123), 1,
      aux_sym_class_definition_repeat1,
    STATE(4124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6132), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [174289] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(6367), 1,
      anon_sym_val,
    ACTIONS(6371), 1,
      anon_sym_inherit,
    ACTIONS(6373), 1,
      anon_sym_method,
    ACTIONS(6451), 1,
      anon_sym_end,
    STATE(4090), 1,
      aux_sym_class_body_type_repeat1,
    STATE(4125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4641), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4621), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [174339] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2897), 1,
      sym__module_expression,
    STATE(2944), 1,
      sym_module_path,
    STATE(4126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174387] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1918), 1,
      anon_sym_let,
    STATE(4127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1920), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174419] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    ACTIONS(6443), 1,
      anon_sym_functor,
    STATE(4128), 1,
      sym_attribute,
    STATE(4249), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174467] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2944), 1,
      sym_module_path,
    STATE(2946), 1,
      sym__module_expression,
    STATE(4129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174515] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6453), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4130), 2,
      sym_attribute,
      aux_sym_type_definition_repeat1,
    ACTIONS(6034), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [174547] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6458), 1,
      anon_sym_let,
    STATE(4131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6456), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174579] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6460), 1,
      anon_sym_constraint,
    ACTIONS(6463), 1,
      anon_sym_val,
    ACTIONS(6466), 1,
      anon_sym_end,
    ACTIONS(6468), 1,
      anon_sym_inherit,
    ACTIONS(6471), 1,
      anon_sym_method,
    ACTIONS(6474), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(6477), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(6480), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4132), 2,
      sym_attribute,
      aux_sym_class_body_type_repeat1,
    STATE(4641), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4621), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [174627] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(27), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(29), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(31), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5188), 1,
      anon_sym_constraint,
    ACTIONS(6367), 1,
      anon_sym_val,
    ACTIONS(6371), 1,
      anon_sym_inherit,
    ACTIONS(6373), 1,
      anon_sym_method,
    ACTIONS(6483), 1,
      anon_sym_end,
    STATE(4081), 1,
      aux_sym_class_body_type_repeat1,
    STATE(4133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4641), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4621), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [174677] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5883), 1,
      anon_sym_functor,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    STATE(4134), 1,
      sym_attribute,
    STATE(4310), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174725] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    ACTIONS(6443), 1,
      anon_sym_functor,
    STATE(4135), 1,
      sym_attribute,
    STATE(4172), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174773] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5883), 1,
      anon_sym_functor,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    STATE(4136), 1,
      sym_attribute,
    STATE(4305), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174821] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2944), 1,
      sym_module_path,
    STATE(3181), 1,
      sym__module_expression,
    STATE(4137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174869] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    ACTIONS(6443), 1,
      anon_sym_functor,
    STATE(4138), 1,
      sym_attribute,
    STATE(4250), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174917] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6419), 1,
      anon_sym_let,
    STATE(4139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6417), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174949] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2944), 1,
      sym_module_path,
    STATE(2947), 1,
      sym__module_expression,
    STATE(4140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174997] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6487), 1,
      anon_sym_let,
    STATE(4141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6485), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175029] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6433), 1,
      anon_sym_and,
    STATE(4130), 1,
      aux_sym_type_definition_repeat1,
    STATE(4142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6038), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175063] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5883), 1,
      anon_sym_functor,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    STATE(4143), 1,
      sym_attribute,
    STATE(4297), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175111] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1926), 1,
      anon_sym_let,
    STATE(4144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1928), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175143] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(6489), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4145), 1,
      sym_attribute,
    STATE(4230), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175191] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(6489), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4146), 1,
      sym_attribute,
    STATE(4231), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175239] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    ACTIONS(6443), 1,
      anon_sym_functor,
    STATE(4147), 1,
      sym_attribute,
    STATE(4191), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175287] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6433), 1,
      anon_sym_and,
    STATE(4084), 1,
      aux_sym_type_definition_repeat1,
    STATE(4148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6048), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175321] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    ACTIONS(6443), 1,
      anon_sym_functor,
    STATE(4149), 1,
      sym_attribute,
    STATE(4236), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175369] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5883), 1,
      anon_sym_functor,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    STATE(4150), 1,
      sym_attribute,
    STATE(4260), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175417] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(4151), 1,
      sym_attribute,
    STATE(4413), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175465] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6435), 1,
      anon_sym_and,
    STATE(4085), 1,
      aux_sym_module_definition_repeat1,
    STATE(4152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6332), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175499] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3263), 1,
      sym__module_expression,
    STATE(4153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175547] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3157), 1,
      sym__module_expression,
    STATE(4154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175595] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1842), 1,
      anon_sym_let,
    STATE(4155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1844), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175627] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6491), 1,
      anon_sym_let,
    STATE(4156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3830), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175659] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5858), 1,
      anon_sym_functor,
    ACTIONS(5860), 1,
      anon_sym_struct,
    STATE(2944), 1,
      sym_module_path,
    STATE(3151), 1,
      sym__module_expression,
    STATE(4157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175707] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    ACTIONS(6443), 1,
      anon_sym_functor,
    STATE(4158), 1,
      sym_attribute,
    STATE(4237), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175755] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5879), 1,
      anon_sym_LPAREN,
    ACTIONS(5885), 1,
      anon_sym_struct,
    ACTIONS(5887), 1,
      sym__uppercase_identifier,
    ACTIONS(6443), 1,
      anon_sym_functor,
    STATE(4159), 1,
      sym_attribute,
    STATE(4238), 1,
      sym__module_expression,
    STATE(4486), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4528), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175803] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6435), 1,
      anon_sym_and,
    STATE(4116), 1,
      aux_sym_module_definition_repeat1,
    STATE(4160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6332), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175837] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5860), 1,
      anon_sym_struct,
    ACTIONS(5963), 1,
      anon_sym_functor,
    STATE(2944), 1,
      sym_module_path,
    STATE(3290), 1,
      sym__module_expression,
    STATE(4161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3124), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175885] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6431), 1,
      anon_sym_and,
    STATE(4123), 1,
      aux_sym_class_definition_repeat1,
    STATE(4162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6318), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175919] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6431), 1,
      anon_sym_and,
    STATE(4124), 1,
      aux_sym_class_definition_repeat1,
    STATE(4163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6318), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175953] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1930), 1,
      anon_sym_let,
    STATE(4164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1932), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [175985] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6433), 1,
      anon_sym_and,
    STATE(4099), 1,
      aux_sym_type_definition_repeat1,
    STATE(4165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6038), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176019] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6493), 1,
      anon_sym_DQUOTE,
    STATE(4166), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(6806), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176060] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6501), 1,
      anon_sym_DQUOTE,
    STATE(4167), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(6482), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176101] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    ACTIONS(6413), 1,
      sym__lowercase_identifier,
    ACTIONS(6503), 1,
      anon_sym_nonrec,
    STATE(4168), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5436), 1,
      sym__type_params,
    STATE(5832), 1,
      sym_type_binding,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6532), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [176156] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6505), 1,
      anon_sym_DQUOTE,
    STATE(4169), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(7194), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176197] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6507), 1,
      anon_sym_DQUOTE,
    STATE(4170), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(6950), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176238] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6509), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4171), 2,
      sym_attribute,
      aux_sym__structure_repeat1,
    ACTIONS(2834), 14,
      ts_builtin_sym_end,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176269] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4172), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4060), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [176308] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6514), 1,
      anon_sym_DQUOTE,
    STATE(4173), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(7087), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176349] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6516), 1,
      anon_sym_DQUOTE,
    STATE(4174), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(6763), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176390] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6518), 1,
      anon_sym_DQUOTE,
    STATE(4175), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(6399), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176431] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6520), 1,
      anon_sym_DQUOTE,
    STATE(4176), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(6345), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176472] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1834), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(1836), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [176503] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6522), 1,
      anon_sym_DQUOTE,
    STATE(4178), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(7030), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176544] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6524), 1,
      anon_sym_DQUOTE,
    STATE(4179), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(6571), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176585] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6526), 1,
      anon_sym_DQUOTE,
    STATE(4180), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(6996), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176626] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6403), 1,
      sym__lowercase_identifier,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    ACTIONS(6528), 1,
      anon_sym_nonrec,
    STATE(4009), 1,
      sym_type_binding,
    STATE(4181), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5322), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6840), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [176681] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6530), 1,
      anon_sym_DQUOTE,
    STATE(4182), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(6362), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176722] = 17,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    ACTIONS(6425), 1,
      sym__lowercase_identifier,
    ACTIONS(6532), 1,
      anon_sym_nonrec,
    STATE(4165), 1,
      sym_type_binding,
    STATE(4183), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5367), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(7077), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [176777] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6534), 1,
      anon_sym_DQUOTE,
    STATE(4184), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(7261), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176818] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6536), 1,
      anon_sym_DQUOTE,
    STATE(4185), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(6898), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176859] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6538), 1,
      anon_sym_DQUOTE,
    STATE(4186), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(7148), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176900] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6540), 1,
      anon_sym_DQUOTE,
    STATE(4187), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(6846), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176941] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6542), 1,
      anon_sym_DQUOTE,
    STATE(4188), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(7041), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176982] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6544), 1,
      anon_sym_DQUOTE,
    STATE(4189), 1,
      sym_attribute,
    STATE(4228), 1,
      aux_sym_string_content_repeat1,
    STATE(4279), 1,
      sym_escape_sequence,
    STATE(7128), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [177023] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1642), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(1644), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [177054] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4191), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4070), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [177093] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    ACTIONS(6413), 1,
      sym__lowercase_identifier,
    STATE(4192), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5436), 1,
      sym__type_params,
    STATE(5799), 1,
      sym_type_binding,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6532), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177145] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    ACTIONS(6413), 1,
      sym__lowercase_identifier,
    STATE(4193), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5436), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6258), 1,
      sym_type_binding,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6532), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177197] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(3808), 1,
      anon_sym_let,
    ACTIONS(6546), 1,
      sym__lowercase_identifier,
    ACTIONS(6548), 1,
      anon_sym_LBRACK,
    STATE(4194), 1,
      sym_attribute,
    STATE(4450), 1,
      sym__simple_class_type,
    STATE(6229), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3472), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177243] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6550), 1,
      sym__lowercase_identifier,
    ACTIONS(6552), 1,
      anon_sym_LBRACK,
    STATE(3566), 1,
      sym__simple_class_type,
    STATE(4195), 1,
      sym_attribute,
    STATE(6330), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3472), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177289] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(6554), 1,
      anon_sym_DASH_GT,
    ACTIONS(6556), 1,
      anon_sym_STAR,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(4196), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 3,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3718), 5,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
  [177333] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2009), 2,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6558), 13,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [177363] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6550), 1,
      sym__lowercase_identifier,
    ACTIONS(6552), 1,
      anon_sym_LBRACK,
    STATE(3681), 1,
      sym__simple_class_type,
    STATE(4198), 1,
      sym_attribute,
    STATE(6330), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3472), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177409] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    ACTIONS(6425), 1,
      sym__lowercase_identifier,
    STATE(4148), 1,
      sym_type_binding,
    STATE(4199), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5367), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(7077), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177461] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3102), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3104), 13,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      sym__uppercase_identifier,
  [177491] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2896), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [177519] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3748), 1,
      anon_sym_let,
    ACTIONS(3754), 1,
      anon_sym_object,
    ACTIONS(6560), 1,
      sym__lowercase_identifier,
    ACTIONS(6562), 1,
      anon_sym_LBRACK,
    STATE(4202), 1,
      sym_attribute,
    STATE(5093), 1,
      sym__simple_class_type,
    STATE(6267), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5335), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177565] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3748), 1,
      anon_sym_let,
    ACTIONS(3754), 1,
      anon_sym_object,
    ACTIONS(6560), 1,
      sym__lowercase_identifier,
    ACTIONS(6562), 1,
      anon_sym_LBRACK,
    STATE(4203), 1,
      sym_attribute,
    STATE(5487), 1,
      sym__simple_class_type,
    STATE(6267), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5335), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177611] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6564), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [177639] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6566), 1,
      anon_sym_DQUOTE,
    STATE(4279), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6571), 2,
      sym__null,
      aux_sym_string_content_token2,
    STATE(4205), 2,
      sym_attribute,
      aux_sym_string_content_repeat1,
    ACTIONS(6574), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6568), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [177675] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    ACTIONS(6425), 1,
      sym__lowercase_identifier,
    STATE(3915), 1,
      sym_type_binding,
    STATE(4206), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5367), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(7077), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177727] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(3808), 1,
      anon_sym_let,
    ACTIONS(6546), 1,
      sym__lowercase_identifier,
    ACTIONS(6548), 1,
      anon_sym_LBRACK,
    STATE(3380), 1,
      sym__simple_class_type,
    STATE(4207), 1,
      sym_attribute,
    STATE(6229), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3472), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177773] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6550), 1,
      sym__lowercase_identifier,
    ACTIONS(6552), 1,
      anon_sym_LBRACK,
    STATE(3380), 1,
      sym__simple_class_type,
    STATE(4208), 1,
      sym_attribute,
    STATE(6330), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3472), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177819] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6403), 1,
      sym__lowercase_identifier,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    STATE(4009), 1,
      sym_type_binding,
    STATE(4209), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5322), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6840), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177871] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6403), 1,
      sym__lowercase_identifier,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    STATE(3922), 1,
      sym_type_binding,
    STATE(4210), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5322), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6840), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177923] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3748), 1,
      anon_sym_let,
    ACTIONS(3754), 1,
      anon_sym_object,
    ACTIONS(6560), 1,
      sym__lowercase_identifier,
    ACTIONS(6562), 1,
      anon_sym_LBRACK,
    STATE(4211), 1,
      sym_attribute,
    STATE(5148), 1,
      sym__simple_class_type,
    STATE(6267), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5335), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177969] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3748), 1,
      anon_sym_let,
    ACTIONS(3754), 1,
      anon_sym_object,
    ACTIONS(6560), 1,
      sym__lowercase_identifier,
    ACTIONS(6562), 1,
      anon_sym_LBRACK,
    STATE(4212), 1,
      sym_attribute,
    STATE(5209), 1,
      sym__simple_class_type,
    STATE(6267), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5335), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178015] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6550), 1,
      sym__lowercase_identifier,
    ACTIONS(6552), 1,
      anon_sym_LBRACK,
    STATE(3608), 1,
      sym__simple_class_type,
    STATE(4213), 1,
      sym_attribute,
    STATE(6330), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3472), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178061] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1468), 1,
      aux_sym_number_token1,
    ACTIONS(1470), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1472), 1,
      anon_sym_SQUOTE,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(6579), 1,
      anon_sym_LBRACE,
    STATE(3059), 1,
      sym__signed_constant,
    STATE(4214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6577), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2949), 2,
      sym__constant,
      sym_signed_number,
    STATE(2953), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [178105] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4270), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [178133] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(6581), 1,
      anon_sym_STAR,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(4216), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3740), 4,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3742), 5,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
  [178175] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3748), 1,
      anon_sym_let,
    ACTIONS(3754), 1,
      anon_sym_object,
    ACTIONS(6560), 1,
      sym__lowercase_identifier,
    ACTIONS(6562), 1,
      anon_sym_LBRACK,
    STATE(4217), 1,
      sym_attribute,
    STATE(5089), 1,
      sym__simple_class_type,
    STATE(6267), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5335), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178221] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2298), 1,
      aux_sym_number_token1,
    ACTIONS(2300), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2302), 1,
      anon_sym_SQUOTE,
    ACTIONS(2304), 1,
      anon_sym_DQUOTE,
    ACTIONS(6585), 1,
      anon_sym_LBRACE,
    STATE(4218), 1,
      sym_attribute,
    STATE(4925), 1,
      sym__signed_constant,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6583), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4788), 2,
      sym__constant,
      sym_signed_number,
    STATE(4738), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [178265] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    ACTIONS(6413), 1,
      sym__lowercase_identifier,
    STATE(4219), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5436), 1,
      sym__type_params,
    STATE(5832), 1,
      sym_type_binding,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6532), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [178317] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3106), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3108), 13,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      sym__uppercase_identifier,
  [178347] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6403), 1,
      sym__lowercase_identifier,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    STATE(3915), 1,
      sym_type_binding,
    STATE(4221), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5322), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6840), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [178399] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6550), 1,
      sym__lowercase_identifier,
    ACTIONS(6552), 1,
      anon_sym_LBRACK,
    STATE(3697), 1,
      sym__simple_class_type,
    STATE(4222), 1,
      sym_attribute,
    STATE(6330), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3472), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178445] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3786), 1,
      anon_sym_let,
    ACTIONS(3792), 1,
      anon_sym_object,
    ACTIONS(5690), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(5692), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6550), 1,
      sym__lowercase_identifier,
    ACTIONS(6552), 1,
      anon_sym_LBRACK,
    STATE(3710), 1,
      sym__simple_class_type,
    STATE(4223), 1,
      sym_attribute,
    STATE(6330), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3472), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178491] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3748), 1,
      anon_sym_let,
    ACTIONS(3754), 1,
      anon_sym_object,
    ACTIONS(6560), 1,
      sym__lowercase_identifier,
    ACTIONS(6562), 1,
      anon_sym_LBRACK,
    STATE(4224), 1,
      sym_attribute,
    STATE(5129), 1,
      sym__simple_class_type,
    STATE(6267), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5335), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178537] = 16,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    ACTIONS(6425), 1,
      sym__lowercase_identifier,
    STATE(4165), 1,
      sym_type_binding,
    STATE(4225), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5367), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(7077), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [178589] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1346), 1,
      aux_sym_number_token1,
    ACTIONS(1348), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1350), 1,
      anon_sym_SQUOTE,
    ACTIONS(1352), 1,
      anon_sym_DQUOTE,
    ACTIONS(6589), 1,
      anon_sym_LBRACE,
    STATE(2047), 1,
      sym__signed_constant,
    STATE(4226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6587), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1857), 2,
      sym__constant,
      sym_signed_number,
    STATE(1861), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [178633] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1076), 1,
      aux_sym_number_token1,
    ACTIONS(1080), 1,
      anon_sym_SQUOTE,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(2252), 1,
      aux_sym_signed_number_token1,
    ACTIONS(6593), 1,
      anon_sym_LBRACE,
    STATE(4227), 1,
      sym_attribute,
    STATE(4330), 1,
      sym__signed_constant,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6591), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4253), 2,
      sym__constant,
      sym_signed_number,
    STATE(4109), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [178677] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6595), 1,
      anon_sym_DQUOTE,
    STATE(4205), 1,
      aux_sym_string_content_repeat1,
    STATE(4228), 1,
      sym_attribute,
    STATE(4279), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6497), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6499), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6495), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [178715] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6597), 15,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [178743] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4060), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [178780] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4070), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [178817] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1854), 1,
      anon_sym_COLON2,
    STATE(4232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1856), 13,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [178846] = 14,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    ACTIONS(6599), 1,
      anon_sym_private,
    ACTIONS(6601), 1,
      anon_sym_DOT_DOT,
    ACTIONS(6603), 1,
      anon_sym_LBRACE,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(4233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3654), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [178893] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    STATE(4234), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5423), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6203), 1,
      sym_type_constructor_path,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [178942] = 14,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6607), 1,
      anon_sym_private,
    ACTIONS(6609), 1,
      anon_sym_DOT_DOT,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    STATE(4235), 1,
      sym_attribute,
    STATE(4938), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4627), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [178989] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4236), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    STATE(5166), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4135), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179032] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4237), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    STATE(5056), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4167), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179075] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4238), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    STATE(5058), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4167), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179118] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1754), 1,
      anon_sym_COLON2,
    STATE(4239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1756), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [179147] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    STATE(4240), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5351), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6244), 1,
      sym_type_constructor_path,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [179196] = 14,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(6603), 1,
      anon_sym_LBRACE,
    ACTIONS(6619), 1,
      anon_sym_private,
    ACTIONS(6621), 1,
      anon_sym_DOT_DOT,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(4241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3501), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [179243] = 14,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(6603), 1,
      anon_sym_LBRACE,
    ACTIONS(6623), 1,
      anon_sym_private,
    ACTIONS(6625), 1,
      anon_sym_DOT_DOT,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(4242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3486), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [179290] = 14,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    ACTIONS(6627), 1,
      anon_sym_private,
    ACTIONS(6629), 1,
      anon_sym_DOT_DOT,
    STATE(4243), 1,
      sym_attribute,
    STATE(4938), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [179337] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(6633), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6631), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179368] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6637), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6635), 13,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179397] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4246), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    STATE(5219), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4173), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179440] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6641), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4245), 1,
      sym__coerced,
    STATE(4247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6639), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179473] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6645), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6643), 13,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179502] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4249), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    STATE(5238), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4173), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179545] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2294), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2296), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4250), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    STATE(5261), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4177), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179588] = 15,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(6405), 1,
      anon_sym_LPAREN,
    ACTIONS(6411), 1,
      anon_sym__,
    STATE(4251), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5464), 1,
      sym__type_params,
    STATE(6049), 1,
      sym__type_param,
    STATE(6238), 1,
      sym_type_constructor_path,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [179637] = 14,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACE,
    ACTIONS(6647), 1,
      anon_sym_private,
    ACTIONS(6649), 1,
      anon_sym_DOT_DOT,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(4252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3594), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [179684] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2970), 1,
      anon_sym_COLON2,
    STATE(4253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2972), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [179712] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4005), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [179738] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4019), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [179764] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6651), 1,
      anon_sym_LPAREN,
    ACTIONS(6653), 1,
      anon_sym_LBRACE,
    ACTIONS(6655), 1,
      anon_sym_begin,
    ACTIONS(6657), 1,
      sym__uppercase_identifier,
    STATE(4256), 1,
      sym_attribute,
    STATE(4439), 1,
      sym__delimited_pattern,
    STATE(4465), 1,
      sym__constructor_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [179806] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6655), 1,
      anon_sym_begin,
    ACTIONS(6657), 1,
      sym__uppercase_identifier,
    ACTIONS(6659), 1,
      anon_sym_LPAREN,
    ACTIONS(6661), 1,
      anon_sym_LBRACE,
    STATE(4257), 1,
      sym_attribute,
    STATE(4465), 1,
      sym__constructor_name,
    STATE(4602), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [179848] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2276), 1,
      anon_sym_LBRACK,
    ACTIONS(2286), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6663), 1,
      anon_sym_LPAREN,
    ACTIONS(6665), 1,
      anon_sym_LBRACE,
    ACTIONS(6667), 1,
      anon_sym_begin,
    ACTIONS(6669), 1,
      sym__uppercase_identifier,
    STATE(4258), 1,
      sym_attribute,
    STATE(4902), 1,
      sym__delimited_pattern,
    STATE(4999), 1,
      sym__constructor_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4959), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [179890] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2314), 1,
      anon_sym_LBRACK,
    ACTIONS(2320), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6667), 1,
      anon_sym_begin,
    ACTIONS(6669), 1,
      sym__uppercase_identifier,
    ACTIONS(6671), 1,
      anon_sym_LPAREN,
    ACTIONS(6673), 1,
      anon_sym_LBRACE,
    STATE(4259), 1,
      sym_attribute,
    STATE(4927), 1,
      sym__delimited_binding_pattern,
    STATE(4999), 1,
      sym__constructor_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4969), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [179932] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4391), 1,
      anon_sym_in,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4260), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    STATE(5329), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179974] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3646), 1,
      anon_sym_as,
    ACTIONS(6675), 1,
      sym__lowercase_identifier,
    ACTIONS(6677), 1,
      anon_sym_POUND,
    STATE(4261), 1,
      sym_attribute,
    STATE(4445), 1,
      sym_type_constructor_path,
    STATE(6168), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3642), 7,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
  [180012] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6653), 1,
      anon_sym_LBRACE,
    ACTIONS(6655), 1,
      anon_sym_begin,
    ACTIONS(6679), 1,
      anon_sym_LPAREN,
    ACTIONS(6681), 1,
      sym__uppercase_identifier,
    STATE(1619), 1,
      sym__constructor_name,
    STATE(4262), 1,
      sym_attribute,
    STATE(4439), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [180054] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6655), 1,
      anon_sym_begin,
    ACTIONS(6661), 1,
      anon_sym_LBRACE,
    ACTIONS(6681), 1,
      sym__uppercase_identifier,
    ACTIONS(6683), 1,
      anon_sym_LPAREN,
    STATE(1619), 1,
      sym__constructor_name,
    STATE(4263), 1,
      sym_attribute,
    STATE(4602), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [180096] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(6687), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4264), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6685), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180126] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3654), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [180152] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3983), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [180178] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    ACTIONS(6689), 1,
      anon_sym_DOT_DOT,
    STATE(4267), 1,
      sym_attribute,
    STATE(4938), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4638), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [180222] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3979), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [180248] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(6603), 1,
      anon_sym_LBRACE,
    ACTIONS(6691), 1,
      anon_sym_DOT_DOT,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(4269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3489), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [180292] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACE,
    ACTIONS(6693), 1,
      anon_sym_DOT_DOT,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(4270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3648), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [180336] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4950), 1,
      anon_sym_as,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4271), 1,
      sym_attribute,
    STATE(4337), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6695), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180370] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6697), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4272), 1,
      sym_attribute,
    STATE(4339), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6699), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180404] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(6603), 1,
      anon_sym_LBRACE,
    ACTIONS(6701), 1,
      anon_sym_DOT_DOT,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(4273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3485), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [180448] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2043), 1,
      anon_sym_LBRACK,
    ACTIONS(2051), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6703), 1,
      anon_sym_LPAREN,
    ACTIONS(6705), 1,
      anon_sym_LBRACE,
    ACTIONS(6707), 1,
      anon_sym_begin,
    ACTIONS(6709), 1,
      sym__uppercase_identifier,
    STATE(2492), 1,
      sym__constructor_name,
    STATE(2513), 1,
      sym__delimited_pattern,
    STATE(4274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2478), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [180490] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6711), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4275), 1,
      sym_attribute,
    STATE(4348), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6713), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180524] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1500), 1,
      anon_sym_LBRACK,
    ACTIONS(1514), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6681), 1,
      sym__uppercase_identifier,
    ACTIONS(6715), 1,
      anon_sym_LPAREN,
    ACTIONS(6717), 1,
      anon_sym_LBRACE,
    ACTIONS(6719), 1,
      anon_sym_begin,
    STATE(1619), 1,
      sym__constructor_name,
    STATE(2154), 1,
      sym__delimited_pattern,
    STATE(4276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2250), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [180566] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6721), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4277), 1,
      sym_attribute,
    STATE(4356), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6723), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180600] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACE,
    ACTIONS(6725), 1,
      anon_sym_DOT_DOT,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(4278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3678), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [180644] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4279), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6729), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6727), 11,
      anon_sym_DQUOTE,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [180672] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4280), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4001), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [180698] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2974), 1,
      anon_sym_COLON2,
    STATE(4281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2976), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [180726] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3742), 1,
      anon_sym_as,
    ACTIONS(6675), 1,
      sym__lowercase_identifier,
    ACTIONS(6677), 1,
      anon_sym_POUND,
    ACTIONS(6731), 1,
      anon_sym_STAR,
    STATE(4282), 1,
      sym_attribute,
    STATE(4445), 1,
      sym_type_constructor_path,
    STATE(6168), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3740), 6,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
  [180766] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3997), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [180792] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1324), 1,
      anon_sym_LBRACK,
    ACTIONS(1334), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6733), 1,
      anon_sym_LPAREN,
    ACTIONS(6735), 1,
      anon_sym_LBRACE,
    ACTIONS(6737), 1,
      anon_sym_begin,
    ACTIONS(6739), 1,
      sym__uppercase_identifier,
    STATE(1798), 1,
      sym__constructor_name,
    STATE(2048), 1,
      sym__delimited_binding_pattern,
    STATE(4284), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2080), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [180834] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3656), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3654), 11,
      anon_sym_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
  [180862] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6741), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4286), 1,
      sym_attribute,
    STATE(4370), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6743), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180896] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6745), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4287), 1,
      sym_attribute,
    STATE(4376), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6747), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180930] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6749), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4288), 1,
      sym_attribute,
    STATE(4378), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6751), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180964] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    ACTIONS(6753), 1,
      anon_sym_DOT_DOT,
    STATE(4289), 1,
      sym_attribute,
    STATE(4938), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4673), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [181008] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6103), 1,
      sym__uppercase_identifier,
    ACTIONS(6653), 1,
      anon_sym_LBRACE,
    ACTIONS(6655), 1,
      anon_sym_begin,
    ACTIONS(6755), 1,
      anon_sym_LPAREN,
    STATE(1761), 1,
      sym__constructor_name,
    STATE(4290), 1,
      sym_attribute,
    STATE(4439), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [181050] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6103), 1,
      sym__uppercase_identifier,
    ACTIONS(6655), 1,
      anon_sym_begin,
    ACTIONS(6661), 1,
      anon_sym_LBRACE,
    ACTIONS(6757), 1,
      anon_sym_LPAREN,
    STATE(1761), 1,
      sym__constructor_name,
    STATE(4291), 1,
      sym_attribute,
    STATE(4602), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181092] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2276), 1,
      anon_sym_LBRACK,
    ACTIONS(2286), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6665), 1,
      anon_sym_LBRACE,
    ACTIONS(6667), 1,
      anon_sym_begin,
    ACTIONS(6759), 1,
      anon_sym_LPAREN,
    ACTIONS(6761), 1,
      sym__uppercase_identifier,
    STATE(3269), 1,
      sym__constructor_name,
    STATE(4292), 1,
      sym_attribute,
    STATE(4902), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4959), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [181134] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2314), 1,
      anon_sym_LBRACK,
    ACTIONS(2320), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6667), 1,
      anon_sym_begin,
    ACTIONS(6673), 1,
      anon_sym_LBRACE,
    ACTIONS(6761), 1,
      sym__uppercase_identifier,
    ACTIONS(6763), 1,
      anon_sym_LPAREN,
    STATE(3269), 1,
      sym__constructor_name,
    STATE(4293), 1,
      sym_attribute,
    STATE(4927), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4969), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181176] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6765), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4294), 1,
      sym_attribute,
    STATE(4399), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6767), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [181210] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2145), 1,
      anon_sym_COLON2,
    ACTIONS(6769), 1,
      anon_sym_DOT_DOT,
    STATE(4295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [181240] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6773), 1,
      anon_sym_as,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4296), 1,
      sym_attribute,
    STATE(4419), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6771), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [181274] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4297), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4060), 4,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [181310] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6775), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4298), 1,
      sym_attribute,
    STATE(4425), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6777), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [181344] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6653), 1,
      anon_sym_LBRACE,
    ACTIONS(6655), 1,
      anon_sym_begin,
    ACTIONS(6755), 1,
      anon_sym_LPAREN,
    ACTIONS(6779), 1,
      sym__uppercase_identifier,
    STATE(3076), 1,
      sym__constructor_name,
    STATE(4299), 1,
      sym_attribute,
    STATE(4439), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [181386] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6783), 1,
      anon_sym_as,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4300), 1,
      sym_attribute,
    STATE(4426), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6781), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [181420] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1116), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6653), 1,
      anon_sym_LBRACE,
    ACTIONS(6655), 1,
      anon_sym_begin,
    ACTIONS(6785), 1,
      anon_sym_LPAREN,
    ACTIONS(6787), 1,
      sym__uppercase_identifier,
    STATE(3222), 1,
      sym__constructor_name,
    STATE(4301), 1,
      sym_attribute,
    STATE(4439), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4431), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [181462] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4349), 1,
      anon_sym_in,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4302), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    STATE(5502), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [181504] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6655), 1,
      anon_sym_begin,
    ACTIONS(6661), 1,
      anon_sym_LBRACE,
    ACTIONS(6787), 1,
      sym__uppercase_identifier,
    ACTIONS(6789), 1,
      anon_sym_LPAREN,
    STATE(3222), 1,
      sym__constructor_name,
    STATE(4303), 1,
      sym_attribute,
    STATE(4602), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181546] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6793), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6791), 11,
      anon_sym_DQUOTE,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [181574] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4305), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4070), 4,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [181610] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6795), 1,
      anon_sym_EQ,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4306), 1,
      sym_attribute,
    STATE(4318), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6797), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [181644] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1050), 1,
      anon_sym_LBRACK,
    ACTIONS(1064), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6655), 1,
      anon_sym_begin,
    ACTIONS(6661), 1,
      anon_sym_LBRACE,
    ACTIONS(6757), 1,
      anon_sym_LPAREN,
    ACTIONS(6779), 1,
      sym__uppercase_identifier,
    STATE(3076), 1,
      sym__constructor_name,
    STATE(4307), 1,
      sym_attribute,
    STATE(4602), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4605), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181686] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3718), 1,
      anon_sym_as,
    ACTIONS(6675), 1,
      sym__lowercase_identifier,
    ACTIONS(6677), 1,
      anon_sym_POUND,
    ACTIONS(6799), 1,
      anon_sym_DASH_GT,
    ACTIONS(6801), 1,
      anon_sym_STAR,
    STATE(4308), 1,
      sym_attribute,
    STATE(4445), 1,
      sym_type_constructor_path,
    STATE(6168), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 5,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
  [181728] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3660), 1,
      anon_sym_as,
    ACTIONS(6675), 1,
      sym__lowercase_identifier,
    ACTIONS(6803), 1,
      anon_sym_POUND,
    STATE(4309), 1,
      sym_attribute,
    STATE(4451), 1,
      sym_type_constructor_path,
    STATE(6168), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3658), 7,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
  [181766] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2248), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2250), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4353), 1,
      anon_sym_in,
    ACTIONS(6512), 1,
      anon_sym_LPAREN,
    STATE(4310), 1,
      sym_attribute,
    STATE(4536), 1,
      sym__simple_module_expression,
    STATE(5401), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4529), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [181808] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [181834] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1434), 1,
      anon_sym_LBRACK,
    ACTIONS(1450), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6805), 1,
      anon_sym_LPAREN,
    ACTIONS(6807), 1,
      anon_sym_LBRACE,
    ACTIONS(6809), 1,
      anon_sym_begin,
    ACTIONS(6811), 1,
      sym__uppercase_identifier,
    STATE(2842), 1,
      sym__constructor_name,
    STATE(3060), 1,
      sym__delimited_binding_pattern,
    STATE(4312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2972), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181876] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3698), 1,
      anon_sym_COLON2,
    STATE(4313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3106), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3108), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [181905] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2400), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [181936] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3028), 1,
      anon_sym_COLON2,
    STATE(4315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3030), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [181963] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4316), 1,
      sym_attribute,
    STATE(4353), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6813), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [181994] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6817), 1,
      anon_sym_COLON2,
    STATE(4317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6815), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182021] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6819), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182052] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3395), 1,
      anon_sym_COLON2,
    STATE(4319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3393), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182079] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6821), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182110] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6817), 1,
      anon_sym_COLON2,
    STATE(4321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6815), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182137] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6825), 1,
      anon_sym_COLON2,
    STATE(4322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6823), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182164] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6829), 1,
      anon_sym_COLON2,
    STATE(4323), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6827), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182191] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3475), 1,
      anon_sym_COLON2,
    STATE(4324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3473), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182218] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6831), 1,
      anon_sym_RPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4325), 1,
      sym_attribute,
    STATE(7302), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182257] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3004), 1,
      anon_sym_COLON2,
    STATE(4326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3006), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182284] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6837), 1,
      anon_sym_COLON2,
    STATE(4327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6835), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182311] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6839), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4328), 1,
      sym_attribute,
    STATE(6373), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182350] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6841), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4329), 1,
      sym_attribute,
    STATE(6431), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182389] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3082), 1,
      anon_sym_COLON2,
    STATE(4330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3084), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182416] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6843), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4331), 1,
      sym_attribute,
    STATE(6932), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182455] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6845), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4332), 1,
      sym_attribute,
    STATE(6436), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182494] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6849), 1,
      anon_sym_COLON2,
    STATE(4333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6847), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182521] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3403), 1,
      anon_sym_COLON2,
    STATE(4334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3401), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182548] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6851), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4335), 1,
      sym_attribute,
    STATE(7037), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182587] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6855), 1,
      anon_sym_COLON2,
    STATE(4336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6853), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182614] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6857), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182645] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4338), 1,
      sym_attribute,
    STATE(4367), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6859), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182676] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6861), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182707] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3062), 1,
      anon_sym_COLON2,
    STATE(4340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3064), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182734] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6863), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4341), 1,
      sym_attribute,
    STATE(6646), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182773] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4342), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6865), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182804] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6867), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182835] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4344), 1,
      sym_attribute,
    STATE(4371), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6869), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182866] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4345), 1,
      sym_attribute,
    STATE(4372), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6871), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182897] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6873), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182928] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4347), 1,
      sym_attribute,
    STATE(4373), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6875), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182959] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6877), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182990] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6879), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4349), 1,
      sym_attribute,
    STATE(6693), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183029] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6881), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183060] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3644), 1,
      sym__lowercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(4365), 1,
      anon_sym_STAR,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(4351), 1,
      sym_attribute,
    STATE(4731), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4363), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4439), 3,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
  [183101] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2394), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183132] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4353), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6883), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183163] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4354), 1,
      sym_attribute,
    STATE(4382), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6885), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183194] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4355), 1,
      sym_attribute,
    STATE(4383), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6887), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183225] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6889), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183256] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6891), 1,
      anon_sym_COLON2,
    STATE(4357), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3106), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3108), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [183285] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6895), 1,
      anon_sym_COLON2,
    STATE(4358), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6893), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183312] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6899), 1,
      anon_sym_COLON2,
    STATE(4359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6897), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183339] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6901), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4360), 1,
      sym_attribute,
    STATE(6848), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183378] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3431), 1,
      anon_sym_COLON2,
    STATE(4361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3429), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183405] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6903), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4362), 1,
      sym_attribute,
    STATE(6535), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183444] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6905), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4363), 1,
      sym_attribute,
    STATE(6457), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183483] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3554), 1,
      anon_sym_COLON2,
    STATE(4364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183510] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6907), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4365), 1,
      sym_attribute,
    STATE(6474), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183549] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    STATE(4366), 1,
      sym_attribute,
    STATE(4955), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    ACTIONS(4461), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [183586] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4367), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6909), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183617] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4368), 1,
      sym_attribute,
    STATE(4389), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6911), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183648] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4369), 1,
      sym_attribute,
    STATE(4390), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6913), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183679] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4370), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6915), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183710] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4371), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6917), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183741] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4372), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6919), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183772] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4373), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6921), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183803] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4374), 1,
      sym_attribute,
    STATE(4392), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6923), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183834] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4375), 1,
      sym_attribute,
    STATE(4393), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6925), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183865] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4376), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6927), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183896] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4377), 1,
      sym_attribute,
    STATE(4395), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6929), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183927] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4378), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6931), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183958] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6935), 1,
      anon_sym_COLON2,
    STATE(4379), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6933), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183985] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6939), 1,
      anon_sym_COLON2,
    STATE(4380), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6937), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184012] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6941), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184043] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4382), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6943), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184074] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4383), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6945), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184105] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4384), 1,
      sym_attribute,
    STATE(4400), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6947), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184136] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3218), 1,
      anon_sym_COLON2,
    STATE(4385), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3220), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184163] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3493), 1,
      anon_sym_COLON2,
    STATE(4386), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3491), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184190] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6951), 1,
      anon_sym_COLON2,
    STATE(4387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6949), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184217] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6955), 1,
      anon_sym_COLON2,
    STATE(4388), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6953), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184244] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4389), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6957), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184275] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6959), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184306] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4391), 1,
      sym_attribute,
    STATE(4404), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6961), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184337] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4392), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6963), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184368] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4393), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6965), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184399] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4394), 1,
      sym_attribute,
    STATE(4405), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6967), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184430] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4395), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6969), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184461] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4396), 1,
      sym_attribute,
    STATE(4406), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6971), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184492] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4397), 1,
      sym_attribute,
    STATE(4407), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6973), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184523] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6977), 1,
      anon_sym_COLON2,
    STATE(4398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6975), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184550] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6979), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184581] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4400), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6981), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184612] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6983), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4401), 1,
      sym_attribute,
    STATE(6679), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [184651] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(6985), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4402), 1,
      sym_attribute,
    STATE(6959), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [184690] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4403), 1,
      sym_attribute,
    STATE(4429), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6987), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184721] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4404), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6989), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184752] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4405), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6991), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184783] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6993), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184814] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6995), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184845] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4408), 1,
      sym_attribute,
    STATE(4411), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6997), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184876] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(4409), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [184903] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7001), 1,
      anon_sym_COLON2,
    STATE(4410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6999), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184930] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7003), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184961] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6381), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184992] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(7005), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4413), 1,
      sym_attribute,
    STATE(6927), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185031] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(4414), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [185058] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7009), 1,
      anon_sym_LPAREN,
    ACTIONS(7011), 1,
      anon_sym_COLON2,
    STATE(4415), 1,
      sym_attribute,
    STATE(4427), 1,
      aux_sym_module_binding_repeat1,
    STATE(4733), 1,
      sym_module_parameter,
    STATE(4763), 1,
      sym__module_typed,
    STATE(5175), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4527), 2,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(7007), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [185101] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(7013), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4416), 1,
      sym_attribute,
    STATE(6604), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185140] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(7015), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4417), 1,
      sym_attribute,
    STATE(7264), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185179] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(7017), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4418), 1,
      sym_attribute,
    STATE(6678), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185218] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4419), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7019), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185249] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4420), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6393), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185280] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(7021), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4421), 1,
      sym_attribute,
    STATE(6743), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185319] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4343), 1,
      aux_sym_expression_item_repeat1,
    STATE(4422), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7023), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185350] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4423), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2372), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185381] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(7025), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4424), 1,
      sym_attribute,
    STATE(6766), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185420] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4425), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7027), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185451] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4426), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7029), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185482] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7009), 1,
      anon_sym_LPAREN,
    ACTIONS(7011), 1,
      anon_sym_COLON2,
    STATE(4427), 1,
      sym_attribute,
    STATE(4599), 1,
      aux_sym_module_binding_repeat1,
    STATE(4733), 1,
      sym_module_parameter,
    STATE(4739), 1,
      sym__module_typed,
    STATE(5042), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4517), 2,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(7031), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [185525] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4346), 1,
      aux_sym_expression_item_repeat1,
    STATE(4428), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7033), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185556] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4429), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7035), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185587] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2145), 1,
      anon_sym_COLON2,
    STATE(4430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185614] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3546), 1,
      anon_sym_COLON2,
    STATE(4431), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3544), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185641] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(7037), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4432), 1,
      sym_attribute,
    STATE(6839), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185680] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1072), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1074), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4064), 1,
      anon_sym_LPAREN,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(7039), 1,
      anon_sym_RPAREN,
    STATE(3028), 1,
      sym__simple_module_expression,
    STATE(4433), 1,
      sym_attribute,
    STATE(6569), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3052), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185719] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3427), 1,
      anon_sym_COLON2,
    STATE(4434), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3425), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185746] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3538), 1,
      anon_sym_COLON2,
    STATE(4435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3536), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185773] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3550), 1,
      anon_sym_COLON2,
    STATE(4436), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3548), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185800] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3347), 1,
      anon_sym_COLON2,
    STATE(4437), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3345), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185827] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3351), 1,
      anon_sym_COLON2,
    STATE(4438), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3349), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185854] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3383), 1,
      anon_sym_COLON2,
    STATE(4439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3381), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185881] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3024), 1,
      anon_sym_COLON2,
    STATE(4440), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3026), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185908] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7043), 1,
      anon_sym_COLON2,
    STATE(4441), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7041), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185935] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    ACTIONS(7047), 1,
      anon_sym_as,
    STATE(4442), 1,
      sym_attribute,
    STATE(4512), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4541), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [185967] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7051), 1,
      anon_sym_COLON,
    ACTIONS(7053), 1,
      anon_sym_COLON2,
    STATE(4443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7049), 9,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185995] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    ACTIONS(7047), 1,
      anon_sym_as,
    STATE(4444), 1,
      sym_attribute,
    STATE(4501), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4600), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [186027] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4445), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3758), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3756), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186053] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4446), 1,
      sym_attribute,
    STATE(4460), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7055), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186083] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7057), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186113] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4448), 1,
      sym_attribute,
    STATE(4462), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7059), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186143] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3806), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3804), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186169] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4447), 1,
      aux_sym_expression_item_repeat1,
    STATE(4450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7061), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186199] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4451), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3887), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3885), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186225] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4452), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3897), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(4017), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
    ACTIONS(3895), 7,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186253] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4453), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3891), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3889), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186279] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3929), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3927), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186305] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4455), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1832), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [186329] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3941), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3939), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186355] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3816), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3814), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186381] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3871), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3869), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186407] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3875), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3873), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186433] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7063), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186463] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4461), 1,
      sym_attribute,
    STATE(4473), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7065), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186493] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4462), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7067), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186523] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4463), 1,
      sym_attribute,
    STATE(4474), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7069), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186553] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4464), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3762), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3760), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186579] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1710), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [186603] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(7071), 1,
      anon_sym_private,
    STATE(3560), 1,
      sym_variant_declaration,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(4466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [186643] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3883), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3881), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186669] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    ACTIONS(7073), 1,
      anon_sym_private,
    STATE(3585), 1,
      sym_variant_declaration,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(4468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [186709] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    ACTIONS(7047), 1,
      anon_sym_as,
    STATE(4469), 1,
      sym_attribute,
    STATE(4517), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4549), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [186741] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4470), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3660), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3658), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186767] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3909), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3907), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186793] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3913), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3911), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186819] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7075), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186849] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2055), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3315), 1,
      aux_sym_expression_item_repeat1,
    STATE(3502), 1,
      sym_item_attribute,
    STATE(4474), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7077), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186879] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4475), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3915), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186905] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3921), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3919), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186931] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4477), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3933), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3931), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186957] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4478), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3945), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3943), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186983] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4479), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3949), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3947), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187009] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    ACTIONS(7079), 1,
      anon_sym_private,
    STATE(3560), 1,
      sym_variant_declaration,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(4480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187049] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1702), 11,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [187073] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4482), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3796), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3794), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187099] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4483), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3800), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3798), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187125] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3108), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    ACTIONS(7081), 1,
      anon_sym_EQ,
    ACTIONS(7083), 1,
      anon_sym_COLON_EQ,
    STATE(4484), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(4659), 1,
      aux_sym_type_binding_repeat1,
    STATE(5158), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4555), 2,
      anon_sym_and,
      anon_sym_in,
  [187167] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4485), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1744), 11,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [187191] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7085), 1,
      anon_sym_DOT,
    STATE(4486), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4161), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [187217] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    ACTIONS(7087), 1,
      anon_sym_private,
    STATE(4487), 1,
      sym_attribute,
    STATE(4938), 1,
      sym_constructor_declaration,
    STATE(5212), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187257] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3937), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3935), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187283] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4489), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3867), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3865), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187309] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4490), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3897), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3895), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187335] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(7089), 1,
      anon_sym_private,
    STATE(3585), 1,
      sym_variant_declaration,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(4491), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187375] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3925), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3923), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187401] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3108), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    ACTIONS(7091), 1,
      anon_sym_EQ,
    ACTIONS(7093), 1,
      anon_sym_COLON_EQ,
    STATE(4493), 1,
      sym_attribute,
    STATE(4654), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5095), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4592), 2,
      anon_sym_and,
      anon_sym_in,
  [187443] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    ACTIONS(7095), 1,
      anon_sym_private,
    STATE(4494), 1,
      sym_attribute,
    STATE(4938), 1,
      sym_constructor_declaration,
    STATE(5073), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187483] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4495), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3901), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3899), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187509] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4496), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3905), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3903), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187535] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4497), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3879), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3877), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187561] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    STATE(3554), 1,
      sym_variant_declaration,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(4498), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187598] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    ACTIONS(7047), 1,
      anon_sym_as,
    ACTIONS(7097), 1,
      anon_sym_EQ,
    STATE(4499), 1,
      sym_attribute,
    STATE(4624), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5104), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4674), 2,
      anon_sym_and,
      anon_sym_in,
  [187637] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4303), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [187660] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4501), 1,
      sym_attribute,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4617), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [187689] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    STATE(3598), 1,
      sym_variant_declaration,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(4502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187726] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6817), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    STATE(4503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6815), 7,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_when,
  [187755] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6817), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    STATE(4504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6815), 8,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [187782] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7103), 1,
      anon_sym_EQ,
    ACTIONS(7105), 1,
      anon_sym_RPAREN,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7109), 1,
      anon_sym_COLON2,
    ACTIONS(7111), 1,
      anon_sym_as,
    STATE(4505), 1,
      sym_attribute,
    STATE(6495), 1,
      sym__typed,
    STATE(6793), 1,
      sym__polymorphic_typed,
    STATE(6925), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [187823] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6825), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    STATE(4506), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6823), 6,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_when,
  [187854] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    STATE(4507), 1,
      sym_attribute,
    STATE(4652), 1,
      sym__attribute,
    STATE(5466), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187891] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    ACTIONS(7047), 1,
      anon_sym_as,
    ACTIONS(7113), 1,
      anon_sym_EQ,
    STATE(4508), 1,
      sym_attribute,
    STATE(4639), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5142), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4697), 2,
      anon_sym_and,
      anon_sym_in,
  [187930] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4509), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4203), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [187953] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7117), 1,
      anon_sym_COLON2,
    STATE(4510), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7115), 8,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_when,
  [187980] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    STATE(4511), 1,
      sym_attribute,
    STATE(4938), 1,
      sym_constructor_declaration,
    STATE(5262), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188017] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4512), 1,
      sym_attribute,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4639), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [188046] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(7119), 1,
      sym__lowercase_identifier,
    ACTIONS(7121), 1,
      anon_sym_LPAREN,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    STATE(2635), 1,
      sym__delimited_type,
    STATE(4513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [188083] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4207), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188106] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3303), 1,
      anon_sym_PIPE,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    STATE(4515), 1,
      sym_attribute,
    STATE(4938), 1,
      sym_constructor_declaration,
    STATE(5214), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188143] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4323), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188166] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4517), 1,
      sym_attribute,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4643), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [188195] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1546), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(1548), 8,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_COLON_GT,
      sym_hash_operator,
  [188220] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4519), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4327), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188243] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4520), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4319), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188266] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1578), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(1580), 8,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_COLON_GT,
      sym_hash_operator,
  [188291] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7125), 1,
      anon_sym_constraint,
    STATE(4657), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4522), 2,
      sym_attribute,
      aux_sym_type_binding_repeat1,
    ACTIONS(4667), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [188318] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4523), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4219), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188341] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(3774), 1,
      sym_constructor_declaration,
    STATE(4524), 1,
      sym_attribute,
    STATE(4653), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188378] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3323), 1,
      anon_sym_PIPE,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    STATE(3554), 1,
      sym_variant_declaration,
    STATE(3890), 1,
      sym_constructor_declaration,
    STATE(4525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188415] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7130), 1,
      anon_sym_COLON2,
    STATE(4526), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7128), 5,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_when,
  [188448] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7134), 1,
      anon_sym_COLON2,
    STATE(4527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7132), 5,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_when,
  [188481] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4528), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4161), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188504] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4255), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188527] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3684), 1,
      anon_sym_LBRACK,
    ACTIONS(3686), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3688), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(7136), 1,
      sym__lowercase_identifier,
    ACTIONS(7138), 1,
      anon_sym_LPAREN,
    STATE(4496), 1,
      sym__delimited_type,
    STATE(4530), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4492), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [188564] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4531), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(1700), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [188589] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3238), 1,
      anon_sym_PIPE,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    STATE(3598), 1,
      sym_variant_declaration,
    STATE(3724), 1,
      sym_constructor_declaration,
    STATE(4532), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188626] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4533), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4299), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188649] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(3774), 1,
      sym_constructor_declaration,
    STATE(4534), 1,
      sym_attribute,
    STATE(4637), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4059), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188686] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4535), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4237), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188709] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4536), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4241), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188732] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6581), 1,
      anon_sym_STAR,
    STATE(4537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3740), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [188757] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(7121), 1,
      anon_sym_LPAREN,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(7140), 1,
      sym__lowercase_identifier,
    STATE(2635), 1,
      sym__delimited_type,
    STATE(4538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [188794] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1318), 1,
      anon_sym_COLON2,
    ACTIONS(6769), 1,
      anon_sym_DOT_DOT,
    STATE(4539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1316), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [188821] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2448), 1,
      anon_sym_LBRACK,
    ACTIONS(2464), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(2466), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(7121), 1,
      anon_sym_LPAREN,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(7142), 1,
      sym__lowercase_identifier,
    STATE(2635), 1,
      sym__delimited_type,
    STATE(4540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2640), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [188858] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3246), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3248), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3714), 1,
      anon_sym_LBRACK,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(7144), 1,
      sym__lowercase_identifier,
    ACTIONS(7146), 1,
      anon_sym_LPAREN,
    STATE(3112), 1,
      sym__delimited_type,
    STATE(4541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3011), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [188895] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(3774), 1,
      sym_constructor_declaration,
    STATE(4542), 1,
      sym_attribute,
    STATE(4661), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188932] = 13,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7109), 1,
      anon_sym_COLON2,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7148), 1,
      anon_sym_EQ,
    ACTIONS(7150), 1,
      anon_sym_RPAREN,
    STATE(4543), 1,
      sym_attribute,
    STATE(6434), 1,
      sym__maybe_polymorphic_typed,
    STATE(6435), 1,
      sym__polymorphic_typed,
    STATE(6881), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [188973] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1578), 1,
      anon_sym_COLON2,
    ACTIONS(1614), 1,
      anon_sym_POUND,
    STATE(4544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1616), 3,
      anon_sym_DOT,
      anon_sym_with,
      sym_hash_operator,
    ACTIONS(1580), 4,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
  [189001] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7152), 1,
      sym__lowercase_identifier,
    ACTIONS(7154), 1,
      anon_sym_LPAREN,
    ACTIONS(7156), 1,
      anon_sym_LBRACK,
    ACTIONS(7158), 1,
      anon_sym_LBRACE,
    ACTIONS(7160), 1,
      sym_indexing_operator,
    STATE(866), 1,
      sym_field_path,
    STATE(4545), 1,
      sym_attribute,
    STATE(5876), 1,
      sym_indexing_operator_path,
    STATE(6687), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [189039] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7160), 1,
      sym_indexing_operator,
    ACTIONS(7162), 1,
      sym__lowercase_identifier,
    ACTIONS(7164), 1,
      anon_sym_LPAREN,
    ACTIONS(7166), 1,
      anon_sym_LBRACK,
    ACTIONS(7168), 1,
      anon_sym_LBRACE,
    STATE(969), 1,
      sym_field_path,
    STATE(4546), 1,
      sym_attribute,
    STATE(5866), 1,
      sym_indexing_operator_path,
    STATE(6610), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [189077] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4547), 1,
      sym_attribute,
    STATE(6844), 1,
      sym_escape_sequence,
    STATE(7259), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189107] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4548), 1,
      sym_attribute,
    STATE(6360), 1,
      sym_character_content,
    STATE(6844), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189137] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3198), 1,
      anon_sym_COLON2,
    STATE(4549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3200), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189161] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7160), 1,
      sym_indexing_operator,
    ACTIONS(7176), 1,
      sym__lowercase_identifier,
    ACTIONS(7178), 1,
      anon_sym_LPAREN,
    ACTIONS(7180), 1,
      anon_sym_LBRACK,
    ACTIONS(7182), 1,
      anon_sym_LBRACE,
    STATE(1796), 1,
      sym_field_path,
    STATE(4550), 1,
      sym_attribute,
    STATE(5845), 1,
      sym_indexing_operator_path,
    STATE(6437), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [189199] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    ACTIONS(7047), 1,
      anon_sym_as,
    STATE(4551), 1,
      sym_attribute,
    STATE(4622), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5211), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4837), 2,
      anon_sym_and,
      anon_sym_in,
  [189235] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3202), 1,
      anon_sym_COLON2,
    STATE(4552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3204), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189259] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3146), 1,
      anon_sym_COLON2,
    STATE(4553), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3148), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189283] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    ACTIONS(7047), 1,
      anon_sym_as,
    STATE(4554), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(4666), 1,
      aux_sym_type_binding_repeat1,
    STATE(5225), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4759), 2,
      anon_sym_and,
      anon_sym_in,
  [189319] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4555), 1,
      sym_attribute,
    STATE(6570), 1,
      sym_character_content,
    STATE(6844), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189349] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7160), 1,
      sym_indexing_operator,
    ACTIONS(7184), 1,
      sym__lowercase_identifier,
    ACTIONS(7186), 1,
      anon_sym_LPAREN,
    ACTIONS(7188), 1,
      anon_sym_LBRACK,
    ACTIONS(7190), 1,
      anon_sym_LBRACE,
    STATE(775), 1,
      sym_field_path,
    STATE(4556), 1,
      sym_attribute,
    STATE(5809), 1,
      sym_indexing_operator_path,
    STATE(6586), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [189387] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3194), 1,
      anon_sym_COLON2,
    STATE(4557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3196), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189411] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3162), 1,
      anon_sym_COLON2,
    STATE(4558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3164), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189435] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3032), 1,
      anon_sym_COLON2,
    STATE(4559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3034), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189459] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3158), 1,
      anon_sym_COLON2,
    STATE(4560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189483] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3110), 1,
      anon_sym_COLON2,
    STATE(4561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3112), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189507] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7160), 1,
      sym_indexing_operator,
    ACTIONS(7192), 1,
      sym__lowercase_identifier,
    ACTIONS(7194), 1,
      anon_sym_LPAREN,
    ACTIONS(7196), 1,
      anon_sym_LBRACK,
    ACTIONS(7198), 1,
      anon_sym_LBRACE,
    STATE(1957), 1,
      sym_field_path,
    STATE(4562), 1,
      sym_attribute,
    STATE(5883), 1,
      sym_indexing_operator_path,
    STATE(6747), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [189545] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4563), 1,
      sym_attribute,
    STATE(6804), 1,
      sym_character_content,
    STATE(6844), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189575] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3166), 1,
      anon_sym_COLON2,
    STATE(4564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3168), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189599] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4565), 1,
      sym_attribute,
    STATE(6844), 1,
      sym_escape_sequence,
    STATE(6893), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189629] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7204), 1,
      sym__right_quoted_string_delimiter,
    STATE(4566), 1,
      sym_attribute,
    STATE(4665), 1,
      aux_sym_quoted_string_content_repeat1,
    STATE(6869), 1,
      sym_quoted_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7202), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7200), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [189659] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4567), 1,
      sym_attribute,
    STATE(6844), 1,
      sym_escape_sequence,
    STATE(6949), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189689] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4568), 1,
      sym_attribute,
    STATE(6844), 1,
      sym_escape_sequence,
    STATE(6995), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189719] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4569), 1,
      sym_attribute,
    STATE(6844), 1,
      sym_escape_sequence,
    STATE(7036), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189749] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4570), 1,
      sym_attribute,
    STATE(6844), 1,
      sym_escape_sequence,
    STATE(7082), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189779] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4571), 1,
      sym_attribute,
    STATE(6844), 1,
      sym_escape_sequence,
    STATE(7145), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189809] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4572), 1,
      sym_attribute,
    STATE(6844), 1,
      sym_escape_sequence,
    STATE(6845), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189839] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1546), 1,
      anon_sym_COLON2,
    ACTIONS(1614), 1,
      anon_sym_POUND,
    STATE(4573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1616), 3,
      anon_sym_DOT,
      anon_sym_with,
      sym_hash_operator,
    ACTIONS(1548), 4,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
  [189867] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4497), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [189889] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7160), 1,
      sym_indexing_operator,
    ACTIONS(7206), 1,
      sym__lowercase_identifier,
    ACTIONS(7208), 1,
      anon_sym_LPAREN,
    ACTIONS(7210), 1,
      anon_sym_LBRACK,
    ACTIONS(7212), 1,
      anon_sym_LBRACE,
    STATE(1062), 1,
      sym_field_path,
    STATE(4575), 1,
      sym_attribute,
    STATE(5906), 1,
      sym_indexing_operator_path,
    STATE(7238), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [189927] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4576), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4483), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [189949] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7160), 1,
      sym_indexing_operator,
    ACTIONS(7178), 1,
      anon_sym_LPAREN,
    ACTIONS(7180), 1,
      anon_sym_LBRACK,
    ACTIONS(7182), 1,
      anon_sym_LBRACE,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(1796), 1,
      sym_field_path,
    STATE(4577), 1,
      sym_attribute,
    STATE(5845), 1,
      sym_indexing_operator_path,
    STATE(7329), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [189987] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4763), 1,
      anon_sym_in,
    ACTIONS(6579), 1,
      anon_sym_LBRACE,
    STATE(4578), 1,
      sym_attribute,
    STATE(4791), 1,
      aux_sym_external_repeat1,
    STATE(5388), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5538), 2,
      sym_string,
      sym_quoted_string,
  [190023] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3114), 1,
      anon_sym_COLON2,
    STATE(4579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190047] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [190069] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3182), 1,
      anon_sym_COLON2,
    STATE(4581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3184), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190093] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3210), 1,
      anon_sym_COLON2,
    STATE(4582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3212), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190117] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3008), 1,
      anon_sym_COLON2,
    STATE(4583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3010), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190141] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4584), 1,
      sym_attribute,
    STATE(6477), 1,
      sym_character_content,
    STATE(6844), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190171] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4533), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [190195] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3138), 1,
      anon_sym_COLON2,
    STATE(4586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190219] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4587), 1,
      sym_attribute,
    STATE(6385), 1,
      sym_character_content,
    STATE(6844), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190249] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7216), 1,
      anon_sym_LPAREN,
    ACTIONS(7218), 1,
      anon_sym_DOT,
    STATE(4588), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4404), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [190275] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3090), 1,
      anon_sym_COLON2,
    STATE(4589), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3092), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190299] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7047), 1,
      anon_sym_as,
    STATE(4590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4509), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [190323] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3138), 1,
      anon_sym_COLON2,
    STATE(4591), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190347] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6554), 1,
      anon_sym_DASH_GT,
    STATE(4592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [190371] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3142), 1,
      anon_sym_COLON2,
    STATE(4593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3144), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190395] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7160), 1,
      sym_indexing_operator,
    ACTIONS(7220), 1,
      sym__lowercase_identifier,
    ACTIONS(7222), 1,
      anon_sym_LPAREN,
    ACTIONS(7224), 1,
      anon_sym_LBRACK,
    ACTIONS(7226), 1,
      anon_sym_LBRACE,
    STATE(1541), 1,
      sym_field_path,
    STATE(4594), 1,
      sym_attribute,
    STATE(5872), 1,
      sym_indexing_operator_path,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [190433] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7160), 1,
      sym_indexing_operator,
    ACTIONS(7228), 1,
      sym__lowercase_identifier,
    ACTIONS(7230), 1,
      anon_sym_LPAREN,
    ACTIONS(7232), 1,
      anon_sym_LBRACK,
    ACTIONS(7234), 1,
      anon_sym_LBRACE,
    STATE(2370), 1,
      sym_field_path,
    STATE(4595), 1,
      sym_attribute,
    STATE(5862), 1,
      sym_indexing_operator_path,
    STATE(6568), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [190471] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4596), 1,
      sym_attribute,
    STATE(6344), 1,
      sym_character_content,
    STATE(6844), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190501] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4597), 1,
      sym_attribute,
    STATE(6844), 1,
      sym_escape_sequence,
    STATE(7127), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190531] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3118), 1,
      anon_sym_COLON2,
    STATE(4598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3120), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190555] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4585), 1,
      anon_sym_COLON2,
    ACTIONS(7236), 1,
      anon_sym_LPAREN,
    STATE(4733), 1,
      sym_module_parameter,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4599), 2,
      sym_attribute,
      aux_sym_module_binding_repeat1,
    ACTIONS(4583), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [190583] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4600), 1,
      sym_attribute,
    STATE(6742), 1,
      sym_character_content,
    STATE(6844), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190613] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7160), 1,
      sym_indexing_operator,
    ACTIONS(7239), 1,
      sym__lowercase_identifier,
    ACTIONS(7241), 1,
      anon_sym_LPAREN,
    ACTIONS(7243), 1,
      anon_sym_LBRACK,
    ACTIONS(7245), 1,
      anon_sym_LBRACE,
    STATE(2750), 1,
      sym_field_path,
    STATE(4601), 1,
      sym_attribute,
    STATE(5879), 1,
      sym_indexing_operator_path,
    STATE(6722), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [190651] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3086), 1,
      anon_sym_COLON2,
    STATE(4602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3088), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190675] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1318), 1,
      anon_sym_COLON2,
    STATE(4603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1316), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190699] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3012), 1,
      anon_sym_COLON2,
    STATE(4604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3014), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190723] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3154), 1,
      anon_sym_COLON2,
    STATE(4605), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190747] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3214), 1,
      anon_sym_COLON2,
    STATE(4606), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3216), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190771] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7172), 1,
      aux_sym_character_content_token2,
    STATE(4607), 1,
      sym_attribute,
    STATE(6542), 1,
      sym_character_content,
    STATE(6844), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7170), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7174), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190801] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4608), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4483), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4505), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [190825] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4833), 1,
      anon_sym_in,
    ACTIONS(6579), 1,
      anon_sym_LBRACE,
    STATE(4609), 1,
      sym_attribute,
    STATE(4791), 1,
      aux_sym_external_repeat1,
    STATE(5377), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5538), 2,
      sym_string,
      sym_quoted_string,
  [190861] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3178), 1,
      anon_sym_COLON2,
    STATE(4610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3180), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190885] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3174), 1,
      anon_sym_COLON2,
    STATE(4611), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3176), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190909] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2994), 1,
      anon_sym_COLON2,
    STATE(4612), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2996), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190933] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4613), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7247), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [190955] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7249), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [190977] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3016), 1,
      anon_sym_COLON2,
    STATE(4615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3018), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191001] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3020), 1,
      anon_sym_COLON2,
    STATE(4616), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3022), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191025] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3036), 1,
      anon_sym_COLON2,
    STATE(4617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3038), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191049] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3040), 1,
      anon_sym_COLON2,
    STATE(4618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3042), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191073] = 12,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7160), 1,
      sym_indexing_operator,
    ACTIONS(7251), 1,
      sym__lowercase_identifier,
    ACTIONS(7253), 1,
      anon_sym_LPAREN,
    ACTIONS(7255), 1,
      anon_sym_LBRACK,
    ACTIONS(7257), 1,
      anon_sym_LBRACE,
    STATE(1899), 1,
      sym_field_path,
    STATE(4619), 1,
      sym_attribute,
    STATE(5855), 1,
      sym_indexing_operator_path,
    STATE(6524), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [191111] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4620), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5907), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [191142] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7261), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191163] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4622), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5198), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4894), 2,
      anon_sym_and,
      anon_sym_in,
  [191196] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7269), 1,
      sym__right_quoted_string_delimiter,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7266), 2,
      sym__null,
      aux_sym_string_content_token2,
    STATE(4623), 2,
      sym_attribute,
      aux_sym_quoted_string_content_repeat1,
    ACTIONS(7263), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [191221] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4624), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5213), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4910), 2,
      anon_sym_and,
      anon_sym_in,
  [191254] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4625), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5762), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [191285] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4626), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5701), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [191316] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4627), 1,
      sym_attribute,
    STATE(4647), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5222), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4869), 2,
      anon_sym_and,
      anon_sym_in,
  [191349] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4628), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(4672), 1,
      aux_sym_type_binding_repeat1,
    STATE(5217), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4914), 2,
      anon_sym_and,
      anon_sym_in,
  [191382] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    STATE(3593), 1,
      sym_constructor_declaration,
    STATE(4629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [191413] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    STATE(4630), 1,
      sym_attribute,
    STATE(5110), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [191444] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6731), 1,
      anon_sym_STAR,
    STATE(4631), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3740), 7,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_as,
      anon_sym_GT,
  [191467] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4632), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(4666), 1,
      aux_sym_type_binding_repeat1,
    STATE(5225), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4759), 2,
      anon_sym_and,
      anon_sym_in,
  [191500] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4633), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5059), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4877), 2,
      anon_sym_and,
      anon_sym_in,
  [191533] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4624), 1,
      aux_sym_type_binding_repeat1,
    STATE(4634), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5104), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4674), 2,
      anon_sym_and,
      anon_sym_in,
  [191566] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4635), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5552), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [191597] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7271), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4636), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(4685), 6,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [191620] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    STATE(3780), 1,
      sym_constructor_declaration,
    STATE(4637), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4059), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [191651] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4638), 1,
      sym_attribute,
    STATE(4648), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5061), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4865), 2,
      anon_sym_and,
      anon_sym_in,
  [191684] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4639), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5240), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4964), 2,
      anon_sym_and,
      anon_sym_in,
  [191717] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4640), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5913), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [191748] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4641), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7274), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191769] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7105), 1,
      anon_sym_RPAREN,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7276), 1,
      anon_sym_COLON2,
    STATE(4642), 1,
      sym_attribute,
    STATE(6495), 1,
      sym__typed,
    STATE(7093), 1,
      sym__polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [191804] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4643), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5551), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [191835] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7278), 1,
      anon_sym_and,
    STATE(4636), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(4644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4932), 6,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [191860] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7150), 1,
      anon_sym_RPAREN,
    ACTIONS(7276), 1,
      anon_sym_COLON2,
    STATE(4645), 1,
      sym_attribute,
    STATE(6881), 1,
      sym__typed,
    STATE(6921), 1,
      sym__polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [191895] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7280), 1,
      anon_sym_COLON,
    STATE(4646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7049), 7,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191918] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4647), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5087), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4906), 2,
      anon_sym_and,
      anon_sym_in,
  [191951] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4648), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5204), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4885), 2,
      anon_sym_and,
      anon_sym_in,
  [191984] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4821), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192005] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192026] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5494), 1,
      anon_sym_COLON2,
    ACTIONS(7282), 1,
      anon_sym_EQ,
    ACTIONS(7284), 1,
      anon_sym_of,
    STATE(4651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4922), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192053] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6605), 1,
      anon_sym_LPAREN,
    ACTIONS(6611), 1,
      anon_sym_LBRACK,
    ACTIONS(6617), 1,
      sym__uppercase_identifier,
    STATE(4652), 1,
      sym_attribute,
    STATE(5495), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6613), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4651), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [192084] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4465), 1,
      anon_sym_LPAREN,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    STATE(3780), 1,
      sym_constructor_declaration,
    STATE(4653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3503), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [192115] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4654), 1,
      sym_attribute,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4873), 2,
      anon_sym_and,
      anon_sym_in,
  [192148] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4655), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5544), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192179] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4841), 1,
      sym__lowercase_identifier,
    ACTIONS(6677), 1,
      anon_sym_POUND,
    ACTIONS(7286), 1,
      anon_sym_SEMI,
    ACTIONS(7288), 1,
      anon_sym_GT,
    STATE(4445), 1,
      sym_type_constructor_path,
    STATE(4656), 1,
      sym_attribute,
    STATE(5604), 1,
      aux_sym_object_type_repeat1,
    STATE(6168), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [192214] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4575), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192235] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4841), 1,
      sym__lowercase_identifier,
    ACTIONS(6677), 1,
      anon_sym_POUND,
    ACTIONS(7290), 1,
      anon_sym_SEMI,
    ACTIONS(7292), 1,
      anon_sym_GT,
    STATE(4445), 1,
      sym_type_constructor_path,
    STATE(4658), 1,
      sym_attribute,
    STATE(5723), 1,
      aux_sym_object_type_repeat1,
    STATE(6168), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [192270] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(4659), 1,
      sym_attribute,
    STATE(5195), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4861), 2,
      anon_sym_and,
      anon_sym_in,
  [192303] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7278), 1,
      anon_sym_and,
    STATE(4644), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(4660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4946), 6,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192328] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    STATE(3780), 1,
      sym_constructor_declaration,
    STATE(4661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [192359] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4622), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(4662), 1,
      sym_attribute,
    STATE(5211), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4837), 2,
      anon_sym_and,
      anon_sym_in,
  [192392] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4795), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192413] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4664), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5835), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192444] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7294), 1,
      sym__right_quoted_string_delimiter,
    STATE(4623), 1,
      aux_sym_quoted_string_content_repeat1,
    STATE(4665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7202), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7200), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [192471] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(4666), 1,
      sym_attribute,
    STATE(5116), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4942), 2,
      anon_sym_and,
      anon_sym_in,
  [192504] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4639), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(4667), 1,
      sym_attribute,
    STATE(5142), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4697), 2,
      anon_sym_and,
      anon_sym_in,
  [192537] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4668), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5798), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192568] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4669), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5804), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192599] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7296), 1,
      anon_sym_EQ,
    ACTIONS(7298), 1,
      anon_sym_COLON2,
    ACTIONS(7302), 1,
      anon_sym_COLON_GT,
    STATE(4248), 1,
      sym__coerced,
    STATE(4670), 1,
      sym_attribute,
    STATE(5113), 1,
      sym__typed,
    STATE(5746), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7300), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [192632] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4671), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5603), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192663] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(4672), 1,
      sym_attribute,
    STATE(5232), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4902), 2,
      anon_sym_and,
      anon_sym_in,
  [192696] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7045), 1,
      anon_sym_constraint,
    STATE(4633), 1,
      aux_sym_type_binding_repeat1,
    STATE(4657), 1,
      sym_type_constraint,
    STATE(4673), 1,
      sym_attribute,
    STATE(5126), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4881), 2,
      anon_sym_and,
      anon_sym_in,
  [192729] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7259), 1,
      anon_sym__,
    STATE(4674), 1,
      sym_attribute,
    STATE(4748), 1,
      aux_sym__type_param_repeat1,
    STATE(5226), 1,
      sym_type_variable,
    STATE(5909), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192760] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_LBRACK,
    ACTIONS(4471), 1,
      sym__uppercase_identifier,
    ACTIONS(4662), 1,
      anon_sym_LPAREN,
    STATE(3593), 1,
      sym_constructor_declaration,
    STATE(4675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4469), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3673), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [192791] = 11,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4841), 1,
      sym__lowercase_identifier,
    ACTIONS(6677), 1,
      anon_sym_POUND,
    ACTIONS(7304), 1,
      anon_sym_SEMI,
    ACTIONS(7306), 1,
      anon_sym_GT,
    STATE(4445), 1,
      sym_type_constructor_path,
    STATE(4676), 1,
      sym_attribute,
    STATE(5628), 1,
      aux_sym_object_type_repeat1,
    STATE(6168), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [192826] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7308), 1,
      anon_sym_COLON,
    STATE(4677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 7,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [192849] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1622), 1,
      anon_sym_PIPE,
    STATE(4678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1624), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [192871] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7310), 1,
      anon_sym_DASH_GT,
    ACTIONS(7312), 1,
      anon_sym_when,
    STATE(4679), 1,
      sym_attribute,
    STATE(6466), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [192903] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4685), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192923] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(3722), 1,
      anon_sym_STAR,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(7314), 1,
      anon_sym_DASH_GT,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(4681), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [192955] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7316), 1,
      anon_sym_COMMA,
    ACTIONS(7318), 1,
      anon_sym_PIPE,
    ACTIONS(7320), 1,
      anon_sym_SEMI,
    ACTIONS(7322), 1,
      anon_sym_as,
    ACTIONS(7324), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7326), 1,
      anon_sym_PIPE_RBRACK,
    STATE(4682), 1,
      sym_attribute,
    STATE(5577), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [192987] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7328), 1,
      anon_sym_RPAREN,
    STATE(4683), 1,
      sym_attribute,
    STATE(6751), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193019] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7330), 1,
      anon_sym_RPAREN,
    STATE(4684), 1,
      sym_attribute,
    STATE(6753), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193051] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7332), 1,
      anon_sym_RPAREN,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(4685), 1,
      sym_attribute,
    STATE(6521), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193083] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7312), 1,
      anon_sym_when,
    ACTIONS(7342), 1,
      anon_sym_DASH_GT,
    STATE(4686), 1,
      sym_attribute,
    STATE(6611), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193115] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5246), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4687), 2,
      sym_attribute,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(4537), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193137] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7346), 1,
      anon_sym__,
    ACTIONS(7348), 1,
      anon_sym_RBRACE,
    STATE(4688), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(6004), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193169] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4689), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1648), 7,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [193189] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4690), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(5915), 1,
      sym_field_pattern,
    STATE(6363), 1,
      sym__quoted_string,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193221] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4691), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(5926), 1,
      sym_field_binding_pattern,
    STATE(6363), 1,
      sym__quoted_string,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193253] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2984), 1,
      anon_sym_PIPE,
    ACTIONS(7350), 1,
      anon_sym_COLON,
    STATE(4692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [193277] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7352), 1,
      anon_sym_RPAREN,
    STATE(4693), 1,
      sym_attribute,
    STATE(6782), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193309] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4819), 1,
      anon_sym_COLON2,
    STATE(4694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4817), 6,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193331] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7354), 1,
      anon_sym_RPAREN,
    STATE(4695), 1,
      sym_attribute,
    STATE(6573), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193363] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7356), 1,
      anon_sym_RPAREN,
    STATE(4696), 1,
      sym_attribute,
    STATE(6579), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193395] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7358), 1,
      anon_sym_RBRACK,
    ACTIONS(7360), 1,
      anon_sym_when,
    STATE(4697), 1,
      sym_attribute,
    STATE(7286), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193427] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    ACTIONS(7364), 1,
      anon_sym_PIPE,
    ACTIONS(7366), 1,
      anon_sym_RBRACK,
    ACTIONS(7368), 1,
      anon_sym_SEMI,
    ACTIONS(7370), 1,
      anon_sym_COLON_COLON,
    STATE(4698), 1,
      sym_attribute,
    STATE(5720), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193459] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7372), 1,
      anon_sym__,
    ACTIONS(7374), 1,
      anon_sym_RBRACE,
    STATE(4699), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(6224), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193491] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7376), 1,
      anon_sym_DASH_GT,
    ACTIONS(7378), 1,
      anon_sym_with,
    STATE(4700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4980), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193515] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    ACTIONS(7364), 1,
      anon_sym_PIPE,
    ACTIONS(7370), 1,
      anon_sym_COLON_COLON,
    STATE(4701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6823), 4,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [193541] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(3722), 1,
      anon_sym_STAR,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(7380), 1,
      anon_sym_DASH_GT,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(4702), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193573] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7382), 1,
      anon_sym__,
    ACTIONS(7384), 1,
      anon_sym_RBRACE,
    STATE(4703), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(6224), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193605] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2984), 1,
      anon_sym_COLON2,
    ACTIONS(7386), 1,
      anon_sym_COLON,
    STATE(4704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [193629] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7388), 1,
      anon_sym__,
    ACTIONS(7390), 1,
      anon_sym_RBRACE,
    STATE(4705), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(6004), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193661] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7392), 1,
      anon_sym_RPAREN,
    STATE(4706), 1,
      sym_attribute,
    STATE(6614), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193693] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7312), 1,
      anon_sym_when,
    ACTIONS(7394), 1,
      anon_sym_DASH_GT,
    STATE(4707), 1,
      sym_attribute,
    STATE(6608), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193725] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7396), 1,
      anon_sym_RPAREN,
    STATE(4708), 1,
      sym_attribute,
    STATE(7313), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193757] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4693), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193777] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4621), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193797] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7398), 1,
      anon_sym_COMMA,
    ACTIONS(7400), 1,
      anon_sym_PIPE,
    ACTIONS(7402), 1,
      anon_sym_as,
    ACTIONS(7404), 1,
      anon_sym_COLON_COLON,
    STATE(4711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3136), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [193825] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(3722), 1,
      anon_sym_STAR,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(7406), 1,
      anon_sym_DASH_GT,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(4712), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193857] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7398), 1,
      anon_sym_COMMA,
    ACTIONS(7404), 1,
      anon_sym_COLON_COLON,
    STATE(4713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 5,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [193881] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7408), 1,
      anon_sym_RPAREN,
    STATE(4714), 1,
      sym_attribute,
    STATE(6619), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193913] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7410), 1,
      anon_sym__,
    ACTIONS(7412), 1,
      anon_sym_RBRACE,
    STATE(4715), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(6224), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193945] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7370), 1,
      anon_sym_COLON_COLON,
    STATE(4716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7115), 6,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [193967] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7376), 1,
      anon_sym_DASH_GT,
    ACTIONS(7378), 1,
      anon_sym_with,
    STATE(4717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5098), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193991] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7326), 1,
      anon_sym_RBRACK,
    ACTIONS(7398), 1,
      anon_sym_COMMA,
    ACTIONS(7400), 1,
      anon_sym_PIPE,
    ACTIONS(7402), 1,
      anon_sym_as,
    ACTIONS(7404), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7414), 1,
      anon_sym_SEMI,
    STATE(4718), 1,
      sym_attribute,
    STATE(5608), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194023] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7416), 1,
      anon_sym_RPAREN,
    STATE(4719), 1,
      sym_attribute,
    STATE(6530), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194055] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7312), 1,
      anon_sym_when,
    ACTIONS(7418), 1,
      anon_sym_DASH_GT,
    STATE(4720), 1,
      sym_attribute,
    STATE(7333), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194087] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7420), 1,
      anon_sym_RPAREN,
    STATE(4721), 1,
      sym_attribute,
    STATE(6724), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194119] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7422), 1,
      anon_sym_RPAREN,
    STATE(4722), 1,
      sym_attribute,
    STATE(6713), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194151] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7312), 1,
      anon_sym_when,
    ACTIONS(7424), 1,
      anon_sym_DASH_GT,
    STATE(4723), 1,
      sym_attribute,
    STATE(6889), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194183] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7426), 1,
      anon_sym__,
    ACTIONS(7428), 1,
      anon_sym_RBRACE,
    STATE(4724), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(6004), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194215] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7398), 1,
      anon_sym_COMMA,
    STATE(4725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 6,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [194237] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7398), 1,
      anon_sym_COMMA,
    ACTIONS(7400), 1,
      anon_sym_PIPE,
    ACTIONS(7404), 1,
      anon_sym_COLON_COLON,
    STATE(4726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3144), 4,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [194263] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7430), 1,
      sym__lowercase_identifier,
    ACTIONS(7432), 1,
      anon_sym_type,
    ACTIONS(7434), 1,
      anon_sym_LBRACK,
    ACTIONS(7436), 1,
      anon_sym_virtual,
    STATE(4082), 1,
      sym_class_binding,
    STATE(4727), 1,
      sym_attribute,
    STATE(5364), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194295] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7438), 1,
      anon_sym_RPAREN,
    STATE(4728), 1,
      sym_attribute,
    STATE(6343), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194327] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7404), 1,
      anon_sym_COLON_COLON,
    STATE(4729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 6,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [194349] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4730), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4533), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [194369] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4365), 1,
      anon_sym_STAR,
    STATE(4687), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(4731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5008), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [194393] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7440), 1,
      anon_sym_RPAREN,
    STATE(4732), 1,
      sym_attribute,
    STATE(6684), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194425] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4849), 1,
      anon_sym_COLON2,
    STATE(4733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4847), 6,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [194447] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4658), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [194467] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7442), 1,
      anon_sym_RPAREN,
    STATE(4735), 1,
      sym_attribute,
    STATE(6653), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194499] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1618), 1,
      anon_sym_PIPE,
    STATE(4736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1620), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [194521] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7053), 1,
      anon_sym_PIPE,
    ACTIONS(7444), 1,
      anon_sym_COLON,
    STATE(4737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7049), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [194545] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1634), 1,
      anon_sym_PIPE,
    STATE(4738), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1636), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [194567] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7446), 1,
      anon_sym_EQ,
    ACTIONS(7448), 1,
      anon_sym_COLON_EQ,
    STATE(4739), 1,
      sym_attribute,
    STATE(5201), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4952), 2,
      anon_sym_and,
      anon_sym_in,
  [194597] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7450), 1,
      anon_sym_RPAREN,
    STATE(4740), 1,
      sym_attribute,
    STATE(6658), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194629] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7452), 1,
      anon_sym__,
    ACTIONS(7454), 1,
      anon_sym_RBRACE,
    STATE(4741), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(6224), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194661] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1650), 1,
      anon_sym_PIPE,
    STATE(4742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1652), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [194683] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7456), 1,
      sym__lowercase_identifier,
    ACTIONS(7458), 1,
      anon_sym_type,
    ACTIONS(7460), 1,
      anon_sym_LBRACK,
    ACTIONS(7462), 1,
      anon_sym_virtual,
    STATE(4743), 1,
      sym_attribute,
    STATE(5450), 1,
      sym__attribute,
    STATE(5824), 1,
      sym_class_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194715] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1826), 1,
      anon_sym_PIPE,
    STATE(4744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1828), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [194737] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1318), 1,
      anon_sym_PIPE,
    ACTIONS(7464), 1,
      anon_sym_DOT_DOT,
    STATE(4745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1316), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [194761] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1842), 1,
      anon_sym_PIPE,
    STATE(4746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1844), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [194783] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1850), 1,
      anon_sym_PIPE,
    STATE(4747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1852), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [194805] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7466), 1,
      anon_sym__,
    STATE(4748), 1,
      sym_attribute,
    STATE(4990), 1,
      aux_sym__type_param_repeat1,
    STATE(5179), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [194833] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7468), 1,
      anon_sym_RPAREN,
    STATE(4749), 1,
      sym_attribute,
    STATE(6351), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194865] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    STATE(4750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6815), 6,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [194887] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7376), 1,
      anon_sym_DASH_GT,
    ACTIONS(7378), 1,
      anon_sym_with,
    STATE(4751), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5110), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [194911] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4629), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [194931] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7470), 1,
      anon_sym__,
    ACTIONS(7472), 1,
      anon_sym_RBRACE,
    STATE(4753), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(6224), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194963] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7312), 1,
      anon_sym_when,
    ACTIONS(7474), 1,
      anon_sym_DASH_GT,
    STATE(4754), 1,
      sym_attribute,
    STATE(7246), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194995] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7476), 1,
      anon_sym__,
    ACTIONS(7478), 1,
      anon_sym_RBRACE,
    STATE(4755), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(6004), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195027] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7480), 1,
      anon_sym_RPAREN,
    STATE(4756), 1,
      sym_attribute,
    STATE(7092), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195059] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7482), 1,
      anon_sym__,
    ACTIONS(7484), 1,
      anon_sym_RBRACE,
    STATE(4757), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(6004), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195091] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7486), 1,
      anon_sym_RPAREN,
    STATE(4758), 1,
      sym_attribute,
    STATE(6691), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195123] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4841), 1,
      sym__lowercase_identifier,
    ACTIONS(6677), 1,
      anon_sym_POUND,
    STATE(4445), 1,
      sym_type_constructor_path,
    STATE(4759), 1,
      sym_attribute,
    STATE(6168), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7488), 2,
      anon_sym_SEMI,
      anon_sym_GT,
  [195153] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7490), 1,
      anon_sym__,
    ACTIONS(7492), 1,
      anon_sym_RBRACE,
    STATE(4760), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(6224), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195185] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7496), 3,
      sym__right_quoted_string_delimiter,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7494), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [195207] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7498), 1,
      anon_sym_RPAREN,
    STATE(4762), 1,
      sym_attribute,
    STATE(6696), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195239] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7500), 1,
      anon_sym_EQ,
    ACTIONS(7502), 1,
      anon_sym_COLON_EQ,
    STATE(4763), 1,
      sym_attribute,
    STATE(5119), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4853), 2,
      anon_sym_and,
      anon_sym_in,
  [195269] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4505), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [195289] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4765), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(5726), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    STATE(7321), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195321] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4766), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(5831), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    STATE(7321), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195353] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4767), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(5637), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    STATE(7073), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195385] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7504), 1,
      anon_sym_rec,
    ACTIONS(7506), 1,
      anon_sym_type,
    STATE(4091), 1,
      sym_module_binding,
    STATE(4768), 1,
      sym_attribute,
    STATE(5431), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7508), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [195415] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4801), 1,
      anon_sym_COLON2,
    STATE(4769), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4799), 6,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [195437] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7376), 1,
      anon_sym_DASH_GT,
    ACTIONS(7378), 1,
      anon_sym_with,
    STATE(4770), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5016), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [195461] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4610), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [195481] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7510), 1,
      anon_sym_RPAREN,
    STATE(4772), 1,
      sym_attribute,
    STATE(6834), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195513] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4773), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(5687), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    STATE(6787), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195545] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7512), 1,
      anon_sym__,
    ACTIONS(7514), 1,
      anon_sym_RBRACE,
    STATE(4774), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(6004), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195577] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4775), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(5692), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    STATE(7126), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195609] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(7314), 1,
      anon_sym_DASH_GT,
    ACTIONS(7516), 1,
      anon_sym_STAR,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(4776), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195641] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7519), 1,
      anon_sym__,
    ACTIONS(7521), 1,
      anon_sym_RBRACE,
    STATE(4777), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(6004), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195673] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7523), 1,
      anon_sym_RPAREN,
    STATE(4778), 1,
      sym_attribute,
    STATE(6446), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195705] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7525), 1,
      anon_sym_RPAREN,
    STATE(4779), 1,
      sym_attribute,
    STATE(6452), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195737] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(3722), 1,
      anon_sym_STAR,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(7527), 1,
      anon_sym_DASH_GT,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(4780), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195769] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7529), 1,
      anon_sym_RPAREN,
    STATE(4781), 1,
      sym_attribute,
    STATE(6536), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195801] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    STATE(4782), 1,
      sym_attribute,
    STATE(6458), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4533), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [195827] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(7464), 1,
      anon_sym_DOT_DOT,
    STATE(4783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [195851] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1616), 7,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [195871] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4785), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(5574), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    STATE(7164), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195903] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6799), 1,
      anon_sym_DASH_GT,
    STATE(4786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 6,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_as,
      anon_sym_GT,
  [195925] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    ACTIONS(7364), 1,
      anon_sym_PIPE,
    ACTIONS(7370), 1,
      anon_sym_COLON_COLON,
    STATE(4787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7132), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [195953] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2970), 1,
      anon_sym_PIPE,
    STATE(4788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2972), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [195975] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7312), 1,
      anon_sym_when,
    ACTIONS(7531), 1,
      anon_sym_DASH_GT,
    STATE(4789), 1,
      sym_attribute,
    STATE(6837), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196007] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7366), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(7533), 1,
      anon_sym_COMMA,
    ACTIONS(7535), 1,
      anon_sym_PIPE,
    ACTIONS(7537), 1,
      anon_sym_SEMI,
    ACTIONS(7539), 1,
      anon_sym_as,
    ACTIONS(7541), 1,
      anon_sym_COLON_COLON,
    STATE(4790), 1,
      sym_attribute,
    STATE(5575), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196039] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7543), 1,
      anon_sym_LBRACE,
    ACTIONS(7546), 1,
      anon_sym_DQUOTE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5118), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    STATE(4791), 2,
      sym_attribute,
      aux_sym_external_repeat1,
    STATE(5538), 2,
      sym_string,
      sym_quoted_string,
  [196065] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7549), 1,
      anon_sym__,
    ACTIONS(7551), 1,
      anon_sym_RBRACE,
    STATE(4792), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(6224), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196097] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7553), 1,
      anon_sym__,
    ACTIONS(7555), 1,
      anon_sym_RBRACE,
    STATE(4793), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(6224), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196129] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7344), 1,
      sym__lowercase_identifier,
    ACTIONS(7557), 1,
      anon_sym__,
    ACTIONS(7559), 1,
      anon_sym_RBRACE,
    STATE(4794), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(6004), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196161] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7561), 1,
      anon_sym_RPAREN,
    STATE(4795), 1,
      sym_attribute,
    STATE(6726), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196193] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(3722), 1,
      anon_sym_STAR,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(7563), 1,
      anon_sym_DASH_GT,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(4796), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196225] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7565), 1,
      anon_sym_rec,
    ACTIONS(7567), 1,
      anon_sym_type,
    STATE(4797), 1,
      sym_attribute,
    STATE(5467), 1,
      sym__attribute,
    STATE(5818), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7569), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [196255] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7571), 1,
      anon_sym_rec,
    ACTIONS(7573), 1,
      anon_sym_type,
    STATE(4050), 1,
      sym_module_binding,
    STATE(4798), 1,
      sym_attribute,
    STATE(5527), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7575), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [196285] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7577), 1,
      anon_sym_COLON2,
    STATE(4799), 1,
      sym_attribute,
    STATE(6076), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 5,
      anon_sym_SEMI,
      anon_sym_GT,
      anon_sym_POUND,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [196309] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7376), 1,
      anon_sym_DASH_GT,
    ACTIONS(7378), 1,
      anon_sym_with,
    STATE(4800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5004), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [196333] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7579), 1,
      anon_sym_RPAREN,
    STATE(4801), 1,
      sym_attribute,
    STATE(6337), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196365] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(3722), 1,
      anon_sym_STAR,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(7581), 1,
      anon_sym_DASH_GT,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(4802), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196397] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1698), 1,
      anon_sym_PIPE,
    ACTIONS(1744), 1,
      anon_sym_DOT,
    STATE(4803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196421] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7583), 1,
      anon_sym_RPAREN,
    STATE(4804), 1,
      sym_attribute,
    STATE(6731), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196453] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7398), 1,
      anon_sym_COMMA,
    ACTIONS(7400), 1,
      anon_sym_PIPE,
    ACTIONS(7402), 1,
      anon_sym_as,
    ACTIONS(7404), 1,
      anon_sym_COLON_COLON,
    STATE(4805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3126), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [196481] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2974), 1,
      anon_sym_PIPE,
    STATE(4806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2976), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196503] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7432), 1,
      anon_sym_type,
    ACTIONS(7585), 1,
      sym__lowercase_identifier,
    ACTIONS(7587), 1,
      anon_sym_LBRACK,
    ACTIONS(7589), 1,
      anon_sym_virtual,
    STATE(3946), 1,
      sym_class_binding,
    STATE(4807), 1,
      sym_attribute,
    STATE(5472), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196535] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    ACTIONS(7364), 1,
      anon_sym_PIPE,
    ACTIONS(7370), 1,
      anon_sym_COLON_COLON,
    STATE(4808), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7128), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [196563] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7312), 1,
      anon_sym_when,
    ACTIONS(7591), 1,
      anon_sym_DASH_GT,
    STATE(4809), 1,
      sym_attribute,
    STATE(6985), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196595] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1698), 1,
      anon_sym_PIPE,
    ACTIONS(1702), 1,
      anon_sym_DOT,
    STATE(4810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1700), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196619] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    ACTIONS(7370), 1,
      anon_sym_COLON_COLON,
    STATE(4811), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6815), 5,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [196643] = 10,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7593), 1,
      anon_sym_RPAREN,
    STATE(4812), 1,
      sym_attribute,
    STATE(6886), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196675] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1834), 1,
      anon_sym_PIPE,
    STATE(4813), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1836), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196696] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5364), 1,
      anon_sym_in,
    ACTIONS(7378), 1,
      anon_sym_with,
    ACTIONS(7595), 1,
      anon_sym_DASH_GT,
    STATE(4814), 1,
      sym_attribute,
    STATE(5421), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196725] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    ACTIONS(7364), 1,
      anon_sym_PIPE,
    ACTIONS(7370), 1,
      anon_sym_COLON_COLON,
    STATE(4815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7597), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [196752] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3138), 1,
      anon_sym_COLON2,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(4816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 3,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_as,
  [196777] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7599), 1,
      sym__lowercase_identifier,
    STATE(2688), 1,
      sym_class_path,
    STATE(4817), 1,
      sym_attribute,
    STATE(5279), 1,
      sym__attribute,
    STATE(6712), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196806] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3024), 1,
      anon_sym_PIPE,
    STATE(4818), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3026), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196827] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3028), 1,
      anon_sym_PIPE,
    STATE(4819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3030), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196848] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7398), 1,
      anon_sym_COMMA,
    ACTIONS(7400), 1,
      anon_sym_PIPE,
    ACTIONS(7402), 1,
      anon_sym_as,
    ACTIONS(7404), 1,
      anon_sym_COLON_COLON,
    STATE(4820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7601), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [196875] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3032), 1,
      anon_sym_PIPE,
    STATE(4821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3034), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196896] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3138), 1,
      anon_sym_COLON2,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    STATE(4822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 4,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [196919] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7603), 1,
      anon_sym_COLON2,
    STATE(4823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4533), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [196942] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3062), 1,
      anon_sym_PIPE,
    STATE(4824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3064), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196963] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3142), 1,
      anon_sym_COLON2,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(4825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3144), 2,
      anon_sym_RPAREN,
      anon_sym_as,
  [196990] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7607), 1,
      anon_sym_AMP,
    STATE(4826), 1,
      sym_attribute,
    STATE(5083), 1,
      aux_sym_tag_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7605), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [197015] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1754), 1,
      anon_sym_PIPE,
    STATE(4827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1756), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197036] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7609), 1,
      anon_sym_PIPE,
    STATE(4828), 1,
      sym_attribute,
    STATE(4835), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5374), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [197059] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7611), 1,
      anon_sym_RPAREN,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(4829), 1,
      sym_attribute,
    STATE(6589), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197088] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3138), 1,
      anon_sym_PIPE,
    ACTIONS(7316), 1,
      anon_sym_COMMA,
    ACTIONS(7324), 1,
      anon_sym_COLON_COLON,
    STATE(4830), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [197113] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7613), 1,
      sym__lowercase_identifier,
    STATE(2355), 1,
      sym_class_path,
    STATE(4831), 1,
      sym_attribute,
    STATE(5293), 1,
      sym__attribute,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197142] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3138), 1,
      anon_sym_PIPE,
    ACTIONS(7316), 1,
      anon_sym_COMMA,
    STATE(4832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 4,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197165] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3554), 1,
      anon_sym_PIPE,
    STATE(4833), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3552), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197186] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3210), 1,
      anon_sym_PIPE,
    STATE(4834), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3212), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197207] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7615), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4835), 2,
      sym_attribute,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5230), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [197228] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7618), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4836), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197257] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3138), 1,
      anon_sym_PIPE,
    STATE(4837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197278] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7620), 1,
      anon_sym_RPAREN,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(4838), 1,
      sym_attribute,
    STATE(7177), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197307] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3174), 1,
      anon_sym_PIPE,
    STATE(4839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3176), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197328] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7622), 1,
      sym__lowercase_identifier,
    ACTIONS(7624), 1,
      anon_sym_LBRACK,
    ACTIONS(7626), 1,
      anon_sym_virtual,
    STATE(3878), 1,
      sym_class_type_binding,
    STATE(4840), 1,
      sym_attribute,
    STATE(5285), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197357] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3178), 1,
      anon_sym_PIPE,
    STATE(4841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3180), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197378] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7628), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4842), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197407] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7630), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4843), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197436] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3150), 1,
      anon_sym_COLON2,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(4844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 4,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_as,
  [197459] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7632), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4845), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197488] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7316), 1,
      anon_sym_COMMA,
    ACTIONS(7318), 1,
      anon_sym_PIPE,
    ACTIONS(7324), 1,
      anon_sym_COLON_COLON,
    STATE(4846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3144), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [197513] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3142), 1,
      anon_sym_PIPE,
    STATE(4847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3144), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197534] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5666), 1,
      anon_sym_in,
    STATE(4848), 1,
      sym_attribute,
    STATE(5313), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7634), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [197561] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7636), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4849), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197590] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3146), 1,
      anon_sym_PIPE,
    STATE(4850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3148), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197611] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3182), 1,
      anon_sym_PIPE,
    STATE(4851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3184), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197632] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3150), 1,
      anon_sym_PIPE,
    ACTIONS(7324), 1,
      anon_sym_COLON_COLON,
    STATE(4852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [197655] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7398), 1,
      anon_sym_COMMA,
    ACTIONS(7400), 1,
      anon_sym_PIPE,
    ACTIONS(7402), 1,
      anon_sym_as,
    ACTIONS(7404), 1,
      anon_sym_COLON_COLON,
    STATE(4853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7638), 2,
      anon_sym_RBRACK,
      anon_sym_SEMI,
  [197682] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3198), 1,
      anon_sym_PIPE,
    STATE(4854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3200), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197703] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7640), 1,
      sym__lowercase_identifier,
    STATE(2040), 1,
      sym_class_path,
    STATE(4855), 1,
      sym_attribute,
    STATE(5301), 1,
      sym__attribute,
    STATE(6744), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197732] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3202), 1,
      anon_sym_PIPE,
    STATE(4856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3204), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197753] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3493), 1,
      anon_sym_PIPE,
    STATE(4857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3491), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197774] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3214), 1,
      anon_sym_PIPE,
    STATE(4858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3216), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197795] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7642), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4859), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197824] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7644), 1,
      sym__lowercase_identifier,
    STATE(1092), 1,
      sym_class_path,
    STATE(4860), 1,
      sym_attribute,
    STATE(5357), 1,
      sym__attribute,
    STATE(6942), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197853] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7646), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4861), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197882] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7316), 1,
      anon_sym_COMMA,
    ACTIONS(7318), 1,
      anon_sym_PIPE,
    ACTIONS(7322), 1,
      anon_sym_as,
    ACTIONS(7324), 1,
      anon_sym_COLON_COLON,
    STATE(4862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7638), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [197909] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3427), 1,
      anon_sym_PIPE,
    STATE(4863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3425), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197930] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5108), 1,
      anon_sym_and,
    STATE(4864), 1,
      sym_attribute,
    STATE(4939), 1,
      aux_sym_expression_item_repeat1,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5106), 2,
      anon_sym_in,
      sym_let_and_operator,
  [197957] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3106), 1,
      anon_sym_PIPE,
    STATE(4865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197978] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5077), 1,
      anon_sym_and,
    STATE(4866), 1,
      sym_attribute,
    STATE(5002), 1,
      aux_sym_expression_item_repeat1,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5075), 2,
      anon_sym_in,
      sym_let_and_operator,
  [198005] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7648), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4867), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198034] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1854), 1,
      anon_sym_PIPE,
    STATE(4868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1856), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198055] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5116), 1,
      anon_sym_and,
    STATE(4869), 1,
      sym_attribute,
    STATE(5004), 1,
      aux_sym_expression_item_repeat1,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5114), 2,
      anon_sym_in,
      sym_let_and_operator,
  [198082] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7650), 1,
      anon_sym_EQ,
    STATE(4870), 1,
      sym_attribute,
    STATE(5174), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5403), 2,
      anon_sym_and,
      anon_sym_in,
  [198109] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7652), 1,
      anon_sym_EQ,
    STATE(4871), 1,
      sym_attribute,
    STATE(5184), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5409), 2,
      anon_sym_and,
      anon_sym_in,
  [198136] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3395), 1,
      anon_sym_PIPE,
    STATE(4872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3393), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198157] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7654), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4873), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198186] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5022), 1,
      anon_sym_and,
    STATE(4874), 1,
      sym_attribute,
    STATE(4939), 1,
      aux_sym_expression_item_repeat1,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5020), 2,
      anon_sym_in,
      sym_let_and_operator,
  [198213] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5056), 1,
      anon_sym_and,
    STATE(4875), 1,
      sym_attribute,
    STATE(4939), 1,
      aux_sym_expression_item_repeat1,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5054), 2,
      anon_sym_in,
      sym_let_and_operator,
  [198240] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5032), 1,
      anon_sym_and,
    STATE(4876), 1,
      sym_attribute,
    STATE(4939), 1,
      aux_sym_expression_item_repeat1,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5030), 2,
      anon_sym_in,
      sym_let_and_operator,
  [198267] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5040), 1,
      anon_sym_and,
    STATE(4877), 1,
      sym_attribute,
    STATE(4908), 1,
      aux_sym_expression_item_repeat1,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5038), 2,
      anon_sym_in,
      sym_let_and_operator,
  [198294] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5036), 1,
      anon_sym_and,
    STATE(4878), 1,
      sym_attribute,
    STATE(5017), 1,
      aux_sym_expression_item_repeat1,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5034), 2,
      anon_sym_in,
      sym_let_and_operator,
  [198321] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4604), 1,
      anon_sym_COLON2,
    STATE(4879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 5,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [198342] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7533), 1,
      anon_sym_COMMA,
    ACTIONS(7535), 1,
      anon_sym_PIPE,
    ACTIONS(7539), 1,
      anon_sym_as,
    ACTIONS(7541), 1,
      anon_sym_COLON_COLON,
    STATE(4880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7128), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [198369] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7398), 1,
      anon_sym_COMMA,
    ACTIONS(7400), 1,
      anon_sym_PIPE,
    ACTIONS(7402), 1,
      anon_sym_as,
    ACTIONS(7404), 1,
      anon_sym_COLON_COLON,
    STATE(4881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7656), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [198396] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6939), 1,
      anon_sym_PIPE,
    STATE(4882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6937), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198417] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6977), 1,
      anon_sym_PIPE,
    STATE(4883), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6975), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198438] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7658), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4884), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198467] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3124), 1,
      anon_sym_COLON2,
    ACTIONS(3126), 1,
      anon_sym_RPAREN,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(4885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198496] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7001), 1,
      anon_sym_PIPE,
    STATE(4886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6999), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198517] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3403), 1,
      anon_sym_PIPE,
    STATE(4887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3401), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198538] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5381), 1,
      anon_sym_in,
    ACTIONS(7378), 1,
      anon_sym_with,
    ACTIONS(7595), 1,
      anon_sym_DASH_GT,
    STATE(4888), 1,
      sym_attribute,
    STATE(5409), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198567] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7660), 1,
      sym__lowercase_identifier,
    STATE(996), 1,
      sym_class_path,
    STATE(4889), 1,
      sym_attribute,
    STATE(5443), 1,
      sym__attribute,
    STATE(6599), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198596] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7613), 1,
      sym__lowercase_identifier,
    STATE(828), 1,
      sym_class_path,
    STATE(4890), 1,
      sym_attribute,
    STATE(5358), 1,
      sym__attribute,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198625] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3538), 1,
      anon_sym_PIPE,
    STATE(4891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3536), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198646] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1614), 1,
      anon_sym_PIPE,
    STATE(4892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1616), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198667] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7662), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4893), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198696] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1646), 1,
      anon_sym_PIPE,
    STATE(4894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1648), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198717] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7664), 1,
      anon_sym_RPAREN,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(4895), 1,
      sym_attribute,
    STATE(6939), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198746] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3550), 1,
      anon_sym_PIPE,
    STATE(4896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3548), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198767] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3162), 1,
      anon_sym_PIPE,
    STATE(4897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3164), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198788] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3347), 1,
      anon_sym_PIPE,
    STATE(4898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3345), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198809] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3166), 1,
      anon_sym_PIPE,
    STATE(4899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3168), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198830] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3351), 1,
      anon_sym_PIPE,
    STATE(4900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3349), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198851] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7666), 1,
      sym__lowercase_identifier,
    STATE(1825), 1,
      sym_class_path,
    STATE(4901), 1,
      sym_attribute,
    STATE(5384), 1,
      sym__attribute,
    STATE(6510), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198880] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3383), 1,
      anon_sym_PIPE,
    STATE(4902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3381), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198901] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7668), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4903), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198930] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6817), 1,
      anon_sym_PIPE,
    STATE(4904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6815), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198951] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7670), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4905), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198980] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7672), 1,
      anon_sym_EQ,
    STATE(4906), 1,
      sym_attribute,
    STATE(5054), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5419), 2,
      anon_sym_and,
      anon_sym_in,
  [199007] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7674), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4907), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199036] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5048), 1,
      anon_sym_and,
    STATE(4908), 1,
      sym_attribute,
    STATE(4939), 1,
      aux_sym_expression_item_repeat1,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5046), 2,
      anon_sym_in,
      sym_let_and_operator,
  [199063] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7676), 1,
      anon_sym_COLON2,
    STATE(4909), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4533), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [199086] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2994), 1,
      anon_sym_PIPE,
    STATE(4910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2996), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199107] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3016), 1,
      anon_sym_PIPE,
    STATE(4911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3018), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199128] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7678), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4912), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199157] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3020), 1,
      anon_sym_PIPE,
    STATE(4913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3022), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199178] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3036), 1,
      anon_sym_PIPE,
    STATE(4914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3038), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199199] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7680), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4915), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199228] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3040), 1,
      anon_sym_PIPE,
    STATE(4916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3042), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199249] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(4917), 1,
      sym_attribute,
    STATE(5277), 1,
      sym__attribute,
    STATE(6247), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199278] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7686), 1,
      sym__lowercase_identifier,
    ACTIONS(7688), 1,
      anon_sym_LBRACK,
    ACTIONS(7690), 1,
      anon_sym_virtual,
    STATE(4918), 1,
      sym_attribute,
    STATE(5273), 1,
      sym__attribute,
    STATE(5858), 1,
      sym_class_type_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199307] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7692), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4919), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199336] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7694), 1,
      anon_sym_RPAREN,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(4920), 1,
      sym_attribute,
    STATE(6876), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199365] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7607), 1,
      anon_sym_AMP,
    STATE(4921), 1,
      sym_attribute,
    STATE(5125), 1,
      aux_sym_tag_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7696), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [199390] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6817), 1,
      anon_sym_PIPE,
    ACTIONS(7533), 1,
      anon_sym_COMMA,
    ACTIONS(7541), 1,
      anon_sym_COLON_COLON,
    STATE(4922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6815), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [199415] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7698), 1,
      sym__lowercase_identifier,
    STATE(1749), 1,
      sym_class_path,
    STATE(4923), 1,
      sym_attribute,
    STATE(5275), 1,
      sym__attribute,
    STATE(6420), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199444] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3431), 1,
      anon_sym_PIPE,
    STATE(4924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3429), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199465] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3082), 1,
      anon_sym_PIPE,
    STATE(4925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3084), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199486] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7700), 1,
      sym__lowercase_identifier,
    STATE(1593), 1,
      sym_class_path,
    STATE(4926), 1,
      sym_attribute,
    STATE(5483), 1,
      sym__attribute,
    STATE(6639), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199515] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    STATE(4927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3088), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199536] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6817), 1,
      anon_sym_PIPE,
    ACTIONS(7533), 1,
      anon_sym_COMMA,
    STATE(4928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6815), 4,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199559] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7533), 1,
      anon_sym_COMMA,
    ACTIONS(7535), 1,
      anon_sym_PIPE,
    ACTIONS(7539), 1,
      anon_sym_as,
    ACTIONS(7541), 1,
      anon_sym_COLON_COLON,
    STATE(4929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7132), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [199586] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(4930), 1,
      sym_attribute,
    STATE(5323), 1,
      sym__attribute,
    STATE(6209), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199615] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7702), 1,
      sym__lowercase_identifier,
    STATE(828), 1,
      sym_class_path,
    STATE(4931), 1,
      sym_attribute,
    STATE(5525), 1,
      sym__attribute,
    STATE(7288), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199644] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7704), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4932), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199673] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    ACTIONS(7364), 1,
      anon_sym_PIPE,
    ACTIONS(7370), 1,
      anon_sym_COLON_COLON,
    STATE(4933), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7706), 2,
      anon_sym_RBRACK,
      anon_sym_SEMI,
  [199700] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7708), 1,
      anon_sym_RPAREN,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(4934), 1,
      sym_attribute,
    STATE(6523), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199729] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6817), 1,
      anon_sym_PIPE,
    STATE(4935), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6815), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199750] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7710), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4936), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199779] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1750), 1,
      anon_sym_PIPE,
    STATE(4937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1752), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199800] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7609), 1,
      anon_sym_PIPE,
    STATE(4938), 1,
      sym_attribute,
    STATE(4956), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4461), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [199823] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4651), 1,
      anon_sym_and,
    ACTIONS(7712), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4649), 2,
      anon_sym_in,
      sym_let_and_operator,
    STATE(4939), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
  [199848] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3090), 1,
      anon_sym_PIPE,
    STATE(4940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3092), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199869] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7715), 1,
      anon_sym_RPAREN,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(4941), 1,
      sym_attribute,
    STATE(6492), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199898] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7533), 1,
      anon_sym_COMMA,
    ACTIONS(7535), 1,
      anon_sym_PIPE,
    ACTIONS(7541), 1,
      anon_sym_COLON_COLON,
    STATE(4942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6823), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [199923] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7717), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4943), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199952] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3218), 1,
      anon_sym_PIPE,
    STATE(4944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3220), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199973] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7719), 1,
      anon_sym_EQ,
    STATE(4945), 1,
      sym_attribute,
    STATE(5159), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5368), 2,
      anon_sym_and,
      anon_sym_in,
  [200000] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7721), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4946), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200029] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7723), 1,
      anon_sym_COLON2,
    STATE(4947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4533), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [200052] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7725), 1,
      anon_sym_COLON2,
    STATE(4948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4533), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [200075] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7727), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4949), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200104] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7729), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4950), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200133] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3291), 1,
      anon_sym_PIPE,
    STATE(4951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3289), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200154] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7731), 1,
      anon_sym_EQ,
    STATE(4952), 1,
      sym_attribute,
    STATE(5118), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5385), 2,
      anon_sym_and,
      anon_sym_in,
  [200181] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    ACTIONS(7364), 1,
      anon_sym_PIPE,
    ACTIONS(7370), 1,
      anon_sym_COLON_COLON,
    STATE(4953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7733), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [200208] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7735), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4954), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200237] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7609), 1,
      anon_sym_PIPE,
    STATE(4828), 1,
      aux_sym_variant_declaration_repeat1,
    STATE(4955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5358), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [200260] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7609), 1,
      anon_sym_PIPE,
    STATE(4835), 1,
      aux_sym_variant_declaration_repeat1,
    STATE(4956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5358), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [200283] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7737), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4957), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200312] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    STATE(4958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2139), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200333] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3546), 1,
      anon_sym_PIPE,
    STATE(4959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3544), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200354] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7739), 1,
      anon_sym_EQ,
    STATE(4960), 1,
      sym_attribute,
    STATE(5143), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5391), 2,
      anon_sym_and,
      anon_sym_in,
  [200381] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6825), 1,
      anon_sym_PIPE,
    STATE(4961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6823), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200402] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7741), 1,
      anon_sym_EQ,
    STATE(4962), 1,
      sym_attribute,
    STATE(5164), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5397), 2,
      anon_sym_and,
      anon_sym_in,
  [200429] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1642), 1,
      anon_sym_PIPE,
    STATE(4963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1644), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200450] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7743), 1,
      anon_sym_RPAREN,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(4964), 1,
      sym_attribute,
    STATE(6371), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200479] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1318), 1,
      anon_sym_PIPE,
    STATE(4965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1316), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200500] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6829), 1,
      anon_sym_PIPE,
    STATE(4966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6827), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200521] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3012), 1,
      anon_sym_PIPE,
    STATE(4967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3014), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200542] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5014), 1,
      anon_sym_and,
    STATE(4939), 1,
      aux_sym_expression_item_repeat1,
    STATE(4968), 1,
      sym_attribute,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5012), 2,
      anon_sym_in,
      sym_let_and_operator,
  [200569] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3154), 1,
      anon_sym_PIPE,
    STATE(4969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200590] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7745), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4970), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200619] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7117), 1,
      anon_sym_PIPE,
    ACTIONS(7541), 1,
      anon_sym_COLON_COLON,
    STATE(4971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7115), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [200642] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7747), 1,
      sym__lowercase_identifier,
    STATE(894), 1,
      sym_class_path,
    STATE(4972), 1,
      sym_attribute,
    STATE(5308), 1,
      sym__attribute,
    STATE(6677), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200671] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1926), 1,
      anon_sym_PIPE,
    STATE(4973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1928), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200692] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7749), 1,
      anon_sym_EQ,
    STATE(4974), 1,
      sym_attribute,
    STATE(5091), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5346), 2,
      anon_sym_and,
      anon_sym_in,
  [200719] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3475), 1,
      anon_sym_PIPE,
    STATE(4975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3473), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200740] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3194), 1,
      anon_sym_PIPE,
    STATE(4976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3196), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200761] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6837), 1,
      anon_sym_PIPE,
    STATE(4977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6835), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200782] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7751), 1,
      anon_sym_RPAREN,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(4978), 1,
      sym_attribute,
    STATE(7332), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200811] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7753), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4979), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200840] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    STATE(4980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1832), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200861] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6855), 1,
      anon_sym_PIPE,
    STATE(4981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6853), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200882] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7755), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4982), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200911] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5644), 1,
      anon_sym_in,
    STATE(4983), 1,
      sym_attribute,
    STATE(5306), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7757), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [200938] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1930), 1,
      anon_sym_PIPE,
    STATE(4984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1932), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200959] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7759), 1,
      anon_sym_COLON2,
    STATE(4985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4533), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [200982] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7761), 1,
      anon_sym_COLON2,
    STATE(4986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4501), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4533), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [201005] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6895), 1,
      anon_sym_PIPE,
    STATE(4987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6893), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201026] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7763), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4988), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201055] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7765), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4989), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201084] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7770), 2,
      anon_sym__,
      aux_sym_type_variable_token1,
    STATE(4990), 2,
      sym_attribute,
      aux_sym__type_param_repeat1,
    ACTIONS(7767), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [201105] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3110), 1,
      anon_sym_PIPE,
    STATE(4991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3112), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201126] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7772), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4992), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201155] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1934), 1,
      anon_sym_PIPE,
    STATE(4993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1936), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201176] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6899), 1,
      anon_sym_PIPE,
    STATE(4994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6897), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201197] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7774), 1,
      anon_sym_EQ,
    STATE(4995), 1,
      sym_attribute,
    STATE(5099), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5332), 2,
      anon_sym_and,
      anon_sym_in,
  [201224] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7776), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(4996), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201253] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3158), 1,
      anon_sym_PIPE,
    STATE(4997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201274] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(3648), 1,
      anon_sym_POUND,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(7778), 1,
      anon_sym_DASH_GT,
    STATE(2593), 1,
      sym_type_constructor_path,
    STATE(4998), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201303] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1708), 1,
      anon_sym_PIPE,
    STATE(4999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1710), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201324] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7780), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(5000), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201353] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7782), 1,
      anon_sym_RPAREN,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5001), 1,
      sym_attribute,
    STATE(6430), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201382] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5096), 1,
      anon_sym_and,
    STATE(4939), 1,
      aux_sym_expression_item_repeat1,
    STATE(5002), 1,
      sym_attribute,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5094), 2,
      anon_sym_in,
      sym_let_and_operator,
  [201409] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1938), 1,
      anon_sym_PIPE,
    STATE(5003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1940), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201430] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5052), 1,
      anon_sym_and,
    STATE(4939), 1,
      aux_sym_expression_item_repeat1,
    STATE(5004), 1,
      sym_attribute,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5050), 2,
      anon_sym_in,
      sym_let_and_operator,
  [201457] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3114), 1,
      anon_sym_PIPE,
    STATE(5005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201478] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7784), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(5006), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201507] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(5007), 1,
      sym_attribute,
    STATE(5383), 1,
      sym__attribute,
    STATE(6084), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201536] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6935), 1,
      anon_sym_PIPE,
    STATE(5008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6933), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201557] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7043), 1,
      anon_sym_PIPE,
    STATE(5009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7041), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201578] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6951), 1,
      anon_sym_PIPE,
    STATE(5010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6949), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201599] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6955), 1,
      anon_sym_PIPE,
    STATE(5011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6953), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201620] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7786), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(5012), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201649] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(5013), 1,
      sym_attribute,
    STATE(5517), 1,
      sym__attribute,
    STATE(6068), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201678] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7533), 1,
      anon_sym_COMMA,
    ACTIONS(7535), 1,
      anon_sym_PIPE,
    ACTIONS(7539), 1,
      anon_sym_as,
    ACTIONS(7541), 1,
      anon_sym_COLON_COLON,
    STATE(5014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7706), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [201705] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5104), 1,
      anon_sym_and,
    STATE(4874), 1,
      aux_sym_expression_item_repeat1,
    STATE(5015), 1,
      sym_attribute,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5102), 2,
      anon_sym_in,
      sym_let_and_operator,
  [201732] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3102), 1,
      anon_sym_PIPE,
    STATE(5016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3104), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201753] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5089), 1,
      anon_sym_and,
    STATE(4939), 1,
      aux_sym_expression_item_repeat1,
    STATE(5017), 1,
      sym_attribute,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5087), 2,
      anon_sym_in,
      sym_let_and_operator,
  [201780] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3118), 1,
      anon_sym_PIPE,
    STATE(5018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3120), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201801] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5044), 1,
      anon_sym_and,
    STATE(4875), 1,
      aux_sym_expression_item_repeat1,
    STATE(5019), 1,
      sym_attribute,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5042), 2,
      anon_sym_in,
      sym_let_and_operator,
  [201828] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5081), 1,
      anon_sym_and,
    STATE(4876), 1,
      aux_sym_expression_item_repeat1,
    STATE(5020), 1,
      sym_attribute,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5079), 2,
      anon_sym_in,
      sym_let_and_operator,
  [201855] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7788), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(5021), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201884] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6849), 1,
      anon_sym_PIPE,
    STATE(5022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6847), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201905] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7790), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(5023), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201934] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7792), 1,
      anon_sym_EQ,
    STATE(5024), 1,
      sym_attribute,
    STATE(5243), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5352), 2,
      anon_sym_and,
      anon_sym_in,
  [201961] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(5025), 1,
      sym_attribute,
    STATE(5524), 1,
      sym__attribute,
    STATE(6171), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201990] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7794), 1,
      sym__lowercase_identifier,
    ACTIONS(7796), 1,
      anon_sym_BANG,
    STATE(4069), 1,
      sym__instance_variable_name,
    STATE(5026), 1,
      sym_attribute,
    STATE(5231), 1,
      aux_sym_instance_variable_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7798), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [202017] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(5027), 1,
      sym_attribute,
    STATE(5526), 1,
      sym__attribute,
    STATE(6271), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202046] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3004), 1,
      anon_sym_PIPE,
    STATE(5028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3006), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202067] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7800), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(5029), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202096] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3008), 1,
      anon_sym_PIPE,
    STATE(5030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3010), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202117] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7316), 1,
      anon_sym_COMMA,
    ACTIONS(7318), 1,
      anon_sym_PIPE,
    ACTIONS(7322), 1,
      anon_sym_as,
    ACTIONS(7324), 1,
      anon_sym_COLON_COLON,
    STATE(5031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3126), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [202144] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7802), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(5032), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202173] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7804), 1,
      anon_sym_and,
    STATE(5033), 1,
      sym_attribute,
    STATE(5034), 1,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4932), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [202196] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7806), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5034), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(4685), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [202217] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    ACTIONS(7809), 1,
      anon_sym_RPAREN,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5035), 1,
      sym_attribute,
    STATE(6607), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202246] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7316), 1,
      anon_sym_COMMA,
    ACTIONS(7318), 1,
      anon_sym_PIPE,
    ACTIONS(7322), 1,
      anon_sym_as,
    ACTIONS(7324), 1,
      anon_sym_COLON_COLON,
    STATE(5036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3136), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [202273] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7804), 1,
      anon_sym_and,
    STATE(5033), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(5037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4946), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [202296] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3134), 1,
      anon_sym_COLON2,
    ACTIONS(3136), 1,
      anon_sym_RPAREN,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(5038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202325] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3138), 1,
      anon_sym_PIPE,
    STATE(5039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202346] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5071), 1,
      anon_sym_and,
    STATE(4864), 1,
      aux_sym_expression_item_repeat1,
    STATE(5040), 1,
      sym_attribute,
    STATE(5327), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5069), 2,
      anon_sym_in,
      sym_let_and_operator,
  [202373] = 9,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    ACTIONS(7811), 1,
      anon_sym_RBRACE,
    STATE(4670), 1,
      sym_field_path,
    STATE(5041), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202402] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5042), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5218), 2,
      anon_sym_and,
      anon_sym_in,
  [202426] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4769), 1,
      anon_sym_DQUOTE,
    ACTIONS(7813), 1,
      anon_sym_LBRACE,
    STATE(3365), 1,
      aux_sym_external_repeat1,
    STATE(5043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
  [202450] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(5044), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(5574), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202476] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7817), 1,
      anon_sym_SEMI,
    ACTIONS(7819), 1,
      anon_sym_GT_RBRACE,
    STATE(5045), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(5681), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202502] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5046), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5437), 2,
      anon_sym_and,
      anon_sym_in,
  [202526] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7396), 1,
      anon_sym_RPAREN,
    STATE(5047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202552] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7821), 1,
      anon_sym_SEMI,
    ACTIONS(7823), 1,
      anon_sym_GT_RBRACE,
    STATE(5048), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(5859), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202578] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4670), 1,
      sym_field_path,
    STATE(5049), 1,
      sym_attribute,
    STATE(5728), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202604] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5050), 1,
      sym_attribute,
    STATE(5068), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5636), 2,
      anon_sym_and,
      anon_sym_in,
  [202628] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7825), 1,
      anon_sym_EQ,
    STATE(5051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202654] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(5052), 1,
      sym_attribute,
    STATE(5082), 1,
      sym_field_path,
    STATE(6224), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202680] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7827), 1,
      anon_sym_SEMI,
    ACTIONS(7829), 1,
      anon_sym_GT_RBRACE,
    STATE(5053), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(5937), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202706] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5054), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5640), 2,
      anon_sym_and,
      anon_sym_in,
  [202730] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5055), 1,
      sym_attribute,
    STATE(6617), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202756] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5056), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5441), 2,
      anon_sym_and,
      anon_sym_in,
  [202780] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3574), 1,
      sym__uppercase_identifier,
    ACTIONS(7831), 1,
      anon_sym_LPAREN,
    STATE(4455), 1,
      sym__constructor_name,
    STATE(5057), 1,
      sym_attribute,
    STATE(5147), 1,
      sym_constructor_path,
    STATE(6422), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202806] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5058), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5441), 2,
      anon_sym_and,
      anon_sym_in,
  [202830] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5059), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5544), 2,
      anon_sym_and,
      anon_sym_in,
  [202854] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7833), 1,
      anon_sym_type,
    ACTIONS(7835), 1,
      anon_sym_module,
    STATE(5060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3331), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [202876] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5061), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4885), 2,
      anon_sym_and,
      anon_sym_in,
  [202900] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(5062), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(5831), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202926] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5063), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5650), 2,
      anon_sym_and,
      anon_sym_in,
  [202950] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(7839), 1,
      anon_sym_PIPE,
    ACTIONS(7841), 1,
      anon_sym_RBRACK,
    STATE(5064), 1,
      sym_attribute,
    STATE(5597), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202976] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5065), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5654), 2,
      anon_sym_and,
      anon_sym_in,
  [203000] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7843), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym__,
      aux_sym_type_variable_token1,
  [203018] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5067), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5658), 2,
      anon_sym_and,
      anon_sym_in,
  [203042] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5068), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5662), 2,
      anon_sym_and,
      anon_sym_in,
  [203066] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5069), 1,
      sym_attribute,
    STATE(5071), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5672), 2,
      anon_sym_and,
      anon_sym_in,
  [203090] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7845), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [203108] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5071), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5676), 2,
      anon_sym_and,
      anon_sym_in,
  [203132] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7847), 1,
      anon_sym_SEMI,
    ACTIONS(7849), 1,
      anon_sym_GT_RBRACE,
    STATE(5072), 1,
      sym_attribute,
    STATE(5543), 1,
      sym_instance_variable_expression,
    STATE(5674), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203158] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5073), 1,
      sym_attribute,
    STATE(5218), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5210), 2,
      anon_sym_and,
      anon_sym_in,
  [203182] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5074), 1,
      sym_attribute,
    STATE(6954), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203208] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5075), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5238), 2,
      anon_sym_and,
      anon_sym_in,
  [203232] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5076), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5242), 2,
      anon_sym_and,
      anon_sym_in,
  [203256] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5077), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5425), 2,
      anon_sym_and,
      anon_sym_in,
  [203280] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5078), 1,
      sym_attribute,
    STATE(6865), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203306] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7851), 1,
      anon_sym_type,
    ACTIONS(7853), 1,
      anon_sym_module,
    STATE(5079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3331), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [203328] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4972), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [203346] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7833), 1,
      anon_sym_type,
    ACTIONS(7835), 1,
      anon_sym_module,
    STATE(5081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3508), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [203368] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7855), 1,
      anon_sym_EQ,
    ACTIONS(7857), 1,
      anon_sym_COLON2,
    STATE(5082), 1,
      sym_attribute,
    STATE(5871), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7859), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [203392] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7607), 1,
      anon_sym_AMP,
    STATE(5083), 1,
      sym_attribute,
    STATE(5086), 1,
      aux_sym_tag_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7861), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [203414] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    STATE(5084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7863), 4,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_AMP,
  [203434] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5085), 1,
      sym_attribute,
    STATE(7017), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203460] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7865), 1,
      anon_sym_AMP,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5086), 2,
      sym_attribute,
      aux_sym_tag_specification_repeat1,
    ACTIONS(7863), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [203480] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5087), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5548), 2,
      anon_sym_and,
      anon_sym_in,
  [203504] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4670), 1,
      sym_field_path,
    STATE(5088), 1,
      sym_attribute,
    STATE(6003), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203530] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5089), 1,
      sym_attribute,
    STATE(5102), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5154), 2,
      anon_sym_and,
      anon_sym_in,
  [203554] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5090), 1,
      sym_attribute,
    STATE(5127), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5158), 2,
      anon_sym_and,
      anon_sym_in,
  [203578] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5091), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5162), 2,
      anon_sym_and,
      anon_sym_in,
  [203602] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7794), 1,
      sym__lowercase_identifier,
    STATE(4068), 1,
      sym__instance_variable_name,
    STATE(5092), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_instance_variable_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7798), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [203626] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5093), 1,
      sym_attribute,
    STATE(5111), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5264), 2,
      anon_sym_and,
      anon_sym_in,
  [203650] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(7839), 1,
      anon_sym_PIPE,
    ACTIONS(7868), 1,
      anon_sym_RBRACK,
    STATE(5094), 1,
      sym_attribute,
    STATE(5874), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203676] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5095), 1,
      sym_attribute,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4873), 2,
      anon_sym_and,
      anon_sym_in,
  [203700] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7440), 1,
      anon_sym_RPAREN,
    STATE(5096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203726] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7047), 1,
      anon_sym_as,
    STATE(5097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6631), 4,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
  [203746] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7870), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4649), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(5098), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
  [203768] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5099), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5415), 2,
      anon_sym_and,
      anon_sym_in,
  [203792] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(6579), 1,
      anon_sym_LBRACE,
    STATE(4609), 1,
      aux_sym_external_repeat1,
    STATE(5100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5538), 2,
      sym_string,
      sym_quoted_string,
  [203816] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7579), 1,
      anon_sym_RPAREN,
    STATE(5101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203842] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5102), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5268), 2,
      anon_sym_and,
      anon_sym_in,
  [203866] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7873), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [203884] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5104), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4910), 2,
      anon_sym_and,
      anon_sym_in,
  [203908] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4841), 1,
      sym__lowercase_identifier,
    ACTIONS(7875), 1,
      anon_sym_POUND,
    STATE(4459), 1,
      sym_type_constructor_path,
    STATE(5105), 1,
      sym_attribute,
    STATE(6168), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203934] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3566), 1,
      sym__uppercase_identifier,
    ACTIONS(7877), 1,
      anon_sym_LPAREN,
    STATE(3026), 1,
      sym__constructor_name,
    STATE(3586), 1,
      sym_constructor_path,
    STATE(5106), 1,
      sym_attribute,
    STATE(6479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203960] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7879), 1,
      anon_sym_SEMI,
    ACTIONS(7881), 1,
      anon_sym_GT_RBRACE,
    STATE(5107), 1,
      sym_attribute,
    STATE(5672), 1,
      sym_instance_variable_expression,
    STATE(5674), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203986] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7851), 1,
      anon_sym_type,
    ACTIONS(7853), 1,
      anon_sym_module,
    STATE(5108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3606), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [204008] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(5082), 1,
      sym_field_path,
    STATE(5109), 1,
      sym_attribute,
    STATE(5687), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204034] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5230), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [204052] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5111), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5445), 2,
      anon_sym_and,
      anon_sym_in,
  [204076] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7883), 1,
      anon_sym_SEMI,
    ACTIONS(7885), 1,
      anon_sym_GT_RBRACE,
    STATE(5112), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(5836), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204102] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7302), 1,
      anon_sym_COLON_GT,
    STATE(4245), 1,
      sym__coerced,
    STATE(5113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6639), 3,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [204124] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5114), 1,
      sym_attribute,
    STATE(6892), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204150] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5115), 1,
      sym_attribute,
    STATE(5130), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5449), 2,
      anon_sym_and,
      anon_sym_in,
  [204174] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5116), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5500), 2,
      anon_sym_and,
      anon_sym_in,
  [204198] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5117), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2095), 2,
      anon_sym_and,
      anon_sym_in,
  [204222] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5118), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5453), 2,
      anon_sym_and,
      anon_sym_in,
  [204246] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5119), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5214), 2,
      anon_sym_and,
      anon_sym_in,
  [204270] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5120), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5166), 2,
      anon_sym_and,
      anon_sym_in,
  [204294] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5121), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5552), 2,
      anon_sym_and,
      anon_sym_in,
  [204318] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2552), 1,
      sym__uppercase_identifier,
    ACTIONS(7877), 1,
      anon_sym_LPAREN,
    STATE(1771), 1,
      sym__constructor_name,
    STATE(3586), 1,
      sym_constructor_path,
    STATE(5122), 1,
      sym_attribute,
    STATE(7293), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204344] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4670), 1,
      sym_field_path,
    STATE(5123), 1,
      sym_attribute,
    STATE(5819), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204370] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3566), 1,
      sym__uppercase_identifier,
    ACTIONS(7887), 1,
      anon_sym_LPAREN,
    STATE(3026), 1,
      sym__constructor_name,
    STATE(3586), 1,
      sym_constructor_path,
    STATE(5124), 1,
      sym_attribute,
    STATE(6480), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204396] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7607), 1,
      anon_sym_AMP,
    STATE(5086), 1,
      aux_sym_tag_specification_repeat1,
    STATE(5125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7605), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [204418] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5126), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4877), 2,
      anon_sym_and,
      anon_sym_in,
  [204442] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5127), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5276), 2,
      anon_sym_and,
      anon_sym_in,
  [204466] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5128), 1,
      sym_attribute,
    STATE(5145), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5280), 2,
      anon_sym_and,
      anon_sym_in,
  [204490] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5121), 1,
      aux_sym_expression_item_repeat1,
    STATE(5129), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5504), 2,
      anon_sym_and,
      anon_sym_in,
  [204514] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5130), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5508), 2,
      anon_sym_and,
      anon_sym_in,
  [204538] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5131), 1,
      sym_attribute,
    STATE(5152), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5284), 2,
      anon_sym_and,
      anon_sym_in,
  [204562] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5132), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5457), 2,
      anon_sym_and,
      anon_sym_in,
  [204586] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5133), 1,
      sym_attribute,
    STATE(6971), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204612] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5134), 1,
      sym_attribute,
    STATE(5151), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5512), 2,
      anon_sym_and,
      anon_sym_in,
  [204636] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1894), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    STATE(5135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1896), 3,
      anon_sym_COMMA,
      anon_sym_with,
      sym_hash_operator,
  [204658] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(7891), 1,
      anon_sym_LBRACE,
    STATE(3745), 1,
      aux_sym_external_repeat1,
    STATE(5136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
  [204682] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5137), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2111), 2,
      anon_sym_and,
      anon_sym_in,
  [204706] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5138), 1,
      sym_attribute,
    STATE(5264), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5516), 2,
      anon_sym_and,
      anon_sym_in,
  [204730] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7150), 1,
      anon_sym_RPAREN,
    STATE(5139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204756] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7893), 1,
      anon_sym_EQ,
    STATE(5140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204782] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4379), 1,
      sym__lowercase_identifier,
    ACTIONS(7895), 1,
      anon_sym_POUND,
    STATE(3125), 1,
      sym_type_constructor_path,
    STATE(5141), 1,
      sym_attribute,
    STATE(5948), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204808] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5142), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4964), 2,
      anon_sym_and,
      anon_sym_in,
  [204832] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5143), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5520), 2,
      anon_sym_and,
      anon_sym_in,
  [204856] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7897), 1,
      anon_sym_type,
    ACTIONS(7899), 1,
      anon_sym_module,
    STATE(5144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4660), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [204878] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5145), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5461), 2,
      anon_sym_and,
      anon_sym_in,
  [204902] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(5146), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(5637), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204928] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4359), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [204946] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5148), 1,
      sym_attribute,
    STATE(5206), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5556), 2,
      anon_sym_and,
      anon_sym_in,
  [204970] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5149), 1,
      sym_attribute,
    STATE(5168), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5524), 2,
      anon_sym_and,
      anon_sym_in,
  [204994] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5150), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5170), 2,
      anon_sym_and,
      anon_sym_in,
  [205018] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5151), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5560), 2,
      anon_sym_and,
      anon_sym_in,
  [205042] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5152), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5465), 2,
      anon_sym_and,
      anon_sym_in,
  [205066] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4670), 1,
      sym_field_path,
    STATE(5153), 1,
      sym_attribute,
    STATE(5921), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205092] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5154), 1,
      sym_attribute,
    STATE(5199), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5469), 2,
      anon_sym_and,
      anon_sym_in,
  [205116] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7901), 1,
      anon_sym_SEMI,
    ACTIONS(7903), 1,
      anon_sym_GT_RBRACE,
    STATE(5155), 1,
      sym_attribute,
    STATE(5635), 1,
      sym_instance_variable_expression,
    STATE(5674), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205142] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7905), 1,
      anon_sym_SEMI,
    ACTIONS(7907), 1,
      anon_sym_GT_RBRACE,
    STATE(5156), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(5791), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205168] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1474), 1,
      anon_sym_DQUOTE,
    ACTIONS(6579), 1,
      anon_sym_LBRACE,
    STATE(4578), 1,
      aux_sym_external_repeat1,
    STATE(5157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5538), 2,
      sym_string,
      sym_quoted_string,
  [205192] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5158), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4861), 2,
      anon_sym_and,
      anon_sym_in,
  [205216] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5159), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5288), 2,
      anon_sym_and,
      anon_sym_in,
  [205240] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7794), 1,
      sym__lowercase_identifier,
    STATE(5160), 1,
      sym_attribute,
    STATE(5203), 1,
      aux_sym_instance_variable_specification_repeat1,
    STATE(6012), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7798), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [205264] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4670), 1,
      sym_field_path,
    STATE(5161), 1,
      sym_attribute,
    STATE(5724), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205290] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5162), 1,
      sym_attribute,
    STATE(6533), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205316] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4670), 1,
      sym_field_path,
    STATE(5163), 1,
      sym_attribute,
    STATE(5757), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205342] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5164), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5528), 2,
      anon_sym_and,
      anon_sym_in,
  [205366] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5165), 1,
      sym_attribute,
    STATE(7309), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205392] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5166), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5144), 2,
      anon_sym_and,
      anon_sym_in,
  [205416] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5167), 1,
      sym_attribute,
    STATE(5210), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5568), 2,
      anon_sym_and,
      anon_sym_in,
  [205440] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5168), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5572), 2,
      anon_sym_and,
      anon_sym_in,
  [205464] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5169), 1,
      sym_attribute,
    STATE(5216), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5576), 2,
      anon_sym_and,
      anon_sym_in,
  [205488] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5170), 1,
      sym_attribute,
    STATE(5221), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5580), 2,
      anon_sym_and,
      anon_sym_in,
  [205512] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(5082), 1,
      sym_field_path,
    STATE(5171), 1,
      sym_attribute,
    STATE(5915), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205538] = 7,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7909), 1,
      anon_sym_LBRACE,
    STATE(3579), 1,
      aux_sym_external_repeat1,
    STATE(5172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
  [205562] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7911), 1,
      anon_sym_SEMI,
    ACTIONS(7913), 1,
      anon_sym_GT_RBRACE,
    STATE(5173), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(5864), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205588] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5174), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5584), 2,
      anon_sym_and,
      anon_sym_in,
  [205612] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5175), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4517), 2,
      anon_sym_and,
      anon_sym_in,
  [205636] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7915), 1,
      anon_sym_DASH_GT,
    ACTIONS(7917), 1,
      anon_sym_with,
    STATE(5176), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5098), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [205658] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7915), 1,
      anon_sym_DASH_GT,
    ACTIONS(7917), 1,
      anon_sym_with,
    STATE(5177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5110), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [205680] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7915), 1,
      anon_sym_DASH_GT,
    ACTIONS(7917), 1,
      anon_sym_with,
    STATE(5178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4980), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [205702] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5179), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7919), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [205720] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7915), 1,
      anon_sym_DASH_GT,
    ACTIONS(7917), 1,
      anon_sym_with,
    STATE(5180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5016), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [205742] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(5181), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(5692), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205768] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5182), 1,
      sym_attribute,
    STATE(5255), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5588), 2,
      anon_sym_and,
      anon_sym_in,
  [205792] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5183), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2083), 2,
      anon_sym_and,
      anon_sym_in,
  [205816] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5184), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5592), 2,
      anon_sym_and,
      anon_sym_in,
  [205840] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4670), 1,
      sym_field_path,
    STATE(5185), 1,
      sym_attribute,
    STATE(5593), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205866] = 7,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7909), 1,
      anon_sym_LBRACE,
    STATE(3578), 1,
      aux_sym_external_repeat1,
    STATE(5186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
  [205890] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5187), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5132), 2,
      anon_sym_and,
      anon_sym_in,
  [205914] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5188), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5532), 2,
      anon_sym_and,
      anon_sym_in,
  [205938] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5189), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5596), 2,
      anon_sym_and,
      anon_sym_in,
  [205962] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5150), 1,
      aux_sym_expression_item_repeat1,
    STATE(5190), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5320), 2,
      anon_sym_and,
      anon_sym_in,
  [205986] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5191), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5481), 2,
      anon_sym_and,
      anon_sym_in,
  [206010] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(5192), 1,
      sym_attribute,
    STATE(5235), 1,
      sym_field_path,
    STATE(5926), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206036] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7897), 1,
      anon_sym_type,
    ACTIONS(7899), 1,
      anon_sym_module,
    STATE(5193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4680), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [206058] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4769), 1,
      anon_sym_DQUOTE,
    ACTIONS(7813), 1,
      anon_sym_LBRACE,
    STATE(3457), 1,
      aux_sym_external_repeat1,
    STATE(5194), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
  [206082] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5195), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5202), 2,
      anon_sym_and,
      anon_sym_in,
  [206106] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5196), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2073), 2,
      anon_sym_and,
      anon_sym_in,
  [206130] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5197), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2133), 2,
      anon_sym_and,
      anon_sym_in,
  [206154] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5198), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5429), 2,
      anon_sym_and,
      anon_sym_in,
  [206178] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5199), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5536), 2,
      anon_sym_and,
      anon_sym_in,
  [206202] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5083), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [206220] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5201), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5148), 2,
      anon_sym_and,
      anon_sym_in,
  [206244] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7420), 1,
      anon_sym_RPAREN,
    STATE(5202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206270] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7794), 1,
      sym__lowercase_identifier,
    STATE(5203), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_instance_variable_specification_repeat1,
    STATE(6261), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7798), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [206294] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5204), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5600), 2,
      anon_sym_and,
      anon_sym_in,
  [206318] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4363), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [206336] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5206), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5604), 2,
      anon_sym_and,
      anon_sym_in,
  [206360] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4670), 1,
      sym_field_path,
    STATE(5207), 1,
      sym_attribute,
    STATE(5565), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206386] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(7839), 1,
      anon_sym_PIPE,
    ACTIONS(7921), 1,
      anon_sym_RBRACK,
    STATE(5208), 1,
      sym_attribute,
    STATE(5779), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206412] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5063), 1,
      aux_sym_expression_item_repeat1,
    STATE(5209), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5608), 2,
      anon_sym_and,
      anon_sym_in,
  [206436] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5210), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5612), 2,
      anon_sym_and,
      anon_sym_in,
  [206460] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5211), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4894), 2,
      anon_sym_and,
      anon_sym_in,
  [206484] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5075), 1,
      aux_sym_expression_item_repeat1,
    STATE(5212), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5136), 2,
      anon_sym_and,
      anon_sym_in,
  [206508] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5213), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5226), 2,
      anon_sym_and,
      anon_sym_in,
  [206532] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5076), 1,
      aux_sym_expression_item_repeat1,
    STATE(5214), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5128), 2,
      anon_sym_and,
      anon_sym_in,
  [206556] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5215), 1,
      sym_attribute,
    STATE(6739), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206582] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5216), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5616), 2,
      anon_sym_and,
      anon_sym_in,
  [206606] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5217), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4902), 2,
      anon_sym_and,
      anon_sym_in,
  [206630] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5218), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5140), 2,
      anon_sym_and,
      anon_sym_in,
  [206654] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5219), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5249), 2,
      anon_sym_and,
      anon_sym_in,
  [206678] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7923), 1,
      anon_sym_SEMI,
    ACTIONS(7925), 1,
      anon_sym_GT_RBRACE,
    STATE(5220), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(5690), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206704] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5221), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5620), 2,
      anon_sym_and,
      anon_sym_in,
  [206728] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5222), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4906), 2,
      anon_sym_and,
      anon_sym_in,
  [206752] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5065), 1,
      aux_sym_expression_item_repeat1,
    STATE(5223), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5624), 2,
      anon_sym_and,
      anon_sym_in,
  [206776] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4670), 1,
      sym_field_path,
    STATE(5224), 1,
      sym_attribute,
    STATE(5651), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206802] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5225), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4942), 2,
      anon_sym_and,
      anon_sym_in,
  [206826] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7927), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [206844] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7794), 1,
      sym__lowercase_identifier,
    STATE(4060), 1,
      sym__instance_variable_name,
    STATE(5092), 1,
      aux_sym_instance_variable_specification_repeat1,
    STATE(5227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7798), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [206868] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3961), 1,
      anon_sym_LPAREN,
    ACTIONS(7929), 1,
      sym__lowercase_identifier,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    STATE(3541), 1,
      sym_parenthesized_operator,
    STATE(4127), 1,
      sym__value_name,
    STATE(5228), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206894] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(5082), 1,
      sym_field_path,
    STATE(5229), 1,
      sym_attribute,
    STATE(5726), 1,
      sym_field_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206920] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    ACTIONS(7933), 1,
      anon_sym_POUND,
    STATE(2620), 1,
      sym_type_constructor_path,
    STATE(5230), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206946] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7794), 1,
      sym__lowercase_identifier,
    STATE(4060), 1,
      sym__instance_variable_name,
    STATE(5231), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_instance_variable_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7798), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [206970] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5232), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5496), 2,
      anon_sym_and,
      anon_sym_in,
  [206994] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7935), 1,
      sym__lowercase_identifier,
    ACTIONS(7939), 1,
      anon_sym_BANG,
    STATE(5233), 1,
      sym_attribute,
    STATE(5456), 1,
      aux_sym_method_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7937), 2,
      anon_sym_private,
      anon_sym_virtual,
  [207018] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5234), 1,
      sym_attribute,
    STATE(5241), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5174), 2,
      anon_sym_and,
      anon_sym_in,
  [207042] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7857), 1,
      anon_sym_COLON2,
    ACTIONS(7941), 1,
      anon_sym_EQ,
    STATE(5235), 1,
      sym_attribute,
    STATE(5794), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7943), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [207066] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7945), 1,
      anon_sym_EQ,
    STATE(5236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207092] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7105), 1,
      anon_sym_RPAREN,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    STATE(5237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207118] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5238), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5249), 2,
      anon_sym_and,
      anon_sym_in,
  [207142] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5239), 1,
      sym_attribute,
    STATE(5251), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5178), 2,
      anon_sym_and,
      anon_sym_in,
  [207166] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5240), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5433), 2,
      anon_sym_and,
      anon_sym_in,
  [207190] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5241), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5304), 2,
      anon_sym_and,
      anon_sym_in,
  [207214] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7947), 1,
      anon_sym_EQ,
    STATE(5242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207240] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5243), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5182), 2,
      anon_sym_and,
      anon_sym_in,
  [207264] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7480), 1,
      anon_sym_RPAREN,
    STATE(5244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207290] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7949), 1,
      anon_sym_EQ,
    STATE(5245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207316] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7951), 1,
      anon_sym_EQ,
    STATE(5246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207342] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7953), 1,
      anon_sym_EQ,
    STATE(5247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207368] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    ACTIONS(7955), 1,
      anon_sym_EQ,
    STATE(5248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207394] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(5235), 1,
      sym_field_path,
    STATE(5249), 1,
      sym_attribute,
    STATE(6004), 1,
      sym_field_binding_pattern,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207420] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(6379), 1,
      anon_sym_COLON2,
    STATE(4247), 1,
      sym__typed,
    STATE(4248), 1,
      sym__coerced,
    STATE(5250), 1,
      sym_attribute,
    STATE(6406), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207446] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5251), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5312), 2,
      anon_sym_and,
      anon_sym_in,
  [207470] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5191), 1,
      aux_sym_expression_item_repeat1,
    STATE(5252), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5316), 2,
      anon_sym_and,
      anon_sym_in,
  [207494] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7851), 1,
      anon_sym_type,
    ACTIONS(7957), 1,
      anon_sym_module,
    STATE(5253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3331), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [207516] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1530), 1,
      anon_sym_DQUOTE,
    ACTIONS(7891), 1,
      anon_sym_LBRACE,
    STATE(3749), 1,
      aux_sym_external_repeat1,
    STATE(5254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3797), 2,
      sym_string,
      sym_quoted_string,
  [207540] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5255), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5628), 2,
      anon_sym_and,
      anon_sym_in,
  [207564] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4670), 1,
      sym_field_path,
    STATE(5256), 1,
      sym_attribute,
    STATE(5605), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207590] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7332), 1,
      anon_sym_RPAREN,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(5257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207616] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5067), 1,
      aux_sym_expression_item_repeat1,
    STATE(5258), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5632), 2,
      anon_sym_and,
      anon_sym_in,
  [207640] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7851), 1,
      anon_sym_type,
    ACTIONS(7957), 1,
      anon_sym_module,
    STATE(5259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5037), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [207662] = 8,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7214), 1,
      sym__lowercase_identifier,
    STATE(4670), 1,
      sym_field_path,
    STATE(5260), 1,
      sym_attribute,
    STATE(5703), 1,
      sym_field_expression,
    STATE(6694), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207688] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5261), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5253), 2,
      anon_sym_and,
      anon_sym_in,
  [207712] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5046), 1,
      aux_sym_expression_item_repeat1,
    STATE(5262), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5234), 2,
      anon_sym_and,
      anon_sym_in,
  [207736] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4976), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [207754] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5264), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5564), 2,
      anon_sym_and,
      anon_sym_in,
  [207778] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(7959), 1,
      anon_sym_DASH_GT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5265), 1,
      sym_attribute,
    STATE(5393), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207801] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7961), 1,
      sym__lowercase_identifier,
    STATE(2471), 1,
      sym_type_constructor_path,
    STATE(5266), 1,
      sym_attribute,
    STATE(6313), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207824] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(7963), 1,
      anon_sym_RBRACK,
    STATE(5267), 1,
      sym_attribute,
    STATE(5395), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207847] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7965), 1,
      sym__lowercase_identifier,
    STATE(3417), 1,
      sym_class_type_path,
    STATE(5268), 1,
      sym_attribute,
    STATE(6330), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207870] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(7967), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207893] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3971), 1,
      anon_sym_DQUOTE,
    ACTIONS(7969), 1,
      anon_sym_LBRACE,
    STATE(5270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3755), 2,
      sym_string,
      sym_quoted_string,
  [207914] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5271), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5756), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [207931] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7971), 1,
      sym__lowercase_identifier,
    STATE(4472), 1,
      sym_class_type_path,
    STATE(5272), 1,
      sym_attribute,
    STATE(6074), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207954] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7686), 1,
      sym__lowercase_identifier,
    ACTIONS(7688), 1,
      anon_sym_LBRACK,
    ACTIONS(7690), 1,
      anon_sym_virtual,
    STATE(5273), 1,
      sym_attribute,
    STATE(5636), 1,
      sym_class_type_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207977] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7973), 1,
      anon_sym_GT_RBRACE,
    STATE(5274), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208000] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7698), 1,
      sym__lowercase_identifier,
    STATE(1794), 1,
      sym_class_path,
    STATE(5275), 1,
      sym_attribute,
    STATE(6420), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208023] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7975), 1,
      anon_sym_GT_RBRACE,
    STATE(5276), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208046] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(5277), 1,
      sym_attribute,
    STATE(6251), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208069] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7977), 1,
      sym__lowercase_identifier,
    STATE(2112), 1,
      sym_type_constructor_path,
    STATE(5278), 1,
      sym_attribute,
    STATE(6034), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208092] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7599), 1,
      sym__lowercase_identifier,
    STATE(2683), 1,
      sym_class_path,
    STATE(5279), 1,
      sym_attribute,
    STATE(6712), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208115] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7979), 1,
      sym__lowercase_identifier,
    STATE(5280), 1,
      sym_attribute,
    STATE(5355), 1,
      aux_sym_method_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7937), 2,
      anon_sym_private,
      anon_sym_virtual,
  [208136] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7456), 1,
      sym__lowercase_identifier,
    ACTIONS(7460), 1,
      anon_sym_LBRACK,
    ACTIONS(7462), 1,
      anon_sym_virtual,
    STATE(5281), 1,
      sym_attribute,
    STATE(6185), 1,
      sym_class_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208159] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7981), 1,
      sym__lowercase_identifier,
    STATE(3068), 1,
      sym_class_type_path,
    STATE(5282), 1,
      sym_attribute,
    STATE(6201), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208182] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7983), 1,
      sym__lowercase_identifier,
    STATE(5283), 1,
      sym_attribute,
    STATE(5404), 1,
      aux_sym__abstract_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7985), 2,
      anon_sym_RPAREN,
      anon_sym_DOT,
  [208203] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7987), 1,
      sym__lowercase_identifier,
    STATE(3417), 1,
      sym_class_type_path,
    STATE(5284), 1,
      sym_attribute,
    STATE(6229), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208226] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7622), 1,
      sym__lowercase_identifier,
    ACTIONS(7624), 1,
      anon_sym_LBRACK,
    ACTIONS(7626), 1,
      anon_sym_virtual,
    STATE(3869), 1,
      sym_class_type_binding,
    STATE(5285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208249] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4791), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [208266] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5287), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5300), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [208283] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5730), 1,
      anon_sym_in,
    ACTIONS(7989), 1,
      anon_sym_and,
    ACTIONS(7991), 1,
      sym_let_and_operator,
    STATE(5288), 1,
      sym_attribute,
    STATE(5441), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208306] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7981), 1,
      sym__lowercase_identifier,
    STATE(3115), 1,
      sym_class_type_path,
    STATE(5289), 1,
      sym_attribute,
    STATE(6201), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208329] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(7993), 1,
      anon_sym_GT_RBRACE,
    STATE(5290), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208352] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7981), 1,
      sym__lowercase_identifier,
    STATE(3071), 1,
      sym_class_type_path,
    STATE(5291), 1,
      sym_attribute,
    STATE(6201), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208375] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7997), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7995), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    STATE(5292), 2,
      sym_attribute,
      aux_sym__type_params_repeat1,
  [208394] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7613), 1,
      sym__lowercase_identifier,
    STATE(2368), 1,
      sym_class_path,
    STATE(5293), 1,
      sym_attribute,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208417] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5700), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5294), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208440] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8000), 1,
      anon_sym_GT_RBRACE,
    STATE(5295), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208463] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4843), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5296), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208486] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8004), 1,
      anon_sym_RBRACK,
    ACTIONS(8006), 1,
      anon_sym_GT,
    STATE(5297), 1,
      sym_attribute,
    STATE(5347), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208509] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8004), 1,
      anon_sym_RBRACK,
    ACTIONS(8006), 1,
      anon_sym_GT,
    STATE(5298), 1,
      sym_attribute,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208532] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5540), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [208549] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4735), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [208566] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7640), 1,
      sym__lowercase_identifier,
    STATE(1955), 1,
      sym_class_path,
    STATE(5301), 1,
      sym_attribute,
    STATE(6744), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208589] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4739), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [208606] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7971), 1,
      sym__lowercase_identifier,
    STATE(4489), 1,
      sym_class_type_path,
    STATE(5303), 1,
      sym_attribute,
    STATE(6074), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208629] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8008), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208652] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8010), 1,
      sym__lowercase_identifier,
    STATE(5280), 1,
      aux_sym_method_specification_repeat1,
    STATE(5305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7937), 2,
      anon_sym_private,
      anon_sym_virtual,
  [208673] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5704), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5306), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208696] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8012), 1,
      anon_sym_GT_RBRACE,
    STATE(5307), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208719] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7747), 1,
      sym__lowercase_identifier,
    STATE(902), 1,
      sym_class_path,
    STATE(5308), 1,
      sym_attribute,
    STATE(6677), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208742] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8014), 1,
      anon_sym_GT_RBRACE,
    STATE(5309), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208765] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4825), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [208782] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8016), 1,
      anon_sym_DOT,
    STATE(5311), 1,
      sym_attribute,
    STATE(5389), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6060), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208805] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1082), 1,
      anon_sym_DQUOTE,
    ACTIONS(6593), 1,
      anon_sym_LBRACE,
    STATE(5312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5271), 2,
      sym_string,
      sym_quoted_string,
  [208826] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5776), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5313), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208849] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8018), 1,
      anon_sym_DOT,
    STATE(5314), 1,
      sym_attribute,
    STATE(5389), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6060), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208872] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7686), 1,
      sym__lowercase_identifier,
    ACTIONS(7688), 1,
      anon_sym_LBRACK,
    ACTIONS(7690), 1,
      anon_sym_virtual,
    STATE(5315), 1,
      sym_attribute,
    STATE(6218), 1,
      sym_class_type_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208895] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7971), 1,
      sym__lowercase_identifier,
    STATE(4495), 1,
      sym_class_type_path,
    STATE(5316), 1,
      sym_attribute,
    STATE(6074), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208918] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8020), 1,
      anon_sym_GT_RBRACE,
    STATE(5317), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208941] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8022), 1,
      anon_sym_GT_RBRACE,
    STATE(5318), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208964] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    STATE(4385), 1,
      sym_type_constructor_path,
    STATE(5319), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208987] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7981), 1,
      sym__lowercase_identifier,
    STATE(3075), 1,
      sym_class_type_path,
    STATE(5320), 1,
      sym_attribute,
    STATE(6201), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209010] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5321), 1,
      sym_attribute,
    STATE(5954), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8024), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [209031] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(8026), 1,
      sym__lowercase_identifier,
    STATE(5322), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6856), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209054] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(5323), 1,
      sym_attribute,
    STATE(6105), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209077] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5824), 1,
      anon_sym_in,
    ACTIONS(7989), 1,
      anon_sym_and,
    ACTIONS(7991), 1,
      sym_let_and_operator,
    STATE(5324), 1,
      sym_attribute,
    STATE(5441), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209100] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8028), 1,
      anon_sym_DOT,
    STATE(5325), 1,
      sym_attribute,
    STATE(5389), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6060), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209123] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8030), 1,
      sym__lowercase_identifier,
    ACTIONS(8032), 1,
      anon_sym_LPAREN,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(3089), 1,
      sym__value_name,
    STATE(5326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209146] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4920), 1,
      anon_sym_and,
    STATE(5327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4918), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [209165] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8034), 1,
      sym__lowercase_identifier,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8036), 2,
      anon_sym_mutable,
      anon_sym_virtual,
    STATE(5328), 2,
      sym_attribute,
      aux_sym_instance_variable_specification_repeat1,
  [209184] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5818), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5329), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209207] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    ACTIONS(8039), 1,
      sym__lowercase_identifier,
    STATE(5330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4428), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [209228] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5338), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [209245] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(8041), 1,
      sym__lowercase_identifier,
    STATE(2174), 1,
      sym_type_constructor_path,
    STATE(5332), 1,
      sym_attribute,
    STATE(6180), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209268] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8043), 1,
      anon_sym_RBRACK,
    STATE(5333), 1,
      sym_attribute,
    STATE(5861), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209291] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5342), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [209308] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4779), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [209325] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7987), 1,
      sym__lowercase_identifier,
    STATE(2619), 1,
      sym_class_type_path,
    STATE(5336), 1,
      sym_attribute,
    STATE(6229), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209348] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    STATE(5337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5004), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [209369] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7971), 1,
      sym__lowercase_identifier,
    STATE(4458), 1,
      sym_class_type_path,
    STATE(5338), 1,
      sym_attribute,
    STATE(6074), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209392] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8047), 1,
      anon_sym_RPAREN,
    STATE(5339), 1,
      sym_attribute,
    STATE(5594), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209415] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3728), 1,
      anon_sym_RBRACK,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8049), 1,
      anon_sym_GT,
    STATE(5340), 1,
      sym_attribute,
    STATE(5361), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209438] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    STATE(5341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8051), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [209457] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5342), 1,
      sym_attribute,
    STATE(6307), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8053), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [209478] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8055), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8051), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    STATE(5343), 2,
      sym_attribute,
      aux_sym_instantiated_class_type_repeat1,
  [209497] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8058), 1,
      sym__lowercase_identifier,
    ACTIONS(8060), 1,
      anon_sym_LPAREN,
    STATE(5344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4966), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [209518] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8062), 1,
      sym__lowercase_identifier,
    STATE(5345), 1,
      sym_attribute,
    STATE(5355), 1,
      aux_sym_method_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7937), 2,
      anon_sym_private,
      anon_sym_virtual,
  [209539] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8064), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8067), 2,
      anon_sym_RBRACK,
      anon_sym_GT,
    STATE(5346), 2,
      sym_attribute,
      aux_sym_polymorphic_variant_type_repeat1,
  [209558] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8069), 1,
      anon_sym_RBRACK,
    ACTIONS(8071), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209581] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(8069), 1,
      anon_sym_RBRACK,
    STATE(5348), 1,
      sym_attribute,
    STATE(5395), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209604] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8073), 1,
      anon_sym_RPAREN,
    STATE(5349), 1,
      sym_attribute,
    STATE(5712), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209627] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8075), 1,
      anon_sym_GT_RBRACE,
    STATE(5350), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209650] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    STATE(5351), 1,
      sym_attribute,
    STATE(5971), 1,
      sym_type_constructor_path,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209673] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(8077), 1,
      sym__lowercase_identifier,
    STATE(5352), 1,
      sym_attribute,
    STATE(5976), 1,
      sym_module_type_path,
    STATE(6058), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209696] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3678), 1,
      anon_sym_RBRACK,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8079), 1,
      anon_sym_GT,
    STATE(5353), 1,
      sym_attribute,
    STATE(5433), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209719] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8081), 1,
      anon_sym_RBRACK,
    STATE(5354), 1,
      sym_attribute,
    STATE(5595), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209742] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8083), 1,
      sym__lowercase_identifier,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8085), 2,
      anon_sym_private,
      anon_sym_virtual,
    STATE(5355), 2,
      sym_attribute,
      aux_sym_method_specification_repeat1,
  [209761] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8060), 1,
      anon_sym_LPAREN,
    ACTIONS(8088), 1,
      sym__lowercase_identifier,
    STATE(4850), 1,
      sym__value_name,
    STATE(4894), 1,
      sym_parenthesized_operator,
    STATE(5356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209784] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7644), 1,
      sym__lowercase_identifier,
    STATE(1097), 1,
      sym_class_path,
    STATE(5357), 1,
      sym_attribute,
    STATE(6942), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209807] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7613), 1,
      sym__lowercase_identifier,
    STATE(829), 1,
      sym_class_path,
    STATE(5358), 1,
      sym_attribute,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209830] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(7841), 1,
      anon_sym_RBRACK,
    STATE(5359), 1,
      sym_attribute,
    STATE(5597), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209853] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8090), 1,
      anon_sym_RBRACK,
    ACTIONS(8092), 1,
      anon_sym_GT,
    STATE(5360), 1,
      sym_attribute,
    STATE(5368), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209876] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8090), 1,
      anon_sym_RBRACK,
    ACTIONS(8092), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209899] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    ACTIONS(8039), 1,
      sym__lowercase_identifier,
    STATE(5362), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4345), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [209920] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5363), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3939), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [209937] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7430), 1,
      sym__lowercase_identifier,
    ACTIONS(7434), 1,
      anon_sym_LBRACK,
    ACTIONS(7436), 1,
      anon_sym_virtual,
    STATE(4163), 1,
      sym_class_binding,
    STATE(5364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209960] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8094), 1,
      anon_sym_GT_RBRACE,
    STATE(5365), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209983] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8096), 1,
      anon_sym_DOT,
    STATE(5366), 1,
      sym_attribute,
    STATE(5389), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6060), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210006] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(8098), 1,
      sym__lowercase_identifier,
    STATE(5367), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(6478), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210029] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8100), 1,
      anon_sym_RBRACK,
    ACTIONS(8102), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5368), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210052] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(8100), 1,
      anon_sym_RBRACK,
    STATE(5369), 1,
      sym_attribute,
    STATE(5395), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210075] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8104), 1,
      anon_sym_DOT,
    STATE(5370), 1,
      sym_attribute,
    STATE(5389), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6060), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210098] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8106), 1,
      anon_sym_GT_RBRACE,
    STATE(5371), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210121] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4900), 1,
      anon_sym_and,
    STATE(5372), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4898), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [210140] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8108), 1,
      sym__lowercase_identifier,
    ACTIONS(8110), 1,
      anon_sym_RBRACE,
    ACTIONS(8112), 1,
      anon_sym_mutable,
    STATE(5373), 1,
      sym_attribute,
    STATE(6114), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210163] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(8114), 1,
      anon_sym_RBRACK,
    STATE(5374), 1,
      sym_attribute,
    STATE(5395), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210186] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7047), 1,
      anon_sym_as,
    STATE(5375), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4968), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [210205] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7047), 1,
      anon_sym_as,
    STATE(5376), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4960), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [210224] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5738), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5377), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210247] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7430), 1,
      sym__lowercase_identifier,
    ACTIONS(7434), 1,
      anon_sym_LBRACK,
    ACTIONS(7436), 1,
      anon_sym_virtual,
    STATE(3941), 1,
      sym_class_binding,
    STATE(5378), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210270] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(8116), 1,
      anon_sym_RBRACK,
    STATE(5379), 1,
      sym_attribute,
    STATE(5395), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210293] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2944), 1,
      anon_sym_COMMA,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    ACTIONS(8118), 1,
      sym_hash_operator,
    STATE(5380), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210316] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4940), 1,
      anon_sym_and,
    STATE(5381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4938), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [210335] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7613), 1,
      sym__lowercase_identifier,
    STATE(2104), 1,
      sym_class_path,
    STATE(5382), 1,
      sym_attribute,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210358] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(5383), 1,
      sym_attribute,
    STATE(6196), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210381] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7666), 1,
      sym__lowercase_identifier,
    STATE(1885), 1,
      sym_class_path,
    STATE(5384), 1,
      sym_attribute,
    STATE(6510), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210404] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7965), 1,
      sym__lowercase_identifier,
    STATE(3358), 1,
      sym_class_type_path,
    STATE(5385), 1,
      sym_attribute,
    STATE(6330), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210427] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8120), 1,
      anon_sym_GT_RBRACE,
    STATE(5386), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210450] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3760), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [210467] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5726), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5388), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210490] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8122), 1,
      anon_sym_DOT,
    ACTIONS(8124), 1,
      aux_sym_type_variable_token1,
    STATE(6060), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5389), 2,
      sym_attribute,
      aux_sym_constructor_declaration_repeat1,
  [210511] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8127), 1,
      sym__lowercase_identifier,
    ACTIONS(8129), 1,
      anon_sym_LPAREN,
    STATE(2043), 1,
      sym__value_name,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(5390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210534] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    ACTIONS(8131), 1,
      anon_sym_EQ,
    ACTIONS(8133), 1,
      anon_sym_RPAREN,
    STATE(5391), 1,
      sym_attribute,
    STATE(5995), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210557] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7987), 1,
      sym__lowercase_identifier,
    STATE(2637), 1,
      sym_class_type_path,
    STATE(5392), 1,
      sym_attribute,
    STATE(6229), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210580] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8135), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5393), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210603] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(8137), 1,
      anon_sym_RBRACK,
    STATE(5394), 1,
      sym_attribute,
    STATE(5395), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210626] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8139), 1,
      anon_sym_RBRACK,
    ACTIONS(8141), 1,
      aux_sym_tag_token1,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5395), 2,
      sym_attribute,
      aux_sym_polymorphic_variant_type_repeat2,
  [210647] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6853), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    STATE(5396), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210670] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6893), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    STATE(5397), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210693] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6897), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    STATE(5398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210716] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4775), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [210733] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    ACTIONS(8144), 1,
      anon_sym_EQ,
    ACTIONS(8146), 1,
      anon_sym_RPAREN,
    STATE(5400), 1,
      sym_attribute,
    STATE(6289), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210756] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5712), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5401), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210779] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    ACTIONS(8039), 1,
      sym__lowercase_identifier,
    STATE(5402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4323), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [210800] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4841), 1,
      sym__lowercase_identifier,
    STATE(4385), 1,
      sym_type_constructor_path,
    STATE(5403), 1,
      sym_attribute,
    STATE(6168), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210823] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8148), 1,
      sym__lowercase_identifier,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8151), 2,
      anon_sym_RPAREN,
      anon_sym_DOT,
    STATE(5404), 2,
      sym_attribute,
      aux_sym__abstract_type_repeat1,
  [210842] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5405), 1,
      sym_attribute,
    STATE(6098), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8153), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [210863] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    ACTIONS(8155), 1,
      sym__lowercase_identifier,
    STATE(4553), 1,
      sym__value_name,
    STATE(4689), 1,
      sym_parenthesized_operator,
    STATE(5406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210886] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4723), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5407), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210909] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7047), 1,
      anon_sym_as,
    STATE(5408), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6685), 3,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [210928] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5760), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5409), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210951] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4731), 1,
      anon_sym_in,
    STATE(5296), 1,
      aux_sym_expression_item_repeat1,
    STATE(5410), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210974] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3184), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(5411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210997] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3200), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(5412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211020] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3204), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(5413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211043] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4731), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5414), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211066] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8157), 1,
      anon_sym_RPAREN,
    STATE(5415), 1,
      sym_attribute,
    STATE(5646), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211089] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8108), 1,
      sym__lowercase_identifier,
    ACTIONS(8112), 1,
      anon_sym_mutable,
    ACTIONS(8159), 1,
      anon_sym_RBRACE,
    STATE(5416), 1,
      sym_attribute,
    STATE(6114), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211112] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7047), 1,
      anon_sym_as,
    STATE(5417), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4689), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [211131] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    STATE(5418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7839), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [211150] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8163), 1,
      anon_sym_of,
    STATE(5419), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8161), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [211169] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5420), 1,
      sym_attribute,
    STATE(6189), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8165), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [211190] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5768), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5421), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211213] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5422), 1,
      sym_attribute,
    STATE(6214), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8167), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [211234] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    STATE(5423), 1,
      sym_attribute,
    STATE(6000), 1,
      sym_type_constructor_path,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211257] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3042), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(5424), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211280] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(8077), 1,
      sym__lowercase_identifier,
    STATE(5425), 1,
      sym_attribute,
    STATE(6058), 1,
      sym_extended_module_path,
    STATE(6065), 1,
      sym_module_type_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211303] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3092), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_PIPE,
    ACTIONS(7338), 1,
      anon_sym_as,
    ACTIONS(7340), 1,
      anon_sym_COLON_COLON,
    STATE(5426), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211326] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3738), 1,
      anon_sym_RBRACK,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8169), 1,
      anon_sym_GT,
    STATE(5298), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5427), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211349] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8171), 1,
      anon_sym_GT_RBRACE,
    STATE(5428), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211372] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(8173), 1,
      anon_sym_RBRACK,
    STATE(5395), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5429), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211395] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7987), 1,
      sym__lowercase_identifier,
    STATE(2633), 1,
      sym_class_type_path,
    STATE(5430), 1,
      sym_attribute,
    STATE(6229), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211418] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8175), 1,
      anon_sym_rec,
    STATE(4152), 1,
      sym_module_binding,
    STATE(5431), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7508), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [211439] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8177), 1,
      anon_sym_RBRACK,
    ACTIONS(8179), 1,
      anon_sym_GT,
    STATE(5432), 1,
      sym_attribute,
    STATE(5468), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211462] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8177), 1,
      anon_sym_RBRACK,
    ACTIONS(8179), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211485] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5434), 1,
      sym_attribute,
    STATE(6280), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8181), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [211506] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8127), 1,
      sym__lowercase_identifier,
    ACTIONS(8129), 1,
      anon_sym_LPAREN,
    STATE(2049), 1,
      sym_parenthesized_operator,
    STATE(2078), 1,
      sym__value_name,
    STATE(5435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211529] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(8183), 1,
      sym__lowercase_identifier,
    STATE(5436), 1,
      sym_attribute,
    STATE(6329), 1,
      sym_extended_module_path,
    STATE(7146), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211552] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5437), 1,
      sym_attribute,
    STATE(6305), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8185), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [211573] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8187), 1,
      anon_sym_GT_RBRACE,
    STATE(5438), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211596] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5439), 1,
      sym_attribute,
    STATE(6314), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8189), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [211617] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8191), 1,
      anon_sym_GT_RBRACE,
    STATE(5440), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211640] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5808), 1,
      anon_sym_in,
    ACTIONS(8193), 1,
      anon_sym_and,
    ACTIONS(8196), 1,
      sym_let_and_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5441), 2,
      sym_attribute,
      aux_sym_value_definition_repeat1,
  [211661] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8108), 1,
      sym__lowercase_identifier,
    ACTIONS(8112), 1,
      anon_sym_mutable,
    ACTIONS(8199), 1,
      anon_sym_RBRACE,
    STATE(5442), 1,
      sym_attribute,
    STATE(6114), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211684] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7660), 1,
      sym__lowercase_identifier,
    STATE(1009), 1,
      sym_class_path,
    STATE(5443), 1,
      sym_attribute,
    STATE(6599), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211707] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8201), 1,
      anon_sym_DASH_GT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5269), 1,
      aux_sym_module_binding_repeat1,
    STATE(5444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211730] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(7921), 1,
      anon_sym_RBRACK,
    STATE(5445), 1,
      sym_attribute,
    STATE(5779), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211753] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8201), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5446), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211776] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(8203), 1,
      sym__lowercase_identifier,
    STATE(2443), 1,
      sym_class_path,
    STATE(5447), 1,
      sym_attribute,
    STATE(6556), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211799] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8205), 1,
      sym__lowercase_identifier,
    STATE(5345), 1,
      aux_sym_method_specification_repeat1,
    STATE(5448), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7937), 2,
      anon_sym_private,
      anon_sym_virtual,
  [211820] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(8207), 1,
      sym__lowercase_identifier,
    STATE(4944), 1,
      sym_type_constructor_path,
    STATE(5449), 1,
      sym_attribute,
    STATE(6254), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211843] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7456), 1,
      sym__lowercase_identifier,
    ACTIONS(7460), 1,
      anon_sym_LBRACK,
    ACTIONS(7462), 1,
      anon_sym_virtual,
    STATE(5450), 1,
      sym_attribute,
    STATE(5943), 1,
      sym_class_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211866] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4771), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5451), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211889] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(8209), 1,
      sym__lowercase_identifier,
    STATE(5452), 1,
      sym_attribute,
    STATE(5491), 1,
      sym_class_type_path,
    STATE(6267), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211912] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(7959), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5453), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211935] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8211), 1,
      anon_sym_DASH_GT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5454), 1,
      sym_attribute,
    STATE(5458), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211958] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8211), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5455), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211981] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8205), 1,
      sym__lowercase_identifier,
    STATE(5355), 1,
      aux_sym_method_specification_repeat1,
    STATE(5456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7937), 2,
      anon_sym_private,
      anon_sym_virtual,
  [212002] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8213), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212025] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8215), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212048] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6999), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    STATE(5459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212071] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    ACTIONS(8118), 1,
      sym_hash_operator,
    ACTIONS(8217), 1,
      anon_sym_with,
    STATE(5460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212094] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6835), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7107), 1,
      anon_sym_PIPE,
    ACTIONS(7111), 1,
      anon_sym_as,
    STATE(5461), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212117] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8219), 1,
      aux_sym_attribute_id_token1,
    STATE(5462), 1,
      sym_attribute,
    STATE(5474), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2754), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
  [212138] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8221), 1,
      anon_sym_GT_RBRACE,
    STATE(5463), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212161] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(4331), 1,
      sym__lowercase_identifier,
    STATE(5464), 1,
      sym_attribute,
    STATE(6216), 1,
      sym_type_constructor_path,
    STATE(6329), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212184] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8030), 1,
      sym__lowercase_identifier,
    ACTIONS(8032), 1,
      anon_sym_LPAREN,
    STATE(3005), 1,
      sym__value_name,
    STATE(3051), 1,
      sym_parenthesized_operator,
    STATE(5465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212207] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5792), 1,
      anon_sym_in,
    STATE(5466), 1,
      sym_attribute,
    STATE(5494), 1,
      aux_sym_expression_item_repeat1,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212230] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8223), 1,
      anon_sym_rec,
    STATE(5467), 1,
      sym_attribute,
    STATE(5844), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7569), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [212251] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8225), 1,
      anon_sym_RBRACK,
    ACTIONS(8227), 1,
      anon_sym_GT,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212274] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8229), 1,
      anon_sym_GT_RBRACE,
    STATE(5469), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212297] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7378), 1,
      anon_sym_with,
    ACTIONS(7595), 1,
      anon_sym_DASH_GT,
    STATE(5470), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5098), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212318] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7378), 1,
      anon_sym_with,
    ACTIONS(7595), 1,
      anon_sym_DASH_GT,
    STATE(5471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5110), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212339] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7585), 1,
      sym__lowercase_identifier,
    ACTIONS(7587), 1,
      anon_sym_LBRACK,
    ACTIONS(7589), 1,
      anon_sym_virtual,
    STATE(4039), 1,
      sym_class_binding,
    STATE(5472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212362] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(8225), 1,
      anon_sym_RBRACK,
    STATE(5395), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5473), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212385] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8219), 1,
      aux_sym_attribute_id_token1,
    STATE(5474), 1,
      sym_attribute,
    STATE(5510), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2744), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
  [212406] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4771), 1,
      anon_sym_in,
    STATE(5407), 1,
      aux_sym_expression_item_repeat1,
    STATE(5475), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212429] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7378), 1,
      anon_sym_with,
    ACTIONS(7595), 1,
      anon_sym_DASH_GT,
    STATE(5476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4980), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212450] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7987), 1,
      sym__lowercase_identifier,
    STATE(2616), 1,
      sym_class_type_path,
    STATE(5477), 1,
      sym_attribute,
    STATE(6229), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212473] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4829), 1,
      anon_sym_in,
    STATE(5414), 1,
      aux_sym_expression_item_repeat1,
    STATE(5478), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212496] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8231), 1,
      anon_sym_DASH_GT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5479), 1,
      sym_attribute,
    STATE(5482), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212519] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8231), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212542] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8233), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212565] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8235), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5482), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212588] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7700), 1,
      sym__lowercase_identifier,
    STATE(1613), 1,
      sym_class_path,
    STATE(5483), 1,
      sym_attribute,
    STATE(6639), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212611] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(7868), 1,
      anon_sym_RBRACK,
    STATE(5484), 1,
      sym_attribute,
    STATE(5874), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212634] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8237), 1,
      anon_sym_GT_RBRACE,
    STATE(5485), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212657] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8108), 1,
      sym__lowercase_identifier,
    ACTIONS(8112), 1,
      anon_sym_mutable,
    ACTIONS(8239), 1,
      anon_sym_RBRACE,
    STATE(5486), 1,
      sym_attribute,
    STATE(6114), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212680] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5487), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4783), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212697] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8241), 1,
      anon_sym_DASH_GT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5488), 1,
      sym_attribute,
    STATE(5492), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212720] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8241), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5489), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212743] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8243), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5490), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212766] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5491), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4813), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212783] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8245), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212806] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    ACTIONS(8118), 1,
      sym_hash_operator,
    ACTIONS(8247), 1,
      anon_sym_with,
    STATE(5493), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212829] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5796), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5494), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212852] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5796), 1,
      anon_sym_in,
    STATE(5294), 1,
      aux_sym_expression_item_repeat1,
    STATE(5495), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212875] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(7987), 1,
      sym__lowercase_identifier,
    STATE(3358), 1,
      sym_class_type_path,
    STATE(5496), 1,
      sym_attribute,
    STATE(6229), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212898] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(8249), 1,
      sym__lowercase_identifier,
    STATE(3129), 1,
      sym_type_constructor_path,
    STATE(5497), 1,
      sym_attribute,
    STATE(6284), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212921] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5730), 1,
      anon_sym_in,
    ACTIONS(7989), 1,
      anon_sym_and,
    ACTIONS(7991), 1,
      sym_let_and_operator,
    STATE(5498), 1,
      sym_attribute,
    STATE(5542), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212944] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7622), 1,
      sym__lowercase_identifier,
    ACTIONS(7624), 1,
      anon_sym_LBRACK,
    ACTIONS(7626), 1,
      anon_sym_virtual,
    STATE(3940), 1,
      sym_class_type_binding,
    STATE(5499), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212967] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    ACTIONS(8118), 1,
      sym_hash_operator,
    ACTIONS(8251), 1,
      anon_sym_with,
    STATE(5500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212990] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    ACTIONS(7929), 1,
      sym__lowercase_identifier,
    STATE(3541), 1,
      sym_parenthesized_operator,
    STATE(4553), 1,
      sym__value_name,
    STATE(5501), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213013] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5696), 1,
      anon_sym_in,
    STATE(5098), 1,
      aux_sym_expression_item_repeat1,
    STATE(5502), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213036] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(8203), 1,
      sym__lowercase_identifier,
    STATE(2431), 1,
      sym_class_path,
    STATE(5503), 1,
      sym_attribute,
    STATE(6556), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213059] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    ACTIONS(8118), 1,
      sym_hash_operator,
    ACTIONS(8253), 1,
      anon_sym_with,
    STATE(5504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213082] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7613), 1,
      sym__lowercase_identifier,
    STATE(2100), 1,
      sym_class_path,
    STATE(5505), 1,
      sym_attribute,
    STATE(6443), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213105] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    ACTIONS(8118), 1,
      sym_hash_operator,
    ACTIONS(8255), 1,
      anon_sym_with,
    STATE(5506), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213128] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7585), 1,
      sym__lowercase_identifier,
    ACTIONS(7587), 1,
      anon_sym_LBRACK,
    ACTIONS(7589), 1,
      anon_sym_virtual,
    STATE(3941), 1,
      sym_class_binding,
    STATE(5507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213151] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    ACTIONS(8118), 1,
      sym_hash_operator,
    ACTIONS(8257), 1,
      anon_sym_with,
    STATE(5508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213174] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2646), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4809), 1,
      anon_sym_in,
    STATE(5451), 1,
      aux_sym_expression_item_repeat1,
    STATE(5509), 1,
      sym_attribute,
    STATE(5919), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213197] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8259), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2758), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
    STATE(5510), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
  [213216] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5788), 1,
      anon_sym_in,
    ACTIONS(7989), 1,
      anon_sym_and,
    ACTIONS(7991), 1,
      sym_let_and_operator,
    STATE(5288), 1,
      aux_sym_value_definition_repeat1,
    STATE(5511), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213239] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    ACTIONS(8209), 1,
      sym__lowercase_identifier,
    STATE(5300), 1,
      sym_class_type_path,
    STATE(5512), 1,
      sym_attribute,
    STATE(6267), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213262] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    ACTIONS(8118), 1,
      sym_hash_operator,
    ACTIONS(8262), 1,
      anon_sym_with,
    STATE(5513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213285] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    ACTIONS(8118), 1,
      sym_hash_operator,
    ACTIONS(8264), 1,
      anon_sym_with,
    STATE(5514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213308] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    ACTIONS(8118), 1,
      sym_hash_operator,
    ACTIONS(8266), 1,
      anon_sym_with,
    STATE(5515), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213331] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2960), 1,
      anon_sym_POUND,
    ACTIONS(7889), 1,
      anon_sym_DOT,
    ACTIONS(8118), 1,
      sym_hash_operator,
    ACTIONS(8268), 1,
      anon_sym_with,
    STATE(5516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213354] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(5517), 1,
      sym_attribute,
    STATE(6077), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213377] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5518), 1,
      sym_attribute,
    STATE(6276), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8270), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [213398] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7378), 1,
      anon_sym_with,
    ACTIONS(7595), 1,
      anon_sym_DASH_GT,
    STATE(5519), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5016), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [213419] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8272), 1,
      anon_sym_GT_RBRACE,
    STATE(5520), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213442] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8274), 1,
      anon_sym_DASH_GT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5521), 1,
      sym_attribute,
    STATE(5523), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213465] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8274), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5522), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213488] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8276), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5523), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213511] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(5524), 1,
      sym_attribute,
    STATE(6177), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213534] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(7702), 1,
      sym__lowercase_identifier,
    STATE(829), 1,
      sym_class_path,
    STATE(5525), 1,
      sym_attribute,
    STATE(7288), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213557] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7682), 1,
      sym__lowercase_identifier,
    ACTIONS(7684), 1,
      anon_sym_LPAREN,
    STATE(1682), 1,
      sym_parenthesized_operator,
    STATE(5526), 1,
      sym_attribute,
    STATE(6275), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213580] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8278), 1,
      anon_sym_rec,
    STATE(4051), 1,
      sym_module_binding,
    STATE(5527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7575), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [213601] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    ACTIONS(8280), 1,
      anon_sym_RBRACK,
    STATE(5395), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5528), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213624] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5784), 1,
      anon_sym_in,
    ACTIONS(7989), 1,
      anon_sym_and,
    ACTIONS(7991), 1,
      sym_let_and_operator,
    STATE(5324), 1,
      aux_sym_value_definition_repeat1,
    STATE(5529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213647] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    ACTIONS(8282), 1,
      anon_sym_GT_RBRACE,
    STATE(5530), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213670] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(8077), 1,
      sym__lowercase_identifier,
    STATE(5531), 1,
      sym_attribute,
    STATE(6058), 1,
      sym_extended_module_path,
    STATE(6283), 1,
      sym_module_type_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213693] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    STATE(5532), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5098), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [213714] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    STATE(5533), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5110), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [213735] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8284), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5534), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213758] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    STATE(5535), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4980), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [213779] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5064), 1,
      sym__uppercase_identifier,
    ACTIONS(8077), 1,
      sym__lowercase_identifier,
    STATE(5536), 1,
      sym_attribute,
    STATE(6058), 1,
      sym_extended_module_path,
    STATE(6318), 1,
      sym_module_type_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213802] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    STATE(5537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5016), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [213823] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5828), 4,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_DQUOTE,
  [213840] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8286), 1,
      anon_sym_DASH_GT,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5539), 1,
      sym_attribute,
    STATE(5541), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213863] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8286), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213886] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    ACTIONS(8288), 1,
      anon_sym_DASH_GT,
    STATE(3270), 1,
      aux_sym_module_binding_repeat1,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213909] = 7,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5784), 1,
      anon_sym_in,
    ACTIONS(7989), 1,
      anon_sym_and,
    ACTIONS(7991), 1,
      sym_let_and_operator,
    STATE(5441), 1,
      aux_sym_value_definition_repeat1,
    STATE(5542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213932] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8290), 1,
      anon_sym_SEMI,
    ACTIONS(8292), 1,
      anon_sym_GT_RBRACE,
    STATE(5543), 1,
      sym_attribute,
    STATE(5622), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213952] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8296), 1,
      anon_sym_RBRACK,
    STATE(5544), 1,
      sym_attribute,
    STATE(5929), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213972] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8298), 3,
      anon_sym_RPAREN,
      anon_sym_DOT,
      sym__lowercase_identifier,
  [213988] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7721), 1,
      anon_sym_RBRACE,
    ACTIONS(8300), 1,
      anon_sym_SEMI,
    STATE(5546), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214008] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8302), 1,
      anon_sym_LPAREN,
    ACTIONS(8304), 1,
      anon_sym_LBRACK,
    ACTIONS(8306), 1,
      anon_sym_LBRACE,
    STATE(5547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214028] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3728), 1,
      anon_sym_RBRACK,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    STATE(5548), 1,
      sym_attribute,
    STATE(5600), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214048] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(8308), 1,
      anon_sym_type,
    STATE(5549), 1,
      sym_attribute,
    STATE(5639), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214068] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8310), 1,
      anon_sym_RBRACK,
    STATE(5292), 1,
      aux_sym__type_params_repeat1,
    STATE(5550), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214088] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8310), 1,
      anon_sym_RBRACK,
    STATE(5551), 1,
      sym_attribute,
    STATE(5572), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214108] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8312), 1,
      anon_sym_RBRACK,
    STATE(5552), 1,
      sym_attribute,
    STATE(5905), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214128] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7484), 1,
      anon_sym_RBRACE,
    ACTIONS(8314), 1,
      anon_sym_SEMI,
    STATE(5553), 1,
      sym_attribute,
    STATE(5815), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214148] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7286), 1,
      anon_sym_SEMI,
    ACTIONS(7288), 1,
      anon_sym_GT,
    STATE(5554), 1,
      sym_attribute,
    STATE(5604), 1,
      aux_sym_object_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214168] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8316), 1,
      anon_sym_SEMI,
    ACTIONS(8319), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5555), 2,
      sym_attribute,
      aux_sym_record_pattern_repeat1,
  [214186] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(8321), 1,
      anon_sym_EQ,
    STATE(5556), 1,
      sym_attribute,
    STATE(6689), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214206] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8069), 1,
      anon_sym_RBRACK,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214226] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5024), 1,
      anon_sym_RPAREN,
    STATE(5558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8323), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [214244] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7745), 1,
      anon_sym_RBRACE,
    ACTIONS(8325), 1,
      anon_sym_SEMI,
    STATE(5559), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214264] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8069), 1,
      anon_sym_RBRACK,
    STATE(5560), 1,
      sym_attribute,
    STATE(5655), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214284] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8067), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [214300] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7993), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8327), 1,
      anon_sym_SEMI,
    STATE(5562), 1,
      sym_attribute,
    STATE(5643), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214320] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(5394), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5563), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214340] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1136), 1,
      anon_sym_end,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5564), 1,
      sym_attribute,
    STATE(6781), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214360] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8329), 1,
      anon_sym_SEMI,
    ACTIONS(8331), 1,
      anon_sym_RBRACE,
    STATE(5565), 1,
      sym_attribute,
    STATE(5585), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214380] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8075), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8333), 1,
      anon_sym_SEMI,
    STATE(5566), 1,
      sym_attribute,
    STATE(5643), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214400] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7488), 1,
      anon_sym_GT,
    ACTIONS(8335), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5567), 2,
      sym_attribute,
      aux_sym_object_type_repeat1,
  [214418] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6103), 1,
      sym__uppercase_identifier,
    ACTIONS(7877), 1,
      anon_sym_LPAREN,
    STATE(1761), 1,
      sym__constructor_name,
    STATE(5568), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214438] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7973), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8338), 1,
      anon_sym_SEMI,
    STATE(5569), 1,
      sym_attribute,
    STATE(5643), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214458] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8340), 1,
      sym__lowercase_identifier,
    ACTIONS(8342), 1,
      anon_sym_LPAREN,
    STATE(2501), 1,
      sym__simple_value_pattern,
    STATE(5570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214478] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3678), 1,
      anon_sym_RBRACK,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    STATE(5571), 1,
      sym_attribute,
    STATE(5717), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214498] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8344), 1,
      anon_sym_RBRACK,
    STATE(5292), 1,
      aux_sym__type_params_repeat1,
    STATE(5572), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214518] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8346), 1,
      sym__lowercase_identifier,
    ACTIONS(8348), 1,
      sym_indexing_operator,
    STATE(5573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214538] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8350), 1,
      anon_sym_SEMI,
    ACTIONS(8352), 1,
      anon_sym_RBRACE,
    STATE(5574), 1,
      sym_attribute,
    STATE(5793), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214558] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2518), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8354), 1,
      anon_sym_SEMI,
    STATE(5575), 1,
      sym_attribute,
    STATE(5598), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214578] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(8356), 1,
      anon_sym_RPAREN,
    STATE(5576), 1,
      sym_attribute,
    STATE(6826), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214598] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2500), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8358), 1,
      anon_sym_SEMI,
    STATE(5577), 1,
      sym_attribute,
    STATE(5599), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214618] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5864), 1,
      anon_sym_in,
    ACTIONS(8360), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5578), 2,
      sym_attribute,
      aux_sym_class_type_definition_repeat1,
  [214636] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(8363), 1,
      anon_sym_EQ,
    STATE(5579), 1,
      sym_attribute,
    STATE(6774), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214656] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7492), 1,
      anon_sym_RBRACE,
    ACTIONS(8365), 1,
      anon_sym_SEMI,
    STATE(5555), 1,
      aux_sym_record_pattern_repeat1,
    STATE(5580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214676] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7290), 1,
      anon_sym_SEMI,
    ACTIONS(7292), 1,
      anon_sym_GT,
    STATE(5581), 1,
      sym_attribute,
    STATE(5723), 1,
      aux_sym_object_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214696] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5897), 1,
      anon_sym_in,
    ACTIONS(8367), 1,
      anon_sym_and,
    STATE(5578), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(5582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214716] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(5473), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5583), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214736] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6338), 1,
      anon_sym_in,
    ACTIONS(8369), 1,
      anon_sym_and,
    STATE(5584), 1,
      sym_attribute,
    STATE(5647), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214756] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7735), 1,
      anon_sym_RBRACE,
    ACTIONS(8371), 1,
      anon_sym_SEMI,
    STATE(5585), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214776] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7710), 1,
      anon_sym_RBRACE,
    ACTIONS(8373), 1,
      anon_sym_SEMI,
    STATE(5586), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214796] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6034), 1,
      anon_sym_in,
    ACTIONS(8375), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5587), 2,
      sym_attribute,
      aux_sym_type_definition_repeat1,
  [214814] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7514), 1,
      anon_sym_RBRACE,
    ACTIONS(8378), 1,
      anon_sym_SEMI,
    STATE(5588), 1,
      sym_attribute,
    STATE(5815), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214834] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7662), 1,
      anon_sym_RBRACE,
    ACTIONS(8380), 1,
      anon_sym_SEMI,
    STATE(5589), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214854] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4579), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [214870] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8004), 1,
      anon_sym_RBRACK,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5591), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214890] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8382), 1,
      anon_sym_RPAREN,
    STATE(5592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214910] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8384), 1,
      anon_sym_SEMI,
    ACTIONS(8386), 1,
      anon_sym_RBRACE,
    STATE(5593), 1,
      sym_attribute,
    STATE(5640), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214930] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8388), 1,
      anon_sym_RPAREN,
    STATE(5343), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214950] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8390), 1,
      anon_sym_RBRACK,
    STATE(5343), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5595), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214970] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8090), 1,
      anon_sym_RBRACK,
    STATE(5596), 1,
      sym_attribute,
    STATE(5615), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214990] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8392), 1,
      anon_sym_RBRACK,
    STATE(5343), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5597), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215010] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7706), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8394), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5598), 2,
      sym_attribute,
      aux_sym__sequence_pattern_content_repeat1,
  [215028] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7638), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8397), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5599), 2,
      sym_attribute,
      aux_sym__sequence_binding_pattern_content_repeat1,
  [215046] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8090), 1,
      anon_sym_RBRACK,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5600), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215066] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(5348), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5601), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215086] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(5528), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5602), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215106] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7995), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [215122] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4479), 1,
      anon_sym_GT,
    ACTIONS(8400), 1,
      anon_sym_SEMI,
    STATE(5567), 1,
      aux_sym_object_type_repeat1,
    STATE(5604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215142] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8402), 1,
      anon_sym_SEMI,
    ACTIONS(8404), 1,
      anon_sym_RBRACE,
    STATE(5605), 1,
      sym_attribute,
    STATE(5620), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215162] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(5267), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5606), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215182] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8406), 1,
      anon_sym_RPAREN,
    STATE(5607), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215202] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2500), 1,
      anon_sym_RBRACK,
    ACTIONS(8408), 1,
      anon_sym_SEMI,
    STATE(5608), 1,
      sym_attribute,
    STATE(5734), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215222] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8348), 1,
      sym_indexing_operator,
    ACTIONS(8410), 1,
      sym__lowercase_identifier,
    STATE(5609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215242] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8412), 1,
      anon_sym_SEMI,
    ACTIONS(8414), 1,
      anon_sym_RBRACE,
    STATE(5610), 1,
      sym_attribute,
    STATE(5941), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215262] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(8416), 1,
      anon_sym_EQ,
    STATE(5611), 1,
      sym_attribute,
    STATE(7325), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215282] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3113), 1,
      sym_attribute_id,
    STATE(5612), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8418), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [215300] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(8420), 1,
      anon_sym_RPAREN,
    STATE(5613), 1,
      sym_attribute,
    STATE(6914), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215320] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8422), 1,
      anon_sym_RPAREN,
    STATE(5614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215340] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8100), 1,
      anon_sym_RBRACK,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215360] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8100), 1,
      anon_sym_RBRACK,
    STATE(5616), 1,
      sym_attribute,
    STATE(5627), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215380] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8348), 1,
      sym_indexing_operator,
    ACTIONS(8424), 1,
      sym__lowercase_identifier,
    STATE(5617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215400] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5933), 1,
      anon_sym_in,
    ACTIONS(8367), 1,
      anon_sym_and,
    STATE(5578), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(5618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215420] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    STATE(5619), 1,
      sym_attribute,
    STATE(5674), 1,
      sym__instance_variable_name,
    STATE(6255), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215440] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7790), 1,
      anon_sym_RBRACE,
    ACTIONS(8426), 1,
      anon_sym_SEMI,
    STATE(5620), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215460] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8430), 1,
      anon_sym_LBRACK,
    STATE(5621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8428), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
  [215478] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8106), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8432), 1,
      anon_sym_SEMI,
    STATE(5622), 1,
      sym_attribute,
    STATE(5643), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215498] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8434), 1,
      anon_sym_SEMI,
    ACTIONS(8437), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5623), 2,
      sym_attribute,
      aux_sym_record_declaration_repeat1,
  [215516] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8439), 1,
      anon_sym_RPAREN,
    STATE(5624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215536] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8441), 1,
      anon_sym_RPAREN,
    STATE(5625), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215556] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8443), 1,
      anon_sym_RPAREN,
    STATE(5626), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215576] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8114), 1,
      anon_sym_RBRACK,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215596] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4434), 1,
      anon_sym_GT,
    ACTIONS(8445), 1,
      anon_sym_SEMI,
    STATE(5567), 1,
      aux_sym_object_type_repeat1,
    STATE(5628), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215616] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8447), 1,
      anon_sym_RPAREN,
    STATE(5292), 1,
      aux_sym__type_params_repeat1,
    STATE(5629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215636] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8449), 1,
      anon_sym_SEMI,
    ACTIONS(8451), 1,
      anon_sym_RBRACE,
    STATE(5630), 1,
      sym_attribute,
    STATE(5641), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215656] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8453), 1,
      anon_sym_SEMI,
    ACTIONS(8455), 1,
      anon_sym_RBRACE,
    STATE(5631), 1,
      sym_attribute,
    STATE(5745), 1,
      aux_sym_record_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215676] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8457), 1,
      anon_sym_SEMI,
    ACTIONS(8459), 1,
      anon_sym_RBRACE,
    STATE(5589), 1,
      aux_sym_record_expression_repeat1,
    STATE(5632), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215696] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8461), 1,
      anon_sym_RPAREN,
    STATE(5633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215716] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6065), 1,
      anon_sym_in,
    ACTIONS(8463), 1,
      anon_sym_and,
    STATE(5587), 1,
      aux_sym_type_definition_repeat1,
    STATE(5634), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215736] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8465), 1,
      anon_sym_SEMI,
    ACTIONS(8467), 1,
      anon_sym_GT_RBRACE,
    STATE(5635), 1,
      sym_attribute,
    STATE(5644), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215756] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5933), 1,
      anon_sym_in,
    ACTIONS(8367), 1,
      anon_sym_and,
    STATE(5582), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(5636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215776] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8469), 1,
      anon_sym_SEMI,
    ACTIONS(8471), 1,
      anon_sym_RBRACE,
    STATE(5553), 1,
      aux_sym_record_binding_pattern_repeat1,
    STATE(5637), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215796] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6779), 1,
      sym__uppercase_identifier,
    ACTIONS(7887), 1,
      anon_sym_LPAREN,
    STATE(3076), 1,
      sym__constructor_name,
    STATE(5638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215816] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4165), 1,
      anon_sym_DOT,
    STATE(5639), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8473), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [215834] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7680), 1,
      anon_sym_RBRACE,
    ACTIONS(8475), 1,
      anon_sym_SEMI,
    STATE(5640), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215854] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7772), 1,
      anon_sym_RBRACE,
    ACTIONS(8477), 1,
      anon_sym_SEMI,
    STATE(5641), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215874] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4117), 1,
      sym_module_binding,
    STATE(5642), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7508), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [215892] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8479), 1,
      anon_sym_SEMI,
    ACTIONS(8482), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5643), 2,
      sym_attribute,
      aux_sym_object_copy_expression_repeat1,
  [215910] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8012), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8484), 1,
      anon_sym_SEMI,
    STATE(5643), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215930] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8486), 1,
      anon_sym_RPAREN,
    ACTIONS(8488), 1,
      anon_sym_COLON2,
    STATE(5645), 1,
      sym_attribute,
    STATE(6714), 1,
      sym__class_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215950] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8490), 1,
      anon_sym_RPAREN,
    STATE(5343), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215970] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6089), 1,
      anon_sym_in,
    ACTIONS(8492), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5647), 2,
      sym_attribute,
      aux_sym_module_definition_repeat1,
  [215988] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6095), 1,
      anon_sym_in,
    ACTIONS(8495), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5648), 2,
      sym_attribute,
      aux_sym_class_definition_repeat1,
  [216006] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6132), 1,
      anon_sym_in,
    ACTIONS(8498), 1,
      anon_sym_and,
    STATE(5648), 1,
      aux_sym_class_definition_repeat1,
    STATE(5649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216026] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(8500), 1,
      anon_sym_EQ,
    STATE(5650), 1,
      sym_attribute,
    STATE(6796), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216046] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8502), 1,
      anon_sym_SEMI,
    ACTIONS(8504), 1,
      anon_sym_RBRACE,
    STATE(5651), 1,
      sym_attribute,
    STATE(5659), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216066] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8506), 1,
      anon_sym_RPAREN,
    STATE(5652), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216086] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(8508), 1,
      anon_sym_RPAREN,
    STATE(5653), 1,
      sym_attribute,
    STATE(6746), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216106] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8004), 1,
      anon_sym_RBRACK,
    STATE(5557), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5654), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216126] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8137), 1,
      anon_sym_RBRACK,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216146] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8348), 1,
      sym_indexing_operator,
    ACTIONS(8510), 1,
      sym__lowercase_identifier,
    STATE(5656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216166] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6338), 1,
      anon_sym_in,
    ACTIONS(8369), 1,
      anon_sym_and,
    STATE(5657), 1,
      sym_attribute,
    STATE(5795), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216186] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(5429), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5658), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216206] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7788), 1,
      anon_sym_RBRACE,
    ACTIONS(8512), 1,
      anon_sym_SEMI,
    STATE(5659), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216226] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8348), 1,
      sym_indexing_operator,
    ACTIONS(8514), 1,
      sym__lowercase_identifier,
    STATE(5660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216246] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(14), 1,
      sym_attribute_id,
    STATE(5661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [216264] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(8), 1,
      sym_attribute_id,
    STATE(5662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [216282] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(27), 1,
      sym_attribute_id,
    STATE(5663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [216300] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(22), 1,
      sym_attribute_id,
    STATE(5664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [216318] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7454), 1,
      anon_sym_RBRACE,
    ACTIONS(8518), 1,
      anon_sym_SEMI,
    STATE(5555), 1,
      aux_sym_record_pattern_repeat1,
    STATE(5665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216338] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8108), 1,
      sym__lowercase_identifier,
    ACTIONS(8112), 1,
      anon_sym_mutable,
    STATE(5666), 1,
      sym_attribute,
    STATE(5944), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216358] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(15), 1,
      sym_attribute_id,
    STATE(5667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [216376] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(34), 1,
      sym_attribute_id,
    STATE(5668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [216394] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8520), 1,
      aux_sym_quoted_extension_token1,
    STATE(5669), 1,
      sym_attribute,
    STATE(6384), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216414] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(8522), 1,
      anon_sym_RPAREN,
    STATE(5670), 1,
      sym_attribute,
    STATE(7014), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216434] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1102), 1,
      anon_sym_end,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5671), 1,
      sym_attribute,
    STATE(6382), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216454] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8524), 1,
      anon_sym_SEMI,
    ACTIONS(8526), 1,
      anon_sym_GT_RBRACE,
    STATE(5672), 1,
      sym_attribute,
    STATE(5838), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216474] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6048), 1,
      anon_sym_in,
    ACTIONS(8463), 1,
      anon_sym_and,
    STATE(5587), 1,
      aux_sym_type_definition_repeat1,
    STATE(5673), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216494] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8528), 1,
      anon_sym_EQ,
    STATE(5674), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8530), 2,
      anon_sym_SEMI,
      anon_sym_GT_RBRACE,
  [216512] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8532), 1,
      anon_sym_SEMI,
    ACTIONS(8534), 1,
      anon_sym_RBRACE,
    STATE(5675), 1,
      sym_attribute,
    STATE(5704), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216532] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7521), 1,
      anon_sym_RBRACE,
    ACTIONS(8536), 1,
      anon_sym_SEMI,
    STATE(5676), 1,
      sym_attribute,
    STATE(5815), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216552] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(4410), 1,
      anon_sym_DOT,
    ACTIONS(8538), 1,
      anon_sym_RPAREN,
    STATE(5677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216572] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(5369), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5678), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216592] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1538), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [216608] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(5374), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5680), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216628] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8540), 1,
      anon_sym_SEMI,
    ACTIONS(8542), 1,
      anon_sym_GT_RBRACE,
    STATE(5681), 1,
      sym_attribute,
    STATE(5710), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216648] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2554), 1,
      aux_sym_tag_token1,
    STATE(5379), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5682), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216668] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8108), 1,
      sym__lowercase_identifier,
    ACTIONS(8112), 1,
      anon_sym_mutable,
    STATE(5683), 1,
      sym_attribute,
    STATE(6114), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216688] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3840), 1,
      sym_attribute_id,
    STATE(5684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8544), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [216706] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8546), 1,
      anon_sym_SEMI,
    ACTIONS(8548), 1,
      anon_sym_RBRACE,
    STATE(5685), 1,
      sym_attribute,
    STATE(5691), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216726] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1044), 1,
      anon_sym_end,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5686), 1,
      sym_attribute,
    STATE(6807), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216746] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8550), 1,
      anon_sym_SEMI,
    ACTIONS(8552), 1,
      anon_sym_RBRACE,
    STATE(5687), 1,
      sym_attribute,
    STATE(5718), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216766] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(8554), 1,
      anon_sym_EQ,
    STATE(5688), 1,
      sym_attribute,
    STATE(6923), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216786] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4152), 1,
      sym_module_binding,
    STATE(5689), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7508), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [216804] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8556), 1,
      anon_sym_SEMI,
    ACTIONS(8558), 1,
      anon_sym_GT_RBRACE,
    STATE(5690), 1,
      sym_attribute,
    STATE(5699), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216824] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7618), 1,
      anon_sym_RBRACE,
    ACTIONS(8560), 1,
      anon_sym_SEMI,
    STATE(5691), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216844] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8562), 1,
      anon_sym_SEMI,
    ACTIONS(8564), 1,
      anon_sym_RBRACE,
    STATE(5692), 1,
      sym_attribute,
    STATE(5725), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216864] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(16), 1,
      sym_attribute_id,
    STATE(5693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [216882] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(17), 1,
      sym_attribute_id,
    STATE(5694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [216900] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8566), 1,
      aux_sym_quoted_extension_token1,
    STATE(5695), 1,
      sym_attribute,
    STATE(6374), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216920] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8159), 1,
      anon_sym_RBRACE,
    ACTIONS(8568), 1,
      anon_sym_SEMI,
    STATE(5623), 1,
      aux_sym_record_declaration_repeat1,
    STATE(5696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216940] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3738), 1,
      anon_sym_RBRACK,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    STATE(5591), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216960] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7839), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [216976] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8020), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8570), 1,
      anon_sym_SEMI,
    STATE(5643), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216996] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5700), 1,
      sym_attribute,
    STATE(5834), 1,
      sym_attribute_id,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [217014] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8574), 1,
      anon_sym_RPAREN,
    STATE(5629), 1,
      aux_sym__type_params_repeat1,
    STATE(5701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217034] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4165), 1,
      anon_sym_DOT,
    STATE(5702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8576), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [217052] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8578), 1,
      anon_sym_SEMI,
    ACTIONS(8580), 1,
      anon_sym_RBRACE,
    STATE(5703), 1,
      sym_attribute,
    STATE(5719), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217072] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7674), 1,
      anon_sym_RBRACE,
    ACTIONS(8582), 1,
      anon_sym_SEMI,
    STATE(5704), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217092] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(2854), 1,
      sym_attribute_id,
    STATE(5705), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8584), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [217110] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8586), 1,
      anon_sym_RPAREN,
    STATE(5706), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217130] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7304), 1,
      anon_sym_SEMI,
    ACTIONS(7306), 1,
      anon_sym_GT,
    STATE(5628), 1,
      aux_sym_object_type_repeat1,
    STATE(5707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217150] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8588), 1,
      anon_sym_RPAREN,
    STATE(5708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217170] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8590), 1,
      anon_sym_RPAREN,
    STATE(5709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217190] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8171), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8592), 1,
      anon_sym_SEMI,
    STATE(5643), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217210] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8348), 1,
      sym_indexing_operator,
    ACTIONS(8594), 1,
      sym__lowercase_identifier,
    STATE(5711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217230] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8596), 1,
      anon_sym_RPAREN,
    STATE(5343), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217250] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6657), 1,
      sym__uppercase_identifier,
    ACTIONS(7831), 1,
      anon_sym_LPAREN,
    STATE(4465), 1,
      sym__constructor_name,
    STATE(5713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217270] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8177), 1,
      anon_sym_RBRACK,
    STATE(5714), 1,
      sym_attribute,
    STATE(5817), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217290] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(8598), 1,
      anon_sym_RPAREN,
    STATE(5715), 1,
      sym_attribute,
    STATE(6584), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217310] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(18), 1,
      sym_attribute_id,
    STATE(5716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [217328] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8177), 1,
      anon_sym_RBRACK,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217348] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7551), 1,
      anon_sym_RBRACE,
    ACTIONS(8600), 1,
      anon_sym_SEMI,
    STATE(5555), 1,
      aux_sym_record_pattern_repeat1,
    STATE(5718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217368] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7632), 1,
      anon_sym_RBRACE,
    ACTIONS(8602), 1,
      anon_sym_SEMI,
    STATE(5719), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217388] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2518), 1,
      anon_sym_RBRACK,
    ACTIONS(8604), 1,
      anon_sym_SEMI,
    STATE(5720), 1,
      sym_attribute,
    STATE(5936), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217408] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8606), 1,
      aux_sym_quoted_extension_token1,
    STATE(5721), 1,
      sym_attribute,
    STATE(6859), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217428] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(8608), 1,
      anon_sym_RPAREN,
    STATE(5722), 1,
      sym_attribute,
    STATE(6651), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217448] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4495), 1,
      anon_sym_GT,
    ACTIONS(8610), 1,
      anon_sym_SEMI,
    STATE(5567), 1,
      aux_sym_object_type_repeat1,
    STATE(5723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217468] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8612), 1,
      anon_sym_SEMI,
    ACTIONS(8614), 1,
      anon_sym_RBRACE,
    STATE(5724), 1,
      sym_attribute,
    STATE(5830), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217488] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7348), 1,
      anon_sym_RBRACE,
    ACTIONS(8616), 1,
      anon_sym_SEMI,
    STATE(5725), 1,
      sym_attribute,
    STATE(5815), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217508] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8618), 1,
      anon_sym_SEMI,
    ACTIONS(8620), 1,
      anon_sym_RBRACE,
    STATE(5665), 1,
      aux_sym_record_pattern_repeat1,
    STATE(5726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217528] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(19), 1,
      sym_attribute_id,
    STATE(5727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [217546] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8622), 1,
      anon_sym_SEMI,
    ACTIONS(8624), 1,
      anon_sym_RBRACE,
    STATE(5728), 1,
      sym_attribute,
    STATE(5740), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217566] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1286), 1,
      anon_sym_RBRACK,
    ACTIONS(8626), 1,
      anon_sym_SEMI,
    STATE(5729), 1,
      sym_attribute,
    STATE(5752), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217586] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8628), 1,
      aux_sym_quoted_extension_token1,
    STATE(5730), 1,
      sym_attribute,
    STATE(7149), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217606] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5731), 1,
      sym_attribute,
    STATE(5844), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7569), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [217624] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(411), 1,
      anon_sym_end,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5732), 1,
      sym_attribute,
    STATE(6403), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217644] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8348), 1,
      sym_indexing_operator,
    ACTIONS(8630), 1,
      sym__lowercase_identifier,
    STATE(5733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217664] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7638), 1,
      anon_sym_RBRACK,
    ACTIONS(8632), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5734), 2,
      sym_attribute,
      aux_sym__sequence_binding_pattern_content_repeat1,
  [217682] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(20), 1,
      sym_attribute_id,
    STATE(5735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [217700] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(8635), 1,
      anon_sym_RPAREN,
    STATE(5736), 1,
      sym_attribute,
    STATE(6628), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217720] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7668), 1,
      anon_sym_RBRACE,
    ACTIONS(8637), 1,
      anon_sym_SEMI,
    STATE(5737), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217740] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8639), 1,
      aux_sym_quoted_extension_token1,
    STATE(5738), 1,
      sym_attribute,
    STATE(6847), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217760] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8348), 1,
      sym_indexing_operator,
    ACTIONS(8641), 1,
      sym__lowercase_identifier,
    STATE(5739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217780] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7630), 1,
      anon_sym_RBRACE,
    ACTIONS(8643), 1,
      anon_sym_SEMI,
    STATE(5740), 1,
      sym_attribute,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217800] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5810), 1,
      anon_sym_and,
    STATE(5741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5808), 2,
      anon_sym_in,
      sym_let_and_operator,
  [217818] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(21), 1,
      sym_attribute_id,
    STATE(5742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [217836] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8645), 1,
      aux_sym_quoted_extension_token1,
    STATE(5743), 1,
      sym_attribute,
    STATE(6498), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217856] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(4410), 1,
      anon_sym_DOT,
    ACTIONS(8647), 1,
      anon_sym_RPAREN,
    STATE(5744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217876] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8199), 1,
      anon_sym_RBRACE,
    ACTIONS(8649), 1,
      anon_sym_SEMI,
    STATE(5623), 1,
      aux_sym_record_declaration_repeat1,
    STATE(5745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217896] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8651), 1,
      anon_sym_EQ,
    STATE(5746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8653), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [217914] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4689), 1,
      anon_sym_EQ,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(6631), 1,
      anon_sym_RPAREN,
    STATE(5747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217934] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(23), 1,
      sym_attribute_id,
    STATE(5748), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [217952] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8655), 1,
      aux_sym_quoted_extension_token1,
    STATE(5749), 1,
      sym_attribute,
    STATE(7304), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217972] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8657), 1,
      aux_sym_quoted_extension_token1,
    STATE(5750), 1,
      sym_attribute,
    STATE(7221), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217992] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8659), 1,
      anon_sym_RPAREN,
    STATE(5751), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218012] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4455), 1,
      anon_sym_RBRACK,
    ACTIONS(8661), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5752), 2,
      sym_attribute,
      aux_sym__sequence_expression_content_repeat1,
  [218030] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4898), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [218046] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8664), 1,
      anon_sym_RPAREN,
    STATE(5754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218066] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(8666), 1,
      anon_sym_type,
    STATE(5702), 1,
      sym_module_path,
    STATE(5755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218086] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8668), 3,
      anon_sym_mutable,
      anon_sym_virtual,
      sym__lowercase_identifier,
  [218102] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8670), 1,
      anon_sym_SEMI,
    ACTIONS(8672), 1,
      anon_sym_RBRACE,
    STATE(5586), 1,
      aux_sym_record_expression_repeat1,
    STATE(5757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218122] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(24), 1,
      sym_attribute_id,
    STATE(5758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218140] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8674), 1,
      aux_sym_quoted_extension_token1,
    STATE(5759), 1,
      sym_attribute,
    STATE(6585), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218160] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3938), 1,
      sym_module_binding,
    STATE(5760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7508), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [218178] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8676), 1,
      anon_sym_RBRACK,
    STATE(5292), 1,
      aux_sym__type_params_repeat1,
    STATE(5761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218198] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8676), 1,
      anon_sym_RBRACK,
    STATE(5762), 1,
      sym_attribute,
    STATE(5821), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218218] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5763), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8678), 3,
      anon_sym_private,
      anon_sym_virtual,
      sym__lowercase_identifier,
  [218234] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(25), 1,
      sym_attribute_id,
    STATE(5764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218252] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1160), 1,
      anon_sym_end,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5765), 1,
      sym_attribute,
    STATE(6862), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218272] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8680), 1,
      aux_sym_quoted_extension_token1,
    STATE(5766), 1,
      sym_attribute,
    STATE(6815), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218292] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(26), 1,
      sym_attribute_id,
    STATE(5767), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218310] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(8682), 1,
      anon_sym_RPAREN,
    STATE(5768), 1,
      sym_attribute,
    STATE(7104), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218330] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8348), 1,
      sym_indexing_operator,
    ACTIONS(8684), 1,
      sym__lowercase_identifier,
    STATE(5769), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218350] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8686), 1,
      aux_sym_quoted_extension_token1,
    STATE(5770), 1,
      sym_attribute,
    STATE(6903), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218370] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6), 1,
      sym_attribute_id,
    STATE(5771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218388] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8688), 1,
      aux_sym_quoted_extension_token1,
    STATE(5772), 1,
      sym_attribute,
    STATE(6953), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218408] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(28), 1,
      sym_attribute_id,
    STATE(5773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218426] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8690), 1,
      anon_sym_SEMI,
    ACTIONS(8692), 1,
      anon_sym_RBRACE,
    STATE(5774), 1,
      sym_attribute,
    STATE(5800), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218446] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8694), 1,
      aux_sym_quoted_extension_token1,
    STATE(5775), 1,
      sym_attribute,
    STATE(7007), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218466] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(29), 1,
      sym_attribute_id,
    STATE(5776), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218484] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8696), 1,
      aux_sym_quoted_extension_token1,
    STATE(5777), 1,
      sym_attribute,
    STATE(7048), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218504] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(30), 1,
      sym_attribute_id,
    STATE(5778), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218522] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8698), 1,
      anon_sym_RBRACK,
    STATE(5343), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5779), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218542] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8700), 1,
      aux_sym_quoted_extension_token1,
    STATE(5780), 1,
      sym_attribute,
    STATE(7099), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218562] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(31), 1,
      sym_attribute_id,
    STATE(5781), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218580] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8702), 1,
      anon_sym_SEMI,
    ACTIONS(8705), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5782), 2,
      sym_attribute,
      aux_sym_record_expression_repeat1,
  [218598] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8707), 1,
      aux_sym_quoted_extension_token1,
    STATE(5783), 1,
      sym_attribute,
    STATE(7154), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218618] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(32), 1,
      sym_attribute_id,
    STATE(5784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218636] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8709), 1,
      aux_sym_quoted_extension_token1,
    STATE(5785), 1,
      sym_attribute,
    STATE(7199), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218656] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(33), 1,
      sym_attribute_id,
    STATE(5786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218674] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8711), 1,
      aux_sym_quoted_extension_token1,
    STATE(5787), 1,
      sym_attribute,
    STATE(7223), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218694] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(7), 1,
      sym_attribute_id,
    STATE(5788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218712] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8713), 1,
      aux_sym_quoted_extension_token1,
    STATE(5789), 1,
      sym_attribute,
    STATE(7243), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218732] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5790), 1,
      sym_attribute,
    STATE(5857), 1,
      sym_attribute_id,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218750] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8715), 1,
      anon_sym_SEMI,
    ACTIONS(8717), 1,
      anon_sym_GT_RBRACE,
    STATE(5791), 1,
      sym_attribute,
    STATE(5808), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218770] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5792), 1,
      sym_attribute,
    STATE(5903), 1,
      sym_attribute_id,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218788] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7472), 1,
      anon_sym_RBRACE,
    ACTIONS(8719), 1,
      anon_sym_SEMI,
    STATE(5555), 1,
      aux_sym_record_pattern_repeat1,
    STATE(5793), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218808] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8721), 1,
      anon_sym_EQ,
    STATE(5794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8723), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [218826] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6345), 1,
      anon_sym_in,
    ACTIONS(8369), 1,
      anon_sym_and,
    STATE(5647), 1,
      aux_sym_module_definition_repeat1,
    STATE(5795), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218846] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4563), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [218862] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6779), 1,
      sym__uppercase_identifier,
    ACTIONS(7877), 1,
      anon_sym_LPAREN,
    STATE(3076), 1,
      sym__constructor_name,
    STATE(5797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218882] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8725), 1,
      anon_sym_RBRACK,
    STATE(5798), 1,
      sym_attribute,
    STATE(5803), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218902] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6048), 1,
      anon_sym_in,
    ACTIONS(8463), 1,
      anon_sym_and,
    STATE(5634), 1,
      aux_sym_type_definition_repeat1,
    STATE(5799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218922] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7717), 1,
      anon_sym_RBRACE,
    ACTIONS(8727), 1,
      anon_sym_SEMI,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    STATE(5800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218942] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8729), 1,
      anon_sym_LPAREN,
    ACTIONS(8731), 1,
      anon_sym_LBRACK,
    ACTIONS(8733), 1,
      anon_sym_LBRACE,
    STATE(5801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218962] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5304), 1,
      aux_sym_module_binding_repeat1,
    STATE(5802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218982] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8735), 1,
      anon_sym_RBRACK,
    STATE(5292), 1,
      aux_sym__type_params_repeat1,
    STATE(5803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219002] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8735), 1,
      anon_sym_RBRACK,
    STATE(5804), 1,
      sym_attribute,
    STATE(5811), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219022] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4715), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [219038] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(2653), 1,
      sym_attribute_id,
    STATE(5806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219056] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1286), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8737), 1,
      anon_sym_SEMI,
    STATE(5807), 1,
      sym_attribute,
    STATE(5814), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219076] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8221), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8739), 1,
      anon_sym_SEMI,
    STATE(5643), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5808), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219096] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8741), 1,
      anon_sym_LPAREN,
    ACTIONS(8743), 1,
      anon_sym_LBRACK,
    ACTIONS(8745), 1,
      anon_sym_LBRACE,
    STATE(5809), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219116] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4567), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [219132] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8747), 1,
      anon_sym_RBRACK,
    STATE(5292), 1,
      aux_sym__type_params_repeat1,
    STATE(5811), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219152] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(8749), 1,
      anon_sym_RPAREN,
    STATE(5812), 1,
      sym_attribute,
    STATE(6951), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219172] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6228), 1,
      anon_sym_in,
    ACTIONS(8463), 1,
      anon_sym_and,
    STATE(5813), 1,
      sym_attribute,
    STATE(5842), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219192] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4455), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8751), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5814), 2,
      sym_attribute,
      aux_sym__sequence_expression_content_repeat1,
  [219210] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8754), 1,
      anon_sym_SEMI,
    ACTIONS(8757), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5815), 2,
      sym_attribute,
      aux_sym_record_binding_pattern_repeat1,
  [219228] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8759), 1,
      anon_sym_RPAREN,
    STATE(5816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219248] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8225), 1,
      anon_sym_RBRACK,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219268] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6351), 1,
      anon_sym_in,
    ACTIONS(8369), 1,
      anon_sym_and,
    STATE(5818), 1,
      sym_attribute,
    STATE(5851), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219288] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8761), 1,
      anon_sym_SEMI,
    ACTIONS(8763), 1,
      anon_sym_RBRACE,
    STATE(5819), 1,
      sym_attribute,
    STATE(5826), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219308] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8225), 1,
      anon_sym_RBRACK,
    STATE(5820), 1,
      sym_attribute,
    STATE(5922), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219328] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8765), 1,
      anon_sym_RBRACK,
    STATE(5292), 1,
      aux_sym__type_params_repeat1,
    STATE(5821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219348] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8348), 1,
      sym_indexing_operator,
    ACTIONS(8767), 1,
      sym__lowercase_identifier,
    STATE(5822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219368] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(8769), 1,
      anon_sym_RPAREN,
    STATE(5823), 1,
      sym_attribute,
    STATE(6981), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219388] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6111), 1,
      anon_sym_in,
    ACTIONS(8498), 1,
      anon_sym_and,
    STATE(5824), 1,
      sym_attribute,
    STATE(5942), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219408] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8771), 1,
      sym__lowercase_identifier,
    ACTIONS(8773), 1,
      anon_sym_LPAREN,
    STATE(2171), 1,
      sym__simple_value_pattern,
    STATE(5825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219428] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7755), 1,
      anon_sym_RBRACE,
    ACTIONS(8775), 1,
      anon_sym_SEMI,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    STATE(5826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219448] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8777), 1,
      anon_sym_SEMI,
    ACTIONS(8779), 1,
      anon_sym_RBRACE,
    STATE(5559), 1,
      aux_sym_record_expression_repeat1,
    STATE(5827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219468] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(9), 1,
      sym_attribute_id,
    STATE(5828), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219486] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5750), 1,
      sym_attribute_id,
    STATE(5829), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219504] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7763), 1,
      anon_sym_RBRACE,
    ACTIONS(8781), 1,
      anon_sym_SEMI,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    STATE(5830), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219524] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8783), 1,
      anon_sym_SEMI,
    ACTIONS(8785), 1,
      anon_sym_RBRACE,
    STATE(5676), 1,
      aux_sym_record_binding_pattern_repeat1,
    STATE(5831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219544] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6038), 1,
      anon_sym_in,
    ACTIONS(8463), 1,
      anon_sym_and,
    STATE(5673), 1,
      aux_sym_type_definition_repeat1,
    STATE(5832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219564] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5481), 1,
      aux_sym_module_binding_repeat1,
    STATE(5833), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219584] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8787), 1,
      aux_sym_quoted_extension_token1,
    STATE(5834), 1,
      sym_attribute,
    STATE(7074), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219604] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8789), 1,
      anon_sym_RBRACK,
    STATE(5761), 1,
      aux_sym__type_params_repeat1,
    STATE(5835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219624] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8791), 1,
      anon_sym_SEMI,
    ACTIONS(8793), 1,
      anon_sym_GT_RBRACE,
    STATE(5569), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219644] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8795), 1,
      anon_sym_RPAREN,
    STATE(5837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219664] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8237), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8797), 1,
      anon_sym_SEMI,
    STATE(5643), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219684] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5669), 1,
      sym_attribute_id,
    STATE(5839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219702] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(10), 1,
      sym_attribute_id,
    STATE(5840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219720] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(11), 1,
      sym_attribute_id,
    STATE(5841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219738] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6038), 1,
      anon_sym_in,
    ACTIONS(8463), 1,
      anon_sym_and,
    STATE(5587), 1,
      aux_sym_type_definition_repeat1,
    STATE(5842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219758] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8799), 1,
      anon_sym_LPAREN,
    ACTIONS(8801), 1,
      anon_sym_LBRACK,
    ACTIONS(8803), 1,
      anon_sym_LBRACE,
    STATE(5843), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219778] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6332), 1,
      anon_sym_in,
    ACTIONS(8369), 1,
      anon_sym_and,
    STATE(5584), 1,
      aux_sym_module_definition_repeat1,
    STATE(5844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219798] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8805), 1,
      anon_sym_LPAREN,
    ACTIONS(8807), 1,
      anon_sym_LBRACK,
    ACTIONS(8809), 1,
      anon_sym_LBRACE,
    STATE(5845), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219818] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5534), 1,
      aux_sym_module_binding_repeat1,
    STATE(5846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219838] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1150), 1,
      anon_sym_end,
    ACTIONS(5000), 1,
      anon_sym_PERCENT,
    STATE(5847), 1,
      sym_attribute,
    STATE(7156), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219858] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(12), 1,
      sym_attribute_id,
    STATE(5848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219876] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5695), 1,
      sym_attribute_id,
    STATE(5849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219894] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5850), 1,
      sym_attribute,
    STATE(6298), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7569), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [219912] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6332), 1,
      anon_sym_in,
    ACTIONS(8369), 1,
      anon_sym_and,
    STATE(5647), 1,
      aux_sym_module_definition_repeat1,
    STATE(5851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219932] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5657), 1,
      sym_module_binding,
    STATE(5852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7569), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [219950] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8811), 1,
      anon_sym_LPAREN,
    ACTIONS(8813), 1,
      anon_sym_LBRACK,
    ACTIONS(8815), 1,
      anon_sym_LBRACE,
    STATE(5853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219970] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8817), 1,
      anon_sym_SEMI,
    ACTIONS(8819), 1,
      anon_sym_RBRACE,
    STATE(5854), 1,
      sym_attribute,
    STATE(5902), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219990] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8821), 1,
      anon_sym_LPAREN,
    ACTIONS(8823), 1,
      anon_sym_LBRACK,
    ACTIONS(8825), 1,
      anon_sym_LBRACE,
    STATE(5855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220010] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5721), 1,
      sym_attribute_id,
    STATE(5856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220028] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8827), 1,
      aux_sym_quoted_extension_token1,
    STATE(5857), 1,
      sym_attribute,
    STATE(7231), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220048] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5953), 1,
      anon_sym_in,
    ACTIONS(8367), 1,
      anon_sym_and,
    STATE(5618), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(5858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220068] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8829), 1,
      anon_sym_SEMI,
    ACTIONS(8831), 1,
      anon_sym_GT_RBRACE,
    STATE(5566), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220088] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8833), 1,
      anon_sym_LPAREN,
    ACTIONS(8835), 1,
      anon_sym_LBRACK,
    ACTIONS(8837), 1,
      anon_sym_LBRACE,
    STATE(5860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220108] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8839), 1,
      anon_sym_RBRACK,
    STATE(5343), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220128] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8841), 1,
      anon_sym_LPAREN,
    ACTIONS(8843), 1,
      anon_sym_LBRACK,
    ACTIONS(8845), 1,
      anon_sym_LBRACE,
    STATE(5862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220148] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5730), 1,
      sym_attribute_id,
    STATE(5863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220166] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8847), 1,
      anon_sym_SEMI,
    ACTIONS(8849), 1,
      anon_sym_GT_RBRACE,
    STATE(5864), 1,
      sym_attribute,
    STATE(5914), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220186] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8851), 1,
      anon_sym_LPAREN,
    ACTIONS(8853), 1,
      anon_sym_LBRACK,
    ACTIONS(8855), 1,
      anon_sym_LBRACE,
    STATE(5865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220206] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8857), 1,
      anon_sym_LPAREN,
    ACTIONS(8859), 1,
      anon_sym_LBRACK,
    ACTIONS(8861), 1,
      anon_sym_LBRACE,
    STATE(5866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220226] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5738), 1,
      sym_attribute_id,
    STATE(5867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220244] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(13), 1,
      sym_attribute_id,
    STATE(5868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220262] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2756), 1,
      aux_sym_attribute_id_token1,
    STATE(5869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2758), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
  [220280] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8863), 1,
      anon_sym_LPAREN,
    ACTIONS(8865), 1,
      anon_sym_LBRACK,
    ACTIONS(8867), 1,
      anon_sym_LBRACE,
    STATE(5870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220300] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8869), 1,
      anon_sym_EQ,
    STATE(5871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8871), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [220318] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8873), 1,
      anon_sym_LPAREN,
    ACTIONS(8875), 1,
      anon_sym_LBRACK,
    ACTIONS(8877), 1,
      anon_sym_LBRACE,
    STATE(5872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220338] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5743), 1,
      sym_attribute_id,
    STATE(5873), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220356] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7837), 1,
      anon_sym_COMMA,
    ACTIONS(8879), 1,
      anon_sym_RBRACK,
    STATE(5343), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220376] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8881), 1,
      anon_sym_LPAREN,
    ACTIONS(8883), 1,
      anon_sym_LBRACK,
    ACTIONS(8885), 1,
      anon_sym_LBRACE,
    STATE(5875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220396] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8887), 1,
      anon_sym_LPAREN,
    ACTIONS(8889), 1,
      anon_sym_LBRACK,
    ACTIONS(8891), 1,
      anon_sym_LBRACE,
    STATE(5876), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220416] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5749), 1,
      sym_attribute_id,
    STATE(5877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220434] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8893), 1,
      anon_sym_LPAREN,
    ACTIONS(8895), 1,
      anon_sym_LBRACK,
    ACTIONS(8897), 1,
      anon_sym_LBRACE,
    STATE(5878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220454] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8899), 1,
      anon_sym_LPAREN,
    ACTIONS(8901), 1,
      anon_sym_LBRACK,
    ACTIONS(8903), 1,
      anon_sym_LBRACE,
    STATE(5879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220474] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5759), 1,
      sym_attribute_id,
    STATE(5880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220492] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8905), 1,
      anon_sym_SEMI,
    ACTIONS(8907), 1,
      anon_sym_RBRACE,
    STATE(5737), 1,
      aux_sym_record_expression_repeat1,
    STATE(5881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220512] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8909), 1,
      anon_sym_LPAREN,
    ACTIONS(8911), 1,
      anon_sym_LBRACK,
    ACTIONS(8913), 1,
      anon_sym_LBRACE,
    STATE(5882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220532] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8915), 1,
      anon_sym_LPAREN,
    ACTIONS(8917), 1,
      anon_sym_LBRACK,
    ACTIONS(8919), 1,
      anon_sym_LBRACE,
    STATE(5883), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220552] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5766), 1,
      sym_attribute_id,
    STATE(5884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220570] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8921), 1,
      anon_sym_LPAREN,
    ACTIONS(8923), 1,
      anon_sym_LBRACK,
    ACTIONS(8925), 1,
      anon_sym_LBRACE,
    STATE(5885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220590] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8929), 1,
      anon_sym_LBRACK,
    STATE(5886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8927), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
  [220608] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5770), 1,
      sym_attribute_id,
    STATE(5887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220626] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8931), 1,
      anon_sym_LPAREN,
    ACTIONS(8933), 1,
      anon_sym_LBRACK,
    ACTIONS(8935), 1,
      anon_sym_LBRACE,
    STATE(5888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220646] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5772), 1,
      sym_attribute_id,
    STATE(5889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220664] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8937), 1,
      anon_sym_LPAREN,
    ACTIONS(8939), 1,
      anon_sym_LBRACK,
    ACTIONS(8941), 1,
      anon_sym_LBRACE,
    STATE(5890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220684] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5775), 1,
      sym_attribute_id,
    STATE(5891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220702] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8943), 1,
      anon_sym_LPAREN,
    ACTIONS(8945), 1,
      anon_sym_LBRACK,
    ACTIONS(8947), 1,
      anon_sym_LBRACE,
    STATE(5892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220722] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5777), 1,
      sym_attribute_id,
    STATE(5893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220740] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8949), 1,
      anon_sym_LPAREN,
    ACTIONS(8951), 1,
      anon_sym_LBRACK,
    ACTIONS(8953), 1,
      anon_sym_LBRACE,
    STATE(5894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220760] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5780), 1,
      sym_attribute_id,
    STATE(5895), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220778] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8955), 1,
      anon_sym_LPAREN,
    ACTIONS(8957), 1,
      anon_sym_LBRACK,
    ACTIONS(8959), 1,
      anon_sym_LBRACE,
    STATE(5896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220798] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5783), 1,
      sym_attribute_id,
    STATE(5897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220816] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8961), 1,
      anon_sym_LPAREN,
    ACTIONS(8963), 1,
      anon_sym_LBRACK,
    ACTIONS(8965), 1,
      anon_sym_LBRACE,
    STATE(5898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220836] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5785), 1,
      sym_attribute_id,
    STATE(5899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220854] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5787), 1,
      sym_attribute_id,
    STATE(5900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220872] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5789), 1,
      sym_attribute_id,
    STATE(5901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8572), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220890] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7646), 1,
      anon_sym_RBRACE,
    ACTIONS(8967), 1,
      anon_sym_SEMI,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    STATE(5902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220910] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8969), 1,
      aux_sym_quoted_extension_token1,
    STATE(5903), 1,
      sym_attribute,
    STATE(6588), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220930] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5457), 1,
      aux_sym_module_binding_repeat1,
    STATE(5904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220950] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8296), 1,
      anon_sym_RBRACK,
    STATE(5292), 1,
      aux_sym__type_params_repeat1,
    STATE(5905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220970] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8971), 1,
      anon_sym_LPAREN,
    ACTIONS(8973), 1,
      anon_sym_LBRACK,
    ACTIONS(8975), 1,
      anon_sym_LBRACE,
    STATE(5906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220990] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8977), 1,
      anon_sym_RBRACK,
    STATE(5907), 1,
      sym_attribute,
    STATE(5908), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221010] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8979), 1,
      anon_sym_RBRACK,
    STATE(5292), 1,
      aux_sym__type_params_repeat1,
    STATE(5908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221030] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8979), 1,
      anon_sym_RBRACK,
    STATE(5909), 1,
      sym_attribute,
    STATE(5911), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221050] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4938), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221066] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8981), 1,
      anon_sym_RBRACK,
    STATE(5292), 1,
      aux_sym__type_params_repeat1,
    STATE(5911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221086] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8983), 1,
      anon_sym_SEMI,
    ACTIONS(8985), 1,
      anon_sym_RBRACE,
    STATE(5546), 1,
      aux_sym_record_expression_repeat1,
    STATE(5912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221106] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(8987), 1,
      anon_sym_RBRACK,
    STATE(5550), 1,
      aux_sym__type_params_repeat1,
    STATE(5913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221126] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8272), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8989), 1,
      anon_sym_SEMI,
    STATE(5643), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221146] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8991), 1,
      anon_sym_SEMI,
    ACTIONS(8993), 1,
      anon_sym_RBRACE,
    STATE(5580), 1,
      aux_sym_record_pattern_repeat1,
    STATE(5915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221166] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(8995), 1,
      anon_sym_RPAREN,
    STATE(5916), 1,
      sym_attribute,
    STATE(7300), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221186] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1506), 1,
      anon_sym_COLON_GT,
    ACTIONS(8997), 1,
      anon_sym_EQ,
    STATE(5917), 1,
      sym_attribute,
    STATE(6348), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221206] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4523), 1,
      anon_sym_LPAREN,
    STATE(3476), 1,
      sym_module_parameter,
    STATE(5490), 1,
      aux_sym_module_binding_repeat1,
    STATE(5918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221226] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4918), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [221242] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(8999), 1,
      anon_sym_RPAREN,
    STATE(5920), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221262] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9001), 1,
      anon_sym_SEMI,
    ACTIONS(9003), 1,
      anon_sym_RBRACE,
    STATE(5921), 1,
      sym_attribute,
    STATE(5930), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221282] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8002), 1,
      anon_sym_PIPE,
    ACTIONS(8280), 1,
      anon_sym_RBRACK,
    STATE(5346), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221302] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(9005), 1,
      anon_sym_RPAREN,
    STATE(5923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221322] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(3938), 1,
      sym_module_binding,
    STATE(5924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7575), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [221340] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4047), 1,
      sym_module_binding,
    STATE(5925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7575), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [221358] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9007), 1,
      anon_sym_SEMI,
    ACTIONS(9009), 1,
      anon_sym_RBRACE,
    STATE(5588), 1,
      aux_sym_record_binding_pattern_repeat1,
    STATE(5926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221378] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8348), 1,
      sym_indexing_operator,
    ACTIONS(9011), 1,
      sym__lowercase_identifier,
    STATE(5927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221398] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    ACTIONS(9013), 1,
      anon_sym_RPAREN,
    STATE(5928), 1,
      sym_attribute,
    STATE(7316), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221418] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8294), 1,
      anon_sym_COMMA,
    ACTIONS(9015), 1,
      anon_sym_RBRACK,
    STATE(5292), 1,
      aux_sym__type_params_repeat1,
    STATE(5929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221438] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7654), 1,
      anon_sym_RBRACE,
    ACTIONS(9017), 1,
      anon_sym_SEMI,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    STATE(5930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221458] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(4051), 1,
      sym_module_binding,
    STATE(5931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7575), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [221476] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9019), 1,
      anon_sym_RPAREN,
    STATE(5932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9021), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [221494] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(9023), 1,
      anon_sym_type,
    STATE(5702), 1,
      sym_module_path,
    STATE(5933), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221514] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4689), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [221530] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8488), 1,
      anon_sym_COLON2,
    ACTIONS(9025), 1,
      anon_sym_RPAREN,
    STATE(5935), 1,
      sym_attribute,
    STATE(7096), 1,
      sym__class_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221550] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7706), 1,
      anon_sym_RBRACK,
    ACTIONS(9027), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5936), 2,
      sym_attribute,
      aux_sym__sequence_pattern_content_repeat1,
  [221568] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9030), 1,
      anon_sym_SEMI,
    ACTIONS(9032), 1,
      anon_sym_GT_RBRACE,
    STATE(5562), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221588] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8108), 1,
      sym__lowercase_identifier,
    ACTIONS(8112), 1,
      anon_sym_mutable,
    STATE(5631), 1,
      sym_field_declaration,
    STATE(5938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221608] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7917), 1,
      anon_sym_with,
    ACTIONS(8045), 1,
      anon_sym_DASH_GT,
    ACTIONS(9034), 1,
      anon_sym_RPAREN,
    STATE(5939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221628] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3973), 1,
      sym__uppercase_identifier,
    ACTIONS(9036), 1,
      anon_sym_type,
    STATE(5702), 1,
      sym_module_path,
    STATE(5940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221648] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7670), 1,
      anon_sym_RBRACE,
    ACTIONS(9038), 1,
      anon_sym_SEMI,
    STATE(5782), 1,
      aux_sym_record_expression_repeat1,
    STATE(5941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221668] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6318), 1,
      anon_sym_in,
    ACTIONS(8498), 1,
      anon_sym_and,
    STATE(5648), 1,
      aux_sym_class_definition_repeat1,
    STATE(5942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221688] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6318), 1,
      anon_sym_in,
    ACTIONS(8498), 1,
      anon_sym_and,
    STATE(5649), 1,
      aux_sym_class_definition_repeat1,
    STATE(5943), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221708] = 6,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9040), 1,
      anon_sym_SEMI,
    ACTIONS(9042), 1,
      anon_sym_RBRACE,
    STATE(5696), 1,
      aux_sym_record_declaration_repeat1,
    STATE(5944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221728] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9044), 1,
      sym__lowercase_identifier,
    ACTIONS(9046), 1,
      anon_sym_LPAREN,
    STATE(5945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221745] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9048), 1,
      anon_sym_then,
    STATE(2766), 1,
      sym_then_clause,
    STATE(5946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221762] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9050), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221777] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9052), 1,
      anon_sym_DOT,
    STATE(5948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221794] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9054), 1,
      anon_sym_COLON2,
    STATE(5949), 1,
      sym_attribute,
    STATE(6836), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221811] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9056), 1,
      anon_sym_SEMI,
    ACTIONS(9058), 1,
      anon_sym_RBRACE,
    STATE(5950), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221828] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9060), 1,
      anon_sym_RPAREN,
    ACTIONS(9062), 1,
      anon_sym_LT_DASH,
    STATE(5951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221845] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9064), 1,
      anon_sym_SEMI,
    ACTIONS(9066), 1,
      anon_sym_RBRACE,
    STATE(5952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221862] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(5953), 1,
      sym_attribute,
    STATE(6654), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221879] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9068), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [221894] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9070), 1,
      sym__lowercase_identifier,
    ACTIONS(9072), 1,
      anon_sym_LPAREN,
    STATE(5955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221911] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(419), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3620), 1,
      anon_sym_RPAREN,
    STATE(5956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221928] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9074), 1,
      anon_sym_do,
    STATE(2313), 1,
      sym_do_clause,
    STATE(5957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221945] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9076), 1,
      anon_sym_RPAREN,
    ACTIONS(9078), 1,
      anon_sym_SEMI,
    STATE(5958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221962] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9076), 1,
      anon_sym_RBRACK,
    ACTIONS(9080), 1,
      anon_sym_SEMI,
    STATE(5959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221979] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9076), 1,
      anon_sym_RBRACE,
    ACTIONS(9082), 1,
      anon_sym_SEMI,
    STATE(5960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221996] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9084), 1,
      anon_sym_RPAREN,
    ACTIONS(9086), 1,
      anon_sym_LT_DASH,
    STATE(5961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222013] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7983), 1,
      sym__lowercase_identifier,
    STATE(5283), 1,
      aux_sym__abstract_type_repeat1,
    STATE(5962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222030] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4689), 1,
      anon_sym_EQ,
    ACTIONS(9088), 1,
      anon_sym_RPAREN,
    STATE(5963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222047] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9090), 1,
      anon_sym_do,
    STATE(2191), 1,
      sym_do_clause,
    STATE(5964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222064] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9092), 1,
      anon_sym_then,
    STATE(3219), 1,
      sym_then_clause,
    STATE(5965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222081] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9074), 1,
      anon_sym_do,
    STATE(2317), 1,
      sym_do_clause,
    STATE(5966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222098] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9094), 1,
      anon_sym_RPAREN,
    ACTIONS(9096), 1,
      anon_sym_LT_DASH,
    STATE(5967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222115] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9098), 1,
      anon_sym_do,
    STATE(2968), 1,
      sym_do_clause,
    STATE(5968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222132] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9046), 1,
      anon_sym_LPAREN,
    ACTIONS(9100), 1,
      sym__lowercase_identifier,
    STATE(5969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222149] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(5970), 1,
      sym_attribute,
    STATE(6369), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222166] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9102), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [222181] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9104), 1,
      anon_sym_then,
    STATE(3319), 1,
      sym_then_clause,
    STATE(5972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222198] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9106), 1,
      anon_sym_DOT,
    STATE(5973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222215] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9108), 1,
      anon_sym_then,
    STATE(2331), 1,
      sym_then_clause,
    STATE(5974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222232] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(9110), 1,
      anon_sym_EQ,
    STATE(5975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222249] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9112), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [222264] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9114), 1,
      anon_sym_RPAREN,
    ACTIONS(9116), 1,
      anon_sym_SEMI,
    STATE(5977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222281] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9114), 1,
      anon_sym_RBRACK,
    ACTIONS(9118), 1,
      anon_sym_SEMI,
    STATE(5978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222298] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9114), 1,
      anon_sym_RBRACE,
    ACTIONS(9120), 1,
      anon_sym_SEMI,
    STATE(5979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222315] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7049), 1,
      anon_sym_COMMA,
    ACTIONS(9122), 1,
      anon_sym_COLON,
    STATE(5980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222332] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9124), 1,
      anon_sym_then,
    STATE(3212), 1,
      sym_then_clause,
    STATE(5981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222349] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3728), 1,
      anon_sym_RBRACK,
    ACTIONS(7839), 1,
      anon_sym_PIPE,
    STATE(5982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222366] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9126), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [222381] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9128), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [222396] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9130), 1,
      anon_sym_RBRACK,
    ACTIONS(9132), 1,
      anon_sym_SEMI,
    STATE(5985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222413] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7390), 1,
      anon_sym_RBRACE,
    ACTIONS(9134), 1,
      anon_sym_SEMI,
    STATE(5986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222430] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9136), 1,
      sym__lowercase_identifier,
    ACTIONS(9138), 1,
      anon_sym_LBRACK,
    STATE(5987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222447] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9140), 1,
      anon_sym_RPAREN,
    ACTIONS(9142), 1,
      anon_sym_COLON_GT,
    STATE(5988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222464] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9144), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [222479] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7555), 1,
      anon_sym_RBRACE,
    ACTIONS(9146), 1,
      anon_sym_SEMI,
    STATE(5990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222496] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(5991), 1,
      sym_attribute,
    STATE(6643), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222513] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8039), 1,
      sym__lowercase_identifier,
    STATE(5400), 1,
      sym__simple_value_pattern,
    STATE(5992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222530] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9148), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [222545] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(5994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9150), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [222560] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9152), 1,
      anon_sym_EQ,
    ACTIONS(9154), 1,
      anon_sym_RPAREN,
    STATE(5995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222577] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9156), 1,
      anon_sym_RPAREN,
    ACTIONS(9158), 1,
      anon_sym_SEMI,
    STATE(5996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222594] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9156), 1,
      anon_sym_RBRACK,
    ACTIONS(9160), 1,
      anon_sym_SEMI,
    STATE(5997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222611] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9156), 1,
      anon_sym_RBRACE,
    ACTIONS(9162), 1,
      anon_sym_SEMI,
    STATE(5998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222628] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(9164), 1,
      sym__lowercase_identifier,
    STATE(5999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222645] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9166), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [222660] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9168), 1,
      sym__lowercase_identifier,
    ACTIONS(9170), 1,
      anon_sym_LPAREN,
    STATE(6001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222677] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9172), 1,
      anon_sym_then,
    STATE(2215), 1,
      sym_then_clause,
    STATE(6002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222694] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8705), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [222709] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8757), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [222724] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7559), 1,
      anon_sym_RBRACE,
    ACTIONS(9174), 1,
      anon_sym_SEMI,
    STATE(6005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222741] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9054), 1,
      anon_sym_COLON2,
    STATE(6006), 1,
      sym_attribute,
    STATE(6926), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222758] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(9176), 1,
      sym__lowercase_identifier,
    STATE(6007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222775] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6008), 1,
      sym_attribute,
    STATE(6827), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222792] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9178), 1,
      anon_sym_RPAREN,
    ACTIONS(9180), 1,
      anon_sym_SEMI,
    STATE(6009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222809] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9178), 1,
      anon_sym_RBRACK,
    ACTIONS(9182), 1,
      anon_sym_SEMI,
    STATE(6010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222826] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9178), 1,
      anon_sym_RBRACE,
    ACTIONS(9184), 1,
      anon_sym_SEMI,
    STATE(6011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222843] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9054), 1,
      anon_sym_COLON2,
    STATE(4446), 1,
      sym__typed,
    STATE(6012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222860] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6013), 1,
      sym_attribute,
    STATE(7297), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222877] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9186), 2,
      anon_sym_RBRACK,
      aux_sym_tag_token1,
  [222892] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9188), 1,
      anon_sym_RPAREN,
    ACTIONS(9190), 1,
      anon_sym_SEMI,
    STATE(6015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222909] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6016), 1,
      sym_attribute,
    STATE(6910), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222926] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9192), 1,
      anon_sym_RPAREN,
    ACTIONS(9194), 1,
      anon_sym_SEMI,
    STATE(6017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222943] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9192), 1,
      anon_sym_RBRACK,
    ACTIONS(9196), 1,
      anon_sym_SEMI,
    STATE(6018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222960] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9192), 1,
      anon_sym_RBRACE,
    ACTIONS(9198), 1,
      anon_sym_SEMI,
    STATE(6019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222977] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9200), 1,
      anon_sym_RPAREN,
    ACTIONS(9202), 1,
      anon_sym_LT_DASH,
    STATE(6020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222994] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9204), 1,
      anon_sym_RPAREN,
    ACTIONS(9206), 1,
      anon_sym_LT_DASH,
    STATE(6021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223011] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9130), 1,
      anon_sym_RBRACE,
    ACTIONS(9208), 1,
      anon_sym_SEMI,
    STATE(6022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223028] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6023), 1,
      sym_attribute,
    STATE(6956), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223045] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9210), 1,
      anon_sym_RPAREN,
    ACTIONS(9212), 1,
      anon_sym_SEMI,
    STATE(6024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223062] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9210), 1,
      anon_sym_RBRACK,
    ACTIONS(9214), 1,
      anon_sym_SEMI,
    STATE(6025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223079] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9210), 1,
      anon_sym_RBRACE,
    ACTIONS(9216), 1,
      anon_sym_SEMI,
    STATE(6026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223096] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9218), 1,
      anon_sym_then,
    STATE(3152), 1,
      sym_then_clause,
    STATE(6027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223113] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9220), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [223128] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9222), 1,
      anon_sym_then,
    STATE(3329), 1,
      sym_then_clause,
    STATE(6029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223145] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6030), 1,
      sym_attribute,
    STATE(7009), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223162] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9224), 1,
      anon_sym_RPAREN,
    ACTIONS(9226), 1,
      anon_sym_SEMI,
    STATE(6031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223179] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9224), 1,
      anon_sym_RBRACK,
    ACTIONS(9228), 1,
      anon_sym_SEMI,
    STATE(6032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223196] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9224), 1,
      anon_sym_RBRACE,
    ACTIONS(9230), 1,
      anon_sym_SEMI,
    STATE(6033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223213] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9232), 1,
      anon_sym_DOT,
    STATE(6034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223230] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9188), 1,
      anon_sym_RBRACK,
    ACTIONS(9234), 1,
      anon_sym_SEMI,
    STATE(6035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223247] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6036), 1,
      sym_attribute,
    STATE(7050), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223264] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9236), 1,
      anon_sym_RPAREN,
    ACTIONS(9238), 1,
      anon_sym_SEMI,
    STATE(6037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223281] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9236), 1,
      anon_sym_RBRACK,
    ACTIONS(9240), 1,
      anon_sym_SEMI,
    STATE(6038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223298] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9236), 1,
      anon_sym_RBRACE,
    ACTIONS(9242), 1,
      anon_sym_SEMI,
    STATE(6039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223315] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(7136), 1,
      sym__lowercase_identifier,
    STATE(6040), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223332] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6041), 1,
      sym_attribute,
    STATE(7175), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223349] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9244), 1,
      anon_sym_RPAREN,
    ACTIONS(9246), 1,
      anon_sym_LT_DASH,
    STATE(6042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223366] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6043), 1,
      sym_attribute,
    STATE(7100), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223383] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9248), 1,
      anon_sym_RPAREN,
    ACTIONS(9250), 1,
      anon_sym_SEMI,
    STATE(6044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223400] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9248), 1,
      anon_sym_RBRACK,
    ACTIONS(9252), 1,
      anon_sym_SEMI,
    STATE(6045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223417] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9248), 1,
      anon_sym_RBRACE,
    ACTIONS(9254), 1,
      anon_sym_SEMI,
    STATE(6046), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223434] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9256), 2,
      anon_sym_to,
      anon_sym_downto,
  [223449] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(9258), 1,
      anon_sym_EQ,
    STATE(6048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223466] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9260), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [223481] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6050), 1,
      sym_attribute,
    STATE(7157), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223498] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9262), 1,
      anon_sym_RPAREN,
    ACTIONS(9264), 1,
      anon_sym_SEMI,
    STATE(6051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223515] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9262), 1,
      anon_sym_RBRACK,
    ACTIONS(9266), 1,
      anon_sym_SEMI,
    STATE(6052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223532] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9262), 1,
      anon_sym_RBRACE,
    ACTIONS(9268), 1,
      anon_sym_SEMI,
    STATE(6053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223549] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9072), 1,
      anon_sym_LPAREN,
    ACTIONS(9270), 1,
      sym__lowercase_identifier,
    STATE(6054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223566] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(9272), 1,
      anon_sym_RPAREN,
    STATE(6055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223583] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6056), 1,
      sym_attribute,
    STATE(7204), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223600] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9274), 1,
      sym__lowercase_identifier,
    ACTIONS(9276), 1,
      anon_sym_LPAREN,
    STATE(6057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223617] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9278), 1,
      anon_sym_DOT,
    STATE(6058), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223634] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6059), 1,
      sym_attribute,
    STATE(7226), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223651] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4017), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
  [223666] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9280), 1,
      anon_sym_then,
    STATE(2130), 1,
      sym_then_clause,
    STATE(6061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223683] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6062), 1,
      sym_attribute,
    STATE(7244), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223700] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7384), 1,
      anon_sym_RBRACE,
    ACTIONS(9282), 1,
      anon_sym_SEMI,
    STATE(6063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223717] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9092), 1,
      anon_sym_then,
    STATE(3242), 1,
      sym_then_clause,
    STATE(6064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223734] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9284), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [223749] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6066), 1,
      sym_attribute,
    STATE(7073), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223766] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9286), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [223781] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(6068), 1,
      sym_attribute,
    STATE(6350), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223798] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6069), 1,
      sym_attribute,
    STATE(6648), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223815] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(9288), 1,
      sym__lowercase_identifier,
    STATE(6070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223832] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9290), 1,
      aux_sym_type_variable_token1,
    STATE(2656), 1,
      sym_type_variable,
    STATE(6071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223849] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(9292), 1,
      anon_sym_EQ,
    STATE(6072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223866] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9294), 1,
      sym__lowercase_identifier,
    ACTIONS(9296), 1,
      anon_sym_LPAREN,
    STATE(6073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223883] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9298), 1,
      anon_sym_DOT,
    STATE(6074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223900] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9300), 1,
      sym__lowercase_identifier,
    ACTIONS(9302), 1,
      anon_sym_LPAREN,
    STATE(6075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223917] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9304), 2,
      anon_sym_SEMI,
      anon_sym_GT,
  [223932] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(6077), 1,
      sym_attribute,
    STATE(6447), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223949] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9306), 1,
      sym__lowercase_identifier,
    ACTIONS(9308), 1,
      anon_sym_LPAREN,
    STATE(6078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223966] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9310), 1,
      anon_sym_RPAREN,
    ACTIONS(9312), 1,
      anon_sym_LT_DASH,
    STATE(6079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223983] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4817), 2,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
  [223998] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9314), 1,
      sym__lowercase_identifier,
    ACTIONS(9316), 1,
      anon_sym_LBRACK,
    STATE(6081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224015] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9318), 1,
      anon_sym_do,
    STATE(2206), 1,
      sym_do_clause,
    STATE(6082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224032] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9320), 1,
      sym__lowercase_identifier,
    ACTIONS(9322), 1,
      anon_sym_LBRACK,
    STATE(6083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224049] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(6084), 1,
      sym_attribute,
    STATE(6557), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224066] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    STATE(3192), 1,
      sym_extended_module_path,
    STATE(6085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224083] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9324), 1,
      anon_sym_RPAREN,
    ACTIONS(9326), 1,
      anon_sym_LT_DASH,
    STATE(6086), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224100] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9328), 1,
      sym__lowercase_identifier,
    ACTIONS(9330), 1,
      anon_sym_LPAREN,
    STATE(6087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224117] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(8410), 1,
      sym__lowercase_identifier,
    STATE(6088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224134] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9332), 1,
      sym__uppercase_identifier,
    STATE(4588), 1,
      sym_extended_module_path,
    STATE(6089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224151] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9334), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224166] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7478), 1,
      anon_sym_RBRACE,
    ACTIONS(9336), 1,
      anon_sym_SEMI,
    STATE(6091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224183] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9104), 1,
      anon_sym_then,
    STATE(3298), 1,
      sym_then_clause,
    STATE(6092), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224200] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9338), 1,
      anon_sym_RPAREN,
    ACTIONS(9340), 1,
      anon_sym_SEMI,
    STATE(6093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224217] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9342), 1,
      anon_sym_RPAREN,
    ACTIONS(9344), 1,
      anon_sym_LT_DASH,
    STATE(6094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224234] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9346), 1,
      sym__lowercase_identifier,
    ACTIONS(9348), 1,
      sym__uppercase_identifier,
    STATE(6095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224251] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9338), 1,
      anon_sym_RBRACK,
    ACTIONS(9350), 1,
      anon_sym_SEMI,
    STATE(6096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224268] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9338), 1,
      anon_sym_RBRACE,
    ACTIONS(9352), 1,
      anon_sym_SEMI,
    STATE(6097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224285] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6098), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9354), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [224300] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9054), 1,
      anon_sym_COLON2,
    STATE(6099), 1,
      sym_attribute,
    STATE(6832), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224317] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9356), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224332] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9358), 1,
      anon_sym_open,
    STATE(6101), 1,
      sym_attribute,
    STATE(6670), 1,
      sym_open_module,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224349] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9360), 1,
      anon_sym_RPAREN,
    ACTIONS(9362), 1,
      anon_sym_LT_DASH,
    STATE(6102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224366] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9364), 1,
      anon_sym_then,
    STATE(3266), 1,
      sym_then_clause,
    STATE(6103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224383] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(9366), 1,
      sym__lowercase_identifier,
    STATE(6104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224400] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9368), 1,
      anon_sym_COLON2,
    STATE(3747), 1,
      sym__maybe_polymorphic_typed,
    STATE(6105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224417] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9188), 1,
      anon_sym_RBRACE,
    ACTIONS(9370), 1,
      anon_sym_SEMI,
    STATE(6106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224434] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(9372), 1,
      sym__lowercase_identifier,
    STATE(6107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224451] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9074), 1,
      anon_sym_do,
    STATE(2305), 1,
      sym_do_clause,
    STATE(6108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224468] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9374), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224483] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9376), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224498] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9378), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224513] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9380), 1,
      anon_sym_RPAREN,
    ACTIONS(9382), 1,
      anon_sym_LT_DASH,
    STATE(6112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224530] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9384), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [224545] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8437), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [224560] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9386), 2,
      anon_sym_to,
      anon_sym_downto,
  [224575] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9388), 2,
      anon_sym_to,
      anon_sym_downto,
  [224590] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9390), 1,
      anon_sym_DOT,
    STATE(6117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224607] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9392), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224622] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9394), 1,
      anon_sym_RPAREN,
    ACTIONS(9396), 1,
      anon_sym_LT_DASH,
    STATE(6119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224639] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7577), 1,
      anon_sym_COLON2,
    STATE(6113), 1,
      sym__maybe_polymorphic_typed,
    STATE(6120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224656] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(9398), 1,
      sym__lowercase_identifier,
    STATE(6121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224673] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9400), 2,
      anon_sym_to,
      anon_sym_downto,
  [224688] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9402), 1,
      anon_sym_RPAREN,
    ACTIONS(9404), 1,
      anon_sym_LT_DASH,
    STATE(6123), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224705] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9406), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224720] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9130), 1,
      anon_sym_RPAREN,
    ACTIONS(9408), 1,
      anon_sym_SEMI,
    STATE(6125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224737] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9410), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224752] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9412), 1,
      sym__lowercase_identifier,
    ACTIONS(9414), 1,
      anon_sym_LPAREN,
    STATE(6127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224769] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9416), 1,
      anon_sym_DOT,
    STATE(6128), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224786] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9418), 1,
      anon_sym_RPAREN,
    ACTIONS(9420), 1,
      anon_sym_LT_DASH,
    STATE(6129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224803] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6130), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9422), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224818] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9424), 1,
      anon_sym_RPAREN,
    ACTIONS(9426), 1,
      anon_sym_LT_DASH,
    STATE(6131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224835] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9428), 1,
      anon_sym_RPAREN,
    ACTIONS(9430), 1,
      anon_sym_LT_DASH,
    STATE(6132), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224852] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9432), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [224867] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9434), 1,
      anon_sym_then,
    STATE(3312), 1,
      sym_then_clause,
    STATE(6134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224884] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9436), 1,
      anon_sym_then,
    STATE(3402), 1,
      sym_then_clause,
    STATE(6135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224901] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9438), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224916] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9440), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224931] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9442), 1,
      anon_sym_then,
    STATE(2326), 1,
      sym_then_clause,
    STATE(6138), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224948] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9444), 1,
      anon_sym_RPAREN,
    ACTIONS(9446), 1,
      anon_sym_LT_DASH,
    STATE(6139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224965] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9448), 1,
      anon_sym_SEMI,
    ACTIONS(9450), 1,
      anon_sym_RBRACE,
    STATE(6140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224982] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9452), 1,
      anon_sym_RPAREN,
    ACTIONS(9454), 1,
      anon_sym_LT_DASH,
    STATE(6141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224999] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9108), 1,
      anon_sym_then,
    STATE(2339), 1,
      sym_then_clause,
    STATE(6142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225016] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(9456), 1,
      anon_sym_RPAREN,
    STATE(6143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225033] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9458), 1,
      anon_sym_SEMI,
    ACTIONS(9460), 1,
      anon_sym_RBRACE,
    STATE(6144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225050] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9318), 1,
      anon_sym_do,
    STATE(2218), 1,
      sym_do_clause,
    STATE(6145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225067] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(9462), 1,
      sym__lowercase_identifier,
    STATE(6146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225084] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2986), 1,
      anon_sym_COMMA,
    ACTIONS(9464), 1,
      anon_sym_COLON,
    STATE(6147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225101] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9466), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225116] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6149), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9468), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225131] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9470), 1,
      sym__lowercase_identifier,
    ACTIONS(9472), 1,
      anon_sym_LPAREN,
    STATE(6150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225148] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9474), 1,
      anon_sym_RPAREN,
    ACTIONS(9476), 1,
      anon_sym_LT_DASH,
    STATE(6151), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225165] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(7358), 1,
      anon_sym_RBRACK,
    STATE(6152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225182] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9478), 1,
      anon_sym_RPAREN,
    ACTIONS(9480), 1,
      anon_sym_LT_DASH,
    STATE(6153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225199] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9482), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225214] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(9484), 1,
      sym__lowercase_identifier,
    STATE(6155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225231] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(9486), 1,
      sym__lowercase_identifier,
    STATE(6156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225248] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9488), 1,
      anon_sym_RPAREN,
    ACTIONS(9490), 1,
      anon_sym_LT_DASH,
    STATE(6157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225265] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9492), 1,
      anon_sym_RPAREN,
    ACTIONS(9494), 1,
      anon_sym_LT_DASH,
    STATE(6158), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225282] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9436), 1,
      anon_sym_then,
    STATE(3471), 1,
      sym_then_clause,
    STATE(6159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225299] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9496), 1,
      anon_sym_RBRACK,
    ACTIONS(9498), 1,
      anon_sym_SEMI,
    STATE(6160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225316] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9496), 1,
      anon_sym_RBRACE,
    ACTIONS(9500), 1,
      anon_sym_SEMI,
    STATE(6161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225333] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9502), 1,
      anon_sym_RPAREN,
    ACTIONS(9504), 1,
      anon_sym_LT_DASH,
    STATE(6162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225350] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9506), 1,
      sym__lowercase_identifier,
    ACTIONS(9508), 1,
      anon_sym_LBRACK,
    STATE(6163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225367] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9434), 1,
      anon_sym_then,
    STATE(3334), 1,
      sym_then_clause,
    STATE(6164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225384] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9510), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225399] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3738), 1,
      anon_sym_RBRACK,
    ACTIONS(7839), 1,
      anon_sym_PIPE,
    STATE(6166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225416] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9124), 1,
      anon_sym_then,
    STATE(3218), 1,
      sym_then_clause,
    STATE(6167), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225433] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9512), 1,
      anon_sym_DOT,
    STATE(6168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225450] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3955), 1,
      anon_sym_RPAREN,
    ACTIONS(9514), 1,
      aux_sym_number_token1,
    STATE(6169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225467] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9516), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225482] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(6171), 1,
      sym_attribute,
    STATE(7241), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225499] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9518), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225514] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(9520), 1,
      sym__lowercase_identifier,
    STATE(6173), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225531] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9054), 1,
      anon_sym_COLON2,
    STATE(6174), 1,
      sym_attribute,
    STATE(6395), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225548] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9098), 1,
      anon_sym_do,
    STATE(2938), 1,
      sym_do_clause,
    STATE(6175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225565] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6176), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9522), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225580] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(6177), 1,
      sym_attribute,
    STATE(7292), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225597] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9280), 1,
      anon_sym_then,
    STATE(2134), 1,
      sym_then_clause,
    STATE(6178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225614] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6179), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9524), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225629] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9526), 1,
      anon_sym_DOT,
    STATE(6180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225646] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9330), 1,
      anon_sym_LPAREN,
    ACTIONS(9528), 1,
      sym__lowercase_identifier,
    STATE(6181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225663] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3620), 1,
      anon_sym_RPAREN,
    ACTIONS(3624), 1,
      anon_sym_COLON_COLON,
    STATE(6182), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225680] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9358), 1,
      anon_sym_open,
    STATE(6183), 1,
      sym_attribute,
    STATE(6930), 1,
      sym_open_module,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225697] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9098), 1,
      anon_sym_do,
    STATE(2956), 1,
      sym_do_clause,
    STATE(6184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225714] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6185), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6095), 2,
      anon_sym_and,
      anon_sym_in,
  [225729] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(9530), 1,
      sym__lowercase_identifier,
    STATE(6186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225746] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(9532), 1,
      sym__lowercase_identifier,
    STATE(6187), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225763] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9172), 1,
      anon_sym_then,
    STATE(2205), 1,
      sym_then_clause,
    STATE(6188), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225780] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9534), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225795] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9536), 2,
      anon_sym_to,
      anon_sym_downto,
  [225810] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9538), 1,
      anon_sym_RPAREN,
    ACTIONS(9540), 1,
      anon_sym_LT_DASH,
    STATE(6191), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225827] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9542), 2,
      anon_sym_to,
      anon_sym_downto,
  [225842] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6193), 1,
      sym_attribute,
    STATE(7321), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225859] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6194), 1,
      sym_attribute,
    STATE(6488), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225876] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(4448), 1,
      sym__maybe_polymorphic_typed,
    STATE(6195), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225893] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(6196), 1,
      sym_attribute,
    STATE(6544), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225910] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9318), 1,
      anon_sym_do,
    STATE(2228), 1,
      sym_do_clause,
    STATE(6197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225927] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9544), 1,
      anon_sym_RPAREN,
    ACTIONS(9546), 1,
      anon_sym_LT_DASH,
    STATE(6198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225944] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7577), 1,
      anon_sym_COLON2,
    STATE(6133), 1,
      sym__maybe_polymorphic_typed,
    STATE(6199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225961] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9548), 1,
      anon_sym_DOT,
    STATE(6200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225978] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9550), 1,
      anon_sym_DOT,
    STATE(6201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225995] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9552), 1,
      sym__lowercase_identifier,
    ACTIONS(9554), 1,
      anon_sym_LBRACK,
    STATE(6202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226012] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9556), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [226027] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9558), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [226042] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(7144), 1,
      sym__lowercase_identifier,
    STATE(6205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226059] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9560), 1,
      anon_sym_RPAREN,
    ACTIONS(9562), 1,
      anon_sym_SEMI,
    STATE(6206), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226076] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6207), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9564), 2,
      anon_sym_to,
      anon_sym_downto,
  [226091] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6208), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9566), 2,
      anon_sym_to,
      anon_sym_downto,
  [226106] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9368), 1,
      anon_sym_COLON2,
    STATE(3764), 1,
      sym__maybe_polymorphic_typed,
    STATE(6209), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226123] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9560), 1,
      anon_sym_RBRACK,
    ACTIONS(9568), 1,
      anon_sym_SEMI,
    STATE(6210), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226140] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9442), 1,
      anon_sym_then,
    STATE(2304), 1,
      sym_then_clause,
    STATE(6211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226157] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9560), 1,
      anon_sym_RBRACE,
    ACTIONS(9570), 1,
      anon_sym_SEMI,
    STATE(6212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226174] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6213), 1,
      sym_attribute,
    STATE(6999), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226191] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9572), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [226206] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9054), 1,
      anon_sym_COLON2,
    STATE(6215), 1,
      sym_attribute,
    STATE(6595), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226223] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9574), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [226238] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3678), 1,
      anon_sym_RBRACK,
    ACTIONS(7839), 1,
      anon_sym_PIPE,
    STATE(6217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226255] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5864), 2,
      anon_sym_and,
      anon_sym_in,
  [226270] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9576), 1,
      aux_sym_type_variable_token1,
    STATE(4266), 1,
      sym_type_variable,
    STATE(6219), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226287] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9296), 1,
      anon_sym_LPAREN,
    ACTIONS(9578), 1,
      sym__lowercase_identifier,
    STATE(6220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226304] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9580), 1,
      sym__lowercase_identifier,
    ACTIONS(9582), 1,
      anon_sym_LPAREN,
    STATE(6221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226321] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9364), 1,
      anon_sym_then,
    STATE(3284), 1,
      sym_then_clause,
    STATE(6222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226338] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9074), 1,
      anon_sym_do,
    STATE(2315), 1,
      sym_do_clause,
    STATE(6223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226355] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8319), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [226370] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9584), 1,
      sym__lowercase_identifier,
    ACTIONS(9586), 1,
      anon_sym_LPAREN,
    STATE(6225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226387] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8039), 1,
      sym__lowercase_identifier,
    STATE(5391), 1,
      sym__simple_value_pattern,
    STATE(6226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226404] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9098), 1,
      anon_sym_do,
    STATE(2940), 1,
      sym_do_clause,
    STATE(6227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226421] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6228), 1,
      sym_attribute,
    STATE(6400), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226438] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9588), 1,
      anon_sym_DOT,
    STATE(6229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226455] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9590), 1,
      anon_sym_then,
    STATE(2738), 1,
      sym_then_clause,
    STATE(6230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226472] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3269), 1,
      anon_sym_COMMA,
    ACTIONS(9592), 1,
      anon_sym_COLON,
    STATE(6231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226489] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(9594), 1,
      anon_sym_RPAREN,
    STATE(6232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226506] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9596), 1,
      sym__lowercase_identifier,
    ACTIONS(9598), 1,
      anon_sym_LPAREN,
    STATE(6233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226523] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    STATE(5677), 1,
      sym_extended_module_path,
    STATE(6234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226540] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9330), 1,
      anon_sym_LPAREN,
    ACTIONS(9600), 1,
      sym__lowercase_identifier,
    STATE(6235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226557] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9602), 1,
      anon_sym_DOT,
    STATE(6236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226574] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(9604), 1,
      sym__lowercase_identifier,
    STATE(6237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226591] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9606), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [226606] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9608), 1,
      anon_sym_RPAREN,
    ACTIONS(9610), 1,
      anon_sym_COLON_GT,
    STATE(6239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226623] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(9612), 1,
      sym__lowercase_identifier,
    STATE(6240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226640] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9614), 1,
      sym__lowercase_identifier,
    ACTIONS(9616), 1,
      anon_sym_LPAREN,
    STATE(6241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226657] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9618), 1,
      anon_sym_SEMI,
    ACTIONS(9620), 1,
      anon_sym_RBRACE,
    STATE(6242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226674] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9048), 1,
      anon_sym_then,
    STATE(2728), 1,
      sym_then_clause,
    STATE(6243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226691] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9622), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [226706] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9624), 1,
      anon_sym_RPAREN,
    ACTIONS(9626), 1,
      anon_sym_LT_DASH,
    STATE(6245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226723] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2476), 1,
      sym__uppercase_identifier,
    STATE(5744), 1,
      sym_extended_module_path,
    STATE(6246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226740] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(3764), 1,
      sym__maybe_polymorphic_typed,
    STATE(6247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226757] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(9628), 1,
      sym__lowercase_identifier,
    STATE(6248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226774] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9630), 1,
      sym__lowercase_identifier,
    ACTIONS(9632), 1,
      anon_sym_LPAREN,
    STATE(6249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226791] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2478), 1,
      aux_sym_type_variable_token1,
    STATE(2656), 1,
      sym_type_variable,
    STATE(6250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226808] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(3747), 1,
      sym__maybe_polymorphic_typed,
    STATE(6251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226825] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9590), 1,
      anon_sym_then,
    STATE(2742), 1,
      sym_then_clause,
    STATE(6252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226842] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9218), 1,
      anon_sym_then,
    STATE(3185), 1,
      sym_then_clause,
    STATE(6253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226859] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9634), 1,
      anon_sym_DOT,
    STATE(6254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226876] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8482), 2,
      anon_sym_SEMI,
      anon_sym_GT_RBRACE,
  [226891] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9358), 1,
      anon_sym_open,
    STATE(6256), 1,
      sym_attribute,
    STATE(7095), 1,
      sym_open_module,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226908] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9090), 1,
      anon_sym_do,
    STATE(2253), 1,
      sym_do_clause,
    STATE(6257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226925] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6034), 2,
      anon_sym_and,
      anon_sym_in,
  [226940] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9636), 1,
      anon_sym_DOT,
    STATE(6259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226957] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(9638), 1,
      sym__lowercase_identifier,
    STATE(6260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226974] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9054), 1,
      anon_sym_COLON2,
    STATE(4461), 1,
      sym__typed,
    STATE(6261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226991] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9640), 1,
      anon_sym_RPAREN,
    ACTIONS(9642), 1,
      anon_sym_LT_DASH,
    STATE(6262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227008] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6263), 1,
      sym_attribute,
    STATE(6363), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227025] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6264), 1,
      sym_attribute,
    STATE(6912), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227042] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9644), 1,
      anon_sym_RPAREN,
    ACTIONS(9646), 1,
      anon_sym_LT_DASH,
    STATE(6265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227059] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(9648), 1,
      sym__lowercase_identifier,
    STATE(6266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227076] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9650), 1,
      anon_sym_DOT,
    STATE(6267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227093] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9652), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227108] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9654), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227123] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9054), 1,
      anon_sym_COLON2,
    STATE(6270), 1,
      sym_attribute,
    STATE(6358), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227140] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(6271), 1,
      sym_attribute,
    STATE(6853), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227157] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    STATE(6272), 1,
      sym_attribute,
    STATE(6600), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227174] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(4463), 1,
      sym__maybe_polymorphic_typed,
    STATE(6273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227191] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9656), 1,
      anon_sym_SEMI,
    ACTIONS(9658), 1,
      anon_sym_RBRACE,
    STATE(6274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227208] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2356), 1,
      anon_sym_COLON2,
    STATE(6275), 1,
      sym_attribute,
    STATE(6880), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227225] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9660), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227240] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9662), 1,
      anon_sym_RPAREN,
    ACTIONS(9664), 1,
      anon_sym_SEMI,
    STATE(6277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227257] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9662), 1,
      anon_sym_RBRACK,
    ACTIONS(9666), 1,
      anon_sym_SEMI,
    STATE(6278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227274] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9662), 1,
      anon_sym_RBRACE,
    ACTIONS(9668), 1,
      anon_sym_SEMI,
    STATE(6279), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227291] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6280), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9670), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [227306] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9672), 1,
      anon_sym_SEMI,
    ACTIONS(9674), 1,
      anon_sym_RBRACE,
    STATE(6281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227323] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(9676), 1,
      sym__lowercase_identifier,
    STATE(6282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227340] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9678), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [227355] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9680), 1,
      anon_sym_DOT,
    STATE(6284), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227372] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9090), 1,
      anon_sym_do,
    STATE(2152), 1,
      sym_do_clause,
    STATE(6285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227389] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    ACTIONS(9682), 1,
      sym__lowercase_identifier,
    STATE(6286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227406] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6833), 1,
      anon_sym_COLON2,
    STATE(6287), 1,
      sym_attribute,
    STATE(6397), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227423] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(9684), 1,
      anon_sym_RPAREN,
    STATE(6288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227440] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9686), 1,
      anon_sym_EQ,
    ACTIONS(9688), 1,
      anon_sym_RPAREN,
    STATE(6289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227457] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1196), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3628), 1,
      anon_sym_RPAREN,
    STATE(6290), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227474] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6291), 1,
      sym_attribute,
    STATE(6852), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227491] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9072), 1,
      anon_sym_LPAREN,
    ACTIONS(9690), 1,
      sym__lowercase_identifier,
    STATE(6292), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227508] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(9692), 1,
      sym__lowercase_identifier,
    STATE(6293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227525] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6294), 1,
      sym_attribute,
    STATE(7126), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227542] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6295), 1,
      sym_attribute,
    STATE(7255), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227559] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9694), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227574] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9696), 1,
      anon_sym_SEMI,
    ACTIONS(9698), 1,
      anon_sym_RBRACE,
    STATE(6297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227591] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6089), 2,
      anon_sym_and,
      anon_sym_in,
  [227606] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    ACTIONS(9700), 1,
      sym__lowercase_identifier,
    STATE(6299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227623] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9318), 1,
      anon_sym_do,
    STATE(2211), 1,
      sym_do_clause,
    STATE(6300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227640] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9702), 1,
      anon_sym_RPAREN,
    ACTIONS(9704), 1,
      anon_sym_SEMI,
    STATE(6301), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227657] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9702), 1,
      anon_sym_RBRACK,
    ACTIONS(9706), 1,
      anon_sym_SEMI,
    STATE(6302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227674] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7119), 1,
      sym__lowercase_identifier,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    STATE(6303), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227691] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9702), 1,
      anon_sym_RBRACE,
    ACTIONS(9708), 1,
      anon_sym_SEMI,
    STATE(6304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227708] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9710), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [227723] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6306), 1,
      sym_attribute,
    STATE(7164), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227740] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6307), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9712), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [227755] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9714), 1,
      anon_sym_RPAREN,
    ACTIONS(9716), 1,
      anon_sym_LT_DASH,
    STATE(6308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227772] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9718), 1,
      sym__lowercase_identifier,
    ACTIONS(9720), 1,
      sym__uppercase_identifier,
    STATE(6309), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227789] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9722), 1,
      anon_sym_RPAREN,
    ACTIONS(9724), 1,
      anon_sym_LT_DASH,
    STATE(6310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227806] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9726), 2,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
  [227821] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9728), 1,
      sym__lowercase_identifier,
    ACTIONS(9730), 1,
      anon_sym_LPAREN,
    STATE(6312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227838] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9732), 1,
      anon_sym_DOT,
    STATE(6313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227855] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9734), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227870] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9736), 1,
      anon_sym_RPAREN,
    ACTIONS(9738), 1,
      anon_sym_LT_DASH,
    STATE(6315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227887] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9090), 1,
      anon_sym_do,
    STATE(2150), 1,
      sym_do_clause,
    STATE(6316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227904] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6317), 1,
      sym_attribute,
    STATE(6872), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227921] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9740), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [227936] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9742), 1,
      anon_sym_RPAREN,
    ACTIONS(9744), 1,
      anon_sym_LT_DASH,
    STATE(6319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227953] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7412), 1,
      anon_sym_RBRACE,
    ACTIONS(9746), 1,
      anon_sym_SEMI,
    STATE(6320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227970] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9222), 1,
      anon_sym_then,
    STATE(3342), 1,
      sym_then_clause,
    STATE(6321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227987] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(9748), 1,
      anon_sym_RPAREN,
    STATE(6322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228004] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6323), 1,
      sym_attribute,
    STATE(7076), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228021] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2556), 1,
      sym__left_quoted_string_delimiter,
    STATE(6324), 1,
      sym_attribute,
    STATE(6451), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228038] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7428), 1,
      anon_sym_RBRACE,
    ACTIONS(9750), 1,
      anon_sym_SEMI,
    STATE(6325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228055] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4950), 1,
      anon_sym_as,
    ACTIONS(9752), 1,
      anon_sym_EQ,
    STATE(6326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228072] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9754), 1,
      anon_sym_RPAREN,
    ACTIONS(9756), 1,
      anon_sym_LT_DASH,
    STATE(6327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228089] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7374), 1,
      anon_sym_RBRACE,
    ACTIONS(9758), 1,
      anon_sym_SEMI,
    STATE(6328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228106] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9760), 1,
      anon_sym_DOT,
    STATE(6329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228123] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4408), 1,
      anon_sym_LPAREN,
    ACTIONS(9762), 1,
      anon_sym_DOT,
    STATE(6330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228140] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9764), 1,
      anon_sym_RPAREN,
    ACTIONS(9766), 1,
      anon_sym_LT_DASH,
    STATE(6331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228157] = 5,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9496), 1,
      anon_sym_RPAREN,
    ACTIONS(9768), 1,
      anon_sym_SEMI,
    STATE(6332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228174] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    STATE(6333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7488), 2,
      anon_sym_SEMI,
      anon_sym_GT,
  [228189] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9770), 1,
      anon_sym_DOT_DOT,
    STATE(6334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228203] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9772), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228217] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9774), 1,
      anon_sym_in,
    STATE(6336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228231] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9776), 1,
      anon_sym_RPAREN,
    STATE(6337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228245] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9778), 1,
      anon_sym_COMMA,
    STATE(6338), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228259] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3989), 1,
      anon_sym_RPAREN,
    STATE(6339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228273] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9780), 1,
      anon_sym_DASH_GT,
    STATE(6340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228287] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7715), 1,
      anon_sym_end,
    STATE(6341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228301] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4015), 1,
      anon_sym_RPAREN,
    STATE(6342), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228315] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9782), 1,
      anon_sym_RPAREN,
    STATE(6343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228329] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9784), 1,
      anon_sym_SQUOTE2,
    STATE(6344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228343] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9786), 1,
      anon_sym_DQUOTE,
    STATE(6345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228357] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1494), 1,
      anon_sym_COMMA,
    STATE(6346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228371] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9788), 1,
      anon_sym_DOT,
    STATE(6347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228385] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9790), 1,
      anon_sym_EQ,
    STATE(6348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228399] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8542), 1,
      anon_sym_GT_RBRACE,
    STATE(6349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228413] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9792), 1,
      anon_sym_EQ,
    STATE(6350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228427] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9794), 1,
      anon_sym_RPAREN,
    STATE(6351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228441] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8292), 1,
      anon_sym_GT_RBRACE,
    STATE(6352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228455] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9796), 1,
      anon_sym_RPAREN,
    STATE(6353), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228469] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9796), 1,
      anon_sym_RBRACK,
    STATE(6354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228483] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9796), 1,
      anon_sym_RBRACE,
    STATE(6355), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228497] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7782), 1,
      anon_sym_end,
    STATE(6356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228511] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9798), 1,
      anon_sym_RBRACK,
    STATE(6357), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228525] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9800), 1,
      anon_sym_RPAREN,
    STATE(6358), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228539] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9802), 1,
      anon_sym_RBRACK,
    STATE(6359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228553] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9804), 1,
      anon_sym_SQUOTE2,
    STATE(6360), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228567] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9806), 1,
      sym__lowercase_identifier,
    STATE(6361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228581] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9808), 1,
      anon_sym_DQUOTE,
    STATE(6362), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228595] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9810), 1,
      anon_sym_RBRACE,
    STATE(6363), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228609] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9812), 1,
      anon_sym_end,
    STATE(6364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228623] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9814), 1,
      anon_sym_DOT,
    STATE(6365), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228637] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9816), 1,
      anon_sym_RBRACK,
    STATE(6366), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228651] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9818), 1,
      sym__lowercase_identifier,
    STATE(6367), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228665] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9820), 1,
      anon_sym_EQ,
    STATE(6368), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228679] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9822), 1,
      anon_sym_RBRACE,
    STATE(6369), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228693] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9824), 1,
      anon_sym_RBRACK,
    STATE(6370), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228707] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9826), 1,
      anon_sym_RPAREN,
    STATE(6371), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228721] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9828), 1,
      anon_sym_COLON2,
    STATE(6372), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228735] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9830), 1,
      anon_sym_RPAREN,
    STATE(6373), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228749] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9832), 1,
      anon_sym_RBRACE,
    STATE(6374), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228763] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9834), 1,
      anon_sym_RPAREN,
    STATE(6375), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228777] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7809), 1,
      anon_sym_end,
    STATE(6376), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228791] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9836), 1,
      anon_sym_RBRACK,
    STATE(6377), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228805] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9838), 1,
      anon_sym_RBRACK,
    STATE(6378), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228819] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9088), 1,
      anon_sym_RPAREN,
    STATE(6379), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228833] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9840), 1,
      anon_sym_EQ,
    STATE(6380), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228847] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9842), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228861] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9844), 1,
      anon_sym_end,
    STATE(6382), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228875] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9846), 1,
      anon_sym_with,
    STATE(6383), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228889] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9848), 1,
      anon_sym_RBRACE,
    STATE(6384), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228903] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9850), 1,
      anon_sym_SQUOTE2,
    STATE(6385), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228917] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9852), 1,
      anon_sym_EQ,
    STATE(6386), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228931] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9854), 1,
      anon_sym_DASH_GT,
    STATE(6387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228945] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9856), 1,
      anon_sym_RBRACK,
    STATE(6388), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228959] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9858), 1,
      anon_sym_PIPE,
    STATE(6389), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228973] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2734), 1,
      anon_sym_DASH_GT,
    STATE(6390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228987] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4187), 1,
      anon_sym_COMMA,
    STATE(6391), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229001] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9860), 1,
      anon_sym_LT_DASH,
    STATE(6392), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229015] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9862), 1,
      anon_sym_RBRACK,
    STATE(6393), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229029] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9864), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6394), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229043] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9866), 1,
      anon_sym_RPAREN,
    STATE(6395), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229057] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9868), 1,
      anon_sym_DOT,
    STATE(6396), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229071] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9870), 1,
      anon_sym_RPAREN,
    STATE(6397), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229085] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9872), 1,
      anon_sym_in,
    STATE(6398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229099] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9874), 1,
      anon_sym_DQUOTE,
    STATE(6399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229113] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9876), 1,
      anon_sym_RBRACE,
    STATE(6400), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229127] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9878), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6401), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229141] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9880), 1,
      anon_sym_COMMA,
    STATE(6402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229155] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9882), 1,
      anon_sym_end,
    STATE(6403), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229169] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9884), 1,
      anon_sym_DOT,
    STATE(6404), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229183] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9886), 1,
      anon_sym_RPAREN,
    STATE(6405), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229197] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9888), 1,
      anon_sym_RPAREN,
    STATE(6406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229211] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9890), 1,
      anon_sym_DOT,
    STATE(6407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229225] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2376), 1,
      anon_sym_COMMA,
    STATE(6408), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229239] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9892), 1,
      anon_sym_DOT,
    STATE(6409), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229253] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9894), 1,
      anon_sym_EQ,
    STATE(6410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229267] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9896), 1,
      anon_sym_DOT,
    STATE(6411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229281] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9898), 1,
      anon_sym_RPAREN,
    STATE(6412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229295] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9900), 1,
      anon_sym_DOT,
    STATE(6413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229309] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7362), 1,
      anon_sym_COMMA,
    STATE(6414), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229323] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9902), 1,
      anon_sym_STAR,
    STATE(6415), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229337] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9904), 1,
      anon_sym_RBRACK,
    STATE(6416), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229351] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9906), 1,
      anon_sym_in,
    STATE(6417), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229365] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9908), 1,
      anon_sym_with,
    STATE(6418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229379] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9910), 1,
      anon_sym_with,
    STATE(6419), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229393] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9912), 1,
      anon_sym_DOT,
    STATE(6420), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229407] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9914), 1,
      anon_sym_RBRACE,
    STATE(6421), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229421] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9916), 1,
      anon_sym_DOT,
    STATE(6422), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229435] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9918), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6423), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229449] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9920), 1,
      anon_sym_in,
    STATE(6424), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229463] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9922), 1,
      anon_sym_end,
    STATE(6425), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229477] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3048), 1,
      anon_sym_COMMA,
    STATE(6426), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229491] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9924), 1,
      anon_sym_RBRACK,
    STATE(6427), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229505] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6554), 1,
      anon_sym_DASH_GT,
    STATE(6428), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229519] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9926), 1,
      anon_sym_STAR,
    STATE(6429), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229533] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9928), 1,
      anon_sym_RPAREN,
    STATE(6430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229547] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9930), 1,
      anon_sym_RPAREN,
    STATE(6431), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229561] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9932), 1,
      anon_sym_with,
    STATE(6432), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229575] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9934), 1,
      anon_sym_with,
    STATE(6433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229589] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9936), 1,
      anon_sym_EQ,
    STATE(6434), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229603] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9938), 1,
      anon_sym_RPAREN,
    STATE(6435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229617] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9940), 1,
      anon_sym_RPAREN,
    STATE(6436), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229631] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9942), 1,
      anon_sym_DOT,
    STATE(6437), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229645] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9944), 1,
      anon_sym_with,
    STATE(6438), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229659] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9946), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229673] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7563), 1,
      anon_sym_DASH_GT,
    STATE(6440), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229687] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9948), 1,
      anon_sym_in,
    STATE(6441), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229701] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9950), 1,
      anon_sym_with,
    STATE(6442), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229715] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9952), 1,
      anon_sym_DOT,
    STATE(6443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229729] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9954), 1,
      anon_sym_RBRACE,
    STATE(6444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229743] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9956), 1,
      anon_sym_COMMA,
    STATE(6445), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229757] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9958), 1,
      anon_sym_RPAREN,
    STATE(6446), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229771] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9960), 1,
      anon_sym_EQ,
    STATE(6447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229785] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9962), 1,
      anon_sym_in,
    STATE(6448), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229799] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9964), 1,
      anon_sym_in,
    STATE(6449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229813] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9032), 1,
      anon_sym_GT_RBRACE,
    STATE(6450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229827] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9966), 1,
      anon_sym_RBRACE,
    STATE(6451), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229841] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9968), 1,
      anon_sym_RPAREN,
    STATE(6452), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229855] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7306), 1,
      anon_sym_GT,
    STATE(6453), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229869] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9970), 1,
      anon_sym_RPAREN,
    STATE(6454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229883] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9970), 1,
      anon_sym_RBRACK,
    STATE(6455), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229897] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9970), 1,
      anon_sym_RBRACE,
    STATE(6456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229911] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9972), 1,
      anon_sym_RPAREN,
    STATE(6457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229925] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9974), 1,
      anon_sym_RPAREN,
    STATE(6458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229939] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9976), 1,
      anon_sym_DASH_GT,
    STATE(6459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229953] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1204), 1,
      anon_sym_RPAREN,
    STATE(6460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229967] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9978), 1,
      sym__lowercase_identifier,
    STATE(6461), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229981] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7708), 1,
      anon_sym_end,
    STATE(6462), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229995] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9980), 1,
      anon_sym_RBRACE,
    STATE(6463), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230009] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9982), 1,
      anon_sym_RPAREN,
    STATE(6464), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230023] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9984), 1,
      anon_sym_RBRACE,
    STATE(6465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230037] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9986), 1,
      anon_sym_DASH_GT,
    STATE(6466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230051] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9988), 1,
      anon_sym_RBRACK,
    STATE(6467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230065] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2730), 1,
      anon_sym_DASH_GT,
    STATE(6468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230079] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7694), 1,
      anon_sym_RPAREN,
    STATE(6469), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230093] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9990), 1,
      sym__lowercase_identifier,
    STATE(6470), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230107] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9992), 1,
      anon_sym_RBRACE,
    STATE(6471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230121] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9994), 1,
      anon_sym_RPAREN,
    STATE(6472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230135] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9996), 1,
      anon_sym_DASH_GT,
    STATE(6473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230149] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9998), 1,
      anon_sym_RPAREN,
    STATE(6474), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230163] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10000), 1,
      anon_sym_in,
    STATE(6475), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230177] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10002), 1,
      anon_sym_DASH_GT,
    STATE(6476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230191] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10004), 1,
      anon_sym_SQUOTE2,
    STATE(6477), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230205] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10006), 1,
      anon_sym_PLUS_EQ,
    STATE(6478), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230219] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10008), 1,
      anon_sym_DOT,
    STATE(6479), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230233] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10010), 1,
      anon_sym_DOT,
    STATE(6480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230247] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10012), 1,
      anon_sym_end,
    STATE(6481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230261] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10014), 1,
      anon_sym_DQUOTE,
    STATE(6482), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230275] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10016), 1,
      anon_sym_RPAREN,
    STATE(6483), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230289] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10018), 1,
      anon_sym_RPAREN,
    STATE(6484), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230303] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7581), 1,
      anon_sym_DASH_GT,
    STATE(6485), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230317] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10020), 1,
      anon_sym_RBRACK,
    STATE(6486), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230331] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10022), 1,
      anon_sym_DOT,
    STATE(6487), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230345] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10024), 1,
      anon_sym_RBRACE,
    STATE(6488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230359] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10026), 1,
      sym__lowercase_identifier,
    STATE(6489), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230373] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3668), 1,
      anon_sym_COMMA,
    STATE(6490), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230387] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10028), 1,
      anon_sym_LT_DASH,
    STATE(6491), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230401] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10030), 1,
      anon_sym_RPAREN,
    STATE(6492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230415] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10032), 1,
      anon_sym_end,
    STATE(6493), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230429] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3955), 1,
      anon_sym_RPAREN,
    STATE(6494), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230443] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10034), 1,
      anon_sym_RPAREN,
    STATE(6495), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230457] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10036), 1,
      anon_sym_in,
    STATE(6496), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230471] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10038), 1,
      anon_sym_RBRACK,
    STATE(6497), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230485] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10040), 1,
      anon_sym_RBRACE,
    STATE(6498), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230499] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10042), 1,
      anon_sym_EQ,
    STATE(6499), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230513] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10044), 1,
      sym__lowercase_identifier,
    STATE(6500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230527] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10046), 1,
      anon_sym_RPAREN,
    STATE(6501), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230541] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10048), 1,
      anon_sym_DOT,
    STATE(6502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230555] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10050), 1,
      anon_sym_in,
    STATE(6503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230569] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10052), 1,
      anon_sym_DOT,
    STATE(6504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230583] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10054), 1,
      anon_sym_DOT,
    STATE(6505), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230597] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10056), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6506), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230611] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1214), 1,
      anon_sym_RPAREN,
    STATE(6507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230625] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10058), 1,
      anon_sym_with,
    STATE(6508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230639] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10060), 1,
      anon_sym_with,
    STATE(6509), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230653] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10062), 1,
      anon_sym_DOT,
    STATE(6510), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230667] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10064), 1,
      anon_sym_in,
    STATE(6511), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230681] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10066), 1,
      anon_sym_RPAREN,
    STATE(6512), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230695] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3624), 1,
      anon_sym_COLON_COLON,
    STATE(6513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230709] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4211), 1,
      anon_sym_COMMA,
    STATE(6514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230723] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10068), 1,
      anon_sym_in,
    STATE(6515), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230737] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7288), 1,
      anon_sym_GT,
    STATE(6516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230751] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4100), 1,
      anon_sym_COMMA,
    STATE(6517), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230765] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7527), 1,
      anon_sym_DASH_GT,
    STATE(6518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230779] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10070), 1,
      anon_sym_with,
    STATE(6519), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230793] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10072), 1,
      anon_sym_with,
    STATE(6520), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230807] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10074), 1,
      anon_sym_RPAREN,
    STATE(6521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230821] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10076), 1,
      anon_sym_DASH_GT,
    STATE(6522), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230835] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10078), 1,
      anon_sym_RPAREN,
    STATE(6523), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230849] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10080), 1,
      anon_sym_DOT,
    STATE(6524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230863] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7398), 1,
      anon_sym_COMMA,
    STATE(6525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230877] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10082), 1,
      anon_sym_with,
    STATE(6526), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230891] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10084), 1,
      anon_sym_RBRACK,
    STATE(6527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230905] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10086), 1,
      anon_sym_EQ,
    STATE(6528), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230919] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10088), 1,
      anon_sym_COMMA,
    STATE(6529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230933] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10090), 1,
      anon_sym_RPAREN,
    STATE(6530), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230947] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10092), 1,
      anon_sym_DASH_GT,
    STATE(6531), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230961] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10094), 1,
      anon_sym_PLUS_EQ,
    STATE(6532), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230975] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10096), 1,
      anon_sym_RPAREN,
    STATE(6533), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230989] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10098), 1,
      anon_sym_DASH_GT,
    STATE(6534), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231003] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10100), 1,
      anon_sym_RPAREN,
    STATE(6535), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231017] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10102), 1,
      anon_sym_RPAREN,
    STATE(6536), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231031] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10104), 1,
      anon_sym_RPAREN,
    STATE(6537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231045] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10104), 1,
      anon_sym_RBRACK,
    STATE(6538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231059] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10104), 1,
      anon_sym_RBRACE,
    STATE(6539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231073] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10106), 1,
      anon_sym_DASH_GT,
    STATE(6540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231087] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10108), 1,
      anon_sym_RBRACE,
    STATE(6541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231101] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10110), 1,
      anon_sym_SQUOTE2,
    STATE(6542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231115] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10112), 1,
      anon_sym_end,
    STATE(6543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231129] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10114), 1,
      anon_sym_EQ,
    STATE(6544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231143] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10116), 1,
      anon_sym_done,
    STATE(6545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231157] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10118), 1,
      anon_sym_DASH_GT,
    STATE(6546), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231171] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10120), 1,
      anon_sym_COMMA,
    STATE(6547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231185] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10122), 1,
      anon_sym_DOT,
    STATE(6548), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231199] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10124), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231213] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10126), 1,
      anon_sym_DOT,
    STATE(6550), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231227] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10128), 1,
      anon_sym_DOT,
    STATE(6551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231241] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10130), 1,
      anon_sym_DASH_GT,
    STATE(6552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231255] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10132), 1,
      anon_sym_done,
    STATE(6553), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231269] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10134), 1,
      anon_sym_with,
    STATE(6554), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231283] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10136), 1,
      anon_sym_with,
    STATE(6555), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231297] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10138), 1,
      anon_sym_DOT,
    STATE(6556), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231311] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10140), 1,
      anon_sym_EQ,
    STATE(6557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231325] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8558), 1,
      anon_sym_GT_RBRACE,
    STATE(6558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231339] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10142), 1,
      anon_sym_RPAREN,
    STATE(6559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231353] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3628), 1,
      anon_sym_RBRACK,
    STATE(6560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231367] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10144), 1,
      anon_sym_LT_DASH,
    STATE(6561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231381] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10146), 1,
      anon_sym_DASH_GT,
    STATE(6562), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231395] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10148), 1,
      anon_sym_with,
    STATE(6563), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231409] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10150), 1,
      anon_sym_with,
    STATE(6564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231423] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10152), 1,
      anon_sym_RPAREN,
    STATE(6565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231437] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10154), 1,
      anon_sym_RBRACK,
    STATE(6566), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231451] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7611), 1,
      anon_sym_end,
    STATE(6567), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231465] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10156), 1,
      anon_sym_DOT,
    STATE(6568), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231479] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10158), 1,
      anon_sym_RPAREN,
    STATE(6569), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231493] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10160), 1,
      anon_sym_SQUOTE2,
    STATE(6570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231507] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10162), 1,
      anon_sym_DQUOTE,
    STATE(6571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231521] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10164), 1,
      anon_sym_with,
    STATE(6572), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231535] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10166), 1,
      anon_sym_RPAREN,
    STATE(6573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231549] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10168), 1,
      anon_sym_RBRACE,
    STATE(6574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231563] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10170), 1,
      anon_sym_RBRACK,
    STATE(6575), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231577] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10172), 1,
      anon_sym_with,
    STATE(6576), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231591] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10174), 1,
      anon_sym_RPAREN,
    STATE(6577), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231605] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10176), 1,
      sym__lowercase_identifier,
    STATE(6578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231619] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10178), 1,
      anon_sym_RPAREN,
    STATE(6579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231633] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10180), 1,
      anon_sym_RPAREN,
    STATE(6580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231647] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10180), 1,
      anon_sym_RBRACK,
    STATE(6581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231661] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10180), 1,
      anon_sym_RBRACE,
    STATE(6582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231675] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10182), 1,
      anon_sym_RBRACK,
    STATE(6583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231689] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10184), 1,
      anon_sym_RPAREN,
    STATE(6584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231703] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10186), 1,
      anon_sym_RBRACE,
    STATE(6585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231717] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10188), 1,
      anon_sym_DOT,
    STATE(6586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231731] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10190), 1,
      anon_sym_in,
    STATE(6587), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231745] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10192), 1,
      anon_sym_RBRACE,
    STATE(6588), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231759] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10194), 1,
      anon_sym_RPAREN,
    STATE(6589), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231773] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10196), 1,
      anon_sym_RPAREN,
    STATE(6590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231787] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10198), 1,
      anon_sym_DOT,
    STATE(6591), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231801] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10200), 1,
      anon_sym_RBRACE,
    STATE(6592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231815] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10202), 1,
      anon_sym_DOT,
    STATE(6593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231829] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10204), 1,
      anon_sym_DOT,
    STATE(6594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231843] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10206), 1,
      anon_sym_RPAREN,
    STATE(6595), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231857] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10208), 1,
      anon_sym_RBRACK,
    STATE(6596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231871] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10210), 1,
      anon_sym_with,
    STATE(6597), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231885] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10212), 1,
      anon_sym_with,
    STATE(6598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231899] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10214), 1,
      anon_sym_DOT,
    STATE(6599), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231913] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10216), 1,
      anon_sym_RPAREN,
    STATE(6600), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231927] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4424), 1,
      anon_sym_GT,
    STATE(6601), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231941] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10218), 1,
      anon_sym_RPAREN,
    STATE(6602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231955] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10220), 1,
      anon_sym_RBRACE,
    STATE(6603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231969] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10222), 1,
      anon_sym_RPAREN,
    STATE(6604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231983] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10224), 1,
      anon_sym_with,
    STATE(6605), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231997] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10226), 1,
      anon_sym_with,
    STATE(6606), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232011] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10228), 1,
      anon_sym_RPAREN,
    STATE(6607), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232025] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10230), 1,
      anon_sym_DASH_GT,
    STATE(6608), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232039] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7123), 1,
      sym__uppercase_identifier,
    STATE(6609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232053] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10232), 1,
      anon_sym_DOT,
    STATE(6610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232067] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10234), 1,
      anon_sym_DASH_GT,
    STATE(6611), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232081] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10236), 1,
      anon_sym_DASH_GT,
    STATE(6612), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232095] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10238), 1,
      sym__lowercase_identifier,
    STATE(6613), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232109] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10240), 1,
      anon_sym_RPAREN,
    STATE(6614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232123] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10242), 1,
      anon_sym_RPAREN,
    STATE(6615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232137] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10244), 1,
      anon_sym_end,
    STATE(6616), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232151] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10246), 1,
      anon_sym_RPAREN,
    STATE(6617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232165] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10248), 1,
      anon_sym_RPAREN,
    STATE(6618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232179] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10250), 1,
      anon_sym_RPAREN,
    STATE(6619), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232193] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10252), 1,
      anon_sym_RPAREN,
    STATE(6620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232207] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10252), 1,
      anon_sym_RBRACK,
    STATE(6621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232221] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10252), 1,
      anon_sym_RBRACE,
    STATE(6622), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232235] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10254), 1,
      anon_sym_RBRACK,
    STATE(6623), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232249] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10256), 1,
      anon_sym_RBRACE,
    STATE(6624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232263] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10258), 1,
      anon_sym_RBRACK,
    STATE(6625), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232277] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10260), 1,
      anon_sym_RBRACE,
    STATE(6626), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232291] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10262), 1,
      anon_sym_RPAREN,
    STATE(6627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232305] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10264), 1,
      anon_sym_RPAREN,
    STATE(6628), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232319] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9058), 1,
      anon_sym_RBRACE,
    STATE(6629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232333] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10266), 1,
      anon_sym_RBRACK,
    STATE(6630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232347] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10268), 1,
      anon_sym_DOT,
    STATE(6631), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232361] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8793), 1,
      anon_sym_GT_RBRACE,
    STATE(6632), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232375] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10270), 1,
      anon_sym_DOT,
    STATE(6633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232389] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10272), 1,
      anon_sym_DOT,
    STATE(6634), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232403] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9066), 1,
      anon_sym_RBRACE,
    STATE(6635), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232417] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10274), 1,
      anon_sym_RBRACK,
    STATE(6636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232431] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10276), 1,
      anon_sym_with,
    STATE(6637), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232445] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10278), 1,
      anon_sym_with,
    STATE(6638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232459] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10280), 1,
      anon_sym_DOT,
    STATE(6639), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232473] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10282), 1,
      anon_sym_end,
    STATE(6640), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232487] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5680), 1,
      anon_sym_in,
    STATE(6641), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232501] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10284), 1,
      anon_sym_end,
    STATE(6642), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232515] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10286), 1,
      anon_sym_RBRACE,
    STATE(6643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232529] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10288), 1,
      anon_sym_with,
    STATE(6644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232543] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10290), 1,
      anon_sym_with,
    STATE(6645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232557] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10292), 1,
      anon_sym_RPAREN,
    STATE(6646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232571] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7099), 1,
      anon_sym_COMMA,
    STATE(6647), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232585] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10294), 1,
      anon_sym_RBRACE,
    STATE(6648), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232599] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10296), 1,
      anon_sym_DOT,
    STATE(6649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232613] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10298), 1,
      anon_sym_RPAREN,
    STATE(6650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232627] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10300), 1,
      anon_sym_RPAREN,
    STATE(6651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232641] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10302), 1,
      sym__lowercase_identifier,
    STATE(6652), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232655] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10304), 1,
      anon_sym_RPAREN,
    STATE(6653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232669] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10306), 1,
      anon_sym_RBRACE,
    STATE(6654), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232683] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10308), 1,
      anon_sym_RBRACE,
    STATE(6655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232697] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2848), 1,
      anon_sym_COMMA,
    STATE(6656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232711] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10310), 1,
      anon_sym_RPAREN,
    STATE(6657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232725] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10312), 1,
      anon_sym_RPAREN,
    STATE(6658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232739] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10314), 1,
      anon_sym_RPAREN,
    STATE(6659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232753] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10314), 1,
      anon_sym_RBRACK,
    STATE(6660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232767] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10314), 1,
      anon_sym_RBRACE,
    STATE(6661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232781] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10316), 1,
      anon_sym_in,
    STATE(6662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232795] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10318), 1,
      anon_sym_LT_DASH,
    STATE(6663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232809] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10320), 1,
      anon_sym_DASH_GT,
    STATE(6664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232823] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10322), 1,
      anon_sym_RBRACE,
    STATE(6665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232837] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10324), 1,
      anon_sym_DASH_GT,
    STATE(6666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232851] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10326), 1,
      anon_sym_RPAREN,
    STATE(6667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232865] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10328), 1,
      anon_sym_RPAREN,
    STATE(6668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232879] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10330), 1,
      anon_sym_DOT,
    STATE(6669), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232893] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10332), 1,
      anon_sym_in,
    STATE(6670), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232907] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10334), 1,
      anon_sym_DOT,
    STATE(6671), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232921] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10336), 1,
      anon_sym_DOT,
    STATE(6672), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232935] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10338), 1,
      anon_sym_RPAREN,
    STATE(6673), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232949] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7292), 1,
      anon_sym_GT,
    STATE(6674), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232963] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10340), 1,
      anon_sym_with,
    STATE(6675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232977] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10342), 1,
      anon_sym_with,
    STATE(6676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232991] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10344), 1,
      anon_sym_DOT,
    STATE(6677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233005] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10346), 1,
      anon_sym_RPAREN,
    STATE(6678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233019] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10348), 1,
      anon_sym_RPAREN,
    STATE(6679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233033] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10350), 1,
      anon_sym_RPAREN,
    STATE(6680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233047] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10352), 1,
      anon_sym_RBRACK,
    STATE(6681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233061] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10354), 1,
      anon_sym_with,
    STATE(6682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233075] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10356), 1,
      anon_sym_with,
    STATE(6683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233089] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10358), 1,
      anon_sym_RPAREN,
    STATE(6684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233103] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10360), 1,
      anon_sym_COMMA,
    STATE(6685), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233117] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10362), 1,
      anon_sym_RBRACE,
    STATE(6686), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233131] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10364), 1,
      anon_sym_DOT,
    STATE(6687), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233145] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10366), 1,
      anon_sym_DASH_GT,
    STATE(6688), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233159] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10368), 1,
      anon_sym_EQ,
    STATE(6689), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233173] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10370), 1,
      anon_sym_EQ,
    STATE(6690), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233187] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10372), 1,
      anon_sym_RPAREN,
    STATE(6691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233201] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7664), 1,
      anon_sym_end,
    STATE(6692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233215] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10374), 1,
      anon_sym_RPAREN,
    STATE(6693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233229] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10376), 1,
      anon_sym_DOT,
    STATE(6694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233243] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7708), 1,
      anon_sym_RPAREN,
    STATE(6695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233257] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10378), 1,
      anon_sym_RPAREN,
    STATE(6696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233271] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10380), 1,
      anon_sym_RPAREN,
    STATE(6697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233285] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10380), 1,
      anon_sym_RBRACK,
    STATE(6698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233299] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10380), 1,
      anon_sym_RBRACE,
    STATE(6699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233313] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10382), 1,
      anon_sym_RPAREN,
    STATE(6700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233327] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10384), 1,
      anon_sym_DASH_GT,
    STATE(6701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233341] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10386), 1,
      anon_sym_DASH_GT,
    STATE(6702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233355] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10388), 1,
      anon_sym_RPAREN,
    STATE(6703), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233369] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10390), 1,
      sym__lowercase_identifier,
    STATE(6704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233383] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10392), 1,
      anon_sym_RPAREN,
    STATE(6705), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233397] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10394), 1,
      sym__lowercase_identifier,
    STATE(6706), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233411] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10396), 1,
      anon_sym_RPAREN,
    STATE(6707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233425] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10398), 1,
      anon_sym_RPAREN,
    STATE(6708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233439] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10400), 1,
      anon_sym_RBRACK,
    STATE(6709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233453] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10402), 1,
      anon_sym_with,
    STATE(6710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233467] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10404), 1,
      anon_sym_with,
    STATE(6711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233481] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10406), 1,
      anon_sym_DOT,
    STATE(6712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233495] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10408), 1,
      anon_sym_RPAREN,
    STATE(6713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233509] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10410), 1,
      anon_sym_RPAREN,
    STATE(6714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233523] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7931), 1,
      sym__uppercase_identifier,
    STATE(6715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233537] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10412), 1,
      anon_sym_RBRACE,
    STATE(6716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233551] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10414), 1,
      anon_sym_with,
    STATE(6717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233565] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10416), 1,
      anon_sym_with,
    STATE(6718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233579] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10418), 1,
      anon_sym_RPAREN,
    STATE(6719), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233593] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10420), 1,
      aux_sym_number_token1,
    STATE(6720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233607] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4088), 1,
      anon_sym_COMMA,
    STATE(6721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233621] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10422), 1,
      anon_sym_DOT,
    STATE(6722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233635] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10424), 1,
      anon_sym_DOT,
    STATE(6723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233649] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10426), 1,
      anon_sym_RPAREN,
    STATE(6724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233663] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4011), 1,
      anon_sym_RPAREN,
    STATE(6725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233677] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10428), 1,
      anon_sym_RPAREN,
    STATE(6726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233691] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10430), 1,
      anon_sym_LT_DASH,
    STATE(6727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233705] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1242), 1,
      anon_sym_RPAREN,
    STATE(6728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233719] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10432), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233733] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10434), 1,
      anon_sym_RPAREN,
    STATE(6730), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233747] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10436), 1,
      anon_sym_RPAREN,
    STATE(6731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233761] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10438), 1,
      anon_sym_RPAREN,
    STATE(6732), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233775] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10438), 1,
      anon_sym_RBRACK,
    STATE(6733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233789] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10438), 1,
      anon_sym_RBRACE,
    STATE(6734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233803] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10440), 1,
      sym__lowercase_identifier,
    STATE(6735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233817] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7380), 1,
      anon_sym_DASH_GT,
    STATE(6736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233831] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10442), 1,
      anon_sym_DASH_GT,
    STATE(6737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233845] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10444), 1,
      anon_sym_RBRACK,
    STATE(6738), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233859] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10446), 1,
      anon_sym_RPAREN,
    STATE(6739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233873] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10448), 1,
      anon_sym_in,
    STATE(6740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233887] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10450), 1,
      aux_sym_number_token1,
    STATE(6741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233901] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10452), 1,
      anon_sym_SQUOTE2,
    STATE(6742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233915] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10454), 1,
      anon_sym_RPAREN,
    STATE(6743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233929] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10456), 1,
      anon_sym_DOT,
    STATE(6744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233943] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10458), 1,
      anon_sym_RPAREN,
    STATE(6745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233957] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10460), 1,
      anon_sym_RPAREN,
    STATE(6746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233971] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10462), 1,
      anon_sym_DOT,
    STATE(6747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233985] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10464), 1,
      anon_sym_RPAREN,
    STATE(6748), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233999] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10466), 1,
      anon_sym_RBRACK,
    STATE(6749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234013] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10468), 1,
      anon_sym_DASH_GT,
    STATE(6750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234027] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10470), 1,
      anon_sym_RPAREN,
    STATE(6751), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234041] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10472), 1,
      anon_sym_in,
    STATE(6752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234055] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10474), 1,
      anon_sym_RPAREN,
    STATE(6753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234069] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10476), 1,
      anon_sym_RPAREN,
    STATE(6754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234083] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10476), 1,
      anon_sym_RBRACK,
    STATE(6755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234097] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10476), 1,
      anon_sym_RBRACE,
    STATE(6756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234111] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10478), 1,
      anon_sym_EQ,
    STATE(6757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234125] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10480), 1,
      anon_sym_RBRACE,
    STATE(6758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234139] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9514), 1,
      aux_sym_number_token1,
    STATE(6759), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234153] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10482), 1,
      anon_sym_DASH_GT,
    STATE(6760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234167] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10484), 1,
      sym__lowercase_identifier,
    STATE(6761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234181] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10486), 1,
      anon_sym_in,
    STATE(6762), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234195] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10488), 1,
      anon_sym_DQUOTE,
    STATE(6763), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234209] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10490), 1,
      anon_sym_RPAREN,
    STATE(6764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234223] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10492), 1,
      anon_sym_DASH_GT,
    STATE(6765), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234237] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10494), 1,
      anon_sym_RPAREN,
    STATE(6766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234251] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10496), 1,
      anon_sym_RPAREN,
    STATE(6767), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234265] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10496), 1,
      anon_sym_RBRACK,
    STATE(6768), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234279] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10496), 1,
      anon_sym_RBRACE,
    STATE(6769), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234293] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10498), 1,
      anon_sym_RPAREN,
    STATE(6770), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234307] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10500), 1,
      anon_sym_RPAREN,
    STATE(6771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234321] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10502), 1,
      anon_sym_RBRACK,
    STATE(6772), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234335] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10504), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234349] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10506), 1,
      anon_sym_EQ,
    STATE(6774), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234363] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10508), 1,
      anon_sym_in,
    STATE(6775), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234377] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10510), 1,
      anon_sym_RBRACK,
    STATE(6776), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234391] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10512), 1,
      anon_sym_RBRACE,
    STATE(6777), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234405] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10514), 1,
      anon_sym_RPAREN,
    STATE(6778), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234419] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10514), 1,
      anon_sym_RBRACK,
    STATE(6779), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234433] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10514), 1,
      anon_sym_RBRACE,
    STATE(6780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234447] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10516), 1,
      anon_sym_end,
    STATE(6781), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234461] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10518), 1,
      anon_sym_RPAREN,
    STATE(6782), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234475] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10520), 1,
      anon_sym_RPAREN,
    STATE(6783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234489] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10522), 1,
      anon_sym_RPAREN,
    STATE(6784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234503] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10522), 1,
      anon_sym_RBRACK,
    STATE(6785), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234517] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10522), 1,
      anon_sym_RBRACE,
    STATE(6786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234531] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10524), 1,
      anon_sym_RBRACE,
    STATE(6787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234545] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10526), 1,
      anon_sym_RPAREN,
    STATE(6788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234559] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10526), 1,
      anon_sym_RBRACK,
    STATE(6789), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234573] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10526), 1,
      anon_sym_RBRACE,
    STATE(6790), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234587] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10528), 1,
      anon_sym_in,
    STATE(6791), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234601] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10530), 1,
      anon_sym_RBRACE,
    STATE(6792), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234615] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10532), 1,
      anon_sym_RPAREN,
    STATE(6793), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234629] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10534), 1,
      anon_sym_in,
    STATE(6794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234643] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10536), 1,
      anon_sym_DASH_GT,
    STATE(6795), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234657] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10538), 1,
      anon_sym_EQ,
    STATE(6796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234671] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10540), 1,
      anon_sym_EQ,
    STATE(6797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234685] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10542), 1,
      anon_sym_RPAREN,
    STATE(6798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234699] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10542), 1,
      anon_sym_RBRACK,
    STATE(6799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234713] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10542), 1,
      anon_sym_RBRACE,
    STATE(6800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234727] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10544), 1,
      anon_sym_DASH_GT,
    STATE(6801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234741] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10546), 1,
      anon_sym_done,
    STATE(6802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234755] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10548), 1,
      anon_sym_RPAREN,
    STATE(6803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234769] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10550), 1,
      anon_sym_SQUOTE2,
    STATE(6804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234783] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10552), 1,
      anon_sym_DASH_GT,
    STATE(6805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234797] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10554), 1,
      anon_sym_DQUOTE,
    STATE(6806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234811] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10556), 1,
      anon_sym_end,
    STATE(6807), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234825] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10558), 1,
      anon_sym_RPAREN,
    STATE(6808), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234839] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10558), 1,
      anon_sym_RBRACK,
    STATE(6809), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234853] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10558), 1,
      anon_sym_RBRACE,
    STATE(6810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234867] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10560), 1,
      anon_sym_RBRACK,
    STATE(6811), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234881] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10562), 1,
      anon_sym_RPAREN,
    STATE(6812), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234895] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10564), 1,
      anon_sym_RBRACK,
    STATE(6813), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234909] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10566), 1,
      anon_sym_RBRACE,
    STATE(6814), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234923] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10568), 1,
      anon_sym_RBRACE,
    STATE(6815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234937] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10570), 1,
      anon_sym_EQ,
    STATE(6816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234951] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10572), 1,
      anon_sym_in,
    STATE(6817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234965] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10574), 1,
      anon_sym_RPAREN,
    STATE(6818), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234979] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10574), 1,
      anon_sym_RBRACK,
    STATE(6819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234993] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10574), 1,
      anon_sym_RBRACE,
    STATE(6820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235007] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10576), 1,
      anon_sym_RBRACK,
    STATE(6821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235021] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10578), 1,
      anon_sym_RBRACE,
    STATE(6822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235035] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10580), 1,
      anon_sym_EQ,
    STATE(6823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235049] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10582), 1,
      anon_sym_DASH_GT,
    STATE(6824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235063] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10584), 1,
      anon_sym_RPAREN,
    STATE(6825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235077] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10586), 1,
      anon_sym_RPAREN,
    STATE(6826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235091] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10588), 1,
      anon_sym_RBRACE,
    STATE(6827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235105] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10590), 1,
      sym__lowercase_identifier,
    STATE(6828), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235119] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10592), 1,
      anon_sym_DASH_GT,
    STATE(6829), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235133] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10594), 1,
      anon_sym_RPAREN,
    STATE(6830), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235147] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10596), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235161] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10598), 1,
      anon_sym_RPAREN,
    STATE(6832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235175] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10600), 1,
      anon_sym_done,
    STATE(6833), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235189] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10602), 1,
      anon_sym_RPAREN,
    STATE(6834), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235203] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10604), 1,
      sym__lowercase_identifier,
    STATE(6835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235217] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10606), 1,
      anon_sym_RPAREN,
    STATE(6836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235231] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10608), 1,
      anon_sym_DASH_GT,
    STATE(6837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235245] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8717), 1,
      anon_sym_GT_RBRACE,
    STATE(6838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235259] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10610), 1,
      anon_sym_RPAREN,
    STATE(6839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235273] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10612), 1,
      anon_sym_PLUS_EQ,
    STATE(6840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235287] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7694), 1,
      anon_sym_end,
    STATE(6841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235301] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10614), 1,
      anon_sym_RPAREN,
    STATE(6842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235315] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10616), 1,
      anon_sym_RBRACK,
    STATE(6843), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235329] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10618), 1,
      anon_sym_SQUOTE2,
    STATE(6844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235343] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10620), 1,
      anon_sym_SQUOTE2,
    STATE(6845), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235357] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10622), 1,
      anon_sym_DQUOTE,
    STATE(6846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235371] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10624), 1,
      anon_sym_RBRACE,
    STATE(6847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235385] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10626), 1,
      anon_sym_RPAREN,
    STATE(6848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235399] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10628), 1,
      anon_sym_RPAREN,
    STATE(6849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235413] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4428), 1,
      anon_sym_GT,
    STATE(6850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235427] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10630), 1,
      anon_sym_RPAREN,
    STATE(6851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235441] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10632), 1,
      anon_sym_RBRACE,
    STATE(6852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235455] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10634), 1,
      anon_sym_EQ,
    STATE(6853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235469] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10636), 1,
      sym__lowercase_identifier,
    STATE(6854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235483] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10638), 1,
      anon_sym_RPAREN,
    STATE(6855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235497] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10640), 1,
      anon_sym_PLUS_EQ,
    STATE(6856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235511] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10642), 1,
      anon_sym_RBRACK,
    STATE(6857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235525] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10644), 1,
      anon_sym_in,
    STATE(6858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235539] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10646), 1,
      anon_sym_RBRACE,
    STATE(6859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235553] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10648), 1,
      anon_sym_of,
    STATE(6860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235567] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10650), 1,
      anon_sym_RPAREN,
    STATE(6861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235581] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10652), 1,
      anon_sym_end,
    STATE(6862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235595] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10654), 1,
      anon_sym_RBRACE,
    STATE(6863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235609] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10656), 1,
      anon_sym_in,
    STATE(6864), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235623] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10658), 1,
      anon_sym_RPAREN,
    STATE(6865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235637] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10660), 1,
      anon_sym_RPAREN,
    STATE(6866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235651] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10662), 1,
      aux_sym_number_token1,
    STATE(6867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235665] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10664), 1,
      anon_sym_in,
    STATE(6868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235679] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10666), 1,
      sym__right_quoted_string_delimiter,
    STATE(6869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235693] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10668), 1,
      sym__lowercase_identifier,
    STATE(6870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235707] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10670), 1,
      anon_sym_RPAREN,
    STATE(6871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235721] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10672), 1,
      anon_sym_RBRACE,
    STATE(6872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235735] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10674), 1,
      anon_sym_in,
    STATE(6873), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235749] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1224), 1,
      anon_sym_RPAREN,
    STATE(6874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235763] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10676), 1,
      anon_sym_RPAREN,
    STATE(6875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235777] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10678), 1,
      anon_sym_RPAREN,
    STATE(6876), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235791] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10680), 1,
      anon_sym_RBRACK,
    STATE(6877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235805] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10682), 1,
      anon_sym_RBRACE,
    STATE(6878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235819] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10684), 1,
      anon_sym_RPAREN,
    STATE(6879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235833] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10686), 1,
      anon_sym_EQ,
    STATE(6880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235847] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10688), 1,
      anon_sym_RPAREN,
    STATE(6881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235861] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10690), 1,
      anon_sym_DOT,
    STATE(6882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235875] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10692), 1,
      anon_sym_DASH_GT,
    STATE(6883), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235889] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10694), 1,
      anon_sym_RPAREN,
    STATE(6884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235903] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10696), 1,
      anon_sym_RPAREN,
    STATE(6885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235917] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10698), 1,
      anon_sym_RPAREN,
    STATE(6886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235931] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10696), 1,
      anon_sym_RBRACK,
    STATE(6887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235945] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10700), 1,
      anon_sym_RBRACE,
    STATE(6888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235959] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10702), 1,
      anon_sym_DASH_GT,
    STATE(6889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235973] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10696), 1,
      anon_sym_RBRACE,
    STATE(6890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235987] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10704), 1,
      anon_sym_end,
    STATE(6891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236001] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10706), 1,
      anon_sym_RPAREN,
    STATE(6892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236015] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10708), 1,
      anon_sym_SQUOTE2,
    STATE(6893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236029] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10710), 1,
      anon_sym_EQ,
    STATE(6894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236043] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7406), 1,
      anon_sym_DASH_GT,
    STATE(6895), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236057] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10712), 1,
      anon_sym_RPAREN,
    STATE(6896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236071] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10714), 1,
      anon_sym_in,
    STATE(6897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236085] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10716), 1,
      anon_sym_DQUOTE,
    STATE(6898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236099] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10718), 1,
      sym__lowercase_identifier,
    STATE(6899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236113] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10720), 1,
      anon_sym_RPAREN,
    STATE(6900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236127] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10722), 1,
      anon_sym_PIPE,
    STATE(6901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236141] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10724), 1,
      anon_sym_RBRACK,
    STATE(6902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236155] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10726), 1,
      anon_sym_RBRACE,
    STATE(6903), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236169] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10728), 1,
      sym__lowercase_identifier,
    STATE(6904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236183] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10730), 1,
      anon_sym_RBRACK,
    STATE(6905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236197] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10732), 1,
      anon_sym_end,
    STATE(6906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236211] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10734), 1,
      anon_sym_RBRACE,
    STATE(6907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236225] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10736), 1,
      sym__lowercase_identifier,
    STATE(6908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236239] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10738), 1,
      anon_sym_RPAREN,
    STATE(6909), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236253] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10740), 1,
      anon_sym_RBRACE,
    STATE(6910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236267] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10742), 1,
      anon_sym_RPAREN,
    STATE(6911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236281] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10744), 1,
      anon_sym_RBRACE,
    STATE(6912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236295] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7809), 1,
      anon_sym_RPAREN,
    STATE(6913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236309] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10746), 1,
      anon_sym_RPAREN,
    STATE(6914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236323] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9450), 1,
      anon_sym_RBRACE,
    STATE(6915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236337] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10748), 1,
      anon_sym_DOT_DOT,
    STATE(6916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236351] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10750), 1,
      anon_sym_DOT_DOT,
    STATE(6917), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236365] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10752), 1,
      anon_sym_DOT_DOT,
    STATE(6918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236379] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10754), 1,
      sym__lowercase_identifier,
    STATE(6919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236393] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10756), 1,
      anon_sym_DASH_GT,
    STATE(6920), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236407] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10758), 1,
      anon_sym_RPAREN,
    STATE(6921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236421] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10760), 1,
      anon_sym_RPAREN,
    STATE(6922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236435] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10762), 1,
      anon_sym_EQ,
    STATE(6923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236449] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10764), 1,
      sym__lowercase_identifier,
    STATE(6924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236463] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10766), 1,
      anon_sym_EQ,
    STATE(6925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236477] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10768), 1,
      anon_sym_RPAREN,
    STATE(6926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236491] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10770), 1,
      anon_sym_RPAREN,
    STATE(6927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236505] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10772), 1,
      sym__lowercase_identifier,
    STATE(6928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236519] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9460), 1,
      anon_sym_RBRACE,
    STATE(6929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236533] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10774), 1,
      anon_sym_in,
    STATE(6930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236547] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10776), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236561] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10778), 1,
      anon_sym_RPAREN,
    STATE(6932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236575] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10780), 1,
      anon_sym_RPAREN,
    STATE(6933), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236589] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10782), 1,
      sym__uppercase_identifier,
    STATE(6934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236603] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4149), 1,
      anon_sym_DASH_GT,
    STATE(6935), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236617] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1216), 1,
      anon_sym_RPAREN,
    STATE(6936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236631] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10784), 1,
      anon_sym_RPAREN,
    STATE(6937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236645] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10786), 1,
      anon_sym_GT,
    STATE(6938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236659] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10788), 1,
      anon_sym_RPAREN,
    STATE(6939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236673] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10790), 1,
      anon_sym_DASH_GT,
    STATE(6940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236687] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10792), 1,
      sym__lowercase_identifier,
    STATE(6941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236701] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10794), 1,
      anon_sym_DOT,
    STATE(6942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236715] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10796), 1,
      anon_sym_RBRACE,
    STATE(6943), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236729] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10798), 1,
      anon_sym_RPAREN,
    STATE(6944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236743] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10800), 1,
      anon_sym_RBRACK,
    STATE(6945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236757] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10802), 1,
      anon_sym_RBRACE,
    STATE(6946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236771] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10804), 1,
      anon_sym_EQ,
    STATE(6947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236785] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10806), 1,
      sym__lowercase_identifier,
    STATE(6948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236799] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10808), 1,
      anon_sym_SQUOTE2,
    STATE(6949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236813] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10810), 1,
      anon_sym_DQUOTE,
    STATE(6950), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236827] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10812), 1,
      anon_sym_RPAREN,
    STATE(6951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236841] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10814), 1,
      anon_sym_RBRACK,
    STATE(6952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236855] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10816), 1,
      anon_sym_RBRACE,
    STATE(6953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236869] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10818), 1,
      anon_sym_RPAREN,
    STATE(6954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236883] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7334), 1,
      anon_sym_COMMA,
    STATE(6955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236897] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10820), 1,
      anon_sym_RBRACE,
    STATE(6956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236911] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10822), 1,
      anon_sym_in,
    STATE(6957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236925] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10824), 1,
      anon_sym_RPAREN,
    STATE(6958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236939] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10826), 1,
      anon_sym_RPAREN,
    STATE(6959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236953] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2450), 1,
      anon_sym_RBRACK,
    STATE(6960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236967] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10828), 1,
      anon_sym_DASH_GT,
    STATE(6961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236981] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10830), 1,
      ts_builtin_sym_end,
    STATE(6962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236995] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4445), 1,
      anon_sym_GT,
    STATE(6963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237009] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10832), 1,
      anon_sym_RBRACE,
    STATE(6964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237023] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10834), 1,
      anon_sym_in,
    STATE(6965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237037] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10836), 1,
      anon_sym_RBRACE,
    STATE(6966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237051] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10838), 1,
      anon_sym_EQ,
    STATE(6967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237065] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10840), 1,
      anon_sym_RPAREN,
    STATE(6968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237079] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10842), 1,
      sym__lowercase_identifier,
    STATE(6969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237093] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10844), 1,
      anon_sym_RPAREN,
    STATE(6970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237107] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10846), 1,
      anon_sym_RPAREN,
    STATE(6971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237121] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6799), 1,
      anon_sym_DASH_GT,
    STATE(6972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237135] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6485), 1,
      anon_sym_in,
    STATE(6973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237149] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7611), 1,
      anon_sym_RPAREN,
    STATE(6974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237163] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10848), 1,
      anon_sym_RPAREN,
    STATE(6975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237177] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10850), 1,
      anon_sym_RBRACK,
    STATE(6976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237191] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10852), 1,
      anon_sym_RBRACE,
    STATE(6977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237205] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10854), 1,
      anon_sym_RPAREN,
    STATE(6978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237219] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10856), 1,
      anon_sym_with,
    STATE(6979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237233] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7782), 1,
      anon_sym_RPAREN,
    STATE(6980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237247] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10858), 1,
      anon_sym_RPAREN,
    STATE(6981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237261] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1228), 1,
      anon_sym_RPAREN,
    STATE(6982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237275] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10860), 1,
      anon_sym_RPAREN,
    STATE(6983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237289] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10862), 1,
      anon_sym_RPAREN,
    STATE(6984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237303] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10864), 1,
      anon_sym_DASH_GT,
    STATE(6985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237317] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3066), 1,
      anon_sym_COMMA,
    STATE(6986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237331] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7358), 1,
      anon_sym_RBRACK,
    STATE(6987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237345] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10866), 1,
      anon_sym_with,
    STATE(6988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237359] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10868), 1,
      anon_sym_RBRACK,
    STATE(6989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237373] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10870), 1,
      ts_builtin_sym_end,
    STATE(6990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237387] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10872), 1,
      anon_sym_RPAREN,
    STATE(6991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237401] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10874), 1,
      anon_sym_RBRACK,
    STATE(6992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237415] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    STATE(6993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237429] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10876), 1,
      anon_sym_RBRACE,
    STATE(6994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237443] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10878), 1,
      anon_sym_SQUOTE2,
    STATE(6995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237457] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10880), 1,
      anon_sym_DQUOTE,
    STATE(6996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237471] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4371), 1,
      anon_sym_COMMA,
    STATE(6997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237485] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10882), 1,
      anon_sym_LT_DASH,
    STATE(6998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237499] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10884), 1,
      anon_sym_RBRACE,
    STATE(6999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237513] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10886), 1,
      anon_sym_of,
    STATE(7000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237527] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10888), 1,
      anon_sym_RPAREN,
    STATE(7001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237541] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10890), 1,
      anon_sym_in,
    STATE(7002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237555] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10892), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237569] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10894), 1,
      anon_sym_DOT,
    STATE(7004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237583] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10896), 1,
      anon_sym_RBRACK,
    STATE(7005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237597] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10898), 1,
      anon_sym_RPAREN,
    STATE(7006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237611] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10900), 1,
      anon_sym_RBRACE,
    STATE(7007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237625] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10902), 1,
      anon_sym_in,
    STATE(7008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237639] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10904), 1,
      anon_sym_RBRACE,
    STATE(7009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237653] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10906), 1,
      anon_sym_DASH_GT,
    STATE(7010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237667] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10908), 1,
      anon_sym_DASH_GT,
    STATE(7011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237681] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10910), 1,
      anon_sym_RPAREN,
    STATE(7012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237695] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10912), 1,
      anon_sym_of,
    STATE(7013), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237709] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10914), 1,
      anon_sym_RPAREN,
    STATE(7014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237723] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10916), 1,
      anon_sym_type,
    STATE(7015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237737] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10918), 1,
      anon_sym_RPAREN,
    STATE(7016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237751] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10920), 1,
      anon_sym_RPAREN,
    STATE(7017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237765] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10922), 1,
      anon_sym_RPAREN,
    STATE(7018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237779] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10924), 1,
      anon_sym_RBRACK,
    STATE(7019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237793] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7751), 1,
      anon_sym_RPAREN,
    STATE(7020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237807] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10926), 1,
      anon_sym_RBRACE,
    STATE(7021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237821] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10928), 1,
      anon_sym_RBRACE,
    STATE(7022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237835] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10930), 1,
      anon_sym_RPAREN,
    STATE(7023), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237849] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10932), 1,
      anon_sym_RBRACK,
    STATE(7024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237863] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10934), 1,
      sym__lowercase_identifier,
    STATE(7025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237877] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10936), 1,
      anon_sym_RPAREN,
    STATE(7026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237891] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10938), 1,
      anon_sym_RBRACK,
    STATE(7027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237905] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10940), 1,
      anon_sym_in,
    STATE(7028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237919] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10942), 1,
      anon_sym_LT_DASH,
    STATE(7029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237933] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10944), 1,
      anon_sym_DQUOTE,
    STATE(7030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237947] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1200), 1,
      anon_sym_RPAREN,
    STATE(7031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237961] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10946), 1,
      anon_sym_RPAREN,
    STATE(7032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237975] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10948), 1,
      anon_sym_RBRACK,
    STATE(7033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237989] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1194), 1,
      anon_sym_RPAREN,
    STATE(7034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238003] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10950), 1,
      anon_sym_RBRACE,
    STATE(7035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238017] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10952), 1,
      anon_sym_SQUOTE2,
    STATE(7036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238031] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10954), 1,
      anon_sym_RPAREN,
    STATE(7037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238045] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9674), 1,
      anon_sym_RBRACE,
    STATE(7038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238059] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10956), 1,
      anon_sym_RBRACK,
    STATE(7039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238073] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10958), 1,
      anon_sym_of,
    STATE(7040), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238087] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10960), 1,
      anon_sym_DQUOTE,
    STATE(7041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238101] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10962), 1,
      sym__lowercase_identifier,
    STATE(7042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238115] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10964), 1,
      sym__lowercase_identifier,
    STATE(7043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238129] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2932), 1,
      anon_sym_COMMA,
    STATE(7044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238143] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10966), 1,
      anon_sym_RBRACK,
    STATE(7045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238157] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10968), 1,
      anon_sym_LT_DASH,
    STATE(7046), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238171] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10970), 1,
      anon_sym_end,
    STATE(7047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238185] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10972), 1,
      anon_sym_RBRACE,
    STATE(7048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238199] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10974), 1,
      anon_sym_RPAREN,
    STATE(7049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238213] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10976), 1,
      anon_sym_RBRACE,
    STATE(7050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238227] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1230), 1,
      anon_sym_RPAREN,
    STATE(7051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238241] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10978), 1,
      sym__lowercase_identifier,
    STATE(7052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238255] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9698), 1,
      anon_sym_RBRACE,
    STATE(7053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238269] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10980), 1,
      anon_sym_PIPE,
    STATE(7054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238283] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10982), 1,
      anon_sym_RPAREN,
    STATE(7055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238297] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10984), 1,
      sym__lowercase_identifier,
    STATE(7056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238311] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10986), 1,
      anon_sym_RBRACK,
    STATE(7057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238325] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10988), 1,
      anon_sym_RPAREN,
    STATE(7058), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238339] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10990), 1,
      anon_sym_DASH_GT,
    STATE(7059), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238353] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10992), 1,
      anon_sym_DOT_DOT,
    STATE(7060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238367] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10994), 1,
      anon_sym_DOT_DOT,
    STATE(7061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238381] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10996), 1,
      anon_sym_DOT_DOT,
    STATE(7062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238395] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(10998), 1,
      anon_sym_RPAREN,
    STATE(7063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238409] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11000), 1,
      anon_sym_in,
    STATE(7064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238423] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11002), 1,
      anon_sym_RBRACK,
    STATE(7065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238437] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11004), 1,
      aux_sym_number_token1,
    STATE(7066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238451] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11006), 1,
      anon_sym_EQ,
    STATE(7067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238465] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11008), 1,
      anon_sym_in,
    STATE(7068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238479] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11010), 1,
      anon_sym_RBRACK,
    STATE(7069), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238493] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1208), 1,
      anon_sym_RPAREN,
    STATE(7070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238507] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11012), 1,
      anon_sym_RPAREN,
    STATE(7071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238521] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11014), 1,
      anon_sym_EQ,
    STATE(7072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238535] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11016), 1,
      anon_sym_RBRACE,
    STATE(7073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238549] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11018), 1,
      anon_sym_RBRACE,
    STATE(7074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238563] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11020), 1,
      sym__lowercase_identifier,
    STATE(7075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238577] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11022), 1,
      anon_sym_RBRACE,
    STATE(7076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238591] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11024), 1,
      anon_sym_PLUS_EQ,
    STATE(7077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238605] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11026), 1,
      anon_sym_in,
    STATE(7078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238619] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11028), 1,
      anon_sym_in,
    STATE(7079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238633] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11030), 1,
      anon_sym_RBRACE,
    STATE(7080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238647] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11032), 1,
      anon_sym_RBRACK,
    STATE(7081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238661] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11034), 1,
      anon_sym_SQUOTE2,
    STATE(7082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238675] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11036), 1,
      sym__lowercase_identifier,
    STATE(7083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238689] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11038), 1,
      anon_sym_RPAREN,
    STATE(7084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238703] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11040), 1,
      anon_sym_RBRACE,
    STATE(7085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238717] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11042), 1,
      sym__lowercase_identifier,
    STATE(7086), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238731] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11044), 1,
      anon_sym_DQUOTE,
    STATE(7087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238745] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11046), 1,
      anon_sym_RBRACK,
    STATE(7088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238759] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11048), 1,
      anon_sym_DOT_DOT,
    STATE(7089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238773] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11050), 1,
      anon_sym_DOT_DOT,
    STATE(7090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238787] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11052), 1,
      anon_sym_DOT_DOT,
    STATE(7091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238801] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11054), 1,
      anon_sym_RPAREN,
    STATE(7092), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238815] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11056), 1,
      anon_sym_RPAREN,
    STATE(7093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238829] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11058), 1,
      anon_sym_RPAREN,
    STATE(7094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238843] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11060), 1,
      anon_sym_in,
    STATE(7095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238857] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11062), 1,
      anon_sym_RPAREN,
    STATE(7096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238871] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11064), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238885] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7664), 1,
      anon_sym_RPAREN,
    STATE(7098), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238899] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11066), 1,
      anon_sym_RBRACE,
    STATE(7099), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238913] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11068), 1,
      anon_sym_RBRACE,
    STATE(7100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238927] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3720), 1,
      anon_sym_DASH_GT,
    STATE(7101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238941] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11070), 1,
      anon_sym_RBRACK,
    STATE(7102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238955] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11072), 1,
      anon_sym_RPAREN,
    STATE(7103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238969] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11074), 1,
      anon_sym_RPAREN,
    STATE(7104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238983] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11076), 1,
      anon_sym_RPAREN,
    STATE(7105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238997] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11078), 1,
      sym__lowercase_identifier,
    STATE(7106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239011] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11080), 1,
      anon_sym_RPAREN,
    STATE(7107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239025] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11082), 1,
      anon_sym_DOT_DOT,
    STATE(7108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239039] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11084), 1,
      anon_sym_DOT_DOT,
    STATE(7109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239053] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11086), 1,
      anon_sym_DOT_DOT,
    STATE(7110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239067] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8526), 1,
      anon_sym_GT_RBRACE,
    STATE(7111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239081] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8849), 1,
      anon_sym_GT_RBRACE,
    STATE(7112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239095] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11088), 1,
      anon_sym_end,
    STATE(7113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239109] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4199), 1,
      anon_sym_COMMA,
    STATE(7114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239123] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11090), 1,
      anon_sym_DOT,
    STATE(7115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239137] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11092), 1,
      anon_sym_STAR,
    STATE(7116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239151] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7620), 1,
      anon_sym_end,
    STATE(7117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239165] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(6793), 1,
      anon_sym_SQUOTE2,
    STATE(7118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239179] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11094), 1,
      anon_sym_LT_DASH,
    STATE(7119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239193] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11096), 1,
      anon_sym_RBRACK,
    STATE(7120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239207] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11098), 1,
      sym__lowercase_identifier,
    STATE(7121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239221] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11100), 1,
      anon_sym_in,
    STATE(7122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239235] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11102), 1,
      anon_sym_DOT_DOT,
    STATE(7123), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239249] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11104), 1,
      anon_sym_DOT_DOT,
    STATE(7124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239263] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11106), 1,
      anon_sym_DOT_DOT,
    STATE(7125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239277] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11108), 1,
      anon_sym_RBRACE,
    STATE(7126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239291] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11110), 1,
      anon_sym_SQUOTE2,
    STATE(7127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239305] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11112), 1,
      anon_sym_DQUOTE,
    STATE(7128), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239319] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(5473), 1,
      anon_sym_in,
    STATE(7129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239333] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1218), 1,
      anon_sym_RPAREN,
    STATE(7130), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239347] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(3620), 1,
      anon_sym_RBRACK,
    STATE(7131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239361] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11114), 1,
      anon_sym_RPAREN,
    STATE(7132), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239375] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11116), 1,
      anon_sym_DASH_GT,
    STATE(7133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239389] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11118), 1,
      anon_sym_GT,
    STATE(7134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239403] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11120), 1,
      sym__lowercase_identifier,
    STATE(7135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239417] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(1196), 1,
      anon_sym_COLON_COLON,
    STATE(7136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239431] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11122), 1,
      anon_sym_DOT_DOT,
    STATE(7137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239445] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11124), 1,
      anon_sym_DOT_DOT,
    STATE(7138), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239459] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11126), 1,
      anon_sym_DOT_DOT,
    STATE(7139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239473] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11128), 1,
      anon_sym_in,
    STATE(7140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239487] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11130), 1,
      anon_sym_RBRACE,
    STATE(7141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239501] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11132), 1,
      anon_sym_DASH_GT,
    STATE(7142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239515] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11134), 1,
      sym__lowercase_identifier,
    STATE(7143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239529] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11136), 1,
      anon_sym_DASH_GT,
    STATE(7144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239543] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11138), 1,
      anon_sym_SQUOTE2,
    STATE(7145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239557] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11140), 1,
      anon_sym_PLUS_EQ,
    STATE(7146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239571] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11142), 1,
      anon_sym_RBRACK,
    STATE(7147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239585] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11144), 1,
      anon_sym_DQUOTE,
    STATE(7148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239599] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11146), 1,
      anon_sym_RBRACE,
    STATE(7149), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239613] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11148), 1,
      anon_sym_RBRACK,
    STATE(7150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239627] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11150), 1,
      anon_sym_DOT_DOT,
    STATE(7151), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239641] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11152), 1,
      anon_sym_DOT_DOT,
    STATE(7152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239655] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11154), 1,
      anon_sym_DOT_DOT,
    STATE(7153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239669] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11156), 1,
      anon_sym_RBRACE,
    STATE(7154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239683] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11158), 1,
      anon_sym_RBRACK,
    STATE(7155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239697] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11160), 1,
      anon_sym_end,
    STATE(7156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239711] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11162), 1,
      anon_sym_RBRACE,
    STATE(7157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239725] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7314), 1,
      anon_sym_DASH_GT,
    STATE(7158), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239739] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9620), 1,
      anon_sym_RBRACE,
    STATE(7159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239753] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11164), 1,
      anon_sym_in,
    STATE(7160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239767] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11166), 1,
      anon_sym_RPAREN,
    STATE(7161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239781] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11168), 1,
      anon_sym_DASH_GT,
    STATE(7162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239795] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11170), 1,
      sym__lowercase_identifier,
    STATE(7163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239809] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11172), 1,
      anon_sym_RBRACE,
    STATE(7164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239823] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11174), 1,
      anon_sym_DOT_DOT,
    STATE(7165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239837] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11176), 1,
      anon_sym_DOT_DOT,
    STATE(7166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239851] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11178), 1,
      anon_sym_DOT_DOT,
    STATE(7167), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239865] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11180), 1,
      anon_sym_RPAREN,
    STATE(7168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239879] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11182), 1,
      anon_sym_RBRACK,
    STATE(7169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239893] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11184), 1,
      sym__lowercase_identifier,
    STATE(7170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239907] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11186), 1,
      anon_sym_RPAREN,
    STATE(7171), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239921] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2726), 1,
      anon_sym_DASH_GT,
    STATE(7172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239935] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11188), 1,
      anon_sym_RBRACK,
    STATE(7173), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239949] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_COMMA,
    STATE(7174), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239963] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11190), 1,
      anon_sym_RBRACE,
    STATE(7175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239977] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11192), 1,
      anon_sym_LT_DASH,
    STATE(7176), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239991] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11194), 1,
      anon_sym_RPAREN,
    STATE(7177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240005] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11196), 1,
      anon_sym_COMMA,
    STATE(7178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240019] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11198), 1,
      anon_sym_DOT_DOT,
    STATE(7179), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240033] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11200), 1,
      anon_sym_DOT_DOT,
    STATE(7180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240047] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11202), 1,
      anon_sym_DOT_DOT,
    STATE(7181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240061] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11204), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7182), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240075] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(9658), 1,
      anon_sym_RBRACE,
    STATE(7183), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240089] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11206), 1,
      anon_sym_RPAREN,
    STATE(7184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240103] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11208), 1,
      anon_sym_RPAREN,
    STATE(7185), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240117] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11210), 1,
      anon_sym_DOT_DOT,
    STATE(7186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240131] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11212), 1,
      anon_sym_DOT_DOT,
    STATE(7187), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240145] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11214), 1,
      anon_sym_DOT_DOT,
    STATE(7188), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240159] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11216), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240173] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11218), 1,
      anon_sym_RBRACE,
    STATE(7190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240187] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11220), 1,
      anon_sym_DOT_DOT,
    STATE(7191), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240201] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11222), 1,
      anon_sym_DOT_DOT,
    STATE(7192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240215] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11224), 1,
      anon_sym_DOT_DOT,
    STATE(7193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240229] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11226), 1,
      anon_sym_DQUOTE,
    STATE(7194), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240243] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11228), 1,
      anon_sym_RBRACK,
    STATE(7195), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240257] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11230), 1,
      anon_sym_DOT_DOT,
    STATE(7196), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240271] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11232), 1,
      anon_sym_DOT_DOT,
    STATE(7197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240285] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11234), 1,
      anon_sym_DOT_DOT,
    STATE(7198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240299] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11236), 1,
      anon_sym_RBRACE,
    STATE(7199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240313] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7743), 1,
      anon_sym_RPAREN,
    STATE(7200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240327] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11238), 1,
      anon_sym_DOT_DOT,
    STATE(7201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240341] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11240), 1,
      anon_sym_DOT_DOT,
    STATE(7202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240355] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11242), 1,
      anon_sym_DOT_DOT,
    STATE(7203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240369] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11244), 1,
      anon_sym_RBRACE,
    STATE(7204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240383] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4076), 1,
      anon_sym_COMMA,
    STATE(7205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240397] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11246), 1,
      anon_sym_DOT_DOT,
    STATE(7206), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240411] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11248), 1,
      anon_sym_DOT_DOT,
    STATE(7207), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240425] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11250), 1,
      anon_sym_DOT_DOT,
    STATE(7208), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240439] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11252), 1,
      anon_sym_LT_DASH,
    STATE(7209), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240453] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11254), 1,
      anon_sym_RPAREN,
    STATE(7210), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240467] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11256), 1,
      anon_sym_DOT_DOT,
    STATE(7211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240481] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11258), 1,
      anon_sym_DOT_DOT,
    STATE(7212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240495] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4025), 1,
      anon_sym_RPAREN,
    STATE(7213), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240509] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8467), 1,
      anon_sym_GT_RBRACE,
    STATE(7214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240523] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4225), 1,
      anon_sym_COMMA,
    STATE(7215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240537] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11260), 1,
      anon_sym_DOT_DOT,
    STATE(7216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240551] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11262), 1,
      anon_sym_DOT_DOT,
    STATE(7217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240565] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11264), 1,
      anon_sym_DOT_DOT,
    STATE(7218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240579] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11266), 1,
      anon_sym_LT_DASH,
    STATE(7219), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240593] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11268), 1,
      anon_sym_RBRACK,
    STATE(7220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240607] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11270), 1,
      anon_sym_RBRACE,
    STATE(7221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240621] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    STATE(7222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240635] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11272), 1,
      anon_sym_RBRACE,
    STATE(7223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240649] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11274), 1,
      anon_sym_LT_DASH,
    STATE(7224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240663] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11276), 1,
      anon_sym_RPAREN,
    STATE(7225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240677] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11278), 1,
      anon_sym_RBRACE,
    STATE(7226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240691] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11280), 1,
      anon_sym_type,
    STATE(7227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240705] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11282), 1,
      anon_sym_in,
    STATE(7228), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240719] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11284), 1,
      sym_indexing_operator,
    STATE(7229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240733] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11286), 1,
      anon_sym_RPAREN,
    STATE(7230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240747] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11288), 1,
      anon_sym_RBRACE,
    STATE(7231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240761] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11290), 1,
      anon_sym_RPAREN,
    STATE(7232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240775] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11292), 1,
      anon_sym_RBRACK,
    STATE(7233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240789] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11294), 1,
      anon_sym_DASH_GT,
    STATE(7234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240803] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11296), 1,
      anon_sym_in,
    STATE(7235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240817] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7715), 1,
      anon_sym_RPAREN,
    STATE(7236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240831] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11298), 1,
      anon_sym_DOT_DOT,
    STATE(7237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240845] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11300), 1,
      anon_sym_DOT,
    STATE(7238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240859] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11302), 1,
      anon_sym_DOT,
    STATE(7239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240873] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11304), 1,
      anon_sym_DOT,
    STATE(7240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240887] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11306), 1,
      anon_sym_EQ,
    STATE(7241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240901] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11308), 1,
      anon_sym_RPAREN,
    STATE(7242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240915] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11310), 1,
      anon_sym_RBRACE,
    STATE(7243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240929] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11312), 1,
      anon_sym_RBRACE,
    STATE(7244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240943] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7751), 1,
      anon_sym_end,
    STATE(7245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240957] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11314), 1,
      anon_sym_DASH_GT,
    STATE(7246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240971] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11316), 1,
      sym__lowercase_identifier,
    STATE(7247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240985] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11318), 1,
      anon_sym_DOT,
    STATE(7248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240999] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7620), 1,
      anon_sym_RPAREN,
    STATE(7249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241013] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11320), 1,
      anon_sym_DOT_DOT,
    STATE(7250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241027] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11322), 1,
      anon_sym_DOT,
    STATE(7251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241041] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11324), 1,
      anon_sym_DOT,
    STATE(7252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241055] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11326), 1,
      anon_sym_end,
    STATE(7253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241069] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11328), 1,
      anon_sym_RPAREN,
    STATE(7254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241083] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11330), 1,
      anon_sym_RBRACE,
    STATE(7255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241097] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7533), 1,
      anon_sym_COMMA,
    STATE(7256), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241111] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11332), 1,
      aux_sym_number_token1,
    STATE(7257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241125] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11334), 1,
      anon_sym_in,
    STATE(7258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241139] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11336), 1,
      anon_sym_SQUOTE2,
    STATE(7259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241153] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11338), 1,
      anon_sym_DOT_DOT,
    STATE(7260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241167] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11340), 1,
      anon_sym_DQUOTE,
    STATE(7261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241181] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11342), 1,
      anon_sym_type,
    STATE(7262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241195] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11344), 1,
      sym_indexing_operator,
    STATE(7263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241209] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11346), 1,
      anon_sym_RPAREN,
    STATE(7264), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241223] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11348), 1,
      anon_sym_LT_DASH,
    STATE(7265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241237] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11350), 1,
      sym_indexing_operator,
    STATE(7266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241251] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11352), 1,
      anon_sym_STAR,
    STATE(7267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241265] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11354), 1,
      anon_sym_type,
    STATE(7268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241279] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11356), 1,
      sym_indexing_operator,
    STATE(7269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241293] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11358), 1,
      sym_indexing_operator,
    STATE(7270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241307] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11360), 1,
      sym_indexing_operator,
    STATE(7271), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241321] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11362), 1,
      sym_indexing_operator,
    STATE(7272), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241335] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11364), 1,
      sym_indexing_operator,
    STATE(7273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241349] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11366), 1,
      sym_indexing_operator,
    STATE(7274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241363] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11368), 1,
      sym_indexing_operator,
    STATE(7275), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241377] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11370), 1,
      sym_indexing_operator,
    STATE(7276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241391] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11372), 1,
      sym_indexing_operator,
    STATE(7277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241405] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11374), 1,
      sym_indexing_operator,
    STATE(7278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241419] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11376), 1,
      sym_indexing_operator,
    STATE(7279), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241433] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11378), 1,
      sym_indexing_operator,
    STATE(7280), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241447] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11380), 1,
      sym_indexing_operator,
    STATE(7281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241461] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11382), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241475] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11384), 1,
      anon_sym_in,
    STATE(7283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241489] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11386), 1,
      anon_sym_with,
    STATE(7284), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241503] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11388), 1,
      anon_sym_in,
    STATE(7285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241517] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11390), 1,
      anon_sym_RBRACK,
    STATE(7286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241531] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11392), 1,
      anon_sym_with,
    STATE(7287), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241545] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11394), 1,
      anon_sym_DOT,
    STATE(7288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241559] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11396), 1,
      sym__lowercase_identifier,
    STATE(7289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241573] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11398), 1,
      anon_sym_DASH_GT,
    STATE(7290), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241587] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11400), 1,
      anon_sym_GT,
    STATE(7291), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241601] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11402), 1,
      anon_sym_EQ,
    STATE(7292), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241615] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11404), 1,
      anon_sym_DOT,
    STATE(7293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241629] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11406), 1,
      anon_sym_RPAREN,
    STATE(7294), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241643] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11408), 1,
      anon_sym_RBRACK,
    STATE(7295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241657] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11410), 1,
      anon_sym_RBRACE,
    STATE(7296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241671] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11412), 1,
      anon_sym_RBRACE,
    STATE(7297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241685] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11414), 1,
      sym__uppercase_identifier,
    STATE(7298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241699] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11416), 1,
      anon_sym_in,
    STATE(7299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241713] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11418), 1,
      anon_sym_RPAREN,
    STATE(7300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241727] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11420), 1,
      anon_sym_RBRACK,
    STATE(7301), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241741] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11422), 1,
      anon_sym_RPAREN,
    STATE(7302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241755] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7316), 1,
      anon_sym_COMMA,
    STATE(7303), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241769] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11424), 1,
      anon_sym_RBRACE,
    STATE(7304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241783] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11426), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241797] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11428), 1,
      anon_sym_COMMA,
    STATE(7306), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241811] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(7743), 1,
      anon_sym_end,
    STATE(7307), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241825] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11430), 1,
      anon_sym_in,
    STATE(7308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241839] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11432), 1,
      anon_sym_RPAREN,
    STATE(7309), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241853] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11434), 1,
      anon_sym_RPAREN,
    STATE(7310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241867] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11436), 1,
      anon_sym_RPAREN,
    STATE(7311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241881] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11438), 1,
      anon_sym_RBRACK,
    STATE(7312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241895] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11440), 1,
      anon_sym_RPAREN,
    STATE(7313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241909] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11442), 1,
      anon_sym_RBRACE,
    STATE(7314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241923] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11444), 1,
      anon_sym_RPAREN,
    STATE(7315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241937] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11446), 1,
      anon_sym_RPAREN,
    STATE(7316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241951] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11448), 1,
      anon_sym_RBRACK,
    STATE(7317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241965] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11450), 1,
      sym_indexing_operator,
    STATE(7318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241979] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11452), 1,
      anon_sym_in,
    STATE(7319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241993] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11454), 1,
      anon_sym_in,
    STATE(7320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242007] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11456), 1,
      anon_sym_RBRACE,
    STATE(7321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242021] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11458), 1,
      anon_sym_with,
    STATE(7322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242035] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(419), 1,
      anon_sym_COLON_COLON,
    STATE(7323), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242049] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11460), 1,
      anon_sym_with,
    STATE(7324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242063] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11462), 1,
      anon_sym_EQ,
    STATE(7325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242077] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11464), 1,
      anon_sym_end,
    STATE(7326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242091] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(8831), 1,
      anon_sym_GT_RBRACE,
    STATE(7327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242105] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11466), 1,
      anon_sym_type,
    STATE(7328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242119] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11468), 1,
      anon_sym_DOT,
    STATE(7329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242133] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11470), 1,
      anon_sym_RPAREN,
    STATE(7330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242147] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11472), 1,
      anon_sym_of,
    STATE(7331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242161] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11474), 1,
      anon_sym_RPAREN,
    STATE(7332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242175] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11476), 1,
      anon_sym_DASH_GT,
    STATE(7333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242189] = 4,
    ACTIONS(405), 1,
      aux_sym_attribute_token1,
    ACTIONS(11478), 1,
      sym__lowercase_identifier,
    STATE(7334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242203] = 1,
    ACTIONS(11480), 1,
      ts_builtin_sym_end,
  [242207] = 1,
    ACTIONS(11482), 1,
      ts_builtin_sym_end,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(1406)] = 0,
  [SMALL_STATE(1407)] = 138,
  [SMALL_STATE(1408)] = 276,
  [SMALL_STATE(1409)] = 414,
  [SMALL_STATE(1410)] = 552,
  [SMALL_STATE(1411)] = 690,
  [SMALL_STATE(1412)] = 822,
  [SMALL_STATE(1413)] = 898,
  [SMALL_STATE(1414)] = 974,
  [SMALL_STATE(1415)] = 1112,
  [SMALL_STATE(1416)] = 1244,
  [SMALL_STATE(1417)] = 1376,
  [SMALL_STATE(1418)] = 1514,
  [SMALL_STATE(1419)] = 1652,
  [SMALL_STATE(1420)] = 1784,
  [SMALL_STATE(1421)] = 1922,
  [SMALL_STATE(1422)] = 2054,
  [SMALL_STATE(1423)] = 2192,
  [SMALL_STATE(1424)] = 2324,
  [SMALL_STATE(1425)] = 2456,
  [SMALL_STATE(1426)] = 2594,
  [SMALL_STATE(1427)] = 2726,
  [SMALL_STATE(1428)] = 2858,
  [SMALL_STATE(1429)] = 2996,
  [SMALL_STATE(1430)] = 3134,
  [SMALL_STATE(1431)] = 3266,
  [SMALL_STATE(1432)] = 3398,
  [SMALL_STATE(1433)] = 3533,
  [SMALL_STATE(1434)] = 3668,
  [SMALL_STATE(1435)] = 3803,
  [SMALL_STATE(1436)] = 3938,
  [SMALL_STATE(1437)] = 4073,
  [SMALL_STATE(1438)] = 4208,
  [SMALL_STATE(1439)] = 4343,
  [SMALL_STATE(1440)] = 4478,
  [SMALL_STATE(1441)] = 4613,
  [SMALL_STATE(1442)] = 4748,
  [SMALL_STATE(1443)] = 4883,
  [SMALL_STATE(1444)] = 5018,
  [SMALL_STATE(1445)] = 5144,
  [SMALL_STATE(1446)] = 5270,
  [SMALL_STATE(1447)] = 5396,
  [SMALL_STATE(1448)] = 5522,
  [SMALL_STATE(1449)] = 5648,
  [SMALL_STATE(1450)] = 5774,
  [SMALL_STATE(1451)] = 5900,
  [SMALL_STATE(1452)] = 6026,
  [SMALL_STATE(1453)] = 6160,
  [SMALL_STATE(1454)] = 6286,
  [SMALL_STATE(1455)] = 6412,
  [SMALL_STATE(1456)] = 6546,
  [SMALL_STATE(1457)] = 6680,
  [SMALL_STATE(1458)] = 6814,
  [SMALL_STATE(1459)] = 6948,
  [SMALL_STATE(1460)] = 7074,
  [SMALL_STATE(1461)] = 7200,
  [SMALL_STATE(1462)] = 7326,
  [SMALL_STATE(1463)] = 7452,
  [SMALL_STATE(1464)] = 7578,
  [SMALL_STATE(1465)] = 7704,
  [SMALL_STATE(1466)] = 7838,
  [SMALL_STATE(1467)] = 7972,
  [SMALL_STATE(1468)] = 8098,
  [SMALL_STATE(1469)] = 8232,
  [SMALL_STATE(1470)] = 8358,
  [SMALL_STATE(1471)] = 8492,
  [SMALL_STATE(1472)] = 8626,
  [SMALL_STATE(1473)] = 8749,
  [SMALL_STATE(1474)] = 8872,
  [SMALL_STATE(1475)] = 8995,
  [SMALL_STATE(1476)] = 9118,
  [SMALL_STATE(1477)] = 9241,
  [SMALL_STATE(1478)] = 9364,
  [SMALL_STATE(1479)] = 9487,
  [SMALL_STATE(1480)] = 9610,
  [SMALL_STATE(1481)] = 9733,
  [SMALL_STATE(1482)] = 9856,
  [SMALL_STATE(1483)] = 9979,
  [SMALL_STATE(1484)] = 10099,
  [SMALL_STATE(1485)] = 10219,
  [SMALL_STATE(1486)] = 10339,
  [SMALL_STATE(1487)] = 10459,
  [SMALL_STATE(1488)] = 10579,
  [SMALL_STATE(1489)] = 10699,
  [SMALL_STATE(1490)] = 10819,
  [SMALL_STATE(1491)] = 10939,
  [SMALL_STATE(1492)] = 11059,
  [SMALL_STATE(1493)] = 11179,
  [SMALL_STATE(1494)] = 11299,
  [SMALL_STATE(1495)] = 11419,
  [SMALL_STATE(1496)] = 11539,
  [SMALL_STATE(1497)] = 11609,
  [SMALL_STATE(1498)] = 11729,
  [SMALL_STATE(1499)] = 11849,
  [SMALL_STATE(1500)] = 11969,
  [SMALL_STATE(1501)] = 12089,
  [SMALL_STATE(1502)] = 12209,
  [SMALL_STATE(1503)] = 12329,
  [SMALL_STATE(1504)] = 12399,
  [SMALL_STATE(1505)] = 12519,
  [SMALL_STATE(1506)] = 12639,
  [SMALL_STATE(1507)] = 12759,
  [SMALL_STATE(1508)] = 12829,
  [SMALL_STATE(1509)] = 12949,
  [SMALL_STATE(1510)] = 13069,
  [SMALL_STATE(1511)] = 13189,
  [SMALL_STATE(1512)] = 13309,
  [SMALL_STATE(1513)] = 13429,
  [SMALL_STATE(1514)] = 13549,
  [SMALL_STATE(1515)] = 13669,
  [SMALL_STATE(1516)] = 13789,
  [SMALL_STATE(1517)] = 13909,
  [SMALL_STATE(1518)] = 14029,
  [SMALL_STATE(1519)] = 14149,
  [SMALL_STATE(1520)] = 14269,
  [SMALL_STATE(1521)] = 14389,
  [SMALL_STATE(1522)] = 14509,
  [SMALL_STATE(1523)] = 14629,
  [SMALL_STATE(1524)] = 14749,
  [SMALL_STATE(1525)] = 14869,
  [SMALL_STATE(1526)] = 14989,
  [SMALL_STATE(1527)] = 15109,
  [SMALL_STATE(1528)] = 15229,
  [SMALL_STATE(1529)] = 15349,
  [SMALL_STATE(1530)] = 15469,
  [SMALL_STATE(1531)] = 15538,
  [SMALL_STATE(1532)] = 15607,
  [SMALL_STATE(1533)] = 15678,
  [SMALL_STATE(1534)] = 15751,
  [SMALL_STATE(1535)] = 15878,
  [SMALL_STATE(1536)] = 16005,
  [SMALL_STATE(1537)] = 16074,
  [SMALL_STATE(1538)] = 16143,
  [SMALL_STATE(1539)] = 16212,
  [SMALL_STATE(1540)] = 16283,
  [SMALL_STATE(1541)] = 16352,
  [SMALL_STATE(1542)] = 16421,
  [SMALL_STATE(1543)] = 16494,
  [SMALL_STATE(1544)] = 16563,
  [SMALL_STATE(1545)] = 16634,
  [SMALL_STATE(1546)] = 16703,
  [SMALL_STATE(1547)] = 16830,
  [SMALL_STATE(1548)] = 16901,
  [SMALL_STATE(1549)] = 16972,
  [SMALL_STATE(1550)] = 17040,
  [SMALL_STATE(1551)] = 17108,
  [SMALL_STATE(1552)] = 17176,
  [SMALL_STATE(1553)] = 17244,
  [SMALL_STATE(1554)] = 17312,
  [SMALL_STATE(1555)] = 17380,
  [SMALL_STATE(1556)] = 17448,
  [SMALL_STATE(1557)] = 17516,
  [SMALL_STATE(1558)] = 17584,
  [SMALL_STATE(1559)] = 17652,
  [SMALL_STATE(1560)] = 17720,
  [SMALL_STATE(1561)] = 17788,
  [SMALL_STATE(1562)] = 17856,
  [SMALL_STATE(1563)] = 17924,
  [SMALL_STATE(1564)] = 17992,
  [SMALL_STATE(1565)] = 18060,
  [SMALL_STATE(1566)] = 18130,
  [SMALL_STATE(1567)] = 18198,
  [SMALL_STATE(1568)] = 18266,
  [SMALL_STATE(1569)] = 18334,
  [SMALL_STATE(1570)] = 18402,
  [SMALL_STATE(1571)] = 18470,
  [SMALL_STATE(1572)] = 18540,
  [SMALL_STATE(1573)] = 18608,
  [SMALL_STATE(1574)] = 18676,
  [SMALL_STATE(1575)] = 18744,
  [SMALL_STATE(1576)] = 18812,
  [SMALL_STATE(1577)] = 18880,
  [SMALL_STATE(1578)] = 18950,
  [SMALL_STATE(1579)] = 19018,
  [SMALL_STATE(1580)] = 19086,
  [SMALL_STATE(1581)] = 19154,
  [SMALL_STATE(1582)] = 19222,
  [SMALL_STATE(1583)] = 19290,
  [SMALL_STATE(1584)] = 19358,
  [SMALL_STATE(1585)] = 19426,
  [SMALL_STATE(1586)] = 19494,
  [SMALL_STATE(1587)] = 19562,
  [SMALL_STATE(1588)] = 19630,
  [SMALL_STATE(1589)] = 19698,
  [SMALL_STATE(1590)] = 19766,
  [SMALL_STATE(1591)] = 19834,
  [SMALL_STATE(1592)] = 19908,
  [SMALL_STATE(1593)] = 19976,
  [SMALL_STATE(1594)] = 20044,
  [SMALL_STATE(1595)] = 20112,
  [SMALL_STATE(1596)] = 20180,
  [SMALL_STATE(1597)] = 20248,
  [SMALL_STATE(1598)] = 20316,
  [SMALL_STATE(1599)] = 20384,
  [SMALL_STATE(1600)] = 20452,
  [SMALL_STATE(1601)] = 20520,
  [SMALL_STATE(1602)] = 20588,
  [SMALL_STATE(1603)] = 20656,
  [SMALL_STATE(1604)] = 20724,
  [SMALL_STATE(1605)] = 20792,
  [SMALL_STATE(1606)] = 20860,
  [SMALL_STATE(1607)] = 20928,
  [SMALL_STATE(1608)] = 20996,
  [SMALL_STATE(1609)] = 21064,
  [SMALL_STATE(1610)] = 21134,
  [SMALL_STATE(1611)] = 21202,
  [SMALL_STATE(1612)] = 21270,
  [SMALL_STATE(1613)] = 21338,
  [SMALL_STATE(1614)] = 21406,
  [SMALL_STATE(1615)] = 21474,
  [SMALL_STATE(1616)] = 21544,
  [SMALL_STATE(1617)] = 21612,
  [SMALL_STATE(1618)] = 21680,
  [SMALL_STATE(1619)] = 21748,
  [SMALL_STATE(1620)] = 21816,
  [SMALL_STATE(1621)] = 21884,
  [SMALL_STATE(1622)] = 21952,
  [SMALL_STATE(1623)] = 22020,
  [SMALL_STATE(1624)] = 22088,
  [SMALL_STATE(1625)] = 22162,
  [SMALL_STATE(1626)] = 22230,
  [SMALL_STATE(1627)] = 22298,
  [SMALL_STATE(1628)] = 22366,
  [SMALL_STATE(1629)] = 22487,
  [SMALL_STATE(1630)] = 22608,
  [SMALL_STATE(1631)] = 22729,
  [SMALL_STATE(1632)] = 22850,
  [SMALL_STATE(1633)] = 22971,
  [SMALL_STATE(1634)] = 23092,
  [SMALL_STATE(1635)] = 23213,
  [SMALL_STATE(1636)] = 23334,
  [SMALL_STATE(1637)] = 23455,
  [SMALL_STATE(1638)] = 23576,
  [SMALL_STATE(1639)] = 23697,
  [SMALL_STATE(1640)] = 23818,
  [SMALL_STATE(1641)] = 23939,
  [SMALL_STATE(1642)] = 24060,
  [SMALL_STATE(1643)] = 24181,
  [SMALL_STATE(1644)] = 24302,
  [SMALL_STATE(1645)] = 24371,
  [SMALL_STATE(1646)] = 24492,
  [SMALL_STATE(1647)] = 24613,
  [SMALL_STATE(1648)] = 24734,
  [SMALL_STATE(1649)] = 24855,
  [SMALL_STATE(1650)] = 24976,
  [SMALL_STATE(1651)] = 25097,
  [SMALL_STATE(1652)] = 25218,
  [SMALL_STATE(1653)] = 25339,
  [SMALL_STATE(1654)] = 25460,
  [SMALL_STATE(1655)] = 25581,
  [SMALL_STATE(1656)] = 25649,
  [SMALL_STATE(1657)] = 25767,
  [SMALL_STATE(1658)] = 25885,
  [SMALL_STATE(1659)] = 25953,
  [SMALL_STATE(1660)] = 26071,
  [SMALL_STATE(1661)] = 26189,
  [SMALL_STATE(1662)] = 26307,
  [SMALL_STATE(1663)] = 26425,
  [SMALL_STATE(1664)] = 26493,
  [SMALL_STATE(1665)] = 26611,
  [SMALL_STATE(1666)] = 26729,
  [SMALL_STATE(1667)] = 26797,
  [SMALL_STATE(1668)] = 26915,
  [SMALL_STATE(1669)] = 26983,
  [SMALL_STATE(1670)] = 27101,
  [SMALL_STATE(1671)] = 27219,
  [SMALL_STATE(1672)] = 27337,
  [SMALL_STATE(1673)] = 27455,
  [SMALL_STATE(1674)] = 27554,
  [SMALL_STATE(1675)] = 27655,
  [SMALL_STATE(1676)] = 27756,
  [SMALL_STATE(1677)] = 27831,
  [SMALL_STATE(1678)] = 27946,
  [SMALL_STATE(1679)] = 28023,
  [SMALL_STATE(1680)] = 28104,
  [SMALL_STATE(1681)] = 28199,
  [SMALL_STATE(1682)] = 28298,
  [SMALL_STATE(1683)] = 28363,
  [SMALL_STATE(1684)] = 28428,
  [SMALL_STATE(1685)] = 28493,
  [SMALL_STATE(1686)] = 28608,
  [SMALL_STATE(1687)] = 28707,
  [SMALL_STATE(1688)] = 28798,
  [SMALL_STATE(1689)] = 28863,
  [SMALL_STATE(1690)] = 28978,
  [SMALL_STATE(1691)] = 29045,
  [SMALL_STATE(1692)] = 29110,
  [SMALL_STATE(1693)] = 29175,
  [SMALL_STATE(1694)] = 29290,
  [SMALL_STATE(1695)] = 29387,
  [SMALL_STATE(1696)] = 29488,
  [SMALL_STATE(1697)] = 29573,
  [SMALL_STATE(1698)] = 29660,
  [SMALL_STATE(1699)] = 29747,
  [SMALL_STATE(1700)] = 29846,
  [SMALL_STATE(1701)] = 29913,
  [SMALL_STATE(1702)] = 29980,
  [SMALL_STATE(1703)] = 30057,
  [SMALL_STATE(1704)] = 30169,
  [SMALL_STATE(1705)] = 30233,
  [SMALL_STATE(1706)] = 30345,
  [SMALL_STATE(1707)] = 30409,
  [SMALL_STATE(1708)] = 30473,
  [SMALL_STATE(1709)] = 30537,
  [SMALL_STATE(1710)] = 30601,
  [SMALL_STATE(1711)] = 30665,
  [SMALL_STATE(1712)] = 30729,
  [SMALL_STATE(1713)] = 30793,
  [SMALL_STATE(1714)] = 30857,
  [SMALL_STATE(1715)] = 30957,
  [SMALL_STATE(1716)] = 31021,
  [SMALL_STATE(1717)] = 31133,
  [SMALL_STATE(1718)] = 31197,
  [SMALL_STATE(1719)] = 31261,
  [SMALL_STATE(1720)] = 31373,
  [SMALL_STATE(1721)] = 31473,
  [SMALL_STATE(1722)] = 31537,
  [SMALL_STATE(1723)] = 31649,
  [SMALL_STATE(1724)] = 31713,
  [SMALL_STATE(1725)] = 31827,
  [SMALL_STATE(1726)] = 31891,
  [SMALL_STATE(1727)] = 31955,
  [SMALL_STATE(1728)] = 32067,
  [SMALL_STATE(1729)] = 32141,
  [SMALL_STATE(1730)] = 32217,
  [SMALL_STATE(1731)] = 32293,
  [SMALL_STATE(1732)] = 32373,
  [SMALL_STATE(1733)] = 32485,
  [SMALL_STATE(1734)] = 32549,
  [SMALL_STATE(1735)] = 32613,
  [SMALL_STATE(1736)] = 32677,
  [SMALL_STATE(1737)] = 32741,
  [SMALL_STATE(1738)] = 32805,
  [SMALL_STATE(1739)] = 32917,
  [SMALL_STATE(1740)] = 33033,
  [SMALL_STATE(1741)] = 33097,
  [SMALL_STATE(1742)] = 33161,
  [SMALL_STATE(1743)] = 33225,
  [SMALL_STATE(1744)] = 33289,
  [SMALL_STATE(1745)] = 33353,
  [SMALL_STATE(1746)] = 33465,
  [SMALL_STATE(1747)] = 33531,
  [SMALL_STATE(1748)] = 33643,
  [SMALL_STATE(1749)] = 33743,
  [SMALL_STATE(1750)] = 33807,
  [SMALL_STATE(1751)] = 33903,
  [SMALL_STATE(1752)] = 34017,
  [SMALL_STATE(1753)] = 34129,
  [SMALL_STATE(1754)] = 34245,
  [SMALL_STATE(1755)] = 34357,
  [SMALL_STATE(1756)] = 34421,
  [SMALL_STATE(1757)] = 34485,
  [SMALL_STATE(1758)] = 34549,
  [SMALL_STATE(1759)] = 34613,
  [SMALL_STATE(1760)] = 34677,
  [SMALL_STATE(1761)] = 34743,
  [SMALL_STATE(1762)] = 34807,
  [SMALL_STATE(1763)] = 34891,
  [SMALL_STATE(1764)] = 34977,
  [SMALL_STATE(1765)] = 35063,
  [SMALL_STATE(1766)] = 35161,
  [SMALL_STATE(1767)] = 35251,
  [SMALL_STATE(1768)] = 35315,
  [SMALL_STATE(1769)] = 35427,
  [SMALL_STATE(1770)] = 35521,
  [SMALL_STATE(1771)] = 35585,
  [SMALL_STATE(1772)] = 35649,
  [SMALL_STATE(1773)] = 35761,
  [SMALL_STATE(1774)] = 35859,
  [SMALL_STATE(1775)] = 35973,
  [SMALL_STATE(1776)] = 36085,
  [SMALL_STATE(1777)] = 36201,
  [SMALL_STATE(1778)] = 36299,
  [SMALL_STATE(1779)] = 36363,
  [SMALL_STATE(1780)] = 36475,
  [SMALL_STATE(1781)] = 36591,
  [SMALL_STATE(1782)] = 36655,
  [SMALL_STATE(1783)] = 36771,
  [SMALL_STATE(1784)] = 36835,
  [SMALL_STATE(1785)] = 36899,
  [SMALL_STATE(1786)] = 36963,
  [SMALL_STATE(1787)] = 37027,
  [SMALL_STATE(1788)] = 37091,
  [SMALL_STATE(1789)] = 37155,
  [SMALL_STATE(1790)] = 37219,
  [SMALL_STATE(1791)] = 37317,
  [SMALL_STATE(1792)] = 37381,
  [SMALL_STATE(1793)] = 37493,
  [SMALL_STATE(1794)] = 37557,
  [SMALL_STATE(1795)] = 37621,
  [SMALL_STATE(1796)] = 37685,
  [SMALL_STATE(1797)] = 37749,
  [SMALL_STATE(1798)] = 37813,
  [SMALL_STATE(1799)] = 37877,
  [SMALL_STATE(1800)] = 37941,
  [SMALL_STATE(1801)] = 38005,
  [SMALL_STATE(1802)] = 38069,
  [SMALL_STATE(1803)] = 38133,
  [SMALL_STATE(1804)] = 38230,
  [SMALL_STATE(1805)] = 38293,
  [SMALL_STATE(1806)] = 38356,
  [SMALL_STATE(1807)] = 38419,
  [SMALL_STATE(1808)] = 38482,
  [SMALL_STATE(1809)] = 38581,
  [SMALL_STATE(1810)] = 38644,
  [SMALL_STATE(1811)] = 38707,
  [SMALL_STATE(1812)] = 38770,
  [SMALL_STATE(1813)] = 38833,
  [SMALL_STATE(1814)] = 38896,
  [SMALL_STATE(1815)] = 38965,
  [SMALL_STATE(1816)] = 39034,
  [SMALL_STATE(1817)] = 39113,
  [SMALL_STATE(1818)] = 39176,
  [SMALL_STATE(1819)] = 39241,
  [SMALL_STATE(1820)] = 39304,
  [SMALL_STATE(1821)] = 39367,
  [SMALL_STATE(1822)] = 39466,
  [SMALL_STATE(1823)] = 39531,
  [SMALL_STATE(1824)] = 39594,
  [SMALL_STATE(1825)] = 39657,
  [SMALL_STATE(1826)] = 39720,
  [SMALL_STATE(1827)] = 39795,
  [SMALL_STATE(1828)] = 39858,
  [SMALL_STATE(1829)] = 39921,
  [SMALL_STATE(1830)] = 39984,
  [SMALL_STATE(1831)] = 40047,
  [SMALL_STATE(1832)] = 40110,
  [SMALL_STATE(1833)] = 40173,
  [SMALL_STATE(1834)] = 40236,
  [SMALL_STATE(1835)] = 40299,
  [SMALL_STATE(1836)] = 40362,
  [SMALL_STATE(1837)] = 40457,
  [SMALL_STATE(1838)] = 40520,
  [SMALL_STATE(1839)] = 40603,
  [SMALL_STATE(1840)] = 40688,
  [SMALL_STATE(1841)] = 40773,
  [SMALL_STATE(1842)] = 40870,
  [SMALL_STATE(1843)] = 40959,
  [SMALL_STATE(1844)] = 41052,
  [SMALL_STATE(1845)] = 41115,
  [SMALL_STATE(1846)] = 41184,
  [SMALL_STATE(1847)] = 41247,
  [SMALL_STATE(1848)] = 41310,
  [SMALL_STATE(1849)] = 41373,
  [SMALL_STATE(1850)] = 41436,
  [SMALL_STATE(1851)] = 41535,
  [SMALL_STATE(1852)] = 41598,
  [SMALL_STATE(1853)] = 41661,
  [SMALL_STATE(1854)] = 41724,
  [SMALL_STATE(1855)] = 41787,
  [SMALL_STATE(1856)] = 41850,
  [SMALL_STATE(1857)] = 41913,
  [SMALL_STATE(1858)] = 41976,
  [SMALL_STATE(1859)] = 42039,
  [SMALL_STATE(1860)] = 42102,
  [SMALL_STATE(1861)] = 42165,
  [SMALL_STATE(1862)] = 42228,
  [SMALL_STATE(1863)] = 42291,
  [SMALL_STATE(1864)] = 42354,
  [SMALL_STATE(1865)] = 42417,
  [SMALL_STATE(1866)] = 42514,
  [SMALL_STATE(1867)] = 42577,
  [SMALL_STATE(1868)] = 42640,
  [SMALL_STATE(1869)] = 42703,
  [SMALL_STATE(1870)] = 42766,
  [SMALL_STATE(1871)] = 42829,
  [SMALL_STATE(1872)] = 42892,
  [SMALL_STATE(1873)] = 42955,
  [SMALL_STATE(1874)] = 43018,
  [SMALL_STATE(1875)] = 43081,
  [SMALL_STATE(1876)] = 43150,
  [SMALL_STATE(1877)] = 43215,
  [SMALL_STATE(1878)] = 43278,
  [SMALL_STATE(1879)] = 43343,
  [SMALL_STATE(1880)] = 43412,
  [SMALL_STATE(1881)] = 43475,
  [SMALL_STATE(1882)] = 43538,
  [SMALL_STATE(1883)] = 43611,
  [SMALL_STATE(1884)] = 43674,
  [SMALL_STATE(1885)] = 43737,
  [SMALL_STATE(1886)] = 43800,
  [SMALL_STATE(1887)] = 43863,
  [SMALL_STATE(1888)] = 43926,
  [SMALL_STATE(1889)] = 43995,
  [SMALL_STATE(1890)] = 44058,
  [SMALL_STATE(1891)] = 44121,
  [SMALL_STATE(1892)] = 44184,
  [SMALL_STATE(1893)] = 44247,
  [SMALL_STATE(1894)] = 44310,
  [SMALL_STATE(1895)] = 44373,
  [SMALL_STATE(1896)] = 44436,
  [SMALL_STATE(1897)] = 44499,
  [SMALL_STATE(1898)] = 44562,
  [SMALL_STATE(1899)] = 44625,
  [SMALL_STATE(1900)] = 44688,
  [SMALL_STATE(1901)] = 44785,
  [SMALL_STATE(1902)] = 44848,
  [SMALL_STATE(1903)] = 44947,
  [SMALL_STATE(1904)] = 45010,
  [SMALL_STATE(1905)] = 45073,
  [SMALL_STATE(1906)] = 45136,
  [SMALL_STATE(1907)] = 45199,
  [SMALL_STATE(1908)] = 45264,
  [SMALL_STATE(1909)] = 45329,
  [SMALL_STATE(1910)] = 45392,
  [SMALL_STATE(1911)] = 45455,
  [SMALL_STATE(1912)] = 45524,
  [SMALL_STATE(1913)] = 45587,
  [SMALL_STATE(1914)] = 45650,
  [SMALL_STATE(1915)] = 45713,
  [SMALL_STATE(1916)] = 45776,
  [SMALL_STATE(1917)] = 45839,
  [SMALL_STATE(1918)] = 45902,
  [SMALL_STATE(1919)] = 45965,
  [SMALL_STATE(1920)] = 46028,
  [SMALL_STATE(1921)] = 46127,
  [SMALL_STATE(1922)] = 46190,
  [SMALL_STATE(1923)] = 46259,
  [SMALL_STATE(1924)] = 46322,
  [SMALL_STATE(1925)] = 46421,
  [SMALL_STATE(1926)] = 46484,
  [SMALL_STATE(1927)] = 46547,
  [SMALL_STATE(1928)] = 46610,
  [SMALL_STATE(1929)] = 46679,
  [SMALL_STATE(1930)] = 46742,
  [SMALL_STATE(1931)] = 46805,
  [SMALL_STATE(1932)] = 46868,
  [SMALL_STATE(1933)] = 46931,
  [SMALL_STATE(1934)] = 46994,
  [SMALL_STATE(1935)] = 47057,
  [SMALL_STATE(1936)] = 47120,
  [SMALL_STATE(1937)] = 47183,
  [SMALL_STATE(1938)] = 47252,
  [SMALL_STATE(1939)] = 47315,
  [SMALL_STATE(1940)] = 47378,
  [SMALL_STATE(1941)] = 47441,
  [SMALL_STATE(1942)] = 47516,
  [SMALL_STATE(1943)] = 47579,
  [SMALL_STATE(1944)] = 47642,
  [SMALL_STATE(1945)] = 47705,
  [SMALL_STATE(1946)] = 47768,
  [SMALL_STATE(1947)] = 47831,
  [SMALL_STATE(1948)] = 47894,
  [SMALL_STATE(1949)] = 47956,
  [SMALL_STATE(1950)] = 48018,
  [SMALL_STATE(1951)] = 48080,
  [SMALL_STATE(1952)] = 48142,
  [SMALL_STATE(1953)] = 48204,
  [SMALL_STATE(1954)] = 48266,
  [SMALL_STATE(1955)] = 48328,
  [SMALL_STATE(1956)] = 48390,
  [SMALL_STATE(1957)] = 48452,
  [SMALL_STATE(1958)] = 48514,
  [SMALL_STATE(1959)] = 48576,
  [SMALL_STATE(1960)] = 48640,
  [SMALL_STATE(1961)] = 48702,
  [SMALL_STATE(1962)] = 48764,
  [SMALL_STATE(1963)] = 48826,
  [SMALL_STATE(1964)] = 48888,
  [SMALL_STATE(1965)] = 48950,
  [SMALL_STATE(1966)] = 49012,
  [SMALL_STATE(1967)] = 49074,
  [SMALL_STATE(1968)] = 49136,
  [SMALL_STATE(1969)] = 49198,
  [SMALL_STATE(1970)] = 49260,
  [SMALL_STATE(1971)] = 49322,
  [SMALL_STATE(1972)] = 49384,
  [SMALL_STATE(1973)] = 49446,
  [SMALL_STATE(1974)] = 49508,
  [SMALL_STATE(1975)] = 49570,
  [SMALL_STATE(1976)] = 49632,
  [SMALL_STATE(1977)] = 49694,
  [SMALL_STATE(1978)] = 49756,
  [SMALL_STATE(1979)] = 49818,
  [SMALL_STATE(1980)] = 49880,
  [SMALL_STATE(1981)] = 49942,
  [SMALL_STATE(1982)] = 50004,
  [SMALL_STATE(1983)] = 50066,
  [SMALL_STATE(1984)] = 50128,
  [SMALL_STATE(1985)] = 50190,
  [SMALL_STATE(1986)] = 50252,
  [SMALL_STATE(1987)] = 50314,
  [SMALL_STATE(1988)] = 50376,
  [SMALL_STATE(1989)] = 50438,
  [SMALL_STATE(1990)] = 50500,
  [SMALL_STATE(1991)] = 50562,
  [SMALL_STATE(1992)] = 50624,
  [SMALL_STATE(1993)] = 50686,
  [SMALL_STATE(1994)] = 50748,
  [SMALL_STATE(1995)] = 50810,
  [SMALL_STATE(1996)] = 50872,
  [SMALL_STATE(1997)] = 50934,
  [SMALL_STATE(1998)] = 50996,
  [SMALL_STATE(1999)] = 51058,
  [SMALL_STATE(2000)] = 51120,
  [SMALL_STATE(2001)] = 51182,
  [SMALL_STATE(2002)] = 51244,
  [SMALL_STATE(2003)] = 51306,
  [SMALL_STATE(2004)] = 51378,
  [SMALL_STATE(2005)] = 51440,
  [SMALL_STATE(2006)] = 51502,
  [SMALL_STATE(2007)] = 51564,
  [SMALL_STATE(2008)] = 51626,
  [SMALL_STATE(2009)] = 51688,
  [SMALL_STATE(2010)] = 51750,
  [SMALL_STATE(2011)] = 51812,
  [SMALL_STATE(2012)] = 51874,
  [SMALL_STATE(2013)] = 51936,
  [SMALL_STATE(2014)] = 51998,
  [SMALL_STATE(2015)] = 52060,
  [SMALL_STATE(2016)] = 52122,
  [SMALL_STATE(2017)] = 52184,
  [SMALL_STATE(2018)] = 52246,
  [SMALL_STATE(2019)] = 52308,
  [SMALL_STATE(2020)] = 52370,
  [SMALL_STATE(2021)] = 52432,
  [SMALL_STATE(2022)] = 52494,
  [SMALL_STATE(2023)] = 52556,
  [SMALL_STATE(2024)] = 52618,
  [SMALL_STATE(2025)] = 52680,
  [SMALL_STATE(2026)] = 52742,
  [SMALL_STATE(2027)] = 52804,
  [SMALL_STATE(2028)] = 52866,
  [SMALL_STATE(2029)] = 52928,
  [SMALL_STATE(2030)] = 52992,
  [SMALL_STATE(2031)] = 53054,
  [SMALL_STATE(2032)] = 53120,
  [SMALL_STATE(2033)] = 53184,
  [SMALL_STATE(2034)] = 53246,
  [SMALL_STATE(2035)] = 53310,
  [SMALL_STATE(2036)] = 53372,
  [SMALL_STATE(2037)] = 53440,
  [SMALL_STATE(2038)] = 53502,
  [SMALL_STATE(2039)] = 53564,
  [SMALL_STATE(2040)] = 53626,
  [SMALL_STATE(2041)] = 53688,
  [SMALL_STATE(2042)] = 53750,
  [SMALL_STATE(2043)] = 53812,
  [SMALL_STATE(2044)] = 53874,
  [SMALL_STATE(2045)] = 53936,
  [SMALL_STATE(2046)] = 54034,
  [SMALL_STATE(2047)] = 54098,
  [SMALL_STATE(2048)] = 54160,
  [SMALL_STATE(2049)] = 54222,
  [SMALL_STATE(2050)] = 54284,
  [SMALL_STATE(2051)] = 54346,
  [SMALL_STATE(2052)] = 54408,
  [SMALL_STATE(2053)] = 54476,
  [SMALL_STATE(2054)] = 54544,
  [SMALL_STATE(2055)] = 54606,
  [SMALL_STATE(2056)] = 54702,
  [SMALL_STATE(2057)] = 54764,
  [SMALL_STATE(2058)] = 54826,
  [SMALL_STATE(2059)] = 54888,
  [SMALL_STATE(2060)] = 54950,
  [SMALL_STATE(2061)] = 55012,
  [SMALL_STATE(2062)] = 55080,
  [SMALL_STATE(2063)] = 55148,
  [SMALL_STATE(2064)] = 55210,
  [SMALL_STATE(2065)] = 55308,
  [SMALL_STATE(2066)] = 55406,
  [SMALL_STATE(2067)] = 55476,
  [SMALL_STATE(2068)] = 55574,
  [SMALL_STATE(2069)] = 55642,
  [SMALL_STATE(2070)] = 55704,
  [SMALL_STATE(2071)] = 55802,
  [SMALL_STATE(2072)] = 55872,
  [SMALL_STATE(2073)] = 55934,
  [SMALL_STATE(2074)] = 56000,
  [SMALL_STATE(2075)] = 56064,
  [SMALL_STATE(2076)] = 56126,
  [SMALL_STATE(2077)] = 56194,
  [SMALL_STATE(2078)] = 56256,
  [SMALL_STATE(2079)] = 56318,
  [SMALL_STATE(2080)] = 56382,
  [SMALL_STATE(2081)] = 56444,
  [SMALL_STATE(2082)] = 56506,
  [SMALL_STATE(2083)] = 56568,
  [SMALL_STATE(2084)] = 56630,
  [SMALL_STATE(2085)] = 56692,
  [SMALL_STATE(2086)] = 56790,
  [SMALL_STATE(2087)] = 56852,
  [SMALL_STATE(2088)] = 56948,
  [SMALL_STATE(2089)] = 57010,
  [SMALL_STATE(2090)] = 57072,
  [SMALL_STATE(2091)] = 57134,
  [SMALL_STATE(2092)] = 57196,
  [SMALL_STATE(2093)] = 57258,
  [SMALL_STATE(2094)] = 57320,
  [SMALL_STATE(2095)] = 57382,
  [SMALL_STATE(2096)] = 57444,
  [SMALL_STATE(2097)] = 57506,
  [SMALL_STATE(2098)] = 57568,
  [SMALL_STATE(2099)] = 57630,
  [SMALL_STATE(2100)] = 57692,
  [SMALL_STATE(2101)] = 57754,
  [SMALL_STATE(2102)] = 57816,
  [SMALL_STATE(2103)] = 57878,
  [SMALL_STATE(2104)] = 57940,
  [SMALL_STATE(2105)] = 58002,
  [SMALL_STATE(2106)] = 58064,
  [SMALL_STATE(2107)] = 58126,
  [SMALL_STATE(2108)] = 58188,
  [SMALL_STATE(2109)] = 58282,
  [SMALL_STATE(2110)] = 58364,
  [SMALL_STATE(2111)] = 58448,
  [SMALL_STATE(2112)] = 58532,
  [SMALL_STATE(2113)] = 58594,
  [SMALL_STATE(2114)] = 58690,
  [SMALL_STATE(2115)] = 58764,
  [SMALL_STATE(2116)] = 58838,
  [SMALL_STATE(2117)] = 58916,
  [SMALL_STATE(2118)] = 59004,
  [SMALL_STATE(2119)] = 59096,
  [SMALL_STATE(2120)] = 59192,
  [SMALL_STATE(2121)] = 59254,
  [SMALL_STATE(2122)] = 59316,
  [SMALL_STATE(2123)] = 59378,
  [SMALL_STATE(2124)] = 59440,
  [SMALL_STATE(2125)] = 59505,
  [SMALL_STATE(2126)] = 59622,
  [SMALL_STATE(2127)] = 59685,
  [SMALL_STATE(2128)] = 59748,
  [SMALL_STATE(2129)] = 59813,
  [SMALL_STATE(2130)] = 59878,
  [SMALL_STATE(2131)] = 59943,
  [SMALL_STATE(2132)] = 60006,
  [SMALL_STATE(2133)] = 60123,
  [SMALL_STATE(2134)] = 60188,
  [SMALL_STATE(2135)] = 60253,
  [SMALL_STATE(2136)] = 60370,
  [SMALL_STATE(2137)] = 60487,
  [SMALL_STATE(2138)] = 60604,
  [SMALL_STATE(2139)] = 60721,
  [SMALL_STATE(2140)] = 60784,
  [SMALL_STATE(2141)] = 60844,
  [SMALL_STATE(2142)] = 60904,
  [SMALL_STATE(2143)] = 60964,
  [SMALL_STATE(2144)] = 61024,
  [SMALL_STATE(2145)] = 61084,
  [SMALL_STATE(2146)] = 61144,
  [SMALL_STATE(2147)] = 61204,
  [SMALL_STATE(2148)] = 61264,
  [SMALL_STATE(2149)] = 61324,
  [SMALL_STATE(2150)] = 61384,
  [SMALL_STATE(2151)] = 61444,
  [SMALL_STATE(2152)] = 61504,
  [SMALL_STATE(2153)] = 61564,
  [SMALL_STATE(2154)] = 61624,
  [SMALL_STATE(2155)] = 61684,
  [SMALL_STATE(2156)] = 61744,
  [SMALL_STATE(2157)] = 61804,
  [SMALL_STATE(2158)] = 61864,
  [SMALL_STATE(2159)] = 61924,
  [SMALL_STATE(2160)] = 61984,
  [SMALL_STATE(2161)] = 62044,
  [SMALL_STATE(2162)] = 62104,
  [SMALL_STATE(2163)] = 62166,
  [SMALL_STATE(2164)] = 62262,
  [SMALL_STATE(2165)] = 62322,
  [SMALL_STATE(2166)] = 62382,
  [SMALL_STATE(2167)] = 62442,
  [SMALL_STATE(2168)] = 62502,
  [SMALL_STATE(2169)] = 62562,
  [SMALL_STATE(2170)] = 62622,
  [SMALL_STATE(2171)] = 62682,
  [SMALL_STATE(2172)] = 62742,
  [SMALL_STATE(2173)] = 62802,
  [SMALL_STATE(2174)] = 62862,
  [SMALL_STATE(2175)] = 62922,
  [SMALL_STATE(2176)] = 62982,
  [SMALL_STATE(2177)] = 63042,
  [SMALL_STATE(2178)] = 63102,
  [SMALL_STATE(2179)] = 63162,
  [SMALL_STATE(2180)] = 63222,
  [SMALL_STATE(2181)] = 63282,
  [SMALL_STATE(2182)] = 63342,
  [SMALL_STATE(2183)] = 63402,
  [SMALL_STATE(2184)] = 63462,
  [SMALL_STATE(2185)] = 63522,
  [SMALL_STATE(2186)] = 63582,
  [SMALL_STATE(2187)] = 63642,
  [SMALL_STATE(2188)] = 63702,
  [SMALL_STATE(2189)] = 63762,
  [SMALL_STATE(2190)] = 63822,
  [SMALL_STATE(2191)] = 63882,
  [SMALL_STATE(2192)] = 63942,
  [SMALL_STATE(2193)] = 64002,
  [SMALL_STATE(2194)] = 64062,
  [SMALL_STATE(2195)] = 64122,
  [SMALL_STATE(2196)] = 64182,
  [SMALL_STATE(2197)] = 64242,
  [SMALL_STATE(2198)] = 64302,
  [SMALL_STATE(2199)] = 64362,
  [SMALL_STATE(2200)] = 64422,
  [SMALL_STATE(2201)] = 64482,
  [SMALL_STATE(2202)] = 64542,
  [SMALL_STATE(2203)] = 64602,
  [SMALL_STATE(2204)] = 64662,
  [SMALL_STATE(2205)] = 64722,
  [SMALL_STATE(2206)] = 64786,
  [SMALL_STATE(2207)] = 64846,
  [SMALL_STATE(2208)] = 64906,
  [SMALL_STATE(2209)] = 64966,
  [SMALL_STATE(2210)] = 65026,
  [SMALL_STATE(2211)] = 65086,
  [SMALL_STATE(2212)] = 65146,
  [SMALL_STATE(2213)] = 65206,
  [SMALL_STATE(2214)] = 65266,
  [SMALL_STATE(2215)] = 65326,
  [SMALL_STATE(2216)] = 65390,
  [SMALL_STATE(2217)] = 65454,
  [SMALL_STATE(2218)] = 65514,
  [SMALL_STATE(2219)] = 65574,
  [SMALL_STATE(2220)] = 65638,
  [SMALL_STATE(2221)] = 65700,
  [SMALL_STATE(2222)] = 65760,
  [SMALL_STATE(2223)] = 65820,
  [SMALL_STATE(2224)] = 65880,
  [SMALL_STATE(2225)] = 65944,
  [SMALL_STATE(2226)] = 66008,
  [SMALL_STATE(2227)] = 66068,
  [SMALL_STATE(2228)] = 66128,
  [SMALL_STATE(2229)] = 66188,
  [SMALL_STATE(2230)] = 66248,
  [SMALL_STATE(2231)] = 66308,
  [SMALL_STATE(2232)] = 66368,
  [SMALL_STATE(2233)] = 66428,
  [SMALL_STATE(2234)] = 66488,
  [SMALL_STATE(2235)] = 66548,
  [SMALL_STATE(2236)] = 66608,
  [SMALL_STATE(2237)] = 66668,
  [SMALL_STATE(2238)] = 66728,
  [SMALL_STATE(2239)] = 66788,
  [SMALL_STATE(2240)] = 66884,
  [SMALL_STATE(2241)] = 66944,
  [SMALL_STATE(2242)] = 67004,
  [SMALL_STATE(2243)] = 67064,
  [SMALL_STATE(2244)] = 67124,
  [SMALL_STATE(2245)] = 67184,
  [SMALL_STATE(2246)] = 67244,
  [SMALL_STATE(2247)] = 67304,
  [SMALL_STATE(2248)] = 67364,
  [SMALL_STATE(2249)] = 67424,
  [SMALL_STATE(2250)] = 67486,
  [SMALL_STATE(2251)] = 67546,
  [SMALL_STATE(2252)] = 67606,
  [SMALL_STATE(2253)] = 67666,
  [SMALL_STATE(2254)] = 67726,
  [SMALL_STATE(2255)] = 67822,
  [SMALL_STATE(2256)] = 67882,
  [SMALL_STATE(2257)] = 67942,
  [SMALL_STATE(2258)] = 68002,
  [SMALL_STATE(2259)] = 68062,
  [SMALL_STATE(2260)] = 68122,
  [SMALL_STATE(2261)] = 68182,
  [SMALL_STATE(2262)] = 68242,
  [SMALL_STATE(2263)] = 68302,
  [SMALL_STATE(2264)] = 68362,
  [SMALL_STATE(2265)] = 68425,
  [SMALL_STATE(2266)] = 68484,
  [SMALL_STATE(2267)] = 68543,
  [SMALL_STATE(2268)] = 68602,
  [SMALL_STATE(2269)] = 68661,
  [SMALL_STATE(2270)] = 68774,
  [SMALL_STATE(2271)] = 68833,
  [SMALL_STATE(2272)] = 68892,
  [SMALL_STATE(2273)] = 68951,
  [SMALL_STATE(2274)] = 69010,
  [SMALL_STATE(2275)] = 69069,
  [SMALL_STATE(2276)] = 69170,
  [SMALL_STATE(2277)] = 69229,
  [SMALL_STATE(2278)] = 69332,
  [SMALL_STATE(2279)] = 69391,
  [SMALL_STATE(2280)] = 69504,
  [SMALL_STATE(2281)] = 69607,
  [SMALL_STATE(2282)] = 69666,
  [SMALL_STATE(2283)] = 69729,
  [SMALL_STATE(2284)] = 69788,
  [SMALL_STATE(2285)] = 69891,
  [SMALL_STATE(2286)] = 69950,
  [SMALL_STATE(2287)] = 70063,
  [SMALL_STATE(2288)] = 70166,
  [SMALL_STATE(2289)] = 70225,
  [SMALL_STATE(2290)] = 70338,
  [SMALL_STATE(2291)] = 70401,
  [SMALL_STATE(2292)] = 70502,
  [SMALL_STATE(2293)] = 70561,
  [SMALL_STATE(2294)] = 70662,
  [SMALL_STATE(2295)] = 70765,
  [SMALL_STATE(2296)] = 70868,
  [SMALL_STATE(2297)] = 70927,
  [SMALL_STATE(2298)] = 70986,
  [SMALL_STATE(2299)] = 71099,
  [SMALL_STATE(2300)] = 71160,
  [SMALL_STATE(2301)] = 71273,
  [SMALL_STATE(2302)] = 71332,
  [SMALL_STATE(2303)] = 71435,
  [SMALL_STATE(2304)] = 71536,
  [SMALL_STATE(2305)] = 71599,
  [SMALL_STATE(2306)] = 71658,
  [SMALL_STATE(2307)] = 71759,
  [SMALL_STATE(2308)] = 71818,
  [SMALL_STATE(2309)] = 71877,
  [SMALL_STATE(2310)] = 71936,
  [SMALL_STATE(2311)] = 72039,
  [SMALL_STATE(2312)] = 72098,
  [SMALL_STATE(2313)] = 72199,
  [SMALL_STATE(2314)] = 72258,
  [SMALL_STATE(2315)] = 72317,
  [SMALL_STATE(2316)] = 72376,
  [SMALL_STATE(2317)] = 72439,
  [SMALL_STATE(2318)] = 72498,
  [SMALL_STATE(2319)] = 72599,
  [SMALL_STATE(2320)] = 72660,
  [SMALL_STATE(2321)] = 72761,
  [SMALL_STATE(2322)] = 72820,
  [SMALL_STATE(2323)] = 72879,
  [SMALL_STATE(2324)] = 72938,
  [SMALL_STATE(2325)] = 73039,
  [SMALL_STATE(2326)] = 73142,
  [SMALL_STATE(2327)] = 73205,
  [SMALL_STATE(2328)] = 73308,
  [SMALL_STATE(2329)] = 73367,
  [SMALL_STATE(2330)] = 73470,
  [SMALL_STATE(2331)] = 73532,
  [SMALL_STATE(2332)] = 73594,
  [SMALL_STATE(2333)] = 73704,
  [SMALL_STATE(2334)] = 73814,
  [SMALL_STATE(2335)] = 73876,
  [SMALL_STATE(2336)] = 73986,
  [SMALL_STATE(2337)] = 74048,
  [SMALL_STATE(2338)] = 74158,
  [SMALL_STATE(2339)] = 74268,
  [SMALL_STATE(2340)] = 74330,
  [SMALL_STATE(2341)] = 74440,
  [SMALL_STATE(2342)] = 74500,
  [SMALL_STATE(2343)] = 74560,
  [SMALL_STATE(2344)] = 74622,
  [SMALL_STATE(2345)] = 74679,
  [SMALL_STATE(2346)] = 74742,
  [SMALL_STATE(2347)] = 74799,
  [SMALL_STATE(2348)] = 74856,
  [SMALL_STATE(2349)] = 74913,
  [SMALL_STATE(2350)] = 74970,
  [SMALL_STATE(2351)] = 75027,
  [SMALL_STATE(2352)] = 75084,
  [SMALL_STATE(2353)] = 75141,
  [SMALL_STATE(2354)] = 75198,
  [SMALL_STATE(2355)] = 75255,
  [SMALL_STATE(2356)] = 75312,
  [SMALL_STATE(2357)] = 75369,
  [SMALL_STATE(2358)] = 75426,
  [SMALL_STATE(2359)] = 75483,
  [SMALL_STATE(2360)] = 75540,
  [SMALL_STATE(2361)] = 75597,
  [SMALL_STATE(2362)] = 75654,
  [SMALL_STATE(2363)] = 75711,
  [SMALL_STATE(2364)] = 75768,
  [SMALL_STATE(2365)] = 75825,
  [SMALL_STATE(2366)] = 75882,
  [SMALL_STATE(2367)] = 75939,
  [SMALL_STATE(2368)] = 75996,
  [SMALL_STATE(2369)] = 76053,
  [SMALL_STATE(2370)] = 76110,
  [SMALL_STATE(2371)] = 76167,
  [SMALL_STATE(2372)] = 76224,
  [SMALL_STATE(2373)] = 76283,
  [SMALL_STATE(2374)] = 76340,
  [SMALL_STATE(2375)] = 76397,
  [SMALL_STATE(2376)] = 76454,
  [SMALL_STATE(2377)] = 76511,
  [SMALL_STATE(2378)] = 76568,
  [SMALL_STATE(2379)] = 76625,
  [SMALL_STATE(2380)] = 76682,
  [SMALL_STATE(2381)] = 76739,
  [SMALL_STATE(2382)] = 76796,
  [SMALL_STATE(2383)] = 76853,
  [SMALL_STATE(2384)] = 76910,
  [SMALL_STATE(2385)] = 76967,
  [SMALL_STATE(2386)] = 77024,
  [SMALL_STATE(2387)] = 77081,
  [SMALL_STATE(2388)] = 77138,
  [SMALL_STATE(2389)] = 77195,
  [SMALL_STATE(2390)] = 77252,
  [SMALL_STATE(2391)] = 77309,
  [SMALL_STATE(2392)] = 77366,
  [SMALL_STATE(2393)] = 77423,
  [SMALL_STATE(2394)] = 77480,
  [SMALL_STATE(2395)] = 77537,
  [SMALL_STATE(2396)] = 77594,
  [SMALL_STATE(2397)] = 77651,
  [SMALL_STATE(2398)] = 77708,
  [SMALL_STATE(2399)] = 77765,
  [SMALL_STATE(2400)] = 77822,
  [SMALL_STATE(2401)] = 77879,
  [SMALL_STATE(2402)] = 77936,
  [SMALL_STATE(2403)] = 77993,
  [SMALL_STATE(2404)] = 78050,
  [SMALL_STATE(2405)] = 78113,
  [SMALL_STATE(2406)] = 78170,
  [SMALL_STATE(2407)] = 78227,
  [SMALL_STATE(2408)] = 78284,
  [SMALL_STATE(2409)] = 78341,
  [SMALL_STATE(2410)] = 78398,
  [SMALL_STATE(2411)] = 78457,
  [SMALL_STATE(2412)] = 78516,
  [SMALL_STATE(2413)] = 78573,
  [SMALL_STATE(2414)] = 78632,
  [SMALL_STATE(2415)] = 78689,
  [SMALL_STATE(2416)] = 78746,
  [SMALL_STATE(2417)] = 78803,
  [SMALL_STATE(2418)] = 78860,
  [SMALL_STATE(2419)] = 78917,
  [SMALL_STATE(2420)] = 78974,
  [SMALL_STATE(2421)] = 79031,
  [SMALL_STATE(2422)] = 79088,
  [SMALL_STATE(2423)] = 79145,
  [SMALL_STATE(2424)] = 79202,
  [SMALL_STATE(2425)] = 79258,
  [SMALL_STATE(2426)] = 79314,
  [SMALL_STATE(2427)] = 79370,
  [SMALL_STATE(2428)] = 79426,
  [SMALL_STATE(2429)] = 79482,
  [SMALL_STATE(2430)] = 79538,
  [SMALL_STATE(2431)] = 79594,
  [SMALL_STATE(2432)] = 79650,
  [SMALL_STATE(2433)] = 79706,
  [SMALL_STATE(2434)] = 79762,
  [SMALL_STATE(2435)] = 79818,
  [SMALL_STATE(2436)] = 79876,
  [SMALL_STATE(2437)] = 79932,
  [SMALL_STATE(2438)] = 79988,
  [SMALL_STATE(2439)] = 80044,
  [SMALL_STATE(2440)] = 80100,
  [SMALL_STATE(2441)] = 80158,
  [SMALL_STATE(2442)] = 80214,
  [SMALL_STATE(2443)] = 80270,
  [SMALL_STATE(2444)] = 80326,
  [SMALL_STATE(2445)] = 80382,
  [SMALL_STATE(2446)] = 80438,
  [SMALL_STATE(2447)] = 80496,
  [SMALL_STATE(2448)] = 80552,
  [SMALL_STATE(2449)] = 80608,
  [SMALL_STATE(2450)] = 80664,
  [SMALL_STATE(2451)] = 80719,
  [SMALL_STATE(2452)] = 80774,
  [SMALL_STATE(2453)] = 80829,
  [SMALL_STATE(2454)] = 80884,
  [SMALL_STATE(2455)] = 80939,
  [SMALL_STATE(2456)] = 80994,
  [SMALL_STATE(2457)] = 81049,
  [SMALL_STATE(2458)] = 81104,
  [SMALL_STATE(2459)] = 81159,
  [SMALL_STATE(2460)] = 81214,
  [SMALL_STATE(2461)] = 81269,
  [SMALL_STATE(2462)] = 81378,
  [SMALL_STATE(2463)] = 81433,
  [SMALL_STATE(2464)] = 81488,
  [SMALL_STATE(2465)] = 81543,
  [SMALL_STATE(2466)] = 81598,
  [SMALL_STATE(2467)] = 81653,
  [SMALL_STATE(2468)] = 81708,
  [SMALL_STATE(2469)] = 81763,
  [SMALL_STATE(2470)] = 81818,
  [SMALL_STATE(2471)] = 81873,
  [SMALL_STATE(2472)] = 81928,
  [SMALL_STATE(2473)] = 81983,
  [SMALL_STATE(2474)] = 82038,
  [SMALL_STATE(2475)] = 82093,
  [SMALL_STATE(2476)] = 82148,
  [SMALL_STATE(2477)] = 82203,
  [SMALL_STATE(2478)] = 82258,
  [SMALL_STATE(2479)] = 82313,
  [SMALL_STATE(2480)] = 82368,
  [SMALL_STATE(2481)] = 82423,
  [SMALL_STATE(2482)] = 82478,
  [SMALL_STATE(2483)] = 82533,
  [SMALL_STATE(2484)] = 82588,
  [SMALL_STATE(2485)] = 82643,
  [SMALL_STATE(2486)] = 82698,
  [SMALL_STATE(2487)] = 82807,
  [SMALL_STATE(2488)] = 82862,
  [SMALL_STATE(2489)] = 82917,
  [SMALL_STATE(2490)] = 82972,
  [SMALL_STATE(2491)] = 83027,
  [SMALL_STATE(2492)] = 83082,
  [SMALL_STATE(2493)] = 83137,
  [SMALL_STATE(2494)] = 83246,
  [SMALL_STATE(2495)] = 83301,
  [SMALL_STATE(2496)] = 83358,
  [SMALL_STATE(2497)] = 83413,
  [SMALL_STATE(2498)] = 83468,
  [SMALL_STATE(2499)] = 83523,
  [SMALL_STATE(2500)] = 83578,
  [SMALL_STATE(2501)] = 83633,
  [SMALL_STATE(2502)] = 83688,
  [SMALL_STATE(2503)] = 83743,
  [SMALL_STATE(2504)] = 83798,
  [SMALL_STATE(2505)] = 83853,
  [SMALL_STATE(2506)] = 83908,
  [SMALL_STATE(2507)] = 83963,
  [SMALL_STATE(2508)] = 84018,
  [SMALL_STATE(2509)] = 84073,
  [SMALL_STATE(2510)] = 84128,
  [SMALL_STATE(2511)] = 84183,
  [SMALL_STATE(2512)] = 84238,
  [SMALL_STATE(2513)] = 84293,
  [SMALL_STATE(2514)] = 84348,
  [SMALL_STATE(2515)] = 84403,
  [SMALL_STATE(2516)] = 84458,
  [SMALL_STATE(2517)] = 84513,
  [SMALL_STATE(2518)] = 84617,
  [SMALL_STATE(2519)] = 84721,
  [SMALL_STATE(2520)] = 84775,
  [SMALL_STATE(2521)] = 84879,
  [SMALL_STATE(2522)] = 84943,
  [SMALL_STATE(2523)] = 84997,
  [SMALL_STATE(2524)] = 85101,
  [SMALL_STATE(2525)] = 85205,
  [SMALL_STATE(2526)] = 85259,
  [SMALL_STATE(2527)] = 85323,
  [SMALL_STATE(2528)] = 85427,
  [SMALL_STATE(2529)] = 85514,
  [SMALL_STATE(2530)] = 85617,
  [SMALL_STATE(2531)] = 85720,
  [SMALL_STATE(2532)] = 85823,
  [SMALL_STATE(2533)] = 85924,
  [SMALL_STATE(2534)] = 86025,
  [SMALL_STATE(2535)] = 86126,
  [SMALL_STATE(2536)] = 86229,
  [SMALL_STATE(2537)] = 86330,
  [SMALL_STATE(2538)] = 86393,
  [SMALL_STATE(2539)] = 86448,
  [SMALL_STATE(2540)] = 86537,
  [SMALL_STATE(2541)] = 86640,
  [SMALL_STATE(2542)] = 86729,
  [SMALL_STATE(2543)] = 86796,
  [SMALL_STATE(2544)] = 86883,
  [SMALL_STATE(2545)] = 86970,
  [SMALL_STATE(2546)] = 87059,
  [SMALL_STATE(2547)] = 87148,
  [SMALL_STATE(2548)] = 87251,
  [SMALL_STATE(2549)] = 87314,
  [SMALL_STATE(2550)] = 87399,
  [SMALL_STATE(2551)] = 87502,
  [SMALL_STATE(2552)] = 87603,
  [SMALL_STATE(2553)] = 87676,
  [SMALL_STATE(2554)] = 87751,
  [SMALL_STATE(2555)] = 87854,
  [SMALL_STATE(2556)] = 87957,
  [SMALL_STATE(2557)] = 88058,
  [SMALL_STATE(2558)] = 88159,
  [SMALL_STATE(2559)] = 88224,
  [SMALL_STATE(2560)] = 88289,
  [SMALL_STATE(2561)] = 88358,
  [SMALL_STATE(2562)] = 88437,
  [SMALL_STATE(2563)] = 88520,
  [SMALL_STATE(2564)] = 88621,
  [SMALL_STATE(2565)] = 88708,
  [SMALL_STATE(2566)] = 88797,
  [SMALL_STATE(2567)] = 88882,
  [SMALL_STATE(2568)] = 88971,
  [SMALL_STATE(2569)] = 89058,
  [SMALL_STATE(2570)] = 89131,
  [SMALL_STATE(2571)] = 89206,
  [SMALL_STATE(2572)] = 89281,
  [SMALL_STATE(2573)] = 89368,
  [SMALL_STATE(2574)] = 89433,
  [SMALL_STATE(2575)] = 89498,
  [SMALL_STATE(2576)] = 89567,
  [SMALL_STATE(2577)] = 89646,
  [SMALL_STATE(2578)] = 89729,
  [SMALL_STATE(2579)] = 89816,
  [SMALL_STATE(2580)] = 89917,
  [SMALL_STATE(2581)] = 90018,
  [SMALL_STATE(2582)] = 90119,
  [SMALL_STATE(2583)] = 90206,
  [SMALL_STATE(2584)] = 90309,
  [SMALL_STATE(2585)] = 90412,
  [SMALL_STATE(2586)] = 90515,
  [SMALL_STATE(2587)] = 90570,
  [SMALL_STATE(2588)] = 90671,
  [SMALL_STATE(2589)] = 90774,
  [SMALL_STATE(2590)] = 90839,
  [SMALL_STATE(2591)] = 90942,
  [SMALL_STATE(2592)] = 91017,
  [SMALL_STATE(2593)] = 91111,
  [SMALL_STATE(2594)] = 91163,
  [SMALL_STATE(2595)] = 91215,
  [SMALL_STATE(2596)] = 91315,
  [SMALL_STATE(2597)] = 91407,
  [SMALL_STATE(2598)] = 91501,
  [SMALL_STATE(2599)] = 91599,
  [SMALL_STATE(2600)] = 91697,
  [SMALL_STATE(2601)] = 91795,
  [SMALL_STATE(2602)] = 91889,
  [SMALL_STATE(2603)] = 91987,
  [SMALL_STATE(2604)] = 92039,
  [SMALL_STATE(2605)] = 92137,
  [SMALL_STATE(2606)] = 92189,
  [SMALL_STATE(2607)] = 92283,
  [SMALL_STATE(2608)] = 92377,
  [SMALL_STATE(2609)] = 92475,
  [SMALL_STATE(2610)] = 92573,
  [SMALL_STATE(2611)] = 92625,
  [SMALL_STATE(2612)] = 92719,
  [SMALL_STATE(2613)] = 92813,
  [SMALL_STATE(2614)] = 92911,
  [SMALL_STATE(2615)] = 92963,
  [SMALL_STATE(2616)] = 93053,
  [SMALL_STATE(2617)] = 93105,
  [SMALL_STATE(2618)] = 93199,
  [SMALL_STATE(2619)] = 93291,
  [SMALL_STATE(2620)] = 93343,
  [SMALL_STATE(2621)] = 93395,
  [SMALL_STATE(2622)] = 93447,
  [SMALL_STATE(2623)] = 93499,
  [SMALL_STATE(2624)] = 93551,
  [SMALL_STATE(2625)] = 93643,
  [SMALL_STATE(2626)] = 93695,
  [SMALL_STATE(2627)] = 93789,
  [SMALL_STATE(2628)] = 93841,
  [SMALL_STATE(2629)] = 93929,
  [SMALL_STATE(2630)] = 93981,
  [SMALL_STATE(2631)] = 94069,
  [SMALL_STATE(2632)] = 94157,
  [SMALL_STATE(2633)] = 94251,
  [SMALL_STATE(2634)] = 94303,
  [SMALL_STATE(2635)] = 94395,
  [SMALL_STATE(2636)] = 94447,
  [SMALL_STATE(2637)] = 94499,
  [SMALL_STATE(2638)] = 94551,
  [SMALL_STATE(2639)] = 94603,
  [SMALL_STATE(2640)] = 94655,
  [SMALL_STATE(2641)] = 94707,
  [SMALL_STATE(2642)] = 94805,
  [SMALL_STATE(2643)] = 94857,
  [SMALL_STATE(2644)] = 94909,
  [SMALL_STATE(2645)] = 95007,
  [SMALL_STATE(2646)] = 95059,
  [SMALL_STATE(2647)] = 95111,
  [SMALL_STATE(2648)] = 95205,
  [SMALL_STATE(2649)] = 95257,
  [SMALL_STATE(2650)] = 95309,
  [SMALL_STATE(2651)] = 95403,
  [SMALL_STATE(2652)] = 95456,
  [SMALL_STATE(2653)] = 95533,
  [SMALL_STATE(2654)] = 95584,
  [SMALL_STATE(2655)] = 95634,
  [SMALL_STATE(2656)] = 95684,
  [SMALL_STATE(2657)] = 95734,
  [SMALL_STATE(2658)] = 95784,
  [SMALL_STATE(2659)] = 95870,
  [SMALL_STATE(2660)] = 95920,
  [SMALL_STATE(2661)] = 95970,
  [SMALL_STATE(2662)] = 96020,
  [SMALL_STATE(2663)] = 96070,
  [SMALL_STATE(2664)] = 96120,
  [SMALL_STATE(2665)] = 96170,
  [SMALL_STATE(2666)] = 96256,
  [SMALL_STATE(2667)] = 96308,
  [SMALL_STATE(2668)] = 96358,
  [SMALL_STATE(2669)] = 96408,
  [SMALL_STATE(2670)] = 96494,
  [SMALL_STATE(2671)] = 96544,
  [SMALL_STATE(2672)] = 96593,
  [SMALL_STATE(2673)] = 96642,
  [SMALL_STATE(2674)] = 96731,
  [SMALL_STATE(2675)] = 96780,
  [SMALL_STATE(2676)] = 96829,
  [SMALL_STATE(2677)] = 96878,
  [SMALL_STATE(2678)] = 96927,
  [SMALL_STATE(2679)] = 96976,
  [SMALL_STATE(2680)] = 97025,
  [SMALL_STATE(2681)] = 97074,
  [SMALL_STATE(2682)] = 97123,
  [SMALL_STATE(2683)] = 97172,
  [SMALL_STATE(2684)] = 97221,
  [SMALL_STATE(2685)] = 97270,
  [SMALL_STATE(2686)] = 97319,
  [SMALL_STATE(2687)] = 97368,
  [SMALL_STATE(2688)] = 97417,
  [SMALL_STATE(2689)] = 97466,
  [SMALL_STATE(2690)] = 97515,
  [SMALL_STATE(2691)] = 97604,
  [SMALL_STATE(2692)] = 97653,
  [SMALL_STATE(2693)] = 97702,
  [SMALL_STATE(2694)] = 97751,
  [SMALL_STATE(2695)] = 97800,
  [SMALL_STATE(2696)] = 97889,
  [SMALL_STATE(2697)] = 97938,
  [SMALL_STATE(2698)] = 97987,
  [SMALL_STATE(2699)] = 98036,
  [SMALL_STATE(2700)] = 98085,
  [SMALL_STATE(2701)] = 98134,
  [SMALL_STATE(2702)] = 98183,
  [SMALL_STATE(2703)] = 98232,
  [SMALL_STATE(2704)] = 98285,
  [SMALL_STATE(2705)] = 98334,
  [SMALL_STATE(2706)] = 98383,
  [SMALL_STATE(2707)] = 98432,
  [SMALL_STATE(2708)] = 98481,
  [SMALL_STATE(2709)] = 98570,
  [SMALL_STATE(2710)] = 98659,
  [SMALL_STATE(2711)] = 98708,
  [SMALL_STATE(2712)] = 98757,
  [SMALL_STATE(2713)] = 98806,
  [SMALL_STATE(2714)] = 98855,
  [SMALL_STATE(2715)] = 98904,
  [SMALL_STATE(2716)] = 98953,
  [SMALL_STATE(2717)] = 99004,
  [SMALL_STATE(2718)] = 99053,
  [SMALL_STATE(2719)] = 99102,
  [SMALL_STATE(2720)] = 99191,
  [SMALL_STATE(2721)] = 99242,
  [SMALL_STATE(2722)] = 99295,
  [SMALL_STATE(2723)] = 99344,
  [SMALL_STATE(2724)] = 99393,
  [SMALL_STATE(2725)] = 99442,
  [SMALL_STATE(2726)] = 99497,
  [SMALL_STATE(2727)] = 99552,
  [SMALL_STATE(2728)] = 99601,
  [SMALL_STATE(2729)] = 99654,
  [SMALL_STATE(2730)] = 99703,
  [SMALL_STATE(2731)] = 99792,
  [SMALL_STATE(2732)] = 99843,
  [SMALL_STATE(2733)] = 99892,
  [SMALL_STATE(2734)] = 99947,
  [SMALL_STATE(2735)] = 99996,
  [SMALL_STATE(2736)] = 100045,
  [SMALL_STATE(2737)] = 100094,
  [SMALL_STATE(2738)] = 100143,
  [SMALL_STATE(2739)] = 100196,
  [SMALL_STATE(2740)] = 100285,
  [SMALL_STATE(2741)] = 100334,
  [SMALL_STATE(2742)] = 100383,
  [SMALL_STATE(2743)] = 100436,
  [SMALL_STATE(2744)] = 100485,
  [SMALL_STATE(2745)] = 100534,
  [SMALL_STATE(2746)] = 100583,
  [SMALL_STATE(2747)] = 100632,
  [SMALL_STATE(2748)] = 100683,
  [SMALL_STATE(2749)] = 100772,
  [SMALL_STATE(2750)] = 100821,
  [SMALL_STATE(2751)] = 100870,
  [SMALL_STATE(2752)] = 100921,
  [SMALL_STATE(2753)] = 100970,
  [SMALL_STATE(2754)] = 101019,
  [SMALL_STATE(2755)] = 101068,
  [SMALL_STATE(2756)] = 101117,
  [SMALL_STATE(2757)] = 101166,
  [SMALL_STATE(2758)] = 101215,
  [SMALL_STATE(2759)] = 101264,
  [SMALL_STATE(2760)] = 101353,
  [SMALL_STATE(2761)] = 101402,
  [SMALL_STATE(2762)] = 101491,
  [SMALL_STATE(2763)] = 101540,
  [SMALL_STATE(2764)] = 101589,
  [SMALL_STATE(2765)] = 101640,
  [SMALL_STATE(2766)] = 101689,
  [SMALL_STATE(2767)] = 101742,
  [SMALL_STATE(2768)] = 101791,
  [SMALL_STATE(2769)] = 101840,
  [SMALL_STATE(2770)] = 101895,
  [SMALL_STATE(2771)] = 101944,
  [SMALL_STATE(2772)] = 101993,
  [SMALL_STATE(2773)] = 102042,
  [SMALL_STATE(2774)] = 102097,
  [SMALL_STATE(2775)] = 102146,
  [SMALL_STATE(2776)] = 102235,
  [SMALL_STATE(2777)] = 102324,
  [SMALL_STATE(2778)] = 102410,
  [SMALL_STATE(2779)] = 102496,
  [SMALL_STATE(2780)] = 102582,
  [SMALL_STATE(2781)] = 102668,
  [SMALL_STATE(2782)] = 102754,
  [SMALL_STATE(2783)] = 102840,
  [SMALL_STATE(2784)] = 102898,
  [SMALL_STATE(2785)] = 102956,
  [SMALL_STATE(2786)] = 103042,
  [SMALL_STATE(2787)] = 103128,
  [SMALL_STATE(2788)] = 103214,
  [SMALL_STATE(2789)] = 103300,
  [SMALL_STATE(2790)] = 103384,
  [SMALL_STATE(2791)] = 103468,
  [SMALL_STATE(2792)] = 103554,
  [SMALL_STATE(2793)] = 103640,
  [SMALL_STATE(2794)] = 103726,
  [SMALL_STATE(2795)] = 103776,
  [SMALL_STATE(2796)] = 103862,
  [SMALL_STATE(2797)] = 103948,
  [SMALL_STATE(2798)] = 104028,
  [SMALL_STATE(2799)] = 104096,
  [SMALL_STATE(2800)] = 104166,
  [SMALL_STATE(2801)] = 104236,
  [SMALL_STATE(2802)] = 104318,
  [SMALL_STATE(2803)] = 104392,
  [SMALL_STATE(2804)] = 104470,
  [SMALL_STATE(2805)] = 104552,
  [SMALL_STATE(2806)] = 104634,
  [SMALL_STATE(2807)] = 104716,
  [SMALL_STATE(2808)] = 104802,
  [SMALL_STATE(2809)] = 104888,
  [SMALL_STATE(2810)] = 104974,
  [SMALL_STATE(2811)] = 105060,
  [SMALL_STATE(2812)] = 105146,
  [SMALL_STATE(2813)] = 105232,
  [SMALL_STATE(2814)] = 105318,
  [SMALL_STATE(2815)] = 105404,
  [SMALL_STATE(2816)] = 105462,
  [SMALL_STATE(2817)] = 105548,
  [SMALL_STATE(2818)] = 105608,
  [SMALL_STATE(2819)] = 105668,
  [SMALL_STATE(2820)] = 105732,
  [SMALL_STATE(2821)] = 105818,
  [SMALL_STATE(2822)] = 105866,
  [SMALL_STATE(2823)] = 105950,
  [SMALL_STATE(2824)] = 106036,
  [SMALL_STATE(2825)] = 106122,
  [SMALL_STATE(2826)] = 106208,
  [SMALL_STATE(2827)] = 106294,
  [SMALL_STATE(2828)] = 106380,
  [SMALL_STATE(2829)] = 106466,
  [SMALL_STATE(2830)] = 106552,
  [SMALL_STATE(2831)] = 106638,
  [SMALL_STATE(2832)] = 106686,
  [SMALL_STATE(2833)] = 106772,
  [SMALL_STATE(2834)] = 106858,
  [SMALL_STATE(2835)] = 106944,
  [SMALL_STATE(2836)] = 107030,
  [SMALL_STATE(2837)] = 107078,
  [SMALL_STATE(2838)] = 107164,
  [SMALL_STATE(2839)] = 107250,
  [SMALL_STATE(2840)] = 107298,
  [SMALL_STATE(2841)] = 107348,
  [SMALL_STATE(2842)] = 107434,
  [SMALL_STATE(2843)] = 107482,
  [SMALL_STATE(2844)] = 107568,
  [SMALL_STATE(2845)] = 107654,
  [SMALL_STATE(2846)] = 107704,
  [SMALL_STATE(2847)] = 107783,
  [SMALL_STATE(2848)] = 107850,
  [SMALL_STATE(2849)] = 107919,
  [SMALL_STATE(2850)] = 107988,
  [SMALL_STATE(2851)] = 108069,
  [SMALL_STATE(2852)] = 108142,
  [SMALL_STATE(2853)] = 108219,
  [SMALL_STATE(2854)] = 108300,
  [SMALL_STATE(2855)] = 108347,
  [SMALL_STATE(2856)] = 108428,
  [SMALL_STATE(2857)] = 108475,
  [SMALL_STATE(2858)] = 108558,
  [SMALL_STATE(2859)] = 108615,
  [SMALL_STATE(2860)] = 108662,
  [SMALL_STATE(2861)] = 108719,
  [SMALL_STATE(2862)] = 108802,
  [SMALL_STATE(2863)] = 108883,
  [SMALL_STATE(2864)] = 108966,
  [SMALL_STATE(2865)] = 109017,
  [SMALL_STATE(2866)] = 109098,
  [SMALL_STATE(2867)] = 109179,
  [SMALL_STATE(2868)] = 109236,
  [SMALL_STATE(2869)] = 109295,
  [SMALL_STATE(2870)] = 109354,
  [SMALL_STATE(2871)] = 109417,
  [SMALL_STATE(2872)] = 109466,
  [SMALL_STATE(2873)] = 109549,
  [SMALL_STATE(2874)] = 109632,
  [SMALL_STATE(2875)] = 109681,
  [SMALL_STATE(2876)] = 109762,
  [SMALL_STATE(2877)] = 109811,
  [SMALL_STATE(2878)] = 109890,
  [SMALL_STATE(2879)] = 109973,
  [SMALL_STATE(2880)] = 110040,
  [SMALL_STATE(2881)] = 110109,
  [SMALL_STATE(2882)] = 110178,
  [SMALL_STATE(2883)] = 110259,
  [SMALL_STATE(2884)] = 110318,
  [SMALL_STATE(2885)] = 110377,
  [SMALL_STATE(2886)] = 110440,
  [SMALL_STATE(2887)] = 110513,
  [SMALL_STATE(2888)] = 110590,
  [SMALL_STATE(2889)] = 110641,
  [SMALL_STATE(2890)] = 110719,
  [SMALL_STATE(2891)] = 110765,
  [SMALL_STATE(2892)] = 110811,
  [SMALL_STATE(2893)] = 110857,
  [SMALL_STATE(2894)] = 110903,
  [SMALL_STATE(2895)] = 110983,
  [SMALL_STATE(2896)] = 111063,
  [SMALL_STATE(2897)] = 111109,
  [SMALL_STATE(2898)] = 111171,
  [SMALL_STATE(2899)] = 111229,
  [SMALL_STATE(2900)] = 111275,
  [SMALL_STATE(2901)] = 111321,
  [SMALL_STATE(2902)] = 111403,
  [SMALL_STATE(2903)] = 111449,
  [SMALL_STATE(2904)] = 111517,
  [SMALL_STATE(2905)] = 111563,
  [SMALL_STATE(2906)] = 111609,
  [SMALL_STATE(2907)] = 111655,
  [SMALL_STATE(2908)] = 111723,
  [SMALL_STATE(2909)] = 111799,
  [SMALL_STATE(2910)] = 111881,
  [SMALL_STATE(2911)] = 111927,
  [SMALL_STATE(2912)] = 111993,
  [SMALL_STATE(2913)] = 112061,
  [SMALL_STATE(2914)] = 112129,
  [SMALL_STATE(2915)] = 112207,
  [SMALL_STATE(2916)] = 112279,
  [SMALL_STATE(2917)] = 112353,
  [SMALL_STATE(2918)] = 112399,
  [SMALL_STATE(2919)] = 112477,
  [SMALL_STATE(2920)] = 112537,
  [SMALL_STATE(2921)] = 112583,
  [SMALL_STATE(2922)] = 112629,
  [SMALL_STATE(2923)] = 112697,
  [SMALL_STATE(2924)] = 112743,
  [SMALL_STATE(2925)] = 112823,
  [SMALL_STATE(2926)] = 112869,
  [SMALL_STATE(2927)] = 112915,
  [SMALL_STATE(2928)] = 112961,
  [SMALL_STATE(2929)] = 113043,
  [SMALL_STATE(2930)] = 113111,
  [SMALL_STATE(2931)] = 113157,
  [SMALL_STATE(2932)] = 113205,
  [SMALL_STATE(2933)] = 113251,
  [SMALL_STATE(2934)] = 113297,
  [SMALL_STATE(2935)] = 113353,
  [SMALL_STATE(2936)] = 113409,
  [SMALL_STATE(2937)] = 113457,
  [SMALL_STATE(2938)] = 113503,
  [SMALL_STATE(2939)] = 113549,
  [SMALL_STATE(2940)] = 113607,
  [SMALL_STATE(2941)] = 113653,
  [SMALL_STATE(2942)] = 113711,
  [SMALL_STATE(2943)] = 113773,
  [SMALL_STATE(2944)] = 113841,
  [SMALL_STATE(2945)] = 113889,
  [SMALL_STATE(2946)] = 113935,
  [SMALL_STATE(2947)] = 113997,
  [SMALL_STATE(2948)] = 114059,
  [SMALL_STATE(2949)] = 114107,
  [SMALL_STATE(2950)] = 114153,
  [SMALL_STATE(2951)] = 114201,
  [SMALL_STATE(2952)] = 114247,
  [SMALL_STATE(2953)] = 114293,
  [SMALL_STATE(2954)] = 114339,
  [SMALL_STATE(2955)] = 114385,
  [SMALL_STATE(2956)] = 114431,
  [SMALL_STATE(2957)] = 114477,
  [SMALL_STATE(2958)] = 114523,
  [SMALL_STATE(2959)] = 114569,
  [SMALL_STATE(2960)] = 114615,
  [SMALL_STATE(2961)] = 114661,
  [SMALL_STATE(2962)] = 114707,
  [SMALL_STATE(2963)] = 114753,
  [SMALL_STATE(2964)] = 114799,
  [SMALL_STATE(2965)] = 114861,
  [SMALL_STATE(2966)] = 114923,
  [SMALL_STATE(2967)] = 114985,
  [SMALL_STATE(2968)] = 115031,
  [SMALL_STATE(2969)] = 115077,
  [SMALL_STATE(2970)] = 115123,
  [SMALL_STATE(2971)] = 115169,
  [SMALL_STATE(2972)] = 115247,
  [SMALL_STATE(2973)] = 115292,
  [SMALL_STATE(2974)] = 115359,
  [SMALL_STATE(2975)] = 115426,
  [SMALL_STATE(2976)] = 115505,
  [SMALL_STATE(2977)] = 115562,
  [SMALL_STATE(2978)] = 115619,
  [SMALL_STATE(2979)] = 115680,
  [SMALL_STATE(2980)] = 115751,
  [SMALL_STATE(2981)] = 115826,
  [SMALL_STATE(2982)] = 115905,
  [SMALL_STATE(2983)] = 115984,
  [SMALL_STATE(2984)] = 116063,
  [SMALL_STATE(2985)] = 116108,
  [SMALL_STATE(2986)] = 116183,
  [SMALL_STATE(2987)] = 116228,
  [SMALL_STATE(2988)] = 116307,
  [SMALL_STATE(2989)] = 116352,
  [SMALL_STATE(2990)] = 116431,
  [SMALL_STATE(2991)] = 116476,
  [SMALL_STATE(2992)] = 116521,
  [SMALL_STATE(2993)] = 116566,
  [SMALL_STATE(2994)] = 116611,
  [SMALL_STATE(2995)] = 116664,
  [SMALL_STATE(2996)] = 116709,
  [SMALL_STATE(2997)] = 116754,
  [SMALL_STATE(2998)] = 116801,
  [SMALL_STATE(2999)] = 116846,
  [SMALL_STATE(3000)] = 116895,
  [SMALL_STATE(3001)] = 116942,
  [SMALL_STATE(3002)] = 116987,
  [SMALL_STATE(3003)] = 117038,
  [SMALL_STATE(3004)] = 117083,
  [SMALL_STATE(3005)] = 117128,
  [SMALL_STATE(3006)] = 117173,
  [SMALL_STATE(3007)] = 117218,
  [SMALL_STATE(3008)] = 117263,
  [SMALL_STATE(3009)] = 117308,
  [SMALL_STATE(3010)] = 117387,
  [SMALL_STATE(3011)] = 117432,
  [SMALL_STATE(3012)] = 117477,
  [SMALL_STATE(3013)] = 117554,
  [SMALL_STATE(3014)] = 117635,
  [SMALL_STATE(3015)] = 117680,
  [SMALL_STATE(3016)] = 117745,
  [SMALL_STATE(3017)] = 117812,
  [SMALL_STATE(3018)] = 117857,
  [SMALL_STATE(3019)] = 117902,
  [SMALL_STATE(3020)] = 117947,
  [SMALL_STATE(3021)] = 117994,
  [SMALL_STATE(3022)] = 118061,
  [SMALL_STATE(3023)] = 118140,
  [SMALL_STATE(3024)] = 118211,
  [SMALL_STATE(3025)] = 118256,
  [SMALL_STATE(3026)] = 118301,
  [SMALL_STATE(3027)] = 118346,
  [SMALL_STATE(3028)] = 118391,
  [SMALL_STATE(3029)] = 118436,
  [SMALL_STATE(3030)] = 118511,
  [SMALL_STATE(3031)] = 118556,
  [SMALL_STATE(3032)] = 118601,
  [SMALL_STATE(3033)] = 118646,
  [SMALL_STATE(3034)] = 118725,
  [SMALL_STATE(3035)] = 118804,
  [SMALL_STATE(3036)] = 118849,
  [SMALL_STATE(3037)] = 118894,
  [SMALL_STATE(3038)] = 118939,
  [SMALL_STATE(3039)] = 118984,
  [SMALL_STATE(3040)] = 119065,
  [SMALL_STATE(3041)] = 119110,
  [SMALL_STATE(3042)] = 119155,
  [SMALL_STATE(3043)] = 119202,
  [SMALL_STATE(3044)] = 119279,
  [SMALL_STATE(3045)] = 119360,
  [SMALL_STATE(3046)] = 119405,
  [SMALL_STATE(3047)] = 119450,
  [SMALL_STATE(3048)] = 119529,
  [SMALL_STATE(3049)] = 119608,
  [SMALL_STATE(3050)] = 119689,
  [SMALL_STATE(3051)] = 119770,
  [SMALL_STATE(3052)] = 119815,
  [SMALL_STATE(3053)] = 119860,
  [SMALL_STATE(3054)] = 119907,
  [SMALL_STATE(3055)] = 119982,
  [SMALL_STATE(3056)] = 120055,
  [SMALL_STATE(3057)] = 120100,
  [SMALL_STATE(3058)] = 120145,
  [SMALL_STATE(3059)] = 120190,
  [SMALL_STATE(3060)] = 120235,
  [SMALL_STATE(3061)] = 120280,
  [SMALL_STATE(3062)] = 120325,
  [SMALL_STATE(3063)] = 120370,
  [SMALL_STATE(3064)] = 120415,
  [SMALL_STATE(3065)] = 120482,
  [SMALL_STATE(3066)] = 120527,
  [SMALL_STATE(3067)] = 120572,
  [SMALL_STATE(3068)] = 120617,
  [SMALL_STATE(3069)] = 120662,
  [SMALL_STATE(3070)] = 120739,
  [SMALL_STATE(3071)] = 120784,
  [SMALL_STATE(3072)] = 120829,
  [SMALL_STATE(3073)] = 120874,
  [SMALL_STATE(3074)] = 120919,
  [SMALL_STATE(3075)] = 120964,
  [SMALL_STATE(3076)] = 121009,
  [SMALL_STATE(3077)] = 121054,
  [SMALL_STATE(3078)] = 121099,
  [SMALL_STATE(3079)] = 121182,
  [SMALL_STATE(3080)] = 121227,
  [SMALL_STATE(3081)] = 121272,
  [SMALL_STATE(3082)] = 121327,
  [SMALL_STATE(3083)] = 121384,
  [SMALL_STATE(3084)] = 121441,
  [SMALL_STATE(3085)] = 121502,
  [SMALL_STATE(3086)] = 121547,
  [SMALL_STATE(3087)] = 121592,
  [SMALL_STATE(3088)] = 121637,
  [SMALL_STATE(3089)] = 121682,
  [SMALL_STATE(3090)] = 121727,
  [SMALL_STATE(3091)] = 121772,
  [SMALL_STATE(3092)] = 121853,
  [SMALL_STATE(3093)] = 121898,
  [SMALL_STATE(3094)] = 121943,
  [SMALL_STATE(3095)] = 121988,
  [SMALL_STATE(3096)] = 122033,
  [SMALL_STATE(3097)] = 122078,
  [SMALL_STATE(3098)] = 122157,
  [SMALL_STATE(3099)] = 122240,
  [SMALL_STATE(3100)] = 122319,
  [SMALL_STATE(3101)] = 122398,
  [SMALL_STATE(3102)] = 122443,
  [SMALL_STATE(3103)] = 122500,
  [SMALL_STATE(3104)] = 122545,
  [SMALL_STATE(3105)] = 122590,
  [SMALL_STATE(3106)] = 122647,
  [SMALL_STATE(3107)] = 122708,
  [SMALL_STATE(3108)] = 122753,
  [SMALL_STATE(3109)] = 122798,
  [SMALL_STATE(3110)] = 122843,
  [SMALL_STATE(3111)] = 122888,
  [SMALL_STATE(3112)] = 122941,
  [SMALL_STATE(3113)] = 122986,
  [SMALL_STATE(3114)] = 123031,
  [SMALL_STATE(3115)] = 123076,
  [SMALL_STATE(3116)] = 123121,
  [SMALL_STATE(3117)] = 123166,
  [SMALL_STATE(3118)] = 123211,
  [SMALL_STATE(3119)] = 123282,
  [SMALL_STATE(3120)] = 123327,
  [SMALL_STATE(3121)] = 123372,
  [SMALL_STATE(3122)] = 123417,
  [SMALL_STATE(3123)] = 123462,
  [SMALL_STATE(3124)] = 123517,
  [SMALL_STATE(3125)] = 123562,
  [SMALL_STATE(3126)] = 123607,
  [SMALL_STATE(3127)] = 123652,
  [SMALL_STATE(3128)] = 123697,
  [SMALL_STATE(3129)] = 123752,
  [SMALL_STATE(3130)] = 123797,
  [SMALL_STATE(3131)] = 123874,
  [SMALL_STATE(3132)] = 123919,
  [SMALL_STATE(3133)] = 123974,
  [SMALL_STATE(3134)] = 124039,
  [SMALL_STATE(3135)] = 124084,
  [SMALL_STATE(3136)] = 124129,
  [SMALL_STATE(3137)] = 124206,
  [SMALL_STATE(3138)] = 124251,
  [SMALL_STATE(3139)] = 124316,
  [SMALL_STATE(3140)] = 124365,
  [SMALL_STATE(3141)] = 124432,
  [SMALL_STATE(3142)] = 124479,
  [SMALL_STATE(3143)] = 124524,
  [SMALL_STATE(3144)] = 124569,
  [SMALL_STATE(3145)] = 124620,
  [SMALL_STATE(3146)] = 124665,
  [SMALL_STATE(3147)] = 124710,
  [SMALL_STATE(3148)] = 124790,
  [SMALL_STATE(3149)] = 124870,
  [SMALL_STATE(3150)] = 124930,
  [SMALL_STATE(3151)] = 124976,
  [SMALL_STATE(3152)] = 125036,
  [SMALL_STATE(3153)] = 125084,
  [SMALL_STATE(3154)] = 125130,
  [SMALL_STATE(3155)] = 125190,
  [SMALL_STATE(3156)] = 125250,
  [SMALL_STATE(3157)] = 125324,
  [SMALL_STATE(3158)] = 125378,
  [SMALL_STATE(3159)] = 125424,
  [SMALL_STATE(3160)] = 125496,
  [SMALL_STATE(3161)] = 125568,
  [SMALL_STATE(3162)] = 125628,
  [SMALL_STATE(3163)] = 125708,
  [SMALL_STATE(3164)] = 125756,
  [SMALL_STATE(3165)] = 125810,
  [SMALL_STATE(3166)] = 125884,
  [SMALL_STATE(3167)] = 125948,
  [SMALL_STATE(3168)] = 126028,
  [SMALL_STATE(3169)] = 126094,
  [SMALL_STATE(3170)] = 126170,
  [SMALL_STATE(3171)] = 126226,
  [SMALL_STATE(3172)] = 126282,
  [SMALL_STATE(3173)] = 126342,
  [SMALL_STATE(3174)] = 126412,
  [SMALL_STATE(3175)] = 126484,
  [SMALL_STATE(3176)] = 126560,
  [SMALL_STATE(3177)] = 126608,
  [SMALL_STATE(3178)] = 126656,
  [SMALL_STATE(3179)] = 126732,
  [SMALL_STATE(3180)] = 126804,
  [SMALL_STATE(3181)] = 126880,
  [SMALL_STATE(3182)] = 126940,
  [SMALL_STATE(3183)] = 127018,
  [SMALL_STATE(3184)] = 127096,
  [SMALL_STATE(3185)] = 127174,
  [SMALL_STATE(3186)] = 127222,
  [SMALL_STATE(3187)] = 127270,
  [SMALL_STATE(3188)] = 127316,
  [SMALL_STATE(3189)] = 127376,
  [SMALL_STATE(3190)] = 127442,
  [SMALL_STATE(3191)] = 127485,
  [SMALL_STATE(3192)] = 127556,
  [SMALL_STATE(3193)] = 127603,
  [SMALL_STATE(3194)] = 127648,
  [SMALL_STATE(3195)] = 127719,
  [SMALL_STATE(3196)] = 127766,
  [SMALL_STATE(3197)] = 127839,
  [SMALL_STATE(3198)] = 127914,
  [SMALL_STATE(3199)] = 127989,
  [SMALL_STATE(3200)] = 128032,
  [SMALL_STATE(3201)] = 128107,
  [SMALL_STATE(3202)] = 128182,
  [SMALL_STATE(3203)] = 128225,
  [SMALL_STATE(3204)] = 128270,
  [SMALL_STATE(3205)] = 128315,
  [SMALL_STATE(3206)] = 128372,
  [SMALL_STATE(3207)] = 128431,
  [SMALL_STATE(3208)] = 128474,
  [SMALL_STATE(3209)] = 128549,
  [SMALL_STATE(3210)] = 128624,
  [SMALL_STATE(3211)] = 128701,
  [SMALL_STATE(3212)] = 128778,
  [SMALL_STATE(3213)] = 128825,
  [SMALL_STATE(3214)] = 128900,
  [SMALL_STATE(3215)] = 128975,
  [SMALL_STATE(3216)] = 129050,
  [SMALL_STATE(3217)] = 129105,
  [SMALL_STATE(3218)] = 129182,
  [SMALL_STATE(3219)] = 129229,
  [SMALL_STATE(3220)] = 129276,
  [SMALL_STATE(3221)] = 129351,
  [SMALL_STATE(3222)] = 129426,
  [SMALL_STATE(3223)] = 129469,
  [SMALL_STATE(3224)] = 129546,
  [SMALL_STATE(3225)] = 129589,
  [SMALL_STATE(3226)] = 129664,
  [SMALL_STATE(3227)] = 129739,
  [SMALL_STATE(3228)] = 129816,
  [SMALL_STATE(3229)] = 129863,
  [SMALL_STATE(3230)] = 129938,
  [SMALL_STATE(3231)] = 129983,
  [SMALL_STATE(3232)] = 130058,
  [SMALL_STATE(3233)] = 130105,
  [SMALL_STATE(3234)] = 130148,
  [SMALL_STATE(3235)] = 130195,
  [SMALL_STATE(3236)] = 130238,
  [SMALL_STATE(3237)] = 130283,
  [SMALL_STATE(3238)] = 130342,
  [SMALL_STATE(3239)] = 130387,
  [SMALL_STATE(3240)] = 130460,
  [SMALL_STATE(3241)] = 130535,
  [SMALL_STATE(3242)] = 130610,
  [SMALL_STATE(3243)] = 130657,
  [SMALL_STATE(3244)] = 130718,
  [SMALL_STATE(3245)] = 130791,
  [SMALL_STATE(3246)] = 130868,
  [SMALL_STATE(3247)] = 130943,
  [SMALL_STATE(3248)] = 131004,
  [SMALL_STATE(3249)] = 131079,
  [SMALL_STATE(3250)] = 131124,
  [SMALL_STATE(3251)] = 131176,
  [SMALL_STATE(3252)] = 131226,
  [SMALL_STATE(3253)] = 131298,
  [SMALL_STATE(3254)] = 131340,
  [SMALL_STATE(3255)] = 131390,
  [SMALL_STATE(3256)] = 131458,
  [SMALL_STATE(3257)] = 131530,
  [SMALL_STATE(3258)] = 131590,
  [SMALL_STATE(3259)] = 131662,
  [SMALL_STATE(3260)] = 131704,
  [SMALL_STATE(3261)] = 131746,
  [SMALL_STATE(3262)] = 131798,
  [SMALL_STATE(3263)] = 131840,
  [SMALL_STATE(3264)] = 131896,
  [SMALL_STATE(3265)] = 131946,
  [SMALL_STATE(3266)] = 131988,
  [SMALL_STATE(3267)] = 132034,
  [SMALL_STATE(3268)] = 132076,
  [SMALL_STATE(3269)] = 132118,
  [SMALL_STATE(3270)] = 132160,
  [SMALL_STATE(3271)] = 132206,
  [SMALL_STATE(3272)] = 132278,
  [SMALL_STATE(3273)] = 132322,
  [SMALL_STATE(3274)] = 132366,
  [SMALL_STATE(3275)] = 132438,
  [SMALL_STATE(3276)] = 132480,
  [SMALL_STATE(3277)] = 132524,
  [SMALL_STATE(3278)] = 132596,
  [SMALL_STATE(3279)] = 132664,
  [SMALL_STATE(3280)] = 132706,
  [SMALL_STATE(3281)] = 132778,
  [SMALL_STATE(3282)] = 132838,
  [SMALL_STATE(3283)] = 132894,
  [SMALL_STATE(3284)] = 132936,
  [SMALL_STATE(3285)] = 132982,
  [SMALL_STATE(3286)] = 133032,
  [SMALL_STATE(3287)] = 133074,
  [SMALL_STATE(3288)] = 133146,
  [SMALL_STATE(3289)] = 133218,
  [SMALL_STATE(3290)] = 133268,
  [SMALL_STATE(3291)] = 133324,
  [SMALL_STATE(3292)] = 133374,
  [SMALL_STATE(3293)] = 133446,
  [SMALL_STATE(3294)] = 133518,
  [SMALL_STATE(3295)] = 133590,
  [SMALL_STATE(3296)] = 133632,
  [SMALL_STATE(3297)] = 133679,
  [SMALL_STATE(3298)] = 133720,
  [SMALL_STATE(3299)] = 133765,
  [SMALL_STATE(3300)] = 133810,
  [SMALL_STATE(3301)] = 133851,
  [SMALL_STATE(3302)] = 133894,
  [SMALL_STATE(3303)] = 133935,
  [SMALL_STATE(3304)] = 133982,
  [SMALL_STATE(3305)] = 134023,
  [SMALL_STATE(3306)] = 134064,
  [SMALL_STATE(3307)] = 134105,
  [SMALL_STATE(3308)] = 134156,
  [SMALL_STATE(3309)] = 134213,
  [SMALL_STATE(3310)] = 134260,
  [SMALL_STATE(3311)] = 134307,
  [SMALL_STATE(3312)] = 134354,
  [SMALL_STATE(3313)] = 134399,
  [SMALL_STATE(3314)] = 134440,
  [SMALL_STATE(3315)] = 134487,
  [SMALL_STATE(3316)] = 134532,
  [SMALL_STATE(3317)] = 134583,
  [SMALL_STATE(3318)] = 134652,
  [SMALL_STATE(3319)] = 134697,
  [SMALL_STATE(3320)] = 134742,
  [SMALL_STATE(3321)] = 134783,
  [SMALL_STATE(3322)] = 134828,
  [SMALL_STATE(3323)] = 134871,
  [SMALL_STATE(3324)] = 134924,
  [SMALL_STATE(3325)] = 134965,
  [SMALL_STATE(3326)] = 135008,
  [SMALL_STATE(3327)] = 135053,
  [SMALL_STATE(3328)] = 135098,
  [SMALL_STATE(3329)] = 135155,
  [SMALL_STATE(3330)] = 135200,
  [SMALL_STATE(3331)] = 135245,
  [SMALL_STATE(3332)] = 135286,
  [SMALL_STATE(3333)] = 135327,
  [SMALL_STATE(3334)] = 135368,
  [SMALL_STATE(3335)] = 135413,
  [SMALL_STATE(3336)] = 135460,
  [SMALL_STATE(3337)] = 135501,
  [SMALL_STATE(3338)] = 135546,
  [SMALL_STATE(3339)] = 135601,
  [SMALL_STATE(3340)] = 135646,
  [SMALL_STATE(3341)] = 135687,
  [SMALL_STATE(3342)] = 135744,
  [SMALL_STATE(3343)] = 135789,
  [SMALL_STATE(3344)] = 135846,
  [SMALL_STATE(3345)] = 135887,
  [SMALL_STATE(3346)] = 135934,
  [SMALL_STATE(3347)] = 135977,
  [SMALL_STATE(3348)] = 136022,
  [SMALL_STATE(3349)] = 136093,
  [SMALL_STATE(3350)] = 136138,
  [SMALL_STATE(3351)] = 136179,
  [SMALL_STATE(3352)] = 136247,
  [SMALL_STATE(3353)] = 136315,
  [SMALL_STATE(3354)] = 136361,
  [SMALL_STATE(3355)] = 136403,
  [SMALL_STATE(3356)] = 136471,
  [SMALL_STATE(3357)] = 136517,
  [SMALL_STATE(3358)] = 136563,
  [SMALL_STATE(3359)] = 136603,
  [SMALL_STATE(3360)] = 136653,
  [SMALL_STATE(3361)] = 136693,
  [SMALL_STATE(3362)] = 136743,
  [SMALL_STATE(3363)] = 136811,
  [SMALL_STATE(3364)] = 136879,
  [SMALL_STATE(3365)] = 136933,
  [SMALL_STATE(3366)] = 136987,
  [SMALL_STATE(3367)] = 137055,
  [SMALL_STATE(3368)] = 137123,
  [SMALL_STATE(3369)] = 137191,
  [SMALL_STATE(3370)] = 137259,
  [SMALL_STATE(3371)] = 137327,
  [SMALL_STATE(3372)] = 137395,
  [SMALL_STATE(3373)] = 137463,
  [SMALL_STATE(3374)] = 137509,
  [SMALL_STATE(3375)] = 137549,
  [SMALL_STATE(3376)] = 137617,
  [SMALL_STATE(3377)] = 137685,
  [SMALL_STATE(3378)] = 137729,
  [SMALL_STATE(3379)] = 137797,
  [SMALL_STATE(3380)] = 137865,
  [SMALL_STATE(3381)] = 137905,
  [SMALL_STATE(3382)] = 137949,
  [SMALL_STATE(3383)] = 138017,
  [SMALL_STATE(3384)] = 138059,
  [SMALL_STATE(3385)] = 138127,
  [SMALL_STATE(3386)] = 138183,
  [SMALL_STATE(3387)] = 138249,
  [SMALL_STATE(3388)] = 138289,
  [SMALL_STATE(3389)] = 138329,
  [SMALL_STATE(3390)] = 138397,
  [SMALL_STATE(3391)] = 138441,
  [SMALL_STATE(3392)] = 138509,
  [SMALL_STATE(3393)] = 138577,
  [SMALL_STATE(3394)] = 138645,
  [SMALL_STATE(3395)] = 138713,
  [SMALL_STATE(3396)] = 138781,
  [SMALL_STATE(3397)] = 138849,
  [SMALL_STATE(3398)] = 138889,
  [SMALL_STATE(3399)] = 138929,
  [SMALL_STATE(3400)] = 138969,
  [SMALL_STATE(3401)] = 139025,
  [SMALL_STATE(3402)] = 139093,
  [SMALL_STATE(3403)] = 139137,
  [SMALL_STATE(3404)] = 139205,
  [SMALL_STATE(3405)] = 139273,
  [SMALL_STATE(3406)] = 139319,
  [SMALL_STATE(3407)] = 139387,
  [SMALL_STATE(3408)] = 139455,
  [SMALL_STATE(3409)] = 139523,
  [SMALL_STATE(3410)] = 139591,
  [SMALL_STATE(3411)] = 139659,
  [SMALL_STATE(3412)] = 139727,
  [SMALL_STATE(3413)] = 139795,
  [SMALL_STATE(3414)] = 139863,
  [SMALL_STATE(3415)] = 139931,
  [SMALL_STATE(3416)] = 139999,
  [SMALL_STATE(3417)] = 140067,
  [SMALL_STATE(3418)] = 140107,
  [SMALL_STATE(3419)] = 140147,
  [SMALL_STATE(3420)] = 140215,
  [SMALL_STATE(3421)] = 140281,
  [SMALL_STATE(3422)] = 140349,
  [SMALL_STATE(3423)] = 140415,
  [SMALL_STATE(3424)] = 140483,
  [SMALL_STATE(3425)] = 140551,
  [SMALL_STATE(3426)] = 140619,
  [SMALL_STATE(3427)] = 140687,
  [SMALL_STATE(3428)] = 140755,
  [SMALL_STATE(3429)] = 140801,
  [SMALL_STATE(3430)] = 140869,
  [SMALL_STATE(3431)] = 140937,
  [SMALL_STATE(3432)] = 141005,
  [SMALL_STATE(3433)] = 141073,
  [SMALL_STATE(3434)] = 141113,
  [SMALL_STATE(3435)] = 141181,
  [SMALL_STATE(3436)] = 141249,
  [SMALL_STATE(3437)] = 141317,
  [SMALL_STATE(3438)] = 141385,
  [SMALL_STATE(3439)] = 141425,
  [SMALL_STATE(3440)] = 141493,
  [SMALL_STATE(3441)] = 141561,
  [SMALL_STATE(3442)] = 141629,
  [SMALL_STATE(3443)] = 141697,
  [SMALL_STATE(3444)] = 141737,
  [SMALL_STATE(3445)] = 141777,
  [SMALL_STATE(3446)] = 141845,
  [SMALL_STATE(3447)] = 141911,
  [SMALL_STATE(3448)] = 141979,
  [SMALL_STATE(3449)] = 142047,
  [SMALL_STATE(3450)] = 142089,
  [SMALL_STATE(3451)] = 142157,
  [SMALL_STATE(3452)] = 142225,
  [SMALL_STATE(3453)] = 142293,
  [SMALL_STATE(3454)] = 142339,
  [SMALL_STATE(3455)] = 142381,
  [SMALL_STATE(3456)] = 142447,
  [SMALL_STATE(3457)] = 142513,
  [SMALL_STATE(3458)] = 142567,
  [SMALL_STATE(3459)] = 142621,
  [SMALL_STATE(3460)] = 142661,
  [SMALL_STATE(3461)] = 142727,
  [SMALL_STATE(3462)] = 142793,
  [SMALL_STATE(3463)] = 142859,
  [SMALL_STATE(3464)] = 142927,
  [SMALL_STATE(3465)] = 142993,
  [SMALL_STATE(3466)] = 143061,
  [SMALL_STATE(3467)] = 143129,
  [SMALL_STATE(3468)] = 143197,
  [SMALL_STATE(3469)] = 143243,
  [SMALL_STATE(3470)] = 143311,
  [SMALL_STATE(3471)] = 143379,
  [SMALL_STATE(3472)] = 143423,
  [SMALL_STATE(3473)] = 143463,
  [SMALL_STATE(3474)] = 143531,
  [SMALL_STATE(3475)] = 143599,
  [SMALL_STATE(3476)] = 143667,
  [SMALL_STATE(3477)] = 143707,
  [SMALL_STATE(3478)] = 143751,
  [SMALL_STATE(3479)] = 143819,
  [SMALL_STATE(3480)] = 143887,
  [SMALL_STATE(3481)] = 143938,
  [SMALL_STATE(3482)] = 143991,
  [SMALL_STATE(3483)] = 144040,
  [SMALL_STATE(3484)] = 144091,
  [SMALL_STATE(3485)] = 144142,
  [SMALL_STATE(3486)] = 144193,
  [SMALL_STATE(3487)] = 144244,
  [SMALL_STATE(3488)] = 144295,
  [SMALL_STATE(3489)] = 144346,
  [SMALL_STATE(3490)] = 144397,
  [SMALL_STATE(3491)] = 144448,
  [SMALL_STATE(3492)] = 144489,
  [SMALL_STATE(3493)] = 144530,
  [SMALL_STATE(3494)] = 144583,
  [SMALL_STATE(3495)] = 144634,
  [SMALL_STATE(3496)] = 144673,
  [SMALL_STATE(3497)] = 144724,
  [SMALL_STATE(3498)] = 144775,
  [SMALL_STATE(3499)] = 144826,
  [SMALL_STATE(3500)] = 144877,
  [SMALL_STATE(3501)] = 144928,
  [SMALL_STATE(3502)] = 144979,
  [SMALL_STATE(3503)] = 145018,
  [SMALL_STATE(3504)] = 145063,
  [SMALL_STATE(3505)] = 145106,
  [SMALL_STATE(3506)] = 145145,
  [SMALL_STATE(3507)] = 145196,
  [SMALL_STATE(3508)] = 145235,
  [SMALL_STATE(3509)] = 145278,
  [SMALL_STATE(3510)] = 145319,
  [SMALL_STATE(3511)] = 145368,
  [SMALL_STATE(3512)] = 145409,
  [SMALL_STATE(3513)] = 145460,
  [SMALL_STATE(3514)] = 145501,
  [SMALL_STATE(3515)] = 145539,
  [SMALL_STATE(3516)] = 145577,
  [SMALL_STATE(3517)] = 145619,
  [SMALL_STATE(3518)] = 145685,
  [SMALL_STATE(3519)] = 145727,
  [SMALL_STATE(3520)] = 145769,
  [SMALL_STATE(3521)] = 145813,
  [SMALL_STATE(3522)] = 145855,
  [SMALL_STATE(3523)] = 145899,
  [SMALL_STATE(3524)] = 145965,
  [SMALL_STATE(3525)] = 146009,
  [SMALL_STATE(3526)] = 146053,
  [SMALL_STATE(3527)] = 146097,
  [SMALL_STATE(3528)] = 146141,
  [SMALL_STATE(3529)] = 146185,
  [SMALL_STATE(3530)] = 146229,
  [SMALL_STATE(3531)] = 146273,
  [SMALL_STATE(3532)] = 146339,
  [SMALL_STATE(3533)] = 146405,
  [SMALL_STATE(3534)] = 146455,
  [SMALL_STATE(3535)] = 146521,
  [SMALL_STATE(3536)] = 146563,
  [SMALL_STATE(3537)] = 146607,
  [SMALL_STATE(3538)] = 146673,
  [SMALL_STATE(3539)] = 146717,
  [SMALL_STATE(3540)] = 146761,
  [SMALL_STATE(3541)] = 146799,
  [SMALL_STATE(3542)] = 146837,
  [SMALL_STATE(3543)] = 146887,
  [SMALL_STATE(3544)] = 146925,
  [SMALL_STATE(3545)] = 146969,
  [SMALL_STATE(3546)] = 147009,
  [SMALL_STATE(3547)] = 147053,
  [SMALL_STATE(3548)] = 147095,
  [SMALL_STATE(3549)] = 147139,
  [SMALL_STATE(3550)] = 147177,
  [SMALL_STATE(3551)] = 147221,
  [SMALL_STATE(3552)] = 147263,
  [SMALL_STATE(3553)] = 147307,
  [SMALL_STATE(3554)] = 147351,
  [SMALL_STATE(3555)] = 147394,
  [SMALL_STATE(3556)] = 147441,
  [SMALL_STATE(3557)] = 147484,
  [SMALL_STATE(3558)] = 147531,
  [SMALL_STATE(3559)] = 147578,
  [SMALL_STATE(3560)] = 147621,
  [SMALL_STATE(3561)] = 147664,
  [SMALL_STATE(3562)] = 147707,
  [SMALL_STATE(3563)] = 147748,
  [SMALL_STATE(3564)] = 147791,
  [SMALL_STATE(3565)] = 147834,
  [SMALL_STATE(3566)] = 147875,
  [SMALL_STATE(3567)] = 147918,
  [SMALL_STATE(3568)] = 147961,
  [SMALL_STATE(3569)] = 148004,
  [SMALL_STATE(3570)] = 148047,
  [SMALL_STATE(3571)] = 148090,
  [SMALL_STATE(3572)] = 148133,
  [SMALL_STATE(3573)] = 148180,
  [SMALL_STATE(3574)] = 148223,
  [SMALL_STATE(3575)] = 148266,
  [SMALL_STATE(3576)] = 148309,
  [SMALL_STATE(3577)] = 148372,
  [SMALL_STATE(3578)] = 148415,
  [SMALL_STATE(3579)] = 148464,
  [SMALL_STATE(3580)] = 148513,
  [SMALL_STATE(3581)] = 148560,
  [SMALL_STATE(3582)] = 148603,
  [SMALL_STATE(3583)] = 148646,
  [SMALL_STATE(3584)] = 148693,
  [SMALL_STATE(3585)] = 148756,
  [SMALL_STATE(3586)] = 148799,
  [SMALL_STATE(3587)] = 148836,
  [SMALL_STATE(3588)] = 148879,
  [SMALL_STATE(3589)] = 148922,
  [SMALL_STATE(3590)] = 148985,
  [SMALL_STATE(3591)] = 149028,
  [SMALL_STATE(3592)] = 149075,
  [SMALL_STATE(3593)] = 149118,
  [SMALL_STATE(3594)] = 149155,
  [SMALL_STATE(3595)] = 149202,
  [SMALL_STATE(3596)] = 149249,
  [SMALL_STATE(3597)] = 149292,
  [SMALL_STATE(3598)] = 149339,
  [SMALL_STATE(3599)] = 149382,
  [SMALL_STATE(3600)] = 149425,
  [SMALL_STATE(3601)] = 149468,
  [SMALL_STATE(3602)] = 149507,
  [SMALL_STATE(3603)] = 149550,
  [SMALL_STATE(3604)] = 149593,
  [SMALL_STATE(3605)] = 149636,
  [SMALL_STATE(3606)] = 149699,
  [SMALL_STATE(3607)] = 149740,
  [SMALL_STATE(3608)] = 149779,
  [SMALL_STATE(3609)] = 149822,
  [SMALL_STATE(3610)] = 149865,
  [SMALL_STATE(3611)] = 149908,
  [SMALL_STATE(3612)] = 149971,
  [SMALL_STATE(3613)] = 150014,
  [SMALL_STATE(3614)] = 150057,
  [SMALL_STATE(3615)] = 150100,
  [SMALL_STATE(3616)] = 150143,
  [SMALL_STATE(3617)] = 150206,
  [SMALL_STATE(3618)] = 150249,
  [SMALL_STATE(3619)] = 150312,
  [SMALL_STATE(3620)] = 150349,
  [SMALL_STATE(3621)] = 150392,
  [SMALL_STATE(3622)] = 150455,
  [SMALL_STATE(3623)] = 150498,
  [SMALL_STATE(3624)] = 150541,
  [SMALL_STATE(3625)] = 150584,
  [SMALL_STATE(3626)] = 150647,
  [SMALL_STATE(3627)] = 150710,
  [SMALL_STATE(3628)] = 150755,
  [SMALL_STATE(3629)] = 150792,
  [SMALL_STATE(3630)] = 150829,
  [SMALL_STATE(3631)] = 150874,
  [SMALL_STATE(3632)] = 150919,
  [SMALL_STATE(3633)] = 150960,
  [SMALL_STATE(3634)] = 151007,
  [SMALL_STATE(3635)] = 151052,
  [SMALL_STATE(3636)] = 151093,
  [SMALL_STATE(3637)] = 151132,
  [SMALL_STATE(3638)] = 151179,
  [SMALL_STATE(3639)] = 151224,
  [SMALL_STATE(3640)] = 151269,
  [SMALL_STATE(3641)] = 151314,
  [SMALL_STATE(3642)] = 151359,
  [SMALL_STATE(3643)] = 151404,
  [SMALL_STATE(3644)] = 151447,
  [SMALL_STATE(3645)] = 151492,
  [SMALL_STATE(3646)] = 151535,
  [SMALL_STATE(3647)] = 151572,
  [SMALL_STATE(3648)] = 151615,
  [SMALL_STATE(3649)] = 151662,
  [SMALL_STATE(3650)] = 151705,
  [SMALL_STATE(3651)] = 151752,
  [SMALL_STATE(3652)] = 151795,
  [SMALL_STATE(3653)] = 151842,
  [SMALL_STATE(3654)] = 151885,
  [SMALL_STATE(3655)] = 151932,
  [SMALL_STATE(3656)] = 151975,
  [SMALL_STATE(3657)] = 152018,
  [SMALL_STATE(3658)] = 152061,
  [SMALL_STATE(3659)] = 152104,
  [SMALL_STATE(3660)] = 152147,
  [SMALL_STATE(3661)] = 152190,
  [SMALL_STATE(3662)] = 152233,
  [SMALL_STATE(3663)] = 152276,
  [SMALL_STATE(3664)] = 152319,
  [SMALL_STATE(3665)] = 152362,
  [SMALL_STATE(3666)] = 152405,
  [SMALL_STATE(3667)] = 152448,
  [SMALL_STATE(3668)] = 152485,
  [SMALL_STATE(3669)] = 152548,
  [SMALL_STATE(3670)] = 152591,
  [SMALL_STATE(3671)] = 152632,
  [SMALL_STATE(3672)] = 152673,
  [SMALL_STATE(3673)] = 152712,
  [SMALL_STATE(3674)] = 152755,
  [SMALL_STATE(3675)] = 152798,
  [SMALL_STATE(3676)] = 152845,
  [SMALL_STATE(3677)] = 152888,
  [SMALL_STATE(3678)] = 152931,
  [SMALL_STATE(3679)] = 152978,
  [SMALL_STATE(3680)] = 153021,
  [SMALL_STATE(3681)] = 153068,
  [SMALL_STATE(3682)] = 153111,
  [SMALL_STATE(3683)] = 153154,
  [SMALL_STATE(3684)] = 153197,
  [SMALL_STATE(3685)] = 153240,
  [SMALL_STATE(3686)] = 153283,
  [SMALL_STATE(3687)] = 153326,
  [SMALL_STATE(3688)] = 153369,
  [SMALL_STATE(3689)] = 153412,
  [SMALL_STATE(3690)] = 153455,
  [SMALL_STATE(3691)] = 153498,
  [SMALL_STATE(3692)] = 153535,
  [SMALL_STATE(3693)] = 153578,
  [SMALL_STATE(3694)] = 153621,
  [SMALL_STATE(3695)] = 153668,
  [SMALL_STATE(3696)] = 153711,
  [SMALL_STATE(3697)] = 153754,
  [SMALL_STATE(3698)] = 153797,
  [SMALL_STATE(3699)] = 153840,
  [SMALL_STATE(3700)] = 153883,
  [SMALL_STATE(3701)] = 153926,
  [SMALL_STATE(3702)] = 153969,
  [SMALL_STATE(3703)] = 154012,
  [SMALL_STATE(3704)] = 154055,
  [SMALL_STATE(3705)] = 154098,
  [SMALL_STATE(3706)] = 154141,
  [SMALL_STATE(3707)] = 154184,
  [SMALL_STATE(3708)] = 154227,
  [SMALL_STATE(3709)] = 154270,
  [SMALL_STATE(3710)] = 154313,
  [SMALL_STATE(3711)] = 154356,
  [SMALL_STATE(3712)] = 154399,
  [SMALL_STATE(3713)] = 154442,
  [SMALL_STATE(3714)] = 154485,
  [SMALL_STATE(3715)] = 154528,
  [SMALL_STATE(3716)] = 154571,
  [SMALL_STATE(3717)] = 154614,
  [SMALL_STATE(3718)] = 154657,
  [SMALL_STATE(3719)] = 154700,
  [SMALL_STATE(3720)] = 154745,
  [SMALL_STATE(3721)] = 154788,
  [SMALL_STATE(3722)] = 154831,
  [SMALL_STATE(3723)] = 154874,
  [SMALL_STATE(3724)] = 154917,
  [SMALL_STATE(3725)] = 154958,
  [SMALL_STATE(3726)] = 155003,
  [SMALL_STATE(3727)] = 155046,
  [SMALL_STATE(3728)] = 155089,
  [SMALL_STATE(3729)] = 155130,
  [SMALL_STATE(3730)] = 155167,
  [SMALL_STATE(3731)] = 155227,
  [SMALL_STATE(3732)] = 155263,
  [SMALL_STATE(3733)] = 155323,
  [SMALL_STATE(3734)] = 155365,
  [SMALL_STATE(3735)] = 155407,
  [SMALL_STATE(3736)] = 155467,
  [SMALL_STATE(3737)] = 155527,
  [SMALL_STATE(3738)] = 155569,
  [SMALL_STATE(3739)] = 155629,
  [SMALL_STATE(3740)] = 155669,
  [SMALL_STATE(3741)] = 155729,
  [SMALL_STATE(3742)] = 155765,
  [SMALL_STATE(3743)] = 155801,
  [SMALL_STATE(3744)] = 155843,
  [SMALL_STATE(3745)] = 155903,
  [SMALL_STATE(3746)] = 155951,
  [SMALL_STATE(3747)] = 155987,
  [SMALL_STATE(3748)] = 156029,
  [SMALL_STATE(3749)] = 156071,
  [SMALL_STATE(3750)] = 156119,
  [SMALL_STATE(3751)] = 156161,
  [SMALL_STATE(3752)] = 156221,
  [SMALL_STATE(3753)] = 156263,
  [SMALL_STATE(3754)] = 156305,
  [SMALL_STATE(3755)] = 156365,
  [SMALL_STATE(3756)] = 156401,
  [SMALL_STATE(3757)] = 156443,
  [SMALL_STATE(3758)] = 156503,
  [SMALL_STATE(3759)] = 156539,
  [SMALL_STATE(3760)] = 156575,
  [SMALL_STATE(3761)] = 156611,
  [SMALL_STATE(3762)] = 156671,
  [SMALL_STATE(3763)] = 156731,
  [SMALL_STATE(3764)] = 156773,
  [SMALL_STATE(3765)] = 156815,
  [SMALL_STATE(3766)] = 156857,
  [SMALL_STATE(3767)] = 156899,
  [SMALL_STATE(3768)] = 156941,
  [SMALL_STATE(3769)] = 157001,
  [SMALL_STATE(3770)] = 157043,
  [SMALL_STATE(3771)] = 157103,
  [SMALL_STATE(3772)] = 157139,
  [SMALL_STATE(3773)] = 157175,
  [SMALL_STATE(3774)] = 157217,
  [SMALL_STATE(3775)] = 157259,
  [SMALL_STATE(3776)] = 157301,
  [SMALL_STATE(3777)] = 157361,
  [SMALL_STATE(3778)] = 157421,
  [SMALL_STATE(3779)] = 157481,
  [SMALL_STATE(3780)] = 157541,
  [SMALL_STATE(3781)] = 157583,
  [SMALL_STATE(3782)] = 157623,
  [SMALL_STATE(3783)] = 157659,
  [SMALL_STATE(3784)] = 157695,
  [SMALL_STATE(3785)] = 157737,
  [SMALL_STATE(3786)] = 157797,
  [SMALL_STATE(3787)] = 157857,
  [SMALL_STATE(3788)] = 157899,
  [SMALL_STATE(3789)] = 157935,
  [SMALL_STATE(3790)] = 157995,
  [SMALL_STATE(3791)] = 158055,
  [SMALL_STATE(3792)] = 158115,
  [SMALL_STATE(3793)] = 158157,
  [SMALL_STATE(3794)] = 158217,
  [SMALL_STATE(3795)] = 158277,
  [SMALL_STATE(3796)] = 158337,
  [SMALL_STATE(3797)] = 158397,
  [SMALL_STATE(3798)] = 158433,
  [SMALL_STATE(3799)] = 158493,
  [SMALL_STATE(3800)] = 158553,
  [SMALL_STATE(3801)] = 158593,
  [SMALL_STATE(3802)] = 158653,
  [SMALL_STATE(3803)] = 158693,
  [SMALL_STATE(3804)] = 158733,
  [SMALL_STATE(3805)] = 158793,
  [SMALL_STATE(3806)] = 158829,
  [SMALL_STATE(3807)] = 158889,
  [SMALL_STATE(3808)] = 158949,
  [SMALL_STATE(3809)] = 159009,
  [SMALL_STATE(3810)] = 159069,
  [SMALL_STATE(3811)] = 159129,
  [SMALL_STATE(3812)] = 159189,
  [SMALL_STATE(3813)] = 159249,
  [SMALL_STATE(3814)] = 159309,
  [SMALL_STATE(3815)] = 159369,
  [SMALL_STATE(3816)] = 159429,
  [SMALL_STATE(3817)] = 159465,
  [SMALL_STATE(3818)] = 159525,
  [SMALL_STATE(3819)] = 159585,
  [SMALL_STATE(3820)] = 159645,
  [SMALL_STATE(3821)] = 159705,
  [SMALL_STATE(3822)] = 159765,
  [SMALL_STATE(3823)] = 159805,
  [SMALL_STATE(3824)] = 159865,
  [SMALL_STATE(3825)] = 159901,
  [SMALL_STATE(3826)] = 159961,
  [SMALL_STATE(3827)] = 160021,
  [SMALL_STATE(3828)] = 160057,
  [SMALL_STATE(3829)] = 160117,
  [SMALL_STATE(3830)] = 160177,
  [SMALL_STATE(3831)] = 160237,
  [SMALL_STATE(3832)] = 160297,
  [SMALL_STATE(3833)] = 160357,
  [SMALL_STATE(3834)] = 160417,
  [SMALL_STATE(3835)] = 160477,
  [SMALL_STATE(3836)] = 160518,
  [SMALL_STATE(3837)] = 160553,
  [SMALL_STATE(3838)] = 160594,
  [SMALL_STATE(3839)] = 160629,
  [SMALL_STATE(3840)] = 160664,
  [SMALL_STATE(3841)] = 160699,
  [SMALL_STATE(3842)] = 160740,
  [SMALL_STATE(3843)] = 160781,
  [SMALL_STATE(3844)] = 160820,
  [SMALL_STATE(3845)] = 160877,
  [SMALL_STATE(3846)] = 160918,
  [SMALL_STATE(3847)] = 160955,
  [SMALL_STATE(3848)] = 160996,
  [SMALL_STATE(3849)] = 161039,
  [SMALL_STATE(3850)] = 161096,
  [SMALL_STATE(3851)] = 161153,
  [SMALL_STATE(3852)] = 161194,
  [SMALL_STATE(3853)] = 161235,
  [SMALL_STATE(3854)] = 161274,
  [SMALL_STATE(3855)] = 161313,
  [SMALL_STATE(3856)] = 161354,
  [SMALL_STATE(3857)] = 161395,
  [SMALL_STATE(3858)] = 161438,
  [SMALL_STATE(3859)] = 161479,
  [SMALL_STATE(3860)] = 161536,
  [SMALL_STATE(3861)] = 161577,
  [SMALL_STATE(3862)] = 161618,
  [SMALL_STATE(3863)] = 161675,
  [SMALL_STATE(3864)] = 161716,
  [SMALL_STATE(3865)] = 161751,
  [SMALL_STATE(3866)] = 161792,
  [SMALL_STATE(3867)] = 161849,
  [SMALL_STATE(3868)] = 161890,
  [SMALL_STATE(3869)] = 161929,
  [SMALL_STATE(3870)] = 161968,
  [SMALL_STATE(3871)] = 162009,
  [SMALL_STATE(3872)] = 162050,
  [SMALL_STATE(3873)] = 162091,
  [SMALL_STATE(3874)] = 162148,
  [SMALL_STATE(3875)] = 162189,
  [SMALL_STATE(3876)] = 162246,
  [SMALL_STATE(3877)] = 162285,
  [SMALL_STATE(3878)] = 162326,
  [SMALL_STATE(3879)] = 162365,
  [SMALL_STATE(3880)] = 162422,
  [SMALL_STATE(3881)] = 162465,
  [SMALL_STATE(3882)] = 162522,
  [SMALL_STATE(3883)] = 162561,
  [SMALL_STATE(3884)] = 162602,
  [SMALL_STATE(3885)] = 162659,
  [SMALL_STATE(3886)] = 162702,
  [SMALL_STATE(3887)] = 162739,
  [SMALL_STATE(3888)] = 162796,
  [SMALL_STATE(3889)] = 162837,
  [SMALL_STATE(3890)] = 162894,
  [SMALL_STATE(3891)] = 162933,
  [SMALL_STATE(3892)] = 162990,
  [SMALL_STATE(3893)] = 163044,
  [SMALL_STATE(3894)] = 163098,
  [SMALL_STATE(3895)] = 163152,
  [SMALL_STATE(3896)] = 163206,
  [SMALL_STATE(3897)] = 163260,
  [SMALL_STATE(3898)] = 163314,
  [SMALL_STATE(3899)] = 163368,
  [SMALL_STATE(3900)] = 163422,
  [SMALL_STATE(3901)] = 163476,
  [SMALL_STATE(3902)] = 163530,
  [SMALL_STATE(3903)] = 163584,
  [SMALL_STATE(3904)] = 163638,
  [SMALL_STATE(3905)] = 163692,
  [SMALL_STATE(3906)] = 163746,
  [SMALL_STATE(3907)] = 163800,
  [SMALL_STATE(3908)] = 163854,
  [SMALL_STATE(3909)] = 163908,
  [SMALL_STATE(3910)] = 163962,
  [SMALL_STATE(3911)] = 164016,
  [SMALL_STATE(3912)] = 164070,
  [SMALL_STATE(3913)] = 164124,
  [SMALL_STATE(3914)] = 164178,
  [SMALL_STATE(3915)] = 164232,
  [SMALL_STATE(3916)] = 164266,
  [SMALL_STATE(3917)] = 164304,
  [SMALL_STATE(3918)] = 164358,
  [SMALL_STATE(3919)] = 164412,
  [SMALL_STATE(3920)] = 164448,
  [SMALL_STATE(3921)] = 164486,
  [SMALL_STATE(3922)] = 164522,
  [SMALL_STATE(3923)] = 164560,
  [SMALL_STATE(3924)] = 164614,
  [SMALL_STATE(3925)] = 164668,
  [SMALL_STATE(3926)] = 164724,
  [SMALL_STATE(3927)] = 164778,
  [SMALL_STATE(3928)] = 164816,
  [SMALL_STATE(3929)] = 164870,
  [SMALL_STATE(3930)] = 164924,
  [SMALL_STATE(3931)] = 164978,
  [SMALL_STATE(3932)] = 165032,
  [SMALL_STATE(3933)] = 165086,
  [SMALL_STATE(3934)] = 165140,
  [SMALL_STATE(3935)] = 165194,
  [SMALL_STATE(3936)] = 165248,
  [SMALL_STATE(3937)] = 165302,
  [SMALL_STATE(3938)] = 165356,
  [SMALL_STATE(3939)] = 165390,
  [SMALL_STATE(3940)] = 165444,
  [SMALL_STATE(3941)] = 165478,
  [SMALL_STATE(3942)] = 165512,
  [SMALL_STATE(3943)] = 165568,
  [SMALL_STATE(3944)] = 165608,
  [SMALL_STATE(3945)] = 165662,
  [SMALL_STATE(3946)] = 165716,
  [SMALL_STATE(3947)] = 165754,
  [SMALL_STATE(3948)] = 165810,
  [SMALL_STATE(3949)] = 165846,
  [SMALL_STATE(3950)] = 165900,
  [SMALL_STATE(3951)] = 165954,
  [SMALL_STATE(3952)] = 166008,
  [SMALL_STATE(3953)] = 166046,
  [SMALL_STATE(3954)] = 166100,
  [SMALL_STATE(3955)] = 166156,
  [SMALL_STATE(3956)] = 166210,
  [SMALL_STATE(3957)] = 166266,
  [SMALL_STATE(3958)] = 166320,
  [SMALL_STATE(3959)] = 166374,
  [SMALL_STATE(3960)] = 166428,
  [SMALL_STATE(3961)] = 166482,
  [SMALL_STATE(3962)] = 166536,
  [SMALL_STATE(3963)] = 166590,
  [SMALL_STATE(3964)] = 166644,
  [SMALL_STATE(3965)] = 166698,
  [SMALL_STATE(3966)] = 166752,
  [SMALL_STATE(3967)] = 166806,
  [SMALL_STATE(3968)] = 166860,
  [SMALL_STATE(3969)] = 166914,
  [SMALL_STATE(3970)] = 166970,
  [SMALL_STATE(3971)] = 167024,
  [SMALL_STATE(3972)] = 167078,
  [SMALL_STATE(3973)] = 167132,
  [SMALL_STATE(3974)] = 167186,
  [SMALL_STATE(3975)] = 167240,
  [SMALL_STATE(3976)] = 167294,
  [SMALL_STATE(3977)] = 167348,
  [SMALL_STATE(3978)] = 167402,
  [SMALL_STATE(3979)] = 167456,
  [SMALL_STATE(3980)] = 167510,
  [SMALL_STATE(3981)] = 167564,
  [SMALL_STATE(3982)] = 167618,
  [SMALL_STATE(3983)] = 167672,
  [SMALL_STATE(3984)] = 167726,
  [SMALL_STATE(3985)] = 167782,
  [SMALL_STATE(3986)] = 167836,
  [SMALL_STATE(3987)] = 167890,
  [SMALL_STATE(3988)] = 167944,
  [SMALL_STATE(3989)] = 167980,
  [SMALL_STATE(3990)] = 168036,
  [SMALL_STATE(3991)] = 168090,
  [SMALL_STATE(3992)] = 168146,
  [SMALL_STATE(3993)] = 168182,
  [SMALL_STATE(3994)] = 168236,
  [SMALL_STATE(3995)] = 168290,
  [SMALL_STATE(3996)] = 168344,
  [SMALL_STATE(3997)] = 168398,
  [SMALL_STATE(3998)] = 168452,
  [SMALL_STATE(3999)] = 168506,
  [SMALL_STATE(4000)] = 168560,
  [SMALL_STATE(4001)] = 168614,
  [SMALL_STATE(4002)] = 168668,
  [SMALL_STATE(4003)] = 168722,
  [SMALL_STATE(4004)] = 168776,
  [SMALL_STATE(4005)] = 168830,
  [SMALL_STATE(4006)] = 168868,
  [SMALL_STATE(4007)] = 168922,
  [SMALL_STATE(4008)] = 168976,
  [SMALL_STATE(4009)] = 169030,
  [SMALL_STATE(4010)] = 169068,
  [SMALL_STATE(4011)] = 169122,
  [SMALL_STATE(4012)] = 169178,
  [SMALL_STATE(4013)] = 169232,
  [SMALL_STATE(4014)] = 169286,
  [SMALL_STATE(4015)] = 169340,
  [SMALL_STATE(4016)] = 169394,
  [SMALL_STATE(4017)] = 169448,
  [SMALL_STATE(4018)] = 169502,
  [SMALL_STATE(4019)] = 169556,
  [SMALL_STATE(4020)] = 169610,
  [SMALL_STATE(4021)] = 169664,
  [SMALL_STATE(4022)] = 169718,
  [SMALL_STATE(4023)] = 169772,
  [SMALL_STATE(4024)] = 169824,
  [SMALL_STATE(4025)] = 169878,
  [SMALL_STATE(4026)] = 169932,
  [SMALL_STATE(4027)] = 169986,
  [SMALL_STATE(4028)] = 170042,
  [SMALL_STATE(4029)] = 170096,
  [SMALL_STATE(4030)] = 170150,
  [SMALL_STATE(4031)] = 170204,
  [SMALL_STATE(4032)] = 170258,
  [SMALL_STATE(4033)] = 170312,
  [SMALL_STATE(4034)] = 170366,
  [SMALL_STATE(4035)] = 170420,
  [SMALL_STATE(4036)] = 170474,
  [SMALL_STATE(4037)] = 170528,
  [SMALL_STATE(4038)] = 170582,
  [SMALL_STATE(4039)] = 170620,
  [SMALL_STATE(4040)] = 170658,
  [SMALL_STATE(4041)] = 170712,
  [SMALL_STATE(4042)] = 170766,
  [SMALL_STATE(4043)] = 170820,
  [SMALL_STATE(4044)] = 170876,
  [SMALL_STATE(4045)] = 170914,
  [SMALL_STATE(4046)] = 170952,
  [SMALL_STATE(4047)] = 170988,
  [SMALL_STATE(4048)] = 171026,
  [SMALL_STATE(4049)] = 171064,
  [SMALL_STATE(4050)] = 171118,
  [SMALL_STATE(4051)] = 171156,
  [SMALL_STATE(4052)] = 171194,
  [SMALL_STATE(4053)] = 171248,
  [SMALL_STATE(4054)] = 171302,
  [SMALL_STATE(4055)] = 171356,
  [SMALL_STATE(4056)] = 171410,
  [SMALL_STATE(4057)] = 171464,
  [SMALL_STATE(4058)] = 171518,
  [SMALL_STATE(4059)] = 171571,
  [SMALL_STATE(4060)] = 171610,
  [SMALL_STATE(4061)] = 171659,
  [SMALL_STATE(4062)] = 171696,
  [SMALL_STATE(4063)] = 171731,
  [SMALL_STATE(4064)] = 171768,
  [SMALL_STATE(4065)] = 171805,
  [SMALL_STATE(4066)] = 171856,
  [SMALL_STATE(4067)] = 171907,
  [SMALL_STATE(4068)] = 171940,
  [SMALL_STATE(4069)] = 171989,
  [SMALL_STATE(4070)] = 172038,
  [SMALL_STATE(4071)] = 172075,
  [SMALL_STATE(4072)] = 172108,
  [SMALL_STATE(4073)] = 172161,
  [SMALL_STATE(4074)] = 172222,
  [SMALL_STATE(4075)] = 172255,
  [SMALL_STATE(4076)] = 172316,
  [SMALL_STATE(4077)] = 172349,
  [SMALL_STATE(4078)] = 172384,
  [SMALL_STATE(4079)] = 172421,
  [SMALL_STATE(4080)] = 172482,
  [SMALL_STATE(4081)] = 172519,
  [SMALL_STATE(4082)] = 172569,
  [SMALL_STATE(4083)] = 172603,
  [SMALL_STATE(4084)] = 172635,
  [SMALL_STATE(4085)] = 172669,
  [SMALL_STATE(4086)] = 172703,
  [SMALL_STATE(4087)] = 172735,
  [SMALL_STATE(4088)] = 172785,
  [SMALL_STATE(4089)] = 172833,
  [SMALL_STATE(4090)] = 172865,
  [SMALL_STATE(4091)] = 172915,
  [SMALL_STATE(4092)] = 172949,
  [SMALL_STATE(4093)] = 172997,
  [SMALL_STATE(4094)] = 173045,
  [SMALL_STATE(4095)] = 173093,
  [SMALL_STATE(4096)] = 173125,
  [SMALL_STATE(4097)] = 173157,
  [SMALL_STATE(4098)] = 173207,
  [SMALL_STATE(4099)] = 173255,
  [SMALL_STATE(4100)] = 173289,
  [SMALL_STATE(4101)] = 173321,
  [SMALL_STATE(4102)] = 173369,
  [SMALL_STATE(4103)] = 173417,
  [SMALL_STATE(4104)] = 173449,
  [SMALL_STATE(4105)] = 173497,
  [SMALL_STATE(4106)] = 173545,
  [SMALL_STATE(4107)] = 173593,
  [SMALL_STATE(4108)] = 173627,
  [SMALL_STATE(4109)] = 173675,
  [SMALL_STATE(4110)] = 173707,
  [SMALL_STATE(4111)] = 173755,
  [SMALL_STATE(4112)] = 173803,
  [SMALL_STATE(4113)] = 173851,
  [SMALL_STATE(4114)] = 173899,
  [SMALL_STATE(4115)] = 173931,
  [SMALL_STATE(4116)] = 173963,
  [SMALL_STATE(4117)] = 173995,
  [SMALL_STATE(4118)] = 174029,
  [SMALL_STATE(4119)] = 174061,
  [SMALL_STATE(4120)] = 174109,
  [SMALL_STATE(4121)] = 174143,
  [SMALL_STATE(4122)] = 174175,
  [SMALL_STATE(4123)] = 174223,
  [SMALL_STATE(4124)] = 174255,
  [SMALL_STATE(4125)] = 174289,
  [SMALL_STATE(4126)] = 174339,
  [SMALL_STATE(4127)] = 174387,
  [SMALL_STATE(4128)] = 174419,
  [SMALL_STATE(4129)] = 174467,
  [SMALL_STATE(4130)] = 174515,
  [SMALL_STATE(4131)] = 174547,
  [SMALL_STATE(4132)] = 174579,
  [SMALL_STATE(4133)] = 174627,
  [SMALL_STATE(4134)] = 174677,
  [SMALL_STATE(4135)] = 174725,
  [SMALL_STATE(4136)] = 174773,
  [SMALL_STATE(4137)] = 174821,
  [SMALL_STATE(4138)] = 174869,
  [SMALL_STATE(4139)] = 174917,
  [SMALL_STATE(4140)] = 174949,
  [SMALL_STATE(4141)] = 174997,
  [SMALL_STATE(4142)] = 175029,
  [SMALL_STATE(4143)] = 175063,
  [SMALL_STATE(4144)] = 175111,
  [SMALL_STATE(4145)] = 175143,
  [SMALL_STATE(4146)] = 175191,
  [SMALL_STATE(4147)] = 175239,
  [SMALL_STATE(4148)] = 175287,
  [SMALL_STATE(4149)] = 175321,
  [SMALL_STATE(4150)] = 175369,
  [SMALL_STATE(4151)] = 175417,
  [SMALL_STATE(4152)] = 175465,
  [SMALL_STATE(4153)] = 175499,
  [SMALL_STATE(4154)] = 175547,
  [SMALL_STATE(4155)] = 175595,
  [SMALL_STATE(4156)] = 175627,
  [SMALL_STATE(4157)] = 175659,
  [SMALL_STATE(4158)] = 175707,
  [SMALL_STATE(4159)] = 175755,
  [SMALL_STATE(4160)] = 175803,
  [SMALL_STATE(4161)] = 175837,
  [SMALL_STATE(4162)] = 175885,
  [SMALL_STATE(4163)] = 175919,
  [SMALL_STATE(4164)] = 175953,
  [SMALL_STATE(4165)] = 175985,
  [SMALL_STATE(4166)] = 176019,
  [SMALL_STATE(4167)] = 176060,
  [SMALL_STATE(4168)] = 176101,
  [SMALL_STATE(4169)] = 176156,
  [SMALL_STATE(4170)] = 176197,
  [SMALL_STATE(4171)] = 176238,
  [SMALL_STATE(4172)] = 176269,
  [SMALL_STATE(4173)] = 176308,
  [SMALL_STATE(4174)] = 176349,
  [SMALL_STATE(4175)] = 176390,
  [SMALL_STATE(4176)] = 176431,
  [SMALL_STATE(4177)] = 176472,
  [SMALL_STATE(4178)] = 176503,
  [SMALL_STATE(4179)] = 176544,
  [SMALL_STATE(4180)] = 176585,
  [SMALL_STATE(4181)] = 176626,
  [SMALL_STATE(4182)] = 176681,
  [SMALL_STATE(4183)] = 176722,
  [SMALL_STATE(4184)] = 176777,
  [SMALL_STATE(4185)] = 176818,
  [SMALL_STATE(4186)] = 176859,
  [SMALL_STATE(4187)] = 176900,
  [SMALL_STATE(4188)] = 176941,
  [SMALL_STATE(4189)] = 176982,
  [SMALL_STATE(4190)] = 177023,
  [SMALL_STATE(4191)] = 177054,
  [SMALL_STATE(4192)] = 177093,
  [SMALL_STATE(4193)] = 177145,
  [SMALL_STATE(4194)] = 177197,
  [SMALL_STATE(4195)] = 177243,
  [SMALL_STATE(4196)] = 177289,
  [SMALL_STATE(4197)] = 177333,
  [SMALL_STATE(4198)] = 177363,
  [SMALL_STATE(4199)] = 177409,
  [SMALL_STATE(4200)] = 177461,
  [SMALL_STATE(4201)] = 177491,
  [SMALL_STATE(4202)] = 177519,
  [SMALL_STATE(4203)] = 177565,
  [SMALL_STATE(4204)] = 177611,
  [SMALL_STATE(4205)] = 177639,
  [SMALL_STATE(4206)] = 177675,
  [SMALL_STATE(4207)] = 177727,
  [SMALL_STATE(4208)] = 177773,
  [SMALL_STATE(4209)] = 177819,
  [SMALL_STATE(4210)] = 177871,
  [SMALL_STATE(4211)] = 177923,
  [SMALL_STATE(4212)] = 177969,
  [SMALL_STATE(4213)] = 178015,
  [SMALL_STATE(4214)] = 178061,
  [SMALL_STATE(4215)] = 178105,
  [SMALL_STATE(4216)] = 178133,
  [SMALL_STATE(4217)] = 178175,
  [SMALL_STATE(4218)] = 178221,
  [SMALL_STATE(4219)] = 178265,
  [SMALL_STATE(4220)] = 178317,
  [SMALL_STATE(4221)] = 178347,
  [SMALL_STATE(4222)] = 178399,
  [SMALL_STATE(4223)] = 178445,
  [SMALL_STATE(4224)] = 178491,
  [SMALL_STATE(4225)] = 178537,
  [SMALL_STATE(4226)] = 178589,
  [SMALL_STATE(4227)] = 178633,
  [SMALL_STATE(4228)] = 178677,
  [SMALL_STATE(4229)] = 178715,
  [SMALL_STATE(4230)] = 178743,
  [SMALL_STATE(4231)] = 178780,
  [SMALL_STATE(4232)] = 178817,
  [SMALL_STATE(4233)] = 178846,
  [SMALL_STATE(4234)] = 178893,
  [SMALL_STATE(4235)] = 178942,
  [SMALL_STATE(4236)] = 178989,
  [SMALL_STATE(4237)] = 179032,
  [SMALL_STATE(4238)] = 179075,
  [SMALL_STATE(4239)] = 179118,
  [SMALL_STATE(4240)] = 179147,
  [SMALL_STATE(4241)] = 179196,
  [SMALL_STATE(4242)] = 179243,
  [SMALL_STATE(4243)] = 179290,
  [SMALL_STATE(4244)] = 179337,
  [SMALL_STATE(4245)] = 179368,
  [SMALL_STATE(4246)] = 179397,
  [SMALL_STATE(4247)] = 179440,
  [SMALL_STATE(4248)] = 179473,
  [SMALL_STATE(4249)] = 179502,
  [SMALL_STATE(4250)] = 179545,
  [SMALL_STATE(4251)] = 179588,
  [SMALL_STATE(4252)] = 179637,
  [SMALL_STATE(4253)] = 179684,
  [SMALL_STATE(4254)] = 179712,
  [SMALL_STATE(4255)] = 179738,
  [SMALL_STATE(4256)] = 179764,
  [SMALL_STATE(4257)] = 179806,
  [SMALL_STATE(4258)] = 179848,
  [SMALL_STATE(4259)] = 179890,
  [SMALL_STATE(4260)] = 179932,
  [SMALL_STATE(4261)] = 179974,
  [SMALL_STATE(4262)] = 180012,
  [SMALL_STATE(4263)] = 180054,
  [SMALL_STATE(4264)] = 180096,
  [SMALL_STATE(4265)] = 180126,
  [SMALL_STATE(4266)] = 180152,
  [SMALL_STATE(4267)] = 180178,
  [SMALL_STATE(4268)] = 180222,
  [SMALL_STATE(4269)] = 180248,
  [SMALL_STATE(4270)] = 180292,
  [SMALL_STATE(4271)] = 180336,
  [SMALL_STATE(4272)] = 180370,
  [SMALL_STATE(4273)] = 180404,
  [SMALL_STATE(4274)] = 180448,
  [SMALL_STATE(4275)] = 180490,
  [SMALL_STATE(4276)] = 180524,
  [SMALL_STATE(4277)] = 180566,
  [SMALL_STATE(4278)] = 180600,
  [SMALL_STATE(4279)] = 180644,
  [SMALL_STATE(4280)] = 180672,
  [SMALL_STATE(4281)] = 180698,
  [SMALL_STATE(4282)] = 180726,
  [SMALL_STATE(4283)] = 180766,
  [SMALL_STATE(4284)] = 180792,
  [SMALL_STATE(4285)] = 180834,
  [SMALL_STATE(4286)] = 180862,
  [SMALL_STATE(4287)] = 180896,
  [SMALL_STATE(4288)] = 180930,
  [SMALL_STATE(4289)] = 180964,
  [SMALL_STATE(4290)] = 181008,
  [SMALL_STATE(4291)] = 181050,
  [SMALL_STATE(4292)] = 181092,
  [SMALL_STATE(4293)] = 181134,
  [SMALL_STATE(4294)] = 181176,
  [SMALL_STATE(4295)] = 181210,
  [SMALL_STATE(4296)] = 181240,
  [SMALL_STATE(4297)] = 181274,
  [SMALL_STATE(4298)] = 181310,
  [SMALL_STATE(4299)] = 181344,
  [SMALL_STATE(4300)] = 181386,
  [SMALL_STATE(4301)] = 181420,
  [SMALL_STATE(4302)] = 181462,
  [SMALL_STATE(4303)] = 181504,
  [SMALL_STATE(4304)] = 181546,
  [SMALL_STATE(4305)] = 181574,
  [SMALL_STATE(4306)] = 181610,
  [SMALL_STATE(4307)] = 181644,
  [SMALL_STATE(4308)] = 181686,
  [SMALL_STATE(4309)] = 181728,
  [SMALL_STATE(4310)] = 181766,
  [SMALL_STATE(4311)] = 181808,
  [SMALL_STATE(4312)] = 181834,
  [SMALL_STATE(4313)] = 181876,
  [SMALL_STATE(4314)] = 181905,
  [SMALL_STATE(4315)] = 181936,
  [SMALL_STATE(4316)] = 181963,
  [SMALL_STATE(4317)] = 181994,
  [SMALL_STATE(4318)] = 182021,
  [SMALL_STATE(4319)] = 182052,
  [SMALL_STATE(4320)] = 182079,
  [SMALL_STATE(4321)] = 182110,
  [SMALL_STATE(4322)] = 182137,
  [SMALL_STATE(4323)] = 182164,
  [SMALL_STATE(4324)] = 182191,
  [SMALL_STATE(4325)] = 182218,
  [SMALL_STATE(4326)] = 182257,
  [SMALL_STATE(4327)] = 182284,
  [SMALL_STATE(4328)] = 182311,
  [SMALL_STATE(4329)] = 182350,
  [SMALL_STATE(4330)] = 182389,
  [SMALL_STATE(4331)] = 182416,
  [SMALL_STATE(4332)] = 182455,
  [SMALL_STATE(4333)] = 182494,
  [SMALL_STATE(4334)] = 182521,
  [SMALL_STATE(4335)] = 182548,
  [SMALL_STATE(4336)] = 182587,
  [SMALL_STATE(4337)] = 182614,
  [SMALL_STATE(4338)] = 182645,
  [SMALL_STATE(4339)] = 182676,
  [SMALL_STATE(4340)] = 182707,
  [SMALL_STATE(4341)] = 182734,
  [SMALL_STATE(4342)] = 182773,
  [SMALL_STATE(4343)] = 182804,
  [SMALL_STATE(4344)] = 182835,
  [SMALL_STATE(4345)] = 182866,
  [SMALL_STATE(4346)] = 182897,
  [SMALL_STATE(4347)] = 182928,
  [SMALL_STATE(4348)] = 182959,
  [SMALL_STATE(4349)] = 182990,
  [SMALL_STATE(4350)] = 183029,
  [SMALL_STATE(4351)] = 183060,
  [SMALL_STATE(4352)] = 183101,
  [SMALL_STATE(4353)] = 183132,
  [SMALL_STATE(4354)] = 183163,
  [SMALL_STATE(4355)] = 183194,
  [SMALL_STATE(4356)] = 183225,
  [SMALL_STATE(4357)] = 183256,
  [SMALL_STATE(4358)] = 183285,
  [SMALL_STATE(4359)] = 183312,
  [SMALL_STATE(4360)] = 183339,
  [SMALL_STATE(4361)] = 183378,
  [SMALL_STATE(4362)] = 183405,
  [SMALL_STATE(4363)] = 183444,
  [SMALL_STATE(4364)] = 183483,
  [SMALL_STATE(4365)] = 183510,
  [SMALL_STATE(4366)] = 183549,
  [SMALL_STATE(4367)] = 183586,
  [SMALL_STATE(4368)] = 183617,
  [SMALL_STATE(4369)] = 183648,
  [SMALL_STATE(4370)] = 183679,
  [SMALL_STATE(4371)] = 183710,
  [SMALL_STATE(4372)] = 183741,
  [SMALL_STATE(4373)] = 183772,
  [SMALL_STATE(4374)] = 183803,
  [SMALL_STATE(4375)] = 183834,
  [SMALL_STATE(4376)] = 183865,
  [SMALL_STATE(4377)] = 183896,
  [SMALL_STATE(4378)] = 183927,
  [SMALL_STATE(4379)] = 183958,
  [SMALL_STATE(4380)] = 183985,
  [SMALL_STATE(4381)] = 184012,
  [SMALL_STATE(4382)] = 184043,
  [SMALL_STATE(4383)] = 184074,
  [SMALL_STATE(4384)] = 184105,
  [SMALL_STATE(4385)] = 184136,
  [SMALL_STATE(4386)] = 184163,
  [SMALL_STATE(4387)] = 184190,
  [SMALL_STATE(4388)] = 184217,
  [SMALL_STATE(4389)] = 184244,
  [SMALL_STATE(4390)] = 184275,
  [SMALL_STATE(4391)] = 184306,
  [SMALL_STATE(4392)] = 184337,
  [SMALL_STATE(4393)] = 184368,
  [SMALL_STATE(4394)] = 184399,
  [SMALL_STATE(4395)] = 184430,
  [SMALL_STATE(4396)] = 184461,
  [SMALL_STATE(4397)] = 184492,
  [SMALL_STATE(4398)] = 184523,
  [SMALL_STATE(4399)] = 184550,
  [SMALL_STATE(4400)] = 184581,
  [SMALL_STATE(4401)] = 184612,
  [SMALL_STATE(4402)] = 184651,
  [SMALL_STATE(4403)] = 184690,
  [SMALL_STATE(4404)] = 184721,
  [SMALL_STATE(4405)] = 184752,
  [SMALL_STATE(4406)] = 184783,
  [SMALL_STATE(4407)] = 184814,
  [SMALL_STATE(4408)] = 184845,
  [SMALL_STATE(4409)] = 184876,
  [SMALL_STATE(4410)] = 184903,
  [SMALL_STATE(4411)] = 184930,
  [SMALL_STATE(4412)] = 184961,
  [SMALL_STATE(4413)] = 184992,
  [SMALL_STATE(4414)] = 185031,
  [SMALL_STATE(4415)] = 185058,
  [SMALL_STATE(4416)] = 185101,
  [SMALL_STATE(4417)] = 185140,
  [SMALL_STATE(4418)] = 185179,
  [SMALL_STATE(4419)] = 185218,
  [SMALL_STATE(4420)] = 185249,
  [SMALL_STATE(4421)] = 185280,
  [SMALL_STATE(4422)] = 185319,
  [SMALL_STATE(4423)] = 185350,
  [SMALL_STATE(4424)] = 185381,
  [SMALL_STATE(4425)] = 185420,
  [SMALL_STATE(4426)] = 185451,
  [SMALL_STATE(4427)] = 185482,
  [SMALL_STATE(4428)] = 185525,
  [SMALL_STATE(4429)] = 185556,
  [SMALL_STATE(4430)] = 185587,
  [SMALL_STATE(4431)] = 185614,
  [SMALL_STATE(4432)] = 185641,
  [SMALL_STATE(4433)] = 185680,
  [SMALL_STATE(4434)] = 185719,
  [SMALL_STATE(4435)] = 185746,
  [SMALL_STATE(4436)] = 185773,
  [SMALL_STATE(4437)] = 185800,
  [SMALL_STATE(4438)] = 185827,
  [SMALL_STATE(4439)] = 185854,
  [SMALL_STATE(4440)] = 185881,
  [SMALL_STATE(4441)] = 185908,
  [SMALL_STATE(4442)] = 185935,
  [SMALL_STATE(4443)] = 185967,
  [SMALL_STATE(4444)] = 185995,
  [SMALL_STATE(4445)] = 186027,
  [SMALL_STATE(4446)] = 186053,
  [SMALL_STATE(4447)] = 186083,
  [SMALL_STATE(4448)] = 186113,
  [SMALL_STATE(4449)] = 186143,
  [SMALL_STATE(4450)] = 186169,
  [SMALL_STATE(4451)] = 186199,
  [SMALL_STATE(4452)] = 186225,
  [SMALL_STATE(4453)] = 186253,
  [SMALL_STATE(4454)] = 186279,
  [SMALL_STATE(4455)] = 186305,
  [SMALL_STATE(4456)] = 186329,
  [SMALL_STATE(4457)] = 186355,
  [SMALL_STATE(4458)] = 186381,
  [SMALL_STATE(4459)] = 186407,
  [SMALL_STATE(4460)] = 186433,
  [SMALL_STATE(4461)] = 186463,
  [SMALL_STATE(4462)] = 186493,
  [SMALL_STATE(4463)] = 186523,
  [SMALL_STATE(4464)] = 186553,
  [SMALL_STATE(4465)] = 186579,
  [SMALL_STATE(4466)] = 186603,
  [SMALL_STATE(4467)] = 186643,
  [SMALL_STATE(4468)] = 186669,
  [SMALL_STATE(4469)] = 186709,
  [SMALL_STATE(4470)] = 186741,
  [SMALL_STATE(4471)] = 186767,
  [SMALL_STATE(4472)] = 186793,
  [SMALL_STATE(4473)] = 186819,
  [SMALL_STATE(4474)] = 186849,
  [SMALL_STATE(4475)] = 186879,
  [SMALL_STATE(4476)] = 186905,
  [SMALL_STATE(4477)] = 186931,
  [SMALL_STATE(4478)] = 186957,
  [SMALL_STATE(4479)] = 186983,
  [SMALL_STATE(4480)] = 187009,
  [SMALL_STATE(4481)] = 187049,
  [SMALL_STATE(4482)] = 187073,
  [SMALL_STATE(4483)] = 187099,
  [SMALL_STATE(4484)] = 187125,
  [SMALL_STATE(4485)] = 187167,
  [SMALL_STATE(4486)] = 187191,
  [SMALL_STATE(4487)] = 187217,
  [SMALL_STATE(4488)] = 187257,
  [SMALL_STATE(4489)] = 187283,
  [SMALL_STATE(4490)] = 187309,
  [SMALL_STATE(4491)] = 187335,
  [SMALL_STATE(4492)] = 187375,
  [SMALL_STATE(4493)] = 187401,
  [SMALL_STATE(4494)] = 187443,
  [SMALL_STATE(4495)] = 187483,
  [SMALL_STATE(4496)] = 187509,
  [SMALL_STATE(4497)] = 187535,
  [SMALL_STATE(4498)] = 187561,
  [SMALL_STATE(4499)] = 187598,
  [SMALL_STATE(4500)] = 187637,
  [SMALL_STATE(4501)] = 187660,
  [SMALL_STATE(4502)] = 187689,
  [SMALL_STATE(4503)] = 187726,
  [SMALL_STATE(4504)] = 187755,
  [SMALL_STATE(4505)] = 187782,
  [SMALL_STATE(4506)] = 187823,
  [SMALL_STATE(4507)] = 187854,
  [SMALL_STATE(4508)] = 187891,
  [SMALL_STATE(4509)] = 187930,
  [SMALL_STATE(4510)] = 187953,
  [SMALL_STATE(4511)] = 187980,
  [SMALL_STATE(4512)] = 188017,
  [SMALL_STATE(4513)] = 188046,
  [SMALL_STATE(4514)] = 188083,
  [SMALL_STATE(4515)] = 188106,
  [SMALL_STATE(4516)] = 188143,
  [SMALL_STATE(4517)] = 188166,
  [SMALL_STATE(4518)] = 188195,
  [SMALL_STATE(4519)] = 188220,
  [SMALL_STATE(4520)] = 188243,
  [SMALL_STATE(4521)] = 188266,
  [SMALL_STATE(4522)] = 188291,
  [SMALL_STATE(4523)] = 188318,
  [SMALL_STATE(4524)] = 188341,
  [SMALL_STATE(4525)] = 188378,
  [SMALL_STATE(4526)] = 188415,
  [SMALL_STATE(4527)] = 188448,
  [SMALL_STATE(4528)] = 188481,
  [SMALL_STATE(4529)] = 188504,
  [SMALL_STATE(4530)] = 188527,
  [SMALL_STATE(4531)] = 188564,
  [SMALL_STATE(4532)] = 188589,
  [SMALL_STATE(4533)] = 188626,
  [SMALL_STATE(4534)] = 188649,
  [SMALL_STATE(4535)] = 188686,
  [SMALL_STATE(4536)] = 188709,
  [SMALL_STATE(4537)] = 188732,
  [SMALL_STATE(4538)] = 188757,
  [SMALL_STATE(4539)] = 188794,
  [SMALL_STATE(4540)] = 188821,
  [SMALL_STATE(4541)] = 188858,
  [SMALL_STATE(4542)] = 188895,
  [SMALL_STATE(4543)] = 188932,
  [SMALL_STATE(4544)] = 188973,
  [SMALL_STATE(4545)] = 189001,
  [SMALL_STATE(4546)] = 189039,
  [SMALL_STATE(4547)] = 189077,
  [SMALL_STATE(4548)] = 189107,
  [SMALL_STATE(4549)] = 189137,
  [SMALL_STATE(4550)] = 189161,
  [SMALL_STATE(4551)] = 189199,
  [SMALL_STATE(4552)] = 189235,
  [SMALL_STATE(4553)] = 189259,
  [SMALL_STATE(4554)] = 189283,
  [SMALL_STATE(4555)] = 189319,
  [SMALL_STATE(4556)] = 189349,
  [SMALL_STATE(4557)] = 189387,
  [SMALL_STATE(4558)] = 189411,
  [SMALL_STATE(4559)] = 189435,
  [SMALL_STATE(4560)] = 189459,
  [SMALL_STATE(4561)] = 189483,
  [SMALL_STATE(4562)] = 189507,
  [SMALL_STATE(4563)] = 189545,
  [SMALL_STATE(4564)] = 189575,
  [SMALL_STATE(4565)] = 189599,
  [SMALL_STATE(4566)] = 189629,
  [SMALL_STATE(4567)] = 189659,
  [SMALL_STATE(4568)] = 189689,
  [SMALL_STATE(4569)] = 189719,
  [SMALL_STATE(4570)] = 189749,
  [SMALL_STATE(4571)] = 189779,
  [SMALL_STATE(4572)] = 189809,
  [SMALL_STATE(4573)] = 189839,
  [SMALL_STATE(4574)] = 189867,
  [SMALL_STATE(4575)] = 189889,
  [SMALL_STATE(4576)] = 189927,
  [SMALL_STATE(4577)] = 189949,
  [SMALL_STATE(4578)] = 189987,
  [SMALL_STATE(4579)] = 190023,
  [SMALL_STATE(4580)] = 190047,
  [SMALL_STATE(4581)] = 190069,
  [SMALL_STATE(4582)] = 190093,
  [SMALL_STATE(4583)] = 190117,
  [SMALL_STATE(4584)] = 190141,
  [SMALL_STATE(4585)] = 190171,
  [SMALL_STATE(4586)] = 190195,
  [SMALL_STATE(4587)] = 190219,
  [SMALL_STATE(4588)] = 190249,
  [SMALL_STATE(4589)] = 190275,
  [SMALL_STATE(4590)] = 190299,
  [SMALL_STATE(4591)] = 190323,
  [SMALL_STATE(4592)] = 190347,
  [SMALL_STATE(4593)] = 190371,
  [SMALL_STATE(4594)] = 190395,
  [SMALL_STATE(4595)] = 190433,
  [SMALL_STATE(4596)] = 190471,
  [SMALL_STATE(4597)] = 190501,
  [SMALL_STATE(4598)] = 190531,
  [SMALL_STATE(4599)] = 190555,
  [SMALL_STATE(4600)] = 190583,
  [SMALL_STATE(4601)] = 190613,
  [SMALL_STATE(4602)] = 190651,
  [SMALL_STATE(4603)] = 190675,
  [SMALL_STATE(4604)] = 190699,
  [SMALL_STATE(4605)] = 190723,
  [SMALL_STATE(4606)] = 190747,
  [SMALL_STATE(4607)] = 190771,
  [SMALL_STATE(4608)] = 190801,
  [SMALL_STATE(4609)] = 190825,
  [SMALL_STATE(4610)] = 190861,
  [SMALL_STATE(4611)] = 190885,
  [SMALL_STATE(4612)] = 190909,
  [SMALL_STATE(4613)] = 190933,
  [SMALL_STATE(4614)] = 190955,
  [SMALL_STATE(4615)] = 190977,
  [SMALL_STATE(4616)] = 191001,
  [SMALL_STATE(4617)] = 191025,
  [SMALL_STATE(4618)] = 191049,
  [SMALL_STATE(4619)] = 191073,
  [SMALL_STATE(4620)] = 191111,
  [SMALL_STATE(4621)] = 191142,
  [SMALL_STATE(4622)] = 191163,
  [SMALL_STATE(4623)] = 191196,
  [SMALL_STATE(4624)] = 191221,
  [SMALL_STATE(4625)] = 191254,
  [SMALL_STATE(4626)] = 191285,
  [SMALL_STATE(4627)] = 191316,
  [SMALL_STATE(4628)] = 191349,
  [SMALL_STATE(4629)] = 191382,
  [SMALL_STATE(4630)] = 191413,
  [SMALL_STATE(4631)] = 191444,
  [SMALL_STATE(4632)] = 191467,
  [SMALL_STATE(4633)] = 191500,
  [SMALL_STATE(4634)] = 191533,
  [SMALL_STATE(4635)] = 191566,
  [SMALL_STATE(4636)] = 191597,
  [SMALL_STATE(4637)] = 191620,
  [SMALL_STATE(4638)] = 191651,
  [SMALL_STATE(4639)] = 191684,
  [SMALL_STATE(4640)] = 191717,
  [SMALL_STATE(4641)] = 191748,
  [SMALL_STATE(4642)] = 191769,
  [SMALL_STATE(4643)] = 191804,
  [SMALL_STATE(4644)] = 191835,
  [SMALL_STATE(4645)] = 191860,
  [SMALL_STATE(4646)] = 191895,
  [SMALL_STATE(4647)] = 191918,
  [SMALL_STATE(4648)] = 191951,
  [SMALL_STATE(4649)] = 191984,
  [SMALL_STATE(4650)] = 192005,
  [SMALL_STATE(4651)] = 192026,
  [SMALL_STATE(4652)] = 192053,
  [SMALL_STATE(4653)] = 192084,
  [SMALL_STATE(4654)] = 192115,
  [SMALL_STATE(4655)] = 192148,
  [SMALL_STATE(4656)] = 192179,
  [SMALL_STATE(4657)] = 192214,
  [SMALL_STATE(4658)] = 192235,
  [SMALL_STATE(4659)] = 192270,
  [SMALL_STATE(4660)] = 192303,
  [SMALL_STATE(4661)] = 192328,
  [SMALL_STATE(4662)] = 192359,
  [SMALL_STATE(4663)] = 192392,
  [SMALL_STATE(4664)] = 192413,
  [SMALL_STATE(4665)] = 192444,
  [SMALL_STATE(4666)] = 192471,
  [SMALL_STATE(4667)] = 192504,
  [SMALL_STATE(4668)] = 192537,
  [SMALL_STATE(4669)] = 192568,
  [SMALL_STATE(4670)] = 192599,
  [SMALL_STATE(4671)] = 192632,
  [SMALL_STATE(4672)] = 192663,
  [SMALL_STATE(4673)] = 192696,
  [SMALL_STATE(4674)] = 192729,
  [SMALL_STATE(4675)] = 192760,
  [SMALL_STATE(4676)] = 192791,
  [SMALL_STATE(4677)] = 192826,
  [SMALL_STATE(4678)] = 192849,
  [SMALL_STATE(4679)] = 192871,
  [SMALL_STATE(4680)] = 192903,
  [SMALL_STATE(4681)] = 192923,
  [SMALL_STATE(4682)] = 192955,
  [SMALL_STATE(4683)] = 192987,
  [SMALL_STATE(4684)] = 193019,
  [SMALL_STATE(4685)] = 193051,
  [SMALL_STATE(4686)] = 193083,
  [SMALL_STATE(4687)] = 193115,
  [SMALL_STATE(4688)] = 193137,
  [SMALL_STATE(4689)] = 193169,
  [SMALL_STATE(4690)] = 193189,
  [SMALL_STATE(4691)] = 193221,
  [SMALL_STATE(4692)] = 193253,
  [SMALL_STATE(4693)] = 193277,
  [SMALL_STATE(4694)] = 193309,
  [SMALL_STATE(4695)] = 193331,
  [SMALL_STATE(4696)] = 193363,
  [SMALL_STATE(4697)] = 193395,
  [SMALL_STATE(4698)] = 193427,
  [SMALL_STATE(4699)] = 193459,
  [SMALL_STATE(4700)] = 193491,
  [SMALL_STATE(4701)] = 193515,
  [SMALL_STATE(4702)] = 193541,
  [SMALL_STATE(4703)] = 193573,
  [SMALL_STATE(4704)] = 193605,
  [SMALL_STATE(4705)] = 193629,
  [SMALL_STATE(4706)] = 193661,
  [SMALL_STATE(4707)] = 193693,
  [SMALL_STATE(4708)] = 193725,
  [SMALL_STATE(4709)] = 193757,
  [SMALL_STATE(4710)] = 193777,
  [SMALL_STATE(4711)] = 193797,
  [SMALL_STATE(4712)] = 193825,
  [SMALL_STATE(4713)] = 193857,
  [SMALL_STATE(4714)] = 193881,
  [SMALL_STATE(4715)] = 193913,
  [SMALL_STATE(4716)] = 193945,
  [SMALL_STATE(4717)] = 193967,
  [SMALL_STATE(4718)] = 193991,
  [SMALL_STATE(4719)] = 194023,
  [SMALL_STATE(4720)] = 194055,
  [SMALL_STATE(4721)] = 194087,
  [SMALL_STATE(4722)] = 194119,
  [SMALL_STATE(4723)] = 194151,
  [SMALL_STATE(4724)] = 194183,
  [SMALL_STATE(4725)] = 194215,
  [SMALL_STATE(4726)] = 194237,
  [SMALL_STATE(4727)] = 194263,
  [SMALL_STATE(4728)] = 194295,
  [SMALL_STATE(4729)] = 194327,
  [SMALL_STATE(4730)] = 194349,
  [SMALL_STATE(4731)] = 194369,
  [SMALL_STATE(4732)] = 194393,
  [SMALL_STATE(4733)] = 194425,
  [SMALL_STATE(4734)] = 194447,
  [SMALL_STATE(4735)] = 194467,
  [SMALL_STATE(4736)] = 194499,
  [SMALL_STATE(4737)] = 194521,
  [SMALL_STATE(4738)] = 194545,
  [SMALL_STATE(4739)] = 194567,
  [SMALL_STATE(4740)] = 194597,
  [SMALL_STATE(4741)] = 194629,
  [SMALL_STATE(4742)] = 194661,
  [SMALL_STATE(4743)] = 194683,
  [SMALL_STATE(4744)] = 194715,
  [SMALL_STATE(4745)] = 194737,
  [SMALL_STATE(4746)] = 194761,
  [SMALL_STATE(4747)] = 194783,
  [SMALL_STATE(4748)] = 194805,
  [SMALL_STATE(4749)] = 194833,
  [SMALL_STATE(4750)] = 194865,
  [SMALL_STATE(4751)] = 194887,
  [SMALL_STATE(4752)] = 194911,
  [SMALL_STATE(4753)] = 194931,
  [SMALL_STATE(4754)] = 194963,
  [SMALL_STATE(4755)] = 194995,
  [SMALL_STATE(4756)] = 195027,
  [SMALL_STATE(4757)] = 195059,
  [SMALL_STATE(4758)] = 195091,
  [SMALL_STATE(4759)] = 195123,
  [SMALL_STATE(4760)] = 195153,
  [SMALL_STATE(4761)] = 195185,
  [SMALL_STATE(4762)] = 195207,
  [SMALL_STATE(4763)] = 195239,
  [SMALL_STATE(4764)] = 195269,
  [SMALL_STATE(4765)] = 195289,
  [SMALL_STATE(4766)] = 195321,
  [SMALL_STATE(4767)] = 195353,
  [SMALL_STATE(4768)] = 195385,
  [SMALL_STATE(4769)] = 195415,
  [SMALL_STATE(4770)] = 195437,
  [SMALL_STATE(4771)] = 195461,
  [SMALL_STATE(4772)] = 195481,
  [SMALL_STATE(4773)] = 195513,
  [SMALL_STATE(4774)] = 195545,
  [SMALL_STATE(4775)] = 195577,
  [SMALL_STATE(4776)] = 195609,
  [SMALL_STATE(4777)] = 195641,
  [SMALL_STATE(4778)] = 195673,
  [SMALL_STATE(4779)] = 195705,
  [SMALL_STATE(4780)] = 195737,
  [SMALL_STATE(4781)] = 195769,
  [SMALL_STATE(4782)] = 195801,
  [SMALL_STATE(4783)] = 195827,
  [SMALL_STATE(4784)] = 195851,
  [SMALL_STATE(4785)] = 195871,
  [SMALL_STATE(4786)] = 195903,
  [SMALL_STATE(4787)] = 195925,
  [SMALL_STATE(4788)] = 195953,
  [SMALL_STATE(4789)] = 195975,
  [SMALL_STATE(4790)] = 196007,
  [SMALL_STATE(4791)] = 196039,
  [SMALL_STATE(4792)] = 196065,
  [SMALL_STATE(4793)] = 196097,
  [SMALL_STATE(4794)] = 196129,
  [SMALL_STATE(4795)] = 196161,
  [SMALL_STATE(4796)] = 196193,
  [SMALL_STATE(4797)] = 196225,
  [SMALL_STATE(4798)] = 196255,
  [SMALL_STATE(4799)] = 196285,
  [SMALL_STATE(4800)] = 196309,
  [SMALL_STATE(4801)] = 196333,
  [SMALL_STATE(4802)] = 196365,
  [SMALL_STATE(4803)] = 196397,
  [SMALL_STATE(4804)] = 196421,
  [SMALL_STATE(4805)] = 196453,
  [SMALL_STATE(4806)] = 196481,
  [SMALL_STATE(4807)] = 196503,
  [SMALL_STATE(4808)] = 196535,
  [SMALL_STATE(4809)] = 196563,
  [SMALL_STATE(4810)] = 196595,
  [SMALL_STATE(4811)] = 196619,
  [SMALL_STATE(4812)] = 196643,
  [SMALL_STATE(4813)] = 196675,
  [SMALL_STATE(4814)] = 196696,
  [SMALL_STATE(4815)] = 196725,
  [SMALL_STATE(4816)] = 196752,
  [SMALL_STATE(4817)] = 196777,
  [SMALL_STATE(4818)] = 196806,
  [SMALL_STATE(4819)] = 196827,
  [SMALL_STATE(4820)] = 196848,
  [SMALL_STATE(4821)] = 196875,
  [SMALL_STATE(4822)] = 196896,
  [SMALL_STATE(4823)] = 196919,
  [SMALL_STATE(4824)] = 196942,
  [SMALL_STATE(4825)] = 196963,
  [SMALL_STATE(4826)] = 196990,
  [SMALL_STATE(4827)] = 197015,
  [SMALL_STATE(4828)] = 197036,
  [SMALL_STATE(4829)] = 197059,
  [SMALL_STATE(4830)] = 197088,
  [SMALL_STATE(4831)] = 197113,
  [SMALL_STATE(4832)] = 197142,
  [SMALL_STATE(4833)] = 197165,
  [SMALL_STATE(4834)] = 197186,
  [SMALL_STATE(4835)] = 197207,
  [SMALL_STATE(4836)] = 197228,
  [SMALL_STATE(4837)] = 197257,
  [SMALL_STATE(4838)] = 197278,
  [SMALL_STATE(4839)] = 197307,
  [SMALL_STATE(4840)] = 197328,
  [SMALL_STATE(4841)] = 197357,
  [SMALL_STATE(4842)] = 197378,
  [SMALL_STATE(4843)] = 197407,
  [SMALL_STATE(4844)] = 197436,
  [SMALL_STATE(4845)] = 197459,
  [SMALL_STATE(4846)] = 197488,
  [SMALL_STATE(4847)] = 197513,
  [SMALL_STATE(4848)] = 197534,
  [SMALL_STATE(4849)] = 197561,
  [SMALL_STATE(4850)] = 197590,
  [SMALL_STATE(4851)] = 197611,
  [SMALL_STATE(4852)] = 197632,
  [SMALL_STATE(4853)] = 197655,
  [SMALL_STATE(4854)] = 197682,
  [SMALL_STATE(4855)] = 197703,
  [SMALL_STATE(4856)] = 197732,
  [SMALL_STATE(4857)] = 197753,
  [SMALL_STATE(4858)] = 197774,
  [SMALL_STATE(4859)] = 197795,
  [SMALL_STATE(4860)] = 197824,
  [SMALL_STATE(4861)] = 197853,
  [SMALL_STATE(4862)] = 197882,
  [SMALL_STATE(4863)] = 197909,
  [SMALL_STATE(4864)] = 197930,
  [SMALL_STATE(4865)] = 197957,
  [SMALL_STATE(4866)] = 197978,
  [SMALL_STATE(4867)] = 198005,
  [SMALL_STATE(4868)] = 198034,
  [SMALL_STATE(4869)] = 198055,
  [SMALL_STATE(4870)] = 198082,
  [SMALL_STATE(4871)] = 198109,
  [SMALL_STATE(4872)] = 198136,
  [SMALL_STATE(4873)] = 198157,
  [SMALL_STATE(4874)] = 198186,
  [SMALL_STATE(4875)] = 198213,
  [SMALL_STATE(4876)] = 198240,
  [SMALL_STATE(4877)] = 198267,
  [SMALL_STATE(4878)] = 198294,
  [SMALL_STATE(4879)] = 198321,
  [SMALL_STATE(4880)] = 198342,
  [SMALL_STATE(4881)] = 198369,
  [SMALL_STATE(4882)] = 198396,
  [SMALL_STATE(4883)] = 198417,
  [SMALL_STATE(4884)] = 198438,
  [SMALL_STATE(4885)] = 198467,
  [SMALL_STATE(4886)] = 198496,
  [SMALL_STATE(4887)] = 198517,
  [SMALL_STATE(4888)] = 198538,
  [SMALL_STATE(4889)] = 198567,
  [SMALL_STATE(4890)] = 198596,
  [SMALL_STATE(4891)] = 198625,
  [SMALL_STATE(4892)] = 198646,
  [SMALL_STATE(4893)] = 198667,
  [SMALL_STATE(4894)] = 198696,
  [SMALL_STATE(4895)] = 198717,
  [SMALL_STATE(4896)] = 198746,
  [SMALL_STATE(4897)] = 198767,
  [SMALL_STATE(4898)] = 198788,
  [SMALL_STATE(4899)] = 198809,
  [SMALL_STATE(4900)] = 198830,
  [SMALL_STATE(4901)] = 198851,
  [SMALL_STATE(4902)] = 198880,
  [SMALL_STATE(4903)] = 198901,
  [SMALL_STATE(4904)] = 198930,
  [SMALL_STATE(4905)] = 198951,
  [SMALL_STATE(4906)] = 198980,
  [SMALL_STATE(4907)] = 199007,
  [SMALL_STATE(4908)] = 199036,
  [SMALL_STATE(4909)] = 199063,
  [SMALL_STATE(4910)] = 199086,
  [SMALL_STATE(4911)] = 199107,
  [SMALL_STATE(4912)] = 199128,
  [SMALL_STATE(4913)] = 199157,
  [SMALL_STATE(4914)] = 199178,
  [SMALL_STATE(4915)] = 199199,
  [SMALL_STATE(4916)] = 199228,
  [SMALL_STATE(4917)] = 199249,
  [SMALL_STATE(4918)] = 199278,
  [SMALL_STATE(4919)] = 199307,
  [SMALL_STATE(4920)] = 199336,
  [SMALL_STATE(4921)] = 199365,
  [SMALL_STATE(4922)] = 199390,
  [SMALL_STATE(4923)] = 199415,
  [SMALL_STATE(4924)] = 199444,
  [SMALL_STATE(4925)] = 199465,
  [SMALL_STATE(4926)] = 199486,
  [SMALL_STATE(4927)] = 199515,
  [SMALL_STATE(4928)] = 199536,
  [SMALL_STATE(4929)] = 199559,
  [SMALL_STATE(4930)] = 199586,
  [SMALL_STATE(4931)] = 199615,
  [SMALL_STATE(4932)] = 199644,
  [SMALL_STATE(4933)] = 199673,
  [SMALL_STATE(4934)] = 199700,
  [SMALL_STATE(4935)] = 199729,
  [SMALL_STATE(4936)] = 199750,
  [SMALL_STATE(4937)] = 199779,
  [SMALL_STATE(4938)] = 199800,
  [SMALL_STATE(4939)] = 199823,
  [SMALL_STATE(4940)] = 199848,
  [SMALL_STATE(4941)] = 199869,
  [SMALL_STATE(4942)] = 199898,
  [SMALL_STATE(4943)] = 199923,
  [SMALL_STATE(4944)] = 199952,
  [SMALL_STATE(4945)] = 199973,
  [SMALL_STATE(4946)] = 200000,
  [SMALL_STATE(4947)] = 200029,
  [SMALL_STATE(4948)] = 200052,
  [SMALL_STATE(4949)] = 200075,
  [SMALL_STATE(4950)] = 200104,
  [SMALL_STATE(4951)] = 200133,
  [SMALL_STATE(4952)] = 200154,
  [SMALL_STATE(4953)] = 200181,
  [SMALL_STATE(4954)] = 200208,
  [SMALL_STATE(4955)] = 200237,
  [SMALL_STATE(4956)] = 200260,
  [SMALL_STATE(4957)] = 200283,
  [SMALL_STATE(4958)] = 200312,
  [SMALL_STATE(4959)] = 200333,
  [SMALL_STATE(4960)] = 200354,
  [SMALL_STATE(4961)] = 200381,
  [SMALL_STATE(4962)] = 200402,
  [SMALL_STATE(4963)] = 200429,
  [SMALL_STATE(4964)] = 200450,
  [SMALL_STATE(4965)] = 200479,
  [SMALL_STATE(4966)] = 200500,
  [SMALL_STATE(4967)] = 200521,
  [SMALL_STATE(4968)] = 200542,
  [SMALL_STATE(4969)] = 200569,
  [SMALL_STATE(4970)] = 200590,
  [SMALL_STATE(4971)] = 200619,
  [SMALL_STATE(4972)] = 200642,
  [SMALL_STATE(4973)] = 200671,
  [SMALL_STATE(4974)] = 200692,
  [SMALL_STATE(4975)] = 200719,
  [SMALL_STATE(4976)] = 200740,
  [SMALL_STATE(4977)] = 200761,
  [SMALL_STATE(4978)] = 200782,
  [SMALL_STATE(4979)] = 200811,
  [SMALL_STATE(4980)] = 200840,
  [SMALL_STATE(4981)] = 200861,
  [SMALL_STATE(4982)] = 200882,
  [SMALL_STATE(4983)] = 200911,
  [SMALL_STATE(4984)] = 200938,
  [SMALL_STATE(4985)] = 200959,
  [SMALL_STATE(4986)] = 200982,
  [SMALL_STATE(4987)] = 201005,
  [SMALL_STATE(4988)] = 201026,
  [SMALL_STATE(4989)] = 201055,
  [SMALL_STATE(4990)] = 201084,
  [SMALL_STATE(4991)] = 201105,
  [SMALL_STATE(4992)] = 201126,
  [SMALL_STATE(4993)] = 201155,
  [SMALL_STATE(4994)] = 201176,
  [SMALL_STATE(4995)] = 201197,
  [SMALL_STATE(4996)] = 201224,
  [SMALL_STATE(4997)] = 201253,
  [SMALL_STATE(4998)] = 201274,
  [SMALL_STATE(4999)] = 201303,
  [SMALL_STATE(5000)] = 201324,
  [SMALL_STATE(5001)] = 201353,
  [SMALL_STATE(5002)] = 201382,
  [SMALL_STATE(5003)] = 201409,
  [SMALL_STATE(5004)] = 201430,
  [SMALL_STATE(5005)] = 201457,
  [SMALL_STATE(5006)] = 201478,
  [SMALL_STATE(5007)] = 201507,
  [SMALL_STATE(5008)] = 201536,
  [SMALL_STATE(5009)] = 201557,
  [SMALL_STATE(5010)] = 201578,
  [SMALL_STATE(5011)] = 201599,
  [SMALL_STATE(5012)] = 201620,
  [SMALL_STATE(5013)] = 201649,
  [SMALL_STATE(5014)] = 201678,
  [SMALL_STATE(5015)] = 201705,
  [SMALL_STATE(5016)] = 201732,
  [SMALL_STATE(5017)] = 201753,
  [SMALL_STATE(5018)] = 201780,
  [SMALL_STATE(5019)] = 201801,
  [SMALL_STATE(5020)] = 201828,
  [SMALL_STATE(5021)] = 201855,
  [SMALL_STATE(5022)] = 201884,
  [SMALL_STATE(5023)] = 201905,
  [SMALL_STATE(5024)] = 201934,
  [SMALL_STATE(5025)] = 201961,
  [SMALL_STATE(5026)] = 201990,
  [SMALL_STATE(5027)] = 202017,
  [SMALL_STATE(5028)] = 202046,
  [SMALL_STATE(5029)] = 202067,
  [SMALL_STATE(5030)] = 202096,
  [SMALL_STATE(5031)] = 202117,
  [SMALL_STATE(5032)] = 202144,
  [SMALL_STATE(5033)] = 202173,
  [SMALL_STATE(5034)] = 202196,
  [SMALL_STATE(5035)] = 202217,
  [SMALL_STATE(5036)] = 202246,
  [SMALL_STATE(5037)] = 202273,
  [SMALL_STATE(5038)] = 202296,
  [SMALL_STATE(5039)] = 202325,
  [SMALL_STATE(5040)] = 202346,
  [SMALL_STATE(5041)] = 202373,
  [SMALL_STATE(5042)] = 202402,
  [SMALL_STATE(5043)] = 202426,
  [SMALL_STATE(5044)] = 202450,
  [SMALL_STATE(5045)] = 202476,
  [SMALL_STATE(5046)] = 202502,
  [SMALL_STATE(5047)] = 202526,
  [SMALL_STATE(5048)] = 202552,
  [SMALL_STATE(5049)] = 202578,
  [SMALL_STATE(5050)] = 202604,
  [SMALL_STATE(5051)] = 202628,
  [SMALL_STATE(5052)] = 202654,
  [SMALL_STATE(5053)] = 202680,
  [SMALL_STATE(5054)] = 202706,
  [SMALL_STATE(5055)] = 202730,
  [SMALL_STATE(5056)] = 202756,
  [SMALL_STATE(5057)] = 202780,
  [SMALL_STATE(5058)] = 202806,
  [SMALL_STATE(5059)] = 202830,
  [SMALL_STATE(5060)] = 202854,
  [SMALL_STATE(5061)] = 202876,
  [SMALL_STATE(5062)] = 202900,
  [SMALL_STATE(5063)] = 202926,
  [SMALL_STATE(5064)] = 202950,
  [SMALL_STATE(5065)] = 202976,
  [SMALL_STATE(5066)] = 203000,
  [SMALL_STATE(5067)] = 203018,
  [SMALL_STATE(5068)] = 203042,
  [SMALL_STATE(5069)] = 203066,
  [SMALL_STATE(5070)] = 203090,
  [SMALL_STATE(5071)] = 203108,
  [SMALL_STATE(5072)] = 203132,
  [SMALL_STATE(5073)] = 203158,
  [SMALL_STATE(5074)] = 203182,
  [SMALL_STATE(5075)] = 203208,
  [SMALL_STATE(5076)] = 203232,
  [SMALL_STATE(5077)] = 203256,
  [SMALL_STATE(5078)] = 203280,
  [SMALL_STATE(5079)] = 203306,
  [SMALL_STATE(5080)] = 203328,
  [SMALL_STATE(5081)] = 203346,
  [SMALL_STATE(5082)] = 203368,
  [SMALL_STATE(5083)] = 203392,
  [SMALL_STATE(5084)] = 203414,
  [SMALL_STATE(5085)] = 203434,
  [SMALL_STATE(5086)] = 203460,
  [SMALL_STATE(5087)] = 203480,
  [SMALL_STATE(5088)] = 203504,
  [SMALL_STATE(5089)] = 203530,
  [SMALL_STATE(5090)] = 203554,
  [SMALL_STATE(5091)] = 203578,
  [SMALL_STATE(5092)] = 203602,
  [SMALL_STATE(5093)] = 203626,
  [SMALL_STATE(5094)] = 203650,
  [SMALL_STATE(5095)] = 203676,
  [SMALL_STATE(5096)] = 203700,
  [SMALL_STATE(5097)] = 203726,
  [SMALL_STATE(5098)] = 203746,
  [SMALL_STATE(5099)] = 203768,
  [SMALL_STATE(5100)] = 203792,
  [SMALL_STATE(5101)] = 203816,
  [SMALL_STATE(5102)] = 203842,
  [SMALL_STATE(5103)] = 203866,
  [SMALL_STATE(5104)] = 203884,
  [SMALL_STATE(5105)] = 203908,
  [SMALL_STATE(5106)] = 203934,
  [SMALL_STATE(5107)] = 203960,
  [SMALL_STATE(5108)] = 203986,
  [SMALL_STATE(5109)] = 204008,
  [SMALL_STATE(5110)] = 204034,
  [SMALL_STATE(5111)] = 204052,
  [SMALL_STATE(5112)] = 204076,
  [SMALL_STATE(5113)] = 204102,
  [SMALL_STATE(5114)] = 204124,
  [SMALL_STATE(5115)] = 204150,
  [SMALL_STATE(5116)] = 204174,
  [SMALL_STATE(5117)] = 204198,
  [SMALL_STATE(5118)] = 204222,
  [SMALL_STATE(5119)] = 204246,
  [SMALL_STATE(5120)] = 204270,
  [SMALL_STATE(5121)] = 204294,
  [SMALL_STATE(5122)] = 204318,
  [SMALL_STATE(5123)] = 204344,
  [SMALL_STATE(5124)] = 204370,
  [SMALL_STATE(5125)] = 204396,
  [SMALL_STATE(5126)] = 204418,
  [SMALL_STATE(5127)] = 204442,
  [SMALL_STATE(5128)] = 204466,
  [SMALL_STATE(5129)] = 204490,
  [SMALL_STATE(5130)] = 204514,
  [SMALL_STATE(5131)] = 204538,
  [SMALL_STATE(5132)] = 204562,
  [SMALL_STATE(5133)] = 204586,
  [SMALL_STATE(5134)] = 204612,
  [SMALL_STATE(5135)] = 204636,
  [SMALL_STATE(5136)] = 204658,
  [SMALL_STATE(5137)] = 204682,
  [SMALL_STATE(5138)] = 204706,
  [SMALL_STATE(5139)] = 204730,
  [SMALL_STATE(5140)] = 204756,
  [SMALL_STATE(5141)] = 204782,
  [SMALL_STATE(5142)] = 204808,
  [SMALL_STATE(5143)] = 204832,
  [SMALL_STATE(5144)] = 204856,
  [SMALL_STATE(5145)] = 204878,
  [SMALL_STATE(5146)] = 204902,
  [SMALL_STATE(5147)] = 204928,
  [SMALL_STATE(5148)] = 204946,
  [SMALL_STATE(5149)] = 204970,
  [SMALL_STATE(5150)] = 204994,
  [SMALL_STATE(5151)] = 205018,
  [SMALL_STATE(5152)] = 205042,
  [SMALL_STATE(5153)] = 205066,
  [SMALL_STATE(5154)] = 205092,
  [SMALL_STATE(5155)] = 205116,
  [SMALL_STATE(5156)] = 205142,
  [SMALL_STATE(5157)] = 205168,
  [SMALL_STATE(5158)] = 205192,
  [SMALL_STATE(5159)] = 205216,
  [SMALL_STATE(5160)] = 205240,
  [SMALL_STATE(5161)] = 205264,
  [SMALL_STATE(5162)] = 205290,
  [SMALL_STATE(5163)] = 205316,
  [SMALL_STATE(5164)] = 205342,
  [SMALL_STATE(5165)] = 205366,
  [SMALL_STATE(5166)] = 205392,
  [SMALL_STATE(5167)] = 205416,
  [SMALL_STATE(5168)] = 205440,
  [SMALL_STATE(5169)] = 205464,
  [SMALL_STATE(5170)] = 205488,
  [SMALL_STATE(5171)] = 205512,
  [SMALL_STATE(5172)] = 205538,
  [SMALL_STATE(5173)] = 205562,
  [SMALL_STATE(5174)] = 205588,
  [SMALL_STATE(5175)] = 205612,
  [SMALL_STATE(5176)] = 205636,
  [SMALL_STATE(5177)] = 205658,
  [SMALL_STATE(5178)] = 205680,
  [SMALL_STATE(5179)] = 205702,
  [SMALL_STATE(5180)] = 205720,
  [SMALL_STATE(5181)] = 205742,
  [SMALL_STATE(5182)] = 205768,
  [SMALL_STATE(5183)] = 205792,
  [SMALL_STATE(5184)] = 205816,
  [SMALL_STATE(5185)] = 205840,
  [SMALL_STATE(5186)] = 205866,
  [SMALL_STATE(5187)] = 205890,
  [SMALL_STATE(5188)] = 205914,
  [SMALL_STATE(5189)] = 205938,
  [SMALL_STATE(5190)] = 205962,
  [SMALL_STATE(5191)] = 205986,
  [SMALL_STATE(5192)] = 206010,
  [SMALL_STATE(5193)] = 206036,
  [SMALL_STATE(5194)] = 206058,
  [SMALL_STATE(5195)] = 206082,
  [SMALL_STATE(5196)] = 206106,
  [SMALL_STATE(5197)] = 206130,
  [SMALL_STATE(5198)] = 206154,
  [SMALL_STATE(5199)] = 206178,
  [SMALL_STATE(5200)] = 206202,
  [SMALL_STATE(5201)] = 206220,
  [SMALL_STATE(5202)] = 206244,
  [SMALL_STATE(5203)] = 206270,
  [SMALL_STATE(5204)] = 206294,
  [SMALL_STATE(5205)] = 206318,
  [SMALL_STATE(5206)] = 206336,
  [SMALL_STATE(5207)] = 206360,
  [SMALL_STATE(5208)] = 206386,
  [SMALL_STATE(5209)] = 206412,
  [SMALL_STATE(5210)] = 206436,
  [SMALL_STATE(5211)] = 206460,
  [SMALL_STATE(5212)] = 206484,
  [SMALL_STATE(5213)] = 206508,
  [SMALL_STATE(5214)] = 206532,
  [SMALL_STATE(5215)] = 206556,
  [SMALL_STATE(5216)] = 206582,
  [SMALL_STATE(5217)] = 206606,
  [SMALL_STATE(5218)] = 206630,
  [SMALL_STATE(5219)] = 206654,
  [SMALL_STATE(5220)] = 206678,
  [SMALL_STATE(5221)] = 206704,
  [SMALL_STATE(5222)] = 206728,
  [SMALL_STATE(5223)] = 206752,
  [SMALL_STATE(5224)] = 206776,
  [SMALL_STATE(5225)] = 206802,
  [SMALL_STATE(5226)] = 206826,
  [SMALL_STATE(5227)] = 206844,
  [SMALL_STATE(5228)] = 206868,
  [SMALL_STATE(5229)] = 206894,
  [SMALL_STATE(5230)] = 206920,
  [SMALL_STATE(5231)] = 206946,
  [SMALL_STATE(5232)] = 206970,
  [SMALL_STATE(5233)] = 206994,
  [SMALL_STATE(5234)] = 207018,
  [SMALL_STATE(5235)] = 207042,
  [SMALL_STATE(5236)] = 207066,
  [SMALL_STATE(5237)] = 207092,
  [SMALL_STATE(5238)] = 207118,
  [SMALL_STATE(5239)] = 207142,
  [SMALL_STATE(5240)] = 207166,
  [SMALL_STATE(5241)] = 207190,
  [SMALL_STATE(5242)] = 207214,
  [SMALL_STATE(5243)] = 207240,
  [SMALL_STATE(5244)] = 207264,
  [SMALL_STATE(5245)] = 207290,
  [SMALL_STATE(5246)] = 207316,
  [SMALL_STATE(5247)] = 207342,
  [SMALL_STATE(5248)] = 207368,
  [SMALL_STATE(5249)] = 207394,
  [SMALL_STATE(5250)] = 207420,
  [SMALL_STATE(5251)] = 207446,
  [SMALL_STATE(5252)] = 207470,
  [SMALL_STATE(5253)] = 207494,
  [SMALL_STATE(5254)] = 207516,
  [SMALL_STATE(5255)] = 207540,
  [SMALL_STATE(5256)] = 207564,
  [SMALL_STATE(5257)] = 207590,
  [SMALL_STATE(5258)] = 207616,
  [SMALL_STATE(5259)] = 207640,
  [SMALL_STATE(5260)] = 207662,
  [SMALL_STATE(5261)] = 207688,
  [SMALL_STATE(5262)] = 207712,
  [SMALL_STATE(5263)] = 207736,
  [SMALL_STATE(5264)] = 207754,
  [SMALL_STATE(5265)] = 207778,
  [SMALL_STATE(5266)] = 207801,
  [SMALL_STATE(5267)] = 207824,
  [SMALL_STATE(5268)] = 207847,
  [SMALL_STATE(5269)] = 207870,
  [SMALL_STATE(5270)] = 207893,
  [SMALL_STATE(5271)] = 207914,
  [SMALL_STATE(5272)] = 207931,
  [SMALL_STATE(5273)] = 207954,
  [SMALL_STATE(5274)] = 207977,
  [SMALL_STATE(5275)] = 208000,
  [SMALL_STATE(5276)] = 208023,
  [SMALL_STATE(5277)] = 208046,
  [SMALL_STATE(5278)] = 208069,
  [SMALL_STATE(5279)] = 208092,
  [SMALL_STATE(5280)] = 208115,
  [SMALL_STATE(5281)] = 208136,
  [SMALL_STATE(5282)] = 208159,
  [SMALL_STATE(5283)] = 208182,
  [SMALL_STATE(5284)] = 208203,
  [SMALL_STATE(5285)] = 208226,
  [SMALL_STATE(5286)] = 208249,
  [SMALL_STATE(5287)] = 208266,
  [SMALL_STATE(5288)] = 208283,
  [SMALL_STATE(5289)] = 208306,
  [SMALL_STATE(5290)] = 208329,
  [SMALL_STATE(5291)] = 208352,
  [SMALL_STATE(5292)] = 208375,
  [SMALL_STATE(5293)] = 208394,
  [SMALL_STATE(5294)] = 208417,
  [SMALL_STATE(5295)] = 208440,
  [SMALL_STATE(5296)] = 208463,
  [SMALL_STATE(5297)] = 208486,
  [SMALL_STATE(5298)] = 208509,
  [SMALL_STATE(5299)] = 208532,
  [SMALL_STATE(5300)] = 208549,
  [SMALL_STATE(5301)] = 208566,
  [SMALL_STATE(5302)] = 208589,
  [SMALL_STATE(5303)] = 208606,
  [SMALL_STATE(5304)] = 208629,
  [SMALL_STATE(5305)] = 208652,
  [SMALL_STATE(5306)] = 208673,
  [SMALL_STATE(5307)] = 208696,
  [SMALL_STATE(5308)] = 208719,
  [SMALL_STATE(5309)] = 208742,
  [SMALL_STATE(5310)] = 208765,
  [SMALL_STATE(5311)] = 208782,
  [SMALL_STATE(5312)] = 208805,
  [SMALL_STATE(5313)] = 208826,
  [SMALL_STATE(5314)] = 208849,
  [SMALL_STATE(5315)] = 208872,
  [SMALL_STATE(5316)] = 208895,
  [SMALL_STATE(5317)] = 208918,
  [SMALL_STATE(5318)] = 208941,
  [SMALL_STATE(5319)] = 208964,
  [SMALL_STATE(5320)] = 208987,
  [SMALL_STATE(5321)] = 209010,
  [SMALL_STATE(5322)] = 209031,
  [SMALL_STATE(5323)] = 209054,
  [SMALL_STATE(5324)] = 209077,
  [SMALL_STATE(5325)] = 209100,
  [SMALL_STATE(5326)] = 209123,
  [SMALL_STATE(5327)] = 209146,
  [SMALL_STATE(5328)] = 209165,
  [SMALL_STATE(5329)] = 209184,
  [SMALL_STATE(5330)] = 209207,
  [SMALL_STATE(5331)] = 209228,
  [SMALL_STATE(5332)] = 209245,
  [SMALL_STATE(5333)] = 209268,
  [SMALL_STATE(5334)] = 209291,
  [SMALL_STATE(5335)] = 209308,
  [SMALL_STATE(5336)] = 209325,
  [SMALL_STATE(5337)] = 209348,
  [SMALL_STATE(5338)] = 209369,
  [SMALL_STATE(5339)] = 209392,
  [SMALL_STATE(5340)] = 209415,
  [SMALL_STATE(5341)] = 209438,
  [SMALL_STATE(5342)] = 209457,
  [SMALL_STATE(5343)] = 209478,
  [SMALL_STATE(5344)] = 209497,
  [SMALL_STATE(5345)] = 209518,
  [SMALL_STATE(5346)] = 209539,
  [SMALL_STATE(5347)] = 209558,
  [SMALL_STATE(5348)] = 209581,
  [SMALL_STATE(5349)] = 209604,
  [SMALL_STATE(5350)] = 209627,
  [SMALL_STATE(5351)] = 209650,
  [SMALL_STATE(5352)] = 209673,
  [SMALL_STATE(5353)] = 209696,
  [SMALL_STATE(5354)] = 209719,
  [SMALL_STATE(5355)] = 209742,
  [SMALL_STATE(5356)] = 209761,
  [SMALL_STATE(5357)] = 209784,
  [SMALL_STATE(5358)] = 209807,
  [SMALL_STATE(5359)] = 209830,
  [SMALL_STATE(5360)] = 209853,
  [SMALL_STATE(5361)] = 209876,
  [SMALL_STATE(5362)] = 209899,
  [SMALL_STATE(5363)] = 209920,
  [SMALL_STATE(5364)] = 209937,
  [SMALL_STATE(5365)] = 209960,
  [SMALL_STATE(5366)] = 209983,
  [SMALL_STATE(5367)] = 210006,
  [SMALL_STATE(5368)] = 210029,
  [SMALL_STATE(5369)] = 210052,
  [SMALL_STATE(5370)] = 210075,
  [SMALL_STATE(5371)] = 210098,
  [SMALL_STATE(5372)] = 210121,
  [SMALL_STATE(5373)] = 210140,
  [SMALL_STATE(5374)] = 210163,
  [SMALL_STATE(5375)] = 210186,
  [SMALL_STATE(5376)] = 210205,
  [SMALL_STATE(5377)] = 210224,
  [SMALL_STATE(5378)] = 210247,
  [SMALL_STATE(5379)] = 210270,
  [SMALL_STATE(5380)] = 210293,
  [SMALL_STATE(5381)] = 210316,
  [SMALL_STATE(5382)] = 210335,
  [SMALL_STATE(5383)] = 210358,
  [SMALL_STATE(5384)] = 210381,
  [SMALL_STATE(5385)] = 210404,
  [SMALL_STATE(5386)] = 210427,
  [SMALL_STATE(5387)] = 210450,
  [SMALL_STATE(5388)] = 210467,
  [SMALL_STATE(5389)] = 210490,
  [SMALL_STATE(5390)] = 210511,
  [SMALL_STATE(5391)] = 210534,
  [SMALL_STATE(5392)] = 210557,
  [SMALL_STATE(5393)] = 210580,
  [SMALL_STATE(5394)] = 210603,
  [SMALL_STATE(5395)] = 210626,
  [SMALL_STATE(5396)] = 210647,
  [SMALL_STATE(5397)] = 210670,
  [SMALL_STATE(5398)] = 210693,
  [SMALL_STATE(5399)] = 210716,
  [SMALL_STATE(5400)] = 210733,
  [SMALL_STATE(5401)] = 210756,
  [SMALL_STATE(5402)] = 210779,
  [SMALL_STATE(5403)] = 210800,
  [SMALL_STATE(5404)] = 210823,
  [SMALL_STATE(5405)] = 210842,
  [SMALL_STATE(5406)] = 210863,
  [SMALL_STATE(5407)] = 210886,
  [SMALL_STATE(5408)] = 210909,
  [SMALL_STATE(5409)] = 210928,
  [SMALL_STATE(5410)] = 210951,
  [SMALL_STATE(5411)] = 210974,
  [SMALL_STATE(5412)] = 210997,
  [SMALL_STATE(5413)] = 211020,
  [SMALL_STATE(5414)] = 211043,
  [SMALL_STATE(5415)] = 211066,
  [SMALL_STATE(5416)] = 211089,
  [SMALL_STATE(5417)] = 211112,
  [SMALL_STATE(5418)] = 211131,
  [SMALL_STATE(5419)] = 211150,
  [SMALL_STATE(5420)] = 211169,
  [SMALL_STATE(5421)] = 211190,
  [SMALL_STATE(5422)] = 211213,
  [SMALL_STATE(5423)] = 211234,
  [SMALL_STATE(5424)] = 211257,
  [SMALL_STATE(5425)] = 211280,
  [SMALL_STATE(5426)] = 211303,
  [SMALL_STATE(5427)] = 211326,
  [SMALL_STATE(5428)] = 211349,
  [SMALL_STATE(5429)] = 211372,
  [SMALL_STATE(5430)] = 211395,
  [SMALL_STATE(5431)] = 211418,
  [SMALL_STATE(5432)] = 211439,
  [SMALL_STATE(5433)] = 211462,
  [SMALL_STATE(5434)] = 211485,
  [SMALL_STATE(5435)] = 211506,
  [SMALL_STATE(5436)] = 211529,
  [SMALL_STATE(5437)] = 211552,
  [SMALL_STATE(5438)] = 211573,
  [SMALL_STATE(5439)] = 211596,
  [SMALL_STATE(5440)] = 211617,
  [SMALL_STATE(5441)] = 211640,
  [SMALL_STATE(5442)] = 211661,
  [SMALL_STATE(5443)] = 211684,
  [SMALL_STATE(5444)] = 211707,
  [SMALL_STATE(5445)] = 211730,
  [SMALL_STATE(5446)] = 211753,
  [SMALL_STATE(5447)] = 211776,
  [SMALL_STATE(5448)] = 211799,
  [SMALL_STATE(5449)] = 211820,
  [SMALL_STATE(5450)] = 211843,
  [SMALL_STATE(5451)] = 211866,
  [SMALL_STATE(5452)] = 211889,
  [SMALL_STATE(5453)] = 211912,
  [SMALL_STATE(5454)] = 211935,
  [SMALL_STATE(5455)] = 211958,
  [SMALL_STATE(5456)] = 211981,
  [SMALL_STATE(5457)] = 212002,
  [SMALL_STATE(5458)] = 212025,
  [SMALL_STATE(5459)] = 212048,
  [SMALL_STATE(5460)] = 212071,
  [SMALL_STATE(5461)] = 212094,
  [SMALL_STATE(5462)] = 212117,
  [SMALL_STATE(5463)] = 212138,
  [SMALL_STATE(5464)] = 212161,
  [SMALL_STATE(5465)] = 212184,
  [SMALL_STATE(5466)] = 212207,
  [SMALL_STATE(5467)] = 212230,
  [SMALL_STATE(5468)] = 212251,
  [SMALL_STATE(5469)] = 212274,
  [SMALL_STATE(5470)] = 212297,
  [SMALL_STATE(5471)] = 212318,
  [SMALL_STATE(5472)] = 212339,
  [SMALL_STATE(5473)] = 212362,
  [SMALL_STATE(5474)] = 212385,
  [SMALL_STATE(5475)] = 212406,
  [SMALL_STATE(5476)] = 212429,
  [SMALL_STATE(5477)] = 212450,
  [SMALL_STATE(5478)] = 212473,
  [SMALL_STATE(5479)] = 212496,
  [SMALL_STATE(5480)] = 212519,
  [SMALL_STATE(5481)] = 212542,
  [SMALL_STATE(5482)] = 212565,
  [SMALL_STATE(5483)] = 212588,
  [SMALL_STATE(5484)] = 212611,
  [SMALL_STATE(5485)] = 212634,
  [SMALL_STATE(5486)] = 212657,
  [SMALL_STATE(5487)] = 212680,
  [SMALL_STATE(5488)] = 212697,
  [SMALL_STATE(5489)] = 212720,
  [SMALL_STATE(5490)] = 212743,
  [SMALL_STATE(5491)] = 212766,
  [SMALL_STATE(5492)] = 212783,
  [SMALL_STATE(5493)] = 212806,
  [SMALL_STATE(5494)] = 212829,
  [SMALL_STATE(5495)] = 212852,
  [SMALL_STATE(5496)] = 212875,
  [SMALL_STATE(5497)] = 212898,
  [SMALL_STATE(5498)] = 212921,
  [SMALL_STATE(5499)] = 212944,
  [SMALL_STATE(5500)] = 212967,
  [SMALL_STATE(5501)] = 212990,
  [SMALL_STATE(5502)] = 213013,
  [SMALL_STATE(5503)] = 213036,
  [SMALL_STATE(5504)] = 213059,
  [SMALL_STATE(5505)] = 213082,
  [SMALL_STATE(5506)] = 213105,
  [SMALL_STATE(5507)] = 213128,
  [SMALL_STATE(5508)] = 213151,
  [SMALL_STATE(5509)] = 213174,
  [SMALL_STATE(5510)] = 213197,
  [SMALL_STATE(5511)] = 213216,
  [SMALL_STATE(5512)] = 213239,
  [SMALL_STATE(5513)] = 213262,
  [SMALL_STATE(5514)] = 213285,
  [SMALL_STATE(5515)] = 213308,
  [SMALL_STATE(5516)] = 213331,
  [SMALL_STATE(5517)] = 213354,
  [SMALL_STATE(5518)] = 213377,
  [SMALL_STATE(5519)] = 213398,
  [SMALL_STATE(5520)] = 213419,
  [SMALL_STATE(5521)] = 213442,
  [SMALL_STATE(5522)] = 213465,
  [SMALL_STATE(5523)] = 213488,
  [SMALL_STATE(5524)] = 213511,
  [SMALL_STATE(5525)] = 213534,
  [SMALL_STATE(5526)] = 213557,
  [SMALL_STATE(5527)] = 213580,
  [SMALL_STATE(5528)] = 213601,
  [SMALL_STATE(5529)] = 213624,
  [SMALL_STATE(5530)] = 213647,
  [SMALL_STATE(5531)] = 213670,
  [SMALL_STATE(5532)] = 213693,
  [SMALL_STATE(5533)] = 213714,
  [SMALL_STATE(5534)] = 213735,
  [SMALL_STATE(5535)] = 213758,
  [SMALL_STATE(5536)] = 213779,
  [SMALL_STATE(5537)] = 213802,
  [SMALL_STATE(5538)] = 213823,
  [SMALL_STATE(5539)] = 213840,
  [SMALL_STATE(5540)] = 213863,
  [SMALL_STATE(5541)] = 213886,
  [SMALL_STATE(5542)] = 213909,
  [SMALL_STATE(5543)] = 213932,
  [SMALL_STATE(5544)] = 213952,
  [SMALL_STATE(5545)] = 213972,
  [SMALL_STATE(5546)] = 213988,
  [SMALL_STATE(5547)] = 214008,
  [SMALL_STATE(5548)] = 214028,
  [SMALL_STATE(5549)] = 214048,
  [SMALL_STATE(5550)] = 214068,
  [SMALL_STATE(5551)] = 214088,
  [SMALL_STATE(5552)] = 214108,
  [SMALL_STATE(5553)] = 214128,
  [SMALL_STATE(5554)] = 214148,
  [SMALL_STATE(5555)] = 214168,
  [SMALL_STATE(5556)] = 214186,
  [SMALL_STATE(5557)] = 214206,
  [SMALL_STATE(5558)] = 214226,
  [SMALL_STATE(5559)] = 214244,
  [SMALL_STATE(5560)] = 214264,
  [SMALL_STATE(5561)] = 214284,
  [SMALL_STATE(5562)] = 214300,
  [SMALL_STATE(5563)] = 214320,
  [SMALL_STATE(5564)] = 214340,
  [SMALL_STATE(5565)] = 214360,
  [SMALL_STATE(5566)] = 214380,
  [SMALL_STATE(5567)] = 214400,
  [SMALL_STATE(5568)] = 214418,
  [SMALL_STATE(5569)] = 214438,
  [SMALL_STATE(5570)] = 214458,
  [SMALL_STATE(5571)] = 214478,
  [SMALL_STATE(5572)] = 214498,
  [SMALL_STATE(5573)] = 214518,
  [SMALL_STATE(5574)] = 214538,
  [SMALL_STATE(5575)] = 214558,
  [SMALL_STATE(5576)] = 214578,
  [SMALL_STATE(5577)] = 214598,
  [SMALL_STATE(5578)] = 214618,
  [SMALL_STATE(5579)] = 214636,
  [SMALL_STATE(5580)] = 214656,
  [SMALL_STATE(5581)] = 214676,
  [SMALL_STATE(5582)] = 214696,
  [SMALL_STATE(5583)] = 214716,
  [SMALL_STATE(5584)] = 214736,
  [SMALL_STATE(5585)] = 214756,
  [SMALL_STATE(5586)] = 214776,
  [SMALL_STATE(5587)] = 214796,
  [SMALL_STATE(5588)] = 214814,
  [SMALL_STATE(5589)] = 214834,
  [SMALL_STATE(5590)] = 214854,
  [SMALL_STATE(5591)] = 214870,
  [SMALL_STATE(5592)] = 214890,
  [SMALL_STATE(5593)] = 214910,
  [SMALL_STATE(5594)] = 214930,
  [SMALL_STATE(5595)] = 214950,
  [SMALL_STATE(5596)] = 214970,
  [SMALL_STATE(5597)] = 214990,
  [SMALL_STATE(5598)] = 215010,
  [SMALL_STATE(5599)] = 215028,
  [SMALL_STATE(5600)] = 215046,
  [SMALL_STATE(5601)] = 215066,
  [SMALL_STATE(5602)] = 215086,
  [SMALL_STATE(5603)] = 215106,
  [SMALL_STATE(5604)] = 215122,
  [SMALL_STATE(5605)] = 215142,
  [SMALL_STATE(5606)] = 215162,
  [SMALL_STATE(5607)] = 215182,
  [SMALL_STATE(5608)] = 215202,
  [SMALL_STATE(5609)] = 215222,
  [SMALL_STATE(5610)] = 215242,
  [SMALL_STATE(5611)] = 215262,
  [SMALL_STATE(5612)] = 215282,
  [SMALL_STATE(5613)] = 215300,
  [SMALL_STATE(5614)] = 215320,
  [SMALL_STATE(5615)] = 215340,
  [SMALL_STATE(5616)] = 215360,
  [SMALL_STATE(5617)] = 215380,
  [SMALL_STATE(5618)] = 215400,
  [SMALL_STATE(5619)] = 215420,
  [SMALL_STATE(5620)] = 215440,
  [SMALL_STATE(5621)] = 215460,
  [SMALL_STATE(5622)] = 215478,
  [SMALL_STATE(5623)] = 215498,
  [SMALL_STATE(5624)] = 215516,
  [SMALL_STATE(5625)] = 215536,
  [SMALL_STATE(5626)] = 215556,
  [SMALL_STATE(5627)] = 215576,
  [SMALL_STATE(5628)] = 215596,
  [SMALL_STATE(5629)] = 215616,
  [SMALL_STATE(5630)] = 215636,
  [SMALL_STATE(5631)] = 215656,
  [SMALL_STATE(5632)] = 215676,
  [SMALL_STATE(5633)] = 215696,
  [SMALL_STATE(5634)] = 215716,
  [SMALL_STATE(5635)] = 215736,
  [SMALL_STATE(5636)] = 215756,
  [SMALL_STATE(5637)] = 215776,
  [SMALL_STATE(5638)] = 215796,
  [SMALL_STATE(5639)] = 215816,
  [SMALL_STATE(5640)] = 215834,
  [SMALL_STATE(5641)] = 215854,
  [SMALL_STATE(5642)] = 215874,
  [SMALL_STATE(5643)] = 215892,
  [SMALL_STATE(5644)] = 215910,
  [SMALL_STATE(5645)] = 215930,
  [SMALL_STATE(5646)] = 215950,
  [SMALL_STATE(5647)] = 215970,
  [SMALL_STATE(5648)] = 215988,
  [SMALL_STATE(5649)] = 216006,
  [SMALL_STATE(5650)] = 216026,
  [SMALL_STATE(5651)] = 216046,
  [SMALL_STATE(5652)] = 216066,
  [SMALL_STATE(5653)] = 216086,
  [SMALL_STATE(5654)] = 216106,
  [SMALL_STATE(5655)] = 216126,
  [SMALL_STATE(5656)] = 216146,
  [SMALL_STATE(5657)] = 216166,
  [SMALL_STATE(5658)] = 216186,
  [SMALL_STATE(5659)] = 216206,
  [SMALL_STATE(5660)] = 216226,
  [SMALL_STATE(5661)] = 216246,
  [SMALL_STATE(5662)] = 216264,
  [SMALL_STATE(5663)] = 216282,
  [SMALL_STATE(5664)] = 216300,
  [SMALL_STATE(5665)] = 216318,
  [SMALL_STATE(5666)] = 216338,
  [SMALL_STATE(5667)] = 216358,
  [SMALL_STATE(5668)] = 216376,
  [SMALL_STATE(5669)] = 216394,
  [SMALL_STATE(5670)] = 216414,
  [SMALL_STATE(5671)] = 216434,
  [SMALL_STATE(5672)] = 216454,
  [SMALL_STATE(5673)] = 216474,
  [SMALL_STATE(5674)] = 216494,
  [SMALL_STATE(5675)] = 216512,
  [SMALL_STATE(5676)] = 216532,
  [SMALL_STATE(5677)] = 216552,
  [SMALL_STATE(5678)] = 216572,
  [SMALL_STATE(5679)] = 216592,
  [SMALL_STATE(5680)] = 216608,
  [SMALL_STATE(5681)] = 216628,
  [SMALL_STATE(5682)] = 216648,
  [SMALL_STATE(5683)] = 216668,
  [SMALL_STATE(5684)] = 216688,
  [SMALL_STATE(5685)] = 216706,
  [SMALL_STATE(5686)] = 216726,
  [SMALL_STATE(5687)] = 216746,
  [SMALL_STATE(5688)] = 216766,
  [SMALL_STATE(5689)] = 216786,
  [SMALL_STATE(5690)] = 216804,
  [SMALL_STATE(5691)] = 216824,
  [SMALL_STATE(5692)] = 216844,
  [SMALL_STATE(5693)] = 216864,
  [SMALL_STATE(5694)] = 216882,
  [SMALL_STATE(5695)] = 216900,
  [SMALL_STATE(5696)] = 216920,
  [SMALL_STATE(5697)] = 216940,
  [SMALL_STATE(5698)] = 216960,
  [SMALL_STATE(5699)] = 216976,
  [SMALL_STATE(5700)] = 216996,
  [SMALL_STATE(5701)] = 217014,
  [SMALL_STATE(5702)] = 217034,
  [SMALL_STATE(5703)] = 217052,
  [SMALL_STATE(5704)] = 217072,
  [SMALL_STATE(5705)] = 217092,
  [SMALL_STATE(5706)] = 217110,
  [SMALL_STATE(5707)] = 217130,
  [SMALL_STATE(5708)] = 217150,
  [SMALL_STATE(5709)] = 217170,
  [SMALL_STATE(5710)] = 217190,
  [SMALL_STATE(5711)] = 217210,
  [SMALL_STATE(5712)] = 217230,
  [SMALL_STATE(5713)] = 217250,
  [SMALL_STATE(5714)] = 217270,
  [SMALL_STATE(5715)] = 217290,
  [SMALL_STATE(5716)] = 217310,
  [SMALL_STATE(5717)] = 217328,
  [SMALL_STATE(5718)] = 217348,
  [SMALL_STATE(5719)] = 217368,
  [SMALL_STATE(5720)] = 217388,
  [SMALL_STATE(5721)] = 217408,
  [SMALL_STATE(5722)] = 217428,
  [SMALL_STATE(5723)] = 217448,
  [SMALL_STATE(5724)] = 217468,
  [SMALL_STATE(5725)] = 217488,
  [SMALL_STATE(5726)] = 217508,
  [SMALL_STATE(5727)] = 217528,
  [SMALL_STATE(5728)] = 217546,
  [SMALL_STATE(5729)] = 217566,
  [SMALL_STATE(5730)] = 217586,
  [SMALL_STATE(5731)] = 217606,
  [SMALL_STATE(5732)] = 217624,
  [SMALL_STATE(5733)] = 217644,
  [SMALL_STATE(5734)] = 217664,
  [SMALL_STATE(5735)] = 217682,
  [SMALL_STATE(5736)] = 217700,
  [SMALL_STATE(5737)] = 217720,
  [SMALL_STATE(5738)] = 217740,
  [SMALL_STATE(5739)] = 217760,
  [SMALL_STATE(5740)] = 217780,
  [SMALL_STATE(5741)] = 217800,
  [SMALL_STATE(5742)] = 217818,
  [SMALL_STATE(5743)] = 217836,
  [SMALL_STATE(5744)] = 217856,
  [SMALL_STATE(5745)] = 217876,
  [SMALL_STATE(5746)] = 217896,
  [SMALL_STATE(5747)] = 217914,
  [SMALL_STATE(5748)] = 217934,
  [SMALL_STATE(5749)] = 217952,
  [SMALL_STATE(5750)] = 217972,
  [SMALL_STATE(5751)] = 217992,
  [SMALL_STATE(5752)] = 218012,
  [SMALL_STATE(5753)] = 218030,
  [SMALL_STATE(5754)] = 218046,
  [SMALL_STATE(5755)] = 218066,
  [SMALL_STATE(5756)] = 218086,
  [SMALL_STATE(5757)] = 218102,
  [SMALL_STATE(5758)] = 218122,
  [SMALL_STATE(5759)] = 218140,
  [SMALL_STATE(5760)] = 218160,
  [SMALL_STATE(5761)] = 218178,
  [SMALL_STATE(5762)] = 218198,
  [SMALL_STATE(5763)] = 218218,
  [SMALL_STATE(5764)] = 218234,
  [SMALL_STATE(5765)] = 218252,
  [SMALL_STATE(5766)] = 218272,
  [SMALL_STATE(5767)] = 218292,
  [SMALL_STATE(5768)] = 218310,
  [SMALL_STATE(5769)] = 218330,
  [SMALL_STATE(5770)] = 218350,
  [SMALL_STATE(5771)] = 218370,
  [SMALL_STATE(5772)] = 218388,
  [SMALL_STATE(5773)] = 218408,
  [SMALL_STATE(5774)] = 218426,
  [SMALL_STATE(5775)] = 218446,
  [SMALL_STATE(5776)] = 218466,
  [SMALL_STATE(5777)] = 218484,
  [SMALL_STATE(5778)] = 218504,
  [SMALL_STATE(5779)] = 218522,
  [SMALL_STATE(5780)] = 218542,
  [SMALL_STATE(5781)] = 218562,
  [SMALL_STATE(5782)] = 218580,
  [SMALL_STATE(5783)] = 218598,
  [SMALL_STATE(5784)] = 218618,
  [SMALL_STATE(5785)] = 218636,
  [SMALL_STATE(5786)] = 218656,
  [SMALL_STATE(5787)] = 218674,
  [SMALL_STATE(5788)] = 218694,
  [SMALL_STATE(5789)] = 218712,
  [SMALL_STATE(5790)] = 218732,
  [SMALL_STATE(5791)] = 218750,
  [SMALL_STATE(5792)] = 218770,
  [SMALL_STATE(5793)] = 218788,
  [SMALL_STATE(5794)] = 218808,
  [SMALL_STATE(5795)] = 218826,
  [SMALL_STATE(5796)] = 218846,
  [SMALL_STATE(5797)] = 218862,
  [SMALL_STATE(5798)] = 218882,
  [SMALL_STATE(5799)] = 218902,
  [SMALL_STATE(5800)] = 218922,
  [SMALL_STATE(5801)] = 218942,
  [SMALL_STATE(5802)] = 218962,
  [SMALL_STATE(5803)] = 218982,
  [SMALL_STATE(5804)] = 219002,
  [SMALL_STATE(5805)] = 219022,
  [SMALL_STATE(5806)] = 219038,
  [SMALL_STATE(5807)] = 219056,
  [SMALL_STATE(5808)] = 219076,
  [SMALL_STATE(5809)] = 219096,
  [SMALL_STATE(5810)] = 219116,
  [SMALL_STATE(5811)] = 219132,
  [SMALL_STATE(5812)] = 219152,
  [SMALL_STATE(5813)] = 219172,
  [SMALL_STATE(5814)] = 219192,
  [SMALL_STATE(5815)] = 219210,
  [SMALL_STATE(5816)] = 219228,
  [SMALL_STATE(5817)] = 219248,
  [SMALL_STATE(5818)] = 219268,
  [SMALL_STATE(5819)] = 219288,
  [SMALL_STATE(5820)] = 219308,
  [SMALL_STATE(5821)] = 219328,
  [SMALL_STATE(5822)] = 219348,
  [SMALL_STATE(5823)] = 219368,
  [SMALL_STATE(5824)] = 219388,
  [SMALL_STATE(5825)] = 219408,
  [SMALL_STATE(5826)] = 219428,
  [SMALL_STATE(5827)] = 219448,
  [SMALL_STATE(5828)] = 219468,
  [SMALL_STATE(5829)] = 219486,
  [SMALL_STATE(5830)] = 219504,
  [SMALL_STATE(5831)] = 219524,
  [SMALL_STATE(5832)] = 219544,
  [SMALL_STATE(5833)] = 219564,
  [SMALL_STATE(5834)] = 219584,
  [SMALL_STATE(5835)] = 219604,
  [SMALL_STATE(5836)] = 219624,
  [SMALL_STATE(5837)] = 219644,
  [SMALL_STATE(5838)] = 219664,
  [SMALL_STATE(5839)] = 219684,
  [SMALL_STATE(5840)] = 219702,
  [SMALL_STATE(5841)] = 219720,
  [SMALL_STATE(5842)] = 219738,
  [SMALL_STATE(5843)] = 219758,
  [SMALL_STATE(5844)] = 219778,
  [SMALL_STATE(5845)] = 219798,
  [SMALL_STATE(5846)] = 219818,
  [SMALL_STATE(5847)] = 219838,
  [SMALL_STATE(5848)] = 219858,
  [SMALL_STATE(5849)] = 219876,
  [SMALL_STATE(5850)] = 219894,
  [SMALL_STATE(5851)] = 219912,
  [SMALL_STATE(5852)] = 219932,
  [SMALL_STATE(5853)] = 219950,
  [SMALL_STATE(5854)] = 219970,
  [SMALL_STATE(5855)] = 219990,
  [SMALL_STATE(5856)] = 220010,
  [SMALL_STATE(5857)] = 220028,
  [SMALL_STATE(5858)] = 220048,
  [SMALL_STATE(5859)] = 220068,
  [SMALL_STATE(5860)] = 220088,
  [SMALL_STATE(5861)] = 220108,
  [SMALL_STATE(5862)] = 220128,
  [SMALL_STATE(5863)] = 220148,
  [SMALL_STATE(5864)] = 220166,
  [SMALL_STATE(5865)] = 220186,
  [SMALL_STATE(5866)] = 220206,
  [SMALL_STATE(5867)] = 220226,
  [SMALL_STATE(5868)] = 220244,
  [SMALL_STATE(5869)] = 220262,
  [SMALL_STATE(5870)] = 220280,
  [SMALL_STATE(5871)] = 220300,
  [SMALL_STATE(5872)] = 220318,
  [SMALL_STATE(5873)] = 220338,
  [SMALL_STATE(5874)] = 220356,
  [SMALL_STATE(5875)] = 220376,
  [SMALL_STATE(5876)] = 220396,
  [SMALL_STATE(5877)] = 220416,
  [SMALL_STATE(5878)] = 220434,
  [SMALL_STATE(5879)] = 220454,
  [SMALL_STATE(5880)] = 220474,
  [SMALL_STATE(5881)] = 220492,
  [SMALL_STATE(5882)] = 220512,
  [SMALL_STATE(5883)] = 220532,
  [SMALL_STATE(5884)] = 220552,
  [SMALL_STATE(5885)] = 220570,
  [SMALL_STATE(5886)] = 220590,
  [SMALL_STATE(5887)] = 220608,
  [SMALL_STATE(5888)] = 220626,
  [SMALL_STATE(5889)] = 220646,
  [SMALL_STATE(5890)] = 220664,
  [SMALL_STATE(5891)] = 220684,
  [SMALL_STATE(5892)] = 220702,
  [SMALL_STATE(5893)] = 220722,
  [SMALL_STATE(5894)] = 220740,
  [SMALL_STATE(5895)] = 220760,
  [SMALL_STATE(5896)] = 220778,
  [SMALL_STATE(5897)] = 220798,
  [SMALL_STATE(5898)] = 220816,
  [SMALL_STATE(5899)] = 220836,
  [SMALL_STATE(5900)] = 220854,
  [SMALL_STATE(5901)] = 220872,
  [SMALL_STATE(5902)] = 220890,
  [SMALL_STATE(5903)] = 220910,
  [SMALL_STATE(5904)] = 220930,
  [SMALL_STATE(5905)] = 220950,
  [SMALL_STATE(5906)] = 220970,
  [SMALL_STATE(5907)] = 220990,
  [SMALL_STATE(5908)] = 221010,
  [SMALL_STATE(5909)] = 221030,
  [SMALL_STATE(5910)] = 221050,
  [SMALL_STATE(5911)] = 221066,
  [SMALL_STATE(5912)] = 221086,
  [SMALL_STATE(5913)] = 221106,
  [SMALL_STATE(5914)] = 221126,
  [SMALL_STATE(5915)] = 221146,
  [SMALL_STATE(5916)] = 221166,
  [SMALL_STATE(5917)] = 221186,
  [SMALL_STATE(5918)] = 221206,
  [SMALL_STATE(5919)] = 221226,
  [SMALL_STATE(5920)] = 221242,
  [SMALL_STATE(5921)] = 221262,
  [SMALL_STATE(5922)] = 221282,
  [SMALL_STATE(5923)] = 221302,
  [SMALL_STATE(5924)] = 221322,
  [SMALL_STATE(5925)] = 221340,
  [SMALL_STATE(5926)] = 221358,
  [SMALL_STATE(5927)] = 221378,
  [SMALL_STATE(5928)] = 221398,
  [SMALL_STATE(5929)] = 221418,
  [SMALL_STATE(5930)] = 221438,
  [SMALL_STATE(5931)] = 221458,
  [SMALL_STATE(5932)] = 221476,
  [SMALL_STATE(5933)] = 221494,
  [SMALL_STATE(5934)] = 221514,
  [SMALL_STATE(5935)] = 221530,
  [SMALL_STATE(5936)] = 221550,
  [SMALL_STATE(5937)] = 221568,
  [SMALL_STATE(5938)] = 221588,
  [SMALL_STATE(5939)] = 221608,
  [SMALL_STATE(5940)] = 221628,
  [SMALL_STATE(5941)] = 221648,
  [SMALL_STATE(5942)] = 221668,
  [SMALL_STATE(5943)] = 221688,
  [SMALL_STATE(5944)] = 221708,
  [SMALL_STATE(5945)] = 221728,
  [SMALL_STATE(5946)] = 221745,
  [SMALL_STATE(5947)] = 221762,
  [SMALL_STATE(5948)] = 221777,
  [SMALL_STATE(5949)] = 221794,
  [SMALL_STATE(5950)] = 221811,
  [SMALL_STATE(5951)] = 221828,
  [SMALL_STATE(5952)] = 221845,
  [SMALL_STATE(5953)] = 221862,
  [SMALL_STATE(5954)] = 221879,
  [SMALL_STATE(5955)] = 221894,
  [SMALL_STATE(5956)] = 221911,
  [SMALL_STATE(5957)] = 221928,
  [SMALL_STATE(5958)] = 221945,
  [SMALL_STATE(5959)] = 221962,
  [SMALL_STATE(5960)] = 221979,
  [SMALL_STATE(5961)] = 221996,
  [SMALL_STATE(5962)] = 222013,
  [SMALL_STATE(5963)] = 222030,
  [SMALL_STATE(5964)] = 222047,
  [SMALL_STATE(5965)] = 222064,
  [SMALL_STATE(5966)] = 222081,
  [SMALL_STATE(5967)] = 222098,
  [SMALL_STATE(5968)] = 222115,
  [SMALL_STATE(5969)] = 222132,
  [SMALL_STATE(5970)] = 222149,
  [SMALL_STATE(5971)] = 222166,
  [SMALL_STATE(5972)] = 222181,
  [SMALL_STATE(5973)] = 222198,
  [SMALL_STATE(5974)] = 222215,
  [SMALL_STATE(5975)] = 222232,
  [SMALL_STATE(5976)] = 222249,
  [SMALL_STATE(5977)] = 222264,
  [SMALL_STATE(5978)] = 222281,
  [SMALL_STATE(5979)] = 222298,
  [SMALL_STATE(5980)] = 222315,
  [SMALL_STATE(5981)] = 222332,
  [SMALL_STATE(5982)] = 222349,
  [SMALL_STATE(5983)] = 222366,
  [SMALL_STATE(5984)] = 222381,
  [SMALL_STATE(5985)] = 222396,
  [SMALL_STATE(5986)] = 222413,
  [SMALL_STATE(5987)] = 222430,
  [SMALL_STATE(5988)] = 222447,
  [SMALL_STATE(5989)] = 222464,
  [SMALL_STATE(5990)] = 222479,
  [SMALL_STATE(5991)] = 222496,
  [SMALL_STATE(5992)] = 222513,
  [SMALL_STATE(5993)] = 222530,
  [SMALL_STATE(5994)] = 222545,
  [SMALL_STATE(5995)] = 222560,
  [SMALL_STATE(5996)] = 222577,
  [SMALL_STATE(5997)] = 222594,
  [SMALL_STATE(5998)] = 222611,
  [SMALL_STATE(5999)] = 222628,
  [SMALL_STATE(6000)] = 222645,
  [SMALL_STATE(6001)] = 222660,
  [SMALL_STATE(6002)] = 222677,
  [SMALL_STATE(6003)] = 222694,
  [SMALL_STATE(6004)] = 222709,
  [SMALL_STATE(6005)] = 222724,
  [SMALL_STATE(6006)] = 222741,
  [SMALL_STATE(6007)] = 222758,
  [SMALL_STATE(6008)] = 222775,
  [SMALL_STATE(6009)] = 222792,
  [SMALL_STATE(6010)] = 222809,
  [SMALL_STATE(6011)] = 222826,
  [SMALL_STATE(6012)] = 222843,
  [SMALL_STATE(6013)] = 222860,
  [SMALL_STATE(6014)] = 222877,
  [SMALL_STATE(6015)] = 222892,
  [SMALL_STATE(6016)] = 222909,
  [SMALL_STATE(6017)] = 222926,
  [SMALL_STATE(6018)] = 222943,
  [SMALL_STATE(6019)] = 222960,
  [SMALL_STATE(6020)] = 222977,
  [SMALL_STATE(6021)] = 222994,
  [SMALL_STATE(6022)] = 223011,
  [SMALL_STATE(6023)] = 223028,
  [SMALL_STATE(6024)] = 223045,
  [SMALL_STATE(6025)] = 223062,
  [SMALL_STATE(6026)] = 223079,
  [SMALL_STATE(6027)] = 223096,
  [SMALL_STATE(6028)] = 223113,
  [SMALL_STATE(6029)] = 223128,
  [SMALL_STATE(6030)] = 223145,
  [SMALL_STATE(6031)] = 223162,
  [SMALL_STATE(6032)] = 223179,
  [SMALL_STATE(6033)] = 223196,
  [SMALL_STATE(6034)] = 223213,
  [SMALL_STATE(6035)] = 223230,
  [SMALL_STATE(6036)] = 223247,
  [SMALL_STATE(6037)] = 223264,
  [SMALL_STATE(6038)] = 223281,
  [SMALL_STATE(6039)] = 223298,
  [SMALL_STATE(6040)] = 223315,
  [SMALL_STATE(6041)] = 223332,
  [SMALL_STATE(6042)] = 223349,
  [SMALL_STATE(6043)] = 223366,
  [SMALL_STATE(6044)] = 223383,
  [SMALL_STATE(6045)] = 223400,
  [SMALL_STATE(6046)] = 223417,
  [SMALL_STATE(6047)] = 223434,
  [SMALL_STATE(6048)] = 223449,
  [SMALL_STATE(6049)] = 223466,
  [SMALL_STATE(6050)] = 223481,
  [SMALL_STATE(6051)] = 223498,
  [SMALL_STATE(6052)] = 223515,
  [SMALL_STATE(6053)] = 223532,
  [SMALL_STATE(6054)] = 223549,
  [SMALL_STATE(6055)] = 223566,
  [SMALL_STATE(6056)] = 223583,
  [SMALL_STATE(6057)] = 223600,
  [SMALL_STATE(6058)] = 223617,
  [SMALL_STATE(6059)] = 223634,
  [SMALL_STATE(6060)] = 223651,
  [SMALL_STATE(6061)] = 223666,
  [SMALL_STATE(6062)] = 223683,
  [SMALL_STATE(6063)] = 223700,
  [SMALL_STATE(6064)] = 223717,
  [SMALL_STATE(6065)] = 223734,
  [SMALL_STATE(6066)] = 223749,
  [SMALL_STATE(6067)] = 223766,
  [SMALL_STATE(6068)] = 223781,
  [SMALL_STATE(6069)] = 223798,
  [SMALL_STATE(6070)] = 223815,
  [SMALL_STATE(6071)] = 223832,
  [SMALL_STATE(6072)] = 223849,
  [SMALL_STATE(6073)] = 223866,
  [SMALL_STATE(6074)] = 223883,
  [SMALL_STATE(6075)] = 223900,
  [SMALL_STATE(6076)] = 223917,
  [SMALL_STATE(6077)] = 223932,
  [SMALL_STATE(6078)] = 223949,
  [SMALL_STATE(6079)] = 223966,
  [SMALL_STATE(6080)] = 223983,
  [SMALL_STATE(6081)] = 223998,
  [SMALL_STATE(6082)] = 224015,
  [SMALL_STATE(6083)] = 224032,
  [SMALL_STATE(6084)] = 224049,
  [SMALL_STATE(6085)] = 224066,
  [SMALL_STATE(6086)] = 224083,
  [SMALL_STATE(6087)] = 224100,
  [SMALL_STATE(6088)] = 224117,
  [SMALL_STATE(6089)] = 224134,
  [SMALL_STATE(6090)] = 224151,
  [SMALL_STATE(6091)] = 224166,
  [SMALL_STATE(6092)] = 224183,
  [SMALL_STATE(6093)] = 224200,
  [SMALL_STATE(6094)] = 224217,
  [SMALL_STATE(6095)] = 224234,
  [SMALL_STATE(6096)] = 224251,
  [SMALL_STATE(6097)] = 224268,
  [SMALL_STATE(6098)] = 224285,
  [SMALL_STATE(6099)] = 224300,
  [SMALL_STATE(6100)] = 224317,
  [SMALL_STATE(6101)] = 224332,
  [SMALL_STATE(6102)] = 224349,
  [SMALL_STATE(6103)] = 224366,
  [SMALL_STATE(6104)] = 224383,
  [SMALL_STATE(6105)] = 224400,
  [SMALL_STATE(6106)] = 224417,
  [SMALL_STATE(6107)] = 224434,
  [SMALL_STATE(6108)] = 224451,
  [SMALL_STATE(6109)] = 224468,
  [SMALL_STATE(6110)] = 224483,
  [SMALL_STATE(6111)] = 224498,
  [SMALL_STATE(6112)] = 224513,
  [SMALL_STATE(6113)] = 224530,
  [SMALL_STATE(6114)] = 224545,
  [SMALL_STATE(6115)] = 224560,
  [SMALL_STATE(6116)] = 224575,
  [SMALL_STATE(6117)] = 224590,
  [SMALL_STATE(6118)] = 224607,
  [SMALL_STATE(6119)] = 224622,
  [SMALL_STATE(6120)] = 224639,
  [SMALL_STATE(6121)] = 224656,
  [SMALL_STATE(6122)] = 224673,
  [SMALL_STATE(6123)] = 224688,
  [SMALL_STATE(6124)] = 224705,
  [SMALL_STATE(6125)] = 224720,
  [SMALL_STATE(6126)] = 224737,
  [SMALL_STATE(6127)] = 224752,
  [SMALL_STATE(6128)] = 224769,
  [SMALL_STATE(6129)] = 224786,
  [SMALL_STATE(6130)] = 224803,
  [SMALL_STATE(6131)] = 224818,
  [SMALL_STATE(6132)] = 224835,
  [SMALL_STATE(6133)] = 224852,
  [SMALL_STATE(6134)] = 224867,
  [SMALL_STATE(6135)] = 224884,
  [SMALL_STATE(6136)] = 224901,
  [SMALL_STATE(6137)] = 224916,
  [SMALL_STATE(6138)] = 224931,
  [SMALL_STATE(6139)] = 224948,
  [SMALL_STATE(6140)] = 224965,
  [SMALL_STATE(6141)] = 224982,
  [SMALL_STATE(6142)] = 224999,
  [SMALL_STATE(6143)] = 225016,
  [SMALL_STATE(6144)] = 225033,
  [SMALL_STATE(6145)] = 225050,
  [SMALL_STATE(6146)] = 225067,
  [SMALL_STATE(6147)] = 225084,
  [SMALL_STATE(6148)] = 225101,
  [SMALL_STATE(6149)] = 225116,
  [SMALL_STATE(6150)] = 225131,
  [SMALL_STATE(6151)] = 225148,
  [SMALL_STATE(6152)] = 225165,
  [SMALL_STATE(6153)] = 225182,
  [SMALL_STATE(6154)] = 225199,
  [SMALL_STATE(6155)] = 225214,
  [SMALL_STATE(6156)] = 225231,
  [SMALL_STATE(6157)] = 225248,
  [SMALL_STATE(6158)] = 225265,
  [SMALL_STATE(6159)] = 225282,
  [SMALL_STATE(6160)] = 225299,
  [SMALL_STATE(6161)] = 225316,
  [SMALL_STATE(6162)] = 225333,
  [SMALL_STATE(6163)] = 225350,
  [SMALL_STATE(6164)] = 225367,
  [SMALL_STATE(6165)] = 225384,
  [SMALL_STATE(6166)] = 225399,
  [SMALL_STATE(6167)] = 225416,
  [SMALL_STATE(6168)] = 225433,
  [SMALL_STATE(6169)] = 225450,
  [SMALL_STATE(6170)] = 225467,
  [SMALL_STATE(6171)] = 225482,
  [SMALL_STATE(6172)] = 225499,
  [SMALL_STATE(6173)] = 225514,
  [SMALL_STATE(6174)] = 225531,
  [SMALL_STATE(6175)] = 225548,
  [SMALL_STATE(6176)] = 225565,
  [SMALL_STATE(6177)] = 225580,
  [SMALL_STATE(6178)] = 225597,
  [SMALL_STATE(6179)] = 225614,
  [SMALL_STATE(6180)] = 225629,
  [SMALL_STATE(6181)] = 225646,
  [SMALL_STATE(6182)] = 225663,
  [SMALL_STATE(6183)] = 225680,
  [SMALL_STATE(6184)] = 225697,
  [SMALL_STATE(6185)] = 225714,
  [SMALL_STATE(6186)] = 225729,
  [SMALL_STATE(6187)] = 225746,
  [SMALL_STATE(6188)] = 225763,
  [SMALL_STATE(6189)] = 225780,
  [SMALL_STATE(6190)] = 225795,
  [SMALL_STATE(6191)] = 225810,
  [SMALL_STATE(6192)] = 225827,
  [SMALL_STATE(6193)] = 225842,
  [SMALL_STATE(6194)] = 225859,
  [SMALL_STATE(6195)] = 225876,
  [SMALL_STATE(6196)] = 225893,
  [SMALL_STATE(6197)] = 225910,
  [SMALL_STATE(6198)] = 225927,
  [SMALL_STATE(6199)] = 225944,
  [SMALL_STATE(6200)] = 225961,
  [SMALL_STATE(6201)] = 225978,
  [SMALL_STATE(6202)] = 225995,
  [SMALL_STATE(6203)] = 226012,
  [SMALL_STATE(6204)] = 226027,
  [SMALL_STATE(6205)] = 226042,
  [SMALL_STATE(6206)] = 226059,
  [SMALL_STATE(6207)] = 226076,
  [SMALL_STATE(6208)] = 226091,
  [SMALL_STATE(6209)] = 226106,
  [SMALL_STATE(6210)] = 226123,
  [SMALL_STATE(6211)] = 226140,
  [SMALL_STATE(6212)] = 226157,
  [SMALL_STATE(6213)] = 226174,
  [SMALL_STATE(6214)] = 226191,
  [SMALL_STATE(6215)] = 226206,
  [SMALL_STATE(6216)] = 226223,
  [SMALL_STATE(6217)] = 226238,
  [SMALL_STATE(6218)] = 226255,
  [SMALL_STATE(6219)] = 226270,
  [SMALL_STATE(6220)] = 226287,
  [SMALL_STATE(6221)] = 226304,
  [SMALL_STATE(6222)] = 226321,
  [SMALL_STATE(6223)] = 226338,
  [SMALL_STATE(6224)] = 226355,
  [SMALL_STATE(6225)] = 226370,
  [SMALL_STATE(6226)] = 226387,
  [SMALL_STATE(6227)] = 226404,
  [SMALL_STATE(6228)] = 226421,
  [SMALL_STATE(6229)] = 226438,
  [SMALL_STATE(6230)] = 226455,
  [SMALL_STATE(6231)] = 226472,
  [SMALL_STATE(6232)] = 226489,
  [SMALL_STATE(6233)] = 226506,
  [SMALL_STATE(6234)] = 226523,
  [SMALL_STATE(6235)] = 226540,
  [SMALL_STATE(6236)] = 226557,
  [SMALL_STATE(6237)] = 226574,
  [SMALL_STATE(6238)] = 226591,
  [SMALL_STATE(6239)] = 226606,
  [SMALL_STATE(6240)] = 226623,
  [SMALL_STATE(6241)] = 226640,
  [SMALL_STATE(6242)] = 226657,
  [SMALL_STATE(6243)] = 226674,
  [SMALL_STATE(6244)] = 226691,
  [SMALL_STATE(6245)] = 226706,
  [SMALL_STATE(6246)] = 226723,
  [SMALL_STATE(6247)] = 226740,
  [SMALL_STATE(6248)] = 226757,
  [SMALL_STATE(6249)] = 226774,
  [SMALL_STATE(6250)] = 226791,
  [SMALL_STATE(6251)] = 226808,
  [SMALL_STATE(6252)] = 226825,
  [SMALL_STATE(6253)] = 226842,
  [SMALL_STATE(6254)] = 226859,
  [SMALL_STATE(6255)] = 226876,
  [SMALL_STATE(6256)] = 226891,
  [SMALL_STATE(6257)] = 226908,
  [SMALL_STATE(6258)] = 226925,
  [SMALL_STATE(6259)] = 226940,
  [SMALL_STATE(6260)] = 226957,
  [SMALL_STATE(6261)] = 226974,
  [SMALL_STATE(6262)] = 226991,
  [SMALL_STATE(6263)] = 227008,
  [SMALL_STATE(6264)] = 227025,
  [SMALL_STATE(6265)] = 227042,
  [SMALL_STATE(6266)] = 227059,
  [SMALL_STATE(6267)] = 227076,
  [SMALL_STATE(6268)] = 227093,
  [SMALL_STATE(6269)] = 227108,
  [SMALL_STATE(6270)] = 227123,
  [SMALL_STATE(6271)] = 227140,
  [SMALL_STATE(6272)] = 227157,
  [SMALL_STATE(6273)] = 227174,
  [SMALL_STATE(6274)] = 227191,
  [SMALL_STATE(6275)] = 227208,
  [SMALL_STATE(6276)] = 227225,
  [SMALL_STATE(6277)] = 227240,
  [SMALL_STATE(6278)] = 227257,
  [SMALL_STATE(6279)] = 227274,
  [SMALL_STATE(6280)] = 227291,
  [SMALL_STATE(6281)] = 227306,
  [SMALL_STATE(6282)] = 227323,
  [SMALL_STATE(6283)] = 227340,
  [SMALL_STATE(6284)] = 227355,
  [SMALL_STATE(6285)] = 227372,
  [SMALL_STATE(6286)] = 227389,
  [SMALL_STATE(6287)] = 227406,
  [SMALL_STATE(6288)] = 227423,
  [SMALL_STATE(6289)] = 227440,
  [SMALL_STATE(6290)] = 227457,
  [SMALL_STATE(6291)] = 227474,
  [SMALL_STATE(6292)] = 227491,
  [SMALL_STATE(6293)] = 227508,
  [SMALL_STATE(6294)] = 227525,
  [SMALL_STATE(6295)] = 227542,
  [SMALL_STATE(6296)] = 227559,
  [SMALL_STATE(6297)] = 227574,
  [SMALL_STATE(6298)] = 227591,
  [SMALL_STATE(6299)] = 227606,
  [SMALL_STATE(6300)] = 227623,
  [SMALL_STATE(6301)] = 227640,
  [SMALL_STATE(6302)] = 227657,
  [SMALL_STATE(6303)] = 227674,
  [SMALL_STATE(6304)] = 227691,
  [SMALL_STATE(6305)] = 227708,
  [SMALL_STATE(6306)] = 227723,
  [SMALL_STATE(6307)] = 227740,
  [SMALL_STATE(6308)] = 227755,
  [SMALL_STATE(6309)] = 227772,
  [SMALL_STATE(6310)] = 227789,
  [SMALL_STATE(6311)] = 227806,
  [SMALL_STATE(6312)] = 227821,
  [SMALL_STATE(6313)] = 227838,
  [SMALL_STATE(6314)] = 227855,
  [SMALL_STATE(6315)] = 227870,
  [SMALL_STATE(6316)] = 227887,
  [SMALL_STATE(6317)] = 227904,
  [SMALL_STATE(6318)] = 227921,
  [SMALL_STATE(6319)] = 227936,
  [SMALL_STATE(6320)] = 227953,
  [SMALL_STATE(6321)] = 227970,
  [SMALL_STATE(6322)] = 227987,
  [SMALL_STATE(6323)] = 228004,
  [SMALL_STATE(6324)] = 228021,
  [SMALL_STATE(6325)] = 228038,
  [SMALL_STATE(6326)] = 228055,
  [SMALL_STATE(6327)] = 228072,
  [SMALL_STATE(6328)] = 228089,
  [SMALL_STATE(6329)] = 228106,
  [SMALL_STATE(6330)] = 228123,
  [SMALL_STATE(6331)] = 228140,
  [SMALL_STATE(6332)] = 228157,
  [SMALL_STATE(6333)] = 228174,
  [SMALL_STATE(6334)] = 228189,
  [SMALL_STATE(6335)] = 228203,
  [SMALL_STATE(6336)] = 228217,
  [SMALL_STATE(6337)] = 228231,
  [SMALL_STATE(6338)] = 228245,
  [SMALL_STATE(6339)] = 228259,
  [SMALL_STATE(6340)] = 228273,
  [SMALL_STATE(6341)] = 228287,
  [SMALL_STATE(6342)] = 228301,
  [SMALL_STATE(6343)] = 228315,
  [SMALL_STATE(6344)] = 228329,
  [SMALL_STATE(6345)] = 228343,
  [SMALL_STATE(6346)] = 228357,
  [SMALL_STATE(6347)] = 228371,
  [SMALL_STATE(6348)] = 228385,
  [SMALL_STATE(6349)] = 228399,
  [SMALL_STATE(6350)] = 228413,
  [SMALL_STATE(6351)] = 228427,
  [SMALL_STATE(6352)] = 228441,
  [SMALL_STATE(6353)] = 228455,
  [SMALL_STATE(6354)] = 228469,
  [SMALL_STATE(6355)] = 228483,
  [SMALL_STATE(6356)] = 228497,
  [SMALL_STATE(6357)] = 228511,
  [SMALL_STATE(6358)] = 228525,
  [SMALL_STATE(6359)] = 228539,
  [SMALL_STATE(6360)] = 228553,
  [SMALL_STATE(6361)] = 228567,
  [SMALL_STATE(6362)] = 228581,
  [SMALL_STATE(6363)] = 228595,
  [SMALL_STATE(6364)] = 228609,
  [SMALL_STATE(6365)] = 228623,
  [SMALL_STATE(6366)] = 228637,
  [SMALL_STATE(6367)] = 228651,
  [SMALL_STATE(6368)] = 228665,
  [SMALL_STATE(6369)] = 228679,
  [SMALL_STATE(6370)] = 228693,
  [SMALL_STATE(6371)] = 228707,
  [SMALL_STATE(6372)] = 228721,
  [SMALL_STATE(6373)] = 228735,
  [SMALL_STATE(6374)] = 228749,
  [SMALL_STATE(6375)] = 228763,
  [SMALL_STATE(6376)] = 228777,
  [SMALL_STATE(6377)] = 228791,
  [SMALL_STATE(6378)] = 228805,
  [SMALL_STATE(6379)] = 228819,
  [SMALL_STATE(6380)] = 228833,
  [SMALL_STATE(6381)] = 228847,
  [SMALL_STATE(6382)] = 228861,
  [SMALL_STATE(6383)] = 228875,
  [SMALL_STATE(6384)] = 228889,
  [SMALL_STATE(6385)] = 228903,
  [SMALL_STATE(6386)] = 228917,
  [SMALL_STATE(6387)] = 228931,
  [SMALL_STATE(6388)] = 228945,
  [SMALL_STATE(6389)] = 228959,
  [SMALL_STATE(6390)] = 228973,
  [SMALL_STATE(6391)] = 228987,
  [SMALL_STATE(6392)] = 229001,
  [SMALL_STATE(6393)] = 229015,
  [SMALL_STATE(6394)] = 229029,
  [SMALL_STATE(6395)] = 229043,
  [SMALL_STATE(6396)] = 229057,
  [SMALL_STATE(6397)] = 229071,
  [SMALL_STATE(6398)] = 229085,
  [SMALL_STATE(6399)] = 229099,
  [SMALL_STATE(6400)] = 229113,
  [SMALL_STATE(6401)] = 229127,
  [SMALL_STATE(6402)] = 229141,
  [SMALL_STATE(6403)] = 229155,
  [SMALL_STATE(6404)] = 229169,
  [SMALL_STATE(6405)] = 229183,
  [SMALL_STATE(6406)] = 229197,
  [SMALL_STATE(6407)] = 229211,
  [SMALL_STATE(6408)] = 229225,
  [SMALL_STATE(6409)] = 229239,
  [SMALL_STATE(6410)] = 229253,
  [SMALL_STATE(6411)] = 229267,
  [SMALL_STATE(6412)] = 229281,
  [SMALL_STATE(6413)] = 229295,
  [SMALL_STATE(6414)] = 229309,
  [SMALL_STATE(6415)] = 229323,
  [SMALL_STATE(6416)] = 229337,
  [SMALL_STATE(6417)] = 229351,
  [SMALL_STATE(6418)] = 229365,
  [SMALL_STATE(6419)] = 229379,
  [SMALL_STATE(6420)] = 229393,
  [SMALL_STATE(6421)] = 229407,
  [SMALL_STATE(6422)] = 229421,
  [SMALL_STATE(6423)] = 229435,
  [SMALL_STATE(6424)] = 229449,
  [SMALL_STATE(6425)] = 229463,
  [SMALL_STATE(6426)] = 229477,
  [SMALL_STATE(6427)] = 229491,
  [SMALL_STATE(6428)] = 229505,
  [SMALL_STATE(6429)] = 229519,
  [SMALL_STATE(6430)] = 229533,
  [SMALL_STATE(6431)] = 229547,
  [SMALL_STATE(6432)] = 229561,
  [SMALL_STATE(6433)] = 229575,
  [SMALL_STATE(6434)] = 229589,
  [SMALL_STATE(6435)] = 229603,
  [SMALL_STATE(6436)] = 229617,
  [SMALL_STATE(6437)] = 229631,
  [SMALL_STATE(6438)] = 229645,
  [SMALL_STATE(6439)] = 229659,
  [SMALL_STATE(6440)] = 229673,
  [SMALL_STATE(6441)] = 229687,
  [SMALL_STATE(6442)] = 229701,
  [SMALL_STATE(6443)] = 229715,
  [SMALL_STATE(6444)] = 229729,
  [SMALL_STATE(6445)] = 229743,
  [SMALL_STATE(6446)] = 229757,
  [SMALL_STATE(6447)] = 229771,
  [SMALL_STATE(6448)] = 229785,
  [SMALL_STATE(6449)] = 229799,
  [SMALL_STATE(6450)] = 229813,
  [SMALL_STATE(6451)] = 229827,
  [SMALL_STATE(6452)] = 229841,
  [SMALL_STATE(6453)] = 229855,
  [SMALL_STATE(6454)] = 229869,
  [SMALL_STATE(6455)] = 229883,
  [SMALL_STATE(6456)] = 229897,
  [SMALL_STATE(6457)] = 229911,
  [SMALL_STATE(6458)] = 229925,
  [SMALL_STATE(6459)] = 229939,
  [SMALL_STATE(6460)] = 229953,
  [SMALL_STATE(6461)] = 229967,
  [SMALL_STATE(6462)] = 229981,
  [SMALL_STATE(6463)] = 229995,
  [SMALL_STATE(6464)] = 230009,
  [SMALL_STATE(6465)] = 230023,
  [SMALL_STATE(6466)] = 230037,
  [SMALL_STATE(6467)] = 230051,
  [SMALL_STATE(6468)] = 230065,
  [SMALL_STATE(6469)] = 230079,
  [SMALL_STATE(6470)] = 230093,
  [SMALL_STATE(6471)] = 230107,
  [SMALL_STATE(6472)] = 230121,
  [SMALL_STATE(6473)] = 230135,
  [SMALL_STATE(6474)] = 230149,
  [SMALL_STATE(6475)] = 230163,
  [SMALL_STATE(6476)] = 230177,
  [SMALL_STATE(6477)] = 230191,
  [SMALL_STATE(6478)] = 230205,
  [SMALL_STATE(6479)] = 230219,
  [SMALL_STATE(6480)] = 230233,
  [SMALL_STATE(6481)] = 230247,
  [SMALL_STATE(6482)] = 230261,
  [SMALL_STATE(6483)] = 230275,
  [SMALL_STATE(6484)] = 230289,
  [SMALL_STATE(6485)] = 230303,
  [SMALL_STATE(6486)] = 230317,
  [SMALL_STATE(6487)] = 230331,
  [SMALL_STATE(6488)] = 230345,
  [SMALL_STATE(6489)] = 230359,
  [SMALL_STATE(6490)] = 230373,
  [SMALL_STATE(6491)] = 230387,
  [SMALL_STATE(6492)] = 230401,
  [SMALL_STATE(6493)] = 230415,
  [SMALL_STATE(6494)] = 230429,
  [SMALL_STATE(6495)] = 230443,
  [SMALL_STATE(6496)] = 230457,
  [SMALL_STATE(6497)] = 230471,
  [SMALL_STATE(6498)] = 230485,
  [SMALL_STATE(6499)] = 230499,
  [SMALL_STATE(6500)] = 230513,
  [SMALL_STATE(6501)] = 230527,
  [SMALL_STATE(6502)] = 230541,
  [SMALL_STATE(6503)] = 230555,
  [SMALL_STATE(6504)] = 230569,
  [SMALL_STATE(6505)] = 230583,
  [SMALL_STATE(6506)] = 230597,
  [SMALL_STATE(6507)] = 230611,
  [SMALL_STATE(6508)] = 230625,
  [SMALL_STATE(6509)] = 230639,
  [SMALL_STATE(6510)] = 230653,
  [SMALL_STATE(6511)] = 230667,
  [SMALL_STATE(6512)] = 230681,
  [SMALL_STATE(6513)] = 230695,
  [SMALL_STATE(6514)] = 230709,
  [SMALL_STATE(6515)] = 230723,
  [SMALL_STATE(6516)] = 230737,
  [SMALL_STATE(6517)] = 230751,
  [SMALL_STATE(6518)] = 230765,
  [SMALL_STATE(6519)] = 230779,
  [SMALL_STATE(6520)] = 230793,
  [SMALL_STATE(6521)] = 230807,
  [SMALL_STATE(6522)] = 230821,
  [SMALL_STATE(6523)] = 230835,
  [SMALL_STATE(6524)] = 230849,
  [SMALL_STATE(6525)] = 230863,
  [SMALL_STATE(6526)] = 230877,
  [SMALL_STATE(6527)] = 230891,
  [SMALL_STATE(6528)] = 230905,
  [SMALL_STATE(6529)] = 230919,
  [SMALL_STATE(6530)] = 230933,
  [SMALL_STATE(6531)] = 230947,
  [SMALL_STATE(6532)] = 230961,
  [SMALL_STATE(6533)] = 230975,
  [SMALL_STATE(6534)] = 230989,
  [SMALL_STATE(6535)] = 231003,
  [SMALL_STATE(6536)] = 231017,
  [SMALL_STATE(6537)] = 231031,
  [SMALL_STATE(6538)] = 231045,
  [SMALL_STATE(6539)] = 231059,
  [SMALL_STATE(6540)] = 231073,
  [SMALL_STATE(6541)] = 231087,
  [SMALL_STATE(6542)] = 231101,
  [SMALL_STATE(6543)] = 231115,
  [SMALL_STATE(6544)] = 231129,
  [SMALL_STATE(6545)] = 231143,
  [SMALL_STATE(6546)] = 231157,
  [SMALL_STATE(6547)] = 231171,
  [SMALL_STATE(6548)] = 231185,
  [SMALL_STATE(6549)] = 231199,
  [SMALL_STATE(6550)] = 231213,
  [SMALL_STATE(6551)] = 231227,
  [SMALL_STATE(6552)] = 231241,
  [SMALL_STATE(6553)] = 231255,
  [SMALL_STATE(6554)] = 231269,
  [SMALL_STATE(6555)] = 231283,
  [SMALL_STATE(6556)] = 231297,
  [SMALL_STATE(6557)] = 231311,
  [SMALL_STATE(6558)] = 231325,
  [SMALL_STATE(6559)] = 231339,
  [SMALL_STATE(6560)] = 231353,
  [SMALL_STATE(6561)] = 231367,
  [SMALL_STATE(6562)] = 231381,
  [SMALL_STATE(6563)] = 231395,
  [SMALL_STATE(6564)] = 231409,
  [SMALL_STATE(6565)] = 231423,
  [SMALL_STATE(6566)] = 231437,
  [SMALL_STATE(6567)] = 231451,
  [SMALL_STATE(6568)] = 231465,
  [SMALL_STATE(6569)] = 231479,
  [SMALL_STATE(6570)] = 231493,
  [SMALL_STATE(6571)] = 231507,
  [SMALL_STATE(6572)] = 231521,
  [SMALL_STATE(6573)] = 231535,
  [SMALL_STATE(6574)] = 231549,
  [SMALL_STATE(6575)] = 231563,
  [SMALL_STATE(6576)] = 231577,
  [SMALL_STATE(6577)] = 231591,
  [SMALL_STATE(6578)] = 231605,
  [SMALL_STATE(6579)] = 231619,
  [SMALL_STATE(6580)] = 231633,
  [SMALL_STATE(6581)] = 231647,
  [SMALL_STATE(6582)] = 231661,
  [SMALL_STATE(6583)] = 231675,
  [SMALL_STATE(6584)] = 231689,
  [SMALL_STATE(6585)] = 231703,
  [SMALL_STATE(6586)] = 231717,
  [SMALL_STATE(6587)] = 231731,
  [SMALL_STATE(6588)] = 231745,
  [SMALL_STATE(6589)] = 231759,
  [SMALL_STATE(6590)] = 231773,
  [SMALL_STATE(6591)] = 231787,
  [SMALL_STATE(6592)] = 231801,
  [SMALL_STATE(6593)] = 231815,
  [SMALL_STATE(6594)] = 231829,
  [SMALL_STATE(6595)] = 231843,
  [SMALL_STATE(6596)] = 231857,
  [SMALL_STATE(6597)] = 231871,
  [SMALL_STATE(6598)] = 231885,
  [SMALL_STATE(6599)] = 231899,
  [SMALL_STATE(6600)] = 231913,
  [SMALL_STATE(6601)] = 231927,
  [SMALL_STATE(6602)] = 231941,
  [SMALL_STATE(6603)] = 231955,
  [SMALL_STATE(6604)] = 231969,
  [SMALL_STATE(6605)] = 231983,
  [SMALL_STATE(6606)] = 231997,
  [SMALL_STATE(6607)] = 232011,
  [SMALL_STATE(6608)] = 232025,
  [SMALL_STATE(6609)] = 232039,
  [SMALL_STATE(6610)] = 232053,
  [SMALL_STATE(6611)] = 232067,
  [SMALL_STATE(6612)] = 232081,
  [SMALL_STATE(6613)] = 232095,
  [SMALL_STATE(6614)] = 232109,
  [SMALL_STATE(6615)] = 232123,
  [SMALL_STATE(6616)] = 232137,
  [SMALL_STATE(6617)] = 232151,
  [SMALL_STATE(6618)] = 232165,
  [SMALL_STATE(6619)] = 232179,
  [SMALL_STATE(6620)] = 232193,
  [SMALL_STATE(6621)] = 232207,
  [SMALL_STATE(6622)] = 232221,
  [SMALL_STATE(6623)] = 232235,
  [SMALL_STATE(6624)] = 232249,
  [SMALL_STATE(6625)] = 232263,
  [SMALL_STATE(6626)] = 232277,
  [SMALL_STATE(6627)] = 232291,
  [SMALL_STATE(6628)] = 232305,
  [SMALL_STATE(6629)] = 232319,
  [SMALL_STATE(6630)] = 232333,
  [SMALL_STATE(6631)] = 232347,
  [SMALL_STATE(6632)] = 232361,
  [SMALL_STATE(6633)] = 232375,
  [SMALL_STATE(6634)] = 232389,
  [SMALL_STATE(6635)] = 232403,
  [SMALL_STATE(6636)] = 232417,
  [SMALL_STATE(6637)] = 232431,
  [SMALL_STATE(6638)] = 232445,
  [SMALL_STATE(6639)] = 232459,
  [SMALL_STATE(6640)] = 232473,
  [SMALL_STATE(6641)] = 232487,
  [SMALL_STATE(6642)] = 232501,
  [SMALL_STATE(6643)] = 232515,
  [SMALL_STATE(6644)] = 232529,
  [SMALL_STATE(6645)] = 232543,
  [SMALL_STATE(6646)] = 232557,
  [SMALL_STATE(6647)] = 232571,
  [SMALL_STATE(6648)] = 232585,
  [SMALL_STATE(6649)] = 232599,
  [SMALL_STATE(6650)] = 232613,
  [SMALL_STATE(6651)] = 232627,
  [SMALL_STATE(6652)] = 232641,
  [SMALL_STATE(6653)] = 232655,
  [SMALL_STATE(6654)] = 232669,
  [SMALL_STATE(6655)] = 232683,
  [SMALL_STATE(6656)] = 232697,
  [SMALL_STATE(6657)] = 232711,
  [SMALL_STATE(6658)] = 232725,
  [SMALL_STATE(6659)] = 232739,
  [SMALL_STATE(6660)] = 232753,
  [SMALL_STATE(6661)] = 232767,
  [SMALL_STATE(6662)] = 232781,
  [SMALL_STATE(6663)] = 232795,
  [SMALL_STATE(6664)] = 232809,
  [SMALL_STATE(6665)] = 232823,
  [SMALL_STATE(6666)] = 232837,
  [SMALL_STATE(6667)] = 232851,
  [SMALL_STATE(6668)] = 232865,
  [SMALL_STATE(6669)] = 232879,
  [SMALL_STATE(6670)] = 232893,
  [SMALL_STATE(6671)] = 232907,
  [SMALL_STATE(6672)] = 232921,
  [SMALL_STATE(6673)] = 232935,
  [SMALL_STATE(6674)] = 232949,
  [SMALL_STATE(6675)] = 232963,
  [SMALL_STATE(6676)] = 232977,
  [SMALL_STATE(6677)] = 232991,
  [SMALL_STATE(6678)] = 233005,
  [SMALL_STATE(6679)] = 233019,
  [SMALL_STATE(6680)] = 233033,
  [SMALL_STATE(6681)] = 233047,
  [SMALL_STATE(6682)] = 233061,
  [SMALL_STATE(6683)] = 233075,
  [SMALL_STATE(6684)] = 233089,
  [SMALL_STATE(6685)] = 233103,
  [SMALL_STATE(6686)] = 233117,
  [SMALL_STATE(6687)] = 233131,
  [SMALL_STATE(6688)] = 233145,
  [SMALL_STATE(6689)] = 233159,
  [SMALL_STATE(6690)] = 233173,
  [SMALL_STATE(6691)] = 233187,
  [SMALL_STATE(6692)] = 233201,
  [SMALL_STATE(6693)] = 233215,
  [SMALL_STATE(6694)] = 233229,
  [SMALL_STATE(6695)] = 233243,
  [SMALL_STATE(6696)] = 233257,
  [SMALL_STATE(6697)] = 233271,
  [SMALL_STATE(6698)] = 233285,
  [SMALL_STATE(6699)] = 233299,
  [SMALL_STATE(6700)] = 233313,
  [SMALL_STATE(6701)] = 233327,
  [SMALL_STATE(6702)] = 233341,
  [SMALL_STATE(6703)] = 233355,
  [SMALL_STATE(6704)] = 233369,
  [SMALL_STATE(6705)] = 233383,
  [SMALL_STATE(6706)] = 233397,
  [SMALL_STATE(6707)] = 233411,
  [SMALL_STATE(6708)] = 233425,
  [SMALL_STATE(6709)] = 233439,
  [SMALL_STATE(6710)] = 233453,
  [SMALL_STATE(6711)] = 233467,
  [SMALL_STATE(6712)] = 233481,
  [SMALL_STATE(6713)] = 233495,
  [SMALL_STATE(6714)] = 233509,
  [SMALL_STATE(6715)] = 233523,
  [SMALL_STATE(6716)] = 233537,
  [SMALL_STATE(6717)] = 233551,
  [SMALL_STATE(6718)] = 233565,
  [SMALL_STATE(6719)] = 233579,
  [SMALL_STATE(6720)] = 233593,
  [SMALL_STATE(6721)] = 233607,
  [SMALL_STATE(6722)] = 233621,
  [SMALL_STATE(6723)] = 233635,
  [SMALL_STATE(6724)] = 233649,
  [SMALL_STATE(6725)] = 233663,
  [SMALL_STATE(6726)] = 233677,
  [SMALL_STATE(6727)] = 233691,
  [SMALL_STATE(6728)] = 233705,
  [SMALL_STATE(6729)] = 233719,
  [SMALL_STATE(6730)] = 233733,
  [SMALL_STATE(6731)] = 233747,
  [SMALL_STATE(6732)] = 233761,
  [SMALL_STATE(6733)] = 233775,
  [SMALL_STATE(6734)] = 233789,
  [SMALL_STATE(6735)] = 233803,
  [SMALL_STATE(6736)] = 233817,
  [SMALL_STATE(6737)] = 233831,
  [SMALL_STATE(6738)] = 233845,
  [SMALL_STATE(6739)] = 233859,
  [SMALL_STATE(6740)] = 233873,
  [SMALL_STATE(6741)] = 233887,
  [SMALL_STATE(6742)] = 233901,
  [SMALL_STATE(6743)] = 233915,
  [SMALL_STATE(6744)] = 233929,
  [SMALL_STATE(6745)] = 233943,
  [SMALL_STATE(6746)] = 233957,
  [SMALL_STATE(6747)] = 233971,
  [SMALL_STATE(6748)] = 233985,
  [SMALL_STATE(6749)] = 233999,
  [SMALL_STATE(6750)] = 234013,
  [SMALL_STATE(6751)] = 234027,
  [SMALL_STATE(6752)] = 234041,
  [SMALL_STATE(6753)] = 234055,
  [SMALL_STATE(6754)] = 234069,
  [SMALL_STATE(6755)] = 234083,
  [SMALL_STATE(6756)] = 234097,
  [SMALL_STATE(6757)] = 234111,
  [SMALL_STATE(6758)] = 234125,
  [SMALL_STATE(6759)] = 234139,
  [SMALL_STATE(6760)] = 234153,
  [SMALL_STATE(6761)] = 234167,
  [SMALL_STATE(6762)] = 234181,
  [SMALL_STATE(6763)] = 234195,
  [SMALL_STATE(6764)] = 234209,
  [SMALL_STATE(6765)] = 234223,
  [SMALL_STATE(6766)] = 234237,
  [SMALL_STATE(6767)] = 234251,
  [SMALL_STATE(6768)] = 234265,
  [SMALL_STATE(6769)] = 234279,
  [SMALL_STATE(6770)] = 234293,
  [SMALL_STATE(6771)] = 234307,
  [SMALL_STATE(6772)] = 234321,
  [SMALL_STATE(6773)] = 234335,
  [SMALL_STATE(6774)] = 234349,
  [SMALL_STATE(6775)] = 234363,
  [SMALL_STATE(6776)] = 234377,
  [SMALL_STATE(6777)] = 234391,
  [SMALL_STATE(6778)] = 234405,
  [SMALL_STATE(6779)] = 234419,
  [SMALL_STATE(6780)] = 234433,
  [SMALL_STATE(6781)] = 234447,
  [SMALL_STATE(6782)] = 234461,
  [SMALL_STATE(6783)] = 234475,
  [SMALL_STATE(6784)] = 234489,
  [SMALL_STATE(6785)] = 234503,
  [SMALL_STATE(6786)] = 234517,
  [SMALL_STATE(6787)] = 234531,
  [SMALL_STATE(6788)] = 234545,
  [SMALL_STATE(6789)] = 234559,
  [SMALL_STATE(6790)] = 234573,
  [SMALL_STATE(6791)] = 234587,
  [SMALL_STATE(6792)] = 234601,
  [SMALL_STATE(6793)] = 234615,
  [SMALL_STATE(6794)] = 234629,
  [SMALL_STATE(6795)] = 234643,
  [SMALL_STATE(6796)] = 234657,
  [SMALL_STATE(6797)] = 234671,
  [SMALL_STATE(6798)] = 234685,
  [SMALL_STATE(6799)] = 234699,
  [SMALL_STATE(6800)] = 234713,
  [SMALL_STATE(6801)] = 234727,
  [SMALL_STATE(6802)] = 234741,
  [SMALL_STATE(6803)] = 234755,
  [SMALL_STATE(6804)] = 234769,
  [SMALL_STATE(6805)] = 234783,
  [SMALL_STATE(6806)] = 234797,
  [SMALL_STATE(6807)] = 234811,
  [SMALL_STATE(6808)] = 234825,
  [SMALL_STATE(6809)] = 234839,
  [SMALL_STATE(6810)] = 234853,
  [SMALL_STATE(6811)] = 234867,
  [SMALL_STATE(6812)] = 234881,
  [SMALL_STATE(6813)] = 234895,
  [SMALL_STATE(6814)] = 234909,
  [SMALL_STATE(6815)] = 234923,
  [SMALL_STATE(6816)] = 234937,
  [SMALL_STATE(6817)] = 234951,
  [SMALL_STATE(6818)] = 234965,
  [SMALL_STATE(6819)] = 234979,
  [SMALL_STATE(6820)] = 234993,
  [SMALL_STATE(6821)] = 235007,
  [SMALL_STATE(6822)] = 235021,
  [SMALL_STATE(6823)] = 235035,
  [SMALL_STATE(6824)] = 235049,
  [SMALL_STATE(6825)] = 235063,
  [SMALL_STATE(6826)] = 235077,
  [SMALL_STATE(6827)] = 235091,
  [SMALL_STATE(6828)] = 235105,
  [SMALL_STATE(6829)] = 235119,
  [SMALL_STATE(6830)] = 235133,
  [SMALL_STATE(6831)] = 235147,
  [SMALL_STATE(6832)] = 235161,
  [SMALL_STATE(6833)] = 235175,
  [SMALL_STATE(6834)] = 235189,
  [SMALL_STATE(6835)] = 235203,
  [SMALL_STATE(6836)] = 235217,
  [SMALL_STATE(6837)] = 235231,
  [SMALL_STATE(6838)] = 235245,
  [SMALL_STATE(6839)] = 235259,
  [SMALL_STATE(6840)] = 235273,
  [SMALL_STATE(6841)] = 235287,
  [SMALL_STATE(6842)] = 235301,
  [SMALL_STATE(6843)] = 235315,
  [SMALL_STATE(6844)] = 235329,
  [SMALL_STATE(6845)] = 235343,
  [SMALL_STATE(6846)] = 235357,
  [SMALL_STATE(6847)] = 235371,
  [SMALL_STATE(6848)] = 235385,
  [SMALL_STATE(6849)] = 235399,
  [SMALL_STATE(6850)] = 235413,
  [SMALL_STATE(6851)] = 235427,
  [SMALL_STATE(6852)] = 235441,
  [SMALL_STATE(6853)] = 235455,
  [SMALL_STATE(6854)] = 235469,
  [SMALL_STATE(6855)] = 235483,
  [SMALL_STATE(6856)] = 235497,
  [SMALL_STATE(6857)] = 235511,
  [SMALL_STATE(6858)] = 235525,
  [SMALL_STATE(6859)] = 235539,
  [SMALL_STATE(6860)] = 235553,
  [SMALL_STATE(6861)] = 235567,
  [SMALL_STATE(6862)] = 235581,
  [SMALL_STATE(6863)] = 235595,
  [SMALL_STATE(6864)] = 235609,
  [SMALL_STATE(6865)] = 235623,
  [SMALL_STATE(6866)] = 235637,
  [SMALL_STATE(6867)] = 235651,
  [SMALL_STATE(6868)] = 235665,
  [SMALL_STATE(6869)] = 235679,
  [SMALL_STATE(6870)] = 235693,
  [SMALL_STATE(6871)] = 235707,
  [SMALL_STATE(6872)] = 235721,
  [SMALL_STATE(6873)] = 235735,
  [SMALL_STATE(6874)] = 235749,
  [SMALL_STATE(6875)] = 235763,
  [SMALL_STATE(6876)] = 235777,
  [SMALL_STATE(6877)] = 235791,
  [SMALL_STATE(6878)] = 235805,
  [SMALL_STATE(6879)] = 235819,
  [SMALL_STATE(6880)] = 235833,
  [SMALL_STATE(6881)] = 235847,
  [SMALL_STATE(6882)] = 235861,
  [SMALL_STATE(6883)] = 235875,
  [SMALL_STATE(6884)] = 235889,
  [SMALL_STATE(6885)] = 235903,
  [SMALL_STATE(6886)] = 235917,
  [SMALL_STATE(6887)] = 235931,
  [SMALL_STATE(6888)] = 235945,
  [SMALL_STATE(6889)] = 235959,
  [SMALL_STATE(6890)] = 235973,
  [SMALL_STATE(6891)] = 235987,
  [SMALL_STATE(6892)] = 236001,
  [SMALL_STATE(6893)] = 236015,
  [SMALL_STATE(6894)] = 236029,
  [SMALL_STATE(6895)] = 236043,
  [SMALL_STATE(6896)] = 236057,
  [SMALL_STATE(6897)] = 236071,
  [SMALL_STATE(6898)] = 236085,
  [SMALL_STATE(6899)] = 236099,
  [SMALL_STATE(6900)] = 236113,
  [SMALL_STATE(6901)] = 236127,
  [SMALL_STATE(6902)] = 236141,
  [SMALL_STATE(6903)] = 236155,
  [SMALL_STATE(6904)] = 236169,
  [SMALL_STATE(6905)] = 236183,
  [SMALL_STATE(6906)] = 236197,
  [SMALL_STATE(6907)] = 236211,
  [SMALL_STATE(6908)] = 236225,
  [SMALL_STATE(6909)] = 236239,
  [SMALL_STATE(6910)] = 236253,
  [SMALL_STATE(6911)] = 236267,
  [SMALL_STATE(6912)] = 236281,
  [SMALL_STATE(6913)] = 236295,
  [SMALL_STATE(6914)] = 236309,
  [SMALL_STATE(6915)] = 236323,
  [SMALL_STATE(6916)] = 236337,
  [SMALL_STATE(6917)] = 236351,
  [SMALL_STATE(6918)] = 236365,
  [SMALL_STATE(6919)] = 236379,
  [SMALL_STATE(6920)] = 236393,
  [SMALL_STATE(6921)] = 236407,
  [SMALL_STATE(6922)] = 236421,
  [SMALL_STATE(6923)] = 236435,
  [SMALL_STATE(6924)] = 236449,
  [SMALL_STATE(6925)] = 236463,
  [SMALL_STATE(6926)] = 236477,
  [SMALL_STATE(6927)] = 236491,
  [SMALL_STATE(6928)] = 236505,
  [SMALL_STATE(6929)] = 236519,
  [SMALL_STATE(6930)] = 236533,
  [SMALL_STATE(6931)] = 236547,
  [SMALL_STATE(6932)] = 236561,
  [SMALL_STATE(6933)] = 236575,
  [SMALL_STATE(6934)] = 236589,
  [SMALL_STATE(6935)] = 236603,
  [SMALL_STATE(6936)] = 236617,
  [SMALL_STATE(6937)] = 236631,
  [SMALL_STATE(6938)] = 236645,
  [SMALL_STATE(6939)] = 236659,
  [SMALL_STATE(6940)] = 236673,
  [SMALL_STATE(6941)] = 236687,
  [SMALL_STATE(6942)] = 236701,
  [SMALL_STATE(6943)] = 236715,
  [SMALL_STATE(6944)] = 236729,
  [SMALL_STATE(6945)] = 236743,
  [SMALL_STATE(6946)] = 236757,
  [SMALL_STATE(6947)] = 236771,
  [SMALL_STATE(6948)] = 236785,
  [SMALL_STATE(6949)] = 236799,
  [SMALL_STATE(6950)] = 236813,
  [SMALL_STATE(6951)] = 236827,
  [SMALL_STATE(6952)] = 236841,
  [SMALL_STATE(6953)] = 236855,
  [SMALL_STATE(6954)] = 236869,
  [SMALL_STATE(6955)] = 236883,
  [SMALL_STATE(6956)] = 236897,
  [SMALL_STATE(6957)] = 236911,
  [SMALL_STATE(6958)] = 236925,
  [SMALL_STATE(6959)] = 236939,
  [SMALL_STATE(6960)] = 236953,
  [SMALL_STATE(6961)] = 236967,
  [SMALL_STATE(6962)] = 236981,
  [SMALL_STATE(6963)] = 236995,
  [SMALL_STATE(6964)] = 237009,
  [SMALL_STATE(6965)] = 237023,
  [SMALL_STATE(6966)] = 237037,
  [SMALL_STATE(6967)] = 237051,
  [SMALL_STATE(6968)] = 237065,
  [SMALL_STATE(6969)] = 237079,
  [SMALL_STATE(6970)] = 237093,
  [SMALL_STATE(6971)] = 237107,
  [SMALL_STATE(6972)] = 237121,
  [SMALL_STATE(6973)] = 237135,
  [SMALL_STATE(6974)] = 237149,
  [SMALL_STATE(6975)] = 237163,
  [SMALL_STATE(6976)] = 237177,
  [SMALL_STATE(6977)] = 237191,
  [SMALL_STATE(6978)] = 237205,
  [SMALL_STATE(6979)] = 237219,
  [SMALL_STATE(6980)] = 237233,
  [SMALL_STATE(6981)] = 237247,
  [SMALL_STATE(6982)] = 237261,
  [SMALL_STATE(6983)] = 237275,
  [SMALL_STATE(6984)] = 237289,
  [SMALL_STATE(6985)] = 237303,
  [SMALL_STATE(6986)] = 237317,
  [SMALL_STATE(6987)] = 237331,
  [SMALL_STATE(6988)] = 237345,
  [SMALL_STATE(6989)] = 237359,
  [SMALL_STATE(6990)] = 237373,
  [SMALL_STATE(6991)] = 237387,
  [SMALL_STATE(6992)] = 237401,
  [SMALL_STATE(6993)] = 237415,
  [SMALL_STATE(6994)] = 237429,
  [SMALL_STATE(6995)] = 237443,
  [SMALL_STATE(6996)] = 237457,
  [SMALL_STATE(6997)] = 237471,
  [SMALL_STATE(6998)] = 237485,
  [SMALL_STATE(6999)] = 237499,
  [SMALL_STATE(7000)] = 237513,
  [SMALL_STATE(7001)] = 237527,
  [SMALL_STATE(7002)] = 237541,
  [SMALL_STATE(7003)] = 237555,
  [SMALL_STATE(7004)] = 237569,
  [SMALL_STATE(7005)] = 237583,
  [SMALL_STATE(7006)] = 237597,
  [SMALL_STATE(7007)] = 237611,
  [SMALL_STATE(7008)] = 237625,
  [SMALL_STATE(7009)] = 237639,
  [SMALL_STATE(7010)] = 237653,
  [SMALL_STATE(7011)] = 237667,
  [SMALL_STATE(7012)] = 237681,
  [SMALL_STATE(7013)] = 237695,
  [SMALL_STATE(7014)] = 237709,
  [SMALL_STATE(7015)] = 237723,
  [SMALL_STATE(7016)] = 237737,
  [SMALL_STATE(7017)] = 237751,
  [SMALL_STATE(7018)] = 237765,
  [SMALL_STATE(7019)] = 237779,
  [SMALL_STATE(7020)] = 237793,
  [SMALL_STATE(7021)] = 237807,
  [SMALL_STATE(7022)] = 237821,
  [SMALL_STATE(7023)] = 237835,
  [SMALL_STATE(7024)] = 237849,
  [SMALL_STATE(7025)] = 237863,
  [SMALL_STATE(7026)] = 237877,
  [SMALL_STATE(7027)] = 237891,
  [SMALL_STATE(7028)] = 237905,
  [SMALL_STATE(7029)] = 237919,
  [SMALL_STATE(7030)] = 237933,
  [SMALL_STATE(7031)] = 237947,
  [SMALL_STATE(7032)] = 237961,
  [SMALL_STATE(7033)] = 237975,
  [SMALL_STATE(7034)] = 237989,
  [SMALL_STATE(7035)] = 238003,
  [SMALL_STATE(7036)] = 238017,
  [SMALL_STATE(7037)] = 238031,
  [SMALL_STATE(7038)] = 238045,
  [SMALL_STATE(7039)] = 238059,
  [SMALL_STATE(7040)] = 238073,
  [SMALL_STATE(7041)] = 238087,
  [SMALL_STATE(7042)] = 238101,
  [SMALL_STATE(7043)] = 238115,
  [SMALL_STATE(7044)] = 238129,
  [SMALL_STATE(7045)] = 238143,
  [SMALL_STATE(7046)] = 238157,
  [SMALL_STATE(7047)] = 238171,
  [SMALL_STATE(7048)] = 238185,
  [SMALL_STATE(7049)] = 238199,
  [SMALL_STATE(7050)] = 238213,
  [SMALL_STATE(7051)] = 238227,
  [SMALL_STATE(7052)] = 238241,
  [SMALL_STATE(7053)] = 238255,
  [SMALL_STATE(7054)] = 238269,
  [SMALL_STATE(7055)] = 238283,
  [SMALL_STATE(7056)] = 238297,
  [SMALL_STATE(7057)] = 238311,
  [SMALL_STATE(7058)] = 238325,
  [SMALL_STATE(7059)] = 238339,
  [SMALL_STATE(7060)] = 238353,
  [SMALL_STATE(7061)] = 238367,
  [SMALL_STATE(7062)] = 238381,
  [SMALL_STATE(7063)] = 238395,
  [SMALL_STATE(7064)] = 238409,
  [SMALL_STATE(7065)] = 238423,
  [SMALL_STATE(7066)] = 238437,
  [SMALL_STATE(7067)] = 238451,
  [SMALL_STATE(7068)] = 238465,
  [SMALL_STATE(7069)] = 238479,
  [SMALL_STATE(7070)] = 238493,
  [SMALL_STATE(7071)] = 238507,
  [SMALL_STATE(7072)] = 238521,
  [SMALL_STATE(7073)] = 238535,
  [SMALL_STATE(7074)] = 238549,
  [SMALL_STATE(7075)] = 238563,
  [SMALL_STATE(7076)] = 238577,
  [SMALL_STATE(7077)] = 238591,
  [SMALL_STATE(7078)] = 238605,
  [SMALL_STATE(7079)] = 238619,
  [SMALL_STATE(7080)] = 238633,
  [SMALL_STATE(7081)] = 238647,
  [SMALL_STATE(7082)] = 238661,
  [SMALL_STATE(7083)] = 238675,
  [SMALL_STATE(7084)] = 238689,
  [SMALL_STATE(7085)] = 238703,
  [SMALL_STATE(7086)] = 238717,
  [SMALL_STATE(7087)] = 238731,
  [SMALL_STATE(7088)] = 238745,
  [SMALL_STATE(7089)] = 238759,
  [SMALL_STATE(7090)] = 238773,
  [SMALL_STATE(7091)] = 238787,
  [SMALL_STATE(7092)] = 238801,
  [SMALL_STATE(7093)] = 238815,
  [SMALL_STATE(7094)] = 238829,
  [SMALL_STATE(7095)] = 238843,
  [SMALL_STATE(7096)] = 238857,
  [SMALL_STATE(7097)] = 238871,
  [SMALL_STATE(7098)] = 238885,
  [SMALL_STATE(7099)] = 238899,
  [SMALL_STATE(7100)] = 238913,
  [SMALL_STATE(7101)] = 238927,
  [SMALL_STATE(7102)] = 238941,
  [SMALL_STATE(7103)] = 238955,
  [SMALL_STATE(7104)] = 238969,
  [SMALL_STATE(7105)] = 238983,
  [SMALL_STATE(7106)] = 238997,
  [SMALL_STATE(7107)] = 239011,
  [SMALL_STATE(7108)] = 239025,
  [SMALL_STATE(7109)] = 239039,
  [SMALL_STATE(7110)] = 239053,
  [SMALL_STATE(7111)] = 239067,
  [SMALL_STATE(7112)] = 239081,
  [SMALL_STATE(7113)] = 239095,
  [SMALL_STATE(7114)] = 239109,
  [SMALL_STATE(7115)] = 239123,
  [SMALL_STATE(7116)] = 239137,
  [SMALL_STATE(7117)] = 239151,
  [SMALL_STATE(7118)] = 239165,
  [SMALL_STATE(7119)] = 239179,
  [SMALL_STATE(7120)] = 239193,
  [SMALL_STATE(7121)] = 239207,
  [SMALL_STATE(7122)] = 239221,
  [SMALL_STATE(7123)] = 239235,
  [SMALL_STATE(7124)] = 239249,
  [SMALL_STATE(7125)] = 239263,
  [SMALL_STATE(7126)] = 239277,
  [SMALL_STATE(7127)] = 239291,
  [SMALL_STATE(7128)] = 239305,
  [SMALL_STATE(7129)] = 239319,
  [SMALL_STATE(7130)] = 239333,
  [SMALL_STATE(7131)] = 239347,
  [SMALL_STATE(7132)] = 239361,
  [SMALL_STATE(7133)] = 239375,
  [SMALL_STATE(7134)] = 239389,
  [SMALL_STATE(7135)] = 239403,
  [SMALL_STATE(7136)] = 239417,
  [SMALL_STATE(7137)] = 239431,
  [SMALL_STATE(7138)] = 239445,
  [SMALL_STATE(7139)] = 239459,
  [SMALL_STATE(7140)] = 239473,
  [SMALL_STATE(7141)] = 239487,
  [SMALL_STATE(7142)] = 239501,
  [SMALL_STATE(7143)] = 239515,
  [SMALL_STATE(7144)] = 239529,
  [SMALL_STATE(7145)] = 239543,
  [SMALL_STATE(7146)] = 239557,
  [SMALL_STATE(7147)] = 239571,
  [SMALL_STATE(7148)] = 239585,
  [SMALL_STATE(7149)] = 239599,
  [SMALL_STATE(7150)] = 239613,
  [SMALL_STATE(7151)] = 239627,
  [SMALL_STATE(7152)] = 239641,
  [SMALL_STATE(7153)] = 239655,
  [SMALL_STATE(7154)] = 239669,
  [SMALL_STATE(7155)] = 239683,
  [SMALL_STATE(7156)] = 239697,
  [SMALL_STATE(7157)] = 239711,
  [SMALL_STATE(7158)] = 239725,
  [SMALL_STATE(7159)] = 239739,
  [SMALL_STATE(7160)] = 239753,
  [SMALL_STATE(7161)] = 239767,
  [SMALL_STATE(7162)] = 239781,
  [SMALL_STATE(7163)] = 239795,
  [SMALL_STATE(7164)] = 239809,
  [SMALL_STATE(7165)] = 239823,
  [SMALL_STATE(7166)] = 239837,
  [SMALL_STATE(7167)] = 239851,
  [SMALL_STATE(7168)] = 239865,
  [SMALL_STATE(7169)] = 239879,
  [SMALL_STATE(7170)] = 239893,
  [SMALL_STATE(7171)] = 239907,
  [SMALL_STATE(7172)] = 239921,
  [SMALL_STATE(7173)] = 239935,
  [SMALL_STATE(7174)] = 239949,
  [SMALL_STATE(7175)] = 239963,
  [SMALL_STATE(7176)] = 239977,
  [SMALL_STATE(7177)] = 239991,
  [SMALL_STATE(7178)] = 240005,
  [SMALL_STATE(7179)] = 240019,
  [SMALL_STATE(7180)] = 240033,
  [SMALL_STATE(7181)] = 240047,
  [SMALL_STATE(7182)] = 240061,
  [SMALL_STATE(7183)] = 240075,
  [SMALL_STATE(7184)] = 240089,
  [SMALL_STATE(7185)] = 240103,
  [SMALL_STATE(7186)] = 240117,
  [SMALL_STATE(7187)] = 240131,
  [SMALL_STATE(7188)] = 240145,
  [SMALL_STATE(7189)] = 240159,
  [SMALL_STATE(7190)] = 240173,
  [SMALL_STATE(7191)] = 240187,
  [SMALL_STATE(7192)] = 240201,
  [SMALL_STATE(7193)] = 240215,
  [SMALL_STATE(7194)] = 240229,
  [SMALL_STATE(7195)] = 240243,
  [SMALL_STATE(7196)] = 240257,
  [SMALL_STATE(7197)] = 240271,
  [SMALL_STATE(7198)] = 240285,
  [SMALL_STATE(7199)] = 240299,
  [SMALL_STATE(7200)] = 240313,
  [SMALL_STATE(7201)] = 240327,
  [SMALL_STATE(7202)] = 240341,
  [SMALL_STATE(7203)] = 240355,
  [SMALL_STATE(7204)] = 240369,
  [SMALL_STATE(7205)] = 240383,
  [SMALL_STATE(7206)] = 240397,
  [SMALL_STATE(7207)] = 240411,
  [SMALL_STATE(7208)] = 240425,
  [SMALL_STATE(7209)] = 240439,
  [SMALL_STATE(7210)] = 240453,
  [SMALL_STATE(7211)] = 240467,
  [SMALL_STATE(7212)] = 240481,
  [SMALL_STATE(7213)] = 240495,
  [SMALL_STATE(7214)] = 240509,
  [SMALL_STATE(7215)] = 240523,
  [SMALL_STATE(7216)] = 240537,
  [SMALL_STATE(7217)] = 240551,
  [SMALL_STATE(7218)] = 240565,
  [SMALL_STATE(7219)] = 240579,
  [SMALL_STATE(7220)] = 240593,
  [SMALL_STATE(7221)] = 240607,
  [SMALL_STATE(7222)] = 240621,
  [SMALL_STATE(7223)] = 240635,
  [SMALL_STATE(7224)] = 240649,
  [SMALL_STATE(7225)] = 240663,
  [SMALL_STATE(7226)] = 240677,
  [SMALL_STATE(7227)] = 240691,
  [SMALL_STATE(7228)] = 240705,
  [SMALL_STATE(7229)] = 240719,
  [SMALL_STATE(7230)] = 240733,
  [SMALL_STATE(7231)] = 240747,
  [SMALL_STATE(7232)] = 240761,
  [SMALL_STATE(7233)] = 240775,
  [SMALL_STATE(7234)] = 240789,
  [SMALL_STATE(7235)] = 240803,
  [SMALL_STATE(7236)] = 240817,
  [SMALL_STATE(7237)] = 240831,
  [SMALL_STATE(7238)] = 240845,
  [SMALL_STATE(7239)] = 240859,
  [SMALL_STATE(7240)] = 240873,
  [SMALL_STATE(7241)] = 240887,
  [SMALL_STATE(7242)] = 240901,
  [SMALL_STATE(7243)] = 240915,
  [SMALL_STATE(7244)] = 240929,
  [SMALL_STATE(7245)] = 240943,
  [SMALL_STATE(7246)] = 240957,
  [SMALL_STATE(7247)] = 240971,
  [SMALL_STATE(7248)] = 240985,
  [SMALL_STATE(7249)] = 240999,
  [SMALL_STATE(7250)] = 241013,
  [SMALL_STATE(7251)] = 241027,
  [SMALL_STATE(7252)] = 241041,
  [SMALL_STATE(7253)] = 241055,
  [SMALL_STATE(7254)] = 241069,
  [SMALL_STATE(7255)] = 241083,
  [SMALL_STATE(7256)] = 241097,
  [SMALL_STATE(7257)] = 241111,
  [SMALL_STATE(7258)] = 241125,
  [SMALL_STATE(7259)] = 241139,
  [SMALL_STATE(7260)] = 241153,
  [SMALL_STATE(7261)] = 241167,
  [SMALL_STATE(7262)] = 241181,
  [SMALL_STATE(7263)] = 241195,
  [SMALL_STATE(7264)] = 241209,
  [SMALL_STATE(7265)] = 241223,
  [SMALL_STATE(7266)] = 241237,
  [SMALL_STATE(7267)] = 241251,
  [SMALL_STATE(7268)] = 241265,
  [SMALL_STATE(7269)] = 241279,
  [SMALL_STATE(7270)] = 241293,
  [SMALL_STATE(7271)] = 241307,
  [SMALL_STATE(7272)] = 241321,
  [SMALL_STATE(7273)] = 241335,
  [SMALL_STATE(7274)] = 241349,
  [SMALL_STATE(7275)] = 241363,
  [SMALL_STATE(7276)] = 241377,
  [SMALL_STATE(7277)] = 241391,
  [SMALL_STATE(7278)] = 241405,
  [SMALL_STATE(7279)] = 241419,
  [SMALL_STATE(7280)] = 241433,
  [SMALL_STATE(7281)] = 241447,
  [SMALL_STATE(7282)] = 241461,
  [SMALL_STATE(7283)] = 241475,
  [SMALL_STATE(7284)] = 241489,
  [SMALL_STATE(7285)] = 241503,
  [SMALL_STATE(7286)] = 241517,
  [SMALL_STATE(7287)] = 241531,
  [SMALL_STATE(7288)] = 241545,
  [SMALL_STATE(7289)] = 241559,
  [SMALL_STATE(7290)] = 241573,
  [SMALL_STATE(7291)] = 241587,
  [SMALL_STATE(7292)] = 241601,
  [SMALL_STATE(7293)] = 241615,
  [SMALL_STATE(7294)] = 241629,
  [SMALL_STATE(7295)] = 241643,
  [SMALL_STATE(7296)] = 241657,
  [SMALL_STATE(7297)] = 241671,
  [SMALL_STATE(7298)] = 241685,
  [SMALL_STATE(7299)] = 241699,
  [SMALL_STATE(7300)] = 241713,
  [SMALL_STATE(7301)] = 241727,
  [SMALL_STATE(7302)] = 241741,
  [SMALL_STATE(7303)] = 241755,
  [SMALL_STATE(7304)] = 241769,
  [SMALL_STATE(7305)] = 241783,
  [SMALL_STATE(7306)] = 241797,
  [SMALL_STATE(7307)] = 241811,
  [SMALL_STATE(7308)] = 241825,
  [SMALL_STATE(7309)] = 241839,
  [SMALL_STATE(7310)] = 241853,
  [SMALL_STATE(7311)] = 241867,
  [SMALL_STATE(7312)] = 241881,
  [SMALL_STATE(7313)] = 241895,
  [SMALL_STATE(7314)] = 241909,
  [SMALL_STATE(7315)] = 241923,
  [SMALL_STATE(7316)] = 241937,
  [SMALL_STATE(7317)] = 241951,
  [SMALL_STATE(7318)] = 241965,
  [SMALL_STATE(7319)] = 241979,
  [SMALL_STATE(7320)] = 241993,
  [SMALL_STATE(7321)] = 242007,
  [SMALL_STATE(7322)] = 242021,
  [SMALL_STATE(7323)] = 242035,
  [SMALL_STATE(7324)] = 242049,
  [SMALL_STATE(7325)] = 242063,
  [SMALL_STATE(7326)] = 242077,
  [SMALL_STATE(7327)] = 242091,
  [SMALL_STATE(7328)] = 242105,
  [SMALL_STATE(7329)] = 242119,
  [SMALL_STATE(7330)] = 242133,
  [SMALL_STATE(7331)] = 242147,
  [SMALL_STATE(7332)] = 242161,
  [SMALL_STATE(7333)] = 242175,
  [SMALL_STATE(7334)] = 242189,
  [SMALL_STATE(7335)] = 242203,
  [SMALL_STATE(7336)] = 242207,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5664),
  [5] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [7] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compilation_unit, 0, 0, 0),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4201),
  [11] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5007),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4079),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4542),
  [17] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4768),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3881),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3534),
  [23] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4727),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4917),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5661),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5663),
  [31] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5700),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(863),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(739),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6220),
  [39] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [41] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_anonymous, 2, 0, 0),
  [43] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression_anonymous, 2, 0, 0),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(300),
  [47] = {.entry = {.count = 1, .reusable = false}}, SHIFT(954),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1415),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3621),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1665),
  [55] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(144),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(139),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1231),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(164),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1152),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(165),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1459),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1447),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4972),
  [75] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5053),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(107),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT(939),
  [81] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5784),
  [83] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5899),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5806),
  [87] = {.entry = {.count = 1, .reusable = true}}, SHIFT(923),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2654),
  [91] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4569),
  [93] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4188),
  [95] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1484),
  [97] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1441),
  [99] = {.entry = {.count = 1, .reusable = true}}, SHIFT(481),
  [101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(886),
  [103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6136),
  [105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(966),
  [107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(740),
  [109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(284),
  [113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(979),
  [115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1431),
  [117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3611),
  [119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1664),
  [121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(141),
  [125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(125),
  [127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1222),
  [129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(158),
  [131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1160),
  [133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(159),
  [135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1463),
  [137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1464),
  [139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4889),
  [141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5173),
  [143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(112),
  [145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5773),
  [149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5891),
  [151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(982),
  [153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4567),
  [155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4170),
  [157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1514),
  [159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1049),
  [163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6109),
  [165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(741),
  [167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1657),
  [169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(146),
  [171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(168),
  [173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1155),
  [175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(138),
  [177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(491),
  [179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1083),
  [181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(744),
  [183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(263),
  [187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1076),
  [189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1427),
  [191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3668),
  [193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1656),
  [195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(161),
  [199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(128),
  [201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1154),
  [203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(129),
  [205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4860),
  [207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5107),
  [209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(100),
  [211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1131),
  [213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5848),
  [215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5792),
  [217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4607),
  [221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4178),
  [223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1501),
  [225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1069),
  [229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6137),
  [231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1688),
  [233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(748),
  [235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1349),
  [237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5027),
  [239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4073),
  [241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1276),
  [243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2462),
  [245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4524),
  [247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4798),
  [249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3844),
  [251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3950),
  [253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4807),
  [255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4930),
  [257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2654),
  [259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1436),
  [261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6100),
  [263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2038),
  [265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3495),
  [267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2204),
  [269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7129),
  [271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5509),
  [273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1091),
  [275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3731),
  [277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3667),
  [279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2199),
  [281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2088),
  [283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5372),
  [285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3279),
  [289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1873),
  [291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2412),
  [293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7335),
  [295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3824),
  [297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4076),
  [299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2993),
  [301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1897),
  [303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3405),
  [305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(983),
  [307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3117),
  [309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2432),
  [311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1585),
  [313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(885),
  [315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2687),
  [317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5753),
  [319] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 4, 0, 0),
  [321] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 4, 0, 0),
  [323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 3, 0, 0),
  [325] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 3, 0, 0),
  [327] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 1, 0, 0),
  [329] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 1, 0, 0),
  [331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3024),
  [333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4535),
  [335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 2, 0, 0),
  [337] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 2, 0, 0),
  [339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(850),
  [341] = {.entry = {.count = 1, .reusable = false}}, SHIFT(743),
  [343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(333),
  [347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(786),
  [349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1426),
  [351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3576),
  [353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1670),
  [355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(140),
  [359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(134),
  [361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1282),
  [363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(149),
  [365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1158),
  [367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(150),
  [369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1449),
  [371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1450),
  [373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4931),
  [375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5072),
  [377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(105),
  [379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(803),
  [381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5716),
  [383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5856),
  [385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(787),
  [387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4572),
  [389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4187),
  [391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1520),
  [393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(808),
  [397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6028),
  [399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(742),
  [401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1669),
  [403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(133),
  [405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5664),
  [407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5716),
  [409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5856),
  [411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4190),
  [413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7318),
  [415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3962),
  [417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7213),
  [419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6830),
  [421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2651),
  [423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1472),
  [425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6460),
  [427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6339),
  [429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6339),
  [431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6725),
  [433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6494),
  [435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6700),
  [437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6700),
  [439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6342),
  [441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1420),
  [443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7269),
  [449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3917),
  [451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6730),
  [453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1479),
  [455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6728),
  [457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1414),
  [459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1892),
  [463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7270),
  [465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6707),
  [467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1476),
  [469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7051),
  [471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1409),
  [473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1906),
  [477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7274),
  [479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3980),
  [481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6879),
  [483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1482),
  [485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6507),
  [487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1418),
  [489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2349),
  [493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7276),
  [495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3981),
  [497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6983),
  [499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1474),
  [501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6874),
  [503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1408),
  [505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1079),
  [509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7229),
  [511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4025),
  [513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7171),
  [515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1481),
  [517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7034),
  [519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1428),
  [521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1017),
  [525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7277),
  [527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3982),
  [529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7032),
  [531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1477),
  [533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6936),
  [535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1406),
  [537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1568),
  [541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7278),
  [543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3985),
  [545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7071),
  [547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1475),
  [549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6982),
  [551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1422),
  [553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(901),
  [557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7279),
  [559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3990),
  [561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7132),
  [563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7031),
  [567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1425),
  [569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2737),
  [573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7280),
  [575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3996),
  [577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7185),
  [579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1478),
  [581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7070),
  [583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1407),
  [585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2009),
  [589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7281),
  [591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3997),
  [593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7230),
  [595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1473),
  [597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7130),
  [599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1417),
  [601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1532),
  [605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(737),
  [607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(292),
  [611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1563),
  [613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1411),
  [615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3618),
  [617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1660),
  [619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(148),
  [623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(137),
  [625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1283),
  [627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(162),
  [629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1157),
  [631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(163),
  [633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1460),
  [635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1461),
  [637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4926),
  [639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5112),
  [641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(98),
  [643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1626),
  [645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5781),
  [647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5897),
  [649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1564),
  [651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4568),
  [653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4180),
  [655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1500),
  [657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1565),
  [661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6126),
  [663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(747),
  [665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1662),
  [667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(143),
  [669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(736),
  [671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1659),
  [673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(147),
  [675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(160),
  [677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1161),
  [679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(114),
  [681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5784),
  [683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5899),
  [685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(745),
  [689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1672),
  [691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(142),
  [693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(746),
  [695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1661),
  [697] = {.entry = {.count = 1, .reusable = false}}, SHIFT(136),
  [699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(154),
  [701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1162),
  [703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(155),
  [705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(738),
  [709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1671),
  [711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(145),
  [713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(166),
  [715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1153),
  [717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(167),
  [719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(749),
  [723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1667),
  [725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(153),
  [727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(156),
  [729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1156),
  [731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(157),
  [733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(462),
  [735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5806),
  [737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(785),
  [739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5969),
  [741] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0),
  [743] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0),
  [745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4556),
  [747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6461),
  [749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1525),
  [751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(941),
  [753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6233),
  [755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4545),
  [757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6489),
  [759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1524),
  [761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(987),
  [763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6127),
  [765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4546),
  [767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7143),
  [769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1517),
  [771] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_application_expression, 2, 0, 34),
  [773] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_application_expression, 2, 0, 34),
  [775] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(785),
  [778] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5969),
  [781] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(45),
  [784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68),
  [786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68),
  [788] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(333),
  [791] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(786),
  [794] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1426),
  [797] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(3576),
  [800] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(334),
  [803] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4931),
  [806] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5072),
  [809] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(105),
  [812] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(803),
  [815] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5716),
  [818] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5856),
  [821] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(787),
  [824] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4572),
  [827] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4187),
  [830] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1520),
  [833] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(808),
  [836] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(6028),
  [839] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(941),
  [842] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(6233),
  [845] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(59),
  [848] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(300),
  [851] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(954),
  [854] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1415),
  [857] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(3621),
  [860] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(301),
  [863] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4972),
  [866] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5053),
  [869] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(107),
  [872] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(939),
  [875] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5784),
  [878] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5899),
  [881] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(923),
  [884] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4569),
  [887] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4188),
  [890] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1484),
  [893] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(886),
  [896] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(6136),
  [899] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(987),
  [902] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(6127),
  [905] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(55),
  [908] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(284),
  [911] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(979),
  [914] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1431),
  [917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(3611),
  [920] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(285),
  [923] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4889),
  [926] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5173),
  [929] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(112),
  [932] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(990),
  [935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5773),
  [938] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5891),
  [941] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(982),
  [944] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4567),
  [947] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4170),
  [950] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1514),
  [953] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1049),
  [956] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(6109),
  [959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1141),
  [961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6221),
  [963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4575),
  [965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6828),
  [967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1512),
  [969] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1141),
  [972] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(6221),
  [975] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(53),
  [978] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(263),
  [981] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1076),
  [984] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1427),
  [987] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(3668),
  [990] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(298),
  [993] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4860),
  [996] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5107),
  [999] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(100),
  [1002] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1131),
  [1005] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5848),
  [1008] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5792),
  [1011] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1105),
  [1014] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4607),
  [1017] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4178),
  [1020] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1501),
  [1023] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1069),
  [1026] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(6137),
  [1029] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5945),
  [1032] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4890),
  [1035] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1522),
  [1038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3540),
  [1040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5955),
  [1042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [1044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2051),
  [1046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5962),
  [1048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7266),
  [1050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1264),
  [1052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4083),
  [1054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4766),
  [1056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1241),
  [1058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5342),
  [1060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5319),
  [1062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7232),
  [1064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1265),
  [1066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1266),
  [1068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5732),
  [1070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1739),
  [1072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5828),
  [1074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5829),
  [1076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4095),
  [1078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6169),
  [1080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4548),
  [1082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4182),
  [1084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7225),
  [1086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1539),
  [1088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6148),
  [1090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4190),
  [1092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1906),
  [1094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2009),
  [1096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3646),
  [1098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6235),
  [1100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [1102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4963),
  [1104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7272),
  [1106] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1262),
  [1108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4765),
  [1110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1269),
  [1112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5422),
  [1114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7012),
  [1116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1263),
  [1118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1270),
  [1120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1724),
  [1122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7105),
  [1124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1568),
  [1126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5405),
  [1128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6703),
  [1130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1079),
  [1132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1892),
  [1134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2349),
  [1136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2229),
  [1138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7263),
  [1140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5321),
  [1142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6667),
  [1144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(792),
  [1146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7184),
  [1148] = {.entry = {.count = 1, .reusable = false}}, SHIFT(901),
  [1150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2514),
  [1152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7275),
  [1154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5437),
  [1156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6937),
  [1158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6770),
  [1160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3095),
  [1162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7273),
  [1164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5434),
  [1166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6771),
  [1168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6991),
  [1170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1017),
  [1172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2737),
  [1174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2244),
  [1176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5773),
  [1178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5891),
  [1180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2322),
  [1182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2892),
  [1184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5848),
  [1186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5792),
  [1188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3399),
  [1190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7210),
  [1192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(817),
  [1194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1124),
  [1196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6512),
  [1198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1139),
  [1200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [1202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(904),
  [1204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3336),
  [1206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4177),
  [1208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2723),
  [1210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2760),
  [1212] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2190),
  [1214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1829),
  [1216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [1218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2018),
  [1220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2012),
  [1222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1034),
  [1224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2417),
  [1226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2214),
  [1228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1553),
  [1230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1819),
  [1232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1917),
  [1234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1574),
  [1236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2360),
  [1238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2314),
  [1240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1931),
  [1242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [1244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2967),
  [1246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2774),
  [1248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1737),
  [1250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1741),
  [1252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1806),
  [1254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1811),
  [1256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1101),
  [1258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2352),
  [1260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2354),
  [1262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(993),
  [1264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(978),
  [1266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1550),
  [1268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1552),
  [1270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1074),
  [1272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(891),
  [1274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(893),
  [1276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2771),
  [1278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1987),
  [1280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2120),
  [1282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [1284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(807),
  [1286] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_content, 2, 0, 0),
  [1288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_content, 3, 0, 0),
  [1290] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6150),
  [1292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6078),
  [1294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6073),
  [1296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(169),
  [1298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4890),
  [1300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1522),
  [1302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6001),
  [1304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1606),
  [1306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6075),
  [1308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4594),
  [1310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6835),
  [1312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1526),
  [1314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2107),
  [1316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_binding_pattern, 1, 0, 0),
  [1318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_binding_pattern, 1, 0, 0),
  [1320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5955),
  [1322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [1324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1253),
  [1326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1868),
  [1328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4767),
  [1330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1240),
  [1332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5278),
  [1334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1284),
  [1336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1255),
  [1338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5686),
  [1340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1780),
  [1342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5693),
  [1344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5849),
  [1346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1894),
  [1348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6867),
  [1350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4596),
  [1352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4176),
  [1354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1701),
  [1356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6204),
  [1358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [1360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5945),
  [1362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1483),
  [1364] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1606),
  [1367] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(6075),
  [1370] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(57),
  [1373] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(292),
  [1376] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1563),
  [1379] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1411),
  [1382] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(3618),
  [1385] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(293),
  [1388] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4926),
  [1391] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5112),
  [1394] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(98),
  [1397] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1626),
  [1400] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5781),
  [1403] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5897),
  [1406] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1564),
  [1409] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4568),
  [1412] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4180),
  [1415] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1500),
  [1418] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1565),
  [1421] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(6126),
  [1424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3003),
  [1426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1443),
  [1428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [1430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5025),
  [1432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4075),
  [1434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1236),
  [1436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2951),
  [1438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4775),
  [1440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4507),
  [1442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4797),
  [1444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3850),
  [1446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4743),
  [1448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5497),
  [1450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1237),
  [1452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1244),
  [1454] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5765),
  [1456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5840),
  [1458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5764),
  [1460] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5884),
  [1462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5841),
  [1464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5790),
  [1466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5705),
  [1468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2952),
  [1470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6741),
  [1472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4547),
  [1474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4184),
  [1476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2731),
  [1478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6179),
  [1480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1434),
  [1482] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2235),
  [1484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 1, 0, 25),
  [1486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 1, 0, 25),
  [1488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [1490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5825),
  [1492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [1494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1292),
  [1496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1330),
  [1498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2585),
  [1500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1277),
  [1502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2217),
  [1504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4785),
  [1506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2788),
  [1508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5390),
  [1510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5332),
  [1512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1331),
  [1514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1278),
  [1516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5564),
  [1518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5868),
  [1520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5667),
  [1522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5839),
  [1524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2221),
  [1526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6720),
  [1528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4587),
  [1530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4175),
  [1532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1435),
  [1534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [1536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1433),
  [1538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_expression, 1, 0, 0),
  [1540] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_expression, 1, 0, 0),
  [1542] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_application, 2, 0, 86),
  [1544] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_application, 2, 0, 86),
  [1546] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_path, 3, 0, 101),
  [1548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_path, 3, 0, 101),
  [1550] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_get_expression, 5, 0, 145),
  [1552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_get_expression, 5, 0, 145),
  [1554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_get_expression, 5, 0, 144),
  [1556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_get_expression, 5, 0, 144),
  [1558] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_get_expression, 6, 0, 181),
  [1560] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_get_expression, 6, 0, 181),
  [1562] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_get_expression, 6, 0, 182),
  [1564] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_get_expression, 6, 0, 182),
  [1566] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bigarray_get_expression, 6, 0, 181),
  [1568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bigarray_get_expression, 6, 0, 181),
  [1570] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bigarray_get_expression, 5, 0, 144),
  [1572] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bigarray_get_expression, 5, 0, 144),
  [1574] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_get_expression, 3, 0, 65),
  [1576] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_get_expression, 3, 0, 65),
  [1578] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_path, 1, 0, 30),
  [1580] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_path, 1, 0, 30),
  [1582] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 2, 0, 0),
  [1584] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 2, 0, 0),
  [1586] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 4, 0, 0),
  [1588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 4, 0, 0),
  [1590] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 4, 0, 63),
  [1592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 4, 0, 63),
  [1594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 7, 0, 134),
  [1596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 7, 0, 134),
  [1598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 7, 0, 170),
  [1600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 7, 0, 170),
  [1602] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 7, 0, 179),
  [1604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 7, 0, 179),
  [1606] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 7, 0, 206),
  [1608] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 7, 0, 206),
  [1610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 5, 0, 132),
  [1612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 5, 0, 132),
  [1614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__value_name, 1, 0, 1),
  [1616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__value_name, 1, 0, 1),
  [1618] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean, 1, 0, 0),
  [1620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean, 1, 0, 0),
  [1622] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_number, 1, 0, 0),
  [1624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_number, 1, 0, 0),
  [1626] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 8, 0, 206),
  [1628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 8, 0, 206),
  [1630] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 5, 0, 44),
  [1632] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 5, 0, 44),
  [1634] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constant, 1, 0, 0),
  [1636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constant, 1, 0, 0),
  [1638] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 5, 0, 134),
  [1640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 5, 0, 134),
  [1642] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unit, 2, 0, 0),
  [1644] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unit, 2, 0, 0),
  [1646] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__value_name, 1, 0, 0),
  [1648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__value_name, 1, 0, 0),
  [1650] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 2, 0, 0),
  [1652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 2, 0, 0),
  [1654] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2405),
  [1656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6225),
  [1658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [1660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(274),
  [1662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2347),
  [1664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1423),
  [1666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3589),
  [1668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [1670] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4831),
  [1672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5156),
  [1674] = {.entry = {.count = 1, .reusable = false}}, SHIFT(102),
  [1676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2350),
  [1678] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5742),
  [1680] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5873),
  [1682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2423),
  [1684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4565),
  [1686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4185),
  [1688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1528),
  [1690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2413),
  [1692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6090),
  [1694] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 5, 0, 0),
  [1696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 5, 0, 0),
  [1698] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_name, 1, 0, 6),
  [1700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_name, 1, 0, 6),
  [1702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_path, 3, 0, 38),
  [1704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_expression, 1, 0, 0),
  [1706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_expression, 1, 0, 0),
  [1708] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_path, 3, 0, 0),
  [1710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_path, 3, 0, 0),
  [1712] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 5, 0, 136),
  [1714] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 5, 0, 136),
  [1716] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_path, 1, 0, 0),
  [1718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_path, 1, 0, 0),
  [1720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 5, 0, 0),
  [1722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 5, 0, 0),
  [1724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_expression, 1, 0, 0),
  [1726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_expression, 1, 0, 0),
  [1728] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_expression, 2, 0, 0),
  [1730] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_expression, 2, 0, 0),
  [1732] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 2, 0, 0),
  [1734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 2, 0, 0),
  [1736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 4, 0, 0),
  [1738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 4, 0, 0),
  [1740] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 2, 0, 0),
  [1742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 2, 0, 0),
  [1744] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_path, 1, 0, 5),
  [1746] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 4, 0, 0),
  [1748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 4, 0, 0),
  [1750] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 3, 0, 0),
  [1752] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 3, 0, 0),
  [1754] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_name, 3, 0, 40),
  [1756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_name, 3, 0, 40),
  [1758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_prefix_expression, 2, 0, 32),
  [1760] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_prefix_expression, 2, 0, 32),
  [1762] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(2405),
  [1765] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(6225),
  [1768] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(52),
  [1771] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(274),
  [1774] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(2347),
  [1777] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1423),
  [1780] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(3589),
  [1783] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(275),
  [1786] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4831),
  [1789] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5156),
  [1792] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(102),
  [1795] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(2350),
  [1798] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5742),
  [1801] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(5873),
  [1804] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(2423),
  [1807] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4565),
  [1810] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(4185),
  [1813] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(1528),
  [1816] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(2413),
  [1819] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 68), SHIFT_REPEAT(6090),
  [1822] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 6, 0, 166),
  [1824] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 6, 0, 166),
  [1826] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_string, 3, 0, 0),
  [1828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_string, 3, 0, 0),
  [1830] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_path, 1, 0, 0),
  [1832] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_path, 1, 0, 0),
  [1834] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unit, 3, 0, 0),
  [1836] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unit, 3, 0, 0),
  [1838] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 31),
  [1840] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 31),
  [1842] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_character, 3, 0, 0),
  [1844] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_character, 3, 0, 0),
  [1846] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 6, 0, 134),
  [1848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 6, 0, 134),
  [1850] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 3, 0, 0),
  [1852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 3, 0, 0),
  [1854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tag, 2, 0, 0),
  [1856] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag, 2, 0, 0),
  [1858] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_expression, 3, 0, 0),
  [1860] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_expression, 3, 0, 0),
  [1862] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 3, 0, 0),
  [1864] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 3, 0, 0),
  [1866] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 3, 0, 0),
  [1868] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 3, 0, 0),
  [1870] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 3, 0, 0),
  [1872] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 3, 0, 0),
  [1874] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_new_expression, 2, 0, 0),
  [1876] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_new_expression, 2, 0, 0),
  [1878] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_new_expression, 3, 0, 0),
  [1880] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_new_expression, 3, 0, 0),
  [1882] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 3, 0, 0),
  [1884] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 3, 0, 0),
  [1886] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 6, 0, 136),
  [1888] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 6, 0, 136),
  [1890] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_invocation, 3, 0, 66),
  [1892] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_invocation, 3, 0, 66),
  [1894] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_expression, 3, 0, 67),
  [1896] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_expression, 3, 0, 67),
  [1898] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_expression, 4, 0, 96),
  [1900] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_expression, 4, 0, 96),
  [1902] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extension, 4, 0, 0),
  [1904] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extension, 4, 0, 0),
  [1906] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_extension, 4, 0, 0),
  [1908] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_extension, 4, 0, 0),
  [1910] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_expression, 3, 0, 63),
  [1912] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_expression, 3, 0, 63),
  [1914] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 6, 0, 170),
  [1916] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 6, 0, 170),
  [1918] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_path, 3, 0, 0),
  [1920] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_path, 3, 0, 0),
  [1922] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_extension, 5, 0, 0),
  [1924] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_extension, 5, 0, 0),
  [1926] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 6, 0, 0),
  [1928] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 6, 0, 0),
  [1930] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 7, 0, 0),
  [1932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 7, 0, 0),
  [1934] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 8, 0, 0),
  [1936] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 8, 0, 0),
  [1938] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 9, 0, 0),
  [1940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 9, 0, 0),
  [1942] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 6, 0, 179),
  [1944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 6, 0, 179),
  [1946] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 4, 0, 18),
  [1948] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 4, 0, 18),
  [1950] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extension, 3, 0, 0),
  [1952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extension, 3, 0, 0),
  [1954] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0),
  [1956] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2235),
  [1959] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0),
  [1961] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5825),
  [1964] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(103),
  [1967] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1277),
  [1970] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2217),
  [1973] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4785),
  [1976] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5332),
  [1979] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1278),
  [1982] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5564),
  [1985] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5667),
  [1988] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5839),
  [1991] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2221),
  [1994] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6720),
  [1997] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4587),
  [2000] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4175),
  [2003] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1539),
  [2006] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6148),
  [2009] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__instance_variable_name, 1, 0, 11),
  [2011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [2013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1245),
  [2015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1782),
  [2017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5764),
  [2019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5884),
  [2021] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_path, 3, 0, 107),
  [2023] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_path, 3, 0, 107),
  [2025] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_path, 1, 0, 10),
  [2027] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_path, 1, 0, 10),
  [2029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2452),
  [2031] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 2, 0, 22),
  [2033] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 2, 0, 22),
  [2035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3362),
  [2037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5570),
  [2039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(108),
  [2041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2597),
  [2043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1234),
  [2045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2466),
  [2047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4773),
  [2049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5266),
  [2051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1235),
  [2053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5847),
  [2055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5662),
  [2057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5771),
  [2059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5889),
  [2061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2490),
  [2063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7066),
  [2065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4563),
  [2067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4166),
  [2069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2446),
  [2071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5947),
  [2073] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 2, 0, 10),
  [2075] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 2, 0, 10),
  [2077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3424),
  [2079] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 1, 0, 33),
  [2081] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 1, 0, 33),
  [2083] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 22),
  [2085] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 22),
  [2087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3466),
  [2089] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 83),
  [2091] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 83),
  [2093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3407),
  [2095] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 83),
  [2097] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 83),
  [2099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3376),
  [2101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 119),
  [2103] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 119),
  [2105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3378),
  [2107] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_argument, 4, 0, 94),
  [2109] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument, 4, 0, 94),
  [2111] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 119),
  [2113] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 119),
  [2115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3421),
  [2117] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 155),
  [2119] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 155),
  [2121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3423),
  [2123] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_argument, 5, 0, 130),
  [2125] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument, 5, 0, 130),
  [2127] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 1, 0, 10),
  [2129] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 1, 0, 10),
  [2131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3412),
  [2133] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 155),
  [2135] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 155),
  [2137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3432),
  [2139] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_pattern, 1, 0, 0),
  [2141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6235),
  [2143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [2145] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_pattern, 1, 0, 0),
  [2147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5319),
  [2149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6759),
  [2151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3437),
  [2153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2617),
  [2155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3363),
  [2157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3448),
  [2159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3369),
  [2161] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_argument, 2, 0, 27),
  [2163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1488),
  [2165] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument, 2, 0, 27),
  [2167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3465),
  [2169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3440),
  [2171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(594),
  [2173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3451),
  [2175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3416),
  [2177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3469),
  [2179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3367),
  [2181] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2452),
  [2184] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5570),
  [2187] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(108),
  [2190] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1234),
  [2193] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2466),
  [2196] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4773),
  [2199] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5266),
  [2202] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1235),
  [2205] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5847),
  [2208] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5771),
  [2211] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5889),
  [2214] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2490),
  [2217] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(7066),
  [2220] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4563),
  [2223] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4166),
  [2226] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2446),
  [2229] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5947),
  [2232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1527),
  [2234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(624),
  [2236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4784),
  [2238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [2240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1271),
  [2242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5403),
  [2244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1226),
  [2246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1753),
  [2248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5748),
  [2250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5877),
  [2252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6759),
  [2254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3150),
  [2256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6176),
  [2258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [2260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1267),
  [2262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1268),
  [2264] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1751),
  [2266] = {.entry = {.count = 1, .reusable = false}}, SHIFT(711),
  [2268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [2270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(110),
  [2272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4951),
  [2274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [2276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1254),
  [2278] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4736),
  [2280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4690),
  [2282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1227),
  [2284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5449),
  [2286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1223),
  [2288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1228),
  [2290] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5671),
  [2292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1774),
  [2294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5758),
  [2296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5880),
  [2298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4678),
  [2300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7257),
  [2302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4584),
  [2304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4167),
  [2306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3193),
  [2308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6268),
  [2310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4892),
  [2312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [2314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1224),
  [2316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4691),
  [2318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1230),
  [2320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1225),
  [2322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1238),
  [2324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1776),
  [2326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1511),
  [2328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(537),
  [2330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [2332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [2334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [2336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1297),
  [2338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1310),
  [2340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1289),
  [2342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1318),
  [2344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1317),
  [2346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1498),
  [2348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1293),
  [2350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1306),
  [2352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1298),
  [2354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [2356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2585),
  [2358] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 3, 0, 104),
  [2360] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 3, 0, 104),
  [2362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [2364] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 4, 0, 140),
  [2366] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 140),
  [2368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [2370] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 3, 0, 60),
  [2372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 3, 0, 60),
  [2374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [2376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1299),
  [2378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1360),
  [2380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5326),
  [2382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1361),
  [2384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5694),
  [2386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5667),
  [2388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5839),
  [2390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [2392] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 5, 0, 140),
  [2394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 140),
  [2396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [2398] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 4, 0, 104),
  [2400] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 104),
  [2402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [2404] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 2, 0, 60),
  [2406] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 2, 0, 60),
  [2408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4924),
  [2410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4839),
  [2412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4841),
  [2414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2453),
  [2416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2473),
  [2418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3086),
  [2420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3087),
  [2422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5693),
  [2424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5849),
  [2426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2089),
  [2428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4863),
  [2430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4434),
  [2432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4361),
  [2434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4611),
  [2436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4610),
  [2438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2538),
  [2440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6969),
  [2442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2529),
  [2444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5013),
  [2446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4079),
  [2448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2533),
  [2450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_payload, 1, 0, 0),
  [2452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4534),
  [2454] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4768),
  [2456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3881),
  [2458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3534),
  [2460] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4727),
  [2462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4917),
  [2464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2517),
  [2466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2536),
  [2468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3226),
  [2470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5477),
  [2472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5828),
  [2474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5829),
  [2476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3235),
  [2478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6118),
  [2480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2173),
  [2482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2175),
  [2484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2090),
  [2486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6181),
  [2488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4317),
  [2490] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_pattern_content, 3, 0, 0),
  [2492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6312),
  [2494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5039),
  [2496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6292),
  [2498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4586),
  [2500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_binding_pattern_content, 2, 0, 0),
  [2502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6241),
  [2504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2020),
  [2506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_binding_pattern_content, 3, 0, 0),
  [2508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2072),
  [2510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6249),
  [2512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4904),
  [2514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6057),
  [2516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3137),
  [2518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_pattern_content, 2, 0, 0),
  [2520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6087),
  [2522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6054),
  [2524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2995),
  [2526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1439),
  [2528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1505),
  [2530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4544),
  [2532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [2534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(231),
  [2536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1430),
  [2538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3584),
  [2540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [2542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4923),
  [2544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5048),
  [2546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(94),
  [2548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1713),
  [2550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1513),
  [2552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1746),
  [2554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5994),
  [2556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4566),
  [2558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(578),
  [2560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(696),
  [2562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1442),
  [2564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2004),
  [2566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [2568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(315),
  [2570] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2015),
  [2572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1421),
  [2574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3626),
  [2576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [2578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4855),
  [2580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5220),
  [2582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(96),
  [2584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2121),
  [2586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5788),
  [2588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5901),
  [2590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5684),
  [2592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2016),
  [2594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4571),
  [2596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4169),
  [2598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1509),
  [2600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2029),
  [2602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6165),
  [2604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1852),
  [2606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [2608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(261),
  [2610] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1889),
  [2612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1416),
  [2614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3605),
  [2616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [2618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4901),
  [2620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5045),
  [2622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [2624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1921),
  [2626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5767),
  [2628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5887),
  [2630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1890),
  [2632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4555),
  [2634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4179),
  [2636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1486),
  [2638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1908),
  [2640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6111),
  [2642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3467),
  [2644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2607),
  [2646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5668),
  [2648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3411),
  [2650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3430),
  [2652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3436),
  [2654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3370),
  [2656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2711),
  [2658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [2660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(212),
  [2662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2734),
  [2664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1419),
  [2666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3625),
  [2668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [2670] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4817),
  [2672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5155),
  [2674] = {.entry = {.count = 1, .reusable = false}}, SHIFT(113),
  [2676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2682),
  [2678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5786),
  [2680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5900),
  [2682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2710),
  [2684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4570),
  [2686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4186),
  [2688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1491),
  [2690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2720),
  [2692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6149),
  [2694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1684),
  [2696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [2698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1936),
  [2700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1424),
  [2702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(101),
  [2704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5735),
  [2706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5867),
  [2708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1938),
  [2710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4597),
  [2712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4189),
  [2714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1502),
  [2716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3392),
  [2718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3366),
  [2720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3419),
  [2722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3441),
  [2724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3409),
  [2726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression, 1, 0, 13),
  [2728] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression, 1, 0, 13),
  [2730] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_anonymous, 4, 0, 0),
  [2732] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression_anonymous, 4, 0, 0),
  [2734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_anonymous, 3, 0, 0),
  [2736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression_anonymous, 3, 0, 0),
  [2738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5742),
  [2740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5873),
  [2742] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_id, 2, 0, 0),
  [2744] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_id, 2, 0, 0),
  [2746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6296),
  [2748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(395),
  [2750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [2752] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_id, 1, 0, 0),
  [2754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_id, 1, 0, 0),
  [2756] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0),
  [2758] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0),
  [2760] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(6296),
  [2763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [2765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(638),
  [2767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(563),
  [2769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(666),
  [2771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3386),
  [2773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [2775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6720),
  [2777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [2779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [2781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [2783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [2785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(441),
  [2787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [2789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [2791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [2793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(471),
  [2795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [2797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [2799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [2801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [2803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [2805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [2807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(548),
  [2809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [2811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [2813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(448),
  [2815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(450),
  [2817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [2819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [2821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [2823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [2825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [2827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [2829] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0),
  [2831] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0), SHIFT_REPEAT(1688),
  [2834] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0),
  [2836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1490),
  [2838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(608),
  [2840] = {.entry = {.count = 1, .reusable = false}}, SHIFT(652),
  [2842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(678),
  [2844] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_then_clause, 2, 0, 31),
  [2846] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_then_clause, 2, 0, 31),
  [2848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [2850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2670),
  [2852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [2854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2659),
  [2856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2657),
  [2858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2663),
  [2860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2663),
  [2862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [2864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(587),
  [2866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2664),
  [2868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2664),
  [2870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(588),
  [2872] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3),
  [2874] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sign_expression, 2, 0, 32),
  [2876] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sign_expression, 2, 0, 32),
  [2878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3393),
  [2880] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_infix_expression, 3, 0, 67),
  [2882] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_infix_expression, 3, 0, 67),
  [2884] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_expression, 3, 0, 64),
  [2886] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_expression, 3, 0, 64),
  [2888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3384),
  [2890] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 2, 0, 31),
  [2892] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2, 0, 31),
  [2894] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat1, 1, 0, 0),
  [2896] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 1, 0, 0),
  [2898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3395),
  [2900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3406),
  [2902] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_expression, 3, 0, 0),
  [2904] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_expression, 3, 0, 0),
  [2906] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression, 1, 0, 0),
  [2908] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression, 1, 0, 0),
  [2910] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_expression, 3, 0, 69),
  [2912] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_expression, 3, 0, 69),
  [2914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(724),
  [2916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(614),
  [2918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4),
  [2920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(615),
  [2922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [2924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(617),
  [2926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [2928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1239),
  [2930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1248),
  [2932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [2934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(702),
  [2936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(703),
  [2938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(704),
  [2940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(705),
  [2942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [2944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element, 4, 0, 94),
  [2946] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element, 4, 0, 94),
  [2948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4619),
  [2950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6367),
  [2952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1523),
  [2954] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_expression, 2, 0, 31),
  [2956] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_expression, 2, 0, 31),
  [2958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4550),
  [2960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6361),
  [2962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1504),
  [2964] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_expression, 3, 0, 63),
  [2966] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_expression, 3, 0, 63),
  [2968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(73),
  [2970] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__signed_constant, 1, 0, 0),
  [2972] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signed_constant, 1, 0, 0),
  [2974] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signed_number, 2, 0, 0),
  [2976] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signed_number, 2, 0, 0),
  [2978] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert_expression, 3, 0, 63),
  [2980] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert_expression, 3, 0, 63),
  [2982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4226),
  [2984] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 2, 0, 27),
  [2986] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 2, 0, 27),
  [2988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2325),
  [2990] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert_expression, 2, 0, 31),
  [2992] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert_expression, 2, 0, 31),
  [2994] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_binding_pattern, 3, 0, 0),
  [2996] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_binding_pattern, 3, 0, 0),
  [2998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4562),
  [3000] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__parenthesized_abstract_type, 3, 0, 0),
  [3002] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__parenthesized_abstract_type, 3, 0, 0),
  [3004] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 4, 0, 38),
  [3006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 4, 0, 38),
  [3008] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 4, 0, 122),
  [3010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 4, 0, 122),
  [3012] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__effect_binding_pattern, 1, 0, 0),
  [3014] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__effect_binding_pattern, 1, 0, 0),
  [3016] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_binding_pattern, 3, 0, 0),
  [3018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_binding_pattern, 3, 0, 0),
  [3020] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 3, 0, 0),
  [3022] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 3, 0, 0),
  [3024] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 5, 0, 159),
  [3026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 5, 0, 159),
  [3028] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 5, 0, 160),
  [3030] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 5, 0, 160),
  [3032] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 5, 0, 162),
  [3034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 5, 0, 162),
  [3036] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_binding_pattern, 3, 0, 0),
  [3038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_binding_pattern, 3, 0, 0),
  [3040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_binding_pattern, 3, 0, 89),
  [3042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_binding_pattern, 3, 0, 89),
  [3044] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_binding_pattern_no_exn, 3, 0, 0),
  [3046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_binding_pattern_no_exn, 3, 0, 0),
  [3048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1295),
  [3050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1355),
  [3052] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__or_binding_pattern_no_exn_anonymous, 3, 0, 0),
  [3054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__or_binding_pattern_no_exn_anonymous, 3, 0, 0),
  [3056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1354),
  [3058] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alias_binding_pattern_no_exn, 3, 0, 93),
  [3060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias_binding_pattern_no_exn, 3, 0, 93),
  [3062] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 6, 0, 194),
  [3064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 6, 0, 194),
  [3066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(611),
  [3068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5),
  [3070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(527),
  [3072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(528),
  [3074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [3076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [3078] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_binding_pattern_no_exn, 3, 0, 69),
  [3080] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_binding_pattern_no_exn, 3, 0, 69),
  [3082] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_pattern, 3, 0, 69),
  [3084] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_pattern, 3, 0, 69),
  [3086] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_binding_pattern, 3, 0, 89),
  [3088] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_binding_pattern, 3, 0, 89),
  [3090] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_binding_pattern, 3, 0, 89),
  [3092] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_binding_pattern, 3, 0, 89),
  [3094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6578),
  [3096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1508),
  [3098] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern_no_exn, 1, 0, 0),
  [3100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern_no_exn, 1, 0, 0),
  [3102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_constructor_path, 3, 0, 37),
  [3104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_constructor_path, 3, 0, 37),
  [3106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_constructor_path, 1, 0, 4),
  [3108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_constructor_path, 1, 0, 4),
  [3110] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern, 1, 0, 0),
  [3112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern, 1, 0, 0),
  [3114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern, 1, 0, 28),
  [3116] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern, 1, 0, 28),
  [3118] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern, 1, 0, 29),
  [3120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern, 1, 0, 29),
  [3122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(74),
  [3124] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_binding_pattern, 2, 0, 51),
  [3126] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_binding_pattern, 2, 0, 51),
  [3128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5435),
  [3130] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_class_expression, 3, 0, 0),
  [3132] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_class_expression, 3, 0, 0),
  [3134] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_binding_pattern, 3, 0, 89),
  [3136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_binding_pattern, 3, 0, 89),
  [3138] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_binding_pattern, 3, 0, 0),
  [3140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_binding_pattern, 3, 0, 0),
  [3142] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__or_binding_pattern_anonymous, 3, 0, 0),
  [3144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__or_binding_pattern_anonymous, 3, 0, 0),
  [3146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alias_binding_pattern, 3, 0, 93),
  [3148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias_binding_pattern, 3, 0, 93),
  [3150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_binding_pattern, 3, 0, 69),
  [3152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_binding_pattern, 3, 0, 69),
  [3154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_binding_pattern, 1, 0, 0),
  [3156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_binding_pattern, 1, 0, 0),
  [3158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_effect_binding_pattern, 4, 0, 131),
  [3160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_effect_binding_pattern, 4, 0, 131),
  [3162] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_binding_pattern, 4, 0, 123),
  [3164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_binding_pattern, 4, 0, 123),
  [3166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 4, 0, 0),
  [3168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 4, 0, 0),
  [3170] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_class_expression, 1, 0, 0),
  [3172] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_class_expression, 1, 0, 0),
  [3174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_binding_pattern, 2, 0, 0),
  [3176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_binding_pattern, 2, 0, 0),
  [3178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_binding_pattern, 2, 0, 0),
  [3180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_binding_pattern, 2, 0, 0),
  [3182] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_binding_pattern, 2, 0, 51),
  [3184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_binding_pattern, 2, 0, 51),
  [3186] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_class_expression, 4, 0, 157),
  [3188] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_class_expression, 4, 0, 157),
  [3190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class, 4, 0, 0),
  [3192] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class, 4, 0, 0),
  [3194] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 5, 0, 0),
  [3196] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 5, 0, 0),
  [3198] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_binding_pattern, 2, 0, 51),
  [3200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_binding_pattern, 2, 0, 51),
  [3202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tag_binding_pattern, 2, 0, 51),
  [3204] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_binding_pattern, 2, 0, 51),
  [3206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class, 5, 0, 0),
  [3208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class, 5, 0, 0),
  [3210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 6, 0, 0),
  [3212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 6, 0, 0),
  [3214] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 7, 0, 0),
  [3216] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 7, 0, 0),
  [3218] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_pattern, 2, 0, 0),
  [3220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_pattern, 2, 0, 0),
  [3222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__match_cases, 3, 0, 0),
  [3224] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__match_cases, 3, 0, 0),
  [3226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1309),
  [3228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2997),
  [3230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2461),
  [3232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5270),
  [3234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2337),
  [3236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3500),
  [3238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3216),
  [3240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2518),
  [3242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3438),
  [3244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5666),
  [3246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2523),
  [3248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2580),
  [3250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3214),
  [3252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5320),
  [3254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5776),
  [3256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5893),
  [3258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3273),
  [3260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6124),
  [3262] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0),
  [3264] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0),
  [3266] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1309),
  [3269] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element, 2, 0, 27),
  [3271] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element, 2, 0, 27),
  [3273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1489),
  [3275] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__match_cases, 1, 0, 0),
  [3277] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__match_cases, 1, 0, 0),
  [3279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__match_cases, 2, 0, 0),
  [3281] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__match_cases, 2, 0, 0),
  [3283] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 4, 0, 106),
  [3285] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 4, 0, 106),
  [3287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [3289] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_value_pattern, 1, 0, 23),
  [3291] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_value_pattern, 1, 0, 23),
  [3293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2312),
  [3295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2493),
  [3297] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5312),
  [3299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2333),
  [3301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4667),
  [3303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4366),
  [3305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2524),
  [3307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4649),
  [3309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5938),
  [3311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4531),
  [3313] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 3, 0, 62),
  [3315] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 3, 0, 62),
  [3317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2486),
  [3319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2335),
  [3321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3558),
  [3323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3323),
  [3325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2520),
  [3327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2340),
  [3329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3483),
  [3331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2332),
  [3333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3580),
  [3335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2338),
  [3337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4634),
  [3339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1515),
  [3341] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_expression, 3, 0, 64),
  [3343] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_expression, 3, 0, 64),
  [3345] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 3, 0, 0),
  [3347] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 3, 0, 0),
  [3349] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_pattern, 3, 0, 0),
  [3351] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_pattern, 3, 0, 0),
  [3353] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_expression, 4, 0, 79),
  [3355] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_expression, 4, 0, 79),
  [3357] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 4, 0, 62),
  [3359] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 4, 0, 62),
  [3361] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, 0, 89),
  [3363] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 4, 0, 89),
  [3365] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_expression, 7, 0, 207),
  [3367] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_expression, 7, 0, 207),
  [3369] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_expression, 5, 0, 129),
  [3371] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_expression, 5, 0, 129),
  [3373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_expression, 8, 0, 222),
  [3375] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_expression, 8, 0, 222),
  [3377] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element, 5, 0, 130),
  [3379] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element, 5, 0, 130),
  [3381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_pattern, 3, 0, 89),
  [3383] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_pattern, 3, 0, 89),
  [3385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 5, 0, 129),
  [3387] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 5, 0, 129),
  [3389] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, 0, 122),
  [3391] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 4, 0, 122),
  [3393] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_pattern, 4, 0, 123),
  [3395] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_pattern, 4, 0, 123),
  [3397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, 0, 123),
  [3399] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 4, 0, 123),
  [3401] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 4, 0, 0),
  [3403] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 4, 0, 0),
  [3405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1521),
  [3407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(76),
  [3409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 12),
  [3411] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 12),
  [3413] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 5, 0, 143),
  [3415] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 5, 0, 143),
  [3417] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 2, 0, 51),
  [3419] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 2, 0, 51),
  [3421] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 5, 0, 106),
  [3423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 5, 0, 106),
  [3425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 2, 0, 0),
  [3427] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_pattern, 2, 0, 0),
  [3429] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_pattern, 2, 0, 0),
  [3431] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_pattern, 2, 0, 0),
  [3433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_clause, 3, 0, 0),
  [3435] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_clause, 3, 0, 0),
  [3437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 6, 0, 180),
  [3439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 6, 0, 180),
  [3441] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_expression, 5, 0, 63),
  [3443] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_expression, 5, 0, 63),
  [3445] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_case, 4, 0, 128),
  [3447] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_case, 4, 0, 128),
  [3449] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_expression, 5, 0, 63),
  [3451] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_expression, 5, 0, 63),
  [3453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 4, 0, 79),
  [3455] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 4, 0, 79),
  [3457] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_clause, 2, 0, 0),
  [3459] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_clause, 2, 0, 0),
  [3461] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_expression, 4, 0, 106),
  [3463] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_expression, 4, 0, 106),
  [3465] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 5, 0, 158),
  [3467] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 5, 0, 158),
  [3469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_expression, 4, 0, 31),
  [3471] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_expression, 4, 0, 31),
  [3473] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 5, 0, 0),
  [3475] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 5, 0, 0),
  [3477] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 6, 0, 193),
  [3479] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 6, 0, 193),
  [3481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(626),
  [3483] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 7, 0, 215),
  [3485] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 7, 0, 215),
  [3487] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 7, 0, 216),
  [3489] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 7, 0, 216),
  [3491] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 7, 0, 0),
  [3493] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 7, 0, 0),
  [3495] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 8, 0, 227),
  [3497] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 8, 0, 227),
  [3499] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 9, 0, 230),
  [3501] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 9, 0, 230),
  [3503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1287),
  [3505] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1287),
  [3508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_expression, 3, 0, 62),
  [3510] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_expression, 3, 0, 62),
  [3512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_expression, 2, 0, 0),
  [3514] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_expression, 2, 0, 0),
  [3516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_expression, 3, 0, 0),
  [3518] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_expression, 3, 0, 0),
  [3520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 1, 0, 0),
  [3522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 1, 0, 0),
  [3524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 1, 0, 25),
  [3526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 1, 0, 25),
  [3528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_refutation_case, 1, 0, 0),
  [3530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_refutation_case, 1, 0, 0),
  [3532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_case, 3, 0, 91),
  [3534] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_case, 3, 0, 91),
  [3536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_pattern, 3, 0, 0),
  [3538] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_pattern, 3, 0, 0),
  [3540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_expression, 4, 0, 31),
  [3542] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_expression, 4, 0, 31),
  [3544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_pattern, 1, 0, 0),
  [3546] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_pattern, 1, 0, 0),
  [3548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 3, 0, 0),
  [3550] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_pattern, 3, 0, 0),
  [3552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 6, 0, 0),
  [3554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 6, 0, 0),
  [3556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1320),
  [3558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [3560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4803),
  [3562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6154),
  [3564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [3566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2931),
  [3568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6110),
  [3570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [3572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [3574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4409),
  [3576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [3578] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1320),
  [3581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(727),
  [3583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [3585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [3587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [3589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5771),
  [3591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5889),
  [3593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1319),
  [3595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(700),
  [3597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3555),
  [3599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4632),
  [3601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3595),
  [3603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3480),
  [3605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4662),
  [3607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3498),
  [3609] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1319),
  [3612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4595),
  [3614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6854),
  [3616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1510),
  [3618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2324),
  [3620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3388),
  [3622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3532),
  [3624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7055),
  [3626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3537),
  [3628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4663),
  [3630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1518),
  [3632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2538),
  [3634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2587),
  [3636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2643),
  [3638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2598),
  [3640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2641),
  [3642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_type, 3, 0, 98),
  [3644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2457),
  [3646] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_type, 3, 0, 98),
  [3648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5430),
  [3650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2557),
  [3652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3073),
  [3654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_variable, 2, 0, 0),
  [3656] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_variable, 2, 0, 0),
  [3658] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_type, 3, 0, 0),
  [3660] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_type, 3, 0, 0),
  [3662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5336),
  [3664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2532),
  [3666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4477),
  [3668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(686),
  [3670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(687),
  [3672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [3674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [3676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(690),
  [3678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4482),
  [3680] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4313),
  [3682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2555),
  [3684] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2563),
  [3686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2527),
  [3688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2579),
  [3690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3209),
  [3692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5303),
  [3694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6269),
  [3696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2644),
  [3698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3244),
  [3700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [3702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [3704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(569),
  [3706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [3708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(571),
  [3710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(572),
  [3712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2590),
  [3714] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2556),
  [3716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type, 1, 0, 0),
  [3718] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type, 1, 0, 0),
  [3720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2808),
  [3722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3294),
  [3724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [3726] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3517),
  [3728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3040),
  [3730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2608),
  [3732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2609),
  [3734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2599),
  [3736] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3461),
  [3738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2603),
  [3740] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_type_rhs, 1, 0, 0),
  [3742] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_type_rhs, 1, 0, 0),
  [3744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3256),
  [3746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3296),
  [3748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6101),
  [3750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2584),
  [3752] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2551),
  [3754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4058),
  [3756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructed_type, 2, 0, 0),
  [3758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructed_type, 2, 0, 0),
  [3760] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_path, 3, 0, 124),
  [3762] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_path, 3, 0, 124),
  [3764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1348),
  [3766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5027),
  [3768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4073),
  [3770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4524),
  [3772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4798),
  [3774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3844),
  [3776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3950),
  [3778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4807),
  [3780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4930),
  [3782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1438),
  [3784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3314),
  [3786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6183),
  [3788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2550),
  [3790] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2534),
  [3792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4072),
  [3794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 3, 0, 0),
  [3796] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 3, 0, 0),
  [3798] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 3, 0, 0),
  [3800] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 3, 0, 0),
  [3802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2530),
  [3804] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 4, 0, 21),
  [3806] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 4, 0, 21),
  [3808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6256),
  [3810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2588),
  [3812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2581),
  [3814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 5, 0, 167),
  [3816] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 5, 0, 167),
  [3818] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(1688),
  [3821] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(1348),
  [3824] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5027),
  [3827] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4073),
  [3830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0),
  [3832] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4524),
  [3835] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4798),
  [3838] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(3844),
  [3841] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(3950),
  [3844] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4807),
  [3847] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4930),
  [3850] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5661),
  [3853] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5663),
  [3856] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5700),
  [3859] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(1438),
  [3862] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(6100),
  [3865] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 2, 0, 0),
  [3867] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 2, 0, 0),
  [3869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 5, 0, 0),
  [3871] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 5, 0, 0),
  [3873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructed_type, 5, 0, 0),
  [3875] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructed_type, 5, 0, 0),
  [3877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 5, 0, 0),
  [3879] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 5, 0, 0),
  [3881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 5, 0, 0),
  [3883] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 5, 0, 0),
  [3885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructed_type, 4, 0, 0),
  [3887] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructed_type, 4, 0, 0),
  [3889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 4, 0, 0),
  [3891] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 4, 0, 0),
  [3893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [3895] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_type, 1, 0, 0),
  [3897] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_type, 1, 0, 0),
  [3899] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 3, 0, 0),
  [3901] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 3, 0, 0),
  [3903] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_type, 3, 0, 100),
  [3905] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_type, 3, 0, 100),
  [3907] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 6, 0, 198),
  [3909] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 6, 0, 198),
  [3911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 6, 0, 0),
  [3913] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 6, 0, 0),
  [3915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 6, 0, 0),
  [3917] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 6, 0, 0),
  [3919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 6, 0, 0),
  [3921] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 6, 0, 0),
  [3923] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_type, 1, 0, 0),
  [3925] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_type, 1, 0, 0),
  [3927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 4, 0, 0),
  [3929] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 4, 0, 0),
  [3931] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 2, 0, 0),
  [3933] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 2, 0, 0),
  [3935] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 2, 0, 0),
  [3937] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 2, 0, 0),
  [3939] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_path, 1, 0, 52),
  [3941] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_path, 1, 0, 52),
  [3943] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 7, 0, 217),
  [3945] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 7, 0, 217),
  [3947] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 7, 0, 0),
  [3949] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 7, 0, 0),
  [3951] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sign_operator, 1, 0, 0),
  [3953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sign_operator, 1, 0, 0),
  [3955] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_add_operator, 1, 0, 0),
  [3957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_toplevel_directive, 1, 0, 0),
  [3959] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_toplevel_directive, 1, 0, 0),
  [3961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2943),
  [3963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4089),
  [3965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6228),
  [3967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4096),
  [3969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4600),
  [3971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4174),
  [3973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2969),
  [3975] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute, 2, 0, 0),
  [3977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute, 2, 0, 0),
  [3979] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__labeled_tuple_type, 3, 0, 0),
  [3981] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__labeled_tuple_type, 3, 0, 0),
  [3983] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_aliased_type, 3, 0, 99),
  [3985] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_aliased_type, 3, 0, 99),
  [3987] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pow_operator, 1, 0, 0),
  [3989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pow_operator, 1, 0, 0),
  [3991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [3993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [3995] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_add_operator, 1, 0, 0),
  [3997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_type_rhs, 3, 0, 0),
  [3999] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_type_rhs, 3, 0, 0),
  [4001] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 5, 0, 168),
  [4003] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 5, 0, 168),
  [4005] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 3, 0, 47),
  [4007] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 3, 0, 47),
  [4009] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mult_operator, 1, 0, 0),
  [4011] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mult_operator, 1, 0, 0),
  [4013] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_or_operator, 1, 0, 0),
  [4015] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_or_operator, 1, 0, 0),
  [4017] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constructor_declaration_repeat1, 1, 0, 0),
  [4019] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__proper_tuple_type, 3, 0, 0),
  [4021] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__proper_tuple_type, 3, 0, 0),
  [4023] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_and_operator, 1, 0, 0),
  [4025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_and_operator, 1, 0, 0),
  [4027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2583),
  [4029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2835),
  [4031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2824),
  [4033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5993),
  [4035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2828),
  [4037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2823),
  [4039] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(5993),
  [4042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4601),
  [4044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7289),
  [4046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1497),
  [4048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(581),
  [4050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2778),
  [4052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [4054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2792),
  [4056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2814),
  [4058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4313),
  [4060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor, 4, 0, 79),
  [4062] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor, 4, 0, 79),
  [4064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3918),
  [4066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5727),
  [4068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5863),
  [4070] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_of, 4, 0, 44),
  [4072] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_of, 4, 0, 44),
  [4074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2997),
  [4076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(628),
  [4078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [4080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [4082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(630),
  [4084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(728),
  [4086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [4088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(656),
  [4090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(657),
  [4092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [4094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(659),
  [4096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(554),
  [4098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(555),
  [4100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(553),
  [4102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(556),
  [4104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(557),
  [4106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [4108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3041),
  [4110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5289),
  [4112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [4114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [4116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6130),
  [4118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1494),
  [4120] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(6130),
  [4123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [4125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [4127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(541),
  [4129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(542),
  [4131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [4133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [4135] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 42),
  [4137] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 42),
  [4139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3287),
  [4141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5282),
  [4143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [4145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6725),
  [4147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6342),
  [4149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2834),
  [4151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3288),
  [4153] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_path, 3, 0, 38),
  [4155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4214),
  [4157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7271),
  [4159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6602),
  [4161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_expression, 1, 0, 0),
  [4163] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__module_expression, 1, 0, 0),
  [4165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6715),
  [4167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 5, 0, 115),
  [4169] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 5, 0, 115),
  [4171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2280),
  [4173] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 75),
  [4175] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 75),
  [4177] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 76),
  [4179] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 76),
  [4181] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_path, 1, 0, 5),
  [4183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(669),
  [4185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [4187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(668),
  [4189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [4191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [4193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(599),
  [4195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(600),
  [4197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(601),
  [4199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(598),
  [4201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(602),
  [4203] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_module_expression, 3, 0, 0),
  [4205] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_module_expression, 3, 0, 0),
  [4207] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure, 3, 0, 0),
  [4209] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_structure, 3, 0, 0),
  [4211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1313),
  [4213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1374),
  [4215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5465),
  [4217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1378),
  [4219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_application, 3, 0, 45),
  [4221] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_application, 3, 0, 45),
  [4223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [4225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [4227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(643),
  [4229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(644),
  [4231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(645),
  [4233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [4235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [4237] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure, 2, 0, 0),
  [4239] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_structure, 2, 0, 0),
  [4241] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_application, 2, 0, 19),
  [4243] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_application, 2, 0, 19),
  [4245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [4247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3456),
  [4249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5013),
  [4251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4534),
  [4253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4771),
  [4255] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_module_expression, 1, 0, 0),
  [4257] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_module_expression, 1, 0, 0),
  [4259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature, 1, 0, 0),
  [4261] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4201),
  [4264] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5013),
  [4267] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4079),
  [4270] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0),
  [4272] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4534),
  [4275] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4768),
  [4278] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(3881),
  [4281] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(3534),
  [4284] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4727),
  [4287] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4917),
  [4290] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5661),
  [4293] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5663),
  [4296] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5700),
  [4299] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 5, 0, 116),
  [4301] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 5, 0, 116),
  [4303] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 7, 0, 184),
  [4305] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 7, 0, 184),
  [4307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3300),
  [4309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3126),
  [4311] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3793),
  [4313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3785),
  [4315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4516),
  [4317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3807),
  [4319] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 6, 0, 149),
  [4321] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 6, 0, 149),
  [4323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 4, 0, 77),
  [4325] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 4, 0, 77),
  [4327] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_module_expression, 4, 0, 78),
  [4329] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_module_expression, 4, 0, 78),
  [4331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2457),
  [4333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2775),
  [4335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_content, 1, 0, 0),
  [4337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(506),
  [4339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 2, 0, 7),
  [4341] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 2, 0, 7),
  [4343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 3, 0, 18),
  [4345] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 3, 0, 18),
  [4347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(640),
  [4349] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 2, 0, 7),
  [4351] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 2, 0, 7),
  [4353] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 3, 0, 18),
  [4355] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 3, 0, 18),
  [4357] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature, 2, 0, 0),
  [4359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 3, 0, 0),
  [4361] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 3, 0, 0),
  [4363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_argument, 1, 0, 0),
  [4365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3446),
  [4367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [4369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1300),
  [4371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(714),
  [4373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [4375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(716),
  [4377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(717),
  [4379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3041),
  [4381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2719),
  [4383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(718),
  [4385] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5007),
  [4388] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4542),
  [4391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 4, 0, 44),
  [4393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 4, 0, 44),
  [4395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [4397] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1300),
  [4400] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 5, 0, 0),
  [4402] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 5, 0, 0),
  [4404] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_module, 4, 0, 151),
  [4406] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_module, 4, 0, 151),
  [4408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6246),
  [4410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6609),
  [4412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1316),
  [4414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4879),
  [4416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2554),
  [4418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4799),
  [4420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2748),
  [4422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7134),
  [4424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2622),
  [4426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6938),
  [4428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2996),
  [4430] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 4, 0, 135),
  [4432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6601),
  [4434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2642),
  [4436] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1316),
  [4439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_argument, 1, 0, 0),
  [4441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3460),
  [4443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7291),
  [4445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4467),
  [4447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6674),
  [4449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4488),
  [4451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [4453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [4455] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__sequence_expression_content_repeat1, 2, 0, 0),
  [4457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6516),
  [4459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3074),
  [4461] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variant_declaration, 1, 0, 0),
  [4463] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variant_declaration, 1, 0, 0),
  [4465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6182),
  [4467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7131),
  [4469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3438),
  [4471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3418),
  [4473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [4475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(684),
  [4477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6850),
  [4479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3135),
  [4481] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 3, 0, 64),
  [4483] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extended_module_path, 3, 0, 38),
  [4485] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extended_module_path, 3, 0, 38),
  [4487] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_expression, 3, 0, 63),
  [4489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6453),
  [4491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2645),
  [4493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6963),
  [4495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4454),
  [4497] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extended_module_path, 4, 0, 0),
  [4499] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extended_module_path, 4, 0, 0),
  [4501] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extended_module_path, 1, 0, 5),
  [4503] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extended_module_path, 1, 0, 5),
  [4505] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_path, 3, 0, 16),
  [4507] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_path, 3, 0, 16),
  [4509] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_constraint, 4, 0, 111),
  [4511] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_constraint, 4, 0, 111),
  [4513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6071),
  [4515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2531),
  [4517] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 2, 0, 5),
  [4519] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 2, 0, 5),
  [4521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4111),
  [4523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5558),
  [4525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3823),
  [4527] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 1, 0, 5),
  [4529] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 1, 0, 5),
  [4531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4126),
  [4533] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_path, 1, 0, 8),
  [4535] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_path, 1, 0, 8),
  [4537] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0),
  [4539] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0),
  [4541] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 5, 0, 185),
  [4543] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 5, 0, 185),
  [4545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2782),
  [4547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4357),
  [4549] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 6, 0, 208),
  [4551] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 6, 0, 208),
  [4553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2586),
  [4555] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 1, 0, 3),
  [4557] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 1, 0, 3),
  [4559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2136),
  [4561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2269),
  [4563] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_open_class_expression, 4, 0, 79),
  [4565] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_open_class_expression, 4, 0, 79),
  [4567] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_function, 4, 0, 79),
  [4569] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_function, 4, 0, 79),
  [4571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2811),
  [4573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(550),
  [4575] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 1, 0, 0),
  [4577] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_binding_repeat1, 1, 0, 0),
  [4579] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_class_expression, 3, 0, 64),
  [4581] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_class_expression, 3, 0, 64),
  [4583] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0),
  [4585] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0),
  [4587] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5558),
  [4590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3042),
  [4592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 2, 0, 14),
  [4594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 2, 0, 14),
  [4596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2125),
  [4598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2298),
  [4600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 4, 0, 150),
  [4602] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 4, 0, 150),
  [4604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3196),
  [4606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(612),
  [4608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1315),
  [4610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 2, 0, 0),
  [4612] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 2, 0, 0),
  [4614] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1315),
  [4617] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 5, 0, 150),
  [4619] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 5, 0, 150),
  [4621] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 3, 0, 0),
  [4623] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 3, 0, 0),
  [4625] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_directive, 2, 0, 0),
  [4627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_directive, 2, 0, 0),
  [4629] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_module_type, 5, 0, 186),
  [4631] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_module_type, 5, 0, 186),
  [4633] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 7, 0, 0),
  [4635] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 7, 0, 0),
  [4637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3834),
  [4639] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 6, 0, 185),
  [4641] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 6, 0, 185),
  [4643] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 7, 0, 208),
  [4645] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 7, 0, 208),
  [4647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [4649] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0),
  [4651] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0),
  [4653] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(5662),
  [4656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1312),
  [4658] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_type, 1, 0, 0),
  [4660] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__module_type, 1, 0, 0),
  [4662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5956),
  [4664] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1312),
  [4667] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0),
  [4669] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2811),
  [4672] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0),
  [4674] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 36),
  [4676] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 36),
  [4678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4241),
  [4680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(655),
  [4682] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2782),
  [4685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0),
  [4687] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0),
  [4689] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__polymorphic_type, 1, 0, 0),
  [4691] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__polymorphic_type, 1, 0, 0),
  [4693] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_module_type, 3, 0, 0),
  [4695] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_module_type, 3, 0, 0),
  [4697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 71),
  [4699] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 71),
  [4701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4242),
  [4703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(859),
  [4705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1274),
  [4707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3445),
  [4709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3403),
  [4711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2807),
  [4713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1752),
  [4715] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__maybe_polymorphic_typed, 2, 0, 35),
  [4717] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__maybe_polymorphic_typed, 2, 0, 35),
  [4719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1280),
  [4721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1779),
  [4723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_extension, 5, 0, 0),
  [4725] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_extension, 5, 0, 0),
  [4727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1281),
  [4729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1792),
  [4731] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_item_extension, 5, 0, 0),
  [4733] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_item_extension, 5, 0, 0),
  [4735] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class_type, 5, 0, 0),
  [4737] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class_type, 5, 0, 0),
  [4739] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 5, 0, 195),
  [4741] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 5, 0, 195),
  [4743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2433),
  [4745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1279),
  [4747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3414),
  [4749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2827),
  [4751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3616),
  [4753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1775),
  [4755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5778),
  [4757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5895),
  [4759] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 112),
  [4761] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 112),
  [4763] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 6, 0, 55),
  [4765] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 6, 0, 55),
  [4767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6323),
  [4769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4173),
  [4771] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_extension, 4, 0, 0),
  [4773] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_extension, 4, 0, 0),
  [4775] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 2, 0, 0),
  [4777] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 2, 0, 0),
  [4779] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_class_type, 1, 0, 0),
  [4781] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_class_type, 1, 0, 0),
  [4783] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_open_class_type, 4, 0, 79),
  [4785] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_open_class_type, 4, 0, 79),
  [4787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2135),
  [4789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2289),
  [4791] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 3, 0, 0),
  [4793] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 3, 0, 0),
  [4795] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_declaration_name, 2, 0, 0),
  [4797] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_declaration_name, 2, 0, 0),
  [4799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_parameter, 2, 0, 0),
  [4801] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_parameter, 2, 0, 0),
  [4803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2137),
  [4805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2279),
  [4807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(726),
  [4809] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_extension, 3, 0, 0),
  [4811] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_extension, 3, 0, 0),
  [4813] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class_type, 4, 0, 0),
  [4815] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class_type, 4, 0, 0),
  [4817] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_parameter, 4, 0, 80),
  [4819] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_parameter, 4, 0, 80),
  [4821] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_declaration_name, 1, 0, 0),
  [4823] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_declaration_name, 1, 0, 0),
  [4825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 6, 0, 195),
  [4827] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 6, 0, 195),
  [4829] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_item_extension, 4, 0, 0),
  [4831] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_item_extension, 4, 0, 0),
  [4833] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 5, 0, 26),
  [4835] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 5, 0, 26),
  [4837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 70),
  [4839] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 70),
  [4841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4220),
  [4843] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_item_extension, 6, 0, 0),
  [4845] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_item_extension, 6, 0, 0),
  [4847] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 1, 0, 0),
  [4849] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_binding_repeat1, 1, 0, 0),
  [4851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4233),
  [4853] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 2, 0, 17),
  [4855] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 2, 0, 17),
  [4857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4105),
  [4859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4108),
  [4861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 2, 0, 3),
  [4863] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 2, 0, 3),
  [4865] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 183),
  [4867] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 183),
  [4869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 148),
  [4871] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 148),
  [4873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 14),
  [4875] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 14),
  [4877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 147),
  [4879] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 147),
  [4881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 147),
  [4883] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 147),
  [4885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 8, 0, 183),
  [4887] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 8, 0, 183),
  [4889] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5060),
  [4892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4252),
  [4894] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 70),
  [4896] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 70),
  [4898] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_attribute, 3, 0, 0),
  [4900] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_attribute, 3, 0, 0),
  [4902] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 110),
  [4904] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 110),
  [4906] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 148),
  [4908] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 148),
  [4910] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 36),
  [4912] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 36),
  [4914] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 110),
  [4916] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 110),
  [4918] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 1, 0, 0),
  [4920] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 1, 0, 0),
  [4922] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 1, 0, 0),
  [4924] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 1, 0, 0),
  [4926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5124),
  [4928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3248),
  [4930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3167),
  [4932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_constraint, 4, 0, 48),
  [4934] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_constraint, 4, 0, 48),
  [4936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5060),
  [4938] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_attribute, 4, 0, 0),
  [4940] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_attribute, 4, 0, 0),
  [4942] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 112),
  [4944] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 112),
  [4946] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_constraint, 3, 0, 48),
  [4948] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_constraint, 3, 0, 48),
  [4950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6250),
  [4952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 43),
  [4954] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 43),
  [4956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4129),
  [4958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4140),
  [4960] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_type, 3, 0, 100),
  [4962] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_type, 3, 0, 100),
  [4964] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 71),
  [4966] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 71),
  [4968] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_type, 3, 0, 108),
  [4970] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_type, 3, 0, 108),
  [4972] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_declaration, 3, 0, 0),
  [4974] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_declaration, 3, 0, 0),
  [4976] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_declaration, 5, 0, 0),
  [4978] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_declaration, 5, 0, 0),
  [4980] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 4, 0, 81),
  [4982] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 4, 0, 81),
  [4984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3828),
  [4986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5081),
  [4988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3324),
  [4990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3523),
  [4992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3818),
  [4994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7268),
  [4996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3069),
  [4998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5539),
  [5000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5612),
  [5002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4947),
  [5004] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_typed, 2, 0, 9),
  [5006] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__module_typed, 2, 0, 9),
  [5008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_argument, 2, 0, 0),
  [5010] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_argument, 2, 0, 0),
  [5012] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 2, 0, 25),
  [5014] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 2, 0, 25),
  [5016] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 2, 0, 20),
  [5018] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 2, 0, 20),
  [5020] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 163),
  [5022] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 163),
  [5024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3397),
  [5026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6287),
  [5028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4782),
  [5030] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 165),
  [5032] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 165),
  [5034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 128),
  [5036] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 128),
  [5038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 196),
  [5040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 196),
  [5042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 164),
  [5044] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 164),
  [5046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 7, 0, 196),
  [5048] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 7, 0, 196),
  [5050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 127),
  [5052] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 127),
  [5054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 164),
  [5056] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 164),
  [5058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4985),
  [5060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3825),
  [5062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5265),
  [5064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3249),
  [5066] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(5868),
  [5069] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 3, 0, 91),
  [5071] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 3, 0, 91),
  [5073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4823),
  [5075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 126),
  [5077] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 126),
  [5079] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 165),
  [5081] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 165),
  [5083] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_declaration, 4, 0, 0),
  [5085] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_declaration, 4, 0, 0),
  [5087] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 128),
  [5089] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 128),
  [5091] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0), SHIFT_REPEAT(3460),
  [5094] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 126),
  [5096] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 126),
  [5098] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 3, 0, 46),
  [5100] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 3, 0, 46),
  [5102] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 163),
  [5104] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 163),
  [5106] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 91),
  [5108] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 91),
  [5110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 3, 0, 47),
  [5112] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 3, 0, 47),
  [5114] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 127),
  [5116] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 127),
  [5118] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0),
  [5120] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0),
  [5122] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(6323),
  [5125] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(4173),
  [5128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 73),
  [5130] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 73),
  [5132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 14),
  [5134] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 14),
  [5136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 72),
  [5138] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 72),
  [5140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 39),
  [5142] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 39),
  [5144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 42),
  [5146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 42),
  [5148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 43),
  [5150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 43),
  [5152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5079),
  [5154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 3, 0, 82),
  [5156] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 3, 0, 82),
  [5158] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 84),
  [5160] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 84),
  [5162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 49),
  [5164] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 49),
  [5166] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 22),
  [5168] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 22),
  [5170] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 50),
  [5172] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 50),
  [5174] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 88),
  [5176] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 88),
  [5178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 90),
  [5180] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 90),
  [5182] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 54),
  [5184] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 54),
  [5186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1377),
  [5188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2809),
  [5190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5026),
  [5192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(805),
  [5194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3348),
  [5196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5233),
  [5198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [5200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6041),
  [5202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 3),
  [5204] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 3),
  [5206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1383),
  [5208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1740),
  [5210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 39),
  [5212] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 39),
  [5214] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 17),
  [5216] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 17),
  [5218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 5),
  [5220] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 5),
  [5222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1385),
  [5224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2353),
  [5226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 36),
  [5228] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 36),
  [5230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0),
  [5232] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0),
  [5234] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 113),
  [5236] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 113),
  [5238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 72),
  [5240] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 72),
  [5242] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 73),
  [5244] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 73),
  [5246] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0), SHIFT_REPEAT(3446),
  [5249] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 5, 0, 75),
  [5251] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 5, 0, 75),
  [5253] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 5, 0, 76),
  [5255] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 5, 0, 76),
  [5257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1389),
  [5259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1909),
  [5261] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5079),
  [5264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 4, 0, 117),
  [5266] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 4, 0, 117),
  [5268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 4, 0, 82),
  [5270] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 4, 0, 82),
  [5272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1391),
  [5274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [5276] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 84),
  [5278] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 84),
  [5280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 120),
  [5282] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 120),
  [5284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 121),
  [5286] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 121),
  [5288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 85),
  [5290] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 85),
  [5292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1393),
  [5294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2434),
  [5296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [5298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1617),
  [5300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_function_type, 3, 0, 47),
  [5302] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_function_type, 3, 0, 47),
  [5304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 88),
  [5306] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 88),
  [5308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1397),
  [5310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(937),
  [5312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 90),
  [5314] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 90),
  [5316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 125),
  [5318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 125),
  [5320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 50),
  [5322] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 50),
  [5324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1401),
  [5326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2746),
  [5328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1403),
  [5330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1982),
  [5332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 2, 0, 24),
  [5334] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 2, 0, 24),
  [5336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3413),
  [5338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_type, 1, 0, 0),
  [5340] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_type, 1, 0, 0),
  [5342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_typed, 2, 0, 53),
  [5344] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_typed, 2, 0, 53),
  [5346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 49),
  [5348] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 49),
  [5350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3442),
  [5352] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 54),
  [5354] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 54),
  [5356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3408),
  [5358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variant_declaration, 2, 0, 0),
  [5360] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variant_declaration, 2, 0, 0),
  [5362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4629),
  [5364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 5, 0, 74),
  [5366] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 5, 0, 74),
  [5368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 85),
  [5370] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 85),
  [5372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3389),
  [5374] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variant_declaration, 3, 0, 0),
  [5376] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variant_declaration, 3, 0, 0),
  [5378] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4629),
  [5381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 6, 0, 114),
  [5383] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 6, 0, 114),
  [5385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 118),
  [5387] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 118),
  [5389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3371),
  [5391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 153),
  [5393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 153),
  [5395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3401),
  [5397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 154),
  [5399] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 154),
  [5401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3415),
  [5403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 191),
  [5405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 191),
  [5407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3427),
  [5409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 192),
  [5411] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 192),
  [5413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3431),
  [5415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 24),
  [5417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 24),
  [5419] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 214),
  [5421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 214),
  [5423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3439),
  [5425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 10),
  [5427] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 10),
  [5429] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 70),
  [5431] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 70),
  [5433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 71),
  [5435] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 71),
  [5437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 113),
  [5439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 113),
  [5441] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 6, 0, 115),
  [5443] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 6, 0, 115),
  [5445] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 5, 0, 117),
  [5447] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 5, 0, 117),
  [5449] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 152),
  [5451] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 152),
  [5453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 118),
  [5455] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 118),
  [5457] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 83),
  [5459] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 83),
  [5461] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 120),
  [5463] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 120),
  [5465] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 121),
  [5467] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 121),
  [5469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 156),
  [5471] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 156),
  [5473] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_floating_attribute, 3, 0, 0),
  [5475] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_floating_attribute, 3, 0, 0),
  [5477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1382),
  [5479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1127),
  [5481] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 125),
  [5483] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 125),
  [5485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5984),
  [5487] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(5984),
  [5490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5106),
  [5492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3215),
  [5494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3147),
  [5496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 110),
  [5498] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 110),
  [5500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 112),
  [5502] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 112),
  [5504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 6, 0, 187),
  [5506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 6, 0, 187),
  [5508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 152),
  [5510] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 152),
  [5512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 188),
  [5514] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 188),
  [5516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 189),
  [5518] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 189),
  [5520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 153),
  [5522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 153),
  [5524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 190),
  [5526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 190),
  [5528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 154),
  [5530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 154),
  [5532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 119),
  [5534] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 119),
  [5536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 156),
  [5538] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 156),
  [5540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_function_type, 5, 0, 168),
  [5542] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_function_type, 5, 0, 168),
  [5544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 8, 0, 147),
  [5546] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 8, 0, 147),
  [5548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 8, 0, 148),
  [5550] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 8, 0, 148),
  [5552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 7, 0, 187),
  [5554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 7, 0, 187),
  [5556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 7, 0, 209),
  [5558] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 7, 0, 209),
  [5560] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 188),
  [5562] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 188),
  [5564] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 189),
  [5566] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 189),
  [5568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 210),
  [5570] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 210),
  [5572] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 190),
  [5574] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 190),
  [5576] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 211),
  [5578] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 211),
  [5580] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 212),
  [5582] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 212),
  [5584] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 191),
  [5586] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 191),
  [5588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 213),
  [5590] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 213),
  [5592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 192),
  [5594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 192),
  [5596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 155),
  [5598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 155),
  [5600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 9, 0, 183),
  [5602] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 9, 0, 183),
  [5604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 8, 0, 209),
  [5606] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 8, 0, 209),
  [5608] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 8, 0, 223),
  [5610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 8, 0, 223),
  [5612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 210),
  [5614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 210),
  [5616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 211),
  [5618] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 211),
  [5620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 212),
  [5622] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 212),
  [5624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 224),
  [5626] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 224),
  [5628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 213),
  [5630] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 213),
  [5632] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 225),
  [5634] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 225),
  [5636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 226),
  [5638] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 226),
  [5640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 214),
  [5642] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 214),
  [5644] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 3, 0, 16),
  [5646] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 3, 0, 16),
  [5648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3799),
  [5650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 9, 0, 223),
  [5652] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 9, 0, 223),
  [5654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 224),
  [5656] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 224),
  [5658] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 225),
  [5660] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 225),
  [5662] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 226),
  [5664] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 226),
  [5666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 4, 0, 41),
  [5668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 4, 0, 41),
  [5670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3806),
  [5672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 229),
  [5674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 229),
  [5676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 11, 0, 229),
  [5678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 11, 0, 229),
  [5680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_floating_attribute, 4, 0, 0),
  [5682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_floating_attribute, 4, 0, 0),
  [5684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3826),
  [5686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7227),
  [5688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5521),
  [5690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5727),
  [5692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5863),
  [5694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4948),
  [5696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 3, 0, 7),
  [5698] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 3, 0, 7),
  [5700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_definition, 4, 0, 0),
  [5702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_definition, 4, 0, 0),
  [5704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 4, 0, 16),
  [5706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 4, 0, 16),
  [5708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3795),
  [5710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5108),
  [5712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 4, 0, 18),
  [5714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 4, 0, 18),
  [5716] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6306),
  [5718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 4, 0, 55),
  [5720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 4, 0, 55),
  [5722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 4, 0, 26),
  [5724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 4, 0, 26),
  [5726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 7, 0, 55),
  [5728] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 7, 0, 55),
  [5730] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 3, 0, 0),
  [5732] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 3, 0, 0),
  [5734] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1437),
  [5736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1437),
  [5738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 6, 0, 26),
  [5740] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 6, 0, 26),
  [5742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4730),
  [5744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3531),
  [5746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3812),
  [5748] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7015),
  [5750] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3043),
  [5752] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5479),
  [5754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4585),
  [5756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_declaration, 2, 0, 0),
  [5758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_declaration, 2, 0, 0),
  [5760] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 7, 0, 114),
  [5762] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 7, 0, 114),
  [5764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3791),
  [5766] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5454),
  [5768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 6, 0, 74),
  [5770] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 6, 0, 74),
  [5772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 3, 0, 26),
  [5774] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 3, 0, 26),
  [5776] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 5, 0, 41),
  [5778] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 5, 0, 41),
  [5780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 5, 0, 55),
  [5782] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 5, 0, 55),
  [5784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 4, 0, 0),
  [5786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 4, 0, 0),
  [5788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 2, 0, 0),
  [5790] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 2, 0, 0),
  [5792] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_definition, 2, 0, 0),
  [5794] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_definition, 2, 0, 0),
  [5796] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_definition, 3, 0, 0),
  [5798] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_definition, 3, 0, 0),
  [5800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4986),
  [5802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3779),
  [5804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7328),
  [5806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5444),
  [5808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0),
  [5810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0),
  [5812] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1437),
  [5815] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1437),
  [5818] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 5, 0, 44),
  [5820] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 5, 0, 44),
  [5822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4909),
  [5824] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 5, 0, 0),
  [5826] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 5, 0, 0),
  [5828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 1, 0, 0),
  [5830] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 1, 0, 0),
  [5832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3813),
  [5834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7262),
  [5836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5488),
  [5838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3735),
  [5840] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 3, 0, 7),
  [5842] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 3, 0, 7),
  [5844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3474),
  [5846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3761),
  [5848] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1440),
  [5851] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1440),
  [5854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4065),
  [5856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4006),
  [5858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5904),
  [5860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [5862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3355),
  [5864] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0),
  [5866] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0),
  [5868] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5499),
  [5871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3447),
  [5873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3751),
  [5875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1390),
  [5877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1930),
  [5879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4066),
  [5881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3945),
  [5883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5846),
  [5885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [5887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4485),
  [5889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3450),
  [5891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1440),
  [5893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1440),
  [5895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4675),
  [5897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_definition, 5, 0, 0),
  [5899] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_definition, 5, 0, 0),
  [5901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5499),
  [5903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3396),
  [5905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 3, 0, 21),
  [5907] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 4, 0, 18),
  [5909] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 4, 0, 18),
  [5911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1384),
  [5913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1791),
  [5915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_item, 2, 0, 0),
  [5917] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_item, 2, 0, 0),
  [5919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1404),
  [5921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1985),
  [5923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3470),
  [5925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1392),
  [5927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1007),
  [5929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_item, 1, 0, 0),
  [5931] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_item, 1, 0, 0),
  [5933] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_definition, 4, 0, 0),
  [5935] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_definition, 4, 0, 0),
  [5937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3434),
  [5939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3473),
  [5941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3435),
  [5943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1387),
  [5945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2366),
  [5947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1394),
  [5949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2441),
  [5951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3368),
  [5953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_definition, 3, 0, 0),
  [5955] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_definition, 3, 0, 0),
  [5957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1396),
  [5959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1608),
  [5961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3999),
  [5963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5802),
  [5965] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(6041),
  [5968] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(4178),
  [5971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1399),
  [5973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(940),
  [5975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 2, 0, 9),
  [5977] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4675),
  [5980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1347),
  [5982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [5984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1402),
  [5986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2677),
  [5988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1379),
  [5990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [5992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(952),
  [5994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [5996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(953),
  [5998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2696),
  [6000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2684),
  [6002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2681),
  [6004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2743),
  [6006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2745),
  [6008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2701),
  [6010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2704),
  [6012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2715),
  [6014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2752),
  [6016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1988),
  [6018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1991),
  [6020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1992),
  [6022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1993),
  [6024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1994),
  [6026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1995),
  [6028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1996),
  [6030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1997),
  [6032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1998),
  [6034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0),
  [6036] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0),
  [6038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 3, 0, 0),
  [6040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 3, 0, 0),
  [6042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4221),
  [6044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3007),
  [6046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [6048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 4, 0, 0),
  [6050] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 4, 0, 0),
  [6052] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4221),
  [6055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1129),
  [6057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [6059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1467),
  [6061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1822),
  [6063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1721),
  [6065] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 5, 0, 0),
  [6067] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 5, 0, 0),
  [6069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1712),
  [6071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1778),
  [6073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1781),
  [6075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1783),
  [6077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1142),
  [6079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1704),
  [6081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1707),
  [6083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1144),
  [6085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1708),
  [6087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1711),
  [6089] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0),
  [6091] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0),
  [6093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1138),
  [6095] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0),
  [6097] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0),
  [6099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [6101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1445),
  [6103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1760),
  [6105] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(6306),
  [6108] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(4175),
  [6111] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 2, 0, 0),
  [6113] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 2, 0, 0),
  [6115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5507),
  [6117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [6119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1453),
  [6121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2410),
  [6123] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5507),
  [6126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4523),
  [6128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(584),
  [6130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2377),
  [6132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 4, 0, 0),
  [6134] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 4, 0, 0),
  [6136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [6138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1451),
  [6140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(797),
  [6142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2384),
  [6144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [6146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1469),
  [6148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1075),
  [6150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2391),
  [6152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1081),
  [6154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2392),
  [6156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [6158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2398),
  [6160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2399),
  [6162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2400),
  [6164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1098),
  [6166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1104),
  [6168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2401),
  [6170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [6172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1454),
  [6174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1013),
  [6176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1944),
  [6178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1914),
  [6180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(809),
  [6182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1935),
  [6184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1887),
  [6186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1934),
  [6188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1883),
  [6190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1848),
  [6192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1855),
  [6194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1912),
  [6196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1146),
  [6198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [6200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [6202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [6204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1446),
  [6206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1577),
  [6208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4573),
  [6210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1024),
  [6212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1031),
  [6214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1038),
  [6216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1039),
  [6218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1040),
  [6220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [6222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1044),
  [6224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1045),
  [6226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1046),
  [6228] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 2, 0, 0),
  [6230] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 2, 0, 0),
  [6232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(831),
  [6234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(838),
  [6236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [6238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1462),
  [6240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(912),
  [6242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2444),
  [6244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2447),
  [6246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2448),
  [6248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2449),
  [6250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2424),
  [6252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2425),
  [6254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2426),
  [6256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2427),
  [6258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2428),
  [6260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(845),
  [6262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1121),
  [6264] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(2809),
  [6267] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5026),
  [6270] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0),
  [6272] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(3348),
  [6275] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5233),
  [6278] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(303),
  [6281] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5661),
  [6284] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5663),
  [6287] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5700),
  [6290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(781),
  [6292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [6294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1444),
  [6296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2751),
  [6298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1581),
  [6300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1592),
  [6302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1595),
  [6304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1597),
  [6306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1598),
  [6308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1600),
  [6310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1602),
  [6312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1604),
  [6314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1605),
  [6316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [6318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 3, 0, 0),
  [6320] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 3, 0, 0),
  [6322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(783),
  [6324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [6326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(62),
  [6328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1448),
  [6330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2032),
  [6332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 3, 0, 0),
  [6334] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 3, 0, 0),
  [6336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5924),
  [6338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 4, 0, 0),
  [6340] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 4, 0, 0),
  [6342] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5924),
  [6345] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 5, 0, 0),
  [6347] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 5, 0, 0),
  [6349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [6351] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 2, 0, 0),
  [6353] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 2, 0, 0),
  [6355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(945),
  [6357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(946),
  [6359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(947),
  [6361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(948),
  [6363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2393),
  [6365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2830),
  [6367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5160),
  [6369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5399),
  [6371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4194),
  [6373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5305),
  [6375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5122),
  [6377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [6379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2795),
  [6381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 3, 0, 0),
  [6383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_toplevel_directive, 2, 0, 0),
  [6385] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_toplevel_directive, 2, 0, 0),
  [6387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5228),
  [6389] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 4, 0, 21),
  [6391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(489),
  [6393] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 0),
  [6395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [6397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 2, 0, 0),
  [6399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2820),
  [6401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3374),
  [6403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3257),
  [6405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4626),
  [6407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4209),
  [6409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5066),
  [6411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5070),
  [6413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4484),
  [6415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4219),
  [6417] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure_item, 1, 0, 0),
  [6419] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure_item, 1, 0, 0),
  [6421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(288),
  [6423] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 3, 0, 9),
  [6425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3400),
  [6427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4225),
  [6429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5310),
  [6431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5378),
  [6433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4206),
  [6435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5760),
  [6437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3387),
  [6439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3444),
  [6441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5286),
  [6443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5833),
  [6445] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5760),
  [6448] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5378),
  [6451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3360),
  [6453] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4206),
  [6456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 1, 0, 0),
  [6458] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat2, 1, 0, 0),
  [6460] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(2809),
  [6463] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5160),
  [6466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0),
  [6468] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(4194),
  [6471] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5305),
  [6474] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5661),
  [6477] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5663),
  [6480] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5700),
  [6483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5302),
  [6485] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__local_structure_item, 1, 0, 0),
  [6487] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__local_structure_item, 1, 0, 0),
  [6489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5918),
  [6491] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0),
  [6493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2485),
  [6495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4279),
  [6497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4279),
  [6499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4304),
  [6501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4742),
  [6503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4192),
  [6505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2025),
  [6507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1019),
  [6509] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0), SHIFT_REPEAT(4201),
  [6512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3949),
  [6514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3836),
  [6516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3758),
  [6518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2233),
  [6520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1834),
  [6522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1122),
  [6524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1913),
  [6526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1569),
  [6528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4210),
  [6530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3805),
  [6532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4199),
  [6534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2954),
  [6536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2358),
  [6538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2697),
  [6540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(794),
  [6542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(933),
  [6544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1898),
  [6546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2646),
  [6548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2785),
  [6550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3742),
  [6552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2779),
  [6554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2796),
  [6556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3274),
  [6558] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__instance_variable_name, 1, 0, 11),
  [6560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5363),
  [6562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2829),
  [6564] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature_item, 1, 0, 0),
  [6566] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0),
  [6568] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4279),
  [6571] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4279),
  [6574] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4304),
  [6577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2951),
  [6579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6294),
  [6581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3280),
  [6583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4736),
  [6585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6263),
  [6587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1868),
  [6589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6066),
  [6591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4083),
  [6593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6193),
  [6595] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_content, 1, 0, 0),
  [6597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 1, 0, 0),
  [6599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4278),
  [6601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3654),
  [6603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5666),
  [6605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6290),
  [6607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4267),
  [6609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4627),
  [6611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6560),
  [6613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4649),
  [6615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5938),
  [6617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4650),
  [6619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4269),
  [6621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3501),
  [6623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4273),
  [6625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3486),
  [6627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4289),
  [6629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4628),
  [6631] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__typed, 2, 0, 35),
  [6633] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__typed, 2, 0, 35),
  [6635] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_constrained, 2, 0, 95),
  [6637] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_constrained, 2, 0, 95),
  [6639] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_constrained, 1, 0, 56),
  [6641] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_constrained, 1, 0, 56),
  [6643] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_constrained, 1, 0, 57),
  [6645] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_constrained, 1, 0, 57),
  [6647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4270),
  [6649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3594),
  [6651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1260),
  [6653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5229),
  [6655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5732),
  [6657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4414),
  [6659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1261),
  [6661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5062),
  [6663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1258),
  [6665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5171),
  [6667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5671),
  [6669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4810),
  [6671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1259),
  [6673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5192),
  [6675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4220),
  [6677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5316),
  [6679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1272),
  [6681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1548),
  [6683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1273),
  [6685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__coerced, 2, 0, 92),
  [6687] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__coerced, 2, 0, 92),
  [6689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4638),
  [6691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3489),
  [6693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3648),
  [6695] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_constraint, 4, 0, 111),
  [6697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [6699] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 137),
  [6701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3485),
  [6703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1242),
  [6705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5109),
  [6707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5847),
  [6709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2435),
  [6711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [6713] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 139),
  [6715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1249),
  [6717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5044),
  [6719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5564),
  [6721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [6723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 142),
  [6725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3678),
  [6727] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 1, 0, 0),
  [6729] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_string_content_repeat1, 1, 0, 0),
  [6731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3252),
  [6733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1250),
  [6735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5146),
  [6737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5686),
  [6739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1690),
  [6741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [6743] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 171),
  [6745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [6747] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 175),
  [6749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [6751] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 176),
  [6753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4673),
  [6755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1246),
  [6757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1247),
  [6759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [6761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3204),
  [6763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1233),
  [6765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [6767] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 204),
  [6769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4227),
  [6771] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 2, 0, 59),
  [6773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5330),
  [6775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(499),
  [6777] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 3, 0, 102),
  [6779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2950),
  [6781] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 3, 0, 103),
  [6783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5362),
  [6785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1256),
  [6787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3153),
  [6789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1257),
  [6791] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_escape_sequence, 1, 0, 0),
  [6793] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_escape_sequence, 1, 0, 0),
  [6795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [6797] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 3, 0, 105),
  [6799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2841),
  [6801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3292),
  [6803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5338),
  [6805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1243),
  [6807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5181),
  [6809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5765),
  [6811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2747),
  [6813] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 141),
  [6815] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_pattern, 3, 0, 0),
  [6817] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_pattern, 3, 0, 0),
  [6819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 105),
  [6821] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 60),
  [6823] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__or_pattern_anonymous, 3, 0, 0),
  [6825] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__or_pattern_anonymous, 3, 0, 0),
  [6827] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias_pattern, 3, 0, 93),
  [6829] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alias_pattern, 3, 0, 93),
  [6831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1028),
  [6833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3793),
  [6835] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_pattern, 3, 0, 89),
  [6837] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_pattern, 3, 0, 89),
  [6839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2685),
  [6841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1123),
  [6843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2374),
  [6845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2724),
  [6847] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_pattern, 5, 0, 162),
  [6849] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_pattern, 5, 0, 162),
  [6851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2991),
  [6853] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_pattern, 2, 0, 51),
  [6855] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_pattern, 2, 0, 51),
  [6857] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_constraint, 5, 0, 111),
  [6859] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 129),
  [6861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 137),
  [6863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1961),
  [6865] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 0),
  [6867] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 79),
  [6869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 172),
  [6871] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 5, 0, 173),
  [6873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 5, 0, 138),
  [6875] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 174),
  [6877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 139),
  [6879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1967),
  [6881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 104),
  [6883] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 141),
  [6885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 177),
  [6887] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 178),
  [6889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 142),
  [6891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3464),
  [6893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_pattern, 2, 0, 51),
  [6895] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_pattern, 2, 0, 51),
  [6897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_pattern, 2, 0, 51),
  [6899] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tag_pattern, 2, 0, 51),
  [6901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1080),
  [6903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1939),
  [6905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(846),
  [6907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1620),
  [6909] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 129),
  [6911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 199),
  [6913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 200),
  [6915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 171),
  [6917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 172),
  [6919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 6, 0, 173),
  [6921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 174),
  [6923] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 201),
  [6925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 202),
  [6927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 175),
  [6929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 203),
  [6931] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 176),
  [6933] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__effect_pattern, 1, 0, 0),
  [6935] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__effect_pattern, 1, 0, 0),
  [6937] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_pattern, 4, 0, 122),
  [6939] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_pattern, 4, 0, 122),
  [6941] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 140),
  [6943] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 177),
  [6945] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 178),
  [6947] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 205),
  [6949] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1, 0, 28),
  [6951] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern, 1, 0, 28),
  [6953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1, 0, 29),
  [6955] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern, 1, 0, 29),
  [6957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 7, 0, 199),
  [6959] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 7, 0, 200),
  [6961] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 7, 0, 218),
  [6963] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 201),
  [6965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 202),
  [6967] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 219),
  [6969] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 203),
  [6971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 220),
  [6973] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 221),
  [6975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_effect_pattern, 4, 0, 131),
  [6977] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_effect_pattern, 4, 0, 131),
  [6979] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 204),
  [6981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 205),
  [6983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1549),
  [6985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2381),
  [6987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_initializer, 2, 0, 61),
  [6989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 8, 0, 218),
  [6991] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 219),
  [6993] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 220),
  [6995] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 221),
  [6997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 228),
  [6999] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_pattern, 3, 0, 89),
  [7001] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_pattern, 3, 0, 89),
  [7003] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 9, 0, 228),
  [7005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(789),
  [7007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4149),
  [7009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5932),
  [7011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3819),
  [7013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1824),
  [7015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1021),
  [7017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(909),
  [7019] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 3, 0, 59),
  [7021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1767),
  [7023] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 79),
  [7025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(917),
  [7027] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 102),
  [7029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 4, 0, 103),
  [7031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4138),
  [7033] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 4, 0, 138),
  [7035] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_initializer, 3, 0, 61),
  [7037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1736),
  [7039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4509),
  [7041] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1, 0, 0),
  [7043] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern, 1, 0, 0),
  [7045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2826),
  [7047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6219),
  [7049] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_pattern, 2, 0, 27),
  [7051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2291),
  [7053] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_pattern, 2, 0, 27),
  [7055] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 3, 0, 26),
  [7057] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_specification, 3, 0, 53),
  [7059] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 3, 0, 105),
  [7061] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_specification, 2, 0, 53),
  [7063] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 4, 0, 26),
  [7065] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 4, 0, 55),
  [7067] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 4, 0, 105),
  [7069] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 4, 0, 139),
  [7071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4532),
  [7073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4525),
  [7075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 5, 0, 55),
  [7077] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 5, 0, 139),
  [7079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4502),
  [7081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2138),
  [7083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2286),
  [7085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6934),
  [7087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4511),
  [7089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4498),
  [7091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2132),
  [7093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2300),
  [7095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4515),
  [7097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4243),
  [7099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1303),
  [7101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1338),
  [7103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [7105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2497),
  [7107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1337),
  [7109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2547),
  [7111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5402),
  [7113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4235),
  [7115] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_pattern, 3, 0, 69),
  [7117] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_pattern, 3, 0, 69),
  [7119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2430),
  [7121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2739),
  [7123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3224),
  [7125] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2826),
  [7128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_pattern, 3, 0, 89),
  [7130] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_pattern, 3, 0, 89),
  [7132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_pattern, 2, 0, 51),
  [7134] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_pattern, 2, 0, 51),
  [7136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4200),
  [7138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2673),
  [7140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3390),
  [7142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3381),
  [7144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3120),
  [7146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2759),
  [7148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [7150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2246),
  [7152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(857),
  [7154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [7156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(418),
  [7158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [7160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5886),
  [7162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(964),
  [7164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [7166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(404),
  [7168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [7170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6844),
  [7172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6844),
  [7174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7118),
  [7176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1919),
  [7178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [7180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(375),
  [7182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [7184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [7186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [7188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(340),
  [7190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [7192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1984),
  [7194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [7196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(432),
  [7198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [7200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4761),
  [7202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4761),
  [7204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6863),
  [7206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1056),
  [7208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [7210] = {.entry = {.count = 1, .reusable = false}}, SHIFT(282),
  [7212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [7214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4521),
  [7216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6234),
  [7218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7298),
  [7220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1531),
  [7222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [7224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(411),
  [7226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [7228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2407),
  [7230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [7232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(397),
  [7234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(398),
  [7236] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5932),
  [7239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2770),
  [7241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(424),
  [7243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(425),
  [7245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [7247] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_field, 1, 0, 0),
  [7249] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 1, 0, 0),
  [7251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1835),
  [7253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [7255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(387),
  [7257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [7259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5070),
  [7261] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_field_specification, 1, 0, 0),
  [7263] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_quoted_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4761),
  [7266] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_quoted_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4761),
  [7269] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_quoted_string_content_repeat1, 2, 0, 0),
  [7271] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5193),
  [7274] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 1, 0, 0),
  [7276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2595),
  [7278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5193),
  [7280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2303),
  [7282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5057),
  [7284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3231),
  [7286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3220),
  [7288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3072),
  [7290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3229),
  [7292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4483),
  [7294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_string_content, 1, 0, 0),
  [7296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(551),
  [7298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2833),
  [7300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 1, 0, 0),
  [7302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2832),
  [7304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3201),
  [7306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2605),
  [7308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2294),
  [7310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [7312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [7314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 3, 0, 98),
  [7316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1288),
  [7318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1350),
  [7320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1305),
  [7322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5356),
  [7324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1351),
  [7326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_binding_pattern_content, 1, 0, 0),
  [7328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3907),
  [7330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3910),
  [7332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3122),
  [7334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1290),
  [7336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1326),
  [7338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5501),
  [7340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1327),
  [7342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [7344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4521),
  [7346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6325),
  [7348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3093),
  [7350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2327),
  [7352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4021),
  [7354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3994),
  [7356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4000),
  [7358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_payload, 2, 0, 0),
  [7360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [7362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1285),
  [7364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1365),
  [7366] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_pattern_content, 1, 0, 0),
  [7368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [7370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1366),
  [7372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6281),
  [7374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2197),
  [7376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3821),
  [7378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5144),
  [7380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2647),
  [7382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6242),
  [7384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4324),
  [7386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2329),
  [7388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6297),
  [7390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2101),
  [7392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4013),
  [7394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [7396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4612),
  [7398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1304),
  [7400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1371),
  [7402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5406),
  [7404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1372),
  [7406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2612),
  [7408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4016),
  [7410] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5950),
  [7412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2488),
  [7414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1291),
  [7416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3971),
  [7418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [7420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4910),
  [7422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3986),
  [7424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [7426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5952),
  [7428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3019),
  [7430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(957),
  [7432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4840),
  [7434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4635),
  [7436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6083),
  [7438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3928),
  [7440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4891),
  [7442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4029),
  [7444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2275),
  [7446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4158),
  [7448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4159),
  [7450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4032),
  [7452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6063),
  [7454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4334),
  [7456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1466),
  [7458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4918),
  [7460] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4668),
  [7462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6202),
  [7464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4218),
  [7466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5103),
  [7468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3931),
  [7470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6328),
  [7472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2160),
  [7474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [7476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6274),
  [7478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4557),
  [7480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4435),
  [7482] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5986),
  [7484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2083),
  [7486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4052),
  [7488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_type_repeat1, 2, 0, 0),
  [7490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5990),
  [7492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4887),
  [7494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_quoted_string_content_repeat1, 1, 0, 0),
  [7496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_quoted_string_content_repeat1, 1, 0, 0),
  [7498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4056),
  [7500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4102),
  [7502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4128),
  [7504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5689),
  [7506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5420),
  [7508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3308),
  [7510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3939),
  [7512] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6005),
  [7514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4899),
  [7516] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_labeled_tuple_element_type, 3, 0, 98), SHIFT(3294),
  [7519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6091),
  [7521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4564),
  [7523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3955),
  [7525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4057),
  [7527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2592),
  [7529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3975),
  [7531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [7533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1296),
  [7535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1328),
  [7537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1302),
  [7539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5344),
  [7541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1321),
  [7543] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(6294),
  [7546] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(4184),
  [7549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6320),
  [7551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2470),
  [7553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6140),
  [7555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4975),
  [7557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6144),
  [7559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4976),
  [7561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3897),
  [7563] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 4, 0, 133),
  [7565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5731),
  [7567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5439),
  [7569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4415),
  [7571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5931),
  [7573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5518),
  [7575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3247),
  [7577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2535),
  [7579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1948),
  [7581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2611),
  [7583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3900),
  [7585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(879),
  [7587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4640),
  [7589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5987),
  [7591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [7593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3923),
  [7595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3810),
  [7597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 4, 0, 161),
  [7599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2763),
  [7601] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 4, 0, 161),
  [7603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3798),
  [7605] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 4, 0, 0),
  [7607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2786),
  [7609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4630),
  [7611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1951),
  [7613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [7615] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4630),
  [7618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1963),
  [7620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1004),
  [7622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6967),
  [7624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4664),
  [7626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6163),
  [7628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1969),
  [7630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1856),
  [7632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1975),
  [7634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3754),
  [7636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1979),
  [7638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__sequence_binding_pattern_content_repeat1, 2, 0, 0),
  [7640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1983),
  [7642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1872),
  [7644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1133),
  [7646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1023),
  [7648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1030),
  [7650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3463),
  [7652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3352),
  [7654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1037),
  [7656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 3, 0, 89),
  [7658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1042),
  [7660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1048),
  [7662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(806),
  [7664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1590),
  [7666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1910),
  [7668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [7670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1623),
  [7672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3375),
  [7674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1805),
  [7676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3768),
  [7678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1570),
  [7680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1587),
  [7682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1684),
  [7684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2922),
  [7686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6386),
  [7688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4620),
  [7690] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6081),
  [7692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1603),
  [7694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2363),
  [7696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 3, 0, 0),
  [7698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1744),
  [7700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1618),
  [7702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [7704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1128),
  [7706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__sequence_pattern_content_repeat1, 2, 0, 0),
  [7708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(897),
  [7710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1137),
  [7712] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(5694),
  [7715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1785),
  [7717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2376),
  [7719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3475),
  [7721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(913),
  [7723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3738),
  [7725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3740),
  [7727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(796),
  [7729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(920),
  [7731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3404),
  [7733] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 3, 0, 89),
  [7735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(928),
  [7737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(935),
  [7739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3426),
  [7741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3429),
  [7743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [7745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1718),
  [7747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(938),
  [7749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3452),
  [7751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2671),
  [7753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2383),
  [7755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2390),
  [7757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3796),
  [7759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3777),
  [7761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3778),
  [7763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [7765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2397),
  [7767] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__type_param_repeat1, 2, 0, 0), SHIFT_REPEAT(5066),
  [7770] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_param_repeat1, 2, 0, 0),
  [7772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2691),
  [7774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3425),
  [7776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1831),
  [7778] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_typed, 2, 0, 35),
  [7780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1709),
  [7782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1860),
  [7784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(780),
  [7786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2732),
  [7788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2768),
  [7790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1770),
  [7792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3410),
  [7794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4197),
  [7796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5227),
  [7798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5756),
  [7800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1755),
  [7802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2744),
  [7804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5253),
  [7806] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5253),
  [7809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(818),
  [7811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1102),
  [7813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6323),
  [7815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4197),
  [7817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6349),
  [7819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1828),
  [7821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7327),
  [7823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1802),
  [7825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [7827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6450),
  [7829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(895),
  [7831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7136),
  [7833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4240),
  [7835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5940),
  [7837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2791),
  [7839] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tag_spec, 1, 0, 0),
  [7841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5452),
  [7843] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_param_repeat1, 1, 0, 0),
  [7845] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 1, 0, 2),
  [7847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6352),
  [7849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(777),
  [7851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4234),
  [7853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5755),
  [7855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1324),
  [7857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2833),
  [7859] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 1, 0, 0),
  [7861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 5, 0, 0),
  [7863] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_tag_specification_repeat1, 2, 0, 0),
  [7865] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tag_specification_repeat1, 2, 0, 0), SHIFT_REPEAT(2786),
  [7868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5268),
  [7870] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(5668),
  [7873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 2, 0, 15),
  [7875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5272),
  [7877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7323),
  [7879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7111),
  [7881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1116),
  [7883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6632),
  [7885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1594),
  [7887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6513),
  [7889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4577),
  [7891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6306),
  [7893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [7895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5291),
  [7897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4251),
  [7899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5549),
  [7901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7214),
  [7903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2706),
  [7905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6838),
  [7907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2356),
  [7909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6041),
  [7911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7112),
  [7913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(997),
  [7915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3815),
  [7917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5259),
  [7919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 2, 0, 0),
  [7921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5284),
  [7923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6558),
  [7925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1949),
  [7927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 1, 0, 0),
  [7929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3540),
  [7931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2927),
  [7933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5392),
  [7935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1204),
  [7937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5763),
  [7939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5448),
  [7941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1329),
  [7943] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 1, 0, 0),
  [7945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [7947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [7949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [7951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(497),
  [7953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(498),
  [7955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(439),
  [7957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5933),
  [7959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3744),
  [7961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2510),
  [7963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4479),
  [7965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3742),
  [7967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3804),
  [7969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6228),
  [7971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4456),
  [7973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1558),
  [7975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1572),
  [7977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2057),
  [7979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6273),
  [7981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3109),
  [7983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5545),
  [7985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__abstract_type, 2, 0, 0),
  [7987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2646),
  [7989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1432),
  [7991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1432),
  [7993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(914),
  [7995] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_params_repeat1, 2, 0, 0),
  [7997] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__type_params_repeat1, 2, 0, 0), SHIFT_REPEAT(4671),
  [8000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(921),
  [8002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2604),
  [8004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2627),
  [8006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5563),
  [8008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4122),
  [8010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6195),
  [8012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2717),
  [8014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2735),
  [8016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2844),
  [8018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3241),
  [8020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1964),
  [8022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1970),
  [8024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5722),
  [8026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3281),
  [8028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2781),
  [8030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3003),
  [8032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2929),
  [8034] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_instance_variable_specification_repeat1, 2, 0, 0),
  [8036] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variable_specification_repeat1, 2, 0, 0), SHIFT_REPEAT(5756),
  [8039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3646),
  [8041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2242),
  [8043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5505),
  [8045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3832),
  [8047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2858),
  [8049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5678),
  [8051] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_instantiated_class_type_repeat1, 2, 0, 0),
  [8053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5653),
  [8055] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_instantiated_class_type_repeat1, 2, 0, 0), SHIFT_REPEAT(2791),
  [8058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4951),
  [8060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2907),
  [8062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1169),
  [8064] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat1, 2, 0, 0), SHIFT_REPEAT(2604),
  [8067] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat1, 2, 0, 0),
  [8069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2621),
  [8071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5658),
  [8073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4309),
  [8075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1784),
  [8077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3324),
  [8079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5583),
  [8081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5447),
  [8083] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_specification_repeat1, 2, 0, 0),
  [8085] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_method_specification_repeat1, 2, 0, 0), SHIFT_REPEAT(5763),
  [8088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4892),
  [8090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3134),
  [8092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5680),
  [8094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1725),
  [8096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3240),
  [8098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3385),
  [8100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3142),
  [8102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5682),
  [8104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2838),
  [8106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(778),
  [8108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6199),
  [8110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3515),
  [8112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6735),
  [8114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3096),
  [8116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3107),
  [8118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1495),
  [8120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1087),
  [8122] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constructor_declaration_repeat1, 2, 0, 0),
  [8124] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_constructor_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6118),
  [8127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2107),
  [8129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2903),
  [8131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [8133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2456),
  [8135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3762),
  [8137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2638),
  [8139] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat2, 2, 0, 0),
  [8141] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat2, 2, 0, 0), SHIFT_REPEAT(5994),
  [8144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [8146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2149),
  [8148] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__abstract_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5545),
  [8151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__abstract_type_repeat1, 2, 0, 0),
  [8153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5576),
  [8155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4784),
  [8157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2526),
  [8159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3543),
  [8161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 1, 0, 0),
  [8163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2709),
  [8165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3835),
  [8167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5715),
  [8169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5601),
  [8171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1812),
  [8173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2649),
  [8175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5642),
  [8177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4453),
  [8179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5602),
  [8181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5812),
  [8183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4493),
  [8185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5916),
  [8187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1833),
  [8189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4983),
  [8191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(802),
  [8193] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1432),
  [8196] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1432),
  [8199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5200),
  [8201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3820),
  [8203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2433),
  [8205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1168),
  [8207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4865),
  [8209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5363),
  [8211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3794),
  [8213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4104),
  [8215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3801),
  [8217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5161),
  [8219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6067),
  [8221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2378),
  [8223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5852),
  [8225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4497),
  [8227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5606),
  [8229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2385),
  [8231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3809),
  [8233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4135),
  [8235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3811),
  [8237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1106),
  [8239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5263),
  [8241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3814),
  [8243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4145),
  [8245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3817),
  [8247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5256),
  [8249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3121),
  [8251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5049),
  [8253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5123),
  [8255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5153),
  [8257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5185),
  [8259] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(6067),
  [8262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5207),
  [8264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5224),
  [8266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5260),
  [8268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5163),
  [8270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3719),
  [8272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1025),
  [8274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3730),
  [8276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3829),
  [8278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5925),
  [8280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4475),
  [8282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1032),
  [8284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4143),
  [8286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3831),
  [8288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3833),
  [8290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5371),
  [8292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(830),
  [8294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4671),
  [8296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7247),
  [8298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__abstract_type_repeat1, 1, 0, 4),
  [8300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4950),
  [8302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6125),
  [8304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5985),
  [8306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6022),
  [8308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5352),
  [8310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7043),
  [8312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7163),
  [8314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4705),
  [8316] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(5052),
  [8319] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_pattern_repeat1, 2, 0, 0),
  [8321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [8323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6287),
  [8325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5000),
  [8327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5295),
  [8329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4954),
  [8331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(918),
  [8333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5365),
  [8335] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_type_repeat1, 2, 0, 0), SHIFT_REPEAT(3317),
  [8338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5276),
  [8340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2440),
  [8342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6226),
  [8344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6706),
  [8346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1530),
  [8348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5621),
  [8350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4753),
  [8352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2143),
  [8354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1307),
  [8356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4326),
  [8358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1308),
  [8360] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5315),
  [8363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [8365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4793),
  [8367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5315),
  [8369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5850),
  [8371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4957),
  [8373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5041),
  [8375] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4193),
  [8378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4794),
  [8380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4949),
  [8382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3094),
  [8384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4915),
  [8386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1562),
  [8388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5141),
  [8390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5503),
  [8392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5512),
  [8394] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1375),
  [8397] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_binding_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1376),
  [8400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3198),
  [8402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5023),
  [8404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1706),
  [8406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2610),
  [8408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1294),
  [8410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4518),
  [8412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4905),
  [8414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1610),
  [8416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(360),
  [8418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2864),
  [8420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4315),
  [8422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3114),
  [8424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(965),
  [8426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5029),
  [8428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_indexing_operator_path, 3, 0, 0),
  [8430] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_indexing_operator_path, 3, 0, 0),
  [8432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5440),
  [8434] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(5683),
  [8437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_declaration_repeat1, 2, 0, 0),
  [8439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3340),
  [8441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3070),
  [8443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3066),
  [8445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3197),
  [8447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6172),
  [8449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4992),
  [8451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2678),
  [8453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5442),
  [8455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5080),
  [8457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4893),
  [8459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(825),
  [8461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3101),
  [8463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4193),
  [8465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5307),
  [8467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2692),
  [8469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4757),
  [8471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2007),
  [8473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6089),
  [8475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4919),
  [8477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5012),
  [8479] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_copy_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5619),
  [8482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_copy_expression_repeat1, 2, 0, 0),
  [8484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5309),
  [8486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2429),
  [8488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2626),
  [8490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5230),
  [8492] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5850),
  [8495] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5281),
  [8498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5281),
  [8500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [8502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5021),
  [8504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2729),
  [8506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2636),
  [8508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1999),
  [8510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(858),
  [8512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5032),
  [8514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1061),
  [8516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1542),
  [8518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4703),
  [8520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6069),
  [8522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2195),
  [8524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5485),
  [8526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1118),
  [8528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(580),
  [8530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_expression, 1, 0, 0),
  [8532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4907),
  [8534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1874),
  [8536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4755),
  [8538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4574),
  [8540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5428),
  [8542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1896),
  [8544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3670),
  [8546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4836),
  [8548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1953),
  [8550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4792),
  [8552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2499),
  [8554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [8556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5317),
  [8558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1956),
  [8560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4842),
  [8562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4688),
  [8564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3035),
  [8566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6194),
  [8568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5373),
  [8570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5318),
  [8572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5462),
  [8574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6170),
  [8576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6085),
  [8578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4845),
  [8580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1968),
  [8582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4996),
  [8584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2703),
  [8586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4449),
  [8588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2648),
  [8590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4709),
  [8592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5438),
  [8594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2702),
  [8596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5105),
  [8598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5028),
  [8600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4715),
  [8602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4849),
  [8604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1286),
  [8606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6264),
  [8608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2157),
  [8610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3208),
  [8612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4988),
  [8614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(791),
  [8616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4724),
  [8618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4741),
  [8620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4437),
  [8622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4843),
  [8624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1830),
  [8626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(517),
  [8628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6295),
  [8630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1813),
  [8632] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_binding_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1323),
  [8635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4819),
  [8637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4932),
  [8639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6324),
  [8641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(768),
  [8643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4859),
  [8645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5953),
  [8647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3233),
  [8649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5486),
  [8651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(682),
  [8653] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 2, 0, 58),
  [8655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5970),
  [8657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6013),
  [8659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4520),
  [8661] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_expression_content_repeat1, 2, 0, 0), SHIFT_REPEAT(531),
  [8664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3110),
  [8666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5425),
  [8668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_instance_variable_specification_repeat1, 1, 0, 0),
  [8670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4936),
  [8672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1126),
  [8674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5991),
  [8676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6941),
  [8678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_specification_repeat1, 1, 0, 0),
  [8680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6008),
  [8682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2010),
  [8684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1990),
  [8686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6016),
  [8688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6023),
  [8690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4943),
  [8692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2365),
  [8694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6030),
  [8696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6036),
  [8698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5496),
  [8700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6043),
  [8702] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5088),
  [8705] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_expression_repeat1, 2, 0, 0),
  [8707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6050),
  [8709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6056),
  [8711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6059),
  [8713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6062),
  [8715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5463),
  [8717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2369),
  [8719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4699),
  [8721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1358),
  [8723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 2, 0, 87),
  [8725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6500),
  [8727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4979),
  [8729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6015),
  [8731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6035),
  [8733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6106),
  [8735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6652),
  [8737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(523),
  [8739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5469),
  [8741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [8743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(343),
  [8745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [8747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6761),
  [8749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3006),
  [8751] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_expression_content_repeat1, 2, 0, 0), SHIFT_REPEAT(595),
  [8754] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_binding_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(5249),
  [8757] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_binding_pattern_repeat1, 2, 0, 0),
  [8759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4457),
  [8761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4982),
  [8763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2382),
  [8765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6948),
  [8767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1881),
  [8769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3014),
  [8771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2131),
  [8773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5992),
  [8775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4989),
  [8777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4970),
  [8779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1788),
  [8781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5006),
  [8783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4777),
  [8785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4616),
  [8787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6317),
  [8789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6704),
  [8791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5274),
  [8793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1551),
  [8795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4500),
  [8797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5386),
  [8799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6093),
  [8801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6096),
  [8803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6097),
  [8805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [8807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(378),
  [8809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [8811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6206),
  [8813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6210),
  [8815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6212),
  [8817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4861),
  [8819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1006),
  [8821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [8823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(390),
  [8825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [8827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6291),
  [8829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5350),
  [8831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1795),
  [8833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6277),
  [8835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6278),
  [8837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6279),
  [8839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5382),
  [8841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [8843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(400),
  [8845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [8847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5520),
  [8849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1010),
  [8851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6301),
  [8853] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6302),
  [8855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6304),
  [8857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [8859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(407),
  [8861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [8863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6332),
  [8865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6160),
  [8867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6161),
  [8869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1357),
  [8871] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 2, 0, 87),
  [8873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [8875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(414),
  [8877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [8879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5385),
  [8881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5958),
  [8883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5959),
  [8885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5960),
  [8887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(420),
  [8889] = {.entry = {.count = 1, .reusable = false}}, SHIFT(421),
  [8891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(422),
  [8893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5977),
  [8895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5978),
  [8897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5979),
  [8899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [8901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(428),
  [8903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [8905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4903),
  [8907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [8909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5996),
  [8911] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5997),
  [8913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5998),
  [8915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [8917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(435),
  [8919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [8921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6009),
  [8923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6010),
  [8925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6011),
  [8927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_indexing_operator_path, 1, 0, 0),
  [8929] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_indexing_operator_path, 1, 0, 0),
  [8931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6017),
  [8933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6018),
  [8935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6019),
  [8937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6024),
  [8939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6025),
  [8941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6026),
  [8943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6031),
  [8945] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6032),
  [8947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6033),
  [8949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6037),
  [8951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6038),
  [8953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6039),
  [8955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6044),
  [8957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6045),
  [8959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6046),
  [8961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6051),
  [8963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6052),
  [8965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6053),
  [8967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4867),
  [8969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6213),
  [8971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [8973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(331),
  [8975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [8977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6908),
  [8979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6919),
  [8981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6924),
  [8983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4946),
  [8985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [8987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6613),
  [8989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5530),
  [8991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4760),
  [8993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4898),
  [8995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2481),
  [8997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [8999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4471),
  [9001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4873),
  [9003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [9005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4478),
  [9007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4774),
  [9009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4913),
  [9011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2409),
  [9013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2484),
  [9015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7025),
  [9017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4884),
  [9019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4769),
  [9021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6272),
  [9023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5531),
  [9025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2069),
  [9027] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1400),
  [9030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5290),
  [9032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(903),
  [9034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2614),
  [9036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5536),
  [9038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4912),
  [9040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5416),
  [9042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3514),
  [9044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1380),
  [9046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6904),
  [9048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(579),
  [9050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2509),
  [9052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6205),
  [9054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2795),
  [9056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6655),
  [9058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2496),
  [9060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2161),
  [9062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6909),
  [9064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6665),
  [9066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3027),
  [9068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5670),
  [9070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6147),
  [9072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6928),
  [9074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [9076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6262),
  [9078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7151),
  [9080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7152),
  [9082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7153),
  [9084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2096),
  [9086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6668),
  [9088] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__polymorphic_typed, 2, 0, 35),
  [9090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [9092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [9094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3025),
  [9096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6472),
  [9098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [9100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(960),
  [9102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2695),
  [9104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [9106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4540),
  [9108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(567),
  [9110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2812),
  [9112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3786),
  [9114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5967),
  [9116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7165),
  [9118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7166),
  [9120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7167),
  [9122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2318),
  [9124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [9126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4265),
  [9128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3771),
  [9130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6042),
  [9132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7250),
  [9134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7053),
  [9136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [9138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4643),
  [9140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3063),
  [9142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3770),
  [9144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3202),
  [9146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6915),
  [9148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2821),
  [9150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1691),
  [9152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [9154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2503),
  [9156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6245),
  [9158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7179),
  [9160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7180),
  [9162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7181),
  [9164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5016),
  [9166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2761),
  [9168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2162),
  [9170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7075),
  [9172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(625),
  [9174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6929),
  [9176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1989),
  [9178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6021),
  [9180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7186),
  [9182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7187),
  [9184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7188),
  [9186] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat2, 1, 0, 0),
  [9188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6331),
  [9190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6916),
  [9192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6094),
  [9194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7191),
  [9196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7192),
  [9198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7193),
  [9200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3036),
  [9202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6484),
  [9204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2505),
  [9206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6849),
  [9208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7260),
  [9210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6119),
  [9212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7196),
  [9214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7197),
  [9216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7198),
  [9218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(639),
  [9220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(822),
  [9222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [9224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6129),
  [9226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7201),
  [9228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7202),
  [9230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7203),
  [9232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6248),
  [9234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6917),
  [9236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6139),
  [9238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7206),
  [9240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7207),
  [9242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7208),
  [9244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3313),
  [9246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6884),
  [9248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6151),
  [9250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7211),
  [9252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7212),
  [9254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6334),
  [9256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [9258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2777),
  [9260] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_params, 1, 0, 0),
  [9262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6157),
  [9264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7216),
  [9266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7217),
  [9268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7218),
  [9270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4677),
  [9272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3103),
  [9274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2948),
  [9276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7083),
  [9278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6095),
  [9280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [9282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7159),
  [9284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3757),
  [9286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5869),
  [9288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [9290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5989),
  [9292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2813),
  [9294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2127),
  [9296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6470),
  [9298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6293),
  [9300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1658),
  [9302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7106),
  [9304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_type, 2, 0, 97),
  [9306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2139),
  [9308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6870),
  [9310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2507),
  [9312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6866),
  [9314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6499),
  [9316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4674),
  [9318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [9320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(958),
  [9322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4655),
  [9324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4115),
  [9326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6855),
  [9328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4443),
  [9330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7135),
  [9332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4580),
  [9334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2346),
  [9336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7183),
  [9338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5951),
  [9340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7060),
  [9342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2418),
  [9344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6922),
  [9346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3302),
  [9348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3236),
  [9350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7061),
  [9352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7062),
  [9354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5613),
  [9356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3305),
  [9358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3850),
  [9360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2421),
  [9362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6933),
  [9364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(549),
  [9366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1945),
  [9368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2540),
  [9370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6918),
  [9372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(995),
  [9374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1018),
  [9376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4232),
  [9378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1820),
  [9380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3333),
  [9382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6577),
  [9384] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, 0, 146),
  [9386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [9388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [9390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4530),
  [9392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2525),
  [9394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1001),
  [9396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6958),
  [9398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5387),
  [9400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [9402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1011),
  [9404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6970),
  [9406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2856),
  [9408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7237),
  [9410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1561),
  [9412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [9414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7086),
  [9416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4541),
  [9418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1554),
  [9420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7016),
  [9422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2891),
  [9424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1556),
  [9426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7026),
  [9428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(841),
  [9430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6978),
  [9432] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 2, 0, 109),
  [9434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(679),
  [9436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(725),
  [9438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(911),
  [9440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1100),
  [9442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(712),
  [9444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [9446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7058),
  [9448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6964),
  [9450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4833),
  [9452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(931),
  [9454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7063),
  [9456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4470),
  [9458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6966),
  [9460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4834),
  [9462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1583),
  [9464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2295),
  [9466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1596),
  [9468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2705),
  [9470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2874),
  [9472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7042),
  [9474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2674),
  [9476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7103),
  [9478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2679),
  [9480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7107),
  [9482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4868),
  [9484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1099),
  [9486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [9488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2021),
  [9490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7161),
  [9492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(843),
  [9494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7006),
  [9496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6198),
  [9498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7138),
  [9500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7139),
  [9502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2023),
  [9504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7168),
  [9506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7067),
  [9508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4625),
  [9510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2026),
  [9512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6040),
  [9514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4281),
  [9516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_params, 3, 0, 0),
  [9518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_params, 4, 0, 0),
  [9520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2594),
  [9522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3190),
  [9524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2839),
  [9526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6240),
  [9528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4646),
  [9530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2699),
  [9532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3146),
  [9534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3842),
  [9536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [9538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2167),
  [9540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6984),
  [9542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [9544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4144),
  [9546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6783),
  [9548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6309),
  [9550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6282),
  [9552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1468),
  [9554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4669),
  [9556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2730),
  [9558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1758),
  [9560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6310),
  [9562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7089),
  [9564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [9566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [9568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7090),
  [9570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7091),
  [9572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5736),
  [9574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2690),
  [9576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5983),
  [9578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6231),
  [9580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1159),
  [9582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7170),
  [9584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2495),
  [9586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7056),
  [9588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6173),
  [9590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [9592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1516),
  [9594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4133),
  [9596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1051),
  [9598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7121),
  [9600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5980),
  [9602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4513),
  [9604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1715),
  [9606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2708),
  [9608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4533),
  [9610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3736),
  [9612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2240),
  [9614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1878),
  [9616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7334),
  [9618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7022),
  [9620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4364),
  [9622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2776),
  [9624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1903),
  [9626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6719),
  [9628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2056),
  [9630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4737),
  [9632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6899),
  [9634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5999),
  [9636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4538),
  [9638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [9640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4973),
  [9642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6825),
  [9644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1809),
  [9646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6745),
  [9648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3782),
  [9650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6121),
  [9652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3275),
  [9654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4285),
  [9656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6444),
  [9658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4582),
  [9660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3725),
  [9662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6132),
  [9664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7108),
  [9666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7109),
  [9668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7110),
  [9670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5823),
  [9672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6463),
  [9674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2263),
  [9676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2984),
  [9678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3808),
  [9680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6187),
  [9682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2445),
  [9684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4125),
  [9686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [9688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2193),
  [9690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4704),
  [9692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4464),
  [9694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1625),
  [9696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6465),
  [9698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2105),
  [9700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2451),
  [9702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6308),
  [9704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7123),
  [9706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7124),
  [9708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7125),
  [9710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5928),
  [9712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5768),
  [9714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1866),
  [9716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7315),
  [9718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4764),
  [9720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4608),
  [9722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2094),
  [9724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6627),
  [9726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_guard, 2, 0, 31),
  [9728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4692),
  [9730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7052),
  [9732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6299),
  [9734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4848),
  [9736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4993),
  [9738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6861),
  [9740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3830),
  [9742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1135),
  [9744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6483),
  [9746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6629),
  [9748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2623),
  [9750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6635),
  [9752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2810),
  [9754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1837),
  [9756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7330),
  [9758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7038),
  [9760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6303),
  [9762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6266),
  [9764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1125),
  [9766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7023),
  [9768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7137),
  [9770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6810),
  [9772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1884),
  [9774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [9776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2082),
  [9778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2310),
  [9780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [9782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3930),
  [9784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1915),
  [9786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1918),
  [9788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3954),
  [9790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [9792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5186),
  [9794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3936),
  [9796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6191),
  [9798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [9800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4333),
  [9802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1601),
  [9804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4121),
  [9806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1797),
  [9808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3746),
  [9810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4744),
  [9812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2718),
  [9814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2780),
  [9816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3505),
  [9818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1901),
  [9820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [9822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3783),
  [9824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2091),
  [9826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1090),
  [9828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3239),
  [9830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2722),
  [9832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2092),
  [9834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6688),
  [9836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2141),
  [9838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2498),
  [9840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4223),
  [9842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2502),
  [9844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4813),
  [9846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1201),
  [9848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2142),
  [9850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2257),
  [9852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4217),
  [9854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [9856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3017),
  [9858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2602),
  [9860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(678),
  [9862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4896),
  [9864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3038),
  [9866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4559),
  [9868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4262),
  [9870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3433),
  [9872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [9874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2258),
  [9876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3759),
  [9878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4900),
  [9880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2284),
  [9882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4177),
  [9884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4284),
  [9886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2207),
  [9888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2153),
  [9890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3947),
  [9892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4263),
  [9894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [9896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4292),
  [9898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2736),
  [9900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4293),
  [9902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3277),
  [9904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2740),
  [9906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [9908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1195),
  [9910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1202),
  [9912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6237),
  [9914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2758),
  [9916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5713),
  [9918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [9920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3372),
  [9922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1559),
  [9924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4911),
  [9926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3271),
  [9928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1946),
  [9930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1067),
  [9932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1220),
  [9934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1179),
  [9936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [9938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2208),
  [9940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2765),
  [9942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5733),
  [9944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1214),
  [9946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4914),
  [9948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3382),
  [9950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1215),
  [9952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6070),
  [9954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4606),
  [9956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2277),
  [9958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3959),
  [9960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5172),
  [9962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [9964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [9966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1895),
  [9968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3964),
  [9970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5961),
  [9972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
  [9974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6080),
  [9976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [9978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [9980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2209),
  [9982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2762),
  [9984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2106),
  [9986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [9988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2727),
  [9990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5074),
  [9992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2700),
  [9994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3032),
  [9996] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 6, 0, 197),
  [9998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1560),
  [10000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3391),
  [10002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2787),
  [10004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4746),
  [10006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4480),
  [10008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5797),
  [10010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5638),
  [10012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1817),
  [10014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4747),
  [10016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1140),
  [10018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3037),
  [10020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5381),
  [10022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2843),
  [10024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2093),
  [10026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [10028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(696),
  [10030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1717),
  [10032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2992),
  [10034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2465),
  [10036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [10038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2414),
  [10040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2415),
  [10042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4202),
  [10044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1452),
  [10046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2454),
  [10048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3969),
  [10050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [10052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4274),
  [10054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4312),
  [10056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1611),
  [10058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1211),
  [10060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1190),
  [10062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6104),
  [10064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3394),
  [10066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4239),
  [10068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [10070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1200),
  [10072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1217),
  [10074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3061),
  [10076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3462),
  [10078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [10080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5822),
  [10082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1188),
  [10084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1952),
  [10086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [10088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2287),
  [10090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3974),
  [10092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [10094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4494),
  [10096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2265),
  [10098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3420),
  [10100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1823),
  [10102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3978),
  [10104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6158),
  [10106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [10108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1927),
  [10110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1145),
  [10112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4710),
  [10114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5254),
  [10116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2180),
  [10118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [10120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2293),
  [10122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3989),
  [10124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1954),
  [10126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4290),
  [10128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4291),
  [10130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [10132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2274),
  [10134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1176),
  [10136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1177),
  [10138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6286),
  [10140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5136),
  [10142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6476),
  [10144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(563),
  [10146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [10148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1178),
  [10150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1182),
  [10152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1846),
  [10154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1847),
  [10156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5927),
  [10158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4519),
  [10160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1932),
  [10162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1942),
  [10164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1189),
  [10166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3998),
  [10168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1851),
  [10170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1130),
  [10172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1199),
  [10174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3344),
  [10176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1958),
  [10178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4003),
  [10180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6327),
  [10182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4067),
  [10184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4818),
  [10186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4071),
  [10188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5739),
  [10190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [10192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1132),
  [10194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1962),
  [10196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [10198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4011),
  [10200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_string, 3, 0, 0),
  [10202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4299),
  [10204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4307),
  [10206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3056),
  [10208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(769),
  [10210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1213),
  [10212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1216),
  [10214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6107),
  [10216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4694),
  [10218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4103),
  [10220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [10222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1853),
  [10224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1218),
  [10226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1219),
  [10228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [10230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [10232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5617),
  [10234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [10236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [10238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [10240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4015),
  [10242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1545),
  [10244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1787),
  [10246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1150),
  [10248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1867),
  [10250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4019),
  [10252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6086),
  [10254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1869),
  [10256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1871),
  [10258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1537),
  [10260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1543),
  [10262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2095),
  [10264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4824),
  [10266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5910),
  [10268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4027),
  [10270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4256),
  [10272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4257),
  [10274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2262),
  [10276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1184),
  [10278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1185),
  [10280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6146),
  [10282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1965),
  [10284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(915),
  [10286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4074),
  [10288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [10290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1165),
  [10292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1966),
  [10294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2148),
  [10296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5573),
  [10298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7142),
  [10300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2194),
  [10302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1455),
  [10304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4031),
  [10306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2416),
  [10308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2500),
  [10310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2212),
  [10312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4035),
  [10314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6315),
  [10316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3351),
  [10318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(594),
  [10320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [10322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3046),
  [10324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3422),
  [10326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2192),
  [10328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2097),
  [10330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4043),
  [10332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4203),
  [10334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4258),
  [10336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4259),
  [10338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6760),
  [10340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1205),
  [10342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1209),
  [10344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6260),
  [10346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(916),
  [10348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1580),
  [10350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1971),
  [10352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1972),
  [10354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1210),
  [10356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1212),
  [10358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4872),
  [10360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2302),
  [10362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1973),
  [10364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5656),
  [10366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2816),
  [10368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [10370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [10372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4055),
  [10374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1974),
  [10376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6088),
  [10378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3892),
  [10380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6020),
  [10382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__infix_operator, 1, 0, 0),
  [10384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [10386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2825),
  [10388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1622),
  [10390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6894),
  [10392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2512),
  [10394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(873),
  [10396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1929),
  [10398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1976),
  [10400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1977),
  [10402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1170),
  [10404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1171),
  [10406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6186),
  [10408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4008),
  [10410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2442),
  [10412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1978),
  [10414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1172),
  [10416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1174),
  [10418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1905),
  [10420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2176),
  [10422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5711),
  [10424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2837),
  [10426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4897),
  [10428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3899),
  [10430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(666),
  [10432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2144),
  [10434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [10436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3903),
  [10438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6265),
  [10440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6120),
  [10442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [10444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4436),
  [10446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2900),
  [10448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3479),
  [10450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2970),
  [10452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4155),
  [10454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1759),
  [10456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6007),
  [10458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1810),
  [10460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2008),
  [10462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5769),
  [10464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [10466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [10468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2793),
  [10470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3909),
  [10472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [10474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3913),
  [10476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6079),
  [10478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4224),
  [10480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(875),
  [10482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2632),
  [10484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1457),
  [10486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [10488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3760),
  [10490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1536),
  [10492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2601),
  [10494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(927),
  [10496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6102),
  [10498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2491),
  [10500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2831),
  [10502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2364),
  [10504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4438),
  [10506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [10508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [10510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1538),
  [10512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1540),
  [10514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6123),
  [10516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2256),
  [10518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4040),
  [10520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4164),
  [10522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6319),
  [10524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2463),
  [10526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6131),
  [10528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [10530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [10532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2455),
  [10534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [10536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [10538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [10540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4211),
  [10542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6141),
  [10544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [10546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2917),
  [10548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(876),
  [10550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2480),
  [10552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3455),
  [10554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2482),
  [10556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2063),
  [10558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6153),
  [10560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3057),
  [10562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1726),
  [10564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1733),
  [10566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1742),
  [10568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3065),
  [10570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4212),
  [10572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [10574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6162),
  [10576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(878),
  [10578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [10580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [10582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [10584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4984),
  [10586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4440),
  [10588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3062),
  [10590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1103),
  [10592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [10594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2859),
  [10596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2367),
  [10598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4821),
  [10600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2184),
  [10602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3966),
  [10604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1614),
  [10606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2011),
  [10608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [10610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1743),
  [10612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4491),
  [10614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2487),
  [10616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1849),
  [10618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_character_content, 1, 0, 0),
  [10620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [10622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [10624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1947),
  [10626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [10628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2506),
  [10630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2213),
  [10632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5410),
  [10634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5194),
  [10636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2371),
  [10638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4086),
  [10640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4466),
  [10642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(835),
  [10644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [10646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(836),
  [10648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4119),
  [10650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5003),
  [10652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3104),
  [10654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_string, 2, 0, 0),
  [10656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [10658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1163),
  [10660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2508),
  [10662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1870),
  [10664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [10666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6592),
  [10668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5250),
  [10670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2489),
  [10672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3356),
  [10674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [10676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1063),
  [10678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2375),
  [10680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1064),
  [10682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1055),
  [10684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1880),
  [10686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5043),
  [10688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2158),
  [10690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4276),
  [10692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [10694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3297),
  [10696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6112),
  [10698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3935),
  [10700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2359),
  [10702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(115),
  [10704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(779),
  [10706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2519),
  [10708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2361),
  [10710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4198),
  [10712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1799),
  [10714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3478),
  [10716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2362),
  [10718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6006),
  [10720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6702),
  [10722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2613),
  [10724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1854),
  [10726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1863),
  [10728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5085),
  [10730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1800),
  [10732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2379),
  [10734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1801),
  [10736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6757),
  [10738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2165),
  [10740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1864),
  [10742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6765),
  [10744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [10746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4340),
  [10748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6784),
  [10750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6785),
  [10752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6786),
  [10754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6797),
  [10756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2650),
  [10758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2159),
  [10760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2420),
  [10762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [10764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6816),
  [10766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [10768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5022),
  [10770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(814),
  [10772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6174),
  [10774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4208),
  [10776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1094),
  [10778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2380),
  [10780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2422),
  [10782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4481),
  [10784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2460),
  [10786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3108),
  [10788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1621),
  [10790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [10792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6947),
  [10794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6155),
  [10796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1033),
  [10798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2386),
  [10800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2387),
  [10802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2388),
  [10804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4222),
  [10806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6380),
  [10808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1041),
  [10810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(980),
  [10812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3010),
  [10814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2467),
  [10816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2469),
  [10818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2170),
  [10820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2474),
  [10822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [10824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1002),
  [10826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2389),
  [10828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [10830] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [10832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4857),
  [10834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(240),
  [10836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4858),
  [10838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4195),
  [10840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7059),
  [10842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6372),
  [10844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1020),
  [10846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1692),
  [10848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2394),
  [10850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2395),
  [10852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2396),
  [10854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [10856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1175),
  [10858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3058),
  [10860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2408),
  [10862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2169),
  [10864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [10866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1206),
  [10868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2006),
  [10870] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compilation_unit, 1, 0, 0),
  [10872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3119),
  [10874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3772),
  [10876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1573),
  [10878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1575),
  [10880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1576),
  [10882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(724),
  [10884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1096),
  [10886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4136),
  [10888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2202),
  [10890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(447),
  [10892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2013),
  [10894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3984),
  [10896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(984),
  [10898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [10900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(985),
  [10902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(449),
  [10904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [10906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [10908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(452),
  [10910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3265),
  [10912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4154),
  [10914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2203),
  [10916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7000),
  [10918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1555),
  [10920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(877),
  [10922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(771),
  [10924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [10926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(773),
  [10928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4386),
  [10930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1068),
  [10932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3729),
  [10934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(971),
  [10936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1557),
  [10938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1786),
  [10940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [10942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(578),
  [10944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1108),
  [10946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(998),
  [10948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6641),
  [10950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(934),
  [10952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(926),
  [10954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3127),
  [10956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3428),
  [10958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4146),
  [10960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(883),
  [10962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5215),
  [10964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [10966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2988),
  [10968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [10970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3304),
  [10972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2990),
  [10974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1058),
  [10976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3008),
  [10978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6099),
  [10980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2600),
  [10982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3507),
  [10984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5114),
  [10986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5475),
  [10988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(930),
  [10990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2606),
  [10992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6353),
  [10994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6354),
  [10996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6355),
  [10998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [11000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [11002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(823),
  [11004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2472),
  [11006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4213),
  [11008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [11010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1059),
  [11012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1584),
  [11014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [11016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1904),
  [11018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3453),
  [11020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5162),
  [11022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3838),
  [11024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4468),
  [11026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [11028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [11030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1060),
  [11032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1005),
  [11034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2693),
  [11036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6215),
  [11038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6473),
  [11040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1858),
  [11042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5055),
  [11044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3839),
  [11046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2436),
  [11048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6454),
  [11050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6455),
  [11052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6456),
  [11054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4319),
  [11056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2459),
  [11058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2479),
  [11060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4207),
  [11062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2099),
  [11064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [11066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2437),
  [11068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2438),
  [11070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2676),
  [11072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2675),
  [11074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2044),
  [11076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4937),
  [11078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5133),
  [11080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2749),
  [11082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6537),
  [11084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6538),
  [11086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6539),
  [11088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1111),
  [11090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3942),
  [11092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3258),
  [11094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [11096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4615),
  [11098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5165),
  [11100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [11102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6580),
  [11104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6581),
  [11106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6582),
  [11108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2957),
  [11110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1933),
  [11112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1804),
  [11114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(884),
  [11116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [11118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2639),
  [11120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6270),
  [11122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6620),
  [11124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6621),
  [11126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6622),
  [11128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [11130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2741),
  [11132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 5, 0, 169),
  [11134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1014),
  [11136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [11138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2019),
  [11140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4487),
  [11142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3295),
  [11144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2754),
  [11146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3259),
  [11148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1586),
  [11150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6659),
  [11152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6660),
  [11154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6661),
  [11156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [11158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7336),
  [11160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2477),
  [11162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1627),
  [11164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [11166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2022),
  [11168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [11170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(970),
  [11172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2252),
  [11174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6697),
  [11176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6698),
  [11178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6699),
  [11180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2024),
  [11182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1084),
  [11184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5078),
  [11186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1071),
  [11188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2260),
  [11190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1086),
  [11192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [11194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1022),
  [11196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2320),
  [11198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6732),
  [11200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6733),
  [11202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6734),
  [11204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2714),
  [11206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3199),
  [11208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2686),
  [11210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6754),
  [11212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6755),
  [11214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6756),
  [11216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1793),
  [11218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2027),
  [11220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6767),
  [11222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6768),
  [11224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6769),
  [11226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2028),
  [11228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(887),
  [11230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6778),
  [11232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6779),
  [11234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6780),
  [11236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(888),
  [11238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6788),
  [11240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6789),
  [11242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6790),
  [11244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(890),
  [11246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6798),
  [11248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6799),
  [11250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6800),
  [11252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(638),
  [11254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4827),
  [11256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6808),
  [11258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6809),
  [11260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6818),
  [11262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6819),
  [11264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6820),
  [11266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(652),
  [11268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2694),
  [11270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2201),
  [11272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2698),
  [11274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [11276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2086),
  [11278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2707),
  [11280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6860),
  [11282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [11284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5801),
  [11286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2037),
  [11288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5478),
  [11290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1710),
  [11292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2039),
  [11294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [11296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [11298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6885),
  [11300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5660),
  [11302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3991),
  [11304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3925),
  [11306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5100),
  [11308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6920),
  [11310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2123),
  [11312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2041),
  [11314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(120),
  [11316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(961),
  [11318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4301),
  [11320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6887),
  [11322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3956),
  [11324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4303),
  [11326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1026),
  [11328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1980),
  [11330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3283),
  [11332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4806),
  [11334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [11336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2958),
  [11338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6890),
  [11340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2959),
  [11342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7040),
  [11344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5843),
  [11346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [11348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(537),
  [11350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5853),
  [11352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3293),
  [11354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7013),
  [11356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5860),
  [11358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5865),
  [11360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5870),
  [11362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5875),
  [11364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5878),
  [11366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5882),
  [11368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5885),
  [11370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5888),
  [11372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5890),
  [11374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5892),
  [11376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5894),
  [11378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5896),
  [11380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5898),
  [11382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4617),
  [11384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [11386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1183),
  [11388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [11390] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_payload, 3, 0, 0),
  [11392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1194),
  [11394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6156),
  [11396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2756),
  [11398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [11400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4476),
  [11402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5157),
  [11404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5568),
  [11406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(973),
  [11408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(974),
  [11410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(975),
  [11412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2145),
  [11414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4576),
  [11416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3379),
  [11418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2483),
  [11420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3788),
  [11422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1036),
  [11424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3827),
  [11426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [11428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2306),
  [11430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [11432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1053),
  [11434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6750),
  [11436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(976),
  [11438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(977),
  [11440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4558),
  [11442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(956),
  [11444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1886),
  [11446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2516),
  [11448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1862),
  [11450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5547),
  [11452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [11454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [11456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3741),
  [11458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1203),
  [11460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1208),
  [11462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(454),
  [11464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4514),
  [11466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7331),
  [11468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5609),
  [11470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1844),
  [11472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4147),
  [11474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2689),
  [11476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [11478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5949),
  [11480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 3, 0, 0),
  [11482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 4, 0, 0),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token_comment = 0,
  ts_external_token__left_quoted_string_delimiter = 1,
  ts_external_token__right_quoted_string_delimiter = 2,
  ts_external_token_DQUOTE = 3,
  ts_external_token_line_number_directive = 4,
  ts_external_token__null = 5,
  ts_external_token__error_sentinel = 6,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token_comment] = sym_comment,
  [ts_external_token__left_quoted_string_delimiter] = sym__left_quoted_string_delimiter,
  [ts_external_token__right_quoted_string_delimiter] = sym__right_quoted_string_delimiter,
  [ts_external_token_DQUOTE] = anon_sym_DQUOTE,
  [ts_external_token_line_number_directive] = sym_line_number_directive,
  [ts_external_token__null] = sym__null,
  [ts_external_token__error_sentinel] = sym__error_sentinel,
};

static const bool ts_external_scanner_states[10][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token_comment] = true,
    [ts_external_token__left_quoted_string_delimiter] = true,
    [ts_external_token__right_quoted_string_delimiter] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
    [ts_external_token__error_sentinel] = true,
  },
  [2] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [3] = {
    [ts_external_token_comment] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [4] = {
    [ts_external_token_comment] = true,
    [ts_external_token__left_quoted_string_delimiter] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [5] = {
    [ts_external_token_comment] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
  },
  [6] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
  },
  [7] = {
    [ts_external_token_comment] = true,
    [ts_external_token__right_quoted_string_delimiter] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
  },
  [8] = {
    [ts_external_token_comment] = true,
    [ts_external_token__left_quoted_string_delimiter] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [9] = {
    [ts_external_token_comment] = true,
    [ts_external_token__right_quoted_string_delimiter] = true,
    [ts_external_token_line_number_directive] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_ocaml_interface_external_scanner_create(void);
void tree_sitter_ocaml_interface_external_scanner_destroy(void *);
bool tree_sitter_ocaml_interface_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_ocaml_interface_external_scanner_serialize(void *, char *);
void tree_sitter_ocaml_interface_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_ocaml_interface(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .supertype_map_slices = ts_supertype_map_slices,
    .supertype_map_entries = ts_supertype_map_entries,
    .supertype_symbols = ts_supertype_symbols,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym__lowercase_identifier,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_ocaml_interface_external_scanner_create,
      tree_sitter_ocaml_interface_external_scanner_destroy,
      tree_sitter_ocaml_interface_external_scanner_scan,
      tree_sitter_ocaml_interface_external_scanner_serialize,
      tree_sitter_ocaml_interface_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "ocaml_interface",
    .reserved_words = &ts_reserved_words[0][0],
    .max_reserved_word_set_size = 55,
    .metadata = {
      .major_version = 0,
      .minor_version = 25,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
